/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2014 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

/**
 * @toc           File_System:
 *
 * @file          efs.h
 *
 * @brief         The Embedded File System (EFS) function library API
 *                and definitions. These functions were formerly part
 *                of the EFS FLIB but are now a part of the C
 *                Run-time library (libcrt).
 *
 */

#ifndef _EFS_H
#define _EFS_H

#include "ose.h"
#include "osetypes.h"
#include "stddef.h"      /* Define size_t. */
#include "ose_spi/efstypes.h" /* Define EfsStatus. */
#include "fm.sig"        /* Define FmHandle. */
#include "stdio.h"       /* Define FILE *. */

#ifdef __cplusplus
extern "C" {
#endif

/**
 * @type EfsFdTag
 *
 * @include "efs.h"
 *
 * @declaration
 * enum EfsFdTag
 * {
 *   FLIB_FD_STALE            = 0x1,
 *   FLIB_FD_SHMEM            = 0x2,
 *   FLIB_FD_APPEND           = 0x3,
 *   FLIB_FD_VMPID            = 0x4,
 *   FLIB_FD_HANDLE           = 0x5,
 *   FLIB_FD_BUFMODE          = 0x6,
 *   FLIB_FD_BUFSIZE          = 0x7,
 *   FLIB_FD_SOCKET           = 0x9,
 *   FLIB_FD_CLONE            = 0xA
 * };
 *
 * @brief The different properties of file descriptors.
 *
 * @long The values of EfsFdTag are used when configuring and examining
 *       file descriptors using @link "" "efs_config_fd" and
 *       @link "" "efs_examine_fd"
 *
 * @field FLIB_FD_STALE           The file descriptor is stale. Stale file
 *                                descriptors are typically caused by broken
 *                                communication link between the client and the
 *                                file system/IP stack.
 * @field FLIB_FD_SHMEM           Shared memory file descriptor. When a file
 *                                descriptor is configured for shared memory,
 *                                data pointers are sent to and from the volume
 *                                managers and the client instead of memory
 *                                buffers. Shared memory increases read/write
 *                                performance but requires a shared memory
 *                                system.
 * @field FLIB_FD_APPEND          Writes are appended at the end of the file.
 *                                If a file descriptor is configured for
 *                                append, data is written at the end of the
 *                                file instead of at the current position. The
 *                                read position of the file after an appending
 *                                write may be changed. The behavior is volume
 *                                manager dependent with regard to read
 *                                position. The default for a file descriptor
 *                                is not to append.
 * @field FLIB_FD_VMPID           The PID of the volume manager managing the
 *                                file referred by the file descriptor. If the
 *                                file descriptor is a socket descriptor, the
 *                                volume manager PID is 0.
 * @field FLIB_FD_HANDLE          The volume manager handle of the file
 *                                descriptor stream. The handle, FmHandle, is
 *                                used to identify an opened file when
 *                                communicating directly to its volume manager.
 * @field FLIB_FD_BUFMODE         The file descriptor buffer mode, defined by
 *                                enum FmBuffEnum in ose_spi/fm.sig.
 *                                The buffer mode describes
 *                                how the file buffering is configured.
 *                                The default buffer mode for
 *                                socket descriptors is FM_BUFF_NONE. The
 *                                default buffer mode for file descriptors is
 *                                volume manager dependent.
 * @field FLIB_FD_BUFSIZE         The file descriptor buffer size (in bytes).
 *                                The buffer size determine how much data can
 *                                be buffered in the client context. The buffer
 *                                size configuration affect the read/write
 *                                performance. The initial buffer size is
 *                                determined by the volume manager of the file.
 * @field FLIB_FD_SOCKET          The file descriptor is a socket descriptor.
 * @field FLIB_FD_CLONE           The file descriptor is possible to clone,
 *                                enabled cloning is default behavior.
 *
 * @seealso efs_examine_fd
 * @seealso efs_config_fd
 */

enum EfsFdTag
{
    FLIB_FD_STALE             = 0x1,
    FLIB_FD_SHMEM             = 0x2,
    FLIB_FD_APPEND            = 0x3,
    FLIB_FD_VMPID             = 0x4,
    FLIB_FD_HANDLE            = 0x5,
    FLIB_FD_BUFMODE           = 0x6,
    FLIB_FD_BUFSIZE           = 0x7,
    FLIB_FD_SOCKET            = 0x9,
    FLIB_FD_CLONE             = 0xA
};

/**
 * @type EfsFileTag
 *
 * @include "efs.h"
 *
 * @declaration
 * enum EfsFileTag
 * {
 *  FLIB_FILE_BUFFER      = 0x1,
 *  FLIB_FILE_BUFSIZE     = 0x2,
 *  FLIB_FILE_CURRENT     = 0x3,
 *  FLIB_FILE_CURRENT_REF = 0x4,
 *  FLIB_FILE_AVAILRD     = 0x5,
 *  FLIB_FILE_AVAILWR     = 0x6,
 *  FLIB_FILE_AVAIL_REF   = 0x7
 * };
 *
 * @brief The different properties of buffered files.
 *
 * @long The values of EfsFileTag are used when examining
 *       buffered files using @link "" "efs_examine_file".
 *
 * @field FLIB_FILE_BUFFER        The file buffer start pointer.
 * @field FLIB_FILE_BUFSIZE       The file buffer size.
 * @field FLIB_FILE_CURRENT       A pointer to the next character to read or
 *                                write.
 * @field FLIB_FILE_CURRENT_REF   A pointer to the current pointer.
 * @field FLIB_FILE_AVAILRD       The number of characters available for read
 *                                (if file is in read mode).
 * @field FLIB_FILE_AVAILWR       The number of characters available for write
 *                                (if file is in write mode).
 * @field FLIB_FILE_AVAIL_REF     A pointer to the available number of
 *                                read/write characters variable.
 *
 * @seealso efs_examine_file
 */

enum EfsFileTag
{
    FLIB_FILE_BUFFER      = 0x1,
    FLIB_FILE_BUFSIZE     = 0x2,
    FLIB_FILE_CURRENT     = 0x3,
    FLIB_FILE_CURRENT_REF = 0x4,
    FLIB_FILE_AVAILRD     = 0x5,
    FLIB_FILE_AVAILWR     = 0x6,
    FLIB_FILE_AVAIL_REF   = 0x7
};

/*
 * STATUS/ERROR CODES
 */

/**
 * @macro         FLIB_EINVALID_FD
 *
 * @include       "efs.h"
 *
 * @declaration   #define FLIB_EINVALID_FD 500
 *
 * @brief         An invalid file descriptor was used in a CRT function call.
 * @long          An invalid file descriptor is a a file descriptor that has
 *                never been opened, or has been opened and then closed.
 *
 */
#define FLIB_EINVALID_FD                500

/**
 * @macro         FLIB_EINVALID_PARAMETER
 *
 * @include       "efs.h"
 *
 * @declaration   #define FLIB_EINVALID_PARAMETER 501
 *
 * @brief         An invalid parameter was used in a
 *                function call.
 *
 */
#define FLIB_EINVALID_PARAMETER         501

/**
 * @macro         FLIB_ESTALE_FD
 *
 * @include       "efs.h"
 *
 * @declaration   #define FLIB_ESTALE_FD 502
 *
 * @brief         A stale file descriptor was used in a function call.
 * @long          When a connection to a server handling the file referred to
 *                by the file descriptor is lost, the file descriptor becomes
 *                stale.
 *
 */
#define FLIB_ESTALE_FD                  502

/**
 * @macro         FLIB_ELOST_CONTACT
 *
 * @include       "efs.h"
 *
 * @declaration   #define FLIB_ELOST_CONTACT 503
 *
 * @brief         Contact with the volume manager was lost during a function
 *                call.
 *
 */
#define FLIB_ELOST_CONTACT              503

/**
 * @macro         FLIB_EUNKNOWN_VOLUME
 *
 * @include       "efs.h"
 *
 * @declaration   #define FLIB_EUNKNOWN_VOLUME 504
 *
 * @brief         An unknown volume was being refereed to in a function call.
 *
 */
#define FLIB_EUNKNOWN_VOLUME            504

/**
 * @macro         FLIB_ESOCKET_ERROR
 *
 * @include       "efs.h"
 *
 * @declaration   #define FLIB_ESOCKET_ERROR 505
 *
 * @brief         A socket error occurred during a
 *                function call.
 *
 */
#define FLIB_ESOCKET_ERROR              505

/**
 * @macro         FLIB_EINVALID_FILE
 *
 * @include       "efs.h"
 *
 * @declaration   #define FLIB_EINVALID_FILE 506
 *
 * @brief         An invalid file pointer was used in a
 *                function call.
 *
 */
#define FLIB_EINVALID_FILE              506

/**
 * @type EfsMatchBitmap
 *
 * @include "efs.h"
 *
 * @declaration
 * enum EfsMatchBitmap
 * {
 *  MATCH_PROCESSES        = 0x01,
 *  MATCH_BLOCKS           = 0x02,
 *  MATCH_SEGMENTS         = 0x04,
 *  MATCH_DISCARD_INSTANCE = 0x08
 * };
 *
 * @brief Bitmap values used when specifying what to match
 * in @link "" "ose_match_processes".
 *
 * @field MATCH_PROCESSES         Match processes.
 * @field MATCH_BLOCKS            Match blocks.
 * @field MATCH_SEGMENTS          Match domains.
 * @field MATCH_DISCARD_INSTANCE  Discard instance name part of process name,
 *                                if it exists. Instance names exists e.g. for
 *                                processes in a program where
 *                                OSE_LM_PROGRAM_NAME is set.
 *
 * @seealso ose_match_processes
 */

enum EfsMatchBitmap
{
    MATCH_PROCESSES        = 0x01,
    MATCH_BLOCKS           = 0x02,
    MATCH_SEGMENTS         = 0x04,
    MATCH_DISCARD_INSTANCE = 0x08
};


/**
 * @function        efs_format
 *
 * @include         "efs.h"
 *
 * @declaration     EfsStatus
 *                  efs_format(const char *volume, const char *params,
 *                             Boolean force);
 *
 * @brief           Format a volume.
 * @long            Formatting a volume prepares it for use by a file manager.
 *                  The volume must first be mounted with @link "" "efs_mount"
 *                  with the file manager format required.
 *
 * @long            With force, formatting can be forced even if there are open
 *                  files on the volume.
 *
 * @long            Which optional parameters to give in the params argument
 *                  depends entirely on the file and volume managers. See the
 *                  example below.
 *
 * @input           volume     Points to the volume name string, as used in
 *                             @link "" "efs_mount" to identify the volume.
 * @input           params     A pointer to a string with multiple optional
 *                             parameter in the form "name1=value1,name2=value2
 *                             .... ". Also boolean parameters can be given,
 *                             i.e. a name without any "=value".
 *                             See the volume manager specific manuals for
 *                             available parameters.
 * @input           force      If true and if there are opened files, it forces
 *                             formatting and removes any opened file streams.
 *                             If False, no action is performed if there are
 *                             any open files are on the volume.
 *
 * @return
 * @tab             EFS_SUCCESS  @tab The volume was successfully formatted.
 * @break
 * @tab             FM_E*        @tab Upon failure the format/volume manager
 *                                    may indicate errors using the
 *                                    status codes defined in fm.sig.
 *                                    Error statuses returned
 *                                    are format/volume manager specific. @break
 * @tab             DDB_E*       @tab Any DDB specific error status.
 *                                    The error statuses returned by
 *                                    the DDBs are specific to each DDB. @break
 * @tab             E*           @tab Upon failure the format/volume manager
 *                                    indicates errors using the OSE
 *                                    errno statuses, see
 *                                    @link "errno.h" "errno"
 *                                    The error statuses returned
 *                                    are format/volume manager specific.@break
 * @tab                          @tab FM_E*, DDB_E* codes can be translated to
 *                                    errno codes using
 *                                    @link "" "efs_status_to_errno". @break
 * @tab                          @tab Global variable  @link "errno.h" "errno"
 *                                    is not updated by this function. @break
 *
 * @seealso         efs_mount
 * @seealso         efs_unmount
 * @seealso         efs_wformat
 * @seealso         default_restrictions_calling_functions
 *
 * @example
 * void
 * format_my_volume(void)
 * {
 *   // formatting a (mounted) disk at /ram with a 16 bit FAT, 4 sectors per
 *   // cluster, 1 copy of the FAT and 64 root sectors (with 16 entries per
 *   // sector). Use force if files are open.
 *
 *   if (efs_format("/ram", "fatsize=16,clustersize=4,fats=1,rootsectors=64",
 *                  True) == EFS_SUCCESS)
 *   {
 *      // Succeeded.
 *   }
 * }
 */
EfsStatus efs_format(const char *volume, const char *params, Boolean force);


/**
 * @function        efs_wformat
 *
 * @include         "efs.h"
 *
 * @declaration     EfsStatus
 *                  efs_wformat(const wchar_t *volume, const char *params,
 *                              Boolean force);
 *
 * @brief           Format a volume.
 * @long            This call is equivalent to @link "" "efs_format" with the
 *                  exception that the volume name is passed as a wchar_t
 *                  string.
 * @long            Formatting a volume prepares it for use by a file manager.
 *                  The volume must first be mounted with e.g. @link ""
 *                  "efs_wmount" with the file manager format required.
 *
 * @long            With force, formatting can be forced even if there are open
 *                  files on the volume.
 *
 * @long            Which optional parameters to give in the params argument
 *                  depends entirely on the file and volume managers. See the
 *                  example below.
 *
 * @input           volume     Points to the volume name string, as used in
 *                             @link "" "efs_wmount" to identify the volume.
 * @input           params     A pointer to a string with multiple optional
 *                             parameter in the form "name1=value1,name2=value2
 *                             .... ". Also boolean parameters can be given,
 *                             i.e. a name without any "=value".
 *                             See the volume manager specific manuals for
 *                             available parameters.
 * @input           force      If true and if there are opened files, it forces
 *                             formatting and removes any opened file streams.
 *                             If False, no action is performed if there are
 *                             any open files are on the volume.
 *
 * @return
 * @tab             EFS_SUCCESS  @tab The volume was successfully formatted.
 * @break
 * @tab             FM_E*        @tab Upon failure the format/volume manager
 *                                    may indicate errors using the
 *                                    status codes defined in fm.sig.
 *                                    Error statuses returned
 *                                    are format/volume manager specific. @break
 * @tab             DDB_E*       @tab Any DDB specific error status.
 *                                    The error statuses returned by
 *                                    the DDBs are specific to each DDB. @break
 * @tab             E*           @tab Upon failure the format/volume manager
 *                                    indicates errors using the OSE
 *                                    errno statuses, see
 *                                    @link "errno.h" "errno"
 *                                    The error statuses returned
 *                                    are format/volume manager specific.@break
 * @tab                          @tab FM_E*, DDB_E* codes can be translated to
 *                                    errno codes using
 *                                    @link "" "efs_status_to_errno". @break
 * @tab                          @tab Global variable  @link "errno.h" "errno"
 *                                    is not updated by this function. @break
 *
 * @seealso         efs_mount
 * @seealso         efs_unmount
 * @seealso         default_restrictions_calling_functions
 *
 * @example
 * void
 * format_my_volume(void)
 * {
 *   // formatting a (mounted) disk at /ram with a 16 bit FAT, 4 sectors per
 *   // cluster, 1 copy of the FAT and 64 root sectors (with 16 entries per
 *   // sector). Use force if files are open.
 *
 *   if (efs_wformat(L"/ram", "fatsize=16,clustersize=4,fats=1,rootsectors=64",
 *                  True) == EFS_SUCCESS)
 *   {
 *      // Succeeded.
 *   }
 * }
 */
EfsStatus efs_wformat(const wchar_t *volume, const char *params, Boolean force);


/**
 * @function        efs_mount
 *
 * @include         "efs.h"
 *
 * @declaration     EfsStatus
 *                  efs_mount(const char *volume, const char *format,
 *                            const char *dev, const char *params);
 *
 * @brief           Mount a volume on the indicated device dev.
 *
 * @long            The name of the volume is set to "volume" and the "format"
 *                  indicates the file system format expected on the volume.
 *                  Optional parameters are given in "params". If the
 *                  volume is not formatted according to "format", it will
 *                  anyhow be mounted, in unformatted state, and can be
 *                  formatted with the function @link "" "efs_format".
 *
 * @long            Upon successful completion, a volume manager (VM) will be
 *                  available through which files on this volume are accessed.
 *
 *                  The volume name is used as the leading part of the absolute
 *                  path for the file names on that volume and the volume name
 *                  must have a leading /.
 *
 * @input           volume     Points to the volume path string, used to
 *                             identify the volume. Must begin with "/".
 * @input           format     A pointer to the name of an existing file
 *                             manager format to use. Examples are "hafsfat",
 *                             "extfat". A file manager (FM) handle the
 *                             specific format must be known to the File System
 *                             Server (ose_fss process).
 * @input           dev        A pointer to the name of the device to mount.
 *                             The use of this string is voume manager specific.
 *                             The typical case is a device registered in the
 *                             file system server (FSS).
 * @input           params     Specifies any required mounting parameters and
 *                             depends on the file manager and volume manager.
 *                             The params argument is a pointer to a string
 *                             with multiple optional parameter in the form
 *                             "name1=value1,name2=value2 .... ". Also boolean
 *                             parameters can be given, i.e. a name without
 *                             any "=value". See the volume manager specific
 *                             manuals for available parameters.
 *
 * @return
 * @tab             EFS_SUCCESS @tab The volume was successfully mounted. @break
 * @tab             FM_E*       @tab Upon failure the mount/volume manager
 *                                   indicates errors using the
 *                                   status codes defined in fm.sig.
 *                                   The error statuses returned
 *                                   are mount/volume manager specific. @break
 * @tab             DDB_E*      @tab Any DDB specific error status.
 *                                   The error statuses returned by
 *                                   the DDBs are specific to each DDB. @break
 * @tab             E*          @tab Upon failure the format/volume manager
 *                                   indicates errors using the OSE
 *                                   errno statuses, see
 *                                   @link "errno.h" "errno".
 *                                   The error statuses returned
 *                                   are format/volume manager specific.@break
 * @tab                         @tab FM_E*, DDB_E* codes can be translated to
 *                                   errno codes using
 *                                   @link "" "efs_status_to_errno". @break
 * @tab                         @tab Global variable  @link "errno.h" "errno"
 *                                   is not updated by this function. @break
 *
 * @seealso         efs_format
 * @seealso         efs_unmount
 * @seealso         default_restrictions_calling_functions
 *
 * @example
 * void
 * mount_my_volume(void)
 * {
 *   // Example mounting a partition (blocks 0..199) of the device named
 *   // 'ramdisk' with 9 extended FAT buffers and 20 directory buffers.
 *   //  The resulting volume will be called '/ram'.
 *   int err;
 *   EfsStatus ret = efs_mount("/ram", "extfat", "ramdisk",
 *       "unit=0,lo=0,hi=199,fatbuffers=9,dirbuffers=20");
 *
 *   if (ret == EFS_SUCCESS)
 *   {
 *     // Succeeded.
 *   }
 *   else
 *   {
 *     // Analyze error.
 *     err = efs_status_to_errno(ret);
 *     // ...
 *   }
 * }
 */
EfsStatus efs_mount(const char *volume, const char *format, const char *dev,
                    const char *params);


/**
 * @function        efs_wmount
 *
 * @include         "efs.h"
 *
 * @declaration     EfsStatus
 *                  efs_wmount(const wchar_t *volume, const char *format,
 *                             const char *dev, const char *params);
 *
 * @brief           Mount a volume on the indicated device dev.
 *
 * @long            This call is equivalent to @link "" "efs_mount" with the
 *                  exception that the volume name is passed as a wchar_t
 *                  string.
 * @long            The name of the volume is set to "volume" and the "format"
 *                  indicates the file system format expected on the volume.
 *                  Optional parameters are given in "params". If the
 *                  volume is not formatted according to "format", it will
 *                  anyhow be mounted, in unformatted state, and can be
 *                  formatted with e.g. the function @link "" "efs_wformat".
 *
 * @long            Upon successful completion, a volume manager (VM) will be
 *                  available through which files on this volume are accessed.
 *
 *                  The volume name is used as the leading part of the absolute
 *                  path for the file names on that volume and the volume name
 *                  must have a leading /.
 *
 * @input           volume     Points to the volume path string, used to
 *                             identify the volume. Must begin with "/".
 * @input           format     A pointer to the format name of an existing file
 *                             manager format to use. Examples are "jeff",
 *                             "vfat". A file manager (FM) handle the
 *                             specific format must be known to the File System
 *                             Server (ose_fss process).
 * @input           dev        A pointer to the name of the device to mount.
 *                             The use of this string is voume manager specific.
 *                             The typical case is a device registered in the
 *                             file system server (FSS).
 * @input           params     Specifies any required mounting parameters and
 *                             depends on the file manager and volume manager.
 *                             The params argument is a pointer to a string
 *                             with multiple optional parameter in the form
 *                             "name1=value1,name2=value2 .... ". Also boolean
 *                             parameters can be given, i.e. a name without
 *                             any "=value". See the volume manager specific
 *                             manuals for available parameters.
 *
 * @return
 * @tab             EFS_SUCCESS @tab The volume was successfully mounted. @break
 * @tab             FM_E*       @tab Upon failure the mount/volume manager
 *                                   indicates errors using the
 *                                   status codes defined in fm.sig.
 *                                   The error statuses returned
 *                                   are mount/volume manager specific. @break
 * @tab             DDB_E*      @tab Any DDB specific error status.
 *                                   The error statuses returned by
 *                                   the DDBs are specific to each DDB. @break
 * @tab             E*          @tab Upon failure the format/volume manager
 *                                   indicates errors using the OSE
 *                                   errno statuses, see
 *                                   @link "errno.h" "errno".
 *                                   The error statuses returned
 *                                   are format/volume manager specific.@break
 * @tab                         @tab FM_E*, DDB_E* codes can be translated to
 *                                   errno codes using
 *                                   @link "" "efs_status_to_errno". @break
 * @tab                         @tab Global variable  @link "errno.h" "errno"
 *                                   is not updated by this function. @break
 *
 * @seealso         efs_format
 * @seealso         efs_unmount
 * @seealso         default_restrictions_calling_functions
 *
 * @example
 * void
 * mount_my_volume(void)
 * {
 *   // Example mounting a partition (blocks 0..199) of the device named
 *   // 'ramdisk' with 9 extended FAT buffers and 20 directory buffers.
 *   //  The resulting volume will be called '/ram'.
 *   int err;
 *   EfsStatus ret = efs_wmount(L"/ram", "extfat", "ramdisk",
 *       "unit=0,lo=0,hi=199,fatbuffers=9,dirbuffers=20");
 *
 *   if (ret == EFS_SUCCESS)
 *   {
 *     // Succeeded.
 *   }
 *   else
 *   {
 *     // Analyze error.
 *     err = efs_status_to_errno(ret);
 *     // ...
 *   }
 * }
 */
EfsStatus efs_wmount(const wchar_t *volume, const char *format, const char *dev,
                     const char *params);


/**
 * @function        efs_relabel
 *
 * @include         "efs.h"
 *
 * @declaration     EfsStatus
 *                  efs_relabel(const char *volume, const char *new_label);
 *
 * @brief           Change the label on a volume.
 *
 * @long            If efs_relabel succeeds or not depends on if the volume
 *                  manager supports a label.
 *
 * @input           volume     Points to the volume name string, as specified
 *                             in @link "" "efs_mount".
 * @input           new_label  A string containing the new label.
 *
 * @return
 * @tab             EFS_SUCCESS  @tab The volume was successfully relabeled.
 * @break
 * @tab             FM_E*        @tab Upon failure the format/volume manager
 *                                    indicates errors using the
 *                                    status codes defined in fm.sig.
 *                                    The error statuses returned
 *                                    are format/volume manager specific. @break
 * @tab             DDB_E*       @tab Any DDB specific error status.
 *                                    The error statuses returned by
 *                                    the DDBs are specific to each DDB. @break
 * @tab             E*           @tab Upon failure the format/volume manager
 *                                    indicates errors using the OSE
 *                                    errno statuses, see
 *                                    @link "errno.h" "errno".
 *                                    The error statuses returned
 *                                    are format/volume manager specific.@break
 * @tab                          @tab FM_E*, DDB_E* codes can be translated to
 *                                    errno codes using
 *                                    @link "" "efs_status_to_errno". @break
 * @tab                          @tab Global variable  @link "errno.h" "errno"
 *                                    is not updated by this function. @break
 *
 * @seealso         efs_format
 * @seealso         efs_mount
 * @seealso         default_restrictions_calling_functions
 *
 * @example
 * void
 * relabel_my_volume(void)
 * {
 *   // Set the label on a volume:
 *
 *   EfsStatus res;
 *   res = efs_relabel("/ram","My company");
 * }
 */
EfsStatus efs_relabel(const char *volume, const char *new_label);


/**
 * @function        efs_wrelabel
 *
 * @include         "efs.h"
 *
 * @declaration     EfsStatus
 *                  efs_wrelabel(const wchar_t *volume, const char *new_label);
 *
 * @brief           Change the label on a volume.
 *
 * @long            This call is equivalent to @link "" "efs_relabel" with the
 *                  exception that the volume name is passed as a wchar_t
 *                  string.
 * @long            If efs_wrelabel succeeds or not depends on if the volume
 *                  manager supports a label.
 * @long            Note that the label is provided as a char string, since
 *                  not all volume managers support UTF-8 labels. The
 *                  application can convert a wide character label to UTF-8
 *                  with e.g. @link "ose_unicode.h" "wcs_to_utf8s"().
 *
 * @input           volume     Points to the volume name string, as specified
 *                             in e.g. @link "" "efs_wmount".
 * @input           new_label  A string containing the new label.
 *
 * @return
 * @tab             EFS_SUCCESS  @tab The volume was successfully relabeled.
 * @break
 * @tab             FM_E*        @tab Upon failure the format/volume manager
 *                                    indicates errors using the
 *                                    status codes defined in fm.sig.
 *                                    The error statuses returned
 *                                    are format/volume manager specific. @break
 * @tab             DDB_E*       @tab Any DDB specific error status.
 *                                    The error statuses returned by
 *                                    the DDBs are specific to each DDB. @break
 * @tab             E*           @tab Upon failure the format/volume manager
 *                                    indicates errors using the OSE
 *                                    errno statuses, see
 *                                    @link "errno.h" "errno".
 *                                    The error statuses returned
 *                                    are format/volume manager specific.@break
 * @tab                          @tab FM_E*, DDB_E* codes can be translated to
 *                                    errno codes using
 *                                    @link "" "efs_status_to_errno". @break
 * @tab                          @tab Global variable  @link "errno.h" "errno"
 *                                    is not updated by this function. @break
 *
 * @seealso         efs_format
 * @seealso         efs_mount
 * @seealso         default_restrictions_calling_functions
 *
 * @example
 * void
 * relabel_my_volume(void)
 * {
 *   // Set the label on a volume:
 *
 *   EfsStatus res;
 *   res = efs_wrelabel(L"/ram","My company");
 * }
 */
EfsStatus efs_wrelabel(const wchar_t *volume, const char *new_label);


/**
 * @function        efs_getlabel
 *
 * @include         "efs.h"
 *
 * @declaration     EfsStatus
 *                  efs_getlabel(const char *volume, char *label,
 *                               size_t *label_size);
 *
 * @brief           Get the label of a volume.
 *
 * @long            If efs_getlabel succeeds or not depends on if the volume
 *                  manager supports a label.
 *
 * @input           volume     Points to the volume name string, as specified
 *                             in @link "" "efs_mount".
 * @input           label      Pointer to where the label name shall be stored.
 * @input           label_size Pointer to size in bytes of memory area pointed
 *                             to by label.
 * @output          label      Null terminated string containing the volume
 *                             label if successful.
 * @output          label_size Pointer to size in bytes of the volume label
 *                             including trailing nul character. Note that
 *                             *label_size is updated also when error code
 *                             E2BIG is returned.
 *
 * @return
 * @tab             EFS_SUCCESS  @tab The operation was successful. @break
 * @tab             E2BIG        @tab *label_size was smaller than the volume
 *                                    label. *label_size is updated with the
 *                                    actual label size on return. @break
 * @tab             E*           @tab Upon failure the format/volume manager
 *                                    indicates errors using the
 *                                    errno statuses.
 *                                    The error statuses returned
 *                                    are format/volume manager specific.@break
 * @tab                          @tab Not errno (FM_E*, DDB_E*,...) codes can be
 *                                    translated to errno codes using
 *                                    @link "" "efs_status_to_errno". @break
 * @tab                          @tab Global variable  @link "errno.h" "errno"
 *                                    is not updated by this function. @break
 *
 * @seealso         efs_format
 * @seealso         efs_mount
 * @seealso         efs_relabel
 * @seealso         default_restrictions_calling_functions
 *
 * @example
 * void
 * get_myvolume_label(void)
 * {
 *   // Get the label on a volume:
 *   char label[512];
 *
 *   EfsStatus res;
 *   size_t label_size = sizeof(label);
 *   res = efs_getlabel("/ram", label, &label_size);
 * }
 */
EfsStatus efs_getlabel(const char *volume, char *label, size_t *label_size);


/**
 * @function        efs_wgetlabel
 *
 * @include         "efs.h"
 *
 * @declaration     EfsStatus
 *                  efs_wgetlabel(const wchar_t *volume, char *label,
 *                                size_t *label_size);
 *
 * @brief           Get the label of a volume.
 *
 * @long            This call is equivalent to @link "" "efs_getlabel" with the
 *                  exception that the volume name is passed as a wchar_t
 *                  string.
 * @long            If efs_wgetlabel succeeds or not depends on if the volume
 *                  manager supports a label.
 * @long            Note that the label is provided as a char string, since
 *                  not all volume managers support UTF-8 labels. The
 *                  application can convert a UTF-8 label to wide characters
 *                  with e.g. @link "ose_unicode.h" "utf8s_to_wcs"().
 *
 * @input           volume     Points to the volume name string, as specified
 *                             in e.g. @link "" "efs_wmount".
 * @input           label      Pointer to where the label name shall be stored.
 * @input           label_size Pointer to size in bytes of memory area pointed
 *                             to by label.
 * @output          label      Null terminated string containing the volume
 *                             label if successful.
 * @output          label_size Pointer to size in bytes of the volume label
 *                             including trailing nul character. Note that
 *                             *label_size is updated also when error code
 *                             E2BIG is returned.
 *
 * @return
 * @tab             EFS_SUCCESS  @tab The operation was successful. @break
 * @tab             E2BIG        @tab *label_size was smaller than the volume
 *                                    label. *label_size is updated with the
 *                                    actual label size on return. @break
 * @tab             E*           @tab Upon failure the format/volume manager
 *                                    indicates errors using the
 *                                    errno statuses.
 *                                    The error statuses returned
 *                                    are format/volume manager specific.@break
 * @tab                          @tab Not errno (FM_E*, DDB_E*,...) codes can be
 *                                    translated to errno codes using
 *                                    @link "" "efs_status_to_errno". @break
 * @tab                          @tab Global variable  @link "errno.h" "errno"
 *                                    is not updated by this function. @break
 *
 * @seealso         efs_format
 * @seealso         efs_mount
 * @seealso         efs_relabel
 * @seealso         default_restrictions_calling_functions
 *
 * @example
 * void
 * get_myvolume_label(void)
 * {
 *   // Get the label on a volume:
 *   char label[512];
 *
 *   EfsStatus res;
 *   size_t label_size = sizeof(label);
 *   res = efs_wgetlabel(L"/ram", label, &label_size);
 * }
 */
EfsStatus efs_wgetlabel(const wchar_t *volume, char *label, size_t *label_size);


/**
 * @function        efs_sync
 *
 * @include         "efs.h"
 *
 * @declaration     EfsStatus
 *                  efs_sync(const char *volume);
 *
 * @brief           Write all data cached in volume manager and device driver
 *                  to media.
 * @long            To really force all data out, first call fflush(NULL) to
 *                  force all open stream buffers out, then call fsync() for all
 *                  open file descriptors.
 * @long            To finally make the entire state of a volume persistent,
 *                  call efs_sync(). efs_sync forces all metadata to the media,
 *                  excluding any user data in non-committed transactional
 *                  file updates. It includes metadata changes such as file
 *                  name changes and access right changes, which are not
 *                  related to open files.
 * @long            efs_sync may cause a significant performance impact, and
 *                  should be used sparingly.
 *
 * @long            NOTE: The efs_sync behavior is dependent on how the
 *                        volume manager implements caching and when the volume
 *                        manager commit data to disk. For further information,
 *                        read the volume manager specific manuals.
 *
 * @input           volume     Points to the volume name string, as specified
 *                             in @link "" "efs_mount".
 *
 * @return
 * @tab             EFS_SUCCESS  @tab The volume was successfully synced. @break
 * @tab             FM_E*        @tab Upon failure the format/volume manager
 *                                    indicates errors using the
 *                                    status codes defined in fm.sig.
 *                                    The error statuses returned
 *                                    are format/volume manager specific. @break
 * @tab             E*           @tab Upon failure the format/volume manager
 *                                    indicates errors using the OSE
 *                                    errno statuses, see
 *                                    @link "errno.h" "errno".
 *                                    The error statuses returned
 *                                    are format/volume manager specific.@break
 * @tab             DDB_E*       @tab Any DDB specific error status.
 *                                    The error statuses returned by
 *                                    the DDBs are specific to each DDB. @break
 * @tab                          @tab FM_E*, DDB_E* codes can be translated to
 *                                    errno codes using
 *                                    @link "" "efs_status_to_errno". @break
 * @tab                          @tab Global variable  @link "errno.h" "errno"
 *                                    is not updated by this function. @break
 *
 * @seealso         fflush
 * @seealso         fsync
 * @seealso         default_restrictions_calling_functions
 *
 * @example
 * void
 * sync_my_volume(void)
 * {
 *   // Sync a volume.
 *
 *   EfsStatus res;
 *   res = efs_sync("/ram");
 * }
 */
EfsStatus efs_sync(const char *volume);


/**
 * @function        efs_wsync
 *
 * @include         "efs.h"
 *
 * @declaration     EfsStatus
 *                  efs_wsync(const wchar_t *volume);
 *
 * @brief           Write all data cached in volume manager and device driver
 *                  to media.
 * @long            This call is equivalent to @link "" "efs_sync" with the
 *                  exception that the volume name is passed as a wchar_t
 *                  string.
 * @long            To really force all data out, first call fflush(NULL) to
 *                  force all open stream buffers out, then call fsync() for all
 *                  open file descriptors.
 * @long            To finally make the entire state of a volume persistent,
 *                  call efs_wsync(). efs_wsync forces all metadata to the
 *                  media, excluding any user data in non-committed
 *                  transactional file updates. It includes metadata changes
 *                  such as file name changes and access right changes, which
 *                  are not related to open files.
 * @long            efs_wsync may cause a significant performance impact, and
 *                  should be used sparingly.
 *
 * @long            NOTE: The efs_wsync behavior is dependent on how the
 *                        volume manager implements caching and when the volume
 *                        manager commit data to disk. For further information,
 *                        read the relevant section in Core Extensions User's
 *                        Guide
 *
 * @input           volume     Points to the volume name string, as specified
 *                             in e.g. @link "" "efs_wmount".
 *
 * @return
 * @tab             EFS_SUCCESS  @tab The volume was successfully synced. @break
 * @tab             FM_E*        @tab Upon failure the format/volume manager
 *                                    indicates errors using the
 *                                    status codes defined in fm.sig.
 *                                    The error statuses returned
 *                                    are format/volume manager specific. @break
 * @tab             E*           @tab Upon failure the format/volume manager
 *                                    indicates errors using the OSE
 *                                    errno statuses, see
 *                                    @link "errno.h" "errno".
 *                                    The error statuses returned
 *                                    are format/volume manager specific.@break
 * @tab             DDB_E*       @tab Any DDB specific error status.
 *                                    The error statuses returned by
 *                                    the DDBs are specific to each DDB. @break
 * @tab                          @tab FM_E*, DDB_E* codes can be translated to
 *                                    errno codes using
 *                                    @link "" "efs_status_to_errno". @break
 * @tab                          @tab Global variable  @link "errno.h" "errno"
 *                                    is not updated by this function. @break
 *
 * @seealso         fflush
 * @seealso         fsync
 * @seealso         default_restrictions_calling_functions
 *
 * @example
 * void
 * sync_my_volume(void)
 * {
 *   // Sync a volume.
 *
 *   EfsStatus res;
 *   res = efs_wsync(L"/ram");
 * }
 */
EfsStatus efs_wsync(const wchar_t *volume);


/**
 * @function        efs_unmount
 *
 * @include         "efs.h"
 *
 * @declaration     EfsStatus
 *                  efs_unmount(const char *volume, Boolean force);
 *
 * @brief           Unmount volume.
 *
 * @input           volume     Points to the volume name string, as specified
 *                             in @link "" "efs_mount".
 * @input           force      With force is set to False, efs_unmount will
 *                             fail if any file is open on the volume. With
 *                             force set True, unmount will be done even if
 *                             there are opened files on the volume.
 *
 * @return
 * @tab             EFS_SUCCESS  @tab The volume was successfully unmounted.
 * @break
 * @tab             FM_E*        @tab Upon failure the format/volume manager
 *                                    indicates errors using the
 *                                    status codes defined in fm.sig.
 *                                    The error statuses returned
 *                                    are format/volume manager specific. @break
 * @tab             E*           @tab Upon failure the format/volume manager
 *                                    indicates errors using the OSE
 *                                    errno statuses, see
 *                                    @link "errno.h" "errno".
 *                                    The error statuses returned
 *                                    are format/volume manager specific.@break
 * @tab             DDB_E*       @tab Any DDB specific error status.
 *                                    The error statuses returned by
 *                                    the DDBs are specific to each DDb. @break
 * @tab                          @tab FM_E*, DDB_E* codes can be translated to
 *                                    errno codes using
 *                                    @link "" "efs_status_to_errno". @break
 * @tab                          @tab Global variable  @link "errno.h" "errno"
 *                                    is not updated by this function. @break
 *
 * @seealso         efs_unmount
 * @seealso         default_restrictions_calling_functions
 *
 * @example
 * void
 * unmount_my_volume(void)
 * {
 *   // Unmount a volume.
 *
 *   if(efs_unmount("/ram", True) == EFS_SUCCESS)
 *   {
 *     // Unmount successful.
 *   }
 * }
 */
EfsStatus efs_unmount(const char *volume, EfsBoolean force);


/**
 * @function        efs_wunmount
 *
 * @include         "efs.h"
 *
 * @declaration     EfsStatus
 *                  efs_wunmount(const wchar_t *volume, Boolean force);
 *
 * @brief           Unmount volume.
 *
 * @long            This call is equivalent to @link "" "efs_unmount" with the
 *                  exception that the volume name is passed as a wchar_t
 *                  string.
 *
 * @input           volume     Points to the volume name string, as specified
 *                             in e.g. @link "" "efs_wmount".
 * @input           force      With force is set to False, efs_wunmount will
 *                             fail if any file is open on the volume. With
 *                             force set True, unmount will be done even if
 *                             there are opened files on the volume.
 *
 * @return
 * @tab             EFS_SUCCESS  @tab The volume was successfully unmounted.
 * @break
 * @tab             FM_E*        @tab Upon failure the format/volume manager
 *                                    indicates errors using the
 *                                    status codes defined in fm.sig.
 *                                    The error statuses returned
 *                                    are format/volume manager specific. @break
 * @tab             E*           @tab Upon failure the format/volume manager
 *                                    indicates errors using the OSE
 *                                    errno statuses, see
 *                                    @link "errno.h" "errno".
 *                                    The error statuses returned
 *                                    are format/volume manager specific.@break
 * @tab             DDB_E*       @tab Any DDB specific error status.
 *                                    The error statuses returned by
 *                                    the DDBs are specific to each DDb. @break
 * @tab                          @tab FM_E*, DDB_E* codes can be translated to
 *                                    errno codes using
 *                                    @link "" "efs_status_to_errno". @break
 * @tab                          @tab Global variable  @link "errno.h" "errno"
 *                                    is not updated by this function. @break
 *
 * @seealso         efs_unmount
 * @seealso         default_restrictions_calling_functions
 *
 * @example
 * void
 * unmount_my_volume(void)
 * {
 *   // Unmount a volume.
 *
 *   if(efs_wunmount(L"/ram", True) == EFS_SUCCESS)
 *   {
 *     // Unmount successful.
 *   }
 * }
 */
EfsStatus efs_wunmount(const wchar_t *volume, EfsBoolean force);


/**
 * @function        efs_shutdown
 *
 * @include         "efs.h"
 *
 * @declaration     EfsStatus
 *                  efs_shutdown(const char *device, EfsEnum type,
 *                               EfsBoolean force);
 *
 * @brief           Shutdown the block device driver.
 *
 * @long            The shutdown function shall be used on an unmounted
 *                  block device driver to ensure fast restart.
 *
 * @input           device     Points to the device name.
 * @input           type       The FSS type (FSS_TYPE_BLOCKDEV).
 * @input           force      Force shutdown if True.
 *
 * @return
 * @tab             EFS_SUCCESS  @tab The volume was successfully shutdown.
 * @break
 * @tab             DDB_E*       @tab Any DDB specific error status.
 *                                    The error statuses returned by
 *                                    the DDBs are specific to each DDB.@break
 * @tab             E*           @tab Upon failure DDB
 *                                    indicates errors using the OSE
 *                                    errno statuses, see
 *                                    @link "errno.h" "errno".
 *                                    The error statuses returned
 *                                    are specific to the DDB.@break
 * @tab                          @tab Not errno (FM_E*, DDB_E*,...) codes can be
 *                                    translated to errno codes using
 *                                    @link "" "efs_status_to_errno". @break
 * @tab                          @tab Global variable  @link "errno.h" "errno"
 *                                    is not updated by this function. @break
 *
 * @seealso         efs_unmount
 * @seealso         efs_mount
 * @seealso         default_restrictions_calling_functions
 *
 * @example
 * void
 * unmount_shutdown_my_volume(void)
 * {
 *   // Unmount a volume.
 *
 *   if(efs_unmount("/ram", True) == EFS_SUCCESS)
 *   {
 *     // Unmount successful.
 *     if(efs_shutdown("ffxddb", FSS_TYPE_BLOCKDEV, True) == EFS_SUCCESS)
 *     {
 *       // Shutdown successful.
 *     }
 *   }
 * }
 */
EfsStatus efs_shutdown(const char *device, EfsEnum type, EfsBoolean force);


/**
 * @function       efs_remove
 *
 * @include        "efs.h"
 *
 * @declaration    EfsStatus
 *                 efs_remove(const char *path, EfsBitmask types,
 *                            EfsBoolean force);
 *
 * @brief          Remove a specified object from a specified volume.
 *
 * @long           efs_remove() removes a subdirectory or file.
 *                 To remove an opened or write protected file or directory the
 *                 "force" flag must be set to True (1). A file or directory
 *                 becomes write protected if the parent directory is write
 *                 protected.
 * @long           Directories need to be empty to be removed.
 *
 * @input path     The path name of the object to remove.
 * @input types    The type of the object to remove. The value has a bit set to
 *                 indicate the type (1 << FM_TYPE_xxx). For the FM_TYPE_xxx
 *                 enum names, see @link "ose_spi/fm.sig" "FmTypeEnum".
 * @input force    If true, objects in use or write protected objects will
 *                 be removed, if supported by the volume manager.
 *
 * @return
 * @tab            EFS_SUCCESS               @tab The removal was successful.
 * @break
 * @tab            FLIB_ELOST_CONTACT or
 *                 ESTALE                    @tab Lost contact with volume
 *                                                manager. @break
 * @tab            FLIB_EUNKNOWN_VOLUME or
 *                 ENOENT                    @tab efs_remove() called with
 *                                                unknown volume. @break
 * @tab            FM_EOBJECT_DELETE_PROTECT
 *                 or EACCES                 @tab The parent directory is
 *                                                write protected. @break
 * @tab            FM_EOBJECT_IN_USE
 *                 or EACCES                 @tab The object to be removed is
 *                                                used and no force flag is set.
 *                                                @break
 * @tab            FM_EDIR_NOT_EMPTY or
 *                 ENOTEMPTY                 @tab A specified directory was not
 *                                                empty. @break
 * @tab            FM_EOBJECT_WRITE_PROTECT
 *                 or EACCES                 @tab The object to be removed was
 *                                                write protected and no force
 *                                                flag set. @break
 * @tab            FM_EOPER_NOT_SUPPORTED
 *                 or ENOTSUP                @tab Tried to perform an
 *                                                unsupported operation. @break
 * @tab            FM_EOBJECT_WRONG_TYPE
 *                 or EINVAL                 @tab Object type not recognized.
 *                                                @break
 * @tab                                      @tab Not errno (FM_E*, DDB_E*,...)
 *                                                codes can be translated to
 *                                                errno codes using
 *                                              @link "" "efs_status_to_errno".
 *                                                @break
 *
 * @tab                                      @tab Global variable
 *                                                @link "errno.h" "errno" is not
 *                                                updated by this function.
 *                                                @break
 *
 * @seealso remove
 * @seealso rmdir
 * @seealso default_restrictions_calling_functions
 *
 * @example
 * static int rmdirforced(const char *directory)
 * {
 *    EfsBitmask types = 1 << FM_TYPE_DIR;
 *    EfsStatus status = efs_remove(directory, types, True);
 *    if (status != EFS_SUCCESS)
 *    {
 *      status = handle_other_return_value();
 *    }
 *    return status;
 *  }
 */
EfsStatus efs_remove(const char *path, EfsBitmask types, EfsBoolean force);


/**
 * @function       efs_wremove
 *
 * @include        "efs.h"
 *
 * @declaration    EfsStatus
 *                 efs_wremove(const wchar_t *path, EfsBitmask types,
 *                             EfsBoolean force);
 *
 * @brief          Remove a specified object from a specified volume.
 *
 * @long            This call is equivalent to @link "" "efs_remove" with the
 *                  exception that the volume name is passed as a wchar_t
 *                  string.
 * @long           efs_wremove() removes a subdirectory or file.
 *                 To remove an opened or write protected file or directory the
 *                 "force" flag must be set to True (1). A file or directory
 *                 becomes write protected if the parent directory is write
 *                 protected.
 * @long           Directories need to be empty to be removed.
 *
 * @input path     The path name of the object to remove.
 * @input types    The type of the object to remove. The value has a bit set to
 *                 indicate the type (1 << FM_TYPE_xxx). For the FM_TYPE_xxx
 *                 enum names, see @link "ose_spi/fm.sig" "FmTypeEnum".
 * @input force    If true, objects in use or write protected objects will
 *                 be removed, if supported by the volume manager.
 *
 * @return
 * @tab            EFS_SUCCESS               @tab The removal was successful.
 * @break
 * @tab            FLIB_ELOST_CONTACT or
 *                 ESTALE                    @tab Lost contact with volume
 *                                                manager. @break
 * @tab            FLIB_EUNKNOWN_VOLUME or
 *                 ENOENT                    @tab efs_wremove() called with
 *                                                unknown volume. @break
 * @tab            FM_EOBJECT_DELETE_PROTECT
 *                 or EACCES                 @tab The parent directory is
 *                                                write protected. @break
 * @tab            FM_EOBJECT_IN_USE
 *                 or EACCES                 @tab The object to be removed is
 *                                                used and no force flag is set.
 *                                                @break
 * @tab            FM_EDIR_NOT_EMPTY or
 *                 ENOTEMPTY                 @tab A specified directory was not
 *                                                empty. @break
 * @tab            FM_EOBJECT_WRITE_PROTECT
 *                 or EACCES                 @tab The object to be removed was
 *                                                write protected and no force
 *                                                flag set. @break
 * @tab            FM_EOPER_NOT_SUPPORTED
 *                 or ENOTSUP                @tab Tried to perform an
 *                                                unsupported operation. @break
 * @tab            FM_EOBJECT_WRONG_TYPE
 *                 or EINVAL                 @tab Object type not recognized.
 *                                                @break
 * @tab                                      @tab Not errno (FM_E*, DDB_E*,...)
 *                                                codes can be translated to
 *                                                errno codes using
 *                                                @link ""
 *                                                "efs_status_to_errno".
 *                                                @break
 *
 * @tab                                      @tab Global variable
 *                                                @link "errno.h" "errno" is not
 *                                                updated by this function.
 *                                                @break
 *
 * @seealso remove
 * @seealso rmdir
 * @seealso default_restrictions_calling_functions
 *
 * @example
 * static int wrmdirforced(const wchar_t *directory)
 * {
 *    EfsBitmask types = 1 << FM_TYPE_DIR;
 *    EfsStatus status = efs_remove(directory, types, True);
 *    if (status != EFS_SUCCESS)
 *    {
 *      status = handle_other_return_value();
 *    }
 *    return status;
 *  }
 */
EfsStatus efs_wremove(const wchar_t *path, EfsBitmask types, EfsBoolean force);


/**
 * @function       eprintf
 *
 * @include        "efs.h"
 *
 * @declaration    int eprintf(const char *fmtstr, ...);
 *
 * @brief          eprintf() formats and prints a string to stderr.
 *
 * @long           The eprintf function creates formatted output on stderr.
 *                 Otherwise it is identical to printf() and fprintf().
 *                 See @link "stdio.h" "fprintf" for details. See also
 *                 ANSI C fprintf() documentation for formats and definitions.
 *
 * @input          fmtstr     The format command string.
 * @input          ...        Parameters to be formatted and printed.
 *                            See @link "stdio.h" "fprintf" for details and
 *                            examples.
 *
 * @return         The number of characters written. EOF will be returned
 *                 on failure. If eprintf fails @link "errno.h" "errno"
 *                 will be set.
 *
 * @seealso        printf
 * @seealso        fprintf
 * @seealso        stderr
 * @seealso        default_restrictions_calling_functions
 *
 */
int eprintf(const char *fmtstr, ...); /* Equals fprintf(stderr, fmtstr, ...); */


/******************************************************************************
 * Functions operating on file descriptors and FILE's. Except for
 * efs_read_line() these functions do not set the errno variable. Instead
 * status is returned directly to the caller.
 *****************************************************************************/


/**
 * @function       efs_read_line
 *
 * @include        "efs.h"
 *
 * @declaration    int efs_read_line(int fd, char *buffer, size_t buffer_size,
 *                                   Boolean echo);
 *
 * @brief          Reads a line of input from file descriptor.
 *
 * @input          fd            File descriptor to read from.
 * @input          buffer_size   The size of the output buffer, including the
 *                               terminating zero.
 * @input          echo          True, if typed characters shall be echoed,
 *                               False, if not.
 *
 * @output         buffer        Zero-terminated output buffer for the read
 *                               characters.
 *
 * @return         The number of read characters or -1 on failure.
 *
 * @seealso        read
 * @seealso        default_restrictions_calling_functions
 */
int efs_read_line(int fd, char *buffer, size_t buffer_size, Boolean echo);


/**
 * @function       efs_examine_file
 *
 * @include        "efs.h"
 *
 * @declaration    EfsStatus efs_examine_file(const FILE * file, ...);
 *
 * @brief          Examine properties of a file pointer (FILE *).
 *
 * @long           The fields of the FILE structure are examined and returned.
 *                 The argument list is used to specify which properties to be
 *                 examined. The argument list consists of pairs of tag IDs
 *                 and a pointer to a variable to receive the value and is
 *                 terminated by a 0.
 *
 * @input          file          The file pointer to examine.
 * @input          ...           The properties of the file pointer to examine.
 *                 The tags available are defined in @link "" "EfsFileTag".
 *                 The syntax of the tag argument list is
 *                 "<tag>, <U32 pointer to value>, <tag2>,
 *                 <U32 pointer to value2>..."
 *                 The following tags can be used when examining a buffered
 *                 file:
 * @break
 * @tab            FLIB_FILE_BUFFER       @tab Retrieve the file buffer pointer.
 *                                             Interpret the returned value as a
 *                                             char **. @break
 * @tab            FLIB_FILE_BUFSIZE      @tab Retrieve the size of the file
 *                                             buffer. Interpret the returned
 *                                             value as a U32 *. @break
 * @tab            FLIB_FILE_CURRENT      @tab Retrieve the current position
 *                                             pointer. Interpret the returned
 *                                             value as a char **. @break
 * @tab            FLIB_FILE_CURRENT_REF  @tab Retrieve a reference pointer to
 *                                             the current position pointer.
 *                                             Interpret the returned value as
 *                                             a char ***. @break
 * @tab            FLIB_FILE_AVAILRD      @tab Retrieve the number of characters
 *                                             available in the buffer for
 *                                             reading before the buffer is
 *                                             empty.
 *                                             Interpret the returned value as a
 *                                             unsigned *. @break
 * @tab            FLIB_FILE_AVAILWR      @tab Retrieve the number of characters
 *                                             available in the buffer for
 *                                             writing before the buffer is
 *                                             full.
 *                                             Interpret the returned value as
 *                                             a unsigned *. @break
 * @tab            FLIB_FILE_AVAIL_REF    @tab Retrieve a reference pointer to
 *                                             the available field in the FILE
 *                                             struct.
 *                                             Interpret the returned value as
 *                                             a unsigned **. @break
 *
 * @output         ... The different output variables will be written with the
 *                   selected values.
 *
 * @return
 * @tab            EFS_SUCCESS             @tab Is returned upon successful
 *                                              execution. @break
 * @tab            FLIB_EINVALID_PARAMETER
 *                 or EINVAL               @tab An invalid parameter was
 *                                              specified in the argument list.
 *                                              @break
 * @tab            EBADF                   @tab An invalid file pointer was
 *                                              specified. @break
 * @tab                                    @tab Not errno (FM_E*, DDB_E*,...)
 *                                              codes can be translated to errno
 *                                              codes using
 *                                              @link "" "efs_status_to_errno".
 *                                              @break
 * @tab                                    @tab Global variable
 *                                              @link "errno.h" "errno" is not
 *                                              updated by this function. @break
 *
 * @seealso        efs_examine_fd
 * @seealso        default_restrictions_calling_functions
 *
 * @example
 * char *buffer;
 * char *current;
 * // Examine the stdin file pointer and retrieve the file buffer pointer in
 * // the buffer variable and the file content position pointer in current
 * // variable.
 * EfsStatus status = efs_examine_file(stdin, FLIB_FILE_BUFFER, &buffer,
 *                                FLIB_FILE_CURRENT, &current, 0);
 */
EfsStatus efs_examine_file(const FILE * file, ...);


/**
 * @function       efs_dup_to
 *
 * @include        "efs.h"
 *
 * @declaration    EfsStatus efs_dup_to(int fd, PROCESS to,
 *                                      FmHandle *new_handle);
 *
 * @brief          Duplicate a file handle and give ownership to another
 *                 process.
 *
 * @long           efs_dup_to is used to duplicate a file handle and give it to
 *                 the process of the "to" parameter. This function call can
 *                 only be used for duplications of file handles between
 *                 processes running on the same node. efs_dup_to() can also
 *                 handle duplications of socket descriptors.
 *
 * @input          fd            File descriptor to duplicate.
 * @input          to            Process to assume ownership of new file
 *                               descriptor.
 *
 * @output         new_handle    Pointer to a variable to be updated with the
 *                               new FmHandle.
 *
 * @return
 * @tab            EFS_SUCCESS             @tab Is returned upon successful
 *                                              execution. @break
 * @tab            FLIB_ESTALE_FD or
 *                 ESTALE                  @tab Will be returned if the contact
 *                                              is lost, or was lost, with the
 *                                              volume manager. @break
 * @tab            FLIB_EINVALID_FD or
 *                 EBADF                   @tab Will be returned if a bad file
 *                                              descriptor is used or if the
 *                                              handle is invalid. @break
 * @tab            FLIB_EINVALID_PARAMETER
 *                 or EINVAL               @tab Will be returned if the process
 *                                              name is invalid. @break
 * @tab            FLIB_ESOCKET_ERROR      @tab Will be returned in the case of
 *                                              a socket error. @break
 * @tab                                    @tab Not errno (FM_E*, DDB_E*,...)
 *                                              codes can be translated to errno
 *                                              codes using
 *                                              @link "" "efs_status_to_errno".
 *                                              @break
 * @tab                                    @tab Global variable
 *                                              @link "errno.h" "errno" is not
 *                                              updated by this function. @break
 *
 * @restriction    Note that efs_dup_to() can not handle an illegal file handle.
 *                 If an illegal file handle is used the application will hang
 *                 waiting for a signal that will never come.
 *
 * @seealso        dup
 * @seealso        dup2
 * @seealso        efs_donate_fd
 * @seealso        default_restrictions_calling_functions
 *
 * @example
 * // executing in process A
 * int fd;
 * FmHandle *handle;
 * fd = open("ose.txt", O_RDWR, S_IRGRP);
 * extern OSENTRYPOINT entry;
 * PROCESS B_ = create_process(OS_BG_PROC, "entry", entry, 1000, (OSPRIORITY) 0,
 *                             (OSTIME) 0, (PROCESS) 0, NULL, (OSVECTOR) 0,
 *                             (OSUSER) 0);
 * start(B_);
 * // Duplicating file descriptor to process B
 * efs_dup_to(fd, B_, handle);
 */
EfsStatus efs_dup_to(int fd, PROCESS to, FmHandle *new_handle);


/**
 * @function       efs_examine_term
 *
 * @include        "efs.h"
 *
 * @declaration    EfsStatus
 * efs_examine_term(int fd, ...)
 *
 * @brief          Examine properties of a terminal.
 *
 * @long           The argument list is used to specify which properties are
 *                 examined. The argument list consist of pairs of a tag ID and
 *                 one or more corresponding pointers that will be set to
 *                 contain the value(s) for that particular examine option. The
 *                 list is terminated by a 0 value.
 *
 * @input          fd     File descriptor representing the terminal.
 * @input          ...    The argument list.
 *                        There are several tags available for the argument
 *                        list:
 * @break
 * @tab            FM_TERM_READMIN    @tab Retrieve read minimum size (bytes),
 *                                         disabled by
 *                                         FM_TIFL_LINE_EDIT.
 *                                         The returned value shall be
 *                                         interpreted as U32 *. @break
 * @tab            FM_TERM_READTMO    @tab Retrieve read time out (in
 *                                         milliseconds),
 *                                         disabled by FM_TIFL_LINE_EDIT.
 *                                         The returned value shall be
 *                                         interpreted as U32 *. @break
 * @tab            FM_TERM_RXSPEED    @tab Retrieve input baud rate.
 *                                         The returned value shall be
 *                                         interpreted as U32 *. @break
 * @tab            FM_TERM_RXBUFFER   @tab Retrieve size of input buffer.
 *                                         The returned value shall be
 *                                         interpreted as U32 *. @break
 * @tab            FM_TERM_TXSPEED    @tab Retrieve output baud rate.
 *                                         The returned value shall be
 *                                         interpreted as U32 *. @break
 * @tab            FM_TERM_RXFLOW     @tab Retrieve flow of receive queue.
 *                                         The returned value shall be
 *                                         interpreted as U32 *. @break
 * @tab            FM_TERM_TXFLOW     @tab Retrieve flow of transmit queue.
 *                                         The returned value shall be
 *                                         interpreted as U32 *. @break
 * @tab            FM_TERM_CHARBITS   @tab Retrieve number of bits per
 *                                         character.
 *                                         The returned value shall be
 *                                         interpreted as U32 *. @break
 * @tab            FM_TERM_STOPBITS   @tab Retrieve number of stop bits.
 *                                         The returned value shall be
 *                                         interpreted as U32 *. @break
 * @tab            FM_TERM_PARITY     @tab Retrieve parity ('O'dd, 'E'ven or
 *                                         'N'o).
 *                                         The returned value shall be
 *                                         interpreted as U32 *. @break
 * @tab            FM_TERM_FLOWCTRL   @tab Retrieve flow control ('N'one or
 *                                         'H'ardware).
 *                                         The returned value shall be
 *                                         interpreted as U32 *. @break
 * @tab            FM_TERM_FLAGS      @tab Retrieve edit flags.
 *                                         The returned value shall be
 *                                         interpreted as U32 *
 *                                         where the bits are defined according
 *                                         to @link "fm.sig" "FmTiflEnum".
 *                                         @break
 * @tab            FM_TERM_WIDTH      @tab Retrieve width of terminal in
 *                                         characters.
 *                                         The returned value shall be
 *                                         interpreted as U32 *. @break
 * @tab            FM_TERM_HEIGHT     @tab Retrieve height of terminal in
 *                                         characters.
 *                                         The returned value shall be
 *                                         interpreted as U32 *. @break
 * @tab            FM_TERM_CTRLCHARS  @tab Retrieve control characters
 *                                         char[FM_TCC_SIZE].
 *                                         The returned value shall be
 *                                         interpreted as
 *                                         char[FM_TCC_SIZE]. @break
 * @tab            FM_TERM_VOLNAME    @tab Retrieve name (label) of the volume
 *                                         and the string length of the label.
 *                                         The returned value shall be
 *                                         interpreted as
 *                                         char * followed by size_t. @break
 * @tab            FM_TERM_DEVNAME    @tab Retrieve name of used device and the
 *                                         string length of the device name.
 *                                         The returned value shall be
 *                                         interpreted as
 *                                         char * followed by size_t. @break
 * @tab            FM_TERM_TERMTYPE   @tab Retrieve name of terminal type and
 *                                         the string
 *                                         length of the terminal type name.
 *                                         The returned value shall be
 *                                         interpreted as
 *                                         char * followed by size_t. @break
 *
 * @tab            FM_TERM_IDLETMO    @tab Retrieve idle time logout interval (in
 *                                         seconds).
 *                                         The returned value shall be
 *                                         interpreted as U32 *. @break
 *
 * @output         ...   The different output variables will be written with
 *                       the selected values.
 *
 * @return
 * @tab            EFS_SUCCESS            @tab Is returned upon successful
 *                                             execution. @break
 * @tab            FM_EINVALID_PARAMETERS
 *                 or EINVAL              @tab Will be returned if an invalid
 *                                             tag is specified in the argument
 *                                             list. @break
 * @tab            FLIB_ELOST_CONTACT or
 *                 ESTALE                 @tab Lost contact with volume manager.
 *                                             @break
 * @tab            FLIB_ESTALE_FD or
 *                 ESTALE                 @tab File descriptor is stale. @break
 * @tab            FLIB_EINVALID_FD or
 *                 EBADF                  @tab An invalid file descriptor is
 *                                             supplied to efs_examine_term().
 *                                             @break
 * @tab                                   @tab Not errno (FM_E*, FLIB_E*,...)
 *                                             codes can be translated to errno
 *                                             codes using
 *                                             @link "" "efs_status_to_errno".
 *                                             @break
 * @tab                                   @tab Global variable
 *                                             @link "errno.h" "errno" is not
 *                                             updated by this function. @break
 * @seealso        efs_config_term
 * @seealso        default_restrictions_calling_functions
 *
 * @example
 * unsigned char keys[FM_TCC_SIZE];
 * EfsStatus status = efs_examine_term(fileno(stdin), FM_TERM_CTRLCHARS,
 *                                     keys, 0);
 * if (status == EFS_SUCCESS)
 * {
 *    // Disable the interrupt key. //
 *      keys[FM_TCC_INTR] = FM_TCC_DISABLED;
 *     status = efs_config_term(fileno(stdin), FM_TERM_CTRLCHARS, keys, 0);
 *     if (status != EFS_SUCCESS)
 *     {
 *        // handle error //
 *     }
 *  }
 * else // handle error //
 */
EfsStatus efs_examine_term(int fd, ...);


/**
 * @function       efs_config_term
 *
 * @include        "efs.h"
 *
 * @declaration    EfsStatus
 * efs_config_term(int fd, ...);
 *
 * @brief          Configure a terminal.
 *
 * @long           efs_config_term() is used to configure the terminal
 *                 specified by the file descriptor. The argument list is used
 *                 to specify which properties to be examined. The argument
 *                 list consists of pairs of a tag IDs and the value of the tag
 *                 and is terminated by a 0 value.
 *
 * @input          fd     File descriptor representing the terminal.
 * @input          ...    The argument list.
 *                 There are several tags available for the argument list:
 * @tab            FM_TERM_READMIN     @tab Configure minimum read size,
 *                                          disabled by
 *                                          FM_TIFL_LINE_EDIT.
 *                                          The configuration value type is U32.
 * @break
 * @tab            FM_TERM_READTMO     @tab Configure read time out (in
 *                                          milliseconds),
 *                                          disabled by FM_TIFL_LINE_EDIT.
 *                                          The configuration value type is U32.
 * @break
 * @tab            FM_TERM_RXSPEED     @tab Configure input baud rate.
 *                                          The configuration value type is U32.
 * @break
 * @tab            FM_TERM_RXBUFFER    @tab Configure size of input buffer.
 *                                          The configuration value type is U32.
 * @break
 * @tab            FM_TERM_TXSPEED     @tab Configure output baud rate.
 *                                          The configuration value type is U32.
 * @break
 * @tab            FM_TERM_RXFLOW      @tab Configure flow of receive queue.
 *                                          The configuration value type is U32.
 * @break
 * @tab            FM_TERM_TXFLOW      @tab Configure flow of transmit queue.
 *                                          The configuration value type is U32.
 * @break
 * @tab            FM_TERM_CHARBITS    @tab Configure 5, 6, 7 or 8 bits per
 *                                          character.
 *                                          The configuration value type is int.
 * @break
 * @tab            FM_TERM_STOPBITS    @tab Configure 1 or 2 stop bits.
 *                                          The configuration value type is int.
 * @break
 * @tab            FM_TERM_PARITY      @tab Configure 'O'dd, 'E'ven or 'N'o
 *                                          parity.
 *                                          The configuration value type is int.
 * @break
 * @tab            FM_TERM_FLOWCTRL    @tab Configure type of flow control
 *                                          ('N'one or 'H'ardware).
 *                                          The configuration value type is int.
 * @break
 * @tab            FM_TERM_FLAGS       @tab Configure edit flags.
 *                                          The configuration value type is U32
 *                                          where the bits are defined according
 *                                          to @link
 *                                          "fm.sig" "FmTiflEnum".
 *                @break
 * @tab            FM_TERM_WIDTH       @tab Configure width of terminal in
 *                                          characters.
 *                                          The configuration value type is int.
 * @break
 * @tab            FM_TERM_HEIGHT      @tab Configure height of terminal in
 *                                          characters.
 *                                          The configuration value type is int.
 * @break
 * @tab            FM_TERM_CTRLCHARS   @tab Configure control characters.
 *                                          The configuration value type is
 *                                          char[FM_TCC_SIZE]. 
 * @break
 * @tab            FM_TERM_IDLETMO     @tab Configure idle logout time interval 
 *                                          (in seconds).
 *                                          The configuration value type is 
 *                                          U32. @break
 *
 * @return
 * @tab            EFS_SUCCESS            @tab Is returned upon successful
 *                                             execution. @break
 * @tab            FM_EINVALID_PARAMETERS
 *                 or EINVAL              @tab Will be returned if an invalid
 *                                             tag is specified in the argument
 *                                             list. @break
 * @tab            FLIB_ELOST_CONTACT or
 *                 ESTALE                 @tab Lost contact with volume manager.
 *                                             @break
 * @tab            FLIB_ESTALE_FD or
 *                 ESTALE                 @tab File descriptor is stale. @break
 * @tab            FLIB_EINVALID_FD or
 *                 EBADF                  @tab An invalid file descriptor is
 *                                             supplied to efs_examine_term().
 *                                             @break
 * @tab                                   @tab Not errno (FLIB_E*, DDB_E*,...)
 *                                             codes can be translated to errno
 *                                             codes using
 *                                             @link "" "efs_status_to_errno".
 *                                             @break
 * @tab                                   @tab Global variable
 *                                             @link "errno.h" "errno" is not
 *                                             updated by this function. @break
 *
 *
 *
 * @seealso        efs_examine_term
 * @seealso        default_restrictions_calling_functions
 */
EfsStatus efs_config_term(int fd, ...);


/**
 * @function       efs_examine_fd
 *
 * @include        "efs.h"
 *
 * @declaration    EfsStatus
 * efs_examine_fd(int fd, ...);
 *
 * @brief          Examine properties of a file descriptor.
 *
 * @long           These properties exist only in the application.
 *                 The argument list is used to specify which properties to be
 *                 examined. The argument list consists of pairs of a tag IDs
 *                 and a pointer to a variable to receive the value and is
 *                 terminated by a 0 value.
 *
 * @error OSE_EFS_FLIB_EUNRESOLVED_INET_FUNCTION INET function could not be
 *                 resolved. This error will be called with fatal bit set.
 *                 Only applies for socket connections.
 *
 * @input          fd    File descriptor to examine.
 * @input          ...   The argument list.
 *                 The tags available are defined in @link "" "EfsFdTag".
 *                 The syntax of the tag argument list is
 *                 "<U32 tag>, <pointer to U32 value>, <U32 tag2>, 
 *                 <pointer to U32 value2>..."
 *                 The following tags can be used when examining a file
 *                 descriptor:
 * @break
 * @tab           FLIB_FD_STALE   @tab Retrieve the stale status. A file
 *                                     descriptor becomes stale when contact
 *                                     with the volume manager is lost.
 *                                     Interpret the returned value as a
 *                                     boolean. @break
 * @tab           FLIB_FD_SHMEM   @tab Retrieve the status of shared memory
 *                                     reads and writes.
 *                                     Interpret the returned value as a
 *                                     boolean. @break
 * @tab           FLIB_FD_APPEND  @tab Retrieve the status of the append flag.
 *                                     Interpret the returned value as a
 *                                     boolean. @break
 * @tab           FLIB_FD_VMPID   @tab Retrieve the PID of the corresponding
 *                                     volume manager.
 *                                     Interpret the returned value as a
 *                                     @link "ose.h" "PROCESS". @break
 * @tab           FLIB_FD_HANDLE  @tab Retrieve the corresponding FmHandle.
 *                                     Interpret the returned value as a
 *                                     FmHandle (ose_spi/fm.sig). @break
 * @tab           FLIB_FD_BUFMODE @tab Retrieve the buffer mode to be used by
 *                                     FILE's opened
 *                                     from this fd.
 *                                     Interpret the returned value as a
 *                                     FmBuffEnum, see ose_spi/fm.sig. @break
 * @tab           FLIB_FD_BUFSIZE @tab Retrieve the buffer size (bytes) to be
 *                                     used by FILE's opened
 *                                     from this fd.
 *                                     Interpret the returned value as a U32.
 *                                     @break
 * @tab           FLIB_FD_SOCKET  @tab Retrieve socket status (is the file
 *                                     descriptor a socket descriptor).
 *                                     Interpret the returned value as a
 *                                     boolean. @break
 * @tab           FLIB_FD_CLONE   @tab Retrieve the clone status.
 *                                     Interpret the returned value as a
 *                                     boolean. @break
 *
 * @output        ... The various arguments in the argument list.
 *
 * @return
 * @tab            EFS_SUCCESS             @tab Is returned upon successful
 *                                              execution. @break
 * @tab            FLIB_EINVALID_PARAMETER
 *                 or EINVAL               @tab Will be returned if an invalid
 *                                              parameter is used. @break
 * @tab            FLIB_EINVALID_FD or
 *                 EBADF                   @tab Will be returned if a bad file
 *                                              descriptor is used. @break
 * @tab                                    @tab Not errno (FLIB_E*, DDB_E*,...)
 *                                              codes can be translated to errno
 *                                              codes using
 *                                              @link "" "efs_status_to_errno".
 *                                              @break
 * @tab                                   @tab Global variable
 *                                              @link "errno.h" "errno" is not
 *                                              updated by this function. @break
 * @seealso        efs_config_fd
 * @seealso        default_restrictions_calling_functions
 *
 * @example
 * // File descriptor fd is a valid fd, vm_pid and handle properly declared //
 * PROCESS vm_pid;
 * FmHandle handle;
 * if (efs_examine_fd(fd, FLIB_FD_VMPID, &vm_pid, FLIB_FD_HANDLE,
 *                    &handle, 0) == EFS_SUCCESS)
 * {
 *    // Send signals directly to volume manager. //
 * }
 */
EfsStatus efs_examine_fd(int fd, ...);


/**
 * @function       efs_config_fd
 *
 * @include        "efs.h"
 *
 * @declaration    EfsStatus efs_config_fd(int fd, ...);
 *
 * @brief          Configure a file descriptor.
 *
 * @long           efs_config_fd() is used to configure the properties of the
 *                 specified file descriptor. The argument list is used to
 *                 specify which properties to be configured and with what
 *                 value. The argument list consists of pairs of a tag IDs and
 *                 the value of the tag and is terminated by a 0 value.
 * @long           The values 1 and 0 represent Enable and Disable for boolean
 *                 values.
 *
 * @input          fd   The file descriptor to configure.
 * @input          ...  The variable argument list
 *                 The tags available are defined in @link "" "EfsFdTag".
 *                 The syntax of the tag argument list is
 *                 "<U32 tag>, <U32 value>, <U32 tag2>, <U32 value2>..."
 *                 The following tags can be used when configuring a file
 *                 descriptor:
 * @break
 * @tab            FLIB_FD_STALE   @tab Set the value to Disable(0), to make an
 *                                      fd stale, if the application has
 *                                      detected that
 *                                      the volume manager has terminated. It is
 *                                      not possible use the tag to make a stale
 *                                      fd active by setting the value to
 *                                      Enable(1). @break
 * @tab            FLIB_FD_SHMEM   @tab Enable(1) or Disable(0) shared memory
 *                                      model for the fd. @break
 * @tab            FLIB_FD_APPEND  @tab Enable(1) or Disable(0) the append mode
 *                                      of the fd. @break
 * @tab            FLIB_FD_BUFMODE @tab Set the buffer mode to be used by FILE's
 *                                      opened from this fd. Compare enum
 *                                      FmBuffEnum in ose_spi/fm.sig. @break
 * @tab            FLIB_FD_BUFSIZE @tab Set the buffer size to be used by FILE's
 *                                      opened from this fd. @break
 * @tab            FLIB_FD_CLONE   @tab Enable(1) or Disable(0) cloning for the
 *                                      fd. @break
 *
 * @output         ... The various arguments in the argument list.
 *
 * @return
 * @tab            EFS_SUCCESS             @tab Is returned upon successful
 *                                              execution. @break
 * @tab            FLIB_EINVALID_PARAMETER
 *                 or EINVAL               @tab Will be returned if an invalid
 *                                              parameter is used. @break
 * @tab            FLIB_EINVALID_FD or
 *                 EBADF                   @tab Will be returned if a bad file
 *                                              descriptor is used. @break
 * @tab                                    @tab Not errno (FLIB_E*, DDB_E*,...)
 *                                              codes can be translated to errno
 *                                              codes using
 *                                              @link "" "efs_status_to_errno".
 *                                              @break
 * @tab                                    @tab Global variable
 *                                              @link "errno.h" "errno" is not
 *                                              updated by this function. @break
 *
 * @seealso        efs_examine_fd
 * @seealso        fcntl
 * @seealso        default_restrictions_calling_functions
 *
 * @example
 * // fd is a valid file descriptor //
 * // Set append mode. Set buffering mode to 1 and buffer size to 80 for fd //
 * efs_config_fd(fd, FLIB_FD_APPEND, 1, FLIB_FD_BUFMODE, 1,
 *               FLIB_FD_BUFSIZE 80, 0);
 */
EfsStatus efs_config_fd(int fd, ...);


/**
 * @function       efs_donate_fd
 *
 * @include        "efs.h"
 *
 * @declaration    EfsStatus efs_donate_fd(int fd, PROCESS to)
 *
 * @brief          Donate a file descriptor to another process.
 *
 * @long           efs_donate_fd() donates the specified file descriptor using
 *                 the DonateFd (efs.sig) signal to the specified
 *                 receiver process. The ownership of the file handle is
 *                 changed in the volume manager. The specified file descriptor
 *                 becomes invalid in the donor and the receiver will receive
 *                 another unrelated file descriptor.
 *
 * @error OSE_EFS_FLIB_EUNRESOLVED_INET_FUNCTION Unresolved INET function.
 *                 This error will be called with fatal bit set.
 *
 * @input          fd  File descriptor to donate to the receiver process.
 * @input          to  PID of the receiving process. This process must execute
 *                 on the same node.
 *
 * @return
 * @tab            EFS_SUCCESS             @tab Is returned upon successful
 *                                              execution. @break
 * @tab            FLIB_ESOCKET_ERROR      @tab will be returned in case of a
 *                                              socket error. @break
 * @tab            FLIB_ESTALE_FD
 *                 or ESTALE               @tab will be returned to indicate a
 *                                              lost contact with the VM or
 *                                              that the file descriptor was
 *                                              stale since before. @break
 * @tab            FLIB_EINVALID_FD or
 *                 EBADF                   @tab will be returned if an invalid
 *                                              fd is used. @break
 * @tab            FLIB_EINVALID_PARAMETER
 *                 or EINVAL               @tab will be returned if the "to"
 *                                              process is invalid. @break
 * @tab                                    @tab Not errno (FLIB_E*, DDB_E*,...)
 *                                              codes can be translated to errno
 *                                              codes using
 *                                              @link "" "efs_status_to_errno".
 *                                              @break
 * @tab                                    @tab Global variable
 *                                              @link "errno.h" "errno" is not
 *                                              updated by this function. @break
 * @seealso        efs_receive_fd
 * @seealso        default_restrictions_calling_functions
 *
 * @example
 *  // executing in process A //
 *  int fd, status;
 *  fd = open("ose.txt", O_RDWR, S_IRGRP);
 *  // check if valid fd //
 *  extern OSENTRYPOINT entry;
 *  PROCESS B_ = create_process(OS_BG_PROC, "B", entry, 1000, (OSPRIORITY) 0,
 *                  (OSTIME) 0, (PROCESS) 0, NULL, (OSVECTOR) 0, (OSUSER) 0);
 * start(B_);
 * // Donating file descriptor to process B //
 * status = efs_donate_fd(fd, B_);
 * // check status //
 * stop(current_process());
 */
EfsStatus efs_donate_fd(int fd, PROCESS to);


/**
 * @function       efs_receive_fd
 *
 * @include        "efs.h"
 *
 * @declaration    int efs_receive_fd(PROCESS from)
 *
 * @brief          Receive a file descriptor donated by some other process.
 *
 * @long           efs_receive_fd() will wait for the
 *                 DonateFd (efs.sig) signal to
 *                 be sent from the 'from' process or from any process if
 *                 'from' equals 0. efs_receive_fd() should be used together
 *                 with efs_donate_fd() which sends the DonateFd (efs.sig)
 *                 signal to the receiver. If everything is OK the received fd
 *                 will be returned from efs_receive_fd(). If the "from"
 *                 process however would terminate, -1 will be returned.
 *
 * @input          from     PID of the process which is sending the
 *                          DonateFd (efs.sig) signal.
 *
 * @return         On success a new file descriptor is returned, on failure
 *                 (the from process terminated before sending the
 *                 DonateFd (efs.sig) signal) -1 is returned.
 *
 * @seealso        efs_donate_fd
 * @seealso        default_restrictions_calling_functions
 */
int efs_receive_fd(PROCESS from);

/**
 * @function       efs_fd_supports_sig
 *
 * @include        "efs.h"
 *
 * @declaration    EfsStatus
 * efs_fd_supports_sig(int fd, SIGSELECT signo);
 *
 * @brief          Find out if the volume manager handling a file descriptor
 *                 supports a certain signal.
 *
 * @long           efs_fd_supports_sig() uses the FmInterface (fm.sig)
 *                 signal to the volume manager to check if signal number
 *                 'signo' is supported by the volume manager. If the signal
 *                 number is supported efs_fd_supports_sig() returns
 *                 EFS_SUCCESS.
 *
 * @input          fd      File descriptor, used to extract the volume manager
 *                         PID.
 * @input          signo   The signal number to be check for.
 *
 * @return         If the interface request fails a VM
 *                 implementation specific reply status will be returned.
 * @tab            EFS_SUCCESS            @tab The volume managers supports the
 *                                             specified signal. @break
 * @tab            FM_EOPER_NOT_SUPPORTED
 *                 or ENOTSUP             @tab if signo was not supported by the
 *                                             VM. @break
 * @tab            FM_EDEVICE_LOST
 *                 or ENXIO               @tab will be returned if the contact
 *                                             with the volume manager is lost.
 *                                             @break
 * @tab            FLIB_EINVALID_FD
 *                 or EBADF               @tab will be returned if an invalid
 *                                             file descriptor is given. @break
 * @tab                                   @tab Not errno (FLIB_E*, FM_E*,...)
 *                                             codes can be translated to errno
 *                                             codes using
 *                                             @link "" "efs_status_to_errno".
 *                                             @break
 * @tab                                   @tab Global variable
 *                                             @link "errno.h" "errno" is not
 *                                             updated by this function. @break
 *
 *
 * @seealso        default_restrictions_calling_functions
 *
 * @example
 * EfsStatus status =
 *       efs_fd_supports_sig(fileno(stdin), FM_SUBSCRIBE_EVENT_REQUEST);
 * if(status == EFS_SUCCESS)
 * {
 *   // signal is supported //
 * }
 *  else // signal is not supported //
 *
 */
EfsStatus efs_fd_supports_sig(int fd, SIGSELECT signo);


/**
 * @function       ose_request_event
 *
 * @include        "efs.h"
 *
 * @declaration    FmHandle
 * ose_request_event(int sd, union SIGNAL **sig, S32 extra);
 *
 * @brief          Request an FmEvent (fm.sig) signal to be sent to the
 *                 caller when a specific event occurs on the file/socket
 *                 associated with the descriptor.
 *
 * @input          fd       The file descriptor of the object the event shall
 *                          be triggered on.
 * @input          sig      A preallocated signal of FmEvent (fm.sig)
 *                          structure.  The signal is sent to the calling
 *                          process when the event occurs. The signal number
 *                          specify what event to trigger on. Information about
 *                          the signal and event types are found in
 *                          ose_spi/fm.sig.
 * @input          extra    The extra parameter is specific to each event type.
 *
 * @return         The handle associated with the fd.
 *
 * @seealso FmEvent
 * @seealso default_restrictions_calling_functions
 */
FmHandle ose_request_event(int fd, union SIGNAL **sig, S32 extra);


/* Functions operating on EFS per process state and child processes. */
/**
 * @function       efs_clone
 *
 * @include        "efs.h"
 *
 * @declaration    EfsStatus
 * efs_clone(PROCESS pid);
 *
 * @brief          Copy all file descriptors, umask and current working
 *                 directory to a target process.
 * @long           This function is intended for initializing the file
 *                 system environment for a newly created process. All
 *                 file descriptors, umask and current working directory
 *                 are cloned to the target process, and the current process
 *                 is set as the parent of the target process. This function
 *                 shall be called before the target process is started.
 * @long           The function should not be used with a target that is
 *                 another process in the same block when operating in the
 *                 shared execution model. Such processes share all file
 *                 descriptors etc with each other and there is no need to
 *                 clone. The function has no effect for such a target. 
 * @long           The targeted process will receive exactly the same file
 *                 descriptors as the parent, which means that file descriptors
 *                 0, 1, 2, ... in the parent will be file descriptors
 *                 0, 1, 2, ... in the targeted process in the same order.
 * @long           The functions @link "" "efs_clone" and @link ""
 *                 "efs_clone_array" are equivalent except that the latter
 *                 allows a selection of which file descriptors to clone.
 *
 * @input   pid    The identity of the target process.
 *
 * @return
 * @tab             EFS_SUCCESS             @tab Upon success. @break
 * @tab             FLIB_ESTALE_FD
 *                  or ESTALE               @tab returned when the contact is
 *                                               lost, or was lost, with the
 *                                               volume manager i.e. a stale
 *                                               file descriptor. @break
 * @tab             FLIB_EINVALID_FD or
 *                  EBADF                   @tab returned when the handle is
 *                                               invalid. @break
 * @tab             FLIB_EINVALID_PARAMETER
 *                  or EINVAL               @tab Is returned when the process
 *                                               name is invalid. @break
 * @tab             FLIB_ESOCKET_ERROR      @tab Is returned in the case of a
 *                                               socket error. @break
 * @tab                                     @tab Not errno (FLIB_E*, DDB_E*,...)
 *                                               codes can be translated to
 *                                               errno codes using
 *                                               @link "" "efs_status_to_errno".
 *                                               @break
 * @tab                                     @tab Global variable
 *                                               @link "errno.h" "errno" is not
 *                                               updated by this function.
 * @break
 *
 * @error           OSE_EFS_FLIB_EINVALID_CLONE  If the destination process is
 *                                               not a prioritized or
 *                                               background process, or if the
 *                                               destination process is not
 *                                               stopped.
 *
 * @restriction    The target process must be a background or prioritized
 *                 process. Only one of the functions @link "" "efs_clone"()
 *                 or @link "" "efs_clone_array"() can be used per target
 *                 process and it shall only be called once per target process.
 *
 *
 * @seealso efs_clone_array
 * @seealso efs_clone_env
 * @seealso default_restrictions_calling_functions
 *
 * @example
 *  // Copy all file descriptors to process new_process_pid.
 *  efs_clone(new_process_pid);
 */
EfsStatus efs_clone(PROCESS pid);


/**
 * @function       efs_clone_array
 *
 * @include        "efs.h"
 *
 * @declaration    EfsStatus
 * efs_clone_array(PROCESS pid, const int *fd_array);
 *
 * @brief          Copy selected file descriptors, umask and current working
 *                 directory to a target process.
 * @long           This function is intended for initializing the file
 *                 system environment for a newly created process. Selected
 *                 file descriptors, umask and current working directory
 *                 are cloned to the target process, and the current process
 *                 is set as the parent of the target process. This function
 *                 shall be called before the target process is started.
 * @long           The function should not be used with a target that is
 *                 another process in the same block when operating in the
 *                 shared execution model. Such processes share all file
 *                 descriptors etc with each other and there is no need to
 *                 clone. The function has no effect for such a target. 
 * @long           The functions @link "" "efs_clone" and @link ""
 *                 "efs_clone_array" are equivalent except that the latter
 *                 allows a selection of which file descriptors to clone.
 *
 * @input          pid        The PID of the target process.
 * @input          fd_array   An array of file descriptors. If set to NULL,
 *                            all file descriptors are cloned as by
 *                            @link "" "efs_clone".
 *                            The fd_array has the following syntax:
 * @break
 *                            fd_array[0] = the number of file descriptors in
 *                            the array.
 * @break                     fd_array[1...n] = the file descriptors.
 * @break                     The cloned file descriptors will not be assigned
 *                            the same descriptor numbers in the target process.
 *                            They will be assigned to ascending fd numbers
 *                            starting with fd 0. For exammple:
 * @break                     fd_array = { 4, 12, 1, 2, 18 } results in
 * @break                     4 => 4 fd's are cloned.
 * @break                     Fd 12 is cloned to fd 0 in target process.
 * @break                     Fd 1  is cloned to fd 1 in target process.
 * @break                     Fd 2  is cloned to fd 2 in target process.
 * @break                     Fd 18 is cloned to fd 3 in target process.
 * @break                     Note that fd's 0, 1 and 2 are reserved for use
 *                            as STDIN, STDOUT and STDERR. A process that
 *                            uses stdio should have those fd's set up for
 *                            that purpose and avoid using those fd's for
 *                            other purposes.
 * @break                     The example above will assign fd 12 from the
 *                            current process as STDIN in the target, copy
 *                            STDOUT and STDERR to the corresponding fd's,
 *                            and copy fd 18 to fd 3 in the target to be used
 *                            for an implementation defined purpose.
 *
 * @return
 * @tab             EFS_SUCCESS             @tab Upon success. @break
 * @tab             FLIB_ESTALE_FD or
 *                  ESTALE                  @tab returned when the contact is
 *                                               lost, or was lost, with the
 *                                               volume manager i.e. a stale
 *                                               file descriptor. @break
 * @tab             FLIB_EINVALID_FD or
 *                  EBADF                   @tab returned when the handle is
 *                                               invalid. @break
 * @tab             FLIB_EINVALID_PARAMETER
 *                  or EINVAL               @tab Is returned when the process
 *                                               name is invalid. @break
 * @tab             FLIB_ESOCKET_ERROR      @tab Is returned in the case of a
 *                                               socket error. @break
 * @tab                                     @tab Not errno (FLIB_E*, DDB_E*,...)
 *                                               codes can be translated to
 *                                               errno codes using
 *                                               @link "" "efs_status_to_errno".
 *                                               @break
 * @tab                                     @tab Global variable
 *                                               @link "errno.h" "errno" is not
 *                                               updated by this function.
 * @break
 *
 * @error           OSE_EFS_FLIB_EINVALID_CLONE  If the destination process is
 *                                               not an prioritized or
 *                                               background process, or if the
 *                                               destination process is not
 *                                               stopped.
 *
 * @restriction    The target process must be a background or prioritized
 *                 process. Only one of the functions @link "" "efs_clone"()
 *                 or @link "" "efs_clone_array"() can be used per target
 *                 process and it shall only be called once per target process.
 *
 * @seealso efs_clone
 * @seealso efs_clone_env
 * @seealso default_restrictions_calling_functions
 *
 * @example
 *  EfsStatus status;
 *  int fds[] = { 3, 0, 1, 2 };
 *  // Clone stdio to the child process.
 *  status = efs_clone_array(child, fds);
 */
EfsStatus efs_clone_array(PROCESS pid, const int *fd_array);


/**
 * @function       efs_clone_env
 *
 * @include        "efs.h"
 *
 * @declaration void
 * efs_clone_env(PROCESS from, PROCESS to);
 *
 * @brief          Clone string environment variables from one process to
 *                 another executing on the same processor.
 *
 * @long           The command also adds the parent's PID, to an environment
 *                 variable called PARENT, in the clone process.
 *
 * @input          from    The process where the environment variables are
 *                         supposed to be cloned from.
 * @input          to      The process which is supposed to get the cloned
 *                         environment variables.
 *
 * @return         void
 *
 * @seealso efs_clone
 * @seealso efs_clone_array
 * @seealso default_restrictions_calling_functions
 */
void    efs_clone_env(PROCESS from, PROCESS to);


/**
 * @function       efs_wait_child
 *
 * @include        "efs.h"
 *
 * @declaration Boolean
 * efs_wait_child(PROCESS child, int *status, int *error);
 *
 * @brief          Waits for a process to terminate using the exit/_exit system
 *                 call.
 *
 * @long           The efs_wait_child function, allow a parent process to
 *                 monitor the termination of a child process. The child
 *                 process shall exit using the exit/_exit system call.
 *
 * @input          child     The child process PID.
 * @input          status    A status value pointer.
 * @input          error     A error value pointer.
 *
 * @return         If the child process terminated using exit/_exit, the
 *                 function returns the value of True. The status and error
 *                 buffers are filled in with the child exit status and child
 *                 errno. If the child does not terminate using exit/_exit,
 *                 the function returns False, and the status and error buffers
 *                 are left intact.
 *
 * @seealso exit
 * @seealso _exit
 * @seealso efs_clone
 * @seealso efs_clone_env
 * @seealso default_restrictions_calling_functions
 *
 * @example
 *  OS_PROCESS(child)
 *  {
 *    int status;
 *    //Do something...and initialize status.
 *    exit(status);
 *  }
 *
 *  OS_PROCESS(parent)
 *  {
 *     PROCESS child_pid;
 *     int ch_status, ch_errno;
 *     child_pid = create_process(OS_BG_PROC, "child", child, 512, 0, 0, 0,
 *                                NULL, 0, 0);
 *     start(child_pid);
 *     if(efs_wait_child(child, &ch_status, &ch_errno))
 *     {
 *       // The child exited in a controlled way using exit/_exit.
 *     }
 *     else
 *     {
 *       // The child terminated without calling exit/_exit.
 *     }
 *     // Do other processing...
 *  }
 */
Boolean efs_wait_child(PROCESS child, int *status, int *error);


/**
 * @function       efs_scan_env
 *
 * @include        "efs.h"
 *
 * @brief          Traverse the environment for the specified process
 *                 and calls the callback function for each variable.
 *
 * @long           For each variable in the environment the callback function
 *                 is called with the name and the value of the environment
 *                 variable together with a user specified parameter.
 *
 * @input          pid    PID for process whose environment to traverse.
 * @input          apply  Callback function to call for each variable.
 * @input          ptr    The first parameter to the callback function.
 *
 * @return         Nothing.
 *
 * @seealso        default_restrictions_calling_functions
 *
 * @example
 *
 * static void
 * print_env(void *pid, const char *name, const char *value)
 * {
 *    printf("%s=%s\n", name, value);
 * }
 *
 * static void
 * print_env_for(PROCESS pid)
 * {
 *    struct OS_pcb *pcb = get_pcb(pid);
 *    if (pcb != NULL)
 *    {
 *       const char *name = &pcb->strings[pcb->name];
 *       printf("Environment variables for %s (0x%lx):\n", name, pid);
 *       efs_scan_env(pid, print_env, (void*)pid);
 *       free_buf((union SIGNAL **)&pcb);
 *    }
 * }
 */
void efs_scan_env(PROCESS pid,
                  void (*apply)(void *, const char *, const char *),
                  void *ptr);


/**
 * @function       efs_copy_param
 *
 * @include        "efs.h"
 *
 * @declaration void
 * efs_copy_param(char *dest, const char *src, size_t size);
 *
 * @brief          Copies the first parameter of the parameter list params to
 *                 the buffer dest.
 *
 * @long           Size specifies the size of the dest buffer. The 'src'
 *                 string is a parameter list of
 *                 "<param>,<param2>,...,<paramN>". If the size of the 'dest'
 *                 buffer is not big enough to hold the entire parameter, the
 *                 parameter is truncated in dest and terminated with '\0'.
 *
 * @input          dest     The destination buffer.
 * @input          src      The parameters list, of syntax
 *                          "<param>,<param2>,...,<paramN>".
 * @input          size     The size of dest.
 *
 * @return         void
 *
 * @seealso        default_restrictions_calling_functions
 */
void efs_copy_param(char *dest, const char *src, size_t size);


/**
 * @function       efs_find_keyword
 *
 * @include        "efs.h"
 *
 * @declaration    const char *efs_find_keyword(const char *str,
 *                                              const char *keyword);
 *
 * @brief          Return a pointer to the first occurrence of the keyword in
 *                 the string.
 *
 * @long           efs_find_keyword() searches for the specified keyword in the
 *                 string and returns a pointer to the first occurrence or NULL
 *                 if not found. Keywords are separated with commas (',') and
 *                 white space is not ignored.
 *
 * @input          str      The string in which to search.
 * @input          keyword  The keyword to search for.
 *
 * @return         Pointer to the first occurrence of the keyword or NULL if not
 *                 found.
 *
 * @seealso        default_restrictions_calling_functions
 *
 * @example
 *
 * Boolean global = efs_find_keyword(parameters, "global") != NULL;
 * if (global)
 * {
 *    // The volume was mounted with global visibility... //
 * }
 */
const char *efs_find_keyword(const char *str, const char *keyword);


/**
 * @function       efs_get_string
 *
 * @include        "efs.h"
 *
 * @brief          Retrieve the string value of the keyword.
 *
 * @long           efs_get_string() searches for the keyword in the parameter
 *                 list and returns True if found together with the keyword
 *                 value (as a string) in the output buffer.
 *
 * @input          params   The parameter list.
 * @input          keyword  The keyword to search for.
 * @input          size     The size of the output buffer for the string.
 *
 * @output         string   The string value of the keyword.
 *
 * @return         True if the keyword was found and a string returned
 *                 otherwise False.
 *
 * @seealso        default_restrictions_calling_functions
 *
 * @example
 *
 * char root[80];
 * Boolean found = efs_get_string(parameters, "root", root, sizeof root);
 * if (found)
 * {
 *    printf("The hostfm volume is mounted on %s\n", root);
 * }
 */
Boolean efs_get_string(const char *params, const char *keyword, char *string,
                       size_t size);


/**
 * @function       efs_get_number
 *
 * @include        "efs.h"
 *
 * @declaration    Boolean efs_get_number(const char *params,
 *                                        const char *keyword, S32 *number);
 *
 * @brief          Retrieve the numerical value of the keyword.
 *
 * @long           efs_get_number() searches for the keyword in the parameter
 *                 list and returns True if found together with the keyword
 *                 value (as a number).
 *
 * @input          params   The parameter list.
 * @input          keyword  The keyword to search for.
 *
 * @output         number   The numerical value of the keyword.
 *
 * @return         True if the keyword was found and a number returned
 *                 otherwise False.
 *
 * @seealso        default_restrictions_calling_functions
 *
 * @example
 *
 * S32 dirbuffers;
 * Boolean found = efs_get_number(parameters, "dirbuffers", &dirbuffers);
 * if (found)
 * {
 *    printf("The volume will have %lu directory buffers\n", dirbuffers);
 * }
 */
Boolean efs_get_number(const char *params, const char *keyword, S32 *number);


/**
 * @function       efs_str_to_long
 *
 * @include        "efs.h"
 *
 * @declaration    long efs_str_to_long(const char *str, const char **err_ptr);
 *
 * @brief          Return the numerical value of a string.
 *
 * @long           efs_str_to_long() converts the string into the corresponding
 *                 number and returns that number. If the conversion ends
 *                 before the string is terminated and err_ptr is not NULL,
 *                 err_ptr will point to the first character in string that
 *                 could not be converted. The returned number will in this
 *                 case be based on all characters up to but excluding
 *                 character pointed to by err_ptr.
 *
 * @input          str      String to convert to number.
 * @input          err_ptr  Pointer to a char pointer to be updated with the
 *                          char pointer that ended the conversion. Can be
 *                          specified as NULL.
 *
 * @return         Numerical value of string.
 *
 * @seealso        atoi
 * @seealso        atol
 * @seealso        strtol
 * @seealso        strtoul
 * @seealso        default_restrictions_calling_functions
 *
 * @example
 * Boolean efs_get_number(const char *params, const char *keyword, S32 *number)
 * {
 *    const char *found = efs_find_keyword(params, keyword);
 *    if (found != NULL)
 *    {
 *       found += strlen(keyword);
 *       if (*found == '=')
 *       {
 *          char *endptr = NULL;
 *          found++;
 *          *number = efs_str_to_long(found, (const char **)&endptr);
 *          return endptr == NULL || *endptr == ',';
 *       }
 *    }
 *    return False;
 * }
 */
long efs_str_to_long(const char *str, const char **err_ptr);


/**
 * @function       efs_split_args
 *
 * @include        "efs.h"
 *
 * @declaration    void efs_split_args(char *arg_line, int *argc,
 *                                     char *** argv);
 *
 * @brief          Split the argument line into an argument vector.
 *
 * @long           efs_split_args() splits the argument line into separate
 *                 items (destroying the input) and puts them into an argument
 *                 vector followed by a NULL pointer. The number of elements in
 *                 the vector is also returned. The argument vector is
 *                 allocated from the heap and must be freed by the caller
 *                 using free().
 *
 * @input          arg_line  The argument line to split up. This parameter is
 *                           destroyed.
 *
 * @output         argc      The number of arguments (valid elements in the
 *                           argv array).
 * @output         argv      Pointer to a NULL terminated array of pointers to
 *                           strings. Allocated by this function as a
 *                           private heap buffer.
 *
 * @return         Nothing.
 *
 * @seealso         default_restrictions_calling_functions
 *
 * @example
 *
 * char buffer[80];
 * if (fgets(buffer, sizeof buffer, stdin) != NULL)
 * {
 *    int i;
 *    int argc;
 *    char **argv;
 *    efs_split_args(buffer, &argc, &argv);
 *    for (i = 0; i < argc; i++)
 *    {
 *       printf("%d : %s\n", i, argv[i]);
 *    }
 *    assert(argv[argc] == NULL);
 *    heap_free_private(argv);
 * }
 */
void efs_split_args(char *arg_line, int *argc, char *** argv);


/**
 *
 * @function efs_str_to_longlong
 *
 * @include        "efs.h"
 *
 * @declaration long long int
 * efs_str_to_longlong(const char *str, const char **err_ptr);
 *
 * @brief Return the numerical value of a string.
 *
 * @long If the conversion ends before the string is terminated and err_ptr is
 *       not NULL, err_ptr will point to the first character in string that
 *       could not be converted. The returned number will in this case be based
 *       on all characters up to but excluding character pointed to by err_ptr.
 *
 * @input str      String to convert to number.
 * @input err_ptr  Pointer to a char pointer to be updated with the char
 *                 pointer that ended the conversion. Can be specified as NULL.
 *
 * @return Numerical value of string.
 *
 * @seealso atoi
 * @seealso atol
 * @seealso strtol
 * @seealso strtoul.
 * @seealso efs_string_to_long
 * @seealso default_restrictions_calling_functions
 */
long long int efs_str_to_longlong(const char *str, const char **err_ptr);



/**
 * @function       efs_status_to_errno
 *
 * @include        "efs.h"
 *
 * @declaration    int efs_status_to_errno(EfsStatus status);
 *
 * @brief          Convert the EfsStatus code into the closest matching errno
 *                 code (from errno.h).
 *
 * @long           The conversion can not be reversed. Can handle DDB_E* (device
 *                 driver), DDC_E* (character device driver), FLIB_E*,
 *                 FSS_E* (file system server), FM_E* status codes.
 *
 * @input          status   The EFS status or error code.
 *
 * @return         The closest matching errno code.
 *
 * @seealso        default_restrictions_calling_functions
 *
 * @example
 * EfsStatus status = efs_mount("/ram", "extfat", "ramdisk", "unit=0");
 * int err = efs_status_to_errno(status);
 * if (status == EFS_SUCCESS)
 * {
 *    // Failed to mount ramdisk. //
 *    eprintf("Failed to mount ramdisk, status %lu, errno %d\n", status, err);
 * }
 */
int efs_status_to_errno(EfsStatus status);


/**
 * @function       ose_hunt_w_tmo
 *
 * @include        "efs.h"
 *
 * @declaration    PROCESS ose_hunt_w_tmo(const char *hunt_path, OSUSER user,
 *                                        OSTIME tmo);
 *
 * @brief          Hunt for the specified process and returns the PID if found
 *                 before the timeout.
 *
 * @long           If the time expires the hunt is canceled and thus
 *                 no hunt signal will be delivered even if the specified
 *                 process is created or found at a later time.
 *
 * @input          hunt_path   Hunt path for the process to hunt for.
 * @input          user        Must be 0.
 * @input          tmo         Timeout for the hunt to be canceled.
 *
 * @return         PID of the hunted process or 0 if not found.
 *
 * @seealso        default_restrictions_calling_functions
 *
 * @example
 * PROCESS pid = ose_hunt_w_tmo("ose_fss", 0, 1000);
 * if (pid != 0)
 * {
 *    // Do something. //
 * }
 * // Else could not find 'ose_fss'. //
 */
PROCESS     ose_hunt_w_tmo(const char *hunt_path, OSUSER user, OSTIME tmo);


/**
 * @function       ose_receive_one
 *
 * @include        "efs.h"
 *
 * @declaration    void * ose_receive_one(SIGSELECT signo, PROCESS from_pid,
 *                                        OSTIME tmo);
 *
 * @brief          Return the selected signal sent from the specified process.
 *
 * @long           If the specified process terminates before sending the
 *                 signal this will be detected and NIL will be returned.
 *
 * @long           ose_receive_one() uses an algorithm with lazy attach to the
 *                 sender. If the timeout expires before the signal has been
 *                 received ose_receive_one() will attach a signal to the
 *                 sender. If the proper signal is returned the attached
 *                 signal will be detached.
 *
 * @input          signo      Signal number of the signal to receive.
 * @input          from_pid   PID of the sending process.
 * @input          tmo        Timeout when to attach to the sender.
 *
 * @return         Pointer to the received signal or NIL.
 *
 * @seealso        default_restrictions_calling_functions
 *
 * @example
 * union SIGNAL *sig = alloc(sizeof(SIGSELECT), REQUEST);
 * send(&sig, remote_);
 * sig = ose_receive_one(REPLY, remote_, 1000);
 * if (sig != NIL)
 * {
 *    // Do something with signal. //
 * }
 * // Else sender terminated before sending reply. //
 */
void *      ose_receive_one(SIGSELECT signo, PROCESS from_pid, OSTIME tmo);


/**
 * @function       ose_match_processes
 *
 * @include        "efs.h"
 *
 * @declaration    PROCESS * ose_match_processes(char *expr, EfsBitmask flags,
 *                                               U32 *num_pids);
 *
 * @brief          Return PIDs of processes that match expression.
 *
 * @long           ose_match_process() parses expr and finds all
 *                 processes/blocks/domains (according to flags) that match.
 *                 The PIDs of the matching processes are returned with a
 *                 pointer to an array of PROCESS. The U32 pointed to by
 *                 num_pids is updated with the number of matching processes
 *                 (i.e. the number of elements in the returned array). The
 *                 returned array must be freed by the caller using the HEAP
 *                 call heap_free_private().
 *
 *
 * @input          expr      Expression used to match processes.
 * @input          flags     Options to control matching, see
 *                           @link "" "EfsMatchBitmap".
 *
 * @output         num_pids  Number of PIDs in returned array.
 *
 * @return         Pointer to an array with PIDs.
 *
 * @seealso        default_restrictions_calling_functions
 *
 * @example
 * EfsBitmask flags = MATCH_PROCESSES;
 * U32 num_pids;
 * PROCESS *pids = ose_match_processes("ose_kalle", flags, &num_pids);
 * if (pids != NULL)
 * {
 *    unsigned i;
 *    for (i = 0; i < num_pids; i++)
 *    {
 *       do_something(pids[i]);
 *    }
 *    heap_free_private(pids);
 * }
 * // Else no processes matched. //
 */
PROCESS *ose_match_processes(char *expr, EfsBitmask flags, U32 *num_pids);


/**
 * @function efs_get_fds
 *
 * @include        "efs.h"
 *
 * @declaration    int efs_get_fds(PROCESS pid);
 *
 * @brief          Return the number of open file descriptors
 *                 in the specified process.
 *
 * @input          pid     PID of the process who opened the files.
 *
 * @return
 * @tab            >0 @tab The number of open file descriptors.
 *                         @break
 * @tab             0 @tab The process has closed all its files.
 *                         @break
 * @tab            -1 @tab The process has not yet performed any operation that
 *                         access file descriptors (or the given PID is
 *                         invalid). @break
 * @tab            -2 @tab The data structures in remote CRT is incompatible.
 *                         @break
 * @seealso        default_restrictions_calling_functions
 */
int efs_get_fds(PROCESS pid);


/**
 * @type EfsFdInfoTag
 *
 * @include        "efs.h"
 *
 * @brief          The different properties of file descriptors.
 *
 * @long           The values of EfsFdInfoTag are used when examining
 *                 an open file using @link "" "efs_get_fd_info".
 *
 * @field FDI_SHMEM     Shared memory file descriptor.
 * @field FDI_SOCKET    The file descriptor is a socket descriptor.
 * @field FDI_APPEND    Writes are appended at the end of the file.
 * @field FDI_CLONE     The file descriptor is possible to clone.
 * @field FDI_SHARED    This file is shared among everyone in sh_context.
 *
 * @seealso efs_get_fd_info
 * @seealso efs_examine_fd
 */
enum EfsFdInfoTag
{
   FDI_SHMEM  = 0x01,
   FDI_SOCKET = 0x02,
   FDI_APPEND = 0x04,
   FDI_CLONE  = 0x08,
   FDI_SHARED = 0x10
};


/**
 * @type FdInfo
 *
 * @include        "efs.h"
 *
 * @declaration
 * struct FdInfo
 * {
 *     int      fd;
 *     PROCESS  vm_pid;
 *     FmHandle handle;
 *     U32      buffer_mode;
 *     U32      buffer_size;
 *     PROCESS  sh_owner;
 *     PROCESS  sh_context;
 *     U32      flags;
 * };
 *
 * @brief Info about an open file obtained by a call to
 *        @link "" "efs_get_fd_info".
 *
 * @field fd            The file descriptor.
 * @field vm_pid        The PID of the volume manager managing the
 *                      file referred by the file descriptor.
 * @field handle        The handle of the file.
 * @field buffer_mode   The file descriptor buffer mode
 *                      Compare FmBuffEnum in ose_spi/fm.sig.
 * @field buffer_size   The file descriptor buffer size (in bytes).
 * @field sh_owner      The PID of the owner of shared files.
 * @field sh_context    The context (a PID) in which files can be shared.
 * @field flags         A bitfield of @link "" "EfsFdInfoTag" bits.
 *
 * @seealso efs_get_fd_info
 * @seealso efs_examine_fd
 */
struct FdInfo
{
    int      fd;
    PROCESS  vm_pid;
    FmHandle handle;
    U32      buffer_mode;
    U32      buffer_size;
    PROCESS  sh_owner;
    PROCESS  sh_context;
    U32      flags;
};


/**
 * @function efs_get_fd_info
 *
 * @include        "efs.h"
 *
 * @declaration    int efs_get_fd_info(struct FdInfo *fdinfo, PROCESS pid,
 *                                     int fd);
 *
 * @brief          Examine properties of a file descriptor.
 *
 * @long           These properties exist only in the application.
 *
 * @input          fdinfo  A pointer to a @link "" "FdInfo" struct in which
 *                         the obtained info will be stored.
 * @input          pid     PID of the process who opened the file.
 * @input          fd      The file descriptor.
 *
 * @return
 * @tab             0 @tab Valid information is returned about the specified
 *                         (or next valid) open file descriptor.
 *                         @break
 * @tab            -1 @tab The given PID is invalid.
 *                         @break
 * @tab            -2 @tab The data structures in remote CRT is incompatible.
 *                         @break
 * @tab            -3 @tab The specified fd is larger than largest open fd.
 *                         @break
 *
 * @seealso efs_examine_fd
 * @seealso default_restrictions_calling_functions
 */
int efs_get_fd_info(struct FdInfo *fdinfo, PROCESS pid, int fd);


/**
 * @function efs_is_abspath
 *
 * @include        "efs.h"
 *
 * @declaration    int efs_is_abspath(const char *path);
 *
 * @brief          Check if a path name is an absolute path name.
 *
 * @long           This function is only needed for legacy purposes.
 *                 Systems that do not use the "link handler path name" feature
 *                 or use a POSIX compliant syntax for specifying a link handler
 *                 path name can forego the use of this function and instead
 *                 compare the first character of the path name to the slash
 *                 ('/') character.
 *
 * @long           Use case: The OSE feature "link handler path name" allows
 *                 direct access to volumes managed by an FSS on a remote node.
 *                 A link handler path name is by definition a special case of
 *                 an absolute path name. There is however a legacy syntax where
 *                 a link handler path name does not start with the slash ('/')
 *                 character, thereby deviating from the POSIX definition of an
 *                 absolute path name. This function can accurately distinguish
 *                 a relative path name from all forms of absolute path name,
 *                 including link handler path name.
 *
 * @input          path    Pointer to null terminated string representing the
 *                         path name to be examined.
 *
 * @return         0 if the path name is a relative path name.
 *                 1 if the path name is an absolute path name.
 *
 * @seealso        default_restrictions_calling_functions
 */
int efs_is_abspath(const char *path);


/**
 * @function efs_get_lnh_syntax
 *
 * @include        "efs.h"
 *
 * @declaration    void efs_get_lnh_syntax(char **lnh_path_prefix,
 *                                         char *lnh_path_separator);
 *
 * @brief          Get link handler path syntax
 *
 * @long           This function can be used to get the prefix and separator
 *                 for the link handler path name, if needed by the user code,
 *                 e.g. to construct direct remote pathnames to use with
 *                 the CRT functions.
 *                 Default both are null and remote pathnames can not be used.
 *                 Instead mount the remote file systems to access them.
 *
 * @long           If they are defined, file name paths in the CRT functions,
 *                 having this prefix, a link handler path and a separator,
 *                 are interpreted to automatically access the remote system
 *                 via the link handler and reach the file.
 *
 * @long           The system parameter krn/lnh_path=prefix:xxx separator:x
 *                 can be used to define the separator and prefix. The prefix
 *                 can be up to 15 characters (plus terminating null).
 *
 * @long           Example: "/lnh/systemA/systemB!/filename" can be used if
 *                 prefix is "/lnh/" and separator is !.
 *
 * @output         lnh_path_prefix  A pointer to a pointer to a string, which
 *                                  will be allocated in this function.
 *                                  The caller must free the memory with
 *                                  heap_free_private.
 *
 * @output         lnh_path_separator A pointer to a character.
 *
 * @return         Nothing.
 *
 * @seealso        default_restrictions_calling_functions
 */
void efs_get_lnh_syntax(char **lnh_path_prefix, char *lnh_path_separator);


/**
 * @type FmValidateVolMsgFunc
 *
 * @include        "efs.h"
 *
 * @declaration
 * typedef void (*FmValidateVolMsgFunc)(const struct FmValidateVolMsg *msg,
 *                                      void *arg);
 *
 * @brief Type for callback function used by
 *        @link "" "efs_validate_volume".
 *
 * @field msg  Received message.
 *
 * @field arg  Extra argument given by funcArg parameter in call to
 *             @link "" "efs_validate_volume".
 *
 * @seealso efs_validate_volume
 */
typedef void (*FmValidateVolMsgFunc)(const struct FmValidateVolMsg *msg,
                                     void *arg);


/**
 * @function efs_validate_volume
 *
 * @include        "efs.h"
 *
 * @declaration    EfsStatus
 *                 efs_validate_volume(const char *vol, Boolean repair,
 *                                     FmValidateVolMsgFunc msgFunc,
 *                                     void *funcArg,
 *                                     U32 *outFixedProblems,
 *                                     U32 *outRemainProblems);
 *
 * @brief          Validate a volume.
 *
 * @long           Validate a volume and look for media errors. Can optionally
 *                 report progress and findings using a callback function.
 *                 Can also attempt to repair found problems.
 *
 * @input          vol  Path to volume to validate.
 *
 * @input          repair  If True, attempt to repair problems found.
 *
 * @input          msgFunc  If not NULL: use function as callback function for
 *                          progress information.
 *
 * @input          funcArg  If msgFunc not NULL: Extra argument given in each
 *                          call to function specified by msgFunc parameter.
 *
 * @output         outFixedProblems  If not NULL: Pointer to where to store the
 *                          number of problems fixed.
 *
 * @output         outRemainProblems  If not NULL: Pointer to where to store
 *                          the number of problems remaining.
 *
 * @return
 * @tab            EFS_SUCCESS             @tab Upon success. @break
 * @tab            FM_EOPER_NOT_ALLOWED    @tab The operation was not allowed.
 * @break
 * @tab            FM_EOBJECT_NOT_FOUND    @tab Unknown volume. @break
 * @tab            FM_EVM_LOST             @tab Lost contact with the volume
 *                                              manager. @break
 *                 Other error status codes come from the volume manager or
 *                 lower level drivers.
 *                 See the status field description of ose_spi/fm.sig @link ""
 *                 "FmValidateVol".
 *
 * @seealso efs_validate_print_callback
 * @seealso default_restrictions_calling_functions
 */
EfsStatus
efs_validate_volume(const char *vol, Boolean repair,
                    FmValidateVolMsgFunc msgFunc, void *funcArg,
                    U32 *outFixedProblems, U32 *outRemainProblems);


/**
 * @function efs_wvalidate_volume
 *
 * @include        "efs.h"
 *
 * @declaration    EfsStatus
 *                 efs_wvalidate_volume(const wchar_t *vol, Boolean repair,
 *                                      FmValidateVolMsgFunc msgFunc,
 *                                      void *funcArg,
 *                                      U32 *outFixedProblems,
 *                                      U32 *outRemainProblems);
 *
 * @brief          Validate a volume.
 *
 * @long           Validate a volume and look for media errors. Can optionally
 *                 report progress and findings using a callback function.
 *                 Can also attempt to repair found problems.
 *
 * @long           This call is equivalent to @link "" "efs_validate_volume"
 *                 with the exception that the volume name is passed as a
 *                 wchar_t string.
 *
 * @input          vol  Path to volume to validate.
 *
 * @input          repair  If True, attempt to repair problems found.
 *
 * @input          msgFunc  If not NULL: use function as callback function for
 *                          progress information.
 *
 * @input          funcArg  If msgFunc not NULL: Extra argument given in each
 *                          call to function specified by msgFunc parameter.
 *
 * @output         outFixedProblems  If not NULL: Pointer to where to store the
 *                          number of problems fixed.
 *
 * @output         outRemainProblems  If not NULL: Pointer to where to store
 *                          the number of problems remaining.
 *
 * @return
 * @tab            EFS_SUCCESS             @tab Upon success. @break
 * @tab            FM_EOPER_NOT_ALLOWED    @tab The operation was not allowed.
 * @break
 * @tab            FM_EOBJECT_NOT_FOUND    @tab Unknown volume. @break
 * @tab            FM_EVM_LOST             @tab Lost contact with the volume
 *                                              manager. @break
 *                 Other error status codes come from the volume manager or
 *                 lower level drivers.
 *                 See the status field description of ose_spi/fm.sig @link ""
 *                 "FmValidateVol".
 *
 * @seealso efs_validate_print_callback
 * @seealso default_restrictions_calling_functions
 */
EfsStatus
efs_wvalidate_volume(const wchar_t *vol, Boolean repair,
                     FmValidateVolMsgFunc msgFunc, void *funcArg,
                     U32 *outFixedProblems, U32 *outRemainProblems);


/**
 * @type EfsValidateVerboseLevel
 *
 * @include "efs.h"
 *
 * @brief Verbosity levels in @link "" "efs_validate_print_callback".
 *
 * @long  This enum defines verbosity levels used by @link ""
 *        "efs_validate_print_callback"
 *        in order to print parts of the message received from @link ""
 *        "efs_validate_volume" to stdout.
 *
 * @field EFS_VALIDATE_VERBOSE_NO_OUTPUT     No message will be printed.
 * @field EFS_VALIDATE_VERBOSE_ERROR         Print only error message.
 * @field EFS_VALIDATE_VERBOSE_WARNING       Print warning and error message.
 * @field EFS_VALIDATE_VERBOSE_INFORMATIONAL Print all informational message.
 *
 * @seealso efs_validate_print_callback
 * @seealso efs_validate_volume
 */
enum EfsValidateVerboseLevel
{
   EFS_VALIDATE_VERBOSE_NO_OUTPUT,
   EFS_VALIDATE_VERBOSE_ERROR,
   EFS_VALIDATE_VERBOSE_WARNING,
   EFS_VALIDATE_VERBOSE_INFORMATIONAL
};


/**
 * @function efs_validate_print_callback
 *
 * @include        "efs.h"
 *
 * @declaration    void
 *                 efs_validate_print_callback(
 *                   const struct FmValidateVolMsg* sig,
 *                   void *verboseLevelPtr);
 *
 * @brief          Callback function for efs_validate_volume().
 *
 * @long           This function can be used as callback function in call to
 *                 efs_validate_volume(). It will write the messages received
 *                 from check disk to stdout.
 *
 * @input          sig     Pointer to signal struct containing message to
 *                         write.
 *
 * @input          verboseLevelPtr Pointer to U32 with a value found in
 *                                 enum EfsValidateVerboseLevel.
 *
 * @seealso        default_restrictions_calling_functions
 *
 * @example
 *
 *   U32 verboseLevel = EFS_VALIDATE_VERBOSE_INFORMATIONAL;
 *   U32 fixedProblems;
 *   U32 remainProblems;
 *   EfsStatus status;
 *
 *   status = efs_validate_volume("/myvol", True,
 *                                efs_validate_print_callback, &verboseLevel,
 *                                &fixedProblems, &remainProblems);
 */
void
efs_validate_print_callback(const struct FmValidateVolMsg* sig,
                            void *verboseLevelPtr);


/**
 * @function      efs_get_ose_shared_block
 *
 * @include       "efs.h"
 *
 * @brief         Get the block ID of the monolith's shared block.
 *
 * @long          The system parameter "core/shared_block=stack_size:4096"
 *                will create a shared mode environment in the monolith.
 *                All static C++ constructors will be executed from within
 *                that environment. If the monolith contains a symbol named
 *                ose_shared_main it will be auto started as a prioritized
 *                process within the shared environment.
 * @break
 *                The efs_get_ose_shared_block() returns the block ID which
 *                contains the monolith's shared environment.
 *
 * @return        Returns the block ID of the monolith's shared block,
 *                or 0 if there is no shared block created in the monolith.
 *
 * @restriction   This system call is only available when linked to the
 *                monolith.
 *
 * @seealso       default_restrictions_calling_functions
 */
PROCESS
efs_get_ose_shared_block(void);


#ifdef __cplusplus
}
#endif

#endif
