/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2011 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */


/**
 * @toc   Standard_C:
 *
 * @file  malloc.h
 *
 * @brief Heap memory allocation functions (malloc et al.).
 *
 * @long  The ISO/IEC C standard defines four storage allocation functions:
 *        malloc, free, realloc, and calloc.
 *
 * @long  Pointers returned by these functions are unique and
 *        discontiguous. Pointers remain valid from the point of
 *        allocation until deallocation. If an allocation fails, NULL is
 *        returned. After deallocation, usage of a pointer has
 *        undefined behavior.
 *
 * @long  The pointer value returned from a successful allocation is suitably
 *        aligned so that it may be assigned to a pointer to any type of
 *        object, including arrays, and used to access such an object.
 *        In practice this typically translates to the minimum alignment
 *        required by a pointer to the primitive type double.
 *
 * @long  The OSE heap manager implements a concept of shared and private
 *        pointers. A shared pointer can be used to resize or deallocate a
 *        memory block from any OSE process whereas a private pointer only
 *        can be operated on by the process that once allocated it.
 *        Private pointers are also automatically garbage collected if not
 *        explicitly deallocated before the owner process terminates.
 *
 * @long  When using the ISO C standard API in OSE it is
 *        sometimes difficult to tell what mode, shared or private,
 *        that is in effect. As a general rule, an allocated pointer
 *        is always private, unless @break
 *        1. The block execution model is shared, or @break
 *        2. The block environment variable
 *           HEAP_SHARED_MODE in ose_heap.h is set or @break
 *        3. The code which calls malloc etc., was compiled with
 *           MALLOC_SHARED set.
 *
 * @long  Developers of applications that share pointers between OSE
 *        processes should acertain that shared heap pointers really are
 *        pointers to shared mode heap buffers.
 *
 * @long  In addition to the shared block execution model and
 *        HEAP_SHARED_MODE, standard API allocations can forcibly be
 *        made shared when the calling module is compiled, by defining 
 *        @link "" "MALLOC_SHARED" before this header file is included,
 *        e.g. via #define MALLOC_SHARED or via the compiler constant
 *        -DMALLOC_SHARED.
 *        Similarly, @link "" "MALLOC_PRIVATE" can be defined to force
 *        private heap mode allocations in a block where shared heap mode
 *        is otherwise set.
 *        Both these macros will change the semantics of the allocation
 *        functions malloc et.al. to never return NULL. Instead an OSE
 *        error will be generated if there is unsufficient memory.
 *        To use these macros, access to system programming interface
 *        (SPI) is required when building the application.
 *
 * @example
 *   gcc -DMALLOC_FILELINE -DMALLOC_SHARED -c example.c -o example.o
 */

#ifndef MALLOC_FILELINE
/**
 * @macro MALLOC_FILELINE
 *
 * @brief Provides the heap with the location (file and line) of a call
 *        to the standard API.
 *
 * @long  The macro only enables the file and line information to be passed
 *        as parameters to the heap manager where the default action is to
 *        do nothing.
 *
 * @long  By installing the heap debug module, the file and line information
 *        will be stored together with each buffer and can later be retrieved
 *        using SPI interfaces, such as HeapBufferInfoRequest defined in heap.sig.
 */
#define MALLOC_FILELINE
#undef  MALLOC_FILELINE
#endif

#ifndef MALLOC_PRIVATE
/**
 * @macro MALLOC_PRIVATE
 *
 * @brief Forcibly enables private heap mode pointers when using standard
 *        allocation functions (malloc, et.al).
 * @long  Define this compile time constant before #include "malloc.h".
 *
 * @restriction
 *        The macro will impose a safe allocation environment where NULL
 *        is never returned. If there is insufficient memory to satisfy
 *        an allocation attempt an OSE error will be generated. This
 *        error cannot be ignored in an error handler.
 * @seealso malloc.h
 */
#define MALLOC_PRIVATE
#undef  MALLOC_PRIVATE
#endif

#ifndef MALLOC_SHARED
/**
 * @macro MALLOC_SHARED
 *
 * @brief Forcibly enables shared heap mode pointers when using standard
 *        allocation functions (malloc, et.al).
 * @long  Define this compile time constant before #include "malloc.h".
 *
 * @restriction
 *        The macro will impose a safe allocation environment where NULL
 *        is never returned. If there is insufficient memory to satisfy
 *        an allocation attempt an OSE error will be generated. This
 *        error cannot be ignored in an error handler.
 * @seealso malloc.h
 */
#define MALLOC_SHARED
#undef  MALLOC_SHARED
#endif


#ifndef _MALLOC_H
#define _MALLOC_H
#define OSE_MALLOC_H

#include <stddef.h>
#include "osarch.h"

#ifdef __cplusplus
extern "C" {
#endif

#if !defined softOSE
/* Non-soft (i.e. real) targets. */

/**
 * @function malloc
 *
 * @brief  Allocate dynamic memory for an object.
 *
 * @long   The function allocates a block of memory of the specified
 *         size. The block will be uninitialized.
 *
 * @long   If 0 bytes are requested, malloc() returns either NULL or
 *         a valid pointer value that can be used in a subsequent
 *         call to free().
 *
 * @input  size The size of the requested buffer.
 *
 * @return A pointer to the allocated buffer, or NULL if there is
 *         insufficient memory.
 *
 * @restriction
 *         Depending on the default mode of operation, allocated
 *         pointers may be private and cannot be shared with other
 *         execution contexts.
 *
 * @seealso calloc
 * @seealso realloc
 * @seealso free
 * @seealso default_restrictions_calling_functions
 */
extern void *malloc(size_t size);

/**
 * @function calloc
 *
 * @brief  Allocate dynamic memory for an array of objects.
 *
 * @long   The function allocates an array of elements of a specified
 *         specified size. The allocated memory is cleared with 0.
 *
 * @long   If the element count or element size is 0, then calloc()
 *         returns either NULL or a valid pointer value that can be
 *         used in a subsequent call to free().
 *
 * @input  nmemb  Number of array elements.
 * @input  size   Size of each element.
 *
 * @return A pointer to the allocated buffer, or NULL if there is
 *         insufficient memory.
 *
 * @restriction
 *         Depending on the default mode of operation, allocated
 *         pointers may be private and cannot be shared with other
 *         execution contexts.
 *
 * @seealso malloc
 * @seealso realloc
 * @seealso free
 * @seealso default_restrictions_calling_functions
 */
extern void *calloc(size_t nmemb, size_t size);

/**
 * @function realloc
 *
 * @brief  Change size of dynamically allocated memory.
 *
 * @long   The function changes the size of a dynamically allocated
 *         block of memory. The data will remain unchanged to the
 *         minimum of the old and the new size. Newly allocated
 *         memory is uninitialized.
 *
 * @input  ptr    Pointer to a block of memory to resize. If NULL, the
 *                call is equivalent to malloc(size).
 * @input  size   Size of the new block. If 0, the call is equivalent
 *                to free(ptr).
 *
 * @return A pointer to the resulting memory block, or NULL if there is
 *         insufficient memory. If NULL is returned, the old memory
 *         block remains unaffected. If a valid pointer is returned,
 *         it may differ from the old pointer if the data had to be
 *         moved to a new location.
 *
 * @restriction
 *         Depending on the default mode of operation, allocated
 *         pointers may be private and cannot be shared, e.g. resized,
 *         by other execution contexts.
 *
 * @seealso malloc
 * @seealso calloc
 * @seealso free
 * @seealso default_restrictions_calling_functions
 */
extern void *realloc(void *ptr, size_t size);

/**
 * @function free
 *
 * @brief  Free dynamically allocated memory.
 *
 * @long   The memory must have been allocated in a previous call to
 *         malloc(), realloc() or calloc(). Attempts to free already
 *         free blocks of memory is illegal.
 *
 * @input  ptr   A pointer to the block of memory to deallocate.
 *               If the pointer is NULL, no operation is performed.
 *
 * @restriction
 *         Depending on the default mode of operation, allocated
 *         pointers may be private and cannot be shared, e.g. freed,
 *         by other execution contexts.
 *
 * @seealso malloc
 * @seealso calloc
 * @seealso realloc
 * @seealso default_restrictions_calling_functions
 */
extern void  free(void *ptr);

#else
/* Soft kernels, uses zzz-prefix to avoid clash with native implementation. */
extern void *zzzmalloc(size_t);
extern void *zzzcalloc(size_t, size_t);
extern void *zzzrealloc(void *, size_t);
extern void  zzzfree(void *);
#endif

/* Prototypes with file and line information, available on all targets. */
extern void *zzmalloc(size_t, const char *, unsigned);
extern void *zzcalloc(size_t, size_t, const char *, unsigned);
extern void *zzrealloc(void *, size_t, const char *, unsigned);

#ifdef __cplusplus
}
#endif

/* Now the macro fest begins! */

#if defined MALLOC_PRIVATE

#include "heapapi.h"
#if defined MALLOC_FILELINE
#define malloc(s)    heap_alloc_private((s), __FILE__, __LINE__)
#define realloc(p,s) heap_realloc_private((p), (s), __FILE__, __LINE__)
#else
#define malloc(s)    heap_alloc_private((s), 0, 0)
#define realloc(p,s) heap_realloc_private((p), (s), 0, 0)
#endif
#define free(p)      heap_free_private((p))
/* Parameters to calloc() evaluated twice!!! */
#define calloc(s,n)  memset(malloc((s) * (n)), 0, (s) * (n))

#elif defined MALLOC_SHARED

#include "heapapi.h"
#if defined MALLOC_FILELINE
#define malloc(s)    heap_alloc_shared((s), __FILE__, __LINE__)
#define realloc(p,s) heap_realloc_shared((p), (s), __FILE__, __LINE__)
#else
#define malloc(s)    heap_alloc_shared((s), 0, 0)
#define realloc(p,s) heap_realloc_shared((p), (s), 0, 0)
#endif
#define free(p)      heap_free_shared((p))
/* Parameters to calloc() evaluated twice!!! */
#define calloc(s,n)  memset(malloc((s) * (n)), 0, (s) * (n))

#elif defined(MALLOC_FILELINE) && !defined(softOSE)

#define malloc(s)    zzmalloc((s), __FILE__, __LINE__)
#define calloc(n,s)  zzcalloc((n), (s), __FILE__, __LINE__)
#define realloc(p,s) zzrealloc((p), (s), __FILE__, __LINE__)
#ifdef softOSE
#define free(p)      zzzfree((p))
#else
/* free() not redefined since it does not need file and line. */
#endif

#elif defined softOSE

#define malloc(s)    zzzmalloc((s))
#define calloc(n,s)  zzzcalloc((n), (s))
#define realloc(p,s) zzzrealloc((p), (s))
#define free(p)      zzzfree((p))

#include "sfk/mkinline.h"

#ifdef ZZINLINE

#undef malloc
ZZMKINLINE_RN1(zzz, void *, malloc, size_t)
#undef calloc
ZZMKINLINE_RN2(zzz, void *, calloc, size_t, size_t)
#undef realloc
ZZMKINLINE_RN2(zzz, void *, realloc, void *, size_t)
#undef free
ZZMKINLINE_N1 (zzz, free, void *)

/* End #ifdef ZZINLINE */
#endif

/* End #ifdef softOSE */
#endif

/* End #ifndef _MALLOC_H */
#endif
