/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2011 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

/**
 * @toc OSE:
 * @file ose_spi/atomic.h
 *
 * @brief Atomic operations.
 *
 * @long Supported from both user and supervisor modes, except
 * for some ARM targets, where supervisor mode is required.
 *
 * @seealso ose_spi/arm/cpu_atomic.h
 * @seealso ose_spi/mips/cpu_atomic.h
 * @seealso ose_spi/powerpc/cpu_atomic.h
 * @seealso ose_spi/sfk/cpu_atomic.h
 */

#include <stdint.h>
#include "compiler.h"


#ifndef _ATOMIC_H
#define _ATOMIC_H
#define OSE_ATOMIC_H

#ifdef __cplusplus
extern "C" {
#endif


/**
 * @type atomic_t
 *
 * @brief The target type of atomic operations defined by atomic.h.
 * @field counter should never be referenced directly
 */
typedef struct
{
  volatile uint32_t counter;
} atomic_t;

/**
 * @macro ATOMIC_INIT
 *
 * @brief Macro to use as an initializer of a atomic_t variable.
 * @input value The wanted value.
 */
#define ATOMIC_INIT(value) {value}

/**
 * @macro atomic_get
 *
 * @brief Returns the value of the variable pointed to by var_ptr.
 * @long This function does not implement any memory barriers.
 *
 * @input ptr       points to the variable to read.
 *
 * @restriction For some ARM targets this function is only available when
 *              executing in supervisor state.
 *    @break
 *              Also available in interrupt and timer interrupt processes.
 *
 * @seealso default_restrictions_calling_functions
 */
#define atomic_get(ptr) (atomic_get_(ptr))

/**
 * @macro atomic_set
 *
 * @brief Atomically set the value of the variable pointed to by var_ptr.
 * @long This function does not implement any memory barriers.
 *
 * @input ptr       points to the variable to set.
 * @input value     the new variable value.
 *
 * @restriction For some ARM targets this function is only available when
 *              executing in supervisor state.
 *    @break
 *              Also available in interrupt and timer interrupt processes.
 *
 * @seealso default_restrictions_calling_functions
 */
#define atomic_set(ptr, value) (atomic_set_(ptr, value))

/**
 * @macro atomic_inc
 *
 * @brief Atomically increments (by one) the variable pointed to by var_ptr.
 * @long This function does not implement any memory barriers.
 *
 * @input ptr       points to the variable to be incremented.
 *
 * @restriction For some ARM targets this function is only available when
 *              executing in supervisor state.
 *    @break
 *              Also available in interrupt and timer interrupt processes.
 *
 * @seealso default_restrictions_calling_functions
 */
#define atomic_inc(ptr)               atomic_add(1, ptr)

/**
 * @macro atomic_dec
 *
 * @brief Atomically decrements (by one) the variable pointed to by var_ptr.
 * @long This function does not implement any memory barriers.
 *
 * @input ptr       points to the variable to be decremented.
 *
 * @restriction For some ARM targets this function is only available when
 *              executing in supervisor state.
 *    @break
 *              Also available in interrupt and timer interrupt processes.
 *
 * @seealso default_restrictions_calling_functions
 */
#define atomic_dec(ptr)               atomic_sub(1, ptr)

/**
 * @macro atomic_inc_return
 *
 * @brief Atomically increment (by one) the variable pointed to by var_ptr, 
 * and return the result.
 * @long This function implements memory barriers.
 *
 * @input ptr       points to the variable to be used
 * @return          the result is returned too
 *
 * @restriction For some ARM targets this function is only available when
 *              executing in supervisor state.
 *    @break
 *              Also available in interrupt and timer interrupt processes.
 *
 * @seealso default_restrictions_calling_functions
 */
#define atomic_inc_return(ptr)        atomic_add_return(1, ptr)

/**
 * @macro atomic_dec_return
 *
 * @brief Atomically decrement (by one) the variable pointed to by var_ptr,
 * and return the result.
 * @long This function implements memory barriers.
 *
 * @input ptr       points to the variable to be used
 * @return          the result is returned too
 *
 * @restriction For some ARM targets this function is only available when
 *              executing in supervisor state.
 *    @break
 *              Also available in interrupt and timer interrupt processes.
 *
 * @seealso default_restrictions_calling_functions
 */
#define atomic_dec_return(ptr)        atomic_sub_return(1, ptr)

/**
 * @function atomic_add
 *
 * @brief Atomically adds a value to the variable pointed to by var_ptr.
 * @long This function does not implement any memory barriers.
 *
 * @input value         The value to be added
 * @input var_ptr       points to the variable to be used
 * @output var_ptr      the result is at the same location
 *
 * @restriction For some ARM targets this function is only available when
 *              executing in supervisor state.
 *    @break
 *              Also available in interrupt and timer interrupt processes.
 *
 * @seealso default_restrictions_calling_functions
 */
static always_inline void
atomic_add(uint32_t value, atomic_t *var_ptr);

/**
 * @function atomic_sub
 *
 * @brief Atomically subtracts a value to the variable pointed to by var_ptr.
 * @long This function does not implement any memory barriers.
 *
 * @input value         The value to be substracted
 * @input var_ptr       points to the variable to be used
 * @output var_ptr      the result is at the same location
 *
 * @restriction For some ARM targets this function is only available when
 *              executing in supervisor state.
 *    @break
 *              Also available in interrupt and timer interrupt processes.
 *
 * @seealso default_restrictions_calling_functions
 */
static always_inline void
atomic_sub(uint32_t value, atomic_t *var_ptr);

/**
 * @function atomic_and
 *
 * @brief Performs atomic logical-and.
 * @long This function does not implement any memory barriers.
 * *var_ptr = *var_ptr & value;
 *
 * @input value         The value to and with
 * @input var_ptr       points to the variable to be used
 * @output var_ptr      the result is at the same location
 *
 * @restriction For some ARM targets this function is only available when
 *              executing in supervisor state.
 *    @break
 *              Also available in interrupt and timer interrupt processes.
 *
 * @seealso default_restrictions_calling_functions
 */
static always_inline void
atomic_and(uint32_t value, atomic_t *var_ptr);

/**
 * @function atomic_or
 *
 * @brief Performs atomic logical-or.
 * @long This function does not implement any memory barriers.
 * *var_ptr = *var_ptr | value;
 *
 * @input value         The value to or with
 * @input var_ptr       points to the variable to be used
 * @output var_ptr      the result is at the same location
 *
 * @restriction For some ARM targets this function is only available when
 *              executing in supervisor state.
 *    @break
 *              Also available in interrupt and timer interrupt processes.
 *
 * @seealso default_restrictions_calling_functions
 */
static always_inline void
atomic_or(uint32_t value, atomic_t *var_ptr); 

/**
 * @function atomic_add_return
 *
 * @brief Atomically adds a value to the variable pointed to by var_ptr,
 * and return the result.
 * @long This function implements memory barriers.
 *
 * @input value         The value to be added
 * @input var_ptr       points to the variable to be used
 * @output var_ptr      the result is at the same location
 * @return              the result is returned too
 *
 * @restriction For some ARM targets this function is only available when
 *              executing in supervisor state.
 *    @break
 *              Also available in interrupt and timer interrupt processes.
 *
 * @seealso default_restrictions_calling_functions
 */
static always_inline uint32_t
atomic_add_return(uint32_t value, atomic_t *var_ptr);

/**
 * @function atomic_sub_return
 *
 * @brief Atomically substracts a value to the variable pointed to by var_ptr,
 * and return the result.
 * @long This function implements memory barriers.
 *
 * @input value         The value to be substracted
 * @input var_ptr       points to the variable to be used
 * @output var_ptr      the result is at the same location
 * @return              the result is returned too
 *
 * @restriction For some ARM targets this function is only available when
 *              executing in supervisor state.
 *    @break
 *              Also available in interrupt and timer interrupt processes.
 *
 * @seealso default_restrictions_calling_functions
 */
static always_inline uint32_t
atomic_sub_return(uint32_t value, atomic_t *var_ptr);

/**
 * @function atomic_and_return
 *
 * @brief Atomically logical and a value to the variable pointed to by var_ptr,
 * and return the result.
 * @long This function implements memory barriers.
 *
 * @input value         The value to and with
 * @input var_ptr       points to the variable to be used
 * @output var_ptr      the result is at the same location
 * @return              the result is returned too
 *
 * @restriction For some ARM targets this function is only available when
 *              executing in supervisor state.
 *    @break
 *              Also available in interrupt and timer interrupt processes.
 *
 * @seealso default_restrictions_calling_functions
 */
static always_inline uint32_t
atomic_and_return(uint32_t value, atomic_t *var_ptr);

/**
 * @function atomic_or_return
 *
 * @brief Atomically logical or a value to the variable pointed to by var_ptr,
 * and return the result.
 * @long This function implements memory barriers.
 *
 * @input value         The value to or with
 * @input var_ptr       points to the variable to be used
 * @output var_ptr      the result is at the same location
 * @return              the result is returned too
 *
 * @restriction For some ARM targets this function is only available when
 *              executing in supervisor state.
 *    @break
 *              Also available in interrupt and timer interrupt processes.
 *
 * @seealso default_restrictions_calling_functions
 */
static always_inline uint32_t
atomic_or_return(uint32_t value, atomic_t *var_ptr); 


/**
 * @function atomic_cas
 *
 * @brief  Atomic compare-and-swap. Compares the variable pointed to by 'ptr'
 * with 'old_value'. If they are equal, the variable is updated to 'new_value'.
 * atomic_cas implements read and write memory barriers.
 *
 * @long atomic_cas() may be used for example to take a multicore lock.
 * To properly release a multicore lock, an EXPORT MEMORY BARRIER is needed.
 * It is possible to release the lock by a second atomic_cas() call.
 *
 * @input var_ptr   points to the variable to be used
 * @input old_value value value
 * @input new_value new value
 * @output var_ptr  the result is at the same location
 * @return          previous value
 *
 * @restriction For some ARM targets this function is only available when
 *              executing in supervisor state.
 *    @break
 *              Also available in interrupt and timer interrupt processes.
 *
 * @seealso default_restrictions_calling_functions
 */
static always_inline uint32_t
atomic_cas(atomic_t *var_ptr, uint32_t old_value, uint32_t new_value);

#ifdef __cplusplus
}
#endif

#include "cpu_atomic.h"

/* End #ifndef _ATOMIC_H */
#endif
