/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2011 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */
/**
 * @toc OSE:
 * @file ose_spi/bios.h
 * @brief OSE Bios offers basic services for use of the System Call / Trap /
 * Software Interrupt mechanism.
 *
 * @long
 * bios.h is part of the System Programming Interface (SPI), which is mainly
 * intended for platform developers.
 * @long
 * Bios handles the system call mechanism. The system call mechanism allows
 * programs to call the executive or other components that have registered
 * with Bios, without knowing the address that the executive or component is
 * located on. It also provides a way for user mode code to switch to
 * supervisor mode in a controlled way.
 * @long
 * The OSE Bios component has nothing to do with the Basic Input/Output System
 * (BIOS) term in the desktop PC world.
 * @long
 * Whenever an application executes a system call instruction (trap, software
 * interrupt, etc., the name varies with the processor architecture), the
 * execution jumps to the Bios component. Bios provides three operations that
 * are essential for the operation of the system; Install, Open and Call.
 * @long
 * The BIOS services are available for all types of processes. Installation
 * operations require Supervisor Mode privileges.
 */

#ifndef _BIOS_H
#define _BIOS_H

#ifdef __cplusplus
extern "C" {
#endif

/*
 ****************************************************************************
 * MANIFEST CONSTANTS.
 ****************************************************************************
 */

/**
 * @macro BIOS_NAME_SIZE
 *
 * @brief Length of Bios name string, with terminator included.
 */
#define BIOS_NAME_SIZE          24

/*
 *---------------------------------------------------------------------------
 * The biosInstall flags
 *---------------------------------------------------------------------------
 */

/**
 * @macro BIOS_SUPERVISOR_ONLY
 *
 * @brief biosInstall flag: Available from Supervisor mode only. (The default)
 */
#define BIOS_SUPERVISOR_ONLY    0

/**
 * @macro BIOS_USERMODE_ALLOWED
 *
 * @brief biosInstall flag: Clients executing in user mode are allowed to
 * open this entry.
 */
#define BIOS_USERMODE_ALLOWED   1

/**
 * @macro BIOS_RESERVED_BY_KERNEL
 *
 * @brief biosInstall flag: Flag reserved by the OSE kernel.
 */
#define BIOS_RESERVED_BY_KERNEL 2

/**
 * @macro BIOS_FORCE_INSTALL
 *
 * @brief biosInstall flag:  Replace function and flags if a function with
 * this name exists.
 */
#define BIOS_FORCE_INSTALL      4

/**
 * @macro BIOS_SUPPORTS_FCDEVCLASS_CONVENTION
 *
 * @brief biosInstall flag: This entry is a device driver, following the OSE
 * BSP specification.
 */
#define BIOS_SUPPORTS_FCDEVCLASS_CONVENTION 8

/**
 * @macro BIOS_SMP_UNSAFE
 *
 * @brief biosInstall flag: This entry is classified as unsafe to use from
 * more than one CPU in an SMP (Symmetric Multi Processing) system. It must
 * only be used on CPU number zero.
 * The intention with the flag is to simplify migration of existing systems
 * into an SMP environment.
 */
#define BIOS_SMP_UNSAFE 16

/*
 *---------------------------------------------------------------------------
 * The biosInstall and biosList success and error codes.
 *---------------------------------------------------------------------------
 */

/**
 * @macro BIOS_SUCCESS
 *
 * @brief Status returned by biosInstall and biosList: operation successful.
 */
#define BIOS_SUCCESS            0

/**
 * @macro BIOS_ENO_TABLE_SPACE
 *
 * @brief Status returned by biosInstall: Internal data area exhausted.
 */
#define BIOS_ENO_TABLE_SPACE    1

/**
 * @macro BIOS_EDUPLICATE_NAME
 *
 * @brief Status returned by biosInstall: Module with this name already
 * present.
 */
#define BIOS_EDUPLICATE_NAME    2

/**
 * @macro BIOS_ENAME_TOO_LONG
 *
 * @brief Status returned by biosInstall: Name too long.
 */
#define BIOS_ENAME_TOO_LONG     3

/**
 * @macro BIOS_EUSER_MODE
 *
 * @brief Status returned by biosInstall: Attempt to install from user mode.
 */
#define BIOS_EUSER_MODE         4

/**
 * @macro BIOS_ENO_ENTRY
 *
 * @brief Status returned by biosList: No find.
 */
#define BIOS_ENO_ENTRY          5

/*
 ****************************************************************************
 * 6  TYPES.
 ****************************************************************************
 */

/**
 * @type BiosFunction
 *
 * @brief The type of a function that can be installed in the BIOS.
 */
typedef long (BiosFunction)(long, long, long, long, long, long, long);


/**
 * @type BiosList
 *
 * @brief The BIOS directory entry as presented by biosList.
 * @field entrypoint Entrypoint. A pointer to a function of 
 *                   type @link "" "BiosFunction".
 * @field runlevel   Runlevel is a number in the range 0 - 255.
 * @field handle     Handle. The handle is used to call the 
 *                   function via @link "" "biosCall".
 * @field flags      Flags. See @link "" "biosInstall".
 * @field name       Name. See @link "" "biosInstall".
 */
struct BiosList {
  BiosFunction *entrypoint;
  long runlevel;
  unsigned long handle;
  unsigned long flags;
  char name[BIOS_NAME_SIZE];
};

/*
 ****************************************************************************
 * 7 GLOBAL FUNCTIONS
 ****************************************************************************
 */

/**
 * @function biosInstall
 * @include "bios.h"
 *
 * @declaration long
 * biosInstall(const char *name, BiosFunction *entrypoint, unsigned long flags);
 *
 * @brief Installs a function with the specified name string.
 * @long
 * The installed function will be available also for other programs via
 * @link "" "biosOpen" and @link "" "biosCall".
 *
 * @input name          Name string with at most 24 characters including the nul
 *                      terminator.
 * @input entrypoint    A pointer to a function of type @link "" "BiosFunction"
 * @input flags         A bit mask of biosInstall flags, describing
 *                      selected installation options. A few flag examples:
 *                      @link "" "BIOS_USERMODE_ALLOWED",
 *                      @link "" "BIOS_SUPPORTS_FCDEVCLASS_CONVENTION",
 *                      @link "" "BIOS_SMP_UNSAFE".
 * @return Returns zero for a successful operation, non-zero for failure.
 *
 * @restriction
 *    Only available in supervisor mode processes
 * @break
 *    The feature is intended to provide permanent services. BiosInstall
 *    by dynamic programs is permitted but the function will no longer work
 *    if the program is killed.
 * @break
 *    The number of entries in the BIOS tables is limited to 64 functions.
 * @seealso biosOpen
 * @seealso biosCall
 * @seealso default_restrictions_calling_functions
 * @example
 *
 * // Install a function for reading CPU registers in user mode.
 * // create_usermode_get_spr() must be called in supervisor mode.
 * // Actually get_spr() is only available in PowerPC
 * #include "bios.h"
 * #include "ose.h"
 * #include "cpu_access.h"
 *
 * // The function to install.
 * static long biosmodule_get_spr(unsigned long spr_no)
 * {
 *    return get_spr(spr_no);
 * }
 *
 * // Install it.
 * void create_usermode_get_spr(void)
 * {
 *    long rc = biosInstall("GET_SPR",
 *                          (BiosFunction *) biosmodule_get_spr,
 *                          BIOS_USERMODE_ALLOWED);
 *    if(rc != BIOS_SUCCESS)
 *       error2(0x17, (OSERRCODE) rc);
 * }
 */

long
biosInstall(const char *name, BiosFunction *entrypoint, unsigned long flags);

/**
 * @function biosOpen
 * @include "bios.h"
 *
 * @declaration unsigned long
 * biosOpen(const char *name);
 *
 * @brief Open a BIOS function for use.
 * @long
 * The function with the given name is located, and a handle to it is returned.
 * The handle is used to call the function via @link "" "biosCall".
 *
 * @input name The name string of the wanted function.
 * @return The handle of the function if non-zero. Zero is returned for
 * failure. Failure can occur if the name is not found or the function
 * requires supervisor mode privileges.
 * @seealso biosCall
 * @seealso biosInstall
 * @seealso default_restrictions_calling_functions
 */

unsigned long
biosOpen(const char *name);

/**
 * @function biosCall
 * @include "bios.h"
 *
 * @declaration long
 * biosCall(unsigned long handle, ...);
 *
 * @brief Call the function specified by handle with the specified
 * parameters (up to seven of them).
 *
 * @input handle Function identity, retrieved by biosOpen.
 * @return The value returned by the function identified by the handle.
 * If the handle is invalid, -1 is returned.
 * @field ... Parameters to the BIOS function.
 * @seealso biosOpen
 * @seealso biosInstall
 * @seealso default_restrictions_calling_functions
 * @example
 * // Assume there is an installed BIOS function named "GET_SPR" to make
 * // some hardware registers readable from user mode, and that function
 * // takes a register identity as parameter. The function returns the
 * // register contents. This example shows how to call the function.
 *
 * #include "bios.h"
 * #include "../ose.h"
 *
 * //
 * unsigned long usermode_get_spr(unsigned long spr_no)
 * {
 *   static unsigned long handle = 0;
 *
 *   // First time, fetch the function handle.
 *   if(handle == 0)
 *     {
 *       handle = biosOpen("GET_SPR");
 *
 *       // Handle the case that the function did not exist.
 *       if(handle == 0)
 *         {
 *           error(0x19);
 *           return 0;
 *         }
 *     }
 *
 *   // Get the wanted register value
 *   return biosCall(handle, spr_no);
 * }
 */

long
biosCall(unsigned long handle, ... /* arg1..arg7 */);

/**
 * @function biosList
 * @include "bios.h"
 *
 * @declaration long
 * biosList(const char * wildcardName, unsigned long prevHandle,
 *          struct BiosList * biosInfo);
 *
 *
 * @brief biosList returns information on installed BIOS functions.
 * @long
 * Each call returns one function. The search starts from
 * the prev_handle, normally returned by the previous call.
 * Set prev_handle to zero in the first call. @break
 * Wildcard expansion rules:
 *              '*' matches zero or more characters.
 *              '?' matches any single character except the string terminator.
 *
 * @input wildcardName  The name pattern wanted.
 * @input prevHandle    The handle returned in the result struct of the
 *                      previous call, or zero to start scanning from the
 *                      beginning of the table.
 * @input biosInfo      A pointer to the result structure.
 * @return              BIOS_SUCCESS when a function is found,
 *                      BIOS_ENO_ENTRY for no find.
 * @seealso biosInstall
 * @seealso default_restrictions_calling_functions
 */

long
biosList(const char * wildcardName, unsigned long prevHandle,
         struct BiosList * biosInfo);

/*
 *===========================================================================
 *                      biosUninstall
 *===========================================================================
 *
 * NOTE: This function is DEPRECATED! Do not use!
 *
 * biosUninstall uninstalls a number of handlers matching the wild card
 * argument. Use this function with care. Later calls to biosInstall will
 * reuse the released handles.
 *
 * Wildcard expansion rules:
 *              '*' matches zero or more characters.
 *              '?' matches any single character except the string terminator.
 *
 * biosUninstall returns the number of functions removed when called from
 * supervisor mode respectively. -1 when illegally called from user mode or when
 * Uninstall is not implemented.
 */
long
biosUninstall(const char * wildcardName);

/*
 *===========================================================================
 *                      biosRunlevel
 *===========================================================================
 *
 * NOTE: This function is DEPRECATED! Do not use!
 *
 * biosRunlevel sets the current run level.
 * Runlevel is a number in the range 0:255.
 *
 * biosRunlevel returns the previous run level.
 * Calls are only permitted from supervisor mode.
 * When called from user mode or when not implemented, -1 is returned.
 *
 *
 */
long
biosRunlevel(long runlevel);

#ifdef __cplusplus
}
#endif
#endif /* # ifndef _BIOS_H */
