/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2011 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */
/**
 * @public
 * @toc OSE:
 * @file ose_spi/boardparam.h
 *
 * @brief Public declaration for the board parameter interface.
 *
 * @long
 * This file contains prototypes and defines for the board
 * parameter interface. The client for to this interface is the
 * kernel and the provider is the customizable part. The functions
 * declared in this interface has to be defined and linked to
 * the core, for example in boardparam.c.
 * The board parameter interface is invoked by the kernel when
 * the system parameter interface (sysparam.h) is used from the
 * application level.
 */

#ifndef OSE_BOARDPARAM_H_
#define OSE_BOARDPARAM_H_

#include "stddef.h"

#ifdef __cplusplus
extern "C" {
#endif

/**
 * @function boardparam_read
 * @brief This function is used to retrieve the string that is the value of
 *        a board parameter stored in flash.
 * @long  The value is copied to the submitted buffer. If the buffer size
 *        is equal to or smaller than the actual value string length, only
 *        "buffer_size - 1" characters are copied. In any case, the function
 *        always null terminates what's copied to the buffer.
 * @input  name        Specifies the name of the parameter to be read.
 * @input buffer       Address of the buffer where the value of the parameter
 *                     should be written. May be NULL, in which case no copy
 *                     will take place.
 * @output buffer      The buffer content is updated.
 * @input buffer_size  Size of buffer.
 * @return The actual length of the parameter value, (regardless if the
 *         buffer was too small to hold it and the value had to be
 *         truncated), without counting the trailing NULL. If the parameter
 *         was not found, the function returns 0.
 * @seealso default_restrictions_calling_functions
 */
size_t boardparam_read(const char *name, char *buffer, size_t buffer_size);


/**
 * @function boardparam_write
 * @brief    This function is used to add, modify or delete one board
 *           parameter.
 * @long     Its usage should normally be restricted to system servicing
 *           code, or boot monitor, since, according to system specific
 *           implementation, this function may have big impact in real-time
 *           characteristics.
 *
 * @long     Warning: Must be used with care, since according to how
 *                         the parameter storage is implemented in the
 *                         actual system, calling this function may have
 *                         deterious impact on real-time characteristics,
 *                         etc.
 * @input  name  The name of the addressed board parameter.
 * @input  value The new value for the addressed board parameter (may
 *		    not be NULL). If it is a pointer to a null-character,
 *		    the parameter will be deleted.
 * @return The function returns 0 for success, != 0 otherwise.
 * @seealso default_restrictions_calling_functions
 */
int boardparam_write(const char *name, const char *value);


/**
 * @function boardparam_list_next
 * @brief    This function supports iterative traversal of the board
 *	        parameter namespace.
 * @long     The function takes the string submitted as 'str' and searches
 *           for the parameter name that would lexicographically follow.
 *           The 'str' and the 'buffer' parameters may be pointers to the
 *           same buffer.
 * @input  str          Specifies the starting point for the search as
 *                      described  above. It may be a be an empty string, in
 *                      which case the (lexicographically) first
 *                      configuration parameter is searched for.
 * @input buffer	   Pointer to the buffer where this function will
 *			   write the found parameter name, up to 'buffer_size'
 *                         characters. May be NULL, in which case no copying
 *			   will take place (but the found parameter name length
 *			   will still be returned).
 * @output buffer	   The buffer content is updated.
 * @input  buffer_size  Size of buffer. Must be 0 if 'buffer' is NULL.
 * @return The function returns the length of the found parameter name,
 *	      regardless if the string copied to 'buffer' was truncated
 *	      (because it was too small to hold it) or 'buffer' was NULL.
 *	      If no parameter name lexicographically following 'str'
 *	      could be found, the function returns zero.
 * @seealso default_restrictions_calling_functions
 * @example
 * This example would print all parameters.
 *
 * char name[NAME_BUF_SIZE];
 * size_t len = boardparam_list_next("", name, NAME_BUF_SIZE);
 * while (len > 0)
 *   {
 *     if (len < NAME_BUF_SIZE)
 *       {
 *         char value[VAL_BUF_SIZE];
 *         size_t val_len = boardparam_read(name, value, VAL_BUF_SIZE);
 *         if (val_len > 0)
 *           {
 *             printf("%s=\"%s\"\n", name, value);
 *           }
 *       }
 *     else
 *       {
 *         break;
 *       }
 *    len = boardparam_list_next(name, name, NAME_BUF_SIZE);
 *  }
 */
size_t boardparam_list_next(const char *str, char *buffer, size_t buffer_size);

#ifdef __cplusplus
}
#endif

#endif /* #ifndef OSE_BOARDPARAM_H_ */
