/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2011 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */
/**
 * @public
 * @toc OSE:
 * @file ose_spi/cache.h
 *
 * @brief     Cache function interface.
 *
 * @long      Provides a generic interface for supported architectures.
 *
 */
#ifndef CACHE_H_
#define CACHE_H_

#include "ose.h"
#include "osetypes.h"
#include "stddef.h"

#ifdef __cplusplus
extern "C" {
#endif

/**
 * @function cache_sync
 *
 * @brief Synchronize caches.
 *
 * @long  Flush data cache and invalidate instruction cache for
 *        a given memory range.
 *
 *
 * @input base    The base address of the memory range.
 *
 * @input size    The size of the memory range.
 *
 * @deprecated    This method is deprecated in OSE 5.5 and will not be
 *                supported in future OSE releases.
 *                See "ose_cache" for similar functionality.
 */
void
cache_sync(OSADDRESS base, size_t size);


/**
 * @function cache_flush_d
 *
 * @brief Flush the data cache for a given memory range.
 *
 * @long  Note that on some architectures this function includes
 *        invalidation of the memory range.
 *
 *
 * @input base    The base address of the memory range.
 *
 * @input size    The size of the memory range.
 *
 * @deprecated    This method is deprecated in OSE 5.5 and will not be
 *                supported in future OSE releases.
 *                See "ose_cache" for similar functionality.
 *
 */
void
cache_flush_d(OSADDRESS base, size_t size);


/**
 * @function cache_invalidate_d
 *
 * @brief Invalidate the data cache for a given memory range.
 *
 * @input base    The base address of the memory range.
 *
 * @input size    The size of the memory range.
 *
 * @deprecated    This method is deprecated in OSE 5.5 and will not be
 *                supported in future OSE releases.
 *                See "ose_cache" for similar functionality.
 *
 */
void
cache_invalidate_d(OSADDRESS base, size_t size);


/**
 * @function cache_flush_invalidate_d
 *
 * @brief Flush and invalidate the data cache for a given memory range.
 *
 * @input base    The base address of the memory range.
 *
 * @input size    The size of the memory range.
 *
 * @deprecated    This method is deprecated in OSE 5.5 and will not be
 *                supported in future OSE releases.
 *                See "ose_cache" for similar functionality.
 *
 */
void
cache_flush_invalidate_d(OSADDRESS base, size_t size);


/**
 * @function cache_invalidate_i
 *
 * @brief Invalidate the instruction cache for a given memory range.
 *
 * @input base    The base address of the memory range.
 *
 * @input size    The size of the memory range.
 *
 * @deprecated    This method is deprecated in OSE 5.5 and will not be
 *                supported in future OSE releases.
 *                See "ose_cache" for similar functionality.
 *
 */
void
cache_invalidate_i(OSADDRESS base, size_t size);


/**
 * @function cache_flush_d_complete
 *
 * @brief Flush the entire data cache.
 *
 * @long  Flush all dirty lines in data cache to main memory.
 *        If <disable_d> is true, disable data cache after flushing is
 *        performed.
 *
 * @input  disable_d   Disable the data cache after flushing if argument
 *                     <disable_d> is non-zero.
 *
 * @deprecated    This method is deprecated in OSE 5.5 and will not be
 *                supported in future OSE releases.
 *                See "ose_cache" for similar functionality.
 */
void
cache_flush_d_complete(int disable_d);

/**
 * @function cache_invalidate_i_complete
 *
 * @brief Invalidate the entire instruction cache.
 *
 * @long  Invalidate the entire instruction cache.
 *        If <disable_i> is true, disable instruction cache after
 *        invalidation is performed.
 *
 * @input  disable_i   Disable the instruction cache after invalidation if
 *                     argument <disable_i> is non-zero.
 *
 * @deprecated    This method is deprecated in OSE 5.5 and will not be
 *                supported in future OSE releases.
 *                See "ose_cache" for similar functionality.
 */
void
cache_invalidate_i_complete(int disable_i);


/**
 * @macro CACHE_QUERY_D_CACHE_LINE_SIZE
 *
 * @brief Get d-cache line size in bytes, 0 if no D-cache exists.
 *        If the query is not supported,
 *        @link "cache.h" "CACHE_QUERY_NOT_SUPPORTED" is returned.
 * @seealso ose_cache_query
 */
#define CACHE_QUERY_D_CACHE_LINE_SIZE 0

/**
 * @macro CACHE_QUERY_I_CACHE_LINE_SIZE
 *
 * @brief Get I-cache line size in bytes, 0 if no I-cache exists.
 *        If the query is not supported,
 *        @link "cache.h" "CACHE_QUERY_NOT_SUPPORTED" is returned.
 * @seealso ose_cache_query
 */
#define CACHE_QUERY_I_CACHE_LINE_SIZE 1

/**
 * @macro CACHE_QUERY_D_CACHE_SIZE
 *
 * @brief Get D-cache size in bytes, 0 if no D-cache exists.
 *        If the query is not supported,
 *        @link "cache.h" "CACHE_QUERY_NOT_SUPPORTED" is returned.
 * @seealso ose_cache_query
 */
#define CACHE_QUERY_D_CACHE_SIZE 2

/**
 * @macro CACHE_QUERY_I_CACHE_SIZE
 *
 * @brief Get I-cache size in bytes, 0 if no I-cache exists.
 *        If the query is not supported,
 *        @link "cache.h" "CACHE_QUERY_NOT_SUPPORTED" is returned.
 * @seealso ose_cache_query
 */
#define CACHE_QUERY_I_CACHE_SIZE 3

/**
 * @macro CACHE_QUERY_D_CACHE_STATUS
 *
 * @brief Return 1 if D-cache exists and is enabled, 0 otherwise.
 *        If the query is not supported,
 *        @link "cache.h" "CACHE_QUERY_NOT_SUPPORTED" is returned.
 * @seealso ose_cache_query
 */
#define CACHE_QUERY_D_CACHE_STATUS 4

/**
 * @macro CACHE_QUERY_I_CACHE_STATUS
 *
 * @brief Return 1 if I-cache exists and is enabled, 0 otherwise.
 *        If the query is not supported,
 *        @link "cache.h" "CACHE_QUERY_NOT_SUPPORTED" is returned.
 * @seealso ose_cache_query
 */
#define CACHE_QUERY_I_CACHE_STATUS 5

/**
 * @macro CACHE_QUERY_L2_CACHE_STATUS
 *
 * @brief Return 1 if there is a L2 cache available and enabled, 0 otherwise.
 *        If the query is not supported,
 *        @link "cache.h" "CACHE_QUERY_NOT_SUPPORTED" is returned.
 * @seealso ose_cache_query
 */
#define CACHE_QUERY_L2_CACHE_STATUS 6

/**
 * @macro CACHE_QUERY_L3_CACHE_STATUS
 *
 * @brief Return 1 if there is a L3 cache available and enabled, 0 otherwise.
 *        If the query is not supported,
 *        @link "cache.h" "CACHE_QUERY_NOT_SUPPORTED" is returned.
 * @seealso ose_cache_query
 */
#define CACHE_QUERY_L3_CACHE_STATUS 7

/**
 * @macro CACHE_QUERY_L2_CACHE_SIZE
 *
 * @brief Returns size of L2 cache.
 *        If the query is not supported,
 *        @link "cache.h" "CACHE_QUERY_NOT_SUPPORTED" is returned.
 * @seealso ose_cache_query
 */
#define CACHE_QUERY_L2_CACHE_SIZE   8

/**
 * @macro CACHE_QUERY_L3_CACHE_SIZE
 *
 * @brief Returns size of L3 cache.
 *        If the query is not supported,
 *        @link "cache.h" "CACHE_QUERY_NOT_SUPPORTED" is returned.
 * @seealso ose_cache_query
 */
#define CACHE_QUERY_L3_CACHE_SIZE   9

/**
 * @macro CACHE_QUERY_INITIALIZED
 *
 * @brief Return 1 if the caches are initialized, 0 otherwise.
 *        If the query is not supported,
 *        @link "cache.h" "CACHE_QUERY_NOT_SUPPORTED" is returned.
 * @seealso ose_cache_query
 */
#define CACHE_QUERY_INITIALIZED    10

/**
 * @macro CACHE_QUERY_LEVEL
 *
 * @brief Return the physical cache level for a logical cache level.
 *
 * @long This request requires an additional parameter to be passed to
 * ose_cache_query defining the logical cache level.
 *
 * @seealso ose_cache_query
 *
 * @example
 * {
 *       U32 phys_cache_level;
 *
 *       phys_cache_level =
 *       ose_cache_query(CACHE_QUERY_LEVEL, CACHE_DMA);
 *
 *       switch (phys_cache_level)
 *       {
 *          case CACHE_L1:
 *          {
 *             // Code //
 *             break;
 *          }
 *          case CACHE_L2:
 *          {
 *             // Code //
 *             break;
 *          }
 *          case CACHE_L3:
 *          {
 *             // Code //
 *             break;
 *          }
 *       }
 *    }
 */
#define CACHE_QUERY_LEVEL          11

/**
 * @macro CACHE_QUERY_NOT_SUPPORTED
 *
 * @brief Returned if cache query not supported.
 * @seealso ose_cache_query
 */
#define CACHE_QUERY_NOT_SUPPORTED   -1


/**
 * @function cache_get_characteristics
 *
 * @brief Returns information requested in the given query identifier.
 *
 * @input query  Requested query to be performed. One of the following
 *
 *               @link "cache.h" "CACHE_QUERY_D_CACHE_LINE_SIZE",
 *               @link "cache.h" "CACHE_QUERY_I_CACHE_LINE_SIZE",
 *               @link "cache.h" "CACHE_QUERY_D_CACHE_SIZE",
 *               @link "cache.h" "CACHE_QUERY_I_CACHE_SIZE",
 *               @link "cache.h" "CACHE_QUERY_D_CACHE_STATUS",
 *               @link "cache.h" "CACHE_QUERY_I_CACHE_STATUS",
 *               @link "cache.h" "CACHE_QUERY_L2_CACHE_STATUS",
 *               @link "cache.h" "CACHE_QUERY_L3_CACHE_STATUS",
 *
 * @return       The answer of the query specified as parameter.
 *
 * @deprecated   This method is deprecated in OSE 5.5 and will not be
 *               supported in future OSE releases.
 *               See "ose_cache_query" for similar functionality.
 */
U32
cache_get_characteristics(U32 query);


/**
 * @function cache_initialized
 *
 * @brief Checks if cache support functions are initialized.
 *
 * @long   Returns true if the CPU HAL is initialized.
 *         If false, the result is unpredictable and the cache functions
 *         must not be called.
 *         NOTE: This functions only purpose is for very early users such
 *         as RAMLOG. When the system is up, this function does not have
 *         to be called.
 *
 * @return True if functions can be used.
 *
 * @deprecated  This method is deprecated in OSE 5.5 and will not be
 *              supported in future OSE releases.
 *              See "ose_cache_query" for similar functionality.
 */
Boolean
cache_initialized(void);




/* CACHE ACTIONS */
#define CACHE_SYS_ACTION_BASE 0x00000000 /* Action range defined by OSE */
#define CACHE_USR_ACTION_BASE 0x80000000 /* Action range for user additions */

/**
 * @macro CACHE_FLUSH_D
 * @brief Flush the data cache.
 *
 * @long  The flush_d action causes modified blocks in the data cache to be
 *        copied from the selected cache level (including all lower levels)
 *        to the next cache level.
 *        This action is translated to a cache clean operation on ARM, a
 *        cache store operation on PPC etc.
 *
 * @long  This operation is typically used prior to performing DMA output.
 *        Flushing the cache ensures that all cache contents is present in
 *        memory before the memory is read by DMA.
 */
#define CACHE_FLUSH_D         (CACHE_SYS_ACTION_BASE+0x10)

/**
 * @macro CACHE_INVALIDATE_D
 * @brief Invalidate the data cache.
 *
 * @long  The invalidate_d action causes blocks in the data cache to be
 *        invalidated for the selected cache level (including all lower
 *        levels).
 *
 * @long  This operation is typically used prior to performing DMA input.
 *        Invalidating the cache ensures that no cache contents will
 *        interfere with the memory contents that gets loaded by DMA.
 *
 * @long  Warning: Applications and drivers that perform this operation
 *        should only operate on buffers where the base address and size
 *        are cache line aligned. Otherwise data adjacent to the buffer
 *        but within the same cache lines will get destroyed. 
 *
 * @long  Warning: Buffers allocated from the heap are typically not cache
 *        line aligned. To safely invalidate heap-allocated memory,
 *        allocate a buffer that is one or two cache lines larger than
 *        needed, and invalidate at an offset into the buffer that is
 *        cache-line aligned. This creates a "safe zone" that prevents
 *        invalidation of heap meta data. A second cache line of safe zone
 *        is needed at the end if the data size is not cache line aligned.
 *        Similar precautions apply to pool-allocated buffers in systems
 *        where the cache line is larger than the pool buffer alignment of
 *        32 bytes.
 */
#define CACHE_INVALIDATE_D    (CACHE_SYS_ACTION_BASE+0x20)

/**
 * @macro CACHE_INVALIDATE_I
 * @brief Invalidate the instruction cache.
 *
 * @long  The invalidate_i action causes blocks in the instruction cache
 *        to be invalidated for the selected cache level (including all
 *        lower levels).
 */
#define CACHE_INVALIDATE_I    (CACHE_SYS_ACTION_BASE+0x30)



/* CACHE LEVELS */
#define CACHE_SYS_LEVEL_BASE  0x00000000 /* Level range defined by OSE */
#define CACHE_USR_LEVEL_BASE  0x80000000 /* Level range for user additions */

/**
 * @macro CACHE_L1
 * @brief Physical cache level 1.
 */
#define CACHE_L1              (CACHE_SYS_LEVEL_BASE+0x1)
/**
 * @macro CACHE_L2
 * @brief Physical cache level 2.
 */
#define CACHE_L2              (CACHE_SYS_LEVEL_BASE+0x2)
/**
 * @macro CACHE_L3
 * @brief Physical cache level 3.
 */
#define CACHE_L3              (CACHE_SYS_LEVEL_BASE+0x3)
/**
 * @macro CACHE_L4
 * @brief Physical cache level 4.
 */
#define CACHE_L4              (CACHE_SYS_LEVEL_BASE+0x4)
/**
 * @macro CACHE_L5
 * @brief Physical cache level 5.
 */
#define CACHE_L5              (CACHE_SYS_LEVEL_BASE+0x5)
/**
 * @macro CACHE_L6
 * @brief Physical cache level 6.
 */
#define CACHE_L6              (CACHE_SYS_LEVEL_BASE+0x6)
/**
 * @macro CACHE_L7
 * @brief Physical cache level 7.
 */
#define CACHE_L7              (CACHE_SYS_LEVEL_BASE+0x7)
/**
 * @macro CACHE_L8
 * @brief Physical cache level 8.
 */
#define CACHE_L8              (CACHE_SYS_LEVEL_BASE+0x8)
/**
 * @macro CACHE_L9
 * @brief Physical cache level 9.
 */
#define CACHE_L9              (CACHE_SYS_LEVEL_BASE+0x9)


/**
 * @macro CACHE_ALL
 * @brief Act on all cache levels.
 *
 * @long Logical cache level defined as the outer most physical
 * cache level. Acting on this level will cause the action to take
 * place on all defined physical levels (from the CPU to the level).
 */
#define CACHE_ALL             (CACHE_SYS_LEVEL_BASE+0x1000)

/**
 * @macro CACHE_DMA
 * @brief Logical cache level defined for DMA.
 */
#define CACHE_DMA             (CACHE_SYS_LEVEL_BASE+0x1010)

/**
 * @macro CACHE_PGM_LOAD
 * @brief Logical cache level defined for synchronization of the
 * I-cache and the D-cache after a program load.
 */
#define CACHE_PGM_LOAD        (CACHE_SYS_LEVEL_BASE+0x1020)

/**
 * @macro CACHE_INNER
 * @brief Logical cache level defined as the inner range.
 */
#define CACHE_INNER           (CACHE_SYS_LEVEL_BASE+0x1030)

/**
 * @macro CACHE_OUTER
 * @brief Logical cache level defined as the outer range.
 */
#define CACHE_OUTER           (CACHE_SYS_LEVEL_BASE+0x1031)


/**
 * @function ose_cache
 *
 * @brief Cache handling function.
 *
 * @long This is the extended OSE cache handling interface. It can be
 * used if needing to act on the cache system to a specific cache
 * level. The requests are passed to the "cache" function in the
 * CPU-HAL where actions and levels can be dealt with by system
 * software.
 *
 * @long The action defines what to do with the cache. The following
 * actions are required by OSE:
 * @link "cache.h" "CACHE_FLUSH_D",
 * @link "cache.h" "CACHE_INVALIDATE_D" and
 * @link "cache.h" "CACHE_INVALIDATE_I".
 *
 * @long The cache level can be physical or logical.
 *
 * @long A physical level identifies an actual physical cache
 * device. Acting on a physical level will affect all levels from the
 * CPU to (and including) the specified device level.
 *
 * @long The following physical levels are defined by OSE:
 * @link "cache.h" "CACHE_L1" to @link "cache.h" "CACHE_L9".
 *
 * @long A logical level identifies a level used for a certain
 * operation within the system. When the physical cache topology is
 * not known by the system software it can use the logical defined
 * levels instead. For logical levels that are not known by the cache
 * implementation (in the CPU HAL) the @link "cache.h" "CACHE_ALL"
 * level can be used.
 *
 * @long Acting on a logical level will cause the cache implementation
 * to act on all levels 'to' the defined logical level. I.e. in a
 * system including a L1 and a L2 cache, if DMA is hardwired to the L3
 * physical level (i.e. main memory), flushing data to DMA level will
 * flush the L1 and L2 physical levels.
 *
 * @long The following logical levels are required by OSE:
 * @link "cache.h" "CACHE_PGM_LOAD",
 * @link "cache.h" "CACHE_DMA" and
 * @link "cache.h" "CACHE_ALL".
 *
 * @long Base and size defines the range to be acted on.
 *
 * @long The complete cache is acted on if base and size are both set
 * to 0 (zero).
 *
 * @long The functionality exported in the ose_cache call is
 * available to all architectures. However, to use the level
 * support, a CPU HAL that supports the cache and cache_level methods
 * is required.
 *
 * @input action  Action to take on cache.
 *
 * @input level   Cache level the action shall take place on.
 *
 * @input base    The base address of the memory range.
 *
 * @input size    The size of the memory range.
 *
 * @restriction Only available in supervisor mode processes.
 *   @break     Also available in interrupt and timer interrupt processes.
 *
 * @seealso default_restrictions_calling_functions
 */
void
ose_cache(U32 action, U32 level, OSADDRESS base, size_t size);


/**
 * @function ose_cache_query
 *
 * @brief Returns information requested for the given query identifier.
 *
 * @input query  Requested query. One of the following:
 *
 *               @link "cache.h" "CACHE_QUERY_D_CACHE_LINE_SIZE",
 *               @link "cache.h" "CACHE_QUERY_I_CACHE_LINE_SIZE",
 *               @link "cache.h" "CACHE_QUERY_D_CACHE_SIZE",
 *               @link "cache.h" "CACHE_QUERY_I_CACHE_SIZE",
 *               @link "cache.h" "CACHE_QUERY_D_CACHE_STATUS",
 *               @link "cache.h" "CACHE_QUERY_I_CACHE_STATUS",
 *               @link "cache.h" "CACHE_QUERY_L2_CACHE_STATUS",
 *               @link "cache.h" "CACHE_QUERY_L3_CACHE_STATUS",
 *               @link "cache.h" "CACHE_QUERY_L2_CACHE_SIZE",
 *               @link "cache.h" "CACHE_QUERY_L3_CACHE_SIZE",
 *               @link "cache.h" "CACHE_QUERY_INITIALIZED",
 *               @link "cache.h" "CACHE_QUERY_LEVEL".
 *
 * @input ...    Additional parameters used for specific requests.
 *               These are defined by each specific request tag.
 *
 * @return       The answer of the query specified as parameter.
 *
 * @restriction Only available in supervisor mode processes.
 *   @break     Also available in interrupt and timer interrupt processes.
 *
 * @seealso default_restrictions_calling_functions
 */
U32
ose_cache_query(U32 query, ...);

#ifdef __cplusplus
}
#endif

#endif /* CACHE_H_ */
