/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2011 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */
/**
 * @toc OSE:
 *
 * @file ose_spi/checksum.h
 *
 * @brief Checksum function library API and definitions.
 *
 * @long These functions are a part of the C Run-time library (libcrt).
 *
 */

#ifndef _CHECKSUM_H
#define _CHECKSUM_H

#include "ose.h"
#include "osetypes.h"
#include "stddef.h"
#include "sys/types.h"

#ifdef __cplusplus
extern "C" {
#endif

/*
 ***********************************************************
 *
 * SYMBOLIC CONSTANTS.
 *
 ***********************************************************/

/*
  @macro OSE_ADLER32_INITIAL_SEED
  @brief The value to be used as seed in the initial
         function call of ose_adler32.
*/

#define OSE_ADLER32_INITIAL_SEED 1

/**
 * @macro CRC16_INIT
 *
 * @declaration #define CRC16_INIT
 *
 * @brief Initial 16-bit CRC value. Passed as input parameter to ose_crc16 as
 * the initial CRC value.
 *
 * @seealso ose_crc16
 */
#define CRC16_INIT      0xffff

/**
 * @macro CRC16_GOOD
 *
 * @declaration #define CRC16_GOOD
 *
 * @brief The expected result from a 16-bit CRC verification.
 *
 * @seealso ose_crc16
 */
#define CRC16_GOOD      0xf0b8

/**
 * @macro CRC32_INIT
 *
 * @declaration #define CRC32_INIT
 *
 * @brief Initial 32-bit CRC value. Passed as input parameter to ose_crc32 as the
 * initial CRC value.
 *
 * @seealso ose_crc32
 */
#define CRC32_INIT  0xffffffff

/**
 * @macro CRC32_GOOD
 *
 * @declaration #define CRC32_GOOD
 *
 * @brief The expected result from a 32-bit CRC verification.
 *
 * @seealso ose_crc32
 */
#define CRC32_GOOD  0xdebb20e3

typedef struct
{
   U32 state[5];
   U32 count[2];
   U8  buffer[64];
} SHA1_CTX;

/*
 ***********************************************************
 *
 * FUNCTION DECLARATIONS.
 *
 ***********************************************************/

/* NOT_USED_SOURCEDOC
 * @function  ose_adler32
 * @include "ose_spi/checksum.h"
 * @brief    Calculate the adler32 checksum of a buffer.
 * @long     The buffer will not be modified during the checksum
 *	     calculation. The buflen value reflects the size (in
 *	     bytes) of the buffer. The seed value is the value of
 * 	     OSE_ADLER32_INITIAL_SEED. If a buffer is
 *	     to be checksumed in intervals, the seed value
 *	     of OSE_ADLER32_INITIAL_SEED shall be used
 *	     in the first call to ose_adler32. The
 *	     return value of the ose_adler32 function
 *	     call is a new seed, to be used in the next
 *	     ose_adler32 function call. When the whole
 *	     buffer checksum has been calculated, the return
 *	     value of ose_adler32 is the calculated
 *	     adler32 checksum.
 * @input     buf    - The buffer to checksum.
 * @input     buflen - The length in bytes of the buffer (buf).
 * @input     seed   - The calculation seed.
 * @returns   The updated adler32 checksum.
 * @library   crt (C Run-time)
 * @standard  RFC1950
 * @restriction Also available in interrupt and timer interrupt processes.
 * @example
 * U32 adler32_sum =
 *   ose_adler32(buf, buflen, OSE_ADLER32_INITIAL_SEED);
 * ose_adler32_sum = ose_adler32(buf2, buflen, adler32_sum);
 *
 * @seealso   OSE_ADLER32_INITIAL_SEED
 * @seealso   default_restrictions_calling_functions
 */

U32
ose_adler32(const void *buf, size_t buflen,
	    U32 seed);

/**
 * @function ose_base64_encode
 *
 * @brief Encodes the input buffer to a base-64 representation.
 *
 * @long The base-64 encoding is designed to represent arbitrary
 * sequences of octets in a form that is not humanly readable. The
 * encoding and decoding algorithms are simple, but the encoded data
 * are consistently only 33 percent larger than the unencoded
 * data. The encoding is defined in RFC 1341.
 *
 * @field dest Pointer to a buffer that will receive the encoded
 * data. The size of this buffer is assumed to be large enough to hold
 * the encoded data, which will require (4*length/3) bytes.
 *
 * @field src Pointer to the buffer of data that will be encoded.
 *
 * @field length Length of the input buffer.
 *
 * @return The number of bytes encoded into the destination buffer.
 *
 * @restriction Also available in interrupt and timer interrupt processes.
 *
 * @seealso ose_base64_decode
 * @seealso default_restrictions_calling_functions
 *
 */
int ose_base64_encode(unsigned char *dest, const void *src, int length);


/**
 * @function ose_base64_decode
 *
 * @brief Decodes the input buffer from a base-64 representation.
 *
 * @long An input buffer of base-64 encoded data is decoded into its
 * binary form.
 *
 * @restriction The input buffer may only contain characters from the
 * base-64 input alphabet [A-Za-z0-9+/=]. If this is violated the
 * result of the decode operation is undefined.
 * @break
 * Also available in interrupt and timer interrupt processes.
 *
 * @field dest Pointer to the buffer that will contain the decoded
 * data. The size of this buffer should be at least (3*length/4).
 *
 * @field src Pointer to the input buffer in base-64 representation.
 *
 * @field length The length of the input buffer. This length bust be
 * evenly dividable by 4 (padding characters ('=') are added to the
 * end of the input buffer to accomplish this).
 *
 * @return The number of bytes decoded into the destination buffer.
 *
 * @seealso ose_base64_encode
 * @seealso default_restrictions_calling_functions
 *
 */
int ose_base64_decode(void *dest, const unsigned char *src, int length);




/**
 * @function ose_url_encode
 *
 * @brief Encodes the input buffer to a URL encoded representation.
 *
 * @long The URL encoding is designed to represent restricted character
 * as a seqence encoded using percent-encoding, a percent followed 
 * by two hexadecimal charaters representing the octet numeric value.
 * The encoding is defined in RFC 3986.
 *
 * @field dest Pointer to a buffer that will receive the encoded
 * data. The size of this buffer is assumed to be large enough to hold
 * the encoded data, which will require a lenght greater then 
 * the length of the source buffer.
 *
 * @field src Pointer to the buffer of data that will be encoded.
 * 
 * @field length Length of the input buffer.
 *
 * @return The number of bytes encoded into the destination buffer.
 *
 * @restriction Also available in interrupt and timer interrupt processes.
 *
 * @seealso ose_url_decode
 * @seealso default_restrictions_calling_functions
 *
 */
int ose_url_encode(char *dest, const char *src, int length);


/**
 * @function ose_url_decode
 *
 * @brief Decodes the input buffer from a URL encoded representation.
 * The encoding is defined in RFC 3986.
 * 
 * @long An input buffer of URL encoded data which will be decoded.
 *
 * @field dest Pointer to the buffer that will contain the decoded
 * data. 
 *
 * @field src Pointer to the input buffer in URL encoded representation.
 *
 * @field length Length of the input buffer.
 *
 * @return The number of bytes decoded into the destination buffer.
 *
 * @restriction Also available in interrupt and timer interrupt processes.
 *
 * @seealso ose_url_encode
 * @seealso default_restrictions_calling_functions
 *
 */
int ose_url_decode(char *dest, const char *src, int length);

/**
 * @function ose_crc16
 *
 * @declaration U16 ose_crc16 (U16 crc, unsigned char * cp, U32 length);
 *
 * @brief Compute the 16-bit CRC over a sequence of bytes.
 * @long  The generator polynomial used by the CRC-16 is:
 *        x^16 + x^12 + x^5 + 1
 *
 * @field crc The initial CRC value.
 * @field cp Pointer to sequence of bytes.
 * @field length Number of bytes in sequence.
 *
 * @return 16-bit CRC value.
 *
 * @restriction Also available in interrupt and timer interrupt processes.
 *
 * @seealso ose_crc32
 * @seealso CRC16_INIT
 * @seealso CRC16_GOOD
 * @seealso default_restrictions_calling_functions
 */
U16 ose_crc16(U16 crc, unsigned char *cp, U32 length);

/**
 * @function ose_crc32
 *
 * @declaration U32 ose_crc32 (U32 crc, unsigned char * cp, U32 length);
 *
 * @brief Compute the 32-bit CRC over a sequence of bytes.
 * @long  The generator polynomial used by the CRC-32 is:
 *        x^32+x^26+x^23+x^22+x^16+x^12+x^11+x^10+x^8+x^7+x^5+x^4+x^2+x+1
 *
 * @field crc The initial CRC value.
 * @field buf Pointer to sequence of bytes.
 * @field length Number of bytes in sequence.
 *
 * @return 32-bit CRC value.
 *
 * @restriction Also available in interrupt and timer interrupt processes.
 *
 * @seealso ose_crc16
 * @seealso CRC32_INIT
 * @seealso CRC32_GOOD
 * @seealso default_restrictions_calling_functions
 *
 * @example
 * // This example will generate the same checksum value in OSE, U-Boot and
 * // Lauterbach TRACE32.
 * //
 * // U-Boot command: crc32 0x80006000 0x8
 * // Lauterbach TRACE32 command: data.sum 0x80006000--0x80006007 /CRC32
 *
 * #include <ose_spi/checksum.h>
 *
 * U32 checksum;
 * unsigned char *addr = (unsigned char *) 0x80006000;
 * int len = 8;
 *
 * checksum = ose_crc32(CRC32_INIT, addr, len);
 * checksum ^= 0xFFFFFFFFL;
 */
U32 ose_crc32(U32 crc, const unsigned char *buf, int length);

/**
 * @function ose_hash_string
 *
 * @brief Compute a 32-bit hash value for a string.
 *
 * @field str Pointer to the null-terminated string.
 *
 * @return 32-bit hash value.
 *
 * @restriction Also available in interrupt and timer interrupt processes.
 *
 * @seealso default_restrictions_calling_functions
 */
U32 ose_hash_string(const char *str);

/**
 * @function ose_sha1_init
 *
 * @brief Initialize the SHA1 context.
 *
 * @long The ose_sha1_ functions calculate a 160-bit cryptographic checksum
 * (digest) for any number of input bytes. A cryptographic checksum is a
 * one-way hash function; that is, it is computationally impractical to find
 * the input corresponding to a particular output. The SHA1 context (SHA1_CTX)
 * is initialized via a call to ose_sha1_init. The input bytes are processed
 * with ose_sha1_update. When all data has been processed, ose_sha1_final is
 * called to compute the 160-bit digest.
 *
 * @field context Pointer to a SHA1 context.
 *
 * @restriction Also available in interrupt and timer interrupt processes.
 *
 * @seealso ose_sha1_update
 * @seealso ose_sha1_final
 * @seealso default_restrictions_calling_functions
 */
void
ose_sha1_init(SHA1_CTX *context);

/**
 * @function ose_sha1_update
 *
 * @brief Applies the SHA1 function on a number of bytes. This function may be
 * called repeatedly to process a large amount of data.
 *
 * @field context Pointer to SHA1 context.
 * @field data Pointer to a buffer of input bytes.
 * @field len Number of bytes to be processed.
 *
 * @restriction Also available in interrupt and timer interrupt processes.
 *
 * @seealso ose_sha1_init
 * @seealso ose_sha1_final
 * @seealso default_restrictions_calling_functions
 */
void
ose_sha1_update(SHA1_CTX *context, const U8 *data, U32 len);

/**
 * @function ose_sha1_final
 *
 * @brief Generate the 160-bit SHA1 digest.
 *
 * @field digest Pointer to  buffer that will receive the 160-bit SHA1 digest.
 * @field context Pointer to SHA1 context.
 *
 * @restriction Also available in interrupt and timer interrupt processes.
 *
 * @seealso ose_sha1_init
 * @seealso ose_sha1_update
 * @seealso default_restrictions_calling_functions
 */
void
ose_sha1_final(U8 digest[20], SHA1_CTX *context);

#ifdef __cplusplus
}
#endif

#endif
