/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2006 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */



/**
 * @public
 * @toc OSE:
 * @file ose_spi/demand_pager.h
 *
 * @brief Interface for Demand Pager and MM.
 *
 * @long
 * This file defines a functional interface implemented in Demand Paging
 * functionality. One user is OSE Memory Manager (MM). To avoid
 * link dependencies from MM to the optional pluggable Demand Paging
 * funtionality function pointers are used for these functions.
 *
 * @long
 * This is generation two of Demand Paging and is developed to
 * support Load Modules loaded as mmap files which are demand paged.
 * Furthermore the page faults should be able to execute in context of the
 * faulting client, possibly with disabled interrupts, but this
 * feature requires support in the underlying file systems and drivers.
 * The Pager functionality also includes a process context which is used
 * for initialization and for page cache increase.
 *
 * @library libpager.a
 *
 */


#ifndef __DEMAND_PAGER_H__
#define __DEMAND_PAGER_H__

#ifdef __cplusplus
extern "C" {
#endif

#include "ose.h"
#include "string.h"
#include "stdint.h"
#include "osetypes.h"
#include "mmtypes.h"

/**
 *
 * @type pager_status_t
 *
 * @brief Enum consisting of pager error codes.
 *
 * @field PAGER_SUCCESS
 *    A status value indicating a successful operation.
 *
 * @field PAGER_ERROR
 *    A status value indicating a generic error.
 *
 * @field PAGER_TOO_MANY_LOCKED
 *    The page was not locked as there are too many locked pages in
 *    the page cache.
 *
 * @field PAGER_CANNOT_DECREASE
 *    The pager cache cannot be decreased anymore.
 *
 * @field PAGER_REGION_NOT_REGISTERED
 *    This region has not been mmaped.
 *
 * @field PAGER_ENON_EXISTING_EVENT
 *    The event does not exist.
 *
 * @field PAGER_ENON_EXISTING_EVENT_HANDLER
 *    The event handler does not exist.
 *
 * @field PAGER_DUPLICATE_ALGORITHM
 *    A replacement algorithm with this name has already been registered.
 *
 * @field PAGER_ALGORITHM_NOT_FOUND
 *    A replacement algorithm with this name cannot be found.
 *
 * @field PAGER_PARAMETER_ERROR
 *    One or more of the input parameters are not correct.
 *
 * @field PAGER_PAGE_CACHE_NOT_FOUND
 *    The supplied page cache does not exist.
 *
 * @field PAGER_RETRY
 *    This is for internal use only.
 *
 */
typedef enum pager_status
{
   PAGER_SUCCESS = 0,
   PAGER_ERROR = 1,
   PAGER_TOO_MANY_LOCKED = 2,
   PAGER_CANNOT_DECREASE = 3,
   PAGER_REGION_NOT_REGISTERED = 4,
   PAGER_ENON_EXISTING_EVENT = 5,
   PAGER_ENON_EXISTING_EVENT_HANDLER = 6,
   PAGER_DUPLICATE_ALGORITHM = 7,
   PAGER_ALGORITHM_NOT_FOUND = 8,
   PAGER_PARAMETER_ERROR = 9,
   PAGER_PAGE_CACHE_NOT_FOUND = 10,
   PAGER_RETRY = 11
} pager_status_t;


/**
 * @type  page_cache_chooser_func_t
 *
 * @brief Callback from the pager when the first access to an mmaped region
 *        is performed and a page cache needs to be associated with that
 *        region.
 * @long  This is an optional callback which can be registered as a parameter
 *        to @link "" "ose_pager_enable", if more than one page cache will
 *        be used.
 * @long  The callback should return the identifier of the page cache, in
 *        which it wants to store the pages in this memory region. This is
 *        a number between 1 and the number of page caches
 *        that have been configured, inclusive, using system parameters
 *        krn/paging/pager/. The callback occurs in
 *        the context of the first faulting process.
 *
 * @input domain The domain this region was created in.
 * @input base   The logical base address of the region.
 * @input size   The size of the region.
 *
 * @return       Should return the identifier of the page cache that
 *               it wants to store the pages in this region. This is
 *               a number between 1 and the number of page caches
 *               that have been created, inclusive.
 *
 */
typedef uint32_t (*page_cache_chooser_func_t)(OSDOMAIN domain,
                                             log_addr_t base,
                                             size_t size);

/**
 * @function ose_pager_enable
 *
 * @brief This function is called at startup to enable Demand Paging
 *        functionality.
 *
 * @long  The pager should be enabled in the OSE start routine.
 *
 * @input chooser_func The function that picks which page cache to
 *                     associate a given mmap with. When there is
 *                     single page cache, this function is not needed and
 *                     NULL can be provided.
 *
 * @return         PAGER_SUCCESS, otherwise the error code
 *                 (@link "" "pager_status_t").
 */
pager_status_t
ose_pager_enable(page_cache_chooser_func_t chooser_func);


/**
 * @function ose_pager_disable
 *
 * @brief This function is called to disable Demand Paging
 *        functionality.
 *
 * @long  Demand Paging function pointers will be uninstalled in MM.
 *        Page fault handler will be uninstalled.
 *
 * @return         PAGER_SUCCESS, otherwise the error code
 *                 (@link "" "pager_status_t").
 *
 */
pager_status_t
ose_pager_disable(void);

/**
 *
 * @type pager_event_no_t
 *
 * @brief Enum consisting of the event numbers of the pager events.
 *
 * @field PAGER_EVENT_PAGE_IN The pager event "page in".
 *
 * @field PAGER_EVENT_PAGE_OUT The pager event "page out".
 *
 * @field PAGER_EVENT_IRQ_DISABLED_PAGE_FAULT The pager event "page fault with
 *                                            interrupts disabled".
 *
 * @field PAGER_EVENT_IRQ_PROCESS_PAGE_FAULT The pager event "page fault from
 *                                           an interrupt or timer process".
 *
 */
typedef enum pager_event_no_t
{
   PAGER_EVENT_PAGE_IN = 0,
   PAGER_EVENT_PAGE_OUT = 1,
   PAGER_EVENT_IRQ_DISABLED_PAGE_FAULT = 2,
   PAGER_EVENT_IRQ_PROCESS_PAGE_FAULT = 3
} pager_event_no_t;

/**
 * @macro PAGER_NO_OF_EVENTS
 * @brief The number of entries in the enum @link "" "pager_event_no_t".
 */
#define PAGER_NO_OF_EVENTS    4

/**
 *
 * @type pager_action_t
 *
 * @brief Enum consisting of the actions the pager should take in response
 *        to an event callback.
 *
 * @field PAGER_CONTINUE The pager should just continue.
 *
 * @field PAGER_TERMINATE_PROCESS The pager should terminate the
 *                                process that got this event.
 *
 * @field PAGER_FATAL_ERROR The pager should exit with a fatal error.
 *
 * @field PAGER_SKIP_OTHER_HANDLERS The pager should skip the rest of the
 *                                  registered handlers and then continue.
 *
 */
typedef enum pager_action_t
{
   PAGER_CONTINUE = 0,
   PAGER_TERMINATE_PROCESS = 1,
   PAGER_FATAL_ERROR = 2,
   PAGER_SKIP_OTHER_HANDLERS = 3
} pager_action_t;
#define PAGER_NO_OF_ACTIONS    4

/**
 * @type  pager_page_in_t
 *
 * @brief Event callback, called for the pager page-in event.
 *
 * @long  This function is called when the pager have paged in the data
 *        of a page to the page cache entry, before returning from the
 *        page fault.
 *        This allows e.g. a custom security check to be performed by the
 *        callback on the new data page, to decide if the data or code is
 *        acceptable or not and ask the pager to perform a suitable
 *        action.
 *
 * @seealso pager_irq_disabled_fault_t
 * @seealso pager_irq_process_fault_t
 *
 * @input domain      The domain the page belongs to.
 * @input fault_addr  The logical address of the paged in page.
 * @input data_addr   The data/text paged in can be examined by reading from
 *                    this address. For security reasons this is not the same
 *                    as fault_addr above.
 * @input size        The size of the data/text paged in. Usually a page.
 *
 * @return       Returns the action the pager should take as a response
 *               to this callback.
 *               PAGER_CONTINUE, PAGER_TERMINATE_PROCESS, PAGER_FATAL_ERROR
 *               and PAGER_SKIP_OTHER_HANDLERS are legal values.
 */
typedef pager_action_t(pager_page_in_t)(OSDOMAIN domain,
                                        log_addr_t fault_addr,
                                        log_addr_t data_addr,
                                        size_t size);

/**
 * @type  pager_page_out_t
 *
 * @brief Event callback, called for the pager page-out event.
 *
 * @long  This function is called before the pager removes a page from the
 *        page cache.
 *
 * @input domain The domain the page belongs to.
 * @input addr   The logical address of the removed page. The data/text
 *               paged in can be examined by reading from this address.
 * @input size   The size of the data/text paged out. Usually a page.
 *
 * @return       Returns the action the pager should take as a response
 *               to this callback.
 *               PAGER_CONTINUE, PAGER_TERMINATE_PROCESS, PAGER_FATAL_ERROR
 *               and PAGER_SKIP_OTHER_HANDLERS are legal values.
 */
typedef pager_action_t(pager_page_out_t)(OSDOMAIN domain,
                                         log_addr_t addr,
                                         size_t size);

/**
 * @type  pager_irq_disabled_fault_t
 *
 * @brief Event callback, called for the pager page-fault with interrupts
 *        disabled event.
 *
 * @long  This function is called when the pager gets a page fault with
 *        interrupts disabled. It is called before the corresponding
 *        @link "" "pager_page_in_t".
 *
 * @input domain The domain the page belongs to.
 * @input addr   The logical address of the page fault.
 *
 * @return       Returns the action the pager should take as a response
 *               to this callback.
 *               PAGER_CONTINUE, PAGER_TERMINATE_PROCESS, PAGER_FATAL_ERROR
 *               and PAGER_SKIP_OTHER_HANDLERS are legal values.
 */
typedef pager_action_t(pager_irq_disabled_fault_t)(OSDOMAIN domain,
                                                   log_addr_t addr);

/**
 * @type  pager_irq_process_fault_t
 *
 * @brief Event callback, called for the pager page-fault from an interrupt or
 *        timer process event.
 *
 * @long  This function is called when the pager gets a page fault from an
 *        interrupt or timer process. It is called before the corresponding
 *        @link "" "pager_page_in_t".
 *
 * @input domain The domain the page belongs to.
 * @input addr   The logical address of the page fault.
 *
 * @return       Returns the action the pager should take as a response
 *               to this callback.
 *               PAGER_CONTINUE, PAGER_TERMINATE_PROCESS, PAGER_FATAL_ERROR
 *               and PAGER_SKIP_OTHER_HANDLERS are legal values.
 */
typedef pager_action_t(pager_irq_process_fault_t)(OSDOMAIN domain,
                                                  log_addr_t addr);

/**
 *
 * @type pager_event_handler_t
 *
 * @brief A union type, describing the possible pager event callback
 *        functions.
 *        Supplied to the @link "" "ose_pager_install_handler" and
 *        @link "" "ose_pager_uninstall_handler" functions.
 *
 * @field page_in_event A page in callback function.
 *
 * @field page_out_event A page out callback function.
 *
 * @field page_irq_disabled_fault_event A page fault with interrupts disabled
 *                                      callback function.
 *
 * @field page_irq_process_fault_event A page fault from an interrupt or
 *                                     timer process callback.
 *
 * @field generic_ptr If C is too strongly typed.
 */
union pager_event_handler_t
{
   pager_page_in_t             *page_in_event;
   pager_page_out_t            *page_out_event;
   pager_irq_disabled_fault_t  *page_irq_disabled_fault_event;
   pager_irq_process_fault_t   *page_irq_process_fault_event;
   void                        *generic_ptr;
};

/**
 * @function ose_pager_install_handler
 *
 * @brief  Install an event handler to a pager event.
 *
 * @input event          The number of the event this handler should
 *                       be called for.
 * @input event_handler  The callback function.
 * @input beginning      1: Insert the handler in the beginning of the event
 *                       queue, instead of the end, which is the default (0).
 *
 * @return           PAGER_SUCCESS if ok, otherwise the pager error code
 *                   (@link "" "pager_status_t").
 */
pager_status_t
ose_pager_install_handler(pager_event_no_t event,
                          union pager_event_handler_t event_handler,
                          int beginning);


/**
 * @function ose_pager_uninstall_handler
 *
 * @brief  Uninstall an event handler for a pager event.
 *
 * @input event      The event this handler should be unistalled for.
 * @input event_handler  The callback function.
 *
 * @return           PAGER_SUCCESS if ok, otherwise the mm error code
 *                   (@link "" "pager_status_t").
 */
pager_status_t
ose_pager_uninstall_handler(pager_event_no_t event,
                            union pager_event_handler_t event_handler);

/**
 * @type page_cache_t
 *
 * @brief Page cache object.
 *
 * @long This opaque object represents one page cache.
 *       The contents of this object is private to the page cache.
 *
 */
struct page_cache_t;

/**
 * @type page_metadata_t
 *
 * @brief Page cache entry metadata object.
 *
 * @long This opaque object represents one entry in the page cache.
 *       The contents of this object is private to the page cache.
 *
 */
struct page_metadata_t;

/**
 * @type repl_alg_metadata_t
 *
 * @brief Replacement algorithm metadata object.
 *
 * @long This object is used by the replacement algorithm to store metadata
 *       associated with one instanciation of it
 * @long This object is only dereferenced by that replacement algorithm,
 *       i.e. it is the replacement algorithm that determines the contents
 *       of this object.
 *
 */
struct repl_alg_metadata_t;

/**
 * @type repl_alg_pp_metadata_t
 *
 * @brief Replacement algorithm per-page metadata object.
 *
 * @long This object is used by the replacement algorithm to store metadata
 *       associated with every page currently in the page cache.
 * @long This object is only dereferenced by that replacement algorithm,
 *       i.e. it is the replacement algorithm that determines the contents
 *       of this object. As an example, this metadata could be the
 *       priority of the process that brought this page into the page cache.
 *       The replacement algorithm could then use this data when choosing
 *       what entry to evict.
 *
 */
struct repl_alg_pp_metadata_t;


/**
 * @function find_next_to_evict_func_t
 *
 * @brief  Callback to the replacement algorithm when it should pick
 *         a page to replace from the page cache.
 *
 * @long   The function should return a reference to the page cache entry 
 *         that should be evicted. It is considered good form to test the page 
 *         cache entry with the @link "" "page_cache_entry_can_be_used"
 *         function before returning it.
 *
 * @input page_cache The page cache to evict a page from.
 * @input metadata   A reference to the metadata for this instance of
 *                   the replacement algorithm.
 * @return           A reference to the page cache entry that should be
 *                   evicted.
 */
typedef struct page_metadata_t* (*find_next_to_evict_func_t)
                                    (struct page_cache_t *page_cache,
                                     struct repl_alg_metadata_t *metadata);

/**
 * @function removed_entry_func_t
 *
 * @brief  Callback function to the replacement algorithm that is
 *         invoked every time an entry is removed from the page
 *         cache by the page cache itself.
 *
 * @long   The replacement algorithm might need to modify its metadata when
 *         this happens.
 *         This can be due to the page cache being shrunk, for example.
 *
 * @input page_cache    The page cache to evict a page from.
 * @input pgm           The page cache entry removed.
 * @input metadata      A reference to the metadata for this instance of
 *                      the replacement algorithm.
 * @input pp_metadata   A reference to the per-page metadata for this
 *                      page cache entry.
 */
typedef void (*removed_entry_func_t)(struct page_cache_t *page_cache,
                                    struct page_metadata_t *pgm,
                                    struct repl_alg_metadata_t *metadata,
                                    struct repl_alg_pp_metadata_t *pp_metadata);

/**
 * @function added_entry_func_t
 *
 * @brief  Callback function to the replacement algorithm that is
 *         invoked every time an entry is added to the page
 *         cache by the page cache itself.
 *
 * @long   The replacement algorithm might need to modify its metadata when
 *         this happens.
 *         This can be due to the page cache being expanded, for example.
 *
 * @long   The replacement algorithm can use the
 *         @link "" "added_entry_func_t" and @link "" "removed_entry_func_t"
 *         callbacks to keep track of which page cache entries exist.
 *
 * @input page_cache    The page cache that the entry is added to.
 * @input pgm           The page cache entry added.
 * @input metadata      A reference to the metadata for this instance of
 *                      the replacement algorithm.
 * @input pp_metadata   A reference to the per-page metadata for this
 *                      page cache entry. It is the algorithm's responsibility
 *                      to initialize and use the metadata area.
 */
typedef void (*added_entry_func_t)(struct page_cache_t *page_cache,
                                   struct page_metadata_t *pgm,
                                   struct repl_alg_metadata_t *metadata,
                                   struct repl_alg_pp_metadata_t *pp_metadata);

/**
 * @function install_algorithm_func_t
 *
 * @brief  Callback function to the replacement algorithm that is
 *         invoked when a new instance of it should be created.
 *
 * @long   If any metadata is needed to implement the replacement algorithm,
 *         then that metadata should be created and a pointer to it returned.
 *         Note that this does not include per-page metadata. That is
 *         created automatically if it has been requested in the
 *         @link "" "register_replacement_algorithm" function.
 *
 * @return A pointer to the metadata associated with this instance of
 *         the replication algorithm.
 */
typedef struct repl_alg_metadata_t * (*install_algorithm_func_t)(void);


/**
 * @function uninstall_algorithm_func_t
 *
 * @brief  Callback function to the replacement algorithm that is
 *         invoked when an instance is destroyed.
 *
 * @long   The callback should deallocate any metadata that was created
 *         when it was installed with @link "" "install_algorithm_func_t".
 *
 * @input metadata  A reference to the metadata for this instance of
 *                  the replacement algorithm. 
 *
 */
typedef void
(*uninstall_algorithm_func_t)(struct repl_alg_metadata_t *metadata);


/**
 * @function register_replacement_algorithm
 *
 * @brief  Registers a new replacement algorithm with the pager.
 *
 * @long   These have to be registered before @link "" "ose_pager_enable"
 *         is called.
 *
 * @input name                   The name of the replacement algorithm.
 *                               Mandatory.
 * @input find_next_to_evict     The @link "" "find_next_to_evict_func_t"
 *                               callback. Mandatory.
 * @input added_entry            The @link "" "added_entry_func_t" callback.
 *                               Optional.
 * @input removed_entry          The @link "" "removed_entry_func_t" callback.
 *                               Optional.
 * @input install_algorithm      The @link "" "install_algorithm_func_t" callback.
 *                               Optional.
 * @input uninstall_algorithm    The @link "" "uninstall_algorithm_func_t" callback.
 *                               Optional.
 * @input per_page_metadata_size The size of any per_page metadata that is
 *                               required by the replacement algorithm, or
 *                               zero if not needed.
 *
 * @return           PAGER_SUCCESS if ok, otherwise the pager error code
 *                   (@link "" "pager_status_t").
 */
pager_status_t
register_replacement_algorithm(const char *name,
                               find_next_to_evict_func_t find_next_to_evict,
                               added_entry_func_t added_entry,
                               removed_entry_func_t removed_entry,
                               install_algorithm_func_t install_algorithm,
                               uninstall_algorithm_func_t uninstall_algorithm,
                               uint32_t per_page_metadata_size);

/**
 * @function page_cache_entry_can_be_used
 *
 * @brief  Checks if a page cache entry can be used or not.
 *
 * @long   It is considered good form for the replacement algorithm to check
 *         an entry before it is returned by the
 *         @link "" "find_next_to_evict_func_t" callback function.
 *         It avoids some cases where the choosen cache entry cannot be
 *         used, but the algorithm can anyhow get a new evict request if the
 *         pager finds that the choosen cache entry cannot be used.
 *
 * @input pgm        The page cache entry to check.
 *
 * @return           Returns one if it can be used, zero otherwise.
 */
int
page_cache_entry_can_be_used(struct page_metadata_t *pgm);


/**
 * @function get_next_metadata
 *
 * @brief  Returns the next page cache entry according to a FIFO order.
 *
 * @long   It will never return NULL.
 * 
 * @long   This is the default replacement algorithm, used if no
 *         algorithm is registered. A registered algorithm could also
 *         use this function, but should instead use the
 *         @link "" "added_entry_func_t" and corresponding
 *         @link "" "removed_entry_func_t" callbacks to keep track of available
 *         page cache entries.
 *
 * @input pg The page cache.
 *
 * @return           The next page cache entry.
 */
struct page_metadata_t*
get_next_metadata(struct page_cache_t *pg);


/**
 * @macro PAGER_HELPER_NAME
 *
 * @brief The name of the pager helper process that you can use to
 *        retrieve statistics.
 *
 * @long The name defined here is ose_pager_helper.
 */
#define PAGER_HELPER_NAME "ose_pager_helper"

/**
 * @signal PAGER_GET_STATISTICS_REQUEST
 *
 * @brief Signal to the pager helper process, used to get pager statistics.
 *
 * @member pager_get_statistics
 */
#define PAGER_GET_STATISTICS_REQUEST 37990

/**
 * @signal PAGER_GET_STATISTICS_REPLY
 *
 * @brief Reply signal from the pager helper process, replying to
 *        @link "demand_pager.h" "PAGER_GET_STATISTICS_REQUEST".
 *
 * @member pager_get_statistics
 */
#define PAGER_GET_STATISTICS_REPLY   37991

/**
 * @type pager_get_statistics
 *
 * @field sig_no Signal number.
 * @input page_cache       The page cache identifier or zero for the
 *                         default one.
 * @output status          Status of the operation,
 * @output used_bytes      The number of bytes used.
 * @output allocated_bytes The number of bytes currently allocated by
 *                         the page cache.
 * @output freeable_bytes  The number of bytes that the page cache can
 *                         potentially free.
 * @output no_of_expansions   The number of times the page cache has
 *                            expanded.
 * @output no_of_contractions The number of times the page cache has
 *                            been contracted.
 * @output no_of_misses       Number of page cache misses.
 * @output no_of_locked_pages Number of locked pages.
 * @output no_of_invalidates  Number of pages that have been invalidated.
 *                            This is mainly due to rearrangements in the
 *                            physical media itself when a new address
 *                            is not available.
 * @output no_of_remaps       Number of pages that have been remapped.
 *                            This is mainly due to rearrangements in the
 *                            physical media itself.
 * @output no_of_updates      Number of pages that have been invalidated.
 *                            This happens when there are simultaneous
 *                            file system writes to an mmaped file. 
 */
struct pager_get_statistics
{
   SIGSELECT      sig_no;
   uint32_t       page_cache;
   pager_status_t status;
   uint32_t       used_bytes;
   uint32_t       allocated_bytes;
   uint32_t       freeable_bytes;
   uint32_t       no_of_expansions;
   uint32_t       no_of_contractions;
   uint32_t       no_of_misses;
   uint32_t       no_of_locked_pages;
   uint32_t       no_of_invalidates;
   uint32_t       no_of_remaps;
   uint32_t       no_of_updates;
};

/**
 * @signal PAGER_CLEAR_STATISTICS_REQUEST
 *
 * @brief Signal used to clear pager statistics.
 *
 * @long Note that statistics vital for the correct function of the
 * pager are not cleared. The signal shall be sent to the pager
 * helper process.
 *
 * @member pager_clear_statistics
 */
#define PAGER_CLEAR_STATISTICS_REQUEST 37992

/**
 * @signal PAGER_CLEAR_STATISTICS_REPLY
 *
 * @brief Reply signal from the pager helper process, replying to
 *        @link "demand_pager.h" "PAGER_CLEAR_STATISTICS_REQUEST".
 *
 * @member pager_clear_statistics
 */
#define PAGER_CLEAR_STATISTICS_REPLY   37993

/**
 * @type pager_clear_statistics
 *
 * @field sig_no Signal number.
 * @input page_cache The page cache identifier or zero for the default one.
 * @output status Status of the operation,
 */
struct pager_clear_statistics
{
   SIGSELECT      sig_no;
   uint32_t       page_cache;
   pager_status_t status;
};

/**
 * @function pager_allocated_memory
 *
 * @brief  Returns the number of bytes allocated by the page cache.
 *
 * @long   This is a fast function call.
 *
 * @input page_cache The page cache identifier or 0 for the default one.
 *
 * @return           Number of bytes allocated by the page cache.
 */
uint32_t
pager_allocated_memory(uint32_t page_cache);

/**
 * @function pager_used_memory
 *
 * @brief  Returns the number of bytes currently in use in the page cache.
 *
 * @long   This is always less than or equal to the size returned by
 *         @link "" "pager_allocated_memory".
 *         This is a fast function call.
 *
 * @input page_cache The page cache identifier or 0 for the default one.
 *
 * @return           Number of bytes currently in use in the page cache.
 */
uint32_t
pager_used_memory(uint32_t page_cache);

/**
 * @function pager_freeable_memory
 *
 * @brief  Returns the number of bytes that the page cache can potentially
 *         free.
 *
 * @long   This number depends on the current size and the number
 *         of mlocked pages.
 *         This is a fast function call.
 *
 * @input page_cache The page cache identifier or 0 for the default one.
 *
 * @return           Number of bytes that the page cache can potentially
 *                   free.
 */
uint32_t
pager_freeable_memory(uint32_t page_cache);


#ifdef __cplusplus
}
#endif

#endif /* __DEMAND_PAGER_H__ */
