/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2006 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */
/**
 * @public
 *
 * @toc OSE:
 *
 * @file ose_spi/efmi.h
 *
 * @brief Executive Freeze-Mode Interface
 *
 * @long
 * This file declares the Executive Freeze-Mode Interface.
 * It may be used to fetch information from the executive in a freeze-mode
 * environment. The purpose is to make OSE Executive concepts accessible
 * for freeze-mode debug monitors that can not use ordinary system calls.
 * @long
 * WARNING: Use of EFMI functionality for any other purpose will severely
 * damage the interrupt latency!
 *
 */

#ifndef _EFMI_H
#define _EFMI_H


#include "stddef.h"
#include "ose.h"
#include "osetypes.h"

#ifdef __cplusplus
extern "C" {
#endif

#if defined(HOST_TOOL_ENVIRONMENT)
#include "host_fmi.h"
#endif

#define EFMI_MAX_FILE_NAMELEN 32 /* Defined in pose.con as MAX_FILE_NAMELEN. */

/* 8 sizes is max according to kernel reference manual. */
#define EFMI_MAX_SIGNAL_BUFFER_SIZE_COUNT 8
#define EFMI_MAX_STACK_BUFFER_SIZE_COUNT  8

#define EFMI_MAX_WANTED_SIGNALS 16 /* Longer arrays are silently truncated. */
#define EFMI_CURSOR_SIZE        (24 / sizeof(U32))

#define EFMI_MAX_BASELINE_LENGTH 64

#define EFMI_ENDIAN_VALUE 0x01020304UL

typedef enum EfmiOSType
{
   EFMI_OS_OSE_5         = 0,
   EFMI_OS_OSE_4,
   EFMI_OS_OSE_EPSILON,
   EFMI_OS_OSE_CK,

   EFMI_OS_UNKNOWN = 0x7fffffff
} EfmiOSType;

typedef enum EfmiCPUType
{
   EFMI_CPU_PPC   = 0,
   EFMI_CPU_ARM,
   EFMI_CPU_MIPS,
   EFMI_CPU_X86,
   EFMI_CPU_SPARC,

   EFMI_CPU_C64X = 0x10000000,
   EFMI_CPU_C64XP,
   EFMI_CPU_SP26XX,
   EFMI_CPU_SP27XX,
   EFMI_CPU_TL3,
   EFMI_CPU_CEVAX,
   EFMI_CPU_M8144,
   EFMI_CPU_M8156,

   EFMI_CPU_UNKNOWN = 0x7fffffff
} EfmiCPUType;

/* Return status of functions */
/**
 *  @type EfmiStatus
 *
 *  @brief enum EfmiStatus defines the set of return values from EFMI
 *         functions.
 *
 *  @field EFMI_OK                  Operation was successful.
 *  @field EFMI_UNAVAILABLE         Requested information not found.
 *  @field EFMI_END_OF_LIST         No more objects found.
 *  @field EFMI_RESULT_TRUNCATED    A list or string was truncated.
 *  @field EFMI_POOL_CORRUPTED      Error found in a non-system pool.
 *  @field EFMI_KERNEL_CORRUPTED    Error found in kernel data structures.
 *  @field EFMI_KERNEL_UNINITIATED  The kernel data structures are not yet
 *                                  initiated.
 *  @field EFMI_CALL_INVALID        Operation preconditions were not met.
**/
typedef enum EfmiStatus
  {
    EFMI_OK                 = 0,
    EFMI_UNAVAILABLE        = 1,
    EFMI_END_OF_LIST        = 2,
    EFMI_RESULT_TRUNCATED   = 3,
    EFMI_POOL_CORRUPTED     = 4,
    EFMI_KERNEL_CORRUPTED   = 5,
    EFMI_KERNEL_UNINITIATED = 6,
    EFMI_CALL_INVALID       = 7
  } EfmiStatus;

struct EfmiSystemInfo
{
   EfmiOSType os_type;
   char baseline[EFMI_MAX_BASELINE_LENGTH];
   EfmiCPUType cpu_type;
   U8 endian[4];
};

/**
 *  @type EfmiSegmentInfo
 *
 *  @brief struct EfmiSegmentInfo contains information on one segment.
 *  @field segment     Segment identity.
 *  @field segid       Domain identity.
**/
struct EfmiSegmentInfo
{
  U32	segment;		/* ose.h PROCESS. */
  U32	segid;			/* ose.h OSSEGMENT. */
};

/**
 *  @type EfmiBlockInfo
 *
 *  @brief struct EfmiBlockInfo contains information on a block.
 *
 *  @field block        Block identity.
 *  @field user         User number of the block.
 *  @field segment      The identity of the segment where the block belongs.
 *  @field sig_pool     Identity of the signal pool.
 *  @field stack_pool   Identity of the stack pool.
 *  @field supervisor_mode Non-zero for supervisor mode blocks.
 *  @field remcall_server Identity of the remote call server, zero if none.
 *  @field cpu_id       Identity of the CPU for a SMP core, otherwise zero.
**/
struct EfmiBlockInfo
{
  U32	block;			/* ose.h PROCESS. */
  U32	user;			/* ose.h OSUSER. */
  U32	segment;		/* ose.h PROCESS. */
  U32	sig_pool;		/* ose.h PROCESS. */
  U32	stack_pool;		/* ose.h PROCESS. */
  U32	supervisor_mode;
  U32	remcall_server;		/* ose.h PROCESS. */
  U32	cpu_id;
};

/**
 *  @type EfmiStackInfo
 *
 *  @brief struct EfmiStackInfo contains information on a stack.
 *  @field top          The lowest address beyond the stack
 *  @field limit        The lowest address inside the stack. If top and limit
 *                      are equal, the stack is not in use.
 *                      The size is top - limit.
 *  @field no_endmark   Set non-zero if the endmark is overwritten.
 *  @seealso EfmiProcessStatistics
 *  @seealso EfmiProcessInfo
**/
struct EfmiStackInfo
{
  U32	top;			/* ose.h OSADDRESS. */
  U32	limit;			/* ose.h OSADDRESS. size == top - limit */
  U32	no_endmark;
};

/**
 *  @type EfmiProcessInfo
 *
 *  @brief struct EfmiProcessInfo contains information on a process.
 *  @field process      Process identity.
 *  @field user         User number.
 *  @field block        Block identity.
 *  @field parent       Parent process identity identifies the creator.
 *  @field entry_point  Process entry point.
 *  @field type         Process type (see enum PROCESS_TYPE).
 *  @field priority     Process priority.
 *  @field status       Process status (see enum PROCESS_STATUS).
 *  @field fsemvalue    Fast semaphore value.
 *  @field sig_cnt_owned  Number of owned signals.
 *  @field line         Source code line number when pending in a system call,
 *                      else zero.
 *  @field process_stack Description of the main stack of the process.
 *  @field supervisor_stack Description of the supervisor stack of a
 *                      user process.
 *  @field wanted       Wanted signals when pending in a receive() call.
 *                      The wanted array is a copy of the SIGSELECT array,
 *                      silently truncated to EFMI_MAX_WANTED_SIGNALS values.
 *                      Only significant when bit OS_RECEIVE is set in status.
 *  @field file         Source file name when pending in a system call,
 *                      else an empty string.
 *  @field cpu_id       Identity of the CPU for a SMP core, otherwise zero.
 *  @seealso EfmiProcessStatistics
 *  @seealso EfmiStackInfo
**/
struct EfmiProcessInfo
{
  U32	process;		/* ose.h PROCESS. */
  U32	user;			/* ose.h OSUSER. */
  U32	block;			/* ose.h PROCESS. */
  U32	parent;			/* ose.h PROCESS. */
  U32	entry_point;		/* ose.h OSADDRESS. */
  U32	type;			/* ose.h enum PROCESS_TYPE. */
  U32	priority;		/* ose.h OSPRIORITY. */
  U32	status;			/* ose.h enum PROCESS_STATUS. */
  S32	fsemvalue;		/* ose.h OSFSEMVAL. */
  U32	sig_cnt_owned;
  U32	line;
  struct EfmiStackInfo process_stack; /* For user and supervisor process. */
  struct EfmiStackInfo supervisor_stack; /* Significant for user process. */
  SIGSELECT wanted[EFMI_MAX_WANTED_SIGNALS];
  char	file[EFMI_MAX_FILE_NAMELEN];
  U32	cpu_id;
};

/**
 *  @type EfmiProcessTag
 *
 *  @brief enum EfmiProcessTag defines process statistics available by
 *         ose_efmi_get_process_statistics()
 *  @field EFMI_SIG_CNT_IN_Q             Number of signals in the receive queue
 *  @field EFMI_PROCESS_STACK_USAGE      Max # bytes used in main stack
 *  @field EFMI_SUPERVISOR_STACK_USAGE   Max # bytes used in supervisor stack
 *                                       for a user mode process
 *  @seealso EfmiProcessStatistics
 *  @seealso ose_efmi_get_process_statistics
**/
enum EfmiProcessTag
  {
    EFMI_SIG_CNT_IN_Q,		/* Number of signals in the receive queue */
    EFMI_PROCESS_STACK_USAGE,	/* Max # bytes used in main stack */
    EFMI_SUPERVISOR_STACK_USAGE	/* Max # bytes used in supervisor stack for a
				 * user mode process */
  };

/**
 *  @type EfmiProcessStatistics
 *
 *  @brief struct EfmiProcessStatistics describes process statistics fetched by
 *         ose_efmi_get_process_statistics()
 *  @field value          The value of the selected statistics item.
 *  @seealso EfmiProcessTag
 *  @seealso ose_efmi_get_process_statistics
**/
struct EfmiProcessStatistics
{
  U32	value;
};

/**
 *  @type EfmiBufSizeInfo
 *
 *  @brief struct EfmiBufSizeInfo presents statistics per size class on stack
 *         and signal use in a pool.
 *  @field configured_size Size of this class of stacks and signals.
 *  @field allocated  Current number of allocated objects of this size class.
 *  @field free Current number of free objects of this size class.
 *  @seealso EfmiPoolInfo
**/
struct EfmiBufSizeInfo
{
  U32	configured_size;
  U32	allocated;
  U32	free;
};

/**
 *  @type EfmiPoolInfo
 *
 *  @brief struct EfmiPoolInfo describes a pool.
 *  @field pool_id             Pool identity.
 *  @field fragment_count      Number of memory fragments.
 *  @field never_used          Size of never used memory in the pool.
 *  @field signal_buf_sizes    Statistics per signal size class.
 *  @field stack_buf_sizes     Statistics per stack size class.
 *  @field segment             The identity of the segment where the pool
 *                             belongs (Cast to PROCESS).
 *  @field stack_alignment     The address alignment of any stack in the pool.
 *  @field stack_admin_size    Difference between actual size of a stack buffer
 *                             in the pool and the corresponding configured
 *                             buffer size. For OSE5 this is 0.
 *  @field stack_internal_admin_size  Difference between the size of a stack
 *                                    and the size of the area that should be
 *                                    used by the application. For OSE 5 this
 *                                    is 1 (endmark).
 *                                
 *  @field signal_alignment    The address alignment of any signal in the pool.
 *  @field signal_admin_size   The size of the signal control block.
 *  @field signal_internal_admin_size Difference between the size of a signal
 *                                    and the size of the area that should be
 *                                    used by the application. For OSE5 this
 *                                    is 0.
 *                             
 *  @seealso EfmiBufSizeInfo
 *  @seealso EfmiFragmentInfo
**/
struct EfmiPoolInfo
{
  U32	pool_id;		/* ose.h PROCESS */
  U32	fragment_count;
  U32	never_used;
  struct EfmiBufSizeInfo signal_buf_sizes[EFMI_MAX_SIGNAL_BUFFER_SIZE_COUNT];
  struct EfmiBufSizeInfo stack_buf_sizes[EFMI_MAX_STACK_BUFFER_SIZE_COUNT];
  U32	segment;
  U16   stack_alignment;
  U8    stack_admin_size;
  U8    stack_internal_admin_size;
  U16   signal_alignment;
  U8    signal_admin_size;
  U8    signal_internal_admin_size;
};


/**
 *  @type EfmiFragmentInfo
 *
 *  @brief struct EfmiFragmentInfo holds statistics per pool fragment.
 *  @field base_address  Base address of the fragment.
 *  @field size          Fragment size in bytes.
 *  @field used_for_signals Number of bytes used for signal buffers. For the
 *  system pool, this figure includes most control blocks used by the system.
 *  Some of these are visible as allocated signals, and some are not.
 *  @field used_for_stacks Number of bytes used for stacks.
 *  @seealso EfmiPoolInfo
**/
struct EfmiFragmentInfo
{
  U32	base_address;		/* ose.h OSADDRESS. */
  U32	size;
  U32	used_for_signals;
  U32	used_for_stacks;
};


/**
 *  @type EfmiSignalStatus
 *
 *  @brief enum EfmiSignalStatus reports status of a signal investigation
 *  @field EFMI_SS_VALID                The signal is valid.
 *  @field EFMI_SS_NO_ENDMARK           The signal has a broken endmark.
 *  @field EFMI_SS_BROKEN_SIGADM        The signal control block is broken.
 *  @field EFMI_SS_WILD_POINTER_INSIDE_POOL The signal pointer does refer into
 *                                      a pool, but not to a signal buffer.
 *  @field EFMI_SS_WILD_POINTER_OUTSIDE_POOL The signal pointer does not refer
 *                                       to any signal pool.
 *
 *  @seealso EfmiSignalInfo
**/
enum EfmiSignalStatus
  {
    EFMI_SS_VALID = 1,
    EFMI_SS_NO_ENDMARK,
    EFMI_SS_BROKEN_SIGADM,
    EFMI_SS_WILD_POINTER_INSIDE_POOL,
    EFMI_SS_WILD_POINTER_OUTSIDE_POOL
  };


/**
 *  @type EfmiSignalInfo
 *
 *  @brief struct EfmiSignalInfo presents information on a signal
 *  @field signal_pointer       Pointer to the signal.
 *  @field sig_no               The signal number.
 *  @field owner                The process identity of the owner, or zero
 *                              if the signal is owned by the system.
 *  @field sender_pid           The process identity of the sender, or zero
 *                              if the signal has not been sent.
 *  @field addressee_pid        The process identity of the addressee.
 *  @field sig_size             The signal size requested.
 *  @field size_in_pool         Total size of the signal, including alignment
 *                              and overhead.
 *  @field sig_status           Status value from enum EfmiSignalStatus.
 *                              When sig_status > EFMI_SS_NO_ENDMARK,
 *                              searching for next signal will probably fail.
 *  @seealso  EfmiSignalStatus
**/
struct EfmiSignalInfo
{
  U32	signal_pointer;		/* ose.h OSADDRESS. */
  U32	sig_no;			/* ose.h SIGSELECT. */
  U32	owner;			/* ose.h PROCESS. */
  U32	sender_pid;		/* ose.h PROCESS. */
  U32	addressee_pid;		/* ose.h PROCESS. */
  U32	sig_size;
  U32	size_in_pool;
  U32	sig_status;
};


/**
 *  @type EfmiRegisterInfo
 *
 *  @brief struct EfmiRegisterInfo reports a register value for a process.
 *  @field value The register value.
**/
struct EfmiRegisterInfo
{
  U32	value;
};


/**
 *  @type EfmiCurrent
 *
 *  @brief struct EfmiCurrent reports identification information of the
 *  currently running process.
 *
 *  @field process      Identity of the current process, cast to (PROCESS).
 *  @field block        Identity of the current block, cast to (PROCESS).
 *  @field segment      Identity of the current segment, cast to (PROCESS).
 *  @field segid        Identity of the current domain, cast to (OSSEGMENT).
 *  @field type         Type of the current process, cast to
 *                      (enum PROCESS_TYPE).
 *  @field ni_process   Identity of the current non-interrupt process,
 *                      cast to (PROCESS).
 *  @field cpu_id       Identity of the CPU for a SMP core, otherwise zero.
**/
struct EfmiCurrent
{
  U32	process;		/* ose.h PROCESS. */
  U32	block;			/* ose.h PROCESS. */
  U32	segment;		/* ose.h PROCESS. */
  U32	segid;			/* ose.h OSSEGMENT. */
  U32	type;			/* ose.h enum PROCESS_TYPE. */
  U32	ni_process;		/* ose.h PROCESS. */
  U32	cpu_id;
};


/**
 *  @type EfmiConfigurationTag
 *
 *  @brief enum EfmiConfigurationTag is used to request a value from the
 *  configuration of the kernel.
 *  @field EFMI_CONF_TRACESIZE Number of bytes reserved for the trace in the
 *  pool.
 *  @field EFMI_CONF_SUPERV_STACK_SIZE Extra space for superv processes.
 *  @field EFMI_CONF_MAX_PROCS Contains max number of allowed processes.
 *  @field EFMI_CONF_MAX_ATTACHED Contains max number of attached signals.
 *  @field EFMI_CONF_TICK_LEN Tick length in microseconds.
 *  @field EFMI_CONF_RESIDENT_PRI Virtual MEM disabled at this priority and up.
 *  @field EFMI_CONF_ALLOW_USER_INTPROC Non-zero to allow user mode processes to
 *  create interrupt processes.
 *  @field EFMI_CONF_ALLOW_ALLOC_NIL Non-zero to allow alloc to return NIL upon
 *  error handler exit for non-system pools.
 *  @field EFMI_CONF_SKIP_STACK_ALLOCATION Non-zero to pass NULL in the stack to
 *  zzcreate_context().
 *  @field EFMI_CONF_LINKHANDLER_TIMEOUT Non-zero enables link handler response
 *  timeout error and sets the time (seconds).
 *  @field EFMI_CONF_ALLOW_POOL_COPY Non-zero enables inter domain copying when
 *  sending between pools.
 *  @field EFMI_CONF_TABLE_EXTENDED Indicates the present table extension
 *         level.
 *
 *  @field EFMI_CONF_LDM_PRIORITY The priority of the ose_ldm/ process.
 *  @field EFMI_CONF_HUNTD_PRIORITY The priority of the ose_huntd process.
 *  @field EFMI_CONF_MAIN_PRIORITY The priority of the zzmain process.
 *  @field EFMI_CONF_SYSD_PRIORITY The priority of the ose_sysd process.
 *  @field EFMI_CONF_ENABLE_GET_PCB_REGS Non-zero to enable registers in
 *         get_pcb().
 *  @field EFMI_CONF_TICKD_PRIORITY The priority of the ose_tickd process.
 *  @field EFMI_CONF_BGD_TIMESLICE The default timeslice for background
 *  processes.
 *  @field EFMI_CONF_ENABLE_REMCALLS Flag to enable remote system calls.
 *  @field EFMI_CONF_NTICK_FREQ Frequency of the ntick timer.
 *  @seealso EfmiConfigurationInfo
**/
enum EfmiConfigurationTag
  {
    EFMI_CONF_TRACESIZE,	/* Number of bytes reserved for the trace in */
                                /* the pool. */
    EFMI_CONF_SUPERV_STACK_SIZE, /* Extra space for superv processes. */
    EFMI_CONF_MAX_PROCS,	/* Contains max number of allowed processes. */
    EFMI_CONF_MAX_ATTACHED,	/* Contains max number of attached signals. */
    EFMI_CONF_TICK_LEN,		/* Tick length in microseconds. */
    EFMI_CONF_RESIDENT_PRI,	/* Virtual MEM disabled at this priority and up. */

    EFMI_CONF_ALLOW_USER_INTPROC, /* Non-zero to allow user mode proc to create
				   *  intprocs. */
    EFMI_CONF_ALLOW_ALLOC_NIL,	/* Non-zero to allow alloc to return NIL upon
				 * error handler exit for non-system pools. */
    EFMI_CONF_SKIP_STACK_ALLOCATION, /* Non-zero to pass NULL in the stack to
				      * zzcreate_context(). */
    EFMI_CONF_LINKHANDLER_TIMEOUT, /* Non-zero enables linkh response timeout
				    *  error and sets the time (seconds). */
    EFMI_CONF_ALLOW_POOL_COPY,	/* Non-zero enables inter domain copying when
				 * sending between pools. */
    EFMI_CONF_TABLE_EXTENDED,	/* Indicates the present table extension */
                                /* level. */

    EFMI_CONF_LDM_PRIORITY,	/* The priority of the ose_ldm/ process. */
    EFMI_CONF_HUNTD_PRIORITY,	/* The priority of the ose_huntd process. */
    EFMI_CONF_MAIN_PRIORITY,	/* The priority of the zzmain process. */
    EFMI_CONF_SYSD_PRIORITY,	/* The priority of the ose_sysd process. */
    EFMI_CONF_ENABLE_GET_PCB_REGS, /* Non-zero to enable registers in get_pcb(). */
    EFMI_CONF_TICKD_PRIORITY,	/* The priority of the ose_tickd process. */
    EFMI_CONF_BGD_TIMESLICE,	/* The default timeslice for background
				 * processes. */
    EFMI_CONF_ENABLE_REMCALLS,	/* Flag to enable remote system calls. */
    EFMI_CONF_NTICK_FREQ        /* Frequency of the ntick timer. */
  };


/**
 *  @type EfmiConfigurationInfo
 *
 *  @brief struct EfmiConfigurationInfo presents a selected item from
 *  the configuration of the kernel.
 *  @field value The value of a kernel configuration item.
 *  @seealso EfmiConfigurationTag
**/
struct EfmiConfigurationInfo
{
  U32	value;
};


/**
 *  @type EfmiSegmentCursor
 *
 *  @brief EFMI Cursor data types
 *  @long Variables of struct types EfmiSegmentCursor, EfmiBlockCursor,
 *  EfmiProcessCursor, EfmiEnvVarCursor, EfmiPoolCursor, EfmiFragmentCursor,
 *  EfmiSignalCursor, EfmiProcNameCursor and EfmiCurrentCursor are provided
 *  by the caller of EFMI functions. They are used by EFMI for internal
 *  book keeping.
 *  @field cursor For internal use by EFMI.
**/
struct EfmiSegmentCursor        { U32 cursor[EFMI_CURSOR_SIZE]; };
struct EfmiBlockCursor          { U32 cursor[EFMI_CURSOR_SIZE]; };
struct EfmiProcessCursor        { U32 cursor[EFMI_CURSOR_SIZE]; };
struct EfmiEnvVarCursor         { U32 cursor[EFMI_CURSOR_SIZE]; };
struct EfmiPoolCursor           { U32 cursor[EFMI_CURSOR_SIZE]; };
struct EfmiFragmentCursor       { U32 cursor[EFMI_CURSOR_SIZE]; };
struct EfmiSignalCursor         { U32 cursor[EFMI_CURSOR_SIZE]; };
struct EfmiProcNameCursor       { U32 cursor[EFMI_CURSOR_SIZE]; };
struct EfmiCurrentCursor        { U32 cursor[EFMI_CURSOR_SIZE]; };


/**
 *  @function ose_efmi_get_status
 *
 *  @brief Get kernel status.
 *  @long Check if the OSE kernel is detected and in such a state that the
 *        information can be retrieved.
 *  @long This function is only available in the kernel image.
 *     It must be called in supervisor mode.
 *  @long EFMI functions are designed for freeze-mode debug purposes. Any other
 *     use will destroy the interrupt latency of the system.
 *  @return EFMI_OK, EFMI_KERNEL_CORRUPTED or EFMI_KERNEL_UNINITIATED.
 *  See enum EfmiStatus.
 *  @error See enum EfmiStatus.
 *  @long EFMI functions are not part of the real-time services.
 *  @seealso EfmiStatus
**/
#ifdef HOST_TOOL_ENVIRONMENT
EfmiStatus
ose_efmi_get_status(HostFmiInstance *i);
#else
EfmiStatus
ose_efmi_get_status(void);
#endif

#ifdef HOST_TOOL_ENVIRONMENT
EfmiStatus
ose_efmi_get_system_info(HostFmiInstance *i,
			 struct EfmiSystemInfo *system_info,
			 size_t system_info_size);
#else
EfmiStatus
ose_efmi_get_system_info(struct EfmiSystemInfo *system_info,
                         size_t system_info_size);
#endif
/**
 *  @function ose_efmi_get_current
 *
 *  @brief Get the current process, block and segment identity.
 *  @long This function is only available in the kernel image.
 *     It must be called in supervisor mode.
 *  @long EFMI functions are designed for freeze-mode debug purposes. Any other
 *     use will destroy the interrupt latency of the system.
 *  @input current_size The number of bytes to write in the current struct.
 *  @output current Information on the currently running context.
 *  @return This function performs very few tests of the kernel data
 *  structures. It may still report failure if the current process control
 *  block or the system pool descriptor can not be found or is severely damaged.
 *  An obvious example is when the kernel is not yet initiated. See enum
 *  EfmiStatus.
 *  @error See enum EfmiStatus
 *  @long EFMI functions are not part of the real-time services.
 *  @seealso EfmiCurrent
 *  @seealso EfmiStatus
**/
#ifdef HOST_TOOL_ENVIRONMENT
EfmiStatus
ose_efmi_get_current(HostFmiInstance *i,
		     struct EfmiCurrent *current,
		     size_t current_size);
#else
EfmiStatus
ose_efmi_get_current(struct EfmiCurrent *current,
                     size_t current_size);
#endif

/**
 *  @function ose_efmi_get_segment_info
 *
 *  @brief Get information on a segment
 *  @long This function is only available in the kernel image.
 *        It must be called in supervisor mode.
 *  @long EFMI functions are designed for freeze-mode debug purposes. Any other
 *        use will destroy the interrupt latency of the system.
 *  @input segment      Segment identity.
 *  @input pri_size     The wanted number of bytes in the output struct.
 *  @output pri         Segment information.
 *  @return See enum EfmiStatus
 *  @error See enum EfmiStatus
 *  @long EFMI functions are not part of the real-time services.
 *  @seealso EfmiSegmentInfo
 *  @seealso EfmiStatus
**/
#ifdef HOST_TOOL_ENVIRONMENT
EfmiStatus
ose_efmi_get_segment_info(HostFmiInstance *i,
			  PROCESS segment,
			  struct EfmiSegmentInfo   *pri,
			  size_t pri_size);
#else
EfmiStatus
ose_efmi_get_segment_info(PROCESS segment,
                          struct EfmiSegmentInfo   *pri,
                          size_t pri_size);
#endif
/**
 *  @function ose_efmi_get_block_info
 *
 *  @brief Get information on a block.
 *  @long This function is only available in the kernel image.
 *        It must be called in supervisor mode.
 *  @long EFMI functions are designed for freeze-mode debug purposes. Any other
 *        use will destroy the interrupt latency of the system.
 *  @input block_id     Block identity.
 *  @input bi_size      The wanted number of bytes in the output struct.
 *  @output bi          Block information.
 *  @return See enum EfmiStatus
 *  @error See enum EfmiStatus
 *  @long EFMI functions are not part of the real-time services.
 *  @seealso EfmiBlockInfo
 *  @seealso EfmiStatus
**/
#ifdef HOST_TOOL_ENVIRONMENT
EfmiStatus
ose_efmi_get_block_info(HostFmiInstance *i,
			PROCESS block_id,
			struct EfmiBlockInfo   *bi,
			size_t bi_size);
#else
EfmiStatus
ose_efmi_get_block_info(PROCESS block_id,
                        struct EfmiBlockInfo   *bi,
                        size_t bi_size);
#endif
/**
 *  @function ose_efmi_get_process_info
 *
 *  @brief Get process information.
 *  @long This function is only available in the kernel image.
 *        It must be called in supervisor mode.
 *  @long EFMI functions are designed for freeze-mode debug purposes. Any other
 *        use will destroy the interrupt latency of the system.
 *  @input pid            Process identity.
 *  @input proc_info_size The wanted number of bytes in the output struct.
 *  @output proc_info     The output struct.
 *  @return See enum EfmiStatus
 *  @error See enum EfmiStatus
 *  @long EFMI functions are not part of the real-time services.
 *  @seealso EfmiProcessInfo
 *  @seealso EfmiStatus
**/
#ifdef HOST_TOOL_ENVIRONMENT
EfmiStatus
ose_efmi_get_process_info(HostFmiInstance *i,
			  PROCESS pid,
			  struct EfmiProcessInfo  *proc_info,
			  size_t proc_info_size);
#else
EfmiStatus
ose_efmi_get_process_info(PROCESS pid,
                          struct EfmiProcessInfo  *proc_info,
                          size_t proc_info_size);
#endif
/**
 *  @function ose_efmi_get_process_statistics
 *
 *  @brief Get selected statistics for a process.
 *  @long This function is only available in the kernel image.
 *     It must be called in supervisor mode.
 *  @long EFMI functions are designed for freeze-mode debug purposes. Any other
 *     use will destroy the interrupt latency of the system.
 *  @input pid          Process identity.
 *  @input tag          Statistics selection tag.
 *  @output statistics  Result structure, with the value representing
 *  the tag.
 *  @return See enum EfmiStatus
 *  @error See enum EfmiStatus
 *  @long EFMI functions are not part of the real-time services.
 *  @seealso EfmiProcessTag
 *  @seealso EfmiProcessStatistics
 *  @seealso EfmiStatus
**/
#ifdef HOST_TOOL_ENVIRONMENT
EfmiStatus
ose_efmi_get_process_statistics(HostFmiInstance *i,
				PROCESS pid,
				enum EfmiProcessTag tag,
				struct EfmiProcessStatistics *statistics);
#else
EfmiStatus
ose_efmi_get_process_statistics(PROCESS pid,
                                enum EfmiProcessTag tag,
                                struct EfmiProcessStatistics *statistics);
#endif
/**
 *  @function ose_efmi_get_pool_info
 *
 *  @brief Get information on a pool.
 *  @long This function is only available in the kernel image.
 *     It must be called in supervisor mode.
 *  @long EFMI functions are designed for freeze-mode debug purposes. Any other
 *     use will destroy the interrupt latency of the system.
 *  @input pool_id             Pool identification.
 *  @input pool_info_size      The wanted number of bytes in the output struct.
 *  @output pool_info          The output struct.
 *  @return See enum EfmiStatus
 *  @error See enum EfmiStatus
 *  @long EFMI functions are not part of the real-time services.
 *  @seealso EfmiPoolInfo
 *  @seealso EfmiStatus
**/
#ifdef HOST_TOOL_ENVIRONMENT
EfmiStatus
ose_efmi_get_pool_info(HostFmiInstance *i,
		       PROCESS pool_id,
		       struct EfmiPoolInfo   *pool_info,
		       size_t pool_info_size);
#else
EfmiStatus
ose_efmi_get_pool_info(PROCESS pool_id,
                       struct EfmiPoolInfo   *pool_info,
                       size_t pool_info_size);
#endif
/**
 *  @function ose_efmi_get_signal_info
 *
 *  @brief Get information on a signal.
 *  @long This function is only available in the kernel image.
 *     It must be called in supervisor mode.
 *  @long EFMI functions are designed for freeze-mode debug purposes. Any other
 *     use will destroy the interrupt latency of the system.
 *  @input signal              The signal pointer.
 *  @input sig_info_size       The wanted number of bytes in the output struct.
 *  @output sig_info           The output struct.
 *  @return See enum EfmiStatus
 *  @error See enum EfmiStatus
 *  @long EFMI functions are not part of the real-time services.
 *  @seealso EfmiSignalInfo
 *  @seealso EfmiStatus
**/
#ifdef HOST_TOOL_ENVIRONMENT
EfmiStatus
ose_efmi_get_signal_info(HostFmiInstance *i,
			 union SIGNAL *signal,
			 struct EfmiSignalInfo   *sig_info,
			 size_t sig_info_size);
#else
EfmiStatus
ose_efmi_get_signal_info(union SIGNAL *signal,
                         struct EfmiSignalInfo   *sig_info,
                         size_t sig_info_size);
#endif
/**
 *  @function ose_efmi_get_process_name
 *
 *  @brief Get the name of a process or a block.
 *  @long Retrieve the name associated with the process or block ID.
 *     If max_name_len is less than the length of the process name, 'name'
 *     is filled in and zero terminated, and EFMI_RESULT_TRUNCATED is returned.
 *  @long This function is only available in the kernel image.
 *     It must be called in supervisor mode.
 *  @long EFMI functions are designed for freeze-mode debug purposes. Any other
 *     use will destroy the interrupt latency of the system.
 *  @input pid            The process or block identity.
 *  @input max_name_len   The maximum number of bytes in the output string
 *                        including the 0 terminator.
 *  @output name          The output string.
 *  @return See enum EfmiStatus
 *  @error See enum EfmiStatus
 *  @long EFMI functions are not part of the real-time services.
 *  @seealso ose_efmi_get_first_process_by_name
 *  @seealso EfmiStatus
**/
#ifdef HOST_TOOL_ENVIRONMENT
EfmiStatus
ose_efmi_get_process_name(HostFmiInstance *i,
			  PROCESS pid, char *name, size_t max_name_len);
#else
EfmiStatus
ose_efmi_get_process_name(PROCESS pid, char *name, size_t max_name_len);
#endif

/**
 *  @function ose_efmi_get_register
 *
 *  @brief Get the value of a process specific register.
 *  @long This function reports the value of a process specific register, for
 *  prioritized and background processes. It will not report values for the
 *  currently executing process. It may also require linking of a kernel
 *  library built for debug purposes.
 *  @long Depending on the state of the process, the set of available
 *  registers may vary from time to time. Therefore, the return status
 *  must be checked for each register access. The minimal information
 *  needed to make a stack back trace should always be available in
 *  images containing debug libraries.
 *  @long This function is only available in the kernel image.
 *  It must be called in supervisor mode.
 *  @long EFMI functions are designed for freeze-mode debug purposes. Any other
 *  use will destroy the interrupt latency of the system.
 *  @input pid                  Process identity
 *  @input register_number      Register identity, as specified by
 *                              enum efmi_process_registers in efmi_registers.h
 *  @output reg                 Result struct, containing the register value.
 *  @return See enum EfmiStatus
 *  @error See enum EfmiStatus
 *  @long EFMI functions are not part of the real-time services.
 *  @seealso EfmiRegisterInfo
 *  @seealso EfmiStatus
**/
#ifdef HOST_TOOL_ENVIRONMENT
EfmiStatus
ose_efmi_get_register(HostFmiInstance *i,
		      PROCESS pid,
		      U32 register_number,
		      struct EfmiRegisterInfo *reg);
#else
EfmiStatus
ose_efmi_get_register(PROCESS pid,
                      U32 register_number,
                      struct EfmiRegisterInfo *reg);
#endif

/**
 *  @function ose_efmi_get_config
 *
 *  @brief Get an item from the kernel configuration.
 *  @long This function is only available in the kernel image.
 *     It must be called in supervisor mode.
 *  @long EFMI functions are designed for freeze-mode debug purposes. Any other
 *     use will destroy the interrupt latency of the system.
 *  @input tag          The identity of the wanted item.
 *  @output conf        The result struct, reporting the wanted value.
 *  @return See enum EfmiStatus
 *  @error See enum EfmiStatus
 *  @long EFMI functions are not part of the real-time services.
 *  @seealso EfmiConfigurationTag
 *  @seealso EfmiConfigurationInfo
 *  @seealso EfmiStatus
**/
#ifdef HOST_TOOL_ENVIRONMENT
EfmiStatus
ose_efmi_get_config(HostFmiInstance *i,
		    enum EfmiConfigurationTag tag,
		    struct EfmiConfigurationInfo *conf);
#else
EfmiStatus
ose_efmi_get_config(enum EfmiConfigurationTag tag,
                    struct EfmiConfigurationInfo *conf);
#endif
/**
 *  @function ose_efmi_get_mem
 *
 *  @brief Copy a memory area.
 *  @long This function copies memory from a given address range to the
 *  buffer provided by the caller. Copying is only performed if the source
 *  address range is known and considered safe by EFMI. In that case EFMI_OK is
 *  returned, otherwise EFMI_UNAVAILABLE.
 *  @long The caller is responsible for passing an accessible destination buffer.
 *  @long This function is only available in the kernel image.
 *     It must be called in supervisor mode.
 *  @long EFMI functions are designed for freeze-mode debug purposes. Any other
 *     use will destroy the interrupt latency of the system.
 *  @input id          Process or block identity for the source context.
 *  @input from        Source address.
 *  @input size        Number of bytes to copy.
 *  @output to         Pointer to the target buffer.
 *  @return See enum EfmiStatus
 *  @error See enum EfmiStatus
 *  @long EFMI functions are not part of the real-time services.
 *  @seealso EfmiRegisterInfo
 *  @seealso EfmiStatus
**/
#ifdef HOST_TOOL_ENVIRONMENT
EfmiStatus
ose_efmi_get_mem(HostFmiInstance *i,
		 PROCESS id, OSADDRESS from, void *to, OSADDRESS size);
#else
EfmiStatus
ose_efmi_get_mem(PROCESS id, OSADDRESS from, void *to, OSADDRESS size);
#endif

/**
 *  @function ose_efmi_get_smp_num_cpus
 *
 *  @brief Get number of CPUs in an SMP system.
 *  @long In a single CPU system, this function returns 1.
 *  @long This function is only available in the kernel image.
 *     It must be called in supervisor mode.
 *  @long EFMI functions are designed for freeze-mode debug purposes. Any other
 *     use will destroy the interrupt latency of the system.
 *  @output num_cpus Number of CPUs.
 *  @return See enum EfmiStatus
 *  @error See enum EfmiStatus
 *  @long EFMI functions are not part of the real-time services.
 *  @seealso EfmiStatus
**/
#ifdef HOST_TOOL_ENVIRONMENT
EfmiStatus
ose_efmi_get_smp_num_cpus(HostFmiInstance *i,
			  U32 *num_cpus);
#else
EfmiStatus
ose_efmi_get_smp_num_cpus(U32 *num_cpus);
#endif

/*
 **===========================================================================
 ** Retrieve a list of current context per CPU in an SMP system.
 **===========================================================================
 */
/**
 *  @function ose_efmi_get_first_current
 *
 *  @brief Get the current context for a CPU in a SMP system.
 *  @long In a single CPU system, this function returns the same information
 *     as ose_efmi_get_current().
 *  @long This function is only available in the kernel image.
 *     It must be called in supervisor mode.
 *  @long EFMI functions are designed for freeze-mode debug purposes. Any other
 *     use will destroy the interrupt latency of the system.
 *  @input current_size The number of bytes to write in the current struct.
 *  @output cc        EFMI cursor for internal use.
 *  @output current   Information on the currently running context on a
 *                    particular CPU.
 *  @return This function performs very few tests of the kernel data
 *  structures. It may still report failure if the current process control
 *  block or the system pool descriptor can not be found or is severely damaged.
 *  An obvious example is when the kernel is not yet initiated. See enum
 *  EfmiStatus.
 *  @error See enum EfmiStatus
 *  @long EFMI functions are not part of the real-time services.
 *  @seealso EfmiCurrent
 *  @seealso EfmiStatus
 *  @seealso ose_efmi_get_current
 *  @example
 *  See ose_efmi_get_first_block
**/
#ifdef HOST_TOOL_ENVIRONMENT
EfmiStatus
ose_efmi_get_first_current(HostFmiInstance *i,
			   struct EfmiCurrentCursor *cc,
			   struct EfmiCurrent *current,
			   size_t current_size);
#else
EfmiStatus
ose_efmi_get_first_current(struct EfmiCurrentCursor *cc,
                           struct EfmiCurrent *current,
                           size_t current_size);
#endif

/**
 *  @function ose_efmi_get_next_current
 *
 *  @brief Get the current context for a CPU in a SMP system.
 *  @long In a single CPU system, this function always returns
 *     EFMI_END_OF_LIST.
 *  @long This function is only available in the kernel image.
 *     It must be called in supervisor mode.
 *  @long EFMI functions are designed for freeze-mode debug purposes. Any other
 *     use will destroy the interrupt latency of the system.
 *  @input cc          EFMI cursor for internal use.
 *  @output cc         EFMI cursor for internal use.
 *  @output current Information on the currently running context on a
 *  particular CPU.
 *  @return This function performs very few tests of the kernel data
 *  structures. It may still report failure if the current process control
 *  block or the system pool descriptor can not be found or is severely damaged.
 *  An obvious example is when the kernel is not yet initiated. See enum
 *  EfmiStatus.
 *  @error See enum EfmiStatus
 *  @long EFMI functions are not part of the real-time services.
 *  @seealso EfmiCurrent
 *  @seealso EfmiStatus
 *  @seealso ose_efmi_get_current
 *  @example
 *  See ose_efmi_get_first_block
**/
#ifdef HOST_TOOL_ENVIRONMENT
EfmiStatus
ose_efmi_get_next_current(HostFmiInstance *i,
			  struct EfmiCurrentCursor *cc,
			  struct EfmiCurrent *current);
#else
EfmiStatus
ose_efmi_get_next_current(struct EfmiCurrentCursor *cc,
                          struct EfmiCurrent *current);
#endif

/*
 **===========================================================================
 ** Retrieve the segment list of the system
 **===========================================================================
 */
/**
 *  @function ose_efmi_get_first_segment
 *
 *  @brief Get information on the first segment in the segment list.
 *  @long This function is only available in the kernel image.
 *     It must be called in supervisor mode.
 *  @long EFMI functions are designed for freeze-mode debug purposes. Any other
 *     use will destroy the interrupt latency of the system.
 *  @input pri_size     The wanted number of bytes in the output struct.
 *  @output prc         EFMI cursor for internal use.
 *  @output pri         The output struct.
 *  @return See enum EfmiStatus
 *  @error See enum EfmiStatus
 *  @long EFMI functions are not part of the real-time services.
 *  @seealso EfmiSegmentInfo
 *  @seealso EfmiStatus
 *  @example
 *  See ose_efmi_get_first_block
**/
#ifdef HOST_TOOL_ENVIRONMENT
EfmiStatus
ose_efmi_get_first_segment(HostFmiInstance *i,
			   struct EfmiSegmentCursor *prc,
			   struct EfmiSegmentInfo   *pri,
			   size_t pri_size);
#else
EfmiStatus
ose_efmi_get_first_segment(struct EfmiSegmentCursor *prc,
                           struct EfmiSegmentInfo   *pri,
                           size_t pri_size);
#endif

/**
 *  @function ose_efmi_get_next_segment
 *
 *  @brief Get information on the next segment in the segment list.
 *  @long This function is only available in the kernel image.
 *     It must be called in supervisor mode.
 *  @long EFMI functions are designed for freeze-mode debug purposes. Any other
 *     use will destroy the interrupt latency of the system.
 *  @input  prc         EFMI cursor for internal use.
 *  @output prc         EFMI cursor for internal use.
 *  @output pri         Segment information.
 *  @return See enum EfmiStatus
 *  @error See enum EfmiStatus
 *  @long EFMI functions are not part of the real-time services.
 *  @seealso EfmiSegmentInfo
 *  @seealso EfmiStatus
 *  @example
 *  See ose_efmi_get_first_block
**/
#ifdef HOST_TOOL_ENVIRONMENT
EfmiStatus
ose_efmi_get_next_segment(HostFmiInstance *i,
			  struct EfmiSegmentCursor *prc,
			  struct EfmiSegmentInfo   *pri);
#else
EfmiStatus
ose_efmi_get_next_segment(struct EfmiSegmentCursor *prc,
                          struct EfmiSegmentInfo   *pri);
#endif

/*
 **===========================================================================
 ** Retrieve block information per segment
 **===========================================================================
 */
/**
 *  @function ose_efmi_get_first_block
 *
 *  @brief Get information on the first block of a segment.
 *  @long This function is only available in the kernel image.
 *     It must be called in supervisor mode.
 *  @long EFMI functions are designed for freeze-mode debug purposes. Any other
 *     use will destroy the interrupt latency of the system.
 *  @input segment_id   The segment to investigate.
 *  @input bi_size      The wanted number of bytes in the output struct.
 *  @output bc          EFMI cursor for internal use.
 *  @output bi          The output struct.
 *  @return See enum EfmiStatus
 *  @error See enum EfmiStatus
 *  @long EFMI functions are not part of the real-time services.
 *  @seealso EfmiBlockInfo
 *  @seealso EfmiStatus
 *  @example
 *   struct EfmiBlockCursor bc;
 *   struct EfmiBlockInfo   block_info;
 *   EfmiStatus rv;
 *
 *   rv = ose_efmi_get_first_block(segment_id, &bc, &block_info,
 *                                 sizeof(block_info);
 *   while(rv == EFMI_OK)
 *   {
 *       do_something_with_block_info(&block_info);
 *       rv = ose_efmi_get_next_block(&bc, &block_info);
 *   }
 *
 *   if(rv != EFMI_END_OF_LIST)
 *   {
 *      fmm_printf("Something went wrong when retrieving block information.\n");
 *   }
**/
#ifdef HOST_TOOL_ENVIRONMENT
EfmiStatus
ose_efmi_get_first_block(HostFmiInstance *i,
			 PROCESS segment_id,
			 struct EfmiBlockCursor *bc,
			 struct EfmiBlockInfo   *bi,
			 size_t bi_size);
#else
EfmiStatus
ose_efmi_get_first_block(PROCESS segment_id,
                         struct EfmiBlockCursor *bc,
                         struct EfmiBlockInfo   *bi,
                         size_t bi_size);
#endif

/**
 *  @function ose_efmi_get_next_block
 *
 *  @brief Get information on the next block of a segment.
 *  @long This function is only available in the kernel image.
 *     It must be called in supervisor mode.
 *  @long EFMI functions are designed for freeze-mode debug purposes. Any other
 *     use will destroy the interrupt latency of the system.
 *  @input  bc          EFMI cursor for internal use.
 *  @output bc          EFMI cursor for internal use.
 *  @output bi          The result structure.
 *  @return See enum EfmiStatus
 *  @error See enum EfmiStatus
 *  @long EFMI functions are not part of the real-time services.
 *  @seealso EfmiBlockInfo
 *  @seealso EfmiStatus
 *  @example
 *  See ose_efmi_get_first_block
**/
#ifdef HOST_TOOL_ENVIRONMENT
EfmiStatus
ose_efmi_get_next_block(HostFmiInstance *i,
			struct EfmiBlockCursor *bc,
			struct EfmiBlockInfo   *bi);
#else
EfmiStatus
ose_efmi_get_next_block(struct EfmiBlockCursor *bc,
                        struct EfmiBlockInfo   *bi);
#endif


/*
 **===========================================================================
 ** Retrieve process information per block
 **===========================================================================
 */
/**
 *  @function ose_efmi_get_first_process
 *
 *  @brief Get information on the first process of a block.
 *  @long This function is only available in the kernel image.
 *     It must be called in supervisor mode.
 *  @long EFMI functions are designed for freeze-mode debug purposes. Any other
 *     use will destroy the interrupt latency of the system.
 *  @input block_id        The block to investigate.
 *  @input proc_info_size  The wanted number of bytes in the output struct.
 *  @output pc             EFMI cursor for internal use.
 *  @output proc_info      The output struct.
 *  @return See enum EfmiStatus
 *  @error See enum EfmiStatus
 *  @long EFMI functions are not part of the real-time services.
 *  @seealso EfmiProcessInfo
 *  @seealso EfmiStatus
 *  @example
 *  See ose_efmi_get_first_block
**/
#ifdef HOST_TOOL_ENVIRONMENT
EfmiStatus
ose_efmi_get_first_process(HostFmiInstance *i,
			   PROCESS block_id,
			   struct EfmiProcessCursor *pc,
			   struct EfmiProcessInfo *proc_info,
			   size_t proc_info_size);
#else
EfmiStatus
ose_efmi_get_first_process(PROCESS block_id,
                           struct EfmiProcessCursor *pc,
                           struct EfmiProcessInfo *proc_info,
                           size_t proc_info_size);
#endif

/**
 *  @function ose_efmi_get_next_process
 *
 *  @brief Get information on the next process of a block.
 *  @long This function is only available in the kernel image.
 *     It must be called in supervisor mode.
 *  @long EFMI functions are designed for freeze-mode debug purposes. Any other
 *     use will destroy the interrupt latency of the system.
 *  @input  pc          EFMI cursor for internal use.
 *  @output pc          EFMI cursor for internal use.
 *  @output proc_info   The result struct.
 *  @return See enum EfmiStatus
 *  @error See enum EfmiStatus
 *  @long EFMI functions are not part of the real-time services.
 *  @seealso EfmiProcessInfo
 *  @seealso EfmiStatus
 *  @example
 *  See ose_efmi_get_first_block
**/
#ifdef HOST_TOOL_ENVIRONMENT
EfmiStatus
ose_efmi_get_next_process(HostFmiInstance *i,
			  struct EfmiProcessCursor *pc,
			  struct EfmiProcessInfo   *proc_info);
#else
EfmiStatus
ose_efmi_get_next_process(struct EfmiProcessCursor *pc,
                          struct EfmiProcessInfo   *proc_info);
#endif

/*
 **===========================================================================
 ** Retrieve environment variable information per process or block
 **===========================================================================
 ** If maxLen is less then the length of a string, the string is filled in
 ** and zero terminated, and EFMI_RESULT_TRUNCATED is returned.
 */
/**
 *  @function ose_efmi_get_first_env_var
 *
 *  @brief Get information on the first environment variable of a process or
 *  block.
 *  @long If maxLen is less then the length of a string, the string is filled
 *  in and zero terminated, and EFMI_RESULT_TRUNCATED is returned.
 *  @long This function is only available in the kernel image.
 *     It must be called in supervisor mode.
 *  @long EFMI functions are designed for freeze-mode debug purposes. Any other
 *     use will destroy the interrupt latency of the system.
 *  @input pid            The process to investigate.
 *  @input max_name_len   The maximum number of bytes in the name string.
 *  @input max_value_len  The maximum number of bytes in the value string.
 *  @output evc           EFMI cursor for internal use.
 *  @output name          The name string.
 *  @output value         The value string.
 *  @return See enum EfmiStatus
 *  @error See enum EfmiStatus
 *  @long EFMI functions are not part of the real-time services.
 *  @seealso EfmiStatus
 *  @example
 *  See ose_efmi_get_first_block
**/
#ifdef HOST_TOOL_ENVIRONMENT
EfmiStatus
ose_efmi_get_first_env_var(HostFmiInstance *i,
			   PROCESS pid,
			   struct EfmiEnvVarCursor *evc,
			   char *name,  size_t max_name_len,
			   char *value, size_t max_value_len);
#else
EfmiStatus
ose_efmi_get_first_env_var(PROCESS pid,
                           struct EfmiEnvVarCursor *evc,
                           char *name,  size_t max_name_len,
                           char *value, size_t max_value_len);
#endif

/**
 *  @function ose_efmi_get_next_env_var
 *
 *  @brief Get information on the next environment variable of a process or
 *  block.
 *  @long If maxLen is less then the length of a string, the string is filled
 *  in and zero terminated, and EFMI_RESULT_TRUNCATED is returned.
 *  @long This function is only available in the kernel image.
 *     It must be called in supervisor mode.
 *  @long EFMI functions are designed for freeze-mode debug purposes. Any other
 *     use will destroy the interrupt latency of the system.
 *  @input evc          EFMI cursor for internal use.
 *  @output evc         EFMI cursor for internal use.
 *  @output name        The name string.
 *  @output value       The value string.
 *  @return See enum EfmiStatus
 *  @error See enum EfmiStatus
 *  @long EFMI functions are not part of the real-time services.
 *  @seealso EfmiStatus
 *  @example
 *  See ose_efmi_get_first_block
**/
#ifdef HOST_TOOL_ENVIRONMENT
EfmiStatus
ose_efmi_get_next_env_var(HostFmiInstance *i,
			  struct EfmiEnvVarCursor *evc,
			  char *name,
			  char *value);
#else
EfmiStatus
ose_efmi_get_next_env_var(struct EfmiEnvVarCursor *evc,
                          char *name,
                          char *value);
#endif

/*
 **===========================================================================
 ** Retrieve pool information
 **===========================================================================
 */
/**
 *  @function ose_efmi_get_first_pool
 *
 *  @brief Get information on the first pool in the pool list of a system.
 *  @long This function is only available in the kernel image.
 *     It must be called in supervisor mode.
 *  @long EFMI functions are designed for freeze-mode debug purposes. Any other
 *     use will destroy the interrupt latency of the system.
 *  @input pool_info_size The wanted number of bytes in the output struct.
 *  @output pc          EFMI cursor for internal use.
 *  @output pool_info   The output struct.
 *  @return See enum EfmiStatus
 *  @error See enum EfmiStatus
 *  @long EFMI functions are not part of the real-time services.
 *  @seealso EfmiPoolInfo
 *  @seealso EfmiStatus
 *  @example
 *  See ose_efmi_get_first_block
**/
#ifdef HOST_TOOL_ENVIRONMENT
EfmiStatus
ose_efmi_get_first_pool(HostFmiInstance *i,
			struct EfmiPoolCursor *pc,
			struct EfmiPoolInfo   *pool_info,
			size_t pool_info_size);
#else
EfmiStatus
ose_efmi_get_first_pool(struct EfmiPoolCursor *pc,
                        struct EfmiPoolInfo   *pool_info,
                        size_t pool_info_size);
#endif

/**
 *  @function ose_efmi_get_next_pool
 *
 *  @brief Get information on the next pool in the pool list of a system.
 *  @long This function is only available in the kernel image.
 *     It must be called in supervisor mode.
 *  @long EFMI functions are designed for freeze-mode debug purposes. Any other
 *     use will destroy the interrupt latency of the system.
 *  @input  pc          EFMI cursor for internal use.
 *  @output pc          EFMI cursor for internal use.
 *  @output pool_info   The output struct.
 *  @return See enum EfmiStatus
 *  @error See enum EfmiStatus
 *  @long EFMI functions are not part of the real-time services.
 *  @seealso EfmiPoolInfo
 *  @seealso EfmiStatus
 *  @example
 * See ose_efmi_get_first_block
**/
#ifdef HOST_TOOL_ENVIRONMENT
EfmiStatus
ose_efmi_get_next_pool(HostFmiInstance *i,
		       struct EfmiPoolCursor *pc,
		       struct EfmiPoolInfo   *pool_info);
#else
EfmiStatus
ose_efmi_get_next_pool(struct EfmiPoolCursor *pc,
                       struct EfmiPoolInfo   *pool_info);
#endif

/*
 **===========================================================================
 ** Retrieve fragment information per pool
 **===========================================================================
 */
/**
 *  @function ose_efmi_get_first_fragment
 *
 *  @brief Get information on the first extension fragment of a pool.
 *  @long This function is only available in the kernel image.
 *     It must be called in supervisor mode.
 *  @long EFMI functions are designed for freeze-mode debug purposes. Any other
 *     use will destroy the interrupt latency of the system.
 *  @input pool_id              The pool to investigate.
 *  @input fragment_info_size   The wanted number of bytes in the output
 *                              struct.
 *  @output fc                  EFMI cursor for internal use.
 *  @output fragment_info       The output struct.
 *  @return See enum EfmiStatus
 *  @error See enum EfmiStatus
 *  @long EFMI functions are not part of the real-time services.
 *  @seealso EfmiFragmentInfo
 *  @seealso EfmiStatus
 *  @example
 *  See ose_efmi_get_first_block
**/
#ifdef HOST_TOOL_ENVIRONMENT
EfmiStatus
ose_efmi_get_first_fragment(HostFmiInstance *i,
			    PROCESS pool_id,
			    struct EfmiFragmentCursor *fc,
			    struct EfmiFragmentInfo   *fragment_info,
			    size_t fragment_info_size);
#else
EfmiStatus
ose_efmi_get_first_fragment(PROCESS pool_id,
                            struct EfmiFragmentCursor *fc,
                            struct EfmiFragmentInfo   *fragment_info,
                            size_t fragment_info_size);
#endif

/**
 *  @function ose_efmi_get_next_fragment
 *
 *  @brief Get information on the next extension fragment of a pool.
 *  @long This function is only available in the kernel image.
 *     It must be called in supervisor mode.
 *  @long EFMI functions are designed for freeze-mode debug purposes. Any other
 *     use will destroy the interrupt latency of the system.
 *  @input  fc                  EFMI cursor for internal use.
 *  @output fc                  EFMI cursor for internal use.
 *  @output fragment_info       The output struct.
 *  @return See enum EfmiStatus
 *  @error See enum EfmiStatus
 *  @long EFMI functions are not part of the real-time services.
 *  @seealso EfmiFragmentInfo
 *  @seealso EfmiStatus
 *  @example
 *  See ose_efmi_get_first_block
**/
#ifdef HOST_TOOL_ENVIRONMENT
EfmiStatus
ose_efmi_get_next_fragment(HostFmiInstance *i,
			   struct EfmiFragmentCursor *fc,
			   struct EfmiFragmentInfo   *fragment_info);
#else
EfmiStatus
ose_efmi_get_next_fragment(struct EfmiFragmentCursor *fc,
                           struct EfmiFragmentInfo   *fragment_info);
#endif

/*
 **===========================================================================
 ** Retrieve information about signals in a pool
 **===========================================================================
 */
/**
 *   @function ose_efmi_get_first_pool_signal
 *
 *  @brief Get information on the first signal in the signal list of a pool.
 *  @long This function is only available in the kernel image.
 *     It must be called in supervisor mode.
 *  @long EFMI functions are designed for freeze-mode debug purposes. Any other
 *     use will destroy the interrupt latency of the system.
 *  @input pool_id       The pool to investigate.
 *  @input ps_info_size  The wanted number of bytes in the output struct.
 *  @output psc          EFMI cursor for internal use.
 *  @output ps_info      The output struct.
 *  @return See enum EfmiStatus
 *  @error See enum EfmiStatus
 *  @long EFMI functions are not part of the real-time services.
 *  @seealso EfmiSignalInfo
 *  @seealso EfmiStatus
 *  @example
 *  See ose_efmi_get_first_block
**/
#ifdef HOST_TOOL_ENVIRONMENT
EfmiStatus
ose_efmi_get_first_pool_signal(HostFmiInstance *i,
			       PROCESS pool_id,
			       struct EfmiSignalCursor *psc,
			       struct EfmiSignalInfo   *ps_info,
			       size_t ps_info_size);
#else
EfmiStatus
ose_efmi_get_first_pool_signal(PROCESS pool_id,
                               struct EfmiSignalCursor *psc,
                               struct EfmiSignalInfo   *ps_info,
                               size_t ps_info_size);
#endif

/**
 *  @function ose_efmi_get_next_pool_signal
 *
 *  @brief Get information on the next signal in the signal list of a pool.
 *  @long This function is only available in the kernel image.
 *     It must be called in supervisor mode.
 *  @long EFMI functions are designed for freeze-mode debug purposes. Any other
 *     use will destroy the interrupt latency of the system.
 *  @input  psc         EFMI cursor for internal use.
 *  @output psc         EFMI cursor for internal use.
 *  @output ps_info     The output struct.
 *  @return See enum EfmiStatus
 *  @error See enum EfmiStatus
 *  @long EFMI functions are not part of the real-time services.
 *  @seealso EfmiSignalInfo
 *  @seealso EfmiStatus
 *  @example
 *  See ose_efmi_get_first_block
**/
#ifdef HOST_TOOL_ENVIRONMENT
EfmiStatus
ose_efmi_get_next_pool_signal(HostFmiInstance *i,
			      struct EfmiSignalCursor *psc,
			      struct EfmiSignalInfo   *ps_info);
#else
EfmiStatus
ose_efmi_get_next_pool_signal(struct EfmiSignalCursor *psc,
                              struct EfmiSignalInfo   *ps_info);
#endif

/*
 **===========================================================================
 ** Retrieve information about signals in a process signal queue
 **===========================================================================
 */
/**
 *  @function ose_efmi_get_first_q_signal
 *
 *  @brief Get information on the first signal in the receive queue of a
 *     process.
 *  @long This function is only available in the kernel image.
 *     It must be called in supervisor mode.
 *  @long EFMI functions are designed for freeze-mode debug purposes. Any other
 *	use will destroy the interrupt latency of the system.
 *  @input pid           The process to investigate.
 *  @input sig_info_size The wanted number of bytes in the output struct.
 *  @output sc           EFMI cursor for internal use.
 *  @output sig_info     The output struct.
 *  @return See enum EfmiStatus
 *  @error See enum EfmiStatus
 *  @long EFMI functions are not part of the real-time services.
 *  @seealso EfmiSignalInfo
 *  @seealso EfmiStatus
 *  @example
 *  See ose_efmi_get_first_block
**/
#ifdef HOST_TOOL_ENVIRONMENT
EfmiStatus
ose_efmi_get_first_q_signal(HostFmiInstance *i,
			    PROCESS pid,
			    struct EfmiSignalCursor *sc,
			    struct EfmiSignalInfo   *sig_info,
			    size_t sig_info_size);
#else
EfmiStatus
ose_efmi_get_first_q_signal(PROCESS pid,
                            struct EfmiSignalCursor *sc,
                            struct EfmiSignalInfo   *sig_info,
                            size_t sig_info_size);
#endif

/**
 *  @function ose_efmi_get_next_q_signal
 *
 *  @brief Get information on the next signal in the receive queue of a
 *     process.
 *  @long This function is only available in the kernel image.
 *     It must be called in supervisor mode.
 *  @long EFMI functions are designed for freeze-mode debug purposes. Any other
 *     use will destroy the interrupt latency of the system.
 *  @input  sc          EFMI cursor for internal use.
 *  @output sc          EFMI cursor for internal use.
 *  @output sig_info    The output struct.
 *  @return See enum EfmiStatus
 *  @error See enum EfmiStatus
 *  @long EFMI functions are not part of the real-time services.
 *  @seealso EfmiSignalInfo
 *  @seealso EfmiStatus
 *  @example
 *  See ose_efmi_get_first_block
**/
#ifdef HOST_TOOL_ENVIRONMENT
EfmiStatus
ose_efmi_get_next_q_signal(HostFmiInstance *i,
			   struct EfmiSignalCursor *sc,
			   struct EfmiSignalInfo   *sig_info);
#else
EfmiStatus
ose_efmi_get_next_q_signal(struct EfmiSignalCursor *sc,
                           struct EfmiSignalInfo   *sig_info);
#endif

/*
 **===========================================================================
 ** Retrieve information about attach() signal usage.
 **===========================================================================
 */
/**
 *  @function ose_efmi_get_first_attach_signal
 *
 *  @brief Get information on the first attach signal per process.
 *  @long This function is only available in the kernel image.
 *     It must be called in supervisor mode.
 *  @long EFMI functions are designed for freeze-mode debug purposes. Any other
 *	use will destroy the interrupt latency of the system.
 *  @input pid           The process to investigate.
 *  @input sig_info_size The wanted number of bytes in the output struct.
 *  @input search_method 1 = Information about attach signals for process 
 *                           'pid' requested by other processes.
 *  @break               2 = Information about attach signals requested by
 *                           process 'pid'.
 *  @input options       0 = Only report number of active attach operations
 *                       watching resp. requested by the process in the
 *                       sig_info member 'sig_no'. No other information
 *                       is returned, and follow-up calls of
 *                       ose_efmi_get_next_attach_signal() are not meaningful.
 *                       @break
 *                       1 = PROCESS information returned in the 
 *                       'sig_info' result struct. Repeat calling 
 *                       ose_efmi_get_next_attach_signal() to traverse
 *                       remaining attach signals.
 *  @output sc           EFMI cursor for internal use.
 *  @output sig_info     The output struct.
 *                       Member 'owner' is the process that performed the 
 *                       attach call. 
 *                       @break
 *                       Member 'sender_pid' is the watched object.
 *                       Note that the 'sender_pid' may point out a 
 *                       non-process object with a PROCESS identifier, 
 *                       like a block or a segment. 
 *                       @break
 *                       Currently, the remaining sig_info fields do not 
 *                       provide any useful information. 
 *  @return See enum EfmiStatus
 *  @error See enum EfmiStatus
 *  @long EFMI functions are not part of the real-time services.
 *  @seealso EfmiSignalInfo
 *  @seealso EfmiStatus
 *  @seealso ose_efmi_get_next_attach_signal
 *  @example
 *  See ose_efmi_get_first_block
**/
#ifdef HOST_TOOL_ENVIRONMENT
EfmiStatus
ose_efmi_get_first_attach_signal(HostFmiInstance *i,
				 PROCESS pid,
				 struct EfmiSignalCursor *sc,
				 struct EfmiSignalInfo   *sig_info,
				 size_t sig_info_size,
				 U32 search_method,
				 U32 options);
#else
EfmiStatus
ose_efmi_get_first_attach_signal(PROCESS pid,
				 struct EfmiSignalCursor *sc,
				 struct EfmiSignalInfo   *sig_info,
				 size_t sig_info_size,
				 U32 search_method,
				 U32 options);
#endif

/**
 *  @function ose_efmi_get_next_attach_signal
 *
 *  @brief Get information on the next attach signal per process.
 *  @long This function is only available in the kernel image.
 *     It must be called in supervisor mode.
 *  @long EFMI functions are designed for freeze-mode debug purposes. Any other
 *     use will destroy the interrupt latency of the system.
 *  @input  sc          EFMI cursor for internal use.
 *  @output sc          EFMI cursor for internal use.
 *  @output sig_info    The output struct.
 *  @return See enum EfmiStatus
 *  @error See enum EfmiStatus
 *  @long EFMI functions are not part of the real-time services.
 *  @seealso EfmiSignalInfo
 *  @seealso EfmiStatus
 *  @seealso ose_efmi_get_first_attach_signal
 *  @example
 *  See ose_efmi_get_first_block
**/
#ifdef HOST_TOOL_ENVIRONMENT
EfmiStatus
ose_efmi_get_next_attach_signal(HostFmiInstance *i,
				struct EfmiSignalCursor *sc,
				struct EfmiSignalInfo   *sig_info);
#else
EfmiStatus
ose_efmi_get_next_attach_signal(struct EfmiSignalCursor *sc,
				struct EfmiSignalInfo   *sig_info);
#endif

/*
 **===========================================================================
 ** Retrieve all the PIDs of the processes named 'name'.
 **===========================================================================
 */
/**
 *   @function ose_efmi_get_first_process_by_name
 *
 *  @brief Get the identity of the first process with a given name in the
 *     system.
 *  @long This function is only available in the kernel image.
 *     It must be called in supervisor mode.
 *  @long EFMI functions are designed for freeze-mode debug purposes. Any other
 *     use will destroy the interrupt latency of the system.
 *  @input  name        The name to search in the system.
 *  @output nc          EFMI cursor for internal use.
 *  @output pid         A process identity with the wanted name.
 *  @return See enum EfmiStatus
 *  @error See enum EfmiStatus
 *  @long EFMI functions are not part of the real-time services.
 *  @seealso ose_efmi_get_process_name
 *  @seealso EfmiStatus
 *  @example
 *  See ose_efmi_get_first_block
**/
#ifdef HOST_TOOL_ENVIRONMENT
EfmiStatus
ose_efmi_get_first_process_by_name(HostFmiInstance *i,
				   const char *name,
				   struct EfmiProcNameCursor *nc,
				   PROCESS *pid);
#else
EfmiStatus
ose_efmi_get_first_process_by_name(const char *name,
                                   struct EfmiProcNameCursor *nc,
                                   PROCESS *pid);
#endif

/**
 *  @function ose_efmi_get_next_process_by_name
 *
 *  @brief Get the identity of the next process with a given name in the
 *     system.
 *  @long This function is only available in the kernel image.
 *     It must be called in supervisor mode.
 *  @long EFMI functions are designed for freeze-mode debug purposes. Any other
 *     use will destroy the interrupt latency of the system.
 *  @input  name        The name to search in the system.
 *  @input  nc          EFMI cursor for internal use.
 *  @output nc          EFMI cursor for internal use.
 *  @output pid         A process identity with the wanted name.
 *  @return See enum EfmiStatus
 *  @error See enum EfmiStatus
 *  @long EFMI functions are not part of the real-time services.
 *  @seealso EfmiStatus
 *  @example
 *  See ose_efmi_get_first_block
**/
#ifdef HOST_TOOL_ENVIRONMENT
EfmiStatus
ose_efmi_get_next_process_by_name(HostFmiInstance *i,
				  const char *name,
				  struct EfmiProcNameCursor *nc,
				  PROCESS *pid);
#else
EfmiStatus
ose_efmi_get_next_process_by_name(const char *name,
                                  struct EfmiProcNameCursor *nc,
                                  PROCESS *pid);
#endif


#ifdef __cplusplus
}
#endif
#endif /* _EFMI_H */
