/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2006 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */
/**
 * @toc File_System:
 *
 * @file ose_spi/efs.sig
 *
 * @brief Internal signals common to all Embedded File System (EFS) modules.
 *
 * @long These signals are not public. The definitions are only provided for
 *       convenience, for instance when debugging.
 *
 *
 */

#ifndef _EFS_SIG
#define _EFS_SIG

#include "ose.h"
#include "osetypes.h"
#include "fm.sig"

#ifdef __cplusplus
extern "C" {
#endif

/** 
 * @macro EFS_SIGBASE
 *
 * @brief Start of internal EFS signal numbers.
 */
#define EFS_SIGBASE (30300)

/**
 * @signal ChildDone
 *
 * @brief Completion signal with status and error codes, sent from child to
 *        parent when child calls @link "stdlib.h" "exit".
 *
 * @member ChildDone
 */

/**
 * @macro CHILD_DONE
 * @brief Signal number used for completion, contains status and error codes.
 * @seealso ChildDone
 */
#define CHILD_DONE	(30301)    /* !-SIGNO(struct ChildDone)-! */

/**
 * @type ChildDone
 *
 * @field sigNo  CHILD_DONE
 * @field status The status of the executed command.
 * @field error  Errno value when the child terminated.
 */
struct ChildDone
{
   SIGSELECT sigNo;
   S32       status;
   S32       error;
};

/**
 * @macro sendChildDone
 *
 * @brief Send signal @link "" "ChildDone".
 *
 * @input statusMA    The status of the executed command.
 * @input errorMA     Errno value when the child terminated.
 * @input parentPidMA PID of the parent to send the signal to.
 *  
 * @seealso ChildDone
 */
#define sendChildDone(statusMA, errorMA, parentPidMA) \
do \
{ \
   /* The union is a trick to avoid breaking the strict-aliasing rules*/\
   union U \
   { \
      union SIGNAL *s;\
      struct ChildDone *x;\
   } u;\
   struct ChildDone *cd;\
   u.s = alloc(sizeof(struct ChildDone), CHILD_DONE); \
   cd = u.x;\
   cd->status = (statusMA); \
   cd->error = (errorMA); \
   send(&u.s, (parentPidMA)); \
} \
while (0)

/**
 * @macro HUNT_FSS
 *
 * @brief Signal number used when hunting for the FSS.
 */
#define HUNT_FSS          (30399)     /* !-SIGNO(SIGSELECT)-! */

/**
 * @macro ATTACH_FSS
 *
 * @brief Signal number used when attaching to the FSS.
 */
#define ATTACH_FSS        (30398)     /* !-SIGNO(SIGSELECT)-! */

/**
 * @macro HUNT_FM
 *
 * @brief Signal number used when hunting for format managers.
 */
#define HUNT_FM           (30397)     /* !-SIGNO(SIGSELECT)-! */

/**
 * @macro ATTACH_FM
 *
 * @brief Signal number used when attaching to format managers.
 */
#define ATTACH_FM         (30396)     /* !-SIGNO(SIGSELECT)-! */

/**
 * @macro HUNT_DD
 *
 * @brief Signal number used when hunting for device drivers.
 */
#define HUNT_DD           (30395)     /* !-SIGNO(SIGSELECT)-! */

/**
 * @macro ATTACH_DD
 *
 * @brief Signal number used when attaching to device drivers.
 */
#define ATTACH_DD         (30394)     /* !-SIGNO(SIGSELECT)-! */

/**
 * @macro HUNT_SHELLD
 *
 * @brief Signal number used when hunting for the shell daemon.
 */
#define HUNT_SHELLD       (30393)     /* !-SIGNO(SIGSELECT)-! */

/**
 * @signal HuntPath
 *
 * @brief Signal used by the FSS when hunting for and attaching to other FSSs.
 *
 * @member HuntPath
 */

/**
 * @macro FHUNT_FSS
 * @brief Signal number used by the FSS when hunting for other FSSs.
 * @seealso HuntPath */
#define FHUNT_FSS         (30392)     /* !-SIGNO(struct HuntPath)-! */

/**
 * @macro FATTACH_FSS
 * @brief Signal number used by the FSS when attaching to other FSSs.
 * @seealso HuntPath */
#define FATTACH_FSS       (30391)     /* !-SIGNO(struct HuntPath)-! */

/**
 * @type HuntPath
 *
 * @field signo     FHUNT_FSS, FATTACH_FSS
 * @field hunt_path A variable sized string array including the hunt path
 *                  string for the process to hunt for. Space for the
 *                  termination byte is already included in the signal size.
 */
struct HuntPath
{
   SIGSELECT signo;
   char      hunt_path[1];
};

/**
 * @signal FTimeout
 *
 * @brief Signal used by the FSS to resolve timeouts.
 *
 * @member FTimeout
 */

/**
 * @type PendReq
 *
 * @brief Definition of structure pointed to by pr in @link "" "FTimeout".
 */
struct PendReq;

/**
 * @macro FTIMEOUT
 * @brief Signal number used by the FSS to resolve timeouts.
 * @seealso FTimeout */
#define FTIMEOUT          (30390)     /* !-SIGNO(struct FTimeout)-! */

/**
 * @type FTimeout
 *
 * @field sigNo  FTIMEOUT
 * @field sigNo2 Internal use.
 * @field pr     Pointer to a @link "" "PendReq" structure.
 */
struct FTimeout
{
   SIGSELECT      sigNo;
   SIGSELECT      sigNo2;
   struct PendReq *pr;
};

/**
 * @macro ATTACH_VM
 *
 * @brief Signal number used by the FLIB when attaching to volume managers.
 */
#define ATTACH_VM        (30389)      /* !-SIGNO(SIGSELECT)-! */

/**
 * @macro ATTACH_CHILD
 *
 * @brief Signal number used by the FLIB when attaching to child process.
 */
#define ATTACH_CHILD     (30388)      /* !-SIGNO(SIGSELECT)-! */

/**
 * @signal DonateFd
 *
 * @brief Signal used when donating file descriptors.
 *
 * @member DonateFd
 */

/**
 * @macro DONATE_FD
 * @brief Signal number used when donating file descriptors.
 * @seealso DonateFd */
#define DONATE_FD        (30387)      /* !-SIGNO(struct DonateFd)-! */

/**
 * @type DonateFd
 *
 * @field sigNo       DONATE_FD
 * @field vmPid       PID of the volume manager.
 * @field handle      @link "fm.sig" "FmHandle" handle to an open object.
 * @field buffer_size Internal use.
 * @field buffer_mode Internal use.
 * @field omode       Internal use.
 * @field extra       A variable sized string array. Internal use.
 *                    Space for the termination byte is already included in the
 *                    signal size.
 */
struct DonateFd
{
   SIGSELECT sigNo;
   PROCESS   vmPid;
   FmHandle  handle;
   U32       buffer_size;
   U32       buffer_mode;
   U32       omode;
   char      extra[1];
};

/**
 * @macro sendDonateFd
 *
 * @brief Send signal @link "" "DonateFd".
 *
 * @input vmPidMA   PID of the volume manager.
 * @input handleMA  @link "fm.sig" "FmHandle" handle to an open object.
 * @input bufModeMA Internal use.
 * @input bufSizeMA Internal use.
 * @input omodeMA   Internal use.
 * @input extraMA   Pointer to a variable sized string array. Internal use.
 * @input toPidMA   PID of the process to send the signal to.
 *  
 * @seealso DonateFd
 */
#define sendDonateFd(vmPidMA, handleMA, bufModeMA, bufSizeMA, omodeMA, extraMA, toPidMA) \
do \
{ \
   /* The union is a trick to avoid breaking the strict-aliasing rules*/\
    union U \
    { \
       union SIGNAL *s;\
       struct DonateFd *x;\
    } u;\
    struct DonateFd *sigMA;\
    const char *_extraMA = (extraMA); \
    u.s = alloc(sizeof(struct DonateFd) + strlen(_extraMA), DONATE_FD); \
    sigMA = u.x;\
    sigMA->vmPid = (vmPidMA); \
    sigMA->handle = (handleMA); \
    sigMA->buffer_size = (bufSizeMA); \
    sigMA->buffer_mode = (bufModeMA); \
    sigMA->omode = (omodeMA); \
    strcpy(sigMA->extra, _extraMA); \
    send(&u.s, (toPidMA)); \
} \
while (0)

/**
 * @macro SUSPEND_TMO_SIG
 *
 * @brief Signal number used by aio_suspend().
 */
#define SUSPEND_TMO_SIG      (30386)     /* !-SIGNO(SIGSELECT)-! */

/**
 * @macro VM_ATTACH_SIG
 *
 * @brief Signal number used when AIO calls attach to a VM.
 */
#define VM_ATTACH_SIG        (30385)     /* !-SIGNO(SIGSELECT)-! */

/**
 * @macro HUNT_SHELLD_CK
 *
 * @brief Signal number used when hunting for the shell daemon on OSEck.
 */
#define HUNT_SHELLD_CK       (30384) /* !-SIGNO(SIGSELECT)-! */

#ifdef __cplusplus
} /* end of extern "C" */
#endif

#endif
