/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2011 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */
/**
 * @toc OSE:
 *
 * @file ose_spi/epm.h
 *
 * @brief Execution Point Module Interface
 *
 * @long This file declares the Execution Point Module Interface.
 */

#ifndef _EPM_H
#define _EPM_H

#include "stddef.h"
#include "stdint.h"
#include "ose.h"
#include "osetypes.h"
#include "cpu_regdump.h"
#include "efmi.h"
#include "exception.h"

#ifdef __cplusplus
extern "C" {
#endif

#define BP_CONF_PARAM_NAME "krn/bp_type"
#define BP_CONF_SW         "sw"
#define BP_CONF_HW         "hw"

/**
 *  @macro OSE_EPM_INTERFACE_VERSION
 *  @brief Manifest constant to identify the version of this interface.
 **/
#define OSE_EPM_INTERFACE_VERSION 1

/**
 *  @macro EPM_INTERNAL_SIZE
 *  @brief Size of the internal_data field of struct EpmExecutionPoint.
 *  @seealso EpmExecutionPoint
 **/
#define EPM_INTERNAL_SIZE (48 / sizeof(uint64_t))

/**
 *  @macro EPM_EP_ANY_PROCESS
 *  @brief The execution point will trigger for all processes in a system.
 *  @long Flag used in flags member of EpmExecutionPoint.
 *  When this flag is set, the PID parameter will be ignored. The execution
 *  point will be valid for all processes in a system.
 **/
#define EPM_EP_ANY_PROCESS (1UL << 0)

/**
 *  @macro EPM_SLIDE_OVER_FUNCTION_CALLS
 *  @brief ose_epm_next_instruction option.
 *  @seealso ose_epm_next_instruction
 **/
#define EPM_SLIDE_OVER_FUNCTION_CALLS (1UL)

/**
 *  @type EpmExecutionPointHandle
 *  @brief EpmExecutionPointHandle contains a handle representing an
 *  execution point.
 **/
typedef OSADDRESS EpmExecutionPointHandle;

/**
 *  @type EpmStatus
 *  @brief enum EpmStatus defines the various return values from EPM functions.
 *
 *  @field EPM_OK                  Operation was successful.
 *  @field EPM_UNKNOWN_FLAGS       Use of unknown flag in the flag field.
 *  @field EPM_INVALID_ARGUMENTS   Bad argument passed to an EPM function.
 *  @field EPM_INVALID_INSTRUCTION It is not possible to set an execution
 *                                 point at specified address.
 *  @field EPM_FUNCTION_NOT_IMPLEMENTED The requested function is not
 *                                 implemented.
 *  @field EPM_INTERNAL_ERROR      Internal error. See RAMLOG for further
 *                                 information.
 *  @field EPM_NO_UNUSED_HW_BREAKPOINTS Failure to set a breakpoint because no
 *                                 unused breakpoint registers was found.
 *  @field EPM_REQUEST_TO_CLEAR_NONEXISTING_BP  Attempt to remove a breakpoint
 *                                 that has not been set.
 *  @field EPM_INIT_FAILED         Initiation of breakpoint failed. Proper
 *                                 breakpoint functionality is not guaranteed.
 *  @field EPM_TOO_MANY_BP         Too many breakpoints have set.
 **/
typedef enum EpmStatus
{
   EPM_OK                  = 0,
   EPM_UNKNOWN_FLAGS       = 1,
   EPM_INVALID_ARGUMENTS   = 2,
   EPM_INVALID_INSTRUCTION = 3,
   EPM_FUNCTION_NOT_IMPLEMENTED = 4,
   EPM_INTERNAL_ERROR                  = 5,
   EPM_NO_UNUSED_HW_BREAKPOINTS        = 6,
   EPM_REQUEST_TO_CLEAR_NONEXISTING_BP = 7,
   EPM_INIT_FAILED                     = 8,
   EPM_TOO_MANY_BP                     = 9
} EpmStatus;

/**
 *  @type EpmExecutionPoint
 *
 *  @brief struct EpmExecutionPoint holds all information for an execution
 *  point.
 *
 *  @long EPM uses the internalData for own bookkeeping, as EPM can not allocate
 *  memory on its own. See example for how a client can associate own data to
 *  the execution point.
 *
 *  @field internal_data Memory placeholder for EPM internal data structures.
 *  @field address       Address of the execution point.
 *  @field pid           Specifies the scope of the execution point
 *                       (process/block/segment). Only processes with matching
 *                       PIDs will trig the execution point and invoke the
 *                       callback function.
 *  @field flags         Various flags.
 *  @field cb            Callback function that is called when an execution
 *                       point is hit within its scope.
 *  @field cb.dump       Register dump at the execution point.
 *  @field cb.current    Process context.
 *
 *  @example
 *  struct MyEpmExecutionPoint
 *  {
 *     struct EpmExecutionPoint epmBp;
 *     U32 myStatusFlag;
 *  };
 **/
struct EpmExecutionPoint
{
   uint64_t internal_data[EPM_INTERNAL_SIZE];
   OSADDRESS address;
   PROCESS pid;
   U32 flags;
   int (*cb)(const struct EpmExecutionPoint *,
	     struct cpu_regdump *dump,
	     struct EfmiCurrent *current);
};

/**
 *  @function ose_epm_add_execution_point
 *  @brief Add an execution point.
 *  @long This function is only available in the core image.
 *  It must be called in supervisor mode.
 *  @input ep Pointer to a struct containing information about the
 *  execution point. This struct must be preserved thru the lifetime of the
 *  execution point. It is used internally by the EPM.
 *  @return  See enum EpmStatus
 *  @seealso ose_epm_remove_execution_point
 *  @seealso default_restrictions_calling_functions
 *
 *  @restriction Only available in supervisor mode processes.
**/
EpmStatus
ose_epm_add_execution_point(struct EpmExecutionPoint *ep);

/**
 *  @function ose_epm_remove_execution_point
 *  @brief Remove an execution point.
 *  @long This function is only available in the core image.
 *  It must be called in supervisor mode.
 *  @input ep Pointer to a struct containing information about the
 *  execution point.
 *  ose_epm_add_execution_point.
 *  @return See enum EpmStatus
 *  @seealso ose_epm_add_execution_point
 *  @seealso default_restrictions_calling_functions
 *
 *  @restriction Only available in supervisor mode processes.
**/
EpmStatus
ose_epm_remove_execution_point(struct EpmExecutionPoint *ep);



/**
 *  @function ose_epm_next_instruction
 *  @brief Calculate the address of next instruction.
 *  @long This function is only available in the core image.
 *  It must be called in supervisor mode.
 *  @long This function is used when a client wants to single step a context
 *  to the next instruction.
 *  @input dump Pointer to a register dump.
 *  @input options Zero to follow, or EPM_SLIDE_OVER_FUNCTION_CALLS to slide
 *  over function calls.
 *  @output next_address The next address that is to be executed by the context
 *  represented by the register dump. To avoid adding execution points to some
 *  special instructions, next_address can be adjusted to point out a better
 *  suited opcode.
 *  @return See enum EpmStatus
 *  @seealso ose_epm_add_execution_point
 *  @seealso default_restrictions_calling_functions
 *
 *  @restriction Only available in supervisor mode processes.
**/
EpmStatus
ose_epm_next_instruction(struct cpu_regdump *dump,
			 U32 options,
			 OSADDRESS *next_address);

/**
 *  @function ose_epm_disable_all
 *  @brief Disable all execution points in the system.
 *  @long This function is only available in the core image.
 *        It must be called in supervisor mode.
 *  @long This function disables all execution points. All instructions in
 *        memory that have been patched to generate an exception are
 *        restored so that the memory will be in original state.
 *  @long If this function is called from within a callback function,
 *        ose_epm_enable_all must be called before the callback returns.
 *  @long When the EPM is in the "disable all" state, new execution points will
 *        also be set as disabled. They will be activated when a call to
 *        ose_epm_enable_all is made.
 *
 *  @return See enum EpmStatus
 *  @seealso ose_epm_enable_all
 *  @seealso default_restrictions_calling_functions
 *
 *  @restriction Only available in supervisor mode processes.
**/
EpmStatus
ose_epm_disable_all(void);

/**
 *  @function ose_epm_enable_all
 *  @brief Enable all execution points in the system.
 *  @long This function is only available in the core image.
 *        It must be called in supervisor mode.
 *  @long Reenable all execution points after a previous call to
 *        ose_epm_disable_all.
 *
 *  @return See enum EpmStatus
 *  @seealso ose_epm_disable_all
 *  @seealso default_restrictions_calling_functions
 *
 *  @restriction Only available in supervisor mode processes.
**/
EpmStatus
ose_epm_enable_all(void);

/**
 *  @function ose_epm_get_reg
 *  @brief Get a register value from a CPU register dump.
 *  @long This function is only available in the core image.
 *  It must be called in supervisor mode.
 *  @input dump Pointer to a CPU register dump.
 *  @input id The number of the register to read according to the file
 *  efmi_registers.h.
 *  @output value Register value.
 *  @return See enum EpmStatus
 *  @seealso ose_epm_set_reg
 *  @seealso default_restrictions_calling_functions
 *
 *  @restriction Only available in supervisor mode processes.
**/
EpmStatus
ose_epm_get_reg(struct cpu_regdump *dump, U32 id, U32 *value);

/**
 *  @function ose_epm_set_reg
 *  @brief Set a register value in a CPU register dump.
 *  @long This function is only available in the core image.
 *  It must be called in supervisor mode.
 *  @input dump Pointer to a CPU register dump.
 *  @input id The number of the register to set according to the file
 *  efmi_registers.h.
 *  @input value Register value to set.
 *  @return See enum EpmStatus
 *  @seealso ose_epm_set_reg
 *  @seealso default_restrictions_calling_functions
 *
 *  @restriction Only available in supervisor mode processes.
**/
EpmStatus
ose_epm_set_reg(struct cpu_regdump *dump, U32 id, U32 value);

/**
 *  @function ose_epm_init
 *  @brief Initiate the Execution Point Module.
 *  @long This function is only available in the core image.
 *  It must be called in supervisor mode.
 *  @return See enum EpmStatus
 *
 *  @seealso default_restrictions_calling_functions
 *
 *  @restriction Only available in supervisor mode processes.
**/
EpmStatus
ose_epm_init(void);


#ifdef __cplusplus
}
#endif

#endif /* _EPM_H */
