/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2011 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */
/**
 * @public
 * @toc OSE:
 * @file ose_spi/exception.h
 *
 * @brief Dynamic Exception Handling Services.
 *
 * @long
 * Declares prototypes for exception handlers and functions
 * to install and uninstall exception handlers. These facilities are available
 * for the core module only.
 *
 * @long
 * The vector identifier, used in this interface to identify an exception vector,
 * is different for different architectures:
 *
 * @tab ARM
 * @tab It is the exception vector address offset counted from the Reset vector.
 * Typically 0x4, 0x8, 0xC, 0x10, 0x18, 0x1C.
 * See @link "" "ose_spi/arm/cpu_regdump.h". @break
 *
 * @tab PowerPC BOOKE
 * @tab It is the IVOR register number.
 * Typically 0 - 15 or 32.
 * See @link "" "ose_spi/powerpc/cpu_regdump.h". @break
 *
 * @tab Other PowerPC types
 * @tab It is the vector address offset.
 * Typically offset values 0x100, 0x200, ... but also some additional
 * special values.
 * See @link "" "ose_spi/powerpc/cpu_regdump.h". @break
 *
 * @tab MIPS
 * @tab Vector 0-31 is the ExcCode in the CAUSE register for general
 *      exceptions. Vector >31 correspond to other exceptions with dedicated
 *      address offsets.
 * See @link "" "ose_spi/mips/cpu_regdump.h". @break
 *
 * @tab X86
 * @tab The x86 exception vector number.
 * Typically values up to 31.
 * See @link "" "ose_spi/x86/cpu_regdump.h".@break
 *
 */

#ifndef _EXCEPTION_H
#define _EXCEPTION_H

#include "stddef.h"
#include "osetypes.h"
#include "cpu_regdump.h"

/*
** C++ needs to know that this is C code.
*/

#ifdef __cplusplus
extern "C" {
#endif


/**
 * @type exception_handler_id
 *
 * @brief Identifier used for uninstall of a C handler.
 */
typedef U32 exception_handler_id;

/**
 * @type exception_handler_attribute
 *
 * @brief Defines attributes for a C exception handler.
 *
 * @restriction The EXCEPTION_MONITOR attribute is not implemented
 *              and ignored for PowerPC or MIPS.
 *
 * @field EXCEPTION_MONITOR To be used for debug monitors and other system
 * components that require early handlers. If multiple handlers are installed
 * for a vector, those marked with this attribute will be executed first.
 *
 * @field EXCEPTION_NORMAL To be used by the majority of the handlers.
 */
enum exception_handler_attribute
  {
    EXCEPTION_NORMAL = 0,
    EXCEPTION_MONITOR = 1
  };

/**
 * @type exception_handler_asm
 *
 * @brief Prototype for an assembler exception handler.
 *
 * @long
 * The assembler exception handler is responsible for the handling of all
 * exceptions related to this vector.
 * Assembler handlers are entered directly from the  exception vector, and
 * complete by making a normal return from exception.
 * For PowerPC, the handler code is copied into the exception
 * vector, leading to a limited length of the code. For MIPS, the code
 * is only copied for the two dedicated vectors "TLB Refill" and "XTLB Refill".
 *
 * @long
 * No stack is available. Handlers requiring a stack must explicitly set
 * it up.
 */
typedef void (exception_handler_asm)(void);

/**
 * @type exception_handler_c
 *
 * @brief Prototype for an exception handler with C language interface.
 *
 * @long
 * The C handlers are called in supervisor mode with interrupts disabled.
 * The handler can read the provided register dump variable to decide if
 * it wants to handle the current exception.
 * In that case, it may enable interrupts and modify the register dump. It
 * shall then return non-zero. It is an error to enable interrupts if the
 * register dump shows that interrupts were disabled before the exception
 * occurred.
 *
 * @long
 * If it does not handle the exception, it must return zero with
 * interrupts still disabled.
 *
 * @long
 * C handlers are chained. As long as all handlers return status zero,
 * the next handler is called. OSE always provides a default handler,
 * to terminate the chain.
 *
 * @long
 * Note that use of many or complex handlers may affect the interrupt
 * latency, as they are called with interrupts disabled.
 *
 * @long
 * The C handlers execute on the supervisor stack of the current process,
 * or, in the early system start, on the system start stack. The register
 * dump structure is already stored on the same stack. Thus, the stack space
 * available to the handler is limited. Supervisor Stack Overflow Errors
 * can be resolved by configuration of larger supervisor stacks. This is
 * however an expensive method, because all processes then require more
 * memory. Another method is to switch to a private handler stack inside
 * the handler. If that stack is statically allocated, interrupts must
 * be disabled while using it.
 *
 * @field exception_id Vector identifier, see leading text in
 *        @link "" "ose_spi/exception.h".
 * @field dump   Register dump. See struct cpu_regdump for the actual
 *        architecture.
 */
typedef U32 (exception_handler_c)(U32 exception_id,
				   struct cpu_regdump *dump);

/**
 * @function ose_install_asm_exception_handler
 *
 * @brief Install an assembler exception handler.
 *
 * @long
 * Install an assembler exception handler on an exception vector.
 * The handler is responsible for the handling of all exceptions related
 * to this vector.
 * @long The handler will overwrite the default exception handler
 * for this vector, which disables any use of a C exception handler.
 *
 * @input exception_id Vector identifier, see leading text in
 *        @link "" "ose_spi/exception.h".
 *
 * @input handler The exception handler to install (a function pointer).
 *
 * @input size The size is used on architectures where the code is copied
 * to the vector area (see @link "" "exception_handler_asm").
 * A longer handler than the hardware defined vector causes an error. The
 * size argument is ignored on architectures using single word vectors,
 * for example ARM.
 *
 * @return @tab ARM: @tab A reference to the previously installed handler.
 *                   It returns 0 if the exception ID is not recognized. @break
 *         @tab PowerPC: @tab 0 at success. It returns -1 at an error, e.g.
 *                   if size is too big or if the exception ID is not
 *                   recognized. @break
 *         @tab MIPS: @tab A reference to the previously installed handler.
 *                   However, it returns NULL for the two dedicated vectors
 *                   "TLB Refill" and "XTLB Refill". @break
 *
 * @restriction Only available in supervisor mode processes.
 *    @break    Not implemented for X86 or SFK.
 *
 * @seealso exception_handler_asm
 * @seealso ose_install_c_exception_handler
 * @seealso default_restrictions_calling_functions
 *
 * @example
 * // To install an ASM machine check handler in PowerPC 750 (not BOOKE)
 * ose_install_asm_exception_handler(0x200,
 *                                   (exception_handler_asm *) myHandler,
 *                                   128);
 *
 * // To install an ASM machine check handler in PowerPC BOOKE, e.g. 440
 * ose_install_asm_exception_handler(1,
 *                                   (exception_handler_asm *) myHandler,
 *                                   128);
 *
 * // To install an ASM data abort handler in ARM
 * ose_install_asm_exception_handler(0x10,
 *                                   (exception_handler_asm *) myHandler,
 *                                   0); // size ignored
 *
 * // To install an ASM machine check handler in MIPS
 * ose_install_asm_exception_handler(24,
 *                                   (exception_handler_asm *) myHandler,
 *                                   0); // size ignored for "24"
 */
exception_handler_asm *
ose_install_asm_exception_handler(U32 exception_id,
				  exception_handler_asm *handler,
				  size_t size);

/**
 * @function ose_install_c_exception_handler
 *
 * @brief Install an exception handler with a C language interface.
 *
 * @long
 * Install an exception handler with a C language interface
 * on a vector. The C handlers are primarily intended for debug,
 * hardware probing, error diagnosis and resolution of other not
 * too frequent exceptions.
 *
 * @long
 * Note that use of many or complex handlers may affect the interrupt
 * latency, as they are called with interrupts disabled.
 *
 * @long
 * The use of C exception handlers requires that the default ASM handlers
 * are used, e.g. C handlers are silently disabled if an ASM handler
 * is installed on the same vector. OSE usually installs own ASM handlers
 * at least on External Interrupt and TLB Miss vectors.
 *
 * @restriction
 * Only available in supervisor mode processes.
 * @break
 * C exception handlers can not be used for system call
 * exceptions, external interrupts including decrementer, and TLB miss
 * vectors.
 *
 * @input exception_id Vector identifier, see leading text in
 *        @link "" "ose_spi/exception.h".
 *
 * @input handler The exception handler to install (function pointer).
 *
 * @input attribute Selected from enum @link "" "exception_handler_attribute".
 *
 * @return Identification of the handler, to be used at uninstall.
 *         It returns 0 if the exception ID is not recognized.
 *
 * @seealso exception_handler_c
 * @seealso ose_uninstall_c_exception_handler
 * @seealso ose_install_asm_exception_handler
 * @seealso default_restrictions_calling_functions
 *
 * @example
 * // To install a data abort C handler in ARM
 * exception_handler_id identity =
 *  ose_install_c_exception_handler(0x10, (exception_handler_c *)myHandler,
 *                                  EXCEPTION_NORMAL);
 *
 * // To install a program exception C handler on PowerPC BOOKE e.g. 440
 * exception_handler_id identity =
 *  ose_install_c_exception_handler(6, (exception_handler_c *)myHandler,
 *                                  EXCEPTION_NORMAL);
 *
 * // To install a program exception C handler on PowerPC 750 (not BOOKE)
 * exception_handler_id identity =
 *  ose_install_c_exception_handler(0x700, (exception_handler_c *)myHandler,
 *                                  EXCEPTION_NORMAL);
 */
exception_handler_id
ose_install_c_exception_handler(U32 exception_id,
				exception_handler_c *handler,
				U32 attribute);

/**
 * @function ose_uninstall_c_exception_handler
 *
 * @brief Uninstall a C language exception handler.
 *
 * @input identity The identifier returned at the installation of the handler.
 *
 * @seealso ose_install_c_exception_handler
 * @seealso default_restrictions_calling_functions
 *
 * @restriction Only available in supervisor mode processes.
 */
void
ose_uninstall_c_exception_handler(exception_handler_id identity);

#ifdef __cplusplus
}
#endif

#endif
/* _EXCEPTION_H */
