/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2011 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */
/**
*  @public
*  @toc File_System:
*  @file ose_spi/flash_api.h
*
*  @brief   Declaration of API for flash access.
*
*  @long
*  The API declared in this file allows application programs to access
*  flash memories.
*
*  @long
*  The user of this API will not need to be concerned with different
*  device drivers. All flash access is channeled through this API.
*  The Flash Access Manager determines the flash instance and
*  delivers the access request to the correct device driver.
*
*  @long
*  This API is thread safe. The Flash Access Manager knows about the
*  hardware limitations of the flash and resolves all potential access
*  conflicts.
*
*  @long
*  An application is not allowed to access the flash without using the
*  API, unless the address range has been configured to allow direct read
*  access. Otherwise this direct access may result in an access conflict
*  in the device with unforeseeable effects.
*
*  @long
*  The allocation of flash regions to different applications is beyond
*  the scope of this API. An application is responsible to only access
*  flash regions that have been assigned to it. This API does not prevent
*  an application from accessing flash regions that are assigned to other
*  applications.
*/


#ifndef _FLASH_API_H
#define _FLASH_API_H

#include <stddef.h>
#include "osetypes.h"

#ifdef __cplusplus
extern "C" {
#endif


/**
 * @macro FAM_INSTANCE_NONE
 * @brief Instance value indicating that instance should be selected by
 *        address.
 * @long  When this value is given in calls where an instance number can be
 *        given, the instance will be selected based on address rather than
 *        by instance number. If no address can be given, this value is
 *        illegal.
 * @seealso flash_read2
 * @seealso flash_write2
 */
#define FAM_INSTANCE_NONE 0xFFFFFFFFul


/**
*  @type    FlashResult
*
*  @declaration
*           typedef enum FlashResult
*           {
*              FLASH_EOK                     = 0,
*              FLASH_ENO_SUCH_INSTANCE       = 0x00CC1000,
*              FLASH_EINVALID_ADDRESS        = 0x00CC1001,
*              FLASH_EUNKNOWN_REQUEST        = 0x00CC1002,
*              FLASH_EUNMATCHED_ERASEBLOCK   = 0x00CC1003,
*              FLASH_EHARDWARE_ERROR         = 0x00CC1004,
*              FLASH_EPROTECTED              = 0x00CC1005,
*              FLASH_EINVALID_DATA           = 0x00CC1006
*           } FlashResult;
*
*  @brief   Result code returned by the flash access functions.
*
*  @field   FLASH_EOK
*           The operation has completed successfully.
*
*  @field   FLASH_ENO_SUCH_INSTANCE
*           There is no installed flash instance that matches the
*           instance number.
*
*  @field   FLASH_EINVALID_ADDRESS
*           A requested flash address does not match any installed flash.
*
*  @field   FLASH_EUNKNOWN_REQUEST
*           The request_type parameter does not match any of the defined
*           request types.
*
*  @field   FLASH_EUNMATCHED_ERASEBLOCK
*           An erase operation has been attempted where eraseblock_addr
*           or length does not match valid erase block boundaries.
*
*  @field   FLASH_EHARDWARE_ERROR
*           The operation could not be completed because of a hardware
*           malfunction. It is possible that part of the operation has been
*           carried out.
*
*  @field   FLASH_EPROTECTED
*           The requested flash address is protected from the requested
*           type of access.
*           Note: The user can not rely on protection to prevent
*           unauthorized access.
*
*  @field   FLASH_EINVALID_DATA
*           Operation could not be performed because the application specified
*           invalid data to write (i.e. tried to change bits the wrong way).
*/
typedef enum FlashResult
{
   FLASH_EOK                     = 0,
   FLASH_ENO_SUCH_INSTANCE       = 0x00CC1000,
   FLASH_EINVALID_ADDRESS        = 0x00CC1001,
   FLASH_EUNKNOWN_REQUEST        = 0x00CC1002,
   FLASH_EUNMATCHED_ERASEBLOCK   = 0x00CC1003,
   FLASH_EHARDWARE_ERROR         = 0x00CC1004,
   FLASH_EPROTECTED              = 0x00CC1005,
   FLASH_EINVALID_DATA           = 0x00CC1006
} FlashResult;


/**
*  @type FlashAddress
*
*  @declaration
*           typedef char *FlashAddress;
*
*  @brief Holds an address within a flash.
*/
typedef char *FlashAddress;


/**
*  @type FlashAccessType
*
*  @brief Indicates access type for read or write.
*
*  @long  Since this enum is not version handled, therefore, members may only
*         be added, not removed. Any code using this enum must be able to
*         gracefully handle unknown values.
*
*  @field FLASH_ACCESS_DATA
*         Read or write chunks of data. In this mode, Sibley expects each write
*         to be done in one block at a time, and no other modifications to the
*         same block is then allowed. For flashes that does not support
*         different types of accesses (e.g. normal NOR flash), this is the
*         only available access type.
*
*  @field FLASH_ACCESS_CONTROL
*         Read or write any amount of data. In this mode, Sibley writes can be
*         of any size, and several writes can be done on the same block as long
*         as ones becomes zeros, i.e. behaves as good old NOR flash does. In
*         mode only half amount of memory is available.
*
*  @seealso flash_read2
*  @seealso flash_write2
*/
typedef enum FlashAccessType
{
   FLASH_ACCESS_DATA = 0,
   FLASH_ACCESS_CONTROL = 1
} FlashAccessType;


/**
*  @function flash_read
*
*  @declaration
*  FlashResult
*  flash_read(void *dest_buf, FlashAddress src_flash_addr, size_t length);
*
*  @brief   Reads data from flash.
*
*  @input   dest_buf
*           Pointer to client supplied buffer that shall
*           receive the read data.
*
*  @input   src_flash_addr
*           Address in flash where data shall be read.
*
*  @input   length
*           Number of bytes that shall be read.
*
*  @return  Result code. See description of enum FlashResult for details.
*
*  @restriction 
*     It is not allowed to kill a calling process while inside a FAM call.
*     This may lead to a deadlock situation, where future FAM calls from
*     other processes would hang forever.
*     The FAM will not detect termination of a FAM client.
*     See also FAM in the OSE Core Extentions User's Guide.
*  @break
*     Only available in supervisor mode processes.
*
*  @seealso default_restrictions_calling_functions
*/
FlashResult
flash_read(void *dest_buf, FlashAddress src_flash_addr, size_t length);


/**
*  @function flash_read2
*
*  @declaration
*  FlashResult
*  flash_read2(void *dest_buf, FlashAddress src_flash_addr, size_t length,
*              U32 instance, FlashAccessType access_type);
*
*  @brief   Reads data from flash.
*
*  @long    When using access type FLASH_ACCESS_CONTROL (used for SIBLEY
*           memories), the parameter src_flash_addr is a logical address.
*           In this mode only half of the page (512 of 1024 bytes) is available.
*           The logical address points to somewhere in the first 512 bytes of
*           the page, i.e. the address consists of the page address + an offset
*           in the range 0-511 (a "logical" address). The logical to physical
*           address translation is done by the FAM driver (see flash_drv.h).
*
*  @input   dest_buf
*           Pointer to client supplied buffer that shall
*           receive the read data.
*
*  @input   src_flash_addr
*           Address in flash where data shall be read.
*
*  @input   length
*           Number of bytes that shall be read.
*
*  @input   instance
*           Which FAM instance to use. If 0xFFFFFFFF the parameter
*           src_flash_addr will be used to determine the correct instance,
*           which is the same behavior as for flash_read().
*
*  @input   access_type
*           Some flash has different types of access depending on what kind of
*           information that is written. E.g. Sibley NOR flash has data access
*           type as well as control. This parameter selects which of these to
*           use. The exact meaning of this parameter is dependent on memory
*           technology.
*           See @link "" "FlashAccessType" for valid values.
*
*  @return  Result code. See description of enum FlashResult for details.
*
*  @restriction 
*     It is not allowed to kill a calling process while inside a FAM call.
*     This may lead to a deadlock situation, where future FAM calls from
*     other processes would hang forever.
*     The FAM will not detect termination of a FAM client.
*     See also FAM in the OSE Core Extentions User's Guide.
*  @break
*     Only available in supervisor mode processes.
*
*  @seealso default_restrictions_calling_functions
*/
FlashResult
flash_read2(void *dest_buf, FlashAddress src_flash_addr, size_t length,
            U32 instance, FlashAccessType access_type);


/**
*  @function flash_write
*
*  @declaration
*  FlashResult
*  flash_write(FlashAddress dest_flash_addr, const void *src_buf, size_t length);
*
*  @brief   Writes data to flash.
*
*  @input   dest_flash_addr
*           Start address in flash where data shall be written.
*
*  @input   src_buf
*           Pointer to buffer containing data that shall be written to flash.
*
*  @input   length
*           Number of bytes that shall be written.
*
*  @return  Result code. See description of enum FlashResult for details.
*
*  @restriction 
*     It is not allowed to kill a calling process while inside a FAM call.
*     This may lead to a deadlock situation, where future FAM calls from
*     other processes would hang forever.
*     The FAM will not detect termination of a FAM client.
*     See also FAM in the OSE Core Extentions User's Guide.
*  @break
*     Only available in supervisor mode processes.
*
*  @seealso default_restrictions_calling_functions
*/
FlashResult
flash_write(FlashAddress dest_flash_addr, const void *src_buf, size_t length);


/**
*  @function flash_write2
*
*  @declaration
*  FlashResult
*  flash_write2(FlashAddress dest_flash_addr, const void *src_buf, size_t length,
*               U32 instance, FlashAccessType access_type);
*
*  @brief   Writes data to flash.
*
*  @long    When using access type FLASH_ACCESS_CONTROL (used for SIBLEY
*           memories), the parameter src_flash_addr is a logical address.
*           In this mode only half of the page (512 of 1024 bytes) is available.
*           The logical address points to somewhere in the first 512 bytes of
*           the page, i.e. The address consists of the page address + an offset
*           in the range 0-511 (a "logical" address). The logical to physical
*           address translation is done by the FAM driver (see flash_drv.h).
*
*  @input   dest_flash_addr
*           Start address in flash where data shall be written.
*
*  @input   src_buf
*           Pointer to buffer containing data that shall be written to flash.
*
*  @input   length
*           Number of bytes that shall be written.
*
*  @input   instance
*           Which FAM instance to use. If 0xFFFFFFFF the parameter
*           src_flash_addr will be used to determine the correct instance,
*           which is the same behavior as for flash_read().
*
*  @input   access_type
*           Some flash has different types of access depending on what kind of
*           information that is written. E.g. Sibley NOR flash has data access
*           type as well as control. This parameter selects which of these to
*           use. The exact meaning of this parameter is dependent on memory
*           technology.
*           See @link "" "FlashAccessType" for valid values.
*
*  @return  Result code. See description of enum FlashResult for details.
*
*  @restriction 
*     It is not allowed to kill a calling process while inside a FAM call.
*     This may lead to a deadlock situation, where future FAM calls from
*     other processes would hang forever.
*     The FAM will not detect termination of a FAM client.
*     See also FAM in the OSE Core Extentions User's Guide.
*  @break
*     Only available in supervisor mode processes.
*
*  @seealso default_restrictions_calling_functions
*/
FlashResult
flash_write2(FlashAddress dest_flash_addr, const void *src_buf, size_t length,
             U32 instance, FlashAccessType access_type);


/**
*  @function flash_erase
*
*  @declaration
*           FlashResult
*           flash_erase(FlashAddress eraseblock_addr, size_t length);
*
*  @brief   Erases a sequence of erase-blocks in flash.
*
*  @input   eraseblock_addr
*           Start address in flash where data shall be erased. This must be
*           the start address of an erase block boundary.
*
*  @input   length
*           Number of bytes that shall be erased. This must be a multiple of
*           the size of one erase block.
*
*  @return  Result code. See description of enum FlashResult for details.
*
*  @restriction 
*     It is not allowed to kill a calling process while inside a FAM call.
*     This may lead to a deadlock situation, where future FAM calls from
*     other processes would hang forever.
*     The FAM will not detect termination of a FAM client.
*     See also FAM in the OSE Core Extentions User's Guide.
*  @break
*     Only available in supervisor mode processes.
*
*  @seealso default_restrictions_calling_functions
*/
FlashResult
flash_erase(FlashAddress eraseblock_addr, size_t length);


/**
*  @type    FlashInstanceInfo
*
*  @declaration
*           typedef struct FlashInstanceInfo
*           {
*              U32            instance_count;
*              FlashAddress   instance_start;
*              size_t         instance_length;
*           } FlashInstanceInfo;
*
*  @brief   Holds the result from the function flash_get_instance.
*
*  @long    See the description of that function for further details.
*
*  @field   instance_count
*           Total number of flash instances that are installed. all instance
*           numbers are in the range zero to (instance_count - 1).
*
*  @field   instance_start
*           The first flash address of the requested flash instance.
*
*  @field   instance_length
*           The number of bytes in the requested flash instance.
*/
typedef struct FlashInstanceInfo
{
   U32            instance_count;
   FlashAddress   instance_start;
   size_t         instance_length;
} FlashInstanceInfo;


/**
*  @function flash_get_instance
*
*  @declaration
*           FlashResult
*           flash_get_instance(FlashInstanceInfo *instance_info,
*                              U32 request_instance_nr);
*
*  @brief   Get the address range of a flash instance and the total number
*           of instances.
*
*  @long
*  A flash instance is continuous range of flash memory. Sometimes a single
*  flash chip, sometimes several flash chips that are accessed at consecutive
*  addresses.
*
*  Use repeated calls to this function to gain information about all instances.
*
*  @input   instance_info
*           Pointer to a struct that will hold the results of the operation.
*           See description of struct FlashInstanceInfo for details.
*
*  @input   request_instance_nr
*           Instance number of the flash instance that shall be examined.
*           The first flash instance is number 0.
*
*  @return  Result code. See description of enum FlashResult for details.
*
*  @restriction 
*     It is not allowed to kill a calling process while inside a FAM call.
*     This may lead to a deadlock situation, where future FAM calls from
*     other processes would hang forever.
*     The FAM will not detect termination of a FAM client.
*     See also FAM in the OSE Core Extentions User's Guide.
*  @break
*     Only available in supervisor mode processes.
*
*  @seealso default_restrictions_calling_functions
*/
FlashResult
flash_get_instance(FlashInstanceInfo *instance_info,
                   U32 request_instance_nr);

/**
 * @type FlashMemoryTypeBits
 *
 * @brief Description of bit field found in @link "" "FlashMemoryType".
 *
 * @long  The bits can be or'ed together to select several features.
 *
 * @field FLASH_MEMTYPE_CAN_REWRITE_ZERO
 *        When a block is rewritten, the flash might expect that if a bit is
 *        already written to zero, the corresponding bit to be written is one.
 *        In this case, this bit is false (0). If this bit is true, the bit
 *        to be written will simply be ignored, since a zero can not become one
 *        when writing.
 *
 * @field FLASH_MEMTYPE_MULTILEVEL_CELL
 *        Indicates whether the flash is multi-level cell or single-level cell.
 *        If flash is multi-level cell, this bit will be true (1).
 *
 * @seealso FlashMemoryType
 */
typedef enum FlashMemoryTypeBits
{
   FLASH_MEMTYPE_CAN_REWRITE_ZERO = 1,
   FLASH_MEMTYPE_MULTILEVEL_CELL = 2
} FlashMemoryTypeBits;

/**
 * @type FlashMemoryTechnologies
 *
 * @brief Valid values indicating memory technology for
 *        @link "" "FlashMemoryType".
 *
 * @long  Since this enum is not version handled, therefore, members may only
 *        be added, not removed. Any code using this enum must be able to
 *        gracefully handle unknown values.
 *
 * @field FLASH_MEMTECH_RAM
 *        Random access memory (RAM).
 *
 * @field FLASH_MEMTECH_NOR
 *        Traditional NOR flash.
 *
 * @field FLASH_MEMTECH_NOR_SIBLEY
 *        Intel Sibley NOR flash.
 *
 * @seealso FlashMemoryType
 */
typedef enum FlashMemoryTechnologies
{
   FLASH_MEMTECH_RAM = 0,
   FLASH_MEMTECH_NOR = 1,
   FLASH_MEMTECH_NOR_SIBLEY = 2
} FlashMemoryTechnologies;

/**
 *  @type FlashMemoryTypeParts
 *
 *  @brief Part of @link "" "FlashMemoryType" union.
 *
 *  @field memory_type
 *         Bitfield containing bits described in
 *         @link "" "FlashMemoryTypeBits".
 *
 *  @field reserved
 *         Reserved for future use. FAM driver must fill this field with the
 *         value zero.
 *
 *  @field memory_technology
 *         Value indicating what type of memory this is (RAM, NOR, etc.).
 *         See @link "" "FlashMemoryTechnologies" for valid values.
 *
 *
 *  @seealso FlashMemoryType
 *  @seealso FlashMemoryTypeBits
 *  @seealso FlashMemoryTechnologies
 */
typedef struct FlashMemoryTypeParts
{
      U16  memory_type;
      U8   reserved;
      U8   memory_technology;
} FlashMemoryTypeParts;

/**
 *  @type FlashMemoryType
 *
 *  @brief Holds the information returned by flash_get_characteristics when
 *         requesting FLASH_MEMORY_TYPE.
 *
 *  @field whole
 *         This field is used when the whole value is referred to.
 *
 *  @field part
 *         Struct containing the actual information. See
 *         @link "" "FlashMemoryTypeParts" for more information.
 *
 *
 *
 *  @seealso FlashMemoryTypeParts
 */
typedef union FlashMemoryType
{
      size_t whole;
      FlashMemoryTypeParts part;
} FlashMemoryType;


/**
*  @type    FlashCharacteristics
*
*  @declaration
*           typedef struct FlashCharacteristics
*           {
*              FlashAddress   range_start;
*              size_t         range_length;
*              size_t         block_length;
*           } FlashCharacteristics;
*
*  @brief   Holds the result from the function flash_get_characteristics.
*
*  @long    See the description of that function for further details and use cases.
*
*  @field   range_start
*           The first flash address of a range, where the entire range has
*           the same specific characteristic.
*
*  @field   range_length
*           The number of bytes in the address range.
*
*  @field   block_length
*           The block size in bytes when the range is a range of equally
*           sized blocks. The number of blocks in the range is
*           (range_length/ block_length). See the description of
*           enum CharacteristicType for more information.
*/
typedef struct FlashCharacteristics
{
   FlashAddress   range_start;
   size_t         range_length;
   size_t         block_length;
} FlashCharacteristics;


/**
*  @type CharacteristicType
*
*  @declaration
*           typedef enum CharacteristicType
*           {
*              FLASH_INSTANCE_SIZE = 0,
*              FLASH_ERASEBLOCK_RANGE = 1,
*              FLASH_MODIFY_PREVENTS_READ = 2,
*              FLASH_MODIFY_PREVENTS_MODIFY = 3,
*              FLASH_PAGE_SIZE = 4,
*              FLASH_MEMORY_TYPE = 5
*           } CharacteristicType;
*
*  @brief   Used with the function get_flash_characteristics to select the
*           type of characteristic desired.
*
*  @long    The interpretation of some fields in struct FlashCharacteristics
*           depend on the requested CharacteristicType.
*
*  @field   FLASH_INSTANCE_SIZE
*           Get the accessible address range of the entire flash instance.
*           block_length is not used by this request. (This can also be
*           retrieved with the function flash_get_instance.)
*
*  @field   FLASH_ERASEBLOCK_RANGE
*           Get an address range of equally sized erase blocks. block_length
*           will hold the size of each erase block in the range.
*
*  @field   FLASH_MODIFY_PREVENTS_READ
*           Get an address range of equally sized blocks, each block of size
*           block_length, where it is not possible to read in a block at the
*           same time as a write or erase is performed in the same block.
*           (It is possible to read in another block.) (These blocks are
*           equivalent of flash partitions in some flash architectures, and
*           entire flash chips in others.)
*
*  @field   FLASH_MODIFY_PREVENTS_MODIFY
*           Get an address range of equally sized blocks, each block of size
*           block_length, where it is not possible to write or erase in a
*           block at the same time as another write or erase is performed in
*           the same block. (It is possible to write or erase in another
*           block. These blocks are equivalent the size of a flash chip in
*           most flash architectures.)
*
*  @field   FLASH_PAGE_SIZE
*           Currently used by Sibley flash only. On Sibley flash this is the
*           size of a program region. The block_length field of the
*           FlashCharacteristics struct is treated as the page size.
*           By taking this value and divide it with 32 on Sibley flash,
*           the result will be the size of a program set regardless of number
*           of Sibley flashes that are in parallel.
*
*  @field   FLASH_MEMORY_TYPE
*           For this characteristic, the block_length field of
*           FlashCharacteristics struct is treated as a struct of type
*           @link "" "FlashMemoryType". The information found in this struct is
*           valid for the whole instance.
*/
typedef enum CharacteristicType
{
   FLASH_INSTANCE_SIZE = 0,
   FLASH_ERASEBLOCK_RANGE = 1,
   FLASH_MODIFY_PREVENTS_READ = 2,
   FLASH_MODIFY_PREVENTS_MODIFY = 3,
   FLASH_PAGE_SIZE = 4,
   FLASH_MEMORY_TYPE = 5
} CharacteristicType;


/**
*  @function flash_get_characteristics
*
*  @declaration
*           FlashResult
*           flash_get_characteristics(FlashCharacteristics *characteristics,
*                                     CharacteristicType request_type,
*                                     FlashAddress request_address);
*
*  @brief Gets the characteristics of an address range in a flash instance.
*
*  @long
*  The behavior is like this:
*  * The requested characteristic is determined at the requested address.
*  * The largest possible range with the same characteristic that surrounds
*    the requested address is determined.
*
*  An example, with arbitrarily selected numbers:
*  With request_type=FLASH_ERASEBLOCK_RANGE and request_address=0x123456
*  it is determined that the erase block size is 0x10000 at that address.
*  Then it is determined that a range of equally sized erase blocks starts
*  at 0x100000 and continues for 0x80000 bytes.
*
*  Most flash types only have one size for each type of block, which will
*  result in the address range being the same size as the entire flash
*  instance. Some flash types implement several block sizes, which will
*  be reflected by several ranges in the flash instance.
*
*  @input
*  characteristics         Pointer to a struct that will hold the results
*                          of the operation. See description of struct
*                          FlashCharacteristics for details.
*
*  @input
*  request_type            A code that selects the type of characteristics
*                          that shall be retrieved. See description of
*                          enum CharacteristicType for details.
*
*  @input
*  request_address         Identifies the flash address range that shall be
*                          examined. Any address within the range is valid.
*                          See use case example below.
*
*  @return
*  Result code. See description of enum FlashResult for details.
*
*  @restriction 
*     It is not allowed to kill a calling process while inside a FAM call.
*     This may lead to a deadlock situation, where future FAM calls from
*     other processes would hang forever.
*     The FAM will not detect termination of a FAM client.
*     See also FAM in the OSE Core Extentions User's Guide.
*  @break
*     Only available in supervisor mode processes.
*
*  @seealso default_restrictions_calling_functions
*
*  @example
*  Use case example:
*  A user needs to acquire the erase block size for all erase blocks in a
*  flash instance. The user knows the start and size of the instance
*  e.g. from following calling flash_get_instance.
*     Step 1:
*        Get the erase block sizes of the first range of equally sized
*        erase blocks. Use request_address = instance start address,
*        request_type = FLASH_ERASEBLOCK_RANGE.
*     Step 2:
*        Check if the end of the erase block range is equal to the end of
*        the entire instance. The procedure ends if they are equal.
*     Step 3:
*        Get the erase block sizes of the next range of equally sized
*        erase blocks. Use request_address = end of last range.
*        Repeat from step 2.
*/
FlashResult
flash_get_characteristics(FlashCharacteristics *characteristics,
                          CharacteristicType request_type,
                          FlashAddress request_address);


/**
*  @function flash_throttle_get_limit
*
*  @declaration
*           int
*           flash_throttle_get_limit();
*
*  @brief   Gets the current maximum number of simultaneous write/erase
*           operations allowed by the FAM throttle.
*
*  @long    The limit may be changed through calls to flash_throttle_down and
*           flash_throttle_up.
*
*           It is an error to call flash_throttle_get_limit without having
*           activated throttling first.
*
*  @return  Current throttle limit. A value <= 0 means that no write/erase
*           operations are allowed.
*
*  @restriction 
*     It is not allowed to kill a calling process while inside a FAM call.
*     This may lead to a deadlock situation, where future FAM calls from
*     other processes would hang forever.
*     The FAM will not detect termination of a FAM client.
*     See also FAM in the OSE Core Extentions User's Guide.
*  @break
*     Only available in supervisor mode processes.
*
*  @seealso default_restrictions_calling_functions
*/
int
flash_throttle_get_limit(void);

/**
*  @function flash_throttle_down
*
*  @declaration
*           int
*           flash_throttle_down();
*
*  @brief   Decreases the throttle limit by one, and returns the new limit.
*
*  @long    The function will not return until the total number of active
*           write/erase operations is within the new limit.
*
*           See description of flash_throttle_activate for further details.
*
*           It is an error to call flash_throttle_down without having
*           activated throttling first.
*
*  @return  New throttle limit. A value <= 0 means that no write/erase
*           operations are allowed.
*
*  @restriction 
*     It is not allowed to kill a calling process while inside a FAM call.
*     This may lead to a deadlock situation, where future FAM calls from
*     other processes would hang forever.
*     The FAM will not detect termination of a FAM client.
*     See also FAM in the OSE Core Extentions User's Guide.
*  @break
*     Only available in supervisor mode processes.
*
*  @seealso default_restrictions_calling_functions
*/
int
flash_throttle_down(void);

/**
*  @function flash_throttle_up
*
*  @declaration
*           int
*           flash_throttle_up();
*
*  @brief   Increases the throttle limit by one, and returns the new limit.
*
*  @long    See description of flash_throttle_activate for further details.
*
*           It is an error to call flash_throttle_up without having
*           activated throttling first.
*
*  @return  New throttle limit. A value <= 0 means that no write/erase
*           operations are allowed.
*
*  @restriction 
*     It is not allowed to kill a calling process while inside a FAM call.
*     This may lead to a deadlock situation, where future FAM calls from
*     other processes would hang forever.
*     The FAM will not detect termination of a FAM client.
*     See also FAM in the OSE Core Extentions User's Guide.
*  @break
*     Only available in supervisor mode processes.
*
*  @seealso default_restrictions_calling_functions
*/
int
flash_throttle_up(void);


/*
* End of declarations
*/

#ifdef __cplusplus
}
#endif

#endif /* ifndef _FLASH_API_H */
