/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2011 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */
/**
*  @public
*  @toc File_System:
*  @file ose_spi/flash_conf.h
*
*  @version
*
*
*  @brief API for configuring the Flash Access Manager (FAM).
*
*  @long The functions in this interface may only be called from code that
*  executes in supervisor mode and that is statically linked to the FAM.
*  Configuration is typically performed in an early OSE start phase.
*/

#ifndef _FLASH_CONF_H
#define _FLASH_CONF_H

#include "flash_drv.h"

#ifdef __cplusplus
extern "C" {
#endif


/**
*  @type FlashConfResult
*
*  @declaration
*           typedef enum FlashConfResult
*           {
*              FLASHCONF_EOK = 0,
*              FLASHCONF_ENOT_INITIALIZED = 0x00CC3000,
*              FLASHCONF_EUNSUPPORTED_DRIVER_VERSION = 0x00CC3001,
*              FLASHCONF_EDRIVER_INIT_FAILURE = 0x00CC3002,
*              FLASHCONF_ENO_SUCH_INSTANCE = 0x00CC3003
*           } FlashConfResult;
*
*  @brief   Result code for flash configuration operations.
*
*  @field   FLASHCONF_EOK
*           Success.
*
*  @field   FLASHCONF_ENOT_INITIALIZED
*           flash_initialize must be called before calling the failed command.
*
*  @field   FLASHCONF_EUNSUPPORTED_DRIVER_VERSION
*           The driver does not support the same interface version as
*           FAM requires.
*
*  @field   FLASHCONF_EDRIVER_INIT_FAILURE
*           The initialize driver call failed.
*
*  @field   FLASHCONF_ENO_SUCH_INSTANCE
*           Attempting to configure flash properties at an address where there
*           is no flash instance.
*/
typedef enum FlashConfResult
{
   FLASHCONF_EOK = 0,
   FLASHCONF_ENOT_INITIALIZED = 0x00CC3000,
   FLASHCONF_EUNSUPPORTED_DRIVER_VERSION = 0x00CC3001,
   FLASHCONF_EDRIVER_INIT_FAILURE = 0x00CC3002,
   FLASHCONF_ENO_SUCH_INSTANCE = 0x00CC3003
} FlashConfResult;


/**
*  @function flash_is_api_available
*
*  @declaration
*           Boolean flash_is_api_available(void)
*
*  @brief   This function determines whether the services provided via
*           flash_api.h are available.
*
*  @long    (In essence whether flash_install_bios_module has been called.)
*
*  @return  True if the functions provided via flash_api.h are available,
*           False otherwise.
*
*  @restriction 
*     It is not allowed to kill a calling process while inside a FAM call.
*     This may lead to a deadlock situation, where future FAM calls from
*     other processes would hang forever.
*     The FAM will not detect termination of a FAM client.
*     See also FAM in the OSE Core Extentions User's Guide.
*  @break
*     Only available in supervisor mode processes.
*
*  @seealso default_restrictions_calling_functions
*/
Boolean
flash_is_api_available(void);


/**
*  @macro IRQ_POLL_NO_PENDING_INTERRUPT
*
*  @brief Return value from interrupt polling function indicating that no
*  interrupt is pending.
*/
#define IRQ_POLL_NO_PENDING_INTERRUPT  (256)


/**
*  @function flash_initialize
*
*  @declaration
*           FlashConfResult flash_initialize(
*              int (*irq_poll_function)(void),
*              void (*micro_delay_timer)(U32 microseconds),
*              U32 max_simultaneous_users)
*
*  @brief   This function performs basic initialization of the flash
*           access module.
*
*  @long    It must be called before any other flash function is called.
*
*  @input   irq_poll_function
*           Pointer to function for polling of interrupt
*           request. NULL if no such function is available. The
*           function shall return IRQ_POLL_NO_PENDING_INTERRUPT
*           if there is NO pending interrupt, any other value
*           otherwise. The function must not alter the state of
*           pending interrupts. (NOTE: the standard OSE
*           INT_VECTOR_HANDLER fulfils these requirements on
*           some OSE architectures.)
*
* @input    micro_delay_timer
*           Pointer to function for hardware assisted micro
*           delay timeouts. NULL if no such function is
*           available.
*
* @input    micro_delay_timer.microseconds
*           Timeout argument in micro seconds.
*
* @input    max_simultaneous_users
*           FAM will fail if more than this number of processes
*           attempts to simultaneously do a flash access call.
*           Recommended to be greater or equal to the number
*           of processes that will use FAM. About 100 bytes of
*           buffer space will be allocated for each.
*
* @return   Result code. See description of enum FlashConfResult for details.
*
*  @restriction 
*     It is not allowed to kill a calling process while inside a FAM call.
*     This may lead to a deadlock situation, where future FAM calls from
*     other processes would hang forever.
*     The FAM will not detect termination of a FAM client.
*     See also FAM in the OSE Core Extentions User's Guide.
*  @break
*     Only available in supervisor mode processes.
*
*  @seealso default_restrictions_calling_functions
*/
FlashConfResult
flash_initialize(int (*irq_poll_function)(void),
                 void (*micro_delay_timer)(U32 microseconds),
                 U32 max_simultaneous_users);


/**
*  @function flash_throttle_activate
*
*  @declaration
*           FlashConfResult flash_throttle_activate(
*              int initial_limit)
*
*  @brief   This function activates and initializes the FAM throttling
*           functionality.
*
*  @long    Throttling imposes a limit on the total number of simultaneous
*           write/erase operations allowed by the FAM. The limit can
*           subsequently be changed through calls to flash_throttle_down
*           and flash_throttle_up.
*
*           Throttling is typically used to limit the total power consumption
*           in a system. E.g. by setting the limit to 1, one never has more than
*           one write/erase operation in progress (through the FAM) at any given
*           time.
*
*           The functions flash_throttle_down and flash_throttle_up can be used
*           to temporarily change the throttle limit, e.g. before starting some
*           power hungry operation in some other part of the system (such as
*           starting up some component or performing some power hungry operation
*           in a NAND flash, which is not controlled by the FAM).
*
*           Must be called after flash_initialize, but before
*           flash_install_bios_module.
*
*  @input   initial_limit
*           The initial limit to the total number of simultaneous
*           write/erase operations allowed by the FAM. A value <= 0 means
*           that no write/erase operations are allowed.
*
*  @return  Result code. See description of enum FlashConfResult for details.
*
*  @restriction 
*     It is not allowed to kill a calling process while inside a FAM call.
*     This may lead to a deadlock situation, where future FAM calls from
*     other processes would hang forever.
*     The FAM will not detect termination of a FAM client.
*     See also FAM in the OSE Core Extentions User's Guide.
*  @break
*     Only available in supervisor mode processes.
*
*  @seealso default_restrictions_calling_functions
*/
FlashConfResult
flash_throttle_activate(int initial_limit);


/**
*  @macro FLASH_DRIVER_MAX_NAME_LEN
*
*  @brief Maximum valid length, excluding terminating null character, for a
*  name string to be used with flash_install_driver and flash_add_instance.
*/
#define FLASH_DRIVER_MAX_NAME_LEN (23)


/**
*  @function flash_install_driver
*
*  @declaration
*           FlashConfResult flash_install_driver(
*              FlashDriver driver,
*              const char *driver_name,
*              const char *parameter_string)
*
*  @brief   This function installs a flash driver with the Flash Access Manager
*           and initializes the driver.
*
*  @input   driver
*           Pointer to the flash driver that shall be installed.
*
*  @input   driver_name
*           Pointer to a text string with a name that the user
*           assigns to the driver.
*
*  @input   parameter_string
*           Text string with initialization parameters. The  string should
*           be on the format TOKEN[=VALUE][,TOKEN[=VALUE]]... where each
*           individual driver defines valid tokens and values. An empty
*           string shall be used if there are no parameters.
*
* @return   Result code. See description of enum FlashConfResult for details.
*
*  @restriction 
*     It is not allowed to kill a calling process while inside a FAM call.
*     This may lead to a deadlock situation, where future FAM calls from
*     other processes would hang forever.
*     The FAM will not detect termination of a FAM client.
*     See also FAM in the OSE Core Extentions User's Guide.
*  @break
*     Only available in supervisor mode processes.
*
*  @seealso default_restrictions_calling_functions
*/
FlashConfResult
flash_install_driver(FlashDriver driver,
                     const char *driver_name,
                     const char *parameter_string);


/**
*  @function flash_add_instance
*
*  @declaration
*           size_t flash_add_instance(
*              const char *driver_name,
*              FlashAddress instance_start,
*              size_t instance_length,
*              const char *parameter_string)
*
*  @brief   Adds a flash device instance to the Flash Access Manager.
*
*  @input   driver_name
*           Pointer to a text string with the name of the desired driver.
*           This must be a name that has been assigned to a driver via the
*           flash_install_driver call. An empty string will cause automatic
*           selection among the installed drivers that support probing.
*
*  @input   instance_start
*           The start address of the flash that shall be installed as the
*           instance. A driver that supports probing for supported device
*           type will probe at this address. A driver that supports scanning
*           for flash size shall start at this address.
*
*  @input   instance_length
*           The size in bytes of the address range allotted to the device
*           instance. A driver that supports scanning for flash size will
*           not scan outside this range.
*
*  @input   parameter_string
*           Text string with instantiation parameters. The string should be
*           on the format TOKEN[=VALUE][,TOKEN[=VALUE]]... where each
*           individual driver defines valid tokens and values.
*           An empty string shall be used if there are no parameters.
*           A driver might support the following configurations and may also
*           support even more:
*           @break
*           @tab write_retry @tab Number of times to retry a write until
*                                 failing. 0 means only try once. @break
*           @tab erase_retry @tab Number of times to retry an erase until
*                                 failing. 0 means only try once. @break
*
*  @return  0 if the instantiation failed. Otherwise the length in bytes
*           of the instance that was installed.
*
*  @restriction 
*     It is not allowed to kill a calling process while inside a FAM call.
*     This may lead to a deadlock situation, where future FAM calls from
*     other processes would hang forever.
*     The FAM will not detect termination of a FAM client.
*     See also FAM in the OSE Core Extentions User's Guide.
*  @break
*     Only available in supervisor mode processes.
*
*  @seealso default_restrictions_calling_functions
*/
size_t
flash_add_instance(const char *driver_name,
                   FlashAddress instance_start,
                   size_t instance_length,
                   const char *parameter_string);


/**
*  @function flash_add_direct_access_region
*
*  @declaration
*           FlashConfResult flash_add_direct_access_region(
*              FlashAddress region_start,
*              size_t region_length)
*
*  @brief   Configures the FAM to provide flash synchronization for direct
*           flash access in a region of flash memory.
*
*  @long    This means that it will be possible to simultaneously execute
*           code in that region and erase or write to the flash.
*
*  @long    It is only possible to do this on a flash instance that have
*           already been added to FAM.
*
*  @input   region_start
*           The start address of the direct access region.
*
*  @input   region_length
*           The size in bytes of the direct access region.
*
*  @return  Result code. See description of enum FlashConfResult for details.
*
*  @restriction 
*     It is not allowed to kill a calling process while inside a FAM call.
*     This may lead to a deadlock situation, where future FAM calls from
*     other processes would hang forever.
*     The FAM will not detect termination of a FAM client.
*     See also FAM in the OSE Core Extentions User's Guide.
*  @break
*     Only available in supervisor mode processes.
*
*  @seealso default_restrictions_calling_functions
*/
FlashConfResult
flash_add_direct_access_region(FlashAddress region_start,
                               size_t region_length);

/**
*  @function flash_install_bios_module
*
*  @declaration
*           FlashConfResult flash_install_bios_module(void)
*
*  @brief   This function makes the Flash Access Manager install itself as
*           a BIOS  module. The services provided via flash_api.h become
*           available after this call.
*
*  @long    All configuration of FAM must be performed before this is done.
*           (Functions flash_install_driver, flash_add_instance and
*           flash_add_direct_access_region.)
*
* @return   Result code. See description of enum FlashConfResult for details.
*
*  @restriction 
*     It is not allowed to kill a calling process while inside a FAM call.
*     This may lead to a deadlock situation, where future FAM calls from
*     other processes would hang forever.
*     The FAM will not detect termination of a FAM client.
*     See also FAM in the OSE Core Extentions User's Guide.
*  @break
*     Only available in supervisor mode processes.
*
*  @seealso default_restrictions_calling_functions
*/
FlashConfResult
flash_install_bios_module(void);


#ifdef __cplusplus
}
#endif

#endif /* ifndef _FLASH_CONF_H */

