/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2006 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */
/**
*  @public
*  @toc File_System:
*  @file ose_spi/flash_drv.h
*
*  @brief   Declaration of API for flash drivers.
*
*  @long
*  This file declares an API that must be exported by flash drivers,
*  if the driver is to be used with the Flash Access Manager (FAM).
*
*  @long
*  HOW TO USE THIS INTERFACE
*
*  @long
*  Unlike most C header files, this file does not declare any function
*  prototypes, only type declarations of pointers to various functions.
*
*  A flash driver implementation must implement all the corresponding
*  functions and provide pointers to the functions via a struct of
*  function pointers. The declarations in this file describe the C
*  language equivalent of an abstract base class that must be implemented
*  by the driver.
*
*  This struct is in turn delivered to the FAM via a function of type
*  FlashDriver, that is installed in the FAM via the configuration
*  interface. (See flash_conf.h). Hence a flash driver will only need
*  to export one global function, of the FlashDriver type.
*
*
*  @long
*  VERSION HANDLING OF THIS INTERFACE
*
*  @long
*  All of the types with a "V1" suffix are part of version 1 of this
*  interface. They will never be changed. The entire interface is
*  defined by the struct with function pointers (FlashDriverInterfaceV1).
*
*  Changes are handled by creating new types with a higher version
*  suffix. A new version will require a new function pointer struct.
*  This new struct is free to re-use "V1" functions that are not changed
*  and to use new version functions that may either be changed versions
*  of previous functions or entirely new functions.
*
*  @long
*
*  BEHAVIOUR OF FLASH DRIVERS
*
*  @long
*  The normal state of the flash is read mode. The flash driver shall
*  always leave the flash in read mode after all calls that do not have
*  a specific purpose to set it in another mode. The flash shall always
*  be in read mode after a call that returns with an error.
*
*/


#ifndef _FLASH_DRV_H
#define _FLASH_DRV_H


#include "flash_api.h"
#include "osetypes.h"
#include <stddef.h>

#ifdef __cplusplus
extern "C" {
#endif


/**
*  @type    FlashDriverResult
*
*  @declaration
*           typedef enum FlashDriverResult
*           {
*              FLASHDRIVER_EOK                  = 0,
*              FLASHDRIVER_EIN_PROGRESS         = 0x00CC2000,
*              FLASHDRIVER_EUNKNOWN_REQUEST     = 0x00CC2001,
*              FLASHDRIVER_EHARDWARE_ERROR      = 0x00CC2002,
*              FLASHDRIVER_ENOT_SUPPORTED       = 0x00CC2003,
*              FLASHDRIVER_EPARAMETER_ERROR     = 0x00CC2004,
*              FLASHDRIVER_EPROTECTED           = 0x00CC2005,
*              FLASHDRIVER_EUNSUPPORTED_DEVICE  = 0x00CC2006,
*              FLASHDRIVER_EINVALID_DATA        = 0x00CC2007
*           } FlashDriverResult;
*
*  @brief   Result codes from flash driver operations.
*
*  @field   FLASHDRIVER_EOK
*           The function has completed successfully.
*
*  @field   FLASHDRIVER_EIN_PROGRESS
*           The operation is in progress. Additional poll calls must be done.
*           This code is only returned by the poll calls.
*
*  @field   FLASHDRIVER_EUNKNOWN_REQUEST
*           Type not recognized.
*
*  @field   FLASHDRIVER_EHARDWARE_ERROR
*           Operation could not be completed because of a hardware malfunction.
*           It is possible that part of the operation has been carried out.
*
*  @field   FLASHDRIVER_ENOT_SUPPORTED
*           The driver does not implement the requested function.
*
*  @field   FLASHDRIVER_EPARAMETER_ERROR
*           Invalid parameter value.
*
*  @field   FLASHDRIVER_EPROTECTED
*           The requested flash address is protected from the requested type
*           of access.
*
*  @field   FLASHDRIVER_EUNSUPPORTED_DEVICE
*           Instantiation failed because device probing did not detect a flash
*           device type that the driver supports.
*
*  @field   FLASHDRIVER_EINVALID_DATA
*           Operation could not be performed because the application specified
*           invalid data to write (i.e. tried to change bits the wrong way).
*/
typedef enum FlashDriverResult
{
   FLASHDRIVER_EOK                  = 0,
   FLASHDRIVER_EIN_PROGRESS         = 0x00CC2000,
   FLASHDRIVER_EUNKNOWN_REQUEST     = 0x00CC2001,
   FLASHDRIVER_EHARDWARE_ERROR      = 0x00CC2002,
   FLASHDRIVER_ENOT_SUPPORTED       = 0x00CC2003,
   FLASHDRIVER_EPARAMETER_ERROR     = 0x00CC2004,
   FLASHDRIVER_EPROTECTED           = 0x00CC2005,
   FLASHDRIVER_EUNSUPPORTED_DEVICE  = 0x00CC2006,
   FLASHDRIVER_EINVALID_DATA        = 0x00CC2007
} FlashDriverResult;


/**
*  @type    FlashInitializeDriverV1
*
*  @declaration
*           typedef FlashDriverResult
*           (*FlashInitializeDriverV1)(const char *parameter_string);
*
*  @brief   Pointer to function for initializing a flash device driver.
*
*  @long    This function shall be called before any of the other flash
*           driver functions are called. (Except FlashDriver).
*
*  @input   parameter_string
*           Text string with initialization parameters. The string should
*           be on the format TOKEN[=VALUE][,TOKEN[=VALUE]]... where each
*           individual driver defines valid tokens and values.
*
*  @return  Result code. See description of FlashDriverResult for details.
*           @link "flash_drv.h" "FlashDriverResult".
*/
typedef FlashDriverResult
(*FlashInitializeDriverV1)(const char *parameter_string);


/**
*  @type    FlashDriverCapabilitiesV1
*
*  @declaration
*           typedef struct FlashDriverCapabilitiesV1
*           {
*              size_t         instance_databuffer_size;
*              Boolean        multiple_instance_support;
*              Boolean        device_probing_support;
*           } FlashDriverCapabilitiesV1;
*
*  @brief   Holds the result from a function of type
*           FlashGetDriverCapabilitiesV1.
*
*  @field   instance_databuffer_size
*           A user supplied buffer of this size shall be passed in the
*           instance parameter in most calls. See description of
*           FlashInstantiateDeviceV1 for details.
*           @link "flash_drv.h" "FlashInstantiateDeviceV1"
*
*  @field   multiple_instance_support
*           True if the driver is capable of handling multiple device
*           instances.
*
*  @field   device_probing_support
*           True if the driver can detect a supported flash device type by
*           probing during instantiation.
*/
typedef struct FlashDriverCapabilitiesV1
{
   size_t         instance_databuffer_size;
   Boolean        multiple_instance_support;
   Boolean        device_probing_support;
} FlashDriverCapabilitiesV1;


/**
*  @type    FlashGetDriverCapabilitiesV1
*
*  @declaration
*           typedef const FlashDriverCapabilitiesV1 *
*           (*FlashGetDriverCapabilitiesV1)(void);
*
*  @brief   Pointer to function for retrieving information about driver
*           capabilities.
*
*  @return  Pointer to a struct containing information about driver
*           capabilities. The struct shall remain accessible for the
*           lifetime of the driver. There is no error return.
*           See description of FlashDriverCapabilitiesV1 for details.
*           @link "flash_drv.h" "FlashDriverCapabilitiesV1".
*/
typedef const FlashDriverCapabilitiesV1 *
(*FlashGetDriverCapabilitiesV1)(void);


/**
*  @type    FlashInstantiateDeviceV1
*
*  @declaration
*           typedef FlashDriverResult
*           (*FlashInstantiateDeviceV1)(void *instance,
*                                       FlashAddress instance_start,
*                                       size_t instance_length,
*                                       const char *parameter_string);
*
*  @brief   Pointer to a function for creating an association with an actual
*           flash device instance.
*
*  @long    This function shall detect the properties of the flash and set
*           up the instance descriptor buffer with all information it needs
*           in order to do further operations on the instance.
*
*           Direct flash access synchronization is not active during this
*           call. Hence the driver should disable interrupts whenever the
*           flash is not in read mode. This is to prevent any interrupt
*           handlers in the system from reading bad data or disturbing
*           the state machine of the flash chip.
*
*  @input   instance
*           Pointer to instance descriptor buffer. The driver will
*           initialize the contents of this with everything it needs to
*           know to keep track of the instance. The user will pass the
*           same struct to all following calls related to the same instance.
*           It is the users responsibility to allocate a size
*           FlashDriverCapabilitiesV1.instance_databuffer_size bytes long
*           buffer before this call.
*
*  @input   instance_start
*           The start address of the flash that shall be installed as the
*           instance. A driver that supports probing for supported device
*           type shall probe at this address. A driver that supports
*           scanning for flash size shall start at this address.
*
*  @input   instance_length
*           The size in bytes of the address range allotted to the device
*           instance. A driver that supports scanning for flash size shall
*           stop scanning after this range.
*
*  @input   parameter_string
*           Text string with instantiation parameters. The string should be
*           on the format TOKEN[=VALUE][,TOKEN[=VALUE]]... where each
*           individual driver defines valid tokens and values.
*           A driver might support the following configurations and may also
*           support even more:
*           @break
*           @tab write_retry @tab Number of times to retry a write until
*                                 failing. 0 means only try once. @break
*           @tab erase_retry @tab Number of times to retry an erase until
*                                 failing. 0 means only try once. @break
*
*  @return  Result code. See description of FlashDriverResult for details.
*           @link "flash_drv.h" "FlashDriverResult".
*/
typedef FlashDriverResult
(*FlashInstantiateDeviceV1)(void *instance,
                            FlashAddress instance_start,
                            size_t instance_length,
                            const char *parameter_string);


/**
*  @type    FlashCharacteristicsV1
*
*  @declaration
*           typedef struct FlashCharacteristicsV1
*           {
*              FlashAddress   range_start;
*              size_t         range_length;
*              size_t         block_length;
*           } FlashCharacteristicsV1;
*
*  @brief   Holds the result from a function of type
*           FlashGetCharacteristicsV1.
*
*  @field   range_start
*           The first flash address of a range.
*
*  @field   range_length
*           The number of bytes in an address range.
*
*  @field   block_length
*           The block size in bytes when the range is a range of equally
*           sized blocks. The characteristics of a block is determined by
*           the requested CharacteristicType.
*/
typedef struct FlashCharacteristicsV1
{
   FlashAddress   range_start;
   size_t         range_length;
   size_t         block_length;
} FlashCharacteristicsV1;


/**
*  @type    FlashGetCharacteristicsV1
*
*  @declaration
*           typedef FlashDriverResult
*           (*FlashGetCharacteristicsV1)(
*              FlashCharacteristicsV1 *characteristics,
*              void *instance,
*              CharacteristicType request_type,
*              FlashAddress request_address);
*
*  @brief   Pointer to function for retrieving the characteristics of an
*           address range in a flash instance.
*
*  @long    The address range is sometimes the size of the entire flash
*           instance, sometimes a portion of the instance. A client must use
*           several calls to get the characteristics of the entire instance
*           in the latter case.
*
*           This function is very similar to get_flash_characteristics in
*           flash_api.h. Refer to that function for further documentation.
*
*           This function must work even if a flash operation is in
*           progress. It is recommended that this function should
*           retrieve all needed information from the instance descriptor
*           buffer.
*
*           It is expected that the FAM driver can handle an unknown
*           value in request_type by returning FLASHDRIVER_EUNKNOWN_REQUEST.
*
*  @input   characteristics
*           Pointer to a struct that will hold the result of the operation.
*           See description of struct FlashCharacteristicsV1 for details.
*           @link "flash_drv.h" "FlashCharacteristicsV1".
*
*  @input   instance
*           Pointer to instance descriptor buffer.
*           See description of FlashInstantiateDeviceV1 for details.
*           @link "flash_drv.h" "FlashInstantiateDeviceV1".
*
*  @input   request_type
*           A code that selects the type of characteristics that shall be
*           retrieved. If an unknown request_type is given, the driver
*           will return FLASHDRIVER_EUNKNOWN_REQUEST.
*           See description of CharacteristicType in flash_api.h for details.
*           @link "flash_api.h" "CharacteristicType".
*
*  @input   request_address
*           Select flash address range that shall be examined. Any address
*           within the range is valid. The parameter is not used if
*           request_type is FLASH_INSTANCE_SIZE.
*
*  @return  Result code. See description of FlashDriverResult for details.
*/
typedef FlashDriverResult
(*FlashGetCharacteristicsV1)(FlashCharacteristicsV1 *characteristics,
                             void *instance,
                             CharacteristicType request_type,
                             FlashAddress request_address);


/**
*  @type    FlashInstanceCapabilitiesV1
*
*  @brief   Holds the result from FlashGetInstanceCapabilitiesV1.
*
*  @field   verify_write
*           Determines if the user should read back data after a write
*           operation, and a recommendation about whether the user should
*           attempt to write again in case of error.
*           * 0 if the user is not recommended to read back.
*           * 1 or higher is a recommendation to read back. The value is the
*             recommended number of write attempts.
*
*  @field   verify_erase
*           The same as verify write, but applied to erase.
*
*  @field   suspend_write_support
*           True if the driver implements write suspend.
*
*  @field   suspend_erase_support
*           True if the driver implements erase suspend.
*/
typedef struct FlashInstanceCapabilitiesV1
{
   unsigned       verify_write;
   unsigned       verify_erase;
   Boolean        suspend_write_support;
   Boolean        suspend_erase_support;
} FlashInstanceCapabilitiesV1;


/**
*  @type    FlashGetInstanceCapabilitiesV1
*
*  @declaration
*           typedef const FlashInstanceCapabilitiesV1 *
*           (*FlashGetInstanceCapabilitiesV1)(void *instance);
*
*  @brief   Pointer to function for retrieving information about the
*           capabilities of a device instance.
*
*  @long    This function must work even if a flash operation is in
*           progress. It is recommended that this function should
*           retrieve all needed information from the instance descriptor
*           buffer.
*
*  @input   instance
*           Pointer to instance descriptor buffer.
*           See description of FlashInstantiateDeviceV1 for details.
*           @link "flash_drv.h" "FlashInstantiateDeviceV1".
*
*  @return  Pointer to a struct containing information about instance
*           capabilities. The struct shall remain accessible for the
*           lifetime of the instance. There is no error return.
*           See description of FlashInstanceCapabilitiesV1 for details.
*           @link "flash_drv.h" "FlashInstanceCapabilitiesV1".
*/
typedef const FlashInstanceCapabilitiesV1 *
(*FlashGetInstanceCapabilitiesV1)(void *instance);


/**
*  @type    FlashOperationParametersV1
*
*  @brief   A block of operation parameters for write and erase.
*
*  @long    This parameter struct is valid for one write or erase operation.
*           FAM sets up a number of input parameters that describe the
*           desired operation at the operation start.
*
*           The driver will then be presented with the same parameter struct
*           for all poll, suspend and resume calls that follow and relate to
*           the same operation.
*
*           The output parameters contain feedback to FAM from the driver.
*           The driver is allowed to modify all the other fields and keep any
*           type of information in them.
*
*           The main purpose of this struct is to let the driver keep state
*           information about the operation between calls without a need for
*           internal memory management. The advantage of using this block
*           instead of a driver-internal buffer is to easier support multiple
*           simultaneous operations.
*
*  @long    This struct has been extended with one field in version two of
*           the interface. This extension is backwards compatible, i.e. a
*           driver expecting version one of this struct can read it as well as
*           a driver expecting version two. Since the change is backwards
*           compatible, the name of the struct has not been changed.
*
*  @field   destination
*           The first flash address of a range that shall be written or
*           erased. (Input to FlashWriteStartV1 and FlashEraseStartV1)
*
*  @field   source_data
*           Pointer to data that shall be written to the flash.
*           (Input to FlashWriteStartV1)
*
*  @field   request_length
*           The number of bytes that shall be written.
*           (Input to FlashWriteStartV1)
*
*  @field   delay_time
*           Recommended time to wait until next poll. Expressed in
*           microseconds for write and milliseconds for erase.
*           (Output from all flash operation calls)
*
*  @field   opaque
*           The driver may use this field for whatever purpose it needs.
*
*  @field   access_type
*           Some flash has different types of access depending on what kind of
*           information that is written. E.g. Sibley NOR flash has data access
*           type as well as control. This parameter selects which of these to
*           use. The exact meaning of this parameter is dependent on memory
*           technology.
*
*           This field is only used for write accesses.
*           If old version of read and write in flash_api.h are used
*           (@link "flash_api.h" "flash_read"() and
*           @link "flash_api.h" "flash_write"()),
*           this field will contain the value FLASH_ACCESS_DATA.
*
*           See @link "flash_api.h" "FlashAccessType" for valid values.
*
*           Note: This field is only available in version 2 of the interface.
*/
typedef struct FlashOperationParametersV1
{
   FlashAddress   destination;
   const void     *source_data;
   size_t          request_length;
   U32             delay_time;
   U32             opaque[12];
   FlashAccessType access_type;
} FlashOperationParametersV1;


/**
*  @type    FlashWriteStartV1
*
*  @declaration
*           typedef FlashDriverResult
*           (*FlashWriteStartV1)(void *instance,
*                                FlashOperationParametersV1 *parameters);
*
*  @brief   Pointer to function for starting a flash write operation.
*
*  @input   instance
*           Pointer to instance descriptor buffer.
*           See description of FlashInstantiateDeviceV1 for details.
*           @link "flash_drv.h" "FlashInstantiateDeviceV1".
*
*  @input   parameters
*           Pointer to a struct of operation command parameters.
*           The input fields are set up by the caller, prior to the call.
*           Output fields shall be valid after the call.
*           See FlashOperationParametersV1 for details.
*           @link "flash_drv.h" "FlashOperationParametersV1".
*
*  @return  Result code. See description of FlashDriverResult for details.
*           @link "flash_drv.h" "FlashDriverResult".
*/
typedef FlashDriverResult
(*FlashWriteStartV1)(void *instance,
                     FlashOperationParametersV1 *parameters);


/**
*  @type    FlashWritePollV1
*
*  @declaration
*           typedef FlashDriverResult
*           (*FlashWritePollV1)(void *instance,
*                               FlashOperationParametersV1 *parameters);
*
*  @brief   Pointer to function for polling to check if a write operation
*           has completed.
*
*  @long    The flash shall be in read mode when this function returns OK.
*
*  @input   instance
*           Pointer to instance descriptor buffer.
*           See description of FlashInstantiateDeviceV1 for details.
*           @link "flash_drv.h" "FlashInstantiateDeviceV1".
*
*  @input   parameters
*           Pointer to a struct of operation command parameters. The caller
*           shall supply the same pointer as was used in the start call.
*           Output fields shall be valid after the call.
*           See FlashOperationParametersV1 for details.
*           @link "flash_drv.h" "FlashOperationParametersV1".
*
*  @return  Result code. See description of FlashDriverResult for details.
*           @link "flash_drv.h" "FlashDriverResult".
*/
typedef FlashDriverResult
(*FlashWritePollV1)(void *instance,
                    FlashOperationParametersV1 *parameters);


/**
*  @type    FlashWriteSuspendV1
*
*  @declaration
*           typedef FlashDriverResult
*           (*FlashWriteSuspendV1)(void *instance,
*                                  FlashOperationParametersV1 *parameters);
*
*  @brief   Pointer to function for suspending a write operation that is in
*           progress.
*
*  @long    The flash shall be in read mode after this call.
*
*  @input   instance
*           Pointer to instance descriptor buffer.
*           See description of FlashInstantiateDeviceV1 for details.
*           @link "flash_drv.h" "FlashInstantiateDeviceV1".
*
*  @input   parameters
*           Pointer to a struct of operation command parameters. The caller
*           shall supply the same pointer as was used in the start call.
*           Output fields shall be valid after the call.
*           See FlashOperationParametersV1 for details.
*           @link "flash_drv.h" "FlashOperationParametersV1".
*
*  @return  Result code. See description of FlashDriverResult for details.
*           @link "flash_drv.h" "FlashDriverResult".
*/
typedef FlashDriverResult
(*FlashWriteSuspendV1)(void *instance,
                       FlashOperationParametersV1 *parameters);


/**
*  @type    FlashWriteResumeV1
*
*  @declaration
*           typedef FlashDriverResult
*           (*FlashWriteResumeV1)(void *instance,
*                                 FlashOperationParametersV1 *parameters);
*
*  @brief   Pointer to function for resuming a write operation that has been
*           suspended.
*
*  @input   instance
*           Pointer to instance descriptor buffer.
*           See description of FlashInstantiateDeviceV1 for details.
*           @link "flash_drv.h" "FlashInstantiateDeviceV1".
 *
*  @input   parameters
*           Pointer to a struct of operation command parameters. The caller
*           shall supply the same pointer as was used in the start call.
*           Output fields shall be valid after the call.
*           See FlashOperationParametersV1 for details.
*           @link "flash_drv.h" "FlashOperationParametersV1".
*
*  @return  Result code. See description of FlashDriverResult for details.
*           @link "flash_drv.h" "FlashDriverResult".
*/
typedef FlashDriverResult
(*FlashWriteResumeV1)(void *instance,
                      FlashOperationParametersV1 *parameters);


/**
*  @type    FlashEraseStartV1
*
*  @declaration
*           typedef FlashDriverResult
*           (*FlashEraseStartV1)(void *instance,
*                                FlashOperationParametersV1 *parameters);
*
*  @brief   Pointer to function for starting a flash erase operation of one
*           erase block.
*
*  @input   instance
*           Pointer to instance descriptor buffer.
*           See description of FlashInstantiateDeviceV1 for details.
*           @link "flash_drv.h" "FlashInstantiateDeviceV1".
*
*  @input   parameters
*           Pointer to a struct of operation command parameters.
*           The input fields are set up by the caller, prior to the call.
*           Output fields shall be valid after the call.
*           See FlashOperationParametersV1 for details.
*           @link "flash_drv.h" "FlashOperationParametersV1"
*
*  @return  Result code. See description of FlashDriverResult for details.
*           @link "flash_drv.h" "FlashDriverResult".
*/
typedef FlashDriverResult
(*FlashEraseStartV1)(void *instance,
                     FlashOperationParametersV1 *parameters);


/**
*  @type    FlashErasePollV1
*
*  @declaration
*           typedef FlashDriverResult
*           (*FlashErasePollV1)(void *instance,
*                               FlashOperationParametersV1 *parameters);
*
*  @brief   Pointer to function for polling to check if an erase operation
*           has completed.
*
*  @long    The flash shall be in read mode when this function returns OK.
*
*  @input   instance
*           Pointer to instance descriptor buffer.
*           See description of FlashInstantiateDeviceV1 for details.
*           @link "flash_drv.h" "FlashInstantiateDeviceV1".
*
*  @input   parameters
*           Pointer to a struct of operation command parameters. The caller
*           shall supply the same pointer as was used in the start call.
*           Output fields shall be valid after the call.
*           See FlashOperationParametersV1 for details.
*           @link "flash_drv.h" "FlashOperationParametersV1".
*
*  @return  Result code. See description of FlashDriverResult for details.
*           @link "flash_drv.h" "FlashDriverResult".
*/
typedef FlashDriverResult
(*FlashErasePollV1)(void *instance,
                    FlashOperationParametersV1 *parameters);


/**
*  @type    FlashEraseSuspendV1
*
*  @declaration
*           typedef FlashDriverResult
*           (*FlashEraseSuspendV1)(void *instance,
*                                  FlashOperationParametersV1 *parameters);
*
*  @brief   Pointer to function for suspending an erase operation that is in
*           progress.
*
*  @long    The flash shall be in read mode after this call.
*
*  @input   instance
*           Pointer to instance descriptor buffer.
*           See description of FlashInstantiateDeviceV1 for details.
*           @link "flash_drv.h" "FlashInstantiateDeviceV1".
*
*  @input   parameters
*           Pointer to a struct of operation command parameters. The caller
*           shall supply the same pointer as was used in the start call.
*           Output fields shall be valid after the call.
*           See FlashOperationParametersV1 for details.
*           @link "flash_drv.h" "FlashOperationParametersV1".
*
*  @return  Result code. See description of FlashDriverResult for details.
*           @link "flash_drv.h" "FlashDriverResult".
*/
typedef FlashDriverResult
(*FlashEraseSuspendV1)(void *instance,
                       FlashOperationParametersV1 *parameters);


/**
*  @type    FlashEraseResumeV1
*
*  @declaration
*           typedef FlashDriverResult
*           (*FlashEraseResumeV1)(void *instance,
*                                 FlashOperationParametersV1 *parameters);
*
*  @brief   Pointer to function for resuming an erase operation that has
*           been suspended.
*
*  @input   instance
*           Pointer to instance descriptor buffer.
*           See description of FlashInstantiateDeviceV1 for details.
*           @link "flash_drv.h" "FlashInstantiateDeviceV1".
*
*  @input   parameters
*           Pointer to a struct of operation command parameters. The caller
*           shall supply the same pointer as was used in the start call.
*           Output fields shall be valid after the call.
*           See FlashOperationParametersV1 for details.
*           @link "flash_drv.h" "FlashOperationParametersV1".
*
*  @return  Result code. See description of FlashDriverResult for details.
*           @link "flash_drv.h" "FlashDriverResult".
*/
typedef FlashDriverResult
(*FlashEraseResumeV1)(void *instance,
                      FlashOperationParametersV1 *parameters);


/**
*  @type    FlashReadV2
*
*  @brief   Pointer to function for reading from flash.
*
*  @long    Some flashes, like Sibley flashes, has different access types.
*           Unlike standard NOR flashes, special measures has to be taken
*           to read data from them. This function has the knowledge how to
*           read from the flash for the different access types.
*
*  @input   instance
*           Pointer to instance descriptor buffer.
*           See description of FlashInstantiateDeviceV1 for details.
*           @link "flash_drv.h" "FlashInstantiateDeviceV1".
*
*  @input   destination
*           Pointer to buffer to store the data read from flash.
*
*  @input   source
*           This first flash address of a range that shall be read.
*
*  @input   nr_bytes
*           Number of bytes to read.
*
*  @input   access_type
*           Some flashes has different types of access depending on what kind
*           of information that is read. E.g. Sibley NOR flash has data access
*           type as well as control. This parameter selects which of these to
*           use. The exact meaning of this parameter is dependent on memory
*           technology.
*
*           If old version of read and write in flash_api.h are used
*           (@link "flash_api.h" "flash_read"() and
*            @link "flash_api.h" "flash_write"()),
*           this field will contain the value FLASH_ACCESS_DATA.
*
*           See @link "flash_api.h" "FlashAccessType" for valid values.
*
*
*  @return  Result code. See description of FlashDriverResult for details.
*           @link "flash_drv.h" "FlashDriverResult".
*/
typedef FlashDriverResult
(*FlashReadV2)(void *instance,
               void *destination,
               FlashAddress source,
               size_t nr_bytes,
               FlashAccessType access_type);


/**
*  @type    FlashCheckContentV2
*
*  @brief   Pointer to function for checking content of flash
*
*  @long    The function pointed to shall compare a supplied buffer
*           against the content of the flash. If a NULL pointer is supplied
*           as the buffer the function shall compare against 0xFF (erased
*           state).The number of bytes to compare is also supplied. The number
*           of matching bytes shall be returned in "nr_matching_bytes".
*           Only driver errors shall be returned as the function's return
*           value.
*
*  @input   instance
*           Pointer to instance descriptor buffer.
*           See description of FlashInstantiateDeviceV1 for details.
*           @link "flash_drv.h" "FlashInstantiateDeviceV1".
*
*  @input   buffer_to_check
*           Pointer to start of data to check
*           (a NULL pointer for an erase check).
*
*  @input   address
*           Pointer to start of flash area to check.
*
*  @input   nr_bytes
*           Number of bytes to compare.
*
*  @output  nr_matching_bytes
*           Pointer to return number of matching bytes found.
*
*  @input   access_type
*           Some flash has different types of access depending on what kind of
*           information that is read. E.g. Sibley NOR flash has data access
*           type as well as control. This parameter selects which of these to
*           use. The exact meaning of this parameter is dependent on memory
*           technology.
*
*           See @link "flash_api.h" "FlashAccessType" for valid values.
*
*
*  @return  Result code. See description of FlashDriverResult for details.
*           @link "flash_drv.h" "FlashDriverResult".
*/
typedef FlashDriverResult
(*FlashCheckContentV2)(void *instance,
                     const void *buffer_to_check,
                     FlashAddress address,
                     size_t nr_bytes,
                     size_t *nr_matching_bytes,
                     FlashAccessType access_type);


/**
*  @type    FlashDriverInterfaceV1
*
*  @declaration
*           typedef struct FlashDriverInterfaceV1
*           {
*              FlashInitializeDriverV1          initialize_driver;
*              FlashGetDriverCapabilitiesV1     get_driver_capabilities;
*              FlashInstantiateDeviceV1         instantiate_device;
*              FlashGetInstanceCapabilitiesV1   get_instance_capabilities;
*              FlashGetCharacteristicsV1        get_characteristics;
*              FlashWriteStartV1                write_start;
*              FlashWritePollV1                 write_poll;
*              FlashWriteSuspendV1              write_suspend;
*              FlashWriteResumeV1               write_resume;
*              FlashEraseStartV1                erase_start;
*              FlashErasePollV1                 erase_poll;
*              FlashEraseSuspendV1              erase_suspend;
*              FlashEraseResumeV1               erase_resume;
*           } FlashDriverInterfaceV1;
*
*  @brief   This struct contains function pointers to all functions that are
*           included in version 1 of the flash driver interface.
*
*  @long    A driver that implements version 1 of the interface must define
*           all functions according to the prototypes.
*
*           The individual fields are described by the corresponding function
*           pointer type descriptions.
*
*  @field initialize_driver The function pointer to the function that
*         initializes the flash driver. This function must be called before any of
*         the other flash driver functions are called (except cfi_flash_driverentry)
*         @link "flash_drv.h" "FlashInitializeDriverV1".
*
*  @field get_driver_capabilities The function pointer to the function that
*         retrieves information about driver capabilities. Returns a pointer to a
*         struct containing information about driver capabilities.
*         You can call this function through out the lifetime of the driver.
*         @link "flash_drv.h" "FlashGetDriverCapabilitiesV1".
*
*  @field instantiate_device The function pointer to the function that
*         instantiates a specific flash driver to a specific address range.
*         I.e. create an association with an actual flash device instance. Also
*         it will prepare for access of that flash instance.
*         @link "flash_drv.h" "FlashInstantiateDeviceV1".
*
*  @field get_instance_capabilities The function pointer to the function through
*         which you can use to get the capabilities of the flash instance. '
*         The information you can get is:
*            base - Start address of the flash image.
*            size - Length of flash instance.
*            flags - A field indicating the device geometry.
*            width - Bus width (only used during query).
*            offset - Endian adaptation.
*            db - CFI query information.
*            caps - Flash capabilities.
*            ops - Flash operations.
*         @link "flash_drv.h" "FlashInstanceCapabilitiesV1".
*
*  @field get_characteristics  The function pointer to a function that
*         retrieves the characteristics of an address range in a flash instance.
*         The address range is sometimes the size of the entire flash instance, sometimes
*         a portion of the instance. A client must use several calls to get the
*         characteristics of the entire instance in the latter case.
*         @link "flash_drv.h" "FlashGetCharacteristicsV1"
*
*  @field write_start The function pointer to the function that will start a flash
*         write operation at a given instance with given parameters.
*         @link "flash_drv.h" "FlashWriteStartV1".
*
*  @field write_poll The function pointer to a function that will poll if a write
*         operation has completed.
*         @link "flash_drv.h" "FlashWritePollV1".
*
*  @field write_suspend The function pointer to a function that will suspend a
*         write operation in progress.
*         @link "flash_drv.h" "FlashWriteSuspendV1"
*
*  @field write_resume The function pointer to the function that will resume
*         a previously suspended write operation.
*         @link "flash_drv.h" "FlashWriteResumeV1".
*
*  @field erase_start The function pointer to the function that initiates a flash
*         erase operation.
*         @link "flash_drv.h" "FlashEraseStartV1".
*
*  @field erase_poll The function pointer to the function that will poll for the
*         completion of an erase operation.
*         @link "flash_drv.h" "FlashErasePollV1".
*
*  @field erase_suspend The function pointer to the function that will suspend an
*         ongoing erase operation.
*         @link "flash_drv.h" "FlashEraseSuspendV1".
*
*  @field erase_resume The function pointer to the function that will resume a
*         suspended erase operation.
*         @link "flash_drv.h" "FlashEraseResumeV1".
*/
typedef struct FlashDriverInterfaceV1
{
   FlashInitializeDriverV1          initialize_driver;
   FlashGetDriverCapabilitiesV1     get_driver_capabilities;
   FlashInstantiateDeviceV1         instantiate_device;
   FlashGetInstanceCapabilitiesV1   get_instance_capabilities;
   FlashGetCharacteristicsV1        get_characteristics;
   FlashWriteStartV1                write_start;
   FlashWritePollV1                 write_poll;
   FlashWriteSuspendV1              write_suspend;
   FlashWriteResumeV1               write_resume;
   FlashEraseStartV1                erase_start;
   FlashErasePollV1                 erase_poll;
   FlashEraseSuspendV1              erase_suspend;
   FlashEraseResumeV1               erase_resume;
} FlashDriverInterfaceV1;


/**
*  @type    FlashDriverInterfaceV2
*
*  @brief   This struct contains function pointers to all functions that are
*           included in version 2 of the flash driver interface.
*
*  @long    A driver that implements version 2 of the interface must define
*           all functions according to the prototypes. Basically this struct is
*           the same as the version 1 interface but it has an addition of the new
*           functions.
*
*           The individual fields are described by the corresponding function
*           pointer type descriptions.
*
*  @field initialize_driver - @link "flash_drv.h" "FlashDriverInterfaceV1"
*  @field get_driver_capabilities - @link "flash_drv.h" "FlashDriverInterfaceV1"
*  @field instantiate_device - @link "flash_drv.h" "FlashDriverInterfaceV1"
*  @field get_instance_capabilities - @link "flash_drv.h" "FlashDriverInterfaceV1"
*  @field get_characteristics - @link "flash_drv.h" "FlashDriverInterfaceV1"
*  @field write_start - @link "flash_drv.h" "FlashDriverInterfaceV1"
*  @field write_poll - @link "flash_drv.h" "FlashDriverInterfaceV1"
*  @field write_suspend - @link "flash_drv.h" "FlashDriverInterfaceV1"
*  @field write_resume - @link "flash_drv.h" "FlashDriverInterfaceV1"
*  @field erase_start - @link "flash_drv.h" "FlashDriverInterfaceV1"
*  @field erase_poll - @link "flash_drv.h" "FlashDriverInterfaceV1"
*  @field erase_suspend - @link "flash_drv.h" "FlashDriverInterfaceV1"
*  @field erase_resume - @link "flash_drv.h" "FlashDriverInterfaceV1"
*  @field read - Function pointer to a function that will read with
*                knowledge of different access types. @link "flash_drv.h" "FlashReadV2".
*  @field check_content - Pointer to function for checking content of flash.
*         @link "flash_drv.h" "FlashCheckContentV2".
*/
typedef struct FlashDriverInterfaceV2
{
   FlashInitializeDriverV1          initialize_driver;
   FlashGetDriverCapabilitiesV1     get_driver_capabilities;
   FlashInstantiateDeviceV1         instantiate_device;
   FlashGetInstanceCapabilitiesV1   get_instance_capabilities;
   FlashGetCharacteristicsV1        get_characteristics;
   FlashWriteStartV1                write_start;
   FlashWritePollV1                 write_poll;
   FlashWriteSuspendV1              write_suspend;
   FlashWriteResumeV1               write_resume;
   FlashEraseStartV1                erase_start;
   FlashErasePollV1                 erase_poll;
   FlashEraseSuspendV1              erase_suspend;
   FlashEraseResumeV1               erase_resume;
   FlashReadV2                      read;
   FlashCheckContentV2              check_content;
} FlashDriverInterfaceV2;


/**
*  @type    FlashDriverInterface
*
*  @brief   A union of all defined flash driver interface versions.
*
*  @long    Version two of the interface contains only a few extensions to the
*           version one, and is backwards compatible.
*
* @field v1 Version 1 of the flash driver interface.
* @field v2 Version 2 of the flash driver interface.
*/
union FlashDriverInterface
{
   FlashDriverInterfaceV1     v1;
   FlashDriverInterfaceV2     v2;
};


/**
*  @type    FlashDriver
*
*  @declaration
*           typedef const union FlashDriverInterface *
*           (*FlashDriver)(U32 interface_version);
*
*  @brief   Pointer to function to get the interface of a flash device driver.
*
*  @long    A flash driver must export a function of this type. All the other
*           driver functions are accessed via the function pointers in the
*           interface struct.
*
*  @input   interface_version
*           The requested interface version number.
*           The value 1 corresponds to interface v1.
*
*  @return  NULL if the requested interface version is not supported by the
*           driver. Otherwise a pointer to a union FlashDriverInterface where
*           the interface version member that corresponds to the requested
*           version number is valid.
*/
typedef const union FlashDriverInterface *
(*FlashDriver)(U32 interface_version);


/*
* End of declarations
*/

#ifdef __cplusplus
}
#endif

#endif /* ifndef _FLASH_DRV_H */


