/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2006 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

/**
 * @public
 *
 * @toc OSE:
 *
 * @file ose_spi/fmd.h
 *
 * @brief Freeze-mode Dumper Interface.
 *
 * @long This file defines the interface to the Freeze-mode Dumper. It can
 * be used in a freeze-mode environment (such as in the system error handler
 * after a fatal error) to fetch information on the system state from OSE core
 * components (such as the Executive, the Program Manager and the Heap) and
 * store the information for later analysis.
 *
 * @long The Freeze-mode Dumper retrieves the information it saves through the
 * freeze-mode interfaces (*fmi.h) of the components it supports.
 *
 * @long The Freeze-mode Dumper uses the RAMDUMP to store the information it
 * retrieves. The RAMDUMP must therefore be configured and included in the core
 * image. The produced dumps can be retrieved either through the RAMDUMP
 * interface or through the signal interface of the Run-Mode Monitor.
 *
 * @seealso ose_spi/efmi.h
 * @seealso ose_spi/pfmi.h
 * @seealso ose_spi/hfmi.h
 * @seealso ose_spi/ramdump.h
 * @seealso ose_spi/monitor_dumps.sig
 */

#ifndef _FMD_H
#define _FMD_H

#ifdef __cplusplus
extern "C" {
#endif

#include "ose.h"

#include "efmi.h"
#include "pfmi.h"
#include "hfmi.h"
#include "cpu.h"

#include "stdint.h"

/**
 * @type FmdStatus
 * @brief Type of status codes.
 *
 * @long Freeze-mode Dumper functions return error information in the form
 * of status codes as defined in the freeze-mode interface of the OSE core
 * component from which the particular Freeze-mode Dumper function retrieves
 * information (>0) or as defined in the RAMDUMP interface (<0).
 *
 * @long Errors retrieving information are thus indicated with positive status
 * codes and errors storing the information retrieved are indicated with
 * negative status codes. Zero indicates success.
 *
 * @seealso EfmiStatus
 * @seealso HfmiStatus
 * @seealso PfmiStatus
 * @seealso RamdumpStatus
 */
typedef int32_t FmdStatus;

/**
 * @type FmdScope
 * @brief Type of scope
 *
 * @seealso FmdScopeValue
 */
typedef uint32_t FmdScope;

/**
 * @type FmdScopeValue
 *
 * @brief Scope type constants used to select the set of processes (or
 * other entities with a PID) for which information or memory contents should
 * be saved.
 *
 * @field FMD_SCOPE_SYSTEM All processes.
 * @field FMD_SCOPE_PROCESS The specified process.
 * @field FMD_SCOPE_BLOCK The processes in the specified block.
 * @field FMD_SCOPE_SEGMENT The processes in the specified segment.
 *
 * @seealso FmdScope
 */
enum FmdScopeValue
{
   FMD_SCOPE_SYSTEM      = 0x00,
   FMD_SCOPE_SEGMENT     = 0x08,
   FMD_SCOPE_BLOCK       = 0x20,
   FMD_SCOPE_PROCESS     = 0x80
};


/**
 * @type FmdInfo
 * @brief Type of information
 *
 * @seealso FmdInfoValue
 */
typedef uint32_t FmdInfo;

/**
 * @type FmdInfoValue
 *
 * @brief Information type constants used to select type and level of detail
 * of the information to be saved.
 *
 * @field FMD_INFO_PROCESS
 * Brief process / block / segment information for processes in the scope.
 * @field FMD_INFO_PROCESS_LONG
 * Detailed process / block / segment information for processes in the scope.
 * @field FMD_INFO_STACK_USAGE
 * Stack location, size and usage information for processes in the scope.
 * @field FMD_INFO_ENV_VARS
 * Environment variables for blocks and processes in the scope.
 * @field FMD_INFO_SIGNAL_QUEUE
 * Signal information for signals in the queue of all processes in the scope.
 * @field FMD_INFO_POOL_INFO
 * Pool information for all pools processes in the scope can use.
 * @field FMD_INFO_POOL_SIGNALS
 * Signal information for all signal buffers in all pools processes in the
 * scope can use.
 * @field FMD_INFO_POOL_SIGNALS_OWNED
 * Signal information for all signal buffers that processes in the scope own.
 * @field FMD_INFO_SYS_PARAM
 * System parameters.
 * @field FMD_INFO_REGISTERS
 * Register values for all processes in the scope.
 * @field FMD_INFO_LOAD_MODULE_INFO
 * Load module and relocation information for the programs to which the
 * processes in the scope belong.
 * @field FMD_INFO_PROGRAM_INFO
 * Load module and relocation information for the programs to which the
 * processes in the scope belong.
 * @field FMD_INFO_TRACE
 * RAMTRACE contents.
 *
 * @seealso FmdInfo
 */
enum FmdInfoValue
{
   FMD_INFO_PROCESS = 0X1,
   FMD_INFO_PROCESS_LONG = 0X2,
   FMD_INFO_STACK_USAGE = 0X4,
   FMD_INFO_ENV_VARS = 0X8,
   FMD_INFO_SIGNAL_QUEUE = 0X10,
   FMD_INFO_POOL_INFO = 0X20,
   FMD_INFO_POOL_SIGNALS = 0X40,
   FMD_INFO_POOL_SIGNALS_OWNED = 0X80,
   FMD_INFO_SYS_PARAM = 0X100,
   FMD_INFO_REGISTERS = 0X200,
   FMD_INFO_LOAD_MODULE_INFO = 0X400,
   FMD_INFO_PROGRAM_INFO = 0X800,
   FMD_INFO_TRACE = 0X1000
};

/**
 * @macro FMD_INFO_SOURCE_DEBUG
 * @brief FmdInfoValue suitable for full source code debugging.
 *
 * @seealso FmdInfo
 */
#define FMD_INFO_SOURCE_DEBUG \
   (FMD_INFO_PROCESS | FMD_INFO_REGISTERS \
    | FMD_INFO_LOAD_MODULE_INFO | FMD_INFO_PROGRAM_INFO)

/**
 * @macro FMD_INFO_SYSTEM_DEBUG
 * @brief FmdInfoValue suitable for full system browsing.
 *
 * @seealso FmdInfo
 */
#define FMD_INFO_SYSTEM_DEBUG \
   (FMD_INFO_PROCESS_LONG | FMD_INFO_ENV_VARS \
    | FMD_INFO_SIGNAL_QUEUE | FMD_INFO_POOL_INFO | FMD_INFO_SYS_PARAM)

/**
 * @macro FMD_INFO_POOL_DEBUG
 * @brief FmdInfoValue suitable for full pool analysis.
 *
 * @seealso FmdInfo
 */
#define FMD_INFO_POOL_DEBUG \
   (FMD_INFO_PROCESS | FMD_INFO_STACK_USAGE \
    | FMD_INFO_POOL_INFO | FMD_INFO_POOL_SIGNALS)

/**
 * @macro FMD_INFO_GENERAL_DEBUG
 * @brief FmdInfoValue suitable for general situations when extensive
 * information is needed.
 *
 * @seealso FmdInfo
 */
#define FMD_INFO_GENERAL_DEBUG \
   (FMD_INFO_PROCESS_LONG | FMD_INFO_STACK_USAGE | FMD_INFO_ENV_VARS \
    | FMD_INFO_SIGNAL_QUEUE | FMD_INFO_POOL_INFO | FMD_INFO_POOL_SIGNALS \
    | FMD_INFO_SYS_PARAM | FMD_INFO_REGISTERS | FMD_INFO_LOAD_MODULE_INFO \
    | FMD_INFO_PROGRAM_INFO)

/**
 * @type FmdMemory
 * @brief Type of memory.
 *
 * @seealso FmdMemoryValue
 */
typedef uint32_t FmdMemory;

/**
 * @type FmdMemoryValue
 * @brief Memory type constants used to select type and amount of memory to
 * be saved.
 *
 * @field FMD_MEMORY_POOL
 * Pool contents for all pools processes in the scope can use.
 * @field FMD_MEMORY_STACKS
 * Pool contents used for stacks for all pools processes in the scope can use.
 * @field FMD_MEMORY_STACKS_OWNED
 * Stacks of all processes in the scope.
 * @field FMD_MEMORY_STACKS_USED
 * The used part of the stacks of all processes in the scope.
 * @field FMD_MEMORY_STACKS_TOP
 * The top of the stacks of all processes in the scope.
 * @field FMD_MEMORY_SIGNALS
 * Pool contents used for signals for all pools processes in the scope can use.
 * @field FMD_MEMORY_SIGNALS_OWNED
 * Signals owned by processes in the scope.
 * @field FMD_MEMORY_SIGNALS_QUEUED
 * Signals in the queues of processes in the scope.
 * @field FMD_MEMORY_SIGNALS_SENT
 * Signals with processes in the scope as sender.
 * @field FMD_MEMORY_HEAP
 * Heap contents for all heaps processes in the scope can use.
 * @field FMD_MEMORY_HEAP_BUFFERS
 * Heap buffers for all heaps processes in the scope can use.
 * @field FMD_MEMORY_HEAP_BUFFERS_OWNED
 * Heap buffers owned by processes in the scope.
 * @field FMD_MEMORY_DATA
 * Data sections (.data, .bss, etc.) for the programs to which the processes
 * in the scope belong.
 * @field FMD_MEMORY_TEXT
 * Text sections (.text, etc.) for the programs to which the processes in the
 * scope belong.
 *
 * @seealso FmdMemory
 */
enum FmdMemoryValue
{
   FMD_MEMORY_POOL = 0X1,
   FMD_MEMORY_STACKS = 0X2,
   FMD_MEMORY_STACKS_OWNED = 0X4,
   FMD_MEMORY_STACKS_USED = 0X8,
   FMD_MEMORY_STACKS_TOP = 0X10,
   FMD_MEMORY_SIGNALS = 0X20,
   FMD_MEMORY_SIGNALS_OWNED = 0X40,
   FMD_MEMORY_SIGNALS_QUEUED = 0X80,
   FMD_MEMORY_SIGNALS_SENT = 0X100,
   FMD_MEMORY_HEAP = 0X200,
   FMD_MEMORY_HEAP_BUFFERS = 0X400,
   FMD_MEMORY_HEAP_BUFFERS_OWNED = 0X800,
   FMD_MEMORY_DATA = 0X1000,
   FMD_MEMORY_TEXT = 0X2000
};

/**
 * @macro FMD_MEMORY_SOURCE_DEBUG
 * @brief FmdMemoryValue suitable for full source code debugging.
 *
 * @seealso FmdMemoryValue
 */
#define FMD_MEMORY_SOURCE_DEBUG \
   (FMD_MEMORY_STACKS_USED | FMD_MEMORY_SIGNALS_OWNED \
    | FMD_MEMORY_HEAP_BUFFERS_OWNED)

/**
 * @macro FMD_MEMORY_SYSTEM_DEBUG
 * @brief FmdMemoryValue suitable for full system browsing.
 *
 * @seealso FmdMemoryValue
 */
#define FMD_MEMORY_SYSTEM_DEBUG \
   (0)

/**
 * @macro FMD_MEMORY_POOL_DEBUG
 * @brief FmdMemoryValue suitable for full pool analysis.
 *
 * @seealso FmdMemoryValue
 */
#define FMD_MEMORY_POOL_DEBUG \
   (0)

/**
 * @macro FMD_MEMORY_GENERAL_DEBUG
 * @brief FmdMemoryValue suitable for general situations when extensive
 * information is needed.
 *
 * @seealso FmdMemoryValue
 */
 #define FMD_MEMORY_GENERAL_DEBUG \
   (FMD_MEMORY_STACKS | FMD_MEMORY_SIGNALS | FMD_MEMORY_HEAP_BUFFERS)

/**
 * @type FmdProperty
 *
 * @brief Fmd Property type constants.
 *
 * @field FMD_NUM_LONG_QUEUE_PROCS - Only include entities for processes
 *                                   that have the longest signal queues.
 *                                   The value is the max number of
 *                                   processes. Applicable to
 *                                   fmd_dump_signal_queue_info and
 *                                   fmd_dump_pool_memory functions.
 * @field FMD_MAX_SIGNALS_IN_QUEUE - Maximum number of signals in the queue to
 *                                   be dumped in the scope. Applicable to 
 *                                   fmd_dump_signal_queue_info and
 *                                   fmd_dump_pool_memory functions.
 * @field FMD_MAX_STACK_TOP_SIZE - Maximum stack top size to be dumped in
 *                                 the scope. Applicable to fmd_dump_pool_memory
 *                                 function.
 *
 * @seealso FmdScope
 */
typedef enum FmdProperty
{
   FMD_NUM_LONG_QUEUE_PROCS = 0,
   FMD_MAX_SIGNALS_IN_QUEUE,
   FMD_MAX_STACK_TOP_SIZE
} FmdProperty;

/**
 * @function fmd_start_handler1
 * @brief Initialize the Freeze-mode Dumper. To be called from krn.con.
 *
 * @example
 * START_HANDLER1(fmd_start_handler1)
 */
void fmd_start_handler1(void);

/**
 * @function fmd_set_watchdog
 * @brief Set a watchdog function to be called at regular intervals during
 * time consuming operations in the Freeze-mode Dumper.
 *
 * @field function The function to be called.
 *
 * @return EFMI_OK if successful.
 *
 * @restriction Can only be called from supervisor mode processes.
 *
 * @seealso EfmiStatus
 * @seealso RamdumpStatus
 */
FmdStatus fmd_set_watchdog(void (*function)(void));

/**
 * @function fmd_start_dump
 * @brief Start a new dump.
 *
 * @long Subsequent calls to fmd_dump_* will append information to this new
 * dump. The default behavior is to start a new dump whenever the system boots
 * (and fmd_start_handler1 is called).
 *
 * @restriction Available in supervisor processes, not in interrupt processes.
 *
 * @return EFMI_OK if successful.
 *
 * @seealso EfmiStatus
 * @seealso RamdumpStatus
 */
FmdStatus fmd_start_dump(void);

/**
 * @function fmd_dump_error
 * @brief Store error information in the current dump.
 *
 * @long This should be one of the first functions called when creating a
 * dump after a fatal error.
 *
 * @restriction If this function returns an EMFI error code, subsequent calls
 * to fmd_dump_*_info and fmd_dump_*_memory functions with scopes other than
 * FMD_SCOPE_SYSTEM are invalid and will fail.
 * @break
 * Available in supervisor processes, not in interrupt processes.
 *
 * @field description String describing the error or NULL.
 * @field user_called Error handler parameter.
 * @field ecode Error handler parameter.
 * @field extra Error handler parameter.
 *
 * @return EFMI_OK if successful.
 *
 * @seealso EfmiStatus
 * @seealso RamdumpStatus
 */
FmdStatus fmd_dump_error(const char *description,
                          OSBOOLEAN user_called,
                          OSERRCODE ecode,
                          OSERRCODE extra);

/**
 * @function fmd_dump_memory
 * @brief Store a region of RAM in the current dump.
 *
 * @long The description string is used in host tools to match the saved
 * blocks to plain text formatters of the saved data. This allows application
 * specific data structures to be stored and interpreted on the host without
 * the need to implement parsers for the entire .pmd file format.
 *
 * @restriction Available in supervisor processes, not in interrupt processes.
 *
 * @field description String describing the data or NULL.
 * @field start_address Start of the region to save.
 * @field length Length of the region to save.
 *
 * @return EFMI_OK if successful.
 *
 * @seealso EfmiStatus
 * @seealso RamdumpStatus
 */
FmdStatus fmd_dump_memory(const char *description,
                           OSADDRESS start_address,
                           uint32_t length);

/**
 * @function fmd_dump_string
 * @brief Store a string with plain text information in the current dump.
 *
 * @restriction Available in supervisor processes, not in interrupt processes.
 *
 * @field description String to store.
 *
 * @return EFMI_OK if successful.
 *
 * @seealso EfmiStatus
 * @seealso RamdumpStatus
 */
FmdStatus fmd_dump_string(const char *description);

/**
 * @function fmd_dump_process_info
 * @brief Store process, block, and segment information in the current dump.
 *
 * @long This should be one of the first functions called (after fmd_dump_error)
 * when creating a dump after a fatal error, to ensure that the process control
 * blocks of the executive are intact, so that the scopes and filter mechanisms
 * of subsequent calls to fmd_dump_*_info and fmd_dump_*_memory are functional.
 *
 * @long If none of the relevant bits (FMD_INFO_PROCESS and
 * FMD_INFO_PROCESS_LONG) are set in info, this function does nothing.
 *
 * @restriction Available in supervisor processes, not in interrupt processes.
 *
 * @field scope The set of processes for which to store information.
 * @field info The kind of information to store.
 *
 * @return EFMI_OK if successful.
 *
 * @seealso EfmiStatus
 * @seealso RamdumpStatus
 * @seealso FmdScopeValue
 * @seealso FmdInfoValue
 */
FmdStatus fmd_dump_process_info(FmdScope scope, FmdInfo info);

/**
 * @function fmd_dump_stack_usage_info
 * @brief Store stack usage information in the current dump.
 *
 * @long If none of the relevant bits (FMD_INFO_STACK_USAGE) are set in info,
 * this function does nothing.
 *
 * @restriction Available in supervisor processes, not in interrupt processes.
 *
 * @field scope The set of processes for which to store information.
 * @field info The kind of information to store.
 *
 * @return EFMI_OK if successful.
 *
 * @seealso EfmiStatus
 * @seealso RamdumpStatus
 * @seealso FmdScopeValue
 * @seealso FmdInfoValue
 */
FmdStatus fmd_dump_stack_usage_info(FmdScope scope, FmdInfo info);

/**
 * @function fmd_dump_env_vars_info
 * @brief Store environment variable information in the current dump.
 *
 * @long If none of the relevant bits (FMD_INFO_ENV_VARS) are set in info,
 * this function does nothing.
 *
 * @restriction Available in supervisor processes, not in interrupt processes.
 *
 * @field scope The set of processes and blocks for which to store information.
 * @field info The kind of information to store.
 *
 * @return EFMI_OK if successful.
 *
 * @seealso EfmiStatus
 * @seealso RamdumpStatus
 * @seealso FmdScopeValue
 * @seealso FmdInfoValue
 */
FmdStatus fmd_dump_env_vars_info(FmdScope scope, FmdInfo info);

/**
 * @function fmd_dump_signal_queue_info
 * @brief Store signal queue information in the current dump.
 *
 * @long If none of the relevant bits (FMD_INFO_SIGNAL_QUEUE) are set in info,
 * this function does nothing.
 *
 * @restriction Available in supervisor processes, not in interrupt processes.
 *
 * @field scope The set of processes for which to store information.
 * @field info The kind of information to store.
 *
 * @return EFMI_OK if successful.
 *
 * @seealso EfmiStatus
 * @seealso RamdumpStatus
 * @seealso FmdScopeValue
 * @seealso FmdInfoValue
 */
FmdStatus fmd_dump_signal_queue_info(FmdScope scope, FmdInfo info);

/**
 * @function fmd_dump_pool_info
 * @brief Store pool information in the current dump.
 *
 * @long If none of the relevant bits (FMD_INFO_POOL_INFO) are set in info,
 * this function does nothing.
 *
 * @restriction Available in supervisor processes, not in interrupt processes.
 *
 * @field scope The set of processes for which to store information.
 * @field info The kind of information to store.
 *
 * @return EFMI_OK if successful.
 *
 * @seealso EfmiStatus
 * @seealso RamdumpStatus
 * @seealso FmdScopeValue
 * @seealso FmdInfoValue
 */
FmdStatus fmd_dump_pool_info(FmdScope scope, FmdInfo info);

/**
 * @function fmd_dump_pool_signals_info
 * @brief Store signal information in the current dump.
 *
 * @long If none of the relevant bits (FMD_INFO_POOL_SIGNALS,
 * FMD_INFO_POOL_SIGNALS_OWNED) are set in info, this function does nothing.
 *
 * @restriction Available in supervisor processes, not in interrupt processes.
 *
 * @field scope The set of processes for which to store information.
 * @field info The kind of information to store.
 *
 * @return EFMI_OK if successful.
 *
 * @seealso EfmiStatus
 * @seealso RamdumpStatus
 * @seealso FmdScopeValue
 * @seealso FmdInfoValue
 */
FmdStatus fmd_dump_pool_signals_info(FmdScope scope, FmdInfo info);

/**
 * @private
 * @function fmd_dump_sys_param_info
 * @brief Store system parameter information in the current dump.
 *
 * @long If none of the relevant bits (FMD_INFO_SYS_PARAM) are set in info,
 * this function does nothing.
 *
 * @restriction Available in supervisor processes, not in interrupt processes.
 * @break
 * Due to limitations in the system parameter implementation
 * this function currently does nothing.
 *
 * @field scope Should be FMD_SCOPE_SYSTEM.
 * @field info The kind of information to store.
 *
 * @return EFMI_OK if successful.
 *
 * @seealso EfmiStatus
 * @seealso RamdumpStatus
 * @seealso FmdScopeValue
 * @seealso FmdInfoValue
 */
FmdStatus fmd_dump_sys_param_info(FmdScope scope, FmdInfo info);

/**
 * @function fmd_dump_registers_info
 * @brief Store register information in the current dump.
 *
 * @long If none of the relevant bits (FMD_INFO_REGISTERS) are set in info,
 * this function does nothing.
 *
 * @restriction Available in supervisor processes, not in interrupt processes.
 *
 * @field scope The set of processes for which to store information.
 * @field info The kind of information to store.
 * @field regdump Optional set of registers for current process.
 *
 * @return EFMI_OK if successful.
 *
 * @seealso EfmiStatus
 * @seealso RamdumpStatus
 * @seealso FmdScopeValue
 * @seealso FmdInfoValue
 */
FmdStatus fmd_dump_registers_info(FmdScope scope, FmdInfo info,
                                   struct cpu_regdump *regdump);

/**
 * @function fmd_dump_program_info
 * @brief Store program information in the current dump.
 *
 * @long If none of the relevant bits (FMD_INFO_PROGRAM_INFO) are set in info,
 * this function does nothing.
 *
 * @restriction Available in supervisor processes, not in interrupt processes.
 *
 * @field scope The set of processes for which to store information.
 * @field info The kind of information to store.
 *
 * @return PFMI_OK if successful.
 *
 * @seealso PfmiStatus
 * @seealso RamdumpStatus
 * @seealso FmdScopeValue
 * @seealso FmdInfoValue
 */
FmdStatus fmd_dump_program_info(FmdScope scope, FmdInfo info);

/**
 * @function fmd_dump_trace_info
 * @brief Store event trace information in the current dump.
 *
 * @long If none of the relevant bits (FMD_INFO_TRACE) are set in info,
 * this function does nothing.
 *
 * @restriction Available in supervisor processes, not in interrupt processes.
 *
 * @field scope Should be FMD_SCOPE_SYSTEM.
 * @field info The kind of information to store.
 *
 * @return EFMI_OK if successful.
 *
 * @seealso EfmiStatus
 * @seealso RamdumpStatus
 * @seealso FmdScopeValue
 * @seealso FmdInfoValue
 */
FmdStatus fmd_dump_trace_info(FmdScope scope, FmdInfo info);

/**
 * @function fmd_dump_load_module_info
 * @brief Store load module information in the current dump.
 *
 * @long If none of the relevant bits (FMD_INFO_LOAD_MODULE_INFO) are set in
 * info, this function does nothing.
 *
 * @restriction Available in supervisor processes, not in interrupt processes.
 *
 * @field scope The set of processes for which to store information.
 * @field info The kind of information to store.
 *
 * @return PFMI_OK if successful.
 *
 * @seealso PfmiStatus
 * @seealso RamdumpStatus
 * @seealso FmdScopeValue
 * @seealso FmdInfoValue
 */
FmdStatus fmd_dump_load_module_info(FmdScope scope, FmdInfo info);

/**
 * @function fmd_dump_pool_memory
 * @brief Store pool memory in the current dump.
 *
 * @long If none of the relevant bits (FMD_MEMORY_POOL, FMD_MEMORY_STACKS,
 * FMD_MEMORY_STACKS_OWNED, FMD_MEMORY_STACKS_USED, FMD_MEMORY_STACKS_TOP,
 * FMD_MEMORY_SIGNALS, FMD_MEMORY_SIGNALS_OWNED, FMD_MEMORY_SIGNALS_QUEUED,
 * or FMD_MEMORY_SIGNALS_SENT) are set in memory, this function does nothing.
 *
 * @restriction Available in supervisor processes, not in interrupt processes.
 *
 * @field scope The set of processes for which to store information.
 * @field memory The kind of memory to store.
 * @field regdump Registers for current process or NULL. Used only for
 * FMD_MEMORY_STACKS_USED and FMD_MEMORY_STACKS_TOP.
 *
 * @return EFMI_OK if successful.
 *
 * @seealso EfmiStatus
 * @seealso RamdumpStatus
 * @seealso FmdScopeValue
 * @seealso FmdMemoryValue
 */
FmdStatus fmd_dump_pool_memory(FmdScope scope, FmdMemory memory,
                               struct cpu_regdump *regdump);

/**
 * @function fmd_dump_heap_memory
 * @brief Store heap memory in the current dump.
 *
 * @long If none of the relevant bits (FMD_MEMORY_HEAP, FMD_MEMORY_HEAP_BUFFERS
 * or FMD_MEMORY_HEAP_BUFFERS_OWNED) are set in memory, this function
 * does nothing.
 *
 * @restriction Available in supervisor processes, not in interrupt processes.
 *
 * @field scope The set of processes for which to store information.
 * @field memory The kind of memory to store.
 *
 * @return HFMI_OK if successful.
 *
 * @seealso HfmiStatus
 * @seealso RamdumpStatus
 * @seealso FmdScopeValue
 * @seealso FmdMemoryValue
 */
FmdStatus fmd_dump_heap_memory(FmdScope scope, FmdMemory memory);

/**
 * @function fmd_dump_program_memory
 * @brief Store program memory in the current dump.
 *
 * @long If none of the relevant bits (FMD_MEMORY_DATA, FMD_MEMORY_TEXT)
 * are set in memory, this function does nothing.
 *
 * @restriction Available in supervisor processes, not in interrupt processes.
 *
 * @field scope The set of processes for which to store information.
 * @field memory The kind of memory to store.
 *
 * @return PFMI_OK if successful.
 *
 * @seealso PfmiStatus
 * @seealso RamdumpStatus
 * @seealso FmdScopeValue
 * @seealso FmdMemoryValue
 */
FmdStatus fmd_dump_program_memory(FmdScope scope, FmdMemory memory);

/**
 * @function fmd_set_property
 * @brief Sets FMD related property.
 *
 * @long This function set's Fmd property value, the property value
 *       controls the amount of information to be dumped in the scope.
 *       See FmdProperty for supported properties.
 *
 * @restriction Available in supervisor processes, not in interrupt processes.
 *
 * @field property The FmdProperty type constant for which value to be set
 * @field value The value of the property
 *
 * @return 0 if property value is set.
 * @return 1 if property value is not set.
 *
 * @seealso FmdProperty
 * @seealso FmdScope
 */
uint32_t fmd_set_property(FmdProperty property, uint32_t value);

/**
 * @function fmd_end_of_dump
 * @brief Dumps "end of dump" text block with appropriate status.
 *
 * @long This function dumps "end of dump" text block with appropriate status.
 * This function should be called at the end of the error handler.
 *
 * @restriction Available in supervisor processes, not in interrupt processes.
 *
 * @field status
 * if status = 0, "End of dump - successful" text will be dumped, else
 * "End of dump - incomplete" text will be dumped.
 *
 * @return FmdStatus.
 * return == 0, if "end of dump" text block is dumped successfully
 *              with appropriate status.
 * return != 0, on failure
 *
 * @seealso FmdStatus
 */
FmdStatus fmd_end_of_dump(uint32_t status);

#ifdef __cplusplus
} /* extern "C" */
#endif

#endif
