/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2006 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */
/**
 * @toc File_System:
 *
 * @file ose_spi/fss.sig
 *
 * @brief Signal interface for FSS, file system server.
 *
 *
 */

#ifndef _FSS_SIG
#define _FSS_SIG

#include "ose.h"
#include "osetypes.h"
#include "ose_spi/efstypes.h"

#ifdef __cplusplus
extern "C" {
#endif

/***********************************************************
 *
 * DATA TYPES
 *
 ***********************************************************/

/* Types of file system objects (enumeration). */
/**
 * @type FssTypeEnum
 * @declaration
 * enum FssTypeEnum
 * {
 *   FSS_TYPE_ANY	= 0x0,
 *   FSS_TYPE_FSS	= 0x1,
 *   FSS_TYPE_FORMAT	= 0x2,
 *   FSS_TYPE_VOLUME	= 0x3,
 *   FSS_TYPE_BLOCKDEV	= 0x4,
 *   FSS_TYPE_CHARDEV	= 0x5,
 *   FSS_TYPE_ALIAS	= 0x6,
 *   FSS_TYPE_SERVICE	= 0x7,
 *   FSS_TYPE_LINK      = 0x8,
 *   FSS_TYPE_SOCKET    = 0x9,
 *   FSS_TYPE_DEVICE    = 0xA,
 *   FSS_TYPE_SYSCTL    = 0xB,
 *   FSS_TYPE_DDADEV    = 0xC,
 *   FSS_TYPE_NETDEV    = 0xD
* };
 * @brief Enum of different types in the FSS.
 * @field FSS_TYPE_ANY - Special, any type.
 * @field FSS_TYPE_FSS - File system server.
 * @field FSS_TYPE_FORMAT - Format manager (template).
 * @field FSS_TYPE_VOLUME - File/volume manager (instantiated).
 * @field FSS_TYPE_BLOCKDEV - Block device driver.
 * @field FSS_TYPE_CHARDEV - Character device driver.
 * @field FSS_TYPE_ALIAS - Volume alias.
 * @field FSS_TYPE_SERVICE - Service provider.
 * @field FSS_TYPE_LINK - Link module.
 * @field FSS_TYPE_SOCKET - Protocol stack with socket interface.
 * @field FSS_TYPE_DEVICE - Generic device (using device.sig).
 * @field FSS_TYPE_SYSCTL - Module that handles requests for a sysctl domain.
 * @field FSS_TYPE_DDADEV - DDA Device Module (using device.h).
 * @field FSS_TYPE_NETDEV - Net Device Module (using ndi.sig).
 */
enum FssTypeEnum
{
  FSS_TYPE_ANY		= 0x0,	/* Special value, any type. */
  FSS_TYPE_FSS		= 0x1,	/* File system server. */
  FSS_TYPE_FORMAT	= 0x2,	/* Format manager (template). */
  FSS_TYPE_VOLUME	= 0x3,	/* File/volume manager (instantiated). */
  FSS_TYPE_BLOCKDEV	= 0x4,	/* Block device driver. */
  FSS_TYPE_CHARDEV	= 0x5,	/* Character device driver. */
  FSS_TYPE_ALIAS	= 0x6,	/* Volume alias. */
  FSS_TYPE_SERVICE	= 0x7,	/* Service provider. */
  FSS_TYPE_LINK         = 0x8,  /* Link module. */
  FSS_TYPE_SOCKET       = 0x9,  /* Protocol stack with socket interface. */
  FSS_TYPE_DEVICE       = 0xA,  /* Generic device (using device.sig). */
  FSS_TYPE_SYSCTL       = 0xB,  /* Sysctl support */
  FSS_TYPE_DDADEV       = 0xC,  /* DDA Device Module (using device.h)*/
  FSS_TYPE_NETDEV       = 0xD   /* Net Device Module (using ndi.sig)*/
};

/* File system object visibility (enumeration). */
/**
 * @type FssTypeVisibilityEnum
 * @declaration
 * enum FssTypeVisibilityEnum
 * {
 *   FSS_VIS_GLOBAL	= 0x0,
 *   FSS_VIS_LOCAL	= 0x1,
 *   FSS_VIS_REMOTE	= 0x2
 * };
 * @brief Enum for the visibility of a FSS object.
 * @field FSS_VIS_GLOBAL - Global (located on this processor).
 * @field FSS_VIS_LOCAL - Local (located on this processor).
 * @field FSS_VIS_REMOTE - Global (located on remote processor, which
 *                         must declare it FSS_VIS_GLOBAL).
 */
enum FssTypeVisibilityEnum
{
  FSS_VIS_GLOBAL	= 0x0,	/* Global (located on this processor). */
  FSS_VIS_LOCAL		= 0x1,	/* Local  (located on this processor). */
  FSS_VIS_REMOTE	= 0x2	/* Global (located on remote processor). */
};

/* Special values for time outs. */
/**
 * @type FssTimeoutEnum
 * @declaration
 * enum FssTimeoutEnum
 * {
 *   FSS_TMO_NONE	= 0UL,
 *   FSS_TMO_FOREVER	= -1
 * };
 * @brief Enum for special timeout values in the FSS.
 * @field FSS_TMO_NONE - No timeout, just poll.
 * @field FSS_TMO_FOREVER - No timeout, block forever.
 */
enum FssTimeoutEnum
{
  FSS_TMO_NONE		= 0UL,	/* No time-out, just poll. */
  FSS_TMO_FOREVER	= -1	/* No time-out, block forever. */
};

/***********************************************************
 *
 * STATUS/ERROR CODES
 *
 ***********************************************************/

/**
 * @macro FSS_ERRBASE
 * @declaration #define FSS_ERRBASE 100
 * @brief Start of the FSS status codes.
 * @long The FSS_E* status codes are applicable to the status field 
 *       of the FSS_*_REPLY signals. These codes are all in the range 100-103.
 */
#define FSS_ERRBASE (100)   

/**
 * @macro FSS_EINVALID_TYPE
 *
 * @brief Invalid type.
 * @long This is a status code applicable to the status field of 
 *       the FSS_*_REPLY signals.
 */
#define FSS_EINVALID_TYPE (100)

/**
 * @macro FSS_EUNKNOWN_OBJECT
 *
 * @brief Unknown object.
 * @long This is a status code applicable to the status field of 
 *       the FSS_*_REPLY signals.
 */
#define FSS_EUNKNOWN_OBJECT (101)

/**
 * @macro FSS_EDUPLICATE_OBJECT
 *
 * @brief Duplicate object.
 * @long This is a status code applicable to the status field of 
 *       the FSS_*_REPLY signals.
 */
#define FSS_EDUPLICATE_OBJECT (102)

/**
 * @macro FSS_ERESOLVE_EXT_VALIDATE_OPERATION_DENIED
 *
 * @brief Error code returned in signal FssResolve indicating operation defined
 *          in extension FssResolveExtValidate is not allowed.
 *          Function @link "fss_validate.h" "fss_validate_volume_access" returned
 *          False.
 * @long This is a status code applicable to the status field of 
 *       the FSS_*_REPLY signals.
 */
#define FSS_ERESOLVE_EXT_VALIDATE_OPERATION_DENIED (103)



/***********************************************************
 *
 * SIGNAL INTERFACE
 *
 ***********************************************************/

/* Obtain information about the FSS interface. */

/**
 * @signal FssInterface
 * @brief Get interface information.
 * @long Obtains information about the File System Server signal interface.
 *	 Unless an error reply status is returned, the reply signal includes
 * 	 a list of supported signals, optionally a BIOS trap handle and a
 *	 descriptive string for this type of format manager.
 * @member FssInterfaceRequest
 * @member FssInterfaceReply
 * @seealso sendFssInterfaceRequest
 * @seealso sendFssInterfaceReply
 */

/**
 * @macro FSS_SIGBASE
 * @declaration #define FSS_SIGBASE 30000
 * @brief Start of the FSS signal numbers.
 */
#define FSS_SIGBASE (30000) /* Start of FSS signal numbers. */

/**
 * @macro FSS_INTERFACE_REQUEST
 * @brief Signal number in FssInterfaceRequest.
 * @declaration #define FSS_INTERFACE_REQUEST	30000
 * @seealso FssInterface
 */
#define FSS_INTERFACE_REQUEST		(30000)	/* !-SIGNO(struct FssInterfaceRequest)-! */

/**
 * @macro FSS_INTERFACE_REPLY
 * @brief Signal number in FssInterfaceReply.
 * @declaration #define FSS_INTERFACE_REPLY	30001
 * @seealso FssInterface
 */
#define FSS_INTERFACE_REPLY		(30001)	/* !-SIGNO(struct FssInterfaceReply)-! */

/**
 * @type FssInterfaceRequest
 * @brief Struct used in the signal FSS_INTERFACE_REQUEST.
 * @declaration
 * struct FssInterfaceRequest
 * {
 * 	SIGSELECT sigNo;
 * };
 * @input sigNo FSS_INTERFACE_REQUEST
 */
struct FssInterfaceRequest
{
	SIGSELECT sigNo;
};

/**
 * @type FssInterfaceReply
 * @brief Struct used in the signal FSS_INTERFACE_REPLY.
 * @declaration
 * struct FssInterfaceReply
 * {
 *	SIGSELECT sigNo;
 *	EfsStatus status;
 *	char whatStr[32];
 *	U32 biosHandle;
 *	SIGSELECT sigs[1];
 * };
 * @output sigNo 	FSS_INTERFACE_REPLY
 * @output status	Status of the operation.
 * @output whatStr	What string of product.
 * @output biosHandle	BIOS handle for trap interface.
 * @output sigs		Supported signals. Variable size.
 */
struct FssInterfaceReply
{
	SIGSELECT sigNo;
	EfsStatus status;		/* O: Status of the operation. */
	char whatStr[32];		/* O: What string of product. */
	U32 biosHandle;			/* O: BIOS handle for trap interface. */
	SIGSELECT sigs[1];		/* O: Supported signals. Variable size. */
};

/**
 * @macro sendFssInterfaceRequest
 * @declaration
 * #define sendFssInterfaceRequest(fssPidMA)
 * @brief Send signal FssInterface request.
 * @input fssPidMA PID of the FSS to send the interface request to.
 * @seealso FssInterface
 * @seealso FssInterfaceReply
 */
#define sendFssInterfaceRequest(fssPidMA) \
do \
{ \
	union SIGNAL *sigMA = alloc(sizeof(struct FssInterfaceRequest), FSS_INTERFACE_REQUEST); \
	send(&sigMA, fssPidMA); \
} \
while (0)

/**
 * @macro sendFssInterfaceReply
 * @declaration
 * #define sendFssInterfaceReply(statusMA, whatMA, handleMA, sigLoMA, sigHiMA, fmPidMA)
 * @brief Send signal FssInterface reply.
 * @input statusMA	Status of the operation.
 * @input whatMA	What string of product.
 * @input handleMA	BIOS handle for a trap interface.
 * @input sigLoMA	Lowest supported signal.
 * @input sigHiMA	Highest supported signal.
 * @input fmPidMA	PID of process to send reply to.
 * @seealso FssInterface
 * @seealso FssInterfaceRequest
 */
#define sendFssInterfaceReply(statusMA, whatMA, handleMA, sigLoMA, sigHiMA, fmPidMA) \
do \
{ \
	/* The union is a trick to avoid breaking the strict-aliasing rules*/\
	union U \
	{ \
		union SIGNAL *s;\
		struct FssInterfaceReply *x;\
	} u;\
	unsigned i; \
	struct FssInterfaceReply *sigMA;\
	U32 numSigs = ((sigHiMA) - (sigLoMA)) / 2 + 1; \
	u.s = alloc(sizeof(struct FssInterfaceReply) + numSigs * sizeof(SIGSELECT), FSS_INTERFACE_REPLY); \
	sigMA = u.x; \
	sigMA->status = (statusMA); \
	strncpy(sigMA->whatStr, (whatMA), sizeof sigMA->whatStr); \
	sigMA->biosHandle = (handleMA); \
	sigMA->sigs[0] = numSigs; \
	for (i = 0; i < numSigs; i++) \
	{ \
            sigMA->sigs[i + 1] = (sigLoMA) + 2 * i; \
	} \
	send(&u.s, (fmPidMA)); \
} \
while (0)

/* Connect to a remote file system server (FSS). */

/**
 * @signal FssConnectFss
 * @brief   Connect to a remote file system server.
 * @long    The local FSS will connect to a remote FSS, and publish all
 *          objects with global visibility to the remote FSS. All global
 *          objects in the locl FSS will become visible as remote objects
 *          in the remote FSS.
 * @member FssConnectFss
 * @seealso sendFssConnectFssRequest
 */

/**
 * @macro FSS_CONNECT_FSS_REQUEST
 * @brief Signal number in FssConnectFss request.
 * @declaration
 * #define FSS_CONNECT_FSS_REQUEST	30002
 * @seealso FssConnectFss
 * @seealso sendFssConnectFssRequest
 */
#define FSS_CONNECT_FSS_REQUEST  (30002) /* !-SIGNO(struct FssConnectFss)-! */

/**
 * @macro FSS_CONNECT_FSS_REPLY
 * @brief Signal number in FssConnectFss reply.
 * @declaration
 * #define FSS_CONNECT_FSS_REPLY 30003
 * @seealso FssConnectFss
 * @seealso sendFssConnectFssRequest
 */
#define FSS_CONNECT_FSS_REPLY    (30003) /* !-SIGNO(struct FssConnectFss)-! */

/**
 * @type FssConnectFss
 * @brief Struct used in the signal FSS_CONNECT_FSS_REQUEST and FSS_CONNECT_FSS_REQUEST.
 * @declaration
 * struct FssConnectFss
 * {
 * 	SIGSELECT sigNo;
 * 	EfsStatus status;
 * 	char huntPath[1];
 * };
 * @input sigNo		FSS_CONNECT_FSS_REPLY or FSS_CONNECT_FSS_REQUEST
 * @output status	Status of the operation.
 * @input huntPath	Full huntpath to the remote FSS. e.g. "a/ose_fss"
 *                      for connecting to the FSS on processor node "a".
 *                      Variable size field.
 */
struct FssConnectFss
{
	SIGSELECT sigNo;
	EfsStatus status;
	char huntPath[1];
};

/* sender(&request) is file system object. */

/**
 * @macro sendFssConnectFssRequest
 * @declaration
 * #define sendFssConnectFssRequest(huntPathMA, fssPidMA)
 * @brief Send signal   FssConnectFssRequest to remote FSS.
 * @input huntPathMA	Huntpath of the remote FSS.
 * @input fssPidMA	    PID of the remote FSS.
 */
#define sendFssConnectFssRequest(huntPathMA, fssPidMA) \
do \
{ \
	/* The union is a trick to avoid breaking the strict-aliasing rules*/\
	union U \
	{ \
		union SIGNAL *s;\
		struct FssConnectFss *x;\
	} u;\
	struct FssConnectFss *sigMA; \
	const char *_huntPathMA = (huntPathMA); \
	u.s = alloc(sizeof(struct FssConnectFss) + strlen(_huntPathMA), FSS_CONNECT_FSS_REQUEST); \
	sigMA = u.x; \
	strcpy(sigMA->huntPath, _huntPathMA); \
	send(&u.s, (fssPidMA)); \
} \
while (0)

/* Register a file system object. */
/**
 * @signal FssRegister
 * @brief Register file system object in the FSS.
 * @member FssRegister
 * @seealso sendFssRegisterRequest
 * @seealso FssTypeVisibilityEnum
 * @seealso FssTypeEnum
 */

/**
 * @macro FSS_REGISTER_REQUEST
 * @brief Signal number in FssRegister request.
 * @declaration
 * #define FSS_REGISTER_REQUEST		30004
 * @seealso FssRegister
 * @seealso sendFssRegisterRequest
 */
#define FSS_REGISTER_REQUEST		(30004)	/* !-SIGNO(struct FssRegister)-! */

/**
 * @macro FSS_REGISTER_REPLY
 * @brief Signal number in FssRegister reply.
 * @declaration
 * #define FSS_REGISTER_REPLY		30005
 * @seealso FssRegister
 * @seealso sendFssRegisterRequest
 */
#define FSS_REGISTER_REPLY		(30005)	/* !-SIGNO(struct FssRegister)-! */

/**
 * @type FssRegister
 * @brief Struct used in the signal FSS_REGISTER_REQUEST,
 *        FSS_REGISTER_SILENT_REQUEST and FSS_REGISTER_REPLY
 * @declaration
 * struct FssRegister
 * {
 *	SIGSELECT sigNo;
 *	EfsStatus status;
 *	EfsEnum local;
 *	EfsEnum type;
 *	char name[1];
 * };
 * @output status	Status of the operation
 * @input sigNo	FSS_REGISTER_REQUEST, FSS_REGISTER_SILENT_REQUEST or FSS_REGISTER_REPLY
 * @input local		Visibility of the object (FSS_VIS_XX).
 * @input type		Type of the object (FSS_TYPE_XX).
 * @input name		Name of the object. Variable size.
 * @seealso FssTypeVisibilityEnum
 * @seealso FssTypeEnum
 * @seealso FSS_ERRBASE
 */
struct FssRegister
{
	SIGSELECT sigNo;
	EfsStatus status;	/* O: Status of the operation. */
	EfsEnum local;		/* I: Visibility of the object (FSS_VIS_XXX). */
	EfsEnum type;		/* I: Type of the object (FSS_TYPE_XX). */
	char name[1];		/* I: Name of the object. Variable size. */
};

/* sender(&request) is file system object. */

/**
 * @macro sendFssRegisterRequest
 * @declaration
 * #define sendFssRegisterRequest(nameMA, typeMA, localMA, fssPidMA)
 * @brief Send a request for a registration of a file system object to the FSS.
 * @input nameMA	Name of the object. Variable size.
 * @input typeMA	Type of the object.
 * @input localMA	Visibility of the object.
 * @input fssPidMA	PID of the file system server.
 * @seealso FssRegister
 * @seealso FssTypeVisibilityEnum
 * @seealso FssTypeEnum
 */
#define sendFssRegisterRequest(nameMA, typeMA, localMA, fssPidMA) \
do \
{ \
	/* The union is a trick to avoid breaking the strict-aliasing rules*/\
	union U \
	{ \
		union SIGNAL *s;\
		struct FssRegister *x;\
	} u;\
	const char *_nameMA = (nameMA); \
	struct FssRegister *sigMA; \
	u.s = alloc(sizeof(struct FssRegister) + strlen(_nameMA), FSS_REGISTER_REQUEST); \
	sigMA = u.x; \
	sigMA->local = (localMA); \
	sigMA->type = (typeMA); \
	strcpy(sigMA->name, _nameMA); \
	send(&u.s, (fssPidMA)); \
} \
while (0)

/* Resolve a name returning file system object. */

/**
 * @signal FssResolve
 * @brief Resolve a name and return a file system object.
 * @member FssResolve
 * @seealso sendFssResolveRequest
 * @seealso FssTypeEnum
 * @seealso FssTimeoutEnum
 */

/**
 * @macro FSS_RESOLVE_REQUEST
 * @brief Signal number in FssResolve request.
 * @declaration
 * #define FSS_RESOLVE_REQUEST		30006
 * @seealso FssResolve
 * @seealso sendFssResolveRequest
 */
#define FSS_RESOLVE_REQUEST		(30006)	/* !-SIGNO(struct FssResolve)-! */

/**
 * @macro FSS_RESOLVE_REPLY
 * @brief Signal number in FssResolve reply.
 * @declaration
 * #define FSS_RESOLVE_REPLY		30007
 * @seealso FssResolve
 * @seealso sendFssResolveRequest
 */
#define FSS_RESOLVE_REPLY		(30007)	/* !-SIGNO(struct FssResolve)-! */

/**
 * @type FssResolve
 * @brief Struct used in the signal FSS_RESOLVE_REQUEST and FSS_RESOLVE_REPLY.
 * @declaration
 * struct FssResolve
 * {
 *	SIGSELECT sigNo;
 *	EfsStatus status;
 *	EfsEnum type;
 *	U32 timeOut;
 *	char name[1];
 * };
 * @input sigNo		FSS_RESOLVE_REQUEST or FSS_RESOLVE_REPLY
 * @output status	Status of the operation.
 * @input type		Requested type of object (FSS_TYPE_XX).
 * @input timeOut	Timeout in ms. (FSS_TMO_XX).
 * @input name		Name to resolve. Variable size.
 * @seealso FssTypeEnum
 * @seealso FssTimeoutEnum
 */
struct FssResolve
{
	SIGSELECT sigNo;
	EfsStatus status;	/* O: Status of the operation. */
	EfsEnum type;		/* I: Requested type of object (FSS_TYPE_XX). */
	U32 timeOut;		/* I: Time-out in milliseconds (FSS_TMO_XX). */
	char name[1];		/* I: Name to resolve. Variable size. */
};

/**
 * @macro sendFssResolveRequest
 * @declaration
 * #define sendFssResolveRequest(nameMA, typeMA, tmoMA, fssPidMA)
 * @brief Send signal FssResolve request.
 * @input nameMA	Name to resolve.
 * @input typeMA	Type of object to resolve.
 * @input tmoMA		Timeout in ms.
 * @input fssPidMA	PID of the file system server.
 * @seealso FssResolve
 */
#define sendFssResolveRequest(nameMA, typeMA, tmoMA, fssPidMA) \
do \
{ \
	/* The union is a trick to avoid breaking the strict-aliasing rules*/\
	union U \
	{ \
		union SIGNAL *s;\
		struct FssResolve *x;\
	} u;\
	struct FssResolve *sigMA; \
	const char *_nameMA = (nameMA); \
	u.s = alloc(sizeof(struct FssResolve) + strlen(_nameMA), FSS_RESOLVE_REQUEST); \
	sigMA = u.x; \
	sigMA->type = (typeMA); \
	sigMA->timeOut = (U32)(tmoMA); \
	strcpy(sigMA->name, _nameMA); \
	send(&u.s, (fssPidMA)); \
} \
while (0)

/* reply->status == EFS_SUCCESS => sender(&reply) is file system object. */

/**
 * @type FssResolveExtHead
 *
 * @brief FssResolve extension header.
 *
 * @long The extension header contains a magic number that indicates that valid
 *       extensions exists in the FssResolve signal and a position of the first
 *       extension.
 * @long The extension header is aligned to U32 and located at the end of the
 *       FssResolve signal. The extension header is present if requested signal
 *       size (sigsize()) is an even multiple of U32 and the last U32 (magic) in
 *       the signal is equal FSS_RESOLVE_EXT_HEAD_MAGIC. The magic number does
 *       not contain string termination character which garanties that the magic
 *       number is not contained at the end of the name member in struct
 *       FssResolve.
 * @long Each extension shall start with the following members:
 * @long - U32 next_ext_position; Position of the next extension in the signal
 *                                allowing several (future) extensions.
 * @long - U32 ext_id;            Extension ID identifying the extension struct.
 * @long - U32 handled;           Flag that indicates if the information is
 *                                processed by FSS.
 * @long All extensions are aligned to U32 in the signal.
 *
 * @input first_ext_position Position of first extension, expressed as bytes
 *                           from beginning of signal.
 * @input magic              Indicates valid extensions exists.
 *
 * @seealso FssResolve
 */
#define FSS_RESOLVE_EXT_HEAD_MAGIC ('E' << 24 | 'X' << 16 | 'T' << 8 | ' ')
struct FssResolveExtHead
{
   U32 first_ext_position;
   U32 magic;
};

/**
 * @type FssResolveExtValidate
 *
 * @brief FssResolve extension validate.
 *
 * @long Extension validate is information added to signal FssResolve that will
 *       be passed to the function @link "fss_validate.h"
 *       "fss_validate_volume_access" for validation of volume access.
 * @long The position of the extension in the signal is given via the extension
 *       header @link "" "FssResolveExtHead" and possibly via other extensions.
 *       The members next_ext_position, ext_id and handled are included in all
 *       extensions. Other members are unique for this extension.
 * @long The extension is aligned to U32 in the signal.
 *
 * @input next_ext_position  Position of next extension, expressed as bytes from
 *                           beginning of signal.
 *                           0 indicates this is the last extension.
 * @input ext_id             Extension id = FSS_RESOLVE_EXT_VALIDATE.
 * @input handled            Set to 0 (not handled).
 * @input operation          Type of operation on the object,
 *                           see @link "fss_validate.h"
 *                           "fss_validate_volume_access".
 * @input obj_path_offset    Offset in string pool for volume name,
 *                           see @link "fss_validate.h"
 *                           "fss_validate_volume_access".
 * @input lnhname_offset     Offset in string pool for link handler name string,
 *                           see @link "fss_validate.h"
 *                           "fss_validate_volume_access".
 * @input opt_string_offset  Offset in string pool for optional string,
 *                           see @link "fss_validate.h"
 *                           "fss_validate_volume_access".
 * @input open_mode          Permission attributes for the opened object,
 *                           see @link "fss_validate.h"
 *                           "fss_validate_volume_access".
 * @input access_mode        Permission attributes for the object,
 *                           see @link "fss_validate.h"
 *                           "fss_validate_volume_access".
 * @input username_offset    Offset in string pool for user name string,
 *                           see @link "fss_validate.h"
 *                           "fss_validate_volume_access".
 * @input termname_offset    Offset in string pool for terminal name string,
 *                           see @link "fss_validate.h"
 *                           "fss_validate_volume_access".
 * @input s                  String pool where the strings are stored. Includes
 *                           5 string terminations.
 *                           Note: The field is of variable size.
 *                                 The actual size must be taken into account
 *                                 when allocating storage.
 *                                 No fields may be added after this field.
 *
 * @output handled           0=Indicates not handled (information not processed)
 *                           by the FSS.
 *                           1=Indicates handled (information processed) by the
 *                           FSS.
 *
 * @seealso FssResolveExtHead
 * @seealso FssResolve
 */
#define FSS_RESOLVE_EXT_VALIDATE      1
struct FssResolveExtValidate
{
   U32 next_ext_position;
   U32 ext_id;
   U32 handled;
   U32 operation;
   U32 obj_path_offset;
   U32 lnhname_offset;
   U32 opt_string_offset;
   U32 open_mode;
   U32 access_mode;
   U32 username_offset;
   U32 termname_offset;
   /* Note: The following field is of variable size.
    *       The actual size must be taken into account when allocating storage.
    *       No fields may be added after this field.
    */
   char s[5];
};

/* List all known file system objects of specified type. */

/**
 * @signal FssList
 * @brief List all known file system objects of a specified type.
 * @member FssListRequest
 * @member FssListReply
 * @seealso sendFssListRequest
 * @seealso FssTypeVisibilityEnum
 * @seealso FssTypeEnum
 */

/**
 * @macro FSS_LIST_REQUEST
 * @brief Signal number in FssListRequest.
 * @declaration
 * #define FSS_LIST_REQUEST		30008
 * @seealso FssList
 * @seealso sendFssListRequest
 */
#define FSS_LIST_REQUEST		(30008)	/* !-SIGNO(struct FssListRequest)-! */

/**
 * @macro FSS_LIST_REPLY
 * @brief Signal number in FssListReply.
 * @declaration
 * #define FSS_LIST_REPLY		30009
 * @seealso FssList
 * @seealso sendFssListRequest
 * @seealso sendFssListReply
 */
#define FSS_LIST_REPLY			(30009)	/* !-SIGNO(struct FssListReply)-! */

/**
 * @type FssListRequest
 * @declaration
 * struct FssListRequest
 * {
 * 	SIGSELECT sigNo;
 * 	EfsEnum type;
 * };
 * @input sigNo FSS_LIST_REQUEST
 * @input type	Type of the objects to list.
 */
struct FssListRequest
{
	SIGSELECT sigNo;
	EfsEnum type;		/* I: Type of the objects to list. */
};

/**
 * @type FssListReply
 * @declaration
 * struct FssListReply
 * {
 *	SIGSELECT sigNo;
 *	EfsStatus status;
 * 	EfsEnum local;
 * 	EfsEnum type;
 * 	char name[1];
 * };
 * @output sigNo	FSS_LIST_REPLY
 * @output status	Status of the operation.
 * @output local	Visibility of the object (FSS_VIS_XX).
 * @output type		Type of object. (FSS_TYPE_XX).
 * @output name		Name of the object. Variable size.
 * @seealso FssTypeVisibilityEnum
 * @seealso FssTypeEnum
 */
struct FssListReply
{
	SIGSELECT sigNo;
	EfsStatus status;	/* O: Status of the operation. */
	EfsEnum local;		/* O: Visibility of the object (FSS_VIS_XX). */
	EfsEnum type;		/* O: Type of the object (FSS_TYPE_XX). */
	char name[1];		/* O: Name of the object. Variable size. */
};

/**
 * @macro sendFssListRequest
 * @declaration
 * #define sendFssListRequest(typeMA, fssPidMA)
 * @brief Send signal FssList request.
 * @input typeMA	Type of the object to list.
 * @input fssPidMA	PID of the FSS.
 */
#define sendFssListRequest(typeMA, fssPidMA) \
do \
{ \
	/* The union is a trick to avoid breaking the strict-aliasing rules*/\
	union U \
	{ \
		union SIGNAL *s;\
		struct FssListRequest *x;\
	} u;\
	struct FssListRequest *sigMA; \
	u.s = alloc(sizeof(struct FssListRequest), FSS_LIST_REQUEST); \
	sigMA = u.x; \
	sigMA->type = (typeMA); \
	send(&u.s, (fssPidMA)); \
} \
while (0)

/**
 * @macro sendFssListReply
 * @declaration
 * #define sendFssListReply(statusMA, nameMA, typeMA, localMA, pidMA, clientPidMA)
 * @brief Send signal FssList reply.
 * @input statusMA	Status of the opration.
 * @input nameMA	  Name of the object. Variable size.
 * @input typeMA	  Type of the object to list.
 * @input localMA	  Visibility of the object.
 * @input pidMA		  PID of the FSS.
 * @input clientPidMA PID of the process to send signal to.
 */
#define sendFssListReply(statusMA, nameMA, typeMA, localMA, pidMA, clientPidMA) \
do \
{ \
	/* The union is a trick to avoid breaking the strict-aliasing rules*/\
	union U \
	{ \
		union SIGNAL *s;\
		struct FssListReply *x;\
	} u;\
	struct FssListReply *sigMA; \
	const char *_nameMA = (nameMA); \
	u.s = alloc(sizeof(struct FssListReply) + strlen(_nameMA), FSS_LIST_REPLY); \
	sigMA = u.x; \
	sigMA->status = (statusMA); \
	sigMA->local = (localMA); \
	sigMA->type = (typeMA); \
	strcpy(sigMA->name, _nameMA); \
	send_w_s(&u.s, (pidMA), (clientPidMA)); \
} \
while (0)
/* reply->status == EFS_SUCCESS => sender(&reply) is file system object. */

/* Extended resolve a name returning file system object. */

/**
 * @signal FssExtResolve
 * @brief Extended resolve a name and return a file system object.
 * @member FssExtResolve
 * @seealso sendFssExtResolveRequest
 * @seealso FssTypeEnum
 * @seealso FssTimeoutEnum
 */

/**
 * @macro FSS_EXT_RESOLVE_REQUEST
 * @brief Signal number in FssExtResolve request.
 * @declaration
 * #define FSS_EXT_RESOLVE_REQUEST	30010
 * @seealso FssExtResolve
 * @seealso sendFssExtResolveRequest
 */
#define FSS_EXT_RESOLVE_REQUEST		(30010)	/* !-SIGNO(struct FssExtResolve)-! */

/**
 * @macro FSS_EXT_RESOLVE_REPLY
 * @brief Signal number in FssExtResolve reply.
 * @declaration
 * #define FSS_EXT_RESOLVE_REPLY	30011
 * @seealso FssExtResolve
 * @seealso sendFssExtResolveRequest
 */
#define FSS_EXT_RESOLVE_REPLY		(30011)	/* !-SIGNO(struct FssExtResolve)-! */

/**
 * @type FssExtResolve
 * @declaration
 * struct FssExtResolve
 * {
 *	SIGSELECT sigNo;
 *	EfsStatus status;
 *	EfsEnum type;
 *	U32 timeOut;
 *      U32 name;
 *      U32 username;
 *      U32 terminal;
 *	char strings[3];
 * };
 * @input sigNo		FSS_EXT_RESOLVE_REQUEST or FSS_EXT_RESOLVE_REPLY.
 * @output status	Status of the operation.
 * @input type		Requested type of object (FSS_TYPE_XX).
 * @input timeOut	Timeout in ms. (FSS_TMO_XX).
 * @input name		Name to resolve. Offset into strings.
 * @input username	Username of client. Offset into strings.
 * @input terminal	Terminal of client. Offset into strings.
 * @input strings       String pool. Variable size.
 * @seealso FssTypeEnum
 * @seealso FssTimeoutEnum
 */
struct FssExtResolve
{
	SIGSELECT sigNo;
	EfsStatus status;	/* O: Status of the operation. */
	EfsEnum type;		/* I: Requested type of object (FSS_TYPE_XX). */
	U32 timeOut;		/* I: Time-out in milliseconds (FSS_TMO_XX). */
        U32 name;               /* I: Name to resolve. */
        U32 username;           /* I: Username of client. */
        U32 terminal;           /* I: Terminal of client. */
	char strings[3];	/* I: String pool. Variable size. */
};

/**
 * @macro sendFssExtResolveRequest
 * @declaration
 * #define sendFssExtResolveRequest(nameMA, typeMA, tmoMA, fssPidMA)
 * @brief Send signal FssExtResolve request.
 * @input nameMA	Name to resolve.
 * @input userMA	Username of client.
 * @input termMA	Terminal of client.
 * @input typeMA	Type of object to resolve.
 * @input tmoMA		Timeout in ms.
 * @input fssPidMA	PID of the file system server.
 * @seealso FssExtResolve
 */
#define sendFssExtResolveRequest(nameMA, userMA, termMA, typeMA, tmoMA, fssPidMA) \
do \
{ \
	/* The union is a trick to avoid breaking the strict-aliasing rules*/\
	union U \
	{ \
		union SIGNAL *s;\
		struct FssExtResolve *x;\
	} u;\
	struct FssExtResolve *sigMA; \
	const char *_nameMA = (nameMA); \
	const char *_userMA = (userMA); \
	const char *_termMA = (termMA); \
	u.s = alloc(sizeof(struct FssExtResolve) + strlen(_nameMA) + strlen(_userMA) + strlen(_termMA), FSS_EXT_RESOLVE_REQUEST); \
	sigMA = u.x; \
	sigMA->type = (typeMA); \
	sigMA->timeOut = (tmoMA); \
	sigMA->name = 0; \
	sigMA->username = strlen(_nameMA) + 1; \
	sigMA->terminal = sigMA->username + 1 + strlen(_userMA); \
	strcpy(&sigMA->strings[sigMA->name], _nameMA); \
	strcpy(&sigMA->strings[sigMA->username], _userMA); \
	strcpy(&sigMA->strings[sigMA->terminal], _termMA); \
	send(&u.s, (fssPidMA)); \
} \
while (0)

/**
 * @macro FSS_REGISTER_SILENT_REQUEST
 * @brief Signal number in FssRegister request. Sender will not
 *        receive the status of the operation.
 * @declaration
 * #define FSS_REGISTER_SILENT_REQUEST	30012
 * @seealso FssRegister
 * @seealso sendFssRegisterSilentRequest
 */
#define FSS_REGISTER_SILENT_REQUEST		(30012)	/* !-SIGNO(struct FssRegister)-! */

/**
 * @macro sendFssRegisterSilentRequest
 * @declaration
 * #define sendFssRegisterSilentRequest(nameMA, typeMA, localMA, fssPidMA)
 * @brief Send a request for a registration of a file system object to the FSS.
 * @input nameMA	Name of the object. Variable size.
 * @input typeMA	Type of the object.
 * @input localMA	Visibility of the object.
 * @input fssPidMA	PID of the file system server.
 * @seealso FssRegister
 * @seealso FssTypeVisibilityEnum
 * @seealso FssTypeEnum
 */
#define sendFssRegisterSilentRequest(nameMA, typeMA, localMA, fssPidMA) \
do \
{ \
	/* The union is a trick to avoid breaking the strict-aliasing rules*/\
	union U \
	{ \
		union SIGNAL *s;\
		struct FssRegister *x;\
	} u;\
	const char *_nameMA = (nameMA); \
	struct FssRegister *sigMA; \
	u.s = alloc(sizeof(struct FssRegister) + strlen(_nameMA), FSS_REGISTER_SILENT_REQUEST); \
	sigMA = u.x; \
	sigMA->local = (localMA); \
	sigMA->type = (typeMA); \
	strcpy(sigMA->name, _nameMA); \
	send(&u.s, (fssPidMA)); \
} \
while (0)


#ifdef __cplusplus
} /* end of extern "C" */
#endif

#endif
