/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2006 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */
/**
 * @toc File_System:
 *
 * @file ose_spi/fss_validate.h
 *
 * @brief Declaration of API for validation of volume access.
 *
 *
 */

#ifndef _FSS_VALIDATE_H
#define _FSS_VALIDATE_H

#include "ose.h"
#include "osetypes.h"

#ifdef __cplusplus
extern "C" {
#endif

/*
 *******************************************************************************
 * Declarations
 *******************************************************************************
 */

/** 
 * @type FssValidateOperationEnum
 *
 * @brief Type of operation on volume or object in volume to be validated.
 *
 * @field FSS_VALIDATE_OPERATION_ACCESS      Set if CRT function access() is
 *                                           called.
 * @field FSS_VALIDATE_OPERATION_CHDIR       Set if CRT function chdir() is
 *                                           called.
 * @field FSS_VALIDATE_OPERATION_CHMOD       Set if CRT function chmod() is
 *                                           called.
 * @field FSS_VALIDATE_OPERATION_FORMAT      Set if CRT function efs_format() is
 *                                           called.
 * @field FSS_VALIDATE_OPERATION_INTERNAL    Internal operation. Set if extra
 *                                           FssResolve needed.
 * @field FSS_VALIDATE_OPERATION_MKDIR       Set if CRT function mkdir() is
 *                                           called.
 * @field FSS_VALIDATE_OPERATION_MOUNT       Set if CRT function efs_mount() is
 *                                           called.
 * @field FSS_VALIDATE_OPERATION_OPEN        Set if CRT function creat(),
 *                                           fopen() or open() is called.
 * @field FSS_VALIDATE_OPERATION_REALPATH    Set if CRT function realpath() is
 *                                           called.
 * @field FSS_VALIDATE_OPERATION_RELABEL     Set if CRT function efs_relabel()
 *                                           is called.
 * @field FSS_VALIDATE_OPERATION_REMOVE      Set if CRT function efs_remove(),
 *                                           remove(), rmdir() or unlink() is
 *                                           called.
 * @field FSS_VALIDATE_OPERATION_RENAME      Set if CRT function rename() is
 *                                           called.
 * @field FSS_VALIDATE_OPERATION_STAT        Set if CRT function stat() is
 *                                           called.
 * @field FSS_VALIDATE_OPERATION_SYNC        Set if CRT function efs_sync() is
 *                                           called.
 * @field FSS_VALIDATE_OPERATION_TRUNCATE    Set if CRT function truncate() is
 *                                           called.
 * @field FSS_VALIDATE_OPERATION_UNKNOWN     Set if a volume look-up is
 *                                           performed from a non-compliant
 *                                           client, e.g. an older CRT.
 * @field FSS_VALIDATE_OPERATION_UNMOUNT     Set if CRT function efs_unmount()
 *                                           is called.
 * @field FSS_VALIDATE_OPERATION_UTIMES      Set if CRT function utimes() is
 *                                           called.
 * @field FSS_VALIDATE_OPERATION_VALIDATE_VOLUME Set if CRT function
 *                                           efs_validate_volume() is called.
 *
 * @seealso fss_validate_volume_access
 */
enum FssValidateOperationEnum
{
   FSS_VALIDATE_OPERATION_ACCESS      = 0,
   FSS_VALIDATE_OPERATION_CHDIR       = 1,
   FSS_VALIDATE_OPERATION_CHMOD       = 2,
   FSS_VALIDATE_OPERATION_FORMAT      = 3,
   FSS_VALIDATE_OPERATION_INTERNAL    = 4,
   FSS_VALIDATE_OPERATION_MKDIR       = 5,
   FSS_VALIDATE_OPERATION_MOUNT       = 6,
   FSS_VALIDATE_OPERATION_OPEN        = 7,
   FSS_VALIDATE_OPERATION_REALPATH    = 8,
   FSS_VALIDATE_OPERATION_RELABEL     = 9,
   FSS_VALIDATE_OPERATION_REMOVE      = 10,
   FSS_VALIDATE_OPERATION_RENAME      = 11,
   FSS_VALIDATE_OPERATION_STAT        = 12,
   FSS_VALIDATE_OPERATION_SYNC        = 13,
   FSS_VALIDATE_OPERATION_TRUNCATE    = 14,
   FSS_VALIDATE_OPERATION_UNKNOWN     = 15,
   FSS_VALIDATE_OPERATION_UNMOUNT     = 16,
   FSS_VALIDATE_OPERATION_UTIMES      = 17,
   FSS_VALIDATE_OPERATION_VALIDATE_VOLUME = 18
};


/*
********************************************************************************
 * FUNCTION PROTOTYPES
 *******************************************************************************
 */

/**
 * @function fss_enable_validate_volume_access
 *
 * @brief Enable or disable validate volume access.
 *
 * @long The function is called from the OSE FSS and should inform FSS if the
 *       function fss_validate_volume_access() shall be called (enabled) or not
 *       (disabled). The function will be called once before the first possible
 *       call to the function fss_validate_volume_access().
 *
 * @long It is not possible to use stdio in the function for additional input
 *       and output to the user.
 *
 * @long This is a function that should be implemented by the developer if it
 *       is wanted. If the function is not implemented a default function in 
 *       the OSE core library is linked into the system. The default function
 *       always returns disable (False).
 *
 * @return The function shall return True or False, if True the function
 *         fss_validate_volume_access() will be called (enabled), if False the
 *         function fss_validate_volume_access() will not be called (disabled).
 *
 * @seealso fss_validate_volume_access
 */
Boolean
fss_enable_validate_volume_access(void);

/**
 * @function fss_validate_volume_access
 *
 * @brief Validate volume access.
 *
 * @long The function is called from the OSE FSS if enabled by the function 
 *       fss_enable_validate_volume_access() and when an OSE CRT function is
 *       called with volume name(s) or volume object (file or directory) name(s)
 *       as argument(s) and sends signal FssResolve to FSS. The function should
 *       perform access control and if the function returns False the operation
 *       (CRT function) will be denied.
 *
 * @long The function may also be used to log information about operations on
 *       the volume and objects in the volume.
 *
 * @long It is not possible to use stdio in the function for additional input
 *       and output to the user.
 *
 * @long This is a function that should be implemented by the developer if it
 *       is wanted. If the function is not implemented a default function in 
 *       the OSE core library is linked into the system. The default function
 *       always returns valid (True).
 *
 * @input operation   Type of operation on the object, see @link "" 
 *                    "FssValidateOperationEnum".
 *
 * @input obj_path    Pointer to a string containing volume name or full path to
 *                    the object in the volume.
 *                    The full path starts with volume name, which starts with 
 *                    '/', followed by path to the object in the volume, e.g. 
 *                    "/ram/mydir/myfile.txt".
 *                    Linkhandler name is not included.
 *                    For operations FSS_VALIDATE_OPERATION_FORMAT, 
 *                    FSS_VALIDATE_OPERATION_MOUNT, 
 *                    FSS_VALIDATE_OPERATION_RELABEL,
 *                    FSS_VALIDATE_OPERATION_SYNC and 
 *                    FSS_VALIDATE_OPERATION_UNMOUNT the string contains the
 *                    volume name only, e.g. "/ram".
 *                    For operation FSS_VALIDATE_OPERATION_INTERNAL the string
 *                    may contain volume name or full path to the object in the
 *                    volume.
 *
 * @input lnhname     Pointer to a string containing linkhandler name if the
 *                    object is accessed with a linkhandler path from another
 *                    cpu node, e.g. "slave" if the path is 
 *                    "slave!/ram/mydir/myfile.txt".
 *                    Empty string if no link handler name defined.
 *
 * @input opt_string  Pointer to a string containing optional string. The string
 *                    contains string arguments given to the CRT function.
 *                    Valid for operations and meaning:
 *                    - FSS_VALIDATE_OPERATION_RENAME: full path of the new file
 *                      name.
 *                    - FSS_VALIDATE_OPERATION_FORMAT: parameter string.
 *                    - FSS_VALIDATE_OPERATION_MOUNT: merged format name string,
 *                      device name string and parameter string.
 *                      Syntax of merged string: 
 *                      "format:<format> dev:<device name> params:<parameter
 *                      string>".
 *                    - FSS_VALIDATE_OPERATION_RELABEL: new name of label.
 *                    Empty string for operations without defined opt_string.
 *
 * @input open_mode   Permission attributes for the opened object.
 *                    This is the open mode argument given to the CRT function.
 *                    Valid for operation FSS_VALIDATE_OPERATION_OPEN.
 *                    Not valid for other operations.
 *
 * @input access_mode Permission attributes for the object.
 *                    This is the access mode argument given to the CRT
 *                    function.
 *                    Valid for operations FSS_VALIDATE_OPERATION_ACCESS,
 *                    FSS_VALIDATE_OPERATION_CHMOD, FSS_VALIDATE_OPERATION_OPEN
 *                    and FSS_VALIDATE_OPERATION_MKDIR.
 *                    For operations FSS_VALIDATE_OPERATION_OPEN and 
 *                    FSS_VALIDATE_OPERATION_MKDIR the file permission bits of
 *                    access_mode are modified by the file mode creation mask of
 *                    the client. The file mode creation mask of the client is 
 *                    set with the function umask().
 *                    Not valid for other operations.
 *               
 * @input username    Pointer to a string containing the name of the user doing
 *                    a volume access.
 *                    Empty string if no user defined.
 *
 * @input termname    Pointer to a string containing the name of the terminal
 *                    which can be used to identify which terminal type the user
 *                    is logged in over. 
 *                    For the serial port this is the name of the volume when
 *                    mounting the console format manager (confm) and serial
 *                    device driver (serdd), e.g. "/tty".
 *                    For the telnet port this is the name of the volume when
 *                    mounting the console format manager for the telnet port,
 *                    syntax: "/telnet_<addr>:<port>".
 *                    For the remote shell connection this is the remote
 *                    terminal name with the linkhandler path added, e.g.
 *                    "master!/tty".
 *                    An FTP-server provides "ftp" as the termname.
 *                    Empty string if no terminal defined.
 *
 * @input caller_pid  Pid of the process that called the CRT function (FSS
 *                    client pid).
 *                    If linkhandler path is used the pid can be used to get the
 *                    linkhandler name of the cpu node where the CRT function
 *                    was called.
 *
 * @return The function shall return True or False, if True the access control
 *         succeeded, if False the access failed.
 *
 * @seealso fss_enable_validate_volume_access
 * @seealso validate_user
 * @seealso validate_cmd
 * @seealso validate_help
 * @seealso logout_user
 */
Boolean
fss_validate_volume_access(U32        operation,
                           const char *obj_path,
                           const char *lnhname,
                           const char *opt_string,
                           U32        open_mode,
                           U32        access_mode,
                           const char *username,
                           const char *termname,
                           PROCESS    caller_pid);

#ifdef __cplusplus
}
#endif

#endif /* #ifndef _FSS_VALIDATE_H */
