/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2012 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */


/**************************************************************************
 * SIGNAL RANGES ASSIGNED TO THIS FILE
 *
 * 33300 - 33389
 * 33400 - 33499
 *
 *************************************************************************/



/**
 * @toc Memory_Management:
 *
 * @file ose_spi/heap.sig
 *
 * @brief The OSE Heap Manager signal interface. This interface provides
 *        functionality to obtain information about heaps and their buffers.
 *
 * @long  The heap concept is changed from OSE 4 to OSE 5.
 *        The heap is in OSE 5 identified by a unique heap reference, instead
 *        of by a segment ID as in OSE 4.
 *        The heap reference can be obtained by the call
 *        @link "heapapi.h" "ose_heap_ref" in heapapi.h,
 *        using either the actual process or a heap owner PID as parameter.
 *
 *
 * @seealso ose_spi/heapapi.h
 */
#ifndef _HEAP_SIG
#define _HEAP_SIG

#include "ose.h"
#include "osetypes.h"
#include "mm.sig"
#include "heapapi.h"

#ifdef __cplusplus
extern "C" {
#endif

/* Status values. */

/**
 * @macro HEAP_SUCCESS
 *
 * @brief Status: Heap information request was successful.
 */
#define HEAP_SUCCESS        0

/**
 * @macro HEAP_NO_MORE_ITEMS
 *
 * @brief Status: There were no more buffers in the heap.
 */
#define HEAP_NO_MORE_ITEMS  1

/**
 * @macro HEAP_ENOT_FOUND
 *
 * @brief Status: No heap was found.
 */
#define HEAP_ENOT_FOUND 2

/**
 * @macro HEAP_INVALID_REQUEST
 *
 * @brief Status: The request was invalid and could not be processed.
 */
#define HEAP_INVALID_REQUEST 3

/**
 * @macro HEAP_EFRAG_NOT_FOUND
 *
 * @brief Status: The heap fragment was not found.
 */
#define HEAP_EFRAG_NOT_FOUND 4

/**
 * @macro HEAP_EOUT_OF_MEMORY
 *
 * @brief Status: The system ran out of memory and the
 *        heap request could not be completed.
 */
#define HEAP_EOUT_OF_MEMORY 5

/**
 * @macro HEAP_ETOO_MANY_ELEMENTS
 *
 * @brief Status: A request for a list of elements would require a reply
 *                signal that is larger than the maximum signal size.
 *                The reply signal is valid with the exception that the
 *                number of elements represents ammount that is returned
 *                in the signal, not the actual number present.
 */
#define HEAP_ETOO_MANY_ELEMENTS           7

/*
 * Legacy defines that gives match on all buffers.
 */

#define HEAP_ALL 0
#define HEAP_SHARED 0

/**
 * @macro HEAP_MAX_FILENAME_LENGTH
 *
 * @brief Max length of filename, returned in the HeapBufferInfoReply's data
 *        section in the signal @link "" "HEAP_BUFFER_INFO_REPLY".
 *
 * @seealso HEAP_BUFFER_INFO_REPLY
 */
#define HEAP_MAX_FILENAME_LENGTH 256

/******************************************************************************
 *
 * HEAP_INTERFACE_REQUEST/HEAP_INTERFACE_REPLY
 *
 ******************************************************************************
 */

/**
 * @signal HEAP_INTERFACE_REQUEST
 * @brief  Signal to query for the supported Heap interface.
 *
 * @member HeapInterfaceRequest
 */
#define HEAP_INTERFACE_REQUEST    (33400) /* !-SIGNO(struct HeapInterfaceRequest)-! */

/**
 * @signal HEAP_INTERFACE_REPLY
 * @brief  Reply signal for @link "" "HEAP_INTERFACE_REQUEST"
 *
 * @member HeapInterfaceReply
 */
#define HEAP_INTERFACE_REPLY      (33401) /* !-SIGNO(struct HeapInterfaceReply)-! */

/**
 * @type HeapInterfaceRequest
 *
 * @brief Signal struct used for heap interface requests.
 * @field sigNo  Signal number.
 */
struct HeapInterfaceRequest
{
    SIGSELECT sigNo;
};

/**
 * @type HeapInterfaceReply
 *
 * @brief Signal struct used for heap interface request replies.
 *
 * @field sigNo    Signal number.
 * @output status  Status of the operation.
 * @output whatStr What-string of the product.
 * @output sigs    Supported signals, first element is count.
 */
struct HeapInterfaceReply
{
    SIGSELECT sigNo;
    U32       status;
    char      whatStr[32];
    SIGSELECT sigs[1];
};

/**
 * @macro sendHeapInterfaceRequest
 *
 * @brief Macro to send heap interface request to the specified heap.
 *
 * @input heapPidMA Heap PID to receive the request.
 *
 * @seealso HEAP_INTERFACE_REQUEST
 */
#define sendHeapInterfaceRequest(heapPidMA) \
    do { \
	union SIGNAL *sigMV = alloc(sizeof(struct HeapInterfaceRequest), HEAP_INTERFACE_REQUEST); \
	send(&sigMV, heapPidMA); \
    } while (0)

/******************************************************************************
 *
 * HEAP_LIST_REQUEST/HEAP_LIST_REPLY
 *
 ******************************************************************************
 */

/**
 * @signal HEAP_LIST_REQUEST
 * @brief  Signal to request a list of heaps in the system.
 *
 * @member HeapListRequest
 */
#define HEAP_LIST_REQUEST         (33402) /* !-SIGNO(struct HeapListRequest)-! */

/**
 * @signal HEAP_LIST_REPLY
 * @brief  Reply signal for @link "" "HEAP_LIST_REQUEST"
 *
 * @member HeapListReply
 */
#define HEAP_LIST_REPLY           (33403) /* !-SIGNO(struct HeapListReply)-! */

/**
 * @type HeapListRequest
 *
 * @field sigNo  Signal number.
 * @brief Signal struct used for heap list requests.
 */
struct HeapListRequest
{
    SIGSELECT sigNo;
};

/**
 * @type HeapListReply
 *
 * @brief Signal struct used for heap list request replies.
 *
 * @field sigNo      Signal number.
 * @output status    Status of the operation.
 * @output num_heaps Number of elements in heaps array.
 * @output heaps     Heaps. Owner PID and heap reference.
 * @field heaps.ownerpid Owning PID.
 * @field heaps.href Heap reference.
 */
struct HeapListReply
{
    SIGSELECT sigNo;
    U32       status;
    U32       num_heaps;
    struct _JustToFoolSigDB1
    {
	PROCESS    ownerpid;   /* Owner PID (main block or phantom). */
	OSHEAP     href;       /* heap reference */
    } heaps[1];
};

/******************************************************************************
 *
 * HEAP_CONFIG_REQUEST/HEAP_CONFIG_REPLY
 *
 ******************************************************************************
 */

/**
 * @signal HEAP_CONFIG_REQUEST
 * @brief  Signal to acquire configuration information for a specified heap.
 * @long  The heap reference is the primary search key.
 *
 * @member HeapConfigRequest
 */
#define HEAP_CONFIG_REQUEST       (33408) /* !-SIGNO(struct HeapConfigRequest)-! */

/**
 * @signal HEAP_CONFIG_REPLY
 * @brief  Reply signal for @link "" "HEAP_CONFIG_REQUEST"
 *
 * @member HeapConfigReply
 */
#define HEAP_CONFIG_REPLY         (33409) /* !-SIGNO(struct HeapConfigReply)-! */

/**
 * @type HeapConfigRequest
 *
 * @brief Signal struct used for heap configuration requests.
 *
 * @field sigNo     Signal number.
 * @input href      Heap reference to query for. href is the primary search key.
 *
 */
struct HeapConfigRequest
{
   SIGSELECT sigNo;
   OSHEAP     href;
};

/**
 * @type HeapConfigReply
 *
 * @brief Signal struct used for heap configuration request replies.
 *
 * @field sigNo             Signal number.
 * @output status           Status of the operation.
 * @output ownerpid         The heap owner (main block or phantom).
 * @output small_init_size  Size of initial fragment.
 * @output small_ext_size   Size of extension fragments.
 * @output large_threshold  Buffers larger than this value breaks from MM.
 * @output max_size         Total max allowed for the heap.
 * @output small_low_limit  Low system threshold when allocating smallbufs.
 * @output large_low_limit  Low system threshold when allocating largebufs.
 * @output domain           MMU domain ID.
 * @output log_mem          Logical memory to allocate heap from.
 *                          If several logical memory types have been
 *                          configured, only the name of the first and primary
 *                          type will be returned in this field.
 * @output permissions      Protection permissions.
 * @output cache_mode       Cache attributes.
 * @output endmark_size     Size of buffer endmark (in bytes).
 * @output annotation_size  Size of annotation space in each buffer.
 * @output wipe_on_free     Boolean flag indicating if wipe-on-free is enabled.
 * @output error_on_null    If True, error() is called from malloc() before
 *                          returning NULL.
 */
struct HeapConfigReply
{
   SIGSELECT sigNo;
   U32       status;
   PROCESS   ownerpid;
   U32       small_init_size;
   U32       small_ext_size;
   U32       large_threshold;
   U32       max_size;
   U32       small_low_limit;
   U32       large_low_limit;
   OSDOMAIN  domain;
   char      log_mem[MM_NAME_LENGTH + 1];
   U32       permissions;
   U32       cache_mode;
   U32       endmark_size;
   U32       annotation_size;
   U32       wipe_on_free;
   U32       error_on_null;
};

/**
 * @macro sendHeapConfigRequest
 *
 * @brief Macro to send heap configuration information requests.
 *
 * @input  hrefMA     Heap reference to query for.
 * @input  heapdPidMA Heapd PID to receive the request.
 *
 * @seealso HEAP_CONFIG_REQUEST
 */
#define sendHeapConfigRequest(hrefMA, heapdPidMA) \
    do { \
      /* The union is a trick to avoid breaking the strict-aliasing rules*/\
      union U \
      { \
         union SIGNAL *s;\
         struct HeapConfigRequest *x;\
      } u;\
      struct HeapConfigRequest *req;\
      u.s = alloc(sizeof(struct HeapConfigRequest), HEAP_CONFIG_REQUEST); \
      req = u.x;\
      req->href = (hrefMA); \
      send(&u.s , (heapdPidMA)); \
    } while (0)

/******************************************************************************
 *
 * HEAP_STATS_REQUEST/HEAP_STATS_REPLY
 *
 * This signal has been obsolete by HEAP_STATSEX_REQUEST and
 * HEAP_FRAGS_REQUEST.
 *
 ******************************************************************************
 */

/**
 * @signal HEAP_STATS_REQUEST
 * @brief  Obsolete. Signal to acquire statistics about a specified heap.
 *
 * @long   Obsolete by @link "" "HEAP_STATSEX_REQUEST" and
 *         @link "" "HEAP_FRAGS_REQUEST".
 *
 * @long   The heap reference is the primary search key.
 *
 * @member HeapStatsRequest
 */
#define HEAP_STATS_REQUEST        (33404) /* !-SIGNO(struct HeapStatsRequest)-! */
/**
 * @signal HEAP_STATS_REPLY
 * @brief  Reply signal for @link "" "HEAP_STATS_REQUEST". Obsolete.
 *
 * @long   Obsolete by @link "" "HEAP_STATSEX_REPLY" and
 *         @link "" "HEAP_FRAGS_REPLY".
 *
 * @member HeapStatsReply
 */
#define HEAP_STATS_REPLY          (33405) /* !-SIGNO(struct HeapStatsReply)-! */

/**
 * @type HeapStatsRequest
 *
 * @brief Obsolete. Signal struct used for heap statistics requests.
 *        See instead HeapStatsExRequest and HeapFragsRequest.
 *
 * @field sigNo    Signal number.
 *
 * @input href     Heap reference. href is the primary search key.
 *
 */
struct HeapStatsRequest
{
   SIGSELECT  sigNo;
   OSHEAP     href;
};

/**
 * @type HeapStatsReply
 *
 * @brief Obsolete. Signal struct used for heap statistics request replies.
 *        See instead HeapStatsExReply and HeapFragsReply.
 *
 * @field sigNo            Signal number.
 * @output status          Status of the operation.
 * @output total_req       Total amount requested.
 * @output total_used      Total amount actually used.
 * @output total_free      Total amount free.
 * @output largest_free    Size of largest free buffer.
 * @output number_private  Number of private buffers.
 * @output number_shared   Number of shared buffers.
 * @output num_frags       Number of elements in frags array.
 * @output frags           Array with fragments in the heap. Variable size.
 * @output frags.addr      Fragment address
 * @output frags.size      Fragment size
 */
struct HeapStatsReply
{
   SIGSELECT sigNo;
   U32       status;
   U32       total_req;
   U32       total_used;
   U32       total_free;
   U32       largest_free;
   U32       number_private;
   U32       number_shared;
   U32       num_frags;

   struct _JustToFoolSigDB2
   {
      U32 addr;              /* Address of fragment. */
      U32 size;              /* Size of fragment. */
   } frags[1];                /* O: Fragments in the heap (variable size). */
};

/**
 * @macro sendHeapStatsRequest
 *
 * @brief Obsolete. Macro to send heap statistics requests.
 *
 * @long  Obsolete by @link "" "sendHeapStatsExRequest" and
 *        @link "" "sendHeapFragsRequest".
 *
 * @input hrefMA     Heap reference to query for.
 * @input heapdPidMA Heapd PID to receive the request.
 */
#define sendHeapStatsRequest(hrefMA, heapdPidMA) \
    do { \
         /* The union is a trick to avoid breaking the strict-aliasing rules*/\
         union U \
         { \
            union SIGNAL *s;\
            struct HeapStatsRequest *x;\
         } u;\
	 struct HeapStatsRequest *req;\
         u.s = alloc(sizeof(struct HeapStatsRequest), HEAP_STATS_REQUEST); \
         req = u.x;\
	 req->href = (hrefMA); \
	send(&u.s, (heapdPidMA)); \
    } while (0)

/******************************************************************************
 *
 * HEAP_BUFFER_INFO_REQUEST/HEAP_BUFFER_INFO_REPLY
 *
 ******************************************************************************
 */

/**
 * @signal HEAP_BUFFER_INFO_REQUEST
 * @brief  Signal to request information about buffers from a specified heap.
 *
 * @long  Using this signal, one can traverse through the list of used buffers,
 *        e.g. to find a specific heap buffer.
 *
 * @long  Many of the fields are to be filled in with info from a
 *        previous @link "" "HEAP_BUFFER_INFO_REPLY" signal,
 *        which uses struct HeapBufferInfoReply.
 *
 * @long  The heap reference is the primary search key.
 *
 * @member HeapBufferInfoRequest
 */
#define HEAP_BUFFER_INFO_REQUEST  (33406) /* !-SIGNO(struct HeapBufferInfoRequest)-! */

/**
 * @signal HEAP_BUFFER_INFO_REPLY
 * @brief  Reply signal for @link "" "HEAP_BUFFER_INFO_REQUEST"
 *
 * @member HeapBufferInfoReply
 */
#define HEAP_BUFFER_INFO_REPLY    (33407) /* !-SIGNO(struct HeapBufferInfoReply)-! */

/**
 * @type HeapBufferInfoRequest
 *
 * @brief Signal struct used for heap buffer information requests. Many fields
 *        should be filled in from previous HeapBufferInfoReply.
 *
 * @field sigNo    Signal number.
 * @input href     Heap reference to query for. href is the primary search key.
 * @input frag     Set to zero first request then retain from last reply.
 * @input path     Set to zero first request then retain from last reply.
 * @input depth    Set to zero first request then retain from last reply.
 * @input owner    List buffers owned by owner. (zero for all)
 * @input max_data Max number of data bytes to return.
 * @input burst    Number of replies requested.
 *
 */
struct HeapBufferInfoRequest
{
    SIGSELECT  sigNo;
    OSHEAP     href;
    U32 frag;
    U32 path;
    U32 depth;
    PROCESS owner;
    U32 max_data;
    U32 burst;
};

/**
 * @type HeapBufferInfoReply
 *
 * @brief Signal struct used for heap buffer information request replies.
 *
 * @field sigNo       Signal number.
 * @output status     Status of the operation.
 * @output frag       Address of heap fragment.
 * @output path       Path to buffer
 * @output depth      Depth in tree of buffer.
 * @output address    Address of buffer, corresponding to the pointer returned
 *                    by malloc.
 * @output owner      PID of owner or 0 for shared.
 * @output file       File name of allocation, index into data or ~0.
 * @output line       Line number of allocation.
 * @output req_size   Requested size of buffer.
 * @output act_size   Actual size of buffer (overhead, internal fragmentation).
 * @output ann_size   Size of annotation part of data (zero if not annotated)
 * @output data_size  Number of bytes sampled, first in data.
 * @output endmark_ok True if endmark ok, otherwise False.
 * @output data       Data sampled from buffer, file name appended and possibly
 *                    beginning with annotation.
 */
struct HeapBufferInfoReply
{
    SIGSELECT sigNo;
    U32       status;
    U32       frag;
    U32       path;
    U32       depth;
    U32       address;
    U32       owner;
    U32       file;
    U32       line;
    U32       req_size;
    U32       act_size;
    U32       ann_size;
    U32       data_size;
    U8        endmark_ok;
    U8        data[1];
};

/**
 * @macro sendFirstHBIRequest
 *
 * @brief Macro to send first buffer information request signal.
 *
 * @input hrefMA     Heap reference to query for.
 * @input ownerMA    Owning PID.
 * @input max_dataMA Max number of data bytes to return.
 * @input burstMA    Number of replies requested.
 * @input heapdPidMA Heapd PID to receive request.
 *
 * @seealso HEAP_BUFFER_INFO_REQUEST
 * @seealso sendNextHBIRequest
 */
#define sendFirstHBIRequest(hrefMA, ownerMA, max_dataMA, burstMA, heapdPidMA) \
    do { \
        /* The union is a trick to avoid breaking the strict-aliasing rules*/\
        union U \
        { \
           union SIGNAL *s;\
           struct HeapBufferInfoRequest *x;\
        } u;\
	struct HeapBufferInfoRequest *req;\
        u.s = alloc(sizeof(struct HeapBufferInfoRequest), HEAP_BUFFER_INFO_REQUEST);\
        req = u.x;\
	req->href = (hrefMA); \
	req->frag = 0; \
	req->path = 0; \
	req->depth = 0; \
	req->owner = (ownerMA); \
	req->max_data = (max_dataMA); \
	req->burst = (burstMA); \
	send(&u.s, (heapdPidMA)); \
    } while (0)

/**
 * @macro sendNextHBIRequest
 *
 * @brief Macro to send next buffer information request signal.
 * @long  Requires previous use of @link "" "sendFirstHBIRequest".
 *
 * @input replyMA    HeapBufferInfoReply struct to use for retaining
 *                   previous data.
 * @input hrefMA     Heap reference to query for.
 * @input ownerMA    Owning PID.
 * @input max_dataMA Max number of data bytes to return.
 * @input burstMA    Number of replies requested.
 * @input heapdPidMA Heapd PID to receive request.
 *
 * @seealso HEAP_BUFFER_INFO_REQUEST
 * @seealso sendFirstHBIRequest
 */
#define sendNextHBIRequest(replyMA, hrefMA, ownerMA, max_dataMA, burstMA, heapdPidMA) \
    do { \
        /* The union is a trick to avoid breaking the strict-aliasing rules*/\
        union U \
        { \
           union SIGNAL *s;\
           struct HeapBufferInfoRequest *x;\
        } u;\
	struct HeapBufferInfoRequest *req;\
        u.s = alloc(sizeof(struct HeapBufferInfoRequest), HEAP_BUFFER_INFO_REQUEST);\
        req = u.x;\
	req->href = (hrefMA); \
	req->frag = (replyMA)->frag; \
	req->path = (replyMA)->path; \
	req->depth = (replyMA)->depth; \
	req->owner = (ownerMA); \
	req->max_data = (max_dataMA); \
	req->burst = (burstMA); \
	send(&u.s, (heapdPidMA)); \
    } while (0)


/******************************************************************************
 *
 * HEAP_STATSEX_REQUEST/HEAP_STATSEX_REPLY
 *
 ******************************************************************************
 */

/**
 * @signal HEAP_STATSEX_REQUEST
 * @brief  Signal to acquire statistics about a specified heap.
 *
 * @long   The heap reference is the primary search key.
 *
 * @long   The HEAP_FRAGS_REQUEST and HEAP_STATSEX_REQUEST
 *         have replaced the earlier HEAP_STATS_REQUEST.
 *
 * @member HeapStatsExRequest
 */
#define HEAP_STATSEX_REQUEST      (33410) /* !-SIGNO(struct HeapStatsExRequest)-! */

/**
 * @signal HEAP_STATSEX_REPLY
 * @brief  Reply signal for @link "" "HEAP_STATSEX_REQUEST"
 *
 * @member HeapStatsExReply
 */
#define HEAP_STATSEX_REPLY        (33411) /* !-SIGNO(struct HeapStatsExReply)-! */
/**
 * @type HeapStatsExRequest
 *
 * @brief Signal struct used for heap statistics requests.
 *
 * @field sigNo       Signal number.
 * @input href        Heap reference to query for. href is the primary search key.
 *
 */
struct HeapStatsExRequest
{
   SIGSELECT  sigNo;
   OSHEAP     href;
};

/**
 * @type HeapStatsExReply
 *
 * @brief Signal struct used for heap statistics request replies.
 *
 * @field sigNo             Signal number.
 * @output status           Status of the operation.
 * @output  total_req       Total amount requested.
 * @output  total_used      Total amount actually used.
 * @output  total_used_hi   High watermark of the above counter
 * @output  total_free      Total amount free.
 * @output  largest_free    Size of largest free buffer.
 * @output  number_private  Number of private buffers.
 * @output  number_shared   Number of shared buffers.
 * @output  malloc_failed   Number of failed malloc() attempts
 * @output  reserved        Reserved. Set to zero.
 * @output  num_sizes       Number of different buffer sizes.
 * @output  size            Size of heap buffers (0..num_sizes-1).
 * @output  num_used        Number of used buffers per size interval,
 *                          (size[i]..size[i+1).
 * @output  num_free        Number of free buffers per size interval,
 *                          (size[i]..size[i+1).
 */
struct HeapStatsExReply
{
    SIGSELECT sigNo;
    U32       status;
    U32       total_req;
    U32       total_used;
    U32       total_used_hi;
    U32       total_free;
    U32       largest_free;
    U32       number_private;
    U32       number_shared;
    U32       malloc_failed;
    U32       reserved[5];
    U32       num_sizes;
    U32       size[32];
    U32       num_used[32];
    U32       num_free[32];
};

/**
 * @macro sendHeapStatsExRequest
 *
 * @brief Macro to send heap statistics requests.
 *
 * @long   The sendHeapFragsRequest and sendHeapStatsExRequest
 *         have replaced the earlier sendHeapStatsRequest.
 *
 * @input hrefMA      Heap reference to query for.
 * @input heapdPidMA  Heapd PID to receive the requests.
 *
 * @seealso HEAP_STATSEX_REQUEST
 */
#define sendHeapStatsExRequest(hrefMA, heapdPidMA) \
    do { \
        /* The union is a trick to avoid breaking the strict-aliasing rules*/\
        union U \
        { \
           union SIGNAL *s;\
           struct HeapStatsExRequest *x;\
        } u;\
	struct HeapStatsExRequest *req;\
        u.s = alloc(sizeof(struct HeapStatsExRequest), HEAP_STATSEX_REQUEST); \
        req = u.x;\
	req->href = (hrefMA); \
	send(&u.s, (heapdPidMA)); \
    } while (0)

/******************************************************************************
 *
 * HEAP_FRAGS_REQUEST/HEAP_FRAGS_REPLY
 *
 ******************************************************************************
 */

/**
 * @signal HEAP_FRAGS_REQUEST
 * @brief  Signal to acquire the fragment list of a specified heap.
 *
 * @long  The heap reference is the primary search key.
 *
 * @member HeapFragsRequest
 */
#define HEAP_FRAGS_REQUEST        (33412) /* !-SIGNO(struct HeapFragsRequest)-! */
/**
 * @signal HEAP_FRAGS_REPLY
 * @brief  Reply signal for @link "" "HEAP_FRAGS_REQUEST"
 *
 * @member HeapFragsReply
 */
#define HEAP_FRAGS_REPLY          (33413) /* !-SIGNO(struct HeapFragsReply)-! */

/**
 * @type HeapFragsRequest
 *
 * @brief Signal struct used for heap fragment requests.
 *
 * @field sigNo     Signal number.
 * @input href      Heap reference to query for. href is the primary search key.
 *
 */
struct HeapFragsRequest
{
   SIGSELECT sigNo;
   OSHEAP     href;
};

/**
 * @type HeapFragsReply
 *
 * @brief Signal struct used for heap fragment request replies.
 *
 * @field sigNo      Signal number.
 * @output status    Status of the operation.
 * @output num_frags Number of elements in frags array.
 * @output frags     Array with fragments in the heap. Variable size.
 * @output frags.addr Fragment address
 * @output frags.size Fragment size
 */
struct HeapFragsReply
{
    SIGSELECT sigNo;
    U32       status;
    U32       num_frags;
    struct _JustToFoolSigDB3
    {
	U32 addr;              /* Address of fragment. */
	U32 size;              /* Size of fragment. */
    } frags[1];                /* O: Fragments in the heap (variable size). */
};

/**
 * @macro sendHeapFragsRequest
 *
 * @brief Macro to send heap fragment list requests.
 *
 * @long   The sendHeapFragsRequest and sendHeapStatsExRequest
 *         have replaced the earlier sendHeapStatsRequest.
 *
 * @input hrefMA     Heap reference to query for.
 * @input heapdPidMA Heapd PID to receive the request.
 *
 * @seealso HEAP_FRAGS_REQUEST
 */
#define sendHeapFragsRequest(hrefMA, heapdPidMA) \
    do { \
        /* The union is a trick to avoid breaking the strict-aliasing rules*/\
        union U \
        { \
           union SIGNAL *s;\
           struct HeapFragsRequest *x;\
        } u;\
	struct HeapFragsRequest *req;\
        u.s = alloc(sizeof(struct HeapFragsRequest), HEAP_FRAGS_REQUEST); \
        req = u.x;\
        req->href = (hrefMA); \
        send(&u.s, (heapdPidMA)); \
    } while (0)

/******************************************************************************
 *
 * HEAP_RESET_STATS_REQUEST/HEAP_RESET_STATS_REPLY
 *
 ******************************************************************************
 */

/**
 * @signal HEAP_RESET_STATS_REQUEST
 * @brief  Signal to reset statistics for a specified heap.
 *
 * @long  The heap reference is the primary search key.
 *
 * @member HeapResetStatsRequest
 */
#define HEAP_RESET_STATS_REQUEST  (33414) /* !-SIGNO(struct HeapResetStatsRequest)-! */

/**
 * @signal HEAP_RESET_STATS_REPLY
 * @brief  Reply signal for @link "" "HEAP_RESET_STATS_REQUEST"
 *
 * @member HeapResetStatsReply
 */
#define HEAP_RESET_STATS_REPLY    (33415) /* !-SIGNO(struct HeapResetStatsReply)-! */

/* Possible values for 'flags': */
/**
 * @macro HEAP_RESET_WATERMARKS
 *
 * @brief Flag to reset watermarks counter, in field flags in
 *        @link "" "HEAP_RESET_STATS_REQUEST".
 */
#define HEAP_RESET_WATERMARKS     0x00000001

/**
 * @type HeapResetStatsRequest
 *
 * @brief Signal struct used for requests to reset heap statistics.
 *
 * @field sigNo    Signal number.
 *
 * @input href     Heap reference of the heap to reset.
 *                 href is the primary search key.
 *
 * @input flags    Flags specifying which counters to reset.
 *
 */
struct HeapResetStatsRequest
{
   SIGSELECT  sigNo;
   OSHEAP     href;
   U32        flags;
};

/**
 * @type HeapResetStatsReply
 *
 * @brief Signal struct used when replying to a request to reset
 *        heap statistics.
 *
 * @output sigNo  Signal number.
 * @output status Status of the operation.
 */
struct HeapResetStatsReply
{
    SIGSELECT sigNo;
    U32       status;
};

/**
 * @macro sendHeapResetStatsRequest
 *
 * @brief Macro to send a request to reset heap statistics.
 *
 * @input hrefMA     Heap reference of the heap to reset.
 * @input flagsMA    Flags specifying which counters to reset.
 * @input heapdPidMA Heapd PID to receive the request.
 *
 * @seealso HEAP_RESET_STATS_REQUEST
 */

#define sendHeapResetStatsRequest(hrefMA, flagsMA, heapdPidMA) \
    do { \
        /* The union is a trick to avoid breaking the strict-aliasing rules*/\
        union U \
        { \
           union SIGNAL *s;\
           struct HeapResetStatsRequest *x;\
        } u;\
	struct HeapResetStatsRequest *req;\
        u.s = alloc(sizeof(struct HeapResetStatsRequest), HEAP_RESET_STATS_REQUEST); \
        req = u.x;\
        req->href = (hrefMA); \
        req->flags = (flagsMA); \
        send(&u.s, (heapdPidMA)); \
    } while (0)


/******************************************************************************
 *
 * HEAP_MONITOR_REQUEST / HEAP_MONITOR_REPLY / HEAP_MONITOR_EVENT
 *
 ******************************************************************************
 */

/**
 * @signal HEAP_MONITOR_REQUEST
 * @brief  Signal to setup a monitor of a specific heap.
 * @long  This signal initiates a monitoring of the heap memory levels. On
 *        allocation and freeing of heap buffers, the total amount of
 *        memory used by the heap (including buffer overhead) is compared to
 *        @link "ose_heap.h" "HEAP_MAX_SIZE". If the amount of available memory
 *        goes under the heap_low_level, an event is triggered and a signal
 *        @link "" "HEAP_MONITOR_EVENT" is sent to the monitor_pid with the
 *        event specifier @link "" "HEAP_EVENT_LOW_LEVEL".
 *        Similarly, when the amount of available memory in the heap increases
 *        and passes the heap_not_low_level, another @link ""
 *        "HEAP_MONITOR_EVENT" is sent with the event specifier @link ""
 *        "HEAP_EVENT_NOT_LOW_LEVEL".
 * @long  The heap reference is the primary search key.
 *
 * @member HeapMonitorRequest
 */
#define HEAP_MONITOR_REQUEST      (33416) /* !-SIGNO(struct HeapMonitorRequest)-! */

/**
 * @signal  HEAP_MONITOR_REPLY
 * @brief   Reply signal for @link "" "HEAP_MONITOR_REQUEST"
 * @long    The reply signal is sent from the heap daemon when a
 *          @link "" "HEAP_MONITOR_REQUEST" signal has been received, reporting
 *          the status for the operation.
 *
 * @member  HeapMonitorReply
 * @seealso HEAP_MONITOR_REQUEST
 */
#define HEAP_MONITOR_REPLY        (33417) /* !-SIGNO(struct HeapMonitorReply)-! */

/**
 * @signal HEAP_MONITOR_EVENT
 * @brief  Signal to report heap events.
 *
 * @member  HeapMonitorEvent
 * @seealso HEAP_MONITOR_REQUEST
 */
#define HEAP_MONITOR_EVENT        (33418) /* !-SIGNO(struct HeapMonitorEvent)-! */

/* Possible values for 'event': */
/**
 * @macro HEAP_EVENT_LOW_LEVEL
 *
 * @brief   Event signaled when available heap memory goes down under low_level.
 * @seealso HEAP_MONITOR_EVENT
 */
#define HEAP_EVENT_LOW_LEVEL     1

/**
 * @macro HEAP_EVENT_NOT_LOW_LEVEL
 *
 * @brief   Event signaled when available heap memory goes up over
 *          not_low_level.
 * @seealso HEAP_MONITOR_EVENT
 */
#define HEAP_EVENT_NOT_LOW_LEVEL 2

/**
 * @type HeapMonitorRequest
 *
 * @brief Signal struct used for heap monitoring requests.
 *
 * @field sigNo              Signal number.
 * @input href               Heap reference to be monitored.
 * @input monitor_pid        PID of the monitor process that will receive heap
 *                           level events.
 * @input heap_low_level     Level to trigger @link "" "HEAP_EVENT_LOW_LEVEL"
 *                           event. Must be smaller than 'heap_not_low_level'.
 * @input heap_not_low_level Level to trigger @link "" "HEAP_EVENT_NOT_LOW_LEVEL"
 *                           event. Must be larger than heap_low_level and
 *                           smaller than @link "ose_heap.h" "HEAP_MAX_SIZE".
 */
struct HeapMonitorRequest
{
   SIGSELECT  sigNo;
   OSHEAP     href;
   PROCESS    monitor_pid;
   U32        heap_low_level;
   U32        heap_not_low_level;
};

/**
 * @type HeapMonitorReply
 *
 * @brief Signal struct used for heap monitoring replies.
 *
 * @field sigNo   Signal number.
 * @output status Status of the operation.
 *                HEAP_SUCCESS if the operation was successful.
 *                HEAP_ENOT_FOUND if the heap referred to by href was not found.
 *                HEAP_INVALID_REQUEST if either the monitor process was invalid,
 *                or if the requested heap levels are infeasible.
*/
struct HeapMonitorReply
{
   SIGSELECT  sigNo;
   U32        status;
};

/**
 * @type HeapMonitorEvent
 *
 * @brief Signal struct used for heap monitor events.
 *
 * @field sigNo        Signal number.
 * @output href        Heap reference that signals the event.
 * @output event       Either @link "" "HEAP_EVENT_LOW_LEVEL"
 *                     or @link "" "HEAP_EVENT_NOT_LOW_LEVEL".
 * @output event_no    Sequence number of the event signal.
 *                     Event signals may arrive out of order and event_no
 *                     is used to verify that the latest received signal is not
 *                     older than any signal received earlier. event_no
 *                     overflows as a regular unsigned 32-bit integer.
 */
struct HeapMonitorEvent
{
   SIGSELECT  sigNo;
   OSHEAP     href;
   U32        event;
   U32        event_no;
};

/**
 * @macro sendHeapMonitorRequest
 *
 * @brief Macro to send heap monitor requests.
 *
 * @input hrefMA       Reference to heap that will be monitored.
 * @input monitorPidMA PID to monitor task that will receive level events.
 * @input lowMA        Level to trigger 'heap low' event.
 * @input notLowMA     Level to trigger 'heap not low' event.
 * @input heapdPidMA   Heapd PID to receive the request.
 *
 * @seealso HEAP_MONITOR_REQUEST
 */
#define sendHeapMonitorRequest(hrefMA, monitorPidMA, lowMA, notLowMA, heapdPidMA) \
    do { \
        /* The union is a trick to avoid breaking the strict-aliasing rules*/\
        union U \
        { \
           union SIGNAL *s;\
           struct HeapMonitorRequest *x;\
        } u;\
	struct HeapMonitorRequest *req;\
        u.s = alloc(sizeof(struct HeapMonitorRequest), HEAP_MONITOR_REQUEST); \
        req = u.x;\
	req->href = (hrefMA); \
	req->monitor_pid = (monitorPidMA); \
	req->heap_low_level = (lowMA); \
	req->heap_not_low_level = (notLowMA); \
	send(&u.s, (heapdPidMA)); \
    } while (0)


/**
 * @macro HEAP_METADATA_SIG
 *
 * @brief Signal number used to identify signal buffers used for heap metadata.
 */
#define HEAP_METADATA_SIG  (33419)


#ifdef __cplusplus
} /* end of extern "C" */
#endif

#endif /* _HEAP_SIG */
