/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2006-2010 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */
/**
 * @public
 * @toc Memory_Management:
 * @file ose_spi/heap_debug.h
 *
 * @brief The heap debug module.
 *
 * @long The file heap_debug.h contains the definitions of the heap debug
 *       module concept.
 *
 * @long For more information on how to use the OSE heap manager, see
 *       the "OSE Core User's Guide".
 *
 */
#ifndef _HEAP_DEBUG_H
#define _HEAP_DEBUG_H
#include <stddef.h>
#include "osetypes.h"
#include "ose.h"

#ifdef __cplusplus
extern "C" {
#endif

#if !defined(OSKERNEL)
static const U32 _endian_checker = 0x12345678;
#ifdef __ARMCC_VERSION
static volatile int get_rid_of_warning = _endian_checker;
#endif
#define _ENDIAN_IS_BIG() (0x78 ^ (*(const U8 *)&_endian_checker))
#else
  /* OSKERNEL */
  #if defined(BIG_ENDIAN) && defined(LITTLE_ENDIAN)
    #if   __BYTE_ORDER == __BIG_ENDIAN
      #undef LITTLE_ENDIAN
    #elif __BYTE_ORDER == __LITTLE_ENDIAN
      #undef BIG_ENDIAN
    #else
      #error Both BIG_ENDIAN and LITTLE_ENDIAN defined
    #endif
  #endif
  #if defined(LITTLE_ENDIAN)
    #define _ENDIAN_IS_BIG() 0
  #else
    #define _ENDIAN_IS_BIG() 1
  #endif
#endif

/************************************************************************
 *
 * GENERIC debug module declarations
 *
 * The declarations in this section are needed for the implementation of
 * the heap and are not only used in the debug module.
 *
 ***********************************************************************/
struct HeapBufferInfoReply;

/**
 * @type OSHEAP
 *
 * @brief Heap reference type.
 */
typedef OSADDRESS OSHEAP;

/**
 * @macro DA_LINE_MASK
 *
 * @brief Line number mask for control field in
 *        @link "" "DefAnnotation".
 */
#define DA_LINE_MASK (0xFFFFFFFE)


/**
 * @macro DA_LINE_OFFSET
 *
 * @brief Offset of line number within control field in
 *        @link "" "DefAnnotation".
 */
#define DA_LINE_OFFSET (1)

/**
 * @macro DA_ANNOT_MASK
 *
 * @brief Mask of annotated flag within control field in
 *        @link "" "DefAnnotation".
 */
#define DA_ANNOT_MASK (0x00000001)

/**
 * @macro HDM_SET_LINE
 *
 * @brief Set line in control field of @link "" "DefAnnotation".
 * @field ctrl Control field of 32 bits that will be updated.
 * @field line Line number to be set in control field.
 */
#define HDM_SET_LINE(ctrl,line) (ctrl = (ctrl & ~DA_LINE_MASK) | ((line << DA_LINE_OFFSET) & DA_LINE_MASK))

/**
 * @macro HDM_GET_LINE
 *
 * @brief Get line number from  control field of @link "" "DefAnnotation".
 * @field ctrl Control field of 32 bits that will be read.
 */
#define HDM_GET_LINE(ctrl)      ((ctrl & DA_LINE_MASK) >> DA_LINE_OFFSET)

/**
 * @macro HDM_IS_ANNOTATED
 *
 * @brief Check whether buffer with control field is annotated in
 *        @link "" "DefAnnotation".
 * @field ctrl Control field of buffer.
 */
#define HDM_IS_ANNOTATED(ctrl)  (ctrl & DA_ANNOT_MASK)

/**
 * @macro HDM_SET_ANNOTATED
 *
 * @brief Set annotation bit in control field of
 *        @link "" "DefAnnotation".
 * @field ctrl Control field of buffer.
 */
#define HDM_SET_ANNOTATED(ctrl) (ctrl |= DA_ANNOT_MASK)

/**
 * @macro SIZEOF_DEF_ANNOTATION
 *
 * @brief The nominal size of the @link "" "DefAnnotation" structure.
 */
#define SIZEOF_DEF_ANNOTATION (4 + sizeof(const char *))

/**
 * @type DefAnnotation
 *
 * @brief Default annotation structure.
 *
 * @long Default structure for annotation data. This structure must be placed at
 * the end of annotation data when annotation is turned on. The "annotated" bit
 * is not to be changed by the debug module.
 *
 * @field file      Name of file where allocation was performed, or NULL.
 * @field control   Includes line number of allocation, masked out by
 *                  DA_LINE_MASK, and an annotated flag, masked out by
 *                  DA_ANNOT_MASK, that is used to test whether a buffer is
 *                  annotated or not.
 * @example
 * DefAnnotation da;
 * char * file;
 * int line;
 * da = (DefaultAnnotation *) ((char*)buf_ptr - SIZEOF_DEF_ANNOTATION);
 * if (da->control & DA_ANNOT_MASK)
 * {
 *    file = da->file;
 *    line = (da->control & DA_LINE_MASK) >> DA_LINE_OFFSET;
 * }
 */
typedef struct
{
    const char *file;
    U32 control;
} DefAnnotation;

/**
 * @type HeapDebugPPData
 *
 * @brief Per process data used for heap debug module.
 *
 * @long If a debug module has been installed, the heap debug PP data contains
 * file and line information about the last allocation performed. The version
 * field is used to reflect the version of this structure used by the system.
 * The debug module can check that this field is equal to HEAP_DEBUG_PP_VERSION
 * to verify binary compatibility between CRT versions.
 *
 * @field sig_no      PPD_MAGIC_SIG
 * @field version     Version of debug pp data. Should equal HEAP_DEBUG_PP_VERSION.
 * @field file        Pointer to the filename or NULL if not available.
 * @field line        Line number of last allocation, or zero if not available.
 *
 */
typedef struct
{
   SIGSELECT sig_no;
   U32 version;
   const char *file;
   U32 line;
} HeapDebugPPData;

/**
 * @macro HEAP_DEBUG_PP_VERSION
 *
 * @brief Version number of @link "" "HeapDebugPPData".
 */
#define HEAP_DEBUG_PP_VERSION 20060302

/**
 * @macro PPD_MAGIC_SIG
 *
 * @brief Signal number of @link "" "HeapDebugPPData".
 */
#define PPD_MAGIC_SIG 31950

/**
 * @type DEBUG_ALLOC_FUNC
 *
 * @brief Prototype for heap allocation hook.
 *
 * @long This is the prototype for the heap allocation hook that needs to be
 * implemented by a heap debug module.
 *
 * @long When heap debug hooks have been properly installed using the call @link
 * "heapapi.h" "ose_heap_set_debug_hooks" the hook defined by this prototype
 * will be called during each allocation of a heap buffer.
 * The realloc function will cause an alloc hook call and free hook call
 * if the new size does not fit into the existing buffer. If the buffer is
 * resized to a smaller buffer, the realloc call will not cause any calls to
 * debug callbacks.
 *
 * @long The area for annotation data, of size annsz, will be available
 * immediately before the user area of the buffer which is pointed to by ptr.
 *
 * @input heap  A reference to the heap from which the allocation was made.
 * @input ptr   A pointer to the user part of the annotated buffer.
 * @input size  Original size requested by user in allocation call.
 * @input annsz Size of annotation area immediately preceding the user data.
 * @input owner Owner of the buffer
 * @input file  Name of file from which allocation was performed.
 * @input line  Line number of allocation call.
 *
 */
typedef void (*DEBUG_ALLOC_FUNC)(OSHEAP heap,
                                 void *ptr,
                                 size_t size,
                                 size_t annsz,
                                 PROCESS owner,
                                 const char *file,
                                 U32 line);

/**
 * @type DEBUG_FREE_FUNC
 *
 * @brief Prototype for free hook.
 *
 * @long This is the prototype for the heap free hook that needs to be
 * implemented by a heap debug module.
 *
 * @long When heap debug hooks have been properly installed using the call @link
 * "heapapi.h" "ose_heap_set_debug_hooks" the hook defined by this prototype
 * will be called during each freeing of a heap buffer.
 *
 * @input heap  A reference to the heap to which the freeing is made.
 * @input ptr   A pointer to the user part of the annotated buffer.
 * @input annsz Size of annotation area immediately preceding the user data.
 * @input owner Owner of the buffer
 *
 * @return The function can instruct the heap to not free the buffer
 *         by returning a non-zero value. The buffer will then have to
 *         be freed by the debug module at a later point in time using
 *         @link "heapapi.h" "heap_forced_free". @break
 *         The buffer is freed immediately if zero is returned.
 */
typedef int (*DEBUG_FREE_FUNC)(OSHEAP heap, void *ptr, size_t annsz,
             PROCESS owner);


/**
 * @type DEBUG_FILELINE_FUNC
 *
 * @brief Prototype for fileline hook.
 *
 * @long This is the prototype for the heap fileline hook that should be
 * implemented by a heap debug module.
 *
 * @long When heap debug hooks have been properly installed using the call @link
 * "heapapi.h" "ose_heap_set_debug_hooks" the hook defined by this prototype
 * will be called to fill in file and line information in the
 * HeapBufferInfoReply struct when either an allocation fails or the heap
 * daemon receives a signal request for heap buffer information.
 *
 * @long When the address field in hbi is zero, the hook was called due to an
 * allocation failure, and file and line information can be retrieved from the
 * heap debug pp data.
 *
 * @long When the address field in hbi is not zero, the hook was called due to
 * an information request from the heap daemon, and the file and line information
 * can be taken from the annotation area of the buffer identified by the
 * hbi's address field.
 *
 * @long The file name and line number is to be filled into 'hbi' updating
 * the fields 'file', 'line', and 'data' in hbi, according to specification
 * in HeapBufferInfoReply.
 *
 * @input heap     Heap reference
 * @input hbi      HeapBufferInfoReply struct. If address field is zero, no
 *                 buffer exists and file and line information need to be taken
 *                 from the heap debug pp data.
 * @output hbi     HeapBufferInfoReply struct. The line number should be copied
 *                 to the line field in hbi. The file name should be copied to
 *                 the data array, at an index given by the offset parameter,
 *                 and with a maximum length of HEAP_MAX_FILENAME_LENGTH.
 *                 The offset should also be filled into the file field of hbi
 *                 or, if no file name is available, the file field should be
 *                 set to ~0.
 * @input offset   Index for file name into data field of hbi.
 */
typedef void (*DEBUG_FILELINE_FUNC)(OSHEAP heap,
                                    struct HeapBufferInfoReply* hbi,
                                    U32 offset);


/**
 * @type DEBUG_CLEANUP_FUNC
 *
 * @brief Prototype for cleanup hook.
 *
 * @long This is the prototype for the cleanup hook that may be
 * implemented by a heap debug module. If not implemented, no cleanup will be
 * performed when a heap debug module is removed.
 *
 * @long This hook is called during an uninstallation of a heap debug module to
 * allow the debug module to cleanup, e.g. by using forced free to free any heap
 * buffers that are held by the module, or to free other resources used by the
 * module.
 *
 * @long The cleanup hook is always the last hook to be called in a heap debug
 * module. Once called, no other hook of that heap will be called again.
 *
 * @long The cleanup hook may be called either from @link "heapapi.h"
 * "ose_heap_remove_debug_hooks" or when the referred heap is removed.
  *
 * @input heap A reference to the heap for which the debug module was installed.
 *
 */
typedef void (*DEBUG_CLEANUP_FUNC)(OSHEAP heap);


/**
 * @type HeapDebugHooks
 * @alias _HeapDebugHooks
 *
 * @brief Struct containing all the heap debug hooks.
 *
 * @field debug_alloc    See @link "" "DEBUG_ALLOC_FUNC".
 * @field debug_free     See @link "" "DEBUG_FREE_FUNC".
 * @field debug_fileline See @link "" "DEBUG_FILELINE_FUNC".
 * @field debug_cleanup  See @link "" "DEBUG_CLEANUP_FUNC".
 *
 */
typedef struct _HeapDebugHooks
{
   DEBUG_ALLOC_FUNC debug_alloc;
   DEBUG_FREE_FUNC debug_free;
   DEBUG_FILELINE_FUNC debug_fileline;
   DEBUG_CLEANUP_FUNC debug_cleanup;
}  HeapDebugHooks;


/************************************************************************
 *
 * DEBUG MODULE SPECIFIC declarations
 *
 * The declarations in this section are only needed for this implementation
 * of the heap debug module.
 *
 ***********************************************************************/

/*
 * Heap debug module control flags.
 *
 */
/**
 * @macro HDM_WIPE_ON_FREE
 *
 * @brief Use wipe on free, heap debug module control flag.
*/
#define HDM_WIPE_ON_FREE        1UL

/**
 * @macro HDM_DEFERRED_FREE
 *
 * @brief Use deferred free, heap debug module control flag.
*/
#define HDM_DEFERRED_FREE       2UL

/**
 * @macro HDM_EXTENDED_ANNOTATION
 *
 * @brief Use extended annotation, heap debug module control flag.
*/
#define HDM_EXTENDED_ANNOTATION 4UL



/* Various defines. */
#define WIPE_PATTERN   0xFF

typedef struct
{
   U32 ra1;   /* Return address to function calling the heap,
               * e.g. calling malloc() or heap_alloc_private().  */
   U32 ra2;   /* Return address to caller of the function in ra1.
               * This is the caller of new() in a C++ system.     */
   U32 pid;   /* Process id of allocating function. */
   DefAnnotation def_annot; /* Must be last in annotation area. */
} ExtAnnotation;

/**
 * @type heap_debug_alloc_hook
 *
 * @brief Implementation of a heap allocation hook.
 *
 * @input heap  A reference to the heap from which the allocation was made.
 * @input ptr   A pointer to the user part of the annotated buffer.
 * @input size  Original size requested by user in allocation call.
 * @input annsz Size of annotation area immediately preceding the user data.
 * @input owner Owner of the buffer.
 * @input file  Name of file from which allocation was performed.
 * @input line  Line number of allocation call.
 *
 * @seealso DEBUG_ALLOC_FUNC
 */
void
heap_debug_alloc_hook(OSHEAP heap, void *ptr, size_t size, size_t annsz,
                      PROCESS owner, const char *file, U32 line);

/**
 * @type heap_debug_free_hook
 *
 * @brief Implementation of a heap free hook.
 *
 * @input heap  A reference to the heap to which the freeing is made.
 * @input ptr   A pointer to the user part of the annotated buffer.
 * @input annsz Size of annotation area immediately preceding the user data.
 * @input owner Owner of the buffer.
 *
 * @seealso DEBUG_FREE_FUNC
 */
int
heap_debug_free_hook(OSHEAP heap,  void *ptr, size_t annsz, PROCESS owner);

/**
 * @type heap_debug_fileline_hook
 *
 * @brief Implementation of a heap fileline hook.
 *
 * @input heap     Heap reference.
 * @input hbi      HeapBufferInfoReply struct. If address field is zero, no
 *                 buffer exists and file and line information need to be taken
 *                 from the heap debug pp data.
 * @output hbi     HeapBufferInfoReply struct. The line number should be copied
 *                 to the line field in hbi. The file name should be copied to
 *                 the data array, at an index given by the offset parameter,
 *                 and with a maximum length of HEAP_MAX_FILENAME_LENGTH.
 *                 The offset should also be filled into the file field of hbi
 *                 or, if no file name is available, the file field should be
 *                 set to ~0.
 * @input offset   Index for file name into data field of hbi.
 *
 * @seealso DEBUG_FILELINE_FUNC
 */
void
heap_debug_fileline_hook(OSHEAP heap, struct HeapBufferInfoReply *hbi,
                         U32 offset);


/**
 * @type heap_debug_cleanup_hook
 *
 * @brief Implementation of a heap cleanup hook.
 *
 * @input heap     Heap reference.
 *
 * @seealso DEBUG_CLEANUP_FUNC
 */
void
heap_debug_cleanup_hook(OSHEAP heap);

/**
 * @type heap_debug_module_init
 *
 * @brief  Installation and initialization of the heap debug module.
 *
 * @long This function installs and initializes a heap debug module using the
 *       available heap debug hooks. It implements some different features for
 *       debugging:
 *
 * @long FILE AND LINE. By default the heap debug module adds file and line
 * information in the annotation area of the heap buffers, when such allocation
 * calls are used that include file and line information. This is achieved for
 * any allowed value on the ctrl parameter.
 *
 * @long WIPE ON FREE. The wipe on free functionality wipes each buffer when it
 * is freed such that late (and faulty) accesses to that buffer are more easily
 * discovered.
 *
 * @long DEFERRED_FREE. With this functionality selected, a buffer that is freed
 * by a user process is taken over by the heap debug module and saved for a while
 * in a circular list. The buffer is freed for real to the heap when the circular
 * list has been completely filled once more. At this point, if WIPE ON FREE is
 * selected, the buffer pattern is compared to the wipe pattern. Too late write
 * accesses can be identified in this way and in such case an error is signaled:
 * HEAP_DEBUG_EBUFFER_TOUCHED_AFTER_FREE.
 *
 * @long EXTENDED ANNOTATION. This functionality extends the annotation data
 * with program counter information for the process performing an allocation.
 * Both the program counter for the calling function, and the function calling
 * that function are included to enable efficient debugging of the new()
 * operator in C++. NOTE: This functionality is not implemented today.
 *
 * @input href         Heap reference.
 * @input ctrl         Control parameter for selecting functionality from the
 *                     debug module. (HDM_WIPE_ON_FREE |
 *                     HDM_DEFERRED_FREE | HDM_EXTENDED_ANNOTATION)
 * @input list_length  Length of the circular deferred free list.
 *
 * @return The size of the annotation field added to each buffer. Zero if init
 *         failed or if module was already installed.
 *
 * @error HEAP_DEBUG_EINIT_FAILED Installation failed because either a debug
 *        module was already installed or the heap was not found.
 * @error HEAP_DEBUG_EILLEGAL_PROC_TYPE Installation failed because the function
 *        was not called from a prioritized or background process.
 *
 * @restriction No more than one heap debug module can be installed in a system.
 *              This restriction is to be removed.
 *
 */
size_t
heap_debug_module_init(OSHEAP href, U32 ctrl, size_t list_length);

#ifdef __cplusplus
}
#endif

#endif
