/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2011 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */
/**
 * @public
 * @toc Memory_Management:
 * @file ose_spi/heapapi.h
 *
 * @brief The Heap Manager non-standard API.
 *
 * @long The file heapapi.h contains the definitions of the heap manager API
 *       that are specific for the OSE heap manager and do not follow any
 *       standards specification. The file malloc.h (included from stdlib.h)
 *       defines the StdC API of the heap manager.
 *
 * @long The number of calls may seem large, but several calls are only
 *       variants of the same call. Four major groups are identified:
 *
 * @long 1) "Foundation" calls; prefixed ose_*, that take both a heap
 *          reference and a PID owner as parameter, which makes them flexible.
 *
 * @long 2) Private calls; calls who always work on the current heap, and
 *       sets the owner PID to current_process().
 *
 * @long 3) Shared calls; calls who also work on the current heap but set
 *       the PID to zero (shared mode). NOTE that private buffers can also be
 *       "forced" to be freed via shared free.
 *
 * @long 4) "Owner independent calls" - heap_alloc_owner/heap_free_owner;
 *       calls which use the current heap, but leave it up to the user to set
 *       the owner.
 *
 * @long An interrupt process is not allowed to make heap calls. Such usage
 *       will generate an error, @link "heap_err.h" "OSE_HEAP_EILLEGAL_USER".
 *
 * @long For more information on how to use the OSE heap manager, see
 *       the "OSE Core User's Guide".
 *
 */
#ifndef _HEAPAPI_H
#define _HEAPAPI_H

#include <stddef.h>
#include "ose.h"
#include "osetypes.h"
#include "heap_debug.h"

#ifdef __cplusplus
extern "C" {
#endif


/**
 * @macro HEAP_NAME_LENGTH
 *
 * @brief The default string length, including zero terminator, for all strings
 *        handled by the heap.
 *
 * @long  Currently this value is 32.
 */
#define HEAP_NAME_LENGTH  32


/**
 * @function heap_alloc_private
 *
 * @brief Allocate a private buffer.
 *
 * @long The function heap_alloc_private() allocates and returns a pointer to a
 *       private buffer suitably aligned for any use. heap_alloc_private()
 *       never returns NULL, instead the error handler will be invoked if
 *       allocation fails. The buffer is owned by the calling process and will
 *       be freed when that process terminates. File and line will be stored
 *       together with the buffer and can be obtained later.
 *
 * @input size The size of the requested buffer.
 *
 * @input file Pointer to the source file name of the allocation (__FILE__) or
 *             NULL (can actually point to any constant string).
 *
 * @input line Line in the source file of the allocation (__LINE__) or 0.
 *
 * @return A pointer to the allocated buffer.
 *
 * @error OSE_HEAP_EFREED_BUFFER_OVERWRITTEN For details, see the error code
 *                                           section.
 *
 * @error OSE_HEAP_ESIZE_TOO_LARGE
 *
 * @error OSE_HEAP_EHEAP_EXHAUSTED
 *
 * @error OSE_HEAP_EILLEGAL_EXTENSION
 *
 * @seealso default_restrictions_calling_functions
 */
void *heap_alloc_private(size_t size, const char *file, unsigned line);


/**
 * @function heap_realloc_private
 *
 * @brief Re-allocate a private buffer with a different size.
 *
 * @long The function heap_realloc_private() reallocates the specified buffer
 *       (owned by the calling process) to obtain a new buffer with the
 *       requested size. It can also be used to allocate a buffer (when ptr
 *       is NULL) or free a buffer (when size is 0). The buffer contents will
 *       be unchanged up to the lesser of the new and old sizes.
 *       heap_realloc_private() never returns NULL, instead the error handler
 *       will be invoked if reallocation fails. The buffer is owned by the
 *       calling process and will be freed when that process terminates. File
 *       and line will be stored together with the buffer and can be obtained
 *       later.
 *
 * @input ptr  Pointer to a buffer (owned by the calling process) or NULL.
 *
 * @input size The new size of the buffer (may be 0).
 *
 * @input file Pointer to the source file name of the allocation (__FILE__) or
 *             NULL (can actually point to any constant string).
 *
 * @input line Line in the source file of the allocation (__LINE__) or 0.
 *
 * @return A pointer to the re-allocated buffer.
 *
 * @error OSE_HEAP_EFREED_BUFFER_OVERWRITTEN For details, see the error code
 *                                           section.
 *
 * @error OSE_HEAP_ESIZE_TOO_LARGE
 *
 * @error OSE_HEAP_EPTR_INVALID
 *
 * @error OSE_HEAP_EBUFFER_FREE
 *
 * @error OSE_HEAP_EPROCESS_NOT_OWNER
 *
 * @error OSE_HEAP_EENDMARK_OVERWRITTEN
 *
 * @error OSE_HEAP_EHEAP_EXHAUSTED
 *
 * @error OSE_HEAP_EILLEGAL_EXTENSION
 *
 * @seealso default_restrictions_calling_functions
 */
void *heap_realloc_private(void *ptr, size_t size,
                           const char *file, unsigned line);


/**
 * @function heap_free_private
 *
 * @brief Free a private buffer.
 *
 * @long The function heap_free_private() frees a buffer owned by the calling
 *       process. The buffer will be merged with other nearby free buffers and
 *       made available for future allocations.
 *
 * @input ptr  Pointer to a buffer (owned by the calling process) or NULL.
 *
 * @error OSE_HEAP_EPTR_INVALID For details, see the error code
 *                              section.
 *
 * @error OSE_HEAP_EBUFFER_FREE
 *
 * @error OSE_HEAP_EPROCESS_NOT_OWNER
 *
 * @error OSE_HEAP_EENDMARK_OVERWRITTEN
 *
 * @seealso default_restrictions_calling_functions
 */
void heap_free_private(void *ptr);


/**
 *
 * @function heap_alloc_shared
 *
 * @brief Allocate a shared buffer.
 *
 * @long The function heap_alloc_shared() allocates and returns a pointer to a
 *       shared buffer suitably aligned for any use. heap_alloc_shared() never
 *       returns NULL, instead the error handler will be invoked if allocation
 *       fails. The buffer is not owned by any specific process. File and line
 *       information will be stored together with the buffer and can be
 *       obtained later.
 *
 * @input size The size of the requested buffer.
 *
 * @input file Pointer to the source file name of the allocation (__FILE__) or
 *             NULL (can actually point to any constant string).
 *
 * @input line Line in the source file of the allocation (__LINE__) or 0.
 *
 * @return A pointer to the allocated buffer.
 *
 * @error OSE_HEAP_EFREED_BUFFER_OVERWRITTEN For details, see the error code
 *                                           section.
 *
 * @error OSE_HEAP_ESIZE_TOO_LARGE
 *
 * @error OSE_HEAP_EHEAP_EXHAUSTED
 *
 * @error OSE_HEAP_EILLEGAL_EXTENSION
 *
 * @seealso default_restrictions_calling_functions
 */
void *heap_alloc_shared(size_t size, const char *file, unsigned line);


/**
 *
 * @function heap_realloc_shared
 *
 * @brief Re-allocate a shared buffer with a different size.
 *
 * @long The function heap_realloc_shared() reallocates the specified shared
 *       buffer to obtain a new buffer with the requested size. It can also be
 *       used to allocate a buffer (when ptr is NULL) or free a buffer (when
 *       size is 0). The buffer contents will be unchanged up to the lesser of
 *       the new and old sizes. heap_realloc_shared() never returns NULL,
 *       instead the error handler will be invoked if reallocation fails. The
 *       buffer is not owned by any specific process. File and line will be
 *       stored together with the buffer and can be obtained later.
 *
 * @input ptr  Pointer to a shared buffer or NULL.
 *
 * @input size The new size of the buffer (may be 0).
 *
 * @input file Pointer to the source file name of the allocation (__FILE__) or
 *             NULL (can actually point to any constant string).
 *
 * @input line Line in the source file of the allocation (__LINE__) or 0.
 *
 * @return A pointer to the re-allocated buffer.
 *
 * @error OSE_HEAP_EFREED_BUFFER_OVERWRITTEN For details, see the error code
 *                                           section.
 *
 * @error OSE_HEAP_ESIZE_TOO_LARGE
 *
 * @error OSE_HEAP_EPTR_INVALID
 *
 * @error OSE_HEAP_EBUFFER_FREE
 *
 * @error OSE_HEAP_EENDMARK_OVERWRITTEN
 *
 * @error OSE_HEAP_EHEAP_EXHAUSTED
 *
 * @error OSE_HEAP_EILLEGAL_EXTENSION
 *
 * @seealso default_restrictions_calling_functions
 */
void *heap_realloc_shared(void *ptr, size_t size,
                          const char *file, unsigned line);


/**
 *
 * @function heap_free_shared
 *
 * @brief Free a shared (or private) buffer.
 *
 * @long The function heap_free_shared() frees a buffer, private or shared.
 *       The buffer will be merged with other nearby free buffers and made
 *       available for future allocations.
 *
 * @input ptr  Pointer to a buffer or NULL.
 *
 * @error OSE_HEAP_EPTR_INVALID For details, see the error code
 *                              section.
 *
 * @error OSE_HEAP_EBUFFER_FREE
 *
 * @error OSE_HEAP_EENDMARK_OVERWRITTEN
 *
 * @seealso default_restrictions_calling_functions
 */
void heap_free_shared(void *ptr);


/**
 * @function heap_alloc_owner
 *
 * @brief Allocate a buffer with the specified process as owner.
 *
 * @long The function heap_alloc_owner() allocates and returns a pointer to a
 *       buffer suitably aligned for any use. heap_alloc_owner() never returns
 *       NULL, instead the error handler will be invoked if allocation fails.
 *       The buffer will be owned by the specific process. File and line
 *       information will be stored together with the buffer and can be
 *       obtained later.
 *
 * @input size The size of the requested buffer.
 *
 * @input owner The PID of the owner of the buffer.
 *
 * @input file Pointer to the source file name of the allocation (__FILE__) or
 *             NULL (can actually point to any constant string).
 *
 * @input line Line in the source file of the allocation (__LINE__) or 0.
 *
 * @return A pointer to the allocated buffer.
 *
 * @error OSE_HEAP_EFREED_BUFFER_OVERWRITTEN For details, see the error code
 *                                           section.
 *
 * @error OSE_HEAP_ESIZE_TOO_LARGE
 *
 * @error OSE_HEAP_EHEAP_EXHAUSTED
 *
 * @error OSE_HEAP_EILLEGAL_EXTENSION
 *
 * @restriction The allocated buffer will not be reclaimed upon termination of
 *              the owner process unless the owner process is registered by the
 *              heap. A process gets registered by the heap when it first
 *              accesses the heap from its own processing scope, e.g. by
 *              performing @link "" "free"(NULL).
 *
 * @seealso default_restrictions_calling_functions
 */
void *heap_alloc_owner(size_t size,
                       PROCESS owner,
                       const char *file,
                       unsigned line);


/**
 * @function heap_free_owner
 *
 * @brief Free a buffer owned by the specific process.
 *
 * @long The function heap_free_owner() frees a buffer owned by the specified
 *       process. The buffer will be merged with other nearby free buffers and
 *       made available for future allocations.
 *
 * @input ptr  Pointer to a buffer (owned by the specified process) or NULL.
 *
 * @input owner The PID of the owner of the buffer. If set to zero, the buffer
 *              will be freed regardless if it was allocated in private mode.
 *
 * @error OSE_HEAP_EPTR_INVALID For details, see the error code section.
 *                              Can only occur if owner != zero.
 *
 * @error OSE_HEAP_EBUFFER_FREE
 *
 * @error OSE_HEAP_EPROCESS_NOT_OWNER
 *
 * @error OSE_HEAP_EENDMARK_OVERWRITTEN
 *
 * @seealso default_restrictions_calling_functions
 */
void heap_free_owner(void *ptr, PROCESS owner);


/**
 * @function heap_alloc_unsafe
 *
 * @brief Try to allocate a buffer with option to specify owner.
 *
 * @long The function heap_alloc_unsafe() attempts to allocate a buffer
 *       suitably aligned for any use and returns a pointer to the buffer
 *       if the allocation is successful. The function returns NULL if
 *       the allocation fails due to a heap exhausted condition. The error
 *       handler will be invoked and the function will not return if the
 *       function fails due to any of the error conditions listed below.
 *
 * @long The buffer will have shared ownership if the heap mode of the
 *       current process is shared or if the 'owner' parameter is 0.
 *       Otherwise the buffer will be owned by the process specified by
 *       the 'owner' parameter.
 *
 * @long File and line information will be stored together with the buffer
 *       and can be obtained later.
 *
 * @input size The size of the requested buffer.
 *
 * @input owner The requested owner PID of the buffer as described above.
 *
 * @input file Pointer to the source file name of the allocation (__FILE__) or
 *             NULL (can actually point to any constant string).
 *
 * @input line Line in the source file of the allocation (__LINE__) or 0.
 *
 * @return A pointer to the allocated buffer or NULL if allocation fails.
 *
 * @error OSE_HEAP_EFREED_BUFFER_OVERWRITTEN For details, see the error code
 *                                           section.
 *
 * @error OSE_HEAP_ESIZE_TOO_LARGE
 *
 * @error OSE_HEAP_EILLEGAL_EXTENSION
 *
 * @restriction The allocated buffer will not be reclaimed upon termination of
 *              the owner process unless the owner process is registered by the
 *              heap. A process gets registered by the heap when it first
 *              accesses the heap from its own processing scope, e.g. by
 *              performing @link "" "free"(NULL).
 *
 * @seealso default_restrictions_calling_functions
 */
void *heap_alloc_unsafe(size_t size, PROCESS owner,
                        const char *file, unsigned line);


/**
 * @function heap_alloc_private_unsafe
 *
 * @brief Try to allocate a private buffer. 
 *
 * @long This function is similar to @link "" "malloc", but overrides the
 *       heap mode to always return a private buffer owned by the current
 *       process.
 *
 * @long The function heap_alloc_unsafe() attempts to allocate a buffer
 *       suitably aligned for any use and returns a pointer to the buffer
 *       if the allocation is successful. The function returns NULL if
 *       the allocation fails due to a heap exhausted condition. The error
 *       handler will be invoked and the function will not return if the
 *       function fails due to any of the error conditions listed below.
 *
 * @long The buffer will be owned by the current process.
 *
 * @long File and line information will be stored together with the buffer
 *       and can be obtained later.
 *
 * @input size The size of the requested buffer.
 *
 * @input file Pointer to the source file name of the allocation (__FILE__) or
 *             NULL (can actually point to any constant string).
 *
 * @input line Line in the source file of the allocation (__LINE__) or 0.
 *
 * @return A pointer to the allocated buffer or NULL if allocation fails.
 *
 * @error OSE_HEAP_EFREED_BUFFER_OVERWRITTEN For details, see the error code
 *                                           section.
 *
 * @error OSE_HEAP_ESIZE_TOO_LARGE
 *
 * @error OSE_HEAP_EILLEGAL_EXTENSION
 *
 * @seealso default_restrictions_calling_functions
 */
void *heap_alloc_private_unsafe(size_t size, const char *file, unsigned line);


/**
 * @function heap_alloc_early_null
 *
 * @deprecated This method is deprecated from OSE 5.5 and will not be supported
 *             in future releases of OSE. The method remains functional but
 *             ignores the watermark level. It may therefore exhaust the heap.
 *
 * @brief Same as @link "" "heap_alloc_unsafe" with the addition of early
 *        null check.
 *
 * @long The function heap_alloc_early_null() can be used in conjunction with
 *       the HEAP_EARLY_NULL_LEVEL configuration in ose_heap.h.
 *       If this variable set to a threshold percentage value,
 *       this function will return NULL if the configured level is reached.
 *
 * @long If multiple levels are configured, heap_alloc_early_null always chooses
 *       the first configured level. To allocate according to a different
 *       configured level, use ose_heap_realloc_watermark instead.
 *
 * @input size The size of the requested buffer.
 *
 * @input owner The requested owner PID of the buffer as described in @link ""
 *              "heap_alloc_unsafe".
 *
 * @input file Pointer to the source file name of the allocation (__FILE__) or
 *             NULL (can actually point to any constant string).
 *
 * @input line Line in the source file of the allocation (__LINE__) or 0.
 *
 * @seealso ose_heap_realloc_watermark
 * @seealso HEAP_EARLY_NULL_LEVEL
 *
 * @restriction The allocated buffer will not be reclaimed upon termination of
 *              the owner process unless the owner process is registered by the
 *              heap. A process gets registered by the heap when it first
 *              accesses the heap from its own processing scope, e.g. by
 *              performing @link "" "free"(NULL).
 *
 * @seealso default_restrictions_calling_functions
 */
void *heap_alloc_early_null(size_t size,
                            PROCESS owner,
                            const char *file,
                            unsigned line);


/**
 * @function heap_buffer_size
 *
 * @brief Return the (requested) size of a buffer.
 *
 * @long The function heap_buffer_size() returns the size specified when the
 *       buffer was allocated (the requested size).
 *
 * @input ptr  Pointer to a buffer.
 *
 * @return The requested size of the buffer.
 *
 * @error OSE_HEAP_EPTR_INVALID For details, see the error code section.
 *
 * @error OSE_HEAP_EBUFFER_FREE
 *
 * @error OSE_HEAP_EPROCESS_NOT_OWNER
 *
 * @error OSE_HEAP_EENDMARK_OVERWRITTEN
 *
 * @restriction The process must own the buffer if the program is in private
 *              mode.
 *
 * @seealso default_restrictions_calling_functions
 */
size_t heap_buffer_size(void *ptr);

/**
 * @function heap_buffer_actual_size
 *
 * @brief Returns the actual size of a buffer.
 *
 * @long  The actual size is the number of bytes occupied by the buffer that
 *        can not be reused until the buffer is freed. The size includes the
 *        requested size, the buffer endmark, administrative overhead and
 *        internal fragmentation.
 *
 * @input ptr  Pointer to a buffer.
 *
 * @return The actual size of the buffer.
 *
 * @error OSE_HEAP_EPTR_INVALID For details, see the error code section.
 *
 * @error OSE_HEAP_EBUFFER_FREE
 *
 * @error OSE_HEAP_EPROCESS_NOT_OWNER
 *
 * @error OSE_HEAP_EENDMARK_OVERWRITTEN
 *
 * @restriction The ownership must match.
 *
 * @seealso default_restrictions_calling_functions
 */
size_t heap_buffer_actual_size(void *ptr);


/**
 * @function heap_buffer_resize
 *
 * @brief Resize a buffer if possible.
 *
 * @long  The function heap_buffer_resize() tries to resize the specified buffer
 *        and returns the size of the buffer after the call (the new size or the
 *        old size).
 *
 * @long  The buffer pointer always remains valid after a call to this function.
 *        It will never be deallocated even if the new size is 0.
 *
 * @input ptr  Pointer to the buffer.
 *
 * @input new_size The new requested size of the buffer.
 * 
 * @return The size of the buffer after the possible resize.
 *
 * @error OSE_HEAP_ESIZE_TOO_LARGE For details, see the error code section.
 *
 * @error OSE_HEAP_EPTR_INVALID
 *
 * @error OSE_HEAP_EBUFFER_FREE
 *
 * @error OSE_HEAP_EPROCESS_NOT_OWNER
 *
 * @error OSE_HEAP_EENDMARK_OVERWRITTEN
 *
 * @restriction The ownership must match.
 *
 * @seealso default_restrictions_calling_functions
 */
size_t heap_buffer_resize(void *ptr, size_t new_size);

/**
 * @function ose_heap_ref
 *
 * @brief Returns a reference to the heap for the PID.
 *
 * @long This function returns the heap reference for a given PID.
 *
 * @long The current heap reference returned (if parameter pid=0) is used
 *       internally in all private/shared heap calls above,
 *       that do not take an OSHEAP reference.
 *       This means that it is possible to always achieve the same
 *       behavior as those calls, by first retrieving the reference via
 *       ose_heap_ref and then call the appropriate ose_* heap
 *       call that takes an OSHEAP reference.
 *
 * @long This call can be used in order to get a reference to heaps in
 *       other programs for a process, for both sharing reasons and
 *       statistics reasons.
 *
 * @input pid The PID of a heap user, either a process/phantom or block PID.
 *        If pid=zero, the current_process is assumed.
 *
 * @return The reference to the heap. If the least significant bit is set,
 *         the block/program that uses the heap executes in shared mode.
 *         If zero is returned, no heap was found.
 *
 * @seealso default_restrictions_calling_functions
 */
OSHEAP ose_heap_ref(PROCESS pid);

/**
 * @function ose_heap_malloc
 *
 * @brief A malloc from an explicit heap.
 *
 * @long The function ose_heap_malloc() performs an unsafe malloc from the heap
 *       given as an argument. This heap reference must be created by a
 *       previous ose_create_heap call.
 *       NOTE that the allocation is performed as a shared allocation.
 *
 * @input heap The reference to the heap.
 * @input size The size of the requested buffer.
 * @input file Pointer to the source file name of the allocation (__FILE__) or
 *             NULL (can actually point to any constant string).
 * @input line Line in the source file of the allocation (__LINE__) or 0.
 *
 * @error OSE_HEAP_EHEAP_EXHAUSTED For details, see the error code section.
 *
 * @error OSE_HEAP_ESIZE_TOO_LARGE
 *
 * @error OSE_HEAP_EILLEGAL_EXTENSION
 *
 * @restriction The buffer will always be shared, i.e. the allocator needs to
 *              keep track of it in order not to get a memory leak.
 *
 * @seealso malloc
 * @seealso ose_create_heap
 * @seealso default_restrictions_calling_functions
 */
void *ose_heap_malloc(OSHEAP heap,
                      size_t size,
                      const char *file,
                      unsigned line);
/**
 * @function ose_heap_free
 *
 * @brief Free buffer from explicit heap
 *
 * @long The function ose_heap_free() frees the specified buffer from the
 *       heap indicated by an argument.
 *
 * @input heap The reference to the heap.
 *
 * @input ptr  Pointer to a buffer or NULL.
 *
 * @error OSE_HEAP_EPTR_INVALID For details, see the error code section.
 *
 * @error OSE_HEAP_EBUFFER_FREE
 *
 * @error OSE_HEAP_EENDMARK_OVERWRITTEN
 *
 * @seealso default_restrictions_calling_functions
 */
void ose_heap_free(OSHEAP heap, void *ptr);


/**
 * @function ose_heap_alloc_tentative
 *
 * @brief Reclaims a tentative buffer from an explicit heap.
 *
 * @long  The function ose_heap_alloc_tentative() reclaims a tentative buffer
 *        previously freed with ose_heap_free_tentative(), unless the handle
 *        used to identify the buffer has been invalidated. The buffer is
 *        returned with its data retained. NULL is returned if the buffer has
 *        been garbage collected.
 *
 * @long  Buffer ownership is retained by a tentative buffer and it is not
 *        legal for a process other than the owner to reclaim it.
 *
 * @input heapref The reference to the heap, or 0 for current heap.
 * @input handle  A handle identifying the tentative buffer.
 * @input file    Pointer to the source file name of the allocation (__FILE__)
 *               or NULL (can actually point to any constant string).
 * @input line   Line in the source file of the allocation (__LINE__) or 0.
 *
 * @return A pointer the buffer.
 *
 * @seealso ose_heap_free_tentative
 * @seealso default_restrictions_calling_functions
 */
void *ose_heap_alloc_tentative(OSHEAP heapref,
                               U32 handle,
                               const char *file,
                               unsigned line);

/**
 * @function ose_heap_free_tentative
 *
 * @brief Frees a buffer tentatively on an explicit heap.
 *
 * @long  The function ose_heap_free_tentative() frees a buffer
 *        tentatively. The caller is assigned a handle to identify the
 *        buffer if it is reclaimed using ose_heap_alloc_tentative(). The
 *        handle can be invalidated and the tentative buffer lost if the
 *        heap or the system is low on memory resources.
 *
 * @input heapref   The reference to the heap, or 0 for current heap.
 * @input ptr       Pointer to a buffer or NULL.
 * @input watermark Deprecated, must be 0.
 * @input file      Pointer to the source file name of the allocation (__FILE__)
 *                  or NULL (can actually point to any constant string).
 * @input line      Line in the source file of the allocation (__LINE__) or 0.
 *
 * @return A handle used to identify the buffer, or
 *         @link "ose_heap.h" "HEAP_INVALID_TENTATIVE_HANDLE" if
 *         the buffer was freed permanently.
 *
 * @seealso ose_heap_alloc_tentative
 * @seealso HEAP_INVALID_TENTATIVE_HANDLE
 * @seealso default_restrictions_calling_functions
 */
U32 ose_heap_free_tentative(OSHEAP heapref,
                            void *ptr,
                            U32 watermark,
                            const char *file,
                            unsigned line);

/**
 * @function ose_heap_realloc
 *
 * @brief Re-allocate a buffer with a different size on a given heap.
 *
 * @long The function ose_heap_realloc() performs realloc on an explicit
 *       heap.
 * @long Equal to heap_realloc_shared except for the extra heap argument
 *       and the optional ownership.
 *
 * @input heap  The reference to the heap.
 *
 * @input ptr  Pointer to a buffer (owned by the calling process) or NULL.
 *
 * @input size The new size of the buffer (may be 0).
 *
 * @input owner The PID of the owner of the buffer or 0 for a shared buffer.
 *
 * @input file Pointer to the source file name of the allocation (__FILE__) or
 *             NULL (can actually point to any constant string).
 *
 * @input line Line in the source file of the allocation (__LINE__) or 0.
 *
 * @seealso heap_realloc_shared
 * @seealso default_restrictions_calling_functions
 */
void * ose_heap_realloc(OSHEAP heap,
                        void *ptr,
                        size_t size,
                        PROCESS owner,
                        const char *file,
                        unsigned line);


/**
 * @function ose_heap_realloc_watermark
 *
 * @deprecated This method is deprecated from OSE 5.5 and will not be supported
 *             in future releases of OSE. The method remains functional but
 *             ignores the watermark level. It may therefore exhaust the heap.
 *
 * @restriction NOT YET AVAILABLE in standard OSE.
 * @break Must not cause a heap extension or large buffer allocations
 *        when used in an interrupt process. In that case an error is called.
 * @break The allocated buffer will not be reclaimed upon termination of the
 *        owner process unless the owner process is registered by the heap.
 *        A process gets registered by the heap when it first accesses the
 *        heap from its own processing scope, e.g. by performing 
 *        @link "" "free"(NULL).
 *
 * @brief Re-allocate a buffer with a different size on a given heap.
 *
 * @long  The function ose_heap_realloc_watermark() performs realloc on an
 *        explicit heap. It is equal to ose_heap_realloc except for the extra
 *        watermark level argument.
 *
 * @long  Watermarks are configured as a percentage of the heap's maximum size
 *        by setting the environment variable HEAP_EARLY_NULL_LEVEL before
 *        creating the heap. They are used to prevent the heap from getting
 *        exhausted. If the operation would result in a heap usage exceeding
 *        the requested watermark level, NULL is returned.
 *
 * @long  A watermark level is chosen by passing its index to the
 *        function. Index 0 is the default level and corresponds to 100% usage.
 *        User configured levels are indexed from 1 to X, where is X is the
 *        number of configured watermarks. They are indexed in order of
 *        appearance. NULL is always returned if an illegal watermark
 *        level index is chosen.
 *
 * @input heap      The reference to the heap, or 0 for current heap.
 * @input ptr       Pointer to a buffer, or NULL to allocate a new buffer.
 * @input size      The new size of the buffer, or 0 to free the buffer.
 * @input watermark The watermark priority level, or 0 for default level.
 * @input owner     The PID of the owner of the buffer or 0 for a shared buffer.
 * @input file      Pointer to the source file name of the allocation (__FILE__)
 *                  or NULL (can actually point to any constant string).
 * @input line      Line in the source file of the allocation (__LINE__) or 0.
 *
 * @seealso heap_alloc_early_null
 * @seealso ose_heap_realloc
 * @seealso default_restrictions_calling_functions
 *
 */
void *ose_heap_realloc_watermark(OSHEAP heap,
                                 void *ptr,
                                 size_t size,
                                 U32 watermark,
                                 PROCESS owner,
                                 const char *file,
                                 unsigned line);

/**
 * @function ose_heap_buffer_size
 *
 * @brief Return the (requested) size of a buffer in an explicit heap.
 *
 * @long The function ose_heap_buffer_size() returns the size specified when the
 *       buffer was allocated (the requested size).
 * @long Equal to heap_buffer_size except for the extra heap argument and
 *       the optional ownership.
 *
 * @input heap  The reference to the heap.
 *
 * @input ptr  Pointer to a buffer.
 *
 * @input owner The PID of the owner of the buffer or 0 for a shared buffer.
 *
 * @seealso heap_buffer_size
 * @seealso default_restrictions_calling_functions
 */
size_t ose_heap_buffer_size(OSHEAP heap, void *ptr, PROCESS owner);

/**
 * @function ose_heap_buffer_actual_size
 *
 * @brief Returns the actual size of a buffer in an explicit heap.
 *
 * @long  The actual size is the number of bytes occupied by the buffer that
 *        can not be reused until the buffer is freed. The size includes the
 *        requested size, the buffer endmark, administrative overhead and
 *        internal fragmentation.
 *
 * @long  Equal to heap_buffer_actual_size except for the extra heap argument
 *        and the optional ownership.
 *
 * @input heap  The reference to the heap.
 *
 * @input ptr  Pointer to a buffer.
 *
 * @input owner The PID of the owner of the buffer or 0 for a shared buffer.
 *
 * @seealso heap_buffer_actual_size
 * @seealso default_restrictions_calling_functions
 *
 */
size_t ose_heap_buffer_actual_size(OSHEAP heap, void *ptr, PROCESS owner);


/**
 * @function ose_heap_buffer_resize
 *
 * @brief Resize a buffer if possible in an explicit heap.
 *
 * @long The function ose_heap_buffer_resize() tries to resize the specified
 *       buffer and returns the size of the buffer after change (new_size or
 *       the old size).
 *
 * @long The buffer pointer always remains valid after a call to this function.
 *       It will never be deallocated even if the new size is 0.
 *
 * @long Equal to heap_buffer_resize except for the extra heap argument and
 *       the optional ownership.
 *
 * @input heap  The reference to the heap.
 *
 * @input ptr  Pointer to a buffer.
 *
 * @input new_size The new requested size of the buffer.
 *
 * @input owner The PID of the owner of the buffer or 0 for a shared buffer.
 *
 * @seealso heap_buffer_resize
 * @seealso default_restrictions_calling_functions
 *
 */
size_t ose_heap_buffer_resize(OSHEAP heap,
                              void *ptr,
                              size_t new_size,
                              PROCESS owner);

/**
 * @function ose_extend_heap
 *
 * @brief Extend a specific (shared) heap.
 *
 * @deprecated This method is deprecated and will not be supported in future
 *             releases of OSE.
 *
 * @long NOTE: This function should normally not need to be used since
 *       heap memory is retrieved from MM by the heap module itself.
 *
 * @long The function ose_extend_heap() adds the specified memory area to the
 *       heap given as an argument. The memory will be fragmented to match the
 *       buffer sizes supported by installed algorithms.
 *
 * @input heap The reference to the heap.
 *
 * @input frag_addr Base address of memory area.
 *
 * @input frag_size Size of memory area (in bytes).
 *
 * @restriction ose_extend_heap() can be called directly from the application
 *              code (prioritized and background processes) or from an error
 *              handler.
 *
 * @seealso default_restrictions_calling_functions
 */
void ose_extend_heap(OSHEAP heap, void *frag_addr, size_t frag_size);

/**
 * @function heap_extend_heap
 *
 * @brief Extend the current heap.
 *
 * @deprecated This method is deprecated and will not be supported in future
 *             releases of OSE.
 *
 * @long NOTE: This function should not need to be used since
 *       heap memory is retrieved from MM by the heap module itself.
 *
 * @long The function heap_extend_heap() adds the specified memory area to the
 *       heap associated with the calling process. The memory will be
 *       fragmented to match the buffer sizes supported by installed
 *       algorithms.
 *
 * @input frag_addr Base address of memory area.
 *
 * @input frag_size Size of memory area (in bytes).
 *
 * @restriction heap_extend_heap() can be called directly from the application
 *              code (prioritized and background processes) or from an error
 *              handler.
 *
 * @seealso default_restrictions_calling_functions
 */
void heap_extend_heap(void *frag_addr, size_t frag_size);

/**
 * @function heap_set_max_size
 *
 * @brief Changes maximum size of the given heap.
 *
 * @long  This function sets a new maximum size for a heap. The value can
 *        be adjusted both upwards and downwards.
 *
 * @input heap  The reference to the heap. Compare @link "" "ose_heap_ref".
 *
 * @input max_size New max size of heap.
 *
 * @seealso default_restrictions_calling_functions
 */
size_t heap_set_max_size(OSHEAP heap, size_t max_size);


/**
 * @function heap_set_threshold
 *
 * @brief  Changes the large buffer threshold size of the given heap.
 *
 * @long   This function sets a new large buffer threshold for a heap. The
 *         value can be adjusted both upwards and downwards.
 *
 * @long   The maximum value allowed is ~8MB when the Fibonacci buddy
 *         algorithm is used and ~16MB when the TLSF algorithm is used.
 *         The threshold can also be set to 0 which effectively disables all
 *         small buffer algorithms (Fibonacci buddy, TLSF).
 *
 * @long   NOTE: If the requested threshold size is too large, the value
 *         will be adjusted down and set to the maximum allowed threshold.
 *
 * @input  heap      The reference to the heap. Compare @link "" "ose_heap_ref".
 * @input  threshold New large buffer threshold size.
 *
 * @return The threshold size in effect after the call.
 *
 * @seealso default_restrictions_calling_functions
 */
size_t heap_set_threshold(OSHEAP heap, size_t threshold);


/**
 * @function heap_set_ext_size
 *
 * @brief  Changes the minimum size of heap memory extensions.
 *
 * @long   This function sets a new minimum size for a heap's memory extension
 *         chunks. The value can be adjusted both upwards and downwards.
 *
 * @input  heap      The reference to the heap. Compare @link "" "ose_heap_ref".
 * @input  ext_size  New extension size of the heap.
 *
 * @return The extension size in effect after the call.
 *
 * @seealso default_restrictions_calling_functions
 */
size_t heap_set_ext_size(OSHEAP heap, size_t ext_size);


/*
 * EXPERIMENTAL HEAP CALLS
 *
 * These calls can be used in order to create an owner session different to the
 * default ownership given in ordinary calls, which is the executing process.
 * The alternative is to represent ownership by a "proxy" phantom, whose name
 * can then be descriptive for any higher layer representation. Example: when
 * listing the owners of the heap, the phantom pid name may be such that it
 * describes an application-level session not necessarily tied to OSE.
 *
 * Scenario:
 * 1) The client creates a phantom process that represents the owner session.
 * 2) The client queries for the current heap (Use "ose_heap_ref" above)
 * 3) The client registers the phantom pid as a "supervised" owner on the heap
 *    reference
 *    received in 2, OR another heap reference retrieved magically from
 *    elsewhere.
 *    Use the call "ose_register_owner" below, and pass the wanted heap and the
 *    phantom pid as parameters.
 *    NOTE: MMU protection may trap you if not set up properly.
 * 4) The client uses the heap calls above that takes a OSHEAP ref and an owner
 *    PID, where the owner is the phantom PID.
 * 5) When the client decides the session is finished, it kills the phantom PID.
 * 6) The heap daemon will reclaim the heap buffers owned by the phantom PID
 *    just as if an ordinary process was killed in private mode.
 */

void ose_register_owner(OSHEAP heap, PROCESS owner);

/**
 * @function ose_heap_buffer_next_size
 *
 * @brief Returns the next supported heap buffer size.
 *
 * @long  The size returned can be used, unadjusted, in any heap allocation
 *        method to obtain a buffer with minimal internal fragmentation.
 *
 * @long  All supported heap buffer sizes may be obtained by calling the
 *        function iteratively with the previously returned value + 1. Note
 *        that the iteration may take a very long time.
 *
 * @input heap The reference to the heap.
 *
 * @input size The start to search for nearest suitable larger size.
 *
 * @return The next suitable buffer size to be used in further allocations.
 *
 * @error OSE_HEAP_EPTR_INVALID For details, see the error code section.
 *
 * @error OSE_HEAP_EBUFFER_FREE
 *
 * @error OSE_HEAP_EPROCESS_NOT_OWNER
 *
 * @error OSE_HEAP_EENDMARK_OVERWRITTEN
 *
 * @seealso default_restrictions_calling_functions
 */
size_t ose_heap_buffer_next_size(OSHEAP heap, size_t size);


/**
 * @function ose_heap_set_debug_hooks
 *
 * @brief Sets the heap debug hooks
 *
 * @long This function installs debug hooks in the heap. These hooks can be used
 * to handle data that is annotated to buffers allocated from the heap, to for
 * example log file and line information of the allocation calls.
 *
 * @long If the requested annotation size is not a multiple of 8, it is adjusted
 *       to the next larger 8-multiple to ensure proper alignment of all
 *       buffers. Further, if debug hooks have already been set and removed
 *       once for this heap, subsequent calls to this function will result in
 *       the same annotation size regardless of the annsz parameter used.
 *
 * @input heap   A reference to the heap, that will get hooks installed.
 *
 * @input annsz  Size of annotation data that is to be added to
 *               each heap buffer.
 * @input hooks  Pointer to a structure containing all debug hooks.
 *               See @link "heap_debug.h" "HeapDebugHooks".
 *               If NULL is passed for the debug_cleanup hook then no
 *               cleanup will be performed before removing the heap
 *               debug module or the heap.
 *
 * @return The resulting size of the annotation field, after alignment
 * adjustment. If debug hooks were earlier installed (and removed) for this heap
 * this value will be the same as in the first installation.
 *
 * @seealso default_restrictions_calling_functions
 */
size_t ose_heap_set_debug_hooks(OSHEAP heap, size_t annsz,
                                HeapDebugHooks *hooks);


/**
 * @function ose_heap_remove_debug_hooks
 *
 * @brief Removes the heap debug hooks
 *
 * @long This function removes the debug hooks that were earlier set with
 * @link "heapapi.h" "ose_heap_set_debug_hooks". This function also calls
 * the hook @link "heap_debug.h" "DEBUG_CLEANUP_FUNC" that allows the heap
 * debug module to cleanup before the module is completely uninstalled.
 *
 * @input heap  A reference to the heap, that will get hooks removed.
 *
 * @seealso default_restrictions_calling_functions
 */
void ose_heap_remove_debug_hooks(OSHEAP heap);

/**
 * @function heap_forced_free
 *
 * @brief Forced free without calling debug hook.
 *
 * @long The function heap_forced_free() frees the specified buffer from the
 *       heap indicated by an argument. This is done without calling a possibly
 *       installed debug hook.
 *
 * @input heap The reference to the heap.
 *
 * @input ptr  The buffer to be freed.
 *
 * @error OSE_HEAP_EPTR_INVALID For details, see the error code section.
 *
 * @error OSE_HEAP_EBUFFER_FREE
 *
 * @error OSE_HEAP_EENDMARK_OVERWRITTEN
 *
 * @seealso default_restrictions_calling_functions
 */
void heap_forced_free(OSHEAP heap,void *ptr);



/**
 * @function heap_alloc_w_tmo
 *
 * @brief   Same as @link "" "heap_alloc_unsafe" with the option to wait for
 *          memory a given time (or forever) if the heap is exhausted.
 *
 * @long The function heap_alloc_w_tmo() will try to allocate, and if it fails,
 *       it will delay for the time of "retry_ms" and then retry.
 *       The function will continue to do such attempts until the time of
 *       "wait_ms" elapsed, when it returns.
 *
 * @input size The size of the requested buffer.
 *
 * @input owner The requested owner PID of the buffer as described in @link ""
 *              "heap_alloc_unsafe".
 *
 * @input wait_ms The total time to wait until allocation is considered
 *        unsuccessful. Set this parameter to (OSTIME)-1 to wait forever.
 *
 * @input retry_ms The time between each allocation retry.
 *
 * @input file Pointer to the source file name of the allocation (__FILE__) or
 *             NULL (can actually point to any constant string).
 *
 * @input line Line in the source file of the allocation (__LINE__) or 0.
 *
 * @restriction The allocated buffer will not be reclaimed upon termination of
 *              the owner process unless the owner process is registered by the
 *              heap. A process gets registered by the heap when it first
 *              accesses the heap from its own processing scope, e.g.
 *              by performing @link "" "free"(NULL).
 *
 * @seealso default_restrictions_calling_functions
 */
void *heap_alloc_w_tmo(size_t size,
		       PROCESS owner,
		       OSTIME wait_ms,
		       OSTIME retry_ms,
		       const char *file,
		       unsigned line);

/**
 * @function heap_alloc_no_extend
 *
 * @brief Same as heap_alloc_unsafe but with no MM extend functionality.
 *
 * @long The function heap_alloc_no_extend() allocates and returns a
 *       pointer to a buffer. However, if the heap is running out of
 *       space, instead of attempting to extend the heap with a new
 *       buffer from MM, NULL will be returned.
 *
 * @input size The size of the requested buffer.
 *
 * @input owner The requested owner PID of the buffer as described in @link ""
 *              "heap_alloc_unsafe".
 *
 * @input file Pointer to the source file name of the allocation (__FILE__) or
 *             NULL (can actually point to any constant string).
 *
 * @input line Line in the source file of the allocation (__LINE__) or 0.
 *
 * @restriction
 *   The allocated buffer will not be reclaimed upon termination of the owner
 *   process unless the owner process is registered by the heap.
 *   A process gets registered by the heap when it first accesses the heap from
 *   its own processing scope, e.g. by performing @link "" "free"(NULL).
 *
 * @seealso default_restrictions_calling_functions
 */
void *heap_alloc_no_extend(size_t size,
			   PROCESS owner,
			   const char *file,
			   unsigned line);


/**
 * @type  HeapServiceInfo
 *
 * @brief Information about a heap service. 
 *
 * @long  The information can be obtained for all active heap services
 *        by using @link "" "ose_heap_get_service".
 *
 * @field  id       A string that uniquely identifies the service.
 *                  Example: "com.enea.heap.tlsf".
 * @field  version  The current service version number.
 *
 * @seealso ose_heap_get_service
 */
struct HeapServiceInfo
{
   char   id[HEAP_NAME_LENGTH];
   U32    version;
};


/**
 * @function ose_heap_get_service
 *
 * @brief   Retrieves information about the next active heap service.
 *
 * @long    The function is called repeatedly with the same argument to
 *          iterate through all active services within a specific heap
 *          instance. Upon the first request, the pointer to the information
 *          structure shall be set to NULL. See example.
 *
 * @input   href   The heap instance for which information about
 *                 services is requested. 
 *
 * @input   svc    Pointer to pointer to structure holding service
 *                 information. The referenced pointer shall be set to
 *                 NULL on first request. See example.
 *                
 * @output  svc    The referenced pointer is updated to point to the
 *                 requested information. 
 *
 * @return  Returns non-zero if new information is available, or zero
 *          if no more services could be found. 
 *
 * @seealso HeapServiceInfo
 *
 * @example
 *
 * const struct HeapServiceInfo *svc = NULL;
 *
 * while (ose_heap_get_service(href, &svc))
 * {
 *    printf("Service: %d - %s\n", svc->version, svc->id);
 * }
 * 
 * @seealso default_restrictions_calling_functions
 */
int ose_heap_get_service(OSHEAP href, const struct HeapServiceInfo **svc);
   
   
#ifdef __cplusplus
}
#endif

#endif

