/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2006 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

/**
 * @public
 * @toc Memory_Management:
 * @file ose_spi/hfmi.h
 *
 * @brief Heap Freeze-mode Interface
 *
 * @long
 * This file declares the Heap Freeze-mode Interface.
 * It may be used to fetch information from the heap in a freeze-mode
 * environment. The purpose is to make OSE Heap information accessible
 * for freeze-mode debug monitors that can not use the signal based interface.
 */
#ifndef _HFMI_H
#define _HFMI_H


#include "stddef.h"
#include "ose.h"
#include "osetypes.h"
#include "heapapi.h"

#ifdef __cplusplus
extern "C" {
#endif

#define HFMI_CURSOR_SIZE (24 / sizeof(U32))

/**
 * @type HfmiStatus
 * @brief enum HfmiStatus defines the set of return values from HFMI
 *   functions. If anything but HFMI_OK is returned from a HMFI function
 *   the contents of all output parameters is undefined
 *   (including cursors).
 *
 * @field HFMI_OK                Operation was successful.
 * @field HFMI_END_OF_LIST       No more objects found.
 * @field HFMI_RESULT_TRUNCATED  A list or string was truncated.
 * @field HFMI_HEAP_CORRUPTED    Unrecoverable error found in the heap. This
 *                               can never happen if the redundant bitmap is
 *                               turned on.
 * @field HFMI_HEAP_UNINITIATED  The heap data structures are not yet
 *                               initiated.
 * @field HFMI_NOT_FOUND         A hfmi_find function could not find what
 *                               was searched for.
 * @field HFMI_CALL_INVALID      Operation preconditions were not met.
 **/
typedef enum HfmiStatus
{
   HFMI_OK                 = 0,
   HFMI_END_OF_LIST        = 2,
   HFMI_RESULT_TRUNCATED   = 3,
   HFMI_HEAP_CORRUPTED     = 4,
   HFMI_HEAP_UNINITIATED   = 6,
   HFMI_NOT_FOUND          = 7,
   HFMI_CALL_INVALID       = 8
} HfmiStatus;

struct HfmiHeapCursor     { U32 cursor[HFMI_CURSOR_SIZE]; };
struct HfmiFragmentCursor { U32 cursor[HFMI_CURSOR_SIZE]; };
struct HfmiBufferCursor   { U32 cursor[64 / sizeof(U32)]; };

/**
 * @type HfmiHeapInfo
 * @brief struct HfmiHeapInfo contains information about a heap.
 *
 * @field href           Heap reference.
 * @field threshold      The threshold for when to allocate
 *                       buffers from the memory manager.
 * @field annotation_size  The size of the annotation area (in bytes).
 * @field endmark_size     The size of an endmark (in bytes).
 * @field wipe_on_free     If non-zero, a buffer is wiped when freed.
 * @field error_on_null    If non-zero, an error is called if malloc fails.
 * @field use_redundant_bitmap If non-zero, all information about the heap
 *                             structures is kept in a copy separate from the
 *                             user data.
 * @field reserved0            NOT_USED
 * @field owner                Heap owner.
 *
 **/
struct HfmiHeapInfo {
   OSHEAP href;
   U32 threshold;
   U32 annotation_size;
   U32 endmark_size;
   unsigned char wipe_on_free;
   unsigned char error_on_null;
   unsigned char use_redundant_bitmap;
   unsigned char reserved0;
   PROCESS owner;
};

/**
 * @type HfmiHeapStatistics
 * @brief struct HfmiHeapStatistics contains statistics about a heap.
 *
 * @field total_req      Total amount of memory requested.
 * @field total_used     Total amount of memory actually used.
 * @field total_used_hi  The maximum value of total_used since last reset.
 * @field total_free     Total amount of free memory.
 * @field largest_free   Size of the largest free buffer.
 * @field number_private Number of private buffers allocated.
 * @field number_shared  Number of shared buffers allocated.
 * @field malloc_fail    Number of failed allocations.
 * @field histo_size     The different sizes.
 *                       Unused buffer indexes are set to zero.
 * @field histo_free     Number of free buffers of each size interval,
 *                       (histo_size[i]...histo_size[i+1].
 * @field histo_used     Number of allocated buffers of each size interval,
 *                       (histo_size[i]...histo_size[i+1].
 **/
struct HfmiHeapStatistics
{
   U32 total_req;
   U32 total_used;
   U32 total_used_hi;
   U32 total_free;
   U32 largest_free;
   U32 number_private;
   U32 number_shared;
   U32 malloc_fail;
   U32 histo_size[32];
   U32 histo_free[32];
   U32 histo_used[32];
};

/**
 * @type HfmiFragmentInfo
 * @brief struct HfmiFragmentInfo contains statistics about a heap fragment.
 *
 * @field address   Address of first byte of heap fragment.
 * @field size      The size of this fragment.
 * @field type      Type of buffer, see HfmiFragmentType.
 * @field status    The status of the buffer (See HfmiFragmentStatus).
 **/
struct HfmiFragmentInfo
{
   U32 address;
   U32 size;
   U32 type;
   U32 status;
};

/**
 * @type HfmiFragmentType
 *
 * @brief Type of fragment.
 *
 * @field HFMI_FRAGTYPE_MM   The fragment was allocated from the memory
 *                           manager and is equivalent to an OSE memory
 *                           region.
 *
 * @field HFMI_FRAGTYPE_FBUDDY   A virtual Fibonacci buddy fragment. Indicates
 *                               how the buddy algorithm has organized its
 *                               buffers inside a real fragment. The virtual
 *                               fragment is therefore a subset of a real
 *                               fragment, allocated from the memory manager
 *                               or manually added.
 *
 * @field HFMI_FRAGTYPE_HEAP     Deprecated. Replaced by HFMI_FRAGTYPE_FBUDDY.
 *
 * @field HFMI_FRAGTYPE_UNKNOWN  The type of fragment is unknown. It may have
 *                               been manually added to the heap by a user,
 *                               using the deprecated method heap_extend_heap().
 **/
enum HfmiFragmentType
{
   HFMI_FRAGTYPE_MM       = 0,
   HFMI_FRAGTYPE_HEAP     = 1,     /* Deprecated. */
   HFMI_FRAGTYPE_FBUDDY   = 1,
   HFMI_FRAGTYPE_UNKNOWN  = 2
};

/**
 * @type HfmiBufferStatus
 * @brief enum HfmiBufferStatus  A bit-mask indicating the status of a
 *                               buffer.
 *
 * @field HFMI_BUFFER_ALLOCATED      This buffer is allocated.
 * @field HFMI_BUFFER_IS_ANNOTATED   The buffer has annotation information.
 * @field HFMI_BUFFER_MM             The buffer was allocated from the memory
 *                  manager.
 * @field HFMI_BUFFER_ALGO_LARGE     The buffer is managed by the large buffer
 *                                    algorithm.
 * @field HFMI_BUFFER_ALGO_FBUDDY    The buffer is managed by the Fibonacci
 *                                    buddy algorithm.
 * @field HFMI_BUFFER_ALGO_TLSF      The buffer is managed by the TLSF
 *                                    algorithm.
 *
 * @field HFMI_BUFFER_BAD_ENDMARK    The buffer's endmark has been overwritten.
 * @field HFMI_BUFFER_BAD_CHECKSUM   The buffer header integrity check failed.
 * @field HFMI_BUFFER_HEADER_CORRUPT The buffer header integrity check failed.
 **/
enum HfmiBufferStatus {
   HFMI_BUFFER_ALLOCATED        = (1UL << 0),  /* 0x00001 */
   HFMI_BUFFER_IS_ANNOTATED     = (1UL << 1),  /* 0x00002 */
   HFMI_BUFFER_MM           = (1UL << 2),  /* 0x00004 */

   HFMI_BUFFER_ALGO_LARGE       = (1UL << 2),  /* 0x00004 */
   HFMI_BUFFER_ALGO_FBUDDY      = (1UL << 3),  /* 0x00008 */
   HFMI_BUFFER_ALGO_TLSF        = (1UL << 4),  /* 0x00010 */

   HFMI_BUFFER_BAD_ENDMARK      = (1UL << 16), /* 0x10000 */
   HFMI_BUFFER_BAD_CHECKSUM     = (1UL << 17), /* 0x20000 */
   HFMI_BUFFER_HEADER_CORRUPT   = (1UL << 18)  /* 0x40000 */
};

/**
 * @type HfmiFragmentStatus
 * @brief enum HfmiFragmentStatus  A bit-mask indicating the status of a
 *                                 fragment.
 *
 * @field HFMI_FRAG_ALLOCATED      The fragment is in use by an algorithm.
 *
 * @field HFMI_FRAG_BAD_HEADER     The fragment header has been damaged.
 *                                 No other information about the fragment
 *                                 should be trusted.
 *
 **/
enum HfmiFragmentStatus
{
   HFMI_FRAG_ALLOCATED         = (1UL << 0),
   HFMI_FRAG_BAD_HEADER        = (1UL << 1)
};


/**
 * @type HfmiBufferInfo
 * @brief struct HfmiBufferInfo contains information about a single heap
 *      buffer.
 *
 * @field status          The status of the buffer (See HfmiBufferStatus).
 * @field data_address    The address of the user data.
 * @field annotation_address  The address of the annotation data.
 * @field actual_size         The actual allocated size.
 * @field requested_size      The requested size.
 * @field owner               The owner.
 **/
struct HfmiBufferInfo
{
   U32 status;
   U32 data_address;
   U32 annotation_address;
   U32 actual_size;
   U32 requested_size;
   PROCESS owner;
};


/**
 * @type HfmiServiceInfo
 *
 * @brief  This structure contains information about a heap service. 
 *
 * @long   The information can be obtained for all active heap services
 *         by using @link "" "ose_hfmi_get_service".
 *
 * @field  id       A string that uniquely identifies the service.
 *                  Example: "com.enea.heap.tlsf".
 * @field  version  The current service version number.
 * 
 * @seealso ose_hfmi_get_service
 */
struct HfmiServiceInfo
{
   char  id[HEAP_NAME_LENGTH];
   U32   version;
};


/**
 * @function ose_hfmi_get_service
 *
 * @brief   Retrieves information about the next active heap service.
 *
 * @long    The function is called repeatedly with the same argument to
 *          iterate through all active services within a specific heap
 *          instance. Upon the first request, the pointer to the information
 *          structure shall be set to NULL. See example.
 *
 * @input   href   The heap instance for which information about
 *                 services is requested. 
 *
 * @input   svc    Pointer to pointer to structure holding service
 *                 information. The referenced pointer shall be set to
 *                 NULL on first request. See example.
 *                
 * @output  svc    The referenced pointer is updated to point to the
 *                 requested information. 
 *
 * @return  @tab HFMI_OK @tab Service information is available. @break
 *          @tab HFMI_END_OF_LIST @tab No more services could be found. @break
 *          @tab HFMI_NOT_FOUND @tab Invalid heap reference. @break
 *
 * @seealso HfmiServiceInfo
 *
 * @example
 *
 * const struct HfmiServiceInfo *svc = NULL;
 *
 * while (ose_hfmi_get_service(href, &svc) == HFMI_OK)
 * {
 *    ramlog_printf("Service: %d - %s\n", svc->version, svc->id);
 * }
 * 
 */
int
ose_hfmi_get_service(OSHEAP href, const struct HfmiServiceInfo **svc);



/**
 * @function  ose_hfmi_get_heap
 * @brief Get information about a heap from a heap reference.
 * @long  This function is only available in the kernel image.
 *        It must be called in supervisor mode.
 * @long  HFMI functions are designed for freeze-mode debug purposes.
 *
 * @input  href    The heap reference.
 * @output hhc     HFMI cursor for internal use
 * @output hi      See HfmiHeapInfo
 * @input  hi_size The maximum number of bytes in the hi struct.
 *                 Normally sizeof(hi)
 * @return         See enum HfmiStatus
 *
 * @seealso   HfmiHeapInfo
 * @seealso   HfmiStatus
 *
 * @example
 * OSHEAP href = ose_heap_ref(bid);
 * struct HfmiHeapCursor hc;
 * struct HfmiHeapInfo   hi;
 * HfmiStatus rv;
 *
 * rv = ose_hfmi_get_heap(href, &hc, &hi, sizeof(hi));
 * assert(rv == HFMI_OK);
 *
 * do_something_with_heap_info(&hi);
 **/
int
ose_hfmi_get_heap(OSHEAP href,
                  struct HfmiHeapCursor* hhc,
                  struct HfmiHeapInfo* hi, size_t hi_size);

/**
 * @function ose_hfmi_get_heap_reference
 *
 * @brief  Obtain a reference to the heap instance used by a specific
 *         process or block.
 *
 * @input  pid   The process or block for which the heap reference is requested.
 *               Use 0 for current process.
 * @output href  The heap reference.
 *
 * @return  @tab HFMI_OK @tab Success. @break
 *          @tab HFMI_NOT_FOUND @tab No heap found. @break
 */
int
ose_hfmi_get_heap_reference(PROCESS pid, OSHEAP *href);


/**
 * @function ose_hfmi_get_first_heap
 * @brief Get information about the first heap of a system.
 * @long  This function is only available in the kernel image.
 *        It must be called in supervisor mode.
 * @long  HFMI functions are designed for freeze-mode debug purposes.
 *
 * @output  hhc      HFMI cursor for internal use
 * @output  hi       See HfmiHeapInfo
 * @input   hi_size  The maximum number of bytes in the hi struct.
 *                   Normally sizeof(hi)
 * @return           See enum HfmiStatus
 *
 * @seealso   HfmiHeapInfo
 * @seealso   HfmiStatus
 * @seealso   ose_hfmi_get_next_heap
 *
 * @example   See ose_hfmi_get_first_fragment
 **/
int
ose_hfmi_get_first_heap(struct HfmiHeapCursor* hhc,
                        struct HfmiHeapInfo* hi, size_t hi_size);

/**
 * @function ose_hfmi_get_next_heap
 * @brief Get information about the next heap.
 * @long This function is only available in the kernel image.
 *       It must be called in supervisor mode.
 * @long HFMI functions are designed for freeze-mode debug purposes.
 *
 * @input  hhc HFMI cursor for internal use.
 * @output hhc HFMI cursor for internal use.
 * @output hi  See HfmiHeapInfo.
 * @input  hi_size The maximum number of bytes in the hi struct. Normally
 *                 sizeof(hi).
 * @return         See enum HfmiStatus.
 *
 * @seealso   HfmiHeapInfo
 * @seealso   HfmiStatus
 * @seealso   ose_hfmi_get_first_heap
 *
 * @example   See ose_hfmi_get_first_fragment
 **/
int
ose_hfmi_get_next_heap(struct HfmiHeapCursor* hhc,
                       struct HfmiHeapInfo* hi, size_t hi_size);


/**
 * @function  ose_hfmi_get_heap_statistics
 * @brief Get statistics for a heap.
 * @long  This function is only available in the kernel image.
 *        It must be called in supervisor mode.
 * @long  HFMI functions are designed for freeze-mode debug purposes.
 *
 * @input  href    The heap reference.
 * @output hs      See HfmiHeapStatistics.
 * @input  hs_size The maximum number of bytes in the hs struct. Normally
 *                 sizeof(hs)
 * @return    See enum HfmiStatus.
 *
 * @seealso   HfmiHeapStatistics
 * @seealso   HfmiStatus
 **/
int
ose_hfmi_get_heap_statistics(OSHEAP href,
                             struct HfmiHeapStatistics* hs, size_t hs_size);


/**
 * @function ose_hfmi_get_first_fragment
 * @brief Get information about the first fragment in a heap.
 * @long  This function is only available in the kernel image.
 *        It must be called in supervisor mode.
 * @long  HFMI functions are designed for freeze-mode debug purposes.
 *
 * @input  href     The heap reference.
 * @output hfc      HFMI cursor for internal use.
 * @output hfi      See HfmiFragmentInfo.
 * @input  hfi_size The maximum number of bytes in the hfi struct.
 *                   Normally sizeof(hfi)
 * @return          See enum HfmiStatus.
 *
 * @seealso   HfmiFragmentInfo
 * @seealso   HfmiStatus
 * @seealso   ose_hfmi_get_next_fragment
 *
 * @example
 * OSHEAP href = ose_heap_ref(current_process());
 * struct HfmiHeapCursor hc;
 * struct HfmiHeapInfo   hi;
 * struct HfmiFragmentCursor fc;
 * struct HfmiFragmentInfo   frag_info;
 * HfmiStatus rv;
 *
 * rv = ose_hfmi_get_heap(href, &hc, &hi, sizeof(hi));
 * assert(rv == HFMI_OK);
 *
 * rv = ose_hfmi_get_first_fragment(hi.href, &fc, &frag_info, sizeof(frag_info);
 * while(rv == HFMI_OK)
 * {
 *     do_something_with_fragment_info(&frag_info);
 *     rv = ose_hfmi_get_next_fragment(&fc, &frag_info);
 * }
 * assert(rv == HFMI_END_OF_LIST);
 *
 **/
int
ose_hfmi_get_first_fragment(OSHEAP href,
                            struct HfmiFragmentCursor* hfc,
                            struct HfmiFragmentInfo* hfi, size_t hfi_size);

/**
 * @function ose_hfmi_get_next_fragment
 * @brief Get information about the next fragment in a heap.
 * @long  This function is only available in the kernel image.
 *        It must be called in supervisor mode.
 * @long  HFMI functions are designed for freeze-mode debug purposes.
 *
 * @input  hfc HFMI cursor for internal use.
 * @output hfc HFMI cursor for internal use.
 * @output hfi See HfmiFragmentInfo.
 * @input  hfi_size The maximum number of bytes in the hfi struct. Normally sizeof(hfi).
 * @return See enum HfmiStatus.
 *
 * @seealso   HfmiFragmentInfo
 * @seealso   HfmiStatus
 *
 * @example   See ose_hfmi_get_first_fragment
 **/
int
ose_hfmi_get_next_fragment(struct HfmiFragmentCursor* hfc,
                           struct HfmiFragmentInfo* hfi, size_t hfi_size);

/**
 * @function  ose_hfmi_get_first_buffer
 * @brief  Get information about the first buffer in a heap.
 * @long  This function is only available in the kernel image.
 *        It must be called in supervisor mode.
 *
 * @input   href The heap reference.
 * @output  bc HMFI cursor for internal use.
 * @output  bi See HfmiBufferInfo.
 * @input   bi_size The maximum number of bytes in the output struct.
 * @return  See enum HfmiStatus.
 *
 * @seealso   HfmiBufferInfo
 * @seealso   HfmiStatus
 * @seealso   ose_hfmi_get_next_buffer
 *
 * @example   See ose_hfmi_get_next_buffer
 **/
int
ose_hfmi_get_first_buffer(OSHEAP href,
                          struct HfmiBufferCursor* bc,
                          struct HfmiBufferInfo* bi, size_t bi_size);

/**
 * @function ose_hfmi_get_next_buffer
 * @brief Get information about the next buffer in a heap.
 * @long  This function is only available in the kernel image.
 *        It must be called in supervisor mode.
 * @long  HFMI functions are designed for freeze-mode debug purposes.
 *
 * @input  bc HFMI cursor for internal use.
 * @output bc HFMI cursor for internal use.
 * @output bi See HfmiFragmentInfo.
 * @input  bi_size The maximum number of bytes in the output struct.
 * @return See enum HfmiStatus.
 *
 * @seealso   HfmiBufferInfo
 * @seealso   HfmiStatus
 * @seealso   ose_hfmi_get_first_buffer
 *
 * @example   See ose_hfmi_get_first_fragment
 **/
int
ose_hfmi_get_next_buffer(struct HfmiBufferCursor* bc,
                         struct HfmiBufferInfo* bi, size_t bi_size);

/**
 * @function ose_hfmi_find_buffer
 * @brief Find a buffer in a heap.
 * @long  This function is only available in the kernel image.
 *        It must be called in supervisor mode.
 *
 * @input  href The heap reference.
 * @input  address Any address within the heap.
 * @output bc HFMI cursor for internal use.
 * @output bi See HfmiBufferInfo.
 * @input  bi_size The maximum number of bytes in the output struct.
 * @return See enum HfmiStatus.
 *
 * @seealso   HfmiBufferInfo
 * @seealso   HfmiStatus
 *
 * @example
 *  struct HfmiHeapCursor hc;
 *  struct HfmiHeapInfo hi;
 *  struct HfmiBufferCursor bc;
 *  struct HfmiBufferInfo bi;
 *  OSHEAP href = ose_heap_ref(current_process());
 *  int rv;
 *
 *  rv = ose_hfmi_find_buffer(href, (OSADDRESS)ptr, &bc, &bi, sizeof(bi));
 *  assert(rv == HFMI_NOT_FOUND || rv == HFMI_OK);
 *
 *  if(rv == HFMI_NOT_FOUND)
 *  {
 *   log("Pointer not pointing into a heap buffer.\n");
 *  }
 *  else {
 *   log("Buffer start = 0x%lx size=%d.\n",
 *          bi.data_address, bi.actual_size);
 *  }
 *
 **/
int
ose_hfmi_find_buffer(OSHEAP href, OSADDRESS address,
                     struct HfmiBufferCursor* bc,
                     struct HfmiBufferInfo* bi, size_t bi_size);



/**
 * @function ose_hfmi_get_buffer_file_line
 * @brief  Get built in annotation information of a buffer.
 * @long   Retrieves the buffer owner, and the file and line information.
 *         This function uses an unvalidated pointer and might not
 *         be safe to use on a corrupted heap.
 * @long   This function is only available in the kernel image.
 *         It must be called in supervisor mode.
 * @long   HFMI functions are designed for freeze-mode debug purposes.
 *
 * @input  bc     HFMI cursor for internal use.
 * @output owner  This is filled in with the owner of the current
 *                buffer. Zero if shared.
 * @output file   This is filled in with file information of the current
 *                buffer. An empty string if no file information is
 *                available.
 * @input  file_len  The length of the string pointed to by file. Not including the nul character.
 * @output line      This is filled in with line information of the current
 *                   buffer.
 * @return   See enum HfmiStatus
 *
 * @seealso   HfmiStatus
 *
 **/
int
ose_hfmi_get_buffer_file_line(struct HfmiBufferCursor* bc,
                              PROCESS* owner,
                              char* file, size_t file_len,
                              int*  line);

#ifdef __cplusplus
} /* extern "C" */
#endif

#endif /* _HFMI_H */

