/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2011 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */
/**
 * @toc OSE:
 * @file ose_spi/list.h
 * @brief OSE doubly linked list package.
 *
 * @long A linked list contains one main element (@link "" "List"),
 *       representing the list itself, and zero or more elements,
 *       arranged in a ring.
 *       All elements, including the main list element, contain a struct
 *       (@link "" "ListLink") with pointers to next and previous
 *       elements in the list.
 *       @break
 *       ... next_last <-> last(tail) <-> main <-> first(head) <-> second ...
 *
 * @long The ListLink struct shall be the first part of each element,
 *       followed by any user data. This enables the entire element to be
 *       reached via the pointers in the ListLink struct.
 *
 * @long The typedef List is used to indicate the main element,
 *       i.e. the list itself. This main element is not counted and
 *       an empty list contains only the main element, where both next
 *       and previous pointers point to the same main element.
 *
 * @long A new empty list is initiated with the @link "" "list_init" function,
 *       The ListLink struct in a user element is initiated with the
 *       @link "" "list_init_link" function, setting both pointers to NULL,
 *       after which the element can be added to a list.
 */

#ifndef _LIST_H
#define _LIST_H

#ifdef OSE4 /* needs old list package for compatibility */

#include "oldlist.h"

#else /* new OSE5 style */

#include <stddef.h>
#ifdef __cplusplus
extern "C" {
#endif

/**
 * @type  ListLink
 * @alias _ListLink
 * @brief  Typedef for a list element, with next/previous pointers.
 * @long   This struct shall be the first part of each element
 *         of a list, optionally followed by user defined data.
 *
 * @field next Pointer to first/next element in list.
 *             If it points to the main list element, this element
 *             is the tail (last element).
 * @field prev Pointer to last/previous element in list.
 *             If it points to the main list element, this element
 *             is the head (first element).
 * @seealso ose_spi/list.h
 */
typedef struct _ListLink
{
    struct _ListLink *next;	/* Ptr to first/next element in list. */
    struct _ListLink *prev;	/* Ptr to last/previous element in list. */
} ListLink;

/**
 * @type List
 * @brief  Typedef for a doubly linked list.
 * @long   This type is used for the main element of a list.
 *         The main element represent the list itself and exists
 *         also when an existing list is empty.
 * @seealso ose_spi/list.h
 */
typedef ListLink List;

/**
 * @function list_init
 * @brief  Initialize an OSE doubly linked list.
 * @long   The main list element is initialized. The function can handle
 *         NULL as parameter, and can e.g. be used like the following
 *         to create a new list:
 *         @break
 *         List *lst = list_init(malloc(...));
 *
 * @input list Pointer to a List structure, to become the main element
 *         of a new list.
 * @return A pointer to the new list, or NULL if the argument is NULL.
 * @seealso ose_spi/list.h
 * @seealso default_restrictions_calling_functions
 */
List * list_init(List *list);

/**
 * @function list_init_link
 * @brief  Initialize the struct ListLink in a new list element.
 * @long   The next/previous pointers are cleared (set to NULL) in
 *         the ListLink structure pointed to by the lnk argument.
 *         The ListLink structure is followed by any user data.
 *         The function can handle NULL as parameter and can
 *         be used like the following to create a new element,
 *         assuming space is allocated with both a ListLink
 *         structure and user data:
 *         @break
 *         MyLink *lnk = list_init_link(malloc(...));
 *
 * @long   The usage is optional, but using this to initialize the pointers
 *         to NULL before using the element, enables other functions to
 *         better test the validity.
 *
 * @input lnk Pointer to a new element, with a leading ListLink
 *         structure.
 * @return A pointer to the new element, or NULL if the argument is NULL.
 * @seealso ose_spi/list.h
 * @seealso default_restrictions_calling_functions
 * @example
 *        #include ose_spi/list.h
 *        // Define element with ListLink and user data
 *        struct MyLink {
 *           ListLink double_link;
 *           MyUserData myUserData;
 *        };
 *        // Create list
 *        List *myList =  list_init(malloc(sizeof(List)));
 *        // Create one element
 *        MyLink *lnk = list_init_link(malloc(sizeof(myLink)));
 *        if (lnk == NULL) {
 *           // ... error ...
 *        }
 *        // ... add user data ...
 *        // Insert element as the first element in the list
 *        list_insert_head(myList, lnk);
 */
void *list_init_link(void *lnk);

/**
 * @function list_unlink
 * @brief  Remove an element from a doubly linked list.
 * @long   The element is removed by modifying the next
 *         and previous list elements, removing pointers to this element.
 * @long   The pointers in the removed element are cleared (set to NULL).
 *
 * @input link_to_remove Pointer to the ListLink structure in the
 *         element to remove.
 * @return Returns a pointer to the same removed element, or NULL if
 *         the argument is NULL.
 * @seealso default_restrictions_calling_functions
 */
ListLink * list_unlink(void *link_to_remove);

/**
 * @function list_insert_head
 * @brief  Insert an element as the first element in a list (the head element).
 * @long   The list can contain other elements, but this function
 *         inserts the new element in the beginning of the list,
 *         i.e. with its "prev" pointer pointing to the main list
 *         element.
 * @long   This function checks the validity of the entire list,
 *         before inserting the element.
 *
 * @input list           List into which to insert the element.
 * @input link_to_insert The element to insert.
 * @seealso default_restrictions_calling_functions
 */
void list_insert_head(List *list, void *link_to_insert);

/**
 * @function list_insert_tail
 * @brief  Inserts an element as the last element in a list (the tail element).
 * @long   The list can contain other elements, but this function
 *         inserts the new element at the end of the list,
 *         i.e. with its "next" pointer pointing to the main list
 *         element.
 * @long   This function checks the validity of the entire list,
 *         before inserting the element.
 *
 * @input list           List into which to insert the element.
 * @input link_to_insert The element to insert.
 * @seealso default_restrictions_calling_functions
 */
void list_insert_tail(List *list, void *link_to_insert);

/**
 * @function list_insert_after
 * @brief  Insert an element after an existing element in a list.
 *
 * @input link_in_list   Existing element in the list.
 * @input link_to_insert The element to insert.
 * @seealso default_restrictions_calling_functions
 */
void list_insert_after(void *link_in_list, void *link_to_insert);

/**
 * @function list_insert_before
 * @brief  Insert an element before an existing element in a list.
 *
 * @input link_in_list   Existing element in the list.
 * @input link_to_insert The element to insert.
 * @seealso default_restrictions_calling_functions
 */
void list_insert_before(void *link_in_list, void *link_to_insert);

#ifdef __cplusplus
/* C++ equivalents of function like macros. */

static inline void *
list_head(const List *list)
{
    return (((list)->next != (list)) ? (list)->next : NULL);
}
static inline void *
list_tail(const List *list)
{
    return (((list)->prev != (list)) ? (list)->prev : NULL);
}
static inline void *
list_next(const List *list, const void *ll)
{
    const ListLink *lnk = (const ListLink *)ll;
    return (((lnk)->next != (list)) ? (lnk)->next : NULL);
}
static inline void *
list_prev(const List *list, const void *ll)
{
    const ListLink *lnk = (const ListLink *)ll;
    return (((lnk)->prev != (list)) ? (lnk)->prev : NULL);
}

#else /* And now C function like macros. */

/**
 * @function list_head
 * @brief  Get the first (head) element of a list.
 *
 * @input list   List in which the first element shall be found.
 * @return Pointer to the first (head) element of a list,
 *         or NULL if the list is empty.
 * @seealso default_restrictions_calling_functions
 */
void * list_head(const List *list);
#define list_head(list) (((list)->next != (list)) ? \
   (void*)((list)->next) : NULL)

/**
 * @function list_tail
 * @brief  Get the last (tail) element of a list.
 *
 * @input list   List in which the last element shall be found.
 * @return Pointer to the last (tail) element of a list,
 *         or NULL if the list is empty.
 * @seealso default_restrictions_calling_functions
 */
void * list_tail(const List *list);
#define list_tail(list) (((list)->prev != (list)) ? \
   (void*)((list)->prev) : NULL)

/**
 * @function list_next
 * @brief  Get the next element in a list, after a given element.
 * @long   If this is the last element (tail), the function returns NULL.
 *
 * @input list   List in which the element is.
 * @input lnk    Existing element in the list.
 * @return Pointer to the next element
 *         or NULL if this is the last element (tail).
 * @seealso default_restrictions_calling_functions
 */
void * list_next(const List *list, const void *lnk);
#define list_next(list, lnk) ((((const ListLink *)(lnk))->next != (list)) ? \
   (void*)(((const ListLink *)(lnk))->next) : NULL)

/**
 * @function list_prev
 * @brief  Get the previous element in a list, before a given element.
 * @long   If this is the first element (head), the function returns NULL.
 *
 * @input list   List in which the element is.
 * @input lnk    Existing element in the list.
 * @return Pointer to the previous element
 *         or NULL if this is the first element (head).
 * @seealso default_restrictions_calling_functions
 */
void * list_prev(const List *list, const void *lnk);
#define list_prev(list, lnk) ((((const ListLink *)(lnk))->prev != (list)) ? \
   (void*)(((const ListLink *)(lnk))->prev) : NULL)

#endif

/**
 * @function list_length
 * @brief  Get the number of elements in a doubly linked list.
 * @long   This does not include the main list element itself.
 *         The elements are counted by traversing the list.
 *         There is no storage of the count, to avoid overhead
 *         when operating on the list.
 *
 * @input list  List in which the elements shall be counted.
 * @return The number of elements in the list.
 * @seealso default_restrictions_calling_functions
 */
unsigned list_length(const List *list);

/**
 * @function list_traverse
 * @brief  Traverse a doubly linked list and call a user supplied function
 *         for each element.
 * @long   The given list is traversed, from head to tail and the given
 *         user defined function (apply) is called for each element.
 * @long   If the apply function returns NULL, the traversal continues,
 *         otherwise the list_traverse returns with the return
 *         value from the function.
 *
 * @input list       List to traverse.
 * @input apply      User defined function to call for each element.
 * @input apply.lnk  Pointer to the element.
 * @input apply.arg  Second argument to the function is a pointer to the arg
 *                   list_traverse argument - user defined.
 * @input arg        Pointer to a user-defined argument, which is used as
 *                   second argument to the user defined function.
 *                   If the argument is in writable memory, its value
 *                   can be modified by the user defined function.
 * @return NULL when the entire list has been traversed. If non-null,
 *         the returned value is the value returned from the last call
 *         to the user defined function.
 * @seealso default_restrictions_calling_functions
 */
void * list_traverse(List *list,
		     void *(*apply) (ListLink *lnk, void *arg),
		     void *arg);

/**
 * @function list_insert_ordered
 * @brief  Inserts an element into an ordered list.
 * @long   A user defined function (compare) is called to define
 *         the order between two elements. It is called repeatedly to
 *         compare the new element with an existing element in the list,
 *         until the new element's location is found.
 * @long   The compare function shall return a positive value if the
 *         new element should be inserted before the existing,
 *         zero if they are on the same level (will be inserted after)
 *         or a negative value if the new should be inserted after the
 *         existing.
 *
 * @input list             List in which the new value will be inserted.
 * @input link_to_insert   New element to be inserted.
 * @input compare          User defined function for comparing two elements.
 * @input compare.new_elem A pointer to the new element.
 * @input compare.current  A pointer to an element in the List.
 * @seealso default_restrictions_calling_functions
 */
void list_insert_ordered(List *list,
     void *link_to_insert,
     int (*compare) (const ListLink *new_elem, const ListLink *current));

/**
 * @function list_assert
 * @brief  Asserts that all elements in a list are valid.
 *
 * @long   This function validates all elements in a list.
 *         If the assertion of any of the next/prev pointers fails,
 *         an @link "ose.h" "error2" is generated.
 *
 * @long   The list_assert function also traverses the list and calls
 *         a user-supplied function for each element, in order from
 *         head to tail, unless the provided function pointer is NULL.
 *         It is the responsibility of the user supplied function
 *         to take care of any found problems.
 *
 * @input list       List, which elements shall be checked.
 * @input check      Pointer to user supplied function.
 * @input check.node The function parameter is a pointer to the element.
 * @seealso default_restrictions_calling_functions
 */
void list_assert(const List *list, void (*check) (const ListLink *node));

#ifdef __cplusplus
}
#endif

#endif /* OSE4/OSE5 */

#endif /* LIST_H */

