
/* COPYRIGHT-ENEA-SRC-R1 *
**************************************************************************
* Copyright (C) 2005 by Enea Embedded Technology AB.
* All rights reserved.
*
* This Software is furnished under a software license agreement and
* may be used only in accordance with the terms of such agreement.
* Any other use or reproduction is prohibited. No title to and
* ownership of the Software is hereby transferred.
*
* PROPRIETARY NOTICE
* This Software consists of confidential information.
* Trade secret law and copyright law protect this Software.
* The above notice of copyright on this Software does not indicate
* any actual or intended publication of such Software.
**************************************************************************
* COPYRIGHT-END */
/**
 * @public
 *
 * @toc Memory_Management:
 *
 * @file ose_spi/mfmi.h
 *
 * @brief Memory Manager Freeze-mode Interface
 *
 * @long
 * This file declares the Memory Manager Freeze-mode Interface.
 * It may be used to fetch information from the Memory Manager (MM) in a
 * freeze-mode environment. The purpose is to make Memory concepts accessible
 * for freeze-mode debug monitors that can not use ordinary system calls.
 * @long
 * These functions performs very few tests of the memory manager data
 * structures. It may report failure if the referred domain or region
 * can not be found, is severely damaged or locked by the memory manager process.
 * An obvious example is when the memory manager is not yet initiated or
 * when the memory manager is busy updating the internal data. Interrupts
 * should be disabled when this interface is used.
 * @long
 * WARNING: Use of MFMI functionality for any other purpose will severely
 * damage the interrupt latency!
 *
 */

#ifndef _MFMI_H
#define _MFMI_H


#include "stddef.h"
#include "ose.h"
#include "stdint.h"
#include "mmtypes.h"
#include "mm.h"


#define MFMI_CURSOR_SIZE (16 / sizeof(U32))


#ifdef __cplusplus
extern "C" {
#endif


/* Return status of functions */
/**
 *  @type MfmiStatus
 *
 *  @brief enum MfmiStatus defines the set of return values from MFMI
 *         functions.
 *
 *  @field MFMI_OK                  Operation was successful.
 *  @field MFMI_UNAVAILABLE         Requested information not found.
 *  @field MFMI_END_OF_LIST         No more objects found.
 *  @field MFMI_DATA_LOCKED         MM data structures are locked by the MM process.
 *  @field MFMI_DATA_CORRUPTED      MM data is corrupted.
 *  @field MFMI_EDOMAIN             Illegal domain.
 *  @field MFMI_EREGION             Illegal region.
 **/
typedef enum MfmiStatus
  {
    MFMI_OK               = 0,
    MFMI_UNAVAILABLE      = 1,
    MFMI_END_OF_LIST      = 2,
    MFMI_DATA_LOCKED      = 3,
    MFMI_DATA_CORRUPTED   = 4,
    MFMI_EDOMAIN          = 5,
    MFMI_EREGION          = 6
  } MfmiStatus;


/**
 *  @type MfmiDomainInfo
 *
 *  @brief struct MfmiDomainInfo contains information for one domain.
 *
 *  @field domain 	Domain number.
 *  @field name 	Domain name.
 **/
struct MfmiDomainInfo
{
  OSDOMAIN domain;
  char     name[MM_NAME_LENGTH + 1];
};

/**
 *  @type MfmiRegionInfo
 *
 *  @brief struct MfmiRegionInfo contains information for one region.
 *
 *  @field name 	Region name.
 *  @field base 	Region base address.
 *  @field size		Region size.
 *  @field permissions  Access permissions.
 *  @field cache_mode	Cache attributes.
 *  @field phys_mem	Name of the physical allocation area.
 *  @field log_mem	Name of the logical allocation area.
 *  @field options      Options used when creating the region
 *                      See @link "mm.sig" "MM_CREATE_REGION_SPECIFY_NAME",
 *                      @link "mm.sig" "MM_CREATE_REGION_SPECIFY_PHYS",
 *                      @link "mm.sig" "MM_CREATE_REGION_SPECIFY_LOG",
 *                      @link "mm.sig" "MM_CREATE_REGION_LOGICAL",
 *                      @link "mm.sig" "MM_CREATE_REGION_SPECIFY_ALIGN",
 *                      @link "mm.sig" "MM_CREATE_REGION_PERSISTENT",
 *                      @link "mm.sig" "MM_CHECK_LOW_LIMIT".
 *                      @link "mm.sig" "MM_LOGMEM_FROM_ADDRESS".
 **/
struct MfmiRegionInfo
{
  char       name[MM_NAME_LENGTH + 1];
  OSADDRESS  base;
  mm_size_t  size;
  uint32_t   permissions;
  uint32_t   cache_mode;
  char       phys_mem[MM_NAME_LENGTH + 1];
  char       log_mem[MM_NAME_LENGTH + 1];
  uint32_t   options;
};

/**
 *  @type MfmiAddressInfo
 *
 *  @brief struct MfmiAddressInfo contains translation information for
 *  one specific logical address.
 *  @field address       Physical address.
 *  @field fragment_base Physical fragment base for address.
 *  @field fragment_size Physical fragment size for address.
 **/
struct MfmiAddressInfo
{
  phys_addr_t address;
  phys_addr_t fragment_base;
  mm_size_t   fragment_size;
};

/**
 *  @type MfmiInfoTag
 *
 *  @brief enum MfmiInfoTag is used to request information from the
 *  memory manager
 *
 *  @field MFMI_INFO_PHYS_MEM_TOTAL     Total physical memory.
 *  @field MFMI_INFO_PHYS_MEM_FREE      Free physical memory.
 *  @field MFMI_INFO_PHYS_MAX_FRAG_SIZE Max block size for SASE allocation.
 *  @field MFMI_INFO_PHYS_MEM_INACTIVE  This field is deprecated and should
 *                                      not be used.
 *  @field MFMI_INFO_PHYS_MEM_WATERMARK Free physical memory watermark.
 **/
enum MfmiInfoTag
  {
    MFMI_INFO_PHYS_MEM_TOTAL,
    MFMI_INFO_PHYS_MEM_FREE,
    MFMI_INFO_PHYS_MAX_FRAG_SIZE,
    MFMI_INFO_PHYS_MEM_INACTIVE,
    MFMI_INFO_PHYS_MEM_WATERMARK
  };


/**
 *  @type MfmiPhysMemTypeInfo
 *
 *  @brief struct MfmiPhysMemTypeInfo contains information
 *  for one physical memory type.
 *
 *  @field name 	  Name of the physical memory type.
 *  @field total_size 	  Total amount of physical memory
 *                        in the physical memory type.
 *  @field free_size	  Amount of free physical memory
 *                        in the physical memory type.
 *  @field max_frag_size  The largest physical fragment used
 *                        in the physical memory type.
 *  @field inactive_size  This field is deprecated and is always set to zero.
 *  @field free_watermark Minimum amount of free memory available so far
 *                        in the physical memory type.
 **/
struct MfmiPhysMemTypeInfo
{
   char name[MM_NAME_LENGTH + 1];
   phys_size_t total_size;
   phys_size_t free_size;
   phys_size_t max_frag_size;
   phys_size_t inactive_size;
   phys_size_t free_watermark;
};

/**
 *  @type MfmiDomainCursor
 *
 *  @brief MFMI Cursor data types
 *
 *  @long Variables of struct types MfmiDomainCursor, MfmiRegionCursor,
 *  are provided by the caller of MFMI functions.
 *  They are used by MFMI for internal book keeping.
 *
 *  @field cursor For internal use by MFMI.
 **/

struct MfmiDomainCursor         { U32 cursor[MFMI_CURSOR_SIZE]; };
struct MfmiRegionCursor         { U32 cursor[MFMI_CURSOR_SIZE]; };
struct MfmiPhysMemTypeCursor    { U32 cursor[MFMI_CURSOR_SIZE]; };


/**
 *  @function ose_mfmi_get_domain_info
 *
 *  @brief Get information on a domain.
 *
 *  @long This function is only available in the kernel image. It must
 *  be called in supervisor mode.
 *  @long  MFMI functions are designed for freeze
 *  mode debug purposes. Any other use will destroy the interrupt latency
 *  of the system. MFMI functions are not part of the real time services.
 *
 *  @input domain   Domain number
 *  @output di		Domain information
 *  @return         See enum MfmiStatus
 *  @error          See enum MfmiStatus
 **/
MfmiStatus
ose_mfmi_get_domain_info(OSDOMAIN domain,
			 struct MfmiDomainInfo *di);

/**
 *  @function ose_mfmi_get_region_info
 *
 *  @brief Get information on a region.
 *
 *  @long This function is only available in the kernel image. It must
 *  be called in supervisor mode.
 *  @long  MFMI functions are designed for freeze-mode debug purposes.
 *  Any other use will destroy the interrupt latency
 *  of the system. MFMI functions are not part of the real-time services.
 *
 *  @input domain	Domain number.
 *  @input address	Region address.
 *  @output ri		Region information.
 *  @return         See enum MfmiStatus.
 *  @error          See enum MfmiStatus.
 *  @seealso        MfmiRegionInfo
 *  @seealso        MfmiStatus
 **/
MfmiStatus
ose_mfmi_get_region_info(OSDOMAIN domain,
			 OSADDRESS address,
			 struct MfmiRegionInfo *ri);

/**
 *  @function ose_mfmi_get_address_info
 *
 *  @brief Get physical address information for a logical address.
 *
 *  @long This function is only available in the kernel image. It must
 *  be called in supervisor mode.
 *  @long  MFMI functions are designed for freeze-mode debug purposes.
 *  Any other use will destroy the interrupt latency
 *  of the system. MFMI functions are not part of the real-time services.
 *
 *  @input domain        The domain that maps the logical address.
 *  @input log_address   The logical address.
 *  @output address_info The physical address information.
 *  @return              See enum MfmiStatus.
 *  @error               See enum MfmiStatus.
 **/
MfmiStatus
ose_mfmi_get_address_info(OSDOMAIN domain,
			  OSADDRESS log_address,
			  struct MfmiAddressInfo *address_info);

/**
 *  @function ose_mfmi_get_memory_info
 *
 *  @brief Get internal information from the memory manager.
 *
 *  @long This function is only available in the kernel image. It must
 *  be called in supervisor mode.
 *  @long  MFMI functions are designed for freeze-mode debug purposes.
 *  Any other use will destroy the interrupt latency
 *  of the system. MFMI functions are not part of the real-time services.
 *
 *  @input tag		    Information selection tag.
 *  @output memory_info	Result value representing the tag.
 *  @return             See enum MfmiStatus.
 *  @error              See enum MfmiStatus.
 *  @seealso            MfmiInfoTag
 **/
MfmiStatus
ose_mfmi_get_memory_info(enum MfmiInfoTag tag,
			 mm_size_t *memory_info);

/* Retrieve the domain list of the system
 */
/**
 *  @function ose_mfmi_get_first_domain
 *
 *  @brief Get information on the first domain in the domain list.
 *
 *  @long This function is only available in the kernel image. It must
 *  be called in supervisor mode.
 *  @long  MMFI functions are designed for freeze-mode debug purposes.
 *  Any other use will destroy the interrupt latency
 *  of the system. MFMI functions are not part of the real-time services.
 *
 *  @input dc		MFMI cursor for internal use.
 *  @output di		Domain information.
 *  @return         See enum MfmiStatus.
 *  @error          See enum MfmiStatus.
 *  @seealso        ose_mfmi_get_next_domain.
 *  @example        See ose_mfmi_get_first_region.
 **/
MfmiStatus
ose_mfmi_get_first_domain(struct MfmiDomainCursor *dc,
			  struct MfmiDomainInfo   *di);

/**
 *  @function ose_mfmi_get_next_domain
 *
 *  @brief Get information for next domain in the domain list.
 *
 *  @long This function is only available in the kernel image. It must
 *  be called in supervisor mode.
 *  @long  MFMI functions are designed for freeze-mode debug purposes.
 *  Any other use will destroy the interrupt latency
 *  of the system. MFMI functions are not part of the real-time services.
 *
 *  @input  dc		MFMI cursor for internal use.
 *  @output di		Domain information.
 *  @return         See enum MfmiStatus.
 *  @error          See enum MfmiStatus.
 *  @seealso        ose_mfmi_get_first_domain.
 *  @example        See ose_mfmi_get_first_region.
 **/
MfmiStatus
ose_mfmi_get_next_domain(struct MfmiDomainCursor *dc,
			 struct MfmiDomainInfo   *di);


/* Retrieve region information per domain
 */
/**
 *  @function ose_mfmi_get_first_region
 *
 *  @brief Get information on the first region in a domain.
 *
 *  @long This function is only available in the kernel image. It must
 *  be called in supervisor mode.
 *  @long  MFMI functions are designed for freeze-mode debug purposes.
 *  Any other use will destroy the interrupt latency
 *  of the system. MFMI functions are not part of the real-time services.
 *
 *  @input domain 	The domain to investigate.
 *  @input rc		MFMI cursor for internal use.
 *  @output ri		Region information.
 *  @return         See enum MfmiStatus.
 *  @error          See enum MfmiStatus.
 *  @seealso        ose_mfmi_get_next_region
 *  @example
 *   struct MfmiRegionCursor rc;
 *   struct MfmiRegionInfo   ri;
 *   MfmiStatus rv;
 *
 *   rv = ose_mfmi_get_first_region(domain, &rc, &ri);
 *
 *   while(rv == MFMI_OK)
 *   {
 *       do_something_with_region_info(&ri);
 *       rv = ose_mfmi_get_next_region(&rc, &ri);
 *   }
 *
 *   if(rv != MFMI_END_OF_LIST)
 *   {
 *      printf("Something went wrong when retrieving region info.\n");
 *   }
 **/
MfmiStatus
ose_mfmi_get_first_region(OSDOMAIN domain,
			  struct MfmiRegionCursor *rc,
			  struct MfmiRegionInfo   *ri);

/**
 *  @function ose_mfmi_get_next_region
 *
 *  @brief Get information for next region in a domain.
 *
 *  @long This function is only available in the kernel image. It must
 *  be called in supervisor mode.
 *  @long  MFMI functions are designed for freeze-mode debug purposes.
 *  Any other use will destroy the interrupt latency
 *  of the system. MFMI functions are not part of the real-time services.
 *
 *  @input  rc		MFMI cursor for internal use.
 *  @output ri		Region information.
 *  @return         See enum MfmiStatus.
 *  @error          See enum MfmiStatus.
 *  @seealso        ose_mfmi_get_first_region
 *  @example        See ose_mfmi_get_first_region.
 **/
MfmiStatus
ose_mfmi_get_next_region(struct MfmiRegionCursor *rc,
			 struct MfmiRegionInfo   *ri);

/**
 *  @function ose_mfmi_get_first_phys_mem_type
 *
 *  @brief Get information on the first physical memory type available.
 *
 *  @long This function is only available in the kernel image. It must
 *  be called in supervisor mode.
 *  @long  MFMI functions are designed for freeze-mode debug purposes.
 *  Any other use will destroy the interrupt latency
 *  of the system. MFMI functions are not part of the real-time services.
 *
 *  @input pmtc		MFMI cursor for internal use.
 *  @output pmti	Physical memory type information.
 *  @return         See enum MfmiStatus.
 *  @error          See enum MfmiStatus.
 *  @seealso        ose_mfmi_get_next_phys_mem_type
 *  @example
 *   struct MfmiPhysMemTypeCursor pmtc;
 *   struct MfmiPhysMemTypeInfo   pmti;
 *   MfmiStatus rv;
 *
 *   rv = ose_mfmi_get_first_phys_mem_type(&pmtc, &pmti);
 *
 *   while(rv == MFMI_OK)
 *   {
 *       do_something_with_phys_mem_type_info(&pmti);
 *       rv = ose_mfmi_get_next_phys_mem_type(&pmtc, &pmti);
 *   }
 *
 *   if(rv != MFMI_END_OF_LIST)
 *   {
 *      printf("Something went wrong when retrieving "
 *              "physical memory type info.\n");
 *   }
 **/
MfmiStatus
ose_mfmi_get_first_phys_mem_type(struct MfmiPhysMemTypeCursor *pmtc,
                                 struct MfmiPhysMemTypeInfo   *pmti);

/**
 *  @function ose_mfmi_get_next_phys_mem_type
 *
 *  @brief Get information on the next physical memory type.
 *
 *  @long This function is only available in the kernel image. It must
 *  be called in supervisor mode.
 *  @long  MFMI functions are designed for freeze-mode debug purposes.
 *  Any other use will destroy the interrupt latency
 *  of the system. MFMI functions are not part of the real-time services.
 *
 *  @input  pmtc        MFMI cursor for internal use.
 *  @output pmti       	Physical memory type information.
 *  @return             See enum MfmiStatus.
 *  @error              See enum MfmiStatus.
 *  @seealso            ose_mfmi_get_first_phys_mem_type
 *  @example            See ose_mfmi_get_first_phys_mem_type.
 **/
MfmiStatus
ose_mfmi_get_next_phys_mem_type(struct MfmiPhysMemTypeCursor *pmtc,
                                struct MfmiPhysMemTypeInfo   *pmti);


/**
 * @function ose_mfmi_get_cache_type
 * 
 * @brief Decompile a binary cache mode bitfield into cache and memory
 *        type identification strings.
 *
 * @long  This function is only available in the kernel image. It must be
 *        called in supervisor mode.
 *
 * @long  MFMI functions are designed for freeze-mode debug purposes.
 *        Any other use will destroy the interrupt latency of the system.
 *        MFMI functions are not part of the real-time services.
 *
 * @input    cache_mode    The cache mode bitfield to decompile.
 * @output   cache_type    Name of the cache type. May be NULL
 *                         if this string is not desired, otherwise
 *                         the string must be able to hold at least
 *                         MM_NAME_LENGTH + 1 characters.
 * @output   memory_type   Name of the memory type. May be NULL
 *                         if this string is not desired, otherwise
 *                         the string must be able to hold at least
 *                         MM_NAME_LENGTH + 1 characters.
 *
 * @return See enum MfmiStatus.
 */
 
MfmiStatus
ose_mfmi_get_cache_type(uint32_t cache_mode,
                        char *cache_type,
                        char *memory_type);
   
#ifdef __cplusplus
}
#endif

#endif /* _MFMI_H */
