/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2015 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */
/**
 * @toc CPU_Dependent:
 * @file ose_spi/mips/cpu_access.h
 *
 * @brief Access of MIPS specific processor features.
 *
 * @long Many of these functions require Supervisor Mode privileges.
 *
 * @long See this header file for details about defines with many macros
 * for CPU registers and masks.
 */
#ifndef _CPU_ACCESS_H
#define _CPU_ACCESS_H

#ifdef __cplusplus
extern "C" {
#endif

#include "cpu_regs.h"
#include "osetypes.h"
#include "cache.h"
#include "cpu_intlock.h"
/* Needed because of FLUSH_CACHE macro. */
#include "cache.h"

#ifndef STRSTR
#define OSE_STR(x) #x
#define STRSTR(x) OSE_STR(x)
#endif

#if defined __GNUC__
#define always_inline __inline__ __attribute__((always_inline))
#endif

#if defined (__GNUC__)
#define INLINE __inline__
#else
#define INLINE
#endif

/**
 * @macro read_cp0
 * @field reg    CP0 register number.
 * @field sel    Select number.
 * @brief Read from CP0 register.
 */
#define read_cp0(reg, sel)                               \
   ({                                                    \
    unsigned long  __result;                             \
    __asm__ volatile (".set push;"                       \
		      ".set mips32;"                     \
		      "mfc0 %0," STRSTR(reg) "," #sel ";" \
		      ".set pop;"                        \
		      : "=r"(__result));                 \
    __result;                                            \
    })

/**
 * @macro write_cp0
 * @brief Write to CP0 register.
 * @field reg    CP0 register number.
 * @field sel    Select number.
 * @field value  New value.
 */
#define write_cp0(reg, sel, value)                          \
   do {                                                     \
      __asm__ volatile (".set push;"                        \
			".set mips32;"                      \
			"mtc0 %0," STRSTR(reg) "," #sel ";"  \
			".set pop;"                         \
			:: "r"(value));                     \
   } while (0)

/**
 * @macro read_cp1
 * @field reg    CP1 register number.
 * @brief Read from CP1 register.
 */
#define read_cp1(reg)                                    \
   ({                                                    \
    unsigned long  __result;                             \
    __asm__ volatile (".set push;"                       \
		      ".set mips32;"                     \
		      "cfc1 %0," STRSTR(reg) ";"         \
		      ".set pop;"                        \
		      : "=r"(__result));                 \
    __result;                                            \
    })


/**
 * @macro write_cp1
 * @brief Write to CP1 register.
 * @field reg    CP1 register number.
 * @field value  New value.
 */
#define write_cp1(reg, value)                       \
   do {                                             \
      __asm__ volatile (".set push;"                \
			".set mips32;"              \
			"ctc1 %0," STRSTR(reg) ";"  \
			".set pop;"                 \
			:: "r"(value));             \
   } while (0)

/**
 * @macro MEMBAR
 * @brief Synchronize shared memory.
 */
#define MEMBAR() __asm__ volatile ("sync")

/**
 * @function write_cp0_status
 * @brief Set the status register in CP0.
 * @long Wait for disable to take effect. Three number of nops recommended in
 * MIPS manuals, CP0 hazards.
 * @input sr Status register to be written to CP0.
 */
#ifdef _MIPS_ARCH_MIPS3
static INLINE void
write_cp0_status(unsigned long sr)
{
   __asm__ volatile
      (".set push;"
       ".set noreorder;"
       " mtc0  %0,$12;"
       " nop;"
       " nop;"
       " nop;"
       ".set pop;"
       :: /*inputs*/ "r" (sr));
}
#else
static INLINE void
write_cp0_status(unsigned long sr)
{
   __asm__ volatile
      (".set push;"
       ".set noreorder;"
       " mtc0  %0,$12;"
       " ssnop;"
       " ssnop;"
       " ssnop;"
       ".set pop;"
       :: /*inputs*/ "r" (sr));
}
#endif

/**
 * @macro FLUSH_CACHE
 *
 * @brief Flush (push and invalidate) specified region of data cache (PowerPC).
 *
 * @input base  Base address to start flushing cache.
 * @input size  Number of bytes to flush.
 * @deprecated  This method is deprecated in OSE5.5 and will not be
 *              supported in future OSE releases.
 *              See "ose_cache" in cache.h for similar functionality.
 */

/* This macro is just here for backwards compatibility. */
#define FLUSH_CACHE(base, size) \
	cache_flush_invalidate_d((OSADDRESS)(base), (size))



/**
 * @macro CPU_ID
 *
 * @brief Return the CPU identify of current CPU.
 *
 */
#define CPU_ID()      ((read_cp0(CP0_PRID, 0) >> 25) & 0x7)


#ifdef __cplusplus
}
#endif

#endif /* # ifndef _CPU_ACCESS_H */
