/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2006 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */
/**
 * @public
 * @toc CPU_Dependent:
 * @file ose_spi/mips/cpu_hal_arch.h
 *
 * @brief This is the CPU HAL interface for MIPS.
 *
 * @long It provides CPU specific functions for the MM HAL and core cache.
 * It also provides defines for exec and mm attributes that is used in the
 * BSP file cpu_hal_xxxx.c, these defines are also used internally by OSE.
 */

#ifndef CPU_HAL_ARCH_H_
#define CPU_HAL_ARCH_H_

#include "osetypes.h"

#ifdef __cplusplus
extern "C" {
#endif

/**
 * @macro CPU_HAL_VERSION
 *
 * @brief Version of the MIPS CPU HAL interface.
 */
#define CPU_HAL_VERSION    2

/*
 ****************************************************************************
 * EXEC attributes.
 ****************************************************************************
 */

/*
 ****************************************************************************
 * MM attributes.
 ****************************************************************************
 */

/**
 * @macro ATTRIBUTE_HW_CACHE_COHERENCY
 *
 * @brief HW cache coherency proto is used by CPU.
 * Use same value for all architectures.
 */
#define ATTRIBUTE_HW_CACHE_COHERENCY 0x00000008

/**
 * @macro ATTRIBUTE_I_CACHE_VIRTUAL_TAGGED
 *
 * @brief The I-cache is virtually tagged.
 * Use same value for all architectures.
 */
#define ATTRIBUTE_I_CACHE_VIRTUAL_TAGGED 0x00000010

/**
 * @macro ATTRIBUTE_D_CACHE_VIRTUAL_TAGGED
 *
 * @brief The D-cache is virtually tagged.
 * Use same value for all architectures.
 */
#define ATTRIBUTE_D_CACHE_VIRTUAL_TAGGED 0x00000020

/**
 * @type cache_info
 *
 * @brief Characteristics for the MIPS cache sub-system.
 *
 * @field d_cache_enabled   Non-zero if data cache is enabled.
 * @field i_cache_enabled   Non-zero if instruction cache is enabled.
 * @field d_cache_line_size Line size of the data cached.
 * @field i_cache_line_size Line size of the instruction cache.
 * @field l1_d_cache_size   Size of the L1 data cache.
 * @field l1_i_cache_size   Size of the L1 instruction cache.
 * @field L2_cache_size     Size of the unified L2 cache.
 * @field L3_cache_size     Size of the unified L3 cache.
 */
struct cache_info
{
   U32 d_cache_enabled;   /* Non-zero if enabled */
   U32 i_cache_enabled;   /* Non-zero if enabled */
   U32 d_cache_line_size;
   U32 i_cache_line_size;
   U32 l1_d_cache_size;
   U32 l1_i_cache_size;
   U32 L2_cache_size;     /* Unified. */
   U32 L3_cache_size;     /* Unified. */
};

/**
 * @type entry_lo
 *
 * @brief Cache coherency attribute
 *
 * @field cache_attr_uncached Bitmask for uncached attribute.
 * @field cache_attr_cachable Bitmask for cached attribute.
 */
struct entry_lo
{
   U32 cache_attr_uncached;
   U32 cache_attr_cachable;
};

/**
 * @type arch_cpu_descriptor
 *
 * @brief MIPS specific CPU information.
 *
 * @field exec_attributes Attributes defined and used by EXEC.
 *
 * @field mm_attributes Attributes defined and used by MM.
 *
 * @field init_cpu Function pointer, performs necessary CPU initialization.
 *
 *                 On Book E processors, address space 0 must be activated
 *                 and one or more TLB entries mapping the core regions must
 *                 be pre-loaded.
 *
 *                 Precondition:
 *                 - Unknown since board-specific firmware has setup the board.
 *
 *                 Postcondition:
 *                 - MMU disabled (or address space 0 activated).
 *
 * @field enable_caches Function pointer, enables D- and/or I-caches.
 *                      Set arguments to non-zero for enable.
 *                      If applicable, all levels (L1, L2, L3, etc.) are to
 *                      be enabled here.
 *
 *                      Since the enable status of the L1 caches control
 *                      whether deeper level caches are to be used, deeper
 *                      level caches must be initialized first.
 *
 *                      Hence the recommended order is:
 *                      - invalidate, configure and enable L3 cache
 *                      - invalidate, configure and enable L2 cache
 *                      - invalidate, configure and enable L1 cache(s).
 *                      @break
 *                      An implementation may choose to do nothing in this
 *                      function, thereby using the cache state established
 *                      by board-resident firmware.
 *
 *                      Precondition:
 *                      - MMU disabled or address space 0 active.
 *
 *                      Postcondition:
 *                      - Cache sub-systems configured according to
 *                        <enable_d_cache> and <enable_i_cache>.
 *
 * @input enable_caches.enable_d_cache True if data cache should be enabled.
 * @input enable_caches.enable_i_cache True if instruction cache should be
 *                                     enabled.
 *
 * @field clean_d_cache Function pointer. Push all dirty lines from D-caches
 *                      into main memory. Disable D-caches if argument
 *                      'disable' is TRUE.
 *
 * @input clean_d_cache.disable True if d cache should be disabled after
 *                              cleaning.
 *
 * @field cache_info    CPU specific cache information,
 *                      see struct @link "" "cache_info".
 *
 * @field entry_lo      Cache coherency attribute,
 *                      see struct @link "" "entry_lo".
 *
 * @field cache Function pointer. Act (flush or invalidate) on the cache
 *                                system to a defined cache level.
 *
 * @input cache.action Action performed on the cache system.
 *
 * @input cache.level Action will take place from CPU to this level.
 *
 * @input cache.base Base address where the action should start.
 *
 * @input cache.size Number of bytes that should be acted on.
 *
 * @field cache_level Function pointer. Returns the physical cache level
 *                                      that corresponds to a logical
 *                                      cache level.
 *
 * @input cache_level.level Request level (logical).
 *
 */
struct arch_cpu_descriptor
{
   U32 exec_attributes;
   U32 mm_attributes;
   void (*init_cpu)(void);
   void (*enable_caches)(U32 enable_d_cache,  /* Non-zero if enable */
                         U32 enable_i_cache); /* Non-zero if enable */
   void (*clean_d_cache)(U32 disable);
   struct cache_info cache_info;
   struct entry_lo entry_lo;
   void (*cache)(U32 action, U32 level, U32 base, U32 size);
   U32  (*cache_level)(U32 level);
};

extern struct arch_cpu_descriptor *hal_arch_cpu_descriptor;

#ifdef __cplusplus
}
#endif

#endif /* #ifndef CPU_HAL_ARCH_H_ */
