/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2007 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */
/**
 * @toc CPU_Dependent:
 * @file ose_spi/mips/cpu_intlock.h
 *
 * @brief Interrupt lock functions (mips).
 *
 * @long The lock functions must only be called from supervisor mode. 
 *       The effect of calls from user mode (or problem state) is unspecified.
 */
#ifndef _CPU_INTLOCK_H
#define _CPU_INTLOCK_H

#ifdef __cplusplus
extern "C" {
#endif

/** 
 * @type Msr
 *
 * @brief Type that can hold a Machine Status Register. 
 */
typedef unsigned long Msr;

/**
 * @macro LOCK_SAVE
 *
 * @brief Disable interrupts and get previous state.
 *
 * @long  LOCK_SAVE is a macro, do not use & operator on the argument.
 *        LOCK_SAVE requires supervisor mode.
 *
 * @output  oldMsr   Will be set to the previous MSR value.
 */

#define LOCK_SAVE(oldMsr)   (oldMsr) = LOCK_SAVE_()
/**
 * @macro IRQ_ENABLED
 * @input msr The msr value to be examined.
 *
 * @brief Evaluates to a nonzero value if interrupts are enabled according to
 * the msr parameter.
 */
#define IRQ_ENABLED(msr) ((msr) & 1)

#ifndef SOURCEDOC
#ifdef _MIPS_ARCH_MIPS3
static __inline__ unsigned long
LOCK_SAVE_(void)
{
    unsigned long sr;
    __asm__ volatile
    (
        ".set push\n\t"
        ".set noreorder\n\t"
        " mfc0	%0,$12\n\t"
        " addiu	$3,$0,0xfffe\n\t" /* 0xfffffffe */
        " and	$3,%0,$3\n\t"
        " mtc0	$3,$12\n\t"
	" nop\n\t"		/* Wait for disable to take effect. */ \
	" nop\n\t"		/* Three number of nops recommended */ \
	" nop\n\t"		/* in MIPS manuals, CP0 hazards.    */ \
        ".set pop\n\t"
        : /*outputs*/ "=r" (sr)
        : /*no inputs*/
        : /*trashes*/ "$3"
    );
    return sr;
}

static __inline__ void
LOCK_RESTORE(unsigned long sr)
{
    __asm__ volatile
    (
       ".set push\n\t"	         \
       ".set noreorder\n\t"	 \
       " mfc0    $8,$12\n\t"     \
       " andi    %0,%0,1\n\t"    \
       " or      %0,%0,$8\n\t"   \
       " mtc0 %0,$12\n\t"        \
       " nop\n\t"		/* Extra precaution to avoid having    */ \
       ".set pop\n\t"	        /* sequence mtco followed by mfco.     */ \
        : /*no outputs*/         \
        : /*inputs*/ "r" (sr)    \
        : /*trashes*/ "$8"
    );
}
#else
static __inline__ unsigned long
LOCK_SAVE_(void)
{
    unsigned long sr;
    __asm__ volatile
    (
        ".set push\n\t"
        ".set noreorder\n\t"
        " mfc0	%0,$12\n\t"
        " addiu	$3,$0,0xfffe\n\t" /* 0xfffffffe */
        " and	$3,%0,$3\n\t"
        " mtc0	$3,$12\n\t"
	" ssnop\n\t"		/* Wait for disable to take effect. */ \
	" ssnop\n\t"		/* Three number of nops recommended */ \
	" ssnop\n\t"		/* in MIPS manuals, CP0 hazards.    */ \
        ".set pop\n\t"
        : /*outputs*/ "=r" (sr)
        : /*no inputs*/
        : /*trashes*/ "$3"
    );
    return sr;
}

static __inline__ void
LOCK_RESTORE(unsigned long sr)
{
    __asm__ volatile
    (
       ".set push\n\t"	         \
       ".set noreorder\n\t"	 \
       " mfc0    $8,$12\n\t"     \
       " andi    %0,%0,1\n\t"    \
       " or      %0,%0,$8\n\t"   \
       " mtc0 %0,$12\n\t"        \
       " ssnop\n\t"		/* Extra precaution to avoid having    */ \
       ".set pop\n\t"	        /* sequence mtco followed by mfco.     */ \
        : /*no outputs*/         \
        : /*inputs*/ "r" (sr)    \
        : /*trashes*/ "$8"
    );
}

#endif
#else /* Only for SOURCEDOC */
/**
 * @macro LOCK_RESTORE
 *
 * @brief Restore old interrupt state from the argument.
 *		
 * @long LOCK_RESTORE requires supervisor mode.
 *
 * @input oldMsr	The old MSR value, which now is restored.
 *
 */
#define LOCK_RESTORE(oldMsr)   LOCK_RESTORE_DUMMY_NOT_USED_(oldMsr)

#endif /* ifndef SOURCEDOC */

#ifdef __cplusplus
}
#endif

#endif /* # ifndef _CPU_INTLOCK_H */
