/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2006 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */
/**
 * @toc CPU_Dependent:
 * @file ose_spi/mips/cpu_regdump.h
 *
 * @brief This header defines the register dump used to report 
 *        Unexpected Exceptions (MIPS).
 *
 * @long Register Dump Struct, used to report Unexpected Exceptions.
 *       The address of the struct ose_regdump is passed in the 'extra' 
 *       parameter to the Kernel Error Handler. 
 */
#ifndef _CPU_REGDUMP_H
#define _CPU_REGDUMP_H

#include "stdint.h"
#include "osetypes.h"

#ifdef __cplusplus
extern "C" {
#endif

/*DEFINE_START*/

/**
 * @macro CPU_REGDUMP_MAGIC
 *
 * @brief Current magic number (MIPS).
 */   
#define CPU_REGDUMP_MAGIC 0x0001daea

/*
 * Exception id's
 */
/**
 * @group EXCEPTION_vectorvalues_MIPS
 *
 * @brief MIPS vector values used in OSE.
 *
 * @long  General exceptions use ExcCode values 0 - 31 in the cause register
 *        to indicate the exception type. The ExcCode vector values are the
 *        EXCEPTION_ID_x macros in this header file.
 *
 * @long  There are also vector values >31 in OSE, used for exceptions with
 *        dedicated exception vector offsets. These dedicated vector
 *        values are the EXCEPTION_VEC_x macros in this header file.
 *        The values are not offsets.
 *
 * @long  The example given for the struct cpu_regdump for MIPS is an extract
 *        from a typical system error handler, to decode the
 *        exception vector values in MIPS. See struct cpu_regdump for MIPS.
 */
#define EXCEPTION_ID_INT          0
#define EXCEPTION_ID_MOD          1
#define EXCEPTION_ID_TLBL         2
#define EXCEPTION_ID_TLBS         3
#define EXCEPTION_ID_ADEL         4
#define EXCEPTION_ID_ADES         5
#define EXCEPTION_ID_IBE          6
#define EXCEPTION_ID_DBE          7     
#define EXCEPTION_ID_SYS          8
#define EXCEPTION_ID_BP           9
#define EXCEPTION_ID_RI          10
#define EXCEPTION_ID_CPU         11
#define EXCEPTION_ID_OV          12
#define EXCEPTION_ID_TR          13
#define EXCEPTION_ID_FPE         15
#define EXCEPTION_ID_C2E         18
#define EXCEPTION_ID_MDMX        22
#define EXCEPTION_ID_WATCH       23
#define EXCEPTION_ID_MCHECK      24
#define EXCEPTION_ID_CACHEERR    30
/* Vectors for dedicated exceptions. */

/* Offset 0x000 */
#define EXCEPTION_VEC_TLB_REFILL 32

/* Offset 0x080 */
#define EXCEPTION_VEC_XTLB       33

/* Offset 0x100 */
#define EXCEPTION_VEC_CACHEERROR 34

/*DEFINE_STOP*/

/**
 * @type cpu_core
 *
 * @brief User model registers (MIPS).
 * 
 * @field r     Array containing the 32 General Purpose Registers, including SP, FP, RA
 * @field lo    Special register LO
 * @field hi    Special register HI
 * @field epc   Program counter at last exception
 * @field sr    Status register
 * @field cause Cause of last general exception, including ExcCode
 */
struct cpu_core
{
   /*INC_PREFIX cpu_core_ */
   /*INC_START*/
   uint64_t r [ 32 ] ;

   uint64_t lo;
   uint64_t hi;

   uint64_t epc;
   U32 sr;
   U32 cause;
#if 0
   U32 size;	/* Used only by .inc file to size. */
#endif
   /*INC_STOP*/
};


/**
 * @type cpu_dumpsection
 *
 * @brief CPU dump section (MIPS).
 *
 * @field identity Identifies contents of this dump part.
 *                 Zero for a set of basic processor core
 *                 registers, saved in form of a struct 
 *                 cpu_core for MIPS.
 *                 Other identities may be allocated to 
 *                 provide more detailed or processors 
 *                 specific information in the future.
 *
 * @field size     Size of dump part, in bytes.
 *
 * @field address  Address of dump part (word aligned).
 */
struct cpu_dumpsection          /* Dump section */
{
  U32   identity;
  U32   size;
  U32   address;
};

/**
 * @type cpu_regdump
 *
 * @brief CPU register dump (MIPS).
 *
 * @field pc     Program counter at exception.
 * @field sp     Stack pointer at exception.
 * @field sr     Status register before exception.
 * @field vector Exception Vector in MIPS.
 *               vector 0-31 is the ExcCode in the CAUSE register for
 *               general exceptions.
 *               vector >31 correspond to other exceptions with dedicated
 *               address offsets.
 *               See also @link "" "EXCEPTION_vectorvalues_MIPS".
 *               The example below is an extract from a typical system
 *               error handler, decoding exception vector values.
 *
 * @field fault_data_set   Nonzero if fault_data_addr is set up.
 * @field fault_data_addr  Faulting address of data exceptions.
 * @field magic            Magic number, to handle compatibility.
 * @field no_sects         Number of parts.
 * @field sect             Basic register dump information 
 *                         (and in the future, more detailed or 
 *                         processor dependent information).
 *                         The field type is cpu_dumpsection for MIPS.
 *
 * @example
 * // Example of decoding MIPS exception vectors in OSE //
 * // NOTE: This is an example! See the vector macros and use //
 * //       the values for the actual architecture. //
 *
 * static const char *exception_name(U32 vector) {
 *  switch (vector)
 *     {
 *     // Standard MIPS exceptions. //
 *     // vector 0-31 is the ExcCode in the CAUSE register for general //
 *     // exceptions. //
 *     // vector >31 correspond to other exceptions with dedicated vector //
 *     // offsets. //
 *         case 0  : return "interrupt";
 *         case 1  : return "TLB modification";
 *         case 2  : return "TLB (load or instruction fetch)";
 *         case 3  : return "TLB (store)";
 *         case 4  : return "address error (load or instruction fetch)";
 *         case 5  : return "address error (store)";
 *         case 6  : return "bus error (instruction fetch)";
 *         case 7  : return "bus error (data reference: load or store)";
 *         case 8  : return "system call";
 *         case 9  : return "breakpoint";
 *         case 10 : return "reserved instruction";
 *         case 11 : return "coprocessor unusable";
 *         case 12 : return "arithmetic overflow";
 *         case 13 : return "trap";
 *         case 15 : return "floating point";
 *         case 18 : return "coprocessor 2";
 *         case 22 : return "MDMX unusable";
 *         case 23 : return "watch";
 *         case 24 : return "machine check";
 *         case 30 : return "cache error";
 *         // Values used for dedicated vectors //
 *         case 32 : return "TLB refill (vector offset 0)";
 *         case 33 : return "XTLB refill (vector offset 0x80)";
 *         case 34 : return "cache error (vector offset 0x100)"; 
 *     }
 * }
 */
struct cpu_regdump              /* Register Dump */
{
  U32   pc;
  U32   sp;
  U32   sr;
  U32   vector;
  U32   fault_data_set;
  U32   fault_data_addr;
  U32   magic;
  U32   no_sects;               /* Number of parts (Currently one) */
  struct cpu_dumpsection sect[1];
};

#ifdef __cplusplus
}
#endif

#endif /* # ifndef _CPU_REGDUMP_H */
