/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2011 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */
/**
 * @public
 * @toc Memory_Management:
 * @file ose_spi/mm.h
 *
 * @brief MM function interface accessible from CRT/LM link scope.
 *
 * @long
 * This file defines a functional interface for the most common
 * usage cases regarding block memory allocations from MM. It hides
 * the log_mem name for system ram in SAS/SASE configurations which
 * simplifies for clients like DDA, etc.
 */

#ifndef OSE_MM_H_
#define OSE_MM_H_

#include "ose.h"
#include "string.h"
#include "mm.sig"
#include "sys/types.h"
#include "sys/mman.h"

#ifdef __cplusplus
extern "C" {
#endif

/**
 * @macro MM_ALLOW_VIRTUAL
 * @brief A flag indicating that SAS memory is allowed.
 * @long  If this option is given, the allocation is
 *        performed from SAS memory if available. If the
 *        option is not given, the memory will be 1:1 mapped.
 */
#define MM_ALLOW_VIRTUAL 0x1

/**
 * @macro MM_ALLOW_CACHED
 * @brief A flag indicating that caching is allowed.
 * @long  If this option is given, the region is marked
 *        using copy_back cache mode. If it is not given,
 *        the mode "inhibit" is used.
 */
#define MM_ALLOW_CACHED 0x2

/**
 * @macro MM_IO_MAPPING
 * @brief A flag indicating that the mmap concerns I/O mapping.
 * @long  If this option is given, the region is mapped 1:1 to
 *        device memory.
 */
#define MM_IO_MAPPING 0x4

/**
 * @function ose_mm_sysram_name
 *
 * @brief   This function returns the logmem name for the 1:1 mapped RAM
 *          memory (SASE).
 * @long    Depending on the configuration (SAS/SASE), the name of the logical
 *          memory that maps ram memory 1:1 varies. This function simplifies
 *          for those that do not bother with the name but wants to operate on
 *          1:1 mapped RAM memory.
 * @return  The log_mem name for the 1:1 mapped ram memory. If MM is
 *          unavailable, NULL is returned.
 *
 * @seealso default_restrictions_calling_functions
 */
const char *ose_mm_sysram_name(void);

/**
 * @function ose_mm_alloc
 *
 * @brief This function is used to allocate a supervisor mode 4k aligned
 *        memory region in the system domain with the access permissions
 *        SU_RW_USR_RO.
 *
 * @input  region_name    Specifies the name of the logical memory region.
 * @input  size           Specifies the size of the region.
 * @input options         option field, see above.
 *
 * @return                The logical address of the region, or zero if the
 *                        operation failed
 *                        (MM will never choose zero as a valid address).
 *
 * @seealso default_restrictions_calling_functions
 */
OSADDRESS ose_mm_alloc(const char *region_name, size_t size, U32 options);

/**
 * @function ose_mm_free
 *
 * @brief This function is used to free a memory region allocated with
 *        ose_mm_alloc.
 * @input addr     Specifies the address of the memory region.
 * @return         Zero if OK, otherwise the MM error code.
 * @seealso default_restrictions_calling_functions
 */
mm_status_t ose_mm_free(OSADDRESS addr);

/* EXPERIMENTAL */
/**
 * @function ose_mmap
 *
 * @brief This function is a more feature rich version of ose_mm_alloc,
 *        with the option to give access rights and optional base address as
 *        well. It automatically finds the log_mem name if base is given.
 *        It always uses system domain!
 *
 * @input  region_name    Specifies the name of the logical memory region.
 * @input  addr           Specifies the desired base of the region. If zero,
 *                        the MM can choose an arbitrary address.
 * @input  size           Specifies the size of the region.
 * @input access          Access permissions, see mm.sig.
 * @input options         Options, see above.
 * @output status         The MM return code (MM_SUCCESS if return != zero.)
 *                        If return == zero, the status shows the MM error.
 *
 * @return                The logical address of the region, or zero if the
 *                        operation failed (MM will never choose zero as a
 *                        valid address).
 *                        On failure a RAMLOG printout is produced.
 *                        In case of a given base address do check the return
 *                        address.
 *
 * @seealso default_restrictions_calling_functions
 */
OSADDRESS ose_mmap(const char *region_name,
                   OSADDRESS addr,
                   size_t size,
                   U32 access,
                   U32 options,
                   mm_status_t *status);

/**
 * @macro MM_VALIDATE_IN_MAS
 * @brief A flag indicating buffer validation in MAS memory.
 * @long  If this option is given, ose_mm_validate_buffer will
 * also check if any part of the buffer resides in MAS mapped memory.
 */
#define MM_VALIDATE_IN_MAS 0x1

/**
 * @function ose_mm_validate_buffer
 *
 * @brief Validates a buffer access without accessing it.
 *
 * @long Size validation is currently only supported on PowerPC, ARM and MIPS.
 *       For all other architectures, the returned number of bytes equals
 *       the requested number of bytes.
 *
 * @input base        The base address of the buffer to be validated.
 *
 * @input size        The size of the buffer to be validated.
 *                    To determine how many bytes that can be accessed from
 *                    a given base address, set a reasonable large size value.
 *                    The maximum value (size_t)-1 might cause long
 *                    validations and could therefor affect performance.
 *
 * @input pid         The process context for which the validation
 *                    should be made. 0 = current process.
 *
 * @input perm        The permission to validate. Should be set to either of
 *                    MM_PERM_READ or MM_PERM_WRITE or OR'ed together for
 *                    validation of both.
 *
 * @input options     Validation options.
 *                    The following validation options are supported:
 *                    MM_VALIDATE_IN_MAS
 *
 * @output result     Result of validation options.
 *                    The requested options are set in the result field
 *                    if evaluated to true.
 *
 * @return            Returns the max number of bytes where validation
 *                    is successful.
 *
 * @seealso default_restrictions_calling_functions
 */
size_t ose_mm_validate_buffer(log_addr_t base,
			      size_t size,
			      PROCESS pid,
			      unsigned int perm,
			      unsigned int options,
			      unsigned int *result);

/**
 * @function ose_mm_l2p
 *
 * @brief Acquire the physical address of the specified
 *        logical address for the specified PID.
 *
 * @long   A MM utility function needed by Pager functionality in order to
 *         execute page faults in the context of the faulting client.
 *           - Not available in SMP lib.
 *
 *
 * @input  pid        Specifies the process ID. 0 = current process.
 * @input  la         Specifies the logical address.
 * @output pa         Physical address that corresponds to given la.
 * @return            Zero if OK, otherwise the MM error code.
 *
 * @seealso default_restrictions_calling_functions
 */
mm_status_t
ose_mm_l2p(PROCESS     pid,
           log_addr_t  la,
           phys_addr_t *pa);


/**
 * @function ose_mm_l2p_range
 *
 * @brief Acquire the physical address of the specified
 *        logical address for the specified PID, as well as the
 *        range of contiguous physical memory.
 *
 * @long  Note that this function can take considerably longer time
 *        to complete its task than ose_mm_l2p, depending on the
 *        size of the physical pages that constitute the memory region.
 *        This is especially the case if a pointer to the value 0
 *        is passed as range parameter, i.e. unlimited range lookup.
 *        - Not available in SMP lib.
 *
 *
 * @input  pid        Specifies the process ID. 0 = current process.
 * @input  la         Specifies the logical address.
 * @output pa         Physical address that corresponds to given la.
 * @input  range      Pointer to expected range of contiguous physical memory.
 *                    If 0, the lookup continues until the physical
 *                    memory is no longer contiguous or end of memory
 *                    is reached. A NULL pointer indicates behavior
 *                    equal to ose_mm_l2p.
 * @output range      Updated with the actual range of
 *                    contiguous physical memory at la and beyond.
 * @return            Zero if OK, otherwise the MM error code.
 *
 * @seealso default_restrictions_calling_functions
 */
mm_status_t
ose_mm_l2p_range(PROCESS     pid,
                 log_addr_t  la,
                 phys_addr_t *pa,
                 size_t      *range);


/**
 * @function ose_mm_get_cache_mode
 *
 * @brief Compile a cache and a memory type name into a 
 *        cache mode bitfield. 
 *
 * @long  The cache mode bitfield can be used directly in OSE MM
 *        external interfaces, such as the mm_create_region signal.
 *
 * @input  cache_type   The requested cache type. Mandatory.
 * @input  memory_type  The requested memory type. Mandatory.
 * @output cache_mode   Compiled cache mode bitfield. Mandatory.
 *
 * @return
 *    @tab MM_SUCCESS
 *       @tab Cache mode was successfully compiled. @break
 *    @tab MM_EILLEGAL_CACHE_MODE
 *       @tab The cache or memory type does not exist,
 *            or their combination is illegal.  @break
 *
 * @seealso ose_mm_get_cache_type
 * @seealso default_restrictions_calling_functions
 */
mm_status_t
ose_mm_get_cache_mode(const char *cache_type,
                      const char *memory_type,
                      U32 *cache_mode);

/**
 * @function ose_mm_get_cache_type
 *
 * @brief Decompile a cache mode bitfield into the original name of
 *        the cache and memory type.
 *
 * @input  cache_mode   The cache mode bitfield to decompile.
 * @output cache_type   Name of the cache mode cache type. May be NULL
 *                      if this string is not desired, otherwise
 *                      the string must be able to hold at least
 *                      MM_NAME_LENGTH + 1 characters.
 * @output memory_type  Name of the cache mode memory type. May be NULL
 *                      if this string is not desired, otherwise
 *                      the string must be able to hold at least
 *                      MM_NAME_LENGTH + 1 characters.
 *
 * @return
 *    @tab MM_SUCCESS
 *       @tab Cache mode was successfully decompiled. @break
 *    @tab MM_EILLEGAL_CACHE_MODE
 *       @tab The cache mode bitfield is illegal. @break
 *
 * @seealso ose_mm_get_cache_mode
 * @seealso default_restrictions_calling_functions
 */
mm_status_t
ose_mm_get_cache_type(U32 cache_mode,
                      char *cache_type,
                      char *memory_type);

#ifdef __cplusplus
}
#endif

#endif /* #ifndef OSE_MM_H_ */
