/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2012 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */
/**
 * @public
 * @toc Memory_Management:
 * @file ose_spi/mm.sig
 * @version All rights reserved.
 *
 * @brief Public signal interface to OSE Memory Manager.
 * @long Signals should be sent to the process named ose_memory_mgr.
 * @long A description of this signal interface can be found in the
 *       Memory Manager (MM) section in the "OSE Core User's Guide".
 *       This section also contains general information
 *       about the OSE Memory Manager and how memory is used by the OSE
 *       executive and program. For fundamentals and memory management
 *       in the core operating system and for programming for
 *       embedded systems, as well as information about program execution
 *       models (private/shared) mode, see the "OSE Architecture User's Guide".
 *
 * @long The main groups of entries in this interface are the following:
 *
 * @long - Create/delete regions or change permissions or
 *       share regions or making regions resident, signals:
 *       MM_CREATE_REGION_*, MM_DELETE_REGION_*, MM_CHANGE_PERMISSIONS_*,
 *       MM_SHARE_REGION_*, MM_MAKE_*.
 * @long - Create/delete domains, signals:
 *       MM_CREATE_DOMAIN_*, MM_DELETE_DOMAIN_*.
 * @long - Get lists or information about regions,
 *       domains, log_mem, phys_mem, page sizes, etc., signals:
 *       MM_LIST_*, MM_GET_*_INFO_*, MM_GET_MAX_*.
 * @long - Get MMU statistics, signals:
 *       MM_GET_MMU_STATISTICS_*.
 * @long - Translate a logical address in a domain to a physical
 *       address, signals:
 *       MM_TRANSLATE_*.
 * @long - Get information about the MM interface and MMU usage, signals:
 *       MM_INTERFACE_INFO_*.
 *
 *
 * @long - Enum with reply status codes in replies from MM:
 *       @link "" "MmStatus" enum with many MM_E* and MM_SUCCESS.
 * @long - Enum with information TAGs used in the MM interface
 *       @link "" "MmTagId" enum with many MM_INFOTAG_*, MM_TAG_INFO_*.
 * @long - Macros for access permissions and cache modes for regions:
 *       SU_*_USR_*, MM_USR_*, MM_SU_*,
 *       MM_CACHE_*, MM_SPECULATIVE_*. MM_ORDERED_*.
 * @long - Enum with mapping types for logical memory areas:
 *       @link "" "MappingType" enum with e.g. SASE, SAS, MAS.
 * @long - Macros for the MM process name (GMM_NAME_STRING) and the
 *       mandatory "RAM" physical memory type identifier (MM_INTERNAL_RAM),
 *       and other constants.
 *
 * @seealso ose_spi/mmtypes.h
 */

#ifndef _MM_SIG
#define _MM_SIG

#ifdef __cplusplus
extern "C" {
#endif

#include "ose.h"
#include "osetypes.h"
#include "mmtypes.h"

/* Checks for the matching version of mmtypes.h */
#if _MMTYPES_H != 500
#error ==========================================================
#error || Wrong version of mmtypes.h                           ||
#error ==========================================================
#endif

/**************************************************************************/
/**
 *
 * @type MmStatus
 * @brief Enum containing Error Codes in the MM signal interface.
 *
 * @long The reply signals, returned from the memory manager, can contain
 * these error codes in the status field.
 *
 * @field MM_SUCCESS
 *   A status value used in reply signals that
 *   indicates a successful operation.
 *
 * @field MM_ESIGNAL_NOT_IMPLEMENTED
 *   Functionality to handle the request of the signal has not been
 *   implemented yet to support this CPU/architecture.
 *
 * @field MM_ENON_EXISTING_REGION
 *   The specified base address and domain does not
 *   specify an existing region.
 *
 * @field MM_ENON_EXISTING_DOMAIN
 *   The specified domain does not exist.
 *
 * @field MM_ENON_EXISTING_PARTITION
 *   The specified logical memory area does not exist.
 *   This error code should not be used since it will not be maintained
 *   in the future. Use MM_ENON_EXISTING_LOG_MEM instead.
 *
 * @field MM_ENON_EXISTING_LOG_MEM
 *   The specified logical memory area does not exist.
 *
 * @field MM_ENON_EXISTING_PHYS_TYPE
 *   The specified physical memory type does not exist.
 *   This error code should not be used since it will not be maintained
 *   in the future. Use MM_ENON_EXISTING_PHYS_MEM instead.
 *
 * @field MM_ENON_EXISTING_PHYS_MEM
 *   The specified physical memory does not exist.
 *
 * @field MM_EILLEGAL_AREA
 *   The specified area (base and size) is illegal.
 *
 * @field MM_ESIZE_IS_ZERO
 *   The size must be greater than zero.
 *
 * @field MM_EOUT_OF_LOG_MEM
 *   Enough logical memory could not be found in the specified
 *   logical memory area.
 *
 * @field MM_EOUT_OF_PHYS_MEM
 *   Enough physical memory could not be found in the specified
 *   physical memory type.
 *
 * @field MM_EOVERLAPPING_ADDRESS
 *   The region could not be created at the requested address, since
 *   it would overlap another region.
 *
 * @field MM_EOUTSIDE_LOG_MEM
 *   The region could not be created at the requested address, since
 *   the address is outside the logical memory area specified.
 *
 * @field MM_EMISMATCH_FRAGMENTS
 *   The total size of the physical fragments did not match the size
 *   of the region to create.
 *
 * @field MM_ENO_WRITE_PERMISSION
 *   No write permissions in the specified region.
 *
 * @field MM_EOUTSIDE_REGION
 *   Write outside region.
 *
 * @field MM_ENO_PHYSICAL
 *   The specified region did not have any physical resources
 *   and can not be shared.
 *
 * @field MM_EOUT_OF_DOMAINS
 *   The maximum number of domains has been reached.
 *   For more information about the maximum number of domains supported
 *   by the system, see @link "mm.sig" "MM_GET_MAX_DOMAINS_REQUEST".
 *
 * @field MM_EILLEGAL_OPERATION
 *   The requested operation is not allowed.
 *
 * @field MM_EOPTION_NOT_SUPPORTED
 *   The options specified in the options field are not supported.
 *
 * @field MM_EILLEGAL_PERMISSION
 *   The specified permissions are not supported. There are two reasons
 *   for this: either the permission parameter has not been correctly
 *   constructed using the predefined macro constants or user mode has more
 *   permissions than supervisor mode.
 *   For MIPS-32, trying to create a region with user-mode permissions
 *   outside KUSEG on MIPS is illegal.
 *
 * @field MM_EILLEGAL_CACHE_MODE
 *   The specified cache mode contains illegal combinations.
 *   For more information about valid combinations, see
 *   @link "mm.sig" "MM_CREATE_REGION_REQUEST".
 *
 * @field MM_EILLEGAL_BASE_ADDRESS
 *   The specified base address is not correctly aligned.
 *
 * @field MM_EWRITE_FAILED
 *   The write operation failed on the physical media.
 *
 * @field MM_EILLEGAL_TAG
 *   The specified tag has not been implemented yet.
 *
 * @field MM_EREGION_LOGICAL
 *   The requested operation could not be carried out on a logical region.
 *
 * @field MM_ESYSTEM_REGION
 *   The operation could not be carried out since the specified region
 *   is a system region. System regions are also knows as static regions
 *
 * @field MM_ESYSTEM_DOMAIN
 *   The operation could not be carried out since the specified domain
 *   is a system domain.
 *
 * @field MM_EWRITE_NOT_IMPLEMENTED
 *   The write operation failed because the specified physical memory type
 *   does not allow write.
 *
 * @field MM_EUSER_SUPPLIED_PHYSICAL
 *   Write operation is not allowed for a region that was created with
 *   physical pages supplied by the user.
 *
 * @field MM_ETOO_MANY_INSTANCES
 *   The operation failed as a conflict related to shared regions occurred.
 *
 * @field MM_EINTERNAL_ERROR
 *   An internal error has occurred in MM.
 *
 * @field MM_EOVERLAPPING_REMAP_ENTRY
 *   The remap operation failed since the area was already remapped.
 *   Note that remap functionality is neither implemented nor supported.
 *
 * @field MM_ENON_EXISTING_REMAP_ENTRY
 *   The unremap operation failed since no remap area exists.
 *   Note that remap functionality is neither implemented nor supported.
 *
 * @field MM_EILLEGAL_DOMAIN
 *   The domain identifier is valid but the operation is illegal.
 *
 * @field MM_EMISMATCH_PERSISTENCE
 *   The persistency configuration option does not match the properties of
 *   the physical memory, i.e. a volatile region can not be created from
 *   flash memory.
 *
 * @field MM_EILLEGAL_PHYS_TYPE
 *   The physical memory type is not allowed in the operation.
 *   This error code should not be used since it will not be maintained
 *   in the future. Use MM_EILLEGAL_PHYS_MEM instead.
 *
 * @field MM_EILLEGAL_PHYS_MEM
 *   The physical memory type is not allowed in the operation.
 *
 * @field MM_EPHYS_FRAG_OVERLAP_PHYS_MEM
 *   The physical fragments overlaps a phys_mem.
 *
 * @field MM_EDOMAIN_NOT_MAS_ACTIVE
 *   The domain is not yet activated in exec.
 *
 * @field MM_EMAS_NOT_SUPPORTED
 *   Multiple address Space (MAS) is not supported for this
 *   operation/architecture.
 *
 * @field MM_EMAS_ATTRIBUTE_MISMATCH
 *   If another MAS region exist with same base and size the second region
 *   must have same regions attributes i.e. caching/permission.
 *
 * @field MM_EMAS_DATA_ONLY
 *   MAS is only supported for data regions.
 *
 * @field MM_ESHARE_UMODE_RESTRICT_FAILED
 *   The requested user mode restriction in the destination domain could
 *   not be used.
 *
 * @field MM_ETOO_MANY_PERSISTENT_REGIONS
 *   The max number of persistent regions was exceeded.
 *
 *
 * @field    MM_EPMM_OUT_OF_MEMORY
 *   The region used for storage of persistent region metadata is full.
 *   If needed increase the RPMM region size in the configuration.
 *
 * @field    MM_EPAGER_PAGEABLE_AREA
 *   An invalid page able area was given or configuration for the
 *   page able area is missing.
 *
 * @field    MM_ENO_PHYSICAL_MAPPING
 *   No physical mapping has been found.
 *
 * @field    MM_EUNAUTHORISED_OPERATION
 *   The caller has not the right to perform the operation.
 *
 * @field    MM_ENON_EXISTING_EVENT
 *   No event of this type exists.
 *
 * @field    MM_ENON_EXISTING_EVENT_HANDLER
 *   This event handler does not exist.
 *
 * @field    MM_EPHYS_FRAG_NO_LOGMEM
 *   When doing a duplicate mapping the supplied phys fragment is not part of
 *   any log_mem.
 *
 * @field    MM_EPHYS_FRAG_ILLEGAL_SASE
 *   When doing a duplicate mapping the SASE region is not SU_NA_USR_NA.
 *
 * @field    MM_EPHYS_FRAG_OUTSIDE_SASE
 *   When doing a duplicate mapping the supplied phys_resource is not part of
 *   any SASE region.
 *
 * @field    MM_EPHYS_FRAG_NOT_SAME_SASE
 *   When doing a duplicate mapping the supplied phys_resource is not part of
 *   the same SASE region.
 *
 * @field    MM_EPHYS_FRAG_NO_VALID_SASE
 *   Reserved for future use.
 *
 * @field    MM_EPHYS_FRAG_TOO_BIG
 *  When doing a duplicate mapping the size of the supplied phys_resource is
 *  larger than the SASE region.
 *
 * @field    MM_EPHYS_FRAG_MISMATCH
 *   When doing a duplicate mapping the number of supplied phys_resources is
 *   less than phys_count states.
 *
 * @field    MM_EILLEGAL_SASE_OPTION
 *   An illegal option for SASE region was given.
 *
 * @field    MM_EILLEGAL_OPTION_COMBINATION
 *   An illegal combination of options was given.
 *
 * @field    MM_EOPTION_DOMAIN_MISMATCH_MAS
 *   An illegal option/domain combination for MAS region was given.
 *
 * @field    MM_EOUT_OF_RESOURCES
 *   The MM is out of resources to meet the request.
 *
 * @field    MM_EVENT_POLICY_ALREADY_INSTALLED
 *   A policy has already been installed for this event.
 *
 * @field    MM_EVENT_POLICY_NOT_INSTALLED
 *   A policy has not been installed for this event.
 *
 * @field    MM_ENO_READ_PERMISSION
 *   No read permissions in the specified region.
 *
 * @field    MM_EINFOTAG_UNSUPPORTED_FOR_CPU
 *   Requested MM_INFOTAG's are valid but
 *   not supported for this architecture/CPU.
 *
 * @field    MM_ETOO_SMALL_SIGNAL_SIZE
 *   MM received a signal which size is too small to contain requested
 *   information. One example is a signal requests using struct mm_tag with
 *   incorrect count field.
 *   @seealso mm_tag
 *   @seealso mm_tag_ext
 *   @seealso mm_get_mmu_statistics
 *
 * @field    MM_EINFOTAG_ONLY_SUPPORTED_CPU0
 *   Requested MM_INFOTAG only supported for cpu0.
 *
 * @field    MM_ENO_INFOTAG_COUNTER_SUPPORT_FOR_CPU
 *   No MM_INFOTAG counter support exists for this CPU.
 *
 * @field    MM_ENOT_GMM_PROCESS
 *  An internal MM function was executed by other process than GMM.
 *
 * @field    MM_EILLEGAL_ALIGNMENT
 *  The specified alignment is not supported.
 *
 * @field    MM_EILLEGAL_CPU_ID
 *  The specified cpu_id is out of range.
 *
 * @field    MM_EINFOTAG_ONLY_LOCAL_COUNTER
 *  The tag_id was requested for a local counter but only a global counter
 *  exists.
 *
  * @field   MM_EINFOTAG_ONLY_GLOBAL_COUNTER
 *  The tag_id was requested for a global counter but only a local counter
 *  exists.
 *
 * @field    MM_EILLEGAL_ARGUMENT
 *  An argument provided to a function was illegal (e.g. NULL pointer). 
 *
 */

enum MmStatus
{
   MM_SUCCESS = 0x0,
   MM_ESIGNAL_NOT_IMPLEMENTED = 0x1,
   MM_ENON_EXISTING_REGION = 0x2,
   MM_ENON_EXISTING_DOMAIN = 0x3,
   MM_ENON_EXISTING_PARTITION = 0x4,
   MM_ENON_EXISTING_LOG_MEM = 0x4,
   MM_ENON_EXISTING_PHYS_TYPE = 0x5,
   MM_ENON_EXISTING_PHYS_MEM = 0x5,
   MM_EILLEGAL_AREA = 0x6,
   MM_ESIZE_IS_ZERO = 0x7,
   MM_EOUT_OF_LOG_MEM = 0x8,
   MM_EOUT_OF_PHYS_MEM = 0x9,
   MM_EOVERLAPPING_ADDRESS = 0xC,
   MM_EOUTSIDE_LOG_MEM = 0xD,
   MM_EMISMATCH_PERSISTENCE = 0xF,
   MM_ENO_WRITE_PERMISSION = 0x10,
   MM_EOUTSIDE_REGION = 0x11,
   MM_ENO_PHYSICAL = 0x12,
   MM_EOUT_OF_DOMAINS = 0x13,
   MM_EILLEGAL_OPERATION = 0x14,
   MM_EOPTION_NOT_SUPPORTED = 0x15,
   MM_EILLEGAL_PERMISSION = 0x16,
   MM_EILLEGAL_CACHE_MODE = 0x17,
   MM_EILLEGAL_BASE_ADDRESS = 0x18,
   MM_EWRITE_FAILED = 0x19,
   MM_EILLEGAL_TAG = 0x1A,
   MM_EREGION_LOGICAL = 0x1C,
   MM_ESYSTEM_REGION = 0x1D,
   MM_ESYSTEM_DOMAIN = 0x1E,
   MM_EWRITE_NOT_IMPLEMENTED = 0x1F,
   MM_EUSER_SUPPLIED_PHYSICAL = 0x20,
   MM_ETOO_MANY_INSTANCES = 0x21,
   MM_EILLEGAL_DOMAIN = 0x22,
   MM_EMISMATCH_FRAGMENTS = 0x24,
   MM_EINTERNAL_ERROR = 0x25,
   MM_EOVERLAPPING_REMAP_ENTRY = 0x26,
   MM_ENON_EXISTING_REMAP_ENTRY = 0x28,
   MM_EILLEGAL_PHYS_TYPE = 0x29,
   MM_EILLEGAL_PHYS_MEM = 0x29,
   MM_EPHYS_FRAG_OVERLAP_PHYS_MEM = 0x30,
   MM_EDOMAIN_NOT_MAS_ACTIVE = 0x31,
   MM_EMAS_NOT_SUPPORTED = 0x32,
   MM_EMAS_ATTRIBUTE_MISMATCH = 0x33,
   MM_EMAS_DATA_ONLY = 0x34,
   MM_ESHARE_UMODE_RESTRICT_FAILED = 0x35,
   MM_ETOO_MANY_PERSISTENT_REGIONS = 0x36,
   MM_EPMM_OUT_OF_MEMORY = 0x37,
   MM_EPAGER_PAGEABLE_AREA = 0x38,
   MM_ENO_PHYSICAL_MAPPING = 0x39,
   MM_EUNAUTHORISED_OPERATION = 0x3A,
   MM_ENON_EXISTING_EVENT = 0x3B,
   MM_ENON_EXISTING_EVENT_HANDLER = 0x3C,
   MM_EPHYS_FRAG_NO_LOGMEM        = 0x3d,
   MM_EPHYS_FRAG_ILLEGAL_SASE     = 0x3e,
   MM_EPHYS_FRAG_OUTSIDE_SASE     = 0x3f,
   MM_EPHYS_FRAG_NOT_SAME_SASE    = 0x40,
   MM_EPHYS_FRAG_NO_VALID_SASE    = 0x41,
   MM_EPHYS_FRAG_TOO_BIG          = 0x42,
   MM_EPHYS_FRAG_MISMATCH         = 0x43,
   MM_EILLEGAL_SASE_OPTION        = 0x44,
   MM_EILLEGAL_OPTION_COMBINATION = 0x45,
   MM_EOPTION_DOMAIN_MISMATCH_MAS = 0x46,
   MM_EOUT_OF_RESOURCES           = 0x47,
   MM_EVENT_POLICY_ALREADY_INSTALLED = 0x48,
   MM_EVENT_POLICY_NOT_INSTALLED     = 0x49,
   MM_ENO_READ_PERMISSION            = 0x4A,
   MM_EINFOTAG_UNSUPPORTED_FOR_CPU   = 0x4B,
   MM_ETOO_SMALL_SIGNAL_SIZE         = 0x4C,
   MM_EINFOTAG_ONLY_SUPPORTED_CPU0   = 0x4D,
   MM_ENO_INFOTAG_COUNTER_SUPPORT_FOR_CPU  = 0x4E,
   MM_ENOT_GMM_PROCESS  = 0x4F,
   MM_EILLEGAL_ALIGNMENT  = 0x50,
   MM_EILLEGAL_ARGUMENT   = 0x51,
   MM_EILLEGAL_CPU_ID     = 0x52,
   MM_EINFOTAG_ONLY_LOCAL_COUNTER = 0x53,
   MM_EINFOTAG_ONLY_GLOBAL_COUNTER = 0x54
};


/**
 *
 * @type MmTagId
 * @brief Enum containing MM TLB info tags and tags for other MM information.
 *
 * @long These enums are used as tag identifiers in
 *       the struct @link "mmtypes.h" "mm_tag"
 *       or @link "mmtypes.h" "mm_tag_ext"
 *       (containing a tag identifier and a tag value).
 *
 * @long Several MM signals contain a tailing field
 *       with an array of this struct,
 *       used for queries/answers. Some examples are
 *       @link "mm.sig" "MM_GET_MMU_STATISTICS_REQUEST",
 *       @link "mm.sig" "MM_GET_PHYS_MEM_TYPE_INFO_REQUEST",
 *       @link "mm.sig" "MM_GET_DOMAIN_INFO_REQUEST",
 *       @link "mm.sig" "MM_GET_LOG_MEM_INFO_REQUEST".
 *
 *
 * @field MM_INFOTAG_TLB_MISS_COUNT
 *   The number of TLB misses. If misses are maintained on Micro-TLB or
 *   Main-TLB level is architecture dependent.
 *
 * @field MM_INFOTAG_DTLB_MISS_COUNT
 *   The number of data TLB misses. If misses are maintained on Micro-TLB or
 *   Main-TLB level is architecture dependent.
 *
 * @field MM_INFOTAG_ITLB_MISS_COUNT
 *   The number of instruction TLB misses. If misses are maintained on
 *   Micro-TLB or Main-TLB level is architecture dependent.
 *
 * @field MM_INFOTAG_DHTAB_MISS_COUNT
 *   The number of data Hash Table misses.
 *
 * @field MM_INFOTAG_IHTAB_MISS_COUNT
 *   The number of instruction Hash Table misses.
 *
 * @field MM_INFOTAG_MMU_ALL
 *   Request all MMU statistics supported for used CPU.
 *   All returned INFOTAG's will be described in mm_tag_ext.
 *               in the struct @link "mmtypes.h" "mm_tag_ext".
 *
 * @field MM_INFOTAG_L1_ICACHE_MISS_COUNT
 *   The number of L1 instruction cache misses.
 *   A separate I-cache is assumed on L1 level.
 *
 * @field MM_INFOTAG_L1_DCACHE_MISS_COUNT
 *   The number of L1 data cache misses.
 *   A separate D-cache is assumed on L1 level.
 *
 * @field MM_INFOTAG_L2_INSTRUCTION_MISS_COUNT
 *   The number of L2 instruction cache misses.
 *   Note that this gives information on instruction misses on L2 cache level
 *   both if the L2 cache have separate I and D caches or if the L2 cache is
 *   unified.
 *
 * @field MM_INFOTAG_L2_DATA_MISS_COUNT
 *   The number of L2 data cache misses.
 *   Note that this gives information on data misses on L2 cache level
 *   both if the L2 cache have separate I and D caches or if the L2 cache is
 *   unified.
 *
 * @field MM_INFOTAG_DTLB_HW_TABLE_WALK_CYCLES
 *   Counts cycles spent performing a hardware table search operation
 *   for the data TLB.
 *
 * @field MM_INFOTAG_ITLB_HW_TABLE_WALK_CYCLES
 *   Counts cycles spent performing a hardware table search operation
 *   for the instruction TLB.
 *
 * @field MM_INFOTAG_INSTRUCTIONS_COMPLETED
 *   Counts completed instructions.
 *
 * @field MM_INFOTAG_L1_DCACHE_HIT_COUNT
 *   The number of L1 data cache hits. (Load, store, or touch hits.)
 *
  * @field MM_INFOTAG_DATA_LOAD_AND_STORES
 *    Counts all load and store instructions completed.
 *
 * @field  MM_INFOTAG_L2_CASTOUT
 *   Number of L2 castouts. Cache blocks that must be written to memory
 *   when a cache miss causes a cache block to be replaced.
 *
 *  @field MM_INFOTAG_L1_DC_LOAD_MISS_CYCLES
 *    L1 data load miss cycles. Counts cycles spent waiting for L1 data cache
 *    misses in the LMQ. It counts from when the miss is detected until
 *    either the data is returned or the request is canceled.
 *
 *  @field MM_INFOTAG_HTAB_OVERFLOW_COUNT
 *    When a free PTE can not be found in neither primary nor secondary PTEG
 *    of the PPC PEM Hash Table it is defined as a hash overflow. The counter
 *    shows how many overflows that has occurred.
 *
 * @field MM_TAG_INFO_PHYS_MEM_INACTIVE_MEMORY
 *   This tag is deprecated and will always return zero.
 *   It was previously a tag for the signal
 *   @link "mm.sig" "MM_GET_PHYS_MEM_TYPE_INFO_REQUEST".
 *
 * @field MM_TAG_INFO_PHYS_MEM_TOTAL_MEMORY
 *   The total amount of free memory of this type.
 *   This is a tag for the signal
 *   @link "mm.sig" "MM_GET_PHYS_MEM_TYPE_INFO_REQUEST".
 *
 * @field MM_TAG_INFO_PHYS_MEM_FREE_MEMORY
 *   The amount of free memory of this type.
 *   This is a tag for the signal
 *   @link "mm.sig" "MM_GET_PHYS_MEM_TYPE_INFO_REQUEST".
 *
 * @field MM_TAG_INFO_PHYS_MEM_FRAG_COUNT
 *   The number of fragments of this type.
 *   This is a tag for the signal
 *   @link "mm.sig" "MM_GET_PHYS_MEM_TYPE_INFO_REQUEST".
 *
 * @field MM_TAG_INFO_PHYS_MEM_BIGGEST_FRAG
 *   The biggest fragment of this type.
 *   This is a tag for the signal
 *   @link "mm.sig" "MM_GET_PHYS_MEM_TYPE_INFO_REQUEST".
 *
 * @field MM_TAG_INFO_PHYS_MEM_FREE_WATERMARK
 *   The smallest amount of free memory available so far
 *   for this type.
 *   This is a tag for the signal
 *   @link "mm.sig" "MM_GET_PHYS_MEM_TYPE_INFO_REQUEST".
 *
 * @field MM_TAG_INFO_PHYS_MEM_METADATA_SIZE
 *   The amount of memory consumed by allocation metadata for this type.
 *   If an error occurs during calculation, (phys_size_t)~0 is returned.
 *   This is a tag for the signal
 *   @link "mm.sig" "MM_GET_PHYS_MEM_TYPE_INFO_REQUEST".
 *
 * @field MM_TAG_INFO_LOG_MEM_BASE_ADDRESS
 *   The base address of a logical memory area. Not valid if the area is
 *   implicit. If multiple banks have been configured the base address
 *   of the first bank is returned.
 *   This is a tag for the signal
 *   @link "mm.sig" "MM_GET_LOG_MEM_INFO_REQUEST".
 *
 * @field MM_TAG_INFO_LOG_MEM_SIZE
 *   The size of  a logical memory area. Not valid if the area is
 *   implicit. If multiple banks have been configured the result
 *   includes all banks (total size).
 *   This is a tag for the signal
 *   @link "mm.sig" "MM_GET_LOG_MEM_INFO_REQUEST".
 *
 * @field MM_TAG_INFO_LOG_MEM_TYPE
 *   The mapping type of a logical memory area.
 *   Compare enum @link "mm.sig" "MappingType".
 *   This is a tag for the signal
 *   @link "mm.sig" "MM_GET_LOG_MEM_INFO_REQUEST".
 *
 * @field MM_TAG_INFO_LOG_MEM_MANAGER_MODE
 *   The manager mode property (TRUE/FALSE) of a logical memory area.
 *   This is a tag for the signal
 *   @link "mm.sig" "MM_GET_LOG_MEM_INFO_REQUEST".
 *   NOT YET AVAILABLE in standard OSE.
 *
 * @field MM_TAG_INFO_LOG_MEM_IMPLICIT
 *   Returns a non-zero value if the log_mem is implicit, otherwise 0 is
 *   returned.
 *   This is a tag for the signal
 *   @link "mm.sig" "MM_GET_LOG_MEM_INFO_REQUEST".
 *
 * @field MM_TAG_INFO_LOG_MEM_BANK_COUNT
 *   Returns the number of banks, or ranges, configured for a logical
 *   memory type.
 *   This is a tag for the signal
 *   @link "mm.sig" "MM_GET_LOG_MEM_INFO_REQUEST".
 *
 * @field MM_TAG_INFO_LOG_MEM_RESIDENT
 *   Returns a non-zero value if the logical memory type is
 *   statically mapped with large resident pages,
 *   otherwise 0 is returned.
 *   This is a tag for the signal
 *   @link "mm.sig" "MM_GET_LOG_MEM_INFO_REQUEST".
 *
 * @field MM_TAG_INFO_LOG_MEM_PERMISSIONS
 *   Returns access permissions configured for a
 *   resident logical memory type. Returns SU_NA_USR_NA if
 *   log_mem type is not resident.
 *   This is a tag for the signal
 *   @link "mm.sig" "MM_GET_LOG_MEM_INFO_REQUEST".
 *
 * @field MM_TAG_INFO_LOG_MEM_CACHE_MODE
 *   Returns cache mode and memory attributes configured for a
 *   resident logical memory type. Returns zero if
 *   log_mem type is not resident.
 *   This is a tag for the signal
 *   @link "mm.sig" "MM_GET_LOG_MEM_INFO_REQUEST".
 *
 * @field MM_TAG_INFO_LOG_MEM_BANK_BASE
 *   Returns the base address of a bank (logical range) used by
 *   a configured logical memory area.
 *   This is a tag for the signal
 *   @link "mm.sig" "MM_GET_LOG_MEM_BANK_INFO_REQUEST".
 *
 * @field MM_TAG_INFO_LOG_MEM_BANK_SIZE
 *   Returns the size of an individual bank (logical range) used by
 *   a configured logical memory area.
 *   memory area.
 *   This is a tag for the signal
 *   @link "mm.sig" "MM_GET_LOG_MEM_BANK_INFO_REQUEST".
 *
 * @seealso mm_tag
 * @seealso MappingType
 * @seealso MM_GET_PHYS_MEM_TYPE_INFO_REQUEST
 * @seealso MM_GET_LOG_MEM_INFO_REQUEST
 * @seealso MM_GET_LOG_MEM_BANK_INFO_REQUEST
 */
enum MmTagId
{
   MM_INFOTAG_TLB_MISS_COUNT                     = 0x1,
   MM_INFOTAG_DTLB_MISS_COUNT                    = 0x2,
   MM_INFOTAG_ITLB_MISS_COUNT                    = 0x3,
   MM_INFOTAG_MMU_ALL                            = 0x13,
   MM_INFOTAG_IHTAB_MISS_COUNT                   = 0x14,
   MM_INFOTAG_DHTAB_MISS_COUNT                   = 0x15,
   MM_INFOTAG_L1_ICACHE_MISS_COUNT               = 0x16,
   MM_INFOTAG_L1_DCACHE_MISS_COUNT               = 0x17,
   MM_INFOTAG_L2_INSTRUCTION_MISS_COUNT          = 0x18,
   MM_INFOTAG_L2_DATA_MISS_COUNT                 = 0x19,
   MM_INFOTAG_DTLB_HW_TABLE_WALK_CYCLES          = 0x20,
   MM_INFOTAG_ITLB_HW_TABLE_WALK_CYCLES          = 0x21,
   MM_INFOTAG_INSTRUCTIONS_COMPLETED             = 0x22,
   MM_INFOTAG_L1_DCACHE_HIT_COUNT                = 0x23,
   MM_INFOTAG_DATA_LOAD_AND_STORES               = 0x24,
   MM_INFOTAG_L2_CASTOUT                         = 0x25,
   MM_INFOTAG_L1_DC_LOAD_MISS_CYCLES             = 0x26,
   MM_INFOTAG_HTAB_OVERFLOW_COUNT                = 0x27,


   /*
    *  Tags for MM_GET_PHYS_MEM_TYPE_INFO.
    */

   MM_TAG_INFO_PHYS_MEM_TOTAL_MEMORY             = 0x4,
   MM_TAG_INFO_PHYS_MEM_FREE_MEMORY              = 0x5,
   MM_TAG_INFO_PHYS_MEM_FRAG_COUNT               = 0x6,
   MM_TAG_INFO_PHYS_MEM_BIGGEST_FRAG             = 0xA,
   MM_TAG_INFO_PHYS_MEM_INACTIVE_MEMORY          = 0xB,
   MM_TAG_INFO_PHYS_MEM_METADATA_SIZE            = 0xF,

   /*
    *  Tag for MM_GET_PHYS_MEM_TYPE_INFO.
    */

   MM_TAG_INFO_PHYS_MEM_FREE_WATERMARK           = 0xC,

   /*
    *  Tags for MM_GET_LOG_MEM_INFO.
    *
    */

   MM_TAG_INFO_LOG_MEM_BASE_ADDRESS              = 0x7,
   MM_TAG_INFO_LOG_MEM_SIZE                      = 0x8,
   MM_TAG_INFO_LOG_MEM_TYPE                      = 0x9,
   MM_TAG_INFO_LOG_MEM_MANAGER_MODE              = 0xD,
   MM_TAG_INFO_LOG_MEM_IMPLICIT                  = 0xE,
   MM_TAG_INFO_LOG_MEM_BANK_COUNT                = 0x10,

   MM_TAG_INFO_LOG_MEM_PERMISSIONS               = 0x28,
   MM_TAG_INFO_LOG_MEM_CACHE_MODE                = 0x29,
   MM_TAG_INFO_LOG_MEM_RESIDENT                  = 0x30,


   /*
    *  Tags for MM_GET_LOG_MEM_BANK_INFO.
    *
    */
   MM_TAG_INFO_LOG_MEM_BANK_BASE                 = 0x11,
   MM_TAG_INFO_LOG_MEM_BANK_SIZE                 = 0x12

};


/**
 * @macro MM_INFOTAG_TLB_MISS_COUNT_DESC
 *
 * @brief String description of the info tag MM_INFOTAG_TLB_MISS_COUNT.
 *
 */
#define MM_INFOTAG_TLB_MISS_COUNT_DESC  "Unified TLB Miss counter."

/**
 * @macro MM_INFOTAG_DTLB_MISS_COUNT_DESC
 *
 * @brief String description of the info tag MM_INFOTAG_DTLB_MISS_COUNT.
 *
 */
#define MM_INFOTAG_DTLB_MISS_COUNT_DESC  "D-TLB Miss counter."

/**
 * @macro MM_INFOTAG_ITLB_MISS_COUNT_DESC
 *
 * @brief String description of the info tag MM_INFOTAG_ITLB_MISS_COUNT.
 *
 */
#define MM_INFOTAG_ITLB_MISS_COUNT_DESC  "I-TLB Miss counter."

/**
 * @macro MM_INFOTAG_INSTRUCTIONS_COMPLETED_DESC
 *
 * @brief String description of the info tag MM_INFOTAG_INSTRUCTIONS_COMPLETED.
 *
 *
 */
#define MM_INFOTAG_INSTRUCTIONS_COMPLETED_DESC  "Instructions completed."

/*
 * Note that the MmTagId MM_INFOTAG_MMU_ALL does not need to have
 * string description, since it is only used as input to MM.
 *
 */

/**
 * @macro MM_INFOTAG_DHTAB_MISS_COUNT_DESC
 *
 * @brief String description of the info tag MM_INFOTAG_DHTAB_MISS_COUNT.
 *
 */
#define MM_INFOTAG_DHTAB_MISS_COUNT_DESC  "D-Hash Table Miss counter."

/**
 * @macro MM_INFOTAG_IHTAB_MISS_COUNT_DESC
 *
 * @brief String description of the info tag MM_INFOTAG_IHTAB_MISS_COUNT.
 *
 */
#define MM_INFOTAG_IHTAB_MISS_COUNT_DESC  "I-Hash Table Miss counter."

/**
 * @macro MM_INFOTAG_L1_ICACHE_MISS_COUNT_DESC
 *
 * @brief String description of the info tag
 *        MM_INFOTAG_L1_ICACHE_MISS_COUNT.
 *
 */
#define MM_INFOTAG_L1_ICACHE_MISS_COUNT_DESC  "L1 I-Cache miss counter."

/**
 * @macro MM_INFOTAG_L1_DCACHE_MISS_COUNT_DESC
 *
 * @brief String description of the info tag
 *        MM_INFOTAG_L1_DCACHE_MISS_COUNT.
 *
 */
#define MM_INFOTAG_L1_DCACHE_MISS_COUNT_DESC  "L1 D-Cache miss counter."

/**
 * @macro MM_INFOTAG_L2_DATA_MISS_COUNT_DESC
 *
 * @brief String description of the info tag
 *        MM_INFOTAG_L2_DATA_MISS_COUNT.
 *
 */
#define MM_INFOTAG_L2_DATA_MISS_COUNT_DESC "L2 Data miss counter."

/**
 * @macro MM_INFOTAG_L2_INSTRUCTION_MISS_COUNT_DESC
 *
 * @brief String description of the info tag
 *        MM_INFOTAG_L2_INSTRUCTION_MISS_COUNT.
 *
 */
#define MM_INFOTAG_L2_INSTRUCTION_MISS_COUNT_DESC "L2 Instruction miss counter."

/**
 * @macro MM_INFOTAG_DTLB_HW_TABLE_WALK_CYCLES_DESC
 *
 * @brief String description of the info tag
 *        MM_INFOTAG_DTLB_HW_TABLE_WALK_CYCLES.
 *
 */
#define MM_INFOTAG_DTLB_HW_TABLE_WALK_CYCLES_DESC \
        "D-TLB HW table search cycles"

/**
 * @macro MM_INFOTAG_ITLB_HW_TABLE_WALK_CYCLES_DESC
 *
 * @brief String description of the info tag
 *        MM_INFOTAG_ITLB_HW_TABLE_WALK_CYCLES.
 *
 */
#define MM_INFOTAG_ITLB_HW_TABLE_WALK_CYCLES_DESC \
        "I-TLB HW table search cycles"

/**
 * @macro MM_INFOTAG_L1_DCACHE_HIT_COUNT_DESC
 *
 * @brief String description of the info tag
 *        MM_INFOTAG_L1_DCACHE_HIT_COUNT.
 *
 */
#define MM_INFOTAG_L1_DCACHE_HIT_COUNT_DESC    "L1 D-Cache hit counter."


/**
 * @macro MM_INFOTAG_DATA_LOAD_AND_STORES_DESC
 *
 * @brief String description of the info tag
 *        MM_INFOTAG_DATA_LOAD_AND_STORES.
 *
 */
#define MM_INFOTAG_DATA_LOAD_AND_STORES_DESC \
   "Counts all load and store instructions completed"

/**
 * @macro MM_INFOTAG_L2_CASTOUT_DESC
 *
 * @brief String description of the info tag
 *        MM_INFOTAG_L2_CASTOUT.
 *
 */
#define MM_INFOTAG_L2_CASTOUT_DESC   "Number of L2 castouts."

/**
 * @macro MM_INFOTAG_L1_DC_LOAD_MISS_CYCLES_DESC
 *
 * @brief String description of the info tag
 *        MM_INFOTAG_L1_DC_LOAD_MISS_CYCLES.
 *
 */
#define MM_INFOTAG_L1_DC_LOAD_MISS_CYCLES_DESC \
        "L1 D-Cache load miss cycles."

/**
 * @macro MM_INFOTAG_HTAB_OVERFLOW_COUNT_DESC
 *
 * @brief String description of the info tag
 *        MM_INFOTAG_HTAB_OVERFLOW_COUNT_DESC.
 *
 */
#define MM_INFOTAG_HTAB_OVERFLOW_COUNT_DESC \
        "Hash Table overflow counter."


/**
 *
 * @type MappingType
 *
 * @brief Mapping enum type for logical memory areas.
 *        Returned by @link "mm.sig" "MM_GET_LOG_MEM_INFO_REQUEST".
 *
 * @long The type indicates the mapping from logical to physical memory.
 *       See the OSE Architecture Guide for more information.
 *
 * @field SASE
 *   Single Address Space Equal - logical and physical addresses are equal.
 *
 * @field SAS
 *   Single Address Space - the mapping from logical to physical addresses is
 *   the same regardless of domain.
 *
 * @field MAS
 *   Multiple Address Space - the mapping from logical to physical addresses
 *   may vary from one protection domain to another.
 */
enum MappingType
{
   SASE = 0x1,
   SAS  = 0x2,
   MAS  = 0x3
};

/**
 * @macro GMM_NAME_STRING
 *
 * @brief The name of the memory manager process.
 *
 * @long It is a macro defined as "ose_memory_mgr".
 */
#define GMM_NAME_STRING          "ose_memory_mgr"


/**
 * @macro MM_INTERNAL_RAM
 *
 * @brief A physical memory type identifier for RAM.
 *        This physical memory type is always available.
 *
 * @long This macro is defined as the string "RAM".
 */
#define MM_INTERNAL_RAM      "RAM"

/**************************************************************************/

/* Internal macros used to set/shift permissions in MM. */
/* Not visible in reference docs. */
#define MM_PERM_READ    (0x1)
#define MM_PERM_WRITE   (0x2)
#define MM_PERM_EXECUTE (0x4)
#define MM_SU_SHIFT     (8*1)

/**
 * @macro MM_USR_EXECUTE
 * @brief User mode execute MM permission bit.
 * @seealso MM_CREATE_REGION_REQUEST
 */
#define MM_USR_EXECUTE  (MM_PERM_EXECUTE)

/**
 * @macro MM_USR_WRITE
 * @brief User mode write MM permission bit.
 * @seealso MM_CREATE_REGION_REQUEST
 */
#define MM_USR_WRITE    (MM_PERM_WRITE)

/**
 * @macro MM_USR_READ
 * @brief User mode read MM permission bit.
 * @seealso MM_CREATE_REGION_REQUEST
 */
#define MM_USR_READ     (MM_PERM_READ)

/**
 * @macro MM_SU_EXECUTE
 * @brief Supervisor mode execute MM permission bit.
 * @seealso MM_CREATE_REGION_REQUEST
 */
#define MM_SU_EXECUTE   (MM_PERM_EXECUTE << MM_SU_SHIFT)

/**
 * @macro MM_SU_WRITE
 * @brief Supervisor mode write MM permission bit.
 * @seealso MM_CREATE_REGION_REQUEST
 */
#define MM_SU_WRITE     (MM_PERM_WRITE << MM_SU_SHIFT)

/**
 * @macro MM_SU_READ
 * @brief Supervisor mode read MM permission bit.
 * @seealso MM_CREATE_REGION_REQUEST
 */
#define MM_SU_READ      (MM_PERM_READ << MM_SU_SHIFT)

/**
 * @macro SU_NA_USR_NA
 * @brief MM permission: No access neither from Supervisor nor user mode.
 * @seealso MM_CREATE_REGION_REQUEST
 */
#define SU_NA_USR_NA (0)

/**
 * @macro SU_RO_USR_NA
 * @brief MM permission: Read only from Supervisor mode.
 *        No access from user mode.
 * @seealso MM_CREATE_REGION_REQUEST
 */
#define SU_RO_USR_NA (MM_SU_READ)

/**
 * @macro SU_RO_USR_RO
 * @brief MM permission: Read only from both Supervisor and user mode.
 * @seealso MM_CREATE_REGION_REQUEST
 */
#define SU_RO_USR_RO (MM_SU_READ | MM_USR_READ)

/**
 * @macro SU_RW_USR_NA
 * @brief MM permission: Read-write from Supervisor mode,
 *        not accessible from user mode.
 * @seealso MM_CREATE_REGION_REQUEST
 */
#define SU_RW_USR_NA (MM_SU_READ | MM_SU_WRITE)

/**
 * @macro SU_RW_USR_RO
 * @brief MM permission: Read-write from Supervisor mode,
 *        read-only from user mode.
 * @seealso MM_CREATE_REGION_REQUEST
 */
#define SU_RW_USR_RO (MM_SU_READ | MM_SU_WRITE | MM_USR_READ)

/**
 * @macro SU_RW_USR_RW
 * @brief MM permission: Read-write from Supervisor mode,
 *        read-write from user mode.
 * @seealso MM_CREATE_REGION_REQUEST
 */
#define SU_RW_USR_RW (MM_SU_READ | MM_SU_WRITE | MM_USR_READ | \
                      MM_USR_WRITE)
/**
 * @macro SU_RX_USR_NA
 * @brief MM permission: Read-execute from Supervisor mode,
 *        not accessible from user mode.
 * @seealso MM_CREATE_REGION_REQUEST
 */
#define SU_RX_USR_NA (MM_SU_READ | MM_SU_EXECUTE)

/**
 * @macro SU_RX_USR_RO
 * @brief MM permission: Read-execute from Supervisor mode,
 *        read-only from user mode.
 * @seealso MM_CREATE_REGION_REQUEST
 */
#define SU_RX_USR_RO (MM_SU_READ | MM_SU_EXECUTE | MM_USR_READ)

/**
 * @macro SU_RX_USR_RX
 * @brief MM permission: Read-execute from Supervisor mode,
 *        read-execute from user mode.
 * @seealso MM_CREATE_REGION_REQUEST
 */
#define SU_RX_USR_RX (MM_SU_READ | MM_SU_EXECUTE | MM_USR_READ | \
                      MM_USR_EXECUTE)

/**
 * @macro SU_RWX_USR_NA
 * @brief MM permission: Read-write-execute from Supervisor mode,
 *        not accessible from user mode.
 * @seealso MM_CREATE_REGION_REQUEST
 */
#define SU_RWX_USR_NA (MM_SU_READ | MM_SU_EXECUTE | MM_SU_WRITE)

/**
 * @macro SU_RWX_USR_RX
 * @brief MM permission: Read-write-execute from Supervisor mode,
 *        read-execute from user mode.
 * @seealso MM_CREATE_REGION_REQUEST
 */
#define SU_RWX_USR_RX (MM_SU_READ | MM_SU_EXECUTE | MM_SU_WRITE | MM_USR_READ | \
                       MM_USR_EXECUTE)

/**
 * @macro SU_RWX_USR_RO
 * @brief MM permission: Read-write-execute from Supervisor mode,
 *        read-only from user mode.
 * @seealso MM_CREATE_REGION_REQUEST
 */
#define SU_RWX_USR_RO (MM_SU_READ | MM_SU_EXECUTE | MM_SU_WRITE | MM_USR_READ)

/**
 * @macro SU_RWX_USR_RW
 * @brief MM permission: Read-write-execute from Supervisor mode,
 *        read-write from user mode.
 * @seealso MM_CREATE_REGION_REQUEST
 */
#define SU_RWX_USR_RW  (MM_SU_READ | MM_SU_EXECUTE | MM_SU_WRITE | MM_USR_READ | \
                       MM_USR_WRITE )

/**
 * @macro SU_RWX_USR_RWX
 * @brief MM permission: Read-write-execute from Supervisor mode,
 *        read-write-execute from user mode.
 * @seealso MM_CREATE_REGION_REQUEST
 */
#define SU_RWX_USR_RWX (MM_SU_READ | MM_SU_EXECUTE | MM_SU_WRITE | MM_USR_READ | \
                       MM_USR_EXECUTE | MM_USR_WRITE )

/**************************************************************************/

/**
 * @macro MM_CACHE_INHIBITED
 * @brief Cache mode attribute: No memory cache activated.
 * @seealso MM_CREATE_REGION_REQUEST
 */
#define MM_CACHE_INHIBITED      0x1

/**
 * @macro MM_CACHE_WRITE_THROUGH
 * @brief Cache mode attribute: Write through cache type.
 * @long  This means that writing always writes both to memory
 *        and cache, and reading is tried from cache.
 * @seealso MM_CREATE_REGION_REQUEST
 */
#define MM_CACHE_WRITE_THROUGH  0x2

/**
 * @macro MM_CACHE_COPY_BACK
 * @brief Cache mode attribute: Copy back cache.
 * @long  This means that writing writes
 *        to the memory cache only, and reading is tried from cache.
 * @seealso MM_CREATE_REGION_REQUEST
 */
#define MM_CACHE_COPY_BACK      0x4

/**
 * @macro MM_SPECULATIVE_ACCESS
 * @brief Cache memory mode attribute: The normal type of MMU mapping that
 *        is optimized for fast access from applications.
 * @seealso MM_CREATE_REGION_REQUEST
 */
#define MM_SPECULATIVE_ACCESS   0x8

/**
 * @macro MM_ORDERED_ACCESS
 * @brief Cache memory mode attribute: A MMU mapping type that is suitable
 *        for accessing memory mapped I/O and flash memory.
 * @seealso MM_CREATE_REGION_REQUEST
 *
 */
#define MM_ORDERED_ACCESS       0x10

/**************************************************************************/

/**
 * @macro MM_SIGNUM_BASE
 * @brief Base number for all MM signals.
 */
#define MM_SIGNUM_BASE  36100

/**
 * @macro MM_SIGERR_BASE
 * @brief Base number for all error codes in MM signals.
 * @seealso MmStatus
 */
#define MM_SIGERR_BASE  0

/**************************************************************************/

/**
 * @signal MM_CREATE_REGION_REQUEST
 *
 * @brief Signal used to create a memory region.
 *
 * @long This signal either creates a region and maps it to physical
 * resources, or creates a logical region without physical resources.
 * A region belongs to exactly one domain.
 *
 * @long Create a Region with Physical Memory:
 *
 * @long The physical memory can either be allocated by MM from the type
 * described in "phys_resources", or it could be supplied by the user.
 *
 * @long If memory is taken from MM administrated physical resources, the name
 * of the physical memory type must be supplied in the "phys_mem" parameter.
 * This is the default behavior.
 * If a region is created this way, the physical memory is automatically freed
 * when there are no regions mapped to it.
 * In all systems, @link "mm.sig" "MM_INTERNAL_RAM" (name string "RAM") is
 * available as a physical memory type.
 * See "OSE Core User's Guide" about the physical memory types.
 *
 * @long Else, if external physical resources are specified, the options
 * parameter must be OR'ed with @link "mm.sig" "MM_CREATE_REGION_SPECIFY_PHYS".
 * Also, a list of the physical resources must be supplied in the
 * "phys_resources" parameter.
 *
 * @long Create a Region without Physical Memory:
 *
 * @long Creating a region without physical memory is only allowed in the
 * system domain. When creating a region that only consists of logical
 * memory, the options field must be OR'ed with
 * @link "mm.sig" "MM_CREATE_REGION_LOGICAL".
 * The request shall not carry any information about physical resources.
 *
 * @long See also @link "mm.sig" "MM_GET_REGION_INFO_REQUEST", which
 * uses the same type of struct and returns information about a region.
 *
 * @long Align and the options parameters are returned
 * with unspecified values.
 *
 * @long Send signals to "ose_memory_mgr".
 *
 * @member mm_create_region
 * @seealso MM_CREATE_REGION_REPLY
 * @seealso MM_GET_REGION_INFO_REQUEST
 */
#define MM_CREATE_REGION_REQUEST     (36100) \
  /* !-SIGNO(struct mm_create_region)-! */

/**
 * @signal MM_CREATE_REGION_REPLY
 * @brief Reply signal from Memory Manager replying to
 *        @link "mm.sig" "MM_CREATE_REGION_REQUEST".
 *
 * @long The reply contains a struct of the same type 'mm_create_region',
 *       as used in the MM_CREATE_REGION_REQUEST request.
 *       See the "output" field descriptions in
 *       @link "mm.sig" "MM_CREATE_REGION_REQUEST".
 *
 * @seealso MM_CREATE_REGION_REQUEST
 */
#define MM_CREATE_REGION_REPLY       (36101) \
  /* !-SIGNO(struct mm_create_region)-! */

/**
 * @macro MM_CREATE_REGION_SPECIFY_NAME
 * @brief Option bit defining that "region name"
 * is specified in the signal.
 * @link "mm.sig" "MM_CREATE_REGION_REQUEST".
 * @seealso MM_CREATE_REGION_REQUEST
 */
#define MM_CREATE_REGION_SPECIFY_NAME  (0x1)

/**
 * @macro MM_CREATE_REGION_SPECIFY_PHYS
 * @brief Option bit defining that "physical resources"
 * are specified in the signal.
 * @link "mm.sig" "MM_CREATE_REGION_REQUEST".
 * @seealso MM_CREATE_REGION_REQUEST
 */
#define MM_CREATE_REGION_SPECIFY_PHYS  (0x2)

/**
 * @macro MM_CREATE_REGION_SPECIFY_LOG
 * @brief Option bit defining that a "logical resource"
 * (field base) is specified in the signal.
 * @link "mm.sig" "MM_CREATE_REGION_REQUEST".
 * In a reply, the resulting base could differ from the requested, if
 * the requested base was not available.
 * Compare also @link "mm.sig" "MM_LOGMEM_FROM_ADDRESS".
 * @seealso MM_CREATE_REGION_REQUEST
 */
#define MM_CREATE_REGION_SPECIFY_LOG   (0x4)

/**
 * @macro MM_CREATE_REGION_LOGICAL
 * @brief Option bit, set to create a region with only logical
 * memory, using the signal.
 * @link "mm.sig" "MM_CREATE_REGION_REQUEST".
 * @seealso MM_CREATE_REGION_REQUEST
 */
#define MM_CREATE_REGION_LOGICAL       (0x8)

/**
 * @macro MM_CREATE_REGION_SPECIFY_ALIGN
 * @brief Option bit defining that "alignment"
 * is specified in the signal.
 * @link "mm.sig" "MM_CREATE_REGION_REQUEST".
 * @seealso MM_CREATE_REGION_REQUEST
 */
#define MM_CREATE_REGION_SPECIFY_ALIGN (0x10)

/**
 * @macro MM_CREATE_REGION_PERSISTENT
 * @brief Option bit, set to create a persistent region,
 * using the signal.
 * @link "mm.sig" "MM_CREATE_REGION_REQUEST".
 * @long Persistent regions must be in the system domain.
 * @seealso MM_CREATE_REGION_REQUEST
 */
#define MM_CREATE_REGION_PERSISTENT    (0x20)

/**
 * @macro MM_CHECK_LOW_LIMIT
 * @brief Option bit, set in
 *        @link "mm.sig" "MM_CREATE_REGION_REQUEST"
 *        if MM should check the total free space against the limit
 *        specified in the low_limit field.
 * @seealso MM_CREATE_REGION_REQUEST
 */
#define MM_CHECK_LOW_LIMIT             (0x40)

/**
 * @macro MM_LOGMEM_FROM_ADDRESS
 * @brief Option bit, set in
 *        @link "mm.sig" "MM_CREATE_REGION_REQUEST"
 *        if MM shall discard the contents of the log_mem field (logical
 *        memory name) and instead use the address in field <base>
 *        to look up a logical memory type.
 *        This option must be used in conjunction with the
 *        option MM_CREATE_REGION_SPECIFY_LOG.
 *
 *        Note that a logical memory entry for address <base> must still
 *        exist.
 * @seealso MM_CREATE_REGION_REQUEST
 */
#define MM_LOGMEM_FROM_ADDRESS         (0x80)

/**
 * @macro MM_CREATE_REGION_TENTATIVE
 * @deprecated This option is deprecated and will have no effect.
 * @seealso MM_CREATE_REGION_REQUEST
 */
#define MM_CREATE_REGION_TENTATIVE     (0x100)

/**
 * @macro MM_CREATE_REGION_FAULTABLE
 * @brief Option bit, set in
 *        @link "mm.sig" "MM_CREATE_REGION_REQUEST"
 *        if MM shall create a faultable region.
 * @seealso MM_CREATE_REGION_REQUEST
 */
#define MM_CREATE_REGION_FAULTABLE     (0x200)

/**
 * @macro MM_CREATE_REGION_DUPLICATE
 * @brief Option bit, set in
 *        @link "mm.sig" "MM_CREATE_REGION_REQUEST"
 *        if MM shall create a duplicate mapping of a region.
 *        Option @link "mm.sig" "MM_CREATE_REGION_SPECIFY_PHYS" must be used in
 *        conjunction with this option.
 * @long  A duplicate mapping is new mapping on an already existing
 *        region. The duplicate mapping is only allowed when the stated
 *        physical memory resides inside a SU_NA_USR_NA SASE region.
 *        The duplicate mapped region can map all or parts of the SASE region.
 * @seealso MM_CREATE_REGION_REQUEST
 */
#define MM_CREATE_REGION_DUPLICATE     (0x400)

/**
 * @macro MM_CREATE_REGION_RETURN_PHYS
 * @brief Option bit, set in
 *        @link "mm.sig" "MM_CREATE_REGION_REQUEST".
 *        if MM shall return the physical resources that was allocated
 *        for the created region.
 * @long  A new reply signal is allocated if not all physical fragment fits
 *        inside the supplied request signal. If the create region request
 *        fails the physical_resources field in the signal is undefined.
 * @seealso MM_CREATE_REGION_REQUEST
 */
#define MM_CREATE_REGION_RETURN_PHYS   (0x800)

/**
 * @macro MM_CREATE_REGION_RESIDENT
 * @brief Option bit, set in
 *        @link "mm.sig" "MM_CREATE_REGION_REPLY".
 *        MM shall return the MM_CREATE_REGION_RESIDENT bit set if the region
 *        is created in resident logical memory
 *        field from mm_create_region structure for the created region.
 * @long  This bit is used only as information regarding regions mapped
 *        in a resident logical memory. It is relevant only for
 *        MM_GET_REGION_INFO_REPLY signal.
 * @seealso MM_CREATE_REGION_REQUEST
 * @seealso MM_GET_REGION_INFO_REPLY
 */
#define MM_CREATE_REGION_RESIDENT   (0x1000)

/**
 * @macro MM_CREATE_REGION_TLB_GLOBAL
 * @brief Option bit, set in
 *        @link "mm.sig" "MM_CREATE_REGION_REPLY".
 *        MM shall return the MM_CREATE_REGION_TLB_GLOBAL bit set if the region
 *        is possible to map with global entries in the TLB.
 *
 * @long  Global in this context means that all MM domains will be able to
 *        share the same TLB entries. This will improve the system performance
 *        by decreasing the number of dynamic TLB misses.
 *        This bit is used only as information and is relevant only for
 *        MM_GET_REGION_INFO_REPLY signal.
 * @seealso MM_GET_REGION_INFO_REPLY
 */
#define MM_CREATE_REGION_TLB_GLOBAL   (0x2000)

/**
 * @macro MM_CREATE_REGION_INITIAL_STATIC
 * @brief Option bit, set in
 *        @link "mm.sig" "MM_CREATE_REGION_REPLY".
 *        MM returns MM_CREATE_REGION_INITIAL_STATIC bit set for a region that
 *        has been defined in the system configuration and is of the
 *        type: static. Static means that is not possible to delete this
 *        region.
 *
 * @long  Regions defined in the system configuration are
 *        created at start up and are called initial regions.
 *        The system start occurs when either booting the system or resetting
 *        the system. Booting is sometimes also referred to as cold start
 *        and reset as warm start.
 *        Static is the default type for regions defined in the system
 *        configuration.
 *        This bit is used only as information and is relevant only for
 *        MM_GET_REGION_INFO_REPLY signal.
 * @seealso MM_GET_REGION_INFO_REPLY
 */
#define MM_CREATE_REGION_INITIAL_STATIC   (0x4000)

/**
 * @macro MM_CREATE_REGION_INITIAL_DYNAMIC
 * @brief Option bit, set in
 *        @link "mm.sig" "MM_CREATE_REGION_REPLY".
 *        MM returns MM_CREATE_REGION_INITIAL_DYNAMIC bit set for a region
 *        that has been defined in the system configuration and are of the
 *        type: dynamic. Dynamic means that is possible to delete this region.
 *        
 * @long  This bit is used only as information and is relevant only for
 *        MM_GET_REGION_INFO_REPLY signal.
 * @seealso MM_CREATE_REGION_INITIAL_STATIC
 * @seealso MM_GET_REGION_INFO_REPLY
 */
#define MM_CREATE_REGION_INITIAL_DYNAMIC (0x8000)

/*
 * The following flag values are reserved for mm_internal.sig.
 *    MM_CREATE_REGION_PRIVATE_FLAG_1    (0x10000)
 *    MM_CREATE_REGION_PRIVATE_FLAG_2    (0x20000)
 */


/**
 * @type mm_create_region
 *
 * @field sig_no Signal number.
 *
 * @input size   Requested size for the region. Mandatory parameter.
 *               When option MM_CREATE_REGION_SPECIFY_PHYS is used the aligned
 *               size value must be the sum of the supplied phys_resources.
 *
 * @input domain Domain identity that this region will be created in.
 *               Observe that persistent regions must be created in
 *               the system domain. Mandatory parameter.
 *
 * @input name   Name of the region.
 *               The name of a region may be useful when debugging.
 *               Store the name string in the name field
 *               and or the options field with
 *               @link "mm.sig" "MM_CREATE_REGION_SPECIFY_NAME".
 *               Optional parameter.
 *
 * @input phys_mem Name of the physical memory type.
 *               Mandatory parameter.
 *               (An exception is if physical memory is supplied by client,
 *               in this case this name is of no significance and will be
 *               ignored. See "phys_resources" parameter for details.)
 *
 * @input log_mem Name of the logical memory area.
 *               Mandatory parameter.
 *               (An exception is if logmem is determined from a given address,
 *               in this case this name is of no significance and will be
 *               ignored See option "MM_LOGMEM_FROM_ADDRESS" for details.)
 *
 * @input permissions Requested user and supervisor mode permissions for
 *               the region.
 *               Mandatory parameter.
 *               @break
 *               See the permission bits:
 *               @link "mm.sig" "MM_USR_EXECUTE",
 *               @link "mm.sig" "MM_USR_WRITE",
 *               @link "mm.sig" "MM_USR_READ",
 *               @link "mm.sig" "MM_SU_EXECUTE",
 *               @link "mm.sig" "MM_SU_WRITE",
 *               @link "mm.sig" "MM_SU_READ"
 *               @break
 *               See also the combined permissions
 *               @link "mm.sig" "SU_NA_USR_NA",
 *               @link "mm.sig" "SU_RO_USR_NA",
 *               @link "mm.sig" "SU_RO_USR_RO",
 *               @link "mm.sig" "SU_RW_USR_NA",
 *               @link "mm.sig" "SU_RW_USR_RO",
 *               @link "mm.sig" "SU_RW_USR_RW",
 *               @link "mm.sig" "SU_RX_USR_NA",
 *               @link "mm.sig" "SU_RX_USR_RO",
 *               @link "mm.sig" "SU_RX_USR_RX"
 *               @break
 *               The combined permission macros names are build with:@break
 *               SU   = Supervisor Mode,
 *               USR  = User Mode,
 *               NA   = Not available i.e. no permissions at all,
 *               R    = Read permissions,
 *               RO   = Read Only,
 *               W    = Write,
 *               X    = Execute.
 *
 * @input cache_mode  Requested cache mode.
 *               Every region must specify a cache_mode that consists of
 *               two parts, a cache attribute and a memory mode attribute.
 *               Mandatory parameter.
 *
 *               Combining one cache attribute with one memory mode
 *               attribute generates a cache_mode.
 *               @break
 *               Cache attributes:
 *               @link "mm.sig" "MM_CACHE_INHIBITED",
 *               @link "mm.sig" "MM_CACHE_WRITE_THROUGH",
 *               @link "mm.sig" "MM_CACHE_COPY_BACK"
 *               @break
 *               Memory mode attributes:
 *               @link "mm.sig" "MM_SPECULATIVE_ACCESS",
 *               @link "mm.sig" "MM_ORDERED_ACCESS"
 *               @break
 *               The following six combination of cache_modes are allowed:
 *               @break
 *               1. (MM_CACHE_INHIBITED     | MM_SPECULATIVE_ACCESS),@break
 *               2. (MM_CACHE_WRITE_THROUGH | MM_SPECULATIVE_ACCESS),@break
 *               3. (MM_CACHE_COPY_BACK     | MM_SPECULATIVE_ACCESS),@break
 *               4. (MM_CACHE_INHIBITED     | MM_ORDERED_ACCESS),@break
 *               5. (MM_CACHE_WRITE_THROUGH | MM_ORDERED_ACCESS),@break
 *               6. (MM_CACHE_COPY_BACK     | MM_ORDERED_ACCESS)
 *
 *
 * @input base   A base address for the region is specified by setting the
 *               base field, and OR'ing the options field with
 *	         @link "mm.sig" "MM_CREATE_REGION_SPECIFY_LOG". The base
 *	         address must be page aligned. 
 *
 *               Optional parameter.
 *
 * @input align  Requested region alignment. Alignment is specified by size.
 *               The alignment of a region is specified by setting the align
 *	         field, and also OR'ing the options field with
 *	         @link "mm.sig" "MM_CREATE_REGION_SPECIFY_ALIGN". The alignment
 *	         value must be a power of two. 
 *
 *               For example a 16 KB alignment is requested by setting align
 *		 to (16*1024).
 *
 *               Optional parameter.
 *
 * @input options Options to use when creating region.
 *               Options of region creation are activated by OR'ing
 *               it with requested option numbers.
 *               If no options are to be used, this parameter should be zero.
 *               (Persistence must be specified when using certain physical
 *               memory types.)
 *               Mandatory parameter.
 *               @break
 *               See the options
 *               @link "mm.sig" "MM_CREATE_REGION_SPECIFY_NAME",
 *               @link "mm.sig" "MM_CREATE_REGION_SPECIFY_PHYS",
 *               @link "mm.sig" "MM_CREATE_REGION_SPECIFY_LOG",
 *               @link "mm.sig" "MM_CREATE_REGION_LOGICAL",
 *               @link "mm.sig" "MM_CREATE_REGION_SPECIFY_ALIGN",
 *               @link "mm.sig" "MM_CREATE_REGION_PERSISTENT",
 *               @link "mm.sig" "MM_CHECK_LOW_LIMIT".
 *               @link "mm.sig" "MM_LOGMEM_FROM_ADDRESS".
 *               @link "mm.sig" "MM_CREATE_REGION_FAULTABLE".
 *               @link "mm.sig" "MM_CREATE_REGION_DUPLICATE".
 *               @link "mm.sig" "MM_CREATE_REGION_RETURN_PHYS".
 *
 * @input phys_count Length of field phys_resources.
 *               Contains the number of elements in phys_resources.
 *
 *               Optional parameter.
 *
 * @input phys_resources List of physical resources - see the struct
 *               @link "mmtypes.h" "phys_resource".
 *               Lists the resource that will be mapped to
 *               this region. This list of non-overlapping
 *               physical resources must span the requested region size.
 *               The number of elements in the list shall be in phys_count.
 *
 *               Optional parameter.
 *
 * @input low_limit The low limit value of the MM total free space that MM
 *                  shall check against when regions are allocated due to
 *                  large blocks allocated by malloc. Only used when the
 *                  @link "mm.sig" "MM_CHECK_LOW_LIMIT"
 *                  option is set.
 *
 *                  Optional parameter.
 *
 *
 * @output status Success or failure reason of operation,
 *                see @link "mm.sig" "MmStatus" for error codes.
 *
 * @output size  Allocated size of region.
 *
 * @output name  Name of the region. Relevant for MM_GET_REGION_INFO_REPLY.
 *
 * @output phys_mem Name of the physical memory type. Relevant for MM_GET_REGION_INFO_REPLY.
 *
 * @output log_mem Name of the logical memory area. Relevant for MM_GET_REGION_INFO_REPLY.
 *
 * @output permissions The actual access permissions.
 *               This may differ from the requested permissions if it
 *               is not supported by the current platform.
 *
 * @output cache_mode Actual cache properties for the region.
 *               The requested cache properties may not be available on
 *	         the current platform.
 *
 * @output base  Actual base address for region.
 *               If this value differs from the requested base address,
 *               the requested base address was not available.
 *
 * @output align Actual align on region.
 *               The actual alignment is always a multiple (>= 1)
 *               of the requested alignment.
 *		 Undefined value with MM_GET_REGION_INFO_REPLY.
 *
 * @output options Options used when creating region.
 *
 * @output phys_count Length of field phys_resources or -1 on incomplete fields.
 *               Only valid if option @link "mm.sig" "MM_CREATE_REGION_RETURN_PHYS" is set.
 *               Contains the number of elements in phys_resources.
 *               If the number of lists is too large to return in one reply signal
 *               phys_count is set to -1.
 *               If -1 is returned use @link "mm.sig" "MM_GET_REGION_INFO_REQUEST"
 *               get a complete list of the physical resources.
 *
 * @output phys_resources List of physical resources - see the struct
 *               @link "mmtypes.h" "phys_resource".
 *               Only valid if option @link "mm.sig" "MM_CREATE_REGION_RETURN_PHYS"
 *               is set.
 *               Lists the resources that is mapped to this region.
 *               The number of elements in the list shall be in phys_count.
 *
 */
struct mm_create_region
{
      SIGSELECT            sig_no;
      mm_status_t          status;
      U32                  size;
      OSDOMAIN             domain;
      char                 name[MM_NAME_LENGTH + 1];
      char                 phys_mem[MM_NAME_LENGTH + 1];
      char                 log_mem[MM_NAME_LENGTH + 1];
      U32                  permissions;
      U32                  cache_mode;
      OSADDRESS            base;
      U32                  align;
      U32                  options;
      U32                  low_limit;
      U32                  phys_count;      /* I: Length of phys_resources. */
      struct phys_resource phys_resources[1]; /* I: List of phys resources. */
};

/**
 * @signal MM_DELETE_REGION_REQUEST
 * @brief Signal used to delete a memory region.
 *
 * @long It is use to delete regions that were created with the create region
 * signal and regions that were made accessible with the share region signal.
 *
 * @long If physical resources were allocated when a region was created, the
 * physical resources will be freed when the last region that uses
 * these physical resources is deleted.
 * Observe that shared regions in a domain are also counted as a
 * user of the physical resources.
 *
 * @long Send signal to "ose_memory_mgr".
 *
 * @member mm_delete_region
 * @seealso MM_DELETE_REGION_REPLY
 * @seealso MM_CREATE_REGION_REQUEST
 * @seealso MM_SHARE_REGION_REQUEST
 */
#define MM_DELETE_REGION_REQUEST (36102) \
  /* !-SIGNO(struct mm_delete_region)-! */

/**
 *
 * @signal MM_DELETE_REGION_REPLY
 * @brief Reply signal from Memory Manager replying to
 *        @link "mm.sig" "MM_DELETE_REGION_REQUEST".
 *
 * @member mm_delete_region
 */
#define MM_DELETE_REGION_REPLY   (36103) \
  /* !-SIGNO(struct mm_delete_region)-! */

/**
 * @type mm_delete_region
 *
 * @field sig_no Signal number.
 * @input domain Domain identity.
 * @input base Base address of region.
 * @output status Status of the operation:
 *         MM_ENON_EXISTING_DOMAIN,
 *         MM_ENON_EXISTING_REGION,
 *         MM_ESYSTEM_REGION
 *         (see @link "mm.sig" "MmStatus").
 */
struct mm_delete_region
{
      SIGSELECT       sig_no;
      mm_status_t     status; /* O: Status of the operation. */
      OSDOMAIN        domain; /* I: Domain identity */
      OSADDRESS       base;   /* I: Base address of region. */
};

/**
 * @signal MM_CHANGE_PERMISSIONS_REQUEST
 * @brief Signal to change access permissions on a non-logical memory region.
 *
 * @long This call is only valid for regions that satisfy all the following
 * conditions:@break
 *   1.  The region must have physical memory.@break
 *   2.  The region must only occur in one domain, i.e. it must not be
 *       shared.
 *
 * @long Send signal to "ose_memory_mgr".
 *
 * @member mm_change_permissions
 * @seealso MM_CHANGE_PERMISSIONS_REPLY
 */
#define MM_CHANGE_PERMISSIONS_REQUEST (36104) \
  /* !-SIGNO(struct mm_change_permissions)-! */

/**
 * @signal MM_CHANGE_PERMISSIONS_REPLY
 * @brief Reply signal from Memory Manager replying to
 *        @link "mm.sig" "MM_CHANGE_PERMISSIONS_REQUEST".
 *
 * @member mm_change_permissions
 */
#define MM_CHANGE_PERMISSIONS_REPLY   (36105) \
  /* !-SIGNO(struct mm_change_permissions)-! */

/**
 * @type mm_change_permissions
 *
 * @field sig_no Signal number.
 * @input base Base address of region.
 * @input domain Domain identity.
 * @input permissions The permissions wanted.
 * @output status Status of the operation.
 * @output permissions Permissions before the change request.
 */
struct mm_change_permissions
{
      SIGSELECT   sig_no;
      mm_status_t status;           /* O: Status of the operation. */
      OSADDRESS   base;             /* I: Base address of region. */
      OSDOMAIN    domain;           /* I: Domain identity */
      U32         permissions;      /* I/O: New permissions/Old permissions */
};

/**
 * @signal MM_SHARE_REGION_REQUEST
 * @brief Signal used to share a memory region.
 *
 * @long When sharing a region, a base address of the region, the
 * domain of the region and the destination domain where the region
 * will be accessible should be specified.
 * Only regions that have physical resources associated with them can
 * be shared.
 *
 * @long When a region R that belongs to domain A is shared to domain B, this
 * has the following effect:
 *
 * @long
 * 1. Region R is accessible from user mode in domain B with the same
 *    access permissions as in domain A
 *
 * @long
 * 2. The physical resources associated with region R will not be freed
 *    until both domain B and A has been deleted.
 *
 * @long Number two is a consequence of the rule that states that physical
 * resources will not be reclaimed until all of the real or shared
 * regions have been deleted.
 *
 * @long It is possible to restrict the user mode permissions for the
 * destination domain. This is done by setting the
 * @link "mm.sig" "MM_SHARE_REGION_UPERM" option
 * in the options parameter in the request signal.
 * The use case is to share another domains data area with read only access
 * for user mode (compared to a normal share which gives write access.)
 *
 * @long The reason for that only user mode permissions in the destination
 * domain may be restricted is that sharing only should affect the destination
 * domain. Supervisor mode permissions should be the same from all domains,
 * and changing it in the destination domain would require changing it in
 * all other domains. The new user mode permission must more restrictive or
 * equal than the original user mode permissions of the region.
 *
 * @long If @link "mm.sig" "MM_SHARE_REGION_UPERM" is set, but the permission
 * has more permissions than the original, or if the permission required
 * changes in the supervisor mode permissions because of restricting the user
 * mode permissions could not be done without changing the supervisor mode
 * permissions, or the HAL could not support that specific permission,
 * status is set to MM_ESHARE_UMODE_RESTRICT_FAILED.
 * In this case, the returned user mode permissions has been set to zero.
 *
 * @long If the returned status is MM_SUCCESS, the resulting user mode
 * permissions in the destination domain is in the permission field in
 * the reply signal.
 *
 * @long Send signals to "ose_memory_mgr".
 *
 * @member mm_share_region
 * @seealso MM_SHARE_REGION_REPLY
 * @seealso MM_CREATE_REGION_REQUEST
 * @seealso MmStatus
 */
#define MM_SHARE_REGION_REQUEST (36106)  \
  /* !-SIGNO(struct mm_share_region)-! */

/**
 * @signal MM_SHARE_REGION_REPLY
 * @brief Reply signal from Memory Manager replying to
 *        @link "mm.sig" "MM_SHARE_REGION_REQUEST".
 *
 * @member mm_share_region
 */
#define MM_SHARE_REGION_REPLY   (36107)  \
  /* !-SIGNO(struct mm_share_region)-! */

/**
 * @macro MM_SHARE_REGION_UPERM
 * @brief Option value used when setting permissions in share requests
 * (@link "mm.sig" "MM_SHARE_REGION_REQUEST").
 */
#define MM_SHARE_REGION_UPERM   (0x1)

/**
 * @type mm_share_region
 *
 * @field sig_no Signal number.
 * @input base Logical base address of region.
 * @input source Domain that contains the mapped region.
 * @input destination Target domain.
 * @input options Share options. @link "mm.sig" "MM_SHARE_REGION_UPERM".
 * @input permission User mode permissions.
 *        @link "mm.sig" "MM_USR_EXECUTE",
 *        @link "mm.sig" "MM_USR_WRITE",
 *        @link "mm.sig" "MM_USR_READ"
 *        or use other combined permission macros.
 *
 * @output status Status of the operation,
 *         MM_ENON_EXISTING_DOMAIN,
 *         MM_ENON_EXISTING_REGION,
 *         MM_EREGION_LOGICAL.
 *         MM_EILLEGAL_DOMAIN.
 * @output permission User mode permissions set.
 */
struct mm_share_region
{
      SIGSELECT   sig_no;
      mm_status_t status;      /* O: Status of the operation */
      OSADDRESS   base;        /* I: Logical base address of region */
      OSDOMAIN    source;      /* I: Domain that contains the */
                               /*    mapped region */
      OSDOMAIN    destination; /* I: Target domain */
      U32         options;     /* I: Share options */
      U32         permission;  /* I/O: Restricted User mode permissions. */
};

/**
 * @signal MM_CREATE_DOMAIN_REQUEST
 * @brief Signal used to create a domain.
 *
 * @long To specify a name on the domain, OR the option parameter
 * with @link "mm.sig" "MM_CREATE_DOMAIN_SPECIFY_NAME" and copy the string to
 * the name parameter.
 *
 * @long A domain is uniquely identified by its domain identity.
 *
 * @long Send signal to "ose_memory_mgr".
 *
 * @member mm_create_domain
 * @seealso MM_CREATE_DOMAIN_REPLY
 */
#define MM_CREATE_DOMAIN_REQUEST (36108)  \
  /* !-SIGNO(struct mm_create_domain)-! */

/**
 * @signal MM_CREATE_DOMAIN_REPLY
 * @brief Reply signal from Memory Manager replying to
 *        @link "mm.sig" "MM_CREATE_DOMAIN_REQUEST".
 *
 * @member mm_create_domain
 */
#define MM_CREATE_DOMAIN_REPLY   (36109)  \
  /* !-SIGNO(struct mm_create_domain)-! */

/**
 * @macro MM_CREATE_DOMAIN_SPECIFY_NAME
 * @brief Option value used when specifying a name in a create domain request.
 */
#define MM_CREATE_DOMAIN_SPECIFY_NAME  (0x1)

/**
 * @type mm_create_domain
 *
 * @field sig_no Signal number.
 * @input name Name of domain.
 * @input options Options on the operation,
 *        @link "mm.sig" "MM_CREATE_DOMAIN_SPECIFY_NAME".
 * @output status Status of the operation,
 *         MM_EOUT_OF_DOMAINS.
 * @output domain Domain identity.
 */
struct mm_create_domain
{
      SIGSELECT   sig_no;
      mm_status_t status;                   /* O: Status of the operation. */
      char        name[MM_NAME_LENGTH + 1]; /* I: Name of domain. */
      OSDOMAIN    domain;                   /* O: Domain identity */
      U32         options;                  /* I: */
};

/**
 * @signal MM_DELETE_DOMAIN_REQUEST
 * @brief Signal used to delete a domain.
 *
 * @long This signal is used to delete a domain.
 * A domain is uniquely identified by its domain identity.
 * It is not required that the domain name is unique.
 *
 * @long Send signal to "ose_memory_mgr".
 *
 * @member mm_delete_domain
 * @seealso MM_DELETE_DOMAIN_REPLY
 * @seealso MM_CREATE_DOMAIN_REQUEST
 */
#define MM_DELETE_DOMAIN_REQUEST (36110)  \
  /* !-SIGNO(struct mm_delete_domain)-! */

/**
 * @signal MM_DELETE_DOMAIN_REPLY
 * @brief Reply signal from Memory Manager replying to
 *        @link "mm.sig" "MM_DELETE_DOMAIN_REQUEST".
 *
 * @member mm_delete_domain
 */
#define MM_DELETE_DOMAIN_REPLY   (36111)  \
  /* !-SIGNO(struct mm_delete_domain)-! */

/**
 * @type mm_delete_domain
 *
 * @field sig_no Signal number.
 * @input domain Domain identity.
 * @output status Status of the operation,
 *         MM_EILLEGAL_OPERATION,
 *         MM_ENON_EXISTING_DOMAIN.
 */
struct mm_delete_domain
{
      SIGSELECT       sig_no;
      mm_status_t     status;   /* O: Status of the operation. */
      OSDOMAIN        domain;   /* I: Domain identity */
};

/**
 * @signal MM_WRITE_REQUEST
 * @brief Signal used to write to a memory region area.
 *
 * @long Send signal to "ose_memory_mgr".
 *
 * @member mm_write
 * @seealso MM_WRITE_REPLY
 */
#define MM_WRITE_REQUEST (36112) \
  /* !-SIGNO(struct mm_write)-! */

/**
 * @signal MM_WRITE_REPLY
 * @brief Reply signal from Memory Manager replying to
 *        @link "mm.sig" "MM_WRITE_REQUEST".
 *
 * @member mm_write
 */
#define MM_WRITE_REPLY   (36113) \
  /* !-SIGNO(struct mm_write)-! */

/**
 * @type mm_write
 *
 * @field sig_no Signal number.
 * @input domain Domain that contains target region.
 * @input target Address in a region to write to.
 * @input source Source address of source buffer. The address must be directly
 *               accessible from the system domain, i.e. it must be mapped with
 *               either SASE or SAS mapping types, MAS is not supported.
 * @input size Size of source buffer.
 * @output status Status of the operation,
 *         MM_ESIZE_IS_ZERO,
 *         MM_ENON_EXISTING_DOMAIN,
 *         MM_ENON_EXISTING_REGION,
 *         MM_ENO_WRITE_PERMISSION,
 *         MM_ENO_READ_PERMISSION,
 *         MM_EWRITE_NOT_IMPLEMENTED,
 *         MM_EREGION_LOGICAL,
 *         MM_EMAS_NOT_SUPPORTED.
 */
struct mm_write
{
      SIGSELECT   sig_no;
      mm_status_t status;     /* O: Status of the operation. */
      OSDOMAIN    domain;     /* I: Domain that contains target region */
      OSADDRESS   target;     /* I: Address in a region to write to */
      OSADDRESS   source;     /* I: Source address of source buffer */
      U32         size;       /* I: Size of source buffer */
};

/*
 ****************************************************************************
 * 7 INFORMATION SIGNALS
 ****************************************************************************
 */

/**
 * @signal MM_INTERFACE_INFO_REQUEST
 * @brief Signal to acquire information about the MM signal and trap interface.
 *
 * @long This signal is sent to Memory Manager to get information about
 * which signal numbers and trap handles the Memory Manager uses.
 *
 * @long Send signal to "ose_memory_mgr".
 *
 * @member mm_interface_info_request
 *
 * @seealso mm_interface_info_reply
 */
#define MM_INTERFACE_INFO_REQUEST (36114)	 \
  /* !-SIGNO(struct mm_interface_info_request)-! */

/**
 * @signal MM_INTERFACE_INFO_REPLY
 * @brief Reply signal from Memory Manager replying to
 *        @link "mm.sig" "MM_INTERFACE_INFO_REQUEST".
 *
 * @member mm_interface_info_reply
 *
 * @seealso mm_interface_info_request
 */
#define MM_INTERFACE_INFO_REPLY (36115)	 \
  /* !-SIGNO(struct mm_interface_info_reply)-! */

/**
 * @type mm_interface_info_request
 *
 * @field sig_no Signal number.
 */
struct mm_interface_info_request
{
      SIGSELECT       sig_no;
};

/**
 * @type mm_interface_info_reply
 *
 * @field sig_no Signal number.
 * @output status Return status code.
 * @output what_str MM what string.
 * @output mmu_mode True if MM uses the MMU.
 * @output sigs Supported signals. Variable size.
 */
struct mm_interface_info_reply
{
      SIGSELECT       sig_no;
      mm_status_t     status;	    /* O: Status of the operation. */
      char            what_str[32]; /* O: What string of product. */
      U32             mmu_mode;     /* O: True if MM actively uses the MMU. */
      SIGSELECT       sigs[1];	    /* O: Supported signals. Variable size. */
};

/**
 * @signal MM_LIST_DOMAINS_REQUEST
 * @brief Signal to retrieve a list of all domains.
 *        The first element is the kernel domain.
 *
 * @long Send signal to "ose_memory_mgr".
 *
 * @member mm_list_domains_request
 *
 * @seealso MM_LIST_DOMAINS_REPLY
 */
#define MM_LIST_DOMAINS_REQUEST (36116)  \
  /* !-SIGNO(struct mm_list_domains_request)-! */

/**
 * @signal MM_LIST_DOMAINS_REPLY
 * @brief Reply signal from Memory Manager replying to
 *        @link "mm.sig" "MM_LIST_DOMAINS_REQUEST".
 *
 * @member mm_list_domains_reply
 *
 * @seealso MM_LIST_DOMAINS_REQUEST
 */
#define MM_LIST_DOMAINS_REPLY   (36117)  \
  /* !-SIGNO(struct mm_list_domains_reply)-! */

/**
 * @type mm_list_domains_request
 *
 * @field sig_no Signal number.
 */
struct mm_list_domains_request
{
      SIGSELECT sig_no;
};

/**
 * @type mm_list_domains_reply
 *
 * @field sig_no Signal number.
 * @output count Number of returned domains.
 * @output domains Domain identities of all domains. Variable size.
 */
struct mm_list_domains_reply
{
      SIGSELECT sig_no;
      U32	count;      /* O: Number of returned domains */
      OSDOMAIN  domains[1]; /* O: Domain identities of all domains. */
};

/**
 * @signal MM_LIST_REGIONS_REQUEST
 * @brief Signal to retrieve a list of all regions of a domain.
 *
 * @long The returned list contains a list of the unique identifiers
 *       of all regions in the domain.
 *
 * @long If wanting to limit the list to only inactive regions,
 *       use instead the signal
 *       @link "mm.sig" "MM_LIST_INACTIVE_REGIONS_REQUEST".
 *
 * @long Send signal to "ose_memory_mgr".
 *
 * @member mm_list_regions_request
 *
 * @seealso MM_LIST_REGIONS_REPLY
 * @seealso MM_LIST_INACTIVE_REGIONS_REQUEST
 */
#define MM_LIST_REGIONS_REQUEST (36118)  \
  /* !-SIGNO(struct mm_list_regions_request)-! */

/**
 *
 * @signal MM_LIST_REGIONS_REPLY
 * @brief Reply signal from Memory Manager replying to
 *        @link "mm.sig" "MM_LIST_REGIONS_REQUEST".
 *
 * @long  The returned list contains a list of the unique identifiers
 *        of all regions in the domain.
 *
 * @member mm_list_regions_reply
 *
 * @seealso MM_LIST_REGIONS_REQUEST
 *
 */
#define MM_LIST_REGIONS_REPLY   (36119)  \
  /* !-SIGNO(struct mm_list_regions_reply)-! */

/**
 * @type mm_list_regions_request
 *
 * @field sig_no Signal number.
 * @input domain The domain where a list of regions is requested.
 *
 */
struct mm_list_regions_request
{
      SIGSELECT sig_no;
      OSDOMAIN  domain;     /* I: Domain requested */
};

/**
 * @type mm_list_regions_reply
 *
 * @field sig_no Signal number.
 * @output status Status of the operation.
 * @output count Number of regions.
 * @output regions Base addresses of returned regions. Variable size.
 *
 */
struct mm_list_regions_reply
{
      SIGSELECT   sig_no;
      mm_status_t status;     /* O: Status of the operation. */
      U32	  count;      /* O: Number of regions */
      OSADDRESS   regions[1]; /* O: Base address of returned regions */
};

/**
 * @signal MM_LIST_LOG_MEM_REQUEST
 * @brief Signal to retrieve a list of names of all logical memory areas.
 *
 * @long Logical Memory Areas are identified by the names.
 * The returned signal contains an array with all name strings in sequence.
 * Each string is null-terminated.
 *
 * @long Send signal to "ose_memory_mgr".
 *
 * @member mm_list_log_mem_request
 *
 * @seealso MM_LIST_LOG_MEM_REPLY
 */
#define MM_LIST_LOG_MEM_REQUEST (36120)  \
  /* !-SIGNO(struct mm_list_log_mem_request)-! */

/**
 * @signal MM_LIST_LOG_MEM_REPLY
 * @brief Reply signal from Memory Manager replying to
 *        @link "mm.sig" "MM_LIST_LOG_MEM_REQUEST".
 *
 * @member mm_list_log_mem_reply
 *
 * @seealso MM_LIST_LOG_MEM_REQUEST
 */
#define MM_LIST_LOG_MEM_REPLY   (36121)  \
  /* !-SIGNO(struct mm_list_log_mem_reply)-! */

/**
 * @type mm_list_log_mem_request
 *
 * @field sig_no Signal number.
 */
struct mm_list_log_mem_request
{
      SIGSELECT sig_no;
};

/**
 * @type mm_list_log_mem_reply
 *
 * @field sig_no Signal number.
 * @output count Number of logical memory areas.
 * @output names List of logical memory area names. Variable size.
 */
struct mm_list_log_mem_reply
{
      SIGSELECT sig_no;
      U32       count;      /* O: Number of logical memory areas */
      char      names[1];   /* O: List of logical memory area names. */
};

/**
 * @signal MM_LIST_PHYS_MEM_REQUEST
 * @brief Signal retrieving a list of all physical memory types.
 *
 * @long Physical Memory Types are identified by names.
 * The returned signal contains an array with all
 * name strings in sequence. Each string is null-terminated.
 *
 * @long Send signal to "ose_memory_mgr".
 *
 * @member mm_list_phys_mem_request
 *
 * @seealso MM_LIST_PHYS_MEM_REPLY
 */

#define MM_LIST_PHYS_MEM_REQUEST (36122)  \
  /* !-SIGNO(struct mm_list_phys_mem_request)-! */

/**
 * @signal MM_LIST_PHYS_MEM_REPLY
 * @brief Reply signal from Memory Manager replying to
 *        @link "mm.sig" "MM_LIST_PHYS_MEM_REQUEST".
 *
 * @member mm_list_phys_mem_reply
 *
 * @seealso MM_LIST_PHYS_MEM_REQUEST
 */
#define MM_LIST_PHYS_MEM_REPLY   (36123)  \
  /* !-SIGNO(struct mm_list_phys_mem_request)-! */

/**
 * @type mm_list_phys_mem_request
 *
 * @field sig_no Signal number.
 */
struct mm_list_phys_mem_request
{
      SIGSELECT sig_no;
};

/**
 * @type mm_list_phys_mem_reply
 *
 * @field sig_no Signal number.
 * @output count Number of physical memory types.
 * @output names List of physical memory type names.
 */
struct mm_list_phys_mem_reply
{
      SIGSELECT sig_no;
      U32       count;      /* O: Number of physical memory types */
      char      names[1];   /* O: List of physical memory type names. */
};

/**
 * @signal MM_GET_REGION_INFO_REQUEST
 * @brief Signal that returns information about a region.
 *
 * @long The same type of struct 'mm_create_region' is used
 *       as when creating a region.
 *       See the "input" field descriptions in
 *       @link "mm.sig" "MM_CREATE_REGION_REQUEST".
 *
 * @long Only the <base> and <domain> input parameters need to be
 *       specified in the request.
 *
 * @long MM can reply with multiple reply signals with sig_no
 *       @link "mm.sig" "MM_GET_REGION_INFO_NOT_LAST",
 *       followed by the last reply signal, which will instead
 *       have sig_no MM_GET_REGION_INFO_REPLY.
 *
 * @seealso MM_GET_REGION_INFO_REPLY
 * @seealso MM_CREATE_REGION_REQUEST
 */
#define MM_GET_REGION_INFO_REQUEST (36124)  \
  /* !-SIGNO(struct mm_create_region)-! */

/**
 * @signal MM_GET_REGION_INFO_REPLY
 * @brief Last reply signal from Memory Manager replying to
 *        @link "mm.sig" "MM_GET_REGION_INFO_REQUEST".
 *
 * @long MM can reply with multiple reply signals with sig_no
 *       @link "mm.sig" "MM_GET_REGION_INFO_NOT_LAST",
 *       followed by the last reply signal, which will instead
 *       be this signal with sig_no MM_GET_REGION_INFO_REPLY.
 *
 * @long The same type of struct 'mm_create_region' is returned
 *       as in the reply from creating a region.
 *       See the "output" field descriptions in
 *       @link "mm.sig" "MM_CREATE_REGION_REQUEST".
 *
 * @seealso MM_CREATE_REGION_REQUEST
 * @seealso MM_GET_REGION_INFO_NOT_LAST
 */
#define MM_GET_REGION_INFO_REPLY   (36125)  \
  /* !-SIGNO(struct mm_create_region)-! */

/**
 * @signal MM_TRANSLATE_ADDRESS_REQUEST
 * @brief Signal to retrieve the physical address of the specified logical
 *        address in a specified domain.
 *
 * @long Send signal to ose_memory_mgr".
 *
 * @member mm_translate_address
 * @seealso MM_TRANSLATE_ADDRESS_REPLY
 */
#define MM_TRANSLATE_ADDRESS_REQUEST (36126)  \
  /* !-SIGNO(struct mm_translate_address)-! */

/**
 * @signal MM_TRANSLATE_ADDRESS_REPLY
 * @brief Reply signal from Memory Manager replying to
 *        @link "mm.sig" "MM_TRANSLATE_ADDRESS_REQUEST".
 *
 * @member mm_translate_address
 */
#define MM_TRANSLATE_ADDRESS_REPLY   (36127)  \
  /* !-SIGNO(struct mm_translate_address)-! */

/**
 * @type mm_translate_address
 *
 * @field sig_no Signal number.
 * @input log_addr Logical address to be translated.
 * @input domain Domain to translate the address.
 *
 * @output status Status of the operation,
 *         MM_EREGION_LOGICAL,
 *         MM_ENON_EXISTING_REGION,
 *         MM_ENON_EXISTING_DOMAIN.
 *
 * @output phys_addr Physical address that corresponds to logical address.
 *
 * @output phys_page Physical base address for MMU page.
 *       NOTE: The phys_page parameter is not set in the returned
 *       signal, and is likely to be replaced in later versions of MM.
 *
 * @output log_page Logical base address for MMU page.
 *       NOTE: The log_page parameter is not set in the returned
 *       signal, and is likely to be replaced in later versions of MM.
 */
struct mm_translate_address
{
      SIGSELECT   sig_no;
      mm_status_t status;    /* O: Status of the operation. */
      OSADDRESS   log_addr;  /* I: Logical address that need to be */
                             /*    translated. */
      OSDOMAIN    domain;    /* I: Domain that translates the */
			     /*    address. */
      phys_addr_t phys_addr; /* O: Physical address that corresponds to */
			     /*    logical address */
      phys_addr_t phys_page; /* O: Physical base address for MMU page */
      OSADDRESS   log_page;  /* O: Logical base address for MMU page */
};

/**
 * @signal MM_LIST_PAGE_SIZES_REQUEST
 * @brief Signal retrieving an array of supported MMU page sizes.
 *
 * @long Send signal to "ose_memory_mgr".
 *
 * @member mm_list_page_sizes_request
 *
 * @seealso MM_LIST_PAGE_SIZES_REPLY
 */
#define MM_LIST_PAGE_SIZES_REQUEST (36128)  \
  /* !-SIGNO(struct mm_list_page_sizes_request)-! */

/**
 * @signal MM_LIST_PAGE_SIZES_REPLY
 * @brief Reply signal from Memory Manager replying to
 *        @link "mm.sig" "MM_LIST_PAGE_SIZES_REQUEST".
 *
 * @member mm_list_page_sizes_reply
 *
 * @seealso MM_LIST_PAGE_SIZES_REQUEST
 */
#define MM_LIST_PAGE_SIZES_REPLY   (36129)  \
  /* !-SIGNO(struct mm_list_page_sizes_reply)-! */

/**
 * @type mm_list_page_sizes_request
 *
 * @field sig_no Signal number.
 */

struct mm_list_page_sizes_request
{
      SIGSELECT sig_no;
};

/**
 * @type mm_list_page_sizes_reply
 *
 * @field sig_no Signal number.
 * @output count Number of page sizes in array.
 * @output sizes Array containing all supported sizes. Variable size.
 *
 */
struct mm_list_page_sizes_reply
{
      SIGSELECT sig_no;
      U32       count;    /* O: Number of page sizes in array */
      U32       sizes[1]; /* O: Array containing all supported sizes */
};

/**
 * @signal MM_GET_MMU_STATISTICS_REQUEST
 * @brief Signal to retrieve statistics about the MMU.
 *
 * @long When HW resources are needed the statistic measuring is enabled
 *       only after the first time that MM receives the
 *       signal MM_GET_MMU_STATISTICS_REQUEST.
 *       HW resources could for example be performance monitor
 *       registers.
 *       The idea is to not occupy these HW resources in systems not using
 *       statistics. @link "mmtypes.h" "MM_TAG_HW_COUNTER".
 *
 * @seealso mm_tag_ext
 * @seealso MmTagId
 *
 * @member mm_get_mmu_statistics
 *
 * @seealso MM_GET_MMU_STATISTICS_REPLY
 */
#define MM_GET_MMU_STATISTICS_REQUEST (36130)  \
  /* !-SIGNO(struct mm_get_mmu_statistics)-! */

/**
 * @signal MM_GET_MMU_STATISTICS_REPLY
 * @brief Reply signal from Memory Manager replying to
 *        @link "mm.sig" "MM_GET_MMU_STATISTICS_REQUEST".
 *
 * @member mm_get_mmu_statistics
 */
#define MM_GET_MMU_STATISTICS_REPLY   (36131)  \
  /* !-SIGNO(struct mm_get_mmu_statistics)-! */

/**
 * @type mm_get_mmu_statistics
 *
 * @field sig_no Signal number.
 * @input count Number of tags queried.
 * @input tags List of tag queries. Variable size.@break
 *             A special case is when the first tag ID field is: @break
 *             tags[0].tag_id == MM_INFOTAG_MMU_ALL.
 *             @break In this case MM will allocate a reply signal and fill
 *             in all available statistics.
 *             This allows clients to be able to retrieve
 *             all available statistics without the need to be updated every
 *             time a new MmTagId is added.
 *
 *
 * @output status Status of the operation.@break
 * MM_SUCCESS @break
 * MM_EILLEGAL_TAG.@break
 * MM_EINFOTAG_UNSUPPORTED_FOR_CPU @break
 * MM_ETOO_SMALL_SIGNAL_SIZE @break
 * MM_ESIGNAL_NOT_IMPLEMENTED @break
 *
 *
 * @output err_index Index of the first illegal or unsupported tag.
 * @output count Number of tag answers returned.
 * @output tags Answer in tag format. Variable size.
 *
 * @seealso mm_tag_ext
 * @seealso MmTagId
 */
struct mm_get_mmu_statistics
{
      SIGSELECT     sig_no;
      mm_status_t   status;    /* O: Status of the operation. */
      U32           count;     /* I/O: Number of tags */
      U32           err_index; /* O: Index of first illegal tag */
      struct mm_tag_ext tags[1];   /* I/O: Queries/Answers in tag format */
};

/**
 * @signal MM_GET_PHYS_MEM_INFO_REQUEST
 * @brief Get detailed physical memory type information.
 *
 * @member mm_get_phys_mem_info
 *
 * @seealso MM_GET_PHYS_MEM_INFO_REPLY
 */
#define MM_GET_PHYS_MEM_INFO_REQUEST (36132)  \
  /* !-SIGNO(struct mm_get_phys_mem_info)-! */

/**
 * @signal MM_GET_PHYS_MEM_INFO_REPLY
 * @brief Reply signal from Memory Manager replying to
 *        @link "mm.sig" "MM_GET_PHYS_MEM_INFO_REQUEST".
 *
 * @member mm_get_phys_mem_info
 */
#define MM_GET_PHYS_MEM_INFO_REPLY   (36133)  \
  /* !-SIGNO(struct mm_get_phys_mem_info)-! */

/**
 * @type mm_get_phys_mem_info
 *
 * @field sig_no Signal number.
 * @input name Name of physical memory of request.
 * @input count Number of query tags in array.
 * @input tags List of query tags. Variable size.
 *
 * @output status Status of the operation.
 *         MM_EILLEGAL_TAG,
 *         MM_ENON_EXISTING_PHYS_MEM.
 *
 * @output err_index Index of first illegal tag.
 * @output count Number of queries in tags array.
 * @output tags List of answer tags.
 */
struct mm_get_phys_mem_info
{
      SIGSELECT     sig_no;
      mm_status_t   status;    /* O: Status of the operation. */
      char          name[MM_NAME_LENGTH + 1]; /* I: Name of physical memory. */
      U32           count;     /* I: Number of query tags in the */
      U32           err_index; /* O: Index of first illegal tag. */
      struct mm_tag tags[1];   /* I/O: Tag list. */
};


/**
 * @signal MM_GET_DOMAIN_INFO_REQUEST
 * @brief Get detailed domain information.
 *
 * @restriction NOTE: This signal has not been implemented. The returning
 *       signal has always the status code MM_ILLEGAL_TAG.
 *
 * @member mm_get_domain_info
 *
 * @seealso MM_GET_DOMAIN_INFO_REPLY
 */
#define MM_GET_DOMAIN_INFO_REQUEST (36134)  \
  /* !-SIGNO(struct mm_get_domain_info)-! */

/**
 * @signal MM_GET_DOMAIN_INFO_REPLY
 * @brief Reply signal from Memory Manager replying to
 *        @link "mm.sig" "MM_GET_DOMAIN_INFO_REQUEST".
 *
 * @member mm_get_domain_info
 */
#define MM_GET_DOMAIN_INFO_REPLY   (36135)  \
  /* !-SIGNO(struct mm_get_domain_info)-! */

/**
 * @type mm_get_domain_info
 *
 * @input domain Domain to request info for.
 * @input count Number of query tags in tags array.
 * @input tags Queries in tag format. Variable size.
 *
 * @field sig_no Signal number.
 * @output status Status of the operation,
 *         MM_ENON_EXISTING_DOMAIN,
 *         MM_EILLEGAL_TAG.
 *
 * @output name Domain name.
 * @output err_index Index of first illegal tag.
 * @output count Number of answers i tags array.
 * @output tags Answers in tag format. Variable size.
 */
struct mm_get_domain_info
{
      SIGSELECT     sig_no;
      mm_status_t   status;    /* O: Status of the operation. */
      OSDOMAIN      domain;    /* I: Protection domain */
      char          name[MM_NAME_LENGTH + 1]; /* O: Domain name */
      U32           count;     /* I: Tag count. */
      U32           err_index; /* O: Index of first illegal tag */
      struct mm_tag tags[1];   /* I: Tag list. */
};


/**
 * @signal MM_GET_LOG_MEM_INFO_REQUEST
 * @brief Get detailed logical memory area information.
 *
 * @member mm_get_log_mem_info
 *
 * @seealso MM_GET_LOG_MEM_INFO_REPLY
 */
#define MM_GET_LOG_MEM_INFO_REQUEST (36136)  \
  /* !-SIGNO(struct mm_get_log_mem_info)-! */

/**
 * @signal MM_GET_LOG_MEM_INFO_REPLY
 * @brief Reply signal from Memory Manager replying to
 *        @link "mm.sig" "MM_GET_LOG_MEM_INFO_REQUEST".
 *
 * @member mm_get_log_mem_info
 */
#define MM_GET_LOG_MEM_INFO_REPLY   (36137)  \
  /* !-SIGNO(struct mm_get_log_mem_info)-! */

/**
 * @type mm_get_log_mem_info
 *
 * @field sig_no Signal number.
 * @input name Name of logical memory area to request info for.
 * @input count Number of query tags in tags array.
 * @input tags Queries in tag format. Variable size.
 *
 * @output status Status of the operation,
 *         MM_ENON_EXISTING_PARTITION,
 *         MM_EILLEGAL_TAG.
 *
 * @output err_index Index of first illegal tag.
 * @output count Number of answers i tags array.
 * @output tags Answers in tag format. Variable size.
 *
 */
struct mm_get_log_mem_info
{
      SIGSELECT     sig_no;
      mm_status_t   status;   /* O: Status of the operation. */
      char          name[MM_NAME_LENGTH + 1];  /* I: Logmem name */
      U32           count;     /* I: Tag count. */
      U32           err_index; /* O: Index of first illegal tag */
      struct mm_tag tags[1];   /* I/O: Tag list. */
};


/**
 * @signal MM_GET_LOG_MEM_BANK_INFO_REQUEST
 * @brief Get detailed logical memory area information.
 *
 * @member mm_get_log_mem_bank_info
 *
 * @seealso MM_GET_LOG_MEM_BANK_INFO_REPLY
 */
#define MM_GET_LOG_MEM_BANK_INFO_REQUEST (36155)  \
  /* !-SIGNO(struct mm_get_log_mem_bank_info)-! */

/**
 * @signal MM_GET_LOG_MEM_BANK_INFO_REPLY
 * @brief Reply signal from Memory Manager replying to
 *        @link "mm.sig" "MM_GET_LOG_MEM_BANK_INFO_REQUEST".
 *
 * @member mm_get_log_mem_bank_info
 */
#define MM_GET_LOG_MEM_BANK_INFO_REPLY   (36156)  \
  /* !-SIGNO(struct mm_get_log_mem_bank_info)-! */

/**
 * @type mm_get_log_mem_bank_info
 *
 * @field sig_no Signal number.
 * @input name Name of logical memory area to request info for.
 * @input count Number of query tags in tags array.
 * @input tags Queries in tag format. Variable size.
 * @input bank Index, starting from 0, selecting the requested bank.
 *
 * @output status Status of the operation,
 *         MM_ENON_EXISTING_PARTITION,
 *         MM_EILLEGAL_TAG.
 *
 * @output err_index Index of first illegal tag.
 * @output count Number of answers i tags array.
 * @output tags Answers in tag format. Variable size.
 *
 */
struct mm_get_log_mem_bank_info
{
      SIGSELECT     sig_no;
      mm_status_t   status;   /* O: Status of the operation. */
      char          name[MM_NAME_LENGTH + 1];  /* I: Logmem name */
      U32           bank;      /* I: Bank index */
      U32           count;     /* I: Tag count. */
      U32           err_index; /* O: Index of first illegal tag */
      struct mm_tag tags[1];   /* I/O: Tag list. */
};



/* NOT_USED_SOURCEDOC
 *
 * @signal MM_REMAP_REGION_REQUEST
 *
 * @brief Signal used to remap parts of an existing memory region to
 *        new physical memory.
 *
 * @restriction NOTE: This signal interface is neither implemented
 *        nor supported.
 *
 * @long This signal remaps parts of an existing region to new
 *       physical memory, located at the same logical address.
 *
 * @long New memory will be allocated from the physical memory type
 *       designated by <phys_mem>.
 *
 * @long The contents of the original memory (located at
 *       <base> up to <base>+<size>) will be copied to the new memory.
 *
 * @long The area designated by the parameters <base>,<size> will be
 *       made writable from supervisor mode.
 *
 * @long The remap parameter <size> must be a supported page size (as
 *       returned by the signal
 *       @link "mm.sig" "MM_LIST_PAGE_SIZES_REQUEST").
 *
 * @long A remap operation is visible in all domains.
 *
 * @long Send signal to "ose_memory_mgr"
 *
 * @member mm_remap_region
 *
 * @seealso MM_REMAP_REGION_REPLY
 * @seealso MM_CREATE_REGION_REQUEST
 */
#define MM_REMAP_REGION_REQUEST (36138)  \
  /* !-SIGNO(struct mm_remap_region)-! */

/* NOT_USED_SOURCEDOC
 *
 * @signal MM_REMAP_REGION_REPLY
 * @brief Reply signal from Memory Manager replying to
 *        @link "mm.sig" "MM_REMAP_REGION_REQUEST".
 *
 * @restriction This signal interface is not implemented.
 *
 * @member mm_remap_region
 */
#define MM_REMAP_REGION_REPLY (36139)  \
  /* !-SIGNO(struct mm_remap_region)-! */

/* NOT_USED_SOURCEDOC
 *
 * @type mm_remap_region
 *
 * @field sig_no Signal number.
 *
 * @input domain Domain identity of region to be remapped.
 *               Mandatory parameter.
 *
 * @input base   Base address of region to be remapped.
 *               Mandatory parameter.
 *
 * @input size   Size of area to be remapped. Must be equal to smallest
 *               supported page size. Mandatory parameter.
 *
 * @input phys_mem Name of a physical memory type to allocate physical
 *               memory from. Mandatory parameter.
 *
 * @input options Future extensions.
 *
 * @output status Success or failure reason of operation,
 *         MM_ENON_EXISTING_REGION
 *         MM_ENON_EXISTING_DOMAIN
 *         MM_ENON_EXISTING_PHYS_TYPE
 *         MM_EILLEGAL_PHYS_TYPE
 *         MM_EREGION_LOGICAL
 *         MM_EILLEGAL_OPERATION
 *         MM_EILLEGAL_AREA
 *         MM_EOVERLAPPING_REMAP_ENTRY
 *
 */
struct mm_remap_region
{
      SIGSELECT   sig_no;
      mm_status_t status;                 /* O: Status of the operation. */
      OSDOMAIN    domain;                 /* I: Domain identity of region. */
      OSADDRESS   base;                   /* I: Base address of remap area. */
      U32         size;       /* I: Size of remap area. */
      char        phys_mem[MM_NAME_LENGTH + 1]; /* Remap memory. */
      U32         options;                /* I: Bit field with options. */
};



/* NOT_USED_SOURCEDOC
 *
 * @signal MM_UNREMAP_REGION_REQUEST
 * @brief Signal to delete a remap area in a region.
 *
 * @restriction This signal is neither implemented nor supported.
 *
 * @long This signal deletes remap parts of an existing region.
 * The physical memory will be restored to the state before the remap
 * operation.
 *
 * @long Send the signal to "ose_memory_mgr"
 *
 * @member mm_unremap_region
 * @seealso MM_UNREMAP_REGION_REPLY
 * @seealso MM_REMAP_REGION_REQUEST
 */
#define MM_UNREMAP_REGION_REQUEST (36140)  \
  /* !-SIGNO(struct mm_unremap_region)-! */

/* NOT_USED_SOURCEDOC
 *
 * @signal MM_UNREMAP_REGION_REPLY
 * @brief Reply signal from Memory Manager replying to
 *        @link "mm.sig" "MM_UNREMAP_REGION_REQUEST".
 *
 * @restriction This signal is neither implemented nor supported.
 *
 * @member mm_unremap_region
 */
#define MM_UNREMAP_REGION_REPLY (36141)  \
  /* !-SIGNO(struct mm_unremap_region)-! */

/* NOT_USED_SOURCEDOC
 *
 * @type mm_unremap_region
 *
 * @field sig_no Signal number.
 * @input domain Domain identity of region to be unremapped.
 *               Mandatory parameter.
 *
 * @input base   Base address of area to be restored.
 *               Mandatory parameter.
 *
 * @output status Success or failure reason of operation,
 *         MM_ENON_EXISTING_REGION
 *         MM_ENON_EXISTING_DOMAIN
 *         MM_ENON_EXISTING_REMAP_ENTRY
 *
 */
struct mm_unremap_region
{
      SIGSELECT   sig_no;
      mm_status_t status;     /* O: Status of the operation. */
      OSDOMAIN    domain;     /* I: Domain identity of region. */
      OSADDRESS   base;       /* I: Base address of region. */
};


/**
 * @signal MM_GET_REGION_INFO_NOT_LAST
 * @brief Reply signal (not the last) from Memory Manager replying to
 *        @link "mm.sig" "MM_GET_REGION_INFO_REQUEST".
 *
 * @long MM can reply with multiple reply signals with sig_no
 *       @link "mm.sig" "MM_GET_REGION_INFO_NOT_LAST",
 *       followed by the last reply signal, which will instead
 *       have sig_no MM_GET_REGION_INFO_REPLY.
 *
 * @long The same type of struct 'mm_create_region' is returned
 *       as in the reply from creating a region.
 *       See the "output" field descriptions in
 *       @link "mm.sig" "MM_CREATE_REGION_REQUEST".
 *
 * @seealso MM_CREATE_REGION_REQUEST
 * @seealso MM_GET_REGION_INFO_REPLY
 */
#define MM_GET_REGION_INFO_NOT_LAST   (36142)  \
  /* !-SIGNO(struct mm_create_region)-! */


/**
 * @signal MM_LIST_INACTIVE_REGIONS_REQUEST
 * @brief Signal to retrieve a list of all "inactive" regions of a domain.
 *
 * @long The returned list contains a list of the unique identifiers
 *        of all "inactive" regions in the domain.
 *
 * @long NOTE: If instead wanting a list of "all" regions in the domain,
 *       use the signal @link "mm.sig" "MM_LIST_REGIONS_REQUEST".
 *
 * @member mm_list_regions_request
 *
 * @seealso MM_LIST_INACTIVE_REGIONS_REPLY
 * @seealso MM_LIST_REGIONS_REQUEST
 */
#define MM_LIST_INACTIVE_REGIONS_REQUEST (36143)  \
  /* !-SIGNO(struct mm_list_regions_request)-! */

/**
 *
 * @signal MM_LIST_INACTIVE_REGIONS_REPLY
 * @brief Reply signal from Memory Manager replying to
 *        @link "mm.sig" "MM_LIST_INACTIVE_REGIONS_REQUEST".
 *
 * @member mm_list_regions_reply
 *
 * @seealso MM_LIST_REGIONS_REQUEST
 */
#define MM_LIST_INACTIVE_REGIONS_REPLY   (36144)  \
  /* !-SIGNO(struct mm_list_regions_reply)-! */

/**
 * @signal MM_GET_PHYS_MEM_TYPE_INFO_REQUEST
 * @brief Get detailed physical memory type information.
 *
 * @member mm_get_phys_mem_type_info
 *
 * @seealso MM_GET_PHYS_MEM_TYPE_INFO_REPLY
 */
#define MM_GET_PHYS_MEM_TYPE_INFO_REQUEST (36147)  \
  /* !-SIGNO(struct mm_get_phys_mem_type_info)-! */

/**
 * @signal MM_GET_PHYS_MEM_TYPE_INFO_REPLY
 * @brief Reply signal from Memory Manager replying to
 *        @link "mm.sig" "MM_GET_PHYS_MEM_TYPE_INFO_REQUEST".
 *
 * @member mm_get_phys_mem_type_info
 */
#define MM_GET_PHYS_MEM_TYPE_INFO_REPLY   (36148)  \
  /* !-SIGNO(struct mm_get_phys_mem_type_info)-! */

/**
 * @type mm_get_phys_mem_type_info
 *
 * @field sig_no Signal number.
 * @input name Name of physical memory type of request.
 * @input count Number of query tags in array.
 * @input info List of information tag-value pairs. Variable size.
 *
 * @output status Status of the operation.
 *         MM_EILLEGAL_TAG,
 *         MM_ENON_EXISTING_PHYS_MEM.
 *
 * @output err_index Index of first illegal information tag-value pair.
 * @output count Number of queries in array containing tag-value pairs.
 * @output info Array of information tag-value pairs.
 */
struct mm_get_phys_mem_type_info
{
      SIGSELECT   sig_no;
      mm_status_t status;       /* O: Status of the operation. */
      char        name[MM_NAME_LENGTH + 1]; /* I: Name of physical memory. */
      U32         count;        /* I: Number of query tags. */
      U32         err_index;    /* O: Index of first illegal
                                 *    tag-value pair. */
      struct mm_phys_mem_info info[1]; /* I/O: List of tag-value pairs. */
};

/**
 * @signal MM_MAKE_RESIDENT_REQUEST
 * @brief Request that the supplied logical area be made resident
 * (i.e. always mapped by MMU).
 *
 * @member mm_make_resident
 *
 * @long The request for making a memory area resident must match
 * the resources used by the target architecture,
 * otherwise the status code returned will be MM_EILLEGAL_AREA.
 *
 * @long This operation requires that the memory area to be made
 * resident be inside a SASE mapped region created in the system domain,
 * otherwise the status code returned will be MM_EILLEGAL_DOMAIN.
 *
 * @long Making regions resident effectively disables the inter-domain
 * memory protection mechanism. The resident area will in large parts
 * behave according to the OSE NOMMU memory model.
 *
 * @long The memory manager will automatically release resident
 * hardware resources associated with a region if the region is
 * deleted. The resources are also released if region permissions
 * change or if the region is shared with restricted user mode
 * permissions.
 *
 * @long If a region that has been shared with restricted user mode
 * permissions is made resident, the user mode restriction
 * will be neglected.
 */
#define MM_MAKE_RESIDENT_REQUEST (36149) \
  /* !-SIGNO(struct mm_make_resident)-! */

/**
 * @signal MM_MAKE_RESIDENT_REPLY
 *
 * @brief Reply signal from Memory Manager replying to
 *        @link "mm.sig" "MM_MAKE_RESIDENT_REQUEST".
 *
 * @member mm_make_resident
 */
#define MM_MAKE_RESIDENT_REPLY (36150) \
  /* !-SIGNO(struct mm_make_resident)-! */

/**
 * @macro MM_MAKE_RESIDENT_CODE
 * @brief Option bit, set in
 *        @link "mm.sig" "MM_MAKE_RESIDENT_REQUEST"
 *        if MM shall make a memory area containing code resident.
 * @seealso MM_MAKE_RESIDENT_REQUEST
 */
#define MM_MAKE_RESIDENT_CODE (0x1)

/**
 * @macro MM_MAKE_RESIDENT_DATA
 * @brief Option bit, set in
 *        @link "mm.sig" "MM_MAKE_RESIDENT_REQUEST"
 *        if MM shall make a memory area containing data resident.
 * @seealso MM_MAKE_RESIDENT_REQUEST
 */
#define MM_MAKE_RESIDENT_DATA (0x2)

/**
 * @type mm_make_resident
 *
 * @field sig_no  Signal number.
 * @input base    Logical base address of memory area to make resident.
 * @input size    Size of memory area to make resident.
 * @input options Bit field containing options for the requested memory area,
 *                i.e. whether the area to make resident contains code or data.
 *
 * @output status Status of the operation.
 * MM_SUCCESS
 * MM_ENON_EXISTING_REGION
 * MM_EOUT_OF_RESOURCES
 * MM_EILLEGAL_AREA
 * MM_EILLEGAL_DOMAIN
 * @output options Bit field containing options for the actual resident mapping.
 *
 */
struct mm_make_resident
{
      SIGSELECT   sig_no;
      mm_status_t status;       /* O: Status of the operation. */
      OSADDRESS   base;         /* I: Logical base address of memory area to map. */
      U32         size;         /* I: Size of memory area to map. */
      U32         options;      /* I: Bit field containing options. */
};

/**
 * @signal MM_MAKE_UNRESIDENT_REQUEST
 * @brief Request that the supplied logical area be made non-resident.
 *
 * @member mm_make_resident
 *
 * @long The request for making a memory area non-resident must match
 * a memory area previously made resident,
 * otherwise the status code returned will be MM_EILLEGAL_AREA.
 */
#define MM_MAKE_UNRESIDENT_REQUEST (36151) \
  /* !-SIGNO(struct mm_make_resident)-! */

/**
 * @signal MM_MAKE_UNRESIDENT_REPLY
 *
 * @brief Reply signal from Memory Manager replying to
 *        @link "mm.sig" "MM_MAKE_UNRESIDENT_REQUEST".
 *
 * @member mm_make_resident
 */
#define MM_MAKE_UNRESIDENT_REPLY (36152) \
  /* !-SIGNO(struct mm_make_resident)-! */

/**
 * @type mm_get_max_domains
 *
 * @field sig_no  Signal number.
 *
 * @output status Status of the operation.
 * MM_SUCCESS
 * @output max_domains The maximum number of domains in the system.
 *
 */
struct mm_get_max_domains
{
      SIGSELECT   sig_no;
      mm_status_t status;       /* O: Status of the operation. */
      U32         max_domains;  /* O: Maximum number of domains. */
};

/**
 * @signal MM_GET_MAX_DOMAINS_REQUEST
 * @brief Request the maximum number of domains in the system.
 *
 * @member mm_get_max_domains
 *
 * @long This signal has been added to make it possible in run-time to retrieve
 * the maximum numbers of domains. For example, the number can differ between
 * an SMP and unicore system.
 */
#define MM_GET_MAX_DOMAINS_REQUEST (36153) \
  /* !-SIGNO(struct mm_get_max_domains)-! */

/**
 * @signal MM_GET_MAX_DOMAINS_REPLY
 *
 * @brief Reply signal from Memory Manager replying to
 *        @link "mm.sig" "MM_GET_MAX_DOMAINS_REQUEST".
 *
 * @member mm_get_max_domains
 */
#define MM_GET_MAX_DOMAINS_REPLY (36154) \
  /* !-SIGNO(struct mm_get_max_domains)-! */


#ifdef __cplusplus
}
#endif

#endif /* #ifdef _MM_SIG */
