/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2011 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */
/**
 * @public
 * @toc Memory_Management:
 * @file ose_spi/mm_access.h
 *
 * @brief     Commands for exception-safe memory transactions.
 *
 * @long      These functions are exception safe, i.e. they intercept
 *            all exceptions caused by memory accesses.
 *
 */
#ifndef OSE_MM_ACCESS_H_
#define OSE_MM_ACCESS_H_

#ifdef __cplusplus
extern "C" {
#endif

#include <stddef.h>
#include "osetypes.h"
#include "mmtypes.h"


/**
 * @function mm_access_safecpy
 * @brief Exception-safe copy of data from one domain to another.
 *
 * @long  Copy <size> bytes from address <src_addr> in domain <src_domain> to
 *        address <dst_addr> in domain <dst_domain>.
 *
 *        This function is exception safe, i.e. it intercepts all exceptions
 *        caused by memory accesses.
 *
 * @input src_domain  The source domain of the copy operation.
 *
 * @input src_addr    The source address in the source domain.
 *
 * @input dst_domain  The destination domain of the copy operation.
 *
 * @input dst_addr    The destination address in the destination domain.
 *
 * @input size        The number of bytes to copy.
 *
 * @input freeze_mode Non-zero if called from freeze-mode context.
 *
 * @return            
 *   @tab 0    @tab Operation successfull. @break
 *   @tab != 0 @tab Operation not successfull. A sub-range of the destination
 *                  range may have been updated. @break
 *
 * @restriction Only available in supervisor mode processes.
 *   @break
 *              Only available after MMU has been enabled.
 *   @break
 *              Either src_domain or dst_domain must be system domain.
 *   @break
 *              If using SMP: destination memory area must not be used by
 *              another CPU.
 *
 * @seealso default_restrictions_calling_functions
 */
int
mm_access_safecpy(OSDOMAIN dst_domain, void *dst_addr,
                 OSDOMAIN src_domain, const void *src_addr,
                 size_t size, int freeze_mode);


/**
 * @function mm_access_forced_write
 *
 * @brief Exception-safe update of write-protected memory.
 *
 * @long <size> bytes are copied from <src> to <dst> effectively ignoring
 *       any write protection in the <dst> area.
 *
 *       This function is exception safe, i.e. it intercepts all exceptions
 *       caused by memory accesses.
 *
 *       If executing on a NOMMU target, the effect of mm_access_forced_write
 *       is equal to a mm_access_safecpy. When executing on a soft target,
 *       the effect is equal to a call to memcpy.
 *
 *       The cache subsystems are effectively synchronized (with regards to
 *       each other and main memory) after a successful operation.
 *
 * @input dst_domain  The domain of the update operation.
 *
 * @input src         Source address of the transaction.
 *
 * @input dst         Destination address of the transaction
 *
 * @input size        The number of bytes to copy.
 *
 * @input freeze_mode Non-zero if called from freeze-mode context.
 *
 * @return
 *   @tab 0    @tab Operation successfull. Caches are synchronized. @break
 *   @tab != 0 @tab Operation not successfull. A sub-range of the destination
 *                  range may have been updated. @break
 *
 * @restriction Only available in supervisor mode processes.
 *   @break
 *              Only available after MMU has been enabled.
 *   @break
 *              the area [<dst>, <dst> + <size> - 1] must not cross a page
 *              boundary.
 *   @break
 *              If using SMP: destination memory area must not be used by
 *              another CPU.
 *
 * @seealso default_restrictions_calling_functions
 */
int
mm_access_forced_write(OSDOMAIN dst_domain,
                       void *dst,
                       const void *src,
                       size_t size,
                       int freeze_mode);

/* The version available from CRT */
int
ose_mm_access_forced_write(OSDOMAIN dst_domain,
                           void *dst,
                           const void *src,
                           size_t size,
                           int freeze_mode);


#ifdef __cplusplus
}
#endif

#endif /* OSE_MM_ACCESS_H_ */
