/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2006 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */
/**
 * @public
 * @toc Memory_Management:
 * @file ose_spi/mm_util.h
 *
 * @brief MM function interface accessible from Core Monolith link scope.
 *
 *
 * @long
 * This file defines a functional interface for the most common
 * usage cases regarding block memory allocations from MM. It hides
 * the log_mem name for system ram in SAS/SASE configurations which
 * simplifies for clients like DDA, etc.
 */

#ifndef OSE_MM_UTIL_H_
#define OSE_MM_UTIL_H_

#include "ose.h"
#include "string.h"
#include "mm.sig"
#include "mmtypes.h"
#include <stdint.h>

#ifdef __cplusplus
extern "C" {
#endif


#define MM_EVENT_ANY_DOMAIN (MM_MAX_DOMAINS + 1)

#define RECLAIM_ALL_INTERNAL (0xffffffff)

/**
 * @function ose_mm_map_range
 *
 * @brief  Map one or a number of continuous 4k pages in the Demand Page Cache
 *         or for NOR.
 *
 * @long   An MM util function needed by Pager functionality in order to
 *         execute page faults in the context of the faulting client.
 *         It is only allowed to use this interface in its own domain.
 *           - Not available in SMP lib.
 *           - Not available for LM/CRT scope.
 *
 * @input domain     Specifies the domain.
 * @input la         Logical address of the first page.
 * @input pa         Physical address of the  first page.
 * @input size       Specifies the size.
 * @input permission Specifies MMU access permissions.
 *                   See create_region field permission in mm.sig for details.
 * @input cache_mode Specifies the cache_mode.
 *                   See create_region field cache_mode in mm.sig for details.
 * @input options    Reserved for future.
 * @return           MM_SUCCESS if OK, otherwise the MM error code.
 */
mm_status_t
ose_mm_map_range(OSDOMAIN    domain,
                 log_addr_t  la,
                 phys_addr_t pa,
                 size_t      size,
                 U32         permission,
                 U32         cache_mode,
                 int         options);



 /**
 * @function ose_mm_unmap_range
 *
 * @brief  Unmap the range [<la>, <la> + <size> - 1] from Demand Page Cache
 *         or NOR.
 *
 * @long   A MM util function needed by Pager functionality in order to
 *         execute page faults in the context of the faulting client.
 *           - Not available in SMP lib.
 *           - Not available for LM/CRT scope.
 *
 * @input domain     Specifies the domain.
 * @input la         Specifies the logical address.
 * @input size       Specifies the size.
 * @input options    Reserved for future.
 * @return           MM_SUCCESS if OK, otherwise the MM error code.
 */
mm_status_t
ose_mm_unmap_range(OSDOMAIN      domain,
                   log_addr_t    la,
                   size_t        size,
                   int           options);


/**
 * @function ose_mm_get_free_mem
 *
 * @brief    Get total amount free MM memory in the RAM memory type.
 *
 * @long     - Not available for LM/CRT scope.
 *
 * @return   The size of the available free memory.
 */
size_t
ose_mm_get_free_mem(void);


/**
 *
 * @type mm_event_no_t
 *
 * @brief Enum consisting of the possible MM events.
 *
 * @field MM_EVENT_REGION_DELETED The MM event generated when a region is
 *                                deleted.
 *
 * @field MM_EVENT_RECLAIM_REGION_PHYSICAL The MM event asking a component to
 *                                         reclaim one region of physical
 *                                         memory.
 */
typedef enum mm_event_no_t
{
   MM_EVENT_REGION_DELETED = 0,
   MM_EVENT_RECLAIM_REGION_PHYSICAL = 1
} mm_event_no_t;
#define MM_NO_OF_EVENTS    2

/**
 *
 * @type ose_component_identifier_t
 *
 * @brief A unique identifier for each single component that participates
 *        in the MM event system.
 *
 * @field UNKNOWN_COMPONENT This is an unknown component, or the component
 *                          could not be identified.
 *
 * @field COMMON_HEAP_COMPONENT Identifier for all common (non-tentative)
 *                              heaps.
 *
 * @field TENTATIVE_HEAP_COMPONENT Identifier for all tentative heaps.
 *
 * @field COMMON_MM_COMPONENT Identifier for all common (non-tentative)
 *                            MM components.
 *
 * @field TENTATIVE_MM_COMPONENT Identifier for all tentative MM regions.
 *
 * @field PAGER_COMPONENT Identifier for the pager and its page cache.
 *
 * @field USER_DEFINED_COMPONENT_START User defined component identifiers
 *                                     start here.
 */
typedef enum ose_component_identifier_t
{
   UNKNOWN_COMPONENT = 0,
   COMMON_HEAP_COMPONENT = 1,
   TENTATIVE_HEAP_COMPONENT = 2,
   COMMON_MM_COMPONENT = 3,
   TENTATIVE_MM_COMPONENT = 4,
   PAGER_COMPONENT = 5,
   USER_DEFINED_COMPONENT_START = 0x1000
} ose_component_identifier_t;


/* Forward declaration. */
union mm_event_t;

/**
 * @type  mm_event_callback_t
 *
 * @brief A user supplied event handler.
 *
 * @long  This function is called whenever an event occurs that this handler
 *        has been registered for.
 *
 * @input event    The event that has been triggered and any input data
 *                 that goes with it.
 * @output event   The data fields in the event might be modified by the event
 *                 handler. See each specific event what fields in the union
 *                 are inputs and outputs.
 */
typedef void(mm_event_callback_t)(union mm_event_t *event);

/**
 * @type  mm_event_policy_t
 *
 * @brief A user supplied MM event policy.
 *
 * @long  This function is called first whenever an event occurs that
 *        this policy has been registered for. This policy shall then
 *        invoke the registered handlers in the order that the policy
 *        implements.
 *
 * @input event    The event that has been triggered and any input data
 *                 that goes with it.
 * @output event   The data fields in the event might be modified by the policy
 *                 handler or one of the event handlers it calls.
 *                 See each specific event what fields in the union
 *                 are inputs and outputs.
 */
typedef void(mm_event_policy_t)(union mm_event_t *event);

/**
 *
 * @type mm_reclaim_physical_event_t
 *
 * @brief A structure describing the reclaim event.
 *
 * @long  This event is called when the MM needs to reclaim some memory
 *        from one or more of the components that have registered for
 *        this event.
 *
 * @input event_no      The event number.
 * @input event_handler The callback function.
 * @input identifier    The component identifier of the callback function.
 * @input internal_identifier A component specific identifier. Set this to
 *                            RECLAIM_ALL_INTERNAL for the component to reclaim
 *                            all of its internal components. For the heap
 *                            this is the heapref of the heap that the memory
 *                            should be reclaimed from. For the page cache,
 *                            this is the page cache number.
 * @input inquiry  Set if the component should only report how many bytes
 *                 it could potentially free and not actually prepare to free
 *                 it.
 * @input phys_mem The physical memory type to be reclaimed.
 *                 At registration, this is the physical memory type
 *                 that this callback will be called for.
 *                 Also need to filled in for ose_mm_get_event_handler and
 *                 ose_mm_get_next_event_handler for proper identification.
 * @output domain  The domain the region can be deleted from.
 * @output base    The start address of the region that can be deleted.
 * @output size    The size of the region that can be deleted, or zero
 *                 if no region can be deleted.
 */
struct mm_reclaim_physical_event_t
{
   mm_event_no_t              event_no;
   mm_event_callback_t       *event_handler;
   ose_component_identifier_t identifier;
   uint32_t                   internal_identifier;
   OSDOMAIN                   domain;
   log_addr_t                 base;
   size_t                     size;
   const char                *phys_mem;
   uint8_t                    inquiry;
};

/**
 * @type  mm_region_deleted_event_t
 *
 * @brief Callback for the MM region deleted event.
 *
 * @long  This function is called when a region is deleted by the MM.
 *
 * @input event_no      The event number.
 * @input event_handler The callback function.
 * @input identifier    The component identifier of the callback function.
 * @input domain        The domain the region was deleted from.
 * @input base          The start address of the region that was deleted.
 */
struct mm_region_deleted_event_t
{
   mm_event_no_t              event_no;
   mm_event_callback_t       *event_handler;
   ose_component_identifier_t identifier;
   OSDOMAIN                   domain;
   log_addr_t                 base;
};

/**
 * @type mm_event_t
 *
 * @declaration union mm_event_t;
 *
 * @brief The MM event union.
 *
 * @long This is the union of all mm_events. Used by the interface to describe
 *       any one event.
 *
 * @field reclaim For accessing a mm_reclaim_physical_event_t structure.
 *
 * @field delete  For accessing a region_deleted_event structure.
 *
*/
union mm_event_t
{
   struct mm_reclaim_physical_event_t reclaim;
   struct mm_region_deleted_event_t    delete;
};

/**
 * @function ose_mm_install_handler
 *
 * @restriction NOT YET AVAILABLE in standard OSE.
 *
 * @brief  Install an MM event handler.
 *
 * @long   For all events it is necessary to fill in event_no, event_handler and
 *         identifier in the mm_event_t union. Fill in the event that
 *         should be registered in event_no, the handler in
 *         event_handler and the identity of their component in the identity
 *         field.
 *         See each event for event specific fields that need to be filled in
 *         at install time.
 *
 * @input event        A pointer to the event to register for.
 * @input beginning    Insert the event in the beginning of the event queue,
 *                     instead of the end, which is the default.
 * @return             MM_SUCCESS if OK, otherwise the MM error code.
 */
mm_status_t
ose_mm_install_handler(union mm_event_t *event,
                       OSBOOLEAN beginning);

/**
 * @function ose_mm_install_policy
 *
 * @restriction NOT YET AVAILABLE in standard OSE.
 *
 * @brief  Install a policy for an MM event.
 *
 * @long   The policy installed should pick what registered components to
 *         call for this event and call them. Once it has gotten a satisfactory
 *         result back (for example, a callback has indicated a region to
 *         reclaim), it should return.
 *         If no policy is installed for an event, FIFO is used.
 *
 * @input event_no        The ID of the event this policy should be called for.
 * @input policy_handler  The policy function called every time an event is
 *                        triggered.
 * @return                MM_SUCCESS if OK, otherwise the MM error code.
 */
mm_status_t
ose_mm_install_policy(mm_event_no_t event_no,
                      mm_event_policy_t *policy_handler);

/**
 * @function ose_mm_uninstall_handler
 *
 * @restriction NOT YET AVAILABLE in standard OSE.
 *
 * @brief  Uninstall an MM event handler.
 *
 * @long   The MM event is identified by the supplied mm_event_t union.
 *         See each event for event specific fields that need to be filled in
 *         for successful identification.
 *
 * @input event   The event this handler should be removed from.
 *                The handler is identified by event_no, event_handler
 *                and identifier plus any event specific fields.
 * @return        MM_SUCCESS if OK, otherwise the MM error code.
 */
mm_status_t
ose_mm_uninstall_handler(union mm_event_t *event);

/**
 * @function ose_mm_uninstall_policy
 *
 * @restriction NOT YET AVAILABLE in standard OSE.
 *
 * @brief  Uninstall a policy for an MM event.
 *
 * @long   Will revert the policy to the default FIFO.
 *
 * @input event_no        The ID of the event the current policy should be
 *                        uninstalled for.
 * @return                MM_SUCCESS if OK, otherwise the MM error code.
 */
mm_status_t
ose_mm_uninstall_policy(mm_event_no_t event_no);

/**
 * @function ose_mm_get_event_handler
 *
 * @restriction NOT YET AVAILABLE in standard OSE.
 *
 * @brief  Gets the event handler identified by the supplied identifier and
 *         event_no.
 *
 * @long   To be used in an MM event policy handler. If the handler is not
 *         registered NULL is returned. Note that certain events might
 *         require that event specific fields in the handler are also filled
 *         in for successful identification. See the specific event for
 *         details.
 *
 * @input event_no     The ID of the event.
 * @input identifier   The identifier of the registered callback that is
 *                     to be retrieved.
 * @input event        Certain events might require that event specific fields
 *                     are also filled in for successful identification.
 * @return             The handler corresponding to the supplied name, or
 *                     NULL if it does not exist.
 */
mm_event_callback_t *
ose_mm_get_event_handler(mm_event_no_t event_no,
                         ose_component_identifier_t identifier,
                         union mm_event_t *event);

/**
 * @function ose_mm_get_next_event_handler
 *
 * @restriction NOT YET AVAILABLE in standard OSE.
 *
 * @brief  Gets the next event handler of the supplied event.
 *
 * @long   To be used in an MM event policy handler. If there are no more
 *         handlers, NULL is returned.
 *
 * @input event_no     The ID of the event.
 * @input handler      The current event handler, or NULL if getting
 *                     the first registered handler.
 * @return             The next handler, or NULL if it does not exist.
 */
mm_event_callback_t *
ose_mm_get_next_event_handler(mm_event_no_t event_no,
                              mm_event_callback_t *handler);

/**
 * @function ose_mm_get_freeable_memory
 *
 * @restriction NOT YET AVAILABLE in standard OSE.
 *
 * @brief    Get total amount free MM memory that is available to a
 *           specified component in the supplied physical memory type.
 *
 * @long     - Not available for LM/CRT scope.
 *
 * @input phys_mem     The name of the physical memory type.
 * @input size         The max amount of memory the caller is interested
 *                     in. Whenever the reclaim framework finds this much
 *                     free memory, it returns without asking any more
 *                     components, thus improving performance.
 * @input identifier   The identifier of the registered callback
 *                     to retrieve.
 *
 * @return             The size of the available free memory.
 */
size_t
ose_mm_get_freeable_memory(const char *phys_mem,
                           size_t size,
                           ose_component_identifier_t identifier);


/**
 * @function ose_mm_multithread
 *
 * @brief  Enables or disables multithreading support in MM.
 *
 * @long   Enables or disables multithreading support in MM. Multithreading
 *         support must be enabled before any use of the mm_util.h interface.
 *         The function may be called several times by different clients.
 *
 * @input enable True (non-zero) to enable multithreading support.
 * @return       MM_SUCCESS if OK, otherwise the MM error code.
 */
mm_status_t
ose_mm_multithread(mm_bool_t enable);


#ifdef __cplusplus
}
#endif

#endif /* #ifndef OSE_MM_UTIL_H_ */
