/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2006 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */
/**
 * @public
 *
 * @toc Memory_Management:
 *
 * @file ose_spi/mmtypes.h
 *
 * @brief Types used by OSE Memory Management (MM).
 *
 */

/* NOTE: this define now sets a version of mmtypes.h to force mm.sig to
         include the correct types. */
#ifndef _MMTYPES_H
#define _MMTYPES_H 500

#include "osetypes.h"

#ifdef __cplusplus
extern "C" {
#endif

/**
 * @macro TRUE
 *
 * @brief Boolean TRUE.
 */
#ifdef TRUE
#undef TRUE
#endif
#define TRUE              1

/**
 * @macro FALSE
 *
 * @brief Boolean FALSE.
 */
#ifdef FALSE
#undef FALSE
#endif
#define FALSE             0

/**
 * @macro MM_NAME_LENGTH
 *
 * @brief The default string length, excluding null byte, for all strings
 *        handled by MM.
 * @long Currently this value is 31.
 */
#define MM_NAME_LENGTH  31


/**
 * @type phys_addr_t
 *
 * @brief Physical memory address.
 *
 */
typedef unsigned long long phys_addr_t;

/**
 * @type phys_size_t
 *
 * @brief Physical memory size.
 *
 */
typedef phys_addr_t phys_size_t;

/**
 * @type log_size_t
 *
 * @brief Logical memory size.
 *
 */
typedef unsigned long log_size_t;

/**
 * @type log_addr_t
 *
 * @brief Logical memory address.
 *
 */
typedef unsigned long log_addr_t;

/**
 * @type mm_size_t
 *
 * @brief MM specific type for size.
 *
 */
typedef unsigned int mm_size_t;

/**
 * @type mm_bool_t
 *
 * @brief MM specific type for Boolean.
 *
 */
typedef int mm_bool_t;

/**
 * @type phys_resource
 *
 * @brief Base and size of physical memory resource.
 *
 * @field base The base of a physical memory resource.
 * @field size The size of a physical memory resource.
 *
 */
struct phys_resource
{
      phys_addr_t base;
      U32         size;
};

/**
 * @type mm_tag
 *
 * @brief Tag identity and tag value.
 *
 * @field tag_id    Question. Values from enum
 *                  @link "mm.sig" "MmTagId".
 * @field tag_value Answer. Assigned by MM.
 */
struct mm_tag
{
      U32 tag_id;      /* Question. Values from enum MmTagId. */
      U32 tag_value;   /* Answer. Assigned by MM. */
};

/**
 * @macro    MM_TAG_SW_COUNTER
 * @brief Tag Option value used as Output By MM when specifying that a
 *        statistic counter is implemented in SW.
 * @long  A SW counter can either be local to a specific CPU or global in a
 *        SMP system.
 */
#define MM_TAG_SW_COUNTER  (0x1)

/**
 * @macro    MM_TAG_HW_COUNTER
 * @brief Tag Option value used as Output By MM when specifying that a
 *        statistic counter is using HW resources.
 * @long  Typically a HW counter is local in a SMP system. I.e. it will
 *        only be updated by the CPUs where the counter is located.
 *        A counter of type MM_TAG_HW_COUNTER is enabled
 *        only after the first time that MM receives the
 *        signal MM_GET_MMU_STATISTICS_REQUEST.
 */
#define MM_TAG_HW_COUNTER  (0x2)

/**
 * @macro    MM_TAG_COUNTER_ENABLED
 * @brief Tag Option value used as Input when a client specifies that a
 *        statistic counter should be enabled.
 * @long  A counter of type MM_TAG_HW_COUNTER is default disabled and
 *        will remain so until MM_TAG_COUNTER_ENABLED is received.
 */
#define MM_TAG_COUNTER_ENABLED  (0x4)

/**
 * @macro    MM_TAG_COUNTER_OVERFLOW
 * @brief Tag Option value used as Output By MM when specifying that a
 *        statistic counter has reached its maximum value without causing
 *        overflow.
 *
 * @long  All HW counters will be stopped/freeze.
 */
#define MM_TAG_COUNTER_OVERFLOW  (0x8)

/**
 * @macro    MM_TAG_COUNTER_FROZEN
 * @brief Tag Option value used as Output By MM when specifying that a
 *        statistic counter has been frozen. 
 *
 * @long  Usually caused by a HW counter that was about to overflow.
 *        When this happens all other HW counters will be frozen,
 *        in order to save the total counter state.
 *        The HW counters will remain frozen until a signal with
 *        option MM_TAG_RESET_COUNTER is received.
 */
#define MM_TAG_COUNTER_FROZEN (0x10)

/**
 * @macro    MM_TAG_RESET_COUNTER
 * @brief Tag Option value used as Input when a client specifies that a
 *        statistic counter should be reset.
 * @long  The counter will be read before reset and the value will be returned
 *        to the client. Note that reset only has effect on enabled counters.
 *        An enabled counter that is reset will still be enabled after the
 *        reset.
 */
#define MM_TAG_RESET_COUNTER  (0x20)

/**
 * @macro    MM_TAG_GLOBAL_COUNTER
 * @brief Tag Option value used as Output By MM when specifying that a
 *        statistic counter is global.
 * @long  In an SMP system this of interest since global counters
 *        are affected by the execution/accesses of all of the CPUs.
 */
#define MM_TAG_GLOBAL_COUNTER  (0x40)

/**
 * @macro    MM_TAG_COUNTER_HANDLER_OV_SUPPORT
 * @brief    This counter is implemented in an exception handler
 *           that supports overflow detection.
 * @long  
 */
#define MM_TAG_COUNTER_HANDLER_OV_SUPPORT  (0x80)

/**
 * @macro    MM_TAG_COUNTER_64
 * @brief    This counter uses 64 bits before overflow.
 * @long  
 */
#define   MM_TAG_COUNTER_64 (0x100)

/**
 * @macro    MM_TAG_PER_CORE_COUNTER
 * @brief    This counter is local to and dedicated to one core.
 * @long  
 */
#define   MM_TAG_PER_CORE_COUNTER (0x200)

/**
 * @macro     MM_TAG_COUNTER_DISABLED
 * @brief Tag Option value used as Input when a client specifies that a
 *        statistic counter should be disabled.
 * @long  Any resources used by counter the will be freed.
 *        A disabled counter can later be enabled with MM_TAG_COUNTER_ENABLED.
 *        Normally only counters of type MM_TAG_HW_COUNTER can be disabled.
 */
#define MM_TAG_COUNTER_DISABLED  (0x400)

/*
 * MM_TAG Option bits > 0x800 reserved for OSE internal used.
 */  

/**
 * @type mm_tag_ext
 *
 * @brief Extended Tag format includes tag identity tag value and
 *        tag string description.
 *
 * @field tag_id    Question. Values from enum
 *                  @link "mm.sig" "MmTagId".
 *
 *
 * @field tag_value Answer. Assigned by MM.
 *
 * @field tag_extra Answer. Assigned by MM. Currently not used.
 *
 * @field tag_options Bit field to use when reading statistics.
 *               Options are activated by OR'ing
 *               requested tag option numbers.@break
 *               If no options are to be used, this parameter should be zero.
 *               @break
 *               Available tag options: @break
 *               @link "mmtypes.h" "MM_TAG_RESET_COUNTER"
 *
 * @field cpu_id    Specifies to which CPU a HW or SW CPU local counter
 *                  belong in a SMP system. In a non-SMP system cpu_id
 *                  should be zero.
 *                  Currently  local counters in an SMP system is
 *                  only supported for CPU 0.
 *
 * @field tag_desc  Tag description in string format. Answer assigned by MM.
 */
/* Note the two first fields in struct mm_tag_ext is and should remain
 * source compatible with struct mm_tag. */
struct mm_tag_ext
{
   U32       tag_id;                  /* Question Values from enum MmTagId. */
   unsigned long long tag_value;      /* Answer. Assigned by MM. */
   U32       tag_extra;               /* Answer. Assigned by MM. */
   U32       tag_options;             /* Question/Answer. */
   U32       cpu_id;                  /* Question/Answer. */
   char      tag_desc[MM_NAME_LENGTH];/* Answer: Tag description given by MM.*/

};

/**
 * @type mm_phys_mem_info
 *
 * @brief Physical memory type information tag-value pair.
 *
 * @field tag   Question. Information tag from enum
 *              @link "mm.sig" "MmTagId".
 * @field value Answer. Value assigned by MM.
 */
struct mm_phys_mem_info
{
      U32         tag;   /* Question. Values from enum MmTagId. */
      phys_size_t value; /* Answer. Assigned by MM. */
};

/**
 * @type OSDOMAIN
 *
 * @brief Type used for a domain identifier.
 *
 */
typedef U32 OSDOMAIN;

/**
 * @type mm_status_t
 *
 * @brief Type used to show status in MM reply signal.
 *
 * @long  Takes values of enum
 *        @link "mm.sig" "MmStatus".
 *        Note that enum should be avoided in signals.
 *
 */
typedef U32 mm_status_t;


/**
 * @macro MM_MAX_DOMAINS
 *
 * @brief Maximum number of domains that can be created including the kernel
 *        domain which is always created.
 *
 * @deprecated This symbol is deprecated from OSE 5.6 and will not be
 *             supported in future releases of OSE. The symbol will remain
 *             but might not indicate the right maximum number of domains
 *             for the system.
 *             Please use the @link "mm.sig" "MM_GET_MAX_DOMAINS_REQUEST"
 *             signal to retrive the maximum number of domains.
 */
#define MM_MAX_DOMAINS  255

/**
 * @macro MM_DOMAIN_RESERVED
 *
 * @brief A domain identity that is guaranteed to never be used by MM.
 */
#define MM_DOMAIN_RESERVED ((OSDOMAIN) ~0u)

/**
 * @macro MM_KERNEL_DOMAIN
 *
 * @brief The kernel domain. Always exists.
 */
#define MM_KERNEL_DOMAIN  ((OSDOMAIN) 0u)

#ifdef __cplusplus
}
#endif

#endif /* _MMTYPES_H */
