/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2009 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

/**
 * @toc OSE:
 * @file ose_spi/monitor.sig
 *
 * @brief OSE System information signal interface
 *        to access OSE run-time monitors.
 *
 * @long The monitor.sig header file defines the main signal interface between
 * debuggers, system supervisors, et. al., and OSE monitors (e.g. FMM and RMM).
 * It provides extensive information for most common run-time OSE entities.
 * The standard process name for the RMM is "ose_monitor", used by in the
 * @link "" "MonitorConnectRequest" request to the RMM and in a few other requests,
 * otherwise a client must use the process ID returned from the connect request
 * to communicate with RMM.
 *
 * @long Additional monitor*.sig interface files contain:
 *
 * @tab monitor_actions.sig
 * @tab This interface provides mechanisms through which the user can interact
 * directly with OSE system entities. Typically to start/stop/kill processes,
 * send signals, set system parameters or similar operations.
 * @break
 *
 * @tab monitor_cpu.sig
 * @tab CPU load profiling and reports from user application profiling are
 * controlled via this interface.
 * @break
 *
 * @tab monitor_dumps.sig
 * @tab Debuggers can via this interface fetch dumps, whose dump components,
 * e.g. the FMD freeze mode dumper, have dumped in the RAMdump area.
 * @break
 *
 * @tab monitor_events.sig
 * @tab The event system in the monitor keeps track of events. The monitor can
 * keep track of each client's event state and automatically perform
 * predefined or user-defined actions when events (predefined events
 * or user events) occurs. Notification reports can be sent to clients.
 * @break
 *
 * @tab monitor_execution.sig
 * @tab This is the interface, used by source code debuggers to control and
 * inspect the OSE system, connected via the monitor. Typical operations
 * are to get/set memory or CPU registers, to control breakpoints
 * and intercept, step or resume processes.
 * It also contains a basic event system, limited to the needs of a
 * source code debugger.
 * @break
 *
 * @long Some request signals may yield multiple reply signals. In this case the
 * monitor finishes the replies by sending a specific endmark signal.
 * The endmark signal also contains the overall status of the request.
 * The monitor will send some replies in multiple reply signals
 * to the client using flow control. With flow control, the monitor sends
 * flow control requests, which the client must respond to.
 *
 * @long Unless otherwise noted in the description of the requests, signals are sent
 * without flow control.
 *
 * @long Signals from the monitor to the client are either:
 * @tab - Sent without flow control.
 * @tab No flow control is used by the monitor for the client
 * to respond to.
 * @break
 * @tab - Sent under blocking flow control.
 * @tab In this case the client should expect and respond to
 * @link "" "MonitorBlockingFlowControlRequest". If the client does not
 * respond in time, the monitor will hold further signals and wait for the
 * response. The configuration sets the maximum number of signals sent to the
 * client after a flow control request before blocking, if the client
 * does not respond.
 * @break
 * @tab - Sent under non-blocking flow control.
 * @tab In this case the client should expect and respond to
 * @link "" "MonitorNonBlockingFlowControlRequest". If the client does not
 * respond in time, the monitor will discard further signals of this kind
 * until it receives the response.
 * @break
 *
 * @long For compatibility reasons, clients must be prepared to accept unknown
 * values in (for instance) enumeration fields in reply signals from the monitor.
 * The reason for this design rule is to allow the concepts represented in this
 * interface to evolve without breaking existing clients.
 *
 * @long Event states are used to allow chained actions, e.g. actions depending on
 * a series of events.
 *
 * @long Clients always need to Connect and in many cases also Attach to serialize
 * requests and actions for multiple clients in the monitor and to limit the
 * operations to a scope of processes.
 * Attachment is only needed when using the event action system (monitor_events.sig)
 * or debugging via the execution control system (monitor_execution.sig).
 * By attaching, the client reserves a scope of processes for which the client
 * is interested.
 *
 * @long The sequence used by a client to communicate with the run-time
 * monitor is:
 *
 * @long 1. The client sends the signal @link "" "MonitorInterfaceRequest" to
 * the monitor ("ose_monitor") to get information about the monitor interface.
 * Especially needed is the monitor's byte order, which is used in all other
 * signals.
 *
 * @long 2. Before connecting, the only other signal that can be used is
 * @link "" "MonitorNameRequest".
 *
 * @long 3. The client connects with the signal
 * @link "" "MonitorConnectRequest" to the monitor,
 * and receives more information about supported features and other details.
 * The client should save the sender ID of @link "" "MonitorConnectReply"
 * and use that PID to send all other signals to the monitor.
 *
 * @long 4. Most information requests can be sent to the monitor in
 * the connected state.
 *
 * @long 5. The client must also attach with a main scope, before using
 * the event action system or debugging features, which are based
 * on different scopes. Also a subscope for intercepts are defined
 * when attaching. The client attaches to the monitor with the
 * @link "monitor_execution.sig" "MonitorAttachRequest", supplying a
 * main scope and an intercept sub-scope.
 *
 * @long 6. When ready, detach the scope
 * @link "monitor_execution.sig" "MonitorDetachRequest" (if attached)
 * and disconnect
 * @link "" "MonitorDisconnectRequest", allowing the monitor to
 * deallocate resources linked to the client.
 *
 * @long System parameters with names beginning with krn/monitor/ can be used
 * to configure non-default values for certain monitor parameters.
 * Examples are configuration of various limits for the resources used by the
 * monitor, initialization of event state and other values or even totally
 * disable the monitor at start to allow later system calls to enable only
 * selected functions.
 * See RMM in the "OSE Core User's Guide".
 *
 * @long Configuration of the kernel for using the
 * monitor are described in the "OSE Core User's Guide" for the OSE Executive.
 * A few examples are: DEBUG(YES) and linkage with the debug libraries.
 * krn/tracesize is the trace buffer size (shared by all RMM clients).
 * POOL_SIZE is the area containing both trace buffer and system pool, so
 * this might need to be increased if a large trace buffer is configured.
 * krn/ldm_prio is the ose_monitor priority (default 1).
 * krn/system_process is a list of processes considered system processes
 * (which should not be intercepted when debugging).
 * Configuration of hardware breakpoint support, if the hardware supports it.
 *
 * @long In addition to these monitor*.sig signal files, additional functions
 * for control of the OSE debug monitor are found in other header files,
 * e.g. in rmm.h and ose.h.
 *
 * @long The rmm.h contains a few functions, complementing the signal
 * interface described in these monitor*.sig files. One example is
 * registering a user action callback with
 * @link "rmm.h" "ose_rmm_set_user_action_handler".
 *
 * @long Examples of functions for debugging in ose.h are
 * @link "ose.h" "ose_create_report" and
 * @link "ose.h" "ose_set_report_val" for user profiling
 * (via @link "monitor_cpu.h" "MonitorSetUserReportsEnabledRequest")
 * or the functions @link "ose.h" "create_event" and
 * @link "ose.h" "signal_event" to produce user events.
 *
 * @long Note that the three features user-action, user-event, and user-reports
 * are three completely separate things.
 *
 * @long See also Run-Mode Monitor (RMM) in the "OSE Core User's Guide".
 *
 * @seealso ose_spi/monitor_actions.sig
 * @seealso ose_spi/monitor_cpu.sig
 * @seealso ose_spi/monitor_events.sig
 * @seealso ose_spi/monitor_dumps.sig
 * @seealso ose_spi/monitor_events.sig
 * @seealso ose_spi/monitor_execution.sig
 * @seealso ose_rmm_set_req_enabled
 * @seealso ose_rmm_set_user_report_handler
 * @seealso ose_rmm_set_user_action_handler
 * @seealso ose_create_report
 * @seealso ose_set_report_val
 * @seealso create_event
 * @seealso signal_event
 * @seealso ose_spi/rmm.h
 * @seealso ose.h
 * @seealso ose_epm_init_hwbp
 */

#ifndef _MONITOR_SIG
#define _MONITOR_SIG

#include "stdint.h"

#ifdef __cplusplus
extern "C" {
#endif

/*******************************************************************************
 *
 * DEFINITIONS
 *
 ******************************************************************************/

/**
 * @type MonitorBoolean
 * @brief Boolean type - this is a 32bit value 1 or 0.
 * @long The byte order is according to the monitor, as seen in
 * @link "" "MonitorInterfaceReply".
 * @seealso MonitorBooleanValue
 */
typedef uint32_t MonitorBoolean;

/**
 * @type  MonitorBooleanValue
 * @brief Enumeration of the boolean values.
 * @long  In the monitor interface, the @link "" "MonitorBoolean" type
 * is actually a 32-bit quantity with 0 or 1 in bit 0,
 thus depending on the byte order.
 * @field MONITOR_BOOLEAN_FALSE False value (0).
 * @field MONITOR_BOOLEAN_TRUE  True value (1).
 */
enum MonitorBooleanValue
{
   MONITOR_BOOLEAN_FALSE = 0,
   MONITOR_BOOLEAN_TRUE  = 1
};

/**
 * @type MonitorStatus
 * @brief Monitor status type with OK or error codes in reply signals from
 * the monitor.
 * @seealso MonitorStatusValue
 */
typedef uint32_t MonitorStatus;

/**
 * @type MonitorStatusValue
 * @brief Enumeration of status values in reply signals.
 * @long Status with OK or error code in replies from monitor requests.
 *
 * @field MONITOR_STATUS_OK                    Status OK.
 * @field MONITOR_STATUS_ID_INVALID            Invalid ID.
 * @field MONITOR_STATUS_SCOPE_STATE_INVALID   Invalid  state for operation.
 * @field MONITOR_STATUS_SCOPE_INVALID         Invalid scope.
 * @field MONITOR_STATUS_SCOPE_OVERLAP         Scope overlap.
 * @field MONITOR_STATUS_SCOPE_EMPTY           Empty scope.
 * @field MONITOR_STATUS_REQUEST_INVALID       Invalid request.
 * @field MONITOR_STATUS_EPM_ERROR             Error returned by the execution
 *                                             point module (EPM).
 * @field MONITOR_STATUS_KERNEL_ERROR          The kernel memory is corrupt or
 *                                             in an unexpected state.
 * @field MONITOR_STATUS_MEMORY_GET_ERROR      Error reading memory.
 * @field MONITOR_STATUS_MEMORY_SET_ERROR      Error writing memory.
 * @field MONITOR_STATUS_MEMORY_EXHAUSTED      Memory allocation failed due to
 *                                             requested buffer too large to
 *                                             fit in pool.
 * @field MONITOR_STATUS_INTERNAL_ERROR        Monitor internal error.
 * @field MONITOR_STATUS_CANCELLED             Operation canceled.
 * @field MONITOR_STATUS_IN_USE                Requested resource is in use.
 * @field MONITOR_STATUS_SYSTEM_PARAM_ERROR    A system parameter operation
 *                                             failed.
 * @field MONITOR_STATUS_RESOURCE_EXHAUSTED    A resource has been exhausted.
 *                                             E.g., too many breakpoints in
 *                                             paged memory.
 * @field MONITOR_STATUS_CPU_INACTIVE          CPU runs other OS than OSE.
 * @field MONITOR_STATUS_MEMORY_LIMIT          Memory allocation failed due to
 *                                             total memory used for profiling
 *                                             is larger than the set limit.
 * @field MONITOR_STATUS_CLIENT_KILLED         Request fails due to a client
 *                                             no longer exist.
 * @field MONITOR_STATUS_BIND_FAILED           Request failed due to that the
 *                                             ose_bind_process system call
 *                                             failed.
 * @field MONITOR_STATUS_KILLED                The referred segment, block, or
 *                                             process was killed.
 * @field MONITOR_STATUS_INVALID_EU            An invalid execution unit id was
 *                                             given.
 */
enum MonitorStatusValue
{
   MONITOR_STATUS_OK                     = 0,
   MONITOR_STATUS_ID_INVALID,            /*  1 */
   MONITOR_STATUS_SCOPE_STATE_INVALID,   /*  2 */
   MONITOR_STATUS_SCOPE_INVALID,         /*  3 */
   MONITOR_STATUS_REQUEST_INVALID,       /*  4 */
   MONITOR_STATUS_EPM_ERROR,             /*  5 */
   MONITOR_STATUS_KERNEL_ERROR,          /*  6 */
   MONITOR_STATUS_MEMORY_GET_ERROR,      /*  7 */
   MONITOR_STATUS_MEMORY_SET_ERROR,      /*  8 */
   MONITOR_STATUS_SCOPE_OVERLAP,         /*  9 */
   MONITOR_STATUS_SCOPE_EMPTY,           /* 10 */
   MONITOR_STATUS_MEMORY_EXHAUSTED,      /* 11 */
   MONITOR_STATUS_INTERNAL_ERROR,        /* 12 */
   MONITOR_STATUS_CANCELLED,             /* 13 */
   MONITOR_STATUS_IN_USE,                /* 14 */
   MONITOR_STATUS_SYSTEM_PARAM_ERROR,    /* 15 */
   MONITOR_STATUS_RESOURCE_EXHAUSTED,    /* 16 */  
   MONITOR_STATUS_CPU_INACTIVE,          /* 17 */
   MONITOR_STATUS_MEMORY_LIMIT,          /* 18 */
   MONITOR_STATUS_CLIENT_KILLED,         /* 19 */
   MONITOR_STATUS_BIND_FAILED,           /* 20 */
   MONITOR_STATUS_KILLED,                /* 21 */
   MONITOR_STATUS_INVALID_EU             /* 22 */
};

/**
 * @type MonitorProcessType
 * @brief Process type.
 * @seealso MonitorProcessTypeValue
 */
typedef uint32_t MonitorProcessType;

/**
 * @type MonitorProcessTypeValue
 * @brief Enumeration of process types.
 * @field MONITOR_PROCESS_TYPE_PRIORITIZED      Prioritized process.
 * @field MONITOR_PROCESS_TYPE_BACKGROUND       Background process.
 * @field MONITOR_PROCESS_TYPE_INTERRUPT        Interrupt process.
 * @field MONITOR_PROCESS_TYPE_TIMER_INTERRUPT  Timer interrupt process.
 * @field MONITOR_PROCESS_TYPE_PHANTOM          Phantom process.
 * @field MONITOR_PROCESS_TYPE_SIGNAL_PORT      NOT USED.
 * @field MONITOR_PROCESS_TYPE_KILLED           Killed process.
 * @field MONITOR_PROCESS_TYPE_UNKNOWN          Unknown process.
 * @field MONITOR_PROCESS_TYPE_IDLE             Idle process (OSEck).
 */
enum MonitorProcessTypeValue
{
   MONITOR_PROCESS_TYPE_PRIORITIZED      =    0x0,
   MONITOR_PROCESS_TYPE_BACKGROUND       =   0x40,
   MONITOR_PROCESS_TYPE_INTERRUPT        =   0x80,
   MONITOR_PROCESS_TYPE_TIMER_INTERRUPT  =  0x100,
   MONITOR_PROCESS_TYPE_PHANTOM          =  0x200,
   MONITOR_PROCESS_TYPE_SIGNAL_PORT      =  0x400,
   MONITOR_PROCESS_TYPE_KILLED           =  0x800,
   MONITOR_PROCESS_TYPE_UNKNOWN          = 0x1000,
   MONITOR_PROCESS_TYPE_IDLE             = 0x10000
};

/**
 * @type MonitorProcessState
 * @brief Process state type.
 * @seealso MonitorProcessStateValue
 */
typedef uint32_t MonitorProcessState;

/**
 * @type MonitorProcessStateValue
 * @brief Enumeration of process states.
 * @field MONITOR_PROCESS_STATE_RECEIVE         Waiting in receive.
 * @field MONITOR_PROCESS_STATE_DELAY           Waiting in delay.
 * @field MONITOR_PROCESS_STATE_SEMAPHORE       Waiting on semaphore.
 * @field MONITOR_PROCESS_STATE_FAST_SEMAPHORE  Waiting on fast semaphore.
 * @field MONITOR_PROCESS_STATE_REMOTE          Waiting on remote call.
 * @field MONITOR_PROCESS_STATE_STOPPED         Stopped process.
 * @field MONITOR_PROCESS_STATE_INTERCEPTED     Intercepted process.
 * @field MONITOR_PROCESS_STATE_MUTEX           NOT USED.
 * @field MONITOR_PROCESS_STATE_RUNNING         Running process.
 * @field MONITOR_PROCESS_STATE_READY           Ready process.
 * @field MONITOR_PROCESS_STATE_KILLED          Killed process.
 */
enum MonitorProcessStateValue
{
   MONITOR_PROCESS_STATE_RECEIVE         =    0x1,
   MONITOR_PROCESS_STATE_DELAY           =    0x2,
   MONITOR_PROCESS_STATE_SEMAPHORE       =    0x4,
   MONITOR_PROCESS_STATE_FAST_SEMAPHORE  =    0x8,
   MONITOR_PROCESS_STATE_REMOTE          =   0x10,
   MONITOR_PROCESS_STATE_STOPPED         =   0x20,
   MONITOR_PROCESS_STATE_INTERCEPTED     =   0x40,
   MONITOR_PROCESS_STATE_MUTEX           =   0x80,
   MONITOR_PROCESS_STATE_RUNNING         =  0x400,
   MONITOR_PROCESS_STATE_READY           =  0x800,
   MONITOR_PROCESS_STATE_KILLED          = 0xFFFF
};

/**
 * @type MonitorSignalStatus
 * @brief Signal administrative status type.
 * @seealso MonitorSignalStatusValue
 */
typedef uint32_t MonitorSignalStatus;

/**
 * @type MonitorSignalStatusValue
 * @brief Enumeration of signal administrative status type values.
 * @long These are the values of the field status in
 * a @link "" "MonitorSignalInfo" struct.
 * @field MONITOR_SIGNAL_STATUS_UNKNOWN          Unknown.
 * @field MONITOR_SIGNAL_STATUS_VALID            Valid.
 * @field MONITOR_SIGNAL_STATUS_ENDMARK          Broken administrative
 *                                               endmark in memory.
 * @field MONITOR_SIGNAL_STATUS_ADMINISTRATION   Broken administration block.
 * @field MONITOR_SIGNAL_STATUS_NOT_A_SIGNAL     Not a signal.
 * @field MONITOR_SIGNAL_STATUS_NOT_A_POOL       Not in pool.
 * @field MONITOR_SIGNAL_STATUS_FREE             In a free list.
 *
 * @seealso get_sig_info
 */
enum MonitorSignalStatusValue
{
   MONITOR_SIGNAL_STATUS_UNKNOWN            = 0,
   MONITOR_SIGNAL_STATUS_VALID              = 1,
   MONITOR_SIGNAL_STATUS_ENDMARK            = 2,
   MONITOR_SIGNAL_STATUS_ADMINISTRATION     = 3,
   MONITOR_SIGNAL_STATUS_NOT_A_SIGNAL       = 4,
   MONITOR_SIGNAL_STATUS_NOT_A_POOL         = 5,
   MONITOR_SIGNAL_STATUS_FREE               = 6
};

/**
 * @type MonitorSignalInfo
 * @brief Signal administrative information type.
 * @long These fields contain administrative information
 * about a signal, which are stored outside the actual signal size
 * in the signal buffer.
 *
 * @field number     Signal number.
 * @field owner      Owner process ID.
 * @field sender     Sender process ID.
 * @field addressee  Addressee process ID.
 * @field size       Signal size. The sigNo and data.
 * @field bufsize    Signal buffer size. One of the configured
 *                   signal buffer sizes.
 * @field address    Signal buffer address.
 * @field status     Signal administrative status.
 *                   See @link "" "MonitorSignalStatusValue".
 * @field reserved0  NOT USED.
 * @field reserved1  NOT USED.
 *
 * @seealso OS_sig_info
 * @seealso get_sig_info
 */
typedef struct MonitorSignalInfo
{
   uint32_t number;
   uint32_t owner;
   uint32_t sender;
   uint32_t addressee;
   uint32_t size;
   uint32_t bufsize;
   uint32_t address;
   MonitorSignalStatus status;
   uint32_t reserved0;
   uint32_t reserved1;
} MonitorSignalInfo;

/**
 * @type MonitorScopeType
 * @brief Type code of scope
 * @seealso MonitorScopeTypeValue
 */
typedef uint32_t MonitorScopeType;

/**
 * @type MonitorScopeTypeValue
 * @brief Enumeration of monitor scope types, for matching a set of entities
 * with a PID.
 *
 * @long The generic MONITOR_SCOPE_ID can select any entity with an ID, but
 * in some requests it is limited to a process ID. In other cases it is
 * limited to either a block or a process.
 *
 * @long For specifying names and entry points, separate scope tags are used
 * in the signal. Compare enum @link "" "MonitorScopeTagValue".
 *
 * @long The general syntax for a MONITOR_SCOPE_NAME_PATTERN name
 * is: "[blockname:][processname]", using the tag value MONITOR_SCOPE_TAG_NAME
 * (from enum @link "" "MonitorScopeTagValue").
 * Both the blockname and processname parts are optional. If a part is left out,
 * it is treated as the wildcard "*" string. Wildcard names consist of string
 * fragments mixed with the usual "*" and "?" characters in any combination.
 * A '*' matches any string of zero or more characters, a '?' matches any
 * single character.
 * Note that a name pattern scope only finds processes which already
 * exist in the target, at the moment when requests with this tag are
 * received by the monitor.
 *
 * See an example in the @link "" "MonitorTag" description.
 *
 * @field MONITOR_SCOPE_SYSTEM       All processes.
 * @field MONITOR_SCOPE_ID           The specified process (or other entity).
 * @field MONITOR_SCOPE_BLOCK_ID     The processes in the specified block.
 * @field MONITOR_SCOPE_SEGMENT_ID   The processes in the specified segment.
 * @field MONITOR_SCOPE_NAME_PATTERN The processes matching the specified pattern
 *                                   "[blockname:][processname]" in the tag list.
 *                                   See above for more details.
 * @field MONITOR_SCOPE_PROCESS_NAME A process with the specified name, given
 *                                   in the tag list. (NOT USED).
 * @field MONITOR_SCOPE_BLOCK_NAME   The processes in a block with a specified
 *                                   name given in the tag list.(NOT USED).
 * @field MONITOR_SCOPE_ENTRYPOINT   The processes with the specified entrypoint
 *                                   in the tag list.
 * @field MONITOR_SCOPE_VOID         No processes. The scope is empty.
 *
 * @seealso MonitorScopeTagValue
 */
enum MonitorScopeTypeValue
{
   MONITOR_SCOPE_SYSTEM         = 0x00,
   MONITOR_SCOPE_NAME_PATTERN   = 0x01,
   MONITOR_SCOPE_ENTRYPOINT     = 0x02,
   MONITOR_SCOPE_SEGMENT_ID     = 0x08,
   MONITOR_SCOPE_BLOCK_NAME     = 0x10,
   MONITOR_SCOPE_BLOCK_ID       = 0x20,
   MONITOR_SCOPE_PROCESS_NAME   = 0x40,
   MONITOR_SCOPE_ID             = 0x80,
   MONITOR_SCOPE_VOID           = 0x100
};

/**
 * @type MonitorScope
 *
 * @brief Type used to define a set (scope) of processes (or other entities
 * with an ID).
 *
 * @long The set of processes matching a scope can be specified
 * in a number of different ways, defined by the scope type.
 *
 * @long The easiest way is to specify an ID of an execution context to indicate
 * all processes in that execution context, e.g. a process, a block or
 * a segment. This is supported in most signals.
 *
 * @long Some requests also accept process entrypoint (MONITOR_SCOPE_ENTRYPOINT)
 * or a process name pattern (MONITOR_SCOPE_NAME_PATTERN). In these cases the
 * scope struct is supplemented with separate tags in the signals, to provide
 * the entrypoint or name pattern to match. The general syntax for a name
 * is: "[block:][process]" - see @link "" "MonitorScopeTypeValue" for details.
 *
 * @field type The type of scope, see enum @link "" "MonitorScopeTypeValue".
 * @field id   The ID. Used if type is MONITOR_SCOPE_ID, MONITOR_SCOPE_BLOCK_ID
 *             or MONITOR_SCOPE_SEGMENT_ID. Otherwise should be set to 0.
 *
 * @seealso MonitorScopeTypeValue
 * @seealso MonitorScopeTagValue
 */

typedef struct MonitorScope
{
   MonitorScopeType type;
   uint32_t id;
} MonitorScope;

/**
 * @type MonitorTag
 *
 * @brief A union type used to encode variable length information - tags.
 *
 * @long Tags are used for complementing information in the
 * monitor signals.
 *
 * @long Each tag can take zero or more arguments, which follow
 * immediately after the tag in the list. The size of the area
 * each argument occupies is always rounded up to a multiple
 * of sizeof(MonitorTag). The total size of the tag and all
 * its arguments (tag.count) is given as the number of
 * sizeof(MonitorTag) entries needed to hold them.
 *
 * @long Usually in signal structs, also the total size of the list
 * of all tags is also given as the number of
 * sizeof(MonitorTag) entries needed to hold all of them.
 *
 * @long Since tag arguments can be null-terminated strings as well as
 * integers, byte order conversion of both tags and arguments has to
 * be handled by the monitor client (the byte order is detected in
 * @link "" "MonitorInterfaceReply").
 *
 * @long The tag.type defines the contents of the tag and can be any of
 * the different Monitor*TagValue enums in the monitor signal files,
 * depending on what the tag is used for. A few examples of the many
 * different tag type value enums are:
 * @link "monitor.sig" "MonitorScopeTagValue",
 * @link "monitor.sig" "MonitorSignalFilterTagValue",
 * @link "monitor_events.sig" "MonitorEventActionpointTagValue",
 * @link "monitor_actions.sig" "MonitorCreateProcessTagValue".
 *
 * @field tag          A tag struct.
 * @field tag.type     The tag type - a value from any of the
 *                     Monitor*TagValue enums.
 * @field tag.count    The length of this tag and its arguments, as the
 *                     number of sizeof(MonitorTag).
 *
 * @field num          An integer (32-bit) tag argument.
 * @field str          A string tag argument (null-terminated). Each tag
 *                     slot can contain 4 characters.
 *
 * @seealso MonitorGetInfoRequest
 * @seealso MonitorScopeTypeValue
 * @seealso MonitorScopeTagValue
 *
 * @example
 * // Request information about all processes in a scope where the process //
 * // name begins with "foo", in any block. Limit (filter) to list only    //
 * // processes with at least 10 signals in the signal queue.              //
 * // Here tagsCount in the signal is the number of tag entries occupied   //
 * // by both tags, including tag entries and tag parameters.              //
 * // 1:st tag is the scope name "*:foo*". The argc gives space for the    //
 * // string with terminating null byte and tag.Count also includes space  //
 * // for the tag value itself.                                            //
 * // 2:nd tag contains two numeric values (min and max), with tag.Count   //
 * // including also the tag value itself.                                 //
 * // The ose_monitor_ is the PID of the "ose_monitor"                     //
 *
 * uint32_t argc = strlen("*:foo*") / sizeof(MonitorTag) + 1;
 * MonitorGetInfoRequest *req = alloc(sizeof(MonitorGetInfoRequest)
 *                                    + argc * sizeof(MonitorTag)
 *                                    + 3    * sizeof(MonitorTag),
 *                                    MONITOR_GET_PROCESS_INFO_REQUEST);
 *
 * req->level = 0;
 * req->scope.type = MONITOR_SCOPE_NAME_PATTERN;
 * req->scope.id = 0;
 * req->tagsCount = 1 + argc + 1 + 2;
 * req->tags[0].tag.type = MONITOR_SCOPE_TAG_NAME;
 * req->tags[0].tag.count = 1 + argc;
 * strcpy(req->tags[1].str, "*:foo*");
 * // [2] is also occupied by the string //
 * req->tags[3].tag.type = MONITOR_PROCESS_FILTER_TAG_SIGNALS_IN_QUEUE;
 * req->tags[3].tag.count = 1 + 2;
 * req->tags[4].num = 10;
 * req->tags[5].num = 100000;
 *
 * send((union SIGNAL *)&req, ose_monitor_);
 */
typedef union MonitorTag
{
   struct
   {
      uint16_t type;
      uint16_t count;
   } tag;
   uint32_t num;
   char str[1];
} MonitorTag;

/**
 * @type MonitorScopeTagValue
 *
 * @brief Enumeration of tags types, to define scope subsets.
 *
 * @long These are used in a tag list, for complementing information
 * to the scope type (@link "" "MonitorScopeTypeValue").
 *
 * @field MONITOR_SCOPE_TAG_ENTRYPOINT   Entrypoint of process (int).
 * @field MONITOR_SCOPE_TAG_NAME         Process name pattern (string).
 * @field MONITOR_SCOPE_TAG_END          Highest possible tag value plus one,
 *                                       should not be used.
 *
 * @seealso MonitorScopeTypeValue
 * @seealso MonitorTag
 */
enum MonitorScopeTagValue
{
   MONITOR_SCOPE_TAG_ENTRYPOINT = 1,
   MONITOR_SCOPE_TAG_NAME,
   MONITOR_SCOPE_TAG_END
};

/**
 * @type MonitorSignalFilterTagValue
 *
 * @brief Tags used to define a filter for signals.
 *
 * @long The filter tags that take two integer arguments A,B all interpret
 * those as the range A <= N < B, where N is the filtered number. B=0 means
 * no upper limit.
 *
 * @long The filter tags that take a single name string all accept
 * the wildcard characters '*' and '?'.
 *
 * @field MONITOR_SIGNAL_FILTER_TAG_NOT           Exclude matches of next tag.
 * @field MONITOR_SIGNAL_FILTER_TAG_NUMBER        Signal number range (num, num).
 * @field MONITOR_SIGNAL_FILTER_TAG_OWNER_ID      Owner PID (num).
 * @field MONITOR_SIGNAL_FILTER_TAG_OWNER_NAME    Owner name (str).
 * @field MONITOR_SIGNAL_FILTER_TAG_SENDER_ID     Sender PID (num).
 * @field MONITOR_SIGNAL_FILTER_TAG_SENDER_NAME   Sender name (str).
 * @field MONITOR_SIGNAL_FILTER_TAG_ADDRESSEE_ID  Addressee PID (num).
 * @field MONITOR_SIGNAL_FILTER_TAG_ADDRESSEE_NAME Addressee name (str).
 * @field MONITOR_SIGNAL_FILTER_TAG_SIZE          Signal size range (num, num).
 * @field MONITOR_SIGNAL_FILTER_TAG_BUFSIZE       Signal buffer size range (num, num).
 * @field MONITOR_SIGNAL_FILTER_TAG_ADDRESS       Signal buffer address range (num, num).
 * @field MONITOR_SIGNAL_FILTER_TAG_STATUS        Administrative Status (num) -
 *                                                see @link "" "MonitorSignalStatusValue".
 * @field MONITOR_SIGNAL_FILTER_TAG_OWNER_BID     Owner block ID (num).
 * @field MONITOR_SIGNAL_FILTER_TAG_OWNER_SID     Owner segment ID (num).
 * @field MONITOR_SIGNAL_FILTER_TAG_END           Highest possible tag value
 *                                                plus one, should not be used.
 *
 * @seealso MonitorSignalStatusValue
 * @seealso MonitorTag
 */
enum MonitorSignalFilterTagValue
{
   MONITOR_SIGNAL_FILTER_TAG_NOT = 10,
   MONITOR_SIGNAL_FILTER_TAG_NUMBER,
   MONITOR_SIGNAL_FILTER_TAG_OWNER_ID,
   MONITOR_SIGNAL_FILTER_TAG_OWNER_NAME,
   MONITOR_SIGNAL_FILTER_TAG_SENDER_ID,
   MONITOR_SIGNAL_FILTER_TAG_SENDER_NAME,
   MONITOR_SIGNAL_FILTER_TAG_ADDRESSEE_ID,
   MONITOR_SIGNAL_FILTER_TAG_ADDRESSEE_NAME,
   MONITOR_SIGNAL_FILTER_TAG_SIZE,
   MONITOR_SIGNAL_FILTER_TAG_BUFSIZE,
   MONITOR_SIGNAL_FILTER_TAG_ADDRESS,
   MONITOR_SIGNAL_FILTER_TAG_STATUS,
   MONITOR_SIGNAL_FILTER_TAG_OWNER_BID,
   MONITOR_SIGNAL_FILTER_TAG_OWNER_SID,
   MONITOR_SIGNAL_FILTER_TAG_END
};

/**
 * @type MonitorStackFilterTagValue
 *
 * @brief Tags used to define a filter for stacks.
 *
 * @long The filter tags that take two integer arguments A,B all interpret
 * those as the range A <= N < B, where N is the filtered number. B=0 means
 * no upper limit.
 *
 * @long The filter tags that take a single name string all accept
 * the wildcard characters '*' and '?'.
 *
 * @field MONITOR_STACK_FILTER_TAG_NOT                Exclude matches of next tag
 * @field MONITOR_STACK_FILTER_TAG_OWNER_ID           Owner PID (num).
 * @field MONITOR_STACK_FILTER_TAG_OWNER_NAME         Owner name (str).
 * @field MONITOR_STACK_FILTER_TAG_SIZE               Stack size range (num, num).
 * @field MONITOR_STACK_FILTER_TAG_BUFSIZE            Buffer size range (num, num).
 * @field MONITOR_STACK_FILTER_TAG_ADDRESS            Buffer address range (num, num).
 * @field MONITOR_STACK_FILTER_TAG_USED               Used stack range (num, num).
 * @field MONITOR_STACK_FILTER_TAG_UNUSED             Unused stack range (num, num).
 * @field MONITOR_STACK_FILTER_TAG_SUPERVISOR_SIZE    Supervisor stack size range (num, num).
 * @field MONITOR_STACK_FILTER_TAG_SUPERVISOR_BUFSIZE Supervisor buffer size range (num, num).
 * @field MONITOR_STACK_FILTER_TAG_SUPERVISOR_ADDRESS Supervisor buffer address range (num, num).
 * @field MONITOR_STACK_FILTER_TAG_SUPERVISOR_USED    Supervisor Used stack range (num, num).
 * @field MONITOR_STACK_FILTER_TAG_SUPERVISOR_UNUSED  Supervisor Unused stack range (num, num).
 * @field MONITOR_STACK_FILTER_TAG_END                Highest possible tag value
 *                                                    plus one, should not be used.
 *
 * @seealso MonitorTag
 */
enum MonitorStackFilterTagValue
{
   MONITOR_STACK_FILTER_TAG_NOT = 100,
   MONITOR_STACK_FILTER_TAG_OWNER_ID,
   MONITOR_STACK_FILTER_TAG_OWNER_NAME,
   MONITOR_STACK_FILTER_TAG_SIZE,
   MONITOR_STACK_FILTER_TAG_BUFSIZE,
   MONITOR_STACK_FILTER_TAG_ADDRESS,
   MONITOR_STACK_FILTER_TAG_USED,
   MONITOR_STACK_FILTER_TAG_UNUSED,
   MONITOR_STACK_FILTER_TAG_SUPERVISOR_SIZE,
   MONITOR_STACK_FILTER_TAG_SUPERVISOR_BUFSIZE,
   MONITOR_STACK_FILTER_TAG_SUPERVISOR_ADDRESS,
   MONITOR_STACK_FILTER_TAG_SUPERVISOR_USED,
   MONITOR_STACK_FILTER_TAG_SUPERVISOR_UNUSED,
   MONITOR_STACK_FILTER_TAG_END
};


/**
 * @type MonitorProcessFilterTagValue
 *
 * @brief Tags used to define a filter for processes.
 *
 * @long The filter tags that take two integer arguments A,B all interpret
 * those as the range A <= N < B, where N is the filtered number. B=0 means
 * no upper limit.
 *
 * @long The filter tags that take a single name string all accept
 * the wildcard characters '*' and '?'.
 *
 * @field MONITOR_PROCESS_FILTER_TAG_NOT                Exclude matches of next tag.
 * @field MONITOR_PROCESS_FILTER_TAG_TYPE               Process type (num) - see @link "" "MonitorProcessTypeValue".
 * @field MONITOR_PROCESS_FILTER_TAG_STATE              Process state (num) - see @link "" "MonitorProcessStateValue".
 * @field MONITOR_PROCESS_FILTER_TAG_PRIORITY           Priority range (num, num).
 * @field MONITOR_PROCESS_FILTER_TAG_UID                User ID range (num, num).
 * @field MONITOR_PROCESS_FILTER_TAG_CREATOR_ID         Creator ID (num). (NOT USED)
 * @field MONITOR_PROCESS_FILTER_TAG_SUPERVISOR         Supervisor mode (num) - 1:supervisor, 0:user mode. (NOT USED)
 * @field MONITOR_PROCESS_FILTER_TAG_SIGNALS_IN_QUEUE   Number of signals in queue range (num, num).
 * @field MONITOR_PROCESS_FILTER_TAG_SIGNALS_ATTACHED   Number of signals attached range (num, num). (NOT USED)
 * @field MONITOR_PROCESS_FILTER_TAG_ENTRYPOINT         Entrypoint range (num, num).
 * @field MONITOR_PROCESS_FILTER_TAG_SEMAPHORE_VALUE    Semaphore value range (num, num).
 * @field MONITOR_PROCESS_FILTER_TAG_STACK_SIZE         Stack size range (num, num).
 * @field MONITOR_PROCESS_FILTER_TAG_SUPERVISOR_STACK_SIZE Supervisor stack size range (num, num). (NOT USED)
 * @field MONITOR_PROCESS_FILTER_TAG_LINE_NUMBER        Source code line number range (num, num).
 * @field MONITOR_PROCESS_FILTER_TAG_SIGNALS_OWNED      Number of signals owned range (num, num).
 * @field MONITOR_PROCESS_FILTER_TAG_TIME_SLICE         Time slice range (num, num). (NOT USED)
 * @field MONITOR_PROCESS_FILTER_TAG_VECTOR             Vector range (num, num). (NOT USED)
 * @field MONITOR_PROCESS_FILTER_TAG_ERROR_HANDLER      Error handler (num). (NOT USED)
 * @field MONITOR_PROCESS_FILTER_TAG_SIGSELECT_COUNT    Sigselect count range (num, num).
 * @field MONITOR_PROCESS_FILTER_TAG_NAME               Process name (str).
 * @field MONITOR_PROCESS_FILTER_TAG_FILE_NAME          Source filename (str).
 * @field MONITOR_PROCESS_FILTER_TAG_SYSTEM             System process 1:system process,
 *                                                      0:user process. Compare configuration
 *                                                      of krn/system_process in OSE.
 * @field MONITOR_PROCESS_FILTER_TAG_EU_ID              Bound to execution unit ID range (num, num).
 * @field MONITOR_PROCESS_FILTER_TAG_END                Highest possible tag value
 *                                                      plus one, should not be used.
 *
 * @seealso MonitorTag
 */
enum MonitorProcessFilterTagValue
{
   MONITOR_PROCESS_FILTER_TAG_NOT = 50,
   MONITOR_PROCESS_FILTER_TAG_TYPE,
   MONITOR_PROCESS_FILTER_TAG_STATE,
   MONITOR_PROCESS_FILTER_TAG_PRIORITY,
   MONITOR_PROCESS_FILTER_TAG_UID,
   MONITOR_PROCESS_FILTER_TAG_CREATOR_ID,
   MONITOR_PROCESS_FILTER_TAG_SUPERVISOR,
   MONITOR_PROCESS_FILTER_TAG_SIGNALS_IN_QUEUE,
   MONITOR_PROCESS_FILTER_TAG_SIGNALS_ATTACHED,
   MONITOR_PROCESS_FILTER_TAG_ENTRYPOINT,
   MONITOR_PROCESS_FILTER_TAG_SEMAPHORE_VALUE,
   MONITOR_PROCESS_FILTER_TAG_STACK_SIZE,
   MONITOR_PROCESS_FILTER_TAG_SUPERVISOR_STACK_SIZE,
   MONITOR_PROCESS_FILTER_TAG_LINE_NUMBER,
   MONITOR_PROCESS_FILTER_TAG_SIGNALS_OWNED,
   MONITOR_PROCESS_FILTER_TAG_TIME_SLICE,
   MONITOR_PROCESS_FILTER_TAG_VECTOR,
   MONITOR_PROCESS_FILTER_TAG_ERROR_HANDLER,
   MONITOR_PROCESS_FILTER_TAG_SIGSELECT_COUNT,
   MONITOR_PROCESS_FILTER_TAG_NAME,
   MONITOR_PROCESS_FILTER_TAG_FILE_NAME,
   MONITOR_PROCESS_FILTER_TAG_SYSTEM,
   MONITOR_PROCESS_FILTER_TAG_EU_ID,
   MONITOR_PROCESS_FILTER_TAG_END
};

/**
 * @type MonitorBlockFilterTagValue
 *
 * @brief Tags used to define a filter for blocks.
 *
 * @long The filter tags that take two integer arguments A,B all interpret
 * those as the range A <= N < B, where N is the filtered number. B=0 means
 * no upper limit.
 *
 * @long The filter tags that take a single name string all accept
 * the wildcard characters '*' and '?'.
 *
 * @field MONITOR_BLOCK_FILTER_TAG_NOT              Exclude matches of next tag.
 * @field MONITOR_BLOCK_FILTER_TAG_UID              User ID range (num, num).
 * @field MONITOR_BLOCK_FILTER_TAG_SUPERVISOR       Supervisor mode (num) - 1:supervisor, 0:user mode. (NOT USED)
 * @field MONITOR_BLOCK_FILTER_TAG_SIGNALS_ATTACHED Number of signals attached range (num, num). (NOT USED)
 * @field MONITOR_BLOCK_FILTER_TAG_ERROR_HANDLER    Error handler (num). (NOT USED)
 * @field MONITOR_BLOCK_FILTER_TAG_MAX_SIG_SIZE     Maximum signal size range (num, num).
 * @field MONITOR_BLOCK_FILTER_TAG_SIG_POOL_ID      Signal pool ID (num).
 * @field MONITOR_BLOCK_FILTER_TAG_STACK_POOL_ID    Stack pool ID (num).
 * @field MONITOR_BLOCK_FILTER_TAG_NAME             Block name (str).
 * @field MONITOR_BLOCK_FILTER_TAG_EU_ID            Bound to execution unit ID range (num, num)
 * @field MONITOR_BLOCK_FILTER_TAG_END              Highest possible tag value
 *                                                  plus one, should not be used.
 *
 * @seealso MonitorTag
 */
enum MonitorBlockFilterTagValue
{
   MONITOR_BLOCK_FILTER_TAG_NOT = 75,
   MONITOR_BLOCK_FILTER_TAG_UID,
   MONITOR_BLOCK_FILTER_TAG_SUPERVISOR,
   MONITOR_BLOCK_FILTER_TAG_SIGNALS_ATTACHED,
   MONITOR_BLOCK_FILTER_TAG_ERROR_HANDLER,
   MONITOR_BLOCK_FILTER_TAG_MAX_SIG_SIZE,
   MONITOR_BLOCK_FILTER_TAG_SIG_POOL_ID,
   MONITOR_BLOCK_FILTER_TAG_STACK_POOL_ID,
   MONITOR_BLOCK_FILTER_TAG_NAME,
   MONITOR_BLOCK_FILTER_TAG_EU_ID,
   MONITOR_BLOCK_FILTER_TAG_END
};

/**
 * @type MonitorHeapBufferFilterTagValue
 *
 * @brief Tags used to define a filter for heap buffers.
 *
 * @long The filter tags that take two integer arguments A,B all interpret
 * those as the range A <= N < B, where N is the filtered number. B=0 means
 * no upper limit.
 *
 * @long The filter tags that take a single name string all accept
 * the wildcard characters '*' and '?'.
 *
 * @field MONITOR_HEAP_BUFFER_FILTER_TAG_NOT          Exclude matches of next tag.
 * @field MONITOR_HEAP_BUFFER_FILTER_TAG_OWNER_ID     Owner process ID (num).
 * @field MONITOR_HEAP_BUFFER_FILTER_TAG_OWNER_NAME   Owner process name (str).
 * @field MONITOR_HEAP_BUFFER_FILTER_TAG_SHARED       Heap buffer shared mode (num),
 *                                                    1 = shared, 0 = private.
 * @field MONITOR_HEAP_BUFFER_FILTER_TAG_ADDRESS      Heap buffer address range (num, num).
 * @field MONITOR_HEAP_BUFFER_FILTER_TAG_SIZE         Actual heap buffer size range (num, num).
 * @field MONITOR_HEAP_BUFFER_FILTER_TAG_REQ_SIZE     Requested heap buffer size range (num, num).
 * @field MONITOR_HEAP_BUFFER_FILTER_TAG_FILE_NAME    Source code file name (str).
 * @field MONITOR_HEAP_BUFFER_FILTER_TAG_LINE_NUMBER  Source code line number range (num, num).
 * @field MONITOR_HEAP_BUFFER_FILTER_TAG_STATUS       Administrative status (num), see
 *                                                    @link "" "MonitorHeapBufferStatusValue".
 * @field MONITOR_HEAP_BUFFER_FILTER_TAG_END          Highest possible tag value
 *                                                    plus one, should not be used.
 *
 * @seealso MonitorHeapBufferStatusValue
 * @seealso MonitorTag
 */
enum MonitorHeapBufferFilterTagValue
{
   MONITOR_HEAP_BUFFER_FILTER_TAG_NOT = 200,
   MONITOR_HEAP_BUFFER_FILTER_TAG_OWNER_ID,
   MONITOR_HEAP_BUFFER_FILTER_TAG_OWNER_NAME,
   MONITOR_HEAP_BUFFER_FILTER_TAG_SHARED,
   MONITOR_HEAP_BUFFER_FILTER_TAG_ADDRESS,
   MONITOR_HEAP_BUFFER_FILTER_TAG_SIZE,
   MONITOR_HEAP_BUFFER_FILTER_TAG_REQ_SIZE,
   MONITOR_HEAP_BUFFER_FILTER_TAG_FILE_NAME,
   MONITOR_HEAP_BUFFER_FILTER_TAG_LINE_NUMBER,
   MONITOR_HEAP_BUFFER_FILTER_TAG_STATUS,
   MONITOR_HEAP_BUFFER_FILTER_TAG_END
};

/**
 * @type MonitorGetInfoRequest
 *
 * @brief Request signal struct for information requests.
 *
 * @input sigNo        Signal number for this Get Info request.
 * @input level        1 for long reply signals or 0 for regular reply signals,
 *                     when this struct is used in signals supporting long
 *                     reply signals. Otherwise should be 0.
 * @input scope        The scope of processes or other entities to examine.
 * @input reserved0    NOT USED. Should be 0.
 * @input tagsCount    The length of the filter list, as the total number of tag
 *                     entries, e.g. the tags area in bytes / sizeof(MonitorTag).
 * @input tags         The filter conditions.
 */
typedef struct MonitorGetInfoRequest
{
   uint32_t sigNo;
   uint32_t level;
   MonitorScope scope;
   uint32_t reserved0;
   uint32_t tagsCount;
   MonitorTag tags[1];
} MonitorGetInfoRequest;

/**
 * @type MonitorGetInfoEndmark
 *
 * @brief Endmark signal struct for information requests that result in multiple
 * reply signals.
 *
 * @output sigNo  Signal number of this endmark reply.
 * @output status MONITOR_STATUS_OK or reason of failure.
 */
typedef struct MonitorGetInfoEndmark
{
   uint32_t sigNo;
   MonitorStatus status;
} MonitorGetInfoEndmark;


/*******************************************************************************
 *
 * CONNECTION/DISCONNECTION SIGNALS
 *
 ******************************************************************************/

/**
 * @signal MonitorInterfaceRequest
 *
 * @brief List monitor interface information request (before connecting).
 *
 * @long This signal should be the first signal sent from the client to the
 * monitor, to obtain basic information about the target and the monitor
 * interface, before connecting to the monitor with the
 * @link "" "MonitorConnectRequest" signal.
 *
 * @long The monitor byte order, indicated in the
 * @link "" "MonitorInterfaceReply", must be used in all monitor signals.
 *
 * @long For extended information, use the
 * @link "" "MonitorConnectRequest" signal.
 *
 * @long A client does not need to be connected to the monitor to send this
 * signal.
 *
 * @long Send this to the "ose_monitor" process.
 *
 * @long This signals is the only signal that can be used even if the RMM
 * signal interface has been disabled (via krn/monitor/interface).
 *
 * @member MONITOR_INTERFACE_REQUEST
 * @member MonitorInterfaceRequest
 *
 * @seealso MonitorInterfaceReply
 * @seealso MonitorConnectRequest
 */

/** @macro MONITOR_INTERFACE_REQUEST */
#define MONITOR_INTERFACE_REQUEST 39000 /* !-SIGNO(MonitorInterfaceRequest)-! */

/**
 * @type MonitorInterfaceRequest
 *
 * @input sigNo                 MONITOR_INTERFACE_REQUEST
 */
typedef struct MonitorInterfaceRequest
{
   uint32_t sigNo;
} MonitorInterfaceRequest;

/**
 * @macro MONITOR_ENDIAN_VALUE
 * @brief Endian constant that declares the monitor byte order.
 *
 * @long This endian integer value is 0x01020304L.
 *
 * @long The value is a long (32-bit) integer, in the monitor byte order,
 * indicated by the monitor in the signal
 * @link "" "MonitorInterfaceReply".
 * The client shall use it to find out the byte order
 * used in all signals to/from the monitor.
 *
 * @seealso MonitorInterfaceReply
 */
#define MONITOR_ENDIAN_VALUE 0x01020304L

/**
 * @type MonitorOSType
 * @brief Type used to indicate the flavor of OSE.
 * @seealso MonitorOSTypeValue
 */
typedef uint32_t MonitorOSType;

/**
 * @type MonitorOSTypeValue
 * @brief Enumeration of OSE Flavors
 * @long This is indicated by the monitor in the signal
 * @link "" "MonitorInterfaceReply".
 * @field MONITOR_OS_OSE_5        OSE 5.x.
 * @field MONITOR_OS_OSE_4        OSE 4.x.
 * @field MONITOR_OS_OSE_EPSILON  OSE Epsilon.
 * @field MONITOR_OS_OSE_CK       OSEck.
 * @field MONITOR_OS_UNKNOWN      The OS is unknown.
 * @seealso MonitorOSType
 */
enum MonitorOSTypeValue
{
   MONITOR_OS_OSE_5         = 0,
   MONITOR_OS_OSE_4,
   MONITOR_OS_OSE_EPSILON,
   MONITOR_OS_OSE_CK,

   MONITOR_OS_UNKNOWN = 0x7fffffff
};

/**
 * @type MonitorCPUClass
 * @brief CPU class type.
 * @seealso MonitorCPUClassValue
 */
typedef uint32_t MonitorCPUClass;

/**
 * @type MonitorCPUClassValue
 * @brief Enumeration of CPU architectures/classes.
 * @long This is indicated by the monitor in the signal
 * @link "" "MonitorInterfaceReply".
 * @field MONITOR_CPU_PPC     PowerPC.
 * @field MONITOR_CPU_M68K    Motorola 68000.
 * @field MONITOR_CPU_ARM     ARM.
 * @field MONITOR_CPU_MIPS    MIPS.
 * @field MONITOR_CPU_X86     Intel x86.
 * @field MONITOR_CPU_SPARC   SPARC.
 * @field MONITOR_CPU_C64X    TMS320c64x.
 * @field MONITOR_CPU_C64XP   TMS320c64x+.
 * @field MONITOR_CPU_SP26XX  SP26xx.
 * @field MONITOR_CPU_SP27XX  SP27xx.
 * @field MONITOR_CPU_TL3     Ceva-TL3.
 * @field MONITOR_CPU_CEVAX   Ceva-X.
 * @field MONITOR_CPU_M8144   MSC8144.
 * @field MONITOR_CPU_M8156   MSC8156.
 * @field MONITOR_CPU_UNKNOWN CPU unknown to the monitor.
 * @seealso MonitorCPUClass
 */
enum MonitorCPUClassValue
{
   MONITOR_CPU_PPC   = 0,
   MONITOR_CPU_M68K,
   MONITOR_CPU_ARM,
   MONITOR_CPU_MIPS,
   MONITOR_CPU_X86,
   MONITOR_CPU_SPARC,
   MONITOR_CPU_C64X,
   MONITOR_CPU_C64XP,
   MONITOR_CPU_SP26XX,
   MONITOR_CPU_SP27XX,
   MONITOR_CPU_TL3,
   MONITOR_CPU_CEVAX,
   MONITOR_CPU_M8144,
   MONITOR_CPU_M8156,

   MONITOR_CPU_UNKNOWN = 0x7fffffff
};

/**
 * @signal MonitorInterfaceReply
 *
 * @brief List monitor interface information reply.
 *
 * @long The list of supported request signals can be limited, e.g.
 * if the RMM was configured with krn/monitor/interface=crippled:1,
 * also depending on any calls to enable selected requests.
 * The MonitorInterfaceRequest is however always supported.
 *
 * @member MONITOR_INTERFACE_REPLY
 * @member MonitorInterfaceReply
 *
 * @seealso MONITOR_ENDIAN_VALUE
 * @seealso MonitorOSTypeValue
 * @seealso MonitorCPUClassValue
 * @seealso MonitorInterfaceRequest
 */

/** @macro MONITOR_INTERFACE_REPLY */
#define MONITOR_INTERFACE_REPLY 39001 /* !-SIGNO(MonitorInterfaceReply)-! */

/**
 * @type MonitorInterfaceReply
 *
 * @output sigNo                 MONITOR_INTERFACE_REPLY
 * @output endian                @link "" "MONITOR_ENDIAN_VALUE" constant in
 *                               the byte order of the monitor, used in all
 *                               monitor signals.
 * @output osType                Flavor of OSE
 *                               - see @link "" "MonitorOSTypeValue".
 * @output cpuClass              CPU class, i.e. MONITOR_CPU_*
 *                               - see @link "" "MonitorCPUClassValue".
 * @output sigsCount          Number of signals numbers in sigs.
 * @output sigs               Array of supported request signal numbers.
 * @output reserved0  NOT USED.
 * @output reserved1  NOT USED.
 * @output reserved2  NOT USED.
 */
typedef struct MonitorInterfaceReply
{
   uint32_t sigNo;
   uint32_t reserved0;
   uint8_t endian[4];
   MonitorOSType osType;
   MonitorCPUClass cpuClass;
   uint32_t reserved1;
   uint32_t reserved2;
   uint32_t sigsCount;
   uint32_t sigs[1];
} MonitorInterfaceReply;

/**
 * @signal MonitorNameRequest
 *
 * @brief List monitor names request.
 *
 * @long This signal obtains names of the system and monitor.
 *
 * @long A client does not need to be connected to the monitor to send this
 * signal, but the client must know the byte order, from previous
 * @link "" "MonitorInterfaceReply".
 *
 * @long Send this to the "ose_monitor" process.
 *
 * @member MONITOR_NAME_REQUEST
 * @member MonitorNameRequest
 *
 * @seealso MonitorNameReply
 */

/** @macro MONITOR_NAME_REQUEST */
#define MONITOR_NAME_REQUEST 39006 /* !-SIGNO(MonitorNameRequest)-! */

/**
 * @type MonitorNameRequest
 *
 * @input sigNo                 MONITOR_NAME_REQUEST
 *
 */
typedef struct MonitorNameRequest
{
   uint32_t sigNo;
} MonitorNameRequest;

/**
 * @signal MonitorNameReply
 *
 * @brief List monitor names reply.
 *
 * @long These are describing null-terminated name strings (not process names),
 * returned by the monitor, to be used by the client to identify the
 * monitor type.
 *
 * @long An example of a monitor name is "com.ose.RunModeMonitor",
 * returned by RMM.
 *
 * @member MONITOR_NAME_REPLY
 * @member MonitorNameReply
 * @seealso MonitorNameRequest
 */

/** @macro MONITOR_NAME_REPLY */
#define MONITOR_NAME_REPLY 39007 /* !-SIGNO(MonitorNameReply)-! */

/**
 * @type MonitorNameReply
 *
 * @output sigNo              MONITOR_NAME_REPLY
 * @output systemNameOffset   Offset in data to system name. (NOT USED)
 * @output monitorNameOffset  Offset in data to monitor name.
 * @output data               The names.
 * @output reserved0          NOT USED.
 * @output reserved1          NOT USED.
 * @output reserved2          NOT USED.
 */
typedef struct MonitorNameReply
{
   uint32_t sigNo;
   uint32_t reserved0;
   uint32_t reserved1;
   uint32_t reserved2;
   uint32_t systemNameOffset;
   uint32_t monitorNameOffset;
   char data[1];
} MonitorNameReply;

/** @private @macro MONITOR_OBSOLETE_CONNECT_REQUEST */
#define MONITOR_OBSOLETE_CONNECT_REQUEST 39002
/** @private @macro MONITOR_OBSOLETE_CONNECT_REPLY */
#define MONITOR_OBSOLETE_CONNECT_REPLY 39003

/**
 * @signal MonitorConnectRequest
 *
 * @brief Monitor connection request.
 *
 * @long Request a connection to the monitor, which is needed to use the
 * monitor. Without a connection, only the
 * @link "" "MonitorInterfaceRequest" and @link "" "MonitorNameRequest"
 * signals and this connect request should be sent to the monitor.
 *
 * @long In addition, also @link "" "MonitorAttachRequest" is needed before
 * using more advanced debugging or event signals.
 *
 * @long The reply @link "" "MonitorConnectReply" will contain extended
 * information about the monitor, like a list of supported features
 * (@link "" "MonitorFeatureValue"), timing information etc.
 *
 * @long For more information about the monitor interface, see
 * @link "" "MonitorInterfaceReply".
 *
 * @long The client's selected name should be unique, if multiple clients
 * are used, mainly used to identify log messages. The name string is
 * truncated to maximum 80 characters by the monitor.
 *
 * @long Send this signal to the "ose_monitor" process.
 *
 * @long After connection, all other signals to the monitor should be sent to
 * the sender PID of the @link "" "MonitorConnectReply".
 *
 * @member MONITOR_CONNECT_REQUEST
 * @member MonitorConnectRequest
 *
 * @seealso MonitorConnectReply
 * @seealso MonitorAttachRequest
 * @seealso MonitorInterfaceRequest
 */

/** @macro MONITOR_CONNECT_REQUEST */
#define MONITOR_CONNECT_REQUEST 39008 /* !-SIGNO(MonitorConnectRequest)-! */

/**
 * @type MonitorConnectRequest
 *
 * @input sigNo                 MONITOR_CONNECT_REQUEST
 * @input name                  The client's selected name describing who and where
 *                              the client is. Should be unique when multiple
 *                              clients are connected.
 */
typedef struct MonitorConnectRequest
{
   uint32_t sigNo;
   char name[1];
} MonitorConnectRequest;

/**
 * @type MonitorFeature
 * @brief Type used for a debug monitor feature in a list of features
 *
 * @long Monitor features are returned in @link "" "MonitorConnectReply".
 * See @link "" "MonitorFeatureValue" for values.
 *
 * @seealso MonitorFeatureValue
 * @seealso MonitorConnectReply
 * @seealso MonitorInterfaceReply
 */
typedef uint16_t MonitorFeature;

/**
 * @type MonitorFeatureValue
 *
 * @brief Enumeration of debug monitor features.
 *
 * @long Monitor features are used to indicate restrictions or additions in the
 * semantics of the operations or the behavior of the monitor. Early versions
 * of the monitor do not support all features.
 *
 * @long The supported features are returned by the monitor in a list in the
 * @link "" "MonitorConnectReply" signal.
 *
 * @long Compare also the list of supported requests, returned by the monitor
 * in the @link "" "MonitorInterfaceReply".
 *
 * @seealso MonitorConnectReply
 * @seealso MonitorAttachReply
 * @seealso MonitorInterfaceReply
 * @seealso MonitorGetInfoRequest
 * @seealso MonitorFeature
 *
 *
 * @field MONITOR_FEATURE_SCOPE
 * The monitor supports main scopes that are smaller than MONITOR_SCOPE_SYSTEM
 * (@link "" "MonitorScopeTypeValue"), to narrow the scope used for
 * notification, trace and trigger operations.
 *
 * @field MONITOR_FEATURE_INTERCEPT_SCOPE
 * The monitor supports intercept scopes (subscope of the main scope) in
 * @link "" "MonitorAttachRequest" that are smaller than
 * MONITOR_SCOPE_SYSTEM (@link "" "MonitorScopeTypeValue").
 *
 * @field MONITOR_FEATURE_AUTOMATIC_SCOPE
 * The monitor supports automatic scope reduction. This means that scopes specified
 * in requests, that require that the client is attached, will automatically
 * be reduced to subsets of the scopes that the client specified when attaching
 * with @link "" "MonitorAttachRequest".
 * If reduction is not supported, the client must make sure that the scopes it
 * sets are subsets of the attached scopes.
 *
 * @field MONITOR_FEATURE_COMPACT_REGISTER_SIGNALS
 * The monitor uses and expects get and set register signals without redundant
 * value fields.
 *
 * @field MONITOR_FEATURE_SYSTEM_PROCESSES
 * The monitor supports the notion of system processes, that is processes that
 * will never be intercepted, and allows the client to tell which processes
 * are system processes, and to exclude or include such processes in filter
 * expressions. See the krn/system_process system configuration in the
 * "OSE Core User's Guide".
 *
 * @field MONITOR_FEATURE_MULTIPLE_EVENT_CLIENTS
 * The monitor supports multiple clients to the event action system at the same
 * time, and does not require the client to set the
 * @link "" "MonitorAttachTagValue" MONITOR_ATTACH_TAG_EVENT_LOCK
 * tag when attaching, to be able to use the event system.
 * (NOT YET IMPLEMENTED in standard OSE).
 *
 * @field MONITOR_FEATURE_EVENT_TRACE
 * The monitor maintains a trace buffer and supports tracing of
 * system events using the @link "monitor_events.sig"
 * "MonitorActionValues" MONITOR_ACTION_TRACE,
 * MONITOR_ACTION_ENABLE_TRACE and MONITOR_ACTION_DISABLE_TRACE.
 *
 * @field MONITOR_FEATURE_EVENT_NOTIFY
 * The monitor supports notification of system events using
 * MONITOR_ACTION_NOTIFY (@link "monitor_events.sig" "MonitorActionValues").
 *
 * @field MONITOR_FEATURE_EVENT_INTERCEPT
 * The monitor supports interception of processes on system events using the
 * MONITOR_ACTION_INTERCEPT (@link "monitor_events.sig" "MonitorActionValues").
 *
 * @field MONITOR_FEATURE_EVENT_USER_ACTION
 * The monitor supports calling a user specified callback function and
 * deliver a pointer to an event report to that function.
 * MONITOR_ACTION_USER (@link "monitor_events.sig" "MonitorActionValues").
 * Compare rmm.h, function @link "rmm.h" "ose_rmm_set_user_report_handler".
 *
 * @field MONITOR_FEATURE_EVENT_STATES
 * The monitor supports the notion of event states in which system event
 * actionpoints can be grouped and activated in response to other system events
 * using the MONITOR_ACTION_SET_STATE (@link "monitor_events.sig"
 * "MonitorActionValues").
 *
 * @field MONITOR_FEATURE_PROCESS_FILTER
 * The monitor supports @link "" "MonitorProcessFilterTagValue" in process
 * information request signals.
 *
 * @field MONITOR_FEATURE_BLOCK_FILTER
 * The monitor supports @link "" "MonitorBlockFilterTagValue" in block
 * information request signals.
 *
 * @field MONITOR_FEATURE_SEGMENT_FILTER
 * The monitor supports MonitorSegmentFilterTagValue in segment
 * information request signals.
 *
 * @field MONITOR_FEATURE_SIGNAL_FILTER
 * The monitor supports @link "" "MonitorSignalFilterTagValue" in signal
 * information request signals.
 *
 * @field MONITOR_FEATURE_STACK_FILTER
 * The monitor supports @link "" "MonitorStackFilterTagValue" in stack
 * information request signals.
 *
 * @field MONITOR_FEATURE_COMPACT_INFO_SIGNALS
 * The monitor supports only short information signals, i.e. the level
 * field in struct MonitorGetInfoRequest should be 0 in request
 * signals, e.g. in @link "" "MonitorGetSysParamRequest" and other
 * signals which support long replies.
 *
 * @field MONITOR_FEATURE_EVENT_UNDO
 * The monitor supports undo of system events using
 * MONITOR_ACTION_UNDO (@link "monitor_events.sig" "MonitorActionValues").
 *
 * @field MONITOR_FEATURE_PROF_PRIORITY_THRESHOLD
 * The monitor supports priority thresholds for 'per block' and 'per process' CPU
 * profiling. Processes with priority lower or equal to the threshold value are
 * ignored in the measurement.
 *
 * @field MONITOR_FEATURE_PROF_ALL_EXECUTION_UNITS
 * The monitor supports profiling of sum of cores
 * MONITOR_ALL_EXECUTION_UNITS.
 *
 * @field MONITOR_FEATURE_CLEAR_ALL_EVENT_ACTIONPOINTS
 * The monitor supports clearing all event actionpoints in one
 * MONITOR_CLEAR_EVENT_ACTIONPOINT_REQUEST.
 *
 * @field MONITOR_FEATURE_TOTAL_CPU_BLOCK_MEASURE
 * The monitor supports total CPU block measurement.
 *
 * @field MONITOR_FEATURE_SEGMENT_INFO_LONG
 * The monitor supports the long reply signal for segment information.
 *
 * @field MONITOR_FEATURE_PER_CORE_PROFILED_PROCESSES
 * The monitor supports per core profiled processes measurement.
 *
 * @field MONITOR_FEATURE_EVENT_HW_TRACE
 * The monitor supports enable/disable hardware trace 
 * on system events by using MONITOR_ACTION_ENABLE_HW_TRACE and 
 * MONITOR_ACTION_DISABLE_HW_TRACE (@link "monitor_events.sig" "MonitorActionValues").
 *
 * @field MONITOR_FEATURE_ACTIONPOINT_DATA_RANGE
 * The monitor supports data filtering for send, receive, free_buf, and
 * user_event events. See tag MONITOR_EVENT_ACTIONPOINT_TAG_DATA_RANGE
 * (@link "monitor_events.sig" "MonitorEventActionpointTagValue").
 *
 * @field MONITOR_FEATURE_TRACEPOINT_ANNOTATIONS
 * The monitor supports setting and getting tracepoint annotations
 * by using MONITOR_TRACEPOINT_SET_ANNOTATIONS_REQUEST and
 * MONITOR_TRACEPOINT_GET_ANNOTATIONS_REQUEST
 * (@link "monitor_tracepoint.sig" "MonitorTracepointSetAnnotationsRequest"
 * and @link "monitor_tracepoint.sig" "MonitorTracepointGetAnnotationsRequest").
 *
 * @field MONITOR_FEATURE_ACCUMULATED_HEAP_PROFILING
 * The monitor supports accumulated heap profiling.
 *
 * @field MONITOR_FEATURE_RESERVED_DO_NOT_USE
 * Reserved. Should not be used.
 *
 * @field MONITOR_FEATURE_END Highest possible tag value
 *                            plus one, should not be used.
 */
enum MonitorFeatureValue
{
   MONITOR_FEATURE_SCOPE =                          1,
   MONITOR_FEATURE_INTERCEPT_SCOPE =                2,
   MONITOR_FEATURE_AUTOMATIC_SCOPE =                3,
   MONITOR_FEATURE_COMPACT_REGISTER_SIGNALS  =      4,
   MONITOR_FEATURE_SYSTEM_PROCESSES =               5,
   MONITOR_FEATURE_MULTIPLE_EVENT_CLIENTS =         6,
   MONITOR_FEATURE_EVENT_TRACE =                    7,
   MONITOR_FEATURE_EVENT_NOTIFY =                   8,
   MONITOR_FEATURE_EVENT_INTERCEPT =                9,
   MONITOR_FEATURE_EVENT_STATES =                  10,
   MONITOR_FEATURE_PROCESS_FILTER =                11,
   MONITOR_FEATURE_BLOCK_FILTER =                  12,
   MONITOR_FEATURE_SEGMENT_FILTER =                13,
   MONITOR_FEATURE_SIGNAL_FILTER =                 14,
   MONITOR_FEATURE_STACK_FILTER =                  15,
   MONITOR_FEATURE_COMPACT_INFO_SIGNALS =          16,
   MONITOR_FEATURE_EVENT_USER_ACTION =             17,
   MONITOR_FEATURE_EVENT_UNDO =                    18,
   MONITOR_FEATURE_PROF_PRIORITY_THRESHOLD =       19,
   MONITOR_FEATURE_PROF_ALL_EXECUTION_UNITS =      20,
   MONITOR_FEATURE_CLEAR_ALL_EVENT_ACTIONPOINTS =  21,
   MONITOR_FEATURE_TOTAL_CPU_BLOCK_MEASURE =       22,
   MONITOR_FEATURE_RESERVED_DO_NOT_USE =           23,
   MONITOR_FEATURE_SEGMENT_INFO_LONG =             24,
   MONITOR_FEATURE_PER_CORE_PROFILED_PROCESSES =   25,
   MONITOR_FEATURE_EVENT_HW_TRACE =                26,
   /* In other versions the FEATURE_64_BIT is 27 */
   MONITOR_FEATURE_ACTIONPOINT_DATA_RANGE =        28, 
   MONITOR_FEATURE_TRACEPOINT_ANNOTATIONS =        29,
   MONITOR_FEATURE_ACCUMULATED_HEAP_PROFILING =    30,
   MONITOR_FEATURE_END
};

/**
 * @type MonitorStartMode
 * 
 * @field MONITOR_START_MODE_UNKNOWN
 * The start mode is unknown or this particular monitor do not support this
 * information.
 * 
 * @field MONITOR_START_MODE_COLD
 * The monitor is started in cold mode meaning that no actionpoints or the
 * system trace are recovered from a previous execution.
 * 
 * @field MONITOR_START_MODE_WARM
 * The monitor is started in warm mode meaning that some settings and data may
 * be recovered from the previous execution.
 * 
 * @seealso MonitorConnectReply
 */
enum MonitorStartMode
{
   MONITOR_START_MODE_UNKNOWN =   0,
   MONITOR_START_MODE_COLD =      1,
   MONITOR_START_MODE_WARM =      2
};

/**
 * @signal MonitorConnectReply
 *
 * @brief Monitor connection reply.
 *
 * @long After the connection to the debug monitor is established, the debug
 * client shall use the sender PID of MonitorConnectReply in all further
 * communication with the debug monitor.
 *
 * @seealso MonitorFeatureValue
 * @seealso MonitorCPUClassValue
 * @seealso MonitorConnectRequest
 * @seealso MonitorStartMode
 *
 * @member MONITOR_CONNECT_REPLY
 * @member MonitorConnectReply
 */

/** @macro MONITOR_CONNECT_REPLY */
#define MONITOR_CONNECT_REPLY 39009 /* !-SIGNO(MonitorConnectReply)-! */

/**
 * @type MonitorConnectReply
 *
 * @output sigNo                   MONITOR_CONNECT_REPLY
 * @output status                  Result of connection attempt.
 * @output tickLength              Number of microseconds for one tick.
 * @output tick                    Number of ticks since system start.
 * @output ntick                   Number of timer steps after the last tick.
 * @output cpuClass                CPU class, i.e. MONITOR_CPU_*.
 * @output cpuType                 CPU type, architecture specific information.
 *                                 Holds PVR value (POWERPC ONLY).
 * @output featuresCount           Count of supportedFeatures array
 * @output features                Array of supported features -
 *                                 @link "" "MonitorFeatureValue".
 * @output errorHandler            Address of the system error handler (NOT USED)
 * @output ntickFrequency          Frequency or the ntick timer in Hz.
 * @output euCount                 Number of execution units (CPU cores) on
 *                                 this target or 0 if the target does not support
 *                                 multiple execution units.
 * @output startMode               Monitor start mode, see enum MonitorStartMode
 * @output reserved0               NOT USED.
 * @output reserved1               NOT USED.
 * @output reserved2               NOT USED.
 * @output reserved3               NOT USED.
 * @output reserved4               NOT USED.
 * @output reserved5               NOT USED.
 * @output reserved6               NOT USED.
 * @output reserved7               NOT USED.
 */
typedef struct MonitorConnectReply
{
   uint32_t sigNo;
   MonitorStatus status;
   uint32_t tickLength;
   uint32_t tick;
   uint32_t ntick;
   uint32_t ntickFrequency;
   MonitorCPUClass cpuClass;
   uint32_t cpuType;
   uint32_t errorHandler;
   uint16_t euCount;
   uint8_t  startMode;
   uint8_t  reserved0;
   uint32_t reserved1;
   uint32_t reserved2;
   uint32_t reserved3;
   uint32_t reserved4;
   uint32_t reserved5;
   uint32_t reserved6;
   uint32_t reserved7;
   uint32_t featuresCount;
   MonitorFeature features[1];
} MonitorConnectReply;

/**
 * @signal MonitorDisconnectRequest
 *
 * @brief Disconnect from debug monitor request.
 *
 * @long It is recommended to detach any active
 * monitor attachment with @link "monitor_execution.sig" "MonitorDetachRequest"
 * before disconnecting.
 *
 * @long There is no reply to this request signal.
 *
 * @member MONITOR_DISCONNECT_REQUEST
 * @member MonitorDisconnectRequest
 */

/** @macro MONITOR_DISCONNECT_REQUEST */
#define MONITOR_DISCONNECT_REQUEST 39004 /* !-SIGNO(MonitorDisconnectRequest)-! */

/**
 * @type MonitorDisconnectRequest
 *
 * @input sigNo MONITOR_DISCONNECT_REQUEST
 */
typedef struct MonitorDisconnectRequest
{
   uint32_t sigNo;
} MonitorDisconnectRequest;

/*******************************************************************************
 *
 * FLOW CONTROL SIGNALS
 *
 ******************************************************************************/

/**
 * @type MonitorFlowControl
 *
 * @brief Type used in flow control - blocking or non-blocking
 * in signals:
 *              MONITOR_BLOCKING_FLOW_CONTROL_REQUEST,
 *              MONITOR_BLOCKING_FLOW_CONTROL_REPLY,
 *              MONITOR_NON_BLOCKING_FLOW_CONTROL_REQUEST,
 *              MONITOR_NON_BLOCKING_FLOW_CONTROL_REPLY
 *
 * @field sigNo Signal number for the flow control request or reply.
 */
typedef struct MonitorFlowControl
{
   uint32_t sigNo;
} MonitorFlowControl;

/**
 * @signal MonitorBlockingFlowControlRequest
 *
 * @brief Sent from monitor to allow the client to control the flow of
 * signals, sent to the client under blocking flow control.
 *
 * @long The monitor will attempt to send this signal in time for the client
 * to respond (with @link "" "MonitorBlockingFlowControlReply")
 * and maintain a uninterrupted flow of signals.
 *
 * @long System parameters (krn/monitor/flow_ctrl) configure the maximum number
 * of signals sent to the client after a MonitorBlockingFlowControlRequest,
 * before the monitor will block, unless the client has responded with a reply.
 *
 * @long Should the client fail to respond to this signal before the maximum
 * number of signals is reached, the monitor will block and hold
 * pending signals to be sent under blocking flow control, and also
 * throw away any signals to be sent under non-blocking flow control and queue
 * new client requests. Once the client responds, operation will be resumed.
 *
 * @seealso MonitorBlockingFlowControlReply
 * @seealso MonitorNonBlockingFlowControlRequest
 *
 * @member MONITOR_BLOCKING_FLOW_CONTROL_REQUEST
 * @member MonitorFlowControl
 */

/** @macro MONITOR_BLOCKING_FLOW_CONTROL_REQUEST */
#define MONITOR_BLOCKING_FLOW_CONTROL_REQUEST 39065 /* !-SIGNO(MonitorFlowControl)-! */

/**
 * @signal MonitorBlockingFlowControlReply
 *
 * @brief Acknowledge sent from client in response to a
 * @link "" "MonitorBlockingFlowControlRequest".
 *
 * @member MONITOR_BLOCKING_FLOW_CONTROL_REPLY
 * @member MonitorFlowControl
 */

/** @macro MONITOR_BLOCKING_FLOW_CONTROL_REPLY */
#define MONITOR_BLOCKING_FLOW_CONTROL_REPLY 39066 /* !-SIGNO(MonitorFlowControl)-! */

/**
 * @signal MonitorNonBlockingFlowControlRequest
 *
 * @brief Sent from monitor to allow the client to control the flow of
 * signals, sent to the client under non-blocking flow control.
 *
 * @long The monitor will attempt to send this signal in time for the client
 * to respond (with @link "" "MonitorNonBlockingFlowControlReply")
 * and maintain a uninterrupted flow of signals.
 *
 * @long System parameters (krn/monitor/flow_ctrl) configure the maximum number
 * of signals sent to the client after a MonitorNonBlockingFlowControlRequest,
 * before the monitor will start discarding signals, unless the client has
 * responded with a reply.
 *
 * @long Should the client fail to respond to this signal before the maximum
 * number of signals is reached, the monitor will throw away signals, that
 * should be sent to the client under non-blocking flow control.
 * Other signal transactions are not affected.
 *
 * @long Note that signals sent with non-blocking flow control will also be
 * thrown away, if the client does not respond to "blocking" flow control
 * requests from the monitor -
 * compare @link "" "MonitorBlockingFlowControlRequest").
 *
 * @seealso MonitorNonBlockingFlowControlReply
 * @seealso MonitorBlockingFlowControlRequest
 *
 * @member MONITOR_NON_BLOCKING_FLOW_CONTROL_REQUEST
 * @member MonitorFlowControl
 */

/** @macro MONITOR_NON_BLOCKING_FLOW_CONTROL_REQUEST */
#define MONITOR_NON_BLOCKING_FLOW_CONTROL_REQUEST 39067 /* !-SIGNO(MonitorFlowControl)-! */

/**
 * @signal MonitorNonBlockingFlowControlReply
 *
 * @brief Acknowledge, sent from client to monitor in response to a
 * @link "" "MonitorNonBlockingFlowControlRequest".
 *
 * @member MONITOR_NON_BLOCKING_FLOW_CONTROL_REPLY
 * @member MonitorFlowControl
 */

/** @macro MONITOR_NON_BLOCKING_FLOW_CONTROL_REPLY */
#define MONITOR_NON_BLOCKING_FLOW_CONTROL_REPLY 39068 /* !-SIGNO(MonitorFlowControl)-! */


/*******************************************************************************
 *
 * PROCESS/BLOCK/SEGMENT INFO SIGNALS
 *
 ******************************************************************************/

/**
 * @signal MonitorGetProcessInfoRequest
 *
 * @brief List processes request.
 *
 * @long Followed by zero or more MonitorGetProcessInfoReply or
 * MonitorGetProcessInfoLongReply (depending on the requested level),
 * MonitorBlockingFlowControl request and one MonitorGetProcessInfoEndmark
 * (with status).
 *
 * @long If level is 0, the process filter tags MONITOR_PROCESS_FILTER_TAG_NOT,
 * MONITOR_PROCESS_FILTER_TAG_TYPE, MONITOR_PROCESS_FILTER_TAG_STATE,
 * MONITOR_PROCESS_FILTER_TAG_PRIORITY, MONITOR_PROCESS_FILTER_TAG_NAME,
 * MONITOR_PROCESS_FILTER_TAG_ENTRYPOINT and MONITOR_PROCESS_FILTER_TAG_SYSTEM
 * are accepted. Short replies are in @link "" "MonitorGetProcessInfoReply" signals.
 *
 * @long If level is 1, all process filter tags are accepted. Long replies
 * are in @link "" "MonitorGetProcessInfoLongReply" signals.
 *
 * @long See a program example in the @link "" "MonitorTag" description.
 *
 * @seealso MonitorProcessFilterTagValue
 * @seealso MonitorGetProcessInfoReply
 * @seealso MonitorGetProcessInfoLongReply
 *
 * @member MONITOR_GET_PROCESS_INFO_REQUEST
 * @member MonitorGetInfoRequest
 */

/** @macro MONITOR_GET_PROCESS_INFO_REQUEST */
#define MONITOR_GET_PROCESS_INFO_REQUEST 39010 /* !-SIGNO(MonitorGetInfoRequest)-! */

/**
 * @signal MonitorGetProcessInfoReply
 *
 * @brief List processes short reply - one reply per process.
 *
 * @long Reply to @link "" "MonitorGetProcessInfoRequest", if short info
 * was requested (level is 0).
 *
 * @long This signal is sent under blocking flow control.
 *
 * @member MONITOR_GET_PROCESS_INFO_REPLY
 * @member MonitorGetProcessInfoReply
 */

/** @macro MONITOR_GET_PROCESS_INFO_REPLY */
#define MONITOR_GET_PROCESS_INFO_REPLY 39011 /* !-SIGNO(MonitorGetProcessInfoReply)-! */

/**
 * @type MonitorGetProcessInfoReply
 *
 * @output sigNo      MONITOR_GET_PROCESS_INFO_REPLY
 * @output pid        ID of the process.
 * @output bid        ID of the parent block.
 * @output sid        ID of the parent segment or 0.
 * @output type       Type of the process, a @link "" "MonitorProcessTypeValue".
 * @output state      State of the process, a @link "" "MonitorProcessStateValue".
 * @output priority   Priority of the process, if valid for this process type.
 * @output entrypoint Entrypoint of the process.
 * @output properties Bit field indicating additional properties of the process.
 *                    Bit 0 is set to 1 for system processes.
 *                    Other bits are undefined.
 * @output name       Null terminated name of the process.
 * @output reserved1  NOT USED.
 */
typedef struct MonitorGetProcessInfoReply
{
   uint32_t sigNo;
   uint32_t pid;
   uint32_t bid;
   uint32_t sid;
   MonitorProcessType type;
   MonitorProcessState state;
   uint32_t priority;
   uint32_t entrypoint;
   uint32_t properties;
   uint32_t reserved1;
   char name[1];
} MonitorGetProcessInfoReply;

/**
 * @signal MonitorGetProcessInfoLongReply
 *
 * @brief List processes long reply - one reply per process.
 *
 * @long Reply to @link "" "MonitorGetProcessInfoRequest", if long info
 * was requested (level is 1).
 *
 * @long This signal is sent under blocking flow control.
 *
 * @member MONITOR_GET_PROCESS_INFO_LONG_REPLY
 * @member MonitorGetProcessInfoLongReply
 */

/** @macro MONITOR_GET_PROCESS_INFO_LONG_REPLY */
#define MONITOR_GET_PROCESS_INFO_LONG_REPLY 39012 /* !-SIGNO(MonitorGetProcessInfoLongReply)-! */

/**
 * @type MonitorGetProcessInfoLongReply
 *
 * @output sigNo               MONITOR_GET_PROCESS_INFO_LONG_REPLY
 * @output pid                 ID of the process.
 * @output bid                 ID of the parent block.
 * @output sid                 ID of the parent segment or 0.
 * @output type                Type of the process, a @link "" "MonitorProcessTypeValue".
 * @output state               State of the process, a @link "" "MonitorProcessStateValue".
 * @output priority            Priority of the process, if valid for this process
 * @output uid                 User ID of the process.
 * @output creator             ID of the creating process. (NOT USED)
 * @output supervisor          True if executing in supervisor mode. (NOT USED)
 *                             False if executing in user mode. (NOT USED)
 * @output signalsInQueue      Number of signals in the receive queue.
 * @output signalsAttached     Number of signals in the attach list. (NOT USED)
 * @output entrypoint          Entrypoint of the process.
 * @output semaphoreValue      Value of the fast semaphore counter.
 * @output stackSize           Stack size of the process.
 * @output supervisorStackSize Supervisor mode stack size for user mode processes.
 *                             (NOT_USED)
 * @output lineNumber          Source file line number if the process is in a
 *                             system call and the source is compiled with
 *                             OS_DEBUG defined
 *                             - see also @link "ose.h" "OS_pcb" in ose.h.
 * @output signalsOwned        Number of signals owned.
 * @output timeSlice           Timeslice for the process (NOT USED)
 * @output vector              Hardware vector for interrupt processes (NOT USED)
 * @output errorHandler        Error handler for the process (NOT USED)
 * @output sigselectOffset     Offset in data of sigselect list.
 * @output sigselectCount      Number of signal numbers + 1 (32-bit words) in
 *                             the sigselect list. Note that the first entry is
 *                             a value with the number of signals in the list.
 * @output processNameOffset   Offset in data of null terminated name of the
 *                             process.
 * @output fileNameOffset      Offset in data of null terminated name of the
 *                             source file if the process is in a system call
 *                             and the source is compiled with OS_DEBUG
 *                             defined.
 * @output properties          Bit field indicating additional properties of
 *                             the process.
 *                             Bit 0 is set to 1 for system processes.
 *                             Other bits are undefined.
 * @output euId                Execution unit ID (CPU core) this process is
 *                             bound to.
 * @output data                Variable sized buffer.
 * @output reserved1           NOT USED.
 * @output reserved2           NOT USED.
 */
typedef struct MonitorGetProcessInfoLongReply
{
   uint32_t sigNo;
   uint32_t pid;
   uint32_t bid;
   uint32_t sid;
   MonitorProcessType type;
   MonitorProcessState state;
   uint32_t priority;
   uint32_t uid;
   uint32_t creator;
   MonitorBoolean supervisor;
   uint32_t signalsInQueue;
   uint32_t signalsAttached;
   uint32_t entrypoint;
   int32_t semaphoreValue;
   uint32_t stackSize;
   uint32_t supervisorStackSize;
   uint32_t lineNumber;
   uint32_t signalsOwned;
   uint32_t timeSlice;
   uint32_t vector;
   uint32_t errorHandler;
   uint32_t properties;
   uint16_t euId;
   uint16_t reserved1;
   uint32_t reserved2;
   uint32_t sigselectOffset;
   uint32_t sigselectCount;
   uint32_t processNameOffset;
   uint32_t fileNameOffset;
   uint8_t data[1];
} MonitorGetProcessInfoLongReply;

/**
 * @signal MonitorGetProcessInfoEndmark
 *
 * @brief List processes endmark, with status.
 *
 * @seealso MonitorGetProcessInfoReply
 * @seealso MonitorGetProcessInfoLongReply
 *
 * @member MONITOR_GET_PROCESS_INFO_ENDMARK
 * @member MonitorGetInfoEndmark
 */

/** @macro MONITOR_GET_PROCESS_INFO_ENDMARK */
#define MONITOR_GET_PROCESS_INFO_ENDMARK 39013 /* !-SIGNO(MonitorGetInfoEndmark)-! */

/**
 * @signal MonitorGetBlockInfoRequest
 *
 * @brief List blocks request.
 *
 * @long Followed by zero or more MonitorGetBlockInfoReply,
 * MonitorBlockingFlowControlRequest, and one MonitorGetBlockInfoEndmark
 * (with status).
 *
 * @long The "level" field is not used and should be set to 0.
 *
 * @long Accepts filter tags MONITOR_BLOCK_FILTER_TAG_NAME,
 * MONITOR_BLOCK_FILTER_TAG_UID, MONITOR_BLOCK_FILTER_TAG_MAX_SIG_SIZE,
 * MONITOR_BLOCK_FILTER_TAG_SIG_POOL_ID, MONITOR_BLOCK_FILTER_TAG_STACK_POOL_ID.
 *
 * @restriction Accepts only scopes of types MONITOR_SCOPE_ID,
 *              MONITOR_SCOPE_SEGMENT_ID and MONITOR_SCOPE_SYSTEM.
 *
 * @seealso MonitorGetBlockInfoReply
 * @seealso MonitorScopeTagValue
 * @seealso MonitorBlockFilterTagValue
 *
 * @member MONITOR_GET_BLOCK_INFO_REQUEST
 * @member MonitorGetInfoRequest
 */

/** @macro MONITOR_GET_BLOCK_INFO_REQUEST */
#define MONITOR_GET_BLOCK_INFO_REQUEST 39014 /* !-SIGNO(MonitorGetInfoRequest)-! */

/**
 * @signal MonitorGetBlockInfoReply
 *
 * @brief List blocks reply - one reply per block.
 *
 * @long This signal is sent under blocking flow control.
 *
 * @seealso MonitorGetBlockInfoRequest
 *
 * @member MONITOR_GET_BLOCK_INFO_REPLY
 * @member MonitorGetBlockInfoReply
 */

/** @macro MONITOR_GET_BLOCK_INFO_REPLY */
#define MONITOR_GET_BLOCK_INFO_REPLY 39015 /* !-SIGNO(MonitorGetBlockInfoReply)-! */

/**
 * @type MonitorGetBlockInfoReply
 *
 * @output sigNo           MONITOR_GET_BLOCK_INFO_REPLY
 * @output bid             Block ID.
 * @output sid             Segment ID.
 * @output uid             User ID.
 * @output supervisor      MONITOR_BOOLEAN_TRUE if supervisor mode block.
 *                         MONITOR_BOOLEAN_FALSE if user mode block. (NOT USED)
 * @output signalsAttached Number of signals in attach list (NOT USED)
 * @output errorHandler    Error handler. (NOT USED)
 * @output maxSigSize      Maximum signal size.
 * @output sigPoolId       Signal pool ID.
 * @output stackPoolId     Stack pool ID.
 * @output euId            Execution unit ID, indicated which execution
 *                         unit (CPU core) this block was last bound to.
 * @output name            Null terminated block name.
 * @output reserved0       NOT USED.
 * @output reserved1       NOT USED.
 */
typedef struct MonitorGetBlockInfoReply
{
   uint32_t sigNo;
   uint32_t bid;
   uint32_t sid;
   uint32_t uid;
   MonitorBoolean supervisor;
   uint32_t signalsAttached;
   uint32_t errorHandler;
   uint32_t maxSigSize;
   uint32_t sigPoolId;
   uint32_t stackPoolId;
   uint16_t euId;
   uint16_t reserved0;
   uint32_t reserved1;
   char name[1];
} MonitorGetBlockInfoReply;

/**
 * @signal MonitorGetBlockInfoEndmark
 *
 * @brief List blocks endmark, with status.
 *
 * @seealso MonitorGetBlockInfoReply
 *
 * @member MONITOR_GET_BLOCK_INFO_ENDMARK
 * @member MonitorGetInfoEndmark
 */

/** @macro MONITOR_GET_BLOCK_INFO_ENDMARK */
#define MONITOR_GET_BLOCK_INFO_ENDMARK 39016 /* !-SIGNO(MonitorGetInfoEndmark)-! */

/**
 * @signal MonitorGetSegmentInfoRequest
 *
 * @brief List segments request.
 *
 * @long Followed by zero or more MonitorGetSegmentInfoReply (or
 * MonitorGetSegmentInfoLongReply), MonitorBlockingFlowControl and one
 * MonitorGetSegmentInfoEndmark (with status).
 *
 * @long The "level" field is set to 0 for MonitorGetSegmentInfoReply and to 1
 * for MonitorGetSegmentInfoLongReply. Note that if the
 * MONITOR_FEATURE_SEGMENT_INFO_LONG feature is not reported the level field
 * will be ignored and the monitor will reply with MonitorGetSegmentInfoReply.
 *
 * @restriction Accepts only scopes of types MONITOR_SCOPE_ID
 *              and MONITOR_SCOPE_SYSTEM.
 *
 * @seealso MonitorGetSegmentInfoReply
 * @seealso MonitorGetSegmentInfoLongReply
 * @seealso MonitorScopeTagValue
 *
 * @member MONITOR_GET_SEGMENT_INFO_REQUEST
 * @member MonitorGetInfoRequest
 */

/** @macro MONITOR_GET_SEGMENT_INFO_REQUEST */
#define MONITOR_GET_SEGMENT_INFO_REQUEST 39017 /* !-SIGNO(MonitorGetInfoRequest)-! */

/**
 * @signal MonitorGetSegmentInfoReply
 *
 * @brief List segment reply.
 *
 * @long This signal is sent under blocking flow control.
 *
 * @seealso MonitorGetSegmentInfoReqest
 *
 * @member MONITOR_GET_SEGMENT_INFO_REPLY
 * @member MonitorGetSegmentInfoReply
 */

/** @macro MONITOR_GET_SEGMENT_INFO_REPLY */
#define MONITOR_GET_SEGMENT_INFO_REPLY 39018 /* !-SIGNO(MonitorGetSegmentInfoReply)-! */

/**
 * @type MonitorGetSegmentInfoReply
 *
 * @output sigNo       MONITOR_GET_SEGMENT_INFO_REPLY
 * @output sid         Segment ID.
 * @output number      Segment number or 0 if not available.
 * @output reserved0   NOT USED.
 * @output reserved1   NOT USED.
 * @output reserved2   NOT USED.
 */
typedef struct MonitorGetSegmentInfoReply
{
   uint32_t sigNo;
   uint32_t sid;
   uint32_t number;
   uint32_t reserved0;
   uint32_t reserved1;
   uint32_t reserved2;
} MonitorGetSegmentInfoReply;

/** @macro MONITOR_GET_SEGMENT_INFO_LONG_REPLY */
#define MONITOR_GET_SEGMENT_INFO_LONG_REPLY 39020 /* !-SIGNO(MonitorGetSegmentInfoReply)-! */
/**
 * @type MonitorGetSegmentInfoLongReply
 * @brief Get segment info.
 *
 * @output sigNo           MONITOR_GET_SEGMENT_INFO_REPLY
 * @output sid             Segment ID.
 * @output number          Segment number or 0 if not available (typical is the MM domain ID).
 * @output signalsAttached Number of signals attached to the segment (NOT USED).
 * @output masMapped       If non-zero the segment contains memory pages that
 *                         are MAS mapped.
 * @output reserved0       NOT USED.
 * @output reserved1       NOT USED.
 * @output reserved2       NOT USED.
 * @output name            Name of the segment, or the name of the main block in
 *                         the segment, or an empty string if no name is
 *                         available.
 */
typedef struct MonitorGetSegmentInfoLongReply
{
   uint32_t sigNo;
   uint32_t sid;
   uint32_t number;
   uint32_t signalsAttached;
   uint32_t masMapped;
   uint32_t reserved0;
   uint32_t reserved1;
   uint32_t reserved2;
   char     name[1];
} MonitorGetSegmentInfoLongReply;


/**
 * @signal MonitorGetSegmentInfoEndmark
 *
 * @brief List blocks endmark, with status.
 *
 * @seealso MonitorGetSegmentInfoReply
 *
 * @member MONITOR_GET_SEGMENT_INFO_ENDMARK
 * @member MonitorGetInfoEndmark
 */

/** @macro MONITOR_GET_SEGMENT_INFO_ENDMARK */
#define MONITOR_GET_SEGMENT_INFO_ENDMARK 39019 /* !-SIGNO(MonitorGetInfoEndmark)-! */


/*******************************************************************************
 *
 * STACK INFO SIGNALS
 *
 ******************************************************************************/

/**
 * @signal MonitorGetStackUsageRequest
 *
 * @brief Get stack usage information request.
 *
 * @long Followed by zero or more MonitorGetStackUsageReply,
 * MonitorBlockingFlowControlRequest, and one MonitorGetStackUsageEndmark.
 *
 * @long The "level" field is not used and should be set to 0.
 *
 * @long Accepts all values of @link "" "MonitorStackFilterTagValue".
 *
 * @seealso MonitorGetStackUsageReply
 * @seealso MonitorScopeTagValue
 *
 * @member MONITOR_GET_STACK_USAGE_REQUEST
 * @member MonitorGetInfoRequest
 */

/** @macro MONITOR_GET_STACK_USAGE_REQUEST */
#define MONITOR_GET_STACK_USAGE_REQUEST 39025 /* !-SIGNO(MonitorGetInfoRequest)-! */

/**
 * @signal MonitorGetStackUsageReply
 *
 * @brief Get stack usage information reply - one per process.
 *
 * @long This signal is sent under blocking flow control.
 *
 * @seealso MonitorStackInfo
 * @seealso MonitorGetStackUsageRequest
 *
 * @member MONITOR_GET_STACK_USAGE_REPLY
 * @member MonitorGetStackUsageReply
 */

/** @macro MONITOR_GET_STACK_USAGE_REPLY */
#define MONITOR_GET_STACK_USAGE_REPLY 39026 /* !-SIGNO(MonitorGetStackUsageReply)-! */

/**
 * @type MonitorStackInfo
 *
 * @brief Stack information.
 * @long Returned in the @link "" "MonitorGetStackUsageReply".
 * @field address Start address of the stack.
 * @field size Size of the stack.
 * @field used Peak value of used stack space.
 * @field bufsize Stack buffer size or 0.
 * @field reserved1  NOT USED.
 * @field reserved2  NOT USED.
 */
typedef struct MonitorStackInfo
{
   uint32_t address;
   uint32_t size;
   uint32_t used;
   uint32_t bufsize;
   uint32_t reserved1;
   uint32_t reserved2;
} MonitorStackInfo;

/**
 * @type MonitorGetStackUsageReply
 *
 * @output sigNo      MONITOR_GET_STACK_USAGE_REPLY
 * @output pid        ID of the process that the info is for.
 * @output main       @link "" "MonitorStackInfo" with information about
 *                    the default stack of the process.
 * @output supervisor @link "" "MonitorStackInfo" with information about
 *                    the supervisor stack of the process
 *                    (for user mode processes). (NOT USED)
 */
typedef struct MonitorGetStackUsageReply
{
   uint32_t sigNo;
   uint32_t pid;
   MonitorStackInfo main;
   MonitorStackInfo supervisor;
} MonitorGetStackUsageReply;

/**
 * @signal MonitorGetStackUsageEndmark
 *
 * @brief Get stack usage information endmark, with status.
 *
 * @seealso MonitorGetStackUsageReply
 *
 * @member MONITOR_GET_STACK_USAGE_ENDMARK
 * @member MonitorGetInfoEndmark
 */

/** @macro MONITOR_GET_STACK_USAGE_ENDMARK */
#define MONITOR_GET_STACK_USAGE_ENDMARK 39027 /* !-SIGNO(MonitorGetInfoEndmark)-! */


/*******************************************************************************
 *
 * ENVIRONMENT VARIABLE SIGNALS
 *
 ******************************************************************************/

/**
 * @signal MonitorGetEnvVarsRequest
 *
 * @brief Get environment variables for a process or block request.
 *
 * @long Followed by any number of MonitorGetEnvVarsReply,
 * MonitorBlockingFlowControlRequest and one MonitorGetEnvVarsEndmark.
 *
 * @long Environment variables set with set_envp() are decoded and
 * presented as text.
 *
 * @long The "level" field is not used and should be set to 0.
 *
 * @restriction Accepts only scope type MONITOR_SCOPE_ID
 * (indicating a block or a process).
 *
 * @member MONITOR_GET_ENV_VARS_REQUEST
 * @member MonitorGetInfoRequest
 */

/** @macro MONITOR_GET_ENV_VARS_REQUEST */
#define MONITOR_GET_ENV_VARS_REQUEST 39030 /* !-SIGNO(MonitorGetInfoRequest)-! */

/**
 * @signal MonitorGetEnvVarsReply
 *
 * @brief Get environment variables for a process/block reply - one reply per variable.
 *
 * @long This signal is sent under blocking flow control.
 *
 * @member MONITOR_GET_ENV_VARS_REPLY
 * @member MonitorGetEnvVarsReply
 */

/** @macro MONITOR_GET_ENV_VARS_REPLY */
#define MONITOR_GET_ENV_VARS_REPLY 39031 /* !-SIGNO(MonitorGetEnvVarsReply)-! */

/**
 * @type MonitorGetEnvVarsReply
 *
 * @output sigNo       MONITOR_GET_ENV_VARS_REPLY
 * @output pid         ID of the process/block that has the environment variable.
 * @output nameOffset  Offset in data to the null terminated environment variable name.
 * @output valueOffset Offset in data to the null terminated environment variable value.
 * @output data        Variable sized buffer.
 * @output reserved0   NOT USED.
 * @output reserved1   NOT USED.
 */
typedef struct MonitorGetEnvVarsReply
{
   uint32_t sigNo;
   uint32_t pid;
   uint32_t nameOffset;
   uint32_t valueOffset;
   uint32_t reserved0;
   uint32_t reserved1;
   char data[1];
} MonitorGetEnvVarsReply;

/**
 * @signal MonitorGetEnvVarsEndmark
 *
 * @brief Get environment variables for a process/block endmark, with status.
 *
 * @seealso MonitorGetEnvVarsReply
 *
 * @member MONITOR_GET_ENV_VARS_ENDMARK
 * @member MonitorGetInfoEndmark
 */

/** @macro MONITOR_GET_ENV_VARS_ENDMARK */
#define MONITOR_GET_ENV_VARS_ENDMARK 39032 /* !-SIGNO(MonitorGetInfoEndmark)-! */


/*******************************************************************************
 *
 * SIGNAL QUEUE SIGNALS
 *
 ******************************************************************************/

/**
 * @signal MonitorGetSignalQueueRequest
 *
 * @brief Get signal queue for a process request.
 *
 * @long Followed by any number of MonitorGetSignalQueueReply or
 * MonitorGetSignalQueueLongReply, MonitorBlockingFlowControlRequest,
 * and one MonitorGetSignalQueueEndmark (with status).
 *
 * @long Long replies contains also the signal data contents.
 *
 * @restriction Accepts only scope type MONITOR_SCOPE_ID
 *
 * @member MONITOR_GET_SIGNAL_QUEUE_REQUEST
 * @member MonitorGetInfoRequest
 */

/** @macro MONITOR_GET_SIGNAL_QUEUE_REQUEST */
#define MONITOR_GET_SIGNAL_QUEUE_REQUEST 39035 /* !-SIGNO(MonitorGetInfoRequest)-! */

/**
 * @signal MonitorGetSignalQueueReply
 *
 * @brief Get signal queue for a process (no signal contents) reply
 *        - one reply per signal.
 *
 * @member MONITOR_GET_SIGNAL_QUEUE_REPLY
 * @member MonitorGetSignalQueueReply
 */

/** @macro MONITOR_GET_SIGNAL_QUEUE_REPLY */
#define MONITOR_GET_SIGNAL_QUEUE_REPLY 39036 /* !-SIGNO(MonitorGetSignalQueueReply)-! */

/**
 * @type MonitorGetSignalQueueReply
 *
 * @output sigNo        MONITOR_GET_SIGNAL_QUEUE_REPLY
 * @output pid          ID of the process holding the signal queue.
 * @output signal       Signal information @link "" "MonitorSignalInfo"
 */
typedef struct MonitorGetSignalQueueReply
{
   uint32_t sigNo;
   uint32_t pid;
   MonitorSignalInfo signal;
} MonitorGetSignalQueueReply;

/**
 * @signal MonitorGetSignalQueueLongReply
 *
 * @brief Get signal queue for a process (with signal contents) long reply
 *        - one reply per signal.
 *
 * @long This signal is sent under blocking flow control.
 *
 * @member MONITOR_GET_SIGNAL_QUEUE_LONG_REPLY
 * @member MonitorGetSignalQueueLongReply
 */

/** @macro MONITOR_GET_SIGNAL_QUEUE_LONG_REPLY */
#define MONITOR_GET_SIGNAL_QUEUE_LONG_REPLY 39037 /* !-SIGNO(MonitorGetSignalQueueLongReply)-! */

/**
 * @type MonitorGetSignalQueueLongReply
 *
 * @output sigNo        MONITOR_GET_SIGNAL_QUEUE_LONG_REPLY
 * @output pid          ID of the process holding the signal queue.
 * @output signal       Signal information @link "" "MonitorSignalInfo"
 * @output dataCount    Number of bytes in data.
 * @output data         Signal data.
 * @output reserved0    NOT USED.
 */
typedef struct MonitorGetSignalQueueLongReply
{
   uint32_t sigNo;
   uint32_t pid;
   MonitorSignalInfo signal;
   uint32_t reserved0;
   uint32_t dataCount;
   uint8_t data[1];
} MonitorGetSignalQueueLongReply;

/**
 * @signal MonitorGetSignalQueueEndmark
 *
 * @brief Get signal queue for a process endmark, with status.
 *
 * @seealso MonitorGetSignalQueueReply
 * @seealso MonitorGetSignalQueueLongReply
 *
 * @member MONITOR_GET_SIGNAL_QUEUE_ENDMARK
 * @member MonitorGetInfoEndmark
 */

/** @macro MONITOR_GET_SIGNAL_QUEUE_ENDMARK */
#define MONITOR_GET_SIGNAL_QUEUE_ENDMARK 39038 /* !-SIGNO(MonitorGetInfoEndmark)-! */

/*******************************************************************************
 *
 * ATTACH LIST SIGNALS
 *
 ******************************************************************************/

/**
 * @signal MonitorGetAttachListRequest
 *
 * @brief Get attach list for a process request. NOT USED.
 *
 * @long This is the list of processes which has attached to a process
 * using the OSE @link "ose.h" "attach"() function.
 *
 * @long In long replies, also the attach signal data is included.
 *
 * @long Followed by any number of MonitorGetAttachListReply or
 * MonitorGetAttachListLongReply, MonitorBlockingFlowControlRequest,
 * and one MonitorGetAttachListEndmark.
 *
 * @restriction Accepts only scope type MONITOR_SCOPE_ID.
 *
 * @member MONITOR_GET_ATTACH_LIST_REQUEST
 * @member MonitorGetInfoRequest
 * @seealso MonitorGetAttachListReply
 * @seealso MonitorGetAttachListLongReply
 */

/** @macro MONITOR_GET_ATTACH_LIST_REQUEST */
#define MONITOR_GET_ATTACH_LIST_REQUEST 39040 /* !-SIGNO(MonitorGetInfoRequest)-! */

/**
 * @signal MonitorGetAttachListReply
 *
 * @brief Get attach for a process (no signal contents) reply
 *        - one attach per reply.
 *
 * @long This signal is sent under blocking flow control.
 *
 * @member MONITOR_GET_ATTACH_LIST_REPLY
 * @member MonitorGetAttachListReply
 */

/** @macro MONITOR_GET_ATTACH_LIST_REPLY */
#define MONITOR_GET_ATTACH_LIST_REPLY 39041 /* !-SIGNO(MonitorGetAttachListReply)-! */

/**
 * @type MonitorGetAttachListReply
 *
 * @output sigNo        MONITOR_GET_ATTACH_LIST_REPLY
 * @output pid          ID of the process holding the attach.
 * @output signal       Signal information @link "" "MonitorSignalInfo"
 */
typedef struct MonitorGetAttachListReply
{
   uint32_t sigNo;
   uint32_t pid;
   MonitorSignalInfo signal;
} MonitorGetAttachListReply;

/**
 * @signal MonitorGetAttachListLongReply
 *
 * @brief Get attach list for a process (with signal contents) long reply
 *        - one attach per reply.
 *
 * @long This signal is sent under blocking flow control.
 *
 * @member MONITOR_GET_ATTACH_LIST_LONG_REPLY
 * @member MonitorGetAttachListLongReply
 */

/** @macro MONITOR_GET_ATTACH_LIST_LONG_REPLY */
#define MONITOR_GET_ATTACH_LIST_LONG_REPLY 39042 /* !-SIGNO(MonitorGetAttachListLongReply)-! */

/**
 * @type MonitorGetAttachListLongReply
 *
 * @output sigNo        MONITOR_GET_ATTACH_LIST_LONG_REPLY
 * @output pid          ID of the process holding the attach.
 * @output signal       Signal information @link "" "MonitorSignalInfo"
 * @output dataCount    Number of bytes in data.
 * @output data         Signal data.
 * @output reserved0    NOT USED.
 */
typedef struct MonitorGetAttachListLongReply
{
   uint32_t sigNo;
   uint32_t pid;
   MonitorSignalInfo signal;
   uint32_t reserved0;
   uint32_t dataCount;
   uint8_t data[1];
} MonitorGetAttachListLongReply;

/**
 * @signal MonitorGetAttachListEndmark
 *
 * @brief Get signal queue for a process endmark, with status.
 *
 * @seealso MonitorGetAttachListReply
 * @seealso MonitorGetAttachListLongReply
 *
 * @member MONITOR_GET_ATTACH_LIST_ENDMARK
 * @member MonitorGetInfoEndmark
 */

/** @macro MONITOR_GET_ATTACH_LIST_ENDMARK */
#define MONITOR_GET_ATTACH_LIST_ENDMARK 39043 /* !-SIGNO(MonitorGetInfoEndmark)-! */


/*******************************************************************************
 *
 * REDIRECTION TABLE SIGNALS
 *
 ******************************************************************************/

/**
 * @signal MonitorGetSignalRedirectionRequest
 *
 * @brief List process redirection table request. NOT USED.
 *
 * @long Followed with zero or more MonitorGetSignalRedirectionReply,
 * MonitorBlockingFlowControlRequest, and one MonitorGetSignalRedirectionEndmark.
 *
 * @long The "level" field is not used and should be set to 0.
 *
 * @restriction Accepts only scope type MONITOR_SCOPE_ID.
 *
 * @seealso MonitorGetSignalRedirectionReply
 * @seealso MonitorRedirection
 *
 * @member MONITOR_GET_SIGNAL_REDIRECTION_REQUEST
 * @member MonitorGetInfoRequest
 */

/** @macro MONITOR_GET_SIGNAL_REDIRECTION_REQUEST */
#define MONITOR_GET_SIGNAL_REDIRECTION_REQUEST 39045 /* !-SIGNO(MonitorGetInfoRequest)-! */

/**
 * @type MonitorRedirection
 * @brief An entry in a signal redirection table.
 *
 * @long In the first entry in a redirection table, the fields contain
 * special values: sig contains a count (number of entries, including
 * first) and PID contains the default process ID, used when the signal number
 * is not found in the redirection table.
 *
 *
 * @field sig The signal number (1:st entry: number of entries).
 * @field pid The process ID, to which signals with number sig shall be sent.
 *            (1:st entry: default process ID).
 *
 * @seealso OS_redir_entry
 */
typedef struct MonitorRedirection
{
   uint32_t sig;
   uint32_t pid;
} MonitorRedirection;

/**
 * @signal MonitorGetSignalRedirectionReply
 *
 * @brief List process redirection table reply.
 *
 * @long This signal is sent under blocking flow control.
 *
 * @seealso MonitorGetSignalRedirectionRequest
 *
 * @member MONITOR_GET_SIGNAL_REDIRECTION_REPLY
 * @member MonitorGetSignalRedirectionReply
 */

/** @macro MONITOR_GET_SIGNAL_REDIRECTION_REPLY */
#define MONITOR_GET_SIGNAL_REDIRECTION_REPLY 39046 /* !-SIGNO(MonitorGetSignalRedirectionReply)-! */

/**
 * @type MonitorGetSignalRedirectionReply
 *
 * @output sigNo             MONITOR_GET_SIGNAL_REDIRECTION_REPLY
 * @output pid               ID of the process.
 * @output redirectionCount  Number of elements in redirection table.
 * @output redirection       Redirection table contents (including count
 *                           and default PID in the first element).
 *                           See type @link "" "MonitorRedirection".
 * @output reserved0  NOT USED.
 */
typedef struct MonitorGetSignalRedirectionReply
{
   uint32_t sigNo;
   uint32_t pid;
   uint32_t reserved0;
   uint32_t redirectionCount;
   MonitorRedirection redirection[1];
} MonitorGetSignalRedirectionReply;

/**
 * @signal MonitorGetSignalRedirectionEndmark
 *
 * @brief List process redirection table endmark, with status.
 *
 * @seealso MonitorGetSignalRedirectionRequest
 *
 * @member MONITOR_GET_SIGNAL_REDIRECTION_ENDMARK
 * @member MonitorGetInfoEndmark
 */

/** @macro MONITOR_GET_SIGNAL_REDIRECTION_ENDMARK */
#define MONITOR_GET_SIGNAL_REDIRECTION_ENDMARK 39047 /* !-SIGNO(MonitorGetInfoEndmark)-! */

/*******************************************************************************
 *
 * POOL INFO SIGNALS
 *
 ******************************************************************************/

/**
 * @signal MonitorGetPoolInfoRequest
 *
 * @brief Get pool information request.
 *
 * @long Followed by any number of MonitorGetPoolInfoReply,
 * MonitorBlockingFlowControl, and one MonitorGetPoolInfoEndmark.
 *
 * @long The "level" field is not used and should be set to 0.
 *
 * @restriction Accepts only scope types MONITOR_SCOPE_ID
 * (indicating a pool), MONITOR_SCOPE_SEGMENT_ID, and MONITOR_SCOPE_SYSTEM.
 *
 * @seealso MonitorGetPoolInfoReply
 *
 * @member MONITOR_GET_POOL_INFO_REQUEST
 * @member MonitorGetInfoRequest
 */

/** @macro MONITOR_GET_POOL_INFO_REQUEST */
#define MONITOR_GET_POOL_INFO_REQUEST 39050 /* !-SIGNO(MonitorGetInfoRequest)-! */

/**
 * @signal MonitorGetPoolInfoReply
 *
 * @brief Get pool information reply.
 *
 * @long This signal is sent under blocking flow control.
 *
 * @member MONITOR_GET_POOL_INFO_REPLY
 * @member MonitorGetPoolInfoReply
 *
 * @seealso MonitorGetPoolInfoRequest
 *
 * @example
 * uint32_t i;
 * MonitorGetPoolInfoReply *reply;
 *
 * // Get info... //
 *
 * MonitorPoolFragmentInfo
 *    *frags = (void *)&reply->data[reply->fragmentsOffset];
 * MonitorPoolBufferSizeInfo
 *    *stacks = (void *)&reply->data[reply->stackSizesOffset];
 * MonitorPoolBufferSizeInfo
 *    *signals = (void *)&reply->data[reply->signalSizesOffset];
 *
 * for (i = 0; i < reply->fragmentsCount; i++)
 * {
 *    printf("Fragment: %lx Size: %ld Used %ld (stacks) %ld (signals)\n",
 *           frags[i].address, frags[i].size, frags[i].stacks, frags[i].signals)
 * }
 * for (i = 0; i < reply->stackSizesCount; i++)
 * {
 *    printf("Stack size: %lx Allocated: %ld Free: %ld\n",
 *           stacks[i].size, stacks[i].allocated, stacks[i].free);
 * }
 * for (i = 0; i < reply->signalSizesCount; i++)
 * {
 *    printf("Signal size: %lx Allocated: %ld Free: %ld\n",
 *           signals[i].size, signals[i].allocated, signals[i].free);
 * }
 */

/** @macro MONITOR_GET_POOL_INFO_REPLY */
#define MONITOR_GET_POOL_INFO_REPLY 39051 /* !-SIGNO(MonitorGetPoolInfoReply)-! */

/**
 * @type MonitorPoolFragmentInfo
 * @brief Pool fragment information type.
 *
 * @seealso MonitorGetPoolSignalsReply
 *
 * @field address   Base address of pool fragment.
 * @field size      Size of pool fragment.
 * @field stacks    Bytes used in the area for stacks in pool fragment.
 * @field signals   Bytes used in the area for signals in pool fragment.
 * @field reserved0  NOT USED.
 * @field reserved1  NOT USED.
 */
typedef struct MonitorPoolFragmentInfo
{
   uint32_t address;
   uint32_t size;
   uint32_t stacks;
   uint32_t signals;
   uint32_t reserved0;
   uint32_t reserved1;
} MonitorPoolFragmentInfo;

/**
 * @type MonitorPoolBufferSizeInfo
 * @brief Pool buffer size info type.
 *
 * @seealso MonitorGetPoolSignalsReply
 *
 * @field size       A configured buffer size in bytes.
 * @field allocated  Number of allocated buffers of this size.
 * @field free       Number of free buffers of this size (NOT USED).
 * @field reserved0  NOT USED.
 */
typedef struct MonitorPoolBufferSizeInfo
{
   uint32_t size;
   uint32_t allocated;
   uint32_t free;
   uint32_t reserved0;
} MonitorPoolBufferSizeInfo;

/**
 * @type MonitorGetPoolInfoReply
 *
 * @output sigNo             MONITOR_GET_POOL_INFO_REPLY
 * @output pid               Pool ID.
 * @output sid               Segment ID or 0.
 * @output total             Total size in bytes of the pool.
 * @output free              Number of bytes available in the pool (does not
 *                           include free buffers).
 * @output fragmentsOffset   Offset into data of first @link "" "MonitorPoolFragmentInfo".
 * @output fragmentsCount    Number of MonitorPoolFragmentInfo:s.
 * @output stackSizesOffset  Offset into data of first @link "" "MonitorPoolBufferSizeInfo"
 *                           for configured stack buffer sizes.
 * @output stackSizesCount   Number of @link "" "MonitorPoolBufferSizeInfo":s for stack
 *                           buffer sizes.
 * @output signalSizesOffset Offset into data of first @link "" "MonitorPoolBufferSizeInfo"
 *                           for configured stack buffer sizes.
 * @output signalSizesCount  Number of @link "" "MonitorPoolBufferSizeInfo":s for signal
 *                           buffer sizes.
 * @output stackAdmSize
 *                           Difference between actual size of a stack buffer in
 *                           the pool and the corresponding configured buffer
 *                           size. For OSE 4/5 this is 0.
 * @output stackInternalAdmSize
 *                           Difference between the size of a stack and the size
 *                           of the area that should be used by the application.
 *                           For OSE 4/5 this is 1 (endmark).
 * @output signalAdmSize
 *                           Difference between actual size of a signal buffer
 *                           in the pool and the corresponding configured buffer
 *                           size. For OSE 4/5 this is 33 (sigadm + endmark).
 * @output signalInternalAdmSize
 *                           Difference between the size of a signal and the size
 *                           of the area that should be used by the application.
 *                           For OSE 4/5 this is 0.
 * @output stackAlignment
 *                           The alignment size of stacks in this pool. For all
 *                           signal sizes the configured size plus the administration
 *                           size must be a multiple of this size.
 * @output signalAlignment
 *                           The alignment size of signals in this pool. For all
 *                           signal sizes the configured size plus the administration
 *                           size must be a multiple of this size.
 * @output reserved0  NOT USED.
 * @output reserved1  NOT USED.
 * @output reserved2  NOT USED.
 *
 * @output data              Data
 */

typedef struct MonitorGetPoolInfoReply
{
   uint32_t sigNo;
   uint32_t pid;
   uint32_t sid;
   uint32_t total;
   uint32_t free;
   uint32_t fragmentsOffset;
   uint32_t fragmentsCount;
   uint32_t stackSizesOffset;
   uint32_t stackSizesCount;
   uint32_t signalSizesOffset;
   uint32_t signalSizesCount;
   uint8_t stackAlignment;
   uint8_t stackAdmSize;
   uint8_t stackInternalAdmSize;
   uint8_t signalAlignment;
   uint8_t signalAdmSize;
   uint8_t signalInternalAdmSize;
   uint8_t reserved0;
   uint8_t reserved1;
   uint32_t reserved2;
   uint32_t data[1];
} MonitorGetPoolInfoReply;

/**
 * @signal MonitorGetPoolInfoEndmark
 *
 * @brief Get pool information endmark, with status.
 *
 * @seealso MonitorGetPoolInfoReply
 *
 * @member MONITOR_GET_POOL_INFO_ENDMARK
 * @member MonitorGetInfoEndmark
 */

/** @macro MONITOR_GET_POOL_INFO_ENDMARK */
#define MONITOR_GET_POOL_INFO_ENDMARK 39052 /* !-SIGNO(MonitorGetInfoEndmark)-! */

/**
 * @signal MonitorGetPoolSignalsRequest
 *
 * @brief Get pool signals request.
 *
 * @long Followed by any number of MonitorGetPoolSignalsReply,
 * MonitorBlockingFlowControlRequest, and one
 * MonitorGetPoolSignalsEndmark.
 *
 * @long Accepts all MonitorSignalFilterTagValues except
 * MONITOR_SIGNAL_FILTER_TAG_OWNER_BID and MONITOR_SIGNAL_FILTER_TAG_OWNER_SID.
 *
 * @long The "level" field is not used and should be set to 0.
 *
 * @restriction Accepts only scope type MONITOR_SCOPE_ID (indicating a pool).
 *
 * @seealso MonitorScopeTagValue
 * @seealso MonitorGetPoolSignalsReply
 *
 * @member MONITOR_GET_POOL_SIGNALS_REQUEST
 * @member MonitorGetInfoRequest
 */

/** @macro MONITOR_GET_POOL_SIGNALS_REQUEST */
#define MONITOR_GET_POOL_SIGNALS_REQUEST 39053 /* !-SIGNO(MonitorGetInfoRequest)-! */

/**
 * @signal MonitorGetPoolSignalsReply
 *
 * @brief Get pool signals reply.
 *
 * @long This signal is sent under blocking flow control.
 *
 * @seealso MonitorGetPoolSignalsRequest
 *
 * @member MONITOR_GET_POOL_SIGNALS_REPLY
 * @member MonitorGetPoolSignalsReply
 */

/** @macro MONITOR_GET_POOL_SIGNALS_REPLY */
#define MONITOR_GET_POOL_SIGNALS_REPLY 39054 /* !-SIGNO(MonitorGetPoolSignalsReply)-! */

/**
 * @type MonitorGetPoolSignalsReply
 *
 * @output sigNo         MONITOR_GET_POOL_SIGNALS_REPLY
 * @output pid           ID of the pool that the info is for.
 * @output signalsCount  Number of signal infos.
 * @output signals       Array of signal infos.
 * @output reserved0  NOT USED
 */
typedef struct MonitorGetPoolSignalsReply
{
   uint32_t sigNo;
   uint32_t pid;
   uint32_t reserved0;
   uint32_t signalsCount;
   MonitorSignalInfo signals[1];
} MonitorGetPoolSignalsReply;

/**
 * @signal MonitorGetPoolSignalsEndmark
 *
 * @brief Get pool signals endmark, with status.
 *
 * @seealso MonitorGetPoolSignalsReply
 *
 * @member MONITOR_GET_POOL_SIGNALS_ENDMARK
 * @member MonitorGetInfoEndmark
 */

/** @macro MONITOR_GET_POOL_SIGNALS_ENDMARK */
#define MONITOR_GET_POOL_SIGNALS_ENDMARK 39055 /* !-SIGNO(MonitorGetInfoEndmark)-! */

/*******************************************************************************
 *
 * HEAP INFO SIGNALS
 *
 ******************************************************************************/

/**
 * @signal MonitorGetHeapInfoRequest
 *
 * @brief Get heap information request.
 *
 * @long Followed by any number of MonitorGetHeapInfoReply,
 * MonitorBlockingFlowControl, and one MonitorGetHeapInfoEndmark.
 *
 * @long The "level" field is not used and should be set to 0.
 *
 * @restriction Accepts only scope types MONITOR_SCOPE_ID
 * (indicating a heap), MONITOR_SCOPE_SEGMENT_ID, and MONITOR_SCOPE_SYSTEM.
 *
 * @member MONITOR_GET_HEAP_INFO_REQUEST
 * @member MonitorGetInfoRequest
 *
 * @seealso MonitorScopeTagValue
 * @seealso MonitorGetHeapInfoReply
 */

/** @macro MONITOR_GET_HEAP_INFO_REQUEST */
#define MONITOR_GET_HEAP_INFO_REQUEST 39140 /* !-SIGNO(MonitorGetInfoRequest)-! */

/**
 * @signal MonitorGetHeapInfoReply
 *
 * @brief Get heap information reply.
 *
 * @long This signal is sent under blocking flow control.
 *
 * @member MONITOR_GET_HEAP_INFO_REPLY
 * @member MonitorGetHeapInfoReply
 *
 * @seealso MonitorGetHeapInfoRequest
 */

/** @macro MONITOR_GET_HEAP_INFO_REPLY */
#define MONITOR_GET_HEAP_INFO_REPLY 39141 /* !-SIGNO(MonitorGetHeapInfoReply)-! */

/**
 * @type MonitorHeapBufferSizeInfo
 *
 * @brief Heap buffer size info type.
 *
 * @field size       An available heap buffer size in bytes.
 * @field allocated  Number of allocated heap buffers of this size.
 * @field free       Number of free heap buffers of this size.
 * @field reserved0  NOT USED.
 *
 * @seealso MonitorGetHeapInfoReply
 */
typedef struct MonitorHeapBufferSizeInfo
{
   uint32_t size;
   uint32_t allocated;
   uint32_t free;
   uint32_t reserved0;
} MonitorHeapBufferSizeInfo;

/**
 * @type MonitorGetHeapInfoReply
 *
 * @output sigNo             MONITOR_GET_HEAP_INFO_REPLY
 * @output id                Heap ID.
 * @output owner             Heap owner ID.
 * @output sid               Segment ID or 0 if not available.
 * @output size              Total heap size in bytes.
 * @output usedSize          Total actually used size in bytes.
 * @output peakUsedSize      High watermark of the above counter or
 *                           0 if not available.
 * @output reqSize           Total requested size in bytes.
 * @output largestFree       Size of largest free heap buffer.
 * @output largeThreshold    Heap buffers larger than this value are
 *                           allocated from another memory manager.
 *                           0 if not applicable.
 * @output priv              Number of private heap buffers.
 * @output shared            Number of shared heap buffers.
 * @output mallocFailed      Number of failed malloc() attempts.
 * @output bufferSizesCount  Number of heap buffer size infos.
 * @output reserved0         NOT USED.
 * @output reserved1         NOT USED.
 * @output reserved2         NOT USED.
 * @output reserved3         NOT USED.
 * @output bufferSizes       Array of heap buffer size infos.
 */
typedef struct MonitorGetHeapInfoReply
{
   uint32_t sigNo;
   uint32_t id;
   uint32_t owner;
   uint32_t sid;
   uint32_t size;
   uint32_t usedSize;
   uint32_t peakUsedSize;
   uint32_t reqSize;
   uint32_t largestFree;
   uint32_t largeThreshold;
   uint32_t priv;
   uint32_t shared;
   uint32_t mallocFailed;
   uint32_t bufferSizesCount;
   uint32_t reserved0;
   uint32_t reserved1;
   uint32_t reserved2;
   uint32_t reserved3;
   MonitorHeapBufferSizeInfo bufferSizes[1];
} MonitorGetHeapInfoReply;

#ifndef __cplusplus
#define private priv
#endif

/**
 * @signal MonitorGetHeapInfoEndmark
 *
 * @brief Get heap information endmark, with status.
 *
 * @member MONITOR_GET_HEAP_INFO_ENDMARK
 * @member MonitorGetInfoEndmark
 *
 * @seealso MonitorGetHeapInfoReply
 */

/** @macro MONITOR_GET_HEAP_INFO_ENDMARK */
#define MONITOR_GET_HEAP_INFO_ENDMARK 39142 /* !-SIGNO(MonitorGetInfoEndmark)-! */

/**
 * @signal MonitorGetHeapFragmentInfoRequest
 *
 * @brief Get heap fragment information request.
 *
 * @long Followed by any number of MonitorGetHeapFragmentInfoReply,
 * MonitorBlockingFlowControl, and one MonitorGetHeapFragmentInfoEndmark.
 *
 * @long The "level" field is not used and should be set to 0.
 *
 * @restriction Accepts only scope type MONITOR_SCOPE_ID (indicating a heap).
 *
 * @member MONITOR_GET_HEAP_FRAGMENT_INFO_REQUEST
 * @member MonitorGetInfoRequest
 *
 * @seealso MonitorScopeTagValue
 * @seealso MonitorGetHeapFragmentInfoReply
 */

/** @macro MONITOR_GET_HEAP_FRAGMENT_INFO_REQUEST */
#define MONITOR_GET_HEAP_FRAGMENT_INFO_REQUEST 39143 /* !-SIGNO(MonitorGetInfoRequest)-! */

/**
 * @signal MonitorGetHeapFragmentInfoReply
 *
 * @brief Get heap fragment information reply.
 *
 * @long This signal is sent under blocking flow control.
 *
 * @member MONITOR_GET_HEAP_FRAGMENT_INFO_REPLY
 * @member MonitorGetHeapFragmentInfoReply
 *
 * @seealso MonitorGetHeapFragmentInfoRequest
 */

/** @macro MONITOR_GET_HEAP_FRAGMENT_INFO_REPLY */
#define MONITOR_GET_HEAP_FRAGMENT_INFO_REPLY 39144 /* !-SIGNO(MonitorGetHeapFragmentInfoReply)-! */

/**
 * @type MonitorHeapFragmentInfo
 *
 * @brief Heap fragment information type.
 *
 * @field address    Heap fragment address.
 * @field size       Total heap fragment size in bytes.
 * @field usedSize   Total actually used size in bytes.
 * @field reqSize    Total requested size in bytes.
 * @field reserved0  NOT USED.
 * @field reserved1  NOT USED.
 *
 * @seealso MonitorGetHeapFragmentInfoReply
 */
typedef struct MonitorHeapFragmentInfo
{
   uint32_t address;
   uint32_t size;
   uint32_t usedSize;
   uint32_t reqSize;
   uint32_t reserved0;
   uint32_t reserved1;
} MonitorHeapFragmentInfo;

/**
 * @type MonitorGetHeapFragmentInfoReply
 *
 * @output sigNo           MONITOR_GET_HEAP_FRAGMENT_INFO_REPLY
 * @output id              ID of the heap that the info is for.
 * @output reserved0       NOT USED.
 * @output fragmentsCount  Number of heap fragment infos.
 * @output fragments       Array of heap fragment infos.
 */
typedef struct MonitorGetHeapFragmentInfoReply
{
   uint32_t sigNo;
   uint32_t id;
   uint32_t reserved0;
   uint32_t fragmentsCount;
   MonitorHeapFragmentInfo fragments[1];
} MonitorGetHeapFragmentInfoReply;

/**
 * @signal MonitorGetHeapFragmentInfoEndmark
 *
 * @brief Get heap fragment information endmark, with status.
 *
 * @member MONITOR_GET_HEAP_FRAGMENT_INFO_ENDMARK
 * @member MonitorGetInfoEndmark
 *
 * @seealso MonitorGetHeapFragmentInfoReply
 */

/** @macro MONITOR_GET_HEAP_FRAGMENT_INFO_ENDMARK */
#define MONITOR_GET_HEAP_FRAGMENT_INFO_ENDMARK 39145 /* !-SIGNO(MonitorGetInfoEndmark)-! */

/**
 * @signal MonitorGetHeapBuffersRequest
 *
 * @brief Get heap buffers request.
 *
 * @long Followed by any number of MonitorGetHeapBuffersReply,
 * MonitorBlockingFlowControlRequest, and one
 * MonitorGetHeapBuffersEndmark.
 *
 * @long Accepts all MonitorHeapBufferFilterTagValues.
 *
 * @long The "level" field is not used and should be set to 0.
 *
 * @restriction Accepts only scope type MONITOR_SCOPE_ID (indicating a heap).
 *
 * @member MONITOR_GET_HEAP_BUFFERS_REQUEST
 * @member MonitorGetInfoRequest
 *
 * @seealso MonitorScopeTagValue
 * @seealso MonitorHeapBufferFilterTagValue
 * @seealso MonitorGetHeapBuffersReply
 */

/** @macro MONITOR_GET_HEAP_BUFFERS_REQUEST */
#define MONITOR_GET_HEAP_BUFFERS_REQUEST 39146 /* !-SIGNO(MonitorGetInfoRequest)-! */

/**
 * @signal MonitorGetHeapBuffersReply
 *
 * @brief Get heap buffers reply.
 *
 * @long This signal is sent under blocking flow control.
 *
 * @member MONITOR_GET_HEAP_BUFFERS_REPLY
 * @member MonitorGetHeapBuffersReply
 *
 * @seealso MonitorGetHeapBuffersRequest
 */

/** @macro MONITOR_GET_HEAP_BUFFERS_REPLY */
#define MONITOR_GET_HEAP_BUFFERS_REPLY 39147 /* !-SIGNO(MonitorGetHeapBuffersReply)-! */

/**
 * @type MonitorHeapBufferStatus
 * @brief Heap buffer administrative status type.
 * @seealso MonitorHeapBufferStatusValue
 */
typedef uint32_t MonitorHeapBufferStatus;

/**
 * @type MonitorHeapBufferStatusValue
 *
 * @brief Enumeration of heap buffer administrative status type values.
 *
 * @long These are the values of the field status in
 * a @link "" "MonitorHeapBufferInfo" struct.
 *
 * @field MONITOR_HEAP_BUFFER_STATUS_UNKNOWN  Unknown.
 * @field MONITOR_HEAP_BUFFER_STATUS_VALID    Valid.
 * @field MONITOR_HEAP_BUFFER_STATUS_ENDMARK  Broken administrative
 *                                            endmark in memory.
 */
enum MonitorHeapBufferStatusValue
{
   MONITOR_HEAP_BUFFER_STATUS_UNKNOWN = 0,
   MONITOR_HEAP_BUFFER_STATUS_VALID,
   MONITOR_HEAP_BUFFER_STATUS_ENDMARK
};

/**
 * @type MonitorHeapBufferInfo
 *
 * @brief Heap buffer administrative information type.
 *
 * @field owner          Owner process ID if a private heap buffer.
 * @field shared         True if the heap buffer is shared, false if it is private.
 * @field address        Heap buffer address.
 * @field size           Actual heap buffer size in bytes (overhead and internal
 *                       fragmentation).
 * @field reqSize        Requested heap buffer size in bytes.
 * @field fileNameCount  Length of the name of the file where the allocation
 *                       took place or 0 if not available. Length is expressed
 *                       in units of MonitorHeapBufferInfo structs.
 * @field lineNumber     Line number in the file where the allocation took
 *                       place or 0 if not available.
 * @field status         Heap buffer administrative status, see
 *                       @link "" "MonitorHeapBufferStatusValue".
 * @field reserved0      NOT USED.
 * @field reserved1      NOT USED.
 */
typedef struct MonitorHeapBufferInfo
{
   uint32_t owner;
   MonitorBoolean shared;
   uint32_t address;
   uint32_t size;
   uint32_t reqSize;
   uint32_t fileNameCount;
   uint32_t lineNumber;
   MonitorHeapBufferStatus status;
   uint32_t reserved0;
   uint32_t reserved1;
} MonitorHeapBufferInfo;

/**
 * @type MonitorHeapBufferInfoHolder
 *
 * @brief Heap buffer administrative information type holder.
 *
 * @long Contains either a MonitorHeapBufferInfo struct or a null-terminated
 * file name (possibly followed by padding characters) or a part thereof. The
 * total length of the file name is specified in the fileNameCount field in
 * the preceding MonitorHeapBufferInfo struct and is expressed in units of
 * MonitorHeapBufferInfo structs.
 *
 * @field info  Heap buffer info.
 * @field str   Null-terminated file name or a part thereof.
 *
 * @seealso MonitorHeapBufferInfo
 */
typedef union MonitorHeapBufferInfoHolder
{
   MonitorHeapBufferInfo info;
   char str[1];
} MonitorHeapBufferInfoHolder;

/**
 * @type MonitorGetHeapBuffersReply
 *
 * @long The buffers field contains MonitorHeapBufferInfo structs possibly
 * followed by null-terminated file names (possibly followed by padding
 * characters) whose length is specified in the fileNameCount field in the
 * preceding MonitorHeapBufferInfo struct and is expressed in units of
 * MonitorHeapBufferInfo structs.
 *
 * @output sigNo         MONITOR_GET_HEAP_BUFFERS_REPLY
 * @output id            ID of the heap that the info is for.
 * @output reserved0     NOT USED.
 * @output buffersCount  Number of heap buffer info holders.
 * @output buffers       Array of heap buffer info holders.
 *
 */
typedef struct MonitorGetHeapBuffersReply
{
   uint32_t sigNo;
   uint32_t id;
   uint32_t reserved0;
   uint32_t buffersCount;
   MonitorHeapBufferInfoHolder buffers[1];
} MonitorGetHeapBuffersReply;

/**
 * @signal MonitorGetHeapBuffersEndmark
 *
 * @brief Get heap buffers endmark, with status.
 *
 * @member MONITOR_GET_HEAP_BUFFERS_ENDMARK
 * @member MonitorGetInfoEndmark
 *
 * @seealso MonitorGetHeapBuffersReply
 */

/** @macro MONITOR_GET_HEAP_BUFFERS_ENDMARK */
#define MONITOR_GET_HEAP_BUFFERS_ENDMARK 39148 /* !-SIGNO(MonitorGetInfoEndmark)-! */

/*******************************************************************************
 *
 * SYSTEM PARAMETER SIGNALS
 *
 ******************************************************************************/

/**
 * @signal MonitorGetSysParamRequest
 *
 * @brief Get system parameters request.
 *
 * @long Followed by any number of MonitorGetSysParamReply,
 * MonitorBlockFlowControlRequest and one
 * MonitorGetSysParamEndmark.
 *
 * @long The "level" field is not used and should be set to 0.
 *
 * @restriction Accepts only scope type MONITOR_SCOPE_SYSTEM
 *
 * @seealso MonitorGetSysParamReply
 *
 * @member MONITOR_GET_SYS_PARAM_REQUEST
 * @member MonitorGetInfoRequest
 */

/** @macro MONITOR_GET_SYS_PARAM_REQUEST */
#define MONITOR_GET_SYS_PARAM_REQUEST 39060 /* !-SIGNO(MonitorGetInfoRequest)-! */

/**
 * @signal MonitorGetSysParamReply
 *
 * @brief Get system parameters reply - one reply per parameter.
 *
 * @long This signal is sent under blocking flow control.
 *
 * @seealso MonitorGetSysParamRequest
 *
 * @member MONITOR_GET_SYS_PARAM_REPLY
 * @member MonitorGetSysParamReply
 */

/** @macro MONITOR_GET_SYS_PARAM_REPLY */
#define MONITOR_GET_SYS_PARAM_REPLY 39061 /* !-SIGNO(MonitorGetSysParamReply)-! */

/**
 * @type MonitorGetSysParamReply
 *
 * @output sigNo       MONITOR_GET_SYS_PARAM_REPLY
 * @output nameOffset  Offset in data to null terminated system parameter name.
 * @output valueOffset Offset in data to null terminated system parameter value.
 * @output data        Variable sized buffer.
 * @field reserved0    NOT USED.
 */
typedef struct MonitorGetSysParamReply
{
   uint32_t sigNo;
   uint32_t reserved0;
   uint32_t nameOffset;
   uint32_t valueOffset;
   char data[1];
} MonitorGetSysParamReply;

/**
 * @signal MonitorGetSysParamEndmark
 *
 * @brief Get system parameters endmark, with status.
 *
 * @seealso MonitorGetSysParamReply
 *
 * @member MONITOR_GET_SYS_PARAM_ENDMARK
 * @member MonitorGetInfoEndmark
 */

/** @macro MONITOR_GET_SYS_PARAM_ENDMARK */
#define MONITOR_GET_SYS_PARAM_ENDMARK 39062 /* !-SIGNO(MonitorGetInfoEndmark)-! */

/*******************************************************************************
 *
 * MEMORY SIGNALS
 *
 ******************************************************************************/

/**
 * @signal MonitorGetMemoryRequest
 *
 * @brief Get memory request.
 *
 * @long Followed by zero or more MonitorGetMemoryReply,
 * MonitorBlockingFlowControlRequest and one MonitorGetMemoryEndmark.
 *
 * @long If all the requested data can not be read, the monitor will attempt
 * to read a smaller amount of data from the specified address. This data will
 * be sent to the client and an error (MONITOR_STATUS_GET_MEMORY_ERROR)
 * will be reported in the
 * @link "" "MonitorStatus" status in the
 * @link  "" "MonitorGetMemoryEndmark" signal.
 *
 * @member MONITOR_GET_MEMORY_REQUEST
 * @member MonitorGetMemoryRequest
 * @seealso MonitorSetMemoryRequest
 * @seealso MonitorGetMemoryReply
 */

/** @macro MONITOR_GET_MEMORY_REQUEST */
#define MONITOR_GET_MEMORY_REQUEST 39100 /* !-SIGNO(MonitorGetMemoryRequest)-! */

/**
 * @type MonitorGetMemoryRequest
 *
 * @input sigNo    MONITOR_GET_MEMORY_REQUEST
 * @input pid      Address is in the context of this process, block or segment.
 * @input address  Memory address to start reading from.
 * @input size     Number of bytes to attempt to read.
 */
typedef struct MonitorGetMemoryRequest
{
   uint32_t sigNo;
   uint32_t pid;
   uint32_t address;
   uint32_t size;
} MonitorGetMemoryRequest;

/**
 * @signal MonitorGetMemoryReply
 *
 * @brief Get memory reply.
 *
 * @long This signal is sent under blocking flow control.
 *
 * @seealso MonitorGetMemoryRequest
 *
 * @member MONITOR_GET_MEMORY_REPLY
 * @member MonitorGetMemoryReply
 * @seealso MonitorGetMemoryRequest
 */

/** @macro MONITOR_GET_MEMORY_REPLY */
#define MONITOR_GET_MEMORY_REPLY 39101 /* !-SIGNO(MonitorGetMemoryReply)-! */

/**
 * @type MonitorGetMemoryReply
 *
 * @output sigNo     MONITOR_GET_MEMORY_REPLY
 * @output pid       Address is in the context of this process, block or segment.
 * @output address   Memory address of first byte in data.
 * @output dataCount Number of bytes in data.
 * @output data      The data.
 */
typedef struct MonitorGetMemoryReply
{
   uint32_t sigNo;
   uint32_t pid;
   uint32_t address;
   uint32_t dataCount;
   uint8_t data[1];
} MonitorGetMemoryReply;

/**
 * @signal MonitorGetMemoryEndmark
 *
 * @brief Get memory endmark, with status.
 *
 * @seealso MonitorGetMemoryReply
 *
 * @member MONITOR_GET_MEMORY_ENDMARK
 * @member MonitorGetMemoryEndmark
 */

/** @macro MONITOR_GET_MEMORY_ENDMARK */
#define MONITOR_GET_MEMORY_ENDMARK 39102 /* !-SIGNO(MonitorGetMemoryEndmark)-! */

/**
 * @type MonitorGetMemoryEndmark
 *
 * @output sigNo    MONITOR_GET_MEMORY_ENDMARK
 * @output status   MONITOR_STATUS_OK or cause of error.
 */
typedef struct MonitorGetMemoryEndmark
{
   uint32_t sigNo;
   MonitorStatus status;
} MonitorGetMemoryEndmark;

/*******************************************************************************
 *
 * CLIENT INFO SIGNALS
 *
 ******************************************************************************/

/**
 * @signal MonitorGetConnectedClientsRequest
 *
 * @brief Request information about the clients that are connected to the monitor.
 *
 * @long This signal is sent under blocking flow control.
 *
 * @member MONITOR_GET_CONNECTED_CLIENTS_REQUEST
 * @member MonitorGetConnectedClientsRequest
 */

/** @macro MONITOR_GET_CONNECTED_CLIENTS_REQUEST   */
#define MONITOR_GET_CONNECTED_CLIENTS_REQUEST 40820 /* !-SIGNO(MonitorGetConnectedClientsRequest)-! */

/**
 * @type MonitorGetConnectedClientsRequest
 *
 * @field sigNo      MONITOR_GET_CONNECTED_CLIENTS_REQUEST
 * @field reserved0  NOT USED. Must be 0.
 * @field reserved1  NOT USED. Must be 0.
 */

typedef struct MonitorGetConnectedClientsRequest
{
   uint32_t sigNo;
   uint32_t reserved0;
   uint32_t reserved1;
} MonitorGetConnectedClientsRequest;

/**
 * @signal MonitorGetConnectedClientsReply
 *
 * @brief  List information for a connected client.
 *
 * @member MONITOR_GET_CONNECTED_CLIENTS_REPLY
 * @member MonitorGetConnectedClientsReply
 */

/** @macro MONITOR_GET_CONNECTED_CLIENTS_REPLY   */
#define MONITOR_GET_CONNECTED_CLIENTS_REPLY 40821 /* !-SIGNO(MonitorGetConnectedClientsReply)-! */

/**
 * @type MonitorGetConnectedClientsReply
 *
 * @field  sigNo                    MONITOR_GET_CONNECTED_CLIENTS_REPLY
 * @output clientPid                Process id of the client process.
 * @output offspringPid             Process id of the offspring process.
 * @field  reserved0                NOT USED.
 * @output scope                    Scope.
 * @output interceptScope           Intercept scope.
 * @output nbrOfBreakpoints         Number of breakpoints in the scope.
 * @output nbrOfTracepoints         Number of tracepoints in the scope.
 * @output nbrOfTracepointActions   Number of tracepoint actions in the scope.
 * @output nbrOfTracepointVariables Number of tracepoint variables in the scope.
 * @output tracepointBufferSize     Size of the tracepoint buffer.
 * @output tracepointBufferType     Specifies if the tracepoint buffer is
 *                                  linear or circular.
 * @field  reserved1                NOT USED.
 * @field  reserved2                NOT USED.
 * @field  reserved3                NOT USED.
 * @output nameLength               Length of the client name
 * @output clientName               Name of the Client.
 */

typedef struct MonitorGetConnectedClientsReply
{
   uint32_t sigNo;
   uint32_t clientPid;
   uint32_t offspringPid;
   uint32_t reserved0;
   MonitorScope scope;
   MonitorScope interceptScope;
   uint32_t nbrOfBreakpoints;
   uint32_t nbrOfTracepoints;
   uint32_t nbrOfTracepointActions;
   uint32_t nbrOfTracepointVariables;
   uint64_t tracepointBufferSize;
   uint32_t tracepointBufferType;
   uint32_t reserved1;
   uint32_t reserved2;
   uint32_t reserved3;
   uint32_t nameLength;
   uint8_t clientName[1];
} MonitorGetConnectedClientsReply;

/** @macro MONITOR_GET_CONNECTED_CLIENTS_ENDMARK */
#define MONITOR_GET_CONNECTED_CLIENTS_ENDMARK 40822 /* !-SIGNO(MonitorGetConnectedClientsEndmark)-! */

/**
 * @type MonitorGetConnectedClientsEndmark
 *
 * @output sigNo  MONITOR_GET_CONNECTED_CLIENTS_ENDMARK
 * @output status MONITOR_STATUS_OK or cause of failure.
 */
typedef struct MonitorGetConnectedClientsEndmark
{
   uint32_t sigNo;
   MonitorStatus status;
} MonitorGetConnectedClientsEndmark;

#ifdef __cplusplus
} /* end of extern "C" */
#endif

#endif /* _MONITOR_SIG */
