/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2011 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

/**
 * @toc OSE:
 * @file ose_spi/monitor_actions.sig
 *
 * @brief OSE System direct action signal interface
 *        to control OSE via OSE run-time monitors.
 *
 * @long This header file defines the signal interface for action requests
 * between debuggers, system supervisors, et. al. and OSE monitors
 * (e.g. FMM and RMM).
 * Typical operations are to create/start/stop/kill processes,
 * send signal or fast semaphores, set system parameters or environment
 * variables
 *
 * @long A connection to the RMM is needed, see monitor.sig for the actual
 * connection procedure using @link "" "MonitorInterfaceRequest" and
 * @link "" "MonitorConnectRequest".
 *
 * @long It is not necessary to attach a scope before using the actions in
 * this header file. Not either is any attached main or intercept scopes
 * affecting the action scope given in these action requests.
 *
 * @long When ready, do disconnect with
 * @link "" "MonitorDisconnectRequest", allowing the monitor to
 * deallocate resources linked to the client.
 *
 * @long Compare the monitor.sig, which is the basic interface, and
 * where a description exists of the sequence to follow at access.
 *
 *
 * @long See also Run-Mode Monitor (RMM) in the "OSE Core User's Guide".
 *
 * @seealso ose_spi/monitor.sig
 * @seealso ose_rmm_set_req_enabled
 */

#ifndef _MONITOR_ACTIONS_SIG
#define _MONITOR_ACTIONS_SIG

#include "stdint.h"
#include "monitor.sig"

#ifdef __cplusplus
extern "C" {
#endif

/*******************************************************************************
 *
 * DEFINITIONS
 *
 ******************************************************************************/

/**
 * @type MonitorActionRequest
 *
 * @brief Request signal struct for action requests.
 *
 * @input sigNo        Signal number.
 * @input scope        The scope of processes or other entities to examine.
 * @input tagsCount    The length of the tags list, as the
 *                     total number of sizeof(MonitorTag)
 *                     entries, with tags and arguments.
 * @input tags         The filter conditions.
 */
typedef struct MonitorActionRequest
{
   uint32_t sigNo;
   MonitorScope scope;
   uint32_t tagsCount;
   MonitorTag tags[1];
} MonitorActionRequest;

/**
 * @type MonitorActionReply
 *
 * @brief Reply signal struct for action requests.
 *
 * @output sigNo Signal number.
 * @output status MONITOR_STATUS_OK or reason of failure.
 */
typedef struct MonitorActionReply
{
   uint32_t sigNo;
   MonitorStatus status;
} MonitorActionReply;

/*******************************************************************************
 *
 * PROCESS MANIPULATION SIGNALS
 *
 ******************************************************************************/

/**
 * @type MonitorCreateProcessTagValue
 *
 * @brief Tag values to use in @link "" "MonitorCreateProcessRequest".
 *
 * @field MONITOR_CREATE_PROCESS_TAG_NAME        Name (str)
 *        The tag argument value is a null terminated string
 *        with the name of a process, occupying one or more
 *        sizeof(MonitorTag) entries in the tag list.
 *
 * @field MONITOR_CREATE_PROCESS_TAG_REDIRECTION Redirection table (X x 2 * num)
 *        The tag argument is a list of table entries (pair of numeric values),
 *        where one field in the first entry contains the number of entries (X),
 *        as described in @link "" "MonitorRedirection".
 *
 * @field MONITOR_CREATE_PROCESS_TAG_END         Highest possible tag value
 *                                               plus one, should not be used.
 *
 * @seealso MonitorTag
 * @seealso OS_redir_entry
 * @seealso MonitorCreateProcessRequest
 */
enum MonitorCreateProcessTagValue
{
   MONITOR_CREATE_PROCESS_TAG_NAME = 3000,
   MONITOR_CREATE_PROCESS_TAG_REDIRECTION,
   MONITOR_CREATE_PROCESS_TAG_END
};

/**
 * @signal MonitorCreateProcessRequest
 *
 * @brief Create process request.
 *
 * @long A new OSE process is created.
 * Compare the description of the OSE
 * function @link "ose.h" "create_process".
 *
 * @long Requires the tag MONITOR_CREATE_PROCESS_TAG_NAME and accepts
 * MONITOR_CREATE_PROCESS_TAG_REDIRECTION
 * - see @link "" "MonitorCreateProcessTagValue".
 *
 * @member MONITOR_CREATE_PROCESS_REQUEST
 * @member MonitorCreateProcessRequest
 *
 * @seealso MonitorCreateProcessReply
 */

/** @macro MONITOR_CREATE_PROCESS_REQUEST */
#define MONITOR_CREATE_PROCESS_REQUEST 39075 /* !-SIGNO(MonitorCreateProcessRequest)-! */

/**
 * @type MonitorCreateProcessRequest
 *
 * @input sigNo             MONITOR_CREATE_PROCESS_REQUEST
 * @input type              Process type.
 * @input entrypoint        Process entrypoint or 0.
 * @input stackSize         Process stacksize or 0.
 * @input priority          Process priority or 0.
 * @input timeSlice         Process timeslice or 0.
 * @input bid               Block in which to create process.
 * @input vector            Process interrupt vector or 0.
 * @input uid               Process user number.
 * @input reserved0         NOT USED. Should be 0.
 * @input reserved1         NOT USED. Should be 0.
 * @input reserved2         NOT USED. Should be 0.
 * @input tagsCount         The length of the tags list, as the
 *                          total number of sizeof(MonitorTag)
 *                          entries, with tags and arguments.
 * @input tags              Process name string and optional
 *                          redirection table - see
 *                          @link "" "MonitorCreateProcessTagValue".
 */
typedef struct MonitorCreateProcessRequest
{
   uint32_t sigNo;
   MonitorProcessType type;
   uint32_t entrypoint;
   uint32_t stackSize;
   uint32_t priority;
   uint32_t timeSlice;
   uint32_t bid;
   uint32_t vector;
   uint32_t uid;
   uint32_t reserved0;
   uint32_t reserved1;
   uint32_t reserved2;
   uint32_t tagsCount;
   MonitorTag tags[1];
} MonitorCreateProcessRequest;

/**
 * @signal MonitorCreateProcessReply
 *
 * @brief Create process reply.
 *
 * @member MONITOR_CREATE_PROCESS_REPLY
 * @member MonitorCreateProcessReply
 *
 * @seealso MonitorCreateProcessRequest
 */

/** @macro MONITOR_CREATE_PROCESS_REPLY */
#define MONITOR_CREATE_PROCESS_REPLY 39076 /* !-SIGNO(MonitorCreateProcessReply)-! */

/**
 * @type MonitorCreateProcessReply
 *
 * @output sigNo     MONITOR_CREATE_PROCESS_REPLY
 * @output status    MONITOR_STATUS_OK or cause of failure.
 * @output pid       ID of the created process.
 * @output bid       ID of the block of the created process.
 * @output sid       ID of the segment of the created process.
 */
typedef struct MonitorCreateProcessReply
{
   uint32_t sigNo;
   MonitorStatus status;
   uint32_t pid;
   uint32_t bid;
   uint32_t sid;
} MonitorCreateProcessReply;

/**
 * @signal MonitorKillScopeRequest
 *
 * @brief Kill all processes in scope request.
 *
 * @long The monitor tries to kill all processes in the scope;
 * no action is taken for processes that can not be killed.
 *
 * @restriction Accepts only scope types MONITOR_SCOPE_ID
 * (process or block ID)
 * and MONITOR_SCOPE_BLOCK_ID (block ID).
 *
 *
 * @member MONITOR_KILL_SCOPE_REQUEST
 * @member MonitorActionRequest
 *
 * @seealso MonitorScopeTypeValue
 * @seealso MonitorKillScopeReply
 */

/** @macro MONITOR_KILL_SCOPE_REQUEST */
#define MONITOR_KILL_SCOPE_REQUEST 39077 /* !-SIGNO(MonitorActionRequest)-! */

/**
 * @signal MonitorKillScopeReply
 *
 * @brief Kill all processes in scope reply.
 *
 * @member MONITOR_KILL_SCOPE_REPLY
 * @member MonitorActionReply
 *
 * @seealso MonitorKillScopeRequest
 */

/** @macro MONITOR_KILL_SCOPE_REPLY */
#define MONITOR_KILL_SCOPE_REPLY 39078 /* !-SIGNO(MonitorActionReply)-! */

/**
 * @signal MonitorStartScopeRequest
 *
 * @brief Start all processes in scope request.
 *
 * @long The monitor tries to start all processes in the scope;
 * no action is taken for processes that can not be started.
 *
 * @restriction Accepts only scope types MONITOR_SCOPE_ID
 * and MONITOR_SCOPE_BLOCK_ID.
 *
 * @member MONITOR_START_SCOPE_REQUEST
 * @member MonitorActionRequest
 *
 * @seealso MonitorScopeTypeValue
 * @seealso MonitorStartScopeReply
 */

/** @macro MONITOR_START_SCOPE_REQUEST */
#define MONITOR_START_SCOPE_REQUEST 39079 /* !-SIGNO(MonitorActionRequest)-! */

/**
 * @signal MonitorStartScopeReply
 *
 * @brief Start all processes in scope reply.
 *
 * @member MONITOR_START_SCOPE_REPLY
 * @member MonitorActionReply
 *
 * @seealso MonitorStartScopeRequest
 */

/** @macro MONITOR_START_SCOPE_REPLY */
#define MONITOR_START_SCOPE_REPLY 39080 /* !-SIGNO(MonitorActionReply)-! */

/**
 * @signal MonitorStopScopeRequest
 *
 * @brief Stop all processes in scope request.
 *
 * @long The monitor tries to stop all processes in the scope;
 * no action is taken for processes that can not be stopped.
 *
 * @restriction Accepts only scope types MONITOR_SCOPE_ID
 * and MONITOR_SCOPE_BLOCK_ID.
 *
 * @member MONITOR_STOP_SCOPE_REQUEST
 * @member MonitorActionRequest
 *
 * @seealso MonitorScopeTypeValue
 * @seealso MonitorStopScopeReply
 */

/** @macro MONITOR_STOP_SCOPE_REQUEST */
#define MONITOR_STOP_SCOPE_REQUEST 39081 /* !-SIGNO(MonitorActionRequest)-! */

/**
 * @signal MonitorStopScopeReply
 *
 * @brief Stop all processes in scope reply.
 *
 * @member MONITOR_STOP_SCOPE_REPLY
 * @member MonitorActionReply
 *
 * @seealso MonitorStopScopeRequest
 */

/** @macro MONITOR_STOP_SCOPE_REPLY */
#define MONITOR_STOP_SCOPE_REPLY 39082 /* !-SIGNO(MonitorActionReply)-! */

/**
 * @signal MonitorSetEnvVarRequest
 *
 * @brief Set environment variable for a process/block request.
 *
 * @long If valueOffset is ~0, the variable is removed.
 *
 * @long If valueOffset is 0, the variable value is in the
 *       numeric field value and the environment variable will be
 *       set with @link "" "set_envp"(value).
 *
 * @member MONITOR_SET_ENV_VAR_REQUEST
 * @member MonitorSetEnvVarRequest
 *
 * @seealso MonitorSetEnvVarReply
 */

/** @macro MONITOR_SET_ENV_VAR_REQUEST */
#define MONITOR_SET_ENV_VAR_REQUEST 39083 /* !-SIGNO(MonitorSetEnvVarRequest)-! */

/**
 * @type MonitorSetEnvVarRequest
 *
 * @input sigNo       MONITOR_SET_ENV_VAR_REQUEST
 * @input pid         ID of the process/block to set the environment variable on.
 * @input value       Value to set with set_envp().
 * @input nameOffset  Offset in data to null terminated environment
 *                    variable name.
 * @input valueOffset Offset in data to null terminated environment
 *                    variable value (or 0 or ~0 as described above).
 * @input data        Variable sized buffer.
 */
typedef struct MonitorSetEnvVarRequest
{
   uint32_t sigNo;
   uint32_t pid;
   uint32_t value;
   uint32_t nameOffset;
   uint32_t valueOffset;
   char data[1];
} MonitorSetEnvVarRequest;

/**
 * @signal MonitorSetEnvVarReply
 *
 * @brief Set environment variable reply.
 *
 * @member MONITOR_SET_ENV_VAR_REPLY
 * @member MonitorActionReply
 *
 * @seealso MonitorSetEnvVarRequest
 */

/** @macro MONITOR_SET_ENV_VAR_REPLY */
#define MONITOR_SET_ENV_VAR_REPLY 39084 /* !-SIGNO(MonitorActionReply)-! */

/**
 * @signal MonitorSendSignalRequest
 *
 * @brief Send signal to a process request.
 *
 * @member MONITOR_SEND_SIGNAL_REQUEST
 * @member MonitorSendSignalRequest
 *
 * @seealso MonitorSendSignalReply
 */

/** @macro MONITOR_SEND_SIGNAL_REQUEST */
#define MONITOR_SEND_SIGNAL_REQUEST 39085 /* !-SIGNO(MonitorSendSignalRequest)-! */

/**
 * @type MonitorSendSignalRequest
 *
 * @input sigNo            MONITOR_SEND_SIGNAL_REQUEST
 * @input number           Signal number.
 * @input addressee        Signal addressee process ID.
 * @input sender           Signal sender process ID.
 * @input dataCount        Size in bytes of the data field.
 * @input data             Optional signal data, not including
 *                         the signal number.
 */
typedef struct MonitorSendSignalRequest
{
   uint32_t sigNo;
   uint32_t number;
   uint32_t sender;
   uint32_t addressee;
   uint32_t dataCount;
   uint8_t data[1];
} MonitorSendSignalRequest;

/**
 * @signal MonitorSendSignalReply
 *
 * @brief Send signal to a process reply.
 *
 * @member MONITOR_SEND_SIGNAL_REPLY
 * @member MonitorActionReply
 *
 * @seealso MonitorSendSignalRequest
 */

/** @macro MONITOR_SEND_SIGNAL_REPLY */
#define MONITOR_SEND_SIGNAL_REPLY 39086 /* !-SIGNO(MonitorActionReply)-! */

/**
 * @signal MonitorSignalSemaphoreRequest
 *
 * @brief Signal fast semaphore of a process request.
 *
 * @member MONITOR_SIGNAL_SEMAPHORE_REQUEST
 * @member MonitorSignalSemaphoreRequest
 *
 * @seealso MonitorSignalSemaphoreReply
 */

/** @macro MONITOR_SIGNAL_SEMAPHORE_REQUEST */
#define MONITOR_SIGNAL_SEMAPHORE_REQUEST 39087 /* !-SIGNO(MonitorSignalSemaphoreRequest)-! */

/**
 * @type MonitorSignalSemaphoreRequest
 *
 * @input sigNo  MONITOR_SIGNAL_SEMAPHORE_REQUEST
 * @input pid    ID of the process whose semaphore to signal.
 */
typedef struct MonitorSignalSemaphoreRequest
{
   uint32_t sigNo;
   uint32_t pid;
} MonitorSignalSemaphoreRequest;

/**
 * @signal MonitorSignalSemaphoreReply
 *
 * @brief Signal fast semaphore of a process reply.
 *
 * @member MONITOR_SIGNAL_SEMAPHORE_REPLY
 * @member MonitorActionReply
 *
 * @seealso MonitorSignalSemaphoreRequest
 */

/** @macro MONITOR_SIGNAL_SEMAPHORE_REPLY */
#define MONITOR_SIGNAL_SEMAPHORE_REPLY 39088 /* !-SIGNO(MonitorActionReply)-! */

/**
 * @signal MonitorSetPriorityRequest
 *
 * @brief Set priority of a process request.
 *
 * @long Compare the OSE call @link "ose.h" "set_pri_for"().
 *
 * @member MONITOR_SET_PRIORITY_REQUEST
 * @member MonitorSetPriorityRequest
 *
 * @seealso MonitorSetPriorityReply
 */

/** @macro MONITOR_SET_PRIORITY_REQUEST */
#define MONITOR_SET_PRIORITY_REQUEST 39089 /* !-SIGNO(MonitorSetPriorityRequest)-! */

/**
 * @type MonitorSetPriorityRequest
 *
 * @input sigNo    MONITOR_SET_PRIORITY_REQUEST
 * @input pid      ID of the process whose priority to set.
 * @input priority The new priority.
 */
typedef struct MonitorSetPriorityRequest
{
   uint32_t sigNo;
   uint32_t pid;
   uint32_t priority;
} MonitorSetPriorityRequest;

/**
 * @signal MonitorSetPriorityReply
 *
 * @brief Set priority of a process reply.
 *
 * @member MONITOR_SET_PRIORITY_REPLY
 * @member MonitorActionReply
 *
 * @seealso MonitorSetPriorityRequest
 */

/** @macro MONITOR_SET_PRIORITY_REPLY */
#define MONITOR_SET_PRIORITY_REPLY 39090 /* !-SIGNO(MonitorActionReply)-! */

/**
 * @signal MonitorSetSysParamRequest
 *
 * @brief Set system parameter variable request.
 *
 * @long If valueOffset is ~0, the system parameter is removed.
 *
 * @long Use this signal with caution. Setting a system parameter to an
 * inappropriate value may render the system unable to boot.
 *
 * @member MONITOR_SET_SYS_PARAM_REQUEST
 * @member MonitorSetSysParamRequest
 *
 * @seealso MonitorSetSysParamReply
 */

/** @macro MONITOR_SET_SYS_PARAM_REQUEST */
#define MONITOR_SET_SYS_PARAM_REQUEST 39091 /* !-SIGNO(MonitorSetSysParamRequest)-! */

/**
 * @type MonitorSetSysParamRequest
 *
 * @input sigNo       MONITOR_SET_SYS_PARAM_REQUEST
 * @input reconfigure If non zero and if the system parameter is read by the
 *                    monitor itself, then reread and reconfigure that
 *                    parameter in the monitor, after setting it.
 * @input reserved0   NOT USED and must contain 0.
 * @input reserved1   NOT USED and must contain 0.
 * @input reserved2   NOT USED and must contain 0.
 * @input reserved3   NOT USED and must contain 0.
 * @input nameOffset  Offset in data to null terminated system parameter name.
 * @input valueOffset Offset in data to null terminated system parameter value
 *                    (or ~0 to delete the system parameter).
 * @input data        Variable sized buffer.
 */
typedef struct MonitorSetSysParamRequest
{
   uint32_t sigNo;
   uint8_t  reconfigure;
   uint8_t  reserved0;
   uint16_t reserved1;
   uint32_t reserved2;
   uint32_t reserved3;
   uint32_t nameOffset;
   uint32_t valueOffset;
   char data[1];
} MonitorSetSysParamRequest;

/**
 * @signal MonitorSetSysParamReply
 *
 * @brief Set system parameter reply
 *
 * @member MONITOR_SET_SYS_PARAM_REPLY
 * @member MonitorActionReply
 *
 * @seealso MonitorSetSysParamRequest
 */

/** @macro MONITOR_SET_SYS_PARAM_REPLY */
#define MONITOR_SET_SYS_PARAM_REPLY 39092 /* !-SIGNO(MonitorActionReply)-! */

/**
 * @signal MonitorSetExecutionUnitRequest
 *
 * @brief Set execution unit (CPU core) of a block/process request.
 *
 * @long Compare with the OSE system call @link "ose_spi/ose_spi.h"
 * "ose_bind_process"().
 *
 * @member MONITOR_SET_EXECUTION_UNIT_REQUEST
 * @member MonitorSetExecutionUnitRequest
 *
 * @seealso MonitorSetExecutionUnitReply
 */

/** @macro MONITOR_SET_EXECUTION_UNIT_REQUEST */
#define MONITOR_SET_EXECUTION_UNIT_REQUEST 39093 /* !-SIGNO(MonitorSetExecutionUnitRequest)-! */

/**
 * @type MonitorSetExecutionUnitRequest
 *
 * @input sigNo    MONITOR_SET_EXECUTION_UNIT_REQUEST
 * @input pid      ID of the block/process whose execution unit to set.
 * @input euId     The new execution unit ID.
 * @input reserved NOT USED. Should be 0.
 */
typedef struct MonitorSetExecutionUnitRequest
{
   uint32_t sigNo;
   uint32_t pid;
   uint16_t euId;
   uint16_t reserved;
} MonitorSetExecutionUnitRequest;

/**
 * @signal MonitorSetExecutionUnitReply
 *
 * @brief Set execution unit (CPU core) of a block/process reply.
 *
 * @member MONITOR_SET_EXECUTION_UNIT_REPLY
 * @member MonitorActionReply
 *
 * @seealso MonitorSetExecutionUnitRequest
 */

/** @macro MONITOR_SET_EXECUTION_UNIT_REPLY */
#define MONITOR_SET_EXECUTION_UNIT_REPLY 39094 /* !-SIGNO(MonitorActionReply)-! */

#ifdef __cplusplus
} /* end of extern "C" */
#endif

#endif
