/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2011 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

/**
 * @toc OSE:
 * @file ose_spi/monitor_counters.sig
 *
 * @brief OSE source code performance counting profiling interface.
 *
 * @long This header file defines signals and types related to hardware
 * performance counter profiling, i.e. profiling based on PC sampling driven
 * by performance counter interrupts. The Run-Mode Monitor (RMM) implements
 * the actual profiling.
 *
 * @long A connection to the RMM is needed. It is not needed to attach a scope
 * before using these signals; these requests are independent of any attached
 * scopes.
 *
 * @long The sequence to collect profiling data is:
 *
 * @long 1. Connect to the monitor using @link "" "MonitorInterfaceRequest" and
 * @link "" "MonitorConnectRequest".
 *
 * @long 2. Retrieve the performance counter types supported by the monitor
 * using @link "" "MonitorGetSupportedCounterTypesRequest".
 *
 * @long 3. Configure and enable the desired performance counter type using
 * @link "" "MonitorSetCounterTypeEnabledRequest".
 *
 * @long 4. Enable performance counter report notifications to this client for
 * the desired performance counter type using @link ""
 * "MonitorSetCounterReportsNotifyEnabledRequest".
 *
 * @long 5. Receive asynchronous @link "" "MonitorCounterReportsNotify"
 * notification signals containing performance counter reports.
 *
 * @long 6. Disable performance counter report notifications to this client for
 * the used performance counter type using @link ""
 * "MonitorSetCounterReportsNotifyEnabledRequest".
 *
 * @long 7. Disable the used performance counter type using @link ""
 * "MonitorSetCounterTypeEnabledRequest".
 *
 * @long 8. Disconnect from the monitor using @link ""
 * "MonitorDisconnectRequest", allowing the monitor to deallocate resources
 * linked to the client.
 *
 * @seealso ose_spi/monitor.sig
 */

#ifndef _MONITOR_COUNTERS_SIG
#define _MONITOR_COUNTERS_SIG

#include "stdint.h"
#include "monitor.sig"

#ifdef __cplusplus
extern "C" {
#endif

/*******************************************************************************
 *
 * DEFINITIONS
 *
 ******************************************************************************/

/**
 * @type MonitorCounterType
 * @brief Performance counter type.
 */
typedef uint32_t MonitorCounterType;

/**
 * @type MonitorCounterReportType
 * @brief Performance counter report type.
 * @seealso MonitorCounterReportTypeValue
 */
typedef uint32_t MonitorCounterReportType;

/**
 * @type MonitorCounterReportTypeValue
 * @brief Enumeration of performance counter report types.
 * @field MONITOR_COUNTER_REPORT      Represents @link ""
 * "MonitorCounterReport" reports.
 * @field MONITOR_COUNTER_REPORT_EXT  Represents @link ""
 * "MonitorCounterReportExt" reports.
 * @seealso MonitorCounterReportType
 */
enum MonitorCounterReportTypeValue
{
   MONITOR_COUNTER_REPORT = 0,
   MONITOR_COUNTER_REPORT_EXT = 1
};

/**
 * @type MonitorCounterReport
 * @brief Performance counter report.
 * @output pc  Program counter.
 */
typedef struct MonitorCounterReport
{
   uint64_t pc;
} MonitorCounterReport;

/**
 * @type MonitorCounterReportExt
 * @brief Extended performance counter report.
 * @output pc         Program counter.
 * @output pid        Current process ID.
 * @output reserved0  NOT USED.
 */
typedef struct MonitorCounterReportExt
{
   uint64_t pc;
   uint32_t pid;
   uint32_t reserved0;
} MonitorCounterReportExt;

/**
 * @macro MONITOR_ALL_EXECUTION_UNITS
 * @brief Value indicating all execution units.
 * Given instead of an execution unit ID.
 */
#ifndef MONITOR_ALL_EXECUTION_UNITS
#define MONITOR_ALL_EXECUTION_UNITS 0xFFFF
#endif

/*******************************************************************************
 *
 * INFORMATION SIGNALS
 *
 ******************************************************************************/

/**
 * @signal MonitorGetSupportedCounterTypesRequest
 *
 * @brief Get supported performance counter types request.
 *
 * @long Get the list of performance counter types that this target supports.
 *
 * @long Followed by any number of MonitorGetSupportedCounterTypesReply,
 * MonitorBlockingFlowControl, and one MonitorGetSupportedCounterTypesEndmark.
 *
 * @member MONITOR_GET_SUPPORTED_COUNTER_TYPES_REQUEST
 * @member MonitorGetSupportedCounterTypesRequest
 */

/** @macro MONITOR_GET_SUPPORTED_COUNTER_TYPES_REQUEST */
#define MONITOR_GET_SUPPORTED_COUNTER_TYPES_REQUEST 40780 /* !-SIGNO(MonitorGetSupportedCounterTypesRequest)-! */

/**
 * @type MonitorGetSupportedCounterTypesRequest
 *
 * @input sigNo      MONITOR_GET_SUPPORTED_COUNTER_TYPES_REQUEST.
 * @input reserved0  NOT USED.
 */
typedef struct MonitorGetSupportedCounterTypesRequest
{
   uint32_t sigNo;
   uint32_t reserved0;
} MonitorGetSupportedCounterTypesRequest;

/**
 * @signal MonitorGetSupportedCounterTypesReply
 *
 * @brief Get supported performance counter types reply.
 *
 * @long This signal is sent under blocking flow control.
 *
 * @member MONITOR_GET_SUPPORTED_COUNTER_TYPES_REPLY
 * @member MonitorGetSupportedCounterTypesReply
 */

/** @macro MONITOR_GET_SUPPORTED_COUNTER_TYPES_REPLY */
#define MONITOR_GET_SUPPORTED_COUNTER_TYPES_REPLY 40781 /* !-SIGNO(MonitorGetSupportedCounterTypesReply)-! */

/**
 * @type MonitorGetSupportedCounterTypesReply
 *
 * @output sigNo        MONITOR_GET_SUPPORTED_COUNTER_TYPES_REPLY.
 * @output counterType  Performance counter type.
 * @output reserved0    NOT USED.
 * @output name         Null terminated name of the performance counter type.
 */
typedef struct MonitorGetSupportedCounterTypesReply
{
   uint32_t sigNo;
   MonitorCounterType counterType;
   uint32_t reserved0;
   char name[1];
} MonitorGetSupportedCounterTypesReply;

/**
 * @signal MonitorGetSupportedCounterTypesEndmark
 *
 * @brief Get supported performance counter types endmark, with status.
 *
 * @member MONITOR_GET_SUPPORTED_COUNTER_TYPES_ENDMARK
 * @member MonitorGetSupportedCounterTypesEndmark
 */

/** @macro MONITOR_GET_SUPPORTED_COUNTER_TYPES_ENDMARK */
#define MONITOR_GET_SUPPORTED_COUNTER_TYPES_ENDMARK 40782 /* !-SIGNO(MonitorGetSupportedCounterTypesEndmark)-! */

/**
 * @type MonitorGetSupportedCounterTypesEndmark
 *
 * @output sigNo   MONITOR_GET_SUPPORTED_COUNTER_TYPES_ENDMARK.
 * @output status  MONITOR_STATUS_OK or cause of error.
 */
typedef struct MonitorGetSupportedCounterTypesEndmark
{
   uint32_t sigNo;
   MonitorStatus status;
} MonitorGetSupportedCounterTypesEndmark;

/*******************************************************************************
 *
 * CONFIGURATION SIGNALS
 *
 ******************************************************************************/

/**
 * @signal MonitorGetCounterTypeEnabledRequest
 *
 * @brief Get performance counter type enabled request.
 *
 * @long Get enablement and configuration information for a specified
 * performance counter type and execution unit.
 *
 * @member MONITOR_GET_COUNTER_TYPE_ENABLED_REQUEST
 * @member MonitorGetCounterTypeEnabledRequest
 */

/** @macro MONITOR_GET_COUNTER_TYPE_ENABLED_REQUEST */
#define MONITOR_GET_COUNTER_TYPE_ENABLED_REQUEST 40783 /* !-SIGNO(MonitorGetCounterTypeEnabledRequest)-! */

/**
 * @type MonitorGetCounterTypeEnabledRequest
 *
 * @input sigNo        MONITOR_GET_COUNTER_TYPE_ENABLED_REQUEST.
 * @input euId         Execution unit or MONITOR_ALL_EXECUTION_UNITS.
 * @input reserved0    NOT USED.
 * @input counterType  Performance counter type.
 * @input reserved1    NOT USED.
 */
typedef struct MonitorGetCounterTypeEnabledRequest
{
   uint32_t sigNo;
   uint16_t euId;
   uint16_t reserved0;
   MonitorCounterType counterType;
   uint32_t reserved1;
} MonitorGetCounterTypeEnabledRequest;

/**
 * @signal MonitorGetCounterTypeEnabledReply
 *
 * @brief Get performance counter type enabled reply.
 *
 * @member MONITOR_GET_COUNTER_TYPE_ENABLED_REPLY
 * @member MonitorGetCounterTypeEnabledReply
 */

/** @macro MONITOR_GET_COUNTER_TYPE_ENABLED_REPLY */
#define MONITOR_GET_COUNTER_TYPE_ENABLED_REPLY 40784 /* !-SIGNO(MonitorGetCounterTypeEnabledReply)-! */

/**
 * @type MonitorGetCounterTypeEnabledReply
 *
 * @output sigNo         MONITOR_GET_COUNTER_TYPE_ENABLED_REPLY.
 * @output status        MONITOR_STATUS_OK or cause of error.
 * @output enabled       True if this performance counter type is enabled, false
 *                       otherwise.
 * @output euId          Execution unit or MONITOR_ALL_EXECUTION_UNITS.
 * @output reserved0     NOT USED.
 * @output reserved1     NOT USED.
 * @output counterType   Performance counter type.
 * @output counterValue  Performance counter trigger value.
 * @output maxReports    The currently configured maximum number of performance
 *                       counter reports in the target for the specified
 *                       performance counter type and execution unit.
 * @output reserved2     NOT USED.
 */
typedef struct MonitorGetCounterTypeEnabledReply
{
   uint32_t sigNo;
   MonitorStatus status;
   MonitorBoolean enabled;
   uint16_t euId;
   uint16_t reserved0;
   uint32_t reserved1;
   MonitorCounterType counterType;
   uint64_t counterValue;
   uint32_t maxReports;
   uint32_t reserved2;
} MonitorGetCounterTypeEnabledReply;

/**
 * @signal MonitorSetCounterTypeEnabledRequest
 *
 * @brief Set performance counter type enabled request.
 *
 * @long Set enablement and configuration information for a specified
 * performance counter type and execution unit.
 *
 * @member MONITOR_SET_COUNTER_TYPE_ENABLED_REQUEST
 * @member MonitorSetCounterTypeEnabledRequest
 */

/** @macro MONITOR_SET_COUNTER_TYPE_ENABLED_REQUEST */
#define MONITOR_SET_COUNTER_TYPE_ENABLED_REQUEST 40785 /* !-SIGNO(MonitorSetCounterTypeEnabledRequest)-! */

/**
 * @type MonitorSetCounterTypeEnabledRequest
 *
 * @input sigNo         MONITOR_SET_COUNTER_TYPE_ENABLED_REQUEST.
 * @input enabled       True if this performance counter type should be enabled,
 *                      false otherwise.
 * @input euId          Execution unit or MONITOR_ALL_EXECUTION_UNITS.
 * @input reserved0     NOT USED.
 * @input counterType   Performance counter type.
 * @input counterValue  Performance counter trigger value.
 * @input maxReports    The suggested maximum number of performance counter
 *                      reports kept in the target for the specified performance
 *                      counter type and execution unit. If set to zero or
 *                      larger than supported by the monitor, the configured
 *                      default value is used instead.
 * @input reserved1     NOT USED.
 */
typedef struct MonitorSetCounterTypeEnabledRequest
{
   uint32_t sigNo;
   MonitorBoolean enabled;
   uint16_t euId;
   uint16_t reserved0;
   MonitorCounterType counterType;
   uint64_t counterValue;
   uint32_t maxReports;
   uint32_t reserved1;
} MonitorSetCounterTypeEnabledRequest;

/**
 * @signal MonitorSetCounterTypeEnabledReply
 *
 * @brief Set performance counter type enabled reply.
 *
 * @member MONITOR_SET_COUNTER_TYPE_ENABLED_REPLY
 * @member MonitorSetCounterTypeEnabledReply
 */

/** @macro MONITOR_SET_COUNTER_TYPE_ENABLED_REPLY */
#define MONITOR_SET_COUNTER_TYPE_ENABLED_REPLY 40786 /* !-SIGNO(MonitorSetCounterTypeEnabledReply)-! */

/**
 * @type MonitorSetCounterTypeEnabledReply
 *
 * @output sigNo   MONITOR_SET_COUNTER_TYPE_ENABLED_REPLY.
 * @output status  MONITOR_STATUS_OK or cause of error.
 */
typedef struct MonitorSetCounterTypeEnabledReply
{
   uint32_t sigNo;
   MonitorStatus status;
} MonitorSetCounterTypeEnabledReply;

/*******************************************************************************
 *
 * NOTIFICATION ENABLEMENT SIGNALS
 *
 ******************************************************************************/

/**
 * @signal MonitorGetCounterReportsNotifyEnabledRequest
 *
 * @brief Get enablement/disablement status of performance counter report
 * notifications to this client for a specified performance counter type and
 * execution unit request.
 *
 * @member MONITOR_GET_COUNTER_REPORTS_NOTIFY_ENABLED_REQUEST
 * @member MonitorGetCounterReportsNotifyEnabledRequest
 */

/** @macro MONITOR_GET_COUNTER_REPORTS_NOTIFY_ENABLED_REQUEST */
#define MONITOR_GET_COUNTER_REPORTS_NOTIFY_ENABLED_REQUEST 40787 /* !-SIGNO(MonitorGetCounterReportsNotifyEnabledRequest)-! */

/**
 * @type MonitorGetCounterReportsNotifyEnabledRequest
 *
 * @input sigNo        MONITOR_GET_COUNTER_REPORTS_NOTIFY_ENABLED_REQUEST.
 * @input euId         Execution unit or MONITOR_ALL_EXECUTION_UNITS.
 * @input reserved0    NOT USED.
 * @input counterType  Performance counter type.
 * @input reserved1    NOT USED.
 */
typedef struct MonitorGetCounterReportsNotifyEnabledRequest
{
   uint32_t sigNo;
   uint16_t euId;
   uint16_t reserved0;
   MonitorCounterType counterType;
   uint32_t reserved1;
} MonitorGetCounterReportsNotifyEnabledRequest;

/**
 * @signal MonitorGetCounterReportsNotifyEnabledReply
 *
 * @brief Get enablement/disablement status of performance counter report
 * notifications to this client for a specified performance counter type and
 * execution unit reply.
 *
 * @member MONITOR_GET_COUNTER_REPORTS_NOTIFY_ENABLED_REPLY
 * @member MonitorGetCounterReportsNotifyEnabledReply
 */

/** @macro MONITOR_GET_COUNTER_REPORTS_NOTIFY_ENABLED_REPLY */
#define MONITOR_GET_COUNTER_REPORTS_NOTIFY_ENABLED_REPLY 40788 /* !-SIGNO(MonitorGetCounterReportsNotifyEnabledReply)-! */

/**
 * @type MonitorGetCounterReportsNotifyEnabledReply
 *
 * @output sigNo        MONITOR_GET_COUNTER_REPORTS_NOTIFY_ENABLED_REPLY.
 * @output status       MONITOR_STATUS_OK or cause of failure.
 * @output enabled      True if performance counter report notifications for the
 *                      specified performance counter type and execution unit is
 *                      enabled for this client, false otherwise.
 * @output euId         Execution unit or MONITOR_ALL_EXECUTION_UNITS.
 * @output reserved0    NOT USED.
 * @output counterType  Performance counter type.
 * @output reserved1    NOT USED.
 */
typedef struct MonitorGetCounterReportsNotifyEnabledReply
{
   uint32_t sigNo;
   MonitorStatus status;
   MonitorBoolean enabled;
   uint16_t euId;
   uint16_t reserved0;
   MonitorCounterType counterType;
   uint32_t reserved1;
} MonitorGetCounterReportsNotifyEnabledReply;

/**
 * @signal MonitorSetCounterReportsNotifyEnabledRequest
 *
 * @brief Enable/disable performance counter report notifications to this client
 * for a specified performance counter type and execution unit request.
 *
 * @long When performance counter report notifications are enabled, the client
 * should be prepared to receive @link "" "MonitorCounterReportsNotify" and
 * @link "" "MonitorCounterReportsLossNotify" signals.
 *
 * @member MONITOR_SET_COUNTER_REPORTS_NOTIFY_ENABLED_REQUEST
 * @member MonitorSetCounterReportsNotifyEnabledRequest
 */

/** @macro MONITOR_SET_COUNTER_REPORTS_NOTIFY_ENABLED_REQUEST */
#define MONITOR_SET_COUNTER_REPORTS_NOTIFY_ENABLED_REQUEST 40789 /* !-SIGNO(MonitorSetCounterReportsNotifyEnabledRequest)-! */

/**
 * @type MonitorSetCounterReportsNotifyEnabledRequest
 *
 * @input sigNo        MONITOR_SET_COUNTER_REPORTS_NOTIFY_ENABLED_REQUEST.
 * @input enabled      True if performance counter report notifications to this
 *                     client should be enabled for the specified performance
 *                     counter type and execution unit, false otherwise.
 * @input euId         Execution unit or MONITOR_ALL_EXECUTION_UNITS.
 * @input reserved0    NOT USED.
 * @input counterType  Performance counter type.
 * @input reserved1    NOT USED.
 */
typedef struct MonitorSetCounterReportsNotifyEnabledRequest
{
   uint32_t sigNo;
   MonitorBoolean enabled;
   uint16_t euId;
   uint16_t reserved0;
   MonitorCounterType counterType;
   uint32_t reserved1;
} MonitorSetCounterReportsNotifyEnabledRequest;

/**
 * @signal MonitorSetCounterReportsNotifyEnabledReply
 *
 * @brief Enable/disable performance counter report notifications to this client
 * for a specified performance counter type and execution unit reply.
 *
 * @member MONITOR_SET_COUNTER_REPORTS_NOTIFY_ENABLED_REPLY
 * @member MonitorSetCounterReportsNotifyEnabledReply
 */

/** @macro MONITOR_SET_COUNTER_REPORTS_NOTIFY_ENABLED_REPLY */
#define MONITOR_SET_COUNTER_REPORTS_NOTIFY_ENABLED_REPLY 40790 /* !-SIGNO(MonitorSetCounterReportsNotifyEnabledReply)-! */

/**
 * @type MonitorSetCounterReportsNotifyEnabledReply
 *
 * @output sigNo   MONITOR_SET_COUNTER_REPORTS_NOTIFY_ENABLED_REPLY.
 * @output status  MONITOR_STATUS_OK or cause of failure.
 */
typedef struct MonitorSetCounterReportsNotifyEnabledReply
{
   uint32_t sigNo;
   MonitorStatus status;
} MonitorSetCounterReportsNotifyEnabledReply;

/*******************************************************************************
 *
 * NOTIFICATION SIGNALS
 *
 ******************************************************************************/

/**
 * @signal MonitorCounterReportsNotify
 *
 * @brief Performance counter reports notify.
 *
 * @long This notification signal contains a batch of performance counter
 * reports for a performance counter type and execution unit.
 *
 * @long This signal is sent under non-blocking flow control.
 *
 * @member MONITOR_COUNTER_REPORTS_NOTIFY
 * @member MonitorCounterReportsNotify
 */

/** @macro MONITOR_COUNTER_REPORTS_NOTIFY */
#define MONITOR_COUNTER_REPORTS_NOTIFY 40791 /* !-SIGNO(MonitorCounterReportsNotify)-! */

/**
 * @type MonitorCounterReportsNotify
 *
 * @output sigNo              MONITOR_COUNTER_REPORTS_NOTIFY.
 * @output euId               Execution unit.
 * @output reserved0          NOT USED.
 * @output counterType        Performance counter type.
 * @output counterReportType  Performance counter report type, specifies the
 * type of the performance counter reports.
 * @output reserved1          NOT USED.
 * @output reportsCount       Number of performance counter reports.
 * @output reports            Array of performance counter reports.
 */
typedef struct MonitorCounterReportsNotify
{
   uint32_t sigNo;
   uint16_t euId;
   uint16_t reserved0;
   MonitorCounterType counterType;
   MonitorCounterReportType counterReportType;
   uint32_t reserved1;
   uint32_t reportsCount;
   MonitorCounterReport reports[1];
} MonitorCounterReportsNotify;

/**
 * @signal MonitorCounterReportsLossNotify
 *
 * @brief Notification that performance counter reports from the monitor for a
 * certain performance counter type and execution unit was lost.
 *
 * @long Notifies the client that the monitor has thrown performance counter
 * report notifications due to flow control.
 *
 * @long This signal is sent under non-blocking flow control.
 *
 * @member MONITOR_COUNTER_REPORTS_LOSS_NOTIFY
 * @member MonitorCounterReportsLossNotify
 */

/** @macro MONITOR_COUNTER_REPORTS_LOSS_NOTIFY */
#define MONITOR_COUNTER_REPORTS_LOSS_NOTIFY 40792 /* !-SIGNO(MonitorCounterReportsLossNotify)-! */

/**
 * @type MonitorCounterReportsLossNotify
 *
 * @output sigNo        MONITOR_COUNTER_REPORTS_LOSS_NOTIFY.
 * @output euId         Execution unit.
 * @output reserved0    NOT USED.
 * @output counterType  Performance counter type.
 * @output reportsLost  Number of performance counter reports lost since the
 *                      last @link "" "MonitorCounterReportsNotify" was sent to
 *                      this client for the specified performance counter type
 *                      and execution unit.
 * @output reserved1    NOT USED.
 */
typedef struct MonitorCounterReportsLossNotify
{
   uint32_t sigNo;
   uint16_t euId;
   uint16_t reserved0;
   MonitorCounterType counterType;
   uint32_t reportsLost;
   uint32_t reserved1;
} MonitorCounterReportsLossNotify;

#ifdef __cplusplus
} /* end of extern "C" */
#endif

#endif
