/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2005-2009 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

/**
 * @toc OSE:
 * @file ose_spi/monitor_cpu.sig
 *
 * @brief OSE System CPU load and user-value profiling interface.
 *
 * @long This header file defines signals and types related to CPU profiling
 * of OSE processes. It also allows user-value profiling, producing reports
 * based on user values, e.g. for profiling applications.
 * The Run-Mode Monitor (RMM) implements the actual profiling.
 *
 * @long A connection to the RMM is needed, see monitor.sig for the actual
 * connection procedure using @link "" "MonitorInterfaceRequest" and
 * by @link "" "MonitorConnectRequest".
 *
 * @long It is not needed to attach a scope before using these signals.
 * These requests are independent of any attached scopes.
 *
 * @long Different types of profiling are supported:
 * @tab - Total CPU load
 * @tab Started by @link "" "MonitorSetCPUReportsEnabledRequest".
 *      Reports @link "" "MonitorCPUReport".
 * @break
 * @tab - CPU load per processes priority level.
 * @tab The levels are Interrupt, Background and 32 different
 *      Prioritized process levels.
 *      Started by @link "" "MonitorSetCPUPriReportsEnabledRequest".
 *      Reports @link "" "MonitorCPUPriReport".
 * @break
 * @tab - CPU load per process.
 * @tab Started by @link "" "MonitorSetCPUProcessReportsEnabledRequest".
 *      Reports @link "" "MonitorCPUProcessReport".
 * @break
 * @tab - CPU load per process, limited to a scope.
 * @tab Define scopes of processes with one or several
 *      @link "" "MonitorSetCPUProcessReportpointRequest",
 *      start by @link "" "MonitorSetCPUProcessReportsEnabledRequest".
 *      Reports are @link "" "MonitorCPUProcessReport".
 *      Reports are sent with blocking flow control.
 *      Report points with scopes can be added or cleared.
 * @break
 * @tab - CPU load per block.
 * @tab Started by @link "" "MonitorSetCPUBlockReportsEnabledRequest".
 *      Reports @link "" "MonitorCPUProcessReport".
 * @break
 * @tab - User value profiling report.
 * @tab Started by @link "" "MonitorSetUserReportsEnabledRequest".
 *      Reports are @link "" "MonitorUserReport"
 *      or @link "" "MonitorMaxMinUserReport".
 * @break
 *
 * @long The actions to collect statistics are enabled (started) or
 * disabled (stopped) with requests in this interface.
 *
 * @long User value profiling depends on reports created and issued by
 * applications. Use the functions @link "ose.h" "ose_create_report" to
 * create a user-report with a given report number, optionally with
 * multiple values with IDs and supporting max/min collection, optionally
 * with values continuing (not reset) between each measurement interval. The
 * application issues @link "ose.h" "ose_set_report_val" calls to update
 * a value (report number and identifier) for which the monitor shall
 * collect statistics and provide reports. The monitor will set a trig
 * indicator for each user-report which is enabled, allowing the application
 * to avoid updating values that nobody is interested in.
 *
 * @long Additionally a user-defined user-report handler can be registered
 * with @link "rmm.h" "ose_rmm_set_user_report_handler"() to be called before
 * the end of each collection interval. This handler is thus called regularly
 * and can issue @link "ose.h" "ose_set_report_val" calls to report any
 * suitable value, e.g. from hardware or the application.
 *
 * @long There is a reserved range of user report identifiers for OSE
 * internal use, e.g. see ose_reports.h, @link "" "OSE_REPORTS_ENEA_END",
 * and in different *_reports.h files.
 *
 * @long The sequence to collect profiling data is:
 *
 * @long 1. @link "" "MonitorInterfaceRequest" and
 *          @link "" "MonitorConnectRequest".
 *
 * @long 2. Optionally define process scopes for process profiling.
 *          This ensures that these processes are first in the
 *          process profiling reports. Otherwise, processes will appear
 *          in the report in the order with the first non-system
 *          executing processes in each interval first in each report.
 *          When all entries in a report are filled, the profiling sum
 *          for all other processes in the interval are in the final
 *          "others" report entry.
 *
 * @long 3. Enable selected profiling data collection.
 *          Optionally enable other profiling types.
 *
 * @long 4. Poll the monitor for the results.
 *
 * @long 5. Disable profiling.
 *
 * @long 6. Disconnect with @link "" "MonitorDisconnectRequest",
 *          allowing the monitor to deallocate resources linked
 *          to the client.
 *
 * @long System parameters with names beginning with krn/monitor/profile
 * and krn/monitor/prof are available to configure the profiling.
 * See RMM in the "OSE Core User's Guide".
 *
 *
 * @restriction The OSE Soft Kernel (SFK) has only limited
 * support for profiling. User-value profiling and user reports are
 * supported. CPU profiling currently is not supported and attempts
 * to use it will return invalid values in replies.
 *
 * @seealso ose_spi/monitor.sig
 * @seealso ose_create_report
 * @seealso ose_set_report_val
 * @seealso ose_rmm_set_user_report_handler
 */

#ifndef _MONITOR_CPU_SIG
#define _MONITOR_CPU_SIG

#include "stdint.h"
#include "monitor.sig"

#ifdef __cplusplus
extern "C" {
#endif

/*******************************************************************************
 *
 * DEFINITIONS
 *
 ******************************************************************************/

/**
 * @type MonitorCPUTimestamp
 *
 * @brief Timestamp used in CPU usage data.
 *
 * @output tick       Number of ticks.
 * @output ntick      Number of timer steps after the last tick.
 */
typedef struct MonitorCPUTimestamp
{
   uint32_t tick;
   uint32_t ntick;
} MonitorCPUTimestamp;

/**
 * @type MonitorCPUReport
 * @brief CPU load measurement report type.
 * @long It is used in @link "" "MonitorGetCPUReportsReply".
 * @output time     Report creation timestamp
 *                  (@link "" "MonitorCPUTimestamp").
 * @output interval Total CPU time available in the measurement interval
 *                  (unspecified unit).
 * @output sum      Time the CPU was occupied in the interval
 *                  (unspecified unit).
 */
typedef struct MonitorCPUReport
{
   MonitorCPUTimestamp time;
   uint32_t interval;
   uint32_t sum;
} MonitorCPUReport;

/**
 * @type MonitorCPUPriReport
 * @brief Priority load measurement report type.
 * @output time           Report creation time.
 * @output interval       Total CPU time available in the measurement interval
 *                        (unspecified unit).
 * @output sumInterrupt   Time the CPU was occupied with interrupt processes
 *                        (unspecified unit).
 * @output sumBackground  Time the CPU was occupied with background processes
 *                        (unspecified unit).
 * @output sumPrioritized Array with the times the CPU was occupied with each of
 *                        the 32 process priority levels.
 *                        One number for each prioritized level
 *                        (unspecified unit).
 * @field reserved0       NOT USED.
 *
 * @seealso MonitorGetCPUPriReportsReply
 */
typedef struct MonitorCPUPriReport
{
   MonitorCPUTimestamp time;
   uint32_t interval;
   uint32_t sumInterrupt;
   uint32_t sumBackground;
   uint32_t reserved0;
   uint32_t sumPrioritized[32];
} MonitorCPUPriReport;

/**
 * @type MonitorCPUProcess
 * @brief CPU load for a certain process (or other execution context).
 * @output id  ID of execution context.
 * @output sum Time the CPU was occupied with this execution context
 *             (unspecified unit).
 * @seealso MonitorCPUProcessReport
 */
typedef struct MonitorCPUProcess
{
   uint32_t id;
   uint32_t sum;
} MonitorCPUProcess;

/**
 * @type MonitorCPUProcessReport
 * @brief Process load measurement report type.
 * @output time           Report creation time
 * @output interval       Total CPU time available in the measurement interval
 *                        (unspecified unit).
 * @output sumOther       Time the CPU was occupied with other processes
 *                        (unspecified unit).
 * @output reserved0      NOT USED.
 * @output processesCount Number of elements in the processes list.
 * @output processes      Elements of @link "" "MonitorCPUProcess" structs,
 *                        with ID and CPU time consumed for each process.
 *
 */
typedef struct MonitorCPUProcessReport
{
   MonitorCPUTimestamp time;
   uint32_t interval;
   uint32_t sumOther;
   uint32_t reserved0;
   uint32_t processesCount;
   MonitorCPUProcess processes[1];
} MonitorCPUProcessReport;


/**
 * @type MonitorUserReportValue
 * @brief User report measurement value.
 * @long Multiple values are reported in each
 * @link "" "MonitorUserReport".
 *
 * @long This struct is also included as a part of the
 * longer report struct @link "" "MonitorMaxMinUserReportValue".
 *
 * @output id  User report ID, identifying this value. Should
 *             be ignored if the created user report does not
 *             contain multiple values.
 * @output sum The sum of reported values during each interval.
 *
 * @seealso MonitorUserReport
 * @seealso MonitorMaxMinUserReportValue
 */
typedef struct MonitorUserReportValue
{
   uint32_t id;
   int32_t sum;
} MonitorUserReportValue;

/**
 * @type MonitorUserReport
 * @brief Struct with user report measurements.
 * @long @link "" "MonitorGetUserReportsReply" signals (one for each
 * user report number) will contain an array of either this struct or
 * the corresponding struct with additional max-min values.
 * @output time                 Report creation time.
 * @output interval             Time of the measurement interval
 *                              (unspecified unit).
 * @output sumOther             The sum of all values not allocated to a
 *                              specific report ID.
 * @output reserved0            NOT USED.
 * @output reportValuesCount    Number of elements in the reportValues list.
 * @output reportValues         Measurement value elements. Each element
 *                              is a @link "" "MonitorUserReportValue"
 *                              struct.
 *
 * @seealso MonitorGetUserReportsReply
 * @seealso MonitorMaxMinUserReport
 */
typedef struct MonitorUserReport
{
   MonitorCPUTimestamp time;
   uint32_t interval;
   int32_t sumOther;
   uint32_t reserved0;
   uint32_t reportValuesCount;
   MonitorUserReportValue reportValues[1];
} MonitorUserReport;

/**
 * @type MonitorMaxMinUserReportValue
 * @brief User report measurement values with min and max values.
 * @long Multiple values are reported in each
 * @link "" "MonitorMaxMinUserReport".
 *
 * @long The leading part of this struct is the same as the
 * basic MonitorUserReportValue struct.
 *
 * @output reportValue    Basic user report measurement value. This is
 *                        a @link "" "MonitorUserReportValue" struct
 *                        with report ID and sum.
 * @output max            The maximum value during the interval.
 * @output min            The minimum value during the interval.
 *
 * @seealso MonitorMaxMinUserReport
 */
typedef struct MonitorMaxMinUserReportValue
{
   MonitorUserReportValue reportValue;
   int32_t max;
   int32_t min;
} MonitorMaxMinUserReportValue;

/**
 * @type MonitorMaxMinUserReport
 * @brief Struct with user report measurements including max/min.
 * @long @link "" "MonitorGetUserReportsReply" signals (one for each
 * user report number) will contain an array of either this struct or
 * the corresponding smaller struct without max-min values.
 * @output time                 Report creation time.
 * @output interval             Time of the measurement interval
 *                              (unspecified unit).
 * @output sumOther             The sum of all values not allocated to a
 *                              specific ID.
 * @output maxOther             The maximum of sumOther during this interval.
 * @output minOther             The minimum of sumOther during this interval.
 * @output reserved2            NOT USED.
 * @output reportValuesCount    Number of elements in reportValues.
 * @output reportValues         Measurement value elements. Each element
 *                              is a @link "" "MonitorMaxMinUserReportValue"
 *                              struct.
 *
 * @seealso MonitorGetUserReportsReply
 * @seealso MonitorUserReport
 */
typedef struct MonitorMaxMinUserReport
{
   MonitorCPUTimestamp time;
   uint32_t interval;
   int32_t sumOther;
   int32_t maxOther;
   int32_t minOther;
   uint32_t reserved2;
   uint32_t reportValuesCount;
   MonitorMaxMinUserReportValue reportValues[1];
} MonitorMaxMinUserReport;

/**
 * @macro MONITOR_ALL_EXECUTION_UNITS
 * @brief Value to get the sum of the execution time for all execution units.
 *        It can be given in the field euId, instead of an execution unit ID.
 * @long  If MONITOR_ALL_EXECUTION_UNITS is given instead of an execution unit,
 *        the reported length of the execution time is the sum of the lengths
 *        of the execution times for all execution units. Thus the time can
 *        be longer than the total interval.
 * @seealso MonitorGetCPUReportsRequest
 * @seealso MonitorGetCPUReportsReply
 * @seealso MonitorGetCPUPriReportsRequest
 * @seealso MonitorGetCPUPriReportsReply
 * @seealso MonitorGetCPUProcessReportsRequest
 * @seealso MonitorGetCPUProcessReportsReply
 */
#define MONITOR_ALL_EXECUTION_UNITS 0xFFFF


/*******************************************************************************
 *
 * CONFIGURATION SIGNALS
 *
 ******************************************************************************/

/**
 * @signal MonitorGetCPUReportsEnabledRequest
 *
 * @brief Get status of CPU measurement request.
 *
 * @member MONITOR_GET_CPU_REPORTS_ENABLED_REQUEST
 * @member MonitorGetCPUReportsEnabledRequest
 */

/** @macro MONITOR_GET_CPU_REPORTS_ENABLED_REQUEST */
#define MONITOR_GET_CPU_REPORTS_ENABLED_REQUEST 39250 /* !-SIGNO(MonitorGetCPUReportsEnabledRequest)-! */

/**
 * @type MonitorGetCPUReportsEnabledRequest
 *
 * @input sigNo MONITOR_GET_CPU_REPORTS_ENABLED_REQUEST.
 */
typedef struct MonitorGetCPUReportsEnabledRequest
{
   uint32_t sigNo;
} MonitorGetCPUReportsEnabledRequest;

/**
 * @signal MonitorGetCPUReportsEnabledReply
 *
 * @brief Get status of CPU measurement reply.
 *
 * @member MONITOR_GET_CPU_REPORTS_ENABLED_REPLY
 * @member MonitorGetCPUReportsEnabledReply
 */

/** @macro MONITOR_GET_CPU_REPORTS_ENABLED_REPLY */
#define MONITOR_GET_CPU_REPORTS_ENABLED_REPLY 39251 /* !-SIGNO(MonitorGetCPUReportsEnabledReply)-! */

/**
 * @type MonitorGetCPUReportsEnabledReply
 *
 * @output sigNo             MONITOR_GET_CPU_REPORTS_ENABLED_REPLY.
 * @output enabled           1 if load info enabled otherwise zero.
 * @output interval          Integration interval in ticks.
 * @output priority          Priority limit. CPU time spent on this
 *                           or lower priority (higher value) is
 *                           disregarded.
 * @output maxReports        The currently configured maximum number of
 *                           reports in the target.
 * @field reserved0          NOT USED.
 * @output sec               Number of seconds since 00:00:00 01 Jan 1970 GMT
 *                           at the system time sectick:secntick, or zero
 *                           if not available.
 * @output sectick           Tick count at time sec.
 * @output secntick          Number of timer steps at time sec.
 */
typedef struct MonitorGetCPUReportsEnabledReply
{
   uint32_t sigNo;
   MonitorBoolean enabled;
   uint32_t interval;
   uint32_t priority;
   uint32_t maxReports;
   uint32_t reserved0;
   uint32_t sec;
   uint32_t sectick;
   uint32_t secntick;
} MonitorGetCPUReportsEnabledReply;

/**
 * @signal MonitorSetCPUReportsEnabledRequest
 *
 * @brief Enable/disable CPU load measurement request.
 *
 * @member MONITOR_SET_CPU_REPORTS_ENABLED_REQUEST
 * @member MonitorSetCPUReportsEnabledRequest
 */

/** @macro MONITOR_SET_CPU_REPORTS_ENABLED_REQUEST */
#define MONITOR_SET_CPU_REPORTS_ENABLED_REQUEST 39252 /* !-SIGNO(MonitorSetCPUReportsEnabledRequest)-! */

/**
 * @type MonitorSetCPUReportsEnabledRequest
 *
 * @input sigNo             MONITOR_SET_CPU_REPORTS_ENABLED_REQUEST.
 * @input enabled           1 if load reports are enabled, otherwise zero.
 * @input interval          Integration interval in ticks. If the interval
 *                          is outside the configured range in the monitor,
 *                          the configured default value is used instead.
 * @input priority          Priority limit. CPU time spent on
 *                          lower priority than this (higher value) is
 *                          disregarded. If the limit is set to a value
 *                          outside the configured range in the monitor,
 *                          the configured default value is used.instead.
 *                          Examples: 0: Report only the load by interrupt
 *                          processes. 32: Interrupt processes and all
 *                          prioritized processes. 33: Also background
 *                          processes.
 * @input maxReports        The suggested maximum number of reports kept in the
 *                          target in a circular buffer. If set to a value
 *                          outside the configured range in the monitor,
 *                          the configured default value is used instead.
 * @field reserved0         Must be set to zero. NOT USED.
 */
typedef struct MonitorSetCPUReportsEnabledRequest
{
   uint32_t sigNo;
   MonitorBoolean enabled;
   uint32_t interval;
   uint32_t priority;
   uint32_t maxReports;
   uint32_t reserved0;
} MonitorSetCPUReportsEnabledRequest;

/**
 * @signal MonitorSetCPUReportsEnabledReply
 *
 * @brief Enable/disable CPU load measurement reply.
 *
 * @member MONITOR_SET_CPU_REPORTS_ENABLED_REPLY
 * @member MonitorSetCPUReportsEnabledReply
 */

/** @macro MONITOR_SET_CPU_REPORTS_ENABLED_REPLY */
#define MONITOR_SET_CPU_REPORTS_ENABLED_REPLY 39253 /* !-SIGNO(MonitorSetCPUReportsEnabledReply)-! */

/**
 * @type MonitorSetCPUReportsEnabledReply
 *
 * @output sigNo MONITOR_SET_CPU_REPORTS_ENABLED_REPLY.
 * @output status MONITOR_STATUS_OK or cause of failure.
 */
typedef struct MonitorSetCPUReportsEnabledReply
{
   uint32_t sigNo;
   MonitorStatus status;
} MonitorSetCPUReportsEnabledReply;

/**
 * @signal MonitorGetCPUPriReportsEnabledRequest
 *
 * @brief Get status of priority level measurement request.
 *
 * @member MONITOR_GET_CPU_PRI_REPORTS_ENABLED_REQUEST
 * @member MonitorGetCPUPriReportsEnabledRequest
 */

/** @macro MONITOR_GET_CPU_PRI_REPORTS_ENABLED_REQUEST */
#define MONITOR_GET_CPU_PRI_REPORTS_ENABLED_REQUEST 39254 /* !-SIGNO(MonitorGetCPUPriReportsEnabledRequest)-! */

/**
 * @type MonitorGetCPUPriReportsEnabledRequest
 *
 * @input sigNo MONITOR_GET_CPU_PRI_REPORTS_ENABLED_REQUEST.
 */
typedef struct MonitorGetCPUPriReportsEnabledRequest
{
   uint32_t sigNo;
} MonitorGetCPUPriReportsEnabledRequest;

/**
 * @signal MonitorGetCPUPriReportsEnabledReply
 *
 * @brief Get status of priority level measurement reply.
 *
 * @member MONITOR_GET_CPU_PRI_REPORTS_ENABLED_REPLY
 * @member MonitorGetCPUPriReportsEnabledReply
 */

/** @macro MONITOR_GET_CPU_PRI_REPORTS_ENABLED_REPLY */
#define MONITOR_GET_CPU_PRI_REPORTS_ENABLED_REPLY 39255 /* !-SIGNO(MonitorGetCPUPriReportsEnabledReply)-! */

/**
 * @type MonitorGetCPUPriReportsEnabledReply
 *
 * @output sigNo             MONITOR_GET_CPU_PRI_REPORTS_ENABLED_REPLY.
 * @output enabled           1 if load info enabled otherwise zero.
 * @output interval          Integration interval in ticks.
 * @output maxReports        The currently configured maximum number of
 *                           reports in the target.
 * @field reserved0          NOT USED.
 * @output sec               Number of seconds since 00:00:00 01 Jan 1970 GMT
 *                           at the system time sectick:secntick, or zero
 *                           if not available.
 * @output sectick           Tick count at time sec.
 * @output secntick          Number of timer steps at time sec.
 */
typedef struct MonitorGetCPUPriReportsEnabledReply
{
   uint32_t sigNo;
   MonitorBoolean enabled;
   uint32_t interval;
   uint32_t maxReports;
   uint32_t reserved0;
   uint32_t sec;
   uint32_t sectick;
   uint32_t secntick;
} MonitorGetCPUPriReportsEnabledReply;

/**
 * @signal MonitorSetCPUPriReportsEnabledRequest
 *
 * @brief Enable/disable priority level measurement request.
 *
 * @member MONITOR_SET_CPU_PRI_REPORTS_ENABLED_REQUEST
 * @member MonitorSetCPUPriReportsEnabledRequest
 */

/** @macro MONITOR_SET_CPU_PRI_REPORTS_ENABLED_REQUEST */
#define MONITOR_SET_CPU_PRI_REPORTS_ENABLED_REQUEST 39256 /* !-SIGNO(MonitorSetCPUPriReportsEnabledRequest)-! */

/**
 * @type MonitorSetCPUPriReportsEnabledRequest
 *
 * @input sigNo             MONITOR_SET_CPU_PRI_REPORTS_ENABLED_REQUEST.
 * @input enabled           1 if load info enabled otherwise zero.
 * @input interval          Integration interval in ticks.
 * @input maxReports        The suggested maximum number of reports kept in
 *                          target in a circular buffer. If set to zero or
 *                          larger than supported by the monitor, the
 *                          configured default value is used instead.
 * @field reserved0         Must be set to zero. NOT USED.
 */
typedef struct MonitorSetCPUPriReportsEnabledRequest
{
   uint32_t sigNo;
   MonitorBoolean enabled;
   uint32_t interval;
   uint32_t maxReports;
   uint32_t reserved0;
} MonitorSetCPUPriReportsEnabledRequest;

/**
 * @signal MonitorSetCPUPriReportsEnabledReply
 *
 * @brief Enable/disable priority level measurement reply.
 *
 * @member MONITOR_SET_CPU_PRI_REPORTS_ENABLED_REPLY
 * @member MonitorSetCPUPriReportsEnabledReply
 */

/** @macro MONITOR_SET_CPU_PRI_REPORTS_ENABLED_REPLY */
#define MONITOR_SET_CPU_PRI_REPORTS_ENABLED_REPLY 39257 /* !-SIGNO(MonitorSetCPUPriReportsEnabledReply)-! */

/**
 * @type MonitorSetCPUPriReportsEnabledReply
 *
 * @output sigNo             MONITOR_SET_CPU_PRI_REPORTS_ENABLED_REPLY.
 * @output status            MONITOR_STATUS_OK or cause of failure.
 */
typedef struct MonitorSetCPUPriReportsEnabledReply
{
   uint32_t sigNo;
   MonitorStatus status;
} MonitorSetCPUPriReportsEnabledReply;

/**
 * @signal MonitorGetCPUProcessReportsEnabledRequest
 *
 * @brief Get status of process level measurement request.
 *
 * @member MONITOR_GET_CPU_PROCESS_REPORTS_ENABLED_REQUEST
 * @member MonitorGetCPUProcessReportsEnabledRequest
 */

/** @macro MONITOR_GET_CPU_PROCESS_REPORTS_ENABLED_REQUEST */
#define MONITOR_GET_CPU_PROCESS_REPORTS_ENABLED_REQUEST 39258 /* !-SIGNO(MonitorGetCPUProcessReportsEnabledRequest)-! */

/**
 * @type MonitorGetCPUProcessReportsEnabledRequest
 *
 * @input sigNo MONITOR_GET_CPU_PROCESS_REPORTS_ENABLED_REQUEST
 */
typedef struct MonitorGetCPUProcessReportsEnabledRequest
{
   uint32_t sigNo;
} MonitorGetCPUProcessReportsEnabledRequest;

/**
 * @signal MonitorGetCPUProcessReportsEnabledReply
 *
 * @brief Get status of process level measurement reply.
 *
 * @member MONITOR_GET_CPU_PROCESS_REPORTS_ENABLED_REPLY
 * @member MonitorGetCPUProcessReportsEnabledReply
 */

/** @macro MONITOR_GET_CPU_PROCESS_REPORTS_ENABLED_REPLY */
#define MONITOR_GET_CPU_PROCESS_REPORTS_ENABLED_REPLY 39259 /* !-SIGNO(MonitorGetCPUProcessReportsEnabledReply)-! */

/**
 * @type MonitorGetCPUProcessReportsEnabledReply
 *
 * @output sigNo            MONITOR_GET_CPU_PROCESS_REPORTS_ENABLED_REPLY.
 * @output enabled          1 if load info enabled otherwise zero.
 * @output interval         Integration interval in ticks.
 * @output maxReports       The currently configured maximum number of
 *                          reports in the target.
 * @output maxProcsReport   The currently configured maximum number of
 *                          processes per report.
 * @input usePriority       A boolean value that decide if the priority parameter is used
 *                          if the value is set to True a priority threshold is used.
 * @input priority          The priority threshold. Only load from processes with higher
 *                          priority is consider.
 * @field reserved1         Must be set to zero. NOT USED.
 * @output sec              Number of seconds since 00:00:00 01 Jan 1970 GMT
 *                          at the system time sectick:secntick, or zero
 *                          if not available.
 * @output sectick          Tick count at time sec.
 * @output secntick         Number of timer steps at time sec.
 */
typedef struct MonitorGetCPUProcessReportsEnabledReply
{
   uint32_t sigNo;
   MonitorBoolean enabled;
   uint32_t interval;
   uint32_t maxReports;
   uint32_t maxProcsReport;
   uint8_t usePriority;
   uint8_t priority;
   uint16_t reserved1;
   uint32_t sec;
   uint32_t sectick;
   uint32_t secntick;
} MonitorGetCPUProcessReportsEnabledReply;

/**
 * @signal MonitorSetCPUProcessReportsEnabledRequest
 *
 * @brief Enable/disable process level measurement request.
 *
 * @long Optionally before this, define one or several
 * report points with process scopes, using
 * @link "" "MonitorSetCPUProcessReportpointRequest".
 * This will force the reports to be first filled with the
 * processes in the scopes.
 *
 * @member MONITOR_SET_CPU_PROCESS_REPORTS_ENABLED_REQUEST
 * @member MonitorSetCPUProcessReportsEnabledRequest
 *
 * @seealso MonitorSetCPUProcessReportpointRequest
 */

/** @macro MONITOR_SET_CPU_PROCESS_REPORTS_ENABLED_REQUEST */
#define MONITOR_SET_CPU_PROCESS_REPORTS_ENABLED_REQUEST 39260 /* !-SIGNO(MonitorSetCPUProcessReportsEnabledRequest)-! */

/**
 * @type MonitorSetCPUProcessReportsEnabledRequest
 *
 * @input sigNo             MONITOR_SET_CPU_PROCESS_REPORTS_ENABLED_REQUEST or
 *                          MONITOR_SET_CPU_BLOCK_REPORTS_ENABLED_REQUEST or
 *                          MONITOR_SET_TOTAL_CPU_BLOCK_REPORTS_ENABLED_REQUEST.
 * @input enabled           1 if load info enabled otherwise zero.
 * @input interval          Integration interval in ticks.
 * @input maxReports        The suggested maximum number of reports kept in
 *                          target in a circular buffer. If set to zero or
 *                          larger than supported by the monitor, the
 *                          configured default value is used instead.
 * @input maxProcsReport    The suggested maximum number of processes per report.
 *                          This does not include the entry "other", which contains
 *                          the sum for all other processes (except those given
 *                          explicitly in the report). If set to zero or
 *                          larger than what is supported by the monitor, the
 *                          configured default value is used instead.
 *                          If scopes are defined in report points, the reports
 *                          are first filled with those processes.
 * @input usePriority       A boolean value that decide if the priority parameter is used
 *                          if the value is set to True a priority threshold is used.
 * @input priority          The priority threshold. Only load from processes with higher
 *                          priority is consider.
 * @field reserved1         If requested for process or block, b0 can optionally be set to 1 to
 *                          let the CPUs individually order measured PIDs.
 *                          If requested for total_cpu_block, b3, b2, b1, and b0 can be used to
 *                          specify the 'oversize' factor.
 *                          Other bits must be 0.
 */
typedef struct MonitorSetCPUProcessReportsEnabledRequest
{
   uint32_t sigNo;
   MonitorBoolean enabled;
   uint32_t interval;
   uint32_t maxReports;
   uint32_t maxProcsReport;
   uint8_t usePriority;
   uint8_t priority;
   uint16_t reserved1;
} MonitorSetCPUProcessReportsEnabledRequest;

/**
 * @signal MonitorSetCPUProcessReportsEnabledReply
 *
 * @brief Enable/disable process level measurement reply.
 *
 * @member MONITOR_SET_CPU_PROCESS_REPORTS_ENABLED_REPLY
 * @member MonitorSetCPUProcessReportsEnabledReply
 */

/** @macro MONITOR_SET_CPU_PROCESS_REPORTS_ENABLED_REPLY */
#define MONITOR_SET_CPU_PROCESS_REPORTS_ENABLED_REPLY 39261 /* !-SIGNO(MonitorSetCPUProcessReportsEnabledReply)-! */

/**
 * @type MonitorSetCPUProcessReportsEnabledReply
 *
 * @output sigNo             MONITOR_SET_CPU_PROCESS_REPORTS_ENABLED_REPLY.
 * @output status            MONITOR_STATUS_OK or cause of failure.
 */
typedef struct MonitorSetCPUProcessReportsEnabledReply
{
   uint32_t sigNo;
   MonitorStatus status;
} MonitorSetCPUProcessReportsEnabledReply;

/**
 * @signal MonitorGetCPUBlockReportsEnabledRequest
 *
 * @brief Get status of block level measurement request.
 *
 * @member MONITOR_GET_CPU_BLOCK_REPORTS_ENABLED_REQUEST
 * @member MonitorGetCPUProcessReportsEnabledRequest
 */

/** @macro MONITOR_GET_CPU_BLOCK_REPORTS_ENABLED_REQUEST */
#define MONITOR_GET_CPU_BLOCK_REPORTS_ENABLED_REQUEST 39290 /* !-SIGNO(MonitorGetCPUProcessReportsEnabledRequest)-! */

/**
 * @signal MonitorGetCPUBlockReportsEnabledReply
 *
 * @brief Get status of block level measurement reply.
 *
 * @member MONITOR_GET_CPU_BLOCK_REPORTS_ENABLED_REPLY
 * @member MonitorGetCPUProcessReportsEnabledReply
 */

/** @macro MONITOR_GET_CPU_BLOCK_REPORTS_ENABLED_REPLY */
#define MONITOR_GET_CPU_BLOCK_REPORTS_ENABLED_REPLY 39291 /* !-SIGNO(MonitorGetCPUProcessReportsEnabledReply)-! */

/**
 * @signal MonitorSetCPUBlockReportsEnabledRequest
 *
 * @brief Enable/disable block level measurement request.
 *
 * @member MONITOR_SET_CPU_BLOCK_REPORTS_ENABLED_REQUEST
 * @member MonitorSetCPUProcessReportsEnabledRequest
 */

/** @macro MONITOR_SET_CPU_BLOCK_REPORTS_ENABLED_REQUEST */
#define MONITOR_SET_CPU_BLOCK_REPORTS_ENABLED_REQUEST 39292 /* !-SIGNO(MonitorSetCPUProcessReportsEnabledRequest)-! */

/**
 * @signal MonitorSetCPUBlockReportsEnabledReply
 *
 * @brief Enable/disable block level measurement reply.
 *
 * @member MONITOR_SET_CPU_BLOCK_REPORTS_ENABLED_REPLY
 * @member MonitorSetCPUProcessReportsEnabledReply
 */

/** @macro MONITOR_SET_CPU_BLOCK_REPORTS_ENABLED_REPLY */
#define MONITOR_SET_CPU_BLOCK_REPORTS_ENABLED_REPLY 39293 /* !-SIGNO(MonitorSetCPUProcessReportsEnabledReply)-! */

/**
 * @type MonitorCPUReportPoint
 * @brief Process report point ID for a scope of processes.
 *
 * @long The ID is returned in reply from
 * @link "" "MonitorSetCPUProcessReportpointReply"
 * to identify the report point.
 *
 */
typedef uint32_t MonitorCPUReportPoint;

/**
 * @signal MonitorSetCPUProcessReportpointRequest
 *
 * @brief Define process report point with a scope for process measurement
 * request.
 *
 * @long Several process report points can be defined with different
 * scopes, which together build a combined scope of processes.
 *
 * @long The first processes in each report will be the processes in
 * the scopes of all defined report points.
 *
 * @long Only if there is space for more processes, other processes
 * can appear in the report, outside the "other" entry.
 *
 * @long An ID is returned in the reply to identify this
 * profiling process report point. This allows selective clearing of
 * report points.
 *
 * @restriction Accepts only MonitorScopeType MONITOR_SCOPE_ID
 * (indicating a process or block) and MONITOR_SCOPE_NAME_PATTERN.
 * Note that a name pattern scope only finds processes which already
 * exist in the target, at the moment when this request is
 * received by the monitor.
 * With a name pattern, also system processes with corresponding names
 * will be selected for profile reporting, independent of the configuration
 * of which processes are considered to be system processes.
 *
 * @member MONITOR_SET_CPU_PROCESS_REPORTPOINT_REQUEST
 * @member MonitorCPUProcessReportpointInfo
 *
 * @seealso MonitorSetCPUProcessReportsEnabledRequest
 */

/** @macro MONITOR_SET_CPU_PROCESS_REPORTPOINT_REQUEST */
#define MONITOR_SET_CPU_PROCESS_REPORTPOINT_REQUEST 39262 /* !-SIGNO(MonitorCPUProcessReportpointInfo)-! */

/**
 * @type MonitorCPUProcessReportpointInfo
 *
 * @field sigNo             MONITOR_SET_CPU_PROCESS_REPORTPOINT_REQUEST
 *                          MONITOR_GET_CPU_PROCESS_REPORTPOINT_INFO_REPLY.
 * @field reserved0         NOT USED. Should be 0.
 * @field scope             Scope of processes to measure.
 * @output id               ID of the report point.
 * @field tagsCount         Number of elements in tags.
 * @field tags              List of tags and their values.
 */
typedef struct MonitorCPUProcessReportpointInfo
{
   uint32_t sigNo;
   uint32_t reserved0;
   MonitorScope scope;
   MonitorCPUReportPoint id;
   uint32_t tagsCount;
   MonitorTag tags[1];
} MonitorCPUProcessReportpointInfo;

/**
 * @signal MonitorSetCPUProcessReportpointReply
 *
 * @brief Define process report point with a scope for process measurement
 * reply.
 *
 * @member MONITOR_SET_CPU_PROCESS_REPORTPOINT_REPLY
 * @member MonitorSetCPUProcessReportpointReply
 */

/** @macro MONITOR_SET_CPU_PROCESS_REPORTPOINT_REPLY */
#define MONITOR_SET_CPU_PROCESS_REPORTPOINT_REPLY 39263 /* !-SIGNO(MonitorSetCPUProcessReportpointReply)-! */

/**
 * @type MonitorSetCPUProcessReportpointReply
 *
 * @output sigNo             MONITOR_SET_CPU_PROCESS_REPORTPOINT_REPLY.
 * @output status            MONITOR_STATUS_OK or cause of failure.
 * @output id                ID of the report point.
 */
typedef struct MonitorSetCPUProcessReportpointReply
{
   uint32_t sigNo;
   MonitorStatus status;
   MonitorCPUReportPoint id;
} MonitorSetCPUProcessReportpointReply;

/**
 * @signal MonitorClearCPUProcessReportpointRequest
 *
 * @brief Clear process report points for process measurement request.
 *
 * @member MONITOR_CLEAR_CPU_PROCESS_REPORTPOINT_REQUEST
 * @member MonitorClearCPUProcessReportpointRequest
 */

/** @macro MONITOR_CLEAR_CPU_PROCESS_REPORTPOINT_REQUEST */
#define MONITOR_CLEAR_CPU_PROCESS_REPORTPOINT_REQUEST 39264 /* !-SIGNO(MonitorClearCPUProcessReportpointRequest)-! */

/**
 * @type MonitorClearCPUProcessReportpointRequest
 *
 * @input sigNo             MONITOR_CLEAR_CPU_PROCESS_REPORTPOINT_REQUEST.
 * @input id                The report point to clear (or ~0 to clear all).
 */
typedef struct MonitorClearCPUProcessReportpointRequest
{
   uint32_t sigNo;
   MonitorCPUReportPoint id;
} MonitorClearCPUProcessReportpointRequest;

/**
 * @signal MonitorClearCPUProcessReportpointReply
 *
 * @brief Clear process report points for process measurement reply.
 *
 * @member MONITOR_CLEAR_CPU_PROCESS_REPORTPOINT_REPLY
 * @member MonitorClearCPUProcessReportpointReply
 */

/** @macro MONITOR_CLEAR_CPU_PROCESS_REPORTPOINT_REPLY */
#define MONITOR_CLEAR_CPU_PROCESS_REPORTPOINT_REPLY 39265 /* !-SIGNO(MonitorClearCPUProcessReportpointReply)-! */

/**
 * @type MonitorClearCPUProcessReportpointReply
 *
 * @output sigNo             MONITOR_CLEAR_CPU_PROCESS_REPORTPOINT_REPLY.
 * @output status            MONITOR_STATUS_OK or cause of failure.
 */
typedef struct MonitorClearCPUProcessReportpointReply
{
   uint32_t sigNo;
   MonitorStatus status;
} MonitorClearCPUProcessReportpointReply;

/**
 * @signal MonitorGetCPUProcessReportpointInfoRequest
 *
 * @brief List process report points with scopes request.
 *
 * @long Followed by zero or more MonitorGetCPUProcessReportpointInfoReply,
 * MonitorBlockingFlowControlRequest and one
 * MonitorGetCPUProcessReportpointInfoEndmark (with status).
 *
 * @member MONITOR_GET_CPU_PROCESS_REPORTPOINT_INFO_REQUEST
 * @member MonitorGetCPUProcessReportpointInfoRequest
 *
 * @seealso MonitorGetCPUProcessReportpointInfoReply
 * @seealso MonitorBlockingFlowControlRequest
 */

/** @macro MONITOR_GET_CPU_PROCESS_REPORTPOINT_INFO_REQUEST */
#define MONITOR_GET_CPU_PROCESS_REPORTPOINT_INFO_REQUEST 39266 /* !-SIGNO(MonitorGetCPUProcessReportpointInfoRequest)-! */

/**
 * @type MonitorGetCPUProcessReportpointInfoRequest
 *
 * @input sigNo MONITOR_GET_CPU_PROCESS_REPORTPOINT_INFO_REQUEST.
 */
typedef struct MonitorGetCPUProcessReportpointInfoRequest
{
   uint32_t sigNo;
} MonitorGetCPUProcessReportpointInfoRequest;

/**
 * @signal MonitorGetCPUProcessReportpointInfoReply
 *
 * @brief List process report points with scopes reply.
 *
 * @long This signal is sent under blocking flow control.
 *
 * @member MONITOR_GET_CPU_PROCESS_REPORTPOINT_INFO_REPLY
 * @member MonitorCPUProcessReportpointInfo
 */

/** @macro MONITOR_GET_CPU_PROCESS_REPORTPOINT_INFO_REPLY */
#define MONITOR_GET_CPU_PROCESS_REPORTPOINT_INFO_REPLY 39267 /* !-SIGNO(MonitorCPUProcessReportpointInfo)-! */

/**
 * @signal MonitorGetCPUProcessReportpointInfoEndmark
 *
 * @brief List process report point with scopes endmark.
 *
 * @member MONITOR_GET_CPU_PROCESS_REPORTPOINT_INFO_ENDMARK
 * @member MonitorGetCPUProcessReportpointInfoEndmark
 */

/** @macro MONITOR_GET_CPU_PROCESS_REPORTPOINT_INFO_ENDMARK */
#define MONITOR_GET_CPU_PROCESS_REPORTPOINT_INFO_ENDMARK 39268 /* !-SIGNO(MonitorGetCPUProcessReportpointInfoEndmark)-! */

/**
 * @type MonitorGetCPUProcessReportpointInfoEndmark
 *
 * @output sigNo MONITOR_GET_CPU_PROCESS_REPORTPOINT_INFO_ENDMARK.
 * @output status MONITOR_STATUS_OK or cause of failure.
 */
typedef struct MonitorGetCPUProcessReportpointInfoEndmark
{
   uint32_t sigNo;
   MonitorStatus status;
} MonitorGetCPUProcessReportpointInfoEndmark;


/*******************************************************************************
 *
 * DATA SIGNALS
 *
 ******************************************************************************/

/**
 * @signal MonitorGetCPUReportsRequest
 *
 * @brief Get CPU load measurement reports request.
 *
 * @long Requests are made for a number of reports that have a later timestamp
 * (tick and microtick), but not equal, than that specified.
 *
 * @member MONITOR_GET_CPU_REPORTS_REQUEST
 * @member MonitorGetCPUReportsRequest
 */

/** @macro MONITOR_GET_CPU_REPORTS_REQUEST */
#define MONITOR_GET_CPU_REPORTS_REQUEST 39270 /* !-SIGNO(MonitorGetCPUReportsRequest)-! */

/**
 * @type MonitorGetCPUReportsRequest
 *
 * @input sigNo             MONITOR_GET_CPU_REPORTS_REQUEST.
 * @input reserved0         (euId) Execution unit for which to get reports.
 *                          If MONITOR_ALL_EXECUTION_UNITS is given the
 *                          sum of the execution times over the execution IDs
 *                          are reported.
 * @input time              Timestamp of the latest report the client has
 *                          received. Set to { 0 , 0 } to get all.
 */
typedef struct MonitorGetCPUReportsRequest
{
   uint32_t sigNo;
   uint32_t reserved0;
   MonitorCPUTimestamp time;
} MonitorGetCPUReportsRequest;

/**
 * @signal MonitorGetCPUReportsReply
 *
 * @brief Get CPU load measurement reports reply.
 *
 * @member MONITOR_GET_CPU_REPORTS_REPLY
 * @member MonitorGetCPUReportsReply
 */

/** @macro MONITOR_GET_CPU_REPORTS_REPLY */
#define MONITOR_GET_CPU_REPORTS_REPLY 39271 /* !-SIGNO(MonitorGetCPUReportsReply)-! */

/**
 * @type MonitorGetCPUReportsReply
 *
 * @output sigNo             MONITOR_GET_CPU_REPORTS_REPLY.
 * @output status            MONITOR_STATUS_OK or cause of failure.
 * @output enabled           1 if load info enabled otherwise zero.
 * @output euId              Execution unit or MONITOR_ALL_EXECUTION_UNITS.
 * @output reserved0         NOT USED.
 * @output reserved1         NOT USED.
 * @output reportsCount      Number of elements in list of reports.
 * @output reports           List of CPU load reports.
 *
 * @seealso MONITOR_ALL_EXECUTION_UNITS
 */
typedef struct MonitorGetCPUReportsReply
{
   uint32_t sigNo;
   MonitorStatus status;
   MonitorBoolean enabled;
   uint16_t euId;
   uint16_t reserved0;
   uint32_t reserved1;
   uint32_t reportsCount;
   MonitorCPUReport reports[1];
} MonitorGetCPUReportsReply;

/**
 * @signal MonitorGetCPUPriReportsRequest
 *
 * @brief Get priority load measurement reports request.
 *
 * @long Requests are made for a number of reports that have a later timestamp
 * (tick and microtick), but not equal, than that specified.
 *
 * @member MONITOR_GET_CPU_PRI_REPORTS_REQUEST
 * @member MonitorGetCPUPriReportsRequest
 */

/** @macro MONITOR_GET_CPU_PRI_REPORTS_REQUEST */
#define MONITOR_GET_CPU_PRI_REPORTS_REQUEST 39272 /* !-SIGNO(MonitorGetCPUPriReportsRequest)-! */

/**
 * @type MonitorGetCPUPriReportsRequest
 *
 * @input sigNo             MONITOR_GET_CPU_PRI_REPORTS_REQUEST.
 * @input reserved0         (euId) Execution unit for which to get reports.
 *                          If MONITOR_ALL_EXECUTION_UNITS is given the
 *                          sum of the execution times over the execution IDs
 *                          are reported.
 * @input time              Timestamp of the latest report the client has
 *                          received. Set to { 0 , 0 } to get all.
 *
 */
typedef struct MonitorGetCPUPriReportsRequest
{
   uint32_t sigNo;
   uint32_t reserved0;
   MonitorCPUTimestamp time;
} MonitorGetCPUPriReportsRequest;

/**
 * @signal MonitorGetCPUPriReportsReply
 *
 * @brief Get priority load measurement reports reply.
 *
 * @member MONITOR_GET_CPU_PRI_REPORTS_REPLY
 * @member MonitorGetCPUPriReportsReply
 */

/** @macro MONITOR_GET_CPU_PRI_REPORTS_REPLY */
#define MONITOR_GET_CPU_PRI_REPORTS_REPLY 39273 /* !-SIGNO(MonitorGetCPUPriReportsReply)-! */

/**
 * @type MonitorGetCPUPriReportsReply
 *
 * @output sigNo             MONITOR_GET_CPU_PRI_REPORTS_REPLY.
 * @output status            MONITOR_STATUS_OK or cause of failure.
 * @output enabled           1 if load info enabled otherwise zero.
 * @output euId              Execution unit or MONITOR_ALL_EXECUTION_UNITS.
 * @output reserved0         NOT USED.
 * @output reserved1         NOT USED.
 * @output reportsCount      Number of elements in list of reports.
 * @output reports           List of CPU pri load reports.
 *
 * @seealso MONITOR_ALL_EXECUTION_UNITS
 */
typedef struct MonitorGetCPUPriReportsReply
{
   uint32_t sigNo;
   MonitorStatus status;
   MonitorBoolean enabled;
   uint16_t euId;
   uint16_t reserved0;
   uint32_t reserved1;
   uint32_t reportsCount;
   MonitorCPUPriReport reports[1];
} MonitorGetCPUPriReportsReply;

/**
 * @signal MonitorGetCPUProcessReportsRequest
 *
 * @brief Get process load measurement reports request.
 *
 * @long Requests are made for a number of reports that have a later timestamp
 * (tick and microtick), but not equal, than that specified.
 *
 * @member MONITOR_GET_CPU_PROCESS_REPORTS_REQUEST
 * @member MonitorGetCPUProcessReportsRequest
 */

/** @macro MONITOR_GET_CPU_PROCESS_REPORTS_REQUEST */
#define MONITOR_GET_CPU_PROCESS_REPORTS_REQUEST 39274 /* !-SIGNO(MonitorGetCPUProcessReportsRequest)-! */

/**
 * @type MonitorGetCPUProcessReportsRequest
 *
 * @input sigNo             MONITOR_GET_CPU_PROCESS_REPORTS_REQUEST.
 * @input reserved0         (euId) Execution unit for which to get reports.
 *                          If MONITOR_ALL_EXECUTION_UNITS is given the
 *                          sum of the execution times over the execution IDs
 *                          are reported.
 * @input time              Timestamp of the latest report the client has
 *                          received. Set to { 0 , 0 } to get all.
 */
typedef struct MonitorGetCPUProcessReportsRequest
{
   uint32_t sigNo;
   uint32_t reserved0;
   MonitorCPUTimestamp time;
} MonitorGetCPUProcessReportsRequest;

/**
 * @signal MonitorGetCPUProcessReportsReply
 *
 * @brief Get process load measurement reports reply.
 *
 * @long This reply contains a variable number of
 * @link "" "MonitorCPUProcessReport" structs with a variable number of
 * @link "" "MonitorCPUProcess" structs each. The example below
 * shows how to iterate over all the data.
 *
 * @member MONITOR_GET_CPU_PROCESS_REPORTS_REPLY
 * @member MonitorGetCPUProcessReportsReply
 *
 * @example
 * MonitorGetCPUProcessReportsReply *reply;
 *
 * // Get info... //
 *
 * MonitorCPUProcess *p  = reply->processes;
 * MonitorCPUProcess *p1 = reply->processes + reply->processesCount;
 *
 * // For each report //
 * while (p < p1)
 * {
 *    MonitorCPUProcessReport *report = (void *)p;
 *    MonitorCPUProcess *pn = report->processes + report->processesCount;
 *
 *    // For each process //
 *    for (p = report->processes; p < pn; p++)
 *    {
 *       printf("Process: %lx Load: %lx%%\n",
 *               p->id, 100 * p->sum / report->interval);
 *    }
 * }
 */

/** @macro MONITOR_GET_CPU_PROCESS_REPORTS_REPLY */
#define MONITOR_GET_CPU_PROCESS_REPORTS_REPLY 39275 /* !-SIGNO(MonitorGetCPUProcessReportsReply)-! */

/**
 * @type MonitorGetCPUProcessReportsReply
 *
 * @output sigNo             MONITOR_GET_CPU_PROCESS_REPORTS_REPLY.
 * @output status            MONITOR_STATUS_OK or cause of failure.
 * @output enabled           1 if load info enabled otherwise zero.
 * @output euId              Execution unit or MONITOR_ALL_EXECUTION_UNITS.
 * @output reserved0         NOT USED.
 * @output reserved1         NOT USED.
 * @output reserved2         NOT USED.
 * @output reportsCount      Number of reports in the processes list.
 * @output processesCount    Length of processes (in MonitorCPUProcess structs).
 * @output processes         List of MonitorCPUProcessReport and their
 *                           MonitorCPUProcess data.
 *
 * @seealso MONITOR_ALL_EXECUTION_UNITS
 */
typedef struct MonitorGetCPUProcessReportsReply
{
   uint32_t sigNo;
   MonitorStatus status;
   MonitorBoolean enabled;
   uint16_t euId;
   uint16_t reserved0;
   uint32_t reserved1;
   uint32_t reserved2;
   uint32_t reportsCount;
   uint32_t processesCount;
   MonitorCPUProcess processes[1];
} MonitorGetCPUProcessReportsReply;

/**
 * @signal MonitorGetCPUBlockReportsRequest
 *
 * @brief Get block load measurement reports request.
 *
 * @long Requests are made for a number of reports that have a later timestamp
 * (tick and microtick), but not equal, than that specified.
 *
 * @member MONITOR_GET_CPU_BLOCK_REPORTS_REQUEST
 * @member MonitorGetCPUProcessReportsRequest
 */

/** @macro MONITOR_GET_CPU_BLOCK_REPORTS_REQUEST */
#define MONITOR_GET_CPU_BLOCK_REPORTS_REQUEST 39276 /* !-SIGNO(MonitorGetCPUProcessReportsRequest)-! */


/**
 * @signal MonitorGetCPUBlockReportsReply
 *
 * @brief Get block load measurement reports reply.
 *
 * @member MONITOR_GET_CPU_BLOCK_REPORTS_REPLY
 * @member MonitorGetCPUProcessReportsReply
 *
 */

/** @macro MONITOR_GET_CPU_BLOCK_REPORTS_REPLY */
#define MONITOR_GET_CPU_BLOCK_REPORTS_REPLY 39277 /* !-SIGNO(MonitorGetCPUProcessReportsReply)-! */


/*******************************************************************************
 *
 * USER REPORT SIGNALS
 *
 ******************************************************************************/

/**
 * @signal MonitorGetUserReportsEnabledRequest
 *
 * @brief Get status of user reports measurement request.
 *
 * @member MONITOR_GET_USER_REPORTS_ENABLED_REQUEST
 * @member MonitorGetUserReportsEnabledRequest
 */

/** @macro MONITOR_GET_USER_REPORTS_ENABLED_REQUEST */
#define MONITOR_GET_USER_REPORTS_ENABLED_REQUEST 39278 /* !-SIGNO(MonitorGetUserReportsEnabledRequest)-! */

/**
 * @type MonitorGetUserReportsEnabledRequest
 *
 * @input sigNo             MONITOR_GET_USER_REPORTS_ENABLED_REQUEST.
 * @input reportNo          User report identifier.
 * @input reserved0         NOT USED.
 */
typedef struct MonitorGetUserReportsEnabledRequest
{
   uint32_t sigNo;
   uint32_t reportNo;
   uint32_t reserved0;
} MonitorGetUserReportsEnabledRequest;

/**
 * @signal MonitorGetUserReportsEnabledReply
 *
 * @brief Get status of user reports measurement reply.
 *
 * @member MONITOR_GET_USER_REPORTS_ENABLED_REPLY
 * @member MonitorGetUserReportsEnabledReply
 */

/** @macro MONITOR_GET_USER_REPORTS_ENABLED_REPLY */
#define MONITOR_GET_USER_REPORTS_ENABLED_REPLY 39279 /* !-SIGNO(MonitorGetUserReportsEnabledReply)-! */

/**
 * @type MonitorGetUserReportsEnabledReply
 *
 * @output sigNo             MONITOR_GET_USER_REPORTS_ENABLED_REPLY.
 * @output reportNo          User report identifier.
 * @output enabled           1 if user report is enabled otherwise zero
 * @output interval          Integration interval in ticks.
 * @output maxReports        The currently configured maximum number of
 *                           reports in the target.
 * @output maxValuesReport   The currently configured maximum number
 *                           of values per report.
 * @output reserved0         NOT USED.
 * @output sec               Number of seconds since 00:00:00 01 Jan 1970 GMT
 *                           at the system time sectick:secntick, or zero
 *                           if not available.
 * @output sectick           Tick count at time sec.
 * @output secntick          Number of timer steps at time sec.
 */

typedef struct MonitorGetUserReportsEnabledReply
{
   uint32_t sigNo;
   uint32_t reportNo;
   MonitorBoolean enabled;
   uint32_t interval;
   uint32_t maxReports;
   uint32_t maxValuesReport;
   uint32_t reserved0;
   uint32_t sec;
   uint32_t sectick;
   uint32_t secntick;
} MonitorGetUserReportsEnabledReply;

/**
 * @signal MonitorSetUserReportsEnabledRequest
 *
 * @brief Enable/disable user value measurement for a user report number request.
 *
 * @long User value profiling is enabled/disabled for the given
 * user report number. If defined, this can contain multiple
 * values.
 *
 * @long Only registered user report numbers are accepted, otherwise an
 * error status is set in the reply.
 *
 * @member MONITOR_SET_USER_REPORTS_ENABLED_REQUEST
 * @member MonitorSetUserReportsEnabledRequest
 */

/** @macro MONITOR_SET_USER_REPORTS_ENABLED_REQUEST */
#define MONITOR_SET_USER_REPORTS_ENABLED_REQUEST 39280 /* !-SIGNO(MonitorSetUserReportsEnabledRequest)-! */

/**
 * @type MonitorSetUserReportsEnabledRequest
 *
 * @input sigNo             MONITOR_SET_USER_REPORTS_ENABLED_REQUEST.
 * @input reportNo          User report identifier.
 * @input enabled           1 if report should be enabled otherwise zero.
 * @input interval          Integration interval in ticks.
 * @input maxReports        The suggested maximum number of reports kept in
 *                          target in a circular buffer. If set to zero or
 *                          larger than supported by the monitor, the
 *                          configured default value is used instead.
 * @input maxValuesReport   The suggested maximum number of values per report.
 *                          If larger than supported by the monitor, the
 *                          configured default value is used instead.
 * @input reserved0         NOT USED.
 */
typedef struct MonitorSetUserReportsEnabledRequest
{
   uint32_t sigNo;
   uint32_t reportNo;
   MonitorBoolean enabled;
   uint32_t interval;
   uint32_t maxReports;
   uint32_t maxValuesReport;
   uint32_t reserved0;
} MonitorSetUserReportsEnabledRequest;

/**
 * @signal MonitorSetUserReportsEnabledReply
 *
 * @brief Enable/disable user value measurement for a user report number reply.
 *
 * @member MONITOR_SET_USER_REPORTS_ENABLED_REPLY
 * @member MonitorSetUserReportsEnabledReply
 */

/** @macro MONITOR_SET_USER_REPORTS_ENABLED_REPLY */
#define MONITOR_SET_USER_REPORTS_ENABLED_REPLY 39281 /* !-SIGNO(MonitorSetUserReportsEnabledReply)-! */

/**
 * @type MonitorSetUserReportsEnabledReply
 *
 * @output sigNo             MONITOR_SET_USER_REPORTS_ENABLED_REPLY.
 * @output status            MONITOR_STATUS_OK or cause of failure.
 */
typedef struct MonitorSetUserReportsEnabledReply
{
   uint32_t sigNo;
   MonitorStatus status;
} MonitorSetUserReportsEnabledReply;

/**
 * @signal MonitorGetUserReportsRequest
 *
 * @brief  Get user reports request.
 *
 * @long Request is made for user reports with a specific report number,
 * that have a later timestamp (tick and microtick) than,
 * but not equal to, the specified timestamp.
 *
 * @member MONITOR_GET_USER_REPORTS_REQUEST
 * @member MonitorGetUserReportsRequest
 */

/** @macro MONITOR_GET_USER_REPORTS_REQUEST */
#define MONITOR_GET_USER_REPORTS_REQUEST  39282 /* !-SIGNO(MonitorGetUserReportsRequest)-! */

/**
 * @type MonitorGetUserReportsRequest
 *
 * @input sigNo             MONITOR_GET_USER_REPORTS_REQUEST.
 * @input reportNo          User report identifier.
 * @input time              Timestamp of the latest report the client has
 *                          already received. Set to { 0 , 0 } to get all.
 * @input reserved0         NOT USED.
 */
typedef struct MonitorGetUserReportsRequest
{
   uint32_t sigNo;
   uint32_t reportNo;
   MonitorCPUTimestamp time;
   uint32_t reserved0;
} MonitorGetUserReportsRequest;

/**
 * @signal MonitorGetUserReportsReply
 *
 * @brief Get user report reply.
 *
 * @long The array reportValues specifies a memory area, seen
 * as an array of slots with slot size sizeof(MonitorUserReportValue).
 * The total number of slots is given in the reportValuesCount parameter.
 * The contents of this reports area can be a variable number of variable
 * sized reports of one of two types.
 * All structs follow immediately after each other.
 *
 * @long If maxmin is 0, the area contains variable sized
 * @link "" "MonitorUserReport" structs (each with parameters and a
 * variable number of @link "" "MonitorUserReportValue" structs).
 *
 * @long If maxmin is set (1), the area contains variable sized
 * @link "" "MonitorMaxMinUserReport" structs (each with parameters and a
 * variable number of @link "" "MonitorMaxMinUserReportValue" structs).
 *
 * @long The first example shows how to iterate over all the data,
 * if the reply contains maxmin = 0, thus MonitorUserReport structs.
 *
 * @long The second example shows the same but where minmax is 1, i.e.
 * the reply contains MonitorMaxMinUserReport structs. Note that the
 * memory area size still is counted in slots with
 * sizeof(MonitorUserReportValue).
 *
 * @member MONITOR_GET_USER_REPORTS_REPLY
 * @member MonitorGetUserReportsReply
 *
 * @example
 * // Example 1: With MonitorUserReport structs in the reply         //
 * //            e.g. maxmin == 0.                                   //
 *
 * MonitorGetUserReportsReply *reply;
 *
 * // Assume the reply is received                                   //
 *
 * // Establish the start and end of the memory area with reports    //
 * MonitorUserReportValue *p  = reply->reportValues;
 * MonitorUserReportValue *p1 = reply->reportValues + reply->reportValuesCount;
 *
 * // For each report, here containing a MonitorUserReport struct    //
 * while (p < p1)
 * {
 *    MonitorUserReport *report = (void *)p;
 *    MonitorUserReportValue *pn = report->reportValues +
 *                                 report->reportValuesCount;
 *
 *    printf("Interval: %u\n", report->interval);
 *
 *    // For each report value, here type MonitorUserReportValue,    //
 *    // each of which is contained on one "slot"                    //
 *    for (p = report->reportValues; p < pn; p++)
 *    {
 *        printf("ID: %u\n", p->id);
 *        printf("Sum: %d\n", p->sum);
 *    }
 * // p has been incremented to point to next report or after end    //
 * }
 *
 *
 *
 * // Example 2: With MonitorMaxMinUserReport structs in the reply   //
 * //            e.g. maxmin == 0.                                   //
 *
 * MonitorGetUserReportsReply *reply;
 *
 * // Assume the reply is received                                   //
 *
 * // Establish the start and end of the memory area with reports    //
 * MonitorMaxMinUserReportValue *p  = reply->reportValues;
 * MonitorMaxMinUserReportValue *p1 = reply->reportValues
 *                                    + reply->reportValuesCount;
 *
 * // For each report, containing a MonitorMaxMinUserReport struct   //
 * while (p < p1)
 * {
 *    MonitorMaxMinUserReport *report = (void *)p;
 *    MonitorMaxMinUserReportValue *pn = report->reportValues +
 *                                 report->reportValuesCount;
 *
 *
 *    printf("Interval: %u\n", report->interval);
 *
 *    // For each report value, type MonitorMaxMinUserReportValue,   //
 *    // each of which occupies sizeof(MonitorMaxMinUserReportValue) //
 *    for (p = report->reportValues; p < pn; p++)
 *    {
 *        printf("ID: %u\n", p->id);
 *        printf("Sum: %d\n", p->sum);
 *        printf("Max: %d\n", p->max);
 *        printf("Min: %d\n", p->min);
 *    }
 *    // p has been incremented to point to next report or after end    //
 * }
 */

/** @macro MONITOR_GET_USER_REPORTS_REPLY */
#define MONITOR_GET_USER_REPORTS_REPLY 39283 /* !-SIGNO(MonitorGetUserReportsReply)-! */

/**
 * @type MonitorGetUserReportsReply
 *
 * @output sigNo              MONITOR_GET_USER_REPORTS_REPLY.
 * @output status             MONITOR_STATUS_OK or cause of failure.
 * @output reportNo           User report identifier.
 * @output enabled            1 if user report is enabled otherwise zero.
 * @output continuous         Set to one to have the previous value copied to
 *                            the next interval. Set to zero to have the value
 *                            cleared at interval start.
 * @output maxmin             Set to one if set_report_val track the maximum
 *                            and minimum during each interval. Set to zero if
 *                            it tracks the current value only.
 * @output multiple           Set to one for a report class with multiple values
 *                            (for instance one per process). Set to zero for a
 *                            report class with a single value.
 * @output reserved0          NOT USED. Should be zero.
 * @output reserved1          NOT USED. Should be zero.
 * @output reportsCount       Number of reports in reportValues.
 * @output reportValuesCount  Length of reportValues
 *                            (in MonitorUserReportValue structs).
 * @output reportValues       List of MonitorUserReport and their
 *                            MonitorUserReportValue data if maxmin is set to
 *                            zero. List of MonitorMaxMinUserReport and their
 *                            MonitorMaxMinUserReportValue data if maxmin is
 *                            set.
 *
 */

typedef struct MonitorGetUserReportsReply
{
   uint32_t sigNo;
   MonitorStatus status;
   uint32_t reportNo;
   MonitorBoolean enabled;
   uint8_t continuous;
   uint8_t maxmin;
   uint8_t multiple;
   uint8_t reserved0;
   uint32_t reserved1;
   uint32_t reportsCount;
   uint32_t reportValuesCount;
   MonitorUserReportValue reportValues[1];
} MonitorGetUserReportsReply;


/* ALLBLOCKS additions by Anders to test */

/** @macro MONITOR_GET_TOTAL_CPU_BLOCK_REPORTS_ENABLED_REQUEST
 *  @brief
 */
#define MONITOR_GET_TOTAL_CPU_BLOCK_REPORTS_ENABLED_REQUEST 39284


/** @macro MONITOR_GET_TOTAL_CPU_BLOCK_REPORTS_ENABLED_REPLY
 *  @brief
 */
#define MONITOR_GET_TOTAL_CPU_BLOCK_REPORTS_ENABLED_REPLY 39285


/** @macro MONITOR_SET_TOTAL_CPU_BLOCK_REPORTS_ENABLED_REQUEST
 *  @brief
 */
#define MONITOR_SET_TOTAL_CPU_BLOCK_REPORTS_ENABLED_REQUEST 39286


/** @macro MONITOR_SET_TOTAL_CPU_BLOCK_REPORTS_ENABLED_REPLY
 *  @brief
 */
#define MONITOR_SET_TOTAL_CPU_BLOCK_REPORTS_ENABLED_REPLY 39287


/** @macro MONITOR_GET_TOTAL_CPU_BLOCK_REPORTS_REQUEST
 *  @brief
 */
#define MONITOR_GET_TOTAL_CPU_BLOCK_REPORTS_REQUEST 39288


/** @macro MONITOR_GET_TOTAL_CPU_BLOCK_REPORTS_REPLY
 *  @brief
 */
#define MONITOR_GET_TOTAL_CPU_BLOCK_REPORTS_REPLY 39289

/**
 * @signal MonitorGetProfilingExecutionUnitEnabledRequest
 *
 * @brief Get the enable/disable status of profiling for a specific core.
 *
 * @long All forms of CPU profiling may be disabled for a specific execution
 * unit (CPU core) and this signal requests the current status.
 *
 * @member MONITOR_GET_PROFILING_EXECUTION_UNIT_ENABLED_REQUEST
 * @member MonitorGetProfilingExecutionUnitEnabledRequest
 *
 * @seealso  MonitorGetProfilingExecutionUnitEnabledReply
 * @seealso  MonitorSetProfilingExecutionUnitEnabledRequest
 */

/** @macro MONITOR_GET_PROFILING_EXECUTION_UNIT_ENABLED_REQUEST */
#define MONITOR_GET_PROFILING_EXECUTION_UNIT_ENABLED_REQUEST  39450 /* !-SIGNO(MonitorGetProfilingExecutionUnitEnabledRequest)-! */

/**
 * @type MonitorGetProfilingExecutionUnitEnabledRequest
 *
 * @input sigNo             MONITOR_GET_PROFILING_EXECUTION_UNIT_ENABLED_REQUEST.
 * @input eu                CPU Execution unit (core)
 *                          if MONITOR_ALL_EXECUTION_UNITS is given it reports
 *                          true if it is possible to profile on all the
 *                          execution units.
 * @input reserved0         NOT USED must be set to 0.
 * @input reserved1         NOT USED must be set to 0.
 */
typedef struct MonitorGetProfilingExecutionUnitEnabledRequest
{
   uint32_t sigNo;
   uint32_t eu;
   uint32_t reserved0;
   uint32_t reserved1;
} MonitorGetProfilingExecutionUnitEnabledRequest;

/**
 * @signal MonitorGetProfilingExecutionUnitEnabledReply
 *
 * @brief Reply with enable/disable status of profiling for a specific core.
 *
 * @member MONITOR_GET_PROFILING_EXECUTION_UNIT_ENABLED_REPLY
 * @member MonitorGetProfilingExecutionUnitEnabledReply
 *
 * @seealso  MonitorGetProfilingExecutionUnitEnabledRequest
 * @seealso  MonitorSetProfilingExecutionUnitEnabledRequest
 */

/** @macro MONITOR_GET_PROFILING_EXECUTION_UNIT_ENABLED_REPLY */
#define MONITOR_GET_PROFILING_EXECUTION_UNIT_ENABLED_REPLY  39451 /* !-SIGNO(MonitorGetProfilingExecutionUnitEnabledReply)-! */

/**
 * @type MonitorGetProfilingExecutionUnitEnabledReply
 *
 * @output sigNo             MONITOR_GET_PROFILING_EXECUTION_UNIT_ENABLED_REPLY.
 * @output status            MONITOR_STATUS_OK or cause of failure.
 * @output enabled           True if it is possible to profile on this eu.
 * @output reserved0         NOT USED.
 */
typedef struct MonitorGetProfilingExecutionUnitEnabledReply
{
   uint32_t       sigNo;
   MonitorStatus  status;
   MonitorBoolean enabled;
   uint32_t       reserved0;
} MonitorGetProfilingExecutionUnitEnabledReply;

/**
 * @signal MonitorSetProfilingExecutionUnitEnabledRequest
 *
 * @brief Set the enable/disable status of profiling for a specific core.
 *
 * @long All forms of CPU profiling may be disabled for a specific execution
 * unit (CPU core) and this signal sets the status. By default profiling is
 * possible for all cores.
 *
 * @long The main reason to disable it, on a specific core, is to lower the
 * profiling overhead if you never will be interested in any CPU profiling on
 * that core.
 *
 * @member MONITOR_SET_PROFILING_EXECUTION_UNIT_ENABLED_REQUEST
 * @member MonitorSetProfilingExecutionUnitEnabledRequest
 *
 * @seealso  MonitorSetProfilingExecutionUnitEnabledReply
 * @seealso  MonitorGetProfilingExecutionUnitEnabledRequest
 */

/** @macro MONITOR_SET_PROFILING_EXECUTION_UNIT_ENABLED_REQUEST */
#define MONITOR_SET_PROFILING_EXECUTION_UNIT_ENABLED_REQUEST  39452 /* !-SIGNO(MonitorSetProfilingExecutionUnitEnabledRequest)-! */

/**
 * @type MonitorSetProfilingExecutionUnitEnabledRequest
 *
 * @input sigNo             MONITOR_SET_PROFILING_EXECUTION_UNIT_ENABLED_REQUEST.
 * @input eu                CPU Execution unit (core).
 *                          It is not possible to change status for
 *                          execution unit 0. If it is set at 0, the reply
 *                          signal status will be MONITOR_STATUS_INVALID_EU.
 *                          If MONITOR_ALL_EXECUTION_UNITS is given all
 *                          execution units, except execution unit 0 are
 *                          affected.
 * @input enabled           True if profiling should be enabled.
 *                          Currently it is only possible to disable cores.
 *                          If enabled is true, reply signal will return
 *                          with status MONITOR_STATUS_REQUEST_INVALID.
 * @input reserved0         NOT USED must be set to 0.
 */
typedef struct MonitorSetProfilingExecutionUnitEnabledRequest
{
   uint32_t       sigNo;
   uint32_t       eu;
   MonitorBoolean enabled;
   uint32_t       reserved0;
} MonitorSetProfilingExecutionUnitEnabledRequest;

/**
 * @signal MonitorSetProfilingExecutionUnitEnabledReply
 *
 * @brief Reply to MonitorSetProfilingExecutionUnitEnabledRequest.
 *
 * @member MONITOR_SET_PROFILING_EXECUTION_UNIT_ENABLED_REPLY
 * @member MonitorSetProfilingExecutionUnitEnabledReply
 *
 * @seealso  MonitorSetProfilingExecutionUnitEnabledRequest
 * @seealso  MonitorGetProfilingExecutionUnitEnabledRequest
 */

/** @macro MONITOR_SET_PROFILING_EXECUTION_UNIT_ENABLED_REPLY */
#define MONITOR_SET_PROFILING_EXECUTION_UNIT_ENABLED_REPLY  39453 /* !-SIGNO(MonitorSetProfilingExecutionUnitEnabledReply)-! */

/**
 * @type MonitorSetProfilingExecutionUnitEnabledReply
 *
 * @output sigNo             MONITOR_SET_PROFILING_EXECUTION_UNIT_ENABLED_REPLY.
 * @output status            MONITOR_STATUS_OK or cause of failure.
 * @output reserved0         NOT USED.
 * @output reserved1         NOT USED.
 */
typedef struct MonitorSetProfilingExecutionUnitEnabledReply
{
   uint32_t       sigNo;
   MonitorStatus  status;
   uint32_t       reserved0;
   uint32_t       reserved1;
} MonitorSetProfilingExecutionUnitEnabledReply;

#ifdef __cplusplus
} /* end of extern "C" */
#endif

#endif
