/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2006 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

/**
 * @toc OSE:
 * @file ose_spi/monitor_dumps.sig
 *
 * @brief OSE System RAM dump access signal interface.
 *
 * @long This header file defines signals and types related to dumps. It
 * provides mechanisms through which the user can retrieve data that was
 * dumped by, for example, the RAM dump component or via FMD (freeze mode
 * dumper, usually via the system error handler).
 *
 * @long The signals in this file use the flow control mechanism described in
 * monitor.sig.
 *
 * @long A connection to the RMM is needed, see monitor.sig for the actual
 * connection procedure using @link "" "MonitorInterfaceRequest" and
 * @link "" "MonitorConnectRequest".
 *
 * @long It is not needed to attach a scope before using the actions in
 * this header file.
 *
 * @long When ready, do disconnect with 
 * @link "" "MonitorConnectRequest", allowing the monitor to
 * deallocate resources linked to the client.
 *
 * @long See also RMM Run Mode Monitor and the FMD Freeze Mode Dumper in
 * in the OSE Core User's Guide.
 *
 * @seealso ose_spi/monitor.sig
 */

#ifndef _MONITOR_DUMPS_SIG
#define _MONITOR_DUMPS_SIG

#include "stdint.h"
#include "monitor.sig"

#ifdef __cplusplus
extern "C" {
#endif

/*******************************************************************************
 *
 * DUMP RETRIEVAL SIGNALS
 *
 ******************************************************************************/

/**
 * @signal MonitorGetRamdumpInfoRequest
 *
 * @brief List RAM dumps request.
 *
 * @long Dumps are listed in reverse chronological order,
 * i.e. most recent dump first.
 * 
 * @long Available dumps are numbered from 0 to N - 1, 
 * where 0 is the most recent dump and N is the total number of dumps.
 * 
 * @long Followed by zero or more MonitorGetRamdumpInfoReply,
 * MonitorBlockingFlowControlRequest and one MonitorGetRamdumpInfoEndmark.
 *
 * @member MONITOR_GET_RAMDUMP_INFO_REQUEST
 * @member MonitorGetRamdumpInfoRequest
 *
 * @seealso MonitorBlockingFlowControlRequest
 * @seealso MonitorGetRamdumpInfoReply
 * @seealso MonitorGetRamdumpInfoEndmark
 */

/** @macro MONITOR_GET_RAMDUMP_INFO_REQUEST */
#define MONITOR_GET_RAMDUMP_INFO_REQUEST 39500 /* !-SIGNO(MonitorGetRamdumpInfoRequest)-! */

/**
 * @type MonitorGetRamdumpInfoRequest
 *
 * @input sigNo  MONITOR_GET_RAMDUMP_INFO_REQUEST
 * @input from   The index of the first dump to get info for. If the number 
 *               is greater than N, no dump info replies will be returned.
 * @input to     The index of the last dump to get info for + 1. If the number
 *               is greater than N, the monitor will return info for dumps up
 *               to the end of the dump list.
 */
typedef struct MonitorGetRamdumpInfoRequest
{
   uint32_t sigNo;
   uint32_t from;
   uint32_t to;
} MonitorGetRamdumpInfoRequest;

/**
 * @signal MonitorGetRamdumpInfoReply
 *
 * @brief List RAM dumps reply.
 *
 * @long Reply to MonitorGetRamdumpInfoRequest. This signal is sent
 * under blocking flow control.
 *
 * @member MONITOR_GET_RAMDUMP_INFO_REPLY
 * @member MonitorGetRamdumpInfoReply
 *
 * @seealso MonitorBlockingFlowControlRequest
 * @seealso MonitorGetRamdumpInfoEndmark
 * @seealso MonitorGetRamdumpInfoRequest
 */

/* @macro MONITOR_GET_RAMDUMP_INFO_REPLY */
#define MONITOR_GET_RAMDUMP_INFO_REPLY 39501 /* !-SIGNO(MonitorGetRamdumpInfoReply)-! */

/**
 * @type MonitorGetRamdumpInfoReply
 *
 * @output sigNo     MONITOR_GET_RAMDUMP_INFO_REPLY
 * @output dumpId    Dump id
 * @output size      Size of dump in bytes
 * @output sec       Seconds since Jan 1, 1970, 00:00:00
 * @output microsec  Additional microseconds
 */
typedef struct MonitorGetRamdumpInfoReply
{
   uint32_t sigNo;
   uint32_t dumpId;
   uint32_t size;
   uint32_t sec;
   uint32_t microsec;
} MonitorGetRamdumpInfoReply;

/**
 * @signal MonitorGetRamdumpInfoEndmark
 *
 * @brief List dumps end mark with status.
 *
 * @member MONITOR_GET_RAMDUMP_INFO_ENDMARK
 * @member MonitorGetRamdumpInfoEndmark
 *
 * @seealso MonitorBlockingFlowControlRequest
 * @seealso MonitorGetRamdumpInfoReply
 * @seealso MonitorGetRamdumpInfoRequest
 */

/* @macro MONITOR_GET_RAMDUMP_INFO_ENDMARK */
#define MONITOR_GET_RAMDUMP_INFO_ENDMARK 39502 /* !-SIGNO(MonitorGetRamdumpInfoEndmark)-! */

/**
 * @type MonitorGetRamdumpInfoEndmark
 *
 * @output sigNo   MONITOR_GET_RAMDUMP_INFO_ENDMARK
 * @output status  MONITOR_STATUS_OK or cause of error
 */
typedef struct MonitorGetRamdumpInfoEndmark
{
   uint32_t sigNo;
   MonitorStatus status;
} MonitorGetRamdumpInfoEndmark;

/**
 * @signal MonitorGetRamdumpRequest
 *
 * @brief Get RAM dump request. Followed by zero or more MonitorGetRamdumpReply,
 * MonitorBlockingFlowControlRequest and one MonitorGetRamdumpEndmark.
 *
 * @member MONITOR_GET_RAMDUMP_REQUEST
 * @member MonitorGetRamdumpRequest
 *
 * @seealso MonitorBlockingFlowControlRequest
 * @seealso MonitorGetRamdumpReply
 * @seealso MonitorGetRamdumpEndmark
 */

/** @macro MONITOR_GET_RAMDUMP_REQUEST */
#define MONITOR_GET_RAMDUMP_REQUEST 39503 /* !-SIGNO(MonitorGetRamdumpRequest)-! */

/**
 * @type MonitorGetRamdumpRequest
 *
 * @input sigNo   MONITOR_GET_RAMDUMP_REQUEST
 * @input dumpId  Dump id, as retrieved by 
 *                @link "" "MonitorGetRamdumpInfoRequest".
 */
typedef struct MonitorGetRamdumpRequest
{
   uint32_t sigNo;
   uint32_t dumpId;
} MonitorGetRamdumpRequest;

/**
 * @signal MonitorGetRamdumpReply
 *
 * @brief Get RAM dump reply.
 *
 * @long This signal is sent under blocking flow control.
 *
 * @member MONITOR_GET_RAMDUMP_REPLY
 * @member MonitorGetRamdumpReply
 *
 * @seealso MonitorGetRamdumpEndmark
 * @seealso MonitorGetRamdumpRequest
 */

/** @macro MONITOR_GET_RAMDUMP_REPLY */
#define MONITOR_GET_RAMDUMP_REPLY 39504 /* !-SIGNO(MonitorGetRamdumpReply)-! */

/**
 * @type MonitorGetRamdumpReply
 *
 * @output sigNo      MONITOR_GET_RAMDUMP_REPLY
 * @output dumpId     Dump id, as retrieved by 
 *                    @link "" "MonitorGetRamdumpInfoRequest".
 * @output offset     Dump offset of the first byte in data.
 * @output dataCount  Number of bytes in data.
 * @output data       The data.
 */
typedef struct MonitorGetRamdumpReply
{
   uint32_t sigNo;
   uint32_t dumpId;
   uint32_t offset;
   uint32_t dataCount;
   uint8_t data[1];
} MonitorGetRamdumpReply;

/**
 * @signal MonitorGetRamdumpEndmark
 *
 * @brief Get RAM dump endmark with status.
 *
 * @member MONITOR_GET_RAMDUMP_ENDMARK
 * @member MonitorGetRamdumpEndmark
 *
 * @seealso MonitorGetRamdumpReply
 * @seealso MonitorGetRamdumpRequest
 */

/** @macro MONITOR_GET_RAMDUMP_ENDMARK */
#define MONITOR_GET_RAMDUMP_ENDMARK 39505 /* !-SIGNO(MonitorGetRamdumpEndmark)-! */

/**
 * @type MonitorGetRamdumpEndmark
 *
 * @output sigNo MONITOR_GET_RAMDUMP_REQUEST
 * @output status MONITOR_STATUS_OK or cause of error
 */
typedef struct MonitorGetRamdumpEndmark
{
   uint32_t sigNo;
   MonitorStatus status;
} MonitorGetRamdumpEndmark;

#ifdef __cplusplus
} /* end of extern "C" */
#endif

#endif
