/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2005-2006 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

/**
 * @toc OSE:
 * @file ose_spi/monitor_events.sig
 *
 * @brief OSE System event action/notification debug signal interface.
 *
 * @long This header file defines the signal interface between OSE debug
 * clients and OSE debug monitors (e.g. FMM and RMM) with regard to
 * event actionpoints.
 *
 * @long A connection to the RMM is needed, see monitor.sig for the actual
 * connection procedure using @link "" "MonitorInterfaceRequest" and
 * @link "" "MonitorConnectRequest".
 *
 * @long The client must also attach a scope by sending a
 * @link "" "MonitorAttachRequest" signal to the monitor,
 * before using this execution control interface. A main scope and
 * an intercept scope is defined.
 *
 * @long The scope tells the monitor what parts of the system the client
 * is interested in. In addition any scope given in the event requests will
 * automatically be reduced to be within the attached main scope,
 * if this feature is supported by the monitor.
 * Each client must attach for a non-overlapping scope.
 *
 * @long Actionpoints are created by clients. The event system in the
 * monitor keeps track of events and performs actions when events occur.
 * Each actionpoint defines what action to perform when an event occurs.
 * Filters limit actions to suitable ranges of events, e.g.
 * from a scope of processes or within a range of event numbers etc.
 * It can be defined that an event shall occur a certain number of times
 * before each time it triggers the action.
 *
 * @long All actionpoints are active at event state 0. Each actionpoint
 * can either be set always active or to be active only at one defined
 * non-zero event state (in addition to state 0). A request can be used to
 * set the event state to any value and actions in actionpoints can change
 * the event state. The initial event state is 0, when a system is started,
 * unless configured otherwise in the target system.
 *
 * @long Event actions can automatically switch between client-defined event
 * states to create a chain of actions and/or send event notification signals
 * to the clients, just trace the events and/or even intercept a process at
 * an event. It can also call a user defined event action callback function,
 * when action type MONITOR_ACTION_USER (@link "" "MonitorActionValues")
 * is requested for an event.
 *
 * @long Multiple actionpoints can be defined for the same event, but only
 * one action per actionpoint. If multiple actionpoints are triggered by
 * a single event, the order of performing the actions are as follows:
 * Trace actions are performed in the order "trace-on", "trace", "trace-off".
 * Actions of type user action, trace and notify are always performed
 * before the action intercept. Otherwise the order is undefined.
 *
 * @long Actions of the type intercept notification will both intercept a
 * scope of processes and send an intercept notifications to the client with
 * information.
 *
 * @long There are predefined events, but applications can also trigger user
 * events, using @link "ose.h" "create_event" and @link "ose.h" "signal_event".
 * User events are one type of event (MONITOR_EVENT_USER) and contain
 * event numbers to identify the events and event data.
 * Actually a range of user event numbers are reserved for OSE,
 * given in ose_events.h, e.g. see @link "ose_events.h" "OSE_EVENTS_ENEA_END".
 * For the OSE heap, a few special user events are defined,
 * e.g. @link "heap_events.h" "HEAP_EVENT_MALLOC".
 *
 * @long User event numbers should be declared in header files using a C comment
 * with the following syntax:  / * !-EVTNO(....)-! * / which can be decoded
 * with the sigdb host tool and used to enable a debugger to interpret and
 * show the contents of the event data.
 *
 * @long For configuration, see monitor.sig and the "OSE Core User's Guide".
 *
 * @long The following requests and replies are deprecated, currently kept
 * only for backwards compatibility and not described in the reference
 * documentation:
 *        MonitorGetEventActionpointInfoRequest,
 *        MonitorGetSendActionpointInfoReply,
 *        MonitorGetReceiveActionpointInfoReply,
 *        MonitorGetSwapActionpointInfoReply,
 *        MonitorGetCreateActionpointInfoReply,
 *        MonitroGetKillActionpointInfoReply,
 *        MonitorGetErrorActionpointInfoReply,
 *        MonitorGetEventActionpointInfoEndmark.
 *
 * @restriction In current implementation the client needs to attach for
 * exclusive (LOCKED) access to the event system before using event
 * actionpoints and thus the event state and the actionpoints are global.
 * This is not needed if the monitor reports the feature code
 * MONITOR_FEATURE_MULTIPLE_EVENT_CLIENTS (@link "" "MonitorFeature")
 * in the reply to the connect request.
 *
 *
 * @seealso ose_spi/monitor.sig
 * @seealso ose_spi/monitor_execution.sig
 * @seealso ose_rmm_set_user_action_handler
 * @seealso create_event
 * @seealso signal_event
 * @seealso ose_events/heap_events.h
 *
 */

#ifndef _MONITOR_EVENTS_SIG
#define _MONITOR_EVENTS_SIG

#include "stdint.h"
#include "monitor.sig"

#ifdef __cplusplus
extern "C" {
#endif

/**
 * @type MonitorAction
 * @brief Event action type.
 * @seealso MonitorActionValues
 */
typedef uint32_t MonitorAction;

/**
 * @type MonitorActionValues
 * @brief Enumeration of event actions.
 *
 * @long See the MONITOR_FEATURE_EVENT_TRACE
 * (@link "monitor.sig" "MonitorFeatureValue") if the monitor
 * supports the trace action.
 *
 * @long The initial event state is default 0, but can be
 * configured to be another value with krn/monitor/action
 * or set by a request or an action to any value.
 *
 * @long To perform "user" actions, MONITOR_ACTION_USER, a
 * user-defined event action handler can be registered in the system.
 * Only one user event action handler can be registered and it must
 * not change the event.
 * See @link "rmm.h" "ose_rmm_set_user_action_handler".
 *
 *
 * @field MONITOR_ACTION_INTERCEPT        Intercept the scope.
 * @field MONITOR_ACTION_NOTIFY           Report (notify) the event.
 * @field MONITOR_ACTION_TRACE            Trace the event.
 * @field MONITOR_ACTION_ENABLE_TRACE     Turn on the trace facility.
 * @field MONITOR_ACTION_DISABLE_TRACE    Turn off the trace facility.
 * @field MONITOR_ACTION_SET_STATE        Go to a new event action state
 *                                        (separate for this client).
 * @field MONITOR_ACTION_COUNT            NOT USED.
 * @field MONITOR_ACTION_USER             Call user specified function,
 *                                        registered as a user event action
 *                                        callback with @link "ose.h"
 *                                        "ose_rmm_set_user_action_handler".
 * @field MONITOR_ACTION_UNDO             Undo the send and receive call.
 * @field MONITOR_ACTION_ENABLE_HW_TRACE  Turn on the hardware trace facility.
 * @field MONITOR_ACTION_DISABLE_HW_TRACE Turn off the hardware trace facility.
 */
enum MonitorActionValues
{
   MONITOR_ACTION_INTERCEPT        = 1,
   MONITOR_ACTION_TRACE            = 3,
   MONITOR_ACTION_ENABLE_TRACE     = 4,
   MONITOR_ACTION_DISABLE_TRACE    = 5,
   MONITOR_ACTION_SET_STATE        = 9,
   MONITOR_ACTION_NOTIFY           = 10,
   MONITOR_ACTION_COUNT            = 11,
   MONITOR_ACTION_USER             = 12,
   MONITOR_ACTION_UNDO             = 13,
   MONITOR_ACTION_ENABLE_HW_TRACE  = 14,
   MONITOR_ACTION_DISABLE_HW_TRACE = 15
};

/**
 * @type MonitorEvent
 * @brief Event type.
 */
typedef uint32_t MonitorEvent;

/**
 * @type MonitorEventValues
 * @brief Enumeration of event types.
 *
 * @long These are different types of events that the monitor can
 * react on. If an action has been enabled, it is performed by the Monitor
 * when the event occurs.
 *
 * @long Events of the type MONITOR_EVENT_USER are user triggered events. They
 * are created and triggered, e.g. by applications, via the OSE function calls
 * @link "ose.h" "create_event"() and @link "ose.h" "signal_event"().
 * There are some user event numbers reserved for OSE in
 * include/ose_events/.
 *
 * @field MONITOR_EVENT_CREATE            A process was created.
 * @field MONITOR_EVENT_KILL              A process was killed.
 * @field MONITOR_EVENT_SWAP              A process was swapped in or out.
 * @field MONITOR_EVENT_SEND              A signal was sent.
 * @field MONITOR_EVENT_RECEIVE           A signal was received.
 * @field MONITOR_EVENT_ERROR             An error was reported.
 * @field MONITOR_EVENT_RESET             A system reset occurred.
 * @field MONITOR_EVENT_LOSS              Notification data was lost. An
 *                                        unknown number of event notifications
 *                                        are missing.
 * @field MONITOR_EVENT_USER              A user event occurred.
 * @field MONITOR_EVENT_BIND              A process, or block was bound
 *                                        (migrated from one execution unit
 *                                        to another).
 * @field MONITOR_EVENT_ALLOC             A process called alloc().
 * @field MONITOR_EVENT_FREE_BUF          A process called free_buf().
 * @field MONITOR_EVENT_SIGSELECT         A process was blocked in receive
 *                                        (NOT USED).
 * @field MONITOR_EVENT_SEMAPHORE_SIGNAL  A process signaled a semaphore
 *                                        (NOT USED).
 * @field MONITOR_EVENT_SEMAPHORE_WAIT    A process was blocked in a wait()
 *                                        call on a semaphore (NOT USED).
 * @field MONITOR_EVENT_SEMAPHORE         A process continued from a wait()
 *                                        call on a semaphore (NOT USED).
 * @field MONITOR_EVENT_MUTEX_LOCK        A process was blocked in a lock()
 *                                        call on a mutex (NOT USED).
 * @field MONITOR_EVENT_MUTEX_UNLOCK      A process unlocked a mutex
 *                                        (NOT USED).
 * @field MONITOR_EVENT_MUTEX             A process continued from a lock()
 *                                        call on a mutex (NOT USED).
 * @field MONITOR_EVENT_TIMEOUT		  The timeout in a system call expired.
 */
enum MonitorEventValues
{
   MONITOR_EVENT_CREATE            = 1,
   MONITOR_EVENT_KILL              = 2,
   MONITOR_EVENT_SWAP              = 3,
   MONITOR_EVENT_SEND              = 4,
   MONITOR_EVENT_RECEIVE           = 5,
   MONITOR_EVENT_RESET             = 10,
   MONITOR_EVENT_ERROR             = 11,
   MONITOR_EVENT_LOSS              = 12,
   MONITOR_EVENT_USER              = 13,
   MONITOR_EVENT_BIND              = 14,
   MONITOR_EVENT_ALLOC             = 15,
   MONITOR_EVENT_FREE_BUF          = 16,
   MONITOR_EVENT_SIGSELECT         = 17,
   MONITOR_EVENT_SEMAPHORE_SIGNAL  = 18,
   MONITOR_EVENT_SEMAPHORE_WAIT    = 19,
   MONITOR_EVENT_SEMAPHORE         = 20,
   MONITOR_EVENT_MUTEX_LOCK        = 21,
   MONITOR_EVENT_MUTEX_UNLOCK      = 22,
   MONITOR_EVENT_MUTEX             = 23,
   MONITOR_EVENT_TIMEOUT	   = 24
};

/*******************************************************************************
 *
 * EVENT ACTIONPOINT SIGNALS
 *
 ******************************************************************************/

/**
 * @type MonitorEventActionpointTagValue
 *
 * @brief Specifier of an event actionpoint tag type.
 *
 * @field MONITOR_EVENT_ACTIONPOINT_TAG_IGNORE
 * Number of times an event is allowed to occur and is ignored, before
 * the action is performed. The counter is restored after each action (num).
 *
 * @field MONITOR_EVENT_ACTIONPOINT_TAG_REMOVE
 * Number of times an event is allowed to occur before the actionpoint is
 * automatically removed, (num) (NOT USED).
 *
 * @field MONITOR_EVENT_ACTIONPOINT_TAG_DATA_RANGE
 * This is a filter tag that compare an unsigned integer, that is part of the
 * event data, to a range. This can for example be a variable in a signal. The
 * tag provides an offset into the data, the integer size (1, 2, 4, or 8 bytes 
 * is supported), a range A <= N < B, where N is the filtered number. B=0 means
 * no upper limit. A and B is 64 bit unsigned integers. The tag is applicable to
 * send, receive, free_buf and user event events. The tag arguments in order is
 * offset as 64 bit num, integer size as 32 bit num, range A as 64 bit num,
 * range B as 64 bit num. 64 bit tag arguments are represented in two 32 bit
 * halves (stored in the num field) and the order of the 32 bit halves follows
 * the target endian. The total tag count size, as specified in tag.count, is 8.
 * 
 * @field MONITOR_EVENT_ACTIONPOINT_TAG_DATA_RANGE_NOT
 * This filter tag works in the same way as
 * MONITOR_EVENT_ACTIONPOINT_TAG_DATA_RANGE but is true if the number is out of
 * the range, i.e. not(A <= N < B). The same parameters shall be included as for
 * MONITOR_EVENT_ACTIONPOINT_TAG_DATA_RANGE.
 * 
 * @field MONITOR_EVENT_ACTIONPOINT_TAG_END
 * Highest possible tag value plus one, should not be used.
 */
enum MonitorEventActionpointTagValue
{
   MONITOR_EVENT_ACTIONPOINT_TAG_IGNORE         = 2000,
   MONITOR_EVENT_ACTIONPOINT_TAG_REMOVE         = 2001,
   MONITOR_EVENT_ACTIONPOINT_TAG_DATA_RANGE     = 2002,
   MONITOR_EVENT_ACTIONPOINT_TAG_DATA_RANGE_NOT = 2003,         
   MONITOR_EVENT_ACTIONPOINT_TAG_END            = 2004
};

/**
 * @type MonitorEventActionpointParameterValue
 * @brief Bit masks for the parameter of an action definition.
 * @long These are mask values to use in the "parameter"
 * field when defining actions in actionpoint requests.
 * Mainly used when the action is MONITOR_ACTION_TRACE,
 * MONITOR_ACTION_NOTIFY or MONITOR_ACTION_INTERCEPT.
 *
 * @field MONITOR_EVENT_ACTIONPOINT_PARAMETER_FILE
 * Save file and line information when possible.
 * @field MONITOR_EVENT_ACTIONPOINT_PARAMETER_DATA
 * Save data when possible.
 *
 * @seealso MonitorEventActionpointInfo
 * @seealso MonitorSetSendActionpointRequest
 */
enum MonitorEventActionpointParameterValue
{
   MONITOR_EVENT_ACTIONPOINT_PARAMETER_FILE = 1,
   MONITOR_EVENT_ACTIONPOINT_PARAMETER_DATA = 2
};

/**
 * @type MonitorEventActionpointInfo
 *
 * @brief Struct used in many signals:
 *      MONITOR_SET_SEND_ACTIONPOINT_REQUEST,
 *      MONITOR_GET_SEND_ACTIONPOINT_INFO_REPLY,
 *      MONITOR_SET_RECEIVE_ACTIONPOINT_REQUEST,
 *      MONITOR_GET_RECEIVE_ACTIONPOINT_INFO_REPLY,
 *      MONITOR_SET_SWAP_ACTIONPOINT_REQUEST,
 *      MONITOR_GET_SWAP_ACTIONPOINT_INFO_REPLY,
 *      MONITOR_SET_CREATE_ACTIONPOINT_REQUEST,
 *      MONITOR_GET_CREATE_ACTIONPOINT_INFO_REPLY,
 *      MONITOR_SET_KILL_ACTIONPOINT_REQUEST,
 *      MONITOR_GET_KILL_ACTIONPOINT_INFO_REPLY,
 *      MONITOR_SET_ERROR_ACTIONPOINT_REQUEST,
 *      MONITOR_GET_ERROR_ACTIONPOINT_INFO_REPLY
 *      MONITOR_SET_BIND_ACTIONPOINT_REQUEST,
 *      MONITOR_GET_BIND_ACTIONPOINT_INFO_REPLY,
 *      MONITOR_SET_ALLOC_ACTIONPOINT_REQUEST,
 *      MONITOR_GET_ALLOC_ACTIONPOINT_INFO_REPLY,
 *      MONITOR_SET_FREE_BUF_ACTIONPOINT_REQUEST,
 *      MONITOR_GET_FREE_BUF_ACTIONPOINT_INFO_REPLY,
 *	MONITOR_SET_TIMEOUT_ACTIONPOINT_REQUEST,
 *	MONITOR_GET_TIMEOUT_ACTIONPOINT_REPLY
 *
 *
 * @field sigNo            Signal number of the request or reply.
 * @field state            If set to zero (0), this event actionpoint is always
 *                         active at all event states. If set to non-zero, the
 *                         actionpoint is only active at this non-zero event
 *                         state, in addition to state 0, where all are active.
 * @field from             Scope of processes to match as signal sender, error
 *                         reporter, creator, killer, or swapped out process.
 * @field to               Scope of processes to match as signal receiver,
 *                         created, killed, rebound, or swapped in process.
 * @field noAction         Set to 0 to perform the action if the event
 *                         matches or 1 to not perform the action if the event
 *                         matches, even if other events indicate this action
 *                         should be performed.
 * @field action           Action to take when this event occurs, i.e.
 *                         one of the @link "" "MonitorActionValues".
 * @field intercept        For MONITOR_ACTION_INTERCEPT, the scope of processes
 *                         to intercept (NOT USED, attached intercept scope
 *                         is used).
 * @field parameter        Parameter to the action.
 *                         @break
 *                         For the actions MONITOR_ACTION_TRACE,
 *                         MONITOR_ACTION_NOTIFY and MONITOR_ACTION_INTERCEPT,
 *                         this is a bit field: @break
 *                         Bit 0: set to 1 if file and line should be incorporated. @break
 *                         Bit 1: set to 1 if signal data should be incorporated. @break
 *                         Other bits should be set to zero. It is
 *                         recommended to use the enum @link ""
 *                         "MonitorEventActionpointParameterValue" instead of the
 *                         actual bit values.
 *                         @break
 *                         For MONITOR_ACTION_SET_STATE this is the state to set.
 * @field client           Set to 0 in requests. The client who set this event
 *                         actionpoint. Created and returned by the monitor
 *                         if actionpoints are listed.
 * @field count            Set to 0 in requests. The number of times this event has
 *                         occurred, without triggering the action. Compare
 *                         @link "" "MonitorEventActionpointTagValue".
 *                         Returned by the monitor if actionpoints
 *                         are listed.
 * @field id               Set to 0 in requests. The ID of this event actionpoint.
 *                         Created and returned by the monitor in the reply and
 *                         returned if actionpoints are listed.
 * @field reserved0        NOT USED.
 * @field reserved1        NOT USED.
 * @field reserved2        NOT USED.
 * @field tagsCount        The length of the tags list, as the
 *                         total number of sizeof(MonitorTag)
 *                         entries, with tags and arguments.
 * @field tags             List of tags and their values.
 */
typedef struct MonitorEventActionpointInfo
{
   uint32_t sigNo;
   uint32_t state;
   MonitorScope from;
   MonitorScope to;
   MonitorBoolean noAction;
   MonitorAction action;
   MonitorScope intercept;
   uint32_t parameter;
   uint32_t client;
   uint32_t id;
   uint32_t count;
   uint32_t reserved0;
   uint32_t reserved1;
   uint32_t reserved2;
   uint32_t tagsCount;
   MonitorTag tags[1];
} MonitorEventActionpointInfo;

#ifndef __cplusplus
#ifndef not
#define not noAction
#endif
#endif

/**
 * @signal MonitorSetSendActionpointRequest
 *
 * @brief Create new send event actionpoint request.
 *
 * @long This defines an action to be performed when a signal is sent
 * from a process matched by the scope and other rules in the actionpoint.
 *
 * @long The maximum number of bytes of signal data to incorporate
 * is controlled with krn/monitor/ group of system parameters.
 * Larger signals will be truncated. See RMM in "OSE User's Guide" for
 * system parameters.
 *
 * Accepts @link "monitor.sig" "MonitorSignalFilterTagValue"
 * MONITOR_SIGNAL_FILTER_TAG_NUMBER only.
 * Accepts @link "" "MonitorEventActionpointTagValue"
 * MONITOR_EVENT_ACTIONPOINT_TAG_IGNORE only.
 *
 * @member MONITOR_SET_SEND_ACTIONPOINT_REQUEST
 * @member MonitorEventActionpointInfo
 */

/** @macro MONITOR_SET_SEND_ACTIONPOINT_REQUEST */
#define MONITOR_SET_SEND_ACTIONPOINT_REQUEST 39150 /* !-SIGNO(MonitorEventActionpointInfo)-! */

/**
 * @signal MonitorSetReceiveActionpointRequest
 *
 * @brief Create new receive event actionpoint request.
 *
 * @long This defines an action to be performed when a signal is received
 * by a process matched by the scope and other rules in the actionpoint.
 *
 * @long The maximum number of bytes of signal data to incorporate
 * is controlled with krn/monitor/ group of system parameters.
 * Larger signals will be truncated. See RMM in the "OSE Core User's Guide" for
 * system parameters.
 *
 * Accepts MonitorSignalFilterTagValue MONITOR_SIGNAL_FILTER_TAG_NUMBER only.
 * Accepts @link "" "MonitorEventActionpointTagValue" MONITOR_EVENT_ACTIONPOINT_TAG_IGNORE only.
 *
 * @member MONITOR_SET_RECEIVE_ACTIONPOINT_REQUEST
 * @member MonitorEventActionpointInfo
 */

/** @macro MONITOR_SET_RECEIVE_ACTIONPOINT_REQUEST */
#define MONITOR_SET_RECEIVE_ACTIONPOINT_REQUEST 39151 /* !-SIGNO(MonitorEventActionpointInfo)-! */

/**
 * @signal MonitorSetSwapActionpointRequest
 *
 * @brief Create new swap (context switch) event actionpoint request.
 *
 * @long This defines an action to be performed when a context switch occurs
 * for a process matched by the scope and other rules in the actionpoint.
 *
 * Accepts @link "" "MonitorEventActionpointTagValue" MONITOR_EVENT_ACTIONPOINT_TAG_IGNORE only.
 *
 * @member MONITOR_SET_SWAP_ACTIONPOINT_REQUEST
 * @member MonitorEventActionpointInfo
 */

/** @macro MONITOR_SET_SWAP_ACTIONPOINT_REQUEST */
#define MONITOR_SET_SWAP_ACTIONPOINT_REQUEST 39152 /* !-SIGNO(MonitorEventActionpointInfo)-! */

/**
 * @signal MonitorSetCreateActionpointRequest
 *
 * @brief Create new create process event actionpoint request.
 *
 * @long This defines an action to be performed when a process is created
 * as matched by the scope and other rules in the actionpoint.
 *
 * Accepts @link "" "MonitorEventActionpointTagValue" MONITOR_EVENT_ACTIONPOINT_TAG_IGNORE only.
 * The from scope is NOT USED.
 *
 * @member MONITOR_SET_CREATE_ACTIONPOINT_REQUEST
 * @member MonitorEventActionpointInfo
 */

/** @macro MONITOR_SET_CREATE_ACTIONPOINT_REQUEST */
#define MONITOR_SET_CREATE_ACTIONPOINT_REQUEST 39153 /* !-SIGNO(MonitorEventActionpointInfo)-! */

/**
 * @signal MonitorSetKillActionpointRequest
 *
 * @brief Create new kill process event actionpoint request.
 *
 * @long This defines an action to be performed when a process is killed
 * as matched by the scope and other rules in the actionpoint.
 *
 * Accepts @link "" "MonitorEventActionpointTagValue" MONITOR_EVENT_ACTIONPOINT_TAG_IGNORE only.
 * The from scope is NOT USED.
 *
 * @member MONITOR_SET_KILL_ACTIONPOINT_REQUEST
 * @member MonitorEventActionpointInfo
 */

/** @macro MONITOR_SET_KILL_ACTIONPOINT_REQUEST */
#define MONITOR_SET_KILL_ACTIONPOINT_REQUEST 39154 /* !-SIGNO(MonitorEventActionpointInfo)-! */

/**
 * @type MonitorErrorFilterTagValue
 *
 * @brief Tags to specify an Error event filter.
 *
 * @long These are tags to use to filter which errors shall trigger an actionpoint.
 *
 * @long The filter tags that take two integer arguments A,B all interpret
 * those as the range A <= N < B, where N is the filtered number. B=0 means
 * no upper limit.
 *
 *
 * @field MONITOR_ERROR_FILTER_TAG_NOT    Exclude matches of next tag (NOT USED).
 * @field MONITOR_ERROR_FILTER_TAG_EXEC   Include errors called from executive (i.e.
 *                                        errors with user_called = 0).
 * @field MONITOR_ERROR_FILTER_TAG_ERROR  Error code range (2 x num).
 * @field MONITOR_ERROR_FILTER_TAG_EXTRA  Extra argument range (2 x num) (NOT USED).
 * @field MONITOR_ERROR_FILTER_TAG_END    Highest possible tag value plus one,
 *                                        should not be used.

 *
 * @seealso MonitorSetErrorActionpointRequest
 */
enum MonitorErrorFilterTagValue
{
   MONITOR_ERROR_FILTER_TAG_NOT = 2010,
   MONITOR_ERROR_FILTER_TAG_EXEC,
   MONITOR_ERROR_FILTER_TAG_ERROR,
   MONITOR_ERROR_FILTER_TAG_EXTRA,
   MONITOR_ERROR_FILTER_TAG_END
};

/**
 * @signal MonitorSetErrorActionpointRequest
 *
 * @brief Create new error event actionpoint request.
 *
 * @long This defines an action to be performed when an error occurs
 * in a process matched by the scope and other rules in the actionpoint.
 *
 * @long To filter specific errors, use error filter tags
 * - see @link "" "MonitorErrorFilterTagValue".
 *
 * Accepts @link "" "MonitorEventActionpointTagValue" MONITOR_EVENT_ACTIONPOINT_TAG_IGNORE only.
 *
 * @seealso MonitorErrorFilterTagValue
 *
 * @member MONITOR_SET_ERROR_ACTIONPOINT_REQUEST
 * @member MonitorEventActionpointInfo
 */

/** @macro MONITOR_SET_ERROR_ACTIONPOINT_REQUEST */
#define MONITOR_SET_ERROR_ACTIONPOINT_REQUEST 39155 /* !-SIGNO(MonitorEventActionpointInfo)-! */

/**
 * @type MonitorUserFilterTagValue
 *
 * @brief Tags to specify a user event filter based on event-number.
 *
 * @long These are tags used for filtering which user events shall trig
 * an actionpoint.
 *
 * @long The filter tags that take two integer arguments A,B interpret
 * those as the range A <= N < B, where N is the filtered number. B=0 means
 * no upper limit.
 *
 * @field MONITOR_USER_FILTER_TAG_NOT    Exclude matches of next tag (NOT USED).
 * @field MONITOR_USER_FILTER_TAG_NUMBER User event number range (2 x num).
 * @field MONITOR_USER_FILTER_TAG_END    Highest possible tag value plus one,
 *                                       should not be used.
 *
 * @seealso MonitorSetUserActionpointRequest
 */
enum MonitorUserFilterTagValue
{
   MONITOR_USER_FILTER_TAG_NOT = 2020,
   MONITOR_USER_FILTER_TAG_NUMBER,
   MONITOR_USER_FILTER_TAG_END
};

/**
 * @signal MonitorSetUserActionpointRequest
 *
 * @brief Create new user event actionpoint request.
 *
 * @long This defines an action to be performed when a user event is generated
 * by a process matched by the scope and other rules in the actionpoint.
 *
 * @long To filter specific user events, use user event filter tags
 * - see @link "" "MonitorUserFilterTagValue".
 *
 * Accepts @link "" "MonitorEventActionpointTagValue"
 * MONITOR_EVENT_ACTIONPOINT_TAG_IGNORE only.
 *
 * @seealso MonitorUserFilterTagValue
 *
 * @member MONITOR_SET_USER_ACTIONPOINT_REQUEST
 * @member MonitorEventActionpointInfo
 */

/** @macro MONITOR_SET_USER_ACTIONPOINT_REQUEST */
#define MONITOR_SET_USER_ACTIONPOINT_REQUEST 39156 /* !-SIGNO(MonitorEventActionpointInfo)-! */

/**
 * @signal MonitorSetBindActionpointRequest
 *
 * @brief Create new bind event actionpoint request.
 *
 * @long This defines an action to be performed when a bind occurs,
 * i.e. a process, matched by the scope and other rules in the actionpoint,
 * is about to be migrated to another execution unit.
 *
 * Accepts @link "" "MonitorEventActionpointTagValue"
 * MONITOR_EVENT_ACTIONPOINT_TAG_IGNORE only.
 *
 * @member MONITOR_SET_BIND_ACTIONPOINT_REQUEST
 * @member MonitorEventActionpointInfo
 */

/** @macro MONITOR_SET_BIND_ACTIONPOINT_REQUEST */
#define MONITOR_SET_BIND_ACTIONPOINT_REQUEST 39157 /* !-SIGNO(MonitorEventActionpointInfo)-! */

/**
 * @signal MonitorSetAllocActionpointRequest
 *
 * @brief Create new alloc event actionpoint request.
 *
 * @long This defines an action to be performed when a memory
 * @link "ose.h" "alloc" is performed,
 * by a process matched by the scope and other rules in the actionpoint.
 *
 * Accepts @link "" "MonitorEventActionpointTagValue"
 * MONITOR_EVENT_ACTIONPOINT_TAG_IGNORE only.
 *
 * @member MONITOR_SET_ALLOC_ACTIONPOINT_REQUEST
 * @member MonitorEventActionpointInfo
 */

/** @macro MONITOR_SET_ALLOC_ACTIONPOINT_REQUEST */
#define MONITOR_SET_ALLOC_ACTIONPOINT_REQUEST 39158 /* !-SIGNO(MonitorEventActionpointInfo)-! */

/**
 * @signal MonitorSetFreeBufActionpointRequest
 *
 * @brief Create new free_buf event actionpoint request.
 *
 * @long This defines an action to be performed when a memory
 * @link "ose.h" "free_buf" is performed,
 * by a process matched by the scope and other rules in the actionpoint.
 *
 * Accepts @link "" "MonitorEventActionpointTagValue"
 * MONITOR_EVENT_ACTIONPOINT_TAG_IGNORE only.
 *
 * @member MONITOR_SET_FREE_BUF_ACTIONPOINT_REQUEST
 * @member MonitorEventActionpointInfo
 */

/** @macro MONITOR_SET_FREE_BUF_ACTIONPOINT_REQUEST */
#define MONITOR_SET_FREE_BUF_ACTIONPOINT_REQUEST 39159 /* !-SIGNO(MonitorEventActionpointInfo)-! */

/**
 * @signal MonitorSetTimeoutActionpointRequest
 *
 * @brief Create new timeout event actionpoint request.
 *
 * @long This defines an action to be performed when a timeout occurs in
 * @link "ose.h" "receive_w_tmo", "receive_from", "receive_sport" or
 * "receive_with", that is performed by a process matching the scope
 * and other rules in the actionpoint.
 *
 * @long See RMM in the "OSE Core User's Guide" for system parameters.
 *
 * Accepts @link "" "MonitorEventActionpointTagValue"
 * MONITOR_EVENT_ACTIONPOINT_TAG_IGNORE only.
 *
 * @member MONITOR_SET_TIMEOUT_ACTIONPOINT_REQUEST
 * @member MonitorEventActionpointInfo
 */

/** @macro MONITOR_SET_TIMEOUT_ACTIONPOINT_REQUEST */
#define MONITOR_SET_TIMEOUT_ACTIONPOINT_REQUEST 39294 /* !-SIGNO(MonitorEventActionpointInfo)-! */


/**
 * @type MonitorEventActionpoint
 * @brief A reference ID for a previously set event actionpoint.
 * @long  It is mainly used in the reply signal from setting
 * and event actionpoint.
 *
 * @seealso MonitorSetEventActionpointReply
 */
typedef uint32_t MonitorEventActionpoint;

/**
 * @signal MonitorSetEventActionpointReply
 *
 * @brief Common reply to all requests to create a new event actionpoint.
 *
 * @long Sent by the server as reply to all requests to create a new
 * event actionpoint. An assigned ID of this actionpoint is returned.
 *
 * @member MONITOR_SET_EVENT_ACTIONPOINT_REPLY
 * @member MonitorSetEventActionpointReply
 */

/** @macro MONITOR_SET_EVENT_ACTIONPOINT_REPLY */
#define MONITOR_SET_EVENT_ACTIONPOINT_REPLY 39160 /* !-SIGNO(MonitorSetEventActionpointReply)-! */

/**
 * @type MonitorSetEventActionpointReply
 *
 * @output sigNo           MONITOR_SET_EVENT_ACTIONPOINT_REPLY
 * @output status          MONITOR_STATUS_OK or cause of error.
 * @output id              Assigned event actionpoint ID. This ID is used
 *                         in following references to this event actionpoint.
 */
typedef struct MonitorSetEventActionpointReply
{
   uint32_t sigNo;
   MonitorStatus status;
   MonitorEventActionpoint id;
} MonitorSetEventActionpointReply;

/**
 * @signal MonitorClearEventActionpointRequest
 *
 * @brief Remove event actionpoint request.
 *
 * @member MONITOR_CLEAR_EVENT_ACTIONPOINT_REQUEST
 * @member MonitorClearEventActionpointRequest
 *
 * @seealso MonitorClearEventActionpointReply
 */

/** @macro MONITOR_CLEAR_EVENT_ACTIONPOINT_REQUEST */
#define MONITOR_CLEAR_EVENT_ACTIONPOINT_REQUEST 39161 /* !-SIGNO(MonitorClearEventActionpointRequest)-! */

/**
 * @type MonitorClearEventActionpointRequest
 *
 * @input sigNo   MONITOR_CLEAR_EVENT_ACTIONPOINT_REQUEST
 * @input id      Event actionpoint ID, returned in previous
 *                @link "" "MonitorSetEventActionpointReply",
 *                or ~0 to clear all event actionpoints if
 *                MONITOR_FEATURE_CLEAR_ALL_EVENT_ACTIONPOINTS
 *                is supported.
 */
typedef struct MonitorClearEventActionpointRequest
{
   uint32_t sigNo;
   MonitorEventActionpoint id;
} MonitorClearEventActionpointRequest;

/**
 * @signal MonitorClearEventActionpointReply
 *
 * @brief Remove event actionpoint reply.
 *
 * @member MONITOR_CLEAR_EVENT_ACTIONPOINT_REPLY
 * @member MonitorClearEventActionpointReply
 */

/** @macro MONITOR_CLEAR_EVENT_ACTIONPOINT_REPLY */
#define MONITOR_CLEAR_EVENT_ACTIONPOINT_REPLY 39162 /* !-SIGNO(MonitorClearEventActionpointReply)-! */

/**
 * @type MonitorClearEventActionpointReply
 *
 * @output sigNo   MONITOR_CLEAR_EVENT_ACTIONPOINT_REPLY
 * @output status  MONITOR_STATUS_OK or cause of failure
 */
typedef struct MonitorClearEventActionpointReply
{
   uint32_t sigNo;
   MonitorStatus status;
} MonitorClearEventActionpointReply;

/* NOT_USED_SOURCEDOC
 * @signal MonitorGetEventActionpointInfoRequest
 *
 * @brief List event actionpoints request.
 *
 * @long Followed by zero or more MonitorGet*ActionpointInfoReply,
 * MonitorBlockingFlowControlRequest and one
 * MonitorGetEventActionpointInfoEndmark.
 *
 * @long Deprecated! Use MonitorGetEventActionpointsRequest instead
 *
 * @member MONITOR_GET_EVENT_ACTIONPOINT_INFO_REQUEST
 * @member MonitorGetEventActionpointInfoRequest
 */

/* @macro MONITOR_GET_EVENT_ACTIONPOINT_INFO_REQUEST */
#define MONITOR_GET_EVENT_ACTIONPOINT_INFO_REQUEST 39163 /* !-SIGNO(MonitorGetEventActionpointInfoRequest)-! */

/* NOT_USED_SOURCEDOC
 * @type MonitorGetEventActionpointInfoRequest
 *
 * @input sigNo MONITOR_GET_EVENT_ACTIONPOINT_INFO_REQUEST
 *
 * @long Deprecated! Use MonitorGetEventActionpointsRequest instead
 */
typedef struct MonitorGetEventActionpointInfoRequest
{
   uint32_t sigNo;
} MonitorGetEventActionpointInfoRequest;

/* NOT_USED_SOURCEDOC
 * @signal MonitorGetSendActionpointInfoReply
 *
 * @brief List send actionpoint reply
 *
 * @long This signal is sent under blocking flow control.
 *
 * @long Deprecated! Use MonitorGetEventActionpointsReply instead
 *
 * @member MONITOR_GET_SEND_ACTIONPOINT_INFO_REPLY
 * @member MonitorEventActionpointInfo
 */

/* @macro MONITOR_GET_SEND_ACTIONPOINT_INFO_REPLY */
#define MONITOR_GET_SEND_ACTIONPOINT_INFO_REPLY 39164 /* !-SIGNO(MonitorEventActionpointInfo)-! */

/* NOT_USED_SOURCEDOC
 * @signal MonitorGetReceiveActionpointInfoReply
 *
 * @brief List receive actionpoint reply
 *
 * @long This signal is sent under blocking flow control.
 *
 * @long Deprecated! Use MonitorGetEventActionpointsReply instead
 *
 * @member MONITOR_GET_RECEIVE_ACTIONPOINT_INFO_REPLY
 * @member MonitorEventActionpointInfo
 */

/* @macro MONITOR_GET_RECEIVE_ACTIONPOINT_INFO_REPLY */
#define MONITOR_GET_RECEIVE_ACTIONPOINT_INFO_REPLY 39165 /* !-SIGNO(MonitorEventActionpointInfo)-! */

/* NOT_USED_SOURCEDOC
 * @signal MonitorGetSwapActionpointInfoReply
 *
 * @brief List swap actionpoint reply
 *
 * @long This signal is sent under blocking flow control.
 *
 * @long Deprecated! Use MonitorGetEventActionpointsReply instead
 *
 * @member MONITOR_GET_SWAP_ACTIONPOINT_INFO_REPLY
 * @member MonitorEventActionpointInfo
 */

/* @macro MONITOR_GET_SWAP_ACTIONPOINT_INFO_REPLY */
#define MONITOR_GET_SWAP_ACTIONPOINT_INFO_REPLY 39166 /* !-SIGNO(MonitorEventActionpointInfo)-! */

/* NOT_USED_SOURCEDOC
 * @signal MonitorGetCreateActionpointInfoReply
 *
 * @brief List create actionpoint reply
 *
 * @long This signal is sent under blocking flow control.
 *
 * @long Deprecated! Use MonitorGetEventActionpointsReply instead
 *
 * @member MONITOR_GET_CREATE_ACTIONPOINT_INFO_REPLY
 * @member MonitorEventActionpointInfo
 */

/* @macro MONITOR_GET_CREATE_ACTIONPOINT_INFO_REPLY */
#define MONITOR_GET_CREATE_ACTIONPOINT_INFO_REPLY 39167 /* !-SIGNO(MonitorEventActionpointInfo)-! */

/* NOT_USED_SOURCEDOC
 * @signal MonitorGetKillActionpointInfoReply
 *
 * @brief List kill actionpoint reply
 *
 * @long This signal is sent under blocking flow control.
 *
 * @long Deprecated! Use MonitorGetEventActionpointsReply instead
 *
 * @member MONITOR_GET_KILL_ACTIONPOINT_INFO_REPLY
 * @member MonitorEventActionpointInfo
 */

/* @macro MONITOR_GET_KILL_ACTIONPOINT_INFO_REPLY */
#define MONITOR_GET_KILL_ACTIONPOINT_INFO_REPLY 39168 /* !-SIGNO(MonitorEventActionpointInfo)-! */

/* NOT_USED_SOURCEDOC
 * @signal MonitorGetErrorActionpointInfoReply
 *
 * @brief List error actionpoint reply
 *
 * @long This signal is sent under blocking flow control.
 *
 * @long Deprecated! Use MonitorGetEventActionpointsReply instead
 *
 * @member MONITOR_GET_ERROR_ACTIONPOINT_INFO_REPLY
 * @member MonitorEventActionpointInfo
 */

/* @macro MONITOR_GET_ERROR_ACTIONPOINT_INFO_REPLY */
#define MONITOR_GET_ERROR_ACTIONPOINT_INFO_REPLY 39169 /* !-SIGNO(MonitorEventActionpointInfo)-! */

/* NOT_USED_SOURCEDOC
 * @signal MonitorGetTimeoutActionpointInfoReply
 * @brief List timeout actionpoint reply
 *
 * @long This signal is sent under blocking flow control.
 *
 * @long Deprecated! Use MonitorGetEventActionpointsReply instead
 *
 * @member MONITOR_GET_TIMEOUT_ACTIONPOINT_INFO_REPLY
 * @member MonitorEventActionpointInfo
 */

/* @macro MONITOR_GET_TIMEOUT_ACTIONPOINT_INFO_REPLY */
#define MONITOR_GET_TIMEOUT_ACTIONPOINT_INFO_REPLY 39295

/* NOT_USED_SOURCEDOC
 * @signal MonitorGetEventActionpointInfoEndmark
 *
 * @brief List event actionpoints endmark, with status.
 *
 * @long Deprecated! Use MonitorGetEventActionpointsEndmark instead
 *
 * @member MONITOR_GET_EVENT_ACTIONPOINT_INFO_ENDMARK
 * @member MonitorGetEventActionpointInfoEndmark
 */

/* @macro MONITOR_GET_EVENT_ACTIONPOINT_INFO_ENDMARK */
#define MONITOR_GET_EVENT_ACTIONPOINT_INFO_ENDMARK 39175 /* !-SIGNO(MonitorGetEventActionpointInfoEndmark)-! */

/* NOT_USED_SOURCEDOC
 * @type MonitorGetEventActionpointInfoEndmark
 *
 * @output sigNo MONITOR_GET_EVENT_ACTIONPOINT_INFO_ENDMARK
 * @output status MONITOR_STATUS_OK or cause of failure.
 */
typedef struct MonitorGetEventActionpointInfoEndmark
{
   uint32_t sigNo;
   MonitorStatus status;
} MonitorGetEventActionpointInfoEndmark;

/**
 * @signal MonitorGetEventActionpointsRequest
 *
 * @brief List event actionpoints request
 *
 * @long Followed by zero or more MonitorGetActionpointsReply,
 * MonitorBlockingFlowControlRequest and one
 * MonitorGetEventActionpointsEndmark.
 *
 * @member MONITOR_GET_EVENT_ACTIONPOINTS_REQUEST
 * @member MonitorGetEventActionpointsRequest
 *
 * @seealso MonitorGetEventActionpointsReply
 * @seealso MonitorBlockingFlowControlRequest
 */

/** @macro MONITOR_GET_EVENT_ACTIONPOINTS_REQUEST */
#define MONITOR_GET_EVENT_ACTIONPOINTS_REQUEST 39170 /* !-SIGNO(MonitorGetEventActionpointsRequest)-! */

/**
 * @type MonitorGetEventActionpointsRequest
 * @input sigNo      MONITOR_GET_EVENT_ACTIONPOINTS_REQUEST
 * @input reserved0  NOT USED. Should be zero.
 */
typedef struct MonitorGetEventActionpointsRequest
{
   uint32_t sigNo;
   uint32_t reserved0;
} MonitorGetEventActionpointsRequest;

/**
 * @signal MonitorGetEventActionpointsReply
 *
 * @brief List event actionpoint reply.
 *
 * @long This signal is sent under blocking flow control.
 *
 * @member MONITOR_GET_EVENT_ACTIONPOINTS_REPLY
 * @member MonitorGetEventActionpointsReply
 *
 * @seealso MonitorGetEventActionpointsRequest
 */

/** @macro MONITOR_GET_EVENT_ACTIONPOINTS_REPLY */
#define MONITOR_GET_EVENT_ACTIONPOINTS_REPLY 39171 /* !-SIGNO(MonitorGetEventActionpointsReply)-! */

/**
 * @type MonitorGetEventActionpointsReply
 * @output sigNo     MONITOR_GET_EVENT_ACTIONPOINTS_REPLY
 * @output reserved0 NOT USED.
 * @output info      Event actionpoint information.
 *                   This is the same struct MonitorEventActionpointInfo
 *                   as included in each MonitorSetXxxxxActionpointRequest,
 *                   e.g. in @link "" "MonitorSetSendActionpointRequest",
 *                   but with actual field values filled in.
 */
typedef struct MonitorGetEventActionpointsReply
{
   uint32_t sigNo;
   uint32_t reserved0;
   MonitorEventActionpointInfo info;
} MonitorGetEventActionpointsReply;

/**
 * @signal MonitorGetEventActionpointsEndmark
 *
 * @brief List event actionpoints endmark, with status
 *
 * @member MONITOR_GET_EVENT_ACTIONPOINTS_ENDMARK
 * @member MonitorGetEventActionpointsEndmark
 *
 * @seealso MonitorGetEventActionpointsRequest
 * @seealso MonitorGetEventActionpointsReply
 */

/** @macro MONITOR_GET_EVENT_ACTIONPOINTS_ENDMARK */
#define MONITOR_GET_EVENT_ACTIONPOINTS_ENDMARK 39172 /* !-SIGNO(MonitorGetEventActionpointsEndmark)-! */

/**
 * @type MonitorGetEventActionpointsEndmark
 *
 * @output sigNo MONITOR_GET_EVENT_ACTIONPOINTS_ENDMARK
 * @output status MONITOR_STATUS_OK or cause of failure.
 */
typedef struct MonitorGetEventActionpointsEndmark
{
   uint32_t sigNo;
   MonitorStatus status;
} MonitorGetEventActionpointsEndmark;


/*******************************************************************************
 *
 * EVENT STATE SIGNALS
 *
 ******************************************************************************/

/**
 * @signal MonitorSetEventStateRequest
 *
 * @brief Set active event state to a specific value request.
 *
 * @long Default the initial state is 0 at system start, unless otherwise
 * configured in the target for the monitor.
 * For full control, it is recommended that a system debugger sets
 * the event state before defining event actionpoints, depending
 * on the state.
 *
 * @long The state can also be changed as an action for an event actionpoint.
 *
 * @member MONITOR_SET_EVENT_STATE_REQUEST
 * @member MonitorSetEventStateRequest
 * @seealso MonitorSetEventStateReply
 */

/** @macro MONITOR_SET_EVENT_STATE_REQUEST */
#define MONITOR_SET_EVENT_STATE_REQUEST 39180 /* !-SIGNO(MonitorSetEventStateRequest)-! */

/**
 * @type MonitorSetEventStateRequest
 *
 * @input sigNo MONITOR_SET_EVENT_STATE_REQUEST
 * @input state New state number.
 */
typedef struct MonitorSetEventStateRequest
{
   uint32_t sigNo;
   uint32_t state;
} MonitorSetEventStateRequest;

/**
 * @signal MonitorSetEventStateReply
 *
 * @brief Set active event state reply.
 *
 * @member MONITOR_SET_EVENT_STATE_REPLY
 * @member MonitorSetEventStateReply
 * @seealso MonitorSetEventStateRequest
 */

/** @macro MONITOR_SET_EVENT_STATE_REPLY */
#define MONITOR_SET_EVENT_STATE_REPLY 39181 /* !-SIGNO(MonitorSetEventStateReply)-! */

/**
 * @type MonitorSetEventStateReply
 *
 * @input sigNo  MONITOR_SET_EVENT_STATE_REPLY
 * @input status MONITOR_STATUS_OK or cause of failure
 */
typedef struct MonitorSetEventStateReply
{
   uint32_t sigNo;
   MonitorStatus status;
} MonitorSetEventStateReply;

/**
 * @signal MonitorGetEventStateRequest
 *
 * @brief Get active state for this client request.
 *
 * @member MONITOR_GET_EVENT_STATE_REQUEST
 * @member MonitorGetEventStateRequest
 * @seealso MonitorSetEventStateRequest
 * @seealso MonitorSetEventStateReply
 */

/** @macro MONITOR_GET_EVENT_STATE_REQUEST */
#define MONITOR_GET_EVENT_STATE_REQUEST 39182 /* !-SIGNO(MonitorGetEventStateRequest)-! */

/**
 * @type MonitorGetEventStateRequest
 * @input sigNo MONITOR_GET_EVENT_STATE_REQUEST
 */
typedef struct MonitorGetEventStateRequest
{
   uint32_t sigNo;
} MonitorGetEventStateRequest;

/**
 * @signal MonitorGetEventStateReply
 *
 * @brief Get active state reply.
 *
 * @member MONITOR_GET_EVENT_STATE_REPLY
 * @member MonitorGetEventStateReply
 * @seealso MonitorSetEventStateRequest
 */

/** @macro MONITOR_GET_EVENT_STATE_REPLY */
#define MONITOR_GET_EVENT_STATE_REPLY 39183 /* !-SIGNO(MonitorGetEventStateReply)-! */

/**
 * @type MonitorGetEventStateReply
 *
 * @output sigNo  MONITOR_GET_EVENT_STATE_REPLY
 * @output status MONITOR_STATUS_OK or cause of failure
 * @output state  New state number.
 */
typedef struct MonitorGetEventStateReply
{
   uint32_t sigNo;
   MonitorStatus status;
   uint32_t state;
} MonitorGetEventStateReply;

/*******************************************************************************
 *
 * EVENT SIGNALS
 *
 ******************************************************************************/

/**
 * @type MonitorProcessInfo
 *
 * @brief Struct used as type and as several signals:
 *    MONITOR_PROCESS_NOTIFY
 *    MONITOR_GET_PROCESS_TRACE_REPLY
 *    MONITOR_PROCESS_INTERCEPT_NOTIFY
 *
 * @output sigNo      Signal number of this notify or reply signal.
 * @output pid        ID of the process.
 * @output bid        ID of the parent block or 0.
 * @output sid        ID of the parent segment (NOT USED).
 * @output type       Type of the process or 0.
 * @output entrypoint Entrypoint of the process or 0.
 * @output properties Bit field indicating additional properties of the process
 *                    (NOT USED).
 *                    Bit 0 is set to 1 for system processes.
 *                    Other bits are undefined.
 * @output reserved0  NOT USED.
 * @output name       Null terminated name of the process.
 */
typedef struct MonitorProcessInfo
{
   uint32_t sigNo;
   uint32_t pid;
   uint32_t bid;
   uint32_t sid;
   MonitorProcessType type;
   uint32_t entrypoint;
   uint32_t properties;
   uint32_t reserved0;
   char name[1];
} MonitorProcessInfo;

/**
 * @type MonitorEventTimestamp
 *
 * @brief Timestamp used in event data.
 *
 * @long The current time can be expressed in two ways,
 * either as <tick>, ticks plus <ntick> timer steps,
 * or as <sec>, seconds plus <sectick> ticks plus <ntick> timer steps.
 *
 * @output tick       Number of ticks.
 * @output ntick      Number of timer steps after the last tick.
 * @output sec        Seconds since 00:00:00 1 Jan 1970 GMT or zero.
 * @output sectick    Number of ticks after the last second.
 */
typedef struct MonitorEventTimestamp
{
   uint32_t tick;
   uint32_t ntick;
   uint32_t sec;
   uint32_t sectick;
} MonitorEventTimestamp;

/**
 * @type MonitorSendReceiveInfo
 *
 * @brief Struct used as several signals:
 *    MONITOR_SEND_NOTIFY,
 *    MONITOR_RECEIVE_NOTIFY,
 *    MONITOR_SEND_INTERCEPT_NOTIFY,
 *    MONITOR_RECEIVE_INTERCEPT_NOTIFY,
 *    MONITOR_GET_SEND_TRACE_REPLY,
 *    MONITOR_GET_RECEIVE_TRACE_REPLY
 *
 * @output sigNo             Signal number of the notification or trace reply.
 * @output reference         Actionpoint ID or trace entry number.
 * @output time              Timestamp (@link "" "MonitorEventTimestamp").
 * @output signalNumber      Signal number.
 * @output signalSender      Sender process ID.
 * @output signalAddressee   Receiver process ID.
 * @output signalSize        Signal size or 0 if this is not known.
 * @output signalAddress     Signal buffer address or 0 if this is not known.
 * @output signalId          Signal ID or 0 if not used (NOT USED).
 * @output lineNumber        line number or 0 if this is not known.
 * @output fileNameOffset    Offset into data of filename (if lineNumber != 0)
 * @output signalDataOffset  Offset into data of signal data.
 * @output signalDataCount   Length of signal data (in bytes). It can be
 *                           less than the size, limited by the configured
 *                           maximum signal size in notifications or traces.
 * @output euId              Execution unit (CPU core) ID of the event source
 *                           or 0 if not available.
 * @output reserved0         NOT USED.
 * @output reserved1         NOT USED.
 * @output data              Filename etc.
 */
typedef struct MonitorSendReceiveInfo
{
   uint32_t sigNo;
   uint32_t reference;
   MonitorEventTimestamp time;
   uint32_t signalNumber;
   uint32_t signalSender;
   uint32_t signalAddressee;
   uint32_t signalSize;
   uint32_t signalAddress;
   uint32_t signalId;
   uint32_t lineNumber;
   uint32_t fileNameOffset;
   uint32_t signalDataOffset;
   uint32_t signalDataCount;
   uint16_t euId;
   uint16_t reserved0;
   uint32_t reserved1;
   uint8_t data[1];
} MonitorSendReceiveInfo;

/**
 * @type MonitorSwapCreateKillInfo
 *
 * @brief Struct used as several signals:
 *   MONITOR_SWAP_NOTIFY
 *   MONITOR_CREATE_NOTIFY
 *   MONITOR_KILL_NOTIFY
 *   MONITOR_SWAP_INTERCEPT_NOTIFY
 *   MONITOR_CREATE_INTERCEPT_NOTIFY
 *   MONITOR_KILL_INTERCEPT_NOTIFY
 *   MONITOR_GET_SWAP_TRACE_REPLY
 *   MONITOR_GET_CREATE_TRACE_REPLY
 *   MONITOR_GET_KILL_TRACE_REPLY
 *
 * @output sigNo             Signal number of the notification or trace reply.
 * @output reference         Actionpoint ID or trace entry number.
 * @output time              Timestamp (@link "" "MonitorEventTimestamp").
 * @output from              Swapped out or caller process ID.
 * @output to                Swapped in, created or killed process ID.
 * @output lineNumber        NOT USED.
 * @output fileNameOffset    NOT USED.
 * @output euId              Execution unit (CPU core) ID of the event source
 *                           or 0 if not available.
 * @output reserved0         NOT USED.
 * @output reserved1         NOT USED.
 * @output data              Filename etc.
 */
typedef struct MonitorSwapCreateKillInfo
{
   uint32_t sigNo;
   uint32_t reference;
   MonitorEventTimestamp time;
   uint32_t from;
   uint32_t to;
   uint32_t lineNumber;
   uint32_t fileNameOffset;
   uint16_t euId;
   uint16_t reserved0;
   uint32_t reserved1;
   uint8_t data[1];
} MonitorSwapCreateKillInfo;

/**
 * @type MonitorErrorInfo
 *
 * @brief Struct used as type and as several signals:
 *   MONITOR_ERROR_NOTIFY
 *   MONITOR_GET_ERROR_TRACE_REPLY
 *   MONITOR_ERROR_INTERCEPT_NOTIFY
 *
 * @output sigNo             Signal number of this notification or trace reply.
 * @output reference         Actionpoint ID or trace entry number.
 * @output time              Timestamp (@link "" "MonitorEventTimestamp").
 * @output from              Caller process ID.
 * @output exec              True if called from the executive (i.e.
 *                           user_called is 0).
 * @output error             The error number.
 * @output extra             The extra information.
 * @output lineNumber        Line number or 0 if not known.
 * @output fileNameOffset    Offset into data of filename (if lineNumber != 0)
 * @output euId              Execution unit (CPU core) ID of the event source
 *                           or 0 if not available.
 * @output reserved0         NOT USED.
 * @output reserved1         NOT USED.
 * @output data              Filename etc.
 */
typedef struct MonitorErrorInfo
{
   uint32_t sigNo;
   uint32_t reference;
   MonitorEventTimestamp time;
   uint32_t from;
   MonitorBoolean exec;
   uint32_t error;
   uint32_t extra;
   uint32_t lineNumber;
   uint32_t fileNameOffset;
   uint16_t euId;
   uint16_t reserved0;
   uint32_t reserved1;
   uint8_t data[1];
} MonitorErrorInfo;

/**
 * @type MonitorResetInfo
 *
 * @brief Struct used as one signal.
 *
 * @output sigNo             MONITOR_GET_RESET_TRACE_REPLY
 * @output reference         Actionpoint ID or trace entry number.
 * @output time              Timestamp (@link "" "MonitorEventTimestamp").
 * @output warm              True if this was a warm reset (RAM contents survived)
 * @output lineNumber        NOT USED.
 * @output fileNameOffset    NOT USED.
 * @output data              NOT USED.
 * @output reserved0         NOT USED.
 * @output reserved1         NOT USED.
 * @output reserved2         NOT USED.
 *
 */
typedef struct MonitorResetInfo
{
   uint32_t sigNo;
   uint32_t reference;
   MonitorEventTimestamp time;
   MonitorBoolean warm;
   uint32_t lineNumber;
   uint32_t fileNameOffset;
   uint32_t reserved0;
   uint32_t reserved1;
   uint32_t reserved2;
   uint8_t data[1];
} MonitorResetInfo;

/**
 * @type MonitorUserInfo
 *
 * @brief Struct used as several signals:
 *   MONITOR_USER_NOTIFY
 *   MONITOR_USER_INTERCEPT_NOTIFY
 *   MONITOR_GET_USER_TRACE_REPLY
 *
 * @output sigNo             Signal number of this notify or reply signal.
 * @output reference         Actionpoint ID or trace entry number.
 * @output time              Timestamp (@link "" "MonitorEventTimestamp").
 * @output from              Caller process ID.
 * @output eventNumber       The user event type identification.
 * @output lineNumber        NOT USED contains 0.
 * @output fileNameOffset    NOT USED.
 * @output euId              Execution unit (CPU core) ID of the event source
 *                           or 0 if not available.
 * @output reserved0         NOT USED.
 * @output eventDataOffset   Offset into data of user data.
 * @output eventDataCount    Number of bytes in user data.
 * @output eventSize         Number of bytes requested for user data.
 * @output data              Data.
 */
typedef struct MonitorUserInfo
{
   uint32_t sigNo;
   uint32_t reference;
   MonitorEventTimestamp time;
   uint32_t from;
   uint32_t eventNumber;
   uint32_t eventSize;
   uint32_t lineNumber;
   uint32_t fileNameOffset;
   uint32_t eventDataOffset;
   uint32_t eventDataCount;
   uint16_t euId;
   uint16_t reserved0;
   uint8_t data[1];
} MonitorUserInfo;

/**
 * @type MonitorBindInfo
 *
 * @brief Struct used as several signals:
 *   MONITOR_BIND_NOTIFY
 *   MONITOR_BIND_INTERCEPT_NOTIFY
 *   MONITOR_GET_BIND_TRACE_REPLY
 *
 * @output sigNo             Signal number of this notify or reply signal.
 * @output reference         Actionpoint ID or trace entry number.
 * @output time              Timestamp (@link "" "MonitorEventTimestamp").
 * @output reserved0         NOT USED.
 * @output bound             Migrated process ID.
 * @output fromEuId          Execution unit ID migrated from.
 * @output toEuId            Execution unit ID migrated to.
 * @output lineNumber        NOT USED.
 * @output fileNameOffset    NOT USED.
 * @output reserved1         NOT USED.
 * @output data              Filename etc.
 */
typedef struct MonitorBindInfo
{
   uint32_t sigNo;
   uint32_t reference;
   MonitorEventTimestamp time;
   uint32_t reserved0;
   uint32_t bound;
   uint16_t fromEuId;
   uint16_t toEuId;
   uint32_t lineNumber;
   uint32_t fileNameOffset;
   uint32_t reserved1;
   uint8_t data[1];
} MonitorBindInfo;

/**
 * @type MonitorAllocFreeBufInfo
 *
 * @brief Struct used as several signals:
 *   MONITOR_ALLOC_NOTIFY
 *   MONITOR_ALLOC_INTERCEPT_NOTIFY
 *   MONITOR_GET_ALLOC_TRACE_REPLY
 *   MONITOR_FREE_BUF_NOTIFY
 *   MONITOR_FREE_BUF_INTERCEPT_NOTIFY
 *  MONITOR_GET_FREE_BUF_TRACE_REPLY
 *
 * @output sigNo             Signal number of this notify or reply signal.
 * @output reference         Actionpoint ID or trace entry number.
 * @output time              Timestamp (@link "" "MonitorEventTimestamp").
 * @output from              ID of process that generated the event.
 * @output signalNumber      Signal number.
 * @output signalSize        Signal size.
 * @output signalAddress     Signal buffer address or 0 if not known.
 * @output signalId          Signal ID or 0 if not used (NOT USED).
 * @output lineNumber        NOT USED.
 * @output fileNameOffset    NOT USED.
 * @output euId              Execution unit of the event
 *                           or 0 if not available.
 * @output reserved0         NOT USED.
 * @output data              Filename etc.
 */
typedef struct MonitorAllocFreeBufInfo
{
   uint32_t sigNo;
   uint32_t reference;
   MonitorEventTimestamp time;
   uint32_t from;
   uint32_t signalNumber;
   uint32_t signalSize;
   uint32_t signalAddress;
   uint32_t signalId;
   uint32_t lineNumber;
   uint32_t fileNameOffset;
   uint16_t euId;
   uint16_t reserved0;
   uint8_t data[1];
} MonitorAllocFreeBufInfo;


/**
 * @type MonitorTimeoutInfo
 *
 * @brief Struct used as several signals:
 *   MONITOR_TIMEOUT_NOTIFY
 *   MONITOR_TIMEOUT_INTERCEPT_NOTIFY
 *   MONITOR_GET_TIMEOUT_TRACE_REPLY
 * @output sigNo             Signal number of the notification or trace reply.
 * @output reference         Actionpoint ID or trace entry number.
 * @output time              Timestamp (@link "" "MonitorEventTimestamp").
 * @output timeout           Timeout length.
 * @output tmoSource	     Contains what it was that timed out. Contains 0 for receive.
 * @output signalAddressee   Receiver process ID.
 * @output lineNumber        Line number or 0 if this is not known.
 * @output fileNameOffset    Offset into data of filename (if lineNumber != 0)
 * @output euId              Execution unit (CPU core) ID of the event source
 *                           or 0 if not available.
 * @output reserved0         NOT USED.
 * @output data              Filename etc.
 */
typedef struct MonitorTimeoutInfo
{
   uint32_t sigNo;
   uint32_t reference;
   MonitorEventTimestamp time;
   uint32_t timeout;
   uint32_t tmoSource;
   uint32_t signalAddressee;
   uint32_t lineNumber;
   uint32_t fileNameOffset;
   uint16_t euId;
   uint16_t reserved0;
   uint8_t data[1];
} MonitorTimeoutInfo;


/*******************************************************************************
 *
 * EVENT NOTIFICATION SIGNALS
 *
 ******************************************************************************/

/**
 * @signal MonitorSetNotifyEnabledRequest
 *
 * @brief Enable / disable notifications to this client request.
 *
 * @member MONITOR_SET_NOTIFY_ENABLED_REQUEST
 * @member MonitorSetNotifyEnabledRequest
 */

/** @macro MONITOR_SET_NOTIFY_ENABLED_REQUEST */
#define MONITOR_SET_NOTIFY_ENABLED_REQUEST 39185 /* !-SIGNO(MonitorSetNotifyEnabledRequest)-! */

/**
 * @type MonitorSetNotifyEnabledRequest
 *
 * @input sigNo           MONITOR_SET_NOTIFY_ENABLED_REQUEST
 * @input enabled         Desired status (MONITOR_BOOLEAN_TRUE
 *                        or MONITOR_BOOLEAN_FALSE)
 */
typedef struct MonitorSetNotifyEnabledRequest
{
   uint32_t sigNo;
   MonitorBoolean enabled;
} MonitorSetNotifyEnabledRequest;

/**
 * @signal MonitorSetNotifyEnabledReply
 *
 * @brief Enable / disable notifications reply.
 *
 * @member MONITOR_SET_NOTIFY_ENABLED_REPLY
 * @member MonitorSetNotifyEnabledReply
 */

/** @macro MONITOR_SET_NOTIFY_ENABLED_REPLY */
#define MONITOR_SET_NOTIFY_ENABLED_REPLY 39186 /* !-SIGNO(MonitorSetNotifyEnabledReply)-! */

/**
 * @type MonitorSetNotifyEnabledReply
 *
 * @output sigNo           MONITOR_SET_NOTIFY_ENABLED_REPLY
 * @output status          MONITOR_STATUS_OK or cause of failure
 */
typedef struct MonitorSetNotifyEnabledReply
{
   uint32_t sigNo;
   MonitorStatus status;
} MonitorSetNotifyEnabledReply;

/**
 * @signal MonitorGetNotifyEnabledRequest
 *
 * @brief Get status of notifications to this client request.
 *
 * @member MONITOR_GET_NOTIFY_ENABLED_REQUEST
 * @member MonitorGetNotifyEnabledRequest
 */

/** @macro MONITOR_GET_NOTIFY_ENABLED_REQUEST */
#define MONITOR_GET_NOTIFY_ENABLED_REQUEST 39187 /* !-SIGNO(MonitorGetNotifyEnabledRequest)-! */

/**
 * @type MonitorGetNotifyEnabledRequest
 *
 * @input sigNo MONITOR_GET_NOTIFY_ENABLED_REQUEST
 */
typedef struct MonitorGetNotifyEnabledRequest
{
   uint32_t sigNo;
} MonitorGetNotifyEnabledRequest;

/**
 * @signal MonitorGetNotifyEnabledReply
 *
 * @brief Get status of notifications reply.
 *
 * @member MONITOR_GET_NOTIFY_ENABLED_REPLY
 * @member MonitorGetNotifyEnabledReply
 */

/** @macro MONITOR_GET_NOTIFY_ENABLED_REPLY */
#define MONITOR_GET_NOTIFY_ENABLED_REPLY 39188 /* !-SIGNO(MonitorGetNotifyEnabledReply)-! */

/**
 * @type MonitorGetNotifyEnabledReply
 *
 * @output sigNo  MONITOR_GET_NOTIFY_ENABLED_REPLY
 * @output status MONITOR_STATUS_OK or cause of failure
 * @output enabled 1 if enabled 0 otherwise.
 */
typedef struct MonitorGetNotifyEnabledReply
{
   uint32_t sigNo;
   MonitorStatus status;
   MonitorBoolean enabled;
} MonitorGetNotifyEnabledReply;

/**
 * @signal MonitorProcessNotify
 *
 * @brief Report with process details - for processes
 *        in later MonitorXxxxxNotify replies.
 *
 * @long To avoid resending details about each process multiple times,
 * the monitor sends process detail reports separately, before
 * notifications are sent referring to these processes.
 *
 * @long The monitor assumes that the client caches at least the last 32
 * reports with process details and avoids resending these.
 * The cache, with 32 reports, is shared by the report signals
 * MonitorProcessNotify and MonitorProcessInterceptNotify, which
 * contain the same process details.
 * The requested size of this cache is actually returned from the
 * monitor to the client as a parameter in @link "" "MonitorAttachReply".
 *
 * @long The monitor will send the MonitorProcessNotify signal for
 * all processes mentioned in a sequence of
 * MonitorXxxxxxNotify signals (but not MonitorXxxxxxInterceptNotify)
 * to provide additional information about the processes.
 * The monitor will ensure that this information has been
 * sent before the first occurrence of the process in other signals. The monitor
 * will make a best effort not to send information about the same process more
 * than once.
 *
 * @long This signal is sent under non-blocking flow control.
 *
 * @member MONITOR_PROCESS_NOTIFY
 * @member MonitorProcessInfo
 *
 * @seealso MonitorProcessInterceptNotify
 * @seealso MonitorGetProcessTraceReply
 */

/** @macro MONITOR_PROCESS_NOTIFY */
#define MONITOR_PROCESS_NOTIFY 39190   /* !-SIGNO(MonitorProcessInfo)-! */

/**
 * @signal MonitorSendNotify
 *
 * @brief Notification that a send has occurred.
 *
 * @long A signal has been sent and the action notify was active
 * in an actionpoint.
 *
 * @long This signal is sent under non-blocking flow control.
 *
 * @long The same struct, MonitorSendReceiveInfo, is used in
 * the signals:
 *    MONITOR_SEND_NOTIFY,
 *    MONITOR_RECEIVE_NOTIFY,
 *    MONITOR_SEND_INTERCEPT_NOTIFY,
 *    MONITOR_RECEIVE_INTERCEPT_NOTIFY,
 *    MONITOR_GET_SEND_TRACE_REPLY,
 *    MONITOR_GET_RECEIVE_TRACE_REPLY
 *
 * @member MONITOR_SEND_NOTIFY
 * @member MonitorSendReceiveInfo
 */

/** @macro MONITOR_SEND_NOTIFY */
#define MONITOR_SEND_NOTIFY 39191 /* !-SIGNO(MonitorSendReceiveInfo)-! */

/**
 * @signal MonitorReceiveNotify
 *
 * @brief Notification that a receive has occurred.
 *
 * @long A signal has been received by the receiving process
 * (from the process signal queue)
 * and the action notify was active in an actionpoint.
 *
 * @long This signal is sent under non-blocking flow control.
 *
 * @member MONITOR_RECEIVE_NOTIFY
 * @member MonitorSendReceiveInfo
 */

/** @macro MONITOR_RECEIVE_NOTIFY */
#define MONITOR_RECEIVE_NOTIFY 39192   /* !-SIGNO(MonitorSendReceiveInfo)-! */

/**
 * @signal MonitorCreateNotify
 *
 * @brief Notification that a create has occurred.
 *
 * @long A process has been created and the action notify was active
 * in an actionpoint.
 *
 * @long This signal is sent under non-blocking flow control.
 *
 * @member MONITOR_CREATE_NOTIFY
 * @member MonitorSwapCreateKillInfo
 */

/** @macro MONITOR_CREATE_NOTIFY */
#define MONITOR_CREATE_NOTIFY 39193 /* !-SIGNO(MonitorSwapCreateKillInfo)-! */

/**
 * @signal MonitorKillNotify
 *
 * @brief Notification that a kill has occurred.
 *
 * @long A process has been killed and the action notify was active
 * in an actionpoint.
 *
 * @long This signal is sent under non-blocking flow control.
 *
 * @member MONITOR_KILL_NOTIFY
 * @member MonitorSwapCreateKillInfo
 */

/** @macro MONITOR_KILL_NOTIFY */
#define MONITOR_KILL_NOTIFY 39194 /* !-SIGNO(MonitorSwapCreateKillInfo)-! */

/**
 * @signal MonitorSwapNotify
 *
 * @brief Notification that a swap (context switch) has occurred.
 *
 * @long A process swap has occurred and the action notify was active
 * in an actionpoint.
 *
 * @long This signal is sent under non-blocking flow control.
 *
 * @member MONITOR_SWAP_NOTIFY
 * @member MonitorSwapCreateKillInfo
 */

/** @macro MONITOR_SWAP_NOTIFY */
#define MONITOR_SWAP_NOTIFY 39195 /* !-SIGNO(MonitorSwapCreateKillInfo)-! */

/**
 * @signal MonitorErrorNotify
 *
 * @brief Notification that an error has occurred.
 *
 * @long An error occurred and the action notify was active
 * in an actionpoint.
 *
 * @long This signal is sent under non-blocking flow control.
 *
 * @member MONITOR_ERROR_NOTIFY
 * @member MonitorErrorInfo
 */

/** @macro MONITOR_ERROR_NOTIFY */
#define MONITOR_ERROR_NOTIFY 39196 /* !-SIGNO(MonitorErrorInfo)-! */

/**
 * @signal MonitorUserNotify
 *
 * @brief Notification that a user event occurred.
 *
 * @long A user event occurred and the action notify was active
 * in an actionpoint.
 *
 * @long This signal is sent under non-blocking flow control.
 *
 * @member MONITOR_USER_NOTIFY
 * @member MonitorUserInfo
 */

/** @macro MONITOR_USER_NOTIFY */
#define MONITOR_USER_NOTIFY 39197 /* !-SIGNO(MonitorUserInfo)-! */

/**
 * @signal MonitorBindNotify
 *
 * @brief Notification that a bind process (process migration)
 * is about to happen.
 *
 * @long A process is about to be migrated to another execution unit
 * and the action notify was active in an actionpoint. The signal
 * is thus sent before the migration actually is performed and one
 * signal is sent only for each process migrating, also if a block
 * or segment is migrated.
 *
 * @long This signal is sent under non-blocking flow control.
 *
 * @member MONITOR_BIND_NOTIFY
 * @member MonitorBindInfo
 */

/** @macro MONITOR_BIND_NOTIFY */
#define MONITOR_BIND_NOTIFY 39199 /* !-SIGNO(MonitorBindInfo)-! */

/**
 * @signal MonitorAllocNotify
 *
 * @brief Notification that a heap alloc occurred.
 *
 * @long An @link "ose.h" "alloc" occurred and the action notify was active
 * in an actionpoint.
 *
 * @long This signal is sent under non-blocking flow control.
 *
 * @member MONITOR_ALLOC_NOTIFY
 * @member MonitorAllocFreeBufInfo
 */

/** @macro MONITOR_ALLOC_NOTIFY */
#define MONITOR_ALLOC_NOTIFY 39200 /* !-SIGNO(MonitorAllocFreeBufInfo)-! */

/**
 * @signal MonitorFreeBufNotify
 *
 * @brief Notification that a heap free_buf occurred.
 *
 * @long An @link "ose.h" "free_buf" occurred and the action notify was active
 * in an actionpoint.
 *
 * @long This signal is sent under non-blocking flow control.
 *
 * @member MONITOR_FREE_BUF_NOTIFY
 * @member MonitorAllocFreeBufInfo
 */

/** @macro MONITOR_FREE_BUF_NOTIFY */
#define MONITOR_FREE_BUF_NOTIFY 39201 /* !-SIGNO(MonitorAllocFreeBufInfo)-! */

/**
 * @signal MonitorTimeoutNotify
 *
 * @brief Notification that a timeout occurred in a system call.
 * was active in an actionpoint.
 *
 * @long This signal is sent under non-blocking flow control.
 *
 * @member MONITOR_TIMEOUT_NOTIFY
 * @member MonitorTimeoutInfo
 * @member MONITOR_TIMEOUT_NOTIFY
 * @member MonitorTimeoutInfo
 */

/** @macro MONITOR_TIMEOUT_NOTIFY */
#define MONITOR_TIMEOUT_NOTIFY 39297 /* !-SIGNO(MonitorTimeoutInfo)-! */

/**
 * @signal MonitorLossNotify
 *
 * @brief Notification that notify signals from the monitor was lost.
 *
 * @long Notifies the client, that the monitor has thrown event notifications
 * due to flow control.
 *
 * @long This signal is sent under non-blocking flow control.
 *
 *
 * @member MONITOR_LOSS_NOTIFY
 * @member MonitorLossNotify
 */

/** @macro MONITOR_LOSS_NOTIFY */
#define MONITOR_LOSS_NOTIFY 39198   /* !-SIGNO(MonitorLossNotify)-! */

/**
 * @type MonitorLossNotify
 *
 * @brief Struct used as type and as several signals:
 *   MONITOR_LOSS_NOTIFY
 *   MONITOR_GET_LOSS_TRACE_REPLY
 *
 *
 * @output sigNo        Signal number of this notify or reply signal.
 * @output reference    Pseudo reference number for ordering of events/losses.
 * @output time         Timestamp (@link "" "MonitorEventTimestamp").
 * @output lostSize     Estimate of the number of bytes lost.
 */
typedef struct MonitorLossNotify
{
   uint32_t sigNo;
   uint32_t reference;
   MonitorEventTimestamp time;
   uint32_t lostSize;
} MonitorLossNotify;

/**
 * @type MonitorEventInfoHeader
 *
 * @brief Header struct used in MonitorEventInfo.
 *
 * @long This struct is used in the beginning of
 * the variable length struct @link "" "MonitorEventInfo",
 * to contain the size of the MonitorEventInfo struct.
 * E.g. see @link "" "MonitorGetTraceMultipleReply".
 *
 * @long In reply signals, containing a list with multiple events,
 * e.g. several variable length MonitorEventInfo structs, also the
 * total list size is given as the number of
 * sizeof(MonitorEventInfoHeaders) slots required to hold the entire list.
 *
 * @field size The size of the struct MonitorEventInfo (where this header
 *             is the first entry, given as the number
 *             of sizeof(MonitorEventInfoHeaders) slots needed to hold the
 *             entire MonitorEventInfo.
 * @field reserved0 NOT USED.
 */
typedef union MonitorEventInfoHeader
{
   uint32_t size;
   uint8_t reserved0;
} MonitorEventInfoHeader;

/**
 * @type MonitorEventInfo
 *
 * @brief Type holding information about an event.
 *
 * @long The @link "" "MonitorGetTraceMultipleReply" contains a list
 * of MonitorEventInfo structs with information about each trace event.
 *
 * @long The MonitorEventInfo has variable length, as it contains a union of
 * different types. It contains a header which is the struct
 * @link "" "MonitorEventInfoHeader". This header contains the actual length.
 * Its total length must always be a multiple of sizeof(MonitorEventInfoHeader)
 * and is actually given as the number of sizeof(MonitorEventInfoHeader) slots.
 *
 * @seealso MonitorEventInfoHeader
 * @seealso MonitorGetTraceMultipleReply
 *
 * @output header                A struct @link "" "MonitorEventInfoHeader",
 *                               containing the total length of this
 *                               MonitorEventInfo event description.
 *                               The length includes this header.
 * @output reserved0             NOT USED.
 * @output info                  Event information union which holds
 *                               one of the following types.
 * @output info.sigNo            Event signal number.
 * @output info.process          Process information.
 * @output info.loss             Information about event loss.
 * @output info.sendReceive      Send or receive information.
 * @output info.swapCreateKill   Swap, create or kill information.
 * @output info.error            Error information.
 * @output info.reset            Reset information.
 * @output info.user             User information.
 * @output info.bind             Bind information.
 * @output info.allocFreeBuf     Alloc or free_buf information.
 * @output info.timeout          Timeout information.
 */
typedef struct MonitorEventInfo
{
   MonitorEventInfoHeader header;
   uint32_t reserved0;
   union
   {
      uint32_t sigNo;
      MonitorProcessInfo process;
      MonitorLossNotify loss;
      MonitorSendReceiveInfo sendReceive;
      MonitorSwapCreateKillInfo swapCreateKill;
      MonitorErrorInfo error;
      MonitorResetInfo reset;
      MonitorUserInfo user;
      MonitorBindInfo bind;
      MonitorAllocFreeBufInfo allocFreeBuf;
      MonitorTimeoutInfo timeout;
   } info;
} MonitorEventInfo;

/*******************************************************************************
 *
 * TRACE SIGNALS
 *
 ******************************************************************************/

/**
 * @signal MonitorSetTraceEnabledRequest
 *
 * @brief Enable tracing for all clients request.
 *
 * @member MONITOR_SET_TRACE_ENABLED_REQUEST
 * @member MonitorSetTraceEnabledRequest
 */

/** @macro MONITOR_SET_TRACE_ENABLED_REQUEST */
#define MONITOR_SET_TRACE_ENABLED_REQUEST 39205 /* !-SIGNO(MonitorSetTraceEnabledRequest)-! */

/**
 * @type MonitorSetTraceEnabledRequest
 *
 * @input sigNo MONITOR_SET_TRACE_ENABLED_REQUEST
 * @input enabled Desired status 1=enabled, 0=disabled
 */
typedef struct MonitorSetTraceEnabledRequest
{
   uint32_t sigNo;
   MonitorBoolean enabled;
} MonitorSetTraceEnabledRequest;

/**
 * @signal MonitorSetTraceEnabledReply
 *
 * @brief Enable trace reply.
 *
 * @member MONITOR_SET_TRACE_ENABLED_REPLY
 * @member MonitorSetTraceEnabledReply
 */

/** @macro MONITOR_SET_TRACE_ENABLED_REPLY */
#define MONITOR_SET_TRACE_ENABLED_REPLY 39206 /* !-SIGNO(MonitorSetTraceEnabledReply)-! */

/**
 * @type MonitorSetTraceEnabledReply
 *
 * @output sigNo           MONITOR_SET_TRACE_ENABLED_REPLY
 * @output status          MONITOR_STATUS_OK or cause of failure
 */
typedef struct MonitorSetTraceEnabledReply
{
   uint32_t sigNo;
   MonitorStatus status;
} MonitorSetTraceEnabledReply;

/**
 * @signal MonitorGetTraceEnabledRequest
 *
 * @brief Get trace-enable status request.
 *
 * @member MONITOR_GET_TRACE_ENABLED_REQUEST
 * @member MonitorGetTraceEnabledRequest
 */

/** @macro MONITOR_GET_TRACE_ENABLED_REQUEST */
#define MONITOR_GET_TRACE_ENABLED_REQUEST 39207 /* !-SIGNO(MonitorGetTraceEnabledRequest)-! */

/**
 * @type MonitorGetTraceEnabledRequest
 *
 * @input sigNo MONITOR_GET_TRACE_ENABLED_REQUEST
 */
typedef struct MonitorGetTraceEnabledRequest
{
   uint32_t sigNo;
} MonitorGetTraceEnabledRequest;

/**
 * @signal MonitorGetTraceEnabledReply
 *
 * @brief Get trace-enable status reply.
 *
 * @member MONITOR_GET_TRACE_ENABLED_REPLY
 * @member MonitorGetTraceEnabledReply
 */

/** @macro MONITOR_GET_TRACE_ENABLED_REPLY */
#define MONITOR_GET_TRACE_ENABLED_REPLY 39208 /* !-SIGNO(MonitorGetTraceEnabledReply)-! */

/**
 * @type MonitorGetTraceEnabledReply
 *
 * @output sigNo           MONITOR_GET_TRACE_ENABLED_REPLY
 * @output status          MONITOR_STATUS_OK or cause of failure.
 * @output enabled         1 if trace is enabled, 0 otherwise.
 */
typedef struct MonitorGetTraceEnabledReply
{
   uint32_t sigNo;
   MonitorStatus status;
   MonitorBoolean enabled;
} MonitorGetTraceEnabledReply;

/**
 * @signal MonitorClearTraceRequest
 *
 * @brief Clear trace buffer contents request.
 *
 * @long This clears the entire trace buffer (common for all clients).
 *
 * @member MONITOR_CLEAR_TRACE_REQUEST
 * @member MonitorClearTraceRequest
 */

/** @macro MONITOR_CLEAR_TRACE_REQUEST */
#define MONITOR_CLEAR_TRACE_REQUEST 39209 /* !-SIGNO(MonitorClearTraceRequest)-! */

/**
 * @type MonitorClearTraceRequest
 *
 * @input sigNo MONITOR_CLEAR_TRACE_REQUEST
 */
typedef struct MonitorClearTraceRequest
{
   uint32_t sigNo;
} MonitorClearTraceRequest;

/**
 * @signal MonitorClearTraceReply
 *
 * @brief Clear trace buffer reply.
 *
 * @member MONITOR_CLEAR_TRACE_REPLY
 * @member MonitorClearTraceReply
 */

/** @macro MONITOR_CLEAR_TRACE_REPLY */
#define MONITOR_CLEAR_TRACE_REPLY 39210 /* !-SIGNO(MonitorClearTraceReply)-! */

/**
 * @type MonitorClearTraceReply
 *
 * @output sigNo   MONITOR_CLEAR_TRACE_REPLY
 * @output status  MONITOR_STATUS_OK or cause of failure.
 */
typedef struct MonitorClearTraceReply
{
   uint32_t sigNo;
   MonitorStatus status;
} MonitorClearTraceReply;


/**
 * @signal MonitorGetTraceMultipleRequest
 *
 * @brief Read trace and get trace entries collected in one reply signal
 * (trace can stay enabled).
 *
 * @long Used to read the trace with multiple entries
 * in one reply signal. This request also supports reading the trace while
 * the trace is enabled. Compare @link "" "MonitorGetTraceRequest", which
 * disables tracing to protect the trace when reading.
 *
 * @long If timeout is set to 0, the reply is sent directly. If timeout is
 * set to ~0, the reply is sent when the reply buffer is full. In all other
 * cases the reply is sent when the reply buffer is full or the timeout is reached.
 *
 * @long The reply buffer size is configurable with the krn/monitor/sigsize
 * system parameter (trace_mgets) - see RMM in the "OSE Core User's Guide".
 * Very large trace signals are dropped, if larger than this configured size,
 * and a MonitorGetLossTraceReply signal is sent to the client.
 *
 * @long For the first request in a sequence, the handle should be set to 0 to
 * read the oldest events in the trace buffer. Subsequent request shall use the
 * handle obtained in the previous reply signal.
 *
 * @long If events are lost, which can happen if the trace is enabled and
 * events are generated at a higher rate than the client is able to read, a
 * @link "" "MonitorGetLossTraceReply" is inserted in the sequence of events.
 *
 * @restriction There should only be one pending request per client.
 *
 * @member MONITOR_GET_TRACE_MULTIPLE_REQUEST
 * @member MonitorGetTraceMultipleRequest
 * @seealso MonitorGetTraceMultipleReply
 */

/** @macro MONITOR_GET_TRACE_MULTIPLE_REQUEST */
#define MONITOR_GET_TRACE_MULTIPLE_REQUEST 39211 /* !-SIGNO(MonitorGetTraceMultipleRequest)-! */

/**
 * @type MonitorGetTraceMultipleRequest
 *
 * @input sigNo   MONITOR_GET_TRACE_MULTIPLE_REQUEST
 * @input timeout 0(immediate), ~0(no timeout) or time out in milliseconds.
 * @input handle  0(first time) or handle obtained from previous reply.
 */
typedef struct MonitorGetTraceMultipleRequest
{
   uint32_t sigNo;
   uint32_t timeout;
   uint32_t handle;
} MonitorGetTraceMultipleRequest;

/**
 * @signal MonitorGetTraceMultipleReply
 *
 * @brief Read trace reply with multiple trace events.
 *
 * @long This reply signal contains multiple trace events. The number
 * depends on the timeout parameter in the request but also on
 * the configured max trace reply buffer size - see more in the description
 * of @link "" "MonitorGetTraceMultipleRequest".
 *
 * @long Each trace event is returned in a MonitorEventInfo struct,
 * which has variable size. Both the total size of each and the
 * total size of the event list are given as the number of
 * sizeof(MonitorEventInfoHeader) slots needed.
 *
 * @member MONITOR_GET_TRACE_MULTIPLE_REPLY
 * @member MonitorGetTraceMultipleReply
 * @seealso MonitorGetTraceMultipleRequest
 *
 * @example
 * MonitorGetTraceMultipleReply *reply;
 *
 * // Get info... //
 *
 * MonitorEventInfoHeader *e  = reply->events;
 * MonitorEventInfoHeader *e1 = reply->events + events->eventsCount;
 *
 * // For each event //
 * while (e < e1)
 * {
 *    MonitorEventInfo *event = (void *)e;
 *
 *    switch (event->info.sigNo)
 *    {
 *       case MONITOR_GET_PROCESS_TRACE_REPLY:
 *       {
 *          MonitorProcessInfo *info = &event->info.process;
 *
 *          printf("Process: %s (%lx) Block: %lx Segment: %lx"
 *                 "Type: %lx Entrypoint: %lx Properties: %lx\n",
 *                 info->name, info->pid, info->bid, info->sid,
 *                 info->type, info->entrypoint, info->properties);
 *          break;
 *       }
 *       case MONITOR_GET_SEND_TRACE_REPLY:
 *       {
 *          MonitorSendReceiveInfo *info = &event->info.sendReceive;
 *          // ... //
 *          break;
 *       }
 *
 *       // ... //
 *
 *       default:
 *          // Skip unknown events //
 *          break;
 *    }
 *    e += event->header.size;
 * }
 */

/** @macro MONITOR_GET_TRACE_MULTIPLE_REPLY */
#define MONITOR_GET_TRACE_MULTIPLE_REPLY 39212 /* !-SIGNO(MonitorGetTraceMultipleReply)-! */

/**
 * @type MonitorGetTraceMultipleReply
 *
 * @brief Multiple events reply.
 *
 * @output sigNo         MONITOR_GET_TRACE_MULTIPLE_REPLY
 * @output status        MONITOR_STATUS_OK or cause of failure.
 * @output handle        A handle to be passed in the next request to continue
 *                       read at the current location.
 * @output reserved0     NOT USED.
 * @output enabled       1 if trace is enabled and 0 otherwise.
 * @output eventsCount   Length of events (given as the number of
 *                       sizeof(MonitorEventInfoHeaders) slots needed.
 * @output events        List of @link "" "MonitorEventInfo" structs.
 */
typedef struct MonitorGetTraceMultipleReply
{
   uint32_t sigNo;
   MonitorStatus status;
   uint32_t handle;
   uint32_t reserved0;
   uint32_t enabled;
   uint32_t eventsCount;
   MonitorEventInfoHeader events[1];
} MonitorGetTraceMultipleReply;

/**
 * @signal MonitorGetTraceRequest
 *
 * @brief Read a range of trace entries request and disable tracing.
 *
 * @long Followed by zero or more of MonitorGetXxxxxxTraceReply signals
 * (depending on the trace type),
 * MonitorBlockingFlowControlRequest, and one MonitorGetTraceEndmark.
 *
 * @long The trace is automatically disabled, to protect the parts being read
 * from being overwritten. The trace needs to be explicitly enabled again if
 * needed. Compare the different @link "" "MonitorGetTraceMultipleRequest",
 * which does not disable tracing.
 *
 * @long Available events in the trace are numbered from 0 to N - 1,
 * where 0 is the oldest event and N is the total number of events
 * in the trace.
 *
 * @member MONITOR_GET_TRACE_REQUEST
 * @member MonitorGetTraceRequest
 */

/** @macro MONITOR_GET_TRACE_REQUEST */
#define MONITOR_GET_TRACE_REQUEST 39215 /* !-SIGNO(MonitorGetTraceRequest)-! */

/**
 * @type MonitorGetTraceRequest
 *
 * @input sigNo MONITOR_GET_TRACE_REQUEST
 *
 * @input from       The index of the first event to read. If the number
 *                   is greater than N, the monitor will return an unspecified
 *                   (but reasonable) number of events from the end of the trace.
 * @input to         The index of the last event to read + 1. If the number
 *                   is greater than N, the monitor will return events up to
 *                   the end of the trace.
 * @input reserved0  NOT USED. Should be 0.
 * @input reserved1  NOT USED. Should be 0.
 * @input reserved2  NOT USED. Should be 0.
 * @input reserved3  NOT USED. Should be 0.
 * @input tagsCount  NOT USED. Should be 0.
 * @input tags       NOT USED.
 */
typedef struct MonitorGetTraceRequest
{
   uint32_t sigNo;
   uint32_t from;
   uint32_t to;
   uint32_t reserved0;
   uint32_t reserved1;
   uint32_t reserved2;
   uint32_t reserved3;
   uint32_t tagsCount;
   MonitorTag tags[1];
} MonitorGetTraceRequest;

/**
 * @signal MonitorGetProcessTraceReply
 *
 * @brief Report with process details - for processes in later
 *        MonitorGetTraceReply signals.
 *
 * @long To avoid resending details about each process multiple times,
 * the monitor sends process detail reports separately, before
 * trace replies are sent referring to these processes.
 *
 * @long The monitor assumes that the client caches at least the last 32
 * reports with process details and avoids resending these. This cache
 * is not shared with other signals.
 * The requested size of this cache is actually returned from the
 * monitor to the client as a parameter in @link "" "MonitorAttachReply".
 *
 * @long The monitor will send this signal for all processes mentioned
 * in a sequence of MonitorGetTraceReply replies to provide additional information
 * about the processes. The monitor will ensure that this information has been
 * sent before the first occurrence of the process in other signals. The monitor
 * will make a best effort not to send information about the same process more
 * than once.
 *
 * @long This signal is sent under blocking flow control.
 *
 * @member MONITOR_GET_PROCESS_TRACE_REPLY
 * @member MonitorProcessInfo
 *
 * @seealso MonitorProcessNotify
 */

/** @macro MONITOR_GET_PROCESS_TRACE_REPLY */
#define MONITOR_GET_PROCESS_TRACE_REPLY 39216 /* !-SIGNO(MonitorProcessInfo)-! */

/**
 * @signal MonitorGetSendTraceReply
 *
 * @brief Trace entry type Send, reply to @link "" "MonitorGetTraceRequest".
 *
 * @long Before this is sent, the monitor will ensure it has sent process
 * details for the processes listed here to the client with
 * @link "" "MonitorGetProcessTraceReply" signals.
 *
 * @long This signal is sent under blocking flow control.
 *
 * @member MONITOR_GET_SEND_TRACE_REPLY
 * @member MonitorSendReceiveInfo
 */

/** @macro MONITOR_GET_SEND_TRACE_REPLY */
#define MONITOR_GET_SEND_TRACE_REPLY 39217 /* !-SIGNO(MonitorSendReceiveInfo)-! */

/**
 * @signal MonitorGetReceiveTraceReply
 *
 * @brief Trace entry type Receive, reply to @link "" "MonitorGetTraceRequest".
 *
 * @long This signal is sent under blocking flow control.
 *
 * @member MONITOR_GET_RECEIVE_TRACE_REPLY
 * @member MonitorSendReceiveInfo
 */

/** @macro MONITOR_GET_RECEIVE_TRACE_REPLY */
#define MONITOR_GET_RECEIVE_TRACE_REPLY 39218 /* !-SIGNO(MonitorSendReceiveInfo)-! */

/**
 * @signal MonitorGetCreateTraceReply
 *
 * @brief Trace entry type Create, reply to @link "" "MonitorGetTraceRequest".
 *
 * @long This signal is sent under blocking flow control.
 *
 * @member MONITOR_GET_CREATE_TRACE_REPLY
 * @member MonitorSwapCreateKillInfo
 */

/** @macro MONITOR_GET_CREATE_TRACE_REPLY */
#define MONITOR_GET_CREATE_TRACE_REPLY 39219 /* !-SIGNO(MonitorSwapCreateKillInfo)-! */

/**
 * @signal MonitorGetKillTraceReply
 *
 * @brief Trace entry type Kill, reply to @link "" "MonitorGetTraceRequest".
 *
 * @long This signal is sent under blocking flow control.
 *
 * @member MONITOR_GET_KILL_TRACE_REPLY
 * @member MonitorSwapCreateKillInfo
 */

/** @macro MONITOR_GET_KILL_TRACE_REPLY */
#define MONITOR_GET_KILL_TRACE_REPLY 39220 /* !-SIGNO(MonitorSwapCreateKillInfo)-! */

/**
 * @signal MonitorGetSwapTraceReply
 *
 * @brief Trace entry type Swap, reply to @link "" "MonitorGetTraceRequest".
 *
 * @long This signal is sent under blocking flow control.
 *
 * @member MONITOR_GET_SWAP_TRACE_REPLY
 * @member MonitorSwapCreateKillInfo
 */

/** @macro MONITOR_GET_SWAP_TRACE_REPLY */
#define MONITOR_GET_SWAP_TRACE_REPLY 39221 /* !-SIGNO(MonitorSwapCreateKillInfo)-! */

/**
 * @signal MonitorGetErrorTraceReply
 *
 * @brief Trace entry type Error, reply to @link "" "MonitorGetTraceRequest".
 *
 * @long This signal is sent under blocking flow control.
 *
 * @member MONITOR_GET_ERROR_TRACE_REPLY
 * @member MonitorErrorInfo
 */

/** @macro MONITOR_GET_ERROR_TRACE_REPLY */
#define MONITOR_GET_ERROR_TRACE_REPLY 39222 /* !-SIGNO(MonitorErrorInfo)-! */

/**
 * @signal MonitorGetResetTraceReply
 *
 * @brief Trace entry type Reset, reply to @link "" "MonitorGetTraceRequest".
 *
 * @long This signal is sent under blocking flow control.
 *
 * @member MONITOR_GET_RESET_TRACE_REPLY
 * @member MonitorResetInfo
 */

/** @macro MONITOR_GET_RESET_TRACE_REPLY */
#define MONITOR_GET_RESET_TRACE_REPLY 39223 /* !-SIGNO(MonitorResetInfo)-! */

/**
 * @signal MonitorGetUserTraceReply
 *
 * @brief Trace entry type User Event, reply to @link "" "MonitorGetTraceRequest".
 *
 * @long This signal is sent under blocking flow control.
 *
 * @member MONITOR_GET_USER_TRACE_REPLY
 * @member MonitorUserInfo
 */

/** @macro MONITOR_GET_USER_TRACE_REPLY */
#define MONITOR_GET_USER_TRACE_REPLY 39224 /* !-SIGNO(MonitorUserInfo)-! */

/**
 * @signal MonitorGetBindTraceReply
 *
 * @brief Trace entry type Bind (process migration), reply to @link "" "MonitorGetTraceRequest".
 *
 * @long This signal is sent under blocking flow control.
 *
 * @member MONITOR_GET_BIND_TRACE_REPLY
 * @member MonitorBindInfo
 */

/** @macro MONITOR_GET_BIND_TRACE_REPLY */
#define MONITOR_GET_BIND_TRACE_REPLY 39225 /* !-SIGNO(MonitorBindInfo)-! */

/**
 * @signal MonitorGetAllocTraceReply
 *
 * @brief Trace entry type Alloc, reply to @link "" "MonitorGetTraceRequest".
 *
 * @long This signal is sent under blocking flow control.
 *
 * @member MONITOR_GET_ALLOC_TRACE_REPLY
 * @member MonitorAllocFreeBufInfo
 */

/** @macro MONITOR_GET_ALLOC_TRACE_REPLY */
#define MONITOR_GET_ALLOC_TRACE_REPLY 39226 /* !-SIGNO(MonitorAllocFreeBufInfo)-! */

/**
 * @signal MonitorGetFreeBufTraceReply
 *
 * @brief Trace entry type FreeBuf, reply to @link "" "MonitorGetTraceRequest".
 *
 * @long This signal is sent under blocking flow control.
 *
 * @member MONITOR_GET_FREE_BUF_TRACE_REPLY
 * @member MonitorAllocFreeBufInfo
 */

/** @macro MONITOR_GET_FREE_BUF_TRACE_REPLY */
#define MONITOR_GET_FREE_BUF_TRACE_REPLY 39227 /* !-SIGNO(MonitorAllocFreeBufInfo)-! */

/**
 * @signal MonitorGetTimeoutTraceReply
 *
 * @brief Trace entry type Timeout, reply to @link "" "MonitorGetTraceRequest".
 *
 * @long This signal is sent under blocking flow control.
 *
 * @member MONITOR_GET_TIMEOUT_TRACE_REPLY
 * @member MonitorTimeoutInfo
 */

/** @macro MONITOR_GET_TIMEOUT_TRACE_REPLY */
#define MONITOR_GET_TIMEOUT_TRACE_REPLY 39296 /* !-SIGNO(MonitorTimeoutInfo)-! */

/**
 * @signal MonitorGetLossTraceReply
 *
 * @brief Lost trace entries reply.
 *
 * @long Notifies the client, that the monitor has thrown events in the trace,
 * which can happen if the trace is enabled and events are generated at a
 * higher rate than the client is able to read in replies to
 * @link "" "MonitorGetTraceRequest".
 *
 * @long This signal is sent under blocking flow control.
 *
 *
 * @member MONITOR_GET_LOSS_TRACE_REPLY
 * @member MonitorLossNotify
 */

/** @macro MONITOR_GET_LOSS_TRACE_REPLY */
#define MONITOR_GET_LOSS_TRACE_REPLY 39228 /* !-SIGNO(MonitorLossNotify)-! */

/**
 * @signal MonitorGetTraceEndmark
 *
 * @brief Read trace endmark, with status.
 *
 * @member MONITOR_GET_TRACE_ENDMARK
 * @member MonitorGetTraceEndmark
 * @seealso MonitorGetTraceRequest
 * @seealso MonitorGetTraceReply
 */

/** @macro MONITOR_GET_TRACE_ENDMARK */
#define MONITOR_GET_TRACE_ENDMARK 39230 /* !-SIGNO(MonitorGetTraceEndmark)-! */

/**
 * @type MonitorGetTraceEndmark
 *
 * @output sigNo   MONITOR_GET_TRACE_ENDMARK
 * @output status  MONITOR_STATUS_OK or cause of failure.
 */
typedef struct MonitorGetTraceEndmark
{
   uint32_t sigNo;
   uint32_t status;
} MonitorGetTraceEndmark;


/*******************************************************************************
 *
 * INTERCEPT RELATED SIGNALS
 *
 ******************************************************************************/

/**
 * @signal MonitorProcessInterceptNotify
 *
 * @brief Report with process details - for processes
 *        in later MonitorXxxxxInterceptNotify replies.
 *
 * @long To avoid resending details about each process multiple times,
 * the monitor sends process detail reports separately, before
 * notifications are sent referring to these processes.
 *
 * @long The monitor assumes that the client caches at least the last 32
 * reports with process details and avoids resending these.
 * The cache, with 32 reports, is shared by the report signals
 * MonitorProcessNotify and MonitorProcessInterceptNotify, which
 * contain the same process details.
 * The requested size of this cache is actually returned from the
 * monitor to the client as a parameter in @link "" "MonitorAttachReply".
 *
 * @long The monitor will send this MonitorProcessInterceptNotify signal for
 * all processes mentioned in a sequence of
 * MonitorXxxxxxInterceptNotify signals (but not MonitorXxxxxxNotify)
 * to provide additional information about the processes.
 * The monitor will ensure that this information has been
 * sent before the first occurrence of the process in other signals. The monitor
 * will make a best effort not to send information about the same process more
 * than once.
 *
 * @member MONITOR_PROCESS_INTERCEPT_NOTIFY
 * @member MonitorProcessInfo
 *
 * @seealso MonitorProcessNotify
 * @seealso MonitorGetProcessTraceReply
 */

/** @macro MONITOR_PROCESS_INTERCEPT_NOTIFY */
#define MONITOR_PROCESS_INTERCEPT_NOTIFY 39231   /* !-SIGNO(MonitorProcessInfo)-! */

/**
 * @signal MonitorSendInterceptNotify
 *
 * @brief Intercept notification for a send, causing intercept.
 *
 * @long A send event occurred, the action intercept notify was active
 * in an actionpoint and the scope in the actionpoint has been intercepted.
 *
 * @member MONITOR_SEND_INTERCEPT_NOTIFY
 * @member MonitorSendReceiveInfo
 */

/** @macro MONITOR_SEND_INTERCEPT_NOTIFY */
#define MONITOR_SEND_INTERCEPT_NOTIFY 39232 /* !-SIGNO(MonitorSendReceiveInfo)-! */

/**
 * @signal MonitorReceiveInterceptNotify
 *
 * @brief Intercept notification for a receive, causing intercept.
 *
 * @long A receive event occurred, the action intercept notify was active
 * in an actionpoint and the scope in the actionpoint has been intercepted.
 *
 * @member MONITOR_RECEIVE_INTERCEPT_NOTIFY
 * @member MonitorSendReceiveInfo
 */

/** @macro MONITOR_RECEIVE_INTERCEPT_NOTIFY */
#define MONITOR_RECEIVE_INTERCEPT_NOTIFY 39233   /* !-SIGNO(MonitorSendReceiveInfo)-! */

/**
 * @signal MonitorCreateInterceptNotify
 *
 * @brief Intercept notification for a create, causing intercept.
 *
 * @long A create event occurred, the action intercept notify was active
 * in an actionpoint and the scope in the actionpoint has been intercepted.
 *
 * @long The caller and the created process IDs are given.
 *
 * @member MONITOR_CREATE_INTERCEPT_NOTIFY
 * @member MonitorSwapCreateKillInfo
 */

/** @macro MONITOR_CREATE_INTERCEPT_NOTIFY */
#define MONITOR_CREATE_INTERCEPT_NOTIFY 39234 /* !-SIGNO(MonitorSwapCreateKillInfo)-! */

/**
 * @signal MonitorKillInterceptNotify
 *
 * @brief Intercept notification for a kill, causing intercept.
 *
 * @long A kill event occurred, the action intercept notify was active
 * in an actionpoint and the scope in the actionpoint has been intercepted.
 *
 * @long The caller and the killed process IDs are given.
 *
 * @member MONITOR_KILL_INTERCEPT_NOTIFY
 * @member MonitorSwapCreateKillInfo
 */

/** @macro MONITOR_KILL_INTERCEPT_NOTIFY */
#define MONITOR_KILL_INTERCEPT_NOTIFY 39235 /* !-SIGNO(MonitorSwapCreateKillInfo)-! */

/**
 * @signal MonitorSwapInterceptNotify
 *
 * @brief Intercept notification for a swap (context switch),
 * causing intercept.
 *
 * @long A swap event occurred, the action intercept notify was active
 * in an actionpoint and the scope in the actionpoint has been intercepted.
 *
 * @long The swapped out and swapped in process IDs are given.
 *
 * @member MONITOR_SWAP_INTERCEPT_NOTIFY
 * @member MonitorSwapCreateKillInfo
 */

/** @macro MONITOR_SWAP_INTERCEPT_NOTIFY */
#define MONITOR_SWAP_INTERCEPT_NOTIFY 39236 /* !-SIGNO(MonitorSwapCreateKillInfo)-! */

/**
 * @signal MonitorErrorInterceptNotify
 *
 * @brief Intercept notification for an error, causing intercept.
 *
 * @long An error event occurred, the action intercept notify was active
 * in an actionpoint and the scope in the actionpoint has been intercepted.
 *
 * @member MONITOR_ERROR_INTERCEPT_NOTIFY
 * @member MonitorErrorInfo
 */

/** @macro MONITOR_ERROR_INTERCEPT_NOTIFY */
#define MONITOR_ERROR_INTERCEPT_NOTIFY 39237 /* !-SIGNO(MonitorErrorInfo)-! */

/**
 * @signal MonitorUserInterceptNotify
 *
 * @brief Intercept notification for a user event, causing intercept.
 *
 * @long A user event occurred, the action intercept notify was active
 * in an actionpoint and the scope in the actionpoint has been intercepted.
 *
 * @member MONITOR_USER_INTERCEPT_NOTIFY
 * @member MonitorUserInfo
 */

/** @macro MONITOR_USER_INTERCEPT_NOTIFY */
#define MONITOR_USER_INTERCEPT_NOTIFY 39238 /* !-SIGNO(MonitorUserInfo)-! */

/**
 * @signal MonitorBindInterceptNotify
 *
 * @brief Intercept notification for a bind (process migration),
 * causing intercept.
 *
 * @long A process migration event occurred, the action intercept notify
 * was active in an actionpoint and the scope in the actionpoint has been intercepted.
 *
 * @member MONITOR_BIND_INTERCEPT_NOTIFY
 * @member MonitorBindInfo
 */

/** @macro MONITOR_BIND_INTERCEPT_NOTIFY */
#define MONITOR_BIND_INTERCEPT_NOTIFY 39239 /* !-SIGNO(MonitorBindInfo)-! */

/**
 * @signal MonitorAllocInterceptNotify
 *
 * @brief Intercept notification for a heap alloc, causing intercept.
 *
 * @long An alloc event occurred, the action intercept notify was active
 * in an actionpoint and the scope in the actionpoint has been intercepted.
 *
 * @member MONITOR_ALLOC_INTERCEPT_NOTIFY
 * @member MonitorAllocFreeBufInfo
 */

/** @macro MONITOR_ALLOC_INTERCEPT_NOTIFY */
#define MONITOR_ALLOC_INTERCEPT_NOTIFY 39244 /* !-SIGNO(MonitorAllocFreeBufInfo)-! */

/**
 * @signal MonitorFreeBufInterceptNotify
 *
 * @brief Intercept notification for a heap free_buf, causing intercept.
 *
 * @long A free_buf event occurred, the action intercept notify was active
 * in an actionpoint and the scope in the actionpoint has been intercepted.
 *
 * @member MONITOR_FREE_BUF_INTERCEPT_NOTIFY
 * @member MonitorAllocFreeBufInfo
 */

/** @macro MONITOR_FREE_BUF_INTERCEPT_NOTIFY */
#define MONITOR_FREE_BUF_INTERCEPT_NOTIFY 39245 /* !-SIGNO(MonitorAllocFreeBufInfo)-! */

/**
 * @signal MonitorTimeoutInterceptNotify
 *
 * @brief Intercept notification for a timeout in a system call.
 * the scope in the actionpoint has been intercepted.
 *
 * @member MONITOR_TIMEOUT_INTERCEPT_NOTIFY
 * @member MonitorTimeoutInfo
 */

/** @macro MONITOR_TIMEOUT_INTERCEPT_NOTIFY */
#define MONITOR_TIMEOUT_INTERCEPT_NOTIFY 39298 /* !-SIGNO(MonitorTimeoutInfo)-! */

/**
 * @signal MonitorSetInterceptEnabledRequest
 *
 * @brief Enable / disable interception.
 *
 * @long Interception is enabled by default.
 *
 * @long Disabling intercept while the attached scope is
 * intercepted will resume the scope.
 *
 * @long The client's intercept subscope will limit the scope
 * of processes that can be intercepted by events with
 * actionpoints defined by this client.
 *
 * @restriction The client must be attached to a scope.
 * @break
 * Another restriction is in current implementation, this is
 * an invalid request, unless the MONITOR_ATTACH_TAG_EVENT_LOCK was
 * set when attaching the scopes.
 *
 * @member MONITOR_SET_INTERCEPT_ENABLED_REQUEST
 * @member MonitorSetInterceptEnabledRequest
 */

/** @macro MONITOR_SET_INTERCEPT_ENABLED_REQUEST */
#define MONITOR_SET_INTERCEPT_ENABLED_REQUEST 39246 /* !-SIGNO(MonitorSetInterceptEnabledRequest)-! */

/**
 * @type MonitorSetInterceptEnabledRequest
 *
 * @input sigNo           MONITOR_SET_INTERCEPT_ENABLED_REQUEST
 * @input enabled         Desired status; 1=enabled, 0=disabled.
 * @input reserved0       NOT USED but must contain zero.
 */
typedef struct MonitorSetInterceptEnabledRequest
{
   uint32_t sigNo;
   MonitorBoolean enabled;
   uint32_t reserved0;
} MonitorSetInterceptEnabledRequest;

/**
 * @signal MonitorSetInterceptEnabledReply
 *
 * @brief Enable / disable intercept reply.
 *
 * @member MONITOR_SET_INTERCEPT_ENABLED_REPLY
 * @member MonitorSetInterceptEnabledReply
 */

/** @macro MONITOR_SET_INTERCEPT_ENABLED_REPLY */
#define MONITOR_SET_INTERCEPT_ENABLED_REPLY 39247 /* !-SIGNO(MonitorSetInterceptEnabledReply)-! */

/**
 * @type MonitorSetInterceptEnabledReply
 *
 * @output sigNo           MONITOR_SET_INTERCEPT_ENABLED_REPLY
 * @output status          MONITOR_STATUS_OK or cause of failure
 * @output reserved0       NOT USED.
 */
typedef struct MonitorSetInterceptEnabledReply
{
   uint32_t sigNo;
   MonitorStatus status;
   uint32_t reserved0;
} MonitorSetInterceptEnabledReply;

/**
 * @signal MonitorGetInterceptEnabledRequest
 *
 * @brief Get status of interception enablement for this client request.
 *
 * @member MONITOR_GET_INTERCEPT_ENABLED_REQUEST
 * @member MonitorGetInterceptEnabledRequest
 *
 * @seealso MonitorSetInterceptEnabledRequest
 */

/** @macro MONITOR_GET_INTERCEPT_ENABLED_REQUEST */
#define MONITOR_GET_INTERCEPT_ENABLED_REQUEST 39248 /* !-SIGNO(MonitorGetInterceptEnabledRequest)-! */

/**
 * @type MonitorGetInterceptEnabledRequest
 *
 * @input sigNo           MONITOR_GET_INTERCEPT_ENABLED_REQUEST
 * @input reserved0       NOT USED but must contain zero.
 */
typedef struct MonitorGetInterceptEnabledRequest
{
   uint32_t sigNo;
   uint32_t reserved0;
} MonitorGetInterceptEnabledRequest;

/**
 * @signal MonitorGetInterceptEnabledReply
 *
 * @brief Get status of interception enablement reply.
 *
 * @member MONITOR_GET_INTERCEPT_ENABLED_REPLY
 * @member MonitorGetInterceptEnabledReply
 *
 * @seealso MonitorGetInterceptEnabledRequest
 * @seealso MonitorSetInterceptEnabledRequest
 */

/** @macro MONITOR_GET_INTERCEPT_ENABLED_REPLY */
#define MONITOR_GET_INTERCEPT_ENABLED_REPLY 39249 /* !-SIGNO(MonitorGetInterceptEnabledReply)-! */

/**
 * @type MonitorGetInterceptEnabledReply
 *
 * @output sigNo           MONITOR_GET_INTERCEPT_ENABLED_REPLY
 * @output status          MONITOR_STATUS_OK or cause of failure
 * @output enabled         1 if enabled 0 otherwise.
 * @output reserved0       NOT USED.
 */
typedef struct MonitorGetInterceptEnabledReply
{
   uint32_t sigNo;
   MonitorStatus status;
   MonitorBoolean enabled;
   uint32_t reserved0;
} MonitorGetInterceptEnabledReply;


/*******************************************************************************
 *
 * SCOPE SIGNALS
 *
 ******************************************************************************/

/**
 * @signal MonitorSetScopeRequest
 *
 * @brief Redefine the debug main scope request.
 *
 * @restriction Accepts only scopes of types MONITOR_SCOPE_ID,
 * MONITOR_SCOPE_BLOCK_ID, MONITOR_SCOPE_SEGMENT_ID and MONITOR_SCOPE_SYSTEM
 * (@link "" "MonitorScopeTypeValue").
 * @break
 * Another restriction is that it is supported only for clients that
 * have attached for exclusive access to the event system
 * with MONITOR_ATTACH_TAG_EVENT_LOCK (@link "monitor_execution.sig"
 * "MonitorAttachTagValue").
 *
 * @member MONITOR_SET_SCOPE_REQUEST
 * @member MonitorSetScopeRequest
 * @seealso MonitorScopeTypeValue
 * @seealso MonitorAttachRequest
 */

/** @macro MONITOR_SET_SCOPE_REQUEST */
#define MONITOR_SET_SCOPE_REQUEST 39240 /* !-SIGNO(MonitorSetScopeRequest)-! */

/**
 * @signal MonitorSetScopeReply
 *
 * @brief Redefine the debug scope reply.
 *
 * @member MONITOR_SET_SCOPE_REPLY
 * @member MonitorSetScopeReply
 */

/** @macro MONITOR_SET_SCOPE_REPLY */
#define MONITOR_SET_SCOPE_REPLY 39241 /* !-SIGNO(MonitorSetScopeReply)-! */

/**
 * @signal MonitorSetInterceptScopeRequest
 *
 * @brief Redefine the intercept scope request.
 *
 *
 * @restriction Accepts only scopes of types MONITOR_SCOPE_ID,
 * MONITOR_SCOPE_BLOCK_ID, MONITOR_SCOPE_SEGMENT_ID and MONITOR_SCOPE_SYSTEM
 * (@link "" "MonitorScopeTypeValue").
 * @break
 * Another restriction is that it is supported only for clients that
 * have attached with for exclusive access to the event system
 * with MONITOR_ATTACH_TAG_EVENT_LOCK (@link "monitor_execution.sig"
 * "MonitorAttachTagValue").
 *
 * @member MONITOR_SET_INTERCEPT_SCOPE_REQUEST
 * @member MonitorSetScopeRequest
 * @seealso MonitorSetInterceptEnabledRequest
 */

/** @macro MONITOR_SET_INTERCEPT_SCOPE_REQUEST */
#define MONITOR_SET_INTERCEPT_SCOPE_REQUEST 39242 /* !-SIGNO(MonitorSetScopeRequest)-! */

/**
 * @signal MonitorSetInterceptScopeReply
 *
 * @brief Redefine the intercept scope reply.
 *
 *
 * @member MONITOR_SET_INTERCEPT_SCOPE_REPLY
 * @member MonitorSetScopeReply
 */

/** @macro MONITOR_SET_INTERCEPT_SCOPE_REPLY */
#define MONITOR_SET_INTERCEPT_SCOPE_REPLY 39243 /* !-SIGNO(MonitorSetScopeReply)-! */

/**
 * @type MonitorSetScopeRequest
 *
 * @input sigNo       MONITOR_SET_SCOPE_REQUEST or
 *                    MONITOR_SET_INTERCEPT_SCOPE_REQUEST
 * @input reserved0   NOT USED. Should be 0.
 * @input scope       The scope to limit reported events to.
 * @input tagsCount   The length of the tags list, as the
 *                    total number of sizeof(MonitorTag)
 *                    entries, with tags and arguments.
 * @input tags        Tags.
 */
typedef struct MonitorSetScopeRequest
{
   uint32_t sigNo;
   uint32_t reserved0;
   MonitorScope scope;
   uint32_t tagsCount;
   MonitorTag tags[1];
} MonitorSetScopeRequest;

/**
 * @type MonitorSetScopeReply
 *
 * @output sigNo     MONITOR_SET_SCOPE_REPLY or
 *                   MONITOR_SET_INTERCEPT_SCOPE_REPLY
 * @output status    MONITOR_STATUS_OK or cause of failure.
 */
typedef struct MonitorSetScopeReply
{
   uint32_t sigNo;
   MonitorStatus status;
} MonitorSetScopeReply;

#ifdef __cplusplus
} /* end of extern "C" */
#endif

#endif /* #ifndef _MONITOR_EVENTS_SIG */
