/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2005-2006 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */
/**
 * @toc OSE:
 * @file ose_spi/monitor_execution.sig
 *
 * @brief OSE System source code debugging and basic events signal interface.
 *
 * @long This header file defines the signal interface to be used by OSE debug
 * clients (typically source code debuggers such as Optima, GDB) that use
 * intercepts, resume, stepping and breakpoints to control the execution of
 * a scope of processes, to set memory or get/set CPU registers.
 *
 * @long It also contains a basic event system, limited to the needs of a
 * source code debugger. A few notification reports can be sent to clients
 * at error events or when processes have been created or killed.
 * Intercept notifications can be sent when breakpoints are fired.
 *
 * @long A connection to the RMM is needed, see monitor.sig for the actual
 * connection procedure using @link "" "MonitorInterfaceRequest" and
 * @link "" "MonitorConnectRequest".
 *
 * @long Attaching is required, to define a main scope and an intercept scope
 * for this connected client, using @link "" "MonitorAttachRequest".
 * Information about the monitor execution support is returned in the reply
 * (@link "" "MonitorAttachReply") after an attach. This contains typically
 * lists of supported request signals of different types and other information.
 * Other monitor interface support information is in the
 * @link "" "MonitorInterfaceReply" and @link "" "MonitorConnectReply".
 *
 * @long All request signals will be limited to within the main scope,
 * i.e. affects only a scope of resources, typically a range of
 * processes (serving to define the client's range of interest).
 * Intercept requests are limited to the given intercept scope, which
 * must be a subset of the main scope.
 *
 * @long When attaching, it might be needed (depending on MonitorFeature)
 * to lock the entire event system for this client, to be able to use the
 * basic event system. See @link "" "MonitorAttachTagValue" for details.
 *
 * @long It is recommended to detach from the scope via @link ""
 * "MonitorDetachRequest" before disconnecting with @link ""
 * "MonitorDisconnectRequest", allowing the monitor to
 * deallocate resources linked to the client.
 *
 * @long Multiple clients can use this interface to fetch information,
 * but only one client at a time can control the system and
 * use the basic event system and intercepting.
 *
 * @long Memory addresses used in the interface are always logical addresses
 * within the affected process or segment.
 *
 * @restriction Currently the basic event signals MonitorBasicXxxxx in this
 * interface can not be used at the same time as the event system
 * in monitor_events.sig.
 * @break
 * SFK (OSE Soft Kernel Environment) has not full support for this debugging
 * interface. SFK does not support the requests to set/get registers
 * or to use breakpoints or stepping.
 *
 *
 * @seealso ose_spi/monitor.sig
 * @seealso ose_spi/monitor_events.sig
 * @seealso ose_rmm_set_req_enabled
 * @seealso ose_epm_init_hwbp
 */

#ifndef _MONITOR_EXECUTION_SIG
#define _MONITOR_EXECUTION_SIG

#include "stdint.h"
#include "monitor.sig"

#ifdef __cplusplus
extern "C" {
#endif

/*******************************************************************************
 *
 * MEMORY SIGNALS
 *
 ******************************************************************************/

/**
 * @signal MonitorSetMemoryRequest
 *
 * @brief Write memory request.
 *
 * @long If all the requested data can not be written, the monitor will attempt
 * to write a smaller amount of data at the specified address. The address range
 * written will be sent to the client and an error
 * (MONITOR_STATUS_SET_MEMORY_ERROR) will be reported in the
 * MonitorSetMemoryReply signal.
 *
 * @member MONITOR_SET_MEMORY_REQUEST
 * @member MonitorSetMemoryRequest
 *
 * @seealso MonitorSetMemoryReply
 * @seealso MonitorGetMemoryRequest
 */

/** @macro MONITOR_SET_MEMORY_REQUEST */
#define MONITOR_SET_MEMORY_REQUEST 39103 /* !-SIGNO(MonitorSetMemoryRequest)-! */

/**
 * @type MonitorSetMemoryRequest
 *
 * @input sigNo     MONITOR_SET_MEMORY_REQUEST
 * @input pid       Address is in the context of this process, block or segment.
 * @input address   Memory address to start writing from.
 * @input dataCount Number of bytes in data.
 * @input data      The data.
 */
typedef struct MonitorSetMemoryRequest
{
   uint32_t sigNo;
   uint32_t pid;
   uint32_t address;
   uint32_t dataCount;
   uint8_t data[1];
} MonitorSetMemoryRequest;

/**
 * @signal MonitorSetMemoryReply
 *
 * @brief Write memory reply.
 *
 * @member MONITOR_SET_MEMORY_REPLY
 * @member MonitorSetMemoryReply
 * @seealso MonitorSetMemoryRequest
 */

/** @macro MONITOR_SET_MEMORY_REPLY */
#define MONITOR_SET_MEMORY_REPLY 39104 /* !-SIGNO(MonitorSetMemoryReply)-! */

/**
 * @type MonitorSetMemoryReply
 *
 * @output sigNo   MONITOR_SET_MEMORY_REPLY
 * @output status  MONITOR_STATUS_OK or cause of error.
 * @output pid     Address is in the context of this process or block.
 * @output address The address at which data was written.
 * @output size    Number of bytes successfully written.
 */
typedef struct MonitorSetMemoryReply
{
   uint32_t sigNo;
   MonitorStatus status;
   uint32_t pid;
   uint32_t address;
   uint32_t size;
} MonitorSetMemoryReply;

/*******************************************************************************
 *
 * REGISTER SIGNALS
 *
 ******************************************************************************/

/**
 * @type MonitorRegisterId
 *
 * @brief Register identity type.
 *
 * @seealso MonitorRegister
 */
typedef uint32_t MonitorRegisterId;

/**
 * @type MonitorRegister
 *
 * @brief Representation of a CPU register value (register ID and value).
 *
 * @long This is used in requests to get or set CPU register values.
 *
 * @field id    The register ID.
 * @field value The value of the register.
 *
 * @seealso MonitorRegisterId
 */
typedef struct MonitorRegister
{
   MonitorRegisterId id;
   uint32_t value;
} MonitorRegister;

/**
 * @macro MONITOR_REGISTER_INVALID
 * @brief The bit mask for the invalid bit in a register ID.
 */
#define MONITOR_REGISTER_INVALID 0x80000000

/**
 * @macro MONITOR_REGISTER_IS_INVALID
 * @brief Macro which is true if the invalid bit in a register ID is set.
 *
 * @input MONITOR_REGISTER The register ID to check.
 * @seealso MONITOR_REGISTER_ID
 * @seealso MonitorRegister
 */
#define MONITOR_REGISTER_IS_INVALID(MONITOR_REGISTER) \
   (((MONITOR_REGISTER) & MONITOR_REGISTER_INVALID) >> 31)

/**
 * @macro MONITOR_REGISTER_ID
 * @brief Macro which strips the invalid bit from a register ID.
 *
 * @input MONITOR_REGISTER The register ID to access.
 * @seealso MONITOR_REGISTER_IS_INVALID
 * @seealso MonitorRegister
 */
#define MONITOR_REGISTER_ID(MONITOR_REGISTER) \
   ((MONITOR_REGISTER) & 0x7fffffff)

/**
 * @type MonitorArmRegisterIdValue
 *
 * @brief Register identities used for register access on ARM.
 *
 * @field    MONITOR_ARM_REGISTER_R0                    R0 - R12 are contiguous
 * @field    MONITOR_ARM_REGISTER_PC                    Program counter
 * @field    MONITOR_ARM_REGISTER_CPSR                  Current Processor Status
 * @field    MONITOR_ARM_REGISTER_SSLIMIT               Supervisor stack location
 * @field    MONITOR_ARM_REGISTER_SP_CURR               Current SP register
 * @field    MONITOR_ARM_REGISTER_LR_CURR               Current LR register
 * @seealso MonitorRegister
 */
enum MonitorArmRegisterIdValue
{
   MONITOR_ARM_REGISTER_R0 = 0,
   MONITOR_ARM_REGISTER_PC = 13,
   MONITOR_ARM_REGISTER_CPSR = 14,

   /* Pseudo registers */
   MONITOR_ARM_REGISTER_SSLIMIT = 255,
   MONITOR_ARM_REGISTER_SP_CURR = 258,
   MONITOR_ARM_REGISTER_LR_CURR = 259
};

/**
 * @type MonitorMipsRegisterIdValue
 *
 * @brief Register identities used for register access on MIPS.
 *
 * @long OSE MIPS targets may use the O32 or N32 calling conventions and if N32
 * is used the common integer registers are 64 bit, for O32 they are 32 bit.
 *
 * @long For 64 bit registers there are two register identifiers, one for the
 * least significant 32 bit word (LSW) and one for the most significant word
 * (MSW). The register IDs for the MSW is postfixed with _HI.
 *
 *
 * @field MONITOR_MIPS_REGISTER_R0        R0 - R31 are contiguous (LSW).
 * @field MONITOR_MIPS_REGISTER_SP        Stack pointer (R29) (LSW).
 * @field MONITOR_MIPS_REGISTER_FP        Frame pointer (R30) (LSW).
 * @field MONITOR_MIPS_REGISTER_F0_LO     HW Floating point, least significant,
 *                                        F0 - F31 are contiguous.
 * @field MONITOR_MIPS_REGISTER_F0_HI     HW Floating point, most significant
 *                                        F0 - F31 are contiguous.
 * @field MONITOR_MIPS_REGISTER_HI        HI register (LSW).
 * @field MONITOR_MIPS_REGISTER_LO        LO register (LSW).
 * @field MONITOR_MIPS_REGISTER_PC        Program counter (LSW).
 * @field MONITOR_MIPS_REGISTER_RA        Deprecated encoding for PC. Recommended
 *                                        encoding is MONITOR_MIPS_REGISTER_PC.
 *
 * @field MONITOR_MIPS_REGISTER_R0_HI     R0 - R31 are contiguous (MSW).
 * @field MONITOR_MIPS_REGISTER_SP_HI     Stack pointer (R29) (MSW).
 * @field MONITOR_MIPS_REGISTER_FP_HI     Frame pointer (R30) (MSW).
 * @field MONITOR_MIPS_REGISTER_RA_HI     Deprecated encoding for PC_HI.
 *                                        Recommended encoding is
 *                                        MONITOR_MIPS_REGISTER_PC_HI.
 *
 * @field MONITOR_MIPS_REGISTER_HI_HI     HI register (MSW).
 * @field MONITOR_MIPS_REGISTER_LO_HI     LO register (MSW).
 * @field MONITOR_MIPS_REGISTER_PC_HI     Program counter (MSW).
 *
 * @field MONITOR_MIPS_REGISTER_SSLIMIT   Supervisor stack location.
 * @field MONITOR_MIPS_REGISTER_SR        Status register.
 * @seealso MonitorRegister
 */
enum MonitorMipsRegisterIdValue
{
   MONITOR_MIPS_REGISTER_R0 = 0,
   MONITOR_MIPS_REGISTER_SP = 29,
   MONITOR_MIPS_REGISTER_FP = 30,
   MONITOR_MIPS_REGISTER_RA = 31,

   MONITOR_MIPS_REGISTER_HI = 64,
   MONITOR_MIPS_REGISTER_LO = 65,
   MONITOR_MIPS_REGISTER_PC = 95,

   MONITOR_MIPS_REGISTER_R0_HI = 256,
   MONITOR_MIPS_REGISTER_SP_HI = 285,
   MONITOR_MIPS_REGISTER_FP_HI = 286,
   MONITOR_MIPS_REGISTER_RA_HI = 287,

   MONITOR_MIPS_REGISTER_HI_HI = 320,
   MONITOR_MIPS_REGISTER_LO_HI = 321,
   MONITOR_MIPS_REGISTER_PC_HI = 351,

   /* Pseudo registers */
   MONITOR_MIPS_REGISTER_SSLIMIT = 1023,
   MONITOR_MIPS_REGISTER_SR =      1036,

   /* Floating point */
   MONITOR_MIPS_REGISTER_F0_LO =   1280,
   MONITOR_MIPS_REGISTER_F0_HI =   1312
};

/**
 * @type MonitorPowerpcRegisterIdValue
 *
 * @brief Register identities used for register access on PowerPC.
 *
 * @field MONITOR_POWERPC_REGISTER_R0         R0 - R31 are contiguous.
 * @field MONITOR_POWERPC_REGISTER_F0_LO      HW Floating point, least significant,
 *                                            F0 - F31 are contiguous.
 * @field MONITOR_POWERPC_REGISTER_F0_HI      HW Floating point, most significant
 *                                            F0 - F31 are contiguous.
 * @field MONITOR_POWERPC_REGISTER_PC         Program counter.
 * @field MONITOR_POWERPC_REGISTER_CR         Condition register.
 * @field MONITOR_POWERPC_REGISTER_FPSCR      Floating point status.
 * @field MONITOR_POWERPC_REGISTER_SPEFSCR    MPC E500 specific register.
 * @field MONITOR_POWERPC_REGISTER_MSR        Machine state.
 *
 * @field MONITOR_POWERPC_REGISTER_R0_HI      For 64 bit R0 - R31 this is the most
 *                                            significant word and registers are
 *                                            continuous.
 * @field MONITOR_POWERPC_REGISTER_ACC_LO     Least significant 32 bit word of ACC.
 * @field MONITOR_POWERPC_REGISTER_ACC_HI     Most significant 32 bit word of ACC.
 *
 * @field MONITOR_POWERPC_REGISTER_XER        Exception register.
 * @field MONITOR_POWERPC_REGISTER_LR         Link register.
 * @field MONITOR_POWERPC_REGISTER_CTR        Counter register.
 * @field MONITOR_POWERPC_REGISTER_RA         Deprecated encoding for PC. Recommended
 *                                            encoding is MONITOR_POWERPC_REGISTER_PC.
 *                                            (pseudo register)
 * @field MONITOR_POWERPC_REGISTER_SSLIMIT    Supervisor stack location (pseudo
 *                                            register).
 * @seealso MonitorRegister
 */
enum MonitorPowerpcRegisterIdValue
{
   MONITOR_POWERPC_REGISTER_R0 = 0,
   MONITOR_POWERPC_REGISTER_F0_LO = 32,
   MONITOR_POWERPC_REGISTER_F0_HI = 64,

   MONITOR_POWERPC_REGISTER_PC = 96,
   MONITOR_POWERPC_REGISTER_CR = 97,
   MONITOR_POWERPC_REGISTER_FPSCR = 98,
   MONITOR_POWERPC_REGISTER_MSR = 99,

   MONITOR_POWERPC_REGISTER_R0_HI = 300,
   MONITOR_POWERPC_REGISTER_SPEFSCR = 332,
   MONITOR_POWERPC_REGISTER_ACC_LO = 333,
   MONITOR_POWERPC_REGISTER_ACC_HI = 334,

   MONITOR_POWERPC_REGISTER_XER = 0x10001,
   MONITOR_POWERPC_REGISTER_LR = 0x10008,
   MONITOR_POWERPC_REGISTER_CTR = 0x10009,

   /* Pseudo registers */
   MONITOR_POWERPC_REGISTER_RA = 200,
   MONITOR_POWERPC_REGISTER_SSLIMIT = 201
};

/**
 * @type MonitorRegisterIds
 * @brief The signal struct for get-register requests / set-register replies.
 *
 * @field sigNo           MONITOR_GET_REGISTERS_REQUEST, MONITOR_SET_REGISTERS_REPLY.
 * @output status         MONITOR_STATUS_OK or cause of error.
 * @field pid             Process from which to read / (reply: to which was written)
 *                        registers.
 * @field registersCount  Number of @link "" "MonitorRegisterId" structs in
 *                        the registers field.
 * @field registers       The registers to read / (reply: that was written).
 */
typedef struct MonitorRegisterIds
{
   uint32_t sigNo;
   uint32_t pid;
   MonitorStatus status;
   uint32_t registersCount;
   MonitorRegisterId registers[1];
} MonitorRegisterIds;

/**
 * @type MonitorRegisterValues
 * @brief The signal struct with register value for set-register requests / get-register replies.
 *
 * @output sigNo          MONITOR_GET_REGISTERS_REPLY, MONITOR_SET_REGISTERS_REQUEST.
 * @output status         MONITOR_STATUS_OK or cause of error.
 * @output pid            Process to which to write / (reply: from which was read) registers.
 * @output registersCount Number of @link "" "MonitorRegister" structs in
 *                        the registers field.
 * @output registers      The registers to write / (reply: that was read).
 */
typedef struct MonitorRegisterValues
{
   uint32_t sigNo;
   uint32_t pid;
   MonitorStatus status;
   uint32_t registersCount;
   MonitorRegister registers[1];
} MonitorRegisterValues;

/**
 * @type MonitorGetRegisterInfoRequest
 * @brief The signal struct for register info requests.
 *
 * @field sigNo           MONITOR_GET_REGISTER_INFO_REQUEST
 * @field reserved0       NOT USED.
 * @field scope           Processes from which to read registers.
 * @field reserved1       NOT USED.
 * @field registersCount  Number of @link "" "MonitorRegisterId" structs
 *                        in the registers field.
 * @field registers       The registers to read.
 */
typedef struct MonitorGetRegisterInfoRequest
{
   uint32_t sigNo;
   uint32_t reserved0;
   MonitorScope scope;
   uint32_t reserved1;
   uint32_t registersCount;
   MonitorRegisterId registers[1];
} MonitorGetRegisterInfoRequest;

/**
 * @signal MonitorGetRegistersRequest
 *
 * @brief Read registers request.
 *
 * @long At least PC and stack pointer should be available, but
 * depending on the system state, some registers might not
 * be accessible.
 *
 * @long The same signal struct is used for get-register requests
 * and set-register replies.
 *
 * @member MONITOR_GET_REGISTERS_REQUEST
 * @member MonitorRegisterIds
 * @seealso MonitorGetRegistersReply
 */

/** @macro MONITOR_GET_REGISTERS_REQUEST */
#define MONITOR_GET_REGISTERS_REQUEST 39105 /* !-SIGNO(MonitorRegisterIds)-! */

/**
 * @signal MonitorGetRegisterInfoRequest
 *
 * @brief Read registers request.
 *
 * @long At least PC and stack pointer should be available, but
 * depending on the system state, some registers might not
 * be accessible.
 *
 * @long Followed by zero or more MonitorGetRegistersReply,
 * MonitorBlockingFlowControlRequest, and one MonitorGetRegisterInfoEndmark.
 *
 * @member MONITOR_GET_REGISTER_INFO_REQUEST
 * @member MonitorGetRegisterInfoRequest
 *
 * @seealso MonitorGetRegisterInfoEndmark
 * @seealso MonitorBlockingFlowControlRequest
 */

/** @macro MONITOR_GET_REGISTER_INFO_REQUEST */
#define MONITOR_GET_REGISTER_INFO_REQUEST 39121 /* !-SIGNO(MonitorGetRegisterInfoRequest)-! */

/**
 * @signal MonitorGetRegistersReply
 *
 * @brief Read registers reply.
 *
 * @long The same signal struct is used for set-register requests
 * and get-register replies.
 *
 * @member MONITOR_GET_REGISTERS_REPLY
 * @member MonitorRegisterValues
 * @seealso MonitorGetRegistersRequest
 */

/** @macro MONITOR_GET_REGISTERS_REPLY */
#define MONITOR_GET_REGISTERS_REPLY 39106 /* !-SIGNO(MonitorRegisterValues)-! */

/**
 * @type MonitorGetRegisterInfoEndmark
 *
 * @brief Endmark signal struct for register requests that result in multiple
 * reply signals.
 *
 * @output sigNo Signal number.
 * @output status MONITOR_STATUS_OK or reason of failure.
 * @seealso MonitorGetRegisterInfoRequest
 */
typedef struct MonitorGetRegisterInfoEndmark
{
   uint32_t sigNo;
   MonitorStatus status;
} MonitorGetRegisterInfoEndmark;

/**
 * @signal MonitorGetRegisterInfoEndmark
 *
 * @brief Endmark signal for register information requests that
 * result in multiple reply signals, with status (NOT USED).
 *
 * @member MONITOR_GET_REGISTER_INFO_ENDMARK
 * @member MonitorGetRegisterInfoEndmark
 */

/** @macro MONITOR_GET_REGISTER_INFO_ENDMARK */
#define MONITOR_GET_REGISTER_INFO_ENDMARK 39120 /* !-SIGNO(MonitorGetRegisterInfoEndmark)-! */

/**
 * @signal MonitorSetRegistersRequest
 *
 * @brief Write registers request.
 *
 * @long Writing registers is only possible if the process is
 * at a breakpoint in intercepted state. Otherwise an error
 * is reported MONITOR_REGISTER_INVALID.
 *
 * @long The same signal struct is used for set-register requests
 * and get-register replies.
 *
 * @member MONITOR_SET_REGISTERS_REQUEST
 * @member MonitorRegisterValues
 *
 * @seealso MonitorRegister
 * @seealso MonitorSetRegistersReply
 */

/** @macro MONITOR_SET_REGISTERS_REQUEST */
#define MONITOR_SET_REGISTERS_REQUEST 39107 /* !-SIGNO(MonitorRegisterValues)-! */

/**
 * @signal MonitorSetRegistersReply
 *
 * @brief Write registers reply.
 *
 * @long The same signal struct is used for get-register requests
 * and set-register replies.
 *
 * @member MONITOR_SET_REGISTERS_REPLY
 * @member MonitorRegisterIds
 * @seealso MonitorSetRegistersRequest
 */

/** @macro MONITOR_SET_REGISTERS_REPLY */
#define MONITOR_SET_REGISTERS_REPLY 39108 /* !-SIGNO(MonitorRegisterIds)-! */


/*******************************************************************************
 *
 * SCOPE MANAGEMENT SIGNALS
 *
 ******************************************************************************/

/**
 * @type MonitorAttachTagValue
 *
 * @brief Tags to set attach options
 *
 * @long This is used in the @link "" "MonitorAttachRequest" to select
 * options.
 *
 * @field MONITOR_ATTACH_TAG_EVENT_LOCK
 * Request exclusive access to the event action system of the monitor
 * (monitor_events.sig).
 * In the current implementation, all clients that intend to use event
 * actionpoints or basic events must set this tag. Other clients must avoid it.
 * @break
 * This tag is not required, if the monitor reports the feature code
 * MONITOR_FEATURE_MULTIPLE_EVENT_CLIENTS (@link "" "MonitorFeature").
 *
 * @field MONITOR_ATTACH_TAG_KEEP_EVENT_TRACEPOINTS
 * Request that event tracepoints set during this session remain active, even
 * after the client disconnects and / or target warmstarts.
 *
 * @field MONITOR_ATTACH_TAG_EVENT_SIGSELECT
 * Indicate that the client will use the signal lists obtained in
 * MonitorAttachReply (or {0}) for the relevant SIGSELECT masks, so that it
 * can handle all events that the monitor can send. If not set, the monitor will
 * throw away signals that the client is not expected to handle. This is a
 * fall back to a limited set of signals supported in earlier OSE monitors,
 * with signals to clients handling send, receive, create, kill, swap and
 * error events.
 *
 * @field MONITOR_ATTACH_TAG_END Always the highest possible tag value plus one
 * and should never be used.
 *
 * @seealso MonitorFeature
 * @seealso MonitorAttachRequest
 */
enum MonitorAttachTagValue
{
   MONITOR_ATTACH_TAG_EVENT_LOCK = 1010,
   MONITOR_ATTACH_TAG_KEEP_EVENT_TRACEPOINTS = 1011,
   MONITOR_ATTACH_TAG_EVENT_SIGSELECT = 1012,
   MONITOR_ATTACH_TAG_END
};

/**
 * @signal MonitorAttachRequest
 *
 * @brief Request to attach this client to a main scope
 * and to a more limited intercept scope.
 *
 * @long The client is required to attach to a scope before it can receive
 * notification about execution events, set breakpoints, or intercept or
 * resume processes.
 * An attached scope is reserved for this client and protected against
 * other clients.
 *
 * @long There are two scopes associated with an attach request.
 *
 * @long The main scope limits the set of processes the client is interested in,
 * that is what it wants to see events for, etc. To minimize the interference
 * with other clients, it is recommended that the client keeps this scope small.
 * If the monitor supports this feature, the main scope will also be used
 * by the event system in monitor_events.sig for automatic scope reduction,
 * compare @link "monitor.sig" "MonitorFeature".
 *
 * @long The intercept scope is a subscope inside the main scope, which limits
 * the set of processes that the client can intercept, either directly or
 * through breakpoints or event actions. The intercept scope must always be
 * equal to or a subset of the main scope.
 *
 * @long Before attaching, check the features returned from the monitor
 * when connecting (see @link "monitor.sig" "MonitorFeature"),
 * for details what is actually supported by the used version of
 * the OSE Run Time Monitor.
 *
 * @long If the client will accept all signals that the monitor can send,
 * the client must set the tag MONITOR_ATTACH_TAG_EVENT_SIGSELECT
 * (see @link "" "MonitorAttachTagValue").
 *
 * @long If the monitor does not support the feature with multiple clients,
 * the tag MONITOR_ATTACH_TAG_EVENT_LOCK (@link "" "MonitorAttachTagValue")
 * must be set.
 *
 * @restriction Accepts only scopes of types MONITOR_SCOPE_ID,
 * MONITOR_SCOPE_BLOCK_ID, MONITOR_SCOPE_SEGMENT_ID and MONITOR_SCOPE_SYSTEM.
 *
 * @member MONITOR_ATTACH_REQUEST
 * @member MonitorAttachRequest
 *
 * @seealso MonitorConnectRequest
 * @seealso MonitorFeature
 * @seealso MonitorAttachReply
 * @seealso MonitorDetachRequest
 * @seealso MonitorScopeTypeValue
 */

/** @macro MONITOR_ATTACH_REQUEST */
#define MONITOR_ATTACH_REQUEST 39109 /* !-SIGNO(MonitorAttachRequest)-! */

/**
 * @type MonitorAttachRequest
 *
 * @input sigNo MONITOR_ATTACH_REQUEST
 * @input scope The main scope to limit reported events to.
 * @input intercept The intercept scope to limit intercept operations to.
 * @input tagsCount The length of the tags list, as the
 *                  total number of sizeof(MonitorTag)
 *                  entries, with tags and arguments.
 * @input tags Tags (@link "" "MonitorAttachTagValue").
 */
typedef struct MonitorAttachRequest
{
   uint32_t sigNo;
   MonitorScope scope;
   MonitorScope intercept;
   uint32_t tagsCount;
   MonitorTag tags[1];
} MonitorAttachRequest;

/**
 * @signal MonitorAttachReply
 *
 * @brief Reply when attaching the client to a main and intercept scope.
 *
 * @long This reply signal contains several parameters from the monitor.
 * If the client will accept to receive all supported signals from the monitor,
 * the client should already in the MonitorAttachRequest request have set
 * the tag MONITOR_ATTACH_TAG_EVENT_SIGSELECT
 * (see @link "" "MonitorAttachTagValue").
 *
 * @member MONITOR_ATTACH_REPLY
 * @member MonitorAttachReply
 *
 * @seealso MonitorAttachRequest
 * @seealso MonitorDetachRequest
 */

/** @macro MONITOR_ATTACH_REPLY */
#define MONITOR_ATTACH_REPLY 39110 /* !-SIGNO(MonitorAttachReply)-! */

/**
 * @type MonitorAttachReply
 *
 * @output sigNo              MONITOR_ATTACH_REPLY
 * @output status             MONITOR_STATUS_OK or cause of error.
 * @output processCacheCount  The monitor maintains a circular buffer of
 *                            process IDs for which it has sent information in
 *                            event trace, intercept, and notification reports.
 *                            If a process is found in the cache information on
 *                            it is not repeated, since it is assumed that the
 *                            client maintains a similar list of at least this
 *                            size.
 * @output reserved0          NOT USED.
 *                            NOTE: Not present in older monitor implementations
 * @output reserved1          NOT USED.
 *                            NOTE: Not present in older monitor implementations
 * @output reserved2          NOT USED.
 *                            NOTE: Not present in older monitor implementations
 * @output notifyOffset       Offset in sigs to a list of notification signals, that
 *                            the monitor can send.
 *                            NOTE: Not present in older monitor implementations
 * @output notifyCount        Number of notification signals.
 *                            NOTE: Not present in older monitor implementations
 * @output interceptOffset    Offset in sigs to a list of intercept notification
 *                            signals, that the monitor can send.
 *                            NOTE: Not present in older monitor implementations
 * @output interceptCount     Number of intercept notification signals.
 *                            NOTE: Not present in older monitor implementations
 * @output traceOffset        Offset in sigs to a list of trace signals, that
 *                            the monitor can send.
 *                            NOTE: Not present in older monitor implementations
 * @output traceCount         Number of trace signals.
 *                            NOTE: Not present in older monitor implementations
 * @output sigs               Signal numbers that the client should be prepared to
 *                            receive from the monitor.
 */
typedef struct MonitorAttachReply
{
   uint32_t sigNo;
   MonitorStatus status;
   uint32_t processCacheCount;
   uint32_t reserved0;
   uint32_t reserved1;
   uint32_t reserved2;
   uint32_t notifyOffset;
   uint32_t notifyCount;
   uint32_t interceptOffset;
   uint32_t interceptCount;
   uint32_t traceOffset;
   uint32_t traceCount;
   uint32_t sigs[1];
} MonitorAttachReply;

/**
 * @signal MonitorDetachRequest
 *
 * @brief Request to detach from the scope.
 *
 * @long It is recommended to detach before disconnecting.
 *
 * @member MONITOR_DETACH_REQUEST
 * @member MonitorDetachRequest
 * @seealso MonitorAttachRequest
 *
 * @seealso MonitorDetachReply
 */

/** @macro MONITOR_DETACH_REQUEST */
#define MONITOR_DETACH_REQUEST 39111 /* !-SIGNO(MonitorDetachRequest)-! */

/**
 * @type MonitorDetachRequest
 *
 * @input sigNo MONITOR_DETACH_REQUEST
 * @input resume True if the attached scope should be resumed.
 *               (NOT USED. Set to 1)
 *
 */
typedef struct MonitorDetachRequest
{
   uint32_t sigNo;
   MonitorBoolean resume;
} MonitorDetachRequest;

/**
 * @signal MonitorDetachReply
 *
 * @brief Confirmation that the scope is now detached.
 *
 * @member MONITOR_DETACH_REPLY
 * @member MonitorDetachReply
 *
 * @seealso MonitorDetachRequest
 */

/** @macro MONITOR_DETACH_REPLY */
#define MONITOR_DETACH_REPLY 39112 /* !-SIGNO(MonitorDetachReply)-! */

/**
 * @type MonitorDetachReply
 *
 * @output sigNo              MONITOR_DETACH_REPLY
 * @output status             MONITOR_STATUS_OK or cause of error.
 */
typedef struct MonitorDetachReply
{
   uint32_t sigNo;
   MonitorStatus status;
} MonitorDetachReply;


/*******************************************************************************
 *
 * BASIC EVENT SIGNALS
 *
 ******************************************************************************/

/**
 * @type MonitorBasicEventAction
 *
 * @brief Event action type
 *
 * @long This is the action to take at an event and is
 * one of the values in enum @link "" "MonitorBasicEventActionValue".
 *
 */
typedef uint32_t MonitorBasicEventAction;

/**
 * @type MonitorBasicEventActionValue
 *
 * @brief Enum of basic event action values.
 *
 * @long The value defines what action the monitor will perform at a
 * requested basic event.
 *
 * @field MONITOR_BASIC_EVENT_ACTION_NOTIFY    Notify the client about the event
 *                                             from the attached main scope.
 *                                             Event notifications are unreliable
 *                                             in the sense that notification can
 *                                             be dropped because of flow control.
 * @field MONITOR_BASIC_EVENT_ACTION_INTERCEPT Intercept the attached intercept
 *                                             scope and notify the event to the
 *                                             client. Event intercept notifications
 *                                             are reliable.
 */
enum MonitorBasicEventActionValue
{
   MONITOR_BASIC_EVENT_ACTION_INTERCEPT = 1,
   MONITOR_BASIC_EVENT_ACTION_NOTIFY    = 10
};

/**
 * @type MonitorBasicEventActionRequest
 *
 * @input sigNo  Signal number of this request.
 * @input action The action to take when the event occurs,
 *               e.g. a notification or an intercept action.
 *               See @link "" "MonitorBasicEventActionValue".
 */
typedef struct MonitorBasicEventActionRequest
{
   uint32_t sigNo;
   MonitorBasicEventAction action;
} MonitorBasicEventActionRequest;

/**
 * @type MonitorBasicEventActionReply
 *
 * @output sigNo  Signal number of this reply.
 * @output status MONITOR_STATUS_OK or cause of error.
 */
typedef struct MonitorBasicEventActionReply
{
   uint32_t sigNo;
   MonitorStatus status;
} MonitorBasicEventActionReply;

/**
 * @signal MonitorBasicErrorActionRequest
 *
 * @brief Request an action to be taken each time an error is reported
 * within the attached main scope.
 *
 * @member MONITOR_BASIC_ERROR_ACTION_REQUEST
 * @member MonitorBasicEventActionRequest
 *
 * @seealso MonitorBasicErrorActionReply
 */

/** @macro MONITOR_BASIC_ERROR_ACTION_REQUEST */
#define MONITOR_BASIC_ERROR_ACTION_REQUEST 39113 /* !-SIGNO(MonitorBasicEventActionRequest)-! */

/**
 * @signal MonitorBasicErrorActionReply
 *
 * @brief Reply from setting basic error event action request.
 *
 * @member MONITOR_BASIC_ERROR_ACTION_REPLY
 * @member MonitorBasicEventActionReply
 *
 * @seealso MonitorBasicErrorActionRequest
 */

/** @macro MONITOR_BASIC_ERROR_ACTION_REPLY */
#define MONITOR_BASIC_ERROR_ACTION_REPLY 39114 /* !-SIGNO(MonitorBasicEventActionReply)-! */

/**
 * @signal MonitorBasicCreateKillActionRequest
 *
 * @brief Request an action to be taken each time a process is
 * created or killed within the attached main scope.
 *
 * @member MONITOR_BASIC_CREATE_KILL_ACTION_REQUEST
 * @member MonitorBasicEventActionRequest
 *
 * @seealso MonitorBasicCreateKillActionReply
 */

/** @macro MONITOR_BASIC_CREATE_KILL_ACTION_REQUEST */
#define MONITOR_BASIC_CREATE_KILL_ACTION_REQUEST 39115 /* !-SIGNO(MonitorBasicEventActionRequest)-! */

/**
 * @signal MonitorBasicCreateKillActionReply
 *
 * @brief Reply from setting basic create or kill action request.
 *
 * @member MONITOR_BASIC_CREATE_KILL_ACTION_REPLY
 * @member MonitorBasicEventActionReply
 * @seealso MonitorBasicCreateKillActionRequest
 */

/** @macro MONITOR_BASIC_CREATE_KILL_ACTION_REPLY */
#define MONITOR_BASIC_CREATE_KILL_ACTION_REPLY 39116 /* !-SIGNO(MonitorBasicEventActionReply)-! */

/**
 * @signal MonitorBasicErrorNotify
 *
 * @brief Notification report to the client that an error event
 * has occurred within the attached main scope.
 *
 * @long This signal is sent under non-blocking, unreliable,
 * flow control (@link "" "MonitorNonBlockingFlowControlRequest").
 *
 * @member MONITOR_BASIC_ERROR_NOTIFY
 * @member MonitorBasicErrorNotify
 *
 * @seealso MonitorBasicErrorActionRequest
 */

/** @macro MONITOR_BASIC_ERROR_NOTIFY */
#define MONITOR_BASIC_ERROR_NOTIFY 39117 /* !-SIGNO(MonitorBasicErrorNotify)-! */

/**
 * @type MonitorBasicErrorNotify
 *
 * @output sigNo MONITOR_BASIC_ERROR_NOTIFY
 * @output pid   ID of the process.
 * @output exec  True if called from the executive (user_called=0).
 * @output error The error number.
 * @output extra The extra error information.
 * @field reserved0       NOT USED.
 */
typedef struct MonitorBasicErrorNotify
{
   uint32_t sigNo;
   uint32_t pid;
   MonitorBoolean exec;
   uint32_t error;
   uint32_t extra;
   uint32_t reserved0;
} MonitorBasicErrorNotify;

/**
 * @signal MonitorBasicCreateKillNotify
 *
 * @brief Notification report to the client that a new process
 * has been created or an existing process killed
 * within the attached main scope.
 *
 * @long This signal is sent under blocking flow control
 *       (@link "" "MonitorBlockingFlowControlRequest").
 *
 * @member MONITOR_BASIC_CREATE_KILL_NOTIFY
 * @member MonitorBasicCreateKillNotify
 *
 * @seealso MonitorBasicCreateKillActionRequest
 */

/** @macro MONITOR_BASIC_CREATE_KILL_NOTIFY */
#define MONITOR_BASIC_CREATE_KILL_NOTIFY 39118 /* !-SIGNO(MonitorBasicCreateKillNotify)-! */

/**
 * @type MonitorBasicCreateKillNotify
 *
 * @output sigNo      MONITOR_BASIC_CREATE_KILL_NOTIFY
 * @output pid        ID of the process.
 * @output bid        ID of the parent block or 0 if not available.
 * @output sid        ID of the parent segment or 0 if not available.
 * @output type       Type of the process, @link "" "MonitorProcessTypeValue".
 * @output state      State of the process @link "" "MonitorProcessStateValue".
 * @output properties Bit field indicating additional properties of the process.
 *                    Bit 0 is set to 1 for system processes.
 *                    Other bits are undefined.
 * @output entrypoint Entrypoint of the process or 0.
 * @output name       Null terminated name of the process.
 */
typedef struct MonitorBasicCreateKillNotify
{
   uint32_t sigNo;
   uint32_t pid;
   uint32_t bid;
   uint32_t sid;
   MonitorProcessType type;
   MonitorProcessState state;
   uint32_t properties;
   uint32_t entrypoint;
   char name[1];
} MonitorBasicCreateKillNotify;

/**
 * @signal MonitorBasicLossNotify
 *
 * @brief Inform the client that reports of events for which the client has
 * requested notifications might have been lost due to flow control.
 *
 * @long This signal is sent under blocking flow control
 *       (@link "" "MonitorBlockingFlowControlRequest").
 *
 * @member MONITOR_BASIC_LOSS_NOTIFY
 * @member MonitorBasicLossNotify
 */

/** @macro MONITOR_BASIC_LOSS_NOTIFY */
#define MONITOR_BASIC_LOSS_NOTIFY 39119 /* !-SIGNO(MonitorBasicLossNotify)-! */

/**
 * @type MonitorBasicLossNotify
 *
 * @output sigNo      MONITOR_BASIC_LOSS_NOTIFY
 */
typedef struct MonitorBasicLossNotify
{
   uint32_t sigNo;
} MonitorBasicLossNotify;


/*******************************************************************************
 *
 * BREAKPOINT AND STEP SIGNALS
 *
 ******************************************************************************/

/**
 * @type MonitorBreakpoint
 *
 * @brief A reference to a previously set breakpoint.
 *
 */
typedef uint32_t MonitorBreakpoint;

/**
 * @type MonitorBreakpointTagValue
 *
 * @brief Tags to specify a breakpoint type.
 *
 * @field MONITOR_BREAKPOINT_TAG_IGNORE Number of times a breakpoint is allowed
 *                                      to fire before the intercept scope is
 *                                      intercepted
 *                                      and the client is notified.
 * @field MONITOR_BREAKPOINT_TAG_REMOVE Number of times a breakpoint is allowed
 *                                      to fire before it is automatically
 *                                      removed (NOT USED).
 * @field MONITOR_BREAKPOINT_TAG_END    Highest possible tag value plus one,
 *                                      should not be used.
 *
 * @seealso MonitorSetBreakpointRequest
 */
enum MonitorBreakpointTagValue
{
   MONITOR_BREAKPOINT_TAG_IGNORE = 1000,
   MONITOR_BREAKPOINT_TAG_REMOVE,
   MONITOR_BREAKPOINT_TAG_END
};

/**
 * @signal MonitorSetBreakpointRequest
 *
 * @brief Set breakpoint request.
 *
 * @long When a breakpoint is set, the monitor creates and returns an
 * ID for the breakpoint. This ID is reported when the breakpoint is fired.
 *
 * @long Breakpoint tags are used, see @link "" "MonitorBreakpointTagValue".
 *
 * @long Both the given scopes are automatically reduced to the
 * attached intercept subscope.
 *
 *
 * @restriction Accepts only scopes of types MONITOR_SCOPE_VOID,
 * MONITOR_SCOPE_ID, MONITOR_SCOPE_BLOCK_ID, MONITOR_SCOPE_SEGMENT_ID
 * and MONITOR_SCOPE_SYSTEM (@link "" "MonitorScopeTypeValue").
 *
 * @member MONITOR_SET_BREAKPOINT_REQUEST
 * @member MonitorSetBreakpointRequest
 *
 * @seealso MonitorSetBreakpointReply
 */

/** @macro MONITOR_SET_BREAKPOINT_REQUEST */
#define MONITOR_SET_BREAKPOINT_REQUEST 39122 /* !-SIGNO(MonitorSetBreakpointRequest)-! */

/**
 * @type MonitorSetBreakpointRequest
 *
 * @input sigNo       MONITOR_SET_BREAKPOINT_REQUEST
 * @input address     Address of the breakpoint.
 * @input pid         The address is in the context of this process or block.
 * @input scope       The set of processes that can fire the breakpoint.
 * @input intercept   The set of processes to intercept when the breakpoint
 *                    is fired. Must either be MONITOR_SCOPE_VOID (to
 *                    indicate only the process that hit the breakpoint), or
 *                    equal to or a superset of the set of processes
 *                    that can fire the breakpoint.
 * @input tagsCount   The length of the tags list, as the
 *                    total number of sizeof(MonitorTag)
 *                    entries, with tags and arguments.
 * @input tags        List of tags and their values.
 */
typedef struct MonitorSetBreakpointRequest
{
   uint32_t sigNo;
   uint32_t pid;
   uint32_t address;
   MonitorScope scope;
   MonitorScope intercept;
   uint32_t tagsCount;
   MonitorTag tags[1];
} MonitorSetBreakpointRequest;

/**
 * @signal MonitorSetBreakpointReply
 *
 * @brief Set breakpoint reply.
 *
 * @member MONITOR_SET_BREAKPOINT_REPLY
 * @member MonitorSetBreakpointReply
 *
 * @seealso MonitorSetBreakpointRequest
 */

/** @macro MONITOR_SET_BREAKPOINT_REPLY */
#define MONITOR_SET_BREAKPOINT_REPLY 39123 /* !-SIGNO(MonitorSetBreakpointReply)-! */

/**
 * @type MonitorSetBreakpointReply
 *
 * @output sigNo  MONITOR_SET_BREAKPOINT_REPLY
 * @output status MONITOR_STATUS_OK or cause of error.
 * @output id     ID of the breakpoint.
 */
typedef struct MonitorSetBreakpointReply
{
   uint32_t sigNo;
   MonitorStatus status;
   MonitorBreakpoint id;
} MonitorSetBreakpointReply;

/**
 * @signal MonitorBreakpointFiredNotify
 *
 * @brief Sent from monitor to responsible client when a breakpoint is fired
 * within the requested scope.
 *
 * @long The breakpoint is identified with the breakpoint ID.
 *
 * @member MONITOR_BREAKPOINT_FIRED_NOTIFY
 * @member MonitorBreakpointFiredNotify
 *
 * @seealso MonitorSetBreakpointRequest
 */

/** @macro MONITOR_BREAKPOINT_FIRED_NOTIFY */
#define MONITOR_BREAKPOINT_FIRED_NOTIFY 39124 /* !-SIGNO(MonitorBreakpointFiredNotify)-! */

/**
 * @type MonitorBreakpointFiredNotify
 *
 * @output sigNo              MONITOR_BREAKPOINT_FIRED_NOTIFY
 * @output pid                Process that hit the breakpoint.
 * @output id                 ID of the breakpoint.
 */
typedef struct MonitorBreakpointFiredNotify
{
   uint32_t sigNo;
   uint32_t pid;
   MonitorBreakpoint id;
} MonitorBreakpointFiredNotify;

/**
 * @signal MonitorClearBreakpointRequest
 *
 * @brief Remove breakpoint request.
 *
 * @member MONITOR_CLEAR_BREAKPOINT_REQUEST
 * @member MonitorClearBreakpointRequest
 *
 * @seealso MonitorClearBreakpointReply
 * @seealso MonitorSetBreakpointRequest
 */

/** @macro MONITOR_CLEAR_BREAKPOINT_REQUEST */
#define MONITOR_CLEAR_BREAKPOINT_REQUEST 39125 /* !-SIGNO(MonitorClearBreakpointRequest)-! */

/**
 * @type MonitorClearBreakpointRequest
 *
 * @input sigNo MONITOR_CLEAR_BREAKPOINT_REQUEST
 * @input id    ID of the breakpoint.
 */
typedef struct MonitorClearBreakpointRequest
{
   uint32_t sigNo;
   MonitorBreakpoint id;
} MonitorClearBreakpointRequest;

/**
 * @signal MonitorClearBreakpointReply
 *
 * @brief Remove breakpoint reply.
 *
 * @member MONITOR_CLEAR_BREAKPOINT_REPLY
 * @member MonitorClearBreakpointReply
 *
 * @seealso MonitorClearBreakpointRequest
 */

/** @macro MONITOR_CLEAR_BREAKPOINT_REPLY */
#define MONITOR_CLEAR_BREAKPOINT_REPLY 39126 /* !-SIGNO(MonitorClearBreakpointReply)-! */

/**
 * @type MonitorClearBreakpointReply
 *
 * @output sigNo  MONITOR_CLEAR_BREAKPOINT_REPLY
 * @output status MONITOR_STATUS_OK or cause of error.
 */
typedef struct MonitorClearBreakpointReply
{
   uint32_t sigNo;
   MonitorStatus status;
} MonitorClearBreakpointReply;


/**
 * @type MonitorStepMode
 *
 * @brief Mode of step operation.
 *
 * @seealso MonitorStepModeValue
 */
typedef uint32_t MonitorStepMode;

/**
 * @type MonitorStepModeValue
 *
 * @long Values to indicate the step mode in
 * @link "" "MonitorStepProcessRequest".
 *
 * @field MONITOR_STEP_INTO         Intercept the scope as soon as the process
 *                                  is about to execute an instruction outside
 *                                  the specified range.
 * @field MONITOR_STEP_OVER         Allow execution to continue into (and out
 *                                  of) functions called from within the
 *                                  specified range.
 * @field MONITOR_STEP_INSTRUCTION  Step the process a single machine
 *                                  instruction, disregarding range.
 */
enum MonitorStepModeValue
{
   MONITOR_STEP_INTO        = 0x0,
   MONITOR_STEP_OVER        = 0x1,
   MONITOR_STEP_INSTRUCTION = 0x2
};

/**
 * @signal MonitorStepProcessRequest
 *
 * @brief Step a process over an address range request.
 *
 * @long When the step has been performed (or an error occurred),
 * a @link "" "MonitorStepProcessNotify" notification is sent to the client.
 *
 * @long Each process may be the subject of at most one step operation at
 * a time. If a process hits a breakpoint or is intercepted in the process
 * of completing a step and a new step operation is requested, the original
 * step operation will be canceled.
 *
 * @restriction Accepts only scopes of types
 * MONITOR_SCOPE_ID, MONITOR_SCOPE_BLOCK_ID, MONITOR_SCOPE_SEGMENT_ID
 * and MONITOR_SCOPE_SYSTEM.
 *
 * @member MONITOR_STEP_PROCESS_REQUEST
 * @member MonitorStepProcessRequest
 *
 * @seealso MonitorStepProcessNotify
 */

/** @macro MONITOR_STEP_PROCESS_REQUEST */
#define MONITOR_STEP_PROCESS_REQUEST 39127 /* !-SIGNO(MonitorStepProcessRequest)-! */

/**
 * @type MonitorStepProcessRequest
 *
 * @input sigNo              MONITOR_STEP_PROCESS_REQUEST
 * @input pid                ID of process to step.
 * @input mode               Flags MONITOR_STEP_OVER or MONITOR_STEP_INSTRUCTION
 *                           or MONITOR_STEP_INTO
 *                           (@link "" "MonitorStepModeValue")
 * @input start              Start address.
 * @input end                End address.
 * @input scope              Which scope of processes to resume and then
 *                           intercept when the step completes.
 *                           Must include the stepped process.
 * @input tagsCount          NOT USED.
 * @input tags               NOT USED.
 *
 */
typedef struct MonitorStepProcessRequest
{
   uint32_t sigNo;
   uint32_t pid;
   MonitorStepMode mode;
   uint32_t start;
   uint32_t end;
   MonitorScope scope;
   uint32_t tagsCount;
   MonitorTag tags[1];
} MonitorStepProcessRequest;

/**
 * @signal MonitorStepProcessNotify
 *
 * @brief Notification that a step process completed or error occurred.
 *
 * @long This notification is only sent at a resulting intercept
 * from an actionpoint. It is not sent after a direct intercept
 * request (MonitorInterceptScopeRequest).
 *
 * @member MONITOR_STEP_PROCESS_NOTIFY
 * @member MonitorStepProcessNotify
 *
 * @seealso MonitorStepProcessRequest
 */

/** @macro MONITOR_STEP_PROCESS_NOTIFY */
#define MONITOR_STEP_PROCESS_NOTIFY 39129 /* !-SIGNO(MonitorStepProcessNotify)-! */

/**
 * @type MonitorStepProcessNotify
 *
 * @output sigNo              MONITOR_STEP_PROCESS_NOTIFY
 * @output pid                ID of process that was stepped.
 * @output status             MONITOR_STATUS_OK or cause of error.
 * @output pc                 Current program counter of the process.
 */
typedef struct MonitorStepProcessNotify
{
   uint32_t sigNo;
   uint32_t pid;
   uint32_t status;
   uint32_t pc;
                              /* REVIEW: Add sp and fp here as well if it
                               * saves the client a register read operation? */
} MonitorStepProcessNotify;


/*******************************************************************************
 *
 * SCOPE STATUS SIGNALS
 *
 ******************************************************************************/

/**
 * @signal MonitorInterceptScopeRequest
 *
 * @brief Intercept a set of processes request.
 *
 * @long The scope given here is automatically reduced to the attached
 * intercept scope, it the given scope is wider.
 *
 * @restriction If the client has attached with the
 * MONITOR_ATTACH_TAG_EVENT_LOCK option, the same scope as in the
 * MonitorAttachRequest or MonitorSetInterceptScopeRequest must be used.
 * If the client has attached without the MONITOR_ATTACH_TAG_EVENT_LOCK option,
 * only the scopes MONITOR_SCOPE_ID, MONITOR_SCOPE_BLOCK_ID,
 * MONITOR_SCOPE_SEGMENT_ID and MONITOR_SCOPE_SYSTEM are supported.
 *
 * @member MONITOR_INTERCEPT_SCOPE_REQUEST
 * @member MonitorInterceptScopeRequest
 *
 * @seealso MonitorInterceptScopeReply
 */

/** @macro MONITOR_INTERCEPT_SCOPE_REQUEST */
#define MONITOR_INTERCEPT_SCOPE_REQUEST 39130 /* !-SIGNO(MonitorInterceptScopeRequest)-! */

/**
 * @type MonitorInterceptScopeRequest
 *
 * @input sigNo              MONITOR_INTERCEPT_SCOPE_REQUEST
 * @input scope              Scope of processes to intercept (auto reduced
*                            to within attached intercept scope).
 * @input tagsCount          The length of the tags list, as the
 *                           total number of sizeof(MonitorTag)
 *                           entries, with tags and arguments.
 * @input tags               Tags (only used if the scope needs it).
 */
typedef struct MonitorInterceptScopeRequest
{
   uint32_t sigNo;
   MonitorScope scope;
   uint32_t tagsCount;
   MonitorTag tags[1];
} MonitorInterceptScopeRequest;

/**
 * @signal MonitorInterceptScopeReply
 *
 * @brief Intercept a set of processes reply.
 *
 * @long All processes within the scope have been intercepted.
 *
 * @long The intercept request will result in this reply,
 * but no intercept notification.
 *
 * @member MONITOR_INTERCEPT_SCOPE_REPLY
 * @member MonitorInterceptScopeReply
 *
 * @seealso MonitorInterceptScopeRequest
 */

/** @macro MONITOR_INTERCEPT_SCOPE_REPLY */
#define MONITOR_INTERCEPT_SCOPE_REPLY 39131 /* !-SIGNO(MonitorInterceptScopeReply)-! */

/**
 * @type MonitorInterceptScopeReply
 *
 * @output sigNo  MONITOR_INTERCEPT_SCOPE_REPLY
 * @output status MONITOR_STATUS_OK or cause of error.
 */
typedef struct MonitorInterceptScopeReply
{
   uint32_t sigNo;
   MonitorStatus status;
} MonitorInterceptScopeReply;

/**
 * @signal MonitorResumeScopeRequest
 *
 * @brief Resume a set of processes request.
 *
 * @long All processes within the scope are resumed (from being
 * earlier intercepted).
 *
 * @restriction If the client has attached with the
 * MONITOR_ATTACH_TAG_EVENT_LOCK option, the same scope as in the
 * MonitorAttachRequest or MonitorSetInterceptScopeRequest must be used.
 * If the client has attached without the MONITOR_ATTACH_TAG_EVENT_LOCK option,
 * only the scopes MONITOR_SCOPE_ID, MONITOR_SCOPE_BLOCK_ID, MONITOR_SCOPE_SEGMENT_ID and
 * MONITOR_SCOPE_SYSTEM are supported.
 *
 * @member MONITOR_RESUME_SCOPE_REQUEST
 * @member MonitorResumeScopeRequest
 * @seealso MonitorResumeScopeReply
 */

/** @macro MONITOR_RESUME_SCOPE_REQUEST */
#define MONITOR_RESUME_SCOPE_REQUEST 39132 /* !-SIGNO(MonitorResumeScopeRequest)-! */

/**
 * @type MonitorResumeScopeRequest
 *
 * @input sigNo     MONITOR_RESUME_SCOPE_REQUEST
 * @input scope     Scope to resume (auto reduced
*                   to within attached intercept scope).
 * @input tagsCount The length of the tags list, as the
 *                  total number of sizeof(MonitorTag)
 *                  entries, with tags and arguments.
 * @input tags Tags
 */
typedef struct MonitorResumeScopeRequest
{
   uint32_t sigNo;
   MonitorScope scope;
   uint32_t tagsCount;
   MonitorTag tags[1];
} MonitorResumeScopeRequest;

/**
 * @signal MonitorResumeScopeReply
 *
 * @brief Resume a set of processes reply.
 *
 * @member MONITOR_RESUME_SCOPE_REPLY
 * @member MonitorResumeScopeReply
 */

/** @macro MONITOR_RESUME_SCOPE_REPLY */
#define MONITOR_RESUME_SCOPE_REPLY 39133 /* !-SIGNO(MonitorResumeScopeReply)-! */

/**
 * @type MonitorResumeScopeReply
 *
 * @output sigNo  MONITOR_RESUME_SCOPE_REPLY
 * @output status MONITOR_STATUS_OK or cause of error.
 * @seealso MonitorResumeScopeRequest
 */
typedef struct MonitorResumeScopeReply
{
   uint32_t sigNo;
   MonitorStatus status;
} MonitorResumeScopeReply;

#ifdef __cplusplus
} /* end of extern "C" */
#endif

#endif
