/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2010 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

/**
 * @toc OSE:
 * @file ose_spi/monitor_tracepoint.sig
 *
 * @brief OSE System source code debugging and tracepoint signal interface.
 *
 * @long This header file defines the signal interface to be used by OSE debug
 * clients (typically source code debuggers such as Optima, GDB) that use
 * tracepoints.
 *
 * @long A connection to the RMM is needed, see monitor.sig for the actual
 * connection procedure using @link "" "MonitorInterfaceRequest" and
 * @link "" "MonitorConnectRequest".
 *
 * @long The client must also attach a scope by sending a
 * @link "" "MonitorAttachRequest" signal to the monitor,
 * before using this tracepoint control interface. A main scope and
 * an intercept scope is defined.
 *
 * @long The scope tells the monitor what parts of the system the client
 * is interested in. In addition any scope given in the event requests will
 * automatically be reduced to be within the attached main scope,
 * if this feature is supported by the monitor.
 * Each client must attach for a non-overlapping scope.
 *
 * @long Some request signals may yield multiple reply signals. In this case
 * the monitor finishes the replies by sending a specific endmark signal.
 * The endmark signal also contains the overall status of the request.
 * The monitor will send some replies in multiple reply signals
 * to the client using flow control. With flow control, the monitor sends
 * flow control requests, which the client must respond to.
 *
 * @long Tracepoints enable the possibility to trace events and data flow
 * in software without rebuilding the code. Tracepoints are
 * sometimes compared with breakpoints, on each point it is possible to look
 * at variables, memory, and registers. But with tracepoints this is
 * done with minimal impact on the running system. Capturing the
 * trace is all done in the target and the result is analyzed offline. For
 * more information about tracepoints, see
 * http://sourceware.org/gdb/talks/esc-west-1999/.
 *
 */


#ifndef _MONITOR_TRACEPOINT_SIG
#define _MONITOR_TRACEPOINT_SIG

#include "stdint.h"
#include "monitor.sig"
#include "monitor_execution.sig"

#ifdef __cplusplus
extern "C"
{
#endif

/**
 * @type MonitorTracepoint
 * @brief Tracepoint type. Holds the tracepoint ID.
 */
typedef uint32_t MonitorTracepoint;

/**
 * @type MonitorTracepointAction
 * @brief Tracepoint action type. Holds the action ID.
 */
typedef uint32_t MonitorTracepointAction;

/**
 * @type MonitorTracepointVariable
 * @brief Tracepoint state variable type. Holds the variable ID.
 */
typedef uint32_t MonitorTracepointVariable;

/**
 * @type MonitorTracepointBufferTypeValue
 *
 * @brief Constants for different tracepoint buffer types.
 *
 * @field MONITOR_TRACEPOINT_BUFFER_NONE Unconfigured buffer.
 *
 * @field MONITOR_TRACEPOINT_BUFFER_LINEAR The tracepoint buffer will run until full
 * and the trace experiment will then stop.
 *
 * @field MONITOR_TRACEPOINT_BUFFER_CIRCULAR The tracepoint buffer will wrap and
 * overwrite the oldest entries rather than stop the experiment.
 *
 * @seealso MonitorTracepointBufferType
 */
typedef enum MonitorTracepointBufferTypeValue
{
   MONITOR_TRACEPOINT_BUFFER_NONE,
   MONITOR_TRACEPOINT_BUFFER_LINEAR,
   MONITOR_TRACEPOINT_BUFFER_CIRCULAR
} MonitorTracepointBufferTypeValue;

/**
 * @type MonitorTracepointBufferType
 * @brief Contains a constant from MonitorTracepointBufferTypeValue
 * @seealso MonitorTracepointBufferTypeValue
 */
typedef uint32_t MonitorTracepointBufferType;


/*-------------------------------------------------------------------------*/

/**
 * @type MonitorTracepointTypes
 *
 * @brief Specify the tracepoint type.
 *
 * @field    MONITOR_TRACEPOINT_TRAP_BASED_TRACEPOINT  Trap based tracepoint
 * @field    MONITOR_TRACEPOINT_FAST_TRACEPOINT        Fast tracepoint. Not
 *                                                     supported.
 * @field    MONITOR_TRACEPOINT_STATIC_TRACEPOINT      Static tracepoint.
 *                                                     Not supported.
 * @seealso MonitorTracepointCreateRequest
 */
enum MonitorTracepointTypes
{
   MONITOR_TRACEPOINT_TRAP_BASED_TRACEPOINT,
   MONITOR_TRACEPOINT_FAST_TRACEPOINT,
   MONITOR_TRACEPOINT_STATIC_TRACEPOINT
};

/*-------------------------------------------------------------------------*/

/**
 * @type MonitorTraceError
 *
 * @brief Specify the trace error.
 *
 * @field    TRACEPOINT_OK                                    Trace status OK.
 * @field    TRACEPOINT_ERROR_DIV_BY_ZERO                     Division by zero in OP code.
 * @field    TRACEPOINT_ERROR_INSTRUCTION_NOT_IMPLEMENTED     OP code has not been implemented.
 * @field    TRACEPOINT_ERROR_UNKNOWN_OP_CODE                 Unknown OP code.
 * @field    TRACEPOINT_ERROR_PROGRAM_TOO_LARGE               OP code program too large.
 * @field    TRACEPOINT_ERROR_ILLEGAL_GOTO_ADDRESS            Illegal goto address in OP code.
 * @field    TRACEPOINT_ERROR_MEMORY_READ                     Memory read error for OP code.
 * @field    TRACEPOINT_ERROR_REGISTER                        Register error for OP code.
 * @field    TRACEPOINT_ERROR_BAD_TVARIABLE_ADDRESS           Variable id error in OP code.
 * @field    TRACEPOINT_ERROR_STACK_OVERFLOW                  OP code stack overflow.
 * @field    TRACEPOINT_ERROR_STACK_UNDERFLOW                 OP code stack underflow.
 * @field    TRACEPOINT_ERROR_TRACE_TOO_LARGE                 Trace is too large for max signal size.
 * @field    TRACEPOINT_ERROR_MEMORY_READ                     Memory read error
 * @field    TRACEPOINT_ERROR_REGISTER                        Register error
 * @field    TRACEPOINT_ERROR_EPM_ERROR                       EPM error in while stepping
 * @field    TRACEPOINT_ERROR_STACK64_NEEDED                  A stack 64 is needed (fast tracepoint error)
 * @field    TRACEPOINT_ERROR_MISSING_END_INSTRUCTION         End instruction is missing
 * @field    TRACEPOINT_ERROR_BYTECODE_MISSING                No Bytecode present (fast tracepoint error)
 * @field    TRACEPOINT_ERROR_INTERNAL                        Internal Error 
 *
 * @seealso MonitorTracepointGetTracebufferReply
 */

enum MonitorTraceError
{
   TRACEPOINT_OK,
   TRACEPOINT_ERROR_DIV_BY_ZERO,
   TRACEPOINT_ERROR_INSTRUCTION_NOT_IMPLEMENTED,
   TRACEPOINT_ERROR_UNKNOWN_OP_CODE,
   TRACEPOINT_ERROR_PROGRAM_TOO_LARGE,
   TRACEPOINT_ERROR_ILLEGAL_GOTO_ADDRESS,
   TRACEPOINT_ERROR_MEMORY_READ,
   TRACEPOINT_ERROR_REGISTER,
   TRACEPOINT_ERROR_BAD_TVARIABLE_ADDRESS,
   TRACEPOINT_ERROR_STACK_OVERFLOW,
   TRACEPOINT_ERROR_STACK_UNDERFLOW,
   TRACEPOINT_ERROR_TRACE_TOO_LARGE,
   TRACEPOINT_ERROR_EPM_ERROR,
   TRACEPOINT_ERROR_STACK64_NEEDED,
   TRACEPOINT_ERROR_MISSING_END_INSTRUCTION,
   TRACEPOINT_ERROR_BYTECODE_MISSING,
   TRACEPOINT_ERROR_INTERNAL 
};

/**
 * @type MonitorTracepointByteCodeError
 * @brief Contains a constant from MonitorTraceError
 * @seealso MonitorTraceError
 */
typedef uint32_t MonitorTracepointByteCodeError;

/*-------------------------------------------------------------------------*/

/**
 * @signal MonitorTracepointCreateRequest
 *
 * @brief Create a new tracepoint.
 *
 * @long  Any number of tracepoints can be created, as well as any number of
 * actions.
 *
 * @member MONITOR_TRACEPOINT_CREATE_TRACEPOINT_REQUEST
 * @member MonitorTracepointCreateTracepointRequest
 *
 * @seealso MonitorTracepointCreateTracepointReply
 * @seealso MonitorTracepointCreateActionRequest
 */

/** @macro MONITOR_TRACEPOINT_CREATE_TRACEPOINT_REQUEST */
#define MONITOR_TRACEPOINT_CREATE_TRACEPOINT_REQUEST 40730  /* !-SIGNO(MonitorTracepointCreateTracepointRequest)-! */

/**
 * @type MonitorTracepointCreateTracepointRequest
 *
 * @field sigNo           MONITOR_TRACEPOINT_CREATE_TRACEPOINT_REQUEST
 * @input enabled         Enable/Disable flag
 * @input address         Address of the tracepoint.
 * @input scope           The set of processes that can fire the tracepoint.
 * @input passCount       Number of passes before stopping
 * @input tracepointType  The type of tracepoint,
 *                        i.e. one of the @link "" "MonitorTracepointTypes".
 * @field reserved0       NOT USED. Must be 0.
 * @field reserved1       NOT USED. Must be 0.
 * @field reserved2       NOT USED. Must be 0.
 * @input byteCodeLength  Number of bytes in the tracepoint condition.
 * @input byteCode        Bytecode for the tracepoint condition.
 */

typedef struct MonitorTracepointCreateTracepointRequest
{
   uint32_t sigNo;
   MonitorBoolean enabled;
   uint64_t address;
   MonitorScope scope;
   uint32_t passCount;
   uint32_t tracepointType;
   uint32_t reserved0;
   uint32_t reserved1;
   uint32_t reserved2;
   uint32_t byteCodeLength;
   uint8_t byteCode[1];
} MonitorTracepointCreateTracepointRequest;

/*-------------------------------------------------------------------------*/

/**
 * @signal MonitorTracepointCreateTracepointReply
 *
 * @brief Replies status of tracepoint creation.
 *
 * @member MONITOR_TRACEPOINT_CREATE_TRACEPOINT_REPLY
 * @member MonitorTracepointCreateTracepointReply
 *
 * @seealso MonitorTracepointCreateTracepointRequest
 *
 */

/** @macro MONITOR_TRACEPOINT_CREATE_TRACEPOINT_REPLY */
#define MONITOR_TRACEPOINT_CREATE_TRACEPOINT_REPLY 40731  /* !-SIGNO(MonitorTracepointCreateTracepointReply)-! */

/**
 * @type MonitorTracepointCreateTracepointReply
 *
 * @output sigNo                      MONITOR_TRACEPOINT_CREATE_TRACEPOINT_REPLY
 * @output status                     MONITOR_STATUS_OK or cause of error.
 * @output id                         ID of the tracepoint.
 * @output byteCodeError              TRACEPOINT_OK or error in Bytecode.
 * @field reserved0                   NOT USED.
 * @field reserved1                   NOT USED.
 */

typedef struct MonitorTracepointCreateTracepointReply
{
   uint32_t sigNo;
   MonitorStatus status;
   MonitorTracepoint id;
   MonitorTracepointByteCodeError byteCodeError;
   uint32_t reserved0;
   uint32_t reserved1;
} MonitorTracepointCreateTracepointReply;

/*-------------------------------------------------------------------------*/

/**
 * @signal MonitorTracepointRemoveTracepointRequest
 *
 * @brief Remove a tracepoint.
 *
 * @long  The tracepoint for the specified ID will be removed.
 *
 * @member MONITOR_TRACEPOINT_REMOVE_TRACEPOINT_REQUEST
 * @member MonitorTracepointRemoveTracepointRequest
 *
 * @seealso MonitorTracepointRemoveTracepointReply
 * @seealso MonitorTracepointCreateTracepointRequest
 *
 */

/** @macro MONITOR_TRACEPOINT_REMOVE_TRACEPOINT_REQUEST */
#define MONITOR_TRACEPOINT_REMOVE_TRACEPOINT_REQUEST 40732  /* !-SIGNO(MonitorTracepointRemoveTracepointRequest)-! */

/**
 * @type MonitorTracepointRemoveTracepointRequest
 *
 * @field sigNo         MONITOR_TRACEPOINT_REMOVE_TRACEPOINT_REQUEST
 * @input id            ID of the tracepoint.
 * @field reserved0     NOT USED. Must be 0.
 * @field reserved1     NOT USED. Must be 0.
 */

typedef struct MonitorTracepointRemoveTracepointRequest
{
   uint32_t sigNo;
   MonitorTracepoint id;
   uint32_t reserved0;
   uint32_t reserved1;
} MonitorTracepointRemoveTracepointRequest;

/*-------------------------------------------------------------------------*/

/**
 * @signal MonitorTracepointRemoveTracepointReply
 *
 * @brief Replies status of tracepoint removal.
 *
 * @member MONITOR_TRACEPOINT_REMOVE_TRACEPOINT_REPLY
 * @member MonitorTracepointRemoveTracepointReply
 *
 * @seealso MonitorTracepointRemoveTracepointRequest
 */

/** @macro MONITOR_TRACEPOINT_REMOVE_TRACEPOINT_REPLY */
#define MONITOR_TRACEPOINT_REMOVE_TRACEPOINT_REPLY 40733  /* !-SIGNO(MonitorTracepointRemoveTracepointReply)-! */

/**
 * @type MonitorTracepointRemoveTracepointReply
 *
 * @output sigNo        MONITOR_TRACEPOINT_REMOVE_TRACEPOINT_REPLY
 * @output status       MONITOR_STATUS_OK or cause of error.
 * @field reserved0     NOT USED.
 * @field reserved1     NOT USED.
 */
typedef struct MonitorTracepointRemoveTracepointReply
{
   uint32_t sigNo;
   MonitorStatus status;
   uint32_t reserved0;
   uint32_t reserved1;
} MonitorTracepointRemoveTracepointReply;

/*-------------------------------------------------------------------------*/

/**
 * @signal MonitorTracepointConfigureTracepointRequest
 *
 * @brief Enable/disable a tracepoint.
 *
 * @member MONITOR_TRACEPOINT_CONFIGURE_TRACEPOINT_REQUEST
 * @member MonitorTracepointConfigureTracepointRequest
 *
 */

/** @macro MONITOR_TRACEPOINT_CONFIGURE_TRACEPOINT_REQUEST */
#define MONITOR_TRACEPOINT_CONFIGURE_TRACEPOINT_REQUEST 40734  /* !-SIGNO(MonitorTracepointConfigureTracepointRequest)-! */

/**
 * @type MonitorTracepointConfigureTracepointRequest
 *
 * @field sigNo         MONITOR_TRACEPOINT_CONFIGURE_TRACEPOINT_REQUEST
 * @input id            Tracepoint ID.
 * @input enabled       Set to one if tracepoint enabled, zero otherwise.
 * @field reserved0     NOT USED. Must be 0.
 * @field reserved1     NOT USED. Must be 0.
 */
typedef struct MonitorTracepointConfigureTracepointRequest
{
   uint32_t sigNo;
   MonitorTracepoint id;
   MonitorBoolean enabled;
   uint32_t reserved0;
   uint32_t reserved1;
} MonitorTracepointConfigureTracepointRequest;

/*-------------------------------------------------------------------------*/

/**
 * @signal MonitorTracepointConfigureTracepointReply
 *
 * @brief Replies status of a request to enable/disable a tracepoint.
 *
 * @member MONITOR_TRACEPOINT_CONFIGURE_TRACEPOINT_REPLY
 * @member MonitorTracepointConfigureTracepointReply
 *
 */

/** @macro MONITOR_TRACEPOINT_CONFIGURE_TRACEPOINT_REPLY */
#define MONITOR_TRACEPOINT_CONFIGURE_TRACEPOINT_REPLY 40735  /* !-SIGNO(MonitorTracepointConfigureTracepointReply)-! */

/**
 * @type MonitorTracepointConfigureTracepointReply
 *
 * @output sigNo        MONITOR_TRACEPOINT_CONFIGURE_TRACEPOINT_REPLY
 * @output status       MONITOR_STATUS_OK or cause of error.
 * @field reserved0     NOT USED.
 * @field reserved1     NOT USED.
 */

typedef struct MonitorTracepointConfigureTracepointReply
{
   uint32_t sigNo;
   MonitorStatus status;
   uint32_t reserved0;
   uint32_t reserved1;
} MonitorTracepointConfigureTracepointReply;

/*-------------------------------------------------------------------------*/

/**
 * @type MonitorTracepointActionTypes
 *
 * @brief Specify the action type.
 *
 * @field MONITOR_TRACEPOINT_ACTION_SAVE_REGISTER     Save all register values.
 * @field MONITOR_TRACEPOINT_ACTION_EXPRESSION        Bytecode expression.
 * @field MONITOR_TRACEPOINT_ACTION_SAVE_MEMORY       Save memory.
 * @field MONITOR_TRACEPOINT_ACTION_WHILE_STEPPING    Repeat action while stepping.
 */
enum MonitorTracepointActionTypes
{
   MONITOR_TRACEPOINT_ACTION_SAVE_REGISTER,
   MONITOR_TRACEPOINT_ACTION_EXPRESSION,
   MONITOR_TRACEPOINT_ACTION_SAVE_MEMORY,
   MONITOR_TRACEPOINT_ACTION_WHILE_STEPPING
};
/**
 * @type MonitorTracepointAccessTypes
 *
 * @brief Specify the type of memory. When accessing hardware the access
 * pattern might have impact on the system
 *
 * @field MONITOR_TRACEPOINT_ACCESS_SIZE_NA   Regular memory, no restriction.
 * @field MONITOR_TRACEPOINT_ACCESS_SIZE_8    Collect data using  8 bit access.
 * @field MONITOR_TRACEPOINT_ACCESS_SIZE_16   Collect data using 16 bit access.
 * @field MONITOR_TRACEPOINT_ACCESS_SIZE_32   Collect data using 32 bit access.
 * @field MONITOR_TRACEPOINT_ACCESS_SIZE_64   Collect data using 64 bit access.
 */
enum MonitorTracepointAccessTypes
{
   MONITOR_TRACEPOINT_ACCESS_SIZE_NA = 0,
   MONITOR_TRACEPOINT_ACCESS_SIZE_8 = 1,
   MONITOR_TRACEPOINT_ACCESS_SIZE_16 = 2,
   MONITOR_TRACEPOINT_ACCESS_SIZE_32 = 4,
   MONITOR_TRACEPOINT_ACCESS_SIZE_64 = 8
};

/*
 * @type MonitorTracepointActionExpression
 *
 * @brief Specifies the bytecode.
 *
 * @long The used bytecode language is identical to the bytecode language
 * generated by the GDB agent expression mechanism. For more information, see
 * http://sourceware.org/gdb/current/onlinedocs/gdb/Agent-Expressions.html.
 *
 * @input ByteCodeLength   Bytecode length.
 * @input byteCode         Bytecode.
 */
typedef struct MonitorTracepointActionExpression
{
   uint32_t byteCodeLength;
   uint8_t byteCode[1];
} MonitorTracepointActionExpression;

/*
 * @type MonitorTracepointActionSaveMemory
 *
 * @brief Specifies the memory to be saved.
 *
 * @field isRegister       If set the base will be interpret as a
 *                         register ID that holds the base address.
 * @input base             Start address.
 * @input length           Number of bytes to be collected.
 */
typedef struct MonitorTracepointActionSaveMemory
{
   uint32_t isRegister;
   uint32_t length;
   uint64_t base;
} MonitorTracepointActionSaveMemory;

/*
 * @type MonitorTracepointActionWhileStepping
 *
 * @brief Specifies how many times an action shall be repeated while stepping.
 *
 * @input stepCount        Number of steps.
 */
typedef struct MonitorTracepointActionWhileStepping
{
   uint32_t stepCount;
} MonitorTracepointActionWhileStepping;

/*
 * @type MonitorTracepointActionData
 *
 * @brief Union used in MonitorTracepointCreateActionRequest to specify an action.
 *
 * @field expression       Specify byte code expressions.
 * @field saveMemory       Specify memory.
 * @field whileStepping    Specifies how many times an action shall be
 *                         repeated while stepping.
 */
typedef union MonitorTracepointActionData
{
   MonitorTracepointActionExpression expression;
   MonitorTracepointActionSaveMemory saveMemory;
   MonitorTracepointActionWhileStepping whileStepping;
} MonitorTracepointActionData;

/**
 * @signal MonitorTracepointCreateActionRequest
 *
 * @brief Request status of tracepoint creation.
 *
 * @member MONITOR_TRACEPOINT_CREATE_ACTION_REQUEST
 * @member MonitorTracepointCreateActionRequest
 *
 */

/** @macro MONITOR_TRACEPOINT_CREATE_ACTION_REQUEST */
#define MONITOR_TRACEPOINT_CREATE_ACTION_REQUEST 40736  /* !-SIGNO(MonitorTracepointCreateActionRequest)-! */

/**
 * @type MonitorTracepointCreateActionRequest
 *
 * @field sigNo               MONITOR_TRACEPOINT_CREATE_ACTION_REQUEST
 * @input tracepointId        ID of the tracepoint.
 * @input actionId            ID of the while stepping action the new action
 *                            shall be performed under or zero,
 *                            if the action only shall be performed once.
 * @field reserved0           NOT USED. Must be 0.
 * @field reserved1           NOT USED. Must be 0.
 * @field actionType          Identifier for the tracepointAction.
 * @field tracepointAction    The action to be performed.
 */
typedef struct MonitorTracepointCreateActionRequest
{
   uint32_t sigNo;
   MonitorTracepoint tracepointId;
   MonitorTracepointAction actionId;
   uint32_t reserved0;
   uint32_t reserved1;
   uint32_t actionType;
   MonitorTracepointActionData tracepointAction;
} MonitorTracepointCreateActionRequest;

/*-------------------------------------------------------------------------*/

/**
 * @signal MonitorTracepointCreateActionReply
 *
 * @brief Replies status of created actions.
 *
 * @member MONITOR_TRACEPOINT_CREATE_ACTION_REPLY
 * @member MonitorTracepointCreateActionReply
 *
 */

/** @macro MONITOR_TRACEPOINT_CREATE_ACTION_REPLY */
#define MONITOR_TRACEPOINT_CREATE_ACTION_REPLY 40737  /* !-SIGNO(MonitorTracepointCreateActionReply)-! */

/**
 * @type MonitorTracepointCreateActionReply
 *
 * @output sigNo        MONITOR_TRACEPOINT_CREATE_ACTION_REPLY
 * @output status       MONITOR_STATUS_OK or cause of error.
 * @output actionId     ID of the created tracepoint action.
 * @field reserved0     NOT USED.
 * @field reserved1     NOT USED.
 */

typedef struct MonitorTracepointCreateActionReply
{
   uint32_t sigNo;
   MonitorStatus status;
   MonitorTracepointAction actionId;
   uint32_t reserved0;
   uint32_t reserved1;
} MonitorTracepointCreateActionReply;

/*-------------------------------------------------------------------------*/

/**
 * @signal MonitorTracepointGetTracepointsRequest
 *
 * @brief Request all tracepoints for a certain scope ID. A list with scope
 * IDs can be retrieved by sending a MonitorTracepointGetAttachScopeRequest.
 *
 * @long This signal is sent under blocking flow control.
 *
 * @member MONITOR_TRACEPOINT_GET_TRACEPOINTS_REQUEST
 * @member MonitorTracepointGetTracepointsRequest
 *
 * @seealso MonitorTracepointGetTracepointsReply
 */

/** @macro MONITOR_TRACEPOINT_GET_TRACEPOINTS_REQUEST   */
#define MONITOR_TRACEPOINT_GET_TRACEPOINTS_REQUEST 40741  /* !-SIGNO(MonitorTracepointGetTracepointsRequest)-! */

/**
 * @type MonitorTracepointGetTracepointsRequest
 *
 * @field sigNo      MONITOR_TRACEPOINT_GET_TRACEPOINTS_REQUEST
 * @input scope      Monitor scope.
 * @field reserved0  NOT USED. Must be 0.
 * @field reserved1  NOT USED. Must be 0.
 */

typedef struct MonitorTracepointGetTracepointsRequest
{
   uint32_t sigNo;
   MonitorScope scope;
   uint32_t reserved0;
   uint32_t reserved1;
} MonitorTracepointGetTracepointsRequest;

/*-------------------------------------------------------------------------*/

/**
 * @signal MonitorTracepointGetTracepointsReply
 *
 * @brief  List tracepoints created for a scope.
 * @long   List tracepoints created for a scope, sent under
 *         blocking flow control.
 *
 * @member MONITOR_TRACEPOINT_GET_TRACEPOINTS_REPLY
 * @member MonitorTracepointGetTracepointsReply
 *
 * @seealso MonitorTracepointGetTracepointsRequest
 * @seealso MonitorTracepointGetTracepointsEndmark
 */

/** @macro MONITOR_TRACEPOINT_GET_TRACEPOINTS_REPLY   */
#define MONITOR_TRACEPOINT_GET_TRACEPOINTS_REPLY 40742  /* !-SIGNO(MonitorTracepointGetTracepointsReply)-! */

/**
 * @type MonitorTracepointGetTracepointsReply
 *
 * @field  sigNo            MONITOR_TRACEPOINT_GET_TRACEPOINTS_REPLY
 * @output id               ID of the tracepoint.
 * @output address          Address of the tracepoint.
 * @output scope            The set of processes that can fire the tracepoint.
 * @output enabled          Enable/disable flag.
 * @output passCount        Number of passes before stopping.
 * @output tracepointType   Type of tracepoint.
 * @field  reserved0        NOT USED.
 * @field  reserved1        NOT USED.
 * @field  reserved2        NOT USED.
 * @output byteCodeLength   Number of bytes in the tracepoint condition.
 * @output byteCode         Bytecode for the tracepoint condition.
 */
typedef struct MonitorTracepointGetTracepointsReply
{
   uint32_t sigNo;
   MonitorTracepoint id;
   uint64_t address;
   MonitorScope scope;
   MonitorBoolean enabled;
   uint32_t passCount;
   uint32_t tracepointType;
   uint32_t reserved0;
   uint32_t reserved1;
   uint32_t reserved2;
   uint32_t byteCodeLength;
   uint8_t byteCode[1];
} MonitorTracepointGetTracepointsReply;

/*-------------------------------------------------------------------------*/

/** @macro MONITOR_TRACEPOINT_GET_TRACEPOINTS_ENDMARK */
#define MONITOR_TRACEPOINT_GET_TRACEPOINTS_ENDMARK 40743 /* !-SIGNO(MonitorTracepointGetTracepointsEndmark)-! */

/**
 * @type MonitorTracepointGetTracepointsEndmark
 * @brief Endmark signal struct for information requests that result in multiple
 * reply signals.
 *
 * @output sigNo MONITOR_TRACEPOINT_GET_TRACEPOINTS_ENDMARK
 * @output status MONITOR_STATUS_OK or cause of failure.
 * @member MONITOR_TRACEPOINT_GET_TRACEPOINTS_ENDMARK
 */
typedef struct MonitorTracepointGetTracepointsEndmark
{
   uint32_t sigNo;
   MonitorStatus status;
} MonitorTracepointGetTracepointsEndmark;

/*-------------------------------------------------------------------------*/

/**
 * @signal MonitorTracepointGetActionsRequest
 *
 * @brief Request all actions for a specified tracepoint located in a certain scope.
 * The tracepoint IDs can be retrieved by sending a
 * MonitorTracepointGetTracepointsRequest.
 *
 * @long This signal is sent under blocking flow control.
 *
 * @member MONITOR_TRACEPOINT_GET_ACTIONS_REQUEST
 * @member MonitorTracepointGetActionsRequest
 *
 * @seealso MonitorTracepointGetActionsReply
 *
 */

/** @macro MONITOR_TRACEPOINT_GET_ACTIONS_REQUEST   */
#define MONITOR_TRACEPOINT_GET_ACTIONS_REQUEST 40744  /* !-SIGNO(MonitorTracepointGetActionsRequest)-! */

/**
 * @type MonitorTracepointGetActionsRequest
 *
 * @field sigNo           MONITOR_TRACEPOINT_GET_ACTIONS_REQUEST
 * @input tracepointId    ID of the tracepoint.
 * @input scope           Monitor scope.
 * @field reserved0       NOT USED. Must be 0.
 * @field reserved1       NOT USED. Must be 0.
 */

typedef struct MonitorTracepointGetActionsRequest
{
   uint32_t sigNo;
   uint32_t tracepointId;
   MonitorScope scope;
   uint32_t reserved0;
   uint32_t reserved1;
} MonitorTracepointGetActionsRequest;

/*-------------------------------------------------------------------------*/

/**
 * @signal MonitorTracepointGetActionsReply
 *
 * @brief  List actions created for a tracepoint.
 * @long   List actions created for a tracepoint, sent under
 *         blocking flow control.
 *
 * @member MONITOR_TRACEPOINT_GET_ACTIONS_REPLY
 * @member MonitorTracepointGetActionsReply
 *
 * @seealso MonitorTracepointGetActionsRequest
 * @seealso MonitorTracepointGetActionsEndmark
 */

/** @macro MONITOR_TRACEPOINT_GET_ACTIONS_REPLY   */
#define MONITOR_TRACEPOINT_GET_ACTIONS_REPLY 40745  /* !-SIGNO(MonitorTracepointGetActionsReply)-! */

/**
 * @type MonitorTracepointGetActionsReply
 *
 * @field  sigNo              MONITOR_TRACEPOINT_GET_ACTIONS_REPLY
 * @input tracepointId        Tracepoint ID.
 * @input actionId            Action ID.
 * @input whileSteppingId     While stepping ID.
 * @field reserved0           NOT USED.
 * @field reserved1           NOT USED.
 * @field actionType          Identifier for the tracepointAction.
 * @field reserved2           NOT USED.
 * @field tracepointAction    The action to be performed.
 */
typedef struct MonitorTracepointGetActionsReply
{
   uint32_t sigNo;
   MonitorTracepoint tracepointId;
   MonitorTracepointAction actionId;
   MonitorTracepointAction whileSteppingId;
   uint32_t reserved0;
   uint32_t reserved1;
   uint32_t actionType;
   uint32_t reserved2;
   MonitorTracepointActionData tracepointAction;
} MonitorTracepointGetActionsReply;

/*-------------------------------------------------------------------------*/

/** @macro MONITOR_TRACEPOINT_GET_ACTIONS_ENDMARK */
#define MONITOR_TRACEPOINT_GET_ACTIONS_ENDMARK 40746 /* !-SIGNO(MonitorTracepointGetActionsEndmark)-! */

/**
 * @type MonitorTracepointGetActionsEndmark
 * @brief Endmark signal struct for information requests that result in multiple
 * reply signals.
 *
 * @output sigNo MONITOR_TRACEPOINT_GET_ACTIONS_ENDMARK
 * @output status MONITOR_STATUS_OK or cause of failure.
 * @member MONITOR_TRACEPOINT_GET_ACTIONS_ENDMARK
 */
typedef struct MonitorTracepointGetActionsEndmark
{
   uint32_t sigNo;
   MonitorStatus status;
} MonitorTracepointGetActionsEndmark;

/*-------------------------------------------------------------------------*/

/**
 * @signal MonitorTracepointGetVariablesRequest
 *
 * @brief Request all variables created for a scope ID.
 * The scope ID can be retrieved by
 * sending a MonitorTracepointGetAttachScopeRequest.
 *
 * @long This signal is sent under blocking flow control.
 *
 * @member MONITOR_TRACEPOINT_GET_VARIABLES_REQUEST
 * @member MonitorTracepointGetVariablesRequest
 */

/** @macro MONITOR_TRACEPOINT_GET_VARIABLES_REQUEST   */
#define MONITOR_TRACEPOINT_GET_VARIABLES_REQUEST 40747  /* !-SIGNO(MonitorTracepointGetVariablesRequest)-! */

/**
 * @type MonitorTracepointGetVariablesRequest
 *
 * @field sigNo           MONITOR_TRACEPOINT_GET_VARIABLES_REQUEST
 * @input scope           Scope ID.
 * @field reserved0       NOT USED. Must be 0.
 * @field reserved1       NOT USED. Must be 0.
 *
 * @seealso MonitorTracepointGetVariablesReply
 */
typedef struct MonitorTracepointGetVariablesRequest
{
   uint32_t sigNo;
   MonitorScope scope;
   uint32_t reserved0;
   uint32_t reserved1;
} MonitorTracepointGetVariablesRequest;

/*-------------------------------------------------------------------------*/

/**
 * @signal MonitorTracepointGetVariablesReply
 *
 * @brief  List static trace variables for a certain scope.
 * @long   List static trace variables for a certain scope, sent under
 *         blocking flow control.
 *
 * @member MONITOR_TRACEPOINT_GET_VARIABLES_REPLY
 * @member MonitorTracepointGetVariablesReply
 *
 * @seealso MonitorTracepointGetVariablesRequest
 * @seealso MonitorTracepointGetVariablesEndmark
 */

/** @macro MONITOR_TRACEPOINT_GET_VARIABLES_REPLY   */
#define MONITOR_TRACEPOINT_GET_VARIABLES_REPLY 40748  /* !-SIGNO(MonitorTracepointGetVariablesReply)-! */

/**
 * @type MonitorTracepointGetVariablesReply
 *
 * @field  sigNo               MONITOR_TRACEPOINT_GET_VARIABLES_REPLY
 * @output variableStatus      if non zero, the value is unknown.
 * @output id                  ID of the variable.
 * @field padding              Padding.
 * @output initialValue        The initial value of the variable.
 * @field reserved0            NOT USED.
 * @field reserved1            NOT USED.
 */

typedef struct MonitorTracepointGetVariablesReply
{
   uint32_t sigNo;
   uint32_t variableStatus;
   MonitorTracepointVariable id;
   uint32_t padding;
   int64_t initialValue;
   uint32_t reserved0;
   uint32_t reserved1;
} MonitorTracepointGetVariablesReply;

/*-------------------------------------------------------------------------*/

/** @macro MONITOR_TRACEPOINT_GET_VARIABLES_ENDMARK */
#define MONITOR_TRACEPOINT_GET_VARIABLES_ENDMARK 40749 /* !-SIGNO(MonitorTracepointGetVariablesEndmark)-! */

/**
 * @type MonitorTracepointGetVariablesEndmark
 * @brief Endmark signal struct for information requests that result in multiple
 * reply signals.
 *
 * @output sigNo MONITOR_TRACEPOINT_GET_VARIABLES_ENDMARK
 * @output status MONITOR_STATUS_OK or cause of failure.
 * @member MONITOR_TRACEPOINT_GET_VARIABLES_ENDMARK
 */
typedef struct MonitorTracepointGetVariablesEndmark
{
   uint32_t sigNo;
   MonitorStatus status;
} MonitorTracepointGetVariablesEndmark;

/*-------------------------------------------------------------------------*/

/**
 * @signal MonitorTracepointCreateVariableRequest
 *
 * @brief  Create a trace variable
 *
 * @member MONITOR_TRACEPOINT_CREATE_VARIABLE_REQUEST
 * @member MonitorTracepointCreateVariableRequest
 */

/** @macro MONITOR_TRACEPOINT_CREATE_VARIABLE_REQUEST */
#define MONITOR_TRACEPOINT_CREATE_VARIABLE_REQUEST 40750  /* !-SIGNO(MonitorTracepointCreateVariableRequest)-! */

/**
 * @type MonitorTracepointCreateVariableRequest
 *
 * @field sigNo      MONITOR_TRACEPOINT_CREATE_VARIABLE_REQUEST
 * @input value      Tracepoint variable value.
 * @field reserved0  Client side ID of the tracepoint variable.
 *                   It is mandatory to set this field.
 *                   If not 0, all the tracepoint variable operations inside
 *                   the bytecode sent by the client during the session
 *                   must be encoded with client side IDs and all
 *                   the tracepoint variables must have client side IDs.
 *                   Everywhere else the IDs returned by the reply signal
 *                   have to be used. Must be set to 0 if the bytecode will use
 *                   the IDs returned by OSE.
 * @field reserved1  NOT USED. Must be 0.
 */

typedef struct MonitorTracepointCreateVariableRequest
{
   uint32_t sigNo;
   int64_t value;
   uint32_t reserved0;
   uint32_t reserved1;
} MonitorTracepointCreateVariableRequest;

/*-------------------------------------------------------------------------*/

/**
 * @signal MonitorTracepointCreateVariableReply
 *
 * @brief Replies status of created variable.
 *
 * @member MONITOR_TRACEPOINT_CREATE_VARIABLE_REPLY
 * @member MonitorTracepointCreateVariableReply
 *
 */

/** @macro MONITOR_TRACEPOINT_CREATE_VARIABLE_REPLY */
#define MONITOR_TRACEPOINT_CREATE_VARIABLE_REPLY 40751  /* !-SIGNO(MonitorTracepointCreateVariableReply)-! */

/**
 * @type MonitorTracepointCreateVariableReply
 *
 * @output sigNo        MONITOR_TRACEPOINT_CREATE_VARIABLE_REPLY
 * @output status       MONITOR_STATUS_OK or cause of error.
 * @output id           ID of the variable.
 * @field reserved0     NOT USED.
 * @field reserved1     NOT USED.
 */

typedef struct MonitorTracepointCreateVariableReply
{
   uint32_t sigNo;
   MonitorStatus status;
   MonitorTracepointVariable id;
   uint32_t reserved0;
   uint32_t reserved1;
} MonitorTracepointCreateVariableReply;

/*-------------------------------------------------------------------------*/

/**
 * @signal MonitorTracepointRemoveVariableRequest
 *
 * @brief  Remove a trace variable.
 *
 * @member MONITOR_TRACEPOINT_REMOVE_VARIABLE_REQUEST
 * @member MonitorTracepointRemoveVariableRequest
 */

/** @macro MONITOR_TRACEPOINT_REMOVE_VARIABLE_REQUEST */
#define MONITOR_TRACEPOINT_REMOVE_VARIABLE_REQUEST 40752  /* !-SIGNO(MonitorTracepointRemoveVariableRequest)-! */

/**
 * @type MonitorTracepointRemoveVariableRequest
 *
 * @field sigNo      MONITOR_TRACEPOINT_REMOVE_VARIABLE_REQUEST
 * @input id         ID of the variable.
 * @field reserved0  NOT USED. Must be 0.
 * @field reserved1  NOT USED. Must be 0.

 */

typedef struct MonitorTracepointRemoveVariableRequest
{
   uint32_t sigNo;
   MonitorTracepointVariable id;
   uint32_t reserved0;
   uint32_t reserved1;
} MonitorTracepointRemoveVariableRequest;

/*-------------------------------------------------------------------------*/

/**
 * @signal MonitorTracepointRemoveVariableReply
 *
 * @brief Replies status of the variable removal.
 *
 * @member MONITOR_TRACEPOINT_REMOVE_VARIABLE_REPLY
 * @member MonitorTracepointRemoveVariableReply
 *
 */

/** @macro MONITOR_TRACEPOINT_REMOVE_VARIABLE_REPLY */
#define MONITOR_TRACEPOINT_REMOVE_VARIABLE_REPLY 40753  /* !-SIGNO(MonitorTracepointRemoveVariableReply)-! */

/**
 * @type MonitorTracepointRemoveVariableReply
 *
 * @output sigNo        MONITOR_TRACEPOINT_REMOVE_VARIABLE_REPLY
 * @output status       MONITOR_STATUS_OK or cause of error.
 * @field reserved0     NOT USED.
 * @field reserved1     NOT USED.
 */

typedef struct MonitorTracepointRemoveVariableReply
{
   uint32_t sigNo;
   MonitorStatus status;
   uint32_t reserved0;
   uint32_t reserved1;
} MonitorTracepointRemoveVariableReply;

/*-------------------------------------------------------------------------*/

/**
 * @signal MonitorTracepointStartRequest
 *
 * @brief  Request start of the trace experiment.
 *
 * @member MONITOR_TRACEPOINT_START_REQUEST
 * @member MonitorTracepointStartRequest
 */

/** @macro MONITOR_TRACEPOINT_START_REQUEST   */
#define MONITOR_TRACEPOINT_START_REQUEST 40754  /* !-SIGNO(MonitorTracepointStartRequest)-! */

/**
 * @type MonitorTracepointStartRequest
 *
 * @field sigNo         MONITOR_TRACEPOINT_START_REQUEST
 * @field reserved0     NOT USED. Must be 0.
 * @field reserved1     NOT USED. Must be 0.
 */

typedef struct MonitorTracepointStartRequest
{
   uint32_t sigNo;
   uint32_t reserved0;
   uint32_t reserved1;
} MonitorTracepointStartRequest;

/*-------------------------------------------------------------------------*/

/**
 * @signal MonitorTracepointStartReply
 *
 * @brief Replies status of started tracepoint experiment.
 *
 * @member MONITOR_TRACEPOINT_START_REPLY
 * @member MonitorTracepointStartReply
 *
 */

/** @macro MONITOR_TRACEPOINT_START_REPLY */
#define MONITOR_TRACEPOINT_START_REPLY 40755  /* !-SIGNO(MonitorTracepointStartReply)-! */

/**
 * @type MonitorTracepointStartReply
 *
 * @output sigNo       MONITOR_TRACEPOINT_START_REPLY
 * @output status      MONITOR_STATUS_OK or cause of error.
 * @field reserved0    NOT USED.
 * @field reserved1    NOT USED.
 */

typedef struct MonitorTracepointStartReply
{
   uint32_t sigNo;
   MonitorStatus status;
   uint32_t reserved0;
   uint32_t reserved1;
} MonitorTracepointStartReply;

/*-------------------------------------------------------------------------*/

/**
 * @signal MonitorTracepointStopRequest
 *
 * @brief  Request stop of the trace experiment.
 *
 * @member MONITOR_TRACEPOINT_STOP_REQUEST
 * @member MonitorTracepointStopRequest
 */

/** @macro MONITOR_TRACEPOINT_STOP_REQUEST   */
#define MONITOR_TRACEPOINT_STOP_REQUEST 40756  /* !-SIGNO(MonitorTracepointStopRequest)-! */

/**
 * @type MonitorTracepointStopRequest
 *
 * @field  sigNo        MONITOR_TRACEPOINT_STOP_REQUEST
 * @field reserved0     NOT USED. Must be 0.
 * @field reserved1     NOT USED. Must be 0.
 */

typedef struct MonitorTracepointStopRequest
{
   uint32_t sigNo;
   uint32_t reserved0;
   uint32_t reserved1;
} MonitorTracepointStopRequest;

/*-------------------------------------------------------------------------*/

/**
 * @signal MonitorTracepointStopReply
 *
 * @brief Replies status of stopped tracepoint experiment.
 *
 * @member MONITOR_TRACEPOINT_STOP_REPLY
 * @member MonitorTracepointStopReply
 *
 */

/** @macro MONITOR_TRACEPOINT_STOP_REPLY */
#define MONITOR_TRACEPOINT_STOP_REPLY 40757  /* !-SIGNO(MonitorTracepointStopReply)-! */

/**
 * @type MonitorTracepointStopReply
 *
 * @output sigNo        MONITOR_TRACEPOINT_STOP_REPLY
 * @output status       MONITOR_STATUS_OK or cause of error.
 * @field reserved0     NOT USED.
 * @field reserved1     NOT USED.
 */

typedef struct MonitorTracepointStopReply
{
   uint32_t sigNo;
   MonitorStatus status;
   uint32_t reserved0;
   uint32_t reserved1;
} MonitorTracepointStopReply;

/*-------------------------------------------------------------------------*/

/**
 * @signal MonitorTracepointClearBufferRequest
 *
 * @brief  Request cleared tracepoint buffers.
 *
 * @member MONITOR_TRACEPOINT_CLEAR_BUFFER_REQUEST
 * @member MonitorTracepointClearBufferRequest
 */

/** @macro MONITOR_TRACEPOINT_CLEAR_BUFFER_REQUEST   */
#define MONITOR_TRACEPOINT_CLEAR_BUFFER_REQUEST 40758  /* !-SIGNO(MonitorTracepointClearBufferRequest)-! */

/**
 * @type MonitorTracepointClearBufferRequest
 *
 * @field sigNo      MONITOR_TRACEPOINT_CLEAR_BUFFER_REQUEST
 * @field reserved0  NOT USED. Must be 0.
 * @field reserved1  NOT USED. Must be 0.
 */

typedef struct MonitorTracepointClearBufferRequest
{
   uint32_t sigNo;
   uint32_t reserved0;
   uint32_t reserved1;
} MonitorTracepointClearBufferRequest;

/*-------------------------------------------------------------------------*/

/**
 * @signal MonitorTracepointClearBufferReply
 *
 * @brief Replies status of cleared tracepoint buffers.
 *
 * @member MONITOR_TRACEPOINT_CLEAR_BUFFER_REPLY
 * @member MonitorTracepointClearBufferReply
 *
 */

/** @macro MONITOR_TRACEPOINT_CLEAR_BUFFER_REPLY */
#define MONITOR_TRACEPOINT_CLEAR_BUFFER_REPLY 40759  /* !-SIGNO(MonitorTracepointClearBufferReply)-! */

/**
 * @type MonitorTracepointClearBufferReply
 *
 * @output sigNo        MONITOR_TRACEPOINT_CLEAR_REPLY
 * @output status       MONITOR_STATUS_OK or cause of error.
 * @field reserved0     NOT USED.
 * @field reserved1     NOT USED.
 */
typedef struct MonitorTracepointClearBufferReply
{
   uint32_t sigNo;
   MonitorStatus status;
   uint32_t reserved0;
   uint32_t reserved1;
} MonitorTracepointClearBufferReply;

/*-------------------------------------------------------------------------*/

/**
 * @type MonitorTraceStatus
 *
 * @brief Specify tracepoint experiment status.
 *
 * @field TRACEPOINT_RUNNING      Tracepoint experiment is running.
 * @field TRACEPOINT_NOTRUN       Tracepoint experiment is no started.
 * @field TRACEPOINT_STOP         Tracepoint experiment is stopped.
 * @field TRACEPOINT_FULL         Tracepoint experiment buffer is full.
 * @field TRACEPOINT_PASSCOUNT    Tracepoint experiment has reached passcount
 *                                limit.
 * @field TRACEPOINT_ERROR        Error was detected during tracepoint.
 *                                experiment.
 */
enum MonitorTraceStatus
{
   TRACEPOINT_RUNNING,
   TRACEPOINT_NOTRUN,
   TRACEPOINT_STOP,
   TRACEPOINT_FULL,
   TRACEPOINT_PASSCOUNT,
   TRACEPOINT_ERROR
};

/*-------------------------------------------------------------------------*/

/**
 * @signal MonitorTracepointGetStatusRequest
 *
 * @brief  Request status of the trace experiment.
 *
 * @member MONITOR_TRACEPOINT_GET_STATUS_REQUEST
 * @member MonitorTracepointGetStatusRequest
 */

/** @macro MONITOR_TRACEPOINT_GET_STATUS_REQUEST */
#define MONITOR_TRACEPOINT_GET_STATUS_REQUEST 40760  /* !-SIGNO(MonitorTracepointGetStatusRequest)-! */

/**
 * @type MonitorTracepointGetStatusRequest
 *
 * @field sigNo       MONITOR_TRACEPOINT_GET_STATUS_REQUEST
 * @field reserved0   NOT USED. Must be 0.
 * @field reserved1   NOT USED. Must be 0.
 */

typedef struct MonitorTracepointGetStatusRequest
{
   uint32_t sigNo;
   uint32_t reserved0;
   uint32_t reserved1;
} MonitorTracepointGetStatusRequest;

/*-------------------------------------------------------------------------*/

/**
 * @signal MonitorTracepointGetStatusReply
 *
 * @brief  Replies the status of the trace experiment.
 *
 * @member MONITOR_TRACEPOINT_GET_STATUS_REPLY
 * @member MonitorTracepointGetStatusReply
 */

/** @macro MONITOR_TRACEPOINT_GET_STATUS_REPLY   */
#define MONITOR_TRACEPOINT_GET_STATUS_REPLY 40761  /* !-SIGNO(MonitorTracepointGetStatusReply)-! */

/**
 * @type MonitorTracepointBufferStatus
 * @brief Structure with tracepoint buffer status returned in
 * the signal @link "" "MonitorTracepointGetStatusReply".
 *
 * @output frames         The number of trace frames in the buffer.
 * @output created        The total number of trace frames created during the run.
 * @output size           The total size of the tracepoint buffer, in bytes.
 * @output free           The number of bytes still unused in the buffer.
 * @output bufferType     Linear or circular buffer type.
 * @field reserved0       NOT USED.
 */
typedef struct MonitorTracepointBufferStatus
{
   uint64_t frames;
   uint64_t created;
   uint64_t size;
   uint64_t free;
   MonitorTracepointBufferType bufferType;
   uint32_t reserved0;
}MonitorTracepointBufferStatus;

/**
 * @type MonitorTracepointGetStatusReply
 *
 * @field  sigNo          MONITOR_TRACEPOINT_GET_STATUS_REPLY
 * @output status         MONITOR_STATUS_OK or cause of error.
 * @output traceStatus    If not running, explain why, see @link "" "MonitorTraceStatus".
 * @field reserved0       NOT USED.
 * @field reserved1       NOT USED.
 * @output cores          Number of cores.
 * @output perCoreStatus  Status information about trace buffer per core.
 */
typedef struct MonitorTracepointGetStatusReply
{
   uint32_t sigNo;
   MonitorStatus status;
   uint32_t traceStatus;
   uint32_t reserved0;
   uint32_t reserved1;
   uint32_t cores;
   MonitorTracepointBufferStatus perCoreStatus[1];
} MonitorTracepointGetStatusReply;

/*-------------------------------------------------------------------------*/

/**
 * @signal MonitorTracepointGetVariableValueRequest
 *
 * @brief  Request variable value.
 *
 * @member MONITOR_TRACEPOINT_GET_VARIABLE_VALUE_REQUEST
 * @member MonitorTracepointGetVariableValueRequest
 */

/** @macro MONITOR_TRACEPOINT_GET_VARIABLE_VALUE_REQUEST   */
#define MONITOR_TRACEPOINT_GET_VARIABLE_VALUE_REQUEST 40762  /* !-SIGNO(MonitorTracepointGetVariableValueRequest)-! */

/**
 * @type MonitorTracepointGetVariableValueRequest
 *
 * @field sigNo      MONITOR_TRACEPOINT_GET_VARIABLE_VALUE_REQUEST
 * @input id         ID of the variable.
 * @field reserved0  NOT USED. Must be 0.
 * @field reserved1  NOT USED. Must be 0.
 */
typedef struct MonitorTracepointGetVariableValueRequest
{
   uint32_t sigNo;
   MonitorTracepointVariable id;
   uint32_t reserved0;
   uint32_t reserved1;
} MonitorTracepointGetVariableValueRequest;

/*-------------------------------------------------------------------------*/

/**
 * @signal MonitorTracepointGetVariableValueReply
 *
 * @brief  Replies variable value.
 *
 * @member MONITOR_TRACEPOINT_GET_VARIABLE_VALUE_REPLY
 * @member MonitorTracepointGetVariableValueReply
 */

/** @macro MONITOR_TRACEPOINT_GET_VARIABLE_VALUE_REPLY   */
#define MONITOR_TRACEPOINT_GET_VARIABLE_VALUE_REPLY 40763  /* !-SIGNO(MonitorTracepointGetVariableValueReply)-! */

/**
 * @type MonitorTracepointGetVariableValueReply
 *
 * @field  sigNo          MONITOR_TRACEPOINT_GET_VARIABLE_VALUE_REPLY
 * @output status         MONITOR_STATUS_OK or cause of error.
 * @output variableStatus If non-zero, the value is unknown.
 * @output value          The value of the variable.
 * @output initialValue   The initial value of the variable.
 * @field reserved0       NOT USED.
 * @field reserved1       NOT USED.
 */

typedef struct MonitorTracepointGetVariableValueReply
{
   uint32_t sigNo;
   MonitorStatus status;
   uint32_t variableStatus;
   int64_t value;
   int64_t initialValue;
   uint32_t reserved0;
   uint32_t reserved1;
} MonitorTracepointGetVariableValueReply;

/*-------------------------------------------------------------------------*/

/**
 * @signal MonitorTracepointGetTracebufferRequest
 *
 * @brief  Request the complete tracepoint buffer.
 *
 * @long This signal requests a raw image of the tracepoint buffer in a
 * number of reply signals. An endmark signal notes the last fragment
 * was sent (or an error occurred).
 *
 * @member MONITOR_TRACEPOINT_GET_TRACEBUFFER_REQUEST
 * @member MonitorTracepointGetTracebufferRequest
 */

/** @macro MONITOR_TRACEPOINT_GET_TRACEBUFFER_REQUEST  */
#define MONITOR_TRACEPOINT_GET_TRACEBUFFER_REQUEST 40764  /* !-SIGNO(MonitorTracepointGetTracebufferRequest)-! */

/**
 * @type MonitorTracepointGetTracebufferRequest
 *
 * @field sigNo         MONITOR_TRACEPOINT_GET_TRACEBUFFER_REQUEST
 * @field reserved0     NOT USED. Must be 0.
 * @field reserved1     NOT USED. Must be 0.
 */

typedef struct MonitorTracepointGetTracebufferRequest
{
   uint32_t sigNo;
   uint32_t reserved0;
   uint32_t reserved1;
} MonitorTracepointGetTracebufferRequest;

/*-------------------------------------------------------------------------*/

/** @macro MONITOR_TRACEPOINT_GET_TRACEBUFFER_ENDMARK */
#define MONITOR_TRACEPOINT_GET_TRACEBUFFER_ENDMARK 40765 /* !-SIGNO(MonitorTracepointGetTracebufferEndmark)-! */

/**
 * @type MonitorTracepointGetTracebufferEndmark
 * @brief Endmark signal struct for information requests that result in multiple
 * reply signals.
 *
 * @output sigNo MONITOR_TRACEPOINT_GET_TRACEBUFFER_ENDMARK
 * @output status MONITOR_STATUS_OK or cause of failure.
 * @member MONITOR_TRACEPOINT_GET_TRACEBUFFER_ENDMARK
 */
typedef struct MonitorTracepointGetTracebufferEndmark
{
   uint32_t sigNo;
   MonitorStatus status;
} MonitorTracepointGetTracebufferEndmark;


/*-------------------------------------------------------------------------*/

/**
 * @signal MonitorTracepointGetTracebufferReply
 *
 * @brief  Replies the complete tracepoint buffer.
 *
 * @long This signal is sent under blocking flow control. One reply for every
 * time a tracepoint address was hit or a while-stepping step was performed. 
 * The data generated when performing an action is given in the buffer
 * traceData. The data consists of a number of chuncks. Each chunk begins
 * with a one byte character to specifying the content.
 *
 * @long 'R' <cpu_core>.                  The register as given as a cpu_core
 *                                        and has the size sizeof(cpu_core).
 *
 * @long 'M' <address> <length> <data>.   A memory block. Before the actual
 *                                        data comes a 64 bit address, which
 *                                        the memory is taken from, followed
 *                                        an 16 bit specifying the number of
 *                                        bytes in data.
 *
 * @long 'V' <id> <value>.                The value of a variable. id is the
 *                                        32 bit variable id and value is the
 *                                        64 bit value.
 *
 * @long 'E' <errorCode> <eCodeExtra>.    Written if an error occurred. The
 *                                        type of error is given in the 32 bit
 *                                        errorCode, see @link "" "MonitorTraceError".
 *                                        Additional information about the error
 *                                        is given in the 32 bit parameter eCodeExtra
 *                                        , see below.
 *
 * @long The field errorCode is set from @link "" "MonitorTraceError".
 * The field eCodeExtra holds different information for each error code,
 * see below:
 *
 * @long TRACEPOINT_ERROR_DIV_BY_ZERO - The parameter eCodeExtra holds no information.
 * Should be ignored.
 *
 * @long TRACEPOINT_ERROR_INSTRUCTION_NOT_IMPLEMENTED - The unimplemented byte code is
 * given in the eCodeExtra parameter.
 *
 * @long TRACEPOINT_ERROR_UNKNOWN_OP_CODE - The unrecognized byte code is
 * given in the eCodeExtra parameter.
 *
 * @long TRACEPOINT_ERROR_PROGRAM_TOO_LARGE - The parameter eCodeExtra is set to the
 * byte code index counter.
 *
 * @long TRACEPOINT_ERROR_ILLEGAL_GOTO_ADDRESS - The parameter eCodeExtra is set to the
 * invalid byte code index.
 *
 * @long TRACEPOINT_ERROR_MEMORY_READ - An error has occurred when reading the
 * memory in the chunk before. The parameter eCodeExtra is then the address
 * where the read failed.
 *
 * @long TRACEPOINT_ERROR_REGISTER - The parameter eCodeExtra is set to the
 * invalid register id.
 *
 * @long TRACEPOINT_ERROR_BAD_TVARIABLE_ADDRESS - The parameter eCodeExtra is set to the
 * invalid variable id.
 *
 * @long TRACEPOINT_ERROR_STACK_OVERFLOW - The parameter eCodeExtra holds no information.
 * Should be ignored.
 *
 * @long TRACEPOINT_ERROR_STACK_UNDERFLOW - The parameter eCodeExtra holds no information.
 * Should be ignored.
 *
 * @long TRACEPOINT_ERROR_TRACE_TOO_LARGE - The parameter eCodeExtra is set to wanted
 * signal size.
 *
 * @long TRACEPOINT_ERROR_EPM_ERROR - The parameter eCodeExtra is set to an epm error code,
 * see @link "" "EpmStatus"
 *
 * @member MONITOR_TRACEPOINT_GET_TRACEBUFFER_REPLY
 * @member MonitorTracepointGetTracebufferReply
 */

/** @macro MONITOR_TRACEPOINT_GET_TRACEBUFFER_REPLY */
#define MONITOR_TRACEPOINT_GET_TRACEBUFFER_REPLY 40766  /* !-SIGNO(MonitorTracepointGetTracebufferReply)-! */

/**
 * @type MonitorTracepointGetTracebufferReply
 *
 * @field  sigNo        MONITOR_TRACEPOINT_GET_TRACEBUFFER_REPLY
 * @output snapshotId   Sequence number of the trace.
 * @output pid          Process ID.
 * @output core         The CPU unit that the snapshot originating from.
 * @output errorCode    TP_OK, or the cause of error, see @link "" "MonitorTraceError".
 * @output eCodeExtra   Contains extra information about the error.
 * @output timeStamp    Timestamp when the trace was written.
 * @output sigLeft      Signals left for this snapshot.
 *                      0 is the last signal for the snapshot.
 *                      All signals will be sent with the same header
 *                      and the data should be concatenated.
 * @field  reserved0    NOT USED.
 * @field  reserved1    NOT USED.
 * @output tracepointId ID of tracepoint.
 * @output dataLength   Number of bytes in trace data.
 * @output traceData    Trace data.
 */

typedef struct MonitorTracepointGetTracebufferReply
{
   uint32_t sigNo;
   uint32_t snapshotId;
   uint32_t pid;
   uint32_t core;
   uint64_t timeStamp;
   uint32_t errorCode;
   uint32_t eCodeExtra;
   uint32_t sigLeft;
   uint32_t reserved0;
   uint32_t reserved1;
   MonitorTracepoint tracepointId;
   uint32_t dataLength;
   uint8_t traceData[1];
} MonitorTracepointGetTracebufferReply;

/*-------------------------------------------------------------------------*/

/**
 * @signal MonitorTracepointConfigureTracebufferRequest
 *
 * @brief Configure the tracepoint manager.
 *
 * @long  Set desired tracepoint buffer size and if a circular buffer
 * should be used. The allocated buffer will be split per core, 
 * so a 16k buffer allocated to a system with 4 cores 
 * will have a 4k buffer for each core.
 *
 * @member MONITOR_TRACEPOINT_CONFIGURE_TRACEBUFFER_REQUEST
 * @member MonitorTracepointConfigureTracebufferRequest
 *
 */

/** @macro MONITOR_TRACEPOINT_CONFIGURE_TRACEBUFFER_REQUEST */
#define MONITOR_TRACEPOINT_CONFIGURE_TRACEBUFFER_REQUEST 40767  /* !-SIGNO(MonitorTracepointConfigureTracebufferRequest)-! */

/**
 * @type MonitorTracepointConfigureTracebufferRequest
 *
 * @field sigNo           MONITOR_TRACEPOINT_CONFIGURE_REQUEST
 * @input bufferType      Linear or circular buffer type.
 * @input bufferSize      Set the size of the trace buffer.
 *                        Optional parameter, set to 0 for default value.
 * @field reserved0       NOT USED. Must be 0.
 * @field reserved1       NOT USED. Must be 0.
 */
typedef struct MonitorTracepointConfigureTracebufferRequest
{
   uint32_t sigNo;
   MonitorTracepointBufferType bufferType;
   uint64_t bufferSize;
   uint32_t reserved0;
   uint32_t reserved1;
} MonitorTracepointConfigureTracebufferRequest;

/*-------------------------------------------------------------------------*/

/**
 * @signal MonitorTracepointConfigureTracebufferReply
 *
 * @brief Replies status of configure tracepoint buffer request.
 *
 * @member MONITOR_TRACEPOINT_CONFIGURE_TRACEBUFFER_REPLY
 * @member MonitorTracepointConfigureTracebufferReply
 *
 */

/** @macro MONITOR_TRACEPOINT_CONFIGURE_TRACEBUFFER_REPLY */
#define MONITOR_TRACEPOINT_CONFIGURE_TRACEBUFFER_REPLY 40768  /* !-SIGNO(MonitorTracepointConfigureTracebufferReply)-! */

/**
 * @type MonitorTracepointConfigureTracebufferReply
 *
 * @output sigNo         MONITOR_TRACEPOINT_CONFIGURE_TRACEBUFFER_REPLY
 * @output status        MONITOR_STATUS_OK or cause of error.
 * @output changeStatus  Status of the change.
 * @field reserved0      NOT USED.
 * @field reserved1      NOT USED.
 */

typedef struct MonitorTracepointConfigureTracebufferReply
{
   uint32_t sigNo;
   MonitorStatus status;
   uint32_t changeStatus;
   uint32_t reserved0;
   uint32_t reserved1;
} MonitorTracepointConfigureTracebufferReply;

/*-------------------------------------------------------------------------*/

/**
 * @signal MonitorTracepointSetAnnotationsRequest
 *
 * @brief Sets optional annotations for a tracepoint.
 *
 * @long There is one string holding all annotations per tracepoint. The
 *       annotations are separated by the end of string character '\0'. Sending
 *       a new MonitorTracepointSetAnnotationsRequest for a specific tracepoint
 *       replaces the old annotations with the new ones.
 *
 * @member MONITOR_TRACEPOINT_SET_ANNOTATIONS_REQUEST
 * @member MonitorTracepointSetAnnotationsRequest
 *
 * @seealso MonitorTracepointSetAnnotationsReply
 */

/** @macro MONITOR_TRACEPOINT_SET_ANNOTATIONS_REQUEST   */
#define MONITOR_TRACEPOINT_SET_ANNOTATIONS_REQUEST 40769  /* !-SIGNO(MonitorTracepointSetAnnotationsRequest)-! */

/**
 * @type MonitorTracepointSetAnnotationsRequest
 *
 * @field sigNo              MONITOR_TRACEPOINT_SET_ANNOTATIONS_REQUEST
 * @input tracepointId       ID of the tracepoint.
 * @input no_of_annotations  The number of annotations in annotations area
 * @input annotations_size   The size of the annotations area
 *                           (number of bytes).
 * @field reserved0          NOT USED. Must be 0.
 * @field reserved1          NOT USED. Must be 0.
 * @input annotations        The annotations separated by the end of string
 *                           character '\0'. There must be no_of_annotations
 *                           annotations contained in this area, all terminated
 *                           by the end of string character and the total size
 *                           of annotations and terminators has to match
 *                           annotations_size.
 */
typedef struct MonitorTracepointSetAnnotationsRequest
{
   uint32_t sigNo;
   MonitorTracepoint tracepointId;
   uint32_t no_of_annotations;
   uint32_t annotations_size;
   uint32_t reserved0;
   uint32_t reserved1;
   char annotations[1];
} MonitorTracepointSetAnnotationsRequest;

/*-------------------------------------------------------------------------*/

/**
 * @signal MonitorTracepointSetAnnotationsReply
 *
 * @brief Replies status of setting annotations.
 *
 * @member MONITOR_TRACEPOINT_SET_ANNOTATIONS_REPLY
 * @member MonitorTracepointSetAnnotationsReply
 *
 * @seealso MonitorTracapointSetAnnotationsRequest
 */

/** @macro MONITOR_TRACEPOINT_SET_ANNOTATIONS_REPLY    */
#define MONITOR_TRACEPOINT_SET_ANNOTATIONS_REPLY 40770  /* !-SIGNO(MonitorTracepointSetAnnotationsReply)-! */

/**
 * @type MonitorTracepointSetAnnotationsReply
 *
 * @field sigNo       MONITOR_TRACEPOINT_SET_ANNOTATIONS_REPLY
 * @output status     MONITOR_STATUS_OK or cause of error
 * @field reserved0   NOT USED. Must be 0.
 * @field reserved1   NOT USED. Must be 0.
 */

typedef struct MonitorTracepointSetAnnotationsReply
{
   uint32_t sigNo;
   MonitorStatus status;
   uint32_t reserved0;
   uint32_t reserved1;
} MonitorTracepointSetAnnotationsReply;

/*-------------------------------------------------------------------------*/

/**
 * @signal MonitorTracepointGetAnnotationsRequest
 *
 * @brief Request to retrieve the annotations for a specified tracepoint
 * located in a certain scope.
 * The tracepoint IDs can be retrieved by sending a
 * MonitorTracepointGetTracepointsRequest.
 *
 * @member MONITOR_TRACEPOINT_GET_ANNOTATIONS_REQUEST
 * @member MonitorTracepointGetAnnotationsRequest
 *
 * @seealso MonitorTracepointGetAnnotationsReply
 */

/** @macro MONITOR_TRACEPOINT_GET_ANNOTATIONS_REQUEST   */
#define MONITOR_TRACEPOINT_GET_ANNOTATIONS_REQUEST 40771 /* !-SIGNO(MonitorTracepointGetAnnotationsRequest)-! */

/**
 * @type MonitorTracepointGetAnnotationsRequest
 *
 * @field sigNo         MONITOR_TRACEPOINT_GET_ANNOTATIONS_REQUEST
 * @input tracepointId  ID of the tracepoint.
 * @input scope         Monitor scope.
 * @field reserved0     NOT USED. Must be 0.
 * @field reserved1     NOT USED. Must be 0.
 */
typedef struct MonitorTracepointGetAnnotationsRequest
{
   uint32_t sigNo;
   MonitorTracepoint tracepointId;
   MonitorScope scope;
   uint32_t reserved0;
   uint32_t reserved1;
} MonitorTracepointGetAnnotationsRequest;

/*-------------------------------------------------------------------------*/

/**
 * @signal MonitorTracepointGetAnnotationsReply
 *
 * @brief Returns the annotations for a tracepoint.
 *
 * @member MONITOR_TRACEPOINT_GET_ANNOTATIONS_REPLY
 * @member MonitorTracepointGetAnnotationsReply
 *
 * @seealso MonitorTracepointGetAnnotationsRequest
 */

/** @macro MONITOR_TRACEPOINT_GET_ANNOTATIONS_REPLY   */
#define MONITOR_TRACEPOINT_GET_ANNOTATIONS_REPLY 40772 /* !-SIGNO(MonitorTracepointGetAnnotationsReply)-! */

/**
 * @type MonitorTracepointGetAnnotationsReply
 *
 * @field sigNo               MONITOR_TRACEPOINT_GET_ANNOTATIONS_REPLY
 * @input tracepointId        Tracepoint ID.
 * @output status             MONIOTOR_STATUS_OK or cause of error.
 * @output no_of_annotations  The number of annotations in annotations area
 * @output annotations_size   The size of the annotations area
 *                            (number of bytes).
 * @field reserved0           NOT USED. Must be 0.
 * @field reserved1           NOT USED. Must be 0.
 * @output annotations        The annotations separated by the end of string
 *                            character '\0'. There are no_of_annotations
 *                            annotations contained in this area, all
 *                            terminated by the end of string character and
 *                            the total size of annotations and terminators
 *                            matches annotations_size.
 */
typedef struct MonitorTracepointGetAnnotationsReply
{
   uint32_t sigNo;
   MonitorTracepoint tracepointId;
   MonitorStatus status;
   uint32_t no_of_annotations;
   uint32_t annotations_size;
   uint32_t reserved0;
   uint32_t reserved1;
   char annotations[1];
} MonitorTracepointGetAnnotationsReply;

/*-------------------------------------------------------------------------*/

#ifdef __cplusplus
} /* end of extern "C" */
#endif

#endif

