/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2011 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */
/**
 * @toc Memory_Management:
 * @file ose_spi/ose_heap.h
 * @brief The heap configuration interface.
 * @long The file ose_heap.h contains the interface that needs to be fulfilled
 *       in order to be able to implement an OSE heap.
 * @long This interface is mainly used for two use cases:
 * @long - Managing the execution environment for separately loaded programs
 *       (performed by the Program Manager).
 * @long - Managing shared, global heaps.
 */

#ifndef _OSE_HEAP_H
#define _OSE_HEAP_H

#include <stddef.h>
#include "ose.h"
#include "osetypes.h"
#include "heapapi.h"
#include "mmtypes.h"

#ifdef __cplusplus
extern "C" {
#endif

/**
 * @private
 *
 * @macro HEAPREF_ENVNAME
 *
 * @brief Environment variable with the heap link entry - set internally.
 * @long Each block that has a valid heap, will have a block/program
 *       variable with this name containing an OSHEAP reference to a valid
 *       heap.
 * @long The variable SHOULD NOT be set by a user,
 *       it will be set by the ose_create_heap() call.
 * @long The name of the environment variable is defined here to: "heap_ref"
 */
#define HEAPREF_ENVNAME     "heap_ref"


/**
 * @macro HEAP_INVALID_TENTATIVE_HANDLE
 *
 * @brief A constant describing invalid tentative buffer handles.
 *
 * @long  The value is returned from
 *        @link "heapapi.h" "ose_heap_free_tentative" if the
 *        tentative buffer cache was full and the buffer was freed
 *        permanently, or if the user supplied buffer pointer was
 *        NULL.
 *
 * @seealso ose_heap_free_tentative
 */
#define HEAP_INVALID_TENTATIVE_HANDLE  ((U32)~0)


/**
 * @private
 *
 * @type  HeapFileLine
 *
 * @brief Structure with information about where a heap request originated.
 *
 * @long  This structure is used internally by the heap implementation and
 *        serves no purpose to regular users.
 *
 * @field file  The file name.
 * @field line  The line number.
 *
 */
typedef struct HeapFileLine HeapFileLine;
struct HeapFileLine
{
   const char *file;
   int         line;
};


typedef struct _HUCB HUCB;
struct _HUCB
{
  U32 version; /* indicates the number of heap functions supported. */
  void   (*register_client)(HUCB *heap, PROCESS owner);
  void * (*alloc_unsafe)(HUCB *heap, size_t size, PROCESS owner, Boolean annotate);
  void * (*alloc_safe)(HUCB *heap, size_t size, PROCESS owner, Boolean annotate);
  void   (*hfree)(HUCB *heap, void *ptr, PROCESS owner);
  size_t (*buffer_size)(HUCB *heap, void *ptr, PROCESS owner);
  size_t (*buffer_actual_size)(HUCB *heap, void *ptr, PROCESS owner);
  size_t (*buffer_resize)(HUCB *heap, void *ptr, PROCESS owner, size_t new_size);
  void   (*extend)(HUCB *heap, void *frag_addr, size_t frag_size);
  void * (*alloc_early_null)(HUCB *heap, size_t size, PROCESS owner, Boolean annotate);
  size_t (*buffer_next_size)(HUCB *heap, size_t size);
  OSHEAP (*handle_create_heap)(PROCESS pid, OSDOMAIN domain,
                               OSBOOLEAN supervisor_mode);
  size_t (*set_debug_hooks)(HUCB *heap, size_t annsz, HeapDebugHooks *hooks);
  void   (*remove_debug_hooks)(HUCB *heap);
  void   (*forced_free)(HUCB *heap, void *ptr);
  void * (*alloc_no_extend)(HUCB *heap, size_t size, PROCESS owner, Boolean annotate);
  DEBUG_ALLOC_FUNC heap_debug_alloc;
  DEBUG_FREE_FUNC  heap_debug_free;
  DEBUG_FILELINE_FUNC heap_debug_fileline;
  DEBUG_CLEANUP_FUNC heap_debug_cleanup;
  size_t annotation_size;
  size_t (*set_max_size)(HUCB *heap, size_t new_max);

  /* Introduced in OSE 5.4. */
  void * (*alloc_tentative)(HUCB *, U32, PROCESS, Boolean);
  U32    (*free_tentative)(HUCB *, void*, U32, PROCESS, Boolean);
  void * (*alloc_watermark)(HUCB *, size_t, U32, PROCESS, Boolean);

  /* Introduced in OSE 5.5. */
  void * (*plug_alloc)  (HUCB *hucb, U32 *params,
                         HeapFileLine *fl, PROCESS pid, size_t size);
  void   (*plug_free)   (HUCB *hucb, U32 *params,
                         HeapFileLine *fl, PROCESS pid, void *ptr);
  void * (*plug_resize) (HUCB *hucb, U32 *params,
                         HeapFileLine *fl, PROCESS pid, void *ptr, size_t size);
  void   (*plug_collect)(HUCB *hucb, U32 *params,
                         HeapFileLine *fl, PROCESS pid);
  void   (*plug_query)  (HUCB *hucb, U32 *params);
  void   (*plug_control)(HUCB *hucb, U32 *params);
};
#define HEAP_VERSION sizeof(HUCB)


/**
* @function ose_reset_heap
* @declaration
* void ose_reset_heap(PROCESS bid);
*
* @brief Reset the state for a block to not having an active heap.
* @long Resets the block's heap link information, which means
*       that the block no longer has a default heap. It is therefore
*       necessary to call ose_create_heap in order to set up a new heap.
*       Note however that all configuration values that are possible to set as
*       environment variables on the block, will be inherited from the parent
*       creator and will still be valid for the newly created block/program.
*
* @input bid The BID of the OSE block that should get a heap reset.
*
* @restriction The function is NOT available outside the RTOS monolith.
*       @break The destination block is expected to exist throughout
*              this call.
*       @break Only available in supervisor mode processes.
*
* @seealso default_restrictions_calling_functions
*/
void ose_reset_heap(PROCESS bid);

/**
 * @function ose_create_heap
 * @declaration
 * OSHEAP ose_create_heap(PROCESS pid, OSDOMAIN domain, OSBOOLEAN supervisor_mode);
 *
 * @brief Create a new heap.
 *
 * @long The default dynamic heap scenario:
 * @long The function ose_create_heap() is called by the program manager to
 *       create a new heap to be used by the given block/program. All processes
 *       created in the block and child blocks will use the created heap.
 * @long This function will set a valid heap reference to the environment of the block,
 *       which will "override" any possible heap inherited via its environment.
 *       The program loader transfers all load module parameters to environment
 *       variables of the created block, so in order to set heap configuration values
 *       they should be defined as load module parameters.
 *
 * @long The user mode global heap scenario:
 * @long A supervisor in the platform creates a phantom process, which is then used
 *       in the ose_create_heap() call. It is the responsibility of the supervisor
 *       to manage this PID, because when the PID is killed, the heap is automatically
 *       deleted and the memory released in the same way as when a block owns the
 *       heap.
 *
 * @long Configuration, all cases:
 * @long The supervisor/creator sets a number of attributes as environment variables on
 *       the owner PID, as for example cache properties and the logical memory (log_mem)
 *       from which the heap memory shall be allocated. See the list of variables below.
 *       The various default values described for the variables are valid for a default
 *       application where the heap is domain private and uses cacheable RAM memory.
 * @long Configuration, global heap:
 * @long When creating a user mode global heap, the domain ID
 *       should be set to zero and the supervisor_mode flag to False.
 *
 * @long The heap configuration works as follows:
 * @long The ose_create_heap() searches for PID HEAP environment variables
 *       as input parameters, which means that parameters set after the ose_create_heap()
 *       call will not have any effect.
 * @long A clean initialization of a new heap is achieved by first calling
 *       the function ose_reset_heap(), then setting the PID environment parameters,
 *       and last calling ose_create_heap.
 * @long NOTE: The initial fragment size and smallbuf extension size
 *       must be tuned to fit, if the max size needs to be fairly low, in the
 *       order of magnitude of the default fragment init and extension sizes.
 *
 * @long HEAP configuration environment variables:
 * @long The configuration variables can optionally be set to suitable values
 *       by the client, before a heap is created. If set by the client,
 *       numeric values will be aligned by the heap module to the smallest page
 *       size. After a heap has been created, these values can be obtained via
 *       the @link "" "ose_spi/heap.sig" management interface.
 *
 * @long @link "" "HEAP_SMALL_BUF_INIT_SIZE",
 *       @link "" "HEAP_SMALL_BUF_EXT_SIZE",
 *       @link "" "HEAP_LARGE_BUF_THRESHOLD",
 *       @link "" "HEAP_MAX_SIZE",
 *       @link "" "HEAP_MAX_TYPE",
 *       @link "" "HEAP_CACHE_MODE",
 *       @link "" "HEAP_PERMISSIONS",
 *       @link "" "HEAP_LOG_MEM",
 *       @link "" "HEAP_PHYS_MEM",
 *       @link "" "HEAP_CLEANUPD_STACK_SIZE",
 *       @link "" "HEAP_FRAG_CHECK",
 *       @link "" "HEAP_LOCK_TYPE",
 *       @link "" "HEAP_ALGO_BUDDY",
 *       @link "" "HEAP_ALGO_TLSF",
 *
 *       @link "" "HEAP_CACHE_MAX_BUFCOUNT",
 *       @link "" "HEAP_CACHE_MAX_BUFSIZE",
 *       @link "" "HEAP_RECLAIM",
 *       @link "" "HEAP_TENTATIVE_BUFCOUNT".
 *
 * @input pid The BID of the main OSE block for the program that should use this heap,
 *        or the phantom PID that will be the owner of a global heap.
 *
 * @input domain The MMU protection domain ID in which the heap fragments shall be
 *        allocated. Should be set to zero for the system heap and global
 *        heaps.
 *
 * @input supervisor_mode Should be set to True if the block (heap clients) runs in
 *        supervisor mode, and set to False if the heap should be possible to access
 *        from user mode.
 *
 * @return A reference to the newly created heap or 0 if creation failed.
 *
 * @restriction Parameters set as block environment variables for this heap will
 *              not have effect after this call is done, i.e. the desired parameters
 *              must be set prior to the ose_create_heap() call.
 *
 *      @break  The destination block is expected to exist throughout this call.
*       @break Only available in supervisor mode processes.
 *
 * @seealso default_restrictions_calling_functions
*/
OSHEAP ose_create_heap(PROCESS pid,
                       OSDOMAIN domain,
                       OSBOOLEAN supervisor_mode);


/**
 * @macro HEAP_SMALL_BUF_INIT_SIZE
 *
 * @brief Environment variable with the size of the initial heap memory chunk.
 *
 * @long  If a static heap is desired, set the initial size equal to the
 *        @link "" "HEAP_MAX_SIZE" size.
 *
 * @long  When the configured small buffer algorithm (Fibonacci buddy, TLSF)
 *        initializes a chunk of memory, it may be split into smaller
 *        fragments if the memory chunk is larger than the largest
 *        supported buffer size.
 *
 * @long  The default initial size is 286720 bytes.
 *
 * @long  The configured value is always rounded upwards to a multiple of the
 *        smallest supported page size (4KB for ARM, PowerPC and SFK, 8KB
 *        for MIPS). If the initial size is larger than the maximum size,
 *        the initial size will be adjusted down to the maximum size.
 *        
 */
#define HEAP_SMALL_BUF_INIT_SIZE "HEAP_SMALL_BUF_INIT_SIZE"

/**
 * @macro HEAP_SMALL_BUF_EXT_SIZE
 *
 * @brief Environment variable with the size of heap memory extensions.
 *
 * @long  The default extension size is 45056 bytes if neither 
 *        @link "" "HEAP_SMALL_BUF_EXT_SIZE" nor
 *        @link "" "HEAP_SMALL_BUF_INIT_SIZE" have been set.
 *        If @link "" "HEAP_SMALL_BUF_INIT_SIZE" is set but not
 *        @link "" "HEAP_SMALL_BUF_EXT_SIZE", the default extension size
 *        is set to one fifth of the initial size and rounded down to the
 *        nearest supported buffer size.
 *        
 * @long  The configured value is always rounded upwards to a multiple of
 *        the smallest supported page size (4KB for ARM, PowerPC and SFK, 8KB
 *        for MIPS).
 */
#define HEAP_SMALL_BUF_EXT_SIZE  "HEAP_SMALL_BUF_EXT_SIZE"

/**
 * @macro HEAP_LARGE_BUF_THRESHOLD
 *
 * @brief Environment variable with the threshold size for large
 *        buffer allocations.
 *
 * @long  Requested buffers, where the buffer size exceeds this threshold,
 *        will be allocated directly from MM instead of in the heap area.
 *        Adjustments of the parameter may help to decrease external
 *        fragmentation.
 *
 * @long  The maximum value allowed is ~8MB when the Fibonacci buddy
 *        algorithm is used and ~16MB when the TLSF algorithm is used.
 *        The threshold can also be set to 0 which effectively disables all
 *        small buffer algorithms (Fibonacci buddy, TLSF). The default
 *        value equals the maximum value.
 *
 * @long  NOTE: If the configured threshold size is too large, the value
 *        will be adjusted down and set to the maximum allowed threshold.
 */
#define HEAP_LARGE_BUF_THRESHOLD "HEAP_LARGE_BUF_THRESHOLD"

/**
 * @macro HEAP_MAX_SIZE
 *
 * @brief Environment variable specifying the maximum size of the heap.
 *
 * @long  The default value is set to 256MB (0x10000000).
 *
 * @seealso HEAP_MAX_TYPE
 */
#define HEAP_MAX_SIZE "HEAP_MAX_SIZE"

/**
 * @macro HEAP_MAX_TYPE
 *
 * @brief Environment variable used to change the behavior of
 *        @link "" "HEAP_MAX_SIZE".
 *
 * @long  The heap can be configured to either limit the amount of
 *        system memory the heap may consume (default behavior), or
 *        the amount of memory applications are allowed to allocate
 *        from the heap.
 *
 * @long  Accepted values are: @break
 *
 *        @link "" "HEAP_MAX_TYPE_FRAGMENTS"  @break
 *        @link "" "HEAP_MAX_TYPE_BUFFERS"
 *
 */
#define HEAP_MAX_TYPE "HEAP_MAX_TYPE"


/**
 * @macro HEAP_MAX_TYPE_FRAGMENTS
 *
 * @brief An option to @link "" "HEAP_MAX_TYPE" that changes the behavior
 *        of @link "" "HEAP_MAX_SIZE".
 *
 * @long  When used, the configured maximum heap size limits the amount of
 *        system memory the heap may consume. This type is the default type.
 */
#define HEAP_MAX_TYPE_FRAGMENTS  "1"

/**
 * @macro HEAP_MAX_TYPE_BUFFERS
 *
 * @brief An option to @link "" "HEAP_MAX_TYPE" that changes the behavior
 *        of @link "" "HEAP_MAX_SIZE".
 *
 * @long  When used, the configured maximum heap size limits the
 *        amount of memory applications are allowed to allocated from
 *        the heap using malloc() and other heap allocation methods.
 *
 * @long  This type may cause consumption of system memory in amounts that
 *        exceed the configured maximum heap size due to fragmentation of
 *        the heap memory.
 *
 * @long  This type is the legacy behavior in previous versions of OSE and
 *        can be used to resolve compatibility issues regarding the heap size.
 */
#define HEAP_MAX_TYPE_BUFFERS    "2"


/**
 * @macro HEAP_CACHE_MODE
 *
 * @brief Environment variable with the cache properties of the
 *        heap memory.
 *
 * @long  This cache attribute is the value given when the heap
 *        allocates memory from the MM. For possible values, see mm.sig.
 *        The value is set as a numeric value, and the default is
 *        (MM_CACHE_COPY_BACK | MM_SPECULATIVE_ACCESS).
 *
 * @seealso ose_spi/mm.sig
 */
#define HEAP_CACHE_MODE          "HEAP_CACHE_MODE"

/**
 * @macro HEAP_PERMISSIONS
 *
 * @brief Environment variable with the permissions of the heap memory.
 *
 * @long  This is the value given when the heap allocates memory from the
 *        MM. For possible values, see mm.sig.
 *
 * @seealso ose_spi/mm.sig
 */

#define HEAP_PERMISSIONS          "HEAP_PERMISSIONS"
/**
 * @macro HEAP_LOG_MEM
 *
 * @brief Environment variable with the logical memory to be used in
 *        heap MM allocations.
 *
 * @long  The heap by default tries to allocate from the mandatory log_mem "RAM",
 *        This variable should only need to be set if the heap is located
 *        in non-standard RAM.
 */
#define HEAP_LOG_MEM          "HEAP_LOG_MEM"

/**
 * @macro HEAP_PHYS_MEM
 *
 * @brief Environment variable with the physical memory to be used in
 *        heap MM allocations.
 *
 * @long  The heap by default tries to allocate from the mandatory phys_mem "RAM".
 *        This variable should only need to be set if the heap logical
 *        memory is located in non-standard RAM.
 */
#define HEAP_PHYS_MEM          "HEAP_PHYS_MEM"

/**
 * @macro HEAP_EARLY_NULL_LEVEL
 *
 * @deprecated This environment variable is deprecated from OSE 5.5 and will
 *             not be supported in future releases of OSE. The variable has no
 *             effect. Watermark allocation methods remain functional
 *             but ignore the watermark level. They may therefore
 *             exhaust the heap.
 *
 * @brief Environment variable with a comma separated list of watermark
 *        levels (% of HEAP_MAX_SIZE) used by priority allocation methods.
 *
 * @long  The heap can be set to return NULL for allocations before the
 *        heap has reached its maximum value set by HEAP_MAX_SIZE. This is
 *        achieved by configuring a low watermark using this variable.
 *        The low watermark is a percentage of the maximum size ranging
 *        from 1-100. The function heap_alloc_early_null returns NULL instead
 *        of new buffer if the heap usage reaches this level. The default level
 *        is 100%, i.e. heap_alloc_early_null will allocate buffers until the
 *        heap is exhausted.
 *
 * @long  Multiple low watermark levels are supported by listing values
 *        as a comma separated list. heap_alloc_early_null always allocates
 *        according to the first configured level. To use additional levels,
 *        ose_heap_realloc_watermark must be used.
 *
 * @long  This "low watermark" feature is useful in systems with a fairly
 *        static heap configuration. In systems where heaps are more dynamic
 *        and shares more of the free memory in the MM, the other threshold values
 *        (low extension, large alloc, etc) are more suited to use in order to
 *        implement low watermark functionality. The HEAP_EARLY_NULL_LEVEL
 *        feature does NOT affect the other allocation functions
 *        (i.e. not either the malloc function.)
 *
 * @seealso heap_alloc_early_null
 * @seealso ose_heap_realloc_watermark
 * @seealso HEAP_ABSOLUTE_WATERMARK
 */
#define HEAP_EARLY_NULL_LEVEL         "HEAP_EARLY_NULL_LEVEL"

/**
 * @macro HEAP_ABSOLUTE_WATERMARK
 *
 * @deprecated This environment variable is deprecated from OSE 5.5 and will
 *             not be supported in future releases of OSE. The variable has no
 *             effect. Watermark allocation methods remain functional
 *             but ignore the watermark level. They may therefore
 *             exhaust the heap.
 *
 * @brief Environment variable with an absolute watermark level used by
 *        priority allocation methods.
 *
 * @long  Instead of using the relative watermarks of HEAP_EARLY_NULL_LEVEL,
 *        an absolute watermark level can be used. An allocation on the
 *        lower priority level will not succeed if the amount of free memory
 *        that the heap could potentially use in the system is less than the
 *        provided absolute watermark level. The same interface are used as for
 *        the HEAP_EARLY_NULL_LEVEL mechanism.
 *
 * @restriction NOT YET AVAILABLE in standard OSE.
 *       @break There are some restrictions on the use. Only one absolute
 *              watermark level is allowed and only on one heap in the whole
 *              system. Absolute and relative watermarks can not be used on
 *              the same heap.
 */
#define HEAP_ABSOLUTE_WATERMARK         "HEAP_ABSOLUTE_WATERMARK"

/**
 * @macro HEAP_TENTATIVE_BUFCOUNT
 *
 * @brief Environment variable with the maximum number of tentative
 *        buffers allowed within a heap.
 *
 * @long  This variable controls the tentative buffer support. If configured,
 *        buffers allocated from the heap can be freed tentatively and later
 *        reclaimed with data retained. A tentatively freed buffer is
 *        garbage collected to accommodate other allocations as a last resort
 *        when the heap no longer can be extended.
 *
 * @long  The default value is 0, i.e. the functionality is disabled.
 *
 */
#define HEAP_TENTATIVE_BUFCOUNT       "HEAP_TENTATIVE_BUFCOUNT"

/**
 * @macro HEAP_CLEANUPD_STACK_SIZE
 * @brief Environment variable with the domain heap cleanup daemon
 *        stack size.
 * @long The heap mechanism creates one cleanup daemon process per program/domain,
 *       which lives together with the program. It is named "ose_heapcd_0x<domain>",
 *       and its stack size can be set via this variable if it is desired to fine-tune
 *       the system. The default is 2000 bytes, which is intended to satisfy
 *       all architectures/compilers. The stack need is deterministic and will
 *       not depend on the heap usage.
 */
#define HEAP_CLEANUPD_STACK_SIZE         "HEAP_CLEANUPD_STACK_SIZE"

/**
 * @macro HEAP_FRAG_CHECK
 *
 * @brief Environment variable which enables the in-fragments pointer
 *        validation (off by default).
 *
 * @long  Boolean, set to "True" to enable the behavior
 *        of validating the buffer pointer to be inside heap
 *        fragments. (Enabling it will give some performance degradation.)
 */
#define HEAP_FRAG_CHECK         "HEAP_FRAG_CHECK"

/**
 * @macro HEAP_RECLAIM
 *
 * @brief Environment variable which enables the heap to return
 *        large unused blocks of memory to the memory manager upon request.
 *
 * @long  Boolean, set to "True" to enable the behavior.
 */
#define HEAP_RECLAIM            "HEAP_RECLAIM"

/**
 * @macro HEAP_CACHE_MAX_BUFCOUNT
 *
 * @brief Environment variable that specifies how many buffers the heap
 *        may cache internally.
 *
 * @long  A non-zero value allows the heap to temporarily store buffers
 *        in a heap cache. The heap cache intercepts buffers that are freed
 *        before they reach the originating heap algorithm. An intercepted
 *        buffer is therefore not coalesced. Subsequent allocation requests
 *        can be directly satisfied by the cache if a suitable buffer has
 *        been previously intercepted. 
 *        
 * @long  In multicore systems, each core has its own cache. Protection
 *        against concurrent execution is therefore unnecessary and the cache
 *        can be used to reduce heap lock contention and increase performance.
 *        The cache is also useful in unicore systems if slower mutex locks
 *        are configured and used by the main algorithm. 
 *
 * @long  The cache organizes intercepted buffers into different buckets.
 *        Each bucket is a linked list holding buffers of a certain logarithmic
 *        size interval, e.g. 16-31 bytes, 32-63 bytes, 64-127 bytes, etc.
 *
 * @long  During an allocation request, the cache will search the list of the
 *        corresponding bucket for a buffer with large enough size. If
 *        no suitable buffer is found the cache will request a new
 *        buffer to be allocated from the main algorithm.
 *
 * @long  The maximum number of buffers the cache may store directly
 *        translates to the maximum length of the bucket lists. If the
 *        value is set to 10, an allocation may involve traversing 10
 *        nodes of the linked list, thereby adding latency to the allocation
 *        if no suitable buffer is found.
 *
 * @long  The parameter applies to each cache instance. If it is set to
 *        10 and there are 8 CPU cores the heap will be able to
 *        cache a total of 80 buffers (10 per core).
 */
#define HEAP_CACHE_MAX_BUFCOUNT "HEAP_CACHE_MAX_BUFCOUNT"

/**
 * @macro HEAP_CACHE_MAX_BUFSIZE
 *
 * @brief Environment variable with the upper limit for the size of
 *        buffers cached in the heap buffer cache.
 *
 * @long This parameter controls the behavior of the heap buffer cache.
 *       Buffers that are freed and are smaller than the size specified
 *       will be cached for fast retrieval during the next allocation.
 *       The default is 32 bytes.
 */
#define HEAP_CACHE_MAX_BUFSIZE  "HEAP_CACHE_MAX_BUFSIZE"


/**
 * @macro HEAP_ALGO_TLSF
 *
 * @brief Environment variable used to enable/disable the TLSF heap
 *        algorithm.
 *
 * @long  The variable is a Boolean. Set to "true" or "false" to enable
 *        or disable the algorithm.
 *
 * @long  TLSF is the default algorithm and it is chosen if no other
 *        algorithm has been configured. The algorithm may be explicitly
 *        disabled. If so, the resulting heap instance will only implement
 *        the mandatory large buffer algorithm, unless another optional
 *        algorithm is enabled.
 *
 * @long  It is illegal to explicitly configure more than one
 *        algorithm. The heap manager will deny creation if such a
 *        configuration is detected and ose_create_heap will return
 *        the invalid heap reference 0.
 *
 * @long  Read more about the TLSF algorithm in the "OSE Core User's
 *        Guide".
 *
 * @seealso HEAP_ALGO_BUDDY
 */
#define HEAP_ALGO_TLSF       "HEAP_ALGO_TLSF"


/**
 * @macro HEAP_ALGO_BUDDY
 *
 * @brief Environment variable used to enable/disable the Fibonacci
 *        buddy heap algorithm.
 *
 * @long  The variable is a Boolean. Set to "true" or "false" to enable
 *        or disable the algorithm.
 *
 * @long  The Fibonacci buddy algorithm is optional. If it is enabled,
 *        the default TLSF algorithm will be disabled since only one
 *        configurable algorithm may be present in a heap instance.
 *
 * @long  It is illegal to explicitly configure more than one
 *        algorithm. The heap manager will deny creation if such a
 *        configuration is detected and ose_create_heap will return
 *        the invalid heap reference 0.
 *
 * @long  Read more about the Fibonacci buddy algorithm in the "OSE
 *        Core User's Guide".
 *
 * @seealso HEAP_ALGO_TLSF
 */
#define HEAP_ALGO_BUDDY      "HEAP_ALGO_BUDDY"


/**
 * @macro HEAP_LOCK_TYPE
 *
 * @brief Environment variable that specifies the type of lock mechanism
 *        used by a heap.
 *
 * @long  The heap can be configured to use different types of locks to
 *        protect itself and its integrity from concurrent accesses that
 *        would otherwise corrupt heap data structure.
 *
 * @long  The types of locks supported have different characteristics
 *        and are suitable for different types of systems and applications.
 *        Available types are: @break
 *
 *        @link "" "HEAP_LOCK_INTERRUPT" @break
 *        @link "" "HEAP_LOCK_MUTEX"     @break
 *        @link "" "HEAP_LOCK_SPINLOCK"
 *
 * @example
 * PROCESS bid;
 * OSHEAP  href;
 *
 * bid = create_block("bl", (OSUSER)0, (OSBOOLEAN)0, (PROCESS)0, (OSBOOLEAN)0);
 *
 * set_env(bid, HEAP_LOCK_TYPE, HEAP_LOCK_MUTEX);
 *
 * href = ose_create_heap(bid, domain, (OSBOOLEAN)0);
 */
#define HEAP_LOCK_TYPE          "HEAP_LOCK_TYPE"

/**
 * @macro HEAP_LOCK_INTERRUPT
 *
 * @brief A heap lock type that disables interrupts to prevent
 *        concurrent access to a critical section of code.
 *
 * @long  This lock type is the default choice for all heaps created. It has
 *        the best performance of the choices available, but
 *        should not be used if clients to the created heap instance will be
 *        bound to multiple distinct CPU cores in a multicore system.
 */
#define HEAP_LOCK_INTERRUPT     "1"

/**
 * @macro HEAP_LOCK_SPINLOCK
 *
 * @brief A heap lock type that uses a combination of interrupt locking and
 *        spinlocks to prevent concurrent access to a critical section of code.
 *
 * @long  This lock type is suitable for multicore applications where clients
 *        to the created heap instance will be bound to multiple distinct CPU
 *        cores.
 *
 * @long  A large disadvantage of this lock type is that all processes on a
 *        CPU core may suffer from starvation if a single process on the
 *        core is spinning to acquire the lock currently held by a
 *        process running on a second core. The process on the first core
 *        will not yield until the lock has been acquired and released.  */
#define HEAP_LOCK_SPINLOCK     "2"

/**
 * @macro HEAP_LOCK_MUTEX
 *
 * @brief A heap lock type that uses a mutex to prevent concurrent access
 *        to a critical section of code.
 *
 * @long  This lock type is suitable for multicore applications where heap
 *        clients are bound to multiple CPU cores. The type can not be used
 *        if there are interrupt processes among the heap's clients.
 *
 * @long  An advantage of this lock type is that processes that attempt to
 *        acquire an already acquired lock will yield until the lock becomes
 *        available.
 *
 * @restriction
 *        Processes are not allowed to terminate during heap requests.
 *        An error, @link "heap_err.h" "OSE_HEAP_EDEADLOCK", will be
 *        generated if a process is killed when it holds the heap lock.
 */
#define HEAP_LOCK_MUTEX        "3"


/*
 * IMPORTANT:
 * The configuration variables below are under development and not fully tested,
 * and therefore not recommended to use in production systems.
 */

/*
 * If this variable is set to True, the behavior for the heap functions
 * malloc/free will be "shared ownership" mode instead of "private ownership" mode (default).
 * This variable substitutes the former "heap_shared_mode" call.
 * NOTE1: The heap "shared ownership" mode will be part of a "shared execution model"
 * from OSE5.1++. DO NOT USE HEAP_SHARED_MODE, it exist only for special customer cases.
 * NOTE2: This "shared ownership" mode is not the same as "shared global" mode,
 * which is officially supported (see "shared global" usage scenario above.).
 */
#define HEAP_SHARED_MODE         "HEAP_SHARED_MODE"

/*
 * Environment variable with the system low mark for small buffer
 * extensions.
 * The default is zero.
 */
#define HEAP_SMALL_LOW_LIMIT     "HEAP_SMALL_LOW_LIMIT"

/*
 * Environment variable with the system low mark for large buffer
 * allocations.
 * The default is zero.
 */
#define HEAP_LARGE_LOW_LIMIT     "HEAP_LARGE_LOW_LIMIT"

/*
 * Not implemented, do not use.
 */
#define HEAP_ANNOTATION_SIZE         "HEAP_ANNOTATION_SIZE"

/*
 * Environment variable with the sizeof endmark,
 * default 1, minimum 1, maximum 4.
 */
#define HEAP_ENDMARK_SIZE         "HEAP_ENDMARK_SIZE"

/*
 * Not implemented, do not use.
 */
#define HEAP_WIPE_ON_FREE         "HEAP_WIPE_ON_FREE"


/**
 * @macro HEAP_MALLOC_ERROR_ON_NULL
 *
 * @brief Environment variable for enabling/disabling error on null.
 *
 * @long  The variable is a boolean. If set to "true", the heap will
 *        call error, reporting @link "" "OSE_HEAP_EHEAP_EXHAUSTED"
 *        before returning a null pointer due to failure to allocate
 *        memory. This applies not only to malloc but to all allocation
 *        and reallocation calls.
 *
 * @long  The default value is false.
 *
 * @long  The main purpose of this feature is to provide backwards
 *        compatibility with older heap implementations. This flag is
 *        only recommended for heaps used with applications that do
 *        not check the return value from allocation calls for null.
 *
 * @long  Note that there is a requirement on the error handler to get
 *        the legacy behavior: The error handler must not return a non-zero
 *        value. Otherwise the heap call will return a null pointer
 *        to the application, whereas the legacy behavior is that the
 *        heap call never returns after allocation failure.
 */
#define HEAP_MALLOC_ERROR_ON_NULL   "HEAP_MALLOC_ERROR_ON_NULL"


/**
 * @private
 * @long EXPERIMENTAL. Intentionally undocumented
 *
 * @macro HEAP_FILTER_FRAG_CACHE
 *
 * @brief Environment variable used to enable/disable the fragment cache.
 *
 * @long  The variable is a Boolean. Set to "true" or "false" to enable
 *        or disable the fragment cache.
 *
 * @long  The fragment cache is disabled by default. If enabled, unused
 *        heap fragments will not be deallocated until the memory manager
 *        requires the memory to satisfy an allocation request. The
 *        performance of the heap's system memory allocations will therefore
 *        improve.
 *
 * @long  Applications that utilize large buffers to a high degree can
 *        benefit from having the fragment cache enabled.
 *
 * @long  The fragment cache implements the deprecated manual heap extension
 *        API, heap_extend_heap and ose_extend_heap.
 */
#define HEAP_FILTER_FRAG_CACHE   "HEAP_FILTER_FRAG_CACHE"

/**
 * @private
 * @long EXPERIMENTAL. Intentionally undocumented
 *
 * @macro HEAP_FILTER_LOG
 *
 * @brief Environment variable used to enable/disable debug logging. 
 *
 * @long  The variable is a bitmask where different bits enable different
 *        logging.
 *
 * @long  Bit 0 - decimal value 1:
 *        Enables ramlog logging when a heap error is detected (default).@break
 *        Bit 1 - decimal value 2:
 *        Enables ramlog logging when a buffer is allocated.@break
 *        Bit 2 - decimal value 4:
 *        Enables ramlog logging when a buffer is freed.@break
 *        Bit 3 - decimal value 8:
 *        Enables ramlog logging when a buffer is resized.@break
 *        Bit 4 - decimal value 16:
 *        Enables ramlog logging of garbage collection start and
 *        end in response to process termination. 
 * 
 */
#define HEAP_FILTER_LOG           "HEAP_FILTER_LOG"

#ifdef __cplusplus
}
#endif

#endif
