/* COPYRIGHT-ENEA-SRC-R2 *
**************************************************************************
* Copyright (C) 2007 by Enea Software AB.
* All rights reserved.
*
* This Software is furnished under a software license agreement and
* may be used only in accordance with the terms of such agreement.
* Any other use or reproduction is prohibited. No title to and
* ownership of the Software is hereby transferred.
*
* PROPRIETARY NOTICE
* This Software consists of confidential information.
* Trade secret law and copyright law protect this Software.
* The above notice of copyright on this Software does not indicate
* any actual or intended publication of such Software.
**************************************************************************
* COPYRIGHT-END */

/**
 * @deprecated
 * @toc File_System:
 *
 * @file ose_spi/ose_service.h
 *
 * @brief Common interfaces for OSE services.
 *
 * @long  This file describes interfaces that can be implemented by any
 *        services. A service is an instantiated function interface, i.e. a
 *        function interface with an associated struct describing a state.
 *
 * @long  All interfaces implemented by a service will inherit from
 *        @link "" "OseBaseApi", and this interface must be implemented by
 *        all services. The OseBaseApi function pointers are
 *        first in all service interface pointer structs, to allow all
 *        interfaces to be cast to OseBaseApi.
 *        See @link "" "OseConfApi" for an example of a valid interface.
 *        All registered services must also support the
 *        @link "ose_spi/ose_service_registry.h" "OseServiceApi" interface.
 *
 * @long  A service is expected to be reference counted and shall keep
 *        track of the reference count. When the reference
 *        count reaches zero, the service shall destroy itself.
 *
 * @long  For information about how to register a service into the service
 *        registry, see the ose_service_registry.h with the function
 *        @link "ose_spi/ose_service_registry.h" "ose_service_register".
 *
 * @restriction NOT YET AVAILABLE in standard OSE.
 *
 * @seealso ose_spi/ose_service_registry.h
 */

#ifndef _OSE_SERVICE_H
#define _OSE_SERVICE_H

#include "stdint.h"
#include "stddef.h"
#include "sys/types.h"
#include "ose.h"

#ifdef __cplusplus
extern "C" {
#endif

struct OseBaseApi;

/******************************************************************************
 * Reference count functions
 ******************************************************************************
 */

/**
 * @function ose_service_reference_add
 *
 * @brief Add a reference to a service instance, incrementing it's reference
 *        count.
 *
 * @long  This function must be used by the client instead of
 *        @link "" "OseBaseApiReferenceAdd". It will call this function on
 *        behalf of the client.
 *
 * @long  The reason why this function and the companion function
 *        @link "" "ose_service_reference_release" are used is for the special
 *        case where a service, registered into the service registry, is to be
 *        destroyed (count reaches zero), and the service is implemented in
 *        a program context. This function gives a good execution context in
 *        which the service registry can receive the destroy indication and
 *        unlock the program, to allow it to be unloaded from memory.
 *
 * @input api        Reference to service interface.
 * @input instance   Reference to service instance data.
 */
void
ose_service_reference_add(const struct OseBaseApi *api,
                          void *instance);


/**
 * @function ose_service_reference_release
 *
 * @brief Release a reference to a service instance, decrementing it's
 *        reference count.
 *
 * @long  This function must be used by the client instead of
 *        @link "" "OseBaseApiReferenceRelease". It will call this function on
 *        behalf of the client.
 *
 * @long  The reason why this function and the companion function
 *        @link "" "ose_service_reference_add" are used is for the special
 *        case where a service, registered into the service registry, is to be
 *        destroyed (count reaches zero), and the service is implemented in
 *        a program context. This function gives a good execution context in
 *        which the service registry can receive the destroy indication and
 *        unlock the program, to allow it to be unloaded from memory.
 *
 * @input api       Reference to service interface.
 * @input instance  Reference to service instance data.
 */
void
ose_service_reference_release(const struct OseBaseApi *api,
                              void *instance);


/******************************************************************************
 * Base class members
 ******************************************************************************
 */

/**
 * @type  OseBaseApiInterfaceQuery
 *
 * @brief Get another named interface, if supported by the indicated service
 *        instance.
 *
 * @long  This function is available in all OSE service interfaces. This makes
 *        it possible to get other interfaces supported by a given service
 *        instance via any interface for this instance.
 *
 * @long  The function returns the requested interface as a pointer to
 *        the interface struct, containing the function pointers to all
 *        functions in the interface, as supported by this service instance.
 *
 * @long  Interface names should be unique and standard name strings for
 *        services included in OSE are usually on the form:
 *        "com.enea.s.z" where s is a descriptive string and z is a version
 *        string. Some standard interface name strings are given in the header
 *        files as macros named *API_NAME.
 *
 * @input  interfaceName  Name of interface. This is a null terminated
 *                        ASCII string.
 * @input  instance       Service instance handle.
 * 
 * @return
 *   @tab Pointer to the interface @tab if interface is supported. @break
 *   @tab Null pointer             @tab if interface is not supported. @break
 *
 * @seealso ose_service_open
 */
typedef const void *(OseBaseApiInterfaceQuery)(
   void *instance, const char *interfaceName);


/**
 * @type  OseBaseApiGetMaxStack
 *
 * @brief Get maximum stack usage for the OSE service.
 *
 * @long  This shall be the worst case amount of free stack space on the
 *        callers stack that is needed when any call is issued to this service.
 *
 * @long  This function is available in all OSE service interfaces.
 *
 * @input instance       Service instance handle.
 *
 * @return Maximum stack usage in bytes.
 */
typedef size_t (OseBaseApiGetMaxStack)(void *instance);


/**
 * @type  OseBaseApiReferenceAdd
 *
 * @brief Increment the reference count of this service instance.
 *
 * @long  NOTE: A client must never call this function directly! Use
 *              @link "" "ose_service_reference_add" instead.
 *
 * @long  This function is available in all OSE service interfaces.
 *
 * @input  instance  Service instance handle.
 *
 * @return Nothing.
 */
typedef void (OseBaseApiReferenceAdd)(void *instance);


/**
 * @type  OseBaseApiReferenceRelease
 *
 * @brief Decrement the reference count of this service instance,
 *        destroying it if the count reaches zero.
 *
 * @long  NOTE: A client must never call this function directly! Use
 *              @link "" "ose_service_reference_release" instead.
 *
 * @long  If the count reaches zero, the service shall destroy itself, i.e.
 *        clean up all resources and return a pointer to it's OseServiceApi.
 *        This return value is a destroy indication and is used by the OSE
 *        registry to perform any other needed cleanup, e.g. unlocking any
 *        unload-lock for a program, if the service is in a program.
 *
 * @long  This function is available in all OSE service interfaces.
 *
 * @input  instance  Service instance handle.
 *
 * @return @tab OseServiceApi interface @tab if reference count became 0 and
 *                  service has been registered into service registry. @break
 *         @tab 0 @tab for all other cases @break
 */
typedef const struct OseServiceApi *(OseBaseApiReferenceRelease)(
   void *instance);


/**
 * @type  OseBaseApi
 *
 * @brief OSE service base interface, leading part of all service interfaces.
 *
 * @long  This interface is inherited by all OSE service interfaces, i.e. a
 *        struct of this type shall be placed first in all service interface
 *        structs. This means that any OSE service interface can be cast to
 *        this interface. See @link "" "OseConfApi" for an example of a
 *        service interface struct.
 *
 * @field b_interface_query   See @link "" "OseBaseApiInterfaceQuery"
 * @field b_get_max_stack     See @link "" "OseBaseApiGetMaxStack"
 * @field b_reference_add     See @link "" "OseBaseApiReferenceAdd" -
 *                            not used by clients.
 * @field b_reference_release See @link "" "OseBaseApiReferenceRelease" -
 *                            not used by clients.
 */
struct OseBaseApi
{
      OseBaseApiInterfaceQuery     *b_interface_query;
      OseBaseApiGetMaxStack        *b_get_max_stack;
      OseBaseApiReferenceAdd       *b_reference_add;
      OseBaseApiReferenceRelease   *b_reference_release;
};


/******************************************************************************
 * Configuration interface
 ******************************************************************************
 */


/**
 * @type OseConfConfigure
 *
 * @brief Configure an OSE service and/or create/start a new service instance.
 *
 * @long  If a new service instance was created as a result of the 
 *        configuration, the new instance shall be registered in the 
 *        OSE service registry.
 *
 * @long  Configuration of a service might create a new service instance.
 *        By suppling non-NULL pointers in inout_created_instance and
 *        out_create_api, this function can return the pointer to the instance
 *        object and the pointer to the instance object's base OseBaseApi
 *        interface.
 *        If no new service was created when inout_created_instance and 
 *        out_create_api are non-NULL, NULL values will be stored.
 *
 * @long  This function is a part of the @link "" "OseConfApi", which can
 *        be supported by services.
 * 
 * @input  instance      Instance handle of the service to be configured.
 * @input  argc          Number of arguments in the argv[] parameter
 *                       configuration for the service. 
 * @input  argv          Text strings pointers that are interpreted by 
 *                       the service, e.g. argv[0] through to argv[argc-1] are 
 *                       pointers to strings whose meaning will be determined 
 *                       by the service usually as configuration settings.
 *                       If a sub command should be passed in the string list 
 *                       argv[0] is usually used for sub commands.
 * @input  inout_created_instance If non-NULL and out_create_api is non-NULL,
 *                       pointer to where a service instance pointer should be
 *                       stored.
 * @output out_created_api If non-NULL and inout_created_instance is non-NULL,
 *                       a pointer to the @link "" "OseBaseApi" interface of
 *                       the newly created service is stored at this address.
 *                       NULL will be stored if no service was created.
 * @output inout_created_instance If non-NULL and out_create_api is non-NULL,
 *                       pointer to the instance object of the newly created
 *                       service is stored at this address.
 *                       NULL will be stored if no service was created.
 *                       Note that if a pointer to an instance object is
 *                       returned, it will be reference counted. When the
 *                       caller is finished with the object, it must be
 *                       released using @link ""
 *                       "ose_service_reference_release".
 *
 * @return
 *   @tab 0 @tab Success. @break
 *   @tab -EINVAL @tab Invalid configuration. @break
 *   @tab -ENOMEM @tab Cannot allocate memory. @break
 *   @tab -errno-code @tab Suitable error codes defined in errno.h, this might
 *                     be passed all the way to user. @break
 *   
 *   
 */

typedef int (OseConfConfigure)(void *instance,
                               int argc,
                               const char *argv[],
                               const struct OseBaseApi **out_created_api,
                               void **inout_created_instance);


/**
 * @macro OSE_CONF_API_NAME
 *
 * @brief Name of the @link "" "OseConfApi"
 *        OSE service configuration interface.
 *
 * @long  This is the name that shall be used in call to
 *        @link "" "OseBaseApiInterfaceQuery" to get interface
 *        @link "" "OseConfApi".
 *
 * @long  The name is on the form "com.enea.ose_conf_api.z", where z is
 *        a version string which can vary.
 *
 * @seealso OseBaseApiInterfaceQuery
 */
#define OSE_CONF_API_NAME ("com.enea.ose_conf_api.v1")

/* Deprecated old name. Will be removed without further notice. */
#define API_NAME_OSE_CONF_API OSE_CONF_API_NAME


/**
 * @type  OseConfApi
 *
 * @brief OSE service configuration interface.
 *
 * @field cfg_base        See @link "" "OseBaseApi"
 * @field cfg_configure   See @link "" "OseConfConfigure"
 *
 * @seealso OSE_CONF_API_NAME
 */
struct OseConfApi
{
      struct OseBaseApi    cfg_base;

      OseConfConfigure *cfg_configure;
};


#ifdef __cplusplus
} /* end of extern "C" */
#endif

#endif /* _OSE_SERVICE_H */
