/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2006 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */
/**
 * @public
 * @toc OSE:
 * @file ose_spi/ose_smp.h
 *
 * @brief     OSE kernel System Programming Interface (SPI),
 *            for developers of SMP (Symmetric Multi Processing)
 *            platforms.
 * @long      This file contains declarations to be included in code
 *            written in C.
 */

#ifndef _OSE_SMP_H
#define _OSE_SMP_H

#include "ose.h"
#include "stdint.h"

/**
 * @macro OSVECTOR_INT_KRN_EVT
 *
 * @brief Indicates that a kernel event is the source of an interrupt.
 * @long The interrupt vector handler of a device handling kernel event
 * interrupts shall return this magic interrupt vector when detecting that the
 * line corresponding to a kernel event was the source of an interrupt.
 */
#define OSVECTOR_INT_KRN_EVT 0xFF01

#ifdef __cplusplus
extern "C" {
#endif

/**
 * @macro OSE_KRN_EVT_MSG_ENTRY_LEN
 *
 * @brief Indicates the max message length of a kernel event.
 */
#define OSE_KRN_EVT_MSG_ENTRY_LEN 2

/**
 * @type OSEVTNOTIFIER
 *
 * @brief EXPERIMENTAL. A function pointer returned by ose_smp_get_evt_notifier.
 *
 * @seealso ose_smp_get_evt_notifier
 */
typedef void (OSEVTNOTIFIER)(void);

/**
 * @type OSEVTINTERRUPTER
 *
 * @brief EXPERIMENTAL. A function pointer used in ose_smp_set_evt_interrupter.
 *
 * @seealso ose_smp_set_evt_interrupter
 */
typedef void (OSEVTINTERRUPTER)(cpuid_t);

/**
 * @function ose_smp_start_other_cpus
 * @include "ose_spi/ose_smp.h"
 *
 * @brief Start CPUs in an SMP system.
 * @long
 * This function is called by the master CPU (cpu0).
 * An SMP system typically starts with only one CPU active while the rest are
 * held in reset state. When the system has reached the point in the start
 * sequence when it is time to start all other CPUs, this function should
 * be called.
 * @long
 * Implemented by the kernel, typically called in an OSE start hook
 * or from the BSP.
 * @restriction
 *
 * @seealso ose_smp_boot_cpu
 */
void
ose_smp_start_other_cpus(void);

/**
 * @function ose_smp_boot_cpu
 * @include "ose_spi/ose_smp.h"
 *
 * @brief Boot a CPU in an SMP system.
 * @long
 * This function is called by the master CPU (cpu0). The specified CPU should
 * be started (released from reset state). There must be some initial code for
 * the started CPU that picks up the supplied stack, sets up the per CPU
 * register and jumps to the ose_smp_cpu_callin entrypoint in the kernel.
 * @long
 * Implemented by the BSP, called by the kernel.
 * @input cpuid The ID of the CPU to boot.
 * @input stack_base Base address of the initial stack.
 * @input stack_size Size of the initial stack.
 * @restriction
 *
 * @seealso ose_smp_cpu_callin
 */
void
ose_smp_boot_cpu(cpuid_t cpuid, void *stack_base, uint32_t stack_size);

/**
 * @function ose_smp_cpu_callin
 * @include "ose_spi/ose_smp.h"
 *
 * @brief Function that is called when a started CPU enters the kernel.
 * @long
 * This function is not called by the master CPU (cpu0).
 * When an additional CPU is started in an SMP system by the master CPU,
 * it should call this function to make the CPU available to the kernel to
 * start scheduling processes.
 * ose_smp_cpu_callin initializes the local CPU registers and the MMU before
 * starting scheduling processes.
 * @long
 * Implemented by the kernel. Called by the BSP in the other CPU after
 * ose_smp_boot_cpu has started the other CPU. It is called from the
 * initialization code executing in the other CPU.
 *
 * @restriction
 *
 * @seealso ose_smp_boot_cpu
 */
void
ose_smp_cpu_callin(void);

/**
 * @function ose_smp_init_other_cpu
 * @include "ose_spi/ose_smp.h"
 *
 * @brief Function that is called when a started CPU enters the kernel.
 * @long
 * This function is not called by the master CPU (cpu0).
 * When an additional CPU is started in an SMP system, by the master CPU,
 * it should call this function in order for the BSP to set up
 * the local CPU and MMU registers.
 * @long
 * Implemented by the BSP, called by the kernel.
 * @input cpuid The ID of the CPU to init.
 * @restriction
 *
 * @seealso ose_smp_cpu_callin
 */
void
ose_smp_init_other_cpu(cpuid_t cpuid);

/**
 * @function ose_smp_intercept_other_cpus
 * @include "ose_spi/ose_smp.h"
 *
 * @brief Temporarily park all other cpus in a multicore system.
 *
 * @long Intended for use by the kernel error handler
 * in an SMP system. Performs no action in single CPU systems.
 * It may be used for example to generate an atomic PMD dump.
 * An intercepted cpu waits, with disabled interrupts, for
 * a matching call of ose_smp_resume_other_cpus. 
 *
 * @seealso ose_smp_resume_other_cpus
 */
void
ose_smp_intercept_other_cpus(void);

/**
 * @function ose_smp_resume_other_cpus
 * @include "ose_spi/ose_smp.h"
 *
 * @brief Release all other cpus that were parked by a
 * ose_smp_intercept_other_cpus call.
 *
 * @long Intended for use by the kernel error handler
 * in an SMP system. Performs no action in single CPU systems.
 *
 * @seealso ose_smp_intercept_other_cpus
 */
void
ose_smp_resume_other_cpus(void);

/**
 * @function ose_smp_get_evt_notifier
 * @include "ose_spi/ose_smp.h"
 *
 * @brief EXPERIMENTAL. Get a function pointer from the kernel to call when
 * receiving an ipi.
 *
 * @long Called by the BSP during initialization.
 *
 * @return A function pointer of the type OSEVTNOTIFIER.
 *
 */
OSEVTNOTIFIER *
ose_smp_get_evt_notifier(void);

/**
 * @function ose_smp_set_evt_interrupter
 * @include "ose_spi/ose_smp.h"
 *
 * @brief EXPERIMENTAL. Install a function in the kernel to be called to
 * generate an ipi.
 *
 * @long Called by the BSP during initialization.
 *
 * @input func A function pointer of the type OSEVTINTERRUPTER.
 *
 * @return True value for success.
 */
OSBOOLEAN
ose_smp_set_evt_interrupter(OSEVTINTERRUPTER *func);

#ifdef __cplusplus
}
#endif

#endif /* _OSE_SMP_H */
