/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2005-2006 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */
/**
 * @public
 *
 * @toc Program_Management:
 *
 * @file ose_spi/pfmi.h
 *
 * @brief Program manager Freeze-mode Interface
 *
 * @long
 * This file declares the Program manager Freeze-mode Interface.
 * It may be used to fetch information from the program manager (PM) in
 * a freeze-mode environment. The purpose is to make Program concepts
 * accessible for freeze-mode debug monitors that can not use ordinary
 * system calls.
 * @long
 * These functions performs very few tests of the program manager data
 * structures. It may report failure if the referred load file or program
 * can not be found, is severely damaged or locked by the program manager
 * process. An obvious example is when the program manager is not yet
 * initiated or when the program manager is busy updating the internal data.
 * Interrupts should be disabled when this interface is used.
 * @long
 * WARNING: Use of PFMI functionality for any other purpose will severely
 * damage the interrupt latency!
 *
 */

#ifndef _PFMI_H
#define _PFMI_H


#include "stddef.h"
#include "ose.h"
#include "stdint.h"
#include "pm.sig"
#include "pm.h"
#include "heapapi.h"

#ifdef __cplusplus
extern "C" {
#endif

#if defined(HOST_TOOL_ENVIRONMENT)
#include "host_fmi.h"
#endif
   
#define PFMI_CURSOR_SIZE (16 / sizeof(uint32_t))


/* Return status of functions */
/**
 *  @type PfmiStatus
 *
 *  @brief enum PfmiStatus         Defines the set of return values from PFMI
 *                                 functions.
 *
 *  @field PFMI_OK                  Operation was successful.
 *  @field PFMI_UNAVAILABLE         Requested information not found.
 *  @field PFMI_END_OF_LIST         No more objects found.
 *  @field PFMI_RESULT_TRUNCATED    A list or string was truncated.
 *  @field PFMI_DATA_LOCKED         PM data structures are locked.
 *  @field PFMI_DATA_CORRUPTED      PM data structures are corrupted.
 *  @field PFMI_EPROGPID            Illegal program PID.
 *  @field PFMI_EINSTHND            Illegal install handle.
 *  @field PFMI_UNKNOWN             The information was unknown by PM.
 *  @field PFMI_CALL_INVALID        Operation preconditions were not met.
 **/
typedef enum PfmiStatus
{
  PFMI_OK                 = 0,
  PFMI_UNAVAILABLE        = 1,
  PFMI_END_OF_LIST        = 2,
  PFMI_RESULT_TRUNCATED   = 3,
  PFMI_DATA_LOCKED        = 4,
  PFMI_DATA_CORRUPTED     = 5,
  PFMI_EPROGPID           = 6,
  PFMI_EINSTHND           = 7,
  PFMI_UNKNOWN            = 8,
  PFMI_CALL_INVALID       = 9
} PfmiStatus;

/**
 *  @type PfmiLoadModuleInfo
 *
 *  @brief struct PfmiLoadModuleInfo contains information for one load module.
 *
 *  @field install_handle   The handle specifying the installed load module
 *  @field options          The installation options of the
 *                          load module (
 *                          @link "pm.sig" "PM_LOAD_MODULE_PERSISTENT"
 *                          and @link "pm.sig" "PM_LOAD_MODULE_ABSOLUTE").
 *  @field entrypoint       The entry point of the load module.
 *  @field text_base        The base address of the text region.
 *  @field text_size        The size of the text region (bytes).
 *  @field data_base        The base address of the data region.
 *  @field data_size        The size of the data region (bytes).
 *  @field no_of_sections   The number of section in the load module.
 *  @field no_of_instances  The number of instances currently created from the
 *                          load module.
 *                          NOTE: The number of instances is increased every
 *                                time a program is created from the load
 *                                module and every time a dynamic library
 *                                (created from the load module) is imported.
 *                                The number is decreased when a program
 *                                (created from the load module) is killed or
 *                                when a program, that previously imported a
 *                                library created from the load module,
 *                                is killed.
 *  @field  file_name       The parent file name (including the absolute path).
 *  @field  file_format     The file format of the load module.
 *
 **/
struct PfmiLoadModuleInfo
{
  char        install_handle[PM_INSTALL_HANDLE_LENGTH];
  OSADDRESS   entrypoint;
  uint32_t    options;
  OSADDRESS   text_base;
  uint32_t    text_size;
  OSADDRESS   data_base;
  uint32_t    data_size;
  uint32_t    no_of_sections;
  uint32_t    no_of_instances;
  char        file_format[PM_FILE_FORMAT_NAME_LENGTH];
  char        file_name[PM_PARENT_FILE_NAME_SIZE];
};

/**
 * @type PfmiLoadModuleSectionInfo
 *
 * @brief struct PfmiLoadModuleSectionInfo contains information for
 * one load module section
 *
 * @long This struct holds information about one single ELF section.
 * @field section_base    The base address of the section.
 * @field section_size    The size of the section (bytes).
 * @field section_attr    The attributes of the section (file format specific,
 *                        for ELF compatible load modules the section_attr
 *                        correlates with the contents of sh_flags of
 *                        Elf32_Shdr).
 * @field section_options These options describe whether the section is part
 *                        of a loaded segment or not. Possible values are:
 *                        @link "pm.sig" "PM_SECTION_LOAD", The section is
 *                        part of a loaded segment.
 *                        @link "pm.sig" "PM_SECTION_NO_LOAD", The section
 *                        is not part of a loaded segment.
 * @field section_name    The name of the section.
 */
struct PfmiLoadModuleSectionInfo
{
  OSADDRESS   section_base;
  uint32_t    section_size;
  uint32_t    section_attr;
  uint32_t    section_options;
  char        section_name[PM_SECTION_NAME_LENGTH];
};


/**
 *  @type PfmiProgramInfo
 *
 *  @brief struct PfmiProgramInfo contains information for one
 *  program created by the program manager.
 *  @field install_handle  The handle referencing the installed load module
 *  @field progpid         The program PID of the program.
 *  @field domain          The domain of the program.
 *  @field segpid          The segment PID of the program.
 *  @field main_block      The main block PID of the program.
 *  @field main_process    The main process PID of the program.
 *  @field stk_poolid      The stack pool identity.
 *  @field stk_pool_base   The base of the programs stack pool.
 *  @field stk_pool_size   The size of the programs stack pool (bytes).
 *  @field sig_poolid      The signal pool identity.
 *  @field sig_pool_base   The base of the programs signal pool.
 *  @field sig_pool_size   The size of the programs signal pool (bytes).
 *  @field uid             Reserved, always 0.
 *  @field heap            The programs heap reference.
 *  @field state           The state of the program
 *                         @link "pm.sig" "PM_PROGRAM_CREATED",
 *                         the program is created but not started.
 *                         @link "pm.sig" "PM_PROGRAM_STARTED",
 *                         the program is started.
 **/
struct PfmiProgramInfo
{
  char        install_handle[PM_INSTALL_HANDLE_LENGTH];
  PROCESS     progpid;
  OSDOMAIN    domain;
  PROCESS     segpid;
  PROCESS     main_block;
  PROCESS     main_process;
  PROCESS     stk_poolid;
  OSADDRESS   stk_pool_base;
  uint32_t    stk_pool_size;
  PROCESS     sig_poolid;
  OSADDRESS   sig_pool_base;
  uint32_t    sig_pool_size;
  OSUSER      uid;
  OSHEAP      heap;
  uint32_t    state;
};


/**
 *  @type PfmiLoadModuleCursor
 *
 *  @brief PFMI Cursor data types
 *  @long Variables of struct types PfmiProgramCursor, PfmiLoadModuleCursor,
 *  are provided by the caller of PFMI functions.
 *  They are used by PFMI for internal book keeping.
 *  @field cursor For internal use by PFMI.
 **/
struct PfmiLoadModuleCursor     { uint32_t cursor[PFMI_CURSOR_SIZE]; };
struct PfmiProgramCursor        { uint32_t cursor[PFMI_CURSOR_SIZE]; };
struct PfmiProgPidCursor        { uint32_t cursor[PFMI_CURSOR_SIZE + 12]; };


/**
 *  @function ose_pfmi_get_progpid
 *
 *  @brief Get the program PID from a process, block or segment PID.
 *
 *  @long The returned progpid represents a program
 *  created and managed by this program manager (PM).
 *  Use ose_pfmi_get_program_info to retrieve information about the program.
 *
 *  @long This function is only available in the kernel image.
 *  It must be called in supervisor mode.
 *  @long PFMI functions are designed for freeze-mode debug purposes.
 *  Any other use will destroy the interrupt latency of the system.
 *  PFMI functions are not part of the real-time services.
 *
 *  @input pid                The process, block or segment PID.
 *  @output progpid           The program PID.
 *  @return                   See enum PfmiStatus.
 *  @error                    See enum PfmiStatus.
 **/
#ifdef HOST_TOOL_ENVIRONMENT
PfmiStatus
ose_pfmi_get_progpid(HostFmiInstance *i,
		     PROCESS pid,
		     PROCESS *progpid);
#else
PfmiStatus
ose_pfmi_get_progpid(PROCESS pid,
                     PROCESS *progpid);
#endif

/**
 *  @function ose_pfmi_get_load_module_info
 *
 *  @brief Get information for one load module.
 *
 *  @long This function is only available in the kernel image.
 *  It must be called in supervisor mode.
 *  @long PFMI functions are designed for freeze-mode debug purposes.
 *  Any other use will destroy the interrupt latency of the system.
 *  PFMI functions are not part of the real-time services.
 *
 *  @input install_handle     The install handle.
 *  @output lmi               Load module information.
 *  @return                   See enum PfmiStatus.
 *  @error                    See enum PfmiStatus.
 **/
#ifdef HOST_TOOL_ENVIRONMENT
PfmiStatus
ose_pfmi_get_load_module_info(HostFmiInstance *i,
			      const char *install_handle,
			      struct PfmiLoadModuleInfo *lmi);
#else
PfmiStatus
ose_pfmi_get_load_module_info(const char *install_handle,
                              struct PfmiLoadModuleInfo *lmi);
#endif

/**
 *  @function ose_pfmi_get_load_module_section_info
 *
 *  @brief Get information for one load module section.
 *
 *  @long This function is only available in the kernel image.
 *  It must be called in supervisor mode.
 *  @long PFMI functions are designed for freeze-mode debug purposes.
 *  Any other use will destroy the interrupt latency of the system.
 *  PFMI functions are not part of the real-time services.
 *
 *  @input install_handle     The install handle.
 *  @input section            Load module section number
 *                            (0 to no_of_sections-1 in PfmiLoadModuleInfo).
 *  @output lmsi              Load module section information.
 *  @return                   See enum PfmiStatus.
 *  @error                    See enum PfmiStatus.
 **/
#ifdef HOST_TOOL_ENVIRONMENT
PfmiStatus
ose_pfmi_get_load_module_section_info(HostFmiInstance *i,
				      const char *install_handle,
				      uint32_t section,
				      struct PfmiLoadModuleSectionInfo *lmsi);
#else
PfmiStatus
ose_pfmi_get_load_module_section_info(const char *install_handle,
                                      uint32_t section,
                                      struct PfmiLoadModuleSectionInfo *lmsi);
#endif

/**
 *  @function ose_pfmi_get_program_info
 *
 *  @brief Get information for a program created by the program manager.
 *
 *  @long An error will be returned if a progpid created by another
 *  program manager than PM is used.
 *
 *  @long This function is only available in the kernel image.
 *  It must be called in supervisor mode.
 *  @long PFMI functions are designed for freeze-mode debug purposes.
 *  Any other use will destroy the interrupt latency of the system.
 *  PFMI functions are not part of the real-time services.
 *
 *  @input progpid	      The program identity. For historical
 *                        reasons a process, block or segment PID
 *                        associated with the segment the program was
 *                        created in may also be used.
 *  @output pi	          Program information.
 *  @return               See enum PfmiStatus.
 *  @error                See enum PfmiStatus.
 **/
#ifdef HOST_TOOL_ENVIRONMENT
PfmiStatus
ose_pfmi_get_program_info(HostFmiInstance *i,
			  PROCESS progpid,
			  struct PfmiProgramInfo *pi);
#else
PfmiStatus
ose_pfmi_get_program_info(PROCESS progpid,
                          struct PfmiProgramInfo *pi);

#endif


/* Retrieve the load module list of the system
 */
/**
 *  @function ose_pfmi_get_first_load_module
 *
 *  @brief Get information on the first load module in the load module list.
 *
 *  @long This function is only available in the kernel image.
 *  It must be called in supervisor mode.
 *  @long PFMI functions are designed for freeze-mode debug purposes.
 *  Any other use will destroy the interrupt latency of the system.
 *  PFMI functions are not part of the real-time services.
 *
 *  @input  lmc	          PFMI cursor for internal use.
 *  @output lmi		      Load module information.
 *  @return               See enum PfmiStatus.
 *  @error                See enum PfmiStatus.
 *  @seealso              ose_pfmi_get_next_load_module
 *  @example              See ose_pfmi_get_first_program.
 **/
#ifdef HOST_TOOL_ENVIRONMENT
PfmiStatus
ose_pfmi_get_first_load_module(HostFmiInstance *i,
			       struct PfmiLoadModuleCursor *lmc,
			       struct PfmiLoadModuleInfo *lmi);
#else
PfmiStatus
ose_pfmi_get_first_load_module(struct PfmiLoadModuleCursor *lmc,
                               struct PfmiLoadModuleInfo *lmi);

#endif

/**
 *  @function ose_pfmi_get_next_load_module
 *
 *  @brief Get information on the next load module in the load module list.
 *
 *  @long This function is only available in the kernel image.
 *  It must be called in supervisor mode.
 *  @long PFMI functions are designed for freeze-mode debug purposes.
 *  Any other use will destroy the interrupt latency of the system.
 *  PFMI functions are not part of the real-time services.
 *
 *  @input  lmc		      PFMI cursor for internal use.
 *  @output lmi		      Load module information.
 *  @return               See enum PfmiStatus.
 *  @error                See enum PfmiStatus.
 *  @seealso              ose_pfmi_get_first_load_module
 *  @example              See ose_pfmi_get_first_program.
 **/
#ifdef HOST_TOOL_ENVIRONMENT
PfmiStatus
ose_pfmi_get_next_load_module(HostFmiInstance *i,
			      struct PfmiLoadModuleCursor *lmc,
			      struct PfmiLoadModuleInfo *lmi);
#else
PfmiStatus
ose_pfmi_get_next_load_module(struct PfmiLoadModuleCursor *lmc,
                              struct PfmiLoadModuleInfo *lmi);
#endif

/* Retrieve the list of programs created by the program loader (PM).
 */
/**
 *  @function ose_pfmi_get_first_program
 *
 *  @brief Get information on the first program (created and managed by PM)
 *  in the program list.
 *
 *  @long This function is only available in the kernel image.
 *  It must be called in supervisor mode.
 *  @long PFMI functions are designed for freeze-mode debug purposes.
 *  Any other use will destroy the interrupt latency of the system.
 *  PFMI functions are not part of the real-time services.
 *
 *  @input  pc		      PFMI cursor for internal use.
 *  @output pi		      Program information.
 *  @return               See enum PfmiStatus.
 *  @error                See enum PfmiStatus.
 *  @seealso              ose_pfmi_get_next_program
 *  @example
 *  struct PfmiProgramCursor pc;
 *  struct PfmiProgramInfo   pi;
 *  PfmiStatus rv;
 *
 *  rv = ose_pfmi_get_first_program(&pc, &pi);
 *
 *  while(rv == PFMI_OK)
 *  {
 *      do_something_with_program_info(&pi);
 *      rv = ose_pfmi_get_next_program(&pc, &pi);
 *  }
 *
 *  if(rv != PFMI_END_OF_LIST)
 *  {
 *     printf("Something went wrong when retrieving program info.\n");
 *  }
 **/
#ifdef HOST_TOOL_ENVIRONMENT
PfmiStatus
ose_pfmi_get_first_program(HostFmiInstance *i,
			   struct PfmiProgramCursor *pc,
			   struct PfmiProgramInfo   *pi);
#else
PfmiStatus
ose_pfmi_get_first_program(struct PfmiProgramCursor *pc,
                           struct PfmiProgramInfo   *pi);
#endif

/**
 *  @function ose_pfmi_get_next_program
 *
 *  @brief Get information on the next program in the program list
 *  managed by PM.
 *
 *  @long This function is only available in the kernel image.
 *  It must be called in supervisor mode.
 *  @long PFMI functions are designed for freeze-mode debug purposes.
 *  Any other use will destroy the interrupt latency of the system.
 *  PFMI functions are not part of the real-time services.
 *
 *  @input  pc		      PFMI cursor for internal use.
 *  @output pi		      Program information.
 *  @return               See enum PfmiStatus.
 *  @error                See enum PfmiStatus.
 *  @seealso              ose_pfmi_get_first_program
 *  @example              See ose_pfmi_get_first_program
 **/
#ifdef HOST_TOOL_ENVIRONMENT
PfmiStatus
ose_pfmi_get_next_program(HostFmiInstance *i,
			  struct PfmiProgramCursor *pc,
			  struct PfmiProgramInfo   *pi);
#else
PfmiStatus
ose_pfmi_get_next_program(struct PfmiProgramCursor *pc,
                          struct PfmiProgramInfo   *pi);
#endif

#ifdef __cplusplus
}
#endif

#endif /* _PFMI_H */
