/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2006 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */
/**
 * @public
 * @toc Memory_Management:
 * @file ose_spi/pmm.sig
 * @version
 *	      All rights reserved.
 *
 * @brief Public signal interface to Persistent Memory Manager (PMM).
 *
 * @long Signals should be sent to the process named ose_mm_pmm.
 * @library core
 */

#ifndef _PMM_SIG
#define _PMM_SIG

#include "ose.h"
#include "osetypes.h"
#include "mmtypes.h"

#ifdef __cplusplus
extern "C" {
#endif

/* Checks for the matching version of mmtypes.h */
#if _MMTYPES_H != 500
#error ==========================================================
#error || Wrong version of mmtypes.h                           ||
#error ==========================================================
#endif

/*
 ****************************************************************************
 * Constants.
 ****************************************************************************
 */

/* The signal space reserved for PMM is 36000 - 36099. */

/**
 * @macro PMM_SIG_VERSION
 *
 * @brief Interface version.
 *
 * @long Incremented for interface changes.
 *
 * @seealso PMM_VERSION
 */
#define PMM_SIG_VERSION 3

/**
 * @macro PMM_MODE_SAS
 *
 * @brief SAS configuration value (only for FLASH).
 *
 * @long Configures PMM to work in SAS mode.
 *
 * @long PMM is configured to work in either SAS or SASE mode.
 *
 * @seealso PMM_MODE_SASE
 * @seealso PMM_INIT
 */
#define PMM_MODE_SAS 1

/**
 * @macro PMM_START_OPTION_FORCE_COLD_START
 *
 * @brief Used for PMM_START_REQUEST->options.
 *
 * @long Forces PMM to perform a cold start.
 *
 */
#define PMM_START_OPTION_FORCE_COLD_START 1

/**
 * @macro PMM_START_WARM
 *
 * @brief Used for PMM_START_REPLY->status.
 *
 * @long Indicates that PMM will attempt a warm start.
 *
 * @seealso PMM_START
 * @seealso PMM_START_COLD
 */
#define PMM_START_WARM 2

/**
 * @macro PMM_START_COLD
 *
 * @brief Used for PMM_START_REPLY->status.
 *
 * @long Indicates that PMM will not perform a warm start.
 *
 * @seealso PMM_START
 * @seealso PMM_START_WARM
 */
#define PMM_START_COLD 3

/**
 * @macro PMM_MODE_SASE
 *
 * @brief SASE configuration value (only for FLASH).
 *
 * @long Configures PMM to work in SASE mode.
 *
 * @long PMM is configured to work in either SAS or SASE mode.
 *
 * @seealso PMM_MODE_SAS
 * @seealso PMM_INIT
 */
#define PMM_MODE_SASE 2

/* Type of region for alloc signals. */

/**
 * @macro PMM_BOUND
 *
 * @brief Indicates that a persistent region has physical memory.
 *
 * @seealso PMM_UNBOUND
 * @seealso PMM_RESTORE_REGION
 */
#define PMM_BOUND 0xBBBB

/**
 * @macro PMM_UNBOUND
 *
 * @brief Indicates that a persistent region lacks physical memory.
 *
 * @seealso PMM_BOUND
 * @seealso PMM_RESTORE_REGION
 */
#define PMM_UNBOUND 0xAAAA

/**
 * @macro PMM_MAX_METALEN
 *
 * @brief The maximum length of the meta data.
 */
#define PMM_MAX_METALEN  128

/**
 * @macro PMM_EOK
 *
 * @brief The operation completed successfully.
 */
#define PMM_EOK            0


#define PMM_SIG_ERR_BASE   (0x00000000)

/**
 * @macro PMM_EOUT_OF_RANGE
 *
 * @brief Error: Address out of range.
 */
#define PMM_EOUT_OF_RANGE                            (0x00000001)

/**
 * @macro PMM_EOUT_OF_MEMORY
 *
 * @brief Error: Out of memory.
 */
#define PMM_EOUT_OF_MEMORY                           (0x00000003)

/**
 * @macro PMM_EALIGNMENT_BASE
 *
 * @brief Error: Misaligned base address.
 */
#define PMM_EALIGNMENT_BASE                          (0x00000004)

/**
 * @macro PMM_EALIGNMENT_SIZE
 *
 * @brief Error: Misaligned size.
 */
#define PMM_EALIGNMENT_SIZE                          (0x00000005)

/**
 * @macro PMM_EOPEN_FAILED_REGION_META_DATA_FILE
 *
 * @brief Error: Failed to open/read metadata file.
 */
#define PMM_EOPEN_FAILED_REGION_META_DATA_FILE       (0x00000006)

/**
 * @macro PMM_EOVERLAP_WITH_EXISTING_REGION
 *
 * @brief Error: Base address overlaps with an existing regions.
 */
#define PMM_EOVERLAP_WITH_EXISTING_REGION            (0x0000000a)

/**
 * @macro PMM_EREGION_SIZE_ZERO
 *
 * @brief Error: Size of region is zero.
 */
#define PMM_EREGION_SIZE_ZERO                        (0x0000000b)

/**
 * @macro PMM_EWRITE_OUTSIDE_RANGE
 *
 * @brief Error: Write operation failed since address out of range.
 */
#define PMM_EWRITE_OUTSIDE_RANGE                     (0x0000001b)

/**
 * @macro PMM_EWRITE_TO_NON_ALLOCATED
 *
 * @brief Error: Write operation failed since base address not allocated.
 */
#define PMM_EWRITE_TO_NON_ALLOCATED                  (0x0000001c)

/**
 * @macro PMM_EWRITE_FAM_FAILED
 *
 * @brief Error: Write operation failed since FAM failed.
 */
#define PMM_EWRITE_FAM_FAILED                        (0x0000001d)

/**
 * @macro PMM_EBAD_REGION_ADDRESS
 *
 * @brief Error: No region at address.
 */
#define PMM_EBAD_REGION_ADDRESS                      (0x0000001f)

/**
 * @macro PMM_ETOO_MANY_REGIONS
 *
 * @brief Error: Too many simultaneous persistent regions. The limit is set by
 * PMM_START_REQUEST->max_nr_of_regions
 */
#define PMM_ETOO_MANY_REGIONS                        (0x00000020)

/**
 * @macro PMM_EGMM_NOT_FOUND
 *
 * @brief Error: The GMM process was not found. This is an indication of a
 * serious system error and must not be ignored by an error handler.
 */
#define PMM_EGMM_NOT_FOUND                           (0x00000021)

/**
 * @macro PMM_ECONFIGURATION
 *
 * @brief Error: The configuration of the RPMM region is missing or incomplete.
 */
#define PMM_ECONFIGURATION                           (0x00000022)


/*
 ****************************************************************************
 * PUBLIC SIGNAL INTERFACE TO PMM
 ****************************************************************************
 */

/**
 * @signal PMM_VERSION
 *
 * @brief Request information about PMM signal interface version.
 *
 * @long This signal is sent to the Persistent Memory Manager to
 * retrieve the current version number of the signal interface.
 *
 * @long This signal is only handled when PMM has completed its
 * re-installation of persistent regions in MM. Hence a reply to this
 * signal may be used as a notification that persistent regions are restored.
 *
 * @member PMM_VERSION_REQUEST
 * @member PMM_VERSION_REPLY
 * @member pmm_version
 */

/**
 * @macro PMM_VERSION_REQUEST
 * @seealso PMM_VERSION
 */
#define PMM_VERSION_REQUEST (36030)   /* !-SIGNO(struct pmm_version)-! */

/**
 * @macro PMM_VERSION_REPLY
 * @seealso PMM_VERSION
 */
#define PMM_VERSION_REPLY   (36031)   /* !-SIGNO(struct pmm_version)-! */

/**
 * @type pmm_version
 *
 * @input sig_no PMM_VERSION_REQUEST
 * @output sig_no PMM_VERSION_REPLY
 * @output sig_version Version of signal interface.
 */
struct pmm_version
{
      SIGSELECT sig_no;
      U32 sig_version;
};

/*
 ****************************************************************************
 * SIGNAL INTERFACE BETWEEN PMM AND MM.
 ****************************************************************************
 */


/**
 * @signal PMM_ASSIGN
 * @brief Signal sent by PMM to notify MM of its existence.
 *
 * @long Sent by PMM to MM during start. Notifies MM that a PMM
 * exists and triggers a restore sequence between PMM and MM.
 *
 * @member PMM_ASSIGN
 * @member pmm_assign
 */

/**
 * @macro PMM_ASSIGN
 * @seealso PMM_ASSIGN
 */
#define PMM_ASSIGN (36000)  /* !-SIGNO(struct pmm_assign)-! */

/**
 * @type pmm_assign
 *
 * @field sig_no PMM_ASSIGN
 * @field version Version of the signal interface.
 * @field max_nr_of_regions The largest number of simultaneous persistent
 * regions. Set at PMM start in PMM_START_REQUEST->max_nr_of_regions.
 * @field memory_type Name of the memory type, variable length (RAM or FLASH)
 */
struct pmm_assign
{
      SIGSELECT sig_no;
      U32 version;
      U32 max_nr_of_regions;
      char memory_type[1];
};


/**
 * @signal PMM_START
 *
 * @brief Start PMM.
 *
 * @long Sent by creator of PMM to PMM during system start.
 *
 * @long Configures PMM with a memory area and operation mode. The path
 * must be absolute and start with the volume name.
 *
 * @long No error code is returned. If PMM can not handle the memory area
 * or if another fatal error is detected, error() is called.
 *
 * @member PMM_START_REQUEST
 * @member PMM_START_REPLY
 * @member pmm_start_request
 * @member pmm_start_reply
 */

/**
 * @macro PMM_START_REQUEST
 * @seealso PMM_START
 */
#define PMM_START_REQUEST (36032) /* !-SIGNO(struct pmm_start_request)-! */

/**
 * @macro PMM_START_REPLY
 * @seealso PMM_START
 */
#define PMM_START_REPLY (36033)   /* !-SIGNO(struct pmm_start_reply)-! */

/**
 * @type pmm_start_request
 *
 * @field sig_no PMM_START_REQUEST
 * @input options Set by OR'ing selected subset of all PMM_START_OPTION_*
 * constants. Use 0 if no options shall be set. Currently, only
 * PMM_START_OPTION_FORCE_COLD_START is available.
 * @input max_nr_of_regions Maximum number of regions.
 */

struct pmm_start_request
{
      SIGSELECT sig_no;
      U32 options;
      U32 max_nr_of_regions;
};

/**
 * @type pmm_start_reply
 *
 * @input sig_no PMM_START_REPLY
 * @output status PMM_START_WARM or  PMM_START_COLD
 */
struct pmm_start_reply
{
      SIGSELECT sig_no;
      U32 status;
};

/**
 * @signal PMM_INIT
 *
 * @brief Configure PMM.
 *
 * @long Sent by MM to PMM during system start.
 *
 * @long Configures PMM with a memory area and operation mode. The path
 * must be absolute and start with the volume name.
 *
 * @long No error code is returned. If PMM can not handle the memory area
 * or if another fatal error is detected, error() is called.
 *
 * @member PMM_INIT_REQUEST
 * @member PMM_INIT_REPLY
 * @member pmm_init
 */

/**
 * @macro PMM_INIT_REQUEST
 * @seealso PMM_INIT
 */
#define PMM_INIT_REQUEST (36002) /* !-SIGNO(struct pmm_init)-! */

/**
 * @macro PMM_INIT_REPLY
 * @seealso PMM_INIT
 */
#define PMM_INIT_REPLY (36003)   /* !-SIGNO(struct pmm_init)-! */

/**
 * @type pmm_init
 *
 *
 * @input sig_no PMM_INIT_REQUEST
 * @output sig_no PMM_INIT_REPLY
 * @input mode @link "pmm.sig" "PMM_MODE_SAS" or
 * @link "pmm.sig" "PMM_MODE_SASE" (valid only for FLASH PMM)
 * @input mem_area Location and size of external memory.
 * @output alignment Alignment requirement for memory regions.
 */
struct pmm_init
{
      SIGSELECT sig_no;
      U32 alignment;
      struct phys_resource mem_area;
      U32 mode;
};


/**
 * @signal PMM_RESTORE_START
 *
 * @brief Start the restore sequence.
 *
 * @long Sent by MM to PMM. The reply contains base addresses for all
 * persistent that may be restored.
 *
 * @member PMM_RESTORE_START_REQUEST
 * @member PMM_RESTORE_START_REPLY
 * @member pmm_restore_start_reply
 */

/**
 * @macro PMM_RESTORE_START_REQUEST
 * @seealso PMM_RESTORE_START
 */
#define PMM_RESTORE_START_REQUEST (36004) /* !-SIGNO(SIGSELECT)-! */

/**
 * @macro PMM_RESTORE_START_REPLY
 * @seealso PMM_RESTORE_START
 */
#define PMM_RESTORE_START_REPLY (36005) \
  /* !-SIGNO(struct pmm_restore_start_reply)-! */

/**
 * @type pmm_restore_start_reply
 *
 * @input sig_no PMM_RESTORE_START_REQUEST
 * @output sig_no PMM_RESTORE_START_REPLY
 * @output no_regions Number of elements in field regions.
 * @output regions Variable length array with base addresses of
 * persistent regions.
 */
struct pmm_restore_start_reply
{
      SIGSELECT sig_no;
      U32 no_regions;
      log_addr_t regions[1];
};


/**
 * @signal PMM_RESTORE_REGION
 * @brief Request metadata for a logical region.
 *
 * @long Sent by MM to PMM. PMM replies with metadata for a
 * persistent region.
 *
 * @long If the region is bound (region_type == @link "" "PMM_BOUND"), this
 * signal must be followed by a @link "" "PMM_RESTORE_BINDING"
 * signal to fetch information about bindings.
 *
 * @member PMM_RESTORE_REGION_REQUEST
 * @member PMM_RESTORE_REGION_REPLY
 * @member pmm_restore_region_request
 * @member pmm_restore_region_reply
 */

/**
 * @macro PMM_RESTORE_REGION_REQUEST
 * @seealso PMM_RESTORE_REGION
 */
#define PMM_RESTORE_REGION_REQUEST (36006) \
  /* !-SIGNO(struct pmm_restore_region_request)-! */

/**
 * @type pmm_restore_region_request
 *
 * @input sig_no PMM_RESTORE_REGION_REQUEST
 * @input base Logical base address of region.
 */
struct pmm_restore_region_request
{
      SIGSELECT sig_no;
      log_addr_t base;
};

/**
 * @macro PMM_RESTORE_REGION_REPLY
 * @seealso PMM_RESTORE_REGION
 */
#define PMM_RESTORE_REGION_REPLY (36007) \
  /* !-SIGNO(struct pmm_restore_region_reply)-! */

/**
 * @type pmm_restore_region_reply
 *
 * @output sig_no PMM_RESTORE_REGION_REPLY
 * @output base Logical base address of region.
 * @output region_type PMM_BOUND/PMM_UNBOUND
 * @output metadata_size Size of field metadata.
 * @output metadata Metadata for the logical region.
 */
struct pmm_restore_region_reply
{
      SIGSELECT sig_no;
      log_addr_t base;
      U32 region_type;
      U32 metadata_size;
      U8 metadata[PMM_MAX_METALEN];
};


/**
 * @signal PMM_RESTORE_BINDING
 *
 * @brief Request information about physical memory fragments for a
 * logical region.
 *
 * @long Sent by MM to PMM during the restore sequence. PMM replies
 * with information about the actual physical fragments.
 *
 * @member PMM_RESTORE_BINDING_REQUEST
 * @member PMM_RESTORE_BINDING_REPLY
 * @member pmm_restore_binding_request
 * @member pmm_restore_binding_reply
 */

/**
 * @macro PMM_RESTORE_BINDING_REQUEST
 * @seealso PMM_RESTORE_BINDING
 */
#define PMM_RESTORE_BINDING_REQUEST (36008) \
  /* !-SIGNO(struct pmm_restore_binding_request)-! */

/**
 * @type pmm_restore_binding_request
 *
 * @input sig_no PMM_RESTORE_BINDING_REQUEST
 * @input base Logical base address of region.
 */
struct pmm_restore_binding_request
{
      SIGSELECT sig_no;
      log_addr_t base;
};

/**
 * @macro PMM_RESTORE_BINDING_REPLY
 * @seealso PMM_RESTORE_BINDING
 */
#define PMM_RESTORE_BINDING_REPLY (36009) \
  /* !-SIGNO(struct pmm_restore_binding_reply)-! */

/**
 * @type pmm_restore_binding_reply
 *
 * @output sig_no PMM_RESTORE_BINDING_REPLY
 * @output base Logical base address of region.
 * @output no_fragments Number of memory fragments.
 * @output fragments Variable length array with physical addresses.
 */
struct pmm_restore_binding_reply
{
      SIGSELECT sig_no;
      log_addr_t base;
      U32 no_fragments;
      struct phys_resource fragments[1];
};


/**
 * @signal PMM_RESTORE_COMPLETE
 *
 * @brief Mark restore sequence as complete.
 *
 * @long Sent by GMM to PMM when it has received all region meta data
 * and all information about bindings. PMM can check if this is
 * consistent with its view.
 *
 * @member PMM_RESTORE_COMPLETE_REQUEST
 * @member PMM_RESTORE_COMPLETE_REPLY
 */

/**
 * @macro PMM_RESTORE_COMPLETE_REQUEST
 * @seealso PMM_RESTORE_COMPLETE
 */
#define PMM_RESTORE_COMPLETE_REQUEST (36010) /* !-SIGNO(SIGSELECT)-! */

/**
 * @macro PMM_RESTORE_COMPLETE_REPLY
 * @seealso PMM_RESTORE_COMPLETE
 */
#define PMM_RESTORE_COMPLETE_REPLY (36011)   /* !-SIGNO(SIGSELECT)-! */


/**
 * @signal PMM_ALLOC_SASE
 *
 * @brief Allocate a SASE region from PMM.
 *
 * @long Sent by GMM to PMM. Allocate memory for a SASE region and
 * store meta data to file system. One memory fragment will be
 * allocated at the address specified by base.
 *
 * @member PMM_ALLOC_SASE_REQUEST
 * @member PMM_ALLOC_SASE_REPLY
 * @member pmm_alloc_sase_request
 * @member pmm_alloc_sase_reply
 */

/**
 * @macro PMM_ALLOC_SASE_REQUEST
 * @seealso PMM_ALLOC_SASE
 */
#define PMM_ALLOC_SASE_REQUEST (36012) \
  /* !-SIGNO(struct pmm_alloc_sase_request)-! */

/**
 * @type pmm_alloc_sase_request
 *
 * @input sig_no PMM_ALLOC_SASE_REQUEST
 * @input base Logical address for the region.
 * @input region_size Requested region size.
 * @input metadata_size Size of data including null terminator.
 * @input metadata Null terminated ASCII string.
 */
struct pmm_alloc_sase_request
{
      SIGSELECT sig_no;
      log_addr_t base;
      mm_size_t region_size;
      U32 metadata_size;
      U8 metadata[PMM_MAX_METALEN];
};

/**
 * @macro PMM_ALLOC_SASE_REPLY
 * @seealso PMM_ALLOC_SASE
 */
#define PMM_ALLOC_SASE_REPLY (36013) \
  /* !-SIGNO(struct pmm_alloc_sase_reply)-! */

/**
 * @type pmm_alloc_sase_reply
 *
 * @output sig_no PMM_ALLOC_SASE_REPLY
 * @output ecode Status code.
 */
struct pmm_alloc_sase_reply
{
      SIGSELECT sig_no;
      U32 ecode;
};

/**
 * @signal PMM_ALLOC_SAS
 *
 * @brief Allocate a SAS region from PMM.
 *
 * @long Sent by GMM to PMM.
 *
 * @long Allocate memory for a SAS region and store meta data in file
 * system.
 *
 * @long Memory will be allocated if the region is bound (region_type
 * == @link "" "PMM_BOUND"), else only meta data will be
 * stored. Memory fragments will be allocated at addresses chosen by
 * PMM.
 *
 * @member PMM_ALLOC_SAS_REQUEST
 * @member PMM_ALLOC_SAS_REPLY
 * @member pmm_alloc_sas_request
 * @member pmm_alloc_sas_reply
 */


/**
 * @macro PMM_ALLOC_SAS_REQUEST
 * @seealso PMM_ALLOC_SAS
 */
#define PMM_ALLOC_SAS_REQUEST (36014) \
  /* !-SIGNO(struct pmm_alloc_sas_request)-! */


/**
 * @type pmm_alloc_sas_request
 *
 * @input sig_no PMM_ALLOC_SAS_REQUEST
 * @input base Logical address for the region.
 * @input region_size Requested region size.
 * @input region_type See @link "" "PMM_BOUND" or @link "" "PMM_UNBOUND".
 * @input metadata_size Size of data including null terminator.
 * @input metadata Variable length array with metadata, null terminated.
 * @input no_fragments Number of fragments.
 * @input fragments The physical fragments.
 */
struct pmm_alloc_sas_request
{
      SIGSELECT sig_no;
      log_addr_t base;
      mm_size_t region_size;
      U32 region_type;
      U32 metadata_size;
      U8 metadata[PMM_MAX_METALEN];
      U32 no_fragments;
      struct phys_resource fragments[1];
};


/**
 * @macro PMM_ALLOC_SAS_REPLY
 * @seealso PMM_ALLOC_SAS
 */
#define PMM_ALLOC_SAS_REPLY (36015) \
  /* !-SIGNO(struct pmm_alloc_sas_reply)-! */

/**
 * @type pmm_alloc_sas_reply
 *
 * @output sig_no PMM_ALLOC_SAS_REPLY
 * @output ecode Status code.
 * @output no_fragments Number of memory fragments (only external/FLASH).
 * @output fragments Array with allocated fragments, variable length.
 */
struct pmm_alloc_sas_reply
{
      SIGSELECT sig_no;
      U32 ecode;
      U32 no_fragments;
      struct phys_resource fragments[1];
};

/**
 * @signal PMM_FREE
 *
 * @brief Free a persistent region.
 *
 * @long Free memory fragments that are associated with a persistent
 * region. Remove metadata for the region.
 *
 * @member PMM_FREE_REQUEST
 * @member PMM_FREE_REPLY
 * @member pmm_free
 */

/**
 * @macro PMM_FREE_REQUEST
 * @seealso PMM_FREE
 */
#define PMM_FREE_REQUEST (36016) /* !-SIGNO(struct pmm_free)-! */
/**
 * @macro PMM_FREE_REPLY
 * @seealso PMM_FREE
 */
#define PMM_FREE_REPLY (36017)   /* !-SIGNO(struct pmm_free)-! */

/**
 * @type pmm_free
 *
 * @input sig_no PMM_FREE_REQUEST
 * @output sig_no PMM_FREE_REPLY
 * @output ecode Status code.
 * @input base Logical address for the region.
 */
struct pmm_free
{
      SIGSELECT sig_no;
      U32 ecode;
      log_addr_t base;
};
/**
 * @signal PMM_CHANGE
 *
 * @brief change metadata on a persistent region.
 *
 * @long When clients changes the cache or permission attributes on
 *       a persistent region, this signal is sent to PMM.
 *
 * @member PMM_CHANGE_REQUEST
 * @member PMM_CHANGE_REPLY
 * @member pmm_change
 */

/**
 * @macro PMM_CHANGE_REQUEST
 * @seealso PMM_CHANGE
 */
#define PMM_CHANGE_REQUEST (36022) /* !-SIGNO(struct pmm_change)-! */
/**
 * @macro PMM_CHANGE_REPLY
 * @seealso PMM_CHANGE
 */
#define PMM_CHANGE_REPLY (36023)   /* !-SIGNO(struct pmm_change)-! */

/**
 * @type pmm_change
 *
 * @input sig_no PMM_CHANGE_REQUEST
 * @output sig_no PMM_CHANGE_REPLY
 * @output ecode Status code.
 * @input base Logical address for the region.
 * @input metadata_size Size of modified data including null terminator.
 * @input metadata Variable length array with metadata, null terminated.
 *
 */
struct pmm_change
{
   SIGSELECT sig_no;
   U32 ecode;
   log_addr_t base;
   U32 metadata_size;
   U8 metadata[PMM_MAX_METALEN];
};

/**
 * @signal PMM_WRITE
 *
 * @brief Write data to a persistent region.
 *
 * @long MM writes data to persistent regions with this signal. A
 * write request from a client may result in several
 * PMM_WRITE_REPLY signals.
 *
 * @long MM translates SAS addresses to SASE addresses before sending
 * PMM_WRITE signals. Thus PMM is able do deduce sector numbers etc.
 * from the addresses.
 *
 * @member PMM_WRITE_REQUEST
 * @member PMM_WRITE_REPLY
 * @member pmm_write
 */

/**
 * @macro PMM_WRITE_REQUEST
 * @seealso PMM_WRITE
 */
#define PMM_WRITE_REQUEST (36018) /* !-SIGNO(struct pmm_write)-! */
/**
 * @macro PMM_WRITE_REPLY
 * @seealso PMM_WRITE
 */
#define PMM_WRITE_REPLY (36019)   /* !-SIGNO(struct pmm_write)-! */
/**
 * @type pmm_write
 *
 * @input sig_no PMM_WRITE_REQUEST
 * @output sig_no PMM_WRITE_REPLY
 * @output ecode Status code.
 * @input dest Destination address.
 * @input source Source address.
 * @input size Number of bytes to write.
 */
struct pmm_write
{
      SIGSELECT sig_no;
      U32 ecode;
      log_addr_t dest;
      log_addr_t source;
      mm_size_t size;
};


/**
 * @signal PMM_CORRUPT_METADATA
 *
 * @brief Notify PMM that a persistent region has corrupt metadata.
 *
 * @long Sent by MM to PMM during the restore phase.
 *
 * @long MM detected corrupt metadata of a region during restore
 * phase. The region is put in state erase and fatal error is
 * called. In next restart this region will be removed and its
 * fragments will be removed from the external memory (erased in the case
 * of flash).
 *
 * @member PMM_CORRUPT_METADATA
 * @member pmm_corrupt_metadata
 */

/**
 * @macro PMM_CORRUPT_METADATA
 * @seealso PMM_CORRUPT_METADATA
 */
#define PMM_CORRUPT_METADATA (36020) \
       /* !-SIGNO(struct pmm_corrupt_metadata)-! */

/**
 * @type pmm_corrupt_metadata
 *
 * @field sig_no PMM_CORRUPT_METADATA
 * @field base Logical address for the region.
 */
struct pmm_corrupt_metadata
{
      SIGSELECT sig_no;
      log_addr_t base;
};

/**
 * @signal pmm_reset
 *
 * @brief Reset PMM.
 *
 * @long This signal (only RAM PMM) resets the persistent information
 * in the PMM.
 * @long NOTE: After this, the MM and PMM will have inconsistent
 * view of persistency.
 * @member pmm_reset
 */

/**
 * @macro PMM_RESET
 * @seealso pmm_reset
 */
#define PMM_RESET (36025) \
       /* !-SIGNO(struct pmm_reset)-! */

/**
 * @type pmm_reset
 *
 * @field sig_no PMM_RESET
 */
struct pmm_reset
{
   SIGSELECT sig_no;
};


/**
 * @signal PMM_HUNT_SIG
 *
 * @brief Internal PMM signal used in the hunt system call.
 *
 * @member PMM_HUNT_SIG
 */


/**
 * @macro PMM_HUNT_SIG
 * @seealso PMM_HUNT_SIG
 */
#define PMM_HUNT_SIG (36096) /* !-SIGNO(SIGSELECT)-! */
/* Contains no data. */


/**
 * @signal PMM_ERASE_BLOCK
 *
 * @brief Erase a number of FLASH blocks.
 *
 * @long Internal PMM signal sent from main process to spawned erase
 * process.
 *
 * @long This signal is optional depending on the implementation of
 * PMM.
 *
 * @member PMM_ERASE_BLOCK_REQUEST
 * @member PMM_ERASE_BLOCK_REPLY
 * @member pmm_erase_block
 */

/**
 * @macro PMM_ERASE_BLOCK_REQUEST
 * @seealso PMM_ERASE_BLOCK
 */
#define PMM_ERASE_BLOCK_REQUEST (36098) \
  /* !-SIGNO(struct pmm_erase_block)-! */

/**
 * @macro PMM_ERASE_BLOCK_REPLY
 * @seealso PMM_ERASE_BLOCK
 */
#define PMM_ERASE_BLOCK_REPLY (36099) \
  /* !-SIGNO(struct pmm_erase_block)-! */

/**
 * @type pmm_erase_block
 *
 * @input sig_no PMM_ERASE_BLOCK_REQUEST
 * @output sig_no PMM_ERASE_BLOCK_REPLY
 * @output ecode Status code.
 * @output no_blocks Number of blocks to erase.
 * @input eraseblocks Variable length array with block numbers.
 * @output eraseblocks Variable length array with block numbers.
 */
struct pmm_erase_block
{
      SIGSELECT sig_no;
      U32 ecode;
      U32 no_blocks;
      struct phys_resource eraseblocks[1];
};
#ifdef __cplusplus
} /* end of extern "C" */

#endif

#endif /* #ifdef _PMM_SIG */
