/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2011 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END *
 */

/**
 * @toc Device_Drivers:
 * @file ose_spi/profiling_counters.h
 * @brief The hardware counter profiling API
 *
 * @long Counter profiling is not supported for all boards,
 *       therefore all functions are implemented using weak linking.
 */

#ifndef PROFILING_COUNTERS_H
#define PROFILING_COUNTERS_H

#include "stdint.h"
#include "compiler.h"

#define NO_EVENT    0xFFFFFFFF
#define COUNTER_OFF 0
#define COUNTER_ON  1

#ifdef __cplusplus
extern "C" {
#endif


/**
 * @type  ProfilingPerfCounterStatus
 * @declaration
 * typedef enum ProfilingPerfCounterStatus {
 *    PROFILING_PC_STATUS_OK = 0,
 *    PROFILING_PC_STATUS_TYPE_NOT_COUNTED,
 *    PROFILING_PC_STATUS_TYPE_ALREADY_COUNTED,
 *    PROFILING_PC_STATUS_NO_COUNTER_AVAILABLE_FOR_TYPE,
 *    PROFILING_PC_STATUS_NO_BUFFERS,
 *    PROFILING_PC_STATUS_INVALID_ARGUMENT
 * }  ProfilingPerfCounterStatus;
 *
 * @brief Enumeration of status values that may be returned.
 *
 * @long
 *        All status values, except for PROFILING_PC_STATUS_OK,
 *        should be considered to be error codes.
 *
 * @field PROFILING_PC_STATUS_OK                Status OK.
 * @field PROFILING_PC_STATUS_TYPE_NOT_COUNTED
 * The specified counter event type
 * is currently not being counted.
 *
 * @field PROFILING_PC_STATUS_TYPE_ALREADY_COUNTED
 *                                              The specified counter event type
 *                                              is already being counted.
 * @field PROFILING_PC_STATUS_NO_COUNTER_AVAILABLE_FOR_TYPE
 * The specified counter event type cannot be counted
 * as there is currently no counter available that can accommodate it.
 *
 * @field PROFILING_PC_STATUS_NO_BUFFERS
 *              the specified counter event type cannot be started due to
 *              the lack of sample buffers (see profiling_add_sample_buffer)
 * @field PROFILING_PC_STATUS_INVALID_ARGUMENT
 *      the user called the driver API function with invalid arguments
 *
 */
typedef enum ProfilingPerfCounterStatus {
    PROFILING_PC_STATUS_OK = 0,
    PROFILING_PC_STATUS_TYPE_NOT_COUNTED,
    PROFILING_PC_STATUS_TYPE_ALREADY_COUNTED,
    PROFILING_PC_STATUS_NO_COUNTER_AVAILABLE_FOR_TYPE,
    PROFILING_PC_STATUS_NO_BUFFERS,
    PROFILING_PC_STATUS_INVALID_ARGUMENT
} ProfilingPerfCounterStatus;

/**
 * @type
 *      ProfilingPerfCounterEventType
 * @declaration
 * typedef struct ProfilingPerfCounterEventType {
 *    uint32_t type;
 *    const char *description;
 * } ProfilingPerfCounterEventType;
 * @brief
 *        Represents a counter event type and its description.
 *        type The CPU specific counter event type ID.
 *        description Null-terminated event type description string.
 *
 * @field type
 * @field description
 *
 *
 */
typedef struct ProfilingPerfCounterEventType {
    uint32_t type;
    const char *description;
} ProfilingPerfCounterEventType;

/**
 * @function
 *        ProfilingBufferFullCB
 * @declaration typedef void (*ProfilingBufferFullCB) (uint32_t event_type,
 *                                  uint64_t * sample_buffer,
 *                                  uint32_t sample_count);
 * @brief Pointer to callback that is called when sample buffer is full.
 * @long
 *        Pointer to callback that is called when sample buffer is full.
 *        If the buffer queue is empty, the full buffer will be reused and
 *        sample_buffer will be NULL. This should be interpreted as loss of
 *        sample data. The sample_count will indicate the number of
 *        samples that were lost.
 * @input event_type    The counter event type count for
 * @input sample_buffer Pointer to the sample buffer
 * @input sample_count  Number of samples lost
 *
 */
typedef void (*ProfilingBufferFullCB) (uint32_t event_type,
                                  uint64_t * sample_buffer,
                                  uint32_t sample_count);


/**
 * @function
 *        ProfilingBufferFreeCB
 * @declaration typedef void (*ProfilingBufferFreeCB) (uint64_t *sample_buffer);
 * @brief Pointer to callback that is called when driver buffer list is flushed.
 * @long
 *        Pointer to callback that is called when driver buffer list is flushed.
 *
 * @input sample_buffer Pointer to the sample buffer
 *
 */
typedef void (*ProfilingBufferFreeCB) (uint64_t * sample_buffer);

/**
 * @function profiling_get_supported_counter_event_types
 *
 * @declaration const ProfilingPerfCounterEventType
 *    *profiling_get_supported_counter_event_types(uint32_t * type_count);
 *
 * @input type_count    Number of elements in the array of
 *                              supported counter event types
 *
 * @return Pointer to the array of supported counter event types.
 *
 * @brief This function gets the array of supported counter event types.
 *
 *
 */
#pragma weak profiling_get_supported_counter_event_types 

const ProfilingPerfCounterEventType *
profiling_get_supported_counter_event_types(uint32_t * type_count);


/**
 * @function profiling_get_count
 * @declaration uint64_t profiling_get_count(uint32_t event_type);
 *
 * @input event_type The counter event type to get the count for it.
 *
 * @return The count or 0 if the event type is not being counted.
 *
 * @brief
 *        Get the current count of the specified counter event type.
 *
 *
 */
#pragma weak profiling_get_count
uint64_t profiling_get_count(uint32_t event_type);

/**
 * @function profiling_reset_count
 * @declaration ProfilingPerfCounterStatus
 *                      profiling_reset_count(uint32_t event_type);
 * @input event_type The counter event type to reset the count for.
 *
 * @return
 *        Returns PROFILING_PC_STATUS_OK if successful, or
 *        PROFILING_PC_STATUS_TYPE_NOT_COUNTED if the specified event type
 *        is not currently being counted.
 *
 * @brief
 *        Resets the current count of the specified counter event type.
 *
 *
 */
#pragma weak profiling_reset_count
ProfilingPerfCounterStatus profiling_reset_count(uint32_t event_type);

/**
 * @function profiling_start_counting
 * @declaration ProfilingPerfCounterStatus
 *              profiling_start_counting(uint32_t event_type);
 * @input event_type The counter event type to count.
 *
 * @return
 *        PROFILING_PC_STATUS_OK if successful, or
 *        PROFILING_PC_STATUS_TYPE_ALREADY_COUNTED if events of
 *        the specified type are already being counted,
 *        or PROFILING_PC_STATUS_NO_COUNTER_AVAILABLE_FOR_TYPE
 *        if there is no counter register available for
 *        counting events of the specified type.
 *
 * @brief
 *        Starts counting events of the specified type.
 *
 *
 */
#pragma weak profiling_start_counting
ProfilingPerfCounterStatus profiling_start_counting(uint32_t event_type);

/**
 * @function profiling_start_counting_with_sampling
 * @declaration ProfilingPerfCounterStatus
 * profiling_start_counting_with_sampling(uint32_t event_type,
 *                                      uint64_t trigger_count,
 *                                      ProfilingBufferFullCB
 *                                      buffer_full_cb);
 *
 * @input       event_type      The counter event type to count.
 * @input       trigger_count   Number of increments between PC samples.
 * @input       buffer_full_cb  Callback function that is called when a
 *                              PC sample buffer is full.
 *
 * return
 *        PROFILING_PC_STATUS_OK if successful, or
 *        PROFILING_PC_STATUS_TYPE_ALREADY_COUNTED if events of
 *        the specified type are already being counted,
 *        or PROFILING_PC_STATUS_NO_COUNTER_AVAILABLE_FOR_TYPE
 *        if there is no counter register available for
 *        counting events of the specified type.
 *        PROFILING_PC_STATUS_INVALID_ARGUMENT if the sample_buffer is
 *        NULL or the sample_slot_count value is invalid.
 *
 * @brief
 *        Counts events and samples PC when trigger count is reached.
 *
 *
 */
#pragma weak profiling_start_counting_with_sampling
ProfilingPerfCounterStatus
profiling_start_counting_with_sampling(uint32_t event_type,
                                  uint64_t trigger_count,
                                  ProfilingBufferFullCB
                                  buffer_full_cb);

/**
 * @function profiling_is_counting
 * @declaration uint8_t profiling_is_counting(uint32_t event_type);
 * @input
 *        event_type - The counter event type to check.
 *
 * @return Returns non-zero if the type is being counted, 0 otherwise.
 *
 * @brief
 *        Checks if the specified counter event type is being counted.
 *
 */
#pragma weak profiling_is_counting
uint8_t profiling_is_counting(uint32_t event_type);

/**
 * @function profiling_get_trigger_count
 * @declaration ProfilingPerfCounterStatus
 *              profiling_get_trigger_count(uint32_t event_type,
 *                      uint64_t *trigger_count, uint32_t *use_trigger);
 * @input event_type            The counter event type to check.
 * @input trigger_count         The trigger value will be stored to this
 * @input use_trigger           The trigger use value will be stored to this
 *
 * @return Returns status.
 *
 * @brief
 *        Returns the specified counter event type trigger configuration.
 *
 *
 */
#pragma weak profiling_get_trigger_count
ProfilingPerfCounterStatus
profiling_get_trigger_count(uint32_t event_type,
        uint64_t *trigger_count, uint32_t *use_trigger);

/**
 * @function profiling_stop_counting
 * @declaration ProfilingPerfCounterStatus
 *              profiling_stop_counting(uint32_t event_type);
 *
 * @input      event_type The counter event type to stop counting.
 *
 * @return
 *        Returns PROFILING_PC_STATUS_OK if successful, or
 *        PROFILING_PC_STATUS_TYPE_NOT_COUNTED if the specified.
 *
 * @brief
 *        Stops counting events of the specified type.
 *
 */
#pragma weak profiling_stop_counting
ProfilingPerfCounterStatus profiling_stop_counting(uint32_t event_type);

/**
 * @function profiling_stop_all_counting
 * @declaration void profiling_stop_all_counting(void);
 * @brief
 *        Stops the counting of all event types.
 *
 */
#pragma weak profiling_stop_all_counting
void profiling_stop_all_counting(void);

/**
 * @function profiling_add_sample_buffer
 * @declaration ProfilingPerfCounterStatus
 *                      profiling_add_sample_buffer(uint64_t * sample_buffer,
 *                              uint32_t sample_slot_count);
 *
 * @input       sample_buffer      Pointer to the sample buffer.
 * @input       sample_slot_count  Number of samples that fit in the buffer.
 *
 * @return
 *        Returns PROFILING_PC_STATUS_OK if successful, or
 *        PROFILING_PC_STATUS_BUFFER_QUEUE_FULL if the buffers queue if full.
 *        PROFILING_PC_STATUS_INVALID_ARGUMENT if the sample_buffer is
 *        NULL or the sample_slot_count value is invalid.
 *
 * @brief
 *        Adds a sample buffer to sample buffer queue associated with event.
 *
 *
 */
#pragma weak profiling_add_sample_buffer
ProfilingPerfCounterStatus
profiling_add_sample_buffer(uint64_t * sample_buffer,
                       uint32_t sample_slot_count);

/**
 * @function profiling_flush_sample_buffers
 * @declaration void profiling_flush_sample_buffers(ProfilingBufferFreeCB *cb);
 *
 * @input       cb  Callback function that is called when a
 *                              a sample buffer should be released.
 *
 * @return
 *        Returns PROFILING_PC_STATUS_OK if successful, or
 *        PROFILING_PC_STATUS_INVALID_ARGUMENT if the callback is NULL.
 *
 * @brief
 *        Frees the sample buffers in the driver buffer queue.
 *
 */
#pragma weak profiling_flush_sample_buffers
ProfilingPerfCounterStatus
profiling_flush_sample_buffers(ProfilingBufferFreeCB cb);

#ifdef __cplusplus
}
#endif

#endif /* PROFILING_COUNTERS_H */
