/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2005-2006 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

/**
 * @toc OSE:
 * @file ose_spi/ramdump.h
 *
 * @brief This file defines the interface to the RAM dump utility.
 *
 * @long This file contains prototypes and macros for the RAM dump utility.
 * The purpose of the RAM dump is to provide a way to dump error information
 * and memory content from the system error handler for later retrieval
 * and analysis. The RAM dump is not a general post mortem tool, as its only
 * functionality is to store arbitrary information.
 *
 * @long A memory area must be reserved for the dump content. The area must
 * be located in a region of memory that is not cleared, initialized or
 * affected in any way during a warm start. Most notably, it must be
 * outside of the .data and .bss sections.
 *
 * @long An additional memory area is required for the ramdump heap
 * which the RAM dump uses as a work area.
 *
 * @long All configuration parameters are specified as system parameters.
 * See core component manual for details on the required configuration.
 *
 * @long The RAM dump file format is described in the core component
 * manual.
 *
 */

#ifndef _RAMDUMP_H
#define _RAMDUMP_H

#ifdef __cplusplus
extern "C" {
#endif

#include "string.h"
#include "ose.h"
#include "osetypes.h"

/* Error codes */
   
/**
 * @macro RAMDUMP_SUCCESS
 *
 * @brief Indicates a successful operation.
 */
#define RAMDUMP_SUCCESS                     (0)

/**
 * @macro RAMDUMP_ECURR_DUMP_TOO_LARGE
 *
 * @brief Error code indicating that a dump operation could not be
 * completed due to insufficient amount of memory for the current dump.
 *
 * @long Error code indicating that a dump operation could not be
 * completed due to insufficient amount of memory for the current dump.
 * A "current" dump will be the target for all the dump actions until
 * the user either explicitly calls ramdump_start_new_dump or performs
 * a restart of the system. The latter case causes a new dump to be started
 * if ramdump_start_handler1 is configured as START_HANDLER1 in the
 * kernel configuration file.
 */
#define RAMDUMP_ECURR_DUMP_TOO_LARGE        (-1)

/**
 * @macro RAMDUMP_EMEMORY_ACCESS_NOK
 *
 * @brief Error code indicating that a memory access failed.
 *
 * @long Error code indicating that a memory access failed due
 * to indirect data input, i.e. passing a wild pointer or an address
 * to non-existent memory may cause this condition.
 */
#define RAMDUMP_EMEMORY_ACCESS_NOK          (-2)

/**
 * @macro RAMDUMP_EPOS_OR_DUMP_NOT_AVAILABLE
 *
 * @brief Error code indicating that a dump or position does not exist.
 *
 * @long Error code indicating that a specified dump does not exist,
 * or that a request to read from a certina position in a dump failed.
 */
#define RAMDUMP_EPOS_OR_DUMP_NOT_AVAILABLE  (-3)

/**
 * @macro RAMDUMP_ERECURSIVE_DUMP_CALL
 *
 * @brief Error code indicating that the dump area has been locked
 * for use by another process.
 *
 * @long Error code indicating that the dump area has been locked
 * for use by another process. This situation typically occurs if
 * a new error is encountered while a dump is in progress.
 */
#define RAMDUMP_ERECURSIVE_DUMP_CALL        (-4)

/**
 * @macro RAMDUMP_ERAMDUMP_NOT_INITIALIZED
 *
 * @brief Error code indicating that the dump area has not been initialized.
 */
#define RAMDUMP_ERAMDUMP_NOT_INITIALIZED    (-5)

/**
 * @macro RAMDUMP_EINVALID_ARGUMENT
 *
 * @brief Error indicating that a particular argument passed to function
 * contained an invalid value.
 *
 * @long Error indicating that a particular argument passed to function
 * contained an invalid value, and was detected as such by the function.
 */
#define RAMDUMP_EINVALID_ARGUMENT   (-6)

/**
 * @macro RAMDUMP_EUNABLE_TO_GET_CURRENT_PROCESS_INFO
 *
 * @brief Error indicating that the ramdump was unable to retrieve the
 * current process information.
 */
#define RAMDUMP_EUNABLE_TO_GET_CURRENT_PROCESS_INFO   (-7)
   
/**
 * @macro RAMDUMP_EUNABLE_TO_GET_CURRENT_BLOCK_INFO
 *
 * @brief Error indicating that the ramdump was unable to retrieve the
 * current block information.
 */
#define RAMDUMP_EUNABLE_TO_GET_CURRENT_BLOCK_INFO   (-8)

/**
 * @macro RAMDUMP_EMUST_BE_SUPERVISOR
 *
 * @brief Error indicating that an operation failed because the current
 * process was not in supervisor mode, when supervisor mode was required.
 */
#define RAMDUMP_EMUST_BE_SUPERVISOR   (-9)

/**
 * @macro RAMDUMP_EOPERATION_REQUIRES_ZLIB
 *
 * @brief Error indicating that an operation failed because
 * the functionality to perform the operation requires zlib.
 */
#define RAMDUMP_EOPERATION_REQUIRES_ZLIB   (-10)

/**
 * @macro RAMDUMP_ENO_COMPRESSION_STREAM
 * 
 * @brief Error indicating that the compression stream could not
 * be initialized.
 */
#define RAMDUMP_ENO_COMPRESSION_STREAM   (-11)

/**
 * @macro RAMDUMP_ECOMPRESSION_FAILED
 * 
 * @brief Error indicating that the compression failed.
 */
#define RAMDUMP_ECOMPRESSION_FAILED   (-12)

/**
 * @macro RAMDUMP_MAX_DUMPS_INFO
 *
 * @brief Maximum number of dumps that can be handled
 * by the dump information retrieval functions.
 */
#define RAMDUMP_MAX_DUMPS_INFO        (32)

/**
 * @type RamdumpDumpId
 *
 * @brief An unsigned integer type used to hold a dump identifier.
 */
typedef U32 RamdumpDumpId;

/**
 * @type RamdumpStatus
 *
 * @brief A signed integral type used for the return status of
 * ramdump functions.
 */
typedef S32 RamdumpStatus;

/**
 * @type RamdumpWatchdogCb
 *
 * @brief The watchdog function type.
 */
typedef void RamdumpWatchdogCb(void);
   
/**
 * @type RamdumpDumpInfo
 *
 * @brief Basic information about a dump.
 *
 * @long Structure containing dump id, size and time stamp
 * information for a dump.
 *
 * @field dump_id Dump identifier.
 * @field size Size in bytes of the dump.
 * @field clock Timestamp in seconds since Jan 1, 1970 00:00:00.
 * @field microseconds Microsecond count part of timestamp.
 */
struct RamdumpDumpInfo {
  RamdumpDumpId dump_id;
  U32 size;
  U32 clock;
  U32 microseconds;
};

/**
 * @function ramdump_start_new_dump
 *
 * @brief Starts a new dump.
 *
 * @long This function starts a new dump after the last one in
 * the post-mortem dump area. Subsequent calls to ramdump_dump_xxx
 * will write information in the same dump.
 *
 * @long If more than one dump in one session is desired, call this function
 * to start a new dump.
 *
 * @long NOTE: Interrupts must be disabled when calling this function.
 *
 * @restriction Available in supervisor processes, not in interrupt processes.
 *
 * @return If the function succeeds, RAMDUMP_SUCCESS is returned.
 * If the dump function was called recursively, the error code
 * RAMDUMP_ERECURSIVE_DUMP_CALL is returned. If the ramdump area is
 * uninitialized, the error code RAMDUMP_ERAMDUMP_NOT_INITIALIZED
 * is returned.
 */
RamdumpStatus ramdump_start_new_dump(void);

/**
 * @function ramdump_clear_dumps
 *
 * @brief Clears dump area.
 *
 * @long This function clears the dump area.
 *
 * @restriction Available in supervisor processes, not in interrupt processes.
 *
 * @return If the function succeeds, RAMDUMP_SUCCESS is returned.
 * If the dump function was called recursively, the error code
 * RAMDUMP_ERECURSIVE_DUMP_CALL is returned. If the ramdump area is
 * uninitialized, the error code RAMDUMP_ERAMDUMP_NOT_INITIALIZED
 * is returned.
 *
 */
RamdumpStatus ramdump_clear_dumps(void);

/**
 * @function ramdump_dump_error_info
 *
 * @brief Writes information about an error event into the current dump.
 *
 * @long NOTE: Interrupts must be disabled when calling this function.
 *
 * @restriction Available in supervisor processes, not in interrupt processes.
 *
 * @input description User-provided optional text. The string is only
 * copied up to the first terminator character.
 * @input user_called Pass value of user_called flag here.
 * @input error_code Pass error code here.
 * @input extra Pass extra parameter here.
 *
 * @return If the function succeeds, RAMDUMP_SUCCESS is returned.
 * If the current dump exceeds its size limit, the error code
 * RAMDUMP_ECURR_DUMP_TOO_LARGE is returned. If the description pointer
 * is invalid, the error code RAMDUMP_EMEMORY_ACCESS_NOK is returned.
 * If the dump function was called recursively, the error code
 * RAMDUMP_ERECURSIVE_DUMP_CALL is returned. If the ramdump area is
 * uninitialized, the error code RAMDUMP_ERAMDUMP_NOT_INITIALIZED
 * is returned. RAMDUMP_EUNABLE_TO_GET_CURRENT_PROCESS_INFO is returned
 * if the function is unable to get information for the current process.
 */
RamdumpStatus ramdump_dump_error_info(const char *description,
				      OSBOOLEAN user_called,
				      OSERRCODE error_code,
				      OSERRCODE extra);

/**
 * @function ramdump_dump_text_info
 *
 * @brief Writes an arbitrary textual description into the current dump.
 *
 * @long NOTE: Interrupts must be disabled when calling this function.
 *
 * @restriction Available in supervisor processes, not in interrupt processes.
 *
 * @input description User-provided optional text. The string
 * is only copied up to the first terminator character.
 *
 * @return If the function succeeds, RAMDUMP_SUCCESS is returned.
 * If the current dump exceeds its size limit, the error code
 * RAMDUMP_ECURR_DUMP_TOO_LARGE is returned. If the description pointer
 * is invalid, the error code RAMDUMP_EMEMORY_ACCESS_NOK is returned.
 * If the dump function was called recursively, the error code
 * RAMDUMP_ERECURSIVE_DUMP_CALL is returned. If the ramdump area is
 * uninitialized, the error code RAMDUMP_ERAMDUMP_NOT_INITIALIZED
 * is returned.
 */
RamdumpStatus ramdump_dump_text_info(const char *description);

/**
 * @function ramdump_dump_memory_image
 *
 * @brief Dumps a section of memory to the current dump.
 *
 * @long NOTE: Interrupts must be disabled when calling this function.
 *
 * @restriction Available in supervisor processes, not in interrupt processes.
 *
 * @input description User-provided optional text. The string is only
 * copied up to the first terminator character.
 * @input start_address Start address of memory to dump.
 * @input length Number of bytes to dump.
 *
 * @return If the function succeeds, RAMDUMP_SUCCESS is returned.
 * If the current dump exceeds its size limit, the function fails
 * and returns the error code RAMDUMP_ECURR_DUMP_TOO_LARGE. If the memory
 * access fails or the description pointer is invalid, the error code
 * RAMDUMP_EMEMROY_ACCESS_NOK is returned. If the dump function was called
 * recursively, the error code RAMDUMP_ERECURSIVE_DUMP_CALL is returned.
 * If the ramdump area is unitialized, the error code
 * RAMDUMP_ERAMDUMP_NOT_INITIALIZED is returned.
 */
RamdumpStatus ramdump_dump_memory_image(const char *description,
					OSADDRESS start_address,
					U32 length);

/**
 * @function ramdump_get_dump_id
 *
 * @brief Retrieves a dump identifier in the dump history.
 *
 * @long This function retrieves a dump identifier in the dump history.
 * The function retrieves the latest dump if nthLast == 0,
 * the second latest dump if nthLast == 1 and so on.
 *
 * @restriction Available in supervisor processes, not in interrupt processes.
 *
 * @input nth_last An index to the nth last dump id to retrieve.
 *
 * @return If the function succeeds, a dump identifier is returned. If
 * no nth_last dump was found, the function returns zero.
 */
RamdumpDumpId ramdump_get_dump_id(U32 nth_last);

/**
 * @function ramdump_get_dump_info
 *
 * @brief Retrieves basic information about a dump.
 *
 * @long This function retrieves some basic information about a dump,
 * such as size and timestamp.
 *
 * @restriction Available in supervisor processes, not in interrupt processes.
 *
 * @input dump_id Dump identifier.
 * @output dump_info User-provided dump information struct for the
 * function to fill out.
 *
 * @return If the function succeeds, RAMDUMP_SUCCESS is returned.
 * If the specified dump id is not valid, the error code
 * RAMDUMP_EPOS_OR_DUMP_NOT_AVAILABLE is returned. If dump_info
 * is a NULL pointer, RAMDUMP_ERAMDUMP_INVALID_ARGUMENT is returned.
 * If the ramdump area is uninitialized, the error code
 * RAMDUMP_ERAMDUMP_NOT_INITIALIZED is returned.
 */
RamdumpStatus ramdump_get_dump_info(RamdumpDumpId dump_id,
				    struct RamdumpDumpInfo *dump_info);

/**
 * @function ramdump_read_dump
 *
 * @brief Copies a number of bytes of a dump into a user-provided buffer.
 *
 * @restriction Available in supervisor processes, not in interrupt processes.
 *
 * @input dump_id Dump identifier.
 * @input pos Byte offset in the dump.
 * @output buffer User-provided buffer into which data is copied.
 * @input bytes Number of bytes to copy.
 *
 * @return If the function succeeds, it returns the number of
 * bytes copied. If memory access to the buffer fails, the error code
 * RAMDUMP_EMEMORY_ACCESS_NOK is returned. If the position does not
 * exist in the dump, or if the dump id is not a valid active dump id,
 * the error code RAMDUMP_EPOS_OR_DUMP_NOT_AVAILABLE is returned.
 * If the ramdump area is uninitialized, the error code
 * RAMDUMP_ERAMDUMP_NOT_INITIALIZED is returned.
 *
 * @example
 * U8 buffer[1024];
 * RamdumpDumpInfo dump_info;
 * RamdumpDumpId dump_id = ramdump_get_dump_id(0);
 * RamdumpStatus status = ramdump_get_dump_info(dump_id, &dump_info);
 * if (status == RAMDUMP_SUCCESS)
 * {
 *    int bytes_read;
 *    int pos;
 *    for (pos = 0; pos < dump_info.size; pos += bytes_read)
 *    {
 *       bytes_read = ramdump_read_dump(dump_id, pos, (void *)buffer,
                               min(sizeof(buffer), dump_info.size - pos));
 *    }
 * }
 *
 */
S32 ramdump_read_dump(RamdumpDumpId dump_id, U32 pos,
		      void *buffer, U32 bytes);

/**
 * @function ramdump_start_handler1
 *
 * @brief Start handler that initializes the ramdump.
 *
 * @long Start handler that initializes the ramdump, specified
 * in the kernel configuration file.
 *
 * @example
 * START_HANDLER1(ramdump_start_handler1)
 */
void ramdump_start_handler1(void);

/**
 * @function ramdump_set_watchdog_function
 *
 * @brief Sets a watchdog callback function to be called during dumping.
 *
 * @long Watchdog support is provided for use when dumps are time
 * consuming. A callback function can be installed in supervisor mode,
 * to be called at regular intervals during the dump phase.
 * It is an error to call this function in non-supervisor mode.
 * Installing a NULL pointer disables watchdog support.
 *
 * @long NOTE: Interrupts must be disabled when calling this function.
 *
 * @input f Pointer to the watchdog callback function to install.
 *
 * @return If the function succeeds in setting the watchdog callback
 * function pointer, the return value is RAMDUMP_SUCCESS. If the ramdump
 * area is uninitialized, the function fails and returns the error code
 * RAMDUMP_ERAMDUMP_NOT_INITIALIZED.
 */
RamdumpStatus ramdump_set_watchdog_function(RamdumpWatchdogCb *f);

/**
 * @function add_ramdump_cmds
 *
 * @brief Adds ramdump shell commands.
 *
 * @long Adds the ramdump shell commands 'ramdumps' and
 * 'ramdumpsave'. The 'ramdumps' command lists the existing
 * dumps and 'ramdumpsave' saves a specified dump to a file.
 */
void add_ramdump_cmds(void);

#ifdef __cplusplus
}
#endif

#endif /* # ifndef _RAMDUMP_H */
