/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2005-2006 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */
/**
 *  @toc OSE:
 *  @file ose_spi/ramtrace.h
 *  @brief This is the system program interface of the OSE trace API.
 * *
 *  @long This trace package supports circular trace buffers and
 *  trace entries of variable size and content. Several different
 *  trace buffers can exist at the same time in the system at
 *  different memory locations. The trace can be used in any supervisor
 *  context (but not from user mode), including interrupt and
 *  exception contexts. Users of this interface must be linked to the
 *  core module.
 *
 *  @long That the trace is circular means that it will never fill up
 *  from a logger's perspective, i.e. as long as the individual trace
 *  entry is smaller than the whole trace buffer, it will always fit.
 *  What happens when the trace is full is that the oldest entry in the
 *  trace is removed (repeated until the new entry fits). This way
 *  there is always a consecutive trail of the most recent events in
 *  the trace.
 *
 *  @long An individual trace entry can not be larger than 65535 bytes, but
 *  a trace entry can have sub trace entries associated with it that allow
 *  larger data structures to be stored in the trace buffer. Also several
 *  trace entries can share one sub trace entry which allows for space
 *  saving since data common to several entries is only stored once.
 *
 *  @long Normally each trace entry has a timestamp (consisting of
 *  the system tick and time count value). The user needs to provide a
 *  function that reads the system tick and the time count value.
 *  The frequency of the time count can be arbitrary and is also specified
 *  in the ose_trace_create call. The OSE system call get_systime can
 *  be used as the timestamp function.
 *
 *  @long Trace entries are considered binary and transparent to
 *  the OSE trace implementation. However by convention each trace
 *  entry has the same layout
 *  as an OSE signal, i.e. it starts with a 32 bit unsigned integer
 *  that uniquely identifies the layout of the trace entry and then
 *  data as described by a struct follows. If this
 *  convention is used, the sigdb tool can be used for generating
 *  symbolic information about the trace content and a trace entry
 *  can thus be easily printed in text using the sigdesc API.
 *
 *  @long This trace buffer SPI is used by the RMM to implement the
 *  OSE kernel event trace that can trace signal send, receive, and
 *  process create, kill, and swap events. The implementation can
 *  however also be used for application specific tracing of events
 *  that are suitable for storing in a RAM based circular buffer.
 *
 */
#ifndef _RAMTRACE_H
#define _RAMTRACE_H

#include "stdlib.h"
#include "stdint.h"
#include "compiler.h"

#ifdef __cplusplus
extern "C" {
#endif

/**
 * @macro TRACE_MIN_SIZE
 * @brief Manifest constant that defines the minimum trace size.
 * @seealso ose_trace_create
 */
#define TRACE_MIN_SIZE 0x1000

/**
 *  @macro TRACE_MAX_NAME_LEN
 *  @brief Manifest constant that defines the maximum trace name length.
 *  @seealso ose_trace_create
 */
#define TRACE_MAX_NAME_LEN 16

/**
 *  @type EntrySize
 *  @brief Type used for trace entry size.
 *
 *  @long The type is used to define the entry_size field in the TraceEntryState
 *  structure and in various trace calls dealing with trace entry sizes.
 */
typedef uint16_t EntrySize;

/**
 * @type TraceAllocFlags
 *
 * @brief Flags passed to ose_trace_alloc.
 *
 * @field T_FORCE    Create trace entry even if tracing is disabled.
 * @field T_NOTIME   Do not write timestamp to the trace entry.
 * @field T_HIDDEN   Hidden trace entries are not returned when reading the trace.
 * @field T_REFERRED The trace entry is referred entry that can not be freed
 * until the trace is completely overwritten.
 *
 * @seealso ose_trace_alloc
 */
enum TraceAllocFlags {
   T_FORCE    = 0x10,
   T_NOTIME   = 0x20,
   T_HIDDEN   = 0x40,
   T_REFERRED = 0x80
};

/**
 * @type TraceEntryState
 * @brief This structure is used to hold temporary data while
 *        examining the trace.
 *
 * @field trace_entry     Pointer to current trace entry.
 * @field tick            System tick of event.
 * @field time_count      A time count related to the tick.
 * @field time_freq       Frequency of time_count in Hz.
 * @field stamp_valid     Non-zero when time_stamp is present.
 * @field entry_size      Number of bytes in the current entry.
 * @field base            Pointer to the base of the trace buffer.
 * @field trace_name      Name of the trace buffer.
 *
 * @seealso ose_trace_find
 * @seealso ose_trace_get_next
 *
 **/
typedef struct
{
   void     *trace_entry;
   uint32_t  tick;
   uint32_t  time_count;
   uint32_t  time_freq;
   uint8_t   stamp_valid;
   EntrySize entry_size;
   void     *base;
   char      trace_name[TRACE_MAX_NAME_LEN + 1];
} TraceEntryState;

/**
 * @type GetTimeFuncPointer
 * @brief Function pointer type to a function returning system tick.
 *
 * @long The client of the trace needs to implement a function that
 * returns the system tick and fills in the time count argument.
 *
 * @long The function is compatible to the get_systime of the OSE kernel.
 *
 * @field time_count      A time count related to the tick.
 *
 * @seealso ose_trace_create
 * @seealso ose_trace_recover
 *
 */
typedef uint32_t (*GetTimeFuncPointer)(uint32_t *time_count);

/**
 * @function ose_trace_create
 *
 * @include ose_spi/ramtrace.h
 *
 * @brief Initializes a trace.
 *
 * @long Sets up a new trace that is cleared but not enabled. This is
 * normally only done once for a trace buffer.
 *
 * @long The caller must make sure the given memory are is not used by
 * anything else and the clients to the trace must be in supervisor mode.
 * Further more the size of the trace buffer must be at least 4KB.
 *
 * @input buffer Buffer is the lowest address of the memory area. The
 *        buffer must be aligned to a cache line boundary.
 *
 * @input size Size is the number of bytes in the memory area.
 *
 * @input name Name is an ASCII string used to identify this particular
 *             trace buffer. The name must not be longer than
 *             TRACE_MAX_NAME_LEN.
 *
 * @input get_time_func A function pointer to the function that returns
 *                      the system tick and the time count value.
 *
 * @input time_freq The frequency of the time count value, as returned
 *                  by the get_time_func function, in Hz.
 *
 * @return 0 if the trace was successfully initialized, less than 0 otherwise.
 *
 * @seealso TRACE_MAX_NAME_LEN
 * @seealso ose_trace_recover
 * @seealso ose_trace_enable
 * @seealso GetTimeFuncPointer
 */
int
ose_trace_create(void *buffer, uint32_t size, const char *name,
		 GetTimeFuncPointer get_time_func, uint32_t time_freq);

/**
 * @function ose_trace_recover
 *
 * @include ose_spi/ramtrace.h
 *
 * @brief Recover a trace after a soft restart.
 *
 * @long After a soft reset (i.e. a reset that keeps content of the trace
 * buffer intact) this function carefully checks all metadata of the trace
 * and if that succeeds it returns 0. Otherwise the trace is reinitialized
 * and a return code greater than 0 is returned.
 *
 * @long Since this function carefully traverses all entries in the trace, the
 * execution time of this function depends upon how many entries there are in
 * the existing trace.
 *
 * @restriction Available in supervisor processes, and in interrupt processes.
 *
 * @input buffer Buffer is the lowest address of the memory area. The
 *        buffer must be aligned to a cache line boundary.
 *
 * @input size Size is the number of bytes in the memory area.
 *
 * @input name Name is an ASCII string used to identify this particular
 *             trace buffer. The name must not be longer than
 *             TRACE_MAX_NAME_LEN.
 *
 * @input get_time_func A function pointer to the function that returns
 *                      the system tick and the time count value.
 *
 * @input time_freq The frequency of the time count value, as returned
 *                  by the get_time_func function, in Hz.
 *
 * @return 0 if trace is recovered
 *         greater than 0 if not recovered but trace is reinitialized
 *         less than 0 if not recovered and trace initialization failed.
 */
int
ose_trace_recover(void *buffer, uint32_t size, const char *name,
		  GetTimeFuncPointer get_time_func, uint32_t time_freq);

/**
 * @function ose_trace_find
 *
 * @include ose_spi/ramtrace.h
 *
 * @brief Searches the given memory area for the specified
 *		trace buffer.
 * @long If the trace is found, data storage is disabled and a
 *		pointer to the trace buffer is returned.
 *
 * @restriction Available in supervisor processes, and in interrupt processes.
 *
 * @input first_mem The lowest address of the memory area to be searched.
 * Using the same value as in the ose_trace_create call is most the efficient
 * way to finding the trace buffer.
 *
 * @input mem_size is the number of bytes in the searched memory area.
 *		The trace head structure must be completely located within
 *		this limit. Using the same value as in the ose_trace_create
 *		call is the most efficient way.
 *
 * @input name The ASCII name used to identify this particular
 *		trace buffer. If name = "", then any trace buffer found is
 *               returned.
 *
 * @return	A pointer to the trace buffer if found, NULL otherwise.
 *
 */
void *
ose_trace_find(void *first_mem, uint32_t mem_size, const char *name);

/**
 * @function ose_trace_clear
 *
 * @include ose_spi/ramtrace.h
 *
 * @brief Removes all data from the trace buffer.
 *
 * @long This is done by incrementing the trace sequence number to
 *	invalidate all previous trace entries.
 *	The trace enable status is not affected.
 *
 * @restriction Available in supervisor processes, and in interrupt processes.
 *
 * @input buffer Buffer is the trace buffer address given to the
 *               ose_trace_create call.
 *
 */
void
ose_trace_clear(void *buffer);

/**
 * @function ose_trace_enable
 *
 * @include ose_spi/ramtrace.h
 *
 * @brief Enables a trace buffer for trace data storage.
 *
 * @long Subsequent calls to ose_trace_alloc causes data to be stored in
 *       the trace buffer.
 *
 * @restriction Available in supervisor processes, and in interrupt processes.
 *
 * @input buffer Buffer is the trace buffer address given to the
 *               ose_trace_create call.
 *
 * @seealso ose_trace_create
 * @seealso ose_trace_disable
 */
void
ose_trace_enable(void *buffer);

/**
 * @function ose_trace_disable
 *
 * @include ose_spi/ramtrace.h
 *
 * @brief Disables a trace buffer for trace data storage.
 *
 * @long Subsequent calls to ose_trace_alloc has no effect.
 *
 * @restriction Available in supervisor processes, and in interrupt processes.
 *
 * @input buffer Buffer is the trace buffer address given to the
 *               ose_trace_create call.
 *
 * @seealso ose_trace_create
 * @seealso ose_trace_enable
 */
void
ose_trace_disable(void *buffer);

/**
 * @function ose_trace_alloc
 *
 * @include ose_spi/ramtrace.h
 *
 * @brief Reserve space in the trace for a trace entry
 *	 if trace data storage is enabled.
 *
 * @long This method is used to put new entries into the trace buffer.
 * An entry is created at the end of the trace if the trace is enabled,
 * if the trace buffer is disabled this method has no effect and NULL
 * is returned. Note that
 * ose_trace_commit must be called for the entry to make it visible.
 *
 * @restriction Available in supervisor processes, and in interrupt processes.
 *
 * @input buffer The trace buffer address given to the
 *               ose_trace_create call.
 *
 * @input size Wanted size of the trace entry.
 *
 * @input flags Flags from @link "" "TraceAllocFlags".
 *
 * @return A pointer to allocated entry or NULL if no entry could be
 * obtained.
 *
 * @seealso ose_trace_commit
 * @seealso ose_trace_create
 * @seealso ose_trace_enable
 */
void *
ose_trace_alloc(void *buffer, EntrySize size, uint32_t flags);

/**
 * @function ose_trace_commit
 *
 * @include ose_spi/ramtrace.h
 *
 * @brief Commit and make an entry part of the trace.
 *
 * @long After an entry has been allocated and filled in with data it
 * must be committed to the trace for the entry to be visible. After
 * the entry has been committed it may no longer be changed.
 *
 * @restriction Available in supervisor processes, and in interrupt processes.
 *
 * @input entry Pointer to the trace entry to be committed.
 *
 * @seealso ose_trace_alloc
 */
void
ose_trace_commit(void *entry);

/*  !! NOTE: No SPI-doc for ose_trace_reclaim any more
 * @function ose_trace_reclaim
 *
 * @include ramtrace.h
 *
 * @brief Marks a trace entry as if it was newly allocated from
 * ose_trace_alloc(). This function can only be used immediately after a
 * successful ose_trace_recover() operation. The purpose is to take back
 * action list entries that were accidentally freed by ose_trace_recover().
 *
 * @restriction Available in supervisor processes, and in interrupt processes.
 *
 * @input entry Pointer to a trace entry to be reclaimed.
 */
void
ose_trace_reclaim(void *entry);

/**
 * @function ose_trace_open
 *
 * @include ramtrace.h
 *
 * @brief This function is used when opening the trace buffer,
 *        after having read it.
 *
 * @restriction Available in supervisor processes, and in interrupt processes.
 *
 * @input buffer Buffer is the trace buffer address given to the
 * ose_trace_create call.
 *
 * @input cursor Pointer to the control structure used when unpacking the
 * trace.
 *
 * @return The function returns zero on success, non-zero on failure.
 */
int
ose_trace_open(void *buffer, TraceEntryState *cursor);

/**
 * @function ose_trace_close
 *
 * @include ose_spi/ramtrace.h
 *
 * @brief This function is used for closing the trace buffer,
 *        preparing it for reading.
 *
 * @restriction Available in supervisor processes, and in interrupt processes.
 *
 * @input cursor Pointer to the control structure used when unpacking the
 * trace.
 *
 * @return The function returns zero on success, non-zero on failure.
 */
int
ose_trace_close(TraceEntryState *cursor);

/**
 * @function ose_trace_read
 *
 * @include ose_spi/ramtrace.h
 *
 * @brief This function is used when reading the trace buffer.
 *
 * @long Read the next valid trace entry information into the buffer pointed to
 * by the 'cursor' parameter.
 *
 * @restriction Available in supervisor processes, and in interrupt processes.
 *
 * @input cursor Pointer to the control structure used when unpacking the
 * trace.
 *
 * @output cursor Information about the next trace entry.
 *
 * @return  0 if nothing to read @break
 *          1 if one entry found @break
 *         >1 if one entry found and data loss (the estimated lost amount in
 *            bytes is returned)
 *
 * @seealso TraceEntryState
 * @seealso ose_trace_open
 *
 * @example
 *    void *trace_buffer;
 *    TraceEntryState state;
 *    if (ose_trace_open(trace_buffer, &state) == 0)
 *    {
 *       while (ose_trace_read(&state)
 *       {
 *          print_my_trace_entry(state.trace_entry);
 *       }
 *    }
 *    ose_trace_close(&state);
 */
int
ose_trace_read(TraceEntryState *cursor);

/* Obsolete functions */
#ifdef USE_OBSOLETE
void *ose_trace_alloc_subentry(void *buffer, EntrySize size);
void  ose_trace_free_subentry(void *entry);
int   ose_trace_get_first(void *buffer, TraceEntryState *cursor);
int   ose_trace_get_next(TraceEntryState *cursor);
#endif

#ifdef __cplusplus
}
#endif

#endif /* ifndef _TRACE_H */
