/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2010 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

/**
 * @public
 *
 * @toc OSE:
 *
 * @file ose_spi/rmd.h
 *
 * @brief Run-Mode Dumper Interface.
 *
 * @long This file defines the interface to the Run-Mode Dumper (RMD). It can
 * be used to fetch information on program state from OSE core components
 * (such as the Executive, the Program Manager and the Heap) and
 * store the information for later analysis.
 *
 * @long See more about RMD in OSE Core User's Guide.
 *
 * @seealso ose_err/rmd_err.h
 */

#ifndef _RMD_H
#define _RMD_H

#ifdef __cplusplus
extern "C" {
#endif

#include "ose.h"
#include "cpu.h"
#include "stdint.h"

/**
 * @type RmdStatus
 * @brief Type of status codes.
 *
 * @seealso RmdStatusValue
 */
typedef int32_t RmdStatus;

/**
 * @type RmdStatusValue
 * @brief Status codes returned in calls to the RMD.
 *
 * @long Errors retrieving information are indicated with positive status
 * codes and errors storing the information retrieved are indicated with
 * negative status codes. Zero indicates success.
 *
 * @field RMD_STATUS_FILE_OPEN_ERROR File could not be opened.
 * @field RMD_STATUS_FILE_ERROR An file system error occurred.
 * @field RMD_STATUS_OK No error
 * @field RMD_STATUS_INTERNAL_ERROR An internal error occurred in the RMD.
 * @field RMD_STATUS_INTERFACE_NOT_SUPPORTED Interface not supported by the
 *        RMD process.
 * @field RMD_STATUS_TOO_MANY_DUMPS There are too many run-mode dumps ongoing
 *        at the same time.
 * @field RMD_STATUS_DUMP_FILE_NAME_GENERATION_ERROR Failed to generate dump
 *        file name.
 * @field RMD_STATUS_CONNECT_TO_RMM_FAILED Failed to connect to RMM process.
 * @field RMD_STATUS_DETACHED_FROM_RMM Detached from RMM process.
 * @field RMD_STATUS_INVALID_PARAMETERS Invalid parameters.
 * @field RMD_STATUS_MEMORY_ALLOCATION_ERROR No memory to allocate internal
 *        buffers.
 * @field RMD_STATUS_ZLIB_ERROR Zlib returned an error.
 * @field RMD_STATUS_DUMP_NOT_FOUND No ongoing dump for the requester PID was
 *        found in the dump list.
 * @field RMD_STATUS_UNKNOWN_INFO_TAG An information tag was used that is
 *        unknown to this RMD implementation.
 * @field RMD_STATUS_UNABLE_TO_RETRIEVE_DATA Failed to retrieve the requested 
 *        data.
 *
 * @seealso RmdStatus
 */
enum RmdStatusValue
{
   RMD_STATUS_FILE_OPEN_ERROR = -2,
   RMD_STATUS_FILE_ERROR = -1,
   RMD_STATUS_OK = 0,
   RMD_STATUS_INTERNAL_ERROR = 1,
   RMD_STATUS_INTERFACE_NOT_SUPPORTED = 2,
   RMD_STATUS_TOO_MANY_DUMPS = 3,
   RMD_STATUS_DUMP_FILE_NAME_GENERATION_ERROR = 4,
   RMD_STATUS_CONNECT_TO_RMM_FAILED = 5,
   RMD_STATUS_DETACHED_FROM_RMM = 6,
   RMD_STATUS_INVALID_PARAMETERS = 7,
   RMD_STATUS_MEMORY_ALLOCATION_ERROR = 8,
   RMD_STATUS_ZLIB_ERROR = 9,
   RMD_STATUS_DUMP_NOT_FOUND                  = 10,
   RMD_STATUS_UNKNOWN_INFO_TAG                = 11,
   RMD_STATUS_UNABLE_TO_RETRIEVE_DATA         = 12
};

/**
 * @type RmdScope
 * @brief Type of scope.
 *
 * @seealso RmdScopeValue
 */
typedef uint32_t RmdScope;

/**
 * @type RmdScopeValue
 *
 * @brief Scope type constants used to select the set of processes (or
 * other entities with a PID) for which information or memory contents should
 * be saved.
 *
 * @field RMD_SCOPE_SYSTEM All processes. Only select this in
 *                         functions documented to support it.
 * @field RMD_SCOPE_PROCESS The specified process.
 * @field RMD_SCOPE_BLOCK The processes in the specified block.
 * @field RMD_SCOPE_SEGMENT The processes in the specified segment.
 *
 * @seealso RmdScope
 */
enum RmdScopeValue
{
   RMD_SCOPE_SYSTEM      = 0x00,
   RMD_SCOPE_SEGMENT     = 0x08,
   RMD_SCOPE_BLOCK       = 0x20,
   RMD_SCOPE_PROCESS     = 0x80
};


/**
 * @type RmdInfo
 * @brief Type of information.
 *
 * @seealso RmdInfoValue
 */
typedef uint32_t RmdInfo;

/**
 * @type RmdInfoValue
 *
 * @brief Information type constants used to select type and level of detail
 * of the information to be saved.
 *
 * @field RMD_INFO_PROCESS
 * Brief process / block / segment information for processes in the scope.
 * @field RMD_INFO_PROCESS_LONG
 * Detailed process / block / segment information for processes in the scope.
 * @field RMD_INFO_STACK_USAGE
 * Stack location, size and usage information for processes in the scope.
 * @field RMD_INFO_ENV_VARS
 * Environment variables for blocks and processes in the scope.
 * @field RMD_INFO_SIGNAL_QUEUE
 * Signal information for signals in the queue of all processes in the scope.
 * @field RMD_INFO_POOL_INFO
 * Pool information for all pools processes in the scope can use.
 * @field RMD_INFO_POOL_SIGNALS
 * Signal information for all signal buffers in all pools processes in the
 * scope can use.
 * @field RMD_INFO_POOL_SIGNALS_OWNED
 * Signal information for all signal buffers that processes in the scope own.
 * @field RMD_INFO_SYS_PARAM
 * System parameters.
 * @field RMD_INFO_REGISTERS
 * Register values for all processes in the scope.
 * @field RMD_INFO_LOAD_MODULE_INFO
 * Load module and relocation information for the programs to which the
 * processes in the scope belong.
 * @field RMD_INFO_PROGRAM_INFO
 * Load module and relocation information for the programs to which the
 * processes in the scope belong.
 * @field RMD_INFO_TRACE
 * Ramtrace contents.
 *
 * @seealso RmdInfo
 */
enum RmdInfoValue
{
   RMD_INFO_PROCESS = 0x1,
   RMD_INFO_PROCESS_LONG = 0x2,
   RMD_INFO_STACK_USAGE = 0x4,
   RMD_INFO_ENV_VARS = 0x8,
   RMD_INFO_SIGNAL_QUEUE = 0x10,
   RMD_INFO_POOL_INFO = 0x20,
   RMD_INFO_POOL_SIGNALS = 0x40,
   RMD_INFO_POOL_SIGNALS_OWNED = 0x80,
   RMD_INFO_SYS_PARAM = 0x100,
   RMD_INFO_REGISTERS = 0x200,
   RMD_INFO_LOAD_MODULE_INFO = 0x400,
   RMD_INFO_PROGRAM_INFO = 0x800,
   RMD_INFO_TRACE = 0x1000
};

/**
 * @macro RMD_INFO_SOURCE_DEBUG
 * @brief RmdInfoValue suitable for full source code debugging.
 *
 * @seealso RmdInfo
 */
#define RMD_INFO_SOURCE_DEBUG \
   (RMD_INFO_PROCESS | RMD_INFO_REGISTERS \
    | RMD_INFO_LOAD_MODULE_INFO | RMD_INFO_PROGRAM_INFO)

/**
 * @macro RMD_INFO_SYSTEM_DEBUG
 * @brief RmdInfoValue suitable for full system browsing.
 *
 * @seealso RmdInfo
 */
#define RMD_INFO_SYSTEM_DEBUG \
   (RMD_INFO_PROCESS_LONG | RMD_INFO_ENV_VARS \
    | RMD_INFO_SIGNAL_QUEUE | RMD_INFO_POOL_INFO | RMD_INFO_SYS_PARAM)

/**
 * @macro RMD_INFO_POOL_DEBUG
 * @brief RmdInfoValue suitable for full pool analysis.
 *
 * @seealso RmdInfo
 */
#define RMD_INFO_POOL_DEBUG \
   (RMD_INFO_PROCESS | RMD_INFO_STACK_USAGE \
    | RMD_INFO_POOL_INFO | RMD_INFO_POOL_SIGNALS)

/**
 * @macro RMD_INFO_GENERAL_DEBUG
 * @brief RmdInfoValue suitable for general situations when extensive
 * information is needed.
 *
 * @seealso RmdInfo
 */
#define RMD_INFO_GENERAL_DEBUG \
   (RMD_INFO_PROCESS_LONG | RMD_INFO_STACK_USAGE | RMD_INFO_ENV_VARS \
    | RMD_INFO_SIGNAL_QUEUE | RMD_INFO_POOL_INFO | RMD_INFO_POOL_SIGNALS \
    | RMD_INFO_SYS_PARAM | RMD_INFO_REGISTERS | RMD_INFO_LOAD_MODULE_INFO \
    | RMD_INFO_PROGRAM_INFO)

/**
 * @type RmdMemory
 * @brief Type of memory.
 *
 * @seealso RmdMemoryValue
 */
typedef uint32_t RmdMemory;

/**
 * @type RmdMemoryValue
 * @brief Memory type constants used to select type and amount of memory to
 * be saved.
 *
 * @field RMD_MEMORY_POOL
 * Pool contents for all pools processes in the scope can use.
 * @field RMD_MEMORY_STACKS
 * Pool contents used for stacks for all pools processes in the scope can use.
 * @field RMD_MEMORY_STACKS_OWNED
 * Stacks of all processes in the scope.
 * @field RMD_MEMORY_STACKS_USED
 * The used part of the stacks of all processes in the scope.
 * @field RMD_MEMORY_STACKS_TOP
 * The top of the stacks of all processes in the scope.
 * @field RMD_MEMORY_SIGNALS
 * Pool contents used for signals for all pools processes in the scope can use.
 * @field RMD_MEMORY_SIGNALS_OWNED
 * Signals owned by processes in the scope.
 * @field RMD_MEMORY_SIGNALS_QUEUED
 * Signals in the queues of processes in the scope.
 * @field RMD_MEMORY_SIGNALS_SENT
 * Signals with processes in the scope as sender.
 * @field RMD_MEMORY_HEAP
 * Heap contents for all heaps processes in the scope can use.
 * @field RMD_MEMORY_HEAP_BUFFERS
 * Heap buffers for all heaps processes in the scope can use.
 * @field RMD_MEMORY_HEAP_BUFFERS_OWNED
 * Heap buffers owned by processes in the scope.
 * @field RMD_MEMORY_DATA
 * Data sections (.data, .bss, etc.) for the programs to which the processes
 * in the scope belong.
 * @field RMD_MEMORY_TEXT
 * Text sections (.text, etc.) for the programs to which the processes in the
 * scope belong.
 *
 * @seealso RmdMemory
 */
enum RmdMemoryValue
{
   RMD_MEMORY_POOL = 0x1,
   RMD_MEMORY_STACKS = 0x2,
   RMD_MEMORY_STACKS_OWNED = 0x4,
   RMD_MEMORY_STACKS_USED = 0x8,
   RMD_MEMORY_STACKS_TOP = 0x10,
   RMD_MEMORY_SIGNALS = 0x20,
   RMD_MEMORY_SIGNALS_OWNED = 0x40,
   RMD_MEMORY_SIGNALS_QUEUED = 0x80,
   RMD_MEMORY_SIGNALS_SENT = 0x100,
   RMD_MEMORY_HEAP = 0x200,
   RMD_MEMORY_HEAP_BUFFERS = 0x400,
   RMD_MEMORY_HEAP_BUFFERS_OWNED = 0x800,
   RMD_MEMORY_DATA = 0x1000,
   RMD_MEMORY_TEXT = 0x2000
};

/**
 * @macro RMD_MEMORY_SOURCE_DEBUG
 * @brief RmdMemoryValue suitable for full source code debugging.
 *
 * @seealso RmdMemoryValue
 */
#define RMD_MEMORY_SOURCE_DEBUG \
   (RMD_MEMORY_STACKS_USED | RMD_MEMORY_SIGNALS_OWNED \
    | RMD_MEMORY_HEAP_BUFFERS_OWNED)

/**
 * @macro RMD_MEMORY_SYSTEM_DEBUG
 * @brief RmdMemoryValue suitable for full system browsing.
 *
 * @seealso RmdMemoryValue
 */
#define RMD_MEMORY_SYSTEM_DEBUG \
   (0)

/**
 * @macro RMD_MEMORY_POOL_DEBUG
 * @brief RmdMemoryValue suitable for full pool analysis.
 *
 * @seealso RmdMemoryValue
 */
#define RMD_MEMORY_POOL_DEBUG \
   (0)

/**
 * @macro RMD_MEMORY_GENERAL_DEBUG
 * @brief RmdMemoryValue suitable for general situations when extensive
 * information is needed.
 *
 * @seealso RmdMemoryValue
 */
 #define RMD_MEMORY_GENERAL_DEBUG \
   (RMD_MEMORY_STACKS | RMD_MEMORY_SIGNALS | RMD_MEMORY_HEAP_BUFFERS)


/**
 * @type RmdProperty
 * @brief Rmd property type constants.
 *
 * @field RMD_NUM_LONG_QUEUE_PROCS - Only include entities for processes
 *                            that have the longest signal queues.
 *                            The value is the max number of processes.
 *                            Applicable to rmd_dump_signal_queue_info and
 *                            rmd_dump_pool_memory functions.
 * @field RMD_MAX_SIGNALS_IN_QUEUE - Max number of signals to be dumped from the queue
 *                            in the scope. Applicable to rmd_dump_signal_queue_info
 *                            and rmd_dump_pool_memory functions.     
 * @field RMD_MAX_STACK_TOP_SIZE - Max stack top size to be dumped in the scope.
 *                          Applicable to rmd_dump_pool_memory function.  
 *
 */

typedef enum RmdProperty
{
   RMD_NUM_LONG_QUEUE_PROCS = 0,
   RMD_MAX_SIGNALS_IN_QUEUE,
   RMD_MAX_STACK_TOP_SIZE
} RmdProperty;


/**
 * @type RmdDumpInfoTag
 * @brief Rmd information tag constants.
 *
 * @field RMD_PATH_NAME_TAG  Used to obtain the file name path of the dump.
 *                           Type of data is a char *.
 *
 * @field RMD_DUMP_ID_TAG    Used to obtain the dump id of the ongoing dump.
 *                           Type of data is uint32_t *.
 *
 */

typedef enum RmdDumpInfoTag
{
   RMD_PATH_NAME_TAG = 0,
   RMD_DUMP_ID_TAG
} RmdDumpInfoTag;


/**
 * @function rmd_start_dump
 * @brief Start and create a new run-mode dump file.
 *
 * @long Subsequent calls to rmd_dump_* will append information to this new
 * dump.
 *
 * @long The dump will be associated with the current process and if the current
 * process already have an ongoing dump then that dump file will be closed. If
 * the calling process is killed the dump file will be closed.
 *
 * @restriction Available in user mode and supervisor processes, not in interrupt processes.
 *
 * @input path The path and file name to the new dump. If NULL then a path name
 * will be generated.
 *
 * @return 0 if successful.
 *
 * @seealso rmd_stop_dump
 * @seealso RamdumpStatus
 */
RmdStatus rmd_start_dump(const char *path);

/**
 * @function rmd_stop_dump
 * @brief Stop and close the current run-mode dump file.
 *
 * @long This function is normally called after an error handler has finished
 * dumping a program.
 *
 * @restriction Available in user mode and supervisor processes, not in interrupt processes.
 *
 * @return 0 if successful.
 *
 * @seealso rmd_start_dump
 * @seealso RamdumpStatus
 */
RmdStatus rmd_stop_dump(void);

/**
 * @function rmd_get_dump_info_allocated
 * @brief Obtain information about the current dump.
 *
 * @long A tag specifies which information to be fetched for the current ongoing
 * dump. rmd_get_dump_info_allocated() must be called between rmd_start_dump() and
 * rmd_stop_dump(). A pointer to pointer must be passed in that will contain
 * a alloced buffer with the requested information. The buffer must be freed
 * after use. The type of the pointer depends on the tag.
 *
 * @input info_tag Tag that specifies the information to get.
 * @output info Pointer to pointer to a buffer with information.
 *
 * @return RMD_STATUS_OK if successful.
 *
 * @restriction Can only be called from the process context that are currently
 * creating a dump.
 *
 * @seealso RmdDumpInfoTag
 *
 *
 * @example
 * char path[128];
 * char *info;
 * if (RMD_STATUS_OK == rmd_get_dump_info_allocated(RMD_PATH_NAME_TAG,
 *                                               (union SIGNAL **)&info))
 * {
 *    strncpy(path, (char *)info, 128);
 *    free_buf((union SIGNAL **)&info);
 * }
 * else
 * {
 *    strncpy(path, "<unknown>", 128);
 * }
 *
 */
RmdStatus rmd_get_dump_info_allocated(RmdDumpInfoTag info_tag, 
                                      union SIGNAL **info);

/**
 * @function rmd_get_dump_info
 * @deprecated This method is deprecated from OSE 5.6.1 and/or from OSE 5.5.4
 *             and will not be supported in future releases of OSE. Instead
 *             use rmd_get_dump_info_allocated() function.
 *
 * @brief Obtain information about the current dump.
 *
 * @long A tag specifies which information to be fetched for the current ongoing
 * dump. rmd_get_dump_info() must be called between rmd_start_dump() and
 * rmd_stop_dump(). A pointer to a pointer must be passed in that will contain
 * a malloced buffer with the requested information. The buffer must be freed
 * after use. The type of the pointer depends on the tag.
 *
 * @input info_tag Tag that specifies the information to get.
 * @output info Pointer to pointer to a buffer with information.
 *
 * @return RMD_STATUS_OK if successful.
 *
 * @restriction Can only be called from the process context that are currently
 * creating a dump.
 * 
 * @seealso RmdDumpInfoTag
 * 
 *
 * @example
 * char path[128];
 * char *info;
 * if (RMD_STATUS_OK == rmd_get_dump_info(RMD_PATH_NAME_TAG, &info))
 * {
 *    strncpy(path, (char *)info, 128);
 *    free(info);
 * }
 * else
 * {
 *    strncpy(path, "<unknown>", 128);
 * }
 * 
 */
RmdStatus rmd_get_dump_info(RmdDumpInfoTag info_tag, void **info);

/**
 * @function rmd_dump_error
 * @brief Store error information in the current dump.
 *
 * @long This should be one of the first functions called when creating a
 * dump after a fatal error.
 *
 * @restriction Available in user mode and supervisor processes, not in interrupt processes.
 *
 * @field description String describing the error or NULL.
 * @field user_called Error handler parameter.
 * @field ecode Error handler parameter.
 * @field extra Error handler parameter.
 *
 * @return 0 if successful.
 *
 * @seealso RamdumpStatus
 */
RmdStatus rmd_dump_error(const char *description,
                          OSBOOLEAN user_called,
                          OSERRCODE ecode,
                          OSERRCODE extra);

/**
 * @function rmd_dump_memory
 * @brief Store a region of RAM in the current dump.
 *
 * @long The description string is used in host tools to match the saved
 * blocks to plain text formatters of the saved data. This allows application
 * specific data structures to be stored and interpreted on the host without
 * the need to implement parsers for the entire .pmd file format.
 *
 * @restriction Available in user mode and supervisor processes, not in interrupt processes.
 *
 * @field description String describing the data or NULL.
 * @field start_address Start of the region to save.
 * @field length Length of the region to save.
 *
 * @return 0 if successful.
 *
 * @seealso RamdumpStatus
 */
RmdStatus rmd_dump_memory(const char *description,
                           OSADDRESS start_address,
                           uint32_t length);

/**
 * @function rmd_dump_string
 * @brief Store a string with plain text information in the current dump.
 *
 * @restriction Available in user mode and supervisor processes, not in interrupt processes.
 *
 * @field description String to store.
 *
 * @return 0 if successful.
 *
 * @seealso RamdumpStatus
 */
RmdStatus rmd_dump_string(const char *description);

/**
 * @function rmd_dump_process_info
 * @brief Store process, block, and segment information in the current dump.
 *
 * @long This should be one of the first functions called (after rmd_dump_error)
 * when creating a dump after a fatal error, to ensure that the process control
 * blocks of the executive are intact, so that the scopes and filter mechanisms
 * of subsequent calls to rmd_dump_*_info and rmd_dump_*_memory are functional.
 *
 * @long If none of the relevant bits (RMD_INFO_PROCESS and
 * RMD_INFO_PROCESS_LONG) are set in info, this function does nothing.
 *
 * @restriction Available in user mode and supervisor processes, not in interrupt processes.
 *
 * @field scope The set of processes for which to store information.
 * Valid values are RMD_SCOPE_SEGMENT, RMD_SCOPE_BLOCK and RMD_SCOPE_PROCESS.
 * @field info The kind of information to store
 *
 * @return 0 if successful.
 *
 * @seealso RmdScopeValue
 * @seealso RmdInfoValue
 */
RmdStatus rmd_dump_process_info(RmdScope scope, RmdInfo info);

/**
 * @function rmd_dump_stack_usage_info
 * @brief Store stack usage information in the current dump.
 *
 * @long If none of the relevant bits (RMD_INFO_STACK_USAGE) are set in info,
 * this function does nothing.
 *
 * @restriction Available in user mode and supervisor processes, not in interrupt processes.
 *
 * @field scope The set of processes for which to store information.
 * Valid values are RMD_SCOPE_SEGMENT, RMD_SCOPE_BLOCK and RMD_SCOPE_PROCESS.
 * @field info The kind of information to store.
 *
 * @return 0 if successful.
 *
 * @seealso RmdScopeValue
 * @seealso RmdInfoValue
 */
RmdStatus rmd_dump_stack_usage_info(RmdScope scope, RmdInfo info);

/**
 * @function rmd_dump_env_vars_info
 * @brief Store environment variable information in the current dump
 *
 * @long If none of the relevant bits (RMD_INFO_ENV_VARS) are set in info,
 * this function does nothing.
 *
 * @restriction Available in user mode and supervisor processes, not in interrupt processes.
 *
 * @field scope The set of processes and blocks for which to store information.
 * Valid values are RMD_SCOPE_SEGMENT, RMD_SCOPE_BLOCK and RMD_SCOPE_PROCESS.
 * @field info The kind of information to store.
 *
 * @return 0 if successful.
 *
 * @seealso RmdScopeValue
 * @seealso RmdInfoValue
 */
RmdStatus rmd_dump_env_vars_info(RmdScope scope, RmdInfo info);

/**
 * @function rmd_dump_signal_queue_info
 * @brief Store signal queue information in the current dump
 *
 * @long If none of the relevant bits (RMD_INFO_SIGNAL_QUEUE) are set in info,
 * this function does nothing.
 *
 * @restriction Available in user mode and supervisor processes, not in interrupt processes.
 *
 * @field scope The set of processes for which to store information.
 * Valid values are RMD_SCOPE_SEGMENT, RMD_SCOPE_BLOCK and RMD_SCOPE_PROCESS.
 * @field info The kind of information to store.
 *
 * @return 0 if successful.
 *
 * @seealso RmdScopeValue
 * @seealso RmdInfoValue
 */
RmdStatus rmd_dump_signal_queue_info(RmdScope scope, RmdInfo info);

/**
 * @function rmd_dump_pool_info
 * @brief Store pool information in the current dump.
 *
 * @long If none of the relevant bits (RMD_INFO_POOL_INFO) are set in info,
 * this function does nothing.
 *
 * @restriction Available in user mode and supervisor processes, not in interrupt processes.
 *
 * @field scope The set of processes for which to store information.
 * Valid values are RMD_SCOPE_SEGMENT, RMD_SCOPE_BLOCK and RMD_SCOPE_PROCESS.
 * @field info The kind of information to store.
 *
 * @return 0 if successful.
 *
 * @seealso RmdScopeValue
 * @seealso RmdInfoValue
 */
RmdStatus rmd_dump_pool_info(RmdScope scope, RmdInfo info);

/**
 * @function rmd_dump_pool_signals_info
 * @brief Store signal information in the current dump.
 *
 * @long If none of the relevant bits (RMD_INFO_POOL_SIGNALS,
 * RMD_INFO_POOL_SIGNALS_OWNED) are set in info, this function does nothing.
 *
 * @restriction Available in user mode and supervisor processes, not in interrupt processes.
 *
 * @field scope The set of processes for which to store information.
 * Valid values are RMD_SCOPE_SEGMENT, RMD_SCOPE_BLOCK and RMD_SCOPE_PROCESS.
 * @field info The kind of information to store.
 *
 * @return 0 if successful.
 *
 * @seealso RmdScopeValue
 * @seealso RmdInfoValue
 */
RmdStatus rmd_dump_pool_signals_info(RmdScope scope, RmdInfo info);

/**
 * @private
 * @function rmd_dump_sys_param_info
 * @brief Store system parameter information in the current dump.
 *
 * @long If none of the relevant bits (RMD_INFO_SYS_PARAM) are set in info,
 * this function does nothing.
 *
 * @restriction Available in user mode and supervisor processes, not in interrupt processes.
 * @break
 * Due to limitations in the system parameter implementation
 * this function currently does nothing.
 *
 * @field scope Should be RMD_SCOPE_SYSTEM.
 * @field info The kind of information to store.
 *
 * @return 0 if successful.
 *
 * @seealso RmdScopeValue
 * @seealso RmdInfoValue
 */
RmdStatus rmd_dump_sys_param_info(RmdScope scope, RmdInfo info);

/**
 * @function rmd_dump_registers_info
 * @brief Store register information in the current dump.
 *
 * @long If none of the relevant bits (RMD_INFO_REGISTERS) are set in info,
 * this function does nothing.
 *
 * @restriction Available in user mode and supervisor processes, not in interrupt processes.
 *
 * @field scope The set of processes for which to store information.
 * Valid values are RMD_SCOPE_SEGMENT, RMD_SCOPE_BLOCK and RMD_SCOPE_PROCESS.
 * @field info The kind of information to store.
 * @field regdump Optional set of registers for current process.
 *
 * @return 0 if successful.
 *
 * @seealso RmdScopeValue
 * @seealso RmdInfoValue
 */
RmdStatus rmd_dump_registers_info(RmdScope scope, RmdInfo info,
                                   struct cpu_regdump *regdump);

/**
 * @function rmd_dump_program_info
 * @brief Store program information in the current dump.
 *
 * @long If none of the relevant bits (RMD_INFO_PROGRAM_INFO) are set in info,
 * this function does nothing.
 *
 * @restriction Available in user mode and supervisor processes, not in interrupt processes.
 *
 * @field scope The set of processes for which to store information.
 * Valid values are RMD_SCOPE_SEGMENT, RMD_SCOPE_BLOCK and RMD_SCOPE_PROCESS.
 * @field info The kind of information to store.
 *
 * @return 0 if successful.
 *
 * @seealso RmdScopeValue
 * @seealso RmdInfoValue
 */
RmdStatus rmd_dump_program_info(RmdScope scope, RmdInfo info);

/**
 * @function rmd_dump_trace_info
 * @brief Store event trace information in the current dump.
 *
 * @long If none of the relevant bits (RMD_INFO_TRACE) are set in info,
 * this function does nothing.
 *
 * @restriction Available in user mode and supervisor processes, not in interrupt processes.
 *
 * @field scope Should be RMD_SCOPE_SYSTEM.
 * @field info The kind of information to store.
 *
 * @return 0 if successful.
 *
 * @seealso RmdScopeValue
 * @seealso RmdInfoValue
 */
RmdStatus rmd_dump_trace_info(RmdScope scope, RmdInfo info);

/**
 * @function rmd_dump_load_module_info
 * @brief Store load module information in the current dump.
 *
 * @long If none of the relevant bits (RMD_INFO_LOAD_MODULE_INFO) are set in
 * info, this function does nothing.
 *
 * @restriction Available in user mode and supervisor processes, not in interrupt processes.
 *
 * @field scope The set of processes for which to store information.
 * Valid values are RMD_SCOPE_SEGMENT, RMD_SCOPE_BLOCK and RMD_SCOPE_PROCESS.
 * @field info The kind of information to store.
 *
 * @return 0 if successful.
 *
 * @seealso RmdScopeValue
 * @seealso RmdInfoValue
 */
RmdStatus rmd_dump_load_module_info(RmdScope scope, RmdInfo info);

/**
 * @function rmd_dump_pool_memory
 * @brief Store pool memory in the current dump.
 *
 * @long If none of the relevant bits (RMD_MEMORY_POOL, RMD_MEMORY_STACKS,
 * RMD_MEMORY_STACKS_OWNED, RMD_MEMORY_STACKS_USED, RMD_MEMORY_STACKS_TOP,
 * RMD_MEMORY_SIGNALS, RMD_MEMORY_SIGNALS_OWNED, RMD_MEMORY_SIGNALS_QUEUED,
 * or RMD_MEMORY_SIGNALS_SENT) are set in memory, this function does nothing.
 *
 * @restriction Available in user mode and supervisor processes, not in interrupt processes.
 *
 * @field scope The set of processes for which to store information.
 * Valid values are RMD_SCOPE_SEGMENT, RMD_SCOPE_BLOCK and RMD_SCOPE_PROCESS.
 * @field memory The kind of memory to store.
 * @field regdump Registers for current process or NULL. Used only for
 * RMD_MEMORY_STACKS_USED and RMD_MEMORY_STACKS_TOP.
 *
 * @return 0 if successful.
 *
 * @seealso RmdScopeValue
 * @seealso RmdMemoryValue
 */
RmdStatus rmd_dump_pool_memory(RmdScope scope, RmdMemory memory,
                               struct cpu_regdump *regdump);

/**
 * @function rmd_dump_heap_memory
 * @brief Store heap memory in the current dump.
 *
 * @long If none of the relevant bits (RMD_MEMORY_HEAP, RMD_MEMORY_HEAP_BUFFERS
 * or RMD_MEMORY_HEAP_BUFFERS_OWNED) are set in memory, this function
 * does nothing.
 *
 * @restriction Available in user mode and supervisor processes, not in interrupt processes.
 *
 * @field scope The set of processes for which to store information.
 * Valid values are RMD_SCOPE_SEGMENT, RMD_SCOPE_BLOCK and RMD_SCOPE_PROCESS.
 * @field memory The kind of memory to store.
 *
 * @return 0 if successful.
 *
 * @seealso RmdScopeValue
 * @seealso RmdMemoryValue
 */
RmdStatus rmd_dump_heap_memory(RmdScope scope, RmdMemory memory);

/**
 * @function rmd_dump_program_memory
 * @brief Store program memory in the current dump.
 *
 * @long If none of the relevant bits (RMD_MEMORY_DATA, RMD_MEMORY_TEXT)
 * are set in memory, this function does nothing.
 *
 * @restriction Available in user mode and supervisor processes, not in interrupt processes.
 *
 * @field scope The set of processes for which to store information.
 * Valid values are RMD_SCOPE_SEGMENT, RMD_SCOPE_BLOCK and RMD_SCOPE_PROCESS.
 * @field memory The kind of memory to store.
 *
 * @return 0 if successful.
 *
 * @seealso RmdScopeValue
 * @seealso RmdMemoryValue
 */
RmdStatus rmd_dump_program_memory(RmdScope scope, RmdMemory memory);

/**
 * @function rmd_set_property
 * @brief Set's RMD property value.
 *
 * @long This function set's the rmd property value, the property value
 * controls the amount of information to be dumped in the scope. 
 * See RmdProperty for supported properties 
 *
 * @restriction Available in user mode and supervisor processes, not in interrupt processes.
 *
 * @field property The RmdProperty type constant for which value to be set
 * @field value The value of the property
 *
 * @return 0 if successful.
 *
 * @seealso RmdProperty
 * @seealso RmdStatus
 */
RmdStatus rmd_set_property(RmdProperty property, uint32_t value);

/**
 * @function rmd_end_of_dump
 * @brief Dumps "end of dump" text block with appropriate status.
 *
 * @long This function dumps "end of dump" text block with appropriate status.
 * This function should be called at the end of the error handler before "rmd_stop_dump" function.
 *
 * @restriction Available in user mode and supervisor processes, not in interrupt processes.
 *
 * @field rmd_dump_status
 * if status = 0, "End of dump - successful" text will be dumped, else
 * "End of dump - incomplete" text will be dumped.
 *
 * @return RmdStatus
 * return == 0, if "end of dump" text block is dumped successfully with appropriate status.
 * return != 0, on failure
 *
 * @seealso RmdStatus
 */
RmdStatus rmd_end_of_dump(uint32_t rmd_dump_status);

#ifdef __cplusplus
} /* extern "C" */
#endif

#endif
