/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2007 - 2009 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */
/**
 * @toc OSE:
 * @file ose_spi/rmm.h
 * @brief
 * Run-Mode Monitor (RMM) system programmer API.
 * @long
 * This file contains declarations to be included in system applications
 * that will use certain run-mode monitor (RMM) functions.
 * @long
 * The main interface to the RMM is its signal interface, see monitor.sig,
 * but the system calls here provide features which can not be done
 * through signals.
 * @long
 * These functions are used to define user callback handlers and to
 * selectively enable request signals in the RMM signal interface, which
 * can be initially disabled by configuration. The initial
 * @link "monitor.sig" "MonitorInterfaceRequest" is always enabled.
 *
 * @long The overall CPU load measurment of the RMM supports a sampled
 * flavor as an alternative to the default context switch based measurement.
 * When in sampled mode there are some limitations, firstly the load from
 * interrupt processes is not included and secondly the priority threshold
 * can not be set to include background processes.
 *
 * @long See more about RMM and configuration in the "RMM" chapter of
 * the "OSE Core User's Guide".
 *
 * @long The client signal interface to access the RMM is covered in the
 * different monitor signal header files.
 *
 * @seealso ose_spi/monitor.sig
 */
#ifndef _RMM_H
#define _RMM_H

#include "stdint.h"
#include "ose.h"

#ifdef __cplusplus
extern "C" {
#endif

/*
** Prototypes.
*/

/**
 * @function ose_rmm_set_user_action_handler
 *
 * @brief Register a callback function for user actions to events.
 * @long
 * This function will provide the RMM
 * with an address to a user specified callback function.
 * This user specified function will be called from the RMM if an event
 * action point is activated with user action - see
 * MONITOR_ACTION_USER in enum @link "monitor_events.sig" "MonitorActionValues".
 * @long
 * The user specified function can be called from different
 * and unusual contexts. This needs to be taken into consideration
 * when it is implemented. It can not make system calls. It must
 * either be reentrant or disable (and enable) interrupts.
 * @long
 * The handler is not allowed to change the event or event data
 * and the event pointers in the arguments are only valid during
 * the call.
 * @long
 * The signalnumbers that can be passed to the handler are:
 * @tab MONITOR_GET_ALLOC_TRACE_REPLY
 * @tab MONITOR_GET_FREE_BUF_TRACE_REPLY @break
 * @tab MONITOR_GET_SEND_TRACE_REPLY
 * @tab MONITOR_GET_RECEIVE_TRACE_REPLY @break
 * @tab MONITOR_GET_USER_TRACE_REPLY
 * @tab MONITOR_GET_ERROR_TRACE_REPLY @break
 * @tab MONITOR_GET_SWAP_TRACE_REPLY
 * @tab MONITOR_GET_CREATE_TRACE_REPLY @break
 * @tab MONITOR_GET_KILL_TRACE_REPLY
 * @tab MONITOR_GET_BIND_TRACE_REPLY @break
 * @tab MONITOR_GET_PROCESS_TRACE_REPLY @break
 * @long
 * To be compatible with future releases of OSE, the handler
 * must however be prepared to handle unknown signal numbers.
 *
 * @input handler The handler function, called with the indicated arguments.
 * @field handler.event_report Information about the event - excluding the event data.
 * @field handler.size      Size of the event.
 * @field handler.data      Pointer to the signal data in send-event,
 *                          receive-event or user-event.
 * @field handler.data_size Size of the data.
 *
 * @restriction
 * It is only possible to register one (1) user defined handler.
 * The registered function must be linked directly to the kernel.
 * One can not unregister the user action.
 * The registered function can not make system calls.
 *
 * @seealso MonitorActionValues
 *
 * @example
 * #include "ose_spi/rmm.h"
 *
 * void user_action_func(union SIGNAL *event_report, uint32_t size,
 *                       void *sig_data, OSBUFSIZE data_size)
 * {
 *    switch(event_report->sigNo)
 *    {
 *       case MONITOR_GET_ALLOC_TRACE_REPLY:
 *	 {
 *          temp1 = event_report->AllocFreeBufInfo.signalNumber;
 *          temp2 = event_report->AllocFreeBufInfo.signalSize;
 *	    /CODE/
 *       }
 *       break;
 *
 *       case MONITOR_GET_SEND_TRACE_REPLY:
 *       {
 *          temp1 = event_report->SendReceiveInfo.signalNumber;
 *          temp2 = event_report->SendReceiveInfo.signalSize;
 *          /CODE/
 *       }
 *       break;
 *
 *       case xxx:
 *       /CODE/
 *       break;
 *
 *       default:
 *	 /CODE/
 *	 break;
 *    }
 * }
 *
 * OS_PROCESS(my_process)
 * {
 *   ose_rmm_set_user_action_handler(user_action_func);
 *
 *   for(;;)
 *     {
 *         /Set any event with action "MONITOR_ACTION_USER"/
 *         /The event will result in the RMM calling /
 *         /user_action_func/
 *     }
 * }
 */

void ose_rmm_set_user_action_handler(void (*handler)(union SIGNAL *event_report,
                                                     uint32_t size,
                                                     void *data,
                                                     OSBUFSIZE data_size));


/**
 * @function ose_rmm_set_user_report_handler
 *
 * @brief Register a callback function for user reports.
 *
 * @long
 * This function will provide the RMM
 * with an address to a user specified callback function.
 * This user specified function will be called from the RMM just
 * before a profiling sample interval for a user report is closed.
 * This is done for all enabled user reports. Compare
 * @link "monitor_cpu.sig" "MonitorSetUserReportsEnabledRequest"
 * in monitor_cpu.sig.
 *
 * @input handler The handler function, called with the indicated argument.
 * @field handler.reportno The report class identification number to handle.
 *
 * @restriction
 * It is only possible to register one(1) user defined handler.
 * The registered function must be linked directly to the kernel.
 * One can not unregister the user report handler.
 *
 * @seealso MonitorSetUserReportsEnabledRequest
 * @seealso ose_create_report
 * @seealso ose_set_report_val
 *
 * @example
 * #include "ose_spi/rmm.h"
 *
 * void user_rpt_func(SIGSELECT reportno)
 * {
 *    switch(reportno)
 *    {
 *       case HWREGA:
 *          if (hwrega_trig)
 *          {
 *             ose_set_report_val(HWREGA, 1, 10);
 *             ose_set_report_val(HWREGA, 2, 10);
 *          }
 *          break;
 *       case xxx:
 *          /CODE/
 *          break;
 *
 *       default:
 *          /CODE/
 *          break;
 *    }
 * }
 *
 * OS_PROCESS(my_process)
 * {
 *    // Code //
 *    ose_create_report(HWREGA, &hwrega_trig, multiple, continuous, maxmin);
 *    ose_rmm_set_user_report_handler(user_rpt_func);
 *
 *    for(;;)
 *    {
 *       hw_value = get_hwrega_value();
 *
 *       // Code //
 *
 *       /The ose_create_report will create a report in the RMM./
 *       /At the end of a sample interval (just before closing)/
 *       /for any enabled user report, user_rpt_func will be called from the RMM./
 *    }
 * }
 */
void ose_rmm_set_user_report_handler(void (*handler)(SIGSELECT reportno));

/**
 * @function ose_rmm_set_req_enabled
 *
 * @brief Control the functionality of the monitor.
 *
 * @long This function is used to enable/disable request signals in the signal
 * interface. Interface signals that are disabled are not replied to (the request
 * signal is silently freed) and the disabled signal numbers do not show up in the
 * list reported in @link "monitor.sig" "MonitorInterfaceReply" in monitor.sig.
 *
 * @long If the system parameter krn/monitor/interface was used to disable all
 * signals to the monitor, this function can be used to selectively enable
 * request signals.
 *
 * @input signo Request signal number.
 *
 * @input enabled Boolean parameter indicating if the specified signal should be
 * enabled(1) or disabled(0).
 *
 * @restriction Available in supervisor processes with user ID 0, 
 * not in interrupt processes.
 *
 * @example
 * // Disable MONITOR_GET_PROCESS_INFO_REQUEST signal. //
 * ose_rmm_set_req_enabled(MONITOR_GET_PROCESS_INFO_REQUEST, 0);
 */
void ose_rmm_set_req_enabled(SIGSELECT signo, int enabled);


/**
 * @function ose_rmm_tracepoint_init
 *
 * @brief Initialize tracepoint support
 *
 * @long Before the signals in the signal interface monitor_tracepoint.sig can
 * be used, this function shall be called.
 *
 * @restriction Available in supervisor processes, not in interrupt processes.
 *
 * @example
 * ose_rmm_tracepoint_init();
 */
void ose_rmm_tracepoint_init(void);


/**
 * @function ose_rmm_set_hw_trace_handlers 
 *
 * @brief Register user defined functions that enable and
 * disable hardware trace.
 *
 * @long Register user defined functions that are called from 
 * RMM when action enable or disable hardware trace is set.
 * The purpose with the user defined functions is to enable or 
 * disable the hardware trace.
 * 
 * @input enbl_handler Function pointer to a user defined function.
 * @input dsbl_handler Function pointer to a user defined function.
 *
 * @return Zero if ok, otherwise 1.
 *
 */
uint32_t ose_rmm_set_hw_trace_handlers(void (*enbl_handler)(void), 
                                       void (*dsbl_handler)(void));

/**
 * @function ose_rmm_init_unique_number
 *
 * @brief Initiate the unique number sequence
 *
 * @long The RMM provide a service for generating unique id numbers. The number
 * sequence will survive a restart if the content of the beginning of the
 * system pool survives a restart and in that case this function do not need to
 * be called. Hence this function could be called early in a cold start
 * scenario.
 * 
 * @long The Run Mode Dumper (RMD) uses this service to generate dump id for
 * each RMD dump.
 * 
 * @input start_number The number that the number sequence will start at.
 *
 * @restriction Available to supervisor processes only. This function is
 * experimentally and can change or be removed in coming OSE5 releases, hence
 * any use of it should contain a plan for update.
 *
 */
void ose_rmm_init_unique_number(uint32_t start_number);

#ifdef __cplusplus
}
#endif

#endif /* _RMM_H */
