/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2010 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

/**
 * @toc OSE:
 * @file ose_spi/rofs_init.h
 * @brief Interface for read-only file system support (ROFS)
 * @long  This interface is used to initialize a ROFS volume manager.
 *        ROFS volumes are read-only file system volumes that can reside in an
 *        ELF file.
 * @long  For information about ROFS, refer to OSE5 Core User's Guide,
 *        chapter "Read-Only File System (ROFS)".
 */

#ifndef _ROFS_INIT_H
#define _ROFS_INIT_H

#ifdef __cplusplus
extern "C"
{
#endif

/**
 * @function ose_rofs_start_handler0
 *
 * @brief Prepare for ROFS mount
 *
 * @long  This function must be run as a START_HANDLER0 in krn.con.
 *        It will configure the memory manager so that the ROFS volume will
 *        not be destroyed before ose_rofs_mount_embedded() has a chance to run.
 * @long  Ramlog is used for trace and error messages.
 */
void ose_rofs_start_handler0(void);

/**
 * @function ose_rofs_mount_embedded
 *
 * @brief Mount ROFS embedded volume
 *
 * @long  Mount a ROFS volume which is embedded into the binary image by using
 *        the rofs_insert host tool.
 * @long  This function requires that the function
 *        @link "" "ose_rofs_start_handler0"() has been called as documented
 *        for that function.
 * @long  On a successful return from this function, a ROFS volume with the
 *        given mount point will be registered in the FSS.
 * @long  Ramlog is used for trace and error messages.
 *
 * @input mount_point A full path to where the volume is to be mounted. Must be
 *                    in the root directory, i.e. have the syntax "/<name>",
 *                    where name must be at least one character and not contain
 *                    any '/' characters.
 *
 * @return Returns 0 on success, or an errno value indicating the error.
 *
 * @error @link "errno.h" "EIO" No volume was found by
 *          @link "" "ose_rofs_start_handler0" or if ose_rofs_start_handler0
 *          wasn't called in time.
 *
 * @error @link "errno.h" "EFAULT" mount_point parameter was a NULL pointer.
 *
 * @error @link "errno.h" "EINVAL" mount_point parameter is not a valid
 *          volume name.
 *
 * @error @link "errno.h" "ENOMEM" Out of memory.
 *
 * @error @link "errno.h" "ESTALE" Process ose_fss could not be found or was
 *          killed, or volume manager died for some reason, or registration
 *          to the FSS failed.
 *
 * @example
 * #include "string.h"
 * #include "errno.h"
 * #include "rofs_init.h"
 *
 * void mymount(void)
 * {
 *    int result = ose_rofs_mount_embedded("/romfs");
 *    if (result != 0)
 *    {
 *       fprintf(stderr, "Failed mounting ROFS: %s\n", strerror(errno));
 *       exit(EXIT_FAILURE);
 *    }
 * }
 */
int ose_rofs_mount_embedded(const char *mount_point);

#ifdef __cplusplus
}
#endif
#endif
