/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2011 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */
/**
 * @public
 * @toc OSE:
 * @file ose_spi/sysparam.h
 *
 * @brief Public declarations for the system parameter interface.
 *
 * @long
 * This file contains prototypes and defines for the system
 * parameter interface. This interface is primarily used for
 * reading run-time configuration parameters at boot time or when load
 * modules are started. It handles system parameters in core or load modules
 * or persistent system or application parameters stored in flash.
 * The sysparam service is a client to the @link "ose_spi/boardparam.h"
 * "ose_spi/boardparam.h" interface, handling hardware dependent storage.
 *
 * @long A board parameter overrides a module parameter with the same
 * name, unless safe mode is selected, see @link "" "sysparam_set_safe_mode".
 */

#ifndef OSE_SYSPARAM_H_
#define OSE_SYSPARAM_H_

#include "stddef.h"
#include "osetypes.h"

#ifdef __cplusplus
extern "C" {
#endif

/**
 * @macro SYSPARAM_MAX_NAME_LENGTH
 * @brief The SYSPARAM_MAX_NAME_LENGTH is the maximum length in bytes
 *        allowed for the system parameter names excluding string
 *        terminator.
 * @long  The current value of SYSPARAM_MAX_NAME_LENGTH is 127.
 */
#define SYSPARAM_MAX_NAME_LENGTH 127

/**
 * @macro SYSPARAM_MAX_VALUE_LENGTH
 * @brief The SYSPARAM_MAX_VALUE_LENGTH is the maximum length in bytes
 *        allowed for the system parameter values excluding string
 *        terminator.
 * @long  The current value of SYSPARAM_MAX_VALUE_LENGTH is 255.
 */
#define SYSPARAM_MAX_VALUE_LENGTH 255

/**
 * @function sysparam_read
 *
 * @brief This function is used to retrieve the string that is the value of
 *        a configuration parameter.
 * @long  The value is copied to the submitted buffer. If the buffer size
 *        is equal to or smaller than the actual value string length, only
 *        "buffer_size - 1" characters are copied. In any case, the function
 *        always null terminates what's copied to the buffer.
 * @input  name        Specifies the name of the parameter to be read.
 * @input buffer       Address of the buffer where the value of the parameter
 *                     should be written.
 * @output buffer      The buffer content is updated.
 * @input buffer_size  Size of buffer.
 * @return The actual length of the parameter value, (regardless if the
 *         buffer was too small to hold it and the value had to be
 *         truncated), without counting the trailing NULL. If the parameter
 *         was not found, the function returns 0.
 * @seealso default_restrictions_calling_functions
 */
size_t sysparam_read(const char *name, char *buffer, size_t buffer_size);


/**
 * @function sysparam_list_next
 *
 * @brief    This function supports iterative traversal of the configuration
 *	        parameter namespace.
 * @long     The function takes the string submitted as 'str' and searches
 *           for the parameter name that would lexicographically follow.
 *           The 'str' and the 'buffer' parameters may be pointers to the
 *           same buffer.
 * @input  str          Specifies the starting point for the search as
 *                      described  above. It may be a NULL pointer, in
 *                      which case the (lexicographically) first
 *                      configuration parameter is searched for.
 *                      It may be the leading part of a name - the first
 *                      matching name is found, otherwise the next name
 *                      lexicographically.
 * @input buffer	   Pointer to the buffer where this function will
 *			   write the found parameter name, up to 'buffer_size'
 *                         characters. May be NULL, in which case no copying
 *			   will take place (but the found parameter name length
 *			   will still be returned).
 * @output buffer	   The buffer content is updated.
 * @input  buffer_size  Size of buffer. Must be 0 if 'buffer' is NULL.
 * @return The function returns the length of the found parameter name,
 *	      regardless if the string copied to 'buffer' was truncated
 *	      (because it was too small to hold it) or 'buffer' was NULL.
 *	      If no parameter name lexicographically following 'str'
 *	      could be found, the function returns zero.
 * @seealso default_restrictions_calling_functions
 * @example
 * This example would print all parameters.
 *
 * char name[SYSPARAM_MAX_NAME_LENGTH + 1];
 * size_t name_len = sysparam_list_next("", name, sizeof name);
 * while (name_len > 0)
 *   {
 *     char value[VALUE_BUFSIZE];
 *     size_t val_len = sysparam_read(name, value, sizeof value);
 *     if ((val_len > 0) && (val_len < sizeof value))
 *       {
 *         printf("%s=\"%s\"\n", name, value);
 *       }
 *     name_len = sysparam_list_next(name, name, sizeof name);
 *   }
 */
size_t sysparam_list_next(const char *str, char *buffer,size_t buffer_size);

/**
 * @function sysparam_write
 *
 * @brief    This function is used to add, modify or delete one board
 *           parameter.
 * @long     Its usage should normally be restricted to system servicing
 *           code, or boot monitor, since, according to system specific
 *           implementation, this function may have big impact in real-time
 *           characteristics.
 *
 * @long     Warning: Must be used with care, since according to how
 *                         the parameter storage is implemented in the
 *                         actual system, calling this function may have
 *                         an impact on real-time characteristics,
 *                         etc.
 * @input  name  The name of the addressed board parameter.
 * @input  value The new value for the addressed board parameter (may
 *		    not be NULL). If it is a pointer to a null-character,
 *		    the parameter will be deleted.
 * @return The function returns 0 for success, != 0 otherwise.
 * @seealso default_restrictions_calling_functions
 */
int sysparam_write(const char *name, const char *value);

/**
 * @type enumeration
 * @brief Type used for translating attribute names to values.
 * @field name The name of the value
 * @field value The corresponding value
 * @seealso sysparam_readattr_enum
 */
struct enumeration
{
  const char *name;
  long value;
};

/**
 * @function sysparam_translate_symbol
 *
 * @brief    Optional user-defined customizable function, implicitly used by
 *           sysparam functions to resolve "@atsymbol" in system parameters.
 *
 * @long     If this function is defined at link time, it will be used at
 *           run-time by the internal system parameter reading functions, to
 *           resolve parameter values which are symbols beginning with "@at",
 *           like for example @atTEXT_BASE, @atTEXT_SIZE.
 * @long     If not present at link-time, the functionality will be ignored
 *           and parameter strings are used as they are.
 *
 * @input  symbol  Name of symbol to resolve, example: "TEXT_SIZE",
 *                 (without the leading @at).
 * @input  value   Pointer to storage for placing the resolved value of symbol,
 *                 if found. The value shall be of correct type.
 * @output value   The content is updated.
 * @return Returns True on successful translation, False otherwise.
 * @seealso default_restrictions_calling_functions
 * @example
 * // In this example the sysparam_translate_symbol function is needed to
 * // translate the parameters to the run-time start and size of the "text"
 * // area of the executable image in memory. The values returned are
 * // of the type long.
 * // The definition line below must be one line (broken only in manual)
 *
 * krn/region/text=base:@TEXT_BASE size:@TEXT_SIZE perm:su_rx_usr_rx
 *      cache:copy_back mem:speculative_access
 */
Boolean sysparam_translate_symbol(const char *symbol, long *value);


/**
 * @function sysparam_readattr_long
 *
 * @brief    Reads a parameter and interprets it as a numerical value.
 * @input  param_name  Name of parameter.
 * @input  attr        Name of attribute. Set to NULL for system parameters 
 *                     without attributes (e.g. myconfig=value).
 * @input  value       Pointer to where to store the result.
 * @output value       The content is updated.
 *
 * @return
 *   @tab EOK     @tab The value was read successfully. @break
 *   @tab EINVAL  @tab Symbol translation failed,
 *                     or the attribute is not numeric. @break
 *   @tab ENOENT  @tab The parameter attribute does not exist. @break
 *   @tab E2BIG   @tab The attribute value string has more than
 *                     @link "" "SYSPARAM_MAX_VALUE_LENGTH" characters. @break
 *
 * @seealso default_restrictions_calling_functions
 * @example
 * Assume having the following system parameter:
 *   myconfig=size:3
 *
 * Then variable "value" will have the value 3 after executing the code
 * below:
 *
 * int result;
 * long value;
 *
 * result = sysparam_readattr_long("myconfig", "size", &value);
 */
int sysparam_readattr_long(const char *param_name, const char *attr,
                           long *value);

/**
 * @function sysparam_readattr_longlong
 *
 * @brief    Reads a parameter and interprets it as a numerical value.
 * @input  param_name  Name of parameter.
 * @input  attr        Name of attribute. Set to NULL for system parameters 
 *                     without attributes (e.g. myconfig=value).
 * @input  value       Pointer to where to store the result.
 * @output value       The content is updated.
 *
 * @return
 *   @tab EOK     @tab The value was read successfully. @break
 *   @tab EINVAL  @tab Symbol translation failed,
 *                     or the attribute is not numeric. @break
 *   @tab ENOENT  @tab The parameter attribute does not exist. @break
 *   @tab E2BIG   @tab The attribute value string has more than
 *                     @link "" "SYSPARAM_MAX_VALUE_LENGTH" characters. @break
 *
 * @seealso default_restrictions_calling_functions
 * @example
 * Assume having the following system parameter:
 *   myconfig=base:0x140000000
 *
 * Then variable "value" will have the value after executing the code
 * below:
 *
 * int result;
 * long long int value;
 *
 * result = sysparam_readattr_longlong("myconfig", "base", &value);
 */
int sysparam_readattr_longlong(const char *param_name, const char *attr,
                               long long int *value);

/**
 * @function sysparam_readattr_string
 *
 * @brief    Reads a parameter and interprets it as a string.
 * @input  param_name  Name of parameter.
 * @input  attr        Name of attribute. Set to NULL for system parameters 
 *                     without attributes (e.g. myconfig=value).
 * @input  value       Pointer to where to store the result.
 * @output value       The content is updated.
 * @input  bufsize     Size of result storage area, including space for
 *                     terminating NULL.
 *
 * @return
 *   @tab EOK     @tab The string was read successfully. @break
 *   @tab ENOSPC  @tab The provided buffer is too small to fit the string. @break
 *   @tab EINVAL  @tab Invalid buffer size. @break
 *   @tab ENOENT  @tab The parameter attribute does not exist. @break
 *   @tab E2BIG   @tab The attribute value string has more than
 *                     @link "" "SYSPARAM_MAX_VALUE_LENGTH" characters. @break
 * 
 * @seealso default_restrictions_calling_functions
 * @example
 * Assume having the following system parameter:
 *   myconfig=size:big
 *
 * Then variable "value" will contain the value "big" after executing
 * the code below:
 *
 * int result;
 * char value[15];
 *
 * result = sysparam_readattr_string("myconfig", "size", value,
 *                                   sizeof (value));
 */
int sysparam_readattr_string(const char *param_name, const char *attr,
                             char *value,
                             size_t bufsize);

/**
 * @function sysparam_readattr_enum
 *
 * @brief    Reads a parameter and translates the attribute to a value.
 * @input  param_name  Name of parameter.
 * @input  attr        Name of attribute. Set to NULL for system parameters 
 *                     without attributes (e.g. myconfig=value).
 * @input  enumeration Pointer to first element in an array used to
 *                     translate attribute name to value. The array is
 *                     terminated by an entry with a null string pointer.
 * @input  value       Pointer to where to store the result.
 * @output value       The content is updated.
 *
 * @return
 *   @tab EOK     @tab The enumaration was successfully read. @break
 *   @tab EINVAL  @tab Invalid enumeration value read. @break
 *   @tab ENOENT  @tab The parameter attribute does not exist. @break
 *   @tab E2BIG   @tab The attribute value string has more than
 *                     @link "" "SYSPARAM_MAX_VALUE_LENGTH" characters. @break
 *
 * @seealso default_restrictions_calling_functions
 * @example
 * Assume having the following system parameter:
 *   myconfig=size:big
 *
 * Then variable "value" will have the value 3 after executing the code
 * below:
 *
 * const struct enumeration enumeration =
 *   {{"big", 3}, {"medium", 2}, {"small", 1}, {NULL, 0}};
 * int result;
 * long value;
 *
 * result = sysparam_readattr_enum("myconfig", "size", enumeration, &value);
 */
int sysparam_readattr_enum(const char *param_name, const char *attr,
                           const struct enumeration *enumeration,
			   long *value);
/**
 * @function sysparam_readattr_bool
 *
 * @brief  Reads a parameter and translates the attribute to a boolean value.
 *         The allowed parameter values are: true/false, yes/no, on/off,
 *         enabled/disabled, 1/0.
 * @input  param_name  Name of parameter.
 * @input  attr        Name of attribute. Set to NULL for system parameters 
 *                     without attributes (e.g. myconfig=value).
 * @input  value       Pointer to where to store the result.
 * @output value       The content is updated.
 *
 * @return
 *   @tab EOK     @tab The boolean was successfully read. @break
 *   @tab EINVAL  @tab Invalid boolean value read. @break
 *   @tab ENOENT  @tab The parameter attribute does not exist. @break
 *   @tab E2BIG   @tab The attribute value string has more than
 *                     @link "" "SYSPARAM_MAX_VALUE_LENGTH" characters. @break
 *
 * @seealso default_restrictions_calling_functions
 * @example
 * Assume having the following system parameter:
 *   myconfig=verbose:true
 *
 * Then variable "value" will have the value True after executing the code
 * below:
 *
 * int result;
 * Boolean verbose;
 *
 * result = sysparam_readattr_bool("myconfig", "verbose", &verbose);
 */
int sysparam_readattr_bool(const char *param_name, const char *attr,
                           Boolean *value);

/**
 * @function sysparam_enter_safe_mode
 *
 * @brief Called by system startup code before any other call to this module
 * if only load module parameters are to be read, and board parameters ignored.
 * This can be wanted for example if the system has entered cyclic reboot
 * because of a fatal configuration error in a board parameter. Same as
 * (void) sysparam_set_safe_mode(1).
 * @seealso sysparam_set_safe_mode
 * @seealso default_restrictions_calling_functions
 */
void sysparam_enter_safe_mode(void);

/**
 * @function sysparam_set_safe_mode
 *
 * @brief This function instructs sysparam whether to select safe mode (in
 * which it ignores board parameters and only reads module parameters).
 * @input new_mode     True for safe mode, False for normal mode.
 * @return The old mode.
 * @seealso sysparam_enter_safe_mode
 * @seealso default_restrictions_calling_functions
 */
Boolean sysparam_set_safe_mode(Boolean new_mode);

/**
 * @function sysparam_add_cmds
 *
 * @brief Adds shell commands for the system parameters.
 *
 * @restriction Also available from START_OSE_HOOK2 context.
 *
 * @seealso default_restrictions_calling_functions
 */
void sysparam_add_cmds(void);

#ifdef __cplusplus
}
#endif

#endif /* #ifndef OSE_SYSPARAM_H_ */
