/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2010 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

/**
 * @public
 * @toc OSE:
 * @file ose_spi/thinlib.h
 *
 * @brief     Thin library initialization functions.
 *
 * @long      A thin library can be used in a load module to decrease its size.
 *            Instead of having the full library code in each load module, the
 *            regular library is only linked against the core module. The load
 *            module is linked against the thin version of the same library.
 *            
 * @long      When a load module calls a function from the thin library, the
 *            execution continues transparently in the full library code in the
 *            core module. Before a load module using such a thin library can
 *            be used, the library's corresponding thin initialization function
 *            must be called from the core module. For example,
 *            ose_thinlib_init_crt must be called before the thin CRT can be
 *            used.
 *
 * @long      The first time a function from the thin library is called, the
 *            function will retrieve a pointer to the real function code
 *            residing in the core module, using a bios interface installed by
 *            the initialization function. The pointer is then stored and used
 *            for all subsequent calls to the function.
 *            
 * @long      This header file contains the initialization functions for all
 *            the thin libraries available in OSE.
 *            
 * @long      This file contains declarations to be included in code
 *            written in C.
 *
 * @library libcore_*.a
 */

#ifndef _THINLIB_H
#define _THINLIB_H

#ifdef __cplusplus
extern "C" {
#endif

/**
 * @macro THINLIB_BIOS_NAME
 * @brief The THINLIB_BIOS_NAME is the common bios handle used by the thin
 *        library functionality.
 * @long  The current value of THINLIB_BIOS_NAME is "ose_thinlib"
 */
#define THINLIB_BIOS_NAME "ose_thinlib"
   
/**
 * @type ThinlibEntry
 * @brief Type mapping a function name to its implementation.
 * @field name The name of the function
 * @field function The corresponding function pointer
 */
struct ThinlibEntry
{
   const char *name;
   void (*function)(void);
};
   
/**
 * @function ose_thinlib_init
 *
 * @brief    This function initializes the generic part of the thin library
 *           functionality.
 *
 * @long     The function is called implicitly by all library specific
 *           initialization functions, so there is no need to call it
 *           separately.
 *
 * @restriction This function is only available in the core module.
 */
void ose_thinlib_init(void);   

/**
 * @function ose_thinlib_init_crt
 *
 * @brief    This function initializes the core module side of the thin CRT.
 *
 * @long     The function must be called before any load module using the thin
 *           CRT is used in the system.
 *
 * @restriction This function is only available in the core module. The thin CRT
 *              can only be used from kernel load modules.
 */
void ose_thinlib_init_crt(void);

#ifdef __cplusplus
}
#endif
   
#endif
