/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2013 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */
/**
 * @public
 * @toc OSE:
 * @file ose_spi/validate_user.h
 *
 * @brief Prototypes for user validation and log, command validation and log,
 *        help text validation, command execution logging and user logout log
 *        in the OSE Shell.
 *
 */

#ifndef _VALIDATE_USER_H
#define _VALIDATE_USER_H

#include "ose.h"
#include "osetypes.h"
#include "stddef.h"
#include "shell.h"

#ifdef __cplusplus
extern "C" {
#endif

/**
 * @function validate_user
 *
 * @brief Validation of the username/password combination and log of user login.
 *
 * @long The OSE login command and the ose_login process performs user
 *       authentication before creating the shell process. The login function or
 *       process reads the username and password from its terminal and calls a
 *       user-supplied function, validate_user, to perform the actual validation
 *       of the supplied credentials.
 *
 * @long The function is also called from the OSE FTP-server to verify logins.
 *       The user identity and process to start are not used by the FTP-server.
 *
 * @long The function may also be used to log information about user login.
 *
 * @long When called from the login command or process, it is possible to use
 *       stdio (stdin, stdout, stderr) in the function for additional input
 *       and output to the user (e.g. when using the console format manager
 *       for the serial port or the Telnet port).
 *       When called from the FTP-server (termname="ftp"), it is not possible
 *       to use stdio for additional input and output to the user.
 *
 * @long This is a function that should be implemented by the developer if it
 *       is wanted. If the function is not implemented a default function in
 *       the OSE CRT library is linked into the system. The default function
 *       always returns valid username/password combination (True).
 *
 * @restriction When called from the FTP-server (termname="ftp"), it is not
 *              possible to use stdio for additional input and output to the
 *              user.
 *
 * @input username    Pointer to a string containing the username.
 *
 * @input password    Pointer to a string containing the password.
 *
 * @input terminal    Pointer to a string containing the name of the terminal
 *                    which can be used to identify which terminal type the user
 *                    is logged in over.
 *                    For the serial port this is the name of the volume when
 *                    mounting the console format manager (confm) and serial
 *                    device driver (serdd), e.g."/tty".
 *                    For the Telnet port this is the name of the volume when
 *                    mounting the console format manager for the Telnet port,
 *                    e.g. "/telnet_<addr>:<port>".
 *                    An FTP-server provides "ftp" as the termname.
 *
 * @input cmdline     Pointer to a command line buffer for output. May be NULL.
 *
 * @input cmdline_len The size of the command line buffer in bytes.
 *
 * @output cmdline    Shell command in command line buffer. The shell command is
 *                    run if valid username/password. May be empty.
 *
 * @return If the username/password is valid, the function shall return non-zero
 *         (True). Otherwise it shall return zero (False).
 *
 * @seealso validate_cmd
 * @seealso validate_help
 * @seealso logout_user
 */
Boolean
validate_user(const char *username,
              const char *password,
              const char *terminal,
              char       *cmdline,
              size_t     cmdline_len);

/**
 * @function validate_cmd
 *
 * @brief Validation and log of OSE shell command.
 *
 * @long The function is called from the OSE shell when the user has entered a
 *       shell command (via a serial port, a Telnet port or a rcmd command in a
 *       remote shell) and should validate that the user is allowed to run
 *       the command.
 *
 * @long The function is also called when an application calls the shell
 *       application program interface function
 *       @link "shell.h" "shell_run_cmd"().
 *
 * @long The function may also be used to log information about entered shell
 *       commands.
 *
 * @long The parameter original_cmd_line is the user entered shell command to be
 *       logged.
 *
 * @long The parameters argc, argv, no_env, env, no_fileredir and fileredir are
 *       extracted when parsing the original_cmd_line in OSE shell. These
 *       parameters should be used for validation of the shell command. If there
 *       is an error found when parsing the string the function validate_cmd()
 *       is called with these parameters set to zero or NULL. If there is an
 *       error found when parsing the string the command will not be executed,
 *       regardless of the result returned from validate_cmd().
 *
 * @long This is a function that should be implemented by the developer if it
 *       is wanted. If the function is not implemented a default function in
 *       the OSE CRT library is linked into the system. The default function
 *       always returns valid (True).
 *
 * @restriction It is not possible to use stdio (stdin, stdout, stderr) in the
 *              function for additional input and output to the user.
 *
 * @input original_cmd_line Pointer to a string to be logged containing the
 *                          received shell command line from the user. The
 *                          string contains environment variable definitions,
 *                          command, parameters and redirection definitions. The
 *                          string should not be parsed because the shell
 *                          command syntax may be changed in the future.
 *                          When the function shell_run_cmd() is used the string
 *                          is the received parameter cmd in function
 *                          shell_run_cmd(). cmd does not contain environment
 *                          variable definitions or redirection definitions.
 *                          These are instead fetched from the parameter
 *                          taglist (tags RUNCMD_ENVLIST and RUNCMD_FDARRAY) in
 *                          function shell_run_cmd().
 *
 * @input username          Pointer to a string containing the username.
 *                          NULL if not defined.
 *                          When the function shell_run_cmd() is used the name
 *                          depends on if the environment variable USERNAME is
 *                          set by the calling application.
 *
 * @input termname          Pointer to a string containing the name of the
 *                          terminal which can be used to identify which
 *                          terminal type the user is logged in over.
 *                          For the serial port this is the name of the volume
 *                          when mounting the console format manager (confm) and
 *                          serial device driver (serdd), e.g. "/tty".
 *                          For the Telnet port this is the name of the volume
 *                          when mounting the console format manager for the
 *                          Telnet port, e.g. "/telnet_<addr>:<port>".
 *                          For the remote connection this is the remote
 *                          terminal name with the linkpath added, e.g.
 *                          "master!/tty".
 *                          When the function shell_run_cmd() is used the name
 *                          depends on if the environment variable TERMINAL is
 *                          set by the calling application.
 *                          NULL if not defined.
 *
 * @input argc              Number of extracted arguments when parsing command
 *                          and parameters in the original_cmd_line. The same
 *                          argc will be passed to the command.
 *                          Zero if an error was found when parsing the
 *                          original_cmd_line.
 *
 * @input argv              Pointer to an array of pointers that points to the
 *                          extracted argument strings when parsing command and
 *                          parameters in the original_cmd_line. The same argv
 *                          will be passed to the command. The first pointer in
 *                          the array will point to a string containing the
 *                          command name.
 *                          NULL if an error was found when parsing the
 *                          original_cmd_line.
 *
 * @input no_env            Number of extracted environment variable definitions
 *                          when parsing environment variable definitions in the
 *                          original_cmd_line.
 *                          When the function shell_run_cmd() is used the
 *                          environment variable definitions are fetched from
 *                          the parameter taglist (tag RUNCMD_ENVLIST) in the
 *                          function shell_run_cmd().
 *                          Zero if no definitions.
 *
 * @input env               Pointer to an array of pointers that points to the
 *                          extracted environment variable definitions strings
 *                          when parsing environment variable definitions in the
 *                          original_cmd_line. An environment variable
 *                          definition string has the following format:
 *                          NAME=value.
 *                          When the function shell_run_cmd() is used the
 *                          environment variable definitions are fetched from
 *                          the parameter taglist (tag RUNCMD_ENVLIST) in the
 *                          function shell_run_cmd().
 *                          NULL if no definitions.
 *
 * @input no_fileredir      Number of file redirections in fileredir.
 *                          Zero if no redirections.
 *
 * @input fileredir         Pointer to an array of pointers that points to the
 *                          extracted redirection strings, containing file
 *                          names, when parsing redirection definitions in the
 *                          original_cmd_line.
 *                          When the function shell_run_cmd() is used the
 *                          redirection definitions are fetched from the
 *                          parameter taglist (tag RUNCMD_FDARRAY) in the
 *                          function shell_run_cmd().
 *                          If there is any redirection the fileredir array will
 *                          always include the following pointers:@break
 *                          Index  : Description@break
 *                          0  :  Pointer to file name to redirect stdin
 *                                    from.@break
 *                          1  :  Pointer to file name to redirect stdout
 *                                     to.@break
 *                          2  :  Pointer to file name to redirect (append)
 *                                     stdout to.@break
 *                          3  :  Pointer to file name to redirect stderr
 *                                     to.@break
 *                          4  :  Pointer to file name to redirect (append)
 *                                stderr to.@break
 *                          The pointer to the filename is NULL if redirection
 *                          is not defined.
 *                          fileredir is NULL if there are no redirections.
 *
 * @input pid               Process ID of the shell process where the command is
 *                          entered or the process ID of the process using the
 *                          function shell_run_cmd(). The process ID gives the
 *                          possibility to get information about the process or
 *                          block, e.g. the name of the process or block and
 *                          e.g. environment variables.
 *                          Zero if not defined, e.g. when called from
 *                          ose_shelld (remote command).
 *
 * @output status           Pointer to a pointer to an output string in a heap
 *                          buffer indicating why a validation of a shell
 *                          command is not valid, e.g.:@break
 *                          - "The command is not allowed."@break
 *                          - "The command argument <xxx> is not allowed."@break
 *                          - "The redirection output to <file> is not allowed."
 *                            @break
 *                          The function validate_cmd() places the string in a
 *                          heap buffer allocated with the function
 *                          @link "ose_spi/heapapi.h" "heap_alloc_private"(). The owner
 *                          process of the heap buffer must be the caller of the
 *                          function validate_cmd() which is accomplished if the
 *                          call to heap_alloc_private() is done in the function
 *                          validate_cmd(). The function validate_cmd() returns
 *                          the heap buffer with the string to the caller.
 *                          If the validation is not valid the string is printed
 *                          (to stderr).
 *                          The heap buffer is freed by the caller of the
 *                          function validate_cmd() with a call to
 *                          @link "ose_spi/heapapi.h" "heap_free_private"().
 *                          If NULL is returned no string is printed and no
 *                          buffer is freed.
 *                          status is NULL if no output string is wanted (e.g.
 *                          when the function shell_run_cmd() is used).
 *
 * @return If the shell command is valid (and the user is allowed to run the
 *         command), the function shall return non-zero (True).
 *         Otherwise it shall return zero (False) and an optional string in
 *         status.
 *
 * @seealso validate_user
 * @seealso validate_help
 * @seealso logout_user
 */
Boolean
validate_cmd(const char *original_cmd_line, /* input */
             const char *username,          /* input */
             const char *termname,          /* input */
             U32        argc,               /* input */
             const char **argv,             /* input */
             U32        no_env,             /* input */
             const char **env,              /* input */
             U32        no_fileredir,       /* input */
             const char **fileredir,        /* input */
             PROCESS    pid,                /* input */
             char       **status);          /* output if not valid */

/**
 * @type CompleteCmd
 *
 * @brief A structure that contains fields which store different
 *        essential command parameters. These parameters are defined
 *        in analogy with the validate_cmd function.
 *              
 * @long  The CompleteCmd structure fields are the ones taken from
 *        the validate_cmd hook parameter list. Full information about
 *        them is available in the validate_cmd function documentation.
 * 
 * @long  The parameters argc, argv, no_env, env, no_fileredir, fileredir
 *        are extracted when parsing the original_cmd_line in
 *        OSE shell. These parameters should be used for valiadtion of the 
 *        shell command. If errors are found, these parameters are 
 *        set to NULL or zero.
 * 
 * @input original_cmd_line Pointer to a string to be logged containing
 *                          the received shell command line from the user.
 *                          The string contains environment variable 
 *                          definitions, command, parameters and redirection
 *                          definitions.
 *
 * @input parsed_cmd_line   Pointer to a string that contains the exact
 *                          command string contained in the original_cmd_line,
 *                          but without redirection and environment definitions
 *                          if this case is present.
 *
 * @input username          Pointer to a string containing the username.
 *                          NULL if not defined.
 *
 * @input termname          Pointer to a string containing the name of the
 *                          terminal type the user is logged in over.
 *                                            
 * @input argc              Number of extracted arguments when parsing command
 *                          and parameters in the original_cmd_line.
 *
 * @input argv              Pointer to an array of pointer that points to the
 *                          extracted argument strings when parsing command                          
 *                          and parameters in the original_cmd_line.
 *
 * @input no_env            Number of extracted environment variable definitions
 *                          when parsing environment variable definitions in the
 *                          original_cmd_line.                          
 *
 * @input env               Pointer to an array of pointers that point to the
 *                          extracted environment variable definitions strings 
 *                          when parsing environment variable definitions in the
 *                          original_cmd_line.
 *
 * @input no_fileredir      Number of file redirections in the fileredir.
 *                          Zero if no redirections.
 *
 * @input fileredir         Pointer to an array of pointers that points to the
 *                          extracted redirection strings, containing file
 *                          names, when parsing redirection definitions in the
 *                          original_cmd_line.
 *
 * @input pid               Process ID of the shell process where the command
 *                          entered.
 *
 * @input cmd_exit_stat     Status value returned by the command's entrypoint.
 *
 * @input duration          Field that stores the precise command execution
 *                          duration.
 *                          The value is expressed in microseconds.
 *
 * @seealso validate_cmd                                                 
 */
struct CompleteCmd
{
   const char             *original_cmd_line;
   const char             *parsed_cmd_line;
   const char             *username;
   const char             *termname;
   U32                    argc;
   const char             **argv;
   U32                    no_env;
   const char             **env;
   U32                    no_fileredir;
   const char             **fileredir;
   PROCESS                pid;
   int                    cmd_exit_stat;
   unsigned long          duration;
};

/**
 * @function shell_hook_log_cmd_completion
 *
 * @brief Provides command information and log after each shell command execution.
 *
 * @long The function is called from the OSE shell when the user has entered a
 *       shell command(via a serial port, a Telnet port or a rcmd command in
 *       remote shell) and should log the duration, status and other command
 *       parameters if the user is allowed to run these commands.
 *
 * @long This hook is OPTIONAL, therefore it is was implemented using
 *       weak linking. The #pragma directive makes sure the hook is called
 *       only when the developer wishes that. If it is implemented,
 *       it will call-up the shell_hook_log_cmd_completion() hook after 
 *       each command gets executed and it will print to the log files.
 *
 * @long If the hook is not implemented, then it will be completly ignored
 *       by the compiler.
 *
 * @input cmd_info Pointer to the CompleteCmd structure type. This end-hook 
 *                 only takes a single parameter. The structure CompleteCmd
 *                 was implemented for this hook so that we can avoid using
 *                 too many function parameters. Being a CompleteCmd type
 *                 parameter, it will store information about the command
 *                 execution in all of the structure fields. 
 *                 The hook is called in the execute_command function inside
 *                 shell/shell.c, after each command execution.
 *
 * @return Returns a True, after each call, passing the command information
 *         to the log file. If it isn't implemented, it will get completly
 *         ignored by the compiler.
 *
 * @seealso CompleteCmd
 * @seealso validate_cmd
 */

#pragma weak shell_hook_log_cmd_completion

extern Boolean
shell_hook_log_cmd_completion(struct CompleteCmd *cmd_info);

/**
 * @function validate_help
 *
 * @brief Validate help text for a OSE shell command.
 *
 * @long The function is called from the OSE shell help command and should
 *       validate if the help text for a shell command is allowed to be printed
 *       to the user.
 *
 * @long It is possible to use stdio in the function for additional input and
 *       output to the user.
 *
 * @long This is a function that should be implemented by the developer if it
 *       is wanted. If the function is not implemented a default function in
 *       the OSE CRT library is linked into the system. The default function
 *       always returns valid (True).
 *
 * @input cmdname  Pointer to a string containing the name of the registered
 *                 command.
 *
 * @input username Pointer to a string containing the username.
 *                 NULL if not defined.
 *                 When the function shell_run_cmd() is used the name depends on
 *                 if the environment variable USERNAME is set by the calling
 *                 application.
 *
 * @input termname Pointer to a string containing the name of the terminal
 *                 which can be used to identify which terminal type the user
 *                 is logged in over.
 *                 For the serial port this is the name of the volume when
 *                 mounting the console format manager (confm) and serial device
 *                 driver (serdd), e.g. "/tty".
 *                 For the Telnet port this is the name of the volume when
 *                 mounting the console format manager for the Telnet port, e.g.
 *                 "/telnet_<addr>:<port>".
 *                 For the remote connection this is the remote terminal name
 *                 with the linkpath added, e.g. "master!/tty".
 *                 When the function shell_run_cmd() is used the name depends on
 *                 if the environment variable TERMINAL is set by the calling
 *                 application.
 *                 NULL if not defined.
 *
 * @return If the command help text is valid (is allowed to be printed to the
 *         user) the function shall return non-zero (True).
 *         Otherwise it shall return zero (False).
 *
 * @seealso validate_user
 * @seealso validate_cmd
 * @seealso logout_user
 */
Boolean
validate_help(const char *cmdname,   /* input */
              const char *username,  /* input */
              const char *termname); /* input */

/**
 * @function logout_user
 *
 * @brief Log user logout.
 *
 * @long The function is called from the OSE shell login process and should log
 *       user logout.
 *
 * @long It is possible to use stdio in the function for additional input and
 *       output to the user.
 *
 * @long This is a function that should be implemented by the developer if it
 *       is wanted. If the function is not implemented a default function in
 *       the OSE CRT library is linked into the system.
 *
 * @input username Pointer to a string containing the username.
 *                 NULL if not defined.
 *
 * @input termname Pointer to a string containing the name of the terminal
 *                 which can be used to identify which terminal type the user
 *                 is logged in over.
 *                 For the serial port this is the name of the volume when
 *                 mounting the console format manager (confm) and serial device
 *                 driver (serdd), e.g. "/tty".
 *                 For the Telnet port this is the name of the volume when
 *                 mounting the console format manager for the Telnet port, e.g.
 *                 "/telnet_<addr>:<port>".
 *                 NULL if not defined.
 *
 * @return None.
 *
 * @seealso validate_user
 * @seealso validate_cmd
 * @seealso validate_help
 */
void
logout_user(const char *username,  /* input */
            const char *termname); /* input */

#define AUTHMODE_USERPASS  1 
#define AUTHMODE_USERONLY  2 
#define AUTHMODE_ANONYMOUS 4

#define AUTHMODE_ACCEPTEMPTY  64

/**
 * @function get_authmode
 *
 * @brief Get configuration for authentication behavior
 *
 * @long This function is called from the login shell command  and returns the
 *       configured flags that control at which points the validate_user
 *       function is called along the login procedure.
 *
 * @long This function should be reimplemented by the developer if needed, to
 *       the desired combination of flags. By default it is simply hardcoded.
 * 
 * @return Login user validation behavior flags
 *
 */
#pragma weak get_authmode
extern int
get_authmode(void); 

#ifdef __cplusplus
}
#endif

#endif /* #ifndef _VALIDATE_USER_H */
