/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2011 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */
/**
 * @toc File_System:
 *
 * @file ose_spi/vfs.h
 * @library libcrt.a
 *
 * @brief Virtual file system framework
 *
 * @long    The VFS framework provides a method to implement an OSE volume
 *          manager by defining a set of callback functions for dealing
 *          with the particulars of the volume format.
 *
 * @long    The function @link "" "vfs_mainloop2" executes an OSE signal
 *          receive loop that manages the details of the fm.sig interface
 *          as needed to implement a volume manager. The loop invokes the
 *          callbacks as needed dependent on signals received from client
 *          applications.
 *
 * @long    VFS is intended for "singleton" volume managers where there
 *          can only exist one volume instance of an implemented volume type.
 *          For that reason, there is no volume handle in the interface, and
 *          the interface does not support the "mount" operation.
 *
 * @long    This framework can be used in two modes: The current mode,
 *          invoked via @link "" "vfs_mainloop2", and the legacy mode, 
 *          invoked via @link "" "vfs_mainloop". The main improvements
 *          with the current mode are in the area of error reporting and
 *          in the support for seeking.
 *          
 * @long    The documentation of the VFS callbacks describe the behaviour
 *          when used with @link "" "vfs_mainloop2". Differences between
 *          the current mode and the legacy mode are noted with
 *          @link "" "vfs_mainloop".
 */

#ifndef _VFS_H
#define _VFS_H

#include "ose.h"
#include "efs.h"

#ifdef __cplusplus
extern "C" {
#endif


/**
 * @type VfsHandle
 *
 * @brief   A handle representing an open file in the the VFS framework.
 *
 * @long    The handle is a pointer to a struct that represents an open file.
 *          Each implementation of a vfs VM defines its own version of the
 *          struct. The intention is that the VM should keep info about the
 *          open file in this struct
 */
typedef void *VfsHandle;


/**
 * @type VfsOpenCallback
 *
 * @brief   Callback for implementing open in the vfs framework.
 *
 * @long    This callback should be defined if the implementation is
 *          to support opening of files (excluding directories).
 *
 * @long     This callback is invoked upon reception of @link "fm.sig"
 *          "FM_OPEN_REQUEST" when the flags don't request a directory.
 *
 * @input   handle      Pointer to empty handle struct. The object will be of
 *                      the size @link "" "VfsOps".handle_size.
 * @input   path        Absolute path name of the file that shall be opened.
 * @input   mode        Mode bits for the open operation. This is equivalent
 *                      to the omode argument of the @link "fcntl.h" "open"()
 *                      call, but the bits are in different positions as
 *                      defined in @link "fm.sig" "FmModeEnum".
 *
 * @output  handle      The callback should update the contents of the handle
 *                      struct to represent the open file. The pointer to
 *                      this struct will be used as handle for future
 *                      operations on the file.
 *
 * @return  0 on success, otherwise status code.
 * @break   Status codes are either FM_E* codes from fm.sig or errno codes
 *          from errno.h. see @link "fm.sig" "FM_ERRBASE" for details.
 *          A negative status code is simply a sign changed code, e.g. -EIO.
 */
typedef int    (VfsOpenCallback)    (VfsHandle handle,
                                     const char * path,
                                     EfsBitmask mode);

/**
 * @type VfsCloseCallback
 *
 * @brief   Callback for implementing close in the vfs framework.
 *
 * @long    This callback is invoked when the last reference to an open
 *          file or directory is closed. This can happen upon reception
 *          of @link "fm.sig" "FM_CLOSE_REQUEST" or if a client dies or
 *          if the volume is unmounted.
 *
 * @long    The object pointed to by handle will be removed after this
 *          function returns, regardless of the return value. I.e. the
 *          status code is informative only; the function is required
 *          to do necessary cleanup even if there is an error.
 *
 * @long    This callback is invoked upon reception of @link "fm.sig"
 *          "FM_CLOSE_REQUEST" and when 
 *
 * @input   handle   Pointer to a handle struct.
 *
 * @return  0 on success, otherwise status code.
 * @break   Status codes are either FM_E* codes from fm.sig or errno codes
 *          from errno.h. see @link "fm.sig" "FM_ERRBASE" for details.
 */
typedef int    (VfsCloseCallback)   (VfsHandle handle);


/**
 * @type VfsReadCallback
 *
 * @brief   Callback for implementing file read in the vfs framework.
 *
 * @long    This callback should be defined if the implementation is
 *          to support reading of files (excluding directories).
 *
 * @long    The read request signals have an option for an embedded seek
 *          request. Such seek requests will however be handled by @link
 *          "" "VfsLseekCallback" and should be ignored here.
 *
 * @long    This callback is invoked upon reception of @link "fm.sig"
 *          "FM_READ_REQUEST" or @link "fm.sig" "FM_READ_PTR_REQUEST"
 *          when referencing an open file.
 *
 * @input   handle      Pointer to a handle struct.
 * @input   position    This parameter should be ignored. It is provided
 *                      only for compatibility with the legacy interface,
 *                      see @link "" "vfs_mainloop" for details.
 * @input   buffer      Pointer to where to place the data.
 * @input   requested   Number of bytes to read.
 * @output  buffer      Contents of buffer should be updated with read data.
 *
 * @return  positive value on success where the value indicates number of
 *          bytes read. Otherwise negative status code.
 * @break   Status codes are either FM_E* codes from fm.sig or errno codes
 *          from errno.h. see @link "fm.sig" "FM_ERRBASE" for details.
 *          A negative status code is simply a sign changed code, e.g. -EIO.
 */
typedef long   (VfsReadCallback)    (VfsHandle handle,
                                     size_t position,
                                     void * buffer,
                                     size_t requested);


/**
 * @type VfsWriteCallback
 *
 * @brief   Callback for implementing regular file write in the vfs framework.
 *
 * @long    This callback should be defined if the the implementation is
 *          to support writing of files.
 *
 * @long    The write request signals have an option for an embedded seek
 *          request. Such seek requests will however be handled by @link
 *          "" "VfsLseekCallback" and should be ignored here. It is assumed
 *          that write at "end of file" is equivalent to "current position"
 *          for VMs that do not support seeking.
 *
 * @long    This callback is invoked upon reception of @link "fm.sig"
 *          "FM_WRITE_REQUEST" or @link "fm.sig" "FM_WRITE_PTR_REQUEST".
 *
 * @input   handle      Pointer to a handle struct.
 * @input   position    This parameter should be ignored. It is provided
 *                      only for compatibility with the legacy interface,
 *                      see @link "" "vfs_mainloop" for details.
 * @input   buffer      Pointer to the data.
 * @output  requested   Number of bytes to write.
 *
 * @return  Positive value on success where the value indicates number of
 *          bytes read. Otherwise negative status code.
 * @break   Status codes are either FM_E* codes from fm.sig or errno codes
 *          from errno.h. see @link "fm.sig" "FM_ERRBASE" for details.
 *          A negative status code is simply a sign changed code, e.g. -EIO.
 */
typedef long   (VfsWriteCallback)   (VfsHandle handle,
                                     size_t position,
                                     const void * buffer,
                                     size_t requested);


/**
 * @type VfsStatCallback
 *
 * @brief   Callback for implementing stat in the vfs framework.
 *
 * @long    This callback should be defined if the the implementation is
 *          to support the stat operation.
 *
 * @long    This callback is invoked upon reception of @link "fm.sig"
 *          "FM_EXAMINE_OBJ_REQUEST".
 *
 * @input   path        Absolute path name of the file that shall be examined.
 * @input   obj_info    Pointer to output buffer.
 * @output  obj_info    The @link "fm.sig" "FmObjectInfo" buffer shall be
 *                      updated to contain info about the file.
 *
 * @return  0 on success, otherwise status code.
 * @break   Status codes are either FM_E* codes from fm.sig or errno codes
 *          from errno.h. see @link "fm.sig" "FM_ERRBASE" for details.
 */
typedef int    (VfsStatCallback)    (const char *path,
                                     struct FmObjectInfo *obj_info);


/**
 * @type VfsOpendirCallback
 *
 * @brief   Callback for implementing opendir in the vfs framework.
 *
 * @long    This callback should be defined if the the implementation is
 *          to support opening of directories.
 *
 * @long    This callback is invoked upon reception of @link "fm.sig"
 *          "FM_OPEN_REQUEST" when the flags request a directory.
 *
 * @input   handle      Pointer to empty handle struct. The object will be of
 *                      the size @link "" "VfsOps".handle_size.
 * @input   path        Absolute path name of the directory to open.
 * @output  handle      The callback should update the contents of the handle
 *                      struct to represent the open directory. The pointer to
 *                      this struct will be used as handle for future
 *                      operations on the directory.
 *
 * @return  0 on success, otherwise status code.
 * @break   Status codes are either FM_E* codes from fm.sig or errno codes
 *          from errno.h. see @link "fm.sig" "FM_ERRBASE" for details.
 */
typedef int    (VfsOpendirCallback) (VfsHandle handle,
                                     const char *path);


/**
 * @type VfsReaddirCallback
 *
 * @brief   Callback for implementing directory read in the vfs framework.
 *
 * @long    This callback should be defined if the the implementation is
 *          to support reading of directories.
 *
 * @long    This callback is invoked upon reception of @link "fm.sig"
 *          "FM_READ_REQUEST" or @link "fm.sig" "FM_READ_PTR_REQUEST"
 *          when referencing an open directory.
 *
 * @input   handle      Pointer to a handle struct.
 * @input   buffer      Pointer to where to place the data.
 * @input   requested   Number of bytes to read.
 * @output  buffer      Contents of buffer should be updated with read data.
 *                      The output should be a sequence of null terminated
 *                      filenames.
 *
 * @return  positive value on success where the value indicates number of
 *          bytes read. Otherwise negative status code.
 * @break   Status codes are either FM_E* codes from fm.sig or errno codes
 *          from errno.h. see @link "fm.sig" "FM_ERRBASE" for details.
 *          A negative status code is simply a sign changed code, e.g. -EIO.
 */
typedef long   (VfsReaddirCallback) (VfsHandle handle,
                                     char *buffer,
                                     size_t requested);

/**
 * @type VfsFmSignalCallback
 *
 * @brief   Callback for handling various fm.sig signals in the vfs framework.
 *
 * @long    This callback gives the option to handle fm.sig signals that
 *          are not handled by other callbacks. If defined, it will be
 *          invoked by the following signals: @link "fm.sig"
 *          "FM_CREATE_OBJ_REQUEST", @link "fm.sig" "FM_REMOVE_OBJ_REQUEST",
 *          @link "fm.sig" "FM_RENAME_OBJ_REQUEST", @link "fm.sig"
 *          "FM_SET_SIZE_OBJ_REQUEST", @link "fm.sig" "FM_SET_TIME_OBJ_REQUEST",
 *          @link "fm.sig" "FM_SET_OWNER_OBJ_REQUEST", @link "fm.sig"
 *          "FM_DRAIN_REQUEST", @link "fm.sig" "FM_POLL_REQUEST", @link "fm.sig"
 *          "FM_CANCEL_POLL", @link "fm.sig" "FM_FCNTL_REQUEST", @link "fm.sig"
 *          "FM_FLUSH_TERM_REQUEST", @link "fm.sig" "FM_EXAMINE_TERM_REQUEST",
 *          @link "fm.sig" "FM_CONFIG_TERM_REQUEST", @link "fm.sig"
 *          "FM_SUBSCRIBE_SIGNAL_REQUEST", @link "fm.sig" "FM_SIGNAL",
 *          @link "fm.sig" "FM_ADD_NOTIFICATION_REQUEST", @link "fm.sig"
 *          "FM_NOTIFICATION", @link "fm.sig" "FM_REMOVE_NOTIFICATION_REQUEST",
 *          @link "fm.sig" "FM_FALLOCATE_REQUEST", and @link "fm.sig"
 *          "FM_FADVISE_REQUEST".
 *
 * @long    Note that the set of signals handled by this function may be
 *          extended in future releases. For unsupported signals, this
 *          function should return a signal modeled on struct @link "fm.sig"
 *          "FmSyncVol", where the sigNo field is incremented by 1 and the
 *          status field contains @link "fm.sig" "FM_EOPER_NOT_SUPPORTED".
 *
 * @long     The callback must create
 *          and return an appropriate reply signal.
 *
 * @input   handle      Pointer to a handle struct.
 * @input   sig         Pointer to request signal. This signal shall either
 *                      be updated as a reply signal and returned or freed.
 *
 * @return  Pointer to reply signal.
 * 
 * @example
 * The following code illustrates a callback example:
 *
 *    // The callback shall be invoked whenever the virtual volume manager
 *    // receives one of the specified signals mentioned above.
 *
 *    static union SIGNAL *
 *    vfs_fm_signal_clbk(Vfshandle VfsH, union SIGNAL *rcv)
 *    {
 *       switch(rcv->sigNo)
 *       {
 *          case FM_CREATE_OBJ_REQUEST:
 *          {
 *             // Do something.
 *          }
 *          case FM_RENAME_OBJ_REQUEST:
 *          {
 *             // Do something else.
 *          }
 *          // Handle more FM signals.
 *       }
 *       rcv->reply.sigNo++;
 *
 *       // Return the reply back to the original sender.
 *       return (union SIGNAL *)rcv;
 *    }
 *                     
 */
typedef union SIGNAL * (VfsFmSignalCallback)(VfsHandle handle,
                                             union SIGNAL *sig);


/**
 * @type VfsTimeoutCallback
 *
 * @brief   Callback for a timeout signal in the vfs framework.
 *
 * @long    If defined, this callback is invoked upon reception of signal
 *          (30116). The intention is that the signal should be used as
 *          a timeout signal in a manner specific to the VM.
 *
 * @long    Note that the signal (30116) is defined by fm.sig, but it is
 *          not used in a context that is supported by VFS.
 *
 * @input   sig      Pointer to signal. The callback assumes ownership of
 *                   this signal.
 *
 * @return  void
 */
typedef void   (VfsTimeoutCallback) (union SIGNAL *sig);


/**
 * @type VfsFstatCallback
 *
 * @brief   Callback for implementing fstat in the vfs framework.
 *
 * @long    This callback should be defined if the implementation is
 *          to support the fstat operation.
 *
 * @long    This callback is invoked upon reception of @link "fm.sig"
 *          "FM_EXAMINE_REQUEST"
 *
 * @input   handle      Pointer to a handle struct.
 * @input   obj_info    Pointer to output buffer.
 * @output  obj_info    The @link "fm.sig" "FmObjectInfo" buffer shall be
 *                      updated to contain info about the file.
 *
 * @return  0 on success, otherwise status code.
 * @break   Status codes are either FM_E* codes from fm.sig or errno codes
 *          from errno.h. see @link "fm.sig" "FM_ERRBASE" for details.
 */
typedef int    (VfsFstatCallback)   (VfsHandle handle,
                                     struct FmObjectInfo *obj_info);



/**
 * @type VfsLseekCallback
 *
 * @brief   Callback for implementing seeking in the vfs framework.
 *
 * @long    This callback should be defined if the implementation is
 *          to support seeking.
 *
 * @long    This callback is invoked by the regular signal for seek:
 *          @link "fm.sig" "FM_SET_POS_REQUEST".
 *
 * @long    This callback is also invoked if a read or write request
 *          contains an embedded seek request encoded in the position
 *          field of the signals @link "fm.sig" "FM_READ_REQUEST",
 *          @link "fm.sig" "FM_READ_PTR_REQUEST", @link "fm.sig"
 *          "FM_WRITE_REQUEST" or @link "fm.sig" "FM_WRITE_PTR_REQUEST".
 *
 * @input   handle      Pointer to a handle struct.
 * @input   whence      Base position to seek from, either @link "unistd.h"
 *                      "SEEK_SET", @link "unistd.h" "SEEK_CUR" or @link
 *                      "unistd.h" "SEEK_END" 
 * @input   offset      Offset from the base position to seek.
 *
 * @return  On success the resulting offset, as measured in bytes from the
 *          beginning of the file. Otherwise a negative status code.
 * @break   Status codes are either FM_E* codes from fm.sig or errno codes
 *          from errno.h. see @link "fm.sig" "FM_ERRBASE" for details.
 *          A negative status code is simply a sign changed code, e.g. -EIO.
 */
typedef long   (VfsLseekCallback)   (VfsHandle handle,
                                     int whence,
                                     long offset);


/**
 * @type VfsOps
 *
 * @brief   Struct with info and callbacks for a VFS VM implementation
 *
 * @long    This struct contains a few info fields and a set of fields
 *          with function pointers to the callbacks that are implemented
 *          by a VFS VM implementation.
 *
 * @long    The callback pointer fields can be either a null pointer or a
 *          function pointer to a callback. A null pointer indicates
 *          that the functionality corresponding to the callback is not
 *          implemented by the VM. (Note that vfs_close is not optional.)
 *    
 * @field   name           Name of the mount-point of the volume, e.g. "/myvol".
 * @field   handle_size    Size in bytes of the object VfsHandle shall point to.
 *                         I.e. typically sizeof a sruct that is used by the
 *                         VM implementation to represent an open file.
 * @field   vfs_open       Pointer to @link "" "VfsOpenCallback" or null
 * @field   vfs_close      Pointer to @link "" "VfsCloseCallback". NOT OPTIONAL!
 * @field   vfs_read       Pointer to @link "" "VfsReadCallback" or null.
 * @field   vfs_write      Pointer to @link "" "VfsWriteCallback" or null.
 * @field   vfs_stat       Pointer to @link "" "VfsStatCallback" or null.
 * @field   vfs_opendir    Pointer to @link "" "VfsOpendirCallback" or null.
 * @field   vfs_readdir    Pointer to @link "" "VfsReaddirCallback" or null.
 * @field   vfs_fm_signal  Pointer to @link "" "VfsFmSignalCallback" or null.
 * @field   vfs_timeout    Pointer to @link "" "VfsTimeoutCallback" or null.
 * @field   vfs_fstat      Pointer to @link "" "VfsFstatCallback" or null.
 * @field   vfs_lseek      Pointer to @link "" "VfsLseekCallback" or null.
 */
struct VfsOps
{
   const char           *name;
   size_t               handle_size;   
   VfsOpenCallback      *vfs_open;
   VfsCloseCallback     *vfs_close;
   VfsReadCallback      *vfs_read;
   VfsWriteCallback     *vfs_write;
   VfsStatCallback      *vfs_stat;
   VfsOpendirCallback   *vfs_opendir;
   VfsReaddirCallback   *vfs_readdir;
   VfsFmSignalCallback  *vfs_fm_signal;
   VfsTimeoutCallback   *vfs_timeout;
   VfsFstatCallback     *vfs_fstat;
   VfsLseekCallback     *vfs_lseek;
   /* Any future extensions will be added after vfs_lseek */
};


/**
 * @function vfs_mainloop2
 * 
 * @brief   Executes a VFS volume manager instance.
 *
 * @long    This function implements an OSE signal receive loop that manages
 *          the details of the fm.sig interface as needed to implement a
 *          volume manager. The loop invokes the callbacks provided in the
 *          vfs_ops parameter as needed dependent on signals received from
 *          client applications.
 *
 * @long    This function implicitly mounts a volume named by vfs_ops->name.
 *          The signal interface for mounting volumes is not supported.
 *
 * @long    The following should be considered when converting a VM
 *          from @link "" "vfs_mainloop" to @link "" "vfs_mainloop2":
 *
 * @break   The functions vfs_ops.vfs_read, vfs_ops.vfs_readdir,
 *          vfs_ops.vfs_write and vfs_ops.vfs_lseek need to return a
 *          negative status code in case of an error; -1 is no longer
 *          acceptable.
 *
 * @break   Support for seeking is changed. It is implemented in
 *          vfs_ops.vfs_lseek only. It should be removed from
 *          vfs_ops.vfs_read, vfs_ops.vfs_write and vfs_ops.vfs_fm_signal.
 *          The entire vfs_ops.vfs_fm_signal should be removed if it
 *          served no other purpose than seeking.
 *
 * @input   vfs_ops  A pointer to a struct @link "" "VfsOps" object,
 *                   typically a static const object. The object is in use
 *                   until the volume is unmounted.
 *
 * @input   ops_size Shall contain sizeof(struct VfsOps). The purpose is
 *                   to allow future enhancements of the interface in a
 *                   backwards compatible manner.
 *
 * @return  The function does not return. The process is killed if the volume
 *          is unmounted.
 *
 * @seealso default_restrictions_calling_functions
 *
 * @example
 * The following code executes a VFS volume manager instance.
 *    OS_PROCESS(my_vfs_vm)
 *    {
 *       vfs_mainloop2(&my_vfs_ops, sizeof(struct VfsOps));
 *    }
 */
void vfs_mainloop2(const struct VfsOps *vfs_ops,
                   size_t ops_size);


/**
 * @function vfs_mainloop
 * 
 * @brief   Executes a VFS volume manager instance in legacy mode.
 *
 * @long    This function provides the functionality of the legacy
 *          version of VFS. It is equivalent to vfs_mainloop2 with the
 *          following exceptions:
 *
 * @long    Older versions of the header file had different types for
 *          the return values of vfs_ops.vfs_read, vfs_ops.vfs_readdir,
 *          vfs_ops.vfs_write and vfs_ops.vfs_lseek, together with the
 *          third argument of vfs_ops.vfs_lseek. Compiling legacy
 *          implementation code with the current header file will result
 *          in compiler warnings unless the code is adapted to the new
 *          function prototypes.
 *
 * @long    The legacy behaviour with vfs_ops.vfs_read and vfs_ops.vfs_readdir
 *          is that all negative return values result in the status code
 *          FM_EINVALID_POSITION.
 *
 * @long    The legacy behaviour with vfs_ops.vfs_write is that the status
 *          code FM_EINVALID_POSITION is generated if the return value
 *          differs from the requested write length.
 *
 * @long    The legacy behaviour with vfs_ops.vfs_lseek is unable to report
 *          an error condition.
 *
 * @long    In the legacy mode, vfs_ops.vfs_fm_signal is supported for
 *          handling FM_SET_POS_REQUEST if vfs_ops.vfs_lseek is NULL. This
 *          is the recommended method to handle seek in legacy mode since
 *          vfs_ops.vfs_lseek is unable to report errors.
 *
 * @long    In the legacy mode, vfs_ops.vfs_lseek will not be invoked for
 *          handling embedded seek operations in read and write requests.
 *          For that reason, vfs_ops.vfs_read and vfs_ops.vfs_write must
 *          handle the position parameter if the implementation supports
 *          seeking. This parameter can contain an embedded seek request
 *          on the format described in @link "fm.sig" "FmPositionEnum".
 *
 * @long    There is no ops_size parameter.
 *
 * @input   vfs_ops  A pointer to a struct @link "" "VfsOps" object,
 *                   typically a static const object. The object is in use
 *                   until the volume is unmounted.
 *
 * @return  The function does not return. The process is killed if the volume
 *          is unmounted.
 *
 * @seealso default_restrictions_calling_functions
 *
 * @example
 * The following code executes a legacy VFS volume manager instance.
 *    OS_PROCESS(my_vfs_vm)
 *    {
 *       vfs_mainloop(&my_vfs_ops);
 *    }
 */
void vfs_mainloop(const struct VfsOps *vfs_ops);


#ifdef __cplusplus
}
#endif

#endif
