/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2011 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

/**
 * @public
 * @toc Standard_C:
 * @file pthread.h
 *
 * @library crt (C run time)
 * 
 * @brief Functions implementing the POSIX thread layer.
 *
 * @long  These functions implement the thread part of POSIX,
 *        and handles: @break
 *        
 *            - Threads  @break
 *            - Mutexes  @break 
 *            - Condition Variables  @break
 *            - Thread specific data
 *
 * @long  OSE implements a few extensions to the POSIX standard to make it
 *        easier to design an OSE system with pthreads: @break
 *
 *        - You can specify which OSE block the pthread will belong to,
 *          see @link "" "pthread_attr_setblock". @break
 *        - The name of the OSE process that implements the pthread can be
 *          set, see @link "" "pthread_attr_setname". @break
 *        - The POSIX standard specifies that pthreads are started
 *          immediately after creation, but this isn't always the preferred
 *          behavior. An OSE pthread can be created in stopped state,
 *          see @link "" "pthread_attr_setstopped".
 *
 * @long  When using OSE pthreads on the soft kernel, the symbols defined
 *        in pthread.h are not allowed to collide with the pthread library
 *        on the host. Therefore, all functions in pthread.h for the
 *        OSE Soft Kernel are prefixed with "ose_". An OSE application that
 *        uses pthreads must include this file which will translate any 
 *        pthread_xxx() call to the corresponding ose_pthread_xxx(). Observe
 *        that pthread.h from the host system should not be included at the
 *        same time as this file.
 *            
 * @long  See the Pthread chapter in OSE Core User's Guide for more
 *        information.
 *
 * @long  A pthread is an OSE prioritized process, so anything that an OSE
 *        prioritized process can do, a pthread can do, with two exceptions:
 *        @break
 *        1) Pthread implementation uses the fast semaphore. Do not use the
 *           fast semaphore for anything else in a phtread. @break
 *        2) Pthreads may only die by a pthread_exit() call or by a simple
 *           return statement from process main scope.
 *
 * @long  There are a few functions that can only be called from a pthread.
 *        These are documented with the below restriction text: @break
 *        "Only available in pthreads.". @break
 *        Functions that do not have this text do not have this restriction.
 */

#ifndef _PTHREAD_H
#define _PTHREAD_H
#define OSE_PTHREAD_H

#include "ose.h"
#include "osetypes.h"
#include "time.h"

#ifdef __cplusplus
extern "C"
{
#endif

/*
 * When using OSE pthreads on the soft kernel, the symbols defined
 * in the library libpthread.a are not allowed to collide with
 * the pthread library on host.
 * Therefore, all functions in libpthread.a for the OSE Soft Kernel
 * are prefixed with "ose_".
 * The OSE application that uses pthreads must include this file,
 * and the following lines will translate any pthread_xxx() call
 * to the corresponding ose_pthread_xxx().
 * Observe that pthread.h from the host system should not be
 * included at the same time as this file.
 *
 */

#ifdef softOSE
#define ZZMK_NAME_MANGLING
#include "sfk/sfk_pthread.h"
#endif

/**
 * @macro SCHED_RR
 *
 * @brief Round robin scheduling.
 *
 */
#define SCHED_RR    (1)

/**
 * @macro SCHED_FIFO
 *
 * @brief First in first out scheduling.
 *
 */
#define SCHED_FIFO  (2)

/**
 * @macro SCHED_OTHER
 *
 * @brief Other kind of scheduling.
 *
 */
#define SCHED_OTHER (2)

/**
 * @macro PTHREAD_COND_INITIALIZER
 *
 * @brief Initializes a static condition variable.
 *
 */
#define PTHREAD_COND_INITIALIZER {0, 0, 0, 0}

/**
 * @macro PTHREAD_MUTEX_INITIALIZER
 *
 * @brief Initializes a static mutex.
 *
 */
#define PTHREAD_MUTEX_INITIALIZER OSE_MUTEX_INITIALIZER(0)

/**
 * @macro PTHREAD_ONCE_INIT
 *
 * @brief Initializes a pthread_once_t control.
 *
 */
#define PTHREAD_ONCE_INIT  PTHREAD_MUTEX_INITIALIZER

/**
 * @macro PTHREAD_CREATE_DETACHED
 *
 * @brief Attribute for creating a detached thread.
 *
 */
#define PTHREAD_CREATE_DETACHED       (1)

/**
 * @macro PTHREAD_CREATE_JOINABLE
 *
 * @brief Attribute for creating a joinable thread.
 *
 */
#define PTHREAD_CREATE_JOINABLE       (2)

/**
 * @macro PTHREAD_CREATE_STOPPED
 *
 * @brief Attribute for creating a stopped thread.
 *
 */
#define PTHREAD_CREATE_STOPPED        (4)

/**
 * @macro PTHREAD_INHERIT_SCHED
 *
 * @brief Attribute for creating a thread that inherits scheduling properties.
 *
 */
#define PTHREAD_INHERIT_SCHED         (0)

/**
 * @macro PTHREAD_EXPLICIT_SCHED
 *
 * @brief Attribute that explicitly defines the scheduling properties.
 *
 */
#define PTHREAD_EXPLICIT_SCHED        (1)

/**
 * @macro PTHREAD_SCOPE_SYSTEM
 *
 * @brief Attribute to make the thread be scheduled system-wide.
 *
 */
#define PTHREAD_SCOPE_SYSTEM          (1)

/**
 * @macro PTHREAD_SCOPE_PROCESS
 *
 * @brief Attribute to make the thread be scheduled process-wide.
 *
 */
#define PTHREAD_SCOPE_PROCESS         (2)

/**
 * @macro PTHREAD_CANCEL_DISABLE
 *
 * @brief Attribute to make the thread impossible to cancel.
 *
 */
#define PTHREAD_CANCEL_DISABLE        (1)

/**
 * @macro PTHREAD_CANCEL_ENABLE
 *
 * @brief Attribute to make it possible to cancel the thread.
 *
 */
#define PTHREAD_CANCEL_ENABLE         (2)

/**
 * @macro PTHREAD_CANCEL_ASYNCHRONOUS
 *
 * @brief Attribute to make it possible to cancel the thread asynchronously.
 *
 */
#define PTHREAD_CANCEL_ASYNCHRONOUS   (1)

/**
 * @macro PTHREAD_CANCEL_DEFERRED
 *
 * @brief Attribute to make it possible to cancel the thread deferred.
 *
 */
#define PTHREAD_CANCEL_DEFERRED       (2)

/**
 * @macro PTHREAD_CANCELED
 *
 * @brief The status when joining a canceled thread.
 *
 */
#define PTHREAD_CANCELED              (3)

/**
 * @macro PTHREAD_KEYS_MAX
 *
 * @brief Maximum number of thread specific data keys that can be created.
 *
 */
#define PTHREAD_KEYS_MAX              (64)

/**
 * @macro PTHREAD_DESTRUCTOR_ITERATIONS
 *
 * @brief Maximum times the iterators are run.
 *
 */
#define PTHREAD_DESTRUCTOR_ITERATIONS (2)

/**
 * @macro DEFAULT_PTHREAD_CREATE_DETACHSTATE
 *
 * @brief Default attribute for detach state.
 *
 */
#define DEFAULT_PTHREAD_CREATE_DETACHSTATE (PTHREAD_CREATE_JOINABLE)

/**
 * @macro DEFAULT_PTHREAD_PRIO
 *
 * @brief Default priority for a thread.
 *
 */
#define DEFAULT_PTHREAD_PRIO               (31)

/**
 * @macro DEFAULT_PTHREAD_INHERIT
 *
 * @brief Default priority inheritance property.
 *
 */
#define DEFAULT_PTHREAD_INHERIT            (PTHREAD_EXPLICIT_SCHED)

/**
 * @macro DEFAULT_PTHREAD_STACKSIZE
 *
 * @brief Default stack size for a thread.
 *
 */
#define DEFAULT_PTHREAD_STACKSIZE          (4095)


#define PTHREAD_MUTEX_RECURSIVE          (0)


/**
 * @macro pthread_cleanup_push
 *
 * @brief Pushes a cleanup handler on the cleanup handler stack.
 *
 * @input func Cleanup handler function.
 *
 * @input arg  Argument (void *) for the cleanup handler function.
 *
 * @seealso default_restrictions_calling_functions
 */
#define pthread_cleanup_push(func,arg) { \
   Cleanelem e; \
   e.cleaner = (func); \
   e.tobecleaned = (arg); \
   xxcleanup_push(&e);

/**
 * @macro pthread_cleanup_pop
 *
 * @brief Pops a cleanup handler from the cleanup handler stack.
 *
 * @input execute If execute is non-zero, the cleanup handler is run
 *                before popping it from the cleanup handler stack.
 *
 * @seealso default_restrictions_calling_functions
 */
#define pthread_cleanup_pop(execute) xxcleanup_pop(execute); }

/**
 * @type pthread_mutex_t
 *
 * @brief Mutex structure.
 */
#if defined (__SVR4) && defined (__sun)
/* See OSECR-11670. */
struct OS_mutex_compat
{
   long long m[3];
};
typedef struct OS_mutex_compat pthread_mutex_t;
#else
typedef MUTEX pthread_mutex_t;
#endif
   
/**
 * @type pthread_mutexattr_t
 *
 * @brief Mutex attribute for mutex creation.
 */
typedef int pthread_mutexattr_t;

/**
 * @type pthread_attr_t
 *
 * @brief Thread attribute for thread creation.
 */
typedef struct pthread_attr_t pthread_attr_t;

/**
 * @type pthread_cond_t
 *
 * @brief Condition variable structure.
 */
typedef struct osepth_condition_variable pthread_cond_t;

/**
 * @type pthread_condattr_t
 *
 * @brief Attribute for condition variable creation.
 */
typedef int pthread_condattr_t;

/**
 * @type pthread_once_t
 *
 * @brief Pthread once control structure.
 */
typedef pthread_mutex_t pthread_once_t;

/**
 * @type pthread_t
 *
 * @brief Thread structure.
 */
typedef PROCESS pthread_t;

typedef struct cleanelem Cleanelem;

typedef struct keydata Keydata;

/**
 * @type pthread_key_t
 *
 * @brief Thread specific data key type.
 */
typedef Keydata *pthread_key_t;

struct keydata
{
   void (*destructor) (void *);
   Keydata **keyref;
};

struct cleanelem
{
   Cleanelem *olderelems;
   void (*cleaner) (void *);
   void *tobecleaned;
};

/**
 * @type sched_param
 *
 * @brief Scheduling parameter.
 *
 * @field sched_priority  Priority for the thread.
 *
 */
struct sched_param
{
   int sched_priority;
};

struct osepth_condition_variable
{
   U32 nwaiters;
   U32 mutex;
   U32 first;
   U32 last;
};

struct pthread_attr_t
{
   U32 detachstate;
   U32 inheritsched;
   struct sched_param schedparam;
   size_t stacksize;
   U32 block_pid;
   const char *name;
};


/**
 * @function pthread_create
 *
 * @brief Creates a new pthread.
 *
 * @long  The new pthread calls the function defined by func() with one
 *        argument, arg. If more than one argument needs to be passed to
 *        func(), the arguments can be packed into a structure, and the
 *        address of that structure can be passed to arg.
 *
 * @long  If func() returns, the pthread will terminate with its exit
 *        status set to the return value of func(). The pthread can
 *        also be terminated through cancellation, using pthread_cancel(),
 *        or by calling pthread_exit().
 *
 * @output thread_id  Thread ID of the created thread. If the pthread
 *                    creation fails, a new pthread is not created and the
 *                    value of thread_id is undefined.
 *
 * @input  attr    The attributes for the new thread or default attributes if
 *                 NULL. If any attributes specified by attr are changed in
 *                 the attribute object prior to the call to pthread_create(),
 *                 the new pthread will acquire those changes. However, if
 *                 any attributes specified by attr are changed after the call
 *                 to pthread_create(), the attributes of existing pthreads
 *                 will not be affected. 
 *
 * @input  func    Function to run as a thread.
 *
 * @input  arg     Argument to the function.
 *
 * @return Zero if successful. Otherwise a non-zero error code.
 *
 * @error EAGAIN  Not enough system resources to create a process for the
 *                new thread.
 *
 * @error EINVAL  More than PTHREAD_THREADS_MAX threads are already active.
 *
 * @seealso default_restrictions_calling_functions
 */
extern int
pthread_create(pthread_t * thread_id,
	       const pthread_attr_t * attr,
	       void *(*func) (void *),
	       void *arg);


/**
 * @function pthread_exit
 *
 * @brief  Terminates the calling pthread.
 *
 * @long   If the calling pthread is not detached, the pthread 
 *         and its exit status are retained. The status value is
 *         available to any successful pthread_join() with the terminating
 *         pthread. If the pthread is detached, status is disregarded allowing
 *         the pthread and its resources to be immediately reclaimed.
 *
 * @long   Upon thread termination, all thread-specific data bindings are
 *         released and its cancellation routines are called. However,
 *         application visible resources, including mutexes, are not released.
 *
 * @long   Appropriate destructor functions will be called in an unspecified
 *         order after all cancellation cleanup handlers have been executed.
 *
 * @long   If any pthread function returns, implicitly or explicitly, the
 *         result is the same as if the pthread called pthread_exit(). 
 *         The return value of the function is used as the exit code. 
 *
 * @input  exitval  The exit value of the exiting thread.
 *
 * @return The function will not return to its caller. 
 * 
 * @restriction Only available in pthreads.
 *
 * @seealso pthread_join
 * @seealso pthread_detach
 * @seealso default_restrictions_calling_functions
 */
extern void
pthread_exit(void *exitval);


/**
 * @function pthread_join
 *
 * @brief Suspends processing until the target thread terminates.
 *
 * @long  Memory resources for a joinable thread can not be reclaimed,
 *        unless another thread joins to it.
 *
 * @input  thread_id   Target thread to join.
 *
 * @output exitval  Exit value of the target thread.
 *
 * @return Zero if successful. Otherwise a non-zero error code.
 *
 * @restriction Only available in pthreads.
 *
 * @error ESRCH   No thread could be found corresponding to that
 *                specified by thread.
 *
 * @error EINVAL  The target thread has been detached.
 *
 * @error EINVAL  Another thread has already joined target thread.
 *
 * @error EDEADLK Current thread is trying to join itself.
 *
 * @seealso default_restrictions_calling_functions
 */
extern int
pthread_join(pthread_t thread_id,
	     void **exitval);


/**
 * @function pthread_detach
 *
 * @brief Sets the target thread in detached state.
 * 
 * @long  When a thread is set in detached state, it implies that the
 *        memory resources for the thread will be reclaimed when the
 *        thread terminates. thread_detach() will not terminate the thread.
 *
 * @input thread_id  Target thread to set in detached state.
 *
 * @return  Zero if successful. Otherwise a non-zero error code.
 *
 * @error   ESRCH  No thread could be found with the specified identity.
 *
 * @error   EINVAL  The specified thread is already in state detached, or
 *                  is already joined by another thread.
 *
 * @seealso default_restrictions_calling_functions
 */
extern int
pthread_detach(pthread_t thread_id);


/**
 * @function pthread_cancel
 *
 * @brief Requests that the target thread is canceled.
 *
 * @input thread_id  Target thread to cancel.
 *
 * @return Zero if successful. Otherwise a non-zero error code.
 *
 * @error ESRCH  No thread could be found with the specified identity.
 *
 * @seealso default_restrictions_calling_functions
 */
extern int
pthread_cancel(pthread_t thread_id);


/**
 * @function pthread_testcancel
 *
 * @brief Initiates self cancellation if thread was canceled previously.
 *
 * @restriction Only available in pthreads.
 *
 * @seealso default_restrictions_calling_functions
 */
extern void
pthread_testcancel(void);


/**
 * @function pthread_self
 *
 * @brief Returns the thread ID of the current thread.
 *
 * @return Thread ID of the current thread.
 *
 * @restriction Only available in pthreads.
 *
 * @seealso default_restrictions_calling_functions
 */
extern pthread_t
pthread_self(void);


/**
 * @function pthread_setcancelstate
 *
 * @brief Sets the cancellation state of the current thread.
 *
 * @input state      New cancellation state.
 *
 * @output oldstate  Old cancellation state.
 *
 * @return  Zero if successful. Otherwise a non-zero error code.
 *
 * @restriction Only available in pthreads.
 *
 * @error EINVAL The state argument is not PTHREAD_CANCEL_ENABLE nor
 *               PTHREAD_CANCEL_DISABLE.
 *
 * @seealso default_restrictions_calling_functions
 */
extern int
pthread_setcancelstate(int state,
		       int *oldstate);


/**
 * @function pthread_mutex_init
 *
 * @brief Initializes the mutex.
 *
 * @input mp    Mutex that will be initialized.
 *
 * @input attr  Attributes for the initialization.
 *
 * @return  Zero if successful. Otherwise a non-zero error code.
 *
 * @error EINVAL The value specified by attr is invalid.
 *
 * @restriction On the OSE soft kernel the fast semaphore is used to
 *              implement mutexes. Non-pthread processes that use mutexes
 *              must therefore initiate their fast semaphore with the value
 *              0 before calling any mutex functions. The result of killing a
 *              process (or a pthread) that is locking or waiting on a
 *              pthread mutex is undefined.
 *
 * @seealso default_restrictions_calling_functions
 */
extern int
pthread_mutex_init(pthread_mutex_t * mp,
		   const pthread_mutexattr_t * attr);


/**
 * @function pthread_mutexattr_init
 *
 * @brief Initializes mutex attributes.
 *
 * @input attr   Attribute variable to initialize.
 * 
 * @return Zero.
 * 
 * @restriction Mutex attributes are not implemented.
 *              Calling this function has no effect.
 *
 * @seealso default_restrictions_calling_functions
 */
extern int
pthread_mutexattr_init(const pthread_mutexattr_t * attr);


/**
 * @function pthread_mutex_lock
 *
 * @brief Locks the mutex.
 *
 * @input mutex  Mutex to lock.
 *
 * @return  Zero if successful. Otherwise a non-zero error code.
 *
 * @error EINVAL   The mutex has not been properly initialized.
 *
 * @error EDEADLK  The mutex is already locked by the calling thread.
 *
 * @restriction On the OSE soft kernel the fast semaphore is used to
 *              implement mutexes. Non-pthread processes that use mutexes
 *              must therefore initiate their fast semaphore with the value
 *              0 before calling any mutex functions. The result of killing a
 *              process (or a pthread) that is locking or waiting on a
 *              pthread mutex is undefined.
 *
 * @seealso default_restrictions_calling_functions
 */
extern int
pthread_mutex_lock(pthread_mutex_t * mutex);


/**
 * @function pthread_mutex_trylock
 *
 * @brief Locks the mutex if it is not currently locked.
 *
 * @input mutex  Mutex that current thread will try to lock.
 *
 * @return  Zero if successful. Otherwise a non-zero error code.
 *
 * @error EBUSY   The mutex was already locked.
 *
 * @error EINVAL  The mutex has not been properly initialized.
 *
 * @error EDEADLK The calling thread has already locked the this mutex.
 *
 * @restriction On the OSE soft kernel the fast semaphore is used to
 *              implement mutexes. Non-pthread processes that use mutexes
 *              must therefore initiate their fast semaphore with the value
 *              0 before calling any mutex functions. The result of killing a
 *              process (or a pthread) that is locking or waiting on a
 *              pthread mutex is undefined.
 *
 * @seealso default_restrictions_calling_functions
 */
extern int
pthread_mutex_trylock(pthread_mutex_t * mutex);


/**
 * @function pthread_mutex_unlock
 *
 * @brief Unlocks the mutex.
 *
 * @input mutex  Mutex to be unlocked.
 *
 * @return  Zero if successful. Otherwise a non-zero error code.
 *
 * @error EINVAL   The mutex has not been properly initialized.
 *
 * @error EPERM    The calling thread does not own the mutex.
 *
 * @restriction On the OSE soft kernel the fast semaphore is used to
 *              implement mutexes. Non-pthread processes that use mutexes
 *              must therefore initiate their fast semaphore with the value
 *              0 before calling any mutex functions. The result of killing a
 *              process (or a pthread) that is locking or waiting on a
 *              pthread mutex is undefined.
 *
 * @seealso default_restrictions_calling_functions
 */
extern int
pthread_mutex_unlock(pthread_mutex_t * mutex);


/**
 * @function pthread_mutex_destroy
 *
 * @brief Destroys the mutex.
 *
 * @input mutex  Mutex to be destroyed.
 *
 * @return  Zero if successful. Otherwise a non-zero error code.
 *
 * @error EBUSY   The mutex is currently locked.
 *
 * @error EINVAL  The mutex has not been properly initialized.
 *
 * @restriction On the OSE soft kernel the fast semaphore is used to
 *              implement mutexes. Non-pthread processes that use mutexes
 *              must therefore initiate their fast semaphore with the value
 *              0 before calling any mutex functions. The result of killing a
 *              process (or a pthread) that is locking or waiting on a
 *              pthread mutex is undefined.
 *
 * @seealso default_restrictions_calling_functions
 */
extern int
pthread_mutex_destroy(pthread_mutex_t * mutex);


/**
 * @function pthread_mutexattr_destroy
 *
 * @brief Destroys a mutex attribute variable.
 *
 * @input attr   Attribute variable to destroy.
 *
 * @return Zero.
 *
 * @restriction Mutex attributes are not implemented.
 *              Calling this function has no effect.
 *
 * @seealso default_restrictions_calling_functions
 */
extern int
pthread_mutexattr_destroy(const pthread_mutexattr_t * attr);


/**
 * @function pthread_mutexattr_settype
 *
 * @brief Sets the mutex type attribute.
 *
 * @input attr  Attribute object.
 *
 * @input type  Requested mutex type.
 * 
 * @return Zero if successful. Otherwise a non-zero error code.
 *
 * @error EINVAL  The mutex attribute object or the type is invalid.
 *
 * @restriction Mutex attributes are not implemented.
 *              Calling this function has no effect.
 *
 * @seealso default_restrictions_calling_functions
 */
extern int
pthread_mutexattr_settype(const pthread_mutexattr_t * attr,
                          int type);


/**
 * @function pthread_cond_init
 *
 * @brief Initializes the condition variable.
 *
 * @input cond       Condition variable to be initialized.
 *
 * @input cond_attr  Attribute for the initialization.
 *
 * @return  Zero.
 *
 * @restriction Only available in pthreads.
 *
 * @seealso default_restrictions_calling_functions
 */
extern int
pthread_cond_init(pthread_cond_t * cond,
		  pthread_condattr_t * cond_attr);


/**
 * @function pthread_cond_destroy
 *
 * @brief Destroys the condition variable.
 *
 * @input cond  Condition variable to be destroyed.
 *
 * @return  Zero if successful. Otherwise a non-zero error code.
 *
 * @restriction Only available in pthreads.
 *
 * @error EBUSY  There are threads waiting on condition variable.
 *
 * @error EINVAL Specified condition variable is invalid.
 *
 * @seealso default_restrictions_calling_functions
 */
extern int
pthread_cond_destroy(pthread_cond_t * cond);


/**
 * @function pthread_cond_signal
 *
 * @brief Signals the condition variable.
 *
 * @input cond  Condition variable to signal.
 *
 * @return  Zero if successful. Otherwise a non-zero error code.
 *
 * @restriction Only available in pthreads.
 *
 * @error EINVAL Specified condition variable is invalid.
 *
 * @seealso default_restrictions_calling_functions
 */
extern int
pthread_cond_signal(pthread_cond_t * cond);


/**
 * @function pthread_cond_wait
 *
 * @brief Waits on the condition variable.
 *
 * @input cond   Condition variable to wait for.
 *
 * @input mutex  Mutex currently used for the condition variable.
 *
 * @return  Zero if successful. Otherwise a non-zero error code.
 *
 * @restriction Only available in pthreads.
 *
 * @error EINVAL Specified condition variable or mutex was invalid,
 *               or the different mutexes were used for the same
 *               condition variable,
 *               or the current thread is not the owner of the mutex.
 *
 * @seealso default_restrictions_calling_functions
 */
extern int
pthread_cond_wait(pthread_cond_t * cond,
		  pthread_mutex_t * mutex);


/**
 * @function pthread_cond_timedwait
 *
 * @brief Waits with timeout on the condition variable.
 *
 * @input cond     Condition variable to wait for.
 *
 * @input mutex    Mutex currently used for the condition variable.
 *
 * @input abstime  Specification of a future moment for an eventual timeout. The
 *                 current absolute system time can be retrieved as a timeval
 *                 structure using the system clock interface gettimeofday().
 *
 * @return  Zero if successful. Otherwise a non-zero error code.
 *
 * @restriction Only available in pthreads.
 *
 * @error ETIMEDOUT  Condition variable was not signaled before
 *                   the specified time.
 *
 * @error EINVAL Specified condition variable or mutex was invalid,
 *               or the different mutexes were used for the same
 *               condition variable,
 *               or the current thread is not the owner of the mutex,
 *               or the specified time was invalid.
 *
 * @seealso default_restrictions_calling_functions
 */
extern int
pthread_cond_timedwait(pthread_cond_t * cond,
		       pthread_mutex_t * mutex,
		       const struct timespec *abstime);


/**
 * @function pthread_cond_broadcast
 * 
 * @brief Performs a broadcast signal on the condition variable.
 *
 * @input cond  Condition variable to be broadcast signaled.
 *
 * @return  Zero if successful. Otherwise a non-zero error code.
 *
 * @restriction Only available in pthreads.
 *
 * @error EINVAL Specified condition variable is invalid.
 *
 * @seealso default_restrictions_calling_functions
 */
extern int
pthread_cond_broadcast(pthread_cond_t * cond);


extern void
xxcleanup_push(Cleanelem * topush);


extern void
xxcleanup_pop(int execute);


/**
 * @function pthread_key_create
 *
 * @brief Creates a thread specific data key.
 *
 * @output key             The created thread specific data key.
 *
 * @input destr_function   Destruction function for the created key.
 *
 * @return Zero if successful. Otherwise a non-zero error code.
 *
 * @error EAGAIN  PTHREAD_KEYS_MAX keys are already created.
 *
 * @restriction  A regular OSE process won't exit or be cancelled like
 *               a pthread and the destructor associated with a key
 *               will never be called.
 *
 * @seealso default_restrictions_calling_functions
 */
extern int
pthread_key_create(pthread_key_t * key,
		   void (*destr_function) (void *));


/**
 * @function pthread_key_delete
 *
 * @brief Deletes a thread specific data key.
 *
 * @input key  Thread specific data key to be deleted.
 *
 * @return Zero if successful. Otherwise a non-zero error code.
 *
 * @error EINVAL  The specified key is not valid.
 *
 * @seealso default_restrictions_calling_functions
 */
extern int
pthread_key_delete(pthread_key_t key);



/**
 * @function pthread_setspecific
 *
 * @brief Sets the value of the given thread specific data key.
 *
 * @input key    Thread specific data key to associate value with.
 *
 * @input value  Actual value to associate to key.
 *
 * @return Zero if successful. Otherwise a non-zero error code.
 *
 * @error EINVAL  The key is invalid.
 *
 * @seealso default_restrictions_calling_functions
 */
extern int
pthread_setspecific(pthread_key_t key,
		    const void *value);

/**
 * @function pthread_getspecific
 *
 * @brief Returns the value corresponding to the given key.
 *
 * @input key  Key used to retrieve thread specific data from.
 *
 * @return  The retrieved value, or NULL if not found.
 *
 * @seealso default_restrictions_calling_functions
 */
extern void *
pthread_getspecific(pthread_key_t key);


/**
 * @function pthread_equal
 * 
 * @brief Returns non-zero if the given thread IDs are equal.
 *
 * @input thread1  First thread ID.
 *
 * @input thread2  Second thread ID.
 *
 * @return Non-zero if the thread IDs are equal, zero otherwise.
 *
 * @restriction Also available in interrupt and timer interrupt processes.
 *
 * @seealso default_restrictions_calling_functions
 */
extern int
pthread_equal(pthread_t thread1,
	      pthread_t thread2);


/**
 * @function pthread_attr_init
 *
 * @brief Initializes the pthread attribute.
 *
 * @input attr  Attribute to initialize according to the default values.
 *
 * @return  Zero.
 *
 * @restriction Also available in interrupt and timer interrupt processes.
 *
 * @seealso default_restrictions_calling_functions
 */
extern int
pthread_attr_init(pthread_attr_t * attr);




/**
 * @function pthread_attr_destroy
 *
 * @brief Destroys the pthread attribute.
 *
 * @input attr  Attribute to destroy.
 *
 * @return  Zero if successful. Otherwise a non-zero error code.
 *
 * @restriction Also available in interrupt and timer interrupt processes.
 *
 * @error EINVAL  The specified attribute was not valid.
 *
 * @seealso default_restrictions_calling_functions
 */
extern int
pthread_attr_destroy(pthread_attr_t * attr);



/**
 * @function pthread_attr_setscope
 * @brief Sets the contention cope of the pthread attribute.
 *
 * @input attr             Attribute to change.
 *
 * @input contentionscope  New contentionscope for the attribute.
 *
 * @return  Zero if successful. Otherwise a non-zero error code.
 *
 * @restriction Also available in interrupt and timer interrupt processes.
 *
 * @error EINVAL  The specified attribute was not valid.
 *
 * @error ENOTSUP The specified scope is not supported.
 *
 * @seealso default_restrictions_calling_functions
 */
extern int
pthread_attr_setscope(pthread_attr_t * attr,
		      int contentionscope);


/**
 * @function pthread_attr_getscope
 *
 * @brief Gets the contention cope of the pthread attribute.
 *
 * @input attr              Attribute to change.
 *
 * @output contentionscope  Current contentionscope in attribute.
 *
 * @return  Zero if successful. Otherwise a non-zero error code.
 *
 * @restriction Also available in interrupt and timer interrupt processes.
 *
 * @error EINVAL  The specified attribute was not valid.
 *
 * @seealso default_restrictions_calling_functions
 */
extern int
pthread_attr_getscope(const pthread_attr_t * attr,
		      int *contentionscope);


/**
 * @function pthread_attr_setstopped
 *
 * @brief The thread created with this attribute will not start when created.
 *
 * @long  The function is an OSE extension to POSIX.
 *
 * @input attr     Attribute to change.
 *
 * @input stopped  Non-zero if created as stopped.
 *
 * @return  Zero if successful. Otherwise a non-zero error code.
 *
 * @restriction Also available in interrupt and timer interrupt processes.
 *
 * @error EINVAL  The specified attribute was not valid.
 *
 * @seealso start
 * @seealso default_restrictions_calling_functions
 */
extern int
pthread_attr_setstopped(pthread_attr_t * attr,
			int stopped);


/**
 * @function pthread_attr_getstopped
 *
 * @brief Returns the value of the pthread attribute stopped.
 *
 * @long  The function is an OSE extension to POSIX.
 *
 * @input attr      Attribute to observe.
 *
 * @output stopped  Non-zero if created as stopped.
 *
 * @return  Zero if successful. Otherwise a non-zero error code.
 *
 * @restriction Also available in interrupt and timer interrupt processes.
 *
 * @error EINVAL  The specified attribute was not valid.
 *
 * @seealso default_restrictions_calling_functions
 */
extern int
pthread_attr_getstopped(const pthread_attr_t * attr,
			int *stopped);


/**
 * @function pthread_attr_setdetachstate
 *
 * @brief Sets the detach state of the pthread attribute.
 *
 * @input attr         Attribute to change.
 *
 * @input detachstate  New detach state.
 *
 * @return  Zero if successful. Otherwise a non-zero error code.
 *
 * @restriction Also available in interrupt and timer interrupt processes.
 *
 * @error EINVAL  The specified attribute was not valid.
 *
 * @seealso default_restrictions_calling_functions
 */
extern int
pthread_attr_setdetachstate(pthread_attr_t * attr,
			    int detachstate);


/**
 * @function pthread_attr_getdetachstate
 *
 * @brief Returns the value of the pthread attribute for detach state.
 *
 * @input attr          Attribute to observe.
 *
 * @output detachstate  Current detach state in attribute.
 *
 * @return  Zero if successful. Otherwise a non-zero error code.
 *
 * @restriction Also available in interrupt and timer interrupt processes.
 *
 * @error EINVAL  The specified attribute was not valid.
 *
 * @seealso default_restrictions_calling_functions
 */
extern int
pthread_attr_getdetachstate(const pthread_attr_t * attr,
			    int *detachstate);


/**
 * @function pthread_attr_setstacksize
 *
 * @brief Sets the stack size of the pthread attribute.
 *
 * @input attr       Attribute to change.
 *
 * @input stacksize  Stack size to configure in attribute.
 *
 * @return  Zero if successful. Otherwise a non-zero error code.
 *
 * @restriction Also available in interrupt and timer interrupt processes.
 *
 * @error EINVAL  The specified attribute was not valid.
 *
 * @seealso default_restrictions_calling_functions
 */
extern int
pthread_attr_setstacksize(pthread_attr_t * attr,
			  size_t stacksize);


/**
 * @function pthread_attr_getstacksize
 *
 * @brief Returns the stack size of the pthread attribute.
 *
 * @input attr        Attribute to observe.
 *
 * @output stacksize  Current stack size in attribute.
 *
 * @return  Zero if successful. Otherwise a non-zero error code.
 *
 * @restriction Also available in interrupt and timer interrupt processes.
 *
 * @error EINVAL  The specified attribute was not valid.
 *
 * @seealso default_restrictions_calling_functions
 */
extern int
pthread_attr_getstacksize(const pthread_attr_t * attr,
			  size_t * stacksize);


/**
 * @function pthread_attr_setschedparam
 *
 * @brief Sets the scheduling parameter of the pthread attribute.
 *
 * @input attr   Attribute to change.
 *
 * @input param  Scheduling parameter to apply to attr.
 *
 * @return  Zero if successful. Otherwise a non-zero error code.
 *
 * @restriction Also available in interrupt and timer interrupt processes.
 *
 * @error EINVAL  The specified attribute was not valid.
 *
 * @seealso default_restrictions_calling_functions
 */
extern int
pthread_attr_setschedparam(pthread_attr_t * attr,
			   const struct sched_param *param);


/**
 * @function pthread_attr_getschedparam
 *
 * @brief Returns the scheduling parameter of the pthread attribute.
 *
 * @input attr    Attribute to observe.
 *
 * @output param  Current scheduling parameter of attribute.
 *
 * @return  Zero if successful. Otherwise a non-zero error code.
 *
 * @restriction Also available in interrupt and timer interrupt processes.
 *
 * @error EINVAL  The specified attribute was not valid.
 *
 * @seealso default_restrictions_calling_functions
 */
extern int
pthread_attr_getschedparam(const pthread_attr_t * attr,
			   struct sched_param *param);


/**
 * @function pthread_attr_setschedpolicy
 *
 * @brief Sets the scheduling policy of the pthread attribute.
 *
 * @input attr    Attribute to change.
 *
 * @input policy  New scheduling policy for attribute.
 *
 * @return  Zero if successful. Otherwise a non-zero error code.
 *
 * @restriction Also available in interrupt and timer interrupt processes.
 *
 * @error EINVAL  The specified attribute was not valid.
 *
 * @error ENOTSUP The specified policy is not supported.
 *
 * @seealso default_restrictions_calling_functions
 */
extern int
pthread_attr_setschedpolicy(pthread_attr_t * attr,
			    int policy);


/**
 * @function pthread_attr_getschedpolicy
 *
 * @brief Returns the scheduling policy of the pthread attribute.
 *
 * @input attr     Attribute to observe.
 *
 * @output policy  Current scheduling policy from attribute.
 *
 * @return  Zero if successful. Otherwise a non-zero error code.
 *
 * @restriction Also available in interrupt and timer interrupt processes.
 *
 * @error EINVAL  The specified attribute was not valid.
 *
 * @seealso default_restrictions_calling_functions
 */
extern int
pthread_attr_getschedpolicy(const pthread_attr_t * attr,
			    int *policy);


/**
 * @function pthread_attr_setinheritsched
 *
 * @brief Sets the inherit sched property of the pthread attribute.
 *
 * @input attr          Attribute to change.
 *
 * @input inheritsched  Inherit scheduling policy to apply to attribute.
 *
 * @return  Zero if successful. Otherwise a non-zero error code.
 *
 * @restriction Also available in interrupt and timer interrupt processes.
 *
 * @error EINVAL  The specified attribute was not valid, or
 *                the inheritsched parameter was not one of
 *                PTHREAD_INHERIT_SCHED or PTHREAD_EXPLICIT_SCHED.
 *
 * @seealso default_restrictions_calling_functions
 */
extern int
pthread_attr_setinheritsched(pthread_attr_t * attr,
			     int inheritsched);


/**
 * @function pthread_attr_getinheritsched
 *
 * @brief Returns the inherit sched property of the pthread attribute.
 *
 * @input attr           Attribute to observe.
 *
 * @output inheritsched  Current inherit scheduling of attribute.
 *
 * @return  Zero if successful. Otherwise a non-zero error code.
 *
 * @restriction Also available in interrupt and timer interrupt processes.
 *
 * @error ENIVAL  The specified attribute was not valid.
 *
 * @seealso default_restrictions_calling_functions
 */
extern int
pthread_attr_getinheritsched(const pthread_attr_t * attr,
			     int *inheritsched);


/**
 * @function pthread_attr_setblock
 *
 * @brief Sets the OSE block of the pthread attribute.
 *
 * @long  The function is an OSE extension to POSIX.
 *
 * @input attr      Attribute to change.
 *
 * @input block_pid  Block PID to set on attribute.
 *
 * @return   Zero if successful. Otherwise a non-zero error code.
 *
 * @restriction Also available in interrupt and timer interrupt processes.
 *
 * @error ENIVAL  The specified attribute was not valid.
 *
 * @seealso default_restrictions_calling_functions
 */
extern int
pthread_attr_setblock(pthread_attr_t * attr,
		      PROCESS block_pid);


/**
 * @function pthread_attr_getblock
 *
 * @brief Returns the OSE block of the pthread attribute.
 *
 * @long  The function is an OSE extension to POSIX.
 *
 * @input attr       Attribute to observe.
 *
 * @output block_pid  Block PID currently set in attribute.
 *
 * @return   Zero if successful. Otherwise a non-zero error code.
 *
 * @restriction Also available in interrupt and timer interrupt processes.
 *
 * @error ENIVAL  The specified attribute was not valid.
 *
 * @seealso default_restrictions_calling_functions
 */
extern int
pthread_attr_getblock(const pthread_attr_t * attr,
		      PROCESS * block_pid);


/**
 * @function pthread_attr_setname
 *
 * @brief Sets the OSE process name of the pthread attribute.
 *
 * @long  The function is an OSE extension to POSIX.
 *
 * @input attr  Attribute to change.
 *
 * @input name  OSE Process name to set in attribute.
 *
 * @return   Zero if successful. Otherwise a non-zero error code.
 *
 * @restriction Also available in interrupt and timer interrupt processes.
 *
 * @error ENIVAL  The specified attribute was not valid.
 *
 * @seealso default_restrictions_calling_functions
 */
extern int
pthread_attr_setname(pthread_attr_t * attr,
		     const char *name);


/**
 * @function pthread_attr_getname
 *
 * @brief Returns the OSE name of the pthread attribute.
 *
 * @long  The function is an OSE extension to POSIX.
 *
 * @input attr   Attribute to observe.
 *
 * @output name  Current OSE Process name set in attribute.
 *
 * @return   Zero if successful. Otherwise a non-zero error code.
 *
 * @restriction Also available in interrupt and timer interrupt processes.
 *
 * @error ENIVAL  The specified attribute was not valid.
 *
 * @seealso default_restrictions_calling_functions
 */
extern int
pthread_attr_getname(const pthread_attr_t * attr,
		     const char **name);


/**
 * @function pthread_setschedparam
 *
 * @brief Sets the sched param of the pthread attribute.
 *
 * @input thread  Thread to change scheduling properties on.
 *
 * @input policy  New scheduling policy to apply to thread.
 *
 * @input param   New scheduling parameter to apply to thread.
 *
 * @return  Zero if successful. Otherwise a non-zero error code.
 *
 * @error ESRCH  No thread could be found with the specified identity.
 *
 * @error EINVAL  The specified policy or the scheduling parameters
 *                are invalid.
 *
 * @error ENOTSUP The specified scheduling policy is not supported.
 *
 * @seealso default_restrictions_calling_functions
 */
extern int
pthread_setschedparam(pthread_t thread,
		      int policy,
		      const struct sched_param *param);

/**
 * @function pthread_getschedparam
 *
 * @brief Returns the sched param of the pthread attribute.
 *
 * @input thread   Thread to retrieve scheduling properties from.
 *
 * @output policy  Current scheduling policy of thread.
 *
 * @output param   Current scheduling parameter of thread.
 *
 * @return  Zero if successful. Otherwise a non-zero error code.
 *
 * @error ESRCH  No thread could be found with the specified identity.
 *
 * @seealso default_restrictions_calling_functions
 */
extern int
pthread_getschedparam(pthread_t thread,
		      int *policy,
		      struct sched_param *param);


/**
 * @function pthread_once
 *
 * @brief Thread safe functionality to execute an init() function once.
 *
 * @input once_control Structure that determines whether init routine
 *                     has been called or not.
 *
 * @input init_func    Function to be called once and not in any subsequent
 *                     calls.
 *
 * @return  Zero if successful. Otherwise a non-zero error code.
 *
 * @error EINVAL  once-control or init_func is NULL.
 *
 * @seealso default_restrictions_calling_functions
 */
extern int
pthread_once(pthread_once_t *once_control,
	     void (*init_func)(void));

/*
 * C++ handling for soft environments.
 */

#ifdef __cplusplus
} /* end of extern "C" */
#endif

#ifdef softOSE
#define ZZMK_INLINE_FUNCTIONS
#include "sfk/sfk_pthread.h"
#endif

/* End #ifndef _PTHREAD_H */
#endif
