/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2011 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */
/**
 * @public
 * @toc OSE:
 * @file ramlog.h
 *
 * @brief This file defines the interface to the RAMLOG utility.
 *
 * @long
 * This file contains prototypes and defines for the RAMLOG
 * utility.
 * The RAMLOG interface is invoked by services in OSE or from
 * application or error handlers when it is desired to safely
 * log an event. The RAMLOG utility must not use any OSE service,
 * it is reentrant and is therefore safe to use from the very
 * beginning and also from interrupt level or exception level.
 *
 * The functions declared in this interface can also be
 * defined by the user in order to override the definitions in the kernel,
 * for example an own ramlog.c.
 *
 */

#ifndef OSE_RAMLOG_H_
#define OSE_RAMLOG_H_

#include <stddef.h>

#ifdef __cplusplus
extern "C" {
#endif

   /*
    *==========================================================================
    *       Pre-defined strings.
    *==========================================================================
    */

/**
 * @macro RAMLOG_STR_SESSION_START
 * @brief RAMLOG_STR_SESSION_START will be written to the RAMLOG buffer by
 *        the RAMLOG implementation after each system (re)start.
 */
#define RAMLOG_STR_SESSION_START "__RAMLOG_SESSION_START__\n"

/**
 * @type RamlogVirtAddr
 * @brief Type for the virtual offset in bytes within the RAMLOG's virtual
 *        address space.
 * @long  The RAMLOG entries are stored in a circular buffer whose virtual
 *        offset moves upwards with each new entry, without being reset between
 *        warm restarts. At any given moment, only the virtual offsets between
 *        the 'oldest' and 'newest' data correspond to
 *        physically present data in the buffer. The virtual offset of the
 *        circular buffer will overflow each time the size of the entries
 *        written exceeds the size representable by an "unsigned int".
 */
   /* Choose unsigned int which should correspond to the machine word
    * for automatic modulo 2^word_bits when 'C' arithmetic is performed.
    */
typedef unsigned int RamlogVirtAddr;


/**
 * @function ramlog_addr_of_oldest
 *
 * @brief  Find out the virtual offset address of the oldest
 *         entry in the RAMLOG buffer.
 * @return The virtual address of oldest data entry in the RAMLOG
 *         (range from 0 to ~0).
 * @seealso default_restrictions_calling_functions
 */
RamlogVirtAddr ramlog_addr_of_oldest(void);


/**
 * @function ramlog_addr_of_newest
 *
 * @brief  Find out the virtual offset address of the newest (next) entry
 *         present in the RAMLOG buffer.
 * @return The virtual address where the next character to be written would
 *	   land (range from 0 to ~0).
 * @seealso default_restrictions_calling_functions
 */
RamlogVirtAddr ramlog_addr_of_newest(void);


/**
 * @function ramlog_find_next
 *
 * @brief Find in the RAMLOG the next occurrence of the submitted pattern.
 * @long  Starts at the virtual address following the one submitted,
 *        i.e. 'whence' + 1.
 * @input whence  The virtual address from which the search starts,
 *		     previous to the first expected hit. If 'whence' is outside
 *		     the actual RAMLOG range at the moment, the search will
 *		     begin at the address of the 'oldest' data.
 * @input pattern  Pointer to the NULL-terminated string to search for in
 *		      the RAMLOG. If 'pattern' is NULL or a pointer to NULL,
 *		      no search will be performed and the value of 'whence'
 *		      will be returned.
 * @return The virtual address of the found pattern, or the value of
 *	      'whence' if the pattern could not be found.
 * @seealso default_restrictions_calling_functions
 *
 * @example
 * To find the oldest occurrence of a given pattern:
 *	addr = ramlog_find_next(ramlog_addr_of_oldest() - 1, pattern);
 */
RamlogVirtAddr ramlog_find_next(RamlogVirtAddr whence, const char *pattern);


/**
 * @function ramlog_find_previous
 * @brief   Find in the RAMLOG the previous occurrence of the submitted
 *	       pattern.
 * @long    Starts at the virtual address previous to the one submitted,
 *          i.e. 'whence - 1'.
 * @input whence   The virtual address from which the search starts,
 *		      higher than the first expected hit. If 'whence' falls
 *		      outside the actual RAMLOG range at the moment, the search
 *		      will begin at the address of 'newest' data.
 * @input pattern  Pointer to the NULL-terminated string to search for in
 *		      the RAMLOG. If 'pattern' is NULL or a pointer to NULL,
 *		      no search will be performed and the value of 'whence'
 *		      will be returned.
 * @return The virtual address of the found pattern, or the value of
 *	      'whence' if the pattern could not be found.
 * @seealso default_restrictions_calling_functions
 */
RamlogVirtAddr ramlog_find_previous(RamlogVirtAddr whence,
				       const char *pattern);

/**
 * @function ramlog_read
 *
 * @brief  Read from and/or beyond the specified virtual address.
 * @long   This function copies 'size' number of characters from the
 *	      RAMLOG to 'buf', starting at 'req_addr', if 'req_addr' is
 *	      within the current RAMLOG range, or from the oldest otherwise.
 *	      The actual size differs from the value of 'size' either if
 *	      there are less RAMLOG contents than 'size', or if 'size' exceeds
 *	      the RAMLOG physical buffer size, which is the maximum read size.
 *	      The variable pointed to by 'actual_addr' is updated with the
 *	      virtual address actually read from.
 * @input buf        Pointer to the buffer where results are to be copied.
 *                   If 'buf' is NULL, no read will be performed.
 * @output buf       The contents of the buffer pointed to by buf is updated.
 * @input  size      Size of the buffer pointed to by 'buf'.
 * @input  req_addr  Requested virtual address for the read.
 * @output actual_addr Pointer to a variable that is updated by the function
 *			with the actual address from which the read was done.
 *			May differ from 'req_addr' as explained above. If this
 *			argument is NULL, it will be ignored.
 * @return The number of characters read.
 * @seealso default_restrictions_calling_functions
 */
size_t ramlog_read(char *buf, size_t size,
                   RamlogVirtAddr req_addr,
                   RamlogVirtAddr *actual_addr);

/**
 * @function ramlog_write
 *
 * @brief   Writes 'size' characters of preformatted ASCII data
 *	       from 'buf' to the RAMLOG.
 * @input  buf  The string to write to the RAMLOG. If 'buf' equals NULL,
 *		   no writing will be done and the function will return 0.
 * @input  size The length of the string. The maximum write size will
 *		   be the RAMLOG buffer size.
 * @return The number of characters written.
 * @restriction This function requires supervisor mode if executed in the
 *         core module, but works in user-mode in separately loaded
 *         programs, where it is implemented by the CRT.
 *    @break
 *         Also available in interrupt and timer interrupt processes.
 * @seealso default_restrictions_calling_functions
 */
size_t ramlog_write(const char *buf, size_t size);


/**
 * @macro RAMLOG_PRINTF_MAXSIZE
 * @brief Maximum number of formatted characters for ramlog_printf.
 * @long  This value is currently 256.
 * @seealso ramlog_printf
 */
#define RAMLOG_PRINTF_MAXSIZE 256

/**
 * @function ramlog_printf
 *
 * @brief Format and write an ASCII string to the RAMLOG.
 * @input fmt Pointer to the string to format and write to the RAMLOG,
 *		 followed by variable number of arguments. If 'fmt' is
 *		 NULL, the function will gracefully return 0. The function
 *		 will have a maximum internal buffering capability of
 *		 'RAMLOG_PRINTF_MAXSIZE' number of bytes (used to format
 *		 the output string). Output strings needing more than that
 *		 will be gracefully truncated.
 * @input ...
 * @return The number of characters written.
 * @restriction This function requires supervisor mode if executed in the
 *         core module, but works in user-mode in separately loaded
 *         programs, where it is implemented by the CRT.
 *   @break
 *         Also available in interrupt and timer interrupt processes.
 *         
 * @seealso RAMLOG_PRINTF_MAXSIZE
 * @seealso default_restrictions_calling_functions
 * @example
 * nr_chars = ramlog_printf("%s: %s\n", app_name, error_str);
 */
size_t ramlog_printf(const char *fmt, ...);

/**
 * @function ramlog_set_oldest
 *
 * @brief Set the oldest virtual offset pointer in the RAMLOG.
 * @input new_oldest The new value of the oldest pointer.
 * @return The actual new oldest pointer.
 * @restriction This function requires supervisor mode if executed in the
 *         core module, but works in user-mode in separately loaded
 *         programs, where it is implemented by the CRT.
 * @seealso ramlog_addr_of_oldest
 * @seealso ramlog_addr_of_newest
 * @seealso ramlog_addr_of_oldest,ramlog_addr_of_newest
 * @seealso default_restrictions_calling_functions
 * @example
 * RamlogVirtAddr newest = ramlog_addr_of_newest();
 * ramlog_set_oldest(newest);
 */
RamlogVirtAddr ramlog_set_oldest(RamlogVirtAddr new_oldest);

#ifdef __cplusplus
}
#endif
#endif
