/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2011 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */
/**
 * @toc File_System:
 *
 * @file shell.h
 *
 * @brief OSE Shell interface.
 */

#ifndef _SHELL_H
#define _SHELL_H

#include "ose.h"
#include "osetypes.h"

#ifdef __cplusplus
extern "C" {
#endif

/**
 * @macro DEFAULT_PROC_TYPE
 *
 * @include        "shell.h"
 *
 * @declaration #define DEFAULT_PROC_TYPE 242
 *
 * @brief The command has same process type as the parent shell.
 *
 * @long Symbolic constant for @link "" "shell_add_cmd_attrs".
 */
#define DEFAULT_PROC_TYPE 242

/**
 * @macro DEFAULT_PRIORITY
 *
 * @include        "shell.h"
 *
 * @declaration #define DEFAULT_PRIORITY  255
 *
 * @brief The command has same priority as the parent shell.
 *
 * @long Symbolic constant for @link "" "shell_add_cmd_attrs".
 */
#define DEFAULT_PRIORITY  255

/**
 * @macro DEFAULT_STACK_SIZE
 *
 * @include        "shell.h"
 *
 * @declaration #define DEFAULT_STACK_SIZE  0
 *
 * @brief Set the command process stack to the default size (4000 bytes).
 *
 * @long Symbolic constant for @link "" "shell_add_cmd_attrs".
 */
#define DEFAULT_STACK_SIZE  0

/**
 * @type CmdType
 *
 * @include        "shell.h"
 *
 * @declaration typedef int (* CmdType) (int argc, char ** argv);
 *
 * @input argc The argument count.
 *
 * @input argv The argument vector.
 *
 * @brief The type for a shell command function.
 */
typedef int (* CmdType) (int argc, char ** argv);

/**
 * @macro RET_SUCCESS
 *
 * @include        "shell.h"
 *
 * @declaration #define RET_SUCCESS 0
 *
 * @brief Command finished successfully.
 *
 * @long Symbolic return code for shell commands.
 */
#define RET_SUCCESS 0

/**
 * @macro RET_WARNING
 *
 * @include        "shell.h"
 *
 * @declaration #define RET_WARNING 10
 *
 * @brief The command finished successfully but with a warning.
 *
 * @long Symbolic return code for shell commands.
 */
#define RET_WARNING 10

/**
 * @macro RET_ERROR
 *
 * @include        "shell.h"
 *
 * @declaration #define RET_ERROR 20
 *
 * @brief Command failed.
 *
 * @long Symbolic return code for shell commands.
 */
#define RET_ERROR 20

/**
 * @type ShellRunCmdTag
 *
 * @include "shell.h"
 *
 * @declaration
 * enum ShellRunCmdTag
 * {
 *  RUNCMD_TAGEND        = 0x0,
 *  RUNCMD_EFS_CLONE     = 0x1,
 *  RUNCMD_CLONE_ENV     = 0x2,
 *  RUNCMD_NOPARENT      = 0x3,
 *  RUNCMD_ADDUSER       = 0x4,
 *  RUNCMD_PIDPTR        = 0x5,
 *  RUNCMD_ASYNCHRONOUS  = 0x6,
 *  RUNCMD_INTERRUPTABLE = 0x8,
 *  RUNCMD_RETURN_STATUS = 0x9,
 *  RUNCMD_RETURN_ERRNO  = 0xA,
 *  RUNCMD_FDARRAY       = 0xC,
 *  RUNCMD_ENVLIST       = 0xD
 * };
 *
 * @brief Definition of tags used in @link "" "shell_run_cmd" signaling.
 * @long  In private mode resources are associated with the
 *        shell command process, whereas in shared mode resources are
 *        associated with the shell command block,
 *        in which the shell command process executes.
 *        Below, the interpretation of the term process/block
 *        depends on the mode applicable to the calling process.
 *
 * @field RUNCMD_TAGEND          The tag list endmark. This must be the last
 *                               tag in the tag list.
 * @field RUNCMD_EFS_CLONE       Clone all open file descriptors, umask and cwd
 *                               from the process/block calling shell_run_cmd()
 *                               to the shell command process/block. Also set
 *                               the calling process as parent to the shell
 *                               command process/block.
 * @field RUNCMD_CLONE_ENV       Clone all string environment variables from the
 *                               process/block calling shell_run_cmd()
 *                               to the shell command process/block.
 * @field RUNCMD_NOPARENT        Do not set a PARENT on the shell command
 *                               process/block.The return value is always success
 *                               when using this tag.The function waits for the
 *                               command to finish or to be killed.                              
 * @field RUNCMD_ADDUSER         Specify a specific user and terminal to run
 *                               the command. The tag RMDCMD_ADDUSER shall be
 *                               followed by two tag elements with a
 *                               char *username and a char *terminal.
 * @field RUNCMD_PIDPTR          Request return of the command process ID in
 *                               the tag structure. The tag RMDCMD_PIDPTR shall
 *                               be followed by a tag element with
 *                               PROCESS *pidptr. The tag is
 *                               mainly used in combination with
 *                               RUNCMD_ASYNCHRONOUS.
 * @field RUNCMD_ASYNCHRONOUS	 Do not wait for the shell command to return
 *                               before shell_run_cmd() returns.
 * @field RUNCMD_INTERRUPTABLE   The shell command is interruptable, "ctrl-c"
 *                               in the shell. The shell command is interrupted
 *                               by sending a signal with the signal number
 *                               "FM_SIGNAL", see ose_spi/fm.sig, to the
 *                               caller of shell_run_cmd(). The tag is not
 *                               compatible with RUNCMD_ASYNCHRONOUS.
 * @field RUNCMD_RETURN_STATUS   Request return of the command return status in
 *                               the tag structure. The tag shall be followed
 *                               by a tag element with an int *status. The tag
 *                               is not compatible with RUNCMD_ASYNCHRONOUS 
 *                               and RUNCMD_NOPARENT.
 * @field RUNCMD_RETURN_ERRNO	 Request return of the command's errno in the
 *                               tag structure. The tag shall be followed
 *                               by a tag element with an int *errno. The tag
 *                               is not compatible with RUNCMD_ASYNCHRONOUS 
 *                               and RUNCMD_NOPARENT.
 * @field RUNCMD_FDARRAY         Clone the file descriptors of the fdarray to
 *                               the shell command process. The tag
 *                               RMDCMD_FDARRAY shall be followed by a
 *                               tag elements with an int *fd_array.
 *                               The fd_array is an array of int. The
 *                               first element of the fdarray specify how many
 *                               fds are defined in the array. The other
 *                               elements, element 1 to n, of the fdarray
 *                               specify the file descriptors of the array.
 * @field RUNCMD_ENVLIST         Set the environment variables of the envlist
 *                               on the shell process. The tag RMDCMD_ENVLIST
 *                               shall be followed by a tag element with a
 *                               char *envlist[]. The
 *                               envlist is an array of char *. Each element of
 *                               the array specify one environment variable
 *                               using the syntax <name>=<value>. The last
 *                               element of the array shall always be a
 *                               NULL pointer.
 *
 * @seealso shell_run_cmd
 */

enum ShellRunCmdTag
{
    RUNCMD_TAGEND        = 0x0,
    RUNCMD_EFS_CLONE     = 0x1,
    RUNCMD_CLONE_ENV     = 0x2,
    RUNCMD_NOPARENT      = 0x3,
    RUNCMD_ADDUSER       = 0x4,
    RUNCMD_PIDPTR        = 0x5,
    RUNCMD_ASYNCHRONOUS	 = 0x6,
    RUNCMD_INTERRUPTABLE = 0x8,
    RUNCMD_RETURN_STATUS = 0x9,
    RUNCMD_RETURN_ERRNO	 = 0xA,
    RUNCMD_FDARRAY       = 0xC,
    RUNCMD_ENVLIST       = 0xD
};


/**
 * @type ShellRunCmdStatus
 *
 * @include "shell.h"
 *
 * @declaration
 * enum ShellRunCmdStatus
 * {
 *  SHELL_SUCCESS         = 0x0,
 *  SHELL_EUNKNOWN_TAG    = 0x1,
 *  SHELL_EMISSING_ARG    = 0x2,
 *  SHELL_EMISSING_CMD    = 0x3,
 *  SHELL_EUNKNOWN_CMD    = 0x4,
 *  SHELL_ETERMINATED     = 0x5,
 *  SHELL_ECLONEFD_FAILED = 0x6,
 *  SHELL_EINVALID_FD	  = 0x7,
 *  SHELL_EILLEGAL_CMD	  = 0x8
 * };
 *
 * @brief Definition of statuses returned by
 *        @link "" "shell_run_cmd".
 *
 * @field SHELL_SUCCESS           Command executed and exited gracefully.
 *                                If RUNCMD_ASYNCHRONOUS was set, this is
 *                                returned when the command has been started,
 *                                but without waiting for it to terminate.
 *                                If RUNCMD_NOPARENT was set this is returned,
 *                                no matter what the command status was and if 
 *                                was succesfully executed.
 * @field SHELL_EUNKNOWN_TAG      Unknown tag specified in taglist.
 * @field SHELL_EMISSING_ARG      Missing argument to option tag.
 * @field SHELL_EMISSING_CMD	  No command specified.
 * @field SHELL_EUNKNOWN_CMD	  Unknown command specified.
 * @field SHELL_ETERMINATED	  Command terminated before exiting gracefully.
 * @field SHELL_ECLONEFD_FAILED   Failed to clone file descriptors.
 * @field SHELL_EINVALID_FD	  Failed to examine file descriptor.
 * @field SHELL_EILLEGAL_CMD	  The command is not allowed to be run.
 * @seealso shell_run_cmd
 */

enum ShellRunCmdStatus
{
    SHELL_SUCCESS         = 0x0,
    SHELL_EUNKNOWN_TAG    = 0x1,
    SHELL_EMISSING_ARG    = 0x2,
    SHELL_EMISSING_CMD	  = 0x3,
    SHELL_EUNKNOWN_CMD	  = 0x4,
    SHELL_ETERMINATED	  = 0x5,
    SHELL_ECLONEFD_FAILED = 0x6,
    SHELL_EINVALID_FD	  = 0x7,
    SHELL_EILLEGAL_CMD	  = 0x8
};

/**
 *
 * @function shell_add_cmd
 *
 * @include        "shell.h"
 *
 * @declaration void
 * shell_add_cmd(const char *cmd_name, const char *usage,
 *               const char *description, CmdType hook)
 *
 * @brief shell_add_cmd() registers a shell command.
 *
 * @long  shell_add_cmd() registers a resident shell command with the shell
 *        daemon overriding any existing command with the same name.
 *
 * @input cmd_name      Name of command.
 * @input usage         Usage string for the command, used by help.
 * @input description   Description string for the command, used by help.
 * @input hook          Entrypoint of the command.
 *
 * @return void
 *
 * @restriction Also available from START_OSE_HOOK2 context.
 *
 * @seealso CmdType
 * @seealso shell_add_cmd_attrs
 * @seealso shell_remove_cmd
 * @seealso default_restrictions_calling_functions
 *
 * @example
 * static int
 * cmd_hello(int argc, char *argv[])
 * {
 *    printf("Hello world\n");
 *    return 0;
 * }
 *
 * void init_hello_cmd(void)
 * {
 *    shell_add_cmd("hello", "hello", "print hello world", cmd_hello);
 * }
 */
void shell_add_cmd(const char *cmd_name, const char *usage,
		   const char *description, CmdType hook);

/**
 *
 * @function shell_add_cmd_attrs
 *
 * @include        "shell.h"
 *
 * @declaration void
 * shell_add_cmd_attrs(const char *cmd_name, const char *usage,
 *                     const char *description, CmdType hook,
 *                     enum PROCESS_TYPE proc_type, OSPRIORITY
 *                     priority, OSADDRESS stack_size);
 *
 * @brief shell_add_cmd_attrs() registers a shell command.
 *
 * @long shell_add_cmd_attrs() registers a resident shell command with the shell
 * daemon overriding any existing command with the same name. The process type,
 * priority and stack size for instances of this shell command can also be
 * specified.
 *
 * @input cmd_name      Name of command.
 * @input usage         Usage string for the command, used by help.
 * @input description   Description string for the command, used by help.
 * @input hook          Entrypoint of the command.
 * @input proc_type     Process type for the command (@link "" "PROCESS_TYPE").
 *                      Use @link "" "DEFAULT_PROC_TYPE" for default (inherited
 *                      from parent shell).
 * @input priority      Process priority for the command, use
 *                      @link "" "DEFAULT_PRIORITY" for default (inherited
 *                      from parent shell).
 * @input stack_size    Stack size for the command, use
 *                      @link "" "DEFAULT_STACK_SIZE" for default (hard-coded
 *                      in shell).
 *
 * @return void
 *
 * @restriction Also available from START_OSE_HOOK2 context.
 *
 * @seealso CmdType
 * @seealso shell_add_cmd
 * @seealso shell_remove_cmd
 * @seealso default_restrictions_calling_functions
 *
 * @example
 * static int cmd_vip(int argc, char *argv[])
 * {
 *    printf("A message from a very important process\n");
 *    return 0;
 * }
 *
 * void init_vip_cmd(void)
 * {
 *    shell_add_cmd_attrs("vip", "vip", "run a very important process",
 *       cmd_vip, OS_PRI_PROC, 1, 1000);
 * }
 */
void shell_add_cmd_attrs(const char *cmd_name, const char *usage,
			 const char *description, CmdType hook,
			 enum PROCESS_TYPE proc_type, OSPRIORITY
			 priority, OSADDRESS stack_size);

/**
 *
 * @function shell_remove_cmd
 *
 * @include        "shell.h"
 *
 * @declaration void
 * shell_remove_cmd(const char *cmd_name, CmdType hook);
 *
 * @brief shell_remove_cmd() removes a shell command.
 *
 * @long shell_remove_cmd() unregisters a resident shell command with the shell
 *       daemon.
 *
 * @input cmd_name       Name of command.
 * @input hook           Entrypoint of the command.
 *
 * @return void
 *
 * @seealso CmdType
 * @seealso shell_add_cmd
 * @seealso default_restrictions_calling_functions
 *
 * @example
 * shell_remove_cmd("hello", cmd_hello);
 */
void shell_remove_cmd(const char *cmd_name, CmdType hook);

/**
 *
 * @function shell_add_user
 *
 * @include        "shell.h"
 *
 * @declaration void
 * shell_add_user(PROCESS pid, const char *username, const char *terminal);
 *
 * @brief shell_add_user() registers a user that has logged in.
 *
 * @long shell_add_user() registers a user (username, terminal and PID of main
 *       process) that has logged in.
 *
 * @input pid        PID of user's main (shell) process.
 * @input username   Name of user.
 * @input terminal   Name of user's terminal.
 *
 * @return void
 *
 * @seealso default_restrictions_calling_functions
 */
void shell_add_user(PROCESS pid, const char *username, const char *terminal);

/**
 *
 * @function shell_run_cmd
 *
 * @include        "shell.h"
 *
 * @declaration U32
 * shell_run_cmd(const char *cmd, const OSADDRESS *taglist)
 *
 * @brief shell_run_cmd() executes the specified shell command.
 *
 * @long shell_run_cmd() executes the specified shell command. The environment
 *       of the command can be modified by specifying extra tags to the
 *       shell_run_cmd() call.
 *
 * @long Only registered commands can be run with shell_run_cmd(), not
 *       built-in shell commands.
 *
 * @input cmd       Command line to execute.
 * @input taglist   Options to modify the environment specified by
 *                  @link "" "ShellRunCmdTag". The tag list array is always
 *                  ended with the RUNCMD_TAGEND tag.
 *
 * @return Status/error code specified by @link "" "ShellRunCmdStatus".
 *
 * @seealso shell_add_cmd
 * @seealso default_restrictions_calling_functions
 *
 * @example
 * //Example 1: Specify a tag list consisting of several tags to demonstrate 
 * //how each tag and its tag arguments are used.
 * OSADDRESS taglist[] = { RUNCMD_EFS_CLONE,
 *                         RUNCMD_CLONE_ENV,
 *                         RUNCMD_NOPARENT,
 *                         RUNCMD_ADDUSER, NULL, NULL,  // *user *terminal
 *                         RUNCMD_PIDPTR, NULL,         // *PID
 *                         RUNCMD_INTERRUPTABLE,
 *                         RUNCMD_RETURN_STATUS, NULL,  // *status
 *                         RUNCMD_RETURN_ERRNO, NULL,   // *errno
 *                         RUNCMD_FDARRAY, NULL,        // *fd_array
 *                         RUNCMD_ENVLIST, NULL,        // *envlist
 *                         RUNCMD_TAGEND
 *                       };
 * int cmd_status;
 * int cmd_errno;
 * PROCESS cmd_parent = current_process(), cmd_pid;
 * char * cmd_envlist[] = {"variable1=value1", "variable2=value2", NULL};
 * char * cmd_username = "user name";
 * char * cmd_terminal = "terminal name";
 * // In this example, stdin, out and err was previously opened.
 * int cmd_fdarray[] = {3, 0, 1, 2};
 * //Initialize the tag arguments.
 *
 * //RUNCMD_ADDUSER arguments.
 * taglist[4] = (OSADDRESS)cmd_username;
 * taglist[5] = (OSADDRESS)cmd_terminal;
 * //RUNCMD_PIDPTR argument.
 * taglist[7] = (OSADDRESS)&cmd_pid;
 * //RUNCMD_RETURN_STATUS argument.
 * taglist[10] = (OSADDRESS)&cmd_status;
 * //RUNCMD_RETURN_ERRNO argument.
 * taglist[12] = (OSADDRESS)&cmd_errno;
 * //RUNCMD_FDARRAY argument.
 * taglist[14] = (OSADDRESS)cmd_fdarray;
 * //RUNCMD_ENVLIST argument.
 * taglist[16] = (OSADDRESS)cmd_envlist;
 *
 * U32 status = shell_run_cmd("ps ose*", taglist);
 * if (status != SHELL_SUCCESS)
 * {
 *    eprintf("Failed to execute command [PID:0x%x], status %lu., errno %lu.\n",
 *       cmd_pid, cmd_status, cmd_errno);
 * }
 * // Else command successfully executed. //
 * 
 * // Example 2: An example of how to configure stdin, stdout and stderr 
 * // redirection with shell_run_cmd.
 * // RUNCMD_CLONE_ENV is used to clone the environment variables.
 * // RUNCMD_CLONE_EFS is not used because the cloned descriptors
 * // would get overwritten by the new fd array. 
 * #include "ose.h"
 * #include "efs.h"
 * #include "stdio.h"
 * #include "string.h"
 * #include "shell.h"
 * #include "unistd.h"
 * #include "sys/stat.h"
 *
 * OSADDRESS tags[15];
 * int fd_array[4];
 * int i = 0, res = 0;
 * fd_array[0] = 3;
 * fd_array[1] = open("stdin_file.txt",O_CREAT | O_RDONLY | O_TRUNC, 
 *                    S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
 * fd_array[2] = open("stdout_file.txt",O_CREAT | O_WRONLY | O_TRUNC, 
 *                    S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
 * fd_array[3] = open("stderr_file.txt",O_CREAT | O_WRONLY | O_TRUNC, 
 *                    S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
 * tags[i++] = RUNCMD_CLONE_ENV;
 * tags[i++] = RUNCMD_FDARRAY;
 * tags[i++] = (OSADDRESS) fd_array;
 * tags[i] = RUNCMD_TAGEND;
 *
 * res = shell_run_cmd("rld", tags);
 * close(fd_array[1]);
 * close(fd_array[2]);
 * close(fd_array[3]);
 * 
 */
U32 shell_run_cmd(const char *cmd, const OSADDRESS *taglist);

#ifdef __cplusplus
}
#endif

#endif /* #ifndef _SHELL_H */
