/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2011 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */
/**
 * @public
 * @toc Standard_C:
 * @file stdlib.h
 *
 * @brief 'ISO/IEC 9899, Programming Languages - C' interface stdlib.h
 * @seealso malloc.h
 */

#ifndef _STDLIB_H
#define _STDLIB_H
#define OSE_STDLIB_H

/* We need to know if this is OSE soft environment or not. */
#include "osarch.h"

/* Include compiler's stddef.h header. */
#include <stddef.h>

#ifdef __cplusplus
#if !defined (NO_NAME_STD) && defined (__ARMCC_VERSION)
namespace std {
#endif /* NO_NAME_STD */
extern "C" {
#endif /* __cplusplus */

#ifdef __GNUC__
/* include gcc extensions */
#include "_ansi.h"
#endif

/* These interfaces should be visible from stdlib.h. */
#include "malloc.h"

#ifdef softOSE
#define ZZMK_NAME_MANGLING
#include "sfk/sfk_stdlib.h"
#endif

/**
 * @macro EXIT_FAILURE
 * @include "stdlib.h"
 *
 * @brief This macro can be used with the exit function to indicate
 *        unsuccessful process completion in a general sense.
 * @long  The value of this macro is 1.
 */
#define EXIT_FAILURE 1

/**
 * @macro EXIT_SUCCESS
 * @include "stdlib.h"
 *
 * @brief This macro can be used with the exit function to indicate
 *        successful process completion.
 * @long  The value of this macro is 0.
 */
#define EXIT_SUCCESS 0

/**
 * @macro RAND_MAX
 * @include "stdlib.h"
 *
 * @brief The maximum return value for random number generators.
 * @long The value of this macro is an integer constant representing
 *       the largest value the rand function can return.
 */
#define RAND_MAX 0x7fffffff

extern int __mb_cur_max(void);

/**
 * @macro MB_CUR_MAX
 * @include "stdlib.h"
 *
 * @brief The maximum number of bytes in a multibyte character.
 * @long The value of MB_CUR_MAX is the maximum number of bytes in a
 *       multibyte character for the current locale.
 */
#define MB_CUR_MAX ( __mb_cur_max() )

/* type of the value returned by the div function. */
/**
 * @type div_t
 * @brief Structure for 'int' division.
 * @long Structure used for division, where "quot" is the integer quotient
 *       and "rem" is the remainder.
 * @field quot Quotient.
 * @field rem Remainder.
 * @seealso ldiv_t
 */
typedef struct div_t { int quot, rem; } div_t;

/* type of the value returned by the ldiv function. */
/**
 * @type ldiv_t
 * @brief Structure for 'long' division.
 * @long Structure used for long division, where "quot" is the integer quotient
 *       and "rem" is the remainder.
 * @field quot Quotient.
 * @field rem Remainder.
 * @seealso div_t
 */
typedef struct ldiv_t { long int quot, rem; } ldiv_t;

/**
 * @function atof
 * @include "stdlib.h"
 *
 * @brief Converts a string to a double.
 *
 * @long atof() converts the initial portion of the string pointed to by "str"
 *       to double. The behavior is the same as strtod(str, (char **)NULL),
 *       except that atof() does not detect errors.
 *
 * @long Standards: SVID 3, POSIX, BSD 4.3, ISO 9899
 *
 * @input str The string to be converted.
 *
 * @return The converted value.
 *
 * @seealso atoi
 * @seealso strtod
 * @seealso strtol
 * @seealso strtoul
 * @seealso default_restrictions_calling_functions
 */
double atof(const char *str);

/**
 * @function atoi
 * @include "stdlib.h"
 *
 * @brief Converts a string to an integer.
 *
 * @long atoi() converts the initial portion of the string pointed to by "str"
 *       to an int. The behavior is the same as
 *       strtol(str, (char **)NULL, 10), except that atoi() does not
 *       detect errors.
 *
 * @long Standards: SVID 3, POSIX.1, BSD 4.3, ISO/IEC 9899.
 *       ISO/IEC 9899:1990 (C89) and POSIX.1 (1996 edition)
 *
 * @input str The string to be converted.
 *
 * @return The converted value.
 *
 * @seealso atof
 * @seealso strtod
 * @seealso strtol
 * @seealso strtoul
 * @seealso default_restrictions_calling_functions
 */
int atoi(const char *str);

/**
 * @function atol
 * @include "stdlib.h"
 *
 * @brief Converts a string to a long.
 *
 * @long atoi() converts the initial portion of the string pointed to by "str"
 *       to a long integer. The behavior is the same as
 *       strtol(str, (char **)NULL, 10), except that atol() does not
 *       detect errors.
 * @long atol() behave the same as atoi(), except that the result is a long.
 *
 * @long Standards: SVID 3, POSIX.1, BSD 4.3, ISO/IEC 9899.
 *       ISO/IEC 9899:1990 (C89) and POSIX.1 (1996  edition)
 *
 * @input str The string to be converted.
 *
 * @return The converted value.
 *
 * @seealso atof
 * @seealso strtod
 * @seealso strtol
 * @seealso strtoul
 * @seealso default_restrictions_calling_functions
 */
long int atol(const char *str);

/**
 * @function strtod
 * @include "stdlib.h"
 *
 * @brief Converts ASCII string to floating point number.
 *
 * @long strtod() converts the initial portion
 *       of the string pointed to by "str" to double..
 * @long The expected form of the (initial portion of the) string is optional
 *       leading white space, an optional plus
 *       ('+') or minus sign ('-') and then either (i) a decimal number, or
 *       (ii) a hexadecimal number, or (iii) an infinity, or (iv) a NAN
 *       (not-a-number).
 * @long A decimal number consists of a non-empty sequence of decimal
 *       digits possibly containing a radix character (decimal point,
 *       locale dependent, usually '.'), optionally followed by a decimal
 *       exponent. A decimal exponent consists of an 'E' or 'e', followed
 *       by an optional plus or minus sign, followed by a non-empty sequence
 *       of decimal digits, and indicates multiplication by a power of 10.
 * @long A hexadecimal number consists of a '0x' or '0X' followed by a
 *       non-empty sequence of hexadecimal digits possibly containing a radix
 *       character, optionally followed by a binary exponent. A binary exponent
 *       consists of a 'P' or 'p', followed by an optional plus or minus
 *       sign, followed by a non-empty sequence of decimal digits, and
 *       indicates multiplication by a power of 2. At least one of radix
 *       character and binary exponent must be present.
 * @long An infinity is either 'INF' or 'INFINITY', disregarding case.
 * @long A NAN is 'NAN' (disregarding case) optionally followed by '(', a
 *       sequence of characters, followed by ')'. The character string
 *       specifies in an implementation-dependent way the type of NAN.
 *
 * @long Standards: ANSI C.
 *
 * @input str The string to be converted.
 * @input endptr If "endptr" is not NULL, a pointer to the character after
 *        the last character used in the conversion will be stored in the
 *        location referenced by "endptr".
 *
 * @error ERANGE The resulting value was out of range.
 *
 * @return The converted value, if any.
 *         If no conversion is performed, zero is returned and the value of
 *         "str" (pointer to first character) is stored in the location
 *         referenced by "endptr", unless "endptr" is NULL.
 *         If the correct value would cause overflow, plus or minus HUGE_VAL
 *         (HUGE_VALF, HUGE_VALL) is returned (according to the sign of the
 *         value), and ERANGE is stored in errno. If the correct value would
 *         cause underflow, zero is returned and ERANGE is stored in errno.
 *
 * @seealso atof
 * @seealso atoi
 * @seealso atol
 * @seealso strtol
 * @seealso strtoul
 * @seealso strtoull
 * @seealso default_restrictions_calling_functions
 */
double strtod(const char *str, char **endptr);

/**
 * @function strtol
 * @include "stdlib.h"
 *
 * @brief Converts a string to a long integer.
 *
 * @long strtol() converts the initial part of the string in "str" to a long
 *       integer value according to the given  base, which must be
 *       between 2 and 36 inclusive, or be the special value 0.
 * @long The string must begin with an arbitrary amount of white space,
 *       followed by a single optional '+' or '-'
 *       sign. If "base" is zero or 16, the string may then include a '0x'
 *       prefix, and the number will be read in base 16; otherwise, a zero base
 *       is taken as 10 (decimal) unless the next character is '0', in which
 *       case it is taken as 8 (octal).
 * @long The remainder of the string is converted to a long int value in the
 *       obvious  manner, stopping at the first character which is not a valid
 *       digit in the given base. (In bases above 10, the letter 'A' in either
 *       upper or lower case represents 10, 'B' represents 11, and so forth,
 *       with 'Z' representing 35.)
 * @long If endptr is not NULL, strtol() stores the address of the first
 *       invalid character in "*endptr". If there were no digits at all,
 *       strtol() stores the original value of "str" in "*endptr"
 *       (and returns 0). In particular, if "*str" is not '\0' but
 *       "**endptr" is '\0' on return, the entire string is valid.
 *
 * @long Standards: SVID 3, BSD 4.3, ISO 9899 (C99) and POSIX
 *
 * @input str The string to be converted.
 * @input endptr If "endptr" is not NULL, a pointer to the character after
 *        the last character used in the conversion will be stored in the
 *        location referenced by "endptr".
 * @input base The base to convert to.
 *
 * @error ERANGE The resulting value was out of range.
 * @error EINVAL (not in C99) The given base contains an unsupported value.
 *        The implementation may also set errno to EINVAL in case no
 *        conversion was performed (no digits seen, and 0 returned).
 * @return strtol() returns the result of the conversion, unless the
 *         value would underflow or overflow. If an underflow occurs, strtol()
 *         returns LONG_MIN. If an overflow occurs, strtol() returns LONG_MAX.
 *         In both cases, errno is set to ERANGE.
 *
 * @seealso atof
 * @seealso atoi
 * @seealso atol
 * @seealso strtod
 * @seealso strtoul
 * @seealso default_restrictions_calling_functions
 */
long int strtol(const char *str, char **endptr, int base);

/**
 * @function strtoul
 * @include "stdlib.h"
 *
 * @brief Converts a string to an unsigned long integer.
 *
 * @long strtoul() converts the initial part of the string in "str"
 *       to an unsigned long integer value according to the given base,
 *       which must be between 2 and 36 inclusive, or be the special
 *       value 0.
 * @long The string must begin with an arbitrary amount of white space,
 *       followed by a single optional '+' or '-'
 *       sign. If "base" is zero or 16, the string may then include a '0x'
 *       prefix, and the number will be read in base 16; otherwise, a
 *       zero base is taken as 10 (decimal) unless the next character is '0',
 *       in which case it is taken as 8 (octal).
 * @long The  remainder of the string is converted to an unsigned long
 *       int value in the obvious manner, stopping at the first character
 *       which is not a valid digit in the given base. (In bases above 10,
 *       the letter 'A' in either upper or lower case represents 10,
 *       'B' represents 11, and so forth, with 'Z' representing 35.)
 * @long If "endptr" is not NULL, strtoul() stores the address of the first
 *       invalid character in "*endptr". If there were no digits at all,
 *       strtoul() stores the original value of "str" in "*endptr"
 *       (and returns 0). In particular, if "*str" is not '\0' but "**endptr"
 *       is '\0' on return, the entire string is valid.
 *
 * @long Standards: SVID 3, BSD 4.3, ISO 9899
 *
 * @input str The string to be converted.
 * @input endptr If "endptr" is not NULL, a pointer to the character after
 *        the last character used in the conversion will be stored in the
 *        location referenced by "endptr".
 * @input base The base to convert to.
 *
 * @error ERANGE The resulting value was out of range.
 * @error EINVAL (not in C99) The given base contains an unsupported value.
 *        The implementation  may also set errno to EINVAL in case no
 *        conversion was performed (no digits seen, and 0 returned).
 *
 * @return strtoul() returns either the result of the conversion or,
 *         if there was a leading minus sign, the negation of the result of
 *         the conversion, unless the original (non-negated) value would
 *         overflow; in the latter case, strtoul() returns ULONG_MAX and
 *         sets the global variable errno to ERANGE. Precisely the same
 *         holds for strtoull() (with ULLONG_MAX instead of ULONG_MAX).
 *
 * @seealso atof
 * @seealso atoi
 * @seealso atol
 * @seealso strtod
 * @seealso strtol
 * @seealso default_restrictions_calling_functions
 */
unsigned long int strtoul(const char *str, char **endptr, int base);
/**
 * @function strtoll
 * @include "stdlib.h"
 *
 * @brief Converts a string to a long long integer.
 *
 * @long strtoll() converts the initial part of the string in "str" to
 *       a long long integer value according to the given  base, which must be
 *       between 2 and 36 inclusive, or be the special value 0.
 * @long The string must begin with an arbitrary amount of white space,
 *       followed by a single optional '+' or '-'
 *       sign. If "base" is zero or 16, the string may then include a '0x'
 *       prefix, and the number will be read in base 16; otherwise, a zero base
 *       is taken as 10 (decimal) unless the next character is '0', in which
 *       case it is taken as 8 (octal).
 * @long The remainder of the string is converted to a long long int value in the
 *       obvious  manner, stopping at the first character which is not a valid
 *       digit in the given base. (In bases above 10, the letter 'A' in either
 *       upper or lower case represents 10, 'B' represents 11, and so forth,
 *       with 'Z' representing 35.)
 * @long If endptr is not NULL, strtol() stores the address of the first
 *       invalid character in "*endptr". If there were no digits at all,
 *       strtol() stores the original value of "str" in "*endptr"
 *       (and returns 0). In particular, if "*str" is not '\0' but
 *       "**endptr" is '\0' on return, the entire string is valid.
 *
 * @long Standards: SVID 3, BSD 4.3, ISO 9899 (C99) and POSIX
 *
 * @input str The string to be converted.
 * @input endptr If "endptr" is not NULL, a pointer to the character after
 *        the last character used in the conversion will be stored in the
 *        location referenced by "endptr".
 * @input base The base to convert to.
 *
 * @error ERANGE The resulting value was out of range.
 * @error EINVAL (not in C99) The given base contains an unsupported value.
 *        The implementation may also set errno to EINVAL in case no
 *        conversion was performed (no digits seen, and 0 returned).
 * @return strtol() returns the result of the conversion, unless the
 *         value would underflow or overflow. If an underflow occurs, strtol()
 *         returns LLONG_MIN. If an overflow occurs, strtoll() returns LLONG_MAX.
 *         In both cases, errno is set to ERANGE.
 *
 * @seealso atof
 * @seealso atoi
 * @seealso atol
 * @seealso strtod
 * @seealso strtoul
 * @seealso strtoull
 * @seealso default_restrictions_calling_functions
 */
long long int strtoll(const char *str, char **endptr, int base);

/**
 * @function strtoull
 * @include "stdlib.h"
 *
 * @brief Converts a string to an unsigned long long integer.
 *
 * @long strtoul() converts the initial part of the string in "str"
 *       to an unsigned long long integer value according to the given base,
 *       which must be between 2 and 36 inclusive, or be the special
 *       value 0.
 * @long The string must begin with an arbitrary amount of white space,
 *       followed by a single optional '+' or '-'
 *       sign. If "base" is zero or 16, the string may then include a '0x'
 *       prefix, and the number will be read in base 16; otherwise, a
 *       zero base is taken as 10 (decimal) unless the next character is '0',
 *       in which case it is taken as 8 (octal).
 * @long The  remainder of the string is converted to an unsigned long long
 *       int value in the obvious manner, stopping at the first character
 *       which is not a valid digit in the given base. (In bases above 10,
 *       the letter 'A' in either upper or lower case represents 10,
 *       'B' represents 11, and so forth, with 'Z' representing 35.)
 * @long If "endptr" is not NULL, strtoul() stores the address of the first
 *       invalid character in "*endptr". If there were no digits at all,
 *       strtoul() stores the original value of "str" in "*endptr"
 *       (and returns 0). In particular, if "*str" is not '\0' but "**endptr"
 *       is '\0' on return, the entire string is valid.
 *
 * @long Standards: SVID 3, BSD 4.3, ISO 9899
 *
 * @input str The string to be converted.
 * @input endptr If "endptr" is not NULL, a pointer to the character after
 *        the last character used in the conversion will be stored in the
 *        location referenced by "endptr".
 * @input base The base to convert to.
 *
 * @error ERANGE The resulting value was out of range.
 * @error EINVAL (not in C99) The given base contains an unsupported value.
 *        The implementation  may also set errno to EINVAL in case no
 *        conversion was performed (no digits seen, and 0 returned).
 *
 * @return strtoul() returns either the result of the conversion or,
 *         if there was a leading minus sign, the negation of the result of
 *         the conversion, unless the original (non-negated) value would
 *         overflow; in the latter case, strtoul() returns ULLONG_MAX and
 *         sets the global variable errno to ERANGE. Precisely the same
 *         holds for strtoull() (with ULLONG_MAX instead of ULLONG_MAX).
 *
 * @seealso atof
 * @seealso atoi
 * @seealso atol
 * @seealso strtod
 * @seealso strtol
 * @seealso strtoll
 * @seealso default_restrictions_calling_functions
 */
unsigned long long int strtoull(const char *str, char **endptr, int base);

/**
 * @function rand
 * @include "stdlib.h"
 *
 * @brief Random number generator.
 *
 * @long rand() returns a pseudo-random integer between 0 and
 *       RAND_MAX. srand() sets its argument as the seed for a
 *       new sequence of pseudo-random integers to be returned by rand().
 *       These sequences are repeatable by calling srand() with the same
 *       seed value.The rand() function need not be reentrant. A function
 *       that is not required to be reentrant is not required to be
 *       thread-safe.
 * @long If no seed value is provided, rand() is automatically
 *       seeded with a value of 1.
 *
 * @long Standards: SVID 3, BSD 4.3, ISO 9899
 *
 * @return rand() returns a value between 0 and RAND_MAX.
 *
 * @seealso default_restrictions_calling_functions
 */
int rand(void);

/**
 * @function srand
 * @include "stdlib.h"
 *
 * @brief Random number generator.
 *
 * @long rand() returns a pseudo-random integer between 0 and
 *       RAND_MAX. srand() sets its argument as the seed for a
 *       new sequence of pseudo-random integers to be returned by rand().
 *       These sequences are repeatable by calling srand() with the same
 *       seed value.
 * @long If no seed value is provided, rand() is automatically
 *       seeded with a value of 1.
 *
 * @long Standards: SVID 3, BSD 4.3, ISO 9899
 *
 * @input seed The seed for the random number generator.
 *
 * @return srand() returns no value.
 *
 * @seealso default_restrictions_calling_functions
 */
void srand(unsigned int seed);

/**
 * @function rand_r
 * @include "stdlib.h"
 *
 * @brief Random number generator.
 *
 * @long The rand_r() function shall compute a sequence of pseudo-random
 *       integers in the range [0, {RAND_MAX}].
 * @long If rand_r() is called with the same initial value for the
 *       object pointed to by seed and that object is not modified between
 *       successive returns and calls to rand_r(), the same sequence shall
 *       be generated.
 *
 * @long Standards: SVID 3, BSD 4.3, ISO 9899
 *
 * @input seed The seed for the random number generator.
 *
 * @return A pseudo-random integer.
 *
 * @seealso default_restrictions_calling_functions
 */
int rand_r(unsigned *seed);

/* Abort is implemented by OSE on target architectures. */

/**
 * @function abort
 * @include "stdlib.h"
 *
 * @brief Causes abnormal process termination
 *
 * @long abort() causes abnormal process termination.
 *       All open streams are closed and flushed.
 *
 * @long abort() enters the error handler via an error2() call
 *       with error code OSE_EFS_EABORT_CALLED.
 *
 * @long Standards: SVID 3, POSIX, BSD 4.3, ISO 9899 (C99)
 *
 * @return Nothing.
 *
 * @seealso exit
 * @seealso default_restrictions_calling_functions
 */
void abort(void);

/**
 * @function atexit
 * @include "stdlib.h"
 *
 * @brief Register a function to be called at normal process termination.
 *
 * @long atexit() registers the given function to be called at normal
 *       process termination, whether via @link "" "exit"() or
 *       returning from main(). Functions so registered are called in
 *       the reverse order of their registration; no arguments are passed.
 *
 * @long The special function @link "unistd.h" "_exit"() avoids
 *       calling functions, registered with atexit.
 *
 * @long Standards: SVID 3, BSD 4.3, ISO 9899
 *
 * @input func Specifies a pointer to the function to be called.
 *
 * @return A 0 if successful, otherwise a non-zero value; errno is not set.
 *
 * @seealso exit
 * @seealso _exit
 * @seealso default_restrictions_calling_functions
 *
 * @example
 * // To add the function myExitFunc():
 *     int OK;
 *     OK = atexit(myExitFunc);
 *     if (OK == 0)
 *     {
 *       // It is added.
 *     }
 */
int atexit(void (* func)(void));

/**
 * @function exit
 *
 * @include "stdlib.h"
 *
 * @brief Terminates the process.
 *
 * @long exit() performs the following actions:
 * @long 1. In C++, static objects are destructed.
 * @long 2. Calls all functions registered by @link "" "atexit"(),
 *          in reverse order of registration.
 * @long 3. Flushes any non-written buffers of all open output streams.
 * @long 4. Closes all open files.
 *          This can remove temporary files, created by
 *          @link "stdio.h" "tmpfile", if supported by the volume.
 * @long 5. Sends a ChildDone signal to parent, if it has a
 *          parent process. See ose_spi/efs.sig for a detailed description
 *          of ChildDone. A parent can use @link "efs.h" "efs_wait_child"()
 *          to wait for a this signal (including status).
 * @long 6. Terminates the calling process.
 * @long If running a program in shared execution model, the entire program
 *       is terminated, otherwise the process is terminated.
 *
 * @long Standards: ISO/IEC 9899:1990 (E), Programming Languages - C.
 *
 * @input status Status to return (if possible) to the host environment.
 *
 * @return The exit function does not return.
 *
 * @seealso atexit
 * @seealso _exit
 * @seealso tmpfile
 * @seealso fflush
 * @seealso fclose
 * @seealso default_restrictions_calling_functions
 */
void exit(int status);


/**
 * @function getenv
 * @include "stdlib.h"
 *
 * @brief Get an environment variable.
 *
 * @long getenv() goes through the environment variable list looking for
 *       a string that matches "name", where the strings are on the form
 *       "string=value".
 *
 * @long getenv causes a hidden free and malloc sequence in the process
 *       context. Once the first getenv for a process is performed,
 *       one heap buffer will always be allocated for the process for
 *       handling the getenv return value. A hidden free/malloc is
 *       done only if a later value needs a larger memory.
 *
 * @long Standards:
 *       ISO/IEC 9899, SVID 3, POSIX, BSD 4.3
 *
 * @input name The desired environment variable.
 *
 * @return A pointer to the value string, or NULL if there is no match.
 *
 * @seealso default_restrictions_calling_functions
 */
char *    getenv(const char * name);

/**
 * @function putenv
 * @include "stdlib.h"
 *
 * @brief Change or add an environment variable.
 *
 * @long putenv() adds or changes the value of an environment variable.
 *       The argument string needs to be on the form "name=value".
 *       If "name" does not already exist in the environment, the string
 *       is added. If "name" does exist, the value of "name" is changed
 *       to value.
 *
 * @long Standards: SVID 3, POSIX, BSD 4.3
 *
 * @restriction   
 *       The environment in OSE is a kernel resource, which cannot be
 *       manipulated without the use of system calls. The Posix standard
 *       stipulates that the string pointed to by "string" becomes part
 *       of the environment, and that manipulating the string after this
 *       call should affect the environment. This is not fulfilled; the
 *       memory occupied by string does not become part of the environment
 *       and manipulating the string after the call will not modify the
 *       environment.
 *
 * @input string The desired environment variable name and value.
 *
 * @error ENOMEM Insufficient space to allocate new environment.
 * @return A 0 on success, or -1 if an error occurs.
 *
 * @seealso default_restrictions_calling_functions
 */
int putenv(const char *string);

/**
 * @function system
 * @include "stdlib.h"
 *
 * @brief Do not use, it is not implemented.
 *
 * @long Do not use, it is not implemented.
 *
 * @long Standards: SVID 3, POSIX, BSD 4.3
 *
 * @input cmd_line -
 *
 * @return -
 */
int system(const char *cmd_line); /* Not implemented by OSE */


/**
 * @function bsearch
 * @include "stdlib.h"
 *
 * @brief bsearch() performs a binary search of a sorted array.
 *
 * @long The bsearch() function goes through an array of nmemb objects
 *       looking for a member that matches the object pointed to by "key".
 *       "base" points to the first nmemb object in the array, and the size
 *       of each member of the array is specified by "size".
 *
 * @long The contents of the array should be in ascending sorted order according
 *       to the comparison function referenced by "compar", which is called
 *       with two arguments that point to the objects being compared.
 *
 * @long The comparison function must return an integer less than,
 *       equal to, or greater than zero if the first argument (key) is
 *       considered to be less than, equal to, or greater than the second
 *       argument (an array member).
 *
 * @long Standards:
 *       SVID 3, BSD 4.3, ISO 9899 (C99)
 *
 * @input key A pointer to the object to search for.
 * @input base A pointer to the first object in the nmemb array.
 * @input nmemb the size of the memb array.
 * @input size The size of each member in the nmemb array.
 * @input compar The comparison function.
 *
 * @return Returns a pointer to an element of the array that matches key.
 *        If more than one matching element is available, the result
 *        may point to any of them. Returns NULL if none is found.
 * @seealso qsort
 * @seealso default_restrictions_calling_functions
 */
void *    bsearch(const void *key,
                  const void *base,
                  size_t nmemb,
                  size_t size,
                  int (*compar)(const void *, const void *));


/**
 * @function qsort
 * @include "stdlib.h"
 *
 * @brief qsort() sorts an array.
 *
 * @long qsort() sorts an array with nmemb elements of size "size".
 *       "base" points to the start of the array.
 *
 * @long The array is sorted in ascending order according to the comparison
 *       function "compar", which is called with two arguments that
 *       point to the objects being compared.
 *
 * @long The comparison function must return an integer less than,
 *       equal to, or greater than zero if the first argument is
 *       considered to be less than, equal to, or greater than the second.
 *       If two members compare as equal, their order in the sorted
 *       array is undefined.
 *
 * @long Standards:
 *        SVID 3, POSIX, BSD 4.3, ISO 9899
 *
 * @input base A pointer to the first object in the nmemb array.
 * @input nmemb the size of the memb array.
 * @input size The size of each member in the nmemb array.
 * @input compar The comparison function.
 *
 * @seealso default_restrictions_calling_functions
 */
void      qsort(void *base,
                size_t nmemb,
                size_t size,
                int (*compar)(const void *, const void *));

/**
 * @function abs
 * @include "stdlib.h"
 *
 * @brief abs() computes the absolute value of an integer.
 *
 * @long abs() computes the absolute value of the integer argument "j".
 * @long Trying to take the absolute value of the most negative integer is not
 *       defined.
 *
 * @long Standards:
 *       SVID 3, POSIX.1, BSD 4.3, ISO/IEC 9899 (C99).
 *       POSIX.1 (1996 edition), ISO/IEC 9899:1990 (C89)
 *
 * @input j The integer to compute the absolute value for.
 *
 * @return The absolute value of the integer argument, of the appropriate
 *         integer type for the function.
 *
 * @seealso default_restrictions_calling_functions
 */
int abs(int j);

/**
 * @function div
 * @include "stdlib.h"
 *
 * @brief div() computes the quotient and remainder of integer division.
 *
 * @long div() computes the value numer/denom and returns the quotient
 *       and remainder in a structure named "div_t" that contains two
 *       integer members named "quot" and "rem".
 *
 * @long Standards:
 *       SVID 3, BSD 4.3, ISO 9899
 *
 * @input numer The integer to divide (numerator).
 * @input denom The integer to divide by (denominator).
 *
 * @return The "div_t" structure.
 * @seealso div_t
 * @seealso ldiv
 * @seealso default_restrictions_calling_functions
 */
div_t div(int numer, int denom);

/**
 * @function labs
 * @include "stdlib.h"
 *
 * @brief labs() computes the absolute value of an integer.
 *
 * @long labs() computes the absolute value of the integer argument "j".
 * @long Trying to take the absolute value of the most negative integer is not
 *       defined.
 *
 * @long Standards:
 *       SVID 3, POSIX.1, BSD 4.3, ISO/IEC 9899 (C99).
 *       POSIX.1 (1996 edition), ISO/IEC 9899:1990 (C89)
 *
 * @input j The integer to compute the absolute value for.
 *
 * @return The absolute value of the integer argument, of the appropriate
 *         integer type for the function.
 *
 * @seealso default_restrictions_calling_functions
 */
long int labs(long int j);

/**
 * @function ldiv
 * @include "stdlib.h"
 *
 * @brief ldiv() computes the quotient and remainder of long integer division.
 *
 * @long ldiv() divides numer by denom and returns the quotient and
 *       remainder in the structure named ldiv_t that contains two
 *       long integer members named "quot" and "rem".
 *       The quotient is rounded towards zero.
 *
 * @long Standards:
 *       SVID 3, POSIX.1, BSD 4.3, ISO/IEC 9899 (C99).
 *       POSIX.1 (1996 edition), ISO/IEC 9899:1990 (C89)
 *
 * @input numer The integer to divide (numerator).
 * @input denom The integer to divide by (denominator).
 *
 * @return The "ldiv_t" structure.
 * @seealso ldiv_t
 * @seealso div
 * @seealso default_restrictions_calling_functions
 */
ldiv_t ldiv(long int numer, long int denom);


/**
 * @function realpath
 * @include "stdlib.h"
 *
 * @brief Convert a pathname to an absolute pathname that does not involve
 *        '.', '..' or symbolic links.
 *
 * @long The pathname pointed to by file_name is converted to an absolute
 *       pathname that resolves to the same directory antry and does not
 *       involve '.', '..' or symbolic links.
 *
 * @long The resolved path is stored in the buffer provided in the
 *       resolved_path parameter, or in a buffer allocated by malloc if
 *       resolved_path is a null pointer.
 *
 * @long If a buffer is provided, it should be at least 256
 *       bytes. The call will fail if a buffer is provided and
 *       the resolved path exceeds this size.
 *
 * @long Note: The ability to accept a null pointer as resolved_path was
 *       introduced in OSE 5.4.2.
 *
 * @input file_name     Pointer to a file name (null terminated string)
 *                      for which the path is to be expanded.
 * @input resolved_path Pointer to the buffer where the result is to be
 *                      stored, or a null pointer which is a request
 *                      that the buffer shall be allocated by the call.
 *
 * @return Upon successfull completion, a pointer to a buffer containing
 *         the resolved name is returned. Otherwise a null pointer is
 *         returned and @link "" "errno" is updated to indicate the error.
 *
 * @seealso getcwd
 * @seealso default_restrictions_calling_functions
 */
char * realpath(const char *file_name, char *resolved_path);


/**
 * @function wrealpath
 * @include "stdlib.h"
 *
 * @brief Convert a pathname to an absolute pathname that does not involve
 *        '.', '..' or symbolic links.
 *
 * @long The pathname pointed to by file_name is converted to an absolute
 *       pathname that resolves to the same directory antry and does not
 *       involve '.', '..' or symbolic links.
 *
 * @long The resolved path is stored in the buffer provided in the
 *       resolved_path parameter, or in a buffer allocated by malloc if
 *       resolved_path is a null pointer.
 *
 * @long If a buffer is provided, it should be at least 256
 *       wchar_t elements. The call will fail if a buffer is
 *       provided and the resolved path exceeds this size.
 *
 * @input file_name     Pointer to a file name (null terminated string)
 *                      for which the path is to be expanded.
 * @input resolved_path Pointer to the buffer where the result is to be
 *                      stored, or a null pointer which is a request
 *                      that the buffer shall be allocated by the call.
 *
 * @return Upon successfull completion, a pointer to a buffer containing
 *         the resolved name is returned. Otherwise a null pointer is
 *         returned and @link "" "errno" is updated to indicate the error.
 *
 * @seealso getcwd
 * @seealso default_restrictions_calling_functions
 */
wchar_t * wrealpath(const wchar_t *file_name, wchar_t *resolved_path);


/*
 * Floating point Math Functions.
 */

/**
 * @function gcvt
 * @include "stdlib.h"
 *
 * @brief Converts a double to a string.
 *
 * @long gcvt() converts "number" to a minimal length NULL terminated
 *       ASCII string and stores the result in "buf".
 *       It produces "ndigit" significant digits in either
 *       printf F format or E format.
 *
 * @input number The number to be converted.
 * @input ndigit The number of desired significant digits.
 * @input buf A pointer pointing to the result.
 *
 * @return The address of the string pointed to by "buf".
 *
 * @seealso ecvt
 * @seealso fcvt
 * @seealso sprintf
 * @seealso default_restrictions_calling_functions
 */
char * gcvt(double number, int ndigit, char *buf);

/**
 * @function gcvtf
 * @include "stdlib.h"
 *
 * @brief Converts a floating point number to a string.
 *
 * @long gcvtf() converts "number" to a minimal length NULL terminated
 *       ASCII string and stores the result in "buf".
 *       It produces "ndigit" significant digits in either
 *       printf F format or E format.
 *
 * @input number The number to be converted.
 * @input ndigit The number of desired significant digits.
 * @input buf A pointer pointing to the result.
 *
 * @return The address of the string pointed to by "buf".
 *
 * @seealso ecvt
 * @seealso fcvt
 * @seealso sprintf
 * @seealso default_restrictions_calling_functions
 */
char * gcvtf(float number, int ndigit, char *buf);

/**
 * @function fcvt
 * @include "stdlib.h"
 *
 * @brief Converts a double to a string.
 *
 * @long fcvt() converts "number" to a NULL terminated string of
 *       "ndigits" digits (where  ndigits specifies the number of
 *       digits after the decimal point.), and returns a pointer to the string.
 *       The high-order digit is non-zero, unless "number" is zero.
 *       The low order digit is rounded.
 * @long The string itself does not contain a decimal point, but
 *       the position of the decimal point relative to the start
 *       of the string is stored in "*decpt". A negative value for *decpt
 *       means that the decimal point is to the left of the start of the
 *       string. If the sign of number is negative, *sign is set to a
 *       non-zero value. If "number" is zero, it is unspecified
 *       whether "*decpt" is 0 or 1.
 * @long Standards: SysVR2, XPG2
 *
 * @input number The number to be converted.
 * @input ndigits The number of desired significant digits.
 * @input decpt The position of the decimal point relative to the start
 *              of the string.
 * @input sign If the sign of "number" is negative, "sign" is set to a
 *             non-zero value.
 *
 * @return A pointer to a static string containing the ASCII
 *         representation of number. The static string is overwritten
 *         by each call to fcvt().
 *
 * @seealso gcvt
 * @seealso sprintf
 * @seealso default_restrictions_calling_functions
 */
char * fcvt(double number, int ndigits, int *decpt, int *sign);

/**
 * @function fcvtf
 * @include "stdlib.h"
 *
 * @brief Converts a floating point number to a string.
 *
 * @long fcvtf() converts "number" to a NULL terminated string of
 *       "ndigits" digits (where  ndigits specifies the number of
 *       digits after the decimal point.), and returns a pointer to the string.
 *       The high-order digit is non-zero, unless "number" is zero.
 *       The low order digit is rounded.
 * @long The string itself does not contain a decimal point, but
 *       the position of the decimal point relative to the start
 *       of the string is stored in "*decpt". A negative value for *decpt
 *       means that the decimal point is to the left of the start of the
 *       string. If the sign of number is negative, *sign is set to a
 *       non-zero value. If "number" is zero, it is unspecified
 *       whether "*decpt" is 0 or 1.
 * @long Standards: SysVR2, XPG2
 *
 * @input number The number to be converted.
 * @input ndigits The number of desired significant digits.
 * @input decpt The position of the decimal point relative to the start
 *              of the string.
 * @input sign If the sign of "number" is negative, "sign" is set to a
 *             non-zero value.
 *
 * @return A pointer to a static string containing the ASCII
 *         representation of number. The static string is overwritten
 *         by each call to fcvtf().
 *
 * @seealso gcvt
 * @seealso gcvtf
 * @seealso sprintf
 * @seealso default_restrictions_calling_functions
 */
char * fcvtf(float number, int ndigits, int *decpt, int *sign);

/**
 * @function ecvt
 * @include "stdlib.h"
 *
 * @brief Converts a double to a string.
 *
 * @long ecvt() converts "number" to a NULL terminated string of
 *       "ndigits" digits (where ndigits is reduced to a
 *       system-specific limit determined by the precision of a double),
 *       and returns a pointer to the string.
 *       The high-order digit is non-zero, unless "number" is zero.
 *       The low order digit is rounded.
 * @long The string itself does not contain a decimal point, but
 *       the position of the decimal point relative to the start
 *       of the string is stored in "*decpt". A negative value for *decpt
 *       means that the decimal point is to the left of the start of the
 *       string. If the sign of number is negative, "*sign" is set to a
 *       non-zero value. If "number" is zero, it is unspecified
 *       whether "*decpt" is 0 or 1.
 * @long Standards: SysVR2, XPG2
 *
 * @input number The number to be converted.
 * @input ndigits The number of desired significant digits,
 *               reduced to a system-specific limit determined
 *               by the precision of a double.
 * @input decpt The position of the decimal point relative to the start
 *              of the string.
 * @input sign If the sign of "number" is negative, "sign" is set to a
 *             non-zero value.
 *
 * @return A pointer to a static string containing the ASCII
 *         representation of number. The static string is overwritten
 *         by each call to ecvt().
 *
 * @seealso gcvt
 * @seealso gcvtf
 * @seealso sprintf
 * @seealso default_restrictions_calling_functions
 */
char * ecvt(double number, int ndigits, int *decpt, int *sign);

/**
 * @function ecvtbuf
 * @include "stdlib.h"
 *
 * @brief Converts a double or a floating point number to a string.
 *
 * @long ecvtbuf() converts "number" to a NULL terminated string of
 *       "ndigits" digits (where ndigits is reduced to a
 *       system-specific limit determined by the precision of a double),
 *       and returns a pointer to the string.
 *       The high-order digit is non-zero, unless "number" is zero.
 *       The low order digit is rounded.
 * @long The string itself does not contain a decimal point, but
 *       the position of the decimal point relative to the start
 *       of the string is stored in "*decpt". A negative value for *decpt
 *       means that the decimal point is to the left of the start of the
 *       string. If the sign of number is negative, "*sign" is set to a
 *       non-zero value. If "number" is zero, it is unspecified
 *       whether "*decpt" is 0 or 1.
 * @long Standards: SysVR2, XPG2
 *
 * @input number The number to be converted.
 * @input ndigits The total number of characters to write
 *                (which is also the number of significant digits in the
 *                formatted string, since the function only write digits).
 * @input decpt The position of the decimal point relative to the start
 *              of the string.
 * @input sign If the sign of "number" is negative, "sign" is set to a
 *             non-zero value.
 * @input buf A pointer to the string containing a character
 *            representation of "number".
 *
 * @return A pointer to "buf", the string containing a character
 *         representation of "number".
 *
 * @seealso gcvt
 * @seealso gcvtf
 * @seealso sprintf
 * @seealso default_restrictions_calling_functions
 */
char * ecvtbuf(double number, int ndigits, int *decpt, int *sign, char *buf);

/**
 * @function fcvtbuf
 * @include "stdlib.h"
 *
 * @brief Converts a double or a floating point number to a string.
 *
 * @long fcvtbuf() converts "number" to a NULL terminated string where "dec"
 *       specifies the number of characters to write after the decimal point.
 *       All digits for the integer part of "number" are always included.
 * @long The string itself does not contain a decimal point, but
 *       the position of the decimal point relative to the start
 *       of the string is stored in "*decpt". A negative value for *decpt
 *       means that the decimal point is to the left of the start of the
 *       string. If the sign of number is negative, "*sign" is set to a
 *       non-zero value. If "number" is zero, it is unspecified
 *       whether "*decpt" is 0 or 1.
 * @long Standards: SysVR2, XPG2
 *
 * @input number The number to be converted.
 * @input dec The number of characters to write after the
 *            decimal point
 * @input decpt The position of the decimal point relative to the start
 *              of the string.
 * @input sign If the sign of "number" is negative, "sign" is set to a
 *             non-zero value.
 * @input buf A pointer to the string containing a character
 *            representation of "number".
 *
 * @return A pointer to "buf", the string containing a character
 *         representation of "number".
 *
 * @seealso gcvt
 * @seealso gcvtf
 * @seealso sprintf
 * @seealso default_restrictions_calling_functions
 */
char * fcvtbuf(double number, int dec, int *decpt, int *sign, char *buf);

/**
 * @function ecvtf
 * @include "stdlib.h"
 *
 * @brief Converts a floating point number to a string.
 *
 * @long ecvtf() converts "number" to a NULL terminated string of
 *       "ndigits" digits (where ndigits is reduced to a
 *       system-specific limit determined by the precision of a double),
 *       and returns a pointer to the string.
 *       The high-order digit is non-zero, unless "number" is zero.
 *       The low order digit is rounded.
 * @long The string itself does not contain a decimal point, but
 *       the position of the decimal point relative to the start
 *       of the string is stored in "*decpt". A negative value for *decpt
 *       means that the decimal point is to the left of the start of the
 *       string. If the sign of number is negative, *sign is set to a
 *       non-zero value. If "number" is zero, it is unspecified
 *       whether "*decpt" is 0 or 1.
 * @long Standards: SysVR2, XPG2
 *
 * @input number The number to be converted.
 * @input ndigits The number of desired significant digits.
 * @input decpt The position of the decimal point relative to the start
 *              of the string.
 * @input sign If the sign of "number" is negative, "sign" is set to a
 *             non-zero value.
 *
 * @return A pointer to a static string containing the ASCII
 *         representation of number. The static string is overwritten
 *         by each call to ecvtf().
 *
 * @seealso gcvt
 * @seealso gcvtf
 * @seealso sprintf
 * @seealso default_restrictions_calling_functions
 */
char * ecvtf(float number, int ndigits, int *decpt, int *sign);

char * dtoa(double, int, int, int *, int*, char**);

/*
 * Multibyte Character Functions.
 */

/**
 * @function mblen
 * @include "stdlib.h"
 *
 * @brief Determines the number of bytes in the next multibyte character.
 *
 * @long If "str" is not a NULL pointer, mblen() inspects at most "n"
 *       bytes of the multibyte string starting at "str", and extracts
 *       the next complete multibyte character.
 * @long mblen() uses a static anonymous shift state only known to the
 *       function itself. If the multibyte character is not the null wide
 *       character, it returns the number of bytes that were consumed
 *       from "str". If the multibyte character is the null wide character,
 *       it returns 0.
 * @long If the "n" bytes starting at "str" do not contain a complete
 *       multibyte character, mblen() returns -1.
 *       This can happen if the multibyte string contains redundant shift
 *       sequences, even if n >= MB_CUR_MAX.
 * @long If the multibyte string starting at "str" contains an invalid
 *       multibyte sequence before the next complete character, mblen()
 *       also returns -1.
 * @long If "str" is a NULL pointer, mblen() resets the shift
 *       state (only known to mblen()) to the initial state, and
 *       returns non-zero if the encoding has a non-trivial shift state,
 *       or zero if the encoding is stateless.
 * @long The behavior of mblen() depends on the LC_CTYPE category of the
 *       current locale. mbrlen() provides a better interface to the
 *       same functionality.
 *
 * @long Standards: ISO/ANSI C, UNIX98
 *
 * @input str A pointer to a multibyte string.
 * @input n The number of characters to inspect.
 *
 * @return The number of bytes parsed from the multibyte sequence starting
 *         at "str", if a non-null wide character was recognized.
 *         0 if a null wide character was recognized, and -1, if an invalid
 *         multibyte sequence was encountered or if it could not parse a
 *         complete multibyte character.
 *
 * @seealso default_restrictions_calling_functions
 */
int   mblen(const char * str, size_t n);

/**
 * @function mbtowc
 * @include "stdlib.h"
 *
 * @brief Converts a multibyte sequence to a wide character.
 *
 * @long If neither "str" or "pwc" is NULL mbtowc() inspects at most
 *       "n" bytes of the multibyte string starting at "str", extracts
 *       the next complete multibyte character, converts it to a
 *       wide character and stores it at "*pwc", and updates an
 *       internal shift state only known to mbtowc().
 * @long If "str" does not point to a '\0' byte, it returns the number
 *       of bytes that were consumed from "str", otherwise it returns 0.
 * @long If the "n" bytes starting at "str" do not contain a complete
 *       multibyte character, or if they contain an invalid multibyte
 *       sequence, mbtowc() returns -1. This can happen if the
 *       multibyte string contains redundant shift sequences,
 *       even if n >= MB_CUR_MAX.
 * @long A different case is when "str" is not NULL but "pwc" is NULL.
 *       In this case the mbtowc() function behaves as above, except
 *       that it does not store the converted wide character in memory.
 * @long A third case is when "str" is NULL. In this case, "pwc" and "n"
 *       are ignored. mbtowc() resets the shift state
 *       (only known to mbtowc()), to the initial state, and returns
 *       non-zero if the encoding has non-trivial shift state, or zero
 *       if the encoding is stateless.
 * @long The behavior of mbtowc() depends on the LC_CTYPE category of
 *       the current locale. This function is not multi-thread safe.
 *       mbrtowc() provides a better interface to the same functionality
 *
 * @long Standards: ISO/ANSI C, UNIX98
 *
 * @input pwc A pointer to the resulting wide character.
 * @input str A pointer to a multibyte string.
 * @input n The number of characters to inspect.
 *
 * @return If "str" is not NULL, mbtowc() returns the number of consumed
 *         bytes starting at "str", or 0 if "str" points to a null byte,
 *         or -1 upon failure. If "str" is NULL, mbtowc() returns non-zero
 *         if the encoding has non-trivial shift state, or zero if the
 *         encoding is stateless.
 *
 * @seealso mbstowcs
 * @seealso MB_CUR_MAX
 * @seealso default_restrictions_calling_functions
 */
int   mbtowc(wchar_t *pwc, const char *str, size_t n);

/**
 * @function wctomb
 * @include "stdlib.h"
 *
 * @brief Converts a wide character to a multibyte sequence.
 *
 * @long If "str" is not NULL, wctomb() converts the wide character "wchar"
 *       to its multibyte representation and stores it at the beginning
 *       of the character array pointed to by "str". It updates the shift
 *       state, which is stored in a static anonymous variable only known
 *       to wctomb(), and returns the length of said multibyte
 *       representation, i.e. the number of bytes written at "str".
 * @long The programmer must ensure that there is room for at least
 *       MB_CUR_MAX bytes at "str".
 * @long If "str" is NULL, wctomb() resets the shift state, to the initial
 *       state, and returns non-zero if the encoding has non-trivial
 *       shift state, or zero if the encoding is stateless.
 *
 * @long Standards: ISO/ANSI C, UNIX98
 *
 * @input str A pointer to a character array where the result is stored.
 * @input wchar The wide character to convert.
 *
 * @return If "str" is not NULL, wctomb() returns the number of bytes that
 *         have been written to the byte array at "str". If "wchar" can not
 *         be represented as a multibyte sequence (according to the current
 *         locale), -1 is returned. If "str" is NULL, wctomb() returns
 *         non-zero if the encoding has non-trivial shift state, or zero
 *         if the encoding is stateless.
 *
 * @seealso wcstombs
 * @seealso MB_CUR_MAX
 * @seealso default_restrictions_calling_functions
 */
int wctomb(char *str, wchar_t wchar);

/*
 * Multibyte String Functions.
 */

/**
 * @function mbstowcs
 * @include "stdlib.h"
 *
 * @brief Converts a multibyte string to a wide character string.
 *
 * @long If "pwcs" is not a NULL pointer, mbstowcs() converts the multibyte
 *       string "s" to a wide character string starting at "pwcs".
 *       At most "n" wide characters are written to "pwcs". The conversion
 *       starts in the initial state, and can stop for three reasons:
 * @long 1. An invalid multibyte sequence has been encountered. In this case
 *       (size_t)(-1) is returned.
 * @long 2. "n" non-L'\0' wide characters have been stored at "pwcs".
 *       In this case the number of wide characters written to "pwcs" is
 *       returned, but the shift state at this point is lost.
 * @long 3. The multibyte string has been completely converted, including the
 *       terminating '\0'. In this case the number of wide characters written
 *       to "pwcs", excluding the terminating L'\0' character, is returned.
 * @long The programmer must ensure that there is room for at least "n"
 *       wide characters at "pwcs".
 * @long If "pwcs" is NULL, "n" is ignored, and the conversion proceeds as
 *       above, except that the converted wide characters are not written
 *       out to memory, and that no length limit exists.
 * @long In order to avoid the case 2 above, the programmer should make
 *       sure n >= mbstowcs(NULL, s, 0) + 1.
 * @long The behavior of mbstowcs() depends on the LC_CTYPE category of
 *       the current locale.
 *
 * @long Standards: ISO/ANSI C, UNIX98
 *
 * @input pwcs A pointer to the wide character string, where the result
 *             is stored.
 * @input s A pointer to a multibyte string.
 * @input n The number of bytes written to "pwcs".
 *
 * @return mbstowcs() returns the number of wide characters that make
 *         up the converted part of the wide character string, not
 *         including the terminating null wide character. If an invalid
 *         multibyte sequence was encountered, (size_t)(-1) is returned.
 *
 * @seealso default_restrictions_calling_functions
 */
size_t mbstowcs(wchar_t *pwcs, const char *s, size_t n);

/**
 * @function wcstombs
 * @include "stdlib.h"
 *
 * @brief Converts a wide character string to a multibyte string
 *
 * @long If "s" is not a NULL pointer, wcstombs() converts the wide character
 *       string "pwcs" to a multibyte string starting at "s". At most "n"
 *       bytes  are written to "s". The conversion starts in the initial state,
 *       and can stop for three reasons:
 * @long 1. A wide character has been encountered that can not be represented
 *       as a multibyte sequence (according to the current locale).
 *       In this case (size_t)(-1) is returned.
 * @long 2. The length limit forces a stop. In this case the number of bytes
 *       written to "s" is returned, but the shift state at this point is lost.
 * @long 3. The wide character string has been completely converted, including
 *       the terminating L'\0'. In this case the conversion ends in the initial
 *       state. The number of bytes written to "s", excluding the terminating
 *       '\0' byte, is returned.
 * @long The programmer must ensure that there is room for at least "n"
 *       bytes at "s".
 * @long If "s" is NULL, "n" is ignored, and the conversion proceeds as above,
 *       except that the converted bytes are not written out to memory,
 *       and that no length limit exists.
 * @long In order to avoid the case 2 above, the programmer should make sure
 *       n >= wcstombs(NULL, pwcs, 0) + 1.
 * @long The behavior of wcstombs depends on the LC_CTYPE category of the
 *       current locale.
 *
 * @long Standards: ISO/ANSI C, UNIX98
 *
 * @input s A pointer to a multibyte string, where the result
 *          is stored.
 * @input pwcs A pointer to the wide character string.
 * @input n The number of bytes written to "s".
 *
 * @return wcstombs() returns the number of bytes that make up the
 *         converted part of multibyte sequence, not including the terminating
 *         null byte. If a wide character was encountered which could not
 *         be converted, (size_t)(-1) is returned
 *
 * @seealso default_restrictions_calling_functions
 */
size_t wcstombs(char *s, const wchar_t *pwcs, size_t n);

/*
 * OSE specific implementation of abort().
 */

/**
 * @function efs_abort
 * @include "stdlib.h"
 *
 * @brief Causes abnormal process termination
 *
 * @long efs_abort() causes abnormal process termination.
 *       All open streams are closed and flushed.
 *
 * @long Standards: SVID 3, POSIX, BSD 4.3, ISO 9899 (C99)
 *
 * @input file Filename.
 * @input line Line number.
 * @return Nothing.
 *
 * @seealso exit
 * @seealso default_restrictions_calling_functions
 */
void efs_abort(const char *file, unsigned line);
#define abort() efs_abort(__FILE__, __LINE__)

#ifdef softOSE
#define ZZMK_INLINE_FUNCTIONS
#include "sfk/sfk_stdlib.h"
#endif

#ifdef __cplusplus
}
#if !defined (NO_NAME_STD) && defined (__ARMCC_VERSION)
}  /* namespace std */
using namespace ::std;
#endif /* NO_NAME_STD */
#endif /* __cplusplus */

/* End #ifndef _STDLIB_H */
#endif
