/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 1991-2011 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

/**
 * @public
 * @toc Standard_C:
 * @file string.h
 *
 * @brief 'ISO/IEC 9899, Programming Languages - C' interface string.h
 *
 * @library crt (C Run-time)
 */

#ifndef _STRING_H_
#define	_STRING_H_

#include <stddef.h>
#include "osarch.h"
#include "strings.h"

#ifdef __cplusplus
#if !defined (NO_NAME_STD) && defined (__ARMCC_VERSION)
namespace std {
#endif /* NO_NAME_STD */
extern "C" {
#endif /* __cplusplus */

#ifndef NULL
#define NULL 0
#endif

/**
 * @function memcpy
 *
 * @brief Copy a memory area.
 *
 * @long  The destination and source memory areas must not overlap.
 *
 * @input dest  Destination.
 * @input src   Source, from which memory is copied.
 * @input n     Number of bytes to copy.
 *
 * @return A pointer to dest.
 *
 * @seealso memmove
 *
 * @example memcpy(destination, source, 18);
 */
void *memcpy(void *dest, const void *src, size_t n);

/**
 * @function memmove
 *
 * @brief Copy a memory area allowing overlap of destination and source.
 *
 * @input dest  Destination.
 * @input src   Source, from which memory is copied.
 * @input n     Number of bytes to copy.
 *
 * @return A pointer to dest.
 *
 * @seealso memcpy
 *
 * @example memmove(destination, source, 18);
 */
void *memmove(void *dest, const void *src, size_t n);

/**
 * @function strcpy
 *
 * @brief Copy a null-terminated string.
 *
 * @long  The destination must be large enough. All are copied
 *        including the null byte.
 * @long  The destination and source memory areas must not overlap.
 *
 * @input dest  Destination.
 * @input src   Source, from which the string is copied.
 *
 * @return A pointer to dest.
 *
 * @seealso strncpy
 *
 * @example
 * if(strlen(source) < sizeof(destination))
 * {
 *    strcpy(destination, source);
 * }
 * else
 * {
 *    // Destination is too small
 * }
 */
char *strcpy(char *dest, const char *src);

/**
 * @function strncpy
 *
 * @brief Copy a null-terminated string, but maximized to n characters.
 *
 * @long  The destination must be at least n characters large!
 * @long  A maximum of n characters are copied, including the terminating
 *        null character. If the string in src is longer than n-1, the
 *        result will not be null-terminated.
 * @long  If the string in src is shorter than n-1, dest will be padded
 *        with NULL characters.
 * @long  The destination and source memory areas must not overlap.
 *
 * @input dest Destination.
 * @input src  Source, from which the string is copied.
 * @input n    Number of characters to copy.
 *
 * @return A pointer to dest.
 *
 * @seealso strcpy
 *
 * @example
 * strncpy(destination, source, sizeof(destination));
 * // Check the result
 * if(destination[sizeof(destination)] != NULL)
 * { // The result is not null-terminated
 * }
 */
char *strncpy(char *dest, const char *src, size_t n);

/**
 * @function strcat
 *
 * @brief Concatenate strings.
 *
 * @long  The destination must be large enough for both strings
 *        and the terminating null.
 * @long  The source string is appended to the destination string,
 *        overwriting the destination string's original terminating null.
 * @long  The destination and source memory areas must not overlap.
 *
 * @input dest  Destination
 * @input src   Source, from which the string is copied.
 *
 * @return A pointer to dest.
 *
 * @seealso strncat
 *
 * @example strcat(destination, source);
 */
char *strcat(char *dest, const char *src);

/**
 * @function strncat
 *
 * @brief Concatenate strings, but maximized to n characters from the source.
 *
 * @long  The result always gets a terminating null. Therefore at most
 *        n characters from source plus a null character are written to dest.
 *
 * @input dest  Destination.
 * @input src   Source, from which the string is copied.
 * @input n     Number of characters to copy.
 *
 * @return A pointer to dest.
 *
 * @seealso strcat
 *
 * @example
 * // Overwrite the initial null and a maximum of 17 more characters in
 * // destination, and add a termination null,
 * strncat(destination, source, 18);
 */
char *strncat(char *dest, const char *src, size_t n);

/**
 * @function memcmp
 *
 * @brief Compare a maximum of n bytes in two memory areas.
 *
 * @long  The function compares the bytes from the beginning of the two areas.
 *        As soon as a byte difference is found, the area with the highest
 *        value of this byte is considered greater than the other.
 *        Bytes are treated as unsigned when comparing.
 *        A maximum of n bytes are compared.
 *
 * @input mfirst   First memory area.
 * @input msecond  Second memory area.
 * @input n        Number of bytes to compare.
 *
 * @return An integer >0 if mfirst is greater than msecond,
 *         an integer =0 if mfirst is equal to msecond,
 *         or an integer <0 if mfirst is less than msecond.
 *
 * @seealso strcmp
 *
 * @example
 * if(memcmp(a1,a2,19) > 0)
 * {  // The first differing byte is greater in a1 than in a2
 * }
 */
int    memcmp(const void *mfirst, const void *msecond, size_t n);

/**
 * @function strcmp
 *
 * @brief Compare two strings.
 *
 * @long  The function compares the characters from the beginning of the
 *        two areas. As soon as a character difference is found, the area with
 *        the highest character value is considered greater than the other.
 *        Characters are treated as unsigned when comparing.
 *        The comparison stops at the end of the shortest string.
 *
 * @input sfirst   First string.
 * @input ssecond  Second string.
 *
 * @return An integer >0 if sfirst is greater than ssecond,
 *         an integer =0 if sfirst is equal to ssecond,
 *         or an integer <0 if sfirst is less than ssecond.
 *
 * @seealso memcmp
 * @seealso strncmp
 *
 * @example
 * if(strcmp(s1,s2) > 0)
 * {  // The first string is lexicographically greater
 * }
 */
int    strcmp(const char *sfirst, const char *ssecond);

/**
 * @function strncmp
 *
 * @brief Compare two strings, but a maximum of n characters.
 *
 * @long  The function compares the characters from the beginning of
 *        the two strings.
 *        As soon as a character difference is found, the string with the
 *        highest value of this character is considered greater than the other.
 *        Characters are treated as unsigned when comparing.
 *        The comparison stops at the end of the shortest string,
 *        or when n characters have been compared.
 *        If n is 0, they are always considered equal.
 *
 * @input sfirst   First string.
 * @input ssecond  Second string.
 * @input n        Maximum number of characters to compare.
 *
 * @return  An integer >0 if sfirst is greater than ssecond,
 *          an integer =0 if sfirst is equal to ssecond,
 *          or an integer <0 if sfirst is less than ssecond.
 *
 * @seealso strcmp
 * @seealso memcmp
 *
 * @example
 * // Comparing a maximum of 3 character in beginning of s1 and s2,
 * if(strncmp(s1,s2,3) >0)
 * { // s1 appears lexicographically greater than s2
 * }
 */
int    strncmp(const char *sfirst, const char *ssecond, size_t n);

/**
 * @function strcoll
 *
 * @brief Compare two strings using active locale.
 *
 * @long See @link "" "strcmp" for details.
 *
 * @input sfirst   First string.
 * @input ssecond  Second string.
 *
 * @return  An integer >0 if sfirst is greater than ssecond,
 *          an integer =0 if sfirst is equal to ssecond,
 *          or an integer <0 if sfirst is less than ssecond.
 * @restriction Only standard "C" locale is supported.
 *
 * @seealso strcmp
 */
int    strcoll(const char *sfirst, const char *ssecond);

/**
 * @function strxfrm
 *
 * @brief Transform a string according to active locale.
 *
 * @long  Characters from src are transformed and stored into dest including
 *        null termination. The transform is such that two transformed strings
 *        can be ordered using strcmp() with the same ordering result as if
 *        using strncoll() on the original non-transformed strings.
 * @long  Only a maximum of n characters (including null termination)
 *        are transformed into dest.
 * @long  If n = 0, nothing is stored to dest, but the return value
 *        indicates the required length of the destination string.
 * @long  The result is undefined if the strings overlap.
 *
 * @input dest     Destination string.
 * @input src      String to transform.
 * @input n        Maximum number of characters to transform, including
 *                 the null terminator character.
 *
 * @return The length of the transformed string in dest (length excluding
 *         null terminator).
 *         If the returned length is >= n, the resulting string is undefined.
 *
 * @restriction Only standard "C" locale is supported, making this function
 *              equivalent to strncpy.
 *
 * @seealso strncpy
 */
size_t strxfrm(char *dest, const char *src, size_t n);

/**
 * @function memchr
 *
 * @brief Find a character in memory.
 *
 * @long  Search memory for the character c and return a pointer to it,
 *        or stop searching after n characters and return NULL.
 *
 * @input src  Pointer to where to start the search in memory.
 * @input c    Character to search for.
 * @input n    Maximum number of characters to search.
 *
 * @return A pointer to the found character is returned, if found,
 *         otherwise NULL.
 *
 * @seealso memcmp
 *
 * @example if(memchr(pmem,10,100) != NULL)
 * {  // Found an LF within 100 characters from pmem
 * }
 */
void *memchr(const void *src, int c, size_t n);

/**
 * @function strchr
 *
 * @brief Find a character in a string.
 *
 * @long  Search the first occurrence of the character c in the string str.
 *        Also c = '\0' is supported to search for the null terminator in str.
 *
 * @input str  String to search.
 * @input c    Character to search for.
 *
 * @return A pointer to the found character in str, or NULL if not found.
 *
 * @seealso memchr
 * @seealso strstr
 *
 * @example
 * if((p = strchr(str,'a')) != NULL)
 * { // p points to the first 'a' in str
 * }
 */
char *strchr(const char *str, int c);

/**
 * @function strcspn
 *
 * @brief Search a string for a set of non-rejected characters.
 *
 * @long  Calculate the length of the initial part of the string str, which
 *        does NOT contain any of the characters in reject.
 *        The terminating null is not considered.
 *
 * @input str     String to search.
 * @input reject  Characters that shall not be in the initial part of the string.
 *
 * @return The length of the initial part of the string.
 *
 * @seealso strspn
 */
size_t strcspn(const char *str, const char *reject);

/**
 * @function strpbrk
 *
 * @brief Find characters in a string.
 *
 * @long  Search the first occurrence of any character in setofchar, found
 *        inside the string str.
 *        The terminating null is not considered.
 *
 * @input str       String to search in.
 * @input setofchar Set of characters to search.
 *
 * @return A pointer to the character found inside str, or NULL if not found.
 *
 * @seealso strchr
 */
char *strpbrk(const char *str, const char *setofchar);

/**
 * @function strrchr
 *
 * @brief Reverse find a character in a string.
 *
 * @long  Search the last occurrence of the character c in the string str.
 *        Also c = '\0' is supported to search for the null terminator in str.
 *
 * @input str  String to search.
 * @input c    Character to search for.
 *
 * @return A pointer to the found character in str, or NULL if not found.
 *
 * @seealso strchr
 */
char *strrchr(const char *str, int c);

/**
 * @function strspn
 *
 * @brief Search a string for a set of accepted characters.
 *
 * @long  Calculate the length of the initial part of the string str, which
 *        does contain only characters in accept.
 *        The terminating null is not considered.
 *
 * @input str    String to search.
 * @input accept Characters that shall be in the initial part of the string.
 *
 * @return The length of the initial part of the string.
 *
 * @seealso strcspn
 */
size_t strspn(const char *str, const char *accept);

/**
 * @function strstr
 *
 * @brief Find a substring within a string.
 *
 * @long  Search the first occurrence of substr within str.
 *        The terminating null is not considered.
 *
 * @input str    String in which to search.
 * @input substr String to search for.
 *
 * @return A pointer to the first character of the first occurrence
 *         of substr within str, or NULL if not found. If substr is a
 *         zero-length string, a pointer to str is returned.
 *
 * @seealso strchr
 */
char *strstr(const char *str, const char *substr);

/**
 * @function strtok
 *
 * @brief Get next token from a string.
 *
 * @long  This function is used to extract tokens from a null-terminated
 *        string with multiple tokens. The tokens are delimited in the string
 *        with characters in the set delimiters (one or more).
 *
 * @input source String with tokens to get the first token in source.
 *               Shall be NULL to continue to extract more tokens from the same
 *               string.
 * @input delimiters A string with a set of delimiter characters,
 *                   separating the tokens in the string. It is possible to
 *                   change the delimiter set each call.
 * @return A pointer (within the source string) to the next token, after
 *         replacing the delimiter with a NULL character. A NULL pointer is
 *         returned if no more tokens remain in the string.
 *
 * @restriction strtok is not thread-safe. See instead @link "" "strtok_r".
 *
 * @seealso strpbrk
 * @seealso strtok_r
 */
char *strtok(char *source, const char *delimiters);

#ifndef __STRICT_ANSI__
/**
 * @function strtok_r
 *
 * @brief Get next token from a string - thread-safe.
 *
 * @long  This function is used to extract tokens from a null-terminated
 *        string with multiple tokens. The tokens are delimited in the string
 *        with characters in the set delimiters (one or more).
 *
 * @input source     String with tokens to get the first token in source.
 *                   Shall be NULL to continue to extract more tokens from the
 *                   same string.
 * @input delimiters A string with a set of delimiter characters,
 *                   separating the tokens in the string. It is possible to
 *                   change the delimiter set each call.
 * @input lasts A pointer to a placeholder, where the last character can
 *              be saved until next call with source set to NULL.
 *
 * @return A pointer (within the source string) to the next token, after
 *         replacing the delimiter with a NULL character. A NULL pointer is
 *         returned if no more tokens remain in the string.
 *
 * @restriction strtok_r is not available if compiling with __STRICT_ANSI__
 *
 * @seealso strpbrk
 * @seealso strtok
 * @seealso default_restrictions_calling_functions
 */
char *strtok_r(char *source, const char *delimiters, char **lasts);
#endif

/**
 * @function memset
 *
 * @brief Fill a memory area.
 *
 * @long  Convert c to unsigned char and fill the entire memory area,
 *        pointed to by dest, with c.
 *
 * @input dest Destination.
 * @input c    Character to fill memory with.
 * @input n    Number of bytes to fill.
 *
 * @return The value of dest.
 *
 * @seealso memcpy
 *
 * @example
 * // Fill memory with nulls, between pmem and pmem+99 inclusive
 * char pmem[100];
 * memset(pmem,(int)'\0', 100);
 */
void *memset(void *dest, int c, size_t n);

/**
 * @function strerror
 *
 * @brief Return a string describing an error number.
 *
 * @long  It is the error numbers in errno.h that are supported.
 * @long  The returned string must not be modified.
 *
 * @input errornumber Error for which the description is requested.
 *
 * @return A pointer to an error description string. If the error number
 *         can not be found, it returns a pointer to "Unknown error".
 *
 * @restriction Only standard "C" locale is supported.
 *
 * @seealso errno.h
 */
char *strerror(int errornumber);

/**
 * @function strlen
 *
 * @brief Length of a string.
 *
 * @long Returns the length of the string str. This is the
 *       number of characters excluding the null terminating character.
 *
 * @input str   String to calculate the length of.
 *
 * @return The length of the string.
 */
size_t strlen(const char *str);

#ifdef __cplusplus
}
#if !defined (NO_NAME_STD) && defined (__ARMCC_VERSION)
}  /* namespace std */
using namespace ::std;
#endif /* NO_NAME_STD */
#endif /* __cplusplus */

#endif /* _STRING_H_ */
