/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2007-2011 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */
/**
 * @public
 * @toc OSE:
 * @file sys/mman.h
 *
 * @brief BSD Unix style mmap commands.
 *
 * @long This declares the mmap system calls for memory mapping,
 * including an additional OSE specific ose_fsmap call.
 *
 * @long If demand paging is enabled in OSE, the memory mapping
 * will use demand paging.
 *
 * @library libcore_*.a
 */

#ifndef	MMAN_H_
#define	MMAN_H_

#include "ose.h"
#include "string.h"
#include "sys/types.h"

#ifdef softOSE
#define ZZMK_NAME_MANGLING
#include "sfk/sys/sfk_mman.h"
#endif

#ifdef __cplusplus
extern "C" {
#endif

/* Return value of `mmap' in case of an error.  */
#define MAP_FAILED	((void *) -1)

#define PROT_READ	0x1		/* Page can be read.  */
#define PROT_WRITE	0x2		/* Page can be written.  */
#define PROT_EXEC	0x4		/* Page can be executed.  */
#define PROT_NONE	0x0		/* Pages cannot be accessed.  */

/* Sharing types (must choose one of MAP_SHARED and MAP_PRIVATE).  */
#define MAP_SHARED	0x01		/* Share changes.  */
#define MAP_PRIVATE	0x02		/* Changes are private.  */
#define MAP_FIXED	0x10		/* Interpret addr exactly.  */

/* The mapping is not backed by any file. The fd and offset arguments
   are ignored and R/W mappings are possible. */
#define MAP_ANONYMOUS   0x20

/**
 * @function mmap
 *
 * @brief Memory maps an area in a file.
 *
 * @long
 * Maps logical addresses starting near map_addr and extending for map_len
 * bytes, from map_offset in the file map_fd with protection and sharing
 * attributes according to map_prot and map_flags. If map_addr is non-zero, it
 * is the desired mapping address. If the MAP_FIXED bit is set in map_flags,
 * the mapping will be at map_addr exactly (which must be page aligned).
 * Otherwise the system chooses a convenient nearby address.
 * The file will be mmap:ed in the domain of the caller.
 * @long
 * The flag MAP_SHARED must be set in map_flags. MAP_PRIVATE is not supported.
 * The flag MAP_ANONYMOUS can be set, if the mapping should not be backed
 * by any file. In this case map_fd and map_offset will be ignored, but
 * read/write mappings are allowed.
 * The protection attributes can be PROT_NONE (no access) or a combination
 * of PROT_READ (read access), PROT_WRITE (write access, only with MAP_ANONYMOUS)
 * and PROT_EXEC (executable).
 * @long
 * The return value is the actual mapping address chosen or MAP_FAILED
 * for errors (in which case 'errno' is set).
 *
 * @long Note that mmap can only map read-only data, except when MAP_ANOYMOUS
 * is set.
 *
 * @input map_addr        A hint; specifies the suggested logical base address
 *                        of the region.
 *                        If NULL, the MM can choose an arbitrary address.
 * @input map_len         Specifies the size of the region.
 * @input map_prot        Protection attribute, see above.
 * @input map_flags       Sharing types, see above
 * @input map_fd          The file descriptor of file to be mapped.
 * @input map_offset      The offset into the file to be mapped.
 *
 * @return                The logical address of the region, or MAP_FAILED (-1)
 *                        and sets 'errno'.
 *
 * @seealso default_restrictions_calling_functions
 */
void *
mmap (void *map_addr,
      size_t map_len,
      int map_prot,
      int map_flags,
      int map_fd,
      off_t map_offset);

/**
 * @function munmap
 *
 * @brief Unmaps a logical memory mapped area.
 *
 * @long
 * Unmaps addresses starting near map_addr and extending for map_len bytes.
 * This function will munmap an entire memory mapped region even
 * if [map_addr, map_addr + map_len) only partially covers the region.
 *
 * @input map_addr        Specifies the logical base address of the region.
 * @input map_len         Specifies the size of the region.
 *
 * @return                0 on success, or MAP_FAILED (-1) and sets 'errno'.
 *
 * @seealso default_restrictions_calling_functions
 */
int
munmap (void *map_addr, size_t map_len);

/**
 * @function mlock
 *
 * @brief Locks pages in a memory mapped area.
 *
 * @long Locks the pages in the address range [map_addr,map_addr+map_len) in the
 *        page cache. If the page is not present,
 *        it will be paged in and then locked.
 *
 * @restriction
 *           - Not available in SMP lib.
 *           - Only supported for ARM arch.
 *           - Only available if Demand Paging is enabled.
 *
 * @input map_addr    Logical address of the first page.
 * @input map_len     Specifies the size.
 * @return            Zero if OK, otherwise -1 and sets 'errno'.
 *
 * @seealso default_restrictions_calling_functions
 */
int
mlock(const void *map_addr, size_t map_len);

/**
 * @function munlock
 *
 * @brief Unlocks pages in a memory mapped area.
 *
 * @long Unlocks the pages in the address range [map_addr,map_addr+map_len)
 *        in the page cache.
 *
 * @restriction
 *           - Not available in SMP lib.
 *           - Only supported for ARM arch.
 *           - Only available if Demand Paging is enabled.
 *
 *
 * @input map_addr    Logical address of the first page.
 * @input map_len     Specifies the size.
 * @return            Zero if OK, otherwise -1 and sets 'errno'.
 *
 * @seealso default_restrictions_calling_functions
 */
int
munlock(const void *map_addr, size_t map_len);

/**
 * @function ose_fsmap
 *
 * @brief Emulates the mmap system call, but with a given name of the
 * logical region.
 *
 * @long This function emulates the @link "" "mmap" system call, but with an
 * additional parameter stating the name of the logical memory region.
 * It maps logical addresses starting near map_addr and extending for map_len
 * bytes, from map_offset in the file map_fd.
 * The attributes are according to permissions and cache_ modes, which are
 * defined in mm.sig, one example of a combined permission constant
 * is "SU_RO_USR_RO" and one example of
 * a useful cache mode is "MM_ORDERED_ACCESS".
 * If map_addr is non-zero, it is the desired mapping address.
 * Otherwise the system chooses a convenient nearby address.
 * @long
 * The return value is the actual mapping address chosen or MAP_FAILED
 * for errors (in which case 'errno' is set).
 * @long
 * This function always uses the system domain. Regions mapped with ose_fsmap
 * can be unmapped with munmap().
 *
 * @input  region_name    Specifies the name of the logical memory region.
 * @input  map_addr       A hint; specifies the suggested logical base address
 *                        of the region.
 *                        If NULL, the MM can choose an arbitrary address
 * @input map_len         Specifies the size of the region.
 * @input permissions     The permissions of the mmaped region.
 *                        The permission syntax is defined in mm.sig.
 * @input cache_mode      The cache attributes of the region type, which
 *                        are a combination of cache and access modes.
 *                        Syntax: The cache syntax defined in mm.sig.
 * @input map_fd          The file descriptor of file to be mapped.
 * @input map_offset      The offset into the file to be mapped.
 *
 * @return                The logical address of the region, or MAP_FAILED.
 *
 * @seealso ose_spi/mm.sig
 * @seealso default_restrictions_calling_functions
 */
void *
ose_fsmap (const char *region_name,
           void *map_addr,
           size_t map_len,
           uint32_t permissions,
           uint32_t cache_mode,
           int map_fd,
           off_t map_offset);


#ifdef __cplusplus
}
#endif

#endif	/* mman.h */
