/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2011 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */
/**
 * @public
 * @toc Standard_C:
 * @file sys/time.h
 * @brief Real time clock interface
 */

/*
 ****************************************************************************

        CONTENTS
        --------

        1  Description
        2  History of development
        3  Include files
        4  Macros
        5  Types
        6  Function prototypes
        7  Constants
        8  Variables



 ****************************************************************************
 */



#ifndef _SYS_TIME_H
#define _SYS_TIME_H
#define OSE_SYS_TIME_H

/*
 ****************************************************************************
 * 1  DESCRIPTION.
 ****************************************************************************
 * RTC Unix interface sys/time.h.
 *
 */

/*
 ****************************************************************************
 * 3  INCLUDE FILES.
 ****************************************************************************
 */

#include "osarch.h"
#include <stddef.h>

#ifdef __cplusplus
extern "C" {
#endif

/*
 ****************************************************************************
 * 4  MACROS.
 ****************************************************************************
 */


/*
 ****************************************************************************
 * 5  TYPES.
 ****************************************************************************
 */

/* The time_t type specifies the real time elapsed since Jan 1 1970 00:00:00.
   This representation is also called 'arithmetic time'. */
/**
 * @type time_t
 * @brief Real time elapsed since Jan 1 1970 00:00:00.
 * @long The type used for time_t and what it means is implementation
 * specific. In this implementation time_t counts the number of seconds.
 * This means time_t can represent time until Feb 7 2106 06:28:15.
 * The type however may change, keeping source code compatibility but
 * breaking binary compatibility.
 *
 */

typedef unsigned long time_t;

/* NOTE: The type used for time_t and what it means is implementation
   specific. In this implementation time_t counts the number of seconds.
   This means time_t can represent time until Feb 7 2106 06:28:15.
   The type however may change, keeping source code compatibility but
   breaking binary compatibility. */

/* struct timeval is defined in RTC sys/time.h, INET inet.h and
 * Windows winsock.h.
 */

#ifndef _WINSOCKAPI_
#ifndef TIMEVAL_DEFINED
#define TIMEVAL_DEFINED

/**
 * @type timeval
 * @brief Time value (seconds and microseconds).
 * @field tv_sec  Seconds (a simple "long" value).
 * @field tv_usec Microseconds.
 *
 */

struct timeval
{
  long tv_sec;    /* Seconds. */
  long tv_usec;   /* Microseconds. */
};
#endif
#endif

/**
 * @type timespec
 * @brief Time value with time_t seconds and nanoseconds.
 * @field tv_sec  Seconds (a time_t value).
 * @field tv_nsec Nanoseconds.
 */

struct timespec
{
  time_t tv_sec;  /* Seconds. */
  long tv_nsec;   /* Nanoseconds. */
};

/**
 * @type timezone
 * @brief Timezone structure.
 * @field tz_minuteswest  Minutes west of Greenwich.
 * @field tz_dsttime      Type of daylight saving time correction.
 */

struct timezone
{
  int tz_minuteswest; /* of Greenwich */
  int tz_dsttime;     /* type of DST correction to apply */
};

/*
 ****************************************************************************
 * 6 FUNCTION PROTOTYPES
 ****************************************************************************
 */

/* utimes() must be declared in sys/time.h due to compatibility with Unix
   header files. utimes() is however not implemented in RTC but in EFS. */

#ifdef softOSE
#define utimes(a,b)  zzutimes(a,b)
#endif

/**
 * @function utimes
 *
 * @include sys/time.h
 *
 * @brief Modify access and modification times.
 *
 * @long Set the access and modification times of the selected file specified
 *       by the pathname argument to the time specified by the times argument.
 *       Times can be specified to the level of a microsecond, but it is up to
 *       the underlying file system implementation to honor this.
 *
 * @field pathname Specifies the file to operate on.
 *
 * @field times   This is a timeval[2] struct where the first member is the
 *                last access time to use and the second member is the last
 *                modification to set. Times in the timeval structure are
 *                specified in seconds and microseconds since the Epoch, but
 *                rounding towards the nearest second may occur. If times is
 *                a null pointer the access and modification times are set to
 *                current time.
 *
 * @return Upon successful completion, 0 shall be returned. Otherwise, -1
 *         shall be returned and errno shall be set to indicate the error,
 *         and the file times shall not be affected.
 *
 * @error EACCES errno @link "errno.h" "EACCES"
 * @error ENAMETOOLONG errno @link "errno.h" "ENAMETOOLONG".
 * @error ENOENT errno @link "errno.h" "ENOENT"
 * @error ENOTDIR errno @link "errno.h" "ENOTDIR"
 * @error EPERM errno @link "errno.h" "EPERM"
 * @error EROFS errno @link "errno.h" "EROFS".
 *
 * @seealso default_restrictions_calling_functions
 */
int utimes(const char *pathname, const struct timeval *times);


/**
 * @function wutimes
 *
 * @include sys/time.h
 *
 * @brief Modify access and modification times.
 *
 * @long This function is equivalent to utimes, with the exception that
 *       the pathname is specified as a wide character string.
 *
 * @long Set the access and modification times of the selected file specified
 *       by the pathname argument to the time specified by the times argument.
 *       Times can be specified to the level of a microsecond, but it is up to
 *       the underlying file system implementation to honor this.
 *
 * @field pathname Specifies the file to operate on.
 *
 * @field times   This is a timeval[2] struct where the first member is the
 *                last access time to use and the second member is the last
 *                modification to set. Times in the timeval structure are
 *                specified in seconds and microseconds since the Epoch, but
 *                rounding towards the nearest second may occur. If times is
 *                a null pointer the access and modification times are set to
 *                current time.
 *
 * @return Upon successful completion, 0 shall be returned. Otherwise, -1
 *         shall be returned and errno shall be set to indicate the error,
 *         and the file times shall not be affected.
 *
 * @error EACCES errno @link "errno.h" "EACCES"
 * @error ENAMETOOLONG errno @link "errno.h" "ENAMETOOLONG".
 * @error ENOENT errno @link "errno.h" "ENOENT"
 * @error ENOTDIR errno @link "errno.h" "ENOTDIR"
 * @error EPERM errno @link "errno.h" "EPERM"
 * @error EROFS errno @link "errno.h" "EROFS".
 *
 * @seealso default_restrictions_calling_functions
 */
int wutimes(const wchar_t *pathname, const struct timeval *times);


/*
 ****************************************************************************
 * 7  CONSTANTS.
 ****************************************************************************
 */

/*
 ****************************************************************************
 * 8  VARIABLES.
 ****************************************************************************
 */

#ifdef __cplusplus
}
#endif

#endif
