/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2011 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */
/**
 * @public
 * @toc Standard_C:
 * @file time.h
 *
 * @brief 'ISO/IEC 9899, Programming Languages - C' interface time.h
 *
 * @long The type time_t is defined in sys/time.h and specifies
 *       the real time elapsed since Jan 1, 1970 00:00:00.
 *       This representation is also called 'arithmetic time'.
 *
 * @long The type struct tm is defined in this file and specifies
 *       time broken down into culturally semantic
 *       components. This representation is also called 'broken down time'.
 *
 * @long The standard formatted broken down time string may look like this
 *       "Thu Nov 24 18:22:48 1986\n\0". This format is used, unless
 *       other formats are specified in the functions.
 *
 * @seealso sys/time.h
 */

#ifndef _TIME_H
#define _TIME_H
#define OSE_TIME_H

#include "osarch.h"   /* To determine softOSE */
#include <stddef.h>

#ifdef __cplusplus
#if !defined(NO_NAME_STD) && defined (__ARMCC_VERSION)
namespace std {
#endif /* NO_NAME_STD */
extern "C" {
#endif /* __cplusplus */

#include "sys/time.h"

#ifdef softOSE
#define ZZMK_NAME_MANGLING
#include "sfk/sfk_time.h"
#endif

/**
 * @macro CLOCKS_PER_SEC
 * @brief The number of clock ticks - as returned by clock() - per second.
 * @long  POSIX etc. requires this value to be 1 million.
 */
#define CLOCKS_PER_SEC 1000000

#ifndef _CLOCK_T
#define	_CLOCK_T
/**
 * @type clock_t
 * @brief The clock_t type specifies the amount of CPU time used.
 */
typedef unsigned long clock_t;
#endif	/* !_CLOCK_T */


  /* The time_t type specifies the real time elapsed since Jan 1 1970 00:00:00.
     This representation is also called 'arithmetic time'. */

  /* time_t is defined in sys/time.h. */


/* The struct tm type specifies time broken down into culturally semantic
   components. This representation is also called 'broken down time'. */

/**
 * @type tm
 * @brief Time broken down into semantic components.
 * @long  The struct tm type specifies time broken down into
 *        culturally semantic components. This representation
 *        is also called 'broken down time'.
 * @field tm_sec   Seconds after the minute [0-60]. Value >59 are for an occasional leap second.
 * @field tm_min   Minutes after the hour [0-59].
 * @field tm_hour  Hours since midnight [0-23].
 * @field tm_mday  Day of the month [1-31].
 * @field tm_mon   Months since January [0-11].
 * @field tm_year  Years since 1900.
 * @field tm_wday  Days since Sunday [0-6].
 * @field tm_yday  Days since January 1 [0-365].
 * @field tm_isdst Daylight Savings Time flag.
 */

struct tm
{
  int tm_sec;
  int tm_min;
  int tm_hour;
  int tm_mday;
  int tm_mon;
  int tm_year;
  int tm_wday;
  int tm_yday;
  int tm_isdst;
};

/* Formatted broken down time may look like this
   "Thu Nov 24 18:22:48 1986\n\0". */

/* The following functions are from Standard C. Relevant functions have been
   complemented with reentrant/recursive _r-functions (thread safe). */

/**
 * @function clock
 * @brief Return (the best approximation of) the CPU time used by the program.
 * @long The best approximation in OSE is the time elapsed since system start.
 * @return The CPU time.
 * @seealso default_restrictions_calling_functions
 */
clock_t clock (void);

/**
 * @function difftime
 * @brief Return the difference (in seconds) between two arithmetic times (time1 - time0).
 * @long Implemented as a macro because of floating point requirements.
 * @input time1 Arithmetic time.
 * @input time0 Arithmetic time.
 * @return The difference.
 * @seealso default_restrictions_calling_functions
 */
#ifdef __cplusplus
inline double difftime(time_t time1, time_t time0) \
	{ return ((double)(time1) - (double)(time0)); }
#else
#define difftime(t1,t0) ((double)(t1) - (double)(t0))
#endif

/**
 * @function mktime
 * @brief Convert from broken down (local) time to arithmetic time.
 * @long The tm structure fields tm_wday and tm_yday are ignored.
 * @input time Broken down time.
 * @return The arithmetic time. At an error, it returns -1.
 * @seealso rtc_mkgmtime
 * @seealso default_restrictions_calling_functions
 */
time_t mktime(struct tm *time);

/**
 * @function time
 * @brief Return the current arithmetic time, i.e. the time elapsed since Jan 1, 1970 00:00:00.
 * @input time Pointer to where the result is returned, in addition to being a return value.
 *             Ignored if NULL.
 * @return The current arithmetic time in seconds.
 * @seealso default_restrictions_calling_functions
 */
time_t time(time_t *time);

/**
 * @function asctime
 * @brief Convert the broken down time to formatted time.
 * @input time Broken down time.
 * @return Pointer to a null-terminated string with formatted time (volatile).
 * @seealso asctime_r
 * @seealso default_restrictions_calling_functions
 */
char *asctime(const struct tm *time);
/**
 * @function asctime_r
 * @brief Convert the broken down time to formatted time (thread safe version).
 * @input time Broken down time.
 * @input buffer Pointer to a buffer for the result.
 * @input buf_size Buffer size in bytes, shall be at least 26,
 *        otherwise the result is truncated.
 * @return Pointer to a null-terminated string with formatted time (volatile).
 * @seealso asctime
 * @seealso default_restrictions_calling_functions
 */
char *asctime_r(const struct tm *time, char *buffer, size_t buf_size);

/**
 * @function ctime
 * @brief Convert the arithmetic time to formatted time.
 * @input time Arithmetic time.
 * @return Pointer to a null-terminated string with formatted time (volatile).
 * @seealso ctime_r
 * @seealso default_restrictions_calling_functions
 */
char *ctime(const time_t *time);
/**
 * @function ctime_r
 * @brief Convert the arithmetic time to formatted time (thread safe version).
 * @input time Arithmetic time.
 * @input buffer Pointer to a buffer for the result as null-terminated string.
 * @input buf_size Buffer size in bytes, shall be at least 26,
 *        otherwise the result is truncated.
 * @return Pointer to a null-terminated string with formatted time (volatile).
 * @seealso ctime
 * @seealso default_restrictions_calling_functions
 */
char *ctime_r(const time_t *time, char *buffer, size_t buf_size);

/**
 * @function gmtime
 * @brief Convert arithmetic time to broken down time according to GMT.
 * @input time Arithmetic time.
 * @return Pointer to a struct with broken down time (volatile)..
 * @seealso gmtime_r
 * @seealso default_restrictions_calling_functions
 */
struct tm *gmtime(const time_t *time);
/**
 * @function gmtime_r
 * @brief Convert arithmetic time to broken down time according to GMT (thread safe version).
 * @input time     Arithmetic time.
 * @input buffer   Pointer to a buffer for the result.
 * @return Pointer to a struct with broken down time (volatile)..
 * @seealso gmtime
 * @seealso default_restrictions_calling_functions
 */
struct tm *gmtime_r(const time_t *time, struct tm *buffer);

/**
 * @function localtime
 * @brief Convert arithmetic time to broken down time according to local time zone
 *        and daylight savings time (DST).
 * @input time Arithmetic time.
 * @return Pointer to a struct with broken down time (volatile)..
 * @restriction
 * This system call may enable interrupts if it is called with interrupts
 * disabled.
 * @seealso localtime_r
 * @seealso default_restrictions_calling_functions
 */
struct tm *localtime(const time_t *time);
/**
 * @function localtime_r
 * @brief Convert arithmetic time to broken down time according to local time zone
 *        and daylight savings time (DST) (thread safe version).
 * @input time Arithmetic time.
 * @input buffer Pointer to a buffer for the result.
 * @return Pointer to a struct with broken down time (volatile)..
 * @seealso localtime
 * @seealso default_restrictions_calling_functions
 */
struct tm *localtime_r(const time_t *time, struct tm *buffer);

/**
 * @function strftime
 * @brief Format broken down time into a buffer according to format.
 * @long  Supported format modifiers: aAbBcCdDehHIjklmMnprRStTUwWxXyYZ
 * @input time Broken down time.
 * @input buffer Pointer to a buffer for the result.
 * @input buf_size Buffer size in bytes.
 * @input format Format specification string. Example: "%Y%m%d %H:%M:%S"
 * @return This is the number of characters returned in the buffer
 *         (excluding the terminating null-byte). If 0, nothing is returned
 *         and the buffer is not valid.
 * @seealso default_restrictions_calling_functions
 */
size_t strftime(char *buffer, size_t buf_size, const char *format,
		 const struct tm *time);

/**
 * @function gettimeofday
 * @brief Return UTC time of day. Local time zone not implemented
 * @input tv Pointer to where the current time value shall be returned.
 * @input tz Pointer to a time zone (not implemented, should be NULL).
 * @return 0 if successful, -1 at error (errno is set).
 * @seealso default_restrictions_calling_functions
 */
int gettimeofday(struct timeval *tv, struct timezone *tz);

/**
 * @function settimeofday
 * @brief Set UTC time of day. Local time zone not implemented
 * @input tv Time value to set as current time (seconds and microseconds).
 * @input tz Time zone (not implemented, should be NULL).
 * @return 0 if successful, -1 at error (errno is set).
 * @seealso default_restrictions_calling_functions
 */
int settimeofday(struct timeval *tv, struct timezone *tz);

/**
 * @function tzset
 * @brief Set local time zone. The call is ignored by OSE.
 * @seealso default_restrictions_calling_functions
 */
void tzset(void);

#ifdef softOSE
#define ZZMK_INLINE_FUNCTIONS
#include "sfk/sfk_time.h"
#endif

#ifdef __cplusplus
}
#if !defined(NO_NAME_STD) && defined (__ARMCC_VERSION)
}  /* namespace std */
using namespace ::std;
#endif /* NO_NAME_STD */
#endif /* __cplusplus */

/* End #ifndef _TIME_H */
#endif

