/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2011 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */
/**
 * @public
 * @toc Standard_C:
 * @file unistd.h
 *
 * @brief The POSIX interface unistd.h
 */

#ifndef _UNISTD_H
#define _UNISTD_H
#define OSE_UNISTD_H

#include <stddef.h>
#include <stdarg.h>
#include "osarch.h"
#include "sys/types.h"

#ifdef softOSE
#define ZZMK_NAME_MANGLING
#include "sfk/sfk_unistd.h"
#endif



#ifdef __cplusplus
extern "C" {
#endif

/*
 ****************************************************************************
 * File access test used in the access() function (int amode).
 ****************************************************************************
 */

/**
 * @macro F_OK
 *
 * @include        "unistd.h"
 *
 * @declaration #define F_OK             0
 *
 * @brief Bit indicating check for existence.
 *
 * @long  File access test used as argument amode in the @link "" "access"()
 *        function.
 */
#define F_OK             0

/**
 * @macro R_OK
 *
 * @include        "unistd.h"
 *
 * @declaration #define R_OK             1
 *
 * @brief Bit indicating check for read permission.
 *
 * @long  File access test used as argument amode in the @link "" "access"()
 *        function.
 */
#define R_OK             1

/**
 * @macro W_OK
 *
 * @include        "unistd.h"
 *
 * @declaration #define W_OK             2
 *
 * @brief Bit indicating check for write permission.
 *
 * @long  File access test used as argument amode in the @link "" "access"()
 *        function.
 */
#define W_OK             2

/**
 * @macro X_OK
 *
 * @include        "unistd.h"
 *
 * @declaration #define X_OK             4
 *
 * @brief Bit indicating check for execution and search permission.
 *
 * @long  File access test used as argument amode in the @link "" "access"()
 *        function.
 */
#define X_OK             4

/*
 * Legacy - All O_XXX macros have been relocated to fcntl.h to conform to
 * POSIX. They are declared here for backwards compatibility only.
 */
#define O_RDONLY         0x00000001
#define O_WRONLY         0x00000002
#define O_RDWR           (O_RDONLY | O_WRONLY)
#define O_APPEND         0x00000004
#define O_CREAT          0x00000008
#define O_EXCL           0x00000010
#define O_NONBLOCK       0x00000020
#define O_TRUNC          0x00000040
#define O_DIR            0x00000080
#define O_LINK           0x00000100
#define O_TEXT           0x00000200
#define O_SOCKET         0x00000400
#define O_NOCTTY         0
#define O_ACCMODE        O_RDWR
#define O_TRANS          0x00000800

/*
 ****************************************************************************
 * Constants to be used as third argument in lseek() function (int whence).
 ****************************************************************************
 */

/**
 * @macro SEEK_CUR
 *
 * @include        "unistd.h"
 *
 * @declaration #define SEEK_CUR         0
 *
 * @brief Seek relative to current position.
 *
 * @long  Constant to be used as argument whence in the @link "" "lseek"()
 *        function or the @link "stdio.h" "fseek"() function.
 */
#define SEEK_CUR         0

/**
 * @macro SEEK_END
 *
 * @include        "unistd.h"
 *
 * @declaration #define SEEK_END         1
 *
 * @brief Seek relative to the end of file.
 *
 * @long  Constant to be used as argument whence in the @link "" "lseek"()
 *        function or the @link "stdio.h" "fseek"() function.
 */
#define SEEK_END         1

/**
 * @macro SEEK_SET
 *
 * @include        "unistd.h"
 *
 * @declaration #define SEEK_SET         2
 *
 * @brief Seek relative to the beginning of file.
 *
 * @long  Constant to be used as argument whence in the @link "" "lseek"()
 *        function or the @link "stdio.h" "fseek"() function.
 */
#define SEEK_SET         2

/*
 ****************************************************************************
 * FUNCTION PROTOTYPES
 ****************************************************************************
 */

/**
 * @function _exit
 *
 * @include        "unistd.h"
 *
 * @declaration void
 * _exit(int status);
 *
 * @brief Terminates the process.
 *
 * @long The _exit function performs the following actions:
 * @long - Close all open files owned by the process being terminated.
 *         This can remove temporary files, created by
 *         @link "stdio.h" "tmpfile", if supported by the volume.
 * @long - Send a ChildDone signal to parent, if it has a parent process,
 *         e.g. the process has an environment variable PARENT with the PID
 *         of the parent.
 *         See ose_spi/efs.sig for the ChildDone signal.
 *         A parent can use @link "efs.h" "efs_wait_child"()
 *         to wait for a this signal (including status).
 * @long - Terminates the calling process.
 * @long If running a program in shared execution model, the entire program
 *       is terminated, otherwise the process is terminated.
 *
 * @long Standards:
 *       Portable Operating System Interface (POSIX.1).
 * @long Conformance:
 *       ISO/IEC 9945-1:1990(IEEE 1003.1-1990).
 *
 * @input status Status to return (if possible) to the host environment.
 *
 * @return The _exit function does not return (_exit can not fail).
 *
 * @seealso exit
 * @seealso atexit
 * @seealso default_restrictions_calling_functions
 */
void     _exit(int status);


/**
 * @function access
 *
 * @include        "unistd.h"
 *
 * @declaration int
 * access(const char *pathname, int amode);
 *
 * @brief Check for existence and access permissions of a file or directory.
 *
 * @long The access function checks whether the specified file or directory
 *       exists and, if it is not a directory, and whether the file can be
 *       accessed in the specified mode.
 * @long amode can take the following bit values:
 * @long - @link "" "F_OK"
 * @long - @link "" "W_OK"
 * @long - @link "" "R_OK"
 * @long - @link "" "X_OK"
 * @long The values can be combined together to check for more than one
 *       condition.
 * @long NOTE: amode is an "int" in this function, but "mode_t" in other
 *       functions.
 *
 * @long Standards:
 *       Portable Operating System Interface (POSIX.1).
 * @long Conformance:
 *       ISO/IEC 9945-1:1990(IEEE 1003.1-1990).
 *
 * @input pathname Specifies the path for the file or directory being checked.
 * @input amode    Specifies the access mode being tested.
 *
 * @return The function returns zero if successful, otherwise it returns -1
 *         and errno is set to one of the values below to indicate the error.
 *         access will fail if one or more of the following error conditions
 *         are true:
 *
 * @error  EACCES       errno @link "errno.h" "EACCES".
 *                      Search permission is denied for any component of the
 *                      path name.
 * @error  EINVAL       errno @link "errno.h" "EINVAL".
 *                      An invalid value was specified for amode.
 * @error  ENAMETOOLONG errno @link "errno.h" "ENAMETOOLONG".
 *                      The length of the path argument exceeds the maximum
 *                      allowed on the volume.
 * @error  ENODEV       errno @link "errno.h" "ENODEV".
 *                      The device driver does not exist or is not available.
 * @error  ENOENT       errno @link "errno.h" "ENOENT".
 *                      The named directory does not exist.
 * @error  ENOTDIR      errno @link "errno.h" "ENOTDIR".
 *                      A component of the path name is not a directory.
 * @error  ENOTMOUNTED  errno @link "errno.h" "ENOTMOUNTED".
 *                      The device driver is not mounted.
 * @error  EROFS        errno @link "errno.h" "EROFS".
 *                      Write access was requested for a file residing on a
 *                      read-only file system.
 * @error  ESTALE       errno @link "errno.h" "ESTALE".
 *                      The path points to a remote machine and the link to
 *                      that machine is no longer active.
 *
 * @seealso creat
 * @seealso dup
 * @seealso fdopen
 * @seealso fopen
 * @seealso open
 * @seealso waccess
 * @seealso default_restrictions_calling_functions
 *
 * @example
 * // The example shows how to check if the user has read permission for
 * // a file "ose.h":
 * int ok;
 *
 * ok = access("ose.h", R_OK);
 * if (ok == 0)
 * {
 *    // User may read file.
 * }
 */
int access(const char *pathname, int amode);


/**
 * @function waccess
 *
 * @include        "unistd.h"
 *
 * @declaration int
 * waccess(const wchar_t *pathname, int amode);
 *
 * @brief Check for existence and access permissions of a file or directory.
 *
 * @long This function is equivalent to access, with the exception that
 *       the pathname is specified as a wide character string.
 *
 * @long The waccess function checks whether the specified file or directory
 *       exists and, if it is not a directory, and whether the file can be
 *       accessed in the specified mode.
 * @long amode can take the following bit values:
 * @long - @link "" "F_OK"
 * @long - @link "" "W_OK"
 * @long - @link "" "R_OK"
 * @long - @link "" "X_OK"
 * @long The values can be combined together to check for more than one
 *       condition.
 * @long NOTE: amode is an "int" in this function, but "mode_t" in other
 *       functions.
 *
 * @long Standards:
 *       None - This is an OSE-specific extension.
 * @long Conformance:
 *       None - This is an OSE-specific extension.
 *
 * @input pathname Specifies the path for the file or directory being checked.
 * @input amode    Specifies the access mode being tested.
 *
 * @return The function returns zero if successful, otherwise it returns -1
 *         and errno is set to one of the values below to indicate the error.
 *         access will fail if one or more of the following error conditions
 *         are true:
 *
 * @error  EACCES       errno @link "errno.h" "EACCES".
 *                      Search permission is denied for any component of the
 *                      path name.
 * @error  EINVAL       errno @link "errno.h" "EINVAL".
 *                      An invalid value was specified for amode.
 * @error  ENAMETOOLONG errno @link "errno.h" "ENAMETOOLONG".
 *                      The length of the path argument exceeds the maximum
 *                      allowed on the volume.
 * @error  ENODEV       errno @link "errno.h" "ENODEV".
 *                      The device driver does not exist or is not available.
 * @error  ENOENT       errno @link "errno.h" "ENOENT".
 *                      The named directory does not exist.
 * @error  ENOTDIR      errno @link "errno.h" "ENOTDIR".
 *                      A component of the path name is not a directory.
 * @error  ENOTMOUNTED  errno @link "errno.h" "ENOTMOUNTED".
 *                      The device driver is not mounted.
 * @error  EROFS        errno @link "errno.h" "EROFS".
 *                      Write access was requested for a file residing on a
 *                      read-only file system.
 * @error  ESTALE       errno @link "errno.h" "ESTALE".
 *                      The path points to a remote machine and the link to
 *                      that machine is no longer active.
 *
 * @seealso wcreat
 * @seealso dup
 * @seealso fdopen
 * @seealso wfopen
 * @seealso wopen
 * @seealso access
 * @seealso default_restrictions_calling_functions
 *
 * @example
 * // The example shows how to check if the user has read permission for
 * // a file "ose.h":
 * int ok;
 *
 * ok = waccess(L"ose.h", R_OK);
 * if (ok == 0)
 * {
 *    // User may read file.
 * }
 */
int waccess(const wchar_t *pathname, int amode);


/**
 * @function chdir
 *
 * @include        "unistd.h"
 *
 * @declaration int
 * chdir(const char *pathname);
 *
 * @brief Change the working directory for a process.
 *
 * @long The chdir function changes the current working directory to that
 *       specified by path, i.e. the starting point for path searches for
 *       pathnames not beginning with a slash.
 *
 * @long Standards:
 *       Portable Operating System Interface (POSIX.1).
 * @long Conformance:
 *       ISO/IEC 9945-1:1990(IEEE 1003.1-1990).
 *
 * @input pathname A pointer to the name of the new directory.
 *
 * @return The function returns zero if successful, otherwise it returns -1
 *         and errno is set to one of the values below to indicate the error.
 *         chdir will fail and the current working directory will be unchanged
 *         if one or more of the following error conditions are true:
 *
 * @error EACCES       errno @link "errno.h" "EACCES".
 *                     Search permission is denied for any component of the
 *                     path name.
 * @error ENAMETOOLONG errno @link "errno.h" "ENAMETOOLONG".
 *                     The length of the path argument exceeds the maximum
 *                     allowed on the volume.
 * @error ENOTDIR      errno @link "errno.h" "ENOTDIR".
 *                     A component of the path name is not a directory.
 * @error ENOENT       errno @link "errno.h" "ENOENT".
 *                     The named directory does not exist.
 *
 * @seealso getcwd
 * @seealso wchdir
 * @seealso default_restrictions_calling_functions
 *
 * @example
 * // To change from current directory to the directory "/ram/foodir":
 * int ok;
 *
 * ok = chdir("/ram/foodir");
 * if (ok == 0)
 * {
 *    // "/ram/foodir" is the new current directory.
 * }
 */
int chdir(const char *pathname);


/**
 * @function wchdir
 *
 * @include        "unistd.h"
 *
 * @declaration int
 * wchdir(const wchar_t *pathname);
 *
 * @brief Change the working directory for a process.
 *
 * @long This function is equivalent to chdir, with the exception that
 *       the pathname is specified as a wide character string.
 *
 * @long The wchdir function changes the current working directory to that
 *       specified by path, i.e. the starting point for path searches for
 *       pathnames not beginning with a slash.
 *
 * @long Standards:
 *       None - This is an OSE-specific extension.
 * @long Conformance:
 *       None - This is an OSE-specific extension.
 *
 * @input pathname A pointer to the name of the new directory.
 *
 * @return The function returns zero if successful, otherwise it returns -1
 *         and errno is set to one of the values below to indicate the error.
 *         wchdir will fail and the current working directory will be unchanged
 *         if one or more of the following error conditions are true:
 *
 * @error EACCES       errno @link "errno.h" "EACCES".
 *                     Search permission is denied for any component of the
 *                     path name.
 * @error ENAMETOOLONG errno @link "errno.h" "ENAMETOOLONG".
 *                     The length of the path argument exceeds the maximum
 *                     allowed on the volume.
 * @error ENOTDIR      errno @link "errno.h" "ENOTDIR".
 *                     A component of the path name is not a directory.
 * @error ENOENT       errno @link "errno.h" "ENOENT".
 *                     The named directory does not exist.
 *
 * @seealso wgetcwd
 * @seealso chdir
 * @seealso default_restrictions_calling_functions
 *
 * @example
 * // To change from current directory to the directory "/ram/foodir":
 * int ok;
 *
 * ok = wchdir(L"/ram/foodir");
 * if (ok == 0)
 * {
 *    // L"/ram/foodir" is the new current directory.
 * }
 */
int wchdir(const wchar_t *pathname);


/**
 * @function close
 *
 * @include        "unistd.h"
 *
 * @declaration int
 * close(int fd);
 *
 * @brief Close a file descriptor opened with open/creat.
 *
 * @long The close function closes an open file, e.g. it closes the file
 *       associated with the file descriptor that is indicated. The close
 *       function deallocates the file descriptor indicated by fd. When all
 *       file descriptors associated with an open file description have been
 *       closed, the open file description is freed.
 * @long Standards:
 *       Portable Operating System Interface (POSIX.1).
 * @long Conformance:
 *       ISO/IEC 9945-1:1990 (IEEE 1003.1-1990).
 *
 * @input fd The file descriptor associated with the file to close.
 *
 * @return If successful, the function returns zero, otherwise it returns -1
 *         and errno is set to one of the values below to indicate the error.
 *         close will fail and the file will be unchanged if any of the
 *         following conditions are true:
 *
 * @error EBADF   errno @link "errno.h" "EBADF".
 *                The fd argument is not a valid file descriptor.
 * @error EINTR   errno @link "errno.h" "EINTR".
 *                The close function was interrupted.
 * @error ESTALE  errno @link "errno.h" "ESTALE".
 *                fd points to a remote machine and the link to that machine is
 *                no longer active.
 *
 * @seealso fclose
 * @seealso remove
 * @seealso open
 * @seealso creat
 * @seealso default_restrictions_calling_functions
 *
 * @example
 * // To close the file ose.txt do the following:
 * // The file descriptor, fd, of ose.txt was given when it was opened.
 * int ok;
 *
 * ok = close(fd);
 * if (ok == 0)
 * {
 *    // The file was successfully closed.
 * }
 */
int close(int fd);

/**
 * @function dup
 *
 * @include        "unistd.h"
 *
 * @declaration int
 * dup(int fd);
 *
 * @brief Duplicate a file descriptor.
 *
 * @long dup() returns a new file descriptor. The new file descriptor refers to
 *       the same open file description as the original file descriptor, fd.
 *       The function call is usually used to redirect the stdin or stdout
 *       stream.
 * @long The dup function call is equivalent to:
 *       @link "fcntl.h" "fcntl"(fd, F_DUPFD, 0);
 *
 * @restriction The stream pointer is shared between the original and the
 *              duplicated file.
 *
 * @long Standards:
 *       Portable Operating System Interface (POSIX.1).
 * @long Conformance:
 *       ISO/IEC 9945-1:1990 (IEEE 1003.1-1990).
 *
 * @input fd The file descriptor to duplicate.
 *
 * @return The function returns another file descriptor if successful,
 *         otherwise it returns -1 and errno is set to one of the values below
 *         to indicate the error. dup will fail if any of the following
 *         conditions are true:
 *
 * @error EBADF  errno @link "errno.h" "EBADF".
 *               The fd argument is not a valid open file descriptor.
 * @error EMFILE errno @link "errno.h" "EMFILE"
 *               The number of file descriptor would exceed
 *               @link "stdio.h" "FOPEN_MAX".
 *
 * @seealso access
 * @seealso dup2
 * @seealso open
 * @seealso default_restrictions_calling_functions
 *
 * @example
 * // To duplicate the file ose.txt file descriptor do the following:
 * // The file descriptor (fd) of ose.txt was given when it was opened.
 * new_fd = dup(fd);
 * if (new_fd >= 0)
 * {
 *    // Then ose.txt file descriptor was successfully duplicated.
 * }
 */
int dup(int fd);

/**
 * @function dup2
 *
 * @include        "unistd.h"
 *
 * @declaration int
 * dup2(int fd, int newfd);
 *
 * @brief Duplicate a file descriptor to given integer.
 *
 * @long The dup2 function duplicates the fd file descriptor to the newfd file
 *       descriptor. The newfd is closed before duplication. The newfd file
 *       descriptor will after the call refer to the same open file descriptor
 *       as the original file descriptor, fd.
 * @long The dup2 function call is equivalent to:
 *       @link "" "close"(newfd); @link "fcntl.h" "fcntl"(fd, F_DUPFD, newfd);
 * @long except for the following:
 * @long - If newfd is negative or greater than @link "stdio.h" "FOPEN_MAX",
 *         the dup2 function will return @link "stdio.h" "EOF" and errno will
 *         be set to @link "errno.h" "EBADF".
 * @long - If fd is a valid file descriptor and is equal to newfd, the dup2
 *         function will return newfd without closing it.
 * @long - If fd is not a valid file descriptor, dup2 will fail and not close
 *         newfd.
 * @long - The value returned will always be equal to newfd upon successful
 *         completion.
 *
 * @long Standards:
 *       Portable Operating System Interface (POSIX.1).
 * @long Conformance:
 *       ISO/IEC9945-1:1990(IEEE 1003.1-1990).
 *
 * @input fd    The file descriptor from which the information is duplicated.
 * @input newfd The file descriptor to which it is copied.
 *
 * @return The function returns a file descriptor if successful, otherwise it
 *         returns -1 and errno is set to one of the values below to indicate
 *         the error. dup2 will fail if any of the following conditions are
 *         true:
 *
 * @error EBADF errno @link "errno.h" "EBADF".
 *              The fd argument is not a valid open file descriptor, or the
 *              newfd argument is negative or greater than or equal to
 *              @link "stdio.h" "FOPEN_MAX".
 * @error EINTR errno @link "errno.h" "EINTR".
 *              The function call was interrupted.
 *
 * @seealso dup
 * @seealso open
 * @seealso default_restrictions_calling_functions
 *
 * @example
 * // To duplicate the file descriptor of ose1.txt, assuming also newfd is open
 * // before the call.
 * fd1 = open("ose1.txt", O_RDWR, S_IRGRP);
 * newfd = open("ose2.txt", O_RDWR, S_IRGRP);
 *
 * ok = dup2(fd1, newfd);
 * // fd1 and newfd now both refer to ose1.txt
 */
int dup2(int fd, int newfd);

/**
 * @function fsync
 *
 * @include        "unistd.h"
 *
 * @declaration int
 * fsync(int fd);
 *
 * @brief Writes buffered (unwritten) data for a file descriptor, all the way
 *        through the volume manager to storage.
 *
 * @long The fsync function forces writing of any cached data in memory buffers
 *       of the file, referred to by the fd file descriptor, all the way
 *       through the volume manager to the storage. The function returns when
 *       all data are persistent on the storage.
 * @long For a transactional file, this call will cause a transaction commit.
 * @long For streams, first use @link "stdio.h" "fflush" before fsync. To make
 *       the entire volume state persistent, see @link "efs.h" "efs_sync".
 *
 * @long Standards:
 *       Portable Operating System Interface (POSIX.1).
 * @long Conformance:
 *       ISO/IEC 9945-1:1990(IEEE 1003.1-1990).
 *
 * @input fd The file descriptor.
 *
 * @return Zero if success, otherwise -1 and errno might be set to:
 *
 * @error EBADF  errno @link "errno.h" "EBADF".
 *               Bad file descriptor.
 * @error EAGAIN errno @link "errno.h" "EAGAIN".
 *               Lost contact with the VM volume manager.
 * @error ESTALE errno @link "errno.h" "ESTALE".
 *               fd indicates a file on a remote machine and the link to that
 *               machine is no longer active.
 *
 * @seealso fflush
 * @seealso efs_sync
 * @seealso default_restrictions_calling_functions
 *
 * @example
 * // To write any cached information from memory to the file
 * // "/ram/mydir/myfile.txt":
 * int res;
 * FILE * myfile = fopen("/ram/mydir/myfile.txt");
 * int myfile_fd = fileno(myfile);
 *
 * // perform file operations on stream myfile, writing to the file ...
 * ....
 * // Ensure information is output to the file:
 * res = fflush(myfile);
 * if (res != 0)
 * {
 *    //  Error from fflush, could check errno ...
 * }
 * res = fsync(myfile_fd);
 * if (res == 0)
 * {
 *    //  All cached data was written to the file.
 * }
 */
int fsync(int fd);

/**
 * @function ftruncate
 *
 * @include        "unistd.h"
 *
 * @declaration int
 * ftruncate(int fd, off_t length);
 *
 * @brief Truncates a file to a specified length i.e. sets the file length to
 *        the value of length.
 *
 * @long Truncates a file to a specified length i.e. sets the file length to
 *       the value of length. If the file truncated is larger than length, data
 *       is lost. If the file is smaller the new file data is set to zero.
 *
 * @restriction The file descriptor to truncate must be a file managed by a
 *              volume manager. It depends on the volume manager if the
 *              function is supported. The normal behavior of OSE's volume
 *              managers is to support ftruncate to length zero or current
 *              size.
 *
 * @long Standards:
 *       Portable Operating System Interface (POSIX.1).
 * @long Conformance:
 *       ISO/IEC 9945-1:1996 (IEEE 1003.1-1996).
 *
 * @input fd     The file descriptor of the file to truncate.
 * @input length The new length of the file.
 *
 * @return If successful, the function returns 0, otherwise it returns -1 and
 *         an errno.
 *
 * @error ESTALE errno @link "errno.h" "ESTALE".
 *               The file descriptor is stale or the contact was lost with the
 *               volume manager handling the file.
 * @error EBADF  errno @link "errno.h" "EBADF".
 *               Invalid file descriptor.
 * @error ENODEV errno @link "errno.h" "ENODEV".
 *               fd is a socket file descriptor.
 * @error ENOSYS errno @link "errno.h" "ENOSYS".
 *               Function not implemented.
 *
 * @seealso truncate
 * @seealso default_restrictions_calling_functions
 *
 * @example
 * // Set the length of fd to 0:
 * ftruncate(fd, 0);
 */
int ftruncate(int fd, off_t length);


/**
 * @function getcwd
 *
 * @include        "unistd.h"
 *
 * @declaration char *
 * getcwd(char *buf, size_t size);
 *
 * @brief Gets the current working directory of a process.
 *
 * @long getcwd() returns the full path of the current working directory for
 *       the client. The current working directory is the path that will be
 *       prepended to all relative pathnames, i.e. path names not beginning
 *       with a slash.
 *
 * @long Standards:
 *       Portable Operating System Interface (POSIX.1).
 * @long Conformance:
 *       ISO/IEC 9945-1:1996 (IEEE 1003.1-1996).
 * @long The ability to allocate a buffer for the result if no buffer is
 *       provided is an OSE-specific extension. The standard does not
 *       specify a behaviour in this situation.
 *
 * @restriction The Posix standard specifies that the returned path shall not
 *          include any extra slashes. The current implementation does
 *          however append an extra slash to the path. The implementation
 *          will be updated to comply with the standard in a future version
 *          of OSE.
 *
 * @input buf  Pointer to a buffer where the path of the current working
 *             directory name shall be stored, or a null pointer. A buffer
 *             that can be returned by free() will be allocated and used
 *             buf is a null pointer.
 *
 * @input size The length of buffer buf. This parameter is ignored if buf
 *             is a null pointer.
 *
 * @return  The function returns a pointer to the buffer containing the 
 *          current working directory if successful, otherwise it
 *          returns NULL and errno is set to the value below to indicate the
 *          error. getcwd will fail if the following condition is true:
 *
 * @error ERANGE errno @link "errno.h" "ERANGE".
 *               The provided buffer is too small to hold the pathname.
 * @error ENOMEM errno @link "errno.h" "ENOMEM".
 *                The function failed to allocate a buffer.
 *
 * @seealso chdir
 * @seealso wgetcwd
 * @seealso default_restrictions_calling_functions
 *
 * @example
 * // To fetch the current directory do the following:
 * char buffer[10];
 * char *dir;
 *
 * dir = getcwd(buffer,10);
 * if (dir != NULL)
 * {
 *    // dir points at current directory
 * }
 * else
 * {
 *    // The current working directory pathname is longer than 9 characters.
 * }
 */
char *getcwd(char *buf, size_t size);


/**
 * @function wgetcwd
 *
 * @include        "unistd.h"
 *
 * @declaration wchar_t *
 * wgetcwd(wchar_t *buf, size_t size);
 *
 * @brief Gets the current working directory of a process.
 *
 * @long This function is equivalent to getcwd, with the exception that
 *       the pathname is returned as a wide character string.
 *
 * @long wgetcwd() returns the full path of the current working directory for
 *       the client. The current working directory is the path that will be
 *       prepended to all relative pathnames, i.e. path names not beginning
 *       with a slash.
 *
 * @long Standards:
 *       None - This is an OSE-specific extension.
 * @long Conformance:
 *       None - This is an OSE-specific extension.
 *
 * @input buf  Pointer to a buffer where the path of the current working
 *             directory name shall be stored, or a null pointer. A buffer
 *             that can be returned by free() will be allocated and used
 *             buf is a null pointer.
 *
 * @input size The length of buffer buf. This parameter is ignored if buf
 *             is a null pointer.
 *
 * @return  The function returns a pointer to the buffer containing the 
 *          current working directory if successful, otherwise it
 *          returns NULL and errno is set to the value below to indicate the
 *          error. getcwd will fail if the following condition is true:
 *
 * @error ERANGE errno @link "errno.h" "ERANGE".
 *               The provided buffer is too small to hold the pathname.
 * @error ENOMEM errno @link "errno.h" "ENOMEM".
 *                The function failed to allocate a buffer.
 *
 * @seealso chdir
 * @seealso wgetcwd
 * @seealso default_restrictions_calling_functions
 *
 * @example
 * // To fetch the current directory do the following:
 * wchar_t *dir;
 *
 * dir = wgetcwd(NULL, 0);
 * if (dir != NULL)
 * {
 *    // dir points at current directory
 * }
 * else
 * {
 *    // Error, presumably out of memory.
 * }
 */
wchar_t *wgetcwd(wchar_t *buf, size_t size);


/**
 * @function gethostname
 *
 * @include "unistd.h"
 *
 * @declaration int gethostname(char *name, size_t namelen);
 *
 * @brief Get host name.
 *
 * @long gethostname() returns the name of the current
 * host. gethostname() will search for a host name starting from
 * volatile store as set by sethostname() followed by persistent
 * storage and last if no host name has been found the string
 * localhost will be returned.
 *
 * @output name Pointer to buffer in which to store the retrieved
 * name. If insufficient space is provided the name is truncated. The
 * buffer is always null-terminated.
 *
 * @input namelen Length of name buffer.
 *
 * @return 0 if successful; otherwise -1 with errno set to indicate the error.
 *
 * @error EFAULT If name is an invalid address.
 * @error EINVAL If namelen is negative.
 *
 * @seealso sethostname
 * @seealso default_restrictions_calling_functions
 */
int gethostname(char *name, size_t namelen);

/**
 * @function isatty
 *
 * @include        "unistd.h"
 *
 * @declaration int
 * isatty(int fd);
 *
 * @brief Checks if a file descriptor refers to a terminal line.
 *
 * @long The isatty function returns non-zero if this file descriptor refers to
 *       a terminal device.
 *
 * @restriction Currently it is considered a terminal if the preferred buffer
 *              mode at open was reported as line buffering
 *              see ose_spi/fm.sig (FmBuffEnum) for details.
 *
 * @long Standards:
 *       Portable Operating System Interface (POSIX.1).
 * @long Conformance:
 *       ISO/IEC 9945-1:1996 (IEEE 1003.1-1996).
 *
 * @input fd The file descriptor.
 *
 * @return The function returns true (a positive value) if it is a terminal
 *         line, or zero (0) if not. At an error it returns -1 and errno is set
 *         to one of the values below to indicate the error.
 *
 * @error EBADF  errno @link "errno.h" "EBADF".
 *               Bad file descriptor.
 * @error ESTALE errno @link "errno.h" "ESTALE".
 *               fd points to a remote machine and the link to that machine is
 *               no longer active.
 *
 * @seealso default_restrictions_calling_functions
 *
 * @example
 * int ok;
 *
 * ok = isatty(fd);
 * if (ok == 0)
 * {
 *    // It is not line-buffered.
 * }
 */
int isatty(int fd);


/**
 * @function link
 *
 * @include        "unistd.h"
 *
 * @brief Create a new hard link for an existing file.
 *
 * @long Create a new hard link named by newpath for an existing file named
 *       by oldpath.
 *
 * @long Standards:
 *       Portable Operating System Interface (POSIX.1).
 * @long Conformance:
 *       ISO/IEC 9945-1:1996 (IEEE 1003.1-1996).
 *
 * @restriction The level of support for this function is dependent on the
 *              volume manager. JEFF has full support. Many of the other
 *              volume managers do not support creation of hard links.
 *
 * @input oldpath    Pointer to pathname string representing an existing file.
 * @input newpath    Pointer to a pathename string naming the new link that
 *                   is to be created.
 * @return           Upon successful completion, link() returns a value of zero.
 *                   Otherwise, a value of -1 is returned and errno is set to
 *                   indicate the error.
 *
 * @error EACCES  errno @link "errno.h" "EACCES".
 *                A component of either path prefix denies search permission;
 *                or the requested link requires writing in a directory with
 *                a mode that denies write permission; or the calling process
 *                does not have permission to access the existing file, and
 *                this is required by the implementation.
 *
 * @error EEXIST  errno @link "errno.h" "EEXIST".
 *                The link named by newpath exists.
 *
 * @error EMLINK  errno @link "errno.h" "EMLINK".
 *                The number of links to the file named by oldpath would
 *                exceed {LINK_MAX}.
 *
 * @error ENAMETOOLONG  errno @link "errno.h" "ENAMETOOLONG".
 *                The length of either the oldpath or the newpath string
 *                exceeds {PATH_MAX}.
 *
 * @error ENOENT  errno @link "errno.h" "ENOENT".
 *                A component of either path prefix does not exist, the file
 *                named by oldpath does not exist, or either path parameter
 *                points to an empty string.
 *
 * @error ENOSPC  errno @link "errno.h" "ENOSPC".
 *                The directory that would contain the link can not be extended.
 *
 * @error ENOTDIR errno @link "errno.h" "ENOTDIR".
 *                A component of either path prefix is not a directory.
 *
 * @error EPERM   errno @link "errno.h" "EPERM".
 *                The file named by existing is a directory, and either the
 *                calling process does not have appropriate privileges or the
 *                implementation prohibits using link() on directories.
 *
 * @error EROFS   errno @link "errno.h" "EROFS".
 *                The requested link requires writing in a directory on a
 *                read-only file system.
 *
 * @error EXDEV   errno @link "errno.h" "EXDEV".
 *                The two paths are on different file systems, and the
 *                implementation does not support links between file systems.
 *
 * @seealso wlink
 * @seealso default_restrictions_calling_functions
 */
int link(const char *oldpath, const char *newpath);


/**
 * @function wlink
 *
 * @include        "unistd.h"
 *
 * @brief Create a new hard link for an existing file.
 *
 * @long Create a new hard link named by newpath for an existing file named
 *       by oldpath.
 *
 * @long This function is equivalent to link, with the exception that
 *       the pathname strings are specified as a wide character strings.
 *
 * @long Standards:
 *       None - This is an OSE-specific extension.
 * @long Conformance:
 *       None - This is an OSE-specific extension.
 *
 * @restriction The level of support for this function is dependent on the
 *              volume manager. JEFF has full support. Many of the other
 *              volume managers do not support creation of hard links.
 *
 * @input oldpath    Pointer to pathname string representing an existing file.
 * @input newpath    Pointer to a pathename string naming the new link that
 *                   is to be created.
 * @return           Upon successful completion, wlink() returns zero.
 *                   Otherwise, -1 is returned and errno is set to indicate
 *                   the error.
 *
 * @error EACCES  errno @link "errno.h" "EACCES".
 *                A component of either path prefix denies search permission;
 *                or the requested link requires writing in a directory with
 *                a mode that denies write permission; or the calling process
 *                does not have permission to access the existing file, and
 *                this is required by the implementation.
 *
 * @error EEXIST  errno @link "errno.h" "EEXIST".
 *                The link named by newpath exists.
 *
 * @error EMLINK  errno @link "errno.h" "EMLINK".
 *                The number of links to the file named by oldpath would
 *                exceed {LINK_MAX}.
 *
 * @error ENAMETOOLONG  errno @link "errno.h" "ENAMETOOLONG".
 *                The length of either the oldpath or the newpath string
 *                exceeds {PATH_MAX}.
 *
 * @error ENOENT  errno @link "errno.h" "ENOENT".
 *                A component of either path prefix does not exist, the file
 *                named by oldpath does not exist, or either path parameter
 *                points to an empty string.
 *
 * @error ENOSPC  errno @link "errno.h" "ENOSPC".
 *                The directory that would contain the link can not be extended.
 *
 * @error ENOTDIR errno @link "errno.h" "ENOTDIR".
 *                A component of either path prefix is not a directory.
 *
 * @error EPERM   errno @link "errno.h" "EPERM".
 *                The file named by existing is a directory, and either the
 *                calling process does not have appropriate privileges or the
 *                implementation prohibits using wlink() on directories.
 *
 * @error EROFS   errno @link "errno.h" "EROFS".
 *                The requested link requires writing in a directory on a
 *                read-only file system.
 *
 * @error EXDEV   errno @link "errno.h" "EXDEV".
 *                The two paths are on different file systems, and the
 *                implementation does not support links between file systems.
 *
 * @seealso link
 * @seealso default_restrictions_calling_functions
 */
int wlink(const wchar_t *oldpath, const wchar_t *newpath);


/**
 * @function lseek
 *
 * @include        "unistd.h"
 *
 * @brief Sets the position of a file descriptor, relative to
 *        beginning/end/current.
 *
 * @long The lseek function is used to place the stream pointer of a specified
 *       file in a new position. The next operation on that file will take
 *       place at that new position.
 *
 * @long The fd argument is an open file descriptor. The lseek function will
 *       set the file offset for the open file description associated with
 *       fd as follows:
 * @break
 *       - If whence is @link "" "SEEK_SET", the offset is set to offset
 *         bytes.
 * @break
 *       - If whence is @link "" "SEEK_CUR", the offset is set to its current
 *         value plus offset bytes.
 * @break
 *       - If whence is @link "" "SEEK_END", the offset is set to the size of
 *         the file plus offset bytes.
 *
 * @long If the volume manager supports this, the lseek function can allow the
 *       file offset to be set beyond the end of existing data in the file. If
 *       data is later written at this point, subsequent reads of data in the
 *       gap could return bytes with the value zero until data is actually
 *       written into the gap. The lseek function will not itself extend the
 *       size of a file.
 *
 * @restriction It depends on the volume manager if the function is supported.
 *       Some volume managers do not support positioning outside the
 *       current file data.
 *
 * @long Standards:
 *       Portable Operating System Interface (POSIX.1).
 * @long Conformance:
 *       ISO/IEC 9945-1:1990 (IEEE 1003.1-1990).
 *
 * @input fd     The file descriptor.
 * @input offset The number of bytes the file position shall be moved.
 * @input whence A position of the file.
 *
 * @return The function returns the offset location if successful, otherwise it
 *         returns ((off_t) -1) and errno is set to one of the values below to
 *         indicate the error. lseek will fail if one or more of the following
 *         conditions are true:
 *
 * @error EBADF  errno @link "errno.h" "EBADF".
 *               The fd argument is not a valid file descriptor.
 * @error EINVAL errno @link "errno.h" "EINVAL".
 *               The whence argument is not a proper value, or the resulting
 *               file offset would be invalid.
 * @error ESPIPE errno @link "errno.h" "ESPIPE".
 *               The fd argument is associated with a pipe or FIFO.
 * @error ESTALE errno @link "errno.h" "ESTALE".
 *               fd points to a remote machine and the link to that machine is
 *               no longer active.
 *
 * @seealso fseek
 * @seealso rewind
 * @seealso fsetpos
 * @seealso default_restrictions_calling_functions
 *
 * @example
 * // To place the stream pointer at the beginning of an open file, do the
 * // following:
 * off_t newpos;
 *
 * newpos = lseek(fd, 0, SEEK_SET);
 * if (newpos != (off_t)-1)
 * {
 *    // The stream position was set to the beginning.
 * }
 */
off_t lseek(int fd, off_t offset, int whence);


/**
 * @function read
 *
 * @include        "unistd.h"
 *
 * @declaration ssize_t
 * read(int fd, void *buffer, size_t nbyte);
 *
 * @brief Reads data from a file using a file descriptor.
 *
 * @long The read function attempts to read nbyte bytes from the file
 *       associated with the open file descriptor, fd, into the buffer pointed
 *       to by buffer. If nbyte is zero, the read function will return zero and
 *       have no other results.
 *
 * @long On a regular file or other file capable of seeking, read will start at
 *       a position in the file given by the file offset associated with fd.
 *       Before successful return from read, the file offset will be
 *       incremented by the number of bytes actually read. On a file not
 *       capable of seeking, read will start from the current position.
 *
 * @long Standards:
 *       Portable Operating System Interface (POSIX.1).
 *
 * @long Conformance:
 *       ISO/IEC 9945-1:1990 (IEEE 1003.1-1990).
 *
 * @input fd     The file descriptor to the associated stream.
 * @input buffer A pointer to the buffer to where the read data is stored.
 * @input nbyte  The number of characters to be read.
 *
 * @return The function returns the number of bytes actually read and placed in
 *         the buffer if successful, otherwise it returns -1 and errno is set
 *         to one of the values below to indicate the error.
 *         read will fail if one or more of the following conditions are true:
 *
 * @error EAGAIN errno @link "errno.h" "EAGAIN".
 *               The @link "" "O_NONBLOCK" flag is set for the file descriptor
 *               and the client would be delayed in the read operation.
 * @error EBADF  errno @link "errno.h" "EBADF".
 *               The fd argument is not a valid file descriptor.
 * @error EISDIR errno @link "errno.h" "EISDIR".
 *               The fd argument is a directory.
 * @error EOVERFLOW  errno @link "errno.h" "EOVERFLOW".
 *               The ranges are to big.
 * @error EINTR  errno @link "errno.h" "EINTR".
 *               The read operation was interrupted, and either no data was
 *               transferred or the file manager does not report partial
 *               transfer for this file.
 * @error ESTALE errno @link "errno.h" "ESTALE".
 *               fd points to a remote machine and the link to that machine is
 *               no longer active.
 * @error EMSGSIZE errno @link "errno.h" "EMSGSIZE".
 *               When reading a notification, it read nothing,
 *               but there is data in the queue. The read
 *               buffer was too small to fit a single notification.
 *
 * @seealso open
 * @seealso write
 * @seealso fopen
 * @seealso default_restrictions_calling_functions
 *
 * @example
 * // To read 10 characters from the ose.txt file do the following:
 * char buffer[10];
 * int ch_read;
 *
 * ch_read = read(fd, &buffer, sizeof(buffer));
 * if (ch_read == sizeof(buffer))
 * {
 *    // Characters were successfully read.
 * }
 */
ssize_t read(int fd, void *buffer, size_t nbyte);


/**
 * @function rmdir
 *
 * @include        "unistd.h"
 *
 * @declaration int
 * rmdir(const char *pathname);
 *
 * @brief Removes a directory.
 *
 * @long The rmdir function removes a directory whose name is given by
 *       pathname. The directory may only be removed if it is an empty
 *       directory.
 *
 * @restriction A directory which is the current directory of a process can
 *              still be removed.
 * @break
 *              It is volume manager dependent if an opened directory is
 *              removed or not.
 *
 * @long Standards:
 *       Portable Operating System Interface (POSIX.1).
 * @long Conformance:
 *       ISO/IEC 9945-1:1990 (IEEE 1003.1-1990).
 *
 * @input pathname The pointer to the directory name associated with the empty
 *                 directory to remove.
 *
 * @return The function returns zero if successful, otherwise it returns -1 and
 *         errno is set to one of the values below to indicate the error.
 *         rmdir will fail if one or more of the following conditions are true:
 *
 * @error EACCES       errno @link "errno.h" "EACCES".
 *                     Search permission is denied for a component of the path
 *                     prefix, or write permission is denied on the parent
 *                     directory of the directory to be removed.
 * @error EBUSY        errno @link "errno.h" "EBUSY".
 *                     The directory named by the path argument can not be
 *                     removed because it is being used by a client.
 * @error ENAMETOOLONG errno @link "errno.h" "ENAMETOOLONG".
 *                     The length of the path argument exceeds the maximum
 *                     allowed on the volume.
 * @error ENODEV       errno @link "errno.h" "ENODEV".
 *                     No such device.
 * @error ENOENT       errno @link "errno.h" "ENOENT".
 *                     The path argument names a non-existent directory or
 *                     points to an empty string.
 * @error ENOTEMPTY    errno @link "errno.h" "ENOTEMPTY".
 *                     The path argument names a directory that is not an empty
 *                     directory.
 * @error ENOTMOUNTED  errno @link "errno.h" "ENOTMOUNTED".
 *                     Not mounted.
 * @error EROFS        errno @link "errno.h" "EROFS".
 *                     The directory entry to be removed resides on a read-only
 *                     file system.
 *
 * @seealso mkdir
 * @seealso wrmdir
 * @seealso default_restrictions_calling_functions
 *
 * @example
 * // To delete the directory "OLDDIR" do the following:
 * int ok;
 *
 * ok = rmdir("OLDDIR");
 * if (ok == 0)
 * {
 *    // Directory was successfully deleted.
 * }
 */
int rmdir(const char *pathname);


/**
 * @function wrmdir
 *
 * @include        "unistd.h"
 *
 * @declaration int
 * wrmdir(const wchar_t *pathname);
 *
 * @brief Removes a directory.
 *
 * @long This function is equivalent to rmdir, with the exception that
 *       the pathname is specified as a wide character string.
 *
 * @long The wrmdir function removes a directory whose name is given by
 *       pathname. The directory may only be removed if it is an empty
 *       directory.
 *
 * @restriction A directory which is the current directory of a process can
 *              still be removed.
 * @break
 *              It is volume manager dependent if an opened directory is
 *              removed or not.
 *
 * @long Standards:
 *       None - This is an OSE-specific extension.
 * @long Conformance:
 *       None - This is an OSE-specific extension.
 *
 * @input pathname The pointer to the directory name associated with the empty
 *                 directory to remove.
 *
 * @return The function returns zero if successful, otherwise it returns -1 and
 *         errno is set to one of the values below to indicate the error.
 *         wrmdir will fail if one or more of the following conditions are true:
 *
 * @error EACCES       errno @link "errno.h" "EACCES".
 *                     Search permission is denied for a component of the path
 *                     prefix, or write permission is denied on the parent
 *                     directory of the directory to be removed.
 * @error EBUSY        errno @link "errno.h" "EBUSY".
 *                     The directory named by the path argument can not be
 *                     removed because it is being used by a client.
 * @error ENAMETOOLONG errno @link "errno.h" "ENAMETOOLONG".
 *                     The length of the path argument exceeds the maximum
 *                     allowed on the volume.
 * @error ENODEV       errno @link "errno.h" "ENODEV".
 *                     No such device.
 * @error ENOENT       errno @link "errno.h" "ENOENT".
 *                     The path argument names a non-existent directory or
 *                     points to an empty string.
 * @error ENOTEMPTY    errno @link "errno.h" "ENOTEMPTY".
 *                     The path argument names a directory that is not an empty
 *                     directory.
 * @error ENOTMOUNTED  errno @link "errno.h" "ENOTMOUNTED".
 *                     Not mounted.
 * @error EROFS        errno @link "errno.h" "EROFS".
 *                     The directory entry to be removed resides on a read-only
 *                     file system.
 *
 * @seealso wmkdir
 * @seealso rmdir
 * @seealso default_restrictions_calling_functions
 *
 * @example
 * // To delete the directory "OLDDIR" do the following:
 * int ok;
 *
 * ok = wrmdir(L"OLDDIR");
 * if (ok == 0)
 * {
 *    // Directory was successfully deleted.
 * }
 */
int wrmdir(const wchar_t *pathname);


/**
 * @function sethostname
 *
 * @include        "unistd.h"
 *
 * @declaration int sethostname(const char *name, size_t namelen);
 *
 * @brief Set host name.
 *
 * @long sethostname() non-persistently sets the name of the current
 * host to name. The name set with sethostname() will override any
 * name configured earlier but will be lost at restart. To permanently
 * change host name, use sysparam_write(), spw or configuration file.
 *
 * @input name Pointer to null terminated name string.
 *
 * @input namelen Length of name-buffer. The maximum allowed name
 * length is 255 bytes.
 *
 * @return 0 if successful; otherwise -1 with errno set to indicate the error.
 *
 * @error EFAULT If name is an invalid address.
 * @error EINVAL If namelen is negative or larger than the maximum
 * allowed size.
 * @error EPERM If host name could not be set.
 *
 * @seealso gethostname
 * @seealso sysparam_write
 * @seealso default_restrictions_calling_functions
 *
 */
int sethostname(const char *name, size_t namelen);


/**
 * @function truncate
 *
 * @include        "unistd.h"
 *
 * @declaration int
 * truncate(const char *pathname, off_t length);
 *
 * @brief Truncates a file to a specified length.
 *
 * @long Truncates a file to a specified length, i.e. sets the file length to
 *       the value of length. If the file truncated is larger than length, data
 *       is lost. If the file is smaller the new file data is set to zero.
 *
 * @restriction The level of support for this function is dependent on the
 *              volume manager. JEFF and VFAT have full support. Many of the
 *              other volume managers only support truncating to length zero.
 *
 * @long Standards:
 *       BSD 4.2.
 * @long Conformance:
 *       BSD 4.2.
 *
 * @input pathname A pointer to the name of a file.
 * @input length   The new length of the file.
 *
 * @return If successful, the function returns 0, otherwise it returns -1 and
 *         an errno.
 *
 * @error ESTALE errno @link "errno.h" "ESTALE".
 *               The file descriptor is stale or the contact was lost with the
 *               volume manager handling the file.
 * @error EBADF  errno @link "errno.h" "EBADF".
 *               Invalid file descriptor.
 * @error ENODEV errno @link "errno.h" "ENODEV".
 *               fd is a socket file descriptor.
 * @error ENOSYS errno @link "errno.h" "ENOSYS".
 *               Function not implemented.
 *
 * @seealso ftruncate
 * @seealso wtruncate
 * @seealso default_restrictions_calling_functions
 *
 * @example
 * // Set the length of "/ram/dummy.txt" to 0:
 * truncate("/ram/dummy.txt", 0);
 */
int truncate(const char *pathname, off_t length);


/**
 * @function wtruncate
 *
 * @include        "unistd.h"
 *
 * @declaration int
 * wtruncate(const wchar_t *pathname, off_t length);
 *
 * @brief Truncates a file to a specified length.
 *
 * @long This function is equivalent to truncate, with the exception that
 *       the pathname is specified as a wide character string.
 *
 * @long Truncates a file to a specified length, i.e. sets the file length to
 *       the value of length. If the file truncated is larger than length, data
 *       is lost. If the file is smaller the new file data is set to zero.
 *
 * @restriction The level of support for this function is dependent on the
 *              volume manager. JEFF and VFAT have full support. Many of the
 *              other volume managers only support truncating to length zero.
 *
 * @long Standards:
 *       None - This is an OSE-specific extension.
 * @long Conformance:
 *       None - This is an OSE-specific extension.
 *
 * @input pathname A pointer to the name of a file.
 * @input length   The new length of the file.
 *
 * @return If successful, the function returns 0, otherwise it returns -1 and
 *         an errno.
 *
 * @error ESTALE errno @link "errno.h" "ESTALE".
 *               The file descriptor is stale or the contact was lost with the
 *               volume manager handling the file.
 * @error EBADF  errno @link "errno.h" "EBADF".
 *               Invalid file descriptor.
 * @error ENODEV errno @link "errno.h" "ENODEV".
 *               fd is a socket file descriptor.
 * @error ENOSYS errno @link "errno.h" "ENOSYS".
 *               Function not implemented.
 *
 * @seealso ftruncate
 * @seealso truncate
 * @seealso default_restrictions_calling_functions
 *
 * @example
 * // Set the length of L"/ram/dummy.txt" to 0:
 * wtruncate(L"/ram/dummy.txt", 0);
 */
int wtruncate(const wchar_t *pathname, off_t length);


/**
 * @function unlink
 *
 * @include        "unistd.h"
 *
 * @declaration int
 * unlink(const char *pathname);
 *
 * @brief Remove a link to a file and possibly the file itself.
 *
 * @long Removes a link to a file. The file itself will be removed if the
 *       link is the only hard link that refers to the file.
 *
 * @long The removal of the file will be delayed until the last open file
 *       handle to the file is closed if there are open file handles to the
 *       file.
 *
 * @restriction The level of support for this function is dependent on the
 *              volume manager. JEFF has full support. Many of the other
 *              volume managers do not support deleting a file that has open
 *              file handles.
 *
 * @input pathname The path of the link.
 *
 * @return If successful, the function returns 0, otherwise it returns -1 and
 *         an errno.
 *
 * @seealso link
 * @seealso wunlink
 * @seealso default_restrictions_calling_functions
 */
int unlink(const char *pathname);


/**
 * @function wunlink
 *
 * @include        "unistd.h"
 *
 * @declaration int
 * wunlink(const wchar_t *pathname);
 *
 * @brief Remove a link to a file and possibly the file itself.
 *
 * @long Removes a link to a file. The file itself will be removed if the
 *       link is the only hard link that refers to the file.
 *
 * @long The removal of the file will be delayed until the last open file
 *       handle to the file is closed if there are open file handles to the
 *       file.
 *
 * @restriction The level of support for this function is dependent on the
 *              volume manager. JEFF has full support. Many of the other
 *              volume managers do not support deleting a file that has open
 *              file handles.
 *
 * @input pathname The path of the link.
 *
 * @return If successful, the function returns 0, otherwise it returns -1 and
 *         an errno.
 *
 * @seealso wlink
 * @seealso unlink
 * @seealso default_restrictions_calling_functions
 */
int wunlink(const wchar_t *pathname);


/**
 * @function write
 *
 * @include        "unistd.h"
 *
 * @declaration ssize_t
 * write(int fd, const void *buffer, size_t nbyte);
 *
 * @brief Writes data to a file using a file descriptor.
 *
 * @long The write function call is used to write a specified number of bytes
 *       from an application buffer to a specified already opened file.
 *
 * @long Standards:
 *       Portable Operating System Interface (POSIX.1).
 * @long Conformance:
 *       ISO/IEC 9945-1:1990 (IEEE 1003.1-1990).
 *
 * @input fd     The file descriptor for the output stream to write to.
 * @input buffer Points to the buffer from where the data bytes are fetched.
 * @input nbyte  The number of bytes to write into the stream from buffer.
 *
 * @return The function returns the number of bytes written if successful,
 *         otherwise it returns -1 and errno is set to one of the values below
 *         to indicate the error. write will fail and the file will be
 *         unchanged if one or more of the following conditions are true:
 *
 * @error EAGAIN errno @link "errno.h" "EAGAIN".
 *               The @link "" "O_NONBLOCK" flag is set for the file descriptor
 *               and the client would be delayed in the write operation.
 * @error EBADF  errno @link "errno.h" "EBADF".
 *               The fd argument is not a valid file descriptor.
 * @error EFBIG  errno @link "errno.h" "EFBIG".
 *               An attempt was made to write a file that exceeds an defined
 *               maximum file size.
 * @error EINTR  errno @link "errno.h" "EINTR".
 *               The write operation was interrupted, and either no data was
 *               transferred or the file manager does not report partial file
 *               transfers for this file.
 * @error ENOSPC errno @link "errno.h" "ENOSPC".
 *               There is no free space remaining on the device driver
 *               containing the file.
 * @error EPIPE  errno @link "errno.h" "EPIPE".
 *               An attempt was made to write to a pipe (or FIFO) that is not
 *               open for reading by any client.
 * @error ESTALE errno @link "errno.h" "ESTALE".
 *               fd points to a remote machine and the link to that machine is
 *               no longer active.
 *
 * @seealso open
 * @seealso read
 * @seealso default_restrictions_calling_functions
 *
 * @example
 * // To write 5 bytes from the buffer input to the file ose.txt do the
 * // following:
 *
 * int fd;
 * char *input;
 *
 * fd = open("ose.txt", O_RDWR);
 * num = write(fd, input, 5);
 * if (num == 5)
 * {
 *    // Operation was successful.
 * }
 */
ssize_t write(int fd, const void *buffer, size_t nbyte);


/**
 * @function pipe
 *
 * @include        "unistd.h"
 *
 * @brief Create a pair of file descriptors connected via a pipe.
 *
 * @long Creates a pair of file descriptors connected via a pipe, and places
 * them in the array pointed to by fd. fd[0] is opened for read and fd[1] is
 * opened for writing.
 *
 * @long Any data written to the write-end will be available for reading at the
 * read end.
 *
 * @restriction A pipe file system mounted as /pipe must be available for
 *              this call to succeed.
 *
 * @long Standards: Portable Operating System Interface (POSIX.1).
 * @long Conformance: ISO/IEC 9945-1:1990 (IEEE 1003.1-1990).
 *
 * @output fd Array of two file descriptors.
 *
 * @return The function returns zero on success. Otherwise it returns -1 and
 * errno is set to one of the values below to indicate the error.
 *
 * @error ENOENT errno @link "errno.h" "ENOENT".
 *               The /pipe file system is not available.
 * @error EMFILE errno @link "errno.h" "EMFILE"
 *               The number of file descriptor would exceed
 *               @link "stdio.h" "FOPEN_MAX".
 *
 * @seealso default_restrictions_calling_functions
 */
int pipe(int fd[2]);


/*
 ****************************************************************************
 * BACKWARDS COMPABILETY FUNCTION PROTOTYPES
 ****************************************************************************
 */


/*
 * chmod()/mkdir()/umask() should be in <sys/stat.h>,
 * but for backwards compatibility reasons they are also declared here.
 */
int chmod(const char *pathname, mode_t attrs);
int mkdir(const char *pathname, mode_t attrs);
mode_t umask(mode_t cmask);

/*
 * create()/open() should be in <fcntl.h>, but for backwards compatibility
 * reasons they are also declared here.
 */

int creat(const char *pathname, mode_t attrs);
int open(const char *pathname, int omode, ...);


#ifdef __cplusplus
} /* end of extern "C" */
#endif

#ifdef softOSE
#define ZZMK_INLINE_FUNCTIONS
#include "sfk/sfk_unistd.h"
#endif

/* End #ifndef _UNISTD_H */
#endif
