/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2006 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

#ifndef _CORE_HASH_H
#define _CORE_HASH_H

#include <stdint.h>
#include "core_list.h"


/************************** DEFINES **********************************/



/************************** STRUCTURES **********************************/

/* The element type stored in a hash */
typedef core_list_link_t  core_hash_link_t;

/* Function for generating a hash value given an entry in the hash.
 * Try to make this function return a value that spreads the entries over
 * all the hash table buckets.*/
typedef uint32_t(*core_hash_value_func_t)(const void *entry);

/* Function for comparing two nodes. This
 * function returns 0 if a does not match b, and 1 if a equals b.*/
typedef int(*core_hash_compare_func_t)(const void *a, const void *b);

/* Function callback used when has been referenced by an
 * iterator and was not immediately removed. */
typedef void (*core_hash_link_deleted_func_t)(core_hash_link_t *link);

/* Set of functions used by hash tables */
typedef struct _core_hash_funcs_t
{
   core_hash_compare_func_t       compare;
   core_hash_value_func_t         value;
   core_hash_link_deleted_func_t  free_link;
} core_hash_funcs_t;


/* Structure describing a hash table. */
struct _core_hash_table
{
   /* An array of lists. One for each bucket in the hash. */
   core_list_t       *bucket;

   /* The number of buckets in this hash. */
   uint32_t           no_of_buckets;

   /* The get hash value function to use for this hash. */
   core_hash_funcs_t *funcs;
};

typedef struct _core_hash_table    core_hash_table_t;

/* A hash table iterator, used to iterate over the hash table. */
struct _core_hash_iterator
{
   /* The hash table the iterator traverses */
   core_hash_table_t   *hash;

   /* The compare function used to match entries with the key */
   core_hash_compare_func_t compare;

   /* Key to find */
   const void *key;

   /* The bucket the iterator points to. */
   uint32_t     bucket;

   /* The entry the iterator points to. */
   core_list_iterator_t iterator;
};

typedef struct _core_hash_iterator core_hash_iterator_t;


/*
 *===========================================================================
 *                         core_hash_table_init
 *===========================================================================
 *
 * Description
 * -----------
 * Initializes a hash table with a specific compare function and
 * hash value generator.
 * Needs to be called before any other operation on the hash table.
 *
 * Parameters
 * ----------
 * no_of_buckets  The number of buckets in this hash table.
 *                Statically defined at initialization.
 * funcs          Set of callbacks required by the hash table.
 *
 */
void
core_hash_table_init(core_hash_table_t   *hash,
                     uint32_t             no_of_buckets,
                     core_hash_funcs_t   *funcs);

/*
 *===========================================================================
 *                         core_hash_link_init
 *===========================================================================
 *
 * Description
 * -----------
 * Initializes a hash table entry.
 *
 */
void
core_hash_link_init(core_hash_link_t *hash);
#define core_hash_link_init(lnk) core_list_link_init(lnk)

/*
 *===========================================================================
 *                         core_hash_table_delete
 *===========================================================================
 *
 * Description
 * -----------
 * Deletes a hash table. It does not delete any of the entries in
 * the hash table.
 *
 */
void
core_hash_table_delete(core_hash_table_t *hash);

/*
 *===========================================================================
 *                        core_hash_find
 *===========================================================================
 *
 * Description
 * -----------
 * Searches for a matching entry in the hash table *hash. An initialized
 * iterator must be provided. The iterator will reference the found entry.
 *
 * Iterators must always be released when the reference to the entry is
 * no longer needed. 
 *
 * Returns
 * -------
 * struct core_hashlink_t* Returns the found entry or NULL if it is not found.
 *
 */
core_list_link_t *
core_hash_find(core_hash_table_t    *hash, 
               core_hash_iterator_t *iterator,
               const void *key);

/*
 *===========================================================================
 *                        core_hash_entry
 *===========================================================================
 *
 * Description
 * -----------
 * Returns the entry an iterator points to. 
 *
 */
int
core_hash_entry(core_hash_iterator_t *it);
#define core_hash_entry(it) ((it)->iterator.current)

/*
 *===========================================================================
 *                        core_hash_insert
 *===========================================================================
 *
 * Description
 * -----------
 * Inserts an entry in the hash_table *hash unless it is already there.
 *
 * Returns
 * -------
 * int      Returns one if the entry was inserted,
 *          or zero is it is already in a hash.
 *
 */
int
core_hash_insert(core_hash_table_t *hash, core_hash_link_t *entry);

/*
 *===========================================================================
 *                        core_hash_delete
 *===========================================================================
 *
 * Description
 * -----------
 * Deletes an entry in the hash table *hash pointed to by an iterator.
 *
 */
void
core_hash_delete(core_hash_iterator_t *iterator);

/*
 *===========================================================================
 *                        core_hash_delete_direct
 *===========================================================================
 *
 * Description
 * -----------
 * Deletes an entry in the hash table immediately, ignoring the reference
 * counter of the entry. Use with caution. 
 *
 */
void
core_hash_delete_direct(core_hash_table_t *hash, core_hash_link_t *entry);

/*
 *===========================================================================
 *                        core_hash_iterator_init
 *===========================================================================
 *
 * Description
 * -----------
 * Initializes a hash table iterator used to reference entries. 
 *
 * The iterator must always be released when it is no longer needed. Failure
 * to do so could result in hash entries being incorrectly referenced and
 * impossible to remove.
 *
 */
void
core_hash_iterator_init(core_hash_iterator_t *iterator,
                        core_hash_compare_func_t compare);

/*
 *===========================================================================
 *                        core_hash_reference
 *===========================================================================
 *
 * Description
 * -----------
 * Used to reference an explicit entry in the hash when only a pointer to
 * the object is available. Useful for situations where a node exists in
 * multiple storage containers and it would be inefficient to search for
 * the node in each container to get hold of an iterator. 
 */
void
core_hash_reference(core_hash_table_t *hash,
                    core_hash_iterator_t *iterator,
                    core_hash_link_t *link);

/*
 *===========================================================================
 *                        core_hash_first
 *===========================================================================
 *
 * Description
 * -----------
 * Retrieves the first entry in the hash that is considered equal to the key
 * by the compare function implemented by the chosen iterator. If the
 * iterator was initialized without compare function, the first entry in the
 * hash is returned.
 * 
 * Returns
 * -------
 * struct core_hash_link_t* Returns the found entry or NULL if it is not found.
 *
 */
core_hash_link_t *
core_hash_first(core_hash_table_t    *hash, 
                core_hash_iterator_t *iterator,
                const void *key);

/*
 *===========================================================================
 *                        core_hash_next
 *===========================================================================
 *
 * Description
 * -----------
 * Iterates over the hash by advancing the iterator. The next entry chosen
 * is considered equal by the compare function implemented by the iterator.
 * If the iterator was initialized without compare function, the next entry
 * in the hash is returned.
 *
 * If the entry currently pointed to has been flagged for deletion and no
 * longer is referenced by any iterator, it is removed permanently.
 *
 * Returns
 * -------
 * struct core_hash_link_t* Returns the found entry or NULL if it is not found.
 *
 */
core_hash_link_t *
core_hash_next(core_hash_iterator_t *iterator);

/*
 *===========================================================================
 *                        core_hash_iterator_release
 *===========================================================================
 *
 * Description
 * -----------
 * Invalidates an iterator. The iterator dereferences the entry currently
 * pointed to and removes it permanently if it has been flagged for deletion.
 *
 * This function MUST ALWAYS be called when an iterator is no longer needed.
 *
 */
void
core_hash_iterator_release(core_hash_iterator_t *iterator);

/*
 *===========================================================================
 *                        core_hash_reserve
 *===========================================================================
 *
 * Description
 * -----------
 * Same functionality as core_list_reserve. 
 */
int
core_hash_reserve(core_hash_iterator_t *it, int enable);

#endif /* _CORE_HASH_H */
