/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2006 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */
/*
 * OSE doubly linked list package.
 */

#ifndef _CORE_LIST_H
#define _CORE_LIST_H

#include <stddef.h>
#ifdef __cplusplus
extern "C" {
#endif

#include <stdint.h>
#ifdef STANDALONE
#include "standalone.h"
#else
#include "spinlock.h"
#endif

/* Option passed to new iterators to configure reverse order traversion */
#define CORE_LIST_ITERATE_REVERSE       0x00000001

struct _core_list_funcs_t;

typedef struct _core_list_link_t
{
   struct _core_list_link_t *next;	/* Ptr to next element in list.     */
   struct _core_list_link_t *prev;	/* Ptr to previous element in list. */
   uint32_t                  no_refs;   /* Number of references to the node */
} core_list_link_t;

typedef struct _core_list_t
{
   struct _core_list_link_t  *next;	/* Ptr to first element in list.    */
   struct _core_list_link_t  *prev;	/* Ptr to last element in list.     */
   struct _core_list_funcs_t *funcs;
   spinlock_t lock;
} core_list_t;

typedef struct _core_list_iterator_t
{
   struct _core_list_t      *list;	/* Ptr to the list.                 */
   struct _core_list_link_t *current;	/* Ptr to current element in list.  */
   uint32_t                  options;	/* List is traversed backwards if 1   */
} core_list_iterator_t;

typedef int   (*core_list_insert_ordered_func_t)(const core_list_link_t *,
                                                 const core_list_link_t *);
typedef void *(*core_list_traverse_func_t)(core_list_link_t *lnk, void *arg);
typedef void  (*core_list_assert_func_t)(const core_list_link_t *lnk);
typedef void  (*core_list_link_deleted_t)(core_list_link_t *link);

typedef struct _core_list_funcs_t
{
   core_list_link_deleted_t  free_link;
} core_list_funcs_t;


/*
 *===========================================================================
 *                         init_list
 *===========================================================================
 *
 * Description
 * -----------
 * Initializes a doubly linked list.
 * Needs to be called before any other operation on the list.
 *
 */
void core_list_init(core_list_t *list, core_list_funcs_t *funcs);

/*
 *===========================================================================
 *                         list_init_link
 *===========================================================================
 *
 * Description
 * -----------
 * Initializes a list entry.
 * Needs to be called before the entry is inserted into a list.
 *
 */
void core_list_link_init(core_list_link_t *link);

/*
 *===========================================================================
 *                         core_list_delete
 *===========================================================================
 *
 * Description
 * -----------
 * Removes the specified entry from the list it is in.
 *
 */
void core_list_delete(core_list_iterator_t *link_to_remove);

/*
 *===========================================================================
 *                         core_list_delete_direct
 *===========================================================================
 *
 * Description
 * -----------
 * Removes the specified entry from the list it is in.
 *
 */
void core_list_delete_direct(core_list_t *list, core_list_link_t *link_to_remove);

/*
 *===========================================================================
 *                         core_list_delete_no_lock
 *===========================================================================
 *
 * Description
 * -----------
 * Removes the specified entry from the list it is in without taking a
 * spin lock or disabling interrupts.
 *
 */
void
core_list_delete_no_lock(core_list_iterator_t *link_to_remove);

/*
 *===========================================================================
 *                         core_list_reserve
 *===========================================================================
 *
 * Description
 * -----------
 * Used by a client to reserve the node for exclusive use, for example when
 * the client intends to remove the node from the list. No other iterator
 * will traverse a reserved node. The reservation will fail and 0 is returned
 * if more than one client is referencing the node.
 */
int
core_list_reserve(core_list_iterator_t *link_to_reserve, uint32_t enable);

/*
 *===========================================================================
 *                         core_list_insert_head
 *===========================================================================
 *
 * Description
 * -----------
 * Inserts the entry first into the list pointed to by 'list'.
 *
 */
void
core_list_insert_head(core_list_t *list, core_list_link_t *link_to_insert);

/*
 *===========================================================================
 *                         insert_list_head_no_lock
 *===========================================================================
 *
 * Description
 * -----------
 * Inserts the entry first into the list pointed to by 'list' without
 * any spin lock or disabling interrupts.
 *
 */
void
core_list_insert_head_no_lock(core_list_t *list, core_list_link_t *link_to_insert);

/*
 *===========================================================================
 *                         core_list_insert_tail
 *===========================================================================
 *
 * Description
 * -----------
 * Inserts the entry last into the list pointed to by 'list'.
 *
 */
void core_list_insert_tail(core_list_t *list, core_list_link_t *link_to_insert);

/*
 *===========================================================================
 *                         core_list_insert_tail_no_lock
 *===========================================================================
 *
 * Description
 * -----------
 * Inserts the entry last into the list pointed to by 'list' without
 * aquiring any spin lock or disabling interrupts.
 *
 */
void
core_list_insert_tail_no_lock(core_list_t *list, core_list_link_t *link_to_insert);

/*
 *===========================================================================
 *                         core_list_insert_after
 *===========================================================================
 *
 * Description
 * -----------
 * Inserts the entry link_to_insert after the entry link_in_list.
 *
 */
void core_list_insert_after(core_list_t *list,
                            core_list_iterator_t *link_in_list,
                            core_list_link_t *link_to_insert);

/*
 *===========================================================================
 *                         core_list_insert_after_no_lock
 *===========================================================================
 *
 * Description
 * -----------
 * Inserts the entry link_to_insert after the entry link_in_list
 * without aquiring any spin lock or disabling interrupts.
 *
 */
void
core_list_insert_after_no_lock(core_list_iterator_t *link_in_list,
                               core_list_link_t *link_to_insert);

/*
 *===========================================================================
 *                         core_list_insert_before
 *===========================================================================
 *
 * Description
 * -----------
 * Inserts the entry link_to_insert before the entry link_in_list.
 *
 */
void core_list_insert_before(core_list_t *list,
                             core_list_iterator_t *link_in_list,
                             core_list_link_t *link_to_insert);

/*
 *===========================================================================
 *                         insert_list_before_no_lock
 *===========================================================================
 *
 * Description
 * -----------
 * Inserts the entry link_to_insert before the entry link_in_list
 * without aquiring any spin lock or disabling interrupts.
 *
 */
void
core_list_insert_before_no_lock(core_list_iterator_t *link_in_list,
                                core_list_link_t *link_to_insert);

/*
 *===========================================================================
 *                         core_list_head
 *===========================================================================
 *
 * Description
 * -----------
 * Returns the first entry in the list.
 *
 */
core_list_link_t *
core_list_head(core_list_t *list, core_list_iterator_t *iterator);

/*
 *===========================================================================
 *                         core_list_tail
 *===========================================================================
 *
 * Description
 * -----------
 * Returns the last entry in the list.
 *
 */
core_list_link_t *
core_list_tail(core_list_t *list, core_list_iterator_t *iterator);

/*
 *===========================================================================
 *                         length_of_list
 *===========================================================================
 *
 * Description
 * -----------
 * Returns the number of entries in the list.
 *
 */
uint32_t
core_list_length(core_list_t *list);

/*
 *===========================================================================
 *                         core_list_insert_ordered
 *===========================================================================
 *
 * Description
 * -----------
 * Inserts the entry ll before the entry for which the compare function is
 * positive.
 *
 */
void
core_list_insert_ordered(core_list_t *list,
                         core_list_link_t *link_to_insert,
                         core_list_insert_ordered_func_t compare);

/*
 *===========================================================================
 *                         core_list_traverse
 *===========================================================================
 *
 * Description
 * -----------
 * Traverses the list and calls the apply function for each entry.
 *
 */
void *
core_list_traverse(core_list_t *list, core_list_traverse_func_t apply, void *arg);

/*
 *===========================================================================
 *                         core_list_iterator_init
 *===========================================================================
 *
 * Description
 * -----------
 * Initializes a list iterator, pointing to the node passed as argument.
 *
 * A node pointed to by an iterator is guaranteed to exist as long it is
 * referenced by any iterator. Deleted nodes will not be removed from the
 * list until all references to the node have disappeared. New iterators will
 * not discover deleted nodes which have not yet been removed. 
 *
 * The iterator must be released, by calling core_list_iterator_release,
 * when it is no longer needed. The third parameter specifies if the
 * iterator traverses the list forwards or backwards.
 *
 * Iterators MUST NOT be shared between execution primitives.
 */
void
core_list_iterator_init(core_list_iterator_t *iterator, uint32_t options);

/*
 *===========================================================================
 *                        core_list_reference
 *===========================================================================
 *
 * Description
 * -----------
 * Used to reference an explicit entry in the hash when only a pointer to
 * the object is available. Useful for situations where a node exists in
 * multiple storage containers and it would be inefficient to search for
 * the node in each container to get hold of an iterator. 
 */
void
core_list_reference(core_list_t *list,
                    core_list_iterator_t *iterator,
                    core_list_link_t *link);

/*
 *===========================================================================
 *                        core_list_reserved_delete
 *===========================================================================
 *
 * Description
 * -----------
 * This function only deletes a node if it is referenced by a single client.
 * If the reference count is higher than one the call aborts and no delayed
 * removal will take place. 
 *
 * Returns
 * -------
 * Non-zero on success.
 */
int
core_list_reserved_delete(core_list_iterator_t *iterator);

/*
 *===========================================================================
 *                         core_list_iterator_node
 *===========================================================================
 *
 * Description
 * -----------
 * Used to retrieve the node currently pointed to by the iterator. It is not
 * legal to reference the node without the iterator object. 
 */
core_list_link_t *
core_list_iterator_node(core_list_iterator_t *list);

/*
 *===========================================================================
 *                         core_list_next
 *===========================================================================
 *
 * Description
 * -----------
 * Advances the iterator to the next node in the list. The node is returned
 * unless end-of-list is reached (NULL). Using iterator_next on reverse-order
 * iterators is not legal and NULL is returned.
 */
core_list_link_t *
core_list_next(core_list_iterator_t *list);

/*
 *===========================================================================
 *                         core_list_prev
 *===========================================================================
 *
 * Description
 * -----------
 * Advances the iterator to the previous node in the list. The node is returned
 * unless end-of-list is reached (NULL). Using iterator_prev on in-order
 * iterators is not legal and NULL is returned.
 */
core_list_link_t *
core_list_prev(core_list_iterator_t *list);

/*
 *===========================================================================
 *                         core_list_iterator_release
 *===========================================================================
 *
 * Description
 * -----------
 * Destroys an iterator. This function must always be called when an iterator
 * is no longer needed. If the iterator references a deleted node, it will
 * remove the node from the list if it is the last refering iterator.
 */
void
core_list_iterator_release(core_list_iterator_t *iterator);


/*
 *===========================================================================
 *                         core_list_assert
 *===========================================================================
 *
 * Description
 * -----------
 * Checks that each entry in the list is intact and that the compare function
 * is valid for each entry.
 *
 */
#ifdef ASSERT
void
core_list_assert(const core_list_t *list, core_list_assert_func_t check);
#endif


#ifdef __cplusplus
}
#endif

#endif /* _CORE_LIST_H */

