/* COPYRIGHT-ENEA-SRC-R2 *
**************************************************************************
* Copyright (C) 2004-2008 by Enea Software AB.
* All rights reserved.
*
* This Software is furnished under a software license agreement and
* may be used only in accordance with the terms of such agreement.
* Any other use or reproduction is prohibited. No title to and
* ownership of the Software is hereby transferred.
*
* PROPRIETARY NOTICE
* This Software consists of confidential information.
* Trade secret law and copyright law protect this Software.
* The above notice of copyright on this Software does not indicate
* any actual or intended publication of such Software.
**************************************************************************
* COPYRIGHT-END */


#ifndef _OSE_RBTREE_LOCKED_H
#define _OSE_RBTREE_LOCKED_H

#include "cpu.h"
#include "ose_spi/spinlock.h"

/**
 * @type rbtree_compare_func_t
 *
 * @brief Function prototype for comparing elements added to a tree.
 */
typedef int (*rbtree_compare_func_t)(const void *a, const void *b);

/**
 * @type rbtree_compare_best_func_t
 *
 * @brief Function prototype for comparing elements added to a tree.
 *
 * @long  The third argument <best> is the best matching node found so far.
 *        The compare function callback can use the pointer when investigating
 *        other nodes in the tree.
 *
 * @long  The function should return
 */
typedef int
(*rbtree_compare_best_func_t)(const void *a, const void *b, const void *best);


/**
 * @type core_rbtree_node_t
 *
 * @brief Representation of a node in a Red-Black Tree.
 *
 * @long Should be instantiated in the data structure held
 *       by the tree.
 *
 * @field parent The node's parent.
 *
 * @field left   The node's left child.
 *
 * @field right  The node's right child.
 */
typedef struct core_rbtree_node_t
{
      struct core_rbtree_node_t  *parent;	/* parent */
      struct core_rbtree_node_t  *left;		/* left child */
      struct core_rbtree_node_t  *right;	/* right child */
} core_rbtree_node_t;


/**
 * @type core_rbtree_t
 *
 * @brief Representation of a Red-Black tree.
 *
 * @field root Root of Red-Black tree.
 *
 * @field nil Representation of a NIL node (bottom-most node).
 *
 * @field compare Callback function for comparing to nodes (keys).
 *
 * @field lock Spinlock protecting concurrent accesses.
 *
 */
typedef struct core_rbtree_t
{
      core_rbtree_node_t  *root;
      core_rbtree_node_t   nil;
      rbtree_compare_func_t compare;
      spinlock_t lock;
} core_rbtree_t;


/**
 * @type rbtree_traverse_func_t
 *
 * @brief Function prototype called for each node during
 *        a traversation of the tree.
 */
typedef void (*rbtree_traverse_func_t)(const void *tree, void *node);


/**
 * @function core_rbtree_init
 *
 * @brief Initializes a new tree.
 *
 * @input tree Pointer to the tree to initialize.
 * @input compare Function used by the tree to compare nodes during insert/remove.
 */
void core_rbtree_init(core_rbtree_t *tree, rbtree_compare_func_t compare);


/**
 * @function core_rbtree_insert
 *
 * @brief Inserts a new node into the Red-Back tree.
 *
 * @input tree Tree to insert the new node into.
 * @input node The new node to insert.
 *
 * @return 0 if the insertion succeeded, -1 otherwise (duplicate node found).
 */
int core_rbtree_insert(core_rbtree_t *tree, void *node);


/**
 * @function core_rbtree_remove
 *
 * @brief Removes a node from a tree.
 *
 * @input tree Tree to remove a node from.
 * @input node The node to remove from the tree.
 * 
 * @return 0 if the operation succeeded, -1 otherwise (node not
 * found).
 */
core_rbtree_node_t *core_rbtree_remove(core_rbtree_t *tree, void *node);


/**
 * @function core_rbtree_find
 *
 * @brief Find a node in the tree given a key.
 *
 * @input tree The tree to search.
 *
 * @input compare Function to use to compare nodes during the search.
 *
 * @input key A key identifying the node to find. 
 *
 * @return Pointer to the node if found, NULL otherwise.
 */
core_rbtree_node_t *core_rbtree_find(core_rbtree_t *tree,
                                     rbtree_compare_func_t compare,
                                     const void *key);

/**
 * @function core_rbtree_find_best
 *
 * @brief Find a node in the tree given a key.
 *
 * @long  Use this function to find nodes in the tree that nearly matches
 *        the requested key.
 *
 * @long  The callback function should return 0 if the key perfectly
 *        matches the node being investigated. The search is then
 *        terminated and the node returned. 
 *
 * @long  The callback function must return -1 (no match, go left) or
 *        1 (no match, go right) if the current node does not match
 *        the search criteria.
 *
 * @long  The callback function should return -2
 *        (match, but continue left) or 2 (match, but continue right)
 *        if a node matches the criteria but there still might be better
 *        nodes in the left or right subtree. 
 *
 * @long  A return value of either -2 or 2 from the callback is 
 *        considered to be a request to select a new best node.
 *        The callback can compare the current node to the currently
 *        selected best node, passed as the third argument,
 *        before deciding to select a new best node. 
 *
 * @input tree The tree to search.
 *
 * @input compare Function to use to compare nodes during the search.
 *
 * @input key A key identifying the node to find. 
 *
 * @return Pointer to the selected best node, NULL otherwise.
 *
 * @seealso rbtree_compare_best_func_t
 */
core_rbtree_node_t *
core_rbtree_find_best(core_rbtree_t *tree,
                      rbtree_compare_best_func_t compare,
                      const void *key);


/**
 * @function core_rbtree_first
 *
 * @brief Return a pointer to the first (smallest) node in the tree.
 *
 * @input tree Pointer to a Red-Black tree.
 *
 * @return Pointer to a tree node or NULL if the tree is empty.
 */
core_rbtree_node_t *core_rbtree_first(core_rbtree_t *tree);


/**
 * @function core_rbtree_next
 *
 * @brief Return the successor for a given node in a Red-Black tree.
 *
 * @input tree Pointer to a Red-Black tree.
 *
 * @input node The node for which the successor is requested.
 *
 * @return Pointer to the node that is the successor to the given node, or
 *         NULL if end of tree has been reached.
 */
core_rbtree_node_t *core_rbtree_next(core_rbtree_t *tree, void *node);


/**
 * @function core_rbtree_last
 *
 * @brief Return a pointer to the last (largest) node in the tree.
 *
 * @input tree Pointer to a Red-Black tree.
 *
 * @return Pointer to a tree node or NULL if the tree is empty.
 */
core_rbtree_node_t *core_rbtree_last(core_rbtree_t *tree);


/**
 * @function core_rbtree_prev
 *
 * @brief Return the predeccessor for a given node in a Red-Black tree.
 *
 * @input tree Pointer to a Red-Black tree.
 *
 * @input node The node for which the predeccessor is requested.
 *
 * @return Pointer to the node that is the predeccessor to the given node, or
 *         NULL if end of tree has been reached.
 */
core_rbtree_node_t *core_rbtree_prev(core_rbtree_t *tree, void *node);


/**
 * @function core_rbtree_traverse
 *
 * @brief Traverses the tree in order from smallest node to largest
 *        and executes a user defined function on each node.
 *
 * @input tree Pointer to a Red-Black tree.
 *
 * @input func User defined function executed for each node.
 */
void core_rbtree_traverse(core_rbtree_t *tree, rbtree_traverse_func_t func);


/**
 * @function core_rbtree_insert_no_lock
 *
 * @brief Inserts a new node into the Red-Back tree.
 *
 * @input tree Tree to insert the new node into.
 * @input node The new node to insert.
 *
 * @return 0 if the insertion succeeded, -1 otherwise (duplicate node found).
 */
int core_rbtree_insert_no_lock(core_rbtree_t *tree, void *node);


/**
 * @function core_rbtree_remove_no_lock
 *
 * @brief Removes a node from a tree.
 *
 * @input tree Tree to remove a node from.
 * @input node The node to remove from the tree.
 * 
 * @return 0 if the operation succeeded, -1 otherwise (node not
 * found).
 */
core_rbtree_node_t *core_rbtree_remove_no_lock(core_rbtree_t *tree, void *node);


/**
 * @function core_rbtree_find_no_lock
 *
 * @brief Find a node in the tree given a key.
 *
 * @input tree The tree to search.
 *
 * @input compare Function to use to compare nodes during the search.
 *
 * @input key A key identifying the node to find. 
 *
 * @return Pointer to the node if found, NULL otherwise.
 */
core_rbtree_node_t *
core_rbtree_find_no_lock(core_rbtree_t *tree,
                         rbtree_compare_func_t compare,
                         const void *key);

/**
 * @function core_rbtree_find_best_no_lock
 *
 * @brief Find a node in the tree given a key.
 *
 * @long  Use this function to find nodes in the tree that nearly matches
 *        the requested key.
 *
 * @long  The callback function should return 0 if the key perfectly
 *        matches the node being investigated. The search is then
 *        terminated and the node returned. 
 *
 * @long  The callback function must return -1 (no match, go left) or
 *        1 (no match, go right) if the current node does not match
 *        the search criteria.
 *
 * @long  The callback function should return -2
 *        (match, but continue left) or 2 (match, but continue right)
 *        if a node matches the criteria but there still might be better
 *        nodes in the left or right subtree. 
 *
 * @long  A return value of either -2 or 2 from the callback is 
 *        considered to be a request to select a new best node.
 *        The callback can compare the current node to the currently
 *        selected best node, passed as the third argument,
 *        before deciding to select a new best node.
 *        
 * @input tree The tree to search.
 *
 * @input compare Function to use to compare nodes during the search.
 *
 * @input key A key identifying the node to find. 
 *
 * @return Pointer to the best node if found, NULL otherwise.
 */
core_rbtree_node_t *
core_rbtree_find_best_no_lock(core_rbtree_t *tree,
                              rbtree_compare_best_func_t compare,
                              const void *key);


/**
 * @function core_rbtree_first_no_lock
 *
 * @brief Return a pointer to the first (smallest) node in the tree.
 *
 * @input tree Pointer to a Red-Black tree.
 *
 * @return Pointer to a tree node or NULL if the tree is empty.
 */
core_rbtree_node_t *core_rbtree_first_no_lock(core_rbtree_t *tree);


/**
 * @function core_rbtree_next_no_lock
 *
 * @brief Return the successor for a given node in a Red-Black tree.
 *
 * @input tree Pointer to a Red-Black tree.
 *
 * @input node The node for which the successor is requested.
 *
 * @return Pointer to the node that is the successor to the given node, or
 *         NULL if end of tree has been reached.
 */
core_rbtree_node_t *core_rbtree_next_no_lock(core_rbtree_t *tree,
                                             void *node);

/**
 * @function core_rbtree_last_no_lock
 *
 * @brief Return a pointer to the last (largest) node in the tree.
 *
 * @input tree Pointer to a Red-Black tree.
 *
 * @return Pointer to a tree node or NULL if the tree is empty.
 */
core_rbtree_node_t *core_rbtree_last_no_lock(core_rbtree_t *tree);


/**
 * @function core_rbtree_prev_no_lock
 *
 * @brief Return the predeccessor for a given node in a Red-Black tree.
 *
 * @input tree Pointer to a Red-Black tree.
 *
 * @input node The node for which the predeccessor is requested.
 *
 * @return Pointer to the node that is the predeccessor to the given node, or
 *         NULL if end of tree has been reached.
 */
core_rbtree_node_t *core_rbtree_prev_no_lock(core_rbtree_t *tree, void *node);


/**
 * @function core_rbtree_traverse_no_lock
 *
 * @brief Traverses the tree in order from smallest node to largest
 *        and executes a user defined function on each node.
 *
 * @input tree Pointer to a Red-Black tree.
 *
 * @input func User defined function executed for each node.
 */
void core_rbtree_traverse_no_lock(core_rbtree_t *tree,
                                  rbtree_traverse_func_t func);


#endif /* _OSE_RBTREE_LOCKED_H */

