/* COPYRIGHT-ENEA-SRC-R2 *
**************************************************************************
* Copyright (C) 2004-2006 by Enea Software AB.
* All rights reserved.
*
* This Software is furnished under a software license agreement and
* may be used only in accordance with the terms of such agreement.
* Any other use or reproduction is prohibited. No title to and
* ownership of the Software is hereby transferred.
*
* PROPRIETARY NOTICE
* This Software consists of confidential information.
* Trade secret law and copyright law protect this Software.
* The above notice of copyright on this Software does not indicate
* any actual or intended publication of such Software.
**************************************************************************
* COPYRIGHT-END */


#ifndef _CORE_SEGTREEE_H
#define _CORE_SEGTREEE_H

#include "core_rbtree.h"

#include "ose.h"

#include <stdint.h>
#include <stdlib.h>


/*
 *===========================================================================
 *                         Segment tree
 *===========================================================================
 *
 * A segment tree is a balanced binary tree for interval bookkeeping. Each
 * interval added to the tree is split or merged into a global, reference
 * counted, representation. Example:
 *
 * User adds the intervals A and B to the tree:
 *
 * A:  |-------|
 * B         |----------|
 *
 * User then removes the interval C:
 *
 * C              |--|
 *
 * The global representation afterwards, with reference counters, will be:
 *
 *     |-----|-|--|  |--|
 *        1   2  1     1
 *
 * The tree is useful for implementations where the global representation
 * is important. The tree will trigger events when the global map changes,
 * i.e. the outside world will notice these changes only:
 *
 *     |-------|            -> first change  (add)
 *
 *             |--------|   -> second change (add)
 *
 *                |--|      -> third change  (remove)
 *
 */


/*
 *===========================================================================
 *                         core_segtree_node_t
 *===========================================================================
 *
 * Description
 * -----------
 * The node structure. Should be instantiated in user defined structures
 * added to the tree. 
 */
typedef struct _core_segtree_node_t
{
   core_rbtree_node_t node;
   uint64_t           start;
   uint64_t           end;
   uint32_t           ref_count;
} core_segtree_node_t;


/*
 *===========================================================================
 *                         Callback function prototypes
 *===========================================================================
 *
 * Description
 * -----------
 *
 */
struct _core_segtree_t;
typedef int (*core_segtree_compare_t)(const void *left, const void *right);
typedef void* (*core_segtree_alloc_t)(size_t size);
typedef int (*core_segtree_free_t)(void *ptr);
typedef int (*core_segtree_ctor_t)(struct _core_segtree_t *tree,
                                   core_segtree_node_t *template,
                                   core_segtree_node_t *new_node, uint32_t create);
typedef int (*core_segtree_dtor_t)(struct _core_segtree_t *tree,
                                   core_segtree_node_t *template,
                                   core_segtree_node_t *old_node, uint32_t destroy);



/*
 *===========================================================================
 *                         core_segtree_funcs_t
 *===========================================================================
 *
 * Description
 * -----------
 * A package of callback functions installed in a segtree. They are used
 * to manage the tree in a user defined manner. 
 *
 * Compare
 * -------
 * Used to order the nodes internally and balance the tree. Same usage
 * as with red-black trees.
 *
 * Add
 * ----
 * Called when a new segment is added to the tree. It can check custom 
 * node metadata not understood by the generic tree implementation, and
 * block new segments if for example, metadata is incompatible with existing
 * segments in the tree.
 * 
 * Alloc
 * -----
 * Called by the tree when it needs to allocate a new internal node object.
 * Allocations can occur both when adding new segments and removing existing
 * segments (due to splits). 
 *
 * Free
 * ----
 * Called when an internal node is no longer needed (due to remove or merge).
 *
 * Constructor
 * -----------
 * Used when a new internal node is created and allows the user to copy
 * custom metadata from a template into the new node. If a node is split,
 * the template is the original. If a new node is added, the template is
 * the object passed to core_segtree_add. If the parameter 'create' is set,
 * the global interval representation has changed. If not set, an existing
 * node has been split and the global representation remains. The user
 * should should always copy custom metadata to the new node.
 *
 * Destructor
 * ----------
 * Called when a segment is removed and allows the user to
 * take custom action. If the parameter 'destroy' is set,
 * the global interval representation has changed. If not set, an existing
 * node has been merged and the global representation remains.   
 */
typedef struct _core_segtree_funcs_t
{
   core_segtree_compare_t compare;
   core_segtree_compare_t add;
   core_segtree_alloc_t   alloc_node;
   core_segtree_free_t    free_node;
   core_segtree_ctor_t    ctor;
   core_segtree_dtor_t    dtor;
} core_segtree_funcs_t;

/*
 *===========================================================================
 *                         core_segtree_t
 *===========================================================================
 *
 * Description
 * -----------
 * The tree structure.
 *
 */
typedef struct _core_segtree_t
{
   core_rbtree_t          tree;
   core_segtree_funcs_t  *funcs;
   uint32_t               node_size;
} core_segtree_t;






/*
 *===========================================================================
 *                         core_segtree_init
 *===========================================================================
 *
 * Description
 * -----------
 * Initializes a segment tree.
 *
 * Parameters
  * ----------
 * tree        - Reference to the tree.
 * functions   - At least 'alloc' and 'free' must be provided.  
 *
 * Returns
 * -------
 * 0 on success, or negative value if an obligatory function is missing. 
 */
int
core_segtree_init(core_segtree_t *tree,
                  core_segtree_funcs_t *functions,
                  uint32_t node_size);


/*
 *===========================================================================
 *                         core_segtree_node_init
 *===========================================================================
 *
 * Description
 * -----------
 * Initializes a segment tree template node. Template nodes are used when
 * adding or removing intervals in the tree. 
 *
 * Parameters
 * ----------
 * tree        - Reference to the tree.
 * node        - Reference to template node allocated by the user. 
 * start       - Start of interval.
 * end         - End of interval. 
 */
void
core_segtree_node_init(core_segtree_t *tree, core_segtree_node_t *node,
                       uint64_t start, uint64_t end);


/*
 *===========================================================================
 *                         core_segtree_find
 *===========================================================================
 *
 * Description
 * -----------
 * Locates a segment node within the tree. Note that a 'segment node' may
 * differ from a 'segment'. A segment may consist of several nodes since
 * all intervals are reference counted. To find all nodes, the user will
 * have to iterate over the requested range.
 *
 * Parameters
 * ----------
 * tree        - Reference to the tree.
 * find_func   - function to compare the key with nodes in the tree.
 * key         - something to look for
 *
 * Returns
 * -------
 * A node matching the key, or NULL if none exist.
 */
core_segtree_node_t *
core_segtree_find(core_segtree_t *tree,
                  core_segtree_compare_t find_func,
                  void *key);


/*
 *===========================================================================
 *                         core_segtree_add
 *===========================================================================
 *
 * Description
 * -----------
 * Add a new interval to the tree. A user can provide an add-callback
 * where overlapping incompatible intervals in the tree can be detected.
 * If so, the callback may abort the function call.
 *
 * Incompatible ranges should be rare, but can exist if the user keeps
 * custom metadata in each node, such as a permission flag. Intervals
 * with different permissions overlapping each other would be illegal, while
 * intervals with equal permission are allowed to overlap.
 *
 * If the global interval representation changes, callback are called to
 * inform the outside world of the changes (ctor, dtor).
 *
 * Paramaters
 * ----------
 * tree        - Reference to the tree.
 * node        - A temporary template node used while adding the new
 *               interval. The tree will not insert the template node into
 *               the tree. It will always allocate and initialize its own
 *               nodes. The 'node' should therefore be deleted afterwards.
 *               The template node must be initialized with
 *               core_segtree_node_init, as well as custom user metadata. 
 *
 * Returns
 * -------
 * 0 on success, negative on failure (errno).
 */
int
core_segtree_add(core_segtree_t *tree, core_segtree_node_t *node);


/*
 *===========================================================================
 *                         core_segtree_remove
 *===========================================================================
 *
 * Description
 * -----------
 * Removes an interval from the tree. The tree will decrease the reference
 * counter of all nodes within the removed interval that are considered
 * equal to the template 'node'.
 *
 * The function will fail of any part of the interval is not present. 
 * 
 * If the global interval representation changes, callback are called to
 * inform the outside world of the changes (ctor, dtor).
 *
 * Parameters
 * ----------
 * tree        - Reference to the tree.
 * node        - A temporary template node used while removing the 
 *               interval. The 'node' should be deleted afterwards.
 *               The template node must be initialized with
 *               core_segtree_node_init, as well as custom user metadata. 
 *
 * Returns
 * -------
 * 0 on success, negative on failure (errno).
 */
int
core_segtree_remove(core_segtree_t *tree, core_segtree_node_t *node);


/*
 *===========================================================================
 *                         core_segtree_empty
 *===========================================================================
 *
 * Description
 * -----------
 * Determines if the tree is empty or not.
 *
 * Returns
 * -------
 * Non-zero if empty.
 */
int
core_segtree_empty(core_segtree_t *tree);

#endif /* _CORE_SEGTREEE_H */
