/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2006 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

#ifndef OSE_EPM_PRIVATE_H_
#define OSE_EPM_PRIVATE_H_

#include "epm.h"

#ifdef __cplusplus
extern "C" {
#endif

/* Writes all execution points to the logical page page_address using the
 * supplied write_ptr. The consistency of the caches has to be taken care
 * by the caller. */
EpmStatus
epm_write_execution_points(OSADDRESS page_address, char *write_ptr);

/* Returns one if there are any execution points on this page. */
int
epm_execution_points_on_page(OSADDRESS page_address);

/**
 *  @function ose_epm_alloc_fast_execution_point
 *  @brief Allocate memory area that is used by a fast execution point.
 *  @long This function allocates an code memory area for an fast ep 
 *  this memory area must be filled with code that is called 
 *  to evaluate if the callback should be called. And this has to be done
 *  before ose_epm_add_fast_execution_point() is called on the ep.
 *  There are some CPU architectual limitation that needs to be fulfilled
 *  to make this possible like for PowerPC there needs to be MM allocatable
 *  executable memory near enough to make a jump in one assembler instructions.
 *  Some architectures might not support placing a fast execution point on every
 *  assembler instructions. If this functions returns NULL a normal execution 
 *  point can be used. The evaluate code that need to be called are executed in 
 *  the same context as the code executing the code at that time. Context might
 *  be checked but it is not mandatory and should not be relied on.
 *  @long This function must only be called once on each fast execution point.
 *  @long This function is only available in the core image.
 *  It must be called in supervisor mode.
 *  @input ep Pointer to a struct containing information about the
 *  execution point. This struct must be preserved thru the lifetime of the
 *  execution point. It is used internally by the EPM.
 *  @input code_size size of code that is allocated this area is 
 *                  called to evaluate if the ep should call the callback.
 *  @return Pointer to memory area or NULL memory area is not available or
 *          something fails. 
 *  @seealso ose_epm_free_fast_execution_point 
 *  @seealso ose_epm_enable_fast_execution_point
 *  @seealso ose_epm_disable_fast_execution_point
**/
void *
ose_epm_alloc_fast_execution_point(struct EpmExecutionPoint *ep, 
                                   size_t code_size);

/**
 *  @function ose_epm_free_fast_execution_point
 *  @brief Free memory area that is used by a fast execution point.
 *  @long This function free's code memory area for an fast ep 
 *  @long This function must only be called once on each fast execution point.
 *  After the return of this function the ep struct is no longer used 
 *  internally by the EPM. The code memory area is also free:ed
 *  @long This function is only available in the core image.
 *  It must be called in supervisor mode.
 *  @input ep Pointer to a struct containing information about the
 *  execution point. 
 *  @return  See enum EpmStatus
 *  @seealso ose_epm_alloc_fast_execution_point
 *  @seealso ose_epm_enable_fast_execution_point
 *  @seealso ose_epm_disable_fast_execution_point
**/
EpmStatus
ose_epm_free_fast_execution_point(struct EpmExecutionPoint *ep);



/**
 *  @function ose_epm_enable_fast_execution_point
 *  @brief Add an fast_execution point.
 *  @long Before this function is called on an ep 
 *  ose_epm_alloc_fast_execution_point() must be called to setup a code
 *  memory area. This memory area must be filled with code that is called 
 *  to evaluate if the callback should be called. 
 *  @long This functions patches the original code in a way that makes 
 *  the code memory area is called and depending on that code it will
 *  call the callback.
 *  @long This function is only available in the core image.
 *  It must be called in supervisor mode. 
 *  @input ep Pointer to a struct containing information about the
 *  execution point. This struct must be preserved thru the lifetime of the
 *  execution point. It is used internally by the EPM.
 *  @return  See enum EpmStatus
 *  @seealso ose_epm_disable_fast_execution_point
 *  @seealso ose_epm_alloc_fast_execution_point 
 *  @seealso ose_epm_free_fast_execution_point
**/
EpmStatus
ose_epm_enable_fast_execution_point(struct EpmExecutionPoint *ep);

/**
 *  @function ose_epm_disable_fast_execution_point
 *  @brief Disable an fast execution point.
 *  @long This function is used to remove the patch to original code
 *  after this function is called ose_epm_enable_fast_execution_point() can be 
 *  called to re-enable the ep or ose_epm_free_fast_execution_point() to 
 *  free up the internal resources used by the ep.
 *  @long This function is only available in the core image.
 *  It must be called in supervisor mode.
 *  @input ep Pointer to a struct containing information about the
 *  execution point.
 *  ose_epm_add_execution_point.
 *  @return See enum EpmStatus
 *  @seealso  ose_epm_enable_fast_execution_point
 *  @seealso ose_epm_alloc_fast_execution_point
 *  @seealso ose_epm_free_fast_execution_point
**/
EpmStatus
ose_epm_disable_fast_execution_point(struct EpmExecutionPoint *ep);

/**
 *  @function ose_epm_fast_execution_point_inside_check_callback
 *  @brief Check if address is inside a callback evaluation code 
 *  @return 0 if it is not.
 *  @seealso ose_epm_alloc_fast_execution_point
 *  @seealso ose_epm_free_fast_execution_point
 *  @seealso ose_epm_enable_fast_execution_point
 *  @seealso ose_epm_disable_fast_execution_point
**/
int
ose_epm_fast_execution_point_inside_check_callback(OSADDRESS address);



#ifdef __cplusplus
}
#endif

#endif /* #ifndef OSE_EPM_PRIVATE_H_ */
