/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2006 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

#ifndef _HASH_H
#define _HASH_H

#include <stdint.h>
#include "list_ext.h"


/************************** DEFINES **********************************/

#define NOT_INITIALIZED 0xffffffff

#define INIT_ITERATOR(iter) (iter).bucket = NOT_INITIALIZED


/************************** STRUCTURES **********************************/

/* Function for generating a hash value given an entry in the hash.
 * Try to make this function return a value that spreads the entries over
 * all the hash table buckets.*/
typedef uint32_t(get_hash_value_func)(const void *entry);

/* Function for comparing two nodes. This
 * function returns 0 if a does not match b, and 1 if a equals b.*/
typedef int(compare_func)(const void *a, const void *b);

/* Structure describing a hash table. */
struct hash_table_t
{
   /* An array of lists. One for each bucket in the hash. */
   list_t *bucket;

   /* The number of buckets in this hash. */
   uint32_t no_of_buckets;

   /* The get hash value function to use for this hash. */
   get_hash_value_func *get_hash_value;

   /* The compare function to use for this hash. */
   compare_func *compare;
};

/* A hash table iterator, used to iterate over the hash table. */
struct _hash_iterator
{
   /* The bucket the iterator points to. */
   uint32_t     bucket;

   /* The entry the iterator points to. */
   list_link_t *entry;
};

typedef struct _hash_iterator hash_iterator;


/*
 *===========================================================================
 *                         init_hash
 *===========================================================================
 *
 * Description
 * -----------
 * Initializes a hash table with a specific compare function and
 * hash value generator.
 * Needs to be called before any other operation on the hash table.
 *
 * Parameters
 * ----------
 * no_of_buckets  The number of buckets in this hash table.
 *                Statically defined at initialization.
 * compare        Callback function for comparing two nodes. This
 *                function returns 0 if key_a does not match key_b,
 *                and 1 if key_a equals key_b.
 * get_hash_value Returns a 32 bit value given an entry. Try to make this
 *                function return a value that spreads the entries over
 *                all the hash table buckets.
 *
 */
void
init_hash(struct hash_table_t *hash, uint32_t no_of_buckets,
          get_hash_value_func *get_hash_value, compare_func *compare);

/*
 *===========================================================================
 *                         delete_hash
 *===========================================================================
 *
 * Description
 * -----------
 * Deletes a hash table. It does not delete any of the entries in
 * the hash table.
 *
 */
void
delete_hash(struct hash_table_t *hash);

/*
 *===========================================================================
 *                        find_in_hash
 *===========================================================================
 *
 * Description
 * -----------
 * Searches for a matching entry in the hash table *hash.
 *
 * Returns
 * -------
 * struct list_link_t *   Returns the found entry or NULL if it is not found.
 *
 */
list_link_t *
find_in_hash(struct hash_table_t *hash, const void *key);

/*
 *===========================================================================
 *                        iterate_over_hash
 *===========================================================================
 *
 * Description
 * -----------
 * Iterates over the hash. If the iterator is initialized with the
 * INIT_ITERATOR() macro, the search starts at
 * the beginning of the hash table, otherwise at the entry after the
 * point the iterator points to. The iterator is modified to point to
 * the position of the mathching entry by this function.
 * It returns an entry matching the key and uses the provided compare function.
 * By repeatedly calling this function
 * and supplying the iterator every time, you will iterate over all matching
 * entries in the hash table.
 * If the compare function is NULL, any entry is considered to match.
 *
 * Returns
 * -------
 * struct list_link_t *   Returns the found entry or NULL if it is not found.
 *
 */
list_link_t *
iterate_over_hash(struct hash_table_t *hash, const void *key,
                  hash_iterator *iterator, compare_func *compare);

/*
 *===========================================================================
 *                        insert_in_hash
 *===========================================================================
 *
 * Description
 * -----------
 * Inserts an entry in the hash_table *hash unless it is already there.
 *
 * Returns
 * -------
 * int      Returns one if the entry was inserted,
 *          or zero is it is already in a hash.
 *
 */
int
insert_in_hash(struct hash_table_t *hash, list_link_t *entry);

/*
 *===========================================================================
 *                        delete_from_hash
 *===========================================================================
 *
 * Description
 * -----------
 * Deletes an entry in the hash table *hash.
 *
 * Returns
 * -------
 * int     Returns one if the entry was deleted,
 *         or zero if it could not be found.
 *
 */
int
delete_from_hash(struct hash_table_t *hash, list_link_t *entry);

#endif /* HASH_H */
