/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2009 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

#ifndef _KSSS_MAP_H
#define _KSSS_MAP_H

/*
 * This file describes the private interface between KSSS and MM components.
 */

#include "sys/types.h"
#include "mmtypes.h"
#include "stdint.h"
#include "ksss_paging.h"

/*
 ***************************************************************************
 * Paging interface
 * This interface is implemented by KSSS if that module is present in the
 * image. If not, the internal functions that runs on top of efs/fm.sig is
 * used (paging on file level)
 ***************************************************************************
 */

/*
 * if this function is referred to by KSSS init, the weak references below
 * is resolved and the MMAN module will use ksss page in instead of using the
 * legacy crt.
 */
void ksss_map_init(void);

struct KsssInode;
struct KsssFsMapApi;
struct FsMap;


/**
 * @type MmanPageMap
 *
 * @brief Object used by mman and ksss to share information about
 *        how to map an inode. 
 * 
 */
struct MmanPageMap
{
   struct KsssInode *inode;             /* Inode to use */
   const struct KsssFsMapApi *map_api;  /* FS map API */
   struct FsMap *map_handle;            /* FS Map handle */
   KsssPageFaultFunc *page_fault_func;  /* Function called to handle the 
                                         * pagefault the page_fault_handle is
                                         * passed as a parameter 
                                         */
   void *page_fault_handle;             /* passed as a parameter to the 
                                         * page_fault_func() 
                                         */
};



/*
 * @function ksss_map_create
 *
 * @brief Creates a page map object with information required to map
 *        blocks of a file. 
 *
 * @long  This function is called by MM when mapping a file for the first time.
 *
 * @input fd                File descriptor for file to be memory mapped.
 *
 * @input cb_api            Collection of pager callbacks needed by ksss.
 * @input page_size_bytes   Size of each page in bytes.
 * @input inout_supports    Which features that the caller of mmap() function
 *                          requested.
 * @output inout_supports   Which features that the volume manager supports.
 *                          Note that this might be more or less than requested.
 * @output out_page_map_ref Reference to inode page map, used by KSSS.
 *
 * @return
 *   @tab 0       @tab if success. @break
 *   @tab -ENOSUP @tab if paging is not supported for this file. @break
 *
 */
typedef int (*KsssMapCreateFunc)(int fd,
                                 struct KsssInode *inode,
                                 const struct KsssPagingCallbackApi *cb_api,
                                 size_t page_size_bytes,
                                 int *inout_supports,
                                 struct MmanPageMap* out_page_map_ref);

/*
 * @function ksss_map_destroy
 *
 * @brief Used to destroy a page map object when the corresponding
 *        file and its blocks no longer are mapped. 
 *
 * @long  Unreference a ksss_fs.h FsMap and FsMap
 *
 * @long  This function is called by MM when a file will no longer
 *        be memory mapped.
 *
 * @input page_map  Inode page map reference for file to be unmapped.
 *
 * @return 
 *   @tab 0       @tab if success. @break
 */
typedef int (*KsssMapDestroyFunc)(struct MmanPageMap *page_map);


/*
 * @function ksss_map_map_range
 *
 * @brief Function that maps sectors of a file.
 *
 * @input page_map  Inode page map reference for file to be mapped.
 *
 * @input offset    Offset into file.
 *
 * @input size      Length of mapping. 
 *
 * @return 
 *   @tab 0       @tab if success. @break
 *   @tab -ENOSUP @tab if paging is not supported for this file. @break
 */
typedef int (*KsssMapMapRangeFunc)(struct MmanPageMap *page_map,
                                   uint64_t offset, size_t size);

/*
 * @function ksss_map_unmap_range
 *
 * @brief Function that unmaps sectors of a file.
 *
 * @input page_map  Inode page map reference for file to be mapped.
 *
 * @input offset    Offset into file.
 *
 * @input size      Length of mapping. 
 *
 * @return 
 *   @tab 0       @tab if success. @break
 *   @tab -ENOSUP @tab if paging is not supported for this file. @break
 */
typedef int (*KsssMapUnmapRangeFunc)(struct MmanPageMap *page_map,
                                     uint64_t offset, size_t size);


/*
 * @function ksss_map_get_inode
 *
 * @brief Function that fetches and references an inode.
 *
 * @input  fd        File descriptor.
 *
 * @output out_inode File inode.
 *
 * @return 
 *   @tab 0       @tab if success. @break
 *   @tab -ENOSUP @tab if paging is not supported for this file. @break
 */
typedef int (*KsssMapGetInodeFunc)(int fd, struct KsssInode **out_inode);

/*
 * @function ksss_map_unref_inode
 *
 * @brief Function that dereferences an inode.
 *
 * @input  inode        The inode.
 */
typedef void (*KsssMapUnrefInodeFunc)(struct KsssInode *inode);

/*
 * @function ksss_map_page_fault
 *
 * @brief Function that handles a page fault.
 *
 * @long  When a page fault occurs, mman calls this function.
 * 
 * @input  page_map    Map instance handle.
 * @input  page_fault  Pointer to a struct describing the page fault.
 *
 * @output out_inode File inode.
 *
 * @return 
 *   @tab 0       @tab if success. @break
 *   @tab -ENOSUP @tab if paging is not supported for this file. @break
 */
typedef int (*KsssMapPageFaultFunc)(struct MmanPageMap *page_map,
                                    struct KsssPageFault *page_fault);


/**
 * @type   KsssMapFunctions
 *
 * @brief  Collections of KSSS callbacks used by MM. 
 *
 * @long   The callbacks are installed by KSSS and used
 *         by MM when mapping and unmapping files in memory.
 *         All of the callbacks must be valid, i.e. it is
 *         illegal to only support a subset of them.
 *
 * @field  create     Called by mman when mapping a file for the first time.
 * @field  destroy    Called by mman when a file no longer is mapped.
 * @field  map        Called by mman when mapping blocks of a file.
 * @field  unmap      Called by mman when unmapping block of a file.
 * @field  get_inode  Used to obtain the inode referenced by a file descriptor.
 * @field  page_fault Called when a page fault occurs.
 */
struct KsssMapFunctions
{
   KsssMapCreateFunc     create;
   KsssMapDestroyFunc    destroy;
   KsssMapMapRangeFunc   map;
   KsssMapUnmapRangeFunc unmap;
   KsssMapGetInodeFunc   get_inode;
   KsssMapUnrefInodeFunc unref_inode;
   KsssMapPageFaultFunc  page_fault;
};


/*
 * @function mman_map_fd
 *
 * @brief Function that maps sectors of a file.
 *
 * @input  fd                File descriptor.
 *
 * @input cb_api             Struct containing needed function pointers
 *                           to handle the demand paging.
 * @input page_size_bytes    Size of each page in bytes.
 * @input inout_supports     Which features that the caller of mmap() function
 *                           requested.
 *
 * @output inout_supports    Which features that the volume manager supports.
 *                           Note that this might be more or less than
 *                           requested.
 *
 * @input  offset            Offset into file where memory map starts.
 *
 * @input  size              Length of memory map. 
 *                           
 * @output out_page_map_ref  Reference to inode page map, used by KSSS.
 *
 * @return 
 *   @tab 0       @tab if success. @break
 *   @tab -ENOSUP @tab if paging is not supported for this file. @break
 *   @tab -ENOMEM @tab if not able to allocate a new page map-object.@break
 */
int mman_map_fd(int fd,
                const struct KsssPagingCallbackApi *cb_api,
                size_t page_size_bytes,
                int *inout_supports,
                uint64_t offset,
                size_t size,
                struct MmanPageMap **out_page_map_ref);


/*
 * @function mman_unmap_fd
 *
 * @brief Function that unmaps sectors of a file.
 *
 * @input  fd        File descriptor.
 *
 * @output out_inode File inode.
 *
 * @return 
 *   @tab 0       @tab if success. @break
 *   @tab -ENOSUP @tab if paging is not supported for this file. @break
 */
int mman_unmap_fd(struct MmanPageMap *page_map,
                  uint64_t offset, size_t size);


/*
 * @function mman_page_fault
 *
 * @brief Handle a page fault
 *
 * @long  This function is called from the pager when a page fault has
 *        occurred.
 *
 * @input  page_map    Page map reference. Identifies the inode.
 * @input  page_fault  Struct describing the page fault. Also contains the
 *                     result at return.
 *
 * @return
 *   @tab 0    @tab if success. @break
 *   @tab -EIO @tab if I/O error occurred. @break
 */
int mman_page_fault(struct MmanPageMap *page_map,
                    struct KsssPageFault *page_fault);


/*
 * @function mman_page_invalidate
 *
 * @brief Handle page invalidation. 
 *
 * @long  Called by filesystem/block device when pages have
 *        been invalidated to force the pager to remap them.
 *
 * @input inode     Handle that KSSS has supplied earlier.
 * @input offset    Starting file offset in bytes of range to invalidate.
 * @input size      Size in bytes of file offset range to invalidate.
 * @input new_data  Address to where pager can copy the new data, or 0 if 
 *                  remapping.
 * @input new_addr  If remapping, this is the new address for the page to be
 *                  remapped. The address corresponds to the offset. If not
 *                  remapping, this field must be 0.
 *
 * @return Nothing.
 */
void
mman_page_invalidate(struct KsssInode *inode, uint64_t offset,
                     size_t size, log_addr_t new_data,
                     phys_addr_t new_phys_addr);


/*
 * @function mman_ksss_init
 *
 * @brief Initiate ksss mmap callbacks in mman
 *
 * @long  This function in MMAN is called from KSSS layer if it is installed.
 *
 * @input ksss_funcs  Collection of callbacks to be installed in mman.
 *
 * @return
 *   @tab 0    @tab if success. @break
 *   @tab -EIO @tab if I/O error occurred. @break
 */
void mman_ksss_init(struct KsssMapFunctions *ksss_funcs);

#endif /* _KSSS_MAP_H */

