/* COPYRIGHT-ENEA-SRC-R2 *
**************************************************************************
* Copyright (C) 2007-2009 by Enea Software AB.
* All rights reserved.
*
* This Software is furnished under a software license agreement and
* may be used only in accordance with the terms of such agreement.
* Any other use or reproduction is prohibited. No title to and
* ownership of the Software is hereby transferred.
*
* PROPRIETARY NOTICE
* This Software consists of confidential information.
* Trade secret law and copyright law protect this Software.
* The above notice of copyright on this Software does not indicate
* any actual or intended publication of such Software.
**************************************************************************
* COPYRIGHT-END */

/**
 * @deprecated
 * @toc File_System:
 *
 * @file ose_spi/ksss_paging.h
 *
 * @brief Interfaces for paging on file systems and partitions.
 *
 * @long This file contains a set of interfaces used by services implementing
 *       demand paging. For more information about what a service is and how
 *       it works, see file ose_spi/ose_service.h. The interfaces defined
 *       in this file are:
 *
 * @tab @link "" "KsssFsPagingApi" @tab
 *       This interface is an extension of
 *       @link "ose_spi/ksss_fs.h" "KsssVolumeApi"
 *       and is used to create a new memory mapping of a file. It uses
 *       @link "ose_spi/ksss_fs.h" "FsVolume" as instance handle and will
 *       return interface @link "" "KsssFsMapApi" to be used for handling the
 *       memory mapped file.                                             @break
 *
 * @tab @link "" "KsssFsMapApi" @tab
 *       This interface handles a memory mapped file. Each memory mapped file
 *       will have zero or one instance of this interface, any further memory
 *       mappings of the same file will use the same instance.           @break
 *
 * @tab @link "" "KsssPartitionPagingApi" @tab
 *       This interface is an extension of
 *       @link "ose_spi/ksss_partition.h" "KsssPartitionIoApi" and is used from
 *       a file system to create a memory map of block device blocks. It uses
 *       @link "ose_spi/ksss_partition.h" "KsssPartitionSession" as instance
 *       handle and will return a @link "" "KsssPartitionMapApi" interface.
 *                                                                       @break
 *
 * @tab @link "" "KsssPartitionMapApi" @tab
 *       This interface handles a memory mapped partition. It is used by the
 *       file system to map and unmap blocks for the memory mapped file. @break
 *
 * @tab @link "" "KsssPagingCallbackApi" @tab
 *       This is an interface with callback functions to KSSS, used by file
 *       systems to perform memory page allocation and invalidation. The
 *       interface is provided to the file system, when the file system is
 *       accessed via the KsssFsPagingApi.                               @break
 *
 * @long All functions in this interface are expected to be re-entrant unless
 *       otherwise specified.
 *
 * @long Unless otherwise specified, non-negative return codes indicate success
 *       and negative return codes failure. Negated POSIX ERRNO codes are used
 *       to indicate failure.
 *
 * @long Sanity checks on parameters to functions are optional. Error codes
 *       to use if checks fail are specified though.
 *
 * @long Paging of a file is initiated by the demand pager. It will ask the
 *       volume manager that manages this file to map the wanted file offset
 *       range. The volume manager may, should, check if the partition on
 *       which the volume resides supports paging. If so, the volume manager
 *       translates the file offset range into block ranges which it will ask
 *       the partition to map.
 * @long However, if the volume manager supports sparse files it cannot, and
 *       will not, ask the partition to map the holes in the sparse file.
 *       There is no media space for those holes.
 *       So there may be parts of a mapped file offset range that do not have
 *       any corresponding mapped block ranges.
 * @long After setting up the map, the volume manager may, for efficiency,
 *       give the partition's page fault handler and its corresponding
 *       page_fault_handle back to the demand pager. If so, the page faults will
 *       be handled directly by the partition with no involvement of the
 *       volume manager.
 * @long Pages used by the demand pager are often several blocks in size. This
 *       means that a page may cover both part of a hole in a sparse file and
 *       written data (see the example below). This has to be considered by the
 *       page fault handlers on both volume level and partition level.
 *       The partition's page fault handler must therefore be able to assemble a
 *       page composed of fragments from several mapped block ranges and
 *       unmapped holes.
 * @long Unmapped blocks shall by the partition's page fault handler be reported
 *       as zero-filled. This can be done because the partition is guaranteed by
 *       the demand pager never to receive page faults for file offsets that are
 *       not mapped.
 *
 * @example
Sparse File:   |------|        |--------|
Page:                      |------|
 *
 * @restriction NOT YET AVAILABLE in standard OSE.
 */

#ifndef _KSSS_PAGING_H
#define _KSSS_PAGING_H

#include "ose_service.h"
#include "stdint.h"
#include "mmtypes.h"

#ifdef __cplusplus
extern "C" {
#endif

/*
 * Types defined elsewhere, but forward declared here to keep include list
 * down to what is really needed.
 */
struct FsVolume;
struct KsssFd;
struct KsssInode;
struct KsssPartition;
struct KsssPartitionSession;
struct MmanPageMap;


/**
 * @type FsMap
 *
 * @brief File map object.
 *
 * @long This object is used by the file system to be able to quickly map
 *       a file offset into a partition access. It is used when paging
 *       a file.
 * @long This object is only dereferenced by the file system, i.e. it is the
 *       file system that determines the contents of this object.
 * @long FsMap is expected to be reference counted, i.e. when the last
 *       reference to FsMap is removed, FsMap must never be used again
 *       and can therefore be removed too.
 * @long Note that there are one FsMap object per mapped file, not per file
 *       mapping.
 */
struct FsMap;


/**
 * @type  KsssMapType
 *
 * @brief Values describing how the page was mapped or can be mapped.
 *
 * @long  The demand pager uses this bitmap to indicate which map types that
 *        are legal for this page fault. The volume manager selects one of
 *        these map types when it has handled the page fault.
 *        See @link "" "KsssPageFault" for more information.
 *
 * @long  Any undefined bit in the flag bit mask must be zero to make it
 *        possible to add new flags in the future in a backwards
 *        compatible way.
 *
 * @restriction This ENUM is not intended to be used as a type, only to define
 *              constants. It is unsafe to use ENUM as a type in public
 *              interfaces, while ENUM constants are safe.
 *
 * @field KSSS_MAP_ZERO  Map pages completely filled with byte value zero.
 *              The number of consecutive pages to map
 *              is given by field @link "" "KsssPageFault" field
 *              pages_mapped.
 *              NOTE: This map type is always available, and must therefore be
 *                    set by the pager when entering the page fault function.
 *
 * @field KSSS_MAP_IN_PLACE  Map pages in place, starting at physical address
 *              given by @link "" "KsssPageFault" field new_page_phys_addr.
 *              Field pages_mapped in the same struct indicates how
 *              many consecutive pages that may be mapped.
 *
 * @field KSSS_MAP_WITH_DMA  The page has only been written using DMA
 *              transfers. This indicates that the page might not need to be
 *              flushed.
 *
 * @field KSSS_MAP_COPY  A new page was allocated and filled with data using
 *              memory copy. If DMA is used to copy parts of the data it's the
 *              responsibility of the partition to invalidate the caches.
 *              NOTE: This map type is always available, and must therefore be
 *                    set by the pager when entering the page fault function.
 */
enum KsssMapType
{
   KSSS_MAP_ZERO     = (1 << 0),
   KSSS_MAP_IN_PLACE = (1 << 1),
   KSSS_MAP_WITH_DMA = (1 << 2),
   KSSS_MAP_COPY     = (1 << 3)
};

/**
 * @type  KsssPageFault
 *
 * @brief Struct describing the page fault handling progress.
 *
 * @long  This object is filled in by the pager and describes the page fault.
 *        Note that the pager might have additional information after this
 *        struct describing state information needed for callback functions.
 *
 * @long  This struct will be updated when the page fault is handled.
 *
 * @field offset_bytes  File offset in bytes where the page fault occurred. The
 *            offset corresponds to the start of the page where the page
 *            fault occurred.
 *
 * @field data_log_addr  Filled in by pager. When coming from the pager this is
 *            the starting address of the page within which the page fault
 *            occurred. When @link "" "KsssPagerAllocPage" is called, this
 *            field is updated to indicate the allocated page's logical start
 *            address. Note that for newly allocated pages, a temporary logical
 *            address is used to solve potential race condition that can occur
 *            if several simultaneous page faults occur on the same page.
 *
 * @field data_phys_addr  Filled in by page allocator, file system or device
 *            driver. When coming from the pager, this field is zero since
 *            there is no physical address associated with the faulted address.
 *            For map type map in place, device driver fills this field with
 *            the physical start address of the page or pages to map in place.
 *            For map type copy the callback function
 *            @link "" "KsssPagerAllocPage" will fill in the allocated page's
 *            physical start address.
 *
 * @field map_type  Filled in by pager, updated by file system or device
 *            driver.
 *            The pager uses this field to indicate which map types it can
 *            allow and understand.
 *            File system and device driver will update this field by selecting
 *            one of the map types set by the pager. See @link "" "KsssMapType"
 *            for valid values.
 *
 * @field pages_mapped  Filled in by pager, updated by file system or
 *            partition. When coming from the pager, this is the number of
 *            pages from and including the one causing a page fault that the
 *            caller has mapped.
 *            Coming from the file system or partition this is the number of
 *            consecutive pages that has been filled with data, starting with
 *            the page causing the page fault. For map in place, this is the
 *            number of consecutive pages that can be mapped in place, starting
 *            with the page causing the page fault.
 *            When returning to the pager, the pages_mapped can be both
 *            greater than or lower than what was provided by the pager.
 *            It is up to the pager what to do if this is more pages than
 *            expected.
 */
struct KsssPageFault
{
   /*
    * Fields filled in by pager
    */
   uint64_t    offset_bytes;

   /*
    * Fields filled in by pager, and updated by file system and device
    * driver
    */
   log_addr_t  data_log_addr;
   phys_addr_t data_phys_addr;
   int         map_type;

   /*
    * Fields filled in by pager, updated by file system or device driver
    */
   unsigned    pages_mapped;
};

/*
 ***************************************************************************
 * Page fault function
 *
 * This function has the same declaration and semantics for both file systems
 * and partitions. This makes it possible in some implementations to bypass
 * file system and call partition directly.
 ***************************************************************************
 */

/**
 * @type  KsssPageFaultFunc
 *
 * @brief Function that handles a page fault.
 *
 * @long  When a page fault occurs, the pager calls this function.
 *
 * @long If the page is returned as an in-place map, the file system or
 *       partition must invalidate the page using
 *       @link "" "KsssInvalidateMapping" before the data in it becomes
 *       invalid.
 *
 * @long NOTE: Recursive page faults for the same process can occur and must be
 *             handled. These recursive page faults will never occur for the
 *             same page though. If an OSE mutex is used for solving this
 *             problem, make sure that it was created with OSE_MUTEX_RECURSIVE
 *             attribute set.
 *
 * @long NOTE: During the scope of this function, most OSE and KSSS functions
 *             are illegal to call. Any such function call must have been done
 *             when @link "" "KsssFsPagingMapCreate" was called.
 *             The ose_mutex_lock() and ose_mutex_unlock() can be called
 *             however, if this is needed to handle recursive page faults.
 *
 * @input  handle      Map instance handle.
 * @input  page_fault  Pointer to a struct describing the page fault.
 *
 * @return
 *   @tab 0          @tab if success. @break
 *   @tab -EIO       @tab if I/O error occurred. @break
 *   @tab -ECANCELED @tab if the operation should be canceled. @break
 */
typedef int (KsssPageFaultFunc)(void *handle,
                                struct KsssPageFault *page_fault);

/*
 ***************************************************************************
 * Callback functions used by file system
 ***************************************************************************
 */

/**
 * @type  KsssInvalidateMapping
 *
 * @brief Invalidate memory map for a file offset range.
 *
 * @long  This is called from a file system when the content of a memory mapped
 *        file has changed, and is memory mapped or when a map in place block
 *        device has moved the data. It should be called after the data has
 *        been copied to the new place but before the old block is modified.
 *
 * @long  If both new_data and new_addr is 0 the mapping is only invalidated,
 *        the updated information, data or new map in place address, will be
 *        updated upon the next page fault.
 *
 * @long  The @link "" "KsssInvalidateMapping" is to be used when
 *        in-place mapped memory is moved, evicted or changed.
 *
 * @input inode     Handle that KSSS has supplied earlier.
 * @input offset    Starting file offset in bytes of range to invalidate.
 * @input size      Size in bytes of file offset range to invalidate.
 * @input new_data  Address to where pager can copy the new data, or 0 if
 *                  remapping.
 * @input new_addr  If remapping, this is the new address for the page to be
 *                  remapped. The address corresponds to the offset. If not
 *                  remapping, this field must be 0.
 *
 * @return Nothing.
 */
typedef void (KsssInvalidateMapping)(
   struct KsssInode *inode,
   uint64_t offset,
   size_t size,
   log_addr_t new_data,
   phys_addr_t new_addr);

/**
 * @type  KsssPagerAllocPage
 *
 * @brief Allocate a memory page from pager cache.
 *
 * @long  This function is called from a file system to allocate a page
 *        when handling a page fault.
 *        The allocated page can be used both for memory copy and DMA
 *        transfers.
 *
 * @long  NOTE: This function might be called even though the end result is
 *              map in place or zero map. If this is the case, the pager will
 *              simply free the allocated page.
 *
 * @long  This function is part of the @link "" "KsssPagingCallbackApi"
 *        interface.
 *
 * @input page_info    Pointer to a struct describing the page fault.
 *
 * @return
 *   @tab 0            @tab If the operation should proceed. @break
 *   @tab -ECANCELED   @tab If the operation should be canceled. @break
 */
typedef int (KsssPagerAllocPage)(struct KsssPageFault *page_info);


/**
 * @type  KsssPagingCallbackApi
 *
 * @brief Callback interface used by file systems towards the pager.
 *
 * @long  This struct contains function pointers that KSSS supplies to the
 *        file system. The interface is provided as a parameter to
 *        @link "" "KsssFsPagingMapCreate" in the KsssFsPagingApi.
 *
 * @field invalidate See @link "" "KsssInvalidateMapping". Note that this
 *                   function is the same for all mappings. It is supplied
 *                   in this struct only to achieve a dynamic linkage.
 * @field alloc_page See @link "" "KsssPagerAllocPage".
 */
struct KsssPagingCallbackApi
{
   KsssInvalidateMapping *invalidate;
   KsssPagerAllocPage    *alloc_page;
};

/*
 ***************************************************************************
 * File system paging interface
 *
 * This interface will create a map object. The interface to handle the map
 * object is described in the File system map interface.
 ***************************************************************************
 */

struct KsssFsMapApi;

/**
 * @type  KsssPagingSupports
 *
 * @brief Bitmap of supported features for a paging device or volume
 *
 * @long  This bitmap is used by the demand pager to indicate which features
 *        that have been requested by the mmap() call. The device and volume
 *        use it to indicate what features they support. It is the demand pager
 *        that in the end decides if the supported features are enough.
 *
 * @long  Any undefined bit in the flag bit mask must be zero to make it
 *        possible to add new flags in the future in a backwards
 *        compatible way.
 *
 * @restriction This ENUM is not intended to be used as a type, only to define
 *              constants. It is unsafe to use ENUM as a type in public
 *              interfaces, while ENUM constants are safe.
 *
 * @field KSSS_PAGING_SUPPORTS_FILE_WRITE Indicates that the file can be opened
 *              for write while being paged.
 * @field KSSS_PAGING_SUPPORTS_INTERRUPTS_DISABLED Indicates that a page fault
 *              can be handled with interrupts disabled.
 * @field KSSS_PAGING_SUPPORTS_DMA When the create map call is issued, this bit
 *              indicates whether the demand pager can allocate memory suitable
 *              for DMA transfers. When exiting the create map call, this bit
 *              indicates whether the driver prefers memory suitable for
 *              DMA transfers. Note that KSSS_MAP_WITH_DMA might be unset when
 *              a page fault occurs even if this bit is set both when entering
 *              and when exiting the create map function.
 */
enum KsssPagingSupports
{
   KSSS_PAGING_SUPPORTS_FILE_WRITE          = (1 << 0),
   KSSS_PAGING_SUPPORTS_INTERRUPTS_DISABLED = (1 << 1),
   KSSS_PAGING_SUPPORTS_DMA                 = (1 << 2)
};


/**
 * @type  KsssFsPagingMapCreate
 *
 * @brief Create a mmap of a file.
 *
 * @long  A mapping object is created and associated with a KsssInode.
 *        Mapped ranges can be added and removed to this object by calls to
 *        @link "" "KsssFsMapMapRange" and @link "" "KsssFsMapUnmapRange".
 *
 * @long  Returns a mapping object and a specified page fault API for demand
 *        pager use.
 *
 * @long  The file to be mapped will always be an opened regular file.
 *
 * @input  vol     Pointer to a mounted file system object.
 * @input  inode   Pointer to the inode object to map.
 * @input  fd      Pointer to volume specific file descriptor object.
 * @input  cb_api  Pointer to a struct with function pointers describing a
 *                 callback interface towards the pager, for use by
 *                 the file system.
 * @input  page_size_bytes Size in bytes for a page.
 * @input  inout_supports Pointer to bitmask describing which features that are
 *                 requested. See @link "" "KsssPagingSupports" for possible
 *                 features.
 * @output inout_supports Bitmask describing which features that the volume
 *                 manager can support for this mapping. This can differ from
 *                 requested in any way, e.g. it might indicate less features
 *                 than requested. It can also indicate more features than
 *                 requested.
 * @output out_api Pointer to an interface to handle a map object.
 * @output out_instance Pointer to a map instance object to be used in calls
 *                 to the interface returned in out_api. Note that this object
 *                 is reference counted, and
 *                 @link "ose_service.h" "OseBaseApiReferenceRelease" must be
 *                 called when finished with the object.
 * @output out_page_fault_func Pointer to the function that will receive the
 *                 page fault calls from the pager.
 * @output out_page_fault_handle Map instance handle used in calls to the
 *                 page fault function above.
 *
 * @return
 *   @tab 0          @tab Success. @break
 *   @tab -EIO       @tab I/O error occurred. @break
 *   @tab -ENOMEM    @tab Not enough memory to fulfill request. @break
 *   @tab -ENOTSUP   @tab The file does not support paging. @break
 */
typedef int (KsssFsPagingMapCreate)(
   struct FsVolume *vol,
   struct KsssInode *inode,
   struct KsssFd *fd,
   const struct KsssPagingCallbackApi *cb_api,
   size_t page_size_bytes,
   int *inout_supports,
   const struct KsssFsMapApi **out_api,
   struct FsMap **out_instance,
   KsssPageFaultFunc **out_page_fault_func,
   void **out_page_fault_handle);

/**
 * @macro KSSS_FS_PAGING_API_NAME
 *
 * @brief Name of the @link "" "KsssFsPagingApi"
 *        KSSS file system paging interface.
 *
 * @long  This is the name that shall be used in call to
 *        @link "" "OseBaseApiInterfaceQuery" to get interface
 *        @link "" "KsssFsPagingApi".
 *
 * @long  The name is on the form "com.enea.ksss_fs_paging_api.z", where z is
 *        a version string which can vary.
 *        See the header file for the full name.
 */
#define KSSS_FS_PAGING_API_NAME ("com.enea.ksss_fs_paging_api.v1")

/* Deprecated old name. Will be removed without further notice. */
#define API_TYPE_KSSS_FS_PAGING_API KSSS_FS_PAGING_API_NAME


/**
 * @type  KsssFsPagingApi
 *
 * @brief Interface to handle demand paging on a file system.
 *
 * @long  This interface will create a map object. The interface to handle
 *        the map object is described in the File system map interface,
 *        returned by this call when the map object has been created.
 *
 * @field fs_base        See @link "" "OseBaseApi"
 * @field fs_map_create  See @link "" "KsssFsPagingMapCreate".
 *
 * @seealso KSSS_FS_PAGING_API_NAME
 */
struct KsssFsPagingApi
{
   struct OseBaseApi       fs_base;
   KsssFsPagingMapCreate  *fs_map_create;
};


/*
 ***************************************************************************
 * File system map interface
 *
 * This interface handles a map object created by KsssFsPagingMapCreate().
 ***************************************************************************
 */

/**
 * @type  KsssFsMapMapRange
 *
 * @brief Add a range to a file mapping.
 *
 * @input instance  Map instance handle.
 * @input offset    Offset in bytes of the range to be added to the mapping.
 * @input size      Size in bytes of the range to be added to the mapping.
 *
 * @return
 *   @tab 0         @tab if success. @break
 *   @tab -ENOMEM   @tab if not enough memory to handle the call. @break
 */
typedef int (KsssFsMapMapRange)(
   struct FsMap *instance,
   uint64_t offset,
   size_t   size);


/**
 * @type  KsssFsMapUnmapRange
 *
 * @brief Remove a range from a file mapping.
 *
 * @input instance  Map instance handle.
 * @input offset    Offset in bytes of the range to be removed from the
 *                  mapping.
 * @input size      Size in bytes of the range to be removed from the mapping.
 *
 * @return
 *   @tab 0         @tab if success. @break
 *   @tab -ENOMEM   @tab if not enough memory to handle the call. @break
 */
typedef int (KsssFsMapUnmapRange)(
   struct FsMap *instance,
   uint64_t offset,
   size_t   size);


/**
 * @macro KSSS_FS_MAP_API_NAME
 *
 * @brief Name of the @link "" "KsssFsMapApi"
 *        KSSS file system map interface.
 *
 * @long  This is the name that shall be used in call to
 *        @link "" "OseBaseApiInterfaceQuery" to get interface
 *        @link "" "KsssFsMapApi".
 *
 * @long  The name is on the form "com.enea.ksss_fs_map_api.z", where z is
 *        a version string which can vary.
 *        See the header file for the full name.
 */
#define KSSS_FS_MAP_API_NAME ("com.enea.ksss_fs_map_api.v1")

/* Deprecated old name. Will be removed without further notice. */
#define API_TYPE_KSSS_FS_MAP_API KSSS_FS_MAP_API_NAME


/**
 * @type  KsssFsMapApi
 *
 * @brief Interface to handle demand paging on a file system.
 *
 * @long  This interface is used to implement demand paging on a file.
 *
 * @field fs_base          See @link "" "OseBaseApi".
 * @field fs_map_range     See @link "" "KsssFsMapMapRange".
 * @field fs_unmap_range   See @link "" "KsssFsMapUnmapRange".
 *
 * @seealso KSSS_FS_MAP_API_NAME
 */
struct KsssFsMapApi
{
   struct OseBaseApi    fs_base;
   KsssFsMapMapRange   *fs_map_range;
   KsssFsMapUnmapRange *fs_unmap_range;
};


/*
 ***************************************************************************
 * Partition paging interface
 ***************************************************************************
 */

struct KsssPartitionMapApi;

/**
 * @type PartitionMap
 *
 * @brief Partition paging map object
 *
 * @long  This object is used as a session for a memory mapping of e.g. the
 *        blocks associated with a file.
 *
 * @long  Blocks can be mapped and unmapped into this session and page faults
 *        can be handled by it.
 *
 * @long  This object is only dereferenced by the partition provider, i.e. it
 *        is the partition provider that determines the contents of this
 *        object.
 *
 * @long  PartitionMap is expected to be reference counted, i.e. when the
 *        last reference to PartitionMap is removed, PartitionMap must never be
 *        used again and can therefore be removed too.
 * @long  Unmapped blocks on the partition shall be reported a zero-filled. This
 *        can be done because the partition is guaranteed by the demand pager
          never to receive page faults for file offsets that are not mapped.
          see @link "" "ose_spi/ksss_paging.h" for more info.
*/
struct PartitionMap;

/**
 * @type  KsssPartitionPagingMapCreate
 *
 * @brief Memory map a file on partition level.
 *
 * @long  When @link "" "KsssFsPagingMapCreate" is called, the file system
 *        will call this function followed by a number of calls to
 *        @link "" "KsssPartitionMapMapBlock" and
 *        @link "" "KsssPartitionMapUnmapBlock".
 *
 * @long  To tear down a paging @link "" "KsssPartitionSession" use
 *        @link "ose_service.h" "ose_service_reference_release".
 *
 * @input  partition       Pointer to a partition object.
 * @input  session         A mounted partition session.
 * @input  inode           Pointer to inode of file to memory map.
 * @input  cb_api          Pointer to a struct with function pointers
 *                         describing a callback interface.
 * @input  page_size_bytes Size in bytes for a page.
 * @input  inout_supports  Pointer to bitmask describing which features that
 *                         are requested. See @link "" "KsssPagingSupports" for
 *                         possible features.
 * @output inout_supports  Bitmask describing which features that the volume
 *                         manager can support for this mapping. This can
 *                         differ from requested in any way, e.g. it might
 *                         indicate less features than requested. It can also
 *                         indicate more features than requested.
 * @output out_api         Pointer to an interface to handle a partition map
 *                         object.
 * @output out_map         Pointer to a partition map instance object  to be
 *                         used in calls to the interface returned in out_api.
 *                         Note that this object is reference counted and
 *                         @link "ose_service.h" "OseBaseApiReferenceRelease"
 *                         must be called when finished with the object.
 * @output out_page_fault_func Pointer to the function that will receive the
 *                         page fault calls from the pager or from the file
 *                         system.
 * @output out_page_fault_handle Map instance handle used in calls to the
 *                         page fault function above.
 *
 * @return
 *   @tab 0         @tab if success. @break
 *   @tab -ENOMEM   @tab if not enough memory to handle the call. @break
 */
typedef int (KsssPartitionPagingMapCreate)(
   struct KsssPartition *partition,
   struct KsssPartitionSession *session,
   struct KsssInode *inode,
   const struct KsssPagingCallbackApi *cb_api,
   size_t page_size_bytes,
   int *inout_supports,
   const struct KsssPartitionMapApi **out_api,
   struct PartitionMap **out_map,
   KsssPageFaultFunc **out_page_fault_func,
   void **out_page_fault_handle);


/**
 * @macro KSSS_PARTITION_PAGING_API_NAME
 *
 * @brief Name of the @link "" "KsssPartitionPagingApi"
 *        KSSS partition paging interface.
 *
 * @long  This is the name that shall be used in call to
 *        @link "" "OseBaseApiInterfaceQuery" to get interface
 *        @link "" "KsssPartitionPagingApi".
 *
 * @seealso KsssPartitionPagingApi
 */
#define KSSS_PARTITION_PAGING_API_NAME \
   ("com.enea.ksss_partition_paging_api.v1")

/* Deprecated old name. Will be removed without further notice. */
#define API_TYPE_KSSS_PARTITION_PAGING_API KSSS_PARTITION_PAGING_API_NAME


/**
 * @type  KsssPartitionPagingApi
 *
 * @brief Partition paging interface
 *
 * @long  This is the interface that the file system uses for demand paging.
 *
 * @field pp_base              See @link "" "OseBaseApi"
 * @field pp_create            See @link "" "KsssPartitionPagingMapCreate"
 *
 * @seealso KSSS_PARTITION_MAP_API_NAME
 */
struct KsssPartitionPagingApi
{
   struct OseBaseApi                  pp_base;
   KsssPartitionPagingMapCreate      *pp_create;
};


/*
 ***************************************************************************
 * Partition map interface
 ***************************************************************************
 */

/**
 * @type  KsssPartitionMapMapBlock
 *
 * @brief Add a block range to a memory mapping of a file.
 *
 * @long  This function is called zero or more times after a
 *        @link "" "KsssPartitionPagingMapCreate" call, or because a mapped
 *        file has been changed.
 *
 * @input  map           A demand paging @link "" "FsMap"
 *                       created by @link "" "KsssPartitionPagingMapCreate".
 * @input  first_block   First block to add to the mapping.
 * @input  block_count   Number of consecutive blocks to add to the mapping.
 * @input  offset        Offset in bytes of this block in the file.
 *
 * @return
 *   @tab 0         @tab  if successful. @break
 *   @tab -ENOMEM   @tab  if not enough memory to perform operation. @break
 *
 * @seealso KSSS_PARTITION_PAGING_API_NAME
 */
typedef int (KsssPartitionMapMapBlock)(
   struct PartitionMap *map,
   uint64_t first_block,
   size_t block_count,
   uint64_t offset);


/**
 * @type  KsssPartitionMapUnmapBlock
 *
 * @brief Remove a block range from a memory mapping of a file.
 *
 * @long  This function is called zero or more times after a
 *        @link "" "KsssPartitionPagingMapCreate" call when a mapped
 *        file has been changed.
 *
 * @input  map           A demand paging @link "" "FsMap"
 *                       created by @link "" "KsssPartitionPagingMapCreate".
 * @input  first_block   First block to remove from the mapping.
 * @input  block_count   Number of consecutive blocks to remove from the
 *                       mapping.
 *
 * @return Nothing.
 */
typedef void (KsssPartitionMapUnmapBlock)(
   struct PartitionMap *map,
   uint64_t first_block,
   size_t block_count);


/**
 * @macro KSSS_PARTITION_MAP_API_NAME
 *
 * @brief Name of the @link "" "KsssPartitionMapApi"
 *        KSSS partition map interface.
 *
 * @long  This is the name that shall be used in call to
 *        @link "" "OseBaseApiInterfaceQuery" to get interface
 *        @link "" "KsssPartitionMapApi".
 *
 * @seealso KsssPartitionMapApi
 */
#define KSSS_PARTITION_MAP_API_NAME ("com.enea.ksss_partition_map_api.v1")

/* Deprecated old name. Will be removed without further notice. */
#define API_TYPE_KSSS_PARTITION_MAP_API KSSS_PARTITION_MAP_API_NAME


/**
 * @type  KsssPartitionMapApi
 *
 * @brief Partition map interface
 *
 * @long  This is the interface that the file system uses for demand paging.
 *
 * @field pm_base             See @link "" "OseBaseApi"
 * @field pm_map_block        See @link "" "KsssPartitionMapMapBlock"
 * @field pm_unmap_block      See @link "" "KsssPartitionMapUnmapBlock"
 *
 * @seealso KSSS_PARTITION_MAP_API_NAME
 */
struct KsssPartitionMapApi
{
   struct OseBaseApi               pm_base;
   KsssPartitionMapMapBlock       *pm_map_block;
   KsssPartitionMapUnmapBlock     *pm_unmap_block;
};


#ifdef __cplusplus
} /* end of extern "C" */
#endif


#endif /* _KSSS_PAGING_H */
