/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2006 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */
/*
 * OSE doubly linked list package.
 */
 
#ifndef _LIST_EXT_H
#define _LIST_EXT_H

#include <stddef.h>
#ifdef __cplusplus
extern "C" {
#endif

#include <stdint.h>
#ifdef STANDALONE
#include "standalone.h"
#else
#include "spinlock.h"
#endif

typedef struct _list_link_t
{
   struct _list_link_t *next;	/* Ptr to next element in list. */
   struct _list_link_t *prev;	/* Ptr to previous element in list. */
} list_link_t;

typedef struct _list_t
{
   struct _list_link_t *next;	/* Ptr to first element in list. */
   struct _list_link_t *prev;	/* Ptr to last element in list. */
   spinlock_t lock;
} list_t;

typedef int (*insert_ordered_func_t)(const list_link_t *, const list_link_t *);
typedef void *(*traverse_list_func_t)(list_link_t *lnk, void *arg);
typedef void (*assert_list_func_t)(const list_link_t *lnk);



/*
 *===========================================================================
 *                         init_list
 *===========================================================================
 *
 * Description
 * -----------
 * Initializes a doubly linked list.
 * Needs to be called before any other operation on the list.
 *
 */
void init_list(list_t *list);

/*
 *===========================================================================
 *                         list_init_link
 *===========================================================================
 *
 * Description
 * -----------
 * Initializes a list entry.
 * Needs to be called before the entry is inserted into a list.
 *
 */
void init_list_link(list_link_t *link);

/*
 *===========================================================================
 *                         delete_from_list
 *===========================================================================
 *
 * Description
 * -----------
 * Removes the specified entry from the list it is in.
 *
 */
void delete_from_list(list_t *list, list_link_t *link_to_remove);

/*
 *===========================================================================
 *                         delete_from_list_no_lock
 *===========================================================================
 *
 * Description
 * -----------
 * Removes the specified entry from the list it is in without taking a
 * spin lock or disabling interrupts.
 *
 */
void
delete_from_list_no_lock(list_link_t *link_to_remove);

/*
 *===========================================================================
 *                         insert_list_head
 *===========================================================================
 *
 * Description
 * -----------
 * Inserts the entry first into the list pointed to by 'list'.
 *
 */
void insert_list_head(list_t *list, list_link_t *link_to_insert);

/*
 *===========================================================================
 *                         insert_list_head_no_lock
 *===========================================================================
 *
 * Description
 * -----------
 * Inserts the entry first into the list pointed to by 'list' without
 * any spin lock or disabling interrupts.
 *
 */
void
insert_list_head_no_lock(list_t *list, list_link_t *link_to_insert);

/*
 *===========================================================================
 *                         insert_list_tail
 *===========================================================================
 *
 * Description
 * -----------
 * Inserts the entry last into the list pointed to by 'list'.
 *
 */
void insert_list_tail(list_t *list, list_link_t *link_to_insert);

/*
 *===========================================================================
 *                         insert_list_tail_no_lock
 *===========================================================================
 *
 * Description
 * -----------
 * Inserts the entry last into the list pointed to by 'list' without
 * aquiring any spin lock or disabling interrupts.
 *
 */
void
insert_list_tail_no_lock(list_t *list, list_link_t *link_to_insert);

/*
 *===========================================================================
 *                         insert_list_after
 *===========================================================================
 *
 * Description
 * -----------
 * Inserts the entry link_to_insert after the entry link_in_list.
 *
 */
void insert_list_after(list_t *list, list_link_t *link_in_list,
                       list_link_t *link_to_insert);

/*
 *===========================================================================
 *                         insert_list_after_no_lock
 *===========================================================================
 *
 * Description
 * -----------
 * Inserts the entry link_to_insert after the entry link_in_list
 * without aquiring any spin lock or disabling interrupts.
 *
 */
void
insert_list_after_no_lock(list_link_t *link_in_list,
                          list_link_t *link_to_insert);

/*
 *===========================================================================
 *                         insert_list_before
 *===========================================================================
 *
 * Description
 * -----------
 * Inserts the entry link_to_insert before the entry link_in_list.
 *
 */
void insert_list_before(list_t *list, list_link_t *link_in_list,
                        list_link_t *link_to_insert);

/*
 *===========================================================================
 *                         insert_list_before_no_lock
 *===========================================================================
 *
 * Description
 * -----------
 * Inserts the entry link_to_insert before the entry link_in_list
 * without aquiring any spin lock or disabling interrupts.
 *
 */
void
insert_list_before_no_lock(list_link_t *link_in_list,
                           list_link_t *link_to_insert);

/*
 *===========================================================================
 *                         list_head
 *===========================================================================
 *
 * Description
 * -----------
 * Returns the first entry in the list.
 *
 */
list_link_t * list_head(const list_t *list);
#define list_head(list) (((list)->next != (const list_link_t *)(list)) ? ((list)->next) : NULL)

/*
 *===========================================================================
 *                         list_tail
 *===========================================================================
 *
 * Description
 * -----------
 * Returns the last entry in the list.
 *
 */
list_link_t * list_tail(const list_t *list);
#define list_tail(list) (((list)->prev != (const list_link_t *)(list)) ? ((list)->prev) : NULL)

/*
 *===========================================================================
 *                         list_next
 *===========================================================================
 *
 * Description
 * -----------
 * Returns the next entry in the list.
 *
 */
list_link_t * list_next(const list_t *list, const list_link_t *lnk);
#define list_next(list, lnk) ((((const list_link_t *)(lnk))->next != (const list_link_t *)(list)) ? (((const list_link_t *)(lnk))->next) : NULL)

/*
 *===========================================================================
 *                         list_prev
 *===========================================================================
 *
 * Description
 * -----------
 * Returns the previous entry in the list.
 *
 */
list_link_t * list_prev(const list_t *list, const list_link_t *lnk);
#define list_prev(list, lnk) ((((const list_link_t *)(lnk))->prev != (const list_link_t *)(list)) ? (((const list_link_t *)(lnk))->prev) : NULL)

/*
 *===========================================================================
 *                         length_of_list
 *===========================================================================
 *
 * Description
 * -----------
 * Returns the number of entries in the list.
 *
 */
uint32_t
length_of_list(const list_t *list);

/*
 *===========================================================================
 *                         insert_list_ordered
 *===========================================================================
 *
 * Description
 * -----------
 * Inserts the entry ll before the entry for which the compare function is
 * positive.
 *
 */
void
insert_list_ordered(list_t *list, list_link_t *link_to_insert,
                    insert_ordered_func_t compare);

/*
 *===========================================================================
 *                         traverse_list
 *===========================================================================
 *
 * Description
 * -----------
 * Traverses the list and calls the apply function for each entry.
 *
 */
void *
traverse_list(list_t *list, traverse_list_func_t apply, void *arg);

/*
 *===========================================================================
 *                         assert_list
 *===========================================================================
 *
 * Description
 * -----------
 * Checks that each entry in the list is intact and that the compare function
 * is valid for each entry.
 *
 */
#ifdef ASSERT
void
assert_list(const list_t *list, assert_list_func_t check);
#endif


#ifdef __cplusplus
}
#endif

#endif /* LIST_EXT_H */

