/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2006 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */
/**
 * @public
 * @toc OSE:
 * @file ose_spi/mman_pager.h
 *
 * @brief Mman function interface accessible from Core Monolith link scope.
 *
 * @long
 * This file defines a functional interface for the Mman component.
 * A pager can be installed, vm area objects can be created and destroyed.
 *
 */

#ifndef OSE_MMAN_PAGER_H_
#define OSE_MMAN_PAGER_H_

#include "ose.h"
#include "string.h"
#include "mm.sig"
#include "ksss_paging.h"
#include "demand_pager.h"


#define NOT_LOCKED 0
#define REGULAR_MLOCK 1
#define COUNTED_MLOCK 2

/* The possible operations for the pager control function used in testing. */
typedef enum pager_control_op_t
{
   PAGER_INVALIDATE_MAPPING = 0,
   PAGER_CLEAN_CACHE = 1,
   PAGER_SET_CHUNK_SIZE = 2,
   PAGER_SET_MIN_SIZE = 3,
   PAGER_SET_MAX_SIZE = 4,
   PAGER_SET_PHYS_MEM = 5,
   PAGER_SET_DISALLOW_LOCKED_PAGES = 6,
   PAGER_SET_REPLACEMENT_ALG = 7,
   PAGER_SET_NO_OF_STACKS = 8,
   PAGER_SET_TEMPORARY_MAPPING = 9,
   PAGER_IS_ENTRY_LOCKED = 10,
   PAGER_INSTALL_HANDLER = 11,
   PAGER_UNINSTALL_HANDLER = 12,
   PAGER_RECLAIM_CACHE = 13
} pager_control_op_t;

#ifdef __cplusplus
extern "C" {
#endif

/* Forward declaration. To be defined by the pager. */
struct page_cache_t;

/**
 * @type  vm_area_t
 *
 * @brief Struct describing the connection between a file descriptor and
 *        a piece of logical memory.
 *
 * @field domain The domain of the vm area.
 *
 * @field base   The logical start address of the vm area.
 *
 * @field size   The size of the vm area.
 *
 * @field permissions The MM permissions of the vm area.
 *
 * @field cache_mode The cache modes of the vm area.
 *
 * @field page_map   The page map to read from when getting a page fault in the
 *                   vm area.
 *
 * @field file_offset The offset in the inode, marking out where to start
 *                    reading.
 *
 * @field page_cache An opaque pointer to the page cache associated with
 *                   the vm area.
 *
 * @field mapping_type Logical memory partition type (SASE/SAS/MAS)
 *
 */
/* One entry for each vm area registered in the pager. */
struct vm_area_t
{
   OSDOMAIN             domain;
   log_addr_t           base;
   size_t               size;
   uint32_t             permissions;
   uint32_t             cache_mode;
   struct MmanPageMap  *page_map;
   uint64_t             file_offset;
   struct page_cache_t *page_cache;
   uint8_t              mapping_type;
};

/**
 * @type  pager_lock_func_t
 *
 * @brief Locks a page from the page cache.
 *
 * @input domain    The domain of the area to be locked.
 * @input base      The start address of the area to be locked.
 * @input size      The size of the area that should be locked.
 * @input lock      Set to LOCKED or LOCKED_COUNTED depending on what type of
 *                  mlock that is desired. If LOCKED_COUNTED, N number of mlocks
 *                  to a page needs N munlocks to unlock the page.
 *                  If LOCKED, one munlock will unlock the page independent
 *                  of how many times mlock was called.
 * @input optimize  If the location of the mlocked pages should be
 *                  optimized or not.
 *
 * @return pager_status_t
 */
typedef pager_status_t(pager_lock_func_t)(OSDOMAIN domain,
                                          log_addr_t base,
                                          size_t size,
                                          uint8_t lock,
                                          uint8_t optimize);

/**
 * @type  pager_unlock_func_t
 *
 * @brief Unlocks a page from the page cache.
 *
 * @input domain    The domain of the area to be unlocked.
 * @input base      The start address of the area to be unlocked.
 * @input size      The size of the area that should be unlocked.
 * @input counting  The munlock should be counted. If so, N number of mlocks
 *                  to a page needs N munlocks to unlock the page.
 *                  If not set, one munlock
 *                  will unlock the page independent of how many times
 *                  mlock was called.
 *
 * @return pager_status_t
 */
typedef pager_status_t(pager_unlock_func_t)(OSDOMAIN domain,
                                            log_addr_t base,
                                            size_t size,
                                            uint8_t counting);

/**
 * @type  pager_invalidate_func_t
 *
 * @brief Called by mman when filesystem requests invalidation of
 *        mapped pages. 
 *        
 * @input page_map  Handle that mman supplies. 
 * @input offset    Starting file offset in bytes of range to invalidate.
 * @input size      Size in bytes of file offset range to invalidate.
 * @input new_data  Address to where pager can copy the new data, or 0 if 
 *                  remapping.
 * @input new_phys_addr  If remapping, this is the new address for the page to be
 *                  remapped. The address corresponds to the offset. If not
 *                  remapping, this field must be 0.
 * @input 
 *
 * @return pager_status_t
 */
typedef void (pager_invalidate_func_t)
      (struct MmanPageMap *page_map, uint64_t offset,
       size_t size, log_addr_t new_data, phys_addr_t new_phys_addr);


/**
 * @type pager_unmap_vma_func_t
 *
 * @brief Called by mman when a memory mapped file is unmapped.
 *
 * @long  The pager should remove and free all resources
 *        associated with the memory mapped area. All pages
 *        must be unmapped and all references to the vm_area_t
 *        structure must be destroyed.
 *
 * @input page_map  Handle that mman supplies. Identifies inode.
 *
 * @input vm_area   The virtual memory area that should be removed.
 *
 */
typedef void (pager_unmap_vma_func_t)(struct MmanPageMap *page_map,
                                      struct vm_area_t *vm_area);


/**
 * @type  pager_control_func_t
 *
 * @brief A catch all pager control function that is used by tests
 *        to configure, flush and torture the pager.
 *
 * @input operation  The pager control operation to be performed.
 * @input arg1       Argument that depends on the operation.
 * @input arg2       Argument that depends on the operation.
 * @input arg3       Argument that depends on the operation.
 * @input arg4       Argument that depends on the operation.
 * @input arg5       Argument that depends on the operation.
 *
 * @return pager_status_t
 */
typedef pager_status_t(pager_control_func_t)(pager_control_op_t operation,
                                             uint32_t arg1,
                                             uint32_t arg2,
                                             uint32_t arg3,
                                             uint32_t arg4,
                                             uint32_t arg5);

/**
 * @type  MmanStatus
 *
 * @brief Value describing the status of an Mman operation.
 *
 * @field MMAN_SUCCESS Mman operation completed successfully.
 *
 * @field MMAN_REGION_OVERLAPPING The supplied mman vm_area is overlapping with
 *        an already registered vm_area.
 *
 * @field MMAN_PAGER_ALREADY_INSTALLED The pager has already been installed.
 *
 * @field MMAN_PAGER_NOT_INSTALLED The pager is not installed.
 *
 * @field MMAN_OUT_OF_MEMORY The mman module or the installed pager is
 *                           out of memory.
 *
 * @field MMAN_INVALID_CALLBACK_FUNCTION One or more of the callback
 *                                       functions are invalid.
 */
enum MmanStatus
{
   MMAN_SUCCESS = 0,
   MMAN_REGION_OVERLAPPING = 1,
   MMAN_PAGER_ALREADY_INSTALLED = 2,
   MMAN_PAGER_NOT_INSTALLED = 3,
   MMAN_OUT_OF_MEMORY = 4,
   MMAN_INVALID_CALLBACK_FUNCTION = 5
};

typedef enum MmanStatus mman_status_t;

/**
 * @function ose_install_pager
 *
 * @brief  Installs a pager.
 *
 * @long   Install a pager for demand paging.
 *
 * @input lock_func   The function in the pager implementing the mlock()
 *                    functionality.
 * @input unlock_func The function in the pager implementing the munlock()
 *                    functionality.
 * @input invalidate_func  The function in the pager implementing the
 *                         functionality that invalidates mappings if
 *                         a physical page has moved.
 * @input alloc_page_func  The function in the pager that allocates a page
 *                         in the page cache. Used by the file system in
 *                         copy page mode.
 * @input unmap_vma_func   The function in the pager that removes and frees
 *                         all resources associated with a virtual memory area.
 *                         Called by mman when a memory mapped file is unmapped.
 * @input control_func     The pager control function used to manipulate
 *                         and configure the pager during tests.
 * @return           0 if ok, otherwise error code.
 */
mman_status_t
ose_install_pager(pager_lock_func_t lock_func,
                  pager_unlock_func_t unlock_func,
                  pager_invalidate_func_t invalidate_func,
                  KsssPagerAllocPage alloc_page_func,
                  pager_unmap_vma_func_t unmap_vma_func,
                  pager_control_func_t control_func);

/**
 * @function ose_uninstall_pager
 *
 * @brief  Uninstalls a pager.
 *
 * @long   Uninstall a pager for demand paging.
 *
 * @return           0 if ok, otherwise an error code.
 */
mman_status_t
ose_uninstall_pager(void);


/**
 * @function ose_find_vma
 *
 * @brief  Gets an mman vma object.
 *
 * @long   Gets an mman vma object
 *
 * @input  domain    The domain.
 * @input  address   The logical address
 * @return           The vma object registered for the logical address in
 *                   the domain supplied, or NULL if not found
 */
struct vm_area_t*
ose_find_vma(OSDOMAIN domain, log_addr_t address);

/**
 * @function find_overlapping_vma
 *
 * @brief  Gets an overlapping mman vma object, if one exists.
 *
 * @long   Gets an overlapping mman vma object, if one exists.
 *
 * @input  domain    The domain
 * @input  address   The logical address
 * @input  size      The size of the region that should be checked
 * @return           The vma object registered for the logical address in
 *                   the domain supplied, or NULL if not found
 */
struct vm_area_t*
find_overlapping_vma(OSDOMAIN domain, log_addr_t address, size_t size);

/**
 * @function get_first_vma
 *
 * @brief  Gets the first vma from mman..
 *
 * @long   Gets the first vma from mman.
 *
 * @return The vma object registered for the logical address in
 *         the domain supplied, or NULL if not found
 */
struct vm_area_t *
get_first_vma(void);

/**
 * @function ose_is_pager_installed
 *
 * @brief  Probes if the pager is installed.
 *
 * @long   Probes if the pager is installed.
 *
 * @return           1 if the pager is installed, otherwise 0.
 */
int
ose_is_pager_installed(void);

/**
 * @function mman_mlock
 *
 * @brief  Lock address range in page cache.
 *
 * @long   This function implements @link "mman.h" "mlock", but lies after
 *         the trap.
 *
 * @input  addr   The logical address
 * @input  len    The size of the region that should be locked.
 *
 * @return Returns zero on success and -1 on failure and sets errno.
 */
int
mman_mlock(log_addr_t addr, size_t len);

/**
 * @function mman_munlock
 *
 * @brief Unlock address range in page cache.
 *
 * @long  This function implements @link "mman.h" "munlock", but lies after
 *        the trap.
 *
 * @input  addr   The logical address
 * @input  len    The size of the region that should be unlocked.
 *
 * @return Returns zero on success and -1 on failure and sets errno.
 */
int
mman_munlock(log_addr_t addr, size_t len);

/*
 *===========================================================================
 *                      ose_mman_mlock_temporary
 *===========================================================================
 *
 * Description
 * -----------
 * Implements an internal mlock functionality only to be used temporarily.
 * This version is much faster then the regular mlock, but will not try
 * to optimize the location of the mlocked page.
 * Locks all pages in the range [addr, addr+len) from the page cache.
 *
 * Returns
 * -------
 * int        Returns zero on success and -1 on failure and sets errno.
 *
 */
int
ose_mman_mlock_temporarily(log_addr_t addr, size_t len);

/*
 *===========================================================================
 *                     ose_mman_munlock_temporarily
 *===========================================================================
 *
 * Description
 * -----------
 * Implements the munlock functionality. Unlocks all pages in the range
 * [addr, addr+len) from the page cache. Accepts non paged aligned
 * addresses.
 *
 * Returns
 * -------
 * int        Returns zero on success and -1 on failure and sets errno.
 *
 */
int
ose_mman_munlock_temporarily(log_addr_t addr, size_t len);


/*
 *===========================================================================
 *                        ose_mman_mlock_counting
 *===========================================================================
 *
 * Description
 * -----------
 * Implements the mlock functionality. Locks all pages in the range
 * [addr, addr+len) from the page cache.
 * Address does not need to be page aligned, and the mlock will
 * have a count associated with it. Every time a page is mlocked, the
 * count is increased by one.
 * Returns
 * -------
 * int        Returns zero on success and -1 on failure and sets errno.
 *
 */
int
ose_mman_mlock_counting(log_addr_t addr, size_t len);

/*
 *===========================================================================
 *                     ose_mman_munlock_counting
 *===========================================================================
 *
 * Description
 * -----------
 * Implements the munlock functionality. Unlocks all pages in the range
 * [addr, addr+len) from the page cache. Accepts non paged aligned
 * addresses. This lock will
 * have a count associated with it. Every time a page is munlocked, the
 * count is decreased by one, and it is not unlocked until it reaches 0.
 *
 * Returns
 * -------
 * int        Returns zero on success and -1 on failure and sets errno.
 *
 */
int
ose_mman_munlock_counting(log_addr_t addr, size_t len);

#ifdef __cplusplus
}
#endif

#endif /* #ifndef OSE_MMAN_PAGER_H_ */
