/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2010 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

/**
 * @public
 *
 * @toc OSE:
 *
 * @file rmd.sig
 *
 * @brief Run Mode Dumper private signal Interface
 *
 * @long This file defines the private signal interface to the Run Mode Dumper
 *       process.
 *
 */

#ifndef _RMD_SIG
#define _RMD_SIG

#include "ose.h"
#include "cpu_regdump.h"
#include "stdint.h"
#include "rmd.h"

/**
 * @type RmdStatusReply
 *
 * @brief Generic type that is used for reply signals that only need status.
 *
 * @output signo                 RMD_START_DUMP_REPLY
 *                               RMD_STOP_DUMP_REPLY
 *                               RMD_DUMP_ERROR_REPLY
 *                               RMD_DUMP_MEMORY_REPLY
 *                               RMD_DUMP_STRING_REPLY
 *                               RMD_DUMP_PROCESS_INFO_REPLY
 *                               RMD_DUMP_STACK_USAGE_INFO_REPLY
 *                               RMD_DUMP_ENV_VARS_INFO_REPLY
 *                               RMD_DUMP_SIGNAL_QUEUE_INFO_REPLY
 *                               RMD_DUMP_POOL_INFO_REPLY
 *                               RMD_DUMP_POOL_SIGNALS_INFO_REPLY
 *                               RMD_DUMP_SYS_PARAM_INFO_REPLY
 *                               RMD_DUMP_REGISTERS_INFO_REPLY
 *                               RMD_DUMP_PROGRAM_INFO_REPLY
 *                               RMD_DUMP_TRACE_INFO_REPLY
 *                               RMD_DUMP_LOAD_MODULE_INFO_REPLY
 *                               RMD_DUMP_POOL_MEMORY_REPLY
 *                               RMD_DUMP_HEAP_MEMORY_REPLY
 *                               RMD_DUMP_PROGRAM_MEMORY_REPLY
 *
 * @output status                Status of the corresponding request
 */
typedef struct
{
   SIGSELECT signo;
   RmdStatus status;
} RmdStatusReply;

/**
 * @signal RmdInterfaceRequest
 *
 * @brief List RMD interface information request (before use).
 *
 * @long This signal should be the first signal sent from the client to the
 * RMD, to obtain basic information about the RMD interface, before sending
 * other requests to the RMD.
 *
 * @member RMD_INTERFACE_REQUEST
 * @member RmdInterfaceRequest
 *
 * @seealso RmdInterfaceReply
 */

/** @macro RMD_INTERFACE_REQUEST */
#define RMD_INTERFACE_REQUEST 33150 /* !-SIGNO(RmdInterfaceRequest)-! */

/**
 * @type RmdInterfaceRequest
 *
 * @input signo                 RMD_INTERFACE_REQUEST
 */
typedef struct RmdInterfaceRequest
{
   SIGSELECT signo;
} RmdInterfaceRequest;

/**
 * @signal RmdInterfaceReply
 *
 * @brief List RMD interface information reply.
 *
 * @member RMD_INTERFACE_REPLY
 * @member RmdInterfaceReply
 *
 * @seealso MonitorInterfaceRequest
 */

/** @macro RMD_INTERFACE_REPLY */
#define RMD_INTERFACE_REPLY 33151

/**
 * @type RmdInterfaceReply
 *
 * @output signo                 RMD_INTERFACE_REPLY
 * @output sigsCount             Number of signals numbers in sigs.
 * @output sigs                  Array of supported request signal numbers.
 * @output reserved0  NOT USED
 * @output reserved1  NOT USED
 * @output reserved2  NOT USED
 */
typedef struct RmdInterfaceReply
{
   uint32_t signo;
   uint32_t reserved0;
   uint32_t reserved1;
   uint32_t reserved2;
   uint32_t sigsCount;
   uint32_t sigs[1];
} RmdInterfaceReply;

/**
 * @signal RmdStartDumpRequest
 *
 * @brief Start and create a new run mode dump file.
 *
 * @long The new dump will be associated with the sending process. If that
 * process is killed the dump will be closed.
 *
 * @member RMD_START_DUMP_REQUEST
 * @member RmdStartDumpRequest
 *
 */

/** @macro RMD_START_DUMP_REQUEST */
#define RMD_START_DUMP_REQUEST 33152 /* !-SIGNO(RmdStartDumpRequest)-! */

/**
 * @type RmdStartDumpRequest
 *
 * @input signo                 RMD_START_DUMP_REQUEST
 * @input dump_pid   The process to create the dump for or zero if sender pid
 *                   is to be used.
 * @input reserved0  NOT USED Must be set to 0
 * @input reserved1  NOT USED Must be set to 0
 * @input reserved2  NOT USED Must be set to 0
 *
 * @input path       A string containing the full path to where the dump should
 *                   be stored.
 *
 */
typedef struct RmdStartDumpRequest
{
   SIGSELECT signo;
   uint32_t dump_pid;
   uint32_t reserved0;
   uint32_t reserved1;
   uint32_t reserved2;
   char path[1]; /* Variable length */
} RmdStartDumpRequest;

/**
 * @signal RmdStartDumpReply
 *
 * @brief Reply with status from RmdStartDumpRequest.
 *
 * @member RMD_START_DUMP_REPLY
 * @member RmdStatusReply
 *
 */

/** @macro RMD_START_DUMP_REPLY */
#define RMD_START_DUMP_REPLY (33153) /* !-SIGNO(RmdStatusReply)-! */

/**
 * @signal RmdStopDumpRequest
 *
 * @brief Stop and close a current run mode dump file.
 *
 * @long This signal is normally sent after an error handler has finished
 * dumping a program.
 *
 * @member RMD_STOP_DUMP_REQUEST
 * @member RmdStopDumpRequest
 *
 */

/** @macro RMD_STOP_DUMP_REQUEST */
#define RMD_STOP_DUMP_REQUEST 33154 /* !-SIGNO(RmdStopDumpRequest)-! */

/**
 * @type RmdStopDumpRequest
 *
 * @input signo                 RMD_STOP_DUMP_REQUEST
 * @input reserved0  NOT USED Must be set to 0
 */
typedef struct RmdStopDumpRequest
{
   SIGSELECT signo;
   uint32_t reserved0;
} RmdStopDumpRequest;

/**
 * @signal RmdStopDumpReply
 *
 * @brief Reply with status from RmdStopDumpRequest.
 *
 * @member RMD_STOP_DUMP_REPLY
 * @member RmdStatusReply
 *
 */

/** @macro RMD_STOP_DUMP_REPLY */
#define RMD_STOP_DUMP_REPLY (33155) /* !-SIGNO(RmdStatusReply)-! */

/**
 * @signal RmdDumpErrorRequest
 *
 * @brief  Store error information in the current dump.
 *
 * @long This should be one of the first signals sent when creating a
 * dump after a fatal error.
 *
 * @member RMD_DUMP_ERROR_REQUEST
 * @member RmdDumpErrorRequest
 *
 */

/** @macro RMD_DUMP_ERROR_REQUEST */
#define RMD_DUMP_ERROR_REQUEST 33156 /* !-SIGNO(RmdDumpErrorRequest)-! */

/**
 * @type RmdDumpErrorRequest
 *
 * @input signo RMD_DUMP_ERROR_REQUEST
 * @input user_called Error handler parameter
 * @input ecode Error handler parameter
 * @input extra Error handler parameter
 * @input reserved0  NOT USED Must be set to 0
 * @input reserved1  NOT USED Must be set to 0
 * @input reserved2  NOT USED Must be set to 0
 * @input reserved3  NOT USED Must be set to 0
 * @input description String describing the error or NULL.
 *
 */
typedef struct RmdDumpErrorRequest
{
   SIGSELECT signo;
   uint32_t user_called;
   uint32_t ecode;
   uint32_t extra;
   uint32_t reserved0;
   uint32_t reserved1;
   uint32_t reserved2;
   uint32_t reserved3;
   char description[1]; /* Variable length */
} RmdDumpErrorRequest;

/**
 * @signal RmdDumpErrorReply
 *
 * @brief Reply with status from RmdDumpErrorRequest.
 *
 * @member RMD_DUMP_ERROR_REPLY
 * @member RmdStatusReply
 *
 */

/** @macro RMD_DUMP_ERROR_REPLY */
#define RMD_DUMP_ERROR_REPLY (33157) /* !-SIGNO(RmdStatusReply)-! */

/**
 * @signal RmdDumpMemoryRequest
 *
 * @brief  Store a region of RAM in the current dump.
 *
 * @long The description string is used in host tools to match the saved
 * blocks to plain text formatters of the saved data. This allows application
 * specific data structures to be stored and interpreted on the host without
 * the need to implement parsers for the entire .pmd file format.
 *
 * @member RMD_DUMP_MEMORY_REQUEST
 * @member RmdDumpMemoryRequest
 *
 */

/** @macro RMD_DUMP_MEMORY_REQUEST */
#define RMD_DUMP_MEMORY_REQUEST 33158 /* !-SIGNO(RmdDumpMemoryRequest)-! */

/**
 * @type RmdDumpMemoryRequest
 *
 * @input signo RMD_DUMP_MEMORY_REQUEST
 * @input start_address Start of the region to save
 * @input length Length of the region to save
 * @input reserved0  NOT USED Must be set to 0
 * @input reserved1  NOT USED Must be set to 0
 * @input reserved2  NOT USED Must be set to 0
 * @input reserved3  NOT USED Must be set to 0
 * @input description String describing the data or NULL.
 *
 */
typedef struct RmdDumpMemoryRequest
{
   SIGSELECT signo;
   uint32_t start_address;
   uint32_t length;
   uint32_t reserved0;
   uint32_t reserved1;
   uint32_t reserved2;
   uint32_t reserved3;
   char description[1]; /* Variable length */
} RmdDumpMemoryRequest;

/**
 * @signal RmdDumpMemoryReply
 *
 * @brief Reply with status from RmdDumpMemoryRequest.
 *
 * @member RMD_DUMP_MEMORY_REPLY
 * @member RmdStatusReply
 *
 */

/** @macro RMD_DUMP_MEMORY_REPLY */
#define RMD_DUMP_MEMORY_REPLY (33159) /* !-SIGNO(RmdStatusReply)-! */

/**
 * @signal RmdDumpStringRequest
 *
 * @brief  Store a string with plain text information in the current dump.
 *
 * @member RMD_DUMP_STRING_REQUEST
 * @member RmdDumpStringRequest
 *
 */

/** @macro RMD_DUMP_STRING_REQUEST */
#define RMD_DUMP_STRING_REQUEST 33160 /* !-SIGNO(RmdDumpStringRequest)-! */

/**
 * @type RmdDumpStringRequest
 *
 * @input signo RMD_DUMP_STRING_REQUEST
 * @input reserved0  NOT USED Must be set to 0
 * @input reserved1  NOT USED Must be set to 0
 * @input reserved2  NOT USED Must be set to 0
 * @input reserved3  NOT USED Must be set to 0
 * @input description String describing the data or NULL.
 *
 */
typedef struct RmdDumpStringRequest
{
   SIGSELECT signo;
   uint32_t reserved0;
   uint32_t reserved1;
   uint32_t reserved2;
   uint32_t reserved3;
   char description[1]; /* Variable length */
} RmdDumpStringRequest;

/**
 * @signal RmdDumpStringReply
 *
 * @brief Reply with status from RmdDumpStringRequest.
 *
 * @member RMD_DUMP_STRING_REPLY
 * @member RmdStatusReply
 *
 */

/** @macro RMD_DUMP_STRING_REPLY */
#define RMD_DUMP_STRING_REPLY (33161) /* !-SIGNO(RmdStatusReply)-! */

/**
 * @signal RmdDumpProcessInfoRequest
 *
 * @brief Store process, block, and segment information in the current dump.
 *
 * @long This should be one of the first signals sent (after RmdDumpErrorRequest)
 * when creating a dump after a fatal error, to ensure that the process control
 * blocks of the executive are intact, so that the scopes and filter mechanisms
 * of subsequent RmdDump*Info or RmdDump*Memory signals are functional.
 *
 * @long If none of the relevant bits (RMD_INFO_PROCESS and
 * RMD_INFO_PROCESS_LONG) are set in info, this function does nothing.
 *
 * @member RMD_DUMP_PROCESS_INFO_REQUEST
 * @member RmdDumpProcessInfoRequest
 *
 */

/** @macro RMD_DUMP_PROCESS_INFO_REQUEST */
#define RMD_DUMP_PROCESS_INFO_REQUEST 33162 /* !-SIGNO(RmdDumpProcessInfoRequest)-! */

/**
 * @type RmdDumpProcessInfoRequest
 *
 * @input signo RMD_DUMP_PROCESS_INFO_REQUEST
 * @input scope The set of processes for which to store information
 * @input info The kind of information to store
 *
 */
typedef struct RmdDumpProcessInfoRequest
{
   SIGSELECT signo;
   uint32_t scope;
   uint32_t info;
} RmdDumpProcessInfoRequest;

/**
 * @signal RmdDumpProcessInfoReply
 *
 * @brief Reply with status from RmdDumpProcessInfoRequest.
 *
 * @member RMD_DUMP_PROCESS_INFO_REPLY
 * @member RmdStatusReply
 *
 */

/** @macro RMD_DUMP_PROCESS_INFO_REPLY */
#define RMD_DUMP_PROCESS_INFO_REPLY (33163) /* !-SIGNO(RmdStatusReply)-! */

/**
 * @signal RmdDumpStackUsageInfoRequest
 *
 * @brief Store stack usage information in the current dump.
 *
 * @long If none of the relevant bits (RMD_INFO_STACK_USAGE) are set in info,
 * the RMD process does nothing.
 *
 * @member RMD_DUMP_STACK_USAGE_INFO_REQUEST
 * @member RmdDumpStackUsageInfoRequest
 *
 */

/** @macro RMD_DUMP_STACK_USAGE_INFO_REQUEST */
#define RMD_DUMP_STACK_USAGE_INFO_REQUEST 33164 /* !-SIGNO(RmdDumpStackUsageInfoRequest)-! */

/**
 * @type RmdDumpStackUsageInfoRequest
 *
 * @input signo RMD_DUMP_STACK_USAGE_INFO_REQUEST
 * @input scope The set of processes for which to store information
 * @input info The kind of information to store
 *
 */
typedef struct RmdDumpStackUsageInfoRequest
{
   SIGSELECT signo;
   uint32_t scope;
   uint32_t info;
} RmdDumpStackUsageInfoRequest;

/**
 * @signal RmdDumpStackUsageInfoReply
 *
 * @brief Reply with status from RmdDumpStackUsageInfoRequest.
 *
 * @member RMD_DUMP_STACK_USAGE_INFO_REPLY
 * @member RmdStatusReply
 *
 */

/** @macro RMD_DUMP_STACK_USAGE_INFO_REPLY */
#define RMD_DUMP_STACK_USAGE_INFO_REPLY (33165) /* !-SIGNO(RmdStatusReply)-! */

/**
 * @signal RmdDumpEnvVarsInfoRequest
 *
 * @brief Store environment variable information in the current dump
 *
 * @long If none of the relevant bits (RMD_INFO_ENV_VARS) are set in info,
 * the RMD process does nothing.
 *
 * @member RMD_DUMP_ENV_VARS_INFO_REQUEST
 * @member RmdDumpEnvVarsInfoRequest
 *
 */

/** @macro RMD_DUMP_ENV_VARS_INFO_REQUEST */
#define RMD_DUMP_ENV_VARS_INFO_REQUEST 33166 /* !-SIGNO(RmdDumpEnvVarsInfoRequest)-! */

/**
 * @type RmdDumpEnvVarsInfoRequest
 *
 * @input signo RMD_DUMP_ENV_VARS_INFO_REQUEST
 * @input scope The set of processes for which to store information
 * @input info The kind of information to store
 *
 */
typedef struct RmdDumpEnvVarsInfoRequest
{
   SIGSELECT signo;
   uint32_t scope;
   uint32_t info;
} RmdDumpEnvVarsInfoRequest;

/**
 * @signal RmdDumpEnvVarsInfoReply
 *
 * @brief Reply with status from RmdDumpEnvVarsInfoRequest.
 *
 * @member RMD_DUMP_ENV_VARS_INFO_REPLY
 * @member RmdStatusReply
 *
 */

/** @macro RMD_DUMP_ENV_VARS_INFO_REPLY */
#define RMD_DUMP_ENV_VARS_INFO_REPLY (33167) /* !-SIGNO(RmdStatusReply)-! */

/**
 * @signal RmdDumpSignalQueueInfoRequest
 *
 * @brief Store signal queue information in the current dump
 *
 * @long If none of the relevant bits (RMD_INFO_SIGNAL_QUEUE) are set in info,
 * the RMD process does nothing.
 *
 * @member RMD_DUMP_SIGNAL_QUEUE_INFO_REQUEST
 * @member RmdDumpSignalQueueInfoRequest
 *
 */

/** @macro RMD_DUMP_SIGNAL_QUEUE_INFO_REQUEST */
#define RMD_DUMP_SIGNAL_QUEUE_INFO_REQUEST 33168 /* !-SIGNO(RmdDumpSignalQueueInfoRequest)-! */

/**
 * @type RmdDumpSignalQueueInfoRequest
 *
 * @input signo RMD_DUMP_SIGNAL_QUEUE_INFO_REQUEST
 * @input scope The set of processes for which to store information
 * @input info The kind of information to store
 *
 */
typedef struct RmdDumpSignalQueueInfoRequest
{
   SIGSELECT signo;
   uint32_t scope;
   uint32_t info;
} RmdDumpSignalQueueInfoRequest;

/**
 * @signal RmdDumpSignalQueueInfoReply
 *
 * @brief Reply with status from RmdDumpSignalQueueInfoRequest.
 *
 * @member RMD_DUMP_SIGNAL_QUEUE_INFO_REPLY
 * @member RmdStatusReply
 *
 */

/** @macro RMD_DUMP_SIGNAL_QUEUE_INFO_REPLY */
#define RMD_DUMP_SIGNAL_QUEUE_INFO_REPLY (33169) /* !-SIGNO(RmdStatusReply)-! */

/**
 * @signal RmdDumpPoolInfoRequest
 *
 * @brief Store pool information in the current dump
 *
 * @long If none of the relevant bits (RMD_INFO_POOL_INFO) are set in info,
 * the RMD process does nothing.
 *
 * @member RMD_DUMP_POOL_INFO_REQUEST
 * @member RmdDumpPoolInfoRequest
 *
 */

/** @macro RMD_DUMP_POOL_INFO_REQUEST */
#define RMD_DUMP_POOL_INFO_REQUEST 33170 /* !-SIGNO(RmdDumpPoolInfoRequest)-! */

/**
 * @type RmdDumpPoolInfoRequest
 *
 * @input signo RMD_DUMP_POOL_INFO_REQUEST
 * @input scope The set of processes for which to store information
 * @input info The kind of information to store
 *
 */
typedef struct RmdDumpPoolInfoRequest
{
   SIGSELECT signo;
   uint32_t scope;
   uint32_t info;
} RmdDumpPoolInfoRequest;

/**
 * @signal RmdDumpPoolInfoReply
 *
 * @brief Reply with status from RmdDumpPoolInfoRequest.
 *
 * @member RMD_DUMP_POOL_INFO_REPLY
 * @member RmdStatusReply
 *
 */

/** @macro RMD_DUMP_POOL_INFO_REPLY */
#define RMD_DUMP_POOL_INFO_REPLY (33171) /* !-SIGNO(RmdStatusReply)-! */

/**
 * @signal RmdDumpPoolSignalsInfoRequest
 *
 * @brief Store pool information in the current dump
 *
 * @long If none of the relevant bits (RMD_INFO_POOL_SIGNALS,
 * RMD_INFO_POOL_SIGNALS_OWNED) are set in info, the RMD process does nothing.
 *
 * @member RMD_DUMP_POOL_SIGNALS_INFO_REQUEST
 * @member RmdDumpPoolSignalsInfoRequest
 *
 */

/** @macro RMD_DUMP_POOL_SIGNALS_INFO_REQUEST */
#define RMD_DUMP_POOL_SIGNALS_INFO_REQUEST 33172 /* !-SIGNO(RmdDumpPoolSignalsInfoRequest)-! */

/**
 * @type RmdDumpPoolSignalsInfoRequest
 *
 * @input signo RMD_DUMP_POOL_SIGNALS_INFO_REQUEST
 * @input scope The set of processes for which to store information
 * @input info The kind of information to store
 *
 */
typedef struct RmdDumpPoolSignalsInfoRequest
{
   SIGSELECT signo;
   uint32_t scope;
   uint32_t info;
} RmdDumpPoolSignalsInfoRequest;

/**
 * @signal RmdDumpPoolSignalsInfoReply
 *
 * @brief Reply with status from RmdDumpPoolSignalsInfoRequest.
 *
 * @member RMD_DUMP_POOL_SIGNALS_INFO_REPLY
 * @member RmdStatusReply
 *
 */

/** @macro RMD_DUMP_POOL_SIGNALS_INFO_REPLY */
#define RMD_DUMP_POOL_SIGNALS_INFO_REPLY (33173) /* !-SIGNO(RmdStatusReply)-! */

/**
 * @signal RmdDumpSysParamInfoRequest
 *
 * @brief Store system parameter information in the current dump
 *
 * @long If none of the relevant bits (RMD_INFO_SYS_PARAM) are set in info,
 * the RMD process does nothing.
 *
 * @member RMD_DUMP_SYS_PARAM_INFO_REQUEST
 * @member RmdDumpSysParamInfoRequest
 *
 */

/** @macro RMD_DUMP_SYS_PARAM_INFO_REQUEST */
#define RMD_DUMP_SYS_PARAM_INFO_REQUEST 33174 /* !-SIGNO(RmdDumpSysParamInfoRequest)-! */

/**
 * @type RmdDumpSysParamInfoRequest
 *
 * @input signo RMD_DUMP_SYS_PARAM_INFO_REQUEST
 * @input scope The set of processes for which to store information
 * @input info The kind of information to store
 *
 */
typedef struct RmdDumpSysParamInfoRequest
{
   SIGSELECT signo;
   uint32_t scope;
   uint32_t info;
} RmdDumpSysParamInfoRequest;

/**
 * @signal RmdDumpSysParamInfoReply
 *
 * @brief Reply with status from RmdDumpSysParamInfoRequest.
 *
 * @member RMD_DUMP_SYS_PARAM_INFO_REPLY
 * @member RmdStatusReply
 *
 */

/** @macro RMD_DUMP_SYS_PARAM_INFO_REPLY */
#define RMD_DUMP_SYS_PARAM_INFO_REPLY (33175) /* !-SIGNO(RmdStatusReply)-! */

/**
 * @signal RmdDumpRegistersInfoRequest
 *
 * @brief Store register information in the current dump
 *
 * @long If none of the relevant bits (RMD_INFO_REGISTERS) are set in info,
 * the RMD process does nothing.
 *
 * @member RMD_DUMP_REGISTERS_INFO_REQUEST
 * @member RmdDumpRegistersInfoRequest
 *
 */

/** @macro RMD_DUMP_REGISTERS_INFO_REQUEST */
#define RMD_DUMP_REGISTERS_INFO_REQUEST 33176 /* !-SIGNO(RmdDumpRegistersInfoRequest)-! */

/**
 * @type RmdDumpRegistersInfoRequest
 *
 * @input signo RMD_DUMP_REGISTERS_INFO_REQUEST
 * @input scope The set of processes for which to store information
 * @input info The kind of information to store
 * @input regdump Optional set of registers for current process
 *
 */
typedef struct RmdDumpRegistersInfoRequest
{
   SIGSELECT signo;
   uint32_t scope;
   uint32_t info;
   struct cpu_regdump regdump;
} RmdDumpRegistersInfoRequest;

/**
 * @signal RmdDumpRegistersInfoReply
 *
 * @brief Reply with status from RmdDumpRegistersInfoRequest.
 *
 * @member RMD_DUMP_REGISTERS_INFO_REPLY
 * @member RmdStatusReply
 *
 */

/** @macro RMD_DUMP_REGISTERS_INFO_REPLY */
#define RMD_DUMP_REGISTERS_INFO_REPLY (33177) /* !-SIGNO(RmdStatusReply)-! */

/**
 * @signal RmdDumpProgramInfoRequest
 *
 * @brief Store program information in the current dump
 *
 * @long If none of the relevant bits (RMD_INFO_PROGRAM_INFO) are set in info,
 * the RMD process does nothing.
 *
 * @member RMD_DUMP_PROGRAM_INFO_REQUEST
 * @member RmdDumpProgramInfoRequest
 *
 */

/** @macro RMD_DUMP_PROGRAM_INFO_REQUEST */
#define RMD_DUMP_PROGRAM_INFO_REQUEST 33178 /* !-SIGNO(RmdDumpProgramInfoRequest)-! */

/**
 * @type RmdDumpProgramInfoRequest
 *
 * @input signo RMD_DUMP_PROGRAM_INFO_REQUEST
 * @input scope The set of processes for which to store information
 * @input info The kind of information to store
 *
 */
typedef struct RmdDumpProgramInfoRequest
{
   SIGSELECT signo;
   uint32_t scope;
   uint32_t info;
} RmdDumpProgramInfoRequest;

/**
 * @signal RmdDumpProgramInfoReply
 *
 * @brief Reply with status from RmdDumpProgramInfoRequest.
 *
 * @member RMD_DUMP_PROGRAM_INFO_REPLY
 * @member RmdStatusReply
 *
 */

/** @macro RMD_DUMP_PROGRAM_INFO_REPLY */
#define RMD_DUMP_PROGRAM_INFO_REPLY (33179) /* !-SIGNO(RmdStatusReply)-! */

/**
 * @signal RmdDumpTraceInfoRequest
 *
 * @long If none of the relevant bits (RMD_INFO_TRACE) are set in info,
 * the RMD process does nothing.
 *
 * @brief Store event trace information in the current dump
 *
 * @member RMD_DUMP_TRACE_INFO_REQUEST
 * @member RmdDumpTraceInfoRequest
 *
 */

/** @macro RMD_DUMP_TRACE_INFO_REQUEST */
#define RMD_DUMP_TRACE_INFO_REQUEST 33180 /* !-SIGNO(RmdDumpTraceInfoRequest)-! */

/**
 * @type RmdDumpTraceInfoRequest
 * @input signo RMD_DUMP_TRACE_INFO_REQUEST
 * @input scope The set of processes for which to store information
 * @input info The kind of information to store
 *
 */
typedef struct RmdDumpTraceInfoRequest
{
   SIGSELECT signo;
   uint32_t scope;
   uint32_t info;
} RmdDumpTraceInfoRequest;

/**
 * @signal RmdDumpTraceInfoReply
 *
 * @brief Reply with status from RmdDumpTraceInfoRequest.
 *
 * @member RMD_DUMP_TRACE_INFO_REPLY
 * @member RmdStatusReply
 *
 */

/** @macro RMD_DUMP_TRACE_INFO_REPLY */
#define RMD_DUMP_TRACE_INFO_REPLY (33181) /* !-SIGNO(RmdStatusReply)-! */

/**
 * @signal RmdDumpLoadModuleInfoRequest
 *
 * @brief Store load module information in the current dump
 *
 * @long If none of the relevant bits (RMD_INFO_LOAD_MODULE_INFO) are set in
 * info, the RMD process does nothing.
 *
 * @member RMD_DUMP_LOAD_MODULE_INFO_REQUEST
 * @member RmdDumpLoadModuleInfoRequest
 *
 */

/** @macro RMD_DUMP_LOAD_MODULE_INFO_REQUEST */
#define RMD_DUMP_LOAD_MODULE_INFO_REQUEST 33182 /* !-SIGNO(RmdDumpLoadModuleInfoRequest)-! */

/**
 * @type RmdDumpLoadModuleInfoRequest
 *
 * @input signo RMD_DUMP_LOAD_MODULE_INFO_REQUEST
 * @input scope The set of processes for which to store information
 * @input info The kind of information to store
 *
 */
typedef struct RmdDumpLoadModuleInfoRequest
{
   SIGSELECT signo;
   uint32_t scope;
   uint32_t info;
} RmdDumpLoadModuleInfoRequest;

/**
 * @signal RmdDumpLoadModuleInfoReply
 *
 * @brief Reply with status from RmdDumpLoadModuleInfoRequest.
 *
 * @member RMD_DUMP_LOAD_MODULE_INFO_REPLY
 * @member RmdStatusReply
 *
 */

/** @macro RMD_DUMP_LOAD_MODULE_INFO_REPLY */
#define RMD_DUMP_LOAD_MODULE_INFO_REPLY (33183) /* !-SIGNO(RmdStatusReply)-! */

/**
 * @signal RmdDumpPoolMemoryRequest
 *
 * @brief Store pool memory in the current dump.
 *
 * @long If none of the relevant bits (RMD_MEMORY_POOL, RMD_MEMORY_STACKS,
 * RMD_MEMORY_STACKS_OWNED, RMD_MEMORY_STACKS_USED, RMD_MEMORY_STACKS_TOP,
 * RMD_MEMORY_SIGNALS, RMD_MEMORY_SIGNALS_OWNED, RMD_MEMORY_SIGNALS_QUEUED,
 * or RMD_MEMORY_SIGNALS_SENT) are set in memory, the RMD process does nothing.
 *
 * @member RMD_DUMP_POOL_MEMORY_REQUEST
 * @member RmdDumpPoolMemoryRequest
 *
 */

/** @macro RMD_DUMP_POOL_MEMORY_REQUEST */
#define RMD_DUMP_POOL_MEMORY_REQUEST 33184 /* !-SIGNO(RmdDumpPoolMemoryRequest)-! */

/**
 * @type RmdDumpPoolMemoryRequest
 *
 * @input signo RMD_DUMP_POOL_MEMORY_REQUEST
 * @input scope The set of processes for which to store information
 * @input memory The kind of memory to store
 * @input reserved0 Must be set to 0. Not used.
 * @input reserved1 Must be set to 0. Not used.
 * @input regdump Registers for current process or NULL. Used only for
 * RMD_MEMORY_STACKS_USED and RMD_MEMORY_STACKS_TOP.
 *
 */
typedef struct RmdDumpPoolMemoryRequest
{
   SIGSELECT signo;
   uint32_t scope;
   uint32_t memory;
   uint32_t reserved0;
   uint32_t reserved1;
   struct cpu_regdump regdump;
} RmdDumpPoolMemoryRequest;

/**
 * @signal RmdDumpPoolMemoryReply
 *
 * @brief Reply with status from RmdDumpPoolMemoryRequest.
 *
 * @member RMD_DUMP_POOL_MEMORY_REPLY
 * @member RmdStatusReply
 *
 */

/** @macro RMD_DUMP_POOL_MEMORY_REPLY */
#define RMD_DUMP_POOL_MEMORY_REPLY (33185) /* !-SIGNO(RmdStatusReply)-! */

/**
 * @signal RmdDumpHeapMemoryRequest
 *
 * @brief Store heap memory in the current dump.
 *
 * @long If none of the relevant bits (RMD_MEMORY_HEAP, RMD_MEMORY_HEAP_BUFFERS
 * or RMD_MEMORY_HEAP_BUFFERS_OWNED) are set in memory, the RMD process
 * does nothing.
 *
 * @member RMD_DUMP_HEAP_MEMORY_REQUEST
 * @member RmdDumpHeapMemoryRequest
 *
 */

/** @macro RMD_DUMP_HEAP_MEMORY_REQUEST */
#define RMD_DUMP_HEAP_MEMORY_REQUEST 33186 /* !-SIGNO(RmdDumpHeapMemoryRequest)-! */

/**
 * @type RmdDumpHeapMemoryRequest
 *
 * @input signo RMD_DUMP_HEAP_MEMORY_REQUEST
 * @input scope The set of processes for which to store information
 * @input memory The kind of memory to store
 *
 */
typedef struct RmdDumpHeapMemoryRequest
{
   SIGSELECT signo;
   uint32_t scope;
   uint32_t memory;
} RmdDumpHeapMemoryRequest;

/**
 * @signal RmdDumpHeapMemoryReply
 *
 * @brief Reply with status from RmdDumpHeapMemoryRequest.
 *
 * @member RMD_DUMP_HEAP_MEMORY_REPLY
 * @member RmdStatusReply
 *
 */

/** @macro RMD_DUMP_HEAP_MEMORY_REPLY */
#define RMD_DUMP_HEAP_MEMORY_REPLY (33187) /* !-SIGNO(RmdStatusReply)-! */

/**
 * @signal RmdDumpProgramMemoryRequest
 *
 * @brief Store heap memory in the current dump.
 *
 * @long If none of the relevant bits (RMD_MEMORY_DATA, RMD_MEMORY_TEXT)
 * are set in memory, the RMD process does nothing.
 *
 * @member RMD_DUMP_PROGRAM_MEMORY_REQUEST
 * @member RmdDumpProgramMemoryRequest
 *
 */

/** @macro RMD_DUMP_PROGRAM_MEMORY_REQUEST */
#define RMD_DUMP_PROGRAM_MEMORY_REQUEST 33188 /* !-SIGNO(RmdDumpProgramMemoryRequest)-! */

/**
 * @type RmdDumpProgramMemoryRequest
 *
 * @input signo RMD_DUMP_PROGRAM_MEMORY_REQUEST
 * @input scope The set of processes for which to store information
 * @input memory The kind of memory to store
 *
 */
typedef struct RmdDumpProgramMemoryRequest
{
   SIGSELECT signo;
   uint32_t scope;
   uint32_t memory;
} RmdDumpProgramMemoryRequest;

/**
 * @signal RmdDumpProgramMemoryReply
 *
 * @brief Reply with status from RmdDumpProgramMemoryRequest.
 *
 * @member RMD_DUMP_PROGRAM_MEMORY_REPLY
 * @member RmdStatusReply
 *
 */

/** @macro RMD_DUMP_PROGRAM_MEMORY_REPLY */
#define RMD_DUMP_PROGRAM_MEMORY_REPLY (33189) /* !-SIGNO(RmdStatusReply)-! */

/** @macro RMD_SET_PROPERTY_REQUEST */
#define RMD_SET_PROPERTY_REQUEST (33190) /* !-SIGNO(RmdSetPropertyRequest)-! */

/**
 * @type RmdSetPropertyRequest
 *
 * @input signo RMD_SET_PROPERTY_REQUEST
 * @input property The type of RmdProperty
 * @input value The value of the property
 *
 */
typedef struct RmdSetPropertyRequest
{
   SIGSELECT signo;
   RmdProperty property;
   uint32_t value;
} RmdSetPropertyRequest;

/**
 * @signal RmdSetPropertyReply
 *
 * @brief Reply with status from RmdSetPropertyRequest.
 *
 * @member RMD_SET_PROPERTY_REPLY
 * @member RmdStatusReply
 *
 */

/** @macro RMD_SET_PROPERTY_REPLY */
#define RMD_SET_PROPERTY_REPLY (33191) /* !-SIGNO(RmdStatusReply)-! */

/** @macro RMD_END_OF_DUMP_REQUEST */
#define RMD_END_OF_DUMP_REQUEST 33192 /* !-SIGNO(RmdEndOfDumpRequest)-! */

/**
 * @type RmdEndOfDumpRequest
 *
 * @input signo RMD_END_OF_DUMP_REQUEST
 * @input status = 0, End of dump,successful
 *        status = 1, End of dump,incomplete
 *
 */
typedef struct RmdEndOfDumpRequest
{
   SIGSELECT signo;
   uint32_t status;
} RmdEndOfDumpRequest;

/**
 * @signal RmdEndOfDumpReply
 *
 * @brief Reply with status from RmdEndOfDumpRequest.
 *
 * @member RMD_END_OF_DUMP_REPLY
 * @member RmdStatusReply
 *
 */

/** @macro RMD_END_OF_DUMP_REPLY */
#define RMD_END_OF_DUMP_REPLY (33193) /* !-SIGNO(RmdStatusReply)-! */


/**
 * @signal RmdGetDumpInfoRequest
 *
 * @brief Get information about the current dump.
 *
 * @member RMD_GET_DUMP_INFO_REQUEST
 * @member RmdGetDumpInfoRequest
 *
 */

/** @macro RMD_GET_DUMP_INFO_REQUEST */
#define RMD_GET_DUMP_INFO_REQUEST (33194) /* !-SIGNO(RmdGetDumpInfoRequest)-! */

/**
 * @type RmdGetDumpInfoRequest
 *
 * @input signo                 RMD_GET_DUMP_INFO_REQUEST
 * @input tag_value             Information tag value
 * @input reserved0             NOT USED Must be set to 0
 */
typedef struct RmdGetDumpInfoRequest
{
   SIGSELECT signo;
   uint32_t tag_value;
   uint32_t reserved0;
} RmdGetDumpInfoRequest;

/**
 * @signal RmdGetDumpInfoReply
 *
 * @brief Reply with requested information from RmdGetDumpInfoRequest.
 *
 * @member RMD_GET_DUMP_INFO_REPLY
 * @member RmdGetDumpInfoReply
 *
 */

/** @macro RMD_GET_DUMP_INFO_REPLY */
#define RMD_GET_DUMP_INFO_REPLY (33195) /* !-SIGNO(RmdGetDumpInfoReply)-! */

/**
 * @type RmdGetDumpInfoRequest
 *
 * @output signo                 RMD_GET_DUMP_INFO_REPLY
 * @output status                RmdStatus of operation.
 * @output length                Length of the returned data in bytes.
 * @field  reserved0             NOT USED
 * @field  reserved1             NOT USED
 * @output data                  The tag dependent data.
 */
typedef struct RmdGetDumpInfoReply
{
   SIGSELECT signo;
   uint32_t status;
   uint32_t length;
   uint32_t reserved0;
   uint32_t reserved1;
   char data[1];
} RmdGetDumpInfoReply;

/* #ifndef _RMD_SIG */
#endif
