/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2010 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

#include "stdint.h"

/*
 * This file describes the media format for ROFS.
 *
 * NOTE: All structs must be 64-bit aligned, since some fields are 64 bits in
 *       size!
 */

/*
 * Which format version this file describes
 */

#define ROFS_FORMAT_VERSION 0

/*
 * This magic 32-bit value must be first value in an ROFS volume.
 * NOTE: In the future, it would be possible to determine the endian by
 *       checking if this value is backwards or not. Might be useful if
 *       ROFS is used for read-only volumes rather than in-elf volumes.
 */
#define ROFS_MAGIC 0xee7897f0ul

#define ROFS_MAGIC_WRONG_ENDIAN 0xf09778eeul 

/*
 * Byte offset from start of struct RofsVolume.
 */
typedef int64_t voloffs_t;

/*
 * Information stored for each file in the ROFS volume. In POSIX terms, this
 * is a combined directory entry (i.e. file name) and inode (i.e. file meta-
 * data).
 */
struct RofsMediaFile
{
      /* Size of file in bytes */
      uint64_t   fileSizeBytes;

      /* Offset from volume start to where file content can be found. */
      voloffs_t dataOffset;

      /* Inode number. Currently a serial number starting from 1, since 0 is
       * reserved to mean illegal inode number. */
      uint64_t   inode;

      /* When traversing files, next file entry is found by taking current
       * offset and increment it with this value.
       * NOTE: This is not the struct size, since there might be some padding
       *       to ensure correct alignment.
       * NOTE: For root directory entry, this is the size of RofsMediaVolume
       *       including the file name. */
      uint32_t  nextOffset;

      /* File creation time in seconds from Epoch. This value is used for all
       * time fields since volume is read-only.
       * For root directory this is the volume creation time.
       * For other files, this is the last status change time for the source
       * file when volume was created. */
      uint32_t  ctime;

      /* File type and attributes as per sys/stat.h. */
      uint32_t  mode;

      /* Number of hard links. If there is a hard link, there will be several
       * file entries that only differs by name. This means that they all have
       * the same inode number and shares the same dataOffset and file size.
       * Link count could seen as a reference count to the data referenced by
       * dataOffset. */
      uint32_t  linkCount;

      /* File name, null terminated UTF8 string. If this is the root directory,
       * this is the volume label. */
      char      name[1]; /* Dynamic size */
};

/* Information about the volume. This is what the volume starts with. */
struct RofsMediaVolume
{
      /* Value ROFS_MAGIC, used for sanity checks. */
      uint32_t  magic;

      /* Major version of the media format. This version number will be changed
       * when media format is changed in a non-backwards compatible way. */
      uint32_t  formatVersion;

      /* Bit field with media version options. It is formatVersion specific
       * what each bit means. For formatVersion 0, no option is defined. */
      uint32_t  options;

      /* Reserved, must be set to 0. */
      uint32_t  reserved;

      /* Size of ROFS volume. */
      uint64_t  volumeSizeBytes;

      /* This is the root directory file. The name of this file is the volume
       * label. Sine the root directory is always created by the image tool
       * the ctime for this file is the volume creation time. */
      struct RofsMediaFile rootDir;
};

/*
 ******************************************************************************
 * The rest of this file describes valid values for mode field in
 * struct RofsMediaFile. The values are the same as for OSE's version of
 * sys/stat.h. They are needed here since the volume can be used both by
 * host tools and by OSE code, and using OSE's sys/stat.h will not work since
 * the names will interfere with the host's sys/stat.h.
 * Only supported file types are included below, but all attributes. This is
 * because ROFS does no actions on attributes, it only provides them when
 * files are examined.
 ******************************************************************************
 */

/**
 * @macro ROFS_S_ISDIR
 *
 * @include        "sys/stat.h"
 *
 * @declaration #define ROFS_S_ISDIR(mode)       ((mode) & ROFS_S_IFDIR)
 *
 * @brief Check if the mode of the stat object is a directory.
 *
 * @field mode
 *
 * @long  Macro for testing object type in the @link "" "stat" structure.
 */
#define ROFS_S_ISDIR(mode)       ((mode) & ROFS_S_IFDIR)

/**
 * @macro ROFS_S_ISREG
 *
 * @include        "sys/stat.h"
 *
 * @declaration #define ROFS_S_ISREG(mode)       ((mode) & ROFS_S_IFREG)
 *
 * @brief Check if the mode of the stat object is a regular file.
 *
 * @field mode
 *
 * @long  Macro for testing object type in the @link "" "stat" structure.
 */
#define ROFS_S_ISREG(mode)       ((mode) & ROFS_S_IFREG)

/**
 * @macro ROFS_S_IFDIR
 *
 * @include        "sys/stat.h"
 *
 * @declaration #define ROFS_S_IFDIR       0x00004000
 *
 * @brief Object is a directory.
 *
 * @long  Bitvalue for object types.
 */
#define ROFS_S_IFDIR       0x00004000

/**
 * @macro ROFS_S_IFREG
 *
 * @include        "sys/stat.h"
 *
 * @declaration #define ROFS_S_IFREG       0x00020000
 *
 * @brief Object is a regular file.
 *
 * @long  Bitvalue for object types.
 */
#define ROFS_S_IFREG       0x00020000

/**
 * @macro ROFS_S_IFMT
 *
 * @include        "sys/stat.h"
 *
 * @brief Bitmask of type information in st_mode
 *
 * @long  Bitmask for object types returned in stat.
 */
#define ROFS_S_IFMT       (ROFS_S_IFDIR |\
                      ROFS_S_IFCHR |\
		      ROFS_S_IFBLK |\
		      ROFS_S_IFREG |\
		      ROFS_S_IFFIFO |\
		      ROFS_S_IFLNK |\
		      ROFS_S_IFSOCK)


/**
 * @macro ROFS_S_IARCH
 *
 * @include        "sys/stat.h"
 *
 * @declaration #define ROFS_S_IARCH       0x00000200
 *
 * @brief Object has been modified.
 *
 * @long  Bitvalue for object access permissions.
 */
#define ROFS_S_IARCH       0x00000200

/**
 * @macro ROFS_S_IRUSR
 *
 * @include        "sys/stat.h"
 *
 * @declaration #define ROFS_S_IRUSR       0x00000100
 *
 * @brief Read access for owner.
 *
 * @long  Bitvalue for object access permissions.
 */
#define ROFS_S_IRUSR       0x00000100

/**
 * @macro ROFS_S_IWUSR
 *
 * @include        "sys/stat.h"
 *
 * @declaration #define ROFS_S_IWUSR       0x00000080
 *
 * @brief Write access for owner.
 *
 * @long  Bitvalue for object access permissions.
 */
#define ROFS_S_IWUSR       0x00000080

/**
 * @macro ROFS_S_IXUSR
 *
 * @include        "sys/stat.h"
 *
 * @declaration #define ROFS_S_IXUSR       0x00000040
 *
 * @brief Execute/search access for owner.
 *
 * @long  Bitvalue for object access permissions.
 */
#define ROFS_S_IXUSR       0x00000040

/**
 * @macro ROFS_S_IRGRP
 *
 * @include        "sys/stat.h"
 *
 * @declaration #define ROFS_S_IRGRP       0x00000020
 *
 * @brief Read access for group.
 *
 * @long  Bitvalue for object access permissions.
 */
#define ROFS_S_IRGRP       0x00000020

/**
 * @macro ROFS_S_IWGRP
 *
 * @include        "sys/stat.h"
 *
 * @declaration #define ROFS_S_IWGRP       0x00000010
 *
 * @brief Write access for group.
 *
 * @long  Bitvalue for object access permissions.
 */
#define ROFS_S_IWGRP       0x00000010

/**
 * @macro ROFS_S_IXGRP
 *
 * @include        "sys/stat.h"
 *
 * @declaration #define ROFS_S_IXGRP       0x00000008
 *
 * @brief Execute/search access for group.
 *
 * @long  Bitvalue for object access permissions.
 */
#define ROFS_S_IXGRP       0x00000008

/**
 * @macro ROFS_S_IROTH
 *
 * @include        "sys/stat.h"
 *
 * @declaration #define ROFS_S_IROTH       0x00000004
 *
 * @brief Read access for other.
 *
 * @long  Bitvalue for object access permissions.
 */
#define ROFS_S_IROTH       0x00000004

/**
 * @macro ROFS_S_IWOTH
 *
 * @include        "sys/stat.h"
 *
 * @declaration #define ROFS_S_IWOTH       0x00000002
 *
 * @brief Write access for other.
 *
 * @long  Bitvalue for object access permissions.
 */
#define ROFS_S_IWOTH       0x00000002

/**
 * @macro ROFS_S_IXOTH
 *
 * @include        "sys/stat.h"
 *
 * @declaration #define ROFS_S_IXOTH       0x00000001
 *
 * @brief Execute/search access for other.
 *
 * @long  Bitvalue for object access permissions.
 */
#define ROFS_S_IXOTH       0x00000001

/**
 * @macro ROFS_S_IRWXU
 *
 * @include        "sys/stat.h"
 *
 * @declaration #define ROFS_S_IRWXU       (ROFS_S_IRUSR | ROFS_S_IWUSR | ROFS_S_IXUSR)
 *
 * @brief All access permissions for owner.
 *
 * @long  Bitvalue for object access permissions.
 */
#define ROFS_S_IRWXU       (ROFS_S_IRUSR | ROFS_S_IWUSR | ROFS_S_IXUSR)

/**
 * @macro ROFS_S_IRWXG
 *
 * @include        "sys/stat.h"
 *
 * @declaration #define ROFS_S_IRWXG       (ROFS_S_IRGRP | ROFS_S_IWGRP | ROFS_S_IXGRP)
 *
 * @brief All access permissions for group.
 *
 * @long  Bitvalue for object access permissions.
 */
#define ROFS_S_IRWXG       (ROFS_S_IRGRP | ROFS_S_IWGRP | ROFS_S_IXGRP)

/**
 * @macro ROFS_S_IRWXO
 *
 * @include        "sys/stat.h"
 *
 * @declaration #define ROFS_S_IRWXO       (ROFS_S_IROTH | ROFS_S_IWOTH | ROFS_S_IXOTH)
 *
 * @brief All access permissions for other.
 *
 * @long  Bitvalue for object access permissions.
 */
#define ROFS_S_IRWXO       (ROFS_S_IROTH | ROFS_S_IWOTH | ROFS_S_IXOTH)

/**
 * @macro ROFS_S_IREAD
 *
 * @include        "sys/stat.h"
 *
 * @declaration #define ROFS_S_IREAD       (ROFS_S_IRUSR | ROFS_S_IRGRP | ROFS_S_IROTH)
 *
 * @brief Read access for owner, group and other.
 *
 * @long  Bitvalue for object access permissions.
 */
#define ROFS_S_IREAD       (ROFS_S_IRUSR | ROFS_S_IRGRP | ROFS_S_IROTH)

/**
 * @macro ROFS_S_IWRITE
 *
 * @include        "sys/stat.h"
 *
 * @declaration #define ROFS_S_IWRITE      (ROFS_S_IWUSR | ROFS_S_IWGRP | ROFS_S_IWOTH)
 *
 * @brief Write access for owner, group and other.
 *
 * @long  Bitvalue for object access permissions.
 */
#define ROFS_S_IWRITE      (ROFS_S_IWUSR | ROFS_S_IWGRP | ROFS_S_IWOTH)

/**
 * @macro ROFS_S_IEXEC
 *
 * @include        "sys/stat.h"
 *
 * @declaration #define ROFS_S_IEXEC       (ROFS_S_IXUSR | ROFS_S_IXGRP | ROFS_S_IXOTH)
 *
 * @brief Execution access for owner, group and other.
 *
 * @long  Bitvalue for object access permissions.
 */
#define ROFS_S_IEXEC       (ROFS_S_IXUSR | ROFS_S_IXGRP | ROFS_S_IXOTH)

/**
 * @macro ROFS_S_ISUID
 *
 * @include        "sys/stat.h"
 *
 * @declaration #define ROFS_S_ISUID 0x10000000
 *
 * @brief Set effective user ID on execution as listed in inode.
 *
 * @long  Bitvalue for object access permissions.
 */
#define ROFS_S_ISUID 0x10000000

/**
 * @macro ROFS_S_ISGID
 *
 * @include        "sys/stat.h"
 *
 * @declaration #define ROFS_S_ISGID 0x20000000
 *
 * @brief Set effective group ID on execution as listed in inode.
 *
 * @long  Bitvalue for object access permissions.
 */
#define ROFS_S_ISGID 0x20000000

/**
 * @macro ROFS_S_ISVTX
 *
 * @include        "sys/stat.h"
 *
 * @declaration #define ROFS_S_ISVTX 0x40000000
 *
 * @brief Sticky bit; for directories, only super user or owner can rename
 *        or unlink files. Implementation specific behavior on files
 *        with bit set.
 *
 * @long  Bitvalue for object access permissions.
 */
#define ROFS_S_ISVTX 0x40000000



/* OSE extension attributes are not documented. */

#define ROFS_S_IHIDDEN   0x00000400  /* The object name should be hidden */
#define ROFS_S_IBOOTFILE 0x00000800  /* The object is designated as a boot file. */
#define ROFS_S_ISTARTUP  ROFS_S_IBOOTFILE /* The object is designated as a startup file.
                                 */
#define ROFS_S_IUSER0    0x01000000  /* User application-defined attribute bit 0. */
#define ROFS_S_IUSER1    0x02000000  /* User application-defined attribute bit 1. */
#define ROFS_S_IUSER2    0x04000000  /* User application-defined attribute bit 2. */
#define ROFS_S_IUSER3    0x08000000  /* User application-defined attribute bit 3. */

