/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2006 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

#ifndef __SLAB_H__
#define __SLAB_H__

#if defined(PAGE_SIZE_4K)
  #define PAGE_SIZE 0x1000
  #define PAGE_SIZE_BITS 12
#elif defined(PAGE_SIZE_8K)
  #define PAGE_SIZE 0x2000
  #define PAGE_SIZE_BITS 13
#else
  #error "No page size set."
#endif

#include <stddef.h>
#include <stdint.h>
#include "mmtypes.h"
#include "list_ext.h"
#include "hash.h"
#ifdef STANDALONE
#include "standalone.h"
#else
#include "compiler.h"
#endif

/**
 * A SLAB allocator for OSE
 */


/*
 * Configuration of the static slab allocator.
 *
 * The static slab uses a set of anonymous caches configured with the
 * sizes listed below. Allocation is performed in a best fit manner.
 * This is the place to add a new size if required. The list must be ordered
 * from smaller sizes to larger. 
 */
#define SLAB_STATIC_SIZES         {0x20, 0x40, 0x60, 0x80}
#define SLAB_STATIC_NUM_SIZES     4
#define SLAB_STATIC_INVALID_SIZE  ((uint32_t)-1)
#define SLAB_STATIC_ALIGN         8
#define SLAB_STATIC_NAME_PREFIX   "slab_static_"


/* When the size of an entry is less than this, we will put the slab control
 * structure at the end of the page. When the size of an entry is larger
 * than this threshold, we use separately allocated slab control structures
 * and might allocate multiple pages from MM at once.
 * The reason for this is to cut down on fragmentation. */
#define EMBED_SLAB_CTRL_THRESHOLD (PAGE_SIZE / 8)

/* The max size that the slab cache will allocate from the MM each time
 * it needs to allocate a new slab. Usually it allocates only one page. */
#define MAX_MM_ALLOC_SIZE 0x4000

#define SLAB_NAME_MAX 32

/* Options */
#define SLAB_SLEEP 1
#define SLAB_NOSLEEP 0  /* Default if no options are supplied */
#define SLAB_NUM_OPTIONS 2

/* One for each slab cache created. */
struct slab_cache
{
   /* The linked list that connects the slab caches. */
   list_link_t link;

   /* The name of the slab cache. */
   char name[SLAB_NAME_MAX];

   /* The size of each entry in this slab cache. */
   size_t size;

   /* The number of entries per slab. Computed from the size. */
   uint32_t entries_per_slab;

   /* The alignment of teach entry in the slab cache, or zero if this is
    * not an issue. */
   uint32_t align;

   /* The number of spare pages to be used when allocating from MM context. */
   uint32_t spare_entries;

   /* The maximum number of entries that are allowed to be in use at any
    * given time without allocating extra entries from MM. */
   uint32_t max_no_of_used_entries;

   /* The size of each slab allocated from the MM.
    * Usually equal to the page size for slabs with small objects. */
   uint32_t mm_alloc_size;

   /* A lock that need to be acquired before manipulating any slab cache
    * metadata. */
   spinlock_t lock;

   /* A pointer to the constructor function called every time an entry
    * is allocated. */
   void (*constructor)(void *, size_t);

   /* A pointer to the destructor function called every time an entry
    * is freed.*/
   void (*destructor)(void *, size_t);

   /* The list of slabs used by this slab cache. */
   list_t slab_list;

   /* The partially_empty pointer points to the first full slab. By inserting
    * an entry before this pointer, you insert it into the partially empty
    * part of the list. */
   struct slab_ctrl *partially_empty;

   /* The hash table that finds the buffer ctrl block given a buffer address.
    * Only used when the slab control structure is not embedded in
    * the page. */
   struct hash_table_t buf_ctrl_hash;

   /* The number of slabs allocated by this slab cache. */
   uint32_t no_of_slabs;

   /* The number of entries in use by this slab cache. */
   uint32_t no_of_used_entries;

   /* The maximum number of spare entries that has ever been used by this
    * slab cache. */
   uint32_t max_used_spare_entries;

   /* A set of options configured for the slab cache during creation. */
   uint32_t options;
};

struct buf_ctrl;

/* One for each slab allocated at the end of the slab. */
struct slab_ctrl
{
   /* For the linked list that connects the slabs. */
   list_link_t link;

   /* A pointer back to the list structure this entry is in. */
   struct slab_cache *slab_cache;

   /* List of free entries in this slab. When the size is
    * smaller than EMBED_SLAB_CTRL_THRESHOLD, the pointers are embedded
    * in the non-allocated slab entries.*/
   struct buf_ctrl *free_list;

   /* Number of entries used in this slab. */
   uint32_t no_of_used_entries;

   /* The list of buffer control blocks. Used for slab caches
    * with a size larger than EMBED_SLAB_CTRL_THRESHOLD. */
   list_t buf_ctrl_list;
};

/* These buffer control blocks are used to locate buffers when the object
 * size is larger than EMBED_SLAB_CTRL_THRESHOLD. */
struct buf_ctrl
{
   /* List of free entries in this slab. When the size is
    * smaller than EMBED_SLAB_CTRL_THRESHOLD, the pointers are embedded
    * in the non-allocated slab entries, and the rest of the entries are
    * not used.*/
   struct buf_ctrl *free_list;

   /* For the linked list that connects the buffer control structures. */
   list_link_t buf_ctrl_link;

   /* For the hash that finds a buffer control block given a buffer address. */
   list_link_t hash_link;

   /* A pointer back to the slab this entry belongs to. */
   struct slab_ctrl *slab;

   /* Pointer to the buffer. */
   void *buf;
};


/*
 *===========================================================================
 *                        create_slab_allocator
 *===========================================================================
 *
 * Description
 * -----------
 * Creates a slab allocator for a specific type of data.
 *
 * Parameter description
 * ---------------------
 * name         Name of the entries in the slab.
 * size         The size of each entry in the slab. Cannot be changed
 *              after creation.
 * align        The alignment of the data. If no alignment is desired,
 *              enter a zero.
 * spares       The number of spare entries that are allocated.
 *              The slab cannot ask for more pages from the MM when it is
 *              called by the MM. Instead, it will use memory from one or
 *              more of these spare entries, and as soon the slab is called
 *              from some other process, it allocates more entries so that
 *              it have the configured amount of spare entries again.
 * constructor  A pointer to a function that is called every time an
 *              object is created. Usefull for setting the initial
 *              state of the entry.
 * destructor   A pointer to a function that is called every time an
 *              object is freed. Usefull for deallocating stuff.
 *
 * options      A bitmask of slab cache options.
 *               - SLAB_SLEEP causes the slab to allocate memory using
 *                 serialized signaling.
 *
 * Restrictions
 * ------------
 * Cannot be called from MM context.
 *
 * Returns
 * -------
 * struct slab_cache*   Returns a pointer to the created slab.
 *
 * Restrictions
 * ------------
 * You cannot create a slab with an entry size larger than
 * EMBED_SLAB_CTRL_THRESHOLD before the system heap has been initialized.
 * This is usually 256 bytes.
 *
 */
struct slab_cache *
create_slab_allocator(const char *name, size_t size, uint32_t align,
                      uint32_t spares,
                      void (*constructor)(void *, size_t),
                      void (*destructor)(void *, size_t),
                      uint32_t options);

/*
 *===========================================================================
 *                        delete_slab_allocator
 *===========================================================================
 *
 * Description
 * -----------
 * Deletes the slab allocator pointed to by *slab_cache. Deallocated all the
 * memory associated with this slab cache.
 *
 */
void
delete_slab_allocator(struct slab_cache *slab_cache);

/*
 *===========================================================================
 *                        slab_alloc
 *===========================================================================
 *
 * Description
 * -----------
 * Allocate an entry from the slab pointed to by *slab
 *
 * Parameter description
 * ---------------------
 * slab       The slab to allocate from.
 * flags      Either SLAB_SLEEP or SLAB_NOSLEEP, indicating if it is
 *            acceptable to wait for memory if none is currently available.
 *
 * Returns
 * -------
 * void*   Returns a pointer to the newly allocated slab entry.
 *
 */
void *
slab_alloc(struct slab_cache *slab_cache, int flags);

/*
 *===========================================================================
 *                        slab_free
 *===========================================================================
 *
 * Description
 * -----------
 * Free the entry *buf from the slab it is located in.
 *
 */
void
slab_free(struct slab_cache *slab_cache, void *buf);

/*
 *===========================================================================
 *                        init_slab
 *===========================================================================
 *
 * Description
 * -----------
 * Initializes the slab to allocate memory from the provided partition.
 * Needs to be called once before the slab is used.
 *
 */
void
init_slab(const char *slab_log_mem_name);

/*
 *===========================================================================
 *                        init_slab2
 *===========================================================================
 *
 * Description
 * -----------
 * The last initialization of the slab. Registers the MM events.
 * The MM event mechanism uses the slab, therefor it cannot be performed
 * in init_slab.
 *
 */
void
init_slab2(void);

#endif /* __SLAB_H__ */
