/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2006 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */
#ifndef _OSE_SPLAYTREE_H_
#define _OSE_SPLAYTREE_H_

/**
 * @public
 *
 * @file splay.h
 *
 * @brief Functions for maintaining a Splay Tree.
 *
 * @long A Splay tree is a Binary Search Tree on which splaying
 * operations are performed. A splaying operation causes a node in the
 * tree to move up the tree and become the root. Splaying of a node is
 * done by rotating the nodes.
 *
 * @long Every time a node is accessed via any of the standard tree
 * operations (find, insert, remove) it is splayed, becoming the new
 * root of the tree. This will allow for a fast lookup if it is
 * immediately searched for again.
*/
#include "osetypes.h"


/**
 * @type SplayNode_t
 *
 * @brief Declaration of a Splay tree node.
 *
 * @long This struct defines the fields that the Splay tree operation
 * uses to maintain the splay tree. The client struct must begin with
 * these fields.
 *
 * @long The fileds that immediately follow are treated as the key to
 * the node. The key field(s) are the ones that make the node unique
 * and maintains an ordering relation within the tree. The order is
 * defined by the compare function described below.
 *
 * @field left Pointer to the left child.
 * @field right Pointer to the right child.
 *
 */
typedef struct SplayNode_t
{
      struct SplayNode_t  *left;	/* left child */
      struct SplayNode_t  *right;	/* right child */
      /* ...client node data... */
} SplayNode_t;


/**
 * @type SplayTree_t
 *
 * @brief Declaration of a Splay tree.
 *
 * @long The client uses a variable of this type to represent a Splay
 * tree. The variable is passed as the first argument to all Splay
 * operations to identify the tree.
 *
 * @field root Pointer to the root of the Splay tree.
 *
 * @field size Number of nodes in tree.
 *
 * @field compare Callback function for comparing two nodes.
 *
 */
typedef struct SplayTree_t
{
      SplayNode_t  *root;
      int           size;
      int         (*compare)(const void *key_a, const void *key_b);
} SplayTree_t;


/**
 * @public
 * @function splaytree_init
 *
 * @brief Initializes a Splay tree. Must be called before any other
 * operation is performed on the tree.
 *
 * @long
 *
 * @field tree
 *
 * @field compare Callback function for comparing two nodes. This
 * function returns a value < 0 if key_a is less than key_b, > 0 if
 * key_a is greater than key_b and == 0 if key_a equals key_b. The
 * pointer key_a and key_b points to the user data portion of the tree
 * nodes (i.e. not including the tree pointers).
 *
 * @library libosl.a
 *
 */
void splaytree_init(SplayTree_t *tree, int (*compare)(const void *key_a, const void *key_b));


/**
 * @public
 * @function splaytree_insert
 *
 * @brief Insert a new node into a Splay tree.
 *
 * @long
 *
 * @field tree Pointer to a Splay tree.
 *
 * @field node Pointer to the node to be inserted.
 *
 * @return 0 if the operation was successful, -1 otherwise (duplicate
 * nod already in the tree).
 *
 * @library libosl.a
 *
*/
int splaytree_insert(SplayTree_t *tree, SplayNode_t *node);


/**
 * @public
 * @function splaytree_remove
 *
 * @brief Remove a node from the Splay tree.
 *
 * @long The remove operation tries to locate the specified node to be
 * removed by searching for it using the key (same procedure as find
 * operation). If the node is found it's removed from the tree and a
 * pointer to it is returned to the caller.
 *
 * @field tree Pointer to a Splay tree.
 *
 * @field key Pointer to the key of the node that should be removed.
 *
 * @return Pointer to the node if operation was successful, NULL
 * othervise (node not found).
 *
 * @library libosl.a
 *
 */
SplayNode_t *splaytree_remove(SplayTree_t *tree, const void *key);


/**
 * @public
 * @function splaytree_find
 *
 * @brief Searches the tree for a node matching a specified key.
 *
 * @long
 *
 * @field tree Pointer to a Splay tree.
 *
 * @field key Pointer to a key.
 *
 * @return Pointer to the node if found, NULL othervise.
 *
 * @library libosl.a
 *
 */
SplayNode_t *splaytree_find(SplayTree_t *tree, const void *key);

#endif /* #ifndef _OSE_SPLAYTREE_H_ */
