/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2009 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

#include "osetypes.h"
#include "osenames.h"
#include "ose.h"
#include "pose.con"
#include "pose.h"
#include "krncon.h"
#include "ramlog.h"
#include "string.h"

#include "intvec.h"

/*
** Space for the vector table should be provided by each target interface.
*/
extern struct intdata *zzint_table[];

/* List of free intdata structs. */
static struct intdata *intdata_free_list = NULL;

/* Configuration struct. */
extern struct Ti_ctlblk zzconf;

/***************************************************************************
 * New-style vector manipulators. These rules apply to all calls:
 *
 * - Should not be called concurrently.
 * - Should be called from the correct core in an SMP system. An
 *   exception is enabling a vector on any core from core 0.
 * - Can currently only add one pcb to a vector.
 * - The functions are called with interrupts enabled. The create
 *   handlers may send signals to other cores if called from an
 *   incorrect core.
 * - Dispatcher should be locked unless the functions are called from
 *   a system process that cannot die.
 **************************************************************************/

/* Check that vecno is a valid vector number. */
static OSBOOLEAN
check_vector(OSVECTOR vecno)
{
   return (vecno <= VECTOR_NUM_ENTRIES);
}

/* Returns true if the vector entry has been allocated. */
static OSBOOLEAN
is_vector_allocated(OSVECTOR vecno)
{
   return (zzint_table[vecno] != NULL);
}

/* Allocates space for a vector entry. */
static void
allocate_vector(OSVECTOR vecno)
{
   struct intdata *item = NULL;
   
   /* Need to create a new vector entry. The caller holds a spinlock. */
   if (intdata_free_list != NULL)
   {
      item = intdata_free_list;
      intdata_free_list = item->pcb;
   }

   /* We may not have had any items on the free list. */
   if (item == NULL)
   {
      item = zzalloc_kernelmemory(sizeof(struct intdata));      
   }

   /* Initialize the entry. */
   memset(item, 0, sizeof(struct intdata));

   /* Install it in the vector table. */
   zzint_table[vecno] = item;
}

/* Frees the given vector entry if it is unused. */
static void
free_vector(OSVECTOR vecno)
{
   struct intdata *item = zzint_table[vecno];
   OSBOOLEAN isr = 0;
   int i;

   /* Check for interrupt process. */
   if (item->pcb != NULL)
   {
      return;
   }
   
   /* Check if there is any ISR (type 1) in the list. */
   for (i = 0; i < ISR_NUM_ENTRIES; ++i)
   {
      if (item->isr[i] == NULL)
      {
         break;
      }
      if (item->type[i] == ISR_TYPE_ISR)
      {
         isr = 1;
         break;
      }
   }

   /* No ISR or intproc means no further use for this vector. */
   if (!isr)
   {
      item->pcb = intdata_free_list;
      intdata_free_list = item;
      zzint_table[vecno] = NULL;
   }
}

/* Assigns a PCB to a vector. The vector must be unused. */
OSBOOLEAN
zzbind_vector_pcb(OSVECTOR vecno, void *pcb)
{
#ifdef softOSE
   /* SFK can only use vectors configured in the kernel configuration. */
   if (!zzcheck_vector_sfk(vecno))
   {
      return False;
   }
#endif
   
   /* Validate the vector number. */
   if (!check_vector(vecno))
   {
      return False;
   }

   /* Allocate a new entry if necessary. */
   if (!is_vector_allocated(vecno))
   {
      allocate_vector(vecno);
   }
   else
   {
      /* If it is allocated already it means that there is already an
       * interrupt process or an ISR assigned to this vector, which
       * means that it cannot be assigned an interrupt process. */
      return False;
   }
   
   /* Populate the new entry. */
   zzint_table[vecno]->pcb = pcb;
   
   return True;
}

/* Removes a PCB from a vector. The vector entry will be freed. */
void
zzrelease_vector_pcb(OSVECTOR vecno, void *pcb)
{
   /* Make sure the vector is valid and allocated. */
   if (!check_vector(vecno) || !is_vector_allocated(vecno))
   {
      return;
   }

   /* Check that it actually has an interrupt process. */
   if (zzint_table[vecno]->pcb == NULL)
   {
      return;
   }

   zzint_table[vecno]->pcb = NULL;
      
   /* Release the entry to the free list. */
   free_vector(vecno);
}

/* Assigns an ISR to a vector entry. It can be either a regular ISR or
 * an interrupt hook, determined by the type argument. */
OSBOOLEAN
zzbind_vector_isr(OSVECTOR vecno, void *isr, uint32_t isr_arg, uint32_t type)
{
   int i;
   int slot = 0;   /* shut-up compiler warning */
   Msr msr;

#ifdef softOSE
   /* SFK can only use vectors configured in the kernel configuration. */
   if (!zzcheck_vector_sfk(vecno))
   {
      return False;
   }
#endif

   /* Validate the vector number. */
   if (!check_vector(vecno))
   {
      return False;
   }

   /* Check if the vector has been allocated. */
   if (!is_vector_allocated(vecno))
   {
      /* Only allocate new entries for ISRs. */
      if (type == ISR_TYPE_ISR)
      {
         allocate_vector(vecno);
      }
      else
      {
         return False;
      }
   }
   else
   {   
      /* An ISR and an interrupt process cannot coexist. */
      if (zzint_table[vecno]->pcb && (type == ISR_TYPE_ISR))
      {
         return False;
      }
   }

   /* Check if all slots are taken already. */
   if (zzint_table[vecno]->isr[ISR_NUM_ENTRIES - 1] != NULL)
   {
      return False;      
   }
   
   if (type == ISR_TYPE_HOOK)
   {
      /* Find a suitable entry (after any other hooks). */      
      for (i = 0; i < ISR_NUM_ENTRIES; ++i)
      {
         if ((zzint_table[vecno]->isr[i] == NULL) ||
             (zzint_table[vecno]->type[i] == ISR_TYPE_ISR))
         {
            slot = i;
            break;
         }
      }

      /* Move any ISR entries out of the way. */
      for (; zzint_table[vecno]->isr[i] != NULL; ++i);

      /* It's simpler to move from the back.. */
      LOCK_SAVE(msr);
      for (; i > slot; --i)
      {
         zzint_table[vecno]->isr[i] = zzint_table[vecno]->isr[i - 1];
         zzint_table[vecno]->arg[i] = zzint_table[vecno]->arg[i - 1];
         zzint_table[vecno]->type[i] = zzint_table[vecno]->type[i - 1];
      }

      /* Fill in the new slot before enabling interrupts again. */
      zzint_table[vecno]->arg[slot] = isr_arg;
      zzint_table[vecno]->isr[slot] = isr;
      zzint_table[vecno]->type[slot] = type;
      
      LOCK_RESTORE(msr);
   }
   else
   {
      /* Just have to find the first empty entry. */
      for (i = 0; i < ISR_NUM_ENTRIES; ++i)
      {
         if (zzint_table[vecno]->isr[i] == NULL)
         {
            slot = i;
            break;
         }      
      }

      /* Set the function pointer last since interrupts may be enabled for
       * this vector already. */
      zzint_table[vecno]->arg[slot] = isr_arg;
      zzint_table[vecno]->type[slot] = type;      
      zzint_table[vecno]->isr[slot] = isr;
   }
      
   return True;
}

/* The ISR is uniquely identified by the pointer, type and the
 * argument. (There may be multiple identical entries but in that case it does
 * not matter which one is removed..) */
void
zzrelease_vector_isr(OSVECTOR vecno, void *isr, uint32_t isr_arg, uint32_t type)
{
   int slot;
   struct intdata *item;
   Msr msr;

   /* Make sure the vector is valid and allocated. */
   if (!check_vector(vecno) || !is_vector_allocated(vecno))
   {
      return;
   }
   
   /* Find the slot to remove. */
   item = zzint_table[vecno];
   for (slot = 0; item->isr[slot] != NULL; ++slot)
   {
      if ((item->isr[slot] == isr) && (item->arg[slot] == isr_arg) && (item->type[slot] == type))
      {
         break;
      }
   }

   /* May not have found one. (Should not happen) */
   if (slot == ISR_NUM_ENTRIES)
   {
      return;
   }

   /* Perform compactation of all slots above the given. Have to
    * disable interrupts since we cannot move an entry atomically. */
   LOCK_SAVE(msr);
   for (; zzint_table[vecno]->isr[slot]; ++slot)
   {
      zzint_table[vecno]->isr[slot] = zzint_table[vecno]->isr[slot + 1];
      zzint_table[vecno]->arg[slot] = zzint_table[vecno]->arg[slot + 1];
      zzint_table[vecno]->type[slot] = zzint_table[vecno]->type[slot + 1];
   }
   LOCK_RESTORE(msr);

   /* Check if the entry should be freed. */
   free_vector(vecno);   
}

/* Calls the configured int_kill_handler to disable an interrupt
 * source. */
void
zzdisable_vector(OSVECTOR vecno)
{
   void (*handler)(OSVECTOR, OSPRIORITY) =
      (void (*)(OSVECTOR, OSPRIORITY)) zzconf.int_kill_handler;
      
   /* Validate the vector index. */
   if (!check_vector(vecno))
   {
      return;
   }

   /* Don't do anything for unused vector entries. */
   if (!is_vector_allocated(vecno))
   {
      return;
   }
   
   /* Call the int_kill_handle to disable the interrupt source. */
   if (handler)
   {
      handler(vecno, zzint_table[vecno]->pri);         
   }

   /* FIXME? SFK could setup an int_kill_handler. */
#ifdef softOSE
   zzrelease_vector_sfk(vecno, NULL, zzint_table[vecno]->pcb);
#endif
}

/* Calls the configured int_create_handler to enable an interrupt
 * source. */
void
zzenable_vector(OSVECTOR vecno, OSPRIORITY pri)
{
   void (*handler)(OSVECTOR, OSPRIORITY) =
      (void (*)(OSVECTOR, OSPRIORITY)) zzconf.int_create_handler;

   /* Validate the vector index. */
   if (!check_vector(vecno))
   {
      return;
   }

   /* Don't enable a vector that is unused. (Should not happen) */
   if (!is_vector_allocated(vecno))
   {
      return;
   }

   zzint_table[vecno]->pri = pri;
   
   /* Enable the interrupt source. */
   if (handler)
   {
      handler(vecno, pri);
   }

#ifdef softOSE
   zzenable_vector_sfk(vecno, NULL, zzint_table[vecno]->pcb, pri);
#endif
}
