/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2014 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */
 
/**
 * @file kernel_queue.c
 *
 * @brief Message queue implementation using OpenPIC IPIs. 
 *
 * @long This driver is used in SMP systems for carrying messages between
 * cores using IPI:s as the mechanism for notifying the destination CPU
 * that a kernel event is available in the message queue.
 *
 */
#include "stdint.h"
#include "string.h"
#include "assert.h"
#include "osenames.h"
#include "ose_smp.h"
#include "ramlog.h"
#include "ose_err.h"
#include "bsperr.h"
#include "compiler.h"
#include "sysparam.h"
#include "pose.h"
#include "cpu_smp.h"
#include "atomic.h"

/* The largest cache line supported. Each queue will get this alignment. */
#define LARGEST_SUPPORTED_CACHE_LINE 128

/* Default length of a kernel event queue */
#define EVENT_QUEUE_DEFAULT_LENGTH 64

/* Max number of signal queues (number of priorities for ose_fsend). */
#define MSG_QUEUE_MAX_COUNT 32

/* State of a message */
#define MSG_FREE    0
#define MSG_READY   1

#define	DEVOLATILE(type, var) \
  ((type)(ptrdiff_t)(volatile void *)(var))

/*
 * There is one event queue for each cpu, and each queue is cache aligned
 * for performance resons.
 * Each message must be  contained in a single cache line. This will guarantee
 * that the valid flag can not be seen set before the payload is also present.
 */

struct msg_entry
{
   volatile uint32_t msg[OSE_KRN_EVT_MSG_ENTRY_LEN];
   volatile uint32_t valid;
   uint32_t _pad;
};

/*
** The queue head.
** The queue is empty when first_ix == last_ix.
** Otherwise, first_ix is the next event to read,
** and last_ix is the last written.
*/
struct queue_head 
{
   volatile uint32_t first_ix;
   volatile uint32_t last_ix;
   uint32_t queue_len;
   uint32_t _pad1;
};

struct msg_queue 
{
  struct queue_head hd;  
  struct msg_entry msg_entries[1]; /* Dynamic limit. */
};

/* Queue pointer array for sendig kernel events between cores. */
static struct msg_queue *evt_queues[OSE_SMP_MAX_CPUS]
#ifdef __GNUC__
__attribute__ ((aligned (LARGEST_SUPPORTED_CACHE_LINE)))
#endif
;

/* Used for sending signals between cores. */
static struct msg_queue *msg_queues[OSE_SMP_MAX_CPUS][MSG_QUEUE_MAX_COUNT];

/* Number of queues for signals. default 1 */
static uint32_t msg_queue_count = 1;

/* Configurable queue length for signals. Default 64 */
static uint32_t msg_queue_len = 64;

/* Array of drop callback hooks, one per priority */
static OSFSENDHOOK *fsend_hook[MSG_QUEUE_MAX_COUNT];

#ifndef softOSE
/* If this symbol exist, install that as fallback since that indicates
 * we have an old BSP with the old event handling.
 */
WEAK_REF(void ipi_kernel_event_generate(cpuid_t dst));
#endif

#if defined(OSE_SMP)
#  if defined(__GNUC__) && defined(OSPP)
/*
** PowerPC GCC
*/
#    if defined(POWERPC_BOOK3E)
#      define ASYNC_WRITE_BARRIER()	__asm__ __volatile__(" lwsync \n" : : )
#      define SYNC_WRITE_BARRIER()	__asm__ __volatile__(" lwsync \n" : : )
#      define READ_READ_BARRIER()	__asm__ __volatile__(" lwsync \n" : : )
#    else
#      define ASYNC_WRITE_BARRIER()	__asm__ __volatile__(" sync \n" : : )
#      define SYNC_WRITE_BARRIER()	__asm__ __volatile__(" sync \n" : : )
#      define READ_READ_BARRIER()	/*__asm__ __volatile__(" isync \n" : : )*/ /*??*/
#    endif
#  elif defined(__GNUC__) && defined(OSMIPS)
/*
** MIPS GCC
** Cavium specific syncw (stype 4) will drain the lazy write buffer
** without stalling the cpu. Needed to make updates visible for 
** other cpus.
** On Broadcom cpus, the SYNCW executes as a normal SYNC.
*/
#      define MIPS_SYNCW		".long 0x10f \n"
#      define ASYNC_WRITE_BARRIER()	__asm__ __volatile__( MIPS_SYNCW : : )
#      define SYNC_WRITE_BARRIER()	__asm__ __volatile__( MIPS_SYNCW : : )
#      define READ_READ_BARRIER()	/*__asm__ __volatile__( MIPS_SYNCW : : )*/ /*??*/
#  elif defined(__GNUC__) && defined(OSARM)
/*
** ARM GCC
** ish: inner shared memory, st: only stores affected. dmb is faster than dsb.
** The read-read barrier is recommended by ARM Litmus Test.
** Note: dmb/dsb argument strings not supported by gcc_solaris2_arm_4.4.3,
** so the wanted options are only comments below!
*/
#      define ASYNC_WRITE_BARRIER()	__asm__ __volatile__(" dmb \n" : : ) /* ishst */
#      define SYNC_WRITE_BARRIER()	__asm__ __volatile__(" dsb \n" : : ) /* ishst */
#      define READ_READ_BARRIER()	__asm__ __volatile__(" dmb \n" : : ) /* ish */
#  elif defined(__ARMCC_VERSION) && defined(OSARM)
/*
** ARM RVCT
** Improvements required, needs a small amount of assembly programming,
** see cpu_atomic.h.
*/
#      define ASYNC_WRITE_BARRIER()	do { __force_stores(); _atomic_dmb(); } while(0)
#      define SYNC_WRITE_BARRIER()	do { __force_stores(); _atomic_dmb(); } while(0)
#      define READ_READ_BARRIER()	do { _atomic_dmb(); } while(0)
#  elif defined(OSX86)
/*
** X86
*/
#      warning "Currently assumed that OSX86 does not require barriers in kernel queues!"
#      define ASYNC_WRITE_BARRIER()
#      define SYNC_WRITE_BARRIER()
#      define READ_READ_BARRIER()  /* ?? */
#  else
#    warning "The used compiler/architecture combination is not supported for OSE_SMP."
#  endif
#else  /* ! defined(OSE_SMP) */
/* Unicore does not require barriers. */
#      define ASYNC_WRITE_BARRIER()
#      define SYNC_WRITE_BARRIER()
#      define READ_READ_BARRIER()
#endif  /* ! defined(OSE_SMP) */


/*
** Ramlog dump of a queue from a frozen context for post mortem analysis.
** Currently only used for kernel event queues in multicore libraries.
*/
static void
dump_queue(struct msg_queue *que, cpuid_t cpu)
{
#ifdef OSE_SMP
  uint32_t first = que->hd.first_ix;
  uint32_t last;
  uint32_t ix;

  last = que->hd.last_ix;
  last = last ? last - 1 : que->hd.queue_len - 1;
  if(cpu >= OSE_SMP_MAX_CPUS  ||
     first >= que->hd.queue_len || last >= que->hd.queue_len)
    {
      return;
    }

  ramlog_printf("DUMPING MESSAGES QUEUED TO cpu%lu (from cpu%lu):\n",
		cpu, zzkrn_cpu_id());
  ramlog_printf("MESSAGE QUEUE next to read: %lu, last written: %lu.\n",
		first, last);
  ramlog_printf("ind q message                ind q message\n");
  for (ix = 0; ix < que->hd.queue_len; ix++)
    {
      char que_flag;

      if(first == last)
	{
	  que_flag = ' ';
	}
      else if(ix == first)
	que_flag = '@';
      else if(first <= last)
	{
	  que_flag = (first < ix && ix <= last) ? '*' : ' ';
	}
      else
	{
	  que_flag = (first < ix || ix <= last) ? '*' : ' ';
	}
      ramlog_printf("%3lu %c %#010lx %#010lx%s",
		    ix, que_flag,
		    que->msg_entries[ix].msg[0], que->msg_entries[ix].msg[1],
		    ((ix & 1) || ix == que->hd.queue_len - 1)  ? " \n" : "  ");
    }
#else
  (void)que;
  (void)cpu;
#endif
}

/* Ramlog dump of a queue from a frozen context for post mortem analysis. */
void
zzkrn_evt_dump_queue(cpuid_t cpu)
{
  dump_queue(evt_queues[cpu], cpu);
}

static int
msg_queue_enqueue(struct msg_queue *msg_queue, const uint32_t *krn_evt_buf)
{
   uint32_t next_ix;
   uint32_t last_ix;

   for (;;)
   {
      last_ix = msg_queue->hd.last_ix;
      next_ix = last_ix + 1;
      if(next_ix == msg_queue->hd.queue_len)
	next_ix = 0;
    
      if (unlikely(msg_queue->hd.first_ix == next_ix))
      {
	 return 1; /* Queue is full. */
      }

      if (atomic_cas(DEVOLATILE(atomic_t *, &msg_queue->hd.last_ix),
		     last_ix, next_ix)
	  == last_ix)
      {   
	 /* We have now reserved this entry, enqueue the event. */  
	 struct msg_entry *msg_entry = &msg_queue->msg_entries[last_ix];
	 uint32_t ix;
	 for (ix = 0; ix < OSE_KRN_EVT_MSG_ENTRY_LEN; ix++)
	 {
	    msg_entry->msg[ix] = krn_evt_buf[ix];
	 }
	 ASYNC_WRITE_BARRIER();
	 msg_entry->valid = MSG_READY;
	 SYNC_WRITE_BARRIER();
	 return 0; /* Success */
      }
   }
}

static uint32_t
msg_queue_dequeue(struct msg_queue *msg_queue, uint32_t *krn_evt_buf)
{
   struct msg_entry *msg_entry;
   uint32_t ix;
   uint32_t next_ix;
   
   if (msg_queue->hd.last_ix == msg_queue->hd.first_ix)
   {
      /* No kernel event available in queue. */
      return 0;
   }
   
   msg_entry = &msg_queue->msg_entries[msg_queue->hd.first_ix];
   
   while (msg_entry->valid == MSG_FREE)
   {
      /* 
       * Loop until the message we want to read is
       * enqueued completely by peer cpu. 
       */
   }

   /*
   ** Read msg after valid flag. Barrier recommended for ARM, 
   ** see ARM Litmus test 6.2.1.1.
   */
   READ_READ_BARRIER();

   for (ix = 0; ix < OSE_KRN_EVT_MSG_ENTRY_LEN; ix++)
   {
      krn_evt_buf[ix] = msg_entry->msg[ix];
   }
    
   /*
   ** FIXME this barrier might not be needed?? No sign of that yet.
   ** If so, use this: ASYNC_WRITE_BARRIER()
   */
   msg_entry->valid = MSG_FREE;
   next_ix = msg_queue->hd.first_ix + 1;
   if(next_ix == msg_queue->hd.queue_len)
     next_ix = 0;
   msg_queue->hd.first_ix = next_ix;
   return OSE_KRN_EVT_MSG_ENTRY_LEN;
}

void
zzkrn_evt_init(void)
{
   cpuid_t num_cpus = zzkrn_num_cpus();
   
   if(num_cpus > OSE_SMP_MAX_CPUS)
   {
      error2(OSE_EFATAL_MASK | OSE_BSP_EKRN_EVT_MAX_CPUS_EXCEEDED,
	     OSE_SMP_MAX_CPUS);
   }

   /* Get kernel event queue configuration. */
#ifdef OSE_SMP
   {
     uint32_t length;
     size_t evt_queue_net_size;
     size_t evt_queue_tot_size;
     size_t bulk_size;
     char *bulk;
     cpuid_t ix;

     length = 0;
     sysparam_readattr_long("krn/event_queue", "length", (long *)&length);
     if(!length)
       length = EVENT_QUEUE_DEFAULT_LENGTH;

     /* Compute padding to fill up some entire cache lines. */
     evt_queue_net_size = sizeof(struct queue_head) +
       sizeof(struct msg_entry) * length;
     evt_queue_tot_size =
       (evt_queue_net_size + (LARGEST_SUPPORTED_CACHE_LINE - 1)) &
       ~(LARGEST_SUPPORTED_CACHE_LINE - 1);
     bulk_size = evt_queue_tot_size * num_cpus;
     bulk = zzalloc_kernelmemory((OSBUFSIZE)bulk_size);
     memset(bulk, 0, bulk_size);
     /* Cache line align the queues. */
     bulk = (char *)((((size_t)bulk) + (LARGEST_SUPPORTED_CACHE_LINE - 1)) &
		     ~(LARGEST_SUPPORTED_CACHE_LINE - 1));
     for(ix = 0; ix < num_cpus; ix++)
       {
	 evt_queues[ix] = (struct msg_queue *)bulk;
	 evt_queues[ix]->hd.queue_len = length;
	 bulk += evt_queue_tot_size;
       }
     ramlog_printf("exec: krn/event_queue using queues of length %lu.\n",
		   length);
   }
#endif

   /* Get fsend msg queue configuration. */
   sysparam_readattr_long("krn/fsend_queue", "count",(long *)&msg_queue_count);
   sysparam_readattr_long("krn/fsend_queue", "length",(long *)&msg_queue_len);

   /* Sanity check of the number of queues. */
   if (msg_queue_count > 32)
   {
      ramlog_printf("exec: krn/fsend_queue=count set to 32.\n");
      msg_queue_count = 32;
   }
   ramlog_printf("exec: krn/fsend_queue using %lu queue(s) of length %lu.\n",
		 msg_queue_count, msg_queue_len);

   /* Allocate memory for msg queues. */
   {
     uint32_t size = sizeof(struct queue_head) +
       sizeof(struct msg_entry) * msg_queue_len; 
      uint32_t buf, prio;
      cpuid_t cpu;

      for (cpu = 0; cpu < num_cpus; cpu++)
      {
	 for (prio = 0; prio < msg_queue_count; prio++)
	 {
	    /* Alloc memory, add space for alignment. */
	    buf = (uint32_t)
	      zzalloc_kernelmemory(size + LARGEST_SUPPORTED_CACHE_LINE);
	    /* Align pointer. */
	    buf = (buf + LARGEST_SUPPORTED_CACHE_LINE) &
	       ~(LARGEST_SUPPORTED_CACHE_LINE - 1);
	    msg_queues[cpu][prio] = (struct msg_queue *)buf;
	    memset(msg_queues[cpu][prio], 0, size);
	    msg_queues[cpu][prio]->hd.queue_len = msg_queue_len;
	 }
      }
   }
   /* Install handlers for pose ose_fsend. */
   zzkrn_set_handler(14, (void *)zzkrn_msg_enqueue);
   zzkrn_set_handler(15, (void *)zzkrn_msg_dequeue);

#ifndef softOSE   
   /* If old ipi handler present in the linker scope, install it. */
   if(ipi_kernel_event_generate != NULL)
   {
      /* Call get_sysconf to avoid dependency to pflib.
       * ose_smp_set_evt_interrupter(ipi_kernel_event_generate);
       */
      get_sysconf(90, 11 << 16, (OSADDRESS)ipi_kernel_event_generate, 0);
      ramlog_printf("exec: older ipi driver found, using that.\n");
   }
#endif   
}

int
zzkrn_set_fsend_hook(int sig_prio, OSFSENDHOOK *hook)
{
   if((sig_prio >= 0) && (sig_prio < MSG_QUEUE_MAX_COUNT))
   {
      fsend_hook[sig_prio] = hook;
      return 0;
   }
   else
   {
      return 1;
   }
}

#ifdef OSE_SMP

uint32_t
zzkrn_evt_enqueue(cpuid_t peer,
                  const uint32_t *krn_evt_buf,
                  uint32_t count)
{
  int rc;

  (void)count;
  assert(OSE_KRN_EVT_MSG_ENTRY_LEN >= count);

  rc = msg_queue_enqueue(evt_queues[peer], krn_evt_buf);
  return rc;
}

uint32_t
zzkrn_evt_dequeue(uint32_t *krn_evt_buf)
{
  uint32_t rc = msg_queue_dequeue(evt_queues[zzkrn_cpu_id()], krn_evt_buf);
  return rc;
}

#endif	/* OSE_SMP */

uint32_t
zzkrn_msg_enqueue(cpuid_t peer, const struct ose_krn_msg *krn_msg_buf,
		  uint32_t flags)
{
   uint32_t prio = flags & 0x1f; /* Range 0-31 */
   struct msg_queue *queue;
   OSFSENDHOOK *hook;
   
   if (prio >= msg_queue_count)
   {
      /* Use lowest prio. */
      prio = msg_queue_count - 1;
   }
   queue = msg_queues[peer][prio]; /* Pick up the selected queue. */
   hook = fsend_hook[prio];

   /* Let the call back function decide if we should enqueue or drop. */
   if (hook)
   {
      uint32_t elems_in_queue;
      uint32_t first = queue->hd.first_ix;
      uint32_t last = queue->hd.last_ix;
      
      if(first == last)
      {
         elems_in_queue = 0;
      }
      else
      {
         elems_in_queue = msg_queue_len + last - first;
         if(elems_in_queue > msg_queue_len)
         {
            elems_in_queue -= msg_queue_len;
         }
      }
      
      if (hook(prio, elems_in_queue))
      {
         return 1; /* Drop due to custom algoritm, which can be for example
                    * WRED or tail-drop.
                    * Drop when full is also a way to define unreliability
                    * on a given priority. */
      }
   }
   return msg_queue_enqueue(queue, krn_msg_buf->data);
   
#if 0
   if (likely(0 == msg_queue_enqueue(queue, krn_msg_buf->data)))
   {
      return 0;
   }
   if(hook)
   {
      ramlog_printf("The fsend hook did not drop signal, ");
   }
   else
   {
      ramlog_printf("There is no fsend hook defined, ");
   }
   ramlog_printf("and the inter-core msg queue became full when ose_fsend "
                 "on prio %u to cpu: %d was done.\n", prio, peer);
   error2(OSE_EKRN_MSG_QUEUE_FULL, peer);
   /* Fatal system error */
   return 1;
#endif   
}

uint32_t
zzkrn_msg_dequeue(struct ose_krn_msg *krn_msg_buf, uint32_t elems)
{
   cpuid_t cpu = zzkrn_cpu_id();
   uint32_t elem_count = 0;
   uint32_t prio;

   /* Check all prio queues, start with prio 0. */
   for (prio = 0; prio < msg_queue_count; prio++)
   {
      /* Loop as long as there are messages in the current prio queue. */
      while (msg_queue_dequeue(msg_queues[cpu][prio], krn_msg_buf->data))
      {
	 if (++elem_count == elems)
	 {
	    /* Max number of elements dequeued. */
	    return elem_count;
	 }
	 krn_msg_buf++;	/* Increment dest pointer. */
      }
   }
   return elem_count;
}
