/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2003-2015 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

/*
 * Description:
 *   Implements the cache SPI for MIPS architecture.
 */



/* Public interfaces. */
#include "cache.h"
#include "ose.h"
#include "assert.h"
#include "stdarg.h"
#include "ramlog.h"

/* Kernel internal interfaces. */
#include "cpu_hal_arch.h"
#include "cache_trap.h"


/* MM internal interfaces. */
#include "hal.h"

/* Cpu register access functions. */
#include "cpu_access.h"

/*
 * Flush dcache one cache line.
 */
static __inline__ void
FLUSH_AND_INVALIDATE_CACHE_LINE_D(unsigned long line_addr)
{
   __asm__ volatile
   (
      ".set push;"
      ".set noreorder;"

      /* Move argument line_addr to register $4. */
      " addi $4,%0,0;"

      /* Flush dcache */
      " cache 0x15,0($4);"
      " sync;"

      ".set pop;"

      :/* output */
      :/* input */ "r" (line_addr)
      :/* modifies */ "$4"
   );
}

/*
 * Invalidate dcache one cache line.
 */
static __inline__ void
INVALIDATE_CACHE_LINE_D(unsigned long line_addr)
{
   __asm__ volatile
   (
      ".set push;"
      ".set noreorder;"

      /* Move argument line_addr to register $4. */
      " addi $4,%0,0;"

      /* Invalid dcache */
      " cache 0x11,0($4);"
      " nop;"  /* Empty pipeline. */
      " nop;"
      " nop;"
      " nop;"

      ".set pop;"

      :/* output */
      :/* input */ "r" (line_addr)
      :/* modifies */ "$4"
   );
}

/*
 * Invalidate icache one cache line.
 */
static __inline__ void
INVALIDATE_CACHE_LINE_I(unsigned long line_addr)
{
   __asm__ volatile
   (
      ".set push;"
      ".set noreorder;"

      /* Move argument line_addr to register $4. */
      " addi $4,%0,0;"

      /*  ICACHE_INVALIDATE_WORKAROUND, taken from yamon monitor src.
       *
       *  Workaround for bug in early revisions of MIPS 4K family of
       *  processors. Only relevant in early engineering samples of test
       *  chips (RTL revision <= 3.0).
       *
       *  The bug is described in :
       *
       *  MIPS32 4K(tm) Processor Core Family RTL Errata Sheet
       *  MIPS Document No: MD00003
       *
       *  The bug is identified as : C16
       */
      " la	$9,999f;"
      " cache 0x14,0($9);"
      " sync;"
      " nop;" /* Empty pipeline. */
      " nop;"
      " nop;"
      " nop;"
      "999:;"

      /* Invalid icache */
      " cache 0x10,0($4);"
      " nop;"  /* Empty pipeline. */
      " nop;"
      " nop;"
      " nop;"

      ".set pop;"

      :/* output */
      :/* input */ "r" (line_addr)
      :/* modifies */ "$4", "$9"
   );
}

/*
 * Invalidate icache one cache line.
 */
static __inline__ void
INVALIDATE_CACHE_INDEX(unsigned long index)
{
   __asm__ volatile
   (
      ".set push;"
      ".set noreorder;"

      /* Move argument index to register $4. */
      " addi $4,%0,0;"

      /*  ICACHE_INVALIDATE_WORKAROUND, taken from yamon monitor src.
       *
       *  Workaround for bug in early revisions of MIPS 4K family of
       *  processors. Only relevant in early engineering samples of test
       *  chips (RTL revision <= 3.0).
       *
       *  The bug is described in :
       *
       *  MIPS32 4K(tm) Processor Core Family RTL Errata Sheet
       *  MIPS Document No: MD00003
       *
       *  The bug is identified as : C16
       */
      " la	$9,999f;"
      " cache 0x14,0($9);"
      " sync;"
      " nop;" /* Empty pipeline. */
      " nop;"
      " nop;"
      " nop;"
      "999:;"

      /* Invalid icache per index */
      " cache 0x0,0($4);"
      " nop;"  /* Empty pipeline. */
      " nop;"
      " nop;"
      " nop;"

      ".set pop;"

      :/* output */
      :/* input */ "r" (index)
      :/* modifies */ "$4", "$9"
   );
}

/* Check if cfg1 is present. */
static int
has_config_sel1(OSADDRESS cfg0)
{
   return (cfg0 & 0x80000000) != 0;
}

#define RELY_ON_CPU_HAL_CACHE_INFO
#ifdef RELY_ON_CPU_HAL_CACHE_INFO

/*
** Return the smallest of L1 D/I cache line sizes.
*/
static __inline__ OSADDRESS
get_cache_line_size(OSADDRESS *cache_line_mask)
{
   OSADDRESS dls;
   OSADDRESS ils;
   OSADDRESS cls;
   
   if (hal_arch_cpu_descriptor != NULL)
   {
      struct cache_info *ci = &hal_arch_cpu_descriptor->cache_info;

      dls = ci->d_cache_line_size;
      ils = ci->i_cache_line_size;
   }
   else
   {
     /*
     ** Extremely early calls in the boot sequence may enter here, 
     ** for example to handle a ramlog write.
     */
     OSADDRESS cfg0 = read_cp0(CP0_CONFIG, 0);

     if (has_config_sel1(cfg0))
     {
	OSADDRESS cfg1 = read_cp0(CP0_CONFIG, 1);

	dls = 2 << ((cfg1 >> 10) & 0x7);
	ils = 2 << ((cfg1 >> 19) & 0x7);
     }
     else
     {
	dls = (cfg0 & 0x10) ? 32 : 16;
	ils = (cfg0 & 0x20) ? 32 : 16;
     }
   }

   /* Use smallest non-zero cache line size. Zero means no such cache. */
   if(0 == ils)
       cls = dls;
   else
     {
       if(0 == dls)
	 cls = ils;
       else
	 cls = dls < ils ? dls : ils;
     }
   *cache_line_mask = ~(cls - 1);
   return cls;
}

#else /* Not RELY_ON_CPU_HAL_CACHE_INFO */

/*
 * get_cache_line_size -Information about cache line size is available in
 * CP0/CONFIG1, bits [21:19] for Icache. bits [12:10] Dcache.
 */
static __inline__ OSADDRESS
get_cache_line_size(OSADDRESS *cache_line_mask)
{
   OSADDRESS cfg0 = read_cp0(CP0_CONFIG, 0);
   OSADDRESS cfg1;
   OSADDRESS dls;
   OSADDRESS ils;
   OSADDRESS cls;

   if (has_config_sel1(cfg0))
   {
      cfg1 = read_cp0(CP0_CONFIG, 1);
      dls  = 2 << ((cfg1 >> 10) & 0x7);
      ils  = 2 << ((cfg1 >> 19) & 0x7);
   }
   else
   {
      dls = (cfg0 & 0x10) ? 32 : 16;
      ils = (cfg0 & 0x20) ? 32 : 16;
   }

   assert(dls == 16 || dls == 32);
   assert(ils == 16 || ils == 32);

   /* Use smallest cache line size. */
   cls = dls < ils ? dls : ils;

   *cache_line_mask = ~(cls - 1);
   return cls;
}

#endif /* Not RELY_ON_CPU_HAL_CACHE_INFO */

/*
 * get_Icache_sets_per_way -Information about cache sets_per_way/associtivity
 * is available in CP0/CONFIG1, bits [24:22]/[18:16].
 */
static __inline__ OSADDRESS
get_Icache_info(OSADDRESS *associativity)
{
   OSADDRESS cfg1 = read_cp0(CP0_CONFIG, 1);
   OSADDRESS sets_per_way  = 64 << ((cfg1 >> 22) & 0x7);

   assert(sets_per_way == 64 || sets_per_way == 128 ||
          sets_per_way == 256);

   *associativity =  ((cfg1 >> 16) & 0x7) + 1;

   assert(*associativity == 1 || *associativity == 2 ||
          *associativity == 3 || *associativity == 4);

   return sets_per_way;
}

/*
 * See cache.h
 */
void
cache_invalidate_d(OSADDRESS base, size_t size)
{
   OSADDRESS cache_line_mask;
   OSADDRESS cache_line_size = get_cache_line_size(&cache_line_mask);
   OSADDRESS addr = base & cache_line_mask;
   OSADDRESS end_line = (base + size - 1) & cache_line_mask;

   do
   {
      INVALIDATE_CACHE_LINE_D(addr);
      addr += cache_line_size;
   } while (addr <= end_line);
}

/*
 * See cache.h
 */
void
cache_flush_invalidate_d(OSADDRESS base, size_t size)
{
   OSADDRESS cache_line_mask;
   OSADDRESS cache_line_size = get_cache_line_size(&cache_line_mask);
   OSADDRESS addr = base & cache_line_mask;
   OSADDRESS end_line = (base + size - 1) & cache_line_mask;

   do
   {
      FLUSH_AND_INVALIDATE_CACHE_LINE_D(addr);
      addr += cache_line_size;
   } while (addr <= end_line);
}

/*
 * See cache.h
 */
void
cache_invalidate_i(OSADDRESS base, size_t size)
{
   OSADDRESS cache_line_mask;
   OSADDRESS cache_line_size = get_cache_line_size(&cache_line_mask);
   OSADDRESS addr = base & cache_line_mask;
   OSADDRESS end_line = (base + size - 1) & cache_line_mask;

   do
   {
      INVALIDATE_CACHE_LINE_I(addr);
      addr += cache_line_size;
   } while (addr <= end_line);
}


void
cache_invalidate_i_complete(int disable_i)
{
   OSADDRESS cfg0 = read_cp0(CP0_CONFIG, 0);
   OSADDRESS cache_line_mask;
   OSADDRESS cache_line_size = get_cache_line_size(&cache_line_mask);
   OSADDRESS addr, line;
   OSADDRESS icache_lines;
   
   (void) disable_i;

   if (has_config_sel1(cfg0))
   {
      OSADDRESS associativity;
      OSADDRESS sets_per_way;
      sets_per_way = get_Icache_info(&associativity);
      icache_lines = sets_per_way * associativity;
   }
   else
   {
      OSADDRESS i_cache_size = 0x1000 << ((cfg0 >> 9) & 0x7);

      icache_lines = i_cache_size / cache_line_size;
   }

   /* Invalidate ICACHE */
   for( line = 0, addr = KSEG0(0);
        line < icache_lines;
        line++, addr += cache_line_size )
   {
      INVALIDATE_CACHE_INDEX( addr );
   }
}

/*
 * See cache.h
 */
void
cache_flush_d(OSADDRESS base, size_t size)
{
   cache_flush_invalidate_d(base, size);
}

/*
 * See cache.h
 */
void
cache_sync(OSADDRESS base, size_t size)
{
   cache_flush_d(base, size);
   cache_invalidate_i(base, size);
}

/*
 * See cache.h
 */
void
cache_flush_d_complete(int disable_d)
{
   /* The CPU HAL knows how to empty the entire data cache subsystem. */
   hal_arch_cpu_descriptor->clean_d_cache(disable_d);

}

/*
 * See cache.h
 */
Boolean
cache_initialized(void)
{
  return (hal_arch_cpu_descriptor != 0);
}

/*
 * See cache.h
 */

U32
cache_get_characteristics(U32 query)
{
   U32 result;
   struct cache_info *cache_info;

   if (query == CACHE_QUERY_INITIALIZED)
   {
     return (U32)cache_initialized();
   }

   result = CACHE_QUERY_NOT_SUPPORTED;

   if (hal_arch_cpu_descriptor != NULL)
   {
      cache_info = &hal_arch_cpu_descriptor->cache_info;

      switch (query)
      {
         case CACHE_QUERY_D_CACHE_LINE_SIZE:
         {
            result = cache_info->d_cache_line_size;
            break;
         }
         case CACHE_QUERY_I_CACHE_LINE_SIZE:
         {
            result = cache_info->i_cache_line_size;
            break;
         }
         case CACHE_QUERY_D_CACHE_SIZE:
         {
            result = cache_info->l1_d_cache_size;
            break;
         }
         case CACHE_QUERY_I_CACHE_SIZE:
         {
            result = cache_info->l1_i_cache_size;
            break;
         }
         case CACHE_QUERY_D_CACHE_STATUS:
         {
            result = cache_info->d_cache_enabled;
            break;
         }
         case CACHE_QUERY_I_CACHE_STATUS:
         {
            result = cache_info->i_cache_enabled;
            break;
         }
         case CACHE_QUERY_L2_CACHE_STATUS:
         {
            result = cache_info->i_cache_enabled;
            break;
         }
         case CACHE_QUERY_L2_CACHE_SIZE:
         {
            result = cache_info->L2_cache_size;
            break;
         }
         default:
         {
            break;
         }
      }
   }
   else
   {
      ramlog_printf("cache_get_characteristics(): "
                    "cache subsystem not yet initialized.\n");
   }
   return result;
}



/*
 * See cache.h
 */
void
zzose_cache(U32 action, U32 level, OSADDRESS base, size_t size)
{
   if (hal_arch_cpu_descriptor->cache != NULL)
   {
      hal_arch_cpu_descriptor->cache(action, level, base, size);
   }
   else
   {
      /*
      ** No CPU-HAL installed that offers the extended cache interface.
      ** Make the best out of it by calling the corresponding internal
      ** functions.
      */
      int complete = (base == 0 && size == 0) ? 1: 0;
      
      switch (action)
      {
	 case CACHE_FLUSH_D:
	    complete ?
	       cache_flush_d_complete(0):	  /* Disable not supported */
	       cache_flush_d(base, size);
	    break;
	 case CACHE_INVALIDATE_D:
	    complete ?
	       cache_invalidate_d(base, (size_t)-1):
	       cache_invalidate_d(base, size);
	    break;
	 case CACHE_INVALIDATE_I:
	    complete ?
	       cache_invalidate_i_complete(0):	  /* Disable not supported */
	       cache_invalidate_i(base, size);
	    break;
	 default:
	    break;
      }
   }
}

/*
 * See cache.h
 */
U32
zzose_cache_query(U32 query, ...)
{
   U32 result = cache_get_characteristics(query);
      
   if (result == (U32)CACHE_QUERY_NOT_SUPPORTED &&
       query == CACHE_QUERY_LEVEL)
   {
      if (hal_arch_cpu_descriptor->cache_level != NULL)
      {
	 /*
	 ** Pass this request to the CPU-HAL
	 */
	 va_list ap;
	 U32 level;
	 
	 va_start(ap, query);
	 level = va_arg(ap, U32);
	 
	 result = hal_arch_cpu_descriptor->cache_level(level);
	 va_end(ap);
      }
      else
      {
	 /*
	 ** No CPU-HAL installed that offers the extended cache interface.
	 ** Make the best out of it by returning the outermost physical level.
	 */
	 result = CACHE_L9;
      }
   }
   return result;
}
