/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2005-2006 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

#include "string.h"
#include "ose.h"
#include "osetypes.h"
#include "exception.h"
#include "cpu.h"
#include "epm.h"
#include "cache.h"
#include "epm_internal.h"
#include "mm_access.h"
#include "mm.sig"
#include "mm.h"
#include "assert.h"
#include "exec_arch.h"
#include "cpu_hal_arch.h"
#include "efmi_registers.h"
#include "ramlog.h"
#include "epm_private.h"

#define OPCODE_J(address)           (2<<26 | ((address>>2) & 0x03FFFFFF))
#define OPCODE_NOP (0)
#define JALR_TO_JR_MASK (~(0x0000F801UL))
#define EPC_MASK (~(0xFFFFFFFFULL))
#define ADDR_MASK (~(0x0FFFFFFFUL))

#define OPCODE_AREA_MIN_SIZE 8

/* For execution points use reserved instruction. */
#define OPCODE_RESERVED_INSTRUCTION 0xec000000

#define FREEZE_MODE 1

#define LINK_ADDRESS_OFFSET 5
#define LINK_REGISTER_OFFSET 7

//#define EPM_ENABLE_PRINTF
#ifdef EPM_ENABLE_PRINTF
#include "stdio.h"
extern uint32_t verbose;
#define ERRLOG(a) if(verbose) ramlog_printf a
#define DLOG(a) if(verbose) ramlog_printf a
#else
#define DLOG(a) ((void) 0)
#define ERRLOG(a) ramlog_printf a
#endif

/* Help struct for intframe parameter to zzpreprocess_trace. */
struct Intframe
{
   U32 pad[4];
   U32 gpr[32];
};

extern PROCESS zzkrn_current_process(void);

void
zzepm_install_exception_handler(exception_handler_c handler)
{
   ose_install_c_exception_handler(EXCEPTION_ID_RI, handler, EXCEPTION_NORMAL);
}

OSADDRESS
zzepm_get_dump_pc(struct cpu_regdump *dump)
{
   return dump->pc;
}

OSADDRESS
zzepm_get_trace_to_pc(struct cpu_regdump *dump, OSADDRESS instr)
{
   struct cpu_core *core = (struct cpu_core *) dump->sect[0].address;
   OSADDRESS next_pc;
   U32 instr_type;

   instr_type = zzpreprocess_trace(instr, core->epc, &next_pc);

   switch (instr_type)
   {
      case INST_NORMAL:
      case INST_TRAP:
      case INST_SYSCALL: /* Step over. */
         return next_pc;
      default:
         if (BR(instr_type) || BC(instr_type))
         {
            /* Generate an intframe pointer. */
            struct Intframe *intframe = (struct Intframe *)
               ((char *) core - offsetof(struct Intframe, gpr));

            /* Calculate target adress. */
            zzget_trace_address(instr_type, instr, core->epc,
                     intframe, &next_pc);
            /*
             * Conditional branches to itself must be stepped over. Example:
             *  0x100:  bdnz 0x100
             */
            if (core->epc == next_pc)
            {
               next_pc += 8; /* Next instruction + branch delay. */
            }
            return next_pc;
         }
         return core->epc + 4;
   }
   return 0;
}

static int
function_call_opcode(OSADDRESS instr)
{
   OSADDRESS masked_instr;

   masked_instr = instr & 0xFC000000;
   if (masked_instr == 0x0C000000) /* jal */
   {
      return 1; /* jal */
   }

   masked_instr = instr & 0xFC1C0000;
   if (masked_instr == 0x04100000) /* bgezal(l) and bltzal(l) */
   {
      return 2; /* bgezal(l) and bltzal(l) */
   }
   return 0;
}

OSADDRESS
zzepm_get_next_pc(struct cpu_regdump *dump, U32 options, OSADDRESS instr)
{
   extern void zzsystem_callh_supervisor(void);
   extern void zzbios_callh(void);

   struct cpu_core *core = (struct cpu_core *) dump->sect[0].address;
   OSADDRESS next_pc;
   OSADDRESS next_opcode;
   U32 instr_type;

   DLOG(("epm_arch.c(%d): zzepm_get_next_pc(): core->epc:0x%08X,"
         "options:0x%08X, instr:0x%08X) \n",
         __LINE__, core->epc, options, instr));

   if ((options & EPM_SLIDE_OVER_FUNCTION_CALLS) &&
       function_call_opcode(instr))
   {
      DLOG(("epm_arch.c(%d): Just slide \n", __LINE__));
      next_pc = core->epc + 8;
   }
   else
   {
      instr_type = zzpreprocess_trace(instr, core->epc, &next_pc);

      switch (instr_type)
      {
         case INST_NORMAL:
         case INST_TRAP:
            break;
         case INST_SYSCALL: /* Step over. */
            DLOG(("epm_arch.c(%d): System call, stepping over \n", __LINE__));
            next_pc += 4;
            break;
         default:
            if (BR(instr_type) || BC(instr_type))
            {
               /* Generate an intframe pointer. */
               struct Intframe *intframe = (struct Intframe *)
                  ((char *) core - offsetof(struct Intframe, gpr));

               /* Calculate target adress. */
               zzget_trace_address(instr_type, instr, core->epc,
                        intframe, &next_pc);
               /*
                * Conditional branches to itself must be stepped over. Example:
                *  0x100:  bdnz 0x100
                */
               if (core->epc == next_pc)
               {
                  DLOG(("epm_arch.c(%d): Conditional branch to self, step over \n",
                           __LINE__));
                  next_pc += 8; /* Next instruction + branch delay. */
               }
               else if (next_pc == (OSADDRESS) zzsystem_callh_supervisor ||
                        next_pc == (OSADDRESS) zzbios_callh)
               {
                  DLOG(("epm_arch.c(%d): Go to return address \n", __LINE__));
                  next_pc = core->r[31];
               }
               else
               {
                  DLOG(("epm_arch.c(%d): Next PC from branch @ 0x%08X = 0x%08X \n",
                           __LINE__, core->epc, next_pc));
               }
            }
            else
            {
               DLOG(("epm_arch.c(%d): No branch, just grab next \n", __LINE__));
               next_pc = core->epc + 4;
            }
            break;
      }
   }

   /* We can not trace a system call! */
   if (EPM_OK == zzepm_get_opcode(next_pc, &next_opcode))
   {
      if ((next_opcode & 0xFC00003F) == 0x0000000C) /* Syscall */
      {
         DLOG(("epm_arch.c(%d): Skip system calls \n", __LINE__));
         next_pc += 4;
      }
   }
   return next_pc;
}

OSADDRESS
zzepm_get_sw_ep(OSADDRESS address, OSADDRESS *opcode, int *bytes_in_opcode)
{
   *opcode = OPCODE_RESERVED_INSTRUCTION;
   *bytes_in_opcode = 4;
   return address;
}

void
zzepm_set_sw_ep(OSADDRESS address)
{
   zzepm_set_opcode(address, OPCODE_RESERVED_INSTRUCTION);
}

void
zzepm_set_opcode(OSADDRESS address, OSADDRESS opcode)
{
   OSADDRESS src = opcode;
   int status;

   DLOG(("epm_arch.c(%d): zzepm_set_opcode(0x%08X, 0x%08X)\n",
            __LINE__, address, opcode));

   status = mm_access_forced_write(MM_KERNEL_DOMAIN, (void *) address,
            &src, 4, FREEZE_MODE);
   if (!(status == MM_SUCCESS || status == MM_ENO_PHYSICAL_MAPPING))
   {
      ERRLOG(("epm_arch.c(%d): forced write failed 0x%lx 0x%lx status:%x\n",
                        __LINE__, address, &src, status));
   }

   ose_cache(CACHE_INVALIDATE_I, CACHE_PGM_LOAD, (OSADDRESS) 0,
            0);
}

EpmStatus
zzepm_get_opcode(OSADDRESS address, OSADDRESS *opcode)
{
   OSADDRESS pre_address = address - 4;
   OSADDRESS next_pc;
   phys_addr_t pa;
   U32 instr_type;
   U32 pre_opcode;

   assert(address != 0);

   /* FIXME This does not work for a paged load module. */
   /* Check that pre_address is mapped. */
   /* Use zzkrn_current_process to avoid executing crt code. */
   if (ose_mm_l2p(zzkrn_current_process(), pre_address, &pa) == MM_SUCCESS)
   {
      /* Call epm to search list for an ep. Avoid reading memory if an ep
       * exists on pre_address.
       */
      if (epm_get_opcode_from_list(pre_address, &pre_opcode) == 0)
      {
         /* No ep exists on pre_address, read memory. */
         pre_opcode = *(OSADDRESS *) pre_address;
      }

      /* Check if address points to a branch delay. */
      instr_type = zzpreprocess_trace(pre_opcode, pre_address, &next_pc);
      if (BR(instr_type) || BC(instr_type))
      {
         /* Tell epm to avoid setting an ep in a branch delay. */
         *opcode = 0;
         return EPM_INVALID_INSTRUCTION;
      }
   }

   *opcode = *(OSADDRESS *) address;

   DLOG(("epm_arch.c(%d): zzepm_get_opcode(0x%08X, 0x%08X)\n",
               __LINE__, address, *opcode));

   return EPM_OK;
}

int
zzepm_emulate_instruction(struct cpu_regdump *dump, OSADDRESS opcode)
{
   struct cpu_core *core = (struct cpu_core *) dump->sect[0].address;
   OSADDRESS next_pc;
   U32 instr_type;

   instr_type = zzpreprocess_trace(opcode, core->epc, &next_pc);
   if (instr_type == INST_MTC0_STATUS_REG)
   {
      U32 rt = RT(opcode);
#if 0
      DLOG("epm_arch.c(%d): MTC0 sr: %x r: %x rt: %d\n",
               __LINE__, core->sr, core->r[rt], rt);
#endif
      core->sr = core->r[rt];
      core->epc += 4;
      return 1; /* Instruction emulated. */
   }
   if (instr_type == INST_MFC0_STATUS_REG)
   {
      U32 rt = RT(opcode);
#if 0
      DLOG("epm_arch.c(%d): MFC0 sr: %x r: %x rt: %d\n",
               __LINE__, core->sr, core->r[rt], rt);
#endif
      core->r[rt] = core->sr;
      core->epc += 4;
      return 1; /* Instruction emulated. */
   }
   return 0; /* Instruction not emulated. */
}

U32
zzepm_lock_save_dump(struct cpu_regdump *dump)
{
   struct cpu_core *core = (struct cpu_core *) dump->sect[0].address;
   U32 old_sr = core->sr;

   core->sr &= ~1; /* Disable interrupts for sr in dump. */
   return old_sr;
}

void
zzepm_lock_restore_dump(struct cpu_regdump *dump, U32 sr)
{
   struct cpu_core *core = (struct cpu_core *) dump->sect[0].address;

   core->sr = sr; /* Restore sr. */
}

/* Public functions. Prototype in epm.h */

EpmStatus
ose_epm_get_reg(struct cpu_regdump *dump, U32 id, U32 *value)
{
   U32 *core = (U32 *) dump->sect[0].address;
   U32 offset = zzget_regdump_core_offset(id);

   if (offset == (U32) ~0UL)
   {
      *value = 0xdeadbeef;
      return EPM_INVALID_ARGUMENTS;
   }
   *value = core[offset];
   return EPM_OK;
}

EpmStatus
ose_epm_set_reg(struct cpu_regdump *dump, U32 id, U32 value)
{
   U32 *core = (U32 *) dump->sect[0].address;
   U32 offset = zzget_regdump_core_offset(id);

   if (offset == (U32) ~0UL)
   {
      return EPM_INVALID_ARGUMENTS;
   }
   core[offset] = value;
   return EPM_OK;
}

void
zzepm_debug_save_regs(void)
{
}

void
zzepm_debug_restore_regs(void)
{
}

/*
 * zzepm_traspose_opcode_to_address()
 * take curr_opcode (that is from address) and create something that execute
 * the samething in opcode_area and then jumps back to the next instruction
 * type is an outvarable where you can put ARCH specific info.
 */
#define EP_VENEER_TYPE_INVALID 0
#define EP_VENEER_TYPE_CODE 1
#define EP_VENEER_TYPE_BRANCH 2
#define EP_VENEER_TYPE_BRANCH_LINK 3
#define EP_VENEER_TYPE_BRANCH_CONDITIONAL 4

static OSADDRESS
create_jump(OSADDRESS fromAddress, OSADDRESS toAddress)
{
   DLOG(("epm_arch.c(%d): Create jump from: 0x%08X to: 0x%08X\n",
                     __LINE__, fromAddress, toAddress));
   /* check if 28 bits addressing suffice */
   if ((fromAddress & ADDR_MASK) == (toAddress & ADDR_MASK))
   {
      /* encode jump as absolute jump */
      return OPCODE_J(toAddress);
   }
   else
   {
      DLOG(("epm_arch.c(%d): Failed comparison of from_nibbel: 0x%08X to_nibbel: 0x%08X\n",
                        __LINE__, (fromAddress & ADDR_MASK), (toAddress & ADDR_MASK)));
      return OPCODE_NOP;
   }
}

static OSADDRESS
get_branch_delay_opcode(OSADDRESS branch_address)
{
   DLOG(("epm_arch.c(%d): bds: 0x%08X @ 0x%08X\n",
            __LINE__, *((OSADDRESS*) (branch_address + 4)),
            branch_address + 4));
   return *((OSADDRESS*) (branch_address + 4));
}

/*
static uint64_t
get_address(uint32_t *buf)
{
   uint64_t ret = 0;

   ret |= (uint64_t)buf[0];
   ret |= ((uint64_t)buf[1])<<32;

   return ret;
}
*/

static void
put_address(uint32_t *buf, OSADDRESS address)
{
   phys_addr_t pa;
   OSADDRESS upper = 0xFFFFFFFF;

   /* TODO: Find a nicer solution to get the upper 32-bits for an address */
   if (ose_mm_l2p(zzkrn_current_process(), (log_addr_t)address, &pa) == MM_SUCCESS)
   {
      upper = (OSADDRESS) (pa >> 32);
   }

   buf[0] = address & 0xFFFFFFFF;
   buf[1] = upper & 0xFFFFFFFF;
}

EpmStatus
zzepm_transpose_opcode(OSADDRESS curr_opcode, OSADDRESS address, uint32_t *type,
         uint32_t *opcode_area, uint32_t opcode_area_size)
{
   OSADDRESS next_pc;
   OSADDRESS jump_opcode;
   U32 instr_type;

   DLOG(("epm_arch.c(%d): Transpose opcode 0x%08X @ 0x%08X\n",
            __LINE__, curr_opcode, address));

   instr_type = zzpreprocess_trace(curr_opcode, address, &next_pc);
   if (INST_ILLEGAL == instr_type)
   {
      return EPM_INTERNAL_ERROR;
   }

   if (opcode_area_size < 8 * 4 || type == NULL)
   {
      return EPM_INTERNAL_ERROR;
   }

   opcode_area[0] = 0;
   opcode_area[1] = 0;
   opcode_area[2] = 0;
   opcode_area[3] = 0;
   opcode_area[4] = 0;
   opcode_area[5] = 0;
   opcode_area[6] = 0;
   opcode_area[7] = 0;

   if (BR(instr_type))
   {
      if ((instr_type == INST_JAL) || (instr_type == INST_JALR))
      {
         *type = EP_VENEER_TYPE_BRANCH_LINK;
      }
      else
      {
         *type = EP_VENEER_TYPE_BRANCH;
      }

      if (INST_JR == instr_type || INST_JALR == instr_type)
      {
         /* We can use the original opcode since address in register
          * is fully encoded
          */
         opcode_area[0] = curr_opcode & JALR_TO_JR_MASK;

         /* The branch delay slot  should be executed regardless (epc+4)*/
         opcode_area[1] = get_branch_delay_opcode(address);

         /*
          * Save link register and link address in order to be able
          * to emulate it in zzepm_handle_transposed_opcode
          */
         if (instr_type == INST_JALR)
         {
            put_address(&opcode_area[LINK_ADDRESS_OFFSET], address + 8);
            opcode_area[LINK_REGISTER_OFFSET] = (curr_opcode & ~JALR_TO_JR_MASK) >> 11;
         }
      }
      else
      {
         /* Jump instruction goes here.
          * Will be NOP if address is out of range and
          * instead the "trap-thing" will be done
          */
         opcode_area[0] = create_jump((OSADDRESS) opcode_area, next_pc);

         /* The branch delay slot  should be executed regardless (epc+4)*/
         opcode_area[1] = get_branch_delay_opcode(address);

         /* code for long jump goes here */
         if(OPCODE_NOP == opcode_area[0])
         {
            opcode_area[2] = OPCODE_RESERVED_INSTRUCTION;
            put_address(&opcode_area[3], next_pc);
         }

         /*
          * Save link register and link address in order to be able
          * to emulate it in zzepm_handle_transposed_opcode
          */
         if (instr_type == INST_JAL)
         {
            put_address(&opcode_area[LINK_ADDRESS_OFFSET], address + 8);
            opcode_area[LINK_REGISTER_OFFSET] = 31;
         }
      }
   }
   else if (BC(instr_type))
   {
      uint32_t branch_not_taken_opcode;
      uint32_t branch_taken_opcode;

      *type = EP_VENEER_TYPE_BRANCH_CONDITIONAL;

      branch_not_taken_opcode = create_jump(
               (OSADDRESS) &opcode_area[2], address + 8);
      branch_taken_opcode = create_jump(
               (OSADDRESS) &opcode_area[4], next_pc);
      /* check for longjump */
      if ((OPCODE_NOP != branch_not_taken_opcode) &&
          (OPCODE_NOP != branch_taken_opcode))
      {
         /* jump within reach */
         curr_opcode = (curr_opcode & 0xffff0000) | (4 - 1);
         opcode_area[0] = curr_opcode;
         /* The branch delay slot */
         opcode_area[1] = get_branch_delay_opcode(address);
         /* branch not taken */
         opcode_area[2] = branch_not_taken_opcode;
         opcode_area[3] = OPCODE_NOP;
         /* branch taken */
         opcode_area[4] = branch_taken_opcode;
         opcode_area[5] = OPCODE_NOP;
      }
      else
      {
         curr_opcode = (curr_opcode & 0xffff0000) | (5 - 1);
         opcode_area[0] = curr_opcode;
         /* The branch delay slot */
         opcode_area[1] = get_branch_delay_opcode(address);
         /* branch not taken */
         opcode_area[2] = OPCODE_RESERVED_INSTRUCTION;
         put_address(&opcode_area[3], address + 8); //skip branch delay slot
         /* branch taken */
         opcode_area[5] = OPCODE_RESERVED_INSTRUCTION;
         put_address(&opcode_area[6], next_pc);
      }
   }
   else
   {
      *type = EP_VENEER_TYPE_CODE;
      /* Add getting back code */
      jump_opcode = create_jump((OSADDRESS) &opcode_area[1], address + 4);
      if (OPCODE_NOP != jump_opcode)
      {
         /* Put instruction in branch delay slot */
         opcode_area[0] = curr_opcode;

         /* encode jump as absolute jump */
         opcode_area[1] = jump_opcode;

         /* NOP in branch delay slot */
         /* TODO: Optimize and put valid instructions in branch delay slot */
         opcode_area[2] = OPCODE_NOP;
      }
      else
      {
         opcode_area[0] = curr_opcode;

         /* long jump: destination is not within reach, use a trap */
         opcode_area[1] = OPCODE_RESERVED_INSTRUCTION;

         put_address(&opcode_area[2], address + 4);
      }
   }

   /* Ensure it reach instruction/data memory for all core's */
   ose_cache(CACHE_FLUSH_D, CACHE_PGM_LOAD, (OSADDRESS) opcode_area,
            opcode_area_size);
   ose_cache(CACHE_INVALIDATE_I, CACHE_PGM_LOAD, (OSADDRESS) 0,
            0);

   return EPM_OK;
}

EpmStatus
zzepm_handle_transposed_opcode(struct cpu_regdump *dump, uint32_t type,
         uint32_t *opcode_area)
{
   struct cpu_core *core = (struct cpu_core *) dump->sect[0].address;
   EpmStatus sts = EPM_INVALID_ARGUMENTS;

   switch (type)
   {
      case EP_VENEER_TYPE_BRANCH_LINK:
         /* simulate a jal/jalr by putting return address in register LR */
         core->r[opcode_area[LINK_REGISTER_OFFSET]] =
                  (core->epc & EPC_MASK) | opcode_area[LINK_ADDRESS_OFFSET];
      case EP_VENEER_TYPE_BRANCH:
      case EP_VENEER_TYPE_BRANCH_CONDITIONAL:
      case EP_VENEER_TYPE_CODE:
         /* jump to veneer code */
         core->epc = (uint64_t) opcode_area;
         sts = EPM_OK;
         break;
      default:
         ERRLOG(("epm_arch.c(%d): Faulty instruction 0x%08X \n", __LINE__, core->epc));
         break;
   }

   dump->pc = core->epc;

   DLOG(("epm_arch.c(%d): Setting pc to 0x%08X: 0x%08X \n",
            __LINE__, dump->pc, *(uint32_t*) dump->pc));

   return sts;
}

void
zzemp_handle_veneer_trap(struct cpu_regdump *dump, OSADDRESS exception_pc)
{
   struct cpu_core *cpucore = (struct cpu_core *) dump->sect[0].address;

   cpucore->epc = (uint64_t) (*(uint32_t *)(((uint8_t *)exception_pc)+8))<<32 |
            (uint64_t) (*(uint32_t *)(((uint8_t *)exception_pc)+4));
   dump->pc = cpucore->epc;

   DLOG(("epm.c(%d): trap in veneer, next pc is 0x%08X: "
            "0x%08X fetched from 0x%08X \n",
            __LINE__, dump->pc, *(uint32_t*)dump->pc,
            (uint32_t *)(((uint8_t *)exception_pc)+4)));
}


/* HW breakpoints not suppported. */
const struct zzepm_bp_conf *zzepm_hwbp_conf = NULL;
