/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2003-2007 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

#include "assert.h"
#include "ose.h"
#include "hal.h"
#include "page.h"
#include "zone.h"
#include "common.h"
#include "mm_err.h"
#include "mmtypes.h"
#include "mm_log.h"
#include "exception.h"
#include "register.h"
#include "cpu_regdump.h"
#include "cpu_access.h"
#include "hal_mmu.h"
#include "ram_allocator.h"
#include "string.h"
#include "gmm.h"
#include "cpu_hal_arch.h"
#include "cpu_smp.h"
#include "spinlock.h"
#include "ose_smp.h"
#include "atomic.h"
#include "mm_private.h"
#include "pose.h" /* To get prototype for zzkrn_evt_enqueue. */

/* Sysconf constants (also defined in tidata.h) . */

/* Install a new function in get_sysconf. */
#define SYSCONF_NEW_FUNCTION           2

/* Set error check by get_sysconf. */
#define SYSCONF_SET_ERROR_CHECK        3

/* get_sysconf codes installed by ti. */
#define SYSCONF_MMU_HAL_GET_ASID      36

#define PRINT_PREFIX "MMU-HAL: "
#define VERBOZE(s) log(CATEGORY_HAL, s)
#define VERBOZE_DETAIL(s) log(CATEGORY_HAL_DETAIL, s)

static mm_bool_t d_cache_enabled;
static mm_bool_t i_cache_enabled;

static log_size_t page_sizes[] = {HAL_SIZE_4K * 2};

typedef struct L2Pte {
   uint64_t entryLo;
} L2Pte_t;

typedef struct L1Pte {
   L2Pte_t *l2Ptp;
   U32 refCount;
}  L1Pte_t;

/* 512 * 64bit = 4K rounded to closest page size 8K*/
#define L1_PAGE_TABLE_SIZE (HAL_SIZE_4K * 2) 

/* 1024 * 64bit = 8K */
#define L2_PAGE_TABLE_SIZE (HAL_SIZE_4K * 2)

struct hal_domain {
   L1Pte_t *l1Ptp;
   U32 asid;
};

L1Pte_t hal_global_page_table[(L1_PAGE_TABLE_INDEX_MASK + 1) * sizeof(L1Pte_t)];

volatile L1Pte_t *hal_local_page_table[OSE_SMP_MAX_CPUS];

static struct zone hal_domain_zone;
#define HAL_DOMAIN_ZONE_ITEM_SIZE (sizeof(struct hal_domain))

#ifdef ASSERT
static mm_bool_t systemDomainInitialized = FALSE;
#endif

static mm_bool_t getAsidForSegnumInstalled = FALSE;

static struct hal_domain systemDomain;

/* Spinlock for SMP exclusion */
static spinlock_t hal_lock;

/* prototypes */
mm_bool_t
mm_mas_mode_active(void);


/*
 * This struct contains information about this HAL implementation.
 * It is used by GMM.
 */
static struct hal_info hal_info =
{
   TRUE,  /* has_mmu */
   TRUE,  /* enable_mmu_early */
   FALSE, /* MMU disabled initially. */
   FALSE, /* MAS not yet supported. */
   sizeof(page_sizes)/sizeof(page_sizes[0]),     /* Number of page sizes. */
   page_sizes,
   0,     /* Number of resident page_sizes. */
   NULL,  /* Pointer to array with resident page sizes.*/
   0,     /* The number of supported tag id's. */
   NULL   /* Pointer to array with supported MmTagIds.*/
};

/* Called by all cores.*/
__attribute__((aligned(4096)))
static void
mmu_hal_invalidate_tlb_local(U32 base, U32 size)
{
   /* Use a non-TLB address, extract current_ASID. */
   U32 va = KSEG0_BASE | (get32_cp0(C0_ENTRYHI) & 0xff);
   U32 mmu_size;
   U32 i;

   (void)base;
   (void)size;
   
   if (get32_cp0(C0_CONFIG) & 0x80000000)
   {
      mmu_size = get32_cp0_sel1(C0_CONFIG);
      mmu_size = (mmu_size >> 25) & 0x3f;
   }
   else
   {
      /* R4000/R4700 have 48 TLB entries. */
      mmu_size = 47;
   }

   VERBOZE((PRINT_PREFIX "mmu_hal_invalidate_tlb_local, size: %d, cpu %d\n",
	    mmu_size, zzkrn_cpu_id()));


   for(i = 0; i <= mmu_size; i++)
   {
      Msr msr;

      LOCK_SAVE(msr);
      set32_cp0(C0_PAGEMASK, 0);
      set64_cp0(C0_ENTRYLO_0, 0);
      set64_cp0(C0_ENTRYLO_1, 0);
      set32_cp0(C0_INDEX, i);
      set32_cp0(C0_ENTRYHI, va);

      __asm__ __volatile__(
         ".set\tmips3\n\t"
         ".set\tnoreorder\n\t"
         "nop;nop;nop;nop\n\t"
         "tlbwi\n\t"
         "nop;nop;nop;nop\n\t"
         ".set\treorder\n\t"
         ".set\tmips0\n\t");
      LOCK_RESTORE(msr);
      
      va += (page_sizes[0]); /* Pagesize */
   }
}

#ifdef OSE_SMP
/* Called by core0 only.*/
static void
mmu_hal_invalidate_tlb_all(void)
{
   cpuid_t num_cpus = ose_num_cpus();
   cpuid_t cpu;

   /* Send event to other cores. This event will make all other cores
      call the function mmu_hal_invalidate_tlb_local. */
   for(cpu = 1; cpu < num_cpus; cpu++)
   {
      /* Invalidate TLB on 'cpu'. */
      get_sysconf(104, cpu, 0, ~0);
   }

   /* Invalidate local tlb, always core0. */
   /* FIXME: Only invalidate changed page */
   mmu_hal_invalidate_tlb_local(0, ~0); 
}
#endif

/* This function is called by TI in order to get the ASID of the domain
 * to which a process belongs (via a block and a segment).
 * The 'seg_id' is really an 'OSSEGMENT segnum' as known by the kernel. */
static U32
getAsidForSegnum(unsigned int segnum, U32 *asid)
{
   /* FIXME: Get the ASID from the ASID vector
    * containing the segnum:s of the created domains. */
   *asid = segnum;

   return 1;
} /* get_asid(...) */


static hal_status_t
l1PageTableAlloc(L1Pte_t **l1Ptp)
{
   if (! mm_page_alloc(L1_PAGE_TABLE_SIZE,
                       (U32 *)l1Ptp))
   {
      return ~HAL_SUCCESS;
   }

   assert((OSADDRESS)*l1Ptp < 0x20000000UL);

   /* Translate the L1 PTP for access from KSEG0 */
   *l1Ptp = (L1Pte_t *)(KSEG0(*l1Ptp));

   assert(mm_log_inside((OSADDRESS)*l1Ptp, KSEG0_BASE, KSEG0_SIZE));

   bzero(*l1Ptp, L1_PAGE_TABLE_SIZE);

   return HAL_SUCCESS;
} /* l1PageTableAlloc(...) */

static void
l1PageTableFree(L1Pte_t *l1Ptp)
{
   assert(NULL != l1Ptp);
   assert(mm_log_inside((OSADDRESS)l1Ptp, KSEG0_BASE, KSEG0_SIZE));

#ifdef ASSERT
   {
      /* Make sure all references to L2 page-tables are cleared before
       * freeing the L1 page-table */
      U32 l1PteIx;
      for (l1PteIx = 0;
           l1PteIx < (L1_PAGE_TABLE_SIZE / sizeof(L1Pte_t));
           l1PteIx++)
      {
         assert(NULL == l1Ptp[l1PteIx].l2Ptp);
         assert(0 == l1Ptp[l1PteIx].refCount);
      }
   }
#endif

   /* Translate the L1 PTP for access from KUSEG */
   l1Ptp = (L1Pte_t *)(((U32)l1Ptp) & ~KSEG_MASK);

   mm_page_free((U32)l1Ptp, L1_PAGE_TABLE_SIZE);
} /* l1PageTableFree(...) */

static L1Pte_t *
getL1Ptp(struct hal_domain *hal_domain)
{
   assert(NULL != hal_domain);
   assert(NULL != hal_domain->l1Ptp);
   return hal_domain->l1Ptp;
} /* getL1Ptp(...) */

static L1Pte_t *
getL1Pte(L1Pte_t *l1Ptp, log_addr_t la)
{
   U32 l1PteIx = L1_PAGE_TABLE_INDEX(la);
   assert(NULL != l1Ptp);
   VERBOZE_DETAIL((PRINT_PREFIX "l1Ptp-0x%x la-0x%x l1PteIx-%x\n",
                   l1Ptp, la, l1PteIx));
   return &(l1Ptp[l1PteIx]);
} /* getL1Pte(...) */

static void
l1PteUpdate(L1Pte_t *l1Pte, L2Pte_t *l2Ptp)
{
   assert(NULL != l1Pte);
   l1Pte->l2Ptp = l2Ptp;
} /* l1PteUpdate(...) */

static L2Pte_t *
getL2Ptp(L1Pte_t *l1Pte)
{
   assert(NULL != l1Pte);
   return l1Pte->l2Ptp;
} /* getL2Ptp(...) */

static L2Pte_t *
getL2Pte(L2Pte_t *l2Ptp, log_addr_t la)
{
   U32 l2PteIx = L2_PAGE_TABLE_INDEX(la);
   L2Pte_t *l2Pte;
   assert(NULL != l2Ptp);
   l2Pte  = &(l2Ptp[l2PteIx]);
   assert(NULL != l2Pte);

   VERBOZE_DETAIL((PRINT_PREFIX "l2Ptp-0x%x la-0x%x l2PteIx-%x\n",
                   l2Ptp, la, l2PteIx));

   return l2Pte;
} /* getL2Pte(...) */

static U32
getL1PteCounter(L1Pte_t *l1Pte)
{
   assert(NULL != l1Pte);
   return (l1Pte->refCount);
}         

static U32
l1PteCounterIncrease(L1Pte_t *l1Pte)
{
   assert(NULL != l1Pte);
   return ++(l1Pte->refCount);
} /* l1PteCounterIncrease(...) */

static U32
l1PteCounterDecrease(L1Pte_t *l1Pte)
{
   assert(NULL != l1Pte);
   assert(0 < l1Pte->refCount);
   return --(l1Pte->refCount);
} /* l1PteCounterDecrease(...) */

static void
l1PteInvalidate(L1Pte_t *l1Pte)
{
   assert(NULL != l1Pte);
   assert(NULL != l1Pte->l2Ptp);
   l1Pte->l2Ptp = 0;
   l1Pte->refCount = 0;
} /* l1PteInvalidate(...) */

static void
l2PteUpdate(L2Pte_t *l2Pte, log_addr_t la, phys_addr_t pa,
            U32 acc_attr, U32 mem_attr)
{
   (void)la;
   
   assert(NULL != l2Pte);
   (void)la;

   VERBOZE_DETAIL((PRINT_PREFIX "acc_attr: 0x%x\n", acc_attr));
   
      l2Pte->entryLo = ((pa & MASK_PFN) >> 6) |
//      (acc_attr & (MM_SU_WRITE | MM_USR_WRITE) ? ENTRY_LO_DIRTY : 0) |
      (acc_attr & MM_USR_READ ? ENTRY_LO_USR_READ : 0) |
      (acc_attr & MM_USR_WRITE ? ENTRY_LO_USR_WRITE : 0) |
      (acc_attr & MM_SU_WRITE ? ENTRY_LO_DIRTY : 0) |
      (acc_attr & (MM_SU_READ | MM_SU_WRITE | MM_SU_EXECUTE |
                   MM_USR_READ | MM_USR_WRITE | MM_USR_EXECUTE) ?
       ENTRY_LO_VALID : 0) |
      (mem_attr & HAL_MIPS_CACHE_ENABLE ?
       hal_arch_cpu_descriptor->entry_lo.cache_attr_cachable :
       hal_arch_cpu_descriptor->entry_lo.cache_attr_uncached);

} /* l2PteUpdate(...) */

static void
l2PteInfo(L2Pte_t *l2Pte, log_addr_t la, phys_addr_t *page_base,
          log_size_t *page_size, U32 *perm)
{
   uint64_t pte = l2Pte->entryLo;;

   if (page_base != NULL)
   {
      *page_base = (pte << 6) & MASK_PFN;
   }
   
   if (page_size != NULL)
   {
      *page_size = page_sizes[0];      
   }
   
   if (perm != NULL)
   {
      if ((mm_log_inside(la, KSEG0_BASE, KSEG0_SIZE)) ||
          (mm_log_inside(la, KSEG1_BASE, KSEG1_SIZE)))
      {
         *perm = MM_SU_READ|MM_SU_EXECUTE|MM_SU_WRITE;
         return;
      }

      *perm = MM_SU_READ|MM_SU_EXECUTE;
      *perm |= pte & ENTRY_LO_USR_READ ? (MM_USR_READ|MM_USR_EXECUTE) : 0;
      *perm |= pte & ENTRY_LO_USR_WRITE ? MM_USR_WRITE : 0;
      *perm |= pte & ENTRY_LO_DIRTY ? MM_SU_WRITE : 0;
   }
}

static void
l2PteInvalidate(L2Pte_t *l2Pte)
{
   assert(NULL != l2Pte);
   l2Pte->entryLo = 0;
} /* l2PteInvalidate(...) */

static void
l2PteDump(L2Pte_t *l2Pte)
{
   assert(NULL != l2Pte);
   (void)l2Pte;
   VERBOZE_DETAIL((PRINT_PREFIX "l2Pte.entryLo=0x%llx\n",  l2Pte->entryLo));
} /* l2PteDump(...) */

static hal_status_t
l2PageTableAlloc(L2Pte_t **l2Ptp)
{
   if (! mm_page_alloc(L2_PAGE_TABLE_SIZE,
                       (U32 *)l2Ptp))
   {
      return HAL_EEND_OF_PHYS_MEMORY;
   }
#ifndef HOST_TEST
   assert((OSADDRESS)*l2Ptp < 0x20000000UL);

   /* Translate the L2 PTP for access from KSEG0 */
   *l2Ptp = (L2Pte_t *)(KSEG0(*l2Ptp));
#endif
   bzero(*l2Ptp, L2_PAGE_TABLE_SIZE);

   return HAL_SUCCESS;
} /* l2PageTableAlloc(...) */

static void
l2PageTableFree(L2Pte_t *l2Ptp)
{
   assert(mm_log_inside((OSADDRESS)l2Ptp, KSEG0_BASE, KSEG0_SIZE));

   /* Translate the L2 PTP for access from KUSEG */
   l2Ptp = (L2Pte_t *)(((U32)l2Ptp) & ~KSEG_MASK);

   mm_page_free((U32)l2Ptp, L2_PAGE_TABLE_SIZE);
} /* l2PageTableFree(...) */

/*
 * See hal.h.
 */
static hal_status_t
mmu_hal_init_1(const struct cpu_descriptor *cpu_descriptor,
               log_addr_t vector_base,
               log_size_t vector_size,
               mm_bool_t enable_d_cache,
               mm_bool_t enable_i_cache)
{
#ifdef LIBWHAT
   LIBWHAT
#endif
   VERBOZE((PRINT_PREFIX "hal_init_1(cpu_descriptor=%p, vector_base=%#lx, "
            "vector_size=%x, enable_d_cache=%d, enable_i_cache=%d)\n",
            cpu_descriptor, vector_base, vector_size,
            enable_d_cache, enable_i_cache));

   d_cache_enabled = enable_d_cache;
   i_cache_enabled = enable_i_cache;

   spin_lock_init(&hal_lock);

   hal_arch_cpu_descriptor = cpu_descriptor->arch_cpu_descriptor;

   /* If supported by cpu hal, call init_cpu function. */
   if (hal_arch_cpu_descriptor->init_cpu)
   {
      hal_arch_cpu_descriptor->init_cpu();
   }

#ifndef _MIPS_ARCH_MIPS3
   /* Enable support for large physical addresses. (bit 29)*/
   set32_cp0_sel1(CO_PAGEGRAIN, 0x20000000);
#endif

   /* Select 4k physical page size. */
   set32_cp0(C0_PAGEMASK, 0);

   mmu_hal_invalidate_tlb_local(0, ~0);

   set32_cp0(C0_ENTRYHI, ASID_SYSTEM);

   return HAL_SUCCESS;
}

/*
 * See hal.h.
 */
static hal_status_t
mmu_hal_init_2(void)
{
   VERBOZE((PRINT_PREFIX "hal_init_2()\n"));

   return HAL_SUCCESS;
}

/*
 * See hal.h.
 */
static hal_status_t
mmu_hal_init_3(struct partition *partition)
{
   VERBOZE((PRINT_PREFIX "hal_init_3(partition=%p)\n", partition));
   (void)partition;

   mm_zinit(&hal_domain_zone, "hal_domain_zone",
            HAL_DOMAIN_ZONE_ITEM_SIZE, 5, 10);

   return HAL_SUCCESS;
}

/*
 * See hal.h.
 */
static void
mmu_hal_init_4(void)
{
   VERBOZE((PRINT_PREFIX "hal_init_4(void)\n"));
}

/*
 * See hal.h.
 */
static void
mmu_hal_init_5(void)
{
   exception_handler_asm hal_tlb_refill_handler;

   VERBOZE((PRINT_PREFIX "hal_init_5(void)\n"));

   ose_install_asm_exception_handler(EXCEPTION_VEC_TLB_REFILL,
                                     hal_tlb_refill_handler,
                                     0x80);
   ose_install_asm_exception_handler(EXCEPTION_VEC_XTLB,
                                     hal_tlb_refill_handler,
                                     0x80);
   hal_info.enabled_mmu = TRUE;
}


/*
 * See hal.h.
 */
static const struct hal_info *
mmu_hal_get_info(void)
{
   VERBOZE((PRINT_PREFIX "hal_get_info()\n"));
   return &hal_info;
}


/*
 * See hal.h.
 */
static hal_status_t
mmu_hal_actual_permissions(U32 *actual_permissions,
                           U32 requested_permissions,
                           log_addr_t base, U32 options)
{
   struct partition_class *p;

   (void)options;
   
   *actual_permissions = requested_permissions;

   VERBOZE((PRINT_PREFIX "mmu_hal_actual_permissions(gmm_perm=0x%x)\n",
            requested_permissions));

   if (requested_permissions & (MM_SU_READ | MM_SU_EXECUTE))
   {
      *actual_permissions |= (MM_SU_READ | MM_SU_EXECUTE);
   }

   if (requested_permissions & (MM_USR_READ | MM_USR_EXECUTE))
   {
      *actual_permissions |= (MM_USR_READ | MM_USR_EXECUTE);
   }


   if (!mm_log_inside(base, KUSEG_BASE, KUSEG_SIZE))
   {
      if ((mm_log_inside(base, KSEG0_BASE, KSEG0_SIZE)) ||
          (mm_log_inside(base, KSEG1_BASE, KSEG1_SIZE)))
      {
         /* We cannot protect regions located in KSEG0/1
          * from any supervisor accesses - hence round up to RWX
          * for supervisor-mode */
         *actual_permissions |= MM_SU_READ | MM_SU_WRITE | MM_SU_EXECUTE;
      }

      if (*actual_permissions & (MM_USR_READ | MM_USR_WRITE | MM_USR_EXECUTE))
      {
         /* Setting user-mode permissions is only valid
          * for regions located inside KUSEG. */
         *actual_permissions &= ~(MM_USR_READ | MM_USR_WRITE | MM_USR_EXECUTE);
         return HAL_EILLEGAL_PERMISSION;
      }
   }

   p = mm_space_is_resident(base, 0);
   if(p != NULL)
   {
      U32 perm_su, perm_usr, r_perm_su, r_perm_usr;

      perm_su  = requested_permissions &
         (MM_SU_EXECUTE  | MM_SU_WRITE  | MM_SU_READ);
      perm_usr = requested_permissions &
         (MM_USR_EXECUTE | MM_USR_WRITE | MM_USR_READ);

      r_perm_su  = p->res_perm &
         (MM_SU_EXECUTE  | MM_SU_WRITE  | MM_SU_READ);
      r_perm_usr = p->res_perm &
         (MM_USR_EXECUTE | MM_USR_WRITE | MM_USR_READ);
  
      /* Check if requested permissions can be satisfied in the log_mem. */
      if (((perm_su & ~r_perm_su) != 0) || ((perm_usr & ~r_perm_usr) != 0))
      {
         /* Requested permissions allow more types of accesses than
          * log_mem permissions, which makes them impossible to fulfill.
          * This condition may have changed after rounding. 
          * 
          * Example of "down" rounded permissions on PPC PEM:
          *   Requested SU_RW_USR_RO -> Actual SU_RW_USR_NA
          */
         perm_su  = *actual_permissions &
            (MM_SU_EXECUTE  | MM_SU_WRITE  | MM_SU_READ);
         perm_usr = *actual_permissions &
            (MM_USR_EXECUTE | MM_USR_WRITE | MM_USR_READ);

         /* Check if rounded permissions can be satisfied in the log_mem. */
         if (((perm_su & ~r_perm_su) != 0) || ((perm_usr & ~r_perm_usr) != 0))
            return HAL_EILLEGAL_PERMISSION;
      }

      *actual_permissions = p->res_perm;
   }

   return HAL_SUCCESS;
} /* mmu_hal_actual_permissions(...) */


/*
 * See hal.h.
 */
static hal_status_t
mmu_hal_create_domain(unsigned int seg_id,
                      mm_bool_t system_domain,
                      struct hal_domain **domain)
{
   Msr sr = 0;
   
   VERBOZE((PRINT_PREFIX "hal_create_domain(seg_id=%x, "\
            "system_domain=%d, hal_domain=%p)\n",
            seg_id, system_domain, domain));

   assert(NULL != domain);

   if (system_domain)
   {
      assert(!systemDomainInitialized);
      *domain = &systemDomain;
      (**domain).l1Ptp = &(hal_global_page_table[0]);

#ifdef ASSERT
      systemDomainInitialized = TRUE;
#endif

   }
   else
   {
      if (TRUE != getAsidForSegnumInstalled)
      {
         /* Install 'getAsidForSegnum(...)'
          * for acces from 'get_sysconf(SYSCONF_MMU_HAL_GET_ASID, ...)'. */
         if (get_sysconf(SYSCONF_NEW_FUNCTION,
                         SYSCONF_MMU_HAL_GET_ASID,
                         (OSADDRESS)getAsidForSegnum, 0))
         {
            /* Overwritten old function installed on
             * 'get_sysconf(SYSCONF_MMU_HAL_GET_ASID)' */
            return HAL_EINTERNAL_ERROR;
         }

         getAsidForSegnumInstalled = TRUE;
      }

      MM_MULTITHREAD_LOCK(sr);

      *domain = mm_zalloc(&hal_domain_zone);
      if (NULL == *domain)
      {
         MM_MULTITHREAD_UNLOCK(sr);
         return HAL_EEND_OF_PHYS_MEMORY;
      }

      if (HAL_SUCCESS != l1PageTableAlloc(&((**domain).l1Ptp)))
      {
         MM_MULTITHREAD_UNLOCK(sr);
         return HAL_EEND_OF_PHYS_MEMORY;
      }

      MM_MULTITHREAD_UNLOCK(sr);
   }

   /* FIXME: Allocate ASID from vector of 255 elements,
      each one containing the seg_id given at creation of the domain */
   (**domain).asid = seg_id;

   return HAL_SUCCESS;
} /* mmu_hal_create_domain(...) */


/*
 * See hal.h.
 */
static void
mmu_hal_delete_domain(unsigned int seg_id,
                      struct hal_domain *domain)
{
   Msr sr = 0;

   VERBOZE((PRINT_PREFIX "hal_delete_domain(seg_id=%x, domain=%p)\n",
            seg_id, domain));
   assert(NULL != domain);
   (void)seg_id;
   (void)domain;

   MM_MULTITHREAD_LOCK(sr);
   l1PageTableFree(domain->l1Ptp);
   /* FIXME: Free ASID from vector of 255 elements,
    * by clearing the entry containing the seg_id. */
   domain->asid = 0;

   mm_zfree(&hal_domain_zone, domain);
   MM_MULTITHREAD_LOCK(sr);
}


/*
 * See hal.h.
 */
static void
mmu_hal_select_domain(struct hal_domain *hal_domain)
{
   VERBOZE((PRINT_PREFIX "hal_select_domain(hal_domain=%p, "\
            "hal_domain->asid=%d)\n",
            hal_domain, hal_domain->asid));
   hal_local_page_table[zzkrn_cpu_id()] = hal_domain->l1Ptp;

   /* Note that unlike powerpc the ASID swap is done in the
    * target interface. */

   if (mm_mas_mode_active())
   {
      /* Invalidation of TLB needed since;
       * Mips does not support supervisor/user mode bits in the TLB entries.
       * The target interface solves this problem by setting MMU ASID = 0
       * to all supervisor mode processes and when a user mode
       * process traps into supervisor mode for example in a system call.
       * I.e. in supervisor mode two processes in differnet domains uses the
       * same ASID (=0) and an invalidation must be issued to prevent them from
       * seeing the wrong MAS data.
       */
      mmu_hal_invalidate_tlb_local(0, ~0); /* FIXME */
   }
}


static hal_status_t
mapPage(struct hal_domain *domain,
        log_addr_t la,
        phys_addr_t pa,
        log_size_t size,
        int options,
        U32 access_attr,
        U32 mem_attr)
{
   L1Pte_t *l1Ptp = getL1Ptp(domain);
   L1Pte_t *l1Pte = getL1Pte(l1Ptp, la);
   L2Pte_t *l2Ptp = getL2Ptp(l1Pte);
   mm_bool_t change = options & HAL_CHANGE;

   (void)size;
   assert(size == page_sizes[0]);
   
   VERBOZE_DETAIL((PRINT_PREFIX "l1Ptp 0x%x\n", l1Ptp));
   VERBOZE_DETAIL((PRINT_PREFIX "l1Pte 0x%x\n", l1Pte));

   if (NULL == l2Ptp)
   {
      /* The L1 page-table for the requested logical address
       * has not yet been referenced, hence does not contain a pointer
       * to a L2 page-table. */
      /* Allocate an L2 page-table and reference
       * it from the L1 page-table entry. */
      if (HAL_SUCCESS == l2PageTableAlloc(&l2Ptp))
      {
         /* A recursive allocation and subsequent map_page might
          * already have created this l2 for us. */
         if (getL2Ptp(l1Pte) != 0)
         {
            l2PageTableFree(l2Ptp);
            l2Ptp = getL2Ptp(l1Pte);
         }
         else
         {
            l1PteUpdate(l1Pte, l2Ptp);
         }         
      }
      else
      {
         return HAL_EEND_OF_PHYS_MEMORY;
      }
   }

   VERBOZE_DETAIL((PRINT_PREFIX "l2Ptp 0x%x\n", l2Ptp));
#ifndef HOST_TEST
   assert(mm_log_inside((OSADDRESS)l2Ptp, KSEG0_BASE, KSEG0_SIZE));
#endif

   if (!change)
   {
      L2Pte_t *l2Pte = getL2Pte(l2Ptp, la);

      /* We are not allowed to overwrite an existing mapping if the user did not
       * explicitly state that. Verify that the page is available...
       */
      if (l2Pte->entryLo != 0) {
         return HAL_EMAPPING_EXISTS;
      }
   }

   {
      L2Pte_t *l2Pte = getL2Pte(l2Ptp, la);
      
      /* map all except SU_NA_USR_NA */
      if(access_attr != SU_NA_USR_NA)
      {
         if (l2Pte->entryLo == 0)
         {
            l1PteCounterIncrease(l1Pte);
         }
         l2PteUpdate(l2Pte, la, pa, access_attr, mem_attr);         
      }
      else /* SU_NA_USR_NA */
      {
         if (l2Pte->entryLo != 0)
         {
            l1PteCounterDecrease(l1Pte);
            l2Pte->entryLo = 0;         
         }
      }
      l2PteDump(l2Pte);
   }

   return HAL_SUCCESS;
} /* mapPage(...) */

static void
unmapRange(struct hal_domain *domain,
           log_addr_t la,
           log_size_t size)
{
   L1Pte_t *l1Ptp = getL1Ptp(domain);
   log_addr_t tmpLa = la;

   for (tmpLa = la;
        tmpLa < la + size;
        tmpLa += page_sizes[0])
   {
      L1Pte_t *l1Pte = getL1Pte(l1Ptp, tmpLa);
      L2Pte_t *l2Ptp;
      L2Pte_t *l2Pte;
      U32 refCounter = getL1PteCounter(l1Pte);
      
      l2Ptp = getL2Ptp(l1Pte);

      if (l2Ptp == NULL)
      {
         continue;
      }
      
      l2Pte = getL2Pte(l2Ptp, tmpLa);
      assert(0 != l2Pte);
      
      if (0 != l2Pte->entryLo)
      {
         refCounter = l1PteCounterDecrease(l1Pte);
      }
      

      if (0 == refCounter)
      {
         l1PteInvalidate(l1Pte);
         l2PageTableFree(l2Ptp);
      }
      else
      {
         l2PteInvalidate(l2Pte);
      }
   }
} /* unmapRange(...) */

/*
 * See hal.h.
 */
static hal_status_t
mmu_hal_map_page(struct hal_domain *domain,
                 log_addr_t la,
                 phys_addr_t pa,
                 log_size_t size,
                 int options,
                 U32 access_attr,
                 U32 mem_attr)
{
   hal_status_t rc = HAL_SUCCESS;
   Msr sr = 0;

   VERBOZE((PRINT_PREFIX "hal_map_page(domain=%p, la=%#lx, pa=%#lx, size=%x, "
            "options=%x, access_attr=%x, mem_attr=%x)\n",
            domain, la, (U32)pa, size, options, access_attr, mem_attr));

   assert(size == page_sizes[0]);

   if (mm_log_inside(la, KUSEG_BASE, KUSEG_SIZE) ||
       mm_log_inside(la, KSSEG_BASE, KSSEG_SIZE) ||
       mm_log_inside(la, KSEG3_BASE, KSEG3_SIZE))
   {
      mm_bool_t applicationDomain = &systemDomain != domain;
      mm_bool_t change = options & HAL_CHANGE;
      mm_bool_t global = !(options & HAL_SHARE);
      mm_bool_t mas = options & HAL_MAS;


      rc = hal_get_cmbits(mem_attr, &mem_attr);
      if (rc != HAL_SUCCESS)
         return HAL_EILLEGAL_CACHE_MODE;
      
      
      MM_MULTITHREAD_LOCK(sr);

      if (applicationDomain)
      {
         /* Map the page into the local application-domain page-table tree */
         rc = mapPage(domain, la, pa, size, change, access_attr, mem_attr);
      }

      /* Map the page into the global system-domain page-table tree */
      if ((HAL_SUCCESS == rc) && global && !mas)
      {
         U32 global_access_attr;

         /* Clear user-mode access if global region. */
         global_access_attr =
            applicationDomain ?
            access_attr & (MM_SU_READ | MM_SU_WRITE | MM_SU_EXECUTE) :
            access_attr;

         if (applicationDomain)
         {
            VERBOZE((PRINT_PREFIX "global_access_attr=%x\n",
                     global_access_attr));
         }

         VERBOZE((PRINT_PREFIX "map in system domain\n"));
         rc = mapPage(&systemDomain, la, pa, size,
                      change, global_access_attr, mem_attr);

         if ((HAL_SUCCESS != rc) && applicationDomain)
         {
            /*
             * The map operation was successful in the application
             * domain but failed in the system domain ->
             * unmap() from the application domain.
             */
            unmapRange(domain, la, size);
            
            MM_MULTITHREAD_UNLOCK(sr);
            return rc;
         }
      }

      /* Purge any stale translations in TLB. */
      if (change)
      {
#ifdef OSE_SMP
	 mmu_hal_invalidate_tlb_all();
#else
         /* FIXME: Only invalidate changed page */
	 mmu_hal_invalidate_tlb_local(la, size); 
#endif
      }

      MM_MULTITHREAD_UNLOCK(sr);

   }
   /* Else is not needed; it is valid to map into KSEG0/1
    * but it should impose no action. */

   return rc;
} /* mmu_hal_map_page(...) */

/*
 * See hal.h.
 */
static void
mmu_hal_unmap_range(struct hal_domain *domain,
                    log_addr_t         la,
                    log_size_t         size,
                    int                options)
{
   Msr sr = 0;

   VERBOZE((PRINT_PREFIX "hal_unmap_range(domain=%p, la=%#lx, size=%x,"
            "options=%x)\n",
            domain, la, size, options));
   assert(NULL != domain);

   if (mm_log_inside(la, KUSEG_BASE, KUSEG_SIZE) ||
       mm_log_inside(la, KSSEG_BASE, KSSEG_SIZE) ||
       mm_log_inside(la, KSEG3_BASE, KSEG3_SIZE))
   {
      MM_MULTITHREAD_LOCK(sr);

      /* First unconditionally un-map from local domain. */
      if (&systemDomain != domain)
      {
         unmapRange(domain, la, size);
      }

      /* Then un-map from system domain if last domain. */
      if (!(options & (HAL_SHARE | HAL_MAS)))
      {
         unmapRange(&systemDomain, la, size);
      }

#ifdef OSE_SMP
      mmu_hal_invalidate_tlb_all();
#else
      mmu_hal_invalidate_tlb_local(la, size); 
#endif

      MM_MULTITHREAD_UNLOCK(sr);
   }
   /* Else is not needed; it is valid to unmap from KSEG0/1
    * but it should impose no action. */
} /* mmu_hal_unmap_range(...) */

/*
 * See hal.h
 */
static hal_status_t
mmu_hal_lookup(OSDOMAIN          domain,
	       log_addr_t         la,
	       phys_addr_t       *pa,
	       log_size_t        *size,
	       U32               *perm,
               U32                options)
{
   struct hal_domain *hal_domain;
   L1Pte_t *l1Ptp;
   L1Pte_t *l1Pte;
   L2Pte_t *l2Ptp;
   L2Pte_t *l2Pte;
   Msr sr = 0;

   VERBOZE((PRINT_PREFIX "hal_lookup(domain=%p, la=%#lx, pa=%x, size=%x,"
            "perm=%x)\n",
            domain, la, pa, size, perm));


   if (!hal_info.enabled_mmu)
   {
      if (pa != NULL)
         *pa = (phys_addr_t)la;

      if (size != NULL)
         *size = HAL_SIZE_8K;

      if (perm != NULL)
         *perm = SU_RWX_USR_RWX;

      return HAL_SUCCESS;
   }
   
   spin_lock_irqsave(&hal_lock, sr);

   if (gmm_state.domains.domain_list[domain] == NULL ||
       gmm_state.domains.domain_list[domain]->hal_domain == NULL)
   {
      spin_unlock_irqrestore(&hal_lock, sr);
      return HAL_ELOOKUP_FAILED;
   }

   hal_domain = gmm_state.domains.domain_list[domain]->hal_domain;

   l1Ptp = getL1Ptp(hal_domain);
   l1Pte = getL1Pte(l1Ptp, la);
   l2Ptp = getL2Ptp(l1Pte);

   if (l2Ptp == NULL)
   {
      if (domain != 0)
      {
         /* Still possible that la is mapped in system domain */
         spin_unlock_irqrestore(&hal_lock, sr);
         return mmu_hal_lookup((OSDOMAIN)0, la, pa, size, perm, options);
      }
      else
      {
         spin_unlock_irqrestore(&hal_lock, sr);
         return HAL_ELOOKUP_FAILED;
      }
   }

   l2Pte = getL2Pte(l2Ptp, la);
   
   if (l2Pte == NULL)
   {
      spin_unlock_irqrestore(&hal_lock, sr);
      return HAL_ELOOKUP_FAILED;
   }

   /* If mapped in the current tree ... */
   if (l2Pte->entryLo != 0)
   {
      l2PteInfo(l2Pte, la, pa, size, perm);
      spin_unlock_irqrestore(&hal_lock, sr);
   }
   else if (domain != 0)
   {
      /* ... otherwise lookup la in system domain */
      spin_unlock_irqrestore(&hal_lock, sr);
      return mmu_hal_lookup((OSDOMAIN)0, la, pa, size, perm, options);
   }
   else
   {
      /* Not mapped anywhere */
      spin_unlock_irqrestore(&hal_lock, sr);
      return HAL_ELOOKUP_FAILED;
   }
      
   return HAL_SUCCESS;
}

#ifdef HOST_TEST
static phys_addr_t
hal_read_pte(OSDOMAIN domain, log_addr_t base)
{
   struct hal_domain *hal_domain;
   L1Pte_t *l1Ptp;
   L1Pte_t *l1Pte;
   L2Pte_t *l2Ptp;
   L2Pte_t *l2Pte;

   if (gmm_state.domains.domain_list[domain] == NULL ||
       gmm_state.domains.domain_list[domain]->hal_domain == NULL)
   {
      return 0;
   }

   hal_domain = gmm_state.domains.domain_list[domain]->hal_domain;

   l1Ptp = getL1Ptp(hal_domain);
   l1Pte = getL1Pte(l1Ptp, base);
   l2Ptp = getL2Ptp(l1Pte);

   if (l2Ptp == NULL)
   {
      if (domain != 0)
      {
         /* Still possible that la is mapped in system domain */
         return hal_read_pte((OSDOMAIN)0, base);
      }
      else
      {
         return 0;
      }
   }

   l2Pte = getL2Pte(l2Ptp, base);
   
   if (l2Pte == NULL)
   {
      return 0;
   }

   /* If mapped in the current tree ... */
   if (l2Pte->entryLo != 0)
   {
      return l2Pte->entryLo;
   }
   else if (domain != 0)
   {
      /* ... otherwise lookup la in system domain */
      return hal_read_pte((OSDOMAIN)0, base);
   }

   return 0;
}
#endif
   
static hal_status_t
uncommit_range(struct hal_domain *hal_domain,
               log_addr_t base, log_size_t size, U32 options)
{
   log_addr_t addr, top;
   L1Pte_t *l1Ptp;
   L1Pte_t *l1Pte;
   L2Pte_t *l2Ptp;
   Msr msr = 0;

   (void)options;
   
   if (size == 0)
   {
      return HAL_SUCCESS;
   }
   
   addr = ALIGN_DOWN(base, HAL_SIZE_8M);
   top  = ALIGN_UP(base + size, HAL_SIZE_8M) - 1;
   base = addr;

   l1Ptp = getL1Ptp(hal_domain);
   
   while (addr < top)
   {
      if (mm_log_inside(addr, KSEG0_BASE, KSEG0_SIZE) ||
          mm_log_inside(addr, KSEG1_BASE, KSEG1_SIZE))
      {
         addr += HAL_SIZE_8M;
         continue;
      }

      MM_MULTITHREAD_LOCK(msr);

      l1Pte = getL1Pte(l1Ptp, addr);
      l2Ptp = getL2Ptp(l1Pte);

      assert(l2Ptp != NULL);
      
      l1Pte->refCount -= (1 << L2_PAGE_TABLE_OFFSET_SHIFT);

      if (l1Pte->refCount == 0)
      {
         l1PteInvalidate(l1Pte);
         l2PageTableFree(l2Ptp);
      }
      
      MM_MULTITHREAD_UNLOCK(msr);
      
      addr += HAL_SIZE_8M;
   }

   return HAL_SUCCESS;
}

static hal_status_t
commit_range(struct hal_domain *hal_domain,
             log_addr_t base, log_size_t size, U32 options)
{
   log_addr_t addr, top;
   hal_status_t status;
   Msr msr = 0;
   L1Pte_t *l1Ptp;
   L1Pte_t *l1Pte;
   L2Pte_t *l2Ptp;

   if (size == 0)
   {
      return HAL_SUCCESS;
   }
   
   addr = ALIGN_DOWN(base, HAL_SIZE_8M);
   top  = ALIGN_UP(base + size, HAL_SIZE_8M) - 1;
   base = addr;

   l1Ptp = getL1Ptp(hal_domain);
   
   while (addr < top)
   {
      if (mm_log_inside(addr, KSEG0_BASE, KSEG0_SIZE) ||
          mm_log_inside(addr, KSEG1_BASE, KSEG1_SIZE))
      {
         addr += HAL_SIZE_8M;
         continue;
      }

      MM_MULTITHREAD_LOCK(msr);

      l1Pte = getL1Pte(l1Ptp, addr);
      l2Ptp = getL2Ptp(l1Pte);

      if (l2Ptp == NULL)
      {
         status = l2PageTableAlloc(&l2Ptp);
         if (status != HAL_SUCCESS)
         {
            uncommit_range(hal_domain, addr, addr - base, options);
            MM_MULTITHREAD_UNLOCK(msr);
            return status;
         }

         /* A recursive allocation and subsequent map_page might
          * already have created this l2 for us. */
         if (getL2Ptp(l1Pte) != 0)
         {
            l2PageTableFree(l2Ptp);
            l2Ptp = getL2Ptp(l1Pte);
         }
         else
         {
            l1PteUpdate(l1Pte, l2Ptp);
         } 
      }

      l1Pte->refCount += (1 << L2_PAGE_TABLE_OFFSET_SHIFT);
      
      MM_MULTITHREAD_UNLOCK(msr);

      addr += HAL_SIZE_8M;
   }

   return HAL_SUCCESS;
}

static hal_status_t
mmu_hal_commit_range(struct hal_domain *hal_domain,
                     log_addr_t base, log_size_t size, U32 options)
{
   hal_status_t status;

   assert(hal_domain != NULL);
   assert(size > 0);

   status = commit_range(&systemDomain, base, size, options);
   if (status != HAL_SUCCESS)
   {
      return status;
   }

   if (hal_domain != &systemDomain)
   {
      status = commit_range(hal_domain, base, size, options);
      if (status != HAL_SUCCESS)
      {
         uncommit_range(&systemDomain, base, size, options);
         return status;
      }
   }

   return HAL_SUCCESS;
}

static hal_status_t
mmu_hal_uncommit_range(struct hal_domain *hal_domain,
                       log_addr_t base, log_size_t size, U32 options)
{
   assert(hal_domain != NULL);
   assert(size > 0);
   
   uncommit_range(&systemDomain, base, size, options);

   if (hal_domain != &systemDomain)
   {
      uncommit_range(hal_domain, base, size, options);
   }
   
   return HAL_SUCCESS;
}


#ifdef OSE_SMP
/*
 * This function should execute on the cpu that is beeing initiated.
 * For example on SMP system with 2 cpu's it is executed only cpu 1 and
 * not on the master cpu 0.
 */
void
hal_init_other_cpu(unsigned long seg_id)
{
   (void) seg_id;

   ramlog_printf("mm: Init cpu %u.\n", zzkrn_cpu_id());

   /* Cpu 0 should never call this function. */
   assert(zzkrn_cpu_id() != 0);

   set32_cp0(C0_PAGEMASK, 0); /* Select 4k physical page size. */

#ifndef _MIPS_ARCH_MIPS3
   /* Enable support for large physical addresses. (bit 29)*/
   set32_cp0_sel1(CO_PAGEGRAIN, 0x20000000);
#endif

   mmu_hal_invalidate_tlb_local(0, ~0); /* FIXME */

   set32_cp0(C0_ENTRYHI, ASID_SYSTEM);
}
#endif


static hal_status_t validate_phys_range(phys_addr_t base,
                                        phys_size_t size)
{
   phys_addr_t top = base + size - 1ULL;

   if (size == 0)
   {
      return HAL_SUCCESS;
   }
   
   /* 
    * base > top => base + size has wrapped.
    * bit 63 and 62 is used to store user read and write permissions
    */
   if (0 != (top >> 62) || base > top)
   {
      ramlog_printf("Unsupported range [0x%08llx-0x%08llx]. "
         "Supported range is [0x00000000-0xffffffff]\n",
         base, top); 
      return HAL_EUNSUPPORTED_PHYSICAL_RANGE;
   } 
   
   return HAL_SUCCESS;
}

/**************************************************************************/

static hal_status_t
mmu_hal_valid_cmbits(U32 *cbits, U32 *mbits, U32 *cmbits, log_addr_t base)
{
   if (cmbits != NULL)
   {
      int kseg0 = mm_log_inside(base, KSEG0_BASE, KSEG0_SIZE);
      int kseg1 = mm_log_inside(base, KSEG1_BASE, KSEG1_SIZE);
      
      if (kseg0 && !(*cmbits & HAL_MIPS_CACHE_ENABLE))
      {
         /* We cannot inhibit cacheability in KSEG0. */
         return HAL_EILLEGAL_CACHE_MODE;
      }
      
      if (kseg1 && (*cmbits & HAL_MIPS_CACHE_ENABLE))
      {
         /* We cannot provide cacheability in KSEG1. */
         return HAL_EILLEGAL_CACHE_MODE;
      }

      if (*cmbits & ~HAL_MIPS_CACHE_ENABLE)
         return HAL_EILLEGAL_CACHE_MODE;
   }

   if (mbits != NULL && *mbits != 0)
      return HAL_EILLEGAL_CACHE_MODE;
   
   if (cbits != NULL && (*cbits & ~HAL_MIPS_CACHE_ENABLE))
      return HAL_EILLEGAL_CACHE_MODE;

   return HAL_SUCCESS;
}

/**************************************************************************/

static hal_status_t
mmu_hal_legacy_cmbits(U32 cmbits, U32 *cache_mode)
{
   if (cmbits & HAL_MIPS_CACHE_ENABLE)
      *cache_mode |= MM_CACHE_COPY_BACK|MM_SPECULATIVE_ACCESS;
   else
      *cache_mode |= MM_CACHE_INHIBITED|MM_ORDERED_ACCESS;
         
   return HAL_SUCCESS;
}

/**************************************************************************/

/* void OSE_MM_HAL_UNIQUELY_DEFINED(void) */
/* { */
/* } */

void install_mmu(void)
{
   hal_init_1 = mmu_hal_init_1;
   hal_init_2 = mmu_hal_init_2;
   hal_init_3 = mmu_hal_init_3;
   hal_init_4 = mmu_hal_init_4;
   hal_init_5 = mmu_hal_init_5;
   hal_get_info = mmu_hal_get_info;
   hal_actual_permissions = mmu_hal_actual_permissions;
   hal_create_domain = mmu_hal_create_domain;
   hal_select_domain = mmu_hal_select_domain;
   hal_delete_domain = mmu_hal_delete_domain;
   hal_map_page = mmu_hal_map_page;
   hal_unmap_range = mmu_hal_unmap_range;
   hal_lookup = mmu_hal_lookup;
   hal_validate_phys_range = validate_phys_range;
   hal_commit_range = mmu_hal_commit_range;
   hal_uncommit_range = mmu_hal_uncommit_range;
   hal_invalidate_tlb_local = mmu_hal_invalidate_tlb_local;

   hal_valid_cmbits  = mmu_hal_valid_cmbits;
   hal_legacy_cmbits = mmu_hal_legacy_cmbits;
} /* install_mmu(...) */
