/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2006 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

/****************************************************************
* $workfile: <file name> $
* Document no: @(#) $FileRevision:   /main/tb_current_ose5/3  $
* $date: $
*
* Purpose: C language program file
* Description: Implementation of the FAM access lock mechanism.
*
*/


#include "ose.h"
#include "cpu.h"
#include "assert.h"
#include "flash_api.h"
#include "flash_drv.h"
#include "flash_err.h"
#include "fam_internal.h"


#define NOTIFY_SIG   36400    /* Sent to notify new lock owner. */
#define TRIG_SIG     36401    /* Sent to request suspend */

#define VERY_LONG_TIMEOUT ((OSTIME)-1)

union SIGNAL
{
   SIGSELECT   sig_no;
};



/****************************************************************
*  STATIC VARIABLES
*/
static const SIGSELECT any_sig[] = { 0 };



/****************************************************************
*  Function queue_access_lock
*
*  Places an AccessLock in queue.
*  A trig signal is sent to the current owner if needed.
*
*  IMPLEMENTATION NOTES:
*  This function cannot place a lock as owner.
*  Shall be called with interrupts DISABLED.
*  Will return with interrupts ENABLED.
*/
static void
queue_access_lock(ModifyLock *modify_lock,
                  AccessLock *my_lock,
                  U32 int_mask_save)
{
   AccessLock  *lock_owner;
   AccessLock  *insert_position;

   /*
   *  Insert my_lock in the wait queue of the current lock owner.
   *  It is inserted at a position after all entries with an activity type
   *  of a higher priority.
   *
   *  Note: This is implemented as a search loop with a theoretical maximum
   *        length equal to the number of threads that are clients to FAM.
   *        The lack of a guaranteed maximum interrupt latency is countered
   *        by the unlikelyhood of a very large number of clients together
   *        with the advantages of a simple design.
   */
   lock_owner = modify_lock->lock_owner;
   insert_position = lock_owner;
   while (insert_position->next_queue != NULL)
   {
      if (my_lock->activity_type > insert_position->next_queue->activity_type)
      {
         /* Finished search if we have a higher priority activity type */
         /* than next in queue. */
         break;
      }

      /* Go to next position. */
      insert_position = insert_position->next_queue;
   }

   /* Insert my_lock after insert_position in the queue. */
   my_lock->next_queue = insert_position->next_queue;
   insert_position->next_queue = my_lock;

   /*
   *  Send a trig signal to the lock owner, if needed.
   */
   if ((lock_owner->sig_wait_state == SWS_EXPECTING_TRIG_SIGNAL) &&
       (lock_owner->activity_type < my_lock->activity_type))
   {
      union SIGNAL *trig_sig;
      lock_owner->sig_wait_state = SWS_TRIG_SIGNAL_SENT;
      FAM_ENABLE_INTERRUPT(int_mask_save);
      trig_sig = alloc(sizeof (SIGSELECT), TRIG_SIG);
      send_w_s(&trig_sig, fam_global.common_pid, lock_owner->owner_pid);
      WHITEBOX_TESTPOINT(TPT_SYNCH_SYSCALL);
   }
   else
   {
      FAM_ENABLE_INTERRUPT(int_mask_save);
   }
}


/****************************************************************
*  Function wait_for_notification
*
*  Wait for notification that we have been placed as owner.
*
*  Should be called with interrupts enabled.
*/
static void
wait_for_notification(void)
{
   union SIGNAL *notify_sig;

   WHITEBOX_TESTPOINT(TPT_SYNCH_SYSCALL);
   notify_sig = receive_from(
      VERY_LONG_TIMEOUT, any_sig, fam_global.common_pid);
   WHITEBOX_TESTPOINT(TPT_SYNCH_SYSCALL);
   assert(notify_sig != NIL);
   assert(notify_sig->sig_no == NOTIFY_SIG);
   free_buf(&notify_sig);
}


/****************************************************************
*  Function fam_acquire_read_lock
*
*  Acquires read access rights.
*  Waits in queue for access rights if necessary. 
*/
void
fam_acquire_read_lock(ReadLock *read_lock)
{
   U32 int_mask_save = 0;  /* Must be initialized to avoid compiler warning. */
   AccessLock *my_lock;
   Boolean is_owner;
   AccessLock *new_owner;

   /*
   *  Check if we can proceed without conflicts.
   */
   FAM_DISABLE_INTERRUPT(int_mask_save);
   if (read_lock->access_count >= 0)
   {
      read_lock->access_count++;
      FAM_ENABLE_INTERRUPT(int_mask_save);
      return;
   }

   /*
   *  There is a conflict
   *  Resolve it by creating an AccessLock and waiting until all other
   *  locks have been released.
   *  Allocate an AccessLock from the linked list of free AccessLocks
   */
   if (fam_global.free_lock_list == NULL)
   {
      fam_internal_error(FLASH_ETOO_MANY_SIMULTANEOUS_USERS, 0);
   }
   my_lock = fam_global.free_lock_list;
   fam_global.free_lock_list = my_lock->next_queue;
   my_lock->next_queue = NULL;
   my_lock->activity_type = AT_READ;
   my_lock->associated_read_lock = read_lock;
   my_lock->owner_pid = 0;    /* Ignore current process until we need it. */
   my_lock->owner_prio = 0;   /* prio=0 disables priority inheritance */ 
   my_lock->prio_raised = False;  /* True=priority inheritance in progress */ 
   my_lock->sig_wait_state = SWS_NOT_EXPECTING_SIGNAL;

   /*
   *  Insert the AccessLock in the queue.
   *  Also enables interrupts
   */
   assert(read_lock->modify_lock->lock_owner != NULL);
   queue_access_lock(read_lock->modify_lock, my_lock, int_mask_save);

   /*
   *  Prepare to receive notification.
   *  It would have been advantageous to do this before enabling interrupts,
   *  but we cannot because OSE has undefined behaviour if called with
   *  disabled interrupts. 
   *  Therefore there is a possibility that we have already been placed as
   *  owners, and that we should not expect notification.
   */
   my_lock->owner_pid = current_process();
   FAM_DISABLE_INTERRUPT(int_mask_save);
   is_owner = (read_lock->modify_lock->lock_owner == my_lock);
   if (!is_owner)
   {
      AccessLock *lock_owner = read_lock->modify_lock->lock_owner;
      my_lock->sig_wait_state = SWS_EXPECTING_NOTIFY_SIGNAL;
      FAM_ENABLE_INTERRUPT(int_mask_save);
      if (lock_owner->owner_prio != 0)
      {
         /* 
          * The lock owner is a prio process (1-31) that should be raised
          * temporarily to reach suspend state if lower than the reader. 
          */
         OSPRIORITY my_prio = get_pri(my_lock->owner_pid);
         if (my_prio < lock_owner->owner_prio)
         {
            /* set flag to avoid unnecessary syscalls */
            lock_owner->prio_raised = True;
            set_pri_for(lock_owner->owner_pid, my_prio);
         }
      }
      wait_for_notification();
      FAM_DISABLE_INTERRUPT(int_mask_save);
      assert(my_lock->sig_wait_state == SWS_NOT_EXPECTING_SIGNAL);
   }

   /*
   *  We now have the lock, which makes it OK to update the access count.
   */
   assert(read_lock->modify_lock->lock_owner == my_lock);
   assert(read_lock->access_count >= 0);
   read_lock->access_count++;

   /*
   *  Remove my_lock from the head of the queue. It is not needed any more.
   *  Make the head of the queue the new lock owner. Note that the new owner
   *  will not be notified until after the read operation is complete.
   */
   new_owner = read_lock->modify_lock->lock_owner->next_queue;
   read_lock->modify_lock->lock_owner = new_owner;

   /* Return my_lock to the linked list of free AccessLocks */
   assert(my_lock != NULL);
   my_lock->next_queue = fam_global.free_lock_list;
   fam_global.free_lock_list = my_lock;

   FAM_ENABLE_INTERRUPT(int_mask_save);

   /*
   *  If we were the last to have the lock, and throttling is activated,
   *  we should release the throttle semaphore. The throttle semaphore is
   *  acquired when a lock_owner that was previously NULL is set, and is
   *  later released when the lock_owner is set to NULL again; it does not
   *  matter if the lock owner is a reader, writer or eraser.
   */
   if ((new_owner == NULL) && IS_FAM_THROTTLE_ACTIVATED())
   {
      LEAVE_FAM_THROTTLE();
   }
}


/****************************************************************
*  Function fam_acquire_modify_lock
*
*  Acquires modify access rights.
*  Waits in queue for access rights if necessary. 
*  Always acquired throttle semaphore (if any) on return.
*/
AccessLock *
fam_acquire_modify_lock(ReadLock *read_lock,
                        ActivityType activity_type)
{
   U32 int_mask_save = 0;  /* Must be initialized to avoid compiler warning. */
   AccessLock *my_lock;

   /*
   *  Allocate and initialize the access lock.
   *  Allocate an AccessLock from the linked list of free AccessLocks
   */
   FAM_DISABLE_INTERRUPT(int_mask_save);
   if (fam_global.free_lock_list == NULL)
   {
      fam_internal_error(FLASH_ETOO_MANY_SIMULTANEOUS_USERS, 0);
   }
   my_lock = fam_global.free_lock_list;
   fam_global.free_lock_list = my_lock->next_queue;
   my_lock->next_queue = NULL;
   my_lock->activity_type = activity_type;
   my_lock->associated_read_lock = read_lock;
   my_lock->owner_pid = 0;    /* Ignore current process until we need it. */
   my_lock->owner_prio = 0;   /* prio=0 disables priority inheritance */ 
   my_lock->prio_raised = False;  /* True=priority inheritance in progress */ 
   my_lock->sig_wait_state = SWS_NOT_EXPECTING_SIGNAL;

   /*
   *  Check if we can get ownership without waiting.
   *  There are two conditions for this:
   *  1. There must not be a modify operation in progress in the modify lock
   *     region, indicated by the presence of a lock owner.
   *  2. There must not be a read operation in progress in the read lock
   *     region, indicated by access_count.
   */
   do
   {
      if (read_lock->modify_lock->lock_owner == NULL)
      {
         /* There is no existing lock owner. Place my_lock as owner. */
         
         if (IS_FAM_THROTTLE_ACTIVATED())
         {
            /* As there is no lock_owner, the modify_lock is not in the throttle.
             * We don't want to place ourselves as lock_owner before getting the
             * throttle semaphore, as we may then prevent reads from taking place
             * in the region while we wait on the semaphore.
             */
            FAM_ENABLE_INTERRUPT(int_mask_save);
            ENTER_FAM_THROTTLE();
            FAM_DISABLE_INTERRUPT(int_mask_save);
            
            /* Now hold the throttle semaphore -- but there is a (small) risk
             * that someone else is now lock_owner of this modify_lock...
             */
            if (read_lock->modify_lock->lock_owner != NULL)
            {
               /* There is a lock owner. Place my_lock in queue. */
               /* Also enables interrupts. */
               queue_access_lock(read_lock->modify_lock, my_lock, int_mask_save);
               
               /* We should not hold the throttle semaphore now, as the owner
                * of the modify_lock already holds it.
                */
               LEAVE_FAM_THROTTLE();
               break;
            }
         }
         assert(read_lock->modify_lock->lock_owner == NULL);
         assert(read_lock->access_count >= 0);
         read_lock->modify_lock->lock_owner = my_lock;
         if (read_lock->access_count == 0)
         {
            /* There is no read access in progress. */
            read_lock->access_count--;
            FAM_ENABLE_INTERRUPT(int_mask_save);
            return my_lock;
         }
         FAM_ENABLE_INTERRUPT(int_mask_save);
      }
      else
      {
         /* There is a lock owner. Place my_lock in queue. */
         /* Also enables interrupts. */
         queue_access_lock(read_lock->modify_lock, my_lock, int_mask_save);
      }
   }
   while (0);
   
   /*
   *  The AccessLock has been placed in queue, or as owner. We shall wait
   *  until we are lock owner, and there is no conflicting read access in
   *  progress. This is done by waiting for a notification unless the 
   *  conditions have already been fulfilled. 
   */
   my_lock->owner_pid = current_process();
   /* 
    * We need to tell any potential higher process that wants to
    * acquire read lock our prio, so it can raise us temporarily
    * in order to reach the suspend state.
    */
   my_lock->owner_prio = get_pri(my_lock->owner_pid);
   FAM_DISABLE_INTERRUPT(int_mask_save);
   while ((read_lock->modify_lock->lock_owner != my_lock) ||
          (read_lock->access_count != 0))
   {
      my_lock->sig_wait_state = SWS_EXPECTING_NOTIFY_SIGNAL;
      FAM_ENABLE_INTERRUPT(int_mask_save);
      wait_for_notification();
      assert(my_lock->sig_wait_state == SWS_NOT_EXPECTING_SIGNAL);
      FAM_DISABLE_INTERRUPT(int_mask_save);
   }

   /* The conditions for modify access are met. */
   read_lock->access_count--;

   FAM_ENABLE_INTERRUPT(int_mask_save);

   return my_lock;
}


/****************************************************************
*  Function notify_lock_owner
*
*  Notifies lock owner if needed.
*
*  IMPLEMENTATION NOTES:
*  Shall be called with interrupts DISABLED.
*  Will return with interrupts ENABLED.
*/
static void
notify_lock_owner(AccessLock *lock_owner,
                  U32 int_mask_save)
{
   /*
   *  Check if we should send a noification to the lock owner. Conditions:
   *     *  There is a lock owner.
   *     *  The lock owner wants a notification.
   *     *  An owner that is modifying the flash should only receive 
   *        notification when the last read access is over.
   *     *  An owner that wants to perform read access should receive
   *        notification even if the last read is not over.
   */
   if ((lock_owner != NULL) &&
       (lock_owner->sig_wait_state == SWS_EXPECTING_NOTIFY_SIGNAL) &&
       ((lock_owner->associated_read_lock->access_count == 0) ||
        (lock_owner->activity_type == AT_READ)))
   {
      union SIGNAL *notify_sig;

      lock_owner->sig_wait_state = SWS_NOT_EXPECTING_SIGNAL;
      FAM_ENABLE_INTERRUPT(int_mask_save);

      notify_sig = alloc(sizeof (SIGSELECT), NOTIFY_SIG);
      send_w_s(&notify_sig, fam_global.common_pid, lock_owner->owner_pid);
      WHITEBOX_TESTPOINT(TPT_SYNCH_SYSCALL);
   }
   else
   {
      FAM_ENABLE_INTERRUPT(int_mask_save);
   }
}


/****************************************************************
*  Function fam_release_read_lock
*
*  Releases read access rights.
*  Notifies next lock owner if needed.
*/
void
fam_release_read_lock(ReadLock *read_lock)
{
   U32 int_mask_save = 0;  /* Must be initialized to avoid compiler warning. */

   FAM_DISABLE_INTERRUPT(int_mask_save);
   assert(read_lock->access_count > 0);
   read_lock->access_count--;

   /* The following enables interrupts. */
   notify_lock_owner(read_lock->modify_lock->lock_owner, int_mask_save);
}


/****************************************************************
*  Function fam_release_modify_lock
*
*  Releases modify access rights.
*  Notifies next lock owner if needed.
*/
void
fam_release_modify_lock(AccessLock *access_lock)
{
   Boolean prio_raised;
   OSPRIORITY owner_prio;
   ModifyLock *modify_lock;
   U32 int_mask_save = 0;  /* Must be initialized to avoid compiler warning. */
   Boolean should_leave_throttle = False;

   FAM_DISABLE_INTERRUPT(int_mask_save);

   /* catch the flag and prio before releasing the lock and enabling ints. */
   prio_raised = access_lock->prio_raised;
   owner_prio = access_lock->owner_prio;

   assert(access_lock->associated_read_lock->access_count == -1);
   access_lock->associated_read_lock->access_count = 0;

   /*
   *  Remove access_lock from the lock owner position.
   *  Make the head of the queue the new lock owner.
   */
   modify_lock = access_lock->associated_read_lock->modify_lock;
   assert(modify_lock->lock_owner == access_lock);
   modify_lock->lock_owner = modify_lock->lock_owner->next_queue;

   /* Check if we should release the throttle semaphore. We should, if we are
    * the last to hold this modify_lock.
    */
   if (IS_FAM_THROTTLE_ACTIVATED())
   {
      if (modify_lock->lock_owner == NULL)
      {
         should_leave_throttle = True;
      }
   }

   /* Return access_lock to the linked list of free AccessLocks */
   assert(access_lock != NULL);
   access_lock->next_queue = fam_global.free_lock_list;
   fam_global.free_lock_list = access_lock;

   /* The following enables interrupts. */
   notify_lock_owner(modify_lock->lock_owner, int_mask_save);

   /* Release the throttle semaphore (if necessary). */
   if (should_leave_throttle)
   {
      LEAVE_FAM_THROTTLE();
   }

   /* If we run on a temporariliy raised level,
    * set back to original priority.
    */
   if (prio_raised)
   {
      set_pri(owner_prio);
   }
}


/****************************************************************
*  Function fam_poll_suspend_request
*
*  RETURNS
*  True if an access lock with an activity type of a higher priority is
*  on queue. False otherwise.
*/
Boolean
fam_poll_suspend_request(AccessLock *access_lock)
{
   Boolean result;

   WHITEBOX_TESTPOINT(TPT_SYNCH);
   result = ((access_lock->next_queue != NULL) &&
      (access_lock->next_queue->activity_type > access_lock->activity_type));
   WHITEBOX_TESTPOINT(TPT_SYNCH);
   return result;
}

/****************************************************************
*  Function fam_wait_suspend_request
*
*  Waits for a suspend request during selected timeout-period.
*
*  RETURNS
*  True if an access lock with an activity type of a higher priority is
*  on queue. False otherwise.
*/
Boolean
fam_wait_suspend_request(AccessLock *access_lock,
                         U32 timeout_ms)
{
   union SIGNAL *trig_sig;
   U32 int_mask_save;
   Boolean result;

   access_lock->owner_pid = current_process();

   FAM_DISABLE_INTERRUPT(int_mask_save);

   /*
   *  Check if a suspend request was issued before entering state 
   *  SWS_EXPECTING_TRIG_SIGNAL.
   */
   if (fam_poll_suspend_request(access_lock))
   {
      FAM_ENABLE_INTERRUPT(int_mask_save);
      return True;
   }

   access_lock->sig_wait_state = SWS_EXPECTING_TRIG_SIGNAL;
   FAM_ENABLE_INTERRUPT(int_mask_save);

   trig_sig = receive_from(
      timeout_ms, any_sig, fam_global.common_pid);
   WHITEBOX_TESTPOINT(TPT_SYNCH_SYSCALL);

   if (trig_sig == NIL)
   {
      /* Timeout occured. */
      FAM_DISABLE_INTERRUPT(int_mask_save);

      if (access_lock->sig_wait_state == SWS_EXPECTING_TRIG_SIGNAL)
      {
         /* No trig signal during timeout. */
         access_lock->sig_wait_state = SWS_NOT_EXPECTING_SIGNAL;
         FAM_ENABLE_INTERRUPT(int_mask_save);
      }
      else
      {
         /* A trig signal was sent during the timeout, but not received.*/
         FAM_ENABLE_INTERRUPT(int_mask_save);
         trig_sig = receive_from(
            VERY_LONG_TIMEOUT, any_sig,
            fam_global.common_pid);
         WHITEBOX_TESTPOINT(TPT_SYNCH_SYSCALL);
         assert(trig_sig != NIL);
      }
   }

   if (trig_sig != NIL)
   {
      /* A trig signal has been received. */
      assert(trig_sig->sig_no == TRIG_SIG);
      assert(access_lock->sig_wait_state == SWS_TRIG_SIGNAL_SENT);
      access_lock->sig_wait_state = SWS_NOT_EXPECTING_SIGNAL;
      free_buf(&trig_sig);
   }

   assert(access_lock->sig_wait_state == SWS_NOT_EXPECTING_SIGNAL);

   result = fam_poll_suspend_request(access_lock);
   return result;
}


/****************************************************************
*  Function fam_wait_during_suspend
*
*  Allows the next process on access queue to perform it's access, and
*  returns when that access is over.
*/
void
fam_wait_during_suspend(AccessLock *access_lock)
{
   ModifyLock *modify_lock;
   U32 int_mask_save = 0;  /* Must be initialized to avoid compiler warning. */

   FAM_DISABLE_INTERRUPT(int_mask_save);

   assert(fam_poll_suspend_request(access_lock));
   assert(access_lock->associated_read_lock->modify_lock->lock_owner ==
          access_lock);
   assert(access_lock->associated_read_lock->access_count == -1);

   /*
   *  Remove access_lock from the lock owner position.
   *  Make the head of the queue the new lock owner.
   */
   modify_lock = access_lock->associated_read_lock->modify_lock;
   modify_lock->lock_owner = modify_lock->lock_owner->next_queue;
   access_lock->associated_read_lock->access_count = 0;

   /*
   *  Change activity type to AT_SUSPENDED_ERASE or AT_SUSPENDED_WRITE.
   */
   access_lock->activity_type++;

   /*
   *  Place the lock in queue until higher priority activities are completed.
   */
   queue_access_lock(modify_lock, access_lock, int_mask_save);
   
   FAM_DISABLE_INTERRUPT(int_mask_save);
   /* The following enables interrupts. */
   notify_lock_owner(modify_lock->lock_owner, int_mask_save);
   
   /* If we were temporariliy raised, set back to original priority
    *  since we are not the lock owner any more.
    */
   if (access_lock->prio_raised)
   {
      assert(access_lock->owner_prio);
      access_lock->prio_raised = 0; /* reset flag. */
      set_pri(access_lock->owner_prio);
   }
   
   /*
   *  The AccessLock has been placed in queue. We shall wait
   *  until we are lock owner, and there is no conflicting read access in
   *  progress. This is done by waiting for a notification, unless the 
   *  conditions have already been fulfilled. 
   */
   access_lock->owner_pid = current_process();
   FAM_DISABLE_INTERRUPT(int_mask_save);
   while ((modify_lock->lock_owner != access_lock) ||
          (access_lock->associated_read_lock->access_count != 0))
   {
      access_lock->sig_wait_state = SWS_EXPECTING_NOTIFY_SIGNAL;
      FAM_ENABLE_INTERRUPT(int_mask_save);
      wait_for_notification();
      assert(access_lock->sig_wait_state == SWS_NOT_EXPECTING_SIGNAL);
      FAM_DISABLE_INTERRUPT(int_mask_save);
   }

   /* The conditions for modify access are met. */
   access_lock->associated_read_lock->access_count--;

   /* Restore activity type. */
   access_lock->activity_type--;

   FAM_ENABLE_INTERRUPT(int_mask_save);
}
