/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2006 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

/****************************************************************
* $workfile: <file name> $
* Document no: @(#) $FileRevision:   /main/tb_current_ose5/3  $
* $date: $
*
* Purpose: C language program file
* Description: Implementation of functions declared in flash_conf.h.
*
*/


#include "flash_conf.h"
#include "flash_err.h"
#include "flash_biosinterface.h"
#include "ose.h"
#include "bios.h"
#include "assert.h"
#include "string.h"
#include "stddef.h"
#include "fam_internal.h"

#define SUPPORTED_DRIVER_VERSION 1

static Boolean is_api_available = False;  /* True if bios module has been 
                                          *  installed.
                                          */

/****************************************************************
*  struct DriverInfo
*
*  Information about an installed driver.
*  The information is only used during flash instantiation.
*/
typedef struct DriverInfo
{
   FlashDriver                   installed_driver;
   const union FlashDriverInterface  *driver_interface;
   U8                            if_version;
   char                          driver_name[FLASH_DRIVER_MAX_NAME_LEN + 1];
} DriverInfo;



/****************************************************************
*  List of installed drivers. This list is not sorted.
*/
static U32           driver_count = 0;    /* Total number of drivers.   */
static DriverInfo    *driver_list = NULL; /* Array of installed drivers.*/



/****************************************************************
*  DEFINITIONS OF GLOBAL VARIABLES
*
*  See fam_internal.h for details.
*****************************************************************
*/
static InstanceIndex    empty_list_end_entry = { 0, 0, NULL };

#ifdef FAM_TEST
static AccessLock       *free_lock_buffer = NULL;  /* Only for debug. */
#endif

FamGlobal fam_global = {
   0,
   &empty_list_end_entry,
   0,
   NULL,
   NULL,
   0,
#ifdef FAM_THROTTLE_TEST
   0,
   0,
#endif /* FAM_THROTTLE_TEST */
   NULL,
   NULL
};

#ifdef FAM_THROTTLE_TEST
const size_t fam_global_size = sizeof(fam_global);
#endif /* FAM_THROTTLE_TEST */



/****************************************************************
*  Functions fam_alloc_mem and fam_free_mem
*
*  Wrappers for memory allocation. Semantics equivalent to malloc and free.
*
*  The purpose of a wrapper is to simplify porting to other memory
*  allocators, e.g. DDA/DRM or heap.
*/
#ifdef ALLOC_FROM_HEAP  /* Allocation using malloc and free. */
#include "stdlib.h"
#endif

static void *
fam_alloc_mem(size_t size)
{
#ifdef ALLOC_FROM_HEAP  /* Allocation using malloc and free. */
   void *mem = malloc(size);
   if (mem == NULL)
      fam_internal_error(FLASH_EMEMORY_ALLOCATION_FAILURE, 0);
   return mem;
#else                   /* Alloc and free of OSE signal buffers. */
   return (void *)alloc(size, 0);
#endif
}

static void
fam_free_mem(void *buffer)
{
#ifdef ALLOC_FROM_HEAP  /* Allocation using malloc and free. */
   free(buffer);
#else                   /* Alloc and free of OSE signal buffers. */
   if (buffer != NULL)
   {
      free_buf((union SIGNAL **)&buffer);
   }
#endif
}


/****************************************************************
 * Function S32_ge_U32(S32 x, U32 y) returns (x >= y). It handles x <0
 * correctly and avoids compiler warnings
 */
static U32 S32_ge_U32(S32 x, U32 y) {
  return (x >= 0) && ((U32)x >= y);
}

/****************************************************************
*  Functions dummy_irq_poll_function and dummy_micro_delay_timer
*
*  Dummy functions to be used if FAM is not configured with corresponding
*  real functions. See function flash_initialize for details.
*/
static int
dummy_irq_poll_function(void)
{
   return IRQ_POLL_NO_PENDING_INTERRUPT;
}

static void
dummy_micro_delay_timer(U32 microseconds)
{
   (void)microseconds;  /* Avoid compiler warning */
}


/****************************************************************
*  Function flash_initialize
*
*  This function performs basic initialization of the flash access module.
*  It must be called before any any other flash function is called.
*
*  See flash_conf.h for details.
*/
extern FlashConfResult
flash_initialize(int (*irq_poll_function)(void),
                 void (*micro_delay_timer)(U32 microseconds),
                 U32 max_simultaneous_users)
{
   U32 i;

   /* Install the irq poll function. */
   if (irq_poll_function == NULL)
   {
      irq_poll_function = dummy_irq_poll_function;
   }
   fam_global.irq_poll_function = irq_poll_function;

   /* Install the micro delay timer. */
   if (micro_delay_timer == NULL)
   {
      micro_delay_timer = dummy_micro_delay_timer;
   }
   fam_global.micro_delay_timer = micro_delay_timer;

   /* Pre-allocate access locks */
   fam_global.free_lock_list =
      fam_alloc_mem(max_simultaneous_users * sizeof (AccessLock));

#ifdef FAM_TEST
   free_lock_buffer = fam_global.free_lock_list; /* Only for debug. */
#endif

   for (i = 0; i < (max_simultaneous_users - 1); i++)
   {
      fam_global.free_lock_list[i].next_queue =
         &fam_global.free_lock_list[i + 1];
   }
   fam_global.free_lock_list[i].next_queue = NULL;

   /* Throttling off by default. No need to initialize fam_global.throttle_limit,
    * as it's only used when fam_global.throttle_sem is non-NULL.
    */
   fam_global.throttle_sem = NULL;

   /* Create a phantom process to be used as sender for FAM internal signals. */
   fam_global.common_pid = create_process(OS_PHANTOM,  /* proc_type */
                                          "FAM_COMMON_SENDER", /* name */
                                          NULL,  /* entrypoint */
                                          0, /* stack_size */
                                          0, /* priority */
                                          0, /* timeslice */
                                          0, /* block. 0 = caller's block. */
                                          NULL, /* redir_table */
                                          0, /* vector */
                                          0); /* user. 0 = default user. */

   return FLASHCONF_EOK;
}


/****************************************************************
*  Function flash_throttle_activate
*
*  This function activates and initializes the FAM throttling
*  functionality
*
*  See flash_conf.h for details.
*/
extern FlashConfResult
flash_throttle_activate(int initial_limit)
{
   /* Verify that flash_initialize has been sucessfully executed. */
   if (fam_global.common_pid == 0)
   {
      return FLASHCONF_ENOT_INITIALIZED;
   }

   /* Set up the throttle. */
   fam_global.throttle_limit = initial_limit;
   fam_global.throttle_sem = create_sem(initial_limit < 0 ? 0 : initial_limit);
#ifdef FAM_THROTTLE_TEST
   fam_global.throttle_modifiers_count = 0;
   fam_global.throttle_max_modifiers_count = 0;
#endif /* FAM_THROTTLE_TEST */

   return FLASHCONF_EOK;
}


/****************************************************************
*  Function flash_install_bios_module
*
*  This function makes The Flash Access Manager install itself as a
*  BIOS module.
*
*  See flash_conf.h for details.
*/
extern FlashConfResult
flash_install_bios_module(void)
{
   long bios_result;

#ifdef LIBWHAT
   LIBWHAT
#endif

   /* Verify that flash_initialize has been sucessfully executed. */
   if (fam_global.common_pid == 0)
   {
      return FLASHCONF_ENOT_INITIALIZED;
   }

   bios_result = biosInstall(FLASH_BIOSMODULE_NAME,
                             fam_biosfunction,
                             BIOS_USERMODE_ALLOWED);

   if (bios_result != 0)
   {
      fam_internal_error(FLASHCONF_EBIOS_INSTALL_FAILED, bios_result);
   }

   is_api_available = True;

   return FLASHCONF_EOK;
}



/****************************************************************
*  Function flash_is_api_available
*
*  This function determines whether the services provided via flash_api.h are
*  available. (In essence whether flash_install_bios_module has been called.)
*
*  See flash_conf.h for details.
*/
extern Boolean
flash_is_api_available(void)
{
   return is_api_available;
}



/****************************************************************
*  Function flash_install_driver
*
*  This function installs a flash driver with the Flash Access Manager and
*  initializes the driver.
*
*  See flash_conf.h for details.
*/
extern FlashConfResult
flash_install_driver(FlashDriver driver,
                     const char *driver_name,
                     const char *parameter_string)
{
   DriverInfo *new_list;
   DriverInfo *old_list;
   const union FlashDriverInterface *driver_interface;
   FlashDriverResult result;
   U8 if_version;

   /*
   *  Initialize the driver.
   */

   do
   {
      /* First try version 2 of the interface. This version is backwards
       * compatible with version 1. */
      driver_interface = driver(2);
      if (driver_interface != NULL)
      {
         if_version = 2;
         break; /* Valid interface found, continue. */
      }

      driver_interface = driver(1);
      if (driver_interface == NULL)
      {
         return FLASHCONF_EUNSUPPORTED_DRIVER_VERSION;
      }

      /* Use version 1 interface */
      if_version = 1;
   } while (0);

   result = driver_interface->v1.initialize_driver(parameter_string);
   if (result != FLASHDRIVER_EOK)
   {
      return FLASHCONF_EDRIVER_INIT_FAILURE;
   }

   /*
   *  The driver_list is an array of DriverInfo structs. Reallocate
   *  driver_list to a buffer that is large enough to hold one more entry.
   */
   new_list = fam_alloc_mem((driver_count + 1) * sizeof(DriverInfo));

   /* Move the old driver list to the new list. */
   if (driver_count > 0)
   {
      memcpy(new_list, driver_list, driver_count * sizeof(DriverInfo));
   }

   /* Initialize the last entry with info about the new driver. */
   new_list[driver_count].installed_driver = driver;
   new_list[driver_count].driver_interface = driver_interface;
   new_list[driver_count].if_version = if_version;
   strncpy(new_list[driver_count].driver_name, driver_name,
      FLASH_DRIVER_MAX_NAME_LEN + 1);
   new_list[driver_count].driver_name[FLASH_DRIVER_MAX_NAME_LEN] = '\0';

   /* Replace the old list with the new list. */
   old_list = driver_list;
   driver_list = new_list;
   driver_count++;

   /* Free the old list, if there was one. */
   if (old_list != NULL)
   {
      assert(driver_count > 1);
      fam_free_mem(old_list);
   }

   return FLASHCONF_EOK;
}


/****************************************************************
*  Function build_access_lock_database
*
*  Builds the access lock database for a Flash instance. The database consists
*  of a number of ModifyLock and ReadLock structures and an array of ReadLock
*  pointers. All structures are initialized and the pointer tree is built.
*/
static void
build_access_lock_database(InstanceControlBlock *instance)
{
   size_t                  smallest_modify_prevents_read_block;
   unsigned                nr_of_read_locks;
   unsigned                nr_of_modify_locks;
   unsigned                nr_of_read_lock_pointers;
   ReadLock                *read_lock_list;
   ModifyLock              *modify_lock_list;
   ReadLock                **read_lock_pointer_list;
   unsigned                read_lock_index;
   unsigned                modify_lock_index;
   size_t                  block_offset;
   FlashDriverResult       result;
   FlashCharacteristicsV1  characteristics;
   size_t                  offset;
   unsigned                i;

   /*
   *  Count the number of FLASH_MODIFY_PREVENTS_READ blocks in the instance
   *  and find the smallest size, by examining each range of such blocks.
   */
   offset = 0;
   smallest_modify_prevents_read_block = instance->instance_length;
   nr_of_read_locks = 0;
   do
   {
      result = instance->driver->v1.get_characteristics(&characteristics,
         instance->driver_data, FLASH_MODIFY_PREVENTS_READ, 
         instance->instance_start + offset);
      if ((result != FLASHDRIVER_EOK) ||
          (characteristics.block_length == 0) ||
          (characteristics.range_start != (instance->instance_start + offset)))
      {
         fam_internal_error(FLASH_EINCONSISTENT_CHARACTERISTICS, result);
      }

      if (characteristics.block_length < smallest_modify_prevents_read_block)
      {
         smallest_modify_prevents_read_block = characteristics.block_length;
      }
      nr_of_read_locks += 
         (characteristics.range_length / characteristics.block_length);
      offset += characteristics.range_length;
   } while (offset < instance->instance_length);

   /*
   *  Count the number of FLASH_MODIFY_PREVENTS_MODIFY blocks in the instance,
   *  by examining each range of such blocks.
   */
   offset = 0;
   nr_of_modify_locks = 0;
   do
   {
      result = instance->driver->v1.get_characteristics(&characteristics,
         instance->driver_data, FLASH_MODIFY_PREVENTS_MODIFY, 
         instance->instance_start + offset);
      if ((result != FLASHDRIVER_EOK) ||
          (characteristics.block_length == 0) ||
          (characteristics.range_start != (instance->instance_start + offset)))
      {
         fam_internal_error(FLASH_EINCONSISTENT_CHARACTERISTICS, result);
      }

      nr_of_modify_locks += 
         (characteristics.range_length / characteristics.block_length);
      offset += characteristics.range_length;
   } while (offset < instance->instance_length);

   /*
   *  Figure out the shift value to find the index of a read lock pointer, and
   *  the number of read lock pointers to use.
   *  This is done by counting trailing zero-bits.
   *  Note: This works best if the block size is a power of two. We get a
   *  smaller shift count and a larger number of read-lock pointers otherwise.
   */ 
   assert(smallest_modify_prevents_read_block != 0);
   for (i = 0; ((smallest_modify_prevents_read_block & (1U << i)) == 0); i++) {}
   instance->read_lock_index_shift = i;
   nr_of_read_lock_pointers = instance->instance_length / (1U << i);

   /*
   *  Allocate buffers to hold three arrays: read lock pointers,
   *  read locks, and modify locks.
   */
   read_lock_pointer_list =
      fam_alloc_mem(nr_of_read_lock_pointers * sizeof (ReadLock *));
   read_lock_list =
      fam_alloc_mem(nr_of_read_locks * sizeof (ReadLock));
   modify_lock_list = 
      fam_alloc_mem(nr_of_modify_locks * sizeof (ModifyLock));

   /*
   *  Initialize all ModifyLocks
   */
   modify_lock_index = 0;
   offset = 0;
   do
   {
      result = instance->driver->v1.get_characteristics(&characteristics,
         instance->driver_data, FLASH_MODIFY_PREVENTS_MODIFY, 
         instance->instance_start + offset);
      if ((result != FLASHDRIVER_EOK) ||
          (characteristics.block_length == 0) ||
          (characteristics.range_start != (instance->instance_start + offset)))
      {
         fam_internal_error(FLASH_EINCONSISTENT_CHARACTERISTICS, result);
      }

      block_offset = 0;
      do
      {
         modify_lock_list[modify_lock_index].base =
            characteristics.range_start + block_offset;
         modify_lock_list[modify_lock_index].length =
            characteristics.block_length;
         modify_lock_list[modify_lock_index].lock_owner = NULL;
         modify_lock_index++;

         block_offset += characteristics.block_length;
      } while (block_offset < characteristics.range_length);

      assert(block_offset == characteristics.range_length);
      offset += characteristics.range_length;
   } while (offset < instance->instance_length);
   assert(modify_lock_index == nr_of_modify_locks);

   /*
   *  Initialize all ReadLocks
   */
   read_lock_index = 0;
   modify_lock_index = 0;  /* The modify lock refered to by the read lock. */
   offset = 0;
   do
   {
      result = instance->driver->v1.get_characteristics(&characteristics,
         instance->driver_data, FLASH_MODIFY_PREVENTS_READ, 
         instance->instance_start + offset);
      if ((result != FLASHDRIVER_EOK) ||
          (characteristics.block_length == 0) ||
          (characteristics.range_start != (instance->instance_start + offset)))
      {
         fam_internal_error(FLASH_EINCONSISTENT_CHARACTERISTICS, result);
      }

      block_offset = 0;
      do
      {
         FlashAddress base = characteristics.range_start + block_offset;

         /*
         *  Check if the ModifyLock at the current modify_lock_index should
         *  be referred to from the current ReadLock. Otherwise we need to go
         *  to the next index.
         */
         if (S32_ge_U32(base - modify_lock_list[modify_lock_index].base,
             modify_lock_list[modify_lock_index].length))
         {
            modify_lock_index++;

            /* The borders of modify lock regions must co-inside with
            *  borders of read lock regions
            */
            if ((modify_lock_index >= nr_of_modify_locks) ||
                (modify_lock_list[modify_lock_index].base != base))
            {
               fam_internal_error(FLASH_EBAD_BLOCK_BORDERS, 0);
            }
         }

         read_lock_list[read_lock_index].base = base;
         read_lock_list[read_lock_index].length = characteristics.block_length;
         read_lock_list[read_lock_index].access_count = 0;
         read_lock_list[read_lock_index].is_direct_access = False;
         read_lock_list[read_lock_index].modify_lock =
            &modify_lock_list[modify_lock_index];
         read_lock_index++;

         block_offset += characteristics.block_length;
      } while (block_offset < characteristics.range_length);

      assert(block_offset == characteristics.range_length);
      offset += characteristics.range_length;
   } while (offset < instance->instance_length);
   assert(modify_lock_index == (nr_of_modify_locks - 1));

   /*
   *  Initialize all ReadLock pointers
   */
   read_lock_index = 0;
   for (i = 0; i < nr_of_read_lock_pointers; i++)
   {
      /*
      *  Check if the ReadLock at the current read_lock_index is accessed via
      *  the pointer that we are about to initialize. Otherwise we need to go
      *  to the next index.
      */
      size_t offset_from_block_start = 
         (size_t)(i * (1 << instance->read_lock_index_shift)) - 
         (size_t)(read_lock_list[read_lock_index].base - 
                  instance->instance_start);
      if (offset_from_block_start >= read_lock_list[read_lock_index].length)
      {
         assert((read_lock_index + 1) < nr_of_read_locks);
         assert(offset_from_block_start ==
                read_lock_list[read_lock_index].length);

         read_lock_index++;
      }

      read_lock_pointer_list[i] = &read_lock_list[read_lock_index];
   }
   assert(read_lock_index == (nr_of_read_locks - 1 ));

   instance->read_lock_pointer_list = read_lock_pointer_list;
}


/****************************************************************
*  Function create_new_instance
*
*  Creates an InstanceControlBlock and initializes it with a new instance.
*
*  RETURNS
*  NULL in case of failure. Otherwise a pointer to a new allocated
*  InstanceControlBlock.
*/
static InstanceControlBlock *
create_new_instance(DriverInfo *driver,
                    FlashAddress instance_start,
                    size_t instance_length,
                    const char *parameter_string,
                    Boolean need_probe_capable)
{
   InstanceControlBlock *new_instance;
   FlashDriverResult result;
   FlashCharacteristicsV1 characteristics;

   /*
   *  Initialize InstanceControlBlock with driver specific information.
   */
   new_instance = fam_alloc_mem(sizeof(InstanceControlBlock));
   new_instance->driver = driver->driver_interface;
   new_instance->if_version = driver->if_version;

   new_instance->driver_capabilities =
      new_instance->driver->v1.get_driver_capabilities();

   new_instance->driver_data = 
      fam_alloc_mem(new_instance->driver_capabilities->instance_databuffer_size);

   /*
   *  Check if a driver with probing support is required.
   */
   if (need_probe_capable && 
       !new_instance->driver_capabilities->device_probing_support)
   {
      goto FAILURE_CLEANUP;
   }

   /*
   *  Instantiate the flash
   */
   result = new_instance->driver->v1.instantiate_device(new_instance->driver_data,
      instance_start, instance_length, parameter_string);

   if (result != FLASHDRIVER_EOK)
   {
      if (result != FLASHDRIVER_EUNSUPPORTED_DEVICE)
      {
         fam_internal_error(FLASH_EUNEXPECTED_DRIVER_ERROR, result);
      }
      goto FAILURE_CLEANUP;   /* Instantiation failed. */
   }

   /*
   *  Get actual instance size.
   */
   result = new_instance->driver->v1.get_characteristics(&characteristics,
      new_instance->driver_data, FLASH_INSTANCE_SIZE, 0);
   new_instance->instance_start = characteristics.range_start;
   new_instance->instance_length = characteristics.range_length;

   if (result != FLASHDRIVER_EOK)
   {
      fam_internal_error(FLASH_EUNEXPECTED_DRIVER_ERROR, result);
   }

   /*
   *  Get instance capabilities.
   */
   new_instance->instance_capabilities =
      new_instance->driver->v1.get_instance_capabilities(
         new_instance->driver_data);

   /*
   *  Build the access lock database.
   */
   build_access_lock_database(new_instance);

   return new_instance;

FAILURE_CLEANUP:
   fam_free_mem(new_instance->driver_data);
   fam_free_mem(new_instance);
   return NULL;
}



/****************************************************************
*  Function flash_add_instance
*
*  Adds a flash device instance to the Flash Access Manager.
*
*  See flash_conf.h for details.
*/
extern size_t
flash_add_instance(const char *driver_name,
                   FlashAddress instance_start,
                   size_t instance_length,
                   const char *parameter_string)
{
   U32 i;
   InstanceControlBlock *new_instance = NULL;
   InstanceIndex *new_list;
   InstanceIndex *old_list;

   /* Check if the address range interferes with an existing instance. */
   for (i = 0; i < fam_global.instance_count; i++)
   {
      if ((instance_start <= (fam_global.instance_list[i].instance_start +
                   (fam_global.instance_list[i].instance_length - 1))) &&
          ((instance_start + (instance_length - 1)) >=
           (fam_global.instance_list[i].instance_start)))
      {
         /*
         *  There is an overlap between the instances.
         *  Unable to provide an informative error code because the function
         *  returns a length value.
         */
         return 0;   
      }
   }

   if (driver_name[0] == '\0')
   {
      /* The requested driver name is an empty string. We shall try to
      *  instantiate with all drivers that support probing.
      */
      for (i = 0; i < driver_count; i++)
      {
         new_instance = create_new_instance(&driver_list[i],
                                            instance_start,
                                            instance_length,
                                            parameter_string,
                                            True);
         if (new_instance != NULL)
         {
            break;
         }
      }
   }
   else
   {
      /* We have a requested driver name.
      *  Search the driver list for the correct driver.
      */
      for (i = 0; i < driver_count; i++)
      {
         if (strcmp(driver_list[i].driver_name, driver_name) == 0)
         {
            /* Found matching driver name. Instantiate it. */
            new_instance = create_new_instance(&driver_list[i],
                                               instance_start,
                                               instance_length,
                                               parameter_string,
                                               False);
            break;
         }
      }
   }

   if (new_instance == NULL)
   {
      /* Failed to instantiate driver. */
      return 0;
   }

   /*
   *  The fam_global.instance_list is an array of InstanceIndex structs, sorted in
   *  reverse. The array is one longer than fam_global.instance_count, in order
   *  to be terminated by an entry of all zeroes.
   *
   *  The following will copy the list to a new buffer, while inserting a new
   *  entry and keeping the list sorted.
   *
   *  Allocate a buffer that is large enough for the new list. It shall be
   *  (fam_global.instance_count + 2) entries long in order to hold both the new
   *  entry and the terminating entry.
   */
   new_list =
      fam_alloc_mem((fam_global.instance_count + 2) * sizeof (InstanceIndex));

   /* copy all old entries with higher start addresses to the new list. */
   for (i = 0; (i < fam_global.instance_count) && 
               ((fam_global.instance_list[i].instance_start) >
                (new_instance->instance_start)); i++)
   {
      memcpy(&new_list[i],
         &fam_global.instance_list[i], sizeof (InstanceIndex));
   }

   /* Create a new InstanceIndex at the correct position. */
   new_list[i].instance_start = new_instance->instance_start;
   new_list[i].instance_length = new_instance->instance_length;
   new_list[i].control_block = new_instance;

   /*
   *  copy all remaining old entries to the new list. 
   *  This includes the end marker from position (instance_count + 1).
   */
   for (; i < (fam_global.instance_count + 1); i++)
   {
      memcpy(&new_list[i + 1],
         &fam_global.instance_list[i], sizeof (InstanceIndex));
   }

   /* replace the old list with the new. */
   old_list = fam_global.instance_list;
   fam_global.instance_list = new_list;
   fam_global.instance_count++;

   /*
   *  Note: The following is not thread safe. There is a potential fault if
   *  another thread is accessing the old fam_global.instance_list at the
   *  same time as we delete it.
   */
   if (fam_global.instance_count > 1)
   {
      /* There was an old list that must be deleted. */
      fam_free_mem(old_list);
   }

   return (new_instance->instance_length);
}



/****************************************************************
*  Function flash_add_direct_access_region
*
*  Configures the FAM to provide flash synchronization for direct flash
*  access in a region of flash memory. This means that it will be possible
*  to simultaneously execute code in that region and erase or write to the
*  flash.
*
*  It is only possible to do this on a flash instance that have aready been
*  added to FAM.
*
*  See flash_conf.h for details.
*/
extern FlashConfResult
flash_add_direct_access_region(FlashAddress region_start,
                               size_t region_length)
{
   InstanceControlBlock *instance;
   ReadLock *read_lock;
   size_t length_in_lock;

   /*
   *  This outer loop is needed in order to allow an operation to pass
   *  the borders of a ReadLock. Instance lookup is also placed inside
   *  the loop to allow for passing of instance borders with negligible
   *  overhead.
   */
   while (region_length > 0)
   {
      /* Find the instance of the flash address. */
      instance = fam_find_instance(region_start, FAM_INSTANCE_NONE);
      if (instance == NULL)
      {
         return FLASHCONF_ENO_SUCH_INSTANCE;
      }

      /* Find the ReadLock associated with the flash address. */
      read_lock = instance->read_lock_pointer_list[
         ((size_t)(region_start - instance->instance_start)) >>
         (instance->read_lock_index_shift)];

      assert(read_lock->base <= region_start);
      assert(((size_t)(region_start - read_lock->base)) < read_lock->length);

      /* Mark the ReadLock for direct access. */
      read_lock->is_direct_access = True;

      /* Figure out how much of the region_length is inside the lock region. */
      length_in_lock = read_lock->length -
         (size_t)(region_start - read_lock->base);
      if (region_length < length_in_lock)
      {
         length_in_lock = region_length;
      }

      region_start += length_in_lock;
      region_length -= length_in_lock;
   }

   return FLASHCONF_EOK;
}


/****************************************************************
*  Function flash_uninstall
*
*  Uninstalls the flash access module and all resources.
*  Only intended for test purposes.
*/
#ifdef FAM_TEST
extern FlashConfResult
flash_uninstall(void)
{
   U32 i;
   long bios_result;

   /* Remove all flash instances. */
   for (i = 0; i < fam_global.instance_count; i++)
   {
      /* Remove all buffers allocated by an instance. */
      fam_free_mem(fam_global.instance_list[i].control_block->driver_data);
      fam_free_mem(fam_global.instance_list[i].control_block->
         read_lock_pointer_list[0]->modify_lock);
      fam_free_mem(fam_global.instance_list[i].control_block->
         read_lock_pointer_list[0]);
      fam_free_mem(fam_global.instance_list[i].control_block);
   }

   /* Remove the instance list. */
   if (fam_global.instance_count > 0)
   {
      fam_free_mem(fam_global.instance_list);
   }
   fam_global.instance_count = 0;
   fam_global.instance_list = &empty_list_end_entry;

   /* Remove list of installed drivers. */
   fam_free_mem(driver_list);
   driver_count = 0;
   driver_list = NULL;

   /* Remove the list of pre-allocated access locks. */
   fam_free_mem(free_lock_buffer);
   fam_global.free_lock_list = NULL;
   free_lock_buffer = NULL;

   /* Remove the FAM phantom process. */
   kill_proc(fam_global.common_pid);
   fam_global.common_pid = 0;

   /* Uninstall the bios module */
   is_api_available = False;
   bios_result = biosUninstall(FLASH_BIOSMODULE_NAME);
   if (bios_result != 1)
   {
      return (FlashConfResult)-1;   /* No need for a proper error code */
   }

   return FLASHCONF_EOK;
}
#endif /* FAM_TEST */
