/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2006 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */
/****************************************************************
* Description: Declarations shared between parts of FAM.
*
*/


#include "fam_whitebox_test.h"


/****************************************************************
*  Macros FAM_DISABLE_INTERRUPT and FAM_ENABLE_INTERRUPT
*
*  Wrappers for LOCK_SAVE and LOCK_RESTORE that additionally provides
*  whitebox test hooks without cluttering the code.
*/
#define FAM_DISABLE_INTERRUPT(int_mask_save) \
   { WHITEBOX_TESTPOINT(TPT_SYNCH_INT_DISABLE); LOCK_SAVE(int_mask_save); }

#define FAM_ENABLE_INTERRUPT(int_mask_save) \
   { LOCK_RESTORE(int_mask_save); WHITEBOX_TESTPOINT(TPT_SYNCH_INT_ENABLE); }


/****************************************************************
*  Helper macros for FAM throttle
*/
#define IS_FAM_THROTTLE_ACTIVATED()          \
   (fam_global.throttle_sem != NULL)

#define ENTER_FAM_THROTTLE()                 \
   do                                        \
   {                                         \
      assert(IS_FAM_THROTTLE_ACTIVATED());   \
      WHITEBOX_TESTPOINT(TPT_SYNCH_SYSCALL); \
      wait_sem(fam_global.throttle_sem);     \
      WHITEBOX_TESTPOINT(TPT_SYNCH_SYSCALL); \
   }                                         \
   while (0)

#define LEAVE_FAM_THROTTLE()                 \
   do                                        \
   {                                         \
      assert(IS_FAM_THROTTLE_ACTIVATED());   \
      signal_sem(fam_global.throttle_sem);   \
      assert(get_sem(fam_global.throttle_sem) <= fam_global.throttle_limit); \
      WHITEBOX_TESTPOINT(TPT_SYNCH_SYSCALL); \
   }                                         \
   while (0)


/****************************************************************
*  Sanity checking macros for FAM throttle
*/

/* These are helpers for the following macros. */
#ifdef FAM_THROTTLE_TEST

   #define FT_CHECK_DEC_MODIFIERS_CNT()    fam_throttle_check_limit(-1)
   #define FT_CHECK_INC_MODIFIERS_CNT()    fam_throttle_check_limit(+1)

#else

   #define FT_CHECK_DEC_MODIFIERS_CNT()    (void)0
   #define FT_CHECK_INC_MODIFIERS_CNT()    (void)0

#endif

/* These should be used before a modifying operation starts or
 * after it has stopped.
 */
#define FT_CHECK_WILL_START_WRITE()    FT_CHECK_INC_MODIFIERS_CNT()
#define FT_CHECK_WILL_START_ERASE()    FT_CHECK_INC_MODIFIERS_CNT()
#define FT_CHECK_DID_STOP_WRITE()      FT_CHECK_DEC_MODIFIERS_CNT()
#define FT_CHECK_DID_STOP_ERASE()      FT_CHECK_DEC_MODIFIERS_CNT()


/****************************************************************
*  enum ActivityType
*
*  Describes the activity, with priority, associated with an access lock.
*/
typedef enum ActivityType
{
   AT_ERASE,            /* Low priority */
   AT_SUSPENDED_ERASE,
   AT_WRITE,
   AT_SUSPENDED_WRITE,
   AT_READ              /* High priority */
} ActivityType;



/****************************************************************
*  enum SigWaitState
*
*  Describes whether the holder of an access lock expects a signal, and
*  what type of signal.
*/
typedef enum SigWaitState
{
   SWS_NOT_EXPECTING_SIGNAL,
   SWS_EXPECTING_TRIG_SIGNAL,
   SWS_TRIG_SIGNAL_SENT,
   SWS_EXPECTING_NOTIFY_SIGNAL
} SigWaitState;


/****************************************************************
*  struct AccessLock
*
*  Provides an association between the process that holds an access lock
*  and the modify and read locks at the accessed address. Provides the
*  ability to queue for a lock.
*
*  Dynamically allocated to a process that performs an access or wants to
*  perform an access.
*
*  Modify accesses hold one of these for the duration of the access. Read
*  accesses only use this if it needs to queue for access rights.
*
*  FIELD NAME           FIELD DESCRIPTION
*  ==========           =================
*
*  next_queue           Pointer to the next access lock in queue. NULL if
*                       there ara no more locks in queue. A lock in queue
*                       represents an operation that has not yet started
*                       or has been suspended. The queue is sorted with
*                       the highest priority operation first.
*
*  activity_type        Describes the activity, with priority, associated
*                       with the access lock.
*
*  associated_read_lock Pointer to the read lock that represents the modify-
*                       prevents-read block that is accessed by the operation.
*
*  owner_pid            The PID of the process that holds the lock. Needed
*                       for sending trig and notify signals. Sometimes 0
*                       when the owner does not expect a signal.
*
*  owner_prio           The priority of the process that holds the lock.
*                       Fam_acquire_read_lock will temporarily raise the prio
*                       of the owner process to its own level in order to 
*                       allow it to reach the state where it releases itself
*                       from the lock and notifies the next.       
*
*  prio_raised          A flag that indicates this process was prio raised. It is
*                       checked at release modify lock and after removing
*                       itself as a lock owner in the fam_wait_during_suspend call. 
*
*  sig_wait_state       Describes whether the holder of the lock expects a
*                       signal, and what type of signal.
*
*  params               Parameter block to be used when accessing the driver.
*                       This is not related to the lock. It is kept here for
*                       convenience because it is not wanted on the stack.
*/
typedef struct AccessLock
{
   struct AccessLock             *next_queue;
   ActivityType                  activity_type;
   struct ReadLock               *associated_read_lock;
   PROCESS                       owner_pid;
   OSPRIORITY                    owner_prio;
   Boolean                       prio_raised;
   SigWaitState                  sig_wait_state;
   FlashOperationParametersV1    params;
} AccessLock;


/****************************************************************
*  struct ModifyLock
*
*  Represents a modify-prevents-modify block in a flash instance.
*
*  Statically allocated as part of the access lock database for an instance.
*
*  FIELD NAME           FIELD DESCRIPTION
*  ==========           =================
*
*  lock_owner           Points to the head of the queue of access locks that
*                       want to perform operations in the associated
*                       modify-prevents-modify block. NULL when the queue is
*                       empty.
*
*  base                 The first flash address of block.
*
*  length               The number of bytes in the block.
*/
typedef struct ModifyLock
{
   AccessLock * volatile   lock_owner;
   FlashAddress            base;
   size_t                  length;
} ModifyLock;


/****************************************************************
*  struct ReadLock
*
*  Represents a modify-prevents-read block in a flash instance.
*
*  Statically allocated as part of the access lock database for an instance.
*
*  FIELD NAME           FIELD DESCRIPTION
*  ==========           =================
*
*  modify_lock          Points to the ModifyLock that is associated with the
*                       address range of this block. Note that several ReadLock
*                       structs can refer to the same ModifyLock.
*
*  access_count         Indicates what type and how many accesses are currently
*                       active in this block. Positive for read accesses,
*                       negative for modify access. A suspended access is not
*                       shown here.
*
*  is_direct_access     True if the flash range controled by this lock is
*                       configured for direct access. False otherwise. Boolean
*                       value stored in U16 to save space.
*
*  base                 The first flash address of block.
*
*  length               The number of bytes in the block.
*/
typedef struct ReadLock
{
   ModifyLock        *modify_lock;
   S16               access_count;
   U16               is_direct_access;
   FlashAddress      base;
   size_t            length;
} ReadLock;


/****************************************************************
*  struct InstanceControlBlock
*
*  Information about a flash instance.
*
*  FIELD NAME              FIELD DESCRIPTION
*  ==========              =================
*
*  instance_start          First flash address of the instance.
*
*  instance_length         Number of bytes in the instance.
*
*  driver                  Pointer to struct of driver function pointers.
*
*  driver_capabilities     Pointer to driver capabilities struct.
*
*  instance_capabilities   Pointer to instance capabilities struct.
*
*  driver_data             Pointer to instance buffer used by driver.
*
*  read_lock_index_shift   Bit shift count used to convert an offset from
*                          the instance start to an index in the array of
*                          read lock pointers. 
*
*  read_lock_pointer_list  Pointer to array of read lock pointers. The array
*                          is for quickly finding a read lock. 
*
*  if_ver                  Which version of the interface that has been
*                          negotiated with the instance.
*/
typedef struct InstanceControlBlock
{
   FlashAddress                        instance_start;
   size_t                              instance_length;
   const union FlashDriverInterface   *driver;
   const FlashDriverCapabilitiesV1    *driver_capabilities;
   const FlashInstanceCapabilitiesV1  *instance_capabilities;
   void                               *driver_data;
   unsigned                            read_lock_index_shift;
   ReadLock                          **read_lock_pointer_list;
   unsigned                            if_version;
} InstanceControlBlock;


/****************************************************************
*  struct InstanceIndex
*
*  Describes an instance for the purpose of finding an InstanceControlBlock.
*
*  An array of these structs forms a list of installed flash instances.
*  The array is sorted in descending (reverse) instance start order.
*  The end of the list is an entry with all zeroes, needed for a faster
*  linear search of the list.
*/
typedef struct InstanceIndex
{
   FlashAddress         instance_start;
   size_t               instance_length;
   InstanceControlBlock *control_block;
} InstanceIndex;


/****************************************************************
*  GLOBAL VARIABLES
*****************************************************************
*/
typedef struct FamGlobal
{
   /* Instance list. See description of struct InstanceIndex for details. */
   U32              instance_count;
   InstanceIndex    *instance_list;

   /* common_pid. Used as sender of all FAM-internal signals. */
   PROCESS          common_pid;

   /* free_lock_list. Pointer to linked list of pre-allocated AccessLocks.*/
   AccessLock       *free_lock_list;

   /* throttle_sem. The semaphore used for throttling. It is non-NULL exactly
    * when throttling is activated.
    */
   SEMAPHORE          *throttle_sem;

   /* throttle_limit. Maximum number of simultaneous write/erase operations
    * allowed.
    */
   int                 throttle_limit;

#ifdef FAM_THROTTLE_TEST
   /* throttle_modifiers_count. Number of simultaneous write/erase operations
    * currently in progress in the flash(es).
    */
   int                 throttle_modifiers_count;

   /* throttle_max_modifiers_count. Largest number of simultaneous write/erase
    * operations in progress in the flash(es) at any time. (I.e. the largest
    * value of throttle_modifiers_count encountered so far.)
    */
   int                 throttle_max_modifiers_count;
#endif /* FAM_THROTTLE_TEST */

   /* irq_poll_function. Pointer to installed irq polling function. */
   int (*irq_poll_function)(void);

   /* micro_delay_timer. Pointer to installed micro delay timer function. */
   void (*micro_delay_timer)(U32 microseconds);
} FamGlobal;

extern FamGlobal fam_global;

#ifdef FAM_THROTTLE_TEST
/*
 * When using fam in a test setting, testing that fam_global_size ==
 * sizeof(fam_global) will (1) fail at link time, if the FAM library wasn't
 * built with FAM_THROTTLE_TEST defined, and (2) fail at run-time if the FAM
 * library is completely incompatible with the test. (The reverse is not true:
 * if the test succeeds, the layout and/or semantics of FamGlobal may still be
 * incompatible with what the test code expects.)
 */
extern const size_t fam_global_size;
#endif /* FAM_THROTTLE_TEST */


/****************************************************************
*  FUNCTION PROTOTYPES
*****************************************************************
*/
long fam_biosfunction(long function_code, long arg2, long arg3, long arg4,
                      long arg5, long arg6, long arg7);
void fam_internal_error(OSERRCODE ecode, OSERRCODE extra);
InstanceControlBlock *fam_find_instance(const FlashAddress request_address,
                                        U32 suggested_instance);
void fam_acquire_read_lock(ReadLock *read_lock);
AccessLock * fam_acquire_modify_lock(ReadLock *read_lock,
                                     ActivityType activity_type);
void fam_release_read_lock(ReadLock *read_lock);
void fam_release_modify_lock(AccessLock *access_lock);
Boolean fam_poll_suspend_request(AccessLock *access_lock);
Boolean fam_wait_suspend_request(AccessLock *access_lock, U32 timeout_ms);
void fam_wait_during_suspend(AccessLock *access_lock);

#ifdef FAM_THROTTLE_TEST
void fam_throttle_check_limit(int change);
#endif /* FAM_THROTTLE_TEST */
