/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2009 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */


/****************************************************************
* $workfile: <file name> $
* Document no: @(#) $FileRevision:   /main/tb_current_ose5/5  $
* $date: $
*
* Purpose: C language program file
* Description: Implementation of flash access module.
*
*/


#include "flash_conf.h"
#include "flash_biosinterface.h"
#include "flash_err.h"
#include "ose.h"
#include "cpu.h"
#include "assert.h"
#include "string.h"
#include "fam_internal.h"



/****************************************************************
*  Function fam_internal_error
*
*  Report an internal error.
*
*  RETURNS
*  The function does not return.
*/
void
fam_internal_error(OSERRCODE ecode, OSERRCODE extra)
{
   WHITEBOX_TESTPOINT(TPT_ERROR);

   /* Give debugger first chanse to find error. */
   assert(0);

   /* Report error to OSE. */
   error2(ecode, extra);

   /* Make sure we do not return. (It is undefined whether error2 returns.) */
   kill_proc(current_process());   
}
#ifndef WHITEBOX_TEST
/* For release build, inline the error call. */
#define fam_internal_error(ecode, extra) \
do \
{ \
   error2(ecode, extra); \
   kill_proc(current_process()); \
} \
while (0)
#endif



/****************************************************************
*  Function fam_throttle_check_limit
*
*  Updates the current number of modifiers (i.e. non-suspended write/erase
*  operations) and checks that it's within the throttle limit (if any).
*
*  RETURNS
*  Nothing.
*
*  Implementation notes:
*  May be called with interrups disabled. We use LOCK_SAVE/LOCK_RESTORE
*  rather than FAM_DISABLE_INTERRUPT/FAM_ENABLE_INTERRUPT, as we don't
*  want to get any WHITEBOX_TESTPOINTs here.
*/
#ifdef FAM_THROTTLE_TEST
void fam_throttle_check_limit(int change)
{
   U32 int_mask_save = 0;  /* Initialized to avoid compiler warning. */
   
   LOCK_SAVE(int_mask_save);

   /* Update the global count. */
   fam_global.throttle_modifiers_count += change;

   /* Check if this is a new max. */
   if (fam_global.throttle_max_modifiers_count <
       fam_global.throttle_modifiers_count)
   {
      fam_global.throttle_max_modifiers_count =
         fam_global.throttle_modifiers_count;
   }
   
   /* Check that values are within limits. */
   if (IS_FAM_THROTTLE_ACTIVATED())
   {
      assert(0 <= fam_global.throttle_modifiers_count);
      assert(fam_global.throttle_modifiers_count <= fam_global.throttle_limit);
   }

   LOCK_RESTORE(int_mask_save);
}
#endif /* FAM_THROTTLE_TEST */



/****************************************************************
*  Function fam_find_instance
*
*  Search for a matching flash instance.
*
*  RETURNS
*  NULL if no match was found. Otherwise a pointer to InstanceControlBlock.
*
*  Implementation notes:
*  The algorithm is linear search. It is chosen because it is very efficient
*  for short lists.
*/
InstanceControlBlock *
fam_find_instance(const FlashAddress request_address,
                  U32 suggested_instance)
{
   InstanceIndex *instance;

   /* Check if explicit instance number requested */
   if (suggested_instance != FAM_INSTANCE_NONE)
   {
      if (suggested_instance >= fam_global.instance_count)
      {
         return NULL;
      }

      return fam_global.instance_list[suggested_instance].control_block;
   }

   /* Search the list until we find an instance that starts
   *  before our request address.
   */
   instance = fam_global.instance_list;
   while (request_address < instance->instance_start)
   {
      instance++;
   }

   /*
   *  Check that the request address is inside the range of the instance.
   */
   if (((size_t)(request_address - instance->instance_start)) >=
       instance->instance_length)
   {
      return NULL;
   }

   WHITEBOX_CALLBACK(TPT_INSTANCE, instance->control_block->instance_start,
      instance->control_block->instance_length);

   return instance->control_block;
}


/****************************************************************
*  Function find_instance_bin
*
*  Search for a matching flash instance.
*
*  RETURNS
*  NULL if no match was found. Otherwise a pointer to InstanceControlBlock.
*
*  Implementation notes:
*  This version uses a binary search, that starts on the most recently used
*  entry.
*
*  An analysis has shown that this search is more efficient on average for
*  random searches than the linear search for lists that are more than 15
*  elements long. Such large numbers of flash instances are not forseen.
*  Hence this algorithm is not used. The code is not tested.
*/
#if 0 /* This is not used, see comment above. */
static InstanceControlBlock *
find_instance_bin(const FlashAddress request_address)
{
   static U32     most_recent_instance_index = 0;
   InstanceIndex  *instance;
   U32            section_start_index;
   U32            section_length;
   U32            instance_index;

   /*
   *  Perform a binary search, starting at the index position of the most
   *  recently used instance.
   */
   section_start_index = 0;
   section_length = fam_global.instance_count;
   instance_index = most_recent_instance_index;
   do
   {
      assert(section_start_index < fam_global.instance_count);
      assert(section_length <=
         (fam_global.instance_count - section_start_index));
      assert(section_length > 0);
      assert(instance_index >= section_start_index);
      assert(instance_index < (section_start_index + section_length));

      instance = &fam_global.instance_list[instance_index];

      /* Check if the request address is below this instance. */
      if (request_address < instance->instance_start)
      {
         /*
         *  The request address is at a lower address than this instance.
         *  This translates to the list half above the current since the
         *  list is sorted in reverse.
         */
         section_start_index = instance_index + 1;
         section_length = section_length - (instance_index + 1);
      }

      /*
      *  Check if the request address is inside the range of the instance.
      */
      else if (((size_t)(request_address - instance->instance_start)) <
               instance->instance_length)
      {
         /*
         *  We have found an instance that matches request_address.
         */
         most_recent_instance_index = instance_index;
         WHITEBOX_CALLBACK(TPT_INSTANCE,
            instance->control_block->instance_start,
            instance->control_block->instance_length);
         return instance->control_block;
      }
      else
      {
         /* The request address is at an address above this instance. */
         section_length = instance_index - section_start_index;
      }

      /*
      *  Get index of element to examine in next iteration.
      *  It is the middle element of the remaining section.
      */
      instance_index = section_start_index + (section_length / 2);
   } while (section_length > 0);

   /* We did not find a match */
   return NULL;
}
#endif



/****************************************************************
*  Function fam_flash_read2
*
*  Implements the bios call for flash_read2. See "flash_api.h" for details.
*/
static FlashResult
fam_flash_read2(void *dest_buf, FlashAddress src_flash_addr, size_t length,
                U32 instance_nr, FlashAccessType access_type)
{
   InstanceControlBlock *instance;
   ReadLock *read_lock;
   size_t length_in_lock;
   FlashDriverResult result = FLASHDRIVER_EOK;

   /*
   *  This outer loop is needed in order to allow an operation to pass
   *  the borders of a ReadLock. Instance lookup is also placed inside
   *  the loop to allow for passing of instance borders with negligible
   *  overhead.
   */
   while (length > 0)
   {
      /* Find the instance of the flash address. */
      instance = fam_find_instance(src_flash_addr, instance_nr);
      if (instance == NULL)
      {
         return FLASH_EINVALID_ADDRESS;
      }

      /* Find the ReadLock associated with the flash address. */
      read_lock = instance->read_lock_pointer_list[
         ((size_t)(src_flash_addr - instance->instance_start)) >>
         (instance->read_lock_index_shift)];

      assert(read_lock->base <= src_flash_addr);
      assert(((size_t)(src_flash_addr - read_lock->base)) < read_lock->length);
      WHITEBOX_CALLBACK(TPT_READ_LOCK, read_lock->base, read_lock->length);
      WHITEBOX_CALLBACK(TPT_MODIFY_LOCK, read_lock->modify_lock->base,
         read_lock->modify_lock->length);

      /* Get read access permissions for the lock region */
      fam_acquire_read_lock(read_lock);

      /*
      *  Figure out how much of the access length is inside the lock region.
      *  First calculate the remaining length from the current position to
      *  the end of the lock region. The remainig length in the region is
      *  the smaller of this and the total remaining length.
      */
      length_in_lock = read_lock->length -
         (size_t)(src_flash_addr - read_lock->base);
      if (length < length_in_lock)
      {
         length_in_lock = length;
      }

      /* Do the read. */
      WHITEBOX_CALLBACK(TPT_SYNCH_READ_ACCESS, src_flash_addr, length_in_lock);
      switch (instance->if_version)
      {
         case 1:
            (void)memcpy(dest_buf, src_flash_addr, length_in_lock);            
            break;

         case 2:
            result = instance->driver->v2.read(instance->driver_data,
                                               dest_buf,
                                               src_flash_addr,
                                               length_in_lock,
                                               access_type);
            break;
         default:
            assert(0);
      }

      /* Count the progress of the read. */
      src_flash_addr += length_in_lock;
      dest_buf = ((char *)dest_buf) + length_in_lock;
      length -= length_in_lock;

      /* Release the read access permissions for the lock region */
      fam_release_read_lock(read_lock);

      /* Check if read failed, if so, bail out. */
      if (result != FLASHDRIVER_EOK)
      {
         break;
      }
   }

   return (result == FLASHDRIVER_EOK)?FLASH_EOK:FLASH_EHARDWARE_ERROR;
}


/****************************************************************
*  Function fam_flash_write2
*
*  Implements the bios call for flash_write2. See "flash_api.h" for details.
*/
static FlashResult
fam_flash_write2(FlashAddress dest_flash_addr,
                 const void *src_buf,
                 size_t length,
                 U32 instance_nr,
                 FlashAccessType access_type)
{
   InstanceControlBlock *instance;
   FlashDriverResult result;
   ReadLock *read_lock;
   AccessLock *access_lock;
   size_t length_in_lock;
   U32 int_mask_save = 0;       /* Initiate to zero to keep compiler happy */
   unsigned attempt_count;

   /*
   *  This outer loop is needed in order to allow an operation to pass the
   *  borders of a ReadLock. Instance lookup is also placed inside the loop
   *  to allow passing of instance borders with negligible overhead.
   */
   while (length > 0)
   {
      /* Find the instance of the flash address. */
      instance = fam_find_instance(dest_flash_addr, instance_nr);
      if (instance == NULL)
      {
         return FLASH_EINVALID_ADDRESS;
      }

      /* Find the ReadLock associated with the flash address. */
      read_lock = instance->read_lock_pointer_list[
         ((size_t)(dest_flash_addr - instance->instance_start)) >>
         (instance->read_lock_index_shift)];

      assert(read_lock->base <= dest_flash_addr);
      assert(((size_t)(dest_flash_addr - read_lock->base)) < read_lock->length);
      WHITEBOX_CALLBACK(TPT_READ_LOCK, read_lock->base, read_lock->length);
      WHITEBOX_CALLBACK(TPT_MODIFY_LOCK, read_lock->modify_lock->base,
         read_lock->modify_lock->length);

      /*
      *  Figure out how much of the access length is inside the lock region.
      *  First calculate the remaining length from the current position to
      *  the end of the lock region. The remainig length in the region is
      *  the smaller of this and the total remaining length.
      */
      length_in_lock = 
         read_lock->length - (size_t)(dest_flash_addr - read_lock->base);
      if (length < length_in_lock)
      {
         length_in_lock = length;
      }

      /* Get modify access permissions for the lock region */
      access_lock = fam_acquire_modify_lock(read_lock, AT_WRITE);

      /*
      *  This loop makes several write attempts if there is a write failure
      *  and the driver has requested read back.
      *  Note that there will not be multiple attempts if the driver reports
      *  an error.
      */
      attempt_count = instance->instance_capabilities->verify_write;
      for (;;)
      {
         /*
         *  Enable the interrupt polling algorithm for direct access.
         */
         if (read_lock->is_direct_access)
         {
            FAM_DISABLE_INTERRUPT(int_mask_save);
         }

         /* Start the write operation. */
         access_lock->params.destination = dest_flash_addr;
         access_lock->params.source_data = src_buf;
         access_lock->params.request_length = length_in_lock;
         /* New in version 2 of the interface */
         access_lock->params.access_type = access_type;
         FT_CHECK_WILL_START_WRITE();
         result = instance->driver->v1.write_start(instance->driver_data, 
                                                &access_lock->params);
         if (result != FLASHDRIVER_EOK)
         {
            FT_CHECK_DID_STOP_WRITE();
            goto FAILURE_CLEANUP_1;
         }

         /*
         *  Select a polling algorithm to poll the driver until the operation
         *  is completed.
         */
         if (read_lock->is_direct_access)
         {
            /*
            *  Polling loop for access of flash that is configured for direct
            *  access. This poll loop operates with disabled interrupts and
            *  polls for interrupt request during the operation.
            */
            do
            {
               /* 
               *  Check for interrupt request.
               *  This is only done if the driver supports suspend.
               */
               if (instance->instance_capabilities->suspend_write_support &&
                   (fam_global.irq_poll_function() !=
                    IRQ_POLL_NO_PENDING_INTERRUPT))
               {
                  /* Interrupt request detected. Suspend the operation. */
                  result = instance->driver->v1.write_suspend(instance->driver_data,
                                                           &access_lock->params);
                  FT_CHECK_DID_STOP_WRITE();
                  if (result != FLASHDRIVER_EOK)
                  {
                     goto FAILURE_CLEANUP_1;
                  }
                  FAM_ENABLE_INTERRUPT(int_mask_save);

                  /*
                  *  The interrupt will be serviced immediately after enable.
                  *  Check if there has been a suspend request during the
                  *  service of the interrupt.
                  */
                  if (fam_poll_suspend_request(access_lock))
                  {
                     fam_wait_during_suspend(access_lock);
                  }

                  /* Continue the operation with interrupts disabled. */
                  FAM_DISABLE_INTERRUPT(int_mask_save);
                  FT_CHECK_WILL_START_WRITE();
                  result = instance->driver->v1.write_resume(instance->driver_data,
                                                          &access_lock->params);
                  if (result != FLASHDRIVER_EOK)
                  {
                     FT_CHECK_DID_STOP_WRITE();
                     goto FAILURE_CLEANUP_1;
                  }
               }

               /* Poll the driver to check if the operation is completed. */
               result = instance->driver->v1.write_poll(instance->driver_data, 
                                                     &access_lock->params);
            } while (result == FLASHDRIVER_EIN_PROGRESS);
            FT_CHECK_DID_STOP_WRITE();
            FAM_ENABLE_INTERRUPT(int_mask_save);
         }
         else
         {
            /*
            *  Polling loop for access of flash that is NOT configured for
            *  direct access. This poll loop checks for a suspend request
            *  each loop turn.
            */
            do
            {
               /*
               *  Do a micro-delay before each poll.
               *  The micro-delay (if installed) allows low priority processes
               *  to execute during the delay interval.
               */
               fam_global.micro_delay_timer(access_lock->params.delay_time);

               /* 
               *  Check for suspend request.
               *  This is only done if the driver supports suspend.
               */
               if (instance->instance_capabilities->suspend_write_support &&
                   fam_poll_suspend_request(access_lock))
               {
                  /* Suspend request detected. Suspend the operation. */
                  result = instance->driver->v1.write_suspend(instance->driver_data,
                                                           &access_lock->params);
                  FT_CHECK_DID_STOP_WRITE();
                  if (result != FLASHDRIVER_EOK)
                  {
                     goto FAILURE_CLEANUP_1;
                  }

                  /* Allow the process that requested suspend to execute. */
                  fam_wait_during_suspend(access_lock);

                  /* Resume the operation. */
                  FT_CHECK_WILL_START_WRITE();
                  result = instance->driver->v1.write_resume(instance->driver_data,
                                                          &access_lock->params);
                  if (result != FLASHDRIVER_EOK)
                  {
                     FT_CHECK_DID_STOP_WRITE();
                     goto FAILURE_CLEANUP_1;
                  }
               }

               /* Poll the driver to check if the operation is completed. */
               result = instance->driver->v1.write_poll(instance->driver_data, 
                                                     &access_lock->params);
            } while (result == FLASHDRIVER_EIN_PROGRESS);
            FT_CHECK_DID_STOP_WRITE();
         }

         /* Check the result from write_poll for errors. */
         if (result != FLASHDRIVER_EOK)
         {
            goto FAILURE_CLEANUP_2;
         }

         /* If no read back is to be done, we're finished */
         if (attempt_count == 0)
         {
            break;
         }

         /* Read back the written data and compare with desired value. */
         if (instance->if_version == 1)
         {
            if (memcmp(dest_flash_addr, src_buf, length_in_lock) == 0)
            {
               /* The data matched. */
               break;   /* Out of for loop */
            }
         }
         else if (instance->if_version == 2)
         {
            size_t nr_matching_bytes = 0;
            
            result = instance->driver->v2.check_content(
                  instance->driver_data,
                  src_buf,
                  dest_flash_addr,
                  length_in_lock,
                  &nr_matching_bytes,
                  access_type);
            
            if (result != FLASHDRIVER_EOK)
            {
               goto FAILURE_CLEANUP_3;
            }
            
            if (nr_matching_bytes == length_in_lock)
            {
               /* The data matched */
               break; /* out of for loop */
            }
         }
         else
         {
            /* unknown interface version */
            assert(0);
         }   

         /* The written data did not match the source data. */
         attempt_count--;
         if (attempt_count == 0)
         {
            goto FAILURE_CLEANUP_3;
         }
      } /* End of for (;;) */

      /* Count the progress of the write. */
      dest_flash_addr += length_in_lock;
      src_buf = ((const char *)src_buf) + length_in_lock;
      assert(length_in_lock <= length);
      length -= length_in_lock;

      /* Release the modify access permissions for the lock region */
      fam_release_modify_lock(access_lock);
   } /* end of while (length > 0) */

   return FLASH_EOK;


FAILURE_CLEANUP_1:
   if (read_lock->is_direct_access)
   {
      FAM_ENABLE_INTERRUPT(int_mask_save);
   }
   /* Fall through to next label. */

FAILURE_CLEANUP_2:
   if (result != FLASHDRIVER_EHARDWARE_ERROR &&
       result != FLASHDRIVER_EINVALID_DATA &&
       result != FLASHDRIVER_EPROTECTED)
   {
      fam_release_modify_lock(access_lock);
      fam_internal_error(FLASH_EUNEXPECTED_DRIVER_ERROR, (OSERRCODE)result);
   }
   /* Fall through to next label. */

FAILURE_CLEANUP_3:
   /* Release the modify access permissions for the lock region */
   fam_release_modify_lock(access_lock);

   /*
   return result == FLASHDRIVER_EHARDWARE_ERROR ?
	  FLASH_EHARDWARE_ERROR : FLASH_EINVALID_DATA;
   */
   
   if (result == FLASHDRIVER_EHARDWARE_ERROR)
     {
       return FLASH_EHARDWARE_ERROR;
     }
   else if (result == FLASHDRIVER_EPROTECTED)
     {
       return FLASH_EPROTECTED;
     }
   else
     {
       return FLASH_EINVALID_DATA;
     }
}


/****************************************************************
*  Function readback_erasecheck
*
*  Checks if a memory region is erased.
*
*  RETURNS
*  True if the memory area was erased. False otherwise.
*/
static Boolean
readback_erasecheck(FlashAddress eraseblock_addr, size_t length)
{
   U32 *addr = (U32 *)eraseblock_addr;
   size_t i;

   for (i = length / sizeof(U32); i != 0; i--)
   {
      if (*addr != (~0UL))
      {
         return False;
      }
      addr++;
   }
   return True;
}


/****************************************************************
*  Function fam_flash_erase
*
*  Implements the bios call for flash_erase. See "flash_api.h" for details.
*/
static FlashResult
fam_flash_erase(FlashAddress eraseblock_addr, size_t length)
{
   InstanceControlBlock *instance;
   FlashDriverResult result;
   ReadLock *read_lock;
   AccessLock *access_lock;
   FlashCharacteristicsV1 inst_characteristics;
   U32 int_mask_save = 0;       /* Initiate to zero to keep compiler happy. */
   unsigned attempt_count;

   /*
   *  This loop erases one eraseblock per turn.
   *
   *  Lookup of instance, eraseblock size and such, together with management
   *  of access locks is done every loop turn in order to correctly cross
   *  borders between such entities. The overhead time is negligible compared
   *  to the time for erase.
   */
   while (length > 0)
   {
      /*
      *  Get the instance of the flash address.
      */
      instance = fam_find_instance(eraseblock_addr, FAM_INSTANCE_NONE);
      if (instance == NULL)
      {
         return FLASH_EINVALID_ADDRESS;
      }

      /*
      *  Get eraseblock characteristics.
      *  We get this from the driver every time in order to avoid keeping it
      *  in an internal structure. 
      */
      result = instance->driver->v1.get_characteristics(&inst_characteristics,
                                                     instance->driver_data,
                                                     FLASH_ERASEBLOCK_RANGE,
                                                     eraseblock_addr);
      /* Check for errors. */
      if (result != FLASHDRIVER_EOK)
      {
         fam_internal_error(FLASH_EUNEXPECTED_DRIVER_ERROR, (OSERRCODE)result);
      }

      /* Verify that we are at an eraseblock boundary. */
      if ((((size_t)(eraseblock_addr - inst_characteristics.range_start)) %
           inst_characteristics.block_length) != 0)
      {
         return FLASH_EUNMATCHED_ERASEBLOCK;
      }

      /* Verify that we are not trying to erase less than one eraseblock. */
      if (length < inst_characteristics.block_length)
      {
         return FLASH_EUNMATCHED_ERASEBLOCK;
      }

      /* Find the ReadLock associated with the flash address. */
      read_lock = instance->read_lock_pointer_list[
         ((size_t)(eraseblock_addr - instance->instance_start)) >>
         (instance->read_lock_index_shift)];

      assert(read_lock->base <= eraseblock_addr);
      assert(((size_t)(eraseblock_addr - read_lock->base)) < read_lock->length);
      WHITEBOX_CALLBACK(TPT_READ_LOCK, read_lock->base, read_lock->length);
      WHITEBOX_CALLBACK(TPT_MODIFY_LOCK, read_lock->modify_lock->base,
         read_lock->modify_lock->length);

      /* Get modify access permissions for the lock region */
      access_lock = fam_acquire_modify_lock(read_lock, AT_ERASE);

      /*
      *  This loop makes several erase attempts if there is an erase failure
      *  and the driver has requested read back.
      *  Note that there will not be multiple attempts if the driver reports
      *  an error.
      */
      attempt_count = instance->instance_capabilities->verify_erase;
      for (;;)
      {
         /*
         *  Enable the interrupt polling algorithm for direct access.
         */
         if (read_lock->is_direct_access)
         {
            FAM_DISABLE_INTERRUPT(int_mask_save);
         }

         /* Start the erase operation. */
         access_lock->params.destination = eraseblock_addr;
         FT_CHECK_WILL_START_ERASE();
         result = instance->driver->v1.erase_start(instance->driver_data,
                                                &access_lock->params);

         if (result != FLASHDRIVER_EOK)
         {
            FT_CHECK_DID_STOP_ERASE();
            goto FAILURE_CLEANUP_1;
         }

         /*
         *  Select a polling algorithm to poll the driver until the operation
         *  is completed.
         */
         if (read_lock->is_direct_access)
         {
            /*
            *  Polling loop for access of flash that is configured for direct
            *  access. This poll loop operates with disabled interrupts and
            *  polls for interrupt request during the operation.
            */
            do
            {
               /* 
               *  Check for interrupt request.
               *  This is only done if the driver supports suspend.
               */
               if (instance->instance_capabilities->suspend_erase_support &&
                   (fam_global.irq_poll_function() !=
                    IRQ_POLL_NO_PENDING_INTERRUPT))
               {
                  /* Interrupt request detected. Suspend the operation. */
                  result = instance->driver->v1.erase_suspend(instance->driver_data,
                                                           &access_lock->params);
                  FT_CHECK_DID_STOP_ERASE();
                  if (result != FLASHDRIVER_EOK)
                  {
                     goto FAILURE_CLEANUP_1;
                  }
                  FAM_ENABLE_INTERRUPT(int_mask_save);

                  /*
                  *  The interrupt will be serviced immediately after enable.
                  *  Check if there has been a suspend request during the
                  *  service of the interrupt.
                  */
                  if (fam_poll_suspend_request(access_lock))
                  {
                     fam_wait_during_suspend(access_lock);
                  }

                  /* Continue the operation with interrupts disabled. */
                  FAM_DISABLE_INTERRUPT(int_mask_save);
                  FT_CHECK_WILL_START_ERASE();
                  result = instance->driver->v1.erase_resume(instance->driver_data,
                                                          &access_lock->params);
                  if (result != FLASHDRIVER_EOK)
                  {
                     FT_CHECK_DID_STOP_ERASE();
                     goto FAILURE_CLEANUP_1;
                  }
               }
               result = instance->driver->v1.erase_poll(instance->driver_data, 
                                                     &access_lock->params);
            } while (result == FLASHDRIVER_EIN_PROGRESS);
            FT_CHECK_DID_STOP_ERASE();
            FAM_ENABLE_INTERRUPT(int_mask_save);
         }
         else
         {
            /*
            *  Polling loop for access of flash that is NOT configured for
            *  direct access. This poll loop checks for a suspend request
            *  each loop turn.
            */
            do
            {
               /*
               *  Handling of suspend request only if the driver supports it.
               */
               if (instance->instance_capabilities->suspend_erase_support)
               {
                  /* Wait for a suspend request during the delay time. */
                  if (fam_wait_suspend_request(access_lock,
                                               access_lock->params.delay_time))
                  {
                     /* Suspend request detected. Suspend the operation. */
                     result = instance->driver->v1.erase_suspend(
                        instance->driver_data, &access_lock->params);
                     FT_CHECK_DID_STOP_ERASE();
                     if (result != FLASHDRIVER_EOK)
                     {
                        goto FAILURE_CLEANUP_1;
                     }

                     /* Allow the process that requested suspend to execute. */
                     fam_wait_during_suspend(access_lock);

                     /* Resume the operation. */
                     FT_CHECK_WILL_START_ERASE();
                     result = instance->driver->v1.erase_resume(
                        instance->driver_data, &access_lock->params);
                     if (result != FLASHDRIVER_EOK)
                     {
                        FT_CHECK_DID_STOP_ERASE();
                        goto FAILURE_CLEANUP_1;
                     }
                  }
               }
               else
               {
                  /* The driver does not support suspend. */
                  WHITEBOX_TESTPOINT(TPT_SYNCH_SYSCALL);
                  delay(access_lock->params.delay_time);
                  WHITEBOX_TESTPOINT(TPT_SYNCH_SYSCALL);
               }

               /* Poll the driver to check if the operation is completed. */
               result = instance->driver->v1.erase_poll(instance->driver_data, 
                                                     &access_lock->params);
            } while (result == FLASHDRIVER_EIN_PROGRESS);
            FT_CHECK_DID_STOP_ERASE();
         }

         /* Check the result from erase_poll for errors. */
         if (result != FLASHDRIVER_EOK)
         {
            goto FAILURE_CLEANUP_2;
         }

         /* Read back the erased data. */
         if ((attempt_count == 0) ||
             readback_erasecheck(eraseblock_addr, 
                                 inst_characteristics.block_length))
         {
            /* The data matched, or we did not use read back. */
            break;   /* Out of for loop */
         }

         /* The eraseblock did not check out as erased. */
         attempt_count--;
         if (attempt_count == 0)
         {
            goto FAILURE_CLEANUP_3;
         }
      } /* End of for (;;) */

      /* Count the progress of the erase. */
      eraseblock_addr += inst_characteristics.block_length;
      length -= inst_characteristics.block_length;

      /* Release the modify access permissions for the lock region */
      fam_release_modify_lock(access_lock);
   } /* end of while (length > 0) */

   return FLASH_EOK;


FAILURE_CLEANUP_1:
   if (read_lock->is_direct_access)
   {
      FAM_ENABLE_INTERRUPT(int_mask_save);
   }
   /* Fall through to next label. */

FAILURE_CLEANUP_2:
   if (result != FLASHDRIVER_EHARDWARE_ERROR &&
       result != FLASHDRIVER_EPROTECTED)
   {
      fam_release_modify_lock(access_lock);
      fam_internal_error(FLASH_EUNEXPECTED_DRIVER_ERROR, (OSERRCODE)result);
   }
   /* Fall through to next label. */

FAILURE_CLEANUP_3:
   /* Release the modify access permissions for the lock region */
   fam_release_modify_lock(access_lock);

   return FLASH_EHARDWARE_ERROR;
}



/****************************************************************
*  Function fam_flash_get_instance
*
*  Implements the bios call for flash_get_instance.
*  See "flash_api.h" for details.
*/
static FlashResult
fam_flash_get_instance(FlashInstanceInfo *instance_info,
                   U32 request_instance_nr)
{
   instance_info->instance_count = fam_global.instance_count;

   if (request_instance_nr >= fam_global.instance_count)
   {
      return FLASH_ENO_SUCH_INSTANCE;
   }

   instance_info->instance_start =
      fam_global.instance_list[request_instance_nr].instance_start;
   instance_info->instance_length =
      fam_global.instance_list[request_instance_nr].instance_length;

   return FLASH_EOK;
}



/****************************************************************
*  Function fam_flash_get_characteristics
*
*  Implements the bios call for flash_get_characteristics.
*  See "flash_api.h" for details.
*/
static FlashResult
fam_flash_get_characteristics(FlashCharacteristics *characteristics,
                              CharacteristicType request_type,
                              const FlashAddress request_address)
{
   InstanceControlBlock *instance;
   FlashDriverResult result;
   FlashCharacteristicsV1 inst_characteristics;

   /* Get the instance of the request address. */
   instance = fam_find_instance(request_address, FAM_INSTANCE_NONE);
   if (instance == NULL)
   {
      return FLASH_EINVALID_ADDRESS;
   }

   /* Pass the call directly to the driver. */
   result = instance->driver->v1.get_characteristics(&inst_characteristics,
                                                  instance->driver_data,
                                                  request_type,
                                                  request_address);


   if (result != FLASHDRIVER_EOK)
   {
      if (result == FLASHDRIVER_EUNKNOWN_REQUEST)
      {
         return FLASH_EUNKNOWN_REQUEST;
      }
      fam_internal_error(FLASH_EUNEXPECTED_DRIVER_ERROR, (OSERRCODE)result);
   }

   /* Copy the result to the user buffer. */
   characteristics->range_start = inst_characteristics.range_start;
   characteristics->range_length = inst_characteristics.range_length;
   characteristics->block_length = inst_characteristics.block_length;

   return FLASH_EOK;
}



/****************************************************************
*  Function fam_flash_throttle_get_limit
*
*  Implements the bios call for flash_throttle_get_limit.
*  See "flash_api.h" for details.
*/
static int
fam_flash_throttle_get_limit(void)
{
   if (!IS_FAM_THROTTLE_ACTIVATED())
   {
      fam_internal_error(FLASH_ETHROTTLE_NOT_ACTIVE,
                         FLASH_FCODE_THROTTLE_GET_LIMIT);
      return 0;
   }

   return fam_global.throttle_limit;
}



/****************************************************************
*  Function fam_flash_throttle_down
*
*  Implements the bios call for flash_throttle_down.
*  See "flash_api.h" for details.
*/
static int
fam_flash_throttle_down(void)
{
   int new_limit;
   U32 int_mask_save = 0;       /* Initiate to zero to keep compiler happy */

   if (!IS_FAM_THROTTLE_ACTIVATED())
   {
      fam_internal_error(FLASH_ETHROTTLE_NOT_ACTIVE,
                         FLASH_FCODE_THROTTLE_DOWN);
      return 0;
   }

   FAM_DISABLE_INTERRUPT(int_mask_save);
   fam_global.throttle_limit--;
   new_limit = fam_global.throttle_limit;
   FAM_ENABLE_INTERRUPT(int_mask_save);

   /* Allow one less operation to run simultaneously, if not previously already
    * at or below zero.
    */
   if (new_limit >= 0)
   {
      ENTER_FAM_THROTTLE();
   }

   return new_limit;
}



/****************************************************************
*  Function fam_flash_throttle_up
*
*  Implements the bios call for flash_throttle_up.
*  See "flash_api.h" for details.
*/
static int
fam_flash_throttle_up(void)
{
   int new_limit;
   U32 int_mask_save = 0;       /* Initiate to zero to keep compiler happy */

   if (!IS_FAM_THROTTLE_ACTIVATED())
   {
      fam_internal_error(FLASH_ETHROTTLE_NOT_ACTIVE,
                         FLASH_FCODE_THROTTLE_UP);
      return 0;
   }

   FAM_DISABLE_INTERRUPT(int_mask_save);
   fam_global.throttle_limit++;
   new_limit = fam_global.throttle_limit;
   FAM_ENABLE_INTERRUPT(int_mask_save);

   /* Allow one more operation to run simultaneously, if not previously below
    * zero.
    */
   if (new_limit > 0)
   {
      LEAVE_FAM_THROTTLE();
   }

   return new_limit;
}



/****************************************************************
*  Function fam_biosfunction
*
*  Call the various implementation functions depending on the function code.
*/
long
fam_biosfunction(long function_code,
                 long arg2,
                 long arg3,
                 long arg4,
                 long arg5,
                 long arg6,
                 long arg7)
{
   long result;

   switch (function_code)
   {
      case FLASH_FCODE_CHECKVERSION:
         /*
         *  This is a consistency check of the access library version
         *  (flash_api.c). Both the function code and the return value have
         *  been generated as random numbers.
         *  THE VALUE OF FLASH_EVERSION_OK SHOULD BE INCREMENTED IF THIS BIOS
         *  API IS CHANGED OR EXTENDED.
         */
         result = (long)FLASH_EVERSION_OK;
         break;

      case FLASH_FCODE_READ:
         result = (long)fam_flash_read2(
            (void *)arg2,
            (FlashAddress)arg3,
            (size_t)arg4,
            FAM_INSTANCE_NONE,
            FLASH_ACCESS_DATA);
         break;

      case FLASH_FCODE_WRITE:
         result = (long)fam_flash_write2(
            (FlashAddress)arg2,
            (const void *)arg3,
            (size_t)arg4,
            FAM_INSTANCE_NONE,
            FLASH_ACCESS_DATA);
         break;

      case FLASH_FCODE_ERASE:
         result = (long)fam_flash_erase(
            (FlashAddress)arg2,
            (size_t)arg3);
         break;

      case FLASH_FCODE_GET_INSTANCE:
         result = (long)fam_flash_get_instance(
            (FlashInstanceInfo *)arg2,
            (U32)arg3);
         break;

      case FLASH_FCODE_GET_CHARACTERISTICS:
         result = (long)fam_flash_get_characteristics(
            (FlashCharacteristics *)arg2,
            (CharacteristicType)arg3,
            (FlashAddress)arg4);
         break;

         /* Functions new in interface version 2 */
      case FLASH_FCODE_READ2:
         result = (long)fam_flash_read2(
            (void *) arg2,
            (FlashAddress) arg3,
            (size_t)arg4,
            (U32) arg5,
            (FlashAccessType) arg6);
         break;

      case FLASH_FCODE_WRITE2:
         result = (long)fam_flash_write2(
            (FlashAddress)arg2,
            (const void *)arg3,
            (size_t)arg4,
            (U32) arg5,
            (FlashAccessType) arg6);
         break;

      case FLASH_FCODE_THROTTLE_GET_LIMIT:
         result = (long)fam_flash_throttle_get_limit();
         break;

      case FLASH_FCODE_THROTTLE_DOWN:
         result = (long)fam_flash_throttle_down();
         break;

      case FLASH_FCODE_THROTTLE_UP:
         result = (long)fam_flash_throttle_up();
         break;

      default:
         /* Function not supported. */
         result = -1;
         (void)arg5;    /* Touch to avoid compiler warning. */
         (void)arg6;    /* Touch to avoid compiler warning. */
         (void)arg7;    /* Touch to avoid compiler warning. */
   }

   return result;
}
