/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2006 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

#include "fmd.h"

#include "ose.h"
#include "cpu.h"
#include "efmi.h"
#include "hfmi.h"
#include "pfmi.h"
#include "cpu.h"

#include "fmm.h"
#include "fmm_efmi.h"
#include "fmm_pfmi.h"
#include "fmm_sysparam.h"
#include "fmm_ramtrace.h"

#include "ramdump.h"
#include "ramdump_private.h"
#include "ramdump_heap.h"
#include "pm.sig"
#include "monitor.sig"
#include "monitor_execution.sig"
#include "monitor_events.sig"

#include "stdint.h"
#include "stdlib.h"
/* pose.h included due to SMP macros. */
#include "pose.h"

/*
** This code heavily uses assignments in conditional expressions. This RVCT
** pragma transforms such compiler warnings into less disturbing remarks.
*/
#if defined(__ARMCC_VERSION)
#pragma diag_remark 1293
#endif

#ifdef softOSE
#define FMD_DO_NOT_SWITCH_STACK
#endif

#define ALIGNED(addr, size) (((unsigned long) (addr) &\
                             (unsigned long) ((size) - 1UL)) == 0)

#define ALIGN_DOWN_TO(value, size) ((value) & ~((size) - 1UL))

#define ALIGN_UP_TO(value, size) (((value) + (size) - 1UL) & ~((size) - 1UL))

/* Text block - memory overflow */
#define DUMP_INCOMPLETE "Dump incomplete due to insufficient memory."

/* Text block - End of dump */
#define END_OF_DUMP_SUCCESS "End of Dump, Status : Successful."

#define END_OF_DUMP_FAILURE "End of Dump, Status : Incomplete."

/* Max number of processes in any scope - required by fmd property 
 * functionality
 */

union Signal
{
   uint32_t signo;
   MonitorInterfaceRequest                  interface_request;
   MonitorInterfaceReply                    interface_reply;
   MonitorNameRequest                       name_request;
   MonitorNameReply                         name_reply;
   MonitorConnectRequest                    connect_request;
   MonitorConnectReply                      connect_reply;
   MonitorGetInfoRequest                    get_info_request;
   MonitorGetRegisterInfoRequest            get_register_info_request;
   MonitorRegisterValues                    register_values;
   MonitorGetRegisterInfoEndmark            get_register_info_endmark;
   MonitorGetTraceRequest                   get_trace_request;
   struct PmProgramsRequest                 programs_request;
   struct PmProgramsReply                   programs_reply;
   struct PmProgramInfo                     program_info;
   struct PmLoadModuleInstallHandlesRequest load_module_install_handle_request;
   struct PmLoadModuleInstallHandlesReply   load_module_install_handle_reply;
   struct PmLoadModuleInfoRequest           load_module_info_request;
   struct PmLoadModuleInfoReply             load_module_info_reply;
   struct PmLoadModuleSectionInfoRequest    load_module_section_info_request;
   struct PmLoadModuleSectionInfoReply      load_module_section_info_reply;
   struct MonitorGetProcessInfoLongReply    get_process_info_long_reply;
};

struct FmmSignal
{
   uint32_t size;
   union Signal data;
};

union ContextInfo
{
   uint32_t id;
   struct EfmiProcessInfo process_info;
   struct EfmiBlockInfo   block_info;
   struct EfmiSegmentInfo segment_info;
   struct EfmiPoolInfo    pool_info;
};

/**
 * @data fmd_properties
 *
 * @brief To store user set property values.
 */
static FmmFmdProperties fmd_properties;

/**
 * @data switch_stack_args
 *
 * @brief Arguments for use with stack switched function.
 */
static FmmSwitchStackArgs switch_stack_args;

static uint32_t dump_empty = 0;

static uint32_t msr;
static uint32_t locked;

/**
 * @data interrupt_stack_dumped
 *
 * @brief Used to see if, and what part of, the interrupt stack is already dumped.
 *
 * @long For a non SMP system there is only one interrupt stack shared by all
 * interrupt processes. This variable is used to avoid dumping the stack more
 * than once. For SMP systems there is one per execution unit.
 *
 * @long Contains 0 if the interrupt stack is not dumped at all,
 * contains FMD_MEMORY_STACKS_TOP if the stack top is dumped,
 * contains FMD_MEMORY_STACKS_USED if the currently used stack is dumped, and
 * contains FMD_MEMORY_STACKS_OWNED if all of the stack is dumped.
 *
 */
static uint32_t interrupt_stack_dumped SMP_ARRAY;

extern uint32_t fmd_regdump_registers(struct cpu_regdump *regdump,
                                      MonitorRegisterId *ids,
                                      MonitorRegister *regs);

static uint32_t
lock(uint32_t interrupts)
{
   LOCK_SAVE(msr);

   /* Recursive locking? */
   if (locked == 0)
   {
      /* Get dump lock */
      locked++;

      /* Release interrupt block, if not handled */
      if (!interrupts)
      {
         LOCK_RESTORE(msr);
      }
      return 1;
   }
   else
   {
      /* Failed to get dump lock */
      LOCK_RESTORE(msr);
      return 0;
   }
}

static void
unlock(uint32_t interrupts)
{
   if (!interrupts)
   {
      LOCK_SAVE(msr);
   }

   locked--;

   LOCK_RESTORE(msr);
}

static uint32_t
start_dump(uint32_t a0, uint32_t a1, uint32_t a2,
           uint32_t a3, uint32_t a4, uint32_t a5)
{
   uint32_t ret = EFMI_OK;

   dump_empty = 1;

   (void)a0;
   (void)a1;
   (void)a2;
   (void)a3;
   (void)a4;
   (void)a5;

   ret = ramdump_private_start_new_dump();

   return ret;
}

static uint32_t
dump_init(void)
{
   uint32_t ret = EFMI_OK;
   FmmSignal *in = fmm_input_signal;
   FmmSignal *out = fmm_output_signal;

   if (dump_empty)
   {
      SMP_CODE(cpuid_t cpu;)

      dump_empty = 0;

      SMP_CODE(for (cpu = 0; cpu < OSE_SMP_MAX_CPUS; cpu++))
      {
	 interrupt_stack_dumped SMP_INDEX_X(cpu) = 0;
      }

      {
         extern MonitorCPUClass fmd_cpu_class;

         MonitorInterfaceRequest *req = &in->data.interface_request;
         MonitorInterfaceReply *rep = &out->data.interface_reply;

         req->sigNo = MONITOR_INTERFACE_REQUEST;
         in->size = sizeof(*req);

         if ((ret = ramdump_private_dump_request(in, RAMDUMP_COMPRESSION_NONE)))
         {
            return ret;
         }

         rep->sigNo = MONITOR_INTERFACE_REPLY;
         rep->reserved0 = 0;
         *(uint32_t *)rep->endian = MONITOR_ENDIAN_VALUE;
         rep->osType = MONITOR_OS_OSE_5;
         rep->cpuClass = fmd_cpu_class;
         rep->reserved1 = 0;
         rep->reserved2 = 0;
         rep->sigsCount = 0;
         out->size = sizeof(*rep);

         if ((ret = ramdump_private_dump_endmark(out, ret)))
         {
            return ret;
         }
      }

      {
         OSTICK ntick;
         MonitorConnectRequest *req = &in->data.connect_request;
         MonitorConnectReply *rep = &out->data.connect_reply;

         req->sigNo = MONITOR_CONNECT_REQUEST;
         req->name[0] = 0;
         in->size = sizeof(*req);

         if ((ret = ramdump_private_dump_request(in, RAMDUMP_COMPRESSION_NONE)))
         {
            return ret;
         }

         rep->sigNo = MONITOR_CONNECT_REPLY;
         rep->status = MONITOR_STATUS_OK;
         rep->tickLength = system_tick();     /* REVIEW: Ok to use? Well, ramdump does */
         rep->tick = get_systime(&ntick);/* REVIEW: Ok to use? Well, ramdump does */
         rep->ntick = ntick;
         rep->ntickFrequency = 1000000;
         rep->reserved0 = 0;
         rep->reserved1 = 0;
         rep->reserved2 = 0;
         rep->reserved3 = 0;
         rep->reserved4 = 0;
         rep->reserved5 = 0;
         rep->reserved6 = 0;
         rep->reserved7 = 0;
         rep->featuresCount = 0;
         out->size = sizeof(*rep);

         if ((ret = ramdump_private_dump_endmark(out, ret)))
         {
            return ret;
         }
      }
   }

   return ret;
}


static uint32_t
dump_error(uint32_t a0, uint32_t a1, uint32_t a2,
           uint32_t a3, uint32_t a4, uint32_t a5)
{
   RamdumpStatus ret;
   struct EfmiCurrent cur;

   (void)a4;
   (void)a5;

   if (dump_empty)
   {
      dump_init();
   }

   if ((ret = ose_efmi_get_current(&cur, sizeof(struct EfmiCurrent))))
   {
      return ret;
   }

   ret = ramdump_private_dump_error_info((const char *)a0, a1, a2, a3);

   return ret;
}

static uint32_t
dump_memory(uint32_t a0, uint32_t a1, uint32_t a2,
            uint32_t a3, uint32_t a4, uint32_t a5)
{
   RamdumpStatus ret;

   (void)a3;
   (void)a4;
   (void)a5;

   if (dump_empty)
   {
      dump_init();
   }

   ret = ramdump_private_dump_memory_image((const char *)a0, a1, a2);

   return ret;
}

static uint32_t
dump_string(uint32_t a0, uint32_t a1, uint32_t a2,
            uint32_t a3, uint32_t a4, uint32_t a5)
{
   RamdumpStatus ret;

   (void)a1;
   (void)a2;
   (void)a3;
   (void)a4;
   (void)a5;

   if (dump_empty)
   {
      dump_init();
   }

   ret = ramdump_private_dump_text_info((const char *)a0);

   return ret;
}

static FmdStatus
list_process(uint32_t id, FmdStatus (*info)(union ContextInfo *ci))
{
   FmdStatus ret;

   struct EfmiProcessInfo pi;

   ret = ose_efmi_get_process_info(id, &pi, sizeof(pi));

   if (!ret)
   {
      ret = info((union ContextInfo *)&pi);

   }

   return ret;
}

static FmdStatus
list_processes_in_block(uint32_t id, FmdStatus (*info)(union ContextInfo *ci))
{
   FmdStatus ret;

   struct EfmiProcessInfo pi;
   struct EfmiProcessCursor pc;

   ret = ose_efmi_get_first_process(id, &pc, &pi, sizeof(pi));

   while (!ret)
   {
      if ((ret = info((union ContextInfo *)&pi)))
      {
         return ret;
      }
      ret = ose_efmi_get_next_process(&pc, &pi);
   }
   return (ret == EFMI_END_OF_LIST) ? EFMI_OK : ret;
}

static FmdStatus
list_processes_in_segment(uint32_t id, FmdStatus (*info)(union ContextInfo *ci))
{
   FmdStatus ret;

   struct EfmiBlockInfo bi;
   struct EfmiBlockCursor bc;

   ret = ose_efmi_get_first_block(id, &bc, &bi, sizeof(bi));

   while (!ret)
   {
      if ((ret = list_processes_in_block(bi.block, info)))
      {
         return ret;
      }
      ret = ose_efmi_get_next_block(&bc, &bi);
   }
   return (ret == EFMI_END_OF_LIST) ? EFMI_OK : ret;
}

static FmdStatus
list_processes_in_system(FmdStatus (*info)(union ContextInfo *ci))
{
   FmdStatus ret;

   struct EfmiSegmentInfo si;
   struct EfmiSegmentCursor sc;

   ret = ose_efmi_get_first_segment(&sc, &si, sizeof(si));

   while (!ret)
   {
      if ((ret = list_processes_in_segment(si.segment, info)))
      {
         return ret;
      }
      ret = ose_efmi_get_next_segment(&sc, &si);
   }
   return (ret == EFMI_END_OF_LIST) ? EFMI_OK : ret;
}

static int32_t
dump_info_reply(FmmSignal *sig)
{
   return ramdump_private_dump_reply(sig);
}

static int fmd_proc_compslot(const void *e1, const void *e2)
{
   const FmmProcQueueCnt *t1 = (const FmmProcQueueCnt*)e1;
   const FmmProcQueueCnt *t2 = (const FmmProcQueueCnt*)e2;
   return  (int)(((int64_t)t2->sig_cnt) - ((int64_t)t1->sig_cnt));
}

static int32_t
process_info_reply(FmmSignal *sig)
{
   /* store process information into the ramdump heap space allocated*/
   if (fmd_properties.process_count < fmd_properties.num_long_queue_procs)
   {
      (fmd_properties.proc_queue_list + fmd_properties.process_count)->pid =
                         sig->data.get_process_info_long_reply.pid;
      (fmd_properties.proc_queue_list + fmd_properties.process_count)->sig_cnt =
                         sig->data.get_process_info_long_reply.signalsInQueue;

      (fmd_properties.process_count)++;

      /* sort the partially filled array in descending order */ 
      qsort(fmd_properties.proc_queue_list, fmd_properties.process_count,
                             sizeof(FmmProcQueueCnt), fmd_proc_compslot); 
      
   }
   else
   {
      /* From now on copy new data into last element of array */
      (fmd_properties.proc_queue_list + fmd_properties.process_count)->pid =
                         sig->data.get_process_info_long_reply.pid;
      (fmd_properties.proc_queue_list + fmd_properties.process_count)->sig_cnt =
                         sig->data.get_process_info_long_reply.signalsInQueue;
      /* sort the array in descending order */
      qsort(fmd_properties.proc_queue_list, fmd_properties.process_count+1,
                             sizeof(FmmProcQueueCnt), fmd_proc_compslot);

   }
   return 0;
}

static int32_t
dump_info_endmark(FmmSignal *sig, int32_t status)
{
   return ramdump_private_dump_endmark(sig, status);
}

static int32_t
process_info_endmark(FmmSignal *sig, int32_t status)
{
   return 0;
}

static int32_t
dump_program_info_request(FmmSignal *sig)
{
   return ramdump_private_dump_request(sig, RAMDUMP_COMPRESSION_ZLIB);
}

static uint32_t
dump_process_info(uint32_t scope, uint32_t info, uint32_t a2, uint32_t a3, uint32_t a4, uint32_t a5)
{
   FmdStatus ret = EFMI_OK;

   FmmSignal *in = fmm_input_signal;

   if (dump_empty)
   {
      dump_init();
   }

   if ((info & FMD_INFO_PROCESS)
       || (info & FMD_INFO_PROCESS_LONG))
   {
      struct EfmiCurrent cur;
      MonitorGetInfoRequest *req = &in->data.get_info_request;

      req->sigNo = MONITOR_GET_PROCESS_INFO_REQUEST;
      req->level = (info & FMD_INFO_PROCESS_LONG) != 0;
      switch(scope)
      {
         case FMD_SCOPE_SYSTEM:
            req->scope.type = scope;
            break;
         case FMD_SCOPE_SEGMENT:
            if ((ret = ose_efmi_get_current(&cur, sizeof(struct EfmiCurrent))))
            {
               return ret;
            }
            req->scope.type = scope;
            req->scope.id = cur.segment;
            break;
         case FMD_SCOPE_BLOCK:
            if ((ret = ose_efmi_get_current(&cur, sizeof(struct EfmiCurrent))))
            {
               return ret;
            }
            req->scope.type = scope;
            req->scope.id = cur.block;
            break;
         case FMD_SCOPE_PROCESS:
            if ((ret = ose_efmi_get_current(&cur, sizeof(struct EfmiCurrent))))
            {
               return ret;
            }
            req->scope.type = scope;
            req->scope.id = cur.process;
            break;
         default:
            return EFMI_CALL_INVALID;
      }
      req->reserved0 = 0;
      req->tagsCount = 0;
      in->size = sizeof(*req);

      if ((ret = ramdump_private_dump_request(in, RAMDUMP_COMPRESSION_ZLIB)))
      {
         return ret;
      }
      if ((ret = fmm_efmi_get_process_info_request(dump_info_reply,
                                                   dump_info_endmark, req)))
      {
         return ret;
      }

      req->sigNo = MONITOR_GET_BLOCK_INFO_REQUEST;
      req->level = 0;
      switch(scope)
      {
         case FMD_SCOPE_SYSTEM:
            req->scope.type = scope;
            break;
         case FMD_SCOPE_SEGMENT:
            req->scope.type = scope;
            req->scope.id = cur.segment;
            break;
         case FMD_SCOPE_BLOCK:
         case FMD_SCOPE_PROCESS:
            req->scope.type = MONITOR_SCOPE_ID;
            req->scope.id = cur.block;
            break;
         default:
            return EFMI_CALL_INVALID;
      }
      req->reserved0 = 0;
      req->tagsCount = 0;
      in->size = sizeof(*req);

      if ((ret = ramdump_private_dump_request(in, RAMDUMP_COMPRESSION_ZLIB)))
      {
         return ret;
      }
      if ((ret = fmm_efmi_get_block_info_request(dump_info_reply,
                                                 dump_info_endmark, req)))
      {
         return ret;
      }

      req->sigNo = MONITOR_GET_SEGMENT_INFO_REQUEST;
      req->level = 0;
      switch(scope)
      {
         case FMD_SCOPE_SYSTEM:
            req->scope.type = scope;
            break;
         case FMD_SCOPE_SEGMENT:
         case FMD_SCOPE_BLOCK:
         case FMD_SCOPE_PROCESS:
            req->scope.type = MONITOR_SCOPE_ID;
            req->scope.id = cur.segment;
            break;
         default:
            return EFMI_CALL_INVALID;
      }
      req->reserved0 = 0;
      req->tagsCount = 0;
      in->size = sizeof(*req);

      if ((ret = ramdump_private_dump_request(in, RAMDUMP_COMPRESSION_ZLIB)))
      {
         return ret;
      }
      if ((ret = fmm_efmi_get_segment_info_request(dump_info_reply,
                                                   dump_info_endmark, req)))
      {
         return ret;
      }
   }
   return ret;
}

static uint32_t
dump_stack_usage_info(uint32_t scope, uint32_t info, uint32_t a2, uint32_t a3, uint32_t a4, uint32_t a5)
{
   FmdStatus ret = EFMI_OK;

   FmmSignal *in = fmm_input_signal;

   if (dump_empty)
   {
      dump_init();
   }

   if (info & FMD_INFO_STACK_USAGE)
   {
      struct EfmiCurrent cur;
      MonitorGetInfoRequest *req = &in->data.get_info_request;

      req->sigNo = MONITOR_GET_STACK_USAGE_REQUEST;
      req->level = 0;
      switch(scope)
      {
         case FMD_SCOPE_SYSTEM:
            req->scope.type = scope;
            break;
         case FMD_SCOPE_SEGMENT:
            if ((ret = ose_efmi_get_current(&cur, sizeof(struct EfmiCurrent))))
            {
               return ret;
            }
            req->scope.type = scope;
            req->scope.id = cur.segment;
            break;
         case FMD_SCOPE_BLOCK:
            if ((ret = ose_efmi_get_current(&cur, sizeof(struct EfmiCurrent))))
            {
               return ret;
            }
            req->scope.type = scope;
            req->scope.id = cur.block;
            break;
         case FMD_SCOPE_PROCESS:
            if ((ret = ose_efmi_get_current(&cur, sizeof(struct EfmiCurrent))))
            {
               return ret;
            }
            req->scope.type = scope;
            req->scope.id = cur.process;
            break;
         default:
            return EFMI_CALL_INVALID;
      }
      req->reserved0 = 0;
      req->tagsCount = 0;
      in->size = sizeof(*req);

      if ((ret = ramdump_private_dump_request(in, RAMDUMP_COMPRESSION_ZLIB)))
      {
         return ret;
      }
      if ((ret = fmm_efmi_get_stack_usage_request(dump_info_reply,
                                                  dump_info_endmark, req)))
      {
         return  ret;
      }
   }
   return ret;
}

static uint32_t
dump_env_vars_info(uint32_t scope, uint32_t info, uint32_t a2, uint32_t a3, uint32_t a4, uint32_t a5)
{
   FmdStatus ret = EFMI_OK;

   FmmSignal *in = fmm_input_signal;

   if (dump_empty)
   {
      dump_init();
   }

   if (info & FMD_INFO_ENV_VARS)
   {
      struct EfmiCurrent cur;
      MonitorGetInfoRequest *req = &in->data.get_info_request;

      req->sigNo = MONITOR_GET_ENV_VARS_REQUEST;
      req->level = 0;
      switch(scope)
      {
         case FMD_SCOPE_SYSTEM:
            req->scope.type = scope;
            break;
         case FMD_SCOPE_SEGMENT:
            if ((ret = ose_efmi_get_current(&cur, sizeof(struct EfmiCurrent))))
            {
               return ret;
            }
            req->scope.type = scope;
            req->scope.id = cur.segment;
            break;
         case FMD_SCOPE_BLOCK:
            if ((ret = ose_efmi_get_current(&cur, sizeof(struct EfmiCurrent))))
            {
               return ret;
            }
            req->scope.type = scope;
            req->scope.id = cur.block;
            break;
         case FMD_SCOPE_PROCESS:
            if ((ret = ose_efmi_get_current(&cur, sizeof(struct EfmiCurrent))))
            {
               return ret;
            }
            req->scope.type = scope;
            req->scope.id = cur.process;
            break;
         default:
            return EFMI_CALL_INVALID;
      }
      req->reserved0 = 0;
      req->tagsCount = 0;
      in->size = sizeof(*req);

      if ((ret = ramdump_private_dump_request(in, RAMDUMP_COMPRESSION_ZLIB)))
      {
         return ret;
      }
      if ((ret = fmm_efmi_get_env_vars_request(dump_info_reply,
                                               dump_info_endmark, req)))
      {
         return ret;
      }
   }
   return ret;
}

static uint32_t
dump_signal_queue_info(uint32_t scope, uint32_t info, uint32_t a2, uint32_t a3, uint32_t a4, uint32_t a5)
{
   FmdStatus ret = EFMI_OK;

   FmmSignal *in = fmm_input_signal;

   if (dump_empty)
   {
      dump_init();
   }

   if (info & FMD_INFO_SIGNAL_QUEUE)
   {
      struct EfmiCurrent cur;
      MonitorGetInfoRequest *req = &in->data.get_info_request;

      req->sigNo = MONITOR_GET_SIGNAL_QUEUE_REQUEST;
      req->level = 0;
      switch(scope)
      {
         case FMD_SCOPE_SYSTEM:
            req->scope.type = scope;
            break;
         case FMD_SCOPE_SEGMENT:
            if ((ret = ose_efmi_get_current(&cur, sizeof(struct EfmiCurrent))))
            {
               return ret;
            }
            req->scope.type = scope;
            req->scope.id = cur.segment;
            break;
         case FMD_SCOPE_BLOCK:
            if ((ret = ose_efmi_get_current(&cur, sizeof(struct EfmiCurrent))))
            {
               return ret;
            }
            req->scope.type = scope;
            req->scope.id = cur.block;
            break;
         case FMD_SCOPE_PROCESS:
            if ((ret = ose_efmi_get_current(&cur, sizeof(struct EfmiCurrent))))
            {
               return ret;
            }
            req->scope.type = scope;
            req->scope.id = cur.process;
            break;
         default:
            return EFMI_CALL_INVALID;
      }
      req->reserved0 = 0;
      req->tagsCount = 0;
      in->size = sizeof(*req);

      if ((ret = ramdump_private_dump_request(in, RAMDUMP_COMPRESSION_ZLIB)))
      {
         return ret;
      }
      /* check whether user requested for
       * signal queue info for longest queue processes only */
      if (fmd_properties.num_long_queue_procs != 0)
      {
         uint32_t flush_cnt;
	 fmd_properties.process_count = 0;

	 /* Flush the ramdump heap allocated */
	 for (flush_cnt =0; flush_cnt < (fmd_properties.num_long_queue_procs + 1);
                      flush_cnt++) 
	 {
	    (fmd_properties.proc_queue_list + flush_cnt)->pid = 0;
	    (fmd_properties.proc_queue_list + flush_cnt)->sig_cnt = 0;
	 }

	 /* Get signal queue count of all the processes in the scope */
	 req->level = 1;
	 req->sigNo = MONITOR_GET_PROCESS_INFO_REQUEST;
	 if ((ret = fmm_efmi_get_process_info_request(process_info_reply,
              process_info_endmark, req)))
	 {
	    return ret;
	 }

	 /* sort all the processes in the scope
	  * on the basis of signal queue count in descending order
          */
	/* qsort(fmd_properties.proc_queue_list, fmd_properties.process_count,
               sizeof(FmmProcQueueCnt), fmd_proc_compslot);*/

         /* pass process count to fmm interface */
         fmm_efmi_set_fmd_properties(&fmd_properties);

      }
      /* dump signal queue info for all the processes in the scope */
      req->sigNo = MONITOR_GET_SIGNAL_QUEUE_REQUEST;
      req->level = 0;

      if ((ret = fmm_efmi_get_signal_queue_request(dump_info_reply,
                                                   dump_info_endmark, req)))
      {
         return ret;
      }
   }
   return ret;
}

static uint32_t
dump_pool_info(uint32_t scope, uint32_t info, uint32_t a2, uint32_t a3, uint32_t a4, uint32_t a5)
{
   FmdStatus ret = EFMI_OK;

   FmmSignal *in = fmm_input_signal;

   if (dump_empty)
   {
      dump_init();
   }

   if (info & FMD_INFO_POOL_INFO)
   {
      struct EfmiCurrent cur;
      MonitorGetInfoRequest *req = &in->data.get_info_request;

      req->sigNo = MONITOR_GET_POOL_INFO_REQUEST;
      req->level = 0;
      switch(scope)
      {
         case FMD_SCOPE_SYSTEM:
            req->scope.type = scope;
            break;
         case FMD_SCOPE_SEGMENT:
         case FMD_SCOPE_BLOCK:
         case FMD_SCOPE_PROCESS:
            if ((ret = ose_efmi_get_current(&cur, sizeof(struct EfmiCurrent))))
            {
               return ret;
            }
            req->scope.type = MONITOR_SCOPE_SEGMENT_ID;
            req->scope.id = cur.segment;
            break;
         default:
            return EFMI_CALL_INVALID;
      }
      req->reserved0 = 0;
      req->tagsCount = 0;
      in->size = sizeof(*req);

      if ((ret = ramdump_private_dump_request(in, RAMDUMP_COMPRESSION_ZLIB)))
      {
         return ret;
      }
      if ((ret = fmm_efmi_get_pool_info_request(dump_info_reply,
                                                dump_info_endmark, req)))
      {
         return ret;
      }
   }
   return ret;
}

static uint32_t
dump_pool_signals_info(uint32_t scope, uint32_t info, uint32_t a2, uint32_t a3, uint32_t a4, uint32_t a5)
{
   FmdStatus ret = EFMI_OK;

   FmmSignal *in = fmm_input_signal;

   if (dump_empty)
   {
      dump_init();
   }

   if (info & FMD_INFO_POOL_SIGNALS)
   {
      struct EfmiCurrent cur;
      MonitorGetInfoRequest *req = &in->data.get_info_request;

      req->sigNo = MONITOR_GET_POOL_SIGNALS_REQUEST;
      req->level = 0;
      switch(scope)
      {
         case FMD_SCOPE_SYSTEM:
            req->scope.type = scope;
            break;
         case FMD_SCOPE_SEGMENT:
         case FMD_SCOPE_BLOCK:
         case FMD_SCOPE_PROCESS:
            if ((ret = ose_efmi_get_current(&cur, sizeof(struct EfmiCurrent))))
            {
               return ret;
            }
            req->scope.type = MONITOR_SCOPE_SEGMENT_ID;
            req->scope.id = cur.segment;
            break;
         default:
            return EFMI_CALL_INVALID;
      }
      req->reserved0 = 0;
      req->tagsCount = 0;
      in->size = sizeof(*req);

      if ((ret = ramdump_private_dump_request(in, RAMDUMP_COMPRESSION_ZLIB)))
      {
         return ret;
      }
      if ((ret = fmm_efmi_get_pool_signals_request(dump_info_reply,
                                                   dump_info_endmark, req)))
      {
         return ret;
      }
   }

   if (info & FMD_INFO_POOL_SIGNALS_OWNED)
   {
      struct EfmiCurrent cur;
      MonitorGetInfoRequest *req = &in->data.get_info_request;

      req->sigNo = MONITOR_GET_POOL_SIGNALS_REQUEST;
      req->level = 0;
      switch(scope)
      {
         case FMD_SCOPE_SYSTEM:
            req->scope.type = scope;
            req->tagsCount = 0;
            break;
         case FMD_SCOPE_SEGMENT:
            if ((ret = ose_efmi_get_current(&cur, sizeof(struct EfmiCurrent))))
            {
               return ret;
            }
            req->scope.type = MONITOR_SCOPE_SEGMENT_ID;
            req->scope.id = cur.segment;
            req->tags[0].tag.type = MONITOR_SIGNAL_FILTER_TAG_OWNER_SID;
            req->tags[0].tag.count = 1+1;
            req->tags[1].num = cur.segment;
            req->tagsCount = 1+1;
            break;
         case FMD_SCOPE_BLOCK:
            if ((ret = ose_efmi_get_current(&cur, sizeof(struct EfmiCurrent))))
            {
               return ret;
            }
            req->scope.type = MONITOR_SCOPE_SEGMENT_ID;
            req->scope.id = cur.segment;
            req->tags[0].tag.type = MONITOR_SIGNAL_FILTER_TAG_OWNER_BID;
            req->tags[0].tag.count = 1+1;
            req->tags[1].num = cur.block;
            req->tagsCount = 1+1;
            break;
         case FMD_SCOPE_PROCESS:
            if ((ret = ose_efmi_get_current(&cur, sizeof(struct EfmiCurrent))))
            {
               return ret;
            }
            req->scope.type = MONITOR_SCOPE_SEGMENT_ID;
            req->scope.id = cur.segment;
            req->tags[0].tag.type = MONITOR_SIGNAL_FILTER_TAG_OWNER_ID;
            req->tags[0].tag.count = 1+1;
            req->tags[1].num = cur.block;
            req->tagsCount = 1+1;
            break;
         default:
            return EFMI_CALL_INVALID;
      }
      req->reserved0 = 0;
      in->size = sizeof(*req) + req->tagsCount * sizeof(MonitorTag) - 1;

      if ((ret = ramdump_private_dump_request(in, RAMDUMP_COMPRESSION_ZLIB)))
      {
         return ret;
      }
      if ((ret = fmm_efmi_get_pool_signals_request(dump_info_reply,
                                                   dump_info_endmark, req)))
      {
         return ret;
      }
   }
   return ret;
}

/* This is not yet supported */
#if 0
static uint32_t
dump_sys_param_info(uint32_t scope, uint32_t info, uint32_t a2, uint32_t a3, uint32_t a4, uint32_t a5)
{
   FmdStatus ret = EFMI_OK;
   FmmSignal *in = fmm_input_signal;

   if (dump_empty)
   {
      dump_init();
   }

   if (info & FMD_INFO_SYS_PARAM)
   {
      MonitorGetInfoRequest *req = &in->data.get_info_request;

      req->sigNo = MONITOR_GET_SYS_PARAM_REQUEST;
      req->level = 0;
      req->scope.type = MONITOR_SCOPE_SYSTEM;
      req->reserved0 = 0;
      in->size = sizeof(*req);

      if ((ret = ramdump_private_dump_request(in, RAMDUMP_COMPRESSION_ZLIB)))
      {
         return ret;
      }
      if ((ret = fmm_sysparam_get_sys_param_request(dump_info_reply,
                                                    dump_info_endmark, req)))
      {
         return ret;
      }
   }
   return ret;
}
#endif

static FmdStatus
dump_regdump(struct cpu_regdump *regdump, struct EfmiCurrent *cur)
{
   FmdStatus ret = EFMI_OK;

   if (dump_empty)
   {
      dump_init();
   }

   if (cur != NULL && regdump != NULL && regdump->magic == CPU_REGDUMP_MAGIC)
   {
      FmmSignal *in = fmm_input_signal;
      FmmSignal *out = fmm_output_signal;
      MonitorGetRegisterInfoRequest *req = &in->data.get_register_info_request;
      MonitorRegisterValues *rep = &out->data.register_values;
      MonitorGetRegisterInfoEndmark *end = &out->data.get_register_info_endmark;
      uint32_t count = fmd_regdump_registers(regdump,
                                             req->registers, rep->registers);

      req->sigNo = MONITOR_GET_REGISTER_INFO_REQUEST;
      req->scope.id = cur->process;
      req->scope.type = MONITOR_SCOPE_ID;
      req->registersCount = count;
      in->size = sizeof(*req)
                 - sizeof(MonitorRegisterId)
                 + count * sizeof(MonitorRegisterId);

      if ((ret = ramdump_private_dump_request(in, RAMDUMP_COMPRESSION_ZLIB)))
      {
         return ret;
      }

      rep->sigNo = MONITOR_GET_REGISTERS_REPLY;
      rep->pid = cur->process;
      rep->status = MONITOR_STATUS_OK;
      rep->registersCount = count;
      out->size = sizeof(*rep)
                  - sizeof(MonitorRegister)
                  + count * sizeof(MonitorRegister);

      if ((ret = ramdump_private_dump_reply(out)))
      {
         return ret;
      }

      end->sigNo = MONITOR_GET_REGISTER_INFO_ENDMARK;
      end->status = MONITOR_STATUS_OK;
      out->size = sizeof(*end);

      if ((ret = ramdump_private_dump_endmark(out, EFMI_OK)))
      {
         return ret;
      }
   }
   return ret;
}


static uint32_t
dump_registers_info(uint32_t scope, uint32_t info, uint32_t regdump, uint32_t a3, uint32_t a4, uint32_t a5)
{
   FmdStatus ret = EFMI_OK;

   if (dump_empty)
   {
      dump_init();
   }

   if (info & FMD_INFO_REGISTERS)
   {
      struct EfmiCurrent cur;
      MonitorGetRegisterInfoRequest *req;
      extern FmmSignal fmd_get_register_info_request;

      req = &fmd_get_register_info_request.data.get_register_info_request;

      switch(scope)
      {
         case FMD_SCOPE_SYSTEM:
            req->scope.type = scope;
            req->scope.id = 0;
            break;
         case FMD_SCOPE_SEGMENT:
            if ((ret = ose_efmi_get_current(&cur, sizeof(struct EfmiCurrent))))
            {
               return ret;
            }
            req->scope.type = scope;
            req->scope.id = cur.segment;
            break;
         case FMD_SCOPE_BLOCK:
            if ((ret = ose_efmi_get_current(&cur, sizeof(struct EfmiCurrent))))
            {
               return ret;
            }
            req->scope.type = scope;
            req->scope.id = cur.block;
            break;
         case FMD_SCOPE_PROCESS:
            if ((ret = ose_efmi_get_current(&cur, sizeof(struct EfmiCurrent))))
            {
               return ret;
            }
            req->scope.type = scope;
            req->scope.id = cur.process;
            break;
         default:
            return EFMI_CALL_INVALID;
      }

      if ((ret = ramdump_private_dump_request(&fmd_get_register_info_request,
                                              RAMDUMP_COMPRESSION_ZLIB)))
      {
         return ret;
      }
      if ((ret = fmm_efmi_get_register_info_request(dump_info_reply,
                                                    dump_info_endmark, req)))
      {
         return ret;
      }

      if ((struct cpu_regdump *)regdump != NULL)
      {
         if (ose_efmi_get_current(&cur, sizeof(struct EfmiCurrent)) == EFMI_OK)
         {
            if ((ret = dump_regdump((struct cpu_regdump *)regdump, &cur)))
            {
               return ret;
            }
         }
      }
   }

   return ret;
}

static uint32_t
dump_program_info(uint32_t scope, uint32_t info, uint32_t a2, uint32_t a3, uint32_t a4, uint32_t a5)
{
   FmdStatus ret = PFMI_OK;

   FmmSignal *in = fmm_input_signal;
   FmmSignal *out = fmm_output_signal;

   if (dump_empty)
   {
      dump_init();
   }

   if (info & FMD_INFO_PROGRAM_INFO)
   {
      FmdStatus ret2 = PFMI_OK;
      struct PmProgramsRequest *req = &in->data.programs_request;

      req->sigNo = PM_PROGRAMS_REQUEST;
      in->size = sizeof (*req);

      switch (scope)
      {
         case FMD_SCOPE_SYSTEM:
            if ((ret = ramdump_private_dump_request(in, RAMDUMP_COMPRESSION_ZLIB)))
            {
               return ret;
            }
            ret = fmm_pfmi_programs_request(dump_info_endmark, req);
            break;
         case FMD_SCOPE_SEGMENT:
         case FMD_SCOPE_BLOCK:
         case FMD_SCOPE_PROCESS:
         {
            struct EfmiCurrent cur;
            struct PmProgramsReply *rep = &out->data.programs_reply;

            if (ose_efmi_get_current(&cur, sizeof(struct EfmiCurrent)))
            {
               return PFMI_CALL_INVALID;
            }

            if ((ret = ramdump_private_dump_request(in, RAMDUMP_COMPRESSION_ZLIB)))
            {
               return ret;
            }

            rep->sigNo = PM_PROGRAMS_REPLY;
            if ((ret = ose_pfmi_get_progpid(cur.process, rep->progpids)))
            {
               rep->status = PM_EPROGPID;
               rep->progpids_size = 0;
            }
            else
            {
               rep->status = PM_SUCCESS;
               rep->progpids_size = 1;
            }
            out->size = sizeof (*rep);

            ret2 = ramdump_private_dump_endmark(out, ret);
            break;
         }
         default:
            return PFMI_CALL_INVALID;
      }
      if (ret || ret2)
      {
         return ret ? ret : ret2;
      }
   }

   if (info & FMD_INFO_PROGRAM_INFO)
   {
      struct PmProgramInfo *req = &in->data.program_info;

      switch (scope)
      {
         case FMD_SCOPE_SYSTEM:
            ret = fmm_pfmi_program_info(dump_program_info_request, dump_info_endmark);
            break;

         case FMD_SCOPE_SEGMENT:
         case FMD_SCOPE_BLOCK:
         case FMD_SCOPE_PROCESS:
         {
            struct EfmiCurrent cur;

            if (ose_efmi_get_current(&cur, sizeof(struct EfmiCurrent)))
            {
               return PFMI_CALL_INVALID;
            }

            req->sigNo = PM_PROGRAM_INFO_REQUEST;
            req->status = PM_SUCCESS;
            req->progpid = cur.block;
            in->size = sizeof(*req);

            if ((ret = ramdump_private_dump_request(in, RAMDUMP_COMPRESSION_ZLIB)))
            {
               return ret;
            }
            ret = fmm_pfmi_program_info_request(dump_info_endmark, req);
            break;
         }
         default:
            return PFMI_CALL_INVALID;
      }
      if (ret)
      {
         return ret;
      }
   }
   return ret;
}

static uint32_t
dump_load_module_info(uint32_t scope, uint32_t info, uint32_t a2, uint32_t a3, uint32_t a4, uint32_t a5)
{
   FmdStatus ret = PFMI_OK;

   FmmSignal *in = fmm_input_signal;
   FmmSignal *out = fmm_output_signal;

   if (dump_empty)
   {
      dump_init();
   }

   if (info & FMD_INFO_LOAD_MODULE_INFO)
   {
      FmdStatus ret2 = PFMI_OK;
      struct PmLoadModuleInstallHandlesRequest *req;

      req = &in->data.load_module_install_handle_request;

      req->sigNo = PM_LOAD_MODULE_INSTALL_HANDLES_REQUEST;
      in->size = sizeof (*req);

      switch (scope)
      {
         case FMD_SCOPE_SYSTEM:
            if ((ret = ramdump_private_dump_request(in, RAMDUMP_COMPRESSION_ZLIB)))
            {
               return ret;
            }
            ret = fmm_pfmi_load_module_install_handles_request(dump_info_endmark, req);
            break;
         case FMD_SCOPE_SEGMENT:
         case FMD_SCOPE_BLOCK:
         case FMD_SCOPE_PROCESS:
         {
            struct EfmiCurrent cur;
            struct PfmiProgramInfo program;
            struct PmLoadModuleInstallHandlesReply *rep;

            if (ose_efmi_get_current(&cur, sizeof(struct EfmiCurrent)))
            {
               return PFMI_CALL_INVALID;
            }

            rep = &out->data.load_module_install_handle_reply;

            if ((ret = ramdump_private_dump_request(in, RAMDUMP_COMPRESSION_ZLIB)))
            {
               return ret;
            }

            rep->sigNo = PM_LOAD_MODULE_INSTALL_HANDLES_REPLY;
            if ((ret = ose_pfmi_get_program_info(cur.block, &program)))
            {
               rep->status = PM_EPROGPID;
               rep->install_handles_no_of_strings = 0;
               rep->install_handles_size = 0;
            }
            else
            {
               rep->status = PM_SUCCESS;
               rep->install_handles_no_of_strings = 1;
               rep->install_handles_size = strlen(program.install_handle) + 1 + 1;
               strcpy(rep->install_handles, program.install_handle);
               rep->install_handles[rep->install_handles_size - 1] = 0;
            }
            out->size = sizeof (*rep) + rep->install_handles_size;

            ret2 = ramdump_private_dump_endmark(out, ret);
            break;
         }
         default:
            return PFMI_CALL_INVALID;
      }
      if (ret || ret2)
      {
         return ret ? ret : ret2;
      }
   }

   if (info & FMD_INFO_LOAD_MODULE_INFO)
   {
      struct PmLoadModuleInfoRequest *req = &in->data.load_module_info_request;

      switch (scope)
      {
         case FMD_SCOPE_SYSTEM:
            ret = fmm_pfmi_load_module_info(dump_program_info_request, dump_info_endmark);
            break;

         case FMD_SCOPE_SEGMENT:
         case FMD_SCOPE_BLOCK:
         case FMD_SCOPE_PROCESS:
         {
            struct EfmiCurrent cur;
            struct PfmiProgramInfo program;

            if (ose_efmi_get_current(&cur, sizeof(struct EfmiCurrent)))
            {
               return PFMI_CALL_INVALID;
            }

            req->sigNo = PM_LOAD_MODULE_INFO_REQUEST;
            if ((ret = ose_pfmi_get_program_info(cur.block, &program)))
            {
               return ret;
            }
            strcpy(req->install_handle, program.install_handle);
            in->size = sizeof (*req);

            if ((ret = ramdump_private_dump_request(in, RAMDUMP_COMPRESSION_ZLIB)))
            {
               return ret;
            }
            ret = fmm_pfmi_load_module_info_request(dump_info_endmark, req);
            break;
         }
         default:
            return PFMI_CALL_INVALID;
      }
      if (ret)
      {
         return ret;
      }
   }

   if (info & FMD_INFO_LOAD_MODULE_INFO)
   {
      struct PmLoadModuleSectionInfoRequest *req = &in->data.load_module_section_info_request;

      switch (scope)
      {
         case FMD_SCOPE_SYSTEM:
            ret = fmm_pfmi_load_module_section_info(dump_program_info_request, dump_info_endmark);
            break;

         case FMD_SCOPE_SEGMENT:
         case FMD_SCOPE_BLOCK:
         case FMD_SCOPE_PROCESS:
         {
            struct EfmiCurrent cur;
            struct PfmiProgramInfo program;
            struct PfmiLoadModuleInfo module;

            if (ose_efmi_get_current(&cur, sizeof(struct EfmiCurrent)))
            {
               return PFMI_CALL_INVALID;
            }

            req->sigNo = PM_LOAD_MODULE_SECTION_INFO_REQUEST;
            if ((ret = ose_pfmi_get_program_info(cur.block, &program)))
            {
               return ret;
            }
            strcpy(req->install_handle, program.install_handle);
            if ((ret = ose_pfmi_get_load_module_info(req->install_handle, &module)))
            {
               return ret;
            }
            req->section_interval_start = 0;
            req->section_interval_end = module.no_of_sections - 1;
            in->size = sizeof (*req);

            if ((ret = ramdump_private_dump_request(in, RAMDUMP_COMPRESSION_ZLIB)))
            {
               return ret;
            }
            ret = fmm_pfmi_load_module_section_info_request(dump_info_endmark, req);
            break;
         }
         default:
            return PFMI_CALL_INVALID;
      }
      if (ret)
      {
         return ret;
      }
   }
   return ret;
}

static uint32_t
dump_trace_info(uint32_t scope, uint32_t info, uint32_t a2, uint32_t a3, uint32_t a4, uint32_t a5)
{
   FmdStatus ret = EFMI_OK;
   FmmSignal *in = fmm_input_signal;

   if (dump_empty)
   {
      dump_init();
   }

   if (info & FMD_INFO_TRACE)
   {
      MonitorGetTraceRequest *req = &in->data.get_trace_request;

      req->sigNo = MONITOR_GET_TRACE_REQUEST;
      req->from = 0;
      req->to = ~0;
      req->reserved0 = 0;
      req->reserved1 = 0;
      req->reserved2 = 0;
      req->tagsCount = 0;

      in->size = sizeof(*req);

      if ((ret = ramdump_private_dump_request(in, RAMDUMP_COMPRESSION_ZLIB)))
      {
         return ret;
      }
      if ((ret = fmm_ramtrace_get_trace_request(dump_info_reply,
                                                dump_info_endmark, req)))
      {
         return ret;
      }
   }
   return ret;
}

static FmdStatus
list_pools_in_system(FmdStatus (*info)(union ContextInfo *ci))
{
   FmdStatus ret;

   struct EfmiPoolCursor cursor;
   struct EfmiPoolInfo pi;

   ret = ose_efmi_get_first_pool(&cursor, &pi, sizeof(pi));

   while (!ret)
   {
      if ((ret = info((union ContextInfo *)&pi)))
      {
         return ret;
      }
      ret = ose_efmi_get_next_pool(&cursor, &pi);
   }

   return ret == EFMI_END_OF_LIST ? EFMI_OK : ret;
}

static FmdStatus
get_pool_sid(uint32_t id, uint32_t *sid)
{
   FmdStatus ret;

   struct EfmiSegmentInfo si;
   struct EfmiSegmentCursor sc;

   ret = ose_efmi_get_first_segment(&sc, &si, sizeof(si));

   while (!ret)
   {
      struct EfmiBlockInfo bi;
      struct EfmiBlockCursor bc;

      ret = ose_efmi_get_first_block(si.segment, &bc, &bi, sizeof(bi));

      while (!ret)
      {
         if (bi.sig_pool == id || bi.stack_pool == id)
         {
            *sid = si.segment;
            return EFMI_OK;
         }
         ret = ose_efmi_get_next_block(&bc, &bi);
      }
      if (ret == EFMI_END_OF_LIST)
      {
         ret = ose_efmi_get_next_segment(&sc, &si);
      }
   }

   return ret == EFMI_END_OF_LIST ? EFMI_UNAVAILABLE : ret;
}

static FmdStatus
list_pools_in_segment(uint32_t id, FmdStatus (*info)(union ContextInfo *ci))
{
   FmdStatus ret;

   struct EfmiPoolCursor cursor;
   struct EfmiPoolInfo pi;

   ret = ose_efmi_get_first_pool(&cursor, &pi, sizeof(pi));

   while (!ret)
   {
      uint32_t sid;

      if ((ret = get_pool_sid(pi.pool_id, &sid)))
      {
         return ret;
      }
      if (sid == id)
      {
         if ((ret = info((union ContextInfo *)&pi)))
         {
            return ret;
         }
      }
      ret = ose_efmi_get_next_pool(&cursor, &pi);
   }

   return ret == EFMI_END_OF_LIST ? EFMI_OK : ret;
}

static FmdStatus
dump_aligned_memory(const char *description,
                    OSADDRESS address,
                    uint32_t size,
                    uint32_t alignment)
{
   FmdStatus ret;
   uint32_t aligned_address = ALIGN_DOWN_TO(address, alignment);
   uint32_t aligned_size = ALIGN_UP_TO(size + (address - aligned_address),
                                       alignment); 
   if ((ret = ramdump_private_dump_memory_image(description,
                                                aligned_address,
                                                aligned_size)))
   {
      return ret;
   }

   return EFMI_OK;
}

#define STACK_ALIGN 256

static FmdStatus
pool(union ContextInfo *ci)
{
   FmdStatus ret;

   struct EfmiFragmentCursor fc;
   struct EfmiFragmentInfo fi;

   ret = ose_efmi_get_first_fragment(ci->id, &fc, &fi, sizeof(fi));

   while (!ret)
   {
      if ((ret = dump_aligned_memory("POOL FRAGMENT",
                                     fi.base_address,
                                     fi.size,
                                     STACK_ALIGN)))
      {
         return ret;
      }

      ret = ose_efmi_get_next_fragment(&fc, &fi);
   }

   return ret == EFMI_END_OF_LIST ? EFMI_OK : ret;
}

static FmdStatus
pool_stacks(union ContextInfo *ci)
{
   FmdStatus ret;

   struct EfmiFragmentCursor fc;
   struct EfmiFragmentInfo fi;

   ret = ose_efmi_get_first_fragment(ci->id, &fc, &fi, sizeof(fi));

   while (!ret)
   {
      if (fi.used_for_stacks)
      {
         /* Assume stacks allocated from high fragment end */
         uint32_t base = fi.base_address + fi.size - fi.used_for_stacks;

         if ((ret = dump_aligned_memory("POOL FRAGMENT STACKS",
                                        base,
                                        fi.used_for_stacks,
                                        STACK_ALIGN)))
         {
            return ret;
         }
      }
      ret = ose_efmi_get_next_fragment(&fc, &fi);
   }

   return ret == EFMI_END_OF_LIST ? EFMI_OK : ret;
}

static FmdStatus
pool_signals(union ContextInfo *ci)
{
   FmdStatus ret;

   struct EfmiFragmentCursor fc;
   struct EfmiFragmentInfo fi;

   ret = ose_efmi_get_first_fragment(ci->id, &fc, &fi, sizeof(fi));

   while (!ret)
   {
      if (fi.used_for_signals)
      {
         /* Assume signals allocated from low fragment end */
         if ((ret = ramdump_private_dump_memory_image("POOL FRAGMENT SIGNALS",
                                        fi.base_address, fi.used_for_signals)))
         {
            return ret;
         }
      }
      ret = ose_efmi_get_next_fragment(&fc, &fi);
   }

   return ret == EFMI_END_OF_LIST ? EFMI_OK : ret;
}

static FmdStatus
stack(union ContextInfo *ci)
{
   FmdStatus ret = EFMI_OK;

   struct EfmiStackInfo *si;

   /* Check if it is an interrupt process and if enough of the interrupt stack is
      already dumped.
   */
   if (ci->process_info.type == OS_INT_PROC ||
       ci->process_info.type == OS_TI_PROC)
   {
      if (interrupt_stack_dumped SMP_INDEX_X(ci->process_info.cpu_id) ==
	  FMD_MEMORY_STACKS_OWNED)
      {
	 /* We have already dumped the interrupt stack. */
	 return ret; /* Bail out. */
      }
      else
      {
	 /* Not enough of the interrupt stack was dumped so dump it. */
	 interrupt_stack_dumped SMP_INDEX_X(ci->process_info.cpu_id) = FMD_MEMORY_STACKS_OWNED;
      }
   }

   si = &ci->process_info.process_stack;
   if (si->top - si->limit > 0)
   {
      if ((ret = dump_aligned_memory("PROCESS STACK",
                                     si->limit,
                                     si->top - si->limit,
                                     STACK_ALIGN)))
      {
         return ret;
      }
   }
   si = &ci->process_info.supervisor_stack;
   if (si->top - si->limit > 0)
   {
      if ((ret = dump_aligned_memory("PROCESS SUPERVISOR STACK",
                                     si->limit,
                                     si->top - si->limit,
                                     STACK_ALIGN)))
      {
         return ret;
      }
   }

   return ret;
}

static FmdStatus
stack_used(union ContextInfo *ci)
{
   FmdStatus ret = EFMI_OK;

   struct EfmiCurrent cur;
   struct EfmiStackInfo *ssi;
   struct EfmiStackInfo *psi;
   struct EfmiRegisterInfo ri;

   extern MonitorRegisterId fmd_stack_pointer_id;

   /* Check if it is an interrupt process and if enough of the interrupt stack is
      already dumped.
      FIXME Needs to be updated for SMP to handle per core interrupt stacks.
   */
   if (ci->process_info.type == OS_INT_PROC ||
       ci->process_info.type == OS_TI_PROC)
   {
      if (interrupt_stack_dumped SMP_INDEX_X(ci->process_info.cpu_id) == FMD_MEMORY_STACKS_OWNED ||
	  interrupt_stack_dumped SMP_INDEX_X(ci->process_info.cpu_id) == FMD_MEMORY_STACKS_USED)
      {
	 /* We have already dumped the interrupt stack. */
	 return ret; /* Bail out. */
      }
      else
      {
	 /* Not enough of the interrupt stack was dumped so dump the used part. */
	 interrupt_stack_dumped SMP_INDEX_X(ci->process_info.cpu_id) = FMD_MEMORY_STACKS_USED;
      }
   }

   psi = &ci->process_info.process_stack;
   ssi = &ci->process_info.supervisor_stack;

   if (psi->top - psi->limit > 0)
   {
      if ((ret = ose_efmi_get_current(&cur, sizeof(cur))) == EFMI_OK)
      {
         if (cur.process == ci->id)
         {
            if ((ret = dump_aligned_memory("PROCESS STACK",
                                           psi->limit, psi->top - psi->limit,
                                           STACK_ALIGN)))
            {
               return ret;
            }
            if (ssi->top - ssi->limit > 0)
            {
               if ((ret = dump_aligned_memory("PROCESS SUPERVISOR STACK",
                                              ssi->limit, ssi->top - ssi->limit,
                                              STACK_ALIGN)))
               {
                  return ret;
               }
            }
         }
         else
         {
            if ((ret = ose_efmi_get_register(ci->id, fmd_stack_pointer_id, &ri)) == EFMI_OK)
            {
               if (psi->limit <= ri.value && ri.value < psi->top)
               {
                  if ((ret = dump_aligned_memory("PROCESS STACK (USED)",
                                                 ri.value,
                                                 psi->top - ri.value,
                                                 STACK_ALIGN)))
                  {
                     return ret;
                  }
               }
               if (ssi->limit <= ri.value && ri.value < ssi->top)
               {
                  if ((ret = dump_aligned_memory("PROCESS STACK",
                                                 psi->limit,
                                                 psi->top - psi->limit,
                                                 STACK_ALIGN)))
                  {
                     return ret;
                  }
                  if ((ret = dump_aligned_memory("PROCESS SUPERVISOR STACK (USED)",
                                                 ri.value,
                                                 ssi->top - ri.value,
                                                 STACK_ALIGN)))
                  {
                     return ret;
                  }
               }
            }
            else
            {
               return 0;
            }
         }
      }
   }

   return ret;
}


static FmdStatus
stack_top(union ContextInfo *ci)
{
   FmdStatus ret = EFMI_OK;

   struct EfmiCurrent cur;
   struct EfmiStackInfo *ssi;
   struct EfmiStackInfo *psi;
   struct EfmiRegisterInfo ri;

   extern MonitorRegisterId fmd_stack_pointer_id;

   /* Check if it is an interrupt process and if enough of the interrupt stack is
      already dumped.
   */
   if (ci->process_info.type == OS_INT_PROC ||
       ci->process_info.type == OS_TI_PROC)
   {
      if (interrupt_stack_dumped SMP_INDEX_X(ci->process_info.cpu_id) == FMD_MEMORY_STACKS_OWNED ||
	  interrupt_stack_dumped SMP_INDEX_X(ci->process_info.cpu_id) == FMD_MEMORY_STACKS_USED ||
	  interrupt_stack_dumped SMP_INDEX_X(ci->process_info.cpu_id) == FMD_MEMORY_STACKS_TOP)
      {
	 /* We have already dumped the interrupt stack. */
	 return ret; /* Bail out. */
      }
      else
      {
	 /* The interrupt stack was not dumped so dump the top of it. */
	 interrupt_stack_dumped SMP_INDEX_X(ci->process_info.cpu_id) = FMD_MEMORY_STACKS_TOP;
      }
   }

   psi = &ci->process_info.process_stack;
   ssi = &ci->process_info.supervisor_stack;
   if (psi->top - psi->limit > 0)
   {
      if ((ret = ose_efmi_get_current(&cur, sizeof(cur))) == EFMI_OK)
      {
         if (cur.process == ci->id)
         {
            if ((ret = dump_aligned_memory("PROCESS STACK",
                                           psi->limit, psi->top - psi->limit,
                                           STACK_ALIGN)))
            {
        
               return ret;
            }
            if (ssi->top - ssi->limit > 0)
            {
               if ((ret = dump_aligned_memory("PROCESS SUPERVISOR STACK",
                                              ssi->limit, ssi->top - ssi->limit,
                                              STACK_ALIGN)))
               {
                  return ret;
               }
            }
         }
         else
         {
            if ((ret = ose_efmi_get_register(ci->id, fmd_stack_pointer_id, &ri)) == EFMI_OK)
            {
               if (psi->limit <= ri.value && ri.value < psi->top)
               {
                  uint32_t top;

                  if ((top = psi->top - ri.value) > fmd_properties.max_stack_top_size)
                  {
                     top = fmd_properties.max_stack_top_size;
                  }
                  if ((ret = dump_aligned_memory("PROCESS STACK (TOP)",
                                                 ri.value, top, STACK_ALIGN)))
                  {
                     return ret;
                  }
               }
               if (ssi->limit <= ri.value && ri.value < ssi->top)
               {
                  uint32_t top;

                  if ((top = ssi->top - ri.value) > fmd_properties.max_stack_top_size)
                  {
                     top = fmd_properties.max_stack_top_size;
                  }
                  if ((ret = dump_aligned_memory("PROCESS SUPERVISOR STACK (TOP)",
                                                 ri.value, top, STACK_ALIGN)))
                  {
                     return ret;
                  }
               }
            }
            else
            {
               return 0;
            }
         }
      }
   }
   return ret;
}


static FmdStatus
pool_signals_owned(union ContextInfo *ci)
{
   FmdStatus ret;

   struct EfmiSignalCursor cursor;
   struct EfmiSignalInfo si;

   ret = ose_efmi_get_first_pool_signal(ci->id, &cursor, &si, sizeof(si));

   while (!ret)
   {
      if (si.owner)
      {
         if ((ret = ramdump_private_dump_memory_image("SIGNAL",
                                           si.signal_pointer, si.sig_size + 1)))
         {
            return ret;
         }
      }
      ret = ose_efmi_get_next_pool_signal(&cursor, &si);
   }

   return ret == EFMI_END_OF_LIST ? EFMI_OK : ret;
}

static FmdStatus
pool_signals_owned_segment(union ContextInfo *ci)
{
   FmdStatus ret;

   struct EfmiCurrent cur;
   struct EfmiBlockInfo bi;
   struct EfmiProcessInfo pi;
   struct EfmiSignalCursor cursor;
   struct EfmiSignalInfo si;

   if ((ret = ose_efmi_get_current(&cur, sizeof(cur))))
   {
      return ret;
   }

   ret = ose_efmi_get_first_pool_signal(ci->id, &cursor, &si, sizeof(si));

   while (!ret)
   {
      if (si.owner)
      {
         uint32_t bid;

         if (ose_efmi_get_process_info(si.owner, &pi, sizeof(pi)) == EFMI_OK)
            bid = pi.block;
         else
            bid = si.owner;

         /* TODO: Return error, skip or include signals with unknown owner? */
         if ((ose_efmi_get_block_info(bid, &bi, sizeof(bi)) == EFMI_OK)
             && (cur.segment == bi.segment))
         {
            if ((ret = ramdump_private_dump_memory_image("SIGNAL",
                                           si.signal_pointer, si.sig_size + 1)))
            {
               return ret;
            }
         }
      }
      ret = ose_efmi_get_next_pool_signal(&cursor, &si);
   }

   return ret == EFMI_END_OF_LIST ? EFMI_OK : ret;
}

static FmdStatus
pool_signals_owned_block(union ContextInfo *ci)
{
   FmdStatus ret;

   struct EfmiCurrent cur;
   struct EfmiProcessInfo pi;
   struct EfmiSignalCursor cursor;
   struct EfmiSignalInfo si;

   if ((ret = ose_efmi_get_current(&cur, sizeof(cur))))
   {
      return ret;
   }

   ret = ose_efmi_get_first_pool_signal(ci->id, &cursor, &si, sizeof(si));

   while (!ret)
   {
      if (si.owner)
      {
         uint32_t bid;

         if (ose_efmi_get_process_info(si.owner, &pi, sizeof(pi)) == EFMI_OK)
            bid = pi.block;
         else
            bid = si.owner;

         /* TODO: Return error, skip or include signals with unknown owner? */
         if (cur.block == bid)
         {
            if ((ret = ramdump_private_dump_memory_image("SIGNAL",
                                           si.signal_pointer, si.sig_size + 1)))
            {
               return ret;
            }
         }
      }
      ret = ose_efmi_get_next_pool_signal(&cursor, &si);
   }

   return ret == EFMI_END_OF_LIST ? EFMI_OK : ret;
}

static FmdStatus
pool_signals_owned_process(union ContextInfo *ci)
{
   FmdStatus ret;

   struct EfmiCurrent cur;
   struct EfmiSignalCursor cursor;
   struct EfmiSignalInfo si;

   if ((ret = ose_efmi_get_current(&cur, sizeof(cur))))
   {
      return ret;
   }

   ret = ose_efmi_get_first_pool_signal(ci->id, &cursor, &si, sizeof(si));

   while (!ret)
   {
      if (si.owner == cur.process)
      {
         if ((ret = ramdump_private_dump_memory_image("SIGNAL",
                                           si.signal_pointer, si.sig_size + 1)))
         {
            return ret;
         }
      }
      ret = ose_efmi_get_next_pool_signal(&cursor, &si);
   }

   return ret == EFMI_END_OF_LIST ? EFMI_OK : ret;
}

static FmdStatus
pool_signals_queued(union ContextInfo *ci)
{
   FmdStatus ret;

   struct EfmiSignalCursor cursor;
   struct EfmiSignalInfo si;
   uint32_t max_signals = fmd_properties.max_signals_in_queue;
   ret = ose_efmi_get_first_q_signal(ci->id, &cursor, &si, sizeof(si));

   while (!ret)
   {
      if ((ret = ramdump_private_dump_memory_image("SIGNAL",
                                        si.signal_pointer, si.sig_size + 1)))
      {
         return ret;
      }
      /* dump max number of signals specified */
      if (!fmd_properties.max_signals_in_queue)

      {
         ret = ose_efmi_get_next_q_signal(&cursor, &si);
      }
      else
      {
	 max_signals--;
	 if (max_signals != 0)
         {
            ret = ose_efmi_get_next_q_signal(&cursor, &si);
         }
	 else
         {
	    return ret = EFMI_OK;
         }
      }

   }

   return ret == EFMI_END_OF_LIST ? EFMI_OK : ret;
}

static FmdStatus
pool_signals_sent(union ContextInfo *ci)
{
   FmdStatus ret;

   struct EfmiSignalCursor cursor;
   struct EfmiSignalInfo si;

   ret = ose_efmi_get_first_pool_signal(ci->id, &cursor, &si, sizeof(si));

   while (!ret)
   {
      if ((si.owner == 0)
          && (si.addressee_pid != ci->id)
          && (si.sender_pid != 0))
      {
         if ((ret = ramdump_private_dump_memory_image("SIGNAL",
                                           si.signal_pointer, si.sig_size + 1)))
         {
            return ret;
         }
      }
      ret = ose_efmi_get_next_pool_signal(&cursor, &si);
   }

   return ret == EFMI_END_OF_LIST ? EFMI_OK : ret;
}

static FmdStatus
pool_signals_sent_segment(union ContextInfo *ci)
{
   FmdStatus ret;

   struct EfmiCurrent cur;
   struct EfmiBlockInfo bi;
   struct EfmiProcessInfo pi;
   struct EfmiSignalCursor cursor;
   struct EfmiSignalInfo si;

   if ((ret = ose_efmi_get_current(&cur, sizeof(cur))))
   {
      return ret;
   }

   ret = ose_efmi_get_first_pool_signal(ci->id, &cursor, &si, sizeof(si));

   while (!ret)
   {
      if ((si.owner == 0)
          && (si.addressee_pid != ci->id)
          && (si.sender_pid != 0))
      {
         uint32_t bid;

         if (ose_efmi_get_process_info(si.sender_pid, &pi, sizeof(pi)) == EFMI_OK)
            bid = pi.block;
         else
            bid = si.sender_pid;

         /* TODO: Return error, skip or include signals with unknown sender? */
         if ((ose_efmi_get_block_info(bid, &bi, sizeof(bi)) == EFMI_OK)
             && (cur.segment == bi.segment))
         {
            if ((ret = ramdump_private_dump_memory_image("SIGNAL",
                                           si.signal_pointer, si.sig_size + 1)))
            {
               return ret;
            }
         }
      }
      ret = ose_efmi_get_next_pool_signal(&cursor, &si);
   }

   return ret == EFMI_END_OF_LIST ? EFMI_OK : ret;
}

static FmdStatus
pool_signals_sent_block(union ContextInfo *ci)
{
   FmdStatus ret;

   struct EfmiCurrent cur;
   struct EfmiProcessInfo pi;
   struct EfmiSignalCursor cursor;
   struct EfmiSignalInfo si;

   if ((ret = ose_efmi_get_current(&cur, sizeof(cur))))
   {
      return ret;
   }

   ret = ose_efmi_get_first_pool_signal(ci->id, &cursor, &si, sizeof(si));

   while (!ret)
   {
      if ((si.owner == 0)
          && (si.addressee_pid != ci->id)
          && (si.sender_pid != 0))
      {
         uint32_t bid;

         if (ose_efmi_get_process_info(si.sender_pid, &pi, sizeof(pi)) == EFMI_OK)
            bid = pi.block;
         else
            bid = si.sender_pid;

         /* TODO: Return error, skip or include signals with unknown sender? */
         if (cur.block == bid)
         {
            if ((ret = ramdump_private_dump_memory_image("SIGNAL",
                                           si.signal_pointer, si.sig_size + 1)))
            {
               return ret;
            }
         }
      }
      ret = ose_efmi_get_next_pool_signal(&cursor, &si);
   }

   return ret == EFMI_END_OF_LIST ? EFMI_OK : ret;
}

static FmdStatus
pool_signals_sent_process(union ContextInfo *ci)
{
   FmdStatus ret;

   struct EfmiCurrent cur;
   struct EfmiSignalCursor cursor;
   struct EfmiSignalInfo si;

   if ((ret = ose_efmi_get_current(&cur, sizeof(cur))))
   {
      return ret;
   }

   ret = ose_efmi_get_first_pool_signal(ci->id, &cursor, &si, sizeof(si));

   while (!ret)
   {
      if ((si.owner == 0)
          && (si.addressee_pid != ci->id)
          && (si.sender_pid == cur.process))
      {
         if ((ret = ramdump_private_dump_memory_image("SIGNAL",
                                           si.signal_pointer, si.sig_size + 1)))
         {
            return ret;
         }
      }
      ret = ose_efmi_get_next_pool_signal(&cursor, &si);
   }

   return ret == EFMI_END_OF_LIST ? EFMI_OK : ret;
}

static FmdStatus
list_heaps_in_system(FmdStatus (*info)(struct HfmiHeapInfo *hi))
{
   FmdStatus ret;

   struct HfmiHeapCursor cursor;
   struct HfmiHeapInfo hi;

   ret = ose_hfmi_get_first_heap(&cursor, &hi, sizeof(hi));

   while (!ret)
   {
      if ((ret = info(&hi)))
      {
         return ret;
      }
      ret = ose_hfmi_get_next_heap(&cursor, &hi, sizeof(hi));
   }

   return ret == HFMI_END_OF_LIST ? HFMI_OK : ret;
}

static FmdStatus
list_heaps_in_segment(uint32_t id, FmdStatus (*info)(struct HfmiHeapInfo *hi))
{
   FmdStatus ret;

   struct EfmiBlockInfo bi;
   struct EfmiProcessInfo pi;
   struct HfmiHeapCursor cursor;
   struct HfmiHeapInfo hi;

   ret = ose_hfmi_get_first_heap(&cursor, &hi, sizeof(hi));

   while (!ret)
   {
      uint32_t bid;

      if (ose_efmi_get_process_info(hi.owner, &pi, sizeof(pi)) == EFMI_OK)
         bid = pi.block;
      else
         bid = hi.owner;

      /* TODO: Return error, skip or include heaps with unknown owner? */
      if ((ose_efmi_get_block_info(bid, &bi, sizeof(bi)) == EFMI_OK)
          && (id == bi.segment))
      {
         if ((ret = info(&hi)))
         {
            return ret;
         }
      }
      ret = ose_hfmi_get_next_heap(&cursor, &hi, sizeof(hi));
   }

   return ret == HFMI_END_OF_LIST ? HFMI_OK : ret;
}

static FmdStatus
heap(struct HfmiHeapInfo *hi)
{
   FmdStatus ret;

   struct HfmiFragmentCursor cursor;
   struct HfmiFragmentInfo fi;

   ret = ose_hfmi_get_first_fragment(hi->href, &cursor, &fi, sizeof(fi));

   while (!ret)
   {
      if ((ret = ramdump_private_dump_memory_image("HEAP FRAGMENT",
                                                        fi.address, fi.size)))
      {
         return ret;
      }
      ret = ose_hfmi_get_next_fragment(&cursor, &fi, sizeof(fi));
   }

   return ret == HFMI_END_OF_LIST ? HFMI_OK : ret;
}

static FmdStatus
heap_buffers(struct HfmiHeapInfo *hi)
{
   FmdStatus ret;

   struct HfmiBufferCursor cursor;
   struct HfmiBufferInfo bi;

   ret = ose_hfmi_get_first_buffer(hi->href, &cursor, &bi, sizeof(bi));

   while (!ret)
   {
      if ((ret = ramdump_private_dump_memory_image("HEAP BUFFER",
                                       bi.data_address, bi.actual_size)))
      {
         return ret;
      }
      ret = ose_hfmi_get_next_buffer(&cursor, &bi, sizeof(bi));
   }

   return ret == HFMI_END_OF_LIST ? HFMI_OK : ret;
}

static FmdStatus
heap_buffers_owned(struct HfmiHeapInfo *hi)
{
   FmdStatus ret;

   struct HfmiBufferCursor cursor;
   struct HfmiBufferInfo bi;

   ret = ose_hfmi_get_first_buffer(hi->href, &cursor, &bi, sizeof(bi));

   while (!ret)
   {
      if ((bi.status & HFMI_BUFFER_ALLOCATED))
      {
         if ((ret = ramdump_private_dump_memory_image("HEAP BUFFER",
                                       bi.data_address, bi.requested_size + 1)))
         {
            return ret;
         }
      }
      ret = ose_hfmi_get_next_buffer(&cursor, &bi, sizeof(bi));
   }

   return ret == HFMI_END_OF_LIST ? HFMI_OK : ret;
}

static FmdStatus
heap_buffers_owned_segment(struct HfmiHeapInfo *hi)
{
   FmdStatus ret;

   struct EfmiCurrent cur;
   struct EfmiProcessInfo epi;
   struct EfmiBlockInfo ebi;
   struct HfmiBufferCursor cursor;
   struct HfmiBufferInfo bi;

   if ((ret = ose_efmi_get_current(&cur, sizeof(cur))))
   {
      return HFMI_CALL_INVALID;
   }

   ret = ose_hfmi_get_first_buffer(hi->href, &cursor, &bi, sizeof(bi));

   while (!ret)
   {
      if (bi.status & HFMI_BUFFER_ALLOCATED)
      {
         uint32_t bid;

         if (ose_efmi_get_process_info(bi.owner, &epi, sizeof(epi)) == EFMI_OK)
            bid = epi.block;
         else
            bid = bi.owner;

         /* TODO: Return error, skip or include buffers with unknown owners? */
         if ((ose_efmi_get_block_info(bid, &ebi, sizeof(ebi)) == EFMI_OK)
             && (cur.segment == ebi.segment))
         {
            if ((ret = ramdump_private_dump_memory_image("HEAP BUFFER",
                                          bi.data_address, bi.requested_size + 1)))
            {
               return ret;
            }
         }
      }
      ret = ose_hfmi_get_next_buffer(&cursor, &bi, sizeof(bi));
   }

   return ret == HFMI_END_OF_LIST ? HFMI_OK : ret;
}

static FmdStatus
heap_buffers_owned_block(struct HfmiHeapInfo *hi)
{
   FmdStatus ret;

   struct EfmiCurrent cur;
   struct EfmiProcessInfo epi;
   struct HfmiBufferCursor cursor;
   struct HfmiBufferInfo bi;

   if ((ret = ose_efmi_get_current(&cur, sizeof(cur))))
   {
      return HFMI_CALL_INVALID;
   }

   ret = ose_hfmi_get_first_buffer(hi->href, &cursor, &bi, sizeof(bi));

   while (!ret)
   {
      if (bi.status & HFMI_BUFFER_ALLOCATED)
      {
         uint32_t bid;

         if (ose_efmi_get_process_info(bi.owner, &epi, sizeof(epi)) == EFMI_OK)
            bid = epi.block;
         else
            bid = bi.owner;

         /* TODO: Return error, skip or include buffers with unknown owners? */
         if (cur.block == bid)
         {
            if ((ret = ramdump_private_dump_memory_image("HEAP BUFFER",
                                          bi.data_address, bi.requested_size + 1)))
            {
               return ret;
            }
         }
      }
      ret = ose_hfmi_get_next_buffer(&cursor, &bi, sizeof(bi));
   }

   return ret == HFMI_END_OF_LIST ? HFMI_OK : ret;
}

static FmdStatus
heap_buffers_owned_process(struct HfmiHeapInfo *hi)
{
   FmdStatus ret;

   struct EfmiCurrent cur;
   struct HfmiBufferCursor cursor;
   struct HfmiBufferInfo bi;

   if ((ret = ose_efmi_get_current(&cur, sizeof(cur))))
   {
      return HFMI_CALL_INVALID;
   }

   ret = ose_hfmi_get_first_buffer(hi->href, &cursor, &bi, sizeof(bi));

   while (!ret)
   {
      if ((bi.status & HFMI_BUFFER_ALLOCATED)
          && (cur.process == bi.owner))
      {
         if ((ret = ramdump_private_dump_memory_image("HEAP BUFFER",
                                       bi.data_address, bi.requested_size + 1)))
         {
            return ret;
         }
      }
      ret = ose_hfmi_get_next_buffer(&cursor, &bi, sizeof(bi));
   }

   return ret == HFMI_END_OF_LIST ? HFMI_OK : ret;
}

static FmdStatus
list_load_modules_in_system(FmdStatus (*info)(struct PfmiLoadModuleInfo *li))
{
   FmdStatus ret;

   struct PfmiLoadModuleCursor cursor;
   struct PfmiLoadModuleInfo lmi;

   ret = ose_pfmi_get_first_load_module(&cursor, &lmi);

   while (!ret)
   {
      if ((ret = info(&lmi)))
      {
         return ret;
      }
      ret = ose_pfmi_get_next_load_module(&cursor, &lmi);
   }

   return ret == PFMI_END_OF_LIST ? PFMI_OK : ret;
}

static FmdStatus
list_load_modules_in_program(uint32_t id, FmdStatus (*info)(struct PfmiLoadModuleInfo *lmi))
{
   FmdStatus ret;

   struct PfmiProgramInfo pi;
   struct PfmiLoadModuleInfo lmi;

   ret = ose_pfmi_get_program_info(id, &pi);

   if (!ret)
   {
      ret = ose_pfmi_get_load_module_info(pi.install_handle, &lmi);

      if (!ret)
      {
         ret = info(&lmi);
      }
   }

   return ret;
}

static FmdStatus
data(struct PfmiLoadModuleInfo *lmi)
{
   return ramdump_private_dump_memory_image("DATA",
                                            lmi->data_base,
                                            lmi->data_size);
}

static FmdStatus
text(struct PfmiLoadModuleInfo *lmi)
{
   return ramdump_private_dump_memory_image("TEXT",
                                            lmi->text_base,
                                            lmi->text_size);
}

static uint32_t
dump_pool_memory(uint32_t scope, uint32_t memory, uint32_t regdump, uint32_t a3, uint32_t a4, uint32_t a5)
{
   FmdStatus ret = EFMI_OK;

   struct EfmiCurrent cur;

   if (dump_empty)
   {
      dump_init();
   }

   if (memory & FMD_MEMORY_POOL)
   {
      switch (scope)
      {
         case FMD_SCOPE_SYSTEM:
            if ((ret = list_pools_in_system(pool)))
            {
               return ret;
            }
            break;
         case FMD_SCOPE_SEGMENT:
         case FMD_SCOPE_BLOCK:
         case FMD_SCOPE_PROCESS:
            if ((ret = ose_efmi_get_current(&cur, sizeof(struct EfmiCurrent))))
            {
               return ret;
            }
            if ((ret = list_pools_in_segment(cur.segment, pool)))
            {
               return ret;
            }
            break;
         default:
            return EFMI_CALL_INVALID;
      }
   }
   else
   {
      if (memory & FMD_MEMORY_STACKS)
      {
         switch (scope)
         {
            case FMD_SCOPE_SYSTEM:
               if ((ret = list_pools_in_system(pool_stacks)))
               {
                  return ret;
               }
               break;
            case FMD_SCOPE_SEGMENT:
            case FMD_SCOPE_BLOCK:
            case FMD_SCOPE_PROCESS:
               if ((ret = ose_efmi_get_current(&cur, sizeof(struct EfmiCurrent))))
               {
                  return ret;
               }
               if ((ret = list_pools_in_segment(cur.segment, pool_stacks)))
               {
                  return ret;
               }
               break;
            default:
               return EFMI_CALL_INVALID;
         }
      }
      else if (memory & FMD_MEMORY_STACKS_OWNED)
      {
         switch (scope)
         {
            case FMD_SCOPE_SYSTEM:
               if ((ret = list_processes_in_system(stack)))
               {
                  return ret;
               }
               break;
            case FMD_SCOPE_SEGMENT:
               if ((ret = ose_efmi_get_current(&cur, sizeof(struct EfmiCurrent))))
               {
                  return ret;
               }
               if ((ret = list_processes_in_segment(cur.segment, stack)))
               {
                  return ret;
               }
               break;
            case FMD_SCOPE_BLOCK:
               if ((ret = ose_efmi_get_current(&cur, sizeof(struct EfmiCurrent))))
               {
                  return ret;
               }
               if ((ret = list_processes_in_block(cur.block, stack)))
               {
                  return ret;
               }
               break;
            case FMD_SCOPE_PROCESS:
               if ((ret = ose_efmi_get_current(&cur, sizeof(struct EfmiCurrent))))
               {
                  return ret;
               }
               if ((ret = list_process(cur.process, stack)))
               {
                  return ret;
               }
               break;
            default:
               return EFMI_CALL_INVALID;
         }
      }
      else if (memory & FMD_MEMORY_STACKS_USED)
      {
         switch (scope)
         {
            case FMD_SCOPE_SYSTEM:
               if ((ret = list_processes_in_system(stack_used)))
               {
                  return ret;
               }
               break;
            case FMD_SCOPE_SEGMENT:
               if ((ret = ose_efmi_get_current(&cur, sizeof(struct EfmiCurrent))))
               {
                  return ret;
               }
               if ((ret = list_processes_in_segment(cur.segment, stack_used)))
               {
                  return ret;
               }
               break;
            case FMD_SCOPE_BLOCK:
               if ((ret = ose_efmi_get_current(&cur, sizeof(struct EfmiCurrent))))
               {
                  return ret;
               }
               if ((ret = list_processes_in_block(cur.block, stack_used)))
               {
                  return ret;
               }
               break;
            case FMD_SCOPE_PROCESS:
               if ((ret = ose_efmi_get_current(&cur, sizeof(struct EfmiCurrent))))
               {
                  return ret;
               }
               if ((ret = list_process(cur.process, stack_used)))
               {
                  return ret;
               }
               break;
            default:
               return EFMI_CALL_INVALID;
         }
      }
      else if (memory & FMD_MEMORY_STACKS_TOP)
      {
         switch (scope)
         {
            case FMD_SCOPE_SYSTEM:
               if ((ret = list_processes_in_system(stack_top)))
               {
                  return ret;
               }
               break;
            case FMD_SCOPE_SEGMENT:
               if ((ret = ose_efmi_get_current(&cur, sizeof(struct EfmiCurrent))))
               {
                  return ret;
               }
               if ((ret = list_processes_in_segment(cur.segment, stack_top)))
               {
                  return ret;
               }
               break;
            case FMD_SCOPE_BLOCK:
               if ((ret = ose_efmi_get_current(&cur, sizeof(struct EfmiCurrent))))
               {
                  return ret;
               }
               if ((ret = list_processes_in_block(cur.block, stack_top)))
               {
            	  return ret;
               }
               break;
            case FMD_SCOPE_PROCESS:
               if ((ret = ose_efmi_get_current(&cur, sizeof(struct EfmiCurrent))))
               {
                  return ret;
               }
               if ((ret = list_process(cur.process, stack_top)))
               {
                  return ret;
               }
               break;
            default:
               return EFMI_CALL_INVALID;
         }
      }
      else
      {
         /* No stack memory requested. */
      }

      if (memory & FMD_MEMORY_SIGNALS)
      {
         switch (scope)
         {
            case FMD_SCOPE_SYSTEM:
               if ((ret = list_pools_in_system(pool_signals)))
               {
                  return ret;
               }
               break;
            case FMD_SCOPE_SEGMENT:
            case FMD_SCOPE_BLOCK:
            case FMD_SCOPE_PROCESS:
               if ((ret = ose_efmi_get_current(&cur, sizeof(struct EfmiCurrent))))
               {
                  return ret;
               }
               if ((ret = list_pools_in_segment(cur.segment, pool_signals)))
               {
                  return ret;
               }
               break;
            default:
               return EFMI_CALL_INVALID;
         }
      }
      else
      {
         if (memory & FMD_MEMORY_SIGNALS_OWNED)
         {
            switch (scope)
            {
               case FMD_SCOPE_SYSTEM:
                  if ((ret = list_pools_in_system(pool_signals_owned)))
                  {
                     return ret;
                  }
                  break;
               case FMD_SCOPE_SEGMENT:
                  if ((ret = ose_efmi_get_current(&cur, sizeof(struct EfmiCurrent))))
                  {
                     return ret;
                  }
                  if ((ret = list_pools_in_system(pool_signals_owned_segment)))
                  {
                     return ret;
                  }
                  break;
               case FMD_SCOPE_BLOCK:
                  if ((ret = ose_efmi_get_current(&cur, sizeof(struct EfmiCurrent))))
                  {
                     return ret;
                  }
                  if ((ret = list_pools_in_system(pool_signals_owned_block)))
                  {
                     return ret;
                  }
                  break;
               case FMD_SCOPE_PROCESS:
                  if ((ret = ose_efmi_get_current(&cur, sizeof(struct EfmiCurrent))))
                  {
                     return ret;
                  }
                  if ((ret = list_pools_in_system(pool_signals_owned_process)))
                  {
                     return ret;
                  }
                  break;
               default:
                  return EFMI_CALL_INVALID;
            }
         }

         if (memory & FMD_MEMORY_SIGNALS_QUEUED)
         {
            /* check whether user requested signal information
             * for longest queue processes only
             */
            if (!fmd_properties.num_long_queue_procs)
            {
 	       switch (scope)
 	       {
 	          case FMD_SCOPE_SYSTEM:
 		     if ((ret = list_processes_in_system(pool_signals_queued)))
 		     {
 		        return ret;
 	             }
 		     break;
 		  case FMD_SCOPE_SEGMENT:
 		     if ((ret = ose_efmi_get_current(&cur, sizeof(struct EfmiCurrent))))
 		     {
 		        return ret;
 		     }
 		     if ((ret = list_processes_in_segment(cur.segment, pool_signals_queued)))
 		     {
	                return ret;
 		     }
 		     break;
 		  case FMD_SCOPE_BLOCK:
 		     if ((ret = ose_efmi_get_current(&cur, sizeof(struct EfmiCurrent))))
 		     {
 		        return ret;
 		     }
 		     if ((ret = list_processes_in_block(cur.block, pool_signals_queued)))
 		     {
 			return ret;
 		     }
 		     break;
 		  case FMD_SCOPE_PROCESS:
 		     if ((ret = ose_efmi_get_current(&cur, sizeof(struct EfmiCurrent))))
 		     {
 		        return ret;
 		     }
 		     if ((ret = list_process(cur.process, pool_signals_queued)))
 		     {
 	                return ret;
 		     }
 		     break;
 		  default:
 		     return EFMI_CALL_INVALID;
 	       }
            }   
            else
            {
               uint32_t i,flush_cnt;
               FmmSignal *in = fmm_input_signal;
 	       MonitorGetInfoRequest *req = &in->data.get_info_request;
	       
               fmd_properties.process_count = 0;
               /* Flush the ramdump heap allocated */
	       for (flush_cnt =0; flush_cnt < (fmd_properties.num_long_queue_procs + 1);
                     flush_cnt++)
	       {
	          (fmd_properties.proc_queue_list + flush_cnt)->pid = 0;
		  (fmd_properties.proc_queue_list + flush_cnt)->sig_cnt = 0;
	       }
               req->level = 1;
 	       req->sigNo = MONITOR_GET_PROCESS_INFO_REQUEST;
 	       if ((scope == FMD_SCOPE_SYSTEM)||(scope == FMD_SCOPE_SEGMENT)||
 			(scope == FMD_SCOPE_BLOCK)||(scope == FMD_SCOPE_PROCESS))
 	       {
 	          req->scope.type = scope;
 	       }
 	       else
 	       {
 		  return EFMI_CALL_INVALID;
 	       }
 	       /* Get signal queue count info
 	        * of all the processes in the scope
                */
 	       if ((ret = fmm_efmi_get_process_info_request(process_info_reply,
                    process_info_endmark, req)))
 	       {
	          return ret;
 	       }
 	       /* sort all the processes in the scope
 	        * on the basis of signal queue count in descending order
                */
 	      /* qsort(fmd_properties.proc_queue_list, fmd_properties.process_count,
                             sizeof(FmmProcQueueCnt), fmd_proc_compslot);*/

               /* Pass process count value to fmm_efmi */
               fmm_efmi_set_fmd_properties(&fmd_properties);
               
               /* dump signals of longest queue processes */
               for (i=0;i < ((fmd_properties.num_long_queue_procs >
                    fmd_properties.process_count)? fmd_properties.process_count:
         	    fmd_properties.num_long_queue_procs); i++)
               {
                  if ((ret = list_process((fmd_properties.proc_queue_list + i)->pid,
                       pool_signals_queued)))
         	  {
 	             return ret;
         	  }
               }
            }
         }

         if (memory & FMD_MEMORY_SIGNALS_SENT)
         {
            switch (scope)
            {
               case FMD_SCOPE_SYSTEM:
                  if ((ret = list_pools_in_system(pool_signals_sent)))
                  {
                     return ret;
                  }
                  break;
               case FMD_SCOPE_SEGMENT:
                  if ((ret = ose_efmi_get_current(&cur, sizeof(struct EfmiCurrent))))
                  {
                     return ret;
                  }
                  if ((ret = list_pools_in_system(pool_signals_sent_segment)))
                  {
                     return ret;
                  }
                  break;
               case FMD_SCOPE_BLOCK:
                  if ((ret = ose_efmi_get_current(&cur, sizeof(struct EfmiCurrent))))
                  {
                     return ret;
                  }
                  if ((ret = list_pools_in_system(pool_signals_sent_block)))
                  {
                     return ret;
                  }
                  break;
               case FMD_SCOPE_PROCESS:
                  if ((ret = ose_efmi_get_current(&cur, sizeof(struct EfmiCurrent))))
                  {
                     return ret;
                  }
                  if ((ret = list_pools_in_system(pool_signals_sent_process)))
                  {
                     return ret;
                  }
                  break;
               default:
                  return EFMI_CALL_INVALID;
            }
         }
      }
   }

   return ret;
}

static uint32_t
dump_heap_memory(uint32_t scope, uint32_t memory, uint32_t a2, uint32_t a3, uint32_t a4, uint32_t a5)
{
   FmdStatus ret = EFMI_OK;

   struct EfmiCurrent cur;

   if (dump_empty)
   {
      dump_init();
   }

   if (memory & FMD_MEMORY_HEAP)
   {
      switch (scope)
      {
         case FMD_SCOPE_SYSTEM:
            if ((ret = list_heaps_in_system(heap)))
            {
               return ret;
            }
            break;
         case FMD_SCOPE_SEGMENT:
         case FMD_SCOPE_BLOCK:
         case FMD_SCOPE_PROCESS:
            if (ose_efmi_get_current(&cur, sizeof(struct EfmiCurrent)))
            {
               return HFMI_CALL_INVALID;
            }
            if ((ret = list_heaps_in_segment(cur.segment, heap)))
            {
               return ret;
            }
            break;
         default:
            return HFMI_CALL_INVALID;
      }
   }
   else if (memory & FMD_MEMORY_HEAP_BUFFERS)
   {
      switch (scope)
      {
         case FMD_SCOPE_SYSTEM:
            if ((ret = list_heaps_in_system(heap_buffers)))
            {
               return ret;
            }
            break;
         case FMD_SCOPE_SEGMENT:
         case FMD_SCOPE_BLOCK:
         case FMD_SCOPE_PROCESS:
            if (ose_efmi_get_current(&cur, sizeof(struct EfmiCurrent)))
            {
               return HFMI_CALL_INVALID;
            }
            if ((ret = list_heaps_in_segment(cur.segment, heap_buffers)))
            {
               return ret;
            }
            break;
         default:
            return HFMI_CALL_INVALID;
      }
   }
   else if (memory & FMD_MEMORY_HEAP_BUFFERS_OWNED)
   {
      switch (scope)
      {
         case FMD_SCOPE_SYSTEM:
            if ((ret = list_heaps_in_system(heap_buffers_owned)))
            {
               return ret;
            }
            break;
         case FMD_SCOPE_SEGMENT:
            if (ose_efmi_get_current(&cur, sizeof(struct EfmiCurrent)))
            {
               return HFMI_CALL_INVALID;
            }
            if ((ret = list_heaps_in_segment(cur.segment, heap_buffers_owned_segment)))
            {
               return ret;
            }
            break;
         case FMD_SCOPE_BLOCK:
            if (ose_efmi_get_current(&cur, sizeof(struct EfmiCurrent)))
            {
               return HFMI_CALL_INVALID;
            }
            if ((ret = list_heaps_in_segment(cur.segment, heap_buffers_owned_block)))
            {
               return ret;
            }
            break;
         case FMD_SCOPE_PROCESS:
            if (ose_efmi_get_current(&cur, sizeof(struct EfmiCurrent)))
            {
               return HFMI_CALL_INVALID;
            }
            if ((ret = list_heaps_in_segment(cur.segment, heap_buffers_owned_process)))
            {
               return ret;
            }
            break;
         default:
            return HFMI_CALL_INVALID;
      }
   }
   else
   {
      /* No heap memory dump was requested. */
   }
   return ret;
}

static uint32_t
dump_program_memory(uint32_t scope, uint32_t memory, uint32_t a2, uint32_t a3, uint32_t a4, uint32_t a5)
{
   FmdStatus ret = PFMI_OK;

   struct EfmiCurrent cur;

   if (dump_empty)
   {
      dump_init();
   }

   if (memory & FMD_MEMORY_DATA)
   {
      switch (scope)
      {
         case FMD_SCOPE_SYSTEM:
            if ((ret = list_load_modules_in_system(data)))
            {
               return ret;
            }
            break;
         case FMD_SCOPE_SEGMENT:
         case FMD_SCOPE_BLOCK:
         case FMD_SCOPE_PROCESS:
            if (ose_efmi_get_current(&cur, sizeof(struct EfmiCurrent)))
            {
               return PFMI_CALL_INVALID;
            }
            if ((ret = list_load_modules_in_program(cur.block, data)))
            {
               return ret;
            }
            break;
         default:
            return PFMI_CALL_INVALID;
      }
   }
   if (memory & FMD_MEMORY_TEXT)
   {
      switch (scope)
      {
         case FMD_SCOPE_SYSTEM:
            if ((ret = list_load_modules_in_system(text)))
            {
               return ret;
            }
            break;
         case FMD_SCOPE_SEGMENT:
         case FMD_SCOPE_BLOCK:
         case FMD_SCOPE_PROCESS:
            if (ose_efmi_get_current(&cur, sizeof(struct EfmiCurrent)))
            {
               return PFMI_CALL_INVALID;
            }
            if ((ret = list_load_modules_in_program(cur.block, text)))
            {
               return ret;
            }
            break;
         default:
            return PFMI_CALL_INVALID;
      }
   }
   return ret;
}

FmdStatus
fmd_start_dump(void)
{
   FmdStatus ret;

   if (lock(1))
   {
#ifdef FMD_DO_NOT_SWITCH_STACK
      ret = start_dump(0, 0, 0, 0, 0, 0);
      fmd_properties.max_stack_top_size = 200;
      /* If error is due to insufficient memory,
       * dump text block stating the same
       */
      if (ret == RAMDUMP_ECURR_DUMP_TOO_LARGE)
      {
         ramdump_private_dump_text_info(DUMP_INCOMPLETE);
      }

#else
      switch_stack_args.arg1 = 0;
      switch_stack_args.arg2 = 0;
      switch_stack_args.arg3 = 0;
      switch_stack_args.arg4 = 0;
      switch_stack_args.arg5 = 0;
      switch_stack_args.arg6 = 0;

      ret = (FmdStatus)fmm_switch_stack(fmm_stack,
                                          start_dump,
                                          &switch_stack_args);
      fmd_properties.max_stack_top_size = 200;
      /* If error is due to insufficient memory,
       * dump text block stating the same
       */
      if (ret == RAMDUMP_ECURR_DUMP_TOO_LARGE)
      {
         ramdump_private_dump_text_info(DUMP_INCOMPLETE);
      }

#endif
      unlock(1);
   }
   else
   {
      ret = EFMI_CALL_INVALID;
   }
   return ret;
}

FmdStatus
fmd_dump_error(const char *description,
               OSBOOLEAN user_called,
               OSERRCODE ecode,
               OSERRCODE extra)
{
   FmdStatus ret;

   if (lock(1))
   {
#ifdef FMD_DO_NOT_SWITCH_STACK
      ret = dump_error((uint32_t)description, user_called, ecode, extra, 0, 0);
      /* If error is due to insufficient memory,
       * dump text block stating the same
       */
      if (ret == RAMDUMP_ECURR_DUMP_TOO_LARGE)
      {
         ramdump_private_dump_text_info(DUMP_INCOMPLETE);
      }

#else
      switch_stack_args.arg1 = (uint32_t)description;
      switch_stack_args.arg2 = (uint32_t)user_called;
      switch_stack_args.arg3 = (uint32_t)ecode;
      switch_stack_args.arg4 = (uint32_t)extra;
      switch_stack_args.arg5 = 0;
      switch_stack_args.arg6 = 0;

      ret = (FmdStatus)fmm_switch_stack(fmm_stack,
                                          dump_error,
                                          &switch_stack_args);
      /* If error is due to insufficient memory,
       * dump text block stating the same
       */
      if (ret == RAMDUMP_ECURR_DUMP_TOO_LARGE)
      {
         ramdump_private_dump_text_info(DUMP_INCOMPLETE);
      }

#endif
      unlock(1);
   }
   else
   {
      ret = EFMI_CALL_INVALID;
   }
   return ret;
}

FmdStatus
fmd_dump_memory(const char *description,
                OSADDRESS start_address,
                uint32_t length)
{
   FmdStatus ret;

   if (lock(0))
   {
#ifdef FMD_DO_NOT_SWITCH_STACK
      ret = dump_memory((uint32_t)description, start_address, length, 0, 0, 0);
      /* If error is due to insufficient memory,
       * dump text block stating the same
       */
      if (ret == RAMDUMP_ECURR_DUMP_TOO_LARGE)
      {
         ramdump_private_dump_text_info(DUMP_INCOMPLETE);
      }

#else
      switch_stack_args.arg1 = (uint32_t)description;
      switch_stack_args.arg2 = (uint32_t)start_address;
      switch_stack_args.arg3 = (uint32_t)length;
      switch_stack_args.arg4 = 0;
      switch_stack_args.arg5 = 0;
      switch_stack_args.arg6 = 0;

      ret = (FmdStatus)fmm_switch_stack(fmm_stack,
                                          dump_memory,
                                          &switch_stack_args);
      /* If error is due to insufficient memory,
       * dump text block stating the same
       */
      if (ret == RAMDUMP_ECURR_DUMP_TOO_LARGE)
      {
         ramdump_private_dump_text_info(DUMP_INCOMPLETE);
      }

#endif
      unlock(0);
   }
   else
   {
      ret = EFMI_CALL_INVALID;
   }
   return ret;
}

FmdStatus
fmd_dump_string(const char *description)
{
   FmdStatus ret;

   if (lock(0))
   {
#ifdef FMD_DO_NOT_SWITCH_STACK
      ret = dump_string((uint32_t)description, 0, 0, 0, 0, 0);
      /* If error is due to insufficient memory,
       * dump text block stating the same
       */
      if (ret == RAMDUMP_ECURR_DUMP_TOO_LARGE)
      {
         ramdump_private_dump_text_info(DUMP_INCOMPLETE);
      }

#else
      switch_stack_args.arg1 = (uint32_t)description;
      switch_stack_args.arg2 = 0;
      switch_stack_args.arg3 = 0;
      switch_stack_args.arg4 = 0;
      switch_stack_args.arg5 = 0;
      switch_stack_args.arg6 = 0;

      ret = (FmdStatus)fmm_switch_stack(fmm_stack,
                                          dump_string,
                                          &switch_stack_args);
      /* If error is due to insufficient memory,
       * dump text block stating the same
       */
      if (ret == RAMDUMP_ECURR_DUMP_TOO_LARGE)
      {
         ramdump_private_dump_text_info(DUMP_INCOMPLETE);
      }

#endif
      unlock(0);
   }
   else
   {
      ret = EFMI_CALL_INVALID;
   }
   return ret;
}

void
fmd_start_handler1(void)
{
   ramdump_start_handler1();
}

FmdStatus
fmd_set_watchdog(void (*function)(void))
{
   return ramdump_set_watchdog_function(function);
}

FmdStatus
fmd_dump_process_info(FmdScope scope, FmdInfo info)
{
   FmdStatus ret;

   if (lock(0))
   {
#ifdef FMD_DO_NOT_SWITCH_STACK
      ret = dump_process_info(scope, info, 0, 0, 0, 0);
      /* If error is due to insufficient memory,
       * dump text block stating the same
       */
      if (ret == RAMDUMP_ECURR_DUMP_TOO_LARGE)
      {
         ramdump_private_dump_text_info(DUMP_INCOMPLETE);
      }

#else
      switch_stack_args.arg1 = (uint32_t)scope;
      switch_stack_args.arg2 = (uint32_t)info;
      switch_stack_args.arg3 = 0;
      switch_stack_args.arg4 = 0;
      switch_stack_args.arg5 = 0;
      switch_stack_args.arg6 = 0;

      ret = (FmdStatus)fmm_switch_stack(fmm_stack,
                                          dump_process_info,
                                          &switch_stack_args);
      /* If error is due to insufficient memory,
       * dump text block stating the same
       */
      if (ret == RAMDUMP_ECURR_DUMP_TOO_LARGE)
      {
         ramdump_private_dump_text_info(DUMP_INCOMPLETE);
      }

#endif
      unlock(0);
   }
   else
   {
      ret = EFMI_CALL_INVALID;
   }
   return ret;
}

FmdStatus
fmd_dump_stack_usage_info(FmdScope scope, FmdInfo info)
{
   FmdStatus ret;

   if (lock(0))
   {
#ifdef FMD_DO_NOT_SWITCH_STACK
      ret = dump_stack_usage_info(scope, info, 0, 0, 0, 0);
      /* If error is due to insufficient memory,
       * dump text block stating the same
       */
      if (ret == RAMDUMP_ECURR_DUMP_TOO_LARGE)
      {
         ramdump_private_dump_text_info(DUMP_INCOMPLETE);
      }

#else
      switch_stack_args.arg1 = (uint32_t)scope;
      switch_stack_args.arg2 = (uint32_t)info;
      switch_stack_args.arg3 = 0;
      switch_stack_args.arg4 = 0;
      switch_stack_args.arg5 = 0;
      switch_stack_args.arg6 = 0;

      ret = (FmdStatus)fmm_switch_stack(fmm_stack,
                                          dump_stack_usage_info,
                                          &switch_stack_args);
      /* If error is due to insufficient memory,
       * dump text block stating the same
       */
      if (ret == RAMDUMP_ECURR_DUMP_TOO_LARGE)
      {
         ramdump_private_dump_text_info(DUMP_INCOMPLETE);
      }

#endif
      unlock(0);
   }
   else
   {
      ret = EFMI_CALL_INVALID;
   }
   return ret;
}

FmdStatus
fmd_dump_env_vars_info(FmdScope scope, FmdInfo info)
{
   FmdStatus ret;

   if (lock(0))
   {
#ifdef FMD_DO_NOT_SWITCH_STACK
      ret = dump_env_vars_info(scope, info, 0, 0, 0, 0);
      /* If error is due to insufficient memory,
       * dump text block stating the same
       */
      if (ret == RAMDUMP_ECURR_DUMP_TOO_LARGE)
      {
         ramdump_private_dump_text_info(DUMP_INCOMPLETE);
      }

#else
      switch_stack_args.arg1 = (uint32_t)scope;
      switch_stack_args.arg2 = (uint32_t)info;
      switch_stack_args.arg3 = 0;
      switch_stack_args.arg4 = 0;
      switch_stack_args.arg5 = 0;
      switch_stack_args.arg6 = 0;

      ret = (FmdStatus)fmm_switch_stack(fmm_stack,
                                          dump_env_vars_info,
                                          &switch_stack_args);
      /* If error is due to insufficient memory,
       * dump text block stating the same
       */
      if (ret == RAMDUMP_ECURR_DUMP_TOO_LARGE)
      {
         ramdump_private_dump_text_info(DUMP_INCOMPLETE);
      }

#endif
      unlock(0);
   }
   else
   {
      ret = EFMI_CALL_INVALID;
   }
   return ret;
}

FmdStatus
fmd_dump_signal_queue_info(FmdScope scope, FmdInfo info)
{
   FmdStatus ret;

   if (lock(0))
   {
#ifdef FMD_DO_NOT_SWITCH_STACK
      ret = dump_signal_queue_info(scope, info, 0, 0, 0, 0);
      /* If error is due to insufficient memory,
       * dump text block stating the same
       */
      if (ret == RAMDUMP_ECURR_DUMP_TOO_LARGE)
      {
         ramdump_private_dump_text_info(DUMP_INCOMPLETE);
      }

#else
      switch_stack_args.arg1 = (uint32_t)scope;
      switch_stack_args.arg2 = (uint32_t)info;
      switch_stack_args.arg3 = 0;
      switch_stack_args.arg4 = 0;
      switch_stack_args.arg5 = 0;
      switch_stack_args.arg6 = 0;

      ret = (FmdStatus)fmm_switch_stack(fmm_stack,
                                          dump_signal_queue_info,
                                          &switch_stack_args);
      /* If error is due to insufficient memory,
       * dump text block stating the same
       */
      if (ret == RAMDUMP_ECURR_DUMP_TOO_LARGE)
      {
         ramdump_private_dump_text_info(DUMP_INCOMPLETE);
      }

#endif
      unlock(0);
   }
   else
   {
      ret = EFMI_CALL_INVALID;
   }
   return ret;
}

FmdStatus
fmd_dump_pool_info(FmdScope scope, FmdInfo info)
{
   FmdStatus ret;

   if (lock(0))
   {
#ifdef FMD_DO_NOT_SWITCH_STACK
      ret = dump_pool_info(scope, info, 0, 0, 0, 0);
      /* If error is due to insufficient memory,
       * dump text block stating the same
       */
      if (ret == RAMDUMP_ECURR_DUMP_TOO_LARGE)
      {
         ramdump_private_dump_text_info(DUMP_INCOMPLETE);
      }

#else
      switch_stack_args.arg1 = (uint32_t)scope;
      switch_stack_args.arg2 = (uint32_t)info;
      switch_stack_args.arg3 = 0;
      switch_stack_args.arg4 = 0;
      switch_stack_args.arg5 = 0;
      switch_stack_args.arg6 = 0;

      ret = (FmdStatus)fmm_switch_stack(fmm_stack,
                                          dump_pool_info,
                                          &switch_stack_args);
      /* If error is due to insufficient memory,
       * dump text block stating the same
       */
      if (ret == RAMDUMP_ECURR_DUMP_TOO_LARGE)
      {
         ramdump_private_dump_text_info(DUMP_INCOMPLETE);
      }

#endif
      unlock(0);
   }
   else
   {
      ret = EFMI_CALL_INVALID;
   }
   return ret;
}

FmdStatus
fmd_dump_pool_signals_info(FmdScope scope, FmdInfo info)
{
   FmdStatus ret;

   if (lock(0))
   {
#ifdef FMD_DO_NOT_SWITCH_STACK
      ret = dump_pool_signals_info(scope, info, 0, 0, 0, 0);
      /* If error is due to insufficient memory,
       * dump text block stating the same
       */
      if (ret == RAMDUMP_ECURR_DUMP_TOO_LARGE)
      {
         ramdump_private_dump_text_info(DUMP_INCOMPLETE);
      }

#else
      switch_stack_args.arg1 = (uint32_t)scope;
      switch_stack_args.arg2 = (uint32_t)info;
      switch_stack_args.arg3 = 0;
      switch_stack_args.arg4 = 0;
      switch_stack_args.arg5 = 0;
      switch_stack_args.arg6 = 0;

      ret = (FmdStatus)fmm_switch_stack(fmm_stack,
                                          dump_pool_signals_info,
                                          &switch_stack_args);
      /* If error is due to insufficient memory,
       * dump text block stating the same
       */
      if (ret == RAMDUMP_ECURR_DUMP_TOO_LARGE)
      {
         ramdump_private_dump_text_info(DUMP_INCOMPLETE);
      }

#endif
      unlock(0);
   }
   else
   {
      ret = EFMI_CALL_INVALID;
   }
   return ret;
}

FmdStatus
fmd_dump_sys_param_info(FmdScope scope, FmdInfo info)
{
#if 0
   FmdStatus ret;

   if (lock(0))
   {
#ifdef FMD_DO_NOT_SWITCH_STACK
      ret = dump_sys_param_info(scope, info, 0, 0, 0, 0);
      /* If error is due to insufficient memory,
       * dump text block stating the same
       */
      if (ret == RAMDUMP_ECURR_DUMP_TOO_LARGE)
      {
         ramdump_private_dump_text_info(DUMP_INCOMPLETE);
      }

#else
      switch_stack_args.arg1 = (uint32_t)scope;
      switch_stack_args.arg2 = (uint32_t)info;
      switch_stack_args.arg3 = 0;
      switch_stack_args.arg4 = 0;
      switch_stack_args.arg5 = 0;
      switch_stack_args.arg6 = 0;

      ret = (FmdStatus)fmm_switch_stack(fmm_stack,
                                          dump_sys_param_info,
                                          &switch_stack_args);
      /* If error is due to insufficient memory,
       * dump text block stating the same
       */
      if (ret == RAMDUMP_ECURR_DUMP_TOO_LARGE)
      {
         ramdump_private_dump_text_info(DUMP_INCOMPLETE);
      }

#endif
      unlock(0);
   }
   else
   {
      ret = EFMI_CALL_INVALID;
   }
   return ret;
#endif

   return EFMI_OK;
}

FmdStatus
fmd_dump_registers_info(FmdScope scope, FmdInfo info, struct cpu_regdump *regdump)
{
   FmdStatus ret;

   if (lock(0))
   {
#ifdef FMD_DO_NOT_SWITCH_STACK
      ret = dump_registers_info(scope, info, 0, 0, 0, 0);
      /* If error is due to insufficient memory,
       * dump text block stating the same
       */
      if (ret == RAMDUMP_ECURR_DUMP_TOO_LARGE)
      {
         ramdump_private_dump_text_info(DUMP_INCOMPLETE);
      }

#else
      switch_stack_args.arg1 = (uint32_t)scope;
      switch_stack_args.arg2 = (uint32_t)info;
      switch_stack_args.arg3 = (uint32_t)regdump;
      switch_stack_args.arg4 = 0;
      switch_stack_args.arg5 = 0;
      switch_stack_args.arg6 = 0;

      ret = (FmdStatus)fmm_switch_stack(fmm_stack,
                                          dump_registers_info,
                                          &switch_stack_args);
      /* If error is due to insufficient memory,
       * dump text block stating the same
       */
      if (ret == RAMDUMP_ECURR_DUMP_TOO_LARGE)
      {
         ramdump_private_dump_text_info(DUMP_INCOMPLETE);
      }

#endif
      unlock(0);
   }
   else
   {
      ret = EFMI_CALL_INVALID;
   }
   return ret;
}

FmdStatus
fmd_dump_program_info(FmdScope scope, FmdInfo info)
{
   FmdStatus ret;

   if (lock(0))
   {
#ifdef FMD_DO_NOT_SWITCH_STACK
      ret = dump_program_info(scope, info, 0, 0, 0, 0);
      /* If error is due to insufficient memory,
       * dump text block stating the same
       */
      if (ret == RAMDUMP_ECURR_DUMP_TOO_LARGE)
      {
         ramdump_private_dump_text_info(DUMP_INCOMPLETE);
      }

#else
      switch_stack_args.arg1 = (uint32_t)scope;
      switch_stack_args.arg2 = (uint32_t)info;
      switch_stack_args.arg3 = 0;
      switch_stack_args.arg4 = 0;
      switch_stack_args.arg5 = 0;
      switch_stack_args.arg6 = 0;

      ret = (FmdStatus)fmm_switch_stack(fmm_stack,
                                          dump_program_info,
                                          &switch_stack_args);
      /* If error is due to insufficient memory,
       * dump text block stating the same
       */
      if (ret == RAMDUMP_ECURR_DUMP_TOO_LARGE)
      {
         ramdump_private_dump_text_info(DUMP_INCOMPLETE);
      }

#endif
      unlock(0);
   }
   else
   {
      ret = PFMI_CALL_INVALID;
   }
   return ret;
}

FmdStatus
fmd_dump_load_module_info(FmdScope scope, FmdInfo info)
{
   FmdStatus ret;

   if (lock(0))
   {
#ifdef FMD_DO_NOT_SWITCH_STACK
      ret = dump_load_module_info(scope, info, 0, 0, 0, 0);
      /* If error is due to insufficient memory,
       * dump text block stating the same
       */
      if (ret == RAMDUMP_ECURR_DUMP_TOO_LARGE)
      {
         ramdump_private_dump_text_info(DUMP_INCOMPLETE);
      }

#else
      switch_stack_args.arg1 = (uint32_t)scope;
      switch_stack_args.arg2 = (uint32_t)info;
      switch_stack_args.arg3 = 0;
      switch_stack_args.arg4 = 0;
      switch_stack_args.arg5 = 0;
      switch_stack_args.arg6 = 0;

      ret = (FmdStatus)fmm_switch_stack(fmm_stack,
                                          dump_load_module_info,
                                          &switch_stack_args);
      /* If error is due to insufficient memory,
       * dump text block stating the same
       */
      if (ret == RAMDUMP_ECURR_DUMP_TOO_LARGE)
      {
         ramdump_private_dump_text_info(DUMP_INCOMPLETE);
      }

#endif
      unlock(0);
   }
   else
   {
      ret = PFMI_CALL_INVALID;
   }
   return ret;
}

FmdStatus
fmd_dump_trace_info(FmdScope scope, FmdInfo info)
{
   FmdStatus ret;

   if (lock(0))
   {
#ifdef FMD_DO_NOT_SWITCH_STACK
      ret = dump_trace_info(scope, info, 0, 0, 0, 0);
      /* If error is due to insufficient memory,
       * dump text block stating the same
       */
      if (ret == RAMDUMP_ECURR_DUMP_TOO_LARGE)
      {
         ramdump_private_dump_text_info(DUMP_INCOMPLETE);
      }

#else
      switch_stack_args.arg1 = (uint32_t)scope;
      switch_stack_args.arg2 = (uint32_t)info;
      switch_stack_args.arg3 = 0;
      switch_stack_args.arg4 = 0;
      switch_stack_args.arg5 = 0;
      switch_stack_args.arg6 = 0;

      ret = (FmdStatus)fmm_switch_stack(fmm_stack,
                                          dump_trace_info,
                                          &switch_stack_args);
      /* If error is due to insufficient memory,
       * dump text block stating the same
       */
      if (ret == RAMDUMP_ECURR_DUMP_TOO_LARGE)
      {
         ramdump_private_dump_text_info(DUMP_INCOMPLETE);
      }

#endif
      unlock(0);
   }
   else
   {
      ret = EFMI_CALL_INVALID;
   }
   return ret;
}

FmdStatus
fmd_dump_pool_memory(FmdScope scope, FmdMemory memory,
                     struct cpu_regdump *regdump)
{
   FmdStatus ret;

   if (lock(0))
   {
#ifdef FMD_DO_NOT_SWITCH_STACK
      ret = dump_pool_memory(scope, memory, 0, 0, 0, 0);
      /* If error is due to insufficient memory,
       * dump text block stating the same
       */
      if (ret == RAMDUMP_ECURR_DUMP_TOO_LARGE)
      {
         ramdump_private_dump_text_info(DUMP_INCOMPLETE);
      }

#else
      switch_stack_args.arg1 = (uint32_t)scope;
      switch_stack_args.arg2 = (uint32_t)memory;
      switch_stack_args.arg3 = (uint32_t)regdump;
      switch_stack_args.arg4 = 0;
      switch_stack_args.arg5 = 0;
      switch_stack_args.arg6 = 0;

      ret = (FmdStatus)fmm_switch_stack(fmm_stack,
                                          dump_pool_memory,
                                          &switch_stack_args);
      /* If error is due to insufficient memory,
       * dump text block stating the same
       */
      if (ret == RAMDUMP_ECURR_DUMP_TOO_LARGE)
      {
         ramdump_private_dump_text_info(DUMP_INCOMPLETE);
      }

#endif
      unlock(0);
   }
   else
   {
      ret = EFMI_CALL_INVALID;
   }
   return ret;
}

FmdStatus
fmd_dump_heap_memory(FmdScope scope, FmdMemory memory)
{
   FmdStatus ret;

   if (lock(0))
   {
#ifdef FMD_DO_NOT_SWITCH_STACK
      ret = dump_heap_memory(scope, memory, 0, 0, 0, 0);
      /* If error is due to insufficient memory,
       * dump text block stating the same
       */
      if (ret == RAMDUMP_ECURR_DUMP_TOO_LARGE)
      {
         ramdump_private_dump_text_info(DUMP_INCOMPLETE);
      }

#else
      switch_stack_args.arg1 = (uint32_t)scope;
      switch_stack_args.arg2 = (uint32_t)memory;
      switch_stack_args.arg3 = 0;
      switch_stack_args.arg4 = 0;
      switch_stack_args.arg5 = 0;
      switch_stack_args.arg6 = 0;

      ret = (FmdStatus)fmm_switch_stack(fmm_stack,
                                          dump_heap_memory,
                                          &switch_stack_args);
      /* If error is due to insufficient memory,
       * dump text block stating the same
       */
      if (ret == RAMDUMP_ECURR_DUMP_TOO_LARGE)
      {
         ramdump_private_dump_text_info(DUMP_INCOMPLETE);
      }

#endif
      unlock(0);
   }
   else
   {
      ret = HFMI_CALL_INVALID;
   }
   return ret;
}

FmdStatus
fmd_dump_program_memory(FmdScope scope, FmdInfo memory)
{
   FmdStatus ret;

   if (lock(0))
   {
#ifdef FMD_DO_NOT_SWITCH_STACK
      ret = dump_program_memory(scope, memory, 0, 0, 0, 0);
      /* If error is due to insufficient memory,
       * dump text block stating the same
       */
      if (ret == RAMDUMP_ECURR_DUMP_TOO_LARGE)
      {
         ramdump_private_dump_text_info(DUMP_INCOMPLETE);
      }

#else
      switch_stack_args.arg1 = (uint32_t)scope;
      switch_stack_args.arg2 = (uint32_t)memory;
      switch_stack_args.arg3 = 0;
      switch_stack_args.arg4 = 0;
      switch_stack_args.arg5 = 0;
      switch_stack_args.arg6 = 0;

      ret = (FmdStatus)fmm_switch_stack(fmm_stack,
                                          dump_program_memory,
                                          &switch_stack_args);
      /* If error is due to insufficient memory,
       * dump text block stating the same
       */
      if (ret == RAMDUMP_ECURR_DUMP_TOO_LARGE)
      {
         ramdump_private_dump_text_info(DUMP_INCOMPLETE);
      }

#endif
      unlock(0);
   }
   else
   {
      ret = PFMI_CALL_INVALID;
   }
   return ret;
}
static uint32_t set_property(uint32_t property, uint32_t value, uint32_t a2, uint32_t a3, uint32_t a4, uint32_t a5)
{
   uint32_t ret = 0;

   switch (property)
   {
      case FMD_NUM_LONG_QUEUE_PROCS:
         /* free the ramdump heap allocated earlier */
         ramdump_heap_free(fmd_properties.proc_queue_list);
         
         /* Allocate ramdump heap memory - Actual size + 1 */
         fmd_properties.proc_queue_list =
            (FmmProcQueueCnt *)ramdump_heap_malloc(
                         sizeof(FmmProcQueueCnt)* (value+1));
         if (fmd_properties.proc_queue_list != NULL)
         {
            fmd_properties.num_long_queue_procs = value;
            fmm_efmi_set_fmd_properties(&fmd_properties);
         }
	 else
	 {
	    fmd_properties.num_long_queue_procs = 0;
            fmm_efmi_set_fmd_properties(&fmd_properties);	
 	    ret = 1;
	 }
         break;

      case FMD_MAX_SIGNALS_IN_QUEUE:
         fmd_properties.max_signals_in_queue = value;
         fmm_efmi_set_fmd_properties(&fmd_properties);
         break;

      case FMD_MAX_STACK_TOP_SIZE:  
         fmd_properties.max_stack_top_size = value;
         fmm_efmi_set_fmd_properties(&fmd_properties);
         break;
   }
   return ret;
}

uint32_t fmd_set_property(FmdProperty property, uint32_t value)
{
   uint32_t ret;
   if (lock(0))
   {
#ifdef FMD_DO_NOT_SWITCH_STACK
      ret = set_property(property, value, 0, 0, 0, 0);

#else
      switch_stack_args.arg1 = (uint32_t)property;
      switch_stack_args.arg2 = (uint32_t)value;
      switch_stack_args.arg3 = 0;
      switch_stack_args.arg4 = 0;
      switch_stack_args.arg5 = 0;
      switch_stack_args.arg6 = 0;

      ret = (FmdStatus)fmm_switch_stack(fmm_stack,
                                          set_property,
                                          &switch_stack_args);

#endif
      unlock(0);
   }
   else
   {
      ret = 1;
   }
   return ret;
}
static uint32_t end_of_dump(uint32_t status, uint32_t a1, uint32_t a2, uint32_t a3, uint32_t a4, uint32_t a5)
{
   FmdStatus ret = EFMI_OK;

   if (status)
   {
      ret = ramdump_private_dump_text_info(END_OF_DUMP_FAILURE);
   }
   else
   {
      ret = ramdump_private_dump_text_info(END_OF_DUMP_SUCCESS);
   }
   return ret;
}

FmdStatus fmd_end_of_dump(uint32_t status)
{
   FmdStatus ret;

   if (lock(0))
   {
#ifdef FMD_DO_NOT_SWITCH_STACK
      ret = end_of_dump(status, 0, 0, 0, 0, 0);
#else
      switch_stack_args.arg1 = (uint32_t)status;
      switch_stack_args.arg2 = 0;
      switch_stack_args.arg3 = 0;
      switch_stack_args.arg4 = 0;
      switch_stack_args.arg5 = 0;
      switch_stack_args.arg6 = 0;

      ret = (FmdStatus)fmm_switch_stack(fmm_stack,
                                          end_of_dump,
                                          &switch_stack_args);

#endif
      unlock(0);
   }
   else
   {
      ret =  EFMI_CALL_INVALID;
   }
   return ret;
}
