/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2006 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

#include "stdint.h"
#include "string.h"

#include "pfmi.h"

#include "pm.sig"

#include "fmm.h"
#include "fmm_pfmi.h"

#define assert(x) if (!(x)) return PFMI_UNAVAILABLE;

union Signal
{
   uint32_t signo;
   struct PmLoadModuleInstallHandlesRequest load_module_install_handles_request;
   struct PmLoadModuleInstallHandlesReply   load_module_install_handles_reply;
   struct PmLoadModuleInfoRequest           load_module_info_request;
   struct PmLoadModuleInfoReply             load_module_info_reply;
   struct PmLoadModuleSectionInfoRequest    load_module_section_info_request;   
   struct PmLoadModuleSectionInfoReply      load_module_section_info_reply;
   struct PmProgramsReply                   programs_reply;
   struct PmProgramInfo                     program_info;
   struct PmGetProgramPid                   get_program_pid;
};  

struct FmmSignal
{
   uint32_t size;
   union Signal data;
};

static PmStatus
map_status(PfmiStatus status)
{
   /* Map pfmi status code to pm status code */
   switch (status)
   {
      case PFMI_OK:
         return PM_SUCCESS;
      case PFMI_END_OF_LIST:
         return PM_SUCCESS;
      case PFMI_EINSTHND:
         return PM_EINSTALL_HANDLE;
      case PFMI_EPROGPID:
         return PM_EPROGPID;
      case PFMI_UNAVAILABLE:
         return PM_ENO_PROGRAM_MANAGER;
      default:
         /* FIXME: What status value should be used here? */
         return PM_ENO_PROGRAM_MANAGER;
   }
}

PfmiStatus
fmm_pfmi_load_module_install_handles_request(
                     int32_t (*reply)(FmmSignal *sig, int32_t status),
                     const struct PmLoadModuleInstallHandlesRequest *request)
{
   PfmiStatus ret = PFMI_UNAVAILABLE;
   FmmSignal *sig;
   struct PmLoadModuleInstallHandlesReply *handles;
   struct PfmiLoadModuleCursor cursor;
   struct PfmiLoadModuleInfo info;
   
   assert(reply);
   assert(request);
   assert(request->sigNo == PM_LOAD_MODULE_INSTALL_HANDLES_REQUEST);
   
   /* Init signal */
   sig = fmm_output_signal;
   handles = &sig->data.load_module_install_handles_reply;
   
   /* Fill out signal struct */
   handles->sigNo = PM_LOAD_MODULE_INSTALL_HANDLES_REPLY;
   
   handles->install_handles_no_of_strings = 0;
   handles->install_handles_size = 0;
   handles->install_handles[0] = 0;
   
   ret = ose_pfmi_get_first_load_module(&cursor, &info);
   
   while (ret == PFMI_OK)
   {
      uint32_t size = strlen(info.install_handle) + 1;
      strcpy(handles->install_handles + handles->install_handles_size,
             info.install_handle);
      handles->install_handles_size += size;
      handles->install_handles_no_of_strings++;
      
      ret = ose_pfmi_get_next_load_module(&cursor, &info);
   }
   
   /* Array is ended with a '\0' */
   handles->install_handles[handles->install_handles_size++] = '\0';
   
   /* Map pfmi status code to pm status code */
   handles->status = map_status(ret);
   
   sig->size = offsetof(struct PmLoadModuleInstallHandlesReply, install_handles) 
               + handles->install_handles_size;
   
   /* Pass on reply signal to reply callback. */
   if (reply(sig, PFMI_END_OF_LIST ? PFMI_OK : ret))
   {
      return PFMI_END_OF_LIST;
   }
      
   return ret == PFMI_END_OF_LIST ? PFMI_OK : ret;
}

PfmiStatus 
fmm_pfmi_load_module_info_request(
                                 int32_t (*reply)(FmmSignal *sig, int32_t status),
                                 const struct PmLoadModuleInfoRequest *request)
{
   PfmiStatus ret = PFMI_UNAVAILABLE;
   FmmSignal *sig;
   struct PmLoadModuleInfoReply *module;
   struct PfmiLoadModuleInfo info;
   
   assert(reply);
   assert(request);
   assert(request->sigNo == PM_LOAD_MODULE_INFO_REQUEST);
   
   /* Init signal */
   sig = fmm_output_signal;
   module = &sig->data.load_module_info_reply;
   sig->size = sizeof(*module);
   
   /* Fill out signal struct */
   module->sigNo = PM_LOAD_MODULE_INFO_REPLY;
      
   ret = ose_pfmi_get_load_module_info((char *)request->install_handle, &info);

   if (ret == PFMI_OK)
   {
      memcpy(module->install_handle, info.install_handle,
             PM_INSTALL_HANDLE_LENGTH);
      module->entrypoint = info.entrypoint;
      module->options = info.options;
      module->text_base = info.text_base;
      module->text_size = info.text_size;
      module->data_base = info.data_base;
      module->data_size = info.data_size;
      module->no_of_sections = info.no_of_sections;
      module->no_of_instances = info.no_of_instances;
      memcpy(module->file_format, info.file_format,
             PM_FILE_FORMAT_NAME_LENGTH);
      memcpy(module->file_name, info.file_name,
             PM_PARENT_FILE_NAME_SIZE);
   }
   
   /* Map pfmi status code to pm status code */
   module->status = map_status(ret);
            
   /* Pass on reply signal to reply callback. */
   if (reply(sig, ret))
   {
      return PFMI_END_OF_LIST;
   }
   
   return ret;
}

PfmiStatus
fmm_pfmi_load_module_section_info_request(
                        int32_t (*reply)(FmmSignal *sig, int32_t status),
                        const struct PmLoadModuleSectionInfoRequest *request)
{
   PfmiStatus ret = PFMI_UNAVAILABLE;
   FmmSignal *sig;
   struct PmLoadModuleSectionInfoReply *sections;
   uint32_t length = 1;
   struct PfmiLoadModuleInfo module_info;
   struct PfmiLoadModuleSectionInfo section_info;
   
   assert(reply);
   assert(request);
   assert(request->sigNo == PM_LOAD_MODULE_SECTION_INFO_REQUEST);
   
   /* Init signal */
   sig = fmm_output_signal;
   sections = &sig->data.load_module_section_info_reply;

   /* Fill out signal struct */
   sections->sigNo = PM_LOAD_MODULE_SECTION_INFO_REPLY;
   sections->status = PM_SUCCESS; /* ...until found unsuccessful. */
   sections->section_interval_start = request->section_interval_start;
   sections->section_interval_end = request->section_interval_end;
   
   ret = ose_pfmi_get_load_module_info((char *)request->install_handle, &module_info);

   if (ret == PFMI_OK)
   {
      if (request->section_interval_start > request->section_interval_end ||
          request->section_interval_start >= module_info.no_of_sections ||
          request->section_interval_end >= module_info.no_of_sections)
      {
         sections->status = PM_ESECTION_INTERVAL;
      }
      else
      {
         uint32_t i;
         
         length = sections->section_interval_end -
                  sections->section_interval_start + 1;
         
         for (i = 0; i < length; i++)
         {
            struct PmLoadModuleSectionInfo *sect;
            
            ret = ose_pfmi_get_load_module_section_info((char *)request->install_handle,
                                                        i,
                                                        &section_info);
            if (ret != PFMI_OK)
            {
               sections->status = PM_ENO_PROGRAM_MANAGER;
               break;
            }
                        
            sect = &sections->sections[i];
            sect->section_base = section_info.section_base;
            sect->section_size = section_info.section_size;
            sect->section_attr = section_info.section_attr;
            sect->section_options = section_info.section_options;
            memcpy(sect->section_name, section_info.section_name,
                   PM_SECTION_NAME_LENGTH);
         }
      }
   }
   else
   {
      sections->status = PM_ENO_PROGRAM_MANAGER;
   }
   
   sig->size = offsetof(struct PmLoadModuleSectionInfoReply, sections) 
               + length * sizeof(struct PmLoadModuleSectionInfo);
               
   /* Pass on reply signal to reply callback. */
   if (reply(sig, ret))
   {
      return PFMI_END_OF_LIST;
   }
      
   return ret;
}

PfmiStatus
fmm_pfmi_programs_request(int32_t (*reply)(FmmSignal *sig, int32_t status),
                          const struct PmProgramsRequest *request)
{
   PfmiStatus ret;
   FmmSignal *sig;
   struct PmProgramsReply *programs;
   struct PfmiProgramCursor cursor;
   struct PfmiProgramInfo info;
   
   assert(reply);
   assert(request);
   assert(request->sigNo == PM_PROGRAMS_REQUEST);
            
   /* Init signal */
   sig = fmm_output_signal;
   programs = &sig->data.programs_reply;
   
   /* Fill out signal struct */
   programs->sigNo = PM_PROGRAMS_REPLY;
   programs->progpids_size = 0;
   programs->progpids[0] = 0;

   ret = ose_pfmi_get_first_program(&cursor, &info);

   while (ret == PFMI_OK)
   {
      programs->progpids[programs->progpids_size++] = info.progpid;
      ret = ose_pfmi_get_next_program(&cursor, &info);
   }

   /* Map pfmi status code to pm status code */
   programs->status = map_status(ret);
   
   /* Calculate size of signal as sum of size of signal struct 
      and size of progpids variable size field. */
   sig->size = offsetof(struct PmProgramsReply, progpids) + 
               programs->progpids_size * sizeof(PROCESS);
   
   /* Pass on reply signal to reply callback. */
   if (reply(sig, PFMI_END_OF_LIST ? PFMI_OK : ret))
   {
      return PFMI_END_OF_LIST;
   }
   
   return ret == PFMI_END_OF_LIST ? PFMI_OK : ret;
}

PfmiStatus
fmm_pfmi_program_info_request(
                    int32_t (*reply)(FmmSignal *sig, int32_t status),
                    const struct PmProgramInfo *request)
{
   PfmiStatus ret = PFMI_UNAVAILABLE;
   FmmSignal *sig;
   struct PmProgramInfo *program;
   struct PfmiProgramInfo info;
   
   assert(reply);
   assert(request);
   assert(request->sigNo == PM_PROGRAM_INFO_REQUEST);
   
   /* Init signal */
   sig = fmm_output_signal;
   program = &sig->data.program_info;
   sig->size = sizeof(*program);
   
   /* Fill out signal struct */
   program->sigNo = PM_PROGRAM_INFO_REPLY;
      
   ret = ose_pfmi_get_program_info(request->progpid, &info);
   if (ret == PFMI_OK)
   {
      memcpy(program->install_handle, info.install_handle,
             PM_INSTALL_HANDLE_LENGTH);
      program->progpid = info.progpid;
      program->domain = info.domain;
      program->main_block = info.main_block;
      program->main_process = info.main_process;
      program->pool_base = info.stk_pool_base;
      program->pool_size = info.stk_pool_size;
      program->uid = info.uid;
      program->state = info.state;
      program->segpid = info.segpid;
      program->stk_poolid = info.stk_poolid;
      program->sig_poolid = info.sig_poolid;
      program->sig_pool_base = info.sig_pool_base;
      program->sig_pool_size = info.sig_pool_size;
      program->heap = info.heap;
   }

   /* Map pfmi status code to pm status code */
   program->status = map_status(ret);
            
   /* Pass on reply signal to reply callback. */
   if (reply(sig, ret))
   {
      ret = PFMI_END_OF_LIST;
   }
         
   return ret;
}

PfmiStatus
fmm_pfmi_get_program_pid_request(
                    int32_t (*reply)(FmmSignal *sig, int32_t status),
                    const struct PmGetProgramPid *request)
{
   PfmiStatus ret = PFMI_UNAVAILABLE;
   FmmSignal *sig;
   struct PmGetProgramPid *program;
   
   assert(reply);
   assert(request);
   assert(request->sigNo == PM_GET_PROGRAM_PID_REQUEST);
   
   /* Init signal */
   sig = fmm_output_signal;
   program = &sig->data.get_program_pid;
   sig->size = sizeof(*program);
   
   /* Fill out signal struct */
   program->sigNo = PM_GET_PROGRAM_PID_REPLY;
   program->pid = request->pid;
   
   ret = ose_pfmi_get_progpid(request->pid, &program->progpid);

   /* Map pfmi status code to pm status code */
   program->status = map_status(ret);
   
   /* Pass on reply signal to reply callback. */
   if (reply(sig, ret))
   {
      ret = PFMI_END_OF_LIST;
   }
   
   return ret;
}

PfmiStatus 
fmm_pfmi_program_info(int32_t (*request)(FmmSignal *sig),
                      int32_t (*reply)(FmmSignal *sig, int32_t status))
{
   PfmiStatus ret;
   
   struct PfmiProgramCursor cursor;
   struct PfmiProgramInfo info;
   
   assert(reply);
   assert(request);

   ret = ose_pfmi_get_first_program(&cursor, &info);
   
   while (!ret)
   {
      FmmSignal *sig;
      struct PmProgramInfo *program;
   
      sig = fmm_output_signal;
      program = &sig->data.program_info;
      program->sigNo = PM_PROGRAM_INFO_REQUEST;
      program->progpid = info.progpid;
      program->status = PM_SUCCESS;
      sig->size = sizeof (*program);

      if (request(sig))
      {
         return PFMI_END_OF_LIST;              
      }

      program->sigNo = PM_PROGRAM_INFO_REPLY;
      memcpy(program->install_handle, 
             info.install_handle, 
             PM_INSTALL_HANDLE_LENGTH);
      program->domain = info.domain;
      program->main_block = info.main_block;
      program->main_process = info.main_process;
      program->pool_base = info.stk_pool_base;
      program->pool_size = info.stk_pool_size;
      program->uid = info.uid;
      program->state = info.state;
      program->segpid = info.segpid;
      program->stk_poolid = info.stk_poolid;
      program->sig_poolid = info.sig_poolid;
      program->sig_pool_base = info.sig_pool_base;
      program->sig_pool_size = info.sig_pool_size;
      program->heap = info.heap;
      sig->size = sizeof (*program);

      if (reply(sig, ret))
      {
         return PFMI_END_OF_LIST;
      }
      
      if (!ret)
      {
         ret = ose_pfmi_get_next_program(&cursor, &info);
      }       
   } 
   
   return ret == PFMI_END_OF_LIST ? PFMI_OK : ret;       
}

PfmiStatus fmm_pfmi_load_module_info(
                    int32_t (*request)(FmmSignal *sig),
                    int32_t (*reply)(FmmSignal *sig, int32_t status))
{
   PfmiStatus ret;
   
   struct PfmiLoadModuleCursor cursor;
   struct PfmiLoadModuleInfo info;
   
   assert(reply);
   assert(request);

   ret = ose_pfmi_get_first_load_module(&cursor, &info);
   
   while (!ret)
   {
      FmmSignal *sig;
      struct PmLoadModuleInfoRequest *req;
      struct PmLoadModuleInfoReply *rep;
   
      sig = fmm_output_signal;
      req = &sig->data.load_module_info_request;
      req->sigNo = PM_LOAD_MODULE_INFO_REQUEST;
      memcpy(req->install_handle, 
             info.install_handle, PM_INSTALL_HANDLE_LENGTH);
      sig->size = sizeof (*req);
      
      if (request(sig))
      {
         return PFMI_END_OF_LIST;
      }
      
      rep = &sig->data.load_module_info_reply;
      rep->sigNo = PM_LOAD_MODULE_INFO_REPLY;
      rep->status = PM_SUCCESS;
      memcpy(rep->install_handle, 
             info.install_handle, PM_INSTALL_HANDLE_LENGTH);
      rep->entrypoint = info.entrypoint;
      rep->options = info.options;
      rep->text_base = info.text_base;
      rep->text_size = info.text_size;
      rep->data_base = info.data_base;
      rep->data_size = info.data_size;
      rep->no_of_sections = info.no_of_sections;
      rep->no_of_instances = info.no_of_instances;
      memcpy(rep->file_format, info.file_format, PM_FILE_FORMAT_NAME_LENGTH);        
      memcpy(rep->file_name, info.file_name, PM_PARENT_FILE_NAME_SIZE);
      sig->size = sizeof (*rep);
      
      if (reply(sig, ret))
      {
         return PFMI_END_OF_LIST;
      }
      
      if (!ret)
      {
         ret = ose_pfmi_get_next_load_module(&cursor, &info);
      }       
   } 
   
   return ret == PFMI_END_OF_LIST ? PFMI_OK : ret;     
}

PfmiStatus fmm_pfmi_load_module_section_info(
                    int32_t (*request)(FmmSignal *sig),
                    int32_t (*reply)(FmmSignal *sig, int32_t status))
{
   PfmiStatus ret;
   
   struct PfmiLoadModuleCursor cursor;
   struct PfmiLoadModuleInfo info;
   
   assert(reply);
   assert(request);

   ret = ose_pfmi_get_first_load_module(&cursor, &info);
   
   while (!ret)
   {
      FmmSignal *sig;
      struct PmLoadModuleSectionInfoRequest *req;
      struct PmLoadModuleSectionInfoReply *rep;
   
      sig = fmm_output_signal;
      req = &sig->data.load_module_section_info_request;
      req->sigNo = PM_LOAD_MODULE_SECTION_INFO_REQUEST;
      memcpy(req->install_handle, 
             info.install_handle, PM_INSTALL_HANDLE_LENGTH);
      req->section_interval_start = 0;
      req->section_interval_end = info.no_of_sections - 1;
      sig->size = sizeof (*req);
      
      if (request(sig))
      {
         return PFMI_END_OF_LIST;
      }
      
      rep = &sig->data.load_module_section_info_reply;
      rep->sigNo = PM_LOAD_MODULE_SECTION_INFO_REPLY;
      rep->status = PM_SUCCESS;
      memcpy(rep->install_handle, 
             info.install_handle, PM_INSTALL_HANDLE_LENGTH);
      /* NOTE: This is absurd if info.no_of_sections == 0 
       * (it will for the core module). Count on ~0 + 1 == 0 
       * to tell the client that there are no sections: */
      rep->section_interval_start = 0;
      rep->section_interval_end = info.no_of_sections - 1;
      {
         uint32_t i;
         struct PfmiLoadModuleSectionInfo sinfo;
         
         for (i = 0; i < info.no_of_sections; i++)
         {
            ret = ose_pfmi_get_load_module_section_info(info.install_handle,
                                                        i, &sinfo);
            if (ret != PFMI_OK)
            {
               break;
            }
            rep->sections[i].section_base = sinfo.section_base;
            rep->sections[i].section_size = sinfo.section_size;
            rep->sections[i].section_attr = sinfo.section_attr;
            rep->sections[i].section_options = sinfo.section_options;
            memcpy(rep->sections[i].section_name, 
                   sinfo.section_name, PM_SECTION_NAME_LENGTH);
         } 
      }      
      rep->status = ret ? PM_ENO_PROGRAM_MANAGER : PM_SUCCESS;         
      sig->size = offsetof(struct PmLoadModuleSectionInfoReply, sections) 
               + info.no_of_sections * sizeof(struct PmLoadModuleSectionInfo);
      
      if (reply(sig, ret))
      {
         return PFMI_END_OF_LIST;
      }
      
      if (!ret)
      {
         ret = ose_pfmi_get_next_load_module(&cursor, &info);
      }       
   } 
   
   return ret == PFMI_END_OF_LIST ? PFMI_OK : ret;         
}                    
