/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 1998-2012 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */
/*

 ****************************************************************************
 *           PROGRAM MODULE
 *
 *     $Workfile:   fss.c  $
 *     Document no: @(#) 511OSEXX-YY $Revision:   1.11  $
 *     $Author:   olli  $
 *     $Date:   09/08/00 14:50:10  $
 *
 *

 ****************************************************************************
 * File System Server; a name server for EFS objects and processes.
 */

#include <stddef.h>
#include <string.h>
#include "assert.h"
#include "ose.h"
#include "ose_rem.h"
#include "ose_err.h"
#include "osetypes.h"
#include "efs.h"
#include "efs_err.h"
#include "heapapi.h"
#include "list.h"
#include "time.h"

#ifdef __ARMCC_VERSION
/*
 ** No one seems to really remember why this is here, but the ARM
 ** compiler was to blame. So I changed OSARM (which affects all Arm
 ** branches) to __ARMCC_VERSION, which only affects the offending
 ** compiler branch. We should investigate further and remove as soon
 ** as possible. --frbr
 */
#pragma O0
#endif

#include "fss.sig"
#include "fm.sig"
#include "efs.sig"
#include "fss_validate.h"

#define INVALID_TMOREF 0

#define STR(x) #x
#define STRSTR(x) STR(x)

#define FSS_WHAT "EFS " STRSTR(VERSION) " file system server"
#define FM_WHAT "EFS " STRSTR(VERSION) " root/null fm"

/* Flag indicating if validation of volume access is enabled. */
static Boolean resolve_validate_enabled = False;

union SIGNAL
{
   SIGSELECT sigNo;
   struct
   {
      SIGSELECT sigNo;
      EfsStatus status;
   } x;
};

OSENTRYPOINT ose_fss;

/*********************
 *
 * some list.h extensions
 *
 *********************/

#undef malloc
#undef free
#ifdef MALLOC_FILELINE
#define malloc(s)   heap_alloc_private((s), __FILE__, __LINE__)
#else
#define malloc(s)   heap_alloc_private((s), 0, 0)
#endif
#define free(p)     heap_free_private((p))

static void *
#ifdef MALLOC_FILELINE
   zzalloc_link(unsigned size, const char *file, unsigned line)
#define alloc_link(s) zzalloc_link((s), __FILE__, __LINE__)
#else
alloc_link(unsigned size)
#endif
{
#ifdef MALLOC_FILELINE
   ListLink *link = heap_alloc_private(size, file, line);

#else
   ListLink *link = heap_alloc_private(size, 0, 0);
#endif
   list_init_link(link);
   return link;
}

static void
free_link(void *link)
{
   heap_free_private(list_unlink(link));
}

/******************************************
 *
 * FileDesc
 *
 ******************************************/

typedef struct
{
   ListLink link;
   PROCESS pid;
   OSATTREF ar;
   FmHandle handle;
   U32 size, current;
   unsigned root:1;
   char buffer[1];
}
FileDesc;

static U32 total_size(List * known);
static void prepare_buffer(List * known, char *buf);

static FileDesc *
alloc_filedesc(struct FmOpen *req, FmHandle handle, List *known, Boolean root)
{
   U32 size = root ? total_size(known) : 0;
   FileDesc *fdp = alloc_link(sizeof (FileDesc) + size);

   fdp->pid = sender((union SIGNAL **) &req);
   fdp->ar = attach(NULL, fdp->pid);
   fdp->handle = handle;
   fdp->size = size;
   fdp->current = 0;
   fdp->root = root != 0;
   if (root)
   {
      prepare_buffer(known, fdp->buffer);
   }
   return fdp;
}

static FileDesc *
dup_filedesc(FileDesc * org, FmHandle new_handle, PROCESS new_pid)
{
   FileDesc *fdp = alloc_link(sizeof (FileDesc) + org->size);
   fdp->pid = new_pid;
   fdp->ar = attach(NULL, fdp->pid);
   fdp->handle = new_handle;
   fdp->size = org->size;
   fdp->current = org->current;
   fdp->root = org->root;
   memcpy(fdp->buffer, org->buffer, org->size);
   return fdp;
}

static void
free_filedesc(FileDesc * fdp)
{
   if (fdp->pid != 0)
   {
      detach(&fdp->ar);
   }
   free_link(fdp);
}

static FileDesc *
find_filedesc(List * filedesc, FmHandle handle)
{
   FileDesc *fdp = list_head(filedesc);

   while (fdp != NULL)
   {
      if (fdp->handle == handle) return fdp;
      fdp = list_next(filedesc, (ListLink *) fdp);
   }
   return NULL;
}

/***************************************
 *
 * Node
 *
 ***************************************/

struct Node
{
   ListLink link;
   EfsEnum local;      /* Visibility: local, global, remote. */
   EfsEnum type;       /* FSS, FM, VM, DDB, DDC etc. */
   PROCESS pid;  /* Pid of object, might be 0 for FSS's. */
   PROCESS ph_pid;     /* Pid of phantom process, might be 0. */
   char name[1];       /* Variable size. */
};

static struct Node *
find_node_by_name(List * list,
                  const char *name,
                  EfsEnum type)
{
   struct Node *node = list_head(list);
   struct Node *remote = NULL;

   while (node != NULL)
   {
      if (strcmp(node->name, name) == 0 &&
            (type == FSS_TYPE_ANY || node->type == type))
      {
         if (node->local == FSS_VIS_REMOTE)
         {
            /* keep the first found entry for later
             * if we won't find a local node */
            if (remote == NULL)
            {
               remote = node;
            }
         }
         else 
         {
            /* found a local node */
            return node;
         }
      }
      node = list_next(list, (ListLink *) node);
   }
   return remote;
}

static struct Node *
new_node(const char *name, PROCESS pid, EfsEnum type, EfsEnum local)
{
   struct Node *node = alloc_link(sizeof (struct Node) + strlen(name));
   assert(type != FSS_TYPE_ANY);
   strcpy(node->name, name);
   node->pid = pid;
   node->ph_pid = 0;
   node->type = type;
   node->local = local;
   return node;
}

/***************************************
 *
 * PendReq
 *
 ***************************************/

struct PendReq
{
   ListLink link;
   struct FssResolve *req;
   OSTMOREF tmoref;
};

static struct PendReq *
new_pendreq(struct FssResolve *req)
{
   struct PendReq *pr = alloc_link(sizeof (struct PendReq));
   pr->req = req;
   if (req->timeOut != (U32) FSS_TMO_FOREVER)
   {
      struct FTimeout *sig = (struct FTimeout *) alloc(sizeof (struct FTimeout), FTIMEOUT);
      sig->pr = pr;
      pr->tmoref = request_tmo_sig(req->timeOut, (union SIGNAL **) &sig);
   }
   else
   {
      pr->tmoref = INVALID_TMOREF;
   }
   return pr;
}

static void
free_pendreq(struct PendReq *pr, struct FTimeout *tmo)
{
   if (tmo != NULL)
   {
      assert(pr->req != NULL);
      free_buf((union SIGNAL **) &tmo);
   }
   else if (pr->tmoref != INVALID_TMOREF)
   {
      assert(pr->req == NULL);
      cancel_tmo(&pr->tmoref);
   }
   free_link(pr);
}

/***************************************
 *
 * Replies to requests
 *
 ***************************************/

static void
interface(struct FssInterfaceRequest *req)
{
   PROCESS sender_ = sender((union SIGNAL **) &req);
   sendFssInterfaceReply(EFS_SUCCESS, FSS_WHAT, 0, FSS_INTERFACE_REQUEST, FSS_EXT_RESOLVE_REQUEST, sender_);
   free_buf((union SIGNAL **) &req);
}

static void
connect_fss(struct FssConnectFss *req, List * known)
{
   if (!find_node_by_name(known, req->huntPath, FSS_TYPE_FSS))
   {
      struct HuntPath *sig = (struct HuntPath *) alloc(sizeof (struct HuntPath)
            + strlen(req->huntPath), FHUNT_FSS);
      strcpy(sig->hunt_path, req->huntPath);
      hunt(req->huntPath, 0, NULL, (union SIGNAL **) &sig);
      list_insert_tail(known, (ListLink *)new_node(req->huntPath, 0,
               FSS_TYPE_FSS, FSS_VIS_LOCAL));
      req->status = EFS_SUCCESS;
   }
   else
      /* Object name already used. */
   {
      req->status = FSS_EDUPLICATE_OBJECT;
   }
   req->sigNo = FSS_CONNECT_FSS_REPLY;
   send((union SIGNAL **) &req, sender((union SIGNAL **) &req));
}

static void
fss_found(struct HuntPath *sig, List * known)
{
   struct Node *fss = find_node_by_name(known, sig->hunt_path,
         FSS_TYPE_FSS);
   assert(fss != NULL);
   assert(fss->pid == 0);
   assert(fss->local == FSS_VIS_LOCAL);
   fss->pid = sender((union SIGNAL **) &sig);
   sig->signo = FATTACH_FSS;   /* Reuse hunt signal as attach signal. */
   attach((union SIGNAL **) &sig, fss->pid);

   /* Send info about all global objects to newly found FSS. */
   {
      struct Node *node = list_head(known);
      while (node != NULL)
      {
         if (node->local == FSS_VIS_GLOBAL && node->type != FSS_TYPE_FSS)
         {
            assert(node->pid != 0);
            sendFssListReply(EFS_SUCCESS, node->name, node->type, node->local,
                  node->pid, fss->pid);
         }
         node = list_next(known, (ListLink *) node);
      }
   }
}

static U32
total_size(List * known)
{
   U32 size = 0;
   struct Node *node = list_head(known);
   while (node != NULL)
   {
      if (node->type == FSS_TYPE_VOLUME && strcmp(node->name, "/") != 0)
      {
         size += strlen(node->name);
      }
      node = list_next(known, (ListLink *) node);
   }
   return size;
}

static void
prepare_buffer(List * known, char *buf)
{
   struct Node *node = list_head(known);
   while (node != NULL)
   {
      if (node->type == FSS_TYPE_VOLUME && strcmp(node->name, "/") != 0)
      {
         strcpy(buf, node->name + 1);
         buf += strlen(node->name);
      }
      node = list_next(known, (ListLink *) node);
   }
}

static void
fss_lost(struct HuntPath *sig, List * known)
{
   struct Node *fss = find_node_by_name(known, sig->hunt_path,
         FSS_TYPE_FSS);
   assert(fss != NULL);
   assert(fss->pid != 0);
   assert(fss->local == FSS_VIS_LOCAL);
   fss->pid = 0;
   sig->signo = FHUNT_FSS;     /* Reuse attach signal as hunt signal. */
   hunt(fss->name, 0, NULL, (union SIGNAL **) &sig);
}

/**
 * @function    resolve_validate
 * @brief       Perform validation of volume access if required (if
 *              enabled and if signal FssResolve contains FssResolveExtValidate
 *              or signal FssResolve contains volume).
 * @input       sig  Pointer to signal FssResolve.
 * @return      True=access granted, False=access denied.
 */
static Boolean
resolve_validate(struct FssResolve *sig)
{
   U32     req_sig_size;
   Boolean valid = True;
   Boolean validated = False;

   /* If validate not enabled. */
   if (!resolve_validate_enabled)
   {
      return valid;
   }

   /* If signal size is even multiple of U32s (see fss.sig). */
   req_sig_size = sigsize( (union SIGNAL **)&sig);

   if (req_sig_size % sizeof(U32) == 0)
   {
      /* If signal contains extension(s). */
      struct FssResolveExtHead *ext_head_p = (struct FssResolveExtHead *)
         ( (U32)sig + req_sig_size - sizeof(struct FssResolveExtHead));

      if (ext_head_p->magic == FSS_RESOLVE_EXT_HEAD_MAGIC)
      {
         /* If signal contains extension FssResolveExtValidate. 
          * Loop because the signal may include more extensions in the future.
          */
         struct FssResolveExtValidate *ext_p = (struct FssResolveExtValidate *)
            ( (U32)sig + ext_head_p->first_ext_position);

         while(1)
         {
            if (ext_p->ext_id == FSS_RESOLVE_EXT_VALIDATE)
            {
               valid = fss_validate_volume_access(ext_p->operation,
                     &ext_p->s[ext_p->obj_path_offset],
                     &ext_p->s[ext_p->lnhname_offset],
                     &ext_p->s[ext_p->opt_string_offset],
                     ext_p->open_mode,
                     ext_p->access_mode,
                     &ext_p->s[ext_p->username_offset],
                     &ext_p->s[ext_p->termname_offset],
                     sender( (union SIGNAL **)&sig));

               /* Output extension handled in reply signal. */
               ext_p->handled = 1;

               validated = True;
               break;
            }

            /* If more extensions. */
            if (ext_p->next_ext_position !=0)
            {
               ext_p = (struct FssResolveExtValidate *)((U32)sig + ext_p->next_ext_position);
            }
            else
            {
               break;
            }
         } /* while... */
      }
   }

   /* If no extension found. Only type VOLUME makes sense for operation UNKNOWN. */
   if ( (!validated) && (sig->type == FSS_TYPE_VOLUME) )
   {
      char    empty_string = '\0';

      valid = fss_validate_volume_access(FSS_VALIDATE_OPERATION_UNKNOWN,
            &sig->name[0],    /* obj_path=volume name */
            &empty_string,    /* lnhname */
            &empty_string,    /* opt_string */
            0,                /* open_mode */
            0,                /* access_mode */
            &empty_string,    /* username */
            &empty_string,    /* termname */
            sender( (union SIGNAL **)&sig));
   }

   return valid;
}

static void
resolve_pending(List * pending, struct Node *new)
{
   /* Traverse list of all pending resolve requests. */
   struct PendReq *pr = list_head(pending);
   while (pr != NULL)
   {
      /* Call list_next() now when 'pending' is still valid. */
      struct PendReq *next = list_next(pending, (ListLink *) pr);
      struct FssResolve *req = pr->req;
      /* Check name and type of pending request for a match. */
      if ((req->type == FSS_TYPE_ANY || req->type == new->type) &&
            strcmp(req->name, new->name) == 0)
      {
         /* A match, try to return resolve request. */
         OSBUFSIZE size = sigsize((union SIGNAL **)&req);
         struct FssResolve *reply = (struct FssResolve *)
            alloc(size, FSS_RESOLVE_REPLY);
         /* Copy signal from signal->type */
         memcpy(&reply->type, &req->type, 
               size - ((char*)&req->type - (char*)req));
         /* Validate. */
         if (resolve_validate(req) != False)
         {
            reply->status = EFS_SUCCESS;
         }
         else
         {
            reply->status = FSS_ERESOLVE_EXT_VALIDATE_OPERATION_DENIED;
         }
         assert(new->pid != 0);
         if (send_w_sender((union SIGNAL **) &reply, new->pid,
                  sender((union SIGNAL **) &req)) == 0)
         {
            /* Succeeded, free pending resolve request. */
            free_buf((union SIGNAL **)&req);
            pr->req = NULL;
            free_pendreq(pr, NULL);
         }
         /* Else failed, ignore match with dead process (new->pid). */
      }
      /* Continue checking next request in list. */
      pr = next;
   }
}

static void
update_remote_fss(List * known, struct Node *new)
{
   struct Node *fss = list_head(known);
   assert(new->local == FSS_VIS_GLOBAL);
   while (fss != NULL)
   {
      if (fss->type == FSS_TYPE_FSS && fss->pid != 0)
      {
         sendFssListReply(EFS_SUCCESS, new->name, new->type, new->local, 
               new->pid, fss->pid);
      }
      fss = list_next(known, (ListLink *) fss);
   }
}

static EfsStatus
register_object(const char *name, EfsEnum type, EfsEnum local,
      PROCESS pid, List * known, List * pending)
{
   EfsStatus status;

   if (type == FSS_TYPE_VOLUME && (name[0] != '/'))
   {
      /* We cannot register a volume to the FSS that does not
       * have a leading slash in its volume name. */
      status = FSS_EINVALID_TYPE;
      goto returnStatus;
   }
   if (type != FSS_TYPE_ANY)
   {
      struct Node *found = find_node_by_name(known, name, type);
      if ( (found == NULL) || (found->local == FSS_VIS_REMOTE))
      {
         /* We did not found a node with this name, or
          * we found one but is remote and we are allowed to
          * register one locally to be prioritised. */
         struct Node *new = new_node(name, pid, type, local);
         list_insert_tail(known, (ListLink *) new);
         attach(NULL, pid);
         resolve_pending(pending, new);
         if (new->local == FSS_VIS_GLOBAL)
         {
            update_remote_fss(known, new);
         }
         status = EFS_SUCCESS;
      }
      else
      {
         /* Device name already used. */
         status = FSS_EDUPLICATE_OBJECT;
      }
   }
   else
      /* Invalid type of EFS object. */
   {
      status = FSS_EINVALID_TYPE;
   }
returnStatus:   
   return status;
}

static void
register_name(struct FssRegister *req, List * known, 
              List * pending, Boolean reply)
{
   PROCESS pid = sender((union SIGNAL **) &req);
   req->status = register_object(req->name, req->type, req->local, pid, 
         known, pending);
   if ( reply )
   {
   req->sigNo = FSS_REGISTER_REPLY;
   send((union SIGNAL **) &req, pid);
}
   else
   {
      free_buf((union SIGNAL **)&req);
   }
}

static void
clean_list(List * list, PROCESS pid)
{
   struct Node *node = list_head(list);
   while (node != NULL)
   {
      struct Node *next = list_next(list, (ListLink *) node);
      if (node->pid == pid)
      {
         if (node->ph_pid != 0)
         {
            /* Kill phantom associated with object. */
            kill_proc(node->ph_pid);
         }
         free_link(node);
      }
      node = next;
   }
}

static void
fm_client_lost(PROCESS pid, List * filedesc)
{
   FileDesc *fdp = list_head(filedesc);
   while (fdp != NULL)
   {
      FileDesc *next = list_next(filedesc, (ListLink *) fdp);
      if (fdp->pid == pid)
      {
         fdp->pid = 0;
         free_filedesc(fdp);
      }
      fdp = next;
   }
}

static void
clear_client(PROCESS pid, List *known, List *filedesc)
{
   clean_list(known, pid);
   fm_client_lost(pid, filedesc);
}

static void
resolve_name(struct FssResolve *req,
             List * known,
             List * pending,
             List * filedesc)
{
   PROCESS sender_ = sender((union SIGNAL **) &req);
   /* Traverse list of all known process. */
   struct Node *node = NULL;


find_new_entry :
   node = find_node_by_name(known, req->name, req->type);

   if (node != NULL) 
   {
      /* A match, try to return resolve request. */
      OSBUFSIZE size = sigsize((union SIGNAL **)&req);
      struct FssResolve *reply = (struct FssResolve *)
         alloc(size, FSS_RESOLVE_REPLY);
      /* Copy signal from signal->type */
      memcpy (&reply->type, &req->type, 
            size - ((char*)&req->type - (char*) req));
      /* Validate. */
      if (resolve_validate(req) != False)
      {
         reply->status = EFS_SUCCESS;
      }
      else
      {
         reply->status = FSS_ERESOLVE_EXT_VALIDATE_OPERATION_DENIED;
      }
      if (send_w_sender((union SIGNAL **) &reply, node->pid, sender_) 
            == 0)
      {
         /* Succeeded. */
         free_buf((union SIGNAL **)&req);
         return; /* Request replied. */
      }
      else
      {
         /* The process process has died but we did not received
          * the attach signal yet. Eliminate the corresponding
          * corresponding entries and try again. */
         clear_client(node->pid, known, filedesc);
         goto find_new_entry;
      }
   }
   else
   {
      /* Name not found among known process. */
      /* Check if this request shall linger for a while (waiting to resolve).*/
      if (req->timeOut != FSS_TMO_NONE)
      {
         /* Yes, add request to queue of pending resolve requests. */
         list_insert_tail(pending, (ListLink *) new_pendreq(req));
         return; /* Request queued. */
      }
   }

   /* No, return request with failed status. */
   /* Validate for logging purpose but ignore result. */
   resolve_validate(req);
   req->status = FSS_EUNKNOWN_OBJECT;
   req->sigNo = FSS_RESOLVE_REPLY;
   send((union SIGNAL **) &req, sender_);

}

static void
list_known(struct FssListRequest *req, List * known)
{
   PROCESS client = sender((union SIGNAL **) &req);
   struct Node *node = list_head(known);
   while (node != NULL)
   {
      if (req->type == FSS_TYPE_ANY || req->type == node->type)
      {
         PROCESS pid = node->pid != 0 ? node->pid : current_process();
         sendFssListReply(EFS_SUCCESS, node->name, node->type, node->local, pid, client);
      }
      node = list_next(known, (ListLink *) node);
   }
   sendFssListReply(EFS_NO_MORE_ITEMS, "", FSS_TYPE_ANY, False, current_process(), client);
   free_buf((union SIGNAL **) &req);
}

static void
ext_resolve_name(struct FssExtResolve *req,
                 List * known,
                 List * pending,
                 List * filedesc)
{
   PROCESS sender_ = sender((union SIGNAL **) &req);
   /* Traverse list of all known process. */
   struct Node *node = NULL;

ext_find_new_entry:

   node = find_node_by_name(known,
            &(req->strings[req->name]),
            req->type);
   if (node != NULL) 
   {
      /* A match, try to return resolve request. */
      OSBUFSIZE size = sigsize((union SIGNAL **)&req);
      struct FssExtResolve *reply = (struct FssExtResolve *)
         alloc(size, FSS_EXT_RESOLVE_REPLY);
      /* Copy signal from signal->type */
      memcpy (&reply->type, &req->type, 
            size - ((char*)&req->type - (char*) req));
      reply->status = EFS_SUCCESS;
      /* Validate. */
      if (send_w_sender((union SIGNAL **) &reply, node->pid, sender_) == 0)
      {
         /* Succeeded. */
         free_buf((union SIGNAL **)&req);
         return; /* Request replied. */
      }
      else
      {
         /* The process process has died but we did not received
          * the attach signal yet. Eliminate the corresponding
          * corresponding entries and try again. */
         free_buf((union SIGNAL **)&reply);
         clear_client(node->pid, known, filedesc);
         goto ext_find_new_entry;
      }
   }
   else
   {
      /* Name not found among known process. */
      /* Check if this request shall linger for a while (waiting to resolve).*/
      if (req->timeOut != FSS_TMO_NONE)
      {
         /* Yes, add request to queue of pending resolve requests. */
         list_insert_tail(pending, (ListLink *) new_pendreq((struct FssResolve *)req));
         return; /* Request queued. */
      }
   }

   /* No, return request with failed status. */
   req->status = FSS_EUNKNOWN_OBJECT;
   req->sigNo = FSS_EXT_RESOLVE_REPLY;
   send((union SIGNAL **) &req, sender_);
}

static void
inherit_remote(struct FssListReply *sig, List * known, List * pending)
{
   PROCESS pid = sender((union SIGNAL **) &sig);
   struct Node *new = new_node(sig->name, pid, sig->type, FSS_VIS_REMOTE);

   list_insert_tail(known, (ListLink *) new);
   attach(NULL, pid);
   resolve_pending(pending, new);

   free_buf((union SIGNAL **) &sig);
}

static void
timeout(struct FTimeout *tmo)
{
   struct PendReq *pr = tmo->pr;
   struct FssResolve *reply = pr->req;
   free_pendreq(pr, tmo);
   /* Validate for logging purpose but ignore result. */
   resolve_validate(reply);
   reply->sigNo = FSS_RESOLVE_REPLY;
   reply->status = FSS_EUNKNOWN_OBJECT;
   send((union SIGNAL **) &reply, sender((union SIGNAL **) &reply));
}

static void
fm_open(struct FmOpen *req, List *filedesc, FmHandle * handle_counter, List * known)
{
   if (strcmp(&req->s[req->name], "/") == 0)
   {
      EfsBitmask supported = FM_MODE_READ | FM_MODE_DIRECTORY;
      if (strcmp(&req->s[req->amode], "") == 0 && (req->modes & ~supported) == 0)
      {
         FileDesc *fdp = alloc_filedesc(req, (*handle_counter)++, known, True);
         list_insert_head(filedesc, (ListLink *) fdp);
         req->status = EFS_SUCCESS;
         req->handle = fdp->handle;
         req->bufferMode = FM_BUFF_FULL;
         req->bufferSize = 512;
      }
      else /* Unsupported ascii modes or binary modes. */
      {
         req->status = FM_EINVALID_PARAMETERS;
      }
   }
   else if (strcmp(&req->s[req->name], "/null") == 0)
   {
      EfsBitmask supported = FM_MODE_READ | FM_MODE_WRITE | FM_MODE_CREATE |
         FM_MODE_EXCLUSIVE | FM_MODE_TRUNCATE | FM_MODE_TEXT |
         FM_MODE_APPEND;
      if (strcmp(&req->s[req->amode], "") == 0 && (req->modes & ~supported) == 0)
      {
         FileDesc *fdp = alloc_filedesc(req, (*handle_counter)++, known, False);
         list_insert_head(filedesc, (ListLink *) fdp);
         req->status = EFS_SUCCESS;
         req->handle = fdp->handle;
         req->bufferMode = FM_BUFF_FULL;
         req->bufferSize = 4096;
      }
      else /* Unsupported ascii modes or binary modes. */
      {
         req->status = FM_EINVALID_PARAMETERS;
      }
   }
   else /* Unsupported object name. */
   {
      req->status = FM_EOBJECT_NOT_FOUND;
   }
   req->sigNo = FM_OPEN_REPLY;
   send_w_s((union SIGNAL **) &req, addressee((union SIGNAL **)&req), 
         sender((union SIGNAL **) &req));
}

static void
fm_close(struct FmClose *req, List * filedesc)
{
   FileDesc *fdp = find_filedesc(filedesc, req->handle);
   if (fdp != NULL)
   {
      free_filedesc(fdp);
      req->status = EFS_SUCCESS;
   }
   else /* Invalid file descriptor. */
   {
      req->status = FM_EINVALID_HANDLE;
   }
   req->sigNo = FM_CLOSE_REPLY;
   send_w_s((union SIGNAL **)&req, addressee((union SIGNAL **)&req), 
         sender((union SIGNAL **)&req));
}

static void
fm_set_owner_process(struct FmSetOwnerProc *req, List * filedesc)
{
   PROCESS sender_ = sender((union SIGNAL **) &req);
   FileDesc *fdp;

   fdp = find_filedesc(filedesc, req->handle);
   if (fdp != NULL)
   {
      PROCESS old_owner = fdp->pid;

      fdp->pid = sender_;
      detach(&fdp->ar);
      fdp->ar = attach(NULL, sender_);
      sender_ = old_owner;    /* Reply to old owner. */
      req->status = EFS_SUCCESS;
   }
   else
   {
      req->status = FM_EINVALID_HANDLE;
   }
   req->sigNo = FM_SET_OWNER_PROC_REPLY;
   send_w_s((union SIGNAL **) &req, addressee((union SIGNAL **)&req), 
         sender_);
}

static void
set_position_zero(struct FmSetPos * req, List * filedesc)
{
   FileDesc *fdp = find_filedesc(filedesc, req->handle);

   if (fdp != NULL && req->offset == 0)
   {
      fdp->current = 0;
      req->position = 0;
      req->status = EFS_SUCCESS;
   }
   else if (req->offset != 0)
   {
      req->status = FM_EOPER_NOT_SUPPORTED;
   }
   else /* fdp = NULL */
   {
      req->status = FM_EINVALID_HANDLE;
   }

   req->sigNo = FM_SET_POS_REPLY;
   send_w_s((union SIGNAL **)&req, addressee((union SIGNAL **)&req),
         sender((union SIGNAL **)&req));
}

static U32
minimum(U32 a, U32 b)
{
   return a < b ? a : b;
}

static void
fm_write(struct FmWriteRequest *req, List * filedesc)
{
   struct FmWriteReply *reply = (struct FmWriteReply *)
      alloc(sizeof(struct FmWriteReply), FM_WRITE_REPLY);
   FileDesc *fdp = find_filedesc(filedesc, req->handle);
   if (fdp != NULL)
   {
      reply->actual = req->requested;
      reply->status = EFS_SUCCESS;
   }
   else
   {
      reply->status = FM_EINVALID_HANDLE;
   }
   reply->handle = req->handle;
   reply->sigNo = FM_WRITE_REPLY;
   send_w_s((union SIGNAL **)&reply, addressee((union SIGNAL **)&req), 
         sender((union SIGNAL **)&req));
   free_buf((union SIGNAL **)&req);
}

static void
fm_read(struct FmReadRequest *req, List * filedesc)
{
   struct FmReadReply *reply;
   FileDesc *fdp = find_filedesc(filedesc, req->handle);
   if (fdp != NULL)
   {
      U32 actual = minimum(req->requested, fdp->size - fdp->current);
      assert(fdp->root || actual == 0);
      reply = (struct FmReadReply *)alloc(sizeof(struct FmReadReply)
            + actual, FM_READ_REPLY);
      if (actual != 0)
      {
         memcpy(reply->buffer, fdp->buffer + fdp->current, actual);
         fdp->current += actual;
      }
      reply->actual = actual;
      reply->status = EFS_SUCCESS;
   }
   else
   {
      reply = (struct FmReadReply *)alloc(sizeof(struct FmReadReply), 
            FM_READ_REPLY);
      reply->status = FM_EINVALID_HANDLE;
   }
   reply->handle = req->handle;
   reply->sigNo = FM_READ_REPLY;
   send_w_s((union SIGNAL **)&reply, addressee((union SIGNAL **)&req), 
         sender((union SIGNAL **)&req));
   free_buf((union SIGNAL **)&req);
}

static void
fm_write_ptr(struct FmWritePtr *req, List * filedesc)
{
   FileDesc *fdp = find_filedesc(filedesc, req->handle);
   if (fdp != NULL)
   {
      req->actual = req->requested;
      req->status = EFS_SUCCESS;
   }
   else
   {
      req->status = FM_EINVALID_HANDLE;
   }
   req->sigNo = FM_WRITE_PTR_REPLY;
   send_w_s((union SIGNAL **)&req, addressee((union SIGNAL **)&req), 
         sender((union SIGNAL **)&req));
}

static void
fm_read_ptr(struct FmReadPtr *req, List * filedesc)
{
   FileDesc *fdp = find_filedesc(filedesc, req->handle);
   if (fdp != NULL)
   {
      U32 actual = minimum(req->requested, fdp->size - fdp->current);
      assert(fdp->root || actual == 0);
      if (actual != 0)
      {
         memcpy(req->buffer, fdp->buffer + fdp->current, actual);
         fdp->current += actual;
      }
      req->actual = actual;
      req->status = EFS_SUCCESS;
   }
   else
   {
      req->status = FM_EINVALID_HANDLE;
   }
   req->sigNo = FM_READ_PTR_REPLY;
   send_w_s((union SIGNAL **)&req, addressee((union SIGNAL **)&req), 
         sender((union SIGNAL **)&req));
}

static void
fm_dup_handle(struct FmDupHandle *req, List *filedesc, FmHandle *handle_counter)
{
   PROCESS sender_ = sender((union SIGNAL **) &req);
   FileDesc *fdp = find_filedesc(filedesc, req->oldHandle);
   if (fdp != NULL)
   {
      FileDesc *new = dup_filedesc(fdp, (*handle_counter)++, sender_);
      list_insert_tail(filedesc, (ListLink *) new);
      sender_ = fdp->pid;     /* Send reply to owner of original FileDesc.  */
      req->newHandle = new->handle;
      req->status = EFS_SUCCESS;
   }
   else
   {
      req->status = FM_EINVALID_HANDLE;
   }
   req->sigNo = FM_DUP_HANDLE_REPLY;
   send_w_s((union SIGNAL **) &req, addressee((union SIGNAL **)&req), 
         sender_);
}

#define MAXNAMLEN 5     /* Max(strlen("/"), strlen("/null"). */

static void
fm_examine_volume(struct FmExamineVolRequest *req, time_t creation_time)
{
   PROCESS to = addressee((union SIGNAL **)&req);
   struct FmExamineVolReply *reply = 
      (struct FmExamineVolReply *)alloc(sizeof(struct FmExamineVolReply) 
            + MAXNAMLEN, FM_EXAMINE_VOL_REPLY);
   struct FmVolumeInfo *info = &reply->info;
   info->bspDevice = False;
   info->unit = ~0U;
   info->loBlock = 0;
   info->hiBlock = 0;
   info->freeBlocks = 0;
   info->blockSize = 0;
   info->creationTime = creation_time;
   info->modificationTime = 0;
   if (to == current_process())
   {
      info->access = FM_ACCESS_FORMAT | FM_ACCESS_READ;
      info->volumeName = 0;
      info->deviceName = strlen("/");
      strcpy(&info->s[info->volumeName], "/");
   }
   else /* to should be equal to null_ph. */
   {
      info->access = FM_ACCESS_FORMAT | FM_ACCESS_READ | FM_ACCESS_WRITE;
      info->volumeName = 0;
      info->deviceName = strlen("/null");
      strcpy(&info->s[info->volumeName], "/null");
   }
   reply->status = EFS_SUCCESS;
   send_w_s((union SIGNAL **)&reply, to, sender((union SIGNAL **)&req));
   free_buf((union SIGNAL **)&req);
}

static void
fm_examine_object(struct FmExamineObjRequest *req, List * filedesc, time_t creation_time)
{
   struct FmExamineObjReply * reply = (struct FmExamineObjReply *)alloc(sizeof(struct FmExamineObjReply) + MAXNAMLEN, FM_EXAMINE_OBJ_REPLY);
   if (strcmp(&req->s[req->name], "/") == 0 || strcmp(&req->s[req->name], "/.") == 0)
   {
      const char *name = "/";
      struct FmObjectInfo *info = &reply->info;
      info->type = FM_TYPE_DIR;
      info->attrs = FM_ATTR_READ | FM_ATTR_NAMEOK;
      info->diskId = 0;
      info->links = 0;
      info->creationTime = creation_time;
      info->dataModTime = 0;
      info->headModTime = 0;
      info->accessTime = 0;
      info->owner = 0;
      info->group = 0;
      info->sizeBytes = 0;
      info->sizeBlocks = 0;
      info->bufferMode = FM_BUFF_FULL;
      info->bufferSize = 512;
      info->openCount = list_length(filedesc); /* LMF: Only count root fd. */
      info->name = 0;
      strcpy(&info->s[info->name], name);
      reply->status = EFS_SUCCESS;
   }
   else if (strcmp(&req->s[req->name], "/null") == 0)
   {
      const char *name = "/null";
      struct FmObjectInfo *info = &reply->info;
      info->type = FM_TYPE_STREAM;
      info->attrs = FM_ATTR_READ | FM_ATTR_WRITE | FM_ATTR_NAMEOK;
      info->diskId = 0;
      info->links = 0;
      info->creationTime = creation_time;
      info->dataModTime = 0;
      info->headModTime = 0;
      info->accessTime = 0;
      info->owner = 0;
      info->group = 0;
      info->sizeBytes = 0;
      info->sizeBlocks = 0;
      info->bufferMode = FM_BUFF_FULL;
      info->bufferSize = 4096;
      info->openCount = list_length(filedesc); /* LMF: Only count null fd. */
      info->name = 0;
      strcpy(&info->s[info->name], name);
      reply->status = EFS_SUCCESS;
   }
   else
   {
      reply->status = FM_EOBJECT_NOT_FOUND;
   }
   send_w_s((union SIGNAL **)&reply, addressee((union SIGNAL **)&req), 
         sender((union SIGNAL **)&req));
   free_buf((union SIGNAL **)&req);
}

static void
fm_examine_by_handle(struct FmExamineRequest *req, List *filedesc)
{
   struct FmExamineReply * reply = (struct FmExamineReply *)alloc(sizeof(struct FmExamineReply), FM_EXAMINE_REPLY);
   struct FmObjectInfo *info = &reply->info;
   PROCESS to = addressee((union SIGNAL **)&req);
   FileDesc *fdp = find_filedesc(filedesc, req->handle);
   
   memset(info, 0, sizeof (struct FmObjectInfo));
   if (fdp != NULL)
   {  
      if (to == current_process())
      {
         strcpy(&info->s[info->name], "/");
         info->bufferMode = FM_BUFF_FULL;
         info->bufferSize = 512;
         info->type = FM_TYPE_DIR;
         info->attrs = FM_ATTR_READ | FM_ATTR_NAMEOK;
      }
      else 
      { 
         strcpy(&info->s[info->name], "/null");
         info->bufferMode = FM_BUFF_FULL;
         info->bufferSize = 4096;
         info->type = FM_TYPE_FILE;
         info->attrs = FM_ATTR_READ | FM_ATTR_WRITE | FM_ATTR_NAMEOK;
      }  

      reply->status = EFS_SUCCESS;
   }
   else
   {
      reply->status = FM_EINVALID_HANDLE;
   }
   reply->handle = req->handle;
   reply->sigNo = FM_EXAMINE_REPLY;
   send_w_s((union SIGNAL **)&reply, to,
         sender((union SIGNAL **)&req));
   
   free_buf((union SIGNAL **)&req);
}

static void
fm_interface(struct FmInterfaceRequest *req)
{
   struct FmInterfaceReply *reply;
   static const SIGSELECT supported_signals[] = {
      36,
      FM_INTERFACE_REQUEST,
      FM_MOUNT_REQUEST,
      FM_UNMOUNT_REQUEST,
      FM_FORMAT_VOL_REQUEST,
      FM_VALIDATE_VOL_REQUEST,
      FM_RELABEL_VOL_REQUEST,
      FM_SYNC_VOL_REQUEST,
      FM_CREATE_OBJ_REQUEST,
      FM_REMOVE_OBJ_REQUEST,
      FM_SET_SIZE_OBJ_REQUEST,
      FM_SET_ATTR_OBJ_REQUEST,
      FM_SET_TIME_OBJ_REQUEST,
      FM_SET_OWNER_OBJ_REQUEST,
      FM_WAIT_REQUEST,
      FM_CANCEL_REQUEST,
      FM_SET_SIZE_REQUEST,
      FM_SET_ATTR_REQUEST,
      FM_DRAIN_REQUEST,
      FM_LOCK_REQUEST,
      FM_FLUSH_TERM_REQUEST,
      FM_EXAMINE_TERM_REQUEST,
      FM_CONFIG_TERM_REQUEST,
      FM_RENAME_OBJ_REQUEST,
      FM_SUBSCRIBE_SIGNAL_REQUEST,
      FM_EXAMINE_VOL_REQUEST,
      FM_EXAMINE_OBJ_REQUEST,
      FM_OPEN_REQUEST,
      FM_CLOSE_REQUEST,
      FM_SET_OWNER_PROC_REQUEST,
      FM_SET_POS_REQUEST,
      FM_WRITE_REQUEST,
      FM_READ_REQUEST,
      FM_WRITE_PTR_REQUEST,
      FM_READ_PTR_REQUEST,
      FM_DUP_HANDLE_REQUEST,
      FM_EXAMINE_REQUEST
   };  
   size_t sig_size = sizeof *reply + sizeof supported_signals;
   
   reply = (struct FmInterfaceReply *)alloc(sig_size, FM_INTERFACE_REPLY);
   reply->status = EFS_SUCCESS;
   strncpy(reply->whatStr, FM_WHAT, sizeof FM_WHAT);
   reply->biosHandle = 0;
   memcpy(reply->sigs, supported_signals, sizeof supported_signals);
   
   send_w_s((union SIGNAL **)&reply, addressee((union SIGNAL **)&req),
         sender((union SIGNAL **)&req));

   free_buf((union SIGNAL **)&req);
}

/***************************************
 *
 * process ose_fss
 *
 ***************************************/

OS_PROCESS(ose_fss)
{
   List known;       /* List with all known file system objects.  */
   List pending;  /* List of pending resolve requests. */
   List filedesc;    /* List of (open) file descriptors. */
   FmHandle handle_counter = 0;
   time_t creation_time = time(0);
   PROCESS self = current_process();
   PROCESS null_ph; /* PID of phantom for null device. */

   list_init(&known);
   list_init(&pending);
   list_init(&filedesc);

   /* Check if validate volume access shall be enabled. */
   resolve_validate_enabled = fss_enable_validate_volume_access();

   /* Register FSS as volume manager for '/' (root). */
   list_insert_tail(&known, (ListLink *) new_node("/", self, FSS_TYPE_VOLUME, FSS_VIS_LOCAL));

   /* Register phantom process as volume manager for '/null' (null device). */
   {
      struct OS_redir_entry redir[1];
      redir[0].sig = 1; /* Number of entries in redir table. */
      redir[0].pid = self;
      null_ph = create_process(OS_PHANTOM, "ose_nulldev", NULL, 0, 0, 0, 0, redir, 0, 0);
      list_insert_tail(&known, (ListLink *) new_node("/null", null_ph, FSS_TYPE_VOLUME, FSS_VIS_LOCAL));
   }

#ifdef LIBWHAT
   LIBWHAT
#endif

      for (;;)
      {
         static const SIGSELECT selall[] = {0};
         union SIGNAL *sig = receive(selall);
         switch (sig->sigNo)
         {
            case FSS_INTERFACE_REQUEST:
               interface((struct FssInterfaceRequest *) sig);
               break;
            case FSS_CONNECT_FSS_REQUEST:
               connect_fss((struct FssConnectFss *) sig, &known);
               break;
            case FSS_REGISTER_REQUEST:
               register_name((struct FssRegister *) sig, &known, &pending, True);
               break;
            case FSS_REGISTER_SILENT_REQUEST:
               register_name((struct FssRegister *) sig, &known, &pending, False);
               break;
            case FSS_RESOLVE_REQUEST:
               resolve_name((struct FssResolve *) sig, &known, &pending, &filedesc);
               break;
            case FSS_LIST_REQUEST:
               list_known((struct FssListRequest *) sig, &known);
               break;
            case FSS_EXT_RESOLVE_REQUEST:
               ext_resolve_name((struct FssExtResolve*) sig, &known, &pending, &filedesc);
               break;
            case FHUNT_FSS:
               fss_found((struct HuntPath *) sig, &known);
               break;
            case FATTACH_FSS:
               fss_lost((struct HuntPath *) sig, &known);
               break;
            case FTIMEOUT:
               timeout((struct FTimeout *) sig);
               break;
            case FSS_LIST_REPLY:
               inherit_remote((struct FssListReply *) sig, &known, &pending);
               break;
            case OS_ATTACH_SIG : /* Client terminated. */
               clear_client(sender(&sig), &known, &filedesc);
               free_buf(&sig);
               break;

            case FM_INTERFACE_REQUEST :
               fm_interface((struct FmInterfaceRequest *) sig);  
               break;
            case FM_MOUNT_REQUEST :
            case FM_UNMOUNT_REQUEST :
            case FM_FORMAT_VOL_REQUEST :
            case FM_VALIDATE_VOL_REQUEST :
            case FM_RELABEL_VOL_REQUEST :
            case FM_SYNC_VOL_REQUEST :
            case FM_CREATE_OBJ_REQUEST :
            case FM_REMOVE_OBJ_REQUEST :
            case FM_SET_SIZE_OBJ_REQUEST :
            case FM_SET_ATTR_OBJ_REQUEST :
            case FM_SET_TIME_OBJ_REQUEST :
            case FM_SET_OWNER_OBJ_REQUEST :
            case FM_WAIT_REQUEST :
            case FM_CANCEL_REQUEST :
            case FM_SET_SIZE_REQUEST :
            case FM_SET_ATTR_REQUEST :
            case FM_DRAIN_REQUEST :
            case FM_LOCK_REQUEST :
            case FM_FLUSH_TERM_REQUEST :
            case FM_EXAMINE_TERM_REQUEST :
            case FM_CONFIG_TERM_REQUEST :
            case FM_RENAME_OBJ_REQUEST :
            case FM_SUBSCRIBE_SIGNAL_REQUEST :
               {
                  union SIGNAL *reply = alloc(sizeof(reply->x), sig->sigNo + 1);
                  reply->x.status = FM_EOPER_NOT_SUPPORTED;
                  send_w_s(&reply, addressee(&sig), sender(&sig));
                  free_buf(&sig);
                  break;
               }

            case FM_EXAMINE_VOL_REQUEST :
               fm_examine_volume((struct FmExamineVolRequest *)sig, creation_time);
               break;
            case FM_EXAMINE_OBJ_REQUEST :
               fm_examine_object((struct FmExamineObjRequest *)sig, &filedesc, creation_time);
               break;
            case FM_OPEN_REQUEST :
               fm_open((struct FmOpen *)sig, &filedesc, &handle_counter, &known);
               break;
            case FM_CLOSE_REQUEST :
               fm_close((struct FmClose *)sig, &filedesc);
               break;
            case FM_SET_OWNER_PROC_REQUEST :
               fm_set_owner_process((struct FmSetOwnerProc *)sig, &filedesc);
               break;
            case FM_SET_POS_REQUEST:
               set_position_zero((struct FmSetPos *) sig, &filedesc);
               break;
            case FM_WRITE_REQUEST :
               fm_write((struct FmWriteRequest *)sig, &filedesc);
               break;
            case FM_READ_REQUEST :
               fm_read((struct FmReadRequest *)sig, &filedesc);
               break;
            case FM_WRITE_PTR_REQUEST :
               fm_write_ptr((struct FmWritePtr *)sig, &filedesc);
               break;
            case FM_READ_PTR_REQUEST :
               fm_read_ptr((struct FmReadPtr *)sig, &filedesc);
               break;
            case FM_DUP_HANDLE_REQUEST :
               fm_dup_handle((struct FmDupHandle *)sig, &filedesc, &handle_counter);
               break;
            case FM_EXAMINE_REQUEST :
               fm_examine_by_handle((struct FmExamineRequest *)sig, &filedesc);
               break;

            default:
               {
                  /*we assume that the reply signal is +1*/
                  sig->sigNo++;
                  sig->x.status = FM_EOPER_NOT_SUPPORTED;
                  send(&sig, sender(&sig));
                  break;
               }
         }
      }
}
