/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2010 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

#include "heap_algo_large.h"
#include <string.h>


/**************************************************************************
 * Plug-in History
 **************************************************************************
 *
 * IMPORTANT
 * 
 * This comment shall be updated with a description of the change each
 * time the extension version number is increased. The version number
 * shall be increased if incompatible modifications are made to internal
 * structures or after behavioral changes. See HeapExtension below.
 * 
 **************************************************************************
 *
 *  Version  Description
 *  -------  ----------------------------------------------------------
 *   101     Renamed struct Algorithm to struct LargeAlgorithm.
 *   100     Initial revision.  
 *
 **************************************************************************/


/***************************************************************************
 * Helper Macros
 ***************************************************************************/
#ifdef OSMIPS
#define PAGE_SIZE  0x2000
#else
#define PAGE_SIZE  0x1000
#endif

#define ALIGN(size)  (((size) + (PAGE_SIZE - 1)) & ~(PAGE_SIZE - 1))


/***************************************************************************
 * Data Structures
 ***************************************************************************/

struct LargeAlgorithm
{
   uint32_t  id;

   struct HeapStatistics *stats;
   struct HeapConfiguration *config;
   struct Buffer *list;

   struct HeapServiceInfo info;           /* Plug-in information. */
};

/***************************************************************************/

struct Buffer
{
   struct Buffer *next;
   struct Buffer *prev;
   size_t         req_size;
   size_t         act_size;
   HeapPID        owner;
   uint32_t       control;

#define BUF_ID_MASK               0xC0000000ul
#define BUF_ID_SHIFT              30
#define BUF_ANNOT_MASK            0x00000001ul
#define BUF_GC_MASK               0x00000002ul

#define buf_id(buf)               ((buf)->control & BUF_ID_MASK)
#define buf_owner(buf)            ((buf)->owner)
#define buf_annotated(buf)        ((buf)->control & BUF_ANNOT_MASK)
#define buf_gc(buf)               ((buf)->control & BUF_GC_MASK)
#define buf_clear_anno(buf)       ((buf)->control &= ~BUF_ANNOT_MASK)
   
#define buf_set_control(buf,id)   ((buf)->control = (id))
#define buf_set_gc(buf)           ((buf)->control |= BUF_GC_MASK)
#define buf_set_owner(buf,pid)    ((buf)->owner = (pid))

#define buf_clear_gc(buf)         ((buf)->control &= ~BUF_GC_MASK)

#define BUF_NEXT_FRAG_MASK        0xFFFFFFFCul
#define BUF_NEXT_ID_MASK          0x00000003ul
#define buf_next_frag(frag)       ((struct Buffer *)((uint32_t)(frag) & BUF_NEXT_FRAG_MASK))
#define buf_next_id(frag)         (((uint32_t)(frag) & BUF_NEXT_ID_MASK) << BUF_ID_SHIFT)
#define buf_next_mask(frag, id)   ((struct Buffer *)((uint32_t)(frag)|((id) >> BUF_ID_SHIFT) ))

#define BUF_CHECKSUM_MAGIC        0x80102040ul
#define buf_checksum(_b)                        \
   (((uint32_t)(_b)->next ^                     \
    (uint32_t)(_b)->prev ^                      \
    BUF_CHECKSUM_MAGIC) & ~(BUF_ID_MASK|BUF_ANNOT_MASK))
#define buf_set_checksum(_b)     ((_b)->control = buf_id(_b)|buf_annotated(_b)|buf_checksum(_b))
#define buf_get_checksum(_b)     ((_b)->control & ~(BUF_ID_MASK|BUF_ANNOT_MASK))
};

/***************************************************************************/

struct BufferCursor
{
   struct Buffer *buf;
   struct Buffer *frag;
   uint32_t       reserved0;
   uint32_t       reserved1;
};

/***************************************************************************/






/***************************************************************************
 * Forward Declarations
 ***************************************************************************/

static inline HeapError
buf_check_header(struct Heap *heap, struct LargeAlgorithm *a,
                 struct Buffer *buf);

/***************************************************************************/




/***************************************************************************
 * Helper Functions
 ***************************************************************************/

static always_inline void
buf_insert(struct Heap *heap, struct LargeAlgorithm *a, struct Buffer *buf)
{
   buf->prev = NULL;
   buf->next = a->list;
   if (buf->next)
   {
      if (buf_checksum(buf->next) != buf_get_checksum(buf->next))
      {
         zzheap_error(heap, HEAP_ERROR_BUF_HEADER, buf->next);
         return;
      }
      
      buf->next->prev = buf;

      buf_set_checksum(buf->next);
   }

   a->list = buf;

   buf_set_checksum(buf);
}

/***************************************************************************/

static always_inline void
buf_remove(struct Heap *heap, struct LargeAlgorithm *a, struct Buffer *buf)
{
   if (buf->prev)
   {
      if (buf_checksum(buf->prev) != buf_get_checksum(buf->prev))
      {
         zzheap_error(heap, HEAP_ERROR_BUF_HEADER, buf->prev);
         return;
      }
      
      buf->prev->next = buf->next;

      buf_set_checksum(buf->prev);
   }
   if (buf->next)
   {
      if (buf_checksum(buf->next) != buf_get_checksum(buf->next))
      {
         zzheap_error(heap, HEAP_ERROR_BUF_HEADER, buf->next);
         return;
      }

      buf->next->prev = buf->prev;

      buf_set_checksum(buf->next);
   }

   if (a->list == buf)
      a->list = buf->next;
}
   
/***************************************************************************/

static always_inline size_t
buf_act_size(struct LargeAlgorithm *a, struct Buffer *buf)
{
   (void)a;

   return buf->act_size;
}

/***************************************************************************/

static always_inline size_t
buf_req_size(struct LargeAlgorithm *a, struct Buffer *buf)
{
   return buf->req_size - sizeof(struct Buffer) - a->config->buf_size_endmark;
}

/***************************************************************************/

static inline struct Buffer *
buf_first(struct Heap *heap, struct LargeAlgorithm *a, HeapError *ecode)
{
   struct Buffer *buf = a->list;
   HeapError tmp_ecode = HEAP_ERROR_NONE;

   if (buf == NULL)
   {
      if (ecode != NULL)
         *ecode = tmp_ecode;
      return NULL;
   }
   
   /* Check integrity of next buffer. */
   tmp_ecode = buf_check_header(heap, a, buf);

   if (ecode != NULL)
      *ecode = tmp_ecode;

   if (tmp_ecode != HEAP_ERROR_NONE)
   {
      if (ecode == NULL)
         zzheap_error(heap, tmp_ecode, buf);

      return NULL;
   }


   return buf;
}


/***************************************************************************/

static inline struct Buffer *
buf_right(struct Heap *heap, struct LargeAlgorithm *a,
          struct Buffer *buf, HeapError *ecode)
{
   HeapError tmp_ecode;

   buf = buf->next;
   if (buf == NULL)
   {
      if (ecode != NULL)
         *ecode = HEAP_ERROR_NONE;
      return NULL;
   }
   
   /* Check integrity of next buddy. */
   tmp_ecode = buf_check_header(heap, a, buf);

   if (ecode != NULL)
      *ecode = tmp_ecode;

   if (tmp_ecode != HEAP_ERROR_NONE)
   {
      if (ecode == NULL)
         zzheap_error(heap, tmp_ecode, buf);

      return NULL;
   }
   
   return buf;
}

/***************************************************************************/

static inline struct Buffer *
buf_find(struct Heap *heap, struct LargeAlgorithm *a,
         struct Buffer *buf, HeapError *ecode)
{
   struct Buffer *next = a->list;

   while (next != buf && next != NULL)
   {
      HeapError tmp_ecode;

      tmp_ecode = buf_check_header(heap, a, next);
      if (tmp_ecode != HEAP_ERROR_NONE)
      {
         if (ecode != NULL)
            *ecode = tmp_ecode;
         else
            zzheap_error(heap, tmp_ecode, next);

         return NULL;
      }

      next = next->next;
   }

   if (next == NULL && ecode != NULL)
      *ecode = HEAP_ERROR_BUF_UNKNOWN;

   return next;
}


/***************************************************************************/

static inline HeapError
buf_check_frag(struct Heap *heap, struct LargeAlgorithm *a,
               struct Buffer *buf)
{
   /* Validates that the buffer pointer resides in one
    * of the algorithm's buffers. If so, it should be
    * safe to access the address. */

   HeapError ecode = HEAP_ERROR_NONE;

   if (!a->config->frag_check)
      return HEAP_ERROR_NONE;

   /* Regions/fragments may disappear at any time due to
    * unbreak requests. Locks are needed. */

   buf = buf_find(heap, a, buf, &ecode);
   if (buf != NULL)
   {
      return HEAP_ERROR_NONE;
   }

   if (ecode != HEAP_ERROR_NONE)
      return ecode;
   
   return HEAP_ERROR_BUF_UNKNOWN;
}

/***************************************************************************/

static inline HeapError
buf_check_header(struct Heap *heap, struct LargeAlgorithm *a,
                 struct Buffer *buf)
{
   /* Validate that the buffer header is reasonable.
    * Free buffers are validated through a checksum
    * of the entire header, including freelist links. 
    * It is stored in the owner field.
    *
    * The validation of allocated buffers is too weak
    * and needs improvment. 
    *
    * If a buffer is not reqognized by an algorithm,
    * the event hook must return the continue action
    * to allow other algorithms to identify it. If no
    * algorithm responds, i.e. all request type event
    * hooks return "continue", the buffer is likely
    * corrupted and the framework will generate an
    * error. */
   (void)heap;
   
   if (((uint32_t)buf) & (HEAP_ALIGNMENT-1))
      return HEAP_ERROR_BUF_UNALIGNED;

   if (buf_id(buf) != a->id)
      return HEAP_ERROR_BUF_UNKNOWN;

   if (buf->req_size > buf->act_size)
      return HEAP_ERROR_BUF_HEADER;
   
   if (buf_get_checksum(buf) != buf_checksum(buf))
      return HEAP_ERROR_BUF_HEADER;
   
   return HEAP_ERROR_NONE;
}

/***************************************************************************/

static inline HeapError
buf_check_owner(struct Heap *heap, struct LargeAlgorithm *a,
                struct Buffer *buf, HeapPID pid1)
{
   /* Validates that the pid passed as parameter is the
    * owner of the buffer. Private buffers can only be
    * operated on by its owner. */
   
   HeapPID pid2 = buf_owner(buf);

   (void)heap;
   (void)a;

   /* A buffer owner value of 0 (pid2) indicates that the
    * buffer is shared and can be operated on by anyone.
    * An originator pid value of 0 (pid1) means that the
    * operation was forced and should be allowed. */

   if (pid1 != 0 && pid2 != 0 && pid1 != pid2)
      return HEAP_ERROR_BUF_OWNER;

   return HEAP_ERROR_NONE;
}

/***************************************************************************/

static inline HeapError
buf_check_endmark(struct Heap *heap, struct LargeAlgorithm *a,
                  struct Buffer *buf)
{
   /* Validates that the buffer endmark, written
    * at the end of the user data area, is still valid
    * when the buffer is freed. Damages to the endmark
    * indicates that the user has used more memory than
    * it originally allocated. The next buffer's header
    * may have been damaged as well. */
   
   char *endm = (char *)&buf[1] + buf_req_size(a, buf);
   int i = a->config->buf_size_endmark;

   (void)heap;

   /* There will be no endmark if the buffer is free. */
   while (--i>=0)
   {
      if (endm[i] != HEAP_ENDMARK)
         return HEAP_ERROR_BUF_ENDMARK;
   }

   return HEAP_ERROR_NONE;
}

/***************************************************************************/

static inline HeapError
buf_check_at_free(struct Heap *heap, struct LargeAlgorithm *a,
                  struct Buffer *buf, HeapPID pid1)
{
   /* A collection of all integrity and error checks
    * that are carried out during a free request. The
    * function is also used for resize request which
    * behaves similarly. */
   
   HeapError status;
   
   status = buf_check_frag(heap, a, buf);
   if (status != HEAP_ERROR_NONE)
      return status;
      
   status = buf_check_header(heap, a, buf);
   if (status != HEAP_ERROR_NONE)
      return status;

   status = buf_check_owner(heap, a, buf, pid1);
   if (status != HEAP_ERROR_NONE)
      return status;

   status = buf_check_endmark(heap, a, buf);
   if (status != HEAP_ERROR_NONE)
      return status;

   return HEAP_ERROR_NONE;
}

/***************************************************************************/

static inline HeapError
buf_check_at_query(struct Heap *heap, struct LargeAlgorithm *a,
                   struct Buffer *buf)
{
   /* A collection of all integrity and error checks
    * that are carried out during a free request. The
    * function is also used for resize request which
    * behaves similarly. */
   
   HeapError status;
   
   /* Frag check is optionally enabled for queries, unless in freeze-mode. */
   if (unlikely(a->config->frag_check || heap->freezed))
   {
      status = buf_check_frag(heap, a, buf);
      if (status != HEAP_ERROR_NONE)
         return status;
   }
      
   status = buf_check_header(heap, a, buf);
   if (status != HEAP_ERROR_NONE)
      return status;

   return HEAP_ERROR_NONE;
}

/***************************************************************************/

static always_inline void
buf_set_endmark(struct LargeAlgorithm *a, struct Buffer *buf)
{
   /* Initializes the buffer endmark, which is used to
    * detect application bugs. If the endmark becomes
    * damaged the application has used more memory in
    * the buffer than it was allowed to. */
   
   char *endm = (char *)&buf[1] + buf_req_size(a, buf);
   unsigned i = 0;

   for (i=0; i<a->config->buf_size_endmark; i++)
      endm[i] = HEAP_ENDMARK;
}

/***************************************************************************/

static inline HeapBufferStatus
buf_status(struct Heap *heap, struct LargeAlgorithm *a, struct Buffer *buf)
{
   /* Helper function that investigates the health of
    * an individual buffer. */
   HeapBufferStatus status = 0;
   HeapError ecode;

   ecode = buf_check_header(heap, a, buf);
   if (ecode != HEAP_ERROR_NONE)
   {
      if (ecode == HEAP_ERROR_BUF_UNALIGNED)
         return HEAP_BUFFER_UNKNOWN;

      if (buf_annotated(buf))
         return HEAP_BUFFER_HEADERS|HEAP_BUFFER_ANNOTATED;

      return HEAP_BUFFER_HEADERS;
   }

   status |= HEAP_BUFFER_ALLOCATED;
   
   if (buf_check_endmark(heap, a, buf) != HEAP_ERROR_NONE)
      status |= HEAP_BUFFER_ENDMARK;
   
   if (buf_annotated(buf))
      status |= HEAP_BUFFER_ANNOTATED;

   return status;
}

/***************************************************************************/

static inline struct Buffer *
buf_next(struct Heap *heap, struct LargeAlgorithm *a,
         struct BufferCursor *cursor, uint8_t advance_frag)
{

   if (buf_next_frag(cursor->frag) == NULL)
      cursor->buf = a->list;
   else
   {
      struct Buffer *buf;
      HeapError ecode;
      
      buf = buf_find(heap, a, buf_next_frag(cursor->frag), &ecode);
      if (buf == NULL)
         return NULL;

      ecode = buf_check_header(heap, a, buf);
      if (ecode != HEAP_ERROR_NONE)
         return NULL;

      if (cursor->buf == NULL)
      {
         cursor->buf = buf;
      }
      else if (advance_frag)
      {
         cursor->buf = buf->next;
      }
      else
      {
         cursor->buf = NULL;
      }
   }

   if (cursor->buf != NULL)
      cursor->frag = buf_next_mask(cursor->buf, a->id);
   else
      cursor->frag = NULL;
   
   return cursor->buf;
}

/***************************************************************************/





/***************************************************************************
 * Private Interface
 ***************************************************************************/

static HeapAction
l_large_alloc(struct Heap *heap, union HeapEvent *event, void *opaque)
{
   struct LargeAlgorithm *a = opaque;
   struct Buffer *buf;
   HeapLock lock;

   size_t req_size =
      event->e_alloc.size +
      sizeof(struct Buffer) +
      a->config->buf_size_endmark;
   size_t act_size = req_size;  /* Requested fragment size.  */
   size_t max_size = req_size;  /* Maximum size of fragment. */

   /* Interrupt processes are not allowed to allocate large buffers. */
   if (!heap_brkallowed()) {
      return HEAP_ACTION_CONTINUE;
   }
   
   /* Query allocator for the size of the fragment we will get. */
   do {
      HeapParam p[] = {
         HPM(QUERY_FRAG_NEXT_SIZE,1), 0,
         HPM_END
      };

      p[1] = (HeapParam)act_size;
      zzheap_query(heap, p, 0);

      if (HPM_IS_HANDLED(p[0]))
         act_size = (size_t)p[1];
   } while (0);

   /* Protect against race condition leading to
    * overcommitment of system memory. */
   heap_brklock(heap, &lock);

   /* Check configured buffer usage limit. */
   if (act_size + a->stats->buf_size_used > a->config->buf_size_max) {
      heap_brkunlock(heap, &lock);
      return HEAP_ACTION_CONTINUE;
   }

   /* Setting min == max means that we get
    * exactly what we asked for and nothing else. */
   max_size = act_size;

   buf = zzheap_break(
      heap,
      &act_size,
      act_size,
      &max_size,
      a->id,
      event->e_alloc.params);
   if (buf == NULL) {
      heap_brkunlock(heap, &lock);
      return HEAP_ACTION_CONTINUE;
   }

   heap_assert(act_size >= req_size,
               "An allocator returned a smaller fragment than requested.");

   heap_lock(heap, &lock, 0);

   buf->req_size = req_size;
   buf->act_size = act_size;
   
   buf_set_owner(buf, event->e_alloc.owner);
   buf_set_control(buf, a->id);
   buf_set_endmark(a, buf);
   buf_insert(heap, a, buf);
   
   a->stats->buf_size_used += buf_act_size(a, buf);
   a->stats->buf_size_req  += buf_req_size(a, buf);

   if (a->stats->buf_size_used > a->stats->buf_size_used_hi)
      a->stats->buf_size_used_hi = a->stats->buf_size_used;

   if (buf_owner(buf))
      a->stats->buf_num_private++;
   else
      a->stats->buf_num_shared++;

   heap_unlock(heap, &lock, 0);

   heap_brkunlock(heap, &lock);

   event->e_alloc.ptr  = &buf[1];
   event->e_alloc.size = act_size;

   return HEAP_ACTION_INTERCEPT;
}

/***************************************************************************/

static HeapAction
l_large_free(struct Heap *heap, union HeapEvent *event, void *opaque)
{
   struct LargeAlgorithm *a = opaque;
   struct Buffer *buf = event->e_free.ptr;
   HeapLock lock;
   HeapError status;
   
   if (buf == NULL)
      return HEAP_ACTION_CONTINUE;

   buf = &buf[-1];

   heap_lock(heap, &lock, 0);

   status = buf_check_at_free(heap, a, buf, event->e_free.owner);
   if (status == HEAP_ERROR_BUF_UNKNOWN)
   {
      heap_unlock(heap, &lock, 0);
      return HEAP_ACTION_CONTINUE;
   }

   if (status != HEAP_ERROR_NONE)
   {
      heap_unlock(heap, &lock, 0);
      zzheap_error(heap, status, &buf[1]);
      return HEAP_ACTION_INTERCEPT;
   }

   if (buf_gc(buf))
   {
      /* The buffer is used by the garbage collector and
       * can't be freed. We simply clear the bit to indicate
       * that we attempted to free it but couldn't. The collector
       * will notice and free the buffer for us. */
      buf_clear_gc(buf);
      heap_unlock(heap, &lock, 0);

      return HEAP_ACTION_INTERCEPT;
   }
   
   a->stats->buf_size_req  -= buf_req_size(a, buf);
   a->stats->buf_size_used -= buf_act_size(a, buf);
   event->e_free.size       = buf_act_size(a, buf);
   
   if (buf_owner(buf))
      a->stats->buf_num_private--;
   else
      a->stats->buf_num_shared--;

   buf_clear_anno(buf);
   buf_remove(heap, a, buf);

   heap_unlock(heap, &lock, 0);

   zzheap_unbreak(heap, buf, event->e_free.params);
   
   return HEAP_ACTION_INTERCEPT;
}

/***************************************************************************/

static HeapAction
l_large_resize(struct Heap *heap, union HeapEvent *event, void *opaque)
{
   struct LargeAlgorithm *a = opaque;
   struct Buffer *buf = event->e_resize.ptr;
   HeapLock lock;
   HeapError status;
   
   size_t req_size =
      event->e_resize.size +
      sizeof(struct Buffer) +
      a->config->buf_size_endmark;

   if (buf == NULL)
      return HEAP_ACTION_CONTINUE;

   buf = &buf[-1];

   heap_lock(heap, &lock, 0);

   status = buf_check_at_free(heap, a, buf, event->e_free.owner);
   if (status == HEAP_ERROR_BUF_UNKNOWN)
   {
      heap_unlock(heap, &lock, 0);
      return HEAP_ACTION_CONTINUE;
   }

   if (status != HEAP_ERROR_NONE)
   {
      heap_unlock(heap, &lock, 0);

      zzheap_error(heap, status, &buf[1]);
      event->e_resize.ptr = NULL;
      return HEAP_ACTION_INTERCEPT;
   }
   
   /* FIXME: Do we want to shrink below previously requested size? */
   if (req_size > buf->act_size || req_size < buf->req_size)
   {
      heap_unlock(heap, &lock, 0);
      event->e_resize.ptr = NULL;
      return HEAP_ACTION_INTERCEPT;
   }

   a->stats->buf_size_req -= buf->req_size;
   buf->req_size = req_size;
   a->stats->buf_size_req += buf->req_size;
   
   buf_set_endmark(a, buf);
   /* The owner field remains unchanged */

   event->e_resize.size = buf->act_size;
   event->e_resize.size_prev = buf->act_size;
   
   heap_unlock(heap, &lock, 0);

   return HEAP_ACTION_INTERCEPT;
}

/***************************************************************************/

static void
buf_collect(struct LargeAlgorithm *a, struct Buffer **head, struct Buffer *buf)
{
   /* This function links the buffer into
    * a list of all buffers that will be
    * freed by the garbage collector at a
    * later point in time.
    *
    * The buffer is also resized to make
    * room for a pointer in the user data
    * area. The pointer will point to the
    * next buffer in the list of collected
    * buffers. The endmark is moved forward. */
   
   const size_t oh =
      sizeof(struct Buffer) +
      sizeof(struct Buffer*) + 
      a->config->buf_size_endmark;

   size_t size = buf->req_size;

   if (size < oh)
   {
      a->stats->buf_size_req += oh - size;
      buf->req_size = oh;
      buf_set_endmark(a, buf);
   }

   buf[1].next = *head;
   *head = buf;
}

/**************************************************************************/

static HeapAction
l_large_collect(struct Heap *heap, union HeapEvent *event, void *opaque)
{
   struct LargeAlgorithm *a = opaque;
   struct Buffer *buf, *head = NULL;
   HeapPID pid = event->e_collect.owner;
   HeapLock lock;
   uint32_t i;

   heap_lock(heap, &lock, 0);

   /* Collect and gather all leaked buffers in a list. */
   buf = buf_first(heap, a, NULL);
   do
   {
      for (i=0; i<100 && buf != NULL; i++)
      {
         if (buf_owner(buf) == pid)
         {
            buf_collect(a, &head, buf);
         }
               
         buf = buf_right(heap, a, buf, NULL);
      }
      
      if (buf == NULL)
      {
         break;
      }
      
      /* Mark the buffer as a hostage of the collector.
       * Attempts to free the buffer will be detected by
       * the collector once heap locks are reacquired. */
      buf_set_gc(buf);

      heap_unlock(heap, &lock, 0);
      heap_lock(heap, &lock, 0);

      if (!buf_gc(buf))
      {
         buf_collect(a, &head, buf);
         buf = buf_right(heap, a, buf, NULL);
      }
      else
      {
         buf_clear_gc(buf);                  
      }
           
   } while (buf != NULL);

   heap_unlock(heap, &lock, 0);

   /* Now free all buffers in the list. */
   while (head != NULL)
   {
      buf = head;
      head = (struct Buffer *)buf[1].next;
      
      zzheap_free(heap, &buf[1], 0, NULL, NULL, 0);
   }
   
   return HEAP_ACTION_CONTINUE;
}

/***************************************************************************/

static HeapAction
l_large_query(struct Heap *heap, union HeapEvent *event, void *opaque)
{
   HeapParam *p = event->e_query.params;
   struct LargeAlgorithm *a = opaque;
   HeapLock lock;

   if (p == NULL)
      return HEAP_ACTION_CONTINUE;
   
   if (event->e_query.freezed)
      heap->freezed++;

   while (*p != HPM_END)
   {
      switch (HPM_ID(*p))
      {
         case HPM_QUERY_SERVICE:
         {
            if (!HPM_IS_HANDLED(p[0]) && p[1] == 0)
            {
               HPM_SET_HANDLED(p[0]);
               HPM_SET_VALUE(p[1], &a->info);
            }
            else if (p[1] == (HeapParam)&a->info)
            {
               HPM_CLEAR_HANDLED(p[0]);
               HPM_SET_VALUE(p[1], 0);
            }

            break;
         }
         case HPM_QUERY_BUF_REQ_SIZE:
         {
            struct Buffer *buf = &((struct Buffer *)p[1])[-1];

            if (!HPM_IS_HANDLED(p[0]))
            {
               heap_lock(heap, &lock, 0);

               if (buf_check_at_query(heap, a, buf) != HEAP_ERROR_NONE)
               {
                  heap_unlock(heap, &lock, 0);
                  break;
               }
               
               HPM_SET_HANDLED(p[0]);
               HPM_SET_VALUE(p[1], buf_req_size(a, buf));

               heap_unlock(heap, &lock, 0);
            }
            break;
         }
         case HPM_QUERY_BUF_ACT_SIZE:
         {
            struct Buffer *buf = &((struct Buffer *)p[1])[-1];

            if (!HPM_IS_HANDLED(p[0]))
            {
               heap_lock(heap, &lock, 0);

               if (buf_check_at_query(heap, a, buf) != HEAP_ERROR_NONE)
               {
                  heap_unlock(heap, &lock, 0);
                  break;
               }

               HPM_SET_HANDLED(p[0]);
               HPM_SET_VALUE(p[1], buf_act_size(a, buf));

               heap_unlock(heap, &lock, 0);
            }
            break;
         }
         case HPM_QUERY_BUF_NUM_SHARED:
         {
            HPM_SET_HANDLED(p[0]);
            HPM_SET_VALUE(p[1], a->stats->buf_num_shared);
            break;
         }
         case HPM_QUERY_BUF_NUM_PRIVATE:
         {
            HPM_SET_HANDLED(p[0]);
            HPM_SET_VALUE(p[1], a->stats->buf_num_private);
            break;
         }
         case HPM_QUERY_BUF_OWNER:
         {
            struct Buffer *buf = &((struct Buffer *)p[1])[-1];

            if (!HPM_IS_HANDLED(p[0]))
            {
               heap_lock(heap, &lock, 0);

               if (buf_check_at_query(heap, a, buf) != HEAP_ERROR_NONE)
               {
                  heap_unlock(heap, &lock, 0);
                  break;
               }

               HPM_SET_HANDLED(p[0]);
               HPM_SET_VALUE(p[1], buf->owner);

               heap_unlock(heap, &lock, 0);
            }

            break;
         }
            
         case HPM_QUERY_BUF_TOT_REQ_SIZE:
         {
            if (!HPM_IS_HANDLED(p[0]))
            {
               HPM_SET_HANDLED(p[0]);
               HPM_SET_VALUE(p[1], a->stats->buf_size_req);
            }
            break;
         }
         case HPM_QUERY_BUF_TOT_USED_SIZE:
         { 
            if (!HPM_IS_HANDLED(p[0]))
            {
               HPM_SET_HANDLED(p[0]);
               HPM_SET_VALUE(p[1], a->stats->buf_size_used);
            }
            break;
         }
         case HPM_QUERY_BUF_MAX_USED_SIZE:
         { 
            if (!HPM_IS_HANDLED(p[0]))
            {
               HPM_SET_HANDLED(p[0]);
               HPM_SET_VALUE(p[1], a->stats->buf_size_used_hi);
            }
            break;
         }

         /* Four unsupported tags. */
         case HPM_QUERY_BUF_TOT_FREE_SIZE:
         case HPM_QUERY_BUF_LARGEST_FREE:
         case HPM_QUERY_BUF_NUM_FREE:
         case HPM_QUERY_BUF_NUM_USED:
         {
            break;
         }

         case HPM_QUERY_BUF_FRAG_NEXT:
         {
            /* This algorithm stores buffers in parameter 2.
             * It doesn't matter if p[2] is a fragment or a buffer since both
             * headers have the algorithm identifier in the same bits.
             * Algorithms may choose which of the two it wants to use for
             * buffer traversals. */

            if (!HPM_IS_HANDLED(p[0]))
            {
               struct BufferCursor *cursor = (struct BufferCursor *)&p[1];
               struct Buffer *buf;

               heap_lock(heap, &lock, 0);
               
               buf = buf_next(heap, a, cursor, 0);
               if (buf == NULL)
               {
                  heap_unlock(heap, &lock, 0);
                  break;
               }

               HPM_SET_HANDLED(p[0]);
               HPM_SET_VALUE(p[1], &buf[1]);

               heap_unlock(heap, &lock, 0);
            }
            break;
         }
         case HPM_QUERY_BUF_NEXT:
         {
            /* This algorithm stores buffers in parameter 2.
             * It doesn't matter if p[2] is a fragment or a buffer since both
             * headers have the algorithm identifier in the same bits.
             * Algorithms may choose which of the two it wants to use for
             * buffer traversals. */

            /* Skip if the request doesn't concern us. */
            if (p[2] != 0 && buf_next_id(p[2]) != a->id)
               break;

            if (!HPM_IS_HANDLED(p[0]))
            {
               struct BufferCursor *cursor = (struct BufferCursor *)&p[1];
               struct Buffer *buf;

               heap_lock(heap, &lock, 0);

               buf = buf_next(heap, a, cursor, 1);
               if (buf == NULL)
               {
                  /* Initialize query for next algorithm. */
                  cursor->frag      = NULL;
                  cursor->reserved0 = 0;
                  cursor->reserved1 = 0;

                  heap_unlock(heap, &lock, 0);
                  break;
               }

               HPM_SET_HANDLED(p[0]);
               HPM_SET_VALUE(p[1], &buf[1]);

               heap_unlock(heap, &lock, 0);
            }
            break;
         }
         case HPM_QUERY_BUF_TYPE:
         {
            struct Buffer *buf = &((struct Buffer *)p[1])[-1];

            if (!HPM_IS_HANDLED(p[0]))
            {
               heap_lock(heap, &lock, 0);
               
               if (buf_check_at_query(heap, a, buf) != HEAP_ERROR_NONE)
               {
                  heap_unlock(heap, &lock, 0);
                  break;
               }

               HPM_SET_HANDLED(p[0]);
               HPM_SET_VALUE(p[1], HEAP_BUFFER_LARGE);

               heap_unlock(heap, &lock, 0);
            }

            break;
         }
         case HPM_QUERY_BUF_DATA:
         {
            struct Buffer *buf = &((struct Buffer *)p[1])[-1];

            if (!HPM_IS_HANDLED(p[0]))
            {
               void *data  = (void *)p[2];
               size_t size = (size_t)p[3];

               heap_lock(heap, &lock, 0);

               if (buf_check_at_query(heap, a, buf) != HEAP_ERROR_NONE)
               {
                  heap_unlock(heap, &lock, 0);
                  break;
               }

               HPM_SET_HANDLED(p[0]);
               memcpy(data, &buf[1], size);
               heap_unlock(heap, &lock, 0);
            }
            break;
         }
         case HPM_QUERY_BUF_STATUS:
         {
            struct Buffer *buf = &((struct Buffer *)p[1])[-1];
            HeapError ecode;

            if (!HPM_IS_HANDLED(p[0]))
            {
               heap_lock(heap, &lock, 0);

               ecode = buf_check_at_query(heap, a, buf);
               if (ecode == HEAP_ERROR_BUF_UNKNOWN)
               {
                  heap_unlock(heap, &lock, 0);
                  break;
               }

               HPM_SET_HANDLED(p[0]);
               HPM_SET_VALUE(p[1], buf_status(heap, a, buf));

               heap_unlock(heap, &lock, 0);
            }
            break;
         }
         case HPM_QUERY_BUF_NEXT_SIZE:
         {
            if (!HPM_IS_HANDLED(p[0]))
            {
               size_t s = (size_t)p[1] +
                  sizeof(struct Buffer) + a->config->buf_size_endmark;

               HeapParam q[] = {
                  HPM(QUERY_FRAG_NEXT_SIZE, 1), 0,
                  HPM_END};

               q[1] = (HeapParam)s;

               /* Query allocator for next fragment size. */
               zzheap_query(heap, q, event->e_query.freezed);

               if (!HPM_IS_HANDLED(q[0]))
                  break;

               s  = (size_t)q[1];
               s -= (sizeof(struct Buffer) + a->config->buf_size_endmark);

               HPM_SET_HANDLED(p[0]);
               HPM_SET_VALUE(p[1], s);
            }
            break;
         }
         case HPM_QUERY_BUF_THRESHOLD:
         {
            if (!HPM_IS_HANDLED(p[0]))
            {
               HPM_SET_HANDLED(p[0]);
               HPM_SET_VALUE(p[1], a->config->buf_size_threshold);
            }
            break;
         }
         case HPM_QUERY_BUF_ENDMARK:
         {
            if (!HPM_IS_HANDLED(p[0]))
            {
               HPM_SET_HANDLED(p[0]);
               HPM_SET_VALUE(p[1], a->config->buf_size_endmark);
            }
            break;
         }
         case HPM_QUERY_BUF_MAX_SIZE:
         {
            if (!HPM_IS_HANDLED(p[0]))
            {
               HPM_SET_HANDLED(p[0]);
               HPM_SET_VALUE(p[1], a->config->buf_size_max);
            }
            break;
         }

         default:
         {
            break;
         }
      }

      p += (1 + HPM_VALUES(p[0]));
   }

   if (event->e_query.freezed)
      heap->freezed--;

   return HEAP_ACTION_CONTINUE;
}

/***************************************************************************/

static HeapAction
l_large_control(struct Heap *heap, union HeapEvent *event, void *opaque)
{
   HeapParam *p = event->e_query.params;
   struct LargeAlgorithm *a = opaque;
   HeapLock lock;

   if (p == NULL)
      return HEAP_ACTION_CONTINUE;
   
   while (*p != HPM_END)
   {
      switch (HPM_ID(*p))
      {
         case HPM_CTRL_BUF_MAX_SIZE:
         {
            if (!HPM_IS_HANDLED(p[0]))
            {
               HPM_SET_HANDLED(p[0]);
               a->config->buf_size_max = (size_t)p[1];
            }
            break;
         }
         case HPM_CTRL_BUF_MAX_USED_SIZE:
         {
            /* Reset used watermark statistics. */
            if (!HPM_IS_HANDLED(p[0]))
            {
               HPM_SET_HANDLED(p[0]);
               a->stats->buf_size_used_hi = 0;
            }
            break;
         }
         case HPM_CTRL_BUF_OWNER:
         {
            struct Buffer *buf = &(((struct Buffer*)p[1])[-1]);
            HeapError ecode;

            if (!HPM_IS_HANDLED(p[0]))
            {
               heap_lock(heap, &lock, 0);
               
               ecode = buf_check_at_query(heap, a, buf);
               if (ecode != HEAP_ERROR_NONE)
               {
                  heap_unlock(heap, &lock, 0);
                  break;
               }

               buf_set_owner(buf, (HeapPID)p[2]);
               HPM_SET_HANDLED(p[0]);

               heap_unlock(heap, &lock, 0);
            }
            break;
         }
         default:
         {
            break;
         }
      }

      p += (1 + HPM_VALUES(p[0]));
   }

   return HEAP_ACTION_CONTINUE;
}

/***************************************************************************/

static HeapAction
l_large_destroy(struct Heap *heap, union HeapEvent *event, void *opaque)
{
   struct LargeAlgorithm *a = opaque;

   (void)event;

   /* Unregister this plug-in. */
   zzheap_detach_hook(heap, HEAP_EVENT_ALLOC_REQ,   l_large_alloc,   a);
   zzheap_detach_hook(heap, HEAP_EVENT_FREE_REQ,    l_large_free,    a);
   zzheap_detach_hook(heap, HEAP_EVENT_RESIZE_REQ,  l_large_resize,  a);
   zzheap_detach_hook(heap, HEAP_EVENT_COLLECT_REQ, l_large_collect, a);
   zzheap_detach_hook(heap, HEAP_EVENT_QUERY_REQ,   l_large_query,   a);
   zzheap_detach_hook(heap, HEAP_EVENT_CONTROL_REQ, l_large_control, a);

   if (a != NULL)
      heap_meta_free(heap, a);

   /* Allocator will free fragments. */

   return HEAP_ACTION_CONTINUE;
}

/***************************************************************************/



/***************************************************************************
 * Public Interface
 ***************************************************************************/

HeapStatus
heap_attach_large_algorithm(struct Heap *heap,
                            struct HeapStatistics *stats,
                            struct HeapConfiguration *config,
                            uint32_t id)
{
   HeapStatus status;
   struct LargeAlgorithm *a = heap_meta_alloc(
      heap, sizeof(struct LargeAlgorithm));

   if (a == NULL)
      return HEAP_EOUT_OF_MEMORY;

   bzero(a, sizeof(struct LargeAlgorithm));

   a->id     = id << BUF_ID_SHIFT;
   a->stats  = stats;
   a->config = config;

   a->info.version = 101;
   strcpy(a->info.id, "com.enea.heap.large");
   
   status = zzheap_attach_hook(
      heap, HEAP_EVENT_ALLOC_REQ, l_large_alloc, a, HEAP_PRIO_ALGO_LOW);
   if (status != HEAP_SUCCESS)
   {
      heap_meta_free(heap, a);
      return status;
   }

   status = zzheap_attach_hook(
      heap, HEAP_EVENT_FREE_REQ, l_large_free, a, HEAP_PRIO_ALGO_LOW);
   if (status != HEAP_SUCCESS)
   {
      heap_meta_free(heap, a);
      return status;
   }

   status = zzheap_attach_hook(
      heap, HEAP_EVENT_RESIZE_REQ, l_large_resize, a, HEAP_PRIO_ALGO_LOW);
   if (status != HEAP_SUCCESS)
   {
      heap_meta_free(heap, a);
      return status;
   }

   status = zzheap_attach_hook(
      heap, HEAP_EVENT_COLLECT_REQ, l_large_collect, a, HEAP_PRIO_ALGO_LOW);
   if (status != HEAP_SUCCESS)
   {
      heap_meta_free(heap, a);
      return status;
   }

   status = zzheap_attach_hook(
      heap, HEAP_EVENT_QUERY_REQ, l_large_query, a, HEAP_PRIO_ALGO_LOW);
   if (status != HEAP_SUCCESS)
   {
      heap_meta_free(heap, a);
      return status;
   }

   status = zzheap_attach_hook(
      heap, HEAP_EVENT_CONTROL_REQ, l_large_control, a, HEAP_PRIO_ALGO_LOW);
   if (status != HEAP_SUCCESS)
   {
      heap_meta_free(heap, a);
      return status;
   }

   status = zzheap_attach_hook(
      heap, HEAP_EVENT_DESTROY_RES, l_large_destroy, a, HEAP_PRIO_ALGO_NORMAL);
   if (status != HEAP_SUCCESS)
   {
      heap_meta_free(heap, a);
      return status;
   }

   return HEAP_SUCCESS;
}

