/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2007 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

#include "heap_algo_pool.h"

#include <string.h>


/***************************************************************************
 * Helper Macros
 ***************************************************************************/
#define  HEAP_POOL_MAGIC  0x100d100dul


/***************************************************************************
 * Data Structures
 ***************************************************************************/

struct Algorithm
{
   uint32_t id;

   struct HeapStatistics *stats;
   struct HeapConfiguration *config;
   
   uint32_t sig_size_count;
   size_t   sig_size[8];
};

/***************************************************************************/

struct Buffer
{
   HeapPID   owner;
   uint32_t  control;

#define BUF_ID_MASK                  0xC0000000ul
#define BUF_FREE_MASK                0x20000000ul
#define BUF_SIZE_MASK                0x1FFFFFFFul
#define BUF_ID(buf)                  ((buf)->control & BUF_ID_MASK)
#define BUF_FREE(buf)                ((buf)->control & BUF_FREE_MASK)
#define BUF_SIZE(buf)                ((buf)->control & BUF_SIZE_MASK)
#define BUF_SET_FREE(buf)            ((buf)->control |= BUF_FREE_MASK)
#define BUF_SET_CONTROL(buf,id,size) ((buf)->control = BUF_FREE_MASK | (size) | (id))
/* FIXME: variable endmark size. */
#define BUF_SET_ENDMARK(buf,size)    (*((char*)(buf) + size - 1) = HEAP_ENDMARK)
#define BUF_SET_OWNER(buf,pid)       ((buf)->owner = (pid))
};

/***************************************************************************/



/***************************************************************************
 * Helper Functions
 ***************************************************************************/

static inline size_t
l_buf_act_size(struct Algorithm *a, struct Buffer *buf)
{
   size_t size = BUF_SIZE(buf) - sizeof(struct Buffer) - a->config->buf_size_endmark; 
   uint32_t ix;

   for (ix=0; ix<a->sig_size_count; ix++)
   {
      if (size < a->sig_size[ix])
         return a->sig_size[ix];
   }
   
   return 0;
}

/***************************************************************************/

static inline size_t
l_buf_req_size(struct Algorithm *a, struct Buffer *buf)
{
   return BUF_SIZE(buf) - sizeof(struct Buffer) - a->config->buf_size_endmark; 
}

/***************************************************************************/




/***************************************************************************
 * Private Interface
 ***************************************************************************/

static HeapAction
l_pool_alloc(struct Heap *heap, union HeapEvent *event, void *opaque)
{
   struct Algorithm *a = opaque;
   struct Buffer *buf = NULL;
   size_t size = event->e_alloc.size + sizeof(struct Buffer) + a->config->buf_size_endmark;

   (void)heap;
   
   if (size > a->sig_size[a->sig_size_count-1])
      return HEAP_ACTION_CONTINUE;

   buf = (void *)alloc(size, 0);

   BUF_SET_OWNER(buf, event->e_alloc.owner);
   BUF_SET_CONTROL(buf, a->id, size);
   BUF_SET_ENDMARK(buf, size);

   a->stats->buf_size_req  += l_buf_req_size(a, buf);
   a->stats->buf_size_used += l_buf_act_size(a, buf);
   
   event->e_alloc.ptr  = &buf[1];
   event->e_alloc.size = size;
   
   return HEAP_ACTION_INTERCEPT;
}

/***************************************************************************/

static HeapAction
l_pool_free(struct Heap *heap, union HeapEvent *event, void *opaque)
{
   struct Algorithm *a = opaque;
   struct Buffer *buf = event->e_free.ptr;

   (void)heap;
   
   if (buf == NULL)
      return HEAP_ACTION_CONTINUE;

   buf = &buf[-1];

   if (BUF_ID(buf) != a->id)
      return HEAP_ACTION_CONTINUE;

   /* FIXME: Check endmark, free status, and owner. */
   
   a->stats->buf_size_req  -= l_buf_req_size(a, buf);
   a->stats->buf_size_used -= l_buf_act_size(a, buf);

   restore((union SIGNAL*)buf);
   free_buf((union SIGNAL**)&buf);
   
   return HEAP_ACTION_INTERCEPT;
}

/***************************************************************************/

static HeapAction
l_pool_query(struct Heap *heap, union HeapEvent *event, void *opaque)
{
   HeapParam *p = event->e_query.params;
   struct Algorithm *a = opaque;

   (void)heap;
   
   if (p == NULL)
      return HEAP_ACTION_CONTINUE;
   
   while (*p != HPM_END)
   {
      switch (HPM_ID(*p))
      {
         case HPM_QUERY_BUF_ACT_SIZE:
         {
            struct Buffer *buf = &((struct Buffer *)p[1])[-1];

            if (!HPM_IS_HANDLED(p[0]) && BUF_ID(buf) == a->id)
            {
               HPM_SET_HANDLED(p[0]);
               HPM_SET_VALUE(p[1], l_buf_act_size(a, buf));
            }
            break;
         }
         case HPM_QUERY_BUF_REQ_SIZE:
         { 
            struct Buffer *buf = &((struct Buffer *)p[1])[-1];

            if (!HPM_IS_HANDLED(p[0]) && BUF_ID(buf) == a->id)
            {
               HPM_SET_HANDLED(p[0]);
               HPM_SET_VALUE(p[1], l_buf_req_size(a, buf));
            }
            break;
         }
         case HPM_QUERY_BUF_OWNER:
         {
            struct Buffer *buf = &((struct Buffer *)p[1])[-1];

            if (!HPM_IS_HANDLED(p[0]) && BUF_ID(buf) == a->id)
            {
               HPM_SET_HANDLED(p[0]);
               HPM_SET_VALUE(p[1], buf->owner);
            }
            break;
         }
         case HPM_QUERY_BUF_TOT_REQ_SIZE:
         {
            if (!HPM_IS_HANDLED(p[0]))
            {
               HPM_SET_HANDLED(p[0]);
               HPM_SET_VALUE(p[1], a->stats->buf_size_req);
            }            
            break;
         }
         case HPM_QUERY_BUF_TOT_USED_SIZE:
         {
            if (!HPM_IS_HANDLED(p[0]))
            {
               HPM_SET_HANDLED(p[0]);
               HPM_SET_VALUE(p[1], a->stats->buf_size_used);
            }            
            break;
         }
         default:
         {
            break;
         }
      }

      p += (1 + HPM_VALUES(*p));
   }

   return HEAP_ACTION_CONTINUE;
}

/***************************************************************************
 * Public Interface
 ***************************************************************************/

HeapStatus
heap_attach_pool_algorithm(struct Heap *heap,
                           struct HeapStatistics *stats,
                           struct HeapConfiguration *config,
                           uint32_t id)
{
   HeapStatus status;
   uint32_t ix;

   PROCESS sigpool_id =
      get_sig_poolid(get_bid(current_process()));

   struct OS_poolcb *poolcb =
      get_poolcb(sigpool_id);

   struct Algorithm *a = heap_meta_alloc(heap, sizeof(struct Algorithm));


   if (poolcb == NULL || poolcb->sig_conf_entries == 0)
   {
      if (a != NULL)
         heap_meta_free(heap, a);

      return HEAP_ENOT_FOUND;
   }

   if (a == NULL)
   {
      free_buf((union SIGNAL**)&poolcb);
      return HEAP_EOUT_OF_MEMORY;
   }

   bzero(a, sizeof(struct Algorithm));

   status = zzheap_attach_hook(
      heap, HEAP_EVENT_ALLOC_REQ, l_pool_alloc, a, HEAP_PRIO_ALGO_NORMAL);
   if (status != HEAP_SUCCESS)
   {
      free_buf((union SIGNAL**)&poolcb);
      heap_meta_free(heap, a);
      return status;
   }

   status = zzheap_attach_hook(
      heap, HEAP_EVENT_FREE_REQ, l_pool_free, a, HEAP_PRIO_ALGO_NORMAL);
   if (status != HEAP_SUCCESS)
   {
      free_buf((union SIGNAL**)&poolcb);
      heap_meta_free(heap, a);
      return status;
   }

   status = zzheap_attach_hook(
      heap, HEAP_EVENT_QUERY_REQ, l_pool_query, a, HEAP_PRIO_ALGO_NORMAL);
   if (status != HEAP_SUCCESS)
   {
      free_buf((union SIGNAL**)&poolcb);
      heap_meta_free(heap, a);
      return status;
   }

   
   a->id             = id << 30;
   a->stats          = stats;
   a->config         = config;
   a->sig_size_count = poolcb->sig_conf_entries;

   
   for (ix=0; ix<a->sig_size_count; ix++)
   {
      a->sig_size[ix] = poolcb->values[poolcb->sig_conf_sizes + ix];
   }

   free_buf((union SIGNAL**)&poolcb);
   
   return HEAP_SUCCESS;
}

