/* COPYRIGHT-ENEA-SRC-R2 *
**************************************************************************
* Copyright (C) 2010-2013 by Enea Software AB.
* All rights reserved.
*
* This Software is furnished under a software license agreement and
* may be used only in accordance with the terms of such agreement.
* Any other use or reproduction is prohibited. No title to and
* ownership of the Software is hereby transferred.
*
* PROPRIETARY NOTICE
* This Software consists of confidential information.
* Trade secret law and copyright law protect this Software.
* The above notice of copyright on this Software does not indicate
* any actual or intended publication of such Software.
**************************************************************************
* COPYRIGHT-END */

 /* Enable fls() function when including heap_osal.h */
#define USE_HEAP_FLS

#include "heap_algo_tlsf.h"
#include "osarch.h"
#include <string.h>


/**************************************************************************
 * Plug-in History
 **************************************************************************
 *
 * IMPORTANT
 * 
 * This comment shall be updated with a description of the change each
 * time the extension version number is increased. The version number
 * shall be increased if incompatible modifications are made to internal
 * structures or after behavioral changes. See HeapExtension below.
 * 
 **************************************************************************
 *
 *  Version  Description
 *  -------  ----------------------------------------------------------
 *   101     Renamed struct Algorithm to struct TlsfAlgorithm.
 *   100     Initial revision.  
 *
 **************************************************************************/


/***************************************************************************
 * Helper Macros
 **************************************************************************/


#define TLSF_L1_COUNT  20
#define TLSF_L1_MIN    4
#define TLSF_L2_BITS   4
#define TLSF_L2_COUNT  16
#define TLSF_L2_MIN    0

#define TLSF_BUF_MIN                                                        \
   ((1 << TLSF_L1_MIN) + (TLSF_L2_MIN << (TLSF_L1_MIN - TLSF_L2_BITS)))

#define TLSF_BUF_MAX                                                    \
   ((1UL<<(TLSF_L1_COUNT+TLSF_L1_MIN)) -                                \
    (1UL<<(TLSF_L1_COUNT+TLSF_L1_MIN-1-TLSF_L2_BITS)))

#define TLSF_FRAG_MAX                                                   \
   (((BUF_SIZE_MASK >> BUF_SIZE_SHIFT) + 1) - sizeof(struct Fragment) -     \
    3 * sizeof(struct Buffer) - TLSF_BUF_MIN)

#define INDEX(l1, l2)                  ((l1) * TLSF_L2_COUNT + (l2))

#define FRAG_NEXT_FRAG_MASK        0xFFFFFFFCul
#define FRAG_NEXT_ID_MASK          0x00000003ul
#define FRAG_NEXT_ID_SHIFT         30
#define FRAG_NEXT_FRAG(frag)       ((struct Fragment*)((uint32_t)(frag) & FRAG_NEXT_FRAG_MASK))
#define FRAG_NEXT_ID(frag)         (((uint32_t)(frag) & FRAG_NEXT_ID_MASK) << FRAG_NEXT_ID_SHIFT)
#define FRAG_NEXT_MASK(frag, id)   ((struct Fragment*)((uint32_t)(frag)|((id) >> FRAG_NEXT_ID_SHIFT) ))


/***************************************************************************
 * Data Structures
 **************************************************************************/

/*
 * Buffer header.
 *
 * The administrative data of a buffer is split into two parts, a header
 * and a tail, that are dislocated in the memory space. Both use the 4 byte
 * 'Buffer' structure, leading to a total of 8 byte metadata.
 *
 * Layout of two adjacent buffers in the memory space:
 * 
 *  -------------------------------------------------------------------
 *  |        buffer                  |        buffer                  |
 *  -------------------------------------------------------------------
 *  | header | user data area | tail | header | user data area | tail |
 *  -------------------------------------------------------------------
 *
 * The header holds the most commonly used state of the buffer.
 *
 * The tail has two purposes:
 *   1) It holds the owner pid when a buffer is allocated.
 *   2) It holds the buffer size when a buffer is free, used 
 *      during merge operations to find the header of the buffer.
 *      A bit in the right buffer's header (see illustration above)
 *      indicates that the left buffer is free and that its tail holds
 *      the size of the left buffer. The left tail is directly adjacent
 *      to the right header. By using the size value in the left tail,
 *      the left buffer's header can be calculated as
 *      (right buffer pointer) - (left buffer size) and the two buffers
 *      can be merged into one.
 *
 * Control bits:
 * FREE    - The buffer is free.
 * PREV    - The left adjacent buffer is free and has a valid tail size.
 * GLUE    - The buffer is too small to be handed out to clients
 *           (only 8 bytes). All buffers must be split to the closest
 *           multiple of 8 bytes, but sometimes the remainder becomes too
 *           small, for example if a 24byte buffer is used to allocate
 *           16 bytes. A "glue" buffer is created for the small 8 byte
 *           remainder and marked as free but not inserted into freelists.
 *           It is merged as soon as one of its neighbours are freed.
 * SIZE    - Size of the buffer. Includes header, tail and endmark when used.
 * ANNOT   - Buffer is annotated with debug information.
 * UCHECK  - Lightweight integrity check for allocated buffers. Kept in header.
 * FCHECK  - Heavyweight integrity check for free buffers.
 *           Kept in tail when free. 
 * GC      - Indicates that the buffer is a hostage of the
 *           garbage collector. The buffer cannot be freed. If it
 *           is, the free operation is deferred and carried out by
 *           the garbage collector later. 
 */
struct Buffer
{
   uint32_t  control;
   
#define BUF_FREE_MASK          0x10000000ul
#define BUF_PREV_MASK          0x20000000ul
#define BUF_UCHECK_SRC_MASK    0x20000000ul
#define BUF_UCHECK_SRC_SHIFT   3
#define BUF_GLUE_MASK          0x08000000ul
#define BUF_ID_MASK            0xC0000000ul
#define BUF_UCHECK_MASK        0x04000000ul
#define BUF_GC_MASK            0x02000000ul
#define BUF_FCHECK_MASK        0xFE000000ul
#define BUF_SIZE_MASK          0x01FFFFFEul
#define BUF_SIZE_SHIFT         1
#define BUF_ID_SHIFT           30
#define BUF_UCHECK_SHIFT       25
#define BUF_FCHECK_SHIFT       25
#define BUF_FCHECK_MAGIC       0x80102040ul
#define BUF_ANNOT_MASK         0x00000001ul

#define buf_gc(buf)            ((buf)->control & BUF_GC_MASK)
#define buf_annotated(buf)     ((buf)->control & BUF_ANNOT_MASK)
#define buf_fbit(buf)          ((buf)->control & BUF_FREE_MASK)
#define buf_pbit(buf)          ((buf)->control & BUF_PREV_MASK)
#define buf_gbit(buf)          ((buf)->control & BUF_GLUE_MASK)
#define buf_size(buf)          (((buf)->control & BUF_SIZE_MASK) >> BUF_SIZE_SHIFT)
#define buf_id(buf)            ((buf)->control & BUF_ID_MASK)
#define buf_not_id(buf)        ((buf)->control & ~BUF_ID_MASK)
#define buf_tail(buf)          (&((struct Buffer *)((char *)buf + (buf_gbit(buf) ? buf_size(buf) : size_align(buf_size(buf)))))[-1])
#define buf_set_gc(buf)        ((buf)->control |= BUF_GC_MASK)
#define buf_set_tag(buf,size)  (buf_set_size(buf_tail(buf),(size)))
#define buf_set_fbit(buf)      ((buf)->control |= BUF_FREE_MASK)
#define buf_set_pbit(buf)      ((buf)->control |= BUF_PREV_MASK)
#define buf_set_gbit(buf)      ((buf)->control |= BUF_GLUE_MASK)
#define buf_set_id(buf, id)    ((buf)->control = buf_not_id(buf) | (id))
#define buf_set_size(buf,size) ((buf)->control = (((buf)->control & ~BUF_SIZE_MASK) | ((size) << BUF_SIZE_SHIFT)))
#define buf_clear_gc(buf)      ((buf)->control &= ~BUF_GC_MASK)
#define buf_clear_fbit(buf)    ((buf)->control &= ~BUF_FREE_MASK)
#define buf_clear_pbit(buf)    ((buf)->control &= ~BUF_PREV_MASK)
#define buf_clear_gbit(buf)    ((buf)->control &= ~BUF_GLUE_MASK)
#define buf_clear_anno(buf)    ((buf)->control &= ~BUF_ANNOT_MASK)
#define buf_set_owner(buf,pid) (buf_tail(buf)->control = (pid))
#define buf_owner(buf)         (buf_tail(buf)->control)
};

/**************************************************************************/

/*
 * The format of a free buffer.
 * Two words of user data are used as links for the freelists.
 */
struct FreeBuffer
{
   struct Buffer      buf;
   struct FreeBuffer *next;
   struct FreeBuffer *prev;
};

/**************************************************************************/

/*
 * Structure used for buffer iteration queries. (list all buffers)
 * All fields shall be zero initialized before the first query and
 * field values must be retained by clients during subsequent queries. 
 */
struct BufferCursor
{
   struct Buffer   *buf_out;
   struct Fragment *frag;
   struct Buffer   *buf;
   uint32_t         reserved1;
};

/***************************************************************************/

struct Region;

/*
 * TLSF Fragment. (part of region=
 *
 * Regions are broken down into fragments due to an upper limit on buffer
 * sizes. If a region is larger than 16MB, it will have more than one
 * fragment. A fragment can have any size below 16MB, there are no alignment
 * requirements or series to follow, except for the minimum heap alignment
 * of 8 bytes.
 */
struct Fragment
{
   /* Purposely unaligned size. Do not change alignment. */
   struct Fragment *next;            /* List link. */
   struct Fragment *prev;            /* List link. */
   struct Buffer   *root;            /* Pointer to first buffer in fragment.  */
   struct Buffer   *tail;            /* Pointer to last buffer in fragment.   *
                                        The last buffer is used to stop merge * 
                                        operations at the root and cannot be  *
                                        allocated.                            */
   struct Region   *region;          /* Region that the fragment belongs to.  */
   uint32_t         chksum;          /* Integrity checksum.                   */
   uint32_t         control;         /* Control bits.                         */
   
#define FRAG_SIZE_MASK           0x1FFFFFFFul
#define frag_size(frag)          ((frag)->control & FRAG_SIZE_MASK)
#define frag_set_size(frag,size) ((frag)->control = (((frag)->control & ~FRAG_SIZE_MASK) | (size)))

#define FRAG_CHECKSUM_MAGIC      0x80102040ul
#define frag_checksum(_r)                       \
     ((uint32_t)(_r)->next ^                    \
      (uint32_t)(_r)->prev ^                    \
      (uint32_t)(_r)->root ^                    \
      (uint32_t)(_r)->tail ^                    \
      (uint32_t)(_r)->region ^                  \
      (uint32_t)(_r)->control ^                 \
      FRAG_CHECKSUM_MAGIC)
#define frag_set_checksum(_r)    ((_r)->chksum = frag_checksum(_r))
#define frag_get_checksum(_r)    ((_r)->chksum)
};

/**************************************************************************/

/*
 * External memory region. 
 */
struct Region
{
   struct Region   *next;           /* List link.                      */
   struct Region   *prev;           /* List link.                      */
   struct Fragment *frag_list;      /* First fragment in region.       */
   struct Fragment *frag_last;      /* End of last fragment in region  *
                                     * (not the last fragment header). */
   uint32_t         chksum;         /* Integrity checksum.             */
   uint32_t         control;        /* Region control bits. See below. */
   
/* Bit descriptions:
 *
 * INITIAL - Initial heap region allocated at heap creation.
 *           Can never be removed.
 * ID      - ID of originating allocator.
 * BUF_ID  - ID of algorithm that allocated the region.
 * LOCK    - Region is being traversed and cannot be deallocated.
 */

#define REGION_BUF_ID_MASK             0xC0000000ul
#define REGION_BUF_ID_SHIFT            30
#define REGION_ID_MASK                 0x30000000ul
#define REGION_ID_SHIFT                28
#define REGION_INITIAL_MASK            0x08000000ul
#define REGION_LOCK_MASK               0x04000000ul
#define region_initial(reg)            ((reg)->control & REGION_INITIAL_MASK)
#define region_set_initial(reg)        ((reg)->control |= REGION_INITIAL_MASK)
#define region_buf_id(reg)             ((reg)->control & REGION_BUF_ID_MASK)
#define region_id(reg)                 ((reg)->control & REGION_ID_MASK)
#define region_set_control(reg,buf_id) ((reg)->control = (buf_id))
#define region_set_lock(reg)           ((reg)->control |= REGION_LOCK_MASK)
#define region_clear_lock(reg)         ((reg)->control &= ~REGION_LOCK_MASK)
#define region_locked(reg)             ((reg)->control & REGION_LOCK_MASK)

#define REGION_CHECKSUM_MAGIC          0x80102040ul
#define region_checksum(_r)                     \
   ((uint32_t)(_r)->next ^                      \
    (uint32_t)(_r)->prev ^                      \
    (uint32_t)(_r)->frag_list ^                 \
    (uint32_t)(_r)->frag_last ^                 \
    (uint32_t)((_r)->control & ~REGION_LOCK_MASK) ^  \
    REGION_CHECKSUM_MAGIC)            
#define region_set_checksum(_r)       ((_r)->chksum = region_checksum(_r))
#define region_get_checksum(_r)       ((_r)->chksum)
};

/**************************************************************************/

/*
 * Administrative data for the algorithm. 
 */
struct TlsfAlgorithm
{
   uint32_t id;                      /* Buffer algorithm identifier. */

   struct HeapStatistics *stats;
   struct HeapConfiguration *config;

   /* Buffers. */
   struct FreeBuffer *list[TLSF_L1_COUNT][TLSF_L2_COUNT]; /* Freelists. */
   size_t             l1_bitmap;                    /* First level bitmap  */
   size_t             l2_bitmap[TLSF_L1_COUNT];     /* Second level bitmap */

   /* Fragments. */
   struct Region     *region_list;                  /* List of regions     */
   uint32_t           region_count;                 /* Number of regions.  */
   uint32_t           frag_count;                   /* Number of fragments.*/

   struct HeapServiceInfo info;                     /* Plug-in information */
};

/**************************************************************************/





/***************************************************************************
 * Forward Declarations
 **************************************************************************/

static inline size_t
buf_act_size(struct TlsfAlgorithm *a, struct Buffer *buf);

static inline size_t
buf_req_size(struct TlsfAlgorithm *a, struct Buffer *buf);

static inline void
buf_set_ucheck(struct Buffer *buf);

static inline void
buf_set_fcheck(struct Buffer *buf);

static inline void
buf_set_check(struct Buffer *buf);

static inline uint32_t
buf_ucheck(struct Buffer *buf);

static always_inline uint32_t
buf_fcheck(struct Buffer *buf);

static inline uint32_t
buf_get_ucheck(struct Buffer *buf);

static always_inline uint32_t
buf_get_fcheck(struct Buffer *buf);

static inline HeapError
buf_check_frag(struct Heap *heap, struct TlsfAlgorithm *a,
               struct Fragment **frag_out, struct Buffer *buf);

static inline HeapError
buf_check_header(struct Heap *heap, struct TlsfAlgorithm *a,
                 struct Fragment *reg, struct Buffer *buf);

static inline HeapError
buf_check_endmark(struct Heap *heap, struct TlsfAlgorithm *a,
                  struct Fragment *frag, struct Buffer *buf);

/**************************************************************************/

static inline struct Region *
region_first(struct Heap *heap, struct TlsfAlgorithm *a, HeapError *ecode);

static inline struct Region *
region_next(struct Heap *heap, struct Region *reg, HeapError *ecode);

static inline struct Region *
region_find(struct Heap *heap, struct TlsfAlgorithm *a,
            void *ptr, uint32_t exact, HeapError *ecode);

/**************************************************************************/

static inline struct Fragment *
frag_first(struct Heap *heap, struct Region *reg, HeapError *ecode);

static inline struct Fragment *
frag_next(struct Heap *heap, struct Fragment *frag, HeapError *ecode);

static inline struct Fragment *
frag_find(struct Heap *heap, struct Region *reg,
          void *ptr, uint32_t exact, HeapError *ecode);

/**************************************************************************/




/***************************************************************************
 * Helper Functions
 **************************************************************************/

#define BITMAP_SET(bitmap, bit)        ((bitmap) |= (0x80000000ul >> bit))
#define BITMAP_CLEAR(bitmap, bit)      ((bitmap) &= ~(0x80000000ul >> bit))

/***************************************************************************/

static always_inline long
BITMAP_FIND(size_t bitmap, size_t min)
{
   /*
    * Find the most significant bit set in a
    * bitmap. Should return -1 if bitmap is 0.
    *
    * The "min" parameter identifies where
    * the search should start.
    *****************************************/

   bitmap &= (~0ul >> (min));
   if (unlikely(!bitmap))
      return -1;
   return 31 - fls(bitmap);
}

/**************************************************************************/

static always_inline size_t
size_align(size_t size)
{
   if (size < TLSF_BUF_MIN)
      return TLSF_BUF_MIN;

   /* Required alignment, sizeof(long double). */
   return (size + HEAP_ALIGNMENT - 1) & ~(HEAP_ALIGNMENT - 1);
}

/**************************************************************************/

static always_inline size_t
size_align_down(size_t size)
{
   if (unlikely(size > TLSF_FRAG_MAX))
      return TLSF_FRAG_MAX;

   return size & ~(HEAP_ALIGNMENT - 1);
}

/**************************************************************************/

static always_inline size_t
size_align_tlsf(size_t size)
{
   int l1 = fls(size) - TLSF_L2_BITS;
   
   /* Aligned to a predefined TLSF bucket size. */
   return (size + ((1 << l1) - 1)) &
      ~((1 << l1) - 1);
}

/**************************************************************************/

static always_inline void
index_search(size_t size, uint32_t *l1, uint32_t *l2)
{
   /* Determine the indecies of a size. */
   
   size = size_align(size);

   *l1 = fls(size);
   *l2 = (size - (1 << (*l1))) >> ((*l1) - TLSF_L2_BITS);
   *l1 = *l1 - TLSF_L1_MIN;
}

/**************************************************************************/

static always_inline void
index_search_align(size_t size, uint32_t *l1, uint32_t *l2)
{
   size = size_align(size);
   size = size_align_tlsf(size);

   /* Determine the indecies of the aligned size. */

   *l1 = fls(size);
   *l2 = (size - (1 << (*l1))) >> ((*l1) - TLSF_L2_BITS);
   *l1 = *l1 - TLSF_L1_MIN;
}

/********************** BUFFER MANIPULATORS *******************************/

static always_inline struct Buffer *
buf_first(struct Heap *heap, struct TlsfAlgorithm *a, struct Fragment *frag)
{
   /* Helper function for linear traversal of
    * a single fragment. The header of each buffer is
    * validated before the buffer pointer is advanced.
    *
    * The heap lock must be aquired for the duration of
    * iterative calls to this function. Any interleaved
    * heap access may make pointers invalid and result in
    * an error or access violation. */

   struct Buffer *first = frag->root;
   HeapError ecode;

   ecode = buf_check_header(heap, a, frag, first);
   if (unlikely(ecode != HEAP_ERROR_NONE))
   {
      zzheap_error(heap, ecode, &first[1]);
      return NULL;
   }

   return first;
}

/**************************************************************************/

static always_inline struct Buffer *
buf_left(struct Heap *heap, struct TlsfAlgorithm *a,
         struct Buffer *buf, uint32_t check)
{
   /* Helper function to locate the header of the left
    * adjacent buffer. Uses the tail size (boundary tag) of
    * the left buffer. Returns NULL if the left buffer is
    * allocated.
    *
    * It is important that the first buffer in the fragment
    * doesn't have the PREV-bit set as that would allow this
    * function to continue and move outside the fragment.
    * This is taken care of in the function l_frag_break()
    * during allocation and insertion of new fragments.
    * (frag->root->control = ...). Since there are no
    * buffers left of the first buffer, the bit cannot
    * be set during free() either. 
    *
    * The heap lock must be aquired for the duration of
    * iterative calls to this function. Any interleaved
    * heap access may make pointers invalid and result in
    * an error or access violation. */

   struct Buffer *left = (struct Buffer *)((char *)buf - buf_size(&buf[-1]));
#ifdef ASSERT
   HeapError ecode;
#else
   (void)heap;
   (void)a;
   (void)check;
#endif

   if (!buf_pbit(buf))
      return NULL;

#ifdef ASSERT
   /* Sanity check. */
   if (likely(check))
   {
      ecode = buf_check_header(heap, a, NULL, left);
      if (unlikely(ecode != HEAP_ERROR_NONE))
      {
         zzheap_error(heap, ecode, &left[1]);
         return NULL;
      }
   }
#endif

   return left;
}

/**************************************************************************/

static always_inline struct Buffer *
buf_right(struct Heap *heap, struct TlsfAlgorithm *a,
          struct Buffer *buf, uint32_t check)
{
   /* Helper function for linear traversal of
    * a single fragment. The header of each buffer is
    * validated before the buffer pointer is advanced.
    *
    * The heap lock must be aquired for the duration of
    * iterative calls to this function. Any interleaved
    * heap access may make pointers invalid and result in
    * an error or access violation. */

   struct Buffer *right = (struct Buffer *)((char *)buf + buf_act_size(a, buf));
   HeapError ecode;

   /* Stop if input buffer is the fragment sentinel buffer.
    * Note that the sentinel has to be returned
    * once to support split and merge operations. */
   if (buf_size(buf) == 0)
      return NULL;

   /* Sanity check. */
   if (likely(check))
   {
      ecode = buf_check_header(heap, a, NULL, right);
      if (unlikely(ecode != HEAP_ERROR_NONE))
      {
         zzheap_error(heap, ecode, &right[1]);
         return NULL;
      }
   }

   return right;
}

/**************************************************************************/

static inline struct Buffer *
buf_next(struct Heap *heap, struct TlsfAlgorithm *a,
         struct BufferCursor *cursor, uint8_t advance_frag)
{
   /* Helper function for safe traversal of all the
    * algorithm's buffers. 
    *
    * This function must not generate any errors since it
    * may be called from ose_heapd in response to signal
    * requests. OSE Optima, for example, may bring down the
    * entire system if an error or exception occurs.
    *
    * If 'advance_frag' is false, iteration will stop when
    * it reaches the end of a fragment. It is set to list
    * buffers in a single fragment only, rather than the
    * whole heap. A fragment to start with must therefore
    * also be provided in the cursor. 
    * */

   struct Region *reg;
   struct Fragment *frag = FRAG_NEXT_FRAG(cursor->frag);
   struct Buffer *buf = cursor->buf, *next;
   HeapError ecode;

   /* Locate the input fragment and make sure it still
    * exists. If it does, advance to the next fragment if
    * there are no more buffers in the current fragment. */
   if (unlikely(buf == NULL))
   {
      if (likely(advance_frag))
      {
         reg = region_first(heap, a, &ecode);
         if (reg == NULL)
            return NULL;

         frag = frag_first(heap, reg, &ecode);
         if (frag == NULL)
            return NULL;

         cursor->frag = FRAG_NEXT_MASK(frag, a->id);
         return cursor->buf = frag->root;
      }
      else
      {
         reg = region_find(heap, a, frag, 1, &ecode);
         if (reg == NULL)
            return NULL;

         frag = frag_first(heap, reg, &ecode);
         if (frag == NULL)
            return NULL;

         cursor->frag = FRAG_NEXT_MASK(frag, a->id);
         return cursor->buf = frag->root;
      }
   }

   ecode = buf_check_frag(heap, a, &frag, buf);
   if (unlikely(ecode != HEAP_ERROR_NONE))
      return NULL;
   
   ecode = buf_check_header(heap, a, frag, buf);
   if (unlikely(ecode != HEAP_ERROR_NONE))
      return NULL;

   next = buf;
   
   do {
      next = buf_right(heap, a, next, 0);

      if (next != NULL)
      {
         ecode = buf_check_header(heap, a, frag, next);
         if (unlikely(ecode != HEAP_ERROR_NONE)) {
            break;
         }
      }
   } while (next != NULL && buf_gbit(next));


   /* Tail buffer reached? */
   if (buf_size(next) == 0 && buf_ucheck(next) == buf_get_ucheck(next))
   {
      reg = frag->region;

      frag = frag_next(heap, frag, &ecode);
      if (frag != NULL)
      {
         cursor->frag = FRAG_NEXT_MASK(frag, a->id);
         return cursor->buf = frag->root;
      }

      if (!advance_frag)
         return NULL;
      
      reg = region_next(heap, reg, &ecode);
      if (reg == NULL)
         return NULL;

      frag = frag_first(heap, reg, &ecode);
      if (frag == NULL)
         return NULL;

      
      cursor->frag = FRAG_NEXT_MASK(frag, a->id);
      return cursor->buf = frag->root;
   }

   cursor->frag = FRAG_NEXT_MASK(frag, a->id);
   return cursor->buf = next;
}

/**************************************************************************/

static always_inline size_t
buf_act_size(struct TlsfAlgorithm *a, struct Buffer *buf)
{
   (void)a;

   /* Return the actual size of the buffer. */
   
   if (buf_gbit(buf))
      return buf_size(buf);
   return size_align(buf_size(buf));
}

/**************************************************************************/

static always_inline size_t
buf_req_size(struct TlsfAlgorithm *a, struct Buffer *buf)
{
   /* Return the requested size of the buffer. */
   return buf_size(buf) - 2 * sizeof(struct Buffer) - a->config->buf_size_endmark;
}

/**************************************************************************/

static inline HeapBufferStatus
buf_status(struct Heap *heap, struct TlsfAlgorithm *a,
           struct Fragment *frag, struct Buffer *buf)
{
   /* Helper function that investigates the health of
    * an individual buffer. */
   
   HeapBufferStatus status = 0;
   HeapError ecode;

   ecode = buf_check_header(heap, a, frag, buf);
   if (unlikely(ecode != HEAP_ERROR_NONE))
   {
      if (ecode == HEAP_ERROR_BUF_UNALIGNED)
         return HEAP_BUFFER_UNKNOWN;

      if (unlikely(buf_annotated(buf)))
         return HEAP_BUFFER_HEADERS|HEAP_BUFFER_ANNOTATED;

      return HEAP_BUFFER_HEADERS;
   }
   
   if (!buf_fbit(buf))
      status |= HEAP_BUFFER_ALLOCATED;

   ecode = buf_check_endmark(heap, a, frag, buf); 
   if (unlikely(ecode != HEAP_ERROR_NONE))
      status |= HEAP_BUFFER_ENDMARK;
   
   if (unlikely(buf_annotated(buf)))
      status |= HEAP_BUFFER_ANNOTATED;

   return status;
}

/**************************************************************************/

static inline void 
buf_insert(struct Heap *heap, struct TlsfAlgorithm *a,
           struct Buffer *buf, uint32_t l1, uint32_t l2)
{
   /* Inserts a buffer into the heap freelists and
    * updates the freelist bitmaps to indicate that
    * the list has a buffer available.
    *
    * The function calculates a header checksum which
    * is stored in the tail and later validated
    * by the buf_check_header() function. */

   struct FreeBuffer *fb = (struct FreeBuffer *)buf;
   struct Buffer *right;
   HeapError ecode;
   
   fb->prev = NULL;
   fb->next = a->list[l1][l2];
   if (fb->next)
   {
#ifdef ASSERT
      ecode = buf_check_header(heap, a, NULL, &fb->next->buf);
#else
      ecode = buf_fcheck(&fb->next->buf) != buf_get_fcheck(&fb->next->buf) ?
         HEAP_ERROR_BUF_LINKS : HEAP_ERROR_NONE;
#endif 
      if (unlikely(ecode != HEAP_ERROR_NONE))
      {
         zzheap_error(heap, HEAP_ERROR_FATAL|ecode, &(&fb->next->buf)[1]);
         return;
      }

      fb->next->prev = fb;

      buf_set_check(&fb->next->buf);
   }

   a->list[l1][l2] = fb;

   buf_set_fbit(buf);
   buf_set_tag(buf, buf_size(buf));
   buf_set_check(buf);

#ifdef ASSERT
   right = buf_right(heap, a, buf, 1);
#else
   right = buf_right(heap, a, buf, 0);
#endif
   buf_set_pbit(right);
   buf_set_ucheck(right);
   
   BITMAP_SET(a->l1_bitmap, l1);
   BITMAP_SET(a->l2_bitmap[l1], l2);

   a->stats->buf_num_used[l1]--;
   a->stats->buf_num_free[l1]++;
   a->stats->buf_size_free += buf_size(buf);
}

/**************************************************************************/

static inline void
buf_remove(struct Heap *heap, struct TlsfAlgorithm *a,
           struct Buffer *buf, uint32_t l1, uint32_t l2)
{
   /* Removes a buffer from the heap freelists and
    * updates the freelist bitmap to indicate that
    * it is empty, if it has become empty.
    *
    * The integrity checksums of the buffer and its
    * adjacent buffers in the list are validate and
    * updated if affected. 
    * */

   struct FreeBuffer *fb = (struct FreeBuffer *)buf;
   struct Buffer *right;
   HeapError ecode;

#ifdef ASSERT
   ecode = buf_check_header(heap, a, NULL, buf);
#else
   ecode = buf_fcheck(buf) != buf_get_fcheck(buf) ?
      HEAP_ERROR_BUF_LINKS : HEAP_ERROR_NONE;
#endif

   if (unlikely(ecode != HEAP_ERROR_NONE))
   {
      zzheap_error(heap, HEAP_ERROR_FATAL|ecode, &buf[1]);
      return;
   }
   
   if (fb->prev)
   {
#ifdef ASSERT
      ecode = buf_check_header(heap, a, NULL, &fb->prev->buf);
#else
      ecode = buf_fcheck(&fb->prev->buf) != buf_get_fcheck(&fb->prev->buf) ?
         HEAP_ERROR_BUF_LINKS : HEAP_ERROR_NONE;
#endif
      if (unlikely(ecode != HEAP_ERROR_NONE))
      {
         zzheap_error(heap, HEAP_ERROR_FATAL|ecode, &(&fb->prev->buf)[1]);
         return;
      }

      fb->prev->next = fb->next;
      buf_set_check(&fb->prev->buf);
   }

   if (fb->next)
   {
#ifdef ASSERT
      ecode = buf_check_header(heap, a, NULL, &fb->next->buf);
#else      
      ecode = buf_fcheck(&fb->next->buf) != buf_get_fcheck(&fb->next->buf) ?
         HEAP_ERROR_BUF_LINKS : HEAP_ERROR_NONE;
#endif
      if (unlikely(ecode != HEAP_ERROR_NONE))
      {
         zzheap_error(heap, HEAP_ERROR_FATAL|ecode, &(&fb->next->buf)[1]);
         return;
      }

      fb->next->prev = fb->prev;
      buf_set_check(&fb->next->buf);
   }

   if (a->list[l1][l2] == fb)
      a->list[l1][l2] = fb->next;

   if (!a->list[l1][l2])
   {
      BITMAP_CLEAR(a->l2_bitmap[l1], l2);

      if (!a->l2_bitmap[l1])
         BITMAP_CLEAR(a->l1_bitmap, l1);
   }

   fb->next = fb->prev = NULL;

   buf_clear_fbit(buf);
   buf_set_ucheck(buf);

#ifdef ASSERT
   right = buf_right(heap, a, buf, 1);
#else
   right = buf_right(heap, a, buf, 0);
#endif
   buf_clear_pbit(right);
   buf_set_ucheck(right);

   a->stats->buf_num_used[l1]++;
   a->stats->buf_num_free[l1]--;
   a->stats->buf_size_free -= buf_size(buf);
}

/**************************************************************************/

static always_inline struct Buffer *
buf_find(struct TlsfAlgorithm *a, uint32_t *l1, uint32_t *l2)
{
   /* Searches the freelist bitmaps for any buffer that
    * has an index equal to or larger than the requested
    * index. */

   long i;

   if (unlikely(*l1 >= TLSF_L1_COUNT))
   {
      return NULL;
   }
   
   i = BITMAP_FIND(a->l2_bitmap[*l1], *l2);
   if (i >= 0)
   {
      *l2 = i;
      return &a->list[*l1][*l2]->buf;
   }
   else
   {
      i = BITMAP_FIND(a->l1_bitmap, *l1 + 1);
      if (i >= 0)
      {
         *l1 = i;
         *l2 = BITMAP_FIND(a->l2_bitmap[*l1], 0);
         return &a->list[*l1][*l2]->buf;
      }
   }

   return NULL;
}

/**************************************************************************/

static inline struct Buffer *
buf_merge_left(struct Heap *heap, struct TlsfAlgorithm *a,
               struct Buffer *buf, size_t req_size)
{
   /* Merges a buffer with its left adjacent neighbour if
    * possible. The resulting buffer is returned. */
   
   struct Buffer *left;
   size_t avail_size = buf_act_size(a, buf);
   uint32_t l1, l2;

   left = buf_left(heap, a, buf, 1);
   if (left == NULL)
      return NULL;
   
   if (!buf_fbit(left) || buf_size(left) + avail_size < req_size)
      return NULL;

   index_search(buf_size(buf), &l1, &l2);
   a->stats->buf_num_used[l1]--;

   if (!buf_gbit(left))
   { 
      index_search(buf_size(left), &l1, &l2);
      a->stats->buf_num_used[l1]--;

      buf_remove(heap, a, left, l1, l2);
   }
   else
   {
      buf_clear_gbit(left);
      buf_clear_fbit(left);
   }

   buf_set_size(left, buf_act_size(a, buf) + buf_size(left));
   buf_set_ucheck(left);

   index_search(buf_size(left), &l1, &l2);
   a->stats->buf_num_used[l1]++;

   return left;
}

/**************************************************************************/

static inline struct Buffer *
buf_merge_right(struct Heap *heap, struct TlsfAlgorithm *a,
                struct Buffer *buf, size_t req_size)
{
   /* Merges a buffer with its right adjacent neighbour if
    * possible. The resulting buffer is returned. */

   struct Buffer *right;
   size_t avail_size = buf_act_size(a, buf);
   uint32_t l1, l2;

#ifdef ASSERT
   right = buf_right(heap, a, buf, 1);
#else
   right = buf_right(heap, a, buf, 0);
#endif
   if (right == NULL || !buf_fbit(right) || buf_size(right) + avail_size < req_size)
      return NULL;

   index_search(buf_size(buf), &l1, &l2);
   a->stats->buf_num_used[l1]--;

   if (!buf_gbit(right))
   {
      index_search(buf_size(right), &l1, &l2);
      a->stats->buf_num_used[l1]--;

      buf_remove(heap, a, right, l1, l2);
   }
   else
   {
      struct Buffer *right_right;

#ifdef ASSERT
      right_right = buf_right(heap, a, right, 1);
#else
      right_right = buf_right(heap, a, right, 0);
#endif
      buf_clear_pbit(right_right);
      buf_set_ucheck(right_right);
   }

   buf_set_size(buf, buf_act_size(a, buf) + buf_size(right));
   buf_set_ucheck(buf);

   index_search(buf_size(buf), &l1, &l2);
   a->stats->buf_num_used[l1]++;

   return buf;
}

/**************************************************************************/

static inline struct Buffer *
buf_split_right(struct Heap *heap, struct TlsfAlgorithm *a,
                struct Buffer *buf, size_t keep_size)
{
   /* Splits a buffer in two parts. The unused remainder
    * is inserted into the freelists and made available
    * for later allocation. */

   struct Buffer *right;
   size_t right_size;
   uint32_t l1, l2;

   keep_size  = size_align(keep_size);
   right_size = buf_act_size(a, buf) - keep_size;

   if (right_size < 2 * sizeof(struct Buffer))
      return NULL;
   

   index_search(buf_size(buf), &l1, &l2);
   a->stats->buf_num_used[l1]--;
   
   buf_set_size(buf, keep_size);

   right = buf_right(heap, a, buf, 0);
   right->control = 0;
   buf_set_id(right, a->id);
   buf_set_size(right, right_size);
   
   /* Only insert the right piece in freelists if it can accomodate
    * the smallest supported buffer size. If it is smaller, the right
    * piece only acts as glue between its siblings. This glue is needed
    * to support "requested size" statistics. */
   if (right_size >= TLSF_BUF_MIN)
   {
      index_search(right_size, &l1, &l2);
      a->stats->buf_num_used[l1]++;

      buf_insert(heap, a, right, l1, l2);
   }
   else
   {
      struct Buffer *right_right;

      /* Create and initialize the glue buffer.
       * Mark it free in its right neighbour's header.
       * The glue buffer is not inserted into freelists. */

      buf_set_fbit(right);
      buf_set_gbit(right);
      buf_set_tag(right, buf_size(right));
      buf_set_ucheck(right);

#ifdef ASSERT
      right_right = buf_right(heap, a, right, 1);
#else
      right_right = buf_right(heap, a, right, 0);
#endif
      buf_set_pbit(right_right);
      buf_set_ucheck(right_right);
   }

   index_search(keep_size, &l1, &l2);
   a->stats->buf_num_used[l1]++;

   return right;
}

/***************************************************************************/

static always_inline void
buf_set_endmark(struct TlsfAlgorithm *a, struct Buffer *buf)
{
   /* Initializes the buffer endmark, which is used to
    * detect application bugs. If the endmark becomes
    * damaged the application has used more memory in
    * the buffer than it was allowed to. */
   
   char *endm = (char *)&buf[1] + buf_req_size(a, buf);
   unsigned i;

   for (i=0; i<a->config->buf_size_endmark; i++)
      endm[i] = HEAP_ENDMARK;
}

/***************************************************************************/

static always_inline uint32_t
buf_ucheck(struct Buffer *buf)
{
#ifdef ASSERT
   uint32_t chk = (buf->control & ~(BUF_UCHECK_MASK|BUF_ANNOT_MASK)) ^ BUF_FCHECK_MAGIC;
   char c, *p = (char *)&chk;   

   c = (p[0] ^ p[2]) ^ (p[1] ^ p[3]);

   chk = c ^ (c >> 2) ^ (c >> 4) ^ (c >> 6) ^ 0x2;

   return (chk & 0x3) << BUF_UCHECK_SHIFT;
#else
   /* "Calculate" the header integrity checksum */
   return ((~buf->control) & BUF_UCHECK_SRC_MASK) >> BUF_UCHECK_SRC_SHIFT;
#endif
}

/***************************************************************************/

static always_inline uint32_t
buf_fcheck(struct Buffer *buf)
{
   struct FreeBuffer *fb = (struct FreeBuffer *)buf;
   struct Buffer *tail = buf_tail(buf);

   uint32_t chk =
      (uint32_t)fb->next ^
      (uint32_t)fb->prev ^
      (uint32_t)(tail->control & ~BUF_FCHECK_MASK) ^
      BUF_FCHECK_MAGIC;

#ifdef ASSERT
   char *p = (char *)&chk;

   chk = (p[0] ^ p[2]) ^ (p[1] ^ p[3]);
   return ((uint32_t)chk << (BUF_FCHECK_SHIFT)) & BUF_FCHECK_MASK;
#else
   /* "Calculate" the header and list link integrity checksum */
   return ((uint32_t)chk << (BUF_FCHECK_SHIFT)) & BUF_FCHECK_MASK;
#endif
}

/***************************************************************************/

static always_inline uint32_t
buf_get_ucheck(struct Buffer *buf)
{
   return buf->control & BUF_UCHECK_MASK;
}

/***************************************************************************/

static always_inline uint32_t
buf_get_fcheck(struct Buffer *buf)
{
   return buf_tail(buf)->control & BUF_FCHECK_MASK;
}

/***************************************************************************/

static always_inline void
buf_set_ucheck(struct Buffer *buf)
{
   buf->control = (buf->control & ~BUF_UCHECK_MASK) | buf_ucheck(buf);
}

/***************************************************************************/

static always_inline void
buf_set_fcheck(struct Buffer *buf)
{
   struct Buffer *tail = buf_tail(buf);
   tail->control = (tail->control & ~BUF_FCHECK_MASK)|buf_fcheck(buf);
}

/***************************************************************************/

static always_inline void
buf_set_check(struct Buffer *buf)
{
   /* Update integrity checksums. */
   
   buf_set_ucheck(buf);
   
   if (buf_fbit(buf) && !buf_gbit(buf))
      buf_set_fcheck(buf);
}

/***************************************************************************/

static always_inline HeapError
buf_check_owner(struct Heap *heap, struct TlsfAlgorithm *a, struct Buffer *buf, HeapPID pid1)
{
   /* Validates that the pid passed as parameter is the
    * owner of the buffer. Private buffers can only be
    * operated on by its owner. */
   
   HeapPID pid2 = buf_owner(buf);

   (void)heap;
   (void)a;
   
   /* A buffer owner value of 0 (pid2) indicates that the
    * buffer is shared and can be operated on by anyone.
    * An originator pid value of 0 (pid1) means that the
    * operation was forced and should be allowed. */

   if (unlikely(pid1 != 0 && pid2 != 0 && pid1 != pid2))
      return HEAP_ERROR_BUF_OWNER;

   return HEAP_ERROR_NONE;
}

/***************************************************************************/

static inline HeapError
buf_check_frag(struct Heap *heap, struct TlsfAlgorithm *a,
               struct Fragment **frag_out, struct Buffer *buf)
{
   /* Validates that the buffer pointer resides in one
    * of the algorithm's fragments. If so, it should be
    * safe to access the address. */
   
   struct Region *reg;
   struct Fragment *frag;
   HeapError ecode = HEAP_ERROR_NONE;

   /* Regions/fragments may disappear at any time due to
    * unbreak requests. Locks are needed. */
   reg = region_find(heap, a, buf, 0, &ecode);
   if (unlikely(reg == NULL))
   {
      if (ecode != HEAP_ERROR_NONE)
         return ecode;
   
      return HEAP_ERROR_BUF_UNKNOWN;
   }

   frag = frag_find(heap, reg, buf, 0, &ecode);
   if (unlikely(frag == NULL))
   {
      if (ecode != HEAP_ERROR_NONE)
         return ecode;
   
      return HEAP_ERROR_BUF_UNKNOWN;
   }

   *frag_out = frag;
   return HEAP_ERROR_NONE;
}

/***************************************************************************/

static inline HeapError
buf_check_header(struct Heap *heap, struct TlsfAlgorithm *a,
                 struct Fragment *frag, struct Buffer *buf)
{
   /* Validate that the buffer header is reasonable.
    * Free buffers are validated through a checksum
    * of the entire header, including freelist links. 
    * It is stored in the owner field.
    *
    * The validation of allocated buffers is too weak
    * and needs improvment. 
    *
    * If a buffer is not reqognized by an algorithm,
    * the event hook must return the continue action
    * to allow other algorithms to identify it. If no
    * algorithm responds, i.e. all request type event
    * hooks return "continue", the buffer is likely
    * corrupted and the framework will generate an
    * error. */

   (void)heap;
   
   if (unlikely((((uintptr_t)buf) & (HEAP_ALIGNMENT-1)) != 0x4 &&
                (((uintptr_t)buf) & (HEAP_ALIGNMENT-1)) != 0xC))
      return HEAP_ERROR_BUF_UNALIGNED;

   /* If the fragment is unknown, the ID must be verified. */
   if (unlikely(frag == NULL && buf_id(buf) != a->id))
      return HEAP_ERROR_BUF_UNKNOWN;
   /* If the fragment is known, the ID must also be. */
   else if (unlikely(frag != NULL && buf_id(buf) != a->id))
      return HEAP_ERROR_BUF_HEADER;
   
   if (unlikely(buf_ucheck(buf) != buf_get_ucheck(buf)))
      return HEAP_ERROR_BUF_HEADER;

   if (unlikely(frag != NULL))
   {
      size_t size = buf_act_size(a, buf);

      if ((uintptr_t)buf + size > (uintptr_t)frag->tail ||
          (uintptr_t)buf + size < (uintptr_t)frag->root)
         return HEAP_ERROR_BUF_HEADER;
   }
   
   if (buf_fbit(buf))
   {
      if (!buf_gbit(buf) && buf_get_fcheck(buf) != buf_fcheck(buf))
         return HEAP_ERROR_BUF_LINKS;

      if (buf_size(buf) > TLSF_FRAG_MAX)
         return HEAP_ERROR_BUF_HEADER;

      if ((buf_gbit(buf) && buf_size(buf) >= TLSF_BUF_MIN) ||
          (!buf_gbit(buf) && buf_size(buf) < TLSF_BUF_MIN))
         return HEAP_ERROR_BUF_HEADER;
   }
   else
   {
      if (buf_gbit(buf))
         return HEAP_ERROR_BUF_HEADER;
      
      if (buf_size(buf) > TLSF_BUF_MAX)
         return HEAP_ERROR_BUF_HEADER;

      if (buf_size(buf) < 2 * sizeof(struct Buffer) + a->config->buf_size_endmark &&
          buf_size(buf) != 0)
         return HEAP_ERROR_BUF_HEADER;
   }

   return HEAP_ERROR_NONE;
}

/***************************************************************************/

static always_inline HeapError
buf_check_endmark(struct Heap *heap, struct TlsfAlgorithm *a,
                  struct Fragment *frag, struct Buffer *buf)
{
   /* Validates that the buffer endmark, written
    * at the end of the user data area, is still valid
    * when the buffer is freed. Damages to the endmark
    * indicates that the user has used more memory than
    * it originally allocated. The next buffer's header
    * may have been damaged as well. */
   
   char *endm = (char *)&buf[1] + buf_req_size(a, buf);
   int i = a->config->buf_size_endmark;

   (void)heap;

   /* There will be no endmark if the buffer is free. */
   if (unlikely(buf_fbit(buf)))
   {
      return HEAP_ERROR_NONE;
   }

   /* Endmark must be inside fragment. */
   if (unlikely(frag != NULL))
   {
      if (unlikely((uint32_t)endm <  (uint32_t)frag->root ||
                   (uint32_t)endm >= (uint32_t)frag->tail))
         return HEAP_ERROR_BUF_HEADER;
   }
   
   while (--i>=0)
   {
      if (unlikely(endm[i] != HEAP_ENDMARK))
         return HEAP_ERROR_BUF_ENDMARK;
   }

   return HEAP_ERROR_NONE;
}

/***************************************************************************/

static inline HeapError
buf_check_at_free(struct Heap *heap, struct TlsfAlgorithm *a,
                  struct Buffer *buf, HeapPID pid)
{
   /* A collection of all integrity and error checks
    * that are carried out during a free request. The
    * function is also used for resize request which
    * behaves similarly. */

   struct Fragment *frag;
   HeapError status;

   /* Frag check is optionally enabled for buffer deallocations. */
   if (unlikely(a->config->frag_check))
   {
      status = buf_check_frag(heap, a, &frag, &buf[1]);
      if (unlikely(status != HEAP_ERROR_NONE))
         return status;
   }
   else
   {
      /* Frag check is disabled, no fragment to validate against. */
      frag = NULL;
   }
      
   status = buf_check_header(heap, a, frag, buf);
   if (unlikely(status != HEAP_ERROR_NONE))
      return status;

   if (unlikely(buf_fbit(buf)))
      return HEAP_ERROR_BUF_FREE;

   status = buf_check_endmark(heap, a, frag, buf);
   if (unlikely(status != HEAP_ERROR_NONE))
      return status;

   status = buf_check_owner(heap, a, buf, pid);
   if (unlikely(status != HEAP_ERROR_NONE))
      return status;

   return HEAP_ERROR_NONE;
}

/***************************************************************************/

static inline HeapError
buf_check_at_query(struct Heap *heap, struct TlsfAlgorithm *a,
                   struct Fragment **frag_out, struct Buffer *buf)
{
   /* A collection of all integrity and error checks
    * that are carried out during queries. */

   struct Fragment *frag = NULL;
   HeapError status;

   /* Frag check is optionally enabled for queries, unless in freeze-mode. */
   if (unlikely(a->config->frag_check || heap->freezed))
   {
      status = buf_check_frag(heap, a, &frag, &buf[1]);
      if (unlikely(status != HEAP_ERROR_NONE))
      {
         *frag_out = NULL;
         return status;
      }
   }

   *frag_out = frag;

   status = buf_check_header(heap, a, frag, buf);
   if (unlikely(status != HEAP_ERROR_NONE))
      return status;

   return HEAP_ERROR_NONE;
}


/********************* FRAGMENT MANIPULATORS ******************************/

static always_inline HeapError
frag_check(struct Fragment *frag)
{
   /* Validate fragment integrity. */
   
   if (unlikely(frag == NULL))
      return HEAP_ERROR_NONE;
   
   if (unlikely(frag_checksum(frag) != frag_get_checksum(frag)))
      return HEAP_ERROR_CORRUPTED;

   return HEAP_ERROR_NONE;
}

/**************************************************************************/

static always_inline void 
frag_insert(struct Region *reg, struct Fragment *f)
{
   /* Registers a fragment descriptor with a region to allow
    * traversal of the fragments. Note that the frag_last field
    * of the region descriptor will become invalid as soon as the
    * region is registered with the heap. Once the region is
    * registered, no more fragments can be added.
    *
    * Fragments must be inserted in the order of increasing addresses,
    * so that frag_last points to the fragment with highest address
    * when the region is registered with the heap. 
    */
   
   f->prev = reg->frag_last;
   f->next = NULL;
   if (f->prev)
   {
      f->prev->next = f;
      frag_set_checksum(f->prev);
   }
   if (reg->frag_list == NULL)
      reg->frag_list = f;
   reg->frag_last = f;

   f->region = reg;
   frag_set_checksum(f);
}

/**************************************************************************/

static inline struct Fragment *
frag_first(struct Heap *heap, struct Region *reg, HeapError *ecode)
{
   /* Returns the first fragment in a region.
    * Fragment integrity is validated. */

   HeapError tmp_ecode = frag_check(reg->frag_list);

   if (ecode != NULL)
      *ecode = tmp_ecode;

   if (unlikely(tmp_ecode != HEAP_ERROR_NONE))
   {
      if (ecode == NULL)
         zzheap_error(heap, tmp_ecode, reg->frag_list);

      return NULL;
   }

   return reg->frag_list;
}

/**************************************************************************/

static inline struct Fragment *
frag_next(struct Heap *heap, struct Fragment *frag, HeapError *ecode)
{
   /* Returns the next fragment in a region.
    * Fragment integrity is validated. */

   HeapError tmp_ecode = frag_check(frag);

   if (ecode != NULL)
      *ecode = tmp_ecode;

   if (unlikely(tmp_ecode != HEAP_ERROR_NONE))
   {
      if (ecode == NULL)
         zzheap_error(heap, tmp_ecode, frag);
      
      return NULL;
   }

   tmp_ecode = frag_check(frag->next);

   if (ecode != NULL)
      *ecode = tmp_ecode;

   if (unlikely(tmp_ecode != HEAP_ERROR_NONE))
   {
      if (ecode == NULL)
         zzheap_error(heap, tmp_ecode, frag->next);

      return NULL;
   }


   return frag->next;
}

/**************************************************************************/

static inline struct Fragment *
frag_find(struct Heap *heap, struct Region *reg,
          void *ptr, uint32_t exact, HeapError *ecode)
{
   /* Tries to find the fragment descriptor that contains the
    * ptr. If exact is set, the ptr must be equal to the
    * pointer of the fragment descriptor. */

   struct Fragment *frag = frag_first(heap, reg, ecode);

   while (frag != NULL)
   {
      if ((uintptr_t)ptr >= (uintptr_t)frag->root  &&
          (uintptr_t)ptr <= (uintptr_t)frag->tail)
      {
         if (exact && frag != ptr)
            return NULL;
         return frag;
      }

      frag = frag_next(heap, frag, ecode);
   }

   return NULL;
}

/**************************************************************************/

static always_inline struct Fragment *
frag_from_buf(struct Buffer *buf)
{
   /* The function locates the fragment descriptor
    * that is associated with a buffer. The fragment
    * is only returned if the buffer is the root buffer
    * of the fragment. The fragment is not of interest
    * otherwise, since this function is called to determine
    * if the fragment can be deallocated.
    */

   struct Fragment *frag = (struct Fragment *)buf;

   frag = &frag[-1];

   if (likely(frag->root != buf))
      return NULL;

   return frag;
}

/********************** REGION MANIPULATORS *******************************/

static always_inline void
region_init(struct TlsfAlgorithm *a, struct Region *reg, uint8_t initial)
{
   region_set_control(reg, a->id);

   if (initial)
      region_set_initial(reg);
   
   reg->frag_list = NULL;
   reg->frag_last = NULL;
}

/**************************************************************************/

static always_inline HeapError
region_check(struct Region *region)
{
   /* Validate region integrity. */
   
   if (unlikely(region == NULL))
      return HEAP_ERROR_NONE;
   
   if (unlikely(region_checksum(region) != region_get_checksum(region)))
      return HEAP_ERROR_CORRUPTED;
   
   return HEAP_ERROR_NONE;
}

/**************************************************************************/

static inline void
region_insert(struct Heap *heap, struct TlsfAlgorithm *a, struct Region *reg)
{
   /* Registers a region descriptor with the heap. The region
    * is inserted into a list and all root buffers within fragments
    * of the region are inserted into the heap freelists.
    *
    * The frag_last field of the descriptor is changed to point at
    * the end of the last fragment (instead of the beginning).
    * The field is used by integrity check methods to determine
    * if pointers reside inside the region or not. */
   
   struct Fragment *f;
   HeapLock lock;

   heap_assert(reg->frag_list  != NULL &&
               reg->frag_last  != NULL,
               "Attempted to insert region with no fragments.");

   heap_lock(heap, &lock, 0);
   reg->prev = NULL;
   reg->next = a->region_list;
   
   if (reg->next)
   {
      HeapError ecode;
      ecode = region_check(reg->next);
      if (unlikely(ecode != HEAP_ERROR_NONE))
      {
         heap_unlock(heap, &lock, 0);

         zzheap_error(heap, HEAP_ERROR_FATAL|ecode, reg->next);
         return;
      }

      reg->next->prev = reg;
      region_set_checksum(reg->next);
   }

   a->region_list = reg;
   a->region_count++;

   /* Adjust frag_last to point to end of last fragment,
    * optimization to speed up searches. */ 
   reg->frag_last = (struct Fragment *)reg->frag_last->tail;

   region_set_checksum(reg);
   
   /* Add all buffers in the region to the heap. */
   f = reg->frag_list;
   while (f != NULL)
   {
      uint32_t l1, l2;

      index_search(buf_size(f->root), &l1, &l2);
      a->stats->buf_num_used[l1]++;
      buf_insert(heap, a, f->root, l1, l2);

      a->frag_count++;
      
      f = f->next;
   }

   heap_unlock(heap, &lock, 0);
}

/**************************************************************************/

static inline void
region_remove(struct Heap *heap, struct TlsfAlgorithm *a, struct Region *reg)
{
   /* Removes/unregisters a region from the heap. All
    * root buffers in the region is removed from freelists.
    * It is assumed that the caller has verified that the
    * region can be removed, i.e. that no buffer is allocated.
    */
   
   struct Fragment *f;
   HeapError ecode = HEAP_ERROR_NONE;
   
   /* Unlink region descriptor. */
   if (reg->prev)
   {
      ecode = region_check(reg->prev);
      if (unlikely(ecode != HEAP_ERROR_NONE))
      {
         zzheap_error(heap, HEAP_ERROR_FATAL|ecode, reg->prev);
         return;
      }

      reg->prev->next = reg->next;
      region_set_checksum(reg->prev);
   }

   if (reg->next)
   {
      ecode = region_check(reg->next);
      if (unlikely(ecode != HEAP_ERROR_NONE))
      {
         zzheap_error(heap, HEAP_ERROR_FATAL|ecode, reg->next);
         return;
      }

      reg->next->prev = reg->prev;
      region_set_checksum(reg->next);
   }

   if (a->region_list == reg)
      a->region_list = reg->next;

   reg->next = reg->prev = NULL;

   a->region_count--;

   /* Remove all buffers from the heap. */
   f = frag_first(heap, reg, &ecode);
   while (f != NULL)
   {
      uint32_t l1, l2;
      
      index_search(buf_size(f->root), &l1, &l2);
      a->stats->buf_num_used[l1]--;
      buf_remove(heap, a, f->root, l1, l2);

      a->frag_count--;
      
      f = frag_next(heap, f, &ecode);
   }

   if (unlikely(ecode != HEAP_ERROR_NONE))
      zzheap_error(heap, HEAP_ERROR_FATAL|ecode, f);
}

/**************************************************************************/

static always_inline struct Region *
region_from_buf(struct Buffer *buf)
{
   /* Locates the region descriptor that is associated
    * with a buffer. The region descriptor can be used
    * to determine if the entire region is unused and
    * can be returned to the allocator. */

   struct Fragment *f = frag_from_buf(buf);

   /* FIXME: Fragment integrity. */

   return f != NULL ? f->region : NULL;
}

/**************************************************************************/

static inline int
region_is_free(struct Heap *heap, struct Region *reg)
{
   /* Traverses all fragments in a region to determine
    * if all fragments are unused. If so, the region can
    * be deallocated using zzheap_unbreak. */
   
   struct Fragment *f;
   HeapError ecode = HEAP_ERROR_NONE;

   if (region_initial(reg))
      return 0;
   
   f = frag_first(heap, reg, &ecode);
   while (f != NULL)
   {
      if (buf_size(f->root) != frag_size(f))
         return 0;

      f = frag_next(heap, f, &ecode);
   }

   if (unlikely(ecode != HEAP_ERROR_NONE))
   {
      zzheap_error(heap, ecode, reg);
      return 0;
   }
   
   return 1;
}

/**************************************************************************/

static inline struct Region *
region_first(struct Heap *heap, struct TlsfAlgorithm *a, HeapError *ecode)
{
   HeapError tmp_ecode = region_check(a->region_list);

   if (ecode != NULL)
      *ecode = tmp_ecode;

   if (unlikely(tmp_ecode != HEAP_ERROR_NONE))
   {
      if (ecode == NULL)
         zzheap_error(heap, tmp_ecode, a->region_list);

      return NULL;
   }

   return a->region_list;
}

/**************************************************************************/

static inline struct Region *
region_next(struct Heap *heap, struct Region *reg, HeapError *ecode)
{
   HeapError tmp_ecode = region_check(reg->next);

   if (ecode != NULL)
      *ecode = tmp_ecode;

   if (unlikely(tmp_ecode != HEAP_ERROR_NONE))
   {
      if (ecode == NULL)
         zzheap_error(heap, tmp_ecode, reg->next);

      return NULL;
   }

   return reg->next;
}

/**************************************************************************/

static inline struct Region *
region_find(struct Heap *heap, struct TlsfAlgorithm *a,
            void *ptr, uint32_t exact, HeapError *ecode)
{
   /* Tries to find the region descriptor that contains the
    * ptr. If exact is set, the ptr must be equal to the
    * pointer of the region descriptor. */

   struct Region *reg = region_first(heap, a, ecode);

   while (reg != NULL)
   {
      if ((uintptr_t)ptr >= (uintptr_t)reg &&
          (uintptr_t)ptr <  (uintptr_t)reg->frag_last)
      {
         if (exact && ptr != reg)
            return NULL;
         return reg;
      }
      
      reg = region_next(heap, reg, ecode);
   }
      
   return NULL;
}

/**************************************************************************/

static inline void
region_reclaim(struct Heap *heap, struct TlsfAlgorithm *a,
               union HeapEvent *event, struct Region *region)
{
   HeapError ecode;
   HeapLock lock;

   /* Interrupt and timer interrupt processes
      are not allowed to return regions. */
   if (!heap_brkallowed()) {
      return;
   }

   /* Grab the break lock forcibly to prevent new extension attempts. */
   heap_brklock(heap, &lock);

   /* Grab general heap lock to protect freelists, etc. */
   heap_lock(heap, &lock, 0);

   /* Check if the region still exists and if it is free. */
   if (region != NULL
       && region_find(heap, a, region, 1, &ecode)
       && !region_locked(region)
       && region_is_free(heap, region))
   {
      /* The region can be safely deleted.*/

      /* Remove region and buffers from freelists.. */
      region_remove(heap, a, region);

      /* Unlock freelists. */
      heap_unlock(heap, &lock, 0);

      /* Delete region. */
      zzheap_unbreak(heap, region, event ? event->e_free.params : NULL);
   }
   else
   {
      heap_unlock(heap, &lock, 0);
   }

   heap_brkunlock(heap, &lock);
}

/**************************************************************************/

static inline int
l_frag_break(struct Heap *heap, struct TlsfAlgorithm *a,
             HeapParam *params, size_t size, int init)
{
   struct Region *reg;
   struct Fragment *frag;

   /* Size of administrative overhead per fragment. */
   size_t shead =
      sizeof(struct Fragment) +
      sizeof(struct Buffer) * 3; /* Sentinel. Root already included. */

   size_t smax = 0;     /* No limit on maximum fragment size. */
   size_t sreq;         /* Preferred fragment size.           */
   size_t smin;         /* Minimum/required fragment size.    */

   /* Add overhead and alignment. */
   if (!init)
   {
      size = size + shead + sizeof(struct Region);
      size = size_align_tlsf(size);
   }

   smin = size;
   sreq = size;

   /* Align to minimum extension size. */
   if (!init && sreq < a->config->frag_size_default)
      sreq = a->config->frag_size_default;

   reg = zzheap_break(heap, &smin, sreq, &smax, a->id, params);
   if (reg == NULL)
   {
      /* Break failed. Possible cause is lack of system memory
       * or that the currently allocated amount of memory has
       * or will reach its configured limit.
       */
      return 0;
   }
   else
   {
      /* Allocation was successful. Update size with
       * the size of the fragment that was acquired. */
      size = smin;
   }


   frag = (struct Fragment *)&reg[1];
   size = size - sizeof(struct Region);

   region_init(a, reg, init);
   
   while (size >= shead + TLSF_BUF_MIN)
   {
      size -= shead;
      
      frag->control = a->id;
      frag_set_size(frag, size_align_down(size));

      /* Initialize buffer. */
      frag->root = (struct Buffer *)&frag[1];
      frag->root->control = a->id;
      buf_set_size(frag->root, frag_size(frag));
      buf_set_fbit(frag->root);
      buf_clear_gbit(frag->root);

      /* Initialize tail. */
      frag->tail = buf_right(heap, a, frag->root, 0);
      frag->tail->control = a->id;
      buf_set_size(frag->tail, 0);
      buf_set_pbit(frag->tail);
      buf_clear_fbit(frag->tail);
      buf_set_ucheck(frag->tail);
   
      frag_insert(reg, frag);

      size = size - frag_size(frag);
      frag = (struct Fragment *)&frag->tail[3];
   }
   
   /* Insert region and all buffers. */
   region_insert(heap, a, reg);
   
   return 1;
}

/***************************************************************************/



   
/***************************************************************************
 * Private Interface
 **************************************************************************/

static HeapAction
l_tlsf_alloc(struct Heap *heap, union HeapEvent *event, void *opaque)
{
   struct TlsfAlgorithm *a = opaque;
   struct Buffer *buf;

   size_t size = event->e_alloc.size;
   size_t act_size;
   
   size += 2 * sizeof(struct Buffer) + a->config->buf_size_endmark;

   if (unlikely(size >= a->config->buf_size_threshold))
      return HEAP_ACTION_CONTINUE;

   do
   {
      HeapLock lock;

      uint32_t l1, l2;

      /* Calculate requested indexes. */
      index_search_align(size, &l1, &l2);

      heap_lock(heap, &lock, 0);

      if (unlikely(size_align(size) + a->stats->buf_size_used > a->config->buf_size_max))
      {
         heap_unlock(heap, &lock, 0);
         return HEAP_ACTION_CONTINUE;
      }
      
      /* Find a large enough buffer. */
      buf = buf_find(a, &l1, &l2);
      if (unlikely(buf == NULL))
      {
         heap_brkprepare(heap);
         heap_unlock(heap, &lock, 0);

         if (!heap_brkallowed())
         {
            return HEAP_ACTION_CONTINUE;
         }

         if (!heap_brklock(heap, &lock))
         {
            heap_brkunlock(heap, &lock);
            continue;
         }
         
         if (unlikely(!l_frag_break(heap, a, event->e_alloc.params, size, 0)))
         {
            heap_brkunlock(heap, &lock);
            return HEAP_ACTION_CONTINUE;
         }

         heap_brkunlock(heap, &lock);
         continue;
      }

      /* Allocate buffer. */
      buf_remove(heap, a, buf, l1, l2);

      /* Split buffer if possible. */
      buf_split_right(heap, a, buf, size);
      buf_set_owner(buf, event->e_alloc.owner);
      buf_set_size(buf, size);
      buf_set_endmark(a, buf);
      buf_set_ucheck(buf);

      /* Update statistics. */
      a->stats->buf_size_used += (act_size = buf_act_size(a, buf));
      a->stats->buf_size_req  += buf_req_size(a, buf);
      if (a->stats->buf_size_used > a->stats->buf_size_used_hi)
         a->stats->buf_size_used_hi = a->stats->buf_size_used;

      if (event->e_alloc.owner)
         a->stats->buf_num_private++;
      else
         a->stats->buf_num_shared++;

      heap_unlock(heap, &lock, 0);
   } while (unlikely(buf == NULL));

   event->e_alloc.ptr  = &buf[1];
   event->e_alloc.size = act_size;
   
   return HEAP_ACTION_INTERCEPT;
}

/***************************************************************************/

static HeapAction
l_tlsf_free(struct Heap *heap, union HeapEvent *event, void *opaque)
{
   struct TlsfAlgorithm *a = opaque;
   struct Buffer *buf = event->e_free.ptr, *left;
   HeapError status;
   HeapLock lock;
   size_t act_size;
   uint32_t l1, l2;
   
   if (unlikely(buf == NULL))
      return HEAP_ACTION_CONTINUE;

   buf = &buf[-1];

   /* Dummy access to catch bad pointers before locking and
    * potentially entering into kill locked state. */
   (void)*(volatile uint32_t *)buf;
   
   heap_lock(heap, &lock, 0);

   status = buf_check_at_free(heap, a, buf, event->e_free.owner);
   if (unlikely(status == HEAP_ERROR_BUF_UNKNOWN))
   {
      heap_unlock(heap, &lock, 0);
      return HEAP_ACTION_CONTINUE;
   }

   if (unlikely(status != HEAP_ERROR_NONE))
   {
      heap_unlock(heap, &lock, 0);
      zzheap_error(heap, status, &buf[1]);
      return HEAP_ACTION_INTERCEPT;
   }

   if (buf_gc(buf))
   {
      /* The buffer is used by the garbage collector and
       * can't be freed. We simply clear the bit to indicate
       * that we attempted to free it but couldn't. The collector
       * will notice and free the buffer for us. */
      buf_clear_gc(buf);
      heap_unlock(heap, &lock, 0);

      return HEAP_ACTION_INTERCEPT;
   }

   /* Update statistics. */
   a->stats->buf_size_used -= (act_size = buf_act_size(a, buf));
   a->stats->buf_size_req  -= buf_req_size(a, buf);

   if (buf_owner(buf))
      a->stats->buf_num_private -= 1;
   else
      a->stats->buf_num_shared  -= 1;

   buf_clear_anno(buf);
   buf_set_size(buf, buf_act_size(a, buf));

   #ifdef OSARM
   buf_set_fbit(buf);  /* Set here to detect double free even if merged left */
   #endif

   buf_set_ucheck(buf);

   left = buf_merge_left(heap, a, buf, 0);
   if (left != NULL)
      buf = left;

   buf_merge_right(heap, a, buf, 0);

   index_search(buf_size(buf), &l1, &l2);
   buf_insert(heap, a, buf, l1, l2);

   /* If this is the root buffer, try to unbreak entire region. */
   if (unlikely(a->config->frag_reclaim))
   {
      struct Region *region = region_from_buf(buf);

      /* Deleting the region requires the heap break lock
       * which can't be locked if the regular heap lock
       * is already locked. The regular lock is therefore
       * unlocked and reacquired in reclaim_region() after
       * grabbing the break lock. */
      heap_unlock(heap, &lock, 0);

      if (region != NULL) {
         region_reclaim(heap, a, event, region);
      }
   } else {
      heap_unlock(heap, &lock, 0);
   }

   event->e_free.size = act_size;

   return HEAP_ACTION_INTERCEPT;
}

/***************************************************************************/

static HeapAction
l_tlsf_resize(struct Heap *heap, union HeapEvent *event, void *opaque)
{
   struct TlsfAlgorithm *a = opaque;
   struct Buffer *buf = event->e_resize.ptr;
   HeapError status;
   HeapLock lock;

   size_t req_size = event->e_resize.size +
      2 * sizeof(struct Buffer) + a->config->buf_size_endmark;
   size_t old_req, old_act, new_act;
   HeapPID old_owner;

   if (unlikely(buf == NULL))
      return HEAP_ACTION_CONTINUE;

   buf = &buf[-1];

   /* Dummy access to catch bad pointers before locking and
    * potentially entering into kill locked state. */
   (void)*(volatile uint32_t *)buf;

   heap_lock(heap, &lock, 0);
   
   status = buf_check_at_free(heap, a, buf, event->e_resize.owner);
   if (unlikely(status == HEAP_ERROR_BUF_UNKNOWN))
   {
      heap_unlock(heap, &lock, 0);
      return HEAP_ACTION_CONTINUE;
   }

   if (unlikely(status != HEAP_ERROR_NONE))
   {
      heap_unlock(heap, &lock, 0);

      zzheap_error(heap, status, &buf[1]);
      event->e_resize.ptr = NULL;
      return HEAP_ACTION_INTERCEPT;
   }

   if (unlikely(req_size > a->config->buf_size_threshold))
   {
      heap_unlock(heap, &lock, 0);

      /* The buffer belongs to this algorithm, but the new size cannot
       * be accomodated. Intercept request and force framework to copy. */
      event->e_resize.ptr = NULL;
      return HEAP_ACTION_INTERCEPT;
   }

   old_req = buf_req_size(a, buf);
   old_act = buf_act_size(a, buf);
   old_owner = buf_owner(buf);

   /* Larger size, need to merge. */
   if (req_size > buf_act_size(a, buf))
   {
      struct Buffer *right;

      right = buf_merge_right(heap, a, buf, req_size);
      if (right == NULL)
      {
         heap_unlock(heap, &lock, 0);

         /* Could not merge. */
         event->e_resize.ptr = NULL;
         return HEAP_ACTION_INTERCEPT;
      }
   }

   /* Smaller size, need to split. */
   if (req_size < buf_act_size(a, buf))
   {
      buf_split_right(heap, a, buf, req_size);
   }
   
   buf_set_size(buf, req_size);
   buf_set_endmark(a, buf);
   buf_set_owner(buf, old_owner);
   buf_set_ucheck(buf);

   /* Update statistics. */
   a->stats->buf_size_req  -= old_req;
   a->stats->buf_size_used -= old_act; 
   a->stats->buf_size_req  += buf_req_size(a, buf);
   a->stats->buf_size_used += (new_act = buf_act_size(a, buf));

   heap_unlock(heap, &lock, 0);
   
   event->e_resize.ptr  = &buf[1];
   event->e_resize.size = new_act;
   event->e_resize.size_prev = old_act;
   
   return HEAP_ACTION_INTERCEPT;
}

/***************************************************************************/

static void
buf_collect(struct TlsfAlgorithm *a, struct Buffer **head, struct Buffer *buf)
{
   /* This function links the buffer into
    * a list of all buffers that will be
    * freed by the garbage collector at a
    * later point in time.
    *
    * The buffer is also resized to make
    * room for a pointer in the user data
    * area. The pointer will point to the
    * next buffer in the list of collected
    * buffers. The endmark is moved forward. */
   
   const size_t oh =
      2 * sizeof(struct Buffer) +
      sizeof(struct Buffer*) + 
      a->config->buf_size_endmark;

   size_t size = buf_size(buf);

   if (size < oh)
   {
      a->stats->buf_size_req += oh - size;

      buf_set_size(buf, oh);
      buf_set_endmark(a, buf);
   }

   buf[1].control = (uint32_t)*head;
   *head = buf;
}

/**************************************************************************/

static HeapAction
l_tlsf_collect(struct Heap *heap, union HeapEvent *event, void *opaque)
{
   struct TlsfAlgorithm *a = opaque;
   HeapPID pid = event->e_collect.owner;
   HeapLock lock;

   struct Region *reg;
   struct Fragment *frag;
   struct Buffer *buf;

   heap_lock(heap, &lock, 0);

   reg = region_first(heap, a, NULL);
   while (reg != NULL)
   {
      /* Prevent reclaim while traversing this region.
       * The minimize interrupt latency, the region
       * is locked which means that it cannot be unbroken
       * if it happens to become unused during traversal.
       * Allows us to traverse it without locks taken. */
      region_set_lock(reg);
      heap_unlock(heap, &lock, 0);

      frag = frag_first(heap, reg, NULL);
      while (frag != NULL)
      {
         struct Buffer *head = NULL;
         uint32_t i;

         /* Mark the first buffer as a hostage of the collector.
          * Attempts to free the buffer will be detected by
          * the collector once heap locks are reacquired. */
         heap_lock(heap, &lock, 0);
         buf = buf_first(heap, a, frag);

         /* Collect and gather all leaked buffers in a list. */
         do {

            /* Work through a hundred buffers at a
             * time before releasing the heap lock. */
            for (i=0; i<100 && buf != NULL && buf_size(buf) != 0; i++)
            {
               if (!buf_fbit(buf) && buf_owner(buf) == pid)
               {
                  buf_collect(a, &head, buf);
               }
               
               buf = buf_right(heap, a, buf, 1);
            }

            /* Scan ahead to the next used buffer.
             * The lock can only be released if.....*/
            while (buf != NULL && buf_fbit(buf))
            {
               buf = buf_right(heap, a, buf, 1);
            }

            if (buf == NULL || buf_size(buf) == 0)
            {
               buf = NULL;
               break;
            }

            /* Mark the buffer as a hostage of the collector.
             * Attempts to free the buffer will be detected by
             * the collector once heap locks are reacquired. */
            buf_set_gc(buf);
            
            heap_unlock(heap, &lock, 0);
            heap_lock(heap, &lock, 0);

            if (!buf_gc(buf))
            {
               buf_collect(a, &head, buf);
               buf = buf_right(heap, a, buf, 1);
            }
            else
            {
               buf_clear_gc(buf);                  
            }
            
         } while (buf != NULL && buf_size(buf) != 0);

         heap_unlock(heap, &lock, 0);

         /* Now free all buffers in the list. */
         while (head != NULL)
         {
            buf = head;
            head = (struct Buffer *)buf[1].control;
            
            zzheap_free(heap, &buf[1], 0, NULL, NULL, 0);
         }

         frag = frag_next(heap, frag, NULL);
      }

      /* Advance to next region. */
      heap_lock(heap, &lock, 0);
      region_clear_lock(reg);

      /* Check if the region became free. Unbreak it if needed. */
      if (unlikely(a->config->frag_reclaim))
      {
         struct Region *tmp = region_next(heap, reg, NULL);
         if (tmp != NULL)
            region_set_lock(tmp);

         /* Deleting the region requires the heap break lock
          * which can't be locked if the regular heap lock
          * is already locked. The regular lock is therefore
          * unlocked and reacquired in reclaim_region() after
          * grabbing the break lock. */
         heap_unlock(heap, &lock, 0);

         region_reclaim(heap, a, NULL, reg);
         
         heap_lock(heap, &lock, 0);
         reg = tmp;
      }
      else
      {
         reg = region_next(heap, reg, NULL);
      }
   }

   heap_unlock(heap, &lock, 0);

   return HEAP_ACTION_CONTINUE;
}

/***************************************************************************/

static HeapAction
l_tlsf_query(struct Heap *heap, union HeapEvent *event, void *opaque)
{
   HeapParam *p = event->e_query.params;
   struct TlsfAlgorithm *a = opaque;
   HeapLock lock;

   if (p == NULL)
      return HEAP_ACTION_CONTINUE;
   
   if (event->e_query.freezed)
      heap->freezed++;

   while (*p != HPM_END)
   {
      switch (HPM_ID(*p))
      {
         case HPM_QUERY_SERVICE:
         {
            if (!HPM_IS_HANDLED(p[0]) && p[1] == 0)
            {
               HPM_SET_HANDLED(p[0]);
               HPM_SET_VALUE(p[1], &a->info);
            }
            else if (p[1] == (HeapParam)&a->info)
            {
               HPM_CLEAR_HANDLED(p[0]);
               HPM_SET_VALUE(p[1], 0);
            }

            break;
         }
         case HPM_QUERY_BUF_ACT_SIZE:
         {
            struct Fragment *frag;
            struct Buffer *buf = &((struct Buffer *)p[1])[-1];

            if (!HPM_IS_HANDLED(p[0]))
            {
               heap_lock(heap, &lock, 0);
               
               if (buf_check_at_query(heap, a, &frag, buf) != HEAP_ERROR_NONE)
               {
                  heap_unlock(heap, &lock, 0);
                  break;
               }

               HPM_SET_HANDLED(p[0]);
               HPM_SET_VALUE(p[1], buf_act_size(a, buf));

               heap_unlock(heap, &lock, 0);
            }
            break;
         }
         case HPM_QUERY_BUF_REQ_SIZE:
         {
            struct Fragment *frag;
            struct Buffer *buf = &((struct Buffer *)p[1])[-1];

            if (!HPM_IS_HANDLED(p[0]))
            {
               heap_lock(heap, &lock, 0);

               if (buf_check_at_query(heap, a, &frag, buf) != HEAP_ERROR_NONE)
               {
                  heap_unlock(heap, &lock, 0);
                  break;
               }

               HPM_SET_HANDLED(p[0]);
               if (!buf_fbit(buf))
                  HPM_SET_VALUE(p[1], buf_req_size(a, buf));
               else
                  HPM_SET_VALUE(p[1], 0);

               heap_unlock(heap, &lock, 0);
            }
            break;
         }
         case HPM_QUERY_BUF_OWNER:
         {
            struct Fragment *frag;
            struct Buffer *buf = &((struct Buffer *)p[1])[-1];

            if (!HPM_IS_HANDLED(p[0]))
            {
               heap_lock(heap, &lock, 0);
               
               if (buf_check_at_query(heap, a, &frag, buf) != HEAP_ERROR_NONE)
               {
                  heap_unlock(heap, &lock, 0);
                  break;
               }

               HPM_SET_HANDLED(p[0]);
               if (!buf_fbit(buf))
                  HPM_SET_VALUE(p[1], buf_owner(buf));
               else
                  HPM_SET_VALUE(p[1], 0);
              
               heap_unlock(heap, &lock, 0);
            }
            break;
         }
         case HPM_QUERY_BUF_TOT_REQ_SIZE:
         { 
            if (!HPM_IS_HANDLED(p[0]))
            {
               HPM_SET_HANDLED(p[0]);
               HPM_SET_VALUE(p[1], a->stats->buf_size_req);
            }
            break;
         }
         case HPM_QUERY_BUF_TOT_USED_SIZE:
         { 
            if (!HPM_IS_HANDLED(p[0]))
            {
               HPM_SET_HANDLED(p[0]);
               HPM_SET_VALUE(p[1], a->stats->buf_size_used);
            }
            break;
         }
         case HPM_QUERY_BUF_MAX_USED_SIZE:
         { 
            if (!HPM_IS_HANDLED(p[0]))
            {
               HPM_SET_HANDLED(p[0]);
               HPM_SET_VALUE(p[1], a->stats->buf_size_used_hi);
            }
            break;
         }
         case HPM_QUERY_BUF_TOT_FREE_SIZE:
         {
            if (!HPM_IS_HANDLED(p[0]))
            {
               HPM_SET_HANDLED(p[0]);
               HPM_SET_VALUE(p[1], a->stats->buf_size_free);
            }
            break;
         }
         case HPM_QUERY_BUF_NUM_SIZE:
         {
            size_t s = (1 << (p[1] + TLSF_L1_MIN));
            
            if (p[1] < TLSF_L1_COUNT && s <= TLSF_BUF_MAX)
            {
               HPM_SET_HANDLED(p[0]);
               HPM_SET_VALUE(p[2], s);
            }
            break;
         }
         case HPM_QUERY_BUF_NUM_FREE:
         {
            if (p[1] < TLSF_L1_COUNT)
            {
               HPM_SET_HANDLED(p[0]);
               HPM_ADD_VALUE(p[2], a->stats->buf_num_free[p[1]]);
            }
            break;
         }
         case HPM_QUERY_BUF_NUM_USED:
         {
            if (p[1] < TLSF_L1_COUNT)
            {
               HPM_SET_HANDLED(p[0]);
               HPM_ADD_VALUE(p[2], a->stats->buf_num_used[p[1]]);
            }
            break;
         }
         case HPM_QUERY_BUF_FRAG_NEXT:
         {
            if (!HPM_IS_HANDLED(p[0]))
            {
               struct BufferCursor *cursor = (struct BufferCursor *)&p[1];
               struct Buffer *buf;

               heap_lock(heap, &lock, 0);

               buf = buf_next(heap, a, cursor, 0);
               if (buf == NULL)
               {
                  heap_unlock(heap, &lock, 0);
                  break;
               }

               HPM_SET_HANDLED(p[0]);
               HPM_SET_VALUE(p[1], &buf[1]);

               heap_unlock(heap, &lock, 0);
            }
            break;
         }
         case HPM_QUERY_BUF_NEXT:
         {
            /* Skip if the request doesn't concern us. */
            if (p[2] != 0 && FRAG_NEXT_ID(p[2]) != a->id)
               break;

            if (!HPM_IS_HANDLED(p[0]))
            {
               struct BufferCursor *cursor = (struct BufferCursor *)&p[1];
               struct Buffer *buf;

               heap_lock(heap, &lock, 0);

               buf = buf_next(heap, a, cursor, 1);
               if (buf == NULL)
               {
                  /* Initialize query for next algorithm. */
                  cursor->buf       = NULL;
                  cursor->frag      = NULL;
                  cursor->reserved1 = 0;
                  heap_unlock(heap, &lock, 0);
                  break;
               }

               HPM_SET_HANDLED(p[0]);
               HPM_SET_VALUE(p[1], &buf[1]);

               heap_unlock(heap, &lock, 0);
            }
            break;
         }
         case HPM_QUERY_BUF_DATA:
         {
            struct Fragment *frag;
            struct Buffer *buf = &((struct Buffer *)p[1])[-1];

            if (!HPM_IS_HANDLED(p[0]))
            {
               void  *data = (void *)p[2];
               size_t size = (size_t)p[3];

               heap_lock(heap, &lock, 0);

               if (buf_check_at_query(heap, a, &frag, buf) != HEAP_ERROR_NONE)
               {
                  heap_unlock(heap, &lock, 0);
                  break;
               }

               HPM_SET_HANDLED(p[0]);
               memcpy(data, &buf[1], size);

               heap_unlock(heap, &lock, 0);
            }
            break;
         }
         case HPM_QUERY_BUF_TYPE:
         {
            struct Fragment *frag;
            struct Buffer *buf = &((struct Buffer *)p[1])[-1];

            if (!HPM_IS_HANDLED(p[0]))
            {
               heap_lock(heap, &lock, 0);

               if (buf_check_at_query(heap, a, &frag, buf) != HEAP_ERROR_NONE)
               {
                  heap_unlock(heap, &lock, 0);
                  break;
               }

               HPM_SET_HANDLED(p[0]);
               HPM_SET_VALUE(p[1], HEAP_BUFFER_TLSF);

               heap_unlock(heap, &lock, 0);
            }
            break;
         }
         case HPM_QUERY_BUF_STATUS:
         {
            struct Buffer *buf = &(((struct Buffer*)p[1])[-1]);
            struct Fragment *frag;
            HeapError ecode;

            if (!HPM_IS_HANDLED(p[0]))
            {
               heap_lock(heap, &lock, 0);

               ecode = buf_check_at_query(heap, a, &frag, buf);
               if (ecode == HEAP_ERROR_BUF_UNKNOWN)
               {
                  heap_unlock(heap, &lock, 0);
                  break;
               }

               HPM_SET_HANDLED(p[0]);
               HPM_SET_VALUE(p[1], buf_status(heap, a, frag, buf));

               heap_unlock(heap, &lock, 0);
            }
            break;
         }
         case HPM_QUERY_BUF_LARGEST_FREE:
         {
            size_t s = 0;
            size_t bitmap = a->l1_bitmap;
            int l1 = 0, l2 = 0;

            heap_lock(heap, &lock, 0);

            if (bitmap)
            {
               while (bitmap <<= 1)
                  l1++;

               bitmap = a->l2_bitmap[l1];
               if (bitmap)
               {
                  while (bitmap <<= 1)
                     l2++;

                  /* FIXME: May not be the largest buffer. Would have to
                   * check the entire free list to find the largest one
                   * with this l1/l2 index. Perhaps just respond with
                   * the natural TLSF size corresponding to the indexes.
                   */
                  s = buf_act_size(a, &a->list[l1][l2]->buf);
               }
            }

            heap_unlock(heap, &lock, 0);

            if (HPM_IS_HANDLED(p[0]))
            {
               if (s >= (size_t)p[1])
               {
                  HPM_SET_VALUE(p[1], s);
               }
            }
            else
            {
               HPM_SET_HANDLED(p[0]);
               HPM_SET_VALUE(p[1], s);
            }
            break;
         }
         case HPM_QUERY_BUF_NEXT_SIZE:
         {
            if (!HPM_IS_HANDLED(p[0]))
            {
               size_t s = (size_t)p[1] +
                  (2 * sizeof(struct Buffer) + a->config->buf_size_endmark);
               
               if (s >= a->config->buf_size_threshold || s >= TLSF_BUF_MAX)
                  break;

               s = size_align(s);
               s-= (2 * sizeof(struct Buffer) + a->config->buf_size_endmark);

               HPM_SET_HANDLED(p[0]);
               HPM_SET_VALUE(p[1], s);
            }
            break;
         }
         case HPM_QUERY_BUF_NUM_SHARED:
         {
            HPM_SET_HANDLED(p[0]);
            HPM_SET_VALUE(p[1], a->stats->buf_num_shared);
            break;
         }
         case HPM_QUERY_BUF_NUM_PRIVATE:
         {
            HPM_SET_HANDLED(p[0]);
            HPM_SET_VALUE(p[1], a->stats->buf_num_private);
            break;
         }
         default:
         {
            break;
         }
      }

      p += (1 + HPM_VALUES(*p));
   }

   if (event->e_query.freezed)
      heap->freezed--;

   return HEAP_ACTION_CONTINUE;
}

/***************************************************************************/

static HeapAction
l_tlsf_control(struct Heap *heap, union HeapEvent *event, void *opaque)
{
   struct TlsfAlgorithm *a = opaque;
   HeapParam *p = event->e_query.params;
   HeapLock lock;

   if (p == NULL)
      return HEAP_ACTION_CONTINUE;
   
   while (HPM_ID(*p) != HPM_END)
   {
      switch (HPM_ID(*p))
      {
         case HPM_CTRL_BUF_THRESHOLD:
         {
            if (!HPM_IS_HANDLED(p[0]))
            {
               size_t s = (size_t)p[1];

               if (s > TLSF_BUF_MAX)
                  s = TLSF_BUF_MAX;
               
               HPM_SET_HANDLED(p[0]);
               a->config->buf_size_threshold = s;
            }
            break;
         }
         case HPM_CTRL_BUF_MAX_SIZE:
         {
            if (!HPM_IS_HANDLED(p[0]))
            {
               HPM_SET_HANDLED(p[0]);
               a->config->buf_size_max = (size_t)p[1];
            }
            break;
         }
         case HPM_CTRL_BUF_MAX_USED_SIZE:
         {
            /* Reset used watermark statistics. */
            
            if (!HPM_IS_HANDLED(p[0]))
            {
               HPM_SET_HANDLED(p[0]);
               a->stats->buf_size_used_hi = 0;
            }
            break;
         }
         case HPM_CTRL_BUF_OWNER:
         {
            struct Buffer *buf = &(((struct Buffer*)p[1])[-1]);
            struct Fragment *frag;
            HeapError ecode;

            if (!HPM_IS_HANDLED(p[0]))
            {
               heap_lock(heap, &lock, 0);
               
               ecode = buf_check_at_query(heap, a, &frag, buf);
               if (ecode != HEAP_ERROR_NONE)
               {
                  heap_unlock(heap, &lock, 0);
                  break;
               }

               buf_set_owner(buf, (HeapPID)p[2]);
               HPM_SET_HANDLED(p[0]);

               heap_unlock(heap, &lock, 0);
            }
            break;
         }
         default:
            break;
      }

      p += (1 + HPM_VALUES(*p));
   }
   
   return HEAP_ACTION_CONTINUE;
}

/***************************************************************************/

static HeapAction
l_tlsf_destroy(struct Heap *heap, union HeapEvent *event, void *opaque)
{
   struct TlsfAlgorithm *a = opaque;

   (void)event;
   
   zzheap_detach_hook(heap, HEAP_EVENT_ALLOC_REQ,   l_tlsf_alloc,   a);
   zzheap_detach_hook(heap, HEAP_EVENT_FREE_REQ,    l_tlsf_free,    a);
   zzheap_detach_hook(heap, HEAP_EVENT_RESIZE_REQ,  l_tlsf_resize,  a);
   zzheap_detach_hook(heap, HEAP_EVENT_COLLECT_REQ, l_tlsf_collect, a);
   zzheap_detach_hook(heap, HEAP_EVENT_QUERY_REQ,   l_tlsf_query,   a);
   zzheap_detach_hook(heap, HEAP_EVENT_CONTROL_REQ, l_tlsf_control, a);

   if (a != NULL)
      heap_meta_free(heap, a);
   
   return HEAP_ACTION_CONTINUE;
}

/***************************************************************************/




/***************************************************************************
 * Public Interface
 ***************************************************************************/

HeapStatus
heap_attach_tlsf_algorithm(struct Heap *heap,
                           struct HeapStatistics *stats,
                           struct HeapConfiguration *config,
                           uint32_t id)
{
   HeapStatus status;

   struct TlsfAlgorithm *a = heap_meta_alloc(
      heap, sizeof(struct TlsfAlgorithm));
   if (a == NULL)
      return HEAP_EOUT_OF_MEMORY;

   
   bzero(a, sizeof(struct TlsfAlgorithm));

   a->id           = id << BUF_ID_SHIFT;
   a->stats        = stats;
   a->config       = config;

   a->info.version = 101;
   strcpy(a->info.id, "com.enea.heap.tlsf");

   if (a->config->buf_size_threshold > TLSF_BUF_MAX)
      a->config->buf_size_threshold = TLSF_BUF_MAX;
   
   if (config->frag_size_init && 
       l_frag_break(heap, a, NULL, config->frag_size_init, 1) == 0)
   {
      heap_meta_free(heap, a);
      return HEAP_EOUT_OF_MEMORY;
   }

   status = zzheap_attach_hook(
      heap, HEAP_EVENT_ALLOC_REQ, l_tlsf_alloc, a, HEAP_PRIO_ALGO_NORMAL);
   if (status != HEAP_SUCCESS)
   {
      heap_meta_free(heap, a);
      return status;
   }

   status = zzheap_attach_hook(
      heap, HEAP_EVENT_FREE_REQ, l_tlsf_free, a, HEAP_PRIO_ALGO_NORMAL);
   if (status != HEAP_SUCCESS)
   {
      heap_meta_free(heap, a);
      return status;
   }

   status = zzheap_attach_hook(
      heap, HEAP_EVENT_RESIZE_REQ, l_tlsf_resize, a, HEAP_PRIO_ALGO_NORMAL);
   if (status != HEAP_SUCCESS)
   {
      heap_meta_free(heap, a);
      return status;
   }

   status = zzheap_attach_hook(
      heap, HEAP_EVENT_COLLECT_REQ, l_tlsf_collect, a, HEAP_PRIO_ALGO_NORMAL);
   if (status != HEAP_SUCCESS)
   {
      heap_meta_free(heap, a);
      return status;
   }

   status = zzheap_attach_hook(
      heap, HEAP_EVENT_QUERY_REQ, l_tlsf_query, a, HEAP_PRIO_ALGO_NORMAL);
   if (status != HEAP_SUCCESS)
   {
      heap_meta_free(heap, a);
      return status;
   }

   status = zzheap_attach_hook(
      heap, HEAP_EVENT_CONTROL_REQ, l_tlsf_control, a, HEAP_PRIO_ALGO_NORMAL);
   if (status != HEAP_SUCCESS)
   {
      heap_meta_free(heap, a);
      return status;
   }

   status = zzheap_attach_hook(
      heap, HEAP_EVENT_DESTROY_RES, l_tlsf_destroy, a, HEAP_PRIO_ALGO_NORMAL);
   if (status != HEAP_SUCCESS)
   {
      heap_meta_free(heap, a);
      return status;
   }
   
   return HEAP_SUCCESS;
}
