/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2007 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */


#include "heap_alloc_region.h"
#include "mm.sig"

#ifdef __STRICT_ANSI__
#undef __STRICT_ANSI__ /* For strtok_r */
#include <string.h>
#define __STRICT_ANSI__
#else
#include <string.h>
#endif

#include <stdlib.h>
#include <stdio.h>


/**************************************************************************
 * Plug-in History
 **************************************************************************
 *
 * IMPORTANT
 * 
 * This comment shall be updated with a description of the change each
 * time the extension version number is increased. The version number
 * shall be increased if incompatible modifications are made to internal
 * structures or after behavioral changes. See HeapExtension below.
 * 
 **************************************************************************
 *
 *  Version  Description
 *  -------  ----------------------------------------------------------
 *   101     Renamed struct Allocator to struct RegionAllocator.
 *   100     Initial revision.  
 *
 **************************************************************************/


/***************************************************************************
 * Helper Macros
 ***************************************************************************/
#ifdef OSMIPS
#define PAGE_SIZE  0x2000
#else
#define PAGE_SIZE  0x1000
#endif

#define ALIGN(size)        (((size) + (PAGE_SIZE - 1)) & ~(PAGE_SIZE - 1))
#define ALIGN_DOWN(size)   ((size) & ~(PAGE_SIZE - 1))

#define FRAG_NEXT_FRAG_MASK        0xFFFFFFFCul
#define FRAG_NEXT_ID_MASK          0x00000003ul
#define FRAG_NEXT_FRAG(frag)       ((struct Fragment*)((uint32_t)(frag) & FRAG_NEXT_FRAG_MASK))
#define FRAG_NEXT_ID(frag)         (((uint32_t)(frag) & FRAG_NEXT_ID_MASK) << FRAG_ID_SHIFT)
#define FRAG_NEXT_MASK(frag, id)   ((struct Fragment*)((uint32_t)(frag)|((id) >> FRAG_ID_SHIFT) ))


/***************************************************************************
 * Data Structures
 ***************************************************************************/

struct PMem
{
   struct PMem *next;

   /* Configuration */
   char         name[MM_NAME_LENGTH];

   /* Statistics. ***/
   size_t       frag_size_used;
   size_t       frag_size_req;
};

/***************************************************************************/

struct LMem
{
   struct LMem *next;

   /* Configuration */
   char         name[MM_NAME_LENGTH];

   /* Statistics. ***/
   size_t       frag_size_used;
   size_t       frag_size_req;
};

/***************************************************************************/

struct RegionAllocator
{
   uint32_t            id;

   struct HeapConfiguration *config;
   struct HeapStatistics    *stats;
   struct HeapServiceInfo    info;           /* Plug-in information. */
   
   /* Fragments. */
   struct Fragment    *frag_list;
   uint32_t            frag_count;
   
   /* Specific configuration. */
   struct LMem        *lmems;
   struct PMem        *pmems;
   uint32_t            cmode;
   uint32_t            pmode;
   uint32_t            domain;
   HeapPID             owner;
   uint8_t             mas;
   size_t              slim;
   size_t              llim;
};

/***************************************************************************/

struct Fragment
{
   struct Fragment *next;
   struct Fragment *prev;
   struct LMem     *lmem;
   struct PMem     *pmem;
   size_t           act_size;
   size_t           req_size;
   uint32_t         chksum;
   uint32_t         control;

#define FRAG_ID_MASK                    0x30000000ul
#define FRAG_ID_SHIFT                   28
#define FRAG_ALGO_ID_MASK               0xC0000000ul
#define FRAG_ALGO_ID_SHIFT              30
#define frag_id(frag)                   ((frag)->control & FRAG_ID_MASK)
#define frag_set_control(frag,id,algo)  ((frag)->control = ((id)|(algo)))
};

/***************************************************************************/

struct FragmentCursor
{
   struct Fragment *frag_out;
   size_t           size_out;
   struct Fragment *frag;
   uint32_t         reserved;
};

/***************************************************************************/

union SIGNAL
{
   SIGSELECT signo;
   struct mm_create_region     create_region;
   struct mm_delete_region     delete_region;
   struct mm_get_log_mem_info  lmem_info;
   struct mm_get_phys_mem_info pmem_info;
};

/***************************************************************************/




/***************************************************************************
 * Forward Declarations
 ***************************************************************************/

static inline struct Fragment *
frag_find(struct Heap *heap, struct RegionAllocator *a,
          void *ptr, int exact, HeapError *ecode);

/***************************************************************************/


   

/***************************************************************************
 * Helper Functions
 ***************************************************************************/

static inline int
lmem_check(struct RegionAllocator *a, const char *name)
{
   static const SIGSELECT sigsel[] = {1, MM_GET_LOG_MEM_INFO_REPLY};
   union SIGNAL *sig = alloc(sizeof(struct mm_get_log_mem_info),
                             MM_GET_LOG_MEM_INFO_REQUEST);

   PROCESS gmm;
   hunt(GMM_NAME_STRING, 0, &gmm, NULL);

   if (gmm == 0)
   {
      free_buf(&sig);
      return -1;
   }
   
   sig->lmem_info.count = 1;
   sig->lmem_info.tags[0].tag_id = MM_TAG_INFO_LOG_MEM_TYPE;
   strlcpy(sig->lmem_info.name, name, MM_NAME_LENGTH);

   send(&sig, gmm);
   sig = receive(sigsel);

   if (sig->lmem_info.status != MM_SUCCESS)
   {
      free_buf(&sig);
      return -1;
   }

   if (sig->lmem_info.tags[0].tag_value == MAS)
      a->config->frag_mas = 1;

   free_buf(&sig);
   return 0;
}


/***************************************************************************/

static inline int
pmem_check(struct RegionAllocator *a, const char *name)
{
   static const SIGSELECT sigsel[] = {1, MM_GET_PHYS_MEM_INFO_REPLY};
   union SIGNAL *sig = alloc(sizeof(struct mm_get_log_mem_info),
                             MM_GET_PHYS_MEM_INFO_REQUEST);

   PROCESS gmm;
   hunt(GMM_NAME_STRING, 0, &gmm, NULL);

   (void)a;

   if (gmm == 0)
   {
      free_buf(&sig);
      return -1;
   }
   
   sig->pmem_info.count = 0;
   strlcpy(sig->pmem_info.name, name, MM_NAME_LENGTH);

   send(&sig, gmm);
   sig = receive(sigsel);

   if (sig->pmem_info.status != MM_SUCCESS)
   {
      free_buf(&sig);
      return -1;
   }

   free_buf(&sig);
   return 0;
}

/***************************************************************************/

static void
mem_cleanup(struct Heap *heap, struct RegionAllocator *a)
{
   struct PMem *pmem_last;
   struct LMem *lmem_last;
   
   while ((pmem_last = a->pmems) != NULL)
   {
      a->pmems = a->pmems->next;
      heap_meta_free(heap, pmem_last);
   }

   while ((lmem_last = a->lmems) != NULL)
   {
      a->lmems = a->lmems->next;
      heap_meta_free(heap, lmem_last);
   }
}

/***************************************************************************/

static inline uint32_t
frag_checksum(struct Fragment *frag)
{
#define FRAG_CHECKSUM_MAGIC 0x1234beef
   return (
      (uint32_t)frag->next ^
      (uint32_t)frag->prev ^
      (uint32_t)frag->lmem ^
      (uint32_t)frag->pmem ^
      (uint32_t)frag->act_size ^
      (uint32_t)frag->req_size ^
      ((uint32_t)frag->control & ~ FRAG_ALGO_ID_MASK) ^
      FRAG_CHECKSUM_MAGIC);
#undef FRAG_CHECKSUM_MAGIC
}

/***************************************************************************/

static inline void
frag_set_checksum(struct Fragment *frag)
{
   frag->chksum = frag_checksum(frag);
}

/***************************************************************************/

static inline uint32_t
frag_get_checksum(struct Fragment *frag)
{
   return frag->chksum;
}

/***************************************************************************/

static inline HeapError
frag_check_exists(struct Heap *heap, struct RegionAllocator *a,
                  struct Fragment *frag)
{
   HeapError ecode = HEAP_ERROR_FRAG_UNKNOWN;

   if (!a->config->frag_check)
      return HEAP_ERROR_NONE;
   
   frag = frag_find(heap, a, frag, 1, &ecode);
   if (frag != NULL)
      return HEAP_ERROR_NONE;

   if (ecode != HEAP_ERROR_NONE)
      return ecode;

   return HEAP_ERROR_FRAG_UNKNOWN;
}

/***************************************************************************/

static inline HeapError
frag_check_header(struct Heap *heap, struct RegionAllocator *a,
                  struct Fragment *frag)
{
   (void)heap;
   
   if (((uint32_t)frag) & (HEAP_ALIGNMENT-1))
      return HEAP_ERROR_FRAG_UNALIGNED;

   if (frag_id(frag) != a->id)
      return HEAP_ERROR_FRAG_UNKNOWN;

   if (frag_checksum(frag) != frag_get_checksum(frag))
      return HEAP_ERROR_FRAG_HEADER;

   if (frag->act_size < frag->req_size)
      return HEAP_ERROR_FRAG_HEADER;

   return HEAP_ERROR_NONE;
}

/***************************************************************************/

static inline HeapError
frag_check(struct Heap *heap, struct RegionAllocator *a,
           struct Fragment *frag)
{
   HeapError ecode;
   
   ecode = frag_check_exists(heap, a, frag);
   if (ecode != HEAP_ERROR_NONE)
      return ecode;
      
   ecode = frag_check_header(heap, a, frag);
   if (ecode != HEAP_ERROR_NONE)
      return ecode;

   return HEAP_ERROR_NONE;
}

/***************************************************************************/

static inline void
frag_insert(struct Heap *heap, struct RegionAllocator *a,
            struct Fragment *f)
{
   HeapError ecode;
   HeapLock lock;
   
   heap_lock(heap, &lock, 0);

   f->prev = NULL;
   f->next = a->frag_list;

   if (f->next)
   {
      ecode = frag_check_header(heap, a, f->next);
      if (ecode != HEAP_ERROR_NONE)
         zzheap_error(heap, HEAP_ERROR_FATAL|ecode, &(f->next)[1]);

      f->next->prev = f;

      frag_set_checksum(f->next);
   }

   a->frag_list = f;
   a->frag_count++;
   a->stats->frag_size_used += f->act_size;
   a->stats->frag_size_req  += f->req_size;
   f->lmem->frag_size_used  += f->act_size;
   f->lmem->frag_size_req   += f->req_size;
   f->pmem->frag_size_used  += f->act_size;
   f->pmem->frag_size_req   += f->req_size;

   frag_set_checksum(f);

   heap_unlock(heap, &lock, 0);
}

/***************************************************************************/

static inline void
frag_remove(struct Heap *heap, struct RegionAllocator *a, struct Fragment *f)
{
   HeapError ecode;

   if (f->prev)
   {
      ecode = frag_check_header(heap, a, f->prev);
      if (ecode != HEAP_ERROR_NONE)
         zzheap_error(heap, HEAP_ERROR_FATAL|ecode, &(f->prev)[1]);

      f->prev->next = f->next;

      frag_set_checksum(f->prev);
   }

   if (f->next)
   {
      ecode = frag_check_header(heap, a, f->next);
      if (ecode != HEAP_ERROR_NONE)
         zzheap_error(heap, HEAP_ERROR_FATAL|ecode, &(f->next)[1]);

      f->next->prev = f->prev;

      frag_set_checksum(f->next);
   }

   if (a->frag_list == f)
      a->frag_list = f->next;

   a->frag_count--;
   a->stats->frag_size_used -= f->act_size;
   a->stats->frag_size_req  -= f->req_size;
   f->lmem->frag_size_used  -= f->act_size;
   f->lmem->frag_size_req   -= f->req_size;
   f->pmem->frag_size_used  -= f->act_size;
   f->pmem->frag_size_req   -= f->req_size;
}

/***************************************************************************/

static inline struct Fragment *
frag_first(struct Heap *heap, struct RegionAllocator *a, HeapError *ecode)
{
   struct Fragment *frag = a->frag_list;

   *ecode = frag_check_header(heap, a, frag);
   if (*ecode != HEAP_ERROR_NONE)
      return NULL;
   
   return frag;
}

/***************************************************************************/

static inline struct Fragment *
frag_right(struct Heap *heap, struct RegionAllocator *a,
           struct Fragment *frag, HeapError *ecode)
{
   struct Fragment *next = frag->next;

   if (next == NULL)
      return NULL;
   
   *ecode = frag_check_header(heap, a, next);
   if (*ecode != HEAP_ERROR_NONE)
      return NULL;
   
   return next;
}

/***************************************************************************/

static inline struct Fragment *
frag_find(struct Heap *heap, struct RegionAllocator *a,
          void *ptr, int exact, HeapError *ecode)
{
   struct Fragment *frag = frag_first(heap, a, ecode);

   while (frag != NULL)
   {
      if ((uint32_t)ptr >= (uint32_t)frag &&
          (uint32_t)ptr <  (uint32_t)frag + frag->act_size)
      {
         if (exact && frag != ptr)
            return NULL;
         return frag;
      }

      frag = frag_right(heap, a, frag, ecode);
   }

   return NULL;
}

/***************************************************************************/

static inline struct Fragment *
frag_next(struct Heap *heap, struct RegionAllocator *a,
          struct FragmentCursor *cursor)
{
   struct Fragment *frag = FRAG_NEXT_FRAG(cursor->frag);
   HeapError ecode;

   /* First fragment? */
   if (frag == NULL)
   {
      frag = a->frag_list;

      if (frag == NULL)
         return NULL;

      cursor->frag = FRAG_NEXT_MASK(frag, a->id);
      cursor->frag_out = &frag[1];
      cursor->size_out = frag->act_size - sizeof(struct Fragment);
      
      return frag;
   }

   ecode = frag_check(heap, a, frag);
   if (ecode != HEAP_ERROR_NONE)
      return NULL;
   
   frag = frag->next;
   if (frag != NULL)
   {   
      cursor->frag = FRAG_NEXT_MASK(frag, a->id);
      cursor->frag_out = &frag[1];
      cursor->size_out = frag->act_size - sizeof(struct Fragment);
   }

   return frag;
}

/***************************************************************************/



/***************************************************************************
 * Private Interface
 ***************************************************************************/

static PROCESS gmm_pid = 0;

static PROCESS
l_hunt_gmm(void)
{
   if (gmm_pid == 0)
      hunt(GMM_NAME_STRING, 0, &gmm_pid, NULL);

   return gmm_pid;
}


/***************************************************************************/

static HeapAction
l_region_break(struct Heap *heap, union HeapEvent *event, void *opaque)
{
   static const SIGSELECT create_reply[] = {1, MM_CREATE_REGION_REPLY};
   union SIGNAL *sig;

   size_t size = event->e_break.min_size + sizeof(struct Fragment);
   struct RegionAllocator *a = opaque;
   struct LMem *lmem;
   struct PMem *pmem;
   HeapLock lock;

   /*
    * First check if the minimum size can be accomodated.
    * There is no reason to proceed any further otherwise. 
    */

   heap_lock(heap, &lock, 0);
   
   size = ALIGN(size);
   if (size + a->stats->frag_size_used > a->config->frag_size_max)
   {
      /* Update max_size field in event to indicate the largest amount
       * the client can allocate without violating the size limit, but
       * only if the client has requested this information. */

      event->e_break.max_size =
         a->config->frag_size_max - a->stats->frag_size_used;

      if (event->e_break.max_size > sizeof(struct Fragment))
         event->e_break.max_size -= sizeof(struct Fragment);
      else
         event->e_break.max_size = 0;

      heap_unlock(heap, &lock, 0);
      return HEAP_ACTION_CONTINUE;
   }

   /*
    * Minimum size is OK. Check if requested/preferred size can
    * be accomodated. If not, choose a smaller size. We know that
    * the requestor will get at least the minimum amount required.
    */
   size = event->e_break.req_size + sizeof(struct Fragment);
   size = ALIGN(size);

   if (size + a->stats->frag_size_used > a->config->frag_size_max)
   {
      /* Requested size is too large. Give the requestor whatever
       * size that remains before the heap becomes exhausted. */
      size = a->config->frag_size_max - a->stats->frag_size_used;
      size = ALIGN_DOWN(size);
   }

   heap_unlock(heap, &lock, 0);

   for (lmem = a->lmems; lmem != NULL; lmem = lmem->next)
   {
      for (pmem = a->pmems; pmem != NULL; pmem = pmem->next)
      {
         sig = alloc(sizeof(struct mm_create_region), MM_CREATE_REGION_REQUEST);

         sig->create_region.domain      = a->domain;
         sig->create_region.size        = size;
         sig->create_region.permissions = a->pmode;
         sig->create_region.cache_mode  = a->cmode;
         sig->create_region.options     = MM_CREATE_REGION_SPECIFY_NAME|MM_CHECK_LOW_LIMIT;

         if (event->e_break.algo_id == 0)
            sig->create_region.low_limit   = a->slim;
         else
            sig->create_region.low_limit   = a->llim;

         snprintf(sig->create_region.name, MM_NAME_LENGTH,
                  "heap_%u:%x", a->domain, a->owner);
         strlcpy(sig->create_region.log_mem, lmem->name, MM_NAME_LENGTH);
         strlcpy(sig->create_region.phys_mem, pmem->name, MM_NAME_LENGTH);

         send(&sig, l_hunt_gmm());
         sig = receive(create_reply);

         if (sig->create_region.status == MM_SUCCESS)
         {
            struct Fragment *frag = (void *)sig->create_region.base;

            frag_set_control(frag, a->id, event->e_break.algo_id);
            frag->act_size = sig->create_region.size;
            frag->req_size = event->e_break.min_size;
            frag->lmem     = lmem;
            frag->pmem     = pmem;

            free_buf(&sig);

            frag_insert(heap, a, frag);
            
            event->e_break.base = &frag[1];
            event->e_break.min_size = frag->act_size - sizeof(struct Fragment);
            event->e_break.max_size = frag->act_size - sizeof(struct Fragment);
            
            return HEAP_ACTION_INTERCEPT;
         }
         else
         {
            event->e_break.status = sig->create_region.status;
            free_buf(&sig);
         }
      }
   }

   event->e_break.base = NULL;
   return HEAP_ACTION_CONTINUE;
}

/***************************************************************************/

static HeapAction
l_region_unbreak(struct Heap *heap, union HeapEvent *event, void *opaque)
{
   static const SIGSELECT delete_reply[] = {1, MM_DELETE_REGION_REPLY};
   union SIGNAL *sig;

   struct Fragment *frag = event->e_unbreak.base;
   struct RegionAllocator *a = opaque;

   HeapError ecode;
   HeapLock lock;


   if (frag == NULL)
      return HEAP_ACTION_CONTINUE;

   frag = &frag[-1];


   heap_lock(heap, &lock, 0);
   
   ecode = frag_check(heap, a, frag);
   if (ecode == HEAP_ERROR_FRAG_UNKNOWN)
   {
      heap_unlock(heap, &lock, 0);
      return HEAP_ACTION_CONTINUE;
   }

   if (ecode != HEAP_ERROR_NONE)
   {
      heap_unlock(heap, &lock, 0);
      zzheap_error(heap, ecode, &frag[1]);
      return HEAP_ACTION_INTERCEPT;
   }
  
   frag_remove(heap, a, frag);

   heap_unlock(heap, &lock, 0);

   sig = alloc(sizeof(struct mm_delete_region), MM_DELETE_REGION_REQUEST);
   sig->delete_region.domain = a->domain;
   sig->delete_region.base   = (OSADDRESS)frag;
   send(&sig, l_hunt_gmm());
   sig = receive(delete_reply);

   if (sig->delete_region.status == MM_SUCCESS)
   {
      free_buf(&sig);
      return HEAP_ACTION_INTERCEPT;
   }

   free_buf(&sig);

   /* Framework will report an error if no other
    * plug-in intercepts and unbreaks this fragment. */
   return HEAP_ACTION_CONTINUE;
}

/***************************************************************************/

static HeapAction
l_region_query(struct Heap *heap, union HeapEvent *event, void *opaque)
{
   HeapParam *p = event->e_query.params;
   struct RegionAllocator *a = opaque;
   HeapLock lock;

   if (p == NULL)
      return HEAP_ACTION_CONTINUE;

   if (event->e_query.freezed)
      heap->freezed++;
   
   while (*p != HPM_END)
   {
      switch (HPM_ID(*p))
      {
         case HPM_QUERY_SERVICE:
         {
            if (!HPM_IS_HANDLED(p[0]) && p[1] == 0)
            {
               HPM_SET_HANDLED(p[0]);
               HPM_SET_VALUE(p[1], &a->info);
            }
            else if (p[1] == (HeapParam)&a->info)
            {
               HPM_CLEAR_HANDLED(p[0]);
               HPM_SET_VALUE(p[1], 0);
            }

            break;
         }
         case HPM_QUERY_FRAG_MAS:
         {
            HPM_SET_HANDLED(p[0]);
            if (a->config->frag_mas != 0)
               HPM_SET_VALUE(p[1], 1);
            break;
         }
         case HPM_QUERY_FRAG_TOT_USED_SIZE:
         {
            HPM_SET_HANDLED(p[0]);
            HPM_SET_VALUE(p[1], a->stats->frag_size_used);
            break;
         }
         case HPM_QUERY_FRAG_TOT_FREE_SIZE:
         {
            /* The allocator never has any free fragments.
             * This information is managed by the fragment cache.*/
            break;
         }
         case HPM_QUERY_FRAG_NEXT:
         {
            struct FragmentCursor *cursor = (struct FragmentCursor *)&p[1];

            /* Skip if the request doesn't concern us. */
            if (cursor->frag != NULL && FRAG_NEXT_ID(cursor->frag) != a->id)
               break;

            if (!HPM_IS_HANDLED(p[0]))
            {
               struct Fragment *frag;

               heap_lock(heap, &lock, 0);
               
               frag = frag_next(heap, a, cursor);
               if (frag == NULL)
               {
                  /* Initialize query for next algorithm. */
                  cursor->frag     = NULL;
                  cursor->frag_out = NULL;
                  cursor->size_out = 0;
                  heap_unlock(heap, &lock, 0);
                  break;
               }

               HPM_SET_HANDLED(p[0]);

               heap_unlock(heap, &lock, 0);
            }
            break;
         }
         case HPM_QUERY_FRAG_TYPE:
         {
            struct Fragment *frag = &((struct Fragment *)p[1])[-1];

            if (!HPM_IS_HANDLED(p[0]))
            {
               heap_lock(heap, &lock, 0);
            
               if (frag_check(heap, a, frag) != HEAP_ERROR_NONE)
               {
                  heap_unlock(heap, &lock, 0);
                  break;
               }

               HPM_SET_HANDLED(p[0]);
               HPM_SET_VALUE(p[1], HEAP_FRAGMENT_REGION);

               heap_unlock(heap, &lock, 0);
            }
            break;
         }
         case HPM_QUERY_FRAG_STATUS:
         {
            struct Fragment *frag = &((struct Fragment *)p[1])[-1];
            HeapError ecode;

            if (!HPM_IS_HANDLED(p[0]))
            {
               heap_lock(heap, &lock, 0);

               ecode = frag_check(heap, a, frag);
               if (ecode == HEAP_ERROR_FRAG_UNKNOWN ||
                   ecode == HEAP_ERROR_FRAG_UNALIGNED)
               {
                  heap_unlock(heap, &lock, 0);
                  break;
               }

               if (ecode == HEAP_ERROR_FRAG_HEADER)
               {
                  HPM_SET_HANDLED(p[0]);
                  HPM_SET_VALUE(p[2], HEAP_FRAGMENT_HEADERS);
               }
               else
               {
                  HPM_SET_HANDLED(p[0]);
                  HPM_SET_VALUE(p[2], HEAP_FRAGMENT_ALLOCATED);
               }

               heap_unlock(heap, &lock, 0);
            }
            break;
         }
         case HPM_QUERY_FRAG_SIZE:
         {
            struct Fragment *frag = &((struct Fragment *)p[1])[-1];

            if (!HPM_IS_HANDLED(p[0]))
            {
               heap_lock(heap, &lock, 0);

               if (frag_check(heap, a, frag) != HEAP_ERROR_NONE)
               {
                  heap_unlock(heap, &lock, 0);
                  break;
               }

               HPM_SET_HANDLED(p[0]);
               HPM_SET_VALUE(p[2], frag->act_size - sizeof(struct Fragment));

               heap_unlock(heap, &lock, 0);
            }
            break;
         }
         case HPM_QUERY_FRAG_PMEM:
         {
            struct Fragment *frag = &((struct Fragment *)p[1])[-1];

            if (!HPM_IS_HANDLED(p[0]))
            {
               char *str = (char *)p[2];
               
               heap_lock(heap, &lock, 0);

               if (frag_check(heap, a, frag) != HEAP_ERROR_NONE)
               {
                  heap_unlock(heap, &lock, 0);
                  break;
               }

               HPM_SET_HANDLED(p[0]);
               strlcpy(str, frag->pmem->name, MM_NAME_LENGTH+1);
            
               heap_unlock(heap, &lock, 0);
            }
            break;
         }
         case HPM_QUERY_FRAG_LMEM:
         {
            struct Fragment *frag = &((struct Fragment *)p[1])[-1];

            if (!HPM_IS_HANDLED(p[0]))
            {
               char *str = (char *)p[2];
               
               heap_lock(heap, &lock, 0);

               if (frag_check(heap, a, frag) != HEAP_ERROR_NONE)
               {
                  heap_unlock(heap, &lock, 0);
                  break;
               }

               HPM_SET_HANDLED(p[0]);
               strlcpy(str, frag->lmem->name, MM_NAME_LENGTH+1);
            
               heap_unlock(heap, &lock, 0);
            }
            break;
         }
         case HPM_QUERY_FRAG_COUNT:
         {
            HPM_SET_HANDLED(p[0]);
            HPM_ADD_VALUE(p[1], a->frag_count);
            break;
         }
         case HPM_QUERY_FRAG_MAX_SIZE:
         {
            if (!HPM_IS_HANDLED(p[0]))
            {
               HPM_SET_HANDLED(p[0]);
               HPM_SET_VALUE(p[1], a->config->frag_size_max);
            }
            break;
         }
         case HPM_QUERY_FRAG_DEFAULT_SIZE:
         {
            if (!HPM_IS_HANDLED(p[0]))
            {
               HPM_SET_HANDLED(p[0]);
               HPM_SET_VALUE(p[1], ALIGN(a->config->frag_size_default));
            }
            break;
         }
         case HPM_QUERY_FRAG_NEXT_SIZE:
         {
            if (!HPM_IS_HANDLED(p[0]))
            {
               size_t s = p[1] + sizeof(struct Fragment);

               s  = ALIGN(s);
               s -= sizeof(struct Fragment);

               HPM_SET_HANDLED(p[0]);
               HPM_SET_VALUE(p[1], s);
            }
            break;
         }
         case HPM_QUERY_FRAG_PREV_SIZE:
         {
            if (!HPM_IS_HANDLED(p[0]))
            {
               size_t s = p[1];

               s = ALIGN_DOWN(s);

               if (s <= sizeof(struct Fragment))
               {
                  HPM_SET_HANDLED(p[0]);
                  HPM_SET_VALUE(p[1], 0);
               }
               else
               {
                  s -= sizeof(struct Fragment);
                  HPM_SET_HANDLED(p[0]);
                  HPM_SET_VALUE(p[1], s);
               }
            }
            break;
         }

         default:
         {
            break;
         }
      }

      p += (1 + HPM_VALUES(*p));
   }

   if (event->e_query.freezed)
      heap->freezed--;

   return HEAP_ACTION_CONTINUE;
}

/***************************************************************************/

static HeapAction
l_region_control(struct Heap *heap, union HeapEvent *event, void *opaque)
{
   HeapParam *p = event->e_query.params;
   struct RegionAllocator *a = opaque;

   if (p == NULL)
      return HEAP_ACTION_CONTINUE;
   
   while (*p != HPM_END)
   {
      switch (HPM_ID(*p))
      {
         case HPM_CTRL_FRAG_MAX_SIZE:
         {
            if (!HPM_IS_HANDLED(p[0]))
            {
               size_t s = ALIGN((size_t)p[1]);
               if (p[1] == 0 || s != 0)
               {
                  HPM_SET_HANDLED(p[0]);
                  a->config->frag_size_max = s;
               }
            }
            break;
         }
         case HPM_CTRL_FRAG_DEFAULT_SIZE:
         {
            if (!HPM_IS_HANDLED(p[0]))
            {
               size_t s = ALIGN((size_t)p[1]);
               if (p[1] == 0 || s != 0)
               {
                  HeapParam q[] = {HPM(QUERY_FRAG_PREV_SIZE,1),0,HPM_END};
                  q[1] = s;
                  zzheap_query(heap, q, 0);
                  
                  HPM_SET_HANDLED(p[0]);
                  a->config->frag_size_default = q[1];
               }
            }
            break;
         }
         case HPM_CTRL_FRAG_REMOVE:
         {
            struct Fragment *frag = (struct Fragment *)p[1];
            HeapLock lock;
            
            if (!HPM_IS_HANDLED(p[0]))
               break;

            if (frag == NULL)
               break;
            
            frag = &frag[-1];
            
            if (frag_id(frag) != a->id)
               break;

            /* Update parameter values. */
            HPM_SET_VALUE(p[1], frag);
            HPM_ADD_VALUE(p[2], sizeof(struct Fragment));
            
            /* Unregister fragment. */
            heap_lock(heap, &lock, 0);
            frag_remove(heap, a, frag);
            heap_unlock(heap, &lock, 0);

            break;
         }
         default:
         {
            break;
         }
      }

      p += (1 + HPM_VALUES(*p));
   }

   return HEAP_ACTION_CONTINUE;
}

/***************************************************************************/

static HeapAction
l_region_destroy(struct Heap *heap, union HeapEvent *event, void *opaque)
{
   struct RegionAllocator *a = opaque;
   struct Fragment *frag = NULL;

   (void)event;
   
   /* Unregister this plug-in. */
   zzheap_detach_hook(heap, HEAP_EVENT_BREAK_REQ,   l_region_break,   a);
   zzheap_detach_hook(heap, HEAP_EVENT_UNBREAK_REQ, l_region_unbreak, a);
   zzheap_detach_hook(heap, HEAP_EVENT_CONTROL_REQ, l_region_control, a);
   zzheap_detach_hook(heap, HEAP_EVENT_QUERY_REQ,   l_region_query,   a);

   if (a == NULL)
      return HEAP_ACTION_CONTINUE;

   /* FIXME: Assumption that the memory manager
    * reclaims all regions in application domains. */
   if (a->domain == 0)
   {
      /* Unbreak all fragments. */
      frag = a->frag_list;
      while (frag != NULL)
      {
         static const SIGSELECT delete_reply[] = {1, MM_DELETE_REGION_REPLY};

         struct Fragment *frag_next = frag->next;

         union SIGNAL *sig =
            alloc(sizeof(struct mm_delete_region), MM_DELETE_REGION_REQUEST);
         sig->delete_region.domain = a->domain;
         sig->delete_region.base   = (OSADDRESS)frag;

         send(&sig, l_hunt_gmm());
         sig = receive(delete_reply);

         free_buf(&sig);
         frag = frag_next;
      }
   }

   /* Free all memory descriptors. */
   mem_cleanup(heap, a);
   heap_meta_free(heap, a);

   return HEAP_ACTION_CONTINUE;
}

/***************************************************************************/




/***************************************************************************
 * Public Interface
 ***************************************************************************/

HeapStatus
heap_attach_region_allocator(struct Heap *heap,
                             struct HeapStatistics *stats,
                             struct HeapConfiguration *config,
                             uint32_t id, uint32_t domain,
                             HeapPID owner, char *lmems, char *pmems,
                             uint32_t cmode, uint32_t pmode,
                             size_t slim, size_t llim)
{
   struct RegionAllocator *a = NULL;
   char *name, *endptr = NULL;
   struct PMem *pmem_last = NULL;
   struct LMem *lmem_last = NULL;
   HeapStatus status = HEAP_SUCCESS;
   static char ram[] = MM_INTERNAL_RAM;
   mm_bool_t add_ram_fall_back = TRUE;

   a = heap_meta_alloc(heap, sizeof(struct RegionAllocator));
   if (a == NULL)
      return HEAP_EOUT_OF_MEMORY;

   bzero(a, sizeof(struct RegionAllocator));

   a->id     = id << FRAG_ID_SHIFT;
   a->stats  = stats;
   a->config = config;

   a->lmems  = NULL;
   a->pmems  = NULL;
   a->cmode  = cmode;
   a->pmode  = pmode;
   a->domain = domain;
   a->owner  = owner;
   a->slim   = slim;
   a->llim   = llim;

   a->info.version = 101;
   strcpy(a->info.id, "com.enea.heap.region");

   if (pmems == NULL)
      goto skip_list;

   if (lmems == NULL)
      lmems = ram;
   
   name = strtok_r(pmems, " ,", &endptr);
   while (name != NULL)
   {
      struct PMem *pmem_next = heap_meta_alloc(heap, sizeof(struct PMem));
      if (pmem_next == NULL)
      {
         status = HEAP_EOUT_OF_MEMORY;
         goto cleanup;
      }
      memset(pmem_next, 0, sizeof(struct PMem));
      
      strlcpy(pmem_next->name, name, MM_NAME_LENGTH);
      pmem_next->next = NULL;

      /* Tail insertion. */
      if (a->pmems == NULL)
         a->pmems = pmem_next;
      else
         pmem_last->next = pmem_next;

      pmem_last = pmem_next;

      /* Verify that the phys_mem exists. */
      if (pmem_check(a, pmem_next->name) < 0)
      {
         status = HEAP_EPHYS_MEM_NOT_FOUND;
         goto cleanup;
      }
      
      if (strcmp(name, MM_INTERNAL_RAM) == 0)
      {
         add_ram_fall_back = FALSE;
      }
      
      name = strtok_r(NULL, " ,", &endptr);
   }

   
  skip_list:
   if (add_ram_fall_back)
   {
      /* Add regular "RAM" as fallback. */
      do {
         struct PMem *pmem_next = heap_meta_alloc(heap, sizeof(struct PMem));
         if (pmem_next == NULL)
         {
            status = HEAP_EOUT_OF_MEMORY;
            goto cleanup;
         }
         memset(pmem_next, 0, sizeof(struct PMem));

         strlcpy(pmem_next->name, MM_INTERNAL_RAM, MM_NAME_LENGTH);
         pmem_next->next = NULL;

         /* Tail insertion. */
         if (a->pmems == NULL)
            a->pmems = pmem_next;
         else
            pmem_last->next = pmem_next;

         pmem_last = pmem_next;

         /* Verify that the phys_mem exists. */
         if (pmem_check(a, pmem_next->name) < 0)
         {
            status = HEAP_EPHYS_MEM_NOT_FOUND;
            goto cleanup;
         }
      } while (0);
   }

   name = strtok_r(lmems, " ,", &endptr);
   while (name != NULL)
   {
      struct LMem *lmem_next = heap_meta_alloc(heap, sizeof(struct LMem));
      if (lmem_next == NULL)
      {
         status = HEAP_EOUT_OF_MEMORY;
         goto cleanup;
      }
      memset(lmem_next, 0, sizeof(struct LMem));
      
      strlcpy(lmem_next->name, name, MM_NAME_LENGTH);

      /* Tail insertion. */
      if (a->lmems == NULL)
         a->lmems = lmem_next;
      else
         lmem_last->next = lmem_next;

      lmem_last = lmem_next;

      /* Verify that the log_mem exists and check if it is MAS. */
      if (lmem_check(a, lmem_next->name) < 0)
      {
         status = HEAP_ELOG_MEM_NOT_FOUND;
         goto cleanup;
      }

      name = strtok_r(NULL, " ,", &endptr);
   }

   /* Creating a MAS heap in the system domain is illegal. */
   if (a->config->frag_mas && domain == MM_KERNEL_DOMAIN)
   {
      status = HEAP_EILLEGAL_MAP_TYPE;
      goto cleanup;
   }
   
   status = zzheap_attach_hook(
      heap, HEAP_EVENT_BREAK_REQ, l_region_break, a, HEAP_PRIO_ALLOC_NORMAL);
   if (status != HEAP_SUCCESS)
      goto cleanup;

   status = zzheap_attach_hook(
      heap, HEAP_EVENT_UNBREAK_REQ, l_region_unbreak, a, HEAP_PRIO_ALLOC_NORMAL);
   if (status != HEAP_SUCCESS)
      goto cleanup;

   status = zzheap_attach_hook(
      heap, HEAP_EVENT_CONTROL_REQ, l_region_control, a, HEAP_PRIO_ALLOC_NORMAL);
   if (status != HEAP_SUCCESS)
      goto cleanup;

   status = zzheap_attach_hook(
      heap, HEAP_EVENT_QUERY_REQ, l_region_query, a, HEAP_PRIO_ALLOC_NORMAL);
   if (status != HEAP_SUCCESS)
      goto cleanup;

   status = zzheap_attach_hook(
      heap, HEAP_EVENT_DESTROY_RES, l_region_destroy, a, HEAP_PRIO_ALLOC_NORMAL);
   if (status != HEAP_SUCCESS)
      goto cleanup;

   return HEAP_SUCCESS;
   
cleanup:
   if (a != NULL)
   {
      mem_cleanup(heap, a);
      heap_meta_free(heap, a);
   }

   return status;
}

