/* COPYRIGHT-ENEA-SRC-R2 *
**************************************************************************
* Copyright (C) 2010 by Enea Software AB.
* All rights reserved.
*
* This Software is furnished under a software license agreement and
* may be used only in accordance with the terms of such agreement.
* Any other use or reproduction is prohibited. No title to and
* ownership of the Software is hereby transferred.
*
* PROPRIETARY NOTICE
* This Software consists of confidential information.
* Trade secret law and copyright law protect this Software.
* The above notice of copyright on this Software does not indicate
* any actual or intended publication of such Software.
**************************************************************************
* COPYRIGHT-END */

#define USE_HEAP_FLS
#include "heap_filt_bcache.h"
#include "heap_list.h"
#include "heap_param.h"
#include "ose.h"
#include "slab.h"
#include "spinlock.h"
#include <string.h>


/**************************************************************************
 * Plug-in History
 **************************************************************************
 *
 * IMPORTANT
 * 
 * This comment shall be updated with a description of the change each
 * time the extension version number is increased. The version number
 * shall be increased if incompatible modifications are made to internal
 * structures or after behavioral changes. See HeapExtension below.
 * 
 **************************************************************************
 *
 *  Version  Description
 *  -------  ----------------------------------------------------------
 *   101     Renamed struct Filter to struct BCacheFilter.
 *   100     Initial revision.  
 *
 **************************************************************************/


#define CACHE_LISTS ose_num_cpus()

#define NUM_LISTS  27
#define MIN_BITS   4
#define MIN_SIZE   16


/**************************************************************************
 * Data Structures
 **************************************************************************/

struct LocalCache;

struct BCacheFilter
{
   struct HeapStatistics    *stats;
   struct HeapConfiguration *config;
   struct HeapServiceInfo    info;        /* Plug-in information. */

   HeapPID   *daemon;   /* Heap daemon identifier. To avoid garbage collected
                         * buffers in the cache -  */
   uint32_t   smax  ;   /* Configured maximum cachable buffer size.         */
   uint32_t   nmax;     /* Configured maximum number of cachable buffers.   */
   HeapBufferType type; /* Cached buffer type.                              */
   uint32_t   mask;     /* Buffer header size mask. (Optimization)          */
   spinlock_t lock;
   struct LocalCache *caches[1];
};

/**************************************************************************/

/* The 'standard' buffer header. Note that TLSF uses
 * a different format where the owner field is missing. */
struct Buffer
{
   HeapPID  owner;
   uint32_t control;

#define BUF_SIZE_MASK          0x01FFFFFEul
#define BUF_SIZE_SHIFT         1
#define BUF_ID_MASK            0xC0000000ul
#define buf_id(buf)            ((buf)->control & BUF_ID_MASK)
};

/**************************************************************************/

struct FreeBuffer
{
   struct Buffer      buf;
   struct FreeBuffer *next;
   struct FreeBuffer *prev;
};

/**************************************************************************/

/*
 * The tread-local buffer cache structure.
 *
 */
struct LocalCache
{
   list_link_t link;          /* Linkage for PID indexed hash table.       */
   struct FreeBuffer *list[NUM_LISTS];  /* Head of freelists.              */
   struct FreeBuffer *tail[NUM_LISTS];  /* Tail of freelists.              */
   uint32_t   ncur;           /* Number of buffers currently cached.       */
   uint32_t   icur;           /* Freelist index for next garbage collect.  */ 
   uint32_t   disable;        /* Temporarily disables the cache if set.    *
                               * Prevents recursion if the cache is full   *
                               * and a buffer is evicted.                  */
   spinlock_t lock;
};

/**************************************************************************/

static struct slab_cache *bcache_slab = NULL;

/**************************************************************************/



/**************************************************************************
 * Helper Functions
 **************************************************************************/

#ifndef ASSERT
/*
 * FIXME!
 * 
 * This is an optimization to increase performance that
 * relies too much on knowledge of specific algorithm
 * implementations. The buffer header should be standardized
 * more strictly for this to be comfortable.
 *
 */
#define buf_size(heap, f, buf) (((buf)->control & (f)->mask) >> BUF_SIZE_SHIFT)
#else
static inline size_t
buf_size(struct Heap *heap, struct BCacheFilter *f, struct Buffer *buf)
{
   HeapParam p[] = {
      HPM(QUERY_BUF_REQ_SIZE, 1), 0,
      HPM_END
   };

   (void)f;

   p[1] = (HeapParam)buf;
   zzheap_query(heap, p, 0);
    
   return HPM_IS_HANDLED(p[0]) ? (size_t)p[1] : 0;
}
#endif

/*****************************************************************************/

static always_inline uint32_t
size_index(size_t size)
{
   if (size < MIN_SIZE)
      return 0;

   return fls(size) - MIN_BITS;
}

/**************************************************************************/

static always_inline uint32_t
size_align(size_t size)
{
   /* Align to closest power of two. */
   
   size--;
   size |= size >> 1;
   size |= size >> 2;
   size |= size >> 4;
   size |= size >> 8;
   size |= size >> 16;
   size++;

   return size;
}

/**************************************************************************/

static always_inline HeapPID *
buf_owner(struct BCacheFilter *f, struct Buffer *buf, size_t size)
{
   /* This is an optimization to increase performance that
    * relies too much on knowledge of specific algorithm
    * implementations. The buffer header should be standardized
    * more strictly for this to be comfortable. */
   
   if (f->type == HEAP_BUFFER_TLSF)
   {
      size = (size < 16) ? 16 : ((size + 7) & ~7);
      return &((struct Buffer*)((char *)buf + size))->owner;
   }
   
   return &buf->owner;
}

/**************************************************************************/

static HeapError
buf_check(struct BCacheFilter *f, struct Buffer *buf,
          size_t size, HeapPID pid1)
{
   int i = f->config->buf_size_endmark;
   char *endm = (char *)buf + size - i;
   HeapPID *pid2 = buf_owner(f, buf, size);

   /* The buffer cache only validates the buffer endmark
    * and correct buffer ownership. A valid endmark most
    * likely indicates a valid buffer header as well and
    * the buffer owner will have to be changed anyway to
    * accomodate the use-case where a cached private
    * buffer is reallocated as shared and later freed by
    * a different process using private APIs. */
   
   while (--i>=0)
   {
      if (unlikely(endm[i] != HEAP_ENDMARK))
         return HEAP_ERROR_BUF_ENDMARK;
   }

   if (pid1 != 0 && *pid2 != 0 && pid1 != *pid2)
      return HEAP_ERROR_BUF_OWNER;

   *pid2 = current_process();
   
   return HEAP_ERROR_NONE;
}

/**************************************************************************/

static inline struct Buffer *
buf_insert(struct Heap *heap,
           struct BCacheFilter *f,
           struct LocalCache *cache,
           struct Buffer *buf, size_t size)
{
   struct FreeBuffer *evict = NULL;
   struct FreeBuffer *fb = (struct FreeBuffer *)buf;
   uint32_t index = size_index(size);
   Msr sr;

   (void)heap;
   
   /* The cache have to be protected from concurrent
    * garbage collection carried out by other processes.
    * Without global GC of all caches, no locks would be
    * required at all since the cache is thread-local. */
   spin_lock_irqsave(&cache->lock, sr);

   *buf_owner(f, buf, size) = 0;

   
   /* Check if the cache is full. If so, one buffer
    * will be evicted in a pseudo-LRU fashion. One
    * could argue that this only adds overhead to
    * almost any free() and that the new buffer simply
    * should be freed permanently. While this is true,
    * the cache was originally designed to boost
    * the malloc-free-malloc-free use-case (same buffer)
    * which it still does very well . */
   if (cache->ncur == f->nmax)
   {
      evict = cache->tail[cache->icur];
      while (evict == NULL)
      {
         cache->icur += 1;
         cache->icur %= NUM_LISTS;
         evict = cache->tail[cache->icur];
      }

      if (evict != NULL)
      {
         if (evict->prev)
            evict->prev->next = evict->next;

         if (evict->next)
            evict->next->prev = evict->prev;

         cache->tail[cache->icur] = evict->prev;

         if (cache->list[cache->icur] == evict)
            cache->list[cache->icur] = NULL;

         cache->ncur--;

         if (cache->icur == index)
         {
            cache->icur += 1;
            cache->icur %= NUM_LISTS;
         }
      }
   }

   /* Insert the new buffer into its freelist. */
   fb->prev = NULL;
   fb->next = cache->list[index];

   if (fb->next)
      fb->next->prev = fb;

   if (cache->tail[index] == NULL)
      cache->tail[index] = fb;

   cache->list[index] = fb;
   cache->ncur++;

   spin_unlock_irqrestore(&cache->lock, sr);

   return &evict->buf;
}

/**************************************************************************/

static inline struct Buffer *
buf_remove_head(struct Heap *heap, struct BCacheFilter *f,
                struct LocalCache *cache, size_t size, HeapPID owner)
{
   /* Round upwards to a power of two and calculate the freelist index.
    * Any buffer in that list will suite the request. */
   uint32_t index = size_index(size_align(size));
   struct FreeBuffer *fb;
   Msr sr;

   (void)heap;
   
   /* The cache have to be protected from concurrent
    * garbage collection carried out by other processes.
    * Without global GC of all caches, no locks would be
    * required at all since the cache is thread-local. */
   spin_lock_irqsave(&cache->lock, sr);

   /* The cache will first check if there is a buffer in
    * the selected larger size freelist. If there isn't,
    * it will have to search the freelist of the smaller
    * size list to find a suitable buffer. If that also
    * fails it gives up and lets the regular algorithms
    * do their magic. */

   /* Quick fetch of larger buffer (sort of half-fit). */
   fb = cache->list[index];
   if (fb == NULL)
   {
      if (index > 0)
      {
         /* Try a smaller size list. Note that the entire list is
          * traversed and that this therefore may take a long time.
          * The user should be made aware that a too large cache
          * can have negative impact on performance. */

         index--;
         for (fb = cache->list[index]; fb != NULL; fb = fb->next)
         {
            /* Stop as soon as we get a hit. */
            if (buf_size(heap, f, &fb->buf) >= size)
               break;
         }
      }

      if (fb == NULL)
      {
         /* Nothing found and we don't want to pick a too large
          * buffer here. Give up and yield to the algorithms. */
         spin_unlock_irqrestore(&cache->lock, sr);
         return NULL;
      }
   }

   /* Got a buffer that will do.
    * Remove it from the cache. */

   if (fb->prev)
      fb->prev->next = fb->next;

   if (fb->next)
      fb->next->prev = fb->prev;

   if (cache->tail[index] == fb)
      cache->tail[index] = fb->prev;

   if (cache->list[index] == fb)
      cache->list[index] = fb->next;

   cache->ncur--;

   /* Change owner in order to cope with private-shared transitions. */
   *buf_owner(f, &fb->buf, buf_size(heap, f, &fb->buf)) = owner;

   spin_unlock_irqrestore(&cache->lock, sr);

   return &fb->buf;
}

/**************************************************************************/

static inline struct LocalCache *
cache_get(struct BCacheFilter *f)
{
   cpuid_t cpu = ose_cpu_id();
   struct LocalCache *cache = f->caches[cpu];

   if (unlikely(cache == NULL))
   {
      Msr sr;

      if (current_process() == *f->daemon)
         return NULL;

      cache = slab_alloc(bcache_slab, SLAB_SLEEP);
      if (cache == NULL)
         return NULL;

      memset(cache, 0, sizeof(struct LocalCache));
      spin_lock_init(&cache->lock);
   
      spin_lock_irqsave(&f->lock, sr);
      if (f->caches[cpu] != NULL)
      {
         slab_free(bcache_slab, cache);
         cache = NULL;
      }
      else
      {
         f->caches[cpu] = cache;
      }
      spin_unlock_irqrestore(&f->lock, sr);
   }

   return cache;
}

/**************************************************************************/

static inline uint32_t
cache_flush(struct Heap *heap,
            struct BCacheFilter *filter, struct LocalCache *cache)
{
   uint32_t index;
   uint32_t success = 0;

   (void)filter;

   /* This function garbage collects all
    * buffers in the specified local cache.
    *
    * The cache have to be temporarily disabled
    * to prevent buffers from reentering the cache
    * during free. It will ultimately lead to an
    * infinite loop.
    */

   for (index = 0; index < NUM_LISTS; index++)
   {
      struct FreeBuffer *fb;
      Msr sr;

      spin_lock_irqsave(&cache->lock, sr);

      fb = cache->list[index];
      while (fb != NULL)
      {
         if (fb->prev)
            fb->prev->next = fb->next;

         if (fb->next)
            fb->next->prev = fb->prev;

         if (cache->tail[index] == fb)
            cache->tail[index] = fb->prev;
   
         cache->list[index] = fb->next;
         cache->ncur--;

         cache->disable++;
         spin_unlock_irqrestore(&cache->lock, sr);

         success = 1;
         zzheap_free(heap, &(&fb->buf)[1], 0, NULL, __FILE__, __LINE__);

         spin_lock_irqsave(&cache->lock, sr);

         cache->disable--;
         fb = cache->list[index];
      }

      spin_unlock_irqrestore(&cache->lock, sr);
   }

   return success;
}

/**************************************************************************/

static inline uint32_t
cache_flush_all(struct Heap *heap, struct BCacheFilter *f)
{
   uint32_t success = 0, i;

   /* Flush all local caches associated with the
    * specified heap. This will typically occur when
    * a heap allocation is about to fail with a NULL
    * pointer or during memory reclaim events that are
    * initiated by the memory manager. */
   
   for (i = 0; i < CACHE_LISTS; i++)
   {
      struct LocalCache *cache = f->caches[i];

      if (cache != NULL)
      {
         success |= cache_flush(heap, f, cache);
      }
   }

   return success;
}

/**************************************************************************/



/**************************************************************************
 * Private Interface
 **************************************************************************/

static HeapAction
l_bcache_alloc(struct Heap *heap, union HeapEvent *event, void *opaque)
{
   struct BCacheFilter *f = opaque;
   struct LocalCache *c = cache_get(f);
   size_t size = event->e_alloc.size;
   struct Buffer *buf;

   (void)heap;

   if (c == NULL || c->disable)
      return HEAP_ACTION_CONTINUE;

   /* Is the requested size included in the cached subset? */
   if (unlikely(size >= f->smax))
      return HEAP_ACTION_CONTINUE;

   /* Is the size large enough to fit freelist links without damaging
    * the buffer endmark? The user area has to be at least 8 bytes. */
   if (size < sizeof(struct FreeBuffer) - sizeof(struct Buffer))
   {
      /* No. Increase the size so that the links will fit
       * in the buffer about to be allocated when it is freed. */
      size = event->e_alloc.size =
         sizeof(struct FreeBuffer) - sizeof(struct Buffer);
   }

   size += sizeof(struct Buffer) + f->config->buf_size_endmark;
   
   buf = buf_remove_head(heap, f, c, size, event->e_alloc.owner);
   if (buf != NULL)
   {
      event->e_alloc.ptr = &buf[1];
      return HEAP_ACTION_INTERCEPT;
   }

   /* No buffer found. Request is forwarded to algorithms. */
   return HEAP_ACTION_CONTINUE;
}

/**************************************************************************/

static HeapAction
l_bcache_alloc_res(struct Heap *heap, union HeapEvent *event, void *opaque)
{
   struct BCacheFilter *filter = opaque;
   struct LocalCache *c = cache_get(filter);
   
   if (c == NULL)
      return HEAP_ACTION_CONTINUE;

   /* If an allocation is about to fail, try to
    * flush all caches associated with the heap.
    * If any buffer is successfully garbage collected
    * retry/repeat the allocation attempt. */
   
   if (event->e_alloc.ptr == NULL)
   {
      if (cache_flush_all(heap, filter))
         return HEAP_ACTION_REPEAT;
   }

   return HEAP_ACTION_CONTINUE;
}

/**************************************************************************/

static HeapAction
l_bcache_free(struct Heap *heap, union HeapEvent *event, void *opaque)
{
   struct BCacheFilter *f = opaque;
   struct LocalCache *c = cache_get(f);
   struct Buffer *buf = event->e_free.ptr;
   HeapError ecode;
   size_t size;

   (void)heap;
   
   if (c == NULL || c->disable)
      return HEAP_ACTION_CONTINUE;

   if (unlikely(buf == NULL))
      return HEAP_ACTION_CONTINUE;

   buf = &buf[-1];

   if (buf_id(buf) == 0)
      return HEAP_ACTION_CONTINUE;

   size = buf_size(heap, f, buf);
   if (size < sizeof(struct FreeBuffer) + f->config->buf_size_endmark)
      return HEAP_ACTION_CONTINUE;

   
   ecode = buf_check(f, buf, size, event->e_free.owner);
   if (ecode != HEAP_ERROR_NONE)
   {
      zzheap_error(heap, ecode, &buf[1]);
      return HEAP_ACTION_INTERCEPT;
   }
      
   buf = buf_insert(heap, f, c, buf, size);
   if (buf != NULL)
   {
      event->e_free.ptr = &buf[1];
      return HEAP_ACTION_CONTINUE;
   }
   
   return HEAP_ACTION_INTERCEPT;
}

/**************************************************************************/

static HeapAction
l_bcache_query(struct Heap *heap, union HeapEvent *event, void *opaque)
{
   struct BCacheFilter *f = opaque;
   HeapParam *p = event->e_query.params;

   (void)heap;
   
   if (p == NULL)
      return HEAP_ACTION_CONTINUE;
   
   while (*p != HPM_END)
   {
      switch (HPM_ID(*p))
      {
         case HPM_QUERY_SERVICE:
         {
            if (!HPM_IS_HANDLED(p[0]) && p[1] == 0)
            {
               HPM_SET_HANDLED(p[0]);
               HPM_SET_VALUE(p[1], &f->info);
            }
            else if (p[1] == (HeapParam)&f->info)
            {
               HPM_CLEAR_HANDLED(p[0]);
               HPM_SET_VALUE(p[1], 0);
            }

            break;
         }
      }

      p += (1 + HPM_VALUES(*p));
   }

   return HEAP_ACTION_CONTINUE;   
}

/**************************************************************************/

static HeapAction
l_bcache_control(struct Heap *heap, union HeapEvent *event, void *opaque)
{
   struct BCacheFilter *f = opaque;
   HeapParam *p = event->e_query.params;

   if (p == NULL)
      return HEAP_ACTION_CONTINUE;
   
   while (*p != HPM_END)
   {
      switch (HPM_ID(*p))
      {
         case HPM_CTRL_BCACHE_FLUSH:
         {
            if (cache_flush_all(heap, f))
               HPM_SET_HANDLED(p[0]);
            break;
         }
      }

      p += (1 + HPM_VALUES(*p));
   }

   return HEAP_ACTION_CONTINUE;   
}

/**************************************************************************/

static HeapAction
l_bcache_destroy(struct Heap *heap, union HeapEvent *event, void *opaque)
{
   struct BCacheFilter *f = opaque;
   uint32_t i;

   (void)event;
   
   zzheap_detach_hook(heap, HEAP_EVENT_ALLOC_REQ,   l_bcache_alloc,    f);
   zzheap_detach_hook(heap, HEAP_EVENT_ALLOC_RES,   l_bcache_alloc_res,f);
   zzheap_detach_hook(heap, HEAP_EVENT_FREE_REQ,    l_bcache_free,     f);
   zzheap_detach_hook(heap, HEAP_EVENT_QUERY_REQ,   l_bcache_query,    f);
   zzheap_detach_hook(heap, HEAP_EVENT_CONTROL_REQ, l_bcache_control,  f);

   for (i = 0; i < CACHE_LISTS; i++)
   {
      struct LocalCache *cache = f->caches[i];
      if (cache != NULL)
      {
         slab_free(bcache_slab, cache);
      }
   }

   heap_meta_free(heap, f);

   return HEAP_ACTION_CONTINUE;
}

/**************************************************************************/




/**************************************************************************
 * Public Interface
 **************************************************************************/

HeapStatus
heap_attach_bcache_filter(struct Heap *heap,
                          struct HeapStatistics *stats,
                          struct HeapConfiguration *config,
                          uint32_t nmax, uint32_t smax,
                          HeapBufferType type, HeapPID *daemon)
{
   struct BCacheFilter *f;
   HeapStatus status;

   /* Check default algorithm. LARGE means that no small
    * buffer algorithm has been configured and that the
    * cache shouldn't be installed. */
   if (type == HEAP_BUFFER_LARGE)
      return HEAP_SUCCESS;

   /* Create global slab cache. */
   if (bcache_slab == NULL)
   {
      bcache_slab = create_slab_allocator(
         "heap.bcache", sizeof(struct LocalCache),
         sizeof(void*), 10, NULL, NULL, SLAB_SLEEP);
      if (bcache_slab == NULL)
         return HEAP_EOUT_OF_MEMORY;
   }


   f = heap_meta_alloc(heap, (sizeof(struct BCacheFilter) +
                              sizeof(struct LocalCache*) * CACHE_LISTS));
   if (f == NULL)
      return HEAP_EOUT_OF_MEMORY;

   memset(f, 0, (sizeof(struct BCacheFilter) +
                 sizeof(struct LocalCache *) * CACHE_LISTS));

   f->stats  = stats;
   f->config = config;
   f->daemon = daemon;
   f->nmax   = nmax;
   f->smax   = smax;
   f->type   = type;

   f->info.version = 101;
   strcpy(f->info.id, "com.enea.heap.bcache");
   
   /* Oh, this is ugly, look another way. */
   f->mask = (type == HEAP_BUFFER_FBUDDY) ? 0x00FFFFFEul : 0x01FFFFFEul;

   spin_lock_init(&f->lock);
   
   status = zzheap_attach_hook(
      heap, HEAP_EVENT_ALLOC_REQ, l_bcache_alloc, f, HEAP_PRIO_FILTER_HIGH);
   if (status != HEAP_SUCCESS)
   {
      heap_meta_free(heap, f);
      return status;
   }

   status = zzheap_attach_hook(
      heap, HEAP_EVENT_ALLOC_RES, l_bcache_alloc_res, f, HEAP_PRIO_FILTER_HIGH);
   if (status != HEAP_SUCCESS)
   {
      heap_meta_free(heap, f);
      return status;
   }

   status = zzheap_attach_hook(
      heap, HEAP_EVENT_FREE_REQ, l_bcache_free, f, HEAP_PRIO_FILTER_HIGH);
   if (status != HEAP_SUCCESS)
   {
      heap_meta_free(heap, f);
      return status;
   }

   status = zzheap_attach_hook(
      heap, HEAP_EVENT_QUERY_REQ, l_bcache_query, f, HEAP_PRIO_FILTER_HIGH);
   if (status != HEAP_SUCCESS)
   {
      heap_meta_free(heap, f);
      return status;
   }

   status = zzheap_attach_hook(
      heap, HEAP_EVENT_CONTROL_REQ, l_bcache_control, f, HEAP_PRIO_FILTER_HIGH);
   if (status != HEAP_SUCCESS)
   {
      heap_meta_free(heap, f);
      return status;
   }

   status = zzheap_attach_hook(
      heap, HEAP_EVENT_DESTROY_RES, l_bcache_destroy, f, HEAP_PRIO_FILTER_HIGH);
   if (status != HEAP_SUCCESS)
   {
      heap_meta_free(heap, f);
      return status;
   }

   return HEAP_SUCCESS;
}

/**************************************************************************/

