/* COPYRIGHT-ENEA-SRC-R2 *
**************************************************************************
* Copyright (C) 2010 by Enea Software AB.
* All rights reserved.
*
* This Software is furnished under a software license agreement and
* may be used only in accordance with the terms of such agreement.
* Any other use or reproduction is prohibited. No title to and
* ownership of the Software is hereby transferred.
*
* PROPRIETARY NOTICE
* This Software consists of confidential information.
* Trade secret law and copyright law protect this Software.
* The above notice of copyright on this Software does not indicate
* any actual or intended publication of such Software.
**************************************************************************
* COPYRIGHT-END */


#include "heap_filt_debug.h"
#include <string.h>


/**************************************************************************
 * Plug-in History
 **************************************************************************
 *
 * IMPORTANT
 * 
 * This comment shall be updated with a description of the change each
 * time the extension version number is increased. The version number
 * shall be increased if incompatible modifications are made to internal
 * structures or after behavioral changes. See HeapExtension below.
 * 
 **************************************************************************
 *
 *  Version  Description
 *  -------  ----------------------------------------------------------
 *   100     Initial revision.  
 *
 **************************************************************************/



/*****************************************************************************
 * Helper Macros
 *****************************************************************************/
#define PATTERN  0xC5
#define LISTSIZE 10


/*****************************************************************************
 * Data Structures
 *****************************************************************************/

struct Buffer
{
   struct Buffer *buf;
   const char    *file;
   HeapPID        pid;
   uint32_t       control;

#define BUF_ID_MASK                   0xC0000000ul
#define BUF_ID_SHIFT                  30
#define BUF_SHARED_MASK               0x10000000ul
#define BUF_LINE_MASK                 0x0FFFFFFEul
#define BUF_LINE_SHIFT                1
#define BUF_ID(buf)                   ((buf)->control & BUF_ID_MASK)
#define BUF_LINE(buf)                 (((buf)->control & BUF_LINE_MASK) >> BUF_LINE_SHIFT)
#define BUF_SHARED(buf)               ((buf)->control & BUF_SHARED_MASK)
#define BUF_SET_CONTROL(buf,id,line)  ((buf)->control = (((line) << BUF_LINE_SHIFT) & BUF_LINE_MASK) | (id))
#define BUF_SET_SHARED(buf)           ((buf)->control |= BUF_SHARED_MASK)
};

/*****************************************************************************/

struct Filter
{
   uint32_t   id;
   uint32_t   cfg;

   struct Buffer *list[LISTSIZE];
   uint32_t   pos;

   struct HeapServiceInfo    info;        /* Plug-in information. */
};

/*****************************************************************************/




/*****************************************************************************
 * Helper Functions
 *****************************************************************************/

static inline int
bcompare(void *ptr, char c, size_t size)
{
   char *begin = ptr, *end = begin + size;
   while (begin < end)
   {
      if (*begin++ != c)
         return -1;
   }
   return 0;
}

/*****************************************************************************/

static inline HeapPID
buf_owner(struct Heap *heap, struct Buffer *buf)
{
   HeapParam p[] = {
      HPM(QUERY_BUF_OWNER, 1), 0,
      HPM_END
   };

   p[1] = (HeapParam)buf;
   zzheap_query(heap, p, 0);
    
   return HPM_IS_HANDLED(p[0]) ? (HeapPID)p[1] : 0;
}

/*****************************************************************************/

static inline size_t
buf_size(struct Heap *heap, struct Buffer *buf)
{
   HeapParam p[] = {
      HPM(QUERY_BUF_REQ_SIZE, 1), 0,
      HPM_END
   };

   p[1] = (HeapParam)buf;
   zzheap_query(heap, p, 0);
    
   return HPM_IS_HANDLED(p[0]) ? (size_t)p[1] : 0;
}

/*****************************************************************************/

static inline struct Buffer *
buf_next(struct Heap *heap, HeapParam *p)
{
   p[0] = HPM(QUERY_BUF_NEXT, 4);
   p[5] = HPM_END;
   zzheap_query(heap, p, 0);
    
   return HPM_IS_HANDLED(p[0]) ? (struct Buffer *)p[1] : NULL;
}

/*****************************************************************************/

static inline void
buf_print(struct Buffer *buf)
{
   heap_log("================= LEAK DETECTED ======================================\n");
   heap_log("Process: %x\n", buf->pid);
   heap_log("File   : %s\n", buf->file);
   heap_log("Line   : %d\n", BUF_LINE(buf));
   heap_log("======================================================================\n");
}

/*****************************************************************************/


/*****************************************************************************
 * Private Interface
 *****************************************************************************/

static HeapAction
l_debug_alloc_req(struct Heap *heap, union HeapEvent *event, void *opaque)
{
   /* The allocation request hook increases the requested 
    * size to make room for the annotation data that will 
    * be stored ahead of the user pointer. */

   (void)heap;
   (void)opaque;

   if (event->e_alloc.size)
      event->e_alloc.size += sizeof(struct Buffer);

   return HEAP_ACTION_CONTINUE;
}

/*****************************************************************************/

static HeapAction
l_debug_alloc_res(struct Heap *heap, union HeapEvent *event, void *opaque)
{
   /* The allocation result hook initializes the annotation
    * data and adjusts the user pointer to point at the end
    * of the annotation structure. */

   struct Filter *f = opaque;
   struct Buffer *buf = event->e_alloc.ptr;

   (void)heap;
   
   /* Skip if no buffer was allocated. */
   if (event->e_alloc.ptr == NULL)
      return HEAP_ACTION_CONTINUE;

   /* Initialize annotation structure. */
   BUF_SET_CONTROL(buf, f->id, event->e_alloc.line);
   if (event->e_alloc.owner == 0) BUF_SET_SHARED(buf);
   buf->pid  = current_process();
   buf->file = event->e_alloc.file;
   buf->buf  = event->e_alloc.ptr;

   /* Adjust user pointer. */
   event->e_alloc.ptr = &buf[1];
    
   return HEAP_ACTION_CONTINUE;
}

/*****************************************************************************/

static HeapAction
l_debug_resize_req(struct Heap *heap, union HeapEvent *event, void *opaque)
{
   /* The resize request hook increases the requested 
    * size to make room for the annotation data that will 
    * be stored ahead of the user pointer. 
    *
    * Note that annotation of the resized buffer only is 
    * possible if the buffer already is annotated. A buffer 
    * that was allocated before the debug module was installed
    * cannot become annotated since user data already occupies 
    * the area where the annotation would reside. 
    * 
    * This limitation could be removed by intercepting this 
    * request with a NULL pointer, which would force the CRT 
    * to allocate a new buffer and copy the data. 
    */

   struct Filter *f = opaque;
   struct Buffer *buf = event->e_resize.ptr;

   (void)heap;

   if (buf == NULL)
      return HEAP_ACTION_CONTINUE;

   buf = &buf[-1];

   /* Make sure that the buffer is annotated. */
   if (BUF_ID(buf) != f->id)
      return HEAP_ACTION_CONTINUE;

   /* Make sure that the annotation data is valid. */
   if (buf->buf == event->e_resize.ptr)
      return HEAP_ACTION_CONTINUE;

   /* Increase size and return adjusted pointer. */
   event->e_resize.size += sizeof(struct Buffer);
   event->e_resize.ptr   = buf;
    
   return HEAP_ACTION_CONTINUE;
}

/*****************************************************************************/

static HeapAction
l_debug_resize_res(struct Heap *heap, union HeapEvent *event, void *opaque)
{
   /* Initializes the annotation data if the resize request
    * was successfully carried out. This is essentially
    * a copy of the allocation result event hook. 
    */

   struct Filter *f = opaque;
   struct Buffer *buf = event->e_resize.ptr;

   (void)heap;
   
   if (buf == NULL)
      return HEAP_ACTION_CONTINUE;

   /* Initialize annotation structure. */
   BUF_SET_CONTROL(buf, f->id, event->e_alloc.line);
   if (event->e_alloc.owner == 0) BUF_SET_SHARED(buf);
   buf->pid  = current_process();
   buf->file = event->e_alloc.file;
   buf->buf  = event->e_alloc.ptr;

   /* Adjust user pointer. */
   event->e_alloc.ptr = &buf[1];

   return HEAP_ACTION_CONTINUE;
}

/*****************************************************************************/

static HeapAction
l_debug_free_req(struct Heap *heap, union HeapEvent *event, void *opaque)
{
   /* The deallocation request hook investigates if the 
    * buffer to be freed has been annotated. If so, the 
    * the supplied pointer is adjusted to point at the 
    * real beginning of the buffer.
    * 
    * If the deferred free list has been activated, the
    * buffer is intercepted and inserted into the fixed
    * size circular list. A special pattern is written in
    * the buffer to detect if it has been accessed after 
    * it was freed and while it remains in the list. 
    * 
    * The hook verifies the pattern when the deferred
    * list has become full and a buffer must be evicted
    * to make room for the lasted deallocated buffer. 
    * If the pattern is damaged an error is generated. 
    */

   struct Filter *f = opaque;
   struct Buffer *buf = event->e_free.ptr;

   (void)heap;

   if (buf == NULL)
      return HEAP_ACTION_CONTINUE;

   buf = &buf[-1];

   /* Make sure that the buffer is annotated. */
   if (BUF_ID(buf) != f->id)
      return HEAP_ACTION_CONTINUE;

   if (f->cfg & HEAP_DEBUG_PATTERN)
   {
      memset(buf, PATTERN, buf_size(heap, buf));
   }

   if (f->cfg & HEAP_DEBUG_DEFER)
   {
      struct Buffer *tmp = f->list[f->pos % LISTSIZE];
      f->list[f->pos++ % LISTSIZE] = buf;
      buf = tmp;
	
      if (buf != NULL && 
          f->cfg & HEAP_DEBUG_PATTERN && 
          bcompare(buf, PATTERN, buf_size(heap, buf)) < 0)
      {
         zzheap_error(heap, HEAP_ERROR_BUF_TOUCHED, buf);
      }
   }

   if (buf == NULL)
      return HEAP_ACTION_INTERCEPT;

   event->e_free.ptr   = buf;
   event->e_free.owner = 0;

   return HEAP_ACTION_CONTINUE;
}

/*****************************************************************************/
static HeapAction
l_debug_collect_req(struct Heap *heap, union HeapEvent *event, void *opaque)
{
   /* The garbage collection hook traverses the entire heap
    * in its search for shared buffers that are leaked due
    * to the termination of a process. If such a buffer is
    * found a message is logged to allow the developer to 
    * find the origin of the leak. These reports may be false
    * if the design allows the buffer to be freed by another 
    * process.
    * 
    * Any private buffers that reside in the deferred free 
    * list are deallocated to avoid duplicate calls to free. */

   struct Filter *f = opaque;
   struct Buffer *buf = NULL;
    
   HeapPID pid = event->e_collect.owner;
   HeapParam next[6] = {0,0,0,0,0,0};
   uint32_t ix;

   (void)heap;

   /* Collect buffers in deferred list. */
   for (ix = 0; ix < LISTSIZE; ix++)
   {
      if (f->list[ix] == NULL)
         continue;

      if (buf_owner(heap, f->list[ix]) == pid)
      {
         zzheap_free(heap, f->list[ix], 0, NULL, __FILE__, __LINE__);
         f->list[ix] = NULL;
      }
   }

   /* Find leaked buffers. */
   buf = buf_next(heap, next);
   while (buf != NULL)
   {
      if (BUF_ID(buf) == f->id && BUF_SHARED(buf) && buf->pid == pid)
         buf_print(buf);
      buf = buf_next(heap, next);
   }

   return HEAP_ACTION_CONTINUE;
}

/*****************************************************************************/

static HeapAction
l_debug_query_req(struct Heap *heap, union HeapEvent *event, void *opaque)
{
   struct Filter *f = opaque;
   HeapParam *p = event->e_query.params;

   (void)heap;
   
   if (p == NULL)
       return HEAP_ACTION_CONTINUE;
   
   while (HPM_ID(p[0]) != HPM_END)
   {
      switch (HPM_ID(p[0]))
      {
         case HPM_QUERY_SERVICE:
         {
            if (!HPM_IS_HANDLED(p[0]) && p[1] == 0)
            {
               HPM_SET_HANDLED(p[0]);
               HPM_SET_VALUE(p[1], &f->info);
            }
            else if (p[1] == (HeapParam)&f->info)
            {
               HPM_CLEAR_HANDLED(p[0]);
               HPM_SET_VALUE(p[1], 0);
            }

            break;
         }
         case HPM_QUERY_BUF_REQ_SIZE:
         case HPM_QUERY_BUF_ACT_SIZE:
         case HPM_QUERY_BUF_OWNER:
         case HPM_QUERY_BUF_TYPE:
         case HPM_QUERY_BUF_STATUS:
         case HPM_QUERY_BUF_DATA:
         {
	    struct Buffer *buf = (struct Buffer *)p[1];
	    
	    if (buf == NULL || BUF_ID(&buf[-1]) != f->id)
               break;

	    p[1] = (HeapParam)&buf[-1];

	    break;
         }

         default:
	    break;
      }

      p += HPM_VALUES(p[0]);
      p += 1;
   }

   return HEAP_ACTION_CONTINUE;
}

/*****************************************************************************/

static HeapAction
l_debug_destroy(struct Heap *heap, union HeapEvent *event, void *opaque)
{
   struct Filter *f = opaque;

   (void)event;

   zzheap_detach_hook(heap, HEAP_EVENT_ALLOC_REQ,   l_debug_alloc_req,   f);
   zzheap_detach_hook(heap, HEAP_EVENT_ALLOC_RES,   l_debug_alloc_res,   f);
   zzheap_detach_hook(heap, HEAP_EVENT_RESIZE_REQ,  l_debug_resize_req,  f);
   zzheap_detach_hook(heap, HEAP_EVENT_RESIZE_RES,  l_debug_resize_res,  f);
   zzheap_detach_hook(heap, HEAP_EVENT_FREE_REQ,    l_debug_free_req,    f);
   zzheap_detach_hook(heap, HEAP_EVENT_COLLECT_REQ, l_debug_collect_req, f);
   zzheap_detach_hook(heap, HEAP_EVENT_QUERY_REQ,   l_debug_query_req,   f);

   if (f != NULL)
      heap_meta_free(heap, f);

   return HEAP_ACTION_CONTINUE;
}

/*****************************************************************************/



/*****************************************************************************
 * Public Interface
 *****************************************************************************/

HeapStatus
heap_attach_debug_filter(struct Heap *heap, uint32_t id, uint32_t options)
{
   struct Filter *f;
   HeapStatus status;

   f = heap_meta_alloc(heap, sizeof(struct Filter));
   if (f == NULL)
      return HEAP_EOUT_OF_MEMORY;

   bzero(f, sizeof(struct Filter));

   f->id  = id << BUF_ID_SHIFT;
   f->cfg = options & (HEAP_DEBUG_PATTERN|HEAP_DEBUG_DEFER);

   f->info.version = 100;
   strcpy(f->info.id, "com.enea.heap.debug");

   status = zzheap_attach_hook(
      heap, HEAP_EVENT_ALLOC_REQ, l_debug_alloc_req, f, HEAP_PRIO_FILTER_HIGH);
   if (status != HEAP_SUCCESS)
   {
      heap_meta_free(heap, f);
      return status;
   }

   status = zzheap_attach_hook(
      heap, HEAP_EVENT_ALLOC_RES, l_debug_alloc_res, f, HEAP_PRIO_FILTER_HIGH);
   if (status != HEAP_SUCCESS)
   {
      heap_meta_free(heap, f);
      return status;
   }

   status = zzheap_attach_hook(
      heap, HEAP_EVENT_RESIZE_REQ, l_debug_resize_req, f, HEAP_PRIO_FILTER_HIGH);
   if (status != HEAP_SUCCESS)
   {
      heap_meta_free(heap, f);
      return status;
   }

   status = zzheap_attach_hook(
      heap, HEAP_EVENT_RESIZE_RES, l_debug_resize_res, f, HEAP_PRIO_FILTER_HIGH);
   if (status != HEAP_SUCCESS)
   {
      heap_meta_free(heap, f);
      return status;
   }

   status = zzheap_attach_hook(
      heap, HEAP_EVENT_FREE_REQ, l_debug_free_req, f, HEAP_PRIO_FILTER_HIGH);
   if (status != HEAP_SUCCESS)
   {
      heap_meta_free(heap, f);
      return status;
   }

   status = zzheap_attach_hook(
      heap, HEAP_EVENT_COLLECT_REQ, l_debug_collect_req, f, HEAP_PRIO_FILTER_HIGH);
   if (status != HEAP_SUCCESS)
   {
      heap_meta_free(heap, f);
      return status;
   }

   status = zzheap_attach_hook(
      heap, HEAP_EVENT_QUERY_REQ, l_debug_query_req, f, HEAP_PRIO_FILTER_HIGH);
   if (status != HEAP_SUCCESS)
   {
      heap_meta_free(heap, f);
      return status;
   }

   status = zzheap_attach_hook(
      heap, HEAP_EVENT_DESTROY_RES, l_debug_destroy, f, HEAP_PRIO_FILTER_HIGH);
   if (status != HEAP_SUCCESS)
   {
      heap_meta_free(heap, f);
      return status;
   }

   return HEAP_SUCCESS;
}

/*****************************************************************************/


