/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2013 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

#include "heap_filt_eventlog.h"
#include "ose_heapd.h"
#include "heap_error.h"


/**************************************************************************
 * Macros and Definitions
 **************************************************************************/

#define LOG_FLAG_ERROR    1U
#define LOG_FLAG_ALLOC    2U
#define LOG_FLAG_FREE     4U
#define LOG_FLAG_RESIZE   8U
#define LOG_FLAG_COLLECT 16U

#define LOG_PFX  "heap: "
#define LOG_SZ   128U
#ifdef OSMIPS
#define PAGE_ALIGN 0x2000UL
#else
#define PAGE_ALIGN 0x1000UL
#endif
#define ALIGN(size) (((size) + PAGE_ALIGN - 1U) & ~(PAGE_ALIGN-1U))


/**************************************************************************
 * Local functions
 **************************************************************************/

static void
buf_size(struct Heap *heap, unsigned long ptr, size_t *req_size, size_t *act_size)
{
   HeapParam p[] = {
      HPM(QUERY_BUF_ACT_SIZE,1), 0,
      HPM(QUERY_BUF_REQ_SIZE,1), 0,
      HPM_END
   };

   p[1] = p[3] = (HeapParam)ptr;

   zzheap_query(heap, p, 1);

   *act_size = HPM_IS_HANDLED(p[0]) ? p[1] : 0;
   *req_size = HPM_IS_HANDLED(p[2]) ? p[3] : 0;
}

/***************************************************************************/

static void
ramlog_dump(const char *info, uintptr_t ptr, size_t size)
{
   uint32_t data[4];
   uintptr_t addr;

   ptr -= (size>>1);

   /* Align to multiples of 4 words. */
   size +=  (ptr & (sizeof(data) - 1));
   size +=  (sizeof(data) - 1);
   size &= ~(sizeof(data) - 1);
   ptr  &= ~(sizeof(data) - 1);

   ramlog_printf(LOG_PFX "%s:\n", info);

   for (addr=ptr; addr<ptr+size; addr += sizeof(data))
   {
      int rv = get_mem(current_process(), addr, data, sizeof(data));
      if (0 == rv)
      {
         ramlog_printf(LOG_PFX "%0*tx: %08x %08x %08x %08x\n",
                       sizeof(addr)*2, addr, data[0], data[1], data[2], data[3]);
      }
      else
      {
         ramlog_printf(LOG_PFX "%0*tx:  pgfault  pgfault  pgfault  pgfault\n",
                       sizeof(addr)*2, addr);
      }
   }
}

/**************************************************************************/

static void
ramlog_info(uintptr_t ptr, size_t req_size, size_t act_size)
{
   ramlog_printf(LOG_PFX "buffer <0x%tx, %zu, %zu, 0x%tx>\n",
                 ptr, req_size, act_size, ptr+req_size);
}

/**************************************************************************/



/***************************************************************************
 * Private Interface
 ***************************************************************************/

static HeapAction
l_log_error(struct Heap *heap, union HeapEvent *event, void *opaque)
{
   struct OseHeap *ose_heap = OSHEAP_FROM_HEAP(heap);

   HeapError ecode = event->e_error.ecode & ~HEAP_ERROR_FATAL;
   unsigned long data = (unsigned long)event->e_error.data;
   size_t req_size, act_size;
   OSADDRESS locked = 0;
   HeapLock lock;

   (void)opaque;

   ramlog_printf(LOG_PFX "detected error %u, h_ref: 0x%x, h_owner: 0x%x\n",
                 ecode, HREF_FROM_HEAP(ose_heap), ose_heap->owner);

   LOCK_SAVE(lock);
   LOCK_RESTORE(lock);

   if (heap_lock_owner(heap) == (current_process() & ~0x80000000))
   {
      locked = 1;

      /* Unlock the heap unconditionally
       * when an error is detected. */
      heap_unlock(heap, &lock, 0);
   }

   switch (ecode)
   {
      case HEAP_ERROR_BUF_HEADER:
      case HEAP_ERROR_BUF_LINKS:
      {
         buf_size(heap, data, &req_size, &act_size);
         ramlog_info(data, req_size, act_size);
         ramlog_dump("buffer header", data, LOG_SZ);
         break;
      }
      case HEAP_ERROR_BUF_TOUCHED:
      {
         buf_size(heap, data, &req_size, &act_size);
         ramlog_info(data, req_size, act_size);
         ramlog_dump("buffer header", data, LOG_SZ);
         ramlog_dump("buffer tail", data+act_size, LOG_SZ);
         break;
      }
      case HEAP_ERROR_BUF_ENDMARK:
      {
         buf_size(heap, data, &req_size, &act_size);
         ramlog_info(data, req_size, act_size);
         ramlog_dump("buffer header", data, LOG_SZ);
         ramlog_dump("buffer endmark", data+req_size, LOG_SZ);
         ramlog_dump("buffer tail", data+act_size, LOG_SZ);
         break;
      }
      default:
      {
         break;
      }
   }

   if (locked)
   {
      heap_lock(heap, &lock, 0);
   }

   return HEAP_ACTION_CONTINUE;
}

/**************************************************************************/

static HeapAction
l_log_event(struct Heap *heap, union HeapEvent *event, void *opaque)
{
   PROCESS curp = current_process();
   
   (void)heap;
   (void)opaque;

   if (event == NULL)
      return HEAP_ACTION_CONTINUE;

   switch (event->id)
   {
      case HEAP_EVENT_ALLOC_REQ:
      {
         struct HeapAllocEvent *e = &event->e_alloc;
         ramlog_printf(LOG_PFX "alloc req %zdb @ %x (%s)\n",
                       e->size,
                       curp,
                       get_pid_info(curp, 14));
         if (e->file)
            ramlog_printf(LOG_PFX "%s:%d\n", e->file, e->line);
         break;
      }
      case HEAP_EVENT_ALLOC_RES:
      {
         struct HeapAllocEvent *e = &event->e_alloc;
         ramlog_printf(LOG_PFX "alloc res 0x%p:%zdb @ %x (%s)\n",
                       e->ptr,
                       e->size,
                       curp,
                       get_pid_info(curp, 14));
         if (e->file)
            ramlog_printf(LOG_PFX "%s:%d\n", e->file, e->line);
         break;
      }
      case HEAP_EVENT_FREE_RES:
      {
         struct HeapFreeEvent *e = &event->e_free;
         ramlog_printf(LOG_PFX "free res 0x%p:%zdb @ %x (%s)\n",
                       e->ptr,
                       e->size,
                       curp,
                       get_pid_info(curp, 14));
         if (e->file)
            ramlog_printf(LOG_PFX "%s:%d\n", e->file, e->line);
         break;
      }
      case HEAP_EVENT_RESIZE_REQ:
      {
         struct HeapResizeEvent *e = &event->e_resize;
         ramlog_printf(LOG_PFX "resize req 0x%p:%zdb @ %x (%s)\n",
                       e->ptr,
                       e->size,
                       curp,
                       get_pid_info(curp, 14));
         if (e->file)
            ramlog_printf(LOG_PFX "%s:%d\n", e->file, e->line);
         break;
      }
      case HEAP_EVENT_RESIZE_RES:
      {
         struct HeapResizeEvent *e = &event->e_resize;
         ramlog_printf(LOG_PFX "resize res 0x%p:%zd->%zdb @ %x (%s)\n",
                       e->ptr,
                       e->size_prev,
                       e->size,
                       curp,
                       get_pid_info(curp, 14));
         if (e->file)
            ramlog_printf(LOG_PFX "%s:%d\n", e->file, e->line);
         break;
      }
      case HEAP_EVENT_COLLECT_REQ:
      {
         struct HeapCollectEvent *e = &event->e_collect;
         ramlog_printf("heap: collect req %x\n", e->owner);
         break;
      }
      case HEAP_EVENT_COLLECT_RES:
      {
         struct HeapCollectEvent *e = &event->e_collect;
         ramlog_printf("heap: collect res %x\n", e->owner);
         break;
      }
      default:
      {
         break;
      }
   }

   return HEAP_ACTION_CONTINUE;
}


/***************************************************************************
 * Public Interface
 ***************************************************************************/

HeapStatus
heap_attach_event_log(struct Heap *heap, uint32_t eflags)
{
   HeapStatus status;

   if (eflags & LOG_FLAG_ALLOC)
   {
      status = zzheap_attach_hook(
         heap, HEAP_EVENT_ALLOC_REQ, l_log_event, NULL, HEAP_PRIO_FILTER_HIGH);
      if (status != HEAP_SUCCESS)
         return status;
      status = zzheap_attach_hook(
         heap, HEAP_EVENT_ALLOC_RES, l_log_event, NULL, HEAP_PRIO_FILTER_HIGH);
      if (status != HEAP_SUCCESS)
         return status;
   }

   if (eflags & LOG_FLAG_FREE)
   {
      status = zzheap_attach_hook(
         heap, HEAP_EVENT_FREE_RES, l_log_event, NULL, HEAP_PRIO_FILTER_HIGH);
      if (status != HEAP_SUCCESS)
         return status;
   }

   if (eflags & LOG_FLAG_RESIZE)
   {
      status = zzheap_attach_hook(
         heap, HEAP_EVENT_RESIZE_REQ, l_log_event, NULL, HEAP_PRIO_FILTER_HIGH);
      if (status != HEAP_SUCCESS)
         return status;
      status = zzheap_attach_hook(
         heap, HEAP_EVENT_RESIZE_RES, l_log_event, NULL, HEAP_PRIO_FILTER_HIGH);
      if (status != HEAP_SUCCESS)
         return status;
   }

   if (eflags & LOG_FLAG_COLLECT)
   {
      status = zzheap_attach_hook(
         heap, HEAP_EVENT_COLLECT_REQ, l_log_event, NULL, HEAP_PRIO_FILTER_HIGH);
      if (status != HEAP_SUCCESS)
         return status;
      status = zzheap_attach_hook(
         heap, HEAP_EVENT_COLLECT_RES, l_log_event, NULL, HEAP_PRIO_FILTER_HIGH);
      if (status != HEAP_SUCCESS)
         return status;
   }

   if (eflags & LOG_FLAG_ERROR)
   {
      status = zzheap_attach_hook(
         heap, HEAP_EVENT_ERROR_RES, l_log_error, NULL, HEAP_PRIO_FILTER_HIGH);
      if (status != HEAP_SUCCESS)
         return status;
   }

#if 0
   status = zzheap_attach_hook(
      heap, HEAP_EVENT_FREE_REQ, l_log_event, NULL, HEAP_PRIO_FILTER_HIGH);
   if (status != HEAP_SUCCESS)
      return status;
   status = zzheap_attach_hook(
      heap, HEAP_EVENT_COLLECT_RES, l_log_event, NULL, HEAP_PRIO_FILTER_HIGH);
   if (status != HEAP_SUCCESS)
      return status;
   status = zzheap_attach_hook(
      heap, HEAP_EVENT_CONTROL_REQ, l_log_event, NULL, HEAP_PRIO_FILTER_HIGH);
   if (status != HEAP_SUCCESS)
      return status;
   status = zzheap_attach_hook(
      heap, HEAP_EVENT_CONTROL_RES, l_log_event, NULL, HEAP_PRIO_FILTER_HIGH);
   if (status != HEAP_SUCCESS)
      return status;
   status = zzheap_attach_hook(
      heap, HEAP_EVENT_QUERY_REQ, l_log_event, NULL, HEAP_PRIO_FILTER_HIGH);
   if (status != HEAP_SUCCESS)
      return status;
   status = zzheap_attach_hook(
      heap, HEAP_EVENT_QUERY_RES, l_log_event, NULL, HEAP_PRIO_FILTER_HIGH);
   if (status != HEAP_SUCCESS)
      return status;
   status = zzheap_attach_hook(
      heap, HEAP_EVENT_BREAK_REQ, l_log_event, NULL, HEAP_PRIO_FILTER_HIGH);
   if (status != HEAP_SUCCESS)
      return status;
   status = zzheap_attach_hook(
      heap, HEAP_EVENT_BREAK_RES, l_log_event, NULL, HEAP_PRIO_FILTER_HIGH);
   if (status != HEAP_SUCCESS)
      return status;
   status = zzheap_attach_hook(
      heap, HEAP_EVENT_UNBREAK_REQ, l_log_event, NULL, HEAP_PRIO_FILTER_HIGH);
   if (status != HEAP_SUCCESS)
      return status;
   status = zzheap_attach_hook(
      heap, HEAP_EVENT_UNBREAK_RES, l_log_event, NULL, HEAP_PRIO_FILTER_HIGH);
   if (status != HEAP_SUCCESS)
      return status;
   status = zzheap_attach_hook(
      heap, HEAP_EVENT_DESTROY_RES, l_log_event, NULL, HEAP_PRIO_FILTER_HIGH);
   if (status != HEAP_SUCCESS)
      return status;
   status = zzheap_attach_hook(
      heap, HEAP_EVENT_ERROR_RES, l_log_event, NULL, HEAP_PRIO_FILTER_HIGH);
   if (status != HEAP_SUCCESS)
      return status;
#endif

   return HEAP_SUCCESS;
}

