/* COPYRIGHT-ENEA-SRC-R2 *
**************************************************************************
* Copyright (C) 2010 by Enea Software AB.
* All rights reserved.
*
* This Software is furnished under a software license agreement and
* may be used only in accordance with the terms of such agreement.
* Any other use or reproduction is prohibited. No title to and
* ownership of the Software is hereby transferred.
*
* PROPRIETARY NOTICE
* This Software consists of confidential information.
* Trade secret law and copyright law protect this Software.
* The above notice of copyright on this Software does not indicate
* any actual or intended publication of such Software.
**************************************************************************
* COPYRIGHT-END */


#include "heap_filt_fcache.h"
#include "heap_rbtree.h"
#include "string.h"
#include "mm.h"

/**************************************************************************
 * Plug-in History
 **************************************************************************
 *
 * IMPORTANT
 * 
 * This comment shall be updated with a description of the change each
 * time the extension version number is increased. The version number
 * shall be increased if incompatible modifications are made to internal
 * structures or after behavioral changes. See HeapExtension below.
 * 
 **************************************************************************
 *
 *  Version  Description
 *  -------  ----------------------------------------------------------
 *   101     Renamed struct Filter to struct FCacheFilter.
 *   100     Initial revision.  
 *
 **************************************************************************/


/*****************************************************************************
 * Helper Macros
 *****************************************************************************/

#define FRAG_NEXT_FRAG_MASK        0xFFFFFFFCul
#define FRAG_NEXT_ID_MASK          0x00000003ul
#define FRAG_NEXT_FRAG(frag)       ((struct Fragment*)((uint32_t)(frag) & FRAG_NEXT_FRAG_MASK))
#define FRAG_NEXT_ID(frag)         (((uint32_t)(frag) & FRAG_NEXT_ID_MASK) << FRAG_ID_SHIFT)
#define FRAG_NEXT_MASK(frag, id)   ((struct Fragment*)((uint32_t)(frag)|((id) >> FRAG_ID_SHIFT) ))


/*****************************************************************************
 * Data Structures
 *****************************************************************************/

struct Fragment
{
   rbtree_node_t node;
   uint32_t      chksum;
   size_t        size;
   uint32_t      control;

#define FRAG_ID_MASK              0x30000000ul
#define FRAG_ID_SHIFT             28
#define FRAG_ALGO_ID_MASK         0xC0000000ul
#define FRAG_ALGO_ID_SHIFT        30
#define FRAG_EXTERNAL_MASK        0x08000000ul
#define FRAG_FBIT_MASK            0x04000000ul
#define frag_id(frag)             ((frag)->control & FRAG_ID_MASK)
#define frag_algo_id(frag)        ((frag)->control & FRAG_ALGO_ID_MASK)
#define frag_set_algo_id(frag,id) ((frag)->control = ((frag)->control & ~FRAG_ALGO_ID_MASK)|(id))
#define frag_set_external(f)      ((f)->control |= FRAG_EXTERNAL_MASK)
#define frag_external(f)          ((f)->control  & FRAG_EXTERNAL_MASK)
#define frag_set_fbit(f)          ((f)->control |= FRAG_FBIT_MASK)
#define frag_clear_fbit(f)        ((f)->control &= ~FRAG_FBIT_MASK)
#define frag_fbit(f)              ((f)->control  & FRAG_FBIT_MASK)
};

/***************************************************************************/

struct FragmentCursor
{
   struct Fragment *frag_out;
   size_t           size_out;
   struct Fragment *frag;
   uint32_t         reserved0;
};

/*****************************************************************************/

struct FCacheFilter
{
   struct HeapStatistics  *stats;
   struct HeapServiceInfo  info;        /* Plug-in information. */

   uint32_t  id;
   uint32_t  frag_count;
   rbtree_t  frag_tree;
};

/*****************************************************************************/




/*****************************************************************************
 * Helper Functions
 *****************************************************************************/

static always_inline uint32_t
frag_checksum(struct Fragment *frag)
{
#define FRAG_CHECKSUM_MAGIC 0x1234beef
   return (
      (uint32_t)frag ^
      (uint32_t)frag->control ^
      (uint32_t)frag->size ^
      FRAG_CHECKSUM_MAGIC);
#undef FRAG_CHECKSUM_MAGIC
}

/***************************************************************************/

static always_inline void
frag_set_checksum(struct Fragment *frag)
{
   frag->chksum = frag_checksum(frag);
}

/***************************************************************************/

static always_inline uint32_t
frag_get_checksum(struct Fragment *frag)
{
   return frag->chksum;
}

/*****************************************************************************/

static always_inline void
frag_init(struct FCacheFilter *filt, struct Fragment *frag,
          size_t size, uint32_t algo_id)
{
   frag->control = filt->id | algo_id;
   frag->size    = size;
   frag_set_checksum(frag);
}

/*****************************************************************************/

static always_inline void
frag_insert(struct FCacheFilter *filt, struct Fragment *frag)
{
   rbtree_insert(&filt->frag_tree, &frag->node);
   filt->frag_count++;
   filt->stats->frag_size_free += frag->size;
   filt->stats->buf_size_free  += frag->size;

   frag_set_fbit(frag);
   frag_set_checksum(frag);
}

/*****************************************************************************/

static always_inline void
frag_remove(struct FCacheFilter *filt, struct Fragment *frag)
{
   rbtree_remove(&filt->frag_tree, &frag->node);
   filt->frag_count--;
   filt->stats->frag_size_free -= frag->size;
   filt->stats->buf_size_free  -= frag->size;

   frag_clear_fbit(frag);
   frag_set_checksum(frag);
}

/*****************************************************************************/

static int
frag_compare(const void *_l, const void *_r)
{
   const struct Fragment *l = (const struct Fragment *)_l;
   const struct Fragment *r = (const struct Fragment *)_r;
   
   if (l->size < r->size)
      return -1;
   if (l->size > r->size)
      return  1;

   if ((uint32_t)l < (uint32_t)r)
      return -1;
   if ((uint32_t)l > (uint32_t)r)
      return  1;
   
   return 0;
}

/*****************************************************************************/

static int
frag_find_best(const void *_s, const void *_f, const void *_b)
{
   const struct Fragment *f = (const struct Fragment *)_f;
   const struct Fragment *b = (const struct Fragment *)_b;
   size_t s = (size_t)_s;

   (void)b;
   
   if (f->size < s)
      return  1;

   if (f->size > s)
      return -2;

   return 0;
}

/*****************************************************************************/

static always_inline HeapError
frag_check_header(struct Heap *heap, struct FCacheFilter *a,
                  struct Fragment *frag)
{
   (void)heap;

   if (((uint32_t)frag) & (HEAP_ALIGNMENT-1))
      return HEAP_ERROR_FRAG_UNALIGNED;

   if (&frag[1] == NULL || frag_id(frag) != a->id)
      return HEAP_ERROR_FRAG_UNKNOWN;

   if (frag_checksum(frag) != frag_get_checksum(frag))
      return HEAP_ERROR_FRAG_HEADER;

   return HEAP_ERROR_NONE;
}

/*****************************************************************************/

static always_inline HeapError
frag_check(struct Heap *heap, struct FCacheFilter *a, struct Fragment *frag)
{
   HeapError ecode;

#if 0 /* FIXME */  
   ecode = frag_check_exists(heap, a, frag);
   if (ecode != HEAP_ERROR_NONE)
      return ecode;
#endif 

   ecode = frag_check_header(heap, a, frag);
   if (ecode != HEAP_ERROR_NONE)
      return ecode;

   return HEAP_ERROR_NONE;
}

/*****************************************************************************/

static inline struct Fragment *
frag_next_free(struct Heap *heap, struct FCacheFilter *filt,
               struct FragmentCursor *cursor)
{
   struct Fragment *frag = FRAG_NEXT_FRAG(cursor->frag);
   HeapError ecode;

   /* First fragment? */
   if (frag == NULL)
   {
      frag = (struct Fragment *)rbtree_first(&filt->frag_tree);

      if (frag == NULL)
         return NULL;

      if (!frag_external(frag))
      {
         cursor->frag = FRAG_NEXT_MASK(frag, filt->id);
         cursor->frag_out = &frag[1];
         cursor->size_out = frag->size;
         
         return frag;
      }
   }

   ecode = frag_check(heap, filt, frag);
   if (ecode != HEAP_ERROR_NONE)
      return NULL;

   do {
      frag = (struct Fragment *)rbtree_next(&filt->frag_tree, &frag->node);
      if (frag != NULL && !frag_external(frag))
      {   
         cursor->frag = FRAG_NEXT_MASK(frag, filt->id);
         cursor->frag_out = &frag[1];
         cursor->size_out = frag->size;
         break;
      }
   } while (frag != NULL);

   return frag;
}

/***************************************************************************/



/*****************************************************************************
 * Private Interface
 *****************************************************************************/

static HeapAction
l_fcache_break_req(struct Heap *heap, union HeapEvent *event, void *opaque)
{
   struct FCacheFilter *filt = opaque;
   struct Fragment *frag;
   HeapLock lock;

   size_t min_size = event->e_break.min_size;
   size_t max_size = event->e_break.max_size;

   heap_lock(heap, &lock, 0);

   frag = (struct Fragment *)rbtree_find_best(
      &filt->frag_tree, frag_find_best, (void *)min_size);

   if (frag == NULL ||
       (frag->size > max_size && max_size != 0))
   {
      /* No suitable fragment found in the cache.
       * Request must be forwarded to allocator. */
      heap_unlock(heap, &lock, 0);

      /* Add room for fragment header. */
      event->e_break.min_size += sizeof(struct Fragment);
      event->e_break.req_size += sizeof(struct Fragment);
      if (event->e_break.max_size != 0)
         event->e_break.max_size += sizeof(struct Fragment);

      return HEAP_ACTION_CONTINUE;
   }

   frag_remove(filt, frag);

   /* Always set owner algorithm in cache fragment descriptor. */
   frag_set_algo_id(&frag[ 0], event->e_break.algo_id);

   /* In case of fragment originated from MM
    * set owner algorithm also in above allocator's fragment descriptor.
    * (external fragments does above fragment descriptor OSECR-9503) */
   if (!frag_external(frag))
   {
      /* FIXME.
         An open question is if writing/knowledge in the
         above fragment descriptor is violating the layered architecture? 
      */ 
      frag_set_algo_id(&frag[-1], event->e_break.algo_id);
   }
   
   heap_unlock(heap, &lock, 0);

   /* FIXME: Statistics. */

   event->e_break.base = &frag[1];
   event->e_break.min_size = frag->size;
   
   return HEAP_ACTION_INTERCEPT;
}

/*****************************************************************************/

static HeapAction
l_fcache_break_res(struct Heap *heap, union HeapEvent *event, void *opaque)
{
   struct FCacheFilter *filt = opaque;
   struct Fragment *frag = (struct Fragment *)event->e_break.base;
   HeapLock lock;

   if (frag != NULL && frag_id(&frag[-1]) == filt->id)
   {
      /* A fragment was allocated from the cache. */
      return HEAP_ACTION_CONTINUE;
   }

   event->e_break.min_size -= sizeof(struct Fragment);
   event->e_break.req_size -= sizeof(struct Fragment);
   if (event->e_break.max_size > sizeof(struct Fragment))
      event->e_break.max_size -= sizeof(struct Fragment);

   if (frag == NULL)
   {
      /* Allocation failed. Possible explanations:
       *   1) MAX_SIZE has been reached.
       *   2) Cached fragments are too small.
       *   3) System memory is exhausted.
       *
       * All problems could be resolved by flushing
       * the cache if there are free fragments available.
       ******************************************/

      heap_lock(heap, &lock, 0);

      /* Fetch first/smallest fragment. Large
       * fragments are considered more valuable
       * from a fragmentation and performance
       * perspective. */

      frag = (struct Fragment *)rbtree_first(&filt->frag_tree);
      while (frag != NULL)
      {
         /* Cannot remove externally supplied fragments. */
         if (!frag_external(frag))
         {
            frag_remove(filt, frag);
            break;
         }

         /* Try next larger fragment. */
         frag = (struct Fragment *)rbtree_next(&filt->frag_tree, &frag->node);
      }

      heap_unlock(heap, &lock, 0);

      if (frag == NULL)
      {
         /* There are no cached fragments that can
          * be released. The client will have to solve
          * the problem on its own by checking the updated
          * max_size field and requesting less memory. */
         return HEAP_ACTION_CONTINUE;
      }

      zzheap_unbreak(heap, frag, event->e_break.params);

      /* A fragment was found in the cache and has been
       * unbroken. However, the heap may still be near the
       * critical point where no memory can be allocated.
       * The maximum size of the fragment is therefore
       * hardwired to the minimum amount the client requires.
       * This should be ok, but could potentially lead to the
       * dismissal of a small enough cached fragment that will
       * appear in the near future before this request has
       * been completed. (if the fragment is smaller than the
       * original max_size but larger than min_size). */
      if (event->e_break.max_size < event->e_break.min_size)
         event->e_break.max_size = event->e_break.min_size;

      return HEAP_ACTION_REPEAT;
   }


   /* Fragment is newly allocated from external source.
    * Initialize the fragment. */
   frag_init(filt, frag,
             event->e_break.min_size,
             event->e_break.algo_id);

   /* Adjust fragment pointer. */
   event->e_break.base = &frag[1];

   return HEAP_ACTION_CONTINUE;
}

/*****************************************************************************/

static HeapAction
l_fcache_unbreak(struct Heap *heap, union HeapEvent *event, void *opaque)
{
   struct FCacheFilter *filt = opaque;
   struct Fragment *frag = (struct Fragment *)event->e_unbreak.base;
   HeapLock lock;

   (void)heap;
   
   if (frag == NULL)
      return HEAP_ACTION_CONTINUE;

   frag = &frag[-1];
   
   if (frag_id(frag) != filt->id)
      return HEAP_ACTION_CONTINUE;

   heap_lock(heap, &lock, 0);
   frag_insert(filt, frag);
   heap_unlock(heap, &lock, 0);
   
   return HEAP_ACTION_INTERCEPT;
}

/*****************************************************************************/

static HeapAction
l_fcache_query_req(struct Heap *heap, union HeapEvent *event, void *opaque)
{
   struct FCacheFilter *filt = opaque;
   HeapParam *p = event->e_query.params;
   HeapLock lock;

   (void)filt;
   
   if (p == NULL)
      return HEAP_ACTION_CONTINUE;
   
   while (HPM_ID(p[0]) != HPM_END)
   {
      switch (HPM_ID(p[0]))
      {
         case HPM_QUERY_SERVICE:
         {
            if (!HPM_IS_HANDLED(p[0]) && p[1] == 0)
            {
               HPM_SET_HANDLED(p[0]);
               HPM_SET_VALUE(p[1], &filt->info);
            }
            else if (p[1] == (HeapParam)&filt->info)
            {
               HPM_CLEAR_HANDLED(p[0]);
               HPM_SET_VALUE(p[1], 0);
            }

            break;
         }
         case HPM_QUERY_FRAG_TOT_FREE_SIZE:
         {
            if (!HPM_IS_HANDLED(p[0]))
            {
               HPM_SET_HANDLED(p[0]);
               HPM_SET_VALUE(p[1], filt->stats->frag_size_free);
            }
            break;
         }
         case HPM_QUERY_BUF_TOT_FREE_SIZE:
         {
            if (!HPM_IS_HANDLED(p[0]))
            {
               HPM_SET_HANDLED(p[0]);
               HPM_SET_VALUE(p[1], filt->stats->buf_size_free);
            }
            break;
         }
         case HPM_QUERY_FRAG_NEXT_SIZE:
         {
            if (!HPM_IS_HANDLED(p[0]))
            {
               /* Add fragment header. */
               HPM_ADD_VALUE(p[1], sizeof(struct Fragment));
            }
            break;
         }
         case HPM_QUERY_FRAG_STATUS:
         {
            struct Fragment *frag = &((struct Fragment *)p[1])[-1];
            HeapError ecode;

            if (!HPM_IS_HANDLED(p[0]))
            {
               ecode = frag_check(heap, filt, frag);
               if (ecode != HEAP_ERROR_NONE)
                  break;

               /* Remove fragment header. */
               HPM_SUB_VALUE(p[1], sizeof(struct Fragment));
            }
            break;
         }
         case HPM_QUERY_FRAG_SIZE:
         {
            struct Fragment *frag = &((struct Fragment *)p[1])[-1];
            HeapError ecode;

            if (!HPM_IS_HANDLED(p[0]))
            {
               ecode = frag_check(heap, filt, frag);
               if (ecode != HEAP_ERROR_NONE)
                  break;

               /* Remove fragment header. */
               HPM_SUB_VALUE(p[1], sizeof(struct Fragment));
            }
            break;
         }
         case HPM_QUERY_FRAG_TYPE:
         {
            struct Fragment *frag = &((struct Fragment *)p[1])[-1];
            HeapError ecode;

            if (!HPM_IS_HANDLED(p[0]))
            {
               ecode = frag_check(heap, filt, frag);
               if (ecode != HEAP_ERROR_NONE)
                  break;

               /* Remove fragment header. */
               HPM_SUB_VALUE(p[1], sizeof(struct Fragment));
            }
            break;
         }
         case HPM_QUERY_FRAG_LMEM:
         {
            struct Fragment *frag = &((struct Fragment *)p[1])[-1];
            HeapError ecode;

            if (!HPM_IS_HANDLED(p[0]))
            {
               ecode = frag_check(heap, filt, frag);
               if (ecode != HEAP_ERROR_NONE)
                  break;

               /* Remove fragment header. */
               HPM_SUB_VALUE(p[1], sizeof(struct Fragment));
            }
            break;
         }
         case HPM_QUERY_FRAG_PMEM:
         {
            struct Fragment *frag = &((struct Fragment *)p[1])[-1];
            HeapError ecode;

            if (!HPM_IS_HANDLED(p[0]))
            {
               ecode = frag_check(heap, filt, frag);
               if (ecode != HEAP_ERROR_NONE)
                  break;

               /* Remove fragment header. */
               HPM_SUB_VALUE(p[1], sizeof(struct Fragment));
            }
            break;
         }
         case HPM_QUERY_FRAG_NEXT_FREE:
         {
            struct FragmentCursor *cursor = (struct FragmentCursor *)&p[1];

            /* Skip if the request doesn't concern us. */
            if (cursor->frag != NULL && FRAG_NEXT_ID(cursor->frag) != filt->id)
               break;

            if (!HPM_IS_HANDLED(p[0]))
            {
               struct Fragment *frag;

               heap_lock(heap, &lock, 0);
               
               frag = frag_next_free(heap, filt, cursor);
               if (frag == NULL)
               {
                  /* Initialize query for next algorithm. */
                  cursor->frag     = NULL;
                  cursor->frag_out = NULL;
                  cursor->size_out = 0;
                  heap_unlock(heap, &lock, 0);
                  break;
               }

               HPM_SET_HANDLED(p[0]);

               heap_unlock(heap, &lock, 0);
            }
            break;
         }
         default:
	    break;
      }

      p += HPM_VALUES(p[0]);
      p += 1;
   }

   return HEAP_ACTION_CONTINUE;
}

/*****************************************************************************/

static HeapAction
l_fcache_query_res(struct Heap *heap, union HeapEvent *event, void *opaque)
{
   struct FCacheFilter *filt = opaque;
   HeapParam *p = event->e_query.params;

   (void)filt;
   
   if (p == NULL)
       return HEAP_ACTION_CONTINUE;
   
   while (HPM_ID(p[0]) != HPM_END)
   {
      switch (HPM_ID(p[0]))
      {
         case HPM_QUERY_FRAG_NEXT_SIZE:
         {
            if (HPM_IS_HANDLED(p[0]))
            {
               /* Remove fragment header. */
               HPM_SUB_VALUE(p[1], sizeof(struct Fragment));
            }
            break;
         }
         case HPM_QUERY_FRAG_PREV_SIZE:
         {
            if (HPM_IS_HANDLED(p[0]))
            {
               /* Remove fragment header. */
               if (p[1] > sizeof(struct Fragment))
                  HPM_SUB_VALUE(p[1], sizeof(struct Fragment));
               else
                  HPM_SET_VALUE(p[1], 0);
            }
            break;
         }
         case HPM_QUERY_FRAG_NEXT:
         {
            struct FragmentCursor *cursor = (struct FragmentCursor *)&p[1];
            HeapError ecode;

            if (cursor->frag_out == NULL)
               break;

            ecode = frag_check(heap, filt, cursor->frag_out);
            if (ecode != HEAP_ERROR_NONE)
               break;

            /* Adjust fragment pointer. */
            cursor->frag_out  = &cursor->frag_out[1];
            cursor->size_out -= sizeof(struct Fragment);

            break;
         }
         case HPM_QUERY_FRAG_STATUS:
         {
            struct Fragment *frag = ((struct Fragment *)p[1]);
            HeapError ecode;

            if (frag == NULL)
               break;
            
            if (HPM_IS_HANDLED(p[0]))
            {
               ecode = frag_check(heap, filt, frag);
               if (ecode != HEAP_ERROR_NONE)
                  break;

               HPM_SET_VALUE(p[1], &frag[1]);

               if (frag_fbit(frag))
               {
                  HPM_SET_VALUE(p[2], p[2] & ~HEAP_FRAGMENT_ALLOCATED);
               }
            }
            break;
         }
         case HPM_QUERY_FRAG_SIZE:
         {
            struct Fragment *frag = ((struct Fragment *)p[1]);
            HeapError ecode;

            if (HPM_IS_HANDLED(p[0]))
            {
               ecode = frag_check(heap, filt, frag);
               if (ecode != HEAP_ERROR_NONE)
                  break;

               /* Remove fragment header. */
               HPM_SUB_VALUE(p[2], sizeof(struct Fragment));
            }
            break;
         }
         default:
	    break;
      }

      p += HPM_VALUES(p[0]);
      p += 1;
   }

   return HEAP_ACTION_CONTINUE;
}

/*****************************************************************************/

static HeapAction
l_fcache_control(struct Heap *heap, union HeapEvent *event, void *opaque)
{
   struct FCacheFilter *filt = opaque;
   HeapParam *p = event->e_query.params;
   HeapError ecode;

   if (p == NULL)
      return HEAP_ACTION_CONTINUE;
   
   while (HPM_ID(p[0]) != HPM_END)
   {
      switch (HPM_ID(p[0]))
      {
         case HPM_CTRL_FRAG_INSERT:
         {
            struct Fragment *frag = (struct Fragment *)p[1];
            size_t size = (size_t)p[2];

            HeapLock lock;

            if (frag == NULL)
               break;
            
            /* Align fragment pointer if necessary. */
            if (((uint32_t)frag) & (HEAP_ALIGNMENT-1))
            {
               struct Fragment *frag_tmp = (struct Fragment *)
                  (((uint32_t)frag + HEAP_ALIGNMENT - 1) & ~(HEAP_ALIGNMENT-1));

               if (size < HEAP_ALIGNMENT)
                  break;

               size -= ((uint32_t)frag_tmp - (uint32_t)frag);
               frag  = frag_tmp;
            }

            /* Align size. */
            size &= ~(HEAP_ALIGNMENT - 1);

            if (size == 0 || size <= sizeof(struct Fragment))
               break;

            /* Remove overhead. */
            size -= sizeof(struct Fragment);
            
            frag_init(filt, frag, size, 0);
            frag_set_external(frag);

            heap_lock(heap, &lock, 0);
            frag_insert(filt, frag);
            heap_unlock(heap, &lock, 0);

            HPM_SET_HANDLED(p[0]);
            break;
         }
         case HPM_CTRL_FRAG_REMOVE:
         {
            struct Fragment *frag = (struct Fragment *)p[1];
            size_t size = (size_t)p[2];

            HeapLock lock;

            heap_lock(heap, &lock, 0);

            if (frag != NULL)
            {
               frag = &frag[-1];

               ecode = frag_check(heap, filt, frag);
               if (ecode != HEAP_ERROR_NONE)
               {
                  heap_unlock(heap, &lock, 0);
                  break;
               }

               if (frag_id(frag) != filt->id)
               {
                  heap_unlock(heap, &lock, 0);
                  break;
               }

               if (!frag_fbit(frag))
               {
                  heap_unlock(heap, &lock, 0);
                  break;
               }

               frag_remove(filt, frag);
            }
            else
            {
               if (size == 0)
               {
                  /* Any fragment requested. */

                  /* Fetch first/smallest fragment. Large
                   * fragments are considered more valuable
                   * from a fragmentation and performance
                   * perspective. */
                  frag = (struct Fragment *)rbtree_first(&filt->frag_tree);
               }
               else
               {
                  /* Any fragment larger than 'size' requested. */

                  frag = (struct Fragment *)rbtree_find_best(
                     &filt->frag_tree, frag_find_best, (void *)size);
               }

               while (frag != NULL)
               {
                  /* Cannot remove externally supplied fragments. */
                  if (!frag_external(frag))
                  {
                     frag_remove(filt, frag);
                     break;
                  }

                  /* Try next larger fragment. */
                  frag = (struct Fragment *)rbtree_next(&filt->frag_tree, &frag->node);
               }
            }

            heap_unlock(heap, &lock, 0);

            if (frag != NULL)
            {
               HPM_SET_HANDLED(p[0]);
               HPM_SET_VALUE(p[1], frag);
               HPM_SET_VALUE(p[2], frag->size);
               HPM_ADD_VALUE(p[2], sizeof(struct Fragment));
            }
            
            break;
         }
         default:
	    break;
      }

      p += HPM_VALUES(p[0]);
      p += 1;
   }

   return HEAP_ACTION_CONTINUE;
}

/*****************************************************************************/

static HeapAction
l_fcache_destroy(struct Heap *heap, union HeapEvent *event, void *opaque)
{
   struct FCacheFilter *f = opaque;

   (void)event;
   
   zzheap_detach_hook(heap, HEAP_EVENT_BREAK_REQ,   l_fcache_break_req, f);
   zzheap_detach_hook(heap, HEAP_EVENT_BREAK_RES,   l_fcache_break_res, f);
   zzheap_detach_hook(heap, HEAP_EVENT_UNBREAK_REQ, l_fcache_unbreak,   f);
   zzheap_detach_hook(heap, HEAP_EVENT_QUERY_REQ,   l_fcache_query_req, f);
   zzheap_detach_hook(heap, HEAP_EVENT_QUERY_RES,   l_fcache_query_res, f);
   zzheap_detach_hook(heap, HEAP_EVENT_CONTROL_REQ, l_fcache_control,   f);

   if (f != NULL)
      heap_meta_free(heap, f);

   return HEAP_ACTION_CONTINUE;
}

/*****************************************************************************/



/*****************************************************************************
 * Public Interface
 *****************************************************************************/

HeapStatus
heap_attach_fcache_filter(struct Heap *heap,
                          struct HeapStatistics *stats,
                          uint32_t id, uint32_t options)
{
   struct FCacheFilter *f;
   HeapStatus status;

   (void)options;
   
   f = heap_meta_alloc(heap, sizeof(struct FCacheFilter));
   if (f == NULL)
      return HEAP_EOUT_OF_MEMORY;

   memset(f, 0, sizeof(struct FCacheFilter));

   f->id    = id << FRAG_ID_SHIFT;
   f->stats = stats;
   f->info.version = 101;
   strcpy(f->info.id, "com.enea.heap.fcache");

   rbtree_init(&f->frag_tree, frag_compare);
    
   status = zzheap_attach_hook(
      heap, HEAP_EVENT_BREAK_REQ, l_fcache_break_req, f, HEAP_PRIO_ALLOC_HIGH);
   if (status != HEAP_SUCCESS)
   {
      heap_meta_free(heap, f);
      return status;
   }

   status = zzheap_attach_hook(
      heap, HEAP_EVENT_BREAK_RES, l_fcache_break_res, f, HEAP_PRIO_ALLOC_HIGH);
   if (status != HEAP_SUCCESS)
   {
      heap_meta_free(heap, f);
      return status;
   }

   status = zzheap_attach_hook(
      heap, HEAP_EVENT_UNBREAK_REQ, l_fcache_unbreak, f, HEAP_PRIO_ALLOC_HIGH);
   if (status != HEAP_SUCCESS)
   {
      heap_meta_free(heap, f);
      return status;
   }

   status = zzheap_attach_hook(
      heap, HEAP_EVENT_QUERY_REQ, l_fcache_query_req, f, HEAP_PRIO_ALLOC_HIGH);
   if (status != HEAP_SUCCESS)
   {
      heap_meta_free(heap, f);
      return status;
   }

   status = zzheap_attach_hook(
      heap, HEAP_EVENT_QUERY_RES, l_fcache_query_res, f, HEAP_PRIO_FILTER_LOW);
   if (status != HEAP_SUCCESS)
   {
      heap_meta_free(heap, f);
      return status;
   }

   status = zzheap_attach_hook(
      heap, HEAP_EVENT_CONTROL_REQ, l_fcache_control, f, HEAP_PRIO_ALLOC_HIGH);
   if (status != HEAP_SUCCESS)
   {
      heap_meta_free(heap, f);
      return status;
   }

   status = zzheap_attach_hook(
      heap, HEAP_EVENT_DESTROY_RES, l_fcache_destroy, f, HEAP_PRIO_ALLOC_HIGH);
   if (status != HEAP_SUCCESS)
   {
      heap_meta_free(heap, f);
      return status;
   }

   return HEAP_SUCCESS;
}

/*****************************************************************************/



