/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2010 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

#include "heap_filt_safe.h"
#include <string.h>


/**************************************************************************
 * Plug-in History
 **************************************************************************
 *
 * IMPORTANT
 * 
 * This comment shall be updated with a description of the change each
 * time the extension version number is increased. The version number
 * shall be increased if incompatible modifications are made to internal
 * structures or after behavioral changes. See HeapExtension below.
 * 
 **************************************************************************
 *
 *  Version  Description
 *  -------  ----------------------------------------------------------
 *   101     Renamed struct Filter to struct SafeFilter.
 *   100     Initial revision.  
 *
 **************************************************************************/



/**************************************************************************
 * Data Structures
 **************************************************************************/

struct SafeFilter
{
   struct HeapServiceInfo info;
   
   /* Statistics.    */
   uint32_t  stats_alloc_failed;    /* Number of failed allocations.    */
   uint32_t  stats_break_failed;    /* Number of failed extensions.     */

   /* Error management. */
   HeapError error;                 /* Last error not yet reported.     */
   size_t    error_size;            /* Size requested at error event.   */

   /* Configuration. */
   uint32_t  cfg_on_alloc;          /* Always generate an error if set. */
   uint32_t  cfg_on_break;          /* Always generate an error if set. */
};

/*****************************************************************************/




/*****************************************************************************
 * Private Interface
 *****************************************************************************/

static HeapAction
l_safe_alloc(struct Heap *heap, union HeapEvent *event, void *opaque)
{
   struct SafeFilter *f = opaque;
   HeapParam *p = event->e_alloc.params;

   if (event->e_alloc.ptr != NULL)
   {
      return HEAP_ACTION_CONTINUE;
   }

   f->stats_alloc_failed++;

   if (p != NULL)
   {
      /* Scan for "safe" allocation paramater. */
      while (HPM_ID(*p) != HPM_END &&
             HPM_ID(*p) != HPM_ALLOC_SAFE &&
             HPM_ID(*p) != HPM_BREAK_NO)
      {
         p += (1 + HPM_VALUES(*p));
      }
   }

   /* Check if break event caused an error. */
   if (f->error != HEAP_ERROR_NONE)
   {
      /* Error detected at break. */
      zzheap_error(heap, f->error, (void *)f->error_size);
      f->error = HEAP_ERROR_NONE;         

      /* Error handler returned. Repeat allocation attempt if
       * it was a safe allocation. If unsafe, then return NULL. */
      if (p != NULL && HPM_ID(*p) == HPM_ALLOC_SAFE)
      {
         HPM_CLEAR_HANDLED(p[0]);
         return HEAP_ACTION_REPEAT;
      }

      return HEAP_ACTION_CONTINUE;
   }

   
   if (p != NULL)
   {
      if (HPM_ID(*p) == HPM_BREAK_NO)
      {
         /* Do not generate an error the
          * "no_extend" allocation method was used. */
         return HEAP_ACTION_CONTINUE;
      }
      else if (HPM_ID(*p) == HPM_ALLOC_SAFE && !HPM_IS_HANDLED(p[0]))
      {
         zzheap_error(heap, HEAP_ERROR_BUF_EXHAUSTED, (void *)event->e_alloc.size);
         return HEAP_ACTION_REPEAT;
      }      
   }

   if (f->cfg_on_alloc)
   {
      zzheap_error(heap, HEAP_ERROR_BUF_EXHAUSTED, (void *)event->e_alloc.size);
      return HEAP_ACTION_CONTINUE;
   }

   return HEAP_ACTION_CONTINUE;
}

/*****************************************************************************/

static HeapAction
l_safe_break(struct Heap *heap, union HeapEvent *event, void *opaque)
{
   struct SafeFilter *f = opaque;
   HeapParam *p = event->e_break.params;
   int generate_error = f->cfg_on_break;

   (void)heap;
   
   if (event->e_break.base != NULL)
      return HEAP_ACTION_CONTINUE;

   f->stats_break_failed++;

   if (p != NULL)
   {
      /* Scan for "safe" allocation paramater. */
      while (HPM_ID(*p) != HPM_END &&
             HPM_ID(*p) != HPM_ALLOC_SAFE)
      {
         p += (1 + HPM_VALUES(*p));
      }
   }

   if (p != NULL)
   {
      if (HPM_ID(*p) == HPM_ALLOC_SAFE)
      {
         /* Set handled bit to prevent call to error when
          * the subsequent alloc event fails too. */
         HPM_SET_HANDLED(p[0]);

         generate_error = 1;
      }
   }

   if (generate_error)
   {
      if (event->e_break.status)
      {
         f->error      = HEAP_ERROR_MEM_EXHAUSTED;
         f->error_size = event->e_break.min_size;
      }
      else
      {
         f->error      = HEAP_ERROR_BUF_EXHAUSTED; 
         f->error_size = event->e_break.min_size;
      }
   }

   return HEAP_ACTION_CONTINUE;
}

/*****************************************************************************/

static HeapAction
l_safe_query(struct Heap *heap, union HeapEvent *event, void *opaque)
{
   struct SafeFilter *f = opaque;
   HeapParam *p = event->e_query.params;

   (void)heap;
   
   if (p == NULL)
      return HEAP_ACTION_CONTINUE;
   
   while (*p != HPM_END)
   {
      switch (HPM_ID(*p))
      {
         case HPM_QUERY_SERVICE:
         {
            if (!HPM_IS_HANDLED(p[0]) && p[1] == 0)
            {
               HPM_SET_HANDLED(p[0]);
               HPM_SET_VALUE(p[1], &f->info);
            }
            else if (p[1] == (HeapParam)&f->info)
            {
               HPM_CLEAR_HANDLED(p[0]);
               HPM_SET_VALUE(p[1], 0);
            }

            break;
         }
         case HPM_QUERY_ALLOC_NULL:
         {
            HPM_SET_HANDLED(p[0]);
            HPM_ADD_VALUE(p[1], f->stats_alloc_failed);
            break;
         }
         case HPM_QUERY_BREAK_NULL:
         {
            HPM_SET_HANDLED(p[0]);
            HPM_ADD_VALUE(p[1], f->stats_break_failed);
         }
         default:
         {
            break;
         }
      }

      p += (1 + HPM_VALUES(*p));
   }

   return HEAP_ACTION_CONTINUE;
}

/*****************************************************************************/

static HeapAction
l_safe_destroy(struct Heap *heap, union HeapEvent *event, void *opaque)
{
   struct SafeFilter *f = opaque;

   (void)event;

   /* This should be optional. ... */
   zzheap_detach_hook(heap, HEAP_EVENT_ALLOC_RES, l_safe_alloc, f);
   zzheap_detach_hook(heap, HEAP_EVENT_BREAK_RES, l_safe_break, f);
   zzheap_detach_hook(heap, HEAP_EVENT_QUERY_REQ, l_safe_query, f);

   heap_meta_free(heap, f);
   
   return HEAP_ACTION_CONTINUE;
}

/*****************************************************************************/




/*****************************************************************************
 * Public Interface
 *****************************************************************************/

HeapStatus
heap_attach_safe_filter(struct Heap *heap, uint8_t on_alloc, uint8_t on_break)
{
   struct SafeFilter *f = heap_meta_alloc(heap, sizeof(struct SafeFilter));
   HeapStatus status;

   if (f == NULL)
      return HEAP_EOUT_OF_MEMORY;

   bzero(f, sizeof(struct SafeFilter));

   f->cfg_on_alloc = on_alloc;
   f->cfg_on_break = on_break;
   
   f->info.version = 101;
   strcpy(f->info.id, "com.enea.heap.safe");
   
   status = zzheap_attach_hook(
      heap, HEAP_EVENT_ALLOC_RES, l_safe_alloc, f, HEAP_PRIO_FILTER_NORMAL);
   if (status != HEAP_SUCCESS)
   {
      heap_meta_free(heap, f);
      return status;
   }

   status = zzheap_attach_hook(
      heap, HEAP_EVENT_BREAK_RES, l_safe_break, f, HEAP_PRIO_FILTER_NORMAL);
   if (status != HEAP_SUCCESS)
   {
      heap_meta_free(heap, f);
      return status;
   }

   status = zzheap_attach_hook(
      heap, HEAP_EVENT_QUERY_REQ, l_safe_query, f, HEAP_PRIO_FILTER_NORMAL);
   if (status != HEAP_SUCCESS)
   {
      heap_meta_free(heap, f);
      return status;
   }

   status = zzheap_attach_hook(
      heap, HEAP_EVENT_DESTROY_RES, l_safe_destroy, f, HEAP_PRIO_FILTER_NORMAL);
   if (status != HEAP_SUCCESS)
   {
      heap_meta_free(heap, f);
      return status;
   }

   return HEAP_SUCCESS;
}


