/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2010 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */


#include "heap_filt_tentative.h"
#include <string.h>

/**************************************************************************
 * Plug-in History
 **************************************************************************
 *
 * IMPORTANT
 * 
 * This comment shall be updated with a description of the change each
 * time the extension version number is increased. The version number
 * shall be increased if incompatible modifications are made to internal
 * structures or after behavioral changes. See HeapExtension below.
 * 
 **************************************************************************
 *
 *  Version  Description
 *  -------  ----------------------------------------------------------
 *   101     Renamed struct Filter to struct TentativeFilter.
 *   100     Initial revision.  
 *
 **************************************************************************/



/*****************************************************************************
 * Macros
 *****************************************************************************/

#define HEAP_PID_INVALID  (-1ul)

/*****************************************************************************/




/*****************************************************************************
 * Data Structures
 *****************************************************************************/

#define slot_invalidate(slot)   ((slot)->avail.handle += 0x10000)
#define slot_index(handle)      ((handle) & 0xFFFF)
#define slot_instance(handle)   ((handle) & 0xFFFF0000ul)
#define slot_set_next(_s, _n)   ((_s)->avail.next = (union Slot *)((uint32_t)(_n)|1))
#define slot_get_next(_s)       ((union Slot *)((uint32_t)(_s)->avail.next & ~1))
#define slot_is_used(_s)        (!((uint32_t)(_s)->avail.next & 1))

union Slot 
{
   /* When slot is unused. */
   struct {
      union Slot *next;          /* Lowest bit indicates availability. */
      uint32_t    handle;
   } avail;

   /* When slot holds a buffer. */
   struct {
      void       *ptr;
      uint32_t    handle;
   } used;
};

/*****************************************************************************/

struct TentativeFilter 
{
   struct HeapServiceInfo info;

   uint32_t    num_bufs;   /* Number of slots.            */
   uint32_t    num_valid;  /* Number of valid slots.      */
   uint32_t    gc_index;   /* Index of next buffer to GC. */
   union Slot *bufs;       /* Array of slots.             */
   union Slot *next;       /* Next available slot.        */
};

/*****************************************************************************/



/*****************************************************************************
 * Helper Functions
 *****************************************************************************/

static inline HeapPID
buf_owner(struct Heap *heap, void *ptr)
{
   HeapParam p[] = {HPM(QUERY_BUF_OWNER, 1), 0, HPM_END};

   p[1] = (HeapParam)ptr;

   zzheap_query(heap, p, 0);

   return HPM_IS_HANDLED(p[0]) ? (HeapPID)p[1] : HEAP_PID_INVALID;
}

/*****************************************************************************/



/*****************************************************************************
 * Private Interface
 *****************************************************************************/

static void
ctrl_tentative_alloc(struct Heap *heap, struct TentativeFilter *f, HeapParam *p)
{
   union {
      struct {
         void    *ptr;
      } output;
      struct {
         uint32_t handle;
         HeapPID  pid;
      } input;
   } *param = (void *)&p[1];

   union Slot *slot;
   HeapLock lock;
   uint32_t index;

   /* Valid slot index ? */
   index = slot_index(param->input.handle);
   if (unlikely(index >= f->num_bufs))
      return;

   slot = &f->bufs[index];
    
   heap_lock(heap, &lock, 0);

   /* Valid instance ? */
   if (unlikely(slot->used.handle != param->input.handle))
   {
      heap_unlock(heap, &lock, 0);
      return;
   }

   param->output.ptr = slot->used.ptr;

   /* Bump handle instance. */
   slot_invalidate(slot);

   /* Insert slot into free list. Also sets the availability bit. */
   slot_set_next(slot, f->next);
   f->next = slot;
   f->num_valid--;

   heap_unlock(heap, &lock, 0);

   HPM_SET_HANDLED(p[0]);
}

/*****************************************************************************/

static void
ctrl_tentative_free(struct Heap *heap, struct TentativeFilter *f, HeapParam *p)
{
   union {
      struct {
         void    *ptr;
         HeapPID  pid;
      } input;
      struct {
         uint32_t handle;
      } output;
   } *param = (void *)&p[1];

   union Slot *slot;
   HeapLock lock;

   heap_lock(heap, &lock, 0);
   while (unlikely(f->next == NULL))
   {
      /* Cache is full. Need to garbage collect
       * one buffer to make room for the new one. */

      uint32_t buf_ix, buf_cnt = 0;
      void *ptr;
      
      do {
         buf_ix = (f->gc_index + (buf_cnt++)) % f->num_bufs;

         slot = &f->bufs[buf_ix];
      } while (!slot_is_used(slot) && buf_cnt < f->num_bufs);

      /* Update next index to garbage collect. */
      f->gc_index = (f->gc_index + buf_cnt) % f->num_bufs;

      if (unlikely(!slot_is_used(slot)))
      {
         /* No buffer to garbage collect was found. It
          * probably disappeared while the cache was iterated. */
         continue;
      }

      ptr = slot->used.ptr;

      /* Bump handle instance. */
      slot_invalidate(slot);

      /* Insert slot into free list. Also sets the availability bit. */
      slot_set_next(slot, f->next);
      f->next = slot;
      f->num_valid--;

      heap_unlock(heap, &lock, 0);

      zzheap_free(heap, ptr, 0, NULL, __FILE__, __LINE__);
      
      heap_lock(heap, &lock, 0);
   }

   slot = f->next;
   f->next = slot_get_next(slot);
   f->num_valid++;
   
   /* Set buffer pointer in the slot. Also clears 
    * the lowest bit indicating availability.  */
   slot->used.ptr = param->input.ptr;
   param->output.handle = slot->used.handle;
   heap_unlock(heap, &lock, 0);
   
   HPM_SET_HANDLED(p[0]);
}

/*****************************************************************************/

static void
ctrl_tentative_flush(struct Heap *heap, struct TentativeFilter *f, HeapParam *p)
{
   /* Flushes the entire tentative cache. */
   union Slot *slot;
   void *ptr;
   uint32_t buf_ix, bufs = 0;
   HeapLock lock;
            
   for (buf_ix = 0; buf_ix < f->num_bufs; buf_ix++)
   {
      slot = &f->bufs[buf_ix];
               
      if (!slot_is_used(slot))
         continue;
               
      ptr = slot->used.ptr;
               
      heap_lock(heap, &lock, 0);
                  
      /* Bump handle generation. */
      slot_invalidate(slot);
                  
      /* Insert slot into free list. Also sets the availability bit. */
      slot_set_next(slot, f->next);
      f->next = slot;
      f->num_valid--;
               
      heap_unlock(heap, &lock, 0);

      zzheap_free(heap, ptr, 0, NULL, __FILE__, __LINE__);

      bufs++;
   }

   /* Only set parameter as handled if one or more buffers were collected. */
   if (bufs > 0)
      HPM_SET_HANDLED(p[0]);
}

/*****************************************************************************/

static HeapAction
l_tentative_alloc_res(struct Heap *heap, union HeapEvent *event, void *opaque)
{
   struct TentativeFilter *f = opaque;
   union Slot *slot;
   uint32_t buf_cnt = 0, buf_ix = 0;
   void *ptr;

   HeapLock lock;

   /* Check if client allocation failed. */
   if (likely(event->e_alloc.ptr != NULL))
      return HEAP_ACTION_CONTINUE;
   
   /* Yes, is there anything to garbage collect? */
   if (f->num_valid == 0)
      return HEAP_ACTION_CONTINUE;

   heap_lock(heap, &lock, 0);

   do {
      buf_ix = (f->gc_index + (buf_cnt++)) % f->num_bufs;

      slot = &f->bufs[buf_ix];
   } while (!slot_is_used(slot) && buf_cnt < f->num_bufs);

   /* Update next index to garbage collect. */
   f->gc_index = (f->gc_index + buf_cnt) % f->num_bufs;

   if (unlikely(!slot_is_used(slot)))
   {
      /* No buffer to garbage collect was found. It
       * probably disappeared while the cache was iterated. */

      heap_unlock(heap, &lock, 0);
      return HEAP_ACTION_CONTINUE;
   }

   ptr = slot->used.ptr;

   /* Bump handle instance. */
   slot_invalidate(slot);

   /* Insert slot into free list. Also sets the availability bit. */
   slot_set_next(slot, f->next);
   f->next = slot;
   f->num_valid--;
   
   heap_unlock(heap, &lock, 0);

   /* Release the memory. */
   zzheap_free(heap, ptr, 0, NULL, __FILE__, __LINE__);

   /* ... and attempt to satisfy the client allocation once more. */
   return HEAP_ACTION_REPEAT;
}

/*****************************************************************************/

static HeapAction
l_tentative_collect(struct Heap *heap, union HeapEvent *event, void *opaque)
{
   struct TentativeFilter *f = opaque;
   union Slot *slot;
   void *ptr;
   HeapPID pid;
   uint32_t buf_ix;

   HeapLock lock;

   if (f->num_valid == 0)
      return HEAP_ACTION_CONTINUE;
   
   for (buf_ix = 0; buf_ix < f->num_bufs; buf_ix++)
   {
      slot = &f->bufs[buf_ix];
      
      if (!slot_is_used(slot))
         continue;

      ptr = slot->used.ptr;
      pid = buf_owner(heap, ptr);

      if (pid == event->e_collect.owner)
      {
         heap_lock(heap, &lock, 0);

         /* Bump handle generation. */
         slot_invalidate(slot);
      
         /* Insert slot into free list. Also sets the availability bit. */
         slot_set_next(slot, f->next);
         f->next = slot;
         f->num_valid--;

         /* No need to free the buffer. It will be collected
          * by its managing algorithm when this hook is finished. */

         heap_unlock(heap, &lock, 0);
      }
   }

   return HEAP_ACTION_CONTINUE;
}

/*****************************************************************************/

static HeapAction
l_tentative_query(struct Heap *heap, union HeapEvent *event, void *opaque)
{
   struct TentativeFilter *f = opaque;
   HeapParam *p = event->e_query.params;

   (void)heap;
   
   if (p == NULL)
      return HEAP_ACTION_CONTINUE;
   
   while (*p != HPM_END)
   {
      switch (HPM_ID(*p))
      {
         case HPM_QUERY_SERVICE:
         {
            if (!HPM_IS_HANDLED(p[0]) && p[1] == 0)
            {
               HPM_SET_HANDLED(p[0]);
               HPM_SET_VALUE(p[1], &f->info);
            }
            else if (p[1] == (HeapParam)&f->info)
            {
               HPM_CLEAR_HANDLED(p[0]);
               HPM_SET_VALUE(p[1], 0);
            }

            break;
         }
      }

      p += (1 + HPM_VALUES(*p));
   }

   return HEAP_ACTION_CONTINUE;
}

/*****************************************************************************/

static HeapAction
l_tentative_control(struct Heap *heap, union HeapEvent *event, void *opaque)
{
   typedef void (*CtrlFunc)(struct Heap *heap, struct TentativeFilter *f, HeapParam *p);

   /* Allocation and deallocation operations are implemented
    * through control events. This is a bit unintuitive, but
    * it is difficult to use the regular alloc and free event
    * for several reasons, the key being that other installed
    * plug-ins might modify the buffer outside this plug-in's
    * control.
    *
    * For example, a debug module may write a debug
    * pattern into the buffer at allocation or deallocation,
    * which would render this plug-in useless since the data
    * is supposed to be retained. 
    */
   
   static const CtrlFunc funcs[] = {
      ctrl_tentative_alloc,
      ctrl_tentative_free,
      ctrl_tentative_flush
   };
   
   HeapParam *p = event->e_query.params;
   struct TentativeFilter *f = opaque;

   if (unlikely(p == NULL))
      return HEAP_ACTION_CONTINUE;
   
   for (; p[0] != HPM_END; p += (1 + HPM_VALUES(p[0])))
   {
      HeapParam id = HPM_ID(p[0]);
      
      if (id >= HPM_CTRL_TENTATIVE_BASE &&
          id <= HPM_CTRL_TENTATIVE_END)
      {
         funcs[id - HPM_CTRL_TENTATIVE_BASE](heap, f, p);
      }
   }

   return HEAP_ACTION_CONTINUE;
}

/*****************************************************************************/

static HeapAction
l_tentative_destroy(struct Heap *heap, union HeapEvent *event, void *opaque)
{
   struct TentativeFilter *f = opaque;

   (void)event;

   zzheap_detach_hook(heap, HEAP_EVENT_ALLOC_RES,   l_tentative_alloc_res, f);
   zzheap_detach_hook(heap, HEAP_EVENT_COLLECT_REQ, l_tentative_collect,   f);
   zzheap_detach_hook(heap, HEAP_EVENT_QUERY_REQ,   l_tentative_query,     f);
   zzheap_detach_hook(heap, HEAP_EVENT_CONTROL_REQ, l_tentative_control,   f);


   if (f != NULL)
   {
      heap_meta_free(heap, f->bufs);
      heap_meta_free(heap, f);
   }

   return HEAP_ACTION_CONTINUE;
}

/*****************************************************************************/





/*****************************************************************************
 * Public Interface
 *****************************************************************************/

HeapStatus
heap_attach_tentative_filter(struct Heap *heap, uint32_t count)
{
   struct TentativeFilter *f = heap_meta_alloc(
      heap, sizeof(struct TentativeFilter));
   HeapStatus status;
   uint32_t ix;

   if (f == NULL)
      return HEAP_EOUT_OF_MEMORY;

   f->num_bufs  = count;
   f->num_valid = 0;
   f->gc_index  = 0;

   f->info.version = 101;
   strcpy(f->info.id, "com.enea.heap.tentative");
   
   f->bufs = heap_meta_alloc(heap, count * sizeof(union Slot));
   if (f->bufs == NULL)
   {
      heap_meta_free(heap, f);
      return HEAP_EOUT_OF_MEMORY;
   }

   /* Initialize all slots. */
   for (ix=count-2; ix < count; ix--)
   {
      slot_set_next(&f->bufs[ix], &f->bufs[ix+1]);
      f->bufs[ix].avail.handle = ix;
      slot_invalidate(&f->bufs[ix]);
   }

   slot_set_next(&f->bufs[count-1], NULL);
   f->bufs[count-1].avail.handle = count - 1;
   
   slot_invalidate(&f->bufs[0]);

   f->next = &f->bufs[0];
    
   status = zzheap_attach_hook(
      heap, HEAP_EVENT_ALLOC_RES, l_tentative_alloc_res, f, HEAP_PRIO_ALGO_HIGH);
   if (status != HEAP_SUCCESS)
      return status;

   status = zzheap_attach_hook(
      heap, HEAP_EVENT_COLLECT_REQ, l_tentative_collect, f, HEAP_PRIO_ALGO_HIGH);
   if (status != HEAP_SUCCESS)
      return status;

   status = zzheap_attach_hook(
      heap, HEAP_EVENT_QUERY_REQ, l_tentative_query, f, HEAP_PRIO_ALGO_HIGH);
   if (status != HEAP_SUCCESS)
      return status;
   
   status = zzheap_attach_hook(
      heap, HEAP_EVENT_CONTROL_REQ, l_tentative_control, f, HEAP_PRIO_ALGO_HIGH);
   if (status != HEAP_SUCCESS)
      return status;
   
   status = zzheap_attach_hook(
      heap, HEAP_EVENT_DESTROY_RES, l_tentative_destroy, f, HEAP_PRIO_ALGO_HIGH);
   if (status != HEAP_SUCCESS)
      return status;

   return HEAP_SUCCESS;
}

/*****************************************************************************/

