/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2007 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */


#include "heap.h"
#include "list_ext.h"

#include <string.h>



/*************************************************************************
 * Datatypes
 *************************************************************************/

typedef struct HeapHookLink 
{
   list_link_t   link;
   HeapHook      hook;
   void         *opaque;
   int           priority;
} HeapHookLink;

/*************************************************************************/




/*************************************************************************
 * Helper Functions
 *************************************************************************/

static always_inline HeapAction
l_heap_event(struct Heap *heap, HeapEventID event_id, union HeapEvent *event)
{
   HeapAction action = HEAP_ACTION_CONTINUE;
   list_t *list = &heap->event_hooks[event_id];
   HeapHookLink *link;

   if (likely(event != NULL))
      event->id = event_id;

   for (link = (struct HeapHookLink *)list_head(list);
        likely(link != NULL);
        link = (struct HeapHookLink *)list_next(list, &link->link))
   {
      heap_assert(link->hook != NULL, "A function pointer is NULL.");
      action = link->hook(heap, event, link->opaque);

      if (action != HEAP_ACTION_CONTINUE)
         break;
   }

   return action;
}

/*************************************************************************/

static int
l_heap_hook_compare(const list_link_t *a_, const list_link_t *b_)
{
   const HeapHookLink *a = (const HeapHookLink *)a_;
   const HeapHookLink *b = (const HeapHookLink *)b_;

   if (a->priority < b->priority)
      return -1;
   if (a->priority > b->priority)
      return +1;

   return 0;
}

/*************************************************************************/





/*************************************************************************
 * Public Interface
 *************************************************************************/

HeapStatus
zzheap_create(struct Heap *heap)
{
   uint32_t ix;

   heap->freezed = 0;
   
   for (ix = 0; ix < HEAP_EVENT_COUNT; ix++)
   {
      init_list(&heap->event_hooks[ix]);
   }
   
   return HEAP_SUCCESS;
}


/*************************************************************************/

void
zzheap_destroy(struct Heap *heap)
{
   HeapAction action;
   HeapEventID event_id;

   action = l_heap_event(heap, HEAP_EVENT_DESTROY_RES, NULL);
#ifdef ASSERT
   heap_assert(action == HEAP_ACTION_CONTINUE,
               "An event hook returned an illegal action for this event.");
#else
   (void)action;
#endif /* ASSERT */

   /* Free administrative hook data that hasn't been detached. */
   for (event_id = 0; event_id < HEAP_EVENT_COUNT; event_id++)
   {
      list_t *list = &heap->event_hooks[event_id];
      list_link_t *link, *link_next;

      link = list_head(list);
      while (link != NULL)
      {
         link_next = list_next(list, link);
         
         heap_meta_free(heap, link);

         link = link_next;
      }
   }

   heap_lock_destroy(heap);
}

/*************************************************************************/

HeapStatus
zzheap_attach_hook(struct Heap *heap,
                   HeapEventID event, HeapHook hook, void *opaque,
                   int priority)
{
   struct HeapHookLink *link;

   heap_assert(event < HEAP_EVENT_COUNT, "Undefined event id");

   link = heap_meta_alloc(heap, sizeof(struct HeapHookLink));
   if (link == NULL)
      return HEAP_EOUT_OF_MEMORY;

   link->hook     = *hook;
   link->opaque   = opaque;
   link->priority = priority;

   insert_list_ordered(&heap->event_hooks[event],
                       &link->link, l_heap_hook_compare);

   return HEAP_SUCCESS;
}

/*************************************************************************/

void
zzheap_detach_hook(struct Heap *heap,
                   HeapEventID event, HeapHook hook, void *opaque)
{
   struct HeapHookLink *link;
   list_t *list = NULL;

   heap_assert(event < HEAP_EVENT_COUNT, "Undefined event id");

   /* Not allowed to call from inside an attached hook. */

   list = &heap->event_hooks[event];
   
   for (link = (struct HeapHookLink *)list_head(list);
        link != NULL;
        link = (struct HeapHookLink *)list_next(list, &link->link))
   {
      if (link->opaque == opaque &&
          link->hook   == *hook)
      {
         delete_from_list(list, &link->link);
         heap_meta_free(heap, link);
         return;
      }
   }
   /* The hook was not found - ignore */
}

/*************************************************************************/

void *
zzheap_alloc(struct Heap *heap,
             size_t size,
             HeapPID pid, HeapParam *params,
             const char *file, int line)
{
   HeapAction action;

   struct HeapAllocEvent event;
   event.size   = size;
   event.owner  = pid;
   event.params = params;
   event.ptr    = NULL;
   event.file   = file;
   event.line   = line;


   do {
      action = l_heap_event(heap, HEAP_EVENT_ALLOC_REQ, (union HeapEvent *)&event);
      if (unlikely(action == HEAP_ACTION_ABORT))
      {
         heap_assert(event.ptr == NULL,
                     "A buffer was leaked when an event was aborted.");
         return NULL;
      }

      heap_assert(action != HEAP_ACTION_REPEAT,
                  "An event hook returned an illegal action for this event.");

      action = l_heap_event(heap, HEAP_EVENT_ALLOC_RES, (union HeapEvent *)&event);
      if (unlikely(action == HEAP_ACTION_ABORT))
      {
         heap_assert(event.ptr == NULL,
                     "A buffer was leaked when an event was aborted.");
         return NULL;
      }
   } while (unlikely(action == HEAP_ACTION_REPEAT));
   
   return event.ptr;
}

/*************************************************************************/

void *
zzheap_realloc(struct Heap *heap,
               void *ptr, size_t size,
               HeapPID pid, HeapParam *params,
               uint32_t resize_only, 
               const char *file, int line)
{
   HeapAction action;

   struct HeapResizeEvent event;
   event.size   = size;
   event.owner  = pid;
   event.params = params;
   event.ptr    = ptr;
   event.file   = file;
   event.line   = line;

   if (unlikely(ptr == NULL))
   {
      if (likely(resize_only))
         return NULL;
      else
         return zzheap_alloc(heap, size, pid, params, file, line);
   }

   /* Note: The buffer cannot be freed if it is a resize-only request.
    * heap_buffer_resize() will query the buffer size immediately
    * afterwards and the pointer should remain valid -> resize
    * but don't free. */
   if (unlikely(size == 0 && resize_only == 0))
   {
      zzheap_free(heap, ptr, pid, params, file, line);
      return NULL;
   }

   do {
      action = l_heap_event(heap, HEAP_EVENT_RESIZE_REQ, (union HeapEvent *)&event);
      if (unlikely(action == HEAP_ACTION_ABORT))
      {
         heap_assert(event.ptr == NULL,
                     "A buffer was leaked when an event was aborted.");
         return NULL;
      }

      if (unlikely(action == HEAP_ACTION_CONTINUE))
         zzheap_error(heap, HEAP_ERROR_BUF_UNKNOWN, ptr);

      heap_assert(action != HEAP_ACTION_REPEAT,
                  "An event hook returned an illegal action for this event.");

      action = l_heap_event(heap, HEAP_EVENT_RESIZE_RES, (union HeapEvent *)&event);
      if (unlikely(action == HEAP_ACTION_ABORT))
      {
         heap_assert(event.ptr == NULL,
                     "A buffer was leaked when an event was aborted.");
         return NULL;
      }
   } while (unlikely(action == HEAP_ACTION_REPEAT));


   /* Copy if resize request was unsuccessful. */
   if (event.ptr == NULL && resize_only == 0)
   {
      size_t copy_size;
      HeapParam p[] = {
         HPM(QUERY_BUF_ACT_SIZE, 1), 0,
         HPM_END};
      
      p[1] = (HeapParam)ptr;
      zzheap_query(heap, p, 0);

      if (unlikely(!HPM_IS_HANDLED(p[0])))
         return NULL;

      copy_size = (size_t)p[1];
      if (copy_size > size)
         copy_size = size;
      
      event.ptr = zzheap_alloc(heap, size, pid, params, file, line);
      if (unlikely(event.ptr == NULL))
         return NULL;
      memcpy(event.ptr, ptr, copy_size);
      zzheap_free(heap, ptr, pid, params, file, line);
   }

   return event.ptr;
}

/*************************************************************************/

void
zzheap_free(struct Heap *heap,
            void *ptr,
            HeapPID pid, HeapParam *params,
            const char *file, int line)
{
   HeapAction action;

   struct HeapFreeEvent event;
   event.owner  = pid;
   event.params = params;
   event.ptr    = ptr;
   event.file   = file;
   event.line   = line;

   action = l_heap_event(heap, HEAP_EVENT_FREE_REQ, (union HeapEvent *)&event);
   if (unlikely(action == HEAP_ACTION_ABORT))
   {
      heap_assert(event.ptr == NULL,
                  "A buffer was leaked when an event was aborted.");
      return;
   }

   if (unlikely(action == HEAP_ACTION_CONTINUE))
      zzheap_error(heap, HEAP_ERROR_BUF_UNKNOWN, ptr);
   
   heap_assert(action != HEAP_ACTION_REPEAT,
               "An event hook returned an illegal action for this event.");

   action = l_heap_event(heap, HEAP_EVENT_FREE_RES, (union HeapEvent *)&event);
   if (unlikely(action == HEAP_ACTION_ABORT))
   {
      heap_assert(event.ptr == NULL,
                  "A buffer was leaked when an event was aborted.");
      return;
   }

   heap_assert(action != HEAP_ACTION_REPEAT,
               "An event hook returned an illegal action for this event.");

}

/*************************************************************************/

void			 
zzheap_collect(struct Heap *heap, HeapPID pid)
{
   HeapAction action;

   struct HeapCollectEvent event;
   event.owner = pid;

   action = l_heap_event(heap, HEAP_EVENT_COLLECT_REQ, (union HeapEvent *)&event);
   if (unlikely(action == HEAP_ACTION_ABORT))
      return;

   heap_assert(action != HEAP_ACTION_REPEAT,
               "An event hook returned an illegal action for this event.");
      
   action = l_heap_event(heap, HEAP_EVENT_COLLECT_RES, (union HeapEvent *)&event);
   if (unlikely(action == HEAP_ACTION_ABORT))
      return;

   heap_assert(action != HEAP_ACTION_REPEAT,
               "An event hook returned an illegal action for this event.");
}

/*************************************************************************/

HeapStatus
zzheap_control(struct Heap *heap, HeapParam *params)
{
   HeapAction action;

   struct HeapControlEvent event;
   event.params = params;

   action = l_heap_event(heap, HEAP_EVENT_CONTROL_REQ, (union HeapEvent *)&event);
   if (unlikely(action == HEAP_ACTION_ABORT))
      return HEAP_EABORTED;

   heap_assert(action != HEAP_ACTION_REPEAT,
               "An event hook returned an illegal action for this event.");
      
   action = l_heap_event(heap, HEAP_EVENT_CONTROL_RES, (union HeapEvent *)&event);
   if (unlikely(action == HEAP_ACTION_ABORT))
      return HEAP_EABORTED;

   heap_assert(action != HEAP_ACTION_REPEAT,
               "An event hook returned an illegal action for this event.");

   return HEAP_SUCCESS;
}

/*************************************************************************/

HeapStatus
zzheap_query(struct Heap *heap, HeapParam *params, uint32_t fm)
{
   HeapAction action;

   struct HeapQueryEvent event;
   event.params  = params;
   event.freezed = fm;

   action = l_heap_event(heap, HEAP_EVENT_QUERY_REQ, (union HeapEvent *)&event);
   if (unlikely(action == HEAP_ACTION_ABORT))
      return HEAP_EABORTED;

   heap_assert(action != HEAP_ACTION_REPEAT,
               "An event hook returned an illegal action for this event.");

   action = l_heap_event(heap, HEAP_EVENT_QUERY_RES, (union HeapEvent *)&event);
   if (unlikely(action == HEAP_ACTION_ABORT))
      return HEAP_EABORTED;

   heap_assert(action != HEAP_ACTION_REPEAT,
               "An event hook returned an illegal action for this event.");


   return HEAP_SUCCESS;
}

/*************************************************************************/

void *
zzheap_break(struct Heap *heap,
             size_t *min_size, size_t req_size, size_t *max_size,
             uint32_t algo, HeapParam *params)
{
   HeapAction action;

   struct HeapBreakEvent event;
   event.base     = NULL;
   event.min_size = *min_size;
   event.req_size = req_size;
   event.max_size = *max_size;
   event.algo_id  = algo;
   event.params   = params;
   event.status   = 0;


   action = l_heap_event(heap, HEAP_EVENT_BREAK_REQ, (union HeapEvent *)&event);
   if (unlikely(action == HEAP_ACTION_ABORT))
   {
      *max_size = event.max_size;
      heap_assert(event.base == NULL,
                  "A fragment was leaked when an event was aborted.");
      return NULL;
   }

   heap_assert(action != HEAP_ACTION_REPEAT,
               "An event hook returned an illegal action for this event.");
      
   action = l_heap_event(heap, HEAP_EVENT_BREAK_RES, (union HeapEvent *)&event);
   if (unlikely(action == HEAP_ACTION_ABORT))
   {
      heap_assert(event.base == NULL,
               "A fragment was leaked when an event was aborted.");
      return NULL;
   }

   heap_assert(action != HEAP_ACTION_REPEAT,
               "An event hook returned an illegal action for this event.");

   *min_size = event.min_size;
   *max_size = event.max_size;
   
   return event.base;
}

/*************************************************************************/

void 
zzheap_unbreak(struct Heap *heap, void *base, HeapParam *params)
{
   HeapAction action;

   struct HeapUnbreakEvent event;
   event.base   = base;
   event.params = params;


   action = l_heap_event(heap, HEAP_EVENT_UNBREAK_REQ, (union HeapEvent *)&event);
   if (unlikely(action == HEAP_ACTION_ABORT))
   {
      heap_assert(event.base != NULL,
                  "A fragment was leaked when an event was aborted.");
      return;
   }

   if (unlikely(action == HEAP_ACTION_CONTINUE))
      zzheap_error(heap, HEAP_ERROR_FRAG_UNKNOWN, base);

   heap_assert(action != HEAP_ACTION_REPEAT,
               "An event hook returned an illegal action for this event.");
      

   action = l_heap_event(heap, HEAP_EVENT_UNBREAK_RES, (union HeapEvent *)&event);
   if (unlikely(action == HEAP_ACTION_ABORT))
   {
      heap_assert(event.base != NULL,
                  "A fragment was leaked when an event was aborted.");
      return;
   }

   heap_assert(action != HEAP_ACTION_REPEAT,
               "An event hook returned an illegal action for this event.");
}

/*************************************************************************/

void 
zzheap_error(struct Heap *heap, HeapError ecode, void *data)
{
   HeapAction action;

   struct HeapErrorEvent event;
   event.ecode = ecode;
   event.data  = data;

   action = l_heap_event(heap, HEAP_EVENT_ERROR_RES, (union HeapEvent *)&event);
#ifdef ASSERT
   heap_assert(action != HEAP_ACTION_REPEAT,
               "An event hook returned an illegal action for this event.");
#else
   (void)action;
#endif /* ASSERT */
}

/*************************************************************************/
