/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2010-2014 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

#ifndef HEAP_H
#define HEAP_H

#include "heap_osal.h"
#include "heap_error.h"
#include "heap_list.h"
#include "heap_param.h"


/*
 *************************************************************************
 * Declarations
 *************************************************************************
 */
struct Heap;
union  HeapEvent;

typedef uint32_t  HeapStatus;
typedef uint32_t  HeapEventID;
typedef uint32_t  HeapAction;
typedef uint32_t  HeapParam;
typedef uint32_t  HeapBufferType;
typedef uint32_t  HeapBufferStatus;
typedef uint32_t  HeapFragmentType;
typedef uint32_t  HeapFragmentStatus;
typedef uint32_t  HeapLockType;


/*
 *************************************************************************
 * Heap Datatypes
 *************************************************************************
 */

/**
 * @type  HeapBufferType
 *
 * @brief Enum containing heap buffer type identifiers.
 *
 * @field HEAP_BUFFER_FBUDDY
 *        A buffer distributed by the fibonacci buddy algorithm.
 *
 * @field HEAP_BUFFER_LARGE
 *        A buffer distributed by the large buffer algorithm
 *        (externally allocated).
 *
 * @field HEAP_BUFFER_TLSF
 *        A buffer distributed by the two-level segregated fit algorithm.
 * 
 * @field HEAP_BUFFER_POOL
 *        Not yet implemented.
 */
enum HeapBufferTypeEnum
{
   HEAP_BUFFER_FBUDDY  = 0x1,
   HEAP_BUFFER_LARGE   = 0x2,
   HEAP_BUFFER_TLSF    = 0x3,
   HEAP_BUFFER_POOL    = 0x4
};

/**
 * @type  HeapFragmentType
 *
 * @brief Enum containing heap fragment type identifiers.
 *
 * @field HEAP_FRAGMENT_REGION
 *        A fragment distributed by the OSE region allocator. 
 *
 * @field HEAP_FRAGMENT_MMAP
 *        A fragment that was allocated using POSIX mmap().
 *
 * @field HEAP_FRAGMENT_SBRK
 *        A fragment that was allocated using UNIX sbrk().
 *
 * @field HEAP_FRAGMENT_BUDDY
 *        A logical Fibonacci Buddy fragment. A l
 *        part of a larNot managed by an allocator. 
 *
 * @field HEAP_FRAGMENT_MANUAL
 *        A fragment that was manually provided the heap, i.e. not
 *        automatically allocated as part of normal operation. 
 *        Example usage is OSE APIs such as heap_extend_heap().
 */
enum HeapFragmentTypeEnum
{
   HEAP_FRAGMENT_REGION = 0x0,
   HEAP_FRAGMENT_MMAP   = 0x1,
   HEAP_FRAGMENT_SBRK   = 0x2,
   HEAP_FRAGMENT_BUDDY  = 0x4,
   HEAP_FRAGMENT_MANUAL = 0x8
};

/**
 * @type  HeapBufferStatus
 * 
 * @brief Enum containing heap buffer status flags.
 *
 * @field HEAP_BUFFER_UNKNOWN
 *        The buffer is unknown and does not belong to the heap
 *        instance or to a specific algorithm within the instance.
 *        Only used for internal communication. 
 *
 * @field HEAP_BUFFER_ENDMARK
 *        The buffer endmark has been damaged.
 *
 * @field HEAP_BUFFER_HEADERS
 *        The buffer header has been damaged.
 *
 * @field HEAP_BUFFER_ALLOCATED
 *        The buffer is allocated to a client.
 *
 * @field HEAP_BUFFER_ANNOTATED
 *        The buffer is annotated with arbitraty metadata.
 *        The annotation most likely contains debug information
 *        such as the file name and line number of the originating
 *        allocation call. 
 */
enum HeapBufferStatusEnum
{
   HEAP_BUFFER_UNKNOWN   = 0x80, 
   HEAP_BUFFER_ENDMARK   = 0x01,
   HEAP_BUFFER_HEADERS   = 0x02,
   HEAP_BUFFER_ALLOCATED = 0x04,
   HEAP_BUFFER_ANNOTATED = 0x08
};


/**
 * @type  HeapFragmentStatus
 *
 * @brief Enum containing heap fragment status flags. 
 *
 * @field HEAP_FRAGMENT_UNKNOWN
 *        The fragment is unknown and does not belong to the heap
 *        instance or to a specific algorithm within the instance. 
 *        Only used for internal communication.
 *
 * @field HEAP_FRAGMENT_ALLOCATED
 *        The fragment is allocated to an algorithm.
 *
 * @field HEAP_FRAGMENT_HEADERS
 *        The fragment header has been damaged. 
 */
enum HeapFragmentStatusEnum
{
   HEAP_FRAGMENT_UNKNOWN   = 0x80,
   HEAP_FRAGMENT_ALLOCATED = 0x01,
   HEAP_FRAGMENT_HEADERS   = 0x02
};


/**
 * @type  HeapLockType
 *
 * @brief Enum containing concurrent lock type identifiers. 
 *
 * @field EHEAP_LOCK_NONE
 *        No lock. 
 *
 * @field EHEAP_LOCK_INT
 *        Interrupt locks.
 *
 * @field EHEAP_LOCK_SPIN
 *        Interrupt and spinlocks.
 *
 * @field EHEAP_LOCK_MUTEX
 *        Mutex lock. 
 *
 */
enum HeapLockTypeEnum
{
   EHEAP_LOCK_NONE  = 0x00,
   EHEAP_LOCK_INT   = 0x01,
   EHEAP_LOCK_SPIN  = 0x02,
   EHEAP_LOCK_MUTEX = 0x03
};


/**
 * @type  HeapStatus
 *
 * @brief Enum containing heap error constants. 
 *
 * @field HEAP_SUCCESS
 *        The operation was successful. 
 *
 * @field HEAP_ENOT_FOUND
 *        The heap instance was not found. 
 * 
 * @field HEAP_EFRAG_NOT_FOUND
 *        The specified heap fragment was not found. 
 *   
 * @field HEAP_EOUT_OF_MEMORY
 *        The system ran out of memory and the heap request
 *        could not be completed. 
 *        
 * @field HEAP_ELOG_MEM_NOT_FOUND
 *        A configured logical memory type does not exist.
 * 
 * @field HEAP_EPHYS_MEM_NOT_FOUND 
 *        A configured physical memory type does not exist.
 *
 * @field HEAP_EPLUG_IN_NOT_FOUND
 *        A configured plug-in has not been linked with the kernel image.
 *        A weak reference to the plug-in installation function is unresolved.
 * 
 * @field HEAP_EABORTED
 *        An event was aborted by an event hook. 
 * 
 * @field HEAP_EILLEGAL_MAP_TYPE
 *        A configured logical memory type has illegal mapping type.
 *        Example: A global heap in MAS.
 *        
 * @field HEAP_EILLEGAL_ALGORITHM
 *        Two or more small buffer algorithms have been enabled. 
 *        
 * @field HEAP_EILLEGAL_VALUE
 *        A configuration value was illegal. For example
 *        used when size strings are not numeric. 
 */
/* WARNING: Values must be identical to those found in heap.sig. */
enum HeapStatusEnum
{
   HEAP_SUCCESS             = 0x00,
   HEAP_ENOT_FOUND          = 0x02,
   HEAP_EFRAG_NOT_FOUND     = 0x04,
   HEAP_EOUT_OF_MEMORY      = 0x05,
   
   HEAP_ELOG_MEM_NOT_FOUND  = 0x11,
   HEAP_EPHYS_MEM_NOT_FOUND = 0x12,
   HEAP_EPLUG_IN_NOT_FOUND  = 0x13,
   HEAP_EABORTED            = 0x14,
   HEAP_EILLEGAL_MAP_TYPE   = 0x15,
   HEAP_EILLEGAL_ALGORITHM  = 0x16,
   HEAP_EILLEGAL_VALUE      = 0x17
};


/**
 * @type  HeapEventID
 *
 * @brief Enum containing heap event identifiers.
 *
 * @long  It is possible to subscribe to different heap events. When an
 *        event is triggered, the heap calls an external hook previously
 *        installed. The hook is allowed to modify data and behavior of
 *        the heap and can therefore be used to extend heap functionality.
 *        There are two categories of events, request and result.
 *
 * @long  Request type events are identified through the "REQ" postfix and are
 *        triggered before a new request has been handled, e.g. after a call
 *        to malloc but before the allocation has been carried out. Request
 *        type events can for example be used to deny requests, such as an
 *        allocation from a process with an overcommitted amount of memory. 
 *
 * @long  Result type events are identified through the "RES" postfix and are
 *        triggered when a request has been handled, e.g. when a buffer has
 *        been allocated but not returned to the client. Result type events
 *        can for example be used to gather allocation statistics, such as
 *        number of allocated buffers, their sizes, number of failed
 *        allocation attempts, etc.  
 *
 * @field HEAP_EVENT_ALLOC_REQ
 *    Request type event. Triggered in response to an allocation.
 *    Valid actions for this event are
 *    @link "heap.h" "HEAP_ACTION_CONTINUE",
 *    @link "heap.h" "HEAP_ACTION_INTERCEPT" and 
 *    @link "heap.h" "HEAP_ACTION_ABORT".
 *    
 * @field HEAP_EVENT_ALLOC_RES
 *    Result type event. Triggered in response to a finished allocation. 
 *    Valid actions for this event are
 *    @link "heap.h" "HEAP_ACTION_CONTINUE",
 *    @link "heap.h" "HEAP_ACTION_ABORT" and 
 *    @link "heap.h" "HEAP_ACTION_REPEAT".
 *
 * @field HEAP_EVENT_FREE_REQ
 *    Request type event. Triggered in response to a deallocation. 
 *    Valid actions for this event are
 *    @link "heap.h" "HEAP_ACTION_CONTINUE",
 *    @link "heap.h" "HEAP_ACTION_INTERCEPT" and 
 *    @link "heap.h" "HEAP_ACTION_ABORT".
 *
 * @field HEAP_EVENT_FREE_RES
 *    Result type event. Triggered in response to a finished deallocation. 
 *    Valid actions for this event are
 *    @link "heap.h" "HEAP_ACTION_CONTINUE",
 *    @link "heap.h" "HEAP_ACTION_ABORT" and 
 *    @link "heap.h" "HEAP_ACTION_REPEAT".
 *
 * @field HEAP_EVENT_RESIZE_REQ
 *    Request type event. Triggered in response to a reallocation.
 *    The HEAP_EVENT_RESIZE_REQ event may be followed by HEAP_EVENT_ALLOC
 *    and HEAP_EVENT_FREE events before HEAP_EVENT_RESIZE_RES, if the buffer
 *    could not be reused.
 *    Valid actions for this event are
 *    @link "heap.h" "HEAP_ACTION_CONTINUE",
 *    @link "heap.h" "HEAP_ACTION_INTERCEPT" and 
 *    @link "heap.h" "HEAP_ACTION_ABORT".
 *
 * @field HEAP_EVENT_RESIZE_RES
 *    Result type event. Triggered in response to a finished reallocation. 
 *    Valid actions for this event are
 *    @link "heap.h" "HEAP_ACTION_CONTINUE",
 *    @link "heap.h" "HEAP_ACTION_ABORT" and 
 *    @link "heap.h" "HEAP_ACTION_REPEAT".
 *
 * @field HEAP_EVENT_COLLECT_REQ
 *    Request type event. Triggered in response to a garbage collect. 
 *    been made.
 *    Valid actions for this event are
 *    @link "heap.h" "HEAP_ACTION_CONTINUE" and
 *    @link "heap.h" "HEAP_ACTION_INTERCEPT".
 *
 * @field HEAP_EVENT_COLLECT_RES
 *    Result type event. Triggered in response to a finished garbage collect. 
 *    Valid action for this event is
 *    @link "heap.h" "HEAP_ACTION_CONTINUE".
 *
 * @field HEAP_EVENT_CONTROL_REQ
 *    Request type event. Triggered in response to a reconfiguration.
 *    Valid actions for this event are
 *    @link "heap.h" "HEAP_ACTION_CONTINUE",
 *    @link "heap.h" "HEAP_ACTION_INTERCEPT" and 
 *    @link "heap.h" "HEAP_ACTION_ABORT".
 *
 * @field HEAP_EVENT_CONTROL_RES
 *    Result type event. Triggered in response to a finished reconfiguration.
 *    Valid actions for this event are
 *    @link "heap.h" "HEAP_ACTION_CONTINUE" and
 *    @link "heap.h" "HEAP_ACTION_ABORT".
 *    
 * @field HEAP_EVENT_QUERY_REQ
 *    Request type event. Triggered in response to a query. 
 *    Valid actions for this event are
 *    @link "heap.h" "HEAP_ACTION_CONTINUE",
 *    @link "heap.h" "HEAP_ACTION_INTERCEPT" and 
 *    @link "heap.h" "HEAP_ACTION_ABORT".
 *
 * @field HEAP_EVENT_QUERY_RES
 *    Result type event. Triggered in response to a finished query. 
 *    Valid actions for this event are
 *    @link "heap.h" "HEAP_ACTION_CONTINUE" and
 *    @link "heap.h" "HEAP_ACTION_ABORT".
 *
 * @field HEAP_EVENT_BREAK_REQ
 *    Request type event. Triggered in response to an allocation of external
 *    memory. 
 *    Valid actions for this event are
 *    @link "heap.h" "HEAP_ACTION_CONTINUE",
 *    @link "heap.h" "HEAP_ACTION_INTERCEPT" and 
 *    @link "heap.h" "HEAP_ACTION_ABORT".
 *
 * @field HEAP_EVENT_BREAK_RES
 *    Result type event. Triggered in response to a finished allocation of
 *    external memory. 
 *    Valid actions for this event are
 *    @link "heap.h" "HEAP_ACTION_CONTINUE",
 *    @link "heap.h" "HEAP_ACTION_ABORT",
 *    @link "heap.h" "HEAP_ACTION_REPEAT".
 *    
 * @field HEAP_EVENT_UNBREAK_REQ
 *    Request type event. Triggered in response to a deallocation of external
 *    memory. 
 *    Valid actions for this event are
 *    @link "heap.h" "HEAP_ACTION_CONTINUE" and
 *    @link "heap.h" "HEAP_ACTION_INTERCEPT".
 *
 * @field HEAP_EVENT_UNBREAK_RES
 *    Result type event. Triggered in response to a finished deallocation of
 *    external memory. 
 *    Valid actions for this event are
 *    @link "heap.h" "HEAP_ACTION_CONTINUE",
 *    @link "heap.h" "HEAP_ACTION_ABORT",
 *    @link "heap.h" "HEAP_ACTION_REPEAT".
 *
 * @field HEAP_EVENT_DESTROY_RES
 *    Result type event. Triggered when the heap is destroyed. Allows
 *    all modules and their hooks to cleanup any allocated resources.
 *    Valid action for this event is
 *    @link "heap.h" "HEAP_ACTION_CONTINUE".
 *
 * @field HEAP_EVENT_COUNT
 *    The total number of event types. 
 *    
 * @seealso HeapHook
 * @seealso HeapAction
 * 
 */
enum HeapEventIDEnum
{
   HEAP_EVENT_ALLOC_REQ      = 0x00,
   HEAP_EVENT_ALLOC_RES      = 0x01,
   HEAP_EVENT_FREE_REQ       = 0x02,
   HEAP_EVENT_FREE_RES       = 0x03,
   HEAP_EVENT_RESIZE_REQ     = 0x04,
   HEAP_EVENT_RESIZE_RES     = 0x05,
   HEAP_EVENT_COLLECT_REQ    = 0x06,
   HEAP_EVENT_COLLECT_RES    = 0x07,
   HEAP_EVENT_CONTROL_REQ    = 0x08,
   HEAP_EVENT_CONTROL_RES    = 0x09,
   HEAP_EVENT_QUERY_REQ      = 0x0A,
   HEAP_EVENT_QUERY_RES      = 0x0B,
   HEAP_EVENT_BREAK_REQ      = 0x0C,
   HEAP_EVENT_BREAK_RES      = 0x0D,
   HEAP_EVENT_UNBREAK_REQ    = 0x0E,
   HEAP_EVENT_UNBREAK_RES    = 0x0F,
   HEAP_EVENT_DESTROY_RES    = 0x10,
   HEAP_EVENT_ERROR_RES      = 0x11,
   HEAP_EVENT_COUNT          = 0x12
};


/**
 * @type  HeapAction
 *
 * @brief Enum containing heap action identifiers.
 *
 * @long  Action identifiers are returned by event hooks to indicate
 *        what action the hook expects the heap to take next. Actions
 *        control the behavior of a specific heap request.
 *
 * @long  All action are not valid for each type of event. For details
 *        regarding usage see @link "heap.h" "HeapEventID" for a description
 *        of available events. If an invalid action is used for a specific
 *        event, an error is generated. 
 *
 * @field HEAP_ACTION_CONTINUE
 *    Continue execution as if the event hook had not been called.
 *    The next hook in the list will be called. When the last hook
 *    has been executed, the original request is handled or finished.
 *    This action can be used for both request and result type events. 
 *
 * @field HEAP_ACTION_INTERCEPT
 *    Terminate execution with a valid result. Used by a hook that has 
 *    fulfilled a request without the assistance of the heap. No futher
 *    hooks will be called and the heap will return the result to the
 *    caller. This action can only be used for request type event.
 *
 * @field HEAP_ACTION_ABORT
 *    Terminate execution without a valid result. Used by a hook to indicate
 *    that the request or the result was unacceptable. No futher hooks will
 *    be called and the heap will return to the caller, indicating an error.
 *    This action is valid for both request and result type events. 
 *
 * @field HEAP_ACTION_REPEAT
 *    Repeat request. Used by a hook to force the heap to retry the
 *    request. No further hooks for the active event will be called. Once
 *    the request is retried, new events will be triggered and invoke the
 *    hooks. This action is only valid for result type events.
 * 
 */
enum HeapActionEnum
{
   HEAP_ACTION_CONTINUE      = 0x01,
   HEAP_ACTION_INTERCEPT     = 0x02,
   HEAP_ACTION_ABORT         = 0x03,
   HEAP_ACTION_REPEAT        = 0x04
};


/**
 * @type  HeapPriority
 *
 * @field HEAP_PRIO_FILTER_HIGH
 *        High priority filter. 
 *        
 * @field HEAP_PRIO_FILTER_NORMAL
 *        Normal priority filter. 
 *
 * @field HEAP_PRIO_FILTER_LOW   
 *        Low priority filter. 
 *
 * @field HEAP_PRIO_ALGO_HIGH    
 *        High priority algorithm. 
 *
 * @field HEAP_PRIO_ALGO_NORMAL  
 *        Normal priority algorithm. 
 *
 * @field HEAP_PRIO_ALGO_LOW     
 *        Low priority algorithm. 
 *
 * @field HEAP_PRIO_ALLOC_HIGH   
 *        High priority allocator. 
 *
 * @field HEAP_PRIO_ALLOC_NORMAL 
 *        Normal priority allocator. 
 *
 * @field HEAP_PRIO_ALLOC_LOW    
 *        Low priority allocator. 
 *
 * @field HEAP_PRIO_USER         
 *        Offset used to create new priority levels from existing ones.
 *        Examples:
 *          HEAP_PRIO_FILTER_HIGH + HEAP_PRIO_USER
 *          HEAP_PRIO_FILTER_HIGH - HEAP_PRIO_USER
 *          HEAP_PRIO_FILTER_HIGH + HEAP_PRIO_USER * 2
 *          
 */
enum HeapPriorityEnum
{
   HEAP_PRIO_FILTER_HIGH     = 0x1000,
   HEAP_PRIO_FILTER_NORMAL   = 0x2000,
   HEAP_PRIO_FILTER_LOW      = 0x3000,
   HEAP_PRIO_ALGO_HIGH       = 0x4000,
   HEAP_PRIO_ALGO_NORMAL     = 0x5000,
   HEAP_PRIO_ALGO_LOW        = 0x6000,
   HEAP_PRIO_ALLOC_HIGH      = 0x7000,
   HEAP_PRIO_ALLOC_NORMAL    = 0x8000,
   HEAP_PRIO_ALLOC_LOW       = 0x9000,
   HEAP_PRIO_USER            = 0x20
};


/**
 * @type HeapStatistics
 *
 * @brief Collection of the most common and generic statistics.
 *
 * @long  The integration layer will allocate and initialize
 *        an instance of this structure. A pointer to the instance
 *        is provided to plug-ins during installation.
 *
 * @long  It is the responsibility of installed plug-ins to update
 *        any field that is affected by a successful operation,
 *        such as a buffer allocation.
 *
 * @field buf_size_used_hi
 *           Peak value of buf_size_used. 
 * 
 * @field buf_size_used
 *           The total amount of used buffer memory.
 *           Includes internal fragmentation and administrative overhead.
 *
 * @field buf_size_req
 *           The total amount of requested buffer memory.
 *           Excludes internal fragmentation and adminstrative overhead. 
 *
 * @field buf_size_free
 *           The total amount of free buffer memory.
 *
 * @field buf_num_shared
 *           Total number of allocated shared buffers.
 *
 * @field buf_num_private
 *           Total number of allocated private buffers.
 *
 * @field buf_num_used
 *           Buffer histogram with the number of allocated
 *           buffers in a specific size interval. There
 *           can be 32 different intervals.
 *
 * @field buf_num_free
 *           Buffer histogram with the number of free
 *           buffers in a specific size interval. There
 *           can be 32 different intervals.
 *
 * @field frag_size_free
 *           The total amount of free fragment memory.
 *
 * @field frag_size_used
 *           The total amount of used fragment memory.
 *
 * @field frag_size_req
 *           The total amount of requested fragment memory.
 *
 * @field meta_size_used
 *           The total amount of memory used for administration data.
 *           Shall be updated by integration layer at each call to
 *           heap_meta_alloc() and heap_meta_free().
 * 
 */
struct HeapStatistics
{
   size_t    buf_size_used_hi;
   size_t    buf_size_used;
   size_t    buf_size_req;
   size_t    buf_size_free;
   uint32_t  buf_num_shared;
   uint32_t  buf_num_private;
   uint32_t  buf_num_used[32];
   uint32_t  buf_num_free[32];

   size_t    frag_size_free;
   size_t    frag_size_used;
   size_t    frag_size_req;

   size_t    meta_size_used;
};


/**
 * @type  HeapConfiguration
 *
 * @brief Collection of the most common and generic configuration.
 *
 * @long  The integration layer will allocate and initialize
 *        an instance of this structure. A pointer to the instance
 *        is provided to plug-ins during installation.
 *
 * @long  These fields define limits that must be enforced by
 *        plug-ins. Plug-ins must test statistics defined
 *        in HeapStatistics against fields in this structure to
 *        determine if an operation is allowed or not. 
 *
 * @field frag_size_max
 *           The maximum amount of used fragment memory allowed. Shall
 *           be enforced by allocators. Example:            
 *           HeapStatistics.frag_size_used < HeapConfiguration.frag_size_used
 * 
 * @field frag_size_init
 *           The size of the initial fragment. Shall be used by algorithms
 *           during the first zzheap_break request.
 * 
 * @field frag_size_default
 *           The default size of extension fragments. The
 *           size used during a break request may be both
 *           smaller and larger, depending on requirements
 *           and available space. 

 * @field frag_check
 *           Enables buffer pointer validation. Algorithms are
 *           required to validate that freed buffer pointers
 *           are inside allocated heap fragments.  
 *
 * @field frag_mas
 *           Indicates that heap fragments may have the MAS mapping type. 
 *
 * @field frag_reclaim
 *           Enables deallocation of unused fragments. If set,
 *           algorithms are obliged to zzheap_unbreak any
 *           fragment that is completely unused. 
 *
 * @field max_type
 *           There are two different methods to limit the total heap size.
 *           1) Size of allocated fragments may not exceed limit.
 *           2) Size of allocated buffers may not exceed limit.
 *           Primarily used by integration layer to initialize
 *           buf_size_max and frag_size_max. 

 * @field buf_size_threshold
 *           Large buffer threshold. The installed small buffer algorithm
 *           shall ignore allocations where the requested buffer size
 *           exceeds this limit. 
 *
 * @field buf_size_max      
 *           The maximum amount of used buffer memory allowed. Shall
 *           be enforced by algorithms. Example:
 *           HeapStatistics.buf_size_used < HeapConfiguration.buf_size_used
 *
 * @field buf_size_endmark
 *
 */
struct HeapConfiguration
{
   size_t    frag_size_max;        
   size_t    frag_size_init;
   size_t    frag_size_default;

   uint8_t   frag_check;
   uint8_t   frag_mas;
   uint8_t   frag_reclaim;
   uint8_t   max_type;

   size_t    buf_size_threshold;   
   size_t    buf_size_max;         
   size_t    buf_size_endmark;
};


/**
 * @type  HeapHook
 *
 * @brief Hook prototype for heap events.
 */
typedef HeapAction
(*HeapHook)(struct Heap *heap, union HeapEvent *event, void *opaque);


/**
 * @type  HeapAllocEvent
 *
 * @brief Data for heap allocation events.
 *
 * @long  This structure is passed to event hooks during execution of
 *        the HEAP_EVENT_ALLOC_REQ and HEAP_EVENT_ALLOC_RES event types.
 *
 * @field id       Identifies the type of event.
 * 
 * @field ptr      Pointer to a buffer. Initially set to NULL before execution
 *                 of the request type event. To be filled in by an algorithm
 *                 and later evaluated during the result type event. 
 *
 * @field size     Requested size of the buffer when passed to request events.
 *                 A hook may change the value to grow or shrink the buffer
 *                 before it is allocated. Once a buffer has been allocated,
 *                 the field is updated with the actual size of the buffer
 *                 before being passed to result events.
 *
 * @field owner    Owner of the buffer. A hook may change the field value to a
 *                 different process identifier or 0 to make the buffer shared. 
 *
 * @field params   Optional heap parameters that are to be considered,
 *                 see heap_param.h.
 *
 * @field file     Name of file where request originated.
 *
 * @field line     Line number in file where request originated. 
 */
struct HeapAllocEvent
{
   HeapEventID   id;
   HeapPID       owner;
   HeapParam    *params;
   void         *ptr;
   size_t        size;
   const char   *file;
   int           line;   
};


/**
 * @type  HeapFreeEvent
 *
 * @brief Data for heap dealloction events.
 *
 * @long  This structure is passed to event hooks during execution of
 *        the HEAP_EVENT_FREE_REQ and HEAP_EVENT_FREE_RES event types. 
 *
 * @field id       Identifies the type of event.
 * 
 * @field ptr      Pointer to a buffer. Initially set to point to the
 *                 deallocation victim before execution of the request
 *                 type event. Must be set to NULL by the algorithm that
 *                 carries out the deallocation. 
 *
 * @field owner    Operation originator. A hook may change the field value
 *                 to a different process identifier or 0 to make operation
 *                 forced. An algorithm shall verify that the owner field
 *                 matches the owner of the buffer. 
 *
 * @field params   Optional heap parameters that are to be considered,
 *                 see heap_param.h.
 *
 * @field file     Name of file where request originated.
 *
 * @field line     Line number in file where request originated.
 *
 * @field size     Actual size of freed buffer.
 */
struct HeapFreeEvent
{
   HeapEventID   id;
   HeapPID       owner;
   HeapParam    *params;
   void         *ptr;
   size_t        size;
   const char   *file;
   int           line;   
};


/**
 * @type  HeapResizeEvent
 *
 * @brief Data for heap buffer resize events.
 *
 * @long  This structure is passed to event hooks during execution of
 *        the HEAP_EVENT_RESIZE_REQ and HEAP_EVENT_RESIZE_RES event types. 
 *
 * @field id       Identifies the type of event.
 * 
 * @field ptr      Pointer to a buffer. Initially set to point to the
 *                 buffer to be resized before execution of the request
 *                 type event.
 *
 *                 Must be set to NULL by the algorithm that
 *                 manages the buffer if the buffer cannot be resized,
 *                 otherwise the pointer values should be retained.  
 *
 * @field size     New size of the buffer. A hook may change the field
 *                 value to grow or shrink the buffer before the resize
 *                 operation is performed. 
 *
 * @field prev_size Previous size of the buffer. Set before execution of
 *                 result type events. 
 *
 * @field owner    Operation originator. A hook may change the field value
 *                 to a different process identifier or 0 to make operation
 *                 forced. An algorithm shall verify that the owner field
 *                 matches the owner of the buffer. 
 *
 * @field params   Optional heap parameters that are to be considered,
 *                 see heap_param.h.
 *
 * @field file     Name of file where request originated.
 *
 * @field line     Line number in file where request originated.
 */
struct HeapResizeEvent
{
   HeapEventID   id;
   HeapPID       owner;
   HeapParam    *params;
   void         *ptr;
   size_t        size;
   size_t        size_prev;
   const char   *file;
   int           line;   
};


/**
 * @type  HeapCollectEvent
 *
 * @brief Data for heap garbage collection events.
 *
 * @long  This structure is passed to event hooks during execution of
 *        the HEAP_EVENT_COLLECT_REQ and HEAP_EVENT_COLLECT_RES event types.
 *
 * @field id       Identifies the type of event.
 *
 * @field owner    Identifies the process that has terminated and
 *                 for which garbage collection is to be carried out. 
 */
struct HeapCollectEvent
{
   HeapEventID   id;
   HeapPID       owner;
};


/**
 * @type  HeapControlEvent
 *
 * @brief Data for heap control events.
 *
 * @long  This structure is passed to event hooks during execution of
 *        the HEAP_EVENT_CONTROL_REQ and HEAP_EVENT_CONTROL_RES event types.
 *
 * @field id       Identifies the type of event.
 *
 * @field params   Heap parameters that are to be controlled,
 *                 see heap_param.h.
 */
struct HeapControlEvent
{
   HeapEventID   id;
   HeapParam    *params;
};


/**
 * @type  HeapQueryEvent
 *
 * @brief Data for heap query events.
 *
 * @long  This structure is passed to event hooks during execution of
 *        the HEAP_EVENT_QUERY_REQ and HEAP_EVENT_QUERY_RES event types.
 *
 * @field id       Identifies the type of event.
 *
 * @field params   Heap parameters that are to be queried,
 *                 see heap_param.h.
 *
 * @field freezed  Set to non-zero if event processing is executed
 *                 in a freeze-mode context. If set, the heap lock must
 *                 not be acquired by plug-ins since that may cause a
 *                 deadlock.
 */
struct HeapQueryEvent
{
   HeapEventID   id;
   HeapParam    *params;
   uint32_t      freezed;
};



/**
 * @type  HeapBreakEvent
 *
 * @brief Data for heap fragment allocation/break events.
 *
 * @long  This structure is passed to event hooks during execution of
 *        the HEAP_EVENT_BREAK_REQ and HEAP_EVENT_BREAK_RES event types.
 * 
 * @field id       Identifies the type of event.
 *
 * @field params   Optional heap parameters that are to be considered,
 *                 see heap_param.h.
 *
 * @field base     Pointer to fragment, result of event. Initially set to NULL
 *                 before execution of the request type event. To be filled in
 *                 by an allocator and later evaluated during the result type
 *                 event. 
 *              
 * @field min_size Minimum size of fragment. The allocated fragment cannot be
 *                 smaller than this size. Event hooks may change the value
 *                 before the fragment is allocated to affect the size. Once
 *                 a fragment is allocated, the allocator must change this
 *                 value to the actual size of the fragment. 
 *
 * @field req_size Desired size of fragment. This is the size that the
 *                 requesting algorithm would prefer. The allocated
 *                 fragment may be smaller or larger. Should not be modified
 *                 by event hooks. 
 *
 * @field max_size Maximum size of fragment. This is the maximum size of a
 *                 fragment that the requesting algorithm will be able to
 *                 handle. Event hooks may change the value before the fragment
 *                 is allocated to affect the size. If the allocation fails due
 *                 to exhaustion (resulting amount of fragment memory would reach
 *                 configured maximum limit), the allocator must change this
 *                 value to a size that is possible to allocate. Otherwise if
 *                 the allocation is successful, this value must be set to
 *                 the size of the allocated fragment. If the size is 0,
 *                 there is no upper limit.
 *
 * @field algo_id  Algorithm identifier. Used to track usage of fragments
 *                 within different algorithms. Allocator shall set the
 *                 mandatory algo_id field of the fragment header to this
 *                 value.
 *
 * @field status   Resulting status of the operation. Should be set by
 *                 the allocator after allocation to indicate success or
 *                 failure. Use 0 for success and any other values for
 *                 failure.
 */
struct HeapBreakEvent
{
   HeapEventID   id;
   HeapParam    *params;
   void         *base;
   size_t        min_size;
   size_t        req_size;
   size_t        max_size;
   uint32_t      algo_id;
   uint32_t      status;
};



/**
 * @type  HeapUnbreakEvent
 *
 * @brief Data for heap fragment deallocation/unbreak events.
 *
 * @long  This structure is passed to event hooks during execution of
 *        the HEAP_EVENT_UNBREAK_REQ and HEAP_EVENT_UNBREAK_RES event types.
 * 
 * @field id       Identifies the type of event.
 *
 * @field params   Optional heap parameters that are to be considered,
 *                 see heap_param.h.
 *
 * @field base     Pointer to fragment to be deallocated. Must be set to
 *                 NULL by the allocator that carries out the deallocation. 

 */
struct HeapUnbreakEvent
{
   HeapEventID   id;
   HeapParam    *params;
   void         *base;
};



/**
 * @type  HeapErrorEvent
 *
 * @brief Data for heap error events.
 *
 * @long  This structure is passed to event hooks during execution of
 *        the HEAP_EVENT_ERROR_RES event type.
 * 
 * @field id       Identifies the type of event.
 *
 * @field params   Optional heap parameters that are to be considered,
 *                 see heap_param.h.
 *
 * @field ecode    Reported heap error code. See heap_error.h. 
 *                 
 * @field data     Specific data that is associated with the error, for
 *                 example the address of a damaged buffer. 
 */
struct HeapErrorEvent
{
   HeapEventID   id;
   HeapParam    *params;
   HeapError     ecode;
   void         *data;
};


/**
 * @type  HeapEvent
 *
 * @brief Union of all heap event data types.
 *
 * @long  This union is passed to event hooks during event execution.
 *        The id field may be used to access the correct member field. 
 * 
 */
union HeapEvent
{
   HeapEventID              id;
   struct HeapAllocEvent    e_alloc;
   struct HeapFreeEvent     e_free;
   struct HeapResizeEvent   e_resize;
   struct HeapCollectEvent  e_collect;
   struct HeapControlEvent  e_control;
   struct HeapQueryEvent    e_query;
   struct HeapBreakEvent    e_break;
   struct HeapUnbreakEvent  e_unbreak;
   struct HeapErrorEvent    e_error;
};


/**
 * @type  Heap
 *
 * @brief The representation of a heap framework instance.
 *
 * @field event_hooks   Array with lists. Each list represents a type of event
 *                      and contains all event hooks attached to it.
 *
 * @field freezed       The heap instance is being accessed from freeze mode.
 *                      The heap lock interface should be disabled if non-zero.
 *
 * @field lock_type     The type of locks used by the instance.
 *
 * @field lock          Instantiation of the lock.
 *                      Defined and implemented by the integration layer.
 */
struct Heap
{
   list_t       event_hooks[HEAP_EVENT_COUNT];

   uint32_t     freezed;

   HeapLockType lock_type;
   struct {
      HeapSpinLock  spin;
      HeapMutexLock mutex;
   } lock;

   HeapPID       brkproc;
   HeapMutexLock brklock;
};

#include "heap_lock.h"


/*
 *************************************************************************
 * Heap Macros
 *************************************************************************
 */

#define HEAP_ALIGNMENT		8  /* bytes */
#define HEAP_ENDMARK            0x12

/*
 *************************************************************************
 * Heap Function Interface
 *************************************************************************
 */


/**
 * @function zzheap_create
 *
 * @brief    Initializes a new heap instance. 
 *
 * @long     The heap instance created is an empty shell; all algorithms,
 *           allocators and filters must be manually created and attached
 *           to the instance before it is usable.
 *
 * @long     The function is called by the integration layer which must
 *           allocate the memory occupied by the instance. The function
 *           does not generate any event. 
 *
 * @input    Pointer to the heap instance to initialize. 
 *           
 * @return   HEAP_SUCCESS.
 * 
 * @seealso  zzheap_destroy
 * 
 */
HeapStatus
zzheap_create(struct Heap *heap);

/**
 * @function zzheap_destroy
 *
 * @brief    Destroys a heap instance.
 *
 * @long     The function generates a destroy event within the heap instance.
 *           All algorithms, allocators and filters installed that subscribe
 *           to the event will be called to release any resources allocated
 *           for use with the instance.
 *
 * @long     The function does not destroy the heap instance object itself.
 *           It is the responsibility of the integration layer to first call
 *           this function to cleanup resources installed within the object
 *           and then to deallocate the object itself.
 *           
 * @input    heap   The heap instance to be destroyed.
 *
 * @seealso  zzheap_create
 */
void
zzheap_destroy(struct Heap *heap);


/**
 * @function zzheap_attach_hook
 *
 * @brief    Attaches an event hook to a heap instance. 
 *
 * @long     A hook is a function called in response to an event within a
 *           heap. Such events include calls to heap allocation methods,
 *           heap extensions, garbage collection, etc.
 *           For a complete list of events see @link "heap.h" "HeapEvent".
 *
 * @long     Multiple hooks can be attached to a single event. If the same hook
 *           is installed multiple times for a single event, 
 *           subsequent invokations of the hook during the event can be
 *           distinguished through the use of hook defined data. The data is
 *           registered along with the hook when it is attached to the event. 
 *           The order in which hooks are called can be controlled when
 *           the hook is installed.
 *
 * @long     Hooks can take different actions and control the continued
 *           flow of execution. For more information, see
 *           @link "heap.h" "HeapAction" and
 *           @link "heap.h" "HeapEventID" for details on valid actions
 *           for specific events.  
 *
 * @input    heap       The heap instance to attach a hook to.
 *
 * @input    event_id   Type of event to attach to.
 *
 * @input    hook       Function to call when the event is triggered.
 *
 * @input    opaque     Hook defined data to be passed to the hook
 *                      when an event occurs. 
 *
 * @input    prio       Hook priority. Determines the order in
 *                      which hooks are called. See @link "" "HeapPriority".
 *
 * @return   HEAP_SUCCESS on success @break
 *           HEAP_EOUT_OF_MEMORY if there was insufficient memory @break
 * 
 * @seealso  HeapEventID
 * @seealso  HeapEvent
 * @seealso  HeapHook
 * @seealso  HeapPriority
 * @seealso  zzheap_detach_hook
 * 
 */
HeapStatus
zzheap_attach_hook(struct Heap *heap,
                   HeapEventID event_id, HeapHook hook, void *opaque,
                   int prio);

/**
 * @function zzheap_detach_hook
 *
 * @brief    Detaches an event hook from a heap instance. 
 *
 * @long     When notification of an event no longer is desired
 *           the hook can be detached. The hook will no longer be called
 *           when the event is triggered.
 *
 * @long     The hook is only removed for the specific event indicated. If
 *           the hook is also registered for other types of events it will
 *           continue to be called when those events are triggered.
 *
 * @long     It is illegal to call this function from inside an event hook,
 *           unless the type of event to detach from differs from the
 *           executing type of event. A hook cannot detach itself. 
 *           Attempts to remove non-existing hooks are ignored.
 *
 * @input    heap       The heap instance to detach a hook from.
 *
 * @input    event_id   Type of event to detach from.
 *
 * @input    hook       Function not to call when the event is triggered.
 *
 * @input    opaque     Hook defined data that was passed to the hook
 *                      when an event occured. 
 *
 * @seealso  zzheap_attach_hook
 * 
 */
void
zzheap_detach_hook(struct Heap *heap,
                   HeapEventID event_id, HeapHook hook, void *opaque);


/**
 * @function zzheap_alloc
 *
 * @brief    Allocate a heap buffer.
 *
 * @long     The function generates two events:
 *           HEAP_EVENT_ALLOC_REQ and
 *           HEAP_EVENT_ALLOC_RES
 *           
 * @input    heap       The heap instance.
 * @input    size       Requested buffer payload size.
 * @input    pid        Requested buffer owner, or 0 for shared buffers.
 * @input    params     Array of allocation parameter IDs and values.
 * @input    file       Optional filename where call originates.
 * @input    line       Optional line number where call originates.
 * 
 * @return   Pointer to a newly allocated buffer, or
 *           NULL if the allocation failed.
 * 
 * @seealso  zzheap_free
 * @seealso  zzheap_realloc
 * 
 */
void *
zzheap_alloc(struct Heap *heap,
             size_t size, HeapPID pid, HeapParam *params,
             const char *file, int line);

/**
 * @function zzheap_realloc
 *
 * @brief    Allocate/resize/free a heap buffer. 
 *
 * @long     The function generates two events:
 *           HEAP_EVENT_RESIZE_REQ and
 *           HEAP_EVENT_RESIZE_RES
 *
 * @input    heap       The heap instance.
 * @input    ptr        Pointer to a previously allocated buffer to resize, or
 *                      NULL to allocate a new buffer. 
 * @input    size       Requested buffer payload size, or
 *                      0 to free the buffer.
 * @input    pid        Requested buffer owner, or 0 for shared buffers.
 * @input    params     Array of allocation parameter IDs and values.
 * @input    resize     Set to non-zero to force the buffer to be resized
 *                      within its current bounds. The buffer data is not
 *                      copied to a new buffer. 
 * @input    file       Optional filename where call originates.
 * @input    line       Optional line number where call originates.
 * 
 * @return   Pointer to a resized or allocated buffer. NULL is returned if
 *           the buffer could not be resized or a new buffer could not be
 *           allocated. NULL is also returned when the buffer is freed. 
 * 
 * @seealso  zzheap_alloc
 * @seealso  zzheap_free
 */
void *
zzheap_realloc(struct Heap *heap,
               void *ptr, size_t size, HeapPID pid, HeapParam *params,
               uint32_t resize, const char *file, int line);

/**
 * @function zzheap_free
 *
 * @brief    Free a heap buffer. 
 *
 * @long     The function generates two events:
 *           HEAP_EVENT_FREE_REQ and
 *           HEAP_EVENT_FREE_RES
 *
 * @input    heap       The heap instance.
 * @input    ptr        Pointer to a previously allocated buffer to free. 
 * @input    pid        Requested buffer owner, or 0 for shared buffers.
 * @input    params     Array of allocation parameter IDs and values.
 * @input    file       Optional filename where call originates.
 * @input    line       Optional line number where call originates.
 * 
 * @seealso  zzheap_free
 * @seealso  zzheap_realloc
 * 
 */
void
zzheap_free(struct Heap *heap,
            void *ptr, HeapPID pid, HeapParam *params,
            const char *file, int line);

/**
 * @function zzheap_collect
 *
 * @brief    Perform a buffer garbage collect. 
 *
 * @long     The function generates two events:
 *           HEAP_EVENT_COLLECT_REQ and
 *           HEAP_EVENT_COLLECT_RES
 *
 * @input    heap       The heap instance. 
 * @input    owner      Owner of buffers to garbage collect. 
 */
void			 
zzheap_collect(struct Heap *heap, HeapPID owner);

/**
 * @function zzheap_control
 *
 * @brief    Configure a heap and its components. 
 *
 * @long     The function generates two events:
 *           HEAP_EVENT_CONTROL_REQ and
 *           HEAP_EVENT_CONTROL_RES
 *
 * @input    heap      The heap instance. 
 * @input    params    Parameters. 
 *
 * @return   HEAP_SUCCESS on success @break
 *           HEAP_EABORTED if the request was aborted by a hook @break
 */
HeapStatus
zzheap_control(struct Heap *heap, HeapParam *params);


/**
 * @function zzheap_query
 *
 * @brief    Query the heap for information.
 *
 * @long     The function generates two events:
 *           HEAP_EVENT_QUERY_REQ and
 *           HEAP_EVENT_QUERY_RES
 *
 * @input    heap      The heap instance.
 * @input    params    Parameters.
 * @input    fm        Boolean indicating call from freeze-mode context. 
 *
 * @return   HEAP_SUCCESS on success @break
 *           HEAP_EABORTED if the request was aborted by a hook @break
 */
HeapStatus
zzheap_query(struct Heap *heap, HeapParam *params, uint32_t fm);

/**
 * @function zzheap_break
 *
 * @brief    Allocate external heap memory.
 *
 * @long     The function generates two events:
 *           HEAP_EVENT_BREAK_REQ and
 *           HEAP_EVENT_BREAK_RES
 *
 * @input    heap      The heap instance.
 * @input    min_size  Minimum size of requested fragment. 
 *                     See description in @link "" "HeapBreakEvent".
 * @input    req_size  Preferred size of requested fragment.
 *                     Typically HeapConfiguration.frag_size_default.
 * @input    max_size  Maximum size of requested fragment. 
 *                     See description in @link "" "HeapBreakEvent".
 * @input    algo      The ID of the algorithm that requests the fragment.
 * @input    params    Parameters. 
 *
 * @return   Pointer to the base of fragment.
 *
 * @seealso  zzheap_unbreak
 */
void *
zzheap_break(struct Heap *heap,
             size_t *min_size, size_t req_size, size_t *max_size, 
             uint32_t algo, HeapParam *params);

/**
 * @function zzheap_unbreak
 *
 * @brief    Deallocate external heap memory.
 *
 * @long     The function generates two events:
 *           HEAP_EVENT_UNBREAK_REQ and
 *           HEAP_EVENT_UNBREAK_RES
 *
 * @input    heap      The heap instance.
 * @input    base      Base address of fragment to deallocate.
 * @input    params    Parameters.
 *
 * @seealso  zzheap_break
 */
void 
zzheap_unbreak(struct Heap *heap, void *base, HeapParam *params);

/**
 * @function zzheap_error
 *
 * @brief    Report an error.
 *
 * @long     The function generates one event:
 *           HEAP_EVENT_ERROR_RES
 *
 * @input    heap      The heap instance.
 * @input    error     The error identifier. 
 * @input    data      Error specific data.
 */
void 
zzheap_error(struct Heap *heap, HeapError err, void *data);


#endif /* HEAP_H */

