/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2013 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */


#include "ose_heap_error.h"
#include "heap.sig"
#include "heap_err.h"
#include "cpu_access.h"
#include "ose_err.h"
#include "ose_heapd.h"


/**************************************************************************
 * Macros and Definitions
 **************************************************************************/

#ifdef OSMIPS
#define PAGE_ALIGN 0x2000UL
#else
#define PAGE_ALIGN 0x1000UL
#endif
#define ALIGN(size) (((size) + PAGE_ALIGN - 1U) & ~(PAGE_ALIGN-1U))

extern void zzkrn_error2(OSERRCODE ecode, OSERRCODE extra);

/***************************************************************************
 * Helper Functions
 ***************************************************************************/

static void
buf_stats(struct Heap *heap, void *ptr, struct HeapBufferInfoReply *hbi)
{
   HeapParam p[] = {
      HPM(QUERY_BUF_ACT_SIZE,1), 0,
      HPM(QUERY_BUF_REQ_SIZE,1), 0,
      HPM(QUERY_BUF_OWNER,1),    0,
      HPM(QUERY_BUF_STATUS,1),   0,
      HPM_END
   };

   p[1] = p[3] = p[5] = p[7] = (HeapParam)ptr;

   zzheap_query(heap, p, 1);

   hbi->address  = (uintptr_t)ptr;
   hbi->act_size = HPM_IS_HANDLED(p[0]) ? (size_t)p[1] : 0;
   hbi->req_size = HPM_IS_HANDLED(p[2]) ? (size_t)p[3] : 0;
   hbi->owner    = HPM_IS_HANDLED(p[4]) ? (HeapPID)p[5] : 0;

   if (HPM_IS_HANDLED(p[6]))
   {
      /* Endmark reported broken? */
      hbi->endmark_ok = (p[7] & OSE_HEAP_EENDMARK_OVERWRITTEN) ? 0 : 1;
      
      /* handle buffer annotation */
      if (p[7] & HEAP_BUFFER_ANNOTATED)
      {
         struct OseHeap *ose_heap = OSHEAP_FROM_HEAP(heap);
         hbi->address  += ose_heap->hucb.annotation_size;
         hbi->req_size -= ose_heap->hucb.annotation_size;
         hbi->ann_size  = ose_heap->hucb.annotation_size;
         
         ose_heap->hucb.heap_debug_fileline((OSHEAP)(uintptr_t)ose_heap, hbi, 0);
      }
   }
}

/***************************************************************************/

static void
frag_stats(struct Heap *heap, size_t size,
           struct HeapBufferInfoReply *hbi,
           HeapError ecode)
{
   struct OseHeap *ose_heap = OSHEAP_FROM_HEAP(heap);

   if (ose_heap->config.max_type == 1 || ecode == HEAP_ERROR_MEM_EXHAUSTED)
   {
      hbi->req_size = size;
      hbi->act_size = ALIGN(size);
   }
   else
   {
      HeapParam p[] = {HPM(QUERY_BUF_NEXT_SIZE,1),0,HPM_END};
      p[1] = size;
      zzheap_query(heap, p, 1);
      
      hbi->req_size = size;
      hbi->act_size = HPM_IS_HANDLED(p[0]) ? p[1] : size;
      hbi->act_size += 8 /* Standard overhead */ + ose_heap->config.buf_size_endmark;
   }

   ose_heap->hucb.heap_debug_fileline((OSHEAP)(uintptr_t)ose_heap, hbi, 0);
}




/***************************************************************************
 * Private Interface
 ***************************************************************************/

static HeapAction
l_heap_error(struct Heap *heap, union HeapEvent *event, void *opaque)
{
   HeapError ecode  = event->e_error.ecode;
   void     *data   = event->e_error.data;
   OSERRCODE fatal  = (ecode & HEAP_ERROR_FATAL) ? OSE_EFATAL_MASK : 0;
   OSADDRESS locked = 0;
   OSADDRESS brklocked = 0;
   HeapLock  lock;

   struct HeapBufferInfoReply *hbi = (struct HeapBufferInfoReply *)
      alloc(sizeof(struct HeapBufferInfoReply) + HEAP_MAX_FILENAME_LENGTH, 0);

   (void)opaque;
   
   hbi->status     = 0; /* O: Status of the operation. */
   hbi->frag       = 0; /* O: Address of heap fragment. */
   hbi->path       = 0; /* O: Path in tree of buffer. */
   hbi->depth      = 0; /* O: Depth in tree of buffer. */
   hbi->address    = 0; /* O: Address of buffer. */
   hbi->owner      = 0; /* O: Pid of owner or 0 for shared. */
   hbi->file       = ~0U;/* O: File name of allocation, index into data[] or ~0. */
   hbi->line       = 0; /* O: Line number of allocation. */
   hbi->req_size   = 0; /* O: Requested size of buffer. */
   hbi->act_size   = 0; /* O: Actual size of buffer (overhead, internal fragmentation). */
   hbi->ann_size   = 0; /* O: Size of annotation data. Zero if no annotation included. */
   hbi->data_size  = 0; /* O: Number of bytes sampled, first in data[]. */
   hbi->endmark_ok = True; /* O: True if endmark ok, otherwise False. */
   *hbi->data      = 0; /* O: Data sampled from buffer, with space for appended file name. */

   LOCK_SAVE(lock);
   LOCK_RESTORE(lock);

   if (heap_lock_owner(heap) == (current_process() & ~0x80000000))
   {
      locked = 1;

      /* Unlock the heap unconditionally
       * when an error is detected. */
      heap_unlock(heap, &lock, 0);
   }

   if (heap_brklock_owner(heap) == (current_process() & ~0x80000000))
   {
      brklocked = 1;

      /* Unlock the heap unconditionally
       * when an error is detected. */
      heap_brkunlock(heap, &lock);
   }

   /* Replaced error2() with zzkrn_error2()
    * This so we won't lose the frame pointer. */
   switch (ecode & ~HEAP_ERROR_FATAL)
   {
      case HEAP_ERROR_BUF_OWNER:
      {
         buf_stats(heap, data, hbi);

         zzkrn_error2(fatal | OSE_HEAP_EPROCESS_NOT_OWNER, (OSERRCODE)hbi);
         break;
      }
      case HEAP_ERROR_BUF_FREE:
      {
         hbi->address = (uintptr_t)data;
         
         zzkrn_error2(fatal | OSE_HEAP_EBUFFER_FREE, (OSERRCODE)hbi);
         break;
      }
      case HEAP_ERROR_BUF_HEADER:
      {
         hbi->address = (uintptr_t)data;
         
         zzkrn_error2(fatal | OSE_HEAP_EHEAP_CORRUPTED, (OSERRCODE)hbi);
         break;
      }
      case HEAP_ERROR_BUF_UNKNOWN:
      case HEAP_ERROR_BUF_UNALIGNED:
      {
         hbi->address = (uintptr_t)data;
         
         zzkrn_error2(fatal | OSE_HEAP_EPTR_INVALID, (OSERRCODE)hbi);
         break;
      }
      case HEAP_ERROR_BUF_LINKS:
      {
         hbi->address = (uintptr_t)data;
         
         zzkrn_error2(fatal | OSE_HEAP_EFREED_BUFFER_OVERWRITTEN, (OSERRCODE)hbi);
         break;
      }
      case HEAP_ERROR_BUF_ENDMARK:
      {
         buf_stats(heap, data, hbi);
         
         zzkrn_error2(fatal | OSE_HEAP_EENDMARK_OVERWRITTEN, (OSERRCODE)hbi);
         break;
      }
      case HEAP_ERROR_BUF_TOUCHED:
      {
         hbi->address = (uintptr_t)data;
         
         zzkrn_error2(fatal | HEAP_DEBUG_EBUFFER_TOUCHED_AFTER_FREE, (OSERRCODE)hbi);
         break;
      }
      case HEAP_ERROR_MEM_EXHAUSTED:
      {
         frag_stats(heap, (size_t)data, hbi, HEAP_ERROR_MEM_EXHAUSTED);

         zzkrn_error2(fatal | OSE_HEAP_ESYSMEM_ALLOC_FAILED, (OSERRCODE)hbi);
         break;
      }
      case HEAP_ERROR_BUF_EXHAUSTED:
      {
         frag_stats(heap, (size_t)data, hbi, HEAP_ERROR_BUF_EXHAUSTED);

         zzkrn_error2(fatal | OSE_HEAP_EHEAP_EXHAUSTED, (OSERRCODE)hbi);
         break;
      }
      case HEAP_ERROR_CORRUPTED:
      {
         /* It's unspecified what the data points to. It's sometimes fragment */
         hbi->frag = (uintptr_t)data;
         zzkrn_error2(fatal | OSE_HEAP_EHEAP_CORRUPTED, (OSERRCODE)hbi);
         break;
      }
      case HEAP_ERROR_FRAG_HEADER:
      case HEAP_ERROR_FRAG_UNKNOWN:
      case HEAP_ERROR_FRAG_UNALIGNED:
      {
         hbi->frag = (uintptr_t)data;
         zzkrn_error2(fatal | OSE_HEAP_EHEAP_CORRUPTED, (OSERRCODE)hbi);
         break;
      }
      case HEAP_ERROR_NONE:
      default:
      {
         zzkrn_error2(fatal | OSE_HEAP_EOTHER, ecode);
         break;
      }
   }

   free_buf((union SIGNAL**)&hbi);

   if (locked) {
      heap_lock(heap, &lock, 0);
   }

   if (brklocked) {
      heap_brklock(heap, &lock);
   }

   return HEAP_ACTION_CONTINUE;
}

/**************************************************************************/




/***************************************************************************
 * Public Interface
 ***************************************************************************/

HeapStatus
heap_attach_ose_error_filter(struct Heap *heap)
{
   HeapStatus status;
   
   status = zzheap_attach_hook(
      heap, HEAP_EVENT_ERROR_RES, l_heap_error, NULL, HEAP_PRIO_FILTER_NORMAL);
   if (status != HEAP_SUCCESS)
      return status;

   return status;
}
