/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2013 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

/* OSE Heap Interfaces */
#include "ose.h"
#include "ose_err.h"
#include "ose_heap.h"
#include "osetypes.h"
#include "heap_err.h"

#include "heap.sig"
#include "heap_private.h"
#include "bios.h"
#include "mm.sig"
#include "mmtypes.h"
#include "mm_private.h"
#include "pose.h"
#include "compiler.h"

/* Conflicting symbols from heap.sig and heap.h. */
#undef HEAP_SUCCESS
#undef HEAP_NO_MORE_ITEMS
#undef HEAP_ENOT_FOUND
#undef HEAP_EFRAG_NOT_FOUND
#undef HEAP_EOUT_OF_MEMORY

#include "heap.h"
#include "ose_heapd.h"
#include "ose_heapd.sig"

#include <string.h>
#include <stdarg.h>



/***************************************************************************
 * Static Data
 ***************************************************************************/
static unsigned long handle = 0;      /* BIOS handle */



/***************************************************************************
 * Forward Declarations
 ***************************************************************************/
static long
l_heap_iface(long fc,
             long arg0, long arg1, long arg2,
             long arg3, long arg4, long arg5);

OSHEAP
create_manager_heap(const char *log_mem, PROCESS pid);

   
/***************************************************************************
 * Helper Functions
 ***************************************************************************/

static inline PROCESS
heapd_pid(void)
{
   static PROCESS pid = 0;
   
   if(pid == 0 && !hunt("ose_heapd", 0, &pid, NULL))
   {
      error(OSE_EFATAL_MASK | OSE_HEAP_EHUNT_HEAPD_FAILED);
   }

   return pid;
}

/***************************************************************************/

static inline int
cpu_umode(void)
{
#if defined(OSARM) && !defined(__thumb)
#  if defined(__ARMCC_VERSION)

   register Msr old_lockstate;	
   __asm { MRS	old_lockstate, CPSR}	
   return !(old_lockstate & 0xF);	

#  elif defined(__GNUC__)

   Msr old_msr;
    __asm__ volatile
    (
        " mrs %0,CPSR;"		
	: /*outputs*/ "=r" (old_msr) 
    );
    return !(old_msr & 0xF);

#  endif
#elif defined(OSPP)
    
    /* User/supervisor mode is checked by biosCall. Simply forward. */
    return 1; 

#elif defined (OSMIPS)
    /* User/supervisor mode is part of CP0_Status.
     * KSEG0 and KUSEG complicates this implementation.
     * Left as is for future improvement. */
    return 1; 

#else
    /* Fallback for sfk, x86, ARM thumb, etc. */
    return 1;
#endif
}

/***************************************************************************/



/***************************************************************************
 * Data Structures
 ***************************************************************************/
struct ReclaimRegion;

union SIGNAL
{
   SIGSELECT                   sig_no;
   struct HeapdHeapCreate      heap_create;
   struct HeapdClientAdd       client_add;
   struct mm_get_log_mem_info  info_log_mem;
};




/***************************************************************************
 * Private Interface
 ***************************************************************************/
static void    l_hucb_register_client(HUCB *, PROCESS client);
static void   *l_hucb_alloc_unsafe(HUCB *, size_t, PROCESS owner, Boolean);
static void   *l_hucb_alloc_no_extend(HUCB *, size_t, PROCESS owner, Boolean);
static void   *l_hucb_alloc_early_null(HUCB *, size_t, PROCESS owner, Boolean);
static void   *l_hucb_alloc_safe(HUCB *, size_t, PROCESS owner, Boolean);
static void    l_hucb_free(HUCB *, void *, PROCESS owner);
static size_t  l_hucb_buffer_size(HUCB *, void *, PROCESS owner);
static size_t  l_hucb_buffer_actual_size(HUCB *, void *, PROCESS owner);
static size_t  l_hucb_buffer_resize(HUCB *, void *, PROCESS owner, size_t);
static void    l_hucb_extend_heap(HUCB *, void *, size_t);
#if 0 /* FIXME Never used. */
static Boolean l_hucb_shrink_heap(HUCB *, struct ReclaimRegion *);
#endif
static size_t  l_hucb_buffer_next_size(HUCB *, size_t);
static size_t  l_hucb_set_max_size(HUCB *, size_t);
static size_t  l_hucb_set_debug_hooks(HUCB *, size_t, HeapDebugHooks*);
static void    l_hucb_remove_debug_hooks(HUCB *);
static void    l_hucb_forced_free(HUCB *, void *);
static void   *l_hucb_alloc_tentative(HUCB *, U32, PROCESS, Boolean);
static U32     l_hucb_free_tentative(HUCB *, void*, U32, PROCESS, Boolean);
static void   *l_hucb_alloc_watermark(HUCB *, size_t, U32, PROCESS, Boolean);

static void *  l_hucb_plug_alloc(HUCB *hucb, U32 *p, HeapFileLine *fl, PROCESS pid, size_t size);
static void    l_hucb_plug_free(HUCB *hucb, U32 *p, HeapFileLine *fl, PROCESS pid, void *ptr);
static void *  l_hucb_plug_resize(HUCB *hucb, U32 *p, HeapFileLine *fl, PROCESS pid, void *ptr, size_t size);
static void    l_hucb_plug_collect(HUCB *hucb, U32 *p, HeapFileLine *fl, PROCESS pid);
static void    l_hucb_plug_query(HUCB *hucb, U32 *p);
static void    l_hucb_plug_control(HUCB *hucb, U32 *p);


static void    l_bios_register_client(HUCB *, PROCESS client);
static void   *l_bios_alloc_unsafe(HUCB *, size_t, PROCESS owner, Boolean);
static void   *l_bios_alloc_no_extend(HUCB *, size_t, PROCESS owner, Boolean);
static void   *l_bios_alloc_early_null(HUCB *, size_t, PROCESS owner, Boolean);
static void   *l_bios_alloc_safe(HUCB *, size_t, PROCESS owner, Boolean);
static void    l_bios_free(HUCB *, void *, PROCESS owner);
static size_t  l_bios_buffer_size(HUCB *, void *, PROCESS owner);
static size_t  l_bios_buffer_actual_size(HUCB *, void *, PROCESS owner);
static size_t  l_bios_buffer_resize(HUCB *, void *, PROCESS owner, size_t);
static void    l_bios_extend_heap(HUCB *, void *, size_t);
#if 0 /* FIXME Never used. */
static Boolean l_bios_shrink_heap(HUCB *, struct ReclaimRegion *);
#endif
static size_t  l_bios_buffer_next_size(HUCB *, size_t);
static size_t  l_bios_set_max_size(HUCB *, size_t);
static size_t  l_bios_set_debug_hooks(HUCB *, size_t, HeapDebugHooks*);
static void    l_bios_remove_debug_hooks(HUCB *);
static void    l_bios_forced_free(HUCB *, void *);
static void   *l_bios_alloc_tentative(HUCB *, U32, PROCESS, Boolean);
static U32     l_bios_free_tentative(HUCB *, void*, U32, PROCESS, Boolean);
static void   *l_bios_alloc_watermark(HUCB *, size_t, U32, PROCESS, Boolean);
static void    l_bios_collect(HUCB *, PROCESS);

static void *  l_bios_plug_alloc(HUCB *hucb, U32 *p, HeapFileLine *fl, PROCESS pid, size_t size);
static void    l_bios_plug_free(HUCB *hucb, U32 *p, HeapFileLine *fl, PROCESS pid, void *ptr);
static void *  l_bios_plug_resize(HUCB *hucb, U32 *p, HeapFileLine *fl, PROCESS pid, void *ptr, size_t size);
static void    l_bios_plug_collect(HUCB *hucb, U32 *p, HeapFileLine *fl, PROCESS pid);
static void    l_bios_plug_query(HUCB *hucb, U32 *p);
static void    l_bios_plug_control(HUCB *hucb, U32 *p);

extern void    default_debug_alloc(OSHEAP, void *, size_t, size_t, PROCESS, const char *, U32);
extern int     default_debug_free(OSHEAP, void *, size_t, PROCESS);
extern void    default_debug_fileline(OSHEAP, struct HeapBufferInfoReply *, U32);
extern void    default_debug_cleanup(OSHEAP);

OSHEAP         ose_handle_create_heap(PROCESS pid, OSDOMAIN domain, OSBOOLEAN supervisor_mode);


/***************************************************************************/


HUCB ose_hucb =
{
   HEAP_VERSION,
   l_hucb_register_client,
   l_hucb_alloc_unsafe,
   l_hucb_alloc_safe,
   l_hucb_free,
   l_hucb_buffer_size,
   l_hucb_buffer_actual_size,
   l_hucb_buffer_resize,
   l_hucb_extend_heap,
   l_hucb_alloc_early_null,
   l_hucb_buffer_next_size,
   ose_handle_create_heap,
   l_hucb_set_debug_hooks,
   l_hucb_remove_debug_hooks,
   l_hucb_forced_free,
   l_hucb_alloc_no_extend,
   default_debug_alloc,
   default_debug_free,
   default_debug_fileline,  
   default_debug_cleanup, 
   0,
   l_hucb_set_max_size,
   l_hucb_alloc_tentative,
   l_hucb_free_tentative,
   l_hucb_alloc_watermark,

   /* Introduced in OSE 5.5 */
   l_hucb_plug_alloc,
   l_hucb_plug_free,
   l_hucb_plug_resize,
   l_hucb_plug_collect,
   l_hucb_plug_query,
   l_hucb_plug_control
};



/***************************************************************************
 * Private HUCB Interface
 ***************************************************************************/

static void
l_hucb_register_client(HUCB *hucb, PROCESS client)
{
   struct OseHeap *heap = (struct OseHeap *)hucb;

   /* Note that the trap only is necessary as long as the heapcd process
    * hasn't been started yet. In order to start it, the heap lock must
    * be acquired which requires supervisor mode priviledges. */
   
   if (heap->daemon == 0)
   {
      biosCall(handle, FC_REGISTER_CLIENT, (long)hucb, (long)client);
   }
   else
   {
      l_bios_register_client(hucb, client);
   }
}

/***************************************************************************/

static void * 
l_hucb_alloc_unsafe(HUCB *hucb, size_t size, PROCESS owner, Boolean annotate)
{
   (void)annotate;

   if (cpu_umode())
   {
      return (void *)biosCall(handle, FC_UNSAFE, (long)hucb, size, owner, annotate);
   }
   else
   {
      return l_bios_alloc_unsafe(hucb, size, owner, annotate);
   }
}

/***************************************************************************/

static void * 
l_hucb_alloc_safe(HUCB *hucb, size_t size, PROCESS owner, Boolean annotate)
{
   (void)annotate;
   
   if (cpu_umode())
   {
      return (void *)biosCall(handle, FC_SAFE, (long)hucb, size, owner, annotate);
   }
   else
   {
      return l_bios_alloc_safe(hucb, size, owner, annotate);
   }
}

/***************************************************************************/

static void
l_hucb_free(HUCB *hucb, void *ptr, PROCESS owner)
{
   if (cpu_umode())
   {
      biosCall(handle, FC_FREE, (long)hucb, ptr, owner);
   }
   else
   {
      l_bios_free(hucb, ptr, owner);
   }
}

/***************************************************************************/

static size_t 
l_hucb_buffer_size(HUCB *hucb, void *ptr, PROCESS owner)
{
   if (cpu_umode())
   {
      return (size_t)biosCall(handle, FC_SIZE, (long)hucb, ptr, owner);
   }
   else
   {
      return l_bios_buffer_size(hucb, ptr, owner);
   }
}

/***************************************************************************/

static size_t 
l_hucb_buffer_actual_size(HUCB *hucb, void *ptr, PROCESS owner)
{   
   if (cpu_umode())
   {
      return (size_t)biosCall(handle, FC_ACTUAL_SIZE, (long)hucb, ptr, owner);
   }
   else
   {
      return l_bios_buffer_actual_size(hucb, ptr, owner);
   }
}

/***************************************************************************/

static size_t 
l_hucb_buffer_resize(HUCB *hucb, void *ptr, PROCESS owner, size_t new_size)
{
   if (cpu_umode())
   {
      return (size_t)biosCall(handle, FC_RESIZE, (long)hucb, ptr, owner, new_size);
   }
   else
   {
      return l_bios_buffer_resize(hucb, ptr, owner, new_size);
   }
}

/***************************************************************************/

static void
l_hucb_extend_heap(HUCB* hucb, void *frag_addr, size_t frag_size)
{
   if (cpu_umode())
   {
      biosCall(handle, FC_EXTEND, (long)hucb, (long)frag_addr, frag_size);
   }
   else
   {
      l_bios_extend_heap(hucb, frag_addr, frag_size);
   }
}

/***************************************************************************/

static void * 
l_hucb_alloc_early_null(HUCB *hucb, size_t size, PROCESS owner, Boolean annotate)
{
   (void)annotate;

   if (cpu_umode())
   {
      return (void *)biosCall(handle, FC_EARLY_NULL, (long)hucb, size, owner, annotate);
   }
   else
   {
      return l_bios_alloc_early_null(hucb, size, owner, annotate);
   }
}

/***************************************************************************/

static size_t 
l_hucb_buffer_next_size(HUCB *hucb, size_t size)
{
   if (cpu_umode())
   {
      return (size_t)biosCall(handle, FC_NEXT_SIZE, (long)hucb, size);
   }
   else
   {
      return l_bios_buffer_next_size(hucb, size);
   }
}

/***************************************************************************/

OSHEAP
ose_handle_create_heap(PROCESS pid, OSDOMAIN domain, OSBOOLEAN supervisor)
{
   static const SIGSELECT create_sig[] = {1, HEAPD_HEAP_CREATE};
   OSHEAP href;
   U32 status;
   
   union SIGNAL *sig = alloc(sizeof(struct HeapdHeapCreate), HEAPD_HEAP_CREATE);
   sig->heap_create.owner      = pid;
   sig->heap_create.domain     = domain;
   sig->heap_create.supervisor = supervisor;

   send(&sig, heapd_pid());
   sig  = receive(create_sig);

   status = sig->heap_create.status;
   href   = sig->heap_create.href;
   free_buf(&sig);

   if (status == HEAP_EPLUG_IN_NOT_FOUND)
   {
      /* The abscence of a plug-in is considered to be 
       * a major design error and is therefore immediately
       * reported as an error. */
      error2(OSE_HEAP_EPLUGIN_NOT_FOUND, 0);
   }
   
   if (handle == 0)
   {
      /* Install HEAP bios module. */
      int err = biosInstall(
         HEAP_BIOS_NAME, l_heap_iface, BIOS_USERMODE_ALLOWED);

      if (err != 0)
      {
         struct HeapBufferInfoReply hbi = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, {0} };
         hbi.status = (U32) err;
         error2(OSE_EFATAL_MASK | OSE_HEAP_EBIOS_INSTALL_FAILED, (OSERRCODE) (&hbi));
         kill_proc(current_process());
      }

      /* Open HEAP bios module. */
      handle = biosOpen(HEAP_BIOS_NAME);
      if (handle == 0)
      {
         error(OSE_EFATAL_MASK | OSE_HEAP_EBIOS_OPEN_FAILED);
         kill_proc(current_process());
      }
   }

   
   return href;
}

/***************************************************************************/

OSHEAP
create_manager_heap(const char *log_mem, PROCESS pid)
{
   SIGSELECT info_reply[] = {1, MM_GET_LOG_MEM_INFO_REPLY};
   PROCESS mm_pid;
   OSHEAP heap;
   union SIGNAL *sig;
   Boolean manager_mode_state;

   /* Verify that this is domain 0 */
   if (get_sysconf(72,pid,16,0) != 0)
   {
      error(OSE_EFATAL_MASK|OSE_HEAP_EILLEGAL_DOMAIN);
   }

   /* Ask MM about the log_mem's manager mode configuration */
   (void)hunt("ose_memory_mgr", 0, &mm_pid, NULL);

   sig = alloc(sizeof(struct mm_get_log_mem_info),
               MM_GET_LOG_MEM_INFO_REQUEST);
   sig->info_log_mem.count = 1;
   sig->info_log_mem.tags[0].tag_id = MM_TAG_INFO_LOG_MEM_MANAGER_MODE;
   strlcpy(sig->info_log_mem.name, log_mem, MM_NAME_LENGTH);
   send(&sig, mm_pid);
   sig = receive(info_reply);

   /* Check MM errors */
   if (sig->info_log_mem.status != MM_SUCCESS)
   {
      error2(OSE_EFATAL_MASK|OSE_HEAP_EINVALID_LOG_MEM,
             sig->info_log_mem.status);
      return (OSHEAP)0;
   }

   /* Verify that the log_mem is a manager mode log_mem */
   if (sig->info_log_mem.tags[0].tag_value != 1)
   {
      error(OSE_EFATAL_MASK|OSE_HEAP_EINVALID_LOG_MEM);
      return (OSHEAP)0;
   }

   free_buf(&sig);

   /* Finally, create the secure heap */
   set_env(pid, HEAP_PERMISSIONS, "0x101"); /* FIXME */
   set_env(pid, HEAP_LOG_MEM, log_mem);

   manager_mode_state = zzset_manager_mode(TRUE);
   heap = ose_create_heap(pid, (OSDOMAIN)0, TRUE);
   zzset_manager_mode(manager_mode_state);

   HEAP_FROM_HREF(heap)->manager = 1;
      
   return heap;
}

/***************************************************************************/

OSHEAP
ose_create_manager_heap(const char *log_mem, PROCESS pid)
{
   return create_manager_heap(log_mem, pid);
}

/***************************************************************************/

static size_t
l_hucb_set_debug_hooks(HUCB *hucb, size_t annsz, HeapDebugHooks * hooks)
{
   if (cpu_umode())
   {
      return (size_t)biosCall(
         handle, FC_SET_DEBUG, (long)hucb, (long)annsz, (long)hooks);   
   }
   else
   {
      return l_bios_set_debug_hooks(hucb, annsz, hooks);
   }
}

/***************************************************************************/

static void
l_hucb_remove_debug_hooks(HUCB *hucb)
{
   if (cpu_umode())
   {
      biosCall(handle, FC_REMOVE_DEBUG, (long)hucb);
   }
   else
   {
      l_bios_remove_debug_hooks(hucb);
   }
}

/***************************************************************************/

static 
void l_hucb_forced_free(HUCB *hucb, void *ptr)
{
   if (cpu_umode())
   {
      (void)biosCall(handle, FC_FORCED_FREE, (long)hucb, ptr);
   }
   else
   {
      l_bios_forced_free(hucb, ptr);
   }
}

/***************************************************************************/

static void * 
l_hucb_alloc_no_extend(HUCB *hucb, size_t size, PROCESS owner, Boolean annotate)
{
   (void)annotate;

   if (cpu_umode())
   {
      return (void *)biosCall(handle, FC_NO_EXTEND, (long)hucb, size, owner, annotate);      
   }
   else
   {
      return l_bios_alloc_no_extend(hucb, size, owner, annotate);
   }
}

/***************************************************************************/

static size_t 
l_hucb_set_max_size(HUCB *hucb, size_t new_size)
{   
   if (cpu_umode())
   {
      return (size_t)biosCall(handle, FC_SETMAX, (long)hucb, new_size);
   }
   else
   {
      return l_bios_set_max_size(hucb, new_size);
   }
}

/***************************************************************************/

static void *
l_hucb_alloc_tentative(HUCB *hucb, U32 bufhandle, PROCESS owner, Boolean annotate)
{
   (void)annotate;
   
   if (cpu_umode())
   {
      return (void*)biosCall(handle, FC_ALLOC_TENTATIVE,
         (long)hucb, (long)bufhandle, (long)owner, (long)annotate);
   }
   else
   {
      return l_bios_alloc_tentative(hucb, bufhandle, owner, annotate);
   }
}

/***************************************************************************/

static U32
l_hucb_free_tentative(HUCB *hucb, void *ptr, U32 watermark,
                      PROCESS owner, Boolean annotate)
{
   (void)annotate;
   
   if (cpu_umode())
   {
      return (U32)biosCall(handle, FC_FREE_TENTATIVE, (long)hucb, (long)ptr,
                           (long)watermark, (long)owner, (long)0);
   }
   else
   {
      return l_bios_free_tentative(hucb, ptr, watermark, owner, 0);
   }
}

/***************************************************************************/

static void *
l_hucb_alloc_watermark(HUCB *hucb, size_t size, U32 watermark,
                       PROCESS owner, Boolean annotate)
{
   (void)annotate;

   if (cpu_umode())
   {
      return (void*)biosCall(handle, FC_ALLOC_WATERMARK, (long)hucb, (long)size,
                             (long)watermark, (long)owner, (long)annotate);
   }
   else
   {
      return l_bios_alloc_watermark(hucb, size, watermark, owner, annotate);
   }
}


/***************************************************************************/

#if 0  /* FIXME: Never used. */
static Boolean
l_hucb_shrink_heap(HUCB *hucb, struct ReclaimRegion *sig)
{
   if (cpu_umode())
   {
      return (Boolean)biosCall(handle, FC_SHRINK, (long)hucb, (long)sig);
   }
   else
   {
      return l_bios_shrink_heap(hucb, sig);
   }
}
#endif

/***************************************************************************/

static void *
l_hucb_plug_alloc(HUCB *hucb, U32 *p,
                  HeapFileLine *fl, PROCESS pid, size_t size)
{
   if (cpu_umode())
   {
      return (void*)biosCall(
         handle, FC_PLUG_ALLOC,
         (long)hucb, (long)p, (long)fl, (long)pid, (size_t)size);
   }
   else
   {
      return l_bios_plug_alloc(hucb, p, fl, pid, size);
   }
}
   
/***************************************************************************/

static void
l_hucb_plug_free(HUCB *hucb, U32 *p,
                 HeapFileLine *fl, PROCESS pid, void *ptr)
{
   if (cpu_umode())
   {
      biosCall(
         handle, FC_PLUG_FREE,
         (long)hucb, (long)p, (long)fl, (long)pid, (long)ptr);
   }
   else
   {
      l_bios_plug_free(hucb, p, fl, pid, ptr);
   }
}
   
/***************************************************************************/

static void *
l_hucb_plug_resize(HUCB *hucb, U32 *p,
                   HeapFileLine *fl, PROCESS pid, void *ptr, size_t size)
{
   if (cpu_umode())
   {
      return (void*)biosCall(
         handle, FC_PLUG_RESIZE,
         (long)hucb, (long)p, (long)fl, (long)pid, (long)ptr, (long)size);
   }
   else
   {
      return l_bios_plug_resize(hucb, p, fl, pid, ptr, size);
   }
}
   
/***************************************************************************/

static void
l_hucb_plug_collect(HUCB *hucb, U32 *p, HeapFileLine *fl, PROCESS pid)
{
   if (cpu_umode())
   {
      biosCall(handle, FC_PLUG_COLLECT,
         (long)hucb, (long)p, (long)fl, (long)pid);
   }
   else
   {
      l_bios_plug_collect(hucb, p, fl, pid);
   }
}
   

/***************************************************************************/

static void
l_hucb_plug_query(HUCB *hucb, U32 *p)
{
   if (cpu_umode())
   {
      biosCall(handle, FC_PLUG_QUERY, (long)hucb, (long)p);
   }
   else
   {
      l_bios_plug_query(hucb, p);
   }
}

/***************************************************************************/

static void
l_hucb_plug_control(HUCB *hucb, U32 *p)
{
   if (cpu_umode())
   {
      biosCall(handle, FC_PLUG_CONTROL, (long)hucb, (long)p);
   }
   else
   {
      l_bios_plug_control(hucb, p);
   }
}   

/***************************************************************************/

static void
l_heap_daemon_start(struct OseHeap *heap, PROCESS client)
{
   if (likely(!heap->daemon))
   {
      struct OS_pcb *pcb;
#ifdef OSMIPS
      /* On MIPS we need to create the heapcd with an entrypoint
       * that is available in user mode. It is done here instead of
       * by the master heap daemon. The heapcd will immediately trap
       * into supervisor mode once it is started. */
         
      OSENTRYPOINT *daemon = (OSENTRYPOINT *)
         get_envp(get_bid(client), HEAP_CLEANUPD_ENTRYPOINT);
         
      heap->daemon_stopped = heapd_daemon_create(
         heap->domain, get_bid(heap->owner), heap->owner, daemon, heap);
#else
      /* On other architectures, the heapcd is created by the master
       * heap daemon, but it is not started because it is uncertain
       * whether the creator will start the block the heapcd is created
       * in or not. So, the first client registering will do it if not
       * already done. */
      (void)client;
#endif

      /* Get process status and start the heapcd if it is stopped. */
      pcb = get_pcb(heap->daemon_stopped);
      if (pcb->status & OS_STOPPED)
         start(heap->daemon_stopped);
      free_buf((union SIGNAL **)&pcb);

      /* Indicate to the master daemon that the heapcd is operational. */
      heap->daemon = heap->daemon_stopped;
   }
}

/***************************************************************************/




/***************************************************************************
 * Private BIOS Interface
 ***************************************************************************/

static void
l_bios_register_client(HUCB *hucb, PROCESS client)
{
   enum PROCESS_TYPE type = get_ptype(client);
   union SIGNAL *sig = alloc(sizeof(struct HeapdClientAdd), HEAPD_CLIENT_ADD);
   struct OseHeap *heap = (struct OseHeap *)hucb;

   /*
    * OSE5.0 compatibility; we did not have the client parameter
    * in OSE5.0, therefore this must be forced to be current process.
    *
    * FIXME: Is this still needed? The heap is only binary interface
    *        compatible back to OSE 5.3. Older versions have an
    *        incompatible hucb interface.
    */
   if (type == OS_ILLEGAL)
   {
      client = current_process();
      type = get_ptype(client);
   }

   /* There are limitations on heap use for interrupt processes */
   if ((OS_INT_PROC == (type)) || (OS_TI_PROC == (type)))
   {
      /* Mutex lock does not work with interrupt processes. */
      if (EHEAP_LOCK_MUTEX == heap_lock_type(&heap->heap))
      {
         error2(OSE_EFATAL_MASK | OSE_HEAP_EILLEGAL_USER, client);
         kill_proc(client);
      }

      /*
      *  We have also considered other limitations on interrupt processes:
      *
      *  It is not possible to extend the heap from interrupt context.
      *  We have therefore considered requiring the heap to be "static"
      *  i.e. configured with initial size equal to max size so that it
      *  cannot be extended anyhow.
      *
      *  We decided against enforcing this limitation because the limitation
      *  does not damage the heap, it just causes an unexpected failed
      *  allocation. The application is free to deal with this in any way
      *  desired, configuration to avoid it or implementing interrupt
      *  processes that can deal with failed allocation. (OSECR-15406)
      *  (There is a check to prevent breaking from interrupt processes
      *  in heap_lock.h, heap_brkallowed(void))
      *
      *  We have previously prohibited use of heap from interrupt processes
      *  in multi core systems. This limitation did not serve any purpose
      *  and was removed when it caused customer pain. (OSECR-15406)
      */
   }

   /* Start dedicated heap daemon if not already done. */
   if (unlikely(!heap->daemon))
   {
      HeapLock lock;

      heap_lock(&heap->heap, &lock, 0);
      l_heap_daemon_start(heap, client);
      heap_unlock(&heap->heap, &lock, 0);
   }

   sig->client_add.href = HREF_FROM_HEAP(heap);
   send_w_s(&sig, client, heap->daemon);
}

/***************************************************************************/

static void *
l_bios_alloc_unsafe(HUCB *hucb, size_t size, PROCESS owner, Boolean ann)
{
   U32 *ptr;

   ptr = zzheap_alloc(HEAP_FROM_HUCB(hucb), size, owner, NULL, __FILE__, __LINE__);

   if (ptr != NULL && ann) 
      HDM_SET_ANNOTATED(ptr[-1]);
      
   return ptr;
}

/***************************************************************************/

static void *
l_bios_alloc_no_extend(HUCB *hucb, size_t size, PROCESS owner, Boolean ann)
{
   HeapParam p[] = {
      HPM(BREAK_NO, 0),
      HPM(END, 0)
   };

   U32 *ptr;

   ptr = zzheap_alloc(HEAP_FROM_HUCB(hucb), size, owner, p, __FILE__, __LINE__);
   
   if (ptr != NULL && ann)
      HDM_SET_ANNOTATED(ptr[-1]);

   return ptr;
}

/***************************************************************************/

static void *
l_bios_alloc_early_null(HUCB *hucb, size_t size, PROCESS owner, Boolean ann)
{
   U32 *ptr;

   ptr = zzheap_alloc(HEAP_FROM_HUCB(hucb), size, owner, NULL, __FILE__, __LINE__);

   if (ptr != NULL && ann)
      HDM_SET_ANNOTATED(ptr[-1]);

   return ptr;
}

/***************************************************************************/

static void *
l_bios_alloc_safe(HUCB *hucb, size_t size, PROCESS owner, Boolean ann)
{
   HeapParam p[] = {
      HPM(ALLOC_SAFE, 0),
      HPM(END, 0)
   };

   U32 *ptr;

   ptr = zzheap_alloc(HEAP_FROM_HUCB(hucb), size, owner, p, __FILE__, __LINE__);

   if (ptr != NULL && ann)
      HDM_SET_ANNOTATED(ptr[-1]);

   return ptr;
}

/***************************************************************************/

static void
l_bios_free(HUCB *hucb, void *ptr, PROCESS owner)
{
   zzheap_free(HEAP_FROM_HUCB(hucb), ptr, owner, NULL, __FILE__, __LINE__);
}

/***************************************************************************/

static size_t
l_bios_buffer_size(HUCB *hucb, void *ptr, PROCESS owner)
{
   HeapParam params[] = {
       HPM(QUERY_BUF_REQ_SIZE, 1), 0,
       HPM_END
   };

   (void)owner; /* FIXME? */

   params[1] = (HeapParam)ptr;
   zzheap_query(HEAP_FROM_HUCB(hucb), params, 0);

   if (!HPM_IS_HANDLED(params[0]))
      return 0;

   return (size_t)params[1];
}

/***************************************************************************/

static size_t
l_bios_buffer_actual_size(HUCB *hucb, void *ptr, PROCESS owner)
{
   HeapParam params[] = {
      HPM(QUERY_BUF_ACT_SIZE, 1), 0,
      HPM_END
   };

   (void)owner; /* FIXME? */

   params[1] = (HeapParam)ptr;
   zzheap_query(HEAP_FROM_HUCB(hucb), params, 0);

   if (!HPM_IS_HANDLED(params[0]))
      return 0;

   return (size_t)params[1];
}

/***************************************************************************/

static size_t
l_bios_buffer_resize(HUCB *hucb, void *ptr, PROCESS owner, size_t new_size)
{
   /* Resize only request. Data not copied. */

   zzheap_realloc(HEAP_FROM_HUCB(hucb), ptr, new_size, owner,
                  NULL, 1, __FILE__, __LINE__);
   
   return l_bios_buffer_size(hucb, ptr, owner);
}

/***************************************************************************/

static void
l_bios_extend_heap(HUCB *hucb, void *base, size_t size)
{
   HeapParam p[] = {
      HPM(CTRL_FRAG_INSERT, 2), 0, 0,
      HPM_END
   };

   if (mm_space_is_mas((log_addr_t)base, (log_size_t)size))
   {
      struct OseHeap *heap = HEAP_FROM_HREF(HREF_FROM_HUCB(hucb));
      heap->config.frag_mas = 1;
   }
   
   p[1] = (HeapParam)base;
   p[2] = (HeapParam)size;

   zzheap_control(HEAP_FROM_HUCB(hucb), p);
}

/***************************************************************************/

#if 0 /* FIXME Never used. */
static Boolean
l_bios_shrink_heap(HUCB *hucb, struct ReclaimRegion *sig)
{
   HeapParam p[] = {
      HPM(CTRL_FRAG_REMOVE, 2), 0, 0,
      HPM_END
   };

   /* FIXME. */
   p[2] = (HeapParam)0;

   zzheap_control(HEAP_FROM_HUCB(hucb), p);

   if (HPM_IS_HANDLED(p[0]))
      return 1;

   return 0;
}

#endif

/***************************************************************************/

static size_t
l_bios_buffer_next_size(HUCB *hucb, size_t size)
{
   HeapParam p[] = {
      HPM(QUERY_BUF_NEXT_SIZE, 1), 0,
      HPM_END
   };

   p[1] = (HeapParam)size;

   zzheap_query(HEAP_FROM_HUCB(hucb), p, 0);

   if (!HPM_IS_HANDLED(p[0]))
      return 0;

   return (size_t)p[1];
}

/***************************************************************************/

static size_t
l_bios_set_max_size(HUCB *hucb, size_t size)
{
   struct OseHeap *heap = (struct OseHeap *)hucb;

   HeapParam p[] = {
      0, 0,
      HPM_END
   };

   if (heap->config.max_type == 1)  /* FRAG_MAX */
   {
      p[0] = HPM(CTRL_FRAG_MAX_SIZE, 1);
   }
   else /* BUF_MAX */
   {
      p[0] = HPM(CTRL_BUF_MAX_SIZE, 1);
   }      
      
   p[1] = size;

   zzheap_control(HEAP_FROM_HUCB(hucb), p);

   if (!HPM_IS_HANDLED(p[0]))
   {
      /* Change failed. Check the current size and return it. */
      if (heap->config.max_type == 1) /* FRAG_MAX */
      {
         p[0] = HPM(QUERY_FRAG_MAX_SIZE, 1);
      }
      else /* BUF_MAX */
      {         
         p[0] = HPM(QUERY_BUF_MAX_SIZE, 1);
      }

      p[2] = HPM_END;

      zzheap_query(HEAP_FROM_HUCB(hucb), p, 0);

      if (!HPM_IS_HANDLED(p[0]))
         return 0;
      
      return (size_t)p[1];
   }

   return (size_t)p[1];
}

/***************************************************************************/

static size_t
l_bios_set_debug_hooks(HUCB *hucb, size_t annsz, HeapDebugHooks *hooks)
{
   /* If annotation size_has been changed once
    * it must always stay the same! -> change
    * it only if it has not already been set.
    * 
    * An annotation size value of -1 indicates
    * setting of default debug hooks and should
    * therefor not affect the configured annotation
    * size
    ***********************************************/
   if (hucb->annotation_size == 0 && annsz != (size_t)-1)
   {
      /* We always need to fit a DefAnnotation in the annotation data! */
      if (annsz < SIZEOF_DEF_ANNOTATION)
      {
         annsz = SIZEOF_DEF_ANNOTATION;
      }
      
      /* Make sure buffer alignment is always OK */
      hucb->annotation_size = (annsz + HEAP_ALIGNMENT - 1) & ~(HEAP_ALIGNMENT - 1);
   }

   hucb->heap_debug_alloc    = hooks->debug_alloc;
   hucb->heap_debug_free     = hooks->debug_free;
   hucb->heap_debug_fileline = hooks->debug_fileline;
   hucb->heap_debug_cleanup  = hooks->debug_cleanup;
   
   return hucb->annotation_size;
}

/***************************************************************************/

static void
l_bios_remove_debug_hooks(HUCB *hucb)
{
   if (hucb->heap_debug_cleanup)
   {
      hucb->heap_debug_cleanup((OSHEAP)HREF_FROM_HUCB(hucb));
   }

#ifndef OSMIPS   
   /*
   ** Even if the CRT now makes an explicit set of the application default
   ** debug hooks after removal, the initialisation below must still be kept
   ** for backward compatibility reasons. On non MIPS targets it is possible
   ** to execute the monoliths default hooks from a separately linked user
   ** mode application context. This is however fatal for MIPS.
   */
   hucb->heap_debug_alloc    = default_debug_alloc;
   hucb->heap_debug_free     = default_debug_free;
   hucb->heap_debug_fileline = default_debug_fileline;
   hucb->heap_debug_cleanup  = default_debug_cleanup;
#endif
}

/***************************************************************************/

static void
l_bios_forced_free(HUCB *hucb, void *ptr)
{
   l_bios_free(hucb, ptr, 0);
}

/***************************************************************************/

static void *
l_bios_alloc_tentative(HUCB *hucb, U32 handle, PROCESS owner, Boolean ann)
{
   HeapParam p[] = {
      HPM(CTRL_TENTATIVE_ALLOC, 2), 0, 0,
      HPM_END
   };

   (void)ann;

   p[1] = (HeapParam)handle;
   p[2] = (HeapParam)owner;

   zzheap_control(HEAP_FROM_HUCB(hucb), p);

   return HPM_IS_HANDLED(p[0]) ? (void *)p[1] : NULL;
}

/***************************************************************************/

static U32
l_bios_free_tentative(HUCB *hucb, void *ptr, U32 wmark, PROCESS owner, Boolean ann)
{
   HeapParam p[] = {
      HPM(CTRL_TENTATIVE_FREE, 2), 0, 0,
      HPM_END
   };

   (void)ann;

   p[1] = (HeapParam)ptr;
   p[2] = (HeapParam)owner;

   if (unlikely(wmark > 0))
   {
      zzheap_free(HEAP_FROM_HUCB(hucb), ptr, owner, NULL, __FILE__, __LINE__);
      return HEAP_INVALID_TENTATIVE_HANDLE;
   }
   
   zzheap_control(HEAP_FROM_HUCB(hucb), p);

   /* The buffer is freed permanently if the tentative
    * plug-in is missing. If so, return an invalid handle. */
   return HPM_IS_HANDLED(p[0]) ? (U32)p[1] : HEAP_INVALID_TENTATIVE_HANDLE;
}

/***************************************************************************/

static void *
l_bios_alloc_watermark(HUCB *hucb, size_t size, U32 wmark, PROCESS owner, Boolean ann)
{
   (void)ann;
   (void)wmark;

   /* There is no watermark plug-in as the feature has is deprecated.
    * All allocations will be allowed as if it is a regular one. */
   return zzheap_alloc(HEAP_FROM_HUCB(hucb), size, owner, NULL, __FILE__, __LINE__);
}

/***************************************************************************/

static void
l_bios_collect(HUCB *hucb, PROCESS pid)
{
   /* This function should not be needed anymore.
    * Kept for backwards compatibilty, just in case. */

   zzheap_collect(HEAP_FROM_HUCB(hucb), pid);
}

/***************************************************************************/

static void *
l_bios_plug_alloc(HUCB *hucb, U32 *p,
                  HeapFileLine *fl, PROCESS pid, size_t size)
{
   return zzheap_alloc(HEAP_FROM_HUCB(hucb),
                       size, (HeapPID)pid, (HeapParam *)p, fl->file, fl->line);
}
   
/***************************************************************************/

static void
l_bios_plug_free(HUCB *hucb, U32 *p,
                 HeapFileLine *fl, PROCESS pid, void *ptr)
{
   (void)fl;

   zzheap_free(HEAP_FROM_HUCB(hucb),
               ptr, (HeapPID)pid, (HeapParam *)p, NULL, 0);
}
   
/***************************************************************************/

static void *
l_bios_plug_resize(HUCB *hucb, U32 *p,
                   HeapFileLine *fl, PROCESS pid, void *ptr, size_t size)
{
   return zzheap_realloc(HEAP_FROM_HUCB(hucb),
      ptr, size, (HeapPID)pid, (HeapParam *)p, 0, fl->file, fl->line);
}
   
/***************************************************************************/

static void
l_bios_plug_collect(HUCB *hucb, U32 *p, HeapFileLine *fl, PROCESS pid)
{
   (void)fl;
   (void)p;

   zzheap_collect(HEAP_FROM_HUCB(hucb), (HeapPID)pid);
}
   

/***************************************************************************/

static void
l_bios_plug_query(HUCB *hucb, U32 *p)
{
   zzheap_query(HEAP_FROM_HUCB(hucb), (HeapParam *)p, 0);
}

/***************************************************************************/

static void
l_bios_plug_control(HUCB *hucb, U32 *p)
{
   zzheap_control(HEAP_FROM_HUCB(hucb), (HeapParam *)p);
}   

/***************************************************************************/



/***************************************************************************
 * Public Interface
 ***************************************************************************/

static long
l_heap_iface(long fc,
             long arg0, long arg1, long arg2,
             long arg3, long arg4, long arg5)
{
   HUCB *hucb = (HUCB *)arg0; /* arg0 is _always_ the HUCB pointer */

   (void)arg5;

   switch (fc)
   {
      /* These first three function codes are kept here
       * for backwards compatibility and to support MIPS. */
      case FC_CREATE_HEAP:
      {
         PROCESS pid = (PROCESS)arg0;
         OSDOMAIN domain = (PROCESS)arg1;
         OSBOOLEAN supervisor_mode = (OSBOOLEAN)arg2;
         return (long)ose_handle_create_heap(pid, domain, supervisor_mode);
      }

      case FC_REGISTER_CLIENT:
      {
         l_bios_register_client((HUCB *)hucb, (PROCESS)arg1);
         return 0;
      }
      case FC_CLEANUPD:
      {
         ose_heapcd();
         assert(0);
         return 0; /* Shall not return */
      }

      case FC_UNSAFE:
      {
         return (long)l_bios_alloc_unsafe(
            hucb, (size_t)arg1, (PROCESS)arg2, (Boolean)arg3);
      }
      case FC_SAFE:
      {
         return (long)l_bios_alloc_safe(
            hucb, (size_t)arg1, (PROCESS)arg2, (Boolean)arg3);
      }            
      case FC_FREE:
      {
         l_bios_free(hucb, (void *)arg1, (PROCESS)arg2);
         return 0;
      }
      case FC_FORCED_FREE:
      {
         l_bios_forced_free(hucb, (void *)arg1);
         return 0;
      }
      case FC_NO_EXTEND:
      {
         return (long)l_bios_alloc_no_extend(
            hucb, (size_t)arg1, (PROCESS)arg2, (Boolean)arg3);
      }
      case FC_EARLY_NULL:
      {
         return (long)l_bios_alloc_early_null(
            hucb, (size_t)arg1, (PROCESS)arg2, (Boolean)arg3);
      }
      case FC_SIZE:
      {
         return (long)l_bios_buffer_size(hucb, (void *)arg1, (PROCESS)arg2);
      }
      case FC_SETMAX:
      {
         return (long)l_bios_set_max_size(hucb, (size_t)arg1);
      }
      case FC_CLEANUP:
      {
         l_bios_collect(hucb, (PROCESS)arg1);
         return 0;
      }
      case FC_RESIZE:
      {
         return (long)l_bios_buffer_resize(
            hucb, (void *)arg1, (PROCESS)arg2, (size_t)arg3);
      }
      case FC_EXTEND:
      {
         l_bios_extend_heap(hucb, (void *)arg1, (size_t)arg2);
         return 0;
      }
      case FC_ACTUAL_SIZE:
      {
         return (long)l_bios_buffer_actual_size(
            hucb, (void *)arg1, (PROCESS)arg2);
      }
      case FC_NEXT_SIZE:
      {
         return (long)l_bios_buffer_next_size(hucb, (size_t)arg1);
      }
      case FC_SET_DEBUG:
      {
         return (long)l_bios_set_debug_hooks(
            hucb, (size_t)arg1, (HeapDebugHooks *)arg2);         
      }
      case FC_REMOVE_DEBUG:
      {
         l_bios_remove_debug_hooks(hucb);
         return 0;
      }
      case FC_SHRINK:
      {
         /* FIXME Never used */
#if 0
         return (long)l_bios_shrink_heap(hucb, (struct ReclaimRegion*)arg1);
#else
         return -1;
#endif
      }
      case FC_ALLOC_TENTATIVE:
      {
         return (long)l_bios_alloc_tentative(
            hucb, (U32)arg1, (PROCESS)arg2, (Boolean)arg3);
      }
      case FC_FREE_TENTATIVE:
      {
         return (long)l_bios_free_tentative(
            hucb, (void *)arg1, (U32)arg2, (PROCESS)arg3, (Boolean)arg4);
      }
      case FC_ALLOC_WATERMARK:
      {
         return (long)l_bios_alloc_watermark(
            hucb, (size_t)arg1, (U32)arg2, (PROCESS)arg3, (Boolean)arg4);
      }



      case FC_PLUG_ALLOC:
      {
         return (long)l_bios_plug_alloc(
            hucb, (U32*)arg1, (HeapFileLine*)arg2, (PROCESS)arg3, (size_t)arg4);
      }
      case FC_PLUG_FREE:
      {
         l_bios_plug_free(
            hucb, (U32*)arg1, (HeapFileLine*)arg2, (PROCESS)arg3, (void *)arg4);
         return 0;
      }
      case FC_PLUG_RESIZE:
      {
         return (long)l_bios_plug_resize(
            hucb, (U32*)arg1, (HeapFileLine*)arg2, (PROCESS)arg3,
            (void*)arg4, (size_t)arg5);
      }
      case FC_PLUG_COLLECT:
      {
         l_bios_plug_collect(
            hucb, (U32*)arg1, (HeapFileLine*)arg2, (PROCESS)arg3);
         return 0;
      }
      case FC_PLUG_QUERY:
      {
         l_bios_plug_query(hucb, (U32*)arg1);
         return 0;
      }
      case FC_PLUG_CONTROL:
      {
         l_bios_plug_control(hucb, (U32*)arg1);
         return 0;
      }

      default:
      {
         struct HeapBufferInfoReply hbi = { 0, 0, 0, 0, 0, 0, 0,
                                            0, 0, 0, 0, 0, 0, 0, {0} };
         hbi.status = (U32) fc;
         error2(OSE_EFATAL_MASK | OSE_HEAP_EUNKNOWN_FCODE, (OSERRCODE) (&hbi));
         kill_proc(current_process());
      }
   }
   return -1;
}

/***************************************************************************/




/***************************************************************************
 * Public Interface
 ***************************************************************************/

void
ose_reset_heap(PROCESS bid)
{
   /*
    * inherit the parameters set by the system heap, and only
    * remove the handle so that a new ose_create_heap is necessary.
    * 
  set_env(bid, HEAP_ANNOTATION_SIZE, NULL);
  set_env(bid, HEAP_ENDMARK_SIZE, NULL);
  set_env(bid, HEAP_WIPE_ON_FREE, NULL);
  set_env(bid, HEAP_FRAG_CHECK, NULL);
  set_env(bid, HEAP_SMALL_BUF_INIT_SIZE, NULL);
  set_env(bid, HEAP_SMALL_BUF_EXT_SIZE, NULL);
  set_env(bid, HEAP_LARGE_BUF_THRESHOLD, NULL);
  set_env(bid, HEAP_MAX_SIZE, NULL);
  set_env(bid, HEAP_SMALL_LOW_LIMIT, NULL);
  set_env(bid, HEAP_LARGE_LOW_LIMIT, NULL);
  set_env(bid, HEAP_CACHE_MODE, NULL);
  set_env(bid, HEAP_LOG_MEM, NULL);
  set_env(bid, HEAP_EARLY_NULL_LEVEL, NULL);
  set_env(bid, HEAP_CLEANUPD_STACK_SIZE, NULL);
   */
   
  set_env(bid, HEAPREF_ENVNAME, NULL);
}

/***************************************************************************/

