/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2010 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */


#include "ose_heap_monitor.h"
#include "heap.sig"
#include <string.h>


/**************************************************************************
 * Plug-in History
 **************************************************************************
 *
 * IMPORTANT
 * 
 * This comment shall be updated with a description of the change each
 * time the extension version number is increased. The version number
 * shall be increased if incompatible modifications are made to internal
 * structures or after behavioral changes. See HeapExtension below.
 * 
 **************************************************************************
 *
 *  Version  Description
 *  -------  ----------------------------------------------------------
 *   100     Initial revision.  
 *
 **************************************************************************/


/*****************************************************************************
 * Data Structures
 *****************************************************************************/

struct HeapMonitor
{
   struct HeapStatistics    *stats;     /* Statistics.          */
   struct HeapConfiguration *config;    /* Configuration.       */
   struct HeapServiceInfo    info;      /* Plug-in information. */

   /* Specific configuration. */
   OSHEAP  cfg_href;                    /* Heap reference.   */
   PROCESS cfg_pid;                     /* Heap monitor pid. */
   size_t  cfg_llevel;                  /* Low level.        */
   size_t  cfg_hlevel;                  /* High level.       */

   /* State. */
   uint32_t st_seq;                     /* Sequence number.  */
   uint8_t  st_low;                     /* Level is low.     */
};

/*****************************************************************************/

union SIGNAL
{
   SIGSELECT sig_no;
   struct HeapMonitorEvent event;
};

/*****************************************************************************/




/*****************************************************************************
 * Private Interface
 *****************************************************************************/

static HeapAction
l_monitor_alloc(struct Heap *heap, union HeapEvent *event, void *opaque)
{
   struct HeapMonitor *m = opaque;
   union SIGNAL *sig;
   HeapLock lock;
   size_t smax = m->config->max_type == 1 ?
      m->config->frag_size_max : m->config->buf_size_max;

   (void)event;

   heap_lock(heap, &lock, 0);

   if (smax - m->stats->buf_size_used < m->cfg_llevel && m->st_low == 0)
   {
      m->st_low  = 1;
      m->st_seq += 1;
      heap_unlock(heap, &lock, 0);
      
      sig = alloc(sizeof(struct HeapMonitorEvent), HEAP_MONITOR_EVENT);
      sig->event.href     = m->cfg_href;
      sig->event.event    = HEAP_EVENT_LOW_LEVEL;
      sig->event.event_no = m->st_seq;
      send(&sig, m->cfg_pid);
   }
   else
   {
      heap_unlock(heap, &lock, 0);
   }

   return HEAP_ACTION_CONTINUE;
}

/*****************************************************************************/

static HeapAction
l_monitor_free(struct Heap *heap, union HeapEvent *event, void *opaque)
{
   struct HeapMonitor *m = opaque;
   union SIGNAL *sig;
   HeapLock lock;
   size_t smax = m->config->max_type == 1 ?
      m->config->frag_size_max : m->config->buf_size_max;

   (void)event;

   heap_lock(heap, &lock, 0);

   if (smax - m->stats->buf_size_used > m->cfg_hlevel && m->st_low == 1)
   {
      m->st_low  = 0;
      m->st_seq += 1;
      heap_unlock(heap, &lock, 0);
      
      sig = alloc(sizeof(struct HeapMonitorEvent), HEAP_MONITOR_EVENT);
      sig->event.href     = m->cfg_href;
      sig->event.event    = HEAP_EVENT_NOT_LOW_LEVEL;
      sig->event.event_no = m->st_seq;
      send(&sig, m->cfg_pid);
   }
   else
   {
      heap_unlock(heap, &lock, 0);
   }

   return HEAP_ACTION_CONTINUE;
}

/*****************************************************************************/

static HeapAction
l_monitor_query(struct Heap *heap, union HeapEvent *event, void *opaque)
{
   struct HeapMonitor *m = opaque;
   HeapParam *p = event->e_query.params;

   (void)heap;
   (void)opaque;
   
   if (p == NULL)
      return HEAP_ACTION_CONTINUE;
   
   while (*p != HPM_END)
   {
      switch (HPM_ID(*p))
      {
         case HPM_QUERY_SERVICE:
         {
            if (!HPM_IS_HANDLED(p[0]) && p[1] == 0)
            {
               HPM_SET_HANDLED(p[0]);
               HPM_SET_VALUE(p[1], &m->info);
            }
            else if (p[1] == (HeapParam)&m->info)
            {
               HPM_CLEAR_HANDLED(p[0]);
               HPM_SET_VALUE(p[1], 0);
            }

            break;
         }
      }

      p += (1 + HPM_VALUES(*p));
   }

   return HEAP_ACTION_CONTINUE;
}

/*****************************************************************************/

static HeapAction
l_monitor_destroy(struct Heap *heap, union HeapEvent *event, void *opaque)
{
   struct HeapMonitor *m = opaque;

   (void)event;

   zzheap_detach_hook(heap, HEAP_EVENT_ALLOC_RES, l_monitor_alloc, m);
   zzheap_detach_hook(heap, HEAP_EVENT_FREE_RES,  l_monitor_free,  m);
   zzheap_detach_hook(heap, HEAP_EVENT_QUERY_REQ, l_monitor_query, m);
   heap_destroy_monitor_filter(heap, m);

   return HEAP_ACTION_CONTINUE;
}

/*****************************************************************************/





/*****************************************************************************
 * Public Interface
 *****************************************************************************/

struct HeapMonitor *
heap_init_monitor_filter(struct Heap *heap,
                         struct HeapStatistics *stats,
                         struct HeapConfiguration *config,
                         OSHEAP href, PROCESS pid,
                         size_t low_level, size_t high_level)
{
   struct HeapMonitor *monitor = heap_meta_alloc(heap, sizeof(struct HeapMonitor));
   if (monitor == NULL)
      return NULL;

   monitor->stats       = stats;
   monitor->config      = config;
   
   monitor->info.version = 100;
   strcpy(monitor->info.id, "com.enea.heap.monitor");

   monitor->cfg_href    = href;
   monitor->cfg_pid     = pid;
   monitor->cfg_llevel  = low_level;
   monitor->cfg_hlevel  = high_level; 

   monitor->st_low      = 0;
   monitor->st_seq      = 0;
   
   return monitor;
}

/*****************************************************************************/

void
heap_destroy_monitor_filter(struct Heap *heap, struct HeapMonitor *monitor)
{
   heap_assert(monitor != NULL, "Invalid monitor instance passed as parameter");

   heap_meta_free(heap, monitor);
}

/*****************************************************************************/

HeapStatus
heap_attach_monitor_filter(struct Heap *heap, struct HeapMonitor *monitor)
{
   HeapStatus status;
   
   heap_assert(heap != NULL, "Invalid heap instance passed as parameter.");
   heap_assert(monitor != NULL, "Invalid monitor instance passed as parameter");

   status = zzheap_attach_hook(
      heap, HEAP_EVENT_ALLOC_RES, l_monitor_alloc, monitor, HEAP_PRIO_FILTER_NORMAL);
   if (status != HEAP_SUCCESS)
      return status;

   status = zzheap_attach_hook(
      heap, HEAP_EVENT_FREE_RES, l_monitor_free, monitor, HEAP_PRIO_FILTER_NORMAL);
   if (status != HEAP_SUCCESS)
      return status;

   status = zzheap_attach_hook(
      heap, HEAP_EVENT_QUERY_REQ, l_monitor_query, monitor, HEAP_PRIO_FILTER_NORMAL);
   if (status != HEAP_SUCCESS)
      return status;

   status = zzheap_attach_hook(
      heap, HEAP_EVENT_DESTROY_RES, l_monitor_destroy, monitor, HEAP_PRIO_FILTER_NORMAL);
   if (status != HEAP_SUCCESS)
      return status;

   return status;
}

/*****************************************************************************/

HeapStatus
heap_detach_monitor_filter(struct Heap *heap, struct HeapMonitor *monitor)
{
   heap_assert(heap != NULL, "Invalid heap instance passed as parameter.");
   heap_assert(monitor != NULL, "Invalid monitor instance passed as parameter");

   zzheap_detach_hook(heap, HEAP_EVENT_ALLOC_RES,  l_monitor_alloc,   monitor);
   zzheap_detach_hook(heap, HEAP_EVENT_FREE_RES,   l_monitor_free,    monitor);
   zzheap_detach_hook(heap, HEAP_EVENT_QUERY_REQ,  l_monitor_query,   monitor);
   zzheap_detach_hook(heap, HEAP_EVENT_DESTROY_RES,l_monitor_destroy, monitor);
 
   return HEAP_SUCCESS;
}

/*****************************************************************************/
