/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2010-2014 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */


#include "ose.h"
#include "ose_err.h"
#include "osetypes.h"
#include "bios.h"
#include "heap.h"
#include "efs.h"
#include "heap_err.h"
#include "heapapi.h"
#include "ose_heap.h"
#include "mm.h"
#include "mm_util.h"
#include "heap_private.h"
#include "heap.sig"
#include "heap_private.sig"
#include "ose_heapd.sig"

#include "heap_algo_buddy.h"
#include "heap_algo_large.h"
#include "heap_algo_pool.h"
#include "heap_algo_tlsf.h"
#include "heap_alloc_region.h"
#include "heap_filt_eventlog.h"
#include "heap_filt_safe.h"
#include "heap_filt_nobreak.h"
#include "heap_filt_debug.h"
#include "heap_filt_bcache.h"
#include "heap_filt_fcache.h"
#include "heap_filt_tentative.h"

#include "ose_heap_error.h"
#include "ose_heap_monitor.h"
#ifdef OSARM
#include "ose_heap_profile.h"
#endif
#include "ose_heapd.h"

/*****************************************************************************
 * Helper Macros
 ****************************************************************************/
#define H_MAX_SIZE           0x10000000
#define H_SMALL_INIT_SIZE    284000
#define H_SMALL_EXT_SIZE     176000
#define H_EARLY_NULL_LEVEL   100
#define H_LARGE_THRESHOLD    0xF0000000  /* Placeholder until the largest
                                            supported small buffer size has
                                            been determined after installation
                                            of the configured algorithm. */
#define H_CACHE_BUFCOUNT     0x80
#define H_ENDMARK            1

#ifdef OSMIPS
#define PAGE_ALIGN 0x2000
#else
#define PAGE_ALIGN 0x1000
#endif

#define ALIGN(size) (((size) + PAGE_ALIGN - 1) & ~(PAGE_ALIGN-1))

/* Internally configurable plug-ins. Do not use. */
#define HEAP_FILTER_DEBUG    "HEAP_FILTER_DEBUG"
#define HEAP_ALGO_LARGE      "HEAP_ALGO_LARGE"



/*****************************************************************************
 * Forward Declarations
 ****************************************************************************/

OSENTRYPOINT ose_heapcd;
OSENTRYPOINT ose_heapcd_trampoline;

extern void ose_heap_init(void);
static void l_frag_reclaim(union SIGNAL *sig, PROCESS pid,
                           struct OseHeap *heap);
static int  l_tentative_flush(struct Heap *heap);
static int  l_bcache_flush(struct Heap *heap);
static PROCESS heapd_pid(void);


/*****************************************************************************
 * Data Structures
 ****************************************************************************/

union SIGNAL
{
   SIGSELECT                    sig_no;

   /* Private signal interface */
   struct HeapdHeapCreate       heap_create;
   struct HeapdHeapDestroy      heap_destroy;
   struct HeapdClientAdd        client_add;
   struct HeapdClientRem        client_rem;
   struct HeapdMonitorRem       mon_rem;
   struct HeapdInitHeapcd       init_heapcd;

   /* Reclaim interface. */
   struct HeapdFragmentRem      frag_rem;
   struct HeapdFragmentInquiry  frag_inquiry;
   
   /* Public signal interface */
   struct HeapConfigRequest     config_req;
   struct HeapConfigReply       config_rep;
   struct HeapInterfaceRequest  if_req;
   struct HeapInterfaceReply    if_rep;
   struct HeapListRequest       list_req;
   struct HeapListReply         list_rep;

   struct HeapStatsRequest      stats_req;
   struct HeapStatsReply        stats_rep;
   struct HeapStatsExRequest    stex_req;
   struct HeapStatsExReply      stex_rep;
   struct HeapResetStatsRequest reset_req;
   struct HeapResetStatsReply   reset_rep;

   struct HeapFragsRequest      frags_req;
   struct HeapFragsReply        frags_rep;
   
   struct HeapBufferInfoRequest buf_req;
   struct HeapBufferInfoReply   buf_rep;
   struct HeapMonitorRequest    mon_req;
   struct HeapMonitorReply      mon_rep;
   struct HeapMonitorEvent      mon_ev;
   struct HeapFragInfo          frag_info;
};

/****************************************************************************/





/*****************************************************************************
 * Static Data
 ****************************************************************************/

static struct OseHeap *heap_list = NULL;
extern HUCB ose_hucb;

/*****************************************************************************
 * Forward Declarations
 ****************************************************************************/

static void l_client_rem(union SIGNAL *sig, PROCESS pid,
			 struct OseHeap *heap);

/*****************************************************************************
 * OS Adaption Functions
 ****************************************************************************/

void *
heap_meta_alloc(struct Heap *heap, size_t size)
{
   SIGSELECT *ptr = (SIGSELECT *)
      alloc(size + sizeof(SIGSELECT), HEAP_METADATA_SIG);

   if (heap)
   {
      OSHEAP_FROM_HEAP(heap)->stats.meta_size_used +=
         sigsize((union SIGNAL **)&ptr);
   }
   
   return (void *)&ptr[1];
}

/****************************************************************************/

void
heap_meta_free(struct Heap *heap, void *p)
{
   SIGSELECT *ptr = (SIGSELECT *)p;

   ptr = &ptr[-1];

   if (heap)
   {
      OSHEAP_FROM_HEAP(heap)->stats.meta_size_used -=
         sigsize((union SIGNAL **)&ptr);
   }

   free_buf((union SIGNAL **)&ptr);
}

/****************************************************************************/




/*****************************************************************************
 * Helper Functions
 ****************************************************************************/

static void
SET_SEGMENT(struct OseHeap *heap)
{
   if (heap->config.frag_mas)
   {
      set_segment(0, heap->domain);
   }
}

/****************************************************************************/

static void
RESTORE_SEGMENT(struct OseHeap *heap)
{
   if (heap->config.frag_mas)
   {
      set_segment(1, 0);
   }
}

/****************************************************************************/

static inline PROCESS
heapd_pid(void)
{
   static PROCESS pid = 0;

   if (unlikely(pid == 0))
   {
      hunt("ose_heapd", 0, &pid, NULL);
      if (unlikely(pid == 0))
      {
         error(OSE_HEAP_EHUNT_HEAPD_FAILED);
         kill_proc(current_process());
      }
   }

   return pid;
}

/****************************************************************************/

static inline void
heap_log_error(const char *str, U32 status)
{
   switch (status)
   {
      case HEAP_EOUT_OF_MEMORY:
         ramlog_printf("HEAP: Error - %s\n", str);
         ramlog_printf("HEAP: Cause - Out of memory. (%d).\n", status);
         break;
      case HEAP_ENOT_FOUND:
         ramlog_printf("HEAP: Error - %s\n", str);
         ramlog_printf("HEAP: Cause - Heap not found (%d).\n", status);
         break;
      case HEAP_ELOG_MEM_NOT_FOUND:
         ramlog_printf("HEAP: Error - %s\n", str);
         ramlog_printf("HEAP: Cause - Logical memory type not found (%d).\n", status);
         break;
      case HEAP_EPHYS_MEM_NOT_FOUND:
         ramlog_printf("HEAP: Error - %s\n", str);
         ramlog_printf("HEAP: Cause - Physical memory type not found (%d).\n", status);
         break;
      case HEAP_EPLUG_IN_NOT_FOUND:
         ramlog_printf("HEAP: Error - %s\n", str);
         ramlog_printf("HEAP: Cause - Configured plug-in not linked with core image. (%d).\n", status);
         break;
      case HEAP_EABORTED:
         ramlog_printf("HEAP: Error - %s\n", str);
         ramlog_printf("HEAP: Cause - Operation was aborted. (%d).\n", status);
         break;
      case HEAP_EILLEGAL_MAP_TYPE:
         ramlog_printf("HEAP: Error - %s\n", str);
         ramlog_printf("HEAP: Cause - Illegal mapping type for logical memory type (%d).\n", status);
         break;
      case HEAP_EILLEGAL_ALGORITHM:
         ramlog_printf("HEAP: Error - %s\n", str);
         ramlog_printf("HEAP: Cause - An illegal combination of algorithms have been configured (%d).\n", status);
         break;
      case HEAP_EILLEGAL_VALUE:
         ramlog_printf("HEAP: Error - %s\n", str);
         ramlog_printf("HEAP: Cause - An illegal value was encountered in the configuration (%d).\n", status);
         break;
      default:
         ramlog_printf("HEAP: Error - %s (%d)\n", str, status);
         break;
   }
}

/****************************************************************************/

static inline struct OseHeap *
l_heap_alloc(HeapPID owner, size_t size)
{
#ifdef OSMIPS
   char hcb_name[MM_NAME_LENGTH];
   sprintf(hcb_name,"heap_ctrl:%x", owner);
   return (struct OseHeap *)ose_mm_alloc(hcb_name, size, MM_ALLOW_CACHED);
#else
   (void)owner;
   return heap_meta_alloc(NULL, size);
#endif
}

/****************************************************************************/

static inline void
l_heap_free(struct OseHeap *heap)
{
#ifdef OSMIPS
   ose_mm_free((OSADDRESS)heap);
#else
   heap_meta_free(NULL, heap);
#endif
}
   
/****************************************************************************/

static inline void
l_heap_insert(struct OseHeap *heap)                       
{                                               
   heap->prev = NULL;                           
   heap->next = heap_list;                      
   if (heap->next)                              
      heap->next->prev = heap;                  
   heap_list = heap;                            
}

/****************************************************************************/

static inline void
l_heap_remove(struct OseHeap *heap)
{                                               
   if (heap->prev)                              
      heap->prev->next = heap->next;            
   if (heap->next)                              
      heap->next->prev = heap->prev;            
   if (heap_list == heap)                       
      heap_list = heap->next;                   
}

/****************************************************************************/

struct OseHeap *
heapd_heap_find(OSHEAP href)
{
   struct OseHeap *heap = NULL;

   href &= ~1; /* Mask global shared mode bit. */

   for (heap  = heap_list;
        href != (OSHEAP)heap && NULL != heap;
        heap  = heap->next);

   return heap;
}

/****************************************************************************/

struct OseHeap *
heapd_heap_first(void)
{
   return heap_list;
}

/****************************************************************************/

struct OseHeap *
heapd_heap_next(struct OseHeap *heap)
{
   if (heap == NULL)
      return NULL;
   return heap->next;
}

/****************************************************************************/

static PROCESS
l_heap_daemon(struct OseHeap *heap)
{
   /* Returns the pid of the heap daemon if it exists and if it hasn't been 
    * stopped,intercepted or are in an error state. The check of error state
    * was added in OSECR-17255 to prohibit a deadlock.
    
    * The deadlock was occuring when a local heapcd found a corruption in the
    * local heap and called the error handler.If the error handler wants to
    * produce a rmd dump it signals to the rmm and then it waits for the reply
    * from the rmm. The rmm then sends signals to the system heapd and it
    * figures out that it must request data from the local heapcd. The system
    * heapd then sends a signal to the local heapcd but this heapcd is still in 
    * the error handler and it only wants receive the reply signal from the rmm. 
    */

   struct OS_pcb *pcb;
   PROCESS pid = heap->daemon;
   OSADDRESS err_level;

   if (pid == 0)
   {
      /* The daemon has never been started, i.e. no client
       * has been registered. */
      return 0;
   }
   
   err_level = get_pid_info(pid, 33);
   if (err_level > 0)
   {
      /* The heap daemon is in an error state */  
      return 0;
   }

   pcb = get_pcb(pid);
   
   assert(pcb->type != OS_ILLEGAL);
   
   if (pcb->type == OS_ZOOMBIE)
   {
      /* The process has died */     
      pid = 0;
   }
   else
   {             
      if (pcb->status & (OS_STOPPED|OS_INTERCEPTED))
      {
         pid = 0;
      }
   }
   free_buf((union SIGNAL **)&pcb);
   
   return pid;
}

/****************************************************************************/


static void
l_heap_reclaim(union mm_event_t *event)
{
   union SIGNAL *sig;
   
   struct OseHeap *heap;
   uint8_t tentative = 0;

   event->reclaim.base = 0;
   event->reclaim.size = 0;

  restart:
   heap = heapd_heap_first();
   while (heap != NULL)
   {
      if (tentative ^ heap->tentative)
      {
         heap = heapd_heap_next(heap);
         continue;
      }

      if (heap->tentative)
      {
         SET_SEGMENT(heap);
         l_tentative_flush(&heap->heap);
         RESTORE_SEGMENT(heap);
      }

      l_bcache_flush(&heap->heap);
      
      if (heap->stats.frag_size_free > 0 && l_heap_daemon(heap) != 0)
      {
         OSATTREF attref = attach(NULL, heap->daemon);

         if (event->reclaim.inquiry)
         {
            static const SIGSELECT sigsel[] = {2, OS_ATTACH_SIG, HEAPD_FRAG_INQUIRY};
         
            sig = alloc(sizeof(struct HeapdFragmentRem), HEAPD_FRAG_INQUIRY);
            sig->frag_inquiry.href = HREF_FROM_HEAP(heap);
            strlcpy(sig->frag_inquiry.phys_mem, event->reclaim.phys_mem, MM_NAME_LENGTH);

            send(&sig, heap->daemon);
            sig = receive(sigsel);
         }
         else
         {
            static const SIGSELECT sigsel[] = {2, OS_ATTACH_SIG, HEAPD_FRAG_REM};

            sig = alloc(sizeof(struct HeapdFragmentRem), HEAPD_FRAG_REM);
            sig->frag_rem.href = HREF_FROM_HEAP(heap);
            strlcpy(sig->frag_rem.phys_mem, event->reclaim.phys_mem, MM_NAME_LENGTH);

            send(&sig, heap->daemon);
            sig = receive(sigsel);
         }

         switch (sig->sig_no)
         {
            case HEAPD_FRAG_REM:
            {
               detach(&attref);
               
               if (sig->frag_rem.size != 0)
               {
                  event->reclaim.domain = (OSDOMAIN)heap->domain;
                  event->reclaim.base   = (log_addr_t)sig->frag_rem.base;
                  event->reclaim.size   = (size_t)sig->frag_rem.size;
                  free_buf(&sig);
                  return;
               }

               break;
            }
            case HEAPD_FRAG_INQUIRY:
            {
               detach(&attref);
               
               event->reclaim.size += (size_t)sig->frag_inquiry.size;
               free_buf(&sig);
               break;
            }
            
            default:
               free_buf(&sig);
               break;
         }
      }

      heap = heapd_heap_next(heap);
   }

   if (!tentative)
   {
      tentative = 1;
      goto restart;
   }
}

/*****************************************************************************/

static inline uint32_t
l_heap_count(void)
{
   struct OseHeap *heap;
   uint32_t count = 0;

   heap = heapd_heap_first();
   while (heap != NULL)
   {
      count++;
      heap = heapd_heap_next(heap);
   }
   
   return count;
}

/****************************************************************************/

static uint32_t
l_heap_cfg(PROCESS pid, const char *name,
           U32 val, OSBOOLEAN align, U32 *errOut)
{
   char *env;

   assert(name != NULL);

   if (errOut != NULL)
      *errOut = 0;
   
   env = get_env(pid, name);
   if (env != NULL)
   {
      const char *errPtr = NULL;
      val = (U32) efs_str_to_long(env, &errPtr);
      free_buf((union SIGNAL **)&env);

      if (errPtr != NULL && errOut != NULL)
         *errOut = 1;
   }

   if(align)
   {
      val = ALIGN(val);
   }

   return val;
}

/****************************************************************************/

static uint32_t
l_heap_cfg32(PROCESS pid, const char *name,
             size_t val, OSBOOLEAN align, U32 *errOut)
{
   val = l_heap_cfg(pid, name, val, align, errOut);
   if (val > ~0U)
   {
      *errOut = 1;
   }
   return (uint32_t)val;
}

/****************************************************************************/

static uint32_t
l_heap_cfg_is_set(PROCESS pid, const char *name)
{
   uint32_t valid = 0;
   char *env;

   assert(name != NULL);

   env = get_env(pid, name);
   if (env != NULL)
   {
      valid = 1;
      free_buf((union SIGNAL **)&env);
   }

   return valid;
}

/****************************************************************************/

static inline uint8_t
l_heap_cfg_bool(PROCESS pid, const char *name, U32 val)
{
   char *env = get_env(pid, name);
   
   if (env != NULL)
   {
      env = strlwr(env);
      if (strcasecmp(env, "true")   == 0 ||
          strcasecmp(env, "yes")    == 0 ||
          strcasecmp(env, "enable") == 0 ||
          strcasecmp(env, "1")      == 0)
      {
         val = 1;
      }
      else
      {
         val = 0;
      }
      free_buf((union SIGNAL **)&env);
   }

   return val;
}

/****************************************************************************/

PROCESS
heapd_daemon_create(OSDOMAIN domain, PROCESS bid,
                    PROCESS ownerpid, OSENTRYPOINT *entry,
                    struct OseHeap *heap)
{
   /* A few notes on the creation of heap daemons:
    *  
    *     Heaps in application domains are assigned to a newly
    *     created slave daemon. This slave must, regardless of
    *     architecture, trap into supervisor mode where it will
    *     operate for the remainder of its life. On MIPS, however,
    *     the slave is not created until the first heap client
    *     is registered, in ose_heap_iface.c. The slave process
    *     trampoline, which makes the trap, must be part of the
    *     CRT since user mode execution in the monolith is illegal.
    *     An environment variable is therefore used to pass the
    *     address of the process entry point to the client registration
    *     function in the monolith where the trampoline is created.
    *     Its really uggly, but backwards compatible. It would have
    *     been better if the CRT had created the trampoline, for
    *     example in ose_create_heap. That way we could have avoided
    *     all special architecture dependent handling of the matter.
    */
   
   int  sstack;
   char name[64];
   PROCESS pid;
   OSPRIORITY pri;
   union SIGNAL *sig;

   if (entry == NULL)
      return 0;

   sstack = l_heap_cfg(ownerpid, HEAP_CLEANUPD_STACK_SIZE, 2000, False, NULL);
   

   /* Match priority of ose_heapd if created in system domain. */
   if (domain == MM_KERNEL_DOMAIN && heapd_pid())
   {
      pri = get_pri(heapd_pid());
   }
   else
   {
      pri = 31;
   }
   
   snprintf(name, 64, "ose_heapcd_%x:%x", domain, ownerpid);

   /* Note that all heaps now are provided with an individual cleanup
    * daemon, even if the heap resides in the system domain. This gives
    * symmetry and recoverability, but also solves a deadlock problem
    * during heap creation if the reclaim mechanism is triggered when
    * the initial fragment is allocated. */

   pid = create_process(OS_PRI_PROC, name, entry, sstack, pri, 0, bid, 0, 0, 0);
   sig = alloc(sizeof (struct HeapdInitHeapcd), HEAPD_INIT_HEAPCD);
   sig->init_heapcd.heap = heap;
   send(&sig, pid);

   /* The process is started when the first client is registered. */

   return pid;
}

/**************************************************************************/

static inline void
l_buf_size(struct Heap *heap, void *buf, size_t *req, size_t *act, int *avail)
{
   HeapParam p[7] = {
      HPM(QUERY_BUF_ACT_SIZE, 1), 0,
      HPM(QUERY_BUF_REQ_SIZE, 1), 0,
      HPM(QUERY_BUF_STATUS,   1), 0,
      HPM_END
   };

   p[1] = (HeapParam)buf;
   p[3] = (HeapParam)buf;
   p[5] = (HeapParam)buf;

   zzheap_query(heap, p, 0);

   if (HPM_IS_HANDLED(p[0]))
      *act = (size_t)p[1];
   if (HPM_IS_HANDLED(p[2]))
      *req = (size_t)p[3];
   if (HPM_IS_HANDLED(p[4]))
      *avail = !(p[5] & HEAP_BUFFER_ALLOCATED);
}

/****************************************************************************/

static inline int
l_tentative_flush(struct Heap *heap)
{
   HeapParam p[] = {HPM(CTRL_TENTATIVE_FLUSH, 0), HPM_END};
   zzheap_control(heap, p);
   return HPM_IS_HANDLED(p[0]);
}

/*****************************************************************************/

static inline int
l_bcache_flush(struct Heap *heap)
{
   HeapParam p[] = {HPM(CTRL_BCACHE_FLUSH, 0), HPM_END};
   zzheap_control(heap, p);
   return HPM_IS_HANDLED(p[0]);
}

/*****************************************************************************/

static inline int
l_frag_next(struct Heap *heap, void **frag, size_t *size, HeapParam *p)
{
   p[0] = HPM(QUERY_FRAG_NEXT, 4);
   p[5] = HPM_END;

   zzheap_query(heap, p, 0);

   if (!HPM_IS_HANDLED(p[0]))
      return 0;

   *frag   = (void *)p[1];
   *size   = (size_t)p[2];

   return 1;
}

/****************************************************************************/

static inline int
l_frag_next_free(struct Heap *heap, void **frag, size_t *size, HeapParam *p)
{
   p[0] = HPM(QUERY_FRAG_NEXT_FREE, 4);
   p[5] = HPM_END;

   zzheap_query(heap, p, 0);

   if (!HPM_IS_HANDLED(p[0]))
      return 0;

   *frag   = (void *)p[1];
   *size   = (size_t)p[2];

   return 1;
}

/****************************************************************************/

static inline size_t
l_frag_size(struct Heap *heap, OSADDRESS base)
{
   HeapParam p[] = {
      HPM(QUERY_FRAG_SIZE, 2), 0, 0,
      HPM_END
   };

   p[1] = (HeapParam)base;

   zzheap_query(heap, p, 0);

   if (!HPM_IS_HANDLED(p[0]))
      return 0;

   return (size_t)p[2];
}

/****************************************************************************/

static inline int
l_frag_buf_next(struct Heap *heap, OSADDRESS base, void **buf, HeapParam *p)
{
   /* Buffer iteration is limited
      to one fragment only. */

   p[0] = HPM(QUERY_BUF_FRAG_NEXT, 4);
   p[2] = (HeapParam)base;
   p[5] = HPM_END;

   zzheap_query(heap, p, 0);

   if (!HPM_IS_HANDLED(p[0]))
      return 0;

   *buf = (void *)p[1];

   return 1;
}

/****************************************************************************/

static inline int
l_frag_status(struct Heap *heap, void *frag, HeapFragmentStatus *status)
{
   HeapParam p[] = {
      HPM(QUERY_FRAG_STATUS, 2), 0, 0,
      HPM_END
   };

   p[1] = (HeapParam)frag;

   zzheap_query(heap, p, 0);

   if (!HPM_IS_HANDLED(p[0]))
      return 0;

   *status = (size_t)p[2];
   
   return 1;
}

/****************************************************************************/

static inline int
l_frag_pmem(struct Heap *heap, void *frag, char *phys_mem)
{
   HeapParam p[] = {
      HPM(QUERY_FRAG_PMEM, 2), 0, 0,
      HPM_END
   };

   p[1] = (HeapParam)frag;
   p[2] = (HeapParam)phys_mem;

   zzheap_query(heap, p, 0);

   if (!HPM_IS_HANDLED(p[0]))
      return 0;

   return 1;
}

/****************************************************************************/

static inline int
l_frag_remove(struct Heap *heap, void **frag, size_t *size)
{
   HeapParam p[] = {
      HPM(CTRL_FRAG_REMOVE, 2), 0, 0,
      HPM_END
   };

   p[1] = (HeapParam)*frag;

   zzheap_control(heap, p);

   if (!HPM_IS_HANDLED(p[0]))
      return 0;

   *frag = (void *)p[1];
   *size = (size_t)p[2];
   
   return 1;
}

/****************************************************************************/


/*****************************************************************************
 * Configuration Helpers
 ****************************************************************************/

static HeapStatus
l_heap_cfg_locks(struct OseHeap *heap, PROCESS pid)
{
   HeapLockType type;
   U32 err;

   type = l_heap_cfg(pid, HEAP_LOCK_TYPE, EHEAP_LOCK_INT, 0, &err);
   if (err) return HEAP_EILLEGAL_VALUE;

   if (type > EHEAP_LOCK_MUTEX)
      type = EHEAP_LOCK_INT;

   heap_lock_init(&heap->heap, type);

   return HEAP_SUCCESS;
}

/****************************************************************************/

static HeapStatus
l_heap_cfg_sizes(struct OseHeap *heap, PROCESS pid)
{
   size_t sinit, sext, slarge, sendm, smax, tmax;
   U32 err;

   slarge = l_heap_cfg(pid, HEAP_LARGE_BUF_THRESHOLD, H_LARGE_THRESHOLD, 1, &err);
   if (err) return HEAP_EILLEGAL_VALUE;
   
   sendm  = l_heap_cfg(pid, HEAP_ENDMARK_SIZE, H_ENDMARK, 0, &err);
   if (err) return HEAP_EILLEGAL_VALUE;
   
   smax   = l_heap_cfg(pid, "OSE_LM_HEAP_SIZE", H_MAX_SIZE, 1, &err);
   if (err) return HEAP_EILLEGAL_VALUE;

   smax   = l_heap_cfg(pid, HEAP_MAX_SIZE, smax, 1, &err);
   if (err) return HEAP_EILLEGAL_VALUE;

   sinit  = l_heap_cfg(pid, "OSE_LM_HEAP_SIZE", H_SMALL_INIT_SIZE, 1, &err);
   if (err) return HEAP_EILLEGAL_VALUE;

   sinit  = l_heap_cfg(pid, HEAP_SMALL_BUF_INIT_SIZE, sinit, 1, &err);
   if (err) return HEAP_EILLEGAL_VALUE;

   tmax   = l_heap_cfg(pid, HEAP_MAX_TYPE, 1 /* FRAG_MAX */, 0, &err);
   if (err) return HEAP_EILLEGAL_VALUE;

   if (sinit > smax)
      sinit = smax;

   if (!l_heap_cfg_is_set(pid, HEAP_SMALL_BUF_EXT_SIZE))
   {
      sext = sinit / 5;
   }
   else
   {
      sext = l_heap_cfg(pid, HEAP_SMALL_BUF_EXT_SIZE,  H_SMALL_EXT_SIZE, 1, &err);
      if (err) return HEAP_EILLEGAL_VALUE;
   }

   if (sext > smax)
      sext = smax;

   if (sinit > smax)
      sinit = smax;

   /* Change large buffer threshold to 0 if no
    * small buffer algorithm will be installed. */

   if (!l_heap_cfg_bool(pid, HEAP_ALGO_TLSF, 1) &&
       !l_heap_cfg_bool(pid, HEAP_ALGO_BUDDY,0))
   {
      slarge = 0;
   }
  
   /* Sizes are further adjusted by individual algorithms
    * to suite their needs. For example, the buddy algorithm
    * adjusts sizes to match the Fibonacci series. */

   switch (tmax)
   {
      
      case 1: /* HEAP_MAX_TYPE_FRAGMENTS */
      default:
         heap->config.buf_size_max  = 0xF0000000;
         heap->config.frag_size_max = smax;
         break;
      case 2: /* HEAP_MAX_TYPE_BUFFERS */
         heap->config.frag_size_max = 0xF0000000;
         heap->config.buf_size_max  = smax;
         break;
   }

   heap->config.max_type           = tmax;
   heap->config.frag_size_init     = sinit;
   heap->config.frag_size_default  = sext;
   heap->config.buf_size_threshold = slarge;

   /* The endmark size is configurable in the interval 1 <= size <= 4 bytes.
    *
    * The user data area must be at least 4 bytes large,
    * due in part to the implementation of garbage collect.
    * The endmark can therefore not be larger than 4 bytes
    * (16 byte buffer - 8 byte header - 4 byte endmark = 4 byte udata) */
   heap->config.buf_size_endmark = sendm >= 1 ? (sendm > 4 ? 4 : sendm) : 1;

   return HEAP_SUCCESS;
}

/****************************************************************************/

static HeapStatus
l_heap_cfg_adjust(struct OseHeap *heap, PROCESS pid)
{
   HeapParam p[] = {0, 0,HPM_END};

   (void)pid;

   /* Adjust the initial size to fit into max size. */
   p[0] = HPM(QUERY_FRAG_PREV_SIZE,1);
   p[1] = ALIGN(heap->config.frag_size_init);
   zzheap_query(&heap->heap, p, 0);
   heap->config.frag_size_init = p[1];

   /* Adjust the extension size. */
   p[0] = HPM(QUERY_FRAG_PREV_SIZE,1);
   p[1] = ALIGN(heap->config.frag_size_default);
   zzheap_query(&heap->heap, p, 0);
   heap->config.frag_size_default = p[1];

   return HEAP_SUCCESS;
}

/****************************************************************************/

static HeapStatus
l_heap_cfg_safe(struct OseHeap *heap, PROCESS pid)
{
   uint8_t ona, onb;

   ona = onb = l_heap_cfg_bool(pid, HEAP_MALLOC_ERROR_ON_NULL, 0);

   /* Cache the configuration locally for faster retrieval. */
   if (ona)
      heap->buf_error_on_null = 1;

   return heap_attach_safe_filter(&heap->heap, ona, onb);
}

/****************************************************************************/

static HeapStatus
l_heap_cfg_eventlog(struct OseHeap *heap, PROCESS pid)
{
   uint32_t eflags;
   
   if (heap_attach_event_log == NULL)
      return HEAP_SUCCESS;

   eflags = l_heap_cfg32(pid, HEAP_FILTER_LOG, 1, 0, NULL);
   if (!eflags)
      return HEAP_SUCCESS;

   return heap_attach_event_log(&heap->heap, eflags);
}

/****************************************************************************/

static HeapStatus
l_heap_cfg_error(struct OseHeap *heap, PROCESS pid)
{
   (void)pid;

   return heap_attach_ose_error_filter(&heap->heap);
}

/****************************************************************************/

static HeapStatus
l_heap_cfg_nobreak(struct OseHeap *heap, PROCESS pid)
{
   (void)pid;

   return heap_attach_nobreak_filter(&heap->heap);
}

/****************************************************************************/

static HeapStatus
l_heap_cfg_debug(struct OseHeap *heap, PROCESS pid, uint32_t *algo)
{
   (void)pid;

   if (!l_heap_cfg_bool(pid, HEAP_FILTER_DEBUG, 0))
      return HEAP_SUCCESS;

   /* FIXME: This is not a critical error. The heap will be
    * functional. Should maybe pass unnoticed? */
   if (heap_attach_debug_filter == NULL)
      return HEAP_EPLUG_IN_NOT_FOUND;

   return heap_attach_debug_filter(
      &heap->heap, (*algo)++, HEAP_DEBUG_DEFER|HEAP_DEBUG_PATTERN);
}

/****************************************************************************/
#ifdef OSARM
#if OSE_DEBUG == 1
static HeapStatus
l_heap_cfg_profile(struct OseHeap *heap, PROCESS pid)
{
   (void)pid;

   if (heap_attach_profile_filter == NULL)
      return HEAP_EPLUG_IN_NOT_FOUND;

   return heap_attach_profile_filter(&heap->heap, &heap->stats,
                                     &heap->config, HREF_FROM_HEAP(heap));
}
#endif
#endif

/****************************************************************************/

static HeapStatus
l_heap_cfg_buddy(struct OseHeap *heap, PROCESS pid, uint32_t *algo, uint32_t *alloc)
{
   HeapStatus status;

   if (!l_heap_cfg_bool(pid, HEAP_ALGO_BUDDY, 0))
      return HEAP_SUCCESS;
   
   /* Large buffer threshold set to 0, disable small buffers. */
   if (heap->config.buf_size_threshold == 0)
      return HEAP_SUCCESS;

   if (heap_attach_buddy_algorithm == NULL)
      return HEAP_EPLUG_IN_NOT_FOUND;

   /* Round down to closest fibonacci number if extend size hasn't
    * been explicitly configured, i.e. round_down((max or init)/5); */
   if (!l_heap_cfg_is_set(pid, HEAP_SMALL_BUF_EXT_SIZE))
   {
      HeapParam p[] = {0, 0,HPM_END};
      
      /* Adjust the extension size. */
      size_t size = heap_buddy_size(heap->config.frag_size_default);

      size = ALIGN(size);

      p[0] = HPM(QUERY_FRAG_PREV_SIZE,1);
      p[1] = size;
      zzheap_query(&heap->heap, p, 0);
      heap->config.frag_size_default = p[1];
   }
   
   status = heap_attach_buddy_algorithm(
      &heap->heap, &heap->stats, &heap->config, (*algo)++, (*alloc)++);
   
   return status;
}

/****************************************************************************/

static HeapStatus
l_heap_cfg_tlsf(struct OseHeap *heap, PROCESS pid, uint32_t *algo)
{
   HeapStatus status;

   if (!l_heap_cfg_bool(pid, HEAP_ALGO_TLSF, 1))
      return HEAP_SUCCESS;
   
   /* Make sure the user hasn't explicitly enabled both algorithms. */
   if (l_heap_cfg_bool(pid, HEAP_ALGO_BUDDY, 0) &&
       l_heap_cfg_bool(pid, HEAP_ALGO_TLSF,  0))
      return HEAP_EILLEGAL_ALGORITHM;

   if (l_heap_cfg_bool(pid, HEAP_ALGO_BUDDY, 0))
      return HEAP_SUCCESS;
   
   /* Large buffer threshold set to 0, disable small buffers. */
   if (heap->config.buf_size_threshold == 0)
      return HEAP_SUCCESS;

   if (heap_attach_tlsf_algorithm == NULL)
      return HEAP_EPLUG_IN_NOT_FOUND;
   
   status = heap_attach_tlsf_algorithm(
      &heap->heap, &heap->stats, &heap->config, (*algo)++);
   
   return status;
}

/****************************************************************************/

#if 0
static HeapStatus
l_heap_cfg_pool(struct OseHeap *heap, PROCESS pid, uint32_t *algo)
{
   if (!l_heap_cfg_bool(pid, HEAP_ALGO_POOL, 0))
      return HEAP_SUCCESS;

   if (heap_attach_pool_algorithm == NULL)
      return HEAP_EPLUG_IN_NOT_FOUND;

   /* Large buffer threshold set to 0, disable small buffers. */
   if (heap->config.buf_size_threshold == 0)
      return HEAP_SUCCESS;

   return heap_attach_pool_algorithm(
      &heap->heap, &heap->stats, &heap->config, (*algo)++);
}
#endif

/****************************************************************************/

static HeapStatus
l_heap_cfg_large(struct OseHeap *heap, PROCESS pid, uint32_t *algo)
{
   if (!l_heap_cfg_bool(pid, HEAP_ALGO_LARGE, 1))
      return HEAP_SUCCESS;

   if (heap_attach_large_algorithm == NULL)
      return HEAP_EPLUG_IN_NOT_FOUND;
   
   /* Large buffer algorithm is always installed as a backup. */
   return heap_attach_large_algorithm(
      &heap->heap, &heap->stats, &heap->config, (*algo)++);
}

/****************************************************************************/

static HeapStatus
l_heap_cfg_region(struct OseHeap *heap, PROCESS pid, uint32_t *alloc_id,
                  OSDOMAIN domain, OSBOOLEAN supervisor)
{
   uint32_t pmode, cmode;
   size_t slim, llim;
   char *lmems, *pmems, *tmp;
   HeapStatus status;
   U32 err;

   if (heap_attach_region_allocator == NULL)
      return HEAP_EPLUG_IN_NOT_FOUND;
   
   pmode = l_heap_cfg(pid, HEAP_PERMISSIONS, supervisor ? SU_RW_USR_RO : SU_RW_USR_RW, False, &err);
   if (err) return HEAP_EILLEGAL_VALUE;

   cmode = l_heap_cfg(pid, HEAP_CACHE_MODE, MM_CACHE_COPY_BACK|MM_SPECULATIVE_ACCESS, False, &err);
   if (err) return HEAP_EILLEGAL_VALUE;

   
   slim = l_heap_cfg(pid, HEAP_SMALL_LOW_LIMIT, 0, True, &err);
   if (err) return HEAP_EILLEGAL_VALUE;

   llim = l_heap_cfg(pid, HEAP_LARGE_LOW_LIMIT, 0, True, &err);
   if (err) return HEAP_EILLEGAL_VALUE;

   lmems = get_env(pid, HEAP_LOG_MEM);
   if (lmems == NULL)
   {
      lmems = (char*)alloc(MM_NAME_LENGTH, 0);
      strlcpy(lmems, MM_INTERNAL_RAM, MM_NAME_LENGTH);
   }

   pmems = get_env(pid, HEAP_PHYS_MEM);
   if (pmems == NULL)
   {
      pmems = (char*)alloc(MM_NAME_LENGTH, 0);
      strlcpy(pmems, MM_INTERNAL_RAM, MM_NAME_LENGTH);
   }
      
   /* Cache some non-generic configuration parameters for later retrieval. */
   heap->frag_slow_limit  = slim;
   heap->frag_llow_limit  = llim;
   heap->frag_permissions = pmode;
   heap->frag_cache_mode  = cmode;

   /* Cache the first configured log_mem in the list. */
   strlcpy(heap->frag_log_mem, lmems, MM_NAME_LENGTH);
   tmp = strchr(heap->frag_log_mem, ',');
   if (tmp != NULL) *tmp = 0;

   status = heap_attach_region_allocator(
      &heap->heap, &heap->stats, &heap->config, (*alloc_id)++, domain,
      pid, lmems, pmems, cmode, pmode, slim, llim);

   if (lmems)
      free_buf((union SIGNAL**) &lmems);
   if (pmems)
      free_buf((union SIGNAL**) &pmems);

   return status;
}

/****************************************************************************/

static HeapStatus
l_heap_cfg_bcache(struct OseHeap *heap, PROCESS pid)
{
   uint32_t nmax, smax;
   U32 err;

   /* HEAP_BUFFER_LARGE will prevent installation
    * the cache if no small buffer algorithm has
    * been configured. */
   HeapBufferType type = HEAP_BUFFER_LARGE; 
   
   if (heap_attach_bcache_filter == NULL)
      return HEAP_EPLUG_IN_NOT_FOUND;

   nmax = l_heap_cfg(pid, HEAP_CACHE_MAX_BUFCOUNT, 0, 0, &err);
   if (err) return HEAP_EILLEGAL_VALUE;
   if (nmax <= 0) return HEAP_SUCCESS;

   smax = l_heap_cfg(pid, HEAP_CACHE_MAX_BUFSIZE, 32, 0, &err);
   if (err) return HEAP_EILLEGAL_VALUE;
   if (smax <= 0) return HEAP_SUCCESS;

   if (l_heap_cfg_bool(pid, HEAP_ALGO_TLSF, 1))
      type = HEAP_BUFFER_TLSF;

   if (l_heap_cfg_bool(pid, HEAP_ALGO_BUDDY,0))
      type = HEAP_BUFFER_FBUDDY;
   
   return heap_attach_bcache_filter(
      &heap->heap, &heap->stats, &heap->config,
      nmax, smax, type, &heap->daemon);
}

/****************************************************************************/

static HeapStatus
l_heap_cfg_fcache(struct OseHeap *heap, PROCESS pid, uint32_t *alloc_id)
{
   if (heap_attach_fcache_filter == NULL)
      return HEAP_EPLUG_IN_NOT_FOUND;
   
   if (!l_heap_cfg_bool(pid, HEAP_FILTER_FRAG_CACHE, 0))
      return HEAP_SUCCESS;

   return heap_attach_fcache_filter(
      &heap->heap, &heap->stats, (*alloc_id)++, 0);
}

/****************************************************************************/

static HeapStatus
l_heap_cfg_tentative(struct OseHeap *heap, PROCESS pid)
{
   uint32_t count;
   HeapStatus status;
   U32 err;

   count = l_heap_cfg(pid, HEAP_TENTATIVE_BUFCOUNT, 0, 0, &err);
   if (err) return HEAP_EILLEGAL_VALUE;

   if (count == 0)
      return HEAP_SUCCESS;
   
   if (heap_attach_tentative_filter == NULL)
      return HEAP_EPLUG_IN_NOT_FOUND;

   status = heap_attach_tentative_filter(&heap->heap, count);

   heap->tentative = (status == HEAP_SUCCESS);

   return status;
}

/****************************************************************************/

#define LOG_CFG(p, name)                                \
do {                                                    \
   char *env;                                           \
                                                        \
   env = get_env(pid, name);                            \
   if (env != NULL)                                     \
   {                                                    \
      ramlog_printf("HEAP: %-40s %s\n", name, env);     \
      free_buf((union SIGNAL **)&env);                  \
   }                                                    \
} while(0)

static void
l_heap_cfg_log(PROCESS pid)
{
   LOG_CFG(pid, HEAP_SMALL_BUF_INIT_SIZE);
   LOG_CFG(pid, HEAP_SMALL_BUF_EXT_SIZE);
   LOG_CFG(pid, HEAP_LARGE_BUF_THRESHOLD);
   LOG_CFG(pid, HEAP_MAX_SIZE);
   LOG_CFG(pid, HEAP_MAX_TYPE);
   LOG_CFG(pid, HEAP_CACHE_MODE);
   LOG_CFG(pid, HEAP_PERMISSIONS);
   LOG_CFG(pid, HEAP_LOG_MEM);
   LOG_CFG(pid, HEAP_PHYS_MEM);
   LOG_CFG(pid, HEAP_EARLY_NULL_LEVEL);
   LOG_CFG(pid, HEAP_ABSOLUTE_WATERMARK);
   LOG_CFG(pid, HEAP_TENTATIVE_BUFCOUNT);
   LOG_CFG(pid, HEAP_CLEANUPD_STACK_SIZE);
   LOG_CFG(pid, HEAP_FRAG_CHECK);
   LOG_CFG(pid, HEAP_RECLAIM);
   LOG_CFG(pid, HEAP_CACHE_MAX_BUFCOUNT);
   LOG_CFG(pid, HEAP_CACHE_MAX_BUFSIZE);
   LOG_CFG(pid, HEAP_ALGO_TLSF);
   LOG_CFG(pid, HEAP_ALGO_BUDDY);
   LOG_CFG(pid, HEAP_LOCK_TYPE);
   LOG_CFG(pid, HEAP_SHARED_MODE);
   LOG_CFG(pid, HEAP_SMALL_LOW_LIMIT);
   LOG_CFG(pid, HEAP_LARGE_LOW_LIMIT);
   LOG_CFG(pid, HEAP_ANNOTATION_SIZE);
   LOG_CFG(pid, HEAP_ENDMARK_SIZE);
   LOG_CFG(pid, HEAP_WIPE_ON_FREE);
   LOG_CFG(pid, HEAP_MALLOC_ERROR_ON_NULL);
   LOG_CFG(pid, HEAP_FILTER_FRAG_CACHE);
}
#undef LOG_CFG

/****************************************************************************/



/*****************************************************************************
 * Private Interface / Signal Implementation
 ****************************************************************************/

static void
l_heapd_interface(union SIGNAL *sig, PROCESS pid)
{
   struct HeapInterfaceReply *reply;
   const U32 numSigs = 10;

   free_buf(&sig);

   reply = (struct HeapInterfaceReply *)
      alloc(sizeof(struct HeapInterfaceReply) + numSigs *
              sizeof(SIGSELECT), HEAP_INTERFACE_REPLY);
   reply->status = HEAP_SUCCESS;
   strncpy(reply->whatStr, "OSE5_nheap", sizeof reply->whatStr);
   reply->sigs[0] = numSigs;
   reply->sigs[1] = HEAP_INTERFACE_REQUEST;
   reply->sigs[2] = HEAP_LIST_REQUEST;
   reply->sigs[3] = HEAP_STATS_REQUEST;
   reply->sigs[4] = HEAP_BUFFER_INFO_REQUEST;
   reply->sigs[5] = HEAP_CONFIG_REQUEST;
   reply->sigs[6] = HEAP_STATSEX_REQUEST;
   reply->sigs[7] = HEAP_FRAGS_REQUEST;
   reply->sigs[8] = HEAP_RESET_STATS_REQUEST;
   reply->sigs[9] = HEAP_MONITOR_REQUEST;
   reply->sigs[10]= HEAP_FRAG_INFO_REQUEST;
   send((union SIGNAL **)&reply, pid);
}

/*****************************************************************************/

static inline union SIGNAL *
l_heap_create(union SIGNAL *sig)
{
   union SIGNAL *att_sig;
   struct OseHeap *heap;
   HeapStatus status;
   uint32_t algo_id  = 0, alloc_id = 0;

   HeapPID   pid = sig->heap_create.owner;
   OSDOMAIN  domain = sig->heap_create.domain;
   OSBOOLEAN supervisor = sig->heap_create.supervisor;


   heap = l_heap_alloc(pid, sizeof(struct OseHeap));
   if (heap == NULL)
   {
      status = HEAP_EOUT_OF_MEMORY;
      goto cleanup;
   }

   memset(heap, 0, sizeof(struct OseHeap));

   status = zzheap_create(&heap->heap);
   if (status != HEAP_SUCCESS)
   {
      l_heap_free(heap);
      heap = NULL;
      goto cleanup;
   }

#ifdef OSMIPS
   heap->stats.meta_size_used += PAGE_ALIGN;
#else
   heap->stats.meta_size_used += sizeof(struct OseHeap);
#endif
   
   /*
    * MISC INIT
    *****************************************/

   memcpy(&heap->hucb, &ose_hucb, sizeof(HUCB));

   status = l_heap_cfg_locks(heap, pid);
   if (status != HEAP_SUCCESS)
   {
      l_heap_free(heap);
      heap = NULL;
      goto cleanup;
   }
   
   status = l_heap_cfg_sizes(heap, pid);
   if (status != HEAP_SUCCESS)
   {
      l_heap_free(heap);
      heap = NULL;
      goto cleanup;
   }
   
   heap->owner   = pid;
   heap->domain  = domain;
   heap->shared  = get_sysconf(64,get_bid(pid),0,0);
   heap->shared  = l_heap_cfg_bool(pid, HEAP_SHARED_MODE, heap->shared);

   /* The heap daemon is started when the first client
    * is registered, see details in ose_heap_iface.c. */
   heap->daemon  = 0; 

#ifdef OSMIPS
   heap->daemon_stopped  = 0;
#else
   /* Only created but not started. */
   heap->daemon_stopped  = heapd_daemon_create(
      domain, get_bid(pid), pid, ose_heapcd_trampoline, heap);
#endif

   heap->config.frag_check   = l_heap_cfg_bool(pid, HEAP_FRAG_CHECK, 0);
   heap->config.frag_reclaim = l_heap_cfg_bool(pid, HEAP_RECLAIM, 0);

   
   /*
    * ALLOCATORS
    *****************************************/
   
   status = l_heap_cfg_region(heap, pid, &alloc_id, domain, supervisor);
   if (status != HEAP_SUCCESS)
      goto cleanup;

   status = l_heap_cfg_bcache(heap, pid);
   if (status != HEAP_SUCCESS)
      goto cleanup;

   status = l_heap_cfg_fcache(heap, pid, &alloc_id);
   if (status != HEAP_SUCCESS)
      goto cleanup;

   /* Adjust some sizes to match capabilities of the allocator. */
   status = l_heap_cfg_adjust(heap, pid);
   if (status != HEAP_SUCCESS)
      goto cleanup;

   /*
    * ALGORITHMS
    *****************************************/
   
   /* It is important that an allocator has been installed
    * before any algorithm is installed. The algorithms may
    * use the incomplete heap to break initial fragments. */

#ifdef OSARM
   zzset_manager_mode(TRUE);
#endif
   set_segment(0, domain);

   /* The large buffer algorithm must be assigned to the first
    * algorithm identifier to make the "low_limit" implementation
    * work. */
   status = l_heap_cfg_large(heap, pid, &algo_id);
   if (status != HEAP_SUCCESS) {
#ifdef OSARM
      zzset_manager_mode(FALSE);
#endif
      set_segment(1, 0);
      goto cleanup;
   }

   status = l_heap_cfg_buddy(heap, pid, &algo_id, &alloc_id);
   if (status != HEAP_SUCCESS) {
#ifdef OSARM
      zzset_manager_mode(FALSE);
#endif
      set_segment(1, 0);
      goto cleanup;
   }

   status = l_heap_cfg_tlsf(heap, pid, &algo_id);
   if (status != HEAP_SUCCESS) {
#ifdef OSARM
      zzset_manager_mode(FALSE);
#endif
      set_segment(1, 0);
      goto cleanup;
   }

#if 0   
   status = l_heap_cfg_pool(heap, pid, &algo_id);
   if (status != HEAP_SUCCESS) {
#ifdef OSARM
      zzset_manager_mode(FALSE);
#endif
      set_segment(1, 0);
      goto cleanup;
   }
#endif

#ifdef OSARM
   zzset_manager_mode(FALSE);
#endif
   set_segment(1, 0);

   /*
    * FILTERS
    *****************************************/
   status = l_heap_cfg_tentative(heap, pid);
   if (status != HEAP_SUCCESS)
      goto cleanup;

   status = l_heap_cfg_safe(heap, pid);
   if (status != HEAP_SUCCESS)
      goto cleanup;
   
   status = l_heap_cfg_eventlog(heap, pid);
   if (status != HEAP_SUCCESS)
      goto cleanup;

   status = l_heap_cfg_nobreak(heap, pid);
   if (status != HEAP_SUCCESS)
      goto cleanup;

   status = l_heap_cfg_error(heap, pid);
   if (status != HEAP_SUCCESS)
      goto cleanup;

   status = l_heap_cfg_debug(heap, pid, &algo_id);
   if (status != HEAP_SUCCESS)
      goto cleanup;

#ifdef OSARM
#if OSE_DEBUG == 1
   status = l_heap_cfg_profile(heap, pid);
   if (status != HEAP_SUCCESS)
   {
      /* Not critical plug-in, continue with heap creation */
      ramlog_printf("HEAP: Profile filter plug-in uninstalled -"
                    " Accumulated heap usage user reports not created\n");
   }
#endif
#endif

   /*
    * REGISTER
    *****************************************/

   l_heap_insert(heap);

   att_sig = alloc(sizeof(struct HeapdHeapDestroy), HEAPD_HEAP_DESTROY);
   att_sig->heap_destroy.href = HREF_FROM_HEAP(heap);
   attach(&att_sig, pid);
   

   /*
    * NOTIFY
    *****************************************/

   set_envp(pid, HEAPREF_ENVNAME, HREF_FROM_HEAP(heap));

   sig->heap_create.status = HEAP_SUCCESS;
   sig->heap_create.href   = HREF_FROM_HEAP(heap);
   
   return sig;

cleanup:
   l_heap_cfg_log(pid);
   
   if (heap != NULL)
   {
      set_segment(0, heap->domain);
      zzheap_destroy(&heap->heap);
      set_segment(1, 0);
   
      l_heap_free(heap);
   }

   heap_log_error("Heap creation failed", status);
   
   sig->heap_create.status = (U32)status;
   sig->heap_create.href   = (OSHEAP)0;

   return sig;
}

/*****************************************************************************/

static inline void
l_heap_destroy(union SIGNAL *sig)
{
   struct OseHeap *heap = heapd_heap_find(sig->heap_destroy.href);

   if (heap == NULL)
   {
      free_buf(&sig);
      return;
   }

   /* Only if in system domain. Application domains
    * are cleaned up automatically by the system. */
   if (heap->domain == MM_KERNEL_DOMAIN)
      heap->hucb.remove_debug_hooks(&heap->hucb);
   
   if (heap->daemon != 0)
      kill_proc(heap->daemon);

   set_segment(0, heap->domain);
   zzheap_destroy(&heap->heap);
   set_segment(1, 0);

   l_heap_remove(heap);
   
   l_heap_free(heap);
   free_buf(&sig);
}


/*****************************************************************************/

static inline void
l_client_add(union SIGNAL *sig, PROCESS pid, struct OseHeap *heap)
{
   /* FIXME: A client is registering on a non-existing heap.
    * This should be signalled somehow. */
   if (heap == NULL)
   {
      free_buf(&sig);
      return;
   }

   /* Check if the client is still alive, if not, garbage collect its
    * buffer right now. */
   if (get_ptype(pid) == OS_ZOOMBIE)
   {
      l_client_rem(sig, pid, heap);
      return;
   }
   
   sig->client_add.sig_no = HEAPD_CLIENT_REM;
   attach(&sig, pid);
}

/* #define MEASURE_GARBAGE_COLLECT */

#ifdef MEASURE_GARBAGE_COLLECT
struct tval
{
   long long tv_sec;
   long long tv_usec;
};

extern void h_timediff(struct tval *diff, struct tval *start,
                       struct tval *end);
extern void h_timeadd(struct tval *sum, struct tval *t1,
                      struct tval *t2);
extern void h_gettime(struct tval *t);
extern void h_inittime(void);
#endif

/*****************************************************************************/

static inline void
l_client_rem(union SIGNAL *sig, PROCESS pid, struct OseHeap *heap)
{
   if (heap == NULL)
   {
      free_buf(&sig);
      return;
   }

   /* If the owner has died.... */
   if (heap->owner == pid)
   {
      /* .. Do nothing.
       *
       * An attach signal will soon arrive to heapd
       * and do all the cleanup that is required. */
   }
   else /* Run garbage collector. */
   {
      /* Check if the heap lock is acquired and if the lock owner has
       * terminated. This is not legal and will leave the heap
       * deadlocked. An error is generated to allow the system
       * to handle the situation. The owner value is missing its
       * most significant bit (due to OSE's mutex implementation) so we
       * have to test with and without it to make sure that the pid
       * is really dead. */
      PROCESS lock_owner = heap_lock_owner(&heap->heap);
      if (lock_owner != 0 &&
          get_pid_info(lock_owner, 9) == OS_ZOOMBIE &&
          get_pid_info(lock_owner|0x80000000, 9) == OS_ZOOMBIE)
      {
         /* Check if the zombie process still owns the lock. */
         if (heap_lock_owner(&heap->heap) == lock_owner)
         {
            error2(OSE_EFATAL_MASK|OSE_HEAP_EDEADLOCK, (OSADDRESS)lock_owner);
            return;
         }
      }
      {
#ifdef MEASURE_GARBAGE_COLLECT               
         struct tval  tm_start;
         struct tval  tm_end;
         struct tval  tm_diff;

         /* Uncommented if not used from a test case that already initiated
            h_inittime() */
         /* h_inittime(); */
         
         h_gettime(&tm_start);
#endif         
         zzheap_collect(&heap->heap, pid);
#ifdef MEASURE_GARBAGE_COLLECT                        
         h_gettime(&tm_end);
         h_timediff(&tm_diff, &tm_start, &tm_end);
            
         ramlog_printf("HEAP_G_COLLECT :%3llu(s) %09llu(ns)  stop:%3llu(s) "
                       "%09llu(ns) diff:%3llu(s) %10llu(ns) (~%10llums)\n", 
                       tm_start.tv_sec, tm_start.tv_usec,
                       tm_end.tv_sec, tm_end.tv_usec,
                       tm_diff.tv_sec,  tm_diff.tv_usec,
                       tm_diff.tv_usec/1000/1000);
#endif      
      }
   }

   free_buf(&sig);
}


/*****************************************************************************/

static inline union SIGNAL *
l_monitor_add(union SIGNAL *sig, PROCESS pid)
{
   union SIGNAL *att_sig;
   
   struct OseHeap *heap = heapd_heap_find(sig->mon_req.href);
   struct HeapMonitor *monitor;
   HeapStatus status;
   size_t smax;
   
   if (heap_attach_monitor_filter == NULL)
   {
      sig->sig_no = HEAP_MONITOR_REPLY;
      sig->mon_rep.status = HEAP_INVALID_REQUEST;
      return sig;
   }
   
   if (heap == NULL)
   {
      sig->sig_no = HEAP_MONITOR_REPLY;
      sig->mon_rep.status = HEAP_ENOT_FOUND;
      return sig;
   }

   smax = heap->config.max_type == 1 ?
      heap->config.frag_size_max : heap->config.buf_size_max;

   if (heap->monitor != NULL ||
       sig->mon_req.monitor_pid == 0 ||
       sig->mon_req.heap_low_level >= sig->mon_req.heap_not_low_level ||
       sig->mon_req.heap_not_low_level >= smax)
   {
      sig->sig_no = HEAP_MONITOR_REPLY;
      sig->mon_rep.status = HEAP_INVALID_REQUEST;
      return sig;
   }

   monitor = heap_init_monitor_filter(
      &heap->heap,
      &heap->stats, &heap->config,
      sig->mon_req.href,
      sig->mon_req.monitor_pid,
      sig->mon_req.heap_low_level,
      sig->mon_req.heap_not_low_level);
   if (monitor == NULL)
   {
      sig->sig_no = HEAP_MONITOR_REPLY;
      sig->mon_rep.status = HEAP_EOUT_OF_MEMORY;   /* FIXME. */
      return sig;
   }

   status = heap_attach_monitor_filter(&heap->heap, monitor);
   if (status != HEAP_SUCCESS)
   {
      sig->sig_no = HEAP_MONITOR_REPLY;
      sig->mon_rep.status = HEAP_EOUT_OF_MEMORY;   /* FIXME. */
      return sig;
   }

   att_sig = alloc(sizeof(struct HeapdMonitorRem), HEAPD_MONITOR_REM);
   att_sig->mon_rem.href = sig->mon_req.href;
   attach(&att_sig, pid);

   heap->monitor       = monitor;
   sig->sig_no         = HEAP_MONITOR_REPLY;
   sig->mon_rep.status = HEAP_SUCCESS;

   return sig;
}

/*****************************************************************************/

static inline void
l_monitor_rem(union SIGNAL *sig, PROCESS pid)
{
   struct OseHeap *heap = heapd_heap_find(sig->mon_rem.href);

   (void)pid;

   free_buf(&sig);

   if (heap == NULL || heap->monitor == NULL)
      return;

   heap_detach_monitor_filter(&heap->heap, heap->monitor);
   heap_destroy_monitor_filter(&heap->heap, heap->monitor);

   heap->monitor = NULL;
}

/*****************************************************************************/

static inline void
l_frag_reclaim(union SIGNAL *sig, PROCESS pid, struct OseHeap *heap)
{
   void *frag = NULL, *frag_out = NULL;
   size_t size = 0;
   char phys_mem[MM_NAME_LENGTH+1];
   HeapParam cursor[6];

   sig->frag_rem.base = 0;
   sig->frag_rem.size = 0;

   if (heap == NULL || l_heap_daemon(heap) == 0 || heap->daemon != current_process())
   {
      send(&sig, pid);
      return;
   }


   SET_SEGMENT(heap);
   
   memset(cursor, 0, sizeof(cursor));
   while (l_frag_next_free(&heap->heap, &frag, &size, cursor))
   {
      if (!l_frag_pmem(&heap->heap, frag, phys_mem))
         continue;

      if (strcmp(phys_mem, sig->frag_rem.phys_mem))
         continue;

      if (!l_frag_remove(&heap->heap, &frag, &size))
         continue;

      frag_out = frag;
      break;
   }

   RESTORE_SEGMENT(heap);

   if (frag_out != NULL)
   {
      sig->frag_rem.base = (OSADDRESS)frag_out;
      sig->frag_rem.size = (OSBUFSIZE)size;
   }

   send(&sig, pid);
}

/*****************************************************************************/

static inline void
l_frag_inquiry(union SIGNAL *sig, PROCESS pid, struct OseHeap *heap)
{
   void *frag = NULL;
   size_t size = 0;
   char phys_mem[MM_NAME_LENGTH+1];
   HeapParam cursor[6];

   sig->frag_inquiry.size = 0;

   if (heap == NULL || l_heap_daemon(heap) == 0 || heap->daemon != current_process())
   {
      send(&sig, pid);
      return;
   }

   SET_SEGMENT(heap);
   
   memset(cursor, 0, sizeof(cursor));
   while (l_frag_next_free(&heap->heap, &frag, &size, cursor))
   {
      if (!l_frag_pmem(&heap->heap, frag, phys_mem))
         continue;

      if (strcmp(phys_mem, sig->frag_inquiry.phys_mem))
         continue;

      sig->frag_inquiry.size += (OSBUFSIZE)size;
   }

   RESTORE_SEGMENT(heap);

   send(&sig, pid);
}

/*****************************************************************************/

/* Get the maximum signal size that can be returned. */
static OSADDRESS
l_get_max_sigsize(void)
{
   struct OS_pcb *pcb = get_pcb(current_process());
   OSADDRESS max_size = pcb->max_sigsize;
   free_buf((union SIGNAL**)&pcb);
   return max_size;
}   


/*****************************************************************************/

static inline union SIGNAL *
l_heap_list(union SIGNAL *sig)
{
   struct OseHeap *heap;
   uint32_t count = l_heap_count();
   uint32_t status = HEAP_SUCCESS;
   
   if (count > 200)
   {
      size_t max_count = 
         (l_get_max_sigsize() - offsetof(struct HeapListReply, heaps)) /
         sizeof(sig->list_rep.heaps[0]);
      if (count > max_count)
      {
         count = max_count;
         status = HEAP_ETOO_MANY_ELEMENTS;
      }
   }

   free_buf(&sig);
   sig = alloc(offsetof(struct HeapListReply, heaps[count]), HEAP_LIST_REPLY);
   
   sig->list_rep.status = status;
   sig->list_rep.num_heaps = count;

   heap = heapd_heap_first();
   while ((heap != NULL) && (count != 0))
   {
      count--;
      sig->list_rep.heaps[count].ownerpid = heap->owner;
      sig->list_rep.heaps[count].href = HREF_FROM_HEAP(heap);

      heap = heapd_heap_next(heap);
   }

   return (union SIGNAL*)sig;
}

/*****************************************************************************/

static inline void
l_heap_stats(union SIGNAL *sig, PROCESS pid, struct OseHeap *heap)
{
   uint32_t frag_count = 0;

   if (heap == NULL)
   {
      free_buf(&sig);
      sig = alloc(sizeof(struct HeapStatsReply), HEAP_STATS_REPLY);
      sig->stats_rep.status = HEAP_ENOT_FOUND;
      send_w_sender(&sig, heapd_pid(), pid);
      return;
   }

   if (l_heap_daemon(heap) != 0 && heap->daemon != current_process())
   {
      /* Forward request to local heap daemon. */
      send_w_sender(&sig, pid, heap->daemon);
      return;
   }
       
   free_buf(&sig);

   SET_SEGMENT(heap);
   
   /*
    * Query Usage Statistics
    ***************************************************/

   do {
      HeapParam p[15];
      uint32_t status = HEAP_SUCCESS;

      memset(p, 0, sizeof(p));
      
      p[0] = HPM(QUERY_BUF_TOT_USED_SIZE, 1);
      p[2] = HPM(QUERY_BUF_TOT_FREE_SIZE, 1);
      p[4] = HPM(QUERY_BUF_TOT_REQ_SIZE,  1);
      p[6] = HPM(QUERY_BUF_LARGEST_FREE,  1);
      p[8] = HPM(QUERY_FRAG_COUNT, 1);
      p[10] = HPM(QUERY_BUF_NUM_SHARED,  1);
      p[12] = HPM(QUERY_BUF_NUM_PRIVATE, 1);
      p[14] = HPM(END, 0);

      zzheap_query(&heap->heap, p, 0);
      
      if (HPM_IS_HANDLED(p[8]))
         frag_count = p[9];

      if (frag_count > 200)
      {
         size_t max_count = 
            (l_get_max_sigsize() - offsetof(struct HeapStatsReply, frags)) /
            sizeof(sig->stats_rep.frags[0]);
         if (frag_count > max_count)
         {
            frag_count = max_count;
            status = HEAP_ETOO_MANY_ELEMENTS;
         }
      }

      sig = alloc(offsetof(struct HeapStatsReply, frags[frag_count]),
                  HEAP_STATS_REPLY);

      sig->stats_rep.status = status;
      sig->stats_rep.num_frags = frag_count;

      if (HPM_IS_HANDLED(p[0]))
         sig->stats_rep.total_used = p[1];
      else
         sig->stats_rep.total_used = 0;
      if (HPM_IS_HANDLED(p[2]))
         sig->stats_rep.total_free = p[3];
      else
         sig->stats_rep.total_free = 0;
      if (HPM_IS_HANDLED(p[4]))
         sig->stats_rep.total_req = p[5];
      else
         sig->stats_rep.total_req = 0;
      if (HPM_IS_HANDLED(p[6]))
         sig->stats_rep.largest_free = p[7];
      else
         sig->stats_rep.largest_free = 0;
      if (HPM_IS_HANDLED(p[10]))
         sig->stats_rep.number_shared = p[11];
      else
         sig->stats_rep.number_shared = 0;
      if (HPM_IS_HANDLED(p[12]))
         sig->stats_rep.number_private = p[13];
      else
         sig->stats_rep.number_private = 0;
   } while (0);


   /*
    * Fragment List
    ***************************************************/

   do {
      HeapParam p[6];
      uint32_t ix;
      
      memset(p, 0, sizeof(p));
         
      for (ix=0; ix<frag_count; ix++)
      {
         l_frag_next(&heap->heap,
            (void  **)&sig->stats_rep.frags[ix].addr,
            (size_t *)&sig->stats_rep.frags[ix].size, p);
      }
      
   } while (0);

   RESTORE_SEGMENT(heap);
      
   send_w_sender(&sig, heapd_pid(), pid);
}

/*****************************************************************************/

static inline void
l_heap_stats_ex(union SIGNAL *sig, PROCESS pid, struct OseHeap *heap)
{
   if (heap == NULL)
   {
      free_buf(&sig);
      sig = alloc(sizeof(struct HeapStatsExReply), HEAP_STATSEX_REPLY);
      sig->stex_rep.status = HEAP_ENOT_FOUND;
      send_w_sender(&sig, heapd_pid(), pid);
      return;
   }

   if (l_heap_daemon(heap) != 0 && heap->daemon != current_process())
   {
      /* Forward request to local heap daemon. */
      send_w_sender(&sig, pid, heap->daemon);
      return;
   }

   free_buf(&sig);


   sig = alloc(sizeof(struct HeapStatsExReply), HEAP_STATSEX_REPLY);
   sig->stex_rep.status = HEAP_SUCCESS;

   SET_SEGMENT(heap);
   
   do {
      HeapParam p[17];
      memset(p, 0, sizeof(p));

      p[0]  = HPM(QUERY_BUF_TOT_USED_SIZE, 1);
      p[2]  = HPM(QUERY_BUF_TOT_FREE_SIZE, 1);
      p[4]  = HPM(QUERY_BUF_TOT_REQ_SIZE,  1);
      p[6]  = HPM(QUERY_BUF_MAX_USED_SIZE, 1);
      p[8]  = HPM(QUERY_BUF_LARGEST_FREE,  1);
      p[10] = HPM(QUERY_BUF_NUM_SHARED,    1);
      p[12] = HPM(QUERY_BUF_NUM_PRIVATE,   1); 
      p[14] = HPM(QUERY_ALLOC_NULL,        1);
      p[16] = HPM(END, 0);

      zzheap_query(&heap->heap, p, 0);

      if (HPM_IS_HANDLED(p[0]))
         sig->stex_rep.total_used = p[1];
      if (HPM_IS_HANDLED(p[2]))
         sig->stex_rep.total_free = p[3];
      if (HPM_IS_HANDLED(p[4]))
         sig->stex_rep.total_req = p[5];
      if (HPM_IS_HANDLED(p[6]))
         sig->stex_rep.total_used_hi = p[7];
      if (HPM_IS_HANDLED(p[8]))
         sig->stex_rep.largest_free = p[9];
      if (HPM_IS_HANDLED(p[10]))
         sig->stex_rep.number_shared = p[11];
      if (HPM_IS_HANDLED(p[12]))
         sig->stex_rep.number_private = p[13];
      if (HPM_IS_HANDLED(p[14]))
         sig->stex_rep.malloc_failed = p[15];
   } while (0);

   do {
      HeapParam p[10];
      uint32_t ix;

      sig->stex_rep.num_sizes = 0;

      memset(&sig->stex_rep.size,     0, sizeof(sig->stex_rep.size));
      memset(&sig->stex_rep.num_used, 0, sizeof(sig->stex_rep.num_used));
      memset(&sig->stex_rep.num_free, 0, sizeof(sig->stex_rep.num_free));
      
      for (ix = 0;
           ix < sizeof(sig->stex_rep.size) / sizeof(sig->stex_rep.size[0]);
           ix++)
      {
         memset(p, 0, sizeof(p));

         p[0] = HPM(QUERY_BUF_NUM_FREE, 2);
         p[3] = HPM(QUERY_BUF_NUM_USED, 2);
         p[6] = HPM(QUERY_BUF_NUM_SIZE, 2);
         p[9] = HPM(END,0);
         p[1] = p[4] = p[7] = (HeapParam)ix;

         zzheap_query(&heap->heap, p, 0);

         if (HPM_IS_HANDLED(p[0]))
            sig->stex_rep.num_free[ix] = (U32)p[2];
         if (HPM_IS_HANDLED(p[3]))
            sig->stex_rep.num_used[ix] = (U32)p[5];
         if (HPM_IS_HANDLED(p[6]))
            sig->stex_rep.size[ix] = (U32)p[8];

         if (sig->stex_rep.size[ix] != 0)
            sig->stex_rep.num_sizes++;
         else
            break;
      }
      
   } while (0);

   RESTORE_SEGMENT(heap);

   send_w_sender(&sig, heapd_pid(), pid);
}

/*****************************************************************************/

static inline void
l_heap_config(union SIGNAL *sig, PROCESS pid, struct OseHeap *heap)
{
   if (heap == NULL)
   {
      free_buf(&sig);
      sig = alloc(sizeof(struct HeapConfigReply), HEAP_CONFIG_REPLY);
      sig->config_rep.status = HEAP_ENOT_FOUND;
      send_w_sender(&sig, heapd_pid(), pid);
      return;
   }

   if (l_heap_daemon(heap) != 0 && heap->daemon != current_process())
   {
      send_w_sender(&sig, pid, heap->daemon);
      return;
   }
       
   free_buf(&sig);
   sig = alloc(sizeof(struct HeapConfigReply), HEAP_CONFIG_REPLY);

   /* These should maybe be queried from the heap
    * instead, but this way is faster and more simple. */
   sig->config_rep.ownerpid        = heap->owner;
   sig->config_rep.small_init_size = ALIGN(heap->config.frag_size_init);
   sig->config_rep.small_ext_size  = ALIGN(heap->config.frag_size_default);
   sig->config_rep.large_threshold = heap->config.buf_size_threshold;
   if (heap->config.max_type == 1) /* HEAP_MAX_TYPE_FRAGMENTS */
      sig->config_rep.max_size     = heap->config.frag_size_max;
   else                            /* HEAP_MAX_TYPE_BUFFERS */
      sig->config_rep.max_size     = heap->config.buf_size_max;
   sig->config_rep.small_low_limit = heap->frag_slow_limit;
   sig->config_rep.large_low_limit = heap->frag_llow_limit;
   sig->config_rep.domain          = heap->domain;
   sig->config_rep.permissions     = heap->frag_permissions;
   sig->config_rep.cache_mode      = heap->frag_cache_mode;
   sig->config_rep.endmark_size    = heap->config.buf_size_endmark;
   sig->config_rep.annotation_size = heap->hucb.annotation_size;
   sig->config_rep.wipe_on_free    = 0;
   sig->config_rep.error_on_null   = heap->buf_error_on_null;
   strlcpy(sig->config_rep.log_mem, heap->frag_log_mem, MM_NAME_LENGTH+1);

   sig->config_rep.status = HEAP_SUCCESS;
   send_w_sender(&sig, heapd_pid(), pid);
}

/*****************************************************************************/

static inline void
l_heap_buffer(union SIGNAL *sig, PROCESS pid, struct OseHeap *heap)
{
   union SIGNAL *reply;
   uint32_t burst;
   size_t data_size, max_size, file_size;
   HeapParam p[6];
   
   
   if (heap == NULL)
   {
      free_buf(&sig);
      reply = alloc(sizeof(struct HeapBufferInfoReply), HEAP_BUFFER_INFO_REPLY);
      reply->buf_rep.status = HEAP_ENOT_FOUND;
      send_w_sender(&reply, heapd_pid(), pid);
      return;
   }

   /* Forward request if there is a local heap daemon. */
   if (l_heap_daemon(heap) != 0 && heap->daemon != current_process())
   {
      send_w_sender(&sig, pid, heap->daemon);
      return;
   }

   /* Get the maximum signal size that can be returned. */
   max_size = l_get_max_sigsize();

   SET_SEGMENT(heap);
   
   /* Prepare buffer iteration request parameters. */
   p[1] = 0;                       /* Buffer       (out)     */
   p[2] = sig->buf_req.frag;       /* Fragment     (in/out)  */
   p[3] = sig->buf_req.path;       /* Misc. data   (in/out). */
   p[4] = sig->buf_req.depth;      /* Misc. data   (in/out)  */

   for (burst = 0; burst < sig->buf_req.burst; burst++)
   {
      union SIGNAL *tmp_sig = alloc(
         sizeof(struct HeapBufferInfoReply) + HEAP_MAX_FILENAME_LENGTH,
         HEAP_BUFFER_INFO_REPLY);

     next_buffer:
      /* Keep data from previous queries. */
      p[0] = HPM(QUERY_BUF_NEXT, 4);
      p[5] = HPM(END, 0);

      /* Query next buffer in iteration. */
      zzheap_query(&heap->heap, p, 0);

      if (!HPM_IS_HANDLED(p[0]))
      {
         tmp_sig->buf_rep.status = HEAP_NO_MORE_ITEMS;
         send_w_sender(&tmp_sig, heapd_pid(), pid);
         break;
      }

      tmp_sig->buf_rep.address = (U32)p[1];
      tmp_sig->buf_rep.frag    = (U32)p[2];
      tmp_sig->buf_rep.path    = (U32)p[3];
      tmp_sig->buf_rep.depth   = (U32)p[4];

      do {
         HeapParam q[9];

         q[0] = HPM(QUERY_BUF_OWNER,1);
         q[2] = HPM(QUERY_BUF_REQ_SIZE,1);
         q[4] = HPM(QUERY_BUF_ACT_SIZE,1);
         q[6] = HPM(QUERY_BUF_STATUS,1);
         q[1] = q[3] = q[5] = q[7] = p[1];
         q[8] = HPM_END;

         zzheap_query(&heap->heap, q, 0);

         /* Only include allocated buffers in reply. */
         if (HPM_IS_HANDLED(q[6]) && !(q[7] & HEAP_BUFFER_ALLOCATED))
            goto next_buffer;

         if (HPM_IS_HANDLED(q[0]))
            tmp_sig->buf_rep.owner = q[1];
         else
            tmp_sig->buf_rep.owner = 0;

         if (sig->buf_req.owner != 0 &&
             sig->buf_req.owner != tmp_sig->buf_rep.owner)
            goto next_buffer;

         if (HPM_IS_HANDLED(q[2]))
            tmp_sig->buf_rep.req_size = q[3];
         else
            tmp_sig->buf_rep.req_size = 0;

         if (HPM_IS_HANDLED(q[4]))
            tmp_sig->buf_rep.act_size = q[5];
         else
            tmp_sig->buf_rep.act_size = 0;

         if (HPM_IS_HANDLED(q[6]))
         {
            tmp_sig->buf_rep.endmark_ok = !(q[7] & HEAP_BUFFER_ENDMARK);
            tmp_sig->buf_rep.ann_size   = (q[7] & HEAP_BUFFER_ANNOTATED) ?
               heap->hucb.annotation_size : 0;
         }
         else
         {
            tmp_sig->buf_rep.ann_size   = 0;
            tmp_sig->buf_rep.endmark_ok = 1;
         }
      } while (0);


      if (tmp_sig->buf_rep.ann_size && heap->hucb.heap_debug_fileline)
      {
         tmp_sig->buf_rep.address = tmp_sig->buf_rep.address + tmp_sig->buf_rep.ann_size;           
         heap->hucb.heap_debug_fileline((OSHEAP)heap, &tmp_sig->buf_rep, 0);
         file_size = strlen((char *)tmp_sig->buf_rep.data) + 1;
      }
      else
      {
         file_size                 = 0;
         tmp_sig->buf_rep.file     = ~0UL;
         tmp_sig->buf_rep.line     = 0;
         tmp_sig->buf_rep.ann_size = 0;
      }

      /* Calculate maximum data size that can be handled. */
      data_size = max_size - sizeof(struct HeapBufferInfoReply) - file_size;

      /* Check if requested data size is too large to handle. */
      data_size = sig->buf_req.max_data > data_size ? data_size : sig->buf_req.max_data;

      /* Allocate reply signal that can handle the requested data payload. */
      reply = alloc(sizeof(struct HeapBufferInfoReply) + file_size + data_size,
                    HEAP_BUFFER_INFO_REPLY);

      /* Copy signal data from temporary buffer. */
      memcpy(&reply->buf_rep, &tmp_sig->buf_rep, sizeof(struct HeapBufferInfoReply));

      /* Copy file name if present. */
      if (file_size > 0)
      {
         reply->buf_rep.file = data_size;
         memcpy(&reply->buf_rep.data[data_size], &tmp_sig->buf_rep.data[0], file_size);
      }

      free_buf(&tmp_sig);

      /* Obtain data sample if requested. */
      if (data_size > 0)
      {
         HeapParam q[] = {
            HPM(QUERY_BUF_DATA, 3), 0,0,0,
            HPM_END
         };

         q[1] = (HeapParam)p[1];
         q[2] = (HeapParam)reply->buf_rep.data;
         q[3] = (HeapParam)data_size;

         zzheap_query(&heap->heap, q, 0);

         if (HPM_IS_HANDLED(q[0]))
            reply->buf_rep.data_size = data_size;
         else
            reply->buf_rep.data_size = 0;
      }
      else
      {
         reply->buf_rep.data_size = 0;
      }


      reply->buf_rep.status = HEAP_SUCCESS;
      send_w_sender(&reply, heapd_pid(), pid);
   }

   RESTORE_SEGMENT(heap);
   
   free_buf(&sig);
}

/*****************************************************************************/

static inline void
l_heap_frags(union SIGNAL *sig, PROCESS pid, struct OseHeap *heap)
{
   uint32_t frag_count = 0;

   if (heap == NULL)
   {
      free_buf(&sig);
      sig = alloc(sizeof(struct HeapFragsReply), HEAP_FRAGS_REPLY);
      sig->frags_rep.status = HEAP_ENOT_FOUND;
      send_w_sender(&sig, heapd_pid(), pid);
      return;
   }

   if (l_heap_daemon(heap) != 0 && heap->daemon != current_process())
   {
      /* Forward to local heap daemon. */
      send_w_sender(&sig, pid, heap->daemon);
      return;
   }
   
   free_buf(&sig);

   SET_SEGMENT(heap);
   
   do {
      HeapParam p[3];
      uint32_t status = HEAP_SUCCESS;

      memset(p, 0, sizeof(p));

      p[0] = HPM(QUERY_FRAG_COUNT, 1);
      p[2] = HPM(END, 0);

      zzheap_query(&heap->heap, p, 0);
      
      if (HPM_IS_HANDLED(p[0]))
         frag_count = (uint32_t)p[1];

      if (frag_count > 200)
      {
         size_t max_count = 
            (l_get_max_sigsize() - offsetof(struct HeapFragsReply, frags)) /
            sizeof(sig->frags_rep.frags[0]);
         if (frag_count > max_count)
         {
            frag_count = max_count;
            status = HEAP_ETOO_MANY_ELEMENTS;
         }
      }

      sig = alloc(offsetof(struct HeapFragsReply, frags[frag_count]),
                  HEAP_FRAGS_REPLY);

      sig->frags_rep.status = status;
      sig->frags_rep.num_frags = frag_count;

   } while (0);


   do {
      HeapParam p[6];
      uint32_t ix = 0;
      
      memset(p, 0, sizeof(p));
         
      for (ix=0; ix<frag_count; ix++)
      {
         l_frag_next(&heap->heap,
            (void  **)&sig->frags_rep.frags[ix].addr,
            (size_t *)&sig->frags_rep.frags[ix].size, p);
      }
   } while (0);

   RESTORE_SEGMENT(heap);
   
   send_w_sender(&sig, heapd_pid(), pid);
}

/*****************************************************************************/

static inline void
l_heap_frag_info(union SIGNAL *sig, PROCESS pid, struct OseHeap *heap)
{
   struct HeapTag *tags = sig->frag_info.tags;
   HeapFragmentStatus frag_stat = 0;
   HeapParam cursor[6];
   void *buf;
   uint32_t ix;
   int res;

   if (heap == NULL)
   {
      free_buf(&sig);
      sig = alloc(sizeof(struct HeapFragsReply), HEAP_FRAG_INFO_REPLY);
      sig->frag_info.status = HEAP_ENOT_FOUND;
      send_w_sender(&sig, heapd_pid(), pid);
      return;
   }

   if (l_heap_daemon(heap) != 0 && heap->daemon != current_process())
   {
      send_w_sender(&sig, pid, heap->daemon);
      return;
   }
   
   SET_SEGMENT(heap);

   /* Check if fragment status. */
   res = l_frag_status(&heap->heap, (void *)sig->frag_info.base, &frag_stat);
   if (!res)
   {
      /* report the error back */
      sig->frag_info.signo     = HEAP_FRAG_INFO_REPLY;
      sig->frag_info.status    = HEAP_EFRAG_NOT_FOUND;
      sig->frag_info.err_index = 0;

      RESTORE_SEGMENT(heap);
      
      /* return faster */
      send_w_sender(&sig, heapd_pid(), pid);
      return; 
   }

   /* Clear tag values. */
   for (ix = 0; ix < sig->frag_info.num_tags; ix++)
   {
      tags[ix].tag_value = 0;
   }

   /* If the fragment is free, fill in size immediately and reply. */
   if (!(frag_stat & HEAP_FRAGMENT_ALLOCATED))
   {
      for (ix = 0; ix < sig->frag_info.num_tags; ix++)
      {
         if (tags[ix].tag_id == HEAP_FRAG_INFO_TAG_FREE_SIZE)
         {
            tags[ix].tag_value = l_frag_size(&heap->heap, sig->frag_info.base);
         }
         else
         {
            tags[ix].tag_value = 0;
         }
      }

      sig->frag_info.signo     = HEAP_FRAG_INFO_REPLY;
      sig->frag_info.status    = HEAP_SUCCESS;
      sig->frag_info.err_index = 0;

      RESTORE_SEGMENT(heap);

      send_w_sender(&sig, heapd_pid(), pid);
      return;
   }
   
   memset(cursor, 0, sizeof(cursor));

   while (l_frag_buf_next(&heap->heap, sig->frag_info.base, &buf, cursor))
   {
      size_t act = 0, req = 0;
      int avail = 0;

      l_buf_size(&heap->heap, buf, &req, &act, &avail);

      for (ix = 0; ix < sig->frag_info.num_tags; ix++)
      {
         switch (tags[ix].tag_id)
         {
            case HEAP_FRAG_INFO_TAG_REQ_SIZE:
            {
               tags[ix].tag_value += req;
               break;
            }
            case HEAP_FRAG_INFO_TAG_USED_SIZE:
            {
               if (!avail)
                  tags[ix].tag_value += act;
               break;
            }
            case HEAP_FRAG_INFO_TAG_FREE_SIZE:
            {
               if (avail)
                  tags[ix].tag_value += act;
               break;
            }
            default:
            {
               sig->frag_info.signo     = HEAP_FRAG_INFO_REPLY;
               sig->frag_info.status    = HEAP_INVALID_REQUEST;
               sig->frag_info.err_index = ix;

               RESTORE_SEGMENT(heap);
               
               send_w_sender(&sig, heapd_pid(), pid);
               return;
            }
         }
      }
   }

   sig->frag_info.signo     = HEAP_FRAG_INFO_REPLY;
   sig->frag_info.status    = HEAP_SUCCESS;
   sig->frag_info.err_index = 0;

   RESTORE_SEGMENT(heap);
   
   send_w_sender(&sig, heapd_pid(), pid);
}

/*****************************************************************************/

static inline void
l_heap_stats_reset(union SIGNAL *sig, PROCESS pid, struct OseHeap *heap)
{
   if (heap == NULL)
   {
      sig->reset_rep.sigNo  = HEAP_RESET_STATS_REPLY;
      sig->reset_rep.status = HEAP_ENOT_FOUND;
      send_w_sender(&sig, heapd_pid(), pid);
      return;
   }

   if (l_heap_daemon(heap) != 0 && heap->daemon != current_process())
   {
      send_w_sender(&sig, pid, heap->daemon);
      return;
   }

   SET_SEGMENT(heap);

   sig->reset_rep.sigNo  = HEAP_RESET_STATS_REPLY;
   sig->reset_rep.status = HEAP_SUCCESS;

   if (sig->reset_req.flags & HEAP_RESET_WATERMARKS)
   {
      HeapParam p[] = {
         HPM(CTRL_BUF_MAX_USED_SIZE, 0),
         HPM_END
      };

      zzheap_control(&heap->heap, p);

      if (!HPM_IS_HANDLED(p[0]))
         sig->reset_rep.status = HEAP_INVALID_REQUEST;
   }

   RESTORE_SEGMENT(heap);
   
   send_w_sender(&sig, heapd_pid(), pid);
}

/*****************************************************************************/



/*****************************************************************************
 * System Heap Daemon
 *****************************************************************************/

OSENTRYPOINT ose_heapd;
OS_PROCESS(ose_heapd)
{
   struct mm_reclaim_physical_event_t mm_event;
   mm_status_t status;

   /*
    * The call is used to ensure that ose_heap.o gets linked in, to resolve
    * weak links in monolith.
    */
   ose_reset_heap(get_bid(current_process()));


   /* must call this function for safety, because if
    * the ose_crt_init call is made after any heap user
    * in the monolith!
    */
   ose_heap_init();

   heap_list = NULL;


   /* Install GMM reclaim event callback. */
   mm_event.base = 0;
   mm_event.size = (log_addr_t)-1;
   mm_event.domain = MM_EVENT_ANY_DOMAIN;
   mm_event.identifier = COMMON_HEAP_COMPONENT;
   mm_event.event_handler = l_heap_reclaim;
   mm_event.event_no = MM_EVENT_RECLAIM_REGION_PHYSICAL;
   mm_event.phys_mem = "RAM";
   status = ose_mm_install_handler((union mm_event_t *)&mm_event, 1);
   if (status != MM_SUCCESS)
   {
      ramlog_printf("Heapd: ose_mm_install_handler failed. Status: %d\n",
                    status);
      error2(OSE_HEAP_EINIT_HEAP_FAILED, (OSERRCODE)status);
   }

#ifdef OSARM
   zzset_manager_mode(TRUE);
#endif
   
   
   for (;;)
   {
      static const SIGSELECT any_sig[] = {0};
      union SIGNAL *sig;
      PROCESS pid;

      sig = receive(any_sig);
      pid = sender(&sig);

      switch (sig->sig_no)
      {
         /* Private signal interface. */

	 case HEAPD_HEAP_CREATE:
            sig = l_heap_create(sig);
            send_w_sender(&sig, heapd_pid(), pid);
	    break;
            
         case HEAPD_HEAP_DESTROY:
            l_heap_destroy(sig);
            break;

         case HEAPD_CLIENT_ADD:
            l_client_add(sig, pid, heapd_heap_find(sig->client_add.href));
            break;

         case HEAPD_CLIENT_REM:
            l_client_rem(sig, pid, heapd_heap_find(sig->client_rem.href));
            break;

         case HEAPD_MONITOR_REM:
            l_monitor_rem(sig, pid);
            break;

         case HEAPD_FRAG_REM:
            l_frag_reclaim(sig, pid, heapd_heap_find(sig->frag_rem.href));
            break;

         case HEAPD_FRAG_INQUIRY:
            l_frag_inquiry(sig, pid, heapd_heap_find(sig->frag_rem.href));
            break;


         /* Public signal interface. */

         case HEAP_INTERFACE_REQUEST:
            l_heapd_interface(sig, pid);
	    break;

         case HEAP_LIST_REQUEST:
            sig = l_heap_list(sig);
            send_w_sender(&sig, heapd_pid(), pid);
	    break;

	 case HEAP_MONITOR_REQUEST:
            sig = l_monitor_add(sig, pid);
            send_w_sender(&sig, heapd_pid(), pid);
	    break;

         case HEAP_STATS_REQUEST:
            l_heap_stats(sig, pid, heapd_heap_find(sig->stats_req.href));
	    break;

         case HEAP_STATSEX_REQUEST:
            l_heap_stats_ex(sig, pid, heapd_heap_find(sig->stats_req.href));
            break;

	 case HEAP_CONFIG_REQUEST:
            l_heap_config(sig, pid, heapd_heap_find(sig->config_req.href));
	    break;
            
         case HEAP_BUFFER_INFO_REQUEST:
            l_heap_buffer(sig, pid, heapd_heap_find(sig->buf_req.href));
	    break;

	 case HEAP_FRAGS_REQUEST:
            l_heap_frags(sig, pid, heapd_heap_find(sig->frags_req.href));
            break;

	 case HEAP_RESET_STATS_REQUEST:
            l_heap_stats_reset(sig, pid, heapd_heap_find(sig->reset_req.href));
            break;
            
         case HEAP_FRAG_INFO_REQUEST:
            l_heap_frag_info(sig, pid, heapd_heap_find(sig->frag_info.href));
            break;

	 default:
	 {
	    static struct HeapBufferInfoReply hbi = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,{0}};
	    hbi.address = (U32)sig;
	    if(sig->sig_no >= 33300 && sig->sig_no < 33400)
	    {
	       ramlog_printf("A heap.sig signal with version < OSE5.2 was received.\n");
	       ramlog_printf("The heap.sig interface is not backward compatible, please recompile clients\n");
	    }
	    error2(OSE_HEAP_EUNKNOWN_SIGNAL, (OSERRCODE)&hbi);
	    free_buf(&sig);
	    break;
	 }
      }
   }
}


/*****************************************************************************
 * Application Heap Daemon
 *****************************************************************************/

OSENTRYPOINT ose_heapcd_trampoline;
OS_PROCESS(ose_heapcd_trampoline)
{
   unsigned long handle = biosOpen(HEAP_BIOS_NAME);
   if (handle == 0)
   {
      error(OSE_EFATAL_MASK | OSE_HEAP_EBIOS_OPEN_FAILED);
      kill_proc(current_process());
   }

   /* Trampoline into supervisor mode and ose_heapcd(). */
   biosCall(handle, FC_CLEANUPD);

   /* Shall never return. */
   error2(OSE_HEAP_EOTHER, __LINE__);
   kill_proc(current_process());
}

/*****************************************************************************/

/*
 * Returns "heap" if "heap" equals "href" with shared bit masked,
 * NULL otherwise.
 */
static struct OseHeap *
l_verify_href(OSHEAP href, struct OseHeap *heap)
{
   if (HEAP_FROM_HREF(href) == heap)
   {
      return heap;
   }

   ramlog_printf("Illegal heap reference 0x%x in signal to process: 0x%x\n",
                 href, current_process());

   return NULL;
}

OSENTRYPOINT ose_heapcd;
OS_PROCESS(ose_heapcd)
{
  static const SIGSELECT any_sig[] = {0};
  static const SIGSELECT init_sig[] = {1, HEAPD_INIT_HEAPCD};
  union SIGNAL *sig;
  struct OseHeap *my_heap;
  PROCESS pid;

#ifdef OSARM
  zzset_manager_mode(TRUE);
#endif

  sig = receive(init_sig);

  my_heap = sig->init_heapcd.heap;
  free_buf(&sig);

  for (;;)
  {
    sig = receive(any_sig);
    pid = sender(&sig);
    
    switch (sig->sig_no)
    {
       case HEAPD_CLIENT_ADD:
          l_client_add(sig, pid, l_verify_href(sig->client_add.href, my_heap));
          break;
       
       case HEAPD_CLIENT_REM:
          l_client_rem(sig, pid, l_verify_href(sig->client_rem.href, my_heap));
          break;

       case HEAPD_FRAG_REM:
          l_frag_reclaim(sig, pid, l_verify_href(sig->frag_rem.href, my_heap));
          break;

       case HEAPD_FRAG_INQUIRY:
          l_frag_inquiry(sig, pid, l_verify_href(sig->frag_inquiry.href,
                                                 my_heap));
          break;
          
       case HEAP_STATS_REQUEST:
          l_heap_stats(sig, pid, l_verify_href(sig->stats_req.href, my_heap));
          break;

       case HEAP_STATSEX_REQUEST:
          l_heap_stats_ex(sig, pid, l_verify_href(sig->stex_req.href,
                                                  my_heap));
          break;

       case HEAP_CONFIG_REQUEST:
          l_heap_config(sig, pid, l_verify_href(sig->config_req.href,
                                                my_heap));
          break;
            
       case HEAP_BUFFER_INFO_REQUEST:
          l_heap_buffer(sig, pid, l_verify_href(sig->buf_req.href, my_heap));
          break;

       case HEAP_FRAGS_REQUEST:
          l_heap_frags(sig, pid, l_verify_href(sig->frags_req.href, my_heap));
          break;

       case HEAP_RESET_STATS_REQUEST:
          l_heap_stats_reset(sig, pid, l_verify_href(sig->reset_req.href,
                                                     my_heap));
          break;

       case HEAP_FRAG_INFO_REQUEST:
          l_heap_frag_info(sig, pid, l_verify_href(sig->frag_info.href,
                                                   my_heap));
          break;

       default:
       {
	  struct HeapBufferInfoReply hbi;
	  memset(&hbi,0, sizeof (struct HeapBufferInfoReply));
	  hbi.address = (U32)sig;
	  error2(OSE_HEAP_EUNKNOWN_SIGNAL, (OSERRCODE) &hbi);
	  break;
       }
    }
  }
}

/*****************************************************************************/
