/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2006 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

#include "ose.h"
#include "osetypes.h"
#include "stdio.h"
#include "unistd.h"
#include "ose_spi/mm.sig"
#include "ose_spi/mmtypes.h"
#include "ose_err/mm_err.h"
#include "ose_spi/mm.h"
#include "ramlog.h"
#include "assert.h"
#include "sys/mman.h"
#include "sys/types.h"
#include "sys/stat.h"
#include "errno.h"
#include "ksss_mm.h"
#include "bios.h"
#include "mman_internal.h"
#include "mman_trap.h"
#include "mm_util.h"
#include "slab.h"
#ifndef STANDALONE
#include "mm_trap_internal.h"
#include "hal.h"
#include "partition.h"
#include "pose.h"
#else
#include "standalone.h"
#endif


/************************** DEFINES **********************************/

#define ALIGN_DOWN(a,alignment)   ((a) - ((a) & (alignment - 1)))
#define ALIGN_UP(a,alignment) ALIGN_DOWN(((a) + (alignment) - 1),alignment)


/************************** STRUCTURES **********************************/

union SIGNAL
{
    SIGSELECT                     sig_no;
    struct mm_create_region       reg;
    struct mm_delete_region       munmap;
    struct mm_change_permissions  perm;
};


/************************** STATIC VARIABLES **************************/



/* The structure storing all the regions in the system. */
static struct vmas_t vmas;

/* For breakpoints in an mmaped region. Required by EPM.
 * If set, prepare_for_page_fault_in_disable_region shall point to a function
 * that checks if a page fault can be allowed although it occurred in a
 * disabled region. If not, 0 shall be returned. Otherwise, it shall make any
 * necessary preparations and then return a nonzero value. If a restore is
 * needed, a pointer to a restore functions can be returned through the
 * argument. */
uint32_t (*prepare_for_page_fault_in_disable_region)(void (**restore)(void))
         = NULL;

static PROCESS mm_pid = 0;

struct pager_info_t pager;


/************************** PROTOTYPES **************************/

#define MIN(a, b)  ((a) < (b) ? (a) : (b))

mman_status_t
add_vma(OSDOMAIN domain, log_addr_t base, size_t size,
        uint32_t permissions, uint32_t cache_mode, struct MmanPageMap *page_map,
        uint64_t file_offset);

static void
delete_vma(struct _vm_area_t *vma);

static int
compare_vmas(const void *a, const void *b);

struct _vm_area_t*
find_vma_internal(OSDOMAIN domain, log_addr_t address);

void
init_mman(void);

void
delete_region_callin(union mm_event_t *event);

int
ose_mman_mlock_temporarily(log_addr_t addr, size_t len);

int
ose_mman_munlock_temporarily(log_addr_t addr, size_t len);

int
ose_mman_mlock_counting(log_addr_t addr, size_t len);

int
ose_mman_munlock_counting(log_addr_t addr, size_t len);


/*
 *===========================================================================
 *                        mmap_mm
 *===========================================================================
 *
 * Description
 * -----------
 * The worker function for mmap. Interacts with MM to create a pageable
 * region.
 *
 * Parameters
 * ----------
 * region_name The name of the region.
 * addr        The start address of the region to be mapped.
 * len         The length of the region to be mapped
 * permissions The permissions of the mmaped region.
 *             The permission syntax is defined in mm.sig.
 * cache_mode  Specifies the cache attributes.
 *             The cache syntax defined in mm.sig.
 * fd          The file descriptor.
 * offset      The offset. Should be a multiple of the page size.
 *
 * Returns
 * -------
 * struct mm_create_region *      Returns the created region.
 *
 */
static struct mm_create_region *
mmap_mm (const char * region_name,
         void *__addr,
         size_t __len,
         uint32_t permissions,
         uint32_t cache_mode,
         int __fd,
         off_t __offset)
{
   static const SIGSELECT reply[] = {1,
                                     MM_CREATE_REGION_REPLY};
   union SIGNAL *sig;
   size_t namelen = strlen(region_name);

   (void) __offset;
   
   if (!mm_pid)
   {
      /* A (virtually) non-blocking hunt. */
      hunt(GMM_NAME_STRING, 0, &mm_pid, NULL);
      if (!mm_pid)
      {
         ramlog_printf("Error: mman could not find the MM.\n");
         error2(OSE_MMAN_INTERNAL_ERROR, 0);
      }
   }

   sig = alloc(sizeof(struct mm_create_region), MM_CREATE_REGION_REQUEST);
   sig->reg.size = __len;
   sig->reg.domain = 0;
   if(namelen < MM_NAME_LENGTH)
   {
      strcpy(sig->reg.name,region_name);
   }
   else
   {
      strlcpy(sig->reg.name,region_name, MM_NAME_LENGTH);
   }
   strcpy(sig->reg.phys_mem, MM_INTERNAL_RAM);
   strcpy(sig->reg.log_mem, "RAM");
   sig->reg.permissions = permissions;
   if (!pager.installed && __fd != -1)
   {
      /* As we need to write the data into this buffer when the pager
       * is not installed. */
      sig->reg.permissions |= MM_SU_WRITE;
   }
   sig->reg.cache_mode = cache_mode;
   sig->reg.options = MM_CREATE_REGION_SPECIFY_NAME;
   if(__addr)
   {
      sig->reg.base = (OSADDRESS) __addr;
      sig->reg.options |= MM_CREATE_REGION_SPECIFY_LOG;
   }

   /* For the anonymous mappings, create a memory region backed up
    * by physical memory. */
   if (pager.installed && __fd != -1)
   {
      /* Not anonymous, so create a region with only logical memory. */
      sig->reg.options |= MM_CREATE_REGION_FAULTABLE | MM_LOGMEM_FROM_ADDRESS;
   }

   send(&sig, mm_pid);
   sig = receive(reply);

   return (struct mm_create_region *) sig;
}

/*
 *===========================================================================
 *                        ose_fsmap
 *===========================================================================
 *
 * Description
 * -----------
 * Mmaps all the pages in the range [addr, addr+length] and gives
 * the region a name.
 *
 * Parameters
 * ----------
 * region_name The name of the region.
 * addr        The start address of the region to be mapped.
 * len         The length of the region to be mapped.
 * permissions The permissions of the mmaped region.
 *             The permission syntax is defined in mm.sig.
 * cache_mode  Specifies the cache attributes.
 *             The cache syntax defined in mm.sig.
 * fd          The file descriptor.
 * offset      The offset. Should be a multiple of the page size.
 *
 * Returns
 * -------
 * int        Returns zero on success and -1 on failure and sets errno.
 *
 */
void *
ose_fsmap (const char *region_name,
                 void *__addr,
                 size_t __len,
                 uint32_t permissions,
                 uint32_t cache_mode,
                 int __fd,
                 off_t __offset)
{
   struct mm_create_region *sig;
   enum PROCESS_TYPE type = get_ptype(current_process());
   struct MmanPageMap *page_map;
   int status;
   int supports = KSSS_PAGING_SUPPORTS_FILE_WRITE |
                  KSSS_PAGING_SUPPORTS_INTERRUPTS_DISABLED;
   mman_status_t mman_status;
   hal_status_t  hal_status;
   OSDOMAIN domain;
   struct KsssPagingCallbackApi cb_api;

   if (!(type == OS_PRI_PROC || type == OS_BLOCK || type == OS_BG_PROC))
   {
      errno = ENOMEM;
      return MAP_FAILED;
   }

   sig = mmap_mm(region_name,__addr, __len, permissions, cache_mode,
                 __fd, __offset);

   if (sig->status == MM_SUCCESS && __addr > 0 &&
       (OSADDRESS)__addr != sig->base)
   {
      ose_mm_free(sig->base);
      sig->status = MM_EOVERLAPPING_ADDRESS;
   }
      
   if (sig->status != MM_SUCCESS)
   {
      free_buf((union SIGNAL **)&sig);
      errno = ENOMEM;
      return MAP_FAILED;
   }

   if (pager.installed && __fd != -1)
   {
      cb_api.invalidate = mman_page_invalidate;
      cb_api.alloc_page = pager.alloc_page_func;
      status = mman_map_fd(__fd, &cb_api, PAGE_SIZE, &supports,
                           __offset, __len, &page_map);
      if (status < 0)
      {
         ose_mm_free((OSADDRESS)__addr);
         errno = -status;
         return MAP_FAILED;
      }

      /* FIXME: supports should be checked here to ensure that what is
       *        needed is fulfilled. */
   }
   else
   {
      page_map = NULL;
   }

   /* FIXME: Hard coded to domain 0 due to SAS and PM.
    *        (Unable to do mmap in other domains.)
    *
    * OSDOMAIN domain = (OSDOMAIN)zzkrn_get_pid_info(current_process(), 16);
    */
   domain = 0;
   __addr = (void *)sig->base;

   free_buf((union SIGNAL **)&sig);

   /* Need to round permissions to make validate_buffer
    * safe on paged regions */
   hal_status = hal_actual_permissions(
      (U32*)&permissions, permissions, (log_addr_t)__addr, 0);
   if (hal_status != HAL_SUCCESS)
   {
      /* Free the mmaped region if we failed to allocate a new
       * vma struct, and return an error. */
      if (page_map != NULL)
      {
         mman_unmap_fd(page_map, __offset, __len);
      }

      ose_mm_free((OSADDRESS)__addr);

      errno = EINVAL;
      return MAP_FAILED;
   }

   mman_status = add_vma(domain, (log_addr_t)__addr, __len, permissions,
                            cache_mode, page_map, (uint64_t)__offset);
   if (mman_status != MMAN_SUCCESS)
   {
      /* Free the mmaped region if we failed to allocate a new
       * vma struct, and return an error. */
      if (page_map != NULL)
      {
         mman_unmap_fd(page_map, __offset, __len);
      }

      ose_mm_free((OSADDRESS)__addr);

      if (mman_status == MMAN_REGION_OVERLAPPING)
      {
         errno = EINVAL;
      }
      else
      {
         errno = ENOMEM;
      }

      return MAP_FAILED;
   }

   if (!pager.installed && __fd != -1)
   {
      /* If the pager is not enabled, copy the file contents over to
       * the region from the file descriptor. This also takes care of
       * the SFK case. */
      int tmp_fd;
      off_t offset;
      union SIGNAL *sig;
      static const SIGSELECT sig_sel[] = {1, MM_CHANGE_PERMISSIONS_REPLY};

      tmp_fd = dup(__fd);
      if (tmp_fd == -1)
      {
         ramlog_printf("Error %s: mman failed to dup the mmaped file.\n",
                       strerror(errno));
         error2(OSE_MMAN_INTERNAL_ERROR, 0);
      }

      offset = lseek(tmp_fd, __offset, SEEK_SET);
      if (offset == (off_t)-1)
      {
         ramlog_printf("Error %s: mman failed to lseek the mmaped file.\n",
                       strerror(errno));
         error2(OSE_MMAN_INTERNAL_ERROR, 0);
      }

      status = read(tmp_fd, __addr, __len);
      if (status == -1)
      {
         ramlog_printf("Error %s: mman failed to copy the mmaped file.\n",
                       strerror(errno));
         error2(OSE_MMAN_INTERNAL_ERROR, 0);
      }

      status = close(tmp_fd);
      if (status == -1)
      {
         ramlog_printf("Error %s: mman failed to close the mmaped file.\n",
                       strerror(errno));
         error2(OSE_MMAN_INTERNAL_ERROR, 0);
      }

      if (!(permissions & MM_SU_WRITE))
      {
         /* Revert the permissions to read-execute. */
         sig = alloc(sizeof(struct mm_change_permissions),
                     MM_CHANGE_PERMISSIONS_REQUEST);
         sig->perm.base = (OSADDRESS)__addr;
         sig->perm.domain = 0;
         sig->perm.permissions = permissions;
         send(&sig, mm_pid);

         sig = receive(sig_sel);
         if (sig->perm.status != MM_SUCCESS)
         {
            ramlog_printf("Error %d: mman failed to change permissions.\n",
                          status);
            error2(OSE_MMAN_INTERNAL_ERROR, 0);
         }
         free_buf(&sig);
      }
   }

   return __addr;
}

/*
 *===========================================================================
 *                        mman_mmap
 *===========================================================================
 *
 * Description
 * -----------
 * Mmaps all the pages in the range [addr, addr+length).
 *
 * Parameters
 * ----------
 * addr     The start address of the region to be mapped.
 * len      The length of the region to be mapped
 * prot     The protection of the mmaped region.
 *          PROT_EXEC, PROT_READ, PROT_WRITE, and PROT_NONE are the valid
 *          protection settings.
 * flags    Specifies the mapping type. Valid flags are:
 *          MAP_FIXED, MAP_SHARED, MAP_ANONYMOUS and MAP_PRIVATE.
 * fd       The file descriptor.
 * offset   The offset. Should be a multiple of the page size.
 *
 * Returns
 * -------
 * int        Returns zero on success and -1 on failure and sets errno.
 *
 */
static void *
mman_mmap (void *addr, size_t len,
           int prot, int flags,
           int fd, off_t offset)
{
   uint32_t permissions;
   
   if (flags & MAP_PRIVATE)
   {
      /* Only MAP_SHARED is supported. */
      errno = ENOTSUP;
      return MAP_FAILED;
   }
   if ((flags & (MAP_PRIVATE | MAP_SHARED)) != MAP_SHARED)
   {
      /* Must provide one of MAP_SHARED and MAP_PRIVATE, but not both. */
      errno = EINVAL;
      return MAP_FAILED;
   }
   if (len == 0)
   {
      /* This is not really correct, as the specification states that
       * mmap should return a pointer to a piece of memory that
       * cannot be used. */
      errno = EINVAL;
      return MAP_FAILED;
   }
   if ((size_t)addr + (size_t)len < (size_t)addr)
   {
      /* The mapped region is outside the possible address space. */
      if (!(flags & MAP_FIXED))
      {
         /* Throw away the hint address. */
         addr = 0;
      }
      else
      {
         errno = ENOMEM;
         return MAP_FAILED;
      }
   }
   if (((ptrdiff_t)addr & (PAGE_SIZE - 1)) != 0)
   {
      /* Address must be page aligned. */
      if (!(flags & MAP_FIXED))
      {
         /* Throw away the hint address. */
         addr = 0;
      }
      else
      {
         errno = EINVAL;
         return MAP_FAILED;
      }
   }

   if (prot & PROT_EXEC)
   {
      /* PROT_EXEC alone is not allowed in OSE. Add read permissions too. */
      prot |= PROT_READ;
   }

   if (!(flags & MAP_ANONYMOUS))
   {
      struct stat buf;
      int status;
      uint32_t tmp;

      if ((prot & PROT_WRITE) || prot == PROT_NONE)
      {
         errno = ENOTSUP;
         return MAP_FAILED;
      }
      if (fd == -1)
      {
         errno = EBADF;
         return MAP_FAILED;
      }

      status = fstat(fd, &buf);
      if (status != 0)
      {
         errno = EACCES;
         return MAP_FAILED;
      }
      /* Check that the fd is not a directory or a socket. */
      if (!(S_ISREG(buf.st_mode) || !S_ISLNK(buf.st_mode)))
      {
         errno = EACCES;
         return MAP_FAILED;
      }
      /* Check that the file supports the requested mmap permissions. */
      if (prot & PROT_READ)
      {
         status = read(fd, &tmp, sizeof(tmp));
         if (status == -1)
         {
            errno = EACCES;
            return MAP_FAILED;
         }
      }
      /* Offset should not be >= file size. */
      if (offset >= buf.st_size)
      {
            errno = EINVAL;
            return MAP_FAILED;
      }
      /* Offset must be page aligned. */
      if ((offset & (PAGE_SIZE - 1)) != 0)
      {
         errno = EINVAL;
         return MAP_FAILED;
      }
      /* Append only files not legal to map with PROT_WRITE.
       * We do not need to check this at the moment. */
   }
   else
   {
      fd = -1;
      offset = 0;
   }

   /* User and supervisor mode permissions will be the same. */
   permissions = prot | (prot << MM_SU_SHIFT);

   if (flags & MAP_FIXED)
   {
      if (addr == 0)
      {
         errno = EINVAL;
         return MAP_FAILED;
      }
      return ose_fsmap("mmap", addr, len, permissions,
                       MM_CACHE_COPY_BACK | MM_SPECULATIVE_ACCESS, fd, offset);
   }
   else
   {
      /* If there is a hint address try it first. */
      if (addr != 0)
      {
         void *return_addr;

         return_addr = ose_fsmap("mmap", addr, len, permissions,
                                 MM_CACHE_COPY_BACK | MM_SPECULATIVE_ACCESS,
                                 fd, offset);
         if (return_addr != MAP_FAILED)
         {
            return return_addr;
         }
      }
      return ose_fsmap("mmap", 0, len, permissions,
                       MM_CACHE_COPY_BACK | MM_SPECULATIVE_ACCESS, fd, offset);
   }
}

/*
 *===========================================================================
 *                        mman_munmap
 *===========================================================================
 *
 * Description
 * -----------
 * Unmaps all the pages in the mmaped range [start, start+length).
 *
 * Returns
 * -------
 * int        Returns zero on success and -1 on failure and sets errno.
 *
 */
static int
mman_munmap(log_addr_t start, size_t length)
{
   struct _vm_area_t *vma;
   mm_status_t status;
   enum PROCESS_TYPE type = get_ptype(current_process());

   /* FIXME: Implement unmapping of a partially mmapped file. */
   (void)length;
   
   if (!(type == OS_PRI_PROC || type == OS_BLOCK || type == OS_BG_PROC))
   {
      errno = ENOMEM;
      return -1;
   }

   /* Check that it is an munmap to a previous mmap. */
   /* FIXME: Hard coded to domain 0 due to SAS and PM.
    *        (Unable to do munmap in other domains.) */
   vma = find_vma_internal(0, start);
   if (!vma)
   {
      errno = EINVAL;
      return -1;
   }

   status = ose_mm_free((OSADDRESS)vma->vm_area.base);
   if (status != MM_SUCCESS)
   {
      errno = EINVAL;
      return -1;
   }

   /* The deletion of the vma is performed by the registered delete_region
    * event callback. */
   return 0;
}

/*
 *===========================================================================
 *                           mman_mlock
 *===========================================================================
 *
 * Description
 * -----------
 * Implements the mlock functionality. Locks all pages in the range
 * [addr, addr+len) from the page cache.
 *
 * Returns
 * -------
 * int        Returns zero on success and -1 on failure and sets errno.
 *
 */
int
mman_mlock(log_addr_t addr, size_t len)
{
   enum PROCESS_TYPE type = get_ptype(current_process());

   if (!(type == OS_PRI_PROC || type == OS_BLOCK || type == OS_BG_PROC))
   {
      errno = ENOMEM;
      return -1;
   }
   if (((ptrdiff_t)addr & (PAGE_SIZE - 1)) != 0)
   {
      /* Address must be page aligned. */
      errno = EINVAL;
      return -1;
   }

   if (pager.installed)
   {
      pager_status_t status;
      OSDOMAIN domain = (OSDOMAIN)zzkrn_get_pid_info(current_process(), 16);

      status = (pager.lock_func)(domain, addr, len, REGULAR_MLOCK, 1);
      if (status == PAGER_SUCCESS || status == PAGER_REGION_NOT_REGISTERED)
      {
         return 0;
      }
      else
      {
         errno = ENOMEM;
         return -1;
      }
   }

   return 0;
}

/*
 *===========================================================================
 *                        ose_mman_mlock_temporarily
 *===========================================================================
 *
 * Description
 * -----------
 * Implements an internal mlock functionality only to be used temporarily.
 * This version is much faster then the regular mlock, but will not try
 * to optimize the location of the mlocked page.
 * Locks all pages in the range [addr, addr+len) from the page cache.
 *
 * Returns
 * -------
 * int        Returns zero on success and -1 on failure and sets errno.
 *
 */
int
ose_mman_mlock_temporarily(log_addr_t addr, size_t len)
{
   enum PROCESS_TYPE type = get_ptype(current_process());

   if (!(type == OS_PRI_PROC || type == OS_BLOCK || type == OS_BG_PROC))
   {
      errno = ENOMEM;
      return -1;
   }

   if (pager.installed)
   {
      pager_status_t status;
      OSDOMAIN domain = (OSDOMAIN)zzkrn_get_pid_info(current_process(), 16);

      /* Page align the address and adjust the size. */
      len += (ptrdiff_t)addr & (PAGE_SIZE - 1);
      addr = ALIGN_DOWN(addr, PAGE_SIZE);

      status = (pager.lock_func)(domain, addr, len, REGULAR_MLOCK, 0);
      if (status == PAGER_SUCCESS || status == PAGER_REGION_NOT_REGISTERED)
      {
         return 0;
      }
      else
      {
         errno = ENOMEM;
         return -1;
      }
   }

   return 0;
}

/*
 *===========================================================================
 *                        ose_mman_mlock_counting
 *===========================================================================
 *
 * Description
 * -----------
 * Implements the mlock functionality. Locks all pages in the range
 * [addr, addr+len) from the page cache.
 * Address does not need to be page aligned, and the mlock will
 * have a count associated with it. Every time a page is mlocked, the
 * count is increased by one.
 * Returns
 * -------
 * int        Returns zero on success and -1 on failure and sets errno.
 *
 */
int
ose_mman_mlock_counting(log_addr_t addr, size_t len)
{
   if (pager.installed)
   {
      pager_status_t status;
      OSDOMAIN domain = (OSDOMAIN)zzkrn_get_pid_info(current_process(), 16);

      /* Page align the address and adjust the size. */
      len += (ptrdiff_t)addr & (PAGE_SIZE - 1);
      addr = ALIGN_DOWN(addr, PAGE_SIZE);

      status = (pager.lock_func)(domain, addr, len, COUNTED_MLOCK, 1);
      if (status == PAGER_SUCCESS || status == PAGER_REGION_NOT_REGISTERED)
      {
         return 0;
      }
      else
      {
         errno = ENOMEM;
         return -1;
      }
   }

   return 0;
}

/*
 *===========================================================================
 *                        mman_munlock
 *===========================================================================
 *
 * Description
 * -----------
 * Implements the munlock functionality. Unlocks all pages in the range
 * [addr, addr+len) from the page cache.
 *
 * Returns
 * -------
 * int        Returns zero on success and -1 on failure and sets errno.
 *
 */
int
mman_munlock(log_addr_t addr, size_t len)
{
   if (((ptrdiff_t)addr & (PAGE_SIZE - 1)) != 0)
   {
      /* Address must be page aligned. */
      errno = EINVAL;
      return -1;
   }

   if (pager.installed)
   {
      OSDOMAIN domain = (OSDOMAIN)zzkrn_get_pid_info(current_process(), 16);

      (pager.unlock_func)(domain, addr, len, REGULAR_MLOCK);
      /* Will always succeed. */
   }

   return 0;
}


/*
 *===========================================================================
 *                     ose_mman_munlock_temporarily
 *===========================================================================
 *
 * Description
 * -----------
 * Implements the munlock functionality. Unlocks all pages in the range
 * [addr, addr+len) from the page cache. Accepts non paged aligned
 * addresses.
 *
 * Returns
 * -------
 * int        Returns zero on success and -1 on failure and sets errno.
 *
 */
int
ose_mman_munlock_temporarily(log_addr_t addr, size_t len)
{
   if (pager.installed)
   {
      OSDOMAIN domain = (OSDOMAIN)zzkrn_get_pid_info(current_process(), 16);

      /* Page align the address and adjust the size. */
      len += (ptrdiff_t)addr & (PAGE_SIZE - 1);
      addr = ALIGN_DOWN(addr, PAGE_SIZE);

      (pager.unlock_func)(domain, addr, len, REGULAR_MLOCK);
      /* Will always succeed. */
   }

   return 0;
}


/*
 *===========================================================================
 *                     ose_mman_munlock_counting
 *===========================================================================
 *
 * Description
 * -----------
 * Implements the munlock functionality. Unlocks all pages in the range
 * [addr, addr+len) from the page cache. Accepts non paged aligned
 * addresses. This lock will
 * have a count associated with it. Every time a page is munlocked, the
 * count is decreased by one, and it is not unlocked until it reaches 0.
 *
 * Returns
 * -------
 * int        Returns zero on success and -1 on failure and sets errno.
 *
 */
int
ose_mman_munlock_counting(log_addr_t addr, size_t len)
{
   if (pager.installed)
   {
      OSDOMAIN domain = (OSDOMAIN)zzkrn_get_pid_info(current_process(), 16);

      /* Page align the address and adjust the size. */
      len += (ptrdiff_t)addr & (PAGE_SIZE - 1);
      addr = ALIGN_DOWN(addr, PAGE_SIZE);

      (pager.unlock_func)(domain, addr, len, COUNTED_MLOCK);
      /* Will always succeed. */
   }

   return 0;
}


/*
 *===========================================================================
 *                        vma_accessible
 *===========================================================================
 *
 * Description
 * -----------
 * Determines if a vma is accessible from a specific domain given a set
 * of permission flags.
 *
 * Returns
 * -------
 * Non-zero if the vma is accessible.
 */
static int
vma_accessible(struct vm_area_t *vma, OSDOMAIN domain,
                  unsigned int perm, OSADDRESS supervisor)
{
   /* FIXME: Currently paging in global regions where supervisor
    * priviledges are irrelevant. Need to check supervisor mode
    * once support for local regions is implemented and used. */
   (void)supervisor; 

   return (vma->permissions & perm) == perm &&
      (domain == vma->domain || (vma->domain == 0 && vma->mapping_type != MAS));
}

/*
 *===========================================================================
 *                        vma_properties
 *===========================================================================
 *
 * Description
 * -----------
 * Used by mman_validate_buffer to optain different properties of a vma,
 * such as if it resides in MAS mapped memory.
 */
static void
vma_properties(struct vm_area_t *vma, unsigned options, unsigned *result)
{
   if (result != NULL)
   {
      if (options & MM_VALIDATE_IN_MAS)
      {
         if (vma->mapping_type == MAS)
         {
            *result |= MM_VALIDATE_IN_MAS;
         }
      }
   }
}

/*
 *===========================================================================
 *                        mman_validate_buffer
 *===========================================================================
 *
 * Description
 * -----------
 * Validates access to a buffer without accessing it.
 * See ose_mm_validate_buffer in mm.h for details.
 *
 * Returns
 * -------
 * size_t        Returns the number of accessible bytes or 0 if requester has
 *               insufficient permissions or the address is unmapped.
 */
size_t
mman_validate_buffer(log_addr_t bbase,	          /* buffer base */
                     size_t bsize,	          /* buffer size */
                     PROCESS pid,		  /* validation context */
                     unsigned int bperm,  	  /* buffer permissions */
                     unsigned int options,	  /* validation options */
                     unsigned int *result)	  /* option results */
{
   OSADDRESS supervisor_mode;
   OSDOMAIN domain;
   struct vm_area_t *vma;                         /* vma for paged mem */
   size_t asize = 0, tsize = 0;

   if (pager.installed != 0)
   {
      unsigned int tperm = bperm;

      if (pid == 0)
      {
	 pid = current_process();
      }
      else
      {
	 /* Check pid. Only allow process and block identifiers */
	 if (zzkrn_get_pid_info(pid, 9) > OS_BLOCK)
	 {
	    return 0;
	 }
      }

      /* Adjust permissions. */
      tperm &= (MM_PERM_READ|MM_PERM_WRITE);

      /* Get execution mode for access process */
      supervisor_mode = zzkrn_get_pid_info(pid, 10);

      if (supervisor_mode)
      {
         tperm <<= MM_SU_SHIFT;
      }

      /* Get domain for process */
      domain = zzkrn_get_pid_info(pid, 16);

      if (domain > MM_MAX_DOMAINS)
      {
         return 0;
      }

      if (result != NULL)
      {
         *result = 0;
      }

      do
      {
         /* FIXME: Domain hardcoded to 0 due to PM and SAS */
         vma = ose_find_vma((OSDOMAIN)0, bbase + asize);

         if (vma != NULL)
         {
            if (vma_accessible(vma, domain, tperm, supervisor_mode))
            {
               vma_properties(vma, options, result);
               tsize += MIN(bsize, vma->size - (bbase - vma->base));
               asize += tsize;
               bsize -= tsize;
            }
            else
            {
               tsize = 0;
            }
         }
         else
         {
            unsigned int result_temp = 0;
            tsize = mm_validate_buffer(bbase+asize, bsize, pid,
                                       bperm, options, &result_temp);
            asize += tsize;
            bsize -= tsize;
            if (result != NULL)
               *result |= result_temp;
         }
      } while ((bbase + asize != 0) && asize < bsize && tsize != 0);

      return asize;
   }

   return mm_validate_buffer(bbase, bsize, pid, bperm, options, result);
}

/***************************************************************************
 * VIRTUAL MEMORY AREA (VMA) MANIPULATION FUNCTIONS
 *
 * You can call:
 *   add_vma
 *   delete_vma
 *   find_vma
 *   get_first_vma
 *
 * MM can call:
 *   delete_region_callin
 *
 ***************************************************************************/

/*
 *===========================================================================
 *                        add_vma
 *===========================================================================
 *
 * Description
 * -----------
 * Adds a new vma and initializes it to the supplied domain, base,
 * size, permissions, cache mode, inode, and file_offset.
 *
 * Returns
 * -------
 * mman_status_t      Returns MMAN_SUCCESS for success.
 *
 */
mman_status_t
add_vma(OSDOMAIN domain, log_addr_t base, size_t size, uint32_t permissions,
        uint32_t cache_mode, struct MmanPageMap *page_map, uint64_t file_offset)
{
   struct vm_area_t *vma;
   struct _vm_area_t *_vma;

   if (!find_overlapping_vma(domain, base, size))
   {
      struct partition_class *partition;
      mapping_type_t map_type;
      mm_status_t status;

      _vma = (struct _vm_area_t *)slab_alloc(vmas.slab, SLAB_SLEEP);
      if (!_vma)
      {
         return MMAN_OUT_OF_MEMORY;
      }

      vma = &_vma->vm_area;
      vma->domain = domain;
      vma->base = base;
      vma->size = size;
      vma->page_map = page_map;
      vma->file_offset = file_offset;
      vma->cache_mode = cache_mode;
      vma->permissions = permissions;
      vma->page_cache = NULL;

      partition = mm_find_partition_by_address(base);
      if (partition == NULL)
      {
         return MMAN_OUT_OF_MEMORY;
      }

      status = mm_find_partition_mapping_type(partition, NULL, &map_type);
      if (status != MM_SUCCESS)
      {
         ramlog_printf("Error: mman failed to determine mapping type for new vma\n");
         error2(OSE_MMAN_INTERNAL_ERROR, 0);
      }

      vma->mapping_type = map_type;
      
      core_rbtree_insert(&vmas.tree, (struct core_rbtree_node_t *)_vma);
   }
   else
   {
      return MMAN_REGION_OVERLAPPING;
   }

   return MMAN_SUCCESS;
}

/*
 *===========================================================================
 *                        find_vma_internal
 *===========================================================================
 *
 * Description
 * -----------
 * Finds a vma uniquely identified by the tuple log_addr and domain.
 *
 * Returns
 * -------
 * struct _vm_area_t*  A pointer to the internal vma entry,
 *                     or NULL if the entry cannot be found.
 *
 */
struct _vm_area_t*
find_vma_internal(OSDOMAIN domain, log_addr_t address)
{
   struct _vm_area_t vma_key;
   struct _vm_area_t *vma;

   /* FIXME: Hard coded to domain 0 due to the installation procedure
    * used by PM and the fact that faultable regions are SAS mapped. */
   (void)domain;

   vma_key.vm_area.base = address;
   vma_key.vm_area.domain = 0;
   vma_key.vm_area.size = 4;
   vma = (struct _vm_area_t *)
      core_rbtree_find(&vmas.tree, compare_vmas, (void *)&vma_key);

   return vma;
}

/*
 *===========================================================================
 *                        ose_find_vma
 *===========================================================================
 *
 * Description
 * -----------
 * Finds a vma uniquely identified by the tuple log_addr and domain.
 *
 * Returns
 * -------
 * struct vm_area_t*  A pointer to the vma entry,
 *                    or NULL if the entry cannot be found.
 *
 */
struct vm_area_t*
ose_find_vma(OSDOMAIN domain, log_addr_t address)
{
   struct _vm_area_t *vma = find_vma_internal(domain, address);

   if (vma)
   {
      return &vma->vm_area;
   }
   else
   {
      return NULL;
   }
}

/*
 *===========================================================================
 *                        find_overlapping_vma
 *===========================================================================
 *
 * Description
 * -----------
 * Finds an already registered vma that overlapps the supplied one, or
 * NULL if none can be found.
 *
 * Returns
 * -------
 * struct vm_area_t*  A pointer to the overlapping vma entry,
 *                    or NULL if such an entry cannot be found.
 *
 */
struct vm_area_t*
find_overlapping_vma(OSDOMAIN domain, log_addr_t address, size_t size)
{
   struct _vm_area_t vma_key;
   struct _vm_area_t *vma;

   vma_key.vm_area.base = address;
   vma_key.vm_area.domain = domain;
   vma_key.vm_area.size = size;
   vma = (struct _vm_area_t *)
      core_rbtree_find(&vmas.tree, compare_vmas, (void *)&vma_key);

   if (vma)
   {
      return &vma->vm_area;
   }
   else
   {
      return NULL;
   }
}

/*
 *===========================================================================
 *                        get_first_vma
 *===========================================================================
 *
 * Description
 * -----------
 * Gets the first vma. Usefull when iterating over all vmas while deleting.
 *
 * Returns
 * -------
 * struct vm_area_t*  A pointer to the vma entry,
 *                    or NULL if there are no more entries.
 *
 */
struct vm_area_t *
get_first_vma(void)
{
   struct _vm_area_t *vma;

   vma = (struct _vm_area_t *)core_rbtree_first(&vmas.tree);
   if (vma)
   {
      return &vma->vm_area;
   }
   else
   {
      return NULL;
   }
}


/*
 *===========================================================================
 *                        delete_vma
 *===========================================================================
 *
 * Description
 * -----------
 * Deletes a vma. Called from MM context.
 *
 */
static void
delete_vma(struct _vm_area_t *vma)
{
   core_rbtree_remove(&vmas.tree, vma);
   if (vma->vm_area.page_map != 0)
   {
      mman_unmap_fd(vma->vm_area.page_map,
                    vma->vm_area.file_offset,
                    vma->vm_area.size);
   }

   if (pager.installed)
   {
      pager.unmap_vma_func(vma->vm_area.page_map, &vma->vm_area);
   }

   slab_free(vmas.slab, (void *)vma);
}

/*
 *===========================================================================
 *                        delete_region_callin
 *===========================================================================
 *
 * Description
 * -----------
 * Called by MM when a region is deleted.
 *
 */
void
delete_region_callin(union mm_event_t *event_raw)
{
   struct _vm_area_t *vma;
   struct mm_region_deleted_event_t *event = &event_raw->delete;

   vma = find_vma_internal(event->domain, event->base);
   if (vma)
   {
      delete_vma(vma);
   }
}

/*
 *===========================================================================
 *                        compare_vmas
 *===========================================================================
 *
 * Description
 * -----------
 * Local function for comparing two rbtree tree nodes containing vma
 * information structures.
 *
 */
static int
compare_vmas(const void *a_raw, const void *b_raw)
{
   const struct vm_area_t *a = &((const struct _vm_area_t *)a_raw)->vm_area;
   const struct vm_area_t *b = &((const struct _vm_area_t *)b_raw)->vm_area;

   if ((a->domain == b->domain || (a->mapping_type != MAS)) &&
       ((a->base >= b->base && a->base < b->base + b->size) ||
        (b->base >= a->base && b->base < a->base + a->size)))
   {
      return 0;
   }
   else if ((a->domain > b->domain) || (a->domain == b->domain &&
                                        a->base > b->base))
   {
      return 1;
   }
   else
   {
      return -1;
   }
}

#if 0
/*
 *===========================================================================
 *                        mman_invalidate_mapping
 *===========================================================================
 *
 * Description
 * -----------
 * Function used by the pager tests to stress the invalidation
 * functionality.
 *
 */
static void
mman_invalidate_mapping(OSDOMAIN domain, log_addr_t address,
                        size_t size, log_addr_t new_data,
                        phys_addr_t new_phys_addr)
{
   if (pager.installed)
   {
      struct vm_area_t *vma;

      vma = ose_find_vma(domain, address);
      if (vma)
      {
         pager.invalidate_func(vma->inode, vma->file_offset, size,
                               new_data, new_phys_addr);
      }
   }
}
#endif

/*
 *===========================================================================
 *                        mman_pager_control
 *===========================================================================
 *
 * Description
 * -----------
 * Function used by the pager tests .
 *
 */
static int
mman_pager_control(pager_control_op_t operation, uint32_t arg1, uint32_t arg2,
                   uint32_t arg3, uint32_t arg4, uint32_t arg5)
{
   if (pager.installed)
   {
      pager_status_t status;

      status = pager.control_func(operation, arg1, arg2, arg3, arg4, arg5);
      if (status != PAGER_SUCCESS)
      {
         return -1;
      }
   }
   return 0;
}

/*
 *===========================================================================
 *                        mman_trap_interface
 *===========================================================================
 *
 * Description
 * -----------
 * The bios trap interface for the mman CRT functions.
 *
 */
static long
mman_trap_interface(long fc, long arg1, long arg2, long arg3,
                    long arg4, long arg5, long arg6)
{
   long result;

   switch (fc)
   {
      case MMAN_TRAP_GET_INTERFACE_VERSION:
      {
	 result = (long)MMAN_TRAP_INTERFACE_VERSION;
	 break;
      }
      case MMAN_TRAP_GET_LAST_FCODE:
      {
	 /* Update when adding function codes to the MM trap interface */
	 result = (long)MMAN_TRAP_PAGER_CONTROL;
	 break;
      }
      case MMAN_TRAP_MMAP:
      {
	 result = (long)mman_mmap((void *) arg1,
                                  (size_t) arg2,
                                  (int) arg3,
                                  (int) arg4,
                                  (int) arg5,
                                  (off_t) arg6);
	 break;
      }
      case MMAN_TRAP_MUNMAP:
      {
	 result = (long)mman_munmap((log_addr_t) arg1,
                                    (size_t) arg2);
	 break;
      }
      case MMAN_TRAP_MLOCK:
      {
	 result = (long)mman_mlock((log_addr_t) arg1,
                                   (size_t) arg2);
	 break;
      }
      case MMAN_TRAP_MUNLOCK:
      {
	 result = (long)mman_munlock((log_addr_t) arg1,
                                     (size_t) arg2);
	 break;
      }
      case MMAN_TRAP_VALIDATE_BUFFER:
      {
         result = (long)mman_validate_buffer((log_addr_t)   arg1,
                                             (size_t)       arg2,
                                             (PROCESS)      arg3,
                                             (unsigned int) arg4,
                                             (unsigned int) arg5,
                                             (unsigned int*)arg6);
	 break;
      }
      case MMAN_TRAP_PAGER_CONTROL:
      {
         result = (long)mman_pager_control((pager_control_op_t) arg1,
                                           (uint32_t) arg2,
                                           (uint32_t) arg3,
                                           (uint32_t) arg4,
                                           (uint32_t) arg5,
                                           (uint32_t) arg6);
	 break;
      }
      default:
      {
	 result = -1;
	 break;
      }
   }
   return result;
}


/*
 *===========================================================================
 *                        init_mman
 *===========================================================================
 *
 * Description
 * -----------
 * Initializes the mman subsystem.
 *
 */
void
init_mman(void)
{
   struct mm_region_deleted_event_t mm_event;
   long status;
   mm_status_t mm_status;
   pager.installed = 0;

   status = biosInstall(MMAN_BIOS_NAME, mman_trap_interface,
                        BIOS_USERMODE_ALLOWED);
   if(status != BIOS_SUCCESS)
   {
      ramlog_printf("Error: mman failed to install bios interface.\n");
      error2(OSE_MMAN_INTERNAL_ERROR, 0);
   }

   /* Initialize the vma storage structure. */
   core_rbtree_init(&vmas.tree, compare_vmas);

   vmas.slab = create_slab_allocator("vm_area slab",
                                     sizeof(struct _vm_area_t),
                                     0, 0, NULL, NULL, 0);

   /* Register callbacks for the MM events we are interested in. */
   mm_event.base = 0;
   mm_event.domain = MM_EVENT_ANY_DOMAIN;
   mm_event.identifier = COMMON_MM_COMPONENT;
   mm_event.event_handler = delete_region_callin;
   mm_event.event_no = MM_EVENT_REGION_DELETED;
   mm_status = ose_mm_install_handler((union mm_event_t *)&mm_event, 0);
   if (mm_status != MM_SUCCESS)
   {
      ramlog_printf("Mman: ose_mm_install_handler failed.\n");
      error2(OSE_MMAN_INTERNAL_ERROR, 0);
   }
}



