/* COPYRIGHT-ENEA-SRC-R2 *
**************************************************************************
* Copyright (C) 2004-2009 by Enea Software AB.
* All rights reserved.
*
* This Software is furnished under a software license agreement and
* may be used only in accordance with the terms of such agreement.
* Any other use or reproduction is prohibited. No title to and
* ownership of the Software is hereby transferred.
*
* PROPRIETARY NOTICE
* This Software consists of confidential information.
* Trade secret law and copyright law protect this Software.
* The above notice of copyright on this Software does not indicate
* any actual or intended publication of such Software.
**************************************************************************
* COPYRIGHT-END */

#include "ose.h"
#include "osetypes.h"
#include "stdio.h"
#include "ksss_mm.h"
#include "sys/types.h"
#include "sys/stat.h"
#include "string.h"
#include "fm.sig"
#include "errno.h"
#include "efs.h"
#include "ksss_paging.h"
#include "mm_err.h"
#include "heapapi.h"
#include "ramlog.h"
#include "mm.h"
#include "slab.h"
#include "list_ext.h"
#include "pose.h"
#include "mman_internal.h"
#include "core_segtree.h"
#include "core_hash.h"


#define KSSS_DELETE_FMHANDLE_REQUEST  0xFEFA
#define KSSS_DELETE_FMHANDLE_REPLY    0xFEFB
#define KSSS_SHARE_FMHANDLE_REQUEST   0xFEFC
#define KSSS_SHARE_FMHANDLE_REPLY     0xFEFD
#define KSSS_PAGE_FAULT_REQUEST       0xFEFE
#define KSSS_PAGE_FAULT_REPLY         0xFEFF


/*
 * @brief Inode cache.
 *
 */
struct KsssInodeInternal
{
   SIGSELECT sig_no;
   FmHandle fmhandle;
   FmHandle proxy_fmhandle;
   PROCESS  vm_pid;
   struct KsssPagingCallbackApi cb_api;
   size_t page_size;
   OSATTREF attach_handle;
   OSADDRESS data_addr;
};

/*
 * @brief struct PageFaultRead
 *
 */
struct PageFaultRead
{
   SIGSELECT sig_no;
   struct KsssInodeInternal *inode;
   struct KsssPageFault *pgfault;
   int result;
};

struct ShareFmHandle
{
   SIGSELECT sig_no;
   FmHandle  fmhandle;
   PROCESS   paging_pid;
   PROCESS   vm_pid;
   OSATTREF  attach_handle;
};

struct DeleteFmHandle
{
   SIGSELECT sig_no;
   FmHandle  fmhandle;
   PROCESS   vm_pid;
   OSATTREF  attach_handle;
};

union SIGNAL
{
   SIGSELECT sig_no;
   struct FmSetPos seek;
   struct FmReadPtr read_ptr;
   struct FmReadReply read_value;
   struct FmExamineReply examine;
   struct PageFaultRead pf_read;
   struct ShareFmHandle share_fmhandle;
   struct DeleteFmHandle delete_fmhandle;
   struct FmDupHandle dup_handle;
   struct FmClose fm_close;
};

struct EmDataInfo
{
   PROCESS vm;
   OSADDRESS addr;
};

struct MmanPageMapInternal
{
   core_hash_link_t     link;
   core_segtree_t       ranges;
   struct MmanPageMap   page_map;
};

struct MmanPageMapRange
{
   core_segtree_node_t  node;
};



/* A reference to the system heap that we need to allocate everything from. */
static OSHEAP                   system_heap;
static PROCESS                  proxy_pid;
static struct KsssMapFunctions  ksss_map_funcs;
static struct slab_cache       *page_map_slab = NULL;
static struct slab_cache       *page_map_range_slab = NULL;
static core_hash_table_t page_map_hash;
static MUTEX                    page_map_mutex;


static int
proxy_copy_page(struct KsssInodeInternal *inode,
                struct KsssPageFault *pgfault)
{
   int res = 0;
   union SIGNAL *reply;
   static const SIGSELECT sigsel2[] = { 1, FM_READ_PTR_REPLY };

   sendFmReadPtrRequest(inode->fmhandle,
                        (unsigned int)pgfault->offset_bytes,
                        False,
                        inode->page_size,
                        (char*) pgfault->data_log_addr,
                        inode->vm_pid);

   reply = receive_from(1000, sigsel2, inode->vm_pid);

   if (reply != NIL)
   {
      if (reply->read_ptr.status == EFS_SUCCESS)
      {
         res = reply->read_ptr.actual;
      }
      else
         /* Read failed in volume manager. */
      {
         /*errno = efs_status_to_errno(reply->read_ptr.status);*/
         errno = (int)(reply->read_ptr.status);
      }
      free_buf(&reply);
   }
   else
      /* Lost contact with volume manager. */
   {
      errno = ESTALE;
   }

   return res;
}



/* IMPORTANT!!!
 * We need this mmap_read_agent process because we cannot send a signal out
 * into the intricate web of OSE processes from the page faulting process
 * context. The reason for this is that that process in itself, might
 * have a number of outstanding signals in the system. If we then sent
 * one more and waited for an answer, we might get an answer that was
 * destined for the application as we are using the same signalling queue.
 * Therefor, all paged read requests must be handled by another processes. */
static void
mmap_read_agent(void)
{
   OSATTREF handle;
   static const SIGSELECT pf_sig[] = {4,
                                      KSSS_PAGE_FAULT_REQUEST,
                                      KSSS_SHARE_FMHANDLE_REQUEST,
                                      KSSS_DELETE_FMHANDLE_REQUEST,
                                      OS_ATTACH_SIG};

   for (;;)
   {
      union SIGNAL *sig, *sig2;
      PROCESS client;

      sig = receive(pf_sig);
      client = sender(&sig);

      switch(sig->sig_no)
      {
         union SIGNAL *attach_sig;
         static const SIGSELECT fm_close_mask[] = {1, FM_CLOSE_REPLY};

         case KSSS_PAGE_FAULT_REQUEST:
            sig->pf_read.result = proxy_copy_page(sig->pf_read.inode,
                                                  sig->pf_read.pgfault);
            sig->sig_no = KSSS_PAGE_FAULT_REPLY;
            send(&sig, client);
            break;

         case KSSS_SHARE_FMHANDLE_REQUEST:
            attach_sig = alloc(sizeof(struct DeleteFmHandle), OS_ATTACH_SIG);
            attach_sig->delete_fmhandle.fmhandle =
               sig->share_fmhandle.fmhandle;
            attach_sig->delete_fmhandle.vm_pid = sig->share_fmhandle.vm_pid;

            handle = attach(&attach_sig, sig->share_fmhandle.paging_pid);

            sig->share_fmhandle.attach_handle = handle;
            sig->sig_no = KSSS_SHARE_FMHANDLE_REPLY;
            send(&sig, client);
            break;

         case KSSS_DELETE_FMHANDLE_REQUEST:
            sendFmCloseRequest(sig->delete_fmhandle.fmhandle, 1,
                               sig->delete_fmhandle.vm_pid);
            sig2 = receive(fm_close_mask);
            if (sig2->sig_no != FM_CLOSE_REPLY)
            {
               ramlog_printf("Error: mmap_read_agent did not receive an FmCloseReply.\n");
               error2(OSE_PAGING_INTERNAL_ERROR, 0);
            }
            else if (!(sig2->fm_close.status != EFS_SUCCESS ||
                       sig2->fm_close.status != FM_EINVALID_HANDLE))
            {
               ramlog_printf("Error: FmClose failed in mmap_read_agent.\n");
               error2(OSE_PAGING_INTERNAL_ERROR, 0);
            }

            if (sig2->fm_close.status == EFS_SUCCESS)
            {
               detach(&sig->delete_fmhandle.attach_handle);
            }
            free_buf(&sig2);
            free_buf(&sig);
            /* Do not send a reply as ksss_unmap() will not receive
             * it anyway. */
            break;

         case OS_ATTACH_SIG:
            sendFmCloseRequest(sig->delete_fmhandle.fmhandle, 1,
                               sig->delete_fmhandle.vm_pid);
            sig2 = receive(fm_close_mask);
            if (sig2->sig_no != FM_CLOSE_REPLY)
            {
               ramlog_printf("Error: mmap_read_agent did not receive an FmCloseReply.\n");
               error2(OSE_PAGING_INTERNAL_ERROR, 0);
            }
            else if (!(sig2->fm_close.status != EFS_SUCCESS ||
                       sig2->fm_close.status != FM_EINVALID_HANDLE))
            {
               ramlog_printf("Error: FmClose failed in mmap_read_agent.\n");
               error2(OSE_PAGING_INTERNAL_ERROR, 0);
            }
            free_buf(&sig2);

            free_buf(&sig);
            break;

         default:
            ramlog_printf("Error: mmap_read_agent recieved an unknown signal.\n");
            error2(OSE_PAGING_INTERNAL_ERROR, 0);
      }
   }
}



static uint32_t
page_map_hash_value(const void *entry)
{
   const struct MmanPageMapInternal *page_map =
      (const struct MmanPageMapInternal*)entry;
   return (uint32_t)page_map->page_map.inode;
}

static int
page_map_hash_compare(const void *_left, const void *_right)
{
   const struct MmanPageMapInternal *left  =
      (const struct MmanPageMapInternal*)_left;
   const struct MmanPageMapInternal *right =
      (const struct MmanPageMapInternal*)_right;

   return (left->page_map.inode == right->page_map.inode);
}

static void
page_map_hash_deleted(core_hash_link_t *link)
{
   slab_free(page_map_slab, link);
}


static int
page_map_range_compare(const void *_left, const void *_right)
{
   const struct MmanPageMapRange *left =
      (const struct MmanPageMapRange*)_left;
   const struct MmanPageMapRange *right =
      (const struct MmanPageMapRange*)_right;

   if ((left->node.start  >= right->node.start &&
        left->node.start  < right->node.end) ||
       (right->node.start >= left->node.start  &&
        right->node.start < left->node.end))
   {
      return 0;
   }
   else if (left->node.start > right->node.start)
   {
      return 1;
   }
   else
   {
      return -1;
   }
}



static void *
page_map_range_alloc(size_t size)
{
   (void)size;

   return slab_alloc(page_map_range_slab, SLAB_SLEEP);
}

static int
page_map_range_free(void *ptr)
{
   slab_free(page_map_range_slab, ptr);
   return 0;
}

static int
page_map_range_ctor(core_segtree_t *tree,
                    core_segtree_node_t *template,
                    core_segtree_node_t *new_node, uint32_t create)
{
   int status = 0;

   (void)template;
   /* When create is set, a new range is added and
    * actions must be taken to map it in the filesystem. */
   if (create)
   {
      struct MmanPageMapInternal *page_map = (struct MmanPageMapInternal*)
         ((char*)tree - offsetof(struct MmanPageMapInternal, ranges));

      status = ksss_map_funcs.map(&page_map->page_map, new_node->start,
                                  new_node->end - new_node->start);
   }

   return status;
}

static int
page_map_range_dtor(core_segtree_t *tree,
                    core_segtree_node_t *template,
                    core_segtree_node_t *old_node, uint32_t destroy)
{
   int status = 0;

   (void)template;
   
   /* When destroy is set, an existing range is removed and
    * actions must be taken to unmap it in the filesystem. */
   if (destroy)
   {
      struct MmanPageMapInternal *page_map = (struct MmanPageMapInternal*)
         ((char*)tree - offsetof(struct MmanPageMapInternal, ranges));

      status = ksss_map_funcs.unmap(&page_map->page_map, old_node->start,
                                    old_node->end - old_node->start);
   }

   return status;
}

static int
page_map_range_add(struct MmanPageMapInternal *page_map,
                   uint64_t offset, size_t size)
{
   core_segtree_node_t template;

   core_segtree_node_init(&page_map->ranges, &template,
                          offset, offset + size);

   return core_segtree_add(&page_map->ranges, &template);
}

static int
page_map_range_remove(struct MmanPageMapInternal *page_map,
                      uint64_t offset, size_t size)
{
   core_segtree_node_t template;

   core_segtree_node_init(&page_map->ranges, &template,
                          offset, offset + size);

   return core_segtree_remove(&page_map->ranges, &template);
}



static core_segtree_funcs_t page_map_funcs =
{
   page_map_range_compare,
   NULL,
   page_map_range_alloc,
   page_map_range_free,
   page_map_range_ctor,
   page_map_range_dtor
};


static struct MmanPageMapInternal *
page_map_find(struct KsssInode *inode, core_hash_iterator_t *it)
{
   struct MmanPageMapInternal page_map_key;
   core_hash_link_t *link;

   page_map_key.page_map.inode = inode;

   core_hash_iterator_init(it, page_map_hash_compare);
   
   link = core_hash_find(&page_map_hash, it, &page_map_key);
   if (link == NULL)
   {
      return NULL;
   }

   return (struct MmanPageMapInternal*)link;
}

static void
page_map_init(struct MmanPageMapInternal *page_map)
{
   memset(page_map, 0, sizeof(struct MmanPageMapInternal));

   core_segtree_init(&page_map->ranges, &page_map_funcs,
                     sizeof(core_segtree_node_t));
}

static int
page_map_register(struct MmanPageMapInternal *page_map)
{
   core_hash_link_init(&page_map->link);
   return core_hash_insert(&page_map_hash, &page_map->link) == 1 ?  0 : -1;
}

static int
page_map_unregister(core_hash_iterator_t *page_map_it)
{
   core_hash_delete(page_map_it);
   return 0;
}

static int
page_map_empty(struct MmanPageMapInternal *page_map)
{
   return core_segtree_empty(&page_map->ranges);
}


static int
mman_map_init(void)
{
   static core_hash_funcs_t page_map_hash_funcs =
      {
         page_map_hash_compare,
         page_map_hash_value,
         page_map_hash_deleted
      };
   
   /* Set up the slab cache used when allocating page map-objects.
    * The cache is anonymous and can be reused by anyone, such as
    * the heap. It must be synchronous (SLAB_SLEEP) since mman is
    * available in systems without a pager and therefore runs with
    * MM unprotected against concurrency. */
   
   page_map_slab
      = create_slab_allocator(NULL, /* Anonymous */
                              sizeof(struct MmanPageMapInternal),
                              0,   /* Alignment (default) */
                              0,   /* Number of spares */
                              NULL, /* Constructor */
                              NULL, /* Destructor  */
                              SLAB_SLEEP);
   if (page_map_slab == NULL)
   {
      ramlog_printf("mman: Could not create slab cache for "
                    "allocation of page map objects\n");
      error2(OSE_MMAN_INTERNAL_ERROR, 0);
   }

   page_map_range_slab
      = create_slab_allocator(NULL, /* Anonymous */
                              sizeof(struct MmanPageMapRange),
                              0,   /* Alignment (default) */
                              0,   /* Number of spares */
                              NULL, /* Constructor */
                              NULL, /* Destructor  */
                              SLAB_SLEEP);
   if (page_map_range_slab == NULL)
   {
      ramlog_printf("mman: Could not create slab cache for "
                    "allocation of page map objects\n");
      error2(OSE_MMAN_INTERNAL_ERROR, 0);
   }

   core_hash_table_init(&page_map_hash, 128, &page_map_hash_funcs);

   return 0;
}



/*
 * @function mman_map_fd
 *
 * @brief Map the blocks of the file pointed to by file desc.
 *
 * @brief THIS FUNCTION IS TO BE USED BY THE CORE 'MMAP' MODULE
 *
 * @return 0 success, <0 negative errno code
 *
 */
int mman_map_fd(int fd,
                const struct KsssPagingCallbackApi *cb_api,
                size_t page_size_bytes,
                int *inout_supports,
                uint64_t offset,
                size_t size,
                struct MmanPageMap **out_page_map)
{
   union SIGNAL *sig, *sig2;
   struct KsssInodeInternal *inode;
   struct MmanPageMapInternal *page_map;
   PROCESS pid;
   static const SIGSELECT pf_sig[] =
      {2, FM_DUP_HANDLE_REPLY, KSSS_SHARE_FMHANDLE_REPLY};
   static const SIGSELECT examine_sig[] =
      {1, FM_EXAMINE_REPLY};

   *out_page_map = NULL;
   
#ifndef softOSE
   /*
    * If KSSS module is present, redirect call to that module
    * instead.
    */
   if(ksss_map_funcs.create && ksss_map_funcs.map)
   {
      struct KsssInode *inode;
      core_hash_iterator_t page_map_it;
      int status;

      if (!page_map_slab)
      {
         mman_map_init();
      }
      
      status = ksss_map_funcs.get_inode(fd, &inode);
      if (status < 0)
      {
         ramlog_printf("mman: Unable to retrieve inode from FS, "
                       "status = %d\n", status);
         return status;
      }

      zzkrn_kill_lock(0);
      ose_mutex_lock(&page_map_mutex);

      page_map = page_map_find(inode, &page_map_it);
      if (page_map == NULL)
      {
         /* This inode has never been mapped. Need to create
          * a new map object and register it along with the new
          * vma soon to be created. */
         page_map = (struct MmanPageMapInternal*)
            slab_alloc(page_map_slab, SLAB_SLEEP);
         if (page_map == NULL)
         {
            ksss_map_funcs.unref_inode(inode);

            /* Done with iterator, unreference any link pointed at */
            core_hash_iterator_release(&page_map_it);

            ose_mutex_unlock(&page_map_mutex);
            zzkrn_kill_unlock(0);
            ramlog_printf("mman: Failed to allocate a new page map-object\n");
            return -ENOMEM;
         }

         page_map_init(page_map);

         status = ksss_map_funcs.create(fd, inode, cb_api, page_size_bytes,
                                        inout_supports, &page_map->page_map);
      
         if (status < 0)
         {
            ksss_map_funcs.unref_inode(inode);
            core_hash_iterator_release(&page_map_it);

            slab_free(page_map_slab, page_map);

            ose_mutex_unlock(&page_map_mutex);
            zzkrn_kill_unlock(0);
            ramlog_printf("mman: Failed to map inode, status = %d\n", status);
            return status;
         }

         status = page_map_register(page_map);
         if (status < 0)
         {
            ksss_map_funcs.unref_inode(inode);
            ksss_map_funcs.destroy(&page_map->page_map);
            core_hash_iterator_release(&page_map_it);

            slab_free(page_map_slab, page_map);

            ose_mutex_unlock(&page_map_mutex);
            zzkrn_kill_unlock(0);
            ramlog_printf("mman: Page map already exists\n");
            error2(OSE_MMAN_INTERNAL_ERROR, 0);
         }

         
         core_hash_iterator_release(&page_map_it);
         page_map = page_map_find(inode, &page_map_it);
      }

      /* We no longer need to reference the inode */
      ksss_map_funcs.unref_inode(inode);

      status = page_map_range_add(page_map, offset, size);
      if (status < 0)
      {
         if (page_map_empty(page_map))
         {
            ksss_map_funcs.destroy(&page_map->page_map);
            page_map_unregister(&page_map_it);
         }

         /* Unreference hash link. If the page_map was unregistered,
          * a deleted-hook is called when the last iterator is released. */
         core_hash_iterator_release(&page_map_it);
         
         ose_mutex_unlock(&page_map_mutex);
         zzkrn_kill_unlock(0);
         return status;
      }

      /* Done with iterator, unreference any link pointed at */
      core_hash_iterator_release(&page_map_it);

      ose_mutex_unlock(&page_map_mutex);
      zzkrn_kill_unlock(0);

      *out_page_map = &page_map->page_map;
      
      return EOK;
   }

   
#endif

   if(!proxy_pid)
   {
      /* Get a reference to the system heap.
       * We need to always allocate from the system heap as we will
       * get calls from many different contexts, including MM. */
      hunt("osemain", 0, &pid, NULL);
      if (pid == 0)
      {
         hunt("main", 0, &pid, NULL);
         if (pid == 0)
         {
            ramlog_printf("ERROR: Could not find the osemain process.\n");
            error2(OSE_PAGING_INTERNAL_ERROR, 0);
         }
      }
      system_heap = ose_heap_ref(pid);
      if (system_heap == 0)
      {
         ramlog_printf("ERROR: Could not find the system heap.\n");
         error2(OSE_PAGING_INTERNAL_ERROR, 0);
      }

      proxy_pid = create_process(OS_PRI_PROC,         /* proc type.   */
                                 "mmap_read_agent",   /* name.        */
                                 mmap_read_agent,     /* entrypoint.  */
                                 0x1000,              /* stack size.  */
				 20,                  /* priority.    */
				 0,            /* timeslice.   */
				 get_bid(pid), /* block.       */
				 NULL,         /* redir table. */
				 0,            /* vector.      */
				 0);           /* user id.     */
      start(proxy_pid);
   }

   inode = (struct KsssInodeInternal *)
      ose_heap_malloc(system_heap, sizeof(struct KsssInodeInternal),
                      __FILE__, __LINE__);
   if (inode == NULL)
   {
      return -ENOMEM;
   }

   memset(inode, 0, sizeof(struct KsssInodeInternal));
   
   efs_examine_fd(fd, FLIB_FD_HANDLE, &inode->fmhandle, 0);
   efs_examine_fd(fd, FLIB_FD_VMPID, &inode->vm_pid, 0);

   /* Obtain map-in-place address if it exists */
   sendFmExamineRequest(inode->fmhandle, inode->vm_pid);
   sig = receive_w_tmo(1000, examine_sig);
   inode->data_addr = 0;

   if (sig != NULL)
   {
      struct EmDataInfo *di = (struct EmDataInfo*)sig->examine.info.s;
      if (inode->vm_pid == di->vm && di->addr != 0)
      {
         phys_addr_t phys_addr;

         if (ose_mm_l2p(current_process(), (log_addr_t)di->addr, &phys_addr)
             == MM_SUCCESS)
         {
            inode->data_addr = (OSADDRESS)phys_addr;
         }
      }
      free_buf(&sig);
   }

   inode->page_size = page_size_bytes;
   *inout_supports = 0;
   memcpy(&inode->cb_api, cb_api, sizeof(struct KsssPagingCallbackApi));

   page_map = ose_heap_malloc(system_heap, sizeof(struct MmanPageMapInternal),
                              __FILE__, __LINE__);
   if (page_map == NULL)
   {
      ose_heap_free(system_heap, inode);
      ramlog_printf("mman: Failed to allocate a new page map object\n");
      return -ENOMEM;
   }

   memset(page_map, 0, sizeof(struct MmanPageMapInternal));
   
   page_map->page_map.inode = (struct KsssInode *)inode;

   /* Duplicate the file handle. */
   sendFmDupHandleRequest(inode->fmhandle, proxy_pid, inode->vm_pid);
   sig = receive_w_tmo(1000, pf_sig);
   if (!sig)
   {
      slab_free(page_map_slab, page_map);
      ose_heap_free(system_heap, inode);
      return -ETIMEDOUT;
   }
   else if (sig->sig_no != FM_DUP_HANDLE_REPLY)
   {
      slab_free(page_map_slab, page_map);
      ose_heap_free(system_heap, inode);
      ramlog_printf("Error: mmap_read_agent recieved an unknown signal.\n");
      return -EAGAIN;
   }
   else if (sig->dup_handle.status != EFS_SUCCESS)
   {
      slab_free(page_map_slab, page_map);
      ose_heap_free(system_heap, inode);
      ramlog_printf("Error: dup failed in mmap_read_agent.\n");
      return -EMFILE;
   }

   inode->proxy_fmhandle = sig->dup_handle.newHandle;

   /* Give the file handle to the mmap proxy. */
   sig2 = alloc(sizeof(struct ShareFmHandle),
                KSSS_SHARE_FMHANDLE_REQUEST);
   sig2->share_fmhandle.fmhandle = sig->dup_handle.newHandle;
   sig2->share_fmhandle.paging_pid = current_process();
   sig2->share_fmhandle.vm_pid = inode->vm_pid;
   free_buf(&sig);
   send(&sig2, proxy_pid);
   sig2 = receive(pf_sig);
   inode->attach_handle = sig2->share_fmhandle.attach_handle;
   free_buf(&sig2);

   *out_page_map = &page_map->page_map;

   return EOK;
}



/*
 * @function mman_unmap_fd
 *
 * @brief Function that unmaps a file.
 *
 * @long Function that unmaps a file.
 *       Unreference a ksss_fs.h FsMap and FsMap.
 *       This function can be called from MM context, so absolutely
 *       no signalling to the MM here.
 *
 * @brief THIS FUNCTION IS TO BE USED BY THE CORE 'MMAP' MODULE
 *
 *
 */
int mman_unmap_fd(struct MmanPageMap *_page_map,
                   uint64_t offset, size_t size)
{
   struct KsssInodeInternal *inode =
      (struct KsssInodeInternal *)_page_map->inode;
   struct MmanPageMapInternal *page_map = (struct MmanPageMapInternal*)
      ((char*)_page_map - offsetof(struct MmanPageMapInternal, page_map));
   union SIGNAL *sig;

#ifndef softOSE   
   int status;
   core_hash_iterator_t page_map_it;

   /*
    * If KSSS module is present, redirect call to that module
    * instead.
    */

   if (ksss_map_funcs.destroy && ksss_map_funcs.unmap)
   {
      if (!page_map_slab)
      {
         mman_map_init();
      }

      zzkrn_kill_lock(0);
      ose_mutex_lock(&page_map_mutex);

      /* Reference the page map to prevent race condition with
       * concurrent page invalidation from block layer */
      core_hash_iterator_init(&page_map_it, page_map_hash_compare);
      core_hash_reference(&page_map_hash, &page_map_it, &page_map->link);
      
      status = page_map_range_remove(page_map, offset, size);
      if (status < 0)
      {
         core_hash_iterator_release(&page_map_it);

         ose_mutex_unlock(&page_map_mutex);
         zzkrn_kill_unlock(0);
         ramlog_printf("mman: Failed to unmap inode, status = %d\n", status);
         return status;
      }

      if (page_map_empty(page_map))
      {
         while (!core_hash_reserve(&page_map_it, 1))
         {
            delay(10);
         }

         ksss_map_funcs.destroy(&page_map->page_map);
         page_map_unregister(&page_map_it);
      }

      core_hash_iterator_release(&page_map_it);

      ose_mutex_unlock(&page_map_mutex);
      zzkrn_kill_unlock(0);

      return EOK;
   }
#endif
   
   /* Send a delete message to the proxy. */
   sig = alloc(sizeof(struct DeleteFmHandle),
               KSSS_DELETE_FMHANDLE_REQUEST);
   sig->delete_fmhandle.fmhandle = inode->proxy_fmhandle;
   sig->delete_fmhandle.attach_handle = inode->attach_handle;
   sig->delete_fmhandle.vm_pid = inode->vm_pid;
   send(&sig, proxy_pid);
   /* Do not wait for any reply, as the file system might also want to use the
    * MM to deallocate some data. */

   ose_heap_free(system_heap, page_map);
   ose_heap_free(system_heap, inode);

   return EOK;
}

/*
 * @function mman_page_fault
 *
 * @brief Handle a page fault
 *
 * @long  This function is called from the pager when a page fault has
 *        occurred. If ksss module is present, it just passes the call.
 *
 * @input  inode       Mapped inode.
 * @input  page_fault  Struct describing the page fault. Also contains the
 *                     result at return.
 *
 * @return
 *   @tab 0          @tab if success. @break
 *   @tab -EIO       @tab if I/O error occurred. @break
 *   @tab -ECANCELED @tab if the operations should be cancelled. @break
 */
int mman_page_fault(struct MmanPageMap *page_map,
                    struct KsssPageFault *pgfault)
{
   union SIGNAL *sig = NULL;
   union SIGNAL *reply = NULL;
   int rv;
   struct KsssInodeInternal *inode =
      (struct KsssInodeInternal *)page_map->inode;
   
   int ksss_status = EOK;
   static const SIGSELECT response_sigsel[] = { 1, KSSS_PAGE_FAULT_REPLY };

#ifndef softOSE
   /*
    * If KSSS module is present, redirect call to that module
    * instead.
    */

   if (ksss_map_funcs.page_fault)
   {
      return ksss_map_funcs.page_fault(page_map, pgfault);
   }
#endif

   if (inode->data_addr != 0 && (pgfault->map_type & KSSS_MAP_IN_PLACE))
   {
#if 0
      if(inode->cb_api.alloc_page)
      {
         ksss_status = inode->cb_api.alloc_page(pgfault);
      }
      if (ksss_status == -ECANCELED)
      {
         /* The operation was cancelled by the alloc_page callback. */
         return ksss_status;
      }
      else if (ksss_status != EOK)
      {
         ramlog_printf("Error: alloc_page failed. Status %u.\n", ksss_status);
         error2(OSE_PAGING_INTERNAL_ERROR, 0);
      }
      pgfault->map_type = KSSS_MAP_COPY;
      pgfault->pages_mapped = 1;
      memcpy((char*)pgfault->data_log_addr,
             (char*)(inode->data_addr + pgfault->offset_bytes),
             inode->page_size);
#else
      pgfault->map_type = KSSS_MAP_IN_PLACE;
      pgfault->pages_mapped = 1;
      pgfault->data_phys_addr = inode->data_addr + pgfault->offset_bytes;
#endif
      return EOK;
   }

   if(inode->cb_api.alloc_page)
   {
      ksss_status = inode->cb_api.alloc_page(pgfault);
   }
   if (ksss_status == -ECANCELED)
   {
      /* The operation was cancelled by the alloc_page callback. */
      return ksss_status;
   }
   else if (ksss_status != EOK)
   {
      ramlog_printf("Error: alloc_page failed. Status %u.\n", ksss_status);
      error2(OSE_PAGING_INTERNAL_ERROR, 0);
   }


   sig = alloc(sizeof(struct PageFaultRead),
               KSSS_PAGE_FAULT_REQUEST);

   sig->pf_read.inode = inode;
   sig->pf_read.pgfault = pgfault;

   send(&sig,proxy_pid);

   do
   {
      reply = receive_from((OSTIME)~0u, response_sigsel, proxy_pid);
   } while(reply == NIL);

   rv = reply->pf_read.result;
   free_buf(&reply);

   pgfault->pages_mapped = 1;
   if (rv >= 0 && rv < (int)inode->page_size)
   {
      /* No more data in file, the rest should be zero. A special case
       * is if the whole page should be zero, in which case we perform a
       * zero map if this is supported by the demand pager. */
      if (((size_t)rv == 0) &&
          ((pgfault->map_type & KSSS_MAP_ZERO) != 0))
      {
         /* Perform zero map */
         pgfault->map_type = KSSS_MAP_ZERO;
      }
      else
      {
         /* Fill the rest of the page with zeroes */
         pgfault->map_type = KSSS_MAP_COPY;
         bzero((void *)(pgfault->data_log_addr + rv), inode->page_size - rv);
      }
   }
   else if (rv == (int)inode->page_size)
   {
      pgfault->map_type = KSSS_MAP_COPY;
   }
   else
   {
      return -EIO;
   }

   return EOK;
}


/*
 * @function mman_page_invalidate
 *
 * @brief Called by filesystem/block device when pages have
 *        been invalidated to force the pager to remap them.
 *
 * @long  See ksss_paging.h for more details.        
 *
 * @return Nothing.
 */
void
mman_page_invalidate(struct KsssInode *inode, uint64_t offset,
                     size_t size, log_addr_t new_data,
                     phys_addr_t new_phys_addr)
{
   struct MmanPageMapInternal *page_map;
   core_hash_iterator_t page_map_it;
  
   page_map = page_map_find(inode, &page_map_it);
   if (page_map == NULL)
   {
      core_hash_iterator_release(&page_map_it);
      return;
   }

   pager.invalidate_func(&page_map->page_map, offset,
                         size, new_data, new_phys_addr);

   /* The file may have been unmapped during the invalidation.
    * In that case, this iterator is the last one to reference the page_map.
    * A deleted-hook is called to free the memory used by the page_map
    * object. All other resources have already been released during unmap. */
   core_hash_iterator_release(&page_map_it);
}


void
mman_ksss_init(struct KsssMapFunctions *ksss_funcs)
{
   ose_mutex_init(&page_map_mutex, OSE_MUTEX_PRIO_INHERIT);

   memcpy(&ksss_map_funcs, ksss_funcs, sizeof(struct KsssMapFunctions));

   if (ksss_map_funcs.create      == NULL ||
       ksss_map_funcs.destroy     == NULL ||
       ksss_map_funcs.map         == NULL ||
       ksss_map_funcs.unmap       == NULL ||
       ksss_map_funcs.get_inode   == NULL ||
       ksss_map_funcs.unref_inode == NULL ||
       ksss_map_funcs.page_fault  == NULL)
   {
      ramlog_printf("mman: KSSS mapping interface is incomplete\n");
      error2(OSE_MMAN_INTERNAL_ERROR, 0);
   }
}


