/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2006 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

/*
               An implementation of top-down splaying.
                    D. Sleator, March 1992.

  This is adapted from simple top-down splay, at the bottom of page 669 of
  "Self-adjusting Binary Search Trees" by Sleator and Tarjan, JACM Volume 32,
  No 3, July 1985.

  The chief modification here is that the splay operation works even if the
  item being splayed is not in the tree, and even if the tree root of the
  tree is NULL.  So the line:

                       t = splay(tree, t, key);

  causes it to search for item with key i in the tree rooted at t.   If it's
  there, it is splayed to the root.  If it isn't there, then the node put
  at the root is the last one before NULL that would have been reached in a
  normal binary search for i.  (It's a neighbor of i in the tree.)   This
  allows insertion, deletion, split and join to be easily implemented.
*/

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include "ramlog.h"
#include "splay_tree.h"

/* Compiling this file with RVCT and optimizations
 * causes the splay tree to be corrupted */
#ifdef __ARMCC_VERSION
#pragma O0
#endif

/***************** FUNCTIONS ************************/

/***************************************************************************
 * SPLAY TREE FUNCTIONS
 *
 * You can call:
 *   init_splay_tree
 *   insert_in_splay_tree
 *   delete_from_splay_tree
 *   find_in_splay_tree
 *
 ***************************************************************************/
#ifdef ASSERT
static int
count_nodes_in_branch(struct splay_node_t *node)
{
   if (node == NULL)
   {
      return 0;
   }
   
   return 1 + count_nodes_in_branch(node->left) +
      count_nodes_in_branch(node->right);
}

#define assert_tree(tree, line) \
{ \
   if (tree->size != count_nodes_in_branch(tree->root)) \
   { \
      ramlog_printf("ERROR: Splay tree is corrupted: %d\n", line); \
   } \
}
#else
#define assert_tree(tree, line)
#endif 

/*
 *===========================================================================
 *                        splay
 *===========================================================================
 *
 * Description
 * -----------
 * Local function for searching in a splay tree.
 * Should be called with the tree lock acquired. 
 *
 * Returns
 * -------
 * struct splay_node_t *   Returns the top of the tree, which is the entry
 *                         if it is found.
 *
 */
static struct splay_node_t *
splay(struct splay_tree_t *tree, struct splay_node_t *t, const void *key)
{
/* Simple top down splay, not requiring t to be in the tree. */
/* What it does is described above. */
   struct splay_node_t N, *l, *r, *y;

   if (t == NULL)
   {
      return t;
   }
   N.left = N.right = NULL;
   l = r = &N;

   for (;;)
   {
      int d = tree->compare(key, t);

      if (d < 0)
      {
         if (t->left == NULL)
         {
            break;
         }
         if (tree->compare(key, t->left) < 0)
         {
            /* rotate right */
            y = t->left;
            t->left = y->right;
            y->right = t;
            t = y;
            if (t->left == NULL)
            {
               break;
            }
         }
         /* link right */
         r->left = t;
         r = t;
         t = t->left;
      }
      else if (d > 0)
      {
         if (t->right == NULL)
         {
            break;
         }
         if (tree->compare(key, t->right) > 0)
         {
            /* rotate left */
            y = t->right;
            t->right = y->left;
            y->left = t;
            t = y;
            if (t->right == NULL)
            {
               break;
            }
         }
         /* link left */
         l->right = t;
         l = t;
         t = t->right;
      }
      else
      {
         break;
      }
   }

   /* assemble */
   l->right = t->left;
   r->left = t->right;
   t->left = N.right;
   t->right = N.left;

   return t;
}

/*
 *===========================================================================
 *                        find_in_splay_tree
 *===========================================================================
 *
 * Description
 * -----------
 * Searches for a matching entry in the splay tree *t.
 *
 * Returns
 * -------
 * struct splay_node_t *   Returns the found entry or NULL if it is not found.
 *
 */
struct splay_node_t *
find_in_splay_tree(struct splay_tree_t *tree, const void *key)
{
   Msr sr;
   struct splay_node_t *root;

   if (tree->root == NULL)
   {
      return NULL;
   }
   
   spin_lock_irqsave(&tree->lock, sr);
   tree->root = splay(tree, tree->root, key);

   if (tree->compare(tree->root, key) == 0)
   {
      root = tree->root;
      assert_tree(tree, __LINE__);
      spin_unlock_irqrestore(&tree->lock, sr);
      return root;
   }

   assert_tree(tree, __LINE__);
   spin_unlock_irqrestore(&tree->lock, sr);
   return NULL;
}


/*
 *===========================================================================
 *                        insert_in_splay_tree
 *===========================================================================
 *
 * Description
 * -----------
 * Inserts an entry in the splay tree *t unless it is already there.
 *
 * Returns
 * -------
 * int      Returns one if node was inserted, or zero is it was a duplicate.
 *
 */
int
insert_in_splay_tree(struct splay_tree_t *tree, struct splay_node_t *node)
{
   struct splay_node_t *x;
   Msr sr;

   spin_lock_irqsave(&tree->lock, sr);

   assert_tree(tree, __LINE__);

   if (tree->root == NULL)
   {
      node->left = NULL;
      node->right = NULL;
   }
   else
   {
      int d;

      x = splay(tree, tree->root, node);

      d = tree->compare(node, x);

      if (d < 0)
      {
         node->right = x;
         node->left = x->left;
         x->left = NULL;
      }
      else if (d > 0)
      {
         node->left = x;
         node->right = x->right;
         x->right = NULL;
      }
      else /* (node == parent) */
      {
         /* Duplicate */
         tree->root = x;
         spin_unlock_irqrestore(&tree->lock, sr);
         return 0;
      }
   }

   tree->root = node;
   tree->size++;

   assert_tree(tree, __LINE__);
   spin_unlock_irqrestore(&tree->lock, sr);
   return 1;
}

/*
 *===========================================================================
 *                        delete_from_splay_tree
 *===========================================================================
 *
 * Description
 * -----------
 * Deletes an entry in the splay tree *tree.
 *
 * Returns
 * -------
 * int     Returns one if node was deleted, or zero if it could not be found.
 *
 */
int
delete_from_splay_tree(struct splay_tree_t *tree, const void *key)
{
   struct splay_node_t *del, *new_root;
   Msr sr;

   if (tree->root == NULL)
   {
      return 0;
   }

   spin_lock_irqsave(&tree->lock, sr);
   assert_tree(tree, __LINE__);

   del = splay(tree, tree->root, key);

   if (tree->compare(del, key) == 0)
   {
      /* Found it! */
      if (del->left == NULL)
      {
         new_root = del->right;
      }
      else
      {
         new_root = splay(tree, del->left, key);
         assert(new_root->right == NULL);
         new_root->right = del->right;
      }
      tree->root = new_root;
      tree->size--;

      assert_tree(tree, __LINE__);
      spin_unlock_irqrestore(&tree->lock, sr);
      return 1;
   }
   else
   {
      tree->root = del;
   }

   spin_unlock_irqrestore(&tree->lock, sr);
   return 0;                         /* it wasn't there */
}

/*
 *===========================================================================
 *                         init_splay_tree
 *===========================================================================
 *
 * Description
 * -----------
 * Initializes a splay tree with a specific compare function.
 * Needs to be called before any other operation on the splay tree.
 *
 */
void
init_splay_tree(struct splay_tree_t *tree,
               int (*compare)(const void *, const void *))
{
   tree->root = NULL;
   tree->size = 0;
   tree->compare = compare;
   spin_lock_init(&tree->lock);
}

