/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2006 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

#ifndef __SPLAY_TREE_H__
#define __SPLAY_TREE_H__

#include <stdint.h>
#ifndef STANDALONE
#include "spinlock.h"
#else
#include "standalone.h"
#endif

struct splay_node_t
{
   struct splay_node_t *left;
   struct splay_node_t *right;
   /* User data here. */
};

struct splay_tree_t
{
   /* A pointer to the root entry in the splay tree. */
   struct splay_node_t  *root;

   /* The number of entries in the splay tree. */
   int                   size;

   /* A pointer to the compare function used to compare entries in the tree. */
   int                 (*compare)(const void *key_a, const void *key_b);

   /* A lock used to provide mutual exclusion for all operations on the
    * splay tree. The main drawback with splay trees are that they
    * require a lock even for find operations, since the tree is splayed. */
   spinlock_t           lock;
};


/*
 *===========================================================================
 *                         init_splay_tree
 *===========================================================================
 *
 * Description
 * -----------
 * Initializes a splay tree with a specific compare function.
 * Needs to be called before any other operation on the splay tree.
 *
 * Parameters
 * ----------
 * compare      Callback function for comparing two nodes. This
 *              function returns a value < 0 if key_a is less than key_b,
 *              > 0 if key_a is greater than key_b and == 0 if key_a
 *              equals key_b. 
 *
 */
void
init_splay_tree(struct splay_tree_t *tree,
                int (*compare)(const void *key_a, const void *key_b));

/*
 *===========================================================================
 *                        insert_in_splay_tree
 *===========================================================================
 *
 * Description
 * -----------
 * Inserts an entry in the splay tree *t unless it is already there.
 *
 * Returns
 * -------
 * int      Returns one if node was inserted, or zero is it was a duplicate.
 *
 */
int
insert_in_splay_tree(struct splay_tree_t *tree, struct splay_node_t *node);

/*
 *===========================================================================
 *                        delete_from_splay_tree
 *===========================================================================
 *
 * Description
 * -----------
 * Deletes an entry in the splay tree *tree.
 *
 * Returns
 * -------
 * int     Returns one if node was deleted, or zero if it could not be found.
 *
 */
int
delete_from_splay_tree(struct splay_tree_t *tree, const void *key);

/*
 *===========================================================================
 *                        find_in_splay_tree
 *===========================================================================
 *
 * Description
 * -----------
 * Searches for a matching entry in the splay tree *t.
 *
 * Returns
 * -------
 * struct splay_node_t *   Returns the found entry or NULL if it is not found.
 *
 */
struct splay_node_t *
find_in_splay_tree(struct splay_tree_t *tree, const void *key);


#endif  /* __SPLAY_TREE_H__ */
