/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2009 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */


/**
 * @private
 *
 * @file pfmi.c
 *
 * @brief Program manager Freeze Mode Interface
 *
 * @long
 * This file implements the Program manager Freeze Mode Interface. 
 * It may be used to fetch information from the executive in a freeze mode
 * environment. The purpose is to make Program concepts accessable
 * for freeze mode debug monitors that can't use ordinary system calls.
 * @long
 * WARNING: Use of pfmi functionality for any other purpose will severely 
 * damage the interrupt latency!
 * 
 */

/*
*****************************************************************************
** INCLUDE FILES
*****************************************************************************
*/

#if defined(HOST_TOOL_ENVIRONMENT)
/* macro for other arglists */
#define ARGLIST(...) (HostFmiInstance *i, __VA_ARGS__)
#else
/* macro for other arglists */
#define ARGLIST(...) (__VA_ARGS__)
#endif

#if defined(HOST_TOOL_ENVIRONMENT)

#include "pfmi_host.h"
#include "pfmi.h"
#include "efmi_host.h"
#include "efmi.h"

/* Updated every time the instance is entered. */
static MAKE_GLOBAL(FMI_TLS_PM_STATE, GlobalPmInternalState);
static SET_TYPE(PmInternalState, GlobalPmInternalState);

/* Avoid having to check for truncation everywhere. */
#define pfmi_strcpy(to,from)						\
do									\
{									\
  int truncated = fmi_strncpy_truncated(to,				\
                                        (const char *)from,		\
                                        sizeof(to));			\
  /* This would indicate memory corruption, so not very common. */	\
  if (truncated)							\
    {									\
      ERROR_PRINT("String truncation detected!\n");			\
      fmi_abort();							\
    }									\
} while (0)
   
#else

/* Make sure we don't use debug versions of system calls internally. */

#ifdef OS_DEBUG
#undef OS_DEBUG
#endif

#include "pfmi.h"		/* Public interface */
#include "pm_programs.h"	/* Program search interface */
#include "string.h"		/* str* and mem* functions */
#include "cpu_intlock.h"	/* Interrupt lock save and restore */
#include "oldlist.h"		/* List handling macros */
#include "stdint.h"		/* Standard integer types */
#include "efmi.h"               /* Freeze mode interface for executive */

extern struct PmInternalState * GlobalPmInternalState;

/* These macros are redefined when building the host version of pfmi. */
#define ENTER_PFMI(a)
#define EXIT_PFMI()

#define SET_TYPE(type, var)
#define SET_TYPE_DEF(type, var)
#define READ_VAL(ptr, member) (ptr).member
#define DEREF_VAL(ptr, member) (ptr)->member
#define READ_PTR(ptr, member) (ptr).member
#define DEREF_PTR(ptr, member) (ptr)->member
#define DEREF_VAL_TYPE(type, ptr, member) (ptr)->member
#define DEREF_PTR_TYPE(type, ptr, member) (ptr)->member

#define efmi_get_process_info_internal ose_efmi_get_process_info
#define efmi_get_block_info_internal ose_efmi_get_block_info

#endif

/*
*****************************************************************************
** MACROS
*****************************************************************************
*/

/*
*****************************************************************************
** TYPES
*****************************************************************************
*/

/*
 *---------------------------------------------------------------------------
 * Size of the local cursor structs must not exeed the size of the public
 * view of the corresponding struct, found in pfmi.h!
 *---------------------------------------------------------------------------
 */

struct ProgramCursor  
{
  struct List *node;
};

struct ProgPidCursor  
{
  struct EfmiSegmentCursor pc;
  struct EfmiBlockCursor bc;
};

struct LoadModuleCursor          
{
  struct List *node;
};

/*
*****************************************************************************
** LOCAL FUNCTIONS
*****************************************************************************
*/

static int
is_pm_initialized(void)
{
#if !defined(HOST_TOOL_ENVIRONMENT)
   return (GlobalPmInternalState != NULL);
#else
   /* In the host environment, this check is performed once during init. */
   return 1;
#endif   
}

static ProgramMetaData *
validate_pid(PROCESS pid)
{
  struct EfmiProcessInfo pi;
  struct EfmiBlockInfo bi;  
  PROCESS segpid;
  
  if (EFMI_OK == efmi_get_process_info_internal(pid, &pi, sizeof(pi)))
    {
      /* Translate process pid to block pid */
       pid = pi.block;
    }
  
  if (EFMI_OK == efmi_get_block_info_internal(pid, &bi, sizeof(bi)))
    {
      /* Translate block pid to segment pid */
      segpid = bi.segment;
    }
  else
    {
      /* Assume segment pid, if not it will not be found among programs */
      segpid = pid;
    }
  
  /* Verify that segpid corresponds to a valid program */
  if (!is_pm_initialized())
    return NULL;
  else
    return (ProgramMetaData*)traverseList1(DEREF_PTR(GlobalPmInternalState, programs), 
                                           pm_program_search_by_segpid,
                                           (void *)segpid);
}

static PfmiStatus
get_progpid(PROCESS pid,
	    PROCESS *progpid)
{
  ProgramMetaData * prog;

  SET_TYPE_DEF(ProgramMetaData, prog);

  /* Verify that program corresponding to PID exists */
  prog = validate_pid(pid);

  /* Check if found */
  if (NULL == prog)
     return PFMI_UNKNOWN;
  
  *progpid = DEREF_VAL(prog, progpid);  
  return PFMI_OK;
}


static void  
copy_program_info(ProgramMetaData *from, 
		  struct PfmiProgramInfo *to)
{
  SET_TYPE_DEF(ProgramMetaData, from);
  
  to->domain       = DEREF_VAL(from, domain);
  to->progpid      = DEREF_VAL(from, progpid);
  to->main_block   = DEREF_VAL(from, mainbid);
  to->main_process = DEREF_VAL(from, mainpid);
  to->stk_pool_base= DEREF_VAL(from, stk_pool_base);
  to->stk_pool_size= DEREF_VAL(from, stk_pool_size);
  to->uid          = DEREF_VAL(from, uid);
  to->state        = DEREF_VAL(from, state);  
  to->segpid       = DEREF_VAL(from, segpid);
  to->stk_poolid   = DEREF_VAL(from, stk_poolid);
  to->sig_poolid   = DEREF_VAL(from, sig_poolid);
  to->sig_pool_base= DEREF_VAL(from, sig_pool_base);
  to->sig_pool_size= DEREF_VAL(from, sig_pool_size);
  to->heap         = DEREF_VAL(from, heap);
#if defined (HOST_TOOL_ENVIRONMENT)
  pfmi_strcpy(to->install_handle, fmi_get_addr("LmRoMetaData",
                                               DEREF_VAL(from, lm),
                                               "install_handle"));
#else
  strcpy (to->install_handle, from->lm->install_handle);
#endif  
}


static PfmiStatus 
get_program_info(PROCESS pid,
		 struct PfmiProgramInfo *pi)
{
  ProgramMetaData *prog;

  /* Check that PM is initiated */
  if(!is_pm_initialized())
    return PFMI_UNAVAILABLE;  

  /* Check that data is not locked by PM process */
  if(!DEREF_VAL(GlobalPmInternalState, data_accessable))
    return PFMI_DATA_LOCKED;  

  /* Validate pid */
  prog = validate_pid(pid);
  
  /* Check if found */
  if(NULL == prog)
    return PFMI_UNKNOWN;
  
  /* Copy program information */
  copy_program_info(prog, pi);
  
  /* Everything went well */
  return PFMI_OK;  
}


static PfmiStatus
get_next_program(struct PfmiProgramCursor *pc,
		 struct PfmiProgramInfo   *pi)
{
  ProgramMetaData *prog;
  struct ProgramCursor* cursor = (struct ProgramCursor*)pc;
  
  prog = (ProgramMetaData *) cursor->node; 

  /* Check that PM is initiated */
  if(!is_pm_initialized())
    return PFMI_UNAVAILABLE;  
  
  /* Check that data is not locked by PM process */
  if(!DEREF_VAL(GlobalPmInternalState, data_accessable))
    return PFMI_DATA_LOCKED;  
  
  /* Check if end of list */
  if (prog == NULL)
    return PFMI_END_OF_LIST;      
  
  /* Copy program information */
  copy_program_info(prog, pi);
  
  /* Move to next */
  cursor->node = nextNode(cursor->node);
  
  return PFMI_OK;
}

static void  
copy_load_module_info(LmRoMetaData *from, 
		      struct PfmiLoadModuleInfo *to)
{
  SET_TYPE_DEF(LmRoMetaData, from);
  
#if !defined (HOST_TOOL_ENVIRONMENT)
  strcpy(to->install_handle, from->install_handle);
  strcpy(to->file_name,      from->parent_file_name);
  strcpy(to->file_format,    from->file_format);
#else
  pfmi_strcpy(to->install_handle, fmi_get_addr("LmRoMetaData", (uint32_t)from,
                                              "install_handle"));
  pfmi_strcpy(to->file_name,      fmi_get_addr("LmRoMetaData", (uint32_t)from,
                                               "parent_file_name"));
  pfmi_strcpy(to->file_format,    fmi_get_addr("LmRoMetaData", (uint32_t)from,
                                               "file_format"));
#endif
  
  to->entrypoint           = DEREF_VAL(from, entrypoint);
#if !defined (HOST_TOOL_ENVIRONMENT)  
  to->text_base            = from->text.base;
  to->text_size            = from->text.size;
  to->data_base            = from->data.base;
  to->data_size            = from->data.size;
#else
  {
    uint32_t text = fmi_get_addr("LmRoMetaData", (uint32_t)from, "text");
    uint32_t data = fmi_get_addr("LmRoMetaData", (uint32_t)from, "data");
    to->text_base = fmi_read_u32(fmi_get_addr("MemRegion", text, "base"));
    to->text_size = fmi_read_u32(fmi_get_addr("MemRegion", text, "size"));
    to->data_base = fmi_read_u32(fmi_get_addr("MemRegion", data, "base"));
    to->data_size = fmi_read_u32(fmi_get_addr("MemRegion", data, "size"));
  }
#endif  
  to->options              = DEREF_VAL(from, options) & 0xf;
#if !defined (HOST_TOOL_ENVIRONMENT)
  to->no_of_sections       = from->section_info.sections_count;
#else
  {
    uint32_t section_info = fmi_get_addr("LmRoMetaData", (uint32_t)from,
                                         "section_info");
    to->no_of_sections = fmi_read_u32(fmi_get_addr("SectionInfo", section_info,
                                                   "sections_count"));
  }
#endif  
  to->no_of_instances      = pm_no_of_programs(GlobalPmInternalState, 
					       to->install_handle);
  to->no_of_instances     += pm_no_of_libraries(GlobalPmInternalState, 
						to->install_handle);
}


static PfmiStatus
get_load_module_info(const char *install_handle,
		     struct PfmiLoadModuleInfo *lmi)
{
  LmMetaData *lm;
  char name_buf[PM_INSTALL_HANDLE_LENGTH];

  SET_TYPE_DEF(LmMetaData, lm);
  
  /* Check that PM is initiated */
  if(!is_pm_initialized())
    return PFMI_UNAVAILABLE;  
  
  /* Check that data is not locked by PM process */
  if(!DEREF_VAL(GlobalPmInternalState, data_accessable))
    return PFMI_DATA_LOCKED;  

  /* Copy the install handle to a non-const buffer. */
  strncpy(name_buf, install_handle, sizeof(name_buf));
  name_buf[sizeof(name_buf) - 1] = '\0';
  
  lm = (LmMetaData *)traverseList1(DEREF_PTR(GlobalPmInternalState, loadmodules),
				   pm_search_lm_install_handle, 
				   name_buf);  
  /* Check if found */
  if(NULL == lm)
    return PFMI_EINSTHND;
  
  /* Copy load module information */
  copy_load_module_info(DEREF_PTR(lm, meta), lmi);

  return PFMI_OK;
}


static PfmiStatus
get_next_load_module(struct PfmiLoadModuleCursor *lmc,
		     struct PfmiLoadModuleInfo   *lmi)
{
  LmMetaData *lm;
  struct LoadModuleCursor* cursor = (struct LoadModuleCursor*)lmc;

  SET_TYPE_DEF(LmMetaData, lm);
  
  lm = (LmMetaData *) cursor->node; 

  /* Check that PM is initiated */
  if(!is_pm_initialized())
    return PFMI_UNAVAILABLE;  
  
  /* Check that data is not locked by PM process */
  if(!DEREF_VAL(GlobalPmInternalState, data_accessable))
    return PFMI_DATA_LOCKED;  
  
  /* Check if end of list */
  if (lm == NULL)
    return PFMI_END_OF_LIST;      
  
  /* Copy load module information */
  copy_load_module_info(DEREF_PTR(lm, meta), lmi);
  
  /* Move to next */
  cursor->node = nextNode(cursor->node);
  
  return PFMI_OK;
}


static PfmiStatus
get_load_module_section_info(const char *install_handle,
			     uint32_t section,
			     struct PfmiLoadModuleSectionInfo *lmsi)
{
  LmMetaData *lm;
  LmRoMetaData *meta;
  Section_Info *si;
  char name_buf[PM_INSTALL_HANDLE_LENGTH];

  SET_TYPE_DEF(LmMetaData, lm);
  SET_TYPE_DEF(Section_Info, si);
  
  /* Check that PM is initiated */
  if(!is_pm_initialized())
    return PFMI_UNAVAILABLE;  
  
  /* Check that data is not locked by PM process */
  if(!DEREF_VAL(GlobalPmInternalState, data_accessable))
    return PFMI_DATA_LOCKED;  

  /* Copy the install handle to a non-const buffer. */
  strncpy(name_buf, install_handle, sizeof(name_buf));
  name_buf[sizeof(name_buf) - 1] = '\0';
  
  lm = (LmMetaData *)traverseList1(DEREF_PTR(GlobalPmInternalState, loadmodules),
				   pm_search_lm_install_handle,
				   name_buf);
  /* Check if found */
  if(NULL == lm)
    return PFMI_EINSTHND;
  
  meta = DEREF_PTR(lm, meta);

  /* Check that section is within range */
#if !defined (HOST_TOOL_ENVIRONMENT)  
  if(section >= meta->section_info.sections_count)
    return PFMI_END_OF_LIST;
  si = &meta->section_info.sections[section];
#else
  {
    uint32_t section_info;
    uint32_t si_size;

    SET_TYPE_DEF(SectionInfo, section_info);
    
    section_info = fmi_get_addr("LmRoMetaData", (uint32_t)meta,
                                         "section_info");
    if (section >= DEREF_VAL(section_info, sections_count))
      return PFMI_END_OF_LIST;

    si_size = fmi_get_size("Section_Info", "#");
    si = (Section_Info *)(DEREF_VAL(section_info, sections) +
                         (si_size * section));
  }
#endif  
  
  /* Copy load modules section information */
  lmsi->section_base       = DEREF_VAL(si, base);
  lmsi->section_size       = DEREF_VAL(si, size);
  lmsi->section_attr       = DEREF_VAL(si, attr);
  lmsi->section_options    = DEREF_VAL(si, options);
#if !defined (HOST_TOOL_ENVIRONMENT)
  strcpy(lmsi->section_name, meta->section_info.sections[section].name);
#else
  pfmi_strcpy(lmsi->section_name, fmi_get_addr("Section_Info",
                                               (uint32_t)si,
                                               "name"));
#endif  
  return PFMI_OK;
}

/*
*****************************************************************************
** PUBLIC FUNCTIONS
*****************************************************************************
*/

PfmiStatus
ose_pfmi_get_progpid
ARGLIST(PROCESS pid,
        PROCESS *progpid)
{  
  uint32_t lock;  
  PfmiStatus status;
  volatile PROCESS tmp_pid = pid;
  
  LOCK_SAVE(lock);
  ENTER_PFMI(i);
  status = get_progpid(tmp_pid, progpid);
  EXIT_PFMI();
  LOCK_RESTORE(lock);    
  
  return status;  
}


PfmiStatus 
ose_pfmi_get_program_info
ARGLIST(PROCESS progpid,
        struct PfmiProgramInfo *pi)
{
  uint32_t lock;  
  PfmiStatus status;
  
  LOCK_SAVE(lock);
  ENTER_PFMI(i);
  status = get_program_info(progpid, pi);
  EXIT_PFMI();
  LOCK_RESTORE(lock);    
  
  return status;  
}


PfmiStatus
ose_pfmi_get_first_program
ARGLIST(struct PfmiProgramCursor *pc,
        struct PfmiProgramInfo   *pi)
{
  uint32_t lock;  
  PfmiStatus status;
  struct ProgramCursor *cursor = (struct ProgramCursor*)pc;
  
  LOCK_SAVE(lock);
  ENTER_PFMI(i);
  cursor->node = firstNode(DEREF_PTR(GlobalPmInternalState, programs));  
  status = get_next_program(pc, pi);
  EXIT_PFMI();
  LOCK_RESTORE(lock);
  
  return status;    
}


PfmiStatus
ose_pfmi_get_next_program
ARGLIST(struct PfmiProgramCursor *pc,
        struct PfmiProgramInfo   *pi)
{
  uint32_t lock;  
  PfmiStatus status;

  LOCK_SAVE(lock);
  ENTER_PFMI(i);
  status = get_next_program(pc, pi);
  EXIT_PFMI();
  LOCK_RESTORE(lock);    
  
  return status;      
}


PfmiStatus
ose_pfmi_get_load_module_info
ARGLIST(const char *install_handle,
        struct PfmiLoadModuleInfo *lmi)
{
  uint32_t lock;
  PfmiStatus status;
  
  LOCK_SAVE(lock);
  ENTER_PFMI(i);
  status = get_load_module_info(install_handle, lmi);
  EXIT_PFMI();
  LOCK_RESTORE(lock);    
  
  return status;  
}


PfmiStatus
ose_pfmi_get_first_load_module
ARGLIST(struct PfmiLoadModuleCursor *lmc,
        struct PfmiLoadModuleInfo *lmi)
{
  uint32_t lock;  
  PfmiStatus status;
  struct LoadModuleCursor *cursor = (struct LoadModuleCursor*)lmc;
  
  LOCK_SAVE(lock);
  ENTER_PFMI(i);
  cursor->node = firstNode(DEREF_PTR(GlobalPmInternalState, loadmodules));
  status = get_next_load_module(lmc, lmi);
  EXIT_PFMI();
  LOCK_RESTORE(lock);
  
  return status;    
}


PfmiStatus
ose_pfmi_get_next_load_module
ARGLIST(struct PfmiLoadModuleCursor *lmc,
        struct PfmiLoadModuleInfo *lmi)
{
  uint32_t lock;  
  PfmiStatus status;
  
  LOCK_SAVE(lock);
  ENTER_PFMI(i);
  status = get_next_load_module(lmc, lmi);
  EXIT_PFMI();
  LOCK_RESTORE(lock);    
  
  return status;      
}


PfmiStatus
ose_pfmi_get_load_module_section_info
ARGLIST(const char *install_handle,
        uint32_t section,
        struct PfmiLoadModuleSectionInfo *lmsi)
{
  uint32_t lock;  
  PfmiStatus status;
  
  LOCK_SAVE(lock);
  ENTER_PFMI(i);
  status =  get_load_module_section_info(install_handle, section, lmsi);
  EXIT_PFMI();
  LOCK_RESTORE(lock);    
  
  return status;      
}


#if 0
PfmiStatus
ose_pfmi_get_load_module_config(char *install_handle,
				char *tag,
				char *data,
				size_t data_size)
{
  uint32_t lock;  
  PfmiStatus status;
  
  LOCK_SAVE(lock);    
  status =  get_load_module_config(install_handle, tag, data, data_size);
  LOCK_RESTORE(lock);    
  
  return status;      
}


PfmiStatus
ose_pfmi_get_load_module_config2(char *install_handle,
				 char *tag,
				 char **data)
{
  uint32_t lock;  
  PfmiStatus status;
  
  LOCK_SAVE(lock);    
  status =  get_load_module_config2(install_handle, tag, data);
  LOCK_RESTORE(lock);    
  
  return status;      
}
#endif

/* ?????????

-What about assert() in called functions???
-Data lock handling? Required if accessed in run mode
-Interrupt locking required? (not in hfmi)

????????? */
