/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2008 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */
/*
 ****************************************************************************
 *			  PM INTERNAL FUNCTIONALITY
 *
 *     $Workfile:   pm_events.c  $
 *
 *
 ****************************************************************************
 * Description: The implementation of the event subscription support in
 *              the Program Manager.
 *
 */

/***********************************************************
 *
 * INCLUDE FILES
 *
 ***********************************************************/

#include "pm_events.h"

/* Program Management block signal interface. */
#include "pm.sig"

/* List handling functionality. */
#include "oldlist.h"

#include "ramlog.h"

/***********************************************************
 *
 * CONSTANTS AND DEFINITIONS
 *
 ***********************************************************/

/* The number of events exported by PM. */   
#define PM_EVENTS_COUNT     1

/* The events exported by PM. */
static const U32 pm_events[PM_EVENTS_COUNT] =
{
  PM_EVENT_UNINSTALL_LOAD_MODULE
};

/***********************************************************
 *
 * STRUCTURES AND GLOBALS
 *
 ***********************************************************/

/***********************************************************
 *
 * FUNCTION DEFINITIONS
 *
 ***********************************************************/

/*
 *===========================================================================
 *                  pm_search_subscriptions_by_event_and_client
 *===========================================================================
 *
 * Description
 * -----------
 * Search for an event subscription (as part of traverseList1).
 * 
 * Parameter description
 * ---------------------
 * node                            The node (in the list) to compare to.
 * req_node                        The node containing the info on 
 *                                 client and event to search for.
 *
 * Returns
 * -------
 * NULL                            The node did not match the event and client.
 *
 * (void *) node                   The node matched the event and client,
 *                                 the pointer to the node is returned.
 *
 */

static void *
pm_search_subscriptions_by_event_and_client(void * node, void * req_node)
{
  if( ( ((EventSubscription *) node)->client == 
        ((EventSubscription *) req_node)->client ) 
      &&
      ( ((EventSubscription *) node)->event ==
        ((EventSubscription *) req_node)->event ) 
    )
  {
    return node;
  }
  else
  {
    return NULL;
  }
}

/*
 *===========================================================================
 *                  pm_search_subscriptions_by_client
 *===========================================================================
 *
 * Description
 * -----------
 * Search for an event subscription (as part of traverseList1).
 * 
 * Parameter description
 * ---------------------
 * node                            The node (in the list) to compare to.
 * client                          The client pid to search for.
 *
 * Returns
 * -------
 * NULL                            The node did not match the client.
 *
 * (void *) node                   The node matched the client,
 *                                 the pointer to the node is returned.
 *
 */

static void *
pm_search_subscriptions_by_client(void * node, void * client)
{
  if(((EventSubscription *) node)->client == (PROCESS)client)
  {
    return node;
  }
  else
  {
    return NULL;
  }
}

/*
 *===========================================================================
 *                        pm_subscribe_event_request
 *===========================================================================
 *
 * Description
 * -----------
 * Handle a PM_SUBSCRIBE_EVENT_REQUEST signal.
 *
 * Parameter description
 * ---------------------
 * PmInternalState                  The PM internal state structure.
 * request                          The PM_SUBSCRIBE_EVENT_REQUEST signal.
 *
 * Returns
 * -------
 * None.
 *
 */

GLOBAL void
pm_subscribe_event_request(struct PmInternalState * PmInternalState, 
                           struct PmSubscribeEventRequest * request)
{
  struct PmSubscribeEventReply * reply;
  EventSubscription * node;
  U32 i;
  union SIGNAL * attach_sig;
  PROCESS client;

  /* Setup reply signal. */
  reply = (struct PmSubscribeEventReply *) 
    alloc(sizeof(struct PmSubscribeEventReply), PM_SUBSCRIBE_EVENT_REPLY);

  /* Need to know client pid at multiple places in this function. */
  client = sender((union SIGNAL **) &request);

  /* Is event requested a valid event? */
  for(i = 0; i < PM_EVENTS_COUNT; i++)
    if(pm_events[i] == request->PmEvent) break;
  if(i == PM_EVENTS_COUNT)
  {
    /* Not a valid event! */
    reply->status = PM_EEVENT_UNKNOWN;
    goto finish_and_return;
  }

  /* Create a new node to possibly insert into list of subscriptions. */
  node = allocNode(sizeof(EventSubscription));
  node->client = client;
  node->event = request->PmEvent;

  /* Has the client already subscribed to the event? */
  {
    EventSubscription * search_node = NULL;
    search_node = (EventSubscription *)
      traverseList1(PmInternalState->event_subscriptions,
                    pm_search_subscriptions_by_event_and_client,
                    (void *)node);
    if(search_node != NULL)
    {
      /* Client has already subscribed to the event! */
      freeNode((void *) node);
      reply->status = PM_EALREADY_SUBSCRIBED;
      goto finish_and_return;
    }
  }

  /* Are we already attached to the client? */
  {
    EventSubscription * list_node = NULL;
    list_node = (EventSubscription *)
      traverseList1(PmInternalState->event_subscriptions,
                    pm_search_subscriptions_by_client,
                    (void *)client);
    if(list_node != NULL)
    {
      /* Yes, so set attach ref. */
      node->attref = list_node->attref;
    }
    else
    {
      /* No, so attach to the client: */
      attach_sig = alloc(sizeof(SIGSELECT), PM_ATTACH_EVENT_SUBSCRIPTION);
      node->attref = attach(&attach_sig, client);
    }
  }

  /* Insert node in list: */
  insertFirst(PmInternalState->event_subscriptions, (void *)node);

  /* Successfully registered the subscription. */
  reply->status = PM_SUCCESS;

 finish_and_return:
  send((union SIGNAL **) &reply, client);
  free_buf((union SIGNAL **) &request);
}

/*
 *===========================================================================
 *                        pm_unsubscribe_event_request
 *===========================================================================
 *
 * Description
 * -----------
 * Handle a PM_UNSUBSCRIBE_EVENT_REQUEST signal.
 *
 * Parameter description
 * ---------------------
 * PmInternalState                  The PM internal state structure.
 * request                          The PM_UNSUBSCRIBE_EVENT_REQUEST signal.
 *
 * Returns
 * -------
 * None.
 *
 */

GLOBAL void
pm_unsubscribe_event_request(struct PmInternalState * PmInternalState, 
                             struct PmUnsubscribeEventRequest * request)
{
  struct PmUnsubscribeEventReply * reply;
  EventSubscription * node = NULL;
  U32 i;
  OSATTREF attref;
  
  /* Setup reply signal. */
  reply = (struct PmUnsubscribeEventReply *) 
    alloc(sizeof(struct PmUnsubscribeEventReply), PM_UNSUBSCRIBE_EVENT_REPLY);

  /* Is event requested a valid event? */
  for(i = 0; i < PM_EVENTS_COUNT; i++)
    if(pm_events[i] == request->PmEvent) break;
  if(i == PM_EVENTS_COUNT)
  {
    /* Not a valid event! */
    reply->status = PM_EEVENT_UNKNOWN;
    goto finish_and_return;
  }
  
  /* Find the client's subscription to this event, if one exists. */
  {
    /* Temporarily allocate a node to carry information into search function */
    EventSubscription * search_node = allocNode(sizeof(EventSubscription));
    search_node->client = sender((union SIGNAL **) &request);
    search_node->event = request->PmEvent;  

    node = (EventSubscription *)
      traverseList1(PmInternalState->event_subscriptions,
                    pm_search_subscriptions_by_event_and_client,
                    (void *)search_node);
    freeNode((void *) search_node);
    if(node == NULL)
    {
      /* Client has not subscribed to the event! */
      reply->status = PM_ENOT_SUBSCRIBED;
      goto finish_and_return;
    }
  }
  
  /* Save for later use. */
  attref = node->attref;

  /* Remove node from list. */
  freeNode((void *) node);

  /* Was this the client's last subscription? */
  node = NULL;
  node = (EventSubscription *)
    traverseList1(PmInternalState->event_subscriptions,
                  pm_search_subscriptions_by_client,
                  (void *) sender((union SIGNAL **) &request));
  if(node == NULL)
  {
    /* Yes, so detach from client. */
    detach(&attref);
  }

  /* Successfully unregistered the subscription. */ 
  reply->status = PM_SUCCESS;
  
 finish_and_return:
  send((union SIGNAL **) &reply, sender((union SIGNAL **) &request));
  free_buf((union SIGNAL **) &request);
}

/*
 *===========================================================================
 *                        pm_event_remove_client
 *===========================================================================
 *
 * Description
 * -----------
 * Handle a PM_ATTACH_EVENT_SUBSCRIPTION signal.
 *
 * Parameter description
 * ---------------------
 * PmInternalState                  The PM internal state structure.
 * request                          The PM_ATTACH_EVENT_SUBSCRIPTION signal.
 *
 * Returns
 * -------
 * None.
 *
 */

GLOBAL void
pm_event_remove_client(struct PmInternalState * PmInternalState, 
                       SIGSELECT * request)
{
  struct Link *current;
  PROCESS client = sender((union SIGNAL **) &request);

  /* Traverse entire list and remove all nodes containing client pid: */

  current = firstNode(PmInternalState->event_subscriptions);
  while (current)
  {
    struct Link *next = nextNode(current);
    if( client == ((EventSubscription *)current)->client)
    {
      /* Remove current node from list. */
      freeNode((void *) current);
    }
    /* Continue searching from next node. */
    current = next;
  }

  free_buf((union SIGNAL **) &request);
}

/*
 *===========================================================================
 *                        pm_event_notify
 *===========================================================================
 *
 * Description
 * -----------
 * Notify all clients of an event occurrence.
 *
 * Parameter description
 * ---------------------
 * PmInternalState                  The PM internal state structure.
 * PmEvent                          The event that occured.
 * extra                            The extra information to pass to clients
 *                                  in PmEventNotify struct.
 * extra_size                       The size of the extra information.
 *
 * Returns
 * -------
 * None.
 *
 */

GLOBAL void
pm_event_notify(struct PmInternalState * PmInternalState, 
                U32 PmEvent, 
                void * extra, 
                size_t extra_size)
{
  struct PmEventNotify * send_sig;
  struct Link *current;
  
  /* Traverse entire list and notify all clients that subscribe to event: */

  current = firstNode(PmInternalState->event_subscriptions);
  while (current)
  {
    struct Link *next = nextNode(current);
    if(PmEvent == ((EventSubscription *)current)->event)
    {
      /* Notify client. */
      send_sig = (struct PmEventNotify *) 
        alloc(sizeof(struct PmEventNotify) 
              + extra_size - 1, 
              PM_EVENT_NOTIFY);
      send_sig->PmEvent = PmEvent;
      memcpy(send_sig->data, extra, extra_size);
      send((union SIGNAL **) &send_sig, 
           ((EventSubscription *)current)->client);
      
    }
    /* Continue searching from next node. */
    current = next;
  }
}

