/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2006 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */
/*
 * Description: The internal structures and definitions general to the PM 
 *              software block.
 * 
 */

#ifndef __PM_INTERNAL_H
#define __PM_INTERNAL_H

/***********************************************************
 *
 * INCLUDE FILES
 *
 ***********************************************************/

#include "oldlist.h"
#include "pm.sig"
#include "rtl.sig"
#include "mmtypes.h"
#include "monitor_cpu.sig"
#include "mm_internal.sig"


/***********************************************************
 *
 * CONSTANTS AND DEFINITIONS
 *
 ***********************************************************/

/*
 * Description: NULL declaration.
 */
#ifndef NULL
#define NULL             0
#endif


/* The internal default number of regions in a program type.
 * Increased to 4 to avoid out of range rvct varning
 * for program_type->regions[PM_HEAP_REGION_TYPE] */
#define PM_DEFAULT_NO_OF_REGIONS      4

/* The name (install handle) of the core module. */
#define PM_CORE_MODULE_NAME          "core_module"

/* Program region names. */
#define PM_REGION_EXTENSION          "Program extension region"
#define PM_REGION_POOL               "Program pool region"

/* FIXME: General Errors, to be moved to a central error 
 * definition file, SE boys... 
 */
#define OSE_EUNKNOWN_SIGNAL             0xfffffffe

/* The segment pid of the system segment. 
 * FIXME: Should be defined by POSE.
 */
#define PM_CORE_SEGPID                 0

/* The segment pool pid. 
 * FIXME: Should be defined by POSE.
 */
#define PM_CORE_POOL_PID                0x10000

/*
 * PM internal bios handle value to indicate a bios open failure.
 */

#define PM_EOPEN_BIOS_FAILURE 0

/*
 * A way to get assert when assert is active, and avoid warnings
 * when asserts are not active.
 */

#ifdef ASSERT
#define ASSERT_WITHOUT_WARNINGS(cond, var) assert(cond)
#else
#define ASSERT_WITHOUT_WARNINGS(cond, var) (void) var
#endif

                                              
                                            
/***********************************************************
 *
 * STRUCTURES AND GLOBALS
 *
 ***********************************************************/

/*
 *===========================================================================
 *                        union SIGNAL
 *===========================================================================
 * 
 * Description
 * -----------
 * The definition of the union SIGNAL structure.
 *
 * Parameter description
 * ---------------------
 * sigNo                 The id of a signal.
 *
 */

union SIGNAL
{
   SIGSELECT sigNo;
   struct MonitorConnectRequest MonitorConnectRequest;
   struct MonitorConnectReply MonitorConnectReply;
   struct MonitorDisconnectRequest MonitorDisconnectRequest;
   struct MonitorSetCPUProcessReportsEnabledRequest MonitorSetCPUProcessReportsEnabledRequest;
   struct MonitorSetCPUProcessReportsEnabledReply MonitorSetCPUProcessReportsEnabledReply;
   struct MonitorGetCPUProcessReportsReply MonitorGetCPUProcessReportsReply;
   struct MonitorGetCPUProcessReportsRequest MonitorGetCPUProcessReportsRequest;
   struct mm_add_phys_resource  add_phys_resource;
   struct mm_remove_phys_resource remove_phys_resource;
};

/*
 *===========================================================================
 *                        struct PmInternalState
 *===========================================================================
 * 
 * Description
 * -----------
 * Definition of the PM internal state parameters.
 *
 * Parameter description
 * ---------------------
 * gmm_pid             The Process ID of the GMM process.
 * gmm_attref          The attach reference to the GMM process.
 * data_accessable     The internal data of PM is accessable in freeze mode
 * pm_bios             The Handle of the PM bios interface.
 * rtl                 The linked list of connected RTLs.
 * programs            The linked list of created programs.
 * installations       The linked list of installed load modules.
 * prog_types          The linked list of program types.
 * event_subscriptions The linked list of event subscriptions.
 *
 */

struct PmInternalState
{
  PROCESS                  gmm_pid;
  OSATTREF              gmm_attref;

  OSBOOLEAN        data_accessable;
  MUTEX                      mutex;
  unsigned long            pm_bios;
  U32               no_of_programs;
   
  struct List                 *rtl;
  struct List            *programs;   
  struct List           *libraries;   
  struct List         *loadmodules;   
  struct List          *prog_types;   
  struct List *event_subscriptions;
};


/*
 *===========================================================================
 *                        ProgramType
 *===========================================================================
 * 
 * Description
 * -----------
 * Definition of the PM program type. A program type consists 
 * of a number of region definitions. Each region configuration 
 * define the memory configuration of the region.
 *
 * Parameter description
 * ---------------------
 * link                  The linked list structure.
 * program_type_name     The name of the program type.
 * super                 Supervisor mode or user mode.
 * no_of_regions         The number of program regions.
 * regions               The program region configurations.
 *
 */

typedef struct
{
  struct Link                link;
  char                       program_type_name[PM_PROGRAM_TYPE_PARAM_LENGTH];
  Boolean                    super;
  U32                        no_of_regions;
  struct PmProgramRegionType regions[PM_DEFAULT_NO_OF_REGIONS];
} ProgramType;


/*
 *===========================================================================
 *                        RtlData
 *===========================================================================
 * 
 * Description
 * -----------
 * Definition of the PM connected RTL information data.
 *
 * Parameter description
 * ---------------------
 * link          The linked list structure.
 * options       The RTL options.
 * pid           The pid of the connected RTL.
 * attref        The attach reference of the connected RTL.
 * name          The file format name supported by the 
 *               connected RTL.
 *
 */

typedef struct
{
  struct Link  link;  
  U32          options;
  PROCESS      pid;      
  OSATTREF     attref;   
  char         name[1]; /* Variable size. */
} RtlData;

/*
 *===========================================================================
 *                        LmMetaData
 *===========================================================================
 * 
 * Description
 * -----------
 * Definition of a load module meta data (in RAM)
 *
 * Parameter description
 * ---------------------
 * link           The linked list structure.
 * meta           Pointer to the load module RO meta data.
 * pers           Pointer to the load module Persistent meta data.
 *
 */

typedef struct
{
   struct Link       link;  
   LmRoMetaData * meta;
   LmRoMetaDataPers * pers;
} LmMetaData;


/*
 *===========================================================================
 *                        Pm_Attach
 *===========================================================================
 * 
 * Description
 * -----------
 * Definition of the list of program attachments.
 *
 * Parameter description
 * ---------------------
 * next           The next Pm_Attach for the program (set to ZERO in the last 
 *                node of the linked list).
 * client         The client (sender of the attach request) pid.
 * sigNo          The signal number requested for the attach notification 
 *                signal.
 * data_len       The length (size) of the client data.
 * data           Client specified data.
 *
 */

typedef struct
{
  void *             next;
  PROCESS          client;
  SIGSELECT         sigNo;
  size_t         data_len;
  char            data[1]; /* Variable size. */
} Pm_Attach;


/*
 *===========================================================================
 *                        ProgramMetaData
 *===========================================================================
 * 
 * Description
 * -----------
 * Definition of the program meta data.
 *
 * Parameter description
 * ---------------------
 * link          The linked list structure.
 * domain        The domain pid of the program.
 * progpid       The program pid of the program.
 * mainbid       The block id of the program main block.
 * mainpid       The process id of the program main process.
 * uid           The user id of the program.
 * pool_base     The base address of the local pool.
 * pool_size     The size of the local pool.
 * lm            The program load module information pointer.
 * attached      The clients attachments to the program information.
 * state         The program state.
 * locked        1 if the program has been locked to a CPU.
 * cleanup       Info about how cleanup shall be performed.
 */

typedef struct
{
  struct Link           link;

  OSDOMAIN            domain;
  PROCESS            progpid;
  PROCESS             segpid;
  PROCESS	     mainbid;
  PROCESS	     mainpid;
  OSUSER                 uid;
  U32                  state;
  U32                 locked;
  U32                cleanup;

  PROCESS         sig_poolid;
  PROCESS         stk_poolid;

  OSADDRESS    stk_pool_base;
  U32	       stk_pool_size;
  OSADDRESS    sig_pool_base;
  U32	       sig_pool_size;

  OSHEAP                heap;

  LmRoMetaData      *     lm;
  Pm_Attach       * attached;
} ProgramMetaData;


/*
 *===========================================================================
 *                        LibraryMetaData
 *===========================================================================
 * 
 * Description
 * -----------
 * Definition of the library meta data.
 *
 * Parameter description
 * ---------------------
 * link          The linked list structure.
 * progpid       The program pid of the program holding the library.
 * lm            The library load module information pointer.
 *
 */

typedef struct
{
  struct Link       link;

  PROCESS        progpid;
  LmRoMetaData      * lm;
} LibraryMetaData;


/***********************************************************
 *
 * FUNCTION DEFINITIONS
 *
 ***********************************************************/

/* FIXME: To be move to string.h. */
size_t
strlcpy(char *dst, const char *src, size_t siz);
size_t
strlcat(char *dst, const char *src, size_t siz);

#endif /* #ifndef __PM_INTERNAL_H */
