/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2001-2007 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */
/*
 ****************************************************************************
 *			  PM INTERNAL FUNCTIONALITY
 *
 *     $Workfile:   pm_load_modules.c  $
 *
 *
 ****************************************************************************
 * Description: The implementation of the load modules support in the
 *              Program Manager.
 *
 */

/***********************************************************
 *
 * INCLUDE FILES
 *
 ***********************************************************/

/* OSE general headers. */
#include "ose.h"
#include "osetypes.h"

#ifndef OUTFMT_NO_SPRINTF
#define OUTFMT_NO_SPRINTF
#include "outfmt.h"
#undef OUTFMT_NO_SPRINTF
#else
#include "outfmt.h"
#endif

/* Standard headers. */
#include "assert.h"
#include "string.h"
#include "stdlib.h"
#include "stdio.h"
#include "malloc.h"

/* Checksum functions. */
#include "checksum.h"

/* List handling functionality. */
#include "oldlist.h"

/* Program Management headers. */
#include "pm.h"
#include "pm_internal.h"

/* Lint functionality .*/
#include "pm_lint.h"

/* Load module configuration handling.*/
#include "ose_conf.h"

/* Program Management block signal interface. */
#include "pm.sig"
#include "rtl.sig"
#include "pm_internal.sig"
#include "pm_err.h"

/* MM interfaces. */
#include "mmtypes.h"
#include "mm.sig"

/* Heap signal interface. */
#include "heap.sig"
#include "ose_heap.h"

/* Bios interface. */
#include "bios.h"

/* The pm function blocks. */
#include "pm_programs.h"
#include "pm_vs_rtl.h"
#include "pm_vs_gmm.h"
#include "pm_conf.h"
#include "pm_load_modules.h"
#include "pm_dynamic_libraries.h"
#include "pm_events.h"

/* The OSE Ramlog utility. */
#include "ramlog.h"

/* Cache handling functions. */
#include "cache.h"

#ifdef OSMIPS       
/* KUSEG offset bits for MIPS. */
#include "cpu_access.h"
#endif

/***********************************************************
 *
 * CONSTANTS AND DEFINITIONS
 *
 ***********************************************************/

/* A orphan region. */
#define  PM_RESURRECTED_ORPHAN_REGION        0
/* A persistent text region. */
#define  PM_RESURRECTED_TEXT_REGION          1
/* A persistent data region. */
#define  PM_RESURRECTED_DATA_REGION          2
/* A static region. */
#define PM_RESURRECTED_STATIC_REGION         3

/* Core module persistency. */
#define PM_CORE_MODULE                       0x8000

/* All known options. */
#define ALL_LOAD_MODULE_OPTIONS \
  (PM_LOAD_MODULE_PERSISTENT\
   | PM_LOAD_MODULE_ABSOLUTE\
   | PM_LOAD_MODULE_PAGED\
   | PM_LOAD_MODULE_KERNEL)

#define PM_WITHIN_REGION(addr, base, size) (addr >= base \
					    && addr < (base + size))

/***********************************************************
 *
 * STRUCTURES AND GLOBALS
 *
 ***********************************************************/

/* Used to detect the Core Module data and bss regions. */
/* NOTE: These variables will never be accessed, only the
 *       addresses of the variables will be used.
 * FIXME: Lint complain about these variables.
 */
static U32 uninitiated_data_location;
static U32 initiated_data_location = 1;

/* OSE Kernel internal configuration struct, used to pick
 * up the system pool_base.
 * FIXME: This extern must go, pose specific.
 */
extern OSADDRESS * zzuu;

/*
 * External reference to core internal function.
 * Retrieves information about the mainpool.
 * Defined in _zzmain.c
 */
extern Boolean ose_has_mainpool(Boolean user_mode);

/***********************************************************
 *
 * FUNCTION DEFINITIONS
 *
 ***********************************************************/

/*
 *===========================================================================
 *                        pm_search_lm_install_handle
 *===========================================================================
 *
 * Description
 * -----------
 * Search for a specific load module using the install handle (as part of
 * traverseList1).
 *
 * Requirement
 * -----------
 * "Install handle".
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * node                             The node (in the list) to compare to.
 * name                             The install handle to search for.
 *
 * Returns
 * -------
 * (void *) 0                       The node did not match the install handle,
 *                                  ZERO is returned.
 * (void *) node                    The node matched the install handle, the
 *                                  pointer to the node is returned.
 *
 *
 */

GLOBAL void *
pm_search_lm_install_handle(void * node, void * install_handle)
{
  LmMetaData * lm = (LmMetaData *) node;

  assert(lm != NULL);
  assert(lm->meta->install_handle != NULL);
  assert(install_handle != NULL);

  if(strcmp(lm->meta->install_handle,
	    (const char *)install_handle) == 0)
    {
      return (void *) node;
    }
  else
    {
      return (void *) 0;
    }
}

/*
 *===========================================================================
 *                        pm_free_load_module_memory
 *===========================================================================
 *
 * Description
 * -----------
 * Free the memory resources associated with a load module.
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * PmInternalState                  The PM internal state structure.
 * install_handle                   The install handle of the load module.
 * meta                             The meta data structure of the load module.
 *
 * Returns
 * -------
 * None.
 *
 */

static PmStatus
pm_free_load_module_memory(struct PmInternalState * state,
			   char * install_handle,
			   LmRoMetaData * meta)
{
  RtlData * rtl;
  PmStatus status;

#ifndef ASSERT
  (void) install_handle;
#endif   
  
  assert(meta != NULL);
  assert(meta->magic[0] == PM_RO_META_DATA_MAGIC_0);
  assert(meta->magic[1] == PM_RO_META_DATA_MAGIC_1);
  assert(meta->magic[2] == PM_RO_META_DATA_MAGIC_2);
  assert(meta->magic[3] == PM_RO_META_DATA_MAGIC_3);
  assert(state != NULL);
  assert(state->rtl != NULL);
  assert(pm_strllen(install_handle,
		    PM_INSTALL_HANDLE_LENGTH) != PM_INSTALL_HANDLE_LENGTH);
  
  rtl = traverseList1(state->rtl,
		      pm_rtl_search_by_ff,
		      meta->file_format);

  if(rtl == NULL)
  {
     return PM_ERTL_LOST;
  }
  
  /* send additional uninstall signal to RTL if needed */
  if((rtl->options & RTL_OPT_UNINSTALL) == RTL_OPT_UNINSTALL)
  {
     struct RtlUninstallLoadModule * uninstall;
     static const SIGSELECT uninstall_reply[] = {1, RTL_UNINSTALL_LOAD_MODULE_REPLY};
     
     uninstall = (struct RtlUninstallLoadModule *)
	alloc(sizeof(struct RtlUninstallLoadModule),
	      RTL_UNINSTALL_LOAD_MODULE_REQUEST);
     uninstall->meta = meta;
     send((union SIGNAL **) &uninstall, rtl->pid);
     uninstall = (struct RtlUninstallLoadModule *)
	receive(uninstall_reply);
     status = uninstall->status;
     free_buf((union SIGNAL **) &uninstall);
  }
  else
  {
    status = PM_SUCCESS;
  }
  {
     PmStatus stat2;

     /* Remove the load module image from memory. */
     
     stat2 = pm_vs_gmm_delete_region(state->gmm_pid, 0,
                                     meta->data.base);
     if(stat2 != PM_SUCCESS)
     {
        for(;;)
        {
           const struct pm_file_line extra = {__FILE__, __LINE__};
           error2(PM_EINTERNAL, (OSERRCODE) &extra);
        } /*lint !e550 */
     }
     
     stat2 = pm_vs_gmm_delete_region(state->gmm_pid, 0,
                                      meta->text.base);
     if(stat2 != PM_SUCCESS)
     {
        for(;;)
        {
           const struct pm_file_line extra = {__FILE__, __LINE__};
           error2(PM_EINTERNAL, (OSERRCODE) &extra);
        } /*lint !e550 */
     }
  }
  return status;
}

/*
 *===========================================================================
 *                        pm_lm_install_handles_request
 *===========================================================================
 *
 * Description
 * -----------
 * Handle a PM_LOAD_MODULES_INSTALL_HANDLES_REQUEST signal. Assemble all
 * install handles and send them to the client.
 *
 * Requirement
 * -----------
 * "load module information".
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * pmInternalState                  The PM internal state structure.
 * request                          The PM_LOAD_MODULE_INSTALL_HANDLES_REQUEST
 *                                  signal.
 *
 * Returns
 * -------
 * None.
 *
 */

GLOBAL void
pm_lm_install_handles_request(struct PmInternalState* pmInternalState,
			      struct PmLoadModuleInstallHandlesRequest *request)
{
  struct PmLoadModuleInstallHandlesReply * reply;
  char * install_handles;
  size_t total_install_handle_size = 0;
  struct List * lm_next;

  assert(pmInternalState != NULL);
  assert(request != NULL);
  assert(pmInternalState->loadmodules !=NULL);

  lm_next = pmInternalState->loadmodules;

  /* Go through the list and count all install_handles combined size. */
  while(lm_next != NULL)
    {

      lm_next = nextNode((void*) lm_next);
      if(NULL != lm_next)
	{
	  size_t size;
	  LmRoMetaData * lm;

	  lm = ((LmMetaData *)lm_next)->meta; /*lint !e740 */
	  size = strlen(lm->install_handle) + (size_t)1;
	  assert(size <= (size_t)PM_INSTALL_HANDLE_LENGTH);
	  total_install_handle_size += size;
	}
    }

  reply = (struct PmLoadModuleInstallHandlesReply *)
    alloc(sizeof(struct PmLoadModuleInstallHandlesReply) +
	  (size_t)total_install_handle_size,
	  PM_LOAD_MODULE_INSTALL_HANDLES_REPLY);

  lm_next = pmInternalState->loadmodules;
  install_handles = reply->install_handles;

  /* Go through the list and copy all install_handles to the reply signal. */
  while(lm_next != NULL)
    {
      lm_next = nextNode((void*) lm_next);
      if(NULL != lm_next)
	{
	  size_t handle_size;
	  LmRoMetaData * lm;

	  lm = (LmRoMetaData *)((LmMetaData *)lm_next)->meta; /*lint !e740 */
	  handle_size = strlen(lm->install_handle) + (size_t)1; /*'\0'*/
	  memcpy(install_handles, lm->install_handle, handle_size);
	  install_handles += handle_size;
	}
    }

  reply->install_handles_size = total_install_handle_size;
  reply->install_handles_no_of_strings = pmInternalState->loadmodules->elements;
  reply->status = PM_SUCCESS;
  send((union SIGNAL **)&reply, sender((union SIGNAL **)&request));
  free_buf((union SIGNAL **)&request);
}


/*
 *===========================================================================
 *                        pm_lm_info_request
 *===========================================================================
 *
 * Description
 * -----------
 * Handle a PM_LOAD_MODULE_INFO_REQUEST signal. Assemble the meta data of the
 * specific load module and send the information back to the client.
 *
 * Requirement
 * -----------
 * "load module information".
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * PmInternalState                  The PM internal state structure.
 * request                          The PM_LOAD_MODULE_INFO_REQUEST signal.
 *
 * Returns
 * -------
 * None.
 *
 */

GLOBAL void
pm_lm_info_request(struct PmInternalState * pmInternalState,
		   struct PmLoadModuleInfoRequest * request)
{
  struct PmLoadModuleInfoReply * reply;
  LmMetaData * lm;
  PmStatus pm_status;

  assert(pmInternalState != NULL);
  assert(request != NULL);
  assert(pmInternalState->loadmodules != NULL);

  if(pm_strllen(request->install_handle,
		PM_INSTALL_HANDLE_LENGTH) == PM_INSTALL_HANDLE_LENGTH)
    {
      lm = NULL;
    }
  else
    {
      lm = (LmMetaData *)traverseList1(pmInternalState->loadmodules,
				       pm_search_lm_install_handle,
				       request->install_handle);
    }

  reply = (struct PmLoadModuleInfoReply *)
    alloc(sizeof(struct PmLoadModuleInfoReply), PM_LOAD_MODULE_INFO_REPLY);

  if(NULL != lm)
    {
      LmRoMetaData * meta = lm->meta;
      size_t str_size;

      str_size = strlcpy(reply->install_handle, meta->install_handle,
			 PM_INSTALL_HANDLE_LENGTH);
      ASSERT_WITHOUT_WARNINGS(str_size < PM_INSTALL_HANDLE_LENGTH, str_size);
      str_size = strlcpy(reply->file_name, meta->parent_file_name,
			 PM_PARENT_FILE_NAME_SIZE);
      ASSERT_WITHOUT_WARNINGS(str_size < PM_PARENT_FILE_NAME_SIZE, str_size);
      str_size = strlcpy(reply->file_format, meta->file_format,
		     PM_FILE_FORMAT_NAME_LENGTH);
      ASSERT_WITHOUT_WARNINGS(str_size < PM_FILE_FORMAT_NAME_LENGTH, str_size);
      reply->entrypoint = meta->entrypoint;
      reply->text_base = meta->text.base;
      reply->text_size = meta->text.size;
      reply->data_base = meta->data.base;
      reply->data_size = meta->data.size;
      reply->options   = (meta->options & 0xf);
      reply->no_of_sections = meta->section_info.sections_count;
      reply->no_of_instances =  pm_no_of_programs(pmInternalState,
						  meta->install_handle);
      reply->no_of_instances += pm_no_of_libraries(pmInternalState,
						   meta->install_handle);

      pm_status = PM_SUCCESS;
    }
  else
    {
      pm_status = PM_EINSTALL_HANDLE;
    }

  reply->status = pm_status;
  (void)strlcpy(reply->install_handle,
		request->install_handle,
		PM_INSTALL_HANDLE_LENGTH);
  send((union SIGNAL **)&reply, sender((union SIGNAL **)&request));
  free_buf((union SIGNAL **) &request);
}


/*
 *===========================================================================
 *                        pm_lm_section_info_request
 *===========================================================================
 *
 * Description
 * -----------
 * Handle a PM_LOAD_MODULE_SECTION_INFO_REQUEST signal. Assemble and reply the
 * specific load module section information to the client.
 *
 * Requirement
 * -----------
 * "load module information".
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * PmInternalState                  The PM internal state structure.
 * request                          The PM_LOAD_MODULE_SECTION_INFO_REQUEST signal.
 *
 * Returns
 * -------
 * None.
 *
 */

GLOBAL void
pm_lm_section_info_request(struct PmInternalState * pmInternalState,
			   struct PmLoadModuleSectionInfoRequest *request)
{
  struct PmLoadModuleSectionInfoReply * reply;
  LmMetaData * lm;
  PmStatus pm_status = PM_SUCCESS;
  U32 interval_length;

  assert(pmInternalState != NULL);
  assert(request != NULL);
  assert(pmInternalState->loadmodules != NULL);

  if(pm_strllen(request->install_handle,
		PM_INSTALL_HANDLE_LENGTH) == PM_INSTALL_HANDLE_LENGTH)
    {
      pm_status = PM_EINSTALL_HANDLE;
      lm = NULL;
    }
  else
    {
      lm = (LmMetaData *)traverseList1(pmInternalState->loadmodules,
				       pm_search_lm_install_handle,
				       request->install_handle);
    }
  if (NULL == lm)
    {
      pm_status = PM_EINSTALL_HANDLE;
    }

  if(request->section_interval_end >= request->section_interval_start)
    {
      interval_length = request->section_interval_end
	- request->section_interval_start + 1;
    }
  else
    {
      pm_status = pm_status == PM_SUCCESS ? PM_ESECTION_INTERVAL : pm_status;
      interval_length = 0;
    }

  reply = (struct PmLoadModuleSectionInfoReply *)
    alloc(sizeof(struct PmLoadModuleSectionInfoReply)
	  + (sizeof(struct PmLoadModuleSectionInfo) * interval_length),
	  (SIGSELECT)PM_LOAD_MODULE_SECTION_INFO_REPLY);

  if(pm_status == PM_SUCCESS)
    {
      U32 i;
      LmRoMetaData * meta;

      meta = lm->meta;

      if((request->section_interval_start < meta->section_info.sections_count)
	 &&
	 (request->section_interval_end < meta->section_info.sections_count))
	{
	  U32 i_start = request->section_interval_start;
	  U32 i_end = request->section_interval_end;
	  for(i = i_start; i <= i_end; i++)
	    {
	      size_t str_size;
	      reply->sections[i-i_start].section_base
		= meta->section_info.sections[i].base;
	      reply->sections[i-i_start].section_size
		= meta->section_info.sections[i].size;
	      reply->sections[i-i_start].section_attr
		= meta->section_info.sections[i].attr;
	      reply->sections[i-i_start].section_options
		= meta->section_info.sections[i].options;
	      str_size = strlcpy(reply->sections[i-i_start].section_name,
				 meta->section_info.sections[i].name,
				 PM_SECTION_NAME_LENGTH);
	      ASSERT_WITHOUT_WARNINGS(str_size < PM_SECTION_NAME_LENGTH, str_size);
	    }
	}
      else
	{
	  pm_status = PM_ESECTION_INTERVAL;
	}
    }

  reply->status = pm_status;
  reply->section_interval_start = request->section_interval_start;
  reply->section_interval_end = request->section_interval_end;
  (void)strlcpy(reply->install_handle,
		request->install_handle,
		PM_INSTALL_HANDLE_LENGTH);

  send((union SIGNAL **)&reply, sender((union SIGNAL **)&request));
  free_buf((union SIGNAL **) &request);
}


/*
 *===========================================================================
 *                        pm_lm_configuration_request
 *===========================================================================
 *
 * Description
 * -----------
 * Handle a PM_LOAD_MODULE_CONFIGURATION_REQUEST signal. Assemble the load
 * modules configuration and send it to the client.
 *
 * Requirement
 * -----------
 * "load module information".
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * PmInternalState                  The PM internal state structure.
 * request                          The PM_LOAD_MODULE_CONFIGURATION_REQUEST signal.
 *
 * Returns
 * -------
 * None.
 *
 */

GLOBAL void
pm_lm_configuration_request(struct PmInternalState * pmInternalState,
			    struct  PmLoadModuleConfigurationRequest *request)
{
  struct  PmLoadModuleConfigurationReply *reply;
  LmMetaData * lm;
  PmStatus pm_status = PM_SUCCESS;

  assert(pmInternalState != NULL);
  assert(request != NULL);
  assert(pmInternalState->loadmodules !=NULL);

  if(pm_strllen(request->install_handle,
		PM_INSTALL_HANDLE_LENGTH) == PM_INSTALL_HANDLE_LENGTH)
    {
      pm_status = PM_EINSTALL_HANDLE;
    }

  if(pm_status == PM_SUCCESS &&
     (NULL != (lm = (LmMetaData *)traverseList1(pmInternalState->loadmodules,
					       pm_search_lm_install_handle,
						request->install_handle))))
    {
      U32 string_size;
      const LmRoMetaData * meta = lm->meta;
      const char * lm_conf = NULL;
      const char * magic = meta->configuration;

      U32 conf_no_of_strings = 0;

      /*This sanity check will be replaced with pm_check_conf()
	in tha future...*/
      /*Sanity check.*/
      if(0 == strncmp(CF_CONF_MAGIC, magic, strlen(CF_CONF_MAGIC)))
	{
	  lm_conf = magic + strlen(CF_CONF_MAGIC); /*Skip the CF_CONF_MAGIC.*/
	  string_size = strlen(lm_conf);
	}
      else
	{
	  string_size = 0;
	  /*OSEHq9624. Give more percise error codes at configuration errors.*/
	  pm_status = PM_ENO_MAGIC;
	}

      reply = (struct  PmLoadModuleConfigurationReply *)
	alloc(sizeof(struct PmLoadModuleConfigurationReply) +
	      (size_t)string_size,
	      PM_LOAD_MODULE_CONFIGURATION_REPLY);

      if(string_size > 0)
	{
	  char * terminator;
	  const char * conf_string;
	  char * reply_conf_buf;

	  reply_conf_buf = reply->conf;
	  conf_string = lm_conf;
	  terminator  = strchr(conf_string, '\n');

	  while(terminator != NULL)
	    {
	      U32 string_length;

	      if(terminator > conf_string)
		{
		  string_length = (U32)(terminator - conf_string);
		  (void)strlcpy(reply_conf_buf, conf_string, string_length + 1);
		  conf_string += string_length + 1;
		  reply_conf_buf += string_length + 1;

		  conf_no_of_strings++;
		}
	      else /*terminator  == conf_string, e.g. a dummy '\n' go past it.*/
	      {
		 conf_string++;
	      }

	      terminator  = strchr(conf_string, '\n');
	    }

	  /* Don't forget the last configuration... */
	  if(strlen(conf_string) > 0)
	    {
	      (void)strcpy(reply_conf_buf, conf_string);
	      conf_no_of_strings++;
	    }
	}
      reply->conf_no_of_strings = conf_no_of_strings;
      reply->conf_size = string_size;
    }
  else
    {
      reply = (struct  PmLoadModuleConfigurationReply *)
	alloc(sizeof(struct PmLoadModuleConfigurationReply),
	      PM_LOAD_MODULE_CONFIGURATION_REPLY);
      reply->conf_size = 0;
      reply->conf_no_of_strings = 0;
      pm_status = PM_EINSTALL_HANDLE;
    }

  reply->status = pm_status;
  (void)strlcpy(reply->install_handle, request->install_handle,
		PM_INSTALL_HANDLE_LENGTH);

  send((union SIGNAL **)&reply, sender((union SIGNAL **)&request));
  free_buf((union SIGNAL **)&request);
}

/*
 *===========================================================================
 *                        pm_get_core_mainpid
 *===========================================================================
 *
 * Description
 * -----------
 * Search and return the mainpid of the core.
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * None.
 *
 * Returns
 * -------
 * The mainpid.
 *
 */

static PROCESS
pm_get_core_mainpid(void)
{
  /* Note: This solution is not acceptable, but the OSE Kernel do not have
   * a method of get the main pid. The pid values are dependant of
   * the current implementation of the OSE Kernel (pose.c). Once the
   * implementation changes hell breaks loose.
   */
  PROCESS pid = current_process(); /* Default to some pid */
  OSBUFSIZE i;
  struct OS_pid_list *plist = get_pid_list(get_bid(current_process()));
  
  /* Assume that the main process has the lowest pid in the block */
  if (plist != NULL)
    {
      for (i=0; i<plist->count; i++)
	{
	  if (plist->list[i] < pid)
	    {
	      pid = plist->list[i];
	    }
	}
      free_buf((union SIGNAL **)&plist);
    }
  return pid;
}

/*
 *===========================================================================
 *                        pm_resurrect_load_modules
 *===========================================================================
 *
 * Description
 * -----------
 * Reinstall load modules after cold/warm reset.
 *
 * Requirement
 * -----------
 * "Restore load modules".
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * PmInternalState                  The PM internal state structure.
 *
 * Returns
 * -------
 * None.
 *
 */

GLOBAL void
pm_resurrect_load_modules(struct PmInternalState * PmInternalState)
{
  struct mm_list_regions_reply * list_reply;
  int i;
  U32 no_of_static;
  LmMetaData * lm;
  LmRoMetaData * lm_core;
  LmRoMetaDataPers * lm_core_pers;
  U32 other_count = 0;
  ProgramMetaData * core_module;
  OSADDRESS text_address;
  OSADDRESS uninitiated_data_address;
  OSADDRESS initiated_data_address;

  assert(PmInternalState != NULL);

  if(pm_vs_gmm_list_regions(PmInternalState->gmm_pid,
			    MM_KERNEL_DOMAIN,
			    &list_reply) != MM_SUCCESS)
    {
      free_buf((union SIGNAL **) &list_reply);
      error(PM_EPERSISTENCY_FAILURE);
      return;
    }

  for(i=0; i < (int)(list_reply->count); i++)
    {
      struct mm_create_region   * get_reply;

      (void)pm_vs_gmm_get_region_info(PmInternalState->gmm_pid,
				MM_KERNEL_DOMAIN,
				(OSADDRESS)list_reply->regions[i],
				&get_reply);

      list_reply->regions[i] = (OSADDRESS) get_reply;

      if(get_reply->status != MM_SUCCESS)
	{
	  int j;
	  for(j=0; j<i+1; j++)
	    {
	      union SIGNAL * free_me = (union SIGNAL *)list_reply->regions[j];
	      free_buf(&free_me);
	    }
	  free_buf((union SIGNAL **) &list_reply);
	  error(PM_EPERSISTENCY_FAILURE);
	  return;
	}
    }

  /* Categorize the regions. */
  no_of_static = 0;
  for(i=0; i<(int)(list_reply->count); i++)
    {
      struct mm_create_region * this_region = (void *)list_reply->regions[i];

      if (0 == strcmp(this_region->name, LM_TEXT))
	{
	  /* A persistent text region. */
	  this_region->status = PM_RESURRECTED_TEXT_REGION;
	}
      else if (0 != strcmp(this_region->name, LM_DATA))
	{
	  /* A static region. */
	  /* FIXME: MM boys will make it possible to detect static regions,
	   *        for now this has to do.
	   */
	  this_region->status = PM_RESURRECTED_STATIC_REGION;
	  no_of_static++;
	}
      else
	{
	  this_region->status = PM_RESURRECTED_ORPHAN_REGION;
	}
    }

  /* Create the System Core Module and persitent load modules. */
  lm_core = heap_alloc_private(sizeof(LmRoMetaData) +
			       (no_of_static * sizeof(MemRegion)) +
			       sizeof(LmRoMetaDataPers),
			       __FILE__, __LINE__);

  memset(lm_core, 0,
	 sizeof(LmRoMetaData) +
	 (no_of_static * sizeof(MemRegion)) +
	 sizeof(LmRoMetaDataPers));

  lm_core->magic[0] = PM_RO_META_DATA_MAGIC_0;
  lm_core->magic[1] = PM_RO_META_DATA_MAGIC_1;
  lm_core->magic[2] = PM_RO_META_DATA_MAGIC_2;
  lm_core->magic[3] = PM_RO_META_DATA_MAGIC_3;

  lm_core_pers = (LmRoMetaDataPers *)(((char *)lm_core) + sizeof(LmRoMetaData)
				     + (no_of_static * sizeof(MemRegion)));
  lm_core->options = PM_CORE_MODULE | PM_LOAD_MODULE_ABSOLUTE;

  /* The entrypoint of the CORE is not retievable without making the
   * PM crt0 dependant, therefore it is set to 0. */
  lm_core->entrypoint = 0;

  /* NOTE: The sections of the core module are not possible to pick up. */
  lm_core->section_info.sections_count = 0;
  lm_core->section_info.sections = NULL;

  /* NOTE: The parent file name and the file format of the core module
   *       can not be retrieved, are using the install_handle to
   *       replace them.
   */
  strcpy(lm_core->parent_file_name, PM_CORE_MODULE_NAME);
  strcpy(lm_core->file_format, PM_CORE_MODULE_NAME);

  strcpy(lm_core->install_handle, PM_CORE_MODULE_NAME);

  /* FIXME: No embedded file support in core modules. */
  lm_core->emfiles = 0;

  /* FIXME: Theses lines shall be removed. */
  lm_core->data.base = 0;
  lm_core->data.size = 0;

  /* Addresses for text and data segments in core_module. */
  text_address = (OSADDRESS)pm_resurrect_load_modules;
  uninitiated_data_address = (OSADDRESS)&uninitiated_data_location;
  initiated_data_address = (OSADDRESS)&initiated_data_location;

#ifdef OSMIPS
  /* For MIPS the KUSEG offset bits have to be masked before address comparisons. */
  text_address &= ~KSEG_MASK;
  uninitiated_data_address &= ~KSEG_MASK; 
  initiated_data_address &= ~KSEG_MASK;
#endif

  for(i=0; i<(int)(list_reply->count);i++)
    {
      struct mm_create_region * this_region = (void *)list_reply->regions[i];

      /* Catagorize the static regions. */
      if(this_region->status == PM_RESURRECTED_STATIC_REGION)
	{
	  if(PM_WITHIN_REGION(text_address,
			      this_region->base, this_region->size))
	    {
	      lm_core->text.base = this_region->base;
	      lm_core->text.size = this_region->size;
#ifdef OSMIPS
              /* For MIPS the KUSEG offset bits have to be restored before storing. */
	      lm_core->text.base |= (KSEG_MASK & (OSADDRESS)pm_resurrect_load_modules);
#endif
	      lm_core->text_conf.permissions = this_region->permissions;
	      lm_core->text_conf.cache_mode = this_region->cache_mode;
	      strlcpy(lm_core->text_conf.phys_mem,
		      this_region->phys_mem,
		      PM_PROGRAM_TYPE_PARAM_LENGTH);
	      strlcpy(lm_core->text_conf.log_mem,
		      this_region->log_mem,
		      PM_PROGRAM_TYPE_PARAM_LENGTH);
	    }
	  else if(PM_WITHIN_REGION(uninitiated_data_address,
				   this_region->base, this_region->size)
		  &&
		  PM_WITHIN_REGION(initiated_data_address,
				   this_region->base, this_region->size))
	    {
	      lm_core->data.base = this_region->base;
	      lm_core->data.size = this_region->size;
#ifdef OSMIPS
              /* For MIPS the KUSEG offset bits have to be restored before storing. */
	      lm_core->data.base |= (KSEG_MASK & (OSADDRESS)&uninitiated_data_location);
#endif
	      lm_core->data_conf.permissions = this_region->permissions;
	      lm_core->data_conf.cache_mode = this_region->cache_mode;
	      strlcpy(lm_core->data_conf.phys_mem,
		      this_region->phys_mem,
		      PM_PROGRAM_TYPE_PARAM_LENGTH);
	      strlcpy(lm_core->data_conf.log_mem,
		      this_region->log_mem,
		      PM_PROGRAM_TYPE_PARAM_LENGTH);
	    }
	  else if(PM_WITHIN_REGION(uninitiated_data_address,
				   this_region->base, this_region->size))
	    {
	      lm_core->uninitiated_data.base = this_region->base;
	      lm_core->uninitiated_data.size = this_region->size;
#ifdef OSMIPS
              /* For MIPS the KUSEG offset bits have to be restored before storing. */
	      lm_core->uninitiated_data.base |= (KSEG_MASK & (OSADDRESS)&uninitiated_data_location);
#endif
	    }
	  else if(PM_WITHIN_REGION(initiated_data_address,
				   this_region->base, this_region->size))

	    {
	      lm_core->initiated_data.base = this_region->base;
	      lm_core->initiated_data.size = this_region->size;
#ifdef OSMIPS
              /* For MIPS the KUSEG offset bits have to be restored before storing. */
	      lm_core->initiated_data.base |= (KSEG_MASK & (OSADDRESS)&initiated_data_location);
#endif
	    }
	  else if(this_region->size != 0)
	    {
	      /* Other region. */
	      lm_core->others[other_count].base = this_region->base;
	      lm_core->others[other_count].size = this_region->size;
	      other_count++;
	    }
	  }

      /* Resurrect the persistent load modules. */
      else if(this_region->status == PM_RESURRECTED_TEXT_REGION)
      {
	  /* A resurrected text region. */
	  int j;
	  struct mm_create_region * data_region_found = NULL;
	  LmRoMetaDataPers * pers = (LmRoMetaDataPers *)(((char *)this_region->base +
							 this_region->size)
							 - sizeof(LmRoMetaDataPers));

	  /* Check the RO region meta magic. */
	  if(NULL != pers
	     && NULL != pers->meta
	     && (pers->magic[0] == PM_RO_META_DATA_MAGIC_0
		 && pers->magic[1] == PM_RO_META_DATA_MAGIC_1
		 && pers->magic[2] == PM_RO_META_DATA_MAGIC_2
		 && pers->magic[3] == PM_RO_META_DATA_MAGIC_3)
	     && PM_WITHIN_REGION((OSADDRESS)pers->meta, this_region->base,
				 this_region->size)
	     && (pers->meta->magic[0] == PM_RO_META_DATA_MAGIC_0
		 && pers->meta->magic[1] == PM_RO_META_DATA_MAGIC_1
		 && pers->meta->magic[2] == PM_RO_META_DATA_MAGIC_2
		 && pers->meta->magic[3] == PM_RO_META_DATA_MAGIC_3))
	  {
	      if(pers->meta->data.size == 0)
		{
		  /* No data in this load module, proceed to the next one. */
		  continue;
		}

	      /* Search for the corresponding data region. */
	      for(j=0;j<(int)(list_reply->count);j++)
		{
		  struct mm_create_region * data_region =
		    (void *)list_reply->regions[j];

		  if(data_region->base == pers->meta->data.base)
		    {
		      /* The data region was persistent. */
		      data_region_found = data_region;
		      data_region->status = PM_RESURRECTED_DATA_REGION;
		    }
		}

	      /* Recreate missing data regions. */
	      if(data_region_found == NULL)
		{
		  U32 permission;
		  U32 cache;
		  U32 data_size = pers->meta->data.size;
		  OSADDRESS data_base = pers->meta->data.base;
                  /* Note that alignment is decided by the the desired
                   * base address not by this redundant align variable.
                   */
                  U32 align;
                  U32 mapping_type = 0;
		  mm_status_t mm_status;
		  PmStatus pm_status;

		  permission = pers->meta->data_conf.permissions;
		  cache = pers->meta->data_conf.cache_mode;

                  /* Don't resurrect data regions in MAS-mapped memory */
                  mm_status = pm_vs_gmm_get_log_mem_mapping_type(
                     PmInternalState->gmm_pid,
                     pers->meta->data_conf.log_mem,
                     &mapping_type);
                  if (mm_status == MM_SUCCESS && mapping_type != MAS)
                  {
                     mm_status = pm_vs_gmm_create_region(PmInternalState->gmm_pid,
                                                         LM_DATA,
                                                         MM_KERNEL_DOMAIN,
                                                         &data_size,
                                                         &data_base,
                                                         pers->meta->data_conf.phys_mem,
                                                         pers->meta->data_conf.log_mem,
                                                         permission,
                                                         cache,
                                                         &align,
                                                         MM_CREATE_REGION_SPECIFY_LOG
                                                         | MM_CREATE_REGION_LOGICAL
                                                         | MM_CREATE_REGION_SPECIFY_NAME);
                     if(mm_status == MM_EOUT_OF_LOG_MEM ||
                        mm_status == MM_EOPTION_NOT_SUPPORTED)
                     {
                        mm_status = pm_vs_gmm_create_region(PmInternalState->gmm_pid,
                                                            LM_DATA,
                                                            MM_KERNEL_DOMAIN,
                                                            &data_size,
                                                            &data_base,
                                                            pers->meta->data_conf.phys_mem,
                                                            pers->meta->data_conf.log_mem,
                                                            permission,
                                                            cache,
                                                            &align,
                                                            MM_CREATE_REGION_SPECIFY_LOG
                                                            | MM_CREATE_REGION_SPECIFY_NAME);
                     }
                     if(mm_status != MM_SUCCESS)
                     {
                        /* The load module could not be restored, remove it. */
                        pm_status = pm_vs_gmm_delete_region(PmInternalState->gmm_pid,
                                                            MM_KERNEL_DOMAIN,
                                                            pers->meta->text.base);
                        if(pm_status != PM_SUCCESS)
                        {
                           for(;;)
                           {
                              const struct pm_file_line extra = {__FILE__, __LINE__};
                              error2(PM_EINTERNAL, (OSERRCODE) &extra);
                           } /*lint !e550 */
                        }

                        continue;
                     }
                  }
		  assert(data_base == pers->meta->data.base);
		  assert(data_size >= pers->meta->data.size);
		}

	      lm = allocNode(sizeof(LmMetaData));
	      lm->meta = pers->meta;
	      lm->pers = pers;
	      insertLast(PmInternalState->loadmodules, lm);
	  }
      else
	  {
	    /* OSEHq10175*/
	    /* The region named 'LM_TEXT', probably created by the RTL,
	     * is corrupt. This region will be removed further below.
	     */

	    struct mm_create_region * this_region =
               (void *)list_reply->regions[i];
	    this_region->status = PM_RESURRECTED_ORPHAN_REGION;

	  }
       }
    }

  for(i=0; i<(int)(list_reply->count);i++)
    {
      struct mm_create_region * this_region = (void *)list_reply->regions[i];

      /* Catagorize the static regions. */
      if(this_region->status == PM_RESURRECTED_ORPHAN_REGION)
	{
	   PmStatus status;
	   /* Orpand data region. */
	   status = pm_vs_gmm_delete_region(PmInternalState->gmm_pid,
				   MM_KERNEL_DOMAIN,
				   this_region->base);
	   if(status != PM_SUCCESS)
	   {
	      for(;;)
	      {
		 const struct pm_file_line extra = {__FILE__, __LINE__};
		 error2(PM_EINTERNAL, (OSERRCODE) &extra);
	      } /*lint !e550 */
	   }
	   (void)ramlog_printf("PM: Deleting orphan region named "
			       "%s, at address: 0x%x , size: 0x%x \n",
			       this_region->name, this_region->base,
			       this_region->size);

	}
    }

  lm_core->configuration = ose_confd;
  lm_core->others[other_count].base = 0;
  lm_core->others[other_count].size = 0;

  lm_core_pers->checksum = 0;
  lm_core_pers->meta = lm_core;

  lm = allocNode(sizeof(LmMetaData));
  lm->meta = lm_core;
  lm->pers = lm_core_pers;
  insertFirst(PmInternalState->loadmodules, lm);

  /* Create Core program. */
  core_module = allocNode(sizeof(ProgramMetaData));
  core_module->domain = MM_KERNEL_DOMAIN;
  core_module->segpid = 0;

  core_module->mainpid = pm_get_core_mainpid();
  core_module->mainbid = get_bid(core_module->mainpid);
  core_module->sig_poolid = get_sig_poolid(core_module->mainbid);
  core_module->stk_poolid = get_stk_poolid(core_module->mainbid);
  core_module->heap = get_envp(core_module->mainbid, HEAPREF_ENVNAME);

  if (ose_has_mainpool(0) == 0)
    {
    struct OS_poolcb * sys_pool;
    sys_pool = get_poolcb(PM_CORE_POOL_PID);

    /* FIXME: For now, with the current pose implementation this
     * is the easiest way to pickup the pool base of the system pool.
     * This is not a permanent solution.
     */
    core_module->stk_pool_base = *((&zzuu)[4]);
    core_module->stk_pool_size = sys_pool->tot_size;
    core_module->sig_pool_base = *((&zzuu)[4]);
    core_module->sig_pool_size = sys_pool->tot_size;
    free_buf((union SIGNAL **) &sys_pool);
    }
  else
    {
    struct OS_poolcb *sys_pool;
    
    sys_pool = get_poolcb(core_module->sig_poolid);
    core_module->sig_pool_base = sys_pool->values[sys_pool->fragment_info+1];
    core_module->sig_pool_size = sys_pool->values[sys_pool->fragment_info+2];
    free_buf((union SIGNAL**)&sys_pool);
    
    sys_pool = get_poolcb(core_module->stk_poolid);
    core_module->stk_pool_base = sys_pool->values[sys_pool->fragment_info+1];
    core_module->stk_pool_size = sys_pool->values[sys_pool->fragment_info+2];
    free_buf((union SIGNAL**)&sys_pool);
    }

  core_module->progpid = core_module->mainbid;
  core_module->lm = lm_core;
  core_module->attached = NULL;
  core_module->state = PM_PROGRAM_STARTED;
  core_module->uid = 0;

  insertFirst(PmInternalState->programs, core_module);
  PmInternalState->no_of_programs++;
  
  for(i=0; i<(int)(list_reply->count); i++)
    {
      union SIGNAL * free_me = (union SIGNAL *)list_reply->regions[i];
      free_buf(&free_me);
    }
  free_buf((union SIGNAL **) &list_reply);
  return;
}


/*
 *===========================================================================
 *                        pm_install_request
 *===========================================================================
 *
 * Description
 * -----------
 * Handle a PM_INSTALL_LOAD_MODULE_REQUEST signal. Find the RTL
 * handling the file format of the load module and send the request to it.
 *
 * Requirement
 * -----------
 * "Install load module".
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * PmInternalState                  The PM internal state structure.
 * request                          The PM_INSTALL_LOAD_MODULE_REQUEST signal.
 *
 * Returns
 * -------
 * None.
 *
 */

GLOBAL void
pm_install_request(struct PmInternalState * PmInternalState,
		   struct PmInstallLoadModuleRequest * request)
{
  RtlData * rtl_found = NULL;
  PmStatus  signal_ok = PM_SUCCESS;
  struct OS_sig_info sig_info;
  PROCESS pool_id;

  assert(PmInternalState != NULL);
  assert(PmInternalState->rtl != NULL);
  assert(request != NULL);

  pool_id = get_sig_poolid(get_bid(current_process()));
  (void)get_sig_info(&sig_info, pool_id, current_process(),
		     (OSADDRESS)request);

  /* Sanity check the signal. */
  if((request->conf_size == 0 && request->conf_no_of_strings != 0)
     ||
     (request->conf_size != 0 && request->conf_no_of_strings == 0))
    {
      /* Inconsistent configuration parameters. */
      /*OSEHq9624. Give more percise error codes at configuration errors.*/
       signal_ok = PM_EPARAMS_CONF_SIZE_CONF_NO_OF_STRINGS;
    }
  /* Since conf_size includes one of the bytes in the struct
   * PmInstallLoadModuleRequest a compensation is needed by one byte
   * to the conf_size, in case the conf_size is larger than 1.
   */
  else if((request->conf_size > 0 ? request->conf_size : 1) +
	  offsetof(struct PmInstallLoadModuleRequest, conf) > sig_info.sig_size)
    {
      /* The configuration parameter conf_size is corrupt. */
      /*OSEHq9624. Give more percise error codes at configuration errors.*/
      signal_ok = PM_EPARAM_CONF_SIZE;
    }
  else if(pm_strcnt(request->conf, request->conf_size) !=
	  request->conf_no_of_strings)
    {
      /* The configuration parameter conf_no_of_strings is corrupt. */
      /*OSEHq9624. Give more percise error codes at configuration errors.*/
       signal_ok = PM_EPARAM_CONF_NO_OF_STRINGS;
    }
  else if(pm_strllen(request->install_handle, PM_INSTALL_HANDLE_LENGTH) ==
	  PM_INSTALL_HANDLE_LENGTH)
    {
      signal_ok = PM_EINSTALL_HANDLE;
    }
  else if(pm_strllen(request->file_name, PM_PARENT_FILE_NAME_SIZE) ==
	  PM_PARENT_FILE_NAME_SIZE)
    {
      ramlog_printf("%s is an incorrect file name.\n", request->file_name);
      signal_ok = PM_EFILE_NAME_CORRUPT;
    }
  else if(pm_strllen(request->file_format,PM_FILE_FORMAT_NAME_LENGTH) ==
	  PM_FILE_FORMAT_NAME_LENGTH)
    {
      ramlog_printf("%s is not a supported file format.\n", request->file_format);
      signal_ok = PM_EFILE_FORMAT_NOT_SUPPORTED;
    }
  else if(request->options & ~ALL_LOAD_MODULE_OPTIONS)
    {
      signal_ok = PM_EOPTION_UNKNOWN;
    }
  else
    {
      /* Check that the install handle is unique. */
      LmMetaData * lm;

      lm = (LmMetaData *) traverseList1(PmInternalState->loadmodules,
                                        pm_search_lm_install_handle,
                                        request->install_handle);
      if(lm)
        {
          /* The install handle was not unique. */
          signal_ok = PM_EINSTALL_HANDLE_ALREADY_INSTALLED;
        }
      else
        {
          /* All checks have passed OK, find a RTL to handle the file format. */
          rtl_found = traverseList1(PmInternalState->rtl, 
                                    pm_rtl_search_by_ff,
		                    request->file_format);
        }
    }

  if(signal_ok != PM_SUCCESS || rtl_found == NULL)
    {
      /* Send PM_INSTALL_LOAD_MODULE_REPLY signal with status failure. */
      struct PmInstallLoadModuleReply * reply;

      reply = (void *) alloc(sizeof(struct PmInstallLoadModuleReply),
			     PM_INSTALL_LOAD_MODULE_REPLY);

      if(signal_ok == PM_SUCCESS && rtl_found == NULL)
	{
          ramlog_printf("%s is not a supported file format.\n", request->file_format);
	  reply->status = PM_EFILE_FORMAT_NOT_SUPPORTED;
	}
      else
	{
	  reply->status = signal_ok;
	}

      (void)strlcpy(reply->install_handle,
		    request->install_handle,
		    PM_INSTALL_HANDLE_LENGTH);

      send((union SIGNAL **) &reply, sender((union SIGNAL **) &request));
      free_buf((union SIGNAL **) &request);
    }
  else
    {
      assert(rtl_found->pid != (PROCESS)0);

      /* Send the request to the RTL. */
      send_w_s((union SIGNAL **)&request,
	       sender((union SIGNAL **) &request),
	       rtl_found->pid);
    }
}

/*
 *===========================================================================
 *                        pm_uninstall_request
 *===========================================================================
 *
 * Description
 * -----------
 * Handle a PM_UNINSTALL_LOAD_MODULE_REQUEST signal. Remove a load module
 * image.
 *
 * Requirement
 * -----------
 * "Uninstall load module".
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * PmInternalState                  The PM internal state structure.
 * request                          The PM_UNINSTALL_LOAD_MODULE_REQUEST signal.
 *
 * Returns
 * -------
 * None.
 *
 */

GLOBAL void
pm_uninstall_request(struct PmInternalState * PmInternalState,
		      struct PmUninstallLoadModule * request)
{
  LmMetaData * lm;
  LmRoMetaData * lm_meta_data;
  PmStatus pm_status = PM_SUCCESS;

  assert(PmInternalState != NULL);
  assert(request != NULL);

  /* Sanity check the signal. */
  if(pm_strllen(request->install_handle,
		PM_INSTALL_HANDLE_LENGTH) == PM_INSTALL_HANDLE_LENGTH)
    {
      /* The install handle is corrupt, fix it, the PM shall never
       * send corrupted signals to the client.
       */
      request->install_handle[PM_INSTALL_HANDLE_LENGTH - 1] = '\0';
      pm_status = PM_EINSTALL_HANDLE;
      goto pm_reply_uninstall;
    }


  /* Check if instances of this load module are running, if so return
   * error code.
   */
  lm = (LmMetaData *)traverseList1(PmInternalState->programs,
				   pm_search_program_by_install_handle,
				   request->install_handle);
  if(lm)
    {
      /*Program(s) exists, set error code.*/
      pm_status = PM_EUNINSTALL_PROGRAM_EXISTS;
    }
  else
    {
      /*Program(s) do not exist. Find the load module, free the GMM memory
       and delete the load module from the list.*/
      lm = (LmMetaData *)traverseList1(PmInternalState->loadmodules,
				       pm_search_lm_install_handle,
				       request->install_handle);

      if(NULL != lm)
	{
	  lm_meta_data = lm->meta;
	  pm_status = pm_free_load_module_memory(PmInternalState,
                                                 request->install_handle,
                                                 lm_meta_data);
	  freeNode(lm);
     
          /* Event PM_EVENT_UNINSTALL_LOAD_MODULE has now occured, 
           * notify all subscribing clients.
           */
          pm_event_notify(PmInternalState, 
                          PM_EVENT_UNINSTALL_LOAD_MODULE, 
                          (void *)request->install_handle,
                          strlen(request->install_handle) + 1);
	}
      else
	{
	  /*The requested lm to be uninstalled is NOT installed.*/
	  pm_status = PM_EINSTALL_HANDLE;
	}
    }
 pm_reply_uninstall:
  request->sigNo = PM_UNINSTALL_LOAD_MODULE_REPLY;
  request->status = pm_status;
  send((union SIGNAL **) &request, sender((union SIGNAL **) &request));

}


/*
 *===========================================================================
 *                        rtl_commit_load_module_data
 *===========================================================================
 *
 * Description
 * -----------
 * Handle a RTL_COMMIT_LOAD_MODULE_DATA signal. Store the data in the meta
 * data structure and inform the client of the installed load module.
 *
 * Requirement
 * -----------
 * "Install load module".
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * PmInternalState                  The PM internal state structure.
 * request                          The RTL_COMMIT_LOAD_MODULE_DATA signal.
 *
 * Returns
 * -------
 * None.
 *
 */

GLOBAL void
rtl_commit_load_module_data(struct PmInternalState * PmInternalState,
			    struct RtlCommitLoadModuleData * reply)
{
  struct PmInstallLoadModuleReply * client_reply;
  PmStatus status;
  char ih_dummy[1] = {'0'};
  LmMetaData * lm = NULL;
  LmRoMetaData * meta = NULL;
  assert(PmInternalState != NULL);
  assert(reply != NULL);

  status = reply->status;

  if(status == PM_SUCCESS)
    {
      const char * program_type_name_ptr;
      char   program_type_name[PM_PROGRAM_TYPE_PARAM_LENGTH];
      U32    program_type_size;
      char * install_handle;
      const char * config;
      U32 text_permission;
      U32 data_permission;
      LmRoMetaDataPers * pers;
      ProgramType *program_type;

      meta = (LmRoMetaData *)reply->load_module_meta_data;

      /* Sanity check the commit signal. */
      if(meta == 0)
	{
	  /*
           * Since the meta data pointer is ZERO the load module can
           * not be cleaned up.
           */
	  install_handle = ih_dummy;
	  status = PM_ERTL_INTERNAL_ERROR;
	  (void)ramlog_printf("RTL_EINTERNAL_ERROR: No meta data found.\n");
	  goto pm_commit_send_reply;
	}
      else if(meta->magic[0] != PM_RO_META_DATA_MAGIC_0
	      || meta->magic[1] != PM_RO_META_DATA_MAGIC_1
	      || meta->magic[2] != PM_RO_META_DATA_MAGIC_2
	      || meta->magic[3] != PM_RO_META_DATA_MAGIC_3)
	{
	  install_handle = ih_dummy;
	  status = PM_ERTL_INTERNAL_ERROR;
	  (void)ramlog_printf("RTL_EINTERNAL_ERROR: Corrupt meta data magic.\n");
	  goto pm_commit_send_reply;
	}
      else if(pm_strllen(meta->install_handle,
			 PM_INSTALL_HANDLE_LENGTH) == PM_INSTALL_HANDLE_LENGTH)
	{
	  install_handle = ih_dummy;
	  status = PM_ERTL_INTERNAL_ERROR;
	  (void)ramlog_printf("RTL_EINTERNAL_ERROR: Too large install "
			      "handle.\n");
	  goto pm_commit_send_reply;
	}
      else if(pm_strllen(meta->parent_file_name,
			 PM_PARENT_FILE_NAME_SIZE) == PM_PARENT_FILE_NAME_SIZE)
	{
	  install_handle = ih_dummy;
	  status = PM_ERTL_INTERNAL_ERROR;
	  (void)ramlog_printf("RTL_EINTERNAL_ERROR: Too large file name.\n");
	  goto pm_commit_send_reply;
	}
      else if(pm_strllen(meta->file_format,
			 PM_FILE_FORMAT_NAME_LENGTH) == PM_FILE_FORMAT_NAME_LENGTH)
	{
	  install_handle = ih_dummy;
	  status = PM_ERTL_INTERNAL_ERROR;
	  (void)ramlog_printf("RTL_EINTERNAL_ERROR: Too large file format name.\n");
	  goto pm_commit_send_reply;
	}

      install_handle = meta->install_handle;
      config = meta->configuration;

      assert(strcmp(install_handle, reply->install_handle) == 0);

      lm = (LmMetaData *)
	traverseList1(PmInternalState->loadmodules,
		      pm_search_lm_install_handle,
		      install_handle);
      if(lm)
	{
	  /* The install handle was not unique. */
	  status = PM_EINSTALL_HANDLE_ALREADY_INSTALLED;
	  goto pm_commit_clean_up_lm;
	}

      /* Make sure the load module at least has a program type configuration */
      program_type_size = pm_lm_conf_get_value(
         config, "OSE_LM_PROGRAM_TYPE", &program_type_name_ptr);
      if(program_type_size == 0)
	{
	  /* Every load module must have a program type... */
	  /*OSEHq9624. Give more percise error codes at configuration errors.*/
	   status = PM_ENO_PROGRAM_TYPE;
	  goto pm_commit_clean_up_lm;
	}

      strlcpy(program_type_name, program_type_name_ptr,
              program_type_size + 1 < PM_PROGRAM_TYPE_PARAM_LENGTH ?
              program_type_size + 1 : PM_PROGRAM_TYPE_PARAM_LENGTH);

      /* Determine the program type. */
      program_type = (ProgramType *)traverseList1(PmInternalState->prog_types,
                                                  pm_search_prog_type,
                                                  program_type_name);
      if(program_type == NULL)
      {
         /* The program type was not found, can not proceed. */
         status = PM_ENO_PROGRAM_TYPE;
         goto pm_commit_clean_up_lm;
      }

      /* Sanity check the program region types. */
      if(pm_sanity_check_program_region_type(PmInternalState,
                                             program_type, &(meta->data_conf),
					     PM_DATA_REGION_TYPE) != PM_SUCCESS ||
	 pm_sanity_check_program_region_type(PmInternalState,
                                             program_type, &(meta->text_conf),
                                             PM_TEXT_REGION_TYPE) != PM_SUCCESS)
	{
	  /* The region configurations was not correct. */
	  status = PM_ERTL_INTERNAL_ERROR;
	  (void)ramlog_printf("RTL_EINTERNAL_ERROR: Corrupt region configuration.\n");
	  goto pm_commit_clean_up_lm;
	}

      /* Region specific sanity checks. */
      text_permission = meta->text_conf.permissions;
      data_permission = meta->data_conf.permissions;

      if((text_permission & MM_USR_EXECUTE) != MM_USR_EXECUTE
	 || (text_permission & MM_USR_READ) != MM_USR_READ)
	{
	  /* Bad user mode permissions for text segment. */
	  status = PM_ERTL_INTERNAL_ERROR;
	  (void)ramlog_printf("RTL_EINTERNAL_ERROR: Bad user mode permissions in the text region.\n");
	  goto pm_commit_clean_up_lm;
	}
      if((data_permission & MM_USR_READ) != MM_USR_READ)
	{
	  /* Bad user mode permissions for data segment. */
	  status = PM_ERTL_INTERNAL_ERROR;
	  (void)ramlog_printf("RTL_EINTERNAL_ERROR: Bad user mode permissions in the data region.\n");
	  goto pm_commit_clean_up_lm;
	}

      if(meta->text.options == PM_REGION_ALIEN
          || meta->data.options == PM_REGION_ALIEN
          || meta->initiated_data.options == PM_REGION_ALIEN
          || meta->uninitiated_data.options == PM_REGION_ALIEN)
      {
         if(meta->options & PM_LOAD_MODULE_PERSISTENT)
         {
            status = PM_ERTL_INTERNAL_ERROR;
            (void)ramlog_printf("RTL_EINTERNAL_ERROR: Persistent ALIEN load modules are not allowed.\n");
            goto pm_commit_clean_up_lm;
         }
         pers = NULL;
      }
      else
      {
         pers = (LmRoMetaDataPers *)((meta->text.base + meta->text.size)
                                     - sizeof(LmRoMetaDataPers));

         if(pers->magic[0] != PM_RO_META_DATA_MAGIC_0
            || pers->magic[1] != PM_RO_META_DATA_MAGIC_1
            || pers->magic[2] != PM_RO_META_DATA_MAGIC_2
            || pers->magic[3] != PM_RO_META_DATA_MAGIC_3)
         {
            status = PM_ERTL_INTERNAL_ERROR;
            (void)ramlog_printf("RTL_EINTERNAL_ERROR: Corrupt persistent meta data magic.\n");
            goto pm_commit_clean_up_lm;
         }

         /* OSEhq12600. Make sure that pres->meta == meta. */
         if(pers->meta != meta)
         {
            status = PM_ERTL_INTERNAL_ERROR;
            (void)ramlog_printf("RTL_EINTERNAL_ERROR: "
                                "LmRoMetaDataPers->meta != LmRoMetaData.\n");
            goto pm_commit_clean_up_lm;
         }

         if(pm_checksum((OSADDRESS)meta->text.base,
                        (U32)(char *)(&(pers->checksum)) -
                        meta->text.base) != pers->checksum)
         {
            /* Bad checksum in text region. */
            status = PM_ERTL_INTERNAL_ERROR;
            (void)ramlog_printf("RTL_EINTERNAL_ERROR: Bad checksum.\n");
            goto pm_commit_clean_up_lm;
         }
      }


      lm = allocNode(sizeof(LmMetaData));
      lm->meta = meta;
      lm->pers = pers;
      insertLast(PmInternalState->loadmodules, lm);

      status = PM_SUCCESS;

      goto pm_commit_send_reply;
    }
  else
    {
      /* The Install failed early in the RTL, just send a reply. */
      goto pm_commit_send_reply;
    }

 pm_commit_clean_up_lm:

  (void)pm_free_load_module_memory(PmInternalState, reply->install_handle,
				   meta);

 pm_commit_send_reply:
  /* Send a reply to the initiator of the install load module. */
  client_reply = (void *) alloc(sizeof(struct PmInstallLoadModuleReply),
				PM_INSTALL_LOAD_MODULE_REPLY);

  client_reply->status = status;
  (void)strlcpy(client_reply->install_handle,
		reply->install_handle, PM_INSTALL_HANDLE_LENGTH);

  send((union SIGNAL **) &client_reply, sender((union SIGNAL **) &reply));
  free_buf((union SIGNAL **) &reply);
}

/*
 *===========================================================================
 *                        pm_checksum
 *===========================================================================
 *
 * Description
 * -----------
 * Calculate an adler32 checksum.
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * base                             The base of the region to checksum.
 * size                             The sizo of the region to checksum.
 *
 * Returns
 * -------
 * The checksum.
 *
 */

GLOBAL U32
pm_checksum(OSADDRESS base, U32 size)
{
  return ose_adler32((void *)base,
		     size,
		     OSE_ADLER32_INITIAL_SEED);
}

/*
 *===========================================================================
 *                        pm_checksum_unt_flush
 *===========================================================================
 *
 * Description
 * -----------
 * Checksum the region and flush the instruction cache.
 *
 * Requirement
 * -----------
 *
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * PmInternalState                  The PM internal state structure.
 * request                          The RTL_FLUSH_AND_CHECKSUM_REQUEST signal.
 *
 * Returns
 * -------
 * None.
 *
 */

GLOBAL void
pm_checksum_unt_flush(struct PmInternalState * pmInternalState,
		      struct RtlFlushAndChecksum *request)
{
#ifndef ASSERT
   (void) pmInternalState;
#endif   
  assert(pmInternalState != NULL);
  assert(request != NULL);

  cache_sync(request->base, request->size);

  request->checksum = pm_checksum(request->base, request->size);
  request->sigNo = RTL_FLUSH_AND_CHECKSUM_REPLY;
  request->status = PM_SUCCESS;
  send((union SIGNAL **)&request, sender((union SIGNAL **)&request));
}

/*
 *===========================================================================
 *                        pm_load_module_checksum_request
 *===========================================================================
 *
 * Description
 * -----------
 * Checksum the region and compare the checksum to the expected value.
 *
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * PmInternalState                  The PM internal state structure.
 * request                          The PM_LOAD_MODULE_CHECKSUM_REQUEST signal.
 *
 * Returns
 * -------
 * None.
 *
 */

GLOBAL void
pm_load_module_checksum_request(struct PmInternalState * pmInternalState,
				struct PmLoadModuleChecksum *request)
{
  PmStatus pm_status;
  U32 checksum = 0;

  assert(pmInternalState != NULL);
  assert(request != NULL);

  if(pm_strllen(request->install_handle,
		PM_INSTALL_HANDLE_LENGTH) == PM_INSTALL_HANDLE_LENGTH)
  {
     pm_status = PM_EINSTALL_HANDLE;
  }
  else
  {
     LmMetaData * lm = (LmMetaData *)traverseList1(pmInternalState->loadmodules,
                                                   pm_search_lm_install_handle,
                                                   request->install_handle);

     if(lm == NULL)
     {
        pm_status = PM_EINSTALL_HANDLE;
     }
     else if(lm->pers == NULL)
     {
        pm_status = PM_ENO_CHECKSUM;
     }
     else if(lm->meta->options & PM_CORE_MODULE)
     {
        checksum = pm_checksum(lm->meta->text.base,
                               lm->meta->text.size);
        if (lm->pers->checksum == 0)
        {
           lm->pers->checksum = checksum;
           pm_status = PM_SUCCESS;
        }
        else if (lm->pers->checksum == checksum)
        {
           pm_status = PM_SUCCESS;
        }
        else
        {
           pm_status = PM_EBAD_CHECKSUM;
        }
     }
     else if(lm->pers->checksum == pm_checksum(lm->meta->text.base,
                                               (U32)&lm->pers->checksum - (U32)lm->meta->text.base))
     {
        checksum = lm->pers->checksum;
        pm_status = PM_SUCCESS;
     }
     else
     {
        pm_status = PM_EBAD_CHECKSUM;
     }
  }

  request->checksum = checksum;
  request->status = pm_status;
  request->sigNo = PM_LOAD_MODULE_CHECKSUM_REPLY;
  send((union SIGNAL **)&request, sender((union SIGNAL **)&request));
}
