/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2001-2013 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */
/*
 ****************************************************************************
 *			  PM INTERNAL FUNCTIONALITY
 *
 *     $Workfile:   pm_master.c  $
 *
 *
 ****************************************************************************
 * Description: The implementation of the master (basic) Program Manager. 
 * 
 */

/***********************************************************
 *
 * INCLUDE FILES
 *
 ***********************************************************/

/* OSE general headers. */
#include "ose.h"
#include "osetypes.h"

/* Bios functionality header. */
#include "bios.h"

/* Standard headers. */
#include "assert.h"
#include "string.h"
#include "stdlib.h"
#include "stdio.h"
#include "malloc.h"

/* File system headers. */
#include "fm.sig"
#include "fss.h"
#include "fss.sig"
#include "efs.h"

/* List handling functionality. */
#include "oldlist.h"

/* Program Management headers. */
#include "pm.h"
#include "pm_internal.h"
#include "pm_err.h"

/* Program Management block signal interface. */
#include "pm.sig"
#include "rtl.sig"
#include "pm_internal.sig"

/* Program Management trap interface. */
#include "pm_trap.h"

/* Lint functionality. */
#include "pm_lint.h"

/* PM function blocks. */
#include "pm_programs.h"
#include "pm_program_types.h"
#include "pm_load_modules.h"
#include "pm_vs_rtl.h"
#include "pm_vs_gmm.h"
#include "pm_dynamic_libraries.h"
#include "pm_measurements.h"
#include "pm_events.h"

/***********************************************************
 *
 * CONSTANTS AND DEFINITIONS
 *
 ***********************************************************/

/* The maximum what string of the PM */
#define PM_MAX_WHATSTRING     31

/* The number of signals in the PM interface. */   
#define PM_INTERFACE_SIZE     29

/* The signals supported in the pm.sig interface. */
static const SIGSELECT pm_sigs[PM_INTERFACE_SIZE+1] =
{
  PM_INTERFACE_SIZE,
  PM_INTERFACE_REQUEST,
  PM_ADD_PROGRAM_TYPE_REQUEST,
  PM_PROGRAM_TYPES_REQUEST,
  PM_PROGRAM_TYPE_INFO_REQUEST,
  PM_FILE_FORMATS_REQUEST,
  PM_INSTALL_LOAD_MODULE_REQUEST,
  PM_UNINSTALL_LOAD_MODULE_REQUEST,
  PM_LOAD_MODULE_INSTALL_HANDLES_REQUEST,
  PM_LOAD_MODULE_INFO_REQUEST,
  PM_LOAD_MODULE_SECTION_INFO_REQUEST,
  PM_LOAD_MODULE_CONFIGURATION_REQUEST,
  PM_CREATE_PROGRAM_REQUEST,
  PM_START_PROGRAM_REQUEST,
  PM_KILL_PROGRAM_REQUEST,
  PM_ATTACH_PROGRAM_REQUEST,
  PM_EXTEND_PROGRAM_MEMORY_REQUEST,
  PM_PROGRAMS_REQUEST,
  PM_PROGRAM_INFO_REQUEST,
  PM_PROGRAM_CONFIGURATION_REQUEST,
  PM_INSTALL_HANDLE_TO_PROGRAM_PID_REQUEST,
  PM_GET_PROGRAM_PID_REQUEST,
  PM_GET_DOMAIN_NUMBER_REQUEST,
  PM_IMPORT_DYNAMIC_LIBRARY_REQUEST,
  PM_DYNAMIC_LIBRARIES_INFO_REQUEST,
  PM_MOVE_PROGRAM_TO_CPU_REQUEST,
  PM_CONFIGURE_LOAD_MEASUREMENT_REQUEST,
  PM_GET_LOAD_MEASUREMENT_REQUEST,
  PM_SUBSCRIBE_EVENT_REQUEST,
  PM_UNSUBSCRIBE_EVENT_REQUEST
};

/***********************************************************
 *
 * STRUCTURES AND GLOBALS
 *
 ***********************************************************/

/* The OSE PM master process. */
void ose_program_mgr(void);

/* The PmInternalState used in the BIOS module. */
struct PmInternalState * GlobalPmInternalState = (void *)0;

/***********************************************************
 *
 * FUNCTION DEFINITIONS
 *
 ***********************************************************/

/*
 *===========================================================================
 *                        pm_initialize_internal_state
 *===========================================================================
 * 
 * Description
 * -----------
 * Initialize the PM internal state structure.
 *
 *
 * Requirement
 * -----------
 * "PM state machine". 
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * None.
 *
 * Returns
 * -------
 * (struct PmInternalState *) state The initialized state.
 *
 */

static 
struct PmInternalState * pm_initialize_internal_state(void)
{
  struct PmInternalState * PmInternalState;
  PmInternalState = (void *)heap_alloc_private(sizeof(struct PmInternalState),
					       __FILE__, __LINE__);

  /* Initial bios handle. */
  PmInternalState->pm_bios = 0;

  PmInternalState->gmm_pid = (PROCESS) 0;
  PmInternalState->gmm_attref = (OSATTREF) 0;

  /* Initialize the list of RTLs. */
  PmInternalState->rtl = allocList();

  /* Initialize the list of Loadmodules. */
  PmInternalState->loadmodules = allocList();

  /* Initialize the list of Libraries. */
  PmInternalState->libraries = allocList();

  /* Initialize the list of programs. */
  PmInternalState->programs = allocList();

  /* Initialize the list of program types. */
  PmInternalState->prog_types = allocList();

  /* Initialize the list of event subscriptions. */
  PmInternalState->event_subscriptions = allocList();

  /* Setup the globally accessible internal state (for BIOS). */
  GlobalPmInternalState = PmInternalState;

  PmInternalState->no_of_programs = 0;

  ose_mutex_init(&(PmInternalState->mutex), 0);
  
  return PmInternalState;
}

/*
 *===========================================================================
 *                        pm_synchronize_to_run_level
 *===========================================================================
 * 
 * Description
 * -----------
 * Synchronize the PM Framework to the current run level.
 *
 * Requirement
 * -----------
 * "PM startup". 
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * PmInternalState                  The PM internal state structure.
 *
 * Returns
 * -------
 * None.
 *
 */

static void
pm_synchronize_to_run_level(struct PmInternalState * PmInternalState)
{
  assert(PmInternalState != NULL);

  /* Synchronize with the Memory Manager Framework. */
  PmInternalState->gmm_attref = pm_attach_gmm(&(PmInternalState->gmm_pid));
}


/*
 *===========================================================================
 *                        pm_trap_interface
 *===========================================================================
 * 
 * Description
 * -----------
 * The PM trap interface.
 *
 *
 * Requirement
 * -----------
 * "PM trap interface". 
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * fc                               The function code.
 * arg1-6                           The function arguments.
 *
 * Returns
 * -------
 * (long) value                     The function return value or -1 if the 
 *                                  fc could not be matched.
 *
 */

static 
long pm_trap_interface(long fc, long arg1, long arg2, 
		       long arg3, long arg4, long arg5, long arg6)
{
  switch (fc)
    {
    case PM_TRAP_INIT_PROGRAM_DATA:
      {
	 pm_init_program_data(GlobalPmInternalState, current_process());
	 return 0;
      }
    case PM_TRAP_GET_PROGID_LIST:
    case PM_TRAP_GET_PROG_INFO:
    default:
      {
	(void) arg1;
	(void) arg2;
	(void) arg3;
	(void) arg4;
	(void) arg5;
	(void) arg6;
	/* Function code not recognized. */
	return -1;
      }
    }
}


/*
 *===========================================================================
 *                        pm_initialize_bios_interface
 *===========================================================================
 * 
 * Description
 * -----------
 * Install the BIOS interface of the PM.
 *
 * Requirement
 * -----------
 * "PM trap interface". 
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * PmInternalState                  The PM internal state structure.
 *
 * Returns
 * -------
 * None.
 *
 */

static void
pm_initialize_bios_interface(struct PmInternalState * PmInternalState)
{
  long status;

  assert(PmInternalState != NULL);
  assert(PmInternalState->pm_bios == 0);

  status = biosInstall(PM_BIOS_NAME, pm_trap_interface, BIOS_USERMODE_ALLOWED);

  if(status != BIOS_SUCCESS)
    {
      /* The install failed. */
      error(PM_EBIOS_INSTALL_FAILED);
      stop(current_process());
    }
  
  PmInternalState->pm_bios = biosOpen(PM_BIOS_NAME);
  assert(PmInternalState->pm_bios != PM_EOPEN_BIOS_FAILURE);
}

/*
 *===========================================================================
 *                        pm_interface_request
 *===========================================================================
 * 
 * Description
 * -----------
 * Handle a PM_INTERFACE_REQUEST signal. Describe the pm.sig interface for 
 * the client.
 *
 * Requirement
 * -----------
 * "PM signal interface". 
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * PmInternalState                  The PM internal state structure.
 * request                          The PM_INTERFACE_REQUEST signal.
 *
 * Returns
 * -------
 * None.
 *
 */

static void
pm_interface_request(struct PmInternalState * PmInternalState,
		     struct PmInterfaceRequest * request)
{
  struct PmInterfaceReply * reply;

  assert(PmInternalState != NULL);
  assert(PmInternalState->pm_bios != (long)NULL);
  assert(request != NULL);

  /* The PM what string, generated from the make files. */
  LIBWHAT

  reply = (struct PmInterfaceReply *) 
    alloc(sizeof(struct PmInterfaceReply)
	  + (sizeof(SIGSELECT) * PM_INTERFACE_SIZE),
	  PM_INTERFACE_REPLY);
  
  reply->status = PM_SUCCESS;

  /* Note: truncation is ignored. */
  (void)strlcpy(reply->whatStr, 
		"OSE Program Manager/VERSION",  
		PM_MAX_WHATSTRING); 

  /* The PM bios inter face handler. */
  reply->biosHandle = PmInternalState->pm_bios;


  /* The copy PM supported signals. */ 
  memcpy(reply->sigs, pm_sigs, sizeof(SIGSELECT) * (PM_INTERFACE_SIZE + 1));

  /* Send the reply. */
  send((union SIGNAL **) &reply, sender((union SIGNAL **)&request));

  /* Free the request buffer, no longer needed. */
  free_buf((union SIGNAL **) &request);	   
}


/*
 *===========================================================================
 *                        ose_program_mgr
 *===========================================================================
 * 
 * Description
 * -----------
 * The main loop of the PGM.
 *
 * Requirement
 * -----------
 * "PGM process". 
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * None.
 *
 * Returns
 * -------
 * None.
 *
 */

OS_PROCESS(ose_program_mgr)
{
  struct PmInternalState * PmInternalState;
  
  assert(sizeof(pm_sigs) == (sizeof(SIGSELECT) * (PM_INTERFACE_SIZE+1)));

  /* Make sure the pm is executing in the system program. */
  if(get_segid(current_process()) != PM_CORE_SEGPID)
    {
      /* The pm process must execute in the core segment. */
      for(;;)
	{
	  error(PM_ENOT_IN_CORE);
	}
    }

  /* Register myself as the one and only Program Manager.
   * This gives the exclusive privilegue of bindind newly created programs
   * to a selected cpu.
   */
  get_sysconf(79, 0 ,0, 0);

  /* Allocate and initialize the PM internal data structure. */
  PmInternalState = pm_initialize_internal_state();

  assert(PmInternalState != NULL);
     
  /* Synchronize the startup of the pm with the 
   * startup of the running system. 
   */
  pm_synchronize_to_run_level(PmInternalState);

  /* Resurrect installed programs. */
  pm_resurrect_load_modules(PmInternalState);

  /* Install the BIOS interface. */
  pm_initialize_bios_interface(PmInternalState);


  /* Start processing requests for operation. */
  for(;;)
    {
      union SIGNAL * request;
      static const SIGSELECT all_requests[] = {0};

      /* Data is accessable from freeze mode interface */
      PmInternalState->data_accessable = TRUE;
      
      request = receive(all_requests);
      
      /* PM uses internal data and is locked for freeze mode */
      PmInternalState->data_accessable = FALSE;

      switch(request->sigNo)
	{
	  /**************************/
	  /* PM interface signals. */
	  /**************************/
 
	  /* Interface/version */

	case PM_INTERFACE_REQUEST:
	  /* Handle versions of the PM public 
	   * signal interface requests. 
	   */
	  pm_interface_request(PmInternalState, 
			       (struct PmInterfaceRequest *)request);
	  break;

	  
	  /* File Formats */
	  
	case PM_FILE_FORMATS_REQUEST:
	  pm_file_formats_request(PmInternalState, 
				  (struct PmFileFormatsRequest *)request);
	  break;
	  
	  /* Program types */
	  
	case PM_ADD_PROGRAM_TYPE_REQUEST:
	  /* Add or modify a program type. */
	  pm_add_program_type_request(PmInternalState, 
				      (struct PmProgramTypeInfo *)request);
	  break;
	  
	case PM_PROGRAM_TYPE_INFO_REQUEST:
	  /* Read a program type. */
	  pm_program_type_info_request(PmInternalState, 
				       (struct PmProgramTypeInfoRequest *)request);
	  break;
	  
	case PM_PROGRAM_TYPES_REQUEST:
	  pm_program_types_request(PmInternalState, 
				   (struct PmProgramTypesRequest *)request);
	  break;
	  
	  /* Load modules */
	  
	case PM_INSTALL_LOAD_MODULE_REQUEST:
	  /* Install a load module (make it resident and in some systems 
	   * multiple instantiable. Results in RTL call.  
	   */
	  pm_install_request(PmInternalState, 
			     (struct PmInstallLoadModuleRequest *)request);
	  break;
	  
	case PM_UNINSTALL_LOAD_MODULE_REQUEST:
	  /* Uninstall a load module, only works if the referenced load module 
	   * is installed and do not have any created programs. Results in RTL call. */
	  pm_uninstall_request(PmInternalState, 
			       (struct PmUninstallLoadModule *)request);
	  break;

	case PM_LOAD_MODULE_INSTALL_HANDLES_REQUEST:
	  pm_lm_install_handles_request(PmInternalState, 
					(struct PmLoadModuleInstallHandlesRequest *)request);
	  break;

	case PM_LOAD_MODULE_INFO_REQUEST:
	  pm_lm_info_request(PmInternalState,
			     (struct PmLoadModuleInfoRequest *)request);
	  break;
	  
	case  PM_LOAD_MODULE_SECTION_INFO_REQUEST:
	  pm_lm_section_info_request(PmInternalState, 
				     (struct PmLoadModuleSectionInfoRequest *)request);
	  break;

	case PM_LOAD_MODULE_CONFIGURATION_REQUEST:
	  pm_lm_configuration_request(PmInternalState, 
				      (struct  PmLoadModuleConfigurationRequest *)request);
	  break;

	case PM_LOAD_MODULE_CHECKSUM_REQUEST:
	  pm_load_module_checksum_request(PmInternalState, 
					  (struct  PmLoadModuleChecksum *)request);
	  break;

	  /* Programs. */
	  
	case PM_CREATE_PROGRAM_REQUEST:
	  /* Create a program program. */
	  pm_create_program_request(PmInternalState, 
				    (struct PmCreateProgramRequest *)request);
	  break;

	case PM_PROGRAM_INFO_REQUEST:
          pm_program_info_request (PmInternalState,
                                   (struct PmProgramInfo *) request);
	  break;

	case PM_PROGRAMS_REQUEST:
          pm_programs_request (PmInternalState,
                               (struct PmProgramsRequest *) request);
	  break;

	case PM_PROGRAM_CONFIGURATION_REQUEST:
	  pm_program_configuration_request(PmInternalState, 
                                           (struct PmProgramConfigurationRequest *)
                                           request);
	  break;

	case PM_GET_DOMAIN_NUMBER_REQUEST:
	  pm_get_domain_number_request(PmInternalState, 
                                       (struct PmGetDomainNumber *) request);
	  break;

	case PM_INSTALL_HANDLE_TO_PROGRAM_PID_REQUEST:
           pm_install_handle_to_program_pid_request (PmInternalState,
                                                     (struct PmInstallHandleToProgramPidRequest *)
                                                     request);
	  break;

	case PM_GET_PROGRAM_PID_REQUEST:
           pm_get_program_pid_request (PmInternalState,
                                       (struct PmGetProgramPid *)
                                       request);
	  break;

	case PM_EXTEND_PROGRAM_MEMORY_REQUEST:
           pm_extend_program_memory_request (PmInternalState,
					     (struct PmExtendProgramMemory *)
					     request);
	  break;

	case PM_ATTACH_PROGRAM_REQUEST:
	  pm_attach_program_request (PmInternalState,
				     (struct PmAttachProgram *)
				     request);	  
	  break;
	  
	case PM_START_PROGRAM_REQUEST:
	  pm_start_program_request(PmInternalState,
				   (struct PmStartProgramRequest *)
				   request);
	  break;

	case PM_KILL_PROGRAM_REQUEST:
	  pm_kill_program_request (PmInternalState,
				   (struct PmKillProgram *)
				   request);
	  break;
	  
	  /* Libraries */

	case PM_IMPORT_DYNAMIC_LIBRARY_REQUEST:
	  pm_import_dyn_lib_request(PmInternalState, 
				    (struct PmImportDynLib *)request);
	  break;
	  
	case PM_DYNAMIC_LIBRARIES_INFO_REQUEST:
	  pm_libraries_info_request(PmInternalState, 
				    (struct PmLibrariesInfoRequest *) request);
	  break;

	  /* SMP */

        case PM_MOVE_PROGRAM_TO_CPU_REQUEST:
	  pm_move_program_to_cpu_request(PmInternalState, 
				    (struct PmMoveProgramToCPU *) request);
	  break;

	  /* Measurement */

        case PM_CONFIGURE_LOAD_MEASUREMENT_REQUEST:
	  pm_configure_load_measurement_request(PmInternalState, 
                                    (struct PmConfigureLoadMeasurement *) request);
	  break;
	
        case PM_GET_LOAD_MEASUREMENT_REQUEST:
	  pm_get_load_measurement_request(PmInternalState, request);
	  break;
	
	  /* Event subscriptions */

        case PM_SUBSCRIBE_EVENT_REQUEST:
          pm_subscribe_event_request(PmInternalState, 
                                     (struct PmSubscribeEventRequest *)request);
          break;

        case PM_UNSUBSCRIBE_EVENT_REQUEST:
          pm_unsubscribe_event_request(PmInternalState, 
                                       (struct PmUnsubscribeEventRequest *)request);
          break;

	  /***********************************/
	  /* PM internal interface signals. */
	  /***********************************/

	case PM_INTERNAL_PROGRAM_KILLED:
	  /* Remove the program information and free its resources when a 
	   * program is killed. 
	   */
	  pm_program_killed(PmInternalState, 
			    (SIGSELECT *)request);
	  break;

	case PM_INTERNAL_RTL_KILLED:
	  /* Remove the rtl information and free its resources when a 
	   * RTL is lost. 
	   */
	  pm_rtl_killed(PmInternalState, 
			(SIGSELECT *)request);
	  break;
	    
	case ATTACH_GMM:
	  /* The GMM process is lost, this should never happen. */
	  error(PM_EGMM_LOST);
	  free_buf(&request);
	  stop(current_process());
	  break;

        case PM_ATTACH_EVENT_SUBSCRIPTION:
          /* A client that subscribes to PM internal event(s) has died. 
           * Remove all of client's subscriptions.
           */
          pm_event_remove_client(PmInternalState, (SIGSELECT *)request);
          break;
          

	  /**************************/
	  /* RTL interface signals. */
	  /**************************/ 

	case RTL_CONNECT_REQUEST:
	  /* Connect RTL to the PM, making it available for operations. */
	  rtl_connect_request(PmInternalState, 
			      (struct RtlConnectRequest *)request);
	  break;

	case RTL_INTERFACE_REQUEST:
	  /* Request for the rtl.sig signal interface version. */ 
	  rtl_interface_request((struct RtlInterfaceRequest *)request);
	  break;
	  
	case RTL_COMMIT_LOAD_MODULE_DATA:
	   /* Commit META data of a newly installed load module. */
	   rtl_commit_load_module_data(PmInternalState, 
				       (struct RtlCommitLoadModuleData *)request);
	  break;
	  
	case RTL_FLUSH_AND_CHECKSUM_REQUEST:
	   pm_checksum_unt_flush(PmInternalState,
				 (struct RtlFlushAndChecksum *) request);
	   break;
	   
	default:
	   error2(OSE_EUNKNOWN_SIGNAL, (OSERRCODE)request);
	   free_buf(&request);	
	}
    }
}
