/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2007-2009 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */
/*
 ****************************************************************************
 *			  PM INTERNAL FUNCTIONALITY
 *
 *     $Workfile:   pm_programs.c  $
 *
 *
 ****************************************************************************
 * Description: The implementation of cpu load measurements in the
 *              Program Manager.
 *
 */

/***********************************************************
 *
 * INCLUDE FILES
 *
 ***********************************************************/

/* Core internal kernel services. */
#include "core_internal.h"

#include "monitor.sig"
#include "monitor_cpu.sig"
#include "ramlog.h"

/***********************************************************
 *
 * DEFINES
 *
 ***********************************************************/
/* The default values and limits for profiling reports for all report types
 * Taken from rmm.c. There is no rmm.h. Hope they never change!!! */
#define MAX_PROF_INTERVAL        0x7FFFFFFF
#define MIN_PROF_INTERVAL        1
#define MIN_PROF_MAXPROCS        1

/***********************************************************
 *
 * STRUCTURES AND GLOBALS
 *
 ***********************************************************/

static PROCESS rmm_pid = 0;


/***********************************************************
 *
 * FUNCTION DEFINITIONS
 *
 ***********************************************************/

static void
disconnect_from_rmm(void)
{
   union SIGNAL *signal;

   /* Disconnect from the RMM. */
   signal = alloc(sizeof(MonitorDisconnectRequest),
                  MONITOR_DISCONNECT_REQUEST);
   send(&signal, rmm_pid);
   rmm_pid = 0;
}


static PmStatus
connect_to_rmm(void)
{
   union SIGNAL *reply, *signal;
   PROCESS ose_monitor_pid = 0;
   static const SIGSELECT selectMonitorReply[] =
       { 1,
         MONITOR_CONNECT_REPLY};

   /* Find the RMM. */
   hunt("ose_monitor", 0, &ose_monitor_pid, 0);
   if (ose_monitor_pid == 0)
   {
      ramlog_printf("Could not find the RMM monitor process.\n");
      return PM_EMEAS_RMM_ERROR;
   }

   /* Register with the RMM. */
   signal = alloc(sizeof(MonitorConnectRequest) + strlen("pm_measurements") + 1,
                     MONITOR_CONNECT_REQUEST);
   strcpy(signal->MonitorConnectRequest.name, "pm_measurements");

   send(&signal, ose_monitor_pid);

   reply = receive(selectMonitorReply);

   if(reply->sigNo != MONITOR_CONNECT_REPLY)
   {
      kill_proc(current_process());
   }
   else if (reply->MonitorConnectReply.status != MONITOR_STATUS_OK)
   {
      ramlog_printf("Could not connect to the RMM.\n");
      return PM_EMEAS_RMM_ERROR;
   }
   else {
      /* success */
      rmm_pid = sender(&reply);
   }
   free_buf(&reply);

   return PM_SUCCESS; /* so far */
}


static PmStatus
configureRMM(struct PmConfigureLoadMeasurement * request, MonitorStatus *status)
{
   union SIGNAL *signal, *reply;
   MonitorSetCPUProcessReportsEnabledRequest *p1;
   static const SIGSELECT selectMonitorReply[] =
   { 2,
     MONITOR_CONNECT_REPLY,
     MONITOR_SET_TOTAL_CPU_BLOCK_REPORTS_ENABLED_REPLY };

   if (request->enable)
   {
      /* Some sanity checks. */
      if (request->sampleInterval < MIN_PROF_INTERVAL ||
          request->sampleInterval > MAX_PROF_INTERVAL)
      {
         ramlog_printf("Illegal sample interval.\n");
         return PM_EMEAS_CONFIG_ERROR;
      }
      if (request->maxProgReports <= MIN_PROF_MAXPROCS ||
          request->maxProgReports > zzkrn_get_pid_info(0, 4) )
      {
         ramlog_printf("Illegal maximum program reports.\n");
         return PM_EMEAS_CONFIG_ERROR;
      }
   }

   /* Enable per block load monitoring. */
   signal = alloc(sizeof(MonitorSetCPUProcessReportsEnabledRequest),
            MONITOR_SET_TOTAL_CPU_BLOCK_REPORTS_ENABLED_REQUEST);
   p1 = &signal->MonitorSetCPUProcessReportsEnabledRequest;
   p1->enabled = request->enable;
   p1->interval = request->sampleInterval;
   /* You need to set this to two, in order to get one report reliably. */
   p1->maxReports = 2;
   p1->maxProcsReport = request->maxProgReports;
   p1->usePriority = True;
   if(sigsize((union SIGNAL **)&request) >= sizeof(struct PmConfigureLoadMeasurement))
   {
      p1->priority = (uint8_t) request->priority;
   }
   else
   {
      p1->priority = 33;
   }
   p1->reserved1 = 0;

   send(&signal, rmm_pid);

   reply = receive(selectMonitorReply);


   if(reply->sigNo != MONITOR_SET_TOTAL_CPU_BLOCK_REPORTS_ENABLED_REPLY)
   {
      kill_proc(current_process());
   }
   else
   {
      *status = reply->MonitorSetCPUProcessReportsEnabledReply.status;
      free_buf(&reply);

      if (*status != MONITOR_STATUS_OK && *status != MONITOR_STATUS_IN_USE)
      {
         ramlog_printf("Could not enable the RMM block load monitoring.\n");
         disconnect_from_rmm();
         return PM_EMEAS_RMM_ERROR;
      }
   }


   return PM_SUCCESS;
}

/* Searches for a PMProgramLoad struct for the supplied program id.
 * If none is found, returns a pointer to a new structure for that id. */
static struct PmProgramLoad *
find_report_struct(struct PmInternalState * PmInternalState,
                   struct PmGetLoadMeasurement *request, U32 id)
{
   unsigned no_scanned = 0;
   struct PmProgramLoad *report = &request->programs[0];

   (void)PmInternalState;
   
   /* Uses a simple linear search. */
   while(report->id != id && no_scanned + 1 <= request->programsCount)
   {
      report = &request->programs[++no_scanned];
   }

   if (report->id != id)
   {
      /* Not found */
      request->programsCount++;
      report = &request->programs[no_scanned];
      report->id = id;
      report->sum = 0;
   }

   return report;
}

/*
 *===========================================================================
 *                        pm_configure_load_measurement_request
 *===========================================================================
 *
 * Description
 * -----------
 * Configures the per program load measurements performed by the PM.
 *
 * Requirement
 * -----------
 *
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * PmInternalState                  The PM internal state structure.
 * request                          The request of the operation.
 *
 * Returns
 * -------
 * None.
 *
 */

GLOBAL void
pm_configure_load_measurement_request(struct PmInternalState * PmInternalState,
                                      struct PmConfigureLoadMeasurement * request)
{
   PmStatus status = PM_SUCCESS;
   MonitorStatus mon_status = MONITOR_STATUS_OK;

   (void)PmInternalState;
   
   if (rmm_pid == 0)
   {
      status = connect_to_rmm();
   }

   if (status == PM_SUCCESS)
   {
      status = configureRMM(request, &mon_status);
      if (mon_status == MONITOR_STATUS_IN_USE) {
         /* OK let's reconfigure */
         /* first disable */
         request->enable = FALSE;
         if (PM_SUCCESS == configureRMM(request, &mon_status))
         {
            /* first enable with our parameters */
            request->enable = TRUE;
            status = configureRMM(request, &mon_status);
         }
      }
   }

   /* make request reply */
   request->sigNo = PM_CONFIGURE_LOAD_MEASUREMENT_REPLY;
   request->status = status;
   send((union SIGNAL **) &request, sender((union SIGNAL **) &request));
}


/*
 *===========================================================================
 *                        pm_get_load_measurement_request
 *===========================================================================
 *
 * Description
 * -----------
 * Generates the per program load measurements.
 *
 * Requirement
 * -----------
 *
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * PmInternalState                  The PM internal state structure.
 * request                          The request of the operation.
 *
 * Returns
 * -------
 * None.
 *
 */

GLOBAL void
pm_get_load_measurement_request(struct PmInternalState * PmInternalState,
                                union SIGNAL * basic_request)
{
   PmStatus status;
   ProgramMetaData *program;
   struct PmGetLoadMeasurement *request =
      (struct PmGetLoadMeasurement *)basic_request;
   union SIGNAL *signal, *reply;
   MonitorGetCPUProcessReportsReply *b;
   MonitorCPUProcess *p;
   PROCESS original_sender = sender(&basic_request);
   uint32_t i;
   /* cpuid_t cpu; */
   static const SIGSELECT selectMonitorReply[] =
      { 1,
        MONITOR_GET_TOTAL_CPU_BLOCK_REPORTS_REPLY };

   /* Check that the RMM has been enabled. */
   if (rmm_pid == 0)
   {
      status = PM_EMEAS_NOT_ENABLED;
      ramlog_printf("rmm_pid was 0.\n");
      goto pm_get_load_measurement_reply;
   }

   /* Free received signal to replace it with another size */
   free_buf(&basic_request);
   /* Allocate the worst case response size. */
   basic_request = alloc(sizeof(struct PmGetLoadMeasurement) +
                         PmInternalState->no_of_programs *
                         sizeof(struct PmProgramLoad),
                         PM_GET_LOAD_MEASUREMENT_REPLY);
   request = (struct PmGetLoadMeasurement *)basic_request;
   bzero(request, sizeof(struct PmGetLoadMeasurement) +
         PmInternalState->no_of_programs *
         sizeof(struct PmProgramLoad));

   /* Measure the load of the blocks on the loaded CPU. */
   signal = alloc(sizeof(MonitorGetCPUProcessReportsRequest),
		  MONITOR_GET_TOTAL_CPU_BLOCK_REPORTS_REQUEST);
   signal->MonitorGetCPUProcessReportsRequest.time.tick = 0;
   signal->MonitorGetCPUProcessReportsRequest.time.ntick = 0;
   signal->MonitorGetCPUProcessReportsRequest.reserved0 = MONITOR_ALL_EXECUTION_UNITS;
   send(&signal, rmm_pid);

   reply = receive(selectMonitorReply);
   b = &reply->MonitorGetCPUProcessReportsReply;

   if (MONITOR_GET_TOTAL_CPU_BLOCK_REPORTS_REPLY == reply->sigNo)
   {
      /* Check that the load info has been enabled. */
      if (b->enabled == 0)
      {
	 ramlog_printf("Load monitoring not enabled.\n");
	 status = PM_EMEAS_NOT_ENABLED;
	 goto pm_get_load_measurement_reply;
      }
   }
   else
   {
      kill_proc(current_process());
   }

   /* Skip all reports except the last one. */
   p = b->processes;
   for (i = 1; i < b->reportsCount; i++)
   {
      MonitorCPUProcessReport *report = (void *)p;

      /* Go to the next report. */
      p = &report->processes[report->processesCount];
   }

   if (b->reportsCount > 0)
   {
      MonitorCPUProcessReport *report = (void *)p;
      MonitorCPUProcess *pn = report->processes + report->processesCount;

      /* For each block */
      for (p = report->processes; p < pn; p++)
      {
	 struct PmProgramLoad *pmLoadReport;

	 program = pm_find_program(PmInternalState, p->id);
	 if (program == NULL)
	 {
	    continue;
	 }

	 /* Aggregate the blocks into programs. */
	 pmLoadReport = find_report_struct(PmInternalState, request,
					   program->progpid);

	 if (pmLoadReport)
	 {
	    pmLoadReport->sum += p->sum;
	    pmLoadReport->locked =
	       zzkrn_segment_migration_status(pmLoadReport->id, 0, 0);
	 }
      }

      request->interval = report->interval;
   }
   else
   {
      request->interval = 0;
   }

   free_buf(&reply);

   status = PM_SUCCESS;

  pm_get_load_measurement_reply:
   request->sigNo = PM_GET_LOAD_MEASUREMENT_REPLY;
   request->status = status;
   send(&basic_request, original_sender);
}
