/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2001-2007 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */
/*
 ****************************************************************************
 *			  PM INTERNAL FUNCTIONALITY
 *
 *     $Workfile:   pm_program_types.c  $
 *
 *
 ****************************************************************************
 * Description: The implementation of the program types support in the 
 *              Program Manager. 
 * 
 */

/***********************************************************
 *
 * INCLUDE FILES
 *
 ***********************************************************/

/* OSE general headers. */
#include "ose.h"
#include "osetypes.h"

/* Outfmt requires a special treatment. */
#ifndef OUTFMT_NO_SPRINTF
#define OUTFMT_NO_SPRINTF
#include "outfmt.h"
#undef OUTFMT_NO_SPRINTF
#else
#include "outfmt.h"
#endif

/* Standard headers. */
#include "assert.h"
#include "string.h"
#include "stdlib.h"
#include "stdio.h"
#include "malloc.h"

/* List handling functionality. */
#include "oldlist.h"
#include "pm_lint.h"

/* Memory manager public interface. */
#include "mm.sig"
#include "mm.h"

/* Program Management headers. */
#include "pm.h"
#include "pm_internal.h"
#include "pm_err.h"

/* Program Management block signal interface. */
#include "pm.sig"

/* The program types interface header. */
#include "pm_program_types_support.h"
#include "pm_program_types.h"
#include "pm_conf.h"

#include "ramlog.h"

/***********************************************************
 *
 * CONSTANTS AND DEFINITIONS
 *
 ***********************************************************/

/* The execution modes available for program types. */
#define PM_EXEC_SUPERVISOR            "SUPERVISOR"

/***********************************************************
 *
 * STRUCTURES AND GLOBALS
 *
 ***********************************************************/

/***********************************************************
 *
 * FUNCTION DEFINITIONS
 *
 ***********************************************************/
  
/*
 *===========================================================================
 *                        pm_search_prog_type
 *===========================================================================
 * 
 * Description
 * -----------
 * Search for a specific program type using the program type name (as part of 
 * traverseList1).
 *
 * Requirement
 * -----------
 * "Program type handling". 
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * node                             The node (in the list) to compare to.
 * name                             The name of a program type to search for.
 *
 * Returns
 * -------
 * (void *) 0                       The node did not match the name, ZERO is 
 *                                  returned.
 * (void *) node                    The node matched the name, the pointer to
 *                                  the node is returned.
 *
 *
 */

GLOBAL void * 
pm_search_prog_type(void * node, void * name)
{
  ProgramType * prog_type;

  assert(node != NULL);
  assert(((ProgramType *)node)->program_type_name != NULL);
  assert(name != NULL);
  
  prog_type = node; 
  if(strcmp(prog_type->program_type_name, (const char *)name) == 0)
    {
      return (void *) node;
    }
  else
    {
      return (void *) 0;
    }
}

/*
 *===========================================================================
 *                        pm_store_program_type
 *===========================================================================
 * 
 * Description
 * -----------
 * Store an added or replaced program type in the list of program types.
 *
 * Requirement
 * -----------
 * "Program type handling". 
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * PmInternalState                  The PM internal state structure.
 * program_type                     The program type to add (stored i a 
 *                                  structure).
 *
 * Returns
 * -------
 * A PmStatus status indication.
 *  PM_SUCCESS                      The operation was successful. 
 *
 */

static PmStatus 
pm_store_program_type(struct PmInternalState * PmInternalState, 
		      ProgramType * program_type)
{
  ProgramType * program;
  U32 i;
  
  assert(PmInternalState != NULL);
  assert(PmInternalState->prog_types != NULL);
  assert(program_type != NULL);
  
  /* Search for the program type. */
  program = (ProgramType *)traverseList1(PmInternalState->prog_types, 
					 pm_search_prog_type, 
					 program_type->program_type_name);

  /* Allocate and initialize a program type buffer if needed. */
  if(program == NULL || program->no_of_regions != program_type->no_of_regions)
    {
      size_t str_size;

      if(program != NULL)
	{
	  freeNode(program);
	}
      
      program = allocNode(sizeof(ProgramType) 
			  +  (sizeof(struct PmProgramRegionType) 
			   * (program_type->no_of_regions 
			      - PM_DEFAULT_NO_OF_REGIONS)));

      program->no_of_regions = program_type->no_of_regions;
      str_size = strlcpy(program->program_type_name, 
			 program_type->program_type_name,
			 PM_PROGRAM_TYPE_PARAM_LENGTH);
      ASSERT_WITHOUT_WARNINGS(str_size < PM_PROGRAM_TYPE_PARAM_LENGTH, str_size);

      insertLast(PmInternalState->prog_types, program);
    }

  program->super = program_type->super;

  for(i=0;i<program_type->no_of_regions;i++)
    {
      size_t str_size;
      str_size = strlcpy(program->regions[i].phys_mem, 
		     program_type->regions[i].phys_mem, 
		     PM_PROGRAM_TYPE_PARAM_LENGTH);
      ASSERT_WITHOUT_WARNINGS(str_size < PM_PROGRAM_TYPE_PARAM_LENGTH, str_size);
      str_size = strlcpy(program->regions[i].log_mem, 
		     program_type->regions[i].log_mem, 
		     PM_PROGRAM_TYPE_PARAM_LENGTH);
      assert(str_size < PM_PROGRAM_TYPE_PARAM_LENGTH);
      program->regions[i].permissions = program_type->regions[i].permissions;
      program->regions[i].cache_mode = program_type->regions[i].cache_mode;
    }

  return PM_SUCCESS;
}



/*
 *===========================================================================
 *                        pm_read_one_conf
 *===========================================================================
 * 
 * Description
 * -----------
 * Read (from configuration) one specific program type configuration.
 *
 * Requirement
 * -----------
 * "Program type handling". 
 *
 * Side effects
 * ------------
 * A getenv cause a hidden free & malloc sequence, once the first getenv for 
 * a process is performed, one heap buffer will always be allocated for the 
 * process for handling getenv.
 *
 * Parameter description
 * ---------------------
 * name                              The name of the configuration.
 * value                             A buffer to store the value in.
 * value_size                        The number of bytes the value buffer can 
 *                                   hold.
 *
 * Returns
 * -------
 * True                              The configuration was read.
 * False                             The configuration was not found, or the 
 *                                   configuration could not be stored.
 *
 */

#if 0
static Boolean
pm_read_one_conf(char * name, char * value, U32 value_size)
{
  char * get_env_value;

  assert(name != NULL);
  assert(value != NULL);
  assert(value_size > 0);

  get_env_value = getenv(name);

  if(get_env_value == NULL || strlen(get_env_value) > (value_size - 1))
    {
      return False;
    }
  else
    {
      strcpy(value, get_env_value);      
      return True;
    }
}
#endif

/*
 *===========================================================================
 *                        pm_sanity_check_program_region_type
 *===========================================================================
 * 
 * Description
 * -----------
 * Sanity check the contents of a program region.
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * PmInternalState               Internal state.
 * program_type                  The program type associated with the region.
 * region                        The region to check.
 * region_type                   Type of region (text, data, heap, etc).
 *
 * Returns
 * -------
 * PM_SUCCESS                    The check was successful.
 * PM_EPROGRAM_TYPE_SYNTAX_ERROR There was a syntax error. 
 * PM_EILLEGAL_REGION_LOG_MEM    Region specified MAS-mapped log_mem for
 *                               supervisor mode program type. 
 */

GLOBAL PmStatus
pm_sanity_check_program_region_type(struct PmInternalState * PmInternalState,
                                    ProgramType *program_type,
                                    struct PmProgramRegionType * region,
				    U32 region_type)
{
  U32 permission_binary;
  mm_status_t status;
  
  if(pm_strllen(region->phys_mem, 
	       PM_PROGRAM_TYPE_PARAM_LENGTH) >= PM_PROGRAM_TYPE_PARAM_LENGTH)
    {
      /* The phys_mem length is to large. */
      return PM_EPROGRAM_TYPE_SYNTAX_ERROR;
    }
  if(pm_strllen(region->log_mem, 
	       PM_PROGRAM_TYPE_PARAM_LENGTH) >= PM_PROGRAM_TYPE_PARAM_LENGTH)
    {
      /* The log_mem length is to large. */
      return PM_EPROGRAM_TYPE_SYNTAX_ERROR;
    }
  permission_binary = region->permissions;


  /* OSEhq09629, added sanity checks for permissions. */
  
  /* TEXT */
  /* READ and EXECUTE must be enabled. */
  if(PM_TEXT_REGION_TYPE == region_type)
  {
     if( ((permission_binary & MM_SU_EXECUTE) != MM_SU_EXECUTE) ||
	 ((permission_binary & MM_SU_READ) != MM_SU_READ))
     {
	return PM_EPROGRAM_TYPE_SYNTAX_ERROR;
     }
  }
  /* DATA POOL or HEAP. */
  /* READ and WRITE must be enabled. */
  else
  {
     if( ((permission_binary & MM_SU_WRITE) != MM_SU_WRITE) ||
	 ((permission_binary & MM_SU_READ) != MM_SU_READ))
     {
	return PM_EPROGRAM_TYPE_SYNTAX_ERROR;
     }
  }

  /* SU permissions larger than USR. */
  if((((permission_binary & MM_USR_EXECUTE) == MM_USR_EXECUTE)
      && ((permission_binary & MM_SU_EXECUTE) != MM_SU_EXECUTE))
     ||
     (((permission_binary & MM_USR_READ) == MM_USR_READ)
      && ((permission_binary & MM_SU_READ) != MM_SU_READ))
     ||
     (((permission_binary & MM_USR_WRITE) == MM_USR_WRITE)
      && ((permission_binary & MM_SU_WRITE) != MM_SU_WRITE)))
  {
     return PM_EPROGRAM_TYPE_SYNTAX_ERROR;
  }

  status = ose_mm_get_cache_type(region->cache_mode, NULL, NULL);
  if (status != MM_SUCCESS)
  {
     return PM_EPROGRAM_TYPE_SYNTAX_ERROR;
  }

  /* Restrict region with MAS-mapped log_mem if
   * program type specifies supervisor mode. */
  if (program_type->super ||
      region_type == PM_TEXT_REGION_TYPE ||
      region_type == PM_POOL_REGION_TYPE)
  {
     mm_status_t status;
     U32 mapping_type = 0;

     status = pm_vs_gmm_get_log_mem_mapping_type(
        PmInternalState->gmm_pid, region->log_mem, &mapping_type);
     if (status == MM_SUCCESS && mapping_type == MAS)
     {
        /* User configured a supervisor mode program type
         * with regions in MAS which is illegal, bail out. */
        return PM_EILLEGAL_REGION_LOG_MEM_TYPE;
     }
  }
  
  return PM_SUCCESS;
}

/*
 *===========================================================================
 *                        pm_add_program_type_request
 *===========================================================================
 * 
 * Description
 * -----------
 * Handle a PM_ADD_PROGRAM_TYPE_REQUEST signal. I.e. determine if the 
 * configuration is adding a new program type or if it modifies an existing 
 * one. Store the program type configuration and reply a status message.  
 *
 * Requirement
 * -----------
 * "program type handling". 
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * PmInternalState                  The PM internal state structure.
 * request                          The PM_ADD_PROGRAM_TYPE_REQUEST signal. 
 *
 * Returns
 * -------
 * None.
 *
 */

GLOBAL void
pm_add_program_type_request(struct PmInternalState * PmInternalState, 
			    struct PmProgramTypeInfo * request)
{
  PmStatus status;
  ProgramType * program_type = NULL;
  U32 i;
  struct PmAddProgramTypeReply * reply;
  struct OS_sig_info sig_info;
  PROCESS pool_id;

  assert(PmInternalState != NULL);
  assert(request != NULL);

  reply = (struct PmAddProgramTypeReply *) 
    alloc(sizeof(struct PmAddProgramTypeReply), 
	  PM_ADD_PROGRAM_TYPE_REPLY);

  /* Sanity checks. */
  pool_id = get_sig_poolid(get_bid(current_process()));
  (void)get_sig_info(&sig_info, pool_id, current_process(),
		     (OSADDRESS)request);

  /* Check the minimal number of regions in the program type. */
  if(request->program_type_size < PM_MINIMAL_NO_OF_REGIONS)
    {
      status = PM_EPROGRAM_TYPE_PARAM_REGIONS;
      goto pm_add_program_type_reply_failure;
    }

  /* Check the string length of the program type name. */
  else if(pm_strllen(request->program_type_name,
		     PM_PROGRAM_TYPE_PARAM_LENGTH) == PM_PROGRAM_TYPE_PARAM_LENGTH)
    {
      status = PM_EPROGRAM_TYPE_PARAM_NAME_CORRUPT;
      goto pm_add_program_type_reply_failure;
    }
  
  /* Check the supervisor flag. */
  else if(request->super != False && request->super != True)
    {
      status = PM_EPROGRAM_TYPE_PARAM_SUPER_CORRUPT;
      goto pm_add_program_type_reply_failure;
    }
  
  /* Check the PmProgramTypeInfo signal size.
   * Note, sizeof signal field  program_type[PM_MINIMAL_NO_OF_REGIONS] */
  else if(sizeof(struct PmProgramTypeInfo) 
	  + (sizeof(struct PmProgramRegionType) * 
	     (request->program_type_size - PM_MINIMAL_NO_OF_REGIONS)
	     ) > sig_info.sig_size)
    {
      status = PM_EPROGRAM_TYPE_PARAM_REGIONS;
      goto pm_add_program_type_reply_failure;
    }

  program_type = heap_alloc_private(sizeof(ProgramType) 
				    + (sizeof(struct PmProgramRegionType) * 
				       (request->program_type_size 
					- PM_DEFAULT_NO_OF_REGIONS)), 
				    __FILE__, __LINE__);

  /* Store the program type in the internal structure */
  program_type->super = (Boolean)request->super;
  program_type->no_of_regions = request->program_type_size;
  if(strlcpy(program_type->program_type_name, 
	     request->program_type_name, 
	     PM_PROGRAM_TYPE_PARAM_LENGTH) >= PM_PROGRAM_TYPE_PARAM_LENGTH)
    {
      /* The program type name is corrupt. */
      status = PM_EPROGRAM_TYPE_SYNTAX_ERROR;
      goto pm_add_program_type_reply_failure;
    }
  for(i=0;i<program_type->no_of_regions;i++)
    {
      status = pm_sanity_check_program_region_type(
         PmInternalState, program_type, &(request->program_type[i]), i);
      if (status != PM_SUCCESS)
        {
	  /* The program type name is corrupt. */
	  goto pm_add_program_type_reply_failure;
	}
      strcpy(program_type->regions[i].phys_mem, 
	     request->program_type[i].phys_mem); 
      strcpy(program_type->regions[i].log_mem, 
	     request->program_type[i].log_mem);
      program_type->regions[i].permissions = request->program_type[i].permissions;
      program_type->regions[i].cache_mode = request->program_type[i].cache_mode;
    }

  /* Store the new program type. */
  status = pm_store_program_type(PmInternalState, program_type);

  heap_free_private(program_type);

  /* Send a status reply to the client. */
  reply->status = status;
  strcpy(reply->program_type_name, request->program_type_name);

  send((union SIGNAL **) &reply, sender((union SIGNAL **) &request));
  free_buf((union SIGNAL **) &request);
  
  return;
  
 pm_add_program_type_reply_failure:

  if(program_type != NULL)
    {
      /* Clean up the heap buffer is necessary. */
      heap_free_private(program_type);
    }
  reply->status = status;
  (void)strlcpy(reply->program_type_name, request->program_type_name, 
		PM_PROGRAM_TYPE_PARAM_LENGTH);

  send((union SIGNAL **) &reply, sender((union SIGNAL **) &request));
  free_buf((union SIGNAL **) &request);  
}


/*
 *===========================================================================
 *                        pm_program_type_info_request
 *===========================================================================
 * 
 * Description
 * -----------
 * Handle a PM_PROGRAM_TYPE_INFO_REQUEST signal
 * 
 * Requirement
 * -----------
 * "program type handling". 
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * PmInternalState                  The PM internal state structure.
 * request                          The PM_PROGRAM_TYPE_INFO_REQUEST signal. 
 *
 * Returns
 * -------
 * None.
 *
 */

GLOBAL void
pm_program_type_info_request(struct PmInternalState * PmInternalState, 
			     struct PmProgramTypeInfoRequest * request)
{
  ProgramType * program;
  struct PmProgramTypeInfo * reply;
  PmStatus status;

  assert(PmInternalState != NULL);
  assert(PmInternalState->prog_types != NULL);

  if(pm_strllen(request->program_type_name,
		PM_PROGRAM_TYPE_PARAM_LENGTH) == PM_PROGRAM_TYPE_PARAM_LENGTH)
    {
      status = PM_EPROGRAM_TYPE_PARAM_NAME_CORRUPT;
      goto pm_program_type_info_request_failure;
    }

  /* Search for the program type. */
  program = (ProgramType *)traverseList1(PmInternalState->prog_types, 
					 pm_search_prog_type, 
					 request->program_type_name);

  if(program == NULL)
    {
      status = PM_EILLEGAL_PROGRAM_TYPE;
      goto pm_program_type_info_request_failure;
    }
  else
    {
      U32 i;
      size_t str_size;

      assert(program->no_of_regions >= PM_MINIMAL_NO_OF_REGIONS);

      /* Alloc PmProgramTypeInfo signal.
       * Note, sizeof signal field program_type[PM_MINIMAL_NO_OF_REGIONS] */
      reply =
	(struct PmProgramTypeInfo *)alloc(sizeof(struct PmProgramTypeInfo) 
					  + (sizeof(struct PmProgramRegionType) 
					     * (program->no_of_regions 
						- PM_MINIMAL_NO_OF_REGIONS)), 
					  PM_PROGRAM_TYPE_INFO_REPLY);

      reply->super = program->super;
      reply->program_type_size = program->no_of_regions;
      str_size = strlcpy(reply->program_type_name, program->program_type_name, 
			 PM_PROGRAM_TYPE_PARAM_LENGTH);
      ASSERT_WITHOUT_WARNINGS(str_size < PM_PROGRAM_TYPE_PARAM_LENGTH, str_size);

      for(i=0;i<reply->program_type_size;i++)
	{
	  str_size = strlcpy(reply->program_type[i].phys_mem, 
			     program->regions[i].phys_mem, 
			     PM_PROGRAM_TYPE_PARAM_LENGTH);
	  assert(str_size < PM_PROGRAM_TYPE_PARAM_LENGTH);
	  str_size = strlcpy(reply->program_type[i].log_mem, 
			 program->regions[i].log_mem, 
			 PM_PROGRAM_TYPE_PARAM_LENGTH);
	  assert(str_size < PM_PROGRAM_TYPE_PARAM_LENGTH);
	  reply->program_type[i].permissions = program->regions[i].permissions;
	  reply->program_type[i].cache_mode = program->regions[i].cache_mode;
	}
      
      reply->status = PM_SUCCESS;      
      send((union SIGNAL **) &reply, sender((union SIGNAL **) &request));
      free_buf((union SIGNAL **) &request);
      return;
    }

 pm_program_type_info_request_failure:
  reply =
    (struct PmProgramTypeInfo *)alloc(sizeof(struct PmProgramTypeInfo), 
				      PM_PROGRAM_TYPE_INFO_REPLY);

  (void)strlcpy(reply->program_type_name, request->program_type_name,
		PM_PROGRAM_TYPE_PARAM_LENGTH);
  
  reply->status = status;
  send((union SIGNAL **) &reply, sender((union SIGNAL **) &request));
  free_buf((union SIGNAL **) &request);
}
 

/*
 *===========================================================================
 *                        pm_program_types_request
 *===========================================================================
 * 
 * Description
 * -----------
 * Handle a PM_PROGRAM_TYPES_REQUEST signal
 * 
 * Requirement
 * -----------
 * "program type handling". 
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * PmInternalState                  The PM internal state structure.
 * request                          The PM_PROGRAM_TYPES_REQUEST signal. 
 *
 * Returns
 * -------
 * None.
 *
 */

GLOBAL void
pm_program_types_request(struct PmInternalState * PmInternalState, 
			 struct PmProgramTypesRequest * request)
{
  ProgramType * program;
  struct PmProgramTypesReply * reply;
  unsigned int i;
  char * program_types;

  assert(PmInternalState != NULL);
  assert(PmInternalState->prog_types != NULL);

  /* loop the program type to get all program type names. */
  i = PmInternalState->prog_types->elements;

  reply = (struct PmProgramTypesReply *) alloc(sizeof(struct PmProgramTypesReply) 
					       + (i * PM_PROGRAM_TYPE_PARAM_LENGTH), 
					       PM_PROGRAM_TYPES_REPLY);

  reply->program_types_no_of_strings = i;

  program_types = reply->program_types;

  program = firstNode(PmInternalState->prog_types);
  while(program != NULL)
    {
      size_t str_size;
      str_size = strlcpy(program_types, program->program_type_name, 
			 PM_PROGRAM_TYPE_PARAM_LENGTH);
      ASSERT_WITHOUT_WARNINGS(str_size < PM_PROGRAM_TYPE_PARAM_LENGTH, str_size);
      program_types += PM_PROGRAM_TYPE_PARAM_LENGTH;
      program = nextNode(program);
    }

  reply->status = PM_SUCCESS;

  send((union SIGNAL **) &reply, sender((union SIGNAL **) &request));
  free_buf((union SIGNAL **) &request);
}
