/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2001-2011 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */
/*
 ****************************************************************************
 *			  PM INTERNAL FUNCTIONALITY
 *
 *     $Workfile:   pm_programs.c  $
 *
 *
 ****************************************************************************
 * Description: The implementation of the programs support in the
 *              Program Manager.
 *
 */

/***********************************************************
 *
 * INCLUDE FILES
 *
 ***********************************************************/

/* OSE general headers. */
#include "ose.h"
#include "osetypes.h"
#include "ose_err/ose_err.h"

#ifndef OUTFMT_NO_SPRINTF
#define OUTFMT_NO_SPRINTF
#include "outfmt.h"
#undef OUTFMT_NO_SPRINTF
#else
#include "outfmt.h"
#endif

/* Standard headers. */
#include "assert.h"
#include "string.h"
#include "stdlib.h"
#include "stdio.h"
#include "malloc.h"
#include "ctype.h"

/* List handling functionality. */
#include "oldlist.h"

/* Lint directives. */
#include "pm_lint.h"

/* EFS functionality. */
#include "efs.h"

/* Bios functionality. */
#include "bios.h"

/* Core internal kernel services. */
#include "core_internal.h"

/* Program Management headers. */
#include "pm.h"
#include "pm_internal.h"
#include "pm_err.h"

/* Program Management block signal interface. */
#include "pm.sig"
#include "pm_internal.sig"
#include "rtl.sig"

/* Include the lm configuration definitions. */
#include "ose_conf.h"

/* Memory manager interface and definitions. */
#include "mm.sig"
#include "mmtypes.h"

/* Heap manager interface and definitions. */
#include "ose_heap.h"

/* The PM support interfaces. */
#include "pm_vs_gmm.h"
#include "pm_conf.h"
#include "pm_load_modules.h"
#include "pm_program_types.h"
#include "pm_program_types_support.h"
#include "pm_programs.h"

/* Low level kernel functions. */
#include "ose_spi.h"
#include "pose.h"

/* SMP  support */
#ifdef OSE_SMP
#include "cpu_smp.h"
#endif

/***********************************************************
 *
 * CONSTANTS AND DEFINITIONS
 *
 ***********************************************************/

/* The background process priority. */
#define PM_BACKGROUND_PROCESS        32

/* The OSE PM directive prefix. */
#define PM_DIRECTIVE_PREFIX          "OSE_LM_"

/* The boolean YES/NO for configuration. */
#define PM_BOOL_YES                  "YES"
#define PM_BOOL_NO                   "NO"

/* Create program default values. */
#define PM_DEFAULT_MAIN_PROC_NAME    "main"
#define PM_DEFAULT_MAIN_PROC_PRIO    "31"
#define PM_DEFAULT_MAIN_PROC_STACK   "1023"
#define PM_DEFAULT_POOL_SIZE         "0x30000"
#define PM_DEFAULT_STACK_SIZES       "256,512,1024,2048,4096,8192,16384,65536"
#define PM_DEFAULT_SIGNAL_SIZES      "31,63,127,255,1023,4095,16383,65535"
#define PM_DEFAULT_POOL_FILL         PM_BOOL_NO

/* ASCII hexadecimal value string maximum length + terminator. */
/* NOTE: The value also holds the maximum 32 bit value (4294967295) in
 * decimal value representation + terminator.
 */
#define PM_HEX_STRING_LEN            11

/* Max number of signal/stack buffer definitions. */
#define PM_MAX_BUFFERS_SIZE          9

/* Minimum size of a pool. Should be defined by pose. */
#define PM_MIN_POOL_SIZE             1024

/* ramlog prefix. */
#define RAMLOG_PREFIX "PM: "

/* Interpretation of the .cleanup field in the ProgramMetaData struct. */
#define PROG_CLEANUP_FREE_PHYS_DATA_RESOURCE    1

/* For automatic loading of load module symbols in simics and GDB. */
#ifdef __ppc
#define MAGIC(n) do {                                   \
  __asm__ __volatile__ ("rlwimi 0,0,0,0," #n);          \
} while (0)

#define SIMICS_MAGIC(text, name_ptr, data) do {                         \
  __asm__ __volatile__ ("stwu r1,-20(r1); "                             \
                        "stw r2,8(r1); "                                \
                        "stw r3,12(r1); "                               \
                        "stw r4,16(r1); "                               \
                        "mr r2,%0; "                                    \
                        "mr r3,%1; "                                    \
                        "mr r4,%2"                                      \
                        : : "r" (text), "r" (name_ptr), "r" (data) :    \
                          "r1", "r2", "r3", "r4");                      \
  MAGIC(0);                                                             \
  __asm__ __volatile__ ("lwz r4,16(r1); "                               \
                        "lwz r3,12(r1); "                               \
                        "lwz r2,8(r1);  "                               \
                        "addi r1,r1,20"                                 \
                        : : : "r1", "r2", "r3", "r4" );                 \
} while(0)
#elif defined(OSARM)
#  ifdef __GNUC__
static void SIMICS_MAGIC(
   uint32_t text, const char * const name_ptr, uint32_t data)
{
   register uint32_t _text __asm__ ("r0") = text;
   register const char * const _name_ptr __asm__ ("r1") = name_ptr;
   register uint32_t _data __asm__ ("r2") = data;
   __asm__ __volatile__ ("orreq r0,r0,r0"
                         : : "r" (_text), "r" (_name_ptr), "r" (_data));
}
#  elif defined(__ARMCC_VERSION)
static __asm void SIMICS_MAGIC(
   uint32_t text, const char * const name_ptr, uint32_t data)
{
   CODE32
   orreq r0,r0,r0; /* Magic 0 */
   bx    lr
}
#  else
/*
 *   If this error occurs, it means that no SIMICS_MAGIC() support has been
 *   implemented for this compiler. Either add it, or define an empty macro
 *   SIMICS_MAGIC(), if Simics magic functionality is not needed.
 */
#    error Unsupported compiler for ARM. See source code for details.
#  endif
#else
#define SIMICS_MAGIC(text, name_ptr, data)
/* We do not have simics for this architecture. */
#endif

/***********************************************************
 *
 * STRUCTURES AND GLOBALS
 *
 ***********************************************************/

/*
 * External reference to core internal function.
 * Retrieves information about the mainpool.
 * Defined in _zzmain.c
 */
extern Boolean ose_has_mainpool(Boolean user_mode);

/*
 *===========================================================================
 *                        Program_Conf
 *===========================================================================
 *
 * Description
 * -----------
 * A structure with program configuration used when creationg a program.
 *
 * Structure
 * ---------
 * struct Program_Conf
 *
 * Parameter description
 * ---------------------
 * program_type           The program type of the program.
 * pool_size              The pool_size of the program.
 * stack                  The stack configuration of the program.
 * signal                 The signal configuration of the program.
 * pool_ext_size          The local pool extension fragment size.
 * pool_max_size          The maximum local pool size.
 * pool_fill              Controls if local pool should be pattern filled.
 * reent                  The reentrancy of the program.
 * shared_model           True if shared execution model
 * user_id                The user id of the program.
 * main_proc_pri          The main process priority.
 * main_proc_stack        The main process stack.
 * main_proc_name         The main process name.
 * main_block_name        The main block name or NULL.
 * intercepted            The program shall be intercepted after it is created.
 * locked                 The program shall not be moved between CPUs.
 * cpu                    The cpu of the program.
 */

struct Program_Conf
{
  char program_type[PM_PROGRAM_TYPE_PARAM_LENGTH];
  U32 pool_size;
  OSBUFSIZE stack[PM_MAX_BUFFERS_SIZE];
  OSBUFSIZE signal[PM_MAX_BUFFERS_SIZE];
  U32 pool_ext_size;
  U32 pool_max_size;
  Boolean pool_fill;
  Boolean reent;
  Boolean shared_model;
  Boolean intercepted;
  Boolean locked;
  OSUSER user_id;
  OSPRIORITY main_proc_pri;
  OSADDRESS main_proc_stack;
  char *main_proc_name;
  char *main_block_name;
  cpuid_t cpu;
};

/***********************************************************
 *
 * FUNCTION DEFINITIONS
 *
 ***********************************************************/

/*
 *===========================================================================
 *                        pm_pool_extender
 *===========================================================================
 *
 * Description
 * -----------
 * The local pool extender that is registered with the kernel.
 * Basically it only allocates a region from MM.
 *
 * Parameter description
 * ---------------------
 * pool                             Pool Id of pool to extend
 * extend_count                     Number of times pool has been extended
 * fragment_size                    Size of desired extend fragment
 * prog_type_p                      Pointer to the program type of this program
 * base                             Pointer to the memory region created
 *
 * Returns
 * -------
 *  0                               The pool extension failed.
 * >0                               The actual size of the created region.
 *
 */

static OSADDRESS
pm_pool_extender(PROCESS pool, OSADDRESS extend_count, OSADDRESS fragment_size,
                 void *prog_type_p, void **base)
{
  char name[32];
  OSDOMAIN domain;
  U32 extend_size = (U32) fragment_size;
  OSADDRESS extend_base = 0;
  ProgramType * prog_type;
  U32 align = 4;

  (void) pool;
  
  sprintf(name, "pool_extend_fragment%u", extend_count);

  /* Get domain id. 
   * FIXME: Perhaps this should be sent as an argument from pose? 
   */
  domain = zzkrn_get_pid_info(current_process(),16);
  
  /* pointer to our program type */
  prog_type = (ProgramType *)prog_type_p;
  
  /* Send signal to MM requesting a region */
  if(MM_SUCCESS == 
     pm_vs_gmm_create_region(
                         GlobalPmInternalState->gmm_pid,
                         name,
                         domain,
                         &extend_size,
                         &extend_base,
                         prog_type->regions[PM_POOL_REGION_TYPE].phys_mem,
                         prog_type->regions[PM_POOL_REGION_TYPE].log_mem,
                         prog_type->regions[PM_POOL_REGION_TYPE].permissions,
                         prog_type->regions[PM_POOL_REGION_TYPE].cache_mode,
                         &align,
                         MM_CREATE_REGION_SPECIFY_ALIGN
                         | MM_CREATE_REGION_SPECIFY_NAME))
  {
    /* On success, return base (0 is a valid base address) and actual size */
    *base = (void*)extend_base;
    return (OSADDRESS)extend_size;
  }

  /* Something went wrong, return NULL to show failure */
  return 0;
}

/*
 *===========================================================================
 *                        pm_user_mode_warning_ignore
 *===========================================================================
 *
 * Description
 * -----------
 * Process error handler for create_block(). We ignore a possible
 * warning for user mode blocks in the monolith if we do use the user mode
 * protection model. Then it is no error, since we will warp the block
 * to another segment by attach_segment().
 */
static OSADDRESS
pm_user_mode_warning_ignore(OSBOOLEAN user_called, OSERRCODE ecode,
			    OSERRCODE extra)
{

   (void) extra;
   
  if(!user_called && (ecode & OSE_EERROR_MASK) == OSE_EILLEGAL_USER_MODE)
  {
/* Accept only if plibs. */
#if MEMORY_MODEL > 0
    return 1;			/* Handled error. */
#else
    ramlog_printf("PGM: Illegal creation of a user mode program!\n");
    ramlog_printf("PGM: This is only possible using the core user protection libraries.\n");
#endif
  }
  return 0;			/* Not handled error. */
}


/*
 *===========================================================================
 *                        pm_create_block_name
 *===========================================================================
 *
 * Description
 * -----------
 * Creates a main block name for newly instantiated programs. The name
 * is formatted as <install_handle>-<program name>. pm_destroy_block_name
 * should be called when the buffer holding the name is no longer needed.
 *
 * Returns
 * -------
 * A buffer holding the block name, or NULL if no program name is supplied,
 * i.e. when the install handle equals the block name.
 */
static char *
pm_create_block_name(const char *install_handle, const char *program_name)
{
   char  *block_name;
   size_t block_name_len;

   if (program_name == NULL)
   {
      return NULL;
   }

   /* block_name  = install_handle + '-' + program_name + '\0' */
   block_name_len = strlen(install_handle) + 1 + strlen(program_name) + 1;
   block_name     = heap_alloc_private(block_name_len, __FILE__, __LINE__);

   snprintf(block_name, block_name_len, "%s-%s", install_handle, program_name);

   block_name[block_name_len - 1] = '\0';

   return block_name;
}

/*
 *===========================================================================
 *                        pm_destroy_block_name
 *===========================================================================
 *
 * Description
 * -----------
 * Deallocates a buffer holding a main block name. 
 */
static void
pm_destroy_block_name(char *block_name)
{
   if (block_name != NULL)
   {
      heap_free_private(block_name);
   }
}


/*
 *===========================================================================
 *                        pm_search_program_by_domain
 *===========================================================================
 *
 * Description
 * -----------
 * Search for a specific program using the programs domain (as
 * part of traverseList1).
 *
 * Requirement
 * -----------
 * "Program domains".
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * node                             The node (in the list) to compare to.
 * domain                           The domain number to search for.
 *
 * Returns
 * -------
 * (void *) 0                       The node did not match the install handle,
 *                                  ZERO is returned.
 * (void *) node                    The node matched the domain, the
 *                                  pointer to the node is returned.
 *
 *
 */

GLOBAL void *
pm_search_program_by_domain(void * node, void * domain)
{
  assert(node != NULL);

  if(((ProgramMetaData *)node)->domain == (PROCESS)domain)
    {
      return (void *) node;
    }
  else
    {
      return NULL;
    }
}

/*
 *===========================================================================
 *                        pm_search_program_by_install_handle
 *===========================================================================
 *
 * Description
 * -----------
 * Search for a specific program using the programs install handle (as
 * part of traverseList1).
 *
 * Requirement
 * -----------
 * "Programs install handle".
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * node                             The node (in the list) to compare to.
 * name                             The install handle to search for.
 *
 * Returns
 * -------
 * NULL                             The node did not match the install handle,
 *                                  ZERO is returned.
 * (void *) node                    The node matched the install handle, the
 *                                  pointer to the node is returned.
 *
 *
 */

GLOBAL void *
pm_search_program_by_install_handle(void * node, void * install_handle)
{
  assert(node != NULL);
  assert(((ProgramMetaData *)node)->lm != NULL);
  assert(install_handle != NULL);

  if(strcmp(((ProgramMetaData *)node)->lm->install_handle,
	    (char *)install_handle) == 0)
    {
      return (void *) node;
    }
  else
    {
      return NULL;
    }
}

/*
 *===========================================================================
 *                        pm_program_search_by_mainpid
 *===========================================================================
 *
 * Description
 * -----------
 * Search for a specific program using the programs mainpid (as
 * part of traverseList1).
 *
 * Requirement
 * -----------
 * "Programs mainpid".
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * node                             The node (in the list) to compare to.
 * name                             The mainpid to search for.
 *
 * Returns
 * -------
 * NULL                             The node did not match the mainpid,
 *                                  ZERO is returned.
 * (void *) node                    The node matched the mainpid, the
 *                                  pointer to the node is returned.
 *
 *
 */

GLOBAL void *
pm_program_search_by_mainpid(void * node, void * main_pid)
{
  ProgramMetaData * program;

  assert(node != NULL);
  assert(((ProgramMetaData *)node)->mainpid != (PROCESS)0);
  assert(main_pid != NULL);

  program = (ProgramMetaData *) node;

  if(program->mainpid == (PROCESS)main_pid)
    {
      return (void *) node;
    }
  else
    {
      return NULL;
    }
}

/*
 *===========================================================================
 *                        pm_program_search_by_progpid
 *===========================================================================
 *
 * Description
 * -----------
 * Search for a specific program using the programs progpid (as
 * part of traverseList1).
 *
 * Requirement
 * -----------
 * "Program search".
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * node                             The node (in the list) to compare to.
 * progpid                          The progpid to search for.
 *
 * Returns
 * -------
 * NULL                             The node did not match the progpid,
 *                                  ZERO is returned.
 * (void *) node                    The node matched the progpid, the
 *                                  pointer to the node is returned.
 *
 *
 */

#if 0
GLOBAL void *
pm_program_search_by_progpid(void * node, void * progpid)
{
  ProgramMetaData * program;

  assert(node != NULL);
  /*The pid nr can not be asserted, since all nr's are legal.*/

  program = (ProgramMetaData *) node;

  if(program->progpid == (PROCESS)progpid)
    {
      return (void *) node;
    }
  else
    {
      return NULL;
    }
}
#endif

/*
 *===========================================================================
 *                        pm_program_search_by_segpid
 *===========================================================================
 *
 * Description
 * -----------
 * Search for a specific program using the programs segpid (as
 * part of traverseList1).
 *
 * Requirement
 * -----------
 * "Program search".
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * node                             The node (in the list) to compare to.
 * segpid                           The segpid to search for.
 *
 * Returns
 * -------
 * NULL                             The node did not match the segpid,
 *                                  ZERO is returned.
 * (void *) node                    The node matched the segpid, the
 *                                  pointer to the node is returned.
 *
 *
 */

GLOBAL void *
pm_program_search_by_segpid(void * node, void * segpid)
{
   ProgramMetaData * program;

   assert(node != NULL);
   /*The segpid nr can not be asserted, since all nr's are legal.*/

   program = (ProgramMetaData *) node;

   if(program->segpid == (PROCESS)segpid)
   {
      return (void *) node;
   }
   else
   {
      return NULL;
   }
}

/*
 *===========================================================================
 *                        pm_program_search_by_mainbid
 *===========================================================================
 *
 * Description
 * -----------
 * Search for a specific program using the programs mainbid (as
 * part of traverseList1).
 *
 * Requirement
 * -----------
 * "Programs"
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * node                             The node (in the list) to compare to.
 * mainbid                          The mainbid to search for.
 *
 * Returns
 * -------
 * NULL                             The node did not match the mainbid,
 *                                  ZERO is returned.
 * (void *) node                    The node matched the mainbid, the
 *                                  pointer to the node is returned.
 *
 *
 */

#if 0
GLOBAL void *
pm_program_search_by_mainbid(void * node, void * mainbid)
{
  ProgramMetaData * program;

  assert(node != NULL);
  assert(((ProgramMetaData *)node)->mainbid != (PROCESS)0);
  assert(mainbid != NULL);

  program = (ProgramMetaData *) node;

  if(program->mainbid == (PROCESS)mainbid)
    {
      return (void *) node;
    }
  else
    {
      return NULL;
    }
}
#endif

/*
 *===========================================================================
 *                        is_line_terminator
 *===========================================================================
 *
 * Description
 * -----------
 * Helper to linelen & linecmp_i. Used to identify end of line.
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * c                    The character to checked.
  *
 * Returns
 * -------
 * Returns 1 if c is a line terminator. 0 otherwise.
 */

static int is_line_terminator(const char c)
{
   return (c == '\0' || c == '\n' || c == '\r');
}

/*
 *===========================================================================
 *                        linelen
 *===========================================================================
 *
 * Description
 * -----------
 * Calculates the length of the line l. Like strlen, but uses
 * is_line_terminator() to identify end of line.
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * l                    The line to checked.
  *
 * Returns
 * -------
 * Returns the length of the line.
 */

static size_t
linelen(const char *line)
{
   const char *line0 = line;

   while (!is_line_terminator(*line))
   {
      line++;
   }
   return line-line0;
}

/*
 *===========================================================================
 *                        line_eq_i
 *===========================================================================
 *
 * Description
 * -----------
 * Checks if the lines l1 & l2 are equal ignoring case.
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * line1, line2      The lines to compared.
  *
 * Returns
 * -------
 * Returns True if line1 is equal to line2. Otherwise False is returned.
 * Uses is_line_terminator() to identify end of line.
 */

static Boolean
line_eq_i(const char *line1, const char *line2)
{
   while (!is_line_terminator(*line1))
   {
      if (tolower(*line1) != tolower(*line2))
      {
         return False;
      }
      line1++;
      line2++;
   }
   
   /* (*s1 == *s2) would be incorrect. One could be '\0' and the other '\n'.
    */
   if (is_line_terminator(*line2))
   {
      return True;
   }
   
   return False;
}

/*
 *===========================================================================
 *                        pm_program_conf_get_value
 *===========================================================================
 *
 * Description
 * -----------
 * Search and read a value of a program parameter, all parameters are '\0'
 * terminated.
 *
 * Requirement
 * -----------
 * "Program parameters"
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * program_conf         The program configuration.
 * no_of_strings        The number of strings in the program configuration.
 * name                 The name of the program parameter to read.
 * value                The value of the program parameter.
 *
 * Returns
 * -------
 * True/False           Read success or read failure.
 */

static Boolean
pm_program_conf_get_value(const char * program_conf, U32 no_of_strings,
			  const char * name, const char ** value)
{
  U32 i;
  const char * curr_var;
  U32 name_len;

  assert(program_conf != NULL);
  assert(name != NULL);
  assert(value != (const char **) NULL);

  curr_var = program_conf;
  name_len = strlen(name);

  for(i=0; i<no_of_strings; i++)
  {
     if(strncmp(curr_var, name, name_len) == 0
        && curr_var[name_len] == '=')
     {
        *value = curr_var + name_len + 1;
        return True;
     }
        
     curr_var += strlen(curr_var) + 1;
  }
  
  return False;
}


/*
 *===========================================================================
 *                        pm_check_binary_32_conf
 *===========================================================================
 *
 * Description
 * -----------
 * Evaluate and turn an value in ASCII format to binary format, allowed
 * terminators are '\0' and '\n'. Note, octal is not supported.
 *
 * Requirement
 * -----------
 * "Program parameters"
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * value                The value of the program parameter.
 * bin_val              The binary translation of value.
 *
 * Returns
 * -------
 * True/False           Read success or read failure.
 */

static Boolean
pm_check_binary_32_conf(const char *value, U32 * bin_val)
{
  char ascii_val[PM_HEX_STRING_LEN];
  char * endmark;
  char * ptr;

  assert(value != NULL);
  assert(bin_val != NULL);

  (void)strlcpy(ascii_val, value, PM_HEX_STRING_LEN);
  assert(NULL == strchr(ascii_val, '='));

  if(((ascii_val[0] == '0' &&
       (ascii_val[1] != 'x' && (ascii_val[1] != '\0' && ascii_val[1] != '\n')))
      || (ascii_val[0] > '9' && ascii_val[0] < '0')))
    {
      return False;
    }

  /* Remove the \n terminator in order to be able to use strtoul. */
  endmark = strchr(ascii_val, '\n');
  if(endmark != NULL)
    {
      *endmark = '\0';
    }

  *bin_val = strtoul(ascii_val, &ptr, 0);

  if(ptr == ascii_val || *ptr != '\0')
    {
      return False;
    }

  return True;
}

/*
 *===========================================================================
 *                        pm_check_binary_8_conf
 *===========================================================================
 *
 * Description
 * -----------
 * Evaluate and turn an value in ASCII format to binary format, allowed
 * terminators are '\0' and '\n'. Note, octal is not supported.
 *
 * Requirement
 * -----------
 * "Program parameters"
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * value                The value of the program parameter.
 * bin_val              The binary translation of value.
 *
 * Returns
 * -------
 * True/False           Read success or read failure.
 */

static Boolean
pm_check_binary_8_conf(const char *value, U8 * bin_val)
{
  char ascii_val[PM_HEX_STRING_LEN];
  char * endmark;
  char * ptr;
  U32 x;

  assert(value != NULL);
  assert(bin_val != NULL);

  (void)strlcpy(ascii_val, value, PM_HEX_STRING_LEN);
  assert(NULL == strchr(ascii_val, '='));

  if(((ascii_val[0] == '0' &&
       (ascii_val[1] != 'x' && (ascii_val[1] != '\0' && ascii_val[1] != '\n')))
      || (ascii_val[0] > '9' && ascii_val[0] < '0')))
    {
      return False;
    }

  /* Remove the \n terminator in order to be able to use strtoul. */
  endmark = strchr(ascii_val, '\n');
  if(endmark != NULL)
    {
      *endmark = '\0';
    }

  x = strtoul(ascii_val, &ptr, 0);

  if(ptr == ascii_val || *ptr != '\0' || x > 0xff)
    {
      return False;
    }
  *bin_val = (U8)x;

  return True;
}


/*
 *===========================================================================
 *                        pm_check_buffers_conf
 *===========================================================================
 *
 * Description
 * -----------
 * Evaluate and turn a pool buffer sizes declaration to binary format.
 * The format of value is "<value0>,<value1>, ...,<valueN>" where N is at most
 * 8.
 *
 * Requirement
 * -----------
 * "Program parameters"
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * value                The value of the program parameter.
 * buffers              The buffer sizes in binary format.
 *
 * Returns
 * -------
 * True/False           Read success or read failure.
 */

static Boolean
pm_check_buffers_conf(const char * value, OSBUFSIZE * buffers)
{
  U32 i;
  const char * cur_var;

  assert(value != NULL);
  assert(buffers != NULL);

  /*
   * Special case - no local pool buffers configuration -> inherit mainpool
   * Note that this implies that the largest signal size, computed as
   * conf->signal[conf->signal[0]], will be zero.
   */
  if (value[0] == '\0' || value[0] == '0')
    {
      buffers[0] = 0;
      return True;
    }

  cur_var = value;

  for(i=1; i<PM_MAX_BUFFERS_SIZE; i++)
    {
      char ascii_val[PM_HEX_STRING_LEN];
      char * endmark;

      (void)strlcpy(ascii_val, cur_var, PM_HEX_STRING_LEN);
      endmark = strchr(ascii_val, ',');
      if(endmark != NULL)
	{
	  *endmark = '\0';
	}

      if(pm_check_binary_32_conf(ascii_val, &buffers[i]) == False)
	{
	  return False;
	}

      if(i > 1 && buffers[i] < buffers[i-1])
	{
	  return False;
	}

      if(endmark == NULL)
	{
	  /* Last value in the buffer. */
	  break;
	}
      cur_var = cur_var + ( endmark + 1 - ascii_val);
    }

  buffers[0] = i;

  return i < PM_MAX_BUFFERS_SIZE ? True : False;
}

/*
 *===========================================================================
 *                        pm_check_boolean_conf
 *===========================================================================
 *
 * Description
 * -----------
 * Evaluate and turn a boolean parameter value to binary format.
 *
 * Requirement
 * -----------
 * "Program parameters"
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * value                The value of the program parameter.
 * reent                The reentrant binary value.
 *
 * Returns
 * -------
 * True/False           Read success or read failure.
 */

static Boolean
pm_check_boolean_conf(const char * value, Boolean * val)
{
   if (line_eq_i(PM_BOOL_YES, value))
   {
      *val = True;
      return True;
   }
   
   if (line_eq_i(PM_BOOL_NO, value))
   {
      *val = False;
      return True;
   }

   return False;
}


/*
 *===========================================================================
 *                        pm_check_main_proc_pri
 *===========================================================================
 *
 * Description
 * -----------
 * Evaluate and turn a main process priority parameter value to binary format.
 *
 * Requirement
 * -----------
 * "Program parameters"
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * value                The value of the program parameter.
 * pri                  The priority of the main process.
 *
 * Returns
 * -------
 * True/False           Read success or read failure.
 */

static Boolean
pm_check_main_proc_pri_conf(const char * value, OSPRIORITY * pri)
{
  assert(value != NULL);
  assert(pri != NULL);

  if(pm_check_binary_8_conf(value, (U8 *)pri) == False
     || *pri > PM_BACKGROUND_PROCESS)
    {
      return False;
    }

  return True;
}



/*
 *===========================================================================
 *                        pm_check_pool_size_conf
 *===========================================================================
 *
 * Description
 * -----------
 * Evaluate and turn a pool size parameter value to binary format.
 *
 * Requirement
 * -----------
 * "Program parameters"
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * value                The value of the program parameter.
 * size                 The binary pool size.
 *
 * Returns
 * -------
 * True/False           Read success or read failure.
 */

static Boolean
pm_check_pool_size_conf(const char * value, U32 * size)
{
  assert(value != NULL);
  assert(size != NULL);

  /* Special case - no local pool configuration -> inherit mainpool */
  if (value[0] == '\0')
    {
      *size = 0;
      return True;
    }
  
  if(pm_check_binary_32_conf(value, size) == False
     || (*size != 0 && *size < PM_MIN_POOL_SIZE))
    {
      return False;
    }

  return True;
}

/*
 *===========================================================================
 *                        pm_check_pool_ext_size_conf
 *===========================================================================
 *
 * Description
 * -----------
 * Evaluate and turn a pool extension size parameter value to 
 * binary format.
 *
 * Requirement
 * -----------
 * "Program parameters"
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * value                The value of the program parameter.
 * size                 The binary pool size.
 *
 * Returns
 * -------
 * True/False           Read success or read failure.
 */

static Boolean
pm_check_pool_ext_size_conf(const char * value, U32 * size)
{
  assert(value != NULL);
  assert(size != NULL);

  /* 0 is used to indicate to pm_check_pool_extension() that 
   * no value was specified.
   */
  if (value[0] == '\0')
    {
      *size = 0;
      return True;
    }
  
  if(pm_check_binary_32_conf(value, size) == False)
    {
      return False;
    }

  return True;
}


/*
 *===========================================================================
 *                        pm_inherit_mainpool
 *===========================================================================
 *
 * Description
 * -----------
 * Adjust loadmodule parameter interpretation to allow inheritance of a 
 * global mainpool if available in the system.
 *
 * Requirement
 * -----------
 *
 * Side effects
 * ------------
 *
 * Parameter description
 * ---------------------
 * Five parameters (all optional) controls the pool setup for a program 
 * (excluding the parameters controlling the pool extension functionality).
 * krn/mainpool         - system parameter defining a common LM pool
 * OSE_LM_POOL_SIZE     - LM par defining the size of a local pool
 * OSE_LM_STACK_SIZES   - LM par defining buffer sizes in a local stack pool
 * OSE_LM_SIGNAL_SIZES  - LM par defining buffer sizes in a local signal pool
 * OSE_LM_POOL_MAX_SIZE - LM par defining the maximum size of a local pool
 *                        (in case pool extension is enabled).
 * 
 * The priority order in case of missing parameters are (high to low):
 * "loadmodule parameter -> system parameter -> default value"
 *
 * If at least one of the OSE_LM_POOL_SIZE or the OSE_LM_POOL_MAX_SIZE 
 * parameters are present, a local pool is created. 
 *
 * The presence of the pool type parameters, OSE_LM_xxx_SIZES, defines 
 * whether allocation should be done for the corresponding type 
 * in the local pool along with the defined buffer sizes.
 *
 * A combination of OSE_LM_POOL_SIZE/OSE_LM_POOL_MAX_SIZE and one (or both) 
 * of the OSE_LM_xxx_SIZES parameters is therefore preferred to create the 
 * local pool. 
 * 
 * If one of the OSE_LM_xxx_SIZES parameters are missing, the 
 * program will inherit the corresponding pool from krn/mainpool.
 *
 * If krn/mainpool is missing and some (or all) loadmodule parameters are 
 * missing, default values will be used to create a local pool for the
 * missing pool type(s).
 * 
 * If both the OSE_LM_POOL_SIZE and the OSE_LM_POOL_MAX_SIZE parameters  
 * are missing and krn/mainpool is present, allocation for both stacks  
 * and signals will be done from krn/mainpool.
 *
 * The mainpool can be created with user mode access permissions or not. 
 *
 * Returns
 * -------
 * True/False       True - OK  
 *                  False - A problem in the configuration was encountered
 *
 */

static Boolean
pm_inherit_mainpool(struct Program_Conf *conf, Boolean user_mode)
{
  /* If no mainpool exists or if the program does not have enough access 
     rights to use the pool - set defaults for the missing LM parameters */
  if (ose_has_mainpool(user_mode) == False) 
  {
    /* User has configured no pool - set default value */
    if ((conf->pool_size == 0 && conf->pool_max_size == 0) &&
        !pm_check_pool_size_conf(PM_DEFAULT_POOL_SIZE, &(conf->pool_size)))
    {
      return False;
    }
    /* User has configured no signal buffers - set default sizes */
    if (conf->signal[0] == 0 &&
        !pm_check_buffers_conf(PM_DEFAULT_SIGNAL_SIZES, conf->signal))
    {
      return False;
    }
    /* User has configured no stack buffers - set default sizes */
    if (conf->stack[0] == 0 &&
        !pm_check_buffers_conf(PM_DEFAULT_STACK_SIZES, conf->stack))
    {
      return False;
    }      
  }  

  /* Special case - a user has configured a pool size but no buffers */
  if ((conf->pool_size >  0 || conf->pool_max_size > 0) && 
      conf->signal[0] == 0 && 
      conf->stack[0]  == 0)
  {
    /* Set default values on buffers */
    if (!pm_check_buffers_conf(PM_DEFAULT_SIGNAL_SIZES, conf->signal) ||
        !pm_check_buffers_conf(PM_DEFAULT_STACK_SIZES,  conf->stack ))
    {
      return False;	  
    }      
  }  
  return True;
}

/*
 *===========================================================================
 *                        pm_check_pool_extension
 *===========================================================================
 *
 * Description
 * -----------
 * Evaluate and adjust the pool extension parameters to ensure consistency.
 * 
 * Requirement
 * -----------
 *
 * Side effects
 * ------------
 *
 * Parameter description
 * ---------------------
 *
 * Returns
 * -------
 * True/False       True - OK  
 *                  False - A problem in the configuration was encountered
 *
 */

#define POOL_ALIGNMENT 32

static Boolean 
pm_check_pool_extension(struct Program_Conf *conf)
{
  /* If pool extension is enabled */
  if (conf->pool_max_size != 0)
  {
    /* The maximum pool size must be larger than the initial pool size */
    if (conf->pool_max_size < conf->pool_size)
    {
      ramlog_printf(RAMLOG_PREFIX "(%s) Error: maximum pool size "
                    "is configured as smaller than initial pool size.\n",
                    conf->main_proc_name);
      return False;
    }

    /*
     * If pool extend fragment size is not defined, set it to default.
     * It must be large enough to hold the largest stack or signal.
     */
    if (conf->pool_ext_size == 0)
    {
      U32 default_ext_size;

      default_ext_size = (conf->pool_max_size - conf->pool_size) / 5;

    if (default_ext_size < conf->signal[conf->signal[0]] + 1024)
      default_ext_size = conf->signal[conf->signal[0]] + 1024;

    if (default_ext_size < conf->stack[conf->stack[0]] + 1024)
      default_ext_size = conf->stack[conf->stack[0]] + 1024;

      /* Align size to a multiple of 32 bytes. */     
      conf->pool_ext_size = ((default_ext_size + (POOL_ALIGNMENT-1)) 
                             /*& (OSADDRESS) ~(POOL_ALIGNMENT-1)*/);
    }
    
    /* If a maximum size was specified but no initial size was, 
     * start with one extension fragment as pool size.
     */
    if(conf->pool_size == 0)
    {
      conf->pool_size = conf->pool_ext_size;
    }
    
    /* Pool extend fragment size must be larger than largest possible
     * pool object (plus 1024 bytes).
     */
    if ((conf->pool_ext_size < conf->signal[conf->signal[0]] + 1024) ||
        (conf->pool_ext_size < conf->stack[conf->stack[0]] + 1024))
    {
      ramlog_printf(RAMLOG_PREFIX "(%s) Error: pool extension fragment "
                    "size is smaller than largest possible pool object "
                    "plus 1024 bytes. Fragment size: 0x%x. "
                    "Largest signal: 0x%x. Largest stack: 0x%x\n",
                    conf->main_proc_name,
                    conf->pool_ext_size,
                    conf->signal[conf->signal[0]], 
                    conf->stack[conf->stack[0]]);
      return False;
    }
  }
  return True;
}

/*
 *===========================================================================
 *                        conf_get_value
 *===========================================================================
 *
 * Description
 * -----------
 * A helper to pm_read_program_config. Reads the value of a given parameter
 * from the program or load module configuration. If the parameter is not
 * found, a provided default value is used.
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * in:
 *  lm_conf             The load module configuration (from file *.lmconf).
 *  program_conf_strs   The program configuration (from request sig).
 *  no_of_string        The number of strings in the program configuration
 *                      above.
 *  param_name          The name of the parameter.
 *  default_value       The default value. Used if the parameter is not found,
 *
 * Returns
 * -------
 * The parameter's value.
 *
 */

static const char *
conf_get_value(const char *lm_conf,
               const char *program_conf_strs,
               U32        no_of_strings,
               const char *param_name,
               const char *default_value)
{
   const char *value;

   if (False == pm_program_conf_get_value(program_conf_strs, no_of_strings,
                                          param_name, &value))
   {
      if (0 == pm_lm_conf_get_value(lm_conf, param_name, &value))
      {
         value = default_value;
      }
   }
   return value;
}

/*
 *===========================================================================
 *                        pm_read_program_config
 *===========================================================================
 *
 * Description
 * -----------
 * Read and sanity check the program configuration to create a program.
 *
 * Requirement
 * -----------
 * "Create program".
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * in:
 * lm_conf             The load module configuration (from file *.lmconf).
 * program_conf_strs   The program configuration (from request sig).
 * no_of_string        The number of strings in the program configuration above.
 *
 * out:
 * conf_x              The programs configuration (with defaults).Allocated.
 *
 * in:
 * client                   The client of the program creation.
 *
 * Returns
 * -------
 * True/False.             True - All configuration was correct and
 *                                found.
 *
 */

static Boolean
pm_read_program_config(const char * lm_conf, const char * program_conf_strs,
		       U32 no_of_strings, struct Program_Conf ** conf_x,
                       PROCESS client)
{
  /* Check the program type */
  const char * value;
  U32 name_len;
  OSUSER usr;
  char user_id_str[PM_HEX_STRING_LEN];
  struct Program_Conf * conf;

  assert(NULL != lm_conf);
  assert(NULL != program_conf_strs);
  assert(NULL != conf_x);
  assert(client != 0);

  conf = heap_alloc_private(sizeof(struct Program_Conf), __FILE__, __LINE__);
  conf->main_proc_name = NULL;
  conf->main_block_name = NULL;

  /* Read the main process name value. */
  value = conf_get_value(lm_conf, program_conf_strs, no_of_strings,
                         OSE_LM_MAIN_NAME, PM_DEFAULT_MAIN_PROC_NAME);
  
  name_len = linelen(value);
  /* Store the configuration */
  conf->main_proc_name = heap_alloc_private(name_len + 1, __FILE__, __LINE__);
  (void)strlcpy(conf->main_proc_name, value, name_len + 1);
  
  /* Read the main block name value. */
  value = conf_get_value(lm_conf, program_conf_strs, no_of_strings,
                         OSE_LM_PROGRAM_NAME, NULL);

  /* Store the configuration */
  if (value != NULL)
    {
      name_len = linelen(value);
      conf->main_block_name = heap_alloc_private(name_len + 1, __FILE__, __LINE__);
      (void)strlcpy(conf->main_block_name, value, name_len + 1);
    }
    
  /* Read the execution model settings. */
  value = conf_get_value(lm_conf, program_conf_strs, no_of_strings,
                         OSE_LM_EXEC_MODEL, "PRIVATE");
  
  name_len = linelen(value);
  
  if(strncmp(value,"SHARED",name_len) == 0)
  {
     conf->shared_model = 1;
  }
  else
  {
     if(strncmp(value,"PRIVATE",name_len) != 0)
     {
        goto pm_read_program_conf_error;
     }
     conf->shared_model = 0;
  }

  /* Decide cpu. */
  value = conf_get_value(lm_conf, program_conf_strs, no_of_strings,
                         OSE_LM_CPU_ID, NULL);

  if (value != NULL)
  {
     conf->cpu = strtoul(value, NULL, 0);
  }
  else
  {
     /* If no cpu configured, inherit it from the current client. */
     conf->cpu = zzkrn_get_pid_info(client, 13);
  }

#ifdef OSE_SMP  
  {
     cpuid_t num_cpus = zzkrn_num_cpus();
          
     if(conf->cpu >= num_cpus)
     {
        ramlog_printf("PGM: cpu_id %u out of range [0..%u] value truncated to %u\n"
                      , conf->cpu, num_cpus - 1, num_cpus - 1);
        conf->cpu = num_cpus - 1;
     }     
  }
#endif    
    
  /* Read the data init value. Should not be in program_conf! */
  if(pm_program_conf_get_value(program_conf_strs, no_of_strings,
			       OSE_LM_DATA_INIT,
			       &value) == True)
    {
       goto pm_read_program_conf_error;
    }

  /* Read the data init value. Should not be in program_conf! */
  if(pm_program_conf_get_value(program_conf_strs, no_of_strings,
			       OSE_LM_BSS_INIT,
			       &value) == True)
    {
       goto pm_read_program_conf_error;
    }

  /* Read the program type value. Should not be in program_conf! */
  if(pm_program_conf_get_value(program_conf_strs, no_of_strings,
			       OSE_LM_PROGRAM_TYPE,
			       &value) == True)
    {
       goto pm_read_program_conf_error;
    }

  name_len = pm_lm_conf_get_value(lm_conf, OSE_LM_PROGRAM_TYPE,
				  &value);
  if(name_len == 0 || name_len >= PM_PROGRAM_TYPE_PARAM_LENGTH)
    {
       goto pm_read_program_conf_error;
    }
  (void)strlcpy(conf->program_type, value, name_len + 1);

  /* Read the rest of the configuration. */

  /* Read the pool size value. */
  value = conf_get_value(lm_conf, program_conf_strs, no_of_strings,
                         OSE_LM_POOL_SIZE, "");
  
  /* Sanity check the value */
  if(pm_check_pool_size_conf(value, &(conf->pool_size)) == False)
    {
       goto pm_read_program_conf_error;
    }

  /* Read the stack sizes value. */
  value = conf_get_value(lm_conf, program_conf_strs, no_of_strings,
                         OSE_LM_STACK_SIZES, "");
  
  /* Sanity check the stack sizes value */
  if(pm_check_buffers_conf(value, conf->stack) == False)
    {
       goto pm_read_program_conf_error;
    }

  /* Read the signal sizes value. */
  value = conf_get_value(lm_conf, program_conf_strs, no_of_strings,
                         OSE_LM_SIGNAL_SIZES, "");
  
  /* Sanity check the signal sizes value */
  if(pm_check_buffers_conf(value, conf->signal) == False)
    {
       goto pm_read_program_conf_error;
    }

  /* Read the pool extension fragment size value.*/
  value = conf_get_value(lm_conf, program_conf_strs, no_of_strings,
                         OSE_LM_POOL_EXT_SIZE, "");

  /* Sanity check the pool extension fragment size value */
  if(pm_check_pool_ext_size_conf(value, &(conf->pool_ext_size)) == False)
  {
    goto pm_read_program_conf_error;
  }

  /* Read the maximum pool size value. */
  value = conf_get_value(lm_conf, program_conf_strs, no_of_strings,
                         OSE_LM_POOL_MAX_SIZE, "");

  /* Sanity check the maximum pool size value */
  if(pm_check_pool_ext_size_conf(value, &(conf->pool_max_size)) == False)
  {
    goto pm_read_program_conf_error;
  }

  /* Read the pool fill value. */
  value = conf_get_value(lm_conf, program_conf_strs, no_of_strings,
                         OSE_LM_POOL_FILL, PM_DEFAULT_POOL_FILL);

  /* Sanity check the pool fill value */
  if(pm_check_boolean_conf(value, &(conf->pool_fill)) == False)
  {
    goto pm_read_program_conf_error;
  }

  /* Read the reentrant value. */
  value = conf_get_value(lm_conf, program_conf_strs, no_of_strings,
                         OSE_LM_REENTRANT, PM_BOOL_NO);
  
  /* Sanity check the reentrant value */
  if(pm_check_boolean_conf(value, &(conf->reent)) == False)
  {
     goto pm_read_program_conf_error;
    }

  /* Read the main_proc_pri value. */
  value = conf_get_value(lm_conf, program_conf_strs, no_of_strings,
                         OSE_LM_MAIN_PRIORITY, PM_DEFAULT_MAIN_PROC_PRIO);

  /* Sanity check the main_proc_pri value */
  if(pm_check_main_proc_pri_conf(value, &(conf->main_proc_pri)) == False)
    {
       goto pm_read_program_conf_error;
    }

  /* Read the main_proc_stack value. */
  value = conf_get_value(lm_conf, program_conf_strs, no_of_strings,
                         OSE_LM_MAIN_STACK_SIZE, PM_DEFAULT_MAIN_PROC_STACK);
  
  /* Sanity check the main_proc_stack value */
  if(pm_check_binary_32_conf(value, &(conf->main_proc_stack)) == False)
    {
       goto pm_read_program_conf_error;
    }

  /* Read the intercepted value. */
  value = conf_get_value(lm_conf, program_conf_strs, no_of_strings,
                         OSE_LM_INTERCEPT, PM_BOOL_NO);
  
  /* Sanity check the reentrant value */
  if(pm_check_boolean_conf(value, &(conf->intercepted)) == False)
    {
       goto pm_read_program_conf_error;
    }

  /* Read the locked to CPU value. */
  value = conf_get_value(lm_conf, program_conf_strs, no_of_strings,
                         OSE_LM_LOCKED_ON_CPU, PM_BOOL_NO);

  if(pm_check_boolean_conf(value, &conf->locked) == False)
  {
     goto pm_read_program_conf_error;
  }

  /* Read the user_id value. */
  value = conf_get_value(lm_conf, program_conf_strs, no_of_strings,
                         OSE_LM_USER_ID, NULL);

  if (value == NULL)
  {
     snprintf(user_id_str, PM_HEX_STRING_LEN, "0x%x", get_uid(client));
     value = user_id_str;
  }

  /* Sanity check the user_id value */
  if(pm_check_binary_32_conf(value, &(conf->user_id)) == False)
    {
       goto pm_read_program_conf_error;
    }

  usr = get_uid(client);

  if(usr != 0 && usr != conf->user_id)
    {
      /* A none super user try to set a user id other than its own user id. */
       goto pm_read_program_conf_error;
    }
  
  *conf_x = conf;
  return True;

 pm_read_program_conf_error:
  /* FIXME: Bad fault detection... */
  /* A configuration error */
  if(conf->main_proc_name != NULL)
    heap_free_private(conf->main_proc_name);
  if(conf->main_block_name != NULL)
    heap_free_private(conf->main_block_name);
  heap_free_private(conf);
  return False;
}

/*
 *===========================================================================
 *                        pm_set_program_configuration
 *===========================================================================
 *
 * Description
 * -----------
 * Read and set the programs configuration as main block environment
 * variables.
 *
 * Requirement
 * -----------
 * "Create program".
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * bid                      The bid of the main block of the program.
 * lm_conf                  The load module configuration.
 * program_conf_strs        The program configuration.
 * no_of_string             The number of strings in the program configuration.
 *
 * Returns
 * -------
 * True/False.
 *
 */

static Boolean
pm_set_program_configuration(PROCESS bid, const char * lm_conf,
                             const char * program_conf_strs,
			     U32 no_of_strings)
{
  unsigned int i;
  const char * current_param_str;

  assert(bid != 0);
  assert(lm_conf != NULL);
  assert(program_conf_strs != NULL);

  /* Check the LM configuration magic. */
  if(lm_conf == NULL ||
     strncmp(lm_conf, CF_CONF_MAGIC, CF_CONF_MAGIC_SIZE) != 0)
    {
      return False;
    }
  assert((strlen(lm_conf) >= CF_CONF_MAGIC_SIZE));
  current_param_str = lm_conf + CF_CONF_MAGIC_SIZE;

  /* Set all LM parameters (overriding monolith inherited environment) */
  while(current_param_str != NULL)
    {
      char * end_of_param;
      end_of_param = strchr(current_param_str, '\n');

      if(*current_param_str != '\n' && *current_param_str != '\0')
	{
	  char * param_buf;
	  U32 param_buf_len;
	  char * equal_char;

	  /* Calculate the lentgh of the parameter + terminator. */
	  if(end_of_param == NULL)
	    {
	      param_buf_len = strlen(current_param_str) + 1;
	    }
	  else
	    {
	      param_buf_len = (U32)(end_of_param - current_param_str) + 1;
	    }

	  /* Copy the parameter to a temp buffer in order to be able
	   * to modify it.
	   */
	  param_buf = heap_alloc_private(param_buf_len, __FILE__, __LINE__);
	  (void)strlcpy(param_buf, current_param_str, param_buf_len);

	  /* Determine the location of the '=' character between the name
	   * and the value of the parameter.
	   */
	  equal_char = strchr(param_buf, '=');

	  /* Make sure the parameter is of correct syntax. */
	  if(equal_char == NULL)
	    {
	      /* Leave the set variables, they will be cleaned up later. */
	      heap_free_private(param_buf);
	      return False;
	    }

	  /* Prepare the parameter for get_env/set_env syntax. */
	  *equal_char = '\0';

	  /* Set the load module parameter as a environment. */
	  (void)set_env(bid, param_buf, equal_char + 1);
	  heap_free_private(param_buf);
	}

      /* Update the current_parm to the next parameter. */
      current_param_str = end_of_param != NULL ? end_of_param + 1 : NULL;
    }

  current_param_str = program_conf_strs;

  /* Set all program parameters (overriding LM parameters) */
  for(i=0; i<no_of_strings; i++)
    {
      char * value = strchr(current_param_str, '=');
      
      if(value == NULL)
	{
	  /* Leave the set variables, they will be cleaned up later. */
	  ramlog_printf("pm_set_program_configuration: No [=] found, current_param_str = %s\n",
			current_param_str);
	  return False;
	}
      
      *value++ = '\0';
      (void)set_env(bid, current_param_str, value);
      *(--value) = '=';
      current_param_str += strlen(current_param_str) + 1;
    }

  return True;
}

/*
 *===========================================================================
 *                        pm_find_program
 *===========================================================================
 *
 * Description
 * -----------
 * Finds the program corresponding to a process, block, or segment pid.
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * PmInternalState                 The PM internal state structure.
 * pid                             A process, block, or segment associated with
 *                                 the program.
 *
 * Returns
 * -------
 * A pointer to the program meta data, or NULL if not found.
 *
 */

static ProgramMetaData *
pm_find_program (struct PmInternalState * state,
                 PROCESS pid)
{
   enum PROCESS_TYPE ptype = get_ptype(pid);
   PROCESS segid;
   
   /* get_ptype() does not identify segment pids. We find the program by
    * assuming that the given pid is a segment pid if get_ptype() returned
    * OS_ILLEGAL. If it was really an illegal pid then we will fail to find the
    * program and return NULL anyway.
    *
    * If the pid was a block or process id then we find the segment containing
    * that pid and search for the corresponding program.
    */
   
   if (ptype == OS_ILLEGAL)
   {
      segid = pid;
   }
   else
   {
      segid = get_segid(pid);
   }
   
   return (ProgramMetaData*)traverseList1(state->programs,
                                          pm_program_search_by_segpid,
                                          (void *)segid);
}

/*
 *===========================================================================
 *                        pm_create_program_request
 *===========================================================================
 *
 * Description
 * -----------
 * Handle a PM_CREATE_PROGRAM_REQUEST signal.
 *
 * Requirement
 * -----------
 * "Create program".
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * PmInternalState                  The PM internal state structure.
 * request                          The PM_CREATE_PROGRAM_REQUEST signal.
 *
 * Returns
 * -------
 * None.
 *
 */

GLOBAL void
pm_create_program_request(struct PmInternalState * PmInternalState,
                          struct PmCreateProgramRequest * request)
{
  /* The status of the create program operation. */
  PmStatus                      status;

  /* The domain number of the create program. */
  OSDOMAIN                      domain = MM_DOMAIN_RESERVED;

  /* The program pid of the create program (segment or block id). */
  PROCESS                       progpid = (PROCESS) 0;

  /* The main block pid of the create program. */
  PROCESS                       main_block = (PROCESS) 0;

  /* The main process pid of the create program. */
  PROCESS                       main_proc;

  /* The master meta data of the load module corresponding to the
   * install handle.
   */
  LmMetaData                  * lm_master;

  /* The meta data of the load module corresponding to the
   * install handle.
   */
  LmRoMetaData                * lm;

  /* The create programs meta data. */
  ProgramMetaData             * prog = NULL;

  /* The create programs program type. */
  ProgramType                 * program_type;

  /* The permission for a program region. */
  U32                           permission;

  /* The cache attributes for a program region. */
  U32                           cache;

  /* The alignment of a create region. */
  U32                           align = 4;

  /* The configuration (and loader directives) of the program to
   * be created.
   */
  struct Program_Conf         * prog_conf = NULL;

  /* The reply signal of the operation. */
  struct PmCreateProgramReply * reply;

  /* The attach signal of the created program. */
  SIGSELECT                   * attachment;

  /* The process type of the main process in the created program. */
  enum PROCESS_TYPE             main_proc_type;

  struct OS_sig_info sig_info;
  PROCESS pool_id;
  U32 kernel_module;
  
  assert(PmInternalState != NULL);
  assert(PmInternalState->programs != NULL);
  assert(PmInternalState->prog_types != NULL);
  assert(PmInternalState->loadmodules != NULL);
  assert(PmInternalState->gmm_pid != 0);
  assert(request != NULL);

  /* Sanity check the signal. */
  pool_id = get_sig_poolid(get_bid(current_process()));
  (void)get_sig_info(&sig_info, pool_id, current_process(),
		     (OSADDRESS)request);

  /* Since conf_size includes one of the bytes in the
   * struct PmCreateProgramRequest a compensation is needed by one byte to the
   * conf_size, in case the conf_size is larger than 1.
   */
  if(((request->conf_size > 0 ? request->conf_size : 1) +
     offsetof(struct PmCreateProgramRequest, conf)) > sig_info.sig_size)
    {
      /* The configuration parameter conf_size is corrupt. */
      status = PM_EPARAM_CONF_SIZE;
      goto pm_create_program_reply;
    }
  else if(pm_strcnt(request->conf, request->conf_size) != request->conf_no_of_strings)
    {
      /* The configuration parameter conf_no_of_strings is corrupt. */
      ramlog_printf(" Wrong number of configuration strings.\n"
		    "Is %d, should be %d [%s]\n",
		    pm_strcnt(request->conf, request->conf_size),
		    request->conf_no_of_strings, request->conf);
      status = PM_EPARAM_CONF_NO_OF_STRINGS;
      goto pm_create_program_reply;
    }
  else if(pm_strllen(request->install_handle, PM_INSTALL_HANDLE_LENGTH) ==
	  PM_INSTALL_HANDLE_LENGTH)
    {
      status = PM_EINSTALL_HANDLE;
      goto pm_create_program_reply;
    }

  /* Search for the load module. */
  lm_master = (LmMetaData *)traverseList1(PmInternalState->loadmodules,
				   pm_search_lm_install_handle,
				   request->install_handle);
  if(lm_master == NULL)
    {
      /* The load module was not found. */
      status = PM_EINSTALL_HANDLE;
      goto pm_create_program_reply;
    }

  lm = lm_master->meta;
  assert(lm != NULL);

  /*
   The load module was found, read and check the configuration.
   The configuration is returned in &prog_conf pointing to a
   Program_Conf struct
  */
  if(pm_read_program_config(lm->configuration, request->conf,
			    request->conf_no_of_strings, &prog_conf,
			    sender((union SIGNAL **)&request)) == False)
    {
       /*FIXME: The error code is not sooo precise,
	 but it will take a major effort...
       */
      /* The configuration was not complete or corrupt. */
      status = PM_EPROGRAM_CONFIGURATION;
       goto pm_create_program_reply;
    }

  kernel_module = (lm->options & PM_LOAD_MODULE_KERNEL);
  if (kernel_module)
  {
     /* Sanity check. */
     if (prog_conf->pool_size != 0)
     {
        ramlog_printf(RAMLOG_PREFIX "(%s) Warning: Parameter "
                      OSE_LM_POOL_SIZE
                      " ignored. Nonzero pool size (0x%08lx) configured for a "
                      "kernel module.\n",
                      prog_conf->main_proc_name, prog_conf->pool_size);
     }
     if (prog_conf->stack[0] != 0)
     {
        ramlog_printf(RAMLOG_PREFIX "(%s) Warning: Parameter "
                      OSE_LM_STACK_SIZES
                      " ignored. Stack sizes configured for a kernel "
                      "module.\n",
                      prog_conf->main_proc_name);
     }
     if (prog_conf->signal[0] != 0)
     {
        ramlog_printf(RAMLOG_PREFIX "(%s) Warning: Parameter "
                      OSE_LM_SIGNAL_SIZES
                      " ignored. Signal sizes configured for a kernel "
                      "module.\n",
                      prog_conf->main_proc_name);
     }
     if (prog_conf->pool_ext_size != 0)
     {
        ramlog_printf(RAMLOG_PREFIX "(%s) Warning: Parameter "
                      OSE_LM_POOL_EXT_SIZE
                      " ignored. Nonzero pool extension fragment size "
                      "(0x%08lx) configured for a kernel module.\n",
                      prog_conf->main_proc_name, prog_conf->pool_ext_size);
     }
     if (prog_conf->pool_max_size != 0)
     {
        ramlog_printf(RAMLOG_PREFIX "(%s) Warning: Parameter "
                      OSE_LM_POOL_MAX_SIZE
                      " ignored. Nonzero maximum pool size (0x%08lx) "
                      "configured for a kernel module.\n",
                      prog_conf->main_proc_name, prog_conf->pool_max_size);
     }
     if (prog_conf->shared_model)
     {
        ramlog_printf(RAMLOG_PREFIX "(%s) Warning: Parameter "
                      OSE_LM_EXEC_MODEL
                      " ignored. Execution mode SHARED is not allowed for a "
                      "kernel module. Using private mode.\n",
                      prog_conf->main_proc_name);
        prog_conf->shared_model = 0;
     }
  }
  
  /* Make an program meta data entry. */
  prog = allocNode(sizeof(ProgramMetaData));

  /* Determine the program type. */
  program_type = (ProgramType *)traverseList1(PmInternalState->prog_types,
					      pm_search_prog_type,
					      prog_conf->program_type);
  if(program_type == NULL)
    {
      /* The program type was not found, can not proceed. */
      status = PM_EILLEGAL_PROGRAM_TYPE;
      goto pm_create_program_reply;
    }

  /* Adjust configuration parameters to inherit mainpool if present */
  if (pm_inherit_mainpool(prog_conf, !program_type->super) == False)
    {
      status = PM_EPROGRAM_CONFIGURATION;
      goto pm_create_program_reply;
    }

  /* Make sure configuration parameters for pool extension are consistent */
  if (!program_type->super && pm_check_pool_extension(prog_conf) == False)
  {
    status = PM_EILLEGAL_POOL_EXTEND_PARAMS;
    goto pm_create_program_reply;
  }

  /* Create the main block. Create an error handler that bypasses the
   * illegal user mode block create in case of plibs.
   * We do want to allow this when using user mode protection.
   */
  {
    OSERRH *old_pr_errhand;
    char *main_block_name;

    old_pr_errhand = create_error_handler(current_process(),
					  pm_user_mode_warning_ignore, 0);

    main_block_name = pm_create_block_name(request->install_handle,
                                           prog_conf->main_block_name);

    main_block = create_block(main_block_name ? 
                                main_block_name 
                                : request->install_handle, 
                              prog_conf->user_id,
			      False, 0, program_type->super);

    pm_destroy_block_name(main_block_name);
    
    if (!kernel_module && !ose_bind_process(main_block, prog_conf->cpu))
    {
       ramlog_printf(RAMLOG_PREFIX "Failed to bind block %s(%#x) to CPU#%u\n",
		     main_block_name ? main_block_name : request->install_handle,
		     main_block, prog_conf->cpu);
    }

    (void) create_error_handler(current_process(), old_pr_errhand, 0);
  }

  if(get_ptype(main_block) == OS_ZOOMBIE)
  {
     for(;;)
     {
        static const struct pm_file_line extra = {__FILE__, __LINE__};
        error2(PM_EINTERNAL, (OSERRCODE)&extra);
     } /*lint !e550 */
  }

  prog->cleanup = 0;
  prog->mainbid = main_block;
  prog->uid = prog_conf->user_id;
  progpid = main_block;
  prog->locked = prog_conf->locked;

  /*
   * Set shared mode for block if configured.
   */
  if(prog_conf->shared_model)
  {
     /* Local, non-public function in crt/src/support/ose_crt_init.c
      * that sets the newly created block in shared mode.
      */
     void ose_set_shared_model(PROCESS bid);

     ose_set_shared_model(main_block);
  }

  if (kernel_module)
  {
     domain = MM_KERNEL_DOMAIN;
  }
  else if(request->domain == PM_NEW_DOMAIN)
  {
     /* Create the domain. */
     if(pm_vs_gmm_create_domain(PmInternalState->gmm_pid,
                                request->install_handle,
                                &domain) != PM_SUCCESS)
     {
        /* Domain creation failed. */
        kill_proc(main_block);
        status = PM_EDOMAIN_CREATION_FAILED;
        goto pm_create_program_reply;
     }
  }
  else if(traverseList1(PmInternalState->programs,
			pm_search_program_by_domain,
			(void *)request->domain) != NULL)
    {
#if 0
       /* FIXME: If the program is loaded in an existing domain, the heap is to
        *        be enherited from the domain it is created in.
        *
        * 1. Create a new block called "pm_spawn".
        * 2. Attach the new block to the domain using attach_segment.
        * 3. Create a new process in the new block called "pm_spawn".
        * 4. Send a signal to the new process requesting a new block in the
        *    domain.
        * 5. Wait for a reply from the process and kill the block.
        * Limit: It is not OK to create a user mode program/block in the
        * system domain.
        */
       domain = request->domain;
#else
       /* Multiple programs in domain not supported yet */
       kill_proc(main_block);
       status = PM_EMULTIPLE_PROGRAMS_IN_DOMAIN;
       goto pm_create_program_reply;
#endif
    }
  else
    {
      /* The domain was not valid. */
      kill_proc(main_block);
      status = PM_EDOMAIN_NOT_FOUND;
      goto pm_create_program_reply;
    }

  prog->domain = domain;
  
  /* Attach the block with the domain. */
  /* FIXME: What happens in the following cases:
   *        1. The kernel is old, i.e. no new segment will be created
   *           if the domain exists. The block is killed, will the kill
   *           clean up the segment as well because of the attach_segment?
   *           If no, this code do not work.
   *        2. Later the creation process fails and the kill_proc(progpid)
   *           call is used to clean up the created execution context, will
   *           other segments in the same domain be killed as well?
   *           If no, this code do not work.
   * NOTE:  Gunnar is the man to ask, do it....
   */
  prog->segpid = attach_segment(main_block, domain);
  
  /* Nail the program to a cpu if required. Use bit mask owned by PM. */
  if(request->domain == PM_NEW_DOMAIN)
    {
      zzkrn_segment_migration_status(prog->segpid, MIGRATION_MASK_PM,
				     prog->locked ? MIGRATION_MASK_PM : 0);
    }

  /* FIXME: move this code to before the create program. */
  /* May not create a user mode program in core. */
  if ((program_type->super != True) && (domain == MM_KERNEL_DOMAIN))
  {
     /* Must be supervisor */
     kill_proc(main_block);
     status = PM_ECORE_PROGRAM_MUST_BE_SUPERVISOR;
     goto pm_create_program_reply;
  }

  if(get_ptype(progpid) == OS_ZOOMBIE)
    {
      /* Compensate for OSEhq08645... */
      if(domain == MM_KERNEL_DOMAIN)
	{
	  status = PM_EMULTIPLE_PROGRAMS_IN_DOMAIN;
          goto pm_create_program_reply;
	}

      for(;;)
	{
	  static const struct pm_file_line extra = {__FILE__, __LINE__};
	  error2(PM_EINTERNAL, (OSERRCODE)&extra);
	} /*lint !e550 */
    }

  prog->progpid = progpid;

  /* Allocate DATA. */
  if(lm->data.options != PM_REGION_ALIEN)
  {
    ProgramMetaData * program_created;
    OSADDRESS data_base = lm->data.base;
    U32 data_size = lm->data.size;
    mm_status_t mm_status = ~MM_SUCCESS;
            
    permission = lm->data_conf.permissions;
    cache = lm->data_conf.cache_mode;

    program_created =
      (ProgramMetaData *)traverseList1(PmInternalState->programs,
                                       pm_search_program_by_install_handle,
                                       request->install_handle);
    if(program_created)
      {
	/* Data must be mas to get the exact specified base address. */
        mm_status =
            pm_vs_gmm_create_region(PmInternalState->gmm_pid,
                                    LM_DATA,
                                    domain,
                                    &data_size,
                                    &data_base,
                                    lm->data_conf.phys_mem,
                                    lm->data_conf.log_mem,
                                    permission,
                                    cache,
                                    &align,
                                    MM_CREATE_REGION_SPECIFY_LOG
                                    | MM_CREATE_REGION_SPECIFY_NAME);
        if ((MM_SUCCESS != mm_status) || (data_base != lm->data.base))         
          {
	    /* The region is not mas, failure. */
            if (MM_SUCCESS == mm_status)
              {
                (void)pm_vs_gmm_delete_region(PmInternalState->gmm_pid,
                                              domain, data_base);
              }            
	    kill_proc(main_block);
	    if(request->domain == PM_NEW_DOMAIN && !kernel_module)
	      {
		(void) pm_vs_gmm_delete_domain(PmInternalState->gmm_pid,
                                               domain);
	      }
	    status = PM_EINSTALL_HANDLE_IN_USE;
            goto pm_create_program_reply;
	  }

	/* Data is created. */
      }
    else
      {
         struct mm_create_region *get_reply;

         /* The status of the create region operation. */
         mm_status = pm_vs_gmm_get_region_info(PmInternalState->gmm_pid,
                                               MM_KERNEL_DOMAIN,
                                               data_base,
                                               &get_reply);

         if (mm_status == MM_ENON_EXISTING_REGION)
         {
            
            /* Region does not exist. A valid reason is that
             * the data region will reside in MAS mapped memory.
             * The reservation of logical MAS memory is removed by the RTL
             * after installation to allow other modules to "reserve" the
             * same logical space for their data region.
             *
             * Create new MAS region with phys mem in wanted domain.
             */
            mm_status =
               pm_vs_gmm_create_region(PmInternalState->gmm_pid,
                                       LM_DATA,
                                       domain,
                                       &data_size,
                                       &data_base,
                                       lm->data_conf.phys_mem,
                                       lm->data_conf.log_mem,
                                       permission,
                                       cache,
                                       &align,
                                       MM_CREATE_REGION_SPECIFY_LOG
                                       | MM_CREATE_REGION_SPECIFY_NAME);
            /* Data is created. */
         }
         else if (mm_status == MM_SUCCESS &&
                  !(get_reply->options & MM_CREATE_REGION_LOGICAL))
         {
            /* Region already exists. */
            if (domain != MM_KERNEL_DOMAIN)
            {
               mm_status = pm_vs_gmm_share_region(PmInternalState->gmm_pid,
                                                  lm->data.base,
                                                  MM_KERNEL_DOMAIN,
                                                  domain,
                                                  MM_SHARE_REGION_UPERM,
                                                  permission);
            }
         }
         else if (mm_status == MM_SUCCESS && domain == MM_KERNEL_DOMAIN)
         {
               
            /* Region is logical only. It is located in the system
             * domain, so a new region can't be created. Add physical
             * resource to existing region.
             */
               
            mm_status =
               pm_vs_gmm_add_phys_resource(PmInternalState->gmm_pid,
                                           MM_KERNEL_DOMAIN,
                                           get_reply->base,
                                           lm->data_conf.phys_mem);
            prog->cleanup |= PROG_CLEANUP_FREE_PHYS_DATA_RESOURCE;
         }
         else if (mm_status == MM_SUCCESS)
         {
               
            /* Region is logical only, create SAS region with phys mem
             * in wanted domain.
             */
               
            mm_status =
               pm_vs_gmm_create_region(PmInternalState->gmm_pid,
                                       LM_DATA,
                                       domain,
                                       &data_size,
                                       &data_base,
                                       lm->data_conf.phys_mem,
                                       lm->data_conf.log_mem,
                                       permission,
                                       cache,
                                       &align,
                                       MM_CREATE_REGION_SPECIFY_LOG
                                       | MM_CREATE_REGION_SPECIFY_NAME);
            /* Data is created. */
         }
         
         free_buf((union SIGNAL **)&get_reply);
         
         /* Common cleanup check for above gmm calls. */
         if(mm_status != MM_SUCCESS)
	  {
             /* Error occured. */
	    kill_proc(main_block);
	    if(request->domain == PM_NEW_DOMAIN)
	      {
               (void) pm_vs_gmm_delete_domain(PmInternalState->gmm_pid,
                                              domain);
	      }
	    status = PM_EREGION_CREATION_FAILED;
            goto pm_create_program_reply;
	  }

	/* The Data region is a SASE region, and is shared to the
         * program domain.
         */
      }

    assert(mm_status == MM_SUCCESS ? data_base == lm->data.base : 1);
    assert(data_size >= lm->data.size);
  }

  /* OSEhq09509, added the "domain != MM_KERNEL_DOMAIN" sanity check. */
  if(lm->text.options != PM_REGION_ALIEN && domain != MM_KERNEL_DOMAIN)
    {
      /* OSEhq09509, Added mm_stat variable. */
      mm_status_t mm_stat;

      permission = lm->text_conf.permissions;

      /* OSEhq09509, the share region return code is not boolean, change this
       * to proper values.
       */
      mm_stat = pm_vs_gmm_share_region(PmInternalState->gmm_pid, lm->text.base,
				       MM_KERNEL_DOMAIN, domain,
                                       MM_SHARE_REGION_UPERM, permission);

      if(mm_stat != MM_SUCCESS)
	{
	  /* Import region failed, cleanup. */
	  kill_proc(progpid);

	  /* Note delete domain will remove the regions inside the
	   * domain (phys data and pool).
	   */
	  if(request->domain == PM_NEW_DOMAIN)
	    {
	      (void) pm_vs_gmm_delete_domain(PmInternalState->gmm_pid, domain);
	    }
	  status = PM_EREGION_CREATION_FAILED;
          goto pm_create_program_reply;
	}
    }

  if(!kernel_module)
  {
     /* Allocate the pool */

     if(prog_conf->pool_size != 0)
     {
        OSADDRESS pool_base = 0;
        U32 pool_size = prog_conf->pool_size;
        
        permission = program_type->regions[PM_POOL_REGION_TYPE].permissions;
        cache = program_type->regions[PM_POOL_REGION_TYPE].cache_mode;
        
        /* Allocate the pool. */
        align = 4;
        if(pm_vs_gmm_create_region(PmInternalState->gmm_pid,
                                   PM_REGION_POOL,
                                   domain,
                                   &pool_size,
                                   &pool_base,
                                   program_type->regions[PM_POOL_REGION_TYPE].phys_mem,
                                   program_type->regions[PM_POOL_REGION_TYPE].log_mem,
                                   permission,
                                   cache,
                                   &align,
                                   MM_CREATE_REGION_SPECIFY_ALIGN
                                   | MM_CREATE_REGION_SPECIFY_NAME) != MM_SUCCESS)
	{
           /* The creation of the pool region failed. */
           kill_proc(progpid);
           
           /* Note delete domain will remove the regions inside the
            * domain (phys data and pool).
            */
           if(request->domain == PM_NEW_DOMAIN)
           {
	      (void) pm_vs_gmm_delete_domain(PmInternalState->gmm_pid, domain);
           }
           
           status = PM_EREGION_CREATION_FAILED;
           goto pm_create_program_reply;
	}
        
        assert(pool_base != 0);
        assert(pool_size >= prog_conf->pool_size);
        assert(align >= 4);

        /* Create the Pool: */
        if(prog_conf->pool_max_size != 0)
        {
          /* Extendible pool is requested: */
          if(program_type->super != True)
          {
            /* User mode program, create extendible pool. */
            zzkrn_create_program_pool(main_block,
                                      pool_base,
                                      prog_conf->pool_size,
                                      prog_conf->signal,
                                      prog_conf->stack,
                                      prog_conf->pool_fill,
                                      pm_pool_extender,
                                      prog_conf->pool_ext_size,
                                      prog_conf->pool_max_size,
                                      (void *)program_type);
          }
          else
          {
            /* Attempted to create an extendible pool in supervisor mode,
             * this causes program creation to fail. 
             */
            kill_proc(progpid);
            
            /* Note delete domain will remove the regions inside the
             * domain (phys data and pool).
             */
            if(request->domain == PM_NEW_DOMAIN)
            {
              (void) pm_vs_gmm_delete_domain(PmInternalState->gmm_pid, domain);
            }

            status = PM_EILLEGAL_POOL_EXTEND_PARAMS;
            goto pm_create_program_reply;
          }
        }
        else
        {
          /* No pool extend requested, create a static pool. */
          zzkrn_create_program_pool(main_block,
                                    pool_base,
                                    prog_conf->pool_size,
                                    prog_conf->signal,
                                    prog_conf->stack,
                                    prog_conf->pool_fill,
                                    NULL,
                                    0,
                                    0,
                                    NULL);
        }
     }
     else if(request->domain != PM_NEW_DOMAIN)
     {
        /* The feature of a local domain pool is not supported
         * in this version of PM. */
        kill_proc(progpid);
        
        /* Note delete domain will remove the regions inside the
         * domain (phys data and pool).
         */
        if(request->domain == PM_NEW_DOMAIN)
	{
           (void) pm_vs_gmm_delete_domain(PmInternalState->gmm_pid, domain);
	}
        status = PM_EFEATURE_NOT_SUPPORTED;
        goto pm_create_program_reply;
     }
  }
  
  prog->sig_poolid = get_sig_poolid(main_block);
  prog->stk_poolid = get_stk_poolid(main_block);
  
  {
    struct OS_poolcb *poolcb;
    
    poolcb = get_poolcb(prog->sig_poolid);
    prog->sig_pool_base = poolcb->values[poolcb->fragment_info+1];
    prog->sig_pool_size = poolcb->values[poolcb->fragment_info+2];
    free_buf((union SIGNAL**)&poolcb);
    
    poolcb = get_poolcb(prog->stk_poolid);
    prog->stk_pool_base = poolcb->values[poolcb->fragment_info+1];
    prog->stk_pool_size = poolcb->values[poolcb->fragment_info+2];
    free_buf((union SIGNAL**)&poolcb);
  }
  
  /* Create a program alias if necessery. */
  if(prog_conf->main_block_name)
  {
     assign_linkhandler(prog_conf->main_block_name, main_block);
  }

  if(prog_conf->main_proc_pri == PM_BACKGROUND_PROCESS)
    {
      main_proc_type = OS_BG_PROC;

      /* 32 in not an OSE accepted background process priority. */
      prog_conf->main_proc_pri = 0;
    }
  else
    {
      main_proc_type = OS_PRI_PROC;
    }

  /* Create the main process. */
  main_proc = create_process(main_proc_type, 
                             prog_conf->main_proc_name,
			     (void (*) (void))lm->entrypoint,
			     prog_conf->main_proc_stack,
			     prog_conf->main_proc_pri, 0, main_block, 0, 0,
			     prog_conf->user_id);

  if(get_ptype(main_proc) == OS_ZOOMBIE)
    {
      for(;;)
	{
	  static const struct pm_file_line extra = {__FILE__, __LINE__};
	  error2(PM_EINTERNAL, (OSERRCODE)&extra);
	} /*lint !e550 */
    }

  prog->mainpid = main_proc;

  if (!kernel_module)
  {
     /*
      * Prepare for creation of the heap. It will not yet be created.
      */
     
     /* Remove all old heap configuration for the block. */
     if(request->domain == PM_NEW_DOMAIN || prog_conf->shared_model)
     {
        ose_reset_heap(main_block);
     }
     
     /* Set up env vars for LM_HEAP_LOG/PHYS_MEM if configured for
        this program type. They will possibly get overridden by
        pm_set_program_configuration() below
     */
     if(program_type->no_of_regions > PM_HEAP_REGION_TYPE)
     {
        char val_str[13]; /* Fit a long in hex 0x12345678 */
        set_env(main_block, HEAP_PHYS_MEM,
                program_type->regions[PM_HEAP_REGION_TYPE].phys_mem);
        
        set_env(main_block, HEAP_LOG_MEM,
                program_type->regions[PM_HEAP_REGION_TYPE].log_mem);
        
        snprintf(val_str, sizeof(val_str),"0x%lx",
                 program_type->regions[PM_HEAP_REGION_TYPE].permissions);
        set_env(main_block, HEAP_PERMISSIONS,val_str);
        
        snprintf(val_str, sizeof(val_str),"0x%lx",
                 program_type->regions[PM_HEAP_REGION_TYPE].cache_mode);
        set_env(main_block, HEAP_CACHE_MODE, val_str);
     }
  }
  
  
  /* Setup the program parameters as environment variables. */
  if(pm_set_program_configuration(main_block, lm->configuration,
                                  request->conf,
                                  request->conf_no_of_strings) != True)
  {
     /* The setting of env vars failed. */
     kill_proc(progpid);
     
     /* Note delete domain will remove the regions inside the
      * domain (phys data and pool).
      */
     if(request->domain == PM_NEW_DOMAIN)
     {
        (void) pm_vs_gmm_delete_domain(PmInternalState->gmm_pid, domain);
     }
     /* FIXME: The error code is not sooo precise,
      * but it will take a major effort...
      */
     status = PM_EPROGRAM_CONFIGURATION;
     goto pm_create_program_reply;
  }
  

  /* Setup heap phys_mem configuration as a last fallback if both
   * lm configuration as well as program type config lacks it. */
  if (!kernel_module)
  {
     char *heap_phys_mem = get_env(main_block, HEAP_PHYS_MEM);
     if (heap_phys_mem == NULL)
     {
        char *rtl_text = get_env(main_block, RTL_PHYS_MEM_TEXT_VAR);
        if (rtl_text != NULL)
        {
           set_env(main_block, HEAP_PHYS_MEM, rtl_text);
           free_buf((union SIGNAL**)&rtl_text);
        }
     }
     else
     {
        free_buf((union SIGNAL**)&heap_phys_mem);
     }
  }


  if (!kernel_module)
  {
     /* Create a new heap if necessery. */
     if(request->domain == PM_NEW_DOMAIN || prog_conf->shared_model)
     {
        prog->heap = ose_create_heap(main_block, domain, program_type->super);
        
        if(prog->heap == 0)
        {
           /* The creation of the heap failed. */
           kill_proc(progpid);
           
           /* Note delete domain will remove the regions inside the
            * domain (phys data and pool).
            */
           
           (void) pm_vs_gmm_delete_domain(PmInternalState->gmm_pid, domain);
           
           status = PM_EHEAP_CREATION_FAILED;
           goto pm_create_program_reply;
        }
     }
  }
  else
  {
     prog->heap = 0;
  }
  
  /* Intercept the created program (for debugging purpuses). */
  if(prog_conf->intercepted)
  {
     zzkrn_dbg_intercept(progpid);
  }

  /* Attach to the program. */
  attachment = (SIGSELECT *) alloc(sizeof(SIGSELECT),
				   PM_INTERNAL_PROGRAM_KILLED);

  (void)attach((union SIGNAL **)&attachment, main_proc);
  prog->attached = NULL;

  /* Store the load module information in the program meta structure. */
  prog->lm = lm;

  /* Set the program state. */
  prog->state = PM_PROGRAM_CREATED;

  /* Increase the number of programs counter. */
  PmInternalState->no_of_programs++;

  /* For debugging purposes, send the load module address to simics. */
  SIMICS_MAGIC(lm->text.base, lm->parent_file_name, lm->data.base);

  /* Store the program. NOTE: The locks are used to avoid a race with
   * pm_init_program_data, the list package is not tread safe. */
  ose_mutex_lock(&(PmInternalState->mutex));
  insertFirst(PmInternalState->programs, (void *)prog);
  ose_mutex_unlock(&(PmInternalState->mutex));

  status = PM_SUCCESS;

 pm_create_program_reply:

  if (prog_conf)
    {
      if (prog_conf->main_proc_name)
        {
          heap_free_private(prog_conf->main_proc_name);
        }
      if (prog_conf->main_block_name)
        {
          heap_free_private(prog_conf->main_block_name);
        }
      heap_free_private(prog_conf);
    }  
 
  if (status != PM_SUCCESS)
    {
      if(prog != NULL)
	freeNode(prog);
    }

  reply = (void *) alloc(sizeof(struct PmCreateProgramReply),
                         PM_CREATE_PROGRAM_REPLY);
  reply->status = status;
  reply->progpid = progpid;
  reply->domain = domain;
  reply->main_bid = main_block;
  (void)strlcpy(reply->install_handle, request->install_handle,
                PM_INSTALL_HANDLE_LENGTH);
  send((union SIGNAL **) &reply, sender((union SIGNAL **) &request));
  free_buf((union SIGNAL **) &request);
}


/*
 *===========================================================================
 *                        pm_program_killed
 *===========================================================================
 *
 * Description
 * -----------
 * Handle a killed program.
 *
 * Requirement
 * -----------
 * "Kill program".
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * PmInternalState                  The PM internal state structure.
 * request                          The attach signal of the killed program.
 *
 * Returns
 * -------
 * None.
 *
 */

GLOBAL void
pm_program_killed(struct PmInternalState * PmInternalState,
                  SIGSELECT * request)
{
  PROCESS zombie_main;
  ProgramMetaData* zombie_node;
  PmStatus status = PM_SUCCESS;
  Pm_Attach * attachment;

  assert(PmInternalState != NULL);
  assert(PmInternalState->programs != NULL);
  assert(PmInternalState->gmm_pid != 0);
  assert(request != NULL);

  zombie_main = sender((union SIGNAL **)&request);

  /*Find the program meta data by using the main pid.*/
  zombie_node  = (ProgramMetaData*)traverseList1(PmInternalState->programs,
						 pm_program_search_by_mainpid,
						 (void *)zombie_main);

  if(zombie_node == NULL)
    {
      for(;;)
	{
	  static const struct pm_file_line extra = {__FILE__, __LINE__};
	  error2(PM_EINTERNAL, (OSERRCODE)&extra);
	} /*lint !e550 */
    }

  /*
   * Do a synchronous block kill
   */
  {
     static const SIGSELECT attsel[] = {1, PM_INTERNAL_OSE_SYSD_SYNC};
     union SIGNAL *sig;

     sig = alloc(sizeof(SIGSELECT), PM_INTERNAL_OSE_SYSD_SYNC);
     (void) attach(&sig, zombie_node->segpid);
     
     kill_proc(zombie_node->progpid);

     /* Wait until ose_sysd is finished killing the program. */
     sig = receive(attsel);

     free_buf(&sig);
  }

  if (zombie_node->domain == MM_KERNEL_DOMAIN)
  {
     if (zombie_node->cleanup & PROG_CLEANUP_FREE_PHYS_DATA_RESOURCE)
     {
        if(MM_SUCCESS !=
           pm_vs_gmm_remove_phys_resource(PmInternalState->gmm_pid,
                                          MM_KERNEL_DOMAIN,
                                          zombie_node->lm->data.base))
        {
           for (;;)
           {
              static const struct pm_file_line extra = {__FILE__, __LINE__};
              error2(PM_EINTERNAL, (OSERRCODE)&extra);
           }
        }
     }
  }
  else
  {
     /*Delete program domain.*/
     status = pm_vs_gmm_delete_domain(PmInternalState->gmm_pid,
                                      zombie_node->domain);
     
     if(status != PM_SUCCESS)
     {
        for(;;)
	{
           static const struct pm_file_line extra = {__FILE__, __LINE__};
           error2(PM_EINTERNAL, (OSERRCODE)&extra);
	} /*lint !e550 */
     }
  }
  
  /* Send attach signals. */
  attachment = zombie_node->attached;
  while(attachment != NULL)
    {
      struct PmAttachProgramNotification * notification;
      Pm_Attach * handled_attachment;

      notification = (struct PmAttachProgramNotification *)
	alloc(sizeof(struct PmAttachProgramNotification) + attachment->data_len,
	      attachment->sigNo);
      notification->progpid = zombie_node->progpid;
      memcpy(notification->data, attachment->data, attachment->data_len);
      send((union SIGNAL **) &notification,
	   attachment->client);
      handled_attachment = attachment;
      attachment = attachment->next;
      heap_free_private(handled_attachment);
    }

  /* Decrease the number of programs counter. */
  assert(PmInternalState->no_of_programs != 0);
  PmInternalState->no_of_programs--;
  
  /* Job done, free the node. The zombie is no more.*/
  /* Note: The locks are used to avoid a race with pm_init_program_data. */
  ose_mutex_lock(&(PmInternalState->mutex));
  unlinkNode(zombie_node);
  ose_mutex_unlock(&(PmInternalState->mutex));

  freeNode(zombie_node);

  free_buf((union SIGNAL **)&request);
}


/*
 *===========================================================================
 *                        pm_no_of_programs
 *===========================================================================
 *
 * Description
 * -----------
 * Count the number of programs created from a specific load module.
 *
 * Requirement
 * -----------
 * "Program information".
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * PmInternalState                  The PM internal state structure.
 * install_handle                   The install handle of the load module.
 *
 * Returns
 * -------
 * The number of programs created from the load module.
 *
 */

GLOBAL U32
pm_no_of_programs(struct PmInternalState * pmInternalState,
		  char * install_handle)
{
   struct Link *link = firstNode(pmInternalState->programs);
   U32 count = 0;

   for (; link != NULL; link = nextNode(link))
   {
      void *result = pm_search_program_by_install_handle(link, install_handle);

      if (NULL != result)
      {
         count++;
      }
   }

   return count;
}


/*
 *===========================================================================
 *                        pm_program_info_request
 *===========================================================================
 *
 * Description
 * -----------
 * Handle a PM_PROGRAM_INFO_REQUEST signal.
 *
 * Requirement
 * -----------
 *
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * state                  The PM internal state structure.
 * request                The request signal.
 *
 * Returns
 * -------
 * None.
 *
 */

GLOBAL void
pm_program_info_request (struct PmInternalState * state,
                         struct PmProgramInfo * request)
{
   const ProgramMetaData * prog;
   struct PmProgramInfo * reply;
   PROCESS prog_pid;
   PROCESS sending_proc;

   assert (state != NULL);
   assert (request != NULL);
   assert (state->programs != NULL);

   /* Find out program PID and the PID of the client. */
   prog_pid = request->progpid;
   sending_proc = sender((union SIGNAL **) &request);
   
   /*
    * Because the client might have been compiled against a version
    * of pm.sig where struct PmProgramInfo was smaller, we now have to free
    * the request signal and re-allocate a new that will match the current
    * implementation.
    */
   free_buf((union SIGNAL **) &request);
   reply = (struct PmProgramInfo *) alloc(sizeof(struct PmProgramInfo),
                                          PM_PROGRAM_INFO_REPLY);

   
   /* Find the program meta data by using the progpid. */
   prog = pm_find_program (state, prog_pid);
   
   /* Check if found */
   if(NULL == prog)
   {
      reply->status = PM_EPROGPID;
      goto pm_program_info_reply;
   }

   assert (NULL != prog->lm);

   /* Copy relevant information */
   reply->progpid      = prog_pid;
   reply->domain       = prog->domain;
   reply->main_block   = prog->mainbid;
   reply->main_process = prog->mainpid;
   reply->pool_base    = prog->stk_pool_base;
   reply->pool_size    = prog->stk_pool_size;
   reply->uid          = prog->uid;
   reply->state        = prog->state;
   reply->segpid       = prog->segpid;
   reply->stk_poolid   = prog->stk_poolid;
   reply->sig_poolid   = prog->sig_poolid;
   reply->sig_pool_base= prog->sig_pool_base;
   reply->sig_pool_size= prog->sig_pool_size;
   reply->heap         = prog->heap;

   assert (pm_strllen(prog->lm->install_handle,
		      PM_INSTALL_HANDLE_LENGTH) < PM_INSTALL_HANDLE_LENGTH);

   strcpy (reply->install_handle, prog->lm->install_handle);

   /* Everything went well */
   reply->status = PM_SUCCESS;

  pm_program_info_reply:
   send ((union SIGNAL **) &reply, sending_proc);
}


/*
 *===========================================================================
 *                        pm_program_configuration_request
 *===========================================================================
 *
 * Description
 * -----------
 * Handle a PM_PROGRAM_CONFIGURATION_REQUEST signal.
 *
 * Requirement
 * -----------
 *
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * state                  The PM internal state structure.
 * request                The request signal.
 *
 * Returns
 * -------
 * None.
 *
 */

extern Boolean efs_binary_env(const char *str);

GLOBAL void
pm_program_configuration_request ( struct PmInternalState * state,
                                   struct PmProgramConfigurationRequest *
                                   request)
{
   const ProgramMetaData * prog;
   struct PmProgramConfigurationReply * reply;
   PmStatus status = PM_SUCCESS;
   U32 conf_no_of_strings = 0;
   S32 conf_size = 0;
   char * last = NULL;
   char * free_env_list = NULL;

#define PTR_VAL_STR_LEN 24 /* "1234567890 (0x12345678)" =  23 + \0 */
   char ptr_val_str[PTR_VAL_STR_LEN]= "";

   assert (state != NULL);
   assert (request != NULL);
   assert (state->programs !=NULL);

   /* Find the program meta data by using the progpid. */
   prog = pm_find_program (state, request->progpid);
   
   /* Check if found */
   if(NULL == prog)
   {
      status = PM_EPROGPID;
      goto pm_program_configuration_reply;
   }

   /*
    * First traverse the list to get the size of the configuration
    */ 
   for(;;)
     {
       char terminator;
       char * curr_env_name;
       char * curr_env_value;
       char * env_list;

       env_list = get_env_list(prog->mainbid, last);

       if(free_env_list != NULL)
	 {
	   free_buf((union SIGNAL **) &free_env_list);
	 }

       /* If list is empty, get out */
       if(env_list[0] == 0)
	 {
	   free_buf((union SIGNAL **) &env_list);
	   break;
	 }

       /* Start at the beginning of the list */
       curr_env_name = env_list;
       while(curr_env_name != NULL)
	 {
	   last = strchr(curr_env_name, ' ');
	   if(last == NULL)
	     {
	       last = curr_env_name + strlen(curr_env_name);
	     }
	   terminator = *last;
	   *last = '\0';

	   /* Get the value associated with this name */
	   curr_env_value = get_env(prog->mainbid, curr_env_name);
	   if(curr_env_value == NULL)
	     {
                goto pm_program_configuration_search_next;
	     }

	   /* Add the name and the equal char to the size. */
	   conf_size += strlen(curr_env_name) + 1;

	   /* Convert to ASCII if value is of pointer type */
	   if (efs_binary_env(curr_env_value))
	   {
	      U32 bin_val = get_envp(prog->mainbid, curr_env_name);
	      snprintf(ptr_val_str, PTR_VAL_STR_LEN,
		       "%lu (0x%lx)", bin_val, bin_val);

	      /* Add the value  and the string terminator char to the size. */
	      conf_size += (strlen(ptr_val_str) + 1);
	   }
	   else
	   { /* Value is of string typ */
	      /* Add the value  and the string terminator char to the size. */
	      conf_size += (strlen(curr_env_value) + 1);
	   }
	   
	   free_buf((union SIGNAL **) &curr_env_value);

	   conf_no_of_strings++;

        pm_program_configuration_search_next:
	   /* Search the next name. */
	   if(terminator != '\0')
	     {
	       curr_env_name = last + 1;
	     }
	   else
	     {
	       last = curr_env_name;
	       curr_env_name = NULL;
	       free_env_list = env_list;
	     }
	 }
     }
 pm_program_configuration_reply:
   assert(conf_size >= 0);
   reply = (struct PmProgramConfigurationReply *)
     alloc (sizeof (struct PmProgramConfigurationReply) + conf_size,
	    PM_PROGRAM_CONFIGURATION_REPLY);

   reply->progpid = request->progpid;
   reply->conf_size = (U32)conf_size;
   reply->conf_no_of_strings = conf_no_of_strings;


   if (PM_SUCCESS == status)
   {
     char * pos = reply->conf;
     last = NULL;
     free_env_list = NULL;

   /*
    * Now traverse the list again to copy the configuration
    */ 
     for(;;)
       {
	 char terminator;
	 char * curr_env_name;
	 char * curr_env_value;
	 char * env_list;

	 env_list = get_env_list(prog->mainbid, last);

	 if(free_env_list != NULL)
	   {
	     free_buf((union SIGNAL **) &free_env_list);
	   }

	 if(((SIGSELECT)env_list[0]) == 0)
	   {
	     free_buf((union SIGNAL **) &env_list);
	     break;
	   }

	 curr_env_name = env_list;
	 while(curr_env_name != NULL)
	   {
	     last = strchr(curr_env_name, ' ');
	     if(last == NULL)
	       {
		 last = curr_env_name + strlen(curr_env_name);
	       }
	     terminator = *last;
	     *last = '\0';

	     curr_env_value = get_env(prog->mainbid, curr_env_name);
	     if(curr_env_value == NULL)
	         {
                    goto pm_program_configuration_search_next_2;
		 }

	     /* Subtract the name and the equal char to the size  */
	     conf_size -= (strlen(curr_env_name) + 1);
	     if(conf_size < 0)
	       {
		 /* Abort, the confsize has increased (new block variables
		  * were added during call).
		  */
		/*FIXME: PM_EPROGRAM_CONFIGURATION is misleading...*/
		  status = PM_EPROGRAM_CONFIGURATION;
		 goto pm_program_configuration_reply_error;
	       }

	     /* Copy env name and '=' to signal buffer */
	     strcpy(pos, curr_env_name);
	     pos += strlen(curr_env_name);
	     strcat(pos++, "=");

	     /* Convert to ASCII if value is of pointer type */
	     if (efs_binary_env(curr_env_value))
             {
		U32 bin_val = get_envp(prog->mainbid, curr_env_name);
		snprintf(ptr_val_str, PTR_VAL_STR_LEN,
			 "%lu (0x%lx)", bin_val, bin_val);
		
		/* Subtract the value  and the string terminator char to the size. */
		conf_size -= (strlen(ptr_val_str) + 1);

		if(conf_size < 0)
		{
		   /* Abort, the confsize has increased (new block variables
		    * were added during call).
		    */
		   /*FIXME: PM_EPROGRAM_CONFIGURATION is misleading...*/
		   status = PM_EPROGRAM_CONFIGURATION;
		   goto pm_program_configuration_reply_error;
		}
		
		strcpy(pos, ptr_val_str);
		
		/* Set the possition to after the terminator of the value. */
		pos += strlen(ptr_val_str) + 1;
	     }
	     else
	     { /* Value is of string type */

		/* Subtract the value  and the string terminator char to the size. */
		conf_size -= (strlen(curr_env_value)+ 1);
		
		if(conf_size < 0)
		{
		   /* Abort, the confsize has increased (new block variables
		    * were added during call).
		  */
		/*FIXME: PM_EPROGRAM_CONFIGURATION is misleading...*/
		 status = PM_EPROGRAM_CONFIGURATION;
		 goto pm_program_configuration_reply_error;
	       }

	     strcpy(pos, curr_env_value);

	     /* Set the possition to after the terminator of the value. */
	     pos += strlen(curr_env_value) + 1;
	     }
	     

	     free_buf((union SIGNAL **) &curr_env_value);

         pm_program_configuration_search_next_2:
	     /* Search the next name. */
	     if(terminator != '\0')
	       {
		 curr_env_name = last + 1;
	       }
	     else
	       {
		 last = curr_env_name;
		 curr_env_name = NULL;
		 free_env_list = env_list;
	       }
	   }
       }
   }
   goto pm_program_configuration_conf_finished;

 pm_program_configuration_reply_error:
   reply->conf_no_of_strings = 0;
   reply->conf_size = 0;

 pm_program_configuration_conf_finished:
   reply->status = status;
   send ((union SIGNAL **) &reply, sender ((union SIGNAL **) &request));
   free_buf((union SIGNAL **) &request);
}

/*
 *===========================================================================
 *                        pm_programs_request
 *===========================================================================
 *
 * Description
 * -----------
 * Handle a PM_PROGRAMS_REQUEST signal.
 *
 * Requirement
 * -----------
 *
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * state                  The PM internal state structure.
 * request                The request signal.
 *
 * Returns
 * -------
 * None.
 *
 */

GLOBAL void
pm_programs_request ( struct PmInternalState * state,
                      struct PmProgramsRequest * request )
{
   U32 nr_programs;
   struct PmProgramsReply * reply;
   const ProgramMetaData * prog;
   U32 progndx;

   assert(state != NULL);
   assert(request != NULL);
   assert(state->programs !=NULL);

   /* Determine the number of created programs */
   nr_programs = howManyNodes (state->programs);

   /* Allocate space for reply signal */
   reply = (struct PmProgramsReply *)
      alloc (sizeof (struct PmProgramsReply)
	     + (sizeof(PROCESS) * (nr_programs != 0 ? nr_programs - 1 : 0)),
             PM_PROGRAMS_REPLY);

   /*
    * Fill in reply progpids array
    */
   prog = (const ProgramMetaData *)
      firstNode (state->programs);

   progndx = 0;

   while (prog != NULL)
   {
      reply->progpids[progndx++] = prog->progpid;

      prog = nextNode ((const void *) prog);
   }

   /*
    * Fill in some additional bookkeeping
    */

   reply->progpids_size = nr_programs;
   reply->status = PM_SUCCESS;

   /*
    * Send reply and free request signals
    */
   send ((union SIGNAL **) &reply, sender ((union SIGNAL **) &request));
   free_buf((union SIGNAL **) &request);
}

/*
 *===========================================================================
 *                        pm_get_program_pid_request
 *===========================================================================
 *
 * Description
 * -----------
 * Handle a PM_GET_PROGRAM_PID_REQUEST signal.
 *
 * Requirement
 * -----------
 *
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * state                  The PM internal state structure.
 * request                The request signal.
 *
 * Returns
 * -------
 * None.
 *
 */

GLOBAL void
pm_get_program_pid_request (struct PmInternalState * state,
                            struct PmGetProgramPid * request )
{
   const ProgramMetaData * prog;

   assert(request != NULL);

   /* Find the program meta data */
   prog = pm_find_program (state, request->pid);
   
   if(prog != NULL)
   {
      request->status = PM_SUCCESS;
      request->progpid = prog->progpid;
   }
   else
   {
      request->status = PM_EPROGPID;
   }

   /*
    * Send reply(reuse the signal).
    */
   request->sigNo = PM_GET_PROGRAM_PID_REPLY;
   send ((union SIGNAL **) &request, sender ((union SIGNAL **) &request));
}


/*
 *===========================================================================
 *                        pm_install_handle_to_program_pid_request
 *===========================================================================
 *
 * Description
 * -----------
 * Handle a PM_INSTALL_HANDLE_TO_PROGRAM_PID_REQUEST signal.
 *
 * Requirement
 * -----------
 *
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * state                  The PM internal state structure.
 * request                The request signal.
 *
 * Returns
 * -------
 * None.
 *
 */

GLOBAL void
pm_install_handle_to_program_pid_request (struct PmInternalState *state,
                                          struct PmInstallHandleToProgramPidRequest *
                                          request )
{
   const ProgramMetaData * prog;
   struct PmInstallHandleToProgramPidReply * reply;
   U32 nr_pids;
   U32 pidndx;
   PmStatus status;

   assert(state != NULL);
   assert(request != NULL);
   assert(state->programs != NULL);

  if(pm_strllen(request->install_handle,
		PM_INSTALL_HANDLE_LENGTH) == PM_INSTALL_HANDLE_LENGTH
     || NULL == traverseList1(state->loadmodules,
                              pm_search_lm_install_handle,
                              request->install_handle))
    {
      status = PM_EINSTALL_HANDLE;
      reply = (struct PmInstallHandleToProgramPidReply *)
         alloc (sizeof (struct PmInstallHandleToProgramPidReply),
                PM_INSTALL_HANDLE_TO_PROGRAM_PID_REPLY);
      reply->progpids_size = 0;
      goto pm_install_handle_to_program_pid_reply;
   }

   /*
    * Count number of matching entries
    */
   nr_pids = 0;
   prog = (const ProgramMetaData *) firstNode (state->programs);

   while (prog != NULL)
   {
      assert (prog->lm != NULL);

      if (strcmp (prog->lm->install_handle,
		  request->install_handle) == 0)
      {
         nr_pids++;
      }
      prog = nextNode ((const void *) prog);
   }

   /* Allocate space for reply signal */
   reply = (struct PmInstallHandleToProgramPidReply *)
      alloc (sizeof (struct PmInstallHandleToProgramPidReply) +
             (sizeof(PROCESS) * (nr_pids - 1)),
             PM_INSTALL_HANDLE_TO_PROGRAM_PID_REPLY);

   /*
    * Fill in pid array
    */
   pidndx = 0;
   prog = (const ProgramMetaData *) firstNode (state->programs);

   while (prog != NULL)
   {
      assert (prog->lm != NULL);

      if (strcmp (prog->lm->install_handle, request->install_handle) == 0)
      {
         reply->progpids[pidndx++] = prog->progpid;
      }
      prog = nextNode ((const void *) prog);
   }

   /*
    * Fill in the rest of the signal
    */

   reply->progpids_size = nr_pids;
   status = PM_SUCCESS;

 pm_install_handle_to_program_pid_reply:
   reply->status = status;
   (void)strlcpy (reply->install_handle, request->install_handle,
		  PM_INSTALL_HANDLE_LENGTH);

   /*
    * Send reply and free request signals
    */
   send ((union SIGNAL **) &reply, sender ((union SIGNAL **) &request));
   free_buf((union SIGNAL **) &request);
}


/*
 *===========================================================================
 *                        pm_get_domain_number_request
 *===========================================================================
 *
 * Description
 * -----------
 * Handle a PM_GET_DOMAIN_NUMBER_REQUEST signal.
 *
 * Requirement
 * -----------
 *
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * state                  The PM internal state structure.
 * request                The request signal.
 *
 * Returns
 * -------
 * None.
 *
 */

extern void
pm_get_domain_number_request ( struct PmInternalState * PmInternalState,
                               struct PmGetDomainNumber * request )
{
   const ProgramMetaData * prog;
   
   assert(PmInternalState != NULL);
   assert(request != NULL);
   assert(PmInternalState->programs != NULL);

   /* Find the program meta data */
  prog = pm_find_program (PmInternalState, request->pid);
   
   if (prog != NULL)
   {
      request->domain = prog->domain;
      request->status = PM_SUCCESS;
   }
   else
   {
      request->status = PM_EPROGPID;
   }

   /*
    * Send reply(reuse the signal).
    */
   request->sigNo = PM_GET_DOMAIN_NUMBER_REPLY;
   send ((union SIGNAL **) &request, sender ((union SIGNAL **) &request));
}

/*
 *===========================================================================
 *                        pm_init_program_data
 *===========================================================================
 *
 * Description
 * -----------
 * Initialize the program data at program start.
 *
 * Requirement
 * -----------
 * "Program start".
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * PmInternalState                  The PM internal state structure.
 * pid                              The started programs mainpid.
 *
 * Returns
 * -------
 * None.
 *
 */

GLOBAL void
pm_init_program_data(struct PmInternalState * PmInternalState, PROCESS mainpid)
{
  ProgramMetaData * program;
  void * data_src_base;
  void * data_dest_base;
  U32    data_size;
  void * bss_base;
  U32    bss_size;

  const char * cfg_str;
  size_t cfg_str_size;

  assert(PmInternalState != NULL);
  assert(PmInternalState->programs != NULL);
  assert(mainpid != 0);

  /* OSEhq09539, lock region to avoid race with pgm. The program/lm structures
   * may be updated by pgm during program initialisation .
   */
  zzkrn_kill_lock(0);
  ose_mutex_lock(&(PmInternalState->mutex));

  program  = (ProgramMetaData *)traverseList1(PmInternalState->programs,
						 pm_program_search_by_mainpid,
						 (void *)mainpid);

  if(program == NULL)
    {
      for(;;)
	{
	  /* FIXME: This is a PM internal error, but the program will
           * report it.
           */
	  static const struct pm_file_line extra = {__FILE__, __LINE__};
	  error2(PM_EINTERNAL, (OSERRCODE)&extra);
	} /*lint !e550 */
    }

  data_src_base = (void *)program->lm->initiated_data.base;
  data_dest_base = (void *)program->lm->data.base;
  data_size = program->lm->initiated_data.size;
  bss_base = (void *)program->lm->uninitiated_data.base;
  bss_size = program->lm->uninitiated_data.size;

  ose_mutex_unlock(&(PmInternalState->mutex));
  zzkrn_kill_unlock(0);
  
  cfg_str_size = pm_lm_conf_get_value(program->lm->configuration,
                                      OSE_LM_DATA_INIT, &cfg_str);

  if((data_size != 0) &&
     (cfg_str_size > 0) &&
     (strncmp(PM_BOOL_NO, cfg_str, cfg_str_size) != 0))
    {
       memcpy(data_dest_base, data_src_base, data_size);
    }

  cfg_str_size = pm_lm_conf_get_value(program->lm->configuration,
                                      OSE_LM_BSS_INIT, &cfg_str);

  if((bss_size != 0) &&
     (cfg_str_size > 0) &&
     (strncmp(PM_BOOL_NO, cfg_str, cfg_str_size) != 0))
    {
       memset(bss_base, 0, bss_size);
    }
}

/*
 *===========================================================================
 *                        pm_start_program_request
 *===========================================================================
 *
 * Description
 * -----------
 * Request to start a program.
 *
 * Requirement
 * -----------
 * "Program start".
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * PmInternalState                  The PM internal state structure.
 * request                          The request of the operation.
 *
 * Returns
 * -------
 * None.
 *
 */

GLOBAL void
pm_start_program_request(struct PmInternalState * PmInternalState,
			 struct PmStartProgramRequest * request)
{
  ProgramMetaData * program;
  PmStatus status;
  struct PmStartProgramReply * reply;

  assert(request != NULL);
  assert(PmInternalState->programs != NULL);

  /* Find the program meta data */
  program = pm_find_program (PmInternalState, request->progpid);
   
  /* Check if found */
  if(NULL == program)
  {
     status = PM_EPROGPID;
     goto pm_send_start_program_reply;
  }
  
  /* Make sure the progpid is not the core_module program pid. */
  if(program->segpid == 0)
  {
     status = PM_EOPERATION_ON_CORE;
     goto pm_send_start_program_reply;
  }

  if(program->state == PM_PROGRAM_STARTED)
    {
      status = PM_EPROGRAM_ALREADY_STARTED;
      goto pm_send_start_program_reply;
    }

  /* Start the main block of the program. */
  start(program->mainbid);

  program->state = PM_PROGRAM_STARTED;

  status = PM_SUCCESS;

 pm_send_start_program_reply:
  reply = (struct PmStartProgramReply *) alloc(sizeof(struct PmStartProgramReply),
					       PM_START_PROGRAM_REPLY);

  reply->status = status;
  reply->progpid = request->progpid;
  send((union SIGNAL **) &reply, sender((union SIGNAL **) &request));
  free_buf((union SIGNAL **) &request);
}


/*
 *===========================================================================
 *                        pm_kill_program_request
 *===========================================================================
 *
 * Description
 * -----------
 * Request to kill a program.
 *
 * Requirement
 * -----------
 * "Program kill".
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * PmInternalState                  The PM internal state structure.
 * request                          The request of the operation.
 *
 * Returns
 * -------
 * None.
 *
 */

GLOBAL void
pm_kill_program_request(struct PmInternalState * PmInternalState,
			struct PmKillProgram * request)
{
  ProgramMetaData * program;
  PmStatus status;
  SIGSELECT * attachment;
  static const SIGSELECT attach_prog[] = {1, PM_ATTACH_PROG};

  assert(request != NULL);
  assert(PmInternalState);
  assert(PmInternalState->programs);

  /* Find the program meta data */
  program = pm_find_program (PmInternalState, request->progpid);
   
  /* Check if found */
  if(NULL == program)
  {
     status = PM_EPROGPID;
     goto pm_send_kill_program_reply;
  }
  
  /* Make sure the progpid is not the core_module program pid. */
  if(program->segpid == 0)
  {
     status = PM_EOPERATION_ON_CORE;
     goto pm_send_kill_program_reply;
  }

  /* Kill the main block of the program, and sync with the killing of the
   * program execution context (segment).
   */
  attachment = (SIGSELECT *) alloc(sizeof(SIGSELECT), PM_ATTACH_PROG);
  (void)attach((union SIGNAL **)&attachment, request->progpid);

  /* Stop the block to prevent the main process to receive attach signals. */
  stop(program->mainbid);
  kill_proc(program->mainbid);
  attachment = (SIGSELECT *) receive(attach_prog);
  free_buf((union SIGNAL **) &attachment);

  status = PM_SUCCESS;

 pm_send_kill_program_reply:
  request->sigNo = PM_KILL_PROGRAM_REPLY;
  request->status = status;
  send((union SIGNAL **) &request, sender((union SIGNAL **) &request));
}


/*
 *===========================================================================
 *                        pm_attach_program_request
 *===========================================================================
 *
 * Description
 * -----------
 * Request to attach to a program.
 *
 * Requirement
 * -----------
 * "Program attach".
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * PmInternalState                  The PM internal state structure.
 * request                          The request of the operation.
 *
 * Returns
 * -------
 * None.
 *
 */

GLOBAL void
pm_attach_program_request(struct PmInternalState * PmInternalState,
			  struct PmAttachProgram * request)
{
  ProgramMetaData * program;
  PmStatus status;
  Pm_Attach * attachment;
  /* Compute the size of the data array in the request signal. */
  OSBUFSIZE data_len = sigsize((union SIGNAL **)&request) -
		       offsetof(struct PmAttachProgram, attach_data);
  /* Crop against the specified size. */
  if (data_len > request->attach_data_size)
  {
    data_len = request->attach_data_size;
  }

  assert(request != NULL);
  assert(PmInternalState);
  assert(PmInternalState->programs);

  /* Find the program meta data */
  program = pm_find_program (PmInternalState, request->progpid);
   
  /* Check if found */
  if(NULL == program)
  {
     status = PM_EPROGPID;
     goto pm_send_attach_program_reply;
  } 

  /* Attach to the main block of the program. */
  attachment = (Pm_Attach *) heap_alloc_private(sizeof(Pm_Attach) + data_len,
						       __FILE__, __LINE__);

  attachment->next = 0;
  attachment->client = sender((union SIGNAL **) &request);
  attachment->sigNo = request->attach_sigNo;
  attachment->data_len = data_len;
  memcpy(attachment->data, request->attach_data, data_len);

  if(program->attached == NULL)
    {
      program->attached = attachment;
    }
  else
    {
      Pm_Attach * attach_list;
      attach_list = program->attached;
      while(attach_list->next != NULL)
	{
	  attach_list = attach_list->next;
	}
      attach_list->next = attachment;
    }

  status = PM_SUCCESS;

 pm_send_attach_program_reply:
  request->sigNo = PM_ATTACH_PROGRAM_REPLY;
  request->status = status;
  send((union SIGNAL **) &request, sender((union SIGNAL **) &request));
}

/*
 *===========================================================================
 *                        pm_extend_program_memory_request
 *===========================================================================
 *
 * Description
 * -----------
 * Request to extend the program memory with a new region.
 *
 * Requirement
 * -----------
 * "Program new memory".
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * PmInternalState                  The PM internal state structure.
 * request                          The request of the operation.
 *
 * Returns
 * -------
 * None.
 *
 */

GLOBAL void
pm_extend_program_memory_request(struct PmInternalState * PmInternalState,
				 struct PmExtendProgramMemory * request)
{
  ProgramMetaData * program;
  PmStatus status;
  U32 options;
  U32 permission;
  U32 cache;
  const char * program_type_conf;
  U32 program_type_len;
  char program_type_name[PM_PROGRAM_TYPE_PARAM_LENGTH];
  ProgramType * program_type;

  /* Find the program meta data */
  program = pm_find_program (PmInternalState, request->progpid);
   
  /* Check if found */
  if(NULL == program)
  {
     status = PM_EPROGPID;
     goto pm_send_extend_program_memory_reply;
  }

  /* Sanity check the options. */
  if((request->options != PM_SPECIFY_BASE
      && request->options != PM_SPECIFY_ALIGN
      && request->options != 0)
     || ((request->options & PM_SPECIFY_ALIGN)
	 && (request->options & PM_SPECIFY_BASE)))
    {
      status = PM_EOPTION_UNKNOWN;
      goto pm_send_extend_program_memory_reply;
    }

  /* Determine the program type. */
  /*The core lm. It does not have the program type defined
    in the configuration. But insted the defualt PM_CORE_PROGRAM_TYPE_NAME
    defined by the core startup is used.*/
  if(program->segpid == 0)
  {
     (void)strcpy(program_type_name, PM_CORE_PROGRAM_TYPE_NAME);
  }
  else
  {
     program_type_len = pm_lm_conf_get_value(program->lm->configuration,
					     OSE_LM_PROGRAM_TYPE,
                                             &program_type_conf);
     (void)strlcpy(program_type_name, program_type_conf, program_type_len + 1);

  }

  program_type = (ProgramType *)traverseList1(PmInternalState->prog_types,
					      pm_search_prog_type,
					      program_type_name);

  if(program_type == NULL)
    {
      for(;;)
	{
	  static const struct pm_file_line extra = {__FILE__, __LINE__};
	  error2(PM_EINTERNAL, (OSERRCODE)&extra);
	} /*lint !e550 */
    }

  if(request->options & PM_SPECIFY_ALIGN)
    {
      options = MM_CREATE_REGION_SPECIFY_ALIGN;
    }
  else if(request->options & PM_SPECIFY_BASE)
    {
      options = MM_CREATE_REGION_SPECIFY_LOG;
    }
  else
    {
      options = 0;
    }

  {
    OSADDRESS ext_base = request->base;
    U32 ext_size = request->size;
    U32 ext_align = request->align;

    permission = program_type->regions[request->program_region_type].permissions;
    cache = program_type->regions[request->program_region_type].cache_mode;
    
    if(pm_vs_gmm_create_region(PmInternalState->gmm_pid,
                               PM_REGION_EXTENSION,
			       program->domain,
                               &ext_size,
			       &ext_base,
			       program_type->regions[request->program_region_type].phys_mem,
			       program_type->regions[request->program_region_type].log_mem,
			       permission,
                               cache,
                               &ext_align,
			       options) != MM_SUCCESS)
      {
        
	status = PM_EREGION_CREATION_FAILED;
	goto pm_send_extend_program_memory_reply;
      }
    assert((options & MM_CREATE_REGION_SPECIFY_LOG) ? ext_base == request->base : ext_base != 0);
    assert((options & MM_CREATE_REGION_SPECIFY_ALIGN) ? ext_align >= request->align : 1);
    assert(ext_size >= request->size);
    
    request->size = ext_size;
    request->base = ext_base;
    request->align = ext_align;
  }

  status = PM_SUCCESS;

 pm_send_extend_program_memory_reply:
  request->sigNo = PM_EXTEND_PROGRAM_MEMORY_REPLY;
  request->status = status;
  send((union SIGNAL **) &request, sender((union SIGNAL **) &request));
}

/*
 *===========================================================================
 *                        pm_move_program_to_cpu_request
 *===========================================================================
 *
 * Description
 * -----------
 * Request to move a program to a specific CPU.
 *
 * Requirement
 * -----------
 * "Program move to CPU".
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * PmInternalState                  The PM internal state structure.
 * request                          The request of the operation.
 *
 * Returns
 * -------
 * None.
 *
 */

GLOBAL void
pm_move_program_to_cpu_request(struct PmInternalState * PmInternalState,
                               struct PmMoveProgramToCPU * request)
{
  ProgramMetaData * program;
  PmStatus status;  
  
  /* Find the program meta data */
  program = pm_find_program (PmInternalState, request->progpid);
   
  /* Check if found */
  if(NULL == program || !program->segpid)
  {
     status = PM_EPROGPID;
     goto pm_move_program_to_cpu_reply;
  }

  /* Sanity check the CPU id */
  if(request->cpu >= ose_num_cpus())
  {
     status = PM_ECPUID;
     goto pm_move_program_to_cpu_reply;
  }

  /* Check if the program is locked by the program configuration or
  ** by the kernel. Note that we MUST CHECK THIS HERE, because the
  ** Program Manager is trusted by the kernel and has special rights
  ** to move stuff.
  */
  if (program->locked ||
      zzkrn_segment_migration_status(request->progpid, 0, 0))
  {
     status = PM_ELOCKED;
     goto pm_move_program_to_cpu_reply;
  }

  /* Try to move the whole segment. */
  if(0 == ose_bind_process(program->segpid, request->cpu))
  {
     /* Strange! Can it really fail? */
     status = PM_ELOCKED;
     goto pm_move_program_to_cpu_reply;
  }
  status = PM_SUCCESS;

 pm_move_program_to_cpu_reply:
  request->sigNo = PM_MOVE_PROGRAM_TO_CPU_REPLY;
  request->status = status;
  send((union SIGNAL **) &request, sender((union SIGNAL **) &request));
}
