/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2001-2007 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */
/*
 ****************************************************************************
 *			  PM INTERNAL FUNCTIONALITY
 *
 *     $Workfile:   pm_vs_gmm.c  $
 *
 *
 ****************************************************************************
 * Description: The implementation of the GMM vs PM interface support in the
 *              Program Manager.
 *
 */

/***********************************************************
 *
 * INCLUDE FILES
 *
 ***********************************************************/

#include "ose.h"
#include "osetypes.h"
#include "mm.sig"

/* OSEhq09509, added the mmtypes.h header to include mm_status_t. */
#include "mmtypes.h"
#include "pm.h"
#include "pm_lint.h"
#include "pm_err.h"
#include "assert.h"
#include "pm_internal.h"
#include "pm_internal.sig"
#include "string.h"

#ifndef OUTFMT_NO_SPRINTF
#define OUTFMT_NO_SPRINTF
#include "outfmt.h"
#undef OUTFMT_NO_SPRINTF
#else
#include "outfmt.h"
#endif

/* The pm vs gmm interface definition. */
#include "pm_vs_gmm.h"

/***********************************************************
 *
 * CONSTANTS AND DEFINITIONS
 *
 ***********************************************************/

/* The process name of the memory manager. */
#define PM_MM_PROCESS_NAME            "ose_memory_mgr"

/* The program domain name. */
#define PM_DOMAIN_NAME                "PM domain"

/***********************************************************
 *
 * STRUCTURES AND GLOBALS
 *
 ***********************************************************/

/***********************************************************
 *
 * FUNCTION DEFINITIONS
 *
 ***********************************************************/

/*
 *===========================================================================
 *                        pm_attach_gmm
 *===========================================================================
 *
 * Description
 * -----------
 * Synchronize the PM with the MM Framework.
 *
 * Requirement
 * -----------
 * "PM memory management usage".
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * gmm                              A pointer to the gmm pid storage.
 *
 * Returns
 * -------
 * (OSATTREF) ref                   The attach reference to the GMM process.
 *
 *
 */

OSATTREF
pm_attach_gmm(PROCESS * gmm)
{
  /* The real sync of the GMM will be done when resurrection of programs is
   * performed.
   */
  union SIGNAL * sig = alloc(sizeof(SIGSELECT), HUNT_GMM);
  static const SIGSELECT hunt_gmm[] = {1, HUNT_GMM};

  assert(strlen(PM_DOMAIN_NAME) < MM_NAME_LENGTH);
  assert((PM_PROGRAM_TYPE_PARAM_LENGTH-1) <= MM_NAME_LENGTH);

  (void)hunt(PM_MM_PROCESS_NAME, 0, 0, &sig);

  sig = receive(hunt_gmm);

  *gmm = sender(&sig);

  sig->sigNo = ATTACH_GMM;

  return attach(&sig, *gmm);
}

/*
 *===========================================================================
 *                        pm_vs_gmm_create_region
 *===========================================================================
 *
 * Description
 * -----------
 * Create a region via the GMM.
 *
 * Requirement
 * -----------
 * "PM memory management usage".
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * gmm                              The gmm pid.
 * region_name                      The name of the region.
 * domain                           The domain number to create the region in.
 * size                             The required size of the region.
 * base                             The required base of the region.
 * phys_mem                         The phys_mem to attach to the region.
 * log_mem                          The log_mem to create the region in.
 * permissions                      The initial permissions to the region.
 * cache                            The cache mode of the region.
 * align                            The required alignment of the region.
 * options                          The additional requirements on the region
 *                                  (see mm.sig for information about the
 *                                  options).
 *
 * Returns
 * -------
 * (mm_status_t) status             The status of the operation.
 *
 *
 */

mm_status_t
pm_vs_gmm_create_region(PROCESS gmm,
			const char * region_name,
			OSDOMAIN domain,
			U32 * size,
			OSADDRESS * base,
			const char * phys_mem,
			const char * log_mem,
			U32 permissions,
			U32 cache,
			U32 *align,
			U32 options)
{
  struct mm_create_region * re;
  static const SIGSELECT gmm_region[] = {2, MM_CREATE_REGION_REPLY,
					 ATTACH_GMM};
  mm_status_t      status = 0xffffffff; /* Make lint happy */;

  assert(gmm != 0);

  re = (struct mm_create_region *) alloc(sizeof(struct mm_create_region),
				  MM_CREATE_REGION_REQUEST);

  re->size   = *size;
  re->domain = domain;
  snprintf(re->name, MM_NAME_LENGTH, region_name);
  snprintf(re->phys_mem, MM_NAME_LENGTH, phys_mem);
  snprintf(re->log_mem, MM_NAME_LENGTH, log_mem);

  if(pm_strllen(re->log_mem, MM_NAME_LENGTH) == MM_NAME_LENGTH
     || pm_strllen(re->phys_mem, MM_NAME_LENGTH) == MM_NAME_LENGTH
     || pm_strllen(re->name, MM_NAME_LENGTH) == MM_NAME_LENGTH)
    {
      /* The name, phys_mem or log_mem of the region is too large. */
      for(;;)
	{
	  const struct pm_file_line extra = {__FILE__, __LINE__};
	  error2(PM_EINTERNAL, (OSERRCODE) &extra);
	} /*lint !e550 */
    }

  re->permissions = permissions;
  re->cache_mode = cache;
  re->base   = *base;
  re->align  = *align;
  re->options  = options;

  send((union SIGNAL **) &re, gmm);

  re = (struct mm_create_region *)receive(gmm_region);

  if(re->sig_no == MM_CREATE_REGION_REPLY)
    {
      *base = re->base;
      *align = re->align;
      *size = re->size;
      status = re->status;

      free_buf((union SIGNAL **) &re);
    }
  else
    {
      /* The GMM process is lost, this should never happen. */
      for(;;)
	{
	  error(PM_EGMM_LOST);
	}
    }

  return status;
}

/*
 *===========================================================================
 *                        pm_vs_gmm_delete_region
 *===========================================================================
 *
 * Description
 * -----------
 * Delete a region via the GMM.
 *
 * Requirement
 * -----------
 * "PM memory management usage".
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * gmm                              The gmm pid.
 * domain                           The domain number to locate the region.
 * base                             The base of the region.
 *
 * Returns
 * -------
 * (PmStatus) status                The status of the operation.
 *
 *
 */

PmStatus
pm_vs_gmm_delete_region(PROCESS gmm,
			OSDOMAIN domain,
			OSADDRESS base)
{
  struct mm_delete_region * re;
  static const SIGSELECT gmm_region[] = {2, MM_DELETE_REGION_REPLY,
					 ATTACH_GMM};
  mm_status_t status = 0xffffffff; /* Make lint happy */;

  assert(base != 0);

  re = (struct mm_delete_region *) alloc(sizeof(struct mm_delete_region),
					 MM_DELETE_REGION_REQUEST);
  re->base = base;
  re->domain = domain;

  send((union SIGNAL **) &re, gmm);

  re = (struct mm_delete_region *)receive(gmm_region);

  if(re->sig_no == MM_DELETE_REGION_REPLY)
    {
      status = re->status;
      free_buf((union SIGNAL **) &re);
    }
  else
    {
      /* The GMM process is lost, this should never happen. */
      for(;;)
	{
	  error(PM_EGMM_LOST);
	}
    }

  if (domain == MM_KERNEL_DOMAIN && status == MM_ENON_EXISTING_REGION)
  {
     /* Failure to remove a non-existing region
      * should not be considered fatal. */
     status = MM_SUCCESS;
  }
  
  return status == MM_SUCCESS ? PM_SUCCESS : PM_EREGION_DELETION_FAILED;
}

/*
 *===========================================================================
 *                        pm_vs_gmm_create_domain
 *===========================================================================
 *
 * Description
 * -----------
 * Create a domain via the GMM.
 *
 * Requirement
 * -----------
 * "PM memory management usage".
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * gmm                              The gmm pid.
 * domain                           The domain number of the created domain.
 *
 * Returns
 * -------
 * (PmStatus) status                The status of the operation.
 *
 *
 */

PmStatus
pm_vs_gmm_create_domain(PROCESS gmm,
                        const char *name,
			OSDOMAIN * domain)
{
  struct mm_create_domain * re;
  static const SIGSELECT gmm_domain[] = {2, MM_CREATE_DOMAIN_REPLY,
					 ATTACH_GMM};
  mm_status_t status = 0xffffffff; /* Make lint happy */

  assert(domain != NULL);
  assert(gmm != 0);
  assert (name != 0);

  re = (struct mm_create_domain *) alloc(sizeof(struct mm_create_domain),
					 MM_CREATE_DOMAIN_REQUEST);

  (void)strlcpy(re->name, name, MM_NAME_LENGTH);

  re->options = MM_CREATE_DOMAIN_SPECIFY_NAME;

  send((union SIGNAL **) &re, gmm);

  re = (struct mm_create_domain *)receive(gmm_domain);

  if(re->sig_no == MM_CREATE_DOMAIN_REPLY)
    {
      *domain = re->domain;
      status = re->status;

      free_buf((union SIGNAL **) &re);
    }
  else
    {
      /* The GMM process is lost, this should never happen. */
      for(;;)
	{
	  error(PM_EGMM_LOST);
	}
    }

  return status == MM_SUCCESS ? PM_SUCCESS : PM_EDOMAIN_CREATION_FAILED;
}


/*
 *===========================================================================
 *                        pm_vs_gmm_delete_domain
 *===========================================================================
 *
 * Description
 * -----------
 * Delete a domain via the GMM.
 *
 * Requirement
 * -----------
 * "PM memory management usage".
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * gmm                              The gmm pid.
 * domain                           The domain number of the domain to delete.
 *
 * Returns
 * -------
 * (PmStatus) status                The status of the operation.
 *
 *
 */

PmStatus
pm_vs_gmm_delete_domain(PROCESS gmm,
			OSDOMAIN domain)
{
  struct mm_delete_domain * re;
  static const SIGSELECT gmm_domain[] = {2, MM_DELETE_DOMAIN_REPLY,
					 ATTACH_GMM};
  mm_status_t status = 0xffffffff; /* Make lint happy */

  assert(domain != 0);
  assert(gmm != 0);

  re = (struct mm_delete_domain *) alloc(sizeof(struct mm_delete_domain),
					 MM_DELETE_DOMAIN_REQUEST);

  re->domain = domain;

  send((union SIGNAL **) &re, gmm);

  re = (struct mm_delete_domain *)receive(gmm_domain);

  if(re->sig_no == MM_DELETE_DOMAIN_REPLY)
    {
      status = re->status;
      free_buf((union SIGNAL **) &re);
    }
  else
    {
      /* The GMM process is lost, this should never happen. */
      for(;;)
	{
	  error(PM_EGMM_LOST);
	}
    }

  return status == MM_SUCCESS ? PM_SUCCESS : PM_EDOMAIN_DELETION_FAILED;
}

/*
 *===========================================================================
 *                        pm_vs_gmm_share_region
 *===========================================================================
 *
 * Description
 * -----------
 * Share a region via the GMM.
 *
 * Requirement
 * -----------
 * "PM memory management usage".
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * gmm                              The gmm pid.
 * source                           The domain number of the source domain.
 * destination                      The domain number of the destination domain.
 * base                             The base address of the region.
 * options                          The share region options.
 * permission                       The user mode permission.
 *
 * Returns
 * -------
 * status                           The MM status of the operation.
 *
 */

/* OSEhq09509, changed return type to MM status type. */
mm_status_t
pm_vs_gmm_share_region(PROCESS gmm, OSADDRESS base, OSDOMAIN source, OSDOMAIN destination,
		       U32 options, U32 permission)
{
  struct mm_share_region * re;
  static const SIGSELECT gmm_share[] = {2, MM_SHARE_REGION_REPLY,
					 ATTACH_GMM};
  mm_status_t  status = 0xffffffff; /* Make lint happy */

  assert(gmm != 0);
  assert(base != 0);

  re = (struct mm_share_region *) alloc(sizeof(struct mm_share_region),
					MM_SHARE_REGION_REQUEST);

  re->base = base;
  re->options = options;
  re->permission = permission;
  re->source = source;
  re->destination = destination;

  send((union SIGNAL **) &re, gmm);

  re = (struct mm_share_region *)receive(gmm_share);

  if(re->sig_no == MM_SHARE_REGION_REPLY)
    {
      status = re->status;
      free_buf((union SIGNAL **) &re);
    }
  else
    {
      /* The GMM process is lost, this should never happen. */
      for(;;)
	{
	  error(PM_EGMM_LOST);
	}
    }

  /* OSEhq09509, changed return value to the MM status. */
  return status;
}


/*
 *===========================================================================
 *                        pm_vs_gmm_list_regions
 *===========================================================================
 *
 * Description
 * -----------
 * List all regions of a specific domain.
 *
 * Requirement
 * -----------
 * "PM resurrection".
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * gmm                              The gmm pid.
 * domain                           The domain to list regions in.
 * reply                            The reply to the request.
 *
 * Returns
 * -------
 * (mm_status_t) status                The status of the operation.
 *
 */

mm_status_t
pm_vs_gmm_list_regions(PROCESS gmm, OSDOMAIN domain,
		       struct mm_list_regions_reply ** reply)
{
  static const SIGSELECT list_regions[] = {1, MM_LIST_REGIONS_REPLY};
  struct mm_list_regions_request * request;

  request = (struct mm_list_regions_request *)
    alloc(sizeof(struct mm_list_regions_request), MM_LIST_REGIONS_REQUEST);
  request->domain = domain;

  send((union SIGNAL **) &request, gmm);
  *reply = (struct mm_list_regions_reply *)receive(list_regions);
  return (*reply)->status;
}


/*
 *===========================================================================
 *                        pm_vs_gmm_get_region_info
 *===========================================================================
 *
 * Description
 * -----------
 * Get information about a region.
 *
 * Requirement
 * -----------
 * "PM resurrection".
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * gmm                              The gmm pid.
 * base                             The base addressof the region.
 * domain                           The domain to list regions in.
 * reply                            The reply to the request.
 *
 * Returns
 * -------
 * (mm_status_t) status                The status of the operation.
 *
 */

mm_status_t
pm_vs_gmm_get_region_info(PROCESS gmm, OSDOMAIN domain, OSADDRESS base,
			  struct mm_create_region ** get_reply)
{
  static const SIGSELECT get_region_info[] = {2, MM_GET_REGION_INFO_REPLY,
					      MM_GET_REGION_INFO_NOT_LAST};
  struct mm_create_region * get_request;

  get_request = (struct mm_create_region *)
    alloc(sizeof(struct mm_create_region) ,MM_GET_REGION_INFO_REQUEST);
  get_request->domain = domain;
  get_request->base = base;

  send((union SIGNAL **) &get_request, gmm);

  do
  {
     *get_reply = (struct mm_create_region *)receive(get_region_info);

     /* Free all "...not_last" signals sent due to long fragment lists */
     /* The physical fragments of the region are of no interest to PM */
     if ((*get_reply)->sig_no == MM_GET_REGION_INFO_NOT_LAST)
     {
	free_buf((union SIGNAL**)(get_reply));
	*get_reply = NULL;
     }
  } while (*get_reply == NULL);

  /* Make sure that the physical fragments are not used,
     as the list might be incomplete */
  (*get_reply)->phys_count = 0;
  
  return (*get_reply)->status;
}

/*
 *===========================================================================
 *                  pm_vs_gmm_get_log_mem_mapping_type
 *===========================================================================
 *
 * Description
 * -----------
 * Get mapping type information about a log_mem.
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * gmm                              The gmm pid.
 * log_mem                          Name of the log_mem.
 * mapping_type                     The mapping type of the log_mem (output).
 *
 * Returns
 * -------
 * (mm_status_t) status             The status of the operation.
 *
 */
mm_status_t
pm_vs_gmm_get_log_mem_mapping_type(PROCESS gmm,
                                   const char *log_mem,
                                   U32 *mapping_type)
{
   static const SIGSELECT info_reply[] = {1, MM_GET_LOG_MEM_INFO_REPLY};
   struct mm_get_log_mem_info *sig;
   mm_status_t status;

   assert(gmm != 0);

   sig = (struct mm_get_log_mem_info *) 
      alloc(sizeof(struct mm_get_log_mem_info),
            MM_GET_LOG_MEM_INFO_REQUEST);

   sig->count = 1;
   sig->tags[0].tag_id = MM_TAG_INFO_LOG_MEM_TYPE;
   strlcpy(sig->name, log_mem, MM_NAME_LENGTH);

   send((union SIGNAL **)&sig, gmm);
   sig = (struct mm_get_log_mem_info *)receive(info_reply);
   status = sig->status;
   
   if (status != MM_SUCCESS)
   {
      free_buf((union SIGNAL **)&sig);
      return status;
   }

   if (mapping_type != NULL)
   {
      *mapping_type = sig->tags[0].tag_value;
   }
   
   free_buf((union SIGNAL **)&sig);

   return status;
}


mm_status_t
pm_vs_gmm_add_phys_resource(PROCESS gmm,
                            OSDOMAIN domain,
                            log_addr_t base,
                            const char *phys_mem)
{
   mm_status_t status;
   static const SIGSELECT sel [] = {1, MM_ADD_PHYS_RESOURCE_REPLY};
   union SIGNAL *sig = alloc(sizeof(struct mm_add_phys_resource),
                             MM_ADD_PHYS_RESOURCE_REQUEST);
   const size_t sz = sizeof sig->add_phys_resource.phys_mem;
   char *last = &sig->add_phys_resource.phys_mem[sz-1];

   (void)domain;

   
   sig->add_phys_resource.domain = MM_KERNEL_DOMAIN;
   sig->add_phys_resource.base   = base;
   strncpy(sig->add_phys_resource.phys_mem, phys_mem, sz);

   if (*last != '\0') {
      *last = '\0';
      ramlog_printf("Warning. Name of phy_mem truncated. '%s' -> '%s'\n",
                    phys_mem, sig->add_phys_resource.phys_mem);
   }
   
   send(&sig, gmm);
   
   sig = receive(sel);
   
   status = sig->add_phys_resource.status;
   free_buf(&sig);
   
   return status;
}

mm_status_t
pm_vs_gmm_remove_phys_resource(PROCESS gmm,
                               OSDOMAIN domain,
                               log_addr_t base)
{
   mm_status_t status;
   static const SIGSELECT sel [] = {1, MM_REMOVE_PHYS_RESOURCE_REPLY};
   union SIGNAL *sig = alloc(sizeof(struct mm_remove_phys_resource),
                             MM_REMOVE_PHYS_RESOURCE_REQUEST);
   
   sig->remove_phys_resource.domain = domain;
   sig->remove_phys_resource.base   = base;
   
   send(&sig, gmm);
   
   sig = receive(sel);
   status = sig->remove_phys_resource.status;
   free_buf(&sig);

   return status;
}
