/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2006 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */
/*
 * Description: The function interface of the pm vs gmm interface in the
 *              Program Manager.
 *
 */

#ifndef __PM_VS_GMM_H
#define __PM_VS_GMM_H

/***********************************************************
 *
 * INCLUDE FILES
 *
 ***********************************************************/

/* Program Management headers. */
#include "pm_internal.h"

/* Program Management block signal interface. */
#include "pm.sig"

/* The memory manages type definitions. */
#include "mmtypes.h"
#include "mm.sig"

/* Note: this file is not adapted to C++ since the file was never intended to be
 * included from a C++ source file.
 */

/***********************************************************
 *
 * CONSTANTS AND DEFINITIONS
 *
 ***********************************************************/

/***********************************************************
 *
 * STRUCTURES AND GLOBALS
 *
 ***********************************************************/

/***********************************************************
 *
 * FUNCTION DEFINITIONS
 *
 ***********************************************************/

/*
 *===========================================================================
 *                        pm_attach_gmm
 *===========================================================================
 *
 * Description
 * -----------
 * Synchronize the PM with the MMS Framework.
 *
 * Requirement
 * -----------
 * "PM memory management usage".
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * gmm                              A pointer to the gmm pid storage.
 *
 * Returns
 * -------
 * (OSATTREF) ref                   The attach reference to the GMM process.
 *
 *
 */

OSATTREF
pm_attach_gmm(PROCESS * gmm);

/*
 *===========================================================================
 *                        pm_vs_gmm_create_region
 *===========================================================================
 *
 * Description
 * -----------
 * Create a region via the GMM.
 *
 * Requirement
 * -----------
 * "PM memory management usage".
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * gmm                              The gmm pid.
 * region_name                      The name of the region.
 * domain                           The domain number to create the region in.
 * size                             The required size of the region.
 * base                             The required base of the region.
 * allocator                        The allocator to attach to the region.
 * partition                        The partition to create the region in.
 * permissions                      The initial permissions to the region.
 * cache                            The cache mode of the region.
 * align                            The required alignment of the region.
 * options                          The additional requirements on the region
 *                                  (see mm.sig for information about the
 *                                  options).
 *
 * Returns
 * -------
 * (mm_status_t) status             The status of the operation.
 *
 *
 */

mm_status_t
pm_vs_gmm_create_region(PROCESS gmm,
			const char * region_name,
			OSDOMAIN domain,
			U32 * size,
			OSADDRESS * base,
			const char * allocator,
			const char * partition,
			U32 permissions,
			U32 cache,
			U32 *align,
			U32 options);

/*
 *===========================================================================
 *                        pm_vs_gmm_delete_region
 *===========================================================================
 *
 * Description
 * -----------
 * Delete a region via the GMM.
 *
 * Requirement
 * -----------
 * "PM memory management usage".
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * gmm                              The gmm pid.
 * domain                           The domain number to locate the region.
 * base                             The base of the region.
 *
 * Returns
 * -------
 * (PmStatus) status                The status of the operation.
 *
 *
 */

PmStatus
pm_vs_gmm_delete_region(PROCESS gmm,
			OSDOMAIN domain,
			OSADDRESS base);

/*
 *===========================================================================
 *                        pm_vs_gmm_create_domain
 *===========================================================================
 *
 * Description
 * -----------
 * Create a domain via the GMM.
 *
 * Requirement
 * -----------
 * "PM memory management usage".
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * gmm                              The gmm pid.
 * domain                           The domain number of the created domain.
 *
 * Returns
 * -------
 * (PmStatus) status                The status of the operation.
 *
 *
 */

PmStatus
pm_vs_gmm_create_domain(PROCESS gmm,
                        const char *name,
			OSDOMAIN * domain);


/*
 *===========================================================================
 *                        pm_vs_gmm_delete_domain
 *===========================================================================
 *
 * Description
 * -----------
 * Delete a domain via the GMM.
 *
 * Requirement
 * -----------
 * "PM memory management usage".
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * gmm                              The gmm pid.
 * domain                           The domain number of the domain to delete.
 *
 * Returns
 * -------
 * (PmStatus) status                The status of the operation.
 *
 *
 */

PmStatus
pm_vs_gmm_delete_domain(PROCESS gmm,
			OSDOMAIN domain);

/*
 *===========================================================================
 *                        pm_vs_gmm_share_region
 *===========================================================================
 *
 * Description
 * -----------
 * Share a region via the GMM.
 *
 * Requirement
 * -----------
 * "PM memory management usage".
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * gmm                              The gmm pid.
 * source                           The domain number of the source domain.
 * destination                      The domain number of the destination domain.
 * base                             The base address of the region.
 * options                          The share region options.
 * permission                       The user mode permission.
 *
 * Returns
 * -------
 * status                           The MM status of the operation.
 *
 */

/* OSEhq09509, changed return type to MM status type. */
mm_status_t
pm_vs_gmm_share_region(PROCESS gmm, OSADDRESS base, OSDOMAIN source, OSDOMAIN destination,
		       U32 options, U32 permission);

/*
 *===========================================================================
 *                        pm_vs_gmm_list_regions
 *===========================================================================
 *
 * Description
 * -----------
 * List all regions of a specific domain.
 *
 * Requirement
 * -----------
 * "PM resurrection".
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * gmm                              The gmm pid.
 * domain                           The domain to list regions in.
 * reply                            The reply to the request.
 *
 * Returns
 * -------
 * (mm_status_t) status                The status of the operation.
 *
 */

mm_status_t
pm_vs_gmm_list_regions(PROCESS gmm, OSDOMAIN domain,
		       struct mm_list_regions_reply ** reply);

/*
 *===========================================================================
 *                        pm_vs_gmm_get_region_info
 *===========================================================================
 *
 * Description
 * -----------
 * Get information about a region.
 *
 * NOTE: The region information (in get_reply) will not include the
 * physical region fragments The phys_count member will be set to 0.
 * This is however not a problem for PM as the current implementation
 * does not make use of this information
 *
 * Requirement
 * -----------
 * "PM resurrection".
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * gmm                              The gmm pid.
 * base                             The base addressof the region.
 * domain                           The domain to list regions in.
 * reply                            The reply to the request.
 *
 * Returns
 * -------
 * (mm_status_t) status                The status of the operation.
 *
 */

mm_status_t
pm_vs_gmm_get_region_info(PROCESS gmm, OSDOMAIN domain, OSADDRESS base,
			  struct mm_create_region ** get_reply);

/*
 *===========================================================================
 *                        pm_vs_gmm_add_phys_resource
 *===========================================================================
 *
 * Description
 * -----------
 * Add a physical resource to a logical only region.
 *
 *
 * Requirement
 * -----------
 *
 * Side effects
 * ------------
 *
 * Parameter description
 * ---------------------
 * gmm                    The gmm pid.
 * domain & base          Used to identify the region.
 * phys_mem               The phys mem from which the physical memory shall be
 *                        allocated.
 *
 * Returns
 * -------
 * (mm_status_t) status   The status of the operation.
 *
 */

mm_status_t
pm_vs_gmm_add_phys_resource(PROCESS gmm,
                            OSDOMAIN domain,
                            log_addr_t base,
                            const char *phys_mem);

/*
 *===========================================================================
 *                        pm_vs_gmm_remove_phys_resource
 *===========================================================================
 *
 * Description
 * -----------
 * Removes a regions physical resource.
 *
 *
 * Requirement
 * -----------
 *
 * Side effects
 * ------------
 *
 * Parameter description
 * ---------------------
 * gmm                    The gmm pid.
 * domain & base          Used to identify the region.
 *
 * Returns
 * -------
 * (mm_status_t) status   The status of the operation.
 *
 */

mm_status_t
pm_vs_gmm_remove_phys_resource(PROCESS gmm,
                               OSDOMAIN domain,
                               log_addr_t base);

/*
 *===========================================================================
 *                  pm_vs_gmm_get_log_mem_mapping_type
 *===========================================================================
 *
 * Description
 * -----------
 * Get mapping type information about a log_mem.
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * gmm                              The gmm pid.
 * log_mem                          Name of the log_mem.
 * mapping_type                     The mapping type of the log_mem (output).
 *
 * Returns
 * -------
 * (mm_status_t) status             The status of the operation.
 *
 */
mm_status_t
pm_vs_gmm_get_log_mem_mapping_type(PROCESS gmm,
                                   const char *log_mem,
                                   U32 *mapping_type);

#endif /* __PM_VS_GMM_H */
