/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2001-2007 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */
/*
 ****************************************************************************
 *			  PM INTERNAL FUNCTIONALITY
 *
 *     $Workfile:   pm_vs_rtl.c  $
 *
 *
 ****************************************************************************
 * Description: The implementation of the rtl management support in the 
 *              Program Manager. 
 * 
 */

/* OSE general headers. */
#include "ose.h"
#include "osetypes.h"

#ifndef OUTFMT_NO_SPRINTF
#define OUTFMT_NO_SPRINTF
#include "outfmt.h"
#undef OUTFMT_NO_SPRINTF
#else
#include "outfmt.h"
#endif

/* Standard headers. */
#include "assert.h"
#include "string.h"
#include "stdlib.h"
#include "stdio.h"
#include "malloc.h"

/* List handling functionality. */
#include "oldlist.h"

/* Program Management headers. */
#include "pm.h"
#include "pm_internal.h"
#include "pm_lint.h"

/* Program Management block signal interface. */
#include "pm.sig"
#include "rtl.sig"
#include "pm_internal.sig"

/* The PM RTL management interface definition. */
#include "pm_vs_rtl.h"

/***********************************************************
 *
 * CONSTANTS AND DEFINITIONS
 *
 ***********************************************************/

/* The number of signals of the RTL interface. */
#define PM_RTL_INTERFACE_SIZE  3

/* The signals supported in the rtl.sig interface. */
static const SIGSELECT rtl_sigs[PM_RTL_INTERFACE_SIZE+1] =
{
  PM_RTL_INTERFACE_SIZE,
  RTL_CONNECT_REQUEST,
  RTL_INTERFACE_REQUEST,
  RTL_UNINSTALL_LOAD_MODULE_REQUEST
};

/* One second wait for a RTL interface reply. */
/* FIXME: This should be configurable. */
#define PM_RTL_CONNECT_TMO     1000

/***********************************************************
 *
 * STRUCTURES AND GLOBALS
 *
 ***********************************************************/

/***********************************************************
 *
 * FUNCTION DEFINITIONS
 *
 ***********************************************************/

/*
 *===========================================================================
 *                        pm_rtl_search_by_ff
 *===========================================================================
 * 
 * Description
 * -----------
 * Search for a specific RTL using the file format (as part of 
 * traverseList1).
 *
 * Requirement
 * -----------
 * "RTL file format". 
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * node                             The node (in the list) to compare to.
 * ff                               The name of the file format to search for.
 *
 * Returns
 * -------
 * (void *) 0                       The node did not match the file format, 
 *                                  ZERO is returned.
 * (void *) node                    The node matched the file format, 
 *                                  the pointer to the node is returned.
 *
 *
 */

GLOBAL void *
pm_rtl_search_by_ff(void * node, void *ff)
{
  RtlData * rtl;
  
  assert(node != NULL);
  assert(((RtlData *)node)->name != NULL);
  assert(ff != NULL);

  rtl = (RtlData *) node; 

  if(strcmp(rtl->name, (const char *)ff) == 0)
    {
      return (void *) node;
    }
  else
    {
      return (void *) 0;
    }
}


/*
 *===========================================================================
 *                        pm_rtl_search_by_pid
 *===========================================================================
 * 
 * Description
 * -----------
 * Search for a specific RTL using the RTL pid (as part of 
 * traverseList1).
 *
 * Requirement
 * -----------
 * "RTL management"
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * node                             The node (in the list) to compare to.
 * pid                              The RTL pid to search for.
 *
 * Returns
 * -------
 * (void *) 0                       The node did not match the pid, 
 *                                  ZERO is returned.
 * (void *) node                    The node matched the pid, 
 *                                  the pointer to the node is returned.
 *
 *
 */

GLOBAL void *
pm_rtl_search_by_pid(void * node, void * pid)
{
  assert(node != NULL);
  assert(((RtlData *)node)->pid != (PROCESS)NULL);
  assert(pid != NULL);

  if(((RtlData *)node)->pid == (PROCESS)pid)
    {
      return (void *) node;
    }
  else
    {
      return (void *) 0;
    }
}

/*
 *===========================================================================
 *                        pm_rtl_killed
 *===========================================================================
 * 
 * Description
 * -----------
 * Handle a killed RTL.
 *
 * Requirement
 * -----------
 * "RTL management"
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * PmInternalState                  The PM internal state structure.
 * sig                              The attach signal of the lost RTL.
 *
 * Returns
 * -------
 * None.
 *
 */

GLOBAL void
pm_rtl_killed(struct PmInternalState * PmInternalState, 
			 SIGSELECT * sig)
{
  RtlData * rtl_info;
  PROCESS rtl_pid;

  assert(PmInternalState != NULL);
  assert(PmInternalState->rtl != NULL);
  assert(sig != NULL);

  rtl_pid = sender((union SIGNAL **) &sig);

  free_buf((union SIGNAL **) &sig);

  rtl_info = traverseList1(PmInternalState->rtl, 
			   pm_rtl_search_by_pid,
			   (void *)rtl_pid);

  if(rtl_info)
    {
      freeNode(rtl_info);
    }
}

/*
 *===========================================================================
 *                        rtl_connect_request
 *===========================================================================
 * 
 * Description
 * -----------
 * Handle a RTL_CONNECT_REQUEST signal.
 *
 * Requirement
 * -----------
 * "RTL management"
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * PmInternalState                  The PM internal state structure.
 * request                          The RTL_CONNECT_REQUEST signal. 
 *
 * Returns
 * -------
 * None.
 *
 */

GLOBAL void
rtl_connect_request(struct PmInternalState * PmInternalState, 
		    struct RtlConnectRequest * request)
{
  PmStatus                      status;
  struct RtlConnectReply       *reply;
  PROCESS                       pid;

  assert(request != NULL);
  assert(PmInternalState != NULL);
  assert(PmInternalState->rtl != NULL);

  pid = sender((union SIGNAL **) &request);

  /* Make sure the RTL is not connected. */
  if(traverseList1(PmInternalState->rtl, 
		   pm_rtl_search_by_pid,
		   (void *) pid))
    {
      /* The RTL is already connected */
      status = RTL_EALREADY_CONNECTED;
    }
  else
    {           
      SIGSELECT  *att_sig;
      RtlData    *rtl_info;
      union SIGNAL * int_rep;
      struct RtlInterfaceRequest * int_req;      
      static const SIGSELECT int_reply[] = {2, RTL_INTERFACE_REPLY, 
					    PM_INTERNAL_RTL_KILLED};
  
      if(strlen(request->ff_name) >= PM_FILE_FORMAT_NAME_LENGTH)
	{
	  status = RTL_EINTERFACE_MISMATCH;
	  goto pm_rtl_interface_reply;
	}

      rtl_info = allocNode(sizeof(RtlData)
			   + strlen(request->ff_name) + 1);

	  
      rtl_info->pid = pid;
      rtl_info->options = 0;	        
      strcpy((char *)rtl_info->name, (const char *)request->ff_name);
	  
      att_sig = (SIGSELECT *) alloc(sizeof(SIGSELECT), 
				    PM_INTERNAL_RTL_KILLED);
      rtl_info->attref = attach((union SIGNAL **) &att_sig, rtl_info->pid);
      
      /* Interface check of the rtls signal interface. */

      int_req = (struct RtlInterfaceRequest *)
	alloc(sizeof(struct RtlInterfaceRequest), RTL_INTERFACE_REQUEST);
	  
      send((union SIGNAL **)&int_req, pid);
      int_rep = receive_w_tmo(PM_RTL_CONNECT_TMO, int_reply);

      if(int_rep == NULL)
	{
	  /* detach will free the attach signal. */
	  detach(&(rtl_info->attref));
	  freeNode(rtl_info);
	  free_buf((union SIGNAL **) &request);
	  return;	  
	}

      if(int_rep->sigNo == RTL_INTERFACE_REPLY)
	{
	  int i;
	  for(i=1; i<PM_RTL_INTERFACE_SIZE+1; i++)
	    {
	      Boolean found = False;
	      unsigned int j;
	      
	      if(((struct RtlInterfaceReply *)int_rep)->sigs[i] 
		 == RTL_UNINSTALL_LOAD_MODULE_REQUEST)
	      {
		rtl_info->options = RTL_OPT_UNINSTALL;
	      }

	      for(j=1; j<((struct RtlInterfaceReply *)int_rep)->sigs[0] + 1; j++)
		{
		  if(rtl_sigs[i] == ((struct RtlInterfaceReply *)int_rep)->sigs[j])
		    {
		      found = True;
		    }
		}

	      if(found == False)
		{
		  /* The interfaces of pgm and rtl do not match. */
		  status = RTL_EINTERFACE_MISMATCH;
		  free_buf((union SIGNAL **) &int_rep);
		  detach(&(rtl_info->attref));
		  freeNode(rtl_info);
		  goto pm_rtl_interface_reply;
		}
	      else
		{		      
		  found = False;
		}
	    }

	  free_buf((union SIGNAL **) &int_rep);
	  insertFirst(PmInternalState->rtl, (void *)rtl_info);
	  status = PM_SUCCESS;
	}
      else
	{
	  /* The Attach signal! */
          free_buf((union SIGNAL **) &int_rep);
          freeNode(rtl_info);
          free_buf((union SIGNAL **) &request);
          return;
	}
    }

 pm_rtl_interface_reply:
  reply = (struct RtlConnectReply *) alloc(sizeof(struct RtlConnectReply), 
					   RTL_CONNECT_REPLY);
  reply->status = status;
  
  send((union SIGNAL **) &reply, pid);

  free_buf((union SIGNAL **) &request);
}

/*
 *===========================================================================
 *                        rtl_interface_request
 *===========================================================================
 * 
 * Description
 * -----------
 * Handle a RTL_INTERFACE_REQUEST signal.
 *
 * Requirement
 * -----------
 * "RTL management"
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * request                          The RTL_INTERFACE_REQUEST signal. 
 *
 * Returns
 * -------
 * None.
 *
 */

GLOBAL void
rtl_interface_request(struct RtlInterfaceRequest * request)
{
  struct RtlInterfaceReply * reply;

  assert(request != NULL);

  reply = (struct RtlInterfaceReply *) alloc(sizeof(struct PmInterfaceReply) 
					   + (sizeof(SIGSELECT) 
					      * (PM_RTL_INTERFACE_SIZE)), 
					   RTL_INTERFACE_REPLY);

  reply->status = PM_SUCCESS;

  /* The PM supported signals. */
  memcpy(reply->sigs, rtl_sigs, (PM_RTL_INTERFACE_SIZE + 1) * sizeof(SIGSELECT));

  /* Send the requrest. */
  send((union SIGNAL **) &reply, sender((union SIGNAL **)&request));

  /* Free the request buffer, no longer needed. */
  free_buf((union SIGNAL **) &request);
}

/*
 *===========================================================================
 *                        pm_file_formats_request
 *===========================================================================
 * 
 * Description
 * -----------
 * Handle a PM_FILE_FORMATS_REQUEST signal.
 *
 * Requirement
 * -----------
 * "RTL management"
 *
 * Side effects
 * ------------
 * None.
 *
 * Parameter description
 * ---------------------
 * PmInternalState                  The PM internal state structure.
 * request                          The PM_FILE_FORMATS_REQUEST signal. 
 *
 * Returns
 * -------
 * None.
 *
 */
GLOBAL void 
pm_file_formats_request(struct PmInternalState * PmInternalState,
			struct PmFileFormatsRequest * request)
{
  RtlData    *rtl_info;
  struct PmFileFormatsReply * reply;
  unsigned int i;
  char * file_formats;
  
  assert(request != NULL);
  assert(PmInternalState != NULL);
  assert(PmInternalState->rtl != NULL);

  /* loop the rtls to get all file format names. */
  i = PmInternalState->rtl->elements;
  
  reply = (struct PmFileFormatsReply *) alloc(sizeof(struct PmFileFormatsReply) 
					       + (i * PM_FILE_FORMAT_NAME_LENGTH), 
					       PM_FILE_FORMATS_REPLY);

  reply->file_formats_no_of_strings = i;

  file_formats = reply->file_formats;

  rtl_info = firstNode(PmInternalState->rtl);
  while(rtl_info != NULL)
    {
      strcpy(file_formats, rtl_info->name);
      file_formats += PM_FILE_FORMAT_NAME_LENGTH;
      rtl_info = nextNode(rtl_info);
    }

  reply->status = PM_SUCCESS;

  send((union SIGNAL **) &reply, sender((union SIGNAL **) &request));
  free_buf((union SIGNAL **) &request);
}


