/**************************************************************************
 * Copyright (C) 2010 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

#include "stdio.h"
#include "stdlib.h"
#include "assert.h"
#include "string.h"
#include "stdint.h"

#include "ose.h"
#include "list.h"
#include "cpu.h"
#include "ramlog.h"
#include "atomic.h"
#include "monitor.sig"
#include "monitor_execution.sig"
#include "monitor_events.sig"
#include "pm.sig"
#include "sysparam.h"
#include "ose_spi/mm.h"

#include "rmd.h"
#include "rmd_private.h"
#include "ose_events/rmd_events.h"
#include "rmd.sig"
#include "rmd_err.h"
#include "rmd_file.h"
#include "stdlib.h"

/* Function to obtain a persistent unique number. */
uint32_t ose_rmm_get_unique_number(void);

#define RMD_RECEIVE_TIMEOUT ~0UL
/* String to default location if rmd.path is not configured, max 63 chars */
#define DEFAULT_DUMP_PATH "/ram"

#define DUMP_INCOMPLETE "Dump incomplete due to insufficient memory."

#define END_OF_DUMP_SUCCESS "End of Dump, Status : Successful."
#define END_OF_DUMP_FAILURE "End of Dump, Status : Incomplete."

#define RMD_CLIENT_ATTACH_SIG 40097
typedef struct RmdClientAttachSig
{
   SIGSELECT signo;
   PROCESS client_pid;
} RmdClientAttachSig;

struct ProcQueueCnt
{
	uint32_t pid;
	uint32_t sig_cnt;
};

#define RMM_HUNT_REPLY (40098)
#define RMM_ATTACH_REPLY (40099)

/*
 ** This code heavily uses assignments in conditional expressions. This RVCT
 ** pragma transforms such compiler warnings into less disturbing remarks.
 */
/*
 #if defined(__ARMCC_VERSION)
 #pragma diag_remark 1293
 #endif
 */

/* Flag to set if the RMD should log events with ramlog_printf */
/* #define RMD_VERBOSE 1 */

/* Number of simultaneously ongoing dumps in all RMD processes. Changes to
 * this value must be handled atomic. */
static atomic_t ongoing_dumps = ATOMIC_INIT(0);

#define ALIGNED(addr, size) (((unsigned long) (addr) &\
                             (unsigned long) ((size) - 1UL)) == 0)

#define ALIGN_DOWN_TO(value, size) ((value) & ~((size) - 1UL))

#define ALIGN_UP_TO(value, size) (((value) + (size) - 1UL) & ~((size) - 1UL))

union SIGNAL
{
   SIGSELECT signo;

   /* Run Mode Dumper signals */
   RmdClientAttachSig rmd_client_attach_sig;
   RmdStatusReply rmd_reply;
   RmdInterfaceRequest rmd_interface_request;
   RmdInterfaceReply rmd_interface_reply;
   RmdStartDumpRequest rmd_start_dump_request;
   RmdStopDumpRequest rmd_stop_dump_request;
   RmdDumpErrorRequest rmd_dump_error_request;
   RmdDumpMemoryRequest rmd_dump_memory_request;
   RmdDumpStringRequest rmd_dump_string_request;
   RmdDumpProcessInfoRequest rmd_dump_process_info_request;
   RmdDumpStackUsageInfoRequest rmd_dump_stack_usage_info_request;
   RmdDumpEnvVarsInfoRequest rmd_dump_env_vars_info_request;
   RmdDumpSignalQueueInfoRequest rmd_dump_signal_queue_info_request;
   RmdDumpPoolInfoRequest rmd_dump_pool_info_request;
   RmdDumpPoolSignalsInfoRequest rmd_dump_pool_signals_info_request;
   RmdDumpSysParamInfoRequest rmd_dump_sys_param_info_request;
   RmdDumpRegistersInfoRequest rmd_dump_registers_info_request;
   RmdDumpProgramInfoRequest rmd_dump_program_info_request;
   RmdDumpTraceInfoRequest rmd_dump_trace_info_request;
   RmdDumpLoadModuleInfoRequest rmd_dump_load_module_info_request;
   RmdDumpPoolMemoryRequest rmd_dump_pool_memory_request;
   RmdDumpHeapMemoryRequest rmd_dump_heap_memory_request;
   RmdDumpProgramMemoryRequest rmd_dump_program_memory_request;
   RmdSetPropertyRequest rmd_set_property_request;
   RmdEndOfDumpRequest rmd_end_of_dump_request;
   RmdGetDumpInfoRequest rmd_get_dump_info_request;
   RmdGetDumpInfoReply rmd_get_dump_info_reply;

   /* Monitor signals */
   MonitorInterfaceRequest monitor_interface_request;
   MonitorInterfaceReply monitor_interface_reply;
   MonitorNameRequest monitor_name_request;
   MonitorNameReply monitor_name_reply;
   MonitorConnectRequest monitor_connect_request;
   MonitorConnectReply monitor_connect_reply;
   MonitorGetInfoRequest monitor_get_info_request;
   MonitorGetRegisterInfoRequest monitor_get_register_info_request;
   MonitorRegisterValues monitor_register_values;
   MonitorGetTraceRequest monitor_get_trace_request;
   MonitorGetInfoEndmark monitor_get_info_endmark;
   MonitorFlowControl monitor_flow_control;
   MonitorGetProcessInfoReply monitor_get_process_info_reply;
   MonitorGetProcessInfoLongReply monitor_get_process_info_long_reply;
   MonitorGetBlockInfoReply monitor_get_block_info_reply;
   MonitorGetSegmentInfoReply monitor_get_segment_info_reply;
   MonitorGetStackUsageReply monitor_get_stack_usage_reply;
   MonitorGetEnvVarsReply monitor_get_env_vars_reply;
   MonitorGetSignalQueueReply monitor_get_signal_queue_reply;
   MonitorGetPoolInfoReply monitor_get_pool_info_reply;
   MonitorGetPoolSignalsReply monitor_get_pool_signals_reply;
   MonitorGetSysParamReply monitor_get_sys_param_reply;
   MonitorGetHeapInfoReply monitor_get_heap_info_reply;
   MonitorGetHeapFragmentInfoReply monitor_get_heap_fragment_info_reply;
   MonitorGetHeapBuffersReply monitor_get_heap_buffers_reply;

   /* Program Manager signals */
   struct PmInterfaceReply pm_interface_reply;
   struct PmProgramsRequest programs_request;
   struct PmProgramsReply programs_reply;
   struct PmProgramInfo program_info;
   struct PmLoadModuleInstallHandlesRequest load_module_install_handle_request;
   struct PmLoadModuleInstallHandlesReply load_module_install_handle_reply;
   struct PmLoadModuleInfoRequest load_module_info_request;
   struct PmLoadModuleInfoReply load_module_info_reply;
   struct PmLoadModuleSectionInfoRequest load_module_section_info_request;
   struct PmLoadModuleSectionInfoReply load_module_section_info_reply;

};

#define RMD_MAGIC 0x7d9832f9

#define DUMP_MAGIC 0x7d9832fa

/* Struct that contains data about ongoing dumps. */
struct Dump
{
   ListLink link;
   uint32_t magic;
   OSATTREF dump_req_attach_ref;

   PROCESS pid;
   PROCESS bid;
   PROCESS sid;

   RmdDumpFile *dump_file;
};

/* Supported signals */
static const SIGSELECT INTERFACE_SIGNALS[] =
{ RMD_INTERFACE_REQUEST, RMD_START_DUMP_REQUEST, RMD_STOP_DUMP_REQUEST,
  RMD_DUMP_ERROR_REQUEST, RMD_DUMP_MEMORY_REQUEST, RMD_DUMP_STRING_REQUEST,
  RMD_DUMP_PROCESS_INFO_REQUEST, RMD_DUMP_STACK_USAGE_INFO_REQUEST,
  RMD_DUMP_ENV_VARS_INFO_REQUEST, RMD_DUMP_SIGNAL_QUEUE_INFO_REQUEST,
  RMD_DUMP_POOL_INFO_REQUEST, RMD_DUMP_POOL_SIGNALS_INFO_REQUEST,
  RMD_DUMP_SYS_PARAM_INFO_REQUEST, RMD_DUMP_REGISTERS_INFO_REQUEST,
  RMD_DUMP_PROGRAM_INFO_REQUEST, RMD_DUMP_TRACE_INFO_REQUEST,
  RMD_DUMP_LOAD_MODULE_INFO_REQUEST, RMD_DUMP_POOL_MEMORY_REQUEST,
  RMD_DUMP_HEAP_MEMORY_REQUEST, RMD_DUMP_PROGRAM_MEMORY_REQUEST,
  RMD_SET_PROPERTY_REQUEST,RMD_END_OF_DUMP_REQUEST};

/**
 * @function rmd_private_interface_request
 *
 * @brief Creates and sends an interface reply signal.
 *
 * @input to_pid Process id to send the interface reply signal to.
 *
 */
static void
rmd_private_interface_request(PROCESS to_pid)
{

   union SIGNAL *sig;
   uint32_t numSigs = sizeof(INTERFACE_SIGNALS) / sizeof(SIGSELECT);

   sig = alloc(sizeof(RmdInterfaceReply) + numSigs * sizeof(uint32_t),
            RMD_INTERFACE_REPLY);

   sig->rmd_interface_reply.sigsCount = numSigs;
   sig->rmd_interface_reply.reserved0 = 0;
   sig->rmd_interface_reply.reserved1 = 0;
   sig->rmd_interface_reply.reserved2 = 0;
   memcpy(sig->rmd_interface_reply.sigs, INTERFACE_SIGNALS, numSigs
            * sizeof(uint32_t));

   send(&sig, to_pid);
} /* rmd_private_interface_request() */

/**
 * @function rmd_file_status_converter
 *
 * @brief Converts status from RmdFileStatus to RmdStatus.
 *
 * @input rmd_file_status The file status.
 *
 * @return The status corresponding to the RmdFileStatus
 *
 */
static RmdStatus
rmd_file_status_converter(RmdFileStatus rmd_file_status,
         struct RmdDumpFile *dump_file)
{
#ifdef RMD_VERBOSE
   switch (rmd_file_status)
   {
      case RMD_FILE_SUCCESS:
      break;

      case RMD_FILE_BLOCK_ERROR:
      ramlog_printf(
               "RMD_FILE_BLOCK_ERROR: Reply or endmark written before request in signal block.\n");
      break;
      case RMD_FILE_MEM_ERROR:
    	  rmd_file_dump_text_block(DUMP_INCOMPLETE,dump_file);
      ramlog_printf(
               "RMD_FILE_MEM_ERROR: No memory to allocate internal buffers for file writer.\n");
      break;
      case RMD_FILE_SEEK_ERROR:
      ramlog_printf("RMD_FILE_SEEK_ERROR: Error in fseek().\n");
      break;
      case RMD_FILE_SYSTEM_ERROR:
      ramlog_printf("RMD_FILE_SYSTEM_ERROR: .\n");
      break;
      case RMD_FILE_TELL_ERROR:
      ramlog_printf("RMD_FILE_TELL_ERROR: Error in ftell().\n");
      break;
      case RMD_FILE_WRITE_ERROR:
      ramlog_printf("RMD_FILE_WRITE_ERROR: Error in fwrite().\n");
      break;
      case RMD_FILE_OPEN_ERROR:
      ramlog_printf("RMD_FILE_OPEN_ERROR: Error in fopen().\n");
      break;
      case RMD_FILE_WRITE_SHORT_ERROR:
      ramlog_printf(
               "RMD_FILE_WRITE_SHORT_ERROR: Data written is shorter than expected, file system full?\n");
      break;
      case RMD_FILE_FLUSH_ERROR:
      ramlog_printf("RMD_FILE_FLUSH_ERROR: Error in fflush.\n");
      break;
      case RMD_FILE_ERROR:
      ramlog_printf("RMD_FILE_ERROR: Error in fflush.\n");
      break;
      case RMD_FILE_ZLIB_ERROR:
      ramlog_printf("RMD_FILE_ZLIB_ERROR: Error in Zlib.\n");
      ramlog_printf("ZLIB error message: %s\n",
               dump_file->compression->zstrm->msg);
      break;
      case RMD_FILE_Z_STREAM_END:
      ramlog_printf(
               "RMD_FILE_Z_STREAM_END: Unexpected end of compression stream.\n");
      ramlog_printf("ZLIB error message: %s\n",
               dump_file->compression->zstrm->msg);
      break;
      case RMD_FILE_Z_STREAM_ERROR:
      ramlog_printf(
               "RMD_FILE_Z_STREAM_ERROR: Error in compression stream.\n");
      ramlog_printf("ZLIB error message: %s\n",
               dump_file->compression->zstrm->msg);
      break;

      default:
      ramlog_printf("RMD: Unknown error.\n");
      break;
   }
#else
   /* Only used for printing, do nothing do get rid of warnings */
   (void) dump_file;
#endif

   switch (rmd_file_status)
   {
      case RMD_FILE_SUCCESS:
         return RMD_STATUS_OK;
      case RMD_FILE_BLOCK_ERROR:
         return RMD_STATUS_INTERNAL_ERROR;
      case RMD_FILE_MEM_ERROR:
    	  rmd_file_dump_text_block(DUMP_INCOMPLETE,dump_file);
         return RMD_STATUS_MEMORY_ALLOCATION_ERROR;
      case RMD_FILE_SEEK_ERROR:
      case RMD_FILE_SYSTEM_ERROR:
      case RMD_FILE_TELL_ERROR:
      case RMD_FILE_WRITE_ERROR:
      case RMD_FILE_WRITE_SHORT_ERROR:
      case RMD_FILE_FLUSH_ERROR:
      case RMD_FILE_ERROR:
         return RMD_STATUS_FILE_ERROR;
      case RMD_FILE_ZLIB_ERROR:
      case RMD_FILE_Z_STREAM_END:
      case RMD_FILE_Z_STREAM_ERROR:
         return RMD_STATUS_ZLIB_ERROR;
      case RMD_FILE_OPEN_ERROR:
         return RMD_STATUS_FILE_OPEN_ERROR;

      default:
         return RMD_STATUS_INTERNAL_ERROR;
   }

} /* rmd_file_status_converter() */

/* Dump file name template */
#define DUMP_FILE_NAME_TEMPLATE "/rmd_%010u_0x%08lx.pmd"
#define DUMP_FILE_NAME_TEMPLATE_LENGTH 30 

/**
 * @function generate_dump_file_name
 *
 * @brief A helper function for rmd_start_dump that generates the file name
 * from the given criteria.
 *
 * @long A helper function for rmd_start_dump that generates the file name
 * from the given criteria. The function allocates memory from the pool and
 * writes the generated or specified name to the file_name buffer. The
 * file_name buffer must be freed with free_buf() by caller
 *
 * @input req The RMD_DUMP_START_REQUEST signal containing the input in .path
 * @input dump_file_serial_nbr Serial number to use as part of automatic generated file name.
 * @input rmd->path Configuration parameter containing default path from system parameter.
 *
 * @output file_name A null terminated character string residing in the pool.
 * Must be free_buf()'ed by the caller.
 *
 * @return RMD_STATUS_OK if OK, RMD_STATUS_DUMP_FILE_NAME_GENERATION_ERROR
 * otherwise.
 *
 * @seealso rmd_start_dump
 *
 */
static RmdStatus
generate_dump_file_name(union SIGNAL *req, char **file_name,
         uint32_t dump_file_serial_nbr, struct Rmd *rmd)
{
   RmdStatus status = RMD_STATUS_OK;
   *file_name = NULL;
   if (strnlen(req->rmd_start_dump_request.path, sigsize(&req)) == 0)
   {
      /* Get rmd path from configuration and append a generated file name */
      char *tmp;
      size_t size = strlen(rmd->path) + DUMP_FILE_NAME_TEMPLATE_LENGTH + 1;

      /* Allocate memory for path and file name */
      *file_name = (char *) alloc(size, 0);

      /* Copy path */
      strcpy(*file_name, rmd->path);

      /* Add generated file name */
      tmp = *file_name + strlen(rmd->path);
      sprintf(tmp, DUMP_FILE_NAME_TEMPLATE, dump_file_serial_nbr, sender(&req));
   }
   else if (req->rmd_start_dump_request.path[0] == '/')
   {
      /* Absolute path in signal, use it as is */
      size_t size = strnlen(req->rmd_start_dump_request.path, sigsize(&req))
               + 1;
      *file_name = (char *) alloc(size, 0);
      strncpy(*file_name, req->rmd_start_dump_request.path, size);
   }
   else
   {
      /* The signal is holding a file name only. Get configured rmd path
       * and append file name from signal */
      size_t size = strnlen(req->rmd_start_dump_request.path, sigsize(&req))
               + 1;
      char *tmp;
      /* Allocate memory for path and file name */
      *file_name = (char *) alloc(strlen(rmd->path) + size + 1, 0);

      /* Copy path */
      strcpy(*file_name, rmd->path);

      /* Add generated file name */
      tmp = *file_name + strlen(rmd->path);
      *tmp = '/';
      tmp++;
      strncpy(tmp, req->rmd_start_dump_request.path, size);
   }
   if (*file_name == NULL)
   {
      /* Handle error */
#ifdef RMD_VERBOSE
      ramlog_printf("RMD: Failed to generate dump file name\n");
#endif
      status = RMD_STATUS_DUMP_FILE_NAME_GENERATION_ERROR;
   }

   return status;
} /* generate_dump_file_name() */

/**
 * get_dump_from_pid
 *
 * @brief Finds the dump associated with a certain PID in the dump list.
 *
 * @input pid PID for the dump.
 * @input dump_list Handle to list of dumps to search through.
 *
 * @return Pointer to the dump corresponding to the requested PID, NULL if no
 * dump is found with that PID.
 *
 */
static struct Dump *
get_dump_from_pid(PROCESS pid, const List *dump_list)
{
   struct Dump *current_dump;
   struct Dump *ret = NULL;

   current_dump = (struct Dump *) list_head(dump_list);

   while (current_dump != NULL)
   {
      /* No fancy search algorithm here since the dump list should be short */
      if (pid == current_dump->pid)
      {
         /* Found the dump with the corresponding pid */
         ret = current_dump;
         break; /* while(current_dump != NULL) */
      }
      current_dump = (struct Dump *) list_next(dump_list, current_dump);
   }

   return ret;
} /* get_dump_from_pid */

/**
 * @function get_supported_interfaces_from_rmm
 *
 * @brief Sends the interface request signal and writes down the request and reply
 * signals to the dump.
 *
 * @long Sends the signal MONITOR_INTERFACE_REQUEST to the process id stored in
 * rmd->rmm_main. Writes the request and reply signal to the dump.
 *
 * @input rmd The Rmd struct that holds the information about the dump sessions for
 * this RMD process.
 *
 * @input dump The active dump.
 *
 * @output rmd->supported_monitor_signals An array of supported signal numbers to be
 * used for subsequent calls to rmm_is_supported. Must be free_buf()'ed by the caller.

 * @output rmd->supported_monitor_signals_count number of signals in
 * rmd->supported_monitor_signals
 *
 * @return
 *
 * @seealso rmm_is_supported
 *
 */
static RmdFileStatus
get_supported_interfaces_from_rmm(struct Rmd *rmd, struct Dump *dump)
{
   RmdStatus status = RMD_STATUS_OK;
   RmdFileStatus file_status = RMD_FILE_SUCCESS;
   union SIGNAL *sig;
   uint32_t i;

   static const SIGSELECT rmm_return_sigs[] =
   { 1, MONITOR_INTERFACE_REPLY };

   sig = alloc(sizeof(MonitorInterfaceRequest), MONITOR_INTERFACE_REQUEST);

   /* Write request signal to filedump */
   file_status = rmd_file_dump_request((OSADDRESS) sig, sigsize(&sig),
            dump->dump_file);
   if (file_status != RMD_FILE_SUCCESS)
   {
      status = rmd_file_status_converter(file_status, dump->dump_file);
      return status;
   }

   send(&sig, rmd->rmm_main);
   sig = receive(rmm_return_sigs);

   /* Store supported interfaces for future use if we're not connected already */
   if (!rmd->rmm_connected)
   {
      rmd->supported_monitor_signals = (uint32_t *) alloc(
               sig->monitor_interface_reply.sigsCount * sizeof(uint32_t), 0);

      rmd->supported_monitor_signals_count
               = sig->monitor_interface_reply.sigsCount;
      for (i = 0; i < sig->monitor_interface_reply.sigsCount; i++)
      {
         rmd->supported_monitor_signals[i]
                  = sig->monitor_interface_reply.sigs[i];
      }
   }

   file_status = rmd_file_dump_endmark((OSADDRESS) sig, sigsize(&sig),
            sig->monitor_get_info_endmark.status, dump->dump_file);

   status = rmd_file_status_converter(file_status, dump->dump_file);
   free_buf(&sig);
   return status;
} /* get_supported_interfaces_from_rmm() */

/**
 * @function connect_to_rmm
 *
 * @brief Sends the interface request signal and writes down the request and reply
 * signals to the dump.
 *
 * @long Sends the signal MONITOR_INTERFACE_REQUEST to the process id stored in
 * rmd->rmm_main. Writes the request and reply signal to the dump.
 *
 * @input rmd The Rmd struct that holds the information about the dump sessions for
 * this RMD process.
 *
 * @input dump The active dump.
 *
 * @output rmd->supported_monitor_signals_count number of signals in
 * rmd->supported_monitor_signals
 *
 * @output rmd->attach_ref Reference to the client RMM process
 *
 * @output rmd->rmm Reference to the client RMM process
 *
 * @output rmd->stored_connect_reply Stored reply signal for use with
 * stored_connect_reply
 *
 * @return
 *
 * @seealso stored_connect_to_rmm
 *
 */
static RmdStatus
connect_to_rmm(struct Rmd *rmd, struct Dump *dump)
{
   RmdStatus status = RMD_STATUS_OK;
   RmdFileStatus file_status = RMD_FILE_SUCCESS;
   char rmd_proc_name[12];
   union SIGNAL *sig;
   const SIGSELECT rmm_return_sigs[] =
   { 1, MONITOR_CONNECT_REPLY };

   sprintf(rmd_proc_name, "ose_rmd%02d", ose_cpu_id());
   sig = alloc(sizeof(MonitorConnectRequest) + 12, MONITOR_CONNECT_REQUEST);
   strncpy(sig->monitor_connect_request.name, rmd_proc_name, 12);

   status = rmd_file_dump_request((OSADDRESS) sig, sigsize(&sig),
            dump->dump_file);
   /*   status = write_file_signal_data(sig, RMD_FILE_WRITE_REQUEST, dump);*/
   if (status != RMD_STATUS_OK)
   {
      /* Error */
      return status;
   }
   send(&sig, rmd->rmm_main);

   sig = receive_w_tmo(RMD_RECEIVE_TIMEOUT, rmm_return_sigs);
   assert(sig);

   /* Write endmark signal to file dump */
   file_status = rmd_file_dump_endmark((OSADDRESS) sig, sigsize(&sig),
            sig->monitor_get_info_endmark.status, dump->dump_file);

   status = rmd_file_status_converter(file_status, dump->dump_file);

   /*status = write_file_signal_data(sig, RMD_FILE_WRITE_ENDMARK, dump);*/
   if (status != RMD_STATUS_OK)
   {
      /* Error */
      return status;
   }

   if (MONITOR_STATUS_OK != sig->monitor_connect_reply.status)
   {
      free_buf(&sig);
      return RMD_STATUS_CONNECT_TO_RMM_FAILED;
   }

   rmd->rmm_connected = TRUE;
   /* Change over to the client rmm process. */
   rmd->rmm = sender(&sig);

   rmd->stored_connect_reply = sig;
   /*free_buf(&sig);*/

   detach(&rmd->attach_ref);
   sig = alloc(sizeof(SIGSELECT), RMM_ATTACH_REPLY);
   rmd->attach_ref = attach(&sig, rmd->rmm);
#ifdef RMD_VERBOSE
   ramlog_printf("RMD: Connected to monitor\n");
#endif
   return status;
}/* connect_to_rmm() */

/**
 * @function stored_connect_to_rmm
 *
 * @brief Writes the connect request and reply signals without sending anything to
 * the RMM.
 *
 * @long Creates the request signal and uses the stored signal in
 * rmd->stored_connect_reply to write to the dump file header. This function is used
 * instead of connect_to_rmm in rmd_start_dump if this RMD process is already connected
 * to the RMM.
 *
 * @input rmd The Rmd struct that holds the information about the dump sessions for
 * this RMD process.
 *
 * @input dump The active dump.
 *
 * @output rmd->supported_monitor_signals_count number of signals in
 * rmd->supported_monitor_signals
 *
 * @output rmd->attach_ref Reference to the client RMM process
 *
 * @output rmd->rmm Reference to the client RMM process
 *
 * @input rmd->stored_connect_reply Stored reply signal from connect_to_rmm
 *
 * @return
 *
 * @seealso connect_to_rmm
 * @seealso rmd_start_dump
 *
 */
static RmdStatus
stored_connect_to_rmm(struct Rmd *rmd, struct Dump *dump)
{
   RmdStatus status = RMD_STATUS_OK;
   char rmd_proc_name[12];
   union SIGNAL *sig;

   if (NIL == rmd->stored_connect_reply)
   {
#ifdef RMD_VERBOSE
      ramlog_printf("RMD: stored_connect_to_rmm(): "
               "stored_connect_reply = NIL, "
               "bailing out\n");
#endif
      return RMD_STATUS_INTERNAL_ERROR;
   }

   sprintf(rmd_proc_name, "ose_rmd%02d", ose_cpu_id());
   sig = alloc(sizeof(MonitorConnectRequest) + 12, MONITOR_CONNECT_REQUEST);
   strncpy(sig->monitor_connect_request.name, rmd_proc_name, 12);

   status = rmd_file_dump_request((OSADDRESS) sig, sigsize(&sig),
            dump->dump_file);

   free_buf(&sig);
   if (status != RMD_STATUS_OK)
   {
      /* Error */
      return status;
   }

   /* Write stored signal to file dump */
   status = rmd_file_dump_endmark((OSADDRESS) rmd->stored_connect_reply,
            sigsize(&rmd->stored_connect_reply),
            rmd->stored_connect_reply->monitor_connect_reply.status,
            dump->dump_file);
   /*   status = write_file_signal_data(rmd->stored_connect_reply,
    RMD_FILE_WRITE_ENDMARK, dump);*/
   return status;
}/* stored_connect_to_rmm() */

/**
 * @function hunt_for_pm
 *
 * @brief A helper function to make the code more readable that hunts for the program
 * manager process (named "ose_prgram_mgr") and returns the PID of that process.
 *
 * @return The PID of the program manager process.
 *
 */
#define PM_HUNT_REPLY 34343
static PROCESS
hunt_for_pm(void)
{
   PROCESS pm_process;
   union SIGNAL *sig;
   static const SIGSELECT pm_hunt_sigs[] =
   { 1, PM_HUNT_REPLY };

   /* Hunt for program manager */
   pm_process = 0;
   sig = alloc(sizeof(SIGSELECT), PM_HUNT_REPLY);
   hunt("ose_program_mgr", 0, 0, &sig);
   sig = receive(pm_hunt_sigs);

   if (sig != NULL)
   {
      pm_process = sender(&sig);
      free_buf(&sig);
   }

   return pm_process;
} /* hunt_for_pm() */

/**
 * @function pm_is_supported
 *
 * @brief Asks the program manager for supported interfaces and checks if
 * a signal number is part of that.
 *
 * @input support_sig The signal number that is tested.
 *
 * @input pm_process PID of the Program Manager.
 *
 * @returns TRUE if the signal is supported by the interface, otherwise FALSE.
 *
 */
static OSBOOLEAN
pm_is_supported(SIGSELECT support_sig, PROCESS pm_process)
{
   OSBOOLEAN supported = FALSE;

   /* Send signal for interface request and verify that the signal we want to send
    * is supported */
   if (pm_process != 0)
   {
      uint32_t i;
      union SIGNAL *sig;
      static const SIGSELECT pm_interface_reply_sigs[] =
      { 1, PM_INTERFACE_REPLY };

      sig = alloc(sizeof(struct PmInterfaceRequest), PM_INTERFACE_REQUEST);
      send(&sig, pm_process);

      sig = receive(pm_interface_reply_sigs);

      if (NIL != sig)
      {
         if (PM_INTERFACE_REPLY == sig->signo) /* What we were expecting to get */
         {
            /* element zero holds the number of elements in the sigs array */
            for (i = 1; i <= sig->pm_interface_reply.sigs[0]; i++)
            {
               if (support_sig == sig->pm_interface_reply.sigs[i])
               {
                  supported = TRUE;
                  break; /*for (i = 1; i <= sig->pm_interface_reply.sigs[0]; i++)*/
               }
            }
         }
         free_buf(&sig);
      }
   }

   return supported;
} /* pm_is_supported() */

/**
 * @function rmm_is_supported
 *
 * @brief Checks the stored interface request if the signal is supported.
 *
 * @input support_sig The signal number that is tested.
 *
 * @input rmd->supported_monitor_signal Array of supported signal numbers.
 *
 * @input rmd->supported_monitor_signals_count Number of signal numbers in
 * rmd->supported_monitor_signal.
 *
 * @returns TRUE if the signal is supported by the interface, otherwise FALSE.
 *
 */
static OSBOOLEAN
rmm_is_supported(SIGSELECT support_sig, const struct Rmd *rmd)
{
   OSBOOLEAN supported = FALSE;
   uint32_t i;

   if (rmd->supported_monitor_signals != NULL)
   {
      /* element zero holds the number of elements in the sigs array */
      for (i = 0; i < rmd->supported_monitor_signals_count; i++)
      {
         if (support_sig == rmd->supported_monitor_signals[i])
         {
            supported = TRUE;
            break; /*for (i = 1; i <= sig->pm_interface_reply.sigs[0]; i++)*/
         }
      }
   }
   return supported;
} /* rmm_is_supported() */


static int
rmd_proc_compslot (const void *e1, const void *e2)
{
   const struct ProcQueueCnt *t1 = (const struct ProcQueueCnt*)e1;
   const struct ProcQueueCnt *t2 = (const struct ProcQueueCnt*)e2;
   return  (int)(((int64_t)t2->sig_cnt) - ((int64_t)t1->sig_cnt));
}

static RmdStatus get_process_long_info(union SIGNAL *req, struct Rmd *rmd, 
                       struct ProcQueueCnt *proc_queue_cnt,uint32_t *count)
{
   RmdStatus status = RMD_STATUS_OK;
   uint32_t signo;
   uint32_t i =0;
   union SIGNAL *reply = NIL;
   const SIGSELECT process_reply[] =
   { 4, RMM_ATTACH_REPLY,
        MONITOR_GET_PROCESS_INFO_LONG_REPLY,
        MONITOR_GET_PROCESS_INFO_ENDMARK,
        MONITOR_BLOCKING_FLOW_CONTROL_REQUEST
   };

   /* Request process information from monitor */
   send((union SIGNAL**)&req, rmd->rmm);


   do /* while (reply->signo != MONITOR_GET_PROCESS_INFO_ENDMARK) */
   {
      reply = receive_from(RMD_RECEIVE_TIMEOUT, process_reply, rmd->rmm);
      assert(reply);
      signo = reply->signo;
      switch (signo)
      {
	 case RMM_ATTACH_REPLY:
	    status = RMD_STATUS_DETACHED_FROM_RMM;
	    free_buf(&reply);
	    goto finally;

	 case MONITOR_GET_PROCESS_INFO_LONG_REPLY:

	    if (i < rmd->rmd_properties.num_long_queue_procs)
	    {
	       (proc_queue_cnt+i)->pid =
                 reply->monitor_get_process_info_long_reply.pid;
	       (proc_queue_cnt+i)->sig_cnt =
                 reply->monitor_get_process_info_long_reply.signalsInQueue;
	       i++;
               qsort(proc_queue_cnt, i, sizeof(struct ProcQueueCnt),
                 rmd_proc_compslot); 
	       *count = i;
	    }
            else
            {
	       (proc_queue_cnt+i)->pid =
                 reply->monitor_get_process_info_long_reply.pid;
	       (proc_queue_cnt+i)->sig_cnt =
                 reply->monitor_get_process_info_long_reply.signalsInQueue;
               qsort(proc_queue_cnt, i+1, sizeof(struct ProcQueueCnt),
                 rmd_proc_compslot); 
            

            }
	    free_buf(&reply);

	    break;
	 case MONITOR_BLOCKING_FLOW_CONTROL_REQUEST:
	    reply->signo = MONITOR_BLOCKING_FLOW_CONTROL_REPLY;
	    send(&reply, rmd->rmm);
	    break;
	 case MONITOR_GET_PROCESS_INFO_ENDMARK:
	    break;
	 default:
	    status = RMD_STATUS_INTERNAL_ERROR;
	    break;
      }
   } while (signo != MONITOR_GET_PROCESS_INFO_ENDMARK);
   free_buf(&reply);

   finally: return status;
}

/**
 * @function rmd_private_start_dump
 *
 * @brief Initialises a dump and sets up the file for subsequent calls to other
 * dump functions.
 *
 * @long
 * All dumps must be initialised by calling rmd_private_start_dump before any
 * other functions in the rmd_private interface can be used. The dump must be closed
 * by calling rmd_private_stop_dump.
 * The initialisation process is the following:
 *  * Generate a file name from the input in req->path.
 *  * Open the dump file and write the file header.
 *  * Create a new dump node and add to the dump list.
 *  * Attach to the process that requested the dump.
 *  * Get supported interfaces from RMM.
 *  * Connect to RMM.
 *  * Signal start dump event.
 *
 * @input req The request signal of type RmdStartDumpRequest.
 *
 * @input rmd The Rmd struct holding information about the dumps in this process.
 *
 * @output Replies with the signal RMD_START_DUMP_REPLY to the sender of the request
 * signal when finished setting up the dump.
 *
 *
 * @seealso generate_dump_file_name
 * @seealso get_supported_interfaces_from_rmm
 * @seealso connect_to_rmm
 * @seealso stored_connect_to_rmm
 * @seealso rmd_private_stop_dump
 * @seealso RmdStartDumpRequest
 */
void
rmd_private_start_dump(union SIGNAL *req, struct Rmd *rmd)
{
   union SIGNAL *sig;
   RmdStatus status = RMD_STATUS_INTERNAL_ERROR;
   char *file_name = NULL;
   struct Dump *dump = NULL;
   RmdFileStatus file_status;
   RmdDumpFile *dump_file = NULL;
   uint32_t file_serial_no;

   /* Validate request signal */
   assert(req);
   assert(RMD_START_DUMP_REQUEST == req->signo);
   assert(rmd);
   assert(RMD_MAGIC == rmd->magic);

   if (0 != req->rmd_start_dump_request.dump_pid)
   {
      status = RMD_STATUS_INTERFACE_NOT_SUPPORTED;
      goto finally;
   }

   file_serial_no = ose_rmm_get_unique_number() | 0x80000000UL;
   status = generate_dump_file_name(req, &file_name, file_serial_no, rmd);
   if (status != RMD_STATUS_OK)
   {
      goto finally;
   }
#ifdef RMD_VERBOSE
   ramlog_printf(
            "RMD: generate_dump_file_name status: %d, dump file name: %s\n",
            status, file_name);
#endif

   /* Check that number of maximum ongoing dumps is not exceeded */
#ifdef RMD_VERBOSE
   ramlog_printf("RMD: max_ongoing_dumps: %d\n", rmd->max_ongoing_dumps);
#endif
   if (atomic_inc_return(&ongoing_dumps) > rmd->max_ongoing_dumps)
   {
      atomic_dec(&ongoing_dumps);
      error2(OSE_RMD_ETOO_MANY_DUMPS, rmd->max_ongoing_dumps);
      status = RMD_STATUS_TOO_MANY_DUMPS;
      goto finally;
   }

   /* Open dump file and check status */
   dump_file = (RmdDumpFile *) alloc(sizeof(RmdDumpFile) + strlen(file_name), 0);
   dump_file->file_name[0] = 0;
#ifdef RMD_VERBOSE
   ramlog_printf("RMD: Calling rmd_file_start_dump()\n");
#endif
   file_status = rmd_file_start_dump(file_name, file_serial_no,
            rmd->use_compression, rmd->window_bits, rmd->mem_level, dump_file);
   if (file_status != RMD_FILE_SUCCESS)
   {
      /* Error */
      status = rmd_file_status_converter(file_status, dump_file);
      goto finally;
   }
   /* Save the file name for later retrieval by rmd_get_dump_info. */
   strcpy(dump_file->file_name, file_name);
   
   /* alloc and insert new Dump node */
   dump = (struct Dump *) alloc(sizeof(struct Dump), 0);
   list_init_link(dump);
   dump->magic = DUMP_MAGIC;
   dump->dump_req_attach_ref = 0;
   dump->pid = sender(&req);
   dump->bid = get_bid(sender(&req));
   dump->sid = get_segid(sender(&req));
   dump->dump_file = dump_file;
   list_insert_head(&rmd->dumps, dump);

   /* Attach to the dump requester (sender) */
   sig = alloc(sizeof(RmdClientAttachSig), RMD_CLIENT_ATTACH_SIG);
   sig->rmd_client_attach_sig.client_pid = dump->pid;
   dump->dump_req_attach_ref = attach(&sig, dump->pid);

   /* Get supported interfaces from RMM */
   status = get_supported_interfaces_from_rmm(rmd, dump);
   if (status != RMD_STATUS_OK)
   {
      goto finally;
   }

   /* Connect to RMM process if not already connected*/
   if (!rmd->rmm_connected)
   {
      status = connect_to_rmm(rmd, dump);
      if (status != RMD_STATUS_OK)
      {
         goto finally;
      }
   }
   else
   {
      status = stored_connect_to_rmm(rmd, dump);
      if (status != RMD_STATUS_OK)
      {
         goto finally;
      }
   }

   /* Signal start dump event */
   if (rmd->start_dump_event_trigger)
   {
      RmdStartDumpEvent *event_data = (RmdStartDumpEvent *) alloc(
               sizeof(RmdStartDumpEvent) + strlen(file_name) + 1, 0);
      strcpy(event_data->file_name, file_name);
      event_data->pid = dump->pid;
      event_data->reserved0 = 0;
      event_data->reserved1 = 0;

      signal_event(RMD_EVENT_START_DUMP, event_data, sizeof(RmdStartDumpEvent)
               + strlen(file_name) + 1);
      free_buf((union SIGNAL **) &event_data);
   }

   finally:
   /* Clean up */
   if (status != RMD_STATUS_OK)
   {
      if (dump_file != NULL)
      {
         free_buf((union SIGNAL **) &dump_file);
         atomic_dec(&ongoing_dumps);
      }
      if (dump != NULL)
      {
         free_buf((union SIGNAL **) &dump);
         list_unlink(dump);
      }
   }

   if (file_name != NULL)
   {
      free_buf((union SIGNAL **) &file_name);
   }

   /* Send reply signal */
   sig = alloc(sizeof(RmdStatusReply), RMD_START_DUMP_REPLY);
   sig->rmd_reply.status = status;
   send(&sig, sender(&req));
} /* rmd_private_start_dump() */

/**
 * @function rmd_private_stop_dump
 *
 * @brief Closes the dump associated with the PID of the requester.
 *
 * @long The function is called in two different cases
 * 1. The dumping process sends the signal RmdStopDumpRequest
 * 2. The client dies and the kernel sends a signal.
 * These two cases must be handled differently when getting the PID and sending a
 * reply signal.
 *
 * @input rmd_req The request signal of type RmdStopDumpRequest or a signal that the
 *        client is detached.
 *
 * @input rmd The Rmd struct holding information about the dumps in this process.
 *
 *
 * @seealso rmd_private_start_dump
 * @seealso RmdStopDumpRequest
 *
 */
void
rmd_private_stop_dump(union SIGNAL *rmd_req, struct Rmd *rmd)
{
   union SIGNAL *sig;
   struct Dump *dump = NULL;
   PROCESS current_pid = 0;
   RmdStatus status = RMD_STATUS_OK;

   /* Validate request signal */
   assert(rmd_req);
   assert(RMD_STOP_DUMP_REQUEST == rmd_req->signo || rmd_req->signo == RMD_CLIENT_ATTACH_SIG);
   assert(rmd);
   assert(RMD_MAGIC == rmd->magic);

#ifdef RMD_VERBOSE
   ramlog_printf("RMD: Stop dump\n");
#endif

   /* This function can be called with two different input signals.
    * The PID is obtained differently in the two cases.
    */
   if (rmd_req->signo == RMD_STOP_DUMP_REQUEST)
      current_pid = sender(&rmd_req);
   else if (rmd_req->signo == RMD_CLIENT_ATTACH_SIG)
      current_pid = rmd_req->rmd_client_attach_sig.client_pid;

   dump = get_dump_from_pid(current_pid, &rmd->dumps);

   if (NULL == dump)
   {
#ifdef RMD_VERBOSE
      ramlog_printf("RMD: Couldn't find a dump from the pid 0x%x\n",
               current_pid);
#endif
      status = RMD_STATUS_DUMP_NOT_FOUND;
      goto finally;
   }

   /* Call file library to flush any pending data and free all resources used */
   rmd_file_end_dump(dump->dump_file);

   /* detach from the dump requester (sender) */
   if (rmd_req->signo == RMD_STOP_DUMP_REQUEST)
   {
      detach(&dump->dump_req_attach_ref);
   }

   /* remove and free Dump node */
   list_unlink(dump);
   free_buf((union SIGNAL **) &dump->dump_file);
   free_buf((union SIGNAL **) &dump);
   atomic_dec(&ongoing_dumps);

   /* disconnect from the ose_monitor if no ongoing dumps left */
   if (0 == list_length(&rmd->dumps))
   {
      /* Disconnect from RMM */
      sig = alloc(sizeof(MonitorDisconnectRequest), MONITOR_DISCONNECT_REQUEST);
      send(&sig, rmd->rmm);
      rmd->rmm_connected = FALSE;

      /* Detach from RMM worker */
      detach(&rmd->attach_ref);
      rmd->rmm = 0;
      /* Attach to main RMM again */
      sig = alloc(sizeof(SIGSELECT), RMM_ATTACH_REPLY);
      rmd->attach_ref = attach(&sig, rmd->rmm_main);

      /* Free up resources not needed when not dumping */
      free_buf(&rmd->stored_connect_reply);
      rmd->stored_connect_reply = NIL;
      free_buf((union SIGNAL **) &rmd->supported_monitor_signals);
      rmd->supported_monitor_signals = NULL;
      rmd->supported_monitor_signals_count = 0;
   }

   /* Signal stop dump event */
   if (rmd->stop_dump_event_trigger)
   {
      RmdStopDumpEvent *event_data = (RmdStopDumpEvent *) alloc(
               sizeof(RmdStopDumpEvent), 0);
      event_data->pid = current_pid;
      event_data->reserved0 = 0;
      event_data->reserved1 = 0;

      signal_event(RMD_EVENT_STOP_DUMP, event_data, sizeof(RmdStopDumpEvent));
      free_buf((union SIGNAL **) &event_data);
   }

   finally:
   /* Send reply signal if we're called with a stop dump request */
   if (rmd_req->signo == RMD_STOP_DUMP_REQUEST)
   {
      /* Send reply signal */
      sig = alloc(sizeof(RmdStatusReply), RMD_STOP_DUMP_REPLY);
      sig->rmd_reply.status = status;
      send(&sig, sender(&rmd_req));
   }
} /* rmd_private_stop_dump() */

/**
 * @function rmd_private_dump_error
 *
 * @brief Dumps error information to the dump.
 *
 * @input req The request signal of type RmdDumpErrorRequest
 *
 * @input rmd The Rmd struct holding information about the dumps in this process.
 *
 * @seealso RmdDumpErrorRequest
 *
 */
void
rmd_private_dump_error(union SIGNAL *req, struct Rmd *rmd)
{
   union SIGNAL *sig;
   struct Dump *dump;
   RmdFileStatus rmd_file_status = RMD_FILE_SUCCESS;
   RmdStatus return_status = RMD_STATUS_OK;

   assert(req);
   assert(RMD_DUMP_ERROR_REQUEST == req->signo);
   assert(rmd);
   assert(RMD_MAGIC == rmd->magic);

#ifdef RMD_VERBOSE
   ramlog_printf("RMD: rmd_private_dump_error()\n");
#endif

   dump = get_dump_from_pid(sender(&req), &rmd->dumps);
   if (NULL == dump)
   {
#ifdef RMD_VERBOSE
      ramlog_printf("RMD: Couldn't find a dump from the pid 0x%x\n", sender(
                        &req));
#endif
      return_status = RMD_STATUS_DUMP_NOT_FOUND;
   }
   else
   {
      rmd_file_status
               = rmd_file_dump_error_block(
                        req->rmd_dump_error_request.description,
                        req->rmd_dump_error_request.user_called,
                        req->rmd_dump_error_request.ecode,
                        req->rmd_dump_error_request.extra, sender(&req),
                        dump->dump_file);

      return_status = rmd_file_status_converter(rmd_file_status,
               dump->dump_file);
   }

   /* Send reply signal */
   sig = alloc(sizeof(RmdStatusReply), RMD_DUMP_ERROR_REPLY);
   sig->rmd_reply.status = return_status;
   send(&sig, sender(&req));
} /* rmd_private_dump_error() */

/**
 * @function rmd_private_dump_memory
 *
 * @brief Dumps general memory.
 *
 * @long Implements two sanity checks on the parameters, if the size is 0 or the address
 * is NULL there will be no output in the dump file.
 *
 * @input req The request signal of type RmdDumpMemoryRequest
 *
 * @input rmd The Rmd struct holding information about the dumps in this process.
 *
 * @seealso RmdDumpMemoryRequest
 */
void
rmd_private_dump_memory(union SIGNAL *req, struct Rmd *rmd)
{
   union SIGNAL *sig;
   struct Dump *dump;
   RmdFileStatus rmd_file_status = RMD_FILE_SUCCESS;
   RmdStatus return_status;

   assert(req);
   assert(RMD_DUMP_MEMORY_REQUEST == req->signo);
   assert(rmd);
   assert(RMD_MAGIC == rmd->magic);

#ifdef RMD_VERBOSE
   ramlog_printf("RMD: Entering rmd_private_dump_memory()\n");
#endif
   dump = get_dump_from_pid(sender(&req), &rmd->dumps);
   if (NULL == dump)
   {
#ifdef RMD_VERBOSE
      ramlog_printf("RMD: Couldn't find a dump from the pid 0x%x\n", sender(
                        &req));
#endif
      return_status = RMD_STATUS_INTERNAL_ERROR;
   }
   else
   {
      uint32_t validated_length;
      unsigned int validation_result;
      /* Verify that the calling process has permission to read the
       * requested memory range */
      validated_length = ose_mm_validate_buffer(
               req->rmd_dump_memory_request.start_address,
               req->rmd_dump_memory_request.length, sender(&req), MM_PERM_READ,
               MM_VALIDATE_IN_MAS, &validation_result);

      if (validated_length != req->rmd_dump_memory_request.length)
      {
#ifdef RMD_VERBOSE
         ramlog_printf("RMD: The calling process (0x%x) does not have"
                  "permisssion to read address: 0x%x, size: %d\n", sender(&req),
                  req->rmd_dump_memory_request.start_address,
                  req->rmd_dump_memory_request.length);
#endif
         /* No access to memory, write empty block to dump */
         req->rmd_dump_memory_request.length = 0;
      }

      rmd_file_status = rmd_file_dump_memory_block(
               dump->bid,
               req->rmd_dump_memory_request.start_address,
               req->rmd_dump_memory_request.length,
               req->rmd_dump_memory_request.description, dump->dump_file);

      switch (rmd_file_status)
      {
         case RMD_FILE_SUCCESS:
            return_status = RMD_STATUS_OK;
            break;
         case RMD_FILE_WRITE_ERROR:
         case RMD_FILE_WRITE_SHORT_ERROR:
         case RMD_FILE_SEEK_ERROR:
            return_status = RMD_STATUS_FILE_ERROR;
            break;
         default:
            return_status = RMD_STATUS_INTERNAL_ERROR;
            break;
      }
   }

   /* Send reply signal */
   sig = alloc(sizeof(RmdStatusReply), RMD_DUMP_MEMORY_REPLY);
   sig->rmd_reply.status = return_status;
   send(&sig, sender(&req));
} /* rmd_private_dump_memory() */

/**
 * @function rmd_private_dump_string
 *
 * @brief Dumps a string.
 *
 * @input req The request signal of type RmdDumpStringRequest
 *
 * @input rmd The Rmd struct holding information about the dumps in this process.
 *
 * @seealso RmdDumpStringRequest
 */
void
rmd_private_dump_string(union SIGNAL *req, struct Rmd *rmd)
{
   union SIGNAL *sig;
   struct Dump *dump;
   RmdFileStatus rmd_file_status = RMD_FILE_SUCCESS;
   RmdStatus return_status;

#ifdef RMD_VERBOSE
   ramlog_printf("RMD: Entering rmd_private_dump_string()\n");
#endif

   dump = get_dump_from_pid(sender(&req), &rmd->dumps);
   if (NULL == dump)
   {
#ifdef RMD_VERBOSE
      ramlog_printf("RMD: Couldn't find a dump from the pid 0x%x\n", sender(
                        &req));
#endif
      return_status = RMD_STATUS_INVALID_PARAMETERS;
   }
   else
   {
      rmd_file_status = rmd_file_dump_text_block(
               req->rmd_dump_string_request.description, dump->dump_file);

      return_status = rmd_file_status_converter(rmd_file_status,
               dump->dump_file);
   }

   /* Send reply signal */
   sig = alloc(sizeof(RmdStatusReply), RMD_DUMP_STRING_REPLY);
   sig->rmd_reply.status = return_status;
   send(&sig, sender(&req));
} /* rmd_private_dump_string() */

/**
 * @function dump_process_info
 *
 * @brief Writes the signals of a MONITOR_GET_PROCESS_INFO_REQUEST transaction
 * to the dump file.
 *
 * @input scope The scope of the dump, can be Segment, Block or Process.
 * @input info The information parameters specifying what to dump.
 * @input dump The current dump.
 * @input rmd The Rmd struct holding information about the dumps in this process.
 *
 * @return  Returns a RmdStatus code
 *
 * @seealso RmdStatus
 *
 */
static RmdStatus
dump_process_info(uint32_t scope, uint32_t info, struct Dump *dump,
         struct Rmd *rmd)
{
   RmdStatus status = RMD_STATUS_OK;
   RmdFileStatus file_status = RMD_FILE_SUCCESS;
   uint32_t signo;
   union SIGNAL *reply = NIL;
   const SIGSELECT process_reply[] =
   { 5, RMM_ATTACH_REPLY, MONITOR_GET_PROCESS_INFO_REPLY,
     MONITOR_GET_PROCESS_INFO_LONG_REPLY, MONITOR_GET_PROCESS_INFO_ENDMARK,
     MONITOR_BLOCKING_FLOW_CONTROL_REQUEST };

   MonitorGetInfoRequest *req = (MonitorGetInfoRequest *) alloc(
            sizeof(MonitorGetInfoRequest), MONITOR_GET_PROCESS_INFO_REQUEST);

   req->level = (info & RMD_INFO_PROCESS_LONG) != 0;
   switch (scope)
   {
      case RMD_SCOPE_SEGMENT:
         req->scope.type = scope;
         req->scope.id = dump->sid;
         break;
      case RMD_SCOPE_BLOCK:
         req->scope.type = scope;
         req->scope.id = dump->bid;
         break;
      case RMD_SCOPE_PROCESS:
         req->scope.type = scope;
         req->scope.id = dump->pid;
         break;
      default:
         /* Handle error */
         free_buf((union SIGNAL **) &req);
         status = RMD_STATUS_INVALID_PARAMETERS;
         goto finally;
   }
   req->reserved0 = 0;
   req->tagsCount = 0;

   /* Write request signal to file dump */
   file_status = rmd_file_dump_request((OSADDRESS) req, sigsize(
            (union SIGNAL **) &req), dump->dump_file);
   if (file_status != RMD_FILE_SUCCESS)
   {
      free_buf((union SIGNAL **) &req);
      status = rmd_file_status_converter(file_status, dump->dump_file);
      goto finally;
   }

   /* Request process information from monitor */
   send((union SIGNAL **) &req, rmd->rmm);

   do /* while (reply->signo != MONITOR_GET_PROCESS_INFO_ENDMARK) */
   {
      reply = receive_from(RMD_RECEIVE_TIMEOUT, process_reply, rmd->rmm);
      assert(reply);
      signo = reply->signo;
      switch (signo)
      {
         case RMM_ATTACH_REPLY:
            status = RMD_STATUS_DETACHED_FROM_RMM;

            free_buf(&reply);
            goto finally;
         case MONITOR_GET_PROCESS_INFO_REPLY:
            /* Write reply signal to file dump */
            file_status = rmd_file_dump_reply((OSADDRESS) reply,
                     sigsize(&reply), dump->dump_file);

            free_buf(&reply);

            if (file_status != RMD_FILE_SUCCESS)
            {
               status = rmd_file_status_converter(file_status, dump->dump_file);
               goto finally;
            }

            break;
         case MONITOR_GET_PROCESS_INFO_LONG_REPLY:
            /* Write reply signal to file dump */
            file_status = rmd_file_dump_reply((OSADDRESS) reply,
                     sigsize(&reply), dump->dump_file);

            free_buf(&reply);

            if (file_status != RMD_FILE_SUCCESS)
            {
               status = rmd_file_status_converter(file_status, dump->dump_file);
               goto finally;
            }
            break;
         case MONITOR_BLOCKING_FLOW_CONTROL_REQUEST:
            reply->signo = MONITOR_BLOCKING_FLOW_CONTROL_REPLY;
            send(&reply, rmd->rmm);
            break;
         case MONITOR_GET_PROCESS_INFO_ENDMARK:
            break;
         default:
            status = RMD_STATUS_INTERNAL_ERROR;
            break;
      }
   } while (signo != MONITOR_GET_PROCESS_INFO_ENDMARK);

   /* Write endmark signal to filedump */
   file_status = rmd_file_dump_endmark((OSADDRESS) reply, sigsize(&reply),
            reply->monitor_get_info_endmark.status, dump->dump_file);
   free_buf(&reply);
   status = rmd_file_status_converter(file_status, dump->dump_file);

   finally: return status;
} /* dump_process_info() */

/**
 * @function dump_block_info
 *
 * @brief Writes the signals of a MONITOR_GET_BLOCK_INFO_REQUEST transaction
 * to the dump file.
 *
 * @input scope The scope of the dump, can be Segment, Block or Process.
 * @input dump The current dump.
 * @input rmd The Rmd struct holding information about the dumps in this process.
 *
 * @return  Returns a RmdStatus code
 *
 * @seealso RmdStatus
 *
 */
static RmdStatus
dump_block_info(uint32_t scope, struct Dump *dump, struct Rmd *rmd)
{
   RmdStatus status = RMD_STATUS_OK;
   RmdFileStatus file_status = RMD_FILE_SUCCESS;
   uint32_t signo;
   union SIGNAL *reply = NIL;

   const SIGSELECT block_reply[] =
   { 4, RMM_ATTACH_REPLY, MONITOR_GET_BLOCK_INFO_REPLY,
     MONITOR_GET_BLOCK_INFO_ENDMARK, MONITOR_BLOCKING_FLOW_CONTROL_REQUEST };

   MonitorGetInfoRequest *req = (MonitorGetInfoRequest *) alloc(
            sizeof(MonitorGetInfoRequest), MONITOR_GET_BLOCK_INFO_REQUEST);

   req->level = 0;
   switch (scope)
   {
      case RMD_SCOPE_SEGMENT:
         req->scope.type = scope;
         req->scope.id = dump->sid;
         break;
      case RMD_SCOPE_BLOCK:
      case RMD_SCOPE_PROCESS:
         req->scope.type = MONITOR_SCOPE_ID;
         req->scope.id = dump->bid;
         break;
      default:
         /* Handle error */
         free_buf((union SIGNAL **) &req);
         status = RMD_STATUS_INVALID_PARAMETERS;
         goto finally;
   }
   req->reserved0 = 0;
   req->tagsCount = 0;

   /* Write request signal to file dump */
   file_status = rmd_file_dump_request((OSADDRESS) req, sigsize(
            (union SIGNAL **) &req), dump->dump_file);
   if (file_status != RMD_FILE_SUCCESS)
   {
      free_buf((union SIGNAL **) &req);
      status = rmd_file_status_converter(file_status, dump->dump_file);
      goto finally;
   }

   /* Request block information from monitor */
   send((union SIGNAL **) &req, rmd->rmm);

   do /* while (reply->signo != MONITOR_GET_BLOCK_INFO_ENDMARK) */
   {
      reply = receive_from(RMD_RECEIVE_TIMEOUT, block_reply, rmd->rmm);
      assert(reply);
      signo = reply->signo;
      switch (signo)
      {
         case RMM_ATTACH_REPLY:
            status = RMD_STATUS_DETACHED_FROM_RMM;
            free_buf(&reply);
            goto finally;
         case MONITOR_GET_BLOCK_INFO_REPLY:
            /* Write reply signal to file dump */
            file_status = rmd_file_dump_reply((OSADDRESS) reply,
                     sigsize(&reply), dump->dump_file);

            free_buf(&reply);

            if (file_status != RMD_FILE_SUCCESS)
            {
               status = rmd_file_status_converter(file_status, dump->dump_file);
               goto finally;
            }

            break;
         case MONITOR_BLOCKING_FLOW_CONTROL_REQUEST:
            reply->signo = MONITOR_BLOCKING_FLOW_CONTROL_REPLY;
            send(&reply, rmd->rmm);
            break;
         case MONITOR_GET_BLOCK_INFO_ENDMARK:
            break;
         default:
            status = RMD_STATUS_INTERNAL_ERROR;
            break;
      }
   } while (signo != MONITOR_GET_BLOCK_INFO_ENDMARK);

   /* Write endmark signal to filedump */
   file_status = rmd_file_dump_endmark((OSADDRESS) reply, sigsize(&reply),
            reply->monitor_get_info_endmark.status, dump->dump_file);

   free_buf(&reply);
   status = rmd_file_status_converter(file_status, dump->dump_file);

   finally: return status;
} /* dump_block_info() */

/**
 * @function dump_segment_info
 *
 * @brief Writes the signals of a MONITOR_GET_SEGMENT_INFO_REQUEST transaction
 * to the dump file.
 *
 * @input scope The scope of the dump, can be Segment, Block or Process.
 * @input dump The current dump.
 * @input rmd The Rmd struct holding information about the dumps in this process.
 *
 * @return  Returns a RmdStatus code
 *
 * @seealso RmdStatus
 *
 */
static RmdStatus
dump_segment_info(uint32_t scope, struct Dump *dump, struct Rmd *rmd)
{
   RmdStatus status = RMD_STATUS_OK;
   RmdFileStatus file_status = RMD_FILE_SUCCESS;
   uint32_t signo;
   union SIGNAL *reply = NIL;

   const SIGSELECT segment_reply[] =
   { 4, RMM_ATTACH_REPLY, MONITOR_GET_SEGMENT_INFO_REPLY,
     MONITOR_GET_SEGMENT_INFO_ENDMARK, MONITOR_BLOCKING_FLOW_CONTROL_REQUEST };

   MonitorGetInfoRequest *req = (MonitorGetInfoRequest *) alloc(
            sizeof(MonitorGetInfoRequest), MONITOR_GET_SEGMENT_INFO_REQUEST);

   req->level = 0;
   switch (scope)
   {
      case RMD_SCOPE_SEGMENT:
      case RMD_SCOPE_BLOCK:
      case RMD_SCOPE_PROCESS:
         req->scope.type = MONITOR_SCOPE_ID;
         req->scope.id = dump->sid;
         break;
      default:
         /* Handle error */
         free_buf((union SIGNAL **) &req);
         status = RMD_STATUS_INVALID_PARAMETERS;
         goto finally;
   }
   req->reserved0 = 0;
   req->tagsCount = 0;

   /* Write request signal to file dump */
   status = rmd_file_dump_request((OSADDRESS) req, sigsize(
            (union SIGNAL **) &req), dump->dump_file);
   if (file_status != RMD_FILE_SUCCESS)
   {
      free_buf((union SIGNAL **) &req);
      status = rmd_file_status_converter(file_status, dump->dump_file);
      goto finally;
   }

   /* Request segment information from monitor */
   send((union SIGNAL **) &req, rmd->rmm);

   do /* while (signo != MONITOR_GET_SEGMENT_INFO_ENDMARK) */
   {
      reply = receive_from(RMD_RECEIVE_TIMEOUT, segment_reply, rmd->rmm);
      assert(reply);
      signo = reply->signo;
      switch (signo)
      {
         case RMM_ATTACH_REPLY:
            status = RMD_STATUS_DETACHED_FROM_RMM;
            free_buf(&reply);
            goto finally;
         case MONITOR_GET_SEGMENT_INFO_REPLY:
            /* Write reply signal to file dump */
            status = rmd_file_dump_reply((OSADDRESS) reply, sigsize(&reply),
                     dump->dump_file);

            free_buf(&reply);

            if (file_status != RMD_FILE_SUCCESS)
            {
               status = rmd_file_status_converter(file_status, dump->dump_file);
               goto finally;
            }

            break;
         case MONITOR_BLOCKING_FLOW_CONTROL_REQUEST:
            reply->signo = MONITOR_BLOCKING_FLOW_CONTROL_REPLY;
            send(&reply, rmd->rmm);
            break;
         case MONITOR_GET_SEGMENT_INFO_ENDMARK:
            break;
         default:
            status = RMD_STATUS_INTERNAL_ERROR;
            break;
      }
   } while (signo != MONITOR_GET_SEGMENT_INFO_ENDMARK);

   /* Write endmark signal to filedump */
   status = rmd_file_dump_endmark((OSADDRESS) reply, sigsize(&reply),
            reply->monitor_get_info_endmark.status, dump->dump_file);
   free_buf(&reply);
   status = rmd_file_status_converter(file_status, dump->dump_file);

   finally: return status;
} /* dump_segment_info() */

/**
 * @function rmd_private_dump_process_info
 *
 * @brief Dumps the informations about the running processes
 *
 * @input rmd_req The request signal of type RmdDumpProcessInfoRequest
 * @input rmd The Rmd struct holding information about the dumps in this process.
 *
 * @seealso RmdDumpProcessInfoRequest
 *
 */
void
rmd_private_dump_process_info(union SIGNAL *rmd_req, struct Rmd *rmd)
{
   RmdStatus status = RMD_STATUS_OK;
   union SIGNAL *sig = NIL;
   struct Dump *dump = NULL;
   uint32_t scope;
   uint32_t info;

   assert(rmd_req);
   assert(RMD_DUMP_PROCESS_INFO_REQUEST == rmd_req->signo);
   assert(rmd);
   assert(RMD_MAGIC == rmd->magic);

   dump = get_dump_from_pid(sender(&rmd_req), &rmd->dumps);
   if (NULL == dump)
   {
      status = RMD_STATUS_DUMP_NOT_FOUND;
      goto finally;
   }

   scope = rmd_req->rmd_dump_process_info_request.scope;
   info = rmd_req->rmd_dump_process_info_request.info;

   if ((info & RMD_INFO_PROCESS) || (info & RMD_INFO_PROCESS_LONG))
   {
      OSBOOLEAN supported = TRUE;
      /* Check if the interface signals are supported in the RMM */
      /* The following signals are sent from this function
       * MONITOR_GET_PROCESS_INFO_REQUEST
       * MONITOR_GET_BLOCK_INFO_REQUEST
       * MONITOR_GET_SEGMENT_INFO_REQUEST
       */
      supported &= rmm_is_supported(MONITOR_GET_PROCESS_INFO_REQUEST, rmd);
      supported &= rmm_is_supported(MONITOR_GET_BLOCK_INFO_REQUEST, rmd);
      supported &= rmm_is_supported(MONITOR_GET_SEGMENT_INFO_REQUEST, rmd);
      if (!supported)
      {
         status = RMD_STATUS_INTERFACE_NOT_SUPPORTED;
         goto finally;
      }

      /* Get process info */
      status = dump_process_info(scope, info, dump, rmd);
      if (RMD_STATUS_OK != status)
      {
         goto finally;
      }

      /* Get block info */
      status = dump_block_info(scope, dump, rmd);
      if (RMD_STATUS_OK != status)
      {
         goto finally;
      }

      /* Get segment info */
      status = dump_segment_info(scope, dump, rmd);
   }

   finally:
   /* Send reply signal */
   sig = alloc(sizeof(RmdStatusReply), RMD_DUMP_PROCESS_INFO_REPLY);
   sig->rmd_reply.status = status;
   send(&sig, sender(&rmd_req));
} /* rmd_private_dump_process_info() */

/**
 * @function rmd_private_dump_stack_usage_info
 *
 * @brief Request information about stack usage from the RMM and writes the
 * response to the dump file.
 *
 * @input rmd_req The request signal of type RmdDumpStackUsageInfoRequest
 * @input rmd The Rmd struct holding information about the dumps in this process.
 *
 * @seealso RmdDumpStackUsageInfoRequest
 *
 */
void
rmd_private_dump_stack_usage_info(union SIGNAL *rmd_req, struct Rmd *rmd)
{
   RmdStatus status = RMD_STATUS_OK;
   union SIGNAL *sig = NIL;
   struct Dump *dump = NULL;
   MonitorGetInfoRequest *req = (MonitorGetInfoRequest *) NIL;
   uint32_t scope;
   uint32_t info;
   uint32_t signo;

   const SIGSELECT select_reply[] =
   { 4, RMM_ATTACH_REPLY, MONITOR_GET_STACK_USAGE_REPLY,
     MONITOR_GET_STACK_USAGE_ENDMARK, MONITOR_BLOCKING_FLOW_CONTROL_REQUEST };

   assert(rmd_req);
   assert(RMD_DUMP_STACK_USAGE_INFO_REQUEST == rmd_req->signo);
   assert(rmd);
   assert(RMD_MAGIC == rmd->magic);

   dump = get_dump_from_pid(sender(&rmd_req), &rmd->dumps);
   if (NULL == dump)
   {
      status = RMD_STATUS_DUMP_NOT_FOUND;
      goto finally;
   }

   scope = rmd_req->rmd_dump_stack_usage_info_request.scope;
   info = rmd_req->rmd_dump_stack_usage_info_request.info;
   if (info & RMD_INFO_STACK_USAGE)
   {
      RmdFileStatus file_status = RMD_FILE_SUCCESS;

      /* Check if the interface signals are supported in the RMM */
      /* The following signals are sent from this function
       * MONITOR_GET_STACK_USAGE_REQUEST
       */
      if (!rmm_is_supported(MONITOR_GET_STACK_USAGE_REQUEST, rmd))
      {
         status = RMD_STATUS_INTERFACE_NOT_SUPPORTED;
         goto finally;
      }

      req = (MonitorGetInfoRequest *) alloc(sizeof(MonitorGetInfoRequest),
               MONITOR_GET_STACK_USAGE_REQUEST);

      req->level = 0;
      switch (scope)
      {
         case RMD_SCOPE_SEGMENT:
            req->scope.type = MONITOR_SCOPE_SEGMENT_ID;
            req->scope.id = dump->sid;
            break;
         case RMD_SCOPE_BLOCK:
            req->scope.type = MONITOR_SCOPE_BLOCK_ID;
            req->scope.id = dump->bid;
            break;
         case RMD_SCOPE_PROCESS:
            req->scope.type = MONITOR_SCOPE_ID;
            req->scope.id = dump->pid;
            break;
         default:
            /* Handle error */
            free_buf((union SIGNAL **) &req);
            status = RMD_STATUS_INVALID_PARAMETERS;
            goto finally;
      }
      req->reserved0 = 0;
      req->tagsCount = 0;

      /* Write request signal to file dump */
      file_status = rmd_file_dump_request((OSADDRESS) req, sigsize(
               (union SIGNAL **) &req), dump->dump_file);
      if (file_status != RMD_FILE_SUCCESS)
      {
         free_buf((union SIGNAL **) &req);
         status = rmd_file_status_converter(file_status, dump->dump_file);
         goto finally;
      }

      /* Request information from monitor */
      send((union SIGNAL **) &req, rmd->rmm);

      do /* while (sig->signo != MONITOR_GET_STACK_USAGE_ENDMARK) */
      {
         sig = receive_from(RMD_RECEIVE_TIMEOUT, select_reply, rmd->rmm);
         assert(sig);
         signo = sig->signo;
         switch (signo)
         {
            case RMM_ATTACH_REPLY:
#ifdef RMD_VERBOSE
               ramlog_printf("RMD: RMM_ATTACH_REPLY received\n");
#endif
               status = RMD_STATUS_DETACHED_FROM_RMM;
               free_buf(&sig);
               goto finally;
            case MONITOR_GET_STACK_USAGE_REPLY:
               /* Write reply signal to file dump */
               file_status = rmd_file_dump_reply((OSADDRESS) sig,
                        sigsize(&sig), dump->dump_file);
               free_buf(&sig);

               if (file_status != RMD_FILE_SUCCESS)
               {
                  status = rmd_file_status_converter(file_status,
                           dump->dump_file);
                  goto finally;
               }

               break;
            case MONITOR_BLOCKING_FLOW_CONTROL_REQUEST:
               sig->signo = MONITOR_BLOCKING_FLOW_CONTROL_REPLY;
               send(&sig, rmd->rmm);
               break;
            case MONITOR_GET_STACK_USAGE_ENDMARK:
               break;
         }
      } while (signo != MONITOR_GET_STACK_USAGE_ENDMARK);

      /* Write endmark signal to filedump */
      file_status = rmd_file_dump_endmark((OSADDRESS) sig, sigsize(&sig),
               sig->monitor_get_info_endmark.status, dump->dump_file);
      status = rmd_file_status_converter(file_status, dump->dump_file);

      free_buf(&sig);
   }

   finally:
   /* Send reply signal */
   sig = alloc(sizeof(RmdStatusReply), RMD_DUMP_STACK_USAGE_INFO_REPLY);
   sig->rmd_reply.status = status;
   send(&sig, sender(&rmd_req));
} /* rmd_private_dump_stack_usage_info() */

/**
 * @brief Function pointer type that specifies the interface for requests that shall
 * be called once for each process in a scope.
 *
 * @returns A RmdStatus code.
 *
 */
typedef RmdStatus
(*HandleProcessFunc)(struct Rmd *rmd, struct Dump *dump, PROCESS pid);

/**
 * @function for_each_processes_in_scope
 *
 * @brief Call handle_process_func once for each process in our scope.
 *
 * @returns A RmdStatus code.
 *
 */
static RmdStatus
for_each_processes_in_scope(RmdScope scope, struct Rmd *rmd, struct Dump *dump,
         HandleProcessFunc handle_process_func)
{
   RmdStatus status = RMD_STATUS_OK;
   union SIGNAL *sig = NIL;
   uint32_t signo;
   MonitorGetInfoRequest *req;
   const SIGSELECT process_reply[] =
   { 4, RMM_ATTACH_REPLY, MONITOR_GET_PROCESS_INFO_REPLY,
     MONITOR_GET_PROCESS_INFO_ENDMARK, MONITOR_BLOCKING_FLOW_CONTROL_REQUEST };

   /* Check if the interface signals are supported in the RMM */
   /* The following signals are sent from this function
    * MONITOR_GET_PROCESS_INFO_REQUEST
    */
   if (!rmm_is_supported(MONITOR_GET_PROCESS_INFO_REQUEST, rmd))
   {
      status = RMD_STATUS_INTERFACE_NOT_SUPPORTED;
      goto finally;
   }

   req = (MonitorGetInfoRequest *) alloc(sizeof(MonitorGetInfoRequest),
            MONITOR_GET_PROCESS_INFO_REQUEST);

   req->level = 0;
   switch (scope)
   {
      case RMD_SCOPE_SEGMENT:
         req->scope.type = scope;
         req->scope.id = dump->sid;
         break;
      case RMD_SCOPE_BLOCK:
         req->scope.type = scope;
         req->scope.id = dump->bid;
         break;
      case RMD_SCOPE_PROCESS:
         req->scope.type = scope;
         req->scope.id = dump->pid;
         break;
      default:
         /* Handle error */
         free_buf((union SIGNAL **) &req);
         status = RMD_STATUS_INVALID_PARAMETERS;
         goto finally;
   }
   req->reserved0 = 0;
   req->tagsCount = 0;

   /* Request process information from monitor */
   send((union SIGNAL **) &req, rmd->rmm);

   do /* while (sig->signo != MONITOR_GET_PROCESS_INFO_ENDMARK) */
   {
      sig = receive_from(RMD_RECEIVE_TIMEOUT, process_reply, rmd->rmm);
      assert(sig);
      signo = sig->signo;
      switch (signo)
      {
         case RMM_ATTACH_REPLY:
            status = RMD_STATUS_DETACHED_FROM_RMM;
            free_buf(&sig);
            goto finally;
         case MONITOR_GET_PROCESS_INFO_REPLY:
            /* Do what we came here for.. */
            status = handle_process_func(rmd, dump,
                     sig->monitor_get_process_info_reply.pid);

            free_buf(&sig);
            if (status != RMD_STATUS_OK)
            {
               goto finally;
            }
            break;
         case MONITOR_BLOCKING_FLOW_CONTROL_REQUEST:
            sig->signo = MONITOR_BLOCKING_FLOW_CONTROL_REPLY;
            send(&sig, rmd->rmm);
            break;
         case MONITOR_GET_PROCESS_INFO_ENDMARK:
            break;
      }
   } while (signo != MONITOR_GET_PROCESS_INFO_ENDMARK);
   free_buf(&sig);

   finally: return status;
} /* for_each_processes_in_scope() */

/**
 * @function for_each_block_in_scope
 *
 * @brief Call handle_block_func once for each block in our scope.
 *
 * @returns A RmdStatus code.
 *
 */
static RmdStatus
for_each_block_in_scope(RmdScope scope, struct Rmd *rmd, struct Dump *dump,
         HandleProcessFunc handle_block_func)
{
   RmdStatus status = RMD_STATUS_OK;
   union SIGNAL *sig = NIL;
   uint32_t signo;
   MonitorGetInfoRequest *req;
   const SIGSELECT block_reply[] =
   { 4, RMM_ATTACH_REPLY, MONITOR_GET_BLOCK_INFO_REPLY,
     MONITOR_GET_BLOCK_INFO_ENDMARK, MONITOR_BLOCKING_FLOW_CONTROL_REQUEST };

   /* Check if the interface signals are supported in the RMM */
   /* The following signals are sent from this function
    * MONITOR_GET_PROCESS_INFO_REQUEST
    */
   if (!rmm_is_supported(MONITOR_GET_BLOCK_INFO_REQUEST, rmd))
   {
      status = RMD_STATUS_INTERFACE_NOT_SUPPORTED;
      goto finally;
   }

   switch (scope)
   {
      case RMD_SCOPE_SEGMENT:
         /* Iterate over all blocks in segment */
         req = (MonitorGetInfoRequest *) alloc(sizeof(MonitorGetInfoRequest),
                  MONITOR_GET_BLOCK_INFO_REQUEST);
         req->level = 0;
         req->scope.type = scope;
         req->scope.id = dump->sid;
         req->reserved0 = 0;
         req->tagsCount = 0;
         break;
      case RMD_SCOPE_BLOCK:
         /* Handle this block and exit */
         status = handle_block_func(rmd, dump, dump->bid);
         goto finally;
      case RMD_SCOPE_PROCESS:
         /* There are no blocks in the process... */
         goto finally;
      default:
         /* Handle error */
         status = RMD_STATUS_INVALID_PARAMETERS;
         goto finally;
   }

   /* Request block information from monitor */
   send((union SIGNAL **) &req, rmd->rmm);

   do /* while (sig->signo != MONITOR_GET_BLOCK_INFO_ENDMARK) */
   {
      sig = receive_from(RMD_RECEIVE_TIMEOUT, block_reply, rmd->rmm);
      assert(sig);
      signo = sig->signo;
      switch (signo)
      {
         case RMM_ATTACH_REPLY:
            status = RMD_STATUS_DETACHED_FROM_RMM;
            free_buf(&sig);
            goto finally;
         case MONITOR_GET_BLOCK_INFO_REPLY:
            /* Do what we came here for.. */
            status = handle_block_func(rmd, dump,
                     sig->monitor_get_process_info_reply.pid);

            free_buf(&sig);
            if (status != RMD_STATUS_OK)
            {
               goto finally;
            }
            break;
         case MONITOR_BLOCKING_FLOW_CONTROL_REQUEST:
            sig->signo = MONITOR_BLOCKING_FLOW_CONTROL_REPLY;
            send(&sig, rmd->rmm);
            break;
         case MONITOR_GET_BLOCK_INFO_ENDMARK:
            break;
      }
   } while (signo != MONITOR_GET_BLOCK_INFO_ENDMARK);
   free_buf(&sig);

   finally: return status;
} /* for_each_block_in_scope() */

/**
 * @function dump_process_env_vars_info
 *
 * @brief This function is of type HandleProcessFunc.
 *        Requests environment variables information related to a process id
 *        and writes down the reply signals to file dump.
 *
 * @returns A RmdStatus code.
 *
 */
static RmdStatus
dump_process_env_vars_info(struct Rmd *rmd, struct Dump *dump, PROCESS pid)
{
   RmdStatus status = RMD_STATUS_OK;
   RmdFileStatus file_status = RMD_FILE_SUCCESS;
   union SIGNAL *sig = NIL;
   uint32_t signo;
   MonitorGetInfoRequest *req;
   const SIGSELECT select_reply[] =
   { 4, RMM_ATTACH_REPLY, MONITOR_GET_ENV_VARS_REPLY,
     MONITOR_GET_ENV_VARS_ENDMARK, MONITOR_BLOCKING_FLOW_CONTROL_REQUEST };

   /* Check if the interface signals are supported in the RMM */
   /* The following signals are sent from this function
    * MONITOR_GET_SIGNAL_QUEUE_REQUEST
    */
   if (!rmm_is_supported(MONITOR_GET_ENV_VARS_REQUEST, rmd))
   {
      status = RMD_STATUS_INTERFACE_NOT_SUPPORTED;
      goto finally;
   }

   req = (MonitorGetInfoRequest *) alloc(sizeof(MonitorGetInfoRequest),
            MONITOR_GET_ENV_VARS_REQUEST);

   req->level = 0;
   req->scope.type = MONITOR_SCOPE_ID;
   req->scope.id = pid;
   req->reserved0 = 0;
   req->tagsCount = 0;

   /* Request information from monitor */
   send((union SIGNAL **) &req, rmd->rmm);

   do /* while (sig->signo != MONITOR_GET_ENV_VARS_ENDMARK) */
   {
      sig = receive_from(RMD_RECEIVE_TIMEOUT, select_reply, rmd->rmm);
      assert(sig);
      signo = sig->signo;
      switch (signo)
      {
         case RMM_ATTACH_REPLY:
#ifdef RMD_VERBOSE
            ramlog_printf("RMD: RMM_ATTACH_REPLY received\n");
#endif
            status = RMD_STATUS_DETACHED_FROM_RMM;
            free_buf(&sig);
            goto finally;
         case MONITOR_GET_ENV_VARS_REPLY:
            /* Write reply signal to file dump */
            file_status = rmd_file_dump_reply((OSADDRESS) sig, sigsize(&sig),
                     dump->dump_file);

            free_buf(&sig);

            if (file_status != RMD_STATUS_OK)
            {
               status = rmd_file_status_converter(file_status, dump->dump_file);
               goto finally;
            }

            break;
         case MONITOR_BLOCKING_FLOW_CONTROL_REQUEST:
            sig->signo = MONITOR_BLOCKING_FLOW_CONTROL_REPLY;
            send(&sig, rmd->rmm);
            break;
         case MONITOR_GET_ENV_VARS_ENDMARK:
            break;
      }
   } while (signo != MONITOR_GET_ENV_VARS_ENDMARK);
   free_buf(&sig);

   finally: return status;
} /* dump_process_env_vars_info() */

/**
 * @function rmd_private_dump_env_vars_info
 *
 * @brief Requests information about the environment variables for the scope
 * of interest from the RMM and writes the response to the dump file.
 *
 * @input rmd_req The request signal of type RmdDumpEnvVarsInfoRequest
 * @input rmd The Rmd struct holding information about the dumps in this process.
 *
 * @seealso RmdDumpEnvVarsInfoRequest
 *
 */
void
rmd_private_dump_env_vars_info(union SIGNAL *rmd_req, struct Rmd *rmd)
{
   RmdStatus status = RMD_STATUS_OK;
   union SIGNAL *sig = NIL;
   struct Dump *dump = NULL;
   assert(rmd_req);
   assert(RMD_DUMP_SIGNAL_QUEUE_INFO_REQUEST == rmd_req->signo);
   assert(rmd);
   assert(RMD_MAGIC == rmd->magic);

   dump = get_dump_from_pid(sender(&rmd_req), &rmd->dumps);
   if (NULL == dump)
   {
      status = RMD_STATUS_DUMP_NOT_FOUND;
      goto finally;
   }

   if (rmd_req->rmd_dump_signal_queue_info_request.info & RMD_INFO_ENV_VARS)
   {
      RmdFileStatus file_status = RMD_FILE_SUCCESS;
      MonitorGetInfoRequest *req = (MonitorGetInfoRequest *) alloc(
               sizeof(MonitorGetInfoRequest), MONITOR_GET_ENV_VARS_REQUEST);
      MonitorGetInfoEndmark *reply;

      req->level = 0;
      switch (rmd_req->rmd_dump_stack_usage_info_request.scope)
      {
         case RMD_SCOPE_SEGMENT:
            req->scope.type = rmd_req->rmd_dump_env_vars_info_request.scope;
            req->scope.id = dump->sid;
            break;
         case RMD_SCOPE_BLOCK:
            req->scope.type = rmd_req->rmd_dump_env_vars_info_request.scope;
            req->scope.id = dump->bid;
            break;
         case RMD_SCOPE_PROCESS:
            req->scope.type = rmd_req->rmd_dump_env_vars_info_request.scope;
            req->scope.id = dump->pid;
            break;
         default:
            /* Handle error */
            free_buf((union SIGNAL **) &req);
            status = RMD_STATUS_INVALID_PARAMETERS;
            goto finally;
      }
      req->reserved0 = 0;
      req->tagsCount = 0;

      /* Write request signal to file dump */
      file_status = rmd_file_dump_request((OSADDRESS) req, sigsize(
               (union SIGNAL **) &req), dump->dump_file);
      free_buf((union SIGNAL **) &req);
      if (file_status != RMD_FILE_SUCCESS)
      {
         status = rmd_file_status_converter(file_status, dump->dump_file);
         goto finally;
      }

      /* Request environment variable information for each process in our
       * scope */
      status = for_each_processes_in_scope(
               rmd_req->rmd_dump_env_vars_info_request.scope, rmd, dump,
               dump_process_env_vars_info);
      if (status != RMD_STATUS_OK)
      {
         goto finally;
      }

      /* Request environment variable information for each block in our
       * scope */
      status = for_each_block_in_scope(
               rmd_req->rmd_dump_env_vars_info_request.scope, rmd, dump,
               dump_process_env_vars_info);
      if (status != RMD_STATUS_OK)
      {
         goto finally;
      }

      /* Create endmark signal and write it */
      reply = (MonitorGetInfoEndmark *) alloc(sizeof(MonitorGetInfoEndmark),
               MONITOR_GET_ENV_VARS_ENDMARK);
      reply->status = MONITOR_STATUS_OK;
      file_status = rmd_file_dump_endmark((OSADDRESS) reply, sigsize(
               (union SIGNAL **) &reply), reply->status, dump->dump_file);
      free_buf((union SIGNAL **) &reply);
      status = rmd_file_status_converter(file_status, dump->dump_file);
   }

   finally:
   /* Send reply signal */
   sig = alloc(sizeof(RmdStatusReply), RMD_DUMP_ENV_VARS_INFO_REPLY);
   sig->rmd_reply.status = status;
   send(&sig, sender(&rmd_req));
} /* rmd_private_dump_env_vars_info() */

/**
 * @function dump_process_signal_queue_info
 *
 * @brief This function is of type HandleProcessFunc.
 *        Requests signal queue information related to a processs id and
 *        writes down the reply signals to file dump.
 *
 * @returns A RmdStatus code.
 *
 */
static RmdStatus
dump_process_signal_queue_info(struct Rmd *rmd, struct Dump *dump, PROCESS pid)
{
   RmdStatus status = RMD_STATUS_OK;
   RmdFileStatus file_status = RMD_FILE_SUCCESS;
   union SIGNAL *sig = NIL;
   uint32_t signo;
   MonitorGetInfoRequest *req;
   const SIGSELECT select_reply[] =
   { 4, RMM_ATTACH_REPLY, MONITOR_GET_SIGNAL_QUEUE_REPLY,
     MONITOR_GET_SIGNAL_QUEUE_ENDMARK, MONITOR_BLOCKING_FLOW_CONTROL_REQUEST };

   /* store signal cap property */
   uint32_t max_signals = rmd->rmd_properties.max_signals_in_queue;

   /* Check if the interface signals are supported in the RMM */
   /* The following signals are sent from this function
    * MONITOR_GET_SIGNAL_QUEUE_REQUEST
    */
   if (!rmm_is_supported(MONITOR_GET_SIGNAL_QUEUE_REQUEST, rmd))
   {
      status = RMD_STATUS_INTERFACE_NOT_SUPPORTED;
      goto finally;
   }

   req = (MonitorGetInfoRequest *) alloc(sizeof(MonitorGetInfoRequest),
            MONITOR_GET_SIGNAL_QUEUE_REQUEST);

   req->level = 0;
   req->scope.type = MONITOR_SCOPE_ID;
   req->scope.id = pid;
   req->reserved0 = 0;
   req->tagsCount = 0;

   /* Request information from monitor */
   send((union SIGNAL **) &req, rmd->rmm);

   do /* while (sig->signo != MONITOR_GET_SIGNAL_QUEUE_ENDMARK) */
   {
      sig = receive_from(RMD_RECEIVE_TIMEOUT, select_reply, rmd->rmm);

      assert(sig);
      signo = sig->signo;

      switch (signo)
      {
         case RMM_ATTACH_REPLY:
#ifdef RMD_VERBOSE
            ramlog_printf("RMD: RMM_ATTACH_REPLY received\n");
#endif
            status = RMD_STATUS_DETACHED_FROM_RMM;
            free_buf(&sig);
            goto finally;
         case MONITOR_GET_SIGNAL_QUEUE_REPLY:
            if (rmd->rmd_properties.max_signals_in_queue)
            {
               if(max_signals)
               {
	          /* Write reply signal to file dump */
		  file_status = rmd_file_dump_reply((OSADDRESS) sig, sigsize(&sig),
						 dump->dump_file);
		  max_signals--;
               }
            }
            else
            {
               /* Write reply signal to file dump */
	       file_status = rmd_file_dump_reply((OSADDRESS) sig, sigsize(&sig),
					 dump->dump_file);
            }
            free_buf(&sig);

            if (file_status != RMD_STATUS_OK)
            {
               status = rmd_file_status_converter(file_status, dump->dump_file);
               goto finally;
            }

            break;
         case MONITOR_BLOCKING_FLOW_CONTROL_REQUEST:
            sig->signo = MONITOR_BLOCKING_FLOW_CONTROL_REPLY;
            send(&sig, rmd->rmm);
            break;
         case MONITOR_GET_SIGNAL_QUEUE_ENDMARK:
            break;
      }
   } while (signo != MONITOR_GET_SIGNAL_QUEUE_ENDMARK);
   free_buf(&sig);
   finally: return status;
} /* dump_process_signal_queue_info() */

/**
 * @function rmd_private_dump_signal_queue_info
 *
 * @brief Requests information about the signal queues from the RMM and writes
 * the results to the dump file.
 *
 * @input rmd_req The request signal of type RmdDumpSignalQueueInfoRequest
 * @input rmd The Rmd struct holding information about the dumps in this process.
 *
 * @seealso RmdDumpSignalQueueInfoRequest
 *
 */
void
rmd_private_dump_signal_queue_info(union SIGNAL *rmd_req, struct Rmd *rmd)
{
   RmdStatus status = RMD_STATUS_OK;
   union SIGNAL *sig = NIL;
   struct Dump *dump = NULL;
   assert(rmd_req);
   assert(RMD_DUMP_SIGNAL_QUEUE_INFO_REQUEST == rmd_req->signo);
   assert(rmd);
   assert(RMD_MAGIC == rmd->magic);

   dump = get_dump_from_pid(sender(&rmd_req), &rmd->dumps);
   if (NULL == dump)
   {
      status = RMD_STATUS_DUMP_NOT_FOUND;
      goto finally;
   }

   if (rmd_req->rmd_dump_signal_queue_info_request.info & RMD_INFO_SIGNAL_QUEUE)
   {
      RmdFileStatus file_status = RMD_FILE_SUCCESS;

      MonitorGetInfoRequest *req = (MonitorGetInfoRequest *) alloc(
               sizeof(MonitorGetInfoRequest), MONITOR_GET_SIGNAL_QUEUE_REQUEST);
      MonitorGetInfoEndmark *reply;

      if (!rmm_is_supported(MONITOR_GET_SIGNAL_QUEUE_REQUEST, rmd))
      {
         status = RMD_STATUS_INTERFACE_NOT_SUPPORTED;
         goto finally;
      }

      req->level = 0;
      switch (rmd_req->rmd_dump_signal_queue_info_request.scope)
      {
         case RMD_SCOPE_SEGMENT:
            req->scope.type = rmd_req->rmd_dump_signal_queue_info_request.scope;
            req->scope.id = dump->sid;
            break;
         case RMD_SCOPE_BLOCK:
            req->scope.type = rmd_req->rmd_dump_signal_queue_info_request.scope;
            req->scope.id = dump->bid;
            break;
         case RMD_SCOPE_PROCESS:
            req->scope.type = rmd_req->rmd_dump_signal_queue_info_request.scope;
            req->scope.id = dump->pid;
            break;
         default:
            /* Handle error */
            free_buf((union SIGNAL **) &req);
            status = RMD_STATUS_INVALID_PARAMETERS;
            goto finally;
      }
      req->reserved0 = 0;
      req->tagsCount = 0;


      /* Write request signal to file dump */
      file_status = rmd_file_dump_request((OSADDRESS) req, sigsize(
               (union SIGNAL **) &req), dump->dump_file);
      free_buf( (union SIGNAL** )&req);

      if (file_status != RMD_FILE_SUCCESS)
      {
         status = rmd_file_status_converter(file_status, dump->dump_file);
         goto finally;
      }
      /* Check for number of long queue processes -rmd property */
      if(rmd->rmd_properties.num_long_queue_procs)
      {
    	 struct ProcQueueCnt *proc_queue_list;
    	 uint32_t i;
    	 uint32_t procs_count = 0;
    	 MonitorGetInfoRequest *req1;

	 if (!rmm_is_supported(MONITOR_GET_PROCESS_INFO_REQUEST, rmd))
    	 {
    	    status = RMD_STATUS_INTERFACE_NOT_SUPPORTED;
    	    goto finally;
    	 }
    	 req1 = (MonitorGetInfoRequest *) alloc(
    	       sizeof(MonitorGetInfoRequest), MONITOR_GET_PROCESS_INFO_REQUEST);
    	 /* get process info of all the processes in the scope */
    	 req1->level = 1;
	 /* set number of tags to zero */
 	 req1->tagsCount = 0;

    	 if ((req1->scope.type = rmd_req->rmd_dump_signal_queue_info_request.scope) ==
               RMD_SCOPE_SEGMENT)
         {
    	    req1->scope.id = dump->sid;
         }
    	 else if (req1->scope.type == RMD_SCOPE_BLOCK)
         { 
    	    req1->scope.id = dump->bid;
         }
    	 else
         {
    	    req1->scope.id = dump->pid;
         }
         
    	 /* create pool buffer to hold pid and signal queue count*/
    	 proc_queue_list = (struct ProcQueueCnt *) alloc(
                          sizeof(struct ProcQueueCnt) *
                          (rmd->rmd_properties.num_long_queue_procs +1), 0);
    	 status = get_process_long_info((union SIGNAL*)req1, rmd, proc_queue_list,
                                         &procs_count);

    	 if (status != RMD_STATUS_OK)
	 {
    	    /* free the proc_queue_list buffer */
    	    free_buf((union SIGNAL **)&proc_queue_list);
    	    goto finally;
	 }


    	 /* dump signal queue info for all the processes in the sorted list*/
    	 for (i=0; i<((rmd->rmd_properties.num_long_queue_procs > procs_count)?
               procs_count:rmd->rmd_properties.num_long_queue_procs); i++)
    	 {
            status = dump_process_signal_queue_info(rmd,dump,((proc_queue_list+i)->pid));

    	    if (status != RMD_STATUS_OK)
	    {
    	       /* free the proc_queue_list buffer */
    	       free_buf((union SIGNAL **)&proc_queue_list);
    	       goto finally;
	    }
    	 }
    	 /* free the proc_queue_list buffer */
    	 free_buf((union SIGNAL **)&proc_queue_list);

      }
      else
      {
         /* Request signal queue information for each process in our scope */
         status = for_each_processes_in_scope(
	     	    rmd_req->rmd_dump_signal_queue_info_request.scope, rmd, dump,
		    dump_process_signal_queue_info);
      }

      if (status != RMD_STATUS_OK)
      {
         goto finally;
      }

      /* Create endmark signal and write it */
      reply = (MonitorGetInfoEndmark *) alloc(sizeof(MonitorGetInfoEndmark),
               MONITOR_GET_SIGNAL_QUEUE_ENDMARK);
      reply->status = MONITOR_STATUS_OK;
      file_status = rmd_file_dump_endmark((OSADDRESS) reply, sigsize(
               (union SIGNAL **) &reply), reply->status, dump->dump_file);
      free_buf((union SIGNAL **) &reply);
      status = rmd_file_status_converter(file_status, dump->dump_file);
   }

   finally:
   /* Send reply signal */
   sig = alloc(sizeof(RmdStatusReply), RMD_DUMP_SIGNAL_QUEUE_INFO_REPLY);
   sig->rmd_reply.status = status;
   send(&sig, sender(&rmd_req));
} /* rmd_private_dump_signal_queue_info() */

/**
 * @function rmd_private_dump_pool_info
 *
 * @brief Requests information about the pools from the RMM and writes the
 * results to the dump file.
 *
 * @input rmd_req The request signal of type RmdDumpPoolInfoRequest
 * @input rmd The Rmd struct holding information about the dumps in this process.
 *
 * @seealso RmdDumpPoolInfoRequest
 *
 */
void
rmd_private_dump_pool_info(union SIGNAL *rmd_req, struct Rmd *rmd)
{
   RmdStatus status = RMD_STATUS_OK;
   union SIGNAL *sig = NIL;
   struct Dump *dump = NULL;
   uint32_t scope;
   uint32_t info;
   uint32_t signo;
   const SIGSELECT select_reply[] =
   { 4, RMM_ATTACH_REPLY, MONITOR_GET_POOL_INFO_REPLY,
     MONITOR_GET_POOL_INFO_ENDMARK, MONITOR_BLOCKING_FLOW_CONTROL_REQUEST };

   assert(rmd_req);
   assert(RMD_DUMP_POOL_INFO_REQUEST == rmd_req->signo);
   assert(rmd);
   assert(RMD_MAGIC == rmd->magic);

   dump = get_dump_from_pid(sender(&rmd_req), &rmd->dumps);
   if (NULL == dump)
   {
      status = RMD_STATUS_DUMP_NOT_FOUND;
      goto finally;
   }

   scope = rmd_req->rmd_dump_pool_info_request.scope;
   info = rmd_req->rmd_dump_pool_info_request.info;
   if (info & RMD_INFO_POOL_INFO)
   {
      RmdFileStatus file_status = RMD_FILE_SUCCESS;
      MonitorGetInfoRequest *req;

      /* Check if the interface signals are supported in the RMM */
      /* The following signals are sent from this function
       * MONITOR_GET_POOL_INFO_REQUEST
       */
      if (!rmm_is_supported(MONITOR_GET_POOL_INFO_REQUEST, rmd))
      {
         status = RMD_STATUS_INTERFACE_NOT_SUPPORTED;
         goto finally;
      }

      req = (MonitorGetInfoRequest *) alloc(sizeof(MonitorGetInfoRequest),
               MONITOR_GET_POOL_INFO_REQUEST);

      req->level = 0;
      switch (scope)
      {
         case RMD_SCOPE_SEGMENT:
         case RMD_SCOPE_BLOCK:
         case RMD_SCOPE_PROCESS:
            req->scope.type = MONITOR_SCOPE_SEGMENT_ID;
            req->scope.id = dump->sid;
            break;
         default:
            /* Handle error */
            free_buf((union SIGNAL **) &req);
            status = RMD_STATUS_INVALID_PARAMETERS;
            goto finally;
      }
      req->reserved0 = 0;
      req->tagsCount = 0;

      /* Write request signal to file dump */
      file_status = rmd_file_dump_request((OSADDRESS) req, sigsize(
               (union SIGNAL **) &req), dump->dump_file);
      if (file_status != RMD_FILE_SUCCESS)
      {
         free_buf((union SIGNAL **) &req);
         status = rmd_file_status_converter(file_status, dump->dump_file);
         goto finally;
      }

      /* Request information from monitor */
      send((union SIGNAL **) &req, rmd->rmm);

      do /* while (sig->signo != MONITOR_GET_POOL_INFO_ENDMARK) */
      {
         sig = receive_from(RMD_RECEIVE_TIMEOUT, select_reply, rmd->rmm);
         assert(sig);
         signo = sig->signo;
         switch (signo)
         {
            case RMM_ATTACH_REPLY:
#ifdef RMD_VERBOSE
               ramlog_printf("RMD: RMM_ATTACH_REPLY received\n");
#endif
               status = RMD_STATUS_DETACHED_FROM_RMM;
               free_buf(&sig);
               goto finally;
            case MONITOR_GET_POOL_INFO_REPLY:
               /* Write reply signal to file dump */
               status = rmd_file_dump_reply((OSADDRESS) sig, sigsize(&sig),
                        dump->dump_file);

               free_buf(&sig);

               if (file_status != RMD_FILE_SUCCESS)
               {
                  status = rmd_file_status_converter(file_status,
                           dump->dump_file);
                  goto finally;
               }

               break;
            case MONITOR_BLOCKING_FLOW_CONTROL_REQUEST:
               sig->signo = MONITOR_BLOCKING_FLOW_CONTROL_REPLY;
               send(&sig, rmd->rmm);
               break;
            case MONITOR_GET_POOL_INFO_ENDMARK:
               break;
         }
      } while (signo != MONITOR_GET_POOL_INFO_ENDMARK);

      /* Write endmark signal to filedump */
      file_status = rmd_file_dump_endmark((OSADDRESS) sig, sigsize(&sig),
               sig->monitor_get_info_endmark.status, dump->dump_file);
      status = rmd_file_status_converter(file_status, dump->dump_file);

      free_buf(&sig);
   }

   finally:
   /* Send reply signal */
   sig = alloc(sizeof(RmdStatusReply), RMD_DUMP_POOL_INFO_REPLY);
   sig->rmd_reply.status = status;
   send(&sig, sender(&rmd_req));
} /* rmd_private_dump_pool_info() */

/**
 * @brief Function pointer type that specifies the interface for requests that shall
 * be called once for each signal in a pool.
 *
 * @returns A RmdStatus code.
 *
 */
typedef RmdStatus
(* SignalHandlerFunction)(union SIGNAL *sig, RmdScope scope, uint32_t memory,
         uint32_t info, struct Dump *dump);

/**
 * @function for_each_signal_in_pool
 *
 * @brief Call signal_handler_function once for each signal in a pool.
 *
 * @input pool_id Pool to traverse.
 * @input scope Send to the handler function.
 * @input memory Send to the handler function.
 * @input info Send to the handler function.
 * @input rmd Send to the handler function.
 * @input dump Send to the handler function.
 * @input signal_handler_function Pointer to function to call for each signal.
 *
 * @output monitor_status The status in the endmark signal.
 *
 * @returns A RmdStatus code.
 *
 */
static RmdStatus
for_each_signal_in_pool(uint32_t pool_id, RmdScope scope, uint32_t memory,
         uint32_t info, MonitorStatus *monitor_status, struct Rmd *rmd,
         struct Dump *dump, SignalHandlerFunction signal_handler_function)
{
   RmdStatus status = RMD_STATUS_OK;
   union SIGNAL *reply = NIL;
   uint32_t signo;
   MonitorGetInfoRequest *req;
   const SIGSELECT select_reply[] =
   { 4, RMM_ATTACH_REPLY, MONITOR_GET_POOL_SIGNALS_REPLY,
     MONITOR_GET_POOL_SIGNALS_ENDMARK, MONITOR_BLOCKING_FLOW_CONTROL_REQUEST };

   req = (MonitorGetInfoRequest *) alloc(sizeof(MonitorGetInfoRequest),
            MONITOR_GET_POOL_SIGNALS_REQUEST);

   switch (scope)
   {
      case RMD_SCOPE_SEGMENT:
      case RMD_SCOPE_BLOCK:
      case RMD_SCOPE_PROCESS:
         req->scope.type = MONITOR_SCOPE_ID;
         req->scope.id = pool_id;
         break;
      default:
         /* Handle error */
         free_buf((union SIGNAL **) &req);
         status = RMD_STATUS_INVALID_PARAMETERS;
         goto finally;
   }
   req->reserved0 = 0;
   req->tagsCount = 0;

   /* Request information from monitor */
   send((union SIGNAL **) &req, rmd->rmm);

   do /* while (signo != MONITOR_GET_POOL_SIGNALS_ENDMARK) */
   {
      reply = receive_from(RMD_RECEIVE_TIMEOUT, select_reply, rmd->rmm);
      assert(reply);
      signo = reply->signo;
      switch (signo)
      {
         case RMM_ATTACH_REPLY:
#ifdef RMD_VERBOSE
            ramlog_printf("RMD: RMM_ATTACH_REPLY received\n");
#endif
            status = RMD_STATUS_DETACHED_FROM_RMM;
            free_buf(&reply);
            goto finally;
         case MONITOR_GET_POOL_SIGNALS_REPLY:
            status = signal_handler_function(reply, scope, memory, info, dump);
            free_buf(&reply);

            if (RMD_STATUS_OK != status)
            {
               goto finally;
            }

            break;
         case MONITOR_BLOCKING_FLOW_CONTROL_REQUEST:
            reply->signo = MONITOR_BLOCKING_FLOW_CONTROL_REPLY;
            send(&reply, rmd->rmm);
            break;
         case MONITOR_GET_POOL_SIGNALS_ENDMARK:
            break;
      }
   } while (signo != MONITOR_GET_POOL_SIGNALS_ENDMARK);

   *monitor_status = reply->monitor_get_info_endmark.status;

   free_buf(&reply);
   finally: return status;
} /* for_each_signal_in_pool() */

/**
 * @brief Function pointer type that specifies the interface for requests that shall
 * be called once for each pooll in a segment.
 *
 * @returns A RmdStatus code.
 *
 */
typedef RmdStatus
(* PoolHandlerFunction)(
         struct MonitorGetPoolInfoReply *monitor_get_pool_info_reply,
         RmdScope scope, uint32_t memory, uint32_t info,
         MonitorStatus *monitor_status, struct Rmd *rmd, struct Dump *dump);

/**
 * @function for_each_pool_in_segment
 *
 * @brief Calls pool_handler_function once for each pool in a segment.
 *
 * @input scope Send to the handler function.
 * @input memory Send to the handler function.
 * @input info Send to the handler function.
 * @input rmd Send to the handler function.
 * @input dump->sid Segment id to traverse
 * @input pool_handler_function Pointer to function to call for each pool.
 *
 * @output monitor_status Output from handler function.
 *
 * @returns A RmdStatus code.
 *
 */
static RmdStatus
for_each_pool_in_segment(RmdScope scope, uint32_t memory, uint32_t info,
         MonitorStatus *monitor_status, struct Rmd *rmd, struct Dump *dump,
         PoolHandlerFunction pool_handler_function)
{
   RmdStatus status = RMD_STATUS_OK;
   /*   RmdFileStatus rmd_file_status = RMD_FILE_SUCCESS;*/
   union SIGNAL *reply = NIL;
   uint32_t signo;

   /* Get information of all pools in segment */
   SIGSELECT select_reply[] =
   { 4, RMM_ATTACH_REPLY, MONITOR_GET_POOL_INFO_REPLY,
     MONITOR_BLOCKING_FLOW_CONTROL_REQUEST, MONITOR_GET_POOL_INFO_ENDMARK };
   MonitorGetInfoRequest *pool_info_req;

   if (!rmm_is_supported(MONITOR_GET_POOL_INFO_REQUEST, rmd))
   {
      status = RMD_STATUS_INTERFACE_NOT_SUPPORTED;
      goto finally;
   }

   pool_info_req = (MonitorGetInfoRequest *) alloc(
            sizeof(MonitorGetInfoRequest), MONITOR_GET_POOL_INFO_REQUEST);

   pool_info_req->level = 0;
   pool_info_req->scope.type = MONITOR_SCOPE_SEGMENT_ID;
   pool_info_req->scope.id = dump->sid;
   pool_info_req->reserved0 = 0;
   pool_info_req->tagsCount = 0;

   /* Request information from monitor */
   send((union SIGNAL **) &pool_info_req, rmd->rmm);

   do /* while (reply->signo != MONITOR_GET_POOL_INFO_ENDMARK) */
   {
      reply = receive_from(RMD_RECEIVE_TIMEOUT, select_reply, rmd->rmm);
      assert(reply);
      signo = reply->signo;
      switch (signo)
      {
         case RMM_ATTACH_REPLY:
#ifdef RMD_VERBOSE
            ramlog_printf("RMD: RMM_ATTACH_REPLY received\n");
#endif
            status = RMD_STATUS_DETACHED_FROM_RMM;
            free_buf(&reply);
            goto finally;

         case MONITOR_GET_POOL_INFO_REPLY:
            /* We have a pool, call handler function. */
            status = pool_handler_function(&reply->monitor_get_pool_info_reply,
                     scope, memory, info, monitor_status, rmd, dump);
            free_buf(&reply);
            if (status != RMD_STATUS_OK)
               goto finally;

            break;

         case MONITOR_BLOCKING_FLOW_CONTROL_REQUEST:
            reply->signo = MONITOR_BLOCKING_FLOW_CONTROL_REPLY;
            send(&reply, rmd->rmm);
            break;

         case MONITOR_GET_POOL_INFO_ENDMARK:
            break;
      }
   } while (signo != MONITOR_GET_POOL_INFO_ENDMARK);
   free_buf(&reply);
   finally: return status;
} /* for_each_pool_in_segment() */

/**
 * @function dump_pool_signals_handler
 *
 * @brief Handler function to be called from for_each_signal_in_pool(),
 * writes the signal to file.
 *
 * @seealso for_each_signal_in_pool
 *
 */
static RmdStatus
dump_pool_signals_handler(union SIGNAL *sig, RmdScope scope, uint32_t memory,
         uint32_t info, struct Dump *dump)
{
   RmdFileStatus file_status;

   /* Not used in this handler function */
   (void) scope;
   (void) memory;
   (void) info;

   /* Write reply signal to file dump */
   file_status = rmd_file_dump_reply((OSADDRESS) sig, sigsize(&sig),
            dump->dump_file);

   return rmd_file_status_converter(file_status, dump->dump_file);
}

/**
 * @function dump_pool_signals_iterator
 *
 * @brief Handler function to be called from for_each_pool_in_segment().
 *
 * @long Iterates over all signals in a pool and calls the
 * dump_pool_signals_handler for each signal.
 *
 * @seealso for_each_pool_in_segment
 * @seealso for_each_signal_in_pool
 * @seealso dump_pool_signals_handler
 *
 */
static RmdStatus
dump_pool_signals_iterator(
         struct MonitorGetPoolInfoReply *monitor_get_pool_info_reply,
         RmdScope scope, uint32_t memory, uint32_t info,
         MonitorStatus *monitor_status, struct Rmd *rmd, struct Dump *dump)
{
   return for_each_signal_in_pool(monitor_get_pool_info_reply->pid, scope,
            memory, info, monitor_status, rmd, dump, dump_pool_signals_handler);
}

/**
 * @function rmd_private_dump_pool_signals_info
 *
 * @brief Finds all the signals in all pools in the current segment and writes
 * them to the dump file.
 *
 * @long A functionality to iterate over all signals in a segment is
 * implemented because the RMM does not accept segment as scope input
 * for this signal. The request signal with the wrong scope is written
 * as the request in the dump file.
 *
 * @input rmd_req Request signal of type RmdDumpPoolSignalsInfoRequest.
 * @input rmd The Rmd struct containing information about ongoing dumps.
 *
 * @seealso RmdDumpPoolSignalsInfoRequest
 * @seealso for_each_pool_in_segment
 * @seealso dump_pool_signals_iterator
 *
 */
void
rmd_private_dump_pool_signals_info(union SIGNAL *rmd_req, struct Rmd *rmd)
{
   RmdStatus status = RMD_STATUS_OK;
   union SIGNAL *sig = NIL;
   struct Dump *dump = NULL;
   uint32_t scope;
   uint32_t info;
   OSBOOLEAN supported = TRUE;

   assert(rmd_req);
   assert(RMD_DUMP_POOL_SIGNALS_INFO_REQUEST == rmd_req->signo);
   assert(rmd);
   assert(RMD_MAGIC == rmd->magic);

   dump = get_dump_from_pid(sender(&rmd_req), &rmd->dumps);
   if (NULL == dump)
   {
      status = RMD_STATUS_DUMP_NOT_FOUND;
      goto finally;
   }

   scope = rmd_req->rmd_dump_pool_signals_info_request.scope;
   info = rmd_req->rmd_dump_pool_signals_info_request.info;
   if (info & RMD_INFO_POOL_SIGNALS)
   {
      RmdFileStatus file_status = RMD_FILE_SUCCESS;
      MonitorStatus monitor_status;
      MonitorGetInfoEndmark *reply;
      MonitorGetInfoRequest *req;

      /* Check if the interface signals are supported in the RMM */
      /* The following signals are sent from this function
       * MONITOR_GET_POOL_SIGNALS_REQUEST
       * MONITOR_GET_POOL_INFO_REQUEST
       */
      supported &= rmm_is_supported(MONITOR_GET_POOL_SIGNALS_REQUEST, rmd);
      supported &= rmm_is_supported(MONITOR_GET_POOL_INFO_REQUEST, rmd);

      if (!supported)
      {
         status = RMD_STATUS_INTERFACE_NOT_SUPPORTED;
         goto finally;
      }

      req = (MonitorGetInfoRequest *) alloc(sizeof(MonitorGetInfoRequest),
               MONITOR_GET_POOL_SIGNALS_REQUEST);

      req->level = 0;
      switch (scope)
      {
         case RMD_SCOPE_SEGMENT:
         case RMD_SCOPE_BLOCK:
         case RMD_SCOPE_PROCESS:
            req->scope.type = MONITOR_SCOPE_SEGMENT_ID;
            req->scope.id = dump->sid;
            break;
         default:
            /* Handle error */
            free_buf((union SIGNAL **) &req);
            status = RMD_STATUS_INVALID_PARAMETERS;
            goto finally;
      }
      req->reserved0 = 0;
      req->tagsCount = 0;

      /* Write request signal to file dump */
      file_status = rmd_file_dump_request((OSADDRESS) req, sigsize(
               (union SIGNAL **) &req), dump->dump_file);
      free_buf((union SIGNAL **) &req);
      if (file_status != RMD_FILE_SUCCESS)
      {
         status = rmd_file_status_converter(file_status, dump->dump_file);
         goto finally;
      }

      /*
       * Search for all signals in segment by iterating over pools and
       * requesting signals for each pool in segment.
       * Handler functions called are first dump_pool_signals_iterator()
       * which calls dump_pool_signals_handler()
       */
      status = for_each_pool_in_segment(scope, 0, info, &monitor_status, rmd,
               dump, dump_pool_signals_iterator);

      /* End signal transaction with an endmark */
      /* Create endmark signal and write it */
      reply = (MonitorGetInfoEndmark *) alloc(sizeof(MonitorGetInfoEndmark),
               MONITOR_GET_POOL_SIGNALS_ENDMARK);
      reply->status = monitor_status;
      file_status = rmd_file_dump_endmark((OSADDRESS) reply, sigsize(
               (union SIGNAL **) &reply), reply->status, dump->dump_file);
      free_buf((union SIGNAL **) &reply);
      status = rmd_file_status_converter(file_status, dump->dump_file);
   }

   finally:
   /* Send reply signal */
   sig = alloc(sizeof(RmdStatusReply), RMD_DUMP_POOL_SIGNALS_INFO_REPLY);
   sig->rmd_reply.status = status;
   send(&sig, sender(&rmd_req));
} /* rmd_private_dump_pool_signals_info() */

/**
 * @function rmd_private_dump_sys_param_info
 *
 * @brief Requests the system parameters from the RMM and writes the replies
 * to the dump file.
 *
 * @input rmd_req Request signal of type RmdDumpSysParamInfoRequest.
 * @input rmd The Rmd struct containing information about ongoing dumps.
 *
 * @seealso RmdDumpSysParamInfoRequest
 *
 */
void
rmd_private_dump_sys_param_info(union SIGNAL *rmd_req, struct Rmd *rmd)
{
   RmdStatus status = RMD_STATUS_OK;
   union SIGNAL *sig = NIL;
   struct Dump *dump = NULL;
   uint32_t info;
   uint32_t signo;
   const SIGSELECT select_reply[] =
   { 4, RMM_ATTACH_REPLY, MONITOR_GET_SYS_PARAM_REPLY,
     MONITOR_GET_SYS_PARAM_ENDMARK, MONITOR_BLOCKING_FLOW_CONTROL_REQUEST };

   assert(rmd_req);
   assert(RMD_DUMP_SYS_PARAM_INFO_REQUEST == rmd_req->signo);
   assert(rmd);
   assert(RMD_MAGIC == rmd->magic);

   /* Verify that rmd_req has scope system */
   if (rmd_req->rmd_dump_sys_param_info_request.scope != RMD_SCOPE_SYSTEM)
   {
      status = RMD_STATUS_INVALID_PARAMETERS;
      goto finally;
   }

   dump = get_dump_from_pid(sender(&rmd_req), &rmd->dumps);
   if (NULL == dump)
   {
      status = RMD_STATUS_DUMP_NOT_FOUND;
      goto finally;
   }

   info = rmd_req->rmd_dump_sys_param_info_request.info;
   if (info & RMD_INFO_SYS_PARAM)
   {
      RmdFileStatus file_status = RMD_FILE_SUCCESS;
      MonitorGetInfoRequest *req = (MonitorGetInfoRequest *) alloc(
               sizeof(MonitorGetInfoRequest), MONITOR_GET_SYS_PARAM_REQUEST);

      /* Check if the interface signals are supported in the RMM */
      /* The following signals are sent from this function
       * MONITOR_GET_SYS_PARAM_REQUEST
       */
      if (!rmm_is_supported(MONITOR_GET_SYS_PARAM_REQUEST, rmd))
      {
         status = RMD_STATUS_INTERFACE_NOT_SUPPORTED;
         goto finally;
      }

      req->level = 0;
      req->scope.type = MONITOR_SCOPE_SYSTEM;
      req->scope.id = 0;
      req->reserved0 = 0;
      req->tagsCount = 0;

      /* Write request signal to file dump */
      file_status = rmd_file_dump_request((OSADDRESS) req, sigsize(
               (union SIGNAL **) &req), dump->dump_file);
      if (file_status != RMD_FILE_SUCCESS)
      {
         free_buf((union SIGNAL **) &req);
         status = rmd_file_status_converter(file_status, dump->dump_file);
         goto finally;
      }

      /* Request information from monitor */
      send((union SIGNAL **) &req, rmd->rmm);

      do /* while (sig->signo != MONITOR_GET_SYS_PARAM_ENDMARK) */
      {
         sig = receive_from(RMD_RECEIVE_TIMEOUT, select_reply, rmd->rmm);
         assert(sig);
         signo = sig->signo;
         switch (signo)
         {
            case RMM_ATTACH_REPLY:
#ifdef RMD_VERBOSE
               ramlog_printf("RMD: RMM_ATTACH_REPLY received\n");
#endif
               status = RMD_STATUS_DETACHED_FROM_RMM;
               free_buf(&sig);
               goto finally;
            case MONITOR_GET_SYS_PARAM_REPLY:
               /* Write reply signal to file dump */
               file_status = rmd_file_dump_reply((OSADDRESS) sig,
                        sigsize(&sig), dump->dump_file);

               free_buf(&sig);

               if (file_status != RMD_FILE_SUCCESS)
               {
                  status = rmd_file_status_converter(file_status,
                           dump->dump_file);
                  goto finally;
               }

               break;
            case MONITOR_BLOCKING_FLOW_CONTROL_REQUEST:
               sig->signo = MONITOR_BLOCKING_FLOW_CONTROL_REPLY;
               send(&sig, rmd->rmm);
               break;
            case MONITOR_GET_SYS_PARAM_ENDMARK:
               break;
         }
      } while (signo != MONITOR_GET_SYS_PARAM_ENDMARK);

      /* Write endmark signal to filedump */
      file_status = rmd_file_dump_endmark((OSADDRESS) sig, sigsize(&sig),
               sig->monitor_get_info_endmark.status, dump->dump_file);
      status = rmd_file_status_converter(file_status, dump->dump_file);

      free_buf(&sig);
   }
   finally:
   /* Send reply signal */
   sig = alloc(sizeof(RmdStatusReply), RMD_DUMP_SYS_PARAM_INFO_REPLY);
   sig->rmd_reply.status = status;
   send(&sig, sender(&rmd_req));
} /* rmd_private_dump_sys_param_info() */

void
rmd_private_dump_registers_info(union SIGNAL *rmd_req, struct Rmd *rmd)
{
   RmdStatus status = RMD_STATUS_OK;
   RmdFileStatus file_status = RMD_FILE_SUCCESS;
   union SIGNAL *sig = NIL;
   struct Dump *dump = NULL;
   uint32_t scope;
   uint32_t info;
   uint32_t signo;

   assert(rmd_req);
   assert(RMD_DUMP_REGISTERS_INFO_REQUEST == rmd_req->signo);
   assert(rmd);
   assert(RMD_MAGIC == rmd->magic);

#ifdef RMD_VERBOSE
   ramlog_printf("RMD: rmd_private_dump_registers_info()\n");
#endif

   dump = get_dump_from_pid(sender(&rmd_req), &rmd->dumps);
   if (NULL == dump)
   {
      status = RMD_STATUS_DUMP_NOT_FOUND;
      goto finally;
   }

   /* Check if the interface signals are supported in the RMM */
   /* The following signals are sent from this function
    * MONITOR_GET_REGISTER_INFO_REQUEST
    */
   if (!rmm_is_supported(MONITOR_GET_REGISTER_INFO_REQUEST, rmd))
   {
      status = RMD_STATUS_INTERFACE_NOT_SUPPORTED;
      goto finally;
   }

   scope = rmd_req->rmd_dump_registers_info_request.scope;
   info = rmd_req->rmd_dump_registers_info_request.info;

   if (info & RMD_INFO_REGISTERS)
   {
      uint32_t
      rmd_get_register_info(MonitorGetRegisterInfoRequest *);
      uint32_t
      rmd_regdump_registers(struct cpu_regdump *);
      uint32_t
      fmd_regdump_registers(struct cpu_regdump *, MonitorRegisterId *,
               MonitorRegister *);

      const SIGSELECT select_reply[] =
      { 4, RMM_ATTACH_REPLY, MONITOR_GET_REGISTERS_REPLY,
        MONITOR_GET_REGISTER_INFO_ENDMARK,
        MONITOR_BLOCKING_FLOW_CONTROL_REQUEST };
      uint32_t reg_count;
      MonitorGetRegisterInfoRequest *req;

      /* Parse rmd_req->rmd_dump_registers_info_request.regdump */
      /* Will probably contain a regdump if we got an OSE_EUNEXPECTED_EXCEPTION_REGDUMP error */
      {
         MonitorGetRegisterInfoRequest *info_req;
         MonitorRegisterValues *reply;
         MonitorGetRegisterInfoEndmark *info_endmark;
         struct cpu_regdump *reg_dump;

         reg_dump = &rmd_req->rmd_dump_registers_info_request.regdump;
         /* Do we have a valid dump? */
         if (reg_dump->magic == CPU_REGDUMP_MAGIC)
         {
            reg_count = rmd_regdump_registers(reg_dump);
#ifdef RMD_VERBOSE
            ramlog_printf("RMD: Found %d registers in OSE_EUNEXPECTED_EXCEPTION_REGDUMP\n", reg_count);
#endif
            info_req = (MonitorGetRegisterInfoRequest *) alloc(
                     sizeof(MonitorGetRegisterInfoRequest) + (reg_count - 1)
                              * sizeof(MonitorRegisterId),
                     MONITOR_GET_REGISTER_INFO_REQUEST);
            reply = (MonitorRegisterValues *) alloc(
                     sizeof(MonitorRegisterValues) + sizeof(MonitorRegister)
                              * reg_count, MONITOR_GET_REGISTERS_REPLY);
            info_endmark = (MonitorGetRegisterInfoEndmark *) alloc(
                     sizeof(MonitorGetRegisterInfoEndmark),
                     MONITOR_GET_REGISTER_INFO_ENDMARK);

            info_req->scope.type = MONITOR_SCOPE_ID;
            info_req->scope.id = dump->pid;
            info_req->registersCount = reg_count;

            reply->status = MONITOR_STATUS_OK;
            reply->pid = dump->pid;
            reply->registersCount = reg_count;

            info_endmark->status = MONITOR_STATUS_OK;

            /* Get the register contents from the freeze mode dumper */
            /* 'ids' is not used but fmd_regdump_registers() do not allow */
            /* NULL pointer arguments */
            fmd_regdump_registers(reg_dump, info_req->registers,
                     reply->registers);

            /* Write MONITOR_GET_REGISTER_INFO_REQUEST to dump */
            file_status = rmd_file_dump_request((OSADDRESS) info_req, sigsize(
                     (union SIGNAL **) &info_req), dump->dump_file);
            free_buf((union SIGNAL **) &info_req);

            if (RMD_FILE_SUCCESS != file_status)
            {
               status = rmd_file_status_converter(file_status, dump->dump_file);
               goto finally;
            }

            /* Write reply signal to file dump */
            file_status = rmd_file_dump_reply((OSADDRESS) reply, sigsize(
                     (union SIGNAL **) &reply), dump->dump_file);
            free_buf((union SIGNAL **) &reply);

            if (RMD_FILE_SUCCESS != file_status)
            {
               status = rmd_file_status_converter(file_status, dump->dump_file);
               goto finally;
            }

            /* Write endmark signal to filedump */
            file_status = rmd_file_dump_endmark((OSADDRESS) info_endmark,
                     sigsize((union SIGNAL **) &info_endmark),
                     info_endmark->status, dump->dump_file);
            free_buf((union SIGNAL **) &info_endmark);

            if (RMD_FILE_SUCCESS != file_status)
            {
               status = rmd_file_status_converter(file_status, dump->dump_file);
               goto finally;
            }
         }
      }

      /* Fetch the number of registers */
      reg_count = rmd_get_register_info(NULL);
      req = (MonitorGetRegisterInfoRequest *) alloc(
               sizeof(MonitorGetRegisterInfoRequest) + (reg_count - 1)
                        * sizeof(MonitorRegisterId),
               MONITOR_GET_REGISTER_INFO_REQUEST);
      /* Initialize the signal with default values and the architecture */
      /* specific register set */
      (void) rmd_get_register_info(req);
      switch (scope)
      {
         case RMD_SCOPE_SEGMENT:
            req->scope.type = MONITOR_SCOPE_SEGMENT_ID;
            req->scope.id = dump->sid;
            break;
         case RMD_SCOPE_BLOCK:
            req->scope.type = MONITOR_SCOPE_BLOCK_ID;
            req->scope.id = dump->bid;
            break;
         case RMD_SCOPE_PROCESS:
            req->scope.type = MONITOR_SCOPE_ID;
            req->scope.id = dump->pid;
            break;
         default:
            /* Handle error */
            free_buf((union SIGNAL **) &req);
            status = RMD_STATUS_INVALID_PARAMETERS;
            goto finally;
      }

      /* Write request signal to file dump */
      status = rmd_file_dump_request((OSADDRESS) req, sigsize(
               (union SIGNAL **) &req), dump->dump_file);

      if (RMD_STATUS_OK != status)
      {
         free_buf((union SIGNAL **) &req);
         goto finally;
      }

      /* Request information from monitor */
      send((union SIGNAL **) &req, rmd->rmm);
      do /* while (sig->signo != MONITOR_GET_REGISTER_INFO_ENDMARK) */
      {
         sig = receive_from(RMD_RECEIVE_TIMEOUT, select_reply, rmd->rmm);
         assert(sig);
         signo = sig->signo;
         switch (signo)
         {
            case RMM_ATTACH_REPLY:
#ifdef RMD_VERBOSE
               ramlog_printf("RMD: RMM_ATTACH_REPLY received\n");
#endif
               status = RMD_STATUS_DETACHED_FROM_RMM;
               free_buf(&sig);
               goto finally;
            case MONITOR_GET_REGISTERS_REPLY:
               /* Write reply signal to file dump */
               status = rmd_file_dump_reply((OSADDRESS) sig, sigsize(
                        (union SIGNAL **) &sig), dump->dump_file);
               free_buf(&sig);

               if (status != RMD_STATUS_OK)
               {
                  goto finally;
               }
               break;
            case MONITOR_BLOCKING_FLOW_CONTROL_REQUEST:
               sig->signo = MONITOR_BLOCKING_FLOW_CONTROL_REPLY;
               send(&sig, rmd->rmm);
               break;
            case MONITOR_GET_REGISTER_INFO_ENDMARK:
               break;
         }
      } while (signo != MONITOR_GET_REGISTER_INFO_ENDMARK);

      /* Write endmark signal to filedump */
      status = rmd_file_dump_endmark((OSADDRESS) sig, sigsize(&sig),
               sig->monitor_get_info_endmark.status, dump->dump_file);

      free_buf(&sig);
   }

   finally:
   /* Clean up */
   if (sig != NIL)
   {
      free_buf(&sig);
   }

   /* Send reply signal */
   sig = alloc(sizeof(RmdStatusReply), RMD_DUMP_REGISTERS_INFO_REPLY);
   sig->rmd_reply.status = status;
   send(&sig, sender(&rmd_req));
}

/* Program Manager helpers */
/**
 * @function get_pm_info_request
 *
 * @brief Requests and returns the PmProgramInfo signal from the program
 * manager.
 *
 * @input sender_pid Process id for the process calling the RMD process.
 *
 * @returns The PmProgramInfoReply signal if ok, NIL if not ok.
 */
static union SIGNAL *
get_pm_info_request(PROCESS sender_pid)
{
   union SIGNAL *sig = NIL;
   PROCESS pm_process;
   static const SIGSELECT pm_program_info[] =
   { 1, PM_PROGRAM_INFO_REPLY };

   /* Hunt for program manager */
   pm_process = hunt_for_pm();
   if (!pm_process)
   {
      /* Failure */
      return NIL;
   }
   if (pm_is_supported(PM_PROGRAM_INFO_REQUEST, pm_process))
   {
      /* Send and receive signal to PM */
      sig = alloc(sizeof(struct PmProgramInfo), PM_PROGRAM_INFO_REQUEST);
      sig->program_info.progpid = sender_pid;
      send (&sig, pm_process);
      sig = receive(pm_program_info);
   }
   return sig;
}

/**
 * @function get_pm_load_module_info_signal
 *
 * @brief Requests and returns the PmLoadModuleInfoReply signal from the
 * program manager.
 *
 * @input install_handle Null terminated string containing the handle
 *                       for the load module.
 *
 * @returns The PmLoadModuleInfoReply signal if ok, NIL if not ok.
 */
static union SIGNAL *
get_pm_load_module_info_signal(char *install_handle)
{
   union SIGNAL *sig = NIL;
   PROCESS pm_process;
   static const SIGSELECT pm_load_module_info[] =
   { 1, PM_LOAD_MODULE_INFO_REPLY };

   /* Hunt for program manager */
   pm_process = hunt_for_pm();
   if (!pm_process)
   {
      /* Failure */
      return NIL;
   }
   if (pm_is_supported(PM_LOAD_MODULE_INFO_REQUEST, pm_process))
   {
      sig = alloc(sizeof(struct PmLoadModuleInfoRequest),
               PM_LOAD_MODULE_INFO_REQUEST);
      strncpy(sig->load_module_info_request.install_handle, install_handle,
               PM_INSTALL_HANDLE_LENGTH);

      send(&sig, pm_process);
      sig = receive(pm_load_module_info);
   }
   return sig;
}

/* End Program Manager helpers */

/**
 * @function rmd_private_dump_program_info
 *
 * @brief Sends the two signals PmLoadModuleInfoRequest and PmProgramInfo to
 * the program manager and writes the results to the dump file.
 *
 * @input rmd_req Request signal of type RmdDumpProgramInfoRequest
 *
 * @seealso RmdDumpProgramInfoRequest
 */
void
rmd_private_dump_program_info(union SIGNAL *rmd_req, struct Rmd *rmd)
{
   RmdStatus status = RMD_STATUS_OK;
   union SIGNAL *sig = NIL;
   struct Dump *dump = NULL;
   uint32_t scope;
   uint32_t info;
   const SIGSELECT select_reply[] =
   { 1, PM_PROGRAMS_REPLY };
   const SIGSELECT select_info_reply[] =
   { 1, PM_PROGRAM_INFO_REPLY };

   assert(rmd_req);
   assert(RMD_DUMP_PROGRAM_INFO_REQUEST == rmd_req->signo);
   assert(rmd);
   assert(RMD_MAGIC == rmd->magic);

   scope = rmd_req->rmd_dump_pool_info_request.scope;
   info = rmd_req->rmd_dump_pool_info_request.info;
   dump = get_dump_from_pid(sender(&rmd_req), &rmd->dumps);
   if (NULL == dump)
   {
      status = RMD_STATUS_DUMP_NOT_FOUND;
      goto finally;
   }

   if (info & RMD_INFO_PROGRAM_INFO)
   {
      RmdFileStatus file_status = RMD_FILE_SUCCESS;
      PROCESS pm_process;
      struct PmProgramsRequest *req;
      struct PmProgramInfo *info_req;
      OSBOOLEAN supported = TRUE;

      /* Hunt for program manager */
      pm_process = hunt_for_pm();
      if (!pm_process)
      {
         status = RMD_STATUS_INTERNAL_ERROR;
         goto finally;
      }

      /* Check that interfaces are supported */
      supported &= pm_is_supported(PM_PROGRAMS_REQUEST, pm_process);
      supported &= pm_is_supported(PM_PROGRAM_INFO_REQUEST, pm_process);
      if (!supported)
      {
         status = RMD_STATUS_INTERFACE_NOT_SUPPORTED;
         goto finally;
      }
      /* PM_PROGRAMS_REQUEST */
      req = (struct PmProgramsRequest *) alloc(
               sizeof(struct PmProgramsRequest), PM_PROGRAMS_REQUEST);

      switch (scope)
      {
         case RMD_SCOPE_SEGMENT:
         case RMD_SCOPE_BLOCK:
         case RMD_SCOPE_PROCESS:
            break;
         default:
            /* Handle error */
            free_buf((union SIGNAL **) &req);
            status = RMD_STATUS_INVALID_PARAMETERS;
            goto finally;
      }

      /* Write request signal to file dump */
      file_status = rmd_file_dump_request((OSADDRESS) req, sigsize(
               (union SIGNAL **) &req), dump->dump_file);
      if (file_status != RMD_FILE_SUCCESS)
      {
         free_buf((union SIGNAL **) &req);
         status = rmd_file_status_converter(file_status, dump->dump_file);
         goto finally;
      }

      /* Request information from program manager */
      send((union SIGNAL **) &req, pm_process);
      sig = receive_from(RMD_RECEIVE_TIMEOUT, select_reply, pm_process);
      assert(sig);

      /* Write endmark signal to filedump */
      file_status = rmd_file_dump_endmark((OSADDRESS) sig, sigsize(&sig),
               sig->monitor_get_info_endmark.status, dump->dump_file);
      free_buf(&sig);
      if (file_status != RMD_FILE_SUCCESS)
      {
         status = rmd_file_status_converter(file_status, dump->dump_file);
         goto finally;
      }

      /* PM_PROGRAM_INFO_REQUEST */
      info_req = (struct PmProgramInfo *) alloc(sizeof(struct PmProgramInfo),
               PM_PROGRAM_INFO_REQUEST);

      info_req->status = PM_SUCCESS;
      info_req->progpid = dump->pid;

      /* Write request signal to file dump */
      file_status = rmd_file_dump_request((OSADDRESS) info_req, sigsize(
               (union SIGNAL **) &info_req), dump->dump_file);
      if (file_status != RMD_FILE_SUCCESS)
      {
         free_buf((union SIGNAL **) &info_req);
         status = rmd_file_status_converter(file_status, dump->dump_file);
         goto finally;
      }

      /* Request information from program manager */
      send((union SIGNAL **) &info_req, pm_process);
      sig = receive_from(RMD_RECEIVE_TIMEOUT, select_info_reply, pm_process);
      assert(sig);

      /* Write endmark signal to filedump */
      file_status = rmd_file_dump_endmark((OSADDRESS) sig, sigsize(&sig),
               sig->monitor_get_info_endmark.status, dump->dump_file);
      /*     status = write_file_signal_data(sig, RMD_FILE_WRITE_ENDMARK, dump);*/
      free_buf(&sig);
      if (file_status != RMD_FILE_SUCCESS)
      {
         status = rmd_file_status_converter(file_status, dump->dump_file);
         goto finally;
      }
   }

   finally:
   /* Send reply signal */
   sig = alloc(sizeof(RmdStatusReply), RMD_DUMP_PROGRAM_INFO_REPLY);
   sig->rmd_reply.status = status;
   send(&sig, sender(&rmd_req));
} /* rmd_private_dump_program_info() */

/**
 * @function rmd_private_dump_trace_info
 *
 * @brief Requests information from the RMM regarding trace information and
 * writes the results to the dump file.
 *
 * @input rmd_req Request signal of type RmdDumpTraceInfoRequest
 *
 * @seealso RmdDumpTraceInfoRequest
 */
void
rmd_private_dump_trace_info(union SIGNAL *rmd_req, struct Rmd *rmd)
{
   RmdStatus status = RMD_STATUS_OK;
   union SIGNAL *sig = NIL;
   struct Dump *dump = NULL;
   uint32_t info;

   assert(rmd_req);
   assert(RMD_DUMP_TRACE_INFO_REQUEST == rmd_req->signo);
   assert(rmd);
   assert(RMD_MAGIC == rmd->magic);

   /* Verify that rmd_req has scope system */
   if (rmd_req->rmd_dump_sys_param_info_request.scope != RMD_SCOPE_SYSTEM)
   {
      status = RMD_STATUS_INVALID_PARAMETERS;
      goto finally;
   }

   info = rmd_req->rmd_dump_sys_param_info_request.info;
   dump = get_dump_from_pid(sender(&rmd_req), &rmd->dumps);
   if (NULL == dump)
   {
      status = RMD_STATUS_DUMP_NOT_FOUND;
      goto finally;
   }

   if (info & RMD_INFO_TRACE)
   {
      RmdFileStatus file_status = RMD_FILE_SUCCESS;
      uint32_t signo;
      const SIGSELECT select_reply[] =
      { 17, RMM_ATTACH_REPLY, MONITOR_GET_PROCESS_TRACE_REPLY,
        MONITOR_GET_SEND_TRACE_REPLY, MONITOR_GET_RECEIVE_TRACE_REPLY,
        MONITOR_GET_CREATE_TRACE_REPLY, MONITOR_GET_KILL_TRACE_REPLY,
        MONITOR_GET_SWAP_TRACE_REPLY, MONITOR_GET_ERROR_TRACE_REPLY,
        MONITOR_GET_RESET_TRACE_REPLY, MONITOR_GET_USER_TRACE_REPLY,
        MONITOR_GET_BIND_TRACE_REPLY, MONITOR_GET_ALLOC_TRACE_REPLY,
        MONITOR_GET_FREE_BUF_TRACE_REPLY, MONITOR_GET_TIMEOUT_TRACE_REPLY,
        MONITOR_GET_LOSS_TRACE_REPLY, MONITOR_GET_TRACE_ENDMARK,
        MONITOR_BLOCKING_FLOW_CONTROL_REQUEST };

      MonitorGetTraceRequest *req = (MonitorGetTraceRequest *) alloc(
               sizeof(MonitorGetTraceRequest), MONITOR_GET_TRACE_REQUEST);

      /* Check if the interface signals are supported in the RMM */
      /* The following signals are sent from this function
       * MONITOR_GET_TRACE_REQUEST
       */
      if (!rmm_is_supported(MONITOR_GET_TRACE_REQUEST, rmd))
      {
         status = RMD_STATUS_INTERFACE_NOT_SUPPORTED;
         goto finally;
      }

      req->from = 0;
      req->to = ~0;
      req->reserved0 = 0;
      req->reserved1 = 0;
      req->reserved2 = 0;
      req->tagsCount = 0;

      /* Write request signal to file dump */
      file_status = rmd_file_dump_request((OSADDRESS) req, sigsize(
               (union SIGNAL **) &req), dump->dump_file);
      if (file_status != RMD_FILE_SUCCESS)
      {
         free_buf((union SIGNAL **) &req);
         status = rmd_file_status_converter(file_status, dump->dump_file);
         goto finally;
      }

      /* Request information from monitor */
      send((union SIGNAL **) &req, rmd->rmm);

      do /* while (sig->signo != MONITOR_GET_TRACE_ENDMARK) */
      {
         sig = receive_from(RMD_RECEIVE_TIMEOUT, select_reply, rmd->rmm);
         assert(sig);
         signo = sig->signo;
         switch (signo)
         {
            case RMM_ATTACH_REPLY:
#ifdef RMD_VERBOSE
               ramlog_printf("RMD: RMM_ATTACH_REPLY received\n");
#endif
               status = RMD_STATUS_DETACHED_FROM_RMM;
               free_buf(&sig);
               goto finally;
            case MONITOR_GET_PROCESS_TRACE_REPLY:
            case MONITOR_GET_SEND_TRACE_REPLY:
            case MONITOR_GET_RECEIVE_TRACE_REPLY:
            case MONITOR_GET_CREATE_TRACE_REPLY:
            case MONITOR_GET_KILL_TRACE_REPLY:
            case MONITOR_GET_SWAP_TRACE_REPLY:
            case MONITOR_GET_ERROR_TRACE_REPLY:
            case MONITOR_GET_RESET_TRACE_REPLY:
            case MONITOR_GET_USER_TRACE_REPLY:
            case MONITOR_GET_BIND_TRACE_REPLY:
            case MONITOR_GET_ALLOC_TRACE_REPLY:
            case MONITOR_GET_FREE_BUF_TRACE_REPLY:
            case MONITOR_GET_TIMEOUT_TRACE_REPLY:
            case MONITOR_GET_LOSS_TRACE_REPLY:
               /* Write reply signal to file dump */
               status = rmd_file_dump_reply((OSADDRESS) sig, sigsize(&sig),
                        dump->dump_file);
               free_buf(&sig);

               if (file_status != RMD_FILE_SUCCESS)
               {
                  status = rmd_file_status_converter(file_status,
                           dump->dump_file);
                  goto finally;
               }

               break;
            case MONITOR_BLOCKING_FLOW_CONTROL_REQUEST:
               sig->signo = MONITOR_BLOCKING_FLOW_CONTROL_REPLY;
               send(&sig, rmd->rmm);
               break;
            case MONITOR_GET_TRACE_ENDMARK:
               break;
         }
      } while (signo != MONITOR_GET_TRACE_ENDMARK);
      /* Write endmark signal to filedump */
      status = rmd_file_dump_endmark((OSADDRESS) sig, sigsize(&sig),
               sig->monitor_get_info_endmark.status, dump->dump_file);
      free_buf(&sig);
      if (file_status != RMD_FILE_SUCCESS)
      {
         status = rmd_file_status_converter(file_status, dump->dump_file);
         goto finally;
      }
   }
   finally:
   /* Send reply signal */
   sig = alloc(sizeof(RmdStatusReply), RMD_DUMP_TRACE_INFO_REPLY);
   sig->rmd_reply.status = status;
   send(&sig, sender(&rmd_req));
} /* rmd_private_dump_trace_info() */

/**
 * @function rmd_private_dump_load_module_info
 *
 * @brief Requests information about the load module from the program manager
 * and writes the results to file.
 *
 * @input rmd_req Request signal of type RmdDumpTraceInfoRequest
 *
 * @seealso RmdDumpTraceInfoRequest
 */
void
rmd_private_dump_load_module_info(union SIGNAL *rmd_req, struct Rmd *rmd)
{
   RmdStatus status = RMD_STATUS_OK;
   RmdFileStatus file_status = RMD_FILE_SUCCESS;
   union SIGNAL *sig = NIL;
   struct Dump *dump = NULL;
   uint32_t scope;
   uint32_t info;
   PROCESS pm_process = 0;

   assert(rmd_req);
   assert(RMD_DUMP_LOAD_MODULE_INFO_REQUEST == rmd_req->signo);
   assert(rmd);
   assert(RMD_MAGIC == rmd->magic);

   dump = get_dump_from_pid(sender(&rmd_req), &rmd->dumps);
   if (NULL == dump)
   {
      status = RMD_STATUS_DUMP_NOT_FOUND;
      goto finally;
   }

   /* Hunt for program manager */
   pm_process = hunt_for_pm();
   if (!pm_process)
   {
      status = RMD_STATUS_INTERNAL_ERROR;
      goto finally;
   }

   scope = rmd_req->rmd_dump_load_module_info_request.scope;
   info = rmd_req->rmd_dump_load_module_info_request.info;
   /* PM_LOAD_MODULE_INSTALL_HANDLES_REQUEST */
   if (info & RMD_INFO_LOAD_MODULE_INFO)
   {
      SIGSELECT select_reply[] =
      { 1, PM_LOAD_MODULE_INSTALL_HANDLES_REPLY };
      struct PmLoadModuleInstallHandlesRequest *req;

      if (!pm_is_supported(PM_LOAD_MODULE_INSTALL_HANDLES_REQUEST, pm_process))
      {
         status = RMD_STATUS_INTERFACE_NOT_SUPPORTED;
         goto finally;
      }
      req = (struct PmLoadModuleInstallHandlesRequest *) alloc(
               sizeof(struct PmLoadModuleInstallHandlesRequest),
               PM_LOAD_MODULE_INSTALL_HANDLES_REQUEST);

      switch (scope)
      {
         case RMD_SCOPE_SEGMENT:
         case RMD_SCOPE_BLOCK:
         case RMD_SCOPE_PROCESS:
            break;
         default:
            /* Handle error */
            free_buf((union SIGNAL **) &req);
            status = RMD_STATUS_INVALID_PARAMETERS;
            goto finally;
      }

      /* Write request signal to file dump */
      file_status = rmd_file_dump_request((OSADDRESS) req, sigsize(
               (union SIGNAL **) &req), dump->dump_file);
      if (file_status != RMD_STATUS_OK)
      {
         free_buf((union SIGNAL **) &req);
         status = rmd_file_status_converter(file_status, dump->dump_file);
         goto finally;
      }

      /* Request information from program manager */
      send((union SIGNAL **) &req, pm_process);
      sig = receive_from(RMD_RECEIVE_TIMEOUT, select_reply, pm_process);
      assert(sig);

      /* Write endmark signal to filedump */
      file_status = rmd_file_dump_endmark((OSADDRESS) sig, sigsize(&sig),
               sig->monitor_get_info_endmark.status, dump->dump_file);
      free_buf(&sig);
      if (file_status != RMD_STATUS_OK)
      {
         status = rmd_file_status_converter(file_status, dump->dump_file);
         goto finally;
      }
   }
   /* PM_LOAD_MODULE_INFO_REQUEST */
   if (info & RMD_INFO_LOAD_MODULE_INFO)
   {
      SIGSELECT select_reply[] =
      { 1, PM_LOAD_MODULE_INFO_REPLY };
      struct PmLoadModuleInfoRequest *req;
      union SIGNAL *pm_info_sig;

      if (!pm_is_supported(PM_LOAD_MODULE_INFO_REQUEST, pm_process))
      {
         status = RMD_STATUS_INTERFACE_NOT_SUPPORTED;
         goto finally;
      }

      pm_info_sig = get_pm_info_request(dump->bid);
      if (NIL == pm_info_sig)
      {
         status = RMD_STATUS_INTERNAL_ERROR;
         goto finally;
      }

      req = (struct PmLoadModuleInfoRequest *) alloc(
               sizeof(struct PmLoadModuleInfoRequest),
               PM_LOAD_MODULE_INFO_REQUEST);

      strcpy(req->install_handle, pm_info_sig->program_info.install_handle);
      free_buf(&pm_info_sig);

      /* Write request signal to file dump */
      file_status = rmd_file_dump_request((OSADDRESS) req, sigsize(
               (union SIGNAL **) &req), dump->dump_file);
      if (file_status != RMD_STATUS_OK)
      {
         status = rmd_file_status_converter(file_status, dump->dump_file);
         free_buf((union SIGNAL **) &req);
         goto finally;
      }

      /* Request information from program manager */
      send((union SIGNAL **) &req, pm_process);
      sig = receive_from(RMD_RECEIVE_TIMEOUT, select_reply, pm_process);
      assert(sig);

      /* Write endmark signal to filedump */
      file_status = rmd_file_dump_endmark((OSADDRESS) sig, sigsize(&sig),
               sig->monitor_get_info_endmark.status, dump->dump_file);
      free_buf(&sig);
      if (file_status != RMD_STATUS_OK)
      {
         status = rmd_file_status_converter(file_status, dump->dump_file);
         goto finally;
      }
   }
   /* PM_LOAD_MODULE_SECTION_INFO_REQUEST */
   if (info & RMD_INFO_LOAD_MODULE_INFO)
   {
      SIGSELECT select_reply[] =
      { 1, PM_LOAD_MODULE_SECTION_INFO_REPLY };
      struct PmLoadModuleSectionInfoRequest *req;

      union SIGNAL *program_info_sig;
      union SIGNAL *module_info_sig;

      if (!pm_is_supported(PM_LOAD_MODULE_SECTION_INFO_REQUEST, pm_process))
      {
         status = RMD_STATUS_INTERFACE_NOT_SUPPORTED;
         goto finally;
      }

      program_info_sig = get_pm_info_request(dump->bid);
      if (NIL == program_info_sig)
      {
         status = RMD_STATUS_INTERNAL_ERROR;
         goto finally;
      }

      module_info_sig = get_pm_load_module_info_signal(
               program_info_sig->program_info.install_handle);
      if (NIL == module_info_sig)
      {
         status = RMD_STATUS_INTERNAL_ERROR;
         free_buf(&program_info_sig);
         goto finally;
      }

      req = (struct PmLoadModuleSectionInfoRequest *) alloc(
               sizeof(struct PmLoadModuleSectionInfoRequest),
               PM_LOAD_MODULE_SECTION_INFO_REQUEST);

      strcpy(req->install_handle, program_info_sig->program_info.install_handle);
      free_buf(&program_info_sig);
      req->section_interval_start = 0;
      req->section_interval_end
               = module_info_sig->load_module_info_reply.no_of_sections - 1;
      free_buf(&module_info_sig);

      /* Write request signal to file dump */
      file_status = rmd_file_dump_request((OSADDRESS) req, sigsize(
               (union SIGNAL **) &req), dump->dump_file);
      if (file_status != RMD_STATUS_OK)
      {
         status = rmd_file_status_converter(file_status, dump->dump_file);
         free_buf((union SIGNAL **) &req);
         goto finally;
      }

      /* Request information from program manager */
      send((union SIGNAL **) &req, pm_process);
      sig = receive_from(RMD_RECEIVE_TIMEOUT, select_reply, pm_process);
      assert(sig);

      /* Write endmark signal to filedump */
      file_status = rmd_file_dump_endmark((OSADDRESS) sig, sigsize(&sig),
               sig->monitor_get_info_endmark.status, dump->dump_file);
      free_buf(&sig);
      if (file_status != RMD_STATUS_OK)
      {
         status = rmd_file_status_converter(file_status, dump->dump_file);
         goto finally;
      }
   }

   finally:
   /* Send reply signal */
   sig = alloc(sizeof(RmdStatusReply), RMD_DUMP_LOAD_MODULE_INFO_REPLY);
   sig->rmd_reply.status = status;
   send(&sig, sender(&rmd_req));
} /* rmd_private_dump_load_module_info() */

/**
 * @function dump_aligned_memory
 *
 * @brief Writes the memory at address aligned (down) to alignment to the
 * dump file.
 *
 * @input description Null terminated string to write to the dump in the memory
 *                    block.
 * @input address Memory address to write to file.
 * @input size Size of memory block to write.
 * @input alignment The value to align to a multiple of.
 * @input rmd_dump_file Dump file to write to.
 *
 */
static RmdStatus
dump_aligned_memory(PROCESS block, const char *description, OSADDRESS address, uint32_t size,
         uint32_t alignment, RmdDumpFile *rmd_dump_file)
{
   RmdFileStatus ret;

   uint32_t aligned_address = ALIGN_DOWN_TO(address, alignment);
   uint32_t aligned_size = ALIGN_UP_TO(size + (address - aligned_address),
            alignment);

   ret = rmd_file_dump_memory_block(block, aligned_address, aligned_size, description,
            rmd_dump_file);
   return rmd_file_status_converter(ret, rmd_dump_file);
} /* dump_aligned_memory() */

#define STACK_ALIGN 256

/**
 * @function pid_belongs_to_scope
 *
 * @brief Checks if the pid belongs to the current dumps scope.
 *
 * @input pid Process id to check
 * @input dump Current dump
 * @input scope Scope to investigate.
 *
 * @returns TRUE if the pid is part of the scope, FALSE otherwise.
 */
static OSBOOLEAN
pid_belongs_to_scope(PROCESS pid, struct Dump *dump, RmdScope scope)
{
   if (pid == 0)
   {
      return TRUE;
   }
   if (get_ptype(pid) != OS_ILLEGAL)
   {
      switch (scope)
      {
         case RMD_SCOPE_SEGMENT:
            if (get_segid(pid) == dump->sid)
               return TRUE;
            return FALSE;
         case RMD_SCOPE_BLOCK:
            if (get_bid(pid) == dump->bid)
               return TRUE;
            return FALSE;
         case RMD_SCOPE_PROCESS:
            if (pid == dump->pid)
               return TRUE;
            return FALSE;
         case RMD_SCOPE_SYSTEM:
         default:
            return TRUE;
      }
   }
   else
   {
      return FALSE;
   }
}

/**
 * @function dump_pool_signals_memory_handler
 *
 * @brief Signal handler for for_each_signal_in_pool
 *
 * @seealso for_each_signal_in_pool
 */
static RmdStatus
dump_pool_signals_memory_handler(union SIGNAL *sig, RmdScope scope,
         uint32_t memory, uint32_t info, struct Dump *dump)
{
   uint32_t iterator;
   RmdFileStatus rmd_file_status = RMD_FILE_SUCCESS;
   RmdStatus status = RMD_STATUS_OK;

   (void) info;

   if (memory & RMD_MEMORY_SIGNALS_OWNED)
   {
      /* Iterate over all signals in pool */
      for (iterator = 0; iterator
               < sig->monitor_get_pool_signals_reply.signalsCount; iterator++)
      {
         /* Make sure signal owner belongs to our scope */
         if (sig->monitor_get_pool_signals_reply.signals[iterator].owner != 0
                  && pid_belongs_to_scope(
                           sig->monitor_get_pool_signals_reply.signals[iterator].owner,
                           dump, scope))
         {
            rmd_file_status
                     = rmd_file_dump_memory_block(dump->bid,
                              sig->monitor_get_pool_signals_reply.signals[iterator].address,
                              sig->monitor_get_pool_signals_reply.signals[iterator].size,
                              "SIGNAL", dump->dump_file);
         }
         if (RMD_FILE_SUCCESS != rmd_file_status)
         {
            status
                     = rmd_file_status_converter(rmd_file_status,
                              dump->dump_file);
            break; /*for (iterator = 0; iterator
             < sig->monitor_get_pool_signals_reply.signalsCount; iterator++)*/
         }
      }
   }

   else if (memory & RMD_MEMORY_SIGNALS_SENT)
   {
      /* Iterate over all signals in pool */
      for (iterator = 0; iterator
               < sig->monitor_get_pool_signals_reply.signalsCount; iterator++)
      {
         /* Make sure signal sender belongs to our scope */
         if (sig->monitor_get_pool_signals_reply.signals[iterator].owner == 0
                  && sig->monitor_get_pool_signals_reply.signals[iterator].sender
                           != 0 && pid_belongs_to_scope(
                  sig->monitor_get_pool_signals_reply.signals[iterator].sender,
                  dump, scope))
         {
            rmd_file_status
                     = rmd_file_dump_memory_block(dump->bid,
                              sig->monitor_get_pool_signals_reply.signals[iterator].address,
                              sig->monitor_get_pool_signals_reply.signals[iterator].size,
                              "SIGNAL", dump->dump_file);
         }
         if (RMD_FILE_SUCCESS != rmd_file_status)
         {
            status
                     = rmd_file_status_converter(rmd_file_status,
                              dump->dump_file);
            break; /*for (iterator = 0; iterator
             < sig->monitor_get_pool_signals_reply.signalsCount; iterator++)*/
         }
      }
   }

   return status;
}

/**
 * @function dump_pool_signals_memory_pool_iterator
 *
 * @brief Signal handler for for_each_pool_in_segment
 *
 * @seealso for_each_pool_in_segment
 */
static RmdStatus
dump_pool_signals_memory_pool_iterator(
         struct MonitorGetPoolInfoReply *monitor_get_pool_info_reply,
         RmdScope scope, uint32_t memory, uint32_t info,
         MonitorStatus *monitor_status, struct Rmd *rmd, struct Dump *dump)
{
   return for_each_signal_in_pool(monitor_get_pool_info_reply->pid, scope,
            memory, info, monitor_status, rmd, dump,
            dump_pool_signals_memory_handler);
}

/**
 * @function dump_pool_fragments_handler
 *
 * @brief Signal handler for for_each_pool_in_segment
 *
 * @seealso for_each_pool_in_segment
 */
static RmdStatus
dump_pool_fragments_handler(
         struct MonitorGetPoolInfoReply *monitor_get_pool_info_reply,
         RmdScope scope, uint32_t memory, uint32_t info,
         MonitorStatus *monitor_status, struct Rmd *rmd, struct Dump *dump)
{
   uint32_t iterator;
   RmdStatus status = RMD_STATUS_OK;
   RmdFileStatus rmd_file_status = RMD_FILE_SUCCESS;

   MonitorPoolFragmentInfo
            *poolFragmentInfo =
                     (void *) &monitor_get_pool_info_reply->data[monitor_get_pool_info_reply->fragmentsOffset];

   /* Not used in this handler */
   (void) scope;
   (void) info;
   (void) monitor_status;
   (void) rmd;

   /* Iterate over all fragments in pool */
   for (iterator = 0; iterator < monitor_get_pool_info_reply->fragmentsCount; iterator++)
   {
      if (memory & RMD_MEMORY_POOL)
      {
         status
                  = dump_aligned_memory(dump->bid,
                           "POOL FRAGMENT",
                           poolFragmentInfo[iterator].address,
                           poolFragmentInfo[iterator].size, STACK_ALIGN,
                           dump->dump_file);
      }
      else if (memory & RMD_MEMORY_STACKS)
      {
         /* Assume stacks allocated from high fragment end */
         uint32_t base = poolFragmentInfo[iterator].address
                  + poolFragmentInfo[iterator].size
                  - poolFragmentInfo[iterator].stacks;

         status = dump_aligned_memory(dump->bid,
                  "POOL FRAGMENT STACKS", base,
                  poolFragmentInfo[iterator].stacks, STACK_ALIGN,
                  dump->dump_file);
      }
      else if (memory & RMD_MEMORY_SIGNALS)
      {
         rmd_file_status = rmd_file_dump_memory_block(dump->bid,
                  poolFragmentInfo[iterator].address,
                  poolFragmentInfo[iterator].signals, "POOL FRAGMENT SIGNALS",
                  dump->dump_file);
      }

      if (RMD_FILE_SUCCESS != rmd_file_status)
      {
         status = rmd_file_status_converter(rmd_file_status, dump->dump_file);
         break; /* for (iterator = 0; iterator
          < monitor_get_pool_info_reply->fragmentsCount; iterator++) */
      }
   }

   return status;
} /* dump_pool_fragments_handler() */

/**
 * @function dump_pool_memory
 *
 * @brief Function to dump different kind of pool memories depending on the
 * scope and memory parameters.
 *
 * @input scope The scope of the operation, can be Segment, Block or Process.
 * @input memory Flags to indicate what to dump of the pool memories.
 * @input rmd The RMD struct that holds information about the ongoing dumps.
 * @input dump Dump to write to.
 *
 * @return RmdStatus code
 *
 * @seealso RmdStatus
 * @seealso for_each_pool_in_segment
 *
 */
static RmdStatus
dump_pool_memory(uint32_t scope, uint32_t memory, struct Rmd *rmd,
         struct Dump *dump)
{
   RmdStatus status = RMD_STATUS_OK;
   /* Reusing the same functionality when dumping pool memory and pool signals,
    * monitor_status (status in endmark signal) is not used in this case */
   MonitorStatus monitor_status;

   if (memory & RMD_MEMORY_SIGNALS_OWNED || memory & RMD_MEMORY_SIGNALS_SENT)
   {
      status = for_each_pool_in_segment(scope, memory, 0, &monitor_status, rmd,
               dump, dump_pool_signals_memory_pool_iterator);
   }
   else
   {
      status = for_each_pool_in_segment(scope, memory, 0, &monitor_status, rmd,
               dump, dump_pool_fragments_handler);
   }

   return status;
} /* dump_pool_memory() */


/**
 * @function dump_stack_usage_memory
 *
 * @brief Dumps the stacks of interest (set by the scope and memory in
 * parameters) to the dump file.
 *
 * @input scope The scope of the operation, can be Segment, Block or Process.
 * @input memory Flags to indicate what to dump of the pool memories.
 * @input rmd The RMD struct that holds information about the ongoing dumps.
 * @input dump Dump to write to.
 *
 * @return RmdStatus code
 *
 * @seealso RmdStatus
 *
 */
static RmdStatus
dump_stack_usage_memory(uint32_t scope, uint32_t memory, struct Rmd *rmd,
         struct Dump *dump)
{
   RmdStatus status = RMD_STATUS_OK;
   RmdFileStatus rmd_file_status = RMD_FILE_SUCCESS;
   union SIGNAL *reply = NIL;
   uint32_t signo;
   MonitorGetInfoRequest *req;
   const SIGSELECT select_reply[] =
   { 4, RMM_ATTACH_REPLY, MONITOR_GET_STACK_USAGE_REPLY,
     MONITOR_GET_STACK_USAGE_ENDMARK, MONITOR_BLOCKING_FLOW_CONTROL_REQUEST };

   if (!rmm_is_supported(MONITOR_GET_STACK_USAGE_REQUEST, rmd))
   {
      status = RMD_STATUS_INTERFACE_NOT_SUPPORTED;
      goto finally;
   }

   req = (MonitorGetInfoRequest *) alloc(sizeof(MonitorGetInfoRequest),
            MONITOR_GET_STACK_USAGE_REQUEST);

   req->level = 0;
   switch (scope)
   {
      case RMD_SCOPE_SEGMENT:
         req->scope.type = MONITOR_SCOPE_SEGMENT_ID;
         req->scope.id = dump->sid;
         break;
      case RMD_SCOPE_BLOCK:
         req->scope.type = MONITOR_SCOPE_BLOCK_ID;
         req->scope.id = dump->bid;
         break;
      case RMD_SCOPE_PROCESS:
         req->scope.type = MONITOR_SCOPE_ID;
         req->scope.id = dump->pid;
         break;
      default:
         /* Handle error */
         free_buf((union SIGNAL **) &req);
         status = RMD_STATUS_INVALID_PARAMETERS;
         goto finally;
   }
   req->reserved0 = 0;
   req->tagsCount = 0;

   /* Request information from monitor */
   send((union SIGNAL **) &req, rmd->rmm);

   do /* while (reply->signo != MONITOR_GET_STACK_USAGE_ENDMARK) */
   {
      reply = receive_from(RMD_RECEIVE_TIMEOUT, select_reply, rmd->rmm);
      assert(reply);
      signo = reply->signo;
      switch (signo)
      {
         case RMM_ATTACH_REPLY:
#ifdef RMD_VERBOSE
            ramlog_printf("RMD: RMM_ATTACH_REPLY received\n");
#endif
            status = RMD_STATUS_DETACHED_FROM_RMM;
            free_buf(&reply);
            goto finally;
         case MONITOR_GET_STACK_USAGE_REPLY:
            /* Dump full stack if requested or if the stack belongs to dumping
             * process */
            if (memory & RMD_MEMORY_STACKS_OWNED
                     || reply->monitor_get_stack_usage_reply.pid == dump->pid)
            {
               /* Dump process stack */
               if (reply->monitor_get_stack_usage_reply.main.address
                        && reply->monitor_get_stack_usage_reply.main.size)
               {
                  rmd_file_status = dump_aligned_memory(dump->bid,
                           "PROCESS STACK",
                           reply->monitor_get_stack_usage_reply.main.address,
                           reply->monitor_get_stack_usage_reply.main.size,
                           STACK_ALIGN, dump->dump_file);
               }

               /* Dump supervisor stack */
               if (reply->monitor_get_stack_usage_reply.supervisor.address
                        && reply->monitor_get_stack_usage_reply.supervisor.size)
               {
                  rmd_file_status
                           = dump_aligned_memory(dump->bid,
                                    "PROCESS SUPERVISOR STACK",
                                    reply->monitor_get_stack_usage_reply.supervisor.address,
                                    reply->monitor_get_stack_usage_reply.supervisor.size,
                                    STACK_ALIGN, dump->dump_file);
               }
            }

            else if (memory & RMD_MEMORY_STACKS_USED)
            {
               /* Dump used part of process stack */
               if (reply->monitor_get_stack_usage_reply.main.address
                        && reply->monitor_get_stack_usage_reply.main.used)
               {
                  rmd_file_status = dump_aligned_memory(dump->bid,
                           "PROCESS STACK (USED)",
                           reply->monitor_get_stack_usage_reply.main.address,
                           reply->monitor_get_stack_usage_reply.main.used,
                           STACK_ALIGN, dump->dump_file);
               }

               /* Dump used part of supervisor stack */
               if (reply->monitor_get_stack_usage_reply.supervisor.address
                        && reply->monitor_get_stack_usage_reply.supervisor.used)
               {
                  rmd_file_status
                           = dump_aligned_memory(dump->bid,
                                    "PROCESS SUPERVISOR STACK (USED)",
                                    reply->monitor_get_stack_usage_reply.supervisor.address,
                                    reply->monitor_get_stack_usage_reply.supervisor.used,
                                    STACK_ALIGN, dump->dump_file);
               }
            }

            else if (memory & RMD_MEMORY_STACKS_TOP)
            {
               /* Dump top part of process stack */
               if (reply->monitor_get_stack_usage_reply.main.address
                        && reply->monitor_get_stack_usage_reply.main.used)
               {
                  uint32_t top = rmd->rmd_properties.max_stack_top_size;
                  if (reply->monitor_get_stack_usage_reply.main.used
                           < top)
                  {
                     top = reply->monitor_get_stack_usage_reply.main.used;
                  }
                  rmd_file_status = dump_aligned_memory(dump->bid,
                           "PROCESS STACK (TOP)",
                           reply->monitor_get_stack_usage_reply.main.address,
                           top, STACK_ALIGN, dump->dump_file);
               }

               /* Dump top part of supervisor stack */
               if (reply->monitor_get_stack_usage_reply.supervisor.address
                        && reply->monitor_get_stack_usage_reply.supervisor.used)
               {
                  uint32_t top = rmd->rmd_properties.max_stack_top_size;
                  if (reply->monitor_get_stack_usage_reply.supervisor.used
                           < top)
                  {
                     top = reply->monitor_get_stack_usage_reply.supervisor.used;
                  }
                  rmd_file_status
                           = dump_aligned_memory(dump->bid,
                                    "PROCESS SUPERVISOR STACK (TOP)",
                                    reply->monitor_get_stack_usage_reply.supervisor.address,
                                    top, STACK_ALIGN, dump->dump_file);
               }
            }

            free_buf(&reply);
            if (rmd_file_status != RMD_FILE_SUCCESS)
            {
               status = rmd_file_status_converter(rmd_file_status,
                        dump->dump_file);
               goto finally;
            }
            break;
         case MONITOR_BLOCKING_FLOW_CONTROL_REQUEST:
            reply->signo = MONITOR_BLOCKING_FLOW_CONTROL_REPLY;
            send(&reply, rmd->rmm);
            break;
         case MONITOR_GET_STACK_USAGE_ENDMARK:
            break;
      }
   } while (signo != MONITOR_GET_STACK_USAGE_ENDMARK);

   free_buf(&reply);

   finally: return status;
} /* dump_stack_usage_memory() */

/**
 * @function dump_process_signal_queue_memory
 *
 * @brief This function is of type HandleProcessFunc.
 *        This function is called once for each process in the scope when calling
 *        dump_pool_memory() with memory parameter == RMD_MEMORY_SIGNALS_QUEUED.
 *
 * @returns A RmdStatus code.
 *
 */
static RmdStatus
dump_process_signal_queue_memory(struct Rmd *rmd, struct Dump *dump,
         PROCESS pid)
{
   RmdStatus status = RMD_STATUS_OK;
   RmdFileStatus rmd_file_status = RMD_FILE_SUCCESS;
   union SIGNAL *reply = NIL;
   uint32_t signo;
   MonitorGetInfoRequest *req;
   uint32_t max_signals = rmd->rmd_properties.max_signals_in_queue;
   const SIGSELECT select_reply[] =
   { 4, RMM_ATTACH_REPLY, MONITOR_GET_SIGNAL_QUEUE_REPLY,
     MONITOR_GET_SIGNAL_QUEUE_ENDMARK, MONITOR_BLOCKING_FLOW_CONTROL_REQUEST };

   req = (MonitorGetInfoRequest *) alloc(sizeof(MonitorGetInfoRequest),
            MONITOR_GET_SIGNAL_QUEUE_REQUEST);
   req->level = 0;
   req->scope.type = MONITOR_SCOPE_ID;
   req->scope.id = pid;
   req->reserved0 = 0;
   req->tagsCount = 0;

   /* Request information from monitor */
   send((union SIGNAL **) &req, rmd->rmm);

   do /* while (reply->signo != MONITOR_GET_SIGNAL_QUEUE_ENDMARK) */
   {
      reply = receive_from(RMD_RECEIVE_TIMEOUT, select_reply, rmd->rmm);
      assert(reply);
      signo = reply->signo;
      switch (signo)
      {
         case RMM_ATTACH_REPLY:
#ifdef RMD_VERBOSE
            ramlog_printf("RMD: RMM_ATTACH_REPLY received\n");
#endif
            status = RMD_STATUS_DETACHED_FROM_RMM;
            free_buf(&reply);
            goto finally;
         case MONITOR_GET_SIGNAL_QUEUE_REPLY:
            /* check for max signals in queue - rmd property */
            if(rmd->rmd_properties.max_signals_in_queue)
            {
               /* dump memory block for only max number of signals in queue */
               if(max_signals)
               {
                  if (reply->monitor_get_signal_queue_reply.signal.address !=
                              (uint32_t) NIL)
                  {
                     rmd_file_status = rmd_file_dump_memory_block(dump->bid,
        	                      reply->monitor_get_signal_queue_reply.signal.address,
        	                      reply->monitor_get_signal_queue_reply.signal.size,
        	                      "SIGNAL", dump->dump_file);
                  }
        	  max_signals--;

               }
            }
            else
            {
               if (reply->monitor_get_signal_queue_reply.signal.address !=
                              (uint32_t) NIL)
               {  
	          rmd_file_status = rmd_file_dump_memory_block(dump->bid,
				 reply->monitor_get_signal_queue_reply.signal.address,
				 reply->monitor_get_signal_queue_reply.signal.size,
				 "SIGNAL", dump->dump_file);
               }
            }

            free_buf(&reply);
            if (rmd_file_status != RMD_FILE_SUCCESS)
            {
               status = RMD_STATUS_FILE_ERROR;
               goto finally;
            }
            break;
         case MONITOR_BLOCKING_FLOW_CONTROL_REQUEST:
            reply->signo = MONITOR_BLOCKING_FLOW_CONTROL_REPLY;
            send(&reply, rmd->rmm);
            break;
         case MONITOR_GET_SIGNAL_QUEUE_ENDMARK:
            break;
      }
   } while (signo != MONITOR_GET_SIGNAL_QUEUE_ENDMARK);

   free_buf(&reply);

   finally: return status;
}

/**
 * @function rmd_private_dump_pool_memory
 *
 * @brief Calls either dump_pool_memory or dump_stack_usage_memory to write
 * the pool memory of interest to the dump file.
 *
 * @input rmd_req Request signal of type RmdDumpPoolMemoryRequest
 *
 * @seealso RmdDumpPoolMemoryRequest
 * @seealso dump_pool_memory
 * @seealso dump_stack_usage_memory
 */
void
rmd_private_dump_pool_memory(union SIGNAL *rmd_req, struct Rmd *rmd)
{
   RmdStatus status = RMD_STATUS_OK;
   union SIGNAL *sig;
   struct Dump *dump;
   uint32_t memory;
   uint32_t scope;
   OSBOOLEAN supported = TRUE;

   assert(rmd_req);
   assert(RMD_DUMP_STRING_REQUEST == rmd_req->signo);
   assert(rmd);
   assert(RMD_MAGIC == rmd->magic);
#ifdef RMD_VERBOSE
   ramlog_printf("RMD: rmd_private_dump_pool_memory()\n");
#endif

   dump = get_dump_from_pid(sender(&rmd_req), &rmd->dumps);
   if (NULL == dump)
   {
      status = RMD_STATUS_DUMP_NOT_FOUND;
      goto finally;
   }

   /* Check if the interface signals are supported in the RMM */
   /* The following signals are sent from this function
    * MONITOR_GET_POOL_INFO_REQUEST
    * MONITOR_GET_SIGNAL_QUEUE_REQUEST (in sub-function)
    * MONITOR_GET_STACK_USAGE_REQUEST
    */
   supported &= rmm_is_supported(MONITOR_GET_POOL_INFO_REQUEST, rmd);
   supported &= rmm_is_supported(MONITOR_GET_SIGNAL_QUEUE_REQUEST, rmd);
   supported &= rmm_is_supported(MONITOR_GET_STACK_USAGE_REQUEST, rmd);
   if (!supported)
   {
      status = RMD_STATUS_INTERFACE_NOT_SUPPORTED;
      goto finally;
   }

   memory = rmd_req->rmd_dump_pool_memory_request.memory;
   scope = rmd_req->rmd_dump_pool_info_request.scope;
   if (rmd_req->rmd_dump_pool_memory_request.memory & RMD_MEMORY_POOL)
   {
      /* Dump full pool memory */
      status = dump_pool_memory(scope, RMD_MEMORY_POOL, rmd, dump);
      if (RMD_STATUS_OK != status)
      {
         goto finally;
      }
   }
   else
   {
      if (rmd_req->rmd_dump_pool_memory_request.memory & RMD_MEMORY_STACKS)
      {
         /* Dump full stack memory */
         status = dump_pool_memory(scope, RMD_MEMORY_STACKS, rmd, dump);
         if (RMD_STATUS_OK != status)
         {
            goto finally;
         }
      }
      else if ((memory & RMD_MEMORY_STACKS_OWNED) ||
               (memory & RMD_MEMORY_STACKS_USED) ||
               (memory & RMD_MEMORY_STACKS_TOP))
      {
         /* Dump part of stack memory */
         status = dump_stack_usage_memory(scope, memory, rmd, dump);
         if (RMD_STATUS_OK != status)
         {
            goto finally;
         }
      }

      if (rmd_req->rmd_dump_pool_memory_request.memory & RMD_MEMORY_SIGNALS)
      {
         /* Dump full signal memory */
         status = dump_pool_memory(scope, RMD_MEMORY_SIGNALS, rmd, dump);
         if (RMD_STATUS_OK != status)
         {
            goto finally;
         }
      }
      else
      {
         if (memory & RMD_MEMORY_SIGNALS_OWNED)
         {
            /* Dump part of signal memory */
            status = dump_pool_memory(scope, RMD_MEMORY_SIGNALS_OWNED, rmd,
                     dump);
            /* status = dump_pool_signals_memory(scope, memory, rmd, dump); */
            if (RMD_STATUS_OK != status)
            {
               goto finally;
            }
         }
         if (memory & RMD_MEMORY_SIGNALS_SENT)
         {
            /* Dump part of signal memory */
            status = dump_pool_memory(scope, RMD_MEMORY_SIGNALS_SENT, rmd,
                              dump);
            /* status = dump_pool_signals_memory(scope, memory, rmd, dump); */
            if (RMD_STATUS_OK != status)
            {
               goto finally;
            }
         }
         if (memory & RMD_MEMORY_SIGNALS_QUEUED)
         {
            /* check for number of long queue processes - rmd property */
            if(rmd->rmd_properties.num_long_queue_procs)
            {
	       struct ProcQueueCnt *proc_queue_list;
	       uint32_t i;
	       uint32_t procs_count = 0;
	       MonitorGetInfoRequest *req1;
	       
	       if (!rmm_is_supported(MONITOR_GET_PROCESS_INFO_REQUEST, rmd))
	       {
	          status = RMD_STATUS_INTERFACE_NOT_SUPPORTED;
		  goto finally;
	       }

	       req1 = (MonitorGetInfoRequest *) alloc(
            	 sizeof(MonitorGetInfoRequest), MONITOR_GET_PROCESS_INFO_REQUEST);
	       /* set request level = process long info */
	       req1->level = 1;
	       /* set number of tags to zero */
	       req1->tagsCount = 0;

	       if ((req1->scope.type =
                    rmd_req->rmd_dump_signal_queue_info_request.scope) ==
                    RMD_SCOPE_SEGMENT)
	       {
                  req1->scope.id = dump->sid;
	       }	
               else if (req1->scope.type == RMD_SCOPE_BLOCK)
               {
	          req1->scope.id = dump->bid;
               }
	       else if (req1->scope.id == RMD_SCOPE_PROCESS)
	       {
		  req1->scope.id = dump->pid;
               }
               else
	       {
		  /* Handle error */
		  free_buf((union SIGNAL **) &req1);
		  status = RMD_STATUS_INVALID_PARAMETERS;
		  goto finally;
	       }
	       /* create pool buffer to hold pid and signal queue count*/
	       proc_queue_list = (struct ProcQueueCnt *) alloc(
                           sizeof(struct ProcQueueCnt) *
                           (rmd->rmd_properties.num_long_queue_procs + 1), 0);
	       /* get process id and signal queue count info of all the processes
                * in the scope
                */
	       status = get_process_long_info(
                        (union SIGNAL*)req1,rmd,proc_queue_list,&procs_count);

	       if (status != RMD_STATUS_OK)
	       {
		  /* free the proc_queue_list buffer */
		  free_buf((union SIGNAL **)&proc_queue_list);
		  goto finally;
	       }


	       /* dump process signal queue info */
	       for (i=0; i<((rmd->rmd_properties.num_long_queue_procs > procs_count)?
                     procs_count:rmd->rmd_properties.num_long_queue_procs);i++)
	       {
	          /* Request signal queue memory for each process in the sorted list */
		  status = dump_process_signal_queue_memory(rmd,dump,((proc_queue_list+i)->pid));
		  if (status != RMD_STATUS_OK)
		  {
		     /* free the proc_queue_list buffer */
		     free_buf((union SIGNAL **)&proc_queue_list);
		     goto finally;
		  }
	       }
	       /* free the proc_queue_list buffer */
	       free_buf((union SIGNAL **)&proc_queue_list);

            }
            else
            {
	       /* Dump part of signal memory */
	       /* Request signal queue memory for each process in our scope */
	       status = for_each_processes_in_scope(scope, rmd, dump,
					 dump_process_signal_queue_memory);
            }
            if (RMD_STATUS_OK != status)
            {
               goto finally;
            }
         }
      }
   }

   finally:
   /* Send reply signal */
   sig = alloc(sizeof(RmdStatusReply), RMD_DUMP_POOL_MEMORY_REPLY);
   sig->rmd_reply.status = status;
   send(&sig, sender(&rmd_req));
} /* rmd_private_dump_pool_memory() */

/**
 * @function dump_memory_heap
 *
 * @brief Request information about the heap from the RMM and writes the reply
 * to the dump file.
 *
 * @input heap_id Heap id to be sent to the RMM
 * @input rmd The rmd struct caontaining information about the dumps in this
 *        process
 * @input dump Current dump containing dump_file.
 *
 * @return RmdStatus code
 */
static RmdStatus
dump_memory_heap(PROCESS heap_id, struct Rmd *rmd, struct Dump *dump)
{
   RmdStatus status = RMD_STATUS_OK;
   RmdFileStatus rmd_file_status = RMD_FILE_SUCCESS;

   union SIGNAL *reply = NIL;
   uint32_t iterator = 0;
   uint32_t signo;
   MonitorGetInfoRequest *req;
   const SIGSELECT select_reply[] =
   { 4, RMM_ATTACH_REPLY, MONITOR_GET_HEAP_FRAGMENT_INFO_REPLY,
     MONITOR_GET_HEAP_FRAGMENT_INFO_ENDMARK,
     MONITOR_BLOCKING_FLOW_CONTROL_REQUEST };

   if (!rmm_is_supported(MONITOR_GET_HEAP_FRAGMENT_INFO_REQUEST, rmd))
   {
      status = RMD_STATUS_INTERFACE_NOT_SUPPORTED;
      goto finally;
   }

   req = (MonitorGetInfoRequest *) alloc(sizeof(MonitorGetInfoRequest),
            MONITOR_GET_HEAP_FRAGMENT_INFO_REQUEST);

   req->level = 0;
   req->scope.type = MONITOR_SCOPE_ID;
   req->scope.id = heap_id;
   req->reserved0 = 0;
   req->tagsCount = 0;

   /* Request information from monitor */
   send((union SIGNAL **) &req, rmd->rmm);

   do /* while (reply->signo != MONITOR_GET_HEAP_FRAGMENT_INFO_ENDMARK) */
   {
      reply = receive_from(RMD_RECEIVE_TIMEOUT, select_reply, rmd->rmm);
      assert(reply);
      signo = reply->signo;
      switch (signo)
      {
         case RMM_ATTACH_REPLY:
#ifdef RMD_VERBOSE
            ramlog_printf("RMD: RMM_ATTACH_REPLY received\n");
#endif
            status = RMD_STATUS_DETACHED_FROM_RMM;
            free_buf(&reply);
            goto finally;
         case MONITOR_GET_HEAP_FRAGMENT_INFO_REPLY:
            /* Write memory fragments to file dump */
            for (iterator = 0; iterator
                     < reply->monitor_get_heap_fragment_info_reply.fragmentsCount; iterator++)
            {
               rmd_file_status
                        = rmd_file_dump_memory_block(dump->bid,
                                 reply->monitor_get_heap_fragment_info_reply.fragments[iterator].address,
                                 reply->monitor_get_heap_fragment_info_reply.fragments[iterator].size,
                                 "HEAP FRAGMENT", dump->dump_file);
               if (RMD_FILE_SUCCESS != rmd_file_status)
               {
                  free_buf(&reply);
                  status = rmd_file_status_converter(rmd_file_status,
                           dump->dump_file);
                  goto finally;
               }
            }
            free_buf(&reply);
            break;
         case MONITOR_BLOCKING_FLOW_CONTROL_REQUEST:
            reply->signo = MONITOR_BLOCKING_FLOW_CONTROL_REPLY;
            send(&reply, rmd->rmm);
            break;
         case MONITOR_GET_HEAP_FRAGMENT_INFO_ENDMARK:
            if (reply->monitor_get_info_endmark.status != MONITOR_STATUS_OK)
            {
               ramlog_printf("RMD: Received invalid status %u in MONITOR_GET_HEAP_FRAGMENT_INFO_ENDMARK\n",
                             reply->monitor_get_info_endmark.status);
               free_buf(&reply);
               status = RMD_STATUS_UNABLE_TO_RETRIEVE_DATA;
               goto finally;
            } 
            break;
      }
   } while (signo != MONITOR_GET_HEAP_FRAGMENT_INFO_ENDMARK);
   free_buf(&reply);

   finally: return status;
} /* dump_memory_heap() */

/**
 * @function dump_memory_heap_buffers
 *
 * @ brief Get memory heap buffers associated with heap_id. The buffers will be
 * filtered to scope. Scope == 0 means no filtering.
 *
 * @input heap_id Heap id to be sent to the RMM.
 * @input rmd The rmd struct caontaining information about the dumps in this
 *        process.
 * @input dump Current dump containing dump_file.
 * @input scope Scope of the call.
 *
 * @return RmdStatus code
 */
static RmdStatus
dump_memory_heap_buffers(PROCESS heap_id, struct Rmd *rmd, struct Dump *dump,
         uint32_t scope)
{
   RmdStatus status = RMD_STATUS_OK;
   RmdFileStatus rmd_file_status = RMD_FILE_SUCCESS;

   union SIGNAL *reply = NIL;
   uint32_t iterator = 0;
   uint32_t signo;
   MonitorGetInfoRequest *req;
   const SIGSELECT select_reply[] =
   { 4, RMM_ATTACH_REPLY, MONITOR_GET_HEAP_BUFFERS_REPLY,
     MONITOR_GET_HEAP_BUFFERS_ENDMARK, MONITOR_BLOCKING_FLOW_CONTROL_REQUEST };

   if (!rmm_is_supported(MONITOR_GET_HEAP_BUFFERS_REQUEST, rmd))
   {
      status = RMD_STATUS_INTERFACE_NOT_SUPPORTED;
      goto finally;
   }

   req = (MonitorGetInfoRequest *) alloc(sizeof(MonitorGetInfoRequest),
            MONITOR_GET_HEAP_BUFFERS_REQUEST);

   req->level = 0;
   req->scope.type = MONITOR_SCOPE_ID;
   req->scope.id = heap_id;
   req->reserved0 = 0;
   req->tagsCount = 0;

   /* Request information from monitor */
   send((union SIGNAL **) &req, rmd->rmm);

   do /* while (reply->signo != MONITOR_GET_HEAP_BUFFERS_ENDMARK) */
   {
      reply = receive_from(RMD_RECEIVE_TIMEOUT, select_reply, rmd->rmm);
      assert(reply);
      signo = reply->signo;
      switch (signo)
      {
         case RMM_ATTACH_REPLY:
#ifdef RMD_VERBOSE
            ramlog_printf("RMD: RMM_ATTACH_REPLY received\n");
#endif
            status = RMD_STATUS_DETACHED_FROM_RMM;
            free_buf(&reply);
            goto finally;
         case MONITOR_GET_HEAP_BUFFERS_REPLY:
            /* Write memory buffers to file dump */
            for (iterator = 0; iterator
                     < reply->monitor_get_heap_buffers_reply.buffersCount; iterator++)
            {
               if (pid_belongs_to_scope(
                        reply->monitor_get_heap_buffers_reply.buffers[iterator].info.owner,
                        dump, scope))
               {
                  rmd_file_status
                           = rmd_file_dump_memory_block(dump->bid,
                                    reply->monitor_get_heap_buffers_reply.buffers[iterator].info.address,
                                    reply->monitor_get_heap_buffers_reply.buffers[iterator].info.size,
                                    "HEAP BUFFER", dump->dump_file);
                  if (RMD_FILE_SUCCESS != rmd_file_status)
                  {
                     free_buf(&reply);
                     status = rmd_file_status_converter(rmd_file_status,
                              dump->dump_file);
                     goto finally;
                  }
               }
            }

            free_buf(&reply);
            break;
         case MONITOR_BLOCKING_FLOW_CONTROL_REQUEST:
            reply->signo = MONITOR_BLOCKING_FLOW_CONTROL_REPLY;
            send(&reply, rmd->rmm);
            break;
         case MONITOR_GET_HEAP_BUFFERS_ENDMARK:
            if (reply->monitor_get_info_endmark.status != MONITOR_STATUS_OK)
            {
               ramlog_printf("RMD: Received invalid status %u in MONITOR_GET_HEAP_BUFFERS_ENDMARK\n",
                             reply->monitor_get_info_endmark.status);
               free_buf(&reply);
               status = RMD_STATUS_UNABLE_TO_RETRIEVE_DATA;
               goto finally;
            }
            break;
      }
   } while (signo != MONITOR_GET_HEAP_BUFFERS_ENDMARK);
   free_buf(&reply);

   finally: return status;
} /* dump_memory_heap_buffers() */

/**
 * @function rmd_private_dump_heap_memory
 *
 * @brief Requests information about the heap from the RMM and writes the
 * results to the dump file by calling dump_memory_heap_buffers or
 * dump_memory_heap.
 *
 * @input rmd_req Request signal of type RmdDumpHeapMemoryRequest.
 *
 * @seealso RmdDumpHeapMemoryRequest
 * @seealso dump_memory_heap
 * @seealso dump_memory_heap_buffers
 *
 */
void
rmd_private_dump_heap_memory(union SIGNAL *rmd_req, struct Rmd *rmd)
{
   RmdStatus status = RMD_STATUS_OK;
   PROCESS heap_id = 0;
   union SIGNAL *sig = NIL;
   struct Dump *dump = NULL;
   uint32_t scope;
   uint32_t memory;
   uint32_t signo;
   MonitorGetInfoRequest *req;
   OSBOOLEAN supported = TRUE;
   const SIGSELECT select_reply[] =
   { 4, RMM_ATTACH_REPLY, MONITOR_GET_HEAP_INFO_REPLY,
     MONITOR_GET_HEAP_INFO_ENDMARK, MONITOR_BLOCKING_FLOW_CONTROL_REQUEST };

   assert(rmd_req);
   assert(RMD_DUMP_HEAP_MEMORY_REQUEST == rmd_req->signo);
   assert(rmd);
   assert(RMD_MAGIC == rmd->magic);

   dump = get_dump_from_pid(sender(&rmd_req), &rmd->dumps);
   if (NULL == dump)
   {
      status = RMD_STATUS_DUMP_NOT_FOUND;
      goto finally;
   }

   /* Check if the interface signals are supported in the RMM */
   /* The following signals are sent from this function
    * MONITOR_GET_HEAP_INFO_REQUEST (in sub-function get_heap_id())
    * MONITOR_GET_HEAP_FRAGMENT_INFO_REQUEST
    * MONITOR_GET_HEAP_BUFFERS_REQUEST
    */
   supported &= rmm_is_supported(MONITOR_GET_HEAP_INFO_REQUEST, rmd);
   supported &= rmm_is_supported(MONITOR_GET_HEAP_FRAGMENT_INFO_REQUEST, rmd);
   supported &= rmm_is_supported(MONITOR_GET_HEAP_BUFFERS_REQUEST, rmd);

   if (!supported)
   {
      status = RMD_STATUS_INTERFACE_NOT_SUPPORTED;
      goto finally;
   }

   memory = rmd_req->rmd_dump_heap_memory_request.memory;
   /* Skip if there is no heap memory requested */
   if (!(memory & (RMD_MEMORY_HEAP | RMD_MEMORY_HEAP_BUFFERS |
                   RMD_MEMORY_HEAP_BUFFERS_OWNED) ))
   {
      goto finally;
   }

   /* Request heap information from monitor */
   req = (MonitorGetInfoRequest *) alloc(sizeof(MonitorGetInfoRequest),
            MONITOR_GET_HEAP_INFO_REQUEST);

#ifdef RMD_VERBOSE
   ramlog_printf("RMD: rmd_private_dump_heap_memory()\n");
#endif

   scope = rmd_req->rmd_dump_heap_memory_request.scope;

   req->level = 0;
   switch (scope)
   {
      case RMD_SCOPE_SEGMENT:
      case RMD_SCOPE_BLOCK:
      case RMD_SCOPE_PROCESS:
         req->scope.type = MONITOR_SCOPE_SEGMENT_ID;
         req->scope.id = dump->sid;
         break;
      default:
         /* Handle error */
         free_buf((union SIGNAL **) &req);
         status = RMD_STATUS_INVALID_PARAMETERS;
         goto finally;
   }
   req->reserved0 = 0;
   req->tagsCount = 0;

   send((union SIGNAL **) &req, rmd->rmm);

   /* Iterate over all heaps in segment */
   do /* while (sig->signo != MONITOR_GET_HEAP_INFO_ENDMARK) */
   {
      sig = receive_from(RMD_RECEIVE_TIMEOUT, select_reply, rmd->rmm);
      assert(sig);
      signo = sig->signo;
      switch (signo)
      {
         case RMM_ATTACH_REPLY:
#ifdef RMD_VERBOSE
            ramlog_printf("RMD: RMM_ATTACH_REPLY received\n");
#endif
            status = RMD_STATUS_DETACHED_FROM_RMM;
            free_buf((union SIGNAL **) &req);
            goto finally;
         case MONITOR_GET_HEAP_INFO_REPLY:
            heap_id = sig->monitor_get_heap_info_reply.id;
            free_buf(&sig);

            if (memory & RMD_MEMORY_HEAP)
            {
               status = dump_memory_heap(heap_id, rmd, dump);
            }
            else if (memory & RMD_MEMORY_HEAP_BUFFERS)
            {
               status = dump_memory_heap_buffers(heap_id, rmd, dump, 0);
            }
            else if (memory & RMD_MEMORY_HEAP_BUFFERS_OWNED)
            {
               status = dump_memory_heap_buffers(heap_id, rmd, dump, scope);
            }
            if (RMD_STATUS_OK != status)
            {
               goto finally;
            }
            break;
         case MONITOR_BLOCKING_FLOW_CONTROL_REQUEST:
            sig->signo = MONITOR_BLOCKING_FLOW_CONTROL_REPLY;
            send(&sig, rmd->rmm);
            break;
         case MONITOR_GET_HEAP_INFO_ENDMARK:
            if (sig->monitor_get_info_endmark.status != MONITOR_STATUS_OK)
            {
               ramlog_printf("RMD: Received invalid status %u in MONITOR_GET_HEAP_INFO_ENDMARK\n",
                             sig->monitor_get_info_endmark.status);
               free_buf(&sig);
               status = RMD_STATUS_UNABLE_TO_RETRIEVE_DATA;
               goto finally;
            }
            break;
      }
   } while (signo != MONITOR_GET_HEAP_INFO_ENDMARK);
   free_buf(&sig);

   finally:
   /* Send reply signal */
   sig = alloc(sizeof(RmdStatusReply), RMD_DUMP_HEAP_MEMORY_REPLY);
   sig->rmd_reply.status = status;
   send(&sig, sender(&rmd_req));
} /* rmd_private_dump_heap_memory() */

/**
 * @function rmd_private_dump_program_memory
 *
 * @brief Request information about the program memory from the PM and writes
 * the memory to the dump file.
 *
 * @input rmd_req Request signal of type RmdDumpProgramMemoryRequest.
 *
 * @seealso RmdDumpProgramMemoryRequest
 */
void
rmd_private_dump_program_memory(union SIGNAL *req, struct Rmd *rmd)
{
   union SIGNAL *sig;
   union SIGNAL *info_sig;
   struct Dump *dump;
   RmdFileStatus rmd_file_status = RMD_FILE_SUCCESS;
   RmdStatus return_status = RMD_STATUS_OK;

   assert(req);
   assert(RMD_DUMP_PROGRAM_MEMORY_REQUEST == req->signo);
   assert(rmd);
   assert(RMD_MAGIC == rmd->magic);
#ifdef RMD_VERBOSE
   ramlog_printf("RMD: rmd_private_dump_program_memory()\n");
#endif

   dump = get_dump_from_pid(sender(&req), &rmd->dumps);
   if (NULL == dump)
   {
#ifdef RMD_VERBOSE
      ramlog_printf("RMD: Couldn't find a dump from the pid 0x%x\n", sender(
                        &req));
#endif
      return_status = RMD_STATUS_DUMP_NOT_FOUND;
      goto finally;
   }

   if (req->rmd_dump_program_memory_request.memory & RMD_MEMORY_DATA)
   {
      info_sig = get_pm_info_request(dump->bid);
      if (NIL == info_sig)
      {
         return_status = RMD_STATUS_INTERNAL_ERROR;
         goto finally;
      }

      sig = get_pm_load_module_info_signal(
               info_sig->program_info.install_handle);
      free_buf(&info_sig);
      if (NIL == sig)
      {
         return_status = RMD_STATUS_INTERNAL_ERROR;
         goto finally;
      }

      /* Now we have the information about the load module,
       * lets dump the requested info */
      rmd_file_status = rmd_file_dump_memory_block(dump->bid,
               sig->load_module_info_reply.data_base,
               sig->load_module_info_reply.data_size, "DATA", dump->dump_file);
      free_buf(&sig);

      return_status = rmd_file_status_converter(rmd_file_status,
               dump->dump_file);
   }

   if (req->rmd_dump_program_memory_request.memory & RMD_MEMORY_TEXT)
   {
      info_sig = get_pm_info_request(dump->bid);
      if (NIL == info_sig)
      {
         return_status = RMD_STATUS_INTERNAL_ERROR;
         goto finally;
      }

      sig = get_pm_load_module_info_signal(
               info_sig->program_info.install_handle);
      free_buf(&info_sig);
      if (NIL == sig)
      {
         return_status = RMD_STATUS_INTERNAL_ERROR;
         goto finally;
      }

      /* Now we have the information about the load module,
       * lets dump the requested info */
      rmd_file_status = rmd_file_dump_memory_block(dump->bid,
               sig->load_module_info_reply.text_base,
               sig->load_module_info_reply.text_size, "TEXT", dump->dump_file);
      free_buf(&sig);

      return_status = rmd_file_status_converter(rmd_file_status,
               dump->dump_file);
   }

   finally:
   /* Send reply signal */
   sig = alloc(sizeof(RmdStatusReply), RMD_DUMP_PROGRAM_MEMORY_REPLY);
   sig->rmd_reply.status = return_status;
   send(&sig, sender(&req));
}

/**
 * @function rmd_private_get_dump_info
 *
 * @brief Request information about the current dump.
 *
 * @input rmd_req Request signal of type RmdGetDumpInfoRequest.
 *
 * @seealso RmdGetDumpInfoRequest
 */
static void
rmd_private_get_dump_info(union SIGNAL *req, struct Rmd *rmd)
{
   union SIGNAL *reply = NULL;
   struct Dump *dump;
   RmdStatus return_status = RMD_STATUS_OK;

   assert(req);
   assert(RMD_GET_DUMP_INFO_REQUEST == req->signo);
   assert(rmd);
   assert(RMD_MAGIC == rmd->magic);
#ifdef RMD_VERBOSE
   ramlog_printf("RMD: rmd_private_get_dump_info()\n");
#endif

   dump = get_dump_from_pid(sender(&req), &rmd->dumps);
   if (NULL == dump)
   {
#ifdef RMD_VERBOSE
      ramlog_printf("RMD: Couldn't find a dump from the pid 0x%x\n", sender(
                        &req));
#endif
      return_status = RMD_STATUS_DUMP_NOT_FOUND;
      goto finally;
   }

   switch (req->rmd_get_dump_info_request.tag_value)
   {
      case RMD_DUMP_ID_TAG:
         if (dump->dump_file)
         {
            reply = alloc(sizeof(RmdGetDumpInfoReply) + 3,
                  RMD_GET_DUMP_INFO_REPLY);
            reply->rmd_get_dump_info_reply.length = sizeof(uint32_t);
            *((uint32_t *)(reply->rmd_get_dump_info_reply.data)) = dump->dump_file->dump_id;
            return_status = RMD_STATUS_OK;
         }
         else
         {
            return_status = RMD_STATUS_DUMP_NOT_FOUND;
         }
         break;
         
      case RMD_PATH_NAME_TAG:
         if (dump->dump_file && strlen(dump->dump_file->file_name) > 0)
         {
            reply = alloc(sizeof(RmdGetDumpInfoReply) +
                  strlen(dump->dump_file->file_name),
                  RMD_GET_DUMP_INFO_REPLY);
            reply->rmd_get_dump_info_reply.length = strlen(dump->dump_file->file_name) + 1;
            strcpy(reply->rmd_get_dump_info_reply.data,
                   dump->dump_file->file_name);
            return_status = RMD_STATUS_OK;
         }
         else
         {
            return_status = RMD_STATUS_DUMP_NOT_FOUND;
         }
         break;
         
      default:
         return_status = RMD_STATUS_UNKNOWN_INFO_TAG;
         break;
   }
   
   finally:
   if (!reply)
   {
      reply = alloc(sizeof(RmdGetDumpInfoReply),
            RMD_GET_DUMP_INFO_REPLY);
      reply->rmd_get_dump_info_reply.length = 0;
   }
   reply->rmd_get_dump_info_reply.status = return_status;
   reply->rmd_get_dump_info_reply.reserved0 = 0;
   reply->rmd_get_dump_info_reply.reserved1 = 0;
   
   /* Send reply signal */
   send(&reply, sender(&req));
}

#include "ramdump.h"

/**
 * @function init_rmd()
 *
 * @brief Code for initialisation of the Rmd struct.
 *
 * @output rmd The Rmd struct is filled with default values from the
 * system parameters
 *
 */
static void
init_rmd(struct Rmd *rmd)
{
   char buf[64];
   long param_value;

   rmd->magic = ~RMD_MAGIC;
   list_init(&rmd->dumps);
   rmd->rmm = 0;
   rmd->magic = RMD_MAGIC;
   rmd->rmm_connected = FALSE;

   if (sysparam_read("krn/rmd/compression", buf, sizeof(buf) - 1))
   {
      rmd->use_compression = TRUE;
   }
   else
   {
      rmd->use_compression = FALSE;
   }

   if (!sysparam_readattr_long("krn/rmd/compression", "window_bits",
            &param_value))
   {
      rmd->window_bits = param_value;
   }
   else
   {
      rmd->window_bits = 15;
   }

   if (!sysparam_readattr_long("krn/rmd/compression", "mem_level", &param_value))
   {
      rmd->mem_level = param_value;
   }
   else
   {
      rmd->mem_level = 8;
   }

   if (sysparam_read("krn/rmd/path", buf, sizeof buf - 1))
   {
      strncpy(rmd->path, buf, sizeof buf);
   }
   else
   {
      /* Default path if not configured */
      strncpy(rmd->path, DEFAULT_DUMP_PATH, sizeof rmd->path);
   }

   if (sysparam_read("krn/rmd/max_number_of_dumps", buf, sizeof buf - 1))
   {
      rmd->max_ongoing_dumps = atoi(buf);
   }
   else
   {
      rmd->max_ongoing_dumps = 10;
   }

   /* Register event triggers */
   rmd->start_dump_event_trigger = ~0;
   rmd->stop_dump_event_trigger = ~0;
   create_event(RMD_EVENT_START_DUMP, &rmd->start_dump_event_trigger);
   create_event(RMD_EVENT_STOP_DUMP, &rmd->stop_dump_event_trigger);

   /* Reset shared information */
   rmd->stored_connect_reply = NIL;
   rmd->supported_monitor_signals = NULL;
   rmd->supported_monitor_signals_count = 0;

   /* Set Rmd Properties to default values */
   rmd->rmd_properties.num_long_queue_procs = 0;
   rmd->rmd_properties.max_signals_in_queue = 0;
   rmd->rmd_properties.max_stack_top_size = 200;
 
   if (ose_cpu_id() == 0)
   {
      ramlog_printf("RMD: use_compression : %d and window_bits : %d\n", 
                     rmd->use_compression, rmd->window_bits);
      ramlog_printf("RMD: mem_level : %d, max_ongoing_dumps : %d and path : %s\n", 
                     rmd->mem_level, rmd->max_ongoing_dumps, rmd->path);
   }
}

void
ose_rmd(void);

/**
 * @function ose_rmd
 *
 * @brief The main process function where the signalling is handled.
 *
 * @long This is where all the requests are made to the RMD system. The ose_rmd
 * calls the rmd_private_* interface implemented above with the request signal
 * and Rmd struct as arguments. The process starts by calling the initialisation
 * of the Rmd struct.
 *
 * @long This process is supposed to be spawned once for each core in the
 * system. The idea of the design is that everything is kept locally if
 * possible. This is stored in the rmd struct. The only information that is
 * shared between the processes is the number of ongoing dumps (ongoing_dumps).
 *
 * @long The startup of the process is that it hunts for the RMM and only
 * listens for the hunt reply signal when doing this. Until the RMM is found
 * and attached to, there is no way to communicate with the RMD process. This
 * is also the case when the RMM goes down and the RMD receives the signal
 * RMM_ARRACH_REPLY.
 *
 */
OS_PROCESS(ose_rmd)
{
   struct Rmd rmd;
   union SIGNAL *sig;
   struct Dump *dump;

   init_rmd(&rmd);

   sig = alloc(sizeof(SIGSELECT), RMM_ATTACH_REPLY);

   for (;;)
   {
      static const SIGSELECT all_sigs[] =
      { 0 };
      static const SIGSELECT rmm_hunt_sigs[] =
      { 1, RMM_HUNT_REPLY };

      const SIGSELECT *sigsel = all_sigs;

      switch (sig->signo)
      {
         /**** Handle administrative signals below ****/

         case RMM_ATTACH_REPLY:
            sig->signo = RMM_HUNT_REPLY;
            rmd.rmm_main = 0;
            rmd.rmm = 0;
            hunt("ose_monitor", 0, 0, &sig);
            /* No monitor yet so block request signals. */
            sigsel = rmm_hunt_sigs;
            break;

         case RMM_HUNT_REPLY:
            sig->signo = RMM_ATTACH_REPLY;
            rmd.rmm_main = sender(&sig);
            rmd.attach_ref = attach(&sig, rmd.rmm_main);
            /* We can accept request signals now. */
            sigsel = all_sigs;
            break;

         case MONITOR_BLOCKING_FLOW_CONTROL_REQUEST:
            /* We can end up here if something went wrong, such as full disk,
             * in a request to RMM. The flow control must be handled or the
             *  RMM will not reply on further requests. */
            sig->signo = MONITOR_BLOCKING_FLOW_CONTROL_REPLY;
            send(&sig, rmd.rmm);
            break;

         case RMD_CLIENT_ATTACH_SIG:
            ramlog_printf("RMD: RMD_CLIENT_ATTACH_SIG received from 0x%08lx\n",
                     sig->rmd_client_attach_sig.client_pid);
            rmd_private_stop_dump(sig, &rmd);
            free_buf(&sig);
            break;

            /**** Handle RMD request signals below ****/

         case RMD_INTERFACE_REQUEST:
            rmd_private_interface_request(sender(&sig));
            free_buf(&sig);
            break;

         case RMD_START_DUMP_REQUEST:
            rmd_private_start_dump(sig, &rmd);
            free_buf(&sig);
            break;

         case RMD_STOP_DUMP_REQUEST:
            rmd_private_stop_dump(sig, &rmd);
            free_buf(&sig);
            break;

         case RMD_DUMP_ERROR_REQUEST:
            rmd_private_dump_error(sig, &rmd);
            free_buf(&sig);
            break;

         case RMD_DUMP_MEMORY_REQUEST:
            rmd_private_dump_memory(sig, &rmd);
            free_buf(&sig);
            break;

         case RMD_DUMP_STRING_REQUEST:
            rmd_private_dump_string(sig, &rmd);
            free_buf(&sig);
            break;

         case RMD_DUMP_PROCESS_INFO_REQUEST:
            rmd_private_dump_process_info(sig, &rmd);
            free_buf(&sig);
            break;

         case RMD_DUMP_STACK_USAGE_INFO_REQUEST:
            rmd_private_dump_stack_usage_info(sig, &rmd);
            free_buf(&sig);
            break;

         case RMD_DUMP_ENV_VARS_INFO_REQUEST:
            rmd_private_dump_env_vars_info(sig, &rmd);
            free_buf(&sig);
            break;

         case RMD_DUMP_SIGNAL_QUEUE_INFO_REQUEST:
            rmd_private_dump_signal_queue_info(sig, &rmd);
            free_buf(&sig);
            break;

         case RMD_DUMP_POOL_INFO_REQUEST:
            rmd_private_dump_pool_info(sig, &rmd);
            free_buf(&sig);
            break;

         case RMD_DUMP_POOL_SIGNALS_INFO_REQUEST:
            rmd_private_dump_pool_signals_info(sig, &rmd);
            free_buf(&sig);
            break;

         case RMD_DUMP_SYS_PARAM_INFO_REQUEST:
            rmd_private_dump_sys_param_info(sig, &rmd);
            free_buf(&sig);
            break;

         case RMD_DUMP_REGISTERS_INFO_REQUEST:
            rmd_private_dump_registers_info(sig, &rmd);
            free_buf(&sig);
            break;

         case RMD_DUMP_PROGRAM_INFO_REQUEST:
            rmd_private_dump_program_info(sig, &rmd);
            free_buf(&sig);
            break;

         case RMD_DUMP_TRACE_INFO_REQUEST:
            rmd_private_dump_trace_info(sig, &rmd);
            free_buf(&sig);
            break;

         case RMD_DUMP_LOAD_MODULE_INFO_REQUEST:
            rmd_private_dump_load_module_info(sig, &rmd);
            free_buf(&sig);
            break;

         case RMD_DUMP_POOL_MEMORY_REQUEST:
            rmd_private_dump_pool_memory(sig, &rmd);
            free_buf(&sig);
            break;

         case RMD_DUMP_HEAP_MEMORY_REQUEST:
            rmd_private_dump_heap_memory(sig, &rmd);
            free_buf(&sig);
            break;

         case RMD_DUMP_PROGRAM_MEMORY_REQUEST:
            rmd_private_dump_program_memory(sig, &rmd);
            free_buf(&sig);
            break;

         case RMD_SET_PROPERTY_REQUEST:
            if (sig->rmd_set_property_request.property ==
                 RMD_NUM_LONG_QUEUE_PROCS)
            { 
       	       rmd.rmd_properties.num_long_queue_procs =
                  sig->rmd_set_property_request.value;
            }
            else if (sig->rmd_set_property_request.property ==
                     RMD_MAX_SIGNALS_IN_QUEUE)
            {
               rmd.rmd_properties.max_signals_in_queue =
                  sig->rmd_set_property_request.value; 
            }
            else if (sig->rmd_set_property_request.property ==
                     RMD_MAX_STACK_TOP_SIZE)
            {
               if (sig->rmd_set_property_request.value < 200)
               {
                  rmd.rmd_properties.max_stack_top_size = 200;
               }
               else
               {
                  rmd.rmd_properties.max_stack_top_size =
                     sig->rmd_set_property_request.value;
               }
            }
            else
            {
               ramlog_printf("RMD: recieved invalid rmd property %d\n",
                             sig->rmd_set_property_request.property); 
            }
            /* Send reply signal */
	    {
	       union SIGNAL *reply;
	       reply = alloc(sizeof(RmdStatusReply), RMD_SET_PROPERTY_REPLY);
	       reply->rmd_reply.status = RMD_STATUS_OK;
	       send(&reply, sender(&sig));
	    }
	    free_buf(&sig);
	    break;
         case RMD_END_OF_DUMP_REQUEST:
	 {
            union SIGNAL *reply;
            /* get dump from pid */
            dump = get_dump_from_pid(sender(&sig), &rmd.dumps);
            reply = alloc(sizeof(RmdStatusReply), RMD_END_OF_DUMP_REPLY);
            /* if status = 0, dump success text block,else,
             * dump incomplete text block
             */
            if(sig->rmd_end_of_dump_request.status != 0)
            {  
               reply->rmd_reply.status =
                 rmd_file_dump_text_block(END_OF_DUMP_FAILURE,dump->dump_file);
            }
            else
            {
               reply->rmd_reply.status =
                 rmd_file_dump_text_block(END_OF_DUMP_SUCCESS,dump->dump_file);
            }
            send(&reply, sender(&sig));
            free_buf(&sig);
            break;
	 }
	 
         case RMD_GET_DUMP_INFO_REQUEST:
            rmd_private_get_dump_info(sig, &rmd);
            free_buf(&sig);
            break;
            
         default:
            ramlog_printf("RMD: received signal %d\n", sig->signo);
            free_buf(&sig);
            break;
      }
      sig = receive(sigsel);
   }
}

