/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2009 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

/**
 * @file rmd_fileformat.c
 *
 * @brief This file contains the dump_file file writing implementation used by the
 * run mode dumper process.
 *
 */

#include "ose.h"
/*#include "malloc.h" Not SMP safe */
#include "mm.h"
#include "stdint.h"
#include "stdlib.h"
#include "stdio.h"
#include "unistd.h"
#include "netinet/in.h" /* Need  htonl */
#include "zlib.h"
#include "string.h"
#include "ramlog.h"
#include "errno.h"

#include "rmd_file.h"
#include "rmd_file_private.h"

/* For debugging.  */
/* Always flush to filesystem after a write */
/*
 #define __FLUSH__
 */
/* Dump alot of trace print-outs in ramlog */
/* #define __RAMLOG_TRACE__ */

/* Signal stub used for free_buf */
union SIGNAL
{
   uint32_t signo;
};

/*
 * Function library for the IFF file format.
 * The implementation is ported from the java implementation
 * /vobs/ose5/tools/eclipse/com.ose.system/src/com/ose/system/DumpFileWriter.java
 * and ramdump functionality
 * /vobs/ose5/core/core/src/krn/ramdump/ramdump.c
 */

/*
 * ****************************************************************************
 * External interface of the RMD file writer
 * ****************************************************************************
 */

/**
 * @function rmd_start_dump
 *
 * @brief Sets up the dump_file structure, opens the dump_file file and writes the
 * file head.
 *
 * @long Sets up the dump_file structure, opens the dump_file file and writes the
 * file head. This is the first function to call in the interface.
 * Implements the same functionality for RMD as
 * @input filename Name of file including full path.
 * @input dump_id ID of the dump_file
 * @input dump_file Dump to work on
 *
 * @return Returns RMD_FILE_SUCCESS if successful , otherwise it returns
 * RMD_FILE_WRITE_ERROR if an error occurred with file writing,
 * RMD_FILE_WRITE_ERROR_SHORT if the data written is too short,
 * or RMD_FILE_ERROR if there were an unforeseen error.
 *
 * @seealso rmd_end_dump
 */
RmdFileStatus
rmd_file_start_dump(const char *filename, uint32_t dump_id,
         uint32_t use_compression, uint32_t window_bits, uint32_t mem_level,
         RmdDumpFile *dump_file)
{
   RmdFileStatus ret;
   size_t size_position;

   union
   {
      struct IFFChunkVERS ch_VERS;
      struct IFFChunkDPID ch_DPID;
      struct IFFHeaderLIST hd_LIST;
   } u;
#ifdef __RAMLOG_TRACE__
   ramlog_printf("RMD_FILE: rmd_file_start_dump()\n");
#endif

   if (use_compression)
   {
      /* Set up compression data structure */
      dump_file->compression = (struct compression_struct *) alloc(
               sizeof(struct compression_struct), 0);
      if (NULL == dump_file->compression)
      {
#ifdef __RAMLOG_TRACE__
         ramlog_printf("RMD_FILE: malloc(compression_struct) failed\n");
#endif
         return RMD_FILE_MEM_ERROR;
      }

      ret = compression_init(window_bits, mem_level, dump_file);

      if (ret != RMD_FILE_SUCCESS)
      {
#ifdef __RAMLOG_TRACE__
         ramlog_printf(
                  "RMD_FILE: compression_init() failed with code 0x%x\n",
                  ret);
#endif
         /* Only free memory allocated in this function if compression_init() fails,
          * it handles internally allocated memory freeing by it self if something fails */
         free(dump_file->compression);

         return ret;
      }
   }
   else
   {
      /* set compression to NULL to indicate the compression is turned off */
      dump_file->compression = NULL;
   }

   /* open a file for writing and truncate if it exist, otherwise create it */
   dump_file->file = fopen(filename, "wb");
   if (NULL == dump_file->file)
   {
#ifdef __RAMLOG_TRACE__
      ramlog_printf("RMD_FILE: fopen() failed with code %d\n", errno);
#endif
      /* free all allocated memory */
      if (dump_file->compression)
      {
         if (dump_file->compression->zstrm)
            free_buf((union SIGNAL **) &dump_file->compression->zstrm);

         if (dump_file->compression->zlib_buffers)
            free_buf((union SIGNAL **) &dump_file->compression->zlib_buffers);

         free_buf((union SIGNAL **) &dump_file->compression);
      }
      return RMD_FILE_OPEN_ERROR;
   }

   dump_file->sizes.dump_size = 40;
   dump_file->sizes.block_list_size = 4;
   dump_file->block_number = 0;
   dump_file->dump_id = dump_id;

   /* Dump header */
   size_position = write_group_header(TAG_FORM, sizeof(struct IFFHeaderFORM)
            - sizeof(struct IFFChunkHeader), TAG_PMD_, dump_file);
  
   if (size_position != DUMP_SIZE_POS)
   {
#ifdef __RAMLOG_TRACE__
      ramlog_printf(
               "RMD_FILE: write_group_header() failed with code %d\n",
               ret);
#endif
      /* free all allocated memory */
      if (dump_file->compression)
      {
         if (dump_file->compression->zstrm)
            free(dump_file->compression->zstrm);

         if (dump_file->compression->zlib_buffers)
            free(dump_file->compression->zlib_buffers);

         free(dump_file->compression);
      }
      return RMD_FILE_WRITE_ERROR;
   }

   /* PMD version */
   u.ch_VERS.chd.id = htonl(TAG_VERS);
   u.ch_VERS.chd.size = htonl(sizeof(struct IFFChunkVERS)
            - sizeof(struct IFFChunkHeader));
   u.ch_VERS.version = PMD_VERSION;

   ret = write_memory((uint32_t *) &u.ch_VERS, sizeof(struct IFFChunkVERS),
            dump_file);
   if (ret != RMD_FILE_SUCCESS)
   {
#ifdef __RAMLOG_TRACE__
      ramlog_printf("RMD_FILE: write_memory() failed with code %d\n",
               ret);
#endif
      /* free all allocated memory */
      if (dump_file->compression)
      {
         if (dump_file->compression->zstrm)
            free(dump_file->compression->zstrm);

         if (dump_file->compression->zlib_buffers)
            free(dump_file->compression->zlib_buffers);

         free(dump_file->compression);
      }
      return ret;
   }

   /* Dump id */
   u.ch_DPID.chd.id = htonl(TAG_DPID);
   u.ch_DPID.chd.size = htonl(sizeof(struct IFFChunkDPID)
            - sizeof(struct IFFChunkHeader));
   u.ch_DPID.dump_id = dump_id;

   ret = write_memory((uint32_t *) &u.ch_DPID, sizeof(struct IFFChunkDPID),
            dump_file);
   if (ret != RMD_FILE_SUCCESS)
   {
#ifdef __RAMLOG_TRACE__
      ramlog_printf("RMD_FILE: write_memory() failed with code %d\n",
               ret);
#endif
      /* free all allocated memory */
      if (dump_file->compression)
      {
         if (dump_file->compression->zstrm)
            free(dump_file->compression->zstrm);

         if (dump_file->compression->zlib_buffers)
            free(dump_file->compression->zlib_buffers);

         free(dump_file->compression);
      }
      return ret;
   }

   /* Empty block list */
   u.hd_LIST.chd.id = htonl(TAG_LIST);
   u.hd_LIST.chd.size = htonl(sizeof(struct IFFHeaderLIST)
            - sizeof(struct IFFChunkHeader));
   u.hd_LIST.cont_type = htonl(TAG_BLOC);

   ret = write_memory((uint32_t *) &u.hd_LIST, sizeof(struct IFFHeaderLIST),
            dump_file);
   if (ret != RMD_FILE_SUCCESS)
   {
#ifdef __RAMLOG_TRACE__
      ramlog_printf("RMD_FILE: write_memory() failed with code %d\n",
               ret);
#endif
      /* free all allocated memory */
      if (dump_file->compression)
      {
         if (dump_file->compression->zstrm)
            free(dump_file->compression->zstrm);

         if (dump_file->compression->zlib_buffers)
            free(dump_file->compression->zlib_buffers);

         free(dump_file->compression);
      }
      return ret;
   }

   /* Flush to file system so that the chunk is written */
   if (fflush(dump_file->file))
      return RMD_FILE_FLUSH_ERROR;
   /* Force writing to file */
   if (fsync(fileno(dump_file->file)))
      return RMD_FILE_FLUSH_ERROR;

   return RMD_FILE_SUCCESS;
}

/**
 * @function rmd_end_dump
 *
 * @brief Closes the dump_file file and frees up memory allocated. Should always be
 * called when a dump_file is finished.
 *
 * @long Closes the dump_file file and frees up memory allocated. Should always be
 * called when a dump_file is finished. The function frees all memory allocated in
 * rmd_start_dump
 *
 * @input dump_file Dump to work on
 *
 * @seealso rmd_start_dump
 *
 */
void
rmd_file_end_dump(RmdDumpFile *dump_file)
{
   if (dump_file != NULL)
   {
#ifdef __RAMLOG_TRACE__
      ramlog_printf("RMD_FILE: End dump_file: Freeing resources\n");
#endif

      /* Flush to filesystem if data is still pending in the stream */
      /* Flush and close file, no need to know if successful */
      /* Flush to file system so that the chunk is written */
      fflush(dump_file->file);
      fsync(fileno(dump_file->file));
      fclose(dump_file->file);

      if (dump_file->compression != NULL)
      {
         /* Free up memory used by deflate */
         deflateEnd(dump_file->compression->zstrm);

         /* Free objects we allocated in rmd_start_dump */
         if (dump_file->compression->zlib_buffers != NULL)
         {
            free_buf((union SIGNAL **) &dump_file->compression->zlib_buffers);
            dump_file->compression->zlib_buffers = NULL;
         }

         if (dump_file->compression->zstrm != NULL)
         {
            free_buf((union SIGNAL **) &dump_file->compression->zstrm);
            dump_file->compression->zstrm = NULL;
         }

         free_buf((union SIGNAL **) &dump_file->compression);
         dump_file->compression = NULL;
      }
   }
}

/**
 * @function rmd_dump_error_block
 *
 * @brief Writes an error block to the dump_file.
 *
 * @long Writes an error block to the dump_file. This is called externally from the
 * code that is implementing the actual dumping.
 *
 * @input error_code Error code.
 * @input extra Extra parameter.
 * @input cpid Current process (the process that is beeing dumped)
 * @input description String with description of error
 * @input dump_file Dump to write to
 *
 * @return Returns RMD_FILE_SUCCESS if successful , otherwise it
 * returns RMD_FILE_WRITE_ERROR if fwrite fails, RMD_FILE_WRITE_SHORT_ERROR
 * if the data written is shorter than expected, RMD_FILE_SEEK_ERROR if fseek
 * fails
 *
 */
RmdFileStatus
rmd_file_dump_error_block(const char *description, uint32_t user_called,
         uint32_t error_code, uint32_t extra, uint32_t cpid,
         RmdDumpFile *dump_file)
{
   RmdFileStatus ret;
   /* Struct for holding the header in memory, written to file in one chunk */
   struct IFFChunkERIN ch_ERIN;

   int padding;
   int desc_size;
   size_t size_position;

#ifdef __RAMLOG_TRACE__
   ramlog_printf("RMD: rmd_file_dump_error_block()\n");
#endif

   desc_size = strnlen(description, RMD_MAX_STRING_LENGTH) + 1;
   padding = get_padding(desc_size);

   ramlog_printf("\t");
   size_position = write_new_block_header(ERROR_BLOCK_SIZE + desc_size + padding,
            dump_file);
   if (0 == size_position)
      return RMD_FILE_ERROR;

   ramlog_printf("\t");
   size_position = write_group_header(TAG_FORM, ERBL_SIZE + desc_size + padding,
            TAG_ERBL, dump_file);
   if (0 == size_position) 
      return RMD_FILE_ERROR;

   /* Error info */
   ch_ERIN.chd.id = htonl(TAG_ERIN);
   ch_ERIN.chd.size = htonl(ERIN_SIZE);
   ch_ERIN.user_called = user_called;
   ch_ERIN.error_code = error_code;
   ch_ERIN.extra = extra;
   ch_ERIN.curr_proc = cpid;

   ramlog_printf("\t");
   ret = write_memory((uint32_t *) &ch_ERIN, sizeof(struct IFFChunkERIN),
            dump_file);
   if (ret != RMD_FILE_SUCCESS)
      return ret;

   ret = write_desc(description, desc_size, padding, dump_file);
   if (ret != RMD_FILE_SUCCESS)
      return ret;

   /* Update previous block sizes with the new sizes, 8 comes from the
    * first FORM and size */
   ret = update_sizes(ERROR_BLOCK_SIZE + 8 + desc_size + padding, dump_file);
   if (ret != RMD_FILE_SUCCESS)
      return ret;

   /* Flush to file system so that the chunk is written */
   if (fflush(dump_file->file))
      return RMD_FILE_FLUSH_ERROR;
   /* Force writing to file */
   if (fsync(fileno(dump_file->file)))
      return RMD_FILE_FLUSH_ERROR;

   /* Everything is ok, return from function with a success code */
   return RMD_FILE_SUCCESS;

}

/**
 * @function rmd_file_dump_text_block
 *
 * @brief Writes a text block to the dump_file.
 *
 * @long Writes a text block to the dump_file. This is called externally from the
 * code that is implementing the actual dumping.
 *
 * @input description String with description of error
 * @input dump_file Dump to write to
 *
 * @return Returns RMD_FILE_SUCCESS if successful , otherwise it
 * returns RMD_FILE_WRITE_ERROR if fwrite fails, RMD_FILE_WRITE_SHORT_ERROR
 * if the data written is shorter than expected, RMD_FILE_SEEK_ERROR if fseek
 * fails
 *
 */
RmdFileStatus
rmd_file_dump_text_block(const char *description, RmdDumpFile *dump_file)
{
   RmdFileStatus ret;
   size_t size_position;
   int padding;
   int desc_size;
   int desc_block_size;

#ifdef __RAMLOG_TRACE__
   ramlog_printf("RMD_FILE: Dump text block: \n");
#endif

   desc_size = strnlen(description, RMD_MAX_STRING_LENGTH) + 1;
   padding = get_padding(desc_size);
   /* size of the description block in the file. 12 is the header data (3 x 32 bit words)*/
   desc_block_size = desc_size + padding + 12;

   /* 36 is the extra space of the headers */
   size_position = write_new_block_header(36 + desc_block_size, dump_file);
   if (0 == size_position)
      return RMD_FILE_ERROR;

   size_position = write_group_header(TAG_FORM, desc_block_size, TAG_TXBL, dump_file);
   if (0 == size_position)
      return RMD_FILE_ERROR;

   ret = write_desc(description, desc_size, padding, dump_file);
   if (ret != RMD_FILE_SUCCESS)
      return ret;

   /* 44 is the extra space of the headers, 8 bytes (2 x 32 bit words) bigger than
    * the size in the block header because the block tag and size of the block is
    * included here */
   ret = update_sizes(44 + desc_block_size, dump_file);
   if (ret != RMD_FILE_SUCCESS)
      return ret;

   /* Flush to file system so that the chunk is written */
   if (fflush(dump_file->file))
      return RMD_FILE_FLUSH_ERROR;
   /* Force writing to file */
   if (fsync(fileno(dump_file->file)))
      return RMD_FILE_FLUSH_ERROR;

   /* Everything is ok, return from function with a success code */
   return RMD_FILE_SUCCESS;
}

/**
 * @function rmd_dump_memory_block
 *
 * @brief Writes a memory block to dump_file including header. Uses compression
 * if dump_file->compression is set (!= NULL).
 *
 * @long Writes a memory block to dump_file including header. Uses compression
 * if dump_file->compression is set (!= 0). External function to the dump_file file
 * interface.
 *
 * @input block OSE Block id for which to dump memory
 * @input start_address Address to get data to compress and write to dump_file
 * @input dataLen Length of data
 * @input description String describing the block.
 * @input dump_file Dump to work on
 *
 * @return Returns RMD_FILE_SUCCESS if successful , otherwise it returns
 * RMD_FILE_MEM_ERROR if there were a memory allocation error,
 * RMD_FILE_Z_STREAM_ERROR if there were an error in the deflate (),
 * RMD_Z_STREAM_END if there were an unexpected end of data to the compressor,
 * RMD_FILE_WRITE_ERROR if an error occured with file writing,
 * RMD_FILE_WRITE_ERROR_SHORT if the data written is too short,
 * RMD_FILE_SEEK_ERROR if there were an error when seeking in the file
 * RMD_FILE_TELL_ERROR if there were an error when storing the current file position
 * or RMD_FILE_ERROR if there were an unforeseen error.
 *
 */
RmdFileStatus
rmd_file_dump_memory_block(PROCESS block, OSADDRESS data, uint32_t data_length,
         const char *description, RmdDumpFile *dump_file)
{
   uint32_t padding = 0;
   RmdFileStatus status = RMD_FILE_SUCCESS;
   uint32_t top_form_size = MHD_TOP_FORM_SIZE;
   uint32_t form_size = MHD_FORM_SIZE;
   uint32_t total_header_size;
   void *tmp_memory_buffer;
   
#ifdef __RAMLOG_TRACE__
   ramlog_printf("RMD FILE: rmd_file_dump_memory_block()\n");
#endif

   dump_file->sizes.compressed_data_size = 0;
   dump_file->sizes.uncompressed_data_size = 0;

   if (dump_file->compression != NULL)
   {
      status = write_data_compressed_reset(dump_file);
      if (status != RMD_FILE_SUCCESS)
         return status;

   }

   total_header_size = dump_memory_block_header(data, data_length, description,
            dump_file);
   
   if (0 == total_header_size)
   {
      return RMD_FILE_ERROR;
   }

   tmp_memory_buffer = (void *)ose_mm_alloc("ose_rmd_mem_buffer",
         (data_length < 4) ? 4 : data_length,
         MM_ALLOW_VIRTUAL | MM_ALLOW_CACHED);
   if (!tmp_memory_buffer)
   {
      return RMD_FILE_MEM_ERROR;
   }
   if (data_length > 0 ? get_mem(block, data, tmp_memory_buffer, data_length) : 0)
   {
      ose_mm_free((OSADDRESS)tmp_memory_buffer);
      return RMD_FILE_MEM_ERROR;
   }

   if (dump_file->compression != NULL)
   {
      status = write_data_compressed((OSADDRESS)tmp_memory_buffer, data_length, dump_file);
      if (status != RMD_FILE_SUCCESS)
      {
         ose_mm_free((OSADDRESS)tmp_memory_buffer);
         return status;
      }

      status = write_data_compressed_end(dump_file);
      if (status != RMD_FILE_SUCCESS)
      {
         ose_mm_free((OSADDRESS)tmp_memory_buffer);
         return status;
      }

      total_header_size += COMPRESSION_HEADER_SIZE;
      top_form_size += COMPRESSION_HEADER_SIZE;
      form_size += COMPRESSION_HEADER_SIZE;
   }
   else
   {
      status = write_data((OSADDRESS)tmp_memory_buffer, data_length, dump_file);
      if (status != RMD_FILE_SUCCESS)
      {
         ose_mm_free((OSADDRESS)tmp_memory_buffer);
         return status;
      }

      dump_file->sizes.compressed_data_size
               = dump_file->sizes.uncompressed_data_size;
   }
   ose_mm_free((OSADDRESS)tmp_memory_buffer);

   padding = get_padding(dump_file->sizes.compressed_data_size);
   status = write_padding(padding, 0, dump_file);
   if (status != RMD_FILE_SUCCESS)
   {
      return status;
   }

   status = update_sizes_all(total_header_size, padding, dump_file);
   if (status != RMD_FILE_SUCCESS)
   {
      return status;
   }

   /* Flush to file system so that the chunk is written */
   if (fflush(dump_file->file))
      return RMD_FILE_FLUSH_ERROR;
   /* Force writing to file */
   if (fsync(fileno(dump_file->file)))
      return RMD_FILE_FLUSH_ERROR;

   return RMD_FILE_SUCCESS;
}

/**
 * @function rmd_file_dump_request
 *
 * @brief Writes header for signal block and initialises dump_file for memory data.
 *
 * @long Writes header for signal block and initialises dump_file for memory data.
 * This is called externally from the code that is implementing the actual
 * dumping.
 *
 * @input request Pointer to the request signal that starts the transaction,
 * this signal will be written as the first signal in the signal block.
 * @input dump_file Dump to write to
 *
 * @return Returns RMD_FILE_SUCCESS if successful , otherwise it returns
 * RMD_FILE_WRITE_ERROR if fwrite fails,
 * RMD_FILE_WRITE_SHORT_ERROR if the data written is shorter than expected,
 * RMD_FILE_SEEK_ERROR if fseek fails
 *
 */
RmdFileStatus
rmd_file_dump_request(OSADDRESS sig_address, uint32_t sig_size,
         RmdDumpFile *dump_file)
{
   RmdFileStatus ret = RMD_FILE_SUCCESS;

#ifdef __RAMLOG_TRACE__
   ramlog_printf("RMD_FILE: rmd_file_dump_request():\n");
#endif
   dump_file->block_type = BLOCK_SIGNAL;

   dump_file->sizes.compressed_data_size = 0;
   dump_file->sizes.uncompressed_data_size = 0;

   if (dump_file->compression != NULL)
   {
      ret = write_data_compressed_reset(dump_file);
      if (ret != RMD_FILE_SUCCESS)
         return ret;
   }

   ret = dump_request_header(sig_address, dump_file);
   if (0 == ret)
   {
      return RMD_FILE_ERROR;
   }

   /* dump_file actual signal data */
   ret = rmd_file_dump_reply(sig_address, sig_size, dump_file);
   if (ret != RMD_FILE_SUCCESS)
   {
      return ret;
   }

   return ret;
}

/**
 * @function rmd_file_dump_reply
 *
 * @brief Writes signal to dump_file.
 *
 * @long Writes signal to dump_file. This is called externally from the code that
 * is implementing the actual dumping.
 *
 * @input sig Pointer to the request signal that starts the transaction
 * @input dump_file Dump to write to
 *
 * @return Returns RMD_FILE_SUCCESS if successful , otherwise it returns
 * RMD_FILE_WRITE_ERROR if fwrite fails,
 * RMD_FILE_WRITE_SHORT_ERROR if the data written is shorter than expected,
 * RMD_FILE_SEEK_ERROR if fseek fails
 * RMD_FILE_BLOCK_ERROR if the block is not initialised
 * or RMD_FILE_ERROR if an unforeseen error occured
 */
RmdFileStatus
rmd_file_dump_reply(OSADDRESS sig_address, uint32_t sig_size,
         RmdDumpFile *dump_file)
{
   RmdFileStatus ret;

#ifdef __RAMLOG_TRACE__
   ramlog_printf(
            "RMD_FILE: rmd_file_dump_reply():\n\tSignal address: %d\n\tSignal size: %d\n",
            (uint32_t) sig_address, sig_size);
#endif
   if (dump_file->block_type != BLOCK_SIGNAL)
   {
      /* Must call rmd_file_dump_request() before this */
      return RMD_FILE_BLOCK_ERROR;
   }

   if ((OSADDRESS)NULL == sig_address)
   {
      /* Nothing to write */
      return RMD_FILE_ERROR;

   }

   if (dump_file->compression != NULL)
   {
      /*
       * Use Compression
       */
      ret = write_data_compressed((OSADDRESS)&sig_size, 4, dump_file);
      if (ret != RMD_FILE_SUCCESS)
      {
         return ret;
      }
      ret = write_data_compressed(sig_address, sig_size, dump_file);
      if (ret != RMD_FILE_SUCCESS)
      {
         return ret;
      }
   }
   else
   {
      /*
       * No compression
       */
      ret = write_data((OSADDRESS)&sig_size, 4, dump_file);
      if (ret != RMD_FILE_SUCCESS)
      {
         return ret;
      }
      ret = write_data(sig_address, sig_size, dump_file);
      if (ret != RMD_FILE_SUCCESS)
      {
         return ret;
      }
   }

   return RMD_FILE_SUCCESS;
}

/**
 * @function rmd_file_dump_endmark
 *
 * @brief Finishes the writing of signals to dump_file.
 *
 * @long Finishes the writing of signals to dump_file. This is called externally
 * from the code that is implementing the actual dumping. Implements the same
 * functionality for RMD as ramdump_dump_endmark do for the FMD.
 *
 * @input endmark Endmark signal. Set to NULL if the signal does not exist,
 * this will not be written to dump_file.
 * @input status Status code for the signal block. This is written to the dump_file.
 * @input dump_file Dump to write to
 *
 * @return Returns RMD_FILE_SUCCESS if successful , otherwise it returns
 * RMD_FILE_WRITE_ERROR if fwrite fails,
 * RMD_FILE_WRITE_SHORT_ERROR if the data written is shorter than expected,
 * RMD_FILE_SEEK_ERROR if fseek fails
 * RMD_FILE_BLOCK_ERROR if the block is not initialised
 * or RMD_FILE_ERROR if an unforeseen error occured
 */
RmdFileStatus
rmd_file_dump_endmark(OSADDRESS sig_address, uint32_t sig_size,
         uint32_t status, RmdDumpFile *dump_file)
{
   uint32_t padding;
   RmdFileStatus ret;
   uint32_t total_header_size = SGHD_SIZE;
   uint32_t top_form_size = SGHD_TOP_FORM_SIZE;
   uint32_t form_size = SGHD_FORM_SIZE;

#ifdef __RAMLOG_TRACE__
   ramlog_printf("RMD_FILE: rmd_file_dump_endmark():\n");
#endif

   /* endmark can be null, if not dump_file the content to file */
   if (sig_address != (OSADDRESS)NULL)
   {
      ret = rmd_file_dump_reply(sig_address, sig_size, dump_file);
      if (ret != RMD_FILE_SUCCESS)
      {
         return ret;
      }
   }

   dump_file->block_type = BLOCK_NONE;

   if (NULL == dump_file->compression)
   {
      /*
       * No compression
       */
      dump_file->sizes.compressed_data_size
               = dump_file->sizes.uncompressed_data_size;
   }
   else
   {
      /* Tell zlib we're done for now */
      ret = write_data_compressed_end(dump_file);
      total_header_size += COMPRESSION_HEADER_SIZE;
      top_form_size += COMPRESSION_HEADER_SIZE;
      form_size += COMPRESSION_HEADER_SIZE;
   }

   /* the data written to file needs to be padded to 32 bit alignment */
   padding = get_padding(dump_file->sizes.compressed_data_size);
   ret = write_padding(padding, 0, dump_file);

   if (ret != RMD_FILE_SUCCESS)
   {
      return ret;
   }

   ret = update_sizes_all(total_header_size, padding, dump_file);
   if (ret != RMD_FILE_SUCCESS)
   {
      return ret;
   }

   fseek(dump_file->file, dump_file->file_pos.content_status_pos, SEEK_SET);
   if (EOF == dump_file->file_pos.zdef_image_length_pos)
   {
      /* ftell error */
      return RMD_FILE_TELL_ERROR;
   }

   ret = write_int(status, dump_file);
   if (ret != RMD_FILE_SUCCESS)
   {
      return ret;
   }

   /* Go to end of file */
   fseek(dump_file->file, 0, SEEK_END);
   if (EOF == dump_file->file_pos.zdef_image_length_pos)
   {
      return RMD_FILE_SEEK_ERROR;
   }

   /* Flush to file system so that the chunk is written */
   if (fflush(dump_file->file))
      return RMD_FILE_FLUSH_ERROR;
   /* Force writing to file */
   if (fsync(fileno(dump_file->file)))
      return RMD_FILE_FLUSH_ERROR;

   return RMD_FILE_SUCCESS;
}

/*
 * ****************************************************************************
 * Internal functions
 * ****************************************************************************
 */

/**
 * Zlib memory allocation functions
 *
 */
static voidpf
zlib_alloc(voidpf opaque, uInt items, uInt size)
{
   /* malloc is not SMP compatible in OSE 5.4 */
   /* return malloc(items * size); */
   (void) opaque;
   /* Use memory manager to get memory */
   return (voidpf) ose_mm_alloc("rmd_zlib_buffer", items * size,
            MM_ALLOW_VIRTUAL | MM_ALLOW_CACHED);

}

static void
zlib_free(voidpf opaque, voidpf address)
{
   /* malloc and free is not SMP safe */
   /* free(address); */
   (void) opaque;
   ose_mm_free((OSADDRESS)address);
}

/**
 * @function write_int
 *
 * @brief Writes a 32 bit word to the dump_file.
 *
 * @long Writes a 32 bit word to the dump_file. dump_file field in
 * the dump_file struct must be initialised using open() before calling
 * this function. This is called from the higher level
 * functions of the interface.
 *
 * @input value 32 bit word to write to file
 * @input dump_file Dump to write to.
 *
 * @return Returns RMD_FILE_SUCCESS if successful , otherwise it
 * returns RMD_FILE_WRITE_ERROR if fwrite fails, RMD_FILE_WRITE_SHORT_ERROR
 * if the data written is shorter than expected
 *
 */
static RmdFileStatus
write_int(uint32_t value, const RmdDumpFile *dump_file)
{
   RmdFileStatus ret;

   ret = fwrite(&value, 4, 1, dump_file->file);
   if (ferror(dump_file->file))
      return RMD_FILE_WRITE_ERROR;

   /* Check if all data has been written */
   if (ret < 1)
      return RMD_FILE_WRITE_SHORT_ERROR;

#ifdef __FLUSH__
   if (fflush(dump_file->file) != 0)
   return RMD_FILE_FLUSH_ERROR;
#endif

   return RMD_FILE_SUCCESS;
}

/**
 * @function write_memory
 *
 * @brief Writes raw memory.
 *
 * @input data Pointer to memory to write
 * @input data_len Length of memory segment to write (in bytes)
 * @input dump_file Dump to write to.
 *
 * @return Returns RMD_FILE_SUCCESS if successful , otherwise it
 * returns RMD_FILE_WRITE_ERROR if fwrite fails, RMD_FILE_WRITE_SHORT_ERROR
 * if the data written is shorter than expected
 *
 * @example Typical use is
 * write_memory((uint32_t *) whatever, sizeof(whatever));
 * where whatever is struct holding formatted data
 *
 */
static RmdFileStatus
write_memory(const void *data, const uint32_t data_len,
         const RmdDumpFile *dump_file)
{
   int bytes_written;

#ifdef __RAMLOG_TRACE__
   ramlog_printf("RMD FILE: write_memory(): ");
#endif

   bytes_written = fwrite(data, 1, data_len, dump_file->file);

#ifdef __RAMLOG_TRACE__
   ramlog_printf("bytes_written = %d\n", bytes_written);
#endif
   if (ferror(dump_file->file))
      return RMD_FILE_WRITE_ERROR;

   if ((uint32_t) bytes_written < data_len)
      return RMD_FILE_WRITE_SHORT_ERROR;

#ifdef __FLUSH__
   if (fflush(dump_file->file) != 0)
   return RMD_FILE_FLUSH_ERROR;
#endif

   return RMD_FILE_SUCCESS;
}

/**
 * @function get_padding
 *
 * @brief Gets the number of bytes to pad in order to get 32 bit words
 * in the dump_file.
 *
 * @long Gets the number of bytes to pad in order to get 32 bit words
 * in the dump_file. This is called from the higher level functions of the
 * interface.
 *
 * @input size Dump size to align to 32 bit
 *
 * @return Returns number of bytes [0, 1, 2, 3] to align to 32 bit words
 *
 */
static uint32_t
get_padding(uint32_t size)
{
#ifdef __RAMLOG_TRACE__
   ramlog_printf("RMD FILE: get_padding()\n");
#endif
   return (4 - (size & 3)) % 4;
}

/**
 * @function write_padding
 *
 * @brief Writes padding bytes to the dump_file.
 *
 * @long Writes padding bytes to the dump_file. This is called from the higher level
 * functions of the interface.
 *
 * @input padding Number of padding bytes to write.
 * @input data Data to write as padding.
 * @input dump_file Dump to write to.
 *
 * @return Returns RMD_FILE_SUCCESS if successful or RMD_FILE_WRITE_ERROR
 * if a file write fails
 *
 */
static RmdFileStatus
write_padding(uint32_t padding, uint8_t data, const RmdDumpFile *dump_file)
{
#ifdef __RAMLOG_TRACE__
   ramlog_printf("RMD FILE: write_padding()\tpadding = %d\n", padding);
#endif
   fwrite(&data, 1, padding, dump_file->file);
   if (ferror(dump_file->file))
   {
      return RMD_FILE_WRITE_ERROR;
   }

#ifdef __FLUSH__
   if (fflush(dump_file->file) != 0)
   return RMD_FILE_FLUSH_ERROR;
#endif

   return RMD_FILE_SUCCESS;
}

/**
 * @function seek_and_write_int
 *
 * @brief Help function to simplify code in update_sizes and update_sizes_all.
 *
 * @input dump_file Dump to write to
 * @input offset Offset in file
 * @input data Data to writeto file
 *
 * @return Returns RMD_FILE_SUCCESS if successful , otherwise it
 * returns RMD_FILE_WRITE_ERROR if fwrite fails, RMD_FILE_WRITE_SHORT_ERROR
 * if the data written is shorter than expected or RMD_FILE_SEEK_ERROR if
 * the fseek fails
 *
 */
static RmdFileStatus
seek_and_write_int(const RmdDumpFile *dump_file, uint32_t offset,
         uint32_t value)
{
   RmdFileStatus ret;
   ret = fseek(dump_file->file, offset, SEEK_SET);
   if (ret != 0)
   {
      /* fseek didn't return 0, the seek failed */
      return RMD_FILE_SEEK_ERROR;
   }
   else
   {
      ret = write_int(value, dump_file);
      if (ret != RMD_FILE_SUCCESS)
      {
         return ret;
      }
   }

   return RMD_FILE_SUCCESS;
}

/**
 * @function update_sizes
 *
 * @brief Updates the size fields in the dump_file.
 *
 * @long Updates the size fields in the dump_file. Seeks back to stored positions
 * in the file to update size data. Updates dump_size and block_list_size in the
 * file. This is called from the higher level functions of the interface.
 *
 * @input size Size of the written block
 * @input dump_file Dump to write to.
 *
 * @return Returns RMD_FILE_SUCCESS if successful , otherwise it
 * returns RMD_FILE_WRITE_ERROR if fwrite fails, RMD_FILE_WRITE_SHORT_ERROR
 * if the data written is shorter than expected or RMD_FILE_SEEK_ERROR if
 * the fseek fails
 *
 */
static RmdFileStatus
update_sizes(uint32_t size, RmdDumpFile *dump_file)
{
   RmdFileStatus filestatus;
   /* Go to offset dump_file->file_pos.dump_sizePos relative file start */
#ifdef __RAMLOG_TRACE__
   ramlog_printf("RMD FILE: update_sizes()\n");
#endif
   dump_file->sizes.block_list_size += size;

   filestatus = seek_and_write_int(dump_file, BLOCK_LIST_SIZE_POS, htonl(
            dump_file->sizes.block_list_size));
   if (filestatus != RMD_FILE_SUCCESS)
   {
      return filestatus;
   }

   dump_file->sizes.dump_size += size;

   filestatus = seek_and_write_int(dump_file, DUMP_SIZE_POS, htonl(
            dump_file->sizes.dump_size));
   if (filestatus != RMD_FILE_SUCCESS)
   {
      return filestatus;
   }

   /* Go back to end of file */
   if (fseek(dump_file->file, 0, SEEK_END))
   {
      /* fseek didn't return 0, the seek failed */
      return RMD_FILE_SEEK_ERROR;
   }

   return RMD_FILE_SUCCESS;
}

/**
 * @function update_sizes_all
 *
 * @brief Updates the size fields in the dump_file.
 *
 * @long Updates the size fields in the dump_file. Seeks back to stored positions
 * in the file to update size data. Updates all sizes in the dump_file. This is
 * called from the higher level functions of the interface.
 *
 * @input total_header_size Size of block header, including description if it exists
 * @input top_form_size Size of top form block
 * @input form_size Size of form block (is always a fixed value smaller than top_form_block)
 * @input padding Padding of the written block
 * @input dump_file Dump to write to. Also used as input for the sizes stored in
 * dump_file->sizes.compressedDAtaSize and dump_file->sizes.uncompressed_data_size
 *
 * @output dump_file->sizes.dump_size
 * @output dump_file->sizes.block_list_size
 *
 * @return Returns RMD_FILE_SUCCESS if successful , otherwise it
 * returns RMD_FILE_WRITE_ERROR if fwrite fails, RMD_FILE_WRITE_SHORT_ERROR
 * if the data written is shorter than expected or RMD_FILE_SEEK_ERROR if
 * the fseek fails
 *
 */

#define TOP_FORM_OFFSET 8
#define FORM_OFFSET 44

static RmdFileStatus
update_sizes_all(uint32_t total_header_size, uint32_t padding,
         RmdDumpFile *dump_file)
{
   RmdFileStatus ret;
   uint32_t size;

   uint32_t top_form_size = total_header_size - TOP_FORM_OFFSET;
   uint32_t form_size = total_header_size - FORM_OFFSET;

#ifdef __RAMLOG_TRACE__
   ramlog_printf("RMD FILE: update_sizes_all()\n");
#endif

   /* Update sizes in the written block first */
   size = dump_file->sizes.compressed_data_size + padding + top_form_size;
   ret = seek_and_write_int(dump_file, dump_file->file_pos.top_form_size_pos,
            htonl(size));
   if (ret != RMD_FILE_SUCCESS)
   {
      return ret;
   }

   size = dump_file->sizes.compressed_data_size + padding + form_size;
   ret = seek_and_write_int(dump_file, dump_file->file_pos.form_size_pos,
            htonl(size));
   if (ret != RMD_FILE_SUCCESS)
   {
      return ret;
   }

   size = dump_file->sizes.uncompressed_data_size;
   ret = seek_and_write_int(dump_file, dump_file->file_pos.content_length_pos,
            size);
   if (ret != RMD_FILE_SUCCESS)
   {
      return ret;
   }

   if (dump_file->compression != NULL)
   {
      size = dump_file->sizes.compressed_data_size + padding
               + COMPRESSION_HEADER_SIZE;
      ret = seek_and_write_int(dump_file, dump_file->file_pos.zdata_size_pos,
               htonl(size));
      if (ret != RMD_FILE_SUCCESS)
      {
         return ret;
      }

      size = dump_file->sizes.compressed_data_size + COMPRESSION_HEADER_SIZE;
      ret = seek_and_write_int(dump_file,
               dump_file->file_pos.zdef_image_length_pos, size);
      if (ret != RMD_FILE_SUCCESS)
      {
         return ret;
      }
   }

   size = dump_file->sizes.compressed_data_size + padding;
   ret = seek_and_write_int(dump_file, dump_file->file_pos.data_size_pos,
            htonl(size));
   if (ret != RMD_FILE_SUCCESS)
   {
      return ret;
   }

   /* Update and write the two total sizes and hope the system
    * won't go down in between */
   dump_file->sizes.block_list_size += dump_file->sizes.compressed_data_size
            + padding + total_header_size;
   ret = seek_and_write_int(dump_file, BLOCK_LIST_SIZE_POS, htonl(
            dump_file->sizes.block_list_size));
   if (ret != RMD_FILE_SUCCESS)
   {
      return ret;
   }

   dump_file->sizes.dump_size += dump_file->sizes.compressed_data_size
            + padding + total_header_size;
   ret = seek_and_write_int(dump_file, DUMP_SIZE_POS, htonl(
            dump_file->sizes.dump_size));
   if (ret != RMD_FILE_SUCCESS)
   {
      return ret;
   }

   /* Go back to end of file */
   if (fseek(dump_file->file, 0, SEEK_END))
   {
      /* fseek didn't return 0, the seek failed */
      return RMD_FILE_SEEK_ERROR;
   }

   return RMD_FILE_SUCCESS;
}

/**
 * @function write_string
 *
 * @brief Writes a string to the dump_file.
 *
 * @long Writes a string to the dump_file. This is called from the higher
 * level functions of the interface.
 *
 * @input string The string to write to file. Must be '\0' terminated.
 * @input dump_file Dump to write to
 *
 * @return Returns RMD_FILE_SUCCESS if successful , otherwise it
 * returns RMD_FILE_WRITE_ERROR if fwrite fails, RMD_FILE_WRITE_SHORT_ERROR
 * if the data written is shorter than expected
 *
 */

/*
 static RmdFileStatus
 write_string(const char *string, uint32_t len, const RmdDumpFile *dump_file)
 {
 RmdFileStatus ret;

 #ifdef __RAMLOG_TRACE__
 ramlog_printf("RMD FILE: write_string()\n");
 #endif
 ret = fwrite(string, 1, len, dump_file->file);
 if (ferror(dump_file->file))
 return RMD_FILE_WRITE_ERROR;

 if (ret < (RmdFileStatus) len)
 return RMD_FILE_WRITE_SHORT_ERROR;

 #ifdef __FLUSH__
 fflush(dump_file->file);
 #endif

 return RMD_FILE_SUCCESS;
 }
 */

/**
 * @function write_desc
 *
 * @brief Writes a description to the dump_file, including tags and length.
 *
 * @long Writes a description to the dump_file, including tags, length and padding. This
 * is called from the higher level functions of the interface.
 *
 * @input description The description string to write to file. Must be '\0'
 * terminated.
 * @input padding Number of characters to pad the data
 * @input dump_file Dump to write to
 *
 * @return Returns RMD_FILE_SUCCESS if successful , otherwise it
 * returns RMD_FILE_WRITE_ERROR if fwrite fails, RMD_FILE_WRITE_SHORT_ERROR
 * if the data written is shorter than expected
 *
 */
static RmdFileStatus
write_desc(const char *description, int desc_size, int padding,
         const RmdDumpFile *dump_file)
{
   RmdFileStatus ret;

#ifdef __RAMLOG_TRACE__
   ramlog_printf("RMD FILE: write_desc()\n");
#endif
   /* For protection, should always point to a string */
   if (NULL == description)
   {
      return RMD_FILE_WRITE_ERROR;
   }

   ret = write_chunk_header(TAG_DESC, desc_size + padding, dump_file);
   if (ret != RMD_FILE_SUCCESS)
      return ret;

   ret = write_memory(description, desc_size, dump_file);
   if (ret != RMD_FILE_SUCCESS)
      return ret;

   ret = write_padding(padding, 0, dump_file);
   if (ret != RMD_FILE_SUCCESS)
      return ret;

   return RMD_FILE_SUCCESS;
}

/**
 * @function write_chunk_header
 *
 * @brief Writes a chunk header to the dump_file.
 *
 * @long Writes a chunk header to the dump_file.
 * @long Converts both the tag and the size to network endian before writing to disk.
 *
 * @input id Chunk id tag.
 * @input data_size Data size of the chunk.
 * @input dump_file Dump where the chunk will be written
 *
 * @return Returns RMD_FILE_SUCCESS if successful , otherwise it
 * returns RMD_FILE_WRITE_ERROR if fwrite fails, RMD_FILE_WRITE_SHORT_ERROR
 * if the data written is shorter than expected
 */
static RmdFileStatus
write_chunk_header(IffId id, uint32_t data_size, const RmdDumpFile *dump_file)
{
   RmdFileStatus ret;

#ifdef __RAMLOG_TRACE__
   ramlog_printf("RMD FILE: write_chunk_header()\n");
#endif
   ret = write_int(htonl(id), dump_file);
   if (ret != RMD_FILE_SUCCESS)
      return ret;
   ret = write_int(htonl(data_size), dump_file);
   if (ret != RMD_FILE_SUCCESS)
      return ret;

   return RMD_FILE_SUCCESS;
}

/**
 * @function write_group_header
 *
 * @brief Writes a group header (FORM, LIST, CAT_ or PROP).
 *
 * @input id Group chunk id.
 * @input data_size Data size of the group chunk.
 * @input form_type Group subchunk id.
 * @input dump_file Dump where the group header will be written.
 *
 * @return Returns position to size field in the file if successful , otherwise it
 * returns 0
 *
 */
static size_t 
write_group_header(IffId id, uint32_t data_size, IffId form_type,
         RmdDumpFile *dump_file)
{
   struct IFFHeaderFORM hd_FORM;
   RmdFileStatus ret = RMD_FILE_SUCCESS;
   long size_position;
#ifdef __RAMLOG_TRACE__
   ramlog_printf("RMD FILE: write_group_header()\n");
#endif

   size_position = ftell(dump_file->file);
   if (EOF == size_position)
   {
#ifdef __RAMLOG_TRACE__
      ramlog_printf(
               "RMD_FILE: rmd_file_dump_request_write_headers(): ftell() failed");
#endif
      return 0;
   }

   hd_FORM.chd.id = htonl(id);
   hd_FORM.chd.size = htonl(data_size);
   hd_FORM.form_type = htonl(form_type);

   ret = write_memory((uint32_t *) &hd_FORM, sizeof(struct IFFHeaderFORM),
            dump_file);

   if (ret == RMD_FILE_SUCCESS)
      return size_position + sizeof(uint32_t);
   else
      return 0;
}

/**
 * @function write_new_block_header
 *
 * @brief Writes the block header to the dump_file.
 *
 * @long Writes the block header to the dump_file. This is called from the higher level
 * functions of the interface.
 *
 * @input size Size of the block
 * @input dump_file Dump to write to.
 *
 * @return Returns file position to size field if successful,
 * otherwise it returns 0
 *
 */
static size_t 
write_new_block_header(uint32_t size, RmdDumpFile *dump_file)
{
   /* Variables to get time */
   OSTICK ticks;
   OSTICK uticks;
   OSTICK abs_time;
   OSTICK ref_ticks;
   OSTICK ref_uticks;
   OSTICK seconds;
   OSTICK microseconds;
   OSTICK tick_len;

   /* structure to hold block header data in memory. This is written to file as
    * binary data at the end of the function */
   struct IFFChunkBLHD ch_BLHD;

   RmdFileStatus ret;
   long size_file_position;

#ifdef __RAMLOG_TRACE__
   ramlog_printf("RMD FILE: write_new_block_header()\n");
#endif

   tick_len = system_tick();

   if (tick_len == 0)
      tick_len = 1000;

   /* Get tick */
   ticks = get_systime(&uticks);

   /* Get abs. time */
   get_abstime(&abs_time, &ref_ticks, &ref_uticks);

   if (uticks < ref_uticks)
   {
      uticks += tick_len;
      ticks--;
   }

   {
      OSTICK delta = (ticks - ref_ticks);
      OSTICK mdelta = delta / 1000000;
      OSTICK rdelta = delta % 1000000;
      OSTICK tmp = rdelta * tick_len + uticks - ref_uticks;
      seconds = abs_time + mdelta * tick_len + tmp / 1000000;
      microseconds = tmp % 1000000;
   }

   /* Dump header */
   size_file_position = write_group_header(TAG_FORM, size, TAG_BLOC, dump_file);
   if (0 == size_file_position)
      return 0;

   /* Use htonl for data to get correct endian */
   ch_BLHD.chd.id = htonl(TAG_BLHD);
   ch_BLHD.chd.size = htonl(sizeof(struct IFFChunkBLHD)
            - sizeof(struct IFFChunkHeader));
   ch_BLHD.dump_id = dump_file->dump_id;
   ch_BLHD.block_no = dump_file->block_number++;
   ch_BLHD.clock = seconds;
   ch_BLHD.microseconds = microseconds;

   ret = write_memory((uint32_t *) &ch_BLHD, sizeof(ch_BLHD), dump_file);

   if (ret != RMD_FILE_SUCCESS)
      return 0;

   return size_file_position;
}

/**
 * @function compression_init
 *
 * @brief Initialises the compression algorithm.
 *
 * @long Initialises the compression algorithm. This is called
 * internally from higher level functions in the implementation
 *
 * @input dump_file Dump to work on
 * @output dump_file->compression->zstrm Compression stream object for consecutive calls to
 * the compression algorithm.
 * @output dump_file->compression->zlib_buffers Working buffers for the compression algorithm
 * @output dump_file->compressed_data_size Data counter for the compression algorithm
 *
 * @return Returns RMD_FILE_SUCCESS if successful , otherwise it returns
 * RMD_MEM_ERROR if there were a memory allocation error, RMD_Z_STREAM_ERROR if
 * there were an error initialising the z_stream object or RMD_FILE_ERROR if there
 * were an unforeseen error.
 *
 */
static RmdFileStatus
compression_init(uint32_t window_bits, uint32_t mem_level,
         RmdDumpFile *dump_file)
{
   int zlib_ret;

#ifdef __RAMLOG_TRACE__
   ramlog_printf("RMD_FILE: compression_init()\n");
#endif
   /*
    * Allocate memory for compression objects, use alloc() instead of malloc()
    * because the Heap is not SMP safe for OSE 5.4
    */
   dump_file->compression->zlib_buffers = (struct ZlibBuffers *) alloc(
            sizeof(struct ZlibBuffers), 0);
   /* No need to check for null pointer, alloc won't return if allocation fails
    if (NULL == dump_file->compression->zlib_buffers)
    {
    #ifdef __RAMLOG_TRACE__
    ramlog_printf(
    "RMD_FILE: compression_init(): malloc(zlibbuffers) failed\n");
    #endif
    return RMD_FILE_MEM_ERROR;
    }
    */

   dump_file->compression->zstrm = (z_stream *) alloc(sizeof(z_stream), 0);
   /* No need to check for null pointer, alloc won't return if allocation fails
    if (NULL == dump_file->compression->zstrm)
    {
    free previously allocated memory
    free_buf(dump_file->compression->zlib_buffers);
    #ifdef __RAMLOG_TRACE__
    ramlog_printf(
    "RMD_FILE: compression_init(): malloc(z_stream) failed\n");
    #endif
    return RMD_FILE_MEM_ERROR;
    }
    */

   /* Memory functions for dynamic memory in zlib */
   dump_file->compression->zstrm->zalloc = &zlib_alloc;
   dump_file->compression->zstrm->zfree = &zlib_free;
   dump_file->compression->zstrm->opaque = (voidpf) 0;

   /* Memory requirements for deflate:
    * 2 ^ (window_bits + 2) + 2 ^ (mem_level+9)
    * Default (window_bits = 15, mem_level = 8) -> ~264kB
    * (1 << (windowBits+2)) + (1 << (memLevel+9)) + a few KB for small
    * objects
    */
   zlib_ret
            = deflateInit2(dump_file->compression->zstrm, Z_BEST_SPEED, Z_DEFLATED,
                     window_bits, mem_level,
                     Z_DEFAULT_STRATEGY);

   /* Init OK? */
   if (zlib_ret != Z_OK)
   {
      /*
       *  deflateInit2 returns Z_OK if success, Z_MEM_ERROR if there was not enough
       * memory, Z_STREAM_ERROR if a parameter is invalid (such as an invalid
       * method). msg is set to null if there is no error message.  deflateInit2 does
       * not perform any compression: this will be done by deflate().
       */

      /* free previously allocated memory */
      free_buf((union SIGNAL **) dump_file->compression->zlib_buffers);
      free_buf((union SIGNAL **) dump_file->compression->zstrm);
#ifdef __RAMLOG_TRACE__
      ramlog_printf(
               "RMD_FILE: compression_init(): deflateInit2() failed with code %d\n",
               zlib_ret);
#endif

      /* decide what error code to return */
      if (Z_MEM_ERROR == zlib_ret)
         return RMD_FILE_MEM_ERROR;

      if (Z_STREAM_ERROR == zlib_ret)
         return RMD_FILE_Z_STREAM_ERROR;

      /* Should not get here, but return the general error if we do */
      return RMD_FILE_ERROR;
   }
   /* Tell zlib where to find the cleared output buffer
    and what size it is. */
   dump_file->compression->zstrm->avail_out = ZLIB_COMPR_OUT_SIZE;
   dump_file->compression->zstrm->next_out
            = (Bytef *) dump_file->compression->zlib_buffers->out_buf;

   /*
    * ZLib set up correctly, reset the compressed_data_size variable
    */
   dump_file->sizes.compressed_data_size = 0;

   return RMD_FILE_SUCCESS;
}

/**
 * @function write_data_compressed_reset
 *
 * @brief Resets the compression algorithm for new data block.
 *
 * @long Resets the compression algorithm for new data block. This is called
 * internally from higher level functions in the implementation
 *
 * @input dump_file Dump to work on
 *
 * @return Returns RMD_FILE_SUCCESS if successful , otherwise it returns
 * RMD_Z_STREAM_ERROR if there were an error with the z_stream object
 *
 */
static RmdFileStatus
write_data_compressed_reset(RmdDumpFile *dump_file)
{
   RmdFileStatus ret;

   ret = deflateReset(dump_file->compression->zstrm);

   /* Init OK? */
   if (ret != Z_OK)
   {
      /*
       * deflateReset returns Z_OK if success, or Z_STREAM_ERROR if the source
       * stream state was inconsistent (such as zalloc or state being NULL).
       */

      return RMD_FILE_Z_STREAM_ERROR;
   }
   /* Tell zlib where to find the cleared output buffer
    and what size it is. */
   dump_file->compression->zstrm->avail_out = ZLIB_COMPR_OUT_SIZE;
   dump_file->compression->zstrm->next_out
            = (Bytef *) dump_file->compression->zlib_buffers->out_buf;

   //   dump_file->sizes.zdata_size = sizeof(struct IFFChunkZDEF) + 3 * sizeof(uint32_t);

   return RMD_FILE_SUCCESS;
}

/**
 * @function refill_input_buffer
 * @brief Fills the input buffer for zlib compression
 */

static uint32_t
refill_input_buffer(z_stream *zstrm, OSADDRESS start_address, uint32_t length,
         uint32_t left_to_proc, RmdDumpFile *dump_file)
{
   uint32_t pad = ALIGN_TO(length, sizeof(uint32_t)) - length;
   uint32_t bytes_left = left_to_proc;
   uint32_t non_pad_left = 0;

   /* Is there more data left to process than can fit in input buffer? */
   if (bytes_left > ZLIB_COMPR_IN_SIZE)
   {
      /* If so, limit the amount of data to process
       in this pass to size of the input buffer. */
      zstrm->avail_in = ZLIB_COMPR_IN_SIZE;
   }
   else
   {
      /* If not, tell zlib how much data there is left to process. */
      zstrm->avail_in = bytes_left;
   }

   if (zstrm->avail_in > 0)
   {
      /* Is there only padding left? */
      if (bytes_left <= pad)
      {
         /* If so, clear the bytes that represent
          the padding in the zlib input buffer. */
         (void) memset((void *) dump_file->compression->zlib_buffers->in_buf,
                  0, zstrm->avail_in);
      }
      else
      {
         /* There is at least some non-padding left. */
         non_pad_left = bytes_left - pad;

         /* Is some of the data to copy in this pass actually padding? */
         if (zstrm->avail_in > non_pad_left)
         {
            /* Copy data from memory, up to the padding,
             to the zlib input buffer. */
            (void) memmove(
                     (void *) dump_file->compression->zlib_buffers->in_buf,
                     (const void *) (start_address + (length - non_pad_left)),
                     non_pad_left);
            /* The rest is padding, so clear the bytes that represent
             the padding in the zlib input buffer. */
            (void) memset(
                     (void *) (dump_file->compression->zlib_buffers->in_buf
                              + non_pad_left), 0, zstrm->avail_in
                              - non_pad_left);
         }
         else
         {
            /* If not, copy data to compress to ramdump zlib input buffer. */
            (void) memmove(
                     (void *) dump_file->compression->zlib_buffers->in_buf,
                     (const void *) (start_address + (length - non_pad_left)),
                     zstrm->avail_in);
         }
      }

      /* Tell zlib where the input buffer is located. */
      zstrm->next_in = (Bytef *) dump_file->compression->zlib_buffers->in_buf;

      /* Update byte count left to process. */
      bytes_left -= zstrm->avail_in;

   }
   return bytes_left;
}

static RmdFileStatus
empty_output_buffer(z_stream *zstrm, RmdDumpFile *dump_file)
{
   RmdFileStatus ret = RMD_FILE_SUCCESS;
   uint32_t size = 0;
   uint32_t bytes_written = 0;

   /* Number of bytes in the output buffer. */
   size = ZLIB_COMPR_OUT_SIZE - zstrm->avail_out;

   if (size > 0)
   {
      /* Write compressed data from output buffer to dump_file. */
      bytes_written = fwrite(dump_file->compression->zlib_buffers->out_buf, 1,
               size, dump_file->file);
      dump_file->sizes.compressed_data_size += size;

      if (bytes_written != size)
      {
#ifdef __RAMLOG_TRACE__
         ramlog_printf(
                  "RMD_FILE: empty_output_buffer(): fwrite() failed with error %d\n",
                  ferror(dump_file->file));
#endif
         if (ferror(dump_file->file))
            ret = RMD_FILE_WRITE_ERROR;

         ret = RMD_FILE_WRITE_SHORT_ERROR;
      }
      else
      {
         /* Tell zlib where to find the cleared output buffer. */
         zstrm->next_out
                  = (Bytef *) dump_file->compression->zlib_buffers->out_buf;
         zstrm->avail_out = ZLIB_COMPR_OUT_SIZE;

         ret = RMD_FILE_SUCCESS;
      }
   }
   else
   {
      ret = RMD_FILE_SUCCESS;
   }

   return ret;
}
/**
 * @function write_data_compressed
 *
 * @brief Compresses and writes data from memory to dump_file.
 *
 * @long Compresses and writes data from memory to dump_file. This is called
 * internally from higher level functions in the implementation
 *
 * @input start_address Address to get data to compress and write to dump_file
 * @ input data_len Length of data
 * @input dump_file Dump to work on
 *
 * @return Returns RMD_FILE_SUCCESS if successful , otherwise it returns
 * RMD_MEM_ERROR if there were a memory allocation error,
 * RMD_Z_STREAM_ERROR if there were an error in the deflate (),
 * RMD_FILE_Z_STREAM_END if there were an unexpected end of data to the compressor,
 * RMD_FILE_WRITE_ERROR if an error occured with file writing,
 * RMD_FILE_WRITE_ERROR_SHORT if the data written is too short or
 * RMD_FILE_ERROR if there were an unforeseen error.
 *
 */
static RmdFileStatus
write_data_compressed(OSADDRESS start_address, uint32_t data_len,
         RmdDumpFile *dump_file)
{
   uint32_t left_to_proc;
   RmdFileStatus ret = RMD_FILE_SUCCESS;

#ifdef __RAMLOG_TRACE__
   ramlog_printf(
            "RMD_FILE: write_data_compressed():\n\tstart_address = 0x%x\n\tdata_len = %d\n",
            start_address, data_len);
#endif
   /* Set the number of bytes to compress,
    aligned up to multiple of 4 bytes. */
   left_to_proc = ALIGN_TO(data_len, sizeof(uint32_t));

   /* Initialise stream input and output data */
   dump_file->compression->zstrm->avail_in = 0;
   dump_file->compression->zstrm->next_in
            = (Bytef *) dump_file->compression->zlib_buffers->in_buf;

   /* Loop until there is no data left to hand over
    to zlib and zlib has no data left to process
    and zlib output buffer is not full */
   while (left_to_proc > 0 || dump_file->compression->zstrm->avail_in > 0
            || dump_file->compression->zstrm->avail_out == 0)
   {
      /* Is the zlib input buffer empty and
       is there any data left to process? */
      if (dump_file->compression->zstrm->avail_in == 0 && left_to_proc > 0)
      {
         /* If so, the input buffer needs a refill. */
         left_to_proc = refill_input_buffer(dump_file->compression->zstrm,
                  start_address, data_len, left_to_proc, dump_file);
      }

      /* Is the output buffer full? */
      if (dump_file->compression->zstrm->avail_out == 0)
      {
         /* If so, the output buffer needs to be emptied
          by writing pending output to dump_file. */
         ret = empty_output_buffer(dump_file->compression->zstrm, dump_file);

         /* Did writing of data fail? */
         if (ret != RMD_FILE_SUCCESS)
         {
#ifdef __RAMLOG_TRACE__
            ramlog_printf(
                     "RMD_FILE: write_data_compressed(): empty_output_buffer() Failed with code %d\n",
                     ret);
#endif
            /* If so, stop processing data. */
            break; /* while (left_to_proc > 0 || dump_file->compression->zstrm->avail_in > 0
             || dump_file->compression->zstrm->avail_out == 0) */
         }
      }

      /* Compress data in input buffer. */
      ret = deflate(dump_file->compression->zstrm, Z_NO_FLUSH);

      /* Did something go wrong during compression? */
      if (ret != Z_OK)
      {
         /* If so, stop processing data. */
#ifdef __RAMLOG_TRACE__
         ramlog_printf(
                  "RMD_FILE: write_data_compressed(): deflate() Failed with code %d\n",
                  ret);
#endif
         ret = RMD_FILE_ZLIB_ERROR;
         break; /* while (left_to_proc > 0 || dump_file->compression->zstrm->avail_in > 0
          || dump_file->compression->zstrm->avail_out == 0) */
      }
   } /* while ((data_len - offset) > 0) */

#ifdef __RAMLOG_TRACE__
   ramlog_printf("RMD_FILE: write_data_compressed(): No more input\n");
#endif

   if (ret == RMD_FILE_SUCCESS)
   {
      dump_file->sizes.uncompressed_data_size
               += ALIGN_TO(data_len, sizeof(uint32_t));

      /* Does the output buffer contain data? */
      if (dump_file->compression->zstrm->avail_out < ZLIB_COMPR_OUT_SIZE)
      {
         /* If so, the output buffer needs to be emptied
          by writing pending output to dump_file. */
         ret = empty_output_buffer(dump_file->compression->zstrm, dump_file);
      }
   }

   return ret;
}

static RmdFileStatus
write_data_compressed_end(RmdDumpFile *dump_file)
{
   RmdFileStatus ret = RMD_FILE_SUCCESS;
   int zlib_status = Z_OK;

#ifdef __RAMLOG_TRACE__
   ramlog_printf("RMD FILE: write_data_compressed_end()\n");
#endif
   /* Tell zlib where to find the cleared input buffer. */
   dump_file->compression->zstrm->avail_in = 0;
   dump_file->compression->zstrm->next_in
            = (Bytef *) dump_file->compression->zlib_buffers->in_buf;

   do
   {
      zlib_status = deflate(dump_file->compression->zstrm, Z_FINISH);

      /* Did something go wrong during compression? */
      if (zlib_status != Z_OK && zlib_status != Z_STREAM_END)
      {

         /* If so, stop processing data. */
         ret = RMD_FILE_ZLIB_ERROR;
         break;
      }

      /* Does the output buffer contain any data? */
      if ((ZLIB_COMPR_OUT_SIZE - dump_file->compression->zstrm->avail_out) > 0)
      {
         /* If so, the output buffer needs to be emptied
          by writing pending output to dump_file. */
         ret = empty_output_buffer(dump_file->compression->zstrm, dump_file);

         /* Did writing of data fail? */
         if (ret != RMD_FILE_SUCCESS)
         {
            /* If so, stop processing data. */
            break;
         }
      }

   } while (zlib_status == Z_OK);

   return ret;
}
/**
 * @function write_data
 *
 * @brief Interface function for write_memory.
 *
 * @input address Address to the memory that is beeing dumped
 * @input data_len Length of memory data to be dumped in the following call
 * @input dump_file Dump to write to
 *
 * @return Returns RMD_FILE_SUCCESS if successful , otherwise it returns
 * RMD_FILE_WRITE_ERROR if fwrite fails,
 * RMD_FILE_WRITE_SHORT_ERROR if the data written is shorter than expected,
 *
 * @seealso write_memory
 * @seealso write_data_compressed
 */
static RmdFileStatus
write_data(OSADDRESS start_address, uint32_t data_len, RmdDumpFile *dump_file)
{
   RmdFileStatus status;
   uint32_t padding;

   status = write_memory((uint32_t *) start_address, data_len, dump_file);
   padding = get_padding(data_len);
   write_padding(padding, 0, (const RmdDumpFile *) dump_file);

   dump_file->sizes.uncompressed_data_size += data_len + padding;

   return status;
}

/**
 * @function dump_memory_block_header
 *
 * @brief Writes header for memory block and initialises dump_file for memory data.
 *
 * @long Writes header for memory block and initialises dump_file for memory data.
 * This is called externally from the code that is implementing the actual
 * dumping.
 *
 * @input address Address to the memory that is beeing dumped
 * @input data_len Length of memory data to be dumped in the following call
 * @input description String with description of error
 * @input dump_file Dump to write to
 *
 * @output dump_file->file_pos.*
 *
 * @return Returns number of bytes in header if successful,
 * otherwise it returns 0
 */
static uint32_t
dump_memory_block_header(OSADDRESS address, uint32_t data_len,
         const char *description, RmdDumpFile *dump_file)
{
   RmdFileStatus ret;

   /* struct to hold memory header data */
   struct IFFChunkMHD_ ch_MHD_;

   /* ftell return value stored in this variable.
    * long int is the type returned by ftell */
   long int current_file_position;

   int padding;
   int desc_size;

#ifdef __RAMLOG_TRACE__
   ramlog_printf("RMD FILE: rmd_file_dump_memory_block_start()\n");
#endif

   desc_size = strnlen(description, RMD_MAX_STRING_LENGTH) + 1;
   padding = get_padding(desc_size);

   /* Size updated through dump_file->file_pos.top_form_size_pos in update_size_all() */
   dump_file->file_pos.top_form_size_pos = write_new_block_header(0, dump_file);
   if (0 == dump_file->file_pos.top_form_size_pos)
   {
      return 0;
   }

   dump_file->file_pos.form_size_pos = write_group_header(TAG_FORM,
            sizeof(struct IFFHeaderFORM) - sizeof(struct IFFChunkHeader),
            TAG_MBL_, dump_file);
   if (0 == dump_file->file_pos.form_size_pos)
   {
      return 0;
   }

   ret = write_desc(description, desc_size, padding, dump_file);
   if (ret != RMD_FILE_SUCCESS)
   {
      return 0;
   }

   current_file_position = ftell(dump_file->file);
   if (EOF == current_file_position)
   {
      /* ftell error */
      return 0;
   }
   dump_file->file_pos.content_length_pos = current_file_position
            + MEM_CONTENT_LENGTH_OFFSET;

   ch_MHD_.chd.id = htonl(TAG_MHD_);
   ch_MHD_.chd.size = htonl(sizeof(struct IFFChunkMHD_)
            - sizeof(struct IFFChunkHeader));
   ch_MHD_.start_address = address;
   ch_MHD_.length = data_len;

   ret = write_memory((uint32_t *) &ch_MHD_, sizeof(struct IFFChunkMHD_),
            dump_file);
   if (ret != RMD_FILE_SUCCESS)
   {
      return 0;
   }

   if (dump_file->compression != NULL)
   {
      /* struct to hold compression header data */
      struct IFFChunkZDEF ch_ZDEF;

      /* Loop variable */
      int i;

      /* Group header */
      dump_file->file_pos.zdata_size_pos = write_group_header(TAG_FORM,
               COMPRESSION_HEADER_SIZE, TAG_ZDAT, dump_file);
      if (0 == current_file_position)
      {
         return 0;
      }

      /* Compression header */
      current_file_position = ftell(dump_file->file);
      if (EOF == current_file_position)
      {
         /* ftell error */
         return 0;
      }
      dump_file->file_pos.zdef_image_length_pos = current_file_position
               + ZDEF_IMAGE_LENGTH_OFFSET;

      ch_ZDEF.chd.id = htonl(TAG_ZDEF);
      ch_ZDEF.chd.size = htonl(sizeof(struct IFFChunkZDEF)
               - sizeof(struct IFFChunkHeader));

      ch_ZDEF.compression = COMPRESS_ZLIB_Z77;
      for (i = 0; i < 4; i++)
      {
         ch_ZDEF.compr_parameters[i] = 0;
      }
      ch_ZDEF.image_length = 0; /* Will be updated later */
      ch_ZDEF._reserved = 0;

      ret = write_memory((uint32_t *) &ch_ZDEF, sizeof(struct IFFChunkZDEF),
               dump_file);
      if (ret != RMD_FILE_SUCCESS)
      {
         return 0;
      }

   }
   current_file_position = ftell(dump_file->file);
   if (EOF == current_file_position)
   {
      return 0;
   }

   ret = write_chunk_header(TAG_DATA, 0, dump_file);
   if (ret != RMD_FILE_SUCCESS)
   {
      return 0;
   }
   dump_file->file_pos.data_size_pos = current_file_position + DATA_SIZE_OFFSET;

   return MHD_SIZE + desc_size + padding;
}

/**
 * @function dump_request_header
 *
 * @brief Writes header for signal block and initialises dump_file for memory data.
 *
 * @long Writes header for memory block and initialises dump_file for memory data.
 * This is called externally from the code that is implementing the actual
 * dumping.
 *
 * @input address Address to the memory that is beeing dumped
 * @input data_len Length of memory data to be dumped in the following call
 * @input description String with description of error
 * @input dump_file Dump to write to
 *
 * @output dump_file->file_pos.*
 *
 * @return Returns number of bytes in header if successful,
 * otherwise it returns 0
 */
static uint32_t
dump_request_header(OSADDRESS request, RmdDumpFile *dump_file)
{
   RmdFileStatus ret = RMD_FILE_SUCCESS;
   struct IFFChunkSGHD ch_SGHD;
   long current_file_position;

   dump_file->file_pos.top_form_size_pos = write_new_block_header(0, dump_file);
   if (0 == dump_file->file_pos.top_form_size_pos)
   {
      return 0;
   }

   dump_file->file_pos.form_size_pos = write_group_header(TAG_FORM,
            sizeof(struct IFFHeaderFORM) - sizeof(struct IFFChunkHeader),
            TAG_SGBL, dump_file);
   if (0 == dump_file->file_pos.form_size_pos)
   {
#ifdef __RAMLOG_TRACE__
      ramlog_printf(
               "RMD_FILE: rmd_file_dump_request_write_headers(): write_group_header() failed with code %d\n",
               ret);
#endif
      return 0;
   }

   ch_SGHD.chd.id = htonl(TAG_SGHD);
   ch_SGHD.chd.size = htonl(sizeof(struct IFFChunkSGHD)
            - sizeof(struct IFFChunkHeader));
   /* First 32 bit word is signo */
   ch_SGHD.signo = *(uint32_t *) request;
   /* Default value. Will be overwritten when endmark is dumped. */
   ch_SGHD.status = 0xBADC0DE;
   ch_SGHD.length = 0;

   current_file_position = ftell(dump_file->file);
   if (EOF == current_file_position)
   {
#ifdef __RAMLOG_TRACE__
      ramlog_printf(
               "RMD_FILE: rmd_file_dump_request_write_headers(): ftell() failed");
#endif
      return 0;
   }

   ret = write_memory((uint32_t *) &ch_SGHD, sizeof(struct IFFChunkSGHD),
            dump_file);
   if (ret != RMD_FILE_SUCCESS)
   {
#ifdef __RAMLOG_TRACE__
      ramlog_printf(
               "RMD_FILE: rmd_file_dump_request_write_headers(): write_memory() failed with code %d\n",
               ret);
#endif
      return 0;
   }

   dump_file->file_pos.content_status_pos = current_file_position
            + CONTENT_STATUS_OFFSET;
   dump_file->file_pos.content_length_pos = current_file_position
            + CONTENT_LENGTH_OFFSET;

   if (dump_file->compression != NULL)
   {
      /* struct to hold compression header data */
      struct IFFChunkZDEF ch_ZDEF;

      /* Loop variable */
      int i;

      /* Group header */
      dump_file->file_pos.zdata_size_pos = write_group_header(TAG_FORM, 11111,
               TAG_ZDAT, dump_file);
      if (0 == dump_file->file_pos.zdata_size_pos)
      {
#ifdef __RAMLOG_TRACE__
         ramlog_printf(
                  "RMD_FILE: rmd_file_dump_request_write_headers(): write_group_header() failed with code %d\n",
                  ret);
#endif
         return 0;
      }

      /* Compression header */
      current_file_position = ftell(dump_file->file);
      if (EOF == current_file_position)
      {
#ifdef __RAMLOG_TRACE__
         ramlog_printf(
                  "RMD_FILE: rmd_file_dump_request_write_headers(): ftell() failed");
#endif
         return 0;
      }

      ch_ZDEF.chd.id = htonl(TAG_ZDEF);
      ch_ZDEF.chd.size = htonl(sizeof(struct IFFChunkZDEF)
               - sizeof(struct IFFChunkHeader));

      ch_ZDEF.compression = COMPRESS_ZLIB_Z77;
      for (i = 0; i < 4; i++)
      {
         ch_ZDEF.compr_parameters[i] = 0;
      }
      ch_ZDEF.image_length = 0; /* Will be updated later */
      ch_ZDEF._reserved = 0;

      ret = write_memory((uint32_t *) &ch_ZDEF, sizeof(struct IFFChunkZDEF),
               dump_file);
      if (ret != RMD_FILE_SUCCESS)
      {
#ifdef __RAMLOG_TRACE__
         ramlog_printf(
                  "RMD_FILE: rmd_file_dump_request_write_headers(): write_memory() failed with code %d\n",
                  ret);
#endif
         return 0;
      }

      dump_file->file_pos.zdef_image_length_pos = current_file_position
               + ZDEF_IMAGE_LENGTH_OFFSET;
   }

   current_file_position = ftell(dump_file->file);
   if (EOF == current_file_position)
   {
#ifdef __RAMLOG_TRACE__
      ramlog_printf(
               "RMD_FILE: rmd_file_dump_request_write_headers(): ftell() failed");
#endif
      return 0;
   }

   ret = write_chunk_header(TAG_DATA, 0, dump_file);
   if (ret != RMD_FILE_SUCCESS)
   {
#ifdef __RAMLOG_TRACE__
      ramlog_printf(
               "RMD_FILE: rmd_file_dump_request_write_headers(): write_chunk_header() failed with code %d\n",
               ret);
#endif
      return 0;
   }
   dump_file->file_pos.data_size_pos = current_file_position + DATA_SIZE_OFFSET;

   return SGHD_SIZE;
}
