/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2005-2006 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

/**
 * @file rmd_fileformat.h
 *
 * @brief Internal header file that contains the rmd file format.
 */

#ifndef _RMD_FILE_H
#define _RMD_FILE_H

#include "ose.h"
#include "stdio.h"
#include "stdint.h"
#include "zlib.h"

#ifdef __cplusplus
extern "C"
{
#endif

/**
 * @type FilePositions
 *
 * @brief This struct holds the file positions to the locations in the IFF file
 * where the sizes are located. This is a part of the dumpstruct.
 *
 * @brief This struct holds the file positions to the locations in the IFF file
 * where the sizes are located. This is a part of the dumpstruct. This struct is
 * built during file writing by instead of writing the size data, the current file
 * position is stored here by using ftell and a place holder value is written to
 * file. When the size is known this is stored in the Sizes struct and is written
 * to the file at the position in this structure.
 *
 * @seealso RmdDumpFile
 * @seealso Sizes
 *
 */
struct FilePositions
{
   /* File positions to seek back to
    * Defined to be long int as input to fseek
    */
/*   long int dump_size_pos;  = DUMP_SIZE_POS */
/*   long int block_list_size_pos;  = BLOCK_LIST_SIZE_POS */
   long int top_form_size_pos;
   long int form_size_pos;
   long int content_status_pos;
   long int content_length_pos;
   long int zdata_size_pos;
   long int zdef_image_length_pos;
   long int data_size_pos;

};

/**
 * @type Sizes
 *
 * @brief This struct holds the sizes of the different blocks in the IFF file. This
 * is a part of the RmdDumpFile.
 *
 * @long This struct holds the sizes of the different blocks in the IFF file. This
 * is a part of the RmdDumpFile. The sizes are updated during the writing of the file
 * and are used together with the positions in the FilePositions struct to seek
 * back in the file and update the sizes.
 *
 * @seealso RmdDumpFile
 * @seealso FilePositions
 *
 */
struct Sizes
{
   uint32_t dump_size;
   uint32_t block_list_size;
   uint32_t compressed_data_size;
   uint32_t uncompressed_data_size;
};

#define ZLIB_COMPR_OUT_SIZE   (0x1000UL) /* 4K */
#define ZLIB_COMPR_IN_SIZE    (0x4000UL) /* 16K */

/**
 * @type ZlibBuffers
 *
 * @brief This struct holds the in/out buffers in the ZLib interface
 *
 * @seealso compression_struct
 *
 */
struct ZlibBuffers
{
   uint8_t out_buf[ZLIB_COMPR_OUT_SIZE];
   uint8_t in_buf[ZLIB_COMPR_IN_SIZE];
};

/**
 * @type compression_struct
 *
 * @brief This struct holds compression information for the current dump_file.
 *
 * @long This struct holds compression information for the current dump_file.
 * The zstrm object passed to all ZLib functions holding all information about
 * the compression algorithm. The zlib_buffers object are the working in/out
 * buffers in the ZLib interface. These two objects are allocated dynamically
 * by the function rmd_start_dump
 *
 * @seealso RmdDumpFile
 * @seealso rmd_start_dump
 * @seealso ZLibBuffers
 *
 */
struct compression_struct
{
   z_stream *zstrm;
   struct ZlibBuffers *zlib_buffers;
};

enum block_type
{
   BLOCK_NONE, BLOCK_SIGNAL

};

/**
 * @type RmdDumpFile
 *
 * @brief This struct holds information about the current dump_file. This is passed to
 * all functions in the interface.
 *
 * @long This struct holds information about the current dump_file. This is passed to
 * all functions in the interface. The structure is created by calling
 * rmd_start_dump and destroyed by calling rmd_end_dump. The actual data allocation
 * and deallocation must be handled outside this interface.
 *
 * @seealso rmd_start_dump
 * @seealso rmd_end_dump
 * @seealso FilePositions
 * @seealso Sizes
 * @seealso compression_struct
 *
 */
typedef struct RmdDumpFile
{
   FILE *file;
   struct FilePositions file_pos;
   struct Sizes sizes;

   enum block_type block_type;

   uint32_t dump_id;
   uint32_t block_number;
   struct compression_struct *compression;
   char file_name[1];
} RmdDumpFile;

#define RMD_MAX_STRING_LENGTH 1024 /* used for strn* functions */

/*
 * Status of the RMD file writer
 */
typedef enum RmdFileStatus
{
   RMD_FILE_SUCCESS = 0,

   /* General errors */
   RMD_FILE_ERROR = -1,
   RMD_FILE_MEM_ERROR = -2,
   RMD_FILE_BLOCK_ERROR = -3,

   /* ZLIB errors */
   RMD_FILE_ZLIB_ERROR = -4,
   RMD_FILE_Z_STREAM_ERROR = -5,
   RMD_FILE_Z_STREAM_END = -6,

   /* File error */
   RMD_FILE_SYSTEM_ERROR = -7,
   RMD_FILE_WRITE_ERROR = -8,
   RMD_FILE_OPEN_ERROR = -9,
   RMD_FILE_SEEK_ERROR = -10,
   RMD_FILE_TELL_ERROR = -11,
   RMD_FILE_FLUSH_ERROR = -12,
   RMD_FILE_WRITE_SHORT_ERROR = -13
} RmdFileStatus;

/******************************************************************************
 *
 * Function declarations of external interface
 *
 *****************************************************************************/

/**
 * @function rmd_start_dump
 *
 * @brief Sets up the dump_file structure, opens the dump_file file and writes the
 * file head.
 *
 * @long Sets up the dump_file structure, opens the dump_file file and writes the
 * file head. This is the first function to call in the interface.
 * Implements the same functionality for RMD as
 * @input filename Name of file including full path.
 * @input dump_id ID of the dump_file
 * @input dump_file Dump to work on
 *
 * @return Returns RMD_FILE_SUCCESS if successful , otherwise it returns
 * RMD_FILE_WRITE_ERROR if an error occurred with file writing,
 * RMD_FILE_WRITE_ERROR_SHORT if the data written is too short,
 * or RMD_FILE_ERROR if there were an unforeseen error.
 *
 * @seealso rmd_end_dump
 */
RmdFileStatus
rmd_file_start_dump(const char *filename, uint32_t dump_id, uint32_t use_compression,
         uint32_t window_bits, uint32_t mem_level, RmdDumpFile *dump_file);

/**
 * @function rmd_end_dump
 *
 * @brief Closes the dump_file file and frees up memory allocated. Should always be
 * called when a dump_file is finished.
 *
 * @long Closes the dump_file file and frees up memory allocated. Should always be
 * called when a dump_file is finished. The function frees all memory allocated in
 * rmd_start_dump
 *
 * @input dump_file Dump to work on
 *
 * @seealso rmd_start_dump
 *
 */
void
rmd_file_end_dump(RmdDumpFile *dump_file);

/**
 * @function rmd_dump_error_block
 *
 * @brief Writes an error block to the dump_file.
 *
 * @long Writes an error block to the dump_file. This is called externally from the
 * code that is implementing the actual dumping.
 *
 * @input error_code Error code.
 * @input extra Extra parameter.
 * @input cpid Current process (the process that is beeing dumped)
 * @input description String with description of error
 * @input dump_file Dump to write to
 *
 * @return Returns RMD_FILE_SUCCESS if successful , otherwise it
 * returns RMD_FILE_WRITE_ERROR if fwrite fails, RMD_FILE_WRITE_SHORT_ERROR
 * if the data written is shorter than expected, RMD_FILE_SEEK_ERROR if fseek
 * fails
 *
 */
RmdFileStatus
rmd_file_dump_error_block(const char *description, uint32_t user_called,
         uint32_t error_code, uint32_t extra, uint32_t cpid, RmdDumpFile *dump_file);

/**
 * @function rmd_file_dump_text_block
 *
 * @brief Writes a text block to the dump_file.
 *
 * @long Writes a text block to the dump_file. This is called externally from the
 * code that is implementing the actual dumping.
 *
 * @input description String with description of error
 * @input dump_file Dump to write to
 *
 * @return Returns RMD_FILE_SUCCESS if successful , otherwise it
 * returns RMD_FILE_WRITE_ERROR if fwrite fails, RMD_FILE_WRITE_SHORT_ERROR
 * if the data written is shorter than expected, RMD_FILE_SEEK_ERROR if fseek
 * fails
 *
 */
RmdFileStatus
rmd_file_dump_text_block(const char *description, RmdDumpFile *dump_file);

/**
 * @function rmd_dump_memory_block
 *
 * @brief Writes a memory block to dump_file including header. Uses compression
 * if dump_file->compression is set (!= NULL).
 *
 * @long Writes a memory block to dump_file including header. Uses compression
 * if dump_file->compression is set (!= 0). External function to the dump_file file
 * interface.
 *
 * @input block OSE Block id for which to dump memory
 * @input start_address Address to get data to compress and write to dump_file
 * @input dataLen Length of data
 * @input description String describing the block.
 * @input dump_file Dump to work on
 *
 * @return Returns RMD_FILE_SUCCESS if successful , otherwise it returns
 * RMD_FILE_MEM_ERROR if there were a memory allocation error,
 * RMD_FILE_Z_STREAM_ERROR if there were an error in the deflate (),
 * RMD_Z_STREAM_END if there were an unexpected end of data to the compressor,
 * RMD_FILE_WRITE_ERROR if an error occured with file writing,
 * RMD_FILE_WRITE_ERROR_SHORT if the data written is too short,
 * RMD_FILE_SEEK_ERROR if there were an error when seeking in the file
 * RMD_FILE_TELL_ERROR if there were an error when storing the current file position
 * or RMD_FILE_ERROR if there were an unforeseen error.
 *
 */
RmdFileStatus
rmd_file_dump_memory_block(PROCESS block, OSADDRESS data, uint32_t dataLen,
         const char *description, RmdDumpFile *dump_file);

/**
 * @function rmd_file_dump_request
 *
 * @brief Writes header for signal block and initialises dump_file for memory data.
 *
 * @long Writes header for signal block and initialises dump_file for memory data.
 * This is called externally from the code that is implementing the actual
 * dumping.
 *
 * @input request Pointer to the request signal that starts the transaction,
 * this signal will be written as the first signal in the signal block.
 * @input dump_file Dump to write to
 *
 * @return Returns RMD_FILE_SUCCESS if successful , otherwise it returns
 * RMD_FILE_WRITE_ERROR if fwrite fails,
 * RMD_FILE_WRITE_SHORT_ERROR if the data written is shorter than expected,
 * RMD_FILE_SEEK_ERROR if fseek fails
 *
 */
RmdFileStatus
rmd_file_dump_request(OSADDRESS sig_address, uint32_t sig_size, RmdDumpFile *dump_file);

/**
 * @function rmd_file_dump_reply
 *
 * @brief Writes signal to dump_file.
 *
 * @long Writes signal to dump_file. This is called externally from the code that
 * is implementing the actual dumping.
 *
 * @input sig Pointer to the request signal that starts the transaction
 * @input dump_file Dump to write to
 *
 * @return Returns RMD_FILE_SUCCESS if successful , otherwise it returns
 * RMD_FILE_WRITE_ERROR if fwrite fails,
 * RMD_FILE_WRITE_SHORT_ERROR if the data written is shorter than expected,
 * RMD_FILE_SEEK_ERROR if fseek fails
 * RMD_FILE_BLOCK_ERROR if the block is not initialised
 * or RMD_FILE_ERROR if an unforeseen error occured
 */
RmdFileStatus
rmd_file_dump_reply(OSADDRESS sig_address, uint32_t sig_size, RmdDumpFile *dump_file);

/**
 * @function rmd_file_dump_endmark
 *
 * @brief Finishes the writing of signals to dump_file.
 *
 * @long Finishes the writing of signals to dump_file. This is called externally
 * from the code that is implementing the actual dumping. Implements the same
 * functionality for RMD as ramdump_dump_endmark do for the FMD.
 *
 * @input emdmark Endmark signal. Set to NULL if the signal does not exist,
 * this will not be written to dump_file.
 * @input status Status code for the signal block. This is written to the dump_file.
 * @input dump_file Dump to write to
 *
 * @return Returns RMD_FILE_SUCCESS if successful , otherwise it returns
 * RMD_FILE_WRITE_ERROR if fwrite fails,
 * RMD_FILE_WRITE_SHORT_ERROR if the data written is shorter than expected,
 * RMD_FILE_SEEK_ERROR if fseek fails
 * RMD_FILE_BLOCK_ERROR if the block is not initialised
 * or RMD_FILE_ERROR if an unforeseen error occured
 */
RmdFileStatus
         rmd_file_dump_endmark(OSADDRESS sig_address, uint32_t sig_size, uint32_t status,
                  RmdDumpFile *dump_file);

#ifdef __cplusplus
}
#endif

#endif /* # ifndef _RMD_FILE_H */
