/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2005-2006 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

/**
 * @file rmd_file_private.h
 *
 * @brief Internal header file that contains the rmd file format.

 The following grammar describes chunks in the dump_file format.
 Chunks are aligned to 32 bits, but the dump_file can be saved in either
 big endian or little endian format.

 Chunk      ::=   ID Size uint32_t*
 ID      ::=   uint32_t
 Size    ::=   uint32_t
 Dump    ::=   'FORM' Size 'PMD ' (Vers | DumpId | BlockList)*
 Vers    ::=   'VERS' Size(4) uint32_t(<version>)
 DumpId  ::=   'DPID' Size(4) uint32_t(<dump_file id>)
 BlockList  ::=   'LIST' Size 'BLOC' (BlockForm)*
 BlockForm  ::=   'FORM' Size 'BLOC' (BlockChunk)*
 BlockChunk ::=   BlockHead | ErrorForm | TextForm | MemoryForm
 BlockHead  ::=   'BLHD' Size(16) uint32_t(<dump_file id>) uint32_t(<block no>)
 uint32_t(<seconds since Jan 1 1970, 00:00>)
 uint32_t(<additional microseconds>)
 ErrorForm  ::=   'FORM' Size 'ERBL' (Descr | ErrInfo)*
 Descr      ::=   'DESC' Size(<size of text>) <text>
 ErrInfo ::=   'ERIN' Size(16) uint32_t(<user_called>) uint32_t(<error_code>)
 uint32_t(<extra>) uint32_t(<curr_proc>)
 TextForm   ::=   'FORM' Size 'TXBL' (Descr)*
 MemoryForm ::=   'FORM' Size 'MBL ' (Descr | (MemDef (Data | Zdata)))*
 MemDef  ::=   'MHD' Size(8) uint32_t(<start>) uint32_t(<length>)
 Data    ::=   'DATA' Size(<length of data>) <data>
 Zdata      ::=   'FORM' Size 'ZDAT' Zdef Data
 Zdef    ::=   'ZDEF' Size(4 + <size of parameters>)
 uint32_t(<method>) <parameters>

 */
#ifndef RMD_FILE_PRIVATE_H_
#define RMD_FILE_PRIVATE_H_

#ifdef __cplusplus
extern "C"
{
#endif

#include "rmd_file.h"
#include "stdint.h"
#include "zlib.h"


/**
 * @macro ID
 *
 * @brief Turns four characters into an Iff Id
 *
 * @input a The first letter of the Iff Id
 * @input b The second letter of the Iff id
 * @input c The third letter of the Iff id
 * @input d The fourth letter of the Iff id
 *
 * @return Returns the letters as an Iff id.
 */
#define ID(a, b, c, d) ( ((uint32_t) (a) << 24) | ((uint32_t) (b) << 16) |\
                             ((uint32_t) (c) << 8) | (uint32_t) (d) )

/**
 * @macro RAMDUMP_TAG_FORM
 *
 * @brief The FORM iff id.
 */
#define TAG_FORM ID('F', 'O', 'R', 'M')

/**
 * @macro TAG_LIST
 *
 * @brief The LIST iff id.
 */
#define TAG_LIST ID('L', 'I', 'S', 'T')

/**
 * @macro TAG_PROP
 *
 * @brief The PROP iff id.
 */
#define TAG_PROP ID('P', 'R', 'O', 'P')

/**
 * @macro TAG_CAT_
 *
 * @brief The CAT_ iff id
 */
#define TAG_CAT_ ID('C', 'A', 'T', ' ')

/**
 * @macro TAG_FILL
 *
 * @brief The FILL iff id
 */
#define TAG_FILL ID(' ', ' ', ' ', ' ')

/**
 * @macro TAG_WILD
 *
 * @brief Used internally for matching chunks with wildcards
 */
#define TAG_WILD ID('*', '*', '*', '*')

/**
 * @macro TAG_PMD_
 *
 * @brief The PMD_ iff id.
 */
#define TAG_PMD_ ID('P', 'M', 'D', ' ')

/**
 * @macro TAG_BLOC
 *
 * @brief The BLOC iff id.
 */
#define TAG_BLOC ID('B', 'L', 'O', 'C')

/**
 * @macro TAG_DPID
 *
 * @brief The DPID iff id.
 */
#define TAG_DPID ID('D', 'P', 'I', 'D')

/**
 * @macro TAG_VERS
 *
 * @brief The VERS iff id.
 */
#define TAG_VERS ID('V', 'E', 'R', 'S')

/**
 * @macro TAG_BLHD
 *
 * @brief The BLHD iff id.
 */
#define TAG_BLHD ID('B', 'L', 'H', 'D')

/**
 * @macro TAG_DESC
 *
 * @brief The DESC iff id.
 */
#define TAG_DESC ID('D', 'E', 'S', 'C')

/**
 * @macro TAG_ERBL
 *
 * @brief The ERBL iff id.
 */
#define TAG_ERBL ID('E', 'R', 'B', 'L')

/**
 * @macro TAG_ERIN
 *
 * @brief The ERIN iff id.
 */
#define TAG_ERIN ID('E', 'R', 'I', 'N')

/**
 * @macro TAG_TXBL
 *
 * @brief The TXBL iff id.
 */
#define TAG_TXBL ID('T', 'X', 'B', 'L')

/**
 * @macro TAG_MBL_
 *
 * @brief The MBL_ iff id.
 */
#define TAG_MBL_ ID('M', 'B', 'L', ' ')

/**
 * @macro TAG_MHD_
 *
 * @brief The MHD_ iff id.
 */
#define TAG_MHD_ ID('M', 'H', 'D', ' ')

/**
 * @macro TAG_DATA
 *
 * @brief The DATA iff id.
 */
#define TAG_DATA ID('D', 'A', 'T', 'A')

/**
 * @macro TAG_ZDAT
 *
 * @brief The ZDAT iff id.
 */
#define TAG_ZDAT ID('Z', 'D', 'A', 'T')

/**
 * @macro TAG_ZDEF
 *
 * @brief The ZDEF iff id
 */
#define TAG_ZDEF ID('Z', 'D', 'E', 'F')

/**
 * @macro TAG_SGBL
 *
 * @brief The SGBL iff id
 */
#define TAG_SGBL ID('S', 'G', 'B', 'L')

/**
 * @macro TAG_SGHD
 *
 * @brief The SGHD iff id
 */
#define TAG_SGHD ID('S', 'G', 'H', 'D')

/**
 * @type IffId
 *
 * @brief Represents an iff chunk id.
 */
typedef uint32_t IffId;

/**
 * @type IFFChunkHeader
 *
 * @brief An IFF chunk header structure.
 *
 * @field id Iff id of the chunk.
 * @field size Size of the chunk minus the header.
 */
struct IFFChunkHeader
{
   IffId id;
   uint32_t size;
};

/**
 * @type IFFHeaderFORM
 *
 * @brief Represents a FORM chunk.
 */
struct IFFHeaderFORM
{
   struct IFFChunkHeader chd;
   IffId form_type;
};

/**
 * @type IFFHeaderLIST
 *
 * @brief Represents a LIST chunk
 */
struct IFFHeaderLIST
{
   struct IFFChunkHeader chd;
   IffId cont_type;
};

/**
 * @type IFFHeaderCAT_
 *
 * @brief Represents a CAT_ chunk
 */
struct IFFHeaderCAT_
{
   struct IFFChunkHeader chd;
   IffId cont_type;
};

/**
 * @type IFFHeaderPROP
 *
 * @brief Represents a PROP chunk.
 */
struct IFFHeaderPROP
{
   struct IFFChunkHeader chd;
   IffId form_type;
};

/**
 * @type IFFChunkVERS
 *
 * @brief Represents a VERS chunk.
 */
struct IFFChunkVERS
{
   struct IFFChunkHeader chd;
   uint32_t version;
};

/**
 * @type IFFChunkDESC
 *
 * @brief Represents a DESC chunk.
 */
struct IFFChunkDESC
{
   struct IFFChunkHeader chd;
   char descr[4];
};

/**
 * @type IFFChunkDPID
 *
 * @brief Represents a DPID chunk
 */
struct IFFChunkDPID
{
   struct IFFChunkHeader chd;
   uint32_t dump_id;
};

/**
 * @type IFFChunkBLHD
 *
 * @brief Represents a BLHD chunk.
 */
struct IFFChunkBLHD
{
   struct IFFChunkHeader chd;
   uint32_t dump_id;
   uint32_t block_no;
   uint32_t clock; /* Seconds since Jan 1, 1970, 00:00:00 */
   uint32_t microseconds;
};

/**
 * @type IFFChunkERIN
 *
 * @brief Represents an ERIN chunk.
 */
struct IFFChunkERIN
{
   struct IFFChunkHeader chd;
   uint32_t user_called;
   uint32_t error_code;
   uint32_t extra;
   uint32_t curr_proc;
};

/**
 * @type IFFChunkMHD_
 *
 * @brief Represents an MHD_ chunk.
 */
struct IFFChunkMHD_
{
   struct IFFChunkHeader chd;
   uint32_t start_address;
   uint32_t length;
};

/**
 * @type IFFChunkZDEF
 *
 * @brief Represents an ZDEF chunk.
 */
struct IFFChunkZDEF
{
   struct IFFChunkHeader chd;
   uint32_t compression;
   uint32_t compr_parameters[4];
   uint32_t image_length; /* Length of compressed image */
   uint32_t _reserved;
};

/**
 * @type IFFChunkSGHD
 *
 * @brief Represents an SGHD chunk.
 */
struct IFFChunkSGHD
{
   struct IFFChunkHeader chd;
   uint32_t signo;
   uint32_t status;
   uint32_t length;
};

union ChunkHeader
{
   struct IFFChunkSGHD ch_SGHD;
   struct IFFChunkMHD_ ch_MHD_;
};
/*
 * These numbers are taken from DumpFileWriter in the optima implementation
 */
#define DUMP_SIZE_POS 4
#define BLOCK_LIST_SIZE_POS 40
#define PMD_VERSION 0x100

#define TOP_FORM_SIZE_OFFSET 1 * sizeof(uint32_t)
#define FORM_SIZE_OFFSET 1 * sizeof(uint32_t)
#define CONTENT_LENGTH_OFFSET 2 * sizeof(uint32_t) + sizeof(struct IFFChunkHeader)
#define MEM_CONTENT_LENGTH_OFFSET 1 * sizeof(uint32_t) + sizeof(struct IFFChunkHeader)
#define CONTENT_STATUS_OFFSET 1 * sizeof(uint32_t) + sizeof(struct IFFChunkHeader)
#define ZDATA_SIZE_OFFSET 1 * sizeof(uint32_t)
#define ZDEF_IMAGE_LENGTH_OFFSET 5 * sizeof(uint32_t) + sizeof(struct IFFChunkHeader)
#define DATA_SIZE_OFFSET 1 * sizeof(uint32_t)

#define ERROR_BLOCK_SIZE 72
#define ERBL_SIZE 36
#define ERIN_SIZE (sizeof(struct IFFChunkERIN) - sizeof(struct IFFChunkHeader))
#define COMPRESSION_HEADER_SIZE 48
#define SGHD_SIZE 76
#define SGHD_TOP_FORM_SIZE 68
#define SGHD_FORM_SIZE 32
#define MHD_SIZE 80
#define MHD_TOP_FORM_SIZE 80
#define MHD_FORM_SIZE 44

/**
 * @macro COMPRESS_ZLIB_Z77
 *
 * @brief Flag that specifies that Z77 should be used for compression.
 */
#define COMPRESS_ZLIB_Z77   ((uint32_t) 0x1)
/**
 * @macro ALIGN_TO
 *
 * @brief Aligns value to the specified size.
 *
 * @long Aligns value to the specified size.
 *
 * @input value The value to align.
 * @input size The size to align value to.
 *
 * @return Returns the size aligned value.
 */
#define ALIGN_TO(value, size) (((value) + (size) - 1UL) & ~((size) - 1UL))

/* RMD file writer internal functions declared here */

/**
 * @function write_int
 *
 * @brief Writes a 32 bit word to the dump_file.
 *
 * @long Writes a 32 bit word to the dump_file. dump_file field in
 * the dump_file struct must be initialised using open() before calling
 * this function. This is called from the higher level
 * functions of the interface.
 *
 * @input value 32 bit word to write to file
 * @input dump_file Dump to write to.
 *
 * @return Returns RMD_FILE_SUCCESS if successful , otherwise it
 * returns RMD_FILE_WRITE_ERROR if fwrite fails, RMD_FILE_WRITE_SHORT_ERROR
 * if the data written is shorter than expected
 *
 */
static RmdFileStatus
write_int(uint32_t value, const RmdDumpFile *dump_file);

/**
 * @function write_memory
 *
 * @brief Writes raw memory.
 *
 * @input data Pointer to memory to write
 * @input data_len Length of memory segment to write (in bytes)
 * @input dump_file Dump to write to.
 *
 * @return Returns RMD_FILE_SUCCESS if successful , otherwise it
 * returns RMD_FILE_WRITE_ERROR if fwrite fails, RMD_FILE_WRITE_SHORT_ERROR
 * if the data written is shorter than expected
 *
 * @example Typical use is
 * write_memory((uint32_t *) whatever, sizeof(whatever));
 * where whatever is struct holding formatted data
 *
 */
static RmdFileStatus
write_memory(const void *data, const uint32_t data_len, const RmdDumpFile *dump_file);

/**
 * @function get_padding
 *
 * @brief Gets the number of bytes to pad in order to get 32 bit words
 * in the dump_file.
 *
 * @long Gets the number of bytes to pad in order to get 32 bit words
 * in the dump_file. This is called from the higher level functions of the
 * interface.
 *
 * @input size Dump size to align to 32 bit
 *
 * @return Returns number of bytes [0, 1, 2, 3] to align to 32 bit words
 *
 */
static uint32_t
get_padding(uint32_t size);

/**
 * @function write_padding
 *
 * @brief Writes padding bytes to the dump_file.
 *
 * @long Writes padding bytes to the dump_file. This is called from the higher level
 * functions of the interface.
 *
 * @input padding Number of padding bytes to write.
 * @input data Data to write as padding.
 * @input dump_file Dump to write to.
 *
 * @return Returns RMD_FILE_SUCCESS if successful or RMD_FILE_WRITE_ERROR
 * if a file write fails
 *
 */
static RmdFileStatus
write_padding(uint32_t padding, uint8_t data, const RmdDumpFile *dump_file);

/**
 * @function seek_and_write_int
 *
 * @brief Help function to simplify code in update_sizes and update_sizes_all.
 *
 * @input dump_file Dump to write to
 * @input offset Offset in file
 * @input data Data to writeto file
 *
 * @return Returns RMD_FILE_SUCCESS if successful , otherwise it
 * returns RMD_FILE_WRITE_ERROR if fwrite fails, RMD_FILE_WRITE_SHORT_ERROR
 * if the data written is shorter than expected or RMD_FILE_SEEK_ERROR if
 * the fseek fails
 *
 */
static RmdFileStatus
seek_and_write_int(const RmdDumpFile *dump_file, uint32_t offset, uint32_t data);

/**
 * @function update_sizes
 *
 * @brief Updates the size fields in the dump_file.
 *
 * @long Updates the size fields in the dump_file. Seeks back to stored positions
 * in the file to update size data. Updates dump_size and block_list_size in the
 * file. This is called from the higher level functions of the interface.
 *
 * @input size Size of the written block
 * @input dump_file Dump to write to.
 *
 * @return Returns RMD_FILE_SUCCESS if successful , otherwise it
 * returns RMD_FILE_WRITE_ERROR if fwrite fails, RMD_FILE_WRITE_SHORT_ERROR
 * if the data written is shorter than expected or RMD_FILE_SEEK_ERROR if
 * the fseek fails
 *
 */
static RmdFileStatus
update_sizes(uint32_t size, RmdDumpFile *dump_file);

/**
 * @function update_sizes_all
 *
 * @brief Updates the size fields in the dump_file.
 *
 * @long Updates the size fields in the dump_file. Seeks back to stored positions
 * in the file to update size data. Updates all sizes in the dump_file. This is
 * called from the higher level functions of the interface.
 *
 * @input total_header_size Size of block header, including description if it exists
 * @input padding Padding of the written block
 * @input dump_file Dump to write to. Also used as input for the sizes stored in
 * dump_file->sizes.compressedDAtaSize and dump_file->sizes.uncompressed_data_size
 *
 * @output dump_file->sizes.dump_size
 * @output dump_file->sizes.block_list_size
 *
 * @return Returns RMD_FILE_SUCCESS if successful , otherwise it
 * returns RMD_FILE_WRITE_ERROR if fwrite fails, RMD_FILE_WRITE_SHORT_ERROR
 * if the data written is shorter than expected or RMD_FILE_SEEK_ERROR if
 * the fseek fails
 *
 */
static RmdFileStatus
update_sizes_all(uint32_t total_header_size, uint32_t padding, RmdDumpFile *dump_file);

/*static RmdFileStatus
 write_string(const char *string, uint32_t len, const RmdDumpFile *dump_file);*/

/**
 * @function write_desc
 *
 * @brief Writes a description to the dump_file, including tags and length.
 *
 * @long Writes a description to the dump_file, including tags, length and padding. This
 * is called from the higher level functions of the interface.
 *
 * @input description The description string to write to file. Must be '\0'
 * terminated.
 * @input padding Number of characters to pad the data
 * @input dump_file Dump to write to
 *
 * @return Returns RMD_FILE_SUCCESS if successful , otherwise it
 * returns RMD_FILE_WRITE_ERROR if fwrite fails, RMD_FILE_WRITE_SHORT_ERROR
 * if the data written is shorter than expected
 *
 */
static RmdFileStatus
write_desc(const char *description, int desc_size, int padding,
         const RmdDumpFile *dump_file);

/**
 * @function write_chunk_header
 *
 * @brief Writes a chunk header to the dump_file.
 *
 * @long Writes a chunk header to the dump_file.
 * @long Converts both the tag and the size to network endian before writing to disk.
 *
 * @input id Chunk id tag.
 * @input data_size Data size of the chunk.
 * @input dump_file Dump where the chunk will be written
 *
 * @return Returns RMD_FILE_SUCCESS if successful , otherwise it
 * returns RMD_FILE_WRITE_ERROR if fwrite fails, RMD_FILE_WRITE_SHORT_ERROR
 * if the data written is shorter than expected
 */
static RmdFileStatus
write_chunk_header(IffId id, uint32_t data_size, const RmdDumpFile *dump_file);

/**
 * @function write_group_header
 *
 * @brief Writes a group header (FORM, LIST, CAT_ or PROP).
 *
 * @input id Group chunk id.
 * @input data_size Data size of the group chunk.
 * @input form_type Group subchunk id.
 * @input dump_file Dump where the group header will be written.
 *
 * @return Returns position to size field in the file if successful , otherwise it
 * returns 0
 *
 */
static size_t 
write_group_header(IffId id, uint32_t data_size, IffId form_type, RmdDumpFile *dump_file);

/**
 * @function write_new_block_header
 *
 * @brief Writes the block header to the dump_file.
 *
 * @long Writes the block header to the dump_file. This is called from the higher level
 * functions of the interface.
 *
 * @input size Size of the block
 * @input dump_file Dump to write to.
 *
 * @return Returns file position to size field if successful,
 * otherwise it returns 0
 *
 */
static size_t 
write_new_block_header(uint32_t size, RmdDumpFile *dump_file);

/**
 * @function compression_init
 *
 * @brief Initialises the compression algorithm.
 *
 * @long Initialises the compression algorithm. This is called
 * internally from higher level functions in the implementation
 *
 * @input dump_file Dump to work on
 * @output dump_file->compression->zstrm Compression stream object for consecutive calls to
 * the compression algorithm.
 * @output dump_file->compression->zlib_buffers Working buffers for the compression algorithm
 * @output dump_file->compressed_data_size Data counter for the compression algorithm
 *
 * @return Returns RMD_FILE_SUCCESS if successful , otherwise it returns
 * RMD_MEM_ERROR if there were a memory allocation error, RMD_Z_STREAM_ERROR if
 * there were an error initialising the z_stream object or RMD_FILE_ERROR if there
 * were an unforeseen error.
 *
 */
static RmdFileStatus
compression_init(uint32_t window_bits, uint32_t mem_level, RmdDumpFile *dump_file);

/**
 * @function write_data_compressed_reset
 *
 * @brief Resets the compression algorithm for new data block.
 *
 * @long Resets the compression algorithm for new data block. This is called
 * internally from higher level functions in the implementation
 *
 * @input dump_file Dump to work on
 *
 * @return Returns RMD_FILE_SUCCESS if successful , otherwise it returns
 * RMD_Z_STREAM_ERROR if there were an error with the z_stream object
 *
 */
static RmdFileStatus
write_data_compressed_reset(RmdDumpFile *dump_file);

static uint32_t
refill_input_buffer(z_stream *zstrm, OSADDRESS start_address, uint32_t length,
         uint32_t left_to_proc, RmdDumpFile *dump_file);

static RmdFileStatus
empty_output_buffer(z_stream *zstrm, RmdDumpFile *dump_file);

/**
 * @function write_data_compressed
 *
 * @brief Compresses and writes data from memory to dump_file.
 *
 * @long Compresses and writes data from memory to dump_file. This is called
 * internally from higher level functions in the implementation
 *
 * @input start_address Address to get data to compress and write to dump_file
 * @ input data_len Length of data
 * @input dump_file Dump to work on
 *
 * @return Returns RMD_FILE_SUCCESS if successful , otherwise it returns
 * RMD_MEM_ERROR if there were a memory allocation error,
 * RMD_Z_STREAM_ERROR if there were an error in the deflate (),
 * RMD_FILE_Z_STREAM_END if there were an unexpected end of data to the compressor,
 * RMD_FILE_WRITE_ERROR if an error occured with file writing,
 * RMD_FILE_WRITE_ERROR_SHORT if the data written is too short or
 * RMD_FILE_ERROR if there were an unforeseen error.
 *
 */
static RmdFileStatus
write_data_compressed(OSADDRESS start_address, uint32_t data_len, RmdDumpFile *dump_file);

static RmdFileStatus
write_data_compressed_end(RmdDumpFile *dump_file);

/**
 * @function write_data
 *
 * @brief Interface function for write_memory.
 *
 * @input address Address to the memory that is beeing dumped
 * @input data_len Length of memory data to be dumped in the following call
 * @input dump_file Dump to write to
 *
 * @return Returns RMD_FILE_SUCCESS if successful , otherwise it returns
 * RMD_FILE_WRITE_ERROR if fwrite fails,
 * RMD_FILE_WRITE_SHORT_ERROR if the data written is shorter than expected,
 *
 * @seealso write_memory
 * @seealso write_data_compressed
 */
static RmdFileStatus
write_data(OSADDRESS data, uint32_t dataLen, RmdDumpFile *dump_file);

/**
 * @function dump_memory_block_header
 *
 * @brief Writes header for memory block and initialises dump_file for memory data.
 *
 * @long Writes header for memory block and initialises dump_file for memory data.
 * This is called externally from the code that is implementing the actual
 * dumping.
 *
 * @input address Address to the memory that is beeing dumped
 * @input data_len Length of memory data to be dumped in the following call
 * @input description String with description of error
 * @input dump_file Dump to write to
 *
 * @output dump_file->file_pos.* All file positions are updated if compression is used
 *
 * @return Returns number of bytes in header if successful,
 * otherwise it returns 0
 */
static uint32_t
dump_memory_block_header(OSADDRESS address, uint32_t data_len,
         const char *description, RmdDumpFile *dump_file);

/**
 * @function dump_request_header
 *
 * @brief Writes header for signal block and initialises dump_file for memory data.
 *
 * @long Writes header for memory block and initialises dump_file for memory data.
 * This is called externally from the code that is implementing the actual
 * dumping.
 *
 * @input address Address to the memory that is beeing dumped
 * @input data_len Length of memory data to be dumped in the following call
 * @input description String with description of error
 * @input dump_file Dump to write to
 *
 * @output dump_file->file_pos.* All file positions are updated if compression is used
 *
 * @return Returns number of bytes in header if successful,
 * otherwise it returns 0
 */
static uint32_t
dump_request_header(OSADDRESS request, RmdDumpFile *dump_file);

/* ZLIB memory functions */
static voidpf
zlib_alloc(voidpf opaque, uInt items, uInt size);
static void
zlib_free(voidpf opaque, voidpf address);

#ifdef __cplusplus
}
#endif

#endif /* RMD_FILEFORMAT_PRIVATE_H_ */
