/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2010 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

/**
 *
 * @toc OSE:
 *
 * @file rmd_private.h
 *
 * @brief Run Mode Dumper private internal interface
 *
 * @long This file defines the internal interface used in the Run Mode Dumper
 * process. The purpose of this interface is to allow unit testing of the
 * internal RMD process functionality.
 */

#ifndef _RMD_PRIVATE_H
#define _RMD_PRIVATE_H

#include "stdint.h"
#include "list.h"
#include "ose.h"

#ifdef __cplusplus
extern "C"
{
#endif

/* Forward declaration */
union SIGNAL;

struct RmdPrivateSetProperties
{
   uint32_t num_long_queue_procs;
   uint32_t max_signals_in_queue;
   uint32_t max_stack_top_size;
};

struct Rmd
{
   uint32_t magic;
   List dumps;
   PROCESS rmm;
   PROCESS rmm_main;
   OSATTREF attach_ref;
   OSBOOLEAN rmm_connected;
   union SIGNAL *stored_connect_reply;
   uint32_t *supported_monitor_signals;
   uint32_t supported_monitor_signals_count;

   /* Signal event */
   OSADDRESS start_dump_event_trigger;
   OSADDRESS stop_dump_event_trigger;

   /* system parameters */
   char path[64];
   uint32_t max_ongoing_dumps;
   OSBOOLEAN use_compression;
   uint32_t window_bits;
   uint32_t mem_level;

   /* Rmd properties */
   struct RmdPrivateSetProperties rmd_properties;
};

/**
 * @function rmd_private_start_dump
 *
 * @brief Initialises a dump and sets up the file for subsequent calls to other
 * dump functions.
 *
 * @long
 * All dumps must be initialised by calling rmd_private_start_dump before any
 * other functions in the rmd_private interface can be used. The dump must be closed
 * by calling rmd_private_stop_dump.
 * The initialisation process is the following:
 *  * Generate a file name from the input in req->path.
 *  * Open the dump file and write the file header.
 *  * Create a new dump node and add to the dump list.
 *  * Attach to the process that requested the dump.
 *  * Get supported interfaces from RMM.
 *  * Connect to RMM.
 *  * Signal start dump event.
 *
 * @input req The request signal of type RmdStartDumpRequest.
 *
 * @input rmd The Rmd struct holding information about the dumps in this process.
 *
 * @output Replies with the signal RMD_START_DUMP_REPLY to the sender of the request
 * signal when finished setting up the dump.
 *
 *
 * @seealso generate_dump_file_name
 * @seealso get_supported_interfaces_from_rmm
 * @seealso connect_to_rmm
 * @seealso stored_connect_to_rmm
 * @seealso rmd_private_stop_dump
 * @seealso RmdStartDumpRequest
 */
void
rmd_private_start_dump(union SIGNAL *req, struct Rmd *rmd);

/**
 * @function rmd_private_stop_dump
 *
 * @brief Closes the dump associated with the PID of the requester.
 *
 * @long The function is called in two different cases
 * 1. The dumping process sends the signal RmdStopDumpRequest
 * 2. The client dies and the kernel sends a signal.
 * These two cases must be handled differently when getting the PID and sending a
 * reply signal.
 *
 * @input rmd_req The request signal of type RmdStopDumpRequest or a signal that the
 *        client is detached.
 *
 * @input rmd The Rmd struct holding information about the dumps in this process.
 *
 *
 * @seealso rmd_private_start_dump
 * @seealso RmdStopDumpRequest
 *
 */
void
rmd_private_stop_dump(union SIGNAL *req, struct Rmd *rmd);

/**
 * @function rmd_private_dump_error
 *
 * @brief Dumps error information to the dump.
 *
 * @input req The request signal of type RmdDumpErrorRequest
 *
 * @input rmd The Rmd struct holding information about the dumps in this process.
 *
 * @seealso RmdDumpErrorRequest
 *
 */
void
rmd_private_dump_error(union SIGNAL *req, struct Rmd *rmd);

/**
 * @function rmd_private_dump_memory
 *
 * @brief Dumps general memory.
 *
 * @long Implements two sanity checks on the parameters, if the size is 0 or the address
 * is NULL there will be no output in the dump file.
 *
 * @input req The request signal of type RmdDumpMemoryRequest
 *
 * @input rmd The Rmd struct holding information about the dumps in this process.
 *
 * @seealso RmdDumpMemoryRequest
 */
void
rmd_private_dump_memory(union SIGNAL *req, struct Rmd *rmd);

/**
 * @function rmd_private_dump_string
 *
 * @brief Dumps a string.
 *
 * @input req The request signal of type RmdDumpStringRequest
 *
 * @input rmd The Rmd struct holding information about the dumps in this process.
 *
 * @seealso RmdDumpStringRequest
 */
void
rmd_private_dump_string(union SIGNAL *req, struct Rmd *rmd);

/**
 * @function rmd_private_dump_process_info
 *
 * @brief Dumps the informations about the running processes
 *
 * @input rmd_req The request signal of type RmdDumpProcessInfoRequest
 * @input rmd The Rmd struct holding information about the dumps in this process.
 *
 * @seealso RmdDumpProcessInfoRequest
 *
 */
void
rmd_private_dump_process_info(union SIGNAL *req, struct Rmd *rmd);

/**
 * @function rmd_private_dump_stack_usage_info
 *
 * @brief Request information about stack usage from the RMM and writes the
 * response to the dump file.
 *
 * @input rmd_req The request signal of type RmdDumpStackUsageInfoRequest
 * @input rmd The Rmd struct holding information about the dumps in this process.
 *
 * @seealso RmdDumpStackUsageInfoRequest
 *
 */
void
rmd_private_dump_stack_usage_info(union SIGNAL *req, struct Rmd *rmd);

/**
 * @function rmd_private_dump_env_vars_info
 *
 * @brief Requests information about the environment variables for the scope
 * of interest from the RMM and writes the response to the dump file.
 *
 * @input rmd_req The request signal of type RmdDumpEnvVarsInfoRequest
 * @input rmd The Rmd struct holding information about the dumps in this process.
 *
 * @seealso RmdDumpEnvVarsInfoRequest
 *
 */
void
rmd_private_dump_env_vars_info(union SIGNAL *req, struct Rmd *rmd);

/**
 * @function rmd_private_dump_signal_queue_info
 *
 * @brief Requests information about the signal queues from the RMM and writes
 * the results to the dump file.
 *
 * @input rmd_req The request signal of type RmdDumpSignalQueueInfoRequest
 * @input rmd The Rmd struct holding information about the dumps in this process.
 *
 * @seealso RmdDumpSignalQueueInfoRequest
 *
 */
void
rmd_private_dump_signal_queue_info(union SIGNAL *req, struct Rmd *rmd);

/**
 * @function rmd_private_dump_pool_info
 *
 * @brief Requests information about the pools from the RMM and writes the
 * results to the dump file.
 *
 * @input rmd_req The request signal of type RmdDumpPoolInfoRequest
 * @input rmd The Rmd struct holding information about the dumps in this process.
 *
 * @seealso RmdDumpPoolInfoRequest
 *
 */
void
rmd_private_dump_pool_info(union SIGNAL *req, struct Rmd *rmd);

/**
 * @function rmd_private_dump_pool_signals_info
 *
 * @brief Finds all the signals in all pools in the current segment and writes
 * them to the dump file.
 *
 * @long A functionality to iterate over all signals in a segment is
 * implemented because the RMM does not accept segment as scope input
 * for this signal. The request signal with the wrong scope is written
 * as the request in the dump file.
 *
 * @input rmd_req Request signal of type RmdDumpPoolSignalsInfoRequest.
 * @input rmd The Rmd struct containing information about ongoing dumps.
 *
 * @seealso RmdDumpPoolSignalsInfoRequest
 * @seealso for_each_pool_in_segment
 * @seealso dump_pool_signals_iterator
 *
 */
void
rmd_private_dump_pool_signals_info(union SIGNAL *req, struct Rmd *rmd);

/**
 * @function rmd_private_dump_sys_param_info
 *
 * @brief Requests the system parameters from the RMM and writes the replies
 * to the dump file.
 *
 * @input rmd_req Request signal of type RmdDumpSysParamInfoRequest.
 * @input rmd The Rmd struct containing information about ongoing dumps.
 *
 * @seealso RmdDumpSysParamInfoRequest
 *
 */
void
rmd_private_dump_sys_param_info(union SIGNAL *req, struct Rmd *rmd);


void
rmd_private_dump_registers_info(union SIGNAL *req, struct Rmd *rmd);

/**
 * @function rmd_private_dump_program_info
 *
 * @brief Sends the two signals PmLoadModuleInfoRequest and PmProgramInfo to
 * the program manager and writes the results to the dump file.
 *
 * @input rmd_req Request signal of type RmdDumpProgramInfoRequest
 *
 * @seealso RmdDumpProgramInfoRequest
 */
void
rmd_private_dump_program_info(union SIGNAL *req, struct Rmd *rmd);

/**
 * @function rmd_private_dump_trace_info
 *
 * @brief Requests information from the RMM regarding trace information and
 * writes the results to the dump file.
 *
 * @input rmd_req Request signal of type RmdDumpTraceInfoRequest
 *
 * @seealso RmdDumpTraceInfoRequest
 */
void
rmd_private_dump_trace_info(union SIGNAL *req, struct Rmd *rmd);

/**
 * @function rmd_private_dump_load_module_info
 *
 * @brief Requests information about the load module from the program manager
 * and writes the results to file.
 *
 * @input rmd_req Request signal of type RmdDumpTraceInfoRequest
 *
 * @seealso RmdDumpTraceInfoRequest
 */
void
rmd_private_dump_load_module_info(union SIGNAL *req, struct Rmd *rmd);

/**
 * @function rmd_private_dump_pool_memory
 *
 * @brief Calls either dump_pool_memory or dump_stack_usage_memory to write
 * the pool memory of interest to the dump file.
 *
 * @input rmd_req Request signal of type RmdDumpPoolMemoryRequest
 *
 * @seealso RmdDumpPoolMemoryRequest
 * @seealso dump_pool_memory
 * @seealso dump_stack_usage_memory
 */
void
rmd_private_dump_pool_memory(union SIGNAL *req, struct Rmd *rmd);

/**
 * @function rmd_private_dump_heap_memory
 *
 * @brief Requests information about the heap from the RMM and writes the
 * results to the dump file by calling dump_memory_heap_buffers or
 * dump_memory_heap.
 *
 * @input rmd_req Request signal of type RmdDumpHeapMemoryRequest.
 *
 * @seealso RmdDumpHeapMemoryRequest
 * @seealso dump_memory_heap
 * @seealso dump_memory_heap_buffers
 *
 */
void
rmd_private_dump_heap_memory(union SIGNAL *req, struct Rmd *rmd);

/**
 * @function rmd_private_dump_program_memory
 *
 * @brief Request information about the program memory from the PM and writes
 * the memory to the dump file.
 *
 * @input rmd_req Request signal of type RmdDumpProgramMemoryRequest.
 *
 * @seealso RmdDumpProgramMemoryRequest
 */
void
rmd_private_dump_program_memory(union SIGNAL *req, struct Rmd *rmd);

#ifdef __cplusplus
} /* extern "C" */
#endif

#endif /* _RMD_PRIVATE_H */
