/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2010 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

/*
 * This is the implementation for Read-Only File System Volume Manager.
 * It is based on VFS framework residing in the CRT.
 */

#include "rofs_media.h"
#include "rofs_init.h"
#include "compiler_macros.h"
#include "fm.sig"
#include "vfs.h"
#include "errno.h"
#include "ose.h"
#include "string.h"
#include "sys/stat.h"
#include "fss.h"
#include "sysparam.h"
#include "ramlog.h"
#include "heapapi.h"
#include "stdlib.h"

/* The process name for ROFS volume manager is <prefix><mount point>.
 * This is the <prefix> part. */
#define ROFS_PROC_NAME_PREFIX "ose_rofs_"

/* Number of bytes of stack to allocate for ROFS volume manager process */
#define ROFS_VM_STACK_SIZE_BYTES 1000

/* Magic value used as ROFS volume image magic when no ROFS volume has been
 * inserted. */
#define ROFS_MAGIC_NO_VOLUME 1

/* Returns minimum of a and b. NOTE: Parameters evaluated twice! */
#define MIN(a,b) ((a)<(b)?(a):(b))

/* ROFS file description. */
struct RofsFd
{
      const struct RofsMediaFile *file;
      off64_t file_pos;
      uint32_t readdir_name_offs;
};

/* Used to cast a VfsHandle to a file description without breaking compiler
 * strict aliasing rules. */
union HandleCast
{
      VfsHandle handle;
      struct RofsFd *fd;
};

union VolumeVolatileCast
{
      volatile struct RofsMediaVolume volatile_vol;
      struct RofsMediaVolume vol;
};

union SIGNAL
{
      SIGSELECT         sig_no;
      struct FssResolve sig_resolve;
};

/* Declaration below requires __attribute__ keyword, which only ARM RVCT and
 * GCC are known for supporting. Support for other compilers could be added
 * when needed, the required functionality is to be able to select in which
 * ELF section the variables shall be put. */
#if (!defined(__ARMCC_VERSION) && !defined(__GNUC__))
#   error Unsupported compiler.
#endif

/* This is the default content of the ROFS volume. If no volume is inserted
 * into the ELF file, the definition below is what ose_rofs_start_handler0 will
 * find. */
static const uint32_t __attribute__ ((section (".rofs")))
   rofs_section = {ROFS_MAGIC_NO_VOLUME};

/*
 * Using an external variable to point to the ROFS volume is a way to assure
 * that the compiler does not optimize for the struct content given above.
 */
const struct RofsMediaVolume *ose_rofs_vol;

/* Set to 1 by ose_rofs_start_handler0() if a valid volume was found.
 * Used by ose_rofs_mount_embedded() to determine if there is anything to
 * mount. */
static int rofs_found = 0;

/* Path to mount point. This is a static variable since current implementation
 * of ROFS only supports one mounted volume. This is a VFS limitation due to
 * the fact that no instance information is provided in the calls to the
 * VFS interface. */
static char *rofs_mount_point = NULL;

/*
 * Helper function to perform a typecast without breaking strict aliasing rules
 */
static struct RofsFd *get_fd(VfsHandle handle)
{
   union HandleCast h;

   h.handle = handle;
   return h.fd;
}


/*
 * Helper function to translate a volume offset to a volume pointer
 */
static const void *
l_offs2ptr(const struct RofsMediaVolume *vol, voloffs_t offs)
{
   return (const uint8_t *)vol + offs;
}


/*
 * Lookup a file in the root directory. Returns a pointer to the file on
 * the media, or NULL if file not found.
 */
static const struct RofsMediaFile *
l_lookup(const struct RofsMediaVolume *vol,
         const char *fname)
{
   /* Start with first file in the directory */
   const struct RofsMediaFile *file =
      l_offs2ptr(vol, vol->rootDir.dataOffset);

   /* One past the last file, used to know when to stop reading directory. */
   const struct RofsMediaFile * const end_file =
      l_offs2ptr(vol, vol->rootDir.dataOffset + vol->rootDir.fileSizeBytes);

   /* Check that first part of fname is the mount path */
   if (strncmp(fname, rofs_mount_point, strlen(rofs_mount_point)) != 0)
   {
      /* Nope, fname not referring to this volume */
      return NULL;
   }

   /* Skip mount path. */
   fname += strlen(rofs_mount_point);

   /* Must either be referring to volume root, or find a delimiter */
   if ((fname[0] != '\0') && (fname[0] != '/'))
   {
      return NULL;
   }

 new_dir_level:
   /* Skip delimiter(s), if any */
   while (fname[0] == '/')
   {
      fname++;
   }

   if ((fname[0] == '.') && ((fname[1] == '/') || (fname[1] == '\0')))
   {
      /* This is a "." directory entry. Skipp to next component. */
      fname++;
      goto new_dir_level;
   }

   if (fname[0] == '\0')
   {
      /* Volume root requested */
      return &ose_rofs_vol->rootDir;
   }

   /*
    * The directory file (inDir) contains a list of struct RofsMediaFile.
    * Traverse those media files until a file with correct name is found.
    */
   while (file < end_file)
   {
      if (strcmp(file->name, fname) == 0)
      {
         /* File found */
         return file;
      }

      /* Go to next file */
      file = (const struct RofsMediaFile *)
         ((const uint8_t *)file + file->nextOffset);
   }

   return NULL;
}


/*
 * VFS open
 */
static int
rofs_open(VfsHandle handle, const char *path, EfsBitmask mode)
{
   struct RofsFd * const fd = get_fd(handle);

   (void) mode;

   /* Check for illegal modes */
   if ((mode & ~(FM_MODE_READ | FM_MODE_TEXT | FM_MODE_TRANS)) != 0)
   {
      return EROFS;
   }

   fd->file = l_lookup(ose_rofs_vol, path);
   if (fd->file == NULL)
   {
      return ENOENT;
   }

   fd->file_pos = 0;
   fd->readdir_name_offs = 0;

   return 0;
}


/*
 * VFS close
 */
static int
rofs_close(VfsHandle handle)
{
   (void)handle;
   
   return 0;
}


/*
 * VFS lseek
 */
static long
rofs_lseek(VfsHandle handle, int whence, int offset)
{
   struct RofsFd * const fd = get_fd(handle);
   off64_t newpos;

   if (whence == SEEK_SET)
   {
      newpos = offset;
   }
   else if (whence == SEEK_CUR)
   {
      newpos = fd->file_pos + offset;
   }
   else if (whence == SEEK_END)
   {
      newpos = fd->file->fileSizeBytes + offset;
   }
   else
   {
      return (int) fd->file_pos;
   }

   if (newpos >= 0)
   {
      fd->file_pos = newpos;
   }

   return (int) fd->file_pos;
}


/*
 * VFS read
 */
static long
rofs_read(VfsHandle handle, size_t offset, void *ptr, size_t length)
{
   struct RofsFd * const fd = get_fd(handle);

   if (offset != (size_t)FM_POSITION_CURRENT)
   {
      fd->file_pos = offset;
   }

   if ((unsigned) fd->file_pos >= fd->file->fileSizeBytes)
   {
      /* No more data in file */
      return 0;
   }

   /* Read requested number of bytes if file is big enough, or the remaining
    * bytes in file otherwise. */
   length = MIN(length, fd->file->fileSizeBytes - fd->file_pos);

   memcpy(ptr,
          l_offs2ptr(ose_rofs_vol, fd->file->dataOffset + fd->file_pos),
          length);

   fd->file_pos += length;

   return length;
}


/*
 * The macro BITMOVE and BITPOS and function l_amode_to_attrs() are shamelessly
 * copied from flib.c in CRT.
 */

/*
 *  Macro BITPOS
 *
 *  Returns the bit position of a bitmask with only one bit.
 *
 *  mask = (1 << pos)   =>   pos == BITPOS(mask)
 *
 *  Note: this algorithm requires that there should only be one bit set in
 *        mask.
 */
#define BITPOS(mask) (((((U32)mask) & 0xAAAAAAAA) ? 1 : 0) + \
      ((((U32)mask) & 0xCCCCCCCC) ? 2 : 0) + \
      ((((U32)mask) & 0xF0F0F0F0) ? 4 : 0) + \
      ((((U32)mask) & 0xFF00FF00) ? 8 : 0) + \
      ((((U32)mask) & 0xFFFF0000) ? 16 : 0))


/*
 *  Macro BITMOVE
 *
 *  Extracts the "from" bit from the bitmask "data", and shifts it to
 *  the position of the "to" bit.
 *
 *  Both "from" and "to" must only contain one set bit each.
 */
#define BITMOVE(data, from, to) \
   ((BITPOS(from) > BITPOS(to)) ? \
    ((((U32)data) & ((U32)from)) >> (BITPOS(from) - BITPOS(to))) :\
    ((((U32)data) & ((U32)from)) << (BITPOS(to) - BITPOS(from))))


static EfsBitmask
l_amode_to_attrs(mode_t amode)
{
   EfsBitmask attrs = 0;

   /* Standard Posix attributes. */
   /* Since ROFS is a read-only file-system, there is no need for
    * the write flag being kept. */
   attrs |= BITMOVE(amode, ROFS_S_IXOTH, FM_ATTR_OTH_EXECUTE);
   attrs |= BITMOVE(amode, ROFS_S_IROTH, FM_ATTR_OTH_READ);
   attrs |= BITMOVE(amode, ROFS_S_IXGRP, FM_ATTR_GRP_EXECUTE);
   attrs |= BITMOVE(amode, ROFS_S_IRGRP, FM_ATTR_GRP_READ);
   attrs |= BITMOVE(amode, ROFS_S_IXUSR, FM_ATTR_OWN_EXECUTE);
   attrs |= BITMOVE(amode, ROFS_S_IRUSR, FM_ATTR_OWN_READ);
   attrs |= BITMOVE(amode, ROFS_S_IARCH, FM_ATTR_ARCHIVED);
   attrs |= BITMOVE(amode, ROFS_S_ISUID, FM_ATTR_SUID);
   attrs |= BITMOVE(amode, ROFS_S_ISGID, FM_ATTR_SGID);
   attrs |= BITMOVE(amode, ROFS_S_ISVTX, FM_ATTR_ISVTX);

   /* OSE Extension attibutes */
   attrs |= BITMOVE(amode, ROFS_S_IHIDDEN, FM_ATTR_HIDDEN);
   attrs |= BITMOVE(amode, ROFS_S_ISTARTUP, FM_ATTR_STARTUP);
   attrs |= BITMOVE(amode, ROFS_S_IUSER0, FM_ATTR_USER0);
   attrs |= BITMOVE(amode, ROFS_S_IUSER1, FM_ATTR_USER1);
   attrs |= BITMOVE(amode, ROFS_S_IUSER2, FM_ATTR_USER2);
   attrs |= BITMOVE(amode, ROFS_S_IUSER3, FM_ATTR_USER3);

   return attrs;
}


/*
 * stat and fstat helper function
 */
static int
l_stat(const struct RofsMediaFile *file, struct FmObjectInfo *info)
{
   if (file == NULL)
   {
      return ENOENT;
   }

   if (ROFS_S_ISDIR(file->mode))
   {
      info->type = FM_TYPE_DIR;
   }
   else
   {
      info->type = FM_TYPE_FILE;
   }

   info->attrs = l_amode_to_attrs(file->mode);
   
   info->sizeBytes = file->fileSizeBytes;
   info->diskId = file->inode;
   info->links = file->linkCount;
   info->creationTime = (U32) file->ctime;
   info->dataModTime = (U32) file->ctime;
   info->headModTime = (U32) file->ctime;
   info->accessTime = (U32) file->ctime;

   return 0;
}


/*
 * VFS stat
 */
static int
rofs_stat(const char *path, struct FmObjectInfo *info)
{
   return l_stat(l_lookup(ose_rofs_vol, path), info);
}


/*
 * VFS fstat
 */
static int
rofs_fstat(VfsHandle handle, struct FmObjectInfo *info)
{
   return l_stat(get_fd(handle)->file, info);
}


/*
 * VFS opendir
 */
static int
rofs_opendir(VfsHandle handle, const char *path)
{
   struct RofsFd * const fd = get_fd(handle);

   fd->file = l_lookup(ose_rofs_vol, path);

   if (fd->file == NULL)
   {
      return ENOENT;
   }

   fd->file_pos = 0;
   fd->readdir_name_offs = 0;

   return 0;
}


/*
 * VFS readdir
 */
static long
rofs_readdir(VfsHandle handle, char *buf, size_t length)
{
   struct RofsFd * const fd = get_fd(handle);
   const struct RofsMediaFile * dirent = (const struct RofsMediaFile *)
      ((const char*) ose_rofs_vol + fd->file->dataOffset + fd->file_pos);
   size_t bytes_read = 0;

   if ((fd->file_pos < 0)
       || ((uint64_t) fd->file_pos >= fd->file->fileSizeBytes))
   {
      /* End of file reached, or file position error. */
      return 0;
   }

   while (((uint64_t) fd->file_pos < fd->file->fileSizeBytes) &&
          (length > 1))
   {
      const size_t name_len =
         strlen(dirent->name) - fd->readdir_name_offs;
      size_t name_frag_len = MIN(name_len, length - 1);

      /* Copy the file name, excluding null terminator */
      memcpy(buf, dirent->name + fd->readdir_name_offs, name_frag_len);

      if (name_len == name_frag_len)
      {
         /* Since the file name is complete, add null terminator */
         buf[name_frag_len] = '\0';

         /* Advance to next file */
         buf += (name_frag_len + 1);
         bytes_read += (name_frag_len + 1);
         length -= (name_frag_len + 1);
         fd->file_pos += dirent->nextOffset;
         dirent = (const struct RofsMediaFile *) (
            (const uint8_t *) dirent + dirent->nextOffset);
         fd->readdir_name_offs = 0;
      }
      else
      {
         /* File name did not fit, remember how much of it that was stored */
         buf += name_frag_len;
         bytes_read += name_frag_len;
         length -= name_frag_len;
         fd->readdir_name_offs += name_frag_len;
      }
   }

   return bytes_read;
}

/*
 * FIXME: move to include header
 */
#if defined(PAGE_SIZE_4K)
    #define PAGE_SHIFT  12
#elif defined(PAGE_SIZE_8K)
    #define PAGE_SHIFT  13
#else
  #error "No page size set."
#endif

#define PAGE_SIZE       (1UL << PAGE_SHIFT)
#define PAGE_MASK       (~(PAGE_SIZE-1))

#define _ALIGN_UP(addr,size)	(((addr)+((size)-1))&(~((size)-1)))
#define _ALIGN_DOWN(addr,size)	((addr)&(~((size)-1)))
#define _ALIGN(addr,size)     _ALIGN_UP(addr,size)

#define PAGE_ALIGN(addr)        _ALIGN(addr, PAGE_SIZE)


/*
 * Public function intended to be called as start handler 0 in krn.con
 * NOTE: OSE is not running when this function is called, so any system call
 *       depending on OSE state is illegal to call.
 */
void
ose_rofs_start_handler0(void)
{
   static char value[256];

   /* Since rofs_section defines an empty volume, the address range occupied by
    * an empty volume will be included at link time. This means that MM will
    * reserve this address range. The range we need to reserve is in addition
    * to this, since it is inserted after the file has been linked.
    * First memory page does not need to be reserved because of this. */
   long base;
   size_t rofs_size;
   int ext_rofs = 0;

   if(!sysparam_readattr_long("krn/region/rofs", "base", &base))
   {
      ext_rofs = 1;
      ose_rofs_vol = (const struct RofsMediaVolume *) base;
      base = (uintptr_t)base + sizeof (rofs_section);
   }
   else
   {
      base = (uintptr_t)&rofs_section + sizeof (rofs_section);
      ose_rofs_vol = (const struct RofsMediaVolume *) &rofs_section;
   }
   if (ose_rofs_vol->magic == ROFS_MAGIC_NO_VOLUME)
   {
      ramlog_printf(
         "ROFS: No embedded volume found. Use rofs_insert host tool to insert "
         "one.\n");
      return;
   }

   if (ose_rofs_vol->magic == ROFS_MAGIC_WRONG_ENDIAN)
   {
      ramlog_printf(
         "ROFS: Magic has the wrong endian. Volume will not be mounted.\n");
      return;
   }

   if (ose_rofs_vol->magic != ROFS_MAGIC)
   {
      ramlog_printf(
         "ROFS: Bad magic in ROFS volume at 0x%x. Volume will not be "
         "mounted.\n",
         ose_rofs_vol);
      return;
   }

   if (ose_rofs_vol->volumeSizeBytes == 0)
   {
      ramlog_printf(
         "ROFS: Empty volume found at 0x%x. Volume will not be mounted.\n",
         ose_rofs_vol);
      return;
   }

   ramlog_printf(
      "ROFS: Embedded FS image found with size %llu bytes.\n",
      ose_rofs_vol->volumeSizeBytes);

   rofs_size = ose_rofs_vol->volumeSizeBytes;

#ifdef OSMIPS
   /*
    * Volume image needs to be reserved in KUSEG or else MM will assume that
    * this memory is free to be used. KSEG0 refers to the same physical memory,
    * it is therefore not necessary to reserve there too.
    */
   base -= 0x80000000; /* Go from KSEG0 to KUSEG */
#endif

   /* Adjust base and size to be aligned to page size. */
   base = PAGE_ALIGN(base);

   /* Skip first page, but round upwards. */
   rofs_size = (rofs_size - 1) & PAGE_MASK;

   if(!ext_rofs)
   {
      sprintf(value,
              "base:0x%x size:0x%x perm:su_ro_usr_ro cache:copy_back "
              "mem:speculative_access",
              base, rofs_size);
      sysparam_write("krn/region/rofs", value);
      
      ramlog_printf(
         "ROFS: Added following system parameter: 'krn/region/rofs = %s'\n",
         value);
   }
   else
   {
      ramlog_printf(
         "ROFS: External ROFS file system at 0x%x size 0x%x found.'\n",
         base, rofs_size);
   }
   rofs_found = 1;
}


/*
 * ROFS interface to be registered to VFS.
 */
static struct VfsOps rofs_if =
{
   NULL,
   sizeof(struct RofsFd),
   rofs_open,
   rofs_close,
   rofs_read,
   NULL,
   rofs_stat,
   rofs_opendir,
   rofs_readdir,
   NULL, /* fm_signal */
   NULL, /* timeout */
   rofs_fstat,
   rofs_lseek
};


/*
 * Process entry point for ROFS volume manager
 */
static OS_PROCESS(ose_rofs_vm)
{
   vfs_mainloop(&rofs_if);
}


/*
 * Public function used to mount and start the ROFS volume.
 */
int
ose_rofs_mount_embedded(const char *mount_point)
{
   char * proc_name = NULL;
   PROCESS vm_pid = 0;
   int result;
   union SIGNAL *sig = NIL;

   if (mount_point == NULL)
   {
      ramlog_printf(
         "ROFS: Trying to mount with NULL pointer, volume not mounted.\n");
      result = EFAULT;
      goto return_result;
   }

   if (!rofs_found)
   {
      ramlog_printf(
         "ROFS: %s: No volume found. "
         "Has ose_rofs_start_handler0() been called?\n",
         mount_point);
      result = EIO;
      goto return_result;
   }

   /* Check correct syntax for mount_point. Must be "/<name>", where name
    * must be at least one character. */
   if ((mount_point[0] != '/') || (mount_point[1] == '\0')
       || (strchr(&mount_point[1], '/') != NULL))
   {
      ramlog_printf("ROFS: %s: Malformed name for mount point, must be "
                    "'/<name>'. Volume not mounted.\n",
                    mount_point);
      result = EINVAL;
      goto return_result;
   }

   proc_name = heap_alloc_unsafe(
      sizeof (ROFS_PROC_NAME_PREFIX) + strlen(mount_point), current_process(),
      __FILE__, __LINE__);

   if (proc_name == NULL)
   {
      ramlog_printf(    
         "ROFS: %s: Out of memory allocating %u bytes, volume not mounted.\n",
         rofs_mount_point,
         sizeof (ROFS_PROC_NAME_PREFIX) + strlen(mount_point));
      result = ENOMEM;
      goto return_result;
   }

   strcpy(proc_name, ROFS_PROC_NAME_PREFIX);
   strcat(proc_name, mount_point);

   /* Create the ROFS volume manager process.*/
   vm_pid = create_process(OS_PRI_PROC,    /* proc type.   */
                                     proc_name, /* name. */
                                     ose_rofs_vm,   /* entrypoint.  */
                                     ROFS_VM_STACK_SIZE_BYTES,
                                     get_pri(current_process()), 
                                     0,              /* timeslice.   */
                                     0,              /* block.       */
                                     NULL,           /* redir table. */
                                     0,              /* vector.      */
                                     0);             /* user id.     */

   rofs_mount_point = heap_alloc_unsafe(
      strlen(mount_point) + 1, vm_pid, __FILE__, __LINE__);

   if (rofs_mount_point == 0)
   {
      ramlog_printf(    
         "ROFS: %s: Out of memory allocating %u bytes, volume not mounted.\n",
         mount_point, strlen(mount_point) + 1);
      result = ENOMEM;
      goto return_result;
   }

   strcpy(rofs_mount_point, mount_point);

   rofs_if.name = rofs_mount_point;

   start(vm_pid);

   /* Check that the volume is correctly registered to FSS */
   {
      static const SIGSELECT sel_sigs[] =
         {2, OS_ATTACH_SIG, FSS_RESOLVE_REPLY};

      OSATTREF vm_attach;
      OSATTREF fss_attach;
      PROCESS fss_pid = 0;

      if (!hunt("ose_fss", 0, &fss_pid, NULL))
      {
         ramlog_printf(
            "ROFS: %s: Could not find process ose_fss, volume not mounted.\n",
            mount_point);
         result = ESTALE;
         goto return_result;
      }

      fss_attach = attach(NULL, fss_pid);
      vm_attach = attach(NULL, vm_pid);

      sig = alloc(sizeof (struct FssResolve) + strlen(rofs_mount_point),
                  FSS_RESOLVE_REQUEST);

      sig->sig_resolve.type = FSS_TYPE_VOLUME;
      sig->sig_resolve.timeOut = (U32) FSS_TMO_FOREVER;
      strcpy(sig->sig_resolve.name, rofs_mount_point);

      send(&sig, fss_pid);

      sig = receive(sel_sigs);

      if (sig->sig_no == FSS_RESOLVE_REPLY)
      {
         detach(&vm_attach);
         detach(&fss_attach);
         /* Check it is the volume we expected, or if another volume is using
          * the same name. */
         if (sender(&sig) != vm_pid)
         {
            ramlog_printf(
               "ROFS: %s: Mount point already mounted by process %x, "
               "volume not mounted.\n",
               mount_point);
            kill_proc(vm_pid);
            vm_pid = 0;
            result = ESTALE;
            goto return_result;
         }
      }
      else /* is OS_ATTACH_SIG */
      {
         if (sender(&sig) == vm_pid)
         {
            vm_pid = 0;
            detach(&fss_attach);
            ramlog_printf("ROFS: %s: Volume died.\n", mount_point);
         }
         else
         {
            fss_pid = 0;
            detach(&vm_pid);
            ramlog_printf("ROFS: %s: ose_fss process died, volume not "
                          "mounted.\n", mount_point);
         }
         result = ESTALE;
         goto return_result;
      }
   }

   result = 0;

  return_result:

   if (proc_name != NULL)
   {
      free(proc_name);
   }

   if (sig != NIL)
   {
      free_buf(&sig);
   }

   if ((vm_pid != 0) && (result != 0))
   {
      kill_proc(vm_pid);
      vm_pid = 0;
   }

   return result; /* returns result code insted of ZERO! */
}
