/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 1999-2012 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */
/*
 ****************************************************************************
 *			      PROGRAM MODULE
 *
 *     $Workfile:   rtcq.c  $
 *     Document no: @(#) 511/OSE90-8 $FileRevision:   /main/tb_current_ose5/5  $
 *     $Author:   thol  $
 *     $Date:   06/02/06 08:42:55  $
 *
 *
 ***************************************************************************
 * DESCRIPTION:
 *
 ***************************************************************************
 * HISTORY OF DEVELOPMENT:
 *
 * OSEhq07188,matl20020217
 *    rtc_cancel_alarm() will now go to error if wrong id been used.
 *    Previous behaviour were to end in an forever loop.
 *    ! struct RtcCancel modified!
 *    
 * OSEhq07381,matl20020128
 *    Changes made to get the old OSE4.1 behaviour back.
 *    Expired alarm are now returned to client.
 *    Changes made in signalSetTime() and request_alarm(). 
 *    A NULL termination added in insert_alarm().
 *    If time is adjusted forward, A periodical alarm will only send the first
 *    expired alarm to client, the rest of them disappear.
 ***************************************************************************/

/* Include public interfaces first to ensure they are valid "stand-alone" */
#include "rtc.h"
#include "rtc.sig"

#include "stddef.h"
#include "assert.h"
#include "string.h"
#include "time.h"

#include "bios.h"
#include "ose.h"
#include "ose_err.h"
#include "osetypes.h"
#include "ose_spi/spinlock.h"
#include "systimer.h"

#include "rtcerr.h"
#include "rtc_private.sig"


union SIGNAL
{
  SIGSELECT            sigNo;
  struct RtcAlarm      alarm;
  struct RtcRequest    request;
  struct RtcCancel_32  cancel_32;
#ifdef OSE64
  struct RtcCancel_64  cancel_64;
#endif
  struct RtcGetTime    getTime;
  struct RtcSetTime    setTime;
  struct RtcAdjustTime adjustTime;
  struct RtcGetTzInfo  getTzInfo;
  struct RtcGetLsInfo  getLsInfo;
  struct RtcGetScale   getScale;
  struct RtcSetScale   setScale;
  struct RtcT2subTypeCorr t2subTypeCorr;

  struct RtcConfig     config;
  struct RtcSetLocalOffset setLocalOffset;
  struct RtcGetDstDiff getDstDiff;
};

#include "timezone.c"

/* ----------------------------------------------------
   --- Defines
   --------------------------------------------------*/

#if 0
 #define RTC_DEBUG
 #define ASSERT
 #define RTC_DEBUG2
 #define RTC_DEBUG_MADNESS
 #define USE_DEBUG_PRINTF
 #include "dbgprintf.h"
 #define printf dbgprintf
#endif

#ifdef RTC_DEBUG
#include "stdio.h"
#endif

#define FIND_AI_ID        0
#define FIND_AI_EXP       1
#define FIND_AI_PID       2
#define ALARM_ID_INVALID  ~0U

#define SCALE_NORMAL      1000000000LL
#define SCALE_TOP         2000000000
#define SCALE_BOT         1

#define CHUNK_SIZE        10

#define GET_SYSTIME(m) ((g.get_systime_whitebox_callback == NULL) ? \
                        get_systime(m) : g.get_systime_whitebox_callback(m))


/* ----------------------------------------------------
   --- Internal variables
   ----------------------------------------------------
*/

#ifdef OSE_SMP
static spinlock_t rtc_spinlock = SPIN_LOCK_INIT;
#endif

static struct RtcGlobal    g = { 0 };     /* declared in rtc_private.h */
static struct timeZoneInfo tzInfo = {(struct state *)NIL, 0, "GMT"};

extern PROCESS ose_rtc_;

/* ----------------------------------------------------
   --- Alarm struct
   --------------------------------------------------*/

struct alarm_info
{
   struct alarm_info *next; 
   U32 id;
   time_t expires;
   U32 period;
   PROCESS client;
   OSATTREF att_ref;
#ifdef ASSERT
   int used;
#endif
   union SIGNAL *alarm_sig;
};

static struct alarm_info *alarm_head = NULL;
static struct alarm_info *free_list = NULL;
static U32 alarm_id_cntr = 0;
OS_PROCESS(ose_rtc_adjust_time);
/* ----------------------------------------------------
   --- alloc_request_chunk
   ---
   --- Allocates a chunk (size CHUNK_SIZE) of Request's
   --- and link them together in a singly linked list
   --- suitable for the alarm_info freeList.
   --------------------------------------------------*/

static struct alarm_info *
alloc_request_chunk(void)
{
  int i;
  struct alarm_info *chunk =
    (struct alarm_info *)alloc(CHUNK_SIZE*sizeof(struct alarm_info),0);

#ifdef ASSERT
  /* Fill requests with garbage. */
  memset(chunk, 0xff, CHUNK_SIZE * sizeof (struct alarm_info));
#endif
  for (i = 0; i < CHUNK_SIZE; i++)
  {
    chunk[i].next = &chunk[i + 1];
#ifdef ASSERT
    chunk[i].used = False;
#endif
    chunk[i].id = alarm_id_cntr++;
  }
  chunk[CHUNK_SIZE - 1].next = NULL;
  return &chunk[0];
}

/* ----------------------------------------------------
   --- get_free_request
   ---
   --- Return an unused Request. Allocate more 
   --- Requests if free list is empty.
   --------------------------------------------------*/

static struct alarm_info *
get_free_request(void)
{
  struct alarm_info *req;

  if (free_list == NULL)
    free_list = alloc_request_chunk();

  req = free_list;
  assert(req != NULL);
  assert(req->used == False);
  free_list = free_list->next;

  if (req->id == ALARM_ID_INVALID)
    req->id = alarm_id_cntr++;

  return req;
}

/* ----------------------------------------------------
   --- Alarm list functions
   --------------------------------------------------*/

static void
insert_alarm(struct alarm_info *ai)
{
   /* Insert sorted */
   struct alarm_info *tai, *otai;
   assert(ai->next == NULL);

   tai = alarm_head;
   otai = NULL;
   while (tai != NULL)
   {
      if (tai->expires < ai->expires)
      {
         otai = tai;
         tai = tai->next;
      }
      else
      {
         break;
      }
   }
   if (tai == NULL && otai == NULL)
   {  /* Empty list */
      alarm_head = ai;
   }
   else if (tai != NULL && otai == NULL)
   {  /* Beginning of list */
      ai->next = tai;
      alarm_head = ai; 
   }
   else if (tai == NULL && otai != NULL) 
   {  /* End of list */
      otai->next = ai;
      ai->next = NULL;
   }
   else
   {  /* Middle of list */
      otai->next = ai;
      ai->next = tai;
   }
}

static struct alarm_info *
new_alarm(time_t expires, U32 period, PROCESS pid,
          union SIGNAL *alarm_sig)
{
   struct alarm_info *ai = get_free_request();
   if (ai == NULL)
      return NULL;
   ai->expires = expires;
   ai->period = period;
   ai->client = pid;
   ai->att_ref = attach(NULL,pid);
   ai->alarm_sig = alarm_sig;
   ai->next = NULL;
   insert_alarm(ai);
   return ai;
}

static void
remove_alarm(struct alarm_info *ai, struct alarm_info *oai)
{
   if (ai == alarm_head)
      alarm_head = ai->next;
   else if (ai->next == NULL)
      oai->next = NULL;
   else
      oai->next = ai->next;
}

static struct alarm_info *
find_alarm(U32 id,
           time_t expires,
           PROCESS pid,
           int mode,
           struct alarm_info **oai)
{
   struct alarm_info *ai = alarm_head;
   *oai = ai;

   while (ai != NULL)
   {
      switch(mode)
      {
         case FIND_AI_ID:
            if (ai->id == id)
               return ai;
            break;
         case FIND_AI_EXP:
            if (ai->expires <= expires)
               return ai;
            else
               return NULL;        /* List sorted after expires */
            /*break; avoid compiler warning about unreachable */
         case FIND_AI_PID:
            if (ai->client == pid)
               return ai;
            /*break; avoid compiler warning about unreachable */
         default:
            return NULL;
            /*break; avoid compiler warning about unreachable */
      }

      *oai = ai;
      ai = ai->next;
   }
   return NULL;
}

static void
replace_alarm(U32 id)
{
   struct alarm_info *ai, *oai;

   ai = find_alarm(id,0,0,FIND_AI_ID,&oai);
   if (ai == NULL)
      return;
   remove_alarm(ai,oai);
   ai->next = NULL; /*This is important if the alarm is inserted at the end.*/
   insert_alarm(ai);
}
#define DEL_DETACH   1
#define DEL_NDETACH  0
#define DEL_FREEBUF  1
#define DEL_NFREEBUF 0

static int
del_alarm(U32 id, time_t expires, PROCESS pid, int mode,
          int do_detach, int do_freebuf)
{
   struct alarm_info *ai,*oai;

   ai = find_alarm(id,expires,pid,mode,&oai);
   if (ai == NULL)
      return 0;

   if (do_detach && ai->att_ref != 0)
      detach(&ai->att_ref);

   if (do_freebuf)
      free_buf(&ai->alarm_sig);

   remove_alarm(ai,oai);

   /* OSEhq12611, expired alarms shall not be reused with old ids. */
   ai->id = ALARM_ID_INVALID;

   /* Put alarm_info back into free list */
   ai->next = free_list;
   free_list = ai;

   return 1;
}

#ifdef RTC_DEBUG
static void
print_alarm_tree(void)
{
   struct alarm_info *ai = alarm_head;

   printf("--(Alarm Tree)----\n");
   while (ai != NULL)
   {
      printf("* (id)0x%x (exp)0x%x (per)0x%x (own)0x%x \n",
             ai->id, ai->expires, ai->period, ai->client);
      ai = ai->next;
   }
   printf("--(Free List)----\n");
   ai = free_list;
   while(ai != NULL)
   {
      printf("* (id)0x%x \n", ai->id);
      ai = ai->next;
   }
   printf("------------------\n");
}
#endif

/* ----------------------------------------------------
   --- assert_rtc
 *
 * Called by:   RTC server process.
   --------------------------------------------------*/

#ifdef ASSERT
static void
assert_rtc(void)
{
   /* Check request free list for consistency. */
   {
      struct alarm_info *req = free_list;
      while (req != NULL)
      {
         assert(req->used == False);
         req = req->next;
      }
   }
}
#else
#define assert_rtc()
#endif

/* ----------------------------------------------------
   --- client_cleanup
   ---
   --- A client died. Nuke it's alarms.
   --------------------------------------------------*/

static void
client_cleanup(PROCESS att_pid)
{
#ifdef RTC_DEBUG
   print_alarm_tree();
#endif
   while( del_alarm(0, 0, att_pid, FIND_AI_PID, DEL_NDETACH, DEL_FREEBUF) )
      (void)0;
#ifdef RTC_DEBUG
   print_alarm_tree();
#endif
}

/*
 *===========================================================================
 *                request_alarm
 *===========================================================================
 * Description: Implements the request of an alarm.
 *
 * Parameters:  alarmInfo -
 *              expires   -
 *              period    -
 *              client    -
 *              alarmSig  -
 *
 * Returns:
 *
 * Called by:   RTC server process.
 */

static void
request_alarm(union SIGNAL *reply, time_t expires, U32 period,
              PROCESS client, union SIGNAL *alarmSig)
{
   struct alarm_info *ai;

   assert_rtc();

   ai = new_alarm(expires, period, client, alarmSig);
   if (ai != NULL)
   {
#ifdef OSE64
      /* Give back cancel information. */
      if (reply->sigNo == RTC_CANCEL_32)
      {
#endif
         reply->cancel_32.alarmInfo.sigNo = alarmSig->alarm.sigNo;
         reply->cancel_32.alarmInfo.instance = ai->id;
         reply->cancel_32.alarmInfo.pointer = (U32)(OSADDRESS)ai;
#ifdef OSE64
      }
      else
      {
         reply->cancel_64.alarmInfo.sigNo = alarmSig->alarm.sigNo;
         reply->cancel_64.alarmInfo.instance = ai->id;
         reply->cancel_64.alarmInfo.pointer = (OSADDRESS)ai;
      }
#endif
   }
   assert_rtc();

#ifdef RTC_DEBUG
   print_alarm_tree();
#endif
}

/* ----------------------------------------------------
   --- find_if_sig_in_que
   ---
   --- try to find if a signal is in a processes signal queue
   --------------------------------------------------*/
static OSBOOLEAN
find_if_sig_in_que(SIGSELECT instance, OSADDRESS pointer)
{
   OSADDRESS next;
   OSBOOLEAN err;
   union SIGNAL *sig = NULL;
   struct OS_pcb *pcb;
   struct alarm_info *ai = (struct alarm_info *)pointer;

   /* may come here with a bad alarminfo */
   if ((ai->id == ALARM_ID_INVALID) || (ai->id != instance))
   {
      return (OSBOOLEAN) 0;
   }
   /* now we know that it was not possible to delete the alarm */
   pcb = get_pcb(ai->client);
   if (pcb->sig_cnt_in_q == 0)
   {
      free_buf((union SIGNAL **)&pcb);
      return (OSBOOLEAN) 0;
   }
   next = pcb->sigqueue;
   free_buf((union SIGNAL **)&pcb);
   err = get_signal(ai->client, next, &sig, &next);
   if (!err && sig)
   {
      do
      {
         if ((ai->alarm_sig->sigNo == sig->sigNo)
             && (ai->id == sig->alarm.instance))
         {
            /* Found the cancel signal in queue */
            free_buf(&sig);
            return (OSBOOLEAN) 1;
         }
         free_buf(&sig);
         if (next)
         {
            err = get_signal(ai->client, next, &sig, &next);
         }
         else
         {
            /* no more signals in queue */
            break;
         }
      } while (!err && sig );
   }

   return (OSBOOLEAN) 0;
}
/*
 *===========================================================================
 *                cancel_alarm
 *===========================================================================
 * Description: Implements the cancellation of an alarm.
 *
 * Parameters:  alarmInfo - identifies the alarm to cancel.
 *
 * Returns:
 *
 * Called by:   RTC server process.
 */


static S32
cancel_alarm(SIGSELECT instance, OSADDRESS pointer)
{
   S32 info = 0;
   struct alarm_info *ai = (struct alarm_info *)pointer;

   assert_rtc();
   if (ai->id != ALARM_ID_INVALID && ai->id == instance)
   {
      /* Remove request from second slot's queue. */
      if (del_alarm(ai->id, 0, 0, FIND_AI_ID, DEL_DETACH, DEL_NFREEBUF))
      {
         /* Send back client's signal.
            If alarm was requested via the signal interface it is the
            proxy process that owns the signal. Still no need for restore
            since it is the proxy process executing cancel_alarm().
            Proxy sends alarm signal to (dead or alive) client.
         */

         /* ai put on free list by del_alarm(), but referenced here again
          * in server context, but then no more.
          */
         /* For safe cancellation. */
         ai->alarm_sig->alarm.instance = instance;
         /* And a proper time stamp. */
         ai->alarm_sig->alarm.time = g.ref_time.real_time.seconds;
         send_w_s(&ai->alarm_sig, ose_rtc_, ai->client);
         /* Alarm signal in client's signal queue. */
         /* Trash instance variable. */
         ai->id = ALARM_ID_INVALID;
      }
      else
      {
         info = OSE_RTC_EWILD_ALARM_ID;
      }
   }
   else
   {
      info = OSE_RTC_EWILD_ALARM_ID;
   }

   /* if the client has an alarm in his q all is fine (tr13305) */
   if ((info == OSE_RTC_EWILD_ALARM_ID) &&
       find_if_sig_in_que(instance, pointer))
   {
      info = 0;
   }

   assert_rtc();

#ifdef RTC_DEBUG
   print_alarm_tree();
#endif

   return info;
}



/*
*  get_time()
*
*  Provide a new TimeRef that corresponds to the current systime.
*
*  The difference between current systime and g.ref_time systime is
*  scaled by the scale factor and added to the g.ref_time.real_time
*  to get a new time reference.
*
*  OSECR-13744
*  The algorithm must use signed arithmetic to avoid overflow errors
*  in the unusual situation when get_systime returns a time that is
*  before the systime in g.ref_time.
*/
static void
get_time(struct TimeRef *out_time)
{
   S32 delta_ticks;
   S32 delta_micros;

   /*
   *  Use *out_time to store a copy of g.ref_time.
   *  g.ref_time must be accessed within a spin locked region to
   *  ensure an atomic copy.
   */
   {
      Msr msr;

      spin_lock_irqsave(&rtc_spinlock, msr);
      *out_time = g.ref_time;
      spin_unlock_irqrestore(&rtc_spinlock, msr);
   }

   /*
   *  Read current systime and compute delta ticks and micros from the
   *  reference we got from g.ref_time. The systime in out_time isn't
   *  needed anymore. Update out_time with current systime
   */
   {   
      OSTICK now_ticks, now_micros;

      now_ticks = GET_SYSTIME(&now_micros);

      delta_ticks = (S32)(now_ticks - out_time->systime_ticks);
      delta_micros = (S32)(now_micros - out_time->systime_micros);

      out_time->systime_ticks = now_ticks;
      out_time->systime_micros = now_micros;
   }

   /*
   *  Combine delta_ticks and delta_micros into microseconds, and apply
   *  the scale factor to the combined delta.
   *  We need 64-bit arithmetic to for the scale conversion.
   */
   {
      long long scaled_micros;

      scaled_micros =
               (long long)(delta_ticks * (S32)g.micros_per_tick + delta_micros);
      scaled_micros *= (long long)out_time->scale_factor;
      scaled_micros /= SCALE_NORMAL;

      delta_micros = (S32)scaled_micros;
   }

   /*
   *  Split delta_micros into seconds and microseconds, and add the result
   *  to the reference real_time in *out_time.
   *
   *  This conversion goes from signed to unsigned, and we need to ensure
   *  that the microsecond value is positive.
   */
   {
      S32 delta_seconds;

      delta_micros += (S32)out_time->real_time.micros;
      delta_seconds = delta_micros / 1000000;
      delta_micros -= delta_seconds * 1000000;

      if (delta_micros < 0)
      {
         delta_micros += 1000000;
         delta_seconds -= 1;
      }

      out_time->real_time.seconds += (time_t)delta_seconds;
      out_time->real_time.micros = (U32)delta_micros;
      assert(out_time->real_time.micros < 1000000);
   }
}


static void
update_ref_time(struct TimeRef *new_time)
{
   /*
   *  Update g.ref_time within a spin locked region to ensure atomic copy.
   */
   {
      Msr msr;

      spin_lock_irqsave(&rtc_spinlock, msr);
      g.ref_time = *new_time;
      spin_unlock_irqrestore(&rtc_spinlock, msr);
   }

   /*
   *  Update the kernel/LDM "abstime" variables.
   *  See ose_spi.h set_abstime() for details.
   *  The purpose of kernel/LDM "abstime" is a similar time reference as
   *  g.ref_time, with the exception that it doesn't handle a scale factor.
   */
   {
      OSTICK micros, ticks;

      /* Convert new_time->real_time.micros into microseconds and ticks. */
      micros = new_time->real_time.micros;
      ticks = micros / g.micros_per_tick;
      micros -= ticks * g.micros_per_tick;

      /* Subtract from systime */
      ticks = new_time->systime_ticks - ticks;
      micros = new_time->systime_micros - micros;

      /* Handle underflow of the microseconds */
      if ((S32)micros < 0)
      {
         micros += g.micros_per_tick; /* Add one tick (in micros). */
         ticks -= 1; /* And subtract one tick. */
      }
      assert(ticks < g.micros_per_tick);

      set_abstime(new_time->real_time.seconds,
                  ticks,
                  micros);
   }
}


/*
 *===========================================================================
 *                trapGetTime
 *===========================================================================
 * Description:
 *
 * Parameters:  tpp - Pointer to struct TimePair to update, NULL if update
 *                    internal time.
 *
 * Returns:     Dummy value for BIOS interface.
 *
 * Called by:   RTC server process and client processes.
 */

static long
trapGetTime(struct TimePair *tpp)
{
   struct TimeRef time_now;

   get_time(&time_now);

   if (tpp != NULL)
   {
      *tpp = time_now.real_time;
   }

   return 0;
}

/*
 *===========================================================================
 *                signalRequest
 *===========================================================================
 * Description: Translate an alarm request through the signal interface to
 *              the internal function.
 *
 * Parameters:  sig - pointer to RtcRequest signal.
 *
 * Returns:     Nothing.
 *
 * Called by:   RTC server process.
 */

static void
signalRequest(union SIGNAL *sig)
{
   time_t expires = sig->request.expires;
   U32 period = sig->request.period;
   PROCESS client = sig->request.client;
   OSBUFSIZE size = sigsize(&sig) - offsetof(struct RtcRequest, alarmSig);
   union SIGNAL *alarmSig = alloc(size, 0);
   union SIGNAL *reply;

   memcpy(alarmSig, &sig->request.alarmSig, size);

#ifdef OSE64
   if (sig->sigNo == RTC_REQUEST_32)
   {
#endif
      reply = alloc(sizeof(struct RtcCancel_32), RTC_CANCEL_32);
#ifdef OSE64
   }
   else /* RTC_REQUEST_64 */
   {
      reply = alloc(sizeof(struct RtcCancel_64), RTC_CANCEL_64);
   }
#endif

   free_buf(&sig);

   request_alarm(reply, expires, period, client, alarmSig);

   /* Send back request signal with cancellation information. */
   send(&reply, client);
}

/*
 *===========================================================================
 *                signalCancel
 *===========================================================================
 * Description:
 *
 * Parameters:
 *
 * Returns:     Nothing.
 *
 * Called by:   RTC server process.
 */

static void
signalCancel(union SIGNAL *sig)
{
#ifdef OSE64
   if (sig->sigNo == RTC_CANCEL_32)
   {
#endif
      /* rtc_cancel_alarm can be of an older version without info in struct */
      OSBUFSIZE sigsize_rtc_cancel = sigsize(&sig);
      S32 info = cancel_alarm(sig->cancel_32.alarmInfo.instance,
                                  (OSADDRESS)sig->cancel_32.alarmInfo.pointer);

      if (sigsize_rtc_cancel == sizeof(struct RtcCancel_32))
      {
         sig->cancel_32.info = info;
      }
#ifdef OSE64
   }
   else /* RTC_CANCEL_64 */
   {
      sig->cancel_64.info =
         cancel_alarm(sig->cancel_64.alarmInfo.instance,
                      (OSADDRESS)sig->cancel_64.alarmInfo.pointer);
   }
#endif

   send(&sig, sender(&sig));
}

/*
 *===========================================================================
 *                signalGetTime
 *===========================================================================
 * Description:
 *
 * Parameters:
 *
 * Returns:     Nothing.
 *
 * Called by:   RTC server process.
 */

static void
signalGetTime(union SIGNAL *sig)
{
   trapGetTime(&sig->getTime.currentTime);
   send(&sig, sender(&sig));
}

/*
 *===========================================================================
 *                signalSetTime
 *===========================================================================
 * Description:
 *
 * Parameters:
 *
 * Returns:     Nothing.
 *
 * Called by:   RTC server process.
 */
static void
signalSetTime(union SIGNAL *sig)
{
   struct TimeRef time_now;

   get_time(&time_now);
   time_now.real_time = sig->setTime.newTime;
   update_ref_time(&time_now);

   /*
   *  Do not Kill historic alarms, let the normal expire_alarms() 
   *  send them back.
   */
   /* ML20020123 */
   send(&sig, sender(&sig));
}


/*
 *===========================================================================
 *                set/get_scale. SOASF
 *===========================================================================
 */

unsigned long bios_module_handle = ~0UL; /* Global state */

static unsigned long get_bh(void)
{
  if (bios_module_handle == ~0UL)
  {
     bios_module_handle = biosOpen(SYSTIMER_NAME);
  }

  return bios_module_handle;
}

static U32 iget_scale(void)
{
   unsigned long bh = get_bh();
   if (bh == 0)
      return g.ref_time.scale_factor;
   else
      return (U32)biosCall(bh, DEVICE_FCGETSCALE);
}

static U32 iset_scale(U32 new_scale)
{
   unsigned long bh = get_bh();
   if (bh == 0)
   { /* No installed module */
      g.ref_time.scale_factor = new_scale;
      if (g.ref_time.scale_factor > SCALE_TOP)
         g.ref_time.scale_factor = SCALE_TOP;
      else if (g.ref_time.scale_factor < SCALE_BOT)
         g.ref_time.scale_factor = SCALE_BOT;
      return g.ref_time.scale_factor;
   }
   else
      /* Module installed */
      return (U32)biosCall(bh, DEVICE_FCSETSCALE, new_scale);
}

static void signal_get_scale(union SIGNAL *sig)
{
   sig->getScale.scale = iget_scale();
   send(&sig, sender(&sig));
}

static void signal_set_scale(union SIGNAL *sig)
{
   sig->setScale.new_scale = iset_scale(sig->setScale.new_scale);
   send(&sig, sender(&sig));
}

/*
 *===========================================================================
 *                signalAdjustTime
 *===========================================================================
 * Description:
 *
 * Parameters:
 *
 * Returns:     Nothing.
 *
 * Called by:   RTC server process.
 */
static void
signalAdjustTime(union SIGNAL *sig)
{
   PROCESS ose_rtc_adjust_time_ =
      create_process(OS_PRI_PROC, "ose_rtc_adjust_time",
                     ose_rtc_adjust_time, 2000, 15, 0,
                     0, 0, 0, 0);

   start(ose_rtc_adjust_time_);

   /* The acknowledge signal will be send by ose_rtc_adjust_time. */
   send_w_s(&sig, sender(&sig), ose_rtc_adjust_time_);
}

/*
 *===========================================================================
 *                expire_alarms
 *===========================================================================
 * Description: Check if alarms has expired. Send back alarm signal 
 *              to client.
 *              Cleanup request or update for multiple alarms.
 *
 * Parameters:  None.
 *
 * Returns:     Nothing.
 *
 * Called by:   RTC server process.
 */

static void
expire_alarms(void)
{
   struct alarm_info *ai, *oai;
#ifdef RTC_DEBUG
   int dirty = 0;
#endif

   ai = find_alarm(0, g.ref_time.real_time.seconds, 0, FIND_AI_EXP, &oai);

   while (ai != NULL)
   {
      union SIGNAL *sig;
#ifdef RTC_DEBUG
      dirty = 1;
#endif
      /* Send alarm signal to client. */
      if (ai->period != 0) /* Multiple alarms. */
      {
         /* Send a copy of the alarm signal to the client process. */
         OSBUFSIZE size = sigsize(&ai->alarm_sig);
         sig = alloc(size, 0);
         memcpy(sig, ai->alarm_sig, size);
         sig->alarm.instance = ai->id; /* For safe cancellation. */
         sig->alarm.time = ai->expires; /* And a proper time stamp. */
         send(&sig, ai->client);
         /* Update request for multiple alarm. */

         ai->expires = g.ref_time.real_time.seconds + ai->period -
            (g.ref_time.real_time.seconds - ai->expires) % ai->period;

         replace_alarm(ai->id);
      }
      else /* Single alarm, send the original alarm signal. */
      {
         sig = ai->alarm_sig;
         sig->alarm.instance = ai->id; /* For safe cancellation. */
         sig->alarm.time = ai->expires; /* And a proper time stamp. */
         send(&sig, ai->client);

         del_alarm(ai->id, 0, 0, FIND_AI_ID, DEL_DETACH, DEL_NFREEBUF);
      }

      ai = find_alarm(0,
                      g.ref_time.real_time.seconds,
                      0,
                      FIND_AI_EXP,
                      &oai);
   }
#ifdef RTC_DEBUG
   if (dirty)
      print_alarm_tree();
#endif
}

/*
 *===========================================================================
 *                trapHandler
 *===========================================================================
 * Description: Route client requests through the BIOS to the corresponding
 *              function.
 *
 * Parameters:  fc         - Request function code.
 *              arg1..arg6 - Arguments for function.
 *
 * Returns:     Return value from called function.
 *
 * Called by:   Client processes.
 */

static long
trapHandler(long fc,
            long arg1,
            long arg2,
            long arg3,
            long arg4,
            long arg5,
            long arg6)
{
   long rc = -1;
   switch (fc)
   {
      case RTC_GET_TIME :
         /* Handles both 32-bit and 64-bit clients. */
         rc = trapGetTime((struct TimePair *)(OSADDRESS)arg1);
         break;
      case RTC_GET_TZ_INFO_REQUEST:
         /* Handles both 32-bit and 64-bit clients. */
         rc = trapGetTzInfo((time_t)arg1,
                            (S32 *)(OSADDRESS)arg2,
                            (S32 *)(OSADDRESS)arg3,
                            (char *)(OSADDRESS)arg4,
                            &tzInfo);
         break;
      case RTC_GET_LS_INFO_REQUEST:
         /* Handles both 32-bit and 64-bit clients. */
         rc = trapGetLsInfo((time_t)arg1,
                            (S32 *)(OSADDRESS)arg2,
                            (S32 *)(OSADDRESS)arg3,
                            &tzInfo);
         break;
      case RTC_T2SUB_TYPE_CORR_REQUEST:
         /* Handles both 32-bit and 64-bit clients. */
         rc = trapTime2subTypeCorr((time_t *)arg1,
                                   (S32)arg2,
                                   (S32)arg3,
                                   &tzInfo);
         break;
      case RTC_GET_DST_DIFF_REQUEST:
         /* Handles both 32-bit and 64-bit clients. */
         rc = trapGetDstDiff((time_t)arg1,
                             (S32)arg2,
                             (S32 *)(OSADDRESS)arg3,
                             &tzInfo);
         break;
      case RTC_WHITEBOX_GET_VERSIONED_GLOBALS:
         rc = (long)&g;
         break;
      default:
         error2(OSE_RTC_EUNKNOWN_FCODE, (OSERRCODE)fc);
         (void) arg5;
         (void) arg6;
         break;
   }

   return rc;
}

/*
 *===========================================================================
 *          init_rtc()
 *===========================================================================
 *
 * Description: Initialises the real time clock.
 *
 * Parameters:  None.
 *
 * Returns:     Nothing.
 *
 * Called by:   ose_rtc.
 */

static void
init_rtc(void)
{
   long err;

   err = biosInstall("RTC", trapHandler,
                     BIOS_USERMODE_ALLOWED
#if (OSE_VERSION >= 570)
#ifdef OSE64
                     | BIOS_64_BIT
#endif
                     | BIOS_32_BIT
#endif
                     );
   if (err != BIOS_SUCCESS)
   {
      error2(OSE_EFATAL_MASK | OSE_RTC_EBIOS_INSTALL_FAILED, (OSERRCODE)err);
   }

#ifdef LIBWHAT
   LIBWHAT
#endif

   g.micros_per_tick = system_tick();
   {
      /* Avoid warning " cast discards qualifiers from pointer target type" */
      OSTICK tmp;
      g.ref_time.systime_ticks = GET_SYSTIME(&tmp);
      g.ref_time.systime_micros = tmp;
   }

   g.ref_time.scale_factor = SCALE_NORMAL;
   g.ref_time.real_time.seconds = 0;
   g.ref_time.real_time.micros = 0;

   assert_rtc();
}

/*
 *===========================================================================
 *          ose_rtc_adjust_time
 *===========================================================================
 *  ose_rtc_adjust_time is a dynamic process invoked my signalAdjusttime()
 *
 *
 *
 */
OS_PROCESS(ose_rtc_adjust_time)
{
#define DELAY_TIME 1000
   static const SIGSELECT sigsel[] = { 0 };
   PROCESS sender_;
   long toAdjust;
   static U32 scaleAtStart = 0;
   union SIGNAL *sig;

   sig = receive(sigsel);
   sender_ = sender(&sig);

   if (scaleAtStart)/*Another adjust_time session is in progress, die!*/
   {
      send_w_s(&sig,ose_rtc_, sender_);/*Send acknowledge.*/
      kill_proc(current_process());
   }
   scaleAtStart = iget_scale();

   if (*(SIGSELECT *)sig != RTC_ADJUST_TIME)
   {
      error2(OSE_RTC_EUNKNOWN_SIGNAL, (OSERRCODE)sig);
      free_buf(&sig);
      kill_proc(current_process());
   }
   toAdjust = sig->adjustTime.micros / 1000000; /*sec*/
   send_w_s(&sig,ose_rtc_, sender_);
   if (toAdjust > 1 )
   {
      iset_scale(SCALE_TOP);
   }
   else if (toAdjust < -1)
   {
      iset_scale(SCALE_BOT);
   }
   else
   {
      scaleAtStart = 0;
      kill_proc(current_process()); /*Adjustments under 1 sec not allowed
				      Since the RTC itself is not more accurate
				      than One second.*/
   }
   if (toAdjust > 1)
   {
      while (toAdjust > 1 )
      {
	 delay(DELAY_TIME);
	 toAdjust -= DELAY_TIME/1000 * (SCALE_TOP/SCALE_NORMAL -
					 scaleAtStart/SCALE_NORMAL);
      }
   }
   else
   {
      while (toAdjust < -1 )
      {
	 delay(DELAY_TIME);
	 toAdjust += DELAY_TIME/1000 * (scaleAtStart/SCALE_NORMAL);
      }
   }

   /* Trim the g.ref_time.scale_factor or back to normal?? */
   iset_scale(scaleAtStart);
   scaleAtStart = 0;
   kill_proc(current_process());  /*Job done!*/
#undef DELAY_TIME
}

/*
 *===========================================================================
 *          ose_rtc
 *===========================================================================
 * ose_rtc is the RTC server process that (memory protected) clients hunts
 * for (via external process in osemain.con). It will function as a proxy and
 * translate signal requests into function calls corresponding to the trap
 * interface.
 */

OSENTRYPOINT ose_rtc;

OS_PROCESS(ose_rtc)
{
   OSTIME ms_per_tick;
   OSTIME tick_per_sec;
   OSTIME tmo;
   OSTICK curr_tick;
   OSTICK next_tmo_tick; /* the tick value when next timeout expires */
   OSTICK diff_tick;

   init_rtc();
   ms_per_tick  = system_tick() / 1000;
   tick_per_sec = 1000 / ms_per_tick;
   next_tmo_tick = get_ticks() + tick_per_sec;

   for (;;)
   {
      static const SIGSELECT sigsel[] = { 0 };
      union SIGNAL *sig;

      /* Calculate the time that remains to the next whole second. */
      curr_tick = get_ticks();

      /*
      *  Two complements aritmetic handles the situation when 
      *  next_tmo_tick < curr_tick
      */
      diff_tick = next_tmo_tick - curr_tick;

      /* if 1 sec late or more, timeout immidiately to catch up the time */
      if (diff_tick > tick_per_sec)
      {
         tmo = 0;
      }
      else
      {
         tmo = ms_per_tick * diff_tick;
      }
      sig = receive_w_tmo(tmo, sigsel);

      /* Update the internal clock variables for new synchronization point. */
      {
         struct TimeRef time_now;

         get_time(&time_now);
         update_ref_time(&time_now);
      }

      if (sig == NULL)
      {
#if defined(RTC_DEBUG) && defined(RTC_DEBUG_MADNESS)
         printf("** RTC: One second passed: 0x%x\n",
                g.ref_time.real_time.seconds);
#endif
         /*
         *  Two complements aritmetic handles the situation when
         *  next_tmo_tick wrap
         */
         next_tmo_tick = next_tmo_tick + tick_per_sec;
         /* Time out */
         expire_alarms();
      }
      else
      {
         switch (sig->sigNo)
         {
            case RTC_CONFIG_REQUEST :
               signalConfigRequest(sig, &tzInfo);
               break;
            case RTC_SET_LOCAL_OFFSET_REQUEST :
               signalSetLocalOffsetRequest(sig, &tzInfo);
               break;
#ifdef OSE64
            case RTC_REQUEST_64:
#endif
            case RTC_REQUEST_32:
               signalRequest(sig);
               break;
#ifdef OSE64
            case RTC_CANCEL_64:
#endif
            case RTC_CANCEL_32:
               signalCancel(sig);
               break;
            case RTC_GET_TIME :
               signalGetTime(sig);
               break;
            case RTC_SET_TIME :
               signalSetTime(sig);
               break;
            case RTC_ADJUST_TIME :
               signalAdjustTime(sig);
               break;
            case RTC_GET_TZ_INFO_REQUEST :
               signalGetTzInfo(sig, &tzInfo);
               break;
            case RTC_GET_LS_INFO_REQUEST :
               signalGetLsInfo(sig, &tzInfo);
               break;
            case RTC_T2SUB_TYPE_CORR_REQUEST:
               signalT2subTypeCorr(sig, &tzInfo);
               break;
            case RTC_GET_DST_DIFF_REQUEST:
               signalGetDstDiff(sig, &tzInfo);
               break;
            case RTC_GET_SCALE:
               signal_get_scale(sig);
               break;
            case RTC_SET_SCALE:
               signal_set_scale(sig);
               break;
            case RTC_READY :
               send(&sig, sender(&sig));
               break;
            case OS_ATTACH_SIG : /* Client Terminated */
               client_cleanup(sender(&sig));
               free_buf(&sig);
               break;
            default :
               error2(OSE_RTC_EUNKNOWN_SIGNAL, (OSERRCODE)sig);
               free_buf(&sig);
               break;
         }
      }
   }
}

/*
****************************************************************************
*                         END OF FILE
****************************************************************************
*/
