/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2000-2012 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */
/*
 ****************************************************************************
 *                PROGRAM MODULE
 *
 *     $Workfile:   timezone.c  $
 *     Document no: @(#) 511/OSE90-6 $Revision:   1.2  $
 *     $Author:   olli  $
 *     $Date:   07/02/00 23:39:00  $
 *
 *
 ***************************************************************************/

/* rtcq.c includes this file */

#include "assert.h"

#ifdef  RTC_DEBUG
#include "stdio.h"
#endif

#include "string.h"
#include "time.h"

#include "ose.h"
#include "ose_err.h"
#include "osetypes.h"

#include "rtc_private.h"
#include "tzfile.h"
#include "timezone.h"

#include "rtc.sig"
#include "rtc_private.sig"

struct rule
{
   int		r_type;		/* type of rule--see below */
   int		r_day;		/* day number of rule */
   int		r_week;		/* week number of rule */
   int		r_mon;		/* month number of rule */
   long		r_time;		/* transition time of rule */
};

#define JULIAN_DAY		0	/* Jn - Julian day */
#define DAY_OF_YEAR		1	/* n - day of year */
#define MONTH_NTH_DAY_OF_WEEK	2	/* Mm.n.d - month, week, day of week */



#ifdef  RTC_DEBUG
static void printTzInfo(struct timeZoneInfo * const tzInfo) 
{
   int i;
   printf("tzInfo:\n");
   if (tzInfo->sp == NULL)
   {
      printf(" gmtOffset = %d\n",tzInfo->gmtOffset);
   }
   else
   {
      printf(" leapcnt = %d\n",tzInfo->sp->leapcnt);
      printf(" timecnt = %d\n",tzInfo->sp->timecnt);
      printf(" typecnt = %d\n",tzInfo->sp->typecnt);
      printf(" charcnt = %d\n",tzInfo->sp->charcnt);
      for (i=0; i<tzInfo->sp->timecnt; i++)
      {
         printf(" Transition %d, type = %d, time = %s",
		i,
		tzInfo->sp->types[i],
		asctime(gmtime(&(tzInfo->sp->ats[i]))));
      }
      for (i=0; i<tzInfo->sp->typecnt; i++)
      {
         printf(" Type %d, abbr = %4s, gmt offset = %d, ",
		i,
		&(tzInfo->sp->chars[tzInfo->sp->ttis[i].tt_abbrind]),
		tzInfo->sp->ttis[i].tt_gmtoff);
         printf("isdst = %d, isstd = %d, isgmt = %d\n",
		tzInfo->sp->ttis[i].tt_isdst,
		tzInfo->sp->ttis[i].tt_ttisstd,
		tzInfo->sp->ttis[i].tt_ttisgmt);
      }
      for (i=0; i<tzInfo->sp->leapcnt; i++)
      {
         printf(" Leap %d, corr = %d, time = %s",
		i,
		tzInfo->sp->lsis[i].ls_corr,
		asctime(gmtime(&(tzInfo->sp->lsis[i].ls_trans))));
      }
      printf("\n\n");
   }
}
#endif



/*
 *===========================================================================
 *                           detzcode
 *===========================================================================
 *
 * Description: Converts an array of four 8 bit char into an S32
 *
 * Parameters:  codep  - Pointer to the array 
 *
 * Returns:     32-bit signed integer
 */
static S32
detzcode(const char *codep)
{
   S32 result = 0;
   int i;

   for (i = 0; i < 4; ++i)
   {
      result = (result << 8) | (codep[i] & 0xff);
   }
   return result;
}



/*
 *===========================================================================
 *                           tzload
 *===========================================================================
 *
 * Description: Read the infile and store the data in a state' struct.
 *
 * Parameters:  tzData  The "infile"
 *              size    Size of tzData.
 *              sp    - pointer to the state struct where the data will be 
 *                      stored
 *
 * Returns:      0      If everything is OK
 *              -1      Otherwise
 */

#include "heapapi.h"

static int
tzload(U8 *tzData, size_t size, struct state *sp)
{
   const char *p;
   int i;

   struct tzhead *tzhp;

   union
   {
      struct tzhead tzhead;
      char buf[sizeof *sp + sizeof *tzhp];
   } *up;
#define u (*up)
   int ttisstdcnt;
   int ttisgmtcnt;

   if (size == 0)
      return -1;
   i = size;

   up = heap_alloc_private(sizeof *up, 0, 0);
   memcpy(u.buf, tzData, (sizeof *up) > size ? size : sizeof* up);

   ttisstdcnt = (int) detzcode(u.tzhead.tzh_ttisgmtcnt);
   ttisgmtcnt = (int) detzcode(u.tzhead.tzh_ttisstdcnt);
   sp->leapcnt = (int) detzcode(u.tzhead.tzh_leapcnt);
   sp->timecnt = (int) detzcode(u.tzhead.tzh_timecnt);
   sp->typecnt = (int) detzcode(u.tzhead.tzh_typecnt);
   sp->charcnt = (int) detzcode(u.tzhead.tzh_charcnt);
   p = u.tzhead.tzh_charcnt + sizeof u.tzhead.tzh_charcnt;
   if (sp->leapcnt < 0 || sp->leapcnt > TZ_MAX_LEAPS ||
       sp->typecnt <= 0 || sp->typecnt > TZ_MAX_TYPES ||
       sp->timecnt < 0 || sp->timecnt > TZ_MAX_TIMES ||
       sp->charcnt < 0 || sp->charcnt > TZ_MAX_CHARS ||
       (ttisstdcnt != sp->typecnt && ttisstdcnt != 0) ||
       (ttisgmtcnt != sp->typecnt && ttisgmtcnt != 0))
   {
#if 0
      return -1;
#else
      goto fail;
#endif
   }

   if (i - (p - u.buf) < sp->timecnt * 4 +	/* ats */
       sp->timecnt +			/* types */
       sp->typecnt * (4 + 2) +		/* ttinfos */
       sp->charcnt +			/* chars */
       sp->leapcnt * (4 + 4) +		/* lsinfos */
       ttisstdcnt +			/* ttisstds */
       ttisgmtcnt)			/* ttisgmts */
   {
#if 0
      return -1;
#else
      goto fail;
   }
#endif

   for (i = 0; i < sp->timecnt; ++i)
   {
      sp->ats[i] = detzcode(p);
      p += 4;
   }
   for (i = 0; i < sp->timecnt; ++i)
   {
      sp->types[i] = (unsigned char) *p++;
      if (sp->types[i] >= sp->typecnt)
      {
#if 0
         return -1;
#else
         goto fail;
#endif
      }
   }

   for (i = 0; i < sp->typecnt; ++i)
   {
      register struct ttinfo *ttisp;

      ttisp = &sp->ttis[i];
      ttisp->tt_gmtoff = detzcode(p);
      p += 4;
      ttisp->tt_isdst = (unsigned char) *p++;
      if (ttisp->tt_isdst != 0 && ttisp->tt_isdst != 1)
         return -1;
      ttisp->tt_abbrind = (unsigned char) *p++;
      if (ttisp->tt_abbrind < 0 ||
          ttisp->tt_abbrind > sp->charcnt)
      {
#if 0
         return -1;
#else
         goto fail;
#endif
      }
   }

   for (i = 0; i < sp->charcnt; ++i)
      sp->chars[i] = *p++;
   sp->chars[i] = '\0';	/* ensure '\0' at end */

   for (i = 0; i < sp->leapcnt; ++i)
   {
      register struct lsinfo *	lsisp;

      lsisp = &sp->lsis[i];
      lsisp->ls_trans = detzcode(p);
      p += 4;
      lsisp->ls_corr = detzcode(p);
      p += 4;
   }

   for (i = 0; i < sp->typecnt; ++i)
   {
      register struct ttinfo *	ttisp;

      ttisp = &sp->ttis[i];
      if (ttisstdcnt == 0)
         ttisp->tt_ttisstd = FALSE;
      else
      {
         ttisp->tt_ttisstd = *p++;
         if (ttisp->tt_ttisstd != TRUE &&
             ttisp->tt_ttisstd != FALSE)
         {
#if 0
            return -1;
#else
            goto fail;
#endif
         }
      }
   }

   for (i = 0; i < sp->typecnt; ++i)
   {
      register struct ttinfo *	ttisp;

      ttisp = &sp->ttis[i];
      if (ttisgmtcnt == 0)
         ttisp->tt_ttisgmt = FALSE;
      else
      {
         ttisp->tt_ttisgmt = *p++;
         if (ttisp->tt_ttisgmt != TRUE &&
             ttisp->tt_ttisgmt != FALSE)
         {
#if 0
            return -1;
#else
            goto fail;
#endif
         }
      }
   }

   heap_free_private(up);
   return 0;

  fail:
   heap_free_private(up);
   return -1;
#undef u
}


static void
signalConfigRequest(union SIGNAL *sig, struct timeZoneInfo *tzInfo)
{
   if (tzInfo->sp == NULL)
   {
      tzInfo->sp = (struct state *) alloc(sizeof(struct state), 0);
   }

   if (tzload(sig->config.tzData, sig->config.size, tzInfo->sp) == 0)
   {
      sig->config.result = RTC_ENORMAL;
   }
   else
   {
      sig->config.result = RTC_EINVALID_TZ_DATA;
      free_buf((union SIGNAL **) &tzInfo->sp);
      tzInfo->gmtOffset = 0;
      strcpy(tzInfo->abbr, "UTC");
   }
#ifdef  RTC_DEBUG
   printTzInfo(tzInfo);
#endif

   sig->sigNo = RTC_CONFIG_REPLY;
   send(&sig, sender(&sig));
}

static void
signalSetLocalOffsetRequest(union SIGNAL *sig,
                            struct timeZoneInfo *tzInfo)
{
   if (tzInfo->sp != NULL)
   {
      free_buf((union SIGNAL **) &tzInfo->sp);
   }
   tzInfo->gmtOffset = sig->setLocalOffset.gmtOffset;
   strncpy(tzInfo->abbr, sig->setLocalOffset.abbr, TZ_ABBR_BUFF_SIZE - 1);
   tzInfo->abbr[TZ_ABBR_BUFF_SIZE - 1] = '\0';

#ifdef  RTC_DEBUG
   printTzInfo(tzInfo);
#endif

   sig->sigNo = RTC_SET_LOCAL_OFFSET_REPLY;
   send(&sig, sender(&sig));
}


/*
 *===========================================================================
 *                trapGetTzInfo
 *===========================================================================
 * Description: Get the local time zone offset. If Daylight savings time is 
 *              active at the given time, it will be included in the offset.
 *
 * Parameters:  time            I: The time
 *              gmtOffset       O: Time zone offset from GMT in seconds.
 *              isdst           O: Flag for Daylight Savings Time
 *              abbr            O: Time zone abbreviation
 *
 * Returns:     Dummy value for BIOS interface.
 *
 * Called by:   RTC server process and client processes.
 */

static long
trapGetTzInfo(time_t t,
	      S32 *gmtOffset,
	      S32 *isdst,
	      char *abbr,
	      struct timeZoneInfo *tzInfo)
{
   int i;
   struct ttinfo *ttisp;

   if (tzInfo->sp == NULL)
   {
      /*
      ** No detailed time zone data given. Use the simple offset.
      */
      *gmtOffset = tzInfo->gmtOffset;

      /*
      ** We dont know if daylight savings time is active
      */
      if (isdst != NULL)
      {
         *isdst = -1;
      }

      if (abbr != NULL)
      {
         strncpy(abbr, tzInfo->abbr, TZ_ABBR_BUFF_SIZE - 1);
         abbr[TZ_ABBR_BUFF_SIZE - 1] = '\0';
      }

      return 0;
   }

   /*
   ** Find the GMT offset ("local time type") for the given time in
   ** the current time zone.
   */

   if (tzInfo->sp->timecnt == 0 || t < tzInfo->sp->ats[0])
   {
      i = 0;

      /*
       * We are before the first transition time. Use the
       * first defined type without DST
       */
      while (tzInfo->sp->ttis[i].tt_isdst)
         if (++i >= tzInfo->sp->typecnt)
         {
            i = 0;
            break;
         }
   }
   else
   {
      /*
       * Loop through the the transition times table to find the index to the
       * time type.
       */
      for (i = 1; i < tzInfo->sp->timecnt; ++i)
      {
         if (t < tzInfo->sp->ats[i])
         {
            break;
         }
      }

      i = tzInfo->sp->types[i - 1];
   }

   ttisp = &tzInfo->sp->ttis[i];
   *gmtOffset = ttisp->tt_gmtoff;

   if (isdst != NULL)
   {
      *isdst = ttisp->tt_isdst;
   }

   if (abbr != NULL)
   {
      strncpy(abbr, &(tzInfo->sp->chars[ttisp->tt_abbrind]), TZ_ABBR_BUFF_SIZE - 1);
      abbr[TZ_ABBR_BUFF_SIZE - 1] = '\0';
   }

   return 0;
}

/*
 *===========================================================================
 *                signalGetTzInfo
 *===========================================================================
 *
 * Description:  See trapGetTzInfo
 *
 * Parameters:
 *
 * Returns:
 */

static void
signalGetTzInfo(union SIGNAL *sig, struct timeZoneInfo *tzInfo)
{
   trapGetTzInfo(sig->getTzInfo.time,
                 &(sig->getTzInfo.gmtOffset),
                 &(sig->getTzInfo.isdst),
                 sig->getTzInfo.abbr,
                 tzInfo);
   sig->sigNo = RTC_GET_TZ_INFO_REPLY;
   send(&sig, sender(&sig));
}


/*
 *===========================================================================
 *                trapGetLsInfo
 *===========================================================================
 *
 * Description: Get the leap second correction factor for a given time.
 *
 * Parameters:  t          I: The time
 *              corr       O: Time correction to be applied
 *              hit        O: Indicates whether this very second is a leap
 *                         second or not.
 *
 * Returns:     Dummy value for BIOS interface.
 *
 * Called by:   RTC server process and client processes.
 */


static long
trapGetLsInfo(time_t t,
	      S32 *corrp,
	      S32 *hitp,
	      struct timeZoneInfo *tzInfo)

{
   const struct lsinfo *lp;
   int i;
   S32 corr = 0;
   int hit = 0;

   /*
   ** Get correction out of leap second data
   */

   i = (tzInfo->sp == NULL) ? 0 : tzInfo->sp->leapcnt;

   while (--i >= 0)
   {
      lp = &tzInfo->sp->lsis[i];
      if (t >= lp->ls_trans)
      {
         if (t == lp->ls_trans)
         {
            hit = ((i == 0 && lp->ls_corr > 0) ||
                   lp->ls_corr > tzInfo->sp->lsis[i - 1].ls_corr);
            if (hit)
            {
               while (i > 0 &&
                      (tzInfo->sp->lsis[i].ls_trans ==
                       tzInfo->sp->lsis[i - 1].ls_trans + 1) &&
                      (tzInfo->sp->lsis[i].ls_corr ==
                       tzInfo->sp->lsis[i - 1].ls_corr + 1))
               {
                  ++hit;
                  --i;
               }
            }
         }
         corr = lp->ls_corr;
         break;
      }
   }
   *corrp = corr;
   *hitp = hit;

   return 0;
}

static void
signalGetLsInfo(union SIGNAL *sig, struct timeZoneInfo *tzInfo)
{
   trapGetLsInfo(sig->getLsInfo.time,
                 &(sig->getLsInfo.corr),
                 &(sig->getLsInfo.hit),
                 tzInfo);
   sig->sigNo = RTC_GET_LS_INFO_REPLY;
   send(&sig, sender(&sig));
}


/*
 *===========================================================================
 *                trapTime2subTypeCorr
 *===========================================================================
 *
 * Description:
 *
 * Parameters:
 *
 * Returns:
 */
static long
trapTime2subTypeCorr(time_t *t,
                     S32 offset,
                     S32 your_isdst,
                     struct timeZoneInfo *tzInfo)
{
   time_t newt;
   int i, j;
   struct tm yourtm;
   struct tm mytm;
   S32 gmtoff;
   S32 corr;
   S32 hit;

   /* Offset appears to not be used. It is not exposed in
    * external interfaces and is always set to 0 when calling
    * this function. It should be safe to ignore the parameter. */
   (void)offset;
   assert(offset == 0);

   trapGetTzInfo(*t, &gmtoff, NULL, NULL, tzInfo);
   trapGetLsInfo(*t, &corr, &hit, tzInfo);

   rtc_timesub(t, gmtoff, &yourtm, corr, hit);

   yourtm.tm_isdst = your_isdst;

   if (tzInfo->sp == NULL)
   {
      *t = (time_t)-1; /* WRONG */
      return 0;
   }

   for (i = tzInfo->sp->typecnt - 1; i >= 0; --i)
   {
      if (tzInfo->sp->ttis[i].tt_isdst != yourtm.tm_isdst)
         continue;
      for (j = tzInfo->sp->typecnt - 1; j >= 0; --j)
      {
         if (tzInfo->sp->ttis[j].tt_isdst == yourtm.tm_isdst)
            continue;
         newt = *t + tzInfo->sp->ttis[j].tt_gmtoff -
            tzInfo->sp->ttis[i].tt_gmtoff;

         trapGetTzInfo(newt, &gmtoff, NULL, NULL, tzInfo);
         trapGetLsInfo(*t, &corr, &hit, tzInfo);

         rtc_timesub(&newt, gmtoff, &mytm, corr, hit);

         if (rtc_tmcomp(&mytm, &yourtm) != 0)
            continue;
         if (mytm.tm_isdst != yourtm.tm_isdst)
            continue;
         /*
         ** We have a match.
         */
         *t = newt;
         return 0;
      }
   }

   *t = (time_t)-1; /* WRONG */
   return 0;
}

static void
signalT2subTypeCorr(union SIGNAL *sig, struct timeZoneInfo *tzInfo)
{
   trapTime2subTypeCorr(&sig->t2subTypeCorr.t,
                        sig->t2subTypeCorr.offset,
                        sig->t2subTypeCorr.your_isdst,
                        tzInfo);
   sig->sigNo = RTC_T2SUB_TYPE_CORR_REPLY;

  send(&sig, sender(&sig));
}



static long
trapGetDstDiff(time_t t,
               S32 same_isdst,
               S32 *dstDiff,
               struct timeZoneInfo *tzInfo)
{
   int i, j;

   if (tzInfo->sp == 0 || tzInfo->sp->timecnt == 0 || t < tzInfo->sp->ats[0])
   {
      /*
       * Time valuess before the first transition time are not supported.
       */
      *dstDiff = 0;
   }
   else
   {
      /*
       * Loop through the the transition times table to find the index to the
       * time type.
       */
      for (i = 1; i < tzInfo->sp->timecnt; ++i)
      {
         if (t < tzInfo->sp->ats[i]) {
            break;
         }
      }
      i--;

      /*
       * Find an earlier time type with other isdst value.
       */

      for (j = i; j >= 0 ; --j)
      {
         if (tzInfo->sp->ttis[tzInfo->sp->types[i]].tt_isdst !=
             tzInfo->sp->ttis[tzInfo->sp->types[j]].tt_isdst)
         {
            break;
         }
      }
      if (j < 0)
      {
         /*
          * No earlier was found, try later
          */
         for (j = i; j < tzInfo->sp->timecnt; ++j)
         {
            if (tzInfo->sp->ttis[tzInfo->sp->types[i]].tt_isdst !=
                tzInfo->sp->ttis[tzInfo->sp->types[j]].tt_isdst)
            {
               break;
            }
         }
      }
      if (j >= 0 && j < tzInfo->sp->timecnt)
      {
         if (tzInfo->sp->ttis[tzInfo->sp->types[i]].tt_isdst == same_isdst)
         {
            *dstDiff = tzInfo->sp->ttis[tzInfo->sp->types[j]].tt_gmtoff -
               tzInfo->sp->ttis[tzInfo->sp->types[i]].tt_gmtoff;
         }
         else
         {
            *dstDiff = tzInfo->sp->ttis[tzInfo->sp->types[i]].tt_gmtoff -
               tzInfo->sp->ttis[tzInfo->sp->types[j]].tt_gmtoff;
         }
      }
      else
      {
         *dstDiff = 0;
      }
   }
   return 0;
}


static void
signalGetDstDiff(union SIGNAL *sig, struct timeZoneInfo *tzInfo)
{
   trapGetDstDiff(sig->getDstDiff.t,
                  sig->getDstDiff.isdst,
                  &sig->getDstDiff.dstDiff,
                  tzInfo);
   sig->sigNo = RTC_GET_DST_DIFF_REPLY;
   send(&sig, sender(&sig));
}
