/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2013 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */
/*
 ****************************************************************************
 *                          PROGRAM MODULE
 *
 *     $Workfile:   shell.c  $
 *     Document no: @(#) 511/OSE52-31 $FileRevision:   /main/tb_current_ose5/13  $
 *     $Author:   thol  $
 *     $Date:   10/07/11 15:00:36  $
 *
 *
 ****************************************************************************
 */

#include <stddef.h>
#include <string.h>
#include <ctype.h>
#undef tolower
#undef toupper
#undef isalnum
#undef isalpha
#undef iscntrl
#undef isdigit
#undef isgraph
#undef islower
#undef isprint
#undef ispunct
#undef isspace
#undef isupper
#undef isxdigit
#include "ose.h"
#include "ose_err.h"
#include "efs.h"
#include "efs_err.h"
#include "stdlib.h"
#include "stdio.h"
#include "unistd.h"
#include "sys/stat.h"
#include "termios.h"
#include "errno.h"
#include "heapapi.h"
#include "assert.h"
#include "shell.h"
#include "list.h"

#include "shelld.sig"
#include "efs.sig"
#include "time.h"

#include "validate_user.h"
#include "shell_private.h"

union SIGNAL
{
    SIGSELECT sigNo;
};

#define BUFSIZE 2048            /* Size of shell input buffer. */

/* Status codes only used internally in struct PreInterpretedCommand.status */
#define SHELL_EINTERNAL_OPEN_REDIRECTION_FAILED    0x20
#define SHELL_EINTERNAL_PARSE_REDIRECTION_FAILED   0x21

#define STR(x) #x
#define STRSTR(y) STR(y)

struct ShellCommand
{
    ListLink link;
    char *name;
    char *usage;
    char *description;
    char s[3];			/* Variable size. */
};

static struct ShellCommand *
alloc_shell_command(const char *name, const char  *usage, const char *description) 
{
    U32 l1 = strlen(name);
    U32 l2 = strlen(usage);
    U32 l3 = strlen(description);
    struct ShellCommand *cmd = list_init_link(heap_alloc_private(sizeof (struct ShellCommand) + l1 + l2 + l3, NULL, 0));
    
    cmd->name = &cmd->s[0];
    cmd->usage = &cmd->s[l1 + 1];
    cmd->description = &cmd->s[l1 + 1 + l2 + 1];
    strcpy(cmd->name, name);
    strcpy(cmd->usage, usage);
    strcpy(cmd->description, description);
    
    return cmd;
}

static char *
skip_non_ws(char *str)
{
    assert(str != NULL);
    while (*str != '\0' && !isspace(*str))
    {
        str++;
    }
    return str;
}

static char *
skip_ws(char *str)
{
    assert(str != NULL);
    while (*str != '\0' && isspace(*str))
    {
        str++;
    }
    return str;
}

static char *
first_arg(char *cmd_line)
{
    char *ptr = skip_non_ws(skip_ws(cmd_line));
    *ptr = '\0';
    return cmd_line;
}

static unsigned
contains_space(const char *str)
{
    for (;;)
    {
        char c = *str++;
        if (c == '\0')
        {
            break;
        }
        if (isspace(c))
        {
            return 1;
        }
    }
    return 0;
}

#define MAX_ENV 16 /* Max # of ENV=VALUE preceeding command. */
struct Redirect
{
    const char *input;
    const char *output;
    const char *errput;
    Boolean output_append;
    Boolean errput_append;
    unsigned nenv;
    char *env[MAX_ENV + 1]; /* One extra element for NULL-termination. */
};

static char *
parse_command(char *cmd_line, struct Redirect *redir, Boolean *async)
{
    size_t size = 1; /* One byte for the NUL character. */
    int i, argc;
    char **argv;
    char *amp = strrchr(cmd_line, '&');

    if (amp != NULL && *skip_ws(amp+1) == '\0')
    {
       *amp = '\0';
       *async = True;
    }
    else
    {
       *async = False;
    }
    efs_split_args(cmd_line, &argc, &argv);
    for (i = 0; i < argc; i++)
    {
        const char *p = argv[i];
        if (*p == '<')
        {
            if (redir->input != NULL)
            {
                eprintf("Syntax error: Duplicate stdin redirection.\n");
                goto failure;
            }
            p++;
            if (*p != '\0')
            {
                redir->input = p;
                argv[i] = NULL;
            }
            else if (i + 1 < argc)
            {
                redir->input = argv[i + 1];
                argv[i] = NULL;
                argv[i + 1] = NULL;
                i++;
            }
            else
            {
                eprintf("Syntax error: '<' missing argument.\n");
                goto failure;
            }
        }
        else if (*p == '>')
        {
            if (redir->output != NULL)
            {
                eprintf("Syntax error: Duplicate stdout redirection.\n");
                goto failure;
            }
            p++;
            if (*p == '>')
            {
                redir->output_append = True;
                p++;
            }
            if (*p != '\0')
            {
                redir->output = p;
                argv[i] = NULL;
            }
            else if (i + 1 < argc)
            {
                redir->output = argv[i + 1];
                argv[i] = NULL;
                argv[i + 1] = NULL;
                i++;
            }
            else
            {
                eprintf("Syntax error: '%s' missing argument.\n", redir->output_append ? ">>" : ">");
                goto failure;
            }
        }
        else if (*p == '2' && *(p+1) == '>')
        {
            if (redir->errput != NULL)
            {
                eprintf("Syntax error: Duplicate stderr redirection.\n");
                goto failure;
            }
            p += 2;
            if (*p == '>')
            {
                redir->errput_append = True;
                p++;
            }
            if (*p != '\0')
            {
                redir->errput = p;
                argv[i] = NULL;
            }
            else if (i + 1 < argc)
            {
                redir->errput = argv[i + 1];
                argv[i] = NULL;
                argv[i + 1] = NULL;
                i++;
            }
            else
            {
                eprintf("Syntax error: '2%s' missing argument.\n", redir->errput_append ? ">>" : ">");
                goto failure;
            }
        }
        else if ((size <= 1) && (strchr(argv[i], '=') != NULL) && (strncmp(argv[i], "=", 1) != 0)
		 && ((strlen(strchr(argv[i], '=')) > 1) || (argc > i+1)))
	   /* argv contains a '=' but does not start with '=' */
	   /* and does not end with a '=' unless there are more arguments  */
        {
	   if (redir->nenv < MAX_ENV)
	   {
               /* Protect env variables USERNAME and TERMINAL from being changed. */
               if (strncmp(argv[i], "USERNAME=", 9) == 0)
               {
                  eprintf("Syntax error: not allowed to set USERNAME.\n");
                  goto failure;
               }
               else if (strncmp(argv[i], "TERMINAL=", 9) == 0)
               {
                  eprintf("Syntax error: not allowed to set TERMINAL.\n");
                  goto failure;
               }

               /* Save env variable. */
                redir->env[redir->nenv++] = argv[i];
                argv[i] = NULL;
            }
            else
            {
                eprintf("Too many assignments, ignoring %s\n", argv[i]);
                argv[i] = NULL;
            }
        }
        else /* Accumulate size of remaining items. */
        {
            size_t len = strlen(argv[i]);
            size += len + 1 + 2 * (len == 0 || contains_space(argv[i]));
        }
    }
    assert(argv[i] == NULL);
    assert(redir->nenv <= MAX_ENV);
    redir->env[redir->nenv++] = NULL;
    /* FIXME: perhaps we should skip this if size == 1 and just return NULL. */
    /* Allocate a new command line buffer and fill it with the remaining items. */
    cmd_line = (char *)heap_alloc_private(size, __FILE__, __LINE__);
    {
        char *ptr = cmd_line;
        for (i = 0; i < argc; i++)
        {
            if (argv[i] != NULL)
            {
                size_t len = strlen(argv[i]);
                Boolean quote = len == 0 || contains_space(argv[i]);
                if (quote)
                {
                    *ptr++ = '"';
                }
                memcpy(ptr, argv[i], len);
                ptr += len;
                if (quote)
                {
                    *ptr++ = '"';
                }
                *ptr++ = ' ';
            }
        }
        *ptr++ = '\0';
        assert(cmd_line + size == ptr);
        /* Remove trailing space. */
        if (size > 1)
        {
            cmd_line[size - 2] = '\0';
        }
    }
    heap_free_private(argv);
    return cmd_line;
failure:
    heap_free_private(argv);
    return NULL;
}

/*
 *  Function validate_and_log_cmd()
 *
 *  Executes the user provided validate_cmd function with correct parameters.
 *
 *  Input command->status together with internal error conditions
 *  determines whether the command is logged and validated or only logged.
 *
 *  Updates command->status in case of error.
 */
static void
validate_and_log_cmd(struct PreInterpretedCommand * command,
                     char *original_cmd_line,
                     struct Redirect *redir)
{
   Boolean    result = False;
   PROCESS    shell_pid;
   char       *username;
   char       *termname;
   char       *tmp_cmd_line;
   int        size;
   int        argc;
   char       **argv;
   U32        no_fileredir = 0;
   const char *fileredir[5] = {NULL, NULL, NULL, NULL, NULL};
   char       *status = NULL;

   shell_pid = current_process();
   username = get_env(shell_pid, "USERNAME");
   termname = get_env(shell_pid, "TERMINAL");

   /* If parsing failed. */
   if ((command->cmd_str == NULL) || (command->status != SHELL_SUCCESS))
   {
      /* Log command. */
      validate_cmd(original_cmd_line,
                   username,
                   termname,
                   0,
                   NULL,
                   0,
                   NULL,
                   0,
                   NULL,
                   shell_pid,
                   NULL);
   }
   else /* Validate and log command. */
   {
      /* Get argc and argv. */
      size = strlen(command->cmd_str) + 1;
      tmp_cmd_line = (char *)heap_alloc_private(size, __FILE__, __LINE__);
      strcpy(tmp_cmd_line, command->cmd_str);
      efs_split_args(tmp_cmd_line, &argc, &argv);

      /* Get no_fileredir and fileredir. */
      if (redir->input != NULL)
      {
         fileredir[0] = redir->input;
         no_fileredir = 5;
      }

      if (redir->output != NULL)
      {
         if (redir->output_append)
         {
            fileredir[2] = redir->output;
         }
         else
         {
            fileredir[1] = redir->output;
         }
         no_fileredir = 5;
      }

      if (redir->errput != NULL)
      {
         if (redir->errput_append)
         {
            fileredir[4] = redir->errput;
         }
         else
         {
            fileredir[3] = redir->errput;
         }
         no_fileredir = 5;
      }

      result = validate_cmd(original_cmd_line,
                            username,
                            termname,
                            argc,
                            (const char **) argv,
                            redir->nenv - 1, /* Don't include the null term element. */
                            (redir->nenv - 1) == 0 ? NULL : (const char **) &redir->env[0],
                            no_fileredir,
                            no_fileredir == 0 ? NULL : (const char **) fileredir,
                            shell_pid,
                            &status);

      /* Print status if validation is false. */
      if (result == False)
      {
         if (status != NULL)
         {
            eprintf("%s", status);
         }
         command->status = SHELL_EILLEGAL_CMD;
      }

      /* Free status heap buffer. */
      if (status != NULL)
      {
         heap_free_private(status);
      }

      heap_free_private(tmp_cmd_line);
      heap_free_private(argv);
   }

   if (username != NULL)
   {
      free_buf((union SIGNAL **)&username);
   }
   if (termname != NULL)
   {
      free_buf((union SIGNAL **)&termname);
   }
}

#define FD0 1
#define FD1 2
#define FD2 3
static void
init_redir(int fd_array[4])
{
    fd_array[0] = 3; /* Number of fd's in array. */
    fd_array[FD0] = 0;
    fd_array[FD1] = 1;
    fd_array[FD2] = 2;
}

static int
open_redir(struct Redirect *redir, int fd_array[4])
{
    if (redir->input != NULL)
    {
        fd_array[FD0] = open(redir->input, O_RDONLY, 0);
        if (fd_array[FD0] == -1)
        {
            int err = errno;
            eprintf("Could not open file '%s' for read: %s.\n", redir->input, strerror(errno));
            return err;
        }
    }
    if (redir->output != NULL)
    {
        fd_array[FD1] = open(redir->output, O_WRONLY | O_TEXT | O_CREAT | (redir->output_append ? O_APPEND : O_TRUNC), S_IREAD | S_IWRITE);
        if (fd_array[FD1] == -1)
        {
            int err = errno;
            eprintf("Could not open file '%s' for write: %s.\n", redir->output, strerror(errno));
            return err;
        }
    }
    if (redir->errput != NULL)
    {
        fd_array[FD2] = open(redir->errput, O_WRONLY | O_TEXT | O_CREAT | (redir->errput_append ? O_APPEND : O_TRUNC), S_IREAD | S_IWRITE);
        if (fd_array[FD2] == -1)
        {
            int err = errno;
            eprintf("Could not open file '%s' for write: %s.\n", redir->errput, strerror(errno));
            return err;
        }
    }
    return EOK;
}

static void
close_redir(int fd_array[4])
{
    if (fd_array[FD0] != 0)
    {
        (void) close(fd_array[FD0]);
    }
    if (fd_array[FD1] != 1)
    {
        (void) close(fd_array[FD1]);
    }
    if (fd_array[FD2] != 2)
    {
        (void) close(fd_array[FD2]);
    }
}
#undef FD0
#undef FD1
#undef FD2


/*
 * cmd_cd
 *
 * Built-in cd command
 *
 */
static int cmd_cd(int argc, char** argv)
{
   const char* pname = argv[0];
   
   if (argc != 2)
   {
      goto failure;
   }
   if (chdir(argv[1]) != 0)
   {
      perror("chdir");
      goto failure;
   }

   return 0;

  failure:
   
   eprintf("usage: %s [new directory]\n", pname);

   return -1;
}


/*
 * cmd_umask
 *
 * Built-in umask command.
 *
 */
static int cmd_umask(int argc, char** argv)
{
   const char* pname = argv[0];
   
   if (argc > 2)
   {
      goto failure;
   }

   if (argc == 2)
   {
      /* We must have at most 4 chars (0XXX) and at least 1 char.
         Everything must start with a 0 */
      unsigned len = strlen(argv[1]);
      char* ap = argv[1];
      mode_t newUmask = 0;
      if (len > 4 || len < 1 || *ap != '0')
      {
         eprintf("Bad mode.\n");
         goto failure;
      }
      
      /* Pass past leading 0 */
      ap ++;

      /* Set umask in a loop */
      while ((unsigned)(*ap - '0') <= (unsigned)('9' - '0'))
      {
         newUmask = (newUmask << 3) | (*ap - '0');
         ap ++;
      }
      if (*ap > '7')
      {
         eprintf("Bad mode.\n");
         goto failure;
      }
      
      umask(newUmask);
   }
   else
   {
      /* Report current umask */
      mode_t oldUmask = umask(0);
      umask(oldUmask);
      printf("0%c%c%c\n", 
             (oldUmask / 64) % 8 + '0', 
             (oldUmask / 8) % 8 + '0',
             oldUmask % 8 + '0');
   }

   return 0;

  failure:
   
   eprintf("usage: %s [octal mode, max 4 chars]\n", pname);

   return -1;
}


/*
 * cmd_exit
 *
 * Built-in exit command
 *
 */
static int
cmd_exit(int argc, char** argv, Boolean* quit)
{
   const char* pname = argv[0];

   if (argc > 2)
      goto failure;

   /* Checks okay, quit the shell */
   *quit = True;

   if (argc == 2)
      return atoi(argv[1]);

   return 0;

failure:

   eprintf("usage: %s [exit code, 0 if not specified]\n", pname);
   return -1;
}

static int
execute_command(char *cmd_line, PROCESS real_parent, Boolean *quit)
{
    int                rc = RET_ERROR;
    int                err = 0;
    struct Redirect    redir;
    char               *original_cmd_line;
    char               *parsed_cmd_line;
    char               *newline;
    int                size;
    struct             PreInterpretedCommand command = { 0, 0, 0, 0 };
    int                fd_array[4];
    Boolean            built_in_cmd = False;
    Boolean            async = False;
    char               *aux_cmd_line   = NULL;
    struct CompleteCmd cmd_info;
    clock_t            duration        = 0;
    clock_t            cmd_clock_stop  = 0;
    clock_t            cmd_clock_start = 0;
    char               *username = NULL;
    char               *termname = NULL;
    U32                no_fileredir  = 0;
    const char         *fileredir[5] = {NULL, NULL, NULL, NULL, NULL};
    PROCESS            shell_pid;
    char               **argv;
    int                argc;
   
    /* Check for command line consisting only of white space. */
    if (*skip_ws(cmd_line) == '\0')
    {
       return RET_SUCCESS; /* Nothing to be done */
    }

    /* Structure initialization */
    cmd_info.original_cmd_line = NULL;
    cmd_info.parsed_cmd_line   = NULL;
    cmd_info.argc              = 0;
    cmd_info.argv              = NULL;
    cmd_info.no_env            = 0;
    cmd_info.env               = NULL;
    cmd_info.no_fileredir      = 0;
    cmd_info.fileredir         = NULL;
    cmd_info.duration          = 0;
    cmd_info.cmd_exit_stat     = 0;

    /* Get the PID, username and terminal name */
    shell_pid  = current_process();
    username   = get_env(shell_pid, "USERNAME");
    termname   = get_env(shell_pid, "TERMINAL");
  
    /* Command execution started - get the arithmetic program time at this point. */
    cmd_clock_start = clock();

    init_redir(fd_array);
    redir.input         = NULL;
    redir.output        = NULL;
    redir.errput        = NULL;
    redir.output_append = False;
    redir.errput_append = False;
    redir.nenv          = 0;
   
    /* Save original command line before parsing (destroys cmd_line) and remove ending '\n'. */
    size = strlen(cmd_line) + 1;
    original_cmd_line = (char *)heap_alloc_private(size, __FILE__, __LINE__);
    strcpy(original_cmd_line, cmd_line);
    newline = strrchr(original_cmd_line, '\n');
    if (newline != NULL) *newline = '\0';
    
    /* Get the original command string for the logging hook.*/
    cmd_info.original_cmd_line = original_cmd_line;

    /* Parse command. Removes file redirection definitions and environment variable definitions
     * from command line and returns the new parsed command line in a new heap buffer.
     */
    parsed_cmd_line = parse_command(cmd_line, &redir, &async);

    /* Get the parsed command line for the logging hook */
    cmd_info.parsed_cmd_line = parsed_cmd_line;

    if (parsed_cmd_line == 0)
    {
       command.status = SHELL_EINTERNAL_PARSE_REDIRECTION_FAILED;
    }
    else if (*skip_ws(parsed_cmd_line) == '\0')
    {
       /* Check for parsed command line consisting only of white space. */
       goto cleanup; /* parsed_cmd_line is on the heap, we need to free it */
    }
    else
    {
        /* Check if command exists in shelld. */
        shell_interpreted_command_init(&command, parsed_cmd_line);
        assert(command.status == SHELL_SUCCESS); /* FIXME Else do what? */
        shell_interpreted_command_resolve(&command, False);
        
        /* If command not found check if built in command. */
        if (command.status != SHELL_SUCCESS)
        {
            assert(command.status == SHELL_EUNKNOWN_CMD);
            
            if ( (strcmp(command.name, "cd") == 0) ||
                 (strcmp(command.name, "exit") == 0) ||
                 (strcmp(command.name, "umask") == 0) )
            {
              /* Set command found.
               * NOTE: command elements status, cmd_str and name are valid.
               *       command element resolve_cmd is not valid.
               */
              command.status = SHELL_SUCCESS;
              built_in_cmd = True;
            }
            else /* Try resolving the default command "" instead. */
            {
               shell_interpreted_command_resolve(&command,
                                                 True); /* Default Command. */
            /* Check if we succeeded in resolving default command, Sona */
               if (command.status != SHELL_SUCCESS)
               {
                   eprintf("Unknown command '%s'.\n", first_arg(parsed_cmd_line));
                   goto cleanup;
               }
                /* Else default command present */
 
            }
        }
    }

    if (redir.input != NULL)
    {
       fileredir[0] = redir.input;
       no_fileredir = 5;
    }
    if (redir.output != NULL)
    {
       if (redir.output_append)
       {
          fileredir[2] = redir.output;
       }
       else
       {
          fileredir[1] = redir.output;
       }
       no_fileredir = 5;
    }
    if (redir.errput != NULL)
    {
       if (redir.errput_append)
       {
          fileredir[4] = redir.errput;
       }
       else
       {
          fileredir[3] = redir.errput;
       }
       no_fileredir = 5;
    }
    
    /* Fill the no_env and env fields */
    cmd_info.no_env = redir.nenv - 1; /* Don't include the null term element.*/
    
    /* If the number of env definitions is a non-zero, store the environment
     * variable definition strings
     */
    if (redir.nenv - 1 == 0)
    {
       cmd_info.env = NULL;
    }
    cmd_info.env = (const char **) &redir.env[0];    
    
    /* Validate and log command.
     * NOTE: Does not and should not use command element resolve_cmd.
     */
    validate_and_log_cmd(&command, original_cmd_line, &redir);
    
    /* Get the original command line and save it for the end hook*/
    size = strlen(original_cmd_line) + 1;
    aux_cmd_line = (char *)heap_alloc_private(size, __FILE__, __LINE__);
    strcpy(aux_cmd_line, original_cmd_line);

    /* Get argc and argv.*/
    efs_split_args(aux_cmd_line, &argc, &argv);
    cmd_info.argc = argc;
    cmd_info.argv = (const char **)argv;
    
    /* Open redirections. */
    if (command.status == SHELL_SUCCESS)
    {
        err = open_redir(&redir, fd_array);
        if (err != EOK)
        {
            command.status = SHELL_EINTERNAL_OPEN_REDIRECTION_FAILED;
        }
    }

    /* Built in commands. */
    if (built_in_cmd &&
        (command.status == SHELL_SUCCESS) )
    {
       /* FIXME: real_parent ignored for built commands. */

       /* Get args */
       int retValue = 0;

       /* efs_split_args allocates memory for argv */
       efs_split_args(parsed_cmd_line, &argc, &argv);
       
       if (argc == 0)
       {
          /* Hey! Why? :-( */
          assert(0);
          goto cleanup;
       }

       /* Go for the built-in command */
       if (strcmp(argv[0], "cd") == 0)
       {
          retValue = cmd_cd(argc, argv);
       }
       else if (strcmp(argv[0], "umask") == 0)
       {
          retValue = cmd_umask(argc, argv);
       }
       else if (strcmp(argv[0], "exit") == 0)
       {
          retValue = cmd_exit(argc, argv, quit);
       }
       else
       {
          assert(0);
       }

       /* Release memory taken */
       heap_free_private(argv);

       rc = RET_SUCCESS;
       if (retValue != 0)
       {
          rc = retValue;
       }
       goto cleanup;
    }

    if (!built_in_cmd &&
        (command.status == SHELL_SUCCESS) )
    {
        OSADDRESS taglist[15];
        unsigned i = 0;
        taglist[i++] = RUNCMD_CLONE_ENV;
        taglist[i++] = RUNCMD_FDARRAY;
        taglist[i++] = (OSADDRESS) fd_array;
	if (async)
	{
	   taglist[i++] = RUNCMD_ASYNCHRONOUS;
	   rc = RET_SUCCESS;
	   err = EOK;
	}
	else
	{
	   taglist[i++] = RUNCMD_RETURN_STATUS;
	   taglist[i++] = (OSADDRESS) &rc;
	   taglist[i++] = RUNCMD_RETURN_ERRNO;
	   taglist[i++] = (OSADDRESS) &err;
	   taglist[i++] = RUNCMD_INTERRUPTABLE;
	}
        taglist[i++] = RUNCMD_ENVLIST;
        taglist[i++] = (OSADDRESS) redir.env;
        taglist[i++] = RUNCMD_TAGEND;
        assert(i <= 15);
        shell_run_cmd_2(&command, taglist);
        switch (command.status)
        {
        case SHELL_SUCCESS :
            {
#if 0
                if (rc > 0)
                {
                    eprintf("'%s' returned %d, errno %u.\n", first_arg(parsed_cmd_line), rc, err);
                }
#else
                char buf[12];
                sprintf(buf, "%ld", rc);
                set_env(current_process(), "?", buf);
#endif
                break;
            }
        case SHELL_ETERMINATED :
            {
                eprintf("'%s' terminated.\n", first_arg(parsed_cmd_line));
                err = ECHILD;
                break;
            }
        case SHELL_EUNKNOWN_CMD :
            {
                eprintf("Unknown command '%s'.\n", first_arg(parsed_cmd_line));
                err = ENOENT;
                break;
            }
        case SHELL_EMISSING_CMD :
            {
              eprintf("Invalid command '%s'.\n", parsed_cmd_line);
              err = EINVAL;
              break;
            }
        case SHELL_ECLONEFD_FAILED :
            {
                eprintf("Failed to clone file descriptors.\n");
                err = EBADF;
                break;
            }
        case SHELL_EILLEGAL_CMD:
            {
              /* Error printout handled where the error was detected. */
              err = EACCES;
              break;
            }
        case SHELL_EINTERNAL_PARSE_REDIRECTION_FAILED:
          {
            /* Error printout handled where the error was detected. */
            err = EINVAL;
            break;
          }
        case SHELL_EINTERNAL_OPEN_REDIRECTION_FAILED:
          {
            /* Error printout handled where the error was detected. */
            /* Error code set up where error was detected. */
            break;
          }

          /* The following should not appear, they are evidence of program errors. */
        case SHELL_EUNKNOWN_TAG:
        case SHELL_EMISSING_ARG:
        case SHELL_EINVALID_FD:
        default :
          {
              assert(0);
              break;
          }
        }
    }

  cleanup:
    /* FIXME: shouldn't we close these files before the child has started? */
    close_redir(fd_array);
    
    /* Finished the command execution - get the arithmetic program clock at this point.*/
    cmd_clock_stop = clock();
    duration = (cmd_clock_stop - cmd_clock_start);
    
    /* Fill up the shell hook completion structure fields with the command details. */
    cmd_info.username          = username;
    cmd_info.termname          = termname;
    cmd_info.no_fileredir      = no_fileredir;
    cmd_info.fileredir         = (const char **) fileredir;
    cmd_info.pid               = shell_pid;
    cmd_info.duration          = duration;
    
    /* Get the command status */
    cmd_info.cmd_exit_stat = rc;
    
    /* These values are useless if the command fails. */    
    if (command.status != SHELL_SUCCESS)
    {
       /* Release allocated memory in case of command failure */
       argv = NULL;
       heap_free_private(cmd_info.argv); 
       
       cmd_info.argc            = 0;
       cmd_info.argv            = NULL;
       cmd_info.no_fileredir    = 0;
       cmd_info.fileredir       = NULL;
       cmd_info.no_env          = 0;
       cmd_info.env             = NULL;
    }

    /* The end-hook is optional - if it is linked, it will call its definition,
     * else, nothing happens*/
    if (shell_hook_log_cmd_completion)
    {
       /* Call up the shell command completion hook */
       shell_hook_log_cmd_completion(&cmd_info);
    }
  
    /* Release allocated memory in case of command success */
    if (cmd_info.argv != NULL)
    {
       argv = NULL;
       heap_free_private(cmd_info.argv);
       cmd_info.argv = NULL;
    }
    
    if (parsed_cmd_line != NULL)
    {
       heap_free_private(parsed_cmd_line);
    }
    heap_free_private(original_cmd_line);
    heap_free_private(aux_cmd_line);
    shell_interpreted_command_delete(&command);

    if (username != NULL)
    {
       free_buf((union SIGNAL **)&username);
    }
    if (termname != NULL)
    {
       free_buf((union SIGNAL **)&termname);
    }

    return rc;
}

int
system(const char *command)
{
   char *newcmd;
   Boolean quit = False;
   int rc;

   newcmd = heap_alloc_private(strlen(command) + 7, NULL, 0);
   /* Execute command in a subshell */
   sprintf(newcmd, "shell %s", command);
   rc = execute_command(newcmd, 0, &quit);
   heap_free_private(newcmd);
   return rc;
}

static int
cmd_forever(int argc, char *argv[])
{
   int rc = SHELL_SUCCESS;
   char *cmdline;

   cmdline = get_env(current_process(), "ARGV");
   if (cmdline != NULL)
   {
      char *args;

      /* Find the command to execute */
      args = skip_ws(skip_non_ws(cmdline));
      if (strlen(args) != 0)
      {
	 for (;;)
	 {

	    rc = system(args);
	    if (rc != 0)
	    {
	       break;
	    }
	 }
      }
      free_buf((union SIGNAL **)&cmdline);
   }
   return rc;
}

static int
open_terminal(void)
{
    int fd = 0; /* Default to stdin. */
    char *terminal = get_env(current_process(), "TERMINAL");
    if (terminal != NULL)
    {
        fd = open(terminal, O_RDONLY, 0);
        if (fd == -1)
        {
            fd = 0; /* Revert to stdin. */
        }
        free_buf((union SIGNAL **)&terminal);
    }
    /* Subscribe to asynchronous signals (if supported). */
    if (efs_fd_supports_sig(fd, FM_SUBSCRIBE_SIGNAL_REQUEST) == EFS_SUCCESS)
    {
        shell_subscribe_signal(fd, True);
    }

    /* clear history */
    /* For the case when a user is logged in and another user logs in,
     * without exiting the first user */
    efs_config_term(fd, FM_TERM_CLEARHIST, 1, 0);

    return fd;
}

static void
close_terminal(int fd)
{
    if (fd != 0) /* Don't close stdin. */
    {
       /* clear history */
       efs_config_term(fd, FM_TERM_CLEARHIST, 1, 0);
       (void) close(fd);
    }
    else /* stdin used as terminal, unsubscribe from signals. */
    {
        if (efs_fd_supports_sig(fd, FM_SUBSCRIBE_SIGNAL_REQUEST) == EFS_SUCCESS)
        {
            shell_subscribe_signal(fd, False);
        }
    }
}

/* Check if the value for CONSOLE_IDLE_TIMEOUT environment variable was set or
 * its value was changed and reconfigure the console with this value. */
static void
set_console_tmo(int fd, long *old_tmo_val)
{
    long new_tmo_val;
    EfsStatus res = EFS_SUCCESS;
    char *tmo_str = get_env(current_process(), "CONSOLE_IDLE_TIMEOUT");
    if(tmo_str != NULL)
    {
        new_tmo_val = efs_str_to_long(tmo_str, NULL);
        if(new_tmo_val != *old_tmo_val)
        {
            res = efs_config_term(fd, FM_TERM_IDLETMO, new_tmo_val, 0);
            if(res == EFS_SUCCESS)
                *old_tmo_val = new_tmo_val;
        }
        free_buf((union SIGNAL **)&tmo_str);
    }
    else
    {
        /* The shell just started and no CONSOLE_IDLE_TIMEOUT environment 
         * variable set. It could be the situation when the console was 
         * configured with an idle timeout value by the previous shell 
         * session. In such situation the console should be reconfigured with
         * this value.*/
         efs_config_term(fd, FM_TERM_IDLETMO, 0, 0);
    }
}

static int
execute_script(FILE *file, Boolean interactive, PROCESS real_parent)
{
    int rc = 20;
    Boolean quit = False;
    EfsBitmask flags = FM_TIFL_LINE_EDIT | FM_TIFL_SAVE_LINE |
                       FM_TIFL_FULL_BELL | FM_TIFL_ECHO | FM_TIFL_ECHO_CTRL |
                       FM_TIFL_FLOW_CTRL | FM_TIFL_ASYNC_SIGS |
                       FM_TIFL_MAP_NL_CRNL;
    char *cmd_line = heap_alloc_private(BUFSIZE, __FILE__, __LINE__);
    int term_fd = open_terminal();
    long current_idle_tmo_val = 0;

    if (interactive)
    {
        EfsStatus status = efs_examine_term(term_fd, FM_TERM_FLAGS, &flags, 0);
        if (status != EFS_SUCCESS)
        {
            eprintf("Could not examine terminal (%lu).\n", status);
        }
        else
            set_console_tmo(term_fd, &current_idle_tmo_val); 
    }

    /* Loop reading and executing commands till someone tells us to quit. */
    while (!quit)
    {
        if (interactive)
        {
            char *prompt = get_env(current_process(), "PROMPT");
            fputs(prompt != NULL ? prompt : "$ ", stdout);
            fflush(stdout);
            if (prompt != NULL)
            {
                free_buf((union SIGNAL**)&prompt);
            }
        }

        if (fgets(cmd_line, BUFSIZE, file) != NULL)
        {
            /* Remove pending (interrupt) signals. */
            static const SIGSELECT sigsel[] = { 1, FM_SIGNAL };
            union SIGNAL * sig;
            while ((sig = receive_w_tmo(0, sigsel)) != NIL)
            {
                free_buf(&sig);
            }
            rc = execute_command(cmd_line, real_parent, &quit);

            /* If the executed command changed the value for the
             * CONSOLE_IDLE_TIMEOUT variable then the corresponding parameter 
             * of the console has to be updated. */
            if(rc == RET_SUCCESS)
                set_console_tmo(term_fd, &current_idle_tmo_val); 
            
            if (interactive)
            {
                /* Restore terminal flags. */
                efs_config_term(term_fd, FM_TERM_FLAGS, flags, 0);
            }
        }
        else /* Reached end of file. */
        {
            /* The fd used for reading is closed either because the idle time
             * expired or the console received an error signal (this is
             * similar to a timeout situation from shell's point of view). 
             * Print the timeout message (if configured) and exit.*/
            char *tmo_msg = get_env(current_process(),
                "CONSOLE_IDLE_TIMEOUT_MSG");
            if(tmo_msg != NULL)
            {
                fputs(tmo_msg, stdout);
                fflush(stdout);
                free_buf((union SIGNAL**)&tmo_msg);
            }
            break;
        }
    }

    close_terminal(term_fd);
    heap_free_private(cmd_line);
    return rc;
}

static int
cmd_shell(int argc, char *argv[])
{
    int rc = 20;
    if (argc == 1)
    {   /* "shell" => interactive session. */
        PROCESS shell_pid = current_process();
        char *msg = get_env(shell_pid, "WELCOME_MSG");
        if (msg == NULL)
        {
            msg = get_env(get_bid(shell_pid), "WELCOME_MSG");
        }
        if (msg != NULL)
        {
            fputs(msg, stdout);
            free_buf((union SIGNAL**)&msg);
        }
        else
        {
            fputs("Welcome to OSE Shell "STRSTR(VERSION)".\n", stdout);
        }
        rc = execute_script(stdin, True, 0);
    }
    else if (argc >= 2 && strcmp(argv[1], "-f") == 0)
    {   /* "shell -f ..." => read commands from file. */
        if (argc == 3)
        {
            FILE *file = fopen(argv[2], "r");
            if (file != NULL)
            {
                rc = execute_script(file, False, 0);
                fclose(file);
            }
            else
            {
                eprintf("Could not open \"%s\": %s\n", argv[2],
                        strerror(errno));
            }
        }
        else
        {
            eprintf("Usage: %s -f <filename>\n", argv[0]);
        }
    }
    else
    {   /* "shell ..." => execute command in subshell. */
        char *cmd_line = get_env(current_process(), "ARGV");
        if (cmd_line != NULL)
        {
            char *args = skip_ws(skip_non_ws(cmd_line));
            if (*args != '\0')
            {
                Boolean quit = False;
                rc = execute_command(args, 0, &quit);
            }
            /* Else no parameters in ARGV, inconsistent with argc/argv. */
            free_buf((union SIGNAL**) &cmd_line);
        }
        /* Else ARGV not set, inconsistent with argc/argv. */
    }
    return rc;
}

/* ose_shell exists only because of backwards compatibility. */
OSENTRYPOINT ose_shell;
OS_PROCESS(ose_shell)
{
    int rc;

    char *cmd_line = get_env(current_process(), "ARGV");
    if (cmd_line != NULL)
    {
        int argc;
        char **argv;
        efs_split_args(cmd_line, &argc, &argv);
        rc = cmd_shell(argc, argv);
        heap_free_private(argv);
        free_buf((union SIGNAL **)&cmd_line);
    }
    else
    {
        static char tmp[] = "shell";
        static char *argv[] = { tmp, NULL };
        rc = cmd_shell(1, argv);
    }
    exit(rc);
}

static int
cmd_dot(int argc, char *argv[]) /* FIXME: does the '.' command still work? */
{
    int rc = 20;
    if (argc == 2)
    {
        FILE *script = fopen(argv[1], "r");
        if (script != NULL)
        {
            char *buf = get_env(current_process(), "PARENT");
            if (buf != NULL)
            {
                PROCESS parent = efs_str_to_long(buf, NULL);
                free_buf((union SIGNAL **)&buf);
                rc = execute_script(script, False, parent);
            }
            else /* Environment variable PARENT not set, very strange. */
            {
                eprintf("Could not determine parent shell, aborting.\n");
            }
            fclose(script);
        }
        else
        {
            eprintf("Could not open \"%s\": %s\n", argv[1], strerror(errno));
        }
    }
    else
    {
        eprintf("Usage: %s <command-file>\n", argv[0]);
    }
    return rc;
}

#define HUNT_TMO 5000

static PROCESS
hunt_shelld(void)
{
    static const SIGSELECT sigsel[] = { 1, HUNT_SHELLD };
    union SIGNAL *sig = alloc(sizeof(SIGSELECT), HUNT_SHELLD);
    hunt("ose_shelld", 0, NULL, &sig);
    sig = receive_w_tmo(HUNT_TMO, sigsel);
    if (sig != NIL)
    {
        PROCESS pid = sender(&sig);
        free_buf(&sig);
        return pid;
    }
    else
    {
        error2(OSE_EFATAL_MASK | OSE_EFS_SHELL_EHUNT_FAILED,
               (OSERRCODE)"ose_shelld");
        kill_proc(current_process());
        return 0; /* Not reached. */
    }
}

static U32
efs_flow_control(FILE *file)
{
    U32 lines = 0;
    char *env = get_env(current_process(), "FLOW_CONTROL");
    if (env != NULL)
    {
        lines = efs_str_to_long(env, NULL);
        if (lines != 0)
        {
            if (isatty(fileno(file)) != 1)
            {
                lines = 0;
            }
            /* Else file is a terminal and lines is ok. */
        }
        free_buf((union SIGNAL **)&env);
    }
    return lines;
}

static void
put_html_char(char c)
{
    switch (c)
    {
        case '<':
            printf("&lt;");
            break;
        case '>':
            printf("&gt;");
            break;
        case '&':
            printf("&amp;");
            break;
        default:
           putchar(c);
   }
}

static void
put_html_str(char *s)
{
    while (*s)
    {
        put_html_char(*s++);
    }
}
static void print_cmd(struct ShellCommand *cmd, char *username, char *termname, 
                      char *rooturl, char *argv[], int argc, Boolean *match, 
                      Boolean brief) {
           
        char *longdesc = NULL;

        longdesc = strchr(cmd->description, '\n');
        if (longdesc != NULL)
        {
            *longdesc++ = '\0';
        }
        if (!validate_help(cmd->name, username, termname))
        {
          /* Not allowed to print help. */
        }
        else if (brief)
        {
                       
            
            if (argc == 1 ||
                strncmp(argv[1], cmd->name, strlen(argv[1]) - 1) == 0)
            {
                if (!rooturl ||
                    strcmp(cmd->name, "reset") == 0 ||
                    strcmp(cmd->name, "shutdown") == 0)
                {
                    printf("%-15s %s\n", cmd->name, cmd->description);
                }
                else
                {
                    printf("<A HREF=\"%s/run?cmd=%s\">%s</A>%*s %s\r\n",
                           rooturl,
                           cmd->name,
                           cmd->name,
                           15 - strlen(cmd->name), "",
                           cmd->description);
                }
                *match = True;
            }
        }
        else /* !brief => print name, synopsis & optionally descrciption
                for specified command. */
        {
            assert(argc > 1);
            assert(argv[1] != NULL);
            if (strcmp(argv[1], cmd->name) == 0)
            {
                printf("NAME\n    %s - ", cmd->name);
                if(rooturl)
                   put_html_str(cmd->description);
                else
                   printf(cmd->description);
                printf("\n\nSYNOPSIS\n    ");
                if(rooturl)
                   put_html_str(cmd->usage);
                else
                   printf(cmd->usage);
                printf("\n\n");
                if (longdesc != NULL)
                {
                    /* I love nroff so I wrote my own. */
                    unsigned col = 0;
                    char last_char = ' ';
                    
                    printf("DESCRIPTION\n");
                    while (*longdesc != '\0')
                    {
                        if (*longdesc == '\n')
                        {
                            /*
                            ** A single '\n' after a '.' is
                            ** interpreted as the beginning of a new
                            ** paragraph.
                            */
                            if (last_char == '.' && longdesc[1] != '\n')
                            {
                                printf("\n\n");
                            }
                            else
                            {
                                printf("\n");
                            }
                            last_char = ' '; /* Reset last_char */
                            col = 0;
                        }
                        else if (*longdesc == ' ' && col >= 70)
                        {
                            printf("\n");
                            col = 0;
                        }
                        else
                        {
                            if (col == 80)
                            {
                                printf("\n");
                                col = 0;
                            }
                            if (col == 0)
                            {
                                printf("    ");
                                col += 4;
                            }
                            last_char = *longdesc;
                            if (rooturl)
                                put_html_char(last_char);
                            else
                                putchar(last_char);
                            col++;
                        }
                        longdesc++;
                    }
                    printf("\n");
                }
                *match = True;
            }
        }

}

static int
cmd_help(int argc, char *argv[])
{
    int rc = RET_SUCCESS;
    static const char format[] = "%-15s %s\n";
    static const char html_format[] = "<B>%-15s %s</B>\n";
    struct ShellListCmd *sig;
    PROCESS shelld_ = hunt_shelld();
    U32 lines = efs_flow_control(stdout);
    U32 line = 0;
    Boolean brief = False;
    char *rooturl;
    PROCESS    help_pid;
    char       *username;
    char       *termname;
    List       internal_cmds;
    struct ShellCommand *intcmd = NULL, *cmd = NULL;
   
   
    if (argc > 2)
    {
        printf("Usage: help [command[*]]\n");
        return RET_ERROR;
    }
    
    /* Manually create the list of internal commands.
     * This is quite unusual but we need to include them in the help.
     */

    list_init(&internal_cmds);
    intcmd = alloc_shell_command("cd" , "cd [new directory]", "change the working directory\nChanges working directory of the shell process. Built in OSE shell.\n");
    list_insert_tail(&internal_cmds, intcmd);
    
    intcmd = alloc_shell_command("exit","exit [exit code, 0 if not specified]","exit the current shell\nExit from current shell session (built in shell command). Terminates current shell session. Default exitcode is 0. \nThis is primarily useful for remote logins.\n");
    list_insert_tail(&internal_cmds, intcmd);
    
    intcmd = alloc_shell_command("umask","umask [octal mode, max 4 chars]","display/set the current file system mask\nDisplay current file system umask value or set a new umask value in octal mode, e.g. 0644 (built in shell command).\n This affects the access permissions on new files or directories, but its effect depends on the volume manager.\n");
    list_insert_tail(&internal_cmds, intcmd);
    
    help_pid = current_process();
    username = get_env(help_pid, "USERNAME");
    termname = get_env(help_pid, "TERMINAL");

    rooturl = get_env(current_process(), "ROOTURL");
    if (rooturl)
    {
       lines = 0;
    }
    
    /* '*' should be last character but we don't explicitly check this. */
    if (argc == 1 || (argc == 2 && strchr(argv[1], '*') != NULL))
    {
        brief = True;
        printf(rooturl ? html_format : format, "Command", "Description");
    }

    sig = (struct ShellListCmd *)alloc(sizeof(struct ShellListCmd),
                                       SHELL_LIST_CMD_REQUEST);
    send((union SIGNAL **)&sig, shelld_);
    for (;;)
    {
        Boolean match = False;
        static const SIGSELECT sigsel[] = { 1, SHELL_LIST_CMD_REPLY };
        sig = (struct ShellListCmd *)receive(sigsel);
        if (!sig->valid)
        {
            free_buf((union SIGNAL **)&sig);
            break;
        }
        cmd = alloc_shell_command(&sig->s[sig->name], &sig->s[sig->usage], 
                                  &sig->s[sig->description]);
        free_buf((union SIGNAL **)&sig);
          
        intcmd      = list_head((&internal_cmds));        
            
        if (intcmd != NULL && strcmp(cmd->name, intcmd->name) == 0) 
        {
               /* A similar command exists in the list. 
                * Remove the current internal command from the list and
                * do not print it, it will print instead 
                * the overwritten command help at the next print_cmd call.
                */
               heap_free_private(list_unlink(intcmd));
               intcmd  = NULL;
        }
            
        if (intcmd != NULL && strcmp(cmd->name, intcmd->name) > 0) 
        {
               /* Print the help for the current internal 
                * command and remove it from the list. 
                */
               print_cmd(intcmd, username, termname, rooturl, argv, 
                         argc, &match, brief);
               heap_free_private(list_unlink(intcmd));
        }
        
        print_cmd(cmd, username, termname, rooturl, argv, 
                  argc, &match, brief);
        heap_free_private(cmd);
        
        if (match && lines != 0 && ++line % lines == 0)
        {
            printf("Press XON to resume"); fflush(stdout);
            tcflow(fileno(stdout), TCOOFF);
            printf("\r                   \r"); fflush(stdout);
        }


   }
    
   cmd = list_head((&internal_cmds));

   while (cmd != NULL)
   {
      struct ShellCommand *next = list_next(&internal_cmds, cmd);
      heap_free_private(list_unlink(intcmd));
      cmd = next;
   }

   if (rooturl != NULL)
    {
        free_buf((union SIGNAL **)&rooturl);
    }

    if (username != NULL)
    {
        free_buf((union SIGNAL **)&username);
    }
    if (termname != NULL)
    {
        free_buf((union SIGNAL **)&termname);
    }

    return rc;

}


#if defined softOSE
static int
cmd_sfkexit(int argc, char *argv[])
{
    extern void (exit)(int);
    (exit)(0);
    return 0; /* NOTREACHED */
}
#endif

void
add_shell_cmds(void)
{
#ifdef LIBWHAT
    LIBWHAT
#endif
    shell_add_cmd("help"   , "help [command[*]]", "print description of commands\n'help' prints a brief description of matching commands.\nWhen a command is specified, 'help' also prints the synopsis and a long description.", cmd_help);
    shell_add_cmd("?"   , "? [command[*]]", "print description of commands\n'?' prints a brief description of matching commands.\nWhen a command is specified, '?' also prints the synopsis and a long description.", cmd_help);
    shell_add_cmd("."      , ". <command-file>", "execute commands from file", cmd_dot);
    shell_add_cmd("forever", "forever <command>", "execute command forever", cmd_forever);
    shell_add_cmd("shell"  , "shell [<cmd> ...] / [-f <file>]", "invoke shell\n'shell' invokes a new shell process [to execute command].\nThe '-f' option specifies a file from which to read and execute shell commands.", cmd_shell);

#if defined softOSE
    shell_add_cmd("sfkexit", "sfkexit", "exit soft kernel environment", cmd_sfkexit);
#endif
}
