/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 1996-2008 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */
/*
 ****************************************************************************
 *                            PROGRAM MODULE
 *
 *     $Workfile:   shelld.c  $
 *     Document no: @(#) 511/OSE52-39 $FileRevision:   /main/tb_current_ose5/9  $
 *     $Author:   thol  $
 *     $Date:   04/21/06 15:00:35  $
 *
 *
 ****************************************************************************
 */

#include <stddef.h>
#include <string.h>
#include "ose.h"
#include "osetypes.h"
#include "list.h"
#include "assert.h"
#include "heapapi.h"
#include "outfmt.h"
#include "shell.h"
#include "shelld.sig"
#include "child.sig"
#include "validate_user.h"
#include "efs.h"

#define STR(literal) #literal
#define STRSTR(macro) STR(macro)

union SIGNAL
{
    SIGSELECT sigNo;
};

struct Proxy
{
    ListLink link;
    S32 refcount;		/* Reference counter for related commands. */
    OSATTREF attref;		/* Attach reference or ~0 if not used. */
    PROCESS pid;		/* PID of command proxy. */
};

static struct Proxy *
alloc_proxy(PROCESS pid)
{
    struct Proxy *proxy =
	list_init_link(heap_alloc_private(sizeof (struct Proxy), NULL, 0));

    proxy->refcount = 0;
    proxy->attref = attach(NULL, pid);
    proxy->pid = pid;
    return proxy;
}

static void
free_proxy(struct Proxy *proxy)
{
    assert(proxy->refcount == 0);
    if (proxy->attref != (OSATTREF) ~ 0)
    {
	detach(&proxy->attref);
    }
    heap_free_private(list_unlink(proxy));
}

static struct Proxy *
find_proxy(const List *list, PROCESS pid)
{
    struct Proxy *proxy = list_head(list);

    while (proxy != NULL)
    {
	if (proxy->pid == pid)
	{
	    return proxy;
	}
	proxy = list_next(list, proxy);
    }
    return NULL;
}

static struct Proxy *
insert_proxy(List *list, PROCESS pid)
{
    struct Proxy *proxy = find_proxy(list, pid);

    if (proxy == NULL)
    {
	proxy = alloc_proxy(pid);
	list_insert_tail(list, proxy);
    }
    proxy->refcount++;
    return proxy;
}

struct Command
{
    ListLink link;
    char *name;
    char *usage;
    char *description;
    CmdType entrypoint;
    U32 proc_type;
    U32 priority;
    U32 stack_size;
    struct Proxy *proxy;
    char s[3];			/* Variable size. */
};

static void *
compare_before(struct Command *node, const char *name)
{
    return strcmp(node->name, name) >= 0 ? node : NULL;
}

static struct Command *
find_position(List *cmds, const char *name)
{
    return (struct Command *) list_traverse(cmds,
					    (void *(*)(ListLink *, void *))
					    compare_before,
                                            (char*) name);
}

static struct Command *
alloc_command(List *proxies, PROCESS proxy, const char *name, const char
	      *usage, const char *description, CmdType entrypoint, U32
	      proc_type, U32 priority, U32 stack_size)
{
    U32 l1 = strlen(name);
    U32 l2 = strlen(usage);
    U32 l3 = strlen(description);
    struct Command *cmd = list_init_link(heap_alloc_private(sizeof (struct Command) + l1 + l2 + l3, NULL, 0));

    cmd->name = &cmd->s[0];
    cmd->usage = &cmd->s[l1 + 1];
    cmd->description = &cmd->s[l1 + 1 + l2 + 1];
    cmd->entrypoint = entrypoint;
    cmd->proc_type = proc_type;
    cmd->priority = priority;
    cmd->stack_size = stack_size;
    cmd->proxy = insert_proxy(proxies, proxy);
    strcpy(cmd->name, name);
    strcpy(cmd->usage, usage);
    strcpy(cmd->description, description);
    return cmd;
}

static void
add_command(List *proxies, List *cmds, struct ShellAddCmd *req)
{
    const char *name = &req->s[req->name];
    const char *usage = &req->s[req->usage];
    const char *description = &req->s[req->description];
    U32 proc_type = req->proc_type;
    U32 priority = req->priority;
    U32 stack_size = req->stack_size;
    struct Command *new = alloc_command(proxies, req->proxy, name, usage,
					description, (CmdType)
					req->entrypoint, proc_type, priority, stack_size);
    struct Command *cmd = find_position(cmds, name);

    if (cmd != NULL)
    {
	list_insert_before(cmd, new);
    }
    else
    {
	list_insert_tail(cmds, new);
    }
    free_buf((union SIGNAL **) &req);
}

static struct Command *
find_command_unique(const List *cmds, const char *name, CmdType entrypoint)
{
    struct Command *cmd = list_head(cmds);

    while (cmd != NULL)
    {
	if (strcmp(cmd->name, name) == 0 && cmd->entrypoint == entrypoint)
	{
	    return cmd;
	}
	cmd = list_next(cmds, cmd);
    }
    return NULL;
}

static void
free_command(struct Command *cmd)
{
    if (--cmd->proxy->refcount == 0)
    {
	free_proxy(cmd->proxy);
    }
    heap_free_private(list_unlink(cmd));
}

static void
remove_command(List *cmds, struct ShellRemoveCmd *req)
{
    struct Command *cmd = find_command_unique(cmds, req->name, (CmdType) req->entrypoint);

    if (cmd != NULL)
    {
	free_command(cmd);
    }
    free_buf((union SIGNAL **) &req);
}

static struct Command *
find_command_byname(const List *cmds, const char *name)
{
    struct Command *cmd =list_head(cmds);

    while (cmd != NULL)
    {
	if (strcmp(cmd->name, name) == 0)
	{
	    return cmd;
	}
	cmd = list_next(cmds, cmd);
    }
    return NULL;
}

static void
command_entrypoint(const List *cmds, struct ShellResolveCmd *req)
{
    PROCESS sender_ = sender((union SIGNAL **) &req);
    struct Command *cmd = find_command_byname(cmds, &req->s[req->cmd_name]);

    req->sigNo = SHELL_RESOLVE_CMD_REPLY;
    if (cmd != NULL)
    {
	req->entrypoint = (OSADDRESS) cmd->entrypoint;
	req->proxy = cmd->proxy->pid;
	req->proc_type = cmd->proc_type;
	req->priority = cmd->priority;
	req->stack_size = cmd->stack_size;
    }
    else
    {
	req->entrypoint = 0;
	req->proxy = 0;
    }
    send((union SIGNAL **) &req, sender_);
}

static void
list_commands(const List *cmds, struct ShellListCmd *req)
{
    PROCESS pid = sender((union SIGNAL **) &req);
    const struct Command *prev = NULL;
    const struct Command *cmd = list_head(cmds);

    while (cmd != NULL)
    {
	if (prev == NULL || strcmp(prev->name, cmd->name) != 0)
	{
	    U32 l1 = strlen(cmd->name);
	    U32 l2 = strlen(cmd->usage);
	    U32 l3 = strlen(cmd->description);
	    struct ShellListCmd *sig = (struct ShellListCmd *) alloc(sizeof (struct
								     ShellListCmd)
								     + 12 + l1
								     + l2 +
								     l3, SHELL_LIST_CMD_REPLY);

	    sig->valid = True;
	    sig->entrypoint = (OSADDRESS)cmd->entrypoint;
	    sig->name = 12;
	    sig->usage = 12 + l1 + 1;
	    sig->description = 12 + l1 + 1 + l2 + 1;
	    sig->proc_type = cmd->proc_type;
	    sig->priority = cmd->priority;
	    sig->stack_size = cmd->stack_size;
	    strcpy(&sig->s[sig->name], cmd->name);
	    strcpy(&sig->s[sig->usage], cmd->usage);
	    strcpy(&sig->s[sig->description], cmd->description);
	    send((union SIGNAL **) &sig, pid);
	}
	prev = cmd;
	cmd = list_next(cmds, cmd);
    }

    req->sigNo = SHELL_LIST_CMD_REPLY;
    req->valid = False;
    send((union SIGNAL **) &req, pid);
}

static void
killed_client(List *proxies, List *cmds, PROCESS pid)
{
    /* If there existed a proxy with this pid then our attach
       reference for that proxy is invalid. */
    {
	struct Proxy *proxy = list_head(proxies);

	while (proxy != NULL)
	{
	    if (proxy->pid == pid)
	    {
		proxy->attref = (OSATTREF) ~ 0;
		break;
	    }
	    proxy = list_next(proxies, proxy);
	}
    }

    /* If there existed a proxy with this pid then all commands
       using this proxy are gone. */
    {
	struct Command *cmd = list_head(cmds);

	while (cmd != NULL)
	{
	    struct Command *next = list_next(cmds, cmd);

	    if (cmd->proxy->pid == pid)
	    {
		free_command(cmd);
	    }
	    cmd = next;
	}
    }
}

struct User
{
    ListLink link;
    PROCESS pid;
    OSUSER userId;
    U32 username;
    U32 terminal;
    char s[2];
};

static void
internal_add_user(List *users, const char *username, const char
		  *terminal, OSUSER user_id, PROCESS pid)
{
    struct User *user = list_init_link(heap_alloc_private(sizeof (struct User) +
	strlen(username) + strlen(terminal), NULL, 0));

    user->pid = pid;
    user->userId = user_id;
    user->username = 0;
    user->terminal = strlen(username) + 1;
    strcpy(&user->s[user->username], username);
    strcpy(&user->s[user->terminal], terminal);
    list_insert_tail(users, user);
    attach(NULL, user->pid);
}

static void
add_user(List *users, struct ShellAddUser *signal)
{
    internal_add_user(users, &signal->s[signal->username],
		      &signal->s[signal->terminal], signal->userId,
		      sender((union SIGNAL **) &signal));
    free_buf((union SIGNAL **) &signal);
}

static void
list_users(List *users, struct ShellListUsersRequest *signal)
{
    PROCESS client = sender((union SIGNAL **) &signal);
    struct User *user = list_head(users);

    while (user != NULL)
    {
	const char *username = &user->s[user->username];
	const char *terminal = &user->s[user->terminal];
	struct ShellListUsersReply *reply =
	     allocShellListUsersReply(strlen(username) + strlen(terminal));

	reply->valid = True;
	reply->userId = user->userId;
	reply->shellPid = user->pid;
	reply->username = 0;
	reply->terminal = strlen(username) + 1;
	strcpy(&reply->s[reply->username], username);
	strcpy(&reply->s[reply->terminal], terminal);
	send((union SIGNAL **) &reply, client);
	user = list_next(users, user);
    }
    {
	struct ShellListUsersReply *reply = allocShellListUsersReply(0);

	reply->valid = False;
	send((union SIGNAL **) &reply, client);
    }
    free_buf((union SIGNAL **) &signal);
}

static void
killed_user(List *users, PROCESS pid)
{
    struct User *user = list_head(users);

    while (user != NULL)
    {
	struct User *next = list_next(users, user);

	if (user->pid == pid)
	{
	    heap_free_private(list_unlink(user));
	    break;
	}
	user = next;
    }
}

static void
interface(struct ShellInterfaceRequest *req)
{
    static const SIGSELECT sigs[] =
    {
	SHELL_ADD_USER, SHELL_LIST_USERS_REQUEST, SHELL_ADD_CMD,
	SHELL_REMOVE_CMD, SHELL_RESOLVE_CMD_REQUEST,
	SHELL_LIST_CMD_REQUEST, SHELL_INTERFACE_REQUEST,
	SHELL_CREATE_CHILD_REQUEST, SHELL_SETENV_CHILD,
	SHELL_SETENV_CHILD_SENDER, SHELL_START_CHILD, SHELL_STOP_CHILD,
	SHELL_KILL_CHILD, SHELL_GETENV_CHILD_REQUEST
    };
    unsigned i;
    unsigned num_sigs = sizeof sigs / sizeof (SIGSELECT);
    struct ShellInterfaceReply *reply = (struct ShellInterfaceReply *)
	     alloc(sizeof (struct ShellInterfaceReply) + num_sigs * sizeof (SIGSELECT), SHELL_INTERFACE_REPLY);

    reply->status = EFS_SUCCESS;
    strncpy(reply->whatStr, "OSE shell daemon " STRSTR(VERSION), sizeof reply->whatStr);
    reply->biosHandle = 0;
    reply->sigs[0] = num_sigs;
    for (i = 0; i < num_sigs; i++)
    {
	reply->sigs[i + 1] = sigs[i];
    }
    send((union SIGNAL **) &reply, sender((union SIGNAL **) &req));
    free_buf((union SIGNAL **) &req);
}

static union SIGNAL *
shell_receive_from(PROCESS from_pid, SIGSELECT signo)
{
    union SIGNAL *sig;
    SIGSELECT sigsel[3];

    sigsel[0] = 2;
    sigsel[1] = OS_ATTACH_SIG;
    sigsel[2] = signo;

    assert(from_pid != 0);
    sig = receive_from(1000, sigsel, from_pid);
    if (sig == NIL)
    {
	OSATTREF ar = attach(NULL, from_pid);

	do
	{
	    sig = receive_from(1000000, sigsel, from_pid);
	}
	while (sig == NIL);
	if (sig->sigNo == OS_ATTACH_SIG)
	{
	    free_buf(&sig);
	    sig = NIL;
	}
	else /* Received wanted signal. */
	{
	    detach(&ar);
	}
    }
    return sig;
}

#define REMOTE_ATTACH 1000 /* FIXME: a better signo? */
struct RemoteAttach
{
    SIGSELECT signo;
    PROCESS child_pid;
};

static void
create_child(List *cmds, List *users, struct ShellCreateChild *req)
{
    PROCESS snd_pid = sender((union SIGNAL **)&req);
    PROCESS child_pid = current_process();	/* Default PID. */
    struct OS_pcb *pcb;
    char          *lhpath;
    char          *slash = NULL;
    char          *terminal;
    int           size;
    char          *tmp_cmd_line;
    int           argc;
    char          **argv;
    Boolean       valid;
    char          *not_valid_status = NULL;

    /* Check if sender name contains lhpath. 
     * The lhpath is everything before last '/'. 
     */
    pcb = get_pcb(sender((union SIGNAL **) &req));
    if (pcb != NULL)
    {
       lhpath = &pcb->strings[pcb->name];
       slash = strrchr(lhpath, '/');

       if (slash != NULL)
       {
          /* lhpath found. */
          *slash = '\0';
       }
    }

    /*
     * We want the string "terminal" to look like: 
     * <lnh_path_prefix><lhpath><lnh_path_separator><terminal_name>
     *
     * We were given either (in the &req->s[req->terminal] variable): 
     * <lnh_path_prefix><lhpath><lnh_path_separator><terminal_name>
     * or:
     * <terminal_name>
     *
     * We have two cases: 
     * insert the lhpath of this node to <lhpath> (and append "/" to it)
     * or:
     * insert prefix (if used), lhpath, separator and terminal name.
     */

    if (slash != NULL)
    {
       char *lnh_path_prefix = NULL;
       char lnh_path_separator[2];
       char *strptr = NULL;
       efs_get_lnh_syntax(&lnh_path_prefix, lnh_path_separator);

       /* Function above thinks lnh_path_separator is a pointer to a character, so
        * it needs to be NULL terminated. */
       lnh_path_separator[1] = 0;

       /* Check if there should be a link handler path at all */
       if ((lnh_path_prefix[0] == 0) && (lnh_path_separator[0] == 0))
       {
          static char defaultTerminal[] = "(remote)";
          terminal = (char*) heap_alloc_private(strlen(defaultTerminal) + 1,
                                                __FILE__, __LINE__);
          strcpy(terminal, defaultTerminal);
          heap_free_private(lnh_path_prefix);
          goto terminal_name_finished;
       }

       if(NULL != strrchr(&req->s[req->terminal], lnh_path_separator[0]))
       {
          /* lnh_path_separator is there, if lnh_path_prefix is used, 
           * it should be there too (we wont look for it).
           */
	  terminal = (char *)heap_alloc_private(strlen(lhpath) + 1 +
						strlen(&req->s[req->terminal]) + 1,
						__FILE__, __LINE__);
	  strcpy(terminal, &req->s[req->terminal]);

          /* Insert '/' + lhpath just after the optional link handler prefix */

	  strptr = terminal + strlen(lnh_path_prefix);
	  memmove(strptr + strlen(lhpath) + 1, strptr, strlen(strptr) + 1);
	  memcpy(strptr, lhpath, strlen(lhpath));
          strptr += strlen(lhpath);
          strptr[0] = '/';
       }
       else
       {
          terminal = (char *)heap_alloc_private(strlen(lnh_path_prefix) + 
                                                strlen(lhpath) + 2 + 
                                                strlen(&req->s[req->terminal]) + 1,
                                                __FILE__, __LINE__);

          strcpy(terminal, lnh_path_prefix);

	  strcat(terminal, lhpath);
	  strcat(terminal, lnh_path_separator);

          /* finally append the terminal name */
          strcat(terminal, &req->s[req->terminal]);
       }
       /* we need to free lnh_path_prefix */
       heap_free_private(lnh_path_prefix);
    }
    else
    {
       terminal = (char *)heap_alloc_private(strlen(&req->s[req->terminal]) + 1,
                                             __FILE__, __LINE__);
       strcpy(terminal, &req->s[req->terminal]);
    }

  terminal_name_finished:

    if (pcb != NULL)
    {
       free_buf((union SIGNAL **) &pcb);
    }

    /* Get argc and argv. */
    size = strlen(&req->s[req->argv]) + 1;
    tmp_cmd_line = (char *)heap_alloc_private(size, __FILE__, __LINE__);
    strcpy(tmp_cmd_line, &req->s[req->argv]);
    efs_split_args(tmp_cmd_line, &argc, &argv);

    /* Validate and log command. */
    valid = validate_cmd(&req->s[req->argv], /* original_cmd_line */
                         &req->s[req->username],
                         terminal,
                         argc,
                         (const char **) argv,
                         0,
                         NULL,
                         0,
                         NULL,
                         0,
                         &not_valid_status);

    if (!valid)
    {
       /* Not allowed to run command. */
       struct ShellCreateChild *reply;

       if (not_valid_status != NULL)
       {
          reply = (struct ShellCreateChild *) alloc(sizeof(struct ShellCreateChild) + 1 +
                                                    strlen(not_valid_status),
                                                    SHELL_CREATE_CHILD_REPLY);
          strcpy(&reply->s[0], not_valid_status);
       }
       else
       {
          reply = (struct ShellCreateChild *) alloc(sizeof(struct ShellCreateChild) + 1,
                                                    SHELL_CREATE_CHILD_REPLY);
          strcpy(&reply->s[0], "");
       }
       reply->status = SHELL_CREATE_CHILD_EILLEGAL_CMD;
       /* FIXME: should use send_w_sender() and check return value */
       send_w_s((union SIGNAL **) &reply, child_pid, snd_pid);

       free_buf((union SIGNAL **) &req);
    }
    else
    {
       /* Allowed to run command. */
       struct Command *cmd = find_command_byname(cmds, &req->s[req->cmd_name]);

       if (cmd != NULL)
       {
	  struct ChildCreateReply *reply;
	  struct ChildCreateRequest *create =
	  (struct ChildCreateRequest *) alloc(sizeof (struct ChildCreateRequest) +
					      strlen(&req->s[req->proc_name]) +
					      strlen(&req->s[req->argv]) +
					      strlen(&req->s[req->wdir]), CHILD_CREATE_REQUEST);

	  create->ptype = cmd->proc_type != DEFAULT_PROC_TYPE ? cmd->proc_type : req->proc_type;
	  create->priority = cmd->priority != DEFAULT_PRIORITY ? cmd->priority : req->priority;
	  create->stack_size = cmd->stack_size != DEFAULT_STACK_SIZE ?
	     cmd->stack_size : 4000;
	  create->entrypoint = (OSADDRESS) cmd->entrypoint;
	  create->user_id = req->user_id;
	  create->name = 0;
	  strcpy(&create->s[create->name], &req->s[req->proc_name]);
	  send((union SIGNAL **) &create, cmd->proxy->pid);
	  reply = (struct ChildCreateReply *)
	     shell_receive_from(cmd->proxy->pid, CHILD_CREATE_REPLY);
	  if (reply != NULL)
	  {
	     child_pid = req->child_pid = reply->child_pid;
	     {
	        /* Supervise child until it has started. */
		struct RemoteAttach *ra = (struct RemoteAttach *)alloc(
		   sizeof(struct RemoteAttach), REMOTE_ATTACH);
		ra->child_pid = child_pid;
		req->att_ref = attach((union SIGNAL **)&ra, snd_pid);
	     }
	     internal_add_user(users, &req->s[req->username], terminal,
			       req->user_id, child_pid);
	     set_env(child_pid, "USERNAME", &req->s[req->username]);
	     set_env(child_pid, "TERMINAL", terminal);
	     set_env(child_pid, "ARGV", &req->s[req->argv]);
	     set_env(child_pid, "CWD", &req->s[req->wdir]);
	     req->status = EFS_SUCCESS;

             free_buf((union SIGNAL **) &reply);
	  }
	  else /* Lost contact with command proxy. */
	  {
             req->status = SHELL_CREATE_CHILD_ECMD_PROXY;
	  }
       }
       else /* Unknown command. */
       {
	  req->status = SHELL_CREATE_CHILD_EUNKNOWN_CMD;
       }
       req->sigNo = SHELL_CREATE_CHILD_REPLY;
       /* FIXME: should use send_w_sender() and check return value */
       send_w_s((union SIGNAL **) &req, child_pid, snd_pid);
    }

    heap_free_private(terminal);
    heap_free_private(tmp_cmd_line);
    heap_free_private(argv);

    if (not_valid_status != NULL)
    {
       heap_free_private(not_valid_status);
    }
}

/* Lost contact with the remote client, kill the orphaned child. */
static void
remote_attach(struct RemoteAttach *ra)
{
    kill_proc(ra->child_pid);
    free_buf((union SIGNAL **)&ra);
}

static void
setenv_child(struct ShellSetenvChild *req)
{
    if (get_ptype(req->child_pid) != OS_ILLEGAL)
    {
	char *name = req->assignment;
	char *value = strchr(name, '=');

	if (value != NULL)
	{
	    *value++ = '\0';
	    set_env(req->child_pid, name, value);
	}
	/* Else invalid assignment ignored. */
    }
    /* No reply sent. */
    free_buf((union SIGNAL **) &req);
}

static void
setenv_child_sender(struct ShellSetenvChildSender *req)
{
    if (get_ptype(req->child_pid) != OS_ILLEGAL)
    {
	char *name = req->assignment;
	char *format = strchr(name, '=');

	if (format != NULL)
	{
	    char value[256];

	    *format++ = '\0';
	    snprintf(value, sizeof value, format, sender((union SIGNAL **) &req));
	    set_env(req->child_pid, name, value);
	}
	/* Else invalid assignment ignored. */
    }
    /* No reply sent. */
    free_buf((union SIGNAL **) &req);
}

static void
start_child(struct ShellStartChild *req)
{
    if (get_ptype(req->child_pid) != OS_ILLEGAL)
    {
	start(req->child_pid);
    }
    /* Since the child has started we need not supervise it any more. */
    detach(&req->att_ref);
    /* No reply sent. */
    free_buf((union SIGNAL **) &req);
}

static void
stop_child(struct ShellStopChild *req)
{
    if (get_ptype(req->child_pid) != OS_ILLEGAL)
    {
	stop(req->child_pid);
    }
    /* No reply sent. */
    free_buf((union SIGNAL **) &req);
}

static void
kill_child(struct ShellKillChild *req)
{
    if (get_ptype(req->child_pid) != OS_ILLEGAL)
    {
	kill_proc(req->child_pid);
    }
    /* No reply sent. */
    free_buf((union SIGNAL **) &req);
}

static void
getenv_child(struct ShellGetenvChildRequest *req)
{
    EfsBoolean success = False;
    const char *value = "";
    struct ShellGetenvChildReply *reply;
    if (get_ptype(req->child_pid) != OS_ILLEGAL)
    {
	char *buf = get_env(req->child_pid, req->name);
	if (buf != NULL)
	{
	    value = buf;
	    success = True;
	}
    }
    reply = (struct ShellGetenvChildReply *)
	alloc(sizeof(struct ShellGetenvChildReply) + strlen(req->name)
	    + strlen(value) + 1, SHELL_GETENV_CHILD_REPLY);
    reply->success = success;
    reply->child_pid = req->child_pid;
    sprintf(reply->assignment, "%s=%s", req->name, value);
    send((union SIGNAL **)&reply, sender((union SIGNAL **)&req));
    if (success)
    {
	free_buf((union SIGNAL **)&value);
    }
    free_buf((union SIGNAL **) &req);
}

OSENTRYPOINT ose_shelld;
OS_PROCESS(ose_shelld)
{
    List proxies;
    List cmds;
    List users;

    list_init(&proxies);
    list_init(&cmds);
    list_init(&users);

    for (;;)
    {
	static const SIGSELECT sigsel[] = {0};
	union SIGNAL *sig = receive(sigsel);

	switch (sig->sigNo)
	{
	    case SHELL_ADD_CMD:
		add_command(&proxies, &cmds, (struct ShellAddCmd *) sig);
		break;
	    case SHELL_REMOVE_CMD:
		remove_command(&cmds, (struct ShellRemoveCmd *) sig);
		break;
	    case SHELL_RESOLVE_CMD_REQUEST:
		command_entrypoint(&cmds, (struct ShellResolveCmd *) sig);
		break;
	    case SHELL_LIST_CMD_REQUEST:
		list_commands(&cmds, (struct ShellListCmd *) sig);
		break;
	    case SHELL_ADD_USER:
		add_user(&users, (struct ShellAddUser *) sig);
		break;
	    case SHELL_LIST_USERS_REQUEST:
		list_users(&users, (struct ShellListUsersRequest *) sig);
		break;
	    case SHELL_INTERFACE_REQUEST:
		interface((struct ShellInterfaceRequest *) sig);
		break;
	    case SHELL_CREATE_CHILD_REQUEST:
		create_child(&cmds, &users, (struct ShellCreateChild *) sig);
		break;
	    case SHELL_SETENV_CHILD:
		setenv_child((struct ShellSetenvChild *) sig);
		break;
	    case SHELL_SETENV_CHILD_SENDER:
		setenv_child_sender((struct ShellSetenvChildSender *) sig);
		break;
	    case SHELL_START_CHILD:
		start_child((struct ShellStartChild *) sig);
		break;
	    case SHELL_STOP_CHILD:
		stop_child((struct ShellStopChild *) sig);
		break;
	    case SHELL_KILL_CHILD:
		kill_child((struct ShellKillChild *) sig);
		break;
	    case SHELL_GETENV_CHILD_REQUEST:
		getenv_child((struct ShellGetenvChildRequest *) sig);
		break;
	    case REMOTE_ATTACH:
		remote_attach((struct RemoteAttach *)sig);
		break;
	    case OS_ATTACH_SIG:
		{
		    PROCESS pid = sender((union SIGNAL **) &sig);

		    killed_user(&users, pid);
		    killed_client(&proxies, &cmds, pid);
		    free_buf((union SIGNAL **) &sig);
		    break;
		}
	    default:
		free_buf(&sig);
	}
    }
}
