/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2010 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

#include "ose.h"
#include "ose_err.h"
#include "stdlib.h"
#include "string.h"
#include "thinlib.h"
#include "thinlib_err.h"
#include "bios.h"
#include "spinlock.h"

struct ThinlibTable
{
   const char *lib_name;
   const struct ThinlibEntry *entries;
   long num_entries;
};

static struct ThinlibTable thin_libs[10];

static int
cmp_thin_entries(const void *e0, const void *e1)
{
   const struct ThinlibEntry *ee0 = (const struct ThinlibEntry *) e0;
   const struct ThinlibEntry *ee1 = (const struct ThinlibEntry *) e1;
   return strcmp(ee0->name, ee1->name);
}

static long
get_function_pointer(const char *lib, const char *function)
{
   struct ThinlibEntry key, *res;
   size_t idx;

   /* Fetch the correct index. */
   for (idx = 0;
        ((idx < sizeof(thin_libs) / sizeof(thin_libs[0])) &&
         (thin_libs[idx].lib_name != NULL));
        ++idx)
   {
      if (strcmp(thin_libs[idx].lib_name, lib) == 0)
      {
         /* Correct index, fetch the correct function pointer. */
         key.name = (const char *)function;
         res = bsearch(&key, thin_libs[idx].entries, thin_libs[idx].num_entries,
                       sizeof(struct ThinlibEntry), cmp_thin_entries);
         if (res != NULL)
         {         
            return (long)(res->function);
         }
         error2(THINLIB_EUNKNOWN_FUNCTION, (OSERRCODE)function);
         return 0;   
      }         
   }

   /* No entries were found for the library in question. */
   error2(THINLIB_ELIB_NOT_FOUND, (OSERRCODE)lib);
   return 0;      
}

/* Returns 1 on success. */
static long
install_function_pointers(const char *lib, const struct ThinlibEntry *entries,
                          long num_entries)
{
   size_t idx;
   Msr sr;
   long res;
#ifdef OSE_SMP   
   static spinlock_t spinlock  = SPIN_LOCK_INIT;
#endif
   
   spin_lock_irqsave(&spinlock, sr);
      
   /* Store the information in the first free entry. */
   for (idx = 0; idx < sizeof(thin_libs) / sizeof(thin_libs[0]); ++idx)
   {
      if (thin_libs[idx].lib_name != NULL)
      {
         /* Check for duplicate installation. */
         if (strcmp(thin_libs[idx].lib_name, lib) == 0)
         {
            res = 3;
            goto done;
         }         
         continue;
      }

      /* Store everything. The caller is not allowed to free or reuse the passed
       * pointers. */
      thin_libs[idx].lib_name = lib;
      thin_libs[idx].entries = entries;
      thin_libs[idx].num_entries = num_entries;

      /* All is well! */
      res = 1;
      goto done;
   }

   /* Out of slot space. */
   res = 2;

   /* Common exit to ensure unlocking. */
  done:
   spin_unlock_irqrestore(&spinlock, sr);
   
   return res;
}

static long
thinlib_trap_handler(long fc, long arg1, long arg2, long arg3,
                     long arg4, long arg5, long arg6)
{
   (void)arg3;
   (void)arg4;
   (void)arg5;
   (void)arg6;
   
   switch (fc)
   {      
      case 1:
         return install_function_pointers((const char *)arg1,
                                          (const struct ThinlibEntry *)arg2,
                                          arg3);
      case 2:
         return get_function_pointer((const char *)arg1,
                                     (const char *)arg2);
      default:
         error2(THINLIB_EBIOS_INVALID, fc);
         return 0;
   }
}

void
ose_thinlib_init(void)
{
   long err = biosInstall(THINLIB_BIOS_NAME, thinlib_trap_handler,
                          BIOS_USERMODE_ALLOWED);

   /* We are ok if we are the first, or if the bios is already present. */
   if ((err != BIOS_SUCCESS) && (err != BIOS_EDUPLICATE_NAME))
   {
      error2(OSE_EFATAL_MASK | THINLIB_EBIOS_INSTALL, err);
   }   
}
