/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2005-2006 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

/**
 * @file ramdump_heap.h
 *
 * @brief This defines the interface to the ramdump freeze mode heap.
 *
 */
#ifndef _RAMDUMP_HEAP_H
#define _RAMDUMP_HEAP_H

#include <stddef.h>

#ifdef __cplusplus
extern "C" {
#endif

/**
 * @macro RAMDUMP_HEAP_ENO_BASE_ADDRESS_CONFIGURED
 *
 * @brief Error code indicating that no base address was configured.
 */
#define RAMDUMP_HEAP_ENO_BASE_ADDRESS_CONFIGURED  (-1)

/**
 * @macro RAMDUMP_HEAP_EHEAP_AREA_SIZE_TOO_SMALL
 *
 * @brief Error code indicating that the configured heap area size
 * is too small.
 */
#define RAMDUMP_HEAP_EHEAP_AREA_SIZE_TOO_SMALL  (-2)

/**
 * @type RamdumpHeapStatus
 *
 * @brief A signed integral type used for the return status
 * of some ramdump heap functions.
 */
typedef int RamdumpHeapStatus;
   
/**
 * @function ramdump_heap_init
 *
 * @brief Initializes heap.
 *
 * @long The function ramdump_heap_init initializes the heap by
 * creating a list of free memory blocks starting at the specified
 * base address. The amount of free memory initially available on
 * the heap is determined by the size parameter. The size of the
 * heap is fixed.
 *
 * @input base The start address of the heap memory area.
 * @input size The size of the heap memory area.
 *
 * @return Returns zero on success or one of
 * RAMDUMP_HEAP_EHEAP_AREA_SIZE_TOO_SMALL and
 * RAMDUMP_HEAP_ENO_BASE_ADDRESS_CONFIGURED on failure.
 *
 * @example
 * RamdumpHeapStatus status = ramdump_heap_init(0x10000, 1000);
 */
RamdumpHeapStatus ramdump_heap_init(char *base, size_t size);

/**
 * @function ramdump_heap_malloc
 *
 * @brief Allocates a region large enough to hold an object
 * of specified size.
 *
 * @long The function ramdump_heap_malloc allocates a region of memory
 * large enough to hold an object whose size (as measured by the
 * sizeof operator) equals the parameter size. A pointer to the
 * first element of the region is returned. The caller may then use
 * a cast operator to convert this pointer to another pointer type.
 * If it is impossible for some reason to perform the requested
 * allocation, then a NULL pointer is returned. If the requested
 * size is 0, then the function will return either a NULL pointer
 * or a non-NULL pointer that nonetheless must not be used to access
 * an object. The allocated memory is not initialized in any way,
 * so the caller cannot depend on its contents.
 *
 * @input size Size of allocated memory region.
 *
 * @return Pointer to the allocated memory region, or NULL if
 * the allocation can not be performed.
 *
 * @example
 * int *ptr = (int *)ramdump_heap_malloc(sizeof(int));
 */
void *ramdump_heap_malloc(size_t size);

/**
 * @function ramdump_heap_calloc
 *
 * @brief Allocates a region of memory large enough to hold
 * an array of element_count elements, each of size element_size.
 *
 * @long The function ramdump_heap_calloc allocates a region of memory
 * large enough to hold an array of element_count elements, each
 * of size element_size (typically given by the sizeof operator).
 * The region of memory is cleared bitwise to zero, and a pointer
 * to the first element of the region is returned. If for some
 * reason it is impossible to perform the requested allocation,
 * or if element_count or element_size is zero, then the return
 * value is NULL.
 *
 * @input element_count Number of elements to allocate space for.
 * @input element_size The size of each element.
 *
 * @return Pointer to the first element in the allocated memory area.
 *
 * @example
 * int *int_array = (int *)ramdump_heap_calloc(100, sizeof(int));
 */
void *ramdump_heap_calloc(size_t element_count, size_t element_size);

/**
 * @function ramdump_heap_realloc
 *
 * @brief Takes a pointer to a previously allocated memory
 * region and changes its size while preserving its contents.
 *
 * @long The function ramdump_heap_realloc takes a pointer to a memory
 * region previously allocated by one of the standard functions
 * and changes its size while preserving its contents. If necessary,
 * the contents are copied to a new memory region. A pointer to the
 * (possibly new) memory region is returned. If the request cannot
 * be satisfied, a null pointer is returned and the old region is
 * not disturbed. If the first argument to ramdump_heap_realloc is a null
 * pointer, then the function behaves like ramdump_heap_malloc. If the
 * pointer parameter is not NULL and the size parameter is zero, then
 * ramdump_heap_realloc returns either a NULL pointer or a pointer that
 * must not be used, and the old region is deallocated. If the new
 * size is smaller than the old size, then some of the old contents
 * at the end of the old region will be discarded. If the new size
 * is larger than the old size, then all of the old contents are
 * preserved and new space is added at the end. The new space is not
 * initialized in any way, and the caller must assume that it contains
 * garbage information. Whenever ramdump_heap_realloc returns a pointer
 * that is different from its first argument, the programmer should
 * assume that the old region of memory was free and should not be used.
 *
 * @input pointer Pointer to the memory region to reallocate.
 * @input size Size of the reallocated region.
 *
 * @return Pointer to the reallocated memory region or NULL if
 * the reallocation could not be performed.
 *
 * @example
 * char *ptr = (char *)ramdump_heap_malloc(100 * sizeof(char));
 * ptr = (char *)ramdump_heap_realloc(ptr, 200 * sizeof(char));
 */
void *ramdump_heap_realloc(void *pointer, size_t size);

/**
 * @function ramdump_heap_free
 *
 * @brief Deallocates a previously allocated memory region.
 *
 * @long The function ramdump_heap_free deallocates a region of memory
 * previously allocated by ramdump_heap_malloc, ramdump_heap_calloc or
 * ramdump_heap_realloc. The argument of ramdump_heap_free must be a pointer
 * that is the same as a pointer previously returned by one of the
 * allocation functions. If the argument is a NULL pointer, then
 * the call has no effect. Once a region of memory has been freed,
 * it must not be used for any other purpose. The use of any pointer
 * into the region (a "dangling pointer") will have unpredictable
 * effects. Likewise, allocating a region of storange once, but
 * freeing it more than once, has unpredictable effects.
 *
 * @input pointer Pointer to a previously allocated region of memory
 * that will be freed.
 *
 * @example
 * char *ptr = (char *)ramdump_heap_malloc(100 * sizeof(char));
 * ramdump_heap_free(ptr);
 */
void ramdump_heap_free(void *pointer);

/**
 * @function ramdump_heap_start_hander1
 *
 * @brief Start handler that initializes the ramdump heap.
 */
void ramdump_heap_start_handler1(void);
   
#ifdef __cplusplus
}
#endif
   
#endif /* _RAMDUMP_HEAP_H */
