/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2006 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

#ifndef _RAMDUMP_PRIVATE_H
#define _RAMDUMP_PRIVATE_H

#include "fmm.h"
#include "ramdump.h"

/**
 * @type RamdumpCompressValue
 * @brief Value indicates whether a certain dump operation should
 * use compression or not. 
 * 
 * @field RAMDUMP_NO_COMPRESSION     Do not use compression.
 * @field RAMDUMP_ZLIB_COMPRESSION   Use zlib compression.
 */
enum RamdumpCompressValue
{
   RAMDUMP_COMPRESSION_NONE = 0,
   RAMDUMP_COMPRESSION_ZLIB = 1,
};

/**
 * @function ramdump_private_start_new_dump
 *
 * @brief Starts a new dump.
 *
 * @long This function starts a new dump after the last one in
 * the post-mortem dump area. Subsequent calls to ramdump_dump_xxx
 * will write information in the same dump.
 *
 * @long If several dumps are preferred, call this function to start a
 * new dump.
 *
 * @long NOTE: This function does not switch the stack.
 *
 * @return If the function succeeds, RAMDUMP_SUCCESS is returned.
 * If the dump function was called recursively, the error code
 * RAMDUMP_ERECURSIVE_DUMP_CALL is returned. If the ramdump area is
 * uninitialized, the error code RAMDUMP_ERAMDUMP_NOT_INITIALIZED
 * is returned.
 *
 */
RamdumpStatus
ramdump_private_start_new_dump(void);

/**
 * @function ramdump_private_clear_dumps
 *
 * @brief Clears dump area.
 *
 * @long This function clears the dump area.
 *
 * @long NOTE: This function does not switch the stack.
 *
 * @return If the function succeeds, RAMDUMP_SUCCESS is returned.
 * If the dump function was called recursively, the error code
 * RAMDUMP_ERECURSIVE_DUMP_CALL is returned. If the ramdump area is
 * uninitialized, the error code RAMDUMP_ERAMDUMP_NOT_INITIALIZED
 * is returned.
 *
 */
RamdumpStatus
ramdump_private_clear_dumps(void);

/**
 * @function ramdump_private_dump_error_info
 *
 * @brief Writes information about an error event into the current dump.
 *
 * @long Writes information about an error event into the current dump.
 *
 * @long NOTE: This function does not switch the stack.
 *
 * @input description User-provided optional text. The string is only
 * copied up to the first terminator character.
 * @input user_called A parameter to the error handler.
 * @input error_code A parameter to the error handler.
 * @input extra A parameter to the error handler
 *
 * @return If the function succeeds, RAMDUMP_SUCCESS is returned.
 * If the current dump exceeds its size limit, the error code
 * RAMDUMP_ECURR_DUMP_TOO_LARGE is returned. If the description pointer
 * is invalid, the error code RAMDUMP_EMEMORY_ACCESS_NOK is returned.
 * If the dump function was called recursively, the error code
 * RAMDUMP_ERECURSIVE_DUMP_CALL is returned. If the ramdump area is
 * uninitialized, the error code RAMDUMP_ERAMDUMP_NOT_INITIALIZED
 * is returned. RAMDUMP_EUNABLE_TO_GET_CURRENT_PROCESS_INFO is returned
 * if the function is unable to get information for the current process.
 */
RamdumpStatus
ramdump_private_dump_error_info(const char *description,
                                OSBOOLEAN user_called,
                                OSERRCODE error_code,
                                OSERRCODE extra);
         
/**
 * @function ramdump_private_dump_text_info
 *
 * @brief Writes an arbitrary textual description into the current dump.
 *
 * @long Writes an arbitrary textual description into the current dump.
 *
 * @long NOTE: This function does not switch the stack.
 *
 * @input description User-provided optional text. The string
 * is only copied up to the first terminator character.
 *
 * @return If the function succeeds, RAMDUMP_SUCCESS is returned.
 * If the current dump exceeds its size limit, the error code
 * RAMDUMP_ECURR_DUMP_TOO_LARGE is returned. If the description pointer
 * is invalid, the error code RAMDUMP_EMEMORY_ACCESS_NOK is returned.
 * If the dump function was called recursively, the error code
 * RAMDUMP_ERECURSIVE_DUMP_CALL is returned. If the ramdump area is
 * uninitialized, the error code RAMDUMP_ERAMDUMP_NOT_INITIALIZED
 * is returned.
 */
RamdumpStatus
ramdump_private_dump_text_info(const char *description);

/**
 * @function ramdump_private_dump_memory_image
 *
 * @brief Dumps a section of memory to the current dump.
 *
 * @long Dumps a section of memory to the current dump.
 *
 * @long NOTE: This function does not switch the stack.
 *
 * @input description User-provided optional text. The string is only
 * copied up to the first terminator character.
 * @input start_address Start address of memory to dump.
 * @input length Number of bytes to dump.
 *
 * @return If the function succeeds, RAMDUMP_SUCCESS is returned.
 * If the current dump exceeds its size limit, the function fails
 * and returns the error code RAMDUMP_ECURR_DUMP_TOO_LARGE. If the memory
 * access fails or the description pointer is invalid, the error code
 * RAMDUMP_EMEMROY_ACCESS_NOK is returned. If the dump function was called
 * recursively, the error code RAMDUMP_ERECURSIVE_DUMP_CALL is returned.
 * If the ramdump area is unitialized, the error code
 * RAMDUMP_ERAMDUMP_NOT_INITIALIZED is returned.
 */
RamdumpStatus
ramdump_private_dump_memory_image(const char *description,
                                  OSADDRESS start_address,
                                  U32 length);
           
/**
 * @function ramdump_private_set_watchdog_function
 *
 * @brief Sets a watchdog callback function to be called during dumping.
 *
 * @long Watchdog support is provided if dumps are time consuming.
 * A callback function can be installed in supervisor mode,
 * to be called at regular intervals during the dump phase.
 * It is an error to call this function in non-supervisor mode.
 * Installing a NULL pointer disables watchdog support.
 *
 * @long NOTE: This function does not switch the stack.
 *
 * @input f Pointer to the watchdog callback function to install.
 *
 * @return If the function succeeds in setting the watchdog callback
 * function pointer, the return value is RAMDUMP_SUCCESS. If the ramdump
 * area is uninitialized, the function fails and returns the error code
 * RAMDUMP_ERAMDUMP_NOT_INITIALIZED.
 *
 * @seealso
 *
 * @example
 *
 */
RamdumpStatus
ramdump_private_set_watchdog_function(RamdumpWatchdogCb *f);

/**
 * @function ramdump_private_dump_request
 * 
 * @brief Writes the request signal of a transaction to the current dump. 
 * 
 * @long This function is meant for opening the dump of a signal
 * transaction (request reply1 [... replyN endmark]) by writing the request
 * signal to the dump. Any subsequent replies should be dumped using the
 * ramdump_private_dump_transaction_reply function. The endmark, if any,
 * should be dumped using the ramdump_private_transaction_end function.
 * Note that the ramdump_private_transaction_end function should be called
 * in order to close the transaction dump block, even if the transaction
 * has no endmark signal.
 * 
 * @long NOTE: This function does not switch the stack before proceeding.
 * 
 * @input request Pointer to a request signal in FmmSignal format. Signal
 * will be written in its entirety to the dump.
 * @input compress Request signal (and subsequent reply signals, if any)
 * will be written to the dump using compression if this parameter is set
 * to the corresponding value.
 * 
 * @return If the function succeeds, RAMDUMP_SUCCESS is returned.
 * If the current dump exceeds its size limit, the function fails
 * and returns the error code RAMDUMP_ECURR_DUMP_TOO_LARGE.
 * If the compression operation fails, RAMDUMP_ECOMPRESSION_FAILED is returned.
 * If the dump function was called recursively, the error code
 * RAMDUMP_ERECURSIVE_DUMP_CALL is returned.
 * If the ramdump area is unitialized, the error code
 * RAMDUMP_ERAMDUMP_NOT_INITIALIZED is returned.
 */
RamdumpStatus
ramdump_private_dump_request(const FmmSignal *request, 
                             enum RamdumpCompressValue compression);

/**
 * @function ramdump_private_dump_reply
 * 
 * @brief Writes a reply signal of a transaction to the current dump. 
 * 
 * @long This function is meant for writing a reply of a transaction
 * (request reply1 [... replyN endmark]). A requirement for using this function
 * is that the last thing written to the current dump was either a request
 * signal (using ramdump_private_dump_transaction_begin) or a reply signal
 * (using ramdump_private_dump_transaction_reply).
 * 
 * @long NOTE: This function does not switch the stack before proceeding.
 * 
 * @input reply Pointer to a reply or endmark signal in FmmSignal format.
 * The signal will be written in its entirety to the dump.
 * 
 * @return If the function succeeds, RAMDUMP_SUCCESS is returned.
 * If the current dump exceeds its size limit, the function fails
 * and returns the error code RAMDUMP_ECURR_DUMP_TOO_LARGE.
 * If the compression operation fails, RAMDUMP_ECOMPRESSION_FAILED is returned.
 * If the dump function was called recursively, the error code
 * RAMDUMP_ERECURSIVE_DUMP_CALL is returned.
 * If the ramdump area is unitialized, the error code
 * RAMDUMP_ERAMDUMP_NOT_INITIALIZED is returned.
 */
RamdumpStatus
ramdump_private_dump_reply(const FmmSignal *reply);

/**
 * @function ramdump_private_dump_endmark
 * 
 * @brief Writes an endmark signal of a transaction to the current dump. 
 * 
 * @long This function is meant for writing an endmark signal, if any, of
 * a transaction (request reply1 [... replyN endmark]) and also for closing
 * the transaction dump block. A requirement for using this function is that
 * the last thing written to the current dump was either a request signal
 * (using ramdump_private_dump_transaction_begin) or a reply signal 
 * (using ramdump_private_dump_transaction_reply). 
 * 
 * @long NOTE: This function does not switch the stack before proceeding.
 * 
 * @input endmark Pointer to a reply or endmark signal in FmmSignal format.
 * The signal will be written in its entirety to the dump. If the transaction
 * has no endmark, this parameter should be NULL.
 * @input status A status value indicating the status of the dumped signal
 * transaction. Zero is reserved for a success status value. Any non-zero
 * value indicates a failure status value. Until this function has been called,
 * the status value present in the dump for the unfinished transaction will
 * contain the default failure status value 0xBADC0DE. Any further
 * differentiation of the non-zero status value space is up to the user of
 * the function.
 * 
 * @return If the function succeeds, RAMDUMP_SUCCESS is returned.
 * If the current dump exceeds its size limit, the function fails
 * and returns the error code RAMDUMP_ECURR_DUMP_TOO_LARGE.
 * If the compression operation fails, RAMDUMP_ECOMPRESSION_FAILED is returned.
 * If the dump function was called recursively, the error code
 * RAMDUMP_ERECURSIVE_DUMP_CALL is returned.
 * If the ramdump area is unitialized, the error code
 * RAMDUMP_ERAMDUMP_NOT_INITIALIZED is returned.
 */
RamdumpStatus
ramdump_private_dump_endmark(const FmmSignal *endmark, int32_t status);

#endif /* _RAMDUMP_PRIVATE_H */
