/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2012 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

/*
** NOTE!
** This module must handle all kinds of external function references with
** extreme care when used from the exception handler. The topic is sharing
** of code between the exception handler and the debugged program.
** Shared code may lead to recursive exceptions.
**
** The exception handler uses a per-process data mechanism to protect itself
** from such recursive exceptions. However, it will not cope with any such
** recursion until it has established that mechanism. The same care is
** needed in the tail of the exception handler, where the mechanism is
** disabled again.
**
** As a consequence of this, there is a local copy of the list packet in
** this module.
** 
** Be aware that different compilers uses different helper functions that
** is not directly visible in the source code. A common example is the
** long long arithmetic helpers. Long long variables should not be used in
** unprotected code areas. Also look out for implicit memcpy calls.
**
** HANDLE_EP_IN_VENEER (OSPP OSMIPS OSARM)
** When a EP is created the memory where the EP is placed is replaced with a
** trap/breakpoint instruction and a memory fragment is allocated (veneer area) 
** where the original instruction is executed together with a way to get back 
** to the instruction after the EP that are still in the right place. 
** Some instruction can not just be moved and in that case a new smal code 
** fragment to replace the instruction is placed in the veneer area.
**
** The veneer area is allocated as a region from MM in PAGE sized chucks (4kb).
** Each chunk fits many (NBR_VENEERS) veneer areas. The chunks are put in a 
** linked list and the veneer areas in them are free for EP in all programs 
** to use. In some (all?) architectures you can make a branch in one 
** instruction if you branch shorter than some range, so it
** is better if you can get a veneer area near enough (+-32Mb for PowerPC), 
** the code tries to get an area within this range in the first place. If
** it is not possible to get memory near enugh it will take whatever memory 
** it gets. To solve the long jump a new trap/breakpoint instruction is put 
** in the in the veneer area that will be detected in the callback and 
** handle the return from the veneer area.
**
** Removing an EP is more tricky as we have no clue if someone is executing 
** in the veneer area when we want to remove it. This is solved by not removing  
** the veneer area directly, we just mark it USED, and send a signal to a 
** garbage collect background process that will take care of removing it 
** and freeing the PAGE chuck if it becomes empty.
** If all veneer area fragments in the memory PAGE chunck 
** are marked FREE the region will be returned to the MM by the GC process.
** The garbage collect background process  will ensure that all that needs 
** to execute on all cores has executed before cleaning out a veneer fragment.
** 
** FAST_EP (PPC only for now)
** A fast ep is an ep that calls some code that will determine if the ep 
** call back should be called. The idea here is some sort of macro code make a 
** really fast check and the "slow" code path "all the way" trough the exception 
** handler is only taken when needed. 
** 
** Behind the scenes the FAST EP code will allocate an area from mm for it's
** own code and date and also for a codearea used by the caller where they place
** the checking code. The memory is placed so near the original code that a 
** one-instruction jump is possible. Instead of a trap the jump instruction 
** is placed that will jump to the allocated memory area. When allocating and 
** there is no memory area near enough it will fail, if there is memory a 
** pointer to an area where to put the checking code is supplied.
** This is used in tracepoint code when using fast tracepoints here the 
** tracepoint opcodes has been JIT:ed (converted to ppc assembler) and placed 
** in this memory area. 
**  
** When enabling a FAST EP the one-instruction-jump is placed on the original 
** address that jumps to some pre/postemble that will call the check-callback 
** code area, and depending on how that code area return it will execute a TRAP 
** or just the original opcode and jump back to the instruction after.
** 
** If a TRAP was executed this is detected inside the EPM callback code and are 
** handled there. 
** 
*/

#include "epm.h"
#include "sysparam.h"
#include "string.h"
#include "epm_private.h"
#ifndef softOSE
#include "inline.h"
#endif
#include "pose.h"


#ifdef softOSE

EpmStatus
ose_epm_init(void)
{
   return EPM_OK;
}

EpmStatus
ose_epm_add_execution_point(struct EpmExecutionPoint *ep)
{
   (void) ep;
   return EPM_FUNCTION_NOT_IMPLEMENTED;
}
EpmStatus
ose_epm_remove_execution_point(struct EpmExecutionPoint *ep)
{
   (void) ep;
   return EPM_FUNCTION_NOT_IMPLEMENTED;
}

EpmStatus
ose_epm_next_instruction(struct cpu_regdump *dump,
			 U32 options,
			 OSADDRESS *next_address)
{
   (void) dump;
   (void) options;
   (void) next_address;

   return EPM_FUNCTION_NOT_IMPLEMENTED;
}

EpmStatus
ose_epm_disable_all(void)
{
   return EPM_FUNCTION_NOT_IMPLEMENTED;
}

EpmStatus
ose_epm_enable_all(void)
{
   return EPM_FUNCTION_NOT_IMPLEMENTED;
}

void *
ose_epm_alloc_fast_execution_point(struct EpmExecutionPoint *ep, 
                                   size_t code_size)
{
   (void) ep;
   (void) code_size;
   return NULL;
}

EpmStatus
ose_epm_free_fast_execution_point(struct EpmExecutionPoint *ep)
{
   (void) ep;
   return EPM_FUNCTION_NOT_IMPLEMENTED;
}

EpmStatus
ose_epm_enable_fast_execution_point(struct EpmExecutionPoint *ep)
{
   (void) ep;
   return EPM_FUNCTION_NOT_IMPLEMENTED;
}

EpmStatus
ose_epm_disable_fast_execution_point(struct EpmExecutionPoint *ep)
{
   (void) ep;
   return EPM_FUNCTION_NOT_IMPLEMENTED;
}


int
ose_epm_fast_execution_point_inside_check_callback(OSADDRESS address)
{
   (void) address;
   return 0;
}

#else

#include "stddef.h"
#include "ose.h"
#include "efmi.h"
#include "efmi_registers.h"
#include "exception.h"
#include "ramlog.h"
#include "assert.h"
#include "epm_internal.h"
#include "cpu_intlock.h"
#include "stdint.h"
#include "atomic.h"

/* TODO 
 * Add support for other architectures then remove the HANDLE_EP_IN_VENEER flag
 * When it will be default remove the non HANDLE_EP_IN_VENEER code as a 
 * cleanup.
 */
/* TODO check EP and fast EP,  what shall we do it there is an normal EP 
 * and also the reverse, check for fast EP when we add a normal EP 
 */



   /*  HANDLE_EP_IN_VENEER
    *  To make EPM work in multicore the SMP_STOP_OTHER_CPUS() and
    *  SMP_START_OTHER_CPUS() must be designed away. The spinlock should
    *  also be removed if possible. This will also make it considerably faster
    *  from around 24000 cycles to around 1600 cycles
    *
    *  The original idea of EPM Multicore was to replace the breakpoint/trap
    *  instruction when the EP is done and do a single step and put the
    *  breakpoint/trap instruction back.
    *
    *  With veneer areas the idea is to create a spare memory area where
    *  the original instruction is executed and then jump back and continue
    *  execution on the instruction after it.
    *
    */
#if defined(OSPP) || defined (OSMIPS)
    #define HANDLE_EP_IN_VENEER
#elif defined(OSARM)
   /*
   ** FIXME:
   ** ARM: HANDLE_EP_IN_VENEER works for ARM and thumb2 but not for thumb1.
   ** This implies that it also works for libcore_smp.a.
   ** The bug is seen with many active, tightly placed eps.
   ** So, we currently enable veneers in multicore and RVCT thumb2 libs.
   */
  #if defined(OSE_SMP) || (defined(__TARGET_ARCH_ARM) && __TARGET_ARCH_ARM >=7)
    #define HANDLE_EP_IN_VENEER
  #endif
#endif

#if defined(OSPP) && defined(HANDLE_EP_IN_VENEER)
   #define FAST_EP
#endif

#if defined(HANDLE_EP_IN_VENEER)
   #define ENABLE_MEMCPY_UNLOCK
   #ifndef FAST_EP
      //#define ENABLE_SYNC_GC
   #endif
   #define ENABLE_INNER_STOP_INVALIDATE
#endif

#ifdef HANDLE_EP_IN_VENEER
#include "mm.sig"
#include "mm_internal.sig"
#include "limits.h"
#include "stdlib.h"
#include "cache.h"
#if defined(OSMIPS) || defined(OSPP)
#include "exec_arch.h"
#include "cpu_regs.h"
#endif
#include "mm.h"
#include "bios.h"

#endif

#ifdef OSE_SMP
#include "cpu_smp.h"
#endif
U8 *zzfind_current_debug_info(void); /* pose */

#ifdef HANDLE_EP_IN_VENEER

/* 
 * Veneer fragment gc process PID, someone might execute from 
 * the veneer area when the remove ep method is called.
 * To ensure that this does not happen the veneer fragment is marked "used"
 * and then switched back to "free" in this process
 */

PROCESS epm_veneer_gc_process_;
PROCESS gmm_;

#define VENEER_GC_STACK 512

#if defined(OSPP)
/* PPC */
/* MAX_JUMP_IN_ONE_INSTRUCTION
 * How far away can the instruction be placed and still be able to jump back
 * to the original place.
 * +-32Mb for PowerPC 
 */
#define MAX_JUMP_IN_ONE_INSTRUCTION (32*1024*1024 - \
        sizeof(struct ep_veneer_fragment))
#define TRAP_VENEER  0x7fe00008

#elif defined(OSMIPS)
/* MIPS */
/* MAX_JUMP_IN_ONE_INSTRUCTION
 * How far away can the instruction be placed and still be able to jump back
 * to the original place.
 * The upper nibble of the instruction address is used to create a 32-bit address
 * from a 26-bit constant shifted left 2
 */
#define MAX_JUMP_IN_ONE_INSTRUCTION (0x0FFFFFFF)
#define TRAP_VENEER 0xEC000000
#define ADDR_MASK (~(0x0FFFFFFFUL))

#elif defined(OSARM)
/* ARM */
/* MAX_JUMP_IN_ONE_INSTRUCTION
 * How far away can the instruction be placed and still be able to jump back
 * to the original place.
 * +-32Mb for ARM
 */
#define MAX_JUMP_IN_ONE_INSTRUCTION (0xFFFFFFFF)
/* Execution points */
#ifdef __BIG_ENDIAN
#define TRAP_VENEER 0xE7FFDEFE /* be ARM mode */
#else
#define TRAP_VENEER 0xE7FDDEFE /* le ARM mode */
#endif

#define TRAP_VENEER_T32 0xDEDEDEDE

#else
#error ARCH constants not defined
#endif
#endif

//#define EPM_ENABLE_PRINTF
#ifdef EPM_ENABLE_PRINTF
#include "stdio.h"
uint32_t verbose = 0;
#define ERRLOG(a) if(verbose) printf a
#define DLOG(a) ((void) 0)
#define DLOG2(a) if(verbose) printf a
uint32_t dlog_counter = 0;
#else
#define DLOG(a) ((void) 0)
#if 1
#define DLOG2(a) ((void) 0)
#else
#define DLOG2(a) ramlog_printf a
#endif
#define ERRLOG(a) ramlog_printf a
#endif

extern void *zzcur_p;

/* For breakpoints in an mmaped region.
 * If set, prepare_for_page_fault_in_disable_region shall point to a function
 * that checks if a page fault can be allowed although it occurred in a
 * disable region. If not, 0 shall be returned. Otherwise, it shall make any
 * necessary preparations and then return a nonzero value. If a restore is
 * needed, a pointer to a restore functions can be returned through the
 * argument. */
extern uint32_t (*prepare_for_page_fault_in_disable_region)(void (**restore)(void));

/*
** START OF LOCAL LIST IMPLEMENTATION
** A local implementation is needed to be able to debug programs using the
** list.h functions.
*/

typedef struct _ListLink
{
    struct _ListLink *next;	/* Ptr to first/next element in list. */
    struct _ListLink *prev;	/* Ptr to last/previous element in list. */
} ListLink;

typedef ListLink List;


void * list_head(const List *list);
#define list_head(list) (((list)->next != (list)) ? (void*)((list)->next) : NULL)

#if 0
void * list_tail(const List *list);
#define list_tail(list) (((list)->prev != (list)) ? (void*)((list)->prev) : NULL)
#endif

void * list_next(const List *list, const void *lnk);
#define list_next(list, lnk) ((((const ListLink *)(lnk))->next != (list)) ? (void*)(((const ListLink *)(lnk))->next) : NULL)

#if 0
void * list_prev(const List *list, const void *lnk);
#define list_prev(list, lnk) ((((const ListLink *)(lnk))->prev != (list)) ? (void*)(((const ListLink *)(lnk))->prev) : NULL)
#endif

#define list_init(list) epm_list_init(list)
static void
epm_list_init(List *list);

#define list_init_link(ll) epm_list_init_link(ll)
static void
epm_list_init_link(void *ll);

#define list_unlink(ll) epm_list_unlink(ll)
static void
epm_list_unlink(void *ll);

#if 0 /* Currently not used */
#define list_insert_head(list, ll) epm_list_insert_head(list, ll)
static void
epm_list_insert_head(List *list, void *ll);
#endif

#define list_insert_tail(list, ll) epm_list_insert_tail(list, ll)
static void
epm_list_insert_tail(List *list, void *ll);

#define list_traverse(list, apply, arg) epm_list_traverse(list, apply, arg)
static void
epm_list_traverse(const List *list,
                  void *(*apply) (ListLink *lnk, void *arg), void *arg);

#ifdef HANDLE_EP_IN_VENEER
#define list_find(list, ll) epm_list_find(list, ll)
static void *
epm_list_find(const List *list, void *ll);
#endif

#ifndef ASSERT
#define list_assert(l,f)
#else
#define list_assert(list, check) epm_list_assert(list, check)

static void
epm_list_assert(const List *list, void (*check) (const ListLink *))
{
    const ListLink *prev;
    const ListLink *next;
    const ListLink *link;
    assert(list != NULL);
    /* Check list head. */
    assert(list->next != NULL);
    assert(list->prev != NULL);
    /* Backward pass. */
    link = list->prev;
    next = (const ListLink *) list;
    while (link != (const ListLink *) list)
    {
	assert(link->next == next);
	next = link;
	link = link->prev;
    }
    /* Forward pass, use user supplied function to check nodes. */
    link = link->next;
    prev = (const ListLink *) list;
    while (link != (const ListLink *) list)
    {
	assert(link->prev == prev);
	if (check)
	{
	    check(link);
	}
	prev = link;
	link = link->next;
    }
}
#endif

static void
epm_list_init(List *list)
{
    if (list != NULL)
    {
	list->next = (ListLink *) list;
	list->prev = (ListLink *) list;
	list_assert(list, NULL);
    }
}
 

static void
epm_list_init_link(void *ll)
{
    ListLink *lnk = (ListLink *)ll;

    if (lnk != NULL)
    {
	lnk->next = NULL;
	lnk->prev = NULL;
    }
}

static void
epm_list_unlink(void *ll)
{
    ListLink *lnk = ll;

    if (lnk != NULL)
    {
	assert(lnk->next != NULL && lnk->prev != NULL);

	lnk->next->prev = lnk->prev;
	lnk->prev->next = lnk->next;
	lnk->next = NULL;
	lnk->prev = NULL;
    }
}

#if 0				/* Currently unused. */
static void
epm_list_insert_head(List *list, void *ll)
{
    ListLink *link_to_insert = ll;

    list_assert(list, NULL);
    assert(link_to_insert != NULL);

    link_to_insert->next = list->next;
    link_to_insert->prev = list;
    list->next->prev = link_to_insert;
    list->next = link_to_insert;
}
#endif

static void
epm_list_insert_tail(List *list, void *ll)
{
    ListLink *link_to_insert = ll;

    list_assert(list, NULL);
    assert(link_to_insert != NULL);

    link_to_insert->next = list;
    link_to_insert->prev = list->prev;
    list->prev->next = link_to_insert;
    list->prev = link_to_insert;
}


static void
epm_list_traverse(const List *list, 
                  void *(*apply) (ListLink *lnk, void *arg), 
                  void *arg)
{
    ListLink *current;

    list_assert(list, NULL);

    current = (ListLink *)list_head(list);
    while (current != NULL)
    {
	ListLink *next = (ListLink *)list_next(list, current);
	void *stp = apply(current, arg);

	if (stp != NULL)
	{
	    return;
	}
	current = next;
    }
}

#ifdef HANDLE_EP_IN_VENEER
/*
 * Find ll in list, return ll if found and NULL if not found 
 */
static void *
epm_list_find(const List *list, void *ll)
{
   ListLink *current;
   list_assert(list, NULL);

   current = (ListLink *)list_head(list);
   while (current != NULL)
   {
      if (current == ll)
      {
         return ll;
      }
      current = (ListLink *)list_next(list, current);
   }
   return NULL;
}
#endif

/*
** END OF LOCAL LIST IMPLEMENTATION
** A local implementation is needed to be able to debug programs using the
** list.h functions.
*/

#if defined(PAGE_SIZE_4K)
  #define PAGE_SIZE 0x1000
  #define PAGE_SIZE_BITS 12
#elif defined(PAGE_SIZE_8K)
  #define PAGE_SIZE 0x2000
  #define PAGE_SIZE_BITS 13
#else
  #error "No page size set."
#endif

#define PAGE_SIZE_MASK ((uintptr_t)~(PAGE_SIZE -1))

#define HASH_SIZE 64
#define HASH_LOOKUP(s) (((s) >> PAGE_SIZE_BITS) & (HASH_SIZE - 1))

#define EPM_MAGIC         0x65706d00
#ifdef HANDLE_EP_IN_VENEER
#define VENEER_MAGIC      0x65706d01
#define FAST_VENEER_MAGIC 0x65706d02

/* signals */
#define EPM_VENEER_SIGNAL_TRIGGER_GC      140000
#define EPM_FAST_VENEER_SIGNAL_TRIGGER_GC 140001


struct epm_gc_address
{
   SIGSELECT sigNo;
   ListLink link; /* MUST always be 2:nd after sigNo so GC can find sig start*/
   struct ep_veneer_mem *veneer_mem;
};

#ifdef FAST_EP
struct epm_gc_fast_veneer
{
   SIGSELECT sigNo;
   ListLink link; /* MUST always be 2:nd after sigNo so GC can find sig start*/
   OSADDRESS mm_region;
};
#endif

union SIGNAL
{
      SIGSELECT sigNo;
      struct mm_create_region create_region;
      struct mm_create_region create_region_reply;
      struct mm_delete_region delete_region;
      struct epm_gc_address   gc;
#ifdef FAST_EP
      struct epm_gc_fast_veneer gc_fast_veneer;
#endif
};



typedef enum veneer_active_state 
{
   EP_VENEER_STATE_FREE = 0,
   EP_VENEER_STATE_ACTIVE,
   EP_VENEER_STATE_USED,   
} veneer_active_state_t;

struct ep_veneer_fragment
{
   /* Set if a soft trace is active. */
   veneer_active_state_t active;
   /* type, ARCH specific */
   uint32_t type;
   /* original state/instruction */
   sub_state_t sub_state;
   /* jump back instructions */
   uint32_t branch_opcodes[12];
};

struct ep_veneer_mem_hdr
{
   ListLink link; /* must be first */
   uint32_t magic;
   
   /* number of active veneers in this page*/
   uint32_t number_available_veneers;

#if defined(OSMIPS)
   OSADDRESS log_addr;
#else
   OSADDRESS requested_address;
#endif
};

/* 
 * NBR_VENEERS define the number of fragments that fit in a page slot 
 * WARNING the fields need to match the "header" in struct ep_veneer_mem
 */
#define NBR_VENEERS ((PAGE_SIZE - sizeof(struct ep_veneer_mem_hdr))/\
         sizeof(struct ep_veneer_fragment))

struct ep_veneer_mem
{
   struct ep_veneer_mem_hdr hdr;  /* must be first */
   struct ep_veneer_fragment veneer[NBR_VENEERS];
};

/* Fast ep callback struct */
struct fast_ep_veneer
{
   ListLink link; /* must be first */
   uint32_t magic;
   uint32_t active;
   unsigned long long instance; /* same as the one of the ep */
   struct EpmExecutionPoint *ep;
   /* address to where to jump to exit and jump back used by callback to 
    * "emulate" instruction */ 
   OSADDRESS emulate_address; 
   size_t size;
   size_t size_callback_code;
   /* opcode used when activated */
   uint32_t fast_ep_opcode;

   /* original state/instruction */
   sub_state_t sub_state;
   
   uint32_t code[45+12]; /* premble/postemble code */
   /* variable size will be branched to to from code must be last as it grows*/
   uint32_t callback_code[1];
};


#endif /* #ifdef HANDLE_EP_IN_VENEER */

/*
 * This is the internal data struct that is part of the struct
 * EpmExecutionPoint.
 */
struct InternalData
{
   ListLink link;
   uint32_t magic;
   unsigned long long instance;

   /* sub_state is a place for set_bp & clear_bp to store internal data. The
    * actual use is to store the original opcode then setting SW breakpoints.
    */
#ifndef HANDLE_EP_IN_VENEER
   sub_state_t sub_state;
#else
   struct fast_ep_veneer *fast_ep_veneer;
   struct ep_veneer_fragment *ep_veneer;
#endif
};

union EP
{
   struct EpmExecutionPoint *ep;
   struct InternalData      *idata;
   ListLink                 *link;
};

/* Counter incremented when a new ep is added. */
static unsigned long long ep_instance = 0;

/* Flag for internal state. */
static int enable_all;

#ifdef HANDLE_EP_IN_VENEER

struct ep_veneer_mem *free_veneer_area(union EP ep);

/* 
 * veneer database 
 * filled with ep_veneer_mem that are pagesized memory regions 
 */

/* 
 * Linked list might not a the best data structure 
 * something that is fast to get nearest address is better 
 * maybe a tree of some kind but since every veneer page fits a lot of 
 * veneer's there will probably only be a few pages so maybe it's not worth 
 * the trouble.
 */
static List ep_veneer_db;

#ifdef FAST_EP

/*
 * Hash array for storing fast execution point data.
 * Each slot contains a linked list. As address the fast ep veneer is used
 * e.g. when i ep callback check exception masked with PAGE_SIZE
 * 
 * finding it from the ep is simple since there is a pointer in that struct
 * that can be used instead of the address
 * 
 */
static List fast_hash[HASH_SIZE];

/*
 * Fast ep veneer areas to removed are place in this list, they may still be
 * triggering callbacks after removed as the program counter might execute 
 * inside them when they are removed. 
 */
static List fast_ep_remove_list;

#endif

#endif /* #ifdef HANDLE_EP_IN_VENEER */

/*
 * Hash array for storing execution point data.
 * Each slot contains a linked list.
 */
static List hash[HASH_SIZE];


#ifndef HANDLE_EP_IN_VENEER

static struct {
/* Set if a soft trace is active. */
   int active;

/* Address of instruction being traced. */
   OSADDRESS address;

   /* opcode at instruction being traced. */
   sub_state_t sub_state;

   /* Internal ep used to set temporary ep when tracing. */
      struct EpmExecutionPoint tmp_ep;

/* Saved status register during trace. */
   uint32_t saved_sr;

   /* Status register before trace. */
   uint32_t sr_before_trace;
} trace = {0};
#endif /* #ifndef HANDLE_EP_IN_VENEER */

/*
** Context specific debug info. Bit mask 1 means active in ep handling of
** the current context. Other bits unused by epm.
*/
static U8 boot_debug_flags = 0;

/* Lock functions for smp. */
/* 
 *  Use LOCK_SAVE() together with SMP_SPIN_LOCK() to ensure no double lock 
 *  like if there are ep's in interrupt path
 */
#ifdef OSE_SMP
/* The epm spin lock, and the source line where it was taken */
static atomic_t epm_spin_lock = ATOMIC_INIT(0);
static int lock_at_line;

#define SMP_SPIN_LOCK()            epm_smp_spin_lock(__LINE__)
#define SMP_SPIN_UNLOCK()          epm_smp_spin_unlock()
#define SMP_START_OTHER_CPUS()               zzkrn_smp_stop_cpu(2, 0)
#define SMP_STOP_OTHER_CPUS()                zzkrn_smp_stop_cpu(1, ~0)
#define SMP_STOP_OTHER_CPUS_INVALIDATE(addr) zzkrn_smp_stop_cpu(1, addr)


static int
epm_smp_spin_lock(int src_line)
{
   int cpu_was_stopped = 0;
   uint32_t lock;
   cpuid_t cpuid;

  /* DLOG(("%d SMP_SPIN_LOCK()\n",dlog_counter++));*/ 
 Retry:
   cpuid = zzkrn_cpu_id() + 1;
   lock = atomic_cas(&epm_spin_lock, 0, cpuid);
   if (lock == 0)
   {
      /* Lock aqcuired, old value was 0. */
      lock_at_line = src_line;
      return cpu_was_stopped;
   }
   if (lock == cpuid)
   {
      /* Error, already owner of the lock. Recursion not handled. */
      ERRLOG(("EPM: spinlock (%u) called from line %d, but already taken at line %d.\n",
	      lock, src_line, lock_at_line));
      error2(EPM_INTERNAL_ERROR, __LINE__);
   }

   /* Check if my cpu has received a stop request. */
#ifdef ENABLE_INNER_STOP_INVALIDATE
   cpu_was_stopped |= zzkrn_smp_stop_cpu(0, 0);
#endif

   goto Retry;
}

static void
epm_smp_spin_unlock(void)
{
   cpuid_t cpuid = zzkrn_cpu_id() + 1;
   uint32_t lock;


   if(atomic_get_(&epm_spin_lock) == cpuid)
      lock_at_line = 0;
   lock = atomic_cas(&epm_spin_lock, cpuid, 0);

   if (lock == cpuid)
   {
      /* Lock released. */
      /* DLOG(("%d SMP_SPIN_UNLOCK()\n",dlog_counter++));*/ 
      return;
   }
   /* Error, not owner or lock free. */
   error2(EPM_INTERNAL_ERROR, __LINE__);
}

#else
#define SMP_SPIN_LOCK()
#define SMP_SPIN_UNLOCK()
#define SMP_START_OTHER_CPUS()
#define SMP_STOP_OTHER_CPUS()
#define SMP_STOP_OTHER_CPUS_INVALIDATE(addr)
#endif

/* Avoid stepping into these functions: */
#if defined(OSARM) || defined(OSMIPS) || defined(OSPP)
extern void zzsystem_callh_supervisor(void);
extern void zzbios_callh(void);
#endif
extern PROCESS zzkrn_current_process(void);
extern OSTICK zzkrn_get_ticks(void);
extern OSTIME zzkrn_system_tick(void);
extern int zzkrn_wake_up(void);
extern OSBOOLEAN zzkrn_kill_lock(PROCESS pid);
extern void zzkrn_kill_unlock(PROCESS pid);
extern void __rt_memcpy(void);

/*
** Avoid stepping into these functions:
** ose_efmi_get_current(): used in unprotected area by epm.c
** current_process(): May be used in L2 I cache invalidate needed during trace.
** Other functions included because they lead into the kernel. These are 
** likely to lead into lock regions, that would break by stepping them.
**
*/
#ifdef __ARMCC_VERSION /* RVCT */
/* remove warning for extended constant initialiser */
#pragma diag_remark 1296
#endif
static const OSADDRESS avoid_step_into_these[] =
{
   (OSADDRESS) ose_efmi_get_current,
#if defined(OSARM)
   (OSADDRESS) zzlock_save,	/* Monolithic crt */
   (OSADDRESS) zzlock_set,	/* Monolithic crt */
#endif
#if defined(OSARM) || defined(OSMIPS) || defined(OSPP)
   (OSADDRESS) zzsystem_callh_supervisor,
   (OSADDRESS) zzbios_callh,
#endif
   (OSADDRESS) zzkrn_current_process, /* Kernel */
   (OSADDRESS) zzkrn_get_ticks, /* Kernel */
   (OSADDRESS) zzkrn_system_tick, /* Kernel */
   (OSADDRESS) zzkrn_wake_up,  /* Kernel */
   (OSADDRESS) zzkrn_kill_lock,  /* Kernel */
   (OSADDRESS) zzkrn_kill_unlock,  /* Kernel */
   (OSADDRESS) current_process, /* Monolithic crt */
   (OSADDRESS) memcpy, /* clear_bp uses memcpy, so "just trace it" will hang */
#ifdef __ARMCC_VERSION
   (OSADDRESS) __rt_memcpy, /* RVCT flavor of memcpy */
#endif
    0
  };

/*
** Check an address for functions to avoid stepping into.
** Return zero if the address is considered harmless.
*/
int
zzepm_avoid_step_into(OSADDRESS address)
{
  const OSADDRESS *entrypoint;
  static int usr_calls_initiated;
  static OSADDRESS *usrcalls;
  static OSADDRESS usrcall_count;
  OSADDRESS *ucalls;
  OSADDRESS ucount;

  /* Check some known named functions */
  for(entrypoint = avoid_step_into_these; *entrypoint; ++entrypoint)
    {
      if (*entrypoint == address)
	return 1;		/* Avoid stepping */
    } /* for */
  
  /*
  ** Also check the user call array of the kernel.
  ** First, check if the the user call pointer is still uninitiated.
  */
  if (!usr_calls_initiated)
    {
      usr_calls_initiated = 1;
      /* Fetch usrcall struct pointer */
      usrcalls = (OSADDRESS *)get_sysconf(54,0,0,0);
      if (usrcalls)
	{
	  usrcall_count = *usrcalls++;
	}
    }
  ucalls = usrcalls;
  for(ucount = usrcall_count; ucount; ucount--)
    {
      if (address == (OSADDRESS)*ucalls)
	return 1;		/* Avoid stepping */
      ucalls++;
    } /* while */
  
  return 0; 			/* Harmless */
} /* zzepm_avoid_step_into */



/* Since the substate is an array, it can not be copied by dst = src. Another
 * way that is also rejected is to use memcpy. See note at the top of the file.
 */
#define CP_SUB_STATE(dst, src)\
   do {\
      U32 i;\
      for (i = 0; i < SUB_STATE_SIZE; i++) {\
         (dst)[i] = (src)[i];\
      }\
   } while (0)


static int epm_init_started   = 0;
static int epm_init_completed = 0;
static const struct zzepm_bp_conf *bp_conf = NULL;

/******************************************************************************
 * Functions for handling soft EPs.
 */

/* Interpretation of sub_state when soft EPs are used. */
struct sw_bp_state {
   uint32_t valid;
   uint32_t saved_opcode;
};

/**********/

#ifdef FAST_EP
static EpmStatus
sw_bp_setstate(OSADDRESS address, sub_state_t state_)
{
   struct sw_bp_state *state;
   EpmStatus status;
   
   /* This assert verifies that the cast in the next statement is safe. */
   assert(sizeof *state <= sizeof(sub_state_t));
   state = (struct sw_bp_state*)(void*)state_;
   
   status = zzepm_get_opcode(address, (OSADDRESS *) &state->saved_opcode);
   
   if (status == EPM_INVALID_INSTRUCTION)
   {
      /* Not possible to set ep at this opcode. */
      state->valid = 0;
      return status;
   }
   state->valid = 1;
   
   return EPM_OK;
}
#endif

static EpmStatus
sw_bp_set(OSADDRESS address, sub_state_t state_)
{
   struct sw_bp_state *state;
   EpmStatus status;
   
   /* This assert verifies that the cast in the next statement is safe. */
   assert(sizeof *state <= sizeof(sub_state_t));
   state = (struct sw_bp_state*)(void*)state_;
   
   status = zzepm_get_opcode(address, (OSADDRESS *) &state->saved_opcode);
   
   if (status == EPM_INVALID_INSTRUCTION)
   {
      /* Not possible to set ep at this opcode. */
      state->valid = 0;
      return status;
   }
   state->valid = 1;
   
   zzepm_set_sw_ep(address);
   return EPM_OK;
}

/**********/

static EpmStatus
sw_bp_clear(OSADDRESS address, sub_state_t state_)
{
   /* This cast is verified in set_sw_ep. In a normal case an ep would be set
    * before it is cleared, so this verification should be enough. */
   struct sw_bp_state *state = (struct sw_bp_state*)(void*)state_;

   if (!state->valid) {
      ERRLOG(("EPM: Attempt to restore a breakpoint without valid opcode.\n"));
      return EPM_INTERNAL_ERROR;
   }
   state->valid = 0;
   zzepm_set_opcode(address, state->saved_opcode);
   return EPM_OK;
}

/**********/

static OSADDRESS
sw_bp_get_opcode_from_state(sub_state_t state_) {
   struct sw_bp_state *state = (struct sw_bp_state*)(void*)state_;
   
   if (!state->valid) {
      ERRLOG(("EPM: Attempt to extract opcode from invalid state.\n"));
      error2(EPM_INTERNAL_ERROR, 0);
   }
   return state->saved_opcode;
}

static EpmStatus sw_bp_init(exception_handler_c eh) {
   zzepm_install_exception_handler(eh);
   return EPM_OK;
}

/******************************************************************************
 * The symbols in this section handles low level EP actions. The default values
 * can be replaced to switch to HW breakpoints. */

static const struct zzepm_bp_conf
sw_bp_conf = {
   sw_bp_set,                   /* set_bp                */
   sw_bp_clear,                 /* clear_bp              */
   sw_bp_get_opcode_from_state, /* get_opcode_from_state */
   sw_bp_init,                  /* init                  */
   NULL,                        /* dump_state            */
};

/*****************************************************************************/

/* Used as start hook. */
EpmStatus ose_epm_init_hwbp(void);
EpmStatus ose_epm_init_hwbp(void) {
   if (epm_init_started) {
      ERRLOG(("ERROR: ose_epm_set_conf() called after initiation "
               "of EPM. Must be called before.\n"));
      return EPM_INIT_FAILED;
   }

   if (zzepm_hwbp_conf == NULL) {
      ERRLOG(("HW breakpoints are not supported.\n"));
      return EPM_INIT_FAILED;
   }

   bp_conf = zzepm_hwbp_conf;

   return EPM_OK;
}

/*****************************************************************************/
#ifndef HANDLE_EP_IN_VENEER
static void dump_bp_info(void) {
   U32 i;
   ERRLOG(("EP list:\n"));
   for (i = 0; i < HASH_SIZE; i++) {
      List *list = &hash[i];
      union EP ep;
      
      ep.link = list_head(list);
      while (ep.link != NULL) {
         ERRLOG(("  0x%08X\n", ep.ep->address));

         ep.link = list_next(list, ep.link);
      }
   }

   /* This function is called if an error occurs. Better be a little extra
    * cautious. */
   if (bp_conf != NULL && bp_conf->dump_state) {
      ERRLOG(("\n"));

      bp_conf->dump_state();
   }
}
#endif
/*****************************************************************************/

#ifdef OSARM
/* Bit 0 must be removed since it is just an indicator for thumb mode */
#  define ADDRESS_MASK ((OSADDRESS)~(OSADDRESS)1)
#else
#  define ADDRESS_MASK ((OSADDRESS)~(OSADDRESS)0)
#endif

/* This macro is really only needed in a paged system. I judge, however,
 * that the introducing more conditions causes more trouble than setting
 * prepare_for_page_fault_in_disable_region a few extra times. */
#  define PREPARE_FOR_PAGE_FAULT(f) \
    (prepare_for_page_fault_in_disable_region = (f))


static List *
list_lookup(OSADDRESS address)
{
   /* Find matching list. */
   return &hash[HASH_LOOKUP(address)];
}

#ifdef FAST_EP
static List *
fast_list_lookup(OSADDRESS address)
{
   /* Find matching list. */
   return &fast_hash[HASH_LOOKUP(address)];
}

static struct fast_ep_veneer *
find_fast_veneer(const List *list, struct fast_ep_veneer * veneer_address)
{
   struct fast_ep_veneer *fast_veneer;

   fast_veneer = (struct fast_ep_veneer *) list_head(list);
   while (fast_veneer != NULL)
   {
      if (veneer_address == fast_veneer)
      {
         break;
      }
      fast_veneer = (struct fast_ep_veneer *) list_next(list, fast_veneer);
   }
   return fast_veneer;
}

static ListLink *
find_fast_ep(const List *list, OSADDRESS ep_address, uint32_t mask)
{
   union EP ep;

   ep.link = list_head(list);
   while (ep.link != NULL)
   {
      if ((ep.idata->fast_ep_veneer) && (ep_address == (ep.ep->address & mask)))
      {
         break;
      }
      ep.link = list_next(list, ep.link);
   }
   return ep.link;
}


#endif /* #ifdef FAST_EP */

/* Find the first execution point in a list that match a specified address. */
static ListLink *
find_first_ep(const List *list, OSADDRESS address, uint32_t mask)
{
   union EP ep;

   ep.link = list_head(list);
   address &= mask;
   while (ep.link != NULL)
   {
      if (address == (ep.ep->address & mask))
      {
         break;
      }
      ep.link = list_next(list, ep.link);
   }
   return ep.link;
}

/* Start a search at a specified link. */
static ListLink *
find_next_ep(const List *list, ListLink *list_link, OSADDRESS address,
             uint32_t mask)
{
   union EP ep;

   address &= mask;
   ep.link = list_link;
   for (;;)
   {
      ep.link = list_next(list, ep.link);
      if (ep.link == NULL)
      {
         return NULL;
      }

      if (address == (ep.ep->address & mask))
      {
         /* Found ep with matching address. */
         return ep.link;
      }
   }
}


#ifdef HANDLE_EP_IN_VENEER

#if defined(OSMIPS)
static uint32_t inKusegRemappable(OSADDRESS addr)
{
   /* 0x00000000 -- 0x7FFFFFFF */
   return (addr < 0x20000000) ? 1 : 0;
}

static uint32_t inKseg0(OSADDRESS addr)
{
   /* 0x80000000 -- 0x9FFFFFFF */
   return ((addr >= KSEG0_BASE) && (addr < KSEG1_BASE)) ? 1 : 0;
}

static OSADDRESS toKseg0(OSADDRESS addr)
{
   if (addr < 0x20000000) /* kuseg: 0x00000000 -- 0x7FFFFFFF */
   {
      return (OSADDRESS) (addr + KSEG0_BASE);
   }
   else if ((addr >= 0x20000000) && (addr < KSEG1_BASE)) /* kusegUnmappable + kseg0: 0x20000000 -- 0x9FFFFFFF */
   {
      return addr;
   }
   else if (addr < KSSEG_BASE) /* kseg1: 0xA0000000 -- 0xBFFFFFFF */
   {
      return (OSADDRESS) (addr - KSEG1_BASE + KSEG0_BASE);
   }
   else if (addr < KSEG3_BASE) /* ksseg: 0xC0000000 -- 0xDFFFFFFF */
   {
      return (OSADDRESS) (addr - KSSEG_BASE + KSEG0_BASE);
   }
   else /* kseg3: 0xE0000000 -- 0xFFFFFFFF */
   {
      return (OSADDRESS) (addr - KSEG3_BASE + KSEG0_BASE);
   }
}
#endif

static struct mm_create_region *
mm_create_region(size_t size,
                 OSDOMAIN domain,
                 const char *name,
                 const char *phys_mem,
                 const char *log_mem,
                 U32 perm,
                 U32 cache_mode,
                 OSADDRESS base,
                 U32 align,
                 U32 options,
                 U32 low_limit,
                 U32 phys_count,
                 const struct phys_resource *phys)
{
   const SIGSELECT sel_reg[] = {1, MM_CREATE_REGION_REPLY};
   union SIGNAL *request, *reply;
   U32 i;

   request = alloc(sizeof(struct mm_create_region) +
                   phys_count * sizeof(struct phys_resource),
                   MM_CREATE_REGION_REQUEST);

   request->create_region.size = size;
   request->create_region.domain = domain;
   strncpy(request->create_region.name, name, MM_NAME_LENGTH);

   if (phys_mem != NULL)
   {
      strncpy(request->create_region.phys_mem, phys_mem, MM_NAME_LENGTH);
   }
   else
   {
      memset(request->create_region.phys_mem, 0, MM_NAME_LENGTH);
   }
   if (log_mem != NULL)
   {
      strncpy(request->create_region.log_mem, log_mem, MM_NAME_LENGTH);
   }
   else
   {
      memset(request->create_region.log_mem, 0, MM_NAME_LENGTH);
   }
   request->create_region.permissions = perm;
   request->create_region.cache_mode = cache_mode;
   request->create_region.base = base;
   request->create_region.align = align;
   request->create_region.options = options;
   request->create_region.low_limit = low_limit;
   request->create_region.phys_count = phys_count;
   {
      const struct phys_resource *cur_phys = phys;
      for (i = 0; i < phys_count; i++)
      {
         request->create_region.phys_resources[i].base = cur_phys->base;
         request->create_region.phys_resources[i].size = cur_phys->size;
         cur_phys++;
      }
   }

   send(&request, gmm_);
   
   reply = receive(sel_reg);
  
   return (struct mm_create_region *)reply;
}

static mm_status_t
mm_delete_region(OSDOMAIN domain, OSADDRESS base)
{
   static const SIGSELECT sigsel[] = {1, MM_DELETE_REGION_REPLY};
   union SIGNAL *reply, *request;
   mm_status_t status;

   request = alloc(sizeof *request, MM_DELETE_REGION_REQUEST);

   request->delete_region.domain = domain;
   request->delete_region.base = base;

   send(&request, gmm_);
   reply = receive(sigsel);
   status = reply->delete_region.status;

   free_buf(&reply);
   return status;
}


/*
 * Get memory with execute permission near ep_address
 *
 * ep_address     The address that we want to be near.
 * size           Ammount of memory to allocate
 * dist           We want memory within this distance from ep_address
 * relaxed        True => Any memory is fine if we can't get within dist.
 *                False => Fail if we can't get memory within dist.
 *
 * Returns NULL on failure. Otherwise pointer to base of allocated mem.
 *
 * Note: The above behavior/semantics does not apply to OSMIPS.
 */
static void *
get_memory_near(OSADDRESS ep_address, size_t size, size_t dist, int relaxed)
{
   /* create_region with option MM_CREATE_REGION_SPECIFY_LOG */
   const struct phys_resource phys_resource = {0,0};
   struct mm_create_region *reply;
   void * mem;

#if defined(OSMIPS)
   (void) ep_address;
   (void) dist;

   reply = mm_create_region(size, /* size - one page*/
            0x0,  /* System Domain */
            "ep_veneer", /* name */
            "RAM", /* phys_mem */
            "RAM_SASE", /* log_mem */
            SU_RWX_USR_RWX, /* perm */
            MM_CACHE_COPY_BACK  | MM_SPECULATIVE_ACCESS, /* cache_mode */
            0, /* base */
            0, /* align */
            0, /* options*/
            0, /* low limit*/
            0, /* phys_count */
            &phys_resource); /* phys_resource * */

   if (MM_SUCCESS == reply->status)
   {
      if (((0xF0000000 & (uint32_t)reply->base) !=
           (0xF0000000 & (uint32_t)ep_address)) &&
          ((0xF0000000 & toKseg0((uint32_t)reply->base)) !=
           (0xF0000000 & (uint32_t)ep_address)))
      {
         mm_delete_region(0x0, reply->base);  /* System Domain */
         free_buf((union SIGNAL **)&reply);
         reply = mm_create_region(size, /* size - one page*/
            0x0,  /* System Domain */
            "ep_veneer", /* name */
            "RAM", /* phys_mem */
            "RAM", /* log_mem */
            SU_RWX_USR_RWX, /* perm */
            MM_CACHE_COPY_BACK  | MM_SPECULATIVE_ACCESS, /* cache_mode */
            0, /* base */
            0, /* align */
            0, /* options*/
            0, /* low limit*/
            0, /* phys_count */
            &phys_resource); /* phys_resource * */
      }
   }
#else
   reply = mm_create_region(size, /* size - one page*/
            0x0,  /* System Domain */
            "ep_veneer", /* name */
            "RAM", /* phys_mem */
            "RAM", /* log_mem */
            SU_RWX_USR_RWX, /* perm */
            MM_CACHE_COPY_BACK  | MM_SPECULATIVE_ACCESS, /* cache_mode */
            ep_address & PAGE_SIZE_MASK, /* base */
            0, /* align */
            /* options */
            MM_CREATE_REGION_SPECIFY_LOG | MM_LOGMEM_FROM_ADDRESS |
            MM_CREATE_REGION_SEARCH_FORWARD | MM_CREATE_REGION_SEARCH_REVERSE,
            0, /* low limit*/
            0, /* phys_count */
            &phys_resource); /* phys_resource * */

   /* Verify that the memory we got is close enough to ep_address. */
   if ((MM_SUCCESS == reply->status) && (!relaxed))
   {
      size_t got_dist = (reply->base >= ep_address) ?
               reply->base + size - ep_address :
               ep_address - (reply->base);
         
      if (got_dist > dist)
      {
         /* No good; throw away the memory and pretend we didn't get any */
         mm_delete_region(0x0, reply->base);  /* System Domain */
         reply->status = MM_EOUT_OF_LOG_MEM;
      }
   }
#endif
   
   if (MM_SUCCESS != reply->status && relaxed)
   {
      /* No good region found just get me anyone */

      free_buf((union SIGNAL **)&reply);
      reply = mm_create_region(size, /* size - one page*/
               0x0,  /* System Domain */
               "ep_region", /* name */
               "RAM", /* phys_mem */
               "RAM", /* log_mem */
               SU_RWX_USR_RWX, /* perm */
               MM_CACHE_COPY_BACK  | MM_SPECULATIVE_ACCESS, /* cache_mode */
               0, /* base */
               0, /* align */
               0, /* options*/
               0, /* low limit*/
               0, /* phys_count */
               &phys_resource); /* phys_resource * */
   }

   if (MM_SUCCESS == reply->status)
   {
      mem = (void *)reply->base;
      DLOG(("epm.c(%d): ALLOCATE veneer area @ 0x%08X\n",
               __LINE__, mem));
   }
   else
   {
      mem = NULL;
   }
   free_buf((union SIGNAL **)&reply);
   return mem;
}


/*
 * Both interrupt and spinlock must be locked when this function is called.
 */

static void *
traverse_veneer_find_near(ListLink *link, void *rc)
{
   union EP ep;
   OSADDRESS ep_address;
   struct ep_veneer_mem *veneer_mem;
   unsigned int i;

   ep.idata = rc;
   ep_address = ep.ep->address;
   veneer_mem = (struct ep_veneer_mem *) link;

   if (veneer_mem->hdr.number_available_veneers > 0)
   {
#ifdef OSMIPS
      if ((ADDR_MASK & (OSADDRESS)veneer_mem) == (ADDR_MASK & (OSADDRESS)ep_address))
#else
      /* there are free slots */
      if (MAX_JUMP_IN_ONE_INSTRUCTION >= 
               (uint32_t) abs((uint32_t)ep_address - (uint32_t)veneer_mem->hdr.requested_address) )
#endif
      {
         /* It's near try to use it */
         /* match check for free slots*/
         /* TODO
          * Doing this while locked is not a completely good design as
          * it touched up to NBR_VENEERS cache lines 
          */
         for (i = 0; i < NBR_VENEERS; i++)
         {
            if (veneer_mem->veneer[i].active == EP_VENEER_STATE_FREE)
            {
               /* Free slot take it */
               veneer_mem->veneer[i].active = EP_VENEER_STATE_ACTIVE;
               ep.idata->ep_veneer = &veneer_mem->veneer[i];
               ep.idata->fast_ep_veneer = NULL;
               DLOG(("epm.c(%d): alloc venner slot MATCH slot %2d used @ 0x%08x"
                     " (0x%08x)\n",
                     __LINE__, i, ep.idata->ep_veneer, veneer_mem));
               return (void*)1; /* Anything != NULL terminates */
            }
         }
      }
   }
   return NULL;
}

/*
 * Both interrupt and spinlock must be locked when this function is called.
 * This function will call system calls, during this calls the locks will be 
 * temporary released.
 */

static struct ep_veneer_mem *
alloc_ep_veneer_area(union EP ep)
{
   struct ep_veneer_mem *veneer_mem = NULL;
   OSADDRESS ep_address = ep.ep->address;
    
   ep.idata->ep_veneer = NULL;
   ep.idata->fast_ep_veneer = NULL;
   /* search available areas */

   list_traverse(&ep_veneer_db, traverse_veneer_find_near, ep.idata);

   if (ep.idata->ep_veneer != NULL)
   {
      /* found slot in available memory area */
      ep.idata->ep_veneer->active = EP_VENEER_STATE_ACTIVE;
     
      return (struct ep_veneer_mem *)
               (((uint32_t)ep.idata->ep_veneer) & PAGE_SIZE_MASK);
   }
   
   /* Unlock while using system commands 
    * It is safe to unlock since datas tructures are not changed nor relied 
    * upon until when we enter this code path
    */
   SMP_SPIN_UNLOCK();
   UNLOCK();
   
   
   /* allocate a new area */

   veneer_mem = (struct ep_veneer_mem *)
     get_memory_near(ep_address, PAGE_SIZE, MAX_JUMP_IN_ONE_INSTRUCTION, TRUE);



   if (veneer_mem)
   {
      unsigned int i;

#if defined(OSMIPS)
      veneer_mem->hdr.log_addr = (OSADDRESS) veneer_mem;
      if (inKseg0(ep_address))
      {
         /*
          *  Run veneer instructions from kseg0 too if possible
          */
         phys_addr_t pa;
         if (ose_mm_l2p(zzkrn_current_process(), (log_addr_t)veneer_mem, &pa) == MM_SUCCESS)
         {
            if (inKusegRemappable((OSADDRESS)pa))
            {
               veneer_mem = (struct ep_veneer_mem *) toKseg0((OSADDRESS)pa);
               DLOG(("epm.c(%d): memory allocated in kuseg 0x%08X pa: 0x%08X veneer_mem: 0x%08X\n",
                                       __LINE__, veneer_mem->hdr.log_addr, (OSADDRESS)pa, veneer_mem));
            }
         }
      }
#else
      /* Remember the requested address */
      veneer_mem->hdr.requested_address = ep_address;
#endif
      /* init veneer mem area */
      list_init_link(veneer_mem);
     
      veneer_mem->hdr.magic = VENEER_MAGIC;
      veneer_mem->hdr.number_available_veneers = NBR_VENEERS;

      /* Setup the first slot */
      veneer_mem->veneer[0].active = EP_VENEER_STATE_ACTIVE;
      /* Setup the rest */
      for(i = 1; i < NBR_VENEERS; i++)
      {
         veneer_mem->veneer[i].active = EP_VENEER_STATE_FREE;
      }
      
      ep.idata->ep_veneer = &veneer_mem->veneer[0];
      ep.idata->fast_ep_veneer = NULL;

      DLOG(("epm.c(%d): alloc_ep_veneer_area NEW slot 0 used @ 0x%08X (0x%08X) for 0x%08X \n",
               __LINE__, ep.idata->ep_veneer, veneer_mem, ep_address));
   }
   else
   {
      ERRLOG(("EPM: alloc_ep_veneer_area Memory ERROR\n"));
   }

   /* 
    * re-lock after using system commands but before putting item 
    * into data structure
    */
   LOCK();
   SMP_SPIN_LOCK();
   
   if (veneer_mem)
   {
      list_insert_tail(&ep_veneer_db, &(veneer_mem->hdr.link));
   } 
   return veneer_mem;
}


static void
doGC(struct ep_veneer_mem *veneer_mem)
{
   union SIGNAL *sig;

   sig = alloc(sizeof(struct epm_gc_address), EPM_VENEER_SIGNAL_TRIGGER_GC);
   sig->gc.veneer_mem = veneer_mem;

   DLOG(("epm.c(%d): Trigger gc, send signal to %d to remove 0x%08X \n",__LINE__,
                     epm_veneer_gc_process_, (uint32_t) veneer_mem));
   send(&sig, epm_veneer_gc_process_);
#ifdef ENABLE_SYNC_GC
   {
      static const SIGSELECT gc_sigsel[] = {1, EPM_VENEER_SIGNAL_TRIGGER_GC};
      sig = receive(gc_sigsel);
      free_buf(&sig);
   }
#endif
}

/* free_veneer_area must be called with interrupts disabled. */
struct ep_veneer_mem*
free_veneer_area(union EP ep)
{
   if (ep.idata->ep_veneer != NULL)
   {
      struct ep_veneer_mem *veneer_mem;
      struct ep_veneer_fragment *veneer_fragment;
      veneer_mem = (struct ep_veneer_mem *)(((uint32_t)(ep.idata->ep_veneer)) &
                                                               PAGE_SIZE_MASK);
      /* found slot in available memory area */

      veneer_fragment = ep.idata->ep_veneer;
      ep.idata->ep_veneer = NULL; /* unlink it */
      veneer_fragment->active = EP_VENEER_STATE_USED;
   
      /* For now logic is still in GC process */
      veneer_mem->hdr.number_available_veneers--;

      if (0 == veneer_mem->hdr.number_available_veneers)
      {
         list_unlink( &(veneer_mem->hdr.link));
         return veneer_mem;
      }
   }

   return NULL;
}

/* 
 * Turn used veneer fragments into free
 * if it becomes empty remove region
*
 * Both interrupt and spinlock must be locked when this function is called.
 * This function will call system calls, during this calls the locks will be 
 * temporary released.
 */
static void
gc_veneer_area(struct ep_veneer_mem *veneer_mem)
{
   /* Check if veneer is ready to be recycled
    */
   if (veneer_mem->hdr.number_available_veneers == 0)
   {
      /* Free memory veneer page area */
      DLOG(("epm.c(%d): GC: REMOVE region veneer_mem:0x%08X \n",
                        __LINE__, veneer_mem));
#if defined(OSMIPS)
         mm_delete_region(0x0, (OSADDRESS) veneer_mem->hdr.log_addr); /* System Domain */
#else
         mm_delete_region(0x0, (OSADDRESS) veneer_mem); /* System Domain */
#endif

   }
}



#ifdef FAST_EP

static struct fast_ep_veneer *
gc_find_fast_ep(const List *list, OSADDRESS address)
{
   struct fast_ep_veneer *fast_veneer;

   fast_veneer = (struct fast_ep_veneer *) list_head(list);
   while (fast_veneer != NULL)
   {
      if (address == (OSADDRESS)fast_veneer)
      {
         break;
      }
      fast_veneer = (struct fast_ep_veneer *) list_next(list, fast_veneer);
   }
   return fast_veneer;
}

static void 
gc_fast_veneer_area(OSADDRESS veneer_mem)
{
   Msr lock;
   ListLink *list_link;
   
   list_link = (ListLink *) gc_find_fast_ep(&fast_ep_remove_list, veneer_mem);

   LOCK_SAVE(lock);
   SMP_SPIN_LOCK();
   if (list_link != NULL)
   {
      list_unlink(list_link);
      SMP_SPIN_UNLOCK();
      LOCK_RESTORE(lock);
   }
   else
   {
      SMP_SPIN_UNLOCK();
      LOCK_RESTORE(lock);
      ERRLOG(("EPM veneer gc process: recived clenup signal of fast ep "
               "veneer area at 0x%08X but its not in the list.\n",
               veneer_mem));
   }
   mm_delete_region(0x0, veneer_mem);  /* System Domain */
}
#endif

static void
gc_veneer_areas(cpuid_t cpus, const List *list)
{
   union SIGNAL *sig;
   void *list_elem = NULL;
   
   unsigned int i;
   cpuid_t cpu;
   /* Ensure all cpu's has executed the venner code if it is in their code path.
    * What we really want to test is if we have idle time on all cpus
    */

   cpu = ose_cpu_id();
   for (i = 0; i < cpus; ++i)
   {
      ose_bind_process(epm_veneer_gc_process_, i);
      /* 
       * Let all other background processes run 
       * Actually we just want to put ourself's last in the bg scheduling list
       * but set_pri() are not available for bg processes. 
       */
      delay(1); 
   }
   ose_bind_process(current_process(), cpu);

   list_elem = list_head(list);
   
   while (list_elem)
   {
      list_unlink(list_elem);
      sig = (union SIGNAL *) (((uint32_t)(list_elem)) - sizeof(SIGSELECT));
      switch (sig->sigNo)
      {
         case EPM_VENEER_SIGNAL_TRIGGER_GC:
         {
            gc_veneer_area(sig->gc.veneer_mem);
            break;
         }
#ifdef FAST_EP
         case EPM_FAST_VENEER_SIGNAL_TRIGGER_GC:
         {
            gc_fast_veneer_area(sig->gc_fast_veneer.mm_region);
            break;
         }
#endif
         default:
            break;
      }
      free_buf(&sig);
      list_elem = list_head(list); /* get next element ...*/
   }   
}


static
OS_PROCESS(epm_veneer_gc_process)
{
   static const SIGSELECT any_sigsel[] = {0};
#if !defined(ENABLE_SYNC_GC)
#ifdef FAST_EP
   static const SIGSELECT gc_sigsel[] = {2,EPM_VENEER_SIGNAL_TRIGGER_GC,
                                           EPM_FAST_VENEER_SIGNAL_TRIGGER_GC};
#else
   static const SIGSELECT gc_sigsel[] = {1,EPM_VENEER_SIGNAL_TRIGGER_GC};
#endif
#endif
   static List ep_gc_list;
   union SIGNAL *sig;
   cpuid_t cpus = ose_num_cpus();
   
   list_init(&ep_gc_list);
   
   for(;;)
   {
      sig = receive(any_sigsel);
      switch (sig->sigNo)
      {
         case EPM_VENEER_SIGNAL_TRIGGER_GC:
#ifdef FAST_EP
         case EPM_FAST_VENEER_SIGNAL_TRIGGER_GC:
#endif

#if !defined(ENABLE_SYNC_GC)
            /* Put up all gc signals in gc list*/
            do {
               list_init_link(&(sig->gc.link));
               list_insert_tail(&ep_gc_list, &(sig->gc.link));
               sig = receive_w_tmo(0, gc_sigsel);
            } while (sig != NIL);

            /* wait on all CPU's cleanout GC objects and then free signals 
             * in other words "do all Da Shit!" */
            gc_veneer_areas(cpus, &ep_gc_list);

#else
            list_init_link(&(sig->gc.link));
            list_insert_tail(&ep_gc_list, &(sig->gc.link));
            {
                union SIGNAL *sig_reply;
                PROCESS s = sender(&sig);

                sig_reply = alloc(sizeof(union SIGNAL), sig->sigNo);

                gc_veneer_areas(cpus, &ep_gc_list);

                send(&sig_reply, s);
             }
#endif
            break;
         default:
            ERRLOG(("EPM veneer gc process: recived wrong signal 0x%08X\n",
                  sig->sigNo));
            free_buf(&sig);
            break;
      }
   }
}

#endif

/* By setting prepare_for_page_fault_in_disable_region to point to this
 * function, a page fault can be allowed in a disabled region. This should
 * of course only be used if the page fault really is allowed.
 */
static uint32_t handle_page_fault(void (**restore_func)(void)) {
   (void) restore_func;
   return 1;
}

/*
 * Both interrupt and spinlock must be locked when this function is called.
 * This function will call alloc_ep_veneer_area(), inside this function
 * the locks will be temporary released.
 */

static EpmStatus
add_execution_point(struct EpmExecutionPoint *ep)
{
   union EP new_ep;
   OSADDRESS address;
   List *list;
   ListLink *list_link;
   EpmStatus status;
#ifdef HANDLE_EP_IN_VENEER
   OSADDRESS curr_opcode;
   struct ep_veneer_mem *veneer;
#endif /* #ifdef HANDLE_EP_IN_VENEER */

   DLOG(("epm.c(%d): add_execution_point() \n", __LINE__));

   new_ep.ep = ep;
   address = new_ep.ep->address;

#ifdef HANDLE_EP_IN_VENEER
   /* might unlock while calling system calls */
   veneer = alloc_ep_veneer_area(new_ep);  
   if(!veneer)
   {
      ERRLOG(("EPM: Failed to alloc veneer area, return error \n", __LINE__));
      return  EPM_INTERNAL_ERROR;
   }
#endif /* #ifdef HANDLE_EP_IN_VENEER */
  
   new_ep.idata->magic = EPM_MAGIC;
   new_ep.idata->instance = ++ep_instance;
#ifndef HANDLE_EP_IN_VENEER
   ((struct sw_bp_state*)(void*)new_ep.idata->sub_state)->valid = 0;
#else
   ((struct sw_bp_state*)(void*)new_ep.idata->ep_veneer->sub_state)->valid = 0;
#endif
   list = list_lookup(address); /* Find matching list. */   
   /* check the list before we add ourself so we don't find ourselfs in it */
   list_link = find_first_ep(list, address, ADDRESS_MASK);
   /*
    * We need to add the ep to the list before we call set_bp as the trap
    * might hit directly after this.
    * 
    * Link ep block last in list.
    * We want increasing instance numbers as we traverse the list.
    */
   list_init_link(new_ep.link);
   list_insert_tail(list, new_ep.link);
   if (list_link == NULL)
   {
      /* No other ep exists at this address */
      /* Don't touch the memory if ep's are disabled. */
      if (enable_all)
      {
         /* 
          * unlock to handle if there is a breakpoint in the 
          * set_bp code path, like memcpy()
          */
#ifndef HANDLE_EP_IN_VENEER
         SMP_SPIN_UNLOCK();
         status = bp_conf->set_bp(address, new_ep.idata->sub_state);
         SMP_SPIN_LOCK();
#else
         /* Get the original opcode at the breakpoint.*/
         PREPARE_FOR_PAGE_FAULT(handle_page_fault);
         status = zzepm_get_opcode(new_ep.ep->address, &curr_opcode);
         PREPARE_FOR_PAGE_FAULT(NULL);

         if (status != EPM_OK)
         {
            ERRLOG(("EPM: Failed to read opcode at adress 0x%08X\n",
                              new_ep.ep->address));
            /*
             * This might be because we tried setting a BP at a branch delay slot
             */
            //error2(EPM_INTERNAL_ERROR, status);
            goto error;
         }

         status = zzepm_transpose_opcode(curr_opcode, address,
                               &new_ep.idata->ep_veneer->type,
                               new_ep.idata->ep_veneer->branch_opcodes,
                               sizeof(new_ep.idata->ep_veneer->branch_opcodes));

         if (status != EPM_OK)
         {
            ERRLOG(("EPM: Failed to transpose opcode at adress 0x%08X\n",
                              new_ep.ep->address));
            /* error2(EPM_INTERNAL_ERROR, status); */
            goto error;
         }

#ifdef ENABLE_INNER_STOP_INVALIDATE
         SMP_STOP_OTHER_CPUS_INVALIDATE(ep->address);
#endif
#ifdef ENABLE_MEMCPY_UNLOCK
         SMP_SPIN_UNLOCK();
#endif
         status = bp_conf->set_bp(address,
                  new_ep.idata->ep_veneer->sub_state);

#ifdef ENABLE_MEMCPY_UNLOCK
         SMP_SPIN_LOCK();
#endif

#ifdef ENABLE_INNER_STOP_INVALIDATE
         SMP_START_OTHER_CPUS();
#endif

#endif /* HANDLE_EP_IN_VENEER */
         if (status != EPM_OK)
         {
            goto error;
         }
      }
   }
   else
   {
      /* Copy state from other ep with same address. */
#ifndef HANDLE_EP_IN_VENEER 
      CP_SUB_STATE(new_ep.idata->sub_state,
                  ((struct InternalData *)list_link)->sub_state);
#else
#ifdef FAST_EP
      if (((struct InternalData *)list_link)->fast_ep_veneer)
      {
         /* since we found a fast ep we need to write the trap to be sure */
         SMP_STOP_OTHER_CPUS_INVALIDATE(address);
#ifdef ENABLE_MEMCPY_UNLOCK
         SMP_SPIN_UNLOCK();
#endif
         status = bp_conf->set_bp(address, new_ep.idata->ep_veneer->sub_state);
#ifdef ENABLE_MEMCPY_UNLOCK
         SMP_SPIN_LOCK();
#endif
         SMP_START_OTHER_CPUS();
         if (status != EPM_OK) 
         {
            goto error;
         }
         /* But we can't use the state as it's can be an active fast ep "jump" 
          * we need to get it from the fast ep meta data instead */
         CP_SUB_STATE(new_ep.idata->ep_veneer->sub_state,
                ((struct InternalData *)list_link)->fast_ep_veneer->sub_state);

      }
      else
#endif
      {
         CP_SUB_STATE(new_ep.idata->ep_veneer->sub_state,
                     ((struct InternalData *)list_link)->ep_veneer->sub_state);
      }
      /* Get the original opcode at the breakpoint.*/
      if (bp_conf->get_opcode_from_state)
      {
         curr_opcode = bp_conf->get_opcode_from_state(
                  new_ep.idata->ep_veneer->sub_state);
      }
      else
      {
         ERRLOG(("EPM: get_opcode_from_state() needed for target to "
                  "support more then one ep on an adress at 0x%08X\n",
                           new_ep.ep->address));
         status = EPM_INVALID_ARGUMENTS;
         goto error;
      }
#ifdef ENABLE_INNER_STOP_INVALIDATE
      SMP_STOP_OTHER_CPUS_INVALIDATE(ep->address);
#endif

      status = zzepm_transpose_opcode(curr_opcode, address,
               &new_ep.idata->ep_veneer->type,
               new_ep.idata->ep_veneer->branch_opcodes,
               sizeof(new_ep.idata->ep_veneer->branch_opcodes));

#ifdef ENABLE_INNER_STOP_INVALIDATE
      SMP_START_OTHER_CPUS();
#endif
      if (status != EPM_OK)
      {
         ERRLOG(("EPM: Failed to transpose opcode at adress 0x%08X\n",
                           new_ep.ep->address));
         goto error;
      }
#endif /* #ifdef HANDLE_EP_IN_VENEER */ 
   }

   return  EPM_OK;
error:
   list_unlink(new_ep.link);
#ifdef HANDLE_EP_IN_VENEER 
   new_ep.idata->ep_veneer->active = EP_VENEER_STATE_FREE;
#endif /* #ifdef HANDLE_EP_IN_VENEER */ 
  return status;
}

/* remove_execution_point must be called with interrupts disabled. */
static EpmStatus
remove_execution_point(struct EpmExecutionPoint *ep,
         void **remove_this)
{
   union EP ep_to_remove;
   OSADDRESS address;
   List *list;
   ListLink *list_link;
   EpmStatus value = EPM_OK;

   ep_to_remove.ep = ep;

   address = ep_to_remove.ep->address;
   list = list_lookup(address);

   assert(ep_to_remove.idata->magic == EPM_MAGIC);

   /* Check that ep is present. */
   list_link = list_head(list);
   
   while (list_link != NULL)
   {
      if (list_link == ep_to_remove.link)
   	{
   	  /* Found ep. */
   	  break;
   	}
      list_link = list_next(list, list_link);
   }

   if (list_link == NULL)
   {
      return EPM_INVALID_ARGUMENTS; /* Error */
   }

   list_unlink(ep_to_remove.link);
   if (enable_all)
   {
      union EP list_link_ep;

#ifndef FAST_EP
      list_link_ep.link = find_first_ep(list, address, ADDRESS_MASK);
#else /* #ifndef FAST_EP */
      union EP found_fast_ep;
      
      found_fast_ep.link = NULL;
      /* Check if there was a fast ep */
      list_link_ep.link = find_first_ep(list, address, ADDRESS_MASK);
      while(list_link_ep.link)
      {
         if (list_link_ep.link && !list_link_ep.idata->fast_ep_veneer)
         {
            break;
         }
         if (list_link_ep.link && list_link_ep.idata->fast_ep_veneer && 
                  (list_link_ep.idata->fast_ep_veneer->active == 1))
         {
            if (found_fast_ep.link)
            {
               ERRLOG(("EPM:ERROR multiple fast ep's on adress 0x%08X\n",
                        address));
               return EPM_INTERNAL_ERROR;
            }
            found_fast_ep = list_link_ep;
         }
         list_link_ep.link = find_next_ep(list,list_link_ep.link,
                  address, ADDRESS_MASK);
      }
#endif /* #ifndef FAST_EP */
      if (!list_link_ep.link)
      {
         /*
          * No more eps exist,
          * lets clear the bp
          */
         list_insert_tail(list, ep_to_remove.link);
         
         /* 
          * unlock to handle if there is a breakpoint in the 
          * clear_bp codepath, like in  memcpy()
          */
#ifndef HANDLE_EP_IN_VENEER 
         SMP_SPIN_UNLOCK();
         value = bp_conf->clear_bp(address, ep_to_remove.idata->sub_state);
         SMP_SPIN_LOCK();
#else

#ifdef ENABLE_MEMCPY_UNLOCK
         SMP_SPIN_UNLOCK();
#endif
#ifdef ENABLE_INNER_STOP_INVALIDATE
         SMP_STOP_OTHER_CPUS_INVALIDATE(ep->address);
#endif
#ifdef FAST_EP
         if (found_fast_ep.link)
         {
            /* Do not put back original opcode but the fast ep JUMP opcode */
            zzepm_set_opcode(address, 
                        found_fast_ep.idata->fast_ep_veneer->fast_ep_opcode);
            value = EPM_OK;
         }
         else
#endif            
         {
            if (ep_to_remove.idata->ep_veneer != NULL)
            {  
               value = bp_conf->clear_bp(address,
                        ep_to_remove.idata->ep_veneer->sub_state);
            }
            else
            {
               value = EPM_INTERNAL_ERROR;
            }

         }
#ifdef ENABLE_INNER_STOP_INVALIDATE
         SMP_START_OTHER_CPUS();
#endif
#ifdef ENABLE_MEMCPY_UNLOCK
         SMP_SPIN_LOCK();
#endif
#endif /* #ifndef HANDLE_EP_IN_VENEER */ 
         list_unlink(ep_to_remove.link);
      }
   }

#ifdef HANDLE_EP_IN_VENEER
   {
      struct ep_veneer_mem *veneer_mem = NULL;
      veneer_mem = free_veneer_area(ep_to_remove);
      if (veneer_mem != NULL && NULL != remove_this)
      {
         *remove_this = veneer_mem;
      }
   }
#endif /* #ifdef HANDLE_EP_IN_VENEER */

   return value;
}

#ifndef HANDLE_EP_IN_VENEER 

static void restore_after_page_fault_during_trace(void) {
   EpmStatus rc;

   /* Temporary remove the breakpoint. This should not fail. */
   rc = bp_conf->clear_bp(trace.address, trace.sub_state);

   if (rc != EPM_OK) {
      ERRLOG(("Failed to temporarity remove ep. "
               "clear_bp(0x%08X) returned %d\n", trace.address, rc));
      dump_bp_info();
      error2(EPM_INTERNAL_ERROR, trace.address);
   }
   
   rc = add_execution_point(&trace.tmp_ep);
   if (rc != EPM_OK) {
      ERRLOG(("Failed to set temporary ep @ 0x%08X, "
                    "add_execution_point returned %ld\n",
               trace.tmp_ep.address, rc));
      dump_bp_info();
      error2(EPM_INTERNAL_ERROR, ((OSERRCODE)&trace));
   }
   /* Trace one instruction. */
   trace.active  = 1;
}

U32 page_fault_during_trace_cnt = 0;

static uint32_t handle_page_fault_during_trace(void (**restore_func)(void)) {
   EpmStatus rc;

   *restore_func = NULL;
   if (!trace.active) {
      return 1;
   }
   
   page_fault_during_trace_cnt++;
   /* Remove temporary ep. */
   rc = remove_execution_point(&trace.tmp_ep, NULL);
   if (rc != EPM_OK) {
      ERRLOG(("Failed to remove temporary ep @ 0x%08X, "
                    "remove_execution_point returned %ld\n",
               trace.tmp_ep.address, rc));
      dump_bp_info();
      error2(EPM_INTERNAL_ERROR, rc);
   }
   
   /* Restore the breakpoint. */
   rc = bp_conf->set_bp(trace.address, trace.sub_state);
   if (rc != EPM_OK) {
      /* A breakpoint has just been removed, so there should be at least
       * one free. */
      ERRLOG(("Failed to restore ep. "
               "set_bp(0x%08X) returned %ld\n", trace.address, rc));
      dump_bp_info();
      error2(EPM_INTERNAL_ERROR, rc);
   }
   
   trace.active = 0;
   *restore_func = restore_after_page_fault_during_trace;
   return 1;
}
#endif /*#ifndef HANDLE_EP_IN_VENEER*/ 

EpmStatus
epm_write_execution_points(OSADDRESS page_address, char *write_ptr)
{
   union EP ep;
   List *list;

   if (!enable_all)
   {
      return EPM_OK;
   }

   list = list_lookup(page_address); /* Find matching list. */
   ep.link = find_first_ep(list, page_address, PAGE_SIZE_MASK);
   while(ep.link)
   {
      OSADDRESS offset;
      OSADDRESS opcode;
      int bytes_in_opcode;
#ifndef HANDLE_EP_IN_VENEER
      struct sw_bp_state* state = (struct sw_bp_state*)(void*)ep.idata->sub_state;
#else
      struct sw_bp_state* state = (struct sw_bp_state*)(void*)ep.idata->ep_veneer->sub_state;
#endif

      if (state->valid)
      {
      offset = zzepm_get_sw_ep(ep.ep->address, &opcode, &bytes_in_opcode) &
         (PAGE_SIZE - 1);
      if (bytes_in_opcode == 4)
      {
         *((uint32_t *)(write_ptr + offset)) = opcode;
      }
      else if (bytes_in_opcode == 2)
      {
         *((uint16_t *)(write_ptr + offset)) = (uint16_t)opcode;
      }
      else
      {
         ERRLOG(("Opcode size not supported in EPM.\n"));
         return EPM_INTERNAL_ERROR;
      }
      }

      ep.link = find_next_ep(list, ep.link, page_address, PAGE_SIZE_MASK);
   }

   return EPM_OK;
}

int
epm_execution_points_on_page(OSADDRESS page_address)
{
   List *list;

   if (!enable_all)
   {
      return 0;
   }

   list = list_lookup(page_address); /* Find matching list. */
   if (find_first_ep(list, page_address, PAGE_SIZE_MASK))
   {
      return 1;
   }
   else
   {
      return 0;
   }
}

static U32
epm_ep_exception_handler(U32 exception_id, struct cpu_regdump *dump)
{
#ifdef OSE_SMP
   int cpu_was_stopped = 0;
#endif
   OSADDRESS exception_pc;
   unsigned long long instance = 0;
   List *list;
   union EP ep;
   struct EfmiCurrent current;
   unsigned char *current_debug_flags;
   U32 clear_debug_flag_at_exit;
   EpmStatus status;
#ifdef FAST_EP
   struct fast_ep_veneer *fast_veneer;
   OSADDRESS org_exception_pc;
   int fast_ep = 0;
   unsigned long long fast_instance = 0;
#endif   
   /* Interrupts are disabled according to the exception-handler concept. */
#if 1
 {
   /* struct cpu_core *core = (struct cpu_core *)dump->sect[0].address; */
      DLOG(("epm.c(%d): exception handler, exceptio_id: 0x%08X, dump->pc: 0x%08X (0x%08X)"
                 "dump->vector: 0x%08X dump->sp: 0x%08X dump->sr: 0x%08X "
                 "dump->fault_data_set: 0x%08X dump->fault_data_addr: 0x%08X "
                 "\n", __LINE__, exception_id, dump->pc, (dump->pc & 1) ? (*((uint16_t*)(dump->pc & ~1))) : (*((uint32_t*)dump->pc)),
                 dump->vector, dump->sp, dump->sr, dump->fault_data_set,
                 dump->fault_data_addr));
 }
#endif
   (void)exception_id;

#ifdef OSE_SMP
#ifndef HANDLE_EP_IN_VENEER
   if (trace.active == 0)
   {
      /* Spin lock is not released while tracing. */
      cpu_was_stopped = SMP_SPIN_LOCK();
   }
#else /* #ifndef HANDLE_EP_IN_VENEER */
   cpu_was_stopped = SMP_SPIN_LOCK();
#endif /* #ifndef HANDLE_EP_IN_VENEER */
#endif
   
   /* The PC member in the dump may not be the exact truth. Let the
   ** architecture layer analyze it.
   */
   exception_pc = zzepm_get_dump_pc(dump);

   /* DLOG(("EPM exception. pc = 0x%08X\n", exception_pc)); */
   
#ifdef FAST_EP
   /* check for fast ep callbacks first*/   
   list = fast_list_lookup(exception_pc & PAGE_SIZE_MASK);
   fast_veneer = (struct fast_ep_veneer *) 
             find_fast_veneer(list, (struct fast_ep_veneer *)  
                                              (exception_pc & PAGE_SIZE_MASK));
   if (fast_veneer)
   {

       /* We only want to detect a trap now 
       * since we now know that the exception_pc in in a fast ep veneer area 
       * it is safe to read it 
       */

#if defined(OSPP)
      /* If exception_pc is in the fast veneer area and is a trap we are a
       * fast ep
       */
      if ((*((uint32_t*)exception_pc) == TRAP_VENEER))
#else
#error Arch not supported
#endif
      {
         fast_ep = 1;
         ep.ep = fast_veneer->ep;
         org_exception_pc = ep.ep->address;
         fast_instance = fast_veneer->instance;
         goto got_ep;
      }
      else
      {
         /* 
          * TODO 
          * Report original PC instead of moved instruction ?? 
          * Can we find it ?? 
          */
         SMP_SPIN_UNLOCK();
         ERRLOG(("EPM: exception in fast veneer trapped "
                       "instruction @ 0x%08X\n",
                       (uint32_t)exception_pc));
         /*
          * Error, propagate to other exception handler.
          */
         DLOG(("EPM: exception handler Done return 0\n"));
         return 0;
      }
   }
   else
   {
      /* Also check if it's a pending remove fast ep */
      struct fast_ep_veneer *removed_fast_veneer = gc_find_fast_ep(&fast_ep_remove_list, 
                                exception_pc & PAGE_SIZE_MASK);
      if(removed_fast_veneer)
      {
         /* silently ignore */
         SMP_SPIN_UNLOCK();
         DLOG(("EPM: exception handler Done Trapped on removed fast ep did "
                  "nothing return 1\n"));
         return 1;
      }
   }
#endif
   
   list = list_lookup(exception_pc & PAGE_SIZE_MASK); /* Find matching list. */
   ep.link = find_first_ep(list, exception_pc, ADDRESS_MASK);

   if (ep.link == NULL)
   {
#ifdef OSE_SMP
      if (cpu_was_stopped)
      {
	      /* The cpu was stopped while waiting to aqcuire the lock. The current
	       * ep can have been removed by another cpu, so retry the instruction.
	       * If the ep is still there we will get a new exception.
	       */
         SMP_SPIN_UNLOCK();
         DLOG(("epm.c(%d): exception handler Done return 1\n", __LINE__));
         return 1;
      }
#endif

#ifdef HANDLE_EP_IN_VENEER 
      /* 
       * Check if trap is from veneer area, if so set PC to the value on the 
       * next address and to a fast exit...
       */
      {
         struct ep_veneer_mem *veneer_mem;
         veneer_mem = (struct ep_veneer_mem *)
                  (((uint32_t)exception_pc) & PAGE_SIZE_MASK);

         /* Check if this address is equal a veneer area         
          * If it is in linked list it's a veneer_mem fragment
          */
         if (list_find(&ep_veneer_db,veneer_mem) == veneer_mem)
         {
            exception_pc = exception_pc & ~1;
            DLOG2(("epm.c(%d): Exception in veneer (0x%08X), is trap if "
		  "0x%08X == 0x%08X @ 0x%08X \n",
		  __LINE__, dump->pc, TRAP_VENEER, 
                  *((uint32_t*)exception_pc), exception_pc));
#if defined(OSPP) || defined(OSMIPS) || defined(OSARM)
           /* If exception_pc is in the veneer area and is a trap we put it
            * there to handle long jumps the address after is the place to
            * execute from
            */
            if ((*((uint32_t*)exception_pc) == TRAP_VENEER) 
#if defined(OSARM)
                || (*((uint32_t*)exception_pc) == TRAP_VENEER_T32)
#endif
               )
#else
#error Arch not supported
#endif
            {
               /* Doubletrap code */
               zzemp_handle_veneer_trap(dump, exception_pc);

               /* PC was changed, continue at new address. */
               goto return_from_exception_veneer_return; /*lint !e801, Use of goto is deprecated */
            }
            else
            {
               /* 
                * TODO 
                * Report original PC instead of moved instruction ?? 
                * Can we find it ?? 
                */
               ERRLOG(("EPM: exception on trapped instruction  @ "
		       "0x%08X: 0x%08X (0x%08X: 0x%08X 0x%08X)\n",
		       (uint32_t)exception_pc, *((uint32_t*)exception_pc),
		       (uint32_t)(exception_pc & ~3u),
		       ((uint32_t*)(exception_pc & ~3u))[0],
		       ((uint32_t*)(exception_pc & ~3u))[1]));
	       stop(current_process());
            }
         }
      }   
#endif
      /*
       * Error, no ep found at current address, propagate to other
       * exception handler.
       */
      SMP_SPIN_UNLOCK();
      DLOG(("epm.c(%d): exception handler Done return 0\n", __LINE__));
      return 0;
   }

#ifdef HANDLE_EP_IN_VENEER

#ifdef FAST_EP
got_ep:
   if(fast_ep)
   {
      /* fast ep veneer */
      DLOG(("EPM: fast ep veneer @ 0x%08X org exception: 0x%08X trigged "
               "by callback check\n", 
            ep.idata->fast_ep_veneer,org_exception_pc));      
   }
   else
#endif
   {
#ifdef FAST_EP
      if (ep.idata->fast_ep_veneer)
      {
         DLOG(("EPM: fast veneer @ 0x%08X (act:%d)\n",
               ep.idata->fast_ep_veneer,ep.idata->fast_ep_veneer->active));
      }
      else
      {
#endif
         if (ep.idata->ep_veneer)
         {
            DLOG(("epm.c(%d): veneer @ 0x%08X (act:%d type:%d st_act:%d at_op:0x%08x "
                     "op(0x%08x):0x%08x 0x%08x 0x%08x 0x%08x 0x%08x )\n",
                     __LINE__,ep.idata->ep_veneer, ep.idata->ep_veneer->active, 
                     ep.idata->ep_veneer->type, 
                     ep.idata->ep_veneer->sub_state[0], 
                     ep.idata->ep_veneer->sub_state[1],
                     &ep.idata->ep_veneer->branch_opcodes[0],
                     ep.idata->ep_veneer->branch_opcodes[0],
                     ep.idata->ep_veneer->branch_opcodes[1],
                     ep.idata->ep_veneer->branch_opcodes[2],
                     ep.idata->ep_veneer->branch_opcodes[3],
                     ep.idata->ep_veneer->branch_opcodes[4],
                     ep.idata->ep_veneer->branch_opcodes[5]));

            if (ep.idata->ep_veneer->active != EP_VENEER_STATE_ACTIVE)
            {
               /* EP hit internal error -> bail out */
               ERRLOG(("EPM ep at 0x%08X using veneer at 0x%08X "
                        "is not in active or setup state.\n",
                        exception_pc, ep.idata->ep_veneer ));
               error2(EPM_INTERNAL_ERROR, EPM_INTERNAL_ERROR);
            }
         }
#ifdef FAST_EP
      }
      org_exception_pc = (OSADDRESS) NULL;
#endif
   }
#else
   
   if (trace.active)
   {
      EpmStatus rc;

      trace.active = 0;
      PREPARE_FOR_PAGE_FAULT(NULL);
      
      /*
      ** Decrement recursive counter since we entered the function with
      ** the lock taken. No unlock should occur.
      */
/*      SMP_SPIN_UNLOCK(); */
      
      /* Remove temporary ep. */
      rc = remove_execution_point(&trace.tmp_ep, NULL);
      if (rc != EPM_OK) {
         ERRLOG(("Exception pc: 0x%08X\n", exception_pc));
         dump_bp_info();
         error2(EPM_INTERNAL_ERROR, rc);
      }
      
      /* Restore the original ep. */
      rc = bp_conf->set_bp(trace.address, trace.sub_state);
      if (rc != EPM_OK) {
         /* A breakpoint has just been removed, so there should be at least
          * one free. */
         error2(EPM_INTERNAL_ERROR, rc);
      }
      
      /* Other cpus have been stopped during the trace. */
      SMP_START_OTHER_CPUS();

      /* Restore status register in the dump. */
      zzepm_lock_restore_dump(dump, trace.saved_sr);

      /* Rescan the list. */
      ep.link = find_first_ep(list, exception_pc, ADDRESS_MASK);

      if (ep.link == NULL)
      {
	    /* No other ep than the temporary was found, continue. */
	    SMP_SPIN_UNLOCK();
        DLOG(("epm.c(%d): exception handler Done return 1\n", __LINE__));
	    return 1;
      }
      /*
       * We have traced to an address with an existing ep that needs to
       * be handled.
       */
   }

   /* What we really want to do is to extract sr without modifications in the
    * arch. specific code. */
   trace.sr_before_trace = zzepm_lock_save_dump(dump);
   zzepm_lock_restore_dump(dump, trace.sr_before_trace);
#endif

   /* Select to use boot phase or process specific debug flags */
   current_debug_flags = (zzcur_p) ?
     zzfind_current_debug_info() : &boot_debug_flags;

   /*
   ** If we are already handling an ep for the current context, the epm or
   ** callback must have hit another ep. We must not get stuck there, we
   ** shall handle it as a non-matching ep.
   */
   if (*current_debug_flags & 1)
   {
      DLOG(("epm.c(%d): Just trace it, return \n", __LINE__));
      clear_debug_flag_at_exit = 0;
      goto just_trace_it; /*lint !e801, Use of goto is deprecated */
   }

   /* Remember that we are handling an ep for the current context. */
   *current_debug_flags |= 1;


   /* Unlock since there can be EP's in this code path */
#ifdef ENABLE_MEMCPY_UNLOCK
   SMP_SPIN_UNLOCK();
#endif
   (void) ose_efmi_get_current(&current, sizeof(current));
#ifdef ENABLE_MEMCPY_UNLOCK
   SMP_SPIN_LOCK();

#ifdef FAST_EP
   if (fast_ep)
   {
      /* Also check if it's a pending remove fast ep */
      struct fast_ep_veneer *removed_fast_veneer = gc_find_fast_ep(&fast_ep_remove_list,
                                exception_pc & PAGE_SIZE_MASK);
      if (removed_fast_veneer)
      {
         /* silently ignore */
         SMP_SPIN_UNLOCK();
         DLOG(("EPM: exception handler Done Trapped on removed fast ep did "
                  "nothing return 1\n"));
         return 1;
      }
   }
   else
#endif
   {
      list = list_lookup(exception_pc & PAGE_SIZE_MASK); /* Find matching list. */
      ep.link = find_first_ep(list, exception_pc, ADDRESS_MASK);

      if (ep.link == NULL)
      {
         SMP_SPIN_UNLOCK();
         return 1;
      }
   }
#endif /* #ifdef ENABLE_MEMCPY_UNLOCK */

   {
      for (;;)
      {
         assert(ep.idata->magic == EPM_MAGIC);
#ifdef FAST_EP
         if (ep.idata->fast_ep_veneer && ep.idata->fast_ep_veneer->active == 0)
         {
            /* Skip non active fast ep's */
            goto skip_cb;
         }
#endif
   	   /* Check if this ep already have been served. */
   	   if (instance < ep.idata->instance)
   	   {
   	      if ((ep.ep->flags & EPM_EP_ANY_PROCESS) ||
   	               ( zzcur_p && (ep.ep->pid == current.process ||
   	                        ep.ep->pid == current.block ||
   	                        ep.ep->pid == current.segment)))
   	      {
   	         /* Found ep with matching scope. */
   	         int ret;

   	         /* Save latest served instance. */
   	         instance = ep.idata->instance;
   	         SMP_SPIN_UNLOCK();

   	         /* Save non gpr registers (float). */
   	         zzepm_debug_save_regs();
#ifdef FAST_EP
   	         if (fast_ep)
   	         {
   	            /* Patch PC so it points out the trap veneer core again 
   	             * but only if the callback didn't change it */
   	            struct cpu_core *core = (struct cpu_core *) dump->sect[0].address;
   	            core->pc = org_exception_pc;
   	            dump->pc = core->pc;
 	         }
#endif   	         
               /* Notify the client, call the callback function. */
               ret = (ep.ep->cb)(ep.ep, dump, &current);
#ifdef EPM_ENABLE_PRINTF
               if(ret == 1)
               {
                  verbose = 1;
               }
               else
#endif
               {
                  /* Return value could be used in future implementations. */
                  assert(ret == 0);
                  (void) ret;
               }

#ifdef FAST_EP

               if (fast_ep && org_exception_pc == zzepm_get_dump_pc(dump))
               {
                  /* restore PC  */
                  struct cpu_core *core = (struct cpu_core *) dump->sect[0].address;
                  core->pc = exception_pc;
                  dump->pc = core->pc;
               }
#endif

               /* Restore non gpr registers (float). */
               zzepm_debug_restore_regs();

               /* Lock in case of a client has unlocked. */
               LOCK();
               SMP_SPIN_LOCK();

               /*
                * Do a new search, the callback can have removed or
                * added ep's.
                */
#ifdef FAST_EP
               /* check for fast ep callbacks first */
               if (fast_ep)
               {
                  goto only_one_fast_ep;
               }
               else
#endif
               {
                  ep.link = find_first_ep(list, exception_pc, ADDRESS_MASK);
               }
   	         if (ep.link == NULL)
   	         {
   	            break; /* End of list. */
   	         }
   	         continue;
   	      }
   	   }
#ifdef FAST_EP
	   skip_cb:
   	   /* check for fast ep callbacks first */
   	   if (fast_ep)
   	   {
   	      goto only_one_fast_ep;
   	   }
#endif
   	   ep.link = find_next_ep(list, ep.link, exception_pc, ADDRESS_MASK);
   	   if (ep.link == NULL)
   	   {
   	      /* No more matching ep's. */
   	      break;
   	   }
      }
   }
#ifdef FAST_EP
only_one_fast_ep:
#endif
   /*
    * We are handling a non-recursive ep for this process. We must remember
    * to clear the debug flag at return from the exception. We want to do
    * that as late as possible, because we want to keep the protection it
    * gives for recursive exceptions.
    */
   clear_debug_flag_at_exit = 1;

   /* Check if a client changed the pc. */
   if (exception_pc != zzepm_get_dump_pc(dump))
   {
      /* PC was changed, continue at new address. */
      DLOG(("epm.c(%d): Client changed PC, return \n", __LINE__));
      goto return_from_exception; /*lint !e801, Use of goto is deprecated */
   }
#ifdef FAST_EP
   /* check for fast ep callbacks first */
   if (fast_ep)
   {
      list = fast_list_lookup(exception_pc & PAGE_SIZE_MASK);
      fast_veneer = (struct fast_ep_veneer *) 
                find_fast_veneer(list, (struct fast_ep_veneer *) 
                                          (exception_pc & PAGE_SIZE_MASK));
      /* Exists? and is it the same instance?? */
      if ( (fast_veneer == NULL) || (fast_instance != fast_veneer->instance))
      {
         /* No or not the same ep exists any more at current address */
         struct cpu_core *core = (struct cpu_core *) dump->sect[0].address;
         core->pc = org_exception_pc;
         dump->pc = core->pc;
         goto return_from_exception; /*lint !e801, Use of goto is deprecated */         
      }
   }
   else
#endif
   {
      ep.link = find_first_ep(list, exception_pc, ADDRESS_MASK);
      if (ep.link == NULL)
      {
         /* No ep exists any more at current address, continue. */
         DLOG(("epm.c(%d): EP is gone, return \n", __LINE__));
         goto return_from_exception; /*lint !e801, Use of goto is deprecated */
      }
   }

 just_trace_it:
#ifdef HANDLE_EP_IN_VENEER 
#ifdef FAST_EP
   if (fast_ep)
   {
      /* fast ep from internal trap 
       * just ignore the TRAP there is code to return after it */
      struct cpu_core *core = (struct cpu_core *) dump->sect[0].address;
      core->pc = core->pc + 4;
      dump->pc = core->pc;
      goto return_from_exception;
   }
   if (ep.idata->fast_ep_veneer)
   {
      /* fast ep from normal ep trap (dig:ed out from database) 
       * just use the emulate_address to exit */
      struct cpu_core *core = (struct cpu_core *) dump->sect[0].address;
      core->pc = ep.idata->fast_ep_veneer->emulate_address;
      dump->pc = core->pc;
      goto return_from_exception;
   }
   else
#endif
   {
      /* normal ep */
      status = zzepm_handle_transposed_opcode(dump, ep.idata->ep_veneer->type, 
                          ep.idata->ep_veneer->branch_opcodes);
      if (status != EPM_OK)
      {
         ERRLOG(("EPM: Failed to handle transposed opcode at adress 0x%08X\n",
                          ep.ep->address));
         error2(EPM_INTERNAL_ERROR, status);
      }
   }
#endif
#ifndef HANDLE_EP_IN_VENEER 
   {
      EpmStatus rc;
      OSADDRESS curr_opcode;

      /* Get the original opcode at the breakpoint. */
      if (bp_conf->get_opcode_from_state) {
         curr_opcode = bp_conf->get_opcode_from_state(ep.idata->sub_state);
      }
      else {
         if (IRQ_ENABLED(trace.sr_before_trace)) 
         {
            /* Interrupts were enabled when the breakpoint was hit => It should
             * be ok to take a page fault here.
             */
            PREPARE_FOR_PAGE_FAULT(handle_page_fault);
         }
         status = zzepm_get_opcode(ep.ep->address, &curr_opcode);
         PREPARE_FOR_PAGE_FAULT(NULL);
         
         if (status != EPM_OK) {
            ERRLOG(("Failed to read opcode at adress 0x%08X\n",
                              ep.ep->address));
            error2(EPM_INTERNAL_ERROR, status);
         }
      }

      /*
       * Some instructions can not be executed with interrupts disabled, such
       * as reading and writing to the status register. These instructions has
       * to be emulated by software. Call function to check if emulation is
       * needed and emulate the instruction if required.
       */
      if (zzepm_emulate_instruction(dump, curr_opcode))
      {
         /* The instruction has been emulated, continue with next instruction.
          */
         DLOG(("epm.c(%d): Emulate instruction ok, return \n", __LINE__));
         goto return_from_exception; /*lint !e801, Use of goto is deprecated */
      }

      
   /*
    * A "soft" trace is required:
    *   Remove the original ep.
    *   Insert a temporary ep at next address.
    *   Execute one instruction and a new exception is handled here.
    *   Restore the original ep.
    *   Remove the temporary ep.
    */

   assert(ep.idata->magic == EPM_MAGIC);

   /*
    * Stop other cpus during trace. Otherwise cpus can execute current
    * address without hitting the bp. Also we don't want other cpus to hit
    * the temporary bp.
    */
   SMP_STOP_OTHER_CPUS();

   /* Temporary remove the breakpoint. This should not fail. */
   CP_SUB_STATE(trace.sub_state, ep.idata->sub_state);
   {
      rc = bp_conf->clear_bp(ep.ep->address, trace.sub_state);
      if (rc != EPM_OK) {
         ERRLOG(("Failed to remove temporary ep @ 0x%08X, "
		       "clear_bp returned %d\n", ep.ep->address, rc));
	 ERRLOG(("Exception pc: 0x%08X\n", exception_pc));
	 dump_bp_info();
         error2(EPM_INTERNAL_ERROR, ep.ep->address);
      }
   }

   /* Add temporary ep. */
      trace.tmp_ep.address = zzepm_get_trace_to_pc(dump, curr_opcode);

      /* Some functions should not be stepped into, check this! */
      if (zzepm_avoid_step_into(trace.tmp_ep.address))
         trace.tmp_ep.address = zzepm_get_next_pc(dump, EPM_SLIDE_OVER_FUNCTION_CALLS,
               curr_opcode);

      rc = add_execution_point(&trace.tmp_ep);
      if (EPM_OK == rc)
      {
         /* Trace one instruction. */
         trace.address = ep.ep->address;
         trace.saved_sr = zzepm_lock_save_dump(dump);

         if (IRQ_ENABLED(trace.sr_before_trace)) {
            PREPARE_FOR_PAGE_FAULT(handle_page_fault_during_trace);
         }

         trace.active = 1;
      }
      else
      {
         ERRLOG(("Failed to set temporary ep @ 0x%08X, "
                           "add_execution_point returned %d\n",
                           trace.tmp_ep.address, rc));
         ERRLOG(("Exception pc: 0x%08X\n", exception_pc));
         dump_bp_info();
         /* Failed to set tracepoint. Just return. */
      }
   }
#endif /* #ifndef HANDLE_EP_IN_VENEER */ 

 return_from_exception:

   /* Remember that we are NOT handling an ep for the current context. */
   if (clear_debug_flag_at_exit)
   {
     *current_debug_flags &= (unsigned char)~1;
   }

#ifndef HANDLE_EP_IN_VENEER 
#ifdef OSE_SMP
   /* Keep lock when tracing. */
   if (trace.active == 0)
   {
      SMP_SPIN_UNLOCK();
   }
#endif  
#else /* #ifndef HANDLE_EP_IN_VENEER */
return_from_exception_veneer_return:
#ifdef OSE_SMP
   SMP_SPIN_UNLOCK();
#endif
#endif /* #ifndef HANDLE_EP_IN_VENEER */

   DLOG(("epm.c(%d): exception handler end of function, Done return 1\n", __LINE__));
   return 1;
}

static U32
bp_conf_ok(const struct zzepm_bp_conf *conf) {
   if (conf == NULL) {
      ERRLOG(("HW breakpoints not supported.\n"));
      return 0;
   }
   if (conf->set_bp == NULL) {
      ERRLOG(("Mandatory breakpoint function set_ep is invalid.\n"));
      return 0;
   }
   if (conf->clear_bp == NULL) {
      ERRLOG(("Mandatory breakpoint function clear_ep is invalid.\n"));

      return 0;
   }

   return 1;
}

EpmStatus
ose_epm_init(void)
{
   int i;
   Msr lock;
#ifdef HANDLE_EP_IN_VENEER
   int ret;
   /* cpuid_t cpus = ose_num_cpus(); */
#endif
   DLOG(("epm.c(%d): ose_epm_init()\n", __LINE__));
   /* Make sure that initiation is performed only once. */
   LOCK_SAVE(lock);
   SMP_SPIN_LOCK();
   if (epm_init_started) {
      EpmStatus rc = epm_init_completed ? EPM_OK : EPM_INIT_FAILED;
      
      SMP_SPIN_UNLOCK();
      LOCK_RESTORE(lock);
      ERRLOG(("EPM: ose_epm_init() ERROR Init allready started return %d\n", rc));
      return rc;
   }
   epm_init_started = 1;
   SMP_SPIN_UNLOCK();
   LOCK_RESTORE(lock);

   /* Check space for struct InternalData in public struct. */
   assert(sizeof(struct InternalData) <=
	  sizeof((*(struct EpmExecutionPoint *) NULL).internal_data));

   enable_all = 1;

   /* Init lists. */
   for (i = 0; i < HASH_SIZE; i++)
   {
      list_init(&hash[i]);
   }
#ifdef HANDLE_EP_IN_VENEER

#ifdef FAST_EP
   for (i = 0; i < HASH_SIZE; i++)
   {
      list_init(&fast_hash[i]);
   }
   list_init(&fast_ep_remove_list);
#endif /* FAST_EP */

   ret = hunt("ose_memory_mgr", 0 , &gmm_, NULL);
   
   if (ret == 0 || gmm_ == 0)
   {
      ERRLOG(("EPM init failed. Can't hunt ose_memory_mgr process\n"));
      ERRLOG(("EPM: ose_epm_init() ERROR Hunt failed return %d\n",
            EPM_INIT_FAILED));
      return EPM_INIT_FAILED;      
   } 

   list_init(&ep_veneer_db);

   /* Setup veneer fragment gc process, someone might execute from 
    * the veneer area when remove ep is called.
    * To ensure that this does not happen the veneer fragment is marked "used"
    * and then switchback to "free" in this process later.
    */

   epm_veneer_gc_process_ = create_process(OS_BG_PROC,
                                "ose_ep_veneer_gc",
                                epm_veneer_gc_process,
                                VENEER_GC_STACK, /*stack */
                                0, /* pri */
                                (OSTIME)0, /* timeslice */
                                0x10001, /* block id 0x10001=OSE block */
                                (struct OS_redir_entry *) NULL,
                                (OSVECTOR)0,
                                (OSUSER)0);
   start(epm_veneer_gc_process_);
   
#else  
   /* Init internal ep. */
   trace.tmp_ep.flags = EPM_EP_ANY_PROCESS;
   trace.tmp_ep.cb    = NULL;
#endif
   
   {
      
      /* The breakpoint type can be set in 3 ways (listed in decreasing
       * priority):
       *
       * 1)  The configuration parameter krn/bp_type.
       * 2)  Calling ose_epm_init_hwbp (e.g. as a start hook)
       * 3)  Using the default value.
       *
       * The configuration parameter overrides a call to ose_epm_init_hwbp. The
       * reason is:
       *  - A configuration parameter can be read by a debugger and that is the
       *    way the debugger extracts information about bp type.
       *  - Configuration by calling ose_epm_init_hwbp is only kept for
       *    backwards compatibility.
       */
      
      char param[SYSPARAM_MAX_VALUE_LENGTH];
      
      if (0 < sysparam_read(BP_CONF_PARAM_NAME, param, sizeof param)) {

         /* The parameter was found. */

         if (!strcmp(param, BP_CONF_HW)) {
            
            /* The value is BP_CONF_HW. */
            
            if (!bp_conf_ok(zzepm_hwbp_conf)) {
               return EPM_INIT_FAILED;
            }
            bp_conf = zzepm_hwbp_conf;
            ERRLOG(("HW breakpoints selected.\n"));
         }
         else if (!strcmp(param, BP_CONF_SW)) {
            
            /* The value is BP_CONF_SW. */
            bp_conf = &sw_bp_conf;
            ERRLOG(("SW breakpoints selected.\n"));
         }
         else {

            ERRLOG(("EPM init Unrecognised value (%s) for configuration "
                     "parameter %s.\n", param, BP_CONF_PARAM_NAME));
            ERRLOG(("EPM: ose_epm_init() ERROR config parameter return %d\n",
                              EPM_INIT_FAILED));
            return EPM_INIT_FAILED;
         }
      }
   }

   if (bp_conf == NULL) {
      
      /* No error and HW breakpoint not selected => Use SW breakpoints. */
      bp_conf = &sw_bp_conf;
      DLOG(("epm.c(%d): SW breakpoints selected. (default)\n", __LINE__));
   }
   
   {
      EpmStatus rc = EPM_OK;
      if (bp_conf && bp_conf->init)
      {
         rc = bp_conf->init(epm_ep_exception_handler);
      }
      epm_init_completed = 1;

      DLOG(("epm.c(%d): ose_epm_init() Done return %d\n", __LINE__, rc));
      return rc;
   }
}

EpmStatus
ose_epm_add_execution_point(struct EpmExecutionPoint *ep)
{
   Msr lock;
   EpmStatus rc;

   /* Make sure that EPM is initialized. */
   if (!epm_init_completed)
   {
      rc = ose_epm_init();
      if (rc != EPM_OK)
      return rc;
   }

   DLOG(("epm.c(%d): ose_epm_add_execution_point() Add ep @ 0x%08X pid: 0x%08X from: %08X \n",
                     __LINE__, ep->address, ep->pid, &rc));

   LOCK_SAVE(lock);
#ifndef ENABLE_INNER_STOP_INVALIDATE
   SMP_STOP_OTHER_CPUS_INVALIDATE(ep->address);
#endif
   SMP_SPIN_LOCK();

   rc = add_execution_point(ep);
   SMP_SPIN_UNLOCK();
#ifndef ENABLE_INNER_STOP_INVALIDATE
   SMP_START_OTHER_CPUS();
#endif
   LOCK_RESTORE(lock);

   DLOG(("epm.c(%d): ose_epm_add_execution_point() Done return %d\n", __LINE__, rc));
   return rc;
}

EpmStatus
ose_epm_remove_execution_point(struct EpmExecutionPoint *ep)
{
   EpmStatus rc;
   Msr lock;
#ifdef HANDLE_EP_IN_VENEER
   struct ep_veneer_mem *remove_mem = NULL;
#endif

   DLOG(("epm.c(%d): ose_epm_remove_execution_point() Remove ep @ 0x%08X pid: 0x%08X.\n",
                     __LINE__, ep->address, ep->pid));

   LOCK_SAVE(lock);
#ifndef ENABLE_INNER_STOP_INVALIDATE
   SMP_STOP_OTHER_CPUS_INVALIDATE(ep->address);
#endif
   SMP_SPIN_LOCK();

#ifdef HANDLE_EP_IN_VENEER
   rc = remove_execution_point(ep, (void**) &remove_mem);
#else
   rc = remove_execution_point(ep, NULL);
#endif

   SMP_SPIN_UNLOCK();

#ifndef ENABLE_INNER_STOP_INVALIDATE
   SMP_START_OTHER_CPUS();
#endif
   LOCK_RESTORE(lock);

   DLOG(("epm.c(%d): ose_epm_remove_execution_point() Done return %d\n",
                     __LINE__, rc));
#ifdef HANDLE_EP_IN_VENEER
   if (NULL != remove_mem)
   {
      doGC(remove_mem);
   }
#endif /* #ifdef HANDLE_EP_IN_VENEER */
   return rc;
}

EpmStatus
ose_epm_next_instruction(struct cpu_regdump *dump, U32 options,
			 OSADDRESS *next_address)
{
   EpmStatus status;
   OSADDRESS address = zzepm_get_dump_pc(dump);
   List *list;
   union EP ep;
   OSADDRESS curr_opcode;
   U32 curr_opcode_valid = 0;
   Msr lock;
   OSADDRESS next;
#ifdef FAST_EP
   struct fast_ep_veneer *fast_veneer;
#endif
   DLOG(("epm.c(%d): ose_epm_next_instruction() options %d address 0x%08X\n",
            __LINE__, options, address));

   LOCK_SAVE(lock);
   SMP_SPIN_LOCK();
#ifdef FAST_EP
   list = list_lookup((OSADDRESS) address);
   ep.link = find_fast_ep(list, address, ADDRESS_MASK);

   if (ep.link)
   {
      /* Get the original opcode at the breakpoint.*/
      if (bp_conf->get_opcode_from_state)
      {
         curr_opcode = bp_conf->get_opcode_from_state(
                  ep.idata->fast_ep_veneer->sub_state);
      }
      else 
      {
         SMP_SPIN_UNLOCK();
         LOCK_RESTORE(lock);
         ERRLOG(("EPM: get_opcode_from_state() needed for target to "
                  "support more then one ep on an adress at 0x%08X\n",
                  ep.ep->address));
         status = EPM_INVALID_ARGUMENTS;
         return status;
      }
      curr_opcode_valid = 1;
   }
    

   if (!curr_opcode_valid)
   {
      /* Check normal breakpoints */
#endif
      list = list_lookup(address);

      if (bp_conf->get_opcode_from_state != NULL)
      {
         ep.link = find_first_ep(list, address, ADDRESS_MASK);
         if (ep.link != NULL)
         {
#ifndef HANDLE_EP_IN_VENEER 
            curr_opcode = bp_conf->get_opcode_from_state(ep.idata->sub_state);
#else
            curr_opcode = bp_conf->get_opcode_from_state(
                     ep.idata->ep_veneer->sub_state);
#endif /* #ifndef HANDLE_EP_IN_VENEER */ 
            curr_opcode_valid = 1;
         }
      }
#ifdef FAST_EP
   }
#endif

   if (!curr_opcode_valid) 
   {
      /* No ep exists at current address, read instruction from memory. */
      /* Allow page faults in zzepm_get_opcode */
      PREPARE_FOR_PAGE_FAULT(handle_page_fault);
      status = zzepm_get_opcode(address, &curr_opcode);
      PREPARE_FOR_PAGE_FAULT(NULL);

      if (status != EPM_OK)
      {
         /* Not possible to read memory at current address. */
         SMP_SPIN_UNLOCK();
         LOCK_RESTORE(lock);
         ERRLOG(("EPM: ose_epm_next_instruction() ERROR return %d\n", status));
         return status;
      }
   }

   next = zzepm_get_next_pc(dump, options, curr_opcode);

   /* Some functions should not be stepped into, check this! */
   if (zzepm_avoid_step_into(next))
   {
      next = zzepm_get_next_pc(dump, options | EPM_SLIDE_OVER_FUNCTION_CALLS,
			   curr_opcode);
   }
#ifdef FAST_EP
   /* 
    * Avoid reporting a PC in a fast veneer area, this could happen if 
    * a blr is used and the link register points to the veneer area 
    */     
   
   list = fast_list_lookup((OSADDRESS) next & PAGE_SIZE_MASK);
   fast_veneer = find_fast_veneer(list, (struct fast_ep_veneer *) 
                                                 (next & PAGE_SIZE_MASK));
   while (fast_veneer != NULL)
   {
      OSADDRESS next_pc;
      /* Inside veneer code area pick next instruction */
      PREPARE_FOR_PAGE_FAULT(handle_page_fault);
      status = zzepm_get_opcode(next, &curr_opcode);
      PREPARE_FOR_PAGE_FAULT(NULL);
      if (status != EPM_OK)
      {
         SMP_SPIN_UNLOCK();
         LOCK_RESTORE(lock);
         /* Not possible to read memory at current address. */
         ERRLOG(("EPM: ose_epm_next_instruction() Cant read fast ep memory area "
                  "ERROR return %d\n", status));
         error2(EPM_INTERNAL_ERROR, status);
         return status;
      }
#if defined(OSPP)
      (void) zzpreprocess_trace(curr_opcode, next, &next_pc);
      next = next_pc;
#else
#error ARCH not supported
#endif
      list = fast_list_lookup((OSADDRESS) next & PAGE_SIZE_MASK);
      fast_veneer = find_fast_veneer(list, (struct fast_ep_veneer *) 
                                     (next & PAGE_SIZE_MASK));
   } /* while */
#endif

   SMP_SPIN_UNLOCK();
   LOCK_RESTORE(lock);

   *next_address = next;

   DLOG(("epm.c(%d): ose_epm_next_instruction() next_address: 0x%08X instr: 0x%08X Done return %d\n",
        __LINE__, *next_address, (next & 1) ? (*((uint16_t*)(next & ~1))) : (*((uint32_t*)next)), EPM_OK));

   return EPM_OK;
}

static void *
traverse_disable_all(ListLink *link, void *rc)
{
   union EP ep;
   EpmStatus status;

   ep.link = link;
   SMP_STOP_OTHER_CPUS_INVALIDATE(ep.ep->address);
#ifndef HANDLE_EP_IN_VENEER
   status = bp_conf->clear_bp(ep.ep->address, ep.idata->sub_state);

#else
   status = bp_conf->clear_bp(ep.ep->address, ep.idata->ep_veneer->sub_state);
#endif /* #ifndef HANDLE_EP_IN_VENEER */
   SMP_START_OTHER_CPUS();

   if (status == EPM_OK) {
      return NULL;
   }
   *(EpmStatus*)rc = status;
   return (void*)1; /* Anything != NULL terminates */
}

EpmStatus
ose_epm_disable_all(void)
{
   Msr lock;
   EpmStatus rc = EPM_OK;

   DLOG(("epm.c(%d): ose_epm_disable_all()\n", __LINE__));

   LOCK_SAVE(lock);
   SMP_SPIN_LOCK();

   if (enable_all)
   {
      int i;

      enable_all = 0;

      /* Traverse all lists and remove all ep's from memory. */
      for (i = 0; i < HASH_SIZE; i++)
      {
	 list_traverse(&hash[i], traverse_disable_all, &rc);
      }
   }

   SMP_SPIN_UNLOCK();
   LOCK_RESTORE(lock);
   DLOG(("epm.c(%d): ose_epm_disable_all() Done\n", __LINE__));
   return rc;
}

static void *
traverse_enable_all(ListLink *link, void *rc)
{
   union EP ep;
   EpmStatus status;

   ep.link = link;
   SMP_STOP_OTHER_CPUS_INVALIDATE(ep.ep->address);
#ifndef HANDLE_EP_IN_VENEER
   status = bp_conf->set_bp(ep.ep->address, ep.idata->sub_state);
#else
   status = bp_conf->set_bp(ep.ep->address, ep.idata->ep_veneer->sub_state);
#endif /* #ifndef HANDLE_EP_IN_VENEER */ 
   SMP_START_OTHER_CPUS();

   if (status == EPM_OK) {
      return NULL;
   }
   *(EpmStatus*)rc = status;
   return (void*)1; /* Anything != NULL terminates */
}

EpmStatus
ose_epm_enable_all(void)
{
   Msr lock;
   EpmStatus rc = EPM_OK;

   DLOG(("epm.c(%d): ose_epm_enable_all()\n", __LINE__));

   LOCK_SAVE(lock);
   SMP_SPIN_LOCK();

   if (enable_all == 0)
   {
      int i;

      enable_all = 1;

      /* Traverse all lists and restore ep's. */
      for (i = 0; i < HASH_SIZE; i++)
      {
	 list_traverse(&hash[i], traverse_enable_all, &rc);
      }
   }

   SMP_SPIN_UNLOCK();
   LOCK_RESTORE(lock);
   DLOG(("epm.c(%d): ose_epm_enable_all() Done\n", __LINE__));
   return rc;
}

/* Returns true if an ep exists on the supplied address.
 * The saved opcode is returned.
 * Called from arch function.
 */
int
epm_get_opcode_from_list(OSADDRESS address, U32 *opcode)
{
   List *list;
   union EP ep;
   
   DLOG(("epm.c(%d): get_opcode_from_list(address: 0x%08X, *viopcode)\n",
            __LINE__, address));

   list = list_lookup(address); /* Find matching list. */
   ep.link = find_first_ep(list, address, ADDRESS_MASK);

   if (ep.link == NULL)
   {
      DLOG(("epm.c(%d): get_opcode_from_list(address: 0x%08X, *opcode) Done Return 0\n",
                        __LINE__, address));
     return 0;
   }
   if (bp_conf->get_opcode_from_state == NULL)
   {
      DLOG(("epm.c(%d): get_opcode_from_list(address: 0x%08X, *opcode) Done Return 0\n",
                        __LINE__, address));
      return 0;
   }
#ifndef HANDLE_EP_IN_VENEER 
   *opcode = bp_conf->get_opcode_from_state(ep.idata->sub_state);
#else
   *opcode = bp_conf->get_opcode_from_state(ep.idata->ep_veneer->sub_state);
#endif /* #ifndef HANDLE_EP_IN_VENEER */ 
   
   DLOG(("epm.c(%d): get_opcode_from_list(address: 0x%08X, *opcode=0x%08X)"
         "Done Return 1\n", 
         __LINE__, address,*opcode));
   return 1;
}

void *
ose_epm_alloc_fast_execution_point(struct EpmExecutionPoint *ep,
                                   size_t code_size)
{
#ifndef FAST_EP
   (void) ep;
   (void) code_size;
   return NULL;
#else
   EpmStatus rc;
   Msr lock;
   union EP new_ep;
   union EP list_link_ep;
   OSADDRESS address;
   List *list_ep;
   List *list_fast_ep;
   List *list_fast_veneer;
   ListLink *list_link_fast_ep;
   struct fast_ep_veneer *fast_veneer;
   size_t total_size;
   void *code_area = NULL;
   OSADDRESS curr_opcode;
   int32_t relative_dest;

   
   /* Make sure that EPM is initialized. */
   if (!epm_init_completed)
   {
      rc = ose_epm_init();
      if (rc != EPM_OK)
      {
         return NULL;
      }
   }

   DLOG(("epm.c(%d): ose_epm_alloc_fast_execution_point() Alloc fast ep @ 0x%08X "
         "pid: 0x%08X code size: 0x%08X\n",
         __LINE__, ep->address, ep->pid, code_size));
   
   total_size = ((code_size + sizeof(struct fast_ep_veneer))
                & PAGE_SIZE_MASK) + PAGE_SIZE;
   
   new_ep.ep = ep;
   address = new_ep.ep->address; 

   list_fast_ep = list_lookup((OSADDRESS) new_ep.ep->address);
   LOCK_SAVE(lock);
   SMP_SPIN_LOCK();
   list_link_fast_ep = find_fast_ep(list_fast_ep, address, ADDRESS_MASK);
  
   if (list_link_fast_ep == NULL)
   {
      /* No other fast ep exists at this address */

      /* Must unlock while calling system calls */
      SMP_SPIN_UNLOCK();
      LOCK_RESTORE(lock);
      fast_veneer = (struct fast_ep_veneer *)
	get_memory_near(address, total_size, MAX_JUMP_IN_ONE_INSTRUCTION,
			FALSE);
      if(!fast_veneer)
      {
         code_area = NULL;
         goto error;
      }

      LOCK_SAVE(lock);
      SMP_SPIN_LOCK();
      list_link_fast_ep = find_fast_ep(list_fast_ep, address, ADDRESS_MASK);
      if (list_link_fast_ep)
      {
	/* Someone has now added a fast ep. We do not need the new region. */
	SMP_SPIN_UNLOCK();
	LOCK_RESTORE(lock);
	mm_delete_region((OSDOMAIN)0, (OSADDRESS)list_link_fast_ep);
	goto fast_ep_already_present;
      }

      list_init_link(new_ep.link);
      new_ep.idata->magic = EPM_MAGIC;
      new_ep.idata->instance = ++ep_instance;
      new_ep.idata->ep_veneer = NULL;
      new_ep.idata->fast_ep_veneer = fast_veneer;
      
     ((struct sw_bp_state*)(void*)new_ep.idata->fast_ep_veneer->sub_state)->valid = 0;
      
      /* init */
      list_init_link((List *)fast_veneer);
      fast_veneer->magic = FAST_VENEER_MAGIC;
      fast_veneer->active = 0;
      fast_veneer->instance = new_ep.idata->instance;
      /* is size needed ??? never used now */
      fast_veneer->ep = new_ep.ep;
      fast_veneer->size = total_size;
      fast_veneer->size_callback_code = code_size;
 
      code_area = &(fast_veneer->callback_code);
  
      /* check if normal ep is set */
      list_ep = list_lookup(address);   
      list_link_ep.link = find_first_ep(list_ep, address, ADDRESS_MASK);
      if (list_link_ep.link != NULL)
      {
         /* Get opcode from EP TRAP/breakpoint set */
         if (list_link_ep.idata->fast_ep_veneer)
         {
            /* fast ep there should be non */
            SMP_SPIN_UNLOCK();
            LOCK_RESTORE(lock);
            ERRLOG(("EPM: ERROR Failed to set fast ep at adress 0x%08X fast ep "
                     "allready present\n",
                     list_link_ep.ep->address));
            code_area = NULL;
            mm_delete_region(0, (OSADDRESS) fast_veneer);
            goto error;
         }
         
         CP_SUB_STATE(new_ep.idata->fast_ep_veneer->sub_state,
                  list_link_ep.idata->ep_veneer->sub_state);
      }
      else
      {
         /* No ep set, get opcode from memory */
         PREPARE_FOR_PAGE_FAULT(handle_page_fault);
         rc = sw_bp_setstate(address,new_ep.idata->fast_ep_veneer->sub_state);
         PREPARE_FOR_PAGE_FAULT(NULL);
         if (rc != EPM_OK) 
         {
            SMP_SPIN_UNLOCK();
            LOCK_RESTORE(lock);
            ERRLOG(("EPM: Failed to read opcode at adress 0x%08X\n",
                     address));
            code_area = NULL;
            mm_delete_region(0, (OSADDRESS) fast_veneer);
            goto error;
         }
      }
      if (bp_conf->get_opcode_from_state)
      {
         curr_opcode = bp_conf->get_opcode_from_state(
                  new_ep.idata->fast_ep_veneer->sub_state);
      }
      else 
      {
         SMP_SPIN_UNLOCK();
         LOCK_RESTORE(lock);
         ERRLOG(("EPM: get_opcode_from_state() needed for target to "
                  "support more then one ep on an adress at 0x%08X\n",
                  new_ep.ep->address));
         code_area = NULL;
         mm_delete_region(0, (OSADDRESS) fast_veneer);
         goto error;
      }
      
#ifdef OSPP   
      rc = zzepm_setup_fast_ep_codearea( curr_opcode,  address,
                                         fast_veneer->code,
                                         sizeof(fast_veneer->code), 
                                         (OSADDRESS)code_area, 
                                         &(fast_veneer->emulate_address));
      if (rc != EPM_OK)
      {
         SMP_SPIN_UNLOCK();
         LOCK_RESTORE(lock);
         ERRLOG(("EPM: Failed to setup fast ep opcodes at "
                  "adress 0x%08X\n", fast_veneer->code));
         code_area = NULL;
         mm_delete_region(0, (OSADDRESS) fast_veneer);
         goto error;
      }
      relative_dest = (uint32_t)&(fast_veneer->code) - (uint32_t)(address);
      fast_veneer->fast_ep_opcode = 0x48000000 | (relative_dest & 0x03fffffc);

#else
#error ARCH not supported
#endif
          
      /*
       * Link ep block last in list.
       * We want increasing instance numbers as we traverse the list.
       */
      /* Find matching lists */
      list_ep = list_lookup((OSADDRESS) new_ep.ep->address);
      list_fast_veneer = fast_list_lookup((OSADDRESS) fast_veneer);

      list_insert_tail(list_ep, new_ep.link);
      list_insert_tail(list_fast_veneer, (List *) fast_veneer);
      SMP_SPIN_UNLOCK();
      LOCK_RESTORE(lock);
   }
   else
   {
      /*
       * Current code can't handle more then one fast ep at the same address
       * they could be chained if one is careful about insert/removal of them
       * as you then need to be able to batch inside a veneer area of another
       * fast ep.
       */
      SMP_SPIN_UNLOCK();
      LOCK_RESTORE(lock);

   fast_ep_already_present:
      ERRLOG(("EPM: Failed to set fast ep at adress 0x%08X fast ep "
               "already present\n",
               address));

      code_area = NULL;
   }

error:
   /* Note that also successful calls will get here! */
   DLOG(("epm.c(%d): ose_epm_alloc_fast_execution_point() Done return 0x%x\n",
          __LINE__, code_area));
   return code_area;
#endif
}

EpmStatus
ose_epm_free_fast_execution_point(struct EpmExecutionPoint *ep)
{
#ifdef FAST_EP
   EpmStatus rc;
   Msr lock;
   union EP tmp_ep;
   union EP free_ep;
   struct fast_ep_veneer *fast_veneer;
   struct fast_ep_veneer *fast_veneer2;
   List *list;
   List *list_fast_veneer;
   union SIGNAL *sig;
   
   /* Make sure that EPM is initialized. */
   if (!epm_init_completed)
   {
      rc = ose_epm_init();
      if (rc != EPM_OK)
         return rc;
   }
   
   DLOG(("epm.c(%d): ose_epm_free_fast_execution_point() Free fast ep @ %#lx "
            "pid: %#lx.\n", __LINE__, ep->address, ep->pid));

   free_ep.ep = ep;
   fast_veneer = free_ep.idata->fast_ep_veneer;
   
   if (fast_veneer == NULL  || fast_veneer->active != 0)
   {
      rc = EPM_INVALID_ARGUMENTS;
      goto error;
   }
      
   LOCK_SAVE(lock);
   SMP_SPIN_LOCK();
   
   /* check the list so there is no usage error and we move it over to the
    * garbage collect process */
   list_fast_veneer = fast_list_lookup((OSADDRESS) fast_veneer);
   fast_veneer2 = find_fast_veneer(list_fast_veneer, fast_veneer);

   list = list_lookup((OSADDRESS) free_ep.ep->address);
   tmp_ep.link = find_fast_ep(list, free_ep.ep->address, ADDRESS_MASK);

   
   
   if ((fast_veneer == fast_veneer2) && (tmp_ep.link != NULL) && 
            (tmp_ep.idata->fast_ep_veneer == fast_veneer) &&
            free_ep.ep == tmp_ep.ep)
   {
      list_unlink(free_ep.link);
      list_unlink(fast_veneer);
      epm_list_init((List *)fast_veneer);
      list_insert_tail(&fast_ep_remove_list, (List *) fast_veneer);
      SMP_SPIN_UNLOCK();
      LOCK_RESTORE(lock);
      /* Trigger gc */
      sig = alloc(sizeof(struct epm_gc_fast_veneer), 
                  EPM_FAST_VENEER_SIGNAL_TRIGGER_GC);
      sig->gc_fast_veneer.mm_region = (OSADDRESS) fast_veneer;
      send(&sig, epm_veneer_gc_process_);
#ifdef ENABLE_SYNC_GC
   {
      static const SIGSELECT gc_sigsel[] = {1, EPM_FAST_VENEER_SIGNAL_TRIGGER_GC};
      sig = receive(gc_sigsel);
      free_buf(&sig);
   }
#endif
      rc = EPM_OK;
   }
   else
   {
      ERRLOG(("EPM: ose_epm_free_fast_execution_point() lookup:"
             " fast_veneer 0x%x == fast_veneer2 0x%x tmp_ep.link=0x%x\n",
             fast_veneer,fast_veneer2,tmp_ep.link));
      SMP_SPIN_UNLOCK();
      LOCK_RESTORE(lock);
      rc = EPM_INVALID_ARGUMENTS;
   }
error:
   DLOG(("epm.c(%d): ose_epm_free_fast_execution_point() Done return %d\n", __LINE__, rc));
   return rc;

#else
   (void) ep;
   return EPM_FUNCTION_NOT_IMPLEMENTED;
#endif
}


EpmStatus
ose_epm_enable_fast_execution_point(struct EpmExecutionPoint *ep)
{
#ifdef FAST_EP
   EpmStatus rc;
   Msr lock;
   List *list_ep;
   union EP list_link_ep;
   union EP enable_ep;
   struct fast_ep_veneer *fast_veneer;
   OSADDRESS address;
   
   /* Make sure that EPM is initialized. */
   if (!epm_init_completed)
   {
      rc = ose_epm_init();
      if (rc != EPM_OK)
         return rc;
   }

   DLOG(("epc.c(%d): ose_epm_enable_fast_execution_point() Enable fast ep @ %#lx "
            "pid: %#lx.\n", __LINE__, ep->address, ep->pid));

   enable_ep.ep = ep;
   address = enable_ep.ep->address;
   fast_veneer = enable_ep.idata->fast_ep_veneer;
   
   if (fast_veneer == NULL  || fast_veneer->active != 0)
   {
      rc = EPM_INVALID_ARGUMENTS;
      goto error;
   }

   /* Only write a branch instruction if not ep (e.g. no TRAP) is present */
   list_ep = list_lookup(address);
   list_link_ep.link = find_first_ep(list_ep, address, ADDRESS_MASK);
   while ((list_link_ep.link != NULL) && 
          (list_link_ep.idata->fast_ep_veneer != NULL) && 
          (list_link_ep.idata->fast_ep_veneer->active == 0) )   
   {
      /* We could find ourself then check one time more */
      list_link_ep.link = find_next_ep(list_ep,list_link_ep.link, 
                                                   address, ADDRESS_MASK);
   }
   fast_veneer->active = 1;

   /* Do not overwrite an ep (trap) */
   if (list_link_ep.link == NULL)
   {
      LOCK_SAVE(lock);
      SMP_SPIN_LOCK();
      SMP_STOP_OTHER_CPUS_INVALIDATE(address);
      zzepm_set_opcode(address, fast_veneer->fast_ep_opcode);
      SMP_START_OTHER_CPUS();
      SMP_SPIN_UNLOCK();
      LOCK_RESTORE(lock);
   }
   rc = EPM_OK;
error:
   DLOG(("epm.c(%d): ose_epm_enable_fast_execution_point() Done return %d\n", __LINE__, rc));
   return rc;
#else
   (void) ep;
   return EPM_FUNCTION_NOT_IMPLEMENTED;
#endif
}

EpmStatus
ose_epm_disable_fast_execution_point(struct EpmExecutionPoint *ep)
{
#ifndef FAST_EP
   (void) ep;

   return EPM_FUNCTION_NOT_IMPLEMENTED;
#else
   EpmStatus rc;
   Msr lock;
   List *list_ep;
   union EP list_link_ep;
   union EP disable_ep;
   struct fast_ep_veneer *fast_veneer;
   OSADDRESS address;

   
   /* Make sure that EPM is initialized. */
   if (!epm_init_completed)
   {
      rc = ose_epm_init();
      if (rc != EPM_OK)
         return rc;
   }
   
   DLOG(("epm.c(%d): ose_epm_disable_fast_execution_point() Disable fast ep @ %#lx "
            "pid: %#lx.\n", __LINE__, ep->address, ep->pid));

   disable_ep.ep = ep;
   address = disable_ep.ep->address;
   fast_veneer = disable_ep.idata->fast_ep_veneer;
   
   if (fast_veneer == NULL || fast_veneer->active == 0)
   {
      rc = EPM_INVALID_ARGUMENTS;
      goto error;
   }

   fast_veneer->active = 0;
   
   /* Only write a branch instruction if not ep (e.g. no TRAP) is present */
   list_ep = list_lookup(address);
   list_link_ep.link = find_first_ep(list_ep, address, ADDRESS_MASK);
   
   while ((list_link_ep.link != NULL) && 
          (list_link_ep.idata->fast_ep_veneer != NULL) && 
          (list_link_ep.idata->fast_ep_veneer->active == 0) )   
   {
      /* We could find ourself then check one time more */
      list_link_ep.link = find_next_ep(list_ep,list_link_ep.link, 
                                                   address, ADDRESS_MASK);
   }

   /* Do not overwrite an ep (trap) */
   if (list_link_ep.link == NULL)
   {
      LOCK_SAVE(lock);
      SMP_SPIN_LOCK();
      SMP_STOP_OTHER_CPUS_INVALIDATE(address);
      zzepm_set_opcode(address,
               ((struct sw_bp_state*)(void*)fast_veneer->sub_state)->saved_opcode);
      SMP_START_OTHER_CPUS();
      SMP_SPIN_UNLOCK();
      LOCK_RESTORE(lock);
   }
   rc = EPM_OK;

error:
   DLOG(("epm.c(%d): ose_epm_disable_fast_execution_point()"
            " Done return %d\n", __LINE__, rc));
   return rc;
#endif   
}

int
ose_epm_fast_execution_point_inside_check_callback(OSADDRESS address)
{
#ifdef FAST_EP
   Msr lock;
   List *list;
   struct fast_ep_veneer *fast_veneer;
   
   DLOG(("EPM: ose_epm_fast_execution_point_inside_check_callback()"
            " address 0x%08X\n", address));
   /* Make sure that EPM is initialized. */
   if (!epm_init_completed)
   {
      DLOG(("EPM: ose_epm_fast_execution_point_inside_check_callback()"
               " done return 0 (epm not initialized yet)\n"));
      return 0;
   }
   
   list = fast_list_lookup(address & PAGE_SIZE_MASK); /* Find matching list. */
   LOCK_SAVE(lock);
   SMP_SPIN_LOCK();
   fast_veneer = (struct fast_ep_veneer *) 
                        find_fast_veneer(list, (struct fast_ep_veneer *)
                                             ( address & PAGE_SIZE_MASK));
   if (fast_veneer == NULL)
   {
      SMP_SPIN_UNLOCK();
      LOCK_RESTORE(lock);
      DLOG(("EPM: ose_epm_fast_execution_point_inside_check_callback()"
               " done return 0 (no fast_veneer)\n"));
      return 0;
   }
   /* address is in an area fast veneer area, check if it is inside 
    * check callback code area*/
   if(address >= (OSADDRESS) &(fast_veneer->callback_code) &&
            (address < (((OSADDRESS)&(fast_veneer->callback_code) 
                       + fast_veneer->size_callback_code))))
   {
      SMP_SPIN_UNLOCK();
      LOCK_RESTORE(lock);
      DLOG(("EPM: ose_epm_fast_execution_point_inside_check_callback()"
               " done return 1 (match)\n"));
      return 1;
   }
   SMP_SPIN_UNLOCK();
   LOCK_RESTORE(lock);
   DLOG(("EPM: ose_epm_fast_execution_point_inside_check_callback()"
            " done return 0 (inside fast_veneer but not in usercode)\n"));
   return 0;
#else
   (void) address;
   return 0;
#endif
}




/* These functions are implemented in the file epm_arch.c
 *
 * EpmStatus
 * ose_epm_get_reg(struct cpu_regdump *dump, U32 id, U32 *value);
 *
 * EpmStatus
 * ose_epm_set_reg(struct cpu_regdump *dump, U32 id, U32 value);
 */
#endif
