/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2009-2013 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */


/**
 * @private
 *
 * @file efmi.c
 *
 * @brief Executive Freeze Mode Interface
 *
 * @long
 * This file implements the Executive Freeze Mode Interface. 
 * It may be used to fetch information from the executive in a freeze mode
 * environment. The purpose is to make OSE Executive concepts accessable
 * for freeze mode debug monitors that can't use ordinary system calls.
 * @long
 * WARNING: Use of efmi functionality for any other purpose will severely 
 * damage the interrupt latency!
 * 
 */

#if defined(HOST_TOOL_ENVIRONMENT)
/* macro for empty arglist i.e. (void) */
#define ARGLIST0(a) (HostFmiInstance *i)
/* macro for other arglists */
#define ARGLIST(...) (HostFmiInstance *i, __VA_ARGS__)
#else
/* macro for empty arglist i.e. (void) */
#define ARGLIST0(a) (void)
/* macro for other arglists */
#define ARGLIST(...) (__VA_ARGS__)
#endif

/*
*****************************************************************************
** INCLUDE FILES
*****************************************************************************
*/

#if defined(HOST_TOOL_ENVIRONMENT)

#include "efmi_host.h"
#include "efmi.h"

/* Global variables found by symbol table lookup. Kept as thread-global
 * variables to simplify the conversion of efmi.c. */
static MAKE_GLOBAL(FMI_TLS_ZZUU, zzuu);
static MAKE_GLOBAL(FMI_TLS_ZZOSVARP, zzosvarp);
static SET_TYPE(world, zzuu);
static SET_TYPE(osvarp, zzosvarp);

/* We must translate register id:s since this code is generic. */
#define ZZREG_SSLIMIT EFMI_REG_SSLIMIT

#else

/* Make sure we don't use debug versions of system calls internally. */

#ifdef OS_DEBUG
#undef OS_DEBUG
#endif

#include "osenames.h"		/* Names for all public symbols.	*/
#include "efmi.h"		/* Public interface */
#include "string.h"		/* str* and mem* functions              */
#include "pose.con"		/* Configurable types and constants.	*/
#include "pose.h"		/* Target interface.		 	*/
#include "pose_internal.h"	/* Manifest constants, types and variables */

#ifndef softOSE
#include "efmi_registers.h"	/* Process specific registers */
#endif

#include "ramlog.h"

/* These macros are redefined when building the host version of efmi. */
#define ENTER_EFMI(a)
#define EXIT_EFMI()

#define SET_TYPE(type, var)
#define READ_VAL(ptr, member) (ptr).member
#define DEREF_VAL(ptr, member) (ptr)->member
#define READ_PTR(ptr, member) (ptr).member
#define DEREF_PTR(ptr, member) (ptr)->member
#define DEREF_VAL_TYPE(type, ptr, member) (ptr)->member
#define DEREF_PTR_TYPE(type, ptr, member) (ptr)->member

#endif

/*
*****************************************************************************
** MACROS
*****************************************************************************
*/

/* Smallest MMU page size used, for any target architecture. */
#define PAGE_SIZE		4096

#define MIN(a,b)		(((a) < (b)) ? (a) : (b))
#define NOT_WORD_ALIGNED(vp)	(3 & (U32) (vp))
#define NOT_POOL_ALIGNED(vp)	(/*((POOL_ALIGNMENT) - 1) &*/ ((U32) vp))

/*
** Convert a pcb pointer to a process id.
**
** The most significant 16 bits of the process id is the instance number.
** The least significant 16 bits is the process index.
*/

#define pcb_to_pid(pcb) ((( (U32) DEREF_VAL_TYPE(struct _pcb, pcb, instance)) << 16) | DEREF_VAL_TYPE(struct _pcb, pcb, pindex))

/*
** Extract a pcb pointer from a process id.
**
** The most significant 16 bits of the process id is the instance number.
** The least significant 16 bits is the process index. The pindex should
** be checked as valid before this function is called.
*/


static struct _pcb *
pid_to_pcb(PROCESS pid)
{
  if(!pid)
    {
      return (struct _pcb *) READ_PTR(zzosvarp, syseg_varp);
    }
#if !defined (HOST_TOOL_ENVIRONMENT)
  return zzuu.pix_list[(PCB_INDEX_T) pid];
#else
  {
    uint32_t addr = READ_VAL(zzuu, pix_list);
    return (struct _pcb *)fmi_read_u32(addr + (PCB_INDEX_T)pid * 4);
  }
#endif
}

/*
** Extract an instance number from a process id.
**
** The most significant 16 bits of the process id is the instance number.
** The least significant 16 bits is the process index.
*/

#define pid_to_instance(pid) ((PCB_INSTANCE_T) (pid >> 16))


/*
*****************************************************************************
** TYPES
*****************************************************************************
*/

/*
 *---------------------------------------------------------------------------
 * Size of the local cursor structs must not exeed the size of the public
 * view of the corresponding struct, found in efmi.h!
 *---------------------------------------------------------------------------
 */

struct CurrentCursor  
{
  U32			next;	/* Next cpu to fetch. */
  size_t 		size;	/* Size of result struct. */
  U32			num_cpus; /* Number of cpus. */
};

struct SegmentCursor  
{
  const struct segcb	*next;	/* Next segment to fetch. */
  size_t 		size;	/* Size of result struct. */
  U32			spare;	/* For future use */
};

struct BlockCursor          
{
  const struct _block	*next;	/* Next block to fetch. */
  PROCESS		segment_id; /* Segment to match. */
  size_t		size;	/* Size of result struct. */
  U32			spare;	/* For future use */
};

struct ProcessCursor        
{
  const struct _pcb	*next;	/* Next process to fetch. */
  size_t		size;	/* Size of result struct. */
  U32			spare;	/* For future use */
};

struct EnvVarCursor         
{
  const struct _pcb	*pcb;	/* Process or block. */
  size_t 		max_name_len;
  size_t 		max_value_len;
  const struct env_buf	*next;
  U32			spare;
};

struct PoolCursor
{
  const struct poolcb	*next;	/*  */
  size_t		size;
  U32			spare;	/* For future use */
};

struct FragmentCursor
{
  const struct poolcb	*poolcb; /*  */
  const struct pool_extension *next; /*  */
  size_t		size;
  U32			spare;	/* For future use */
};

struct SignalCursor
{
  U32			traverse_pool; /* 1==traverse pool, 0==traverse queue */
  const struct _signal	*next;	/* sigadm pointer */
  const struct poolcb	*poolcb; /* if traverse_pool != 0 */
  size_t		size;
  U32			spare;
};

struct AttachCursor
{
  U32			search_method;  /* 0=system 1=match victim, 2=match owner */
  U32			options;	/* 0=PROCESS info only, 1=also signal info */
  PROCESS		pid;	/* owner or victim */
  OSATTREF		next;
  size_t		size;
};

struct ProcNameCursor       
{
  U32			spare;
  PCB_INDEX_T		next;	/* Next pixlist entry to check */
};

/*
*****************************************************************************
** LOCAL FUNCTIONS
*****************************************************************************
*/

/* SIMPLE ERROR TRACE TO RAMLOG - START */
#if 0
#define RAMLOGTRACE 1
#define ERRTRACE(p,v) ramlog_printf("EFMI call failed, line: %d data: 0x%x 0x%x \n", __LINE__, (U32)p, (U32)v)
#if 0
#define CALLTRACE(name) ramlog_printf(".%s\n", name)
#else
#define CALLTRACE(name) 
#endif
#else
#define RAMLOGTRACE 0
#define ERRTRACE(p,v)
#define CALLTRACE(name)
#endif

static void
if_trace_dump(const void *vp, const char *name, int words)
{
#if RAMLOGTRACE == 0
  (void)vp;
  (void)words;
#else
  const U32 *p = (const U32 *)vp;

  if(RAMLOGTRACE)
    {
      ramlog_printf("EFMI: %s @ 0x%8.8x:", name, (U32)p);
      while(words)
	{
#if !defined (HOST_TOOL_ENVIRONMENT)           
	  ramlog_printf(" %8.8x", *p);
#else
	  ramlog_printf(" %8.8x", fmi_read_u8((uint32_t)p));
#endif          
	  p++;
	  words--;
	} /* while */

      ramlog_printf("\n");
    }
#endif
  (void)name;  
}

/* SIMPLE ERROR TRACE TO RAMLOG - END */

static U32 get_smp_num_cpus(void)
{
#if defined(OSE_SMP) || defined(HOST_TOOL_ENVIRONMENT)
  return zzkrn_num_cpus();
#else
  return 1;
#endif
}

/* 
** Truncating copy, returns nonzero if truncated. Size > 0 assumed.
*/
static int 
strncpy_truncated(char * dest, const char *source, size_t size)
{
#if !defined (HOST_TOOL_ENVIRONMENT)   
  if(size == 0)
    {
      return 1;			/* Bad size argument */
    }
  size--;
  while(size--)
    {
      char c;
      c = *source++;
      if(0 == (*dest++ = c))
	{
	  return 0;		/* Complete copy. */
	}
    } /* while */
  *dest = 0;
  return *source;		/* Nonzero if truncated. */
#else
  return fmi_strncpy_truncated(dest, source, size);
#endif  
} /* strncpy_truncated */

static U32
public_priority(U32 pcb_priority)
{
  return pcb_priority & 0x1F;	/* External representation of priority,
				 * works for all process types */
} /* public_priority */

static enum PROCESS_TYPE
prio_to_type(U32 priority)
{
  if(priority < 32)
    return OS_PRI_PROC;
  else if(priority < 33)
    return OS_BG_PROC;
  else if(priority < 96)
    return OS_INT_PROC;
  else if(priority < 128)
    return OS_TI_PROC;
  else
    return OS_PHANTOM;
} /* prio_to_type */

static EfmiStatus
check_word_pointer(const void *vp)
{
  if(!vp)
    {
      return EFMI_UNAVAILABLE;	/* Mission failed */
    }
  if(NOT_WORD_ALIGNED(vp))
    {
      return EFMI_KERNEL_CORRUPTED; /* Mission failed */
    }
  return EFMI_OK;
} /* check_word_pointer */

/* This is a helper for find_pool() and check_pool_pointer() */
static EfmiStatus
helper_find_pool(const void *vp, const struct poolcb **poolcb)
{
  const struct poolcb *pool;
  const struct pool_extension *frag;

  SET_TYPE(poolcb, pool);
  SET_TYPE(pool_extension, frag);
  
  if(!vp)
    {
      ERRTRACE(vp, 0);
      return EFMI_UNAVAILABLE;	/* Mission failed */
    }
  if(NOT_POOL_ALIGNED(vp))
    {
      ERRTRACE(vp, 0);
      return EFMI_UNAVAILABLE;	/* Mission failed */
    }
  pool = *poolcb;
  if(pool)			/* This pool only */
    {
       for(frag = DEREF_PTR(pool, frag_list); frag;
           frag = DEREF_PTR(frag, frag_list))
	{
	  if(DEREF_VAL(frag, magic_frag) != MAGIC_POOL)
	    {
	      ERRTRACE(frag, DEREF_PTR(frag, magic_frag));
	      return EFMI_KERNEL_CORRUPTED;
	    }
  
	  if(DEREF_VAL(frag, base) <= (OSADDRESS) vp &&
             DEREF_VAL(frag, limit) > (OSADDRESS) vp)
	    {
	      return EFMI_OK;
	    }
	} /* for */
    }
  else				/* Any pool */
    {
       for(pool = READ_PTR(zzuu, pool_list); pool;
           pool = DEREF_PTR(pool, next))
	{
	  if(DEREF_VAL(pool, magic) != MAGIC_POOL)
	    {
	      ERRTRACE(pool, DEREF_VAL(pool, magic));
	      return EFMI_KERNEL_CORRUPTED;
	    }

	  for(frag = DEREF_PTR(pool, frag_list); frag;
              frag = DEREF_PTR(frag, frag_list))
	    {
	      if(DEREF_VAL(frag, magic_frag) != MAGIC_POOL)
		{
		  ERRTRACE(frag, DEREF_PTR(frag, magic_frag));
		  return EFMI_KERNEL_CORRUPTED;
		}
	      if(DEREF_VAL(frag, base) <= (OSADDRESS) vp &&
                 DEREF_VAL(frag, limit) > (OSADDRESS) vp)
		{
		  *poolcb = pool;
		  return EFMI_OK;
		}
	    } /* for */
	} /* for */
    }
  ERRTRACE(vp, *poolcb);
  return EFMI_UNAVAILABLE;	/* Failed: not found in this/any pool */
} /* helper_find_pool */

/* 
** Find the proper pool for non-NULL and POOL_ALIGNED pointer.
**
**
** Returns:	EFMI_KERNEL_CORRUPTED if errors found in pool control blocks.
**		EFMI_UNAVAILABLE for NULL pointer or outside the pool/pools,
**		or called with a pointer that is not POOL_ALIGN aligned 
**		(all kernel objects have that alignment).
**		EFMI_OK if inside the selected resp. any pool.
*/

static EfmiStatus
find_pool(const void *vp, const struct poolcb **poolcb)
{
  *poolcb = NULL;
  return helper_find_pool(vp, poolcb);
} /* find_pool */
/* 
** Check for non-NULL and POOL_ALIGNED pointer in one specified or in any pool.
**
** Returns:	EFMI_KERNEL_CORRUPTED if errors found in pool control blocks.
**		EFMI_UNAVAILABLE for NULL pointer or outside the pool/pools,
**		or called with a pointer that is not POOL_ALIGN aligned 
**		(all kernel objects have that alignment).
**		EFMI_OK if inside the selected resp. any pool.
*/

static EfmiStatus
check_pool_pointer(const void *vp, const struct poolcb *poolcb)
{
  const struct poolcb *pool;

  pool = poolcb;
  return helper_find_pool(vp, &pool);
} /* check_pool_pointer */

/* 
** Check for non-NULL and POOL_ALIGNED pointer in the system pool only.
*/
static EfmiStatus
check_syspool_pointer(const void *vp)
{

  /* 
  ** These specific objects are logically part of the system pool, even if not 
  ** physically.
  */
  if(vp == READ_PTR(zzosvarp, syspool_varp) || /* system pool cb */
     vp == READ_PTR(zzosvarp, syseg_varp)) /* system segment cb */
    {
      return EFMI_OK;
    }
  return check_pool_pointer(vp, READ_PTR(zzosvarp, syspool_varp));
} /* check_syspool_pointer */

/* 
 * This function is used by the target interface to validate pointers for 
 * register access.
 */
static OSBOOLEAN 
is_syspool_pointer(const void *vp)
{
  OSADDRESS a = (OSADDRESS)vp;

  return  (check_pool_pointer((void *) (a /*& ~((POOL_ALIGNMENT) - 1)*/),
			      READ_PTR(zzosvarp, syspool_varp)) == EFMI_OK) ? 1 : 0;
} /* is_syspool_pointer */

static EfmiStatus
check_stkcb(const struct stkcb *stkcb, const struct poolcb *poolcb)
{
  void		*adj_stk_top;
  EfmiStatus	status;

  SET_TYPE(stkcb, stkcb);
  
  if((status = check_syspool_pointer(stkcb)) != EFMI_OK)
    {
      return status;
    }
  
  /* 
  ** The last stack in a pool has a top equal to the pool limit, so we
  ** must subtract a POOL_ALIGNMENT to get a pool_aligned  address that 
  ** is INSIDE the pool limits.
  */
  adj_stk_top = ((char *) DEREF_PTR(stkcb, top)) - POOL_ALIGNMENT;
  if(DEREF_PTR(stkcb, limit) >= DEREF_PTR(stkcb, top) ||
     ((status = check_pool_pointer(adj_stk_top, poolcb)) != EFMI_OK) ||
     ((status = check_pool_pointer(DEREF_PTR(stkcb, limit), poolcb)) != EFMI_OK))
    {
      return status;
    }
  return EFMI_OK;
} /* check_stkcb */

/* 
** Check pool, segment, process and block descriptors
** These shall all be in the system pool, except a few static objects,
** that are checked separately.
*/
static EfmiStatus
check_pix_pointer(const void *vp, U32 magic)
{
  EfmiStatus		status;
  const struct _pcb	*pcb = vp;

  SET_TYPE(_pcb, pcb);
  
  if(vp != READ_PTR(zzosvarp, syspool_varp) && /* system pool cb */
     vp != READ_PTR(zzosvarp, syseg_varp)) /* system segment cb */
    {
      status = check_syspool_pointer(pcb);
      if(status != EFMI_OK)
	{
	  ERRTRACE(vp, magic);
	  return status;
	}
    }
  
  if(DEREF_VAL(pcb, magic) != magic)
    {
      ERRTRACE(DEREF_VAL(pcb, magic), magic);
      if_trace_dump(pcb, "pcb", 8);
      if_trace_dump(-32 + (const char *)pcb, "pcb-0x20", 8);
      return EFMI_UNAVAILABLE; /* Mission failed */
    }
  return EFMI_OK;
} /* check_pix_pointer */

static EfmiStatus 
check_pix_identity(PROCESS pid)
{
  EfmiStatus		status;
  const struct _pcb	*pcb;

  SET_TYPE(_pcb, pcb);
  
  if(READ_VAL(zzuu, next_free_pix) <= (PCB_INDEX_T) pid)
    {
      return EFMI_UNAVAILABLE;
    }

  pcb = pid_to_pcb(pid);

  if((status = check_syspool_pointer(pcb)) != EFMI_OK)
    {
      return status;		/* Mission failed */
    }
  
  if(DEREF_VAL(pcb, instance) != pid_to_instance(pid))
    {
      return EFMI_UNAVAILABLE;	/* Mission failed */
    }
  return EFMI_OK;
} /* check_pix_identity */



static EfmiStatus
check_segcb(const struct segcb *segcb)
{
  EfmiStatus		status;

  status = check_pix_pointer(segcb, MAGIC_SEG);
  return status;
} /* check_segcb */

static EfmiStatus
check_poolcb(const struct poolcb *poolcb)
{
  EfmiStatus		status;

  SET_TYPE(poolcb, poolcb);
  
  status = check_pix_pointer(poolcb, MAGIC_POOL);
  if(status == EFMI_OK)
    {
      /* There should be at least one fragment */
      status = check_syspool_pointer(DEREF_PTR(poolcb, frag_list));
    }
  return status;
} /* check_poolcb */

static EfmiStatus
check_block(const struct _block *block)
{
  EfmiStatus		status = EFMI_OK;

  SET_TYPE(_block, block);
  
  if((status = check_pix_pointer(block, MAGIC_BCB)) != EFMI_OK)
    {
      return status;		/* Mission failed */
    }
  if((status = check_poolcb(DEREF_PTR(block, stackpool))) != EFMI_OK)
    {
      return status;		/* Mission failed */
    }
  if(DEREF_PTR(block, sigpool) != DEREF_PTR(block, stackpool))
    {
      status = check_poolcb(DEREF_PTR(block, sigpool));
    }
  return status;
} /* check_block */

static EfmiStatus
check_process(const struct _pcb *pcb)
{
  EfmiStatus		status;

  SET_TYPE(_pcb, pcb);
  
  if((status = check_pix_pointer(pcb, MAGIC_PCB)) != EFMI_OK)
    {
      return status;		/* Mission failed */
    }
  if(DEREF_VAL(pcb, pid) != pcb_to_pid(pcb) ||
     DEREF_VAL(pcb, priority) > 128 ||
     ((!(U32) DEREF_VAL(pcb, stack)) != (DEREF_VAL(pcb, priority) > 32)))
    {
      return EFMI_KERNEL_CORRUPTED;
    }
  if((status = check_block(DEREF_PTR(pcb, block))) != EFMI_OK)
    {
      return status;		/* Mission failed */
    }
  if(DEREF_PTR(pcb, uarea) &&
     ((status = check_syspool_pointer(DEREF_PTR(pcb, uarea))) != EFMI_OK))
    {
      return status;		/* Mission failed */
    }
  if(DEREF_PTR(pcb, stack) &&
     ((status = check_stkcb(DEREF_PTR(pcb, stack), DEREF_PTR_TYPE(struct _block, DEREF_PTR(pcb, block), stackpool))) != EFMI_OK))
    {
      return status;		/* Mission failed */
    }
  if((DEREF_VAL(pcb, status) & ST_RECEIVE) &&
     ((status = check_word_pointer(DEREF_PTR(pcb, wanted))) != EFMI_OK))
    {
      return status;		/* Mission failed */
    }
  return EFMI_OK;
} /* check_process */

static EfmiStatus
check_kernel(void)
{
  if(!READ_VAL(zzuu, kernel_initialized))	/* Cleared early in the start sequence */
    {
      ERRTRACE(0,EFMI_KERNEL_UNINITIATED);
      return  EFMI_KERNEL_UNINITIATED; /* Mission failed */
    }

  if(READ_VAL(zzuu, kernel_initialized) != MAGIC_KERNEL_TYPE || /* .bss-like */
     READ_VAL(zzosvarp, magic_kernel_type) != MAGIC_KERNEL_TYPE ||	/* .text-like */
     !READ_VAL(zzuu, sysctlblk) || NOT_WORD_ALIGNED(READ_VAL(zzuu, sysctlblk)) ||
     READ_VAL(zzuu, max_proc) != DEREF_VAL_TYPE(struct sysctlblk, READ_VAL(zzuu, sysctlblk), max_procs) ||
     READ_VAL(zzuu, next_free_pix) > READ_VAL(zzuu, max_proc))
    {
      ERRTRACE(0,EFMI_KERNEL_CORRUPTED);
      return  EFMI_KERNEL_CORRUPTED; /* Mission failed */
    }
  return EFMI_OK;
} /* check_kernel */

/* Macros for converting the current baseline label to a string constant. */
#define XXBL_LABEL_STR(LABEL) #LABEL
#define XBL_LABEL_STR(LABEL) XXBL_LABEL_STR(LABEL)
#define BL_LABEL_STR XBL_LABEL_STR(BL_LABEL)

/*
** Retrieve system information
*/
static EfmiStatus
get_system_info(struct EfmiSystemInfo *info_p, size_t size)
{
  EfmiStatus status = EFMI_OK;
  struct EfmiSystemInfo info;

  info.os_type = EFMI_OS_OSE_5;

#if !defined (HOST_TOOL_ENVIRONMENT)  
  strncpy(info.baseline, BL_LABEL_STR, sizeof(info.baseline));

  /* Check for truncation. */
  if (info.baseline[sizeof(info.baseline) - 1] != '\0')
    {
      status = EFMI_RESULT_TRUNCATED;
      info.baseline[sizeof(info.baseline) - 1] = '\0';
    }

#if defined(OSPP)
  info.cpu_type = EFMI_CPU_PPC;
#elif defined(OSARM)
  info.cpu_type = EFMI_CPU_ARM;
#elif defined(OSMIPS)
  info.cpu_type = EFMI_CPU_MIPS;
#elif defined(OSX86)
  info.cpu_type = EFMI_CPU_X86;
#elif defined(softOSE)

#if defined(LINUX) || defined(WIN32)
  info.cpu_type = EFMI_CPU_X86;
#elif defined(SUN5)
  info.cpu_type = EFMI_CPU_SPARC;
#else
#error Cannot detect soft kernel architecture!
#endif

#else  
#error Cannot detect architecture!
#endif

  *(uint32_t *)&info.endian = EFMI_ENDIAN_VALUE;

#else
  {
    uint32_t baseline;
    HostFmiStatus ret;

    ret = fmi_get_core_baseline(&baseline);
    if (ret != HOST_FMI_OK)
      {
        /* Everything except this value is pretty much known by the caller, so
         * there's no point in returning anything else. */
        return EFMI_UNAVAILABLE;
      }
    else
      {
        /* We know that there is enough room for this string. */
        if (baseline)
          {
            sprintf(info.baseline, "BL%d", baseline);
          }
        else
          {
            strcpy(info.baseline, "Unknown");
          }
      }

    /* Map FMI arch to EFMI cpu type. */
    switch (fmi_cur_instance->arch)
      {
      case FMI_ARCH_PPC:
        info.cpu_type = EFMI_CPU_PPC;
        break;
      case FMI_ARCH_ARM:
        info.cpu_type = EFMI_CPU_ARM;
        break;
      case FMI_ARCH_MIPS:
        info.cpu_type = EFMI_CPU_MIPS;
        break;
      case FMI_ARCH_X86:
        info.cpu_type = EFMI_CPU_X86;
        break;
      default:
        info.cpu_type = EFMI_CPU_UNKNOWN;
        break;
      }

    /* Simply return what we were told during init. */
    memcpy(info.endian, fmi_cur_instance->endian.u8, sizeof(info.endian));
  }
#endif
  
  memcpy(info_p, &info, MIN(size, sizeof(info)));
  return status;
}

/*
** Retrieve segment information
*/
static EfmiStatus
get_segment_info(struct SegmentCursor *cursor, /* May be NULL */
		 PROCESS segment, /* Used when cursor is NULL */
		 struct EfmiSegmentInfo   *info_p,
		 size_t size)
{
  EfmiStatus			status;
  struct EfmiSegmentInfo	info;
  const struct segcb		*segcb;

  SET_TYPE(segcb, segcb);
  
  if((status = check_kernel()) != EFMI_OK)
    {
      return status;		/* Mission failed */
    }
  
  if(cursor)
    {
      /* Find next non-killed segment using the cursor. */
      for(segcb = cursor->next; ; segcb = DEREF_PTR(segcb, next))
	{
	  if(!segcb)
	    {
	      return EFMI_END_OF_LIST;
	    }
	  if(EFMI_OK != (status = check_syspool_pointer(segcb)))
	    {
	      return status;
	    }
	  if(!(DEREF_VAL(segcb, magic) & MAGIC_KILLED))
	    {
	      break;		/* Still alive. */
	    }
	} /* for */
    }
  else
    {
      if(READ_VAL(zzuu, next_free_pix) <= (PCB_INDEX_T) segment)
	{
	  return EFMI_UNAVAILABLE;
	}
      segcb = (struct segcb *) pid_to_pcb(segment);
    }

  if((status = check_segcb(segcb)) != EFMI_OK)
    {
      return status;		/* Mission failed */
    }
  
  if(cursor)
    {
      cursor->next = DEREF_PTR(segcb, next);
    }
  else if(DEREF_VAL(segcb, instance) != pid_to_instance(segment))
    {
      return EFMI_UNAVAILABLE;	/* Mission failed */
    }
  
  info.segment = pcb_to_pid(segcb);
  info.segid = DEREF_VAL(segcb, segid);
  memcpy(info_p, &info, MIN(size, sizeof(info)));
  return EFMI_OK;
} /* get_segment_info */

/*
** Retrieve block information
*/
static EfmiStatus
get_block_info(struct BlockCursor *cursor, /* May be NULL */
	       PROCESS block, /* Used when cursor is NULL */
	       struct EfmiBlockInfo   *info_p,
	       size_t size)
{
  EfmiStatus			status;
  struct EfmiBlockInfo		info;
  const struct _block		*bcb;

  SET_TYPE(_block, bcb);
  
  if((status = check_kernel()) != EFMI_OK)
    {
      return status;		/* Mission failed */
    }
  
  if(cursor)
    {
      for(bcb = cursor->next; ; bcb = DEREF_PTR(bcb, next))
	{
	  if(!bcb)
	    {
	      return EFMI_END_OF_LIST;
	    }
	  if(EFMI_OK != (status = check_syspool_pointer(bcb)))
	    {
	      return status;
	    }
	  if(DEREF_VAL(bcb, magic) & MAGIC_KILLED)
	    {
	      continue;		/* Try next non-killed block. */
	    }
	  if((status = check_block(bcb)) != EFMI_OK)
	    {
	      return status;	/* Mission failed */
	    }
	  if(DEREF_VAL(bcb, segpid) == cursor->segment_id)
	    {
	      break;
	    }
	} /* for */
      cursor->next = DEREF_PTR(bcb, next);
    }
  else
    {
      if(READ_VAL(zzuu, next_free_pix) <= (PCB_INDEX_T) block)
	{
	  return EFMI_UNAVAILABLE;
	}
      bcb = (struct _block *) pid_to_pcb(block);
      if((status = check_block(bcb)) != EFMI_OK)
	{
	  return status;		/* Mission failed */
	}
      if(DEREF_VAL(bcb, instance) != pid_to_instance(block))
	{
	  return EFMI_UNAVAILABLE;	/* Mission failed */
	}
    }  

  info.block = pcb_to_pid(bcb);
  info.user = DEREF_VAL(bcb, user);
  info.segment = DEREF_VAL(bcb, segpid);
  info.sig_pool = pcb_to_pid(DEREF_VAL(bcb, sigpool));
  info.stack_pool = pcb_to_pid(DEREF_VAL(bcb, stackpool));
  info.supervisor_mode = DEREF_VAL(bcb, supervisor_mode);
  info.remcall_server = DEREF_VAL(bcb, remcall_server);
  info.cpu_id = DEREF_VAL(bcb, cpu_id);
  memcpy(info_p, &info, MIN(size, sizeof(info)));
  return EFMI_OK;
} /* get_block_info */

/*
** Retrieve process information
*/
static EfmiStatus
get_process_info(struct ProcessCursor *cursor, /* May be NULL */
		 PROCESS process, /* Used when cursor is NULL */
		 struct EfmiProcessInfo   *info_p,
		 size_t size)
{
  EfmiStatus			status;
  const struct _pcb		*pcb;
  OSADDRESS			line;
/* Don't want these large items on the stack. We are single-threaded here. */
#if !defined(HOST_TOOL_ENVIRONMENT)  
  static struct EfmiProcessInfo	info;
  static char			file[MAX_FILE_NAMELEN];
#else
  struct EfmiProcessInfo	info;
  char				file[MAX_FILE_NAMELEN];
#endif
  
  SET_TYPE(_pcb, pcb);
  
  if((status = check_kernel()) != EFMI_OK)
    {
      return status;		/* Mission failed */
    }
  
  if(cursor)
    {
      /* Find next non-killed pcb using the cursor. */
      for(pcb = cursor->next; ; pcb = DEREF_PTR(pcb, proclist))
	{
	  if(!pcb)
	    {
	      return EFMI_END_OF_LIST;
	    }
	  if(EFMI_OK != (status = check_syspool_pointer(pcb)))
	    {
	      ERRTRACE(pcb, status);
	      return status;
	    }
	  if(!(DEREF_VAL(pcb, magic) & MAGIC_KILLED))
	    {
	      break;		/* Still alive. */
	    }
	} /* for */
    }
  else
    {
      if(READ_VAL(zzuu, next_free_pix) <= (PCB_INDEX_T) process)
	{
	  ERRTRACE(READ_VAL(zzuu, next_free_pix), process);
	  return EFMI_UNAVAILABLE;
	}
      pcb = pid_to_pcb(process);
    }

  if((status = check_process(pcb)) != EFMI_OK)
    {
      ERRTRACE(status, process);
      return status;		/* Mission failed */
    }
  
  if(cursor)
    {
      cursor->next = DEREF_PTR(pcb, proclist);
    }
  else if(DEREF_VAL(pcb, instance) != pid_to_instance(process))
    {
      ERRTRACE(DEREF_VAL(pcb, instance), process);
      return EFMI_UNAVAILABLE;	/* Mission failed */
    }
  
  memset(&info, 0, sizeof(info)); /* safe values */

  info.process = DEREF_VAL(pcb, pid);	/* ose.h PROCESS. */
  info.user = DEREF_VAL(pcb, user);	/* ose.h OSUSER. */
  info.block = pcb_to_pid(DEREF_PTR(pcb, block)); /* ose.h PROCESS. */
  info.parent = DEREF_VAL(pcb, parentpid); /* ose.h PROCESS. */
  info.entry_point = (U32) DEREF_PTR(pcb, entrypoint); /* ose.h OSADDRESS. */
  info.type = prio_to_type(DEREF_VAL(pcb, priority)); /* ose.h enum PROCESS_TYPE. */
  info.priority = public_priority(DEREF_VAL(pcb, priority)); /* ose.h OSPRIORITY. */
  info.status = (U32) ((DEREF_VAL(pcb, status) & 0x1fc) >> 2); /* ose.h 
						     * enum PROCESS_STATUS. */
  info.fsemvalue = DEREF_VAL(pcb, fsemval); /* ose.h OSFSEMVAL. */
  info.sig_cnt_owned = DEREF_VAL(pcb, sig_cnt_owned);
  info.cpu_id = DEREF_VAL(pcb, cpu_id);


  /* 
   * If there is an uarea, get file and line.
   * With paging ".text" enabled, references to the filename may 
   * cause a page fault.
   * FIXME filename in MAS .data or .bss may cause exception! ??
   */
  if(DEREF_PTR(pcb, uarea))
    {
      /* If uarea exists, check_process() has already checked the pointer. */
      if(0 == READ_VAL(zzuu, pagesize))
	{
	  zzget_filename(DEREF_PTR(pcb, uarea), &line, file);
	  if(line)
	    {
	      info.line = line;
	      strncpy(info.file, file, sizeof(info.file));
	      info.file[(sizeof(info.file)) - 1] = 0;
	    }
	}
#if ZZGET_FILE_LINE
      else
	{
	  /* Paging is actived. Avoid pagefault on the filename. */
	  char *filename;
	  zzget_file_line(DEREF_PTR(pcb, uarea), &line, &filename);
	  if(line)
	    {
	      info.line = line;
	      strcpy(info.file, "Unknown");
	    }
	}
#endif
    }
  
  if(info.type == OS_PRI_PROC || info.type == OS_BG_PROC)
    {
      /* Error checks on pcb->stack are already done through check_process() */
      info.process_stack.top = (U32) DEREF_PTR_TYPE(struct stkcb, DEREF_PTR(pcb, stack), top); 
      info.process_stack.limit = (U32) DEREF_PTR_TYPE(struct stkcb, DEREF_PTR(pcb, stack), limit);
#if !defined (HOST_TOOL_ENVIRONMENT)      
      info.process_stack.no_endmark = ((unsigned char) END_MARK) !=
	*(unsigned char *) pcb->stack->limit;
#else
      info.process_stack.no_endmark = ((unsigned char) END_MARK) !=
	fmi_read_u8(DEREF_VAL_TYPE(struct stkcb, DEREF_PTR(pcb, stack), limit));
#endif

#if !defined(softOSE) || defined(HOST_TOOL_ENVIRONMENT)
      {
        U32	limit;
        
        if(DEREF_PTR(pcb, uarea) &&
           0 == zzget_register(DEREF_PTR(pcb, uarea), ZZREG_SSLIMIT, &limit, is_syspool_pointer) &&
           EFMI_OK == check_syspool_pointer((void *)limit) &&
           limit != (U32) DEREF_PTR_TYPE(struct stkcb, DEREF_PTR(pcb, stack), limit))
          {
            info.supervisor_stack.limit = limit;
	    /* Calculate the top as all the TIs do! */
            info.supervisor_stack.top = limit +
              (DEREF_VAL_TYPE(struct sysctlblk, READ_PTR(zzuu, sysctlblk), superv_stack_size) & ~0xf);
#if !defined (HOST_TOOL_ENVIRONMENT)            
            info.supervisor_stack.no_endmark = ((unsigned char) END_MARK) !=
              *(unsigned char *) limit;
#else
            info.supervisor_stack.no_endmark = ((unsigned char) END_MARK) !=
              fmi_read_u8(limit);
#endif            
          }
      }
#endif
    }
  else if ((info.type == OS_INT_PROC || info.type == OS_TI_PROC) &&
	   READ_VAL(zzuu, interrupt_stack_size))
    {
      struct stkcb *int_stk;

      SET_TYPE(stkcb, int_stk);

#if !defined (HOST_TOOL_ENVIRONMENT)      
#ifdef OSE_SMP
      int_stk = zzcpu[pcb->cpu_id].interrupt_stack;
#else
      int_stk = zzuu.interrupt_stack;
#endif
#else
      int_stk = fmi_get_zzzu_interrupt_stack(DEREF_VAL(pcb, cpu_id));
#endif      
      /*
      ** Top is just outside the stack, so it may point one byte after
      ** the end of the pool fragment.
      */
      if (!(is_syspool_pointer(int_stk) &&
	    is_syspool_pointer((char *)DEREF_PTR(int_stk, top) - POOL_ALIGNMENT) &&
	    is_syspool_pointer(DEREF_PTR(int_stk, limit))))
        {
	  return EFMI_KERNEL_CORRUPTED;
        }
      
      info.process_stack.top = (U32) DEREF_PTR(int_stk, top);
      info.process_stack.limit = (U32) DEREF_PTR(int_stk, limit);
#if !defined (HOST_TOOL_ENVIRONMENT)      
      info.process_stack.no_endmark = ((unsigned char) END_MARK) !=
	*(unsigned char *) int_stk->limit;
#else
      info.process_stack.no_endmark = ((unsigned char) END_MARK) !=
        fmi_read_u8(DEREF_VAL(int_stk, limit));
#endif      
    }

  if(info.status & OS_RECEIVE)
    {
      /* Modify process status if we are in a remote system call. */
      if(DEREF_VAL(pcb, in_rem_call))
	info.status = (OSADDRESS)
	  ((info.status & ~OS_RECEIVE) | OS_REMOTE);
	      
      /*
       * About PAGING .text and/or MAS mapped data:
       * 1) We can not read the wanted array without risc of a page fault if
       * paging is enabled. When paging, we must return an emty wanted array.
       * 2) Sigselect in MAS .data/.bss may cause exception, so if MAS is
       * enabled, then we can only get sigselect arrays from the currently
       * selected segement, or a non mas_mapped one! 
       */
      if(check_word_pointer(DEREF_PTR(pcb, wanted)) == EFMI_OK && 0 == READ_VAL(zzuu, pagesize)) 
	{
	  PROCESS segpid;
	  struct segcb *segcb;

          SET_TYPE(segcb, segcb);

	  /* The _block struct is implicitely checked by check_pcb(). */
	  segpid = DEREF_VAL_TYPE(struct _block, DEREF_PTR(pcb, block), segpid);
	  segcb = segpid ? (struct segcb*)pid_to_pcb(segpid) :
	    READ_PTR(zzosvarp, syseg_varp);
	  if((EFMI_OK == check_segcb(segcb) && !DEREF_VAL(segcb, mas_mapped)) ||
	     (EFMI_OK == check_process(zzcur_p) &&
	      DEREF_VAL_TYPE(struct _pcb, zzcur_p, segid) == DEREF_VAL_TYPE(struct _block, DEREF_PTR(pcb, block), segid)))
	    {
	      /* Either the current segment or another non-mas_mapped one. */
#if !defined (HOST_TOOL_ENVIRONMENT)
              long count = (long) *pcb->wanted;               
#else
              uint32_t addr = DEREF_VAL(pcb, wanted);
              long count = fmi_read_u32(addr);
#endif              
	      size_t size;
	      int neg = 0;

	      if (count < 0)
		{
		  count = -count;
		  neg = 1;
		}
	      size = (size_t) (count + 1) * sizeof(SIGSELECT);
	      if (size > sizeof(info.wanted))
		{
		  size = sizeof(info.wanted);
		  count = size / sizeof(SIGSELECT) - 1;
		}
#if !defined (HOST_TOOL_ENVIRONMENT)
	      memcpy(info.wanted, pcb->wanted, size);
#else
              {
                int i;
                for (i = 0; i < count; ++i)
                  {
                    info.wanted[i] = fmi_read_u32(addr + i * 4);
                  }
              }
#endif              
	      info.wanted[0] = neg ? -count : count;
	    }
	}
    }
  
  memcpy(info_p, &info, MIN(size, sizeof(info)));
  return EFMI_OK;
} /* get_process_info */

/* with ERROR_CHECKS 0, get_watermark will report full stacks, as there is 
** no endmark pattern.
*/
static EfmiStatus
get_watermark_at(const char *sp, const char *top, U32 *watermark)
{
  char end_mark = END_MARK;
  
#if !defined (HOST_TOOL_ENVIRONMENT)  
  while(sp < top && *sp == end_mark)
#else
  while ((sp < top) && ((char)fmi_read_u8((uint32_t)sp) == end_mark))
#endif
    {
      sp++;
    } /* while */
  
  *watermark = top - sp;
  return EFMI_OK;
} /* get_watermark_at */

static EfmiStatus
get_watermark(const struct stkcb *stack, U32 *watermark)
{
  SET_TYPE(stkcb, stack);
  
  if(NULL == stack)
    {
      *watermark = 0;
      return EFMI_OK;
    }
  
  if(DEREF_VAL(stack, limit) > DEREF_VAL(stack, top))
    {
      return EFMI_UNAVAILABLE;
    }

  return get_watermark_at(DEREF_PTR(stack, limit), DEREF_PTR(stack, top), watermark);
} /* get_watermark */

static EfmiStatus
get_queue_count(const struct _pcb *pcb, U32 *queue_count)
{
  EfmiStatus		status = EFMI_OK;
  U32			count = 0;
  const struct _signal	*sigadm;

  SET_TYPE(_pcb, pcb);
  SET_TYPE(_signal, sigadm);
  
  sigadm = DEREF_PTR(pcb, mailbox);

  while(sigadm)
    {
      if (MEMORY_MODEL > 0)
        {
          if((status = check_syspool_pointer(sigadm)) != EFMI_OK)
            {
              return status;
            }
        }
      else
        {
          if((status = check_pool_pointer(sigadm, NULL)) != EFMI_OK)
            {
              return status;
            }
        }
      if (ERROR_CHECKS > 0)
      {
        if(DEREF_VAL(sigadm, magic_signo) != MAGIC_SIGNO)
          {
            return EFMI_KERNEL_CORRUPTED;
          }
      }
      count++;
      sigadm = DEREF_PTR(sigadm, next);
    } /* while */
      
  *queue_count = count;
  return status;
} /* get_queue_count */

/*
** Retrieve pool information
*/
static EfmiStatus
get_pool_info(struct PoolCursor *cursor, /* May be NULL */
	      PROCESS pool, /* Used when cursor is NULL */
	      struct EfmiPoolInfo   *info_p,
	      size_t size)
{
  EfmiStatus			status;
  const struct poolcb		*poolcb;
  int				sizei;
  struct EfmiBufSizeInfo	*size_info;
  const struct pool_list_entry	*size_cb;
  const struct _signal		*sigadm;
  const struct stkcb		*stkcb;
  /* Don't want this lagre chunk on the stack. We are single threaded here. */
#if !defined(HOST_TOOL_ENVIRONMENT)  
  static struct EfmiPoolInfo	info;
#else
  struct EfmiPoolInfo		info;
#endif
  
  SET_TYPE(poolcb, poolcb);
  SET_TYPE(pool_list_entry, size_cb);
  SET_TYPE(_signal, sigadm);
  SET_TYPE(stkcb, stkcb);
  
  if((status = check_kernel()) != EFMI_OK)
    {
      return status;		/* Mission failed */
    }
  
  if(cursor)
    {
      for(poolcb = cursor->next; ; poolcb = DEREF_PTR(poolcb, next))
	{
	  if(!poolcb)
	    {
	      return EFMI_END_OF_LIST;
	    }
	  if(EFMI_OK != (status = check_syspool_pointer(poolcb)))
	    {
	      return status;
	    }
	  if(!(DEREF_VAL(poolcb, magic) & MAGIC_KILLED))
	    {
	      break;		/* Still alive. */
	    }
	} /* for */
    }
  else
    {
      if(READ_VAL(zzuu, next_free_pix) <= (PCB_INDEX_T) pool)
	{
	  return EFMI_UNAVAILABLE;
	}
      poolcb = (struct poolcb *) pid_to_pcb(pool);
    }

  if((status = check_poolcb(poolcb)) != EFMI_OK)
    {
      return status;		/* Mission failed */
    }
  
  if(cursor)
    {
      cursor->next = DEREF_PTR(poolcb, next);
    }
  else if(DEREF_VAL(poolcb, instance) != pid_to_instance(pool))
    {
      return EFMI_UNAVAILABLE;	/* Mission failed */
    }
  
  info.pool_id = pcb_to_pid(poolcb);
  info.fragment_count = (DEREF_VAL(poolcb, extend_count)) + 1;
  info.never_used = DEREF_VAL(poolcb, pool_hi_p) - DEREF_VAL(poolcb, pool_low_p);
  info.segment = DEREF_VAL(poolcb, segpid);

  for(sizei = 0; sizei < EFMI_MAX_SIGNAL_BUFFER_SIZE_COUNT; sizei++)
    {
      size_info = &info.signal_buf_sizes[sizei];
#if !defined (HOST_TOOL_ENVIRONMENT)      
      size_cb = &poolcb->signal_list[sizei];
#else
      {
        uint32_t type_size = fmi_get_size("struct poolcb", "signal_list") / EFMI_MAX_SIGNAL_BUFFER_SIZE_COUNT;
        size_cb = (struct pool_list_entry *)(fmi_get_addr("struct poolcb", (uint32_t)poolcb, "signal_list") + sizei * type_size);
      }
#endif      
      size_info->configured_size = DEREF_VAL(size_cb, size);
      size_info->allocated = DEREF_VAL(size_cb, alloc_count);
      
      size_info->free = 0;
      for(sigadm = (struct _signal *) DEREF_PTR(size_cb, hd); sigadm;
	  sigadm = DEREF_PTR(sigadm, next))
	{
          if (MEMORY_MODEL > 0)
            {
              if (ERROR_CHECKS > 0)
                {
                  if(check_syspool_pointer(sigadm) != EFMI_OK ||
                     DEREF_PTR(sigadm, next) == sigadm ||
                     DEREF_VAL(sigadm, magic_signo) != MAGIC_SIGNO)
                    {
                      return EFMI_KERNEL_CORRUPTED;
                    }
                }
              else
                {
                  if(check_syspool_pointer(sigadm) != EFMI_OK ||
                     DEREF_PTR(sigadm, next) == sigadm)
                    {
                      return EFMI_KERNEL_CORRUPTED;
                    }
                }
            }
          else
            {
              /* MEMORY_MODEL 0 */
              if (ERROR_CHECKS > 0)
                {
                  if(check_pool_pointer(sigadm, poolcb) != EFMI_OK ||
                     DEREF_PTR(sigadm, next) == sigadm ||
                     DEREF_VAL(sigadm, magic_signo) != MAGIC_SIGNO)
                    {
                      if(poolcb == READ_PTR(zzosvarp, syspool_varp))
                        {
                          return EFMI_KERNEL_CORRUPTED;
                        }
                      return EFMI_POOL_CORRUPTED;
                    }
                }
              else
                {
                  if(check_pool_pointer(sigadm, poolcb) != EFMI_OK ||
                     DEREF_PTR(sigadm, next) == sigadm)
                    {
                      if(poolcb == READ_PTR(zzosvarp, syspool_varp))
                        {
                          return EFMI_KERNEL_CORRUPTED;
                        }
                      return EFMI_POOL_CORRUPTED;
                    }
                }
            }
	  size_info->free++;
	} /* for */
    } /* for */
  
  for(sizei = 0; sizei < EFMI_MAX_STACK_BUFFER_SIZE_COUNT; sizei++)
    {
      size_info = &info.stack_buf_sizes[sizei];
#if !defined (HOST_TOOL_ENVIRONMENT)      
      size_cb = &poolcb->stack_list[sizei];
#else
      {
        uint32_t type_size = fmi_get_size("struct poolcb", "stack_list") / EFMI_MAX_STACK_BUFFER_SIZE_COUNT;
        size_cb = (struct pool_list_entry *)(fmi_get_addr("struct poolcb", (uint32_t)poolcb, "stack_list") + sizei * type_size);
      }
#endif      
      size_info->configured_size = DEREF_VAL(size_cb, size);
      size_info->allocated = DEREF_VAL(size_cb, alloc_count);
      
      size_info->free = 0;
      for(stkcb = (struct stkcb *) DEREF_PTR(size_cb, hd); stkcb;
          stkcb = DEREF_PTR(stkcb, next))
	{
	  if(check_syspool_pointer(stkcb) != EFMI_OK ||
	     DEREF_PTR(stkcb, next) == stkcb)
	    {
	      return EFMI_KERNEL_CORRUPTED;
	    }
	  size_info->free++;
	} /* for */
    } /* for */

  info.stack_alignment = READ_VAL(zzuu, pool_alignment);
  info.stack_admin_size = 1;
  info.stack_internal_admin_size = 0;
  info.signal_alignment = READ_VAL(zzuu, pool_alignment);
#if !defined (HOST_TOOL_ENVIRONMENT)  
  info.signal_admin_size = 1 + sizeof(struct _signal);
#else
  info.signal_admin_size = 1 + fmi_get_size("struct _signal", "#");
#endif  
  info.signal_internal_admin_size = 0;
  
  memcpy(info_p, &info, MIN(size, sizeof(info)));
  return EFMI_OK;
} /* get_pool_info */

/*
** Retrieve pool fragment information
*/
static EfmiStatus
get_fragment_info(struct FragmentCursor *cursor,
	      struct EfmiFragmentInfo   *info_p,
	      size_t size)
{
  EfmiStatus			status;
  struct EfmiFragmentInfo	info;
  const struct pool_extension	*frag;

  SET_TYPE(pool_extension, frag);
  
  if((status = check_kernel()) != EFMI_OK)
    {
      return status;		/* Mission failed */
    }
  
  if((status = check_poolcb(cursor->poolcb)) != EFMI_OK)
    {
      return status;		/* Mission failed */
    }

  frag = cursor->next;
  if(!frag)
    {
      return EFMI_END_OF_LIST;
    }

  if((status = check_syspool_pointer(frag)) != EFMI_OK)
    {
      return status;		/* Mission failed */
    }

  if(DEREF_VAL(frag, magic_frag) != MAGIC_POOL)
    {
      return EFMI_KERNEL_CORRUPTED; /* Mission failed */
    }

  cursor->next = DEREF_PTR(frag, frag_list);

  info.base_address = DEREF_VAL(frag, base);
  info.size = DEREF_VAL(frag, limit) - DEREF_VAL(frag, base);

  /* Current fragment uses break pointers in the poolcb */
  if(DEREF_PTR_TYPE(struct poolcb, cursor->poolcb, frag_list) == frag)
    {
      info.used_for_stacks = DEREF_VAL(frag, limit) - DEREF_VAL_TYPE(struct poolcb, cursor->poolcb, pool_hi_p);
      info.used_for_signals = DEREF_VAL_TYPE(struct poolcb, cursor->poolcb, pool_low_p) - DEREF_VAL(frag, base);
    }
  else
    {
      info.used_for_stacks = DEREF_VAL(frag, limit) - DEREF_VAL(frag, pool_hi_p);
      info.used_for_signals = DEREF_VAL(frag, pool_low_p) - DEREF_VAL(frag, base);
    }
  
  memcpy(info_p, &info, MIN(size, sizeof(info)));
  return EFMI_OK;
} /* get_fragment_info */

/*
** Retrieve signal information. 
*/
static EfmiStatus
get_signal_info(struct SignalCursor *cursor,
		struct EfmiSignalInfo   *info_p,
		size_t size)
{
  EfmiStatus			status;
  U32				sig_status = EFMI_SS_VALID;
  PROCESS			owner;
  struct EfmiSignalInfo		info;
  const struct _signal		*sigadm;
  const struct _signal		*usigadm;
  const SIGSELECT		*sig_no_ptr;
  const struct poolcb		*poolcb = NULL;	/* NULL for signal queue op. */

  SET_TYPE(_signal, sigadm);
  SET_TYPE(_signal, usigadm);
  
  if((status = check_kernel()) != EFMI_OK)
    {
      return status;		/* Mission failed */
    }

  if(cursor->traverse_pool)
    {
      poolcb = cursor->poolcb;
      if((status = check_poolcb(poolcb)) != EFMI_OK)
	{
	  return status;		/* Mission failed */
	}
    }

  sigadm = cursor->next;
  if(!sigadm)
    {
      return EFMI_END_OF_LIST;
    }

  if (MEMORY_MODEL > 0)
    {
      if((status = check_syspool_pointer(sigadm)) != EFMI_OK)
        {
          return status;		/* Mission failed */
        }

      usigadm = - 1 + (struct _signal *) DEREF_PTR(sigadm, udata);
      if(
         (ERROR_CHECKS == 0 || DEREF_VAL(sigadm, magic_signo) != MAGIC_SIGNO) ||
         check_pool_pointer(usigadm, poolcb) != EFMI_OK)
        {
          return EFMI_KERNEL_CORRUPTED; /* Mission failed */
        }

      if(
         (ERROR_CHECKS == 0 || DEREF_VAL(usigadm, magic_signo) != MAGIC_SIGNO) ||
         (struct _signal *) DEREF_PTR(usigadm, udata) != sigadm)
        {
          sig_status = EFMI_SS_BROKEN_SIGADM;
        }
    }
  else
    {
      if((status = check_pool_pointer(sigadm, poolcb)) != EFMI_OK)
        {
          return status;		/* Mission failed */
        }
      if (ERROR_CHECKS > 0)
        {
          if(DEREF_VAL(sigadm, magic_signo) != MAGIC_SIGNO)
            {
              sig_status = EFMI_SS_BROKEN_SIGADM;
            }
        }
      usigadm = sigadm;
    }
  
  cursor->next = cursor->traverse_pool ? DEREF_PTR(sigadm, siglist) : DEREF_PTR(sigadm, next);

  sig_no_ptr = (const SIGSELECT *) (usigadm + 1);
  if (ERROR_CHECKS > 0)
    {
      if(sig_status == EFMI_SS_VALID)
        {
#if !defined (HOST_TOOL_ENVIRONMENT)          
          uint8_t end_mark = *(sigadm->size + (const U8 *) sig_no_ptr);
#else
          uint8_t end_mark = fmi_read_u8((DEREF_VAL(sigadm, size) + (uint32_t)sig_no_ptr));
#endif
          if (end_mark != END_MARK)
            {
              sig_status = EFMI_SS_NO_ENDMARK;
            }
        }
    }

  owner = 0;
  if(DEREF_VAL(sigadm, owner) >= READ_VAL(zzuu, next_free_pix))
    {
      sig_status = EFMI_SS_BROKEN_SIGADM;
    }
  else
    {
#if !defined (HOST_TOOL_ENVIRONMENT)        
      owner = sigadm->owner ? zzuu.pix_list[sigadm->owner]->pid : 0;
#else
      {
        uint32_t sigadm_owner = DEREF_VAL(sigadm, owner);
        if (sigadm_owner)
          {
            uint32_t addr = READ_VAL(zzuu, pix_list);
            struct _pcb *pcb = (struct _pcb *)fmi_read_u32(addr + sigadm_owner * 4);
            owner = DEREF_VAL_TYPE(struct _pcb, pcb, pid);
          }
        else          
          owner = 0;
      }
#endif      
    }

  info.signal_pointer = (U32) sig_no_ptr;
#if !defined (HOST_TOOL_ENVIRONMENT)  
  info.sig_no = *sig_no_ptr;
#else
  info.sig_no = fmi_read_u32((uint32_t)sig_no_ptr);
#endif  
  info.owner = owner;
  info.sender_pid = DEREF_VAL(sigadm, sender_id) ? DEREF_VAL(sigadm, sender_id) : owner;
  info.addressee_pid = DEREF_VAL(sigadm, dest_id) ? DEREF_VAL(sigadm, dest_id) : owner;
  info.sig_size = DEREF_VAL(sigadm, size);
#if !defined (HOST_TOOL_ENVIRONMENT)  
  info.size_in_pool = sigadm->buffer_head->size + sizeof(struct _signal) + 1;
#else
  info.size_in_pool = DEREF_VAL_TYPE(struct pool_list_entry, DEREF_PTR(sigadm, buffer_head), size) + 33;
#endif
  info.sig_status = sig_status;
  
  memcpy(info_p, &info, MIN(size, sizeof(info)));
  return EFMI_OK;
} /* get_signal_info */

/* Get an environment variable name and value for a block or process. */
static EfmiStatus
get_env_var(struct EnvVarCursor *cursor, char *name, char *value)
{
  EfmiStatus		status;
  const struct env_buf	*env_buf = cursor->next;

  SET_TYPE(env_buf, env_buf);
  
 restart:
  if(!env_buf)
    {
      return EFMI_END_OF_LIST;
    }
  if((status = check_syspool_pointer(env_buf)) != EFMI_OK)
    {
      return status;
    }

  /* If the variable has been deleted, then skip to the next! */
  if(!DEREF_PTR(env_buf, value))
    {
      env_buf = DEREF_PTR(env_buf, next);
      goto restart;
    }
  
  if((status = check_syspool_pointer(DEREF_PTR(env_buf, value))) != EFMI_OK)
    {
      return status;
    }
#if !defined (HOST_TOOL_ENVIRONMENT)  
  if(strncpy_truncated(name, env_buf->name, cursor->max_name_len))
#else
  if(strncpy_truncated(name, (char *)fmi_get_addr("struct env_buf", (uint32_t)env_buf, "name"), cursor->max_name_len))
#endif    
    {
      status = EFMI_RESULT_TRUNCATED;
    }
  if(strncpy_truncated(value, DEREF_PTR(env_buf, value), cursor->max_value_len))
    {
      status = EFMI_RESULT_TRUNCATED;
    }
  
  cursor->next = DEREF_PTR(env_buf, next);
  return status;
} /* get_env_var */

static EfmiStatus
get_process_by_name(const char *name,
		    struct ProcNameCursor *nc,
		    PROCESS *pid)
{
  EfmiStatus			status;
  PCB_INDEX_T			next;
  const struct _pcb		*pcb;

  SET_TYPE(_pcb, pcb);
  
  status = check_kernel();
  if(status != EFMI_OK)
    {
      return status;
    }
  if(!name)
    {
      return EFMI_END_OF_LIST;
    }

  for(next = nc->next; next < READ_VAL(zzuu, next_free_pix); next++)
    {
#if !defined (HOST_TOOL_ENVIRONMENT)      
      pcb = zzuu.pix_list[next];
#else
      uint32_t addr = READ_VAL(zzuu, pix_list);
      pcb = (struct _pcb *)fmi_read_u32(addr + next * 4);      
#endif      
      if((status = check_syspool_pointer(pcb)) != EFMI_OK)
	{
	  return status;
	}
      
      if(DEREF_VAL(pcb, magic) == MAGIC_PCB && DEREF_PTR(pcb, name))
	{
	  if((status = check_syspool_pointer(DEREF_PTR(pcb, name))) != EFMI_OK)
	    {
	      /* We skip idle and other kernel processes here.
	      ** They may have name strings that are not signals.
	      */
	      ;			
	    }
	  else
	    {
              int match;
#if !defined (HOST_TOOL_ENVIRONMENT)              
	      match = (0 == strcmp(pcb->name, name));
#else
              {
                int i = 0;
                uint32_t addr = DEREF_VAL(pcb, name);
                char c = fmi_read_u8(addr);
                while ((c != '\0') && (name[i] != '\0') && (c == name[i]))
                  {
                    addr++;
                    i++;
                    c = fmi_read_u8(addr);
                  }
                /* If both are equal (0) now, then we have a match. */
                match = (c == name[i]);
              }
#endif
              if (match)
		{
		  *pid = DEREF_VAL(pcb, pid);
		  nc->next = next + 1;
		  return EFMI_OK;
		}
	    }
	}
    } /* for */
  return EFMI_END_OF_LIST;

} /* get_process_by_name */

/* Helper function that differs between target and host EFMI. */
static const struct _signal *
get_sigadm_list_from_poolcb(const struct poolcb *pool)
{
#if !defined (HOST_TOOL_ENVIRONMENT)
  return pool->firstsig.siglist;
#else  
  uint32_t addr = fmi_get_addr("struct poolcb", (uint32_t)pool, "firstsig");
  return (const struct _signal *)fmi_read("struct short_sigadm", addr, "siglist");
#endif  
}

/*
*****************************************************************************
** PUBLIC FUNCTIONS
*****************************************************************************
*/

/*
** Check if the ose kernel is detected and such a state that we can
** retrieve information
**
** Returns: EFMI_OK | EFMI_KERNEL_CORRUPTED | EFMI_KERNEL_UNINITIATED
**
*/
EfmiStatus 
ose_efmi_get_status
ARGLIST0(void)
{
  EfmiStatus		status;
  U32			lock;

  LOCK_SAVE(lock);
  ENTER_EFMI(i);
  status = check_kernel();
  EXIT_EFMI();
  LOCK_RESTORE(lock);
  
  return status;
} /* ose_efmi_get_status */

/*
** Retrieve system information
*/
EfmiStatus
ose_efmi_get_system_info
ARGLIST(struct EfmiSystemInfo *system_info,
        size_t system_info_size)
{
  EfmiStatus		status;
  U32			lock;

  LOCK_SAVE(lock);
  CALLTRACE("ose_efmi_get_system_info");
  ENTER_EFMI(i);
  status = get_system_info(system_info, system_info_size);
  EXIT_EFMI();
  LOCK_RESTORE(lock);
  return status;
} /* ose_efmi_get_system_info */

/*
** Retrieve segment information
*/
EfmiStatus
ose_efmi_get_segment_info
ARGLIST(PROCESS segment,
        struct EfmiSegmentInfo   *pri,
        size_t size)
{
  EfmiStatus		status;
  U32			lock;

  LOCK_SAVE(lock);
  CALLTRACE("ose_efmi_get_segment_info");
  ENTER_EFMI(i);
  status = get_segment_info(NULL, segment, pri, size);
  EXIT_EFMI();
  LOCK_RESTORE(lock);
  return status;
} /* ose_efmi_get_segment_info */

/*
** Retrieve block information
*/
EfmiStatus
ose_efmi_get_block_info
ARGLIST(PROCESS block_id,
        struct EfmiBlockInfo   *bi,
        size_t size)
{
  EfmiStatus		status;
  U32			lock;

  LOCK_SAVE(lock);
  CALLTRACE("ose_efmi_get_block_info");
  ENTER_EFMI(i);
  status = get_block_info(NULL, block_id, bi, size);
  EXIT_EFMI();
  LOCK_RESTORE(lock);
  return status;
} /* ose_efmi_get_block_info */

#if defined(HOST_TOOL_ENVIRONMENT)
int
efmi_get_block_info_internal(uint32_t block_id,
                             struct EfmiBlockInfo *bi,
                             size_t size)
{
  return get_block_info(NULL, block_id, bi, size);
}
#endif

/*
** Retrieve process information
*/
EfmiStatus
ose_efmi_get_process_info
ARGLIST(PROCESS pid,
        struct EfmiProcessInfo  *proc_info,
        size_t size)
{
  EfmiStatus		status;
  U32			lock;

  LOCK_SAVE(lock);
  CALLTRACE("ose_efmi_get_process_info");
  ENTER_EFMI(i);
  status = get_process_info(NULL, pid, proc_info, size);
  EXIT_EFMI();
  LOCK_RESTORE(lock);
  return status;
} /* ose_efmi_get_process_info */

#if defined(HOST_TOOL_ENVIRONMENT)
int
efmi_get_process_info_internal(uint32_t pid,
                               struct EfmiProcessInfo *proc_info,
                               size_t size)
{
  return get_process_info(NULL, pid, proc_info, size);
}
#endif

EfmiStatus
ose_efmi_get_process_statistics
ARGLIST(PROCESS pid,
        enum EfmiProcessTag tag,
        struct EfmiProcessStatistics *statistics)
{
  EfmiStatus		status;
  U32			lock;
  const struct _pcb	*pcb;

  SET_TYPE(_pcb, pcb);
  
  LOCK_SAVE(lock);
  CALLTRACE("ose_efmi_get_process_statistics");
  ENTER_EFMI(i);
  status = check_kernel();
  if(status == EFMI_OK)
    {
      status = check_pix_identity(pid);
    }
  if(status == EFMI_OK)
    {
      pcb = pid_to_pcb(pid);
      status = check_process(pcb);
    }
  if(status == EFMI_OK)
    {
      switch(tag)
	{
	case EFMI_SIG_CNT_IN_Q:	/* Number of signals in the receive queue */
	  status = get_queue_count(pcb, &statistics->value);
	  break;
	  
	case EFMI_PROCESS_STACK_USAGE: /* Max # bytes used in main stack */
#if !defined(softOSE) || defined(HOST_TOOL_ENVIRONMENT)
	  status = get_watermark(DEREF_PTR(pcb, stack), &statistics->value);
#else          
          status = EFMI_UNAVAILABLE;
#endif          
	  break;
	  
	case EFMI_SUPERVISOR_STACK_USAGE: /* Max # bytes used in supervisor 
					   * stack for a user mode process */
	  {
	    U32 limit;
	    status = EFMI_UNAVAILABLE; 
#if !defined(softOSE) || defined(HOST_TOOL_ENVIRONMENT)
	    if(!DEREF_VAL_TYPE(struct _block, DEREF_PTR(pcb, block), supervisor_mode) &&
	       DEREF_VAL(pcb, priority) <= 32 &&
	       0 == zzget_register(DEREF_PTR(pcb, uarea), ZZREG_SSLIMIT, &limit,
				   is_syspool_pointer))
	    {
	      char *top;
	      OSADDRESS utop;

	      /* Calculate the top as all the TIs do! */
	      top = (char *) (limit +
			      (DEREF_VAL_TYPE(struct sysctlblk, READ_PTR(zzuu, sysctlblk), superv_stack_size) & ~0xf));
	      /*
	       * Top may point after the end of the pool, so we decrement
	       * with POOL_ALIGNMENT in the pointer check.
	       */
	      utop = ((OSADDRESS)top - 1)/* & ~(POOL_ALIGNMENT - 1)*/;
	      status = check_syspool_pointer((void *)utop);
	      if(status == EFMI_OK)
		{
		  status = get_watermark_at((char *)limit, top, &statistics->value);
		}
	    }
#endif
	  }
	  break;
	  
	default:
	  status = EFMI_UNAVAILABLE;
	  break;
	} /* switch */
    }
  EXIT_EFMI();
  LOCK_RESTORE(lock);
  return status;
} /* ose_efmi_get_process_statistics */

/*
** Retrieve pool information
*/
EfmiStatus
ose_efmi_get_pool_info
ARGLIST(PROCESS pool_id,
        struct EfmiPoolInfo   *pool_info,
        size_t size)
{
  EfmiStatus		status;
  U32			lock;

  LOCK_SAVE(lock);
  CALLTRACE("ose_efmi_get_pool_info");
  ENTER_EFMI(i);
  status = get_pool_info(NULL, pool_id, pool_info, size);
  EXIT_EFMI();
  LOCK_RESTORE(lock);
  return status;
} /* ose_efmi_get_pool_info */

/*
** Retrieve information on a signal.
*/
EfmiStatus
ose_efmi_get_signal_info
ARGLIST(union SIGNAL *signal,	
        struct EfmiSignalInfo *sig_info,
        size_t size)
{
  U32				lock;
  EfmiStatus			status;
  U32				sig_status;
  PROCESS			owner;
  struct EfmiSignalInfo		info;
  const struct _signal		*sigadm = NULL;
  const struct _signal		*usigadm = NULL;
  const struct poolcb		*poolcb;
  union SIGNAL			*pool_aligned_signal;

  SET_TYPE(_signal, sigadm);
  SET_TYPE(_signal, usigadm);
  
  sig_status = EFMI_SS_WILD_POINTER_OUTSIDE_POOL;

  memset(&info, 0, sizeof(info)); /* safe values */
  LOCK_SAVE(lock);
  CALLTRACE("ose_efmi_get_signal_info");
  ENTER_EFMI(i);
  
  status = check_kernel();

  if(status != EFMI_OK)
    goto done;			/* >>kernel NOK */

  if(signal == NIL)
    goto done;			/* >>NIL pointer */

  pool_aligned_signal = (union SIGNAL *)
    (/*(~(POOL_ALIGNMENT - 1)) &*/ (OSADDRESS)signal);
  status = find_pool(pool_aligned_signal, &poolcb);

  if(status == EFMI_UNAVAILABLE)
    {
      status = EFMI_OK;
      goto done;		/* >>wild outside */
    }
  
  if(status != EFMI_OK)
    goto done;			/* >>kernel NOK */

  /* Now we know that the pointer is properly aligned and inside the pool. */
  sig_status = EFMI_SS_WILD_POINTER_INSIDE_POOL;

  if(pool_aligned_signal != signal)
    goto done;			/* >>wild inside */


  /* Check if user sigadm reached via the pool siglist. */
  usigadm = - 1 + (struct _signal *) signal;
    
  if (MEMORY_MODEL > 0)
    {
      /* Memory model 1 */
      for(sigadm = get_sigadm_list_from_poolcb(poolcb); 1; sigadm = DEREF_PTR(sigadm, siglist))
        {
          const struct _signal *tmp_usigadm;
          
          if(!sigadm)
            break;			/* >>wild inside */
      
          status = check_syspool_pointer(sigadm);
          if(EFMI_OK != status)
            break;			/* >>kernel NOK */

          if (ERROR_CHECKS > 0)
            {
              if(DEREF_VAL(sigadm, magic_signo) != MAGIC_SIGNO)
                {
                  status =  EFMI_KERNEL_CORRUPTED;
                  break;		/* >>kernel NOK */
                }
            }

          tmp_usigadm = -1 + (struct _signal*)DEREF_PTR(sigadm, udata);
          if(tmp_usigadm == usigadm)
            {

              if((struct _signal*)DEREF_PTR(usigadm, udata) != sigadm
                 || (ERROR_CHECKS > 0 && DEREF_VAL(usigadm, magic_signo) != MAGIC_SIGNO)
                 )
                {
                  sig_status = EFMI_SS_BROKEN_SIGADM;
                  break;		/* >>broken usigadm */
                }

              /* ENDMARK checked below. */
              sig_status = EFMI_SS_VALID;
              break;		/* >>valid */
            } /* signal found */

          if(sigadm == DEREF_PTR(sigadm, siglist))
            {
              status =  EFMI_KERNEL_CORRUPTED;
              break;		/* >>kernel NOK */
            }
        } /* for */
    }
  else
    {
      /* Memory model 0 */

      sigadm = get_sigadm_list_from_poolcb(poolcb);
      if(!sigadm)
        goto done;			/* >>wild inside */
  
      for(;;)
        {
          status = check_pool_pointer(sigadm, poolcb);
          if(EFMI_UNAVAILABLE == status)
            {
              status = (poolcb == READ_PTR(zzosvarp, syspool_varp)) ?
                EFMI_KERNEL_CORRUPTED : EFMI_POOL_CORRUPTED;
              break;		/* >>kernel/pool NOK */
            }
          if(EFMI_OK != status)
            break;			/* >>kernel NOK */

          if(sigadm == usigadm)
            {
              if (ERROR_CHECKS > 0)
                {
                  if(DEREF_VAL(sigadm, magic_signo) != MAGIC_SIGNO)
                    {
                      sig_status = EFMI_SS_BROKEN_SIGADM;
                      break;		/* >>broken sigadm */
                    }
                }
              /* ENDMARK checked below. */
              sig_status = EFMI_SS_VALID;
              break;		/* >>valid */
            } /* signal found */

          if(!DEREF_PTR(sigadm, siglist))
            break;			/* >>wild inside */
      
          /* The signal list must be in address order for MODEL 0! */
          if((struct _signal *)DEREF_PTR(sigadm, siglist) <= sigadm)
            {
              status = (poolcb == READ_PTR(zzosvarp, syspool_varp)) ?
                EFMI_KERNEL_CORRUPTED : EFMI_POOL_CORRUPTED;
              break;		/* >>kernel/pool NOK */
            }
          sigadm = DEREF_PTR(sigadm, siglist);
        } /* for */
  /* end MEMORY_MODEL 0. */
    }

  owner = 0;
  
  if(sig_status == EFMI_SS_VALID)
    {
      if(DEREF_VAL(sigadm, owner) >= READ_VAL(zzuu, next_free_pix))
	{
	  sig_status = EFMI_SS_BROKEN_SIGADM; /* >>broken sigadm */
	}
      else
	{
#if !defined (HOST_TOOL_ENVIRONMENT)        
          owner = sigadm->owner ? zzuu.pix_list[sigadm->owner]->pid : 0;
#else
          {
            uint32_t sigadm_owner = DEREF_VAL(sigadm, owner);
            if (sigadm_owner)
              {
                uint32_t addr = READ_VAL(zzuu, pix_list);
                struct _pcb *pcb = (struct _pcb *)fmi_read_u32(addr + sigadm_owner * 4);
                owner = DEREF_VAL_TYPE(struct _pcb, pcb, pid);
              }
            else          
              owner = 0;
          }
#endif                
	}
    }

  if(sig_status == EFMI_SS_VALID)
    {
#if !defined (HOST_TOOL_ENVIRONMENT)      
      info.sig_no = *(SIGSELECT *)signal;
#else
      info.sig_no = fmi_read_u32((uint32_t)signal);
#endif      
      info.owner = owner;
      info.sender_pid = DEREF_VAL(sigadm, sender_id) ? DEREF_VAL(sigadm, sender_id) : owner;
      info.addressee_pid = DEREF_VAL(sigadm, dest_id) ? DEREF_VAL(sigadm, dest_id) : owner;
      info.sig_size = DEREF_VAL(sigadm, size);
#if !defined (HOST_TOOL_ENVIRONMENT)  
  info.size_in_pool = sigadm->buffer_head->size + sizeof(struct _signal) + 1;
#else
  info.size_in_pool = DEREF_VAL_TYPE(struct pool_list_entry, DEREF_PTR(sigadm, buffer_head), size) + 33;
#endif
      if (ERROR_CHECKS > 0)
        {
          const U8 *endmark_addr = (DEREF_VAL(sigadm, size) + (const U8 *) signal);
#if !defined (HOST_TOOL_ENVIRONMENT)          
          if(END_MARK != *endmark_addr)
#else
          if (END_MARK != fmi_read_u8((uint32_t)endmark_addr))
#endif      
            {
              sig_status = EFMI_SS_NO_ENDMARK; /* >>no endmark */
            }
        }
    }
  

 done:
  info.signal_pointer = (U32) signal;
  info.sig_status = sig_status;
  
  memcpy(sig_info, &info, MIN(size, sizeof(info)));
  EXIT_EFMI();
  LOCK_RESTORE(lock);
  return status;
} /* ose_efmi_get_signal_info */

/* 
** Copy a memory area. 
** Copying is only performed if the source address range is known
** and considered safe by efmi. 
*/
EfmiStatus 
ose_efmi_get_mem
ARGLIST(PROCESS id, OSADDRESS from, void *to, OSADDRESS size)
{
  U32				lock;
  EfmiStatus			status;
  OSADDRESS			page, last_page;
  OSADDRESS			to_addr = (OSADDRESS)to;

  (void)id;
  
  LOCK_SAVE(lock);
  CALLTRACE("ose_efmi_get_mem");
  ENTER_EFMI(i);
  status = check_kernel();
  if(status == EFMI_OK)
    {
      /*
       * Check that to and from are nonzero and the ranges does not wrap.
       */
      if(!from || !to || from + size < from || to_addr + size < to_addr)
	{
	  status = EFMI_UNAVAILABLE;
	}
      else
	{
	  /* Check that each source page is present in a pool. */
	  page = from & ~(PAGE_SIZE - 1);
	  last_page = (from + size) & ~(PAGE_SIZE - 1);
	  for(;;)
	    {
#if !defined (HOST_TOOL_ENVIRONMENT)
              /* In a host environment we allow addresses outside pools. */
	      status = check_pool_pointer((void *)page, NULL);
	      if(status != EFMI_OK)
		{
		  status = EFMI_UNAVAILABLE;
		  break;	/* FAILS */
		}
#endif                  
	      if(page == last_page)
		{
#if !defined (HOST_TOOL_ENVIRONMENT)                  
		  memcpy(to, (void *)from, size);
#else
                  {
                    uint32_t idx;
                    for (idx = 0; idx < size; ++idx)
                      {
                        ((uint8_t *)to)[idx] = fmi_read_u8(from + idx);
                      }
                  }
#endif                  
		  break;	/* SUCCESS */
		}
	      page = page + PAGE_SIZE;
	    } /* for */
	}
    }
  EXIT_EFMI();
  LOCK_RESTORE(lock);
  return status;
} /* ose_efmi_get_mem */

/*
** Get number of cpus in an SMP system.
*/

EfmiStatus 
ose_efmi_get_smp_num_cpus
ARGLIST(U32 *num_cpus)
{
  ENTER_EFMI(i);
  *num_cpus = get_smp_num_cpus();
  EXIT_EFMI();
  return EFMI_OK;
}

/*
** Example on how to use iterator functions.
**  
**  struct EfmiBlockCursor pc;
**  struct EfmiBlockInfo   block_info;
**
**  EfmiStatus rv;
**
**  rv = ose_efmi_get_first_block(segment_id, &pc, &block_info,
**                                sizeof(block_info);
**  while(rv == EFMI_OK) {
**      do_something_with_block_info(&block_info);
**      rv = ose_efmi_get_next_block(&pc, &block_info);
**  }
**
**  if(rv != EFMI_END_OF_LIST) {
**     fmm_printf("Something went wrong when retrieving block information.\n");
**  }
*/

/*
** Retrieve the current context list per smp cpu.
*/

static EfmiStatus
get_current(cpuid_t cpu_id, struct EfmiCurrent *current, size_t size)
{
  EfmiStatus		status;
  struct EfmiCurrent	cur;
  const struct _block	*block;
  U32			priority;
  const struct _pcb	*pcb;
  const struct _pcb	*cur_ni;

  SET_TYPE(_pcb, pcb);
  SET_TYPE(_block, block);
  SET_TYPE(_pcb, cur_ni);
  
  if(!READ_VAL(zzuu, kernel_initialized))
    {
      status = EFMI_KERNEL_UNINITIATED;
    }
  else if(cpu_id >= get_smp_num_cpus())
    {
      status = EFMI_UNAVAILABLE;
    }
  else
    {
      pcb = zzcur_p_x(cpu_id);
      status = check_pix_pointer(pcb, MAGIC_PCB);

      if(status == EFMI_OK)
	{
	  block = DEREF_PTR(pcb, block);
	  status = check_pix_pointer(block, MAGIC_BCB);
	}
    }
  if(status != EFMI_OK)
    {
      ERRTRACE(status, 0);
      return status;		/* Mission failed */
    }
  cur.process = DEREF_VAL(pcb, pid);		/* ose.h PROCESS. */
  cur.block = pcb_to_pid(block);	/* ose.h PROCESS. */
  cur.segment = DEREF_VAL(block, segpid);		/* ose.h PROCESS. */
  cur.segid = DEREF_VAL(block, segid);		/* ose.h OSSEGMENT. */
  cur.ni_process = cur.process;		/* ose.h PROCESS. */
  cur.cpu_id = DEREF_VAL(pcb, cpu_id);
  cur_ni = zzcur_ni_x(cpu_id);
  if(cur_ni != pcb)
    {
      /*
      ** Current process is an interrupt process.
      ** Report the idle process as ni_process if we really was idle.
      ** If a context switch is pending, report last_cur_ni instead.
      ** This scheme may improve pin-pointing bandits in a software
      ** watchdog, with a cyclic context switch scenarie.
      **
      ** So, how can we detect that we are idle or just switching process?
      ** zzdp_look_aside shows if there are any ready prioritized
      ** processes. We may ignore background processes, as we try
      ** to solve a watchdog problem here, and the background level
      ** is not suitable for watchdog supervision anyway.
      */
      if(cur_ni == zzidle_pcb_x(cpu_id) &&
	 dp_look_aside_x(cpu_id))
	{
	  /* We are not idle, just swapping to a new conext. */
	  cur_ni = zzlast_cur_ni_x(cpu_id);
	}
      
      if(EFMI_OK == check_pix_pointer(cur_ni, MAGIC_PCB))
	{
	  cur.ni_process = DEREF_VAL(cur_ni, pid); /* ose.h PROCESS. */
	}
    }
  
  /* Extract process type from process priority. */
  priority = DEREF_VAL(pcb, priority);
  cur.type = prio_to_type(priority);	/* ose.h enum PROCESS_TYPE. */
  memcpy(current, &cur, MIN(size, sizeof(cur)));
  return EFMI_OK;
} /* get_current */

EfmiStatus
ose_efmi_get_first_current
ARGLIST(struct EfmiCurrentCursor *cc,
        struct EfmiCurrent *current,
        size_t current_size)
{
  struct CurrentCursor		*cursor = (struct CurrentCursor *) cc;

  ENTER_EFMI(i);
  cursor->size = current_size;
  cursor->next = 0;
  cursor->num_cpus = get_smp_num_cpus();
  EXIT_EFMI();
#if !defined (HOST_TOOL_ENVIRONMENT)  
  return ose_efmi_get_next_current((struct EfmiCurrentCursor *)cursor,                                   
				   current);
#else
  return ose_efmi_get_next_current(i, (struct EfmiCurrentCursor *)cursor,                                   
				   current);
#endif  
} /* ose_efmi_get_first_current */

EfmiStatus
ose_efmi_get_next_current
ARGLIST(struct EfmiCurrentCursor *cc,
        struct EfmiCurrent *current)
{
  EfmiStatus			status;
  U32				lock;
  struct CurrentCursor		*cursor = (struct CurrentCursor *) cc;

  LOCK_SAVE(lock);
  CALLTRACE("ose_efmi_get_next_current");
  ENTER_EFMI(i);

  for(;;)
    {
      if(cursor->next >= cursor->num_cpus)
	{
	  status = EFMI_END_OF_LIST;
	  break;
	}
      status = get_current(cursor->next, current, cursor->size);
      cursor->next++;

      if(status == EFMI_OK)
	{
	  break;
	}
      
      /* For cpu #0, return actual failure status. */
      if(cursor->next == 1)
	{
	  break;
	}
    } /* for */

  EXIT_EFMI();
  LOCK_RESTORE(lock);
  return status;
} /* ose_efmi_get_next_current */

/*
** Retrieve the segment list of the system
*/
EfmiStatus
ose_efmi_get_first_segment
ARGLIST(struct EfmiSegmentCursor *prc, 
        struct EfmiSegmentInfo   *pri,
        size_t size)
{
  struct SegmentCursor		*cursor = (struct SegmentCursor *) prc;

  CALLTRACE("ose_efmi_get_first_segment");
  ENTER_EFMI(i);
  cursor->next = READ_PTR(zzuu, segment_list);
  cursor->size = size;
  EXIT_EFMI();
#if !defined (HOST_TOOL_ENVIRONMENT)  
  return ose_efmi_get_next_segment((struct EfmiSegmentCursor *) cursor, pri);
#else
  return ose_efmi_get_next_segment(i, (struct EfmiSegmentCursor *) cursor, pri);
#endif
} /* ose_efmi_get_first_segment */

EfmiStatus
ose_efmi_get_next_segment
ARGLIST(struct EfmiSegmentCursor *prc,
        struct EfmiSegmentInfo   *pri)
{
  EfmiStatus			status;
  U32				lock;
  struct SegmentCursor		*cursor = (struct SegmentCursor *) prc;

  LOCK_SAVE(lock);
  CALLTRACE("ose_efmi_get_next_segment");
  ENTER_EFMI(i);
  status = get_segment_info(cursor, 0, pri, cursor->size);
  EXIT_EFMI();
  LOCK_RESTORE(lock);
  return status;
} /* ose_efmi_get_next_segment */


/*
** Retrieve block information per segment
*/
EfmiStatus
ose_efmi_get_first_block
ARGLIST(PROCESS segment_id,
        struct EfmiBlockCursor *bc, 
        struct EfmiBlockInfo   *bi,
        size_t size)
{
  struct BlockCursor		*cursor = (struct BlockCursor *) bc;

  CALLTRACE("ose_efmi_get_first_block");
  ENTER_EFMI(i);
  cursor->next = READ_PTR(zzuu, bcb_list);
  cursor->segment_id = segment_id;
  cursor->size = size;
  EXIT_EFMI();
#if !defined (HOST_TOOL_ENVIRONMENT)  
  return ose_efmi_get_next_block((struct EfmiBlockCursor *) cursor, bi);
#else
  return ose_efmi_get_next_block(i, (struct EfmiBlockCursor *) cursor, bi);
#endif  
} /* ose_efmi_get_first_block */

EfmiStatus
ose_efmi_get_next_block
ARGLIST(struct EfmiBlockCursor *bc,
        struct EfmiBlockInfo   *bi)
{
  EfmiStatus			status;
  U32				lock;
  struct BlockCursor		*cursor = (struct BlockCursor *) bc;

  LOCK_SAVE(lock);
  CALLTRACE("ose_efmi_get_next_block");
  ENTER_EFMI(i);
  status = get_block_info(cursor, 0, bi, cursor->size);
  EXIT_EFMI();
  LOCK_RESTORE(lock);
  return status;
} /* ose_efmi_get_next_block */

/*
** Retrieve process information per block
*/
EfmiStatus
ose_efmi_get_first_process
ARGLIST(PROCESS block_id,
        struct EfmiProcessCursor *pc, 
        struct EfmiProcessInfo  *proc_info,
        size_t size)
{
  EfmiStatus			status;
  U32				lock;
  struct ProcessCursor		*cursor = (struct ProcessCursor *) pc;
  const struct _block		*bcb;

  SET_TYPE(_block, bcb);
  
  CALLTRACE("ose_efmi_get_first_process");
  ENTER_EFMI(i);
  if(READ_VAL(zzuu, next_free_pix) <= (PCB_INDEX_T) block_id)
    {
      EXIT_EFMI();
      return EFMI_UNAVAILABLE;
    }
  bcb = (struct _block *) pid_to_pcb(block_id);

  LOCK_SAVE(lock);
  if((status = check_block(bcb)) != EFMI_OK)
    {
      EXIT_EFMI();
      LOCK_RESTORE(lock);
      return status;
    }
  
  cursor->next = DEREF_PTR(bcb, proclist);
  cursor->size = size;
  status = get_process_info(cursor, 0, proc_info, size);
  EXIT_EFMI();
  LOCK_RESTORE(lock);
  return status;
} /* ose_efmi_get_first_process */

EfmiStatus
ose_efmi_get_next_process
ARGLIST(struct EfmiProcessCursor *pc,
        struct EfmiProcessInfo   *proc_info)
{
  EfmiStatus			status;
  U32				lock;
  struct ProcessCursor		*cursor = (struct ProcessCursor *) pc;

  LOCK_SAVE(lock);
  CALLTRACE("ose_efmi_get_next_process");
  ENTER_EFMI(i);
  status = get_process_info(cursor, 0, proc_info, cursor->size);
  EXIT_EFMI();
  LOCK_RESTORE(lock);
  return status;
} /* ose_efmi_get_next_process */

/*
** Retrieve environment variable information per process or block
** If maxLen is less then the length of a string, the string is filled in 
** and zero terminated, and EFMI_RESULT_TRUNCATED is returned.
*/
EfmiStatus 
ose_efmi_get_first_env_var
ARGLIST(PROCESS pid,
        struct EfmiEnvVarCursor *evc,
        char *name,  size_t max_name_len,
        char *value, size_t max_value_len)
{
  EfmiStatus			status;
  U32				lock;
  struct EnvVarCursor		*cursor = (struct EnvVarCursor *) evc;

  LOCK_SAVE(lock);
  CALLTRACE("ose_efmi_get_first_env_var");
  ENTER_EFMI(i);
  
  status = check_kernel();
  if(status == EFMI_OK)
    {
      status = check_pix_identity(pid);
    }
  if(status == EFMI_OK)		/* Process does still exist. */
    {
      struct _pcb *pcb = pid_to_pcb(pid);

      SET_TYPE(_pcb, pcb);
      
      if(DEREF_VAL(pcb, magic) != MAGIC_PCB && DEREF_VAL(pcb, magic) != MAGIC_BCB)
	{
	  status = EFMI_UNAVAILABLE; /* Neither process nor block. */
	}
      else
	{
	  cursor->pcb = pcb;
	  cursor->next = DEREF_PTR_TYPE(struct _pcb, cursor->pcb, env_list);
	  cursor->max_name_len = max_name_len;
	  cursor->max_value_len = max_value_len;

	  status = get_env_var(cursor, name, value);
	}
      EXIT_EFMI();
      LOCK_RESTORE(lock);
      return status;
    }
  EXIT_EFMI();
  LOCK_RESTORE(lock);
  return EFMI_UNAVAILABLE;
} /* ose_efmi_get_first_env_var */

EfmiStatus
ose_efmi_get_next_env_var
ARGLIST(struct EfmiEnvVarCursor *evc,
        char *name,
        char *value)
{
  EfmiStatus			status;
  U32				lock;
  struct EnvVarCursor		*cursor = (struct EnvVarCursor *) evc;

  LOCK_SAVE(lock);
  CALLTRACE("ose_efmi_get_next_env_var");
  ENTER_EFMI(i);
  status = get_env_var(cursor, name, value);
  EXIT_EFMI();
  LOCK_RESTORE(lock);
  return status;
} /* ose_efmi_get_next_env_var */

/*
** Retrieve pool information
*/
EfmiStatus
ose_efmi_get_first_pool
ARGLIST(struct EfmiPoolCursor *pc,
        struct EfmiPoolInfo   *pool_info,
        size_t size)
{
  struct PoolCursor		*cursor = (struct PoolCursor *) pc;

  CALLTRACE("ose_efmi_get_first_pool");
  ENTER_EFMI(i);
  cursor->next = READ_PTR(zzuu, pool_list);
  cursor->size = size;
  EXIT_EFMI();
#if !defined (HOST_TOOL_ENVIRONMENT)  
  return ose_efmi_get_next_pool((struct EfmiPoolCursor *) cursor, pool_info);
#else  
  return ose_efmi_get_next_pool(i, (struct EfmiPoolCursor *) cursor, pool_info);
#endif
} /* ose_efmi_get_first_pool */

EfmiStatus
ose_efmi_get_next_pool
ARGLIST(struct EfmiPoolCursor *pc,
        struct EfmiPoolInfo   *pool_info)
{
  EfmiStatus			status;
  U32				lock;
  struct PoolCursor		*cursor = (struct PoolCursor *) pc;

  LOCK_SAVE(lock);
  CALLTRACE("ose_efmi_get_next_pool");
  ENTER_EFMI(i);
  status = get_pool_info(cursor, 0, pool_info, cursor->size);
  EXIT_EFMI();
  LOCK_RESTORE(lock);
  return status;
} /* ose_efmi_get_next_pool */

/*
** Retrieve fragment information per pool
*/
EfmiStatus
ose_efmi_get_first_fragment
ARGLIST(PROCESS pool_id,
        struct EfmiFragmentCursor *fc,
        struct EfmiFragmentInfo   *fragment_info,
        size_t size)
{
  EfmiStatus			status;
  U32				lock;
  struct FragmentCursor		*cursor = (struct FragmentCursor *) fc;

  LOCK_SAVE(lock);
  CALLTRACE("ose_efmi_get_first_fragment");
  ENTER_EFMI(i);
  status = check_pix_identity(pool_id);
  if(status == EFMI_OK)
    {
      cursor->poolcb = (struct poolcb *) pid_to_pcb(pool_id);
      cursor->next = DEREF_PTR_TYPE(struct poolcb, cursor->poolcb, frag_list);
      cursor->size = size;

      status = get_fragment_info(cursor, fragment_info, cursor->size);
    }
  EXIT_EFMI();
  LOCK_RESTORE(lock);
  return status;
} /* ose_efmi_get_first_fragment */

EfmiStatus
ose_efmi_get_next_fragment
ARGLIST(struct EfmiFragmentCursor *fc,
        struct EfmiFragmentInfo   *fragment_info)
{
  EfmiStatus			status;
  U32				lock;
  struct FragmentCursor		*cursor = (struct FragmentCursor *) fc;

  LOCK_SAVE(lock);
  CALLTRACE("ose_efmi_get_next_fragment");
  ENTER_EFMI(i);
  status = get_fragment_info(cursor, fragment_info, cursor->size);
  EXIT_EFMI();
  LOCK_RESTORE(lock);
  return status;
} /* ose_efmi_get_next_fragment */

/*
** Retrieve information about signals in a pool
*/
EfmiStatus
ose_efmi_get_first_pool_signal
ARGLIST(PROCESS pool_id,
        struct EfmiSignalCursor *psc,
        struct EfmiSignalInfo   *ps_info,
        size_t size)
{
  EfmiStatus			status;
  U32				lock;
  struct SignalCursor		*cursor = (struct SignalCursor *) psc;

  LOCK_SAVE(lock);
  CALLTRACE("ose_efmi_get_first_pool_signal");
  ENTER_EFMI(i);
  status = check_kernel();
  if(status == EFMI_OK)
    {
      status = check_pix_identity(pool_id);
    }

  if(status == EFMI_OK)
    {
      cursor->traverse_pool = 1;
      cursor->poolcb = (struct poolcb *) pid_to_pcb(pool_id);

      /* First one is dummy */
      cursor->next = get_sigadm_list_from_poolcb(cursor->poolcb);
      cursor->size = size;

      status = get_signal_info(cursor, ps_info, cursor->size);
      EXIT_EFMI();
      LOCK_RESTORE(lock);
      return status;
    }
  EXIT_EFMI();
  LOCK_RESTORE(lock);
  return EFMI_UNAVAILABLE;
} /* ose_efmi_get_first_pool_signal */

EfmiStatus
ose_efmi_get_next_pool_signal
ARGLIST(struct EfmiSignalCursor *psc,
        struct EfmiSignalInfo   *ps_info)
{
  EfmiStatus			status;
  U32				lock;
  struct SignalCursor		*cursor = (struct SignalCursor *) psc;

  LOCK_SAVE(lock);
  CALLTRACE("ose_efmi_get_next_pool_signal");
  ENTER_EFMI(i);
  status = get_signal_info(cursor, ps_info, cursor->size);
  EXIT_EFMI();
  LOCK_RESTORE(lock);
  return status;
} /* ose_efmi_get_next_pool_signal */

/*
** Retrieve information about signals in a process signal queue
*/
EfmiStatus 
ose_efmi_get_first_q_signal
ARGLIST(PROCESS pid,
        struct EfmiSignalCursor *sc,
        struct EfmiSignalInfo   *sig_info,
        size_t size)
{
  EfmiStatus			status;
  U32				lock;
  struct SignalCursor		*cursor = (struct SignalCursor *) sc;

  LOCK_SAVE(lock);
  CALLTRACE("ose_efmi_get_first_q_signal");
  ENTER_EFMI(i);
  status = check_kernel();
  if(status == EFMI_OK)
    {
      status = check_pix_identity(pid);
    }
  if(status == EFMI_OK)		/* Process does still exist. */
    {
      cursor->traverse_pool = 0;
      cursor->poolcb = NULL;
      cursor->next = DEREF_PTR_TYPE(struct _pcb, pid_to_pcb(pid), mailbox);
      cursor->size = size;

      status = get_signal_info(cursor, sig_info, cursor->size);
      EXIT_EFMI();
      LOCK_RESTORE(lock);
      return status;
    }
  EXIT_EFMI();
  LOCK_RESTORE(lock);
  return EFMI_UNAVAILABLE;
} /* ose_efmi_get_first_q_signal */

EfmiStatus 
ose_efmi_get_next_q_signal
ARGLIST(struct EfmiSignalCursor *sc,
        struct EfmiSignalInfo   *sig_info)
{
  EfmiStatus			status;
  U32				lock;
  struct SignalCursor		*cursor = (struct SignalCursor *) sc;

  LOCK_SAVE(lock);
  CALLTRACE("ose_efmi_get_next_q_signal");
  ENTER_EFMI(i);
  status = get_signal_info(cursor, sig_info, cursor->size);
  EXIT_EFMI();
  LOCK_RESTORE(lock);
  return status;
} /* ose_efmi_get_next_q_signal */

/*
** Retrieve information about attach signals
*/
#if !defined(HOST_TOOL_ENVIRONMENT)

static EfmiStatus
get_attach_info(struct AttachCursor *cursor,
		struct EfmiSignalInfo *sig_info,
		size_t sig_info_size)
{
  struct EfmiSignalInfo sigi;
  struct _attach_entry *attcb;
  struct _attach_entry *next_attcb = NULL;
  const U32 attcb_align_bits = POOL_ALIGNMENT - 1;
  int found = 0;

  memset(&sigi, 0, sizeof(sigi));

  if(OSE_ILLEGAL_ATTREF == cursor->next)
    {
      goto quitting;
    }

  /* Sanity check */
  if(cursor->next >= zzuu.next_free_attach)
    return EFMI_KERNEL_CORRUPTED;
  attcb = zzuu.attach_pool[cursor->next];

  if(attcb == NULL ||
     ((OSADDRESS)attcb) & attcb_align_bits ||
     attcb->attrid != cursor->next)
    return EFMI_KERNEL_CORRUPTED;
    
  while(attcb)
    {
      /* Currently implementing a subset only ?? */
      if(cursor->search_method)
	{
	  if(cursor->pid == attcb->victim)  /* Is victim */
	    {
	      next_attcb = attcb->vic_fq;
	      if(cursor->search_method == 1)  /* Search victim */
		found = 1;
	    }
	  else if(cursor->pid == attcb->owner)  /* Is owner */
	    {
	      next_attcb = attcb->own_fq;
	      if(cursor->search_method == 2)  /* Search owner */
		found = 1;
	    }
	  else
	    return EFMI_KERNEL_CORRUPTED;  /* Neither victim nor owner */

	  /* Sanity check next attcb */
	  if(next_attcb)
	    {
	      if(((OSADDRESS)next_attcb) & attcb_align_bits ||
		 next_attcb->attrid >= zzuu.next_free_attach)
		{
		  ramlog_printf("efmi err:%d: cursor->pid %x attcb %x\n",
				__LINE__, cursor->pid, attcb);
		  return EFMI_KERNEL_CORRUPTED;
		}
	      cursor->next = next_attcb->attrid;
	    }
	  else
	    cursor->next = OSE_ILLEGAL_ATTREF;
	}
      else
	{
	  /* search_method 0: System scan */
	  return EFMI_UNAVAILABLE; /* Not implemented ?? */
	}

      if(found)
	{
	  if(cursor->options)
	    {
	      /* Return information on the current attcb. */
	      sigi.owner = attcb->owner;
	      sigi.addressee_pid = attcb->owner;
	      sigi.sender_pid = attcb->victim;
	      sigi.sig_status = EFMI_SS_VALID;
	      memcpy(sig_info, &sigi, MIN(sig_info_size, sizeof(sigi)));
	      return EFMI_OK;
	    }
	  /* Options zero, just count the number of attachments */
	  found = 0;
	  sigi.sig_no++;
	}
      attcb = next_attcb;
    } /* while */

 quitting:
  if(0 == cursor->options)
    {
      /* Return number of attachments */
      sigi.owner = cursor->pid;
      sigi.addressee_pid = cursor->pid;
      sigi.sender_pid = cursor->pid;
      sigi.sig_status = EFMI_SS_VALID;
      memcpy(sig_info, &sigi, MIN(sig_info_size,sizeof(sigi)));
      cursor->next = OSE_ILLEGAL_ATTREF;
      return EFMI_OK;
    }
  return EFMI_END_OF_LIST;
}

EfmiStatus
ose_efmi_get_first_attach_signal
ARGLIST(PROCESS pid,
        struct EfmiSignalCursor *sc,
        struct EfmiSignalInfo   *sig_info,
        size_t sig_info_size,
	U32 search_method,
	U32 options)
{
  EfmiStatus			status;
  U32				lock;
  struct _pcb			*pcb;
  struct _attach_entry		*attached;
  struct AttachCursor		*cursor = (struct AttachCursor *) sc;

  LOCK_SAVE(lock);
  CALLTRACE("ose_efmi_get_first_attach_signal");
  ENTER_EFMI(i);
  status = check_kernel();
  if(status == EFMI_OK)
    {
      /* ?? Currently we implement a subset only - check the arguments! ?? */
      if((search_method >= 1 && search_method <= 2) && options <= 1)
	{
	  status = check_pix_identity(pid);
	  if(EFMI_OK == status)
	    {
	      pcb = pid_to_pcb(pid);
	      if(MAGIC_PCB != pcb->magic || pcb_to_pid(pcb) != pid)
		status = EFMI_UNAVAILABLE;
	    }
	}
      else
	status = EFMI_CALL_INVALID;
    }

  if(status == EFMI_OK)		/* Process does still exist. */
    {
      cursor->search_method = search_method;
      cursor->options = options;
      cursor->size = sig_info_size;
      if(search_method == 1 || search_method == 2)
	{
	  cursor->pid = pid;
	  attached = pcb->attached;
	  cursor->next = attached ? attached->attrid : OSE_ILLEGAL_ATTREF;
	}
      else
	{
	  cursor->pid = 0;
	  if(zzuu.next_free_attach)
	    {
	      cursor->next = zzuu.next_free_attach - 1;
	    }
	  else
	    {
	      cursor->next = OSE_ILLEGAL_ATTREF;
	    }
	}

      status = get_attach_info(cursor, sig_info, cursor->size);
    }
  EXIT_EFMI();
  LOCK_RESTORE(lock);
  return status;
} /* ose_efmi_get_first_attach_signal */

EfmiStatus
ose_efmi_get_next_attach_signal
ARGLIST(struct EfmiSignalCursor *sc,
        struct EfmiSignalInfo   *sig_info)
{
  EfmiStatus			status;
  U32				lock;
  struct AttachCursor		*cursor = (struct AttachCursor *) sc;

  LOCK_SAVE(lock);
  CALLTRACE("ose_efmi_get_next_attach_signal");
  ENTER_EFMI(i);
  status = get_attach_info(cursor, sig_info, cursor->size);
  EXIT_EFMI();
  LOCK_RESTORE(lock);
  return status;
} /* ose_efmi_get_next_attach_signal */

#endif   /* !defined(HOST_TOOL_ENVIRONMENT) */

/*
** Retrieve the process, block and segment identity of the currently running 
** process.
*/
EfmiStatus
ose_efmi_get_current
ARGLIST(struct EfmiCurrent *current,
        size_t size)
{
  EfmiStatus		status;
  U32			lock;

  LOCK_SAVE(lock);
  CALLTRACE("ose_efmi_get_current");
  /* check_kernel() is not called by this function because we want 
   * ose_efmi_get_current() to report success whenever possible!
   */
  ENTER_EFMI(i);
  status = get_current(zzkrn_cpu_id(), current, size);
  EXIT_EFMI();
  LOCK_RESTORE(lock);
  
  return status;
} /* ose_efmi_get_current */

/*
** Retrieve all the pids of the processes named 'name'.
*/

EfmiStatus
ose_efmi_get_first_process_by_name
ARGLIST(const char *name,
        struct EfmiProcNameCursor *nc,
        PROCESS *pid)
{
  U32				lock;
  EfmiStatus			status;
  struct ProcNameCursor 	*cursor = (struct ProcNameCursor *) nc;

  LOCK_SAVE(lock);
  CALLTRACE("ose_efmi_get_first_process_by_name");
  ENTER_EFMI(i);
  cursor->next = 1;	
  status = get_process_by_name(name, cursor, pid);
  EXIT_EFMI();
  LOCK_RESTORE(lock);
  return status;
} /* ose_efmi_get_first_process_by_name */

EfmiStatus
ose_efmi_get_next_process_by_name
ARGLIST(const char *name,
        struct EfmiProcNameCursor *nc,
        PROCESS *pid)
{
  U32				lock;
  EfmiStatus			status;
  struct ProcNameCursor 	*cursor = (struct ProcNameCursor *) nc;

  LOCK_SAVE(lock);
  CALLTRACE("ose_efmi_get_next_process_by_name");
  ENTER_EFMI(i);
  status = get_process_by_name(name, cursor, pid);
  EXIT_EFMI();
  LOCK_RESTORE(lock);
  return status;
} /* ose_efmi_get_next_process_by_name */

/*
** Retrieve the name associated with process or block 'id'. If nameLen is less 
** than the length of the process name 'name' is filled in and zero terminated,
** and EFMI_RESULT_TRUNCATED is returned.
*/
EfmiStatus 
ose_efmi_get_process_name
ARGLIST(PROCESS id, char *name, size_t nameLen)
{
  EfmiStatus			status;
  U32				lock;
  const struct _pcb		*pcb;
  const char			*tmp_name = NULL;

  SET_TYPE(_pcb, pcb);
  
  status = EFMI_UNAVAILABLE;
  LOCK_SAVE(lock);
  CALLTRACE("ose_efmi_get_process_name");
  ENTER_EFMI(i);
  if(check_kernel() == EFMI_OK &&
     check_pix_identity(id) == EFMI_OK)
    {
      pcb = pid_to_pcb(id);
      if(DEREF_VAL(pcb, magic) == MAGIC_PCB)
	{
	  tmp_name = DEREF_PTR(pcb, name);
	}
      if(DEREF_VAL(pcb, magic) == MAGIC_BCB)
	{
	  tmp_name = DEREF_PTR_TYPE(struct _block, (const struct _block *) pcb, name);
	}

      /* Names must either be in the system pool, or be names of known pose 
      * objects.
      */
      {      
#if !defined (HOST_TOOL_ENVIRONMENT)
         struct _pcb *idle_varp = *zzosvarp.idle_varp;
#else
         uint32_t addr = READ_VAL(zzosvarp, idle_varp);
         struct _pcb *idle_varp = (struct _pcb *)fmi_read_u32(addr);
#endif      
         if(check_syspool_pointer(tmp_name) == EFMI_OK ||
            id == READ_VAL(zzuu, tickd_) || id == READ_VAL(zzuu, sysd_) || id == READ_VAL(zzuu, huntd_) ||
            id == READ_VAL(zzuu, huntd2_) || id == READ_VAL(zzuu, ldm_) || pcb == idle_varp ||
            ((const struct _block *) pcb) == READ_PTR(zzuu, system_block))
         {
            status = EFMI_OK;
            if(strncpy_truncated(name, tmp_name, nameLen))
	    {
               status = EFMI_RESULT_TRUNCATED;
	    }
         }
      }
    }
  EXIT_EFMI();
  LOCK_RESTORE(lock);
  return status;
} /* ose_efmi_get_process_name */

/*
** Retrieve information about the value of a register.
*/

EfmiStatus 
ose_efmi_get_register
ARGLIST(PROCESS pid,
        U32 register_number,
        struct EfmiRegisterInfo *reg)
{
  EfmiStatus			status;
  U32				lock;
  struct _pcb			*pcb;

  (void) register_number;
  (void) reg;
  
  SET_TYPE(_pcb, pcb);
  
  status = EFMI_UNAVAILABLE;
  LOCK_SAVE(lock);
  CALLTRACE("ose_efmi_get_register");
  ENTER_EFMI(i);
  if(check_pix_identity(pid) == EFMI_OK)
    {
      pcb = pid_to_pcb(pid);
      /*
      ** Check that this pcb is an alive process and either current process
      ** or of type PRI or BACKGROUND.
      */
      if(DEREF_VAL(pcb, magic) == MAGIC_PCB && (DEREF_VAL(pcb, priority) < 33 || pcb == zzcur_p))
	{
	   if (check_syspool_pointer(DEREF_PTR(pcb, uarea)) == EFMI_OK)
	   {
	      if(0 == zzget_register(DEREF_PTR(pcb, uarea), register_number, &reg->value,
				     is_syspool_pointer))
		   {
		      status = EFMI_OK;
		   }
	   }
	}
    }

  EXIT_EFMI();
  LOCK_RESTORE(lock);
  return status;
} /* ose_efmi_get_register */

/*
** Retrieve information about a kernel configuration attribute
*/
EfmiStatus 
ose_efmi_get_config
ARGLIST(enum EfmiConfigurationTag tag,
        struct EfmiConfigurationInfo *conf)
{
  EfmiStatus			status;
  U32				lock;
  const struct sysctlblk	*sysc;

  SET_TYPE(sysctlblk, sysc);
  
  LOCK_SAVE(lock);
  CALLTRACE("ose_efmi_get_config");
  ENTER_EFMI(i);
  status = check_kernel();
  LOCK_RESTORE(lock);

  sysc = READ_PTR(zzuu, sysctlblk);
  
  if(status != EFMI_OK)
    {
      EXIT_EFMI();
      return status;		/* Mission failed */
    }

  switch(tag)
    {
    case EFMI_CONF_TRACESIZE:	/* Number of bytes reserved for the trace in the pool. */
      conf->value = DEREF_VAL(sysc, tracesize);
      break;

    case EFMI_CONF_SUPERV_STACK_SIZE: /* Extra space for superv processes. */
      conf->value = DEREF_VAL(sysc, superv_stack_size);
      break;

    case EFMI_CONF_MAX_PROCS:	/* Contains max number of allowed processes. */
      conf->value = DEREF_VAL(sysc, max_procs);
      break;

    case EFMI_CONF_MAX_ATTACHED: /* Contains max number of attached signals. */
      conf->value = DEREF_VAL(sysc, max_attached);
      break;

    case EFMI_CONF_TICK_LEN:	/* Tick length in microseconds. */
      conf->value = DEREF_VAL(sysc, tick_len);
      break;

    case EFMI_CONF_RESIDENT_PRI: /* Virtual mem disabled at this pri and up. */
      conf->value = DEREF_VAL(sysc, resident_pri);
      break;

    case EFMI_CONF_ALLOW_USER_INTPROC: /* Nonzero to allow user mode proc to create
					*  intprocs. */
      conf->value = DEREF_VAL(sysc, allow_user_intproc);
      break;

    case EFMI_CONF_ALLOW_ALLOC_NIL: /* Nonzero to allow alloc to return NIL upon 
				     * error handler exit for non sys pools. */
      conf->value = DEREF_VAL(sysc, allow_alloc_nil);
      break;

    case EFMI_CONF_SKIP_STACK_ALLOCATION: /* Nonzero to pass NULL in the stack to
					   * zzcreate_context(). */
      conf->value = DEREF_VAL(sysc, skip_stack_allocation);
      break;

    case EFMI_CONF_LINKHANDLER_TIMEOUT: /* Nonzero enables linkh response timeout
					 *  error and sets the time (sek). */
      conf->value = DEREF_VAL(sysc, linkhandler_timeout);
      break;

    case EFMI_CONF_ALLOW_POOL_COPY: /* Nonzero enables intersegment copying when 
				     * sending between pools. */
      conf->value = DEREF_VAL(sysc, allow_pool_copy);
      break;

    case EFMI_CONF_TABLE_EXTENDED: /* The present table extension level. */
      conf->value = DEREF_VAL(sysc, table_extended);
      break;

    case EFMI_CONF_LDM_PRIORITY: /* The priority of the ose_ldm/ process. */
      conf->value = DEREF_VAL(sysc, ldm_priority);
      break;

    case EFMI_CONF_HUNTD_PRIORITY: /* The priority of the ose_huntd process. */
      conf->value = DEREF_VAL(sysc, huntd_priority);
      break;

    case EFMI_CONF_MAIN_PRIORITY: /* The priority of the zzmain process. */
      conf->value = DEREF_VAL(sysc, main_priority);
      break;

    case EFMI_CONF_SYSD_PRIORITY: /* The priority of the ose_sysd process. */
      conf->value = DEREF_VAL(sysc, sysd_priority);
      break;

    case EFMI_CONF_ENABLE_GET_PCB_REGS: /* Nonzero enables registers strings in
					 * get_pcb() */
      conf->value = DEREF_VAL(sysc, enable_get_pcb_regs);
      break;

    case EFMI_CONF_TICKD_PRIORITY: /* The priority of the ose_tickd process. */
      conf->value = DEREF_VAL(sysc, tickd_priority);
      break;

    case EFMI_CONF_BGD_TIMESLICE: /* The default timeslice for background
				   * processes. */
      conf->value = DEREF_VAL(sysc, bgd_timeslice);
      break;

    case EFMI_CONF_ENABLE_REMCALLS: /* Flag to enable remote system calls. */
      conf->value = DEREF_VAL(sysc, enable_remcalls);
      break;

    case EFMI_CONF_NTICK_FREQ: /* RMM ntick frequency. */
      conf->value = 1000000; /* This is hardcoded in rmm.c! */
      break;
      
    default:
      status = EFMI_UNAVAILABLE;
      break;
    } /* switch */

  EXIT_EFMI();
  return status;
} /* ose_efmi_get_config */
