/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2009 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */


/**
 * @private
 *
 * @brief Convert efmi calls to dbgserver interface messages.
 *
 * @long To be used by JTAG debuggers and freeze mode monitors only
 * 
 */

/*
*****************************************************************************
** INCLUDE FILES
*****************************************************************************
*/

#include "assert.h"

#include "osetypes.h"
#include "efmi.h"
/* #include "ramlog.h" */
#include "ose_err.h"
#include "fm_dbgserver.h"

#define _GLOBAL static
#define IS_INCLUDED
#define IS_DBGSERVER_CLIENT
/* We are not going to do any byte order conversion so hard code endian. */
#ifndef BIG_ENDIAN
#define BIG_ENDIAN
#endif
#ifdef LITTLE_ENDIAN
#undef LITTLE_ENDIAN
#endif
#include "dbgserverinterface.sig"

#define alloc ose_dbg_fm_alloc
#include "dbgserverinterface.c"
#undef alloc
#undef _GLOBAL

/*
*****************************************************************************
** Types
*****************************************************************************
*/

union SIGNAL
{
   SIGSELECT                   sigNo;
   struct SigHeader            SigHeader;
   struct DBGPing              DBGPing;
   struct DBGInitConnection    DBGInitConnection;
   struct DBGGetBlockInfo      DBGGetBlockInfo;
   struct DBGGetBlockProcInfo  DBGGetBlockProcInfo;
   struct DBGGetProcessInfo    DBGGetProcessInfo;
   struct DBGGetSignalQueue    DBGGetSignalQueue;
   struct DBGGetStackUsage     DBGGetStackUsage;
   struct DBGGetEnvList        DBGGetEnvList;
   struct DBGGetPoolInfo       DBGGetPoolInfo;
   struct DBGGetPoolSignals    DBGGetPoolSignals;
   struct DBGGetBlockFullInfo  DBGGetBlockFullInfo;
   struct DBGReadRegister      DBGReadRegister;
   struct DBGWriteRegister     DBGWriteRegister;
};

typedef enum
{
   DBG_OK,
   DBG_EFMI_ERROR,
   DBG_TRM_BUFFER_FULL
} DBGStatus;

/*
*****************************************************************************
** Prototypes
*****************************************************************************
*/

void ose_dbg_fm_transfer_bp(void);
extern void ose_dbg_fm_transfer_bp(void);

/*
*****************************************************************************
** Constants and variables
*****************************************************************************
*/

#define MAX_PROC_NAME_LENGTH 1024
#define MAX_BLOCK_NAME_LENGTH 128
#define MAX_ENVVAR_NAME_LENGTH 128
#define MAX_ENVVAR_VALUE_LENGTH 1024

/**
 * @macro MAX_NO_SIG_SIZES
 * @brief The max number of pre defined signal sizes in a pool.
 */
#define MAX_NO_SIG_SIZES 8

/**
 * @macro MAX_POOL_FRAGMENTS
 * @brief The max number of pool fragments that can be retrived for one pool.
 */
#define MAX_POOL_FRAGMENTS 32

/**
 * @macro SIGNAL_BATCH_COUNT
 * @brief Number of signals to put in one DBGGetPoolSignalsReply.
 */
#define SIGNAL_BATCH_COUNT 8

/**
 * @macro MAX_TRANSFER_SIG_SIZE
 * @brief Max size for a signal to be transfered.
 *
 * @long Must be four bytes smaller than TRANSFER_BUFFER_SIZE.
 */
#define MAX_TRANSFER_SIG_SIZE 1024

/**
 * @macro TRANSFER_BUFFER_SIZE
 * @brief The transfer buffer size.
 * @seealso ose_dbg_transfer_buffer
 */
#define TRANSFER_BUFFER_SIZE 0x2000

char ose_dbg_transfer_buffer[TRANSFER_BUFFER_SIZE];
/**
 * @data ose_dbg_transfer_buffer_size
 *
 * @brief The amount of data in the transfer buffer, used to tell how many
 * messages to read from the buffer.
 *
 * @long This variable holds how much data there is in the transfer buffer, it
 * is used in both directions, i.e. both "sending" and "receiving" messages
 * sizes are specified in this variable. If two messages of size 8 byte is in
 * the buffer this variable should have the value 16.
 */
U32 ose_dbg_transfer_buffer_size;
U32 ose_dbg_transfer_error;

/* Used by the allocation mechanism for telling where to break memory
 in the transfer buffer. */
static size_t pos;

/*
*****************************************************************************
** Private Functions
*****************************************************************************
*/

static int
check_error(void)
{
    if(0 != ose_dbg_transfer_error)
    {
	/* Cludge fix: this is an illegal signal that will be picked up by
	 * the client that will drop the connection. We need to extend
	 * the dbgserverinterface with an asynchronous error message.
	 */
	memset(ose_dbg_transfer_buffer, 0, 8);
	return 0;
    }
    else
    {
	return 1;
    }
}

/**
 * @function init_connection
 * @brief Send reply signal to client.
 */
static void
init_connection(struct DBGInitConnection *signal)
{
   union SIGNAL *send_sig;
   U32 tick_length;
   int con_accepted;
   
   assert(signal);
   
   ose_dbg_transfer_error = 0;
   if(EFMI_OK == ose_efmi_get_status())
   {
      struct EfmiConfigurationInfo conf;
      EfmiStatus status = ose_efmi_get_config(EFMI_CONF_TICK_LEN,
					      &conf);
      if(EFMI_OK == status)
      {
	 con_accepted = (MAJOR_INTERFACE_VERSION ==
			 signal->majorInterfaceVersion? 1: 0);
	 tick_length = conf.value / 1000; /* Tick length in ms */
      }
      else
      {
	 con_accepted = 0;
	 tick_length = 0;
      }
   }
   else
   {
      /* Kernel is not accessible so fail the connection */
      con_accepted = 0;
      tick_length = 0;
   }
   /* Set the transfer buffer size so it can be read by the debugger. */
   ose_dbg_transfer_buffer_size = TRANSFER_BUFFER_SIZE;
   create_DBGInitConnectionReply (con_accepted,
				  MAJOR_INTERFACE_VERSION,
				  MINOR_INTERFACE_VERSION,
				  tick_length,
				  0, /* Trace and evact is not supported */
				  0,
				  &send_sig);
   assert(send_sig != NULL);
   ose_dbg_fm_send(send_sig);

   if(signal->minorInterfaceVersion > 0)
   {
      U16 endian_test = 0x1;
      U32 endian;
      U32 features[] = { OSE_FEATURE_SLD_READ_REGISTERS };
      
      if(0 == *((U8 *)&endian_test))
      {
         /* We are a big endian system */
	 endian = 0x1020304;
      }
      else
      {
         /* We are a little endian system */
	 endian = 0x04030201;
      }
      /* Interface specifies an additional signal */
      create_DBGTargetInfo(endian,
			   TRANSFER_BUFFER_SIZE,
			   0,
			   0,
			   0,
			   0,
			   features,
			   4,
			   &send_sig);
      assert(send_sig != NULL);
      ose_dbg_fm_send(send_sig);
   }
}

static U16
efmi_state2debug_state(U32 pid, U32 efmi_state)
{
   U32 state = efmi_state;
   if(0 == state)
   {
      /* Let see if the process is running or if it is just ready. */
      struct EfmiCurrent current_info;
      if(EFMI_OK == ose_efmi_get_current(&current_info, sizeof(current_info)))
      {
	 if(current_info.process == pid)
	 {
	    state |= DBG_RUNNING;
	 }
	 else
	 {
	    state |= DBG_READY;
	 }
      }
      else
      {
	 /* FIXME: corrupt kernel? cheat for now */
	 state |= DBG_READY;
      }
   }
   assert(state > 0);
   return state;
}

static DBGStatus
create_proc_info(U32 pid, int full)
{
   struct EfmiProcessInfo proc_info;
   static char name[1024];
   EfmiStatus status;
   
   assert(pid > 0x10000);
   
   status = ose_efmi_get_process_info(pid, &proc_info,
				      sizeof(proc_info));
   if(EFMI_OK == status)
   {
      status = ose_efmi_get_process_name(pid, name, sizeof(name));
      if(EFMI_OK == status)
      {
	 union SIGNAL *sig;
	 U16 state = efmi_state2debug_state(pid, proc_info.status);
	 if(full)
	 {
	    S32 wanted_cnt = (S32)proc_info.wanted[0];
	    if(wanted_cnt < 0) wanted_cnt *= -1;
	    create_DBGGetProcessFullInfoReply(pid,
					      proc_info.block,
					      proc_info.parent,
					      proc_info.user,
					      (U16)proc_info.type,
					      state,
					      (U16)proc_info.priority,
					      0, /* FIXME signals in queue */
					      proc_info.entry_point,
					      0, 0, /* Creation time */
					      proc_info.fsemvalue,
					      proc_info.process_stack.top -
					      proc_info.process_stack.limit,
					      proc_info.line,
					      proc_info.sig_cnt_owned,
					      &proc_info.wanted[1],
					      wanted_cnt * sizeof(SIGSELECT),
					      name,
					      proc_info.file,
					      "", /* CPU Registers */
					      &sig);
	    ose_dbg_fm_send(sig);
	    return DBG_OK;
	 }
	 else
	 {
	    create_DBGGetProcessInfoReply(pid, proc_info.block,
					  proc_info.parent,
					  (U16)proc_info.type,
					  state,
					  (U16)proc_info.priority, 0, 0,
					  name, &sig);
	    ose_dbg_fm_send(sig);
	    return DBG_OK;
	 }
      }
   }
   return DBG_EFMI_ERROR;
}

static void
get_process_info(struct DBGGetProcessInfo *insig)
{
   union SIGNAL *sig;
   U32 pid;
   int full;
   
   assert(insig);
   assert(insig->sigNo == DBGGETPROCESSINFO);

   pid = insig->pid;
   full = insig->fullInfo;
   
   if(pid != 0)
   {
      /* One process requested */
      create_proc_info(pid, full);
   }
   else
   {
      /* All processes */
      struct EfmiSegmentCursor prc;
      struct EfmiSegmentInfo pri;
      EfmiStatus espr;
      for(espr = ose_efmi_get_first_segment(&prc, &pri, sizeof(pri));
	  EFMI_OK == espr;
	  espr = ose_efmi_get_next_segment(&prc, &pri))
      {
	 struct EfmiBlockCursor bc;
	 struct EfmiBlockInfo bi;
	 EfmiStatus esb;
	 for(esb = ose_efmi_get_first_block(pri.segment, &bc, &bi, sizeof(bi));
	     EFMI_OK == esb;
	     esb = ose_efmi_get_next_block(&bc, &bi))
	 {
	    struct EfmiProcessCursor pc;
	    struct EfmiProcessInfo pi;
	    EfmiStatus esp;
	    for(esp = ose_efmi_get_first_process(bi.block, &pc, &pi, sizeof(pi));
		EFMI_OK == esp;
		esp = ose_efmi_get_next_process(&pc, &pi))
	    {
	       create_proc_info(pi.process, full);
	    }
	 }
      }
   }
   create_DBGInfoProcEndMark(&sig);
   ose_dbg_fm_send(sig);
}

static void
get_current_process_info(void)
{
   union SIGNAL *sig;
   struct EfmiCurrent current;
   EfmiStatus status;

   status = ose_efmi_get_current(&current, sizeof(current));
   if(EFMI_OK == status)
   {
      /* One process requested */
      create_proc_info(current.process, 0);
   }
   create_DBGInfoProcEndMark(&sig);
   ose_dbg_fm_send(sig);
}

static void
get_block_proc_info(struct DBGGetBlockProcInfo *insig)
{
   union SIGNAL *sig;
   U32 bid;
   U32 full;
   EfmiStatus status;
   struct EfmiProcessCursor pc;
   struct EfmiProcessInfo pi;
   
   assert(insig);
   assert(insig->sigNo == DBGGETBLOCKPROCINFO);
   assert(insig->bid > 0x10000);
   assert(insig->fullInfo == 0 || insig->fullInfo == 1);

   bid = insig->bid;
   full = insig->fullInfo;

   for(status = ose_efmi_get_first_process(bid, &pc, &pi, sizeof(pi));
       EFMI_OK == status;
       status = ose_efmi_get_next_process(&pc, &pi))
   {
      create_proc_info(pi.process, full);
   }
   create_DBGInfoProcEndMark(&sig);
   ose_dbg_fm_send(sig);
}

static DBGStatus
create_block_info(U32 bid)
{
   char name[MAX_BLOCK_NAME_LENGTH];
   struct EfmiBlockInfo bi;
   EfmiStatus status;

   assert(bid > 0x10000);
   
   status = ose_efmi_get_block_info(bid, &bi, sizeof(bi));
   if(EFMI_OK == status)
   {
      status = ose_efmi_get_process_name(bid, name, 128);
      if(EFMI_OK == status)
      {
	 union SIGNAL *sig;
	 create_DBGGetBlockInfoReply(bid, 0, 0, 0, bi.user, name, &sig);
	 assert(sig);
	 ose_dbg_fm_send(sig);
	 return DBG_OK;
      }
   }
   return DBG_EFMI_ERROR;
}

static DBGStatus
create_block_full_info(U32 bid)
{
   char name[MAX_BLOCK_NAME_LENGTH];
   struct EfmiBlockInfo bi;
   EfmiStatus status;

   assert(bid > 0x10000);
   
   status = ose_efmi_get_block_info(bid, &bi, sizeof(bi));
   if(EFMI_OK == status)
   {
      status = ose_efmi_get_process_name(bid, name, 128);
      if(EFMI_OK == status)
      {
	 union SIGNAL *sig;
	 create_DBGGetBlockFullInfoReply(bid, 0, 0, 0, bi.user,
					 0, /* FIXME Max sig size, unknown */
					 bi.segment,
					 bi.sig_pool,
					 bi.stack_pool,
					 name, &sig);
	 assert(sig);
	 ose_dbg_fm_send(sig);
	 return DBG_OK;
      }
   }
   return DBG_EFMI_ERROR;
}

static void
ose_dbg_list_blocks(struct DBGGetBlockInfo *insig)
{
   union SIGNAL *sig;
   U32 bid;

   assert(insig);
   assert(insig->sigNo == DBGGETBLOCKINFO);

   bid = insig->bid;

   if(bid)
   {
      create_block_info(bid);
   }
   else
   {
      struct EfmiSegmentCursor prc;
      struct EfmiSegmentInfo pri;
      EfmiStatus espr;
      
      for(espr = ose_efmi_get_first_segment(&prc, &pri, sizeof(pri));
	  EFMI_OK == espr;
	  espr = ose_efmi_get_next_segment(&prc, &pri))
      {
	 struct EfmiBlockCursor bc;
	 struct EfmiBlockInfo bi;
	 EfmiStatus esb;
	 for(esb = ose_efmi_get_first_block(pri.segment, &bc, &bi, sizeof(bi));
	     EFMI_OK == esb;
	     esb = ose_efmi_get_next_block(&bc, &bi))
	 {
	    create_block_info(bi.block);
	 }
      }
   }
   create_DBGInfoBlockEndMark(&sig);
   ose_dbg_fm_send(sig);
}

static void
ose_dbg_list_blocks_full(struct DBGGetBlockFullInfo *insig)
{
   union SIGNAL *sig;
   U32 bid;

   assert(insig);
   assert(insig->sigNo == DBGGETBLOCKFULLINFO);

   bid = insig->bid;

   if(bid)
   {
      create_block_full_info(bid);
   }
   else
   {
      struct EfmiSegmentCursor prc;
      struct EfmiSegmentInfo pri;
      EfmiStatus espr;
      
      for(espr = ose_efmi_get_first_segment(&prc, &pri, sizeof(pri));
	  EFMI_OK == espr;
	  espr = ose_efmi_get_next_segment(&prc, &pri))
      {
	 struct EfmiBlockCursor bc;
	 struct EfmiBlockInfo bi;
	 EfmiStatus esb;
	 for(esb = ose_efmi_get_first_block(pri.segment, &bc, &bi, sizeof(bi));
	     EFMI_OK == esb;
	     esb = ose_efmi_get_next_block(&bc, &bi))
	 {
	    if(DBG_OK != create_block_full_info(bi.block))
	    {
	       goto break_out;
	    }
	 }
      }
   }
  break_out:
   create_DBGInfoBlockEndMark(&sig);
   ose_dbg_fm_send(sig);
}

static void
get_signal_queue(struct DBGGetSignalQueue *insig)
{
   union SIGNAL *sig;
   struct EfmiSignalCursor sc;
   struct EfmiSignalInfo si;
   EfmiStatus status;
   U32 pid;
   int entry_no = 0;
   
   assert(insig);
   assert(DBGGETSIGNALQUEUE == insig->sigNo);
   assert(insig->pid > 0x10000);

   pid = insig->pid;
   
   for(status = ose_efmi_get_first_q_signal(pid, &sc, &si, sizeof(si));
       EFMI_OK == status;
       status = ose_efmi_get_next_q_signal(&sc, &si))
   {
      create_DBGGetSignalQueueReply(pid,
				    si.sender_pid,
				    si.addressee_pid,
				    si.sig_size,
				    si.sig_no,
				    entry_no,
				    &sig);
      ose_dbg_fm_send(sig);
      entry_no++;
   }
   
   create_DBGInfoQueueEndMark(pid, &sig);
   ose_dbg_fm_send(sig);
}

static EfmiStatus
create_stack_usage(U32 pid)
{
   struct EfmiProcessInfo pi;
   EfmiStatus status;
   union SIGNAL *sig;

   status = ose_efmi_get_process_info(pid, &pi, sizeof(pi));
   if(EFMI_OK == status)
   {
      struct EfmiProcessStatistics stat;
      status = ose_efmi_get_process_statistics(pid,
					       EFMI_PROCESS_STACK_USAGE,
					       &stat);
      if(EFMI_OK == status)
      {
	 create_DBGGetStackUsageReply(pid, 0,
				      pi.process_stack.limit,
				      pi.process_stack.top,
				      stat.value,
				      &sig);
	 ose_dbg_fm_send(sig);
      }
   }
   return status;
}

static void
get_stack_usage(struct DBGGetStackUsage *insig)
{
   union SIGNAL *sig;
   U32 pid;
   
   assert(insig);
   assert(DBGGETSTACKUSAGE == insig->sigNo);
   assert(insig->pid > 0x10000 || insig->pid == 0);

   pid = insig->pid;
   
   if(pid != 0)
   {
      if(EFMI_OK != create_stack_usage(pid))
      {
	 /* Send a failure message back since create_stack_usage failed */
	 create_DBGGetStackUsageReply(pid, -1, 0, 0, 0, &sig);
	 ose_dbg_fm_send(sig);
      }
   }
   else
   {
      /* All processes */
      struct EfmiSegmentCursor prc;
      struct EfmiSegmentInfo pri;
      EfmiStatus espr;
      for(espr = ose_efmi_get_first_segment(&prc, &pri, sizeof(pri));
	  EFMI_OK == espr;
	  espr = ose_efmi_get_next_segment(&prc, &pri))
      {
	 struct EfmiBlockCursor bc;
	 struct EfmiBlockInfo bi;
	 EfmiStatus esb;
	 for(esb = ose_efmi_get_first_block(pri.segment, &bc, &bi, sizeof(bi));
	     EFMI_OK == esb;
	     esb = ose_efmi_get_next_block(&bc, &bi))
	 {
	    struct EfmiProcessCursor pc;
	    struct EfmiProcessInfo pi;
	    EfmiStatus esp;
	    for(esp = ose_efmi_get_first_process(bi.block, &pc, &pi, sizeof(pi));
		EFMI_OK == esp;
		esp = ose_efmi_get_next_process(&pc, &pi))
	    {
	       create_stack_usage(pi.process);
	    }
	 }
      }
      /* Send an end mark message with OK status. */
      create_DBGGetStackUsageReply(0, 1, 0, 0, 0, &sig);
      ose_dbg_fm_send(sig);
   }
}

static void
dbg_get_env_list(struct DBGGetEnvList *insig)
{
   union SIGNAL *sig;
   EfmiStatus status;
   struct EfmiEnvVarCursor evc;
   static char name[MAX_ENVVAR_NAME_LENGTH];
   static char value[MAX_ENVVAR_VALUE_LENGTH];
   U32 pid;
   
   assert(insig);
   assert(DBGGETENVLIST == insig->sigNo);
   assert(insig->pid > 0x10000);

   pid = insig->pid;
   
   for(status = ose_efmi_get_first_env_var(pid,
					   &evc,
					   name, MAX_ENVVAR_NAME_LENGTH,
					   value, MAX_ENVVAR_VALUE_LENGTH);
       EFMI_OK == status;
       status = ose_efmi_get_next_env_var(&evc, name, value))
   {
      create_DBGGetEnvListReply(pid, name, value, &sig);
      ose_dbg_fm_send(sig);
   }

   create_DBGGetEnvListEndMark(pid, &sig);
   ose_dbg_fm_send(sig);   
}

static void
dbg_read_register(struct DBGReadRegister *insig)
{
   union SIGNAL *sig;
   U32 pid;
   U32 size;
   int i;
   U32 *reg_pair;
   int count;
   
   assert(insig);
   assert(DBGREADREGISTER == insig->sigNo);
   assert(insig->pid > 0x10000);

   pid = insig->pid;

   DBGReadRegister_get_regVal(insig, reg_pair);
   DBGReadRegister_get_regVal_size(insig, size);

   count = size / sizeof(U32); /* Number of U32 in array, 2 * reg count */
   /* Loop over the register pairs (each register has two U32, id and value) */
   for(i = 0; i < count; i += 2)
   {
      struct EfmiRegisterInfo regvalue;
      EfmiStatus status = ose_efmi_get_register(pid, reg_pair[i], &regvalue);
      if(EFMI_OK == status)
      {
	 reg_pair[i + 1] = regvalue.value;
      }
      else
      {
	 reg_pair[i] |= 0x80000000; /* Set MSB of reg id to indicate failure. */
      }
   }
   create_DBGReadRegisterReply(0, reg_pair, size, &sig);
   ose_dbg_fm_send(sig);
}

static void
get_one_pool_info(struct EfmiPoolInfo *pool_info)
{
   union SIGNAL *reply;
   struct EfmiFragmentCursor fragc;
   struct EfmiFragmentInfo fragi;
   U32 fr_base_addr[MAX_POOL_FRAGMENTS];
   U32 fr_size[MAX_POOL_FRAGMENTS];
   U32 fr_used_stack[MAX_POOL_FRAGMENTS];
   U32 fr_used_sig[MAX_POOL_FRAGMENTS];
   U32 tot_size;
   U32 tot_free;
   U32 sig_conf_size[MAX_NO_SIG_SIZES];
   U32 stk_conf_size[MAX_NO_SIG_SIZES];
   U32 stk_alloc_cnt[MAX_NO_SIG_SIZES];
   U32 sig_alloc_cnt[MAX_NO_SIG_SIZES];
   U32 stk_free_cnt[MAX_NO_SIG_SIZES];
   U32 sig_free_cnt[MAX_NO_SIG_SIZES];
   EfmiStatus status;
   int fr_cnt = 0;
   
   status = ose_efmi_get_first_fragment(pool_info->pool_id,
					&fragc,
					&fragi,
					sizeof(fragi));
   if(EFMI_OK == status)
   {
      fr_base_addr[0] = fragi.base_address;
      fr_size[0] = fragi.size;
      fr_used_stack[0] = fragi.used_for_stacks;
      fr_used_sig[0] = fragi.used_for_signals;
      tot_size = fragi.size;
      tot_free = fragi.size - fragi.used_for_stacks - fragi.used_for_signals;
      
      for(fr_cnt = 1; fr_cnt < MAX_POOL_FRAGMENTS && EFMI_OK == status;
	  fr_cnt++)
      {
	 status = ose_efmi_get_next_fragment(&fragc, &fragi);
	 if(EFMI_OK == status)
	 {
	    fr_base_addr[fr_cnt] = fragi.base_address;
	    fr_size[fr_cnt] = fragi.size;
	    fr_used_stack[fr_cnt] = fragi.used_for_stacks;
	    fr_used_sig[fr_cnt] = fragi.used_for_signals;
	    tot_size += fragi.size;
	    tot_free += fragi.size - fragi.used_for_stacks -
	       fragi.used_for_signals;
	 }
      }
      if(EFMI_OK == status || EFMI_END_OF_LIST == status)
      {
	 int i;
	 
	 fr_cnt--;
	 fr_cnt *= sizeof(U32);

	 for(i = 0; i < MAX_NO_SIG_SIZES; i++)
	 {
	    sig_conf_size[i] = pool_info->signal_buf_sizes[i].configured_size;
	    stk_conf_size[i] = pool_info->stack_buf_sizes[i].configured_size;
	    sig_alloc_cnt[i] = pool_info->signal_buf_sizes[i].allocated;
	    stk_alloc_cnt[i] = pool_info->stack_buf_sizes[i].allocated;
	    sig_free_cnt[i] = pool_info->signal_buf_sizes[i].free;
	    stk_free_cnt[i] = pool_info->stack_buf_sizes[i].free;
	 }
	 create_DBGGetPoolInfoReply(pool_info->pool_id,
				    tot_size,
				    tot_free,
				    fr_base_addr, fr_cnt,
				    fr_size, fr_cnt,
				    fr_used_stack, fr_cnt,
				    fr_used_sig, fr_cnt,
				    stk_conf_size, 32,
				    sig_conf_size, 32,
				    stk_alloc_cnt, 32,
				    sig_alloc_cnt, 32,
				    stk_free_cnt, 32,
				    sig_free_cnt, 32,
				    &reply);
	 ose_dbg_fm_send(reply);
      }
   }
}

static void
get_pool_info(struct DBGGetPoolInfo *insig)
{
   union SIGNAL *reply;
   struct EfmiPoolInfo pooli;
   EfmiStatus status;
   
   assert(insig);
   assert(DBGGETPOOLINFO == insig->sigNo);
   assert((0 == insig->pid || insig->pid > 0x10000));

   if(0 == insig->pid)
   {
      struct EfmiPoolCursor poolc;

      status = ose_efmi_get_first_pool(&poolc, &pooli, sizeof(pooli));
      while(EFMI_OK == status)
      {
	 get_one_pool_info(&pooli);
	 status = ose_efmi_get_next_pool(&poolc, &pooli);
      }
   }
   else
   {
      if(EFMI_OK == ose_efmi_get_pool_info(insig->pid, &pooli, sizeof(pooli)))
      {
	 get_one_pool_info(&pooli);
      }
   }

   create_DBGInfoPoolEndMark(&reply);
   ose_dbg_fm_send(reply);
}

static void
get_pool_signals(struct DBGGetPoolSignals *insig)
{
   union SIGNAL *reply;
   EfmiStatus status;
   struct EfmiSignalCursor psc;
   struct EfmiSignalInfo ps_info;
   U32 data[SIGNAL_BATCH_COUNT * 8];
   int cnt = 0;
   
   assert(insig);
   assert(DBGGETPOOLSIGNALS == insig->sigNo);
   assert(0x10000 <= insig->pid);
   
   status = ose_efmi_get_first_pool_signal(insig->pid, &psc, &ps_info,
					   sizeof(ps_info));
   while(EFMI_OK == status)
   {
      data[SIGNAL_BATCH_COUNT * cnt + 0] = ps_info.sig_no;
      data[SIGNAL_BATCH_COUNT * cnt + 1] = ps_info.owner;
      data[SIGNAL_BATCH_COUNT * cnt + 2] = ps_info.sender_pid;
      data[SIGNAL_BATCH_COUNT * cnt + 3] = ps_info.addressee_pid;
      data[SIGNAL_BATCH_COUNT * cnt + 4] = ps_info.sig_size;
      data[SIGNAL_BATCH_COUNT * cnt + 5] = ps_info.size_in_pool;
      data[SIGNAL_BATCH_COUNT * cnt + 6] = ps_info.signal_pointer;
      data[SIGNAL_BATCH_COUNT * cnt + 7] = ps_info.sig_status;
      cnt++;
      status = ose_efmi_get_next_pool_signal(&psc, &ps_info);
      if(status != EFMI_OK || cnt > 7)
      {
	 create_DBGGetPoolSignalsReply(insig->pid,
				       data,
				       32 * cnt,
				       &reply);
	 ose_dbg_fm_send(reply);
	 cnt = 0;
      }
   }
   create_DBGPoolSignalsEndMark(insig->pid, &reply);
   ose_dbg_fm_send(reply);
}

/*
*****************************************************************************
** Public Functions
*****************************************************************************
*/

void
ose_dbg_fm_reset_trf_buf(void)
{
   pos = 0;
}

union SIGNAL *
ose_dbg_fm_alloc(U32 size, SIGSELECT sigNo)
{
   static char buffer[MAX_TRANSFER_SIG_SIZE]; /* Largest transfer signal */
   if(size <= MAX_TRANSFER_SIG_SIZE)
   {
      union SIGNAL *sig = (union SIGNAL *)buffer;
      if(size >= 4)
      {
	 sig->sigNo = sigNo;
      }
      return sig;
   }
   else
   {
      ose_dbg_transfer_error = OSE_ENO_KERN_SPACE;
      ose_dbg_fm_transfer_bp(); /* Should not return ! */
      assert(0);
      return NULL;
   }
}

void
ose_dbg_fm_send(union SIGNAL *sig)
{
   struct SigHeader *s;
   assert(sig);
   s = (struct SigHeader *)sig;
   assert(s->size < TRANSFER_BUFFER_SIZE - 4);
   if(pos + s->size > TRANSFER_BUFFER_SIZE - 4)
   {
      ose_dbg_fm_transfer_bp();
      ose_dbg_fm_reset_trf_buf();
   }
   memcpy(&ose_dbg_transfer_buffer[pos], s, s->size);
   pos += s->size;
   if(pos % 4) pos += 4 - (pos % 4); /* Align to next word */
   /* Indicate last signal in buffer */
   *((U32 *)&ose_dbg_transfer_buffer[pos]) = 0;
}

void
ose_dbg_fm_process_message(void)
{
   U32 sigNo;
   U32 size;
   
   ose_dbg_fm_reset_trf_buf();
   ose_dbg_transfer_error = 0;
   sigNo = *((U32 *)ose_dbg_transfer_buffer);
   size = *((U32 *)&ose_dbg_transfer_buffer[4]);

   if(size >= 8 && size <= TRANSFER_BUFFER_SIZE)
   {
      switch(sigNo)
      {
	 case DBGINITCONNECTION:
	    init_connection((struct DBGInitConnection *)
			    ose_dbg_transfer_buffer);
	    break;
	 case DBGGETPROCESSINFO:
	    get_process_info((struct DBGGetProcessInfo *)
			     ose_dbg_transfer_buffer);
	    break;
	 case DBGGETBLOCKPROCINFO:
	    get_block_proc_info((struct DBGGetBlockProcInfo *)
				ose_dbg_transfer_buffer);
	    break;
	 case DBGGETBLOCKINFO:
	    ose_dbg_list_blocks((struct DBGGetBlockInfo *)
				ose_dbg_transfer_buffer);
	    break;
	 case DBGGETBLOCKFULLINFO:
	    ose_dbg_list_blocks_full((struct DBGGetBlockFullInfo *)
				     ose_dbg_transfer_buffer);
	    break;
	 case DBGGETSIGNALQUEUE:
	    get_signal_queue((struct DBGGetSignalQueue *)
			     ose_dbg_transfer_buffer);
	    break;
	 case DBGGETSTACKUSAGE:
	    get_stack_usage((struct DBGGetStackUsage *)
			    ose_dbg_transfer_buffer);
	    break;
	 case DBGGETENVLIST:
	    dbg_get_env_list((struct DBGGetEnvList *)ose_dbg_transfer_buffer);
	    break;
	 case DBGREADREGISTER:
	    dbg_read_register((struct DBGReadRegister *)
			      ose_dbg_transfer_buffer);
	    break;
	 case DBGGETPOOLINFO:
	    get_pool_info((struct DBGGetPoolInfo *)
			  ose_dbg_transfer_buffer);
	    break;
	 case DBGGETPOOLSIGNALS:
	    get_pool_signals((struct DBGGetPoolSignals *)
			     ose_dbg_transfer_buffer);
	    break;
	 case DBGGETCURRENTPROCESS:
	    get_current_process_info();
	    break;
	 default:
	    /*ramlog_printf("ose fm dbgserver: received unknown message: %u\n",
	      sigNo);*/
	    break;
      }
   }
   else
   {
      /* ramlog_printf("ose fm dbgserver: received message with illegal size:"
	 " sigNo: %u size %u\n", sigNo, size);*/
   }
}
