/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2006 - 2011 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

#include "string.h"
#include "stdint.h"
#include "stdio.h"

#include "efmi.h"

#include "monitor.sig"
#include "monitor_execution.sig"

#include "fmm.h"
#include "fmm_efmi.h"

/*
** This code heavily uses assignments in conditional expressions. This RVCT
** pragma transforms such compiler warnings into less disturbing remarks.
*/
#if defined(__ARMCC_VERSION)
#pragma diag_remark 1293
#endif

#define assert(x) if (!(x)) return EFMI_UNAVAILABLE;

static FmmFmdProperties fmd_properties;


union Signal
{
   uint32_t signo;
   MonitorGetInfoRequest               get_info_request;
   MonitorGetProcessInfoReply          get_process_info_reply;
   MonitorGetProcessInfoLongReply      get_process_info_long_reply;
   MonitorGetBlockInfoReply            get_block_info_reply;
   MonitorGetSegmentInfoReply          get_segment_info_reply;
   MonitorGetStackUsageReply           get_stack_usage_reply;
   MonitorGetEnvVarsReply              get_env_vars_reply;
   MonitorGetSignalQueueReply          get_signal_queue_reply;
   MonitorGetSignalQueueLongReply      get_signal_queue_long_reply;
   MonitorGetPoolInfoReply             get_pool_info_reply;
   MonitorGetPoolSignalsReply          get_pool_signals_reply;
   MonitorGetInfoEndmark               get_info_endmark;
   MonitorGetRegisterInfoRequest       get_register_info_request;
   MonitorRegisterValues               get_registers_reply;
   MonitorGetRegisterInfoEndmark       get_register_info_endmark;
};

struct FmmSignal
{
   uint32_t size;
   union Signal data;
};

union ContextInfo
{
   uint32_t id;
   struct EfmiProcessInfo process_info;
   struct EfmiBlockInfo block_info;
   struct EfmiSegmentInfo segment_info;
};

static int
match(const char *p, const char *s)
{
   const char *p0 = (char *)0;

   while(*p || *s)
   {
      if(*p == '*')       /* Wildcard: Set p0 */
      {
         p0 = ++p;

         if(!*p)          /* Wildcard at end: Return match */
         {
            return 1;
         }
      }
      else if((*p == *s) || (*p == '?' && *s)) /* Match: Check next */
      {
         p++;
         s++;
      }
      else if(!*s || !p0) /* No match and s over or no match and no wildcard: */
      {                   /* Return fail */
         return 0;
      }
      else if(p != p0)    /* No match but after wildcard: Rewind to p0 */
      {
         s -= p - p0 - 1;
         p = p0;
      }
      else                /* No match but in wildcard: Check next */
      {
        s++;
      }
   }
   return 1;
}

static EfmiStatus
list_process(uint32_t id,
             EfmiStatus (*info)(int32_t (*reply)(FmmSignal *sig),
                                union ContextInfo *context,
                                const union Signal *request),
             int32_t (*reply)(FmmSignal *sig),
             const union Signal *request)
{
   EfmiStatus ret;
   union ContextInfo context;

   ret = ose_efmi_get_process_info(id, &context.process_info,
                                   sizeof(context.process_info));

   if (!ret)
   {
      if((ret = info(reply, &context, request)))
      {
         return ret;
      }
   }

   return ret;
}

static EfmiStatus
list_processes_in_block(uint32_t id,
                        EfmiStatus (*info)(int32_t (*reply)(FmmSignal *sig),
                                           union ContextInfo *context,
                                           const union Signal *request),
                        int32_t (*reply)(FmmSignal *sig),
                        const union Signal *request)
{
   EfmiStatus ret;
   union ContextInfo context;
   struct EfmiProcessCursor pc;

   ret = ose_efmi_get_first_process(id, &pc, &context.process_info,
                                    sizeof(context.process_info));

   while (!ret)
   {
      if ((ret = info(reply, &context, request)))
      {
         return ret;
      }
      ret = ose_efmi_get_next_process(&pc, &context.process_info);
   }

   return (ret == EFMI_END_OF_LIST) ? EFMI_OK : ret;
}

static EfmiStatus
list_processes_in_segment(uint32_t id,
                          EfmiStatus (*info)(int32_t (*reply)(FmmSignal *sig),
                                             union ContextInfo *context,
                                             const union Signal *request),
                          int32_t (*reply)(FmmSignal *sig),
                          const union Signal *request)
{
   EfmiStatus ret;

   struct EfmiBlockInfo bi;
   struct EfmiBlockCursor bc;

   ret = ose_efmi_get_first_block(id, &bc, &bi, sizeof(bi));

   while (!ret)
   {
      if ((ret = list_processes_in_block(bi.block, info, reply, request)))
      {
         return ret;
      }
      ret = ose_efmi_get_next_block(&bc, &bi);
   }

   return (ret == EFMI_END_OF_LIST) ? EFMI_OK : ret;
}

static EfmiStatus
list_processes_in_system(EfmiStatus (*info)(int32_t (*reply)(FmmSignal *sig),
                                            union ContextInfo *context,
                                            const union Signal *request),
                         int32_t (*reply)(FmmSignal *sig),
                         const union Signal *request)
{
   EfmiStatus ret;

   struct EfmiSegmentInfo si;
   struct EfmiSegmentCursor sc;

   ret = ose_efmi_get_first_segment(&sc, &si, sizeof(si));

   while (!ret)
   {
      if ((ret = list_processes_in_segment(si.segment, info, reply, request)))
      {
         return ret;
      }
      ret = ose_efmi_get_next_segment(&sc, &si);
   }

   return (ret == EFMI_END_OF_LIST) ? EFMI_OK : ret;
}

static EfmiStatus
process_info_set_sid(uint32_t *sid, struct EfmiProcessInfo *pi)
{
   EfmiStatus ret;
   struct EfmiBlockInfo bi;

   if (!(ret = ose_efmi_get_block_info(pi->block, &bi, sizeof(bi))))
   {
      *sid = bi.segment;
   }
   return ret;
}

static EfmiStatus
process_info_set_state(MonitorProcessState *state, struct EfmiProcessInfo *pi)
{
   EfmiStatus ret;
   struct EfmiCurrent cur;

   ret = EFMI_OK;

   if (!(*state = pi->status))
   {
      if(!(ret = ose_efmi_get_current(&cur, sizeof(cur))))
      {
		   if (pi->process == cur.process)
		      *state = MONITOR_PROCESS_STATE_RUNNING;
		   else
		      *state = MONITOR_PROCESS_STATE_READY;
      }
   }
   return ret;
}

static EfmiStatus
process_info_set_name(MonitorGetProcessInfoReply *info,
                      struct EfmiProcessInfo *pi)
{
   return ose_efmi_get_process_name(pi->process, info->name,
                                    FMM_OUTPUT_SIGNAL_SIZE - sizeof(*info));
}

static EfmiStatus
process_info_filter(MonitorGetProcessInfoReply *info,
                    uint32_t tagscount, const MonitorTag *tags)
{
   uint32_t i;
   uint32_t not = 0;

   for (i = 0; i < tagscount; i += tags[i].tag.count)
   {
      switch (tags[i].tag.type)
      {
         case MONITOR_PROCESS_FILTER_TAG_TYPE:
            if (not ^ (info->type != tags[i+1].num))
            {
               return EFMI_END_OF_LIST;
            }
            break;
         case MONITOR_PROCESS_FILTER_TAG_STATE:
            if (not ^ (info->state != tags[i+1].num))
            {
               return EFMI_END_OF_LIST;
            }
            break;
         case MONITOR_PROCESS_FILTER_TAG_PRIORITY:
            if (not ^ ((info->priority < tags[i+1].num)
                       || (info->priority >= tags[i+2].num)))
            {
               return EFMI_END_OF_LIST;
            }
            break;
         case MONITOR_PROCESS_FILTER_TAG_NAME:
            if (not ^ !match(tags[i+1].str, info->name))
            {
               return EFMI_END_OF_LIST;
            }
            break;
         case MONITOR_PROCESS_FILTER_TAG_SYSTEM:
            if (not ^ !(info->properties & 1))
            {
               return EFMI_END_OF_LIST;
            }
            break;
         default:
            break;
      }
      if (MONITOR_PROCESS_FILTER_TAG_NOT == tags[i].tag.type)
         not = 1;
      else
         not = 0;
   }
   return EFMI_OK;
}

static EfmiStatus
process_info(int32_t (*reply)(FmmSignal *sig),
             union ContextInfo *context,
             const union Signal *request)
{
   EfmiStatus ret;

   FmmSignal *sig;
   MonitorGetProcessInfoReply *info;

   sig = fmm_output_signal;

   info = &sig->data.get_process_info_reply;
   info->sigNo = MONITOR_GET_PROCESS_INFO_REPLY;
   info->pid = context->process_info.process;
   info->bid = context->process_info.block;
   if ((ret = process_info_set_sid(&info->sid, &context->process_info)))
   {
      return ret;
   }
   info->type = context->process_info.type;
   if ((ret = process_info_set_state(&info->state, &context->process_info)))
   {
      return ret;
   }
   info->priority = context->process_info.priority;
   info->entrypoint = context->process_info.entry_point;
   info->properties = 0;                            /* TODO: System process? */
   info->reserved1 = 0;
   if ((ret = process_info_set_name(info, &context->process_info)))
   {
      return ret;
   }
   sig->size = offsetof(MonitorGetProcessInfoReply, name) +
               strlen(info->name) + 1;

   if (!(ret = process_info_filter(info,
                                   request->get_info_request.tagsCount,
                                   request->get_info_request.tags)))
   {
      if (reply(sig))
      {
         return EFMI_END_OF_LIST;
      }
   }
   return ret == EFMI_END_OF_LIST ? EFMI_OK : ret;
}

static EfmiStatus
process_info_set_supervisor(MonitorBoolean *supervisor,
                            struct EfmiProcessInfo *pi)
{
   /* Unused supervisor stack means supervisor mode process */
   if (pi->supervisor_stack.top == pi->supervisor_stack.limit)
      *supervisor = 1;
   else
      *supervisor = 0;

   return EFMI_OK;
}

static EfmiStatus
process_info_set_signals_in_queue(uint32_t *inqueue, struct EfmiProcessInfo *pi)
{
   EfmiStatus ret;
   struct EfmiProcessStatistics s;

   if (!(ret = ose_efmi_get_process_statistics(pi->process,
                                               EFMI_SIG_CNT_IN_Q,
                                               &s)))
   {
      *inqueue = s.value;
   }
   return ret;
}

static EfmiStatus
process_info_set_sigselect(MonitorGetProcessInfoLongReply *info,
                           struct EfmiProcessInfo *pi)
{
   info->sigselectOffset = 0;

   if (pi->status & MONITOR_PROCESS_STATE_RECEIVE)
   {
      int32_t n, m;

      n = pi->wanted[0];
      n = (n < 0) ? -n : n;
      n = (n > EFMI_MAX_WANTED_SIGNALS) ? EFMI_MAX_WANTED_SIGNALS : n;
      m = (FMM_OUTPUT_SIGNAL_SIZE - sizeof(*info)) / sizeof(uint32_t);

      info->sigselectCount = (n > m) ? m : n;
      if (info->sigselectCount > 0)
      {
         memcpy(info->data, &pi->wanted[1], info->sigselectCount * sizeof(uint32_t));
      }
   }
   else
   {
      info->sigselectCount = 0;
   }
   return EFMI_OK;
}

static EfmiStatus
process_info_set_process_name(MonitorGetProcessInfoLongReply *info,
                              struct EfmiProcessInfo *pi)
{
   EfmiStatus ret;

   uint32_t n;

   n = info->sigselectCount * sizeof(uint32_t);

   ret = ose_efmi_get_process_name(pi->process, (char *)&info->data[n],
                                   FMM_OUTPUT_SIGNAL_SIZE - sizeof(*info) - n);
   info->processNameOffset = n;

   return ret;
}

static EfmiStatus
process_info_set_file_name(MonitorGetProcessInfoLongReply *info,
                           struct EfmiProcessInfo *pi)
{
   uint32_t n;

   n = info->processNameOffset;
   n += strlen((char *)&info->data[info->processNameOffset]) + 1;

   strncpy((char *)&info->data[n], pi->file, FMM_OUTPUT_SIGNAL_SIZE
                                                - sizeof(*info) - n - 1);
   info->data[FMM_OUTPUT_SIGNAL_SIZE - sizeof(*info)] = '\0';
   info->fileNameOffset = n;

   return EFMI_OK;
}

static EfmiStatus
process_info_long_filter(MonitorGetProcessInfoLongReply *info,
                         uint32_t tagscount, const MonitorTag *tags)
{
   uint32_t i;
   uint32_t not = 0;

   for (i = 0; i < tagscount; i += tags[i].tag.count)
   {
      switch (tags[i].tag.type)
      {
         case MONITOR_PROCESS_FILTER_TAG_TYPE:
            if (not ^ (info->type != tags[i+1].num))
            {
               return EFMI_END_OF_LIST;
            }
            break;
         case MONITOR_PROCESS_FILTER_TAG_STATE:
            if (not ^ (info->state != tags[i+1].num))
            {
               return EFMI_END_OF_LIST;
            }
            break;
         case MONITOR_PROCESS_FILTER_TAG_PRIORITY:
            if (not ^ ((info->priority < tags[i+1].num)
                       || (info->priority >= tags[i+2].num)))
            {
               return EFMI_END_OF_LIST;
            }
            break;
         case MONITOR_PROCESS_FILTER_TAG_UID:
            if (not ^ ((info->uid < tags[i+1].num)
                       || (info->uid >= tags[i+2].num)))
            {
               return EFMI_END_OF_LIST;
            }
            break;
         case MONITOR_PROCESS_FILTER_TAG_SIGNALS_IN_QUEUE:
            if (not ^ ((info->signalsInQueue < tags[i+1].num)
                       || (info->signalsInQueue >= tags[i+2].num)))
            {
               return EFMI_END_OF_LIST;
            }
            break;
         case MONITOR_PROCESS_FILTER_TAG_ENTRYPOINT:
            if (not ^ ((info->entrypoint < tags[i+1].num)
                       || (info->entrypoint >= tags[i+2].num)))
            {
               return EFMI_END_OF_LIST;
            }
            break;
         case MONITOR_PROCESS_FILTER_TAG_SEMAPHORE_VALUE:
            if (not ^ ((info->semaphoreValue < (int32_t)tags[i+1].num)
                       || (info->semaphoreValue >= (int32_t)tags[i+2].num)))
            {
               return EFMI_END_OF_LIST;
            }
            break;
         case MONITOR_PROCESS_FILTER_TAG_STACK_SIZE:
            if (not ^ ((info->stackSize < (tags[i+1].num)
                        || (info->stackSize >= tags[i+2].num))))
            {
               return EFMI_END_OF_LIST;
            }
            break;
         case MONITOR_PROCESS_FILTER_TAG_SUPERVISOR_STACK_SIZE:
            if (not ^ ((info->supervisorStackSize < tags[i+1].num)
                        || (info->supervisorStackSize >= tags[i+2].num)))
            {
               return EFMI_END_OF_LIST;
            }
            break;
         case MONITOR_PROCESS_FILTER_TAG_LINE_NUMBER:
            if (not ^ ((info->lineNumber < tags[i+1].num)
                        || (info->lineNumber >= tags[i+2].num)))
            {
               return EFMI_END_OF_LIST;
            }
            break;
         case MONITOR_PROCESS_FILTER_TAG_SIGNALS_OWNED:
            if (not ^ ((info->signalsOwned < tags[i+1].num)
                        || (info->signalsOwned >= tags[i+2].num)))
            {
               return EFMI_END_OF_LIST;
            }
            break;
         case MONITOR_PROCESS_FILTER_TAG_SIGSELECT_COUNT:
            if (not ^ ((info->sigselectCount == 0)
                       || (*(uint32_t *)info->data < tags[i+1].num)
                       || (*(uint32_t *)info->data >= tags[i+2].num)))
            {
               return EFMI_END_OF_LIST;
            }
            break;
         case MONITOR_PROCESS_FILTER_TAG_NAME:
            if (not ^ !match(tags[i+1].str,
                             (char *)&info->data[info->processNameOffset]))
            {
               return EFMI_END_OF_LIST;
            }
            break;
         case MONITOR_PROCESS_FILTER_TAG_FILE_NAME:
            if (not ^ ((info->fileNameOffset == 0)
                       || !match(tags[i+1].str,
                                 (char *)&info->data[info->fileNameOffset])))
            {
               return EFMI_END_OF_LIST;
            }
            break;
         case MONITOR_PROCESS_FILTER_TAG_SYSTEM:
            if (not ^ !(info->properties & 1))
            {
               return EFMI_END_OF_LIST;
            }
            break;
         default:
            break;
      }
      if (MONITOR_PROCESS_FILTER_TAG_NOT == tags[i].tag.type)
         not = 1;
      else
         not = 0;
   }
   return EFMI_OK;
}

static EfmiStatus
process_info_long(int32_t (*reply)(FmmSignal *sig),
                  union ContextInfo *context,
                  const union Signal *request)
{
   EfmiStatus ret;

   FmmSignal *sig;
   MonitorGetProcessInfoLongReply *info;

   sig = fmm_output_signal;

   info = &sig->data.get_process_info_long_reply;
   info->sigNo = MONITOR_GET_PROCESS_INFO_LONG_REPLY;
   info->pid = context->process_info.process;
   info->bid = context->process_info.block;
   if ((ret = process_info_set_sid(&info->sid, &context->process_info)))
   {
      return ret;
   }
   info->type = context->process_info.type;
   if ((ret = process_info_set_state(&info->state, &context->process_info)))
   {
      return ret;
   }
   info->priority = context->process_info.priority;
   info->uid = context->process_info.user;
   info->creator = context->process_info.parent;
   if ((ret = process_info_set_supervisor(&info->supervisor,
        &context->process_info)))
   {
      return ret;
   }
   if ((ret = process_info_set_signals_in_queue(&info->signalsInQueue,
        &context->process_info)))
   {
      return ret;
   }
   info->signalsAttached = 0; /* TODO: Signals attached */
   info->entrypoint = context->process_info.entry_point;
   info->semaphoreValue = context->process_info.fsemvalue;
   info->stackSize = context->process_info.process_stack.top -
                     context->process_info.process_stack.limit;
   info->supervisorStackSize = context->process_info.supervisor_stack.top -
                               context->process_info.supervisor_stack.limit;
   info->lineNumber = context->process_info.line;
   info->signalsOwned = context->process_info.sig_cnt_owned;
   info->timeSlice = 0;       /* TODO: Timeslice */
   info->vector = 0;          /* TODO: Vector */
   info->errorHandler = 0;    /* TODO: Errorhandler */
   info->properties = 0;      /* TODO: System process */
   info->reserved1 = 0;
   info->reserved2 = 0;
   if ((ret = process_info_set_sigselect(info, &context->process_info)))
   {
      return ret;
   }
   if ((ret = process_info_set_process_name(info, &context->process_info)))
   {
      return ret;
   }
   if ((ret = process_info_set_file_name(info, &context->process_info)))
   {
      return ret;
   }
   sig->size = offsetof(MonitorGetProcessInfoLongReply, data)
               + info->sigselectCount * sizeof(uint32_t)
               + strlen((char *)&info->data[info->processNameOffset]) + 1
               + strlen((char *)&info->data[info->fileNameOffset]) + 1;
   if (!(ret = process_info_long_filter(info,
                                        request->get_info_request.tagsCount,
                                        request->get_info_request.tags)))
   {
      if ((ret = reply(sig)))
      {
         return ret;
      }
   }
   return ret == EFMI_END_OF_LIST ? EFMI_OK : ret;
}


/*
 * Requests from monitor.sig
 */

EfmiStatus
fmm_efmi_get_process_info_request(
                           int32_t (*reply)(FmmSignal *sig),
                           int32_t (*endmark)(FmmSignal *sig, int32_t status),
                           const MonitorGetInfoRequest *request)
{
   EfmiStatus ret;
   EfmiStatus (*info)(int32_t (*reply)(FmmSignal *sig),
                      union ContextInfo *context,
                      const union Signal *request);

   FmmSignal *sig;

   assert(request != NULL);
   assert(request->sigNo == MONITOR_GET_PROCESS_INFO_REQUEST);

   info = request->level ? process_info_long : process_info;

   switch (request->scope.type)
   {
      case MONITOR_SCOPE_SYSTEM:
         ret = list_processes_in_system(info, reply,
                                        (const union Signal *)request);
         break;
      case MONITOR_SCOPE_SEGMENT_ID:
         ret = list_processes_in_segment(request->scope.id, info, reply,
                                         (const union Signal *)request);
         break;
      case MONITOR_SCOPE_BLOCK_ID:
         ret = list_processes_in_block(request->scope.id, info, reply,
                                       (const union Signal *)request);
         break;
      case MONITOR_SCOPE_ID:
         ret = list_process(request->scope.id, info, reply,
                            (const union Signal *)request);
         break;
      /* TODO: case MONITOR_SCOPE_NAME_PATTERN: (Not used in PMD) */
      /* TODO: case MONITOR_SCOPE_ENTRYPOINT: (Not used in PMD) */
      default:
         ret = EFMI_UNAVAILABLE;
         break;
   }

   sig = fmm_output_signal;
   sig->size = sizeof(MonitorGetInfoEndmark);
   sig->data.get_info_endmark.sigNo = MONITOR_GET_PROCESS_INFO_ENDMARK;

   switch (ret)
   {
      case EFMI_OK:
         sig->data.get_info_endmark.status = MONITOR_STATUS_OK;
         break;
      case EFMI_END_OF_LIST:
         sig->data.get_info_endmark.status = MONITOR_STATUS_CANCELLED;
         break;
      default:
         sig->data.get_info_endmark.status = MONITOR_STATUS_KERNEL_ERROR;
         break;
   }

   if (endmark(sig, ret))
   {
      return (ret) ? ret : EFMI_END_OF_LIST;
   }

   return ret;
}

static EfmiStatus
list_block(uint32_t id,
             EfmiStatus (*info)(int32_t (*reply)(FmmSignal *sig),
                                union ContextInfo *context,
                                const union Signal *request),
             int32_t (*reply)(FmmSignal *sig),
             const union Signal *request)
{
   EfmiStatus ret;

   union ContextInfo context;

   ret = ose_efmi_get_block_info(id, &context.block_info,
                                 sizeof(context.block_info));

   if (!ret)
   {
      if((ret = info(reply, &context, request)))
      {
         return ret;
      }
   }

   return ret;
}

static EfmiStatus
list_blocks_in_segment(uint32_t id,
                       EfmiStatus (*info)(int32_t (*reply)(FmmSignal *sig),
                                          union ContextInfo *context,
                                          const union Signal *request),
                       int32_t (*reply)(FmmSignal *sig),
                       const union Signal *request)
{
   EfmiStatus ret;

   struct EfmiBlockInfo bi;
   struct EfmiBlockCursor bc;

   ret = ose_efmi_get_first_block(id, &bc, &bi, sizeof(bi));

   while (!ret)
   {
      if ((ret = list_block(bi.block, info, reply, request)))
      {
         return ret;
      }
      ret = ose_efmi_get_next_block(&bc, &bi);
   }

   return (ret == EFMI_END_OF_LIST) ? EFMI_OK : ret;
}

static EfmiStatus
list_blocks_in_system(EfmiStatus (*info)(int32_t (*reply)(FmmSignal *sig),
                                         union ContextInfo *context,
                                         const union Signal *request),
                      int32_t (*reply)(FmmSignal *sig),
                      const union Signal *request)
{
   EfmiStatus ret;

   struct EfmiSegmentInfo si;
   struct EfmiSegmentCursor sc;

   ret = ose_efmi_get_first_segment(&sc, &si, sizeof(si));

   while (!ret)
   {
      if ((ret = list_blocks_in_segment(si.segment, info, reply, request)))
      {
         return ret;
      }
      ret = ose_efmi_get_next_segment(&sc, &si);
   }

   return (ret == EFMI_END_OF_LIST) ? EFMI_OK : ret;
}

static EfmiStatus
block_info_set_name(MonitorGetBlockInfoReply *info,
                    struct EfmiBlockInfo *bi)
{
   return ose_efmi_get_process_name(bi->block, info->name,
                                    FMM_OUTPUT_SIGNAL_SIZE - sizeof(*info));
}

static EfmiStatus
block_info_set_supervisor(MonitorBoolean *supervisor, struct EfmiBlockInfo *bi)
{
   if (bi->supervisor_mode)
      *supervisor = MONITOR_BOOLEAN_TRUE;
   else
      *supervisor = MONITOR_BOOLEAN_FALSE;

   return EFMI_OK;
}

static EfmiStatus
block_info_filter(MonitorGetBlockInfoReply *info,
                  uint32_t tagscount, const MonitorTag *tags)
{
   uint32_t i;
   uint32_t not = 0;

   for (i = 0; i < tagscount; i += tags[i].tag.count)
   {
      switch (tags[i].tag.type)
      {
         case MONITOR_BLOCK_FILTER_TAG_STACK_POOL_ID:
            if (not ^ ((info->stackPoolId < tags[i+1].num)
                       || (info->stackPoolId >= tags[i+2].num)))
            {
               return EFMI_END_OF_LIST;
            }
            break;
         case MONITOR_BLOCK_FILTER_TAG_SIG_POOL_ID:
            if (not ^ ((info->sigPoolId < tags[i+1].num)
                       || (info->sigPoolId >= tags[i+2].num)))
            {
               return EFMI_END_OF_LIST;
            }
            break;
         case MONITOR_BLOCK_FILTER_TAG_MAX_SIG_SIZE:
            if (not ^ ((info->maxSigSize < tags[i+1].num)
                       || (info->maxSigSize >= tags[i+2].num)))
            {
               return EFMI_END_OF_LIST;
            }
            break;
         case MONITOR_BLOCK_FILTER_TAG_UID:
            if (not ^ ((info->uid < tags[i+1].num)
                       || (info->uid >= tags[i+2].num)))
            {
               return EFMI_END_OF_LIST;
            }
            break;
         case MONITOR_PROCESS_FILTER_TAG_NAME:
            if (not ^ !match(tags[i+1].str, info->name))
            {
               return EFMI_END_OF_LIST;
            }
            break;
         default:
            break;
      }
      if (MONITOR_PROCESS_FILTER_TAG_NOT == tags[i].tag.type)
         not = 1;
      else
         not = 0;
   }
   return EFMI_OK;
}

static EfmiStatus
block_info(int32_t (*reply)(FmmSignal *sig),
           union ContextInfo *context,
           const union Signal *request)
{
   EfmiStatus ret;

   FmmSignal *sig;
   MonitorGetBlockInfoReply *info;

   sig = fmm_output_signal;

   info = &sig->data.get_block_info_reply;
   info->sigNo = MONITOR_GET_BLOCK_INFO_REPLY;
   info->bid = context->block_info.block;
   info->sid = context->block_info.segment;
   info->uid = context->block_info.user;
   if ((ret = block_info_set_supervisor(&info->supervisor, &context->block_info)))
   {
      return ret;
   }
   info->signalsAttached = 0; /* TODO: Signals attached */
   info->errorHandler = 0;    /* TODO: Errorhandler */
   info->maxSigSize = 0;      /* TODO: Max sig size */
   info->sigPoolId = context->block_info.sig_pool;
   info->stackPoolId = context->block_info.stack_pool;
   info->reserved0 = 0;
   info->reserved1 = 0;
   if ((ret = block_info_set_name(info, &context->block_info)))
   {
      return ret;
   }
   sig->size = offsetof(MonitorGetBlockInfoReply, name) +
               strlen(info->name) + 1;

   if (!(ret = block_info_filter(info,
                                 request->get_info_request.tagsCount,
                                 request->get_info_request.tags)))
   {
      if ((ret = reply(sig)))
      {
         return ret;
      }
   }
   return ret == EFMI_END_OF_LIST ? EFMI_OK : ret;
}

EfmiStatus
fmm_efmi_get_block_info_request(int32_t (*reply)(FmmSignal *sig),
                                int32_t (*endmark)(FmmSignal *sig, int32_t status),
                                const MonitorGetInfoRequest *request)
{
   EfmiStatus ret;
   FmmSignal *sig;

   assert(request != NULL);
   assert(request->sigNo == MONITOR_GET_BLOCK_INFO_REQUEST);

   switch (request->scope.type)
   {
      case MONITOR_SCOPE_SYSTEM:
         ret = list_blocks_in_system(block_info, reply,
                                     (const union Signal *)request);
         break;
      case MONITOR_SCOPE_SEGMENT_ID:
         ret = list_blocks_in_segment(request->scope.id, block_info, reply,
                                      (const union Signal *)request);
         break;
      case MONITOR_SCOPE_BLOCK_ID:
         ret = EFMI_UNAVAILABLE;
         break;
      case MONITOR_SCOPE_ID:
         ret = list_block(request->scope.id, block_info, reply,
                          (const union Signal *)request);
         break;
      /* TODO: case MONITOR_SCOPE_NAME_PATTERN: (Not used in PMD) */
      /* TODO: case MONITOR_SCOPE_ENTRYPOINT: (Not used in PMD) */
      default:
         ret = EFMI_UNAVAILABLE;
         break;
   }

   sig = fmm_output_signal;
   sig->size = sizeof(MonitorGetInfoEndmark);
   sig->data.get_info_endmark.sigNo = MONITOR_GET_BLOCK_INFO_ENDMARK;

   switch (ret)
   {
      case EFMI_OK:
         sig->data.get_info_endmark.status = MONITOR_STATUS_OK;
         break;
      case EFMI_END_OF_LIST:
         sig->data.get_info_endmark.status = MONITOR_STATUS_CANCELLED;
         break;
      default:
         sig->data.get_info_endmark.status = MONITOR_STATUS_KERNEL_ERROR;
         break;
   }

   if (endmark(sig, ret))
   {
      return (ret) ? ret : EFMI_END_OF_LIST;
   }

   return ret;
}

static EfmiStatus
list_segment(uint32_t id,
             EfmiStatus (*info)(int32_t (*reply)(FmmSignal *sig),
                                union ContextInfo *context,
                                const union Signal *request),
             int32_t (*reply)(FmmSignal *sig),
             const union Signal *request)
{
   EfmiStatus ret;

   union ContextInfo context;

   ret = ose_efmi_get_segment_info(id, &context.segment_info,
                                   sizeof(context.segment_info));

   if (!ret)
   {
      if((ret = info(reply, &context, (const union Signal *)request)))
      {
         return ret;
      }
   }

   return ret;
}

static EfmiStatus
list_segments_in_system(EfmiStatus (*info)(int32_t (*reply)(FmmSignal *sig),
                                           union ContextInfo *context,
                                           const union Signal *request),
                        int32_t (*reply)(FmmSignal *sig),
                        const union Signal *request)
{
   EfmiStatus ret;

   struct EfmiSegmentInfo si;
   struct EfmiSegmentCursor sc;

   ret = ose_efmi_get_first_segment(&sc, &si, sizeof(si));

   while (!ret)
   {
      if ((ret = list_segment(si.segment, info, reply,
                              (const union Signal *)request)))
      {
         return ret;
      }
      ret = ose_efmi_get_next_segment(&sc, &si);
   }

   return (ret == EFMI_END_OF_LIST) ? EFMI_OK : ret;
}

static EfmiStatus
segment_info(int32_t (*reply)(FmmSignal *sig),
             union ContextInfo *context,
             const union Signal *request)
{
   EfmiStatus ret = EFMI_OK;

   FmmSignal *sig;
   MonitorGetSegmentInfoReply *info;

   (void)request; /* Not used */
   
   sig = fmm_output_signal;

   info = &sig->data.get_segment_info_reply;
   info->sigNo = MONITOR_GET_SEGMENT_INFO_REPLY;
   info->sid = context->segment_info.segment;
   info->number = context->segment_info.segid; /* FIXME: Is info->number same as si->segid? */
   info->reserved0 = 0;
   info->reserved1 = 0;
   info->reserved2 = 0;
   sig->size = sizeof(*info);

   if ((ret = reply(sig)))
   {
      return ret;
   }

   return ret == EFMI_END_OF_LIST ? EFMI_OK : ret;
}

EfmiStatus
fmm_efmi_get_segment_info_request(int32_t (*reply)(FmmSignal *sig),
                                  int32_t (*endmark)(FmmSignal *sig, int32_t status),
                                  const MonitorGetInfoRequest *request)
{
   EfmiStatus ret;

   FmmSignal *sig;

   assert(request != NULL);
   assert(request->sigNo == MONITOR_GET_SEGMENT_INFO_REQUEST);

   switch (request->scope.type)
   {
      case MONITOR_SCOPE_SYSTEM:
         ret = list_segments_in_system(segment_info, reply,
                                       (const union Signal *)request);
         break;
      case MONITOR_SCOPE_SEGMENT_ID:
         ret = EFMI_UNAVAILABLE;
         break;
      case MONITOR_SCOPE_BLOCK_ID:
         ret = EFMI_UNAVAILABLE;
         break;
      case MONITOR_SCOPE_ID:
         ret = list_segment(request->scope.id, segment_info, reply,
                            (const union Signal *)request);
         break;
      /* TODO: case MONITOR_SCOPE_NAME_PATTERN: (Not used in PMD) */
      /* TODO: case MONITOR_SCOPE_ENTRYPOINT: (Not used in PMD) */
      default:
         ret = EFMI_UNAVAILABLE;
         break;
   }

   sig = fmm_output_signal;
   sig->size = sizeof(MonitorGetInfoEndmark);
   sig->data.get_info_endmark.sigNo = MONITOR_GET_SEGMENT_INFO_ENDMARK;

   switch (ret)
   {
      case EFMI_OK:
         sig->data.get_info_endmark.status = MONITOR_STATUS_OK;
         break;
      case EFMI_END_OF_LIST:
         sig->data.get_info_endmark.status = MONITOR_STATUS_CANCELLED;
         break;
      default:
         sig->data.get_info_endmark.status = MONITOR_STATUS_KERNEL_ERROR;
         break;
   }

   if (endmark(sig, ret))
   {
      return (ret) ? ret : EFMI_END_OF_LIST;
   }

   return ret;
}

static EfmiStatus
stack_usage_set_main_used(uint32_t *used, struct EfmiProcessInfo *pi)
{
   EfmiStatus ret;
   struct EfmiProcessStatistics s;

   if (pi->type == OS_BG_PROC || pi->type == OS_PRI_PROC)
   {
      if (!(ret = ose_efmi_get_process_statistics(pi->process, EFMI_PROCESS_STACK_USAGE, &s)))
      {
         *used = s.value;
      }
   }
   else
   {
      *used = 0;
      ret = EFMI_OK;
   }

   return ret;
}

static EfmiStatus
stack_usage_set_main_bufsize(uint32_t *bufsize, struct EfmiProcessInfo *pri)
{
   EfmiStatus ret;
   struct EfmiBlockInfo bi;

   uint32_t size = pri->process_stack.top - pri->process_stack.limit;

   if (!(ret = ose_efmi_get_block_info(pri->block, &bi, sizeof(bi))))
   {
      struct EfmiPoolInfo poi;

      if (!(ret = ose_efmi_get_pool_info(bi.stack_pool, &poi, sizeof(poi))))
      {
         uint32_t i;

         for (i = 0; i < EFMI_MAX_STACK_BUFFER_SIZE_COUNT; i++)
         {
            if (size <= poi.stack_buf_sizes[i].configured_size)
            {
               *bufsize = poi.stack_buf_sizes[i].configured_size;
               break;
            }
         }
      }
   }
   return ret;
}

static EfmiStatus
stack_usage_set_supervisor_used(uint32_t *used, struct EfmiProcessInfo *pi)
{
   EfmiStatus ret = EFMI_OK;

#if 0
   struct EfmiProcessStatistics s;

   if (!(ret = ose_efmi_get_process_statistics(pi->process, EFMI_SUPERVISOR_STACK_USAGE, &s)))
   {
      *used = s.value;
   }
#else
   (void)pi; /* Not used */
#endif

   /* EFMI is currently unable to get
      the supervisor stack usage
      information, so set it to 0 for now. */
   *used = 0;

   return ret;
}

static EfmiStatus
stack_usage_set_supervisor_bufsize(uint32_t *bufsize, struct EfmiProcessInfo *pi)
{
   (void)pi; /* Not used */
   
   *bufsize = 0;     /* TODO: Supervisor bufsize? We don't know which pool,
                      * or if it is in the stack or signal part of that pool,
                      * so there is not much to do. */
   return EFMI_OK;
}

static EfmiStatus
stack_usage_filter(MonitorGetStackUsageReply *info,
                   uint32_t tagscount, const MonitorTag *tags)
{
   uint32_t i;
   uint32_t not = 0;

   for (i = 0; i < tagscount; i += tags[i].tag.count)
   {
      switch (tags[i].tag.type)
      {
         case MONITOR_STACK_FILTER_TAG_OWNER_ID:
            if (not ^ (info->pid != tags[i+1].num))
            {
               return EFMI_END_OF_LIST;
            }
            break;
         case MONITOR_STACK_FILTER_TAG_OWNER_NAME:
         {
            EfmiStatus ret;
            void *name = &info[1];

            if ((ret = ose_efmi_get_process_name(info->pid, name,
                              FMM_OUTPUT_SIGNAL_SIZE - sizeof(*info))))
            {
               return ret;
            }
            if (not ^ (!match(tags[i+1].str, name)))
            {
               return EFMI_END_OF_LIST;
            }
            break;
         }
         case MONITOR_STACK_FILTER_TAG_SIZE:
            if (not ^ ((info->main.size < tags[i+1].num)
                       || (info->main.size >= tags[i+2].num)))
            {
               return EFMI_END_OF_LIST;
            }
            break;
         case MONITOR_STACK_FILTER_TAG_BUFSIZE:
            if (not ^ ((info->main.bufsize < tags[i+1].num)
                       || (info->main.bufsize >= tags[i+2].num)))
            {
               return EFMI_END_OF_LIST;
            }
            break;
         case MONITOR_STACK_FILTER_TAG_ADDRESS:
            if (not ^ ((info->main.address < tags[i+1].num)
                       || (info->main.address >= tags[i+2].num)))
            {
               return EFMI_END_OF_LIST;
            }
            break;
         case MONITOR_STACK_FILTER_TAG_USED:
            if (not ^ ((info->main.used < tags[i+1].num)
                       || (info->main.used >= tags[i+2].num)))
            {
               return EFMI_END_OF_LIST;
            }
            break;
         case MONITOR_STACK_FILTER_TAG_UNUSED:
         {
            uint32_t unused = info->main.size - info->main.used;

            if (not ^ ((unused < tags[i+1].num)
                       || (unused >= tags[i+2].num)))
            {
               return EFMI_END_OF_LIST;
            }
            break;
         }
         case MONITOR_STACK_FILTER_TAG_SUPERVISOR_SIZE:
            if (not ^ ((info->supervisor.size < tags[i+1].num)
                       || (info->supervisor.size >= tags[i+2].num)))
            {
               return EFMI_END_OF_LIST;
            }
            break;
         case MONITOR_STACK_FILTER_TAG_SUPERVISOR_BUFSIZE:
            if (not ^ ((info->supervisor.bufsize < tags[i+1].num)
                       || (info->supervisor.bufsize >= tags[i+2].num)))
            {
               return EFMI_END_OF_LIST;
            }
            break;
         case MONITOR_STACK_FILTER_TAG_SUPERVISOR_ADDRESS:
            if (not ^ ((info->supervisor.address < tags[i+1].num)
                       || (info->supervisor.address >= tags[i+2].num)))
            {
               return EFMI_END_OF_LIST;
            }
            break;
         case MONITOR_STACK_FILTER_TAG_SUPERVISOR_USED:
            if (not ^ ((info->supervisor.used < tags[i+1].num)
                       || (info->supervisor.used >= tags[i+2].num)))
            {
               return EFMI_END_OF_LIST;
            }
            break;
         case MONITOR_STACK_FILTER_TAG_SUPERVISOR_UNUSED:
         {
            uint32_t unused = info->supervisor.size - info->supervisor.used;

            if (not ^ ((unused < tags[i+1].num)
                       || (unused >= tags[i+2].num)))
            {
               return EFMI_END_OF_LIST;
            }
            break;
         }
         default:
            break;
      }
      if (MONITOR_STACK_FILTER_TAG_NOT == tags[i].tag.type)
         not = 1;
      else
         not = 0;
   }
   return EFMI_OK;
}

static EfmiStatus
stack_usage(int32_t (*reply)(FmmSignal *sig),
            union ContextInfo *context,
            const union Signal *request)
{
   EfmiStatus ret;

   FmmSignal *sig;
   MonitorGetStackUsageReply *info;

   sig = fmm_output_signal;

   info = &sig->data.get_stack_usage_reply;
   info->sigNo = MONITOR_GET_STACK_USAGE_REPLY;
   info->pid = context->process_info.process;
   info->main.address = context->process_info.process_stack.limit;
   info->main.size = context->process_info.process_stack.top -
                     context->process_info.process_stack.limit;
   if ((ret = stack_usage_set_main_used(&info->main.used,
                                        &context->process_info)))
   {
      return ret;
   }
   if ((ret = stack_usage_set_main_bufsize(&info->main.bufsize,
                                           &context->process_info)))
   {
      return ret;
   }
   info->main.reserved1 = 0;
   info->main.reserved2 = 0;
   info->supervisor.address = context->process_info.supervisor_stack.limit;
   info->supervisor.size = context->process_info.supervisor_stack.top -
                           context->process_info.supervisor_stack.limit;
   if ((ret = stack_usage_set_supervisor_used(&info->supervisor.used,
                                              &context->process_info)))
   {
      return ret;
   }
   if ((ret = stack_usage_set_supervisor_bufsize(&info->supervisor.bufsize,
                                                 &context->process_info)))
   {
      return ret;
   }
   info->supervisor.reserved1 = 0;
   info->supervisor.reserved2 = 0;
   sig->size = sizeof(*info);

   if (!(ret = stack_usage_filter(info,
                                  request->get_info_request.tagsCount,
                                  request->get_info_request.tags)))
   {
      if ((ret = reply(sig)))
      {
         return ret;
      }
   }
   return ret == EFMI_END_OF_LIST ? EFMI_OK : ret;
}

EfmiStatus
fmm_efmi_get_stack_usage_request(int32_t (*reply)(FmmSignal *sig),
                                 int32_t (*endmark)(FmmSignal *sig, int32_t status),
                                 const MonitorGetInfoRequest *request)
{
   EfmiStatus ret;

   FmmSignal *sig;

   assert(request != NULL);
   assert(request->sigNo == MONITOR_GET_STACK_USAGE_REQUEST);

   switch (request->scope.type)
   {
      case MONITOR_SCOPE_SYSTEM:
         ret = list_processes_in_system(stack_usage, reply,
                                        (const union Signal *)request);
         break;
      case MONITOR_SCOPE_SEGMENT_ID:
         ret = list_processes_in_segment(request->scope.id, stack_usage, reply,
                                         (const union Signal *)request);
         break;
      case MONITOR_SCOPE_BLOCK_ID:
         ret = list_processes_in_block(request->scope.id, stack_usage, reply,
                                       (const union Signal *)request);
         break;
      case MONITOR_SCOPE_ID:
         ret = list_process(request->scope.id, stack_usage, reply,
                            (const union Signal *)request);
         break;
      /* TODO: case MONITOR_SCOPE_NAME_PATTERN: (Not used in PMD) */
      /* TODO: case MONITOR_SCOPE_ENTRYPOINT: (Not used in PMD) */
      default:
         ret = EFMI_UNAVAILABLE;
         break;
   }

   sig = fmm_output_signal;
   sig->size = sizeof(MonitorGetInfoEndmark);
   sig->data.get_info_endmark.sigNo = MONITOR_GET_STACK_USAGE_ENDMARK;

   switch (ret)
   {
      case EFMI_OK:
         sig->data.get_info_endmark.status = MONITOR_STATUS_OK;
         break;
      case EFMI_END_OF_LIST:
         sig->data.get_info_endmark.status = MONITOR_STATUS_CANCELLED;
         break;
      default:
         sig->data.get_info_endmark.status = MONITOR_STATUS_KERNEL_ERROR;
         break;
   }

   if (endmark(sig, ret))
   {
      return (ret) ? ret : EFMI_END_OF_LIST;
   }

   return ret;
}

static EfmiStatus
env_vars(int32_t (*reply)(FmmSignal *sig),
         union ContextInfo *context,
         const union Signal *request)
{
   EfmiStatus ret;

   FmmSignal *sig;
   MonitorGetEnvVarsReply *info;
   struct EfmiEnvVarCursor cursor;

   size_t total_size;
   char *name;
   size_t max_name_len;
   char *value;
   size_t max_value_len;

   (void)request; /* Not used */
   
   sig = fmm_output_signal;
   info = &sig->data.get_env_vars_reply;

   total_size = FMM_OUTPUT_SIGNAL_SIZE - sizeof(*info);
   max_name_len = total_size / 2;
   max_value_len = total_size - max_name_len;
   name = (char *)info->data;
   value = &name[max_name_len];

   ret = ose_efmi_get_first_env_var(context->id, &cursor, name, max_name_len,
                                    value, max_value_len);
   while (ret == EFMI_OK)
   {
      sig = fmm_output_signal;
      info = &sig->data.get_env_vars_reply;
      info->sigNo = MONITOR_GET_ENV_VARS_REPLY;
      info->pid = context->id;
      info->reserved0 = 0;
      info->reserved1 = 0;
      info->nameOffset = 0;
      info->valueOffset = strlen(name) + 1;

      memmove(&info->data[info->valueOffset], value, max_value_len);
      value = &info->data[info->valueOffset];

      /* Is value an envp value? */
      if (strlen(value) == 8
          && (value[4] == 0x40 || value[4] == 0x60)
          && (value[5] == 0x40 || value[5] == 0x60)
          && (value[6] == 0x40 || value[6] == 0x60)
          && (value[7] == 0x40 || value[7] == 0x60))
      {
         /*
          * The pointer value is obtained by xor-ing the two
          * words in the value.
          */
         OSADDRESS a;
         OSADDRESS b;
         OSADDRESS pval;
         memcpy(&a, value, sizeof(OSADDRESS));
         memcpy(&b, value, sizeof(OSADDRESS));
         pval = a ^ b;
         sprintf(value, "0x%X", pval);
      }

      sig->size = offsetof(MonitorGetEnvVarsReply, data) +
                  strlen(name) + 1 + strlen(value) + 1;

      if (reply(sig))
      {
         ret = EFMI_END_OF_LIST;
         break;
      }

      /* Reset value pointer */
      value = &name[max_name_len];
      ret = ose_efmi_get_next_env_var(&cursor, name, value);
   }

   return ret == EFMI_END_OF_LIST ? EFMI_OK : ret;
}


EfmiStatus
fmm_efmi_get_env_vars_request(int32_t (*reply)(FmmSignal *sig),
                              int32_t (*endmark)(FmmSignal *sig, int32_t status),
                              const MonitorGetInfoRequest *request)
{
   EfmiStatus ret = EFMI_UNAVAILABLE;
   union ContextInfo context;

   FmmSignal *sig;

   assert(request != NULL);
   assert(request->sigNo == MONITOR_GET_ENV_VARS_REQUEST);

   switch (request->scope.type)
   {
      case MONITOR_SCOPE_SYSTEM:
         ret = list_blocks_in_system(env_vars, reply,
                                     (const union Signal *)request);
         if (ret != EFMI_OK)
         {
            break;
         }
         ret = list_processes_in_system(env_vars, reply,
                                        (const union Signal *)request);
         break;
      case MONITOR_SCOPE_SEGMENT_ID:
         ret = list_blocks_in_segment(request->scope.id, env_vars, reply,
                                      (const union Signal *)request);
         if (ret != EFMI_OK)
         {
            break;
         }
         ret = list_processes_in_segment(request->scope.id, env_vars, reply,
                                         (const union Signal *)request);
         break;
      case MONITOR_SCOPE_BLOCK_ID:
         context.id = request->scope.id;
         ret = env_vars(reply, &context, (const union Signal *)request);
         if (ret != EFMI_OK)
         {
            break;
         }
         ret = list_processes_in_block(request->scope.id, env_vars, reply,
                                       (const union Signal *)request);
         break;
      case MONITOR_SCOPE_ID:
         context.id = request->scope.id;
         ret = env_vars(reply, &context, (const union Signal *)request);
         break;
      /* TODO: case MONITOR_SCOPE_NAME_PATTERN: (Not used in PMD) */
      /* TODO: case MONITOR_SCOPE_ENTRYPOINT: (Not used in PMD) */
      default:
         ret = EFMI_UNAVAILABLE;
         break;
   }

   sig = fmm_output_signal;
   sig->size = sizeof(MonitorGetInfoEndmark);
   sig->data.get_info_endmark.sigNo = MONITOR_GET_ENV_VARS_ENDMARK;

   switch (ret)
   {
      case EFMI_OK:
         sig->data.get_info_endmark.status = MONITOR_STATUS_OK;
         break;
      case EFMI_END_OF_LIST:
         sig->data.get_info_endmark.status = MONITOR_STATUS_CANCELLED;
         break;
      default:
         sig->data.get_info_endmark.status = MONITOR_STATUS_KERNEL_ERROR;
         break;
   }

   if (endmark(sig, ret))
   {
      return (ret) ? ret : EFMI_END_OF_LIST;
   }
   return ret;
}

static EfmiStatus
signal_queue_filter(MonitorGetSignalQueueReply *info,
                    uint32_t tagscount, const MonitorTag *tags)
{
   uint32_t i;
   uint32_t not = 0;

   for (i = 0; i < tagscount; i += tags[i].tag.count)
   {
      switch (tags[i].tag.type)
      {
         case MONITOR_SIGNAL_FILTER_TAG_NUMBER:
            if (not ^ (info->sigNo != tags[i+1].num))
            {
               return EFMI_END_OF_LIST;
            }
            break;
         case MONITOR_SIGNAL_FILTER_TAG_OWNER_ID:
            if (not ^ (info->signal.owner != tags[i+1].num))
            {
               return EFMI_END_OF_LIST;
            }
            break;
         case MONITOR_SIGNAL_FILTER_TAG_OWNER_NAME:
         {
            EfmiStatus ret;
            void *name = &info[1];

            if ((ret = ose_efmi_get_process_name(info->signal.owner, name,
                              FMM_OUTPUT_SIGNAL_SIZE - sizeof(*info))))
            {
               return ret;
            }
            if (not ^ (!match(tags[i+1].str, name)))
            {
               return EFMI_END_OF_LIST;
            }
            break;
         }
         case MONITOR_SIGNAL_FILTER_TAG_SENDER_ID:
            if (not ^ (info->signal.sender != tags[i+1].num))
            {
               return EFMI_END_OF_LIST;
            }
            break;
         case MONITOR_SIGNAL_FILTER_TAG_SENDER_NAME:
         {
            EfmiStatus ret;
            void *name = &info[1];

            if ((ret = ose_efmi_get_process_name(info->signal.sender, name,
                              FMM_OUTPUT_SIGNAL_SIZE - sizeof(*info))))
            {
               return ret;
            }
            if (not ^ (!match(tags[i+1].str, name)))
            {
               return EFMI_END_OF_LIST;
            }
            break;
         }
         case MONITOR_SIGNAL_FILTER_TAG_ADDRESSEE_ID:
            if (not ^ (info->signal.addressee != tags[i+1].num))
            {
               return EFMI_END_OF_LIST;
            }
            break;
         case MONITOR_SIGNAL_FILTER_TAG_ADDRESSEE_NAME:
         {
            EfmiStatus ret;
            void *name = &info[1];

            if ((ret = ose_efmi_get_process_name(info->signal.addressee, name,
                              FMM_OUTPUT_SIGNAL_SIZE - sizeof(*info))))
            {
               return ret;
            }
            if (not ^ (!match(tags[i+1].str, name)))
            {
               return EFMI_END_OF_LIST;
            }
            break;
         }
         case MONITOR_SIGNAL_FILTER_TAG_SIZE:
            if (not ^ ((info->signal.size < tags[i+1].num)
                       || (info->signal.size >= tags[i+2].num)))
            {
               return EFMI_END_OF_LIST;
            }
            break;
         case MONITOR_SIGNAL_FILTER_TAG_BUFSIZE:
            if (not ^ ((info->signal.bufsize < tags[i+1].num)
                       || (info->signal.bufsize >= tags[i+2].num)))
            {
               return EFMI_END_OF_LIST;
            }
            break;
         case MONITOR_SIGNAL_FILTER_TAG_ADDRESS:
            if (not ^ ((info->signal.address < tags[i+1].num)
                       || (info->signal.address >= tags[i+2].num)))
            {
               return EFMI_END_OF_LIST;
            }
            break;
         case MONITOR_SIGNAL_FILTER_TAG_STATUS:
            if (not ^ (info->signal.status != tags[i+1].num))
            {
               return EFMI_END_OF_LIST;
            }
            break;
         default:
            break;
      }
      if (MONITOR_STACK_FILTER_TAG_NOT == tags[i].tag.type)
         not = 1;
      else
         not = 0;
   }
   return EFMI_OK;
}

static EfmiStatus
signal_queue_set_status(MonitorSignalStatus *status,
                        const struct EfmiSignalInfo *sig_info,
                        const MonitorScope *scope)
{
   if (!sig_info->owner && sig_info->addressee_pid == scope->id)
   {
      *status = MONITOR_SIGNAL_STATUS_FREE;
   }
   else
   {
      switch (sig_info->sig_status)
      {
         case EFMI_SS_VALID:
            *status = MONITOR_SIGNAL_STATUS_VALID;
            break;
         case EFMI_SS_NO_ENDMARK:
            *status = MONITOR_SIGNAL_STATUS_ENDMARK;
            break;
         case EFMI_SS_BROKEN_SIGADM:
            *status = MONITOR_SIGNAL_STATUS_ADMINISTRATION;
            break;
         case EFMI_SS_WILD_POINTER_INSIDE_POOL:
            *status = MONITOR_SIGNAL_STATUS_NOT_A_SIGNAL;
            break;
         case EFMI_SS_WILD_POINTER_OUTSIDE_POOL:
            *status = MONITOR_SIGNAL_STATUS_NOT_A_POOL;
            break;
      }
   }

   return EFMI_OK;
}

static EfmiStatus
signal_queue(int32_t (*reply)(FmmSignal *sig),
                  union ContextInfo *context,
                  const union Signal *request)
{
   EfmiStatus ret;

   FmmSignal *sig;
   MonitorGetSignalQueueReply *info;
   uint32_t max_signals = fmd_properties.max_signals_in_queue;
   struct EfmiSignalCursor cursor;
   struct EfmiSignalInfo sig_info;

   ret = ose_efmi_get_first_q_signal(context->process_info.process, &cursor,
                                     &sig_info, sizeof(sig_info));
   while (ret == EFMI_OK)
   {
      sig = fmm_output_signal;
      info = &sig->data.get_signal_queue_reply;
      info->sigNo = MONITOR_GET_SIGNAL_QUEUE_REPLY;
      info->pid = context->process_info.process;
      info->signal.number = sig_info.sig_no;
      info->signal.owner = sig_info.owner;
      info->signal.sender = sig_info.sender_pid;
      info->signal.addressee = sig_info.addressee_pid;
      info->signal.size = sig_info.sig_size;
      info->signal.bufsize = sig_info.size_in_pool;
      info->signal.address = sig_info.signal_pointer;

      ret = signal_queue_set_status(&info->signal.status, &sig_info,
                                    &request->get_info_request.scope);
      if (ret != EFMI_OK)
      {
         break;
      }
      sig->size = sizeof(*info);

      ret = signal_queue_filter(info, request->get_info_request.tagsCount,
                                request->get_info_request.tags);
      if (ret == EFMI_OK)
      {
         if ((ret =reply(sig)))
         {
            return ret;
         }
      }
      else
      {
         break;
      }
      /* dump signal info for max number of signals specified */
      if (!fmd_properties.max_signals_in_queue)
      {
         ret = ose_efmi_get_next_q_signal(&cursor, &sig_info);
      }
      else
      {

         max_signals--;
	 if (max_signals != 0)
         {
	    ret = ose_efmi_get_next_q_signal(&cursor, &sig_info);
         }
	 else
         {
	    return ret = EFMI_OK;
         }
      }

   }

   return ret == EFMI_END_OF_LIST ? EFMI_OK : ret;
}

static EfmiStatus
signal_queue_long_filter(MonitorGetSignalQueueLongReply *info,
                    uint32_t tagscount, const MonitorTag *tags)
{
   uint32_t i;
   uint32_t not = 0;

   for (i = 0; i < tagscount; i += tags[i].tag.count)
   {
      switch (tags[i].tag.type)
      {
         case MONITOR_SIGNAL_FILTER_TAG_NUMBER:
            if (not ^ (info->sigNo != tags[i+1].num))
            {
               return EFMI_END_OF_LIST;
            }
            break;
         case MONITOR_SIGNAL_FILTER_TAG_OWNER_ID:
            if (not ^ (info->pid != tags[i+1].num))
            {
               return EFMI_END_OF_LIST;
            }
            break;
         case MONITOR_SIGNAL_FILTER_TAG_OWNER_NAME:
         {
            EfmiStatus ret;
            void *name = &info[1];

            if ((ret = ose_efmi_get_process_name(info->pid, name,
                              FMM_OUTPUT_SIGNAL_SIZE - sizeof(*info))))
            {
               return ret;
            }
            if (not ^ (!match(tags[i+1].str, name)))
            {
               return EFMI_END_OF_LIST;
            }
            break;
         }
         case MONITOR_SIGNAL_FILTER_TAG_SENDER_ID:
            if (not ^ (info->signal.sender != tags[i+1].num))
            {
               return EFMI_END_OF_LIST;
            }
            break;
         case MONITOR_SIGNAL_FILTER_TAG_SENDER_NAME:
         {
            EfmiStatus ret;
            void *name = &info[1];

            if ((ret = ose_efmi_get_process_name(info->signal.sender, name,
                              FMM_OUTPUT_SIGNAL_SIZE - sizeof(*info))))
            {
               return ret;
            }
            if (not ^ (!match(tags[i+1].str, name)))
            {
               return EFMI_END_OF_LIST;
            }
            break;
         }
         case MONITOR_SIGNAL_FILTER_TAG_ADDRESSEE_ID:
            if (not ^ (info->signal.addressee != tags[i+1].num))
            {
               return EFMI_END_OF_LIST;
            }
            break;
         case MONITOR_SIGNAL_FILTER_TAG_ADDRESSEE_NAME:
         {
            EfmiStatus ret;
            void *name = &info[1];

            if ((ret = ose_efmi_get_process_name(info->signal.addressee, name,
                              FMM_OUTPUT_SIGNAL_SIZE - sizeof(*info))))
            {
               return ret;
            }
            if (not ^ (!match(tags[i+1].str, name)))
            {
               return EFMI_END_OF_LIST;
            }
            break;
         }
         case MONITOR_SIGNAL_FILTER_TAG_SIZE:
            if (not ^ ((info->signal.size < tags[i+1].num)
                       || (info->signal.size >= tags[i+2].num)))
            {
               return EFMI_END_OF_LIST;
            }
            break;
         case MONITOR_SIGNAL_FILTER_TAG_BUFSIZE:
            if (not ^ ((info->signal.bufsize < tags[i+1].num)
                       || (info->signal.bufsize >= tags[i+2].num)))
            {
               return EFMI_END_OF_LIST;
            }
            break;
         case MONITOR_SIGNAL_FILTER_TAG_ADDRESS:
            if (not ^ ((info->signal.address < tags[i+1].num)
                       || (info->signal.address >= tags[i+2].num)))
            {
               return EFMI_END_OF_LIST;
            }
            break;
         case MONITOR_SIGNAL_FILTER_TAG_STATUS:
            if (not ^ (info->signal.status != tags[i+1].num))
            {
               return EFMI_END_OF_LIST;
            }
            break;
         default:
            break;
      }
      if (MONITOR_STACK_FILTER_TAG_NOT == tags[i].tag.type)
         not = 1;
      else
         not = 0;
   }
   return EFMI_OK;
}

static EfmiStatus
signal_queue_long_set_data(MonitorGetSignalQueueLongReply *info,
                           uint32_t pid,
                           const struct EfmiSignalInfo *sig_info)
{
   EfmiStatus ret = EFMI_OK;

   assert(info != NULL);
   assert(sig_info != NULL);

   if (sig_info->sig_status == EFMI_SS_VALID ||
       sig_info->sig_status == EFMI_SS_NO_ENDMARK)
   {
      info->dataCount = sig_info->sig_size;
      ret = ose_efmi_get_mem(pid, (OSADDRESS)sig_info->signal_pointer,
                             &info->data, (OSADDRESS)sig_info->sig_size);
      if (ret != EFMI_OK)
      {
         info->dataCount = 0;
      }
   }

   return ret;
}

static EfmiStatus
signal_queue_long(int32_t (*reply)(FmmSignal *sig),
                       union ContextInfo *context,
                       const union Signal *request)
{
   EfmiStatus ret;
   FmmSignal *sig;
   MonitorGetSignalQueueLongReply *info;
   uint32_t max_signals = fmd_properties.max_signals_in_queue;

   struct EfmiSignalCursor cursor;
   struct EfmiSignalInfo sig_info;

   ret = ose_efmi_get_first_q_signal(context->process_info.process, &cursor,
                                     &sig_info, sizeof(sig_info));
   while (ret == EFMI_OK)
   {
      sig = fmm_output_signal;
      info = &sig->data.get_signal_queue_long_reply;
      info->sigNo = MONITOR_GET_SIGNAL_QUEUE_LONG_REPLY;
      info->pid = context->process_info.process;
      info->signal.number = sig_info.sig_no;
      info->signal.owner = sig_info.owner;
      info->signal.sender = sig_info.sender_pid;
      info->signal.addressee = sig_info.addressee_pid;
      info->signal.size = sig_info.sig_size;
      info->signal.bufsize = sig_info.size_in_pool;
      info->signal.address = sig_info.signal_pointer;

      ret = signal_queue_set_status(&info->signal.status, &sig_info,
                                    &request->get_info_request.scope);
      if (ret != EFMI_OK)
      {
         break;
      }

      ret = signal_queue_long_set_data(info, context->process_info.process,
                                       &sig_info);
      if (ret != EFMI_OK)
      {
         break;
      }


      if (info->dataCount > 0)
      {
         sig->size = offsetof(MonitorGetSignalQueueLongReply, data) +
                     info->dataCount * sizeof(uint8_t);
      }
      else
      {
         sig->size = sizeof(*info);
      }

      ret = signal_queue_long_filter(info, request->get_info_request.tagsCount,
                                     request->get_info_request.tags);
      if (ret == EFMI_OK)
      {
         if ((ret = reply(sig)))
         {
            return ret;
         }
      }
      else
      {
         break;
      }
      /* dump signal info for max number of signals specified */
      if (!fmd_properties.max_signals_in_queue)
      {
         ret = ose_efmi_get_next_q_signal(&cursor, &sig_info);
      }
      else
      {
      	 max_signals--;
    	 if (max_signals != 0)
         { 
    	    ret = ose_efmi_get_next_q_signal(&cursor, &sig_info);
         }
    	 else
         { 
    	    return ret = EFMI_OK;
         }
      }

   }

   return ret == EFMI_END_OF_LIST ? EFMI_OK : ret;
}


EfmiStatus
fmm_efmi_get_signal_queue_request(int32_t (*reply)(FmmSignal *sig),
                                  int32_t (*endmark)(FmmSignal *sig, int32_t status),
                                  const MonitorGetInfoRequest *request)
{
   EfmiStatus ret;
   EfmiStatus (*info)(int32_t (*reply)(FmmSignal *sig),
                      union ContextInfo *context,
                      const union Signal *request);

   FmmSignal *sig;

   assert(request != NULL);
   assert(request->sigNo == MONITOR_GET_SIGNAL_QUEUE_REQUEST);

   info = request->level ? signal_queue_long : signal_queue;

   /* check whether user requested,
    * signal queue info for longest queue processes only */
   if (!fmd_properties.num_long_queue_procs)
   {
      switch (request->scope.type)
      {
         case MONITOR_SCOPE_SYSTEM:
	    ret = list_processes_in_system(info, reply,
			(const union Signal *)request);
	    break;
         case MONITOR_SCOPE_SEGMENT_ID:
	    ret = list_processes_in_segment(request->scope.id, info, reply,
					 (const union Signal *)request);
	    break;
	 case MONITOR_SCOPE_BLOCK_ID:
	    ret = list_processes_in_block(request->scope.id, info, reply,
					   (const union Signal *)request);
	    break;
	 case MONITOR_SCOPE_ID:
	    ret = list_process(request->scope.id, info, reply,
				(const union Signal *)request);
	    break;
	  /* TODO: case MONITOR_SCOPE_NAME_PATTERN: (Not used in PMD) */
	  /* TODO: case MONITOR_SCOPE_ENTRYPOINT: (Not used in PMD) */
	 default:
	    ret = EFMI_UNAVAILABLE;
	    break;
      }
   }
   else
   {
      uint32_t i;
      ret = EFMI_OK;
      /* store signal queue info for all the longest queue processes in the scope */
      for (i = 0; i < ((fmd_properties.num_long_queue_procs >
                    fmd_properties.process_count)? fmd_properties.process_count
		    :fmd_properties.num_long_queue_procs); i++)
      {

         ret = list_process((fmd_properties.proc_queue_list + i)->pid,
                            info, reply, (const union Signal *)request); 

         if (ret != EFMI_OK)
         { 
            return ret;
         }
      }
   }

   sig = fmm_output_signal;
   sig->size = sizeof(MonitorGetInfoEndmark);
   sig->data.get_info_endmark.sigNo = MONITOR_GET_SIGNAL_QUEUE_ENDMARK;

   switch (ret)
   {
      case EFMI_OK:
         sig->data.get_info_endmark.status = MONITOR_STATUS_OK;
         break;
      case EFMI_END_OF_LIST:
         sig->data.get_info_endmark.status = MONITOR_STATUS_CANCELLED;
         break;
      default:
         sig->data.get_info_endmark.status = MONITOR_STATUS_KERNEL_ERROR;
         break;
   }

   if (endmark(sig, ret))
   {
      return (ret) ? ret : EFMI_END_OF_LIST;
   }
   return ret;
}


static EfmiStatus
get_pool_sid(uint32_t pool_id, uint32_t *sid)
{
   EfmiStatus ret;

   struct EfmiSegmentInfo si;
   struct EfmiSegmentCursor sc;

   *sid = 0;

   ret = ose_efmi_get_first_segment(&sc, &si, sizeof(si));

   while (!ret)
   {
      struct EfmiBlockInfo bi;
      struct EfmiBlockCursor bc;

      ret = ose_efmi_get_first_block(si.segment, &bc, &bi, sizeof(bi));

      while (!ret)
      {
         if (bi.sig_pool == pool_id || bi.stack_pool == pool_id)
         {
            *sid = si.segment;
            return EFMI_OK;
         }

         ret = ose_efmi_get_next_block(&bc, &bi);
      }

      if (ret != EFMI_END_OF_LIST)
      {
         return ret;
      }

      ret = ose_efmi_get_next_segment(&sc, &si);
   }

   if (ret == EFMI_END_OF_LIST)
   {
      ret = EFMI_UNAVAILABLE;
   }

   return ret;
}

static EfmiStatus
list_pool(uint32_t id,
          EfmiStatus (*info)(int32_t (*reply)(FmmSignal *sig),
                             struct EfmiPoolInfo *pi,
                             uint32_t sid,
                             const union Signal *request),
          int32_t (*reply)(FmmSignal *sig),
          const union Signal *request)
{
   EfmiStatus ret;
   struct EfmiPoolInfo pi;
   uint32_t sid = 0;

   ret = get_pool_sid(id, &sid);
   if (ret != EFMI_OK)
   {
      return ret;
   }

   ret = ose_efmi_get_pool_info(id, &pi, sizeof(pi));
   if (ret != EFMI_OK)
   {
      return ret;
   }

   ret = info(reply, &pi, sid, request);

   return ret;
}

static EfmiStatus
list_pools_in_segment(uint32_t id,
                      EfmiStatus (*info)(int32_t (*reply)(FmmSignal *sig),
                                         struct EfmiPoolInfo *pi,
                                         uint32_t sid,
                                         const union Signal *request),
                      int32_t (*reply)(FmmSignal *sig),
                      const union Signal *request)
{
   EfmiStatus ret;
   uint32_t sid;

   struct EfmiPoolCursor cursor;
   struct EfmiPoolInfo pi;

   ret = ose_efmi_get_first_pool(&cursor, &pi, sizeof(pi));
   while (ret == EFMI_OK)
   {
      ret = get_pool_sid(pi.pool_id, &sid);
      if (ret != EFMI_OK)
      {
         return ret;
      }

      if (sid == id)
      {
         ret = list_pool(pi.pool_id, info, reply, request);
         if (ret != EFMI_OK)
         {
            return ret;
         }
      }

      ret = ose_efmi_get_next_pool(&cursor, &pi);
   }

   return ret == EFMI_END_OF_LIST ? EFMI_OK : ret;
}

static EfmiStatus
list_pools_in_system(EfmiStatus (*info)(int32_t (*reply)(FmmSignal *sig),
                                        struct EfmiPoolInfo *pi,
                                        uint32_t sid,
                                        const union Signal *request),
                     int32_t (*reply)(FmmSignal *sig),
                     const union Signal *request)
{
   EfmiStatus ret = EFMI_UNAVAILABLE;

   struct EfmiPoolCursor cursor;
   struct EfmiPoolInfo pi;

   ret = ose_efmi_get_first_pool(&cursor, &pi, sizeof(pi));
   while (ret == EFMI_OK)
   {
      ret = list_pool(pi.pool_id, info, reply, request);
      if (ret != EFMI_OK)
      {
         return ret;
      }

      ret = ose_efmi_get_next_pool(&cursor, &pi);
   }

   return ret == EFMI_END_OF_LIST ? EFMI_OK : ret;
}

static EfmiStatus
pool_info_set_fragments(MonitorGetPoolInfoReply *info,
                        struct EfmiPoolInfo *pi,
                        uint32_t *data_size)
{
   EfmiStatus ret;
   struct EfmiFragmentCursor cursor;
   struct EfmiFragmentInfo fi;

   MonitorPoolFragmentInfo *frags;

   info->fragmentsCount = 0;
   info->fragmentsOffset = *data_size;
   frags = (void *)&info->data[*data_size];

   ret = ose_efmi_get_first_fragment(pi->pool_id, &cursor, &fi, sizeof(fi));
   while (ret == EFMI_OK)
   {
      frags[info->fragmentsCount].address = fi.base_address;
      frags[info->fragmentsCount].size = fi.size;
      frags[info->fragmentsCount].stacks = fi.used_for_stacks;
      frags[info->fragmentsCount].signals = fi.used_for_signals;
      frags[info->fragmentsCount].reserved0 = 0;
      frags[info->fragmentsCount].reserved1 = 0;

      info->total += fi.size;
      info->free += fi.size - fi.used_for_stacks - fi.used_for_signals;

      info->fragmentsCount++;
      *data_size += sizeof(MonitorPoolFragmentInfo) / 4;

      ret = ose_efmi_get_next_fragment(&cursor, &fi);
   }

   if (ret == EFMI_OK)
   {
      if (info->fragmentsCount != pi->fragment_count)
      {
         return EFMI_END_OF_LIST;
      }
   }

   return ret == EFMI_END_OF_LIST ? EFMI_OK : ret;
}

static EfmiStatus
pool_info_set_stack_sizes(MonitorGetPoolInfoReply *info,
                          struct EfmiPoolInfo *pi,
                          uint32_t *data_size)
{
   uint32_t i;
   MonitorPoolBufferSizeInfo *stacks;
   info->stackSizesCount = 0;
   info->stackSizesOffset = *data_size;
   stacks = (void *)&info->data[*data_size];

   for(i = 0; i < EFMI_MAX_STACK_BUFFER_SIZE_COUNT; i++)
   {
      stacks[i].size = pi->stack_buf_sizes[i].configured_size;
      stacks[i].allocated = pi->stack_buf_sizes[i].allocated;
      stacks[i].free = pi->stack_buf_sizes[i].free;
      stacks[i].reserved0 = 0;
      info->stackSizesCount++;
      *data_size += sizeof(MonitorPoolBufferSizeInfo) / 4;
   }

   if (info->stackSizesCount != EFMI_MAX_STACK_BUFFER_SIZE_COUNT)
   {
      return EFMI_END_OF_LIST;
   }

   return EFMI_OK;
}

static EfmiStatus
pool_info_set_signal_sizes(MonitorGetPoolInfoReply *info,
                           struct EfmiPoolInfo *pi,
                           uint32_t *data_size)
{
   uint32_t i;
   MonitorPoolBufferSizeInfo *signals;
   info->signalSizesCount = 0;
   info->signalSizesOffset = *data_size;
   signals = (void *)&info->data[*data_size];

   for(i = 0; i < EFMI_MAX_SIGNAL_BUFFER_SIZE_COUNT; i++)
   {
      signals[i].size = pi->signal_buf_sizes[i].configured_size;
      signals[i].allocated = pi->signal_buf_sizes[i].allocated;
      signals[i].free = pi->signal_buf_sizes[i].free;
      signals[i].reserved0 = 0;
      info->signalSizesCount++;
      *data_size += sizeof(MonitorPoolBufferSizeInfo) / 4;
   }

   if (info->signalSizesCount != EFMI_MAX_SIGNAL_BUFFER_SIZE_COUNT)
   {
      return EFMI_END_OF_LIST;
   }

   return EFMI_OK;
}

static EfmiStatus
pool_info(int32_t (*reply)(FmmSignal *sig),
          struct EfmiPoolInfo *pi,
          uint32_t sid,
          const union Signal *request)
{
   EfmiStatus ret = EFMI_OK;
   FmmSignal *sig;
   MonitorGetPoolInfoReply *info;
   uint32_t data_size = 0;

   (void)request; /* Not used */
   
   sig = fmm_output_signal;

   info = &sig->data.get_pool_info_reply;
   info->sigNo = MONITOR_GET_POOL_INFO_REPLY;
   info->pid = pi->pool_id;
   info->sid = sid;

   /* FIXME: Stack alignment should not be hardcoded,
             but POOL_ALIGNMENT is currently not available from EFMI. */
#ifdef softOSE
   info->stackAlignment = 16;
#else
   info->stackAlignment = 32;
#endif
   info->stackAdmSize = 0;
   info->stackInternalAdmSize = 1;

   /* FIXME: Signal alignment should not be hardcoded,
             but POOL_ALIGNMENT is currently not available from EFMI. */
#ifdef softOSE
   info->signalAlignment = 16;
#else
   info->signalAlignment = 32;
#endif
   info->signalAdmSize = 1 + 32; /* TODO: Should be 1 + sizeof(struct _signal) from pose_internal.h */
   info->signalInternalAdmSize = 0;

   info->reserved0 = 0;
   info->reserved1 = 0;
   info->reserved2 = 0;


   info->total = 0; /* Sum of fragment sizes (?) */
   info->free = 0; /* Sum of fragment sizes - used in all fragments (?) */

   ret = pool_info_set_fragments(info, pi, &data_size);
   if (ret != EFMI_OK)
   {
      return ret;
   }

   ret = pool_info_set_stack_sizes(info, pi, &data_size);
   if (ret != EFMI_OK)
   {
      return ret;
   }

   ret = pool_info_set_signal_sizes(info, pi, &data_size);
   if (ret != EFMI_OK)
   {
      return ret;
   }

   sig->size = offsetof(MonitorGetPoolInfoReply, data) +
               info->fragmentsCount * sizeof(MonitorPoolFragmentInfo) +
               info->stackSizesCount * sizeof(MonitorPoolBufferSizeInfo) +
               info->signalSizesCount * sizeof(MonitorPoolBufferSizeInfo);

   if (ret == EFMI_OK)
   {
      if (reply(sig))
      {
         ret = EFMI_END_OF_LIST;
      }
   }

   return ret;
}


EfmiStatus
fmm_efmi_get_pool_info_request(int32_t (*reply)(FmmSignal *sig),
                               int32_t (*endmark)(FmmSignal *sig, int32_t status),
                               const MonitorGetInfoRequest *request)
{
   EfmiStatus ret;
   FmmSignal *sig;

   assert(request != NULL);
   assert(request->sigNo == MONITOR_GET_POOL_INFO_REQUEST);

   switch (request->scope.type)
   {
      case MONITOR_SCOPE_SYSTEM:
         ret = list_pools_in_system(pool_info, reply,
                                    (const union Signal *)request);
         break;
      case MONITOR_SCOPE_SEGMENT_ID:
         ret = list_pools_in_segment(request->scope.id, pool_info, reply,
                                     (const union Signal *)request);
         break;
      case MONITOR_SCOPE_BLOCK_ID:
         ret = EFMI_UNAVAILABLE;
         break;
      case MONITOR_SCOPE_ID:
         ret = list_pool(request->scope.id, pool_info, reply,
                         (const union Signal *)request);
         break;
      /* TODO: case MONITOR_SCOPE_NAME_PATTERN: (Not used in PMD) */
      /* TODO: case MONITOR_SCOPE_ENTRYPOINT: (Not used in PMD) */
      default:
         ret = EFMI_UNAVAILABLE;
         break;
   }

   sig = fmm_output_signal;
   sig->size = sizeof(MonitorGetInfoEndmark);
   sig->data.get_info_endmark.sigNo = MONITOR_GET_POOL_INFO_ENDMARK;

   switch (ret)
   {
      case EFMI_OK:
         sig->data.get_info_endmark.status = MONITOR_STATUS_OK;
         break;
      case EFMI_END_OF_LIST:
         sig->data.get_info_endmark.status = MONITOR_STATUS_CANCELLED;
         break;
      default:
         sig->data.get_info_endmark.status = MONITOR_STATUS_KERNEL_ERROR;
         break;
   }

   if (endmark(sig, ret))
   {
      return (ret) ? ret : EFMI_END_OF_LIST;
   }

   return ret;
}


static EfmiStatus
pool_signals_filter(MonitorGetPoolSignalsReply *info, uint32_t index,
                    uint32_t sid, uint32_t tagscount,
                    const MonitorTag *tags)
{
   uint32_t i;
   uint32_t not = 0;

   for (i = 0; i < tagscount; i += tags[i].tag.count)
   {
      switch (tags[i].tag.type)
      {
         case MONITOR_SIGNAL_FILTER_TAG_NUMBER:
            if (not ^ (info->signals[index].number != tags[i+1].num))
            {
               return EFMI_END_OF_LIST;
            }
            break;
         case MONITOR_SIGNAL_FILTER_TAG_OWNER_ID:
            if (not ^ (info->signals[index].owner != tags[i+1].num))
            {
               return EFMI_END_OF_LIST;
            }
            break;
         case MONITOR_SIGNAL_FILTER_TAG_OWNER_BID:
         {
            EfmiStatus ret;
            struct EfmiProcessInfo pi;

            ret = ose_efmi_get_process_info(info->signals[index].owner,
                                            &pi, sizeof(pi));
            if (ret != EFMI_OK)
            {
               return ret;
            }
            if (not ^ (pi.block != tags[i+1].num))
            {
               return EFMI_END_OF_LIST;
            }
            break;
         }
         case MONITOR_SIGNAL_FILTER_TAG_OWNER_SID:
            if (not ^ (sid != tags[i+1].num))
            {
               return EFMI_END_OF_LIST;
            }
            break;
         case MONITOR_SIGNAL_FILTER_TAG_OWNER_NAME:
         {
            EfmiStatus ret;
            void *name = &info->signals[index + 1];

            /* NOTE: The space available for the name shrinks. */
            if ((ret = ose_efmi_get_process_name(info->signals[index].owner,
                              name,
                              FMM_OUTPUT_SIGNAL_SIZE -
                              sizeof(*info) -
                              (index - 1) * sizeof(MonitorSignalInfo))))
            {
               return ret;
            }
            if (not ^ (!match(tags[i+1].str, name)))
            {
               return EFMI_END_OF_LIST;
            }
            break;
         }
         case MONITOR_SIGNAL_FILTER_TAG_SENDER_ID:
            if (not ^ (info->signals[index].sender != tags[i+1].num))
            {
               return EFMI_END_OF_LIST;
            }
            break;
         case MONITOR_SIGNAL_FILTER_TAG_SENDER_NAME:
         {
            EfmiStatus ret;
            void *name = &info->signals[index + 1];

            /* NOTE: The space available for the name shrinks. */
            if ((ret = ose_efmi_get_process_name(info->signals[index].sender,
                              name,
                              FMM_OUTPUT_SIGNAL_SIZE -
                              sizeof(*info) -
                              (index - 1) * sizeof(MonitorSignalInfo))))
            {
               return ret;
            }
            if (not ^ (!match(tags[i+1].str, name)))
            {
               return EFMI_END_OF_LIST;
            }
            break;
         }
         case MONITOR_SIGNAL_FILTER_TAG_ADDRESSEE_ID:
            if (not ^ (info->signals[index].addressee != tags[i+1].num))
            {
               return EFMI_END_OF_LIST;
            }
            break;
         case MONITOR_SIGNAL_FILTER_TAG_ADDRESSEE_NAME:
         {
            EfmiStatus ret;
            void *name = &info->signals[index + 1];

            /* NOTE: The space available for the name shrinks. */
            if ((ret = ose_efmi_get_process_name(info->signals[index].addressee,
                              name,
                              FMM_OUTPUT_SIGNAL_SIZE -
                              sizeof(*info) -
                              (index - 1) * sizeof(MonitorSignalInfo))))
            {
               return ret;
            }
            if (not ^ (!match(tags[i+1].str, name)))
            {
               return EFMI_END_OF_LIST;
            }
            break;
         }
         case MONITOR_SIGNAL_FILTER_TAG_SIZE:
            if (not ^ ((info->signals[index].size < tags[i+1].num)
                       || (info->signals[index].size >= tags[i+2].num)))
            {
               return EFMI_END_OF_LIST;
            }
            break;
         case MONITOR_SIGNAL_FILTER_TAG_BUFSIZE:
            if (not ^ ((info->signals[index].bufsize < tags[i+1].num)
                       || (info->signals[index].bufsize >= tags[i+2].num)))
            {
               return EFMI_END_OF_LIST;
            }
            break;
         case MONITOR_SIGNAL_FILTER_TAG_ADDRESS:
            if (not ^ ((info->signals[index].address < tags[i+1].num)
                       || (info->signals[index].address >= tags[i+2].num)))
            {
               return EFMI_END_OF_LIST;
            }
            break;
         case MONITOR_SIGNAL_FILTER_TAG_STATUS:
            if (not ^ (info->signals[index].status != tags[i+1].num))
            {
               return EFMI_END_OF_LIST;
            }
            break;
         default:
            break;
      }
      if (MONITOR_STACK_FILTER_TAG_NOT == tags[i].tag.type)
         not = 1;
      else
         not = 0;
   }
   return EFMI_OK;
}

static EfmiStatus
pool_signals_set_status(MonitorSignalStatus *status,
                        struct EfmiSignalInfo *sig_info,
                        uint32_t id)
{
   if (!sig_info->owner && sig_info->addressee_pid == id)
   {
      *status = MONITOR_SIGNAL_STATUS_FREE;
   }
   else
   {
      switch (sig_info->sig_status)
      {
         case EFMI_SS_VALID:
            *status = MONITOR_SIGNAL_STATUS_VALID;
            break;
         case EFMI_SS_NO_ENDMARK:
            *status = MONITOR_SIGNAL_STATUS_ENDMARK;
            break;
         case EFMI_SS_BROKEN_SIGADM:
            *status = MONITOR_SIGNAL_STATUS_ADMINISTRATION;
            break;
         case EFMI_SS_WILD_POINTER_INSIDE_POOL:
            *status = MONITOR_SIGNAL_STATUS_NOT_A_SIGNAL;
            break;
         case EFMI_SS_WILD_POINTER_OUTSIDE_POOL:
            *status = MONITOR_SIGNAL_STATUS_NOT_A_POOL;
            break;
      }
   }

   return EFMI_OK;
}

static EfmiStatus
pool_signals(int32_t (*reply)(FmmSignal *sig),
             struct EfmiPoolInfo *pi,
             uint32_t sid,
             const union Signal *request)
{
   EfmiStatus ret;
   static const uint32_t sigs_per_reply = 10; /* FIXME: Set to reasonable number */
   uint32_t i;

   FmmSignal *sig;
   MonitorGetPoolSignalsReply *info;

   struct EfmiSignalCursor cursor;
   struct EfmiSignalInfo sig_info;

   sig = fmm_output_signal;
   info = &sig->data.get_pool_signals_reply;
   info->sigNo = MONITOR_GET_POOL_SIGNALS_REPLY;
   info->pid = pi->pool_id;
   info->reserved0 = 0;

   i = 0;
   ret = ose_efmi_get_first_pool_signal(pi->pool_id, &cursor, &sig_info,
                                        sizeof(sig_info));
   while (ret == EFMI_OK)
   {
      info->signals[i].number = sig_info.sig_no;
      info->signals[i].owner = sig_info.owner;
      info->signals[i].sender = sig_info.sender_pid;
      info->signals[i].addressee = sig_info.addressee_pid;
      info->signals[i].size = sig_info.sig_size;
      info->signals[i].bufsize = sig_info.size_in_pool;
      info->signals[i].address = sig_info.signal_pointer;
      if ((ret = pool_signals_set_status(&info->signals[i].status, &sig_info, pi->pool_id)))
      {
         return ret;
      }

      if (!(ret = pool_signals_filter(info, i, sid,
                                      request->get_info_request.tagsCount,
                                      request->get_info_request.tags)))
      {
         i++;
         if (i == sigs_per_reply)
         {
            info->signalsCount = i;
            sig->size = offsetof(MonitorGetPoolSignalsReply, signals) +
                        sizeof(MonitorSignalInfo) * i;
            i = 0;
            if (reply(sig))
            {
               return EFMI_END_OF_LIST;
            }
         }
      }
      else
      {
         break;
      }

      ret = ose_efmi_get_next_pool_signal(&cursor, &sig_info);
   }
   if (i > 0)
   {
      /* Send the "left overs". */
      info->signalsCount = i;
      sig->size = offsetof(MonitorGetPoolSignalsReply, signals) +
         sizeof(MonitorSignalInfo) * i;
      if (reply(sig))
      {
         return EFMI_END_OF_LIST;
      }
   }

   return ret == EFMI_END_OF_LIST ? EFMI_OK : ret;
}

EfmiStatus
fmm_efmi_get_pool_signals_request(int32_t (*reply)(FmmSignal *sig),
                                  int32_t (*endmark)(FmmSignal *sig, int32_t status),
                                  const MonitorGetInfoRequest *request)
{
   EfmiStatus ret;
   FmmSignal *sig;

   assert(request != NULL);
   assert(request->sigNo == MONITOR_GET_POOL_SIGNALS_REQUEST);

   switch (request->scope.type)
   {
      case MONITOR_SCOPE_SYSTEM:
         ret = list_pools_in_system(pool_signals, reply,
                                    (const union Signal *)request);
         break;
      case MONITOR_SCOPE_SEGMENT_ID:
         ret = list_pools_in_segment(request->scope.id, pool_signals, reply,
                                     (const union Signal *)request);
         break;
      case MONITOR_SCOPE_BLOCK_ID:
         ret = EFMI_UNAVAILABLE;
         break;
      case MONITOR_SCOPE_ID:
         ret = list_pool(request->scope.id, pool_signals, reply,
                         (const union Signal *)request);
         break;
      /* TODO: case MONITOR_SCOPE_NAME_PATTERN: (Not used in PMD) */
      /* TODO: case MONITOR_SCOPE_ENTRYPOINT: (Not used in PMD) */
      default:
         ret = EFMI_UNAVAILABLE;
         break;
   }

   sig = fmm_output_signal;
   sig->size = sizeof(MonitorGetInfoEndmark);
   sig->data.get_info_endmark.sigNo = MONITOR_GET_POOL_SIGNALS_ENDMARK;

   switch (ret)
   {
      case EFMI_OK:
         sig->data.get_info_endmark.status = MONITOR_STATUS_OK;
         break;
      case EFMI_END_OF_LIST:
         sig->data.get_info_endmark.status = MONITOR_STATUS_CANCELLED;
         break;
      default:
         sig->data.get_info_endmark.status = MONITOR_STATUS_KERNEL_ERROR;
         break;
   }

   if (endmark(sig, ret))
   {
      return (ret) ? ret : EFMI_END_OF_LIST;
   }
   return ret;
}

static EfmiStatus
registers_set_registers(MonitorRegisterValues *info,
                        uint32_t pid,
                        uint32_t reg_count,
                        const MonitorRegisterId *regs)
{
   uint32_t i;

   for (i = 0; i < reg_count; i++)
   {
      EfmiStatus err;
      struct EfmiRegisterInfo reg;

      err = ose_efmi_get_register(pid, MONITOR_REGISTER_ID(regs[i]), &reg);

      if (err != EFMI_OK)
      {
         info->registers[i].id = MONITOR_REGISTER_INVALID | regs[i];
         info->registers[i].value = 0xffffffff;
      }
      else
      {
         info->registers[i].id = regs[i];
         info->registers[i].value = reg.value;
      }
   }
   info->registersCount = reg_count;

   return EFMI_OK;
}


static EfmiStatus
registers(int32_t (*reply)(FmmSignal *sig),
          union ContextInfo *context,
          const union Signal *request)
{
   EfmiStatus ret = EFMI_OK;

   FmmSignal *sig;
   MonitorRegisterValues *info;

   struct EfmiProcessInfo pi;

   ret = ose_efmi_get_process_info(context->process_info.process, &pi,
                                   sizeof(pi));
   if (ret != EFMI_OK)
   {
      return ret;
   }

   if (pi.type == OS_BG_PROC || pi.type == OS_PRI_PROC
       || pi.type == OS_INT_PROC || pi.type == OS_TI_PROC)
   {
      sig = fmm_output_signal;

      info = &sig->data.get_registers_reply;
      info->sigNo = MONITOR_GET_REGISTERS_REPLY;
      info->pid = context->process_info.process;
      info->status = MONITOR_STATUS_OK;

      ret = registers_set_registers(info,
                                    context->process_info.process,
                                    request->get_register_info_request.registersCount,
                                    request->get_register_info_request.registers);
      if (ret != EFMI_OK)
      {
         return ret;
      }

      sig->size = offsetof(MonitorRegisterValues, registers) +
                  info->registersCount * sizeof(MonitorRegister);

      if (reply(sig))
      {
         ret = EFMI_END_OF_LIST;
      }
   }

   return ret;
}


/*
 * Requests from monitor_execution.sig
 */

EfmiStatus
fmm_efmi_get_register_info_request(int32_t (*reply)(FmmSignal *sig),
                                   int32_t (*endmark)(FmmSignal *sig, int32_t status),
                                   const MonitorGetRegisterInfoRequest *request)
{
   EfmiStatus ret;
   FmmSignal *sig;

   assert(request != NULL);
   assert(request->sigNo == MONITOR_GET_REGISTER_INFO_REQUEST);


   switch (request->scope.type)
   {
      case MONITOR_SCOPE_SYSTEM:
         ret = list_processes_in_system(registers, reply,
                                        (const union Signal *)request);
         break;
      case MONITOR_SCOPE_SEGMENT_ID:
         ret = list_processes_in_segment(request->scope.id, registers, reply,
                                         (const union Signal *)request);
         break;
      case MONITOR_SCOPE_BLOCK_ID:
         ret = list_processes_in_block(request->scope.id, registers, reply,
                                       (const union Signal *)request);
         break;
      case MONITOR_SCOPE_ID:
         ret = list_process(request->scope.id, registers, reply,
                            (const union Signal *)request);
         break;
      /* TODO: case MONITOR_SCOPE_NAME_PATTERN: (Not used in PMD) */
      /* TODO: case MONITOR_SCOPE_ENTRYPOINT: (Not used in PMD) */
      default:
         ret = EFMI_UNAVAILABLE;
         break;
   }

   sig = fmm_output_signal;
   sig->size = sizeof(MonitorGetRegisterInfoEndmark);
   sig->data.get_register_info_endmark.sigNo = MONITOR_GET_REGISTER_INFO_ENDMARK;

   switch (ret)
   {
      case EFMI_OK:
         sig->data.get_register_info_endmark.status = MONITOR_STATUS_OK;
         break;
      case EFMI_END_OF_LIST:
         sig->data.get_register_info_endmark.status = MONITOR_STATUS_CANCELLED;
         break;
      default:
         sig->data.get_register_info_endmark.status = MONITOR_STATUS_KERNEL_ERROR;
         break;
   }

   if (endmark(sig, ret))
   {
      return (ret) ? ret : EFMI_END_OF_LIST;
   }
   return ret;
}

void fmm_efmi_set_fmd_properties(FmmFmdProperties *fmd_set_properties)
{
   fmd_properties.num_long_queue_procs = fmd_set_properties->num_long_queue_procs;
   fmd_properties.max_signals_in_queue = fmd_set_properties->max_signals_in_queue;
   fmd_properties.max_stack_top_size = fmd_set_properties->max_stack_top_size;
   fmd_properties.process_count = fmd_set_properties->process_count;
   fmd_properties.proc_queue_list = fmd_set_properties->proc_queue_list;
}
