/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2006 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

#include "stdint.h"
#include "string.h"

#include "ramtrace.h"
#include "rmm_ramtrace.h"

#include "efmi.h"
#include "monitor.sig"
#include "monitor_events.sig"

#include "fmm.h"
#include "fmm_ramtrace.h"

#define assert(x) if (!(x)) return EFMI_UNAVAILABLE;

union Signal
{
   uint32_t signo;
   MonitorGetTraceRequest              get_trace_request;   
   MonitorProcessInfo                  process_info;
   MonitorSendReceiveInfo              send_receive_info;
   MonitorSwapCreateKillInfo           swap_create_kill_info;
   MonitorErrorInfo                    error_info;
   MonitorResetInfo                    reset_info;
   MonitorUserInfo                     user_info;
   MonitorBindInfo                     bind_info;
   MonitorAllocFreeBufInfo             alloc_free_buf_info;
   MonitorTimeoutInfo                  timeout_info;
   MonitorGetTraceEndmark              get_trace_endmark;
};

struct FmmSignal
{
   uint32_t size;
   union Signal data;
};

#define MAX_IDBUF_SIZE 32

struct idbuf
{
   int        i0;
   int        i1;
   uint32_t   pid[MAX_IDBUF_SIZE];
};



static void
idbuf_init(struct idbuf *buf)
{
   buf->i0 = 0;
   buf->i1 = 0;
}

static int
idbuf_contains(struct idbuf *buf, uint32_t pid)
{
   int        i;

   /* Search buffer for 'pid' */
   for (i = buf->i0; i != buf->i1; i = (i + 1) % MAX_IDBUF_SIZE)
   {
      if (buf->pid[i] == pid)
      {
         return 1;
      }
   }

   buf->pid[buf->i1] = pid;
   buf->i1 = (buf->i1 + 1) % MAX_IDBUF_SIZE;

   if (buf->i1 == buf->i0)
   {
      buf->i0 = (buf->i0 + 1) % MAX_IDBUF_SIZE;
   }

   /* Return not found */
   return 0;
}

static uint32_t 
get_trace_len(void)
{
   struct EfmiConfigurationInfo ci;
   
   if (ose_efmi_get_config(EFMI_CONF_TRACESIZE, &ci))
   {
      return 0;
   }
   /* TODO: This should be validated when read, not here, in pose, in rmm... */
   if (ci.value < 8192)
   {
      ci.value = 8192;
   }
   return ci.value;
}

static void *
get_trace(uint32_t len)
{
   EfmiStatus ret;
   
   struct EfmiFragmentInfo fi;
   struct EfmiFragmentCursor fc;
   
   /* TODO: Do not assume 0x10000 is system pool */
   ret = ose_efmi_get_first_fragment(0x10000, &fc, &fi, sizeof(fi));
   
   while (!ret)
   {
      ret = ose_efmi_get_next_fragment(&fc, &fi);

      /* TODO: Do not assume trace located before first fragement */
      /* TODO: Do not assume first fragment is last fragment returned */
      if (ret == EFMI_END_OF_LIST)
      {
         void *trace;
         
         if ((trace = ose_trace_find((char *)fi.base_address - len, len, SYS_TRACE_NAME)))
         {
            return trace;
         }
      }
   }
      
   return NULL;
}

static int
is_public_event(uint16_t no)
{
   switch (no)
   {
      case TRACE_SEND:
      case TRACE_RECEIVE:
      case TRACE_USER_ERROR:
      case TRACE_OSERROR:
      case TRACE_CREATE:
      case TRACE_KILL:
      case TRACE_SWAP:
      case TRACE_USER_EVENT:
      case TRACE_ALLOC:
      case TRACE_FREE_BUF:
      case TRACE_BIND:
      case TRACE_SYSTEM_START:
      case TRACE_RECEIVE_TMO:
         return 1;                  
      default:
         return 0;                 
   }
}

static void
process_info(FmmSignal *sig, struct trace_procname *tp)
{
   static const uint32_t type[] = 
   { 
      MONITOR_PROCESS_TYPE_PRIORITIZED, 
      MONITOR_PROCESS_TYPE_BACKGROUND,
      MONITOR_PROCESS_TYPE_INTERRUPT, 
      MONITOR_PROCESS_TYPE_TIMER_INTERRUPT, 
      MONITOR_PROCESS_TYPE_PHANTOM, 
      MONITOR_PROCESS_TYPE_UNKNOWN, 
      MONITOR_PROCESS_TYPE_UNKNOWN, 
      MONITOR_PROCESS_TYPE_UNKNOWN 
   };
   
   sig->data.process_info.sigNo = MONITOR_GET_PROCESS_TRACE_REPLY;
   sig->data.process_info.pid = tp->pid;
   sig->data.process_info.bid = 0;
   sig->data.process_info.sid = 0;
   /* NOTE: Relies on internal POSE representation (see pose_internal.h) */
   sig->data.process_info.type = type[tp->priority >> 5];
   sig->data.process_info.entrypoint = 0;
   sig->data.process_info.properties = 0;
   sig->data.process_info.reserved0 = 0;
   
   strcpy((char *)&sig->data.process_info.name, tp->name);
   
   sig->size = sizeof(MonitorProcessInfo) + tp->namelen;
}

static size_t 
fmm_strnlen(const char *s, size_t n)
{
   size_t i;
   
   for (i = 0; i < n; i++)
   {
      if (!s[i])
      {
         return i;
      }
   }
   return n;
}

static int 
is_send_receive_ok(struct trace_send_receive *tsr, 
                   char *trace, uint32_t trace_len)
{ 
   return tsr->sigsize >= sizeof(uint32_t) 
      && (char *)tsr->to >= trace && (char *)tsr->to < trace + trace_len 
      && (char *)tsr->from >= trace && (char *)tsr->from < trace + trace_len 
      && (!tsr->abstime || ((char *)tsr->abstime >= trace && (char *)tsr->abstime < trace + trace_len)) 
      && fmm_strnlen(tsr->to->name, tsr->to->namelen + 1) == tsr->to->namelen 
      && fmm_strnlen(tsr->from->name, tsr->from->namelen + 1) == tsr->from->namelen
      && (char *)tsr->data + tsr->sigsize <= trace + trace_len;
}

static void
send_receive_info(FmmSignal *sig, TraceEntryState *tes, uint32_t i)
{
   struct trace_send_receive *tsr = tes->trace_entry;
   
   sig->data.signo = tsr->event_no == TRACE_SEND 
                        ? MONITOR_GET_SEND_TRACE_REPLY 
                        : MONITOR_GET_RECEIVE_TRACE_REPLY;
   sig->data.send_receive_info.reference = i;
   sig->data.send_receive_info.time.tick = tes->tick;
   sig->data.send_receive_info.time.ntick = tes->time_count;
   if (tsr->abstime)
   {
      sig->data.send_receive_info.time.sec = tsr->abstime->abs_sec;
      sig->data.send_receive_info.time.sectick = tsr->abstime->abs_tick;
   }
   else
   {
      sig->data.send_receive_info.time.sec = 0;
      sig->data.send_receive_info.time.sectick = 0;
   }
   sig->data.send_receive_info.signalNumber = *(uint32_t *)tsr->data;
   sig->data.send_receive_info.signalSender = tsr->from->pid;
   sig->data.send_receive_info.signalAddressee = tsr->to->pid;
   sig->data.send_receive_info.signalSize = 0;
   sig->data.send_receive_info.signalAddress = 0;
   sig->data.send_receive_info.signalId = 0;
   sig->data.send_receive_info.lineNumber = 0;
   sig->data.send_receive_info.fileNameOffset = 0;
   sig->data.send_receive_info.signalDataOffset = 0;
   sig->data.send_receive_info.signalDataCount = tsr->sigsize;
   memcpy(sig->data.send_receive_info.data, tsr->data, tsr->sigsize);
   sig->data.send_receive_info.euId = tsr->cpu; 
   sig->data.send_receive_info.reserved0 = 0;
   sig->data.send_receive_info.reserved1 = 0;

   sig->size = offsetof(MonitorSendReceiveInfo, data) + tsr->sigsize;
}

static int
is_ose_error_ok(struct trace_ose_error *toe,
                char *trace, uint32_t trace_len)
{
   return (char *)toe->process >= trace && (char *)toe->process < trace + trace_len
      && (!toe->abstime || ((char *)toe->abstime >= trace && (char *)toe->abstime < trace + trace_len))
      && fmm_strnlen(toe->process->name, toe->process->namelen + 1) == toe->process->namelen; 
}                

static void
error_info(FmmSignal *sig, TraceEntryState *tes, uint32_t i)
{
   struct trace_ose_error *toe = tes->trace_entry;
   
   sig->data.signo = MONITOR_GET_ERROR_TRACE_REPLY;
   sig->data.error_info.reference = i;
   sig->data.error_info.time.tick = tes->tick;
   sig->data.error_info.time.ntick = tes->time_count;
   if (toe->abstime)
   {
      sig->data.error_info.time.sec = toe->abstime->abs_sec;
      sig->data.error_info.time.sectick = toe->abstime->abs_tick;
   }
   else
   {
      sig->data.error_info.time.sec = 0;
      sig->data.error_info.time.sectick = 0;     
   }
   sig->data.error_info.from = toe->process->pid;
   sig->data.error_info.exec = toe->event_no == TRACE_OSERROR;
   sig->data.error_info.error = toe->ecode;
   sig->data.error_info.extra = toe->extra;
   sig->data.error_info.lineNumber = 0;
   sig->data.error_info.fileNameOffset = 0;
   sig->data.error_info.euId = toe->cpu;
   sig->data.error_info.reserved0 = 0;
   sig->data.error_info.reserved1 = 0;
   
   sig->size = offsetof(MonitorErrorInfo, data);   
}

static int
is_create_kill_ok(struct trace_create_kill *tck,
                  char *trace, uint32_t trace_len)
{
   return (char *)tck->process >= trace && (char *)tck->process < trace + trace_len
      && (!tck->abstime || ((char *)tck->abstime >= trace && (char *)tck->abstime < trace + trace_len))   
      && fmm_strnlen(tck->process->name, tck->process->namelen + 1) == tck->process->namelen; 
}

static void
create_kill_info(FmmSignal *sig, TraceEntryState *tes, uint32_t i)
{
   struct trace_create_kill *tck = tes->trace_entry;
   
   sig->data.signo = tck->event_no == TRACE_CREATE
                        ? MONITOR_GET_CREATE_TRACE_REPLY
                        : MONITOR_GET_KILL_TRACE_REPLY;
   sig->data.swap_create_kill_info.reference = i;
   sig->data.swap_create_kill_info.time.tick = tes->tick;
   sig->data.error_info.time.ntick = tes->time_count;
   if (tck->abstime)
   {
      sig->data.send_receive_info.time.sec = tck->abstime->abs_sec;
      sig->data.send_receive_info.time.sectick = tck->abstime->abs_tick;
   }
   else
   {
      sig->data.send_receive_info.time.sec = 0;
      sig->data.send_receive_info.time.sectick = 0;      
   }
   sig->data.swap_create_kill_info.from = 0;
   sig->data.swap_create_kill_info.to = tck->process->pid;
   sig->data.swap_create_kill_info.lineNumber = 0;
   sig->data.swap_create_kill_info.fileNameOffset = 0;
   sig->data.swap_create_kill_info.euId = tck->cpu;
   sig->data.swap_create_kill_info.reserved0 = 0;
   sig->data.swap_create_kill_info.reserved1 = 0;   

   sig->size = offsetof(MonitorSwapCreateKillInfo, data);
}

static int
is_swap_ok(struct trace_swap *ts,
           char *trace, uint32_t trace_len)
{
   return (char *)ts->to_process >= trace && (char *)ts->to_process < trace + trace_len 
      && (char *)ts->from_process >= trace && (char *)ts->from_process < trace + trace_len 
      && (!ts->abstime || ((char *)ts->abstime >= trace && (char *)ts->abstime < trace + trace_len)) 
      && fmm_strnlen(ts->to_process->name, ts->to_process->namelen + 1) == ts->to_process->namelen 
      && fmm_strnlen(ts->from_process->name, ts->from_process->namelen + 1) == ts->from_process->namelen;     
}

static void
swap_info(FmmSignal *sig, TraceEntryState *tes, uint32_t i)
{
   struct trace_swap *ts = tes->trace_entry;
   
   sig->data.signo = MONITOR_GET_SWAP_TRACE_REPLY;
   sig->data.swap_create_kill_info.reference = i;
   sig->data.swap_create_kill_info.time.tick = tes->tick;
   sig->data.swap_create_kill_info.time.ntick = tes->time_count;
   if (ts->abstime)
   {
      sig->data.swap_create_kill_info.time.sec = ts->abstime->abs_sec;
      sig->data.swap_create_kill_info.time.sectick = ts->abstime->abs_tick;
   }
   else
   {
      sig->data.swap_create_kill_info.time.sec = 0;
      sig->data.swap_create_kill_info.time.sectick = 0;
   }
   sig->data.swap_create_kill_info.from = ts->from_process->pid;
   sig->data.swap_create_kill_info.to = ts->to_process->pid;
   sig->data.swap_create_kill_info.lineNumber = 0;
   sig->data.swap_create_kill_info.fileNameOffset = 0;
   sig->data.swap_create_kill_info.euId = ts->cpu; 
   sig->data.swap_create_kill_info.reserved0 = 0;
   sig->data.swap_create_kill_info.reserved1 = 0;

   sig->size = offsetof(MonitorSwapCreateKillInfo, data);
}

static int
is_user_event_ok(struct trace_user_event *tue,
                 char *trace, uint32_t trace_len)
{
   return (char *)tue->process >= trace && (char *)tue->process < trace + trace_len
      && (!tue->abstime || ((char *)tue->abstime >= trace && (char *)tue->abstime < trace + trace_len))   
      && fmm_strnlen(tue->process->name, tue->process->namelen + 1) == tue->process->namelen 
      && (char *)tue->data + tue->event_size < trace + trace_len;
}

static void
user_info(FmmSignal *sig, TraceEntryState *tes, uint32_t i)
{
   struct trace_user_event *tue = tes->trace_entry;
   
   sig->data.signo = MONITOR_GET_USER_TRACE_REPLY;
   sig->data.user_info.reference = i;
   sig->data.user_info.time.tick = tes->tick;
   sig->data.user_info.time.ntick = tes->time_count;
   if (tue->abstime)
   {
      sig->data.user_info.time.sec = tue->abstime->abs_sec;
      sig->data.user_info.time.sectick = tue->abstime->abs_tick;
   }
   else
   {
      sig->data.user_info.time.sec = 0;
      sig->data.user_info.time.sectick = 0;
   }
   sig->data.user_info.from = tue->process->pid;
   sig->data.user_info.eventNumber = tue->event_id;
   sig->data.user_info.eventSize = tue->requested_size + sizeof(tue->event_id);
   sig->data.user_info.lineNumber = 0;
   sig->data.user_info.fileNameOffset = 0;
   sig->data.user_info.eventDataOffset = 0;
   sig->data.user_info.eventDataCount = tue->event_size + sizeof(tue->event_id);
   sig->data.user_info.euId = tue->cpu;
   sig->data.user_info.reserved0 = 0;
   memcpy(sig->data.user_info.data, &tue->event_id, sizeof(tue->event_id));
   memcpy(sig->data.user_info.data + sizeof(tue->event_id), tue->data, tue->event_size);

   sig->size = offsetof(MonitorUserInfo, data) + sizeof(tue->event_id) + tue->event_size;
}

static int
is_alloc_free_buf_ok(struct trace_alloc_free_buf *tafb,
                     char *trace, uint32_t trace_len)
{
   return tafb->sigsize >= sizeof(uint32_t)
      && (char *)tafb->process >= trace && (char *)tafb->process < trace + trace_len
      && (!tafb->abstime || ((char *)tafb->abstime >= trace && (char *)tafb->abstime < trace + trace_len))   
      && fmm_strnlen(tafb->process->name, tafb->process->namelen + 1) == tafb->process->namelen;
}

static void
alloc_free_buf_info(FmmSignal *sig, TraceEntryState *tes, uint32_t i)
{
   struct trace_alloc_free_buf *tafb = tes->trace_entry;
   
   sig->data.signo = tafb->event_no == TRACE_ALLOC 
                        ? MONITOR_GET_ALLOC_TRACE_REPLY
                        : MONITOR_GET_FREE_BUF_TRACE_REPLY;
   sig->data.alloc_free_buf_info.reference = i;
   sig->data.alloc_free_buf_info.time.tick = tes->tick;
   sig->data.alloc_free_buf_info.time.ntick = tes->time_count;
   if (tafb->abstime)
   {
      sig->data.alloc_free_buf_info.time.sec = tafb->abstime->abs_sec;
      sig->data.alloc_free_buf_info.time.sectick = tafb->abstime->abs_tick;
   }
   else
   {
      sig->data.alloc_free_buf_info.time.sec = 0;
      sig->data.alloc_free_buf_info.time.sectick = 0;
   }
   sig->data.alloc_free_buf_info.from = tafb->process->pid;
   sig->data.alloc_free_buf_info.signalNumber = *(uint32_t *)tafb->data; 
   sig->data.alloc_free_buf_info.signalSize = tafb->requested_size;
   sig->data.alloc_free_buf_info.signalAddress = tafb->address;
   sig->data.alloc_free_buf_info.lineNumber = 0;
   sig->data.alloc_free_buf_info.fileNameOffset = 0;
   sig->data.alloc_free_buf_info.euId = tafb->cpu;
   sig->data.alloc_free_buf_info.reserved0 = 0;
   
   sig->size = offsetof(MonitorAllocFreeBufInfo, data);
}

static int
is_bind_ok(struct trace_bind *tb, 
           char *trace, uint32_t trace_len)
{
   return (char *)tb->process >= trace && (char *)tb->process < trace + trace_len
      && (!tb->abstime || ((char *)tb->abstime >= trace && (char *)tb->abstime < trace + trace_len))   
      && fmm_strnlen(tb->process->name, tb->process->namelen + 1) == tb->process->namelen;    
}

static void
bind_info(FmmSignal *sig, TraceEntryState *tes, uint32_t i)
{
   struct trace_bind *tb = tes->trace_entry;
   
   sig->data.signo = MONITOR_GET_BIND_TRACE_REPLY;
   sig->data.bind_info.reference = i;
   sig->data.bind_info.time.tick = tes->tick;
   sig->data.bind_info.time.ntick = tes->time_count;
   if (tb->abstime)
   {
      sig->data.bind_info.time.sec = tb->abstime->abs_sec;
      sig->data.bind_info.time.sectick = tb->abstime->abs_tick;
   }
   else
   {
      sig->data.bind_info.time.sec = 0;
      sig->data.bind_info.time.sectick = 0;        
   }
   sig->data.bind_info.reserved0 = 0;
   sig->data.bind_info.bound = tb->process->pid;
   sig->data.bind_info.fromEuId = tb->from_eu;
   sig->data.bind_info.toEuId = tb->to_eu;
   sig->data.bind_info.lineNumber = 0;
   sig->data.bind_info.fileNameOffset = 0;
   sig->data.bind_info.reserved1 = 0;
   
   sig->size = offsetof(MonitorBindInfo, data);
}

static int
is_receive_tmo_ok(struct trace_receive_tmo *trt,
                     char *trace, uint32_t trace_len)
{
   return (char *)trt->process >= trace && (char *)trt->process < trace + trace_len
      && (!trt->abstime || ((char *)trt->abstime >= trace && (char *)trt->abstime < trace + trace_len))   
      && fmm_strnlen(trt->process->name, trt->process->namelen + 1) == trt->process->namelen;
}

static void
timeout_info(FmmSignal *sig, TraceEntryState *tes, uint32_t i)
{
   struct trace_receive_tmo *trt = tes->trace_entry;
   
   sig->data.signo = MONITOR_GET_TIMEOUT_TRACE_REPLY;
   sig->data.timeout_info.reference = i;
   sig->data.timeout_info.time.tick = tes->tick;
   sig->data.timeout_info.time.ntick = tes->time_count;
   if (trt->abstime)
   {
      sig->data.timeout_info.time.sec = trt->abstime->abs_sec;
      sig->data.timeout_info.time.sectick = trt->abstime->abs_tick;
   }
   else
   {
      sig->data.timeout_info.time.sec = 0;
      sig->data.timeout_info.time.sectick = 0;
   }
   sig->data.timeout_info.signalAddressee = trt->process->pid;
   sig->data.timeout_info.timeout = trt->timeout;
   sig->data.timeout_info.tmoSource = 0;
   sig->data.alloc_free_buf_info.lineNumber = 0;
   sig->data.alloc_free_buf_info.fileNameOffset = 0;
   sig->data.alloc_free_buf_info.euId = trt->cpu;
   sig->data.alloc_free_buf_info.reserved0 = 0;
   
   sig->size = offsetof(MonitorTimeoutInfo, data);
}

static int
is_system_start_ok(struct trace_system_start *tss,
                   char *trace, uint32_t trace_len)
{
   return (!tss->abstime || ((char *)tss->abstime >= trace && (char *)tss->abstime < trace + trace_len));
}

static void
reset_info(FmmSignal *sig, TraceEntryState *tes, uint32_t i)
{
   struct trace_system_start *tss = tes->trace_entry;
   
   sig->data.signo = MONITOR_GET_RESET_TRACE_REPLY;
   sig->data.reset_info.reference = i;
   sig->data.reset_info.time.tick = tes->tick;
   sig->data.reset_info.time.ntick = tes->time_count;
   if (tss->abstime)
   {
      sig->data.reset_info.time.sec = tss->abstime->abs_sec;
      sig->data.reset_info.time.sectick = tss->abstime->abs_tick;
   }
   else
   {
      sig->data.reset_info.time.sec = 0;
      sig->data.reset_info.time.sectick = 0;
   }
   sig->data.reset_info.warm = tss->warm_flag;
   sig->data.reset_info.lineNumber = 0;
   sig->data.reset_info.fileNameOffset = 0;
   sig->data.reset_info.reserved0 = 0;
   sig->data.reset_info.reserved1 = 0;
   sig->data.reset_info.reserved2 = 0;
   
   sig->size = offsetof(MonitorResetInfo, data);
}

static EfmiStatus
list_ramtrace(int32_t (*reply)(FmmSignal *sig), 
              const MonitorGetTraceRequest *request)
{
   EfmiStatus ret; 
   FmmSignal *sig;
      
   char *trace;
   uint32_t trace_len;
   
   ret = EFMI_OK;
   
   sig = fmm_output_signal;
   
   if ((trace_len = get_trace_len()) > 0 
       && (trace = get_trace(trace_len)) != NULL)
   {      
      TraceEntryState tes0, tes1;
      
      if (!ose_trace_open(trace, &tes0))
      {
         uint32_t i0, i1, in;
         static struct idbuf pids;
                           
         in = 0;
         tes1 = tes0;         
         idbuf_init(&pids);
         
         /* Find the total number of events */
         while (ose_trace_read(&tes0))
         {      
            union TRACE *te = tes0.trace_entry;
            
            if (is_public_event(te->event_no))
            {
               in++;
            }
         }
         
         i1 = (in > request->to) ? request->to : in;
         i0 = (in > request->from) ? request->from : (in > 48) ? in - 48 : 0;
         in = 0; 
         
         /* Skip past to the first event to return */
         while (in < i0)
         {
            if (ose_trace_read(&tes1))
            {
               union TRACE *te = tes1.trace_entry;

               if (is_public_event(te->event_no))
               {
                  in++;
               }   
            }            
         }

         /* Now, return the events requested */         
         while (in < i1)
         {
            if (ose_trace_read(&tes1))
            {
               union TRACE *te = tes1.trace_entry;

               switch (te->event_no)
               {
                  case TRACE_SEND:
                  case TRACE_RECEIVE:
                  {
                     struct trace_send_receive *tsr = &te->trace_send_receive;
                 
                     if (!is_send_receive_ok(tsr, trace, trace_len))
                     {
                        ret = EFMI_KERNEL_CORRUPTED;
                        break;
                     }                                          
                     if (!idbuf_contains(&pids, tsr->from->pid))
                     {                           
                        process_info(sig, tsr->from);

                        if (reply(sig))
                        {   
                           ret = EFMI_END_OF_LIST;
                           break;
                        }
                     }
                     if (!idbuf_contains(&pids, tsr->to->pid))
                     {                           
                        process_info(sig, tsr->to);

                        if (reply(sig))
                        {   
                           ret = EFMI_END_OF_LIST;
                           break;
                        }
                     }
                     send_receive_info(sig, &tes1, in);

                     if (reply(sig))
                     {   
                        ret = EFMI_END_OF_LIST;
                        break;
                     }
                     break;
                  }
                  case TRACE_USER_ERROR:
                  case TRACE_OSERROR:
                  {
                     struct trace_ose_error *toe = &te->trace_ose_error;
                     
                     if (!is_ose_error_ok(toe, trace, trace_len))
                     {
                        ret = EFMI_KERNEL_CORRUPTED;
                        break;
                     }                                          
                     if (!idbuf_contains(&pids, toe->process->pid))
                     {                           
                        process_info(sig, toe->process);
                        
                        if (reply(sig))
                        {   
                           ret = EFMI_END_OF_LIST;
                           break;
                        }
                     }
                     error_info(sig, &tes1, in);

                     if (reply(sig))
                     {   
                        ret = EFMI_END_OF_LIST;
                        break;
                     }
                     break;                     
                  }
                  case TRACE_CREATE:
                  case TRACE_KILL:
                  {
                     struct trace_create_kill *tck = &te->trace_create_kill;
                     
                     if (!is_create_kill_ok(tck, trace, trace_len))
                     {
                        ret = EFMI_KERNEL_CORRUPTED;
                        break;
                     }
                     if (!idbuf_contains(&pids, tck->process->pid))
                     {                           
                        process_info(sig, tck->process);

                        if (reply(sig))
                        {   
                           ret = EFMI_END_OF_LIST;
                           break;
                        }
                     }
                     create_kill_info(sig, &tes1, in);

                     if (reply(sig))
                     {   
                        ret = EFMI_END_OF_LIST;
                        break;
                     }
                     break;                                          
                  }
                  case TRACE_SWAP:
                  {
                     struct trace_swap *ts = &te->trace_swap;
                 
                     if (!is_swap_ok(ts, trace, trace_len))
                     {
                        ret = EFMI_KERNEL_CORRUPTED;
                        break;
                     }                                          
                     if (!idbuf_contains(&pids, ts->from_process->pid))
                     {                           
                        process_info(sig, ts->from_process);

                        if (reply(sig))
                        {   
                           ret = EFMI_END_OF_LIST;
                           break;
                        }
                     }
                     if (!idbuf_contains(&pids, ts->to_process->pid))
                     {                           
                        process_info(sig, ts->to_process);

                        if (reply(sig))
                        {   
                           ret = EFMI_END_OF_LIST;
                           break;
                        }
                     }
                     swap_info(sig, &tes1, in);

                     if (reply(sig))
                     {   
                        ret = EFMI_END_OF_LIST;
                        break;
                     }
                     break;
                  }                  
                  case TRACE_USER_EVENT:
                  {
                     struct trace_user_event *tue = &te->trace_user_event;
                     
                     if (!is_user_event_ok(tue, trace, trace_len))
                     {
                        ret = EFMI_KERNEL_CORRUPTED;
                        break;
                     }
                     if (!idbuf_contains(&pids, tue->process->pid))
                     {                           
                        process_info(sig, tue->process);

                        if (reply(sig))
                        {   
                           ret = EFMI_END_OF_LIST;
                           break;
                        }
                     }
                     user_info(sig, &tes1, in);

                     if (reply(sig))
                     {   
                        ret = EFMI_END_OF_LIST;
                        break;
                     }
                     break;                                                               
                  }
                  case TRACE_ALLOC:
                  case TRACE_FREE_BUF:
                  {
                     struct trace_alloc_free_buf *tafb = &te->trace_alloc_free_buf;
                     
                     if (!is_alloc_free_buf_ok(tafb, trace, trace_len))
                     {
                        ret = EFMI_KERNEL_CORRUPTED;
                        break;
                     }
                     if (!idbuf_contains(&pids, tafb->process->pid))
                     {                           
                        process_info(sig, tafb->process);

                        if (reply(sig))
                        {   
                           ret = EFMI_END_OF_LIST;
                           break;
                        }
                     }
                     alloc_free_buf_info(sig, &tes1, in);

                     if (reply(sig))
                     {   
                        ret = EFMI_END_OF_LIST;
                        break;
                     }                                  
                     break;                                                               
                  }
                  case TRACE_BIND:
                  {
                     struct trace_bind *tb = &te->trace_bind;
                     
                     if (!is_bind_ok(tb, trace, trace_len))
                     {
                        ret = EFMI_KERNEL_CORRUPTED;
                        break;
                     }
                     if (!idbuf_contains(&pids, tb->process->pid))
                     {                           
                        process_info(sig, tb->process);

                        if (reply(sig))
                        {   
                           ret = EFMI_END_OF_LIST;
                           break;
                        }
                     }
                     bind_info(sig, &tes1, in);

                     if (reply(sig))
                     {   
                        ret = EFMI_END_OF_LIST;
                        break;
                     }                                  
                     break;                                                               
                  }
		  case TRACE_RECEIVE_TMO:
                  {
                     struct trace_receive_tmo *trt = &te->trace_receive_tmo;
                     
                     if (!is_receive_tmo_ok(trt, trace, trace_len))
                     {
                        ret = EFMI_KERNEL_CORRUPTED;
                        break;
                     }
                     if (!idbuf_contains(&pids, trt->process->pid))
                     {                           
                        process_info(sig, trt->process);

                        if (reply(sig))
                        {   
                           ret = EFMI_END_OF_LIST;
                           break;
                        }
                     }
                     timeout_info(sig, &tes1, in);

                     if (reply(sig))
                     {   
                        ret = EFMI_END_OF_LIST;
                        break;
                     }                                  
                     break;                                                               
                  }
                  case TRACE_SYSTEM_START:
                  {
                     struct trace_system_start *tss = &te->trace_system_start;
                     
                     if (!is_system_start_ok(tss, trace, trace_len))
                     {
                        ret = EFMI_KERNEL_CORRUPTED;
                        break;
                     }
                     reset_info(sig, &tes1, in);

                     if (reply(sig))
                     {   
                        ret = EFMI_END_OF_LIST;
                        break;
                     }                                  
                     break;                                                               
                  }
                  default:
                     break;
               }
               
               if (ret)
               {
                  break;
               }
               in++;
            }   
         }

         ose_trace_close(&tes1);
      }
   }
   else
   {
      ret = EFMI_UNAVAILABLE;
   }
   
   return ret;
}

EfmiStatus fmm_ramtrace_get_trace_request(
                 int32_t (*reply)(FmmSignal *sig),
                 int32_t (*endmark)(FmmSignal *sig, int32_t status),
                 const MonitorGetTraceRequest *request)
{
   EfmiStatus ret;
   FmmSignal *sig;
   
   assert(request != NULL);
   assert(request->sigNo == MONITOR_GET_TRACE_REQUEST);
   
   ret = list_ramtrace(reply, request);   
   
   sig = fmm_output_signal;
   sig->size = sizeof(MonitorGetTraceEndmark);
   sig->data.get_trace_endmark.sigNo = MONITOR_GET_TRACE_ENDMARK;
   
   switch (ret)
   {
      case EFMI_OK:
         sig->data.get_trace_endmark.status = MONITOR_STATUS_OK;
         break;
      case EFMI_END_OF_LIST:
         sig->data.get_trace_endmark.status = MONITOR_STATUS_CANCELLED;
         break;
      default:
         sig->data.get_trace_endmark.status = MONITOR_STATUS_KERNEL_ERROR;
         break;
   }
   
   if (endmark(sig, ret))
   {
      return (ret) ? ret : EFMI_END_OF_LIST;
   }
            
   return ret;
}
