/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2006 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

#ifndef _ALLOCATOR_H
#define _ALLOCATOR_H

/* C runtime include files. */
#include <stddef.h>

/* OSE public include files. */
#include "osetypes.h"
#include "mm.sig"

/* OSE internal include files. */

/* MM internal include files. */
#include "partition.h"

/* Inform phys_object.h that there is a struct allocator. */
struct allocator;
#include "phys_object.h"

/*
 * Avoid free macro from OSE Heap. Observe that the heap
 * is not used in MM, but stddef.h includes malloc.h that
 * causes the problem.
 */
#ifdef free
#  undef free
#endif

/*
 * Allocator is a data structure for used for managing allocation and
 * writing to physical memory.
 * This interface has been implemented for allocation from ram and
 * also when allocating from a PMM.
 *
 * These functions are implemented for an allocator:
 *
 *   alloc       - Allocation of physical memory.
 *   free        - Freeing physical memory.
 *   write       - Writing to physical memory. Observe that this physical
 *                 memory is assumed to be mapped 1-1 to logical memory, and
 *                 that inadequate access permissions for writing are not
 *                 considered.
 *   info        - Implement the signal interface that deals with info tags
 *                 for physical memory types.
 *   within_area - Returns true if the given physical address is within
 *                 range of a physical allocator.
 *
 * Allmost all of these functions must be implemented, and their entries in
 * the corresponding allocator struct must be set.
 * The only exception is write, which should be set to zero, if no write
 * function has been implemented.
 *
 * All functions get a pointer to their own allocator struct as an argument.
 * The functions alloc, free, write and info returns a status code
 * that is declared in enum MmStatus.
 *
 * Besides the functions, each allocator must also set the "alignment"
 * field to the alignment accepted by the allocator implementation.
 * This alignment should be:
 *
 *    MAX(allocator minimum alignment, minimum page size from HAL) FIXME
 */

/* Each allocator defines its own structure for internal data. This
 * declaration is just to state that such a structure exists. */
struct allocator_internal_data;

struct allocator
{
   struct allocator *next;

   log_size_t alignment;                       /* Alignment requirement */

   void
   (*add_sas_bank)(struct allocator *allocator,
                   phys_addr_t base, phys_size_t size);

   void
   (*add_sase_bank)(struct allocator *allocator,
                    log_addr_t base, log_size_t size);
   
   /* Allocate scattered phys mem. */
   mm_status_t
   (*alloc_s)(const struct allocator *allocator,
              struct phys_object *phys_object,
              phys_size_t size,
              phys_size_t alignend);
   
   /* Allocate continuous range of physical memory.
    * Memory is primary allocated from SAS banks,
    * secondly from SASE banks.*/
   mm_status_t
   (*alloc_c)(const struct allocator *allocator,
              struct phys_object *phys_object,
              phys_size_t size,
              phys_size_t alignend,
              phys_addr_t range_base,
              phys_size_t range_size);
   
   /* Allocate continuous range of physical memory from SASE bank,
    * i.e. logical memory addresses equals physical memory addresses
    * (physical memory to be used for SASE log_mem). */
   mm_status_t
   (*alloc_e)(const struct allocator *allocator,
              struct phys_object *phys_object,
              phys_size_t size,
              phys_size_t alignend,
              phys_addr_t range_base,
              phys_size_t range_size);
   
   /* Allocate continuous phys mem at fixed address. */
   mm_status_t
   (*alloc_f)(const struct allocator *allocator,
              struct fragment_container **fragment_container,
              phys_addr_t base,
              phys_size_t size);
   
   /* Free allocated memory. */
   void
   (*free)(const struct allocator *allocator,
           struct phys_object *phys_object);
   
   /*
    * Copy <size> bytes from <source> to <destination>.
    * It is assumed that the given addresses are mapped to physical memory
    * that belongs to this allocator, but this is not checked.
    * Further, there are no alignment requirements on either the addresses
    * or on the size. FIXME_is this so?
    */
   mm_status_t
   (*write)(struct allocator *allocator,
            log_addr_t source,
            log_addr_t destination,
            log_size_t size);

   /*
    * Process a signal with a tag list.
    */
   void
   (*info)(struct allocator *allocator,
           union SIGNAL *tag_sig,
           union SIGNAL **result);

   /*
    * Get the total amount of memory available to the allocator.
    */
   phys_size_t
   (*get_total_mem)(const struct allocator *allocator);

   /*
    * Get the amount of free memory available to the allocator.
    */
   phys_size_t
   (*get_free_mem)(const struct allocator *allocator);

   /*
    * Get the amount of physical fragments used by the allocator.
    */
   U32
   (*get_frag_count)(const struct allocator *allocator);

   /*
    * Get the biggest physical fragment used by the allocator.
    */
   phys_size_t
   (*get_biggest_frag)(const struct allocator *allocator);

   /*
    * Get the smallest amount of free memory so far for this allocator.
    */
   phys_size_t
   (*get_free_watermark)(const struct allocator *allocator);

   /*
    * Returns TRUE if and only if the area indicated by (<base>, <size>)
    * overlaps with the area handled by the allocator.
    */
   mm_bool_t
   (*overlap)(const struct allocator *allocator,
              phys_addr_t base,
              phys_size_t size);

   /* Identical to overlap, except that base and size are updated with the base
    * and size of the overlapping bank if there is an overlap. Otherwise base
    * and size will be unchanged.
    */
   mm_bool_t
   (*overlap2)(const struct allocator *allocator,
               phys_addr_t *base,
               phys_size_t *size);

   /*
    * Returns TRUE if and only if the area indicated by (<base>, <size>)
    * is a subarea of the area handled by the allocator.
    */
   mm_bool_t
   (*sub_area)(const struct allocator *allocator,
              phys_addr_t base,
              phys_size_t size);

   /*
    * Return amount of free space for SASE or SAS physical memory.
    * Null if method not supported.
    */
   phys_size_t
   (*size_of_free_mem)(const struct allocator *allocator,
                       mm_bool_t sas);

   char name[MM_NAME_LENGTH + 1];             /* Name of allocator. */

   struct page_info *
   (*get_page_stat)(void);

   /* dump the allocators internal state using the print function print. */
   void
   (*dump_state)(const struct allocator*allocator,
                 int (*print)(const char *, ...));

   /* Returns amount of memory consumed by allocator metadata, in bytes. */
   phys_size_t
   (*get_metadata_size)(const struct allocator *allocator);
     
   struct allocator_internal_data *state;    /* Private state. */

    /*
    * Find smallest possible super set that cover all memory banks.
    * The super set is returned in the area (<base>, <top>)
    */
   /* FIXME placing find_super_set field last in struct eases
    * but does not solve/explain OSECR-8985 */
   void
   (*find_super_set)(const struct allocator *allocator,
                     phys_addr_t *base,
                     phys_size_t *top);

   /*
    * Retrieves information about the location of a phys_mem bank. Returns
    * 0 when the bank index number is out of bounds. Indexes start at 0.
    */
   int
   (*get_bank_info)(const struct allocator *allocator,
                    int index, phys_addr_t *base, phys_size_t *size);
};

/*
 * Initialize sub-system. To be called before any other function described in
 * this file.
 */
void
mm_allocator_init(void);

/*
 * Register an allocator to allow it to be found by find_allocator_by_X
 */
void mm_register_allocator(struct allocator *allocator);

/*
 * Returns the allocator with the given name.
 * Return NULL on no match.
 */
struct allocator *
mm_find_allocator_by_name(const char * name);


/*
 * Search for an allocator that handles physical memory within the range
 * indicated by (<base>,<size>). To be more specific: Finds an allocator
 * that handles a subrange of the specified range.
 * If no such allocator is found, NULL is returned.
 */
struct allocator *
mm_find_allocator_by_range(phys_addr_t base, phys_size_t size);


/*
 * Print states for all allocators to a shell of to the RAM log. (ramlog_printf
 * returns size_t instead of int, but the return value is not used so the cast
 * is safe):
 *  dump_allocator_states((int (*)(const char *, ...))ramlog_printf);
 *  dump_allocator_states(printf);
 */
void
dump_allocator_states(int (*print)(const char *, ...));

void
ram_allocator_prepare_for_warm_start(U32 nr_of_regions);

#endif /* _ALLOCATOR_ */
