/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2002-2007 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */
/*
 */

/* C runtime include files. */
#include "stddef.h"
#include "string.h"

/* OSE public include files. */
#include "assert.h"
#include "infmt.h"
#include "mm_err.h"
#include "outfmt.h"
#include "ramlog.h"

/* OSE internal include files. */

/* MM internal include files. */
#include "codec.h"
#include "common.h"
#include "partition.h"
#include "gmm.h"


/*
**  Note!!! If the meta data string format (defined by ENCODE_FMT,
**  DECODE_FMT, FS, FS_VAL etc) needs update, the version number
**  in pmm.sig must be incremented as well.
**  Custom PMM implementations (Razor) might depend on this format.
*/

/* Field separator in the metadata string. Used between the string elements.
 * Any character that does not occur in a name would do. */
#define FS "\034"
#define FS_VAL 034

/* Metadata format strings */
#define ENCODE_FMT \
  "0x%lx 0x%lx 0x%lx 0x%lx"\
  FS ">%s" FS ">%s"

#define DECODE_FMT \
"%lx %lx %lx %lx " FS "%n%*[^" FS "]" FS "%n%*[^" FS "]"

/*
 * Saves the following state:
 * - base address
 * - size
 * - cache-mode
 * - permissions
 * - partition name
 * - region name
 * - allocator name
 *
 */
char *
mm_encode_metadata(struct region *region)
{
  char *buf;
  size_t buf_size;
  struct partition *partition =
     mm_log_object_partition(region->log_object, region->domain);

  assert(partition != NULL);
  
  buf_size = 50
    +strlen(partition->class->name)
    +strlen(region->name);
  
  buf = (char*)alloc(buf_size, 0);
  buf[buf_size-1] = 0;
  
  snprintf(buf,
           buf_size,
           ENCODE_FMT,
           region->log_object->base,
           region->log_object->size,
           region->cache_mode,
           region->permissions,
           region->log_mem,
           region->name);
  
  if (buf[buf_size-1] != 0) {
    /* Buffer was not large enough */
    free_buf((union SIGNAL**)&buf);
    return 0;
  }
  return buf;
}

static U32
indicesOk(size_t bufLen, U32 part_name, U32 reg_name) {
  if (part_name == 0) return 0;
  if (part_name >= reg_name) return 0;
  if (reg_name >= bufLen) return 0;
  return 1;
}

/*
 * See codec.h
 */
struct region *
mm_decode_metadata(U8 *buf0, log_addr_t in_base)
{
  /* The input comes from the field PMM_RESTORE_REGION_REPLY->metadata,
   * which is of type U8[] */

   char *buf = (char*)buf0;
   size_t bufLen;
   log_addr_t base;
   log_size_t size;
   U32 permissions;
   U32 cache_mode;
   struct region *region;
   U32 partitionIx = 0, regNameIx = 0;
   char *partitionName, *regionName;
   struct partition *partition;
   struct log_object *lobj;
   struct domain *domain;
   mm_status_t status;

   assert(buf != NULL);
   bufLen = strlen((char*)buf);

   /* Decode meta-data for a logical region. */
   sscanf(buf, DECODE_FMT, &base, &size, &cache_mode, &permissions,
          &partitionIx, &regNameIx);
   
   if (!indicesOk(bufLen, partitionIx, regNameIx))
     return NULL;
   
   /* buf[x-1]now contains a FS. Replace with a string terminator. */
   assert(buf[regNameIx-1] == FS_VAL);
   buf[regNameIx-1] = 0;
   
   /* Since the pattern in DECODE_FMT, "%*[^FS]", requires a nonempty string,
    * a '>' was added by the encoder at the beginning of each string. The
    * purpose is only to avoid zero length strings, and these characters are
    * not a parts of the saved strings.
    */
   partitionName = &buf[partitionIx+1];
   regionName    = &buf[regNameIx+1];
   
   /*
    * Make sure base address in meta-data equals base address in
    * PMM_RESTORE_START_REPLY signal.
    */
   if (in_base != base)
     return NULL;

   domain = mm_find_domain_by_domid(&gmm_state.domains, MM_KERNEL_DOMAIN);
   if (domain == NULL)
   {
      INTERNAL_ERROR();
   }
   
   /*
    * Make sure the partition exists and that the region is located
    * entirely inside the partition.
    */
   partition = mm_find_domain_partition_by_name(domain, partitionName);
   
   if (partition == NULL) {
     MM_FATAL_ERROR2(("Partition '%s' not found\n", partitionName),
                     OSE_MM_EPMM_LOG_MEM_ERROR,
                     (OSERRCODE) partitionName);
   }

   if (!partition->class->is_superset_of(
          partition->class, base, size))
   {
     /*
      * This error case has to be handled separately since PMM is not
      * able to deal with this problem. Why? Because the problem is
      * caused by an inconsistent change in the system configuration.
      *
      * Consider the following scenario:
      *   1 A persistent region R is created in partition A
      *   2 The system shuts down
      *   3 The system configuration is modified; partition A is
      *     removed.
      *   4 The system start up. During the RESTORE phase region R
      *     is to be re-created but since partition A no longer exists
      *     the logical region for R can not be created.
      */
     
     
     MM_FATAL_ERROR2(("No log_mem element covering persistent region %s,"
                      "base = %#.8lx, size = %#.8lx\n",
                      regionName, base, size),
                     OSE_MM_EPMM_LOG_MEM_ERROR,
                     (OSERRCODE) partitionName);
   }

   /* Create the logical memory container */
   lobj = mm_create_log_object();
   if (lobj == NULL)
   {
      INTERNAL_ERROR();
   }

   /* Allocate logical memory */
   status = partition->class->alloc_f(
      partition, lobj, size, base, MM_PARTITION_OPT_EXCLUSIVE);
   if (status != MM_SUCCESS)
   {
      INTERNAL_ERROR();
   }

   /* Mark logical memory as used */
   mm_attach_log_object(lobj, domain);

   /* Create the region */
   status = mm_add_region(MM_KERNEL_DOMAIN,
                          lobj,
                          NULL,     /* Logical only region so far */
                          permissions,
                          cache_mode,
                          0,        /* Options only for tentative regions */
                          MM_REGION_OPT_PERSISTENT,
                          regionName,
                          partitionName,
                          &region);
   if (status != MM_SUCCESS)
   {
      INTERNAL_ERROR();
   }

   
   return region;
}
