/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2005-2006 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

#include "assert.h"
#include "compat.h"
#include "gmm.h"
#include "hal.h"
#include "common.h"
#include "mm_profile.h"

#ifdef ASSERT
static void
assert_size(log_size_t size)
{
   unsigned int i;

   /* Traverse through the page sizes supported by the HAL module. */
   for (i = 0; i < hal_get_info()->no_page_sizes; i++)
   {
      if (size == hal_get_info()->page_sizes[i])
      {
         return;
      }
   }
   assert(0);
}
#else
#define assert_size(size)
#endif


/*
 * Compute Largest Page Size.
 */
static log_size_t
lps(log_addr_t la, phys_addr_t pa, phys_size_t phys_size)
{
   unsigned int i;
   log_size_t size;
   log_size_t page_size;

   assert(mm_log_aligned(la, hal_get_info()->page_sizes[0]));
   assert(mm_phys_aligned(pa, hal_get_info()->page_sizes[0]));
   assert(ALIGNED(phys_size, hal_get_info()->page_sizes[0]));

   /* Traverse through the page sizes supported by the HAL module. */
   for (i = 0, page_size = hal_get_info()->page_sizes[0];
        i < hal_get_info()->no_page_sizes;
        i++)
   {
      size = hal_get_info()->page_sizes[i];

      /* The logical and the physical addresses must be naturally aligned. */
      if (phys_size >= size &&
          mm_log_aligned(la, size) && mm_phys_aligned(pa, size))
      {
         page_size = size;
      }
      else
      {
         break;
      }
   }
   assert_size(page_size);
   return page_size;
}


hal_status_t
mm_map_area(struct hal_domain *hal_domain,
            struct phys_object *phys_object,
            log_addr_t la,
            int options,
            U32 access_attr,
            U32 mem_attr)
{
   log_addr_t curr_la;
   struct phys_mem_cursor cursor;
   log_size_t total_size;
   total_size = 0;

   assert(phys_object != NULL);
   curr_la = la;

#ifdef OSPP
   if (options & HAL_CHANGE)
   {
      options |= HAL_SKIP_INVALIDATE_TLB;
   }
#endif

   
   /* Check that all physical fragment are within the HAL's capability.
    * This is done in a separate loop to avoid checking the function pointer
    * hal_validate_phys_range more than once.
    */
   if (NULL != hal_validate_phys_range)
   {
      mm_phys_mem_get_first(phys_object, &cursor);
      
      do
      {
         hal_status_t rc = hal_validate_phys_range(cursor.phys_page.pa,
                                                   cursor.phys_page.size);
         if (HAL_SUCCESS != rc)
         {
            return rc;
         }
      } while (mm_phys_mem_get_next(&cursor));
   }
   
   MM_PROFILE_CHECKPOINT("validate_phys_range", __FILE__, __LINE__);
   
#ifdef MM_PROFILE
#if defined(OSPP) || defined(OSMIPS)
   {
      mm_bool_t global = !(options & HAL_SHARE);
      mm_bool_t mas = options & HAL_MAS;
      mm_bool_t map_global_page_tree = global && !mas;
      mm_bool_t application_domain = (gmm_state.system_domain->hal_domain !=
                                      hal_domain);

      MM_PROFILE_CHECKPOINT("map_area",XSTR(map_local_page_tree),
                            application_domain);
      MM_PROFILE_CHECKPOINT("map_area",XSTR(map_global_page_tree),
                            map_global_page_tree);
   }
#elif defined(OSARM)
   {
      mm_bool_t global = !(options & HAL_SHARE);
      MM_PROFILE_CHECKPOINT("map_area","broadcast_page_mapping",global);
   }
#else
   MM_PROFILE_CHECKPOINT("map_area", __FILE__, __LINE__);
#endif
#endif   
   
   /*
    * Map all fragments.
    */
   mm_phys_mem_get_first(phys_object, &cursor);
   do
   {
      struct phys_page phys_page;
      log_size_t page_sum;

      page_sum = 0;
      phys_page = cursor.phys_page;

      /* Used as end-of-vector internally in phys_container.c. */
      assert(phys_page.size != 0);

      while (page_sum < phys_page.size)
      {
         hal_status_t rc;
         log_size_t page_size;

         page_size = lps(curr_la, phys_page.pa, phys_page.size - page_sum);
         rc = hal_map_page(hal_domain,
                           curr_la,
                           phys_page.pa,
                           page_size,
                           options,
                           access_attr,
                           mem_attr);
         if (rc != HAL_SUCCESS)
         {
	    if (ose_mm_reclaim_component_memory(&gmm_state, page_size,
						MM_INTERNAL_RAM))
	    {
	       continue;
	    }
	    
            if (0 != curr_la - la)
            {
               /* Only unmap in case at least one page
                * has been successfully mapped. */
               hal_unmap_range(hal_domain, la, curr_la - la, options);
            }
            return rc;
         }
         curr_la += page_size;
         phys_page.pa += page_size;
         page_sum += page_size;
      }
      total_size += page_sum;
      assert(page_sum == phys_page.size);
   } while (mm_phys_mem_get_next(&cursor));

   MM_PROFILE_CHECKPOINT("map_area", __FILE__, __LINE__);
   
   assert(curr_la == total_size + la);

#ifdef OSPP
   if (options & HAL_CHANGE)
   {
      /*
       * Avoid per-fragment tlb-invalidation, instead order hal to
       * invalidate the entire range at one sweep. (OSECR-3154)
       * This will on PPC PEM with hash table give possibilty to
       * clean entire hash table only once per region which will boost
       * performance of program loading. Note that PGM when
       * the text-writing of a LM is complete the PGM will issue a
       * change_permission() for the whole
       * text region to write-protection.
       *
       * This makes it for example possible  to take a decision to
       * clean entire hash table.
       */
      MM_PROFILE_CHECKPOINT("TLB invalidate(ALL)", __FILE__, __LINE__);

      if (options & HAL_GLOBAL_MAPPING ||
          options & HAL_GLOBAL_MAPPING_TRANSIENT) {
         /* PPC:
            A (system domain) region with tlb global attribute,
            can only exist in the system domain page table. (OSECR-8881)
            Avoid unnecessary tlb invalidation looping all domains.
         */
         hal_invalidate_tlb_range(la, total_size, hal_domain, TRUE);
      } else
      {
         hal_invalidate_tlb_range(la, total_size, hal_domain, FALSE);
      }
   }
   
   else if (options & HAL_SHARE)
   {
      extern int ose_mm_segnum_cnt;

      /* Must invalidate tlb if the newly mapped area is a result of
       * a share region request.
       *
       * A supervisor access can have occurred in the mapped
       * area prior to this area was mapped and added to the domain local
       * page table.
       * Such a supervisor access will have been translated via the global page
       * table and inserted in tlb/hash with supervisor permission only. To
       * ensure that the PTE's from new mapping (with possible added usr mode
       * will be used in TLB/hash, we need to do tlb/hash invalidation.
       *
       *  The problem only arises if the a usr block was created in super visor
       *  block. This is default not permitted by the exec but
       *  OSE_EILLEGAL_USER_MODE is non fatal and can be ignored by the client.
       *  (However it would be possible to share data between exec and
       *  mm, since the exec detects OSE_EILLEGAL_USER_MODE, and tlb
       *  invalidate in domains where a OSE_EILLEGAL_USER_MODE has occurred.)
       */

      if (ose_mm_segnum_cnt > 1)
      {
      /* Why if (ose_mm_segnum_cnt > 1) ?
       * Explanation:
       * In a monolitic system with only one page table no sharing
       * can take place. 
       * 
       * (Also note that PPC440 did not boot if tlb invalidation occured during
       * system boot.)
       */
         MM_PROFILE_CHECKPOINT("TLB invalidate(1)", __FILE__, __LINE__);
         hal_invalidate_tlb_range(la, total_size, hal_domain, TRUE);
         
         
      }
   }
#endif

   MM_PROFILE_CHECKPOINT("TLB invalidate_done", __FILE__, __LINE__);

   return HAL_SUCCESS;
}
