/* COPYRIGHT-ENEA-SRC-R2 *
**************************************************************************
* Copyright (C) 2004-2007 by Enea Software AB.
* All rights reserved.
*
* This Software is furnished under a software license agreement and
* may be used only in accordance with the terms of such agreement.
* Any other use or reproduction is prohibited. No title to and
* ownership of the Software is hereby transferred.
*
* PROPRIETARY NOTICE
* This Software consists of confidential information.
* Trade secret law and copyright law protect this Software.
* The above notice of copyright on this Software does not indicate
* any actual or intended publication of such Software.
**************************************************************************
* COPYRIGHT-END */

/* C runtime include files. */
#include <stddef.h>
#include <stdlib.h>
#include <string.h>

/* OSE public include files. */
#include "assert.h"
#include "efs.h"
#include "errno.h"
#include "mm.sig"
#include "mm_err.h"
#include "ramlog.h"
#include "sysparam.h"
#ifndef HOST_TEST
#include "cpu_hal_arch.h"
#  ifdef OSARM
#  define ARM_HAL_IS_V6 (hal_arch_cpu_descriptor->exec_attributes & CPU_HAS_MMU_V6)
#  endif
#endif

/* OSE internal include files. */
#include "zone.h"
#include "slab.h"

/* MM internal include files. */
#include "allocator.h"
#include "common.h"
#include "configure.h"
#include "compat.h"
#include "hal.h"
#include "gmm.h"
#include "partition.h"
#include "implicit_partition.h"
#include "pmm.sig"
#include "ram_allocator.h"
#include "mm_log.h"

#if defined(OSMIPS) && !defined(HOST_TEST)
#include "cpu_access.h"
#endif

/*
 * Prefixes for board parameters holding MM data.
 */
#define CACHETYPE_PREFIX     "krn/cache_type/"
#define MEMTYPE_PREFIX       "krn/mem_type/"
#define LOGMEM_PREFIX        "krn/log_mem/"
#define PHYS_MEM_PREFIX      "krn/phys_mem/"
#define RES_MEM_PREFIX       "krn/res_mem/"
#define REGION_PREFIX        "krn/region/"
#define PHYS_RAM_PREFIX      PHYS_MEM_PREFIX "RAM"
#define PARAM_NAME_INHIBIT_MM_CONF_CHECK "krn/inhibit_mm_conf_check"

/*
 * MM boot heap defines.
 */
#define BOOT_REGION_NAME   "MM-meta-data"

/* Keywords for attributes. */
#define TOP_ATTRIBUTE_KEY           "top"
#define MAPPING_ATTRIBUTE_KEY       "mapping_type"
#define PERMISSION_ATTRIBUTE_KEY    "perm"
#define CACHE_ATTRIBUTE_KEY         "cache"
#define MEMORY_ATTRIBUTE_KEY        "mem"
#define PHYS_BASE_ATTRIBUTE_KEY_OLD "phys_base"
#define PHYS_BASE_ATTRIBUTE_KEY_U   "phys_base_unreserved"
#define PHYS_BASE_ATTRIBUTE_KEY_R   "phys_base_reserved"
#define MANAGER_MODE_ATTRIBUTE_KEY  "manager_mode"
#define IMPLICIT_ATTRIBUTE_KEY      "implicit"
#define ALIAS_ATTRIBUTE_KEY         "alias"

#ifdef OSARM
#  define HAL_PREFIX            "krn/mm/hal_options"
#  define STATIC_PAGETABLES_KEY "static_pagetables"
#endif

#define INVALID_LOG_MEM_BASE    ((log_addr_t)0xffffffff)
#define INVALID_LOG_MEM_SIZE    ((log_size_t)0xffffffff)


/******************************************************************************
 * Iterate over all system parameters matching key, using buf as current.
 */
#define MAP_SYSPARAM(key, buf) \
   strcpy(buf, key); \
   while (sysparam_list_next(buf, buf, sizeof(buf)) && \
         0 == strncmp(buf, key, strlen(key)))

/******************************************************************************
 * Like MAP_SYSPARAM, except for the filter parameter. It is a pointer to
 * function that takes the parameter name as argument. If filter is not NULL
 * and it returns 0, then the parameter is excluded. */

#define MAP_SYSPARAM_FILTER(key, buf, filter) \
   MAP_SYSPARAM(key, buf) /*lint -save -e506 Constant value Boolean */\
     if (NULL == (filter) || (filter)(buf)) /*lint -restore */

/*****************************************************************************/

#define HAS_MMU    (hal_get_info()->has_mmu)

/*****************************************************************************/

#define ATTRIBUTE_ERROR(param, attr, ecode) \
do {\
   MM_FATAL_ERROR2(("Attribute '%s' for configuration parameter %s is %s.\n",\
                    (attr), (param), \
                    ENOENT == (ecode) ? "missing": "invalid"),\
                   OSE_MM_ECONFIGURATION_ERROR, \
                   (OSERRCODE)"Invalid attribute.");\
} while (0)

/******************************************************************************
 * Read the attribute for the parameter. Does not return unless the operation
 * was successful.
 *
 * The temporay variable temp_var is needed since sizeof enum is not always 
 * the same as sizeof long.
 */
#define REQUIRE_ATTR_STRING(var, length, param, attr) \
   do {\
      int _rc_ = sysparam_readattr_string((param), (attr), (var),\
                                          (length));\
      if (EOK != _rc_) {\
         ATTRIBUTE_ERROR((param), (attr), _rc_);\
      }\
   } while (0)

/******************************************************************************
 * Read the attribute for the parameter. Does not return unless the operation
 * was successful.
 *
 * The temporay variable temp_var is needed since sizeof enum is not always 
 * the same as sizeof long.
 */
#define REQUIRE_ATTR_ENUM(var, param, attr, encoding) \
   do {\
      long temp_var;\
      int _rc_ = sysparam_readattr_enum((param), (attr), (encoding),\
                                        &(temp_var));\
      var = temp_var;\
      if (EOK != _rc_) {\
         ATTRIBUTE_ERROR((param), (attr), _rc_);\
      }\
   } while (0)

/******************************************************************************
 * Read the mapping type attribute for the parameter. Does not return unless
 * the operation was successful.
 *
 * The temporay variable temp_var is needed since sizeof enum is not always 
 * the same as sizeof long.
 *
 * If the mapping_type attribute is missing, i.e. not invalid, SAS is
 * chosen as the default mapping type. 
 */
#define REQUIRE_ATTR_ENUM_MAP_TYPE(var, param, attr, encoding) \
   do { \
      long temp_var; \
      int _rc_ = sysparam_readattr_enum((param), (attr), (encoding), \
                                        &(temp_var)); \
      if (_rc_ == ENOENT) \
         var = SAS; \
      else \
      { \
         var = temp_var; \
         if (EOK != _rc_) { \
            ATTRIBUTE_ERROR((param), (attr), _rc_); \
         } \
      } \
   } while (0)
        
/*****************************************************************************/

#define VERBOZE(s)  log(CATEGORY_START, s)
#define VERBOZE_2(s)  log(CATEGORY_START_DETAIL, s)

typedef char param_t[SYSPARAM_MAX_NAME_LENGTH+1];

struct log_mem_param {
   const char     *name;
   char           alias[MM_NAME_LENGTH+1];
   char           bank[MM_NAME_LENGTH+1];
   Boolean        implicit;
   mapping_type_t mapping_type;
   log_addr_t     base;
   log_size_t     size;
   mm_bool_t      manager_mode;
};

static const struct enumeration cache_encoding[] =
{
   {"copy_back", MM_CACHE_COPY_BACK},
   {"inhibited", MM_CACHE_INHIBITED},
   {"write_through", MM_CACHE_WRITE_THROUGH},
   {"cache_copy_back", MM_CACHE_COPY_BACK},
   {"cache_inhibited", MM_CACHE_INHIBITED},
   {"cache_write_through", MM_CACHE_WRITE_THROUGH},
   { NULL, 0}
};

static const struct enumeration permission_encoding[] =
{
   {"su_na_usr_na",   SU_NA_USR_NA},
   {"su_ro_usr_na",   SU_RO_USR_NA},
   {"su_ro_usr_ro",   SU_RO_USR_RO},
   {"su_rw_usr_na",   SU_RW_USR_NA},
   {"su_rw_usr_ro",   SU_RW_USR_RO},
   {"su_rw_usr_rw",   SU_RW_USR_RW},
   {"su_rx_usr_na",   SU_RX_USR_NA},
   {"su_rx_usr_ro",   SU_RX_USR_RO},
   {"su_rx_usr_rx",   SU_RX_USR_RX},
   {"su_rwx_usr_na",  SU_RWX_USR_NA},
   {"su_rwx_usr_ro",  SU_RWX_USR_RO},
   {"su_rwx_usr_rx",  SU_RWX_USR_RX},
   {"su_rwx_usr_rw",  SU_RWX_USR_RW},
   {"su_rwx_usr_rwx", SU_RWX_USR_RWX},
   { NULL, 0}
};


static const struct enumeration mem_encoding[] =
{
   {"speculative_access", MM_SPECULATIVE_ACCESS},
   {"ordered_access", MM_ORDERED_ACCESS},
   { NULL, 0}
};

static const struct enumeration mapping_type_encoding[] =
{
   {"SASE", SASE},
   {"SAS", SAS},
   {"MAS", MAS},
   { NULL, 0}
};

#if 0
/* currently unused. */
static const struct enumeration bat_type_encoding[] =
{
   {"data",      (HAL_CONTENTS_TYPE_DATA)},
   {"text",      (HAL_CONTENTS_TYPE_TEXT)},
   {"text_data", (HAL_CONTENTS_TYPE_TEXT|HAL_CONTENTS_TYPE_DATA)},
   {NULL, 0}
};
#endif

static const char *
num_to_str(const struct enumeration *enu, U32 value)
{
   while (enu->name != NULL) {
      if ((U32)enu->value == value) {
         return enu->name;
      }
      enu++;
   }

   return "???";
}

/*****************************************************************************/

struct region_conf {
   log_addr_t log_base;
   log_addr_t log_top;
   phys_addr_t phys_base;
   phys_addr_t phys_top;
   log_size_t size;
   mm_bool_t reserve_phys_mem;
};

/*****************************************************************************/

static struct boot_heap_data {
   log_addr_t init_base;
   log_addr_t init_top;
   log_addr_t curr_base;

   struct fragment_container *frag_list;

   U32 active;
   U32 auto_configured;
} boot_heap_data;

static void
read_logmem_entry(const char *param, struct log_mem_param *data);

/*****************************************************************************/

static void print_boot_heap(void) {
   struct fragment_container *f;

   if (!boot_heap_data.active) {
      ramlog_printf("Boot heap is not active\n");
      return;
   }
   
   ramlog_printf("Initial range : [0x%08lx-0x%08lx]\n",
                 boot_heap_data.init_base, boot_heap_data.init_top);
   
   ramlog_printf("curr_base     : 0x%08lx\n", boot_heap_data.curr_base);
   ramlog_printf("phys_frag     :\n");
   for (f = boot_heap_data.frag_list; f; f = FC_GET_NEXT(f)) {
      ramlog_printf("[0x%08llx-0x%08llx]\n", FC_GET_BASE(f), FC_GET_TOP(f));
   }
}

/*****************************************************************************/

static const char *
ram_sase_param_name(void)
{
   static const char *param_name = NULL;

   if (NULL == param_name)
   {
      param_name = HAS_MMU
         ? LOGMEM_PREFIX RAM_SASE_NAME
         : LOGMEM_PREFIX RAM_NAME;
   }
   return param_name;
}

/*****************************************************************************/

static int
is_alias_log_mem(const char *param)
{
   char alias_name[MM_NAME_LENGTH];

   int status = sysparam_readattr_string(
      param, ALIAS_ATTRIBUTE_KEY, alias_name, MM_NAME_LENGTH);
   
   if (status == EOK)
      return 1;

   return 0;
}

/*****************************************************************************/

static int
is_implicit_log_mem(const char *param)
{
   char alias_name[MM_NAME_LENGTH];

   int status = sysparam_readattr_string(
      param, IMPLICIT_ATTRIBUTE_KEY, alias_name, MM_NAME_LENGTH);
   
   if (status == EOK)
      return 1;

   return 0;
}

/*****************************************************************************/

#ifdef HOST_TEST
static void (*test_init_boot_heap)(struct boot_heap_data*,
                                   log_addr_t *base,
                                   log_addr_t *top) = NULL;
#endif

static void
init_boot_heap(log_addr_t base, log_addr_t top) {
   ramlog_printf("init_boot_heap(0x%08lx, 0x%08lx)\n", base, top);

#if defined(OSMIPS) && !defined(HOST_TEST)
   base |= KSEG0_BASE;
   top  |= KSEG0_BASE;
#endif
   
#ifdef HOST_TEST
#  pragma BullseyeCoverage off

   if (NULL != test_init_boot_heap) {
   test_init_boot_heap(&boot_heap_data, &base, &top);
   }
   
#  pragma BullseyeCoverage on
#endif
   
   boot_heap_data.init_base = base;
   boot_heap_data.init_top  = top;
   boot_heap_data.curr_base = top+1;
   boot_heap_data.frag_list = NULL;
   boot_heap_data.active = 1;
   
   print_boot_heap();
}

/*****************************************************************************/

static void reserve_boot_heap_mem(void) {
   struct allocator *ram = gmm_state.ram_allocator;
   log_addr_t frag_base;
   log_size_t frag_size;
   struct fragment_container *f;
   
   if (boot_heap_data.auto_configured) {
      /* Reserve what we have used so far. */
      frag_base = ALIGN_DOWN(boot_heap_data.curr_base,
                             hal_get_info()->page_sizes[0]);
   }
   else {
      /* Reserve the entire region. */
      frag_base = ALIGN_DOWN(boot_heap_data.init_base,
                             hal_get_info()->page_sizes[0]);
   }
   
   frag_size = boot_heap_data.init_top + 1 - frag_base;
   
#ifdef OSMIPS
   frag_base &= ~KSEG0_BASE;
#endif
   
   if (MM_SUCCESS != ram->alloc_f(ram, &f, frag_base, frag_size))
   {
      MM_FATAL_ERROR2(("Failed to reserve physical memory for the boot "
                       "heap. [0x%08lx-0x%08lx]\n",
                       frag_base, frag_base + frag_size - 1),
                      OSE_MM_EINTERNAL_ERROR, (OSERRCODE) 0);
   }
   assert(boot_heap_data.frag_list == NULL);
   FC_SET_NEXT(f, NULL);
   boot_heap_data.frag_list = f;
   set_boot_heap_range(frag_base, frag_size);
}

/*****************************************************************************/

int
mm_alloc_boot_space(size_t size,
                    log_addr_t *out_base)
{
   log_addr_t base;
   
   if (!boot_heap_data.active) {
      *out_base = 0;
      return 0;
   }

   /* The boot heap grows downwards. */
   base = boot_heap_data.curr_base - size;

   /* Make sure that base is 64 bit aligned (Mips N32 requirement). */
   base &= ~(log_addr_t)7;
   
   /* (base > boot_heap_data.curr_base) corresponds to wrap 0 -> 0xf..f. */
   if (base < boot_heap_data.init_base || base > boot_heap_data.curr_base) {
      *out_base = 0;
      return 0;
   }
   
   boot_heap_data.curr_base = base;

   /* If the heap is explicitly configured, the physical memory has already
    * been allocated. */
   if (boot_heap_data.auto_configured) {
      static U32 recurse = 0;

   /* The boot heap grows down and any new physical fragments are inserted
       * at the beginning of the list (boot_heap_data.frag_list) => The lowest
       * address will be the base of the first fragment. */
      if (!recurse && (boot_heap_data.frag_list != NULL)
          && (base < FC_GET_BASE(boot_heap_data.frag_list)))
      {
         struct allocator *ram = gmm_state.ram_allocator;
         struct fragment_container *f;
         log_addr_t frag_base = ALIGN_DOWN(base, hal_get_info()->page_sizes[0]);
         log_size_t frag_size = FC_GET_BASE(boot_heap_data.frag_list)-frag_base;
         
#ifdef OSMIPS
         frag_base &= ~KSEG0_BASE;
#endif
         recurse = 1;
         if (MM_SUCCESS != ram->alloc_f(ram, &f, frag_base, frag_size))
         {
            print_boot_heap();
            MM_FATAL_ERROR2(("Failed to reserve physical memory for the boot "
                             "heap. [0x%08lx-0x%08lx]\n",
                             boot_heap_data.init_base, size),
                            OSE_MM_EINTERNAL_ERROR, (OSERRCODE) 0);
         }
         /* Insert new fragment at the beginning. See comment above. */
         FC_SET_NEXT(f, boot_heap_data.frag_list);
         boot_heap_data.frag_list = f;
         recurse = 0;
      }
   }
   
   *out_base = base;
   return 1;
}

/******************************************************************************
 *
 * This function extracts base, size and top for a memory area specified by
 * the parameter <param> in the configuration file.
 *
 * The configuration must contain enough information (base, size and top) to
 * uniquely and consistently define the area.
 * E.g:
 * + base:0x1000 size:0x1000 is enough to calculate the missing value for top.
 * + base:0x1000 size:0x1000 top:0x3fff is inconsistent.
 *
 * Base and size will probably cause no confusion, but it's a good idea to be
 * clear about top:
 *   !! top refers to the highest address in the area. !!
 *
 * Due to alignment restrictions, top normally ends with f..f, but for
 * convenience, the address immediately above top can be used instead.
 * E.g:
 *   top:0x3fff is the same as top:0x4000.
 * This allows the K, M & G prefixes to be used for top.
 * Note that this simplification is possible due to the fact that top:0x4000
 * interpreted exactly would be illegal due to alignment restrictions.
 */
static void readBaseSizeAndTop64(const char *param,
                                 phys_addr_t *base_,
                                 phys_size_t *size_,
                                 phys_addr_t *top_)
{
#define BASE_MASK 1UL
#define SIZE_MASK 2UL
#define TOP_MASK  4UL

   
   phys_addr_t base = 0, top = 0;
   phys_size_t size = 0;
   U32 tmp = 0;
   int rc;
   log_size_t min_alignment = hal_get_info()->page_sizes[0];

   /* Read attributes. */
   rc = sysparam_readattr_longlong(param, BASE_ATTRIBUTE_KEY,
                                   (long long *)&base);
   if (rc == EOK) {
      if (!ALIGNED(base, min_alignment)) {
         MM_FATAL_ERROR2(
            ("Attribute " BASE_ATTRIBUTE_KEY " (0x%llx) in parameter %s is "
	     "not properly aligned. Min alignment = 0x%lx.\n",
             base, param, min_alignment),
             OSE_MM_ECONFIGURATION_ERROR,
             (OSERRCODE)("Attribute " BASE_ATTRIBUTE_KEY " is not properly "
                         "aligned."));
      }
      tmp += BASE_MASK;
   }
   
   rc = sysparam_readattr_longlong(param, SIZE_ATTRIBUTE_KEY,
                                   (long long *)&size);
   if (rc == EOK) {
      if (!ALIGNED(size, min_alignment)) {
         MM_FATAL_ERROR2(
            ("Attribute " SIZE_ATTRIBUTE_KEY " (0x%llx) in parameter %s is "
             "not properly aligned. Min alignment = 0x%lx.\n",
             size, param, min_alignment),
            OSE_MM_ECONFIGURATION_ERROR,
            (OSERRCODE)("Attribute " SIZE_ATTRIBUTE_KEY " is not properly "
	                "aligned."));
      }
      tmp += SIZE_MASK;
   }
   
   rc = sysparam_readattr_longlong(param, TOP_ATTRIBUTE_KEY,
                                   (long long *)&top);
   if (rc == EOK) {
      /* It is nice to be able to write "top:1M" although highest address
       * really is 0xfffff.
       */
      if ((top & 1) == 0) top--;

      if (!ALIGNED((top+1), min_alignment)) {
         MM_FATAL_ERROR2(
            ("Attribute " TOP_ATTRIBUTE_KEY " (0x%llx) in parameter %s is not "
             "properly aligned. Min alignment (for top+1) = 0x%lx.\n",
             top, param, min_alignment),
            OSE_MM_ECONFIGURATION_ERROR,
            (OSERRCODE)("Attribute " TOP_ATTRIBUTE_KEY " is not properly "
                        "aligned."));
      }
      tmp += TOP_MASK;
   }

   switch (tmp) {

      case BASE_MASK + SIZE_MASK + TOP_MASK:
         if (base + size != top+(phys_addr_t)1) {
            MM_FATAL_ERROR2(
               ("Inconsistent data for \"%s\". Base = 0x%08llx, "
                "size = 0x%08llx, top = 0x%08llx\n",
                param, base, size, top),
               OSE_MM_ECONFIGURATION_ERROR,
               (OSERRCODE)("Inconsistent data combination for base, size "
			   "and top."));
         }
         break;

      case SIZE_MASK + TOP_MASK:
         base = top+(phys_addr_t)1-size;
         break;

      case BASE_MASK + TOP_MASK:
         size = top+(phys_addr_t)1-base;
         break;

      case BASE_MASK + SIZE_MASK:
         top = base + size - (phys_addr_t)1;
         break;

      default:
         MM_FATAL_ERROR2(
            ("Not enough information (base, size, top) to extract location"
             " for \"%s\".\n", param),
            OSE_MM_ECONFIGURATION_ERROR,
            (OSERRCODE)("Not enough information (base, size, top) to extract "
			"location."));
   }

   /* Check if base + size-1 wraps */
   if ((size > 0) && (top < base)) {
      MM_FATAL_ERROR2(
         ("%s: Top is less than base or base+size wraps.\n"
          "  base:0x%08llx size:0x%08llx, top = 0x%08llx\n",
          param, base, size, top),
         OSE_MM_ECONFIGURATION_ERROR,
         (OSERRCODE)("Top is less than base or base+size wraps."));
   }

   if (base_) *base_ = base;
   if (size_) *size_ = size;
   if (top_)  *top_ = top;

   VERBOZE_2(("\"%s\" base=0x%08llx size=0x%08llx top=0x%08llx\n",
              param, base, size, top));

#undef BASE_MASK
#undef SIZE_MASK
#undef TOP_MASK
}

/*****************************************************************************/
static void readBaseSizeAndTop32(const char *param,
                                 log_addr_t *base_,
                                 log_size_t *size_,
                                 log_addr_t *top_)
{
   phys_addr_t base64;
   phys_size_t size64;
   phys_addr_t top64;

   readBaseSizeAndTop64(param, &base64, &size64, &top64);
   
   if (base64 != (log_addr_t)base64) {
      MM_FATAL_ERROR2(
         ("Attribute base for parameter %s (0x%08llx) exceeds the range of "
          "32 bits\n", param, base64),
         OSE_MM_ECONFIGURATION_ERROR,
         (OSERRCODE)("Attribute base exceeds the range of 32 bits."));
   }

   if (size64 != (log_size_t)size64) {
      MM_FATAL_ERROR2(
         ("Attribute size for parameter %s (0x%08llx) exceeds the range of "
          "32 bits\n", param, size64),
         OSE_MM_ECONFIGURATION_ERROR,
         (OSERRCODE)("Attribute size exceeds the range of 32 bits."));
   }

   if (top64 != (log_addr_t)top64) {
      MM_FATAL_ERROR2(
         ("Attribute top for parameter %s (0x%08llx) exceeds the range of "
          "32 bits\n", param, top64),
         OSE_MM_ECONFIGURATION_ERROR,
         (OSERRCODE)("Attribute top exceeds the range of 32 bits."));
   }
   
   if (base_) {
      *base_ = (log_addr_t)base64;
   }
   if (size_) {
      *size_ = (log_size_t)size64;
   }
   if (top_) {
      *top_ = (log_addr_t)top64;
   }
}

/*****************************************************************************/

static void read_region_conf(const char *param, struct region_conf *reg) {
   U32 explicit_phys_base = 0;
   
   readBaseSizeAndTop32(param, &reg->log_base, &reg->size, &reg->log_top);
   
   /* Check phys_base attributes. */
   do {
      const char *attribute;
      mm_status_t rc;
      
      reg->phys_base        = reg->log_base;
      reg->reserve_phys_mem = TRUE;
      
      /* Check for unreserved phys_base.
       * Unreserved phys_base means that the physical memory area
       * by the (phys_base base address, size) is not marked as occupied
       * by MM.
       *
       * The only restriction is that phys_base may not be inside the physical
       * memory area covered by a SASE partition.
       */
      attribute = PHYS_BASE_ATTRIBUTE_KEY_U; /* phys_base_unreserved */
      rc = sysparam_readattr_longlong(param, attribute,
                                      (long long int *)&reg->phys_base);
      
      if (rc == EOK)
      {
         explicit_phys_base = 1;
         reg->reserve_phys_mem = FALSE;
         break;
      }
      if (rc != ENOENT)
      {
         ATTRIBUTE_ERROR(param, attribute, rc);
      }
      
      /* Try the old phys_base configuration. */
      attribute = PHYS_BASE_ATTRIBUTE_KEY_OLD; /* phys_base */
      rc = sysparam_readattr_longlong(param, attribute,
                                      (long long int *)&reg->phys_base);
      
      if (rc == EOK)
      {
         explicit_phys_base = 1;
         reg->reserve_phys_mem = FALSE;
         break;
      }
      if (rc != ENOENT)
      {
         ATTRIBUTE_ERROR(param, attribute, rc);
      }
      
      /* Resource_Info attribute - reserve_phys_mem */
      attribute = PHYS_BASE_ATTRIBUTE_KEY_R;
      rc = sysparam_readattr_longlong(param, attribute,
                                      (long long int *)&reg->phys_base);
      
      if (rc == EOK) {
         explicit_phys_base = 1;
      }
      else if (rc != ENOENT) {
         ATTRIBUTE_ERROR(param, attribute, rc);
      }
   } while (0);

#ifdef OSMIPS

   do {
      phys_addr_t expected_phys_base;
      
      if ((reg->log_base & KSEG_MASK) == KSEG0_BASE) {
         expected_phys_base = reg->log_base - KSEG0_BASE;
      }
      else if ((reg->log_base & KSEG_MASK) == KSEG1_BASE) { 
         expected_phys_base = reg->log_base - KSEG1_BASE;
      }
      else {
         break;
      }
      
      if (!explicit_phys_base) {
         reg->phys_base = expected_phys_base;
         break;
      }
      
      if (reg->phys_base != expected_phys_base) {
         MM_FATAL_ERROR2(("%s: Combination of logical base (0x%08lx) and "
                          "physical base (0x%08llx) is incompatible with the "
                          "MIPS architecture.\n",
                          param, reg->log_base, reg->phys_base),
                         OSE_MM_ECONFIGURATION_ERROR,
                         (OSERRCODE)
                         ("Combination of logical and physical "
			  "base is incompatible with the MIPS arch."));
      }
   } while (0);
   
#else

   /* Putting #ifdef at every occurence makes the code too noisy. */
   (void)explicit_phys_base;

#endif

   reg->phys_top = reg->phys_base + reg->size - 1;
}

/******************************************************************************
 * Filter for excluding the non resident mapped log mem.
 */

static int
include_resident(const char *param) 
{
   struct partition_class *
         p_class = mm_find_partition_by_name(param + strlen(LOGMEM_PREFIX));
   return (p_class->options & MM_PARTITION_OPT_RESIDENT);
}

/******************************************************************************
 * Filter for including only the PHYS_RAM_PREFIX krn/phys/RAM.
 */

static int
include_phys_ram(const char *param)
{
   return (strcmp(param, PHYS_RAM_PREFIX) == 0);
}

/******************************************************************************
 * Filter for excluding the SASE mapped log mem RAM_SASE/RAM.
 */

static int
exclude_ram_sase(const char *param) {
   return strcmp(param, ram_sase_param_name()) && !is_alias_log_mem(param);
}

/******************************************************************************
 * Filter for inclusion of implicit log mems only.
 */

static int
include_implicit_only(const char *param) {
   return is_implicit_log_mem(param);
}

/******************************************************************************
 * Filter for excluding alias and implicit log mem.
 */

static int
exclude_aliases_and_implicit(const char *param) {
   return !is_alias_log_mem(param) && !is_implicit_log_mem(param);
}

/******************************************************************************
 * Filter for inclusion of alias log mems only.
 */

static int
include_aliases_only(const char *param) {
   return is_alias_log_mem(param);
}

/******************************************************************************
 * Filter for excluding outer cache type configuration. 
 */

static int
exclude_outer(const char *param) {
   const char *name = param + strlen(CACHETYPE_PREFIX);
   const char *level_name = strchr(name, '/');
   if (level_name != NULL)
   {
      level_name++;
      if (!strcmp(level_name, "outer") ||
          !strcmp(level_name, "2"))
      {
         return 0;
      }
   }
   return 1;
}

/******************************************************************************
 * Filter for including the outer cache type configuration only. 
 */
static int
include_outer(const char *param) {
   const char *name = param + strlen(CACHETYPE_PREFIX);
   const char *level_name = strchr(name, '/');
   if (level_name != NULL)
   {
      level_name++;
      if (!strcmp(level_name, "outer") ||
          !strcmp(level_name, "2"))
      {
         return 1;
      }
   }
   return 0;
}

/*****************************************************************************
 * Search for the largest gap in the range specified by base & size. The
 * avoid_params lists the parameters that specifies areas that shall be
 * removed.
 * If a gap with a nonzero size is found, then  base and size are updated to
 * contain the largest gap and a nonzero value is returned.
 * Otherwise 0 is returned.
 *
 * The function first creates a list of all gaps and then picks the largest.
 * Since we do not have access to dynamically allocatable memory yet, the list
 * elements are allocated in a static arrary. If the number of elements is too
 * small, then an element from the gap list is used. The consequence is that if
 * the system contains many regions, then this function might not find the
 * largest.
 */

struct sysparam_selector {
   const char *param;
   int (*filter)(const char*);
};

/******************************************************************************
 * find_largest_gap
 *
 * This function tries to find the largest gap in the range specified by base
 * & size. On success, 1 is returned and the parameters base & size are
 * updated with the values for the found gap.
 * The gaps are all areas in the specified range that is not covered by a range
 * specified by any of the parameters in the avoid_params list.
 */

/* First some helpers */

struct log_gap {
   struct log_gap *next;
   log_addr_t base;
   log_addr_t top;
};

static struct log_gap *free_gap_list;

/**********/

static struct log_gap **
insert_new_gap_element(struct log_gap **insertion_point,
                       log_addr_t base, log_addr_t top)
{
   struct log_gap *new_gap;
   
   /* Allocate new element */
   new_gap       = free_gap_list;
   free_gap_list = free_gap_list->next;
   
   /* Update it */
   new_gap->base = base;
   new_gap->top  = top;
   
   /* Insert it at insertion point */
   new_gap->next    = *insertion_point;
   *insertion_point = new_gap;

   /* Return the insertion point to use for succeeding elements */
   return &new_gap->next;
}

/**********/

static struct log_gap *
unlink_smallest_gap(struct log_gap **list, const struct log_gap *curr)
{
   struct log_gap **rm = list, **tmp = &(*list)->next, *rv;
   log_size_t rm_size = (*rm)->top+1-(*rm)->base;
   
   while (*tmp)
   {
      /* Dont drop the element that we are operating on. */
      if (curr != *tmp)
      {
         log_size_t tmp_size = (*tmp)->top+1-(*tmp)->base;
                           
         if (tmp_size > rm_size)
         {
            rm = tmp;
            rm_size = tmp_size;
         }
      }
      tmp = &(*tmp)->next;
   }

   rv = *rm;
   *rm = rv->next;   /* Unlink rv */

   return rv;
}

/********** The main attraction! **********/
static U32
find_largest_gap(log_addr_t *base, log_size_t *size,
                 const struct sysparam_selector *avoid_params)
{
#define NR_OF_GAP_ELEMENTS 100

   static struct log_gap gap_elements[NR_OF_GAP_ELEMENTS];
   const struct sysparam_selector *avoid;
   
   U32 next_unused_elem = 1;
   struct log_gap *gap_list = &gap_elements[0];

   free_gap_list = NULL;

   if (0 == *size)
   {
      return 0;
   }
   
   gap_list->base = *base;
   gap_list->top  = *base+*size-1;
   gap_list->next = NULL;
   
   /* Init done. gap_list now contains the entire range (*base, *size). In the
    * next step memory belonging to the parameters in avoid_params is removed.
    */
   
   for (avoid = avoid_params; NULL != avoid->param; avoid++)
   {
      param_t param;

      /* Areas specified by the parameter avoid->param are considered as used
       *  => remove from the gap list.
       */
      MAP_SYSPARAM_FILTER(avoid->param, param, avoid->filter) {
         
         /* base & top of the area to be removed (rm) */
         phys_addr_t rm_base, rm_top;
         
         struct log_gap **curr_gap;
         
         readBaseSizeAndTop64(param, &rm_base, NULL, &rm_top);
         
         /* Check if param overlaps any of the gaps. */
         curr_gap = &gap_list;
         while (NULL != *curr_gap)
         {
            log_addr_t gap_base = (*curr_gap)->base;
            log_addr_t gap_top  = (*curr_gap)->top;
            
            if (rm_base > gap_top || rm_top < gap_base) {
               /* Area is entirely outside the current gap. */
               curr_gap = &(*curr_gap)->next;
               continue;
            }
            
            /* Region overlaps current gap. A tempting optimization might be to
             * exit the loop over the gap list, but at this early stage, it is
             * not yet ensured that no overlaps exist (if
             * krn/inhibit_mm_conf_check=yes). It is therefore possible for a
             * region etc. to overlap more then one gap.
             */
            
            {
               /* Ensure that there are at least two elements in free_gap_list.
                *
                * Two gap element will be needed if the region splits the gap
                * in two pieces. The current gap element can be reused => one
                * new element might be needed.
                */

               /* First make sure that free_gap_list contains at least one
                * element */
               if (NULL == free_gap_list)
               {
                  struct log_gap *new_elem;
                  
                  if (next_unused_elem < NR_OF_GAP_ELEMENTS) {
                     /* Use a previously unused element. */
                     new_elem = &gap_elements[next_unused_elem++];
                  }
                  else
                  {
                     /* (Rare case)
                      * We need to get a gap element. Drop the smallest. If
                      * this happens, then gap returned by find_largest_gap
                      * might not be the largest. Hopefully, it be enough
                      * anyway.
                      */
                     new_elem = unlink_smallest_gap(&gap_list, *curr_gap);
                  }
                  
                  /* Insert new_elem into free_gap_list. */
                  new_elem->next = free_gap_list;
                  free_gap_list = new_elem;
               }
               
               /* Move *curr_gap to free_gap_list. */
               {
                  struct log_gap *tmp = *curr_gap;
                  *curr_gap = (*curr_gap)->next;
                  tmp->next = free_gap_list;
                  free_gap_list = tmp;
               }
                  
               /* free_gap_list now contains at lest 2 elements. */
            }
            
            {
               /* Add remaining parts of the current gap to the gap list. */

               /* Is there a remaining piece of the gap below the region? */
               if (rm_base > gap_base)
               {
                  curr_gap = insert_new_gap_element(curr_gap,
                                                    gap_base,
                                                    rm_base-1);
               }
               
               /* Is there a remaining piece of the gap above the region? */
               if (rm_top < gap_top)
               {
                  curr_gap = insert_new_gap_element(curr_gap,
                                                    rm_top+1,
                                                    gap_top);
               }
            }
            
            if (NULL == gap_list) {
               /* This case would be caught later, but there is no use
                * in spinning over the remaining regions if we are sure
                * to fail. */
               return 0; /* <=> No gap found */
            }
         }
      }
   }
   
   /* Now gap_list holds a list of the gaps between the regions. Some gaps
    * might have been dropped, but the importang thing is that no memory in
    * the gaps are used by static regions. */
   
   if (NULL == gap_list) {
      return 0; /* <=> No gap found */
   }
   
   /* Pick the largest */
   {
      struct log_gap *gap = gap_list;
      log_addr_t best_base = (log_addr_t)-1;
      log_size_t best_size = 0;
      
      while (gap)
      {
         if (gap->top+1-gap->base > best_size)
         {
            best_base = gap->base;
            best_size = gap->top+1-gap->base;
         }
         gap = gap->next;
      }

      assert(0 != best_size);

      *base = best_base;
      *size = best_size;
      return 1;
   }
}

/*****************************************************************************/

static void
auto_configure_boot_heap_implicit_ram_sase(void)
{
   log_addr_t max_top;

   param_t physmem_param;
   log_addr_t best_base = 0;
   log_size_t best_size = 0;
   U32 gap_cnt  = 0;
   
#ifdef OSMIPS
   max_top = KSEG0_SIZE-1;
#else
   max_top = 0xffffffff;
#endif
   
   /* The "/" at the end of the parameter name makes  MAP_SYSPARAM match only
    * phys_mem/RAM, excluding other phys types, e.g. phys_mem/RAM-X.
    */
   MAP_SYSPARAM(PHYS_RAM_PREFIX "/", physmem_param) {

      /* The search criteria is:
       * Memory that is covered by phys_mem of type "RAM", but not by regions
       * or log_mems. Currently we no not allow any log_mem other than RAM_SASE
       * to cover physical memory, but that might change, so we include
       * krn/log_mem in the avoid list.
       */
      static const struct sysparam_selector avoid[]
         = {
            /* All regions and all log_mems except RAM_SASE. */
            { REGION_PREFIX, NULL             },
            { LOGMEM_PREFIX, exclude_ram_sase },
            
            { NULL, NULL }/* Terminator */
         };

      phys_addr_t bank_base64, bank_top64;
      log_addr_t base;
      log_size_t size;
      
      readBaseSizeAndTop64(physmem_param, &bank_base64, NULL, &bank_top64);
      
      if (max_top < bank_base64)
      {
         /* Bank is located entirely above the limit. */
         continue;
      }
      
      if (max_top < bank_top64)
      {
         /* Bank ends above the limit, but has some part below. */
         bank_top64 = max_top;
      }

      /* Now both bank_base64 & bank_top64 is <= max_top. */
      base = (log_addr_t)bank_base64;
      size = (log_size_t)(bank_top64+1-bank_base64);
      if (!find_largest_gap(&base, &size, avoid))
      {
         continue;
      }
      
      /* best_base is initiated to 0 and find_largest_gap should not return
       * true unless size > 0 => It is not necessary to check if best_base
       * & best_size are valid. */
      if (size > best_size)
      {
         best_base = base;
         best_size = size;
      }
      gap_cnt++;
   }
   
   if (0 == gap_cnt)
   {
      MM_FATAL_ERROR2(("No space for boot heap found. Try explicit"
                       " configuration of MM's meta data region:\n"
                       "  " REGION_PREFIX BOOT_REGION_NAME "="
                       BASE_ATTRIBUTE_KEY ":<...> "
                       SIZE_ATTRIBUTE_KEY ":<...>\n"
                         ),
                      OSE_MM_EINTERNAL_ERROR, 0);
      }
      
   VERBOZE(("Boot heap automatically configured. [0x%08lx-0x%08lx]\n",
            best_base, best_base+best_size-1));
   
   init_boot_heap(best_base, best_base+best_size-1);
   boot_heap_data.auto_configured  = 1;
}
   
/*****************************************************************************/

static void
auto_configure_boot_heap_explicit_ram_sase(void)
{
   /* The search criteria is:
    * Memory that is part of RAM_SASE, but not by regions.
    */
   static const struct sysparam_selector avoid[]
      = {
         { REGION_PREFIX, NULL }, /* All regions. */
         { NULL, NULL }           /* Terminator */
      };
   
   log_addr_t base;
   log_size_t size;
   
   readBaseSizeAndTop32(ram_sase_param_name(), &base, &size, NULL);
   
   if (!find_largest_gap(&base, &size, avoid))
   {
      MM_FATAL_ERROR2(("Error searching for boot heap"
                       " space. Try explicit configuration of MM's"
                       " meta data region:\n"
                       "  " REGION_PREFIX BOOT_REGION_NAME "="
                       BASE_ATTRIBUTE_KEY ":<...> "
                       SIZE_ATTRIBUTE_KEY ":<...>\n"
                         ),
                      OSE_MM_EINTERNAL_ERROR, 0);
   }
   
   VERBOZE(("Boot heap automatically configured. [0x%08lx-0x%08lx]\n",
            base, base+size-1));
   
   init_boot_heap(base, base+size-1);
   boot_heap_data.auto_configured  = 1;
}

/*****************************************************************************/

#ifdef OSPP
static void
fail_if_tlb1_threshold_not_supported(void)
{
   /* Check if the TLB1 threshold is configured. If so,
    * verify that we are running on e500 with extended physical
    * addressing. */

   long threshold_in_bytes;
   int rc = sysparam_readattr_long(
      "krn/mm/tlb1", "threshold", &threshold_in_bytes);
   if (rc == EOK)
   {
      if (hal_arch_cpu_descriptor->ppc_family != ppc_book_e ||
          !(hal_arch_cpu_descriptor->mm_attributes & MM_ATTRIBUTE_BOOK_E_E500) ||
          !(hal_arch_cpu_descriptor->mm_attributes & MM_ATTRIBUTE_BOOK_E_EXT_ADDR))
      {
         
         MM_FATAL_ERROR2(("TLB1 threshold is not supported on this CPU.\n"
                          "TLB1 threshold can only be configured on PowerPC e500 "
                          "processors with extended physical addressing enabled.\n"),
                         OSE_MM_ECONFIGURATION_ERROR,
                         (OSERRCODE)"TLB1 threshold not supported on CPU.");
      }
   }
}
#endif

/**************************************************************************/

static void fail_if_overlap2(const char *param_name1,
                             int (*filter1)(const char *),
                             const char *param_name2,
                             int (*filter2)(const char *),
                             const char *msg)
{
   param_t param1, param2;

   MAP_SYSPARAM_FILTER(param_name1, param1, filter1) {
      phys_addr_t base1, size1, top1;

      readBaseSizeAndTop64(param1, &base1, &size1, &top1);
      if(size1 == 0)
      {
        continue;
      }
      
      /* Check overlap */
      MAP_SYSPARAM_FILTER(param_name2, param2, filter2) {
         phys_addr_t base2, size2, top2;

         if (!strcmp(param1, param2))
         {
            continue;
         }

         readBaseSizeAndTop64(param2, &base2, &size2, &top2);

         if (top1 < base2 || top2 < base1)
            continue;

         if (NULL == msg)
         {
            msg = "Configuration overlap.";
         }
         
         MM_FATAL_ERROR2(
            ("Overlap.\n"
             "  %s [0x%08llx - 0x%08llx]\n"
             "  %s [0x%08llx - 0x%08llx]\n"
             "  %s\n",
             param1, base1, top1,
             param2, base2, top2,
             msg),
            OSE_MM_ECONFIGURATION_ERROR,
            (OSERRCODE)msg);
      }
   }
}

/*****************************************************************************/

static void fail_if_overlap(const char *param_name,
                            int (*exclude_func)(const char *),
                            const char *msg)
{
   fail_if_overlap2(param_name, exclude_func, param_name, exclude_func, msg);
}

/*****************************************************************************/

static void
fail_if_above_4G(const char *param_name,
                 int (*filter)(const char *))
{
   param_t param;
   
   MAP_SYSPARAM_FILTER(param_name, param, filter) {
      phys_addr_t base, top;

      readBaseSizeAndTop64(param, &base, NULL, &top);
      
      if (top >= 0x100000000ULL)
      {
         MM_FATAL_ERROR2(
            ("Range of %s, [0x%08llx - 0x%08llx], exceeds 4G.\n",
             param, base, top),
            OSE_MM_ECONFIGURATION_ERROR,
            (OSERRCODE)("Outside logical address space."));
      }
   }
}

/*****************************************************************************/

static void verify_configuration(void)
{
   Boolean implicit_ram_sase = False;
   param_t param;

   /* Filter function used together with MAP_SYSPARAM_FILTER to exclude
    * RAM_SASE in the implicit case. */
   int (*ram_sase_filter)(const char*) = NULL;
   
   if (0 < sysparam_read(PARAM_NAME_INHIBIT_MM_CONF_CHECK,
                         param, sizeof param))
   {
      
      /* The parameter was found. */
      
      if (!strcmp(param, "yes")) {
         ramlog_printf("Verification of MM configuration aborted. ("
                       PARAM_NAME_INHIBIT_MM_CONF_CHECK "=yes)\n");
         return;
      }
   }

#ifndef HOST_TEST
   {
      struct log_mem_param log_mem_param;
      /* Verify NO MMU RAM is set up to SASE at an early stage. */
      read_logmem_entry(ram_sase_param_name(), &log_mem_param);
   }
#endif
   
   {
      U32 hit = 0;
      MAP_SYSPARAM(PHYS_RAM_PREFIX "/", param) {
         hit = 1;
         break;
      }
      if (!hit) {
         MM_FATAL_ERROR2(
            ("No " PHYS_RAM_PREFIX "/* found.\n"),
            OSE_MM_ECONFIGURATION_ERROR,
            (OSERRCODE)("No " PHYS_RAM_PREFIX "/* found."));
      }
   }

   fail_if_overlap(PHYS_MEM_PREFIX, NULL, NULL);
   fail_if_overlap(REGION_PREFIX,   NULL, NULL);

   fail_if_above_4G(REGION_PREFIX,   NULL);

   /* Verify that multiple RAM_SASE banks haven't been configured. */
   {
      int sase_bank_count = 0;
      size_t sase_length = strlen(ram_sase_param_name());
      
      MAP_SYSPARAM(ram_sase_param_name(), param) {
         
         if (strncmp(ram_sase_param_name(), param, sase_length) == 0)
         {
            if (param[sase_length] == '\0' ||
                param[sase_length] == '/')
               sase_bank_count++;
         }
      }

      if (sase_bank_count > 1)
      {
         MM_FATAL_ERROR2(("log_mem '%s' cannot have multiple banks.\n",
                          ram_sase_param_name()),
                         OSE_MM_ECONFIGURATION_ERROR, (OSERRCODE)
                         "Multiple SASE log_mem banks.");
      }
   }
   
   /* Verify that alias log_mem:s don't specify a logical range */
   MAP_SYSPARAM_FILTER(LOGMEM_PREFIX, param, include_aliases_only) {
      long temp;
      int has_base, has_size, has_top;

      has_base = sysparam_readattr_long(param, BASE_ATTRIBUTE_KEY, &temp);
      has_size = sysparam_readattr_long(param, SIZE_ATTRIBUTE_KEY, &temp);
      has_top  = sysparam_readattr_long(param, TOP_ATTRIBUTE_KEY, &temp);

      if (has_base == EOK || has_size == EOK || has_top == EOK)
      {
         MM_FATAL_ERROR2(("Alias log_mem '%s' cannot specify its own"
                          " logical range.\n", param),
                         OSE_MM_ECONFIGURATION_ERROR,
                         (OSERRCODE)"Alias log_mem cannot specify a logical range.");
      }
   }


   /* This parameter is already read, and that could be used. However, one
    * purpose of this function is to allow creation of a host tool to verify
    * configurations and for that, it is good to keep it as independant as
    * possible.
    */
   if (EOK != sysparam_readattr_bool(ram_sase_param_name(),
                                     IMPLICIT_ATTRIBUTE_KEY,
                                     &implicit_ram_sase))
   {
      implicit_ram_sase = False;
   }
   
   if (implicit_ram_sase)
   {
      /* Implicit log_mem is OK. Check that there is some physical memory
       * in the range of the logical address space [0-4G]. */

      U32 hit_flag = 0;
      ram_sase_filter = exclude_ram_sase;
      
      MAP_SYSPARAM(PHYS_MEM_PREFIX, param) {
         phys_addr_t ph_base, ph_top;
         
         readBaseSizeAndTop64(param, &ph_base, NULL, &ph_top);
         if (ph_base >= 0x100000000ULL)
         {
            /* Memory above 4G can't be used for RAM_SASE. */
            continue;
         }
         
         /* The system has some memory below 4G. */
         hit_flag = 1;
         break;
      }
      if (hit_flag != 1) {
         MM_FATAL_ERROR2(
            ("No physical memory below 4G found."
             " (Needed for SASE mappings).\n"),
            OSE_MM_ECONFIGURATION_ERROR,
            (OSERRCODE)("No physical memory below 4G found."
                        " (Needed for SASE mappings)."));
      }
      fail_if_overlap2(LOGMEM_PREFIX,   exclude_ram_sase,
                       PHYS_MEM_PREFIX, NULL,
                       "When an implicit SASE mapped log_mem is used, no other"
                       " log_mems are allowed to overlap physical memory.");
   }
   else
   {
      /* Check that RAM_SASE is covered physical memory. */

      U32 hit_flag = 0;
      phys_addr_t sase_base64, sase_top64;
      log_addr_t sase_base, sase_top;
      
      ram_sase_filter = exclude_aliases_and_implicit;

      /* Use the 64 bit version together with a local check to get a better
       * diagnostic message if the range exceeds the 32 bit space. */
      readBaseSizeAndTop64(ram_sase_param_name(),
                           &sase_base64, NULL, &sase_top64);
      sase_top  = (log_addr_t)sase_top64;
      sase_base = (log_addr_t)sase_base64;

      if (sase_top != sase_top64) {
         MM_FATAL_ERROR2(
            ("Top of %s must not exceed 0x%08lx, found 0x%08llx.\n",
             ram_sase_param_name(), (log_addr_t)-1, sase_top64),
            OSE_MM_ECONFIGURATION_ERROR,
            (OSERRCODE)("Top of SASE configuration too high."));
      }

      MAP_SYSPARAM(PHYS_MEM_PREFIX, param) {
         phys_addr_t ph_base, ph_top;

         readBaseSizeAndTop64(param, &ph_base, NULL, &ph_top);
         if (sase_base > ph_top || ph_base > sase_top)
         {
            continue;
         }

         if (sase_base < ph_base || sase_top > ph_top) {
            MM_FATAL_ERROR2(
               ("SASE mapped RAM crosses phys mem border.\n"
                "  %s [0x%08lx-0x%08lx].\n"
                "  %s [0x%08llx-0x%08llx].\n",
                ram_sase_param_name(), sase_base, sase_top,
                param, ph_base, ph_top),
               OSE_MM_ECONFIGURATION_ERROR,
               (OSERRCODE)("SASE mapped RAM crosses phys mem border."));
         }

         /* Since phys_mem has already been checked for overlap, it is not
          * possible for RAM_SASE to be located in more than one
          * phys_mem. => There is no need to count matches, just set the
          * flag to 1. */
         hit_flag = 1;
      }
      if (hit_flag != 1) {
         MM_FATAL_ERROR2(
            ("SASE mapped RAM, [0x%08lx-0x%08lx], is not located within "
             "any physical memory.\n",
             sase_base, sase_top),
            OSE_MM_ECONFIGURATION_ERROR,
            (OSERRCODE)("SASE mapped RAM is not located within any physical"
                        " memory."));
      }
   }

   fail_if_above_4G(LOGMEM_PREFIX, ram_sase_filter);
   fail_if_overlap(LOGMEM_PREFIX,  ram_sase_filter, NULL);
   

   {
      param_t reg_param;
      MAP_SYSPARAM(REGION_PREFIX, reg_param) {
         phys_addr_t reg_base, reg_top;

         readBaseSizeAndTop64(reg_param, &reg_base, NULL, &reg_top);

         /* Check region location with respect to phys_mem */
         MAP_SYSPARAM_FILTER(LOGMEM_PREFIX, param, ram_sase_filter) {
            phys_addr_t phys_base, phys_top;

            readBaseSizeAndTop64(param, &phys_base, NULL, &phys_top);

            if (reg_base >= phys_base && reg_top <= phys_top) {
               /* Region is located entirely within the phys mem. */
               continue;
            }
            if (phys_base > reg_top || reg_base > phys_top) {
               /* Region is located entirely outside the phys mem. */
               continue;
            }

            MM_FATAL_ERROR2(
               ("Region crosses log mem border.\n"
                "  %s [0x%08llx-0x%08llx].\n"
                "  %s [0x%08llx-0x%08llx].\n",
                reg_param, reg_base, reg_top,
                param, phys_base, phys_top),
               OSE_MM_ECONFIGURATION_ERROR,
               (OSERRCODE)("Region crosses log mem border."));
         }
      }
   }

#ifdef OSPP
   /* Verify correct use of krn/mm/tlb1 */
   fail_if_tlb1_threshold_not_supported();
#endif
}

/*****************************************************************************
 * Create a complete region in the system domain.
 * Upon failure, print error information to ramlog.
 */
static struct region *
add_static_region(const char *param,
                  const char *name,
                  log_addr_t log_base,
                  log_size_t size,
                  U32 cache,
                  U32 permission,
                  phys_addr_t phys_base,
                  mm_bool_t reserve_phys_mem)
{
   mm_status_t status;
   struct region *region;
   struct log_object *lobj;
   struct phys_object *pobj;
   struct partition *partition;
   struct partition_alias *p_alias;
   const char *log_mem_name = "";
   U32 meta_options = 0;

   /***** Reserve logical resource. *****/
   {
      
      partition = mm_find_domain_partition_by_addr(
         gmm_state.system_domain, log_base);
      
      if (NULL == partition)
      {
         struct partition_class *p_class;

         /* Partition not found. Try to create one.
          * The new configuration will be instantiated in the system domain. */
         if (phys_base != (phys_addr_t)log_base)
         {
            p_class = mm_add_partition(MM_PARTITION_OPT_EXCLUSIVE|
                                       MM_PARTITION_OPT_VIRTUAL);
         }
         else
         {
            p_class = mm_add_partition(MM_PARTITION_OPT_EXCLUSIVE|
                                       MM_PARTITION_OPT_PHYSICAL);
         }
         if (p_class == NULL)
         {
            MM_FATAL_ERROR2(("no log_mem element for '%s'\n", param),
                            OSE_MM_ECONFIGURATION_ERROR, (OSERRCODE) param);
         }

         status = mm_add_partition_alias(p_class, name,
                                         MM_PARTITION_OPT_EXCLUSIVE |
                                         MM_PARTITION_OPT_PHYSICAL);
         if (status != MM_SUCCESS)
         {
            MM_FATAL_ERROR2(("no log_mem element for '%s'\n", param),
                            OSE_MM_ECONFIGURATION_ERROR, (OSERRCODE)
                            "Name collision between between log_mem:s.");
         }
         
         status = mm_add_partition_bank(p_class, log_base, size);
         if (status != MM_SUCCESS)
         {
            MM_FATAL_ERROR2(("no log_mem element for '%s'\n", param),
                            OSE_MM_ECONFIGURATION_ERROR, (OSERRCODE)
                            "Partial overlap between region and log_mem.");
         }

         ramlog_printf("mm: log_mem  [0x%08lx-0x%08lx]     %4s %-32s\n",
                       log_base, log_base+size-1, "SASE", name);

         /* Created. Get the instantiated partition from system domain. */
         partition = mm_find_domain_partition_by_addr(
            gmm_state.system_domain, log_base);
         if (partition == NULL)
         {
            MM_FATAL_ERROR2(("no log_mem element for '%s'\n", param),
                            OSE_MM_ECONFIGURATION_ERROR, (OSERRCODE) param);
         }
      }

      /* Lookup default alias for the partition and make sure it has
       * either SAS or SASE mapping type. */
      p_alias = mm_find_partition_alias(partition->class, NULL);
      if (p_alias == NULL)
      {
         MM_FATAL_ERROR2(("No log_mem element for '%s'\n", param),
                         OSE_MM_ECONFIGURATION_ERROR, (OSERRCODE) param);
      }

      switch (p_alias->options & MM_PARTITION_MAPTYPE_MASK)
      {
         case MM_PARTITION_OPT_SASE:
            break;
         case MM_PARTITION_OPT_SAS:
         {
            if (!partition_is_combined(partition->class))
            {
               break;
            }
         }
         default:
            MM_FATAL_ERROR2(("log_mem element '%s' has illegal mapping type for \n"
                             "    static region '%s'\n",
                             p_alias->name, param),
                            OSE_MM_ECONFIGURATION_ERROR, (OSERRCODE)
                            "Illegal mapping type for static region.");
      }

      log_mem_name = p_alias->name;
      
      /* Partition ok. */
      
      if (!partition->class->is_superset_of(partition->class, log_base, size))
      {
         MM_FATAL_ERROR2(("'%s', [0x%08lx-0x%08lx], only partially inside "
                          "a log_mem (%s).\n",
                          param, log_base, log_base + size - 1, name),
                         OSE_MM_ECONFIGURATION_ERROR, (OSERRCODE)
                         "Partial overlap between region and log_mem.");
      }
      
      /* Create logical memory container */
      lobj = mm_create_log_object();

      /* Allocate the logical memory exclusively (no MAS) */
      status = partition->class->alloc_f(
         partition, lobj, size, log_base,
         MM_PARTITION_OPT_EXCLUSIVE|MM_PARTITION_OPT_PHYSICAL);

      if (status != MM_SUCCESS)
      {
         MM_FATAL_ERROR2(
            ("Failed to create region '%s'.\n", name),
            OSE_MM_ECONFIGURATION_ERROR,
            (OSERRCODE)"Allocation of logical memory failed.");
      }
   } /* Logical resource OK. */


   /* Mark logical memory as used */
   mm_attach_log_object(lobj, gmm_state.system_domain);


   /***** Read static option. *****/
   {
      int rc;
      long val;
      static const struct enumeration enum_type[] = {
         {"static",   MM_REGION_OPT_INITIAL_STATIC},
         {"dynamic",  MM_REGION_OPT_INITIAL_DYNAMIC}};
      
      rc = sysparam_readattr_enum(param, "type", enum_type, &val);
      
      if (rc == EOK)
      {
         ramlog_printf("Detected a region in system parameters which "
                       " explicitly defines type:\n '%s' type %lx\n",
                       param, val);

         meta_options |= val;
      }
      else if (rc == EINVAL)
      {
         ATTRIBUTE_ERROR(param, "static", rc);
      }
      else
      {
         meta_options |= MM_REGION_OPT_INITIAL_STATIC;
         /* Default: conf regions are static. */
      }
   }

   /***** Reserve physical resource. *****/
   {
      struct allocator *allocator = NULL;

      if (reserve_phys_mem) {
         allocator = mm_find_allocator_by_range(phys_base, (size_t)size);
         /* It's ok if mm_find_allocator_by_range returns NULL. */
      }
      
      pobj = mm_create_phys_object(allocator, meta_options);
      if (pobj == NULL)
      {
         /* No need to clean up since this is a fatal error. */
         MM_FATAL_ERROR2(("Failed to allocate meta data for physical "
                          "resource. (%s)\n",
                          param), OSE_MM_ECONFIGURATION_ERROR,
                         (OSERRCODE) param);
      }
      
      /*
       * Two cases exist here.
       * Either an allocator was found that manages the area (<base>, <size>).
       * Then <size> bytes must be allocated from that allocator in order to
       * mark it as <allocated>.
       *
       * Otherwise no allocator was found. This is permitted and the phys_mem
       * object has to be created manually.
       */
      if (allocator != NULL)
      {
         pobj->data_representation = PHYS_OBJ_FRAG_LIST;
         /* An allocator was found. Allocate <size> bytes of memory from it. */
         status = allocator->alloc_f(allocator,
                                     &pobj->data.frag_list,
                                     phys_base,
                                     size);
         if (status != MM_SUCCESS)
         {
            /* No need to clean up since this is a fatal error. */
            MM_FATAL_ERROR2(("Allocation of physical memory failed for '%s'\n",
                             param), OSE_MM_ECONFIGURATION_ERROR, (OSERRCODE)
                            "Allocation of physical memory failed.");
         }
      }
      else
      {
         pobj->data_representation = PHYS_OBJ_DIRECT_PHYS_FRAG;
         pobj->data.direct_phys_frag.pa = phys_base;
         pobj->data.direct_phys_frag.size = size;
      }
   } /* Physical resource OK. */   
   
   mm_attach_phys_object(pobj);

   /*****  Tie it all together in a region. *****/
   status = mm_add_region(
      MM_KERNEL_DOMAIN,
      lobj,
      pobj,
      permission,
      cache,
      0,
      meta_options,
      name,
      log_mem_name,
      &region);

   if (status != MM_SUCCESS)
   {
      switch (status)
      {
         case MM_EILLEGAL_PERMISSION:
            MM_FATAL_ERROR2(
               ("Illegal permissions (%s) for region '%s'.\n",
                num_to_str(permission_encoding, permission), name),
               OSE_MM_ECONFIGURATION_ERROR,
               (OSERRCODE)"Failed to create region (permissions).");
            break; /*lint !e527 Unreachable */
               
         case MM_EILLEGAL_CACHE_MODE:
            /* The cache value is made up of 3 bits for cache mode and 2
             * bits for access type. FIXME. This should not be hard coded.
             */
            MM_FATAL_ERROR2(
               ("Illegal cache mode (%s) or access type (%s) for "
                "region '%s'.\n",
                num_to_str(cache_encoding, cache & 0x7),
                num_to_str(mem_encoding, cache & (0x3 << 3)),
                name),
               OSE_MM_ECONFIGURATION_ERROR,
               (OSERRCODE)"Failed to create region (cache mode).");
            break; /*lint !e527 Unreachable */
               
         case MM_EOUTSIDE_LOG_MEM:
            MM_FATAL_ERROR2(
               ("Region '%s' [0x%08lx-0x%08lx] is not located in "
                "free log_mem.\n",
                name, log_base, log_base+size-1),
               OSE_MM_ECONFIGURATION_ERROR,
               (OSERRCODE) "Failed to create region (outside log_mem).");
            break; /*lint !e527 Unreachable */
               
         default:
            MM_FATAL_ERROR2(
               ("Failed to create region '%s'.\n", name),
               OSE_MM_ECONFIGURATION_ERROR,
               (OSERRCODE)"Failed to create region (generic).");
      }
   }
   
   if (log_base == phys_base) {
      VERBOZE(("%s: [0x%08lx-0x%08lx]\n", param, log_base, log_base+size-1));
   }
   else {
      VERBOZE(("%s: [0x%08lx-0x%08lx] phys_base = 0x%08llx.\n",
               param, log_base, log_base + size - 1, phys_base));
   }

   return region;
}



/*
 * Determine whether the mapping_type is supported or not.
 */
static mm_bool_t
mapping_type_supported(mapping_type_t mapping_type)
{
   switch (mapping_type)
   {
      case SASE: return TRUE;
      case SAS: return hal_get_info()->has_mmu;
      case MAS: return hal_get_info()->has_mas;
      default: return FALSE;
   }
}


/*
 * Create a log_mem entry from a system parameter.
 */
static void
read_logmem_entry(const char *param, struct log_mem_param *data)
{
   VERBOZE(("mm: start parsing log_mem string: %s\n", param));

   data->name = param + strlen(LOGMEM_PREFIX); /*lint !e613 */

   if (EOK != sysparam_readattr_string(param,
                                       ALIAS_ATTRIBUTE_KEY,
                                       data->alias, MM_NAME_LENGTH))
   {
      data->alias[0] = '\0';
   }
   
   if (EOK != sysparam_readattr_bool(param,
                                     IMPLICIT_ATTRIBUTE_KEY,
                                     &data->implicit))
   {
      data->implicit = False;
   }
   
   REQUIRE_ATTR_ENUM_MAP_TYPE(data->mapping_type, 
                              param, MAPPING_ATTRIBUTE_KEY,
                              mapping_type_encoding);

   /* Check that the mapping type is supported. */
   if (!mapping_type_supported(data->mapping_type))
   {
      MM_FATAL_ERROR2(("Unsupported mapping type '%s' for this architecture "\
                       "in log_mem entry '%s' (MMU=%d)\n",
                       num_to_str(mapping_type_encoding,
                                  (U32)data->mapping_type),
                       data->name, HAS_MMU),
                      OSE_MM_ECONFIGURATION_ERROR, (OSERRCODE)
                      "Unsupported mapping type for this architecture.");
   }

   /* Read manager mode */
   {
      int rc = sysparam_readattr_bool(param, MANAGER_MODE_ATTRIBUTE_KEY,
                                      (Boolean*)&data->manager_mode);
      if (rc == ENOENT)
      {
         data->manager_mode = FALSE;
      }
      else if (rc != EOK)
      {
         ATTRIBUTE_ERROR(param, MANAGER_MODE_ATTRIBUTE_KEY, rc);
      }
   }
   
#ifndef OSARM
   if (data->manager_mode)
   {
      MM_FATAL_ERROR2(("manager_mode is not supported by this architecture.\n"),
                      OSE_MM_ECONFIGURATION_ERROR, (OSERRCODE)
                      "manager_mode not supported.");
   }
#endif
   
   if (data->implicit)
   {
      return;
   }

   /* Range */
   data->base = INVALID_LOG_MEM_BASE;
   data->size = INVALID_LOG_MEM_SIZE;

   if (!is_alias_log_mem(param))
   {
      readBaseSizeAndTop32(param, &data->base, &data->size, NULL);
   }
}

static void
create_logmem_entry(struct log_mem_param *log_mem_param,
                    U32 implicit_ram_sase)
{
   struct partition_class *new_part_class;
   mm_status_t status;
   U32 options = 0;
   char name[MM_NAME_LENGTH];
   char *end_of_name = strchr(log_mem_param->name, '/');

   /* Pointer to the first SASE mapped partition that overlaps physical ram.
    * Used to detect if there are more than one, and in that case provides
    * error information. */
   static struct partition_class *ram_sase = NULL;


   /* Copy name so we can modify it */
   strlcpy(name, log_mem_param->name, MM_NAME_LENGTH);

   /* Check if the log_mem specifies a bank */
   if (end_of_name != NULL)
   {
      U32 index = (U32)end_of_name - (U32)log_mem_param->name;
      name[index] = '\0';
   }
   
   /* Refuse to create a partition with size 0 */
   if (!log_mem_param->implicit && log_mem_param->size == 0)
   {
      ramlog_printf("mm: rejected partition %s. size = 0.\n", log_mem_param->name);
      return;
   }

   switch (log_mem_param->mapping_type)
   {
      case MAS:
         options |= MM_PARTITION_OPT_NON_EXCLUSIVE;
         options |= MM_PARTITION_OPT_VIRTUAL;
         break;
      case SAS:
         options |= MM_PARTITION_OPT_EXCLUSIVE;
         options |= MM_PARTITION_OPT_VIRTUAL;
         break;
      case SASE:
         options |= MM_PARTITION_OPT_EXCLUSIVE;
         options |= MM_PARTITION_OPT_PHYSICAL;
         break;
      default:
         MM_FATAL_ERROR2(("log_mem %s has an illegal mapping type.\n", name),
                         OSE_MM_ECONFIGURATION_ERROR,
                         (OSERRCODE)log_mem_param);
   }

   /* If the log_mem name includes a bank specifier then check
    * if the partition is already created and simply attach a
    * new bank to it */
   if (end_of_name != NULL)
   {
      struct partition_class *p_class = mm_find_partition_by_name(name);

      /* Log_mem aliases cannot be banked */
      if (log_mem_param->alias[0] != '\0')
      {
         MM_FATAL_ERROR2(("Failed to create log_mem '%s'. "
                          "Alias log_mem:s cannot have banks\n", name),
                         OSE_MM_ECONFIGURATION_ERROR,
                         (OSERRCODE)"Alias log_mem:s cannot have banks.");
      }

  
      /* Add a new partition bank to existing log_mem */
      if (p_class != NULL)
      {
         /* Bank must specify same mapping type as existing log_mem banks */
         if ((options & MM_PARTITION_MAPTYPE_MASK) !=
             (p_class->options & MM_PARTITION_MAPTYPE_MASK))
         {
            MM_FATAL_ERROR2(("Failed to create bank for log_mem '%s'. "
                             "Mapping types mismatch.\n", name),
                            OSE_MM_ECONFIGURATION_ERROR,
                            (OSERRCODE)"log_mem bank mapping type mismatch.");
         }

         /* Banks must not be configured for SASE log_mem:s */
         if ((p_class->options & MM_PARTITION_MAPTYPE_MASK) ==
             MM_PARTITION_OPT_SASE)
         {
            MM_FATAL_ERROR2(("Failed to create bank for log_mem '%s'.\n"
                             "A SASE log_mem cannot have multiple banks.\n", name),
                            OSE_MM_ECONFIGURATION_ERROR,
                            (OSERRCODE)"Multiple SASE log_mem banks.");
         }
         
         status = mm_add_partition_bank(
            p_class, log_mem_param->base, log_mem_param->size);
         if (status != MM_SUCCESS)
         {
            MM_FATAL_ERROR2(("Failed to create new bank for log_mem '%s'."
                             "Possible overlap in the logical space. \n", name),
                            OSE_MM_ECONFIGURATION_ERROR,
                            (OSERRCODE)"Failed to create log_mem bank.");
         }

         ramlog_printf("mm: log_mem  [0x%08lx-0x%08lx]     %4s %s\n",
                       log_mem_param->base,
                       log_mem_param->base + log_mem_param->size - 1,
                       num_to_str(mapping_type_encoding,
                                  log_mem_param->mapping_type),
                       name);
         return;
      }
   }

   if (log_mem_param->implicit)
   {
      const char *ram_sase_name = HAS_MMU
         ? RAM_SASE_NAME
         : RAM_NAME;
      
      if (strcmp(ram_sase_name, name))
      {
         MM_FATAL_ERROR2(("Attempt to create implicit log_mem %s."
                          " Only %s is allowed to be implicit.\n",
                          name, ram_sase_name),
                         OSE_MM_ECONFIGURATION_ERROR,
                         (OSERRCODE)"Illegal use of implicit log_mem.");
      }

      if (SASE != log_mem_param->mapping_type)
      {
         MM_FATAL_ERROR2(("Mapping type of implicit log_mem must be SASE."
                          " (%s)\n", name),
                         OSE_MM_ECONFIGURATION_ERROR,
                         (OSERRCODE)"Mapping type of implicit log_mem must be SASE.");
      }

      new_part_class = mm_add_implicit_sase_partition(ram_sase_name);
      if (new_part_class == NULL)
      {
         MM_FATAL_ERROR2(("Failed to create log_mem '%s'\n", name),
                         OSE_MM_ECONFIGURATION_ERROR,
                         (OSERRCODE)log_mem_param);
      }

      status = mm_add_partition_alias(new_part_class, name, options);
      if (status != MM_SUCCESS)
      {
         MM_FATAL_ERROR2(("Failed to create log_mem '%s'.\n", name),
                         OSE_MM_ECONFIGURATION_ERROR,
                         (OSERRCODE)log_mem_param);
      }

      ramlog_printf("mm: log_mem  [      <implicit>     ]     %4s %-32s %s\n",
                    num_to_str(mapping_type_encoding,
                               log_mem_param->mapping_type),
                    name,
                    log_mem_param->manager_mode ? "manager_mode:yes": "");
   }
   else
   {
      if (implicit_ram_sase && log_mem_param->alias[0] == '\0')
      {
         /* When implicit RAM_SASE is used, then not other log_mems are
          * allowed to overlap with the physical memory.
          */
         struct allocator *pa
            = mm_find_allocator_by_range(log_mem_param->base,
                                         log_mem_param->size);
         if (NULL != pa)
         {
            MM_FATAL_ERROR2(("log_mem %s overlaps physical allocator %s. This"
                             " is not allowed for a SAS/MAS mapped log_mem.\n",
                             name, pa->name),
                            OSE_MM_ECONFIGURATION_ERROR,
                            (OSERRCODE)"log_mem overlaps phys_mem.");
         }
      }
      
#if 0
      ramlog_printf("mm: log_mem %#010lx->%#010lx:%s type:%s %s.\n",
                    log_mem_param->base,
                    log_mem_param->size,
                    name,
                    log_mem_param->mapping_type,
                    log_mem_param->manager_mode ? "manager_mode:yes": "");
#endif

      if (log_mem_param->manager_mode)
      {
         options |= MM_PARTITION_OPT_MANAGER_MODE;
      }
      
      if (log_mem_param->alias[0] != '\0')
      {
         struct partition_alias *p_alias;
         struct partition_class *p_class =
            mm_find_partition_by_name(log_mem_param->alias);
         if (p_class == NULL)
         {
            MM_FATAL_ERROR2(("Failed to create alias log_mem %s from "
                             "non-existing log_mem %s.\n",
                             name, log_mem_param->alias),
                            OSE_MM_ECONFIGURATION_ERROR,
                            (OSERRCODE)"Cannot create alias log_mem "
                            "from non-existing log_mem.");
         }

         p_alias = mm_find_partition_alias(p_class, log_mem_param->alias);
         if (p_alias == NULL)
         {
            MM_FATAL_ERROR2(("Failed to create alias log_mem %s from "
                             "non-existing log_mem %s.\n",
                             name, log_mem_param->alias),
                            OSE_MM_ECONFIGURATION_ERROR,
                            (OSERRCODE)"Cannot create alias log_mem "
                            "from non-existing log_mem.");
         }

         /* Verify that the existing partition can be used to
          * create the new alias:
          *  - Must not be another alias
          */
         if (!(p_alias->options & MM_PARTITION_OPT_DEFAULT))
         {
            MM_FATAL_ERROR2(("Cannot create alias log_mem '%s' from "
                             "existing alias log_mem '%s'.\n",
                             name, log_mem_param->alias),
                            OSE_MM_ECONFIGURATION_ERROR,
                            (OSERRCODE)"Cannot create alias log_mem "
                            "from another alias log_mem.");
         }

         /*  - Must not be implicit */
         if (partition_is_implicit(p_class))
         {
            MM_FATAL_ERROR2(("Cannot create alias log_mem '%s' from "
                             "implicit log_mem '%s'.\n",
                             name, log_mem_param->alias),
                            OSE_MM_ECONFIGURATION_ERROR,
                            (OSERRCODE)"Cannot create alias log_mem "
                            "from implicit log_mem.");
         }

         /*  - Must not be SASE */
         if ((p_class->options & MM_PARTITION_OPT_SASE) == MM_PARTITION_OPT_SASE)
         {
            MM_FATAL_ERROR2(("Cannot create alias log_mem '%s' from "
                             "SASE-mapped log_mem '%s'.\n",
                             name, log_mem_param->alias),
                            OSE_MM_ECONFIGURATION_ERROR,
                            (OSERRCODE)"Cannot create alias log_mem "
                            "from SASE-mapped log_mem.");
         }

         
         /* New alias cannot be SASE, only SAS and MAS aliases are allowed */
         if (log_mem_param->mapping_type != MAS &&
             log_mem_param->mapping_type != SAS)
         {
            MM_FATAL_ERROR2(("Cannot create alias log_mem '%s' with "
                             "SASE mapping type.\n", name),
                            OSE_MM_ECONFIGURATION_ERROR,
                            (OSERRCODE)"Cannot create alias log_mem "
                            "with SASE mapping type.");
         }

#if 0 /* Already verified by verify_configuration */
         /* Base and size must not be specified */
         if (log_mem_param->base != INVALID_LOG_MEM_BASE ||
             log_mem_param->size != INVALID_LOG_MEM_SIZE)
         {
            MM_FATAL_ERROR2(("Logical range specified for alias log_mem '%s'\n",
                             name),
                            OSE_MM_ECONFIGURATION_ERROR,
                            (OSERRCODE)log_mem_param);
         }
#endif
         
         new_part_class = p_class;
         
         status = mm_add_partition_alias(
            new_part_class, name, options);
         if (status != MM_SUCCESS)
         {
            MM_FATAL_ERROR2(("Failed to create alias log_mem '%s'\n", name),
                            OSE_MM_ECONFIGURATION_ERROR,
                            (OSERRCODE)log_mem_param);
         }

         ramlog_printf("mm: log_mem  [       <alias>       ]     %4s %s %s\n",
                       num_to_str(mapping_type_encoding,
                                  log_mem_param->mapping_type),
                       name,
                       log_mem_param->manager_mode ? "manager_mode:yes": "");
         ramlog_printf("mm: log_mem                                   %s\n",
                       log_mem_param->alias);
      }
      else
      {
         new_part_class = mm_add_partition(options);
         if (new_part_class == NULL)
         {
            MM_FATAL_ERROR2(("Failed to create log_mem '%s'\n", name),
                            OSE_MM_ECONFIGURATION_ERROR,
                            (OSERRCODE)log_mem_param);
         }

         status = mm_add_partition_alias(new_part_class, name, options);
         if (status != MM_SUCCESS)
         {
            MM_FATAL_ERROR2(("Failed to create log_mem '%s'\n", name),
                            OSE_MM_ECONFIGURATION_ERROR,
                            (OSERRCODE)log_mem_param);
         }

         status = mm_add_partition_bank(
            new_part_class, log_mem_param->base, log_mem_param->size);
         if (status != MM_SUCCESS)
         {
            MM_FATAL_ERROR2(("Failed to create bank for log_mem '%s'. "
                             "Possible overlap in logical space.\n", name),
                            OSE_MM_ECONFIGURATION_ERROR,
                            (OSERRCODE)"Failed to create log_mem bank.");
         }

         ramlog_printf("mm: log_mem  [0x%08lx-0x%08lx]     %4s %s %s\n",
                       log_mem_param->base,
                       log_mem_param->base + log_mem_param->size - 1,
                       num_to_str(mapping_type_encoding,
                                  log_mem_param->mapping_type),
                       name,
                       log_mem_param->manager_mode ? "manager_mode:yes": "");
      }
      
   }
   
   /* Check that only one SASE logmem is created inside RAM. */
   if (log_mem_param->mapping_type == SASE &&
       gmm_state.ram_allocator->sub_area(gmm_state.ram_allocator,
                                         log_mem_param->base,
                                         log_mem_param->size))
   {
      if (NULL != ram_sase)
      {       
         MM_FATAL_ERROR2(("Only one log_mem entry with mapping type "
                          "SASE is allowed inside RAM. Found %s & %s.\n",
                          new_part_class->name, ram_sase->name),
                         OSE_MM_ECONFIGURATION_ERROR,
                         (OSERRCODE)"Only one log_mem per phys_mem.");
      }
      ram_sase = new_part_class;
   }
}


/*
 * Process a buffer with character encoded data for a region.
 * Create a system region.
 */
static void
configure_region(const char *param)
{
   const char *name;
   struct region_conf reg;
   long perm;
   char cache_type[MM_NAME_LENGTH+1];
   char mem_type[MM_NAME_LENGTH+1];
   U32 cache_mode;

   assert(param != NULL);

   name = strrchr(param, '/') + 1; /*lint !e613 */

   /* Read attributes. */
   read_region_conf(param, &reg);

   /* Refuse to create a region with size 0 */
   if(reg.size == 0)
   {
      ramlog_printf("mm: rejected region %s. size = 0.\n", param);
      return;
   }         
     
   REQUIRE_ATTR_ENUM(perm, param, PERMISSION_ATTRIBUTE_KEY,
                     permission_encoding);

   {
      hal_status_t status;

      REQUIRE_ATTR_STRING(cache_type, MM_NAME_LENGTH,
                          param, CACHE_ATTRIBUTE_KEY);
      REQUIRE_ATTR_STRING(mem_type, MM_NAME_LENGTH,
                          param, MEMORY_ATTRIBUTE_KEY);

      status = hal_get_cache_mode(cache_type, mem_type, &cache_mode);
      if (status != HAL_SUCCESS)
      {
         MM_FATAL_ERROR2(("Illegal cache and memory type combination detected "
                          "for static region %s\n", param),
                         OSE_MM_ECONFIGURATION_ERROR, (OSERRCODE)
                         "Illegal cache and memory type combination detected.");
      }
   }

   if (!HAS_MMU && reg.phys_base != reg.log_base)
   {
      MM_FATAL_ERROR2(("Illegal 'phys_base' attribute for static region '%s'\n"
                       "The physical address is not allowed to differ from "
                       "the logical address in a system without MMU HAL "
                       "support.\n", param),
                      OSE_MM_ECONFIGURATION_ERROR, (OSERRCODE)
                      "Illegal 'phys_base' attribute for region (nommu).");
   }
   
   if (!reg.reserve_phys_mem)
   {
      /* We have a phys_base, indicating a piece of physical memory
       * that MUST NOT exist in any logical memory.
       */
      
      /* Check that (phys_base,size) does not overlap with any SASE logmem
       * area in ram. This is to prevent double mappinging in ram.
       * For example on arm a part of ram has to be mapped to vector area
       * in the configuration using physbase. Usually address 0-4k of ram.
       * Unless this check is made that physical area will be
       * be possible to cover with a sase logmem and the same address 0-4k area
       * might later be mapped to a new logical region sase. When
       * this happens the creator is free to write to this new logical region
       * and as consequence to destroy the vector area.
       */
      if (mm_find_allocator_by_range(reg.phys_base, reg.size))
      {
         ramlog_printf("mm: Warning: Physical range (unreserved) for %s, "
                       "[0x%08llx-0x%08llx]  overlaps physical memory handled "
                       "by MM. Data may be damaged if the memory is allocated "
                       "by an application.\n",
                       param, reg.phys_base, reg.phys_base + reg.size - 1);
      }
   }

   if (reg.phys_base != reg.log_base)
   {
      struct partition_class *p;
      struct allocator *a = mm_find_allocator_by_range(reg.log_base, reg.size);
      if (a != NULL)
      {
         /* Phys_mem backed logical range. Potentially dangerous if
          * there are explicitly configured SASE log_mem types for
          * the range as well. */
            
         p = mm_find_sase_partition_by_range(reg.log_base, reg.size);
         if (p != NULL)
         {
            MM_FATAL_ERROR2(
               ("Logical range of static region %s [0x%08lx-0x%08lx] "
                "covers unallocated physical memory that may be used "
                "to create SASE type regions. Run-time SASE allocations "
                "from phys_mem %s will fail.\n",
                param,
                reg.log_base,
                reg.log_base + reg.size - 1,
                a->name),
               OSE_MM_ECONFIGURATION_ERROR, (OSERRCODE)
               "phys_based region hides physical SASE resources.");
         }
      }
   }
   
   
   VERBOZE(("mm: region \"%s\": [0x%08lx-0x%08lx], (0x%08lx) %s %s %s\n",
            name, reg.log_base, reg.log_base + reg.size -1, reg.size,
            num_to_str(permission_encoding, perm),
            cache_type, mem_type));
   
   /* FIXME. Move this check to ramlog.c */
#ifdef OSMIPS
   if (!strcmp(name, "ramlog"))
   {
      if (!((SUB_AREA(reg.log_base, reg.size, KSEG0_BASE, KSEG0_SIZE)) ||
            (SUB_AREA(reg.log_base, reg.size, KSEG1_BASE, KSEG1_SIZE))))
      {
         MM_FATAL_ERROR2(("region '%s' [%#010llx, %#010llx] "\
                          "must be located in KSEG0 [%#010lx, %#010lx] "\
                          "or KSEG1 [%#010lx, %#010lx]\n",
                          name,
                          reg.log_base, reg.log_base + reg.size,
                          KSEG0_BASE, KSEG0_BASE + KSEG0_SIZE - 1,
                          KSEG1_BASE, KSEG1_BASE + KSEG1_SIZE - 1),
                         OSE_MM_ECONFIGURATION_ERROR, (OSERRCODE)
                         "Ramlog must be placed within KSEG0/1.");
      }
   }

   if ((SUB_AREA(reg.log_base, reg.size, KSEG0_BASE, KSEG0_SIZE) ||
        SUB_AREA(reg.log_base, reg.size, KSEG1_BASE, KSEG1_SIZE)) &&
       (perm & (MM_USR_READ|MM_USR_EXECUTE|MM_USR_WRITE)) != 0)
   {
      MM_FATAL_ERROR2(("Creation of user accessible static regions "\
                       "located in KSEG0 or KSEG1 is not allowed.\n"),
                      OSE_MM_ECONFIGURATION_ERROR, (OSERRCODE)
                      "User mode permissions not allowed for KSEG region.");
   }
#endif
   
   (void)add_static_region(param, name, reg.log_base, reg.size,
                           cache_mode, (U32)perm,
                           reg.phys_base, reg.reserve_phys_mem);
   return;
}


/*****************************************************************************
 *
 * Check log_mem entries
 *   - RAM:      Exists and has mapping type SAS.
 *   - RAM_SASE: Exists, has mapping type SASE and is covered by RAM memory.
 *   - HAL:      Exists and has mapping type SAS.
 */
static void
check_logmem_entries_mmu(void)
{
   struct partition_class *ram_sase = mm_find_partition_by_name(RAM_SASE_NAME);

   /* Check log_mem RAM */
   {
      struct partition_class *ram = mm_find_partition_by_name(RAM_NAME);
      if (ram == NULL)
      {
         MM_FATAL_ERROR2(("Missing mandatory log_mem 'RAM'\n"),\
                         OSE_MM_ECONFIGURATION_ERROR,
                         (OSERRCODE)("Missing mandatory log_mem 'RAM'."));
      }
      
      if  ((ram->options & MM_PARTITION_OPT_SAS) != MM_PARTITION_OPT_SAS)
      {
         MM_FATAL_ERROR2(("Mandatory log_mem 'RAM' "
                          "must have mapping type SAS\n"),
                         OSE_MM_ECONFIGURATION_ERROR, (OSERRCODE)
                         "Mandatory log_mem 'RAM' must have mapping type SAS.");
      }

#ifdef OSMIPS
      /* MMU -> log_mem "RAM" must be SAS and inside KUSEG. */
      if (!ram->is_subset_of(ram, KUSEG_BASE, KUSEG_SIZE))
      {
         MM_FATAL_ERROR2(("%s must be located in KUSEG [0x%08lx-0x%08lx]\n",
                          LOGMEM_PREFIX RAM_NAME,
                          KUSEG_BASE, KUSEG_BASE + KUSEG_SIZE),
                         OSE_MM_ECONFIGURATION_ERROR,
                         (OSERRCODE)"Mandatory log_mem RAM outside KUSEG.");
      }
#endif
      
   }

   /* Check log_mem RAM_SASE */
   {
      if (ram_sase == NULL)
      {
         MM_FATAL_ERROR2(("Missing mandatory log_mem '" RAM_SASE_NAME "'\n"),
                         OSE_MM_ECONFIGURATION_ERROR,
                         (OSERRCODE)("Missing mandatory log_mem '"
                         RAM_SASE_NAME "'."));
      }
      
      /* log_mem "RAM_SASE" must be SASE. */
      if  ((ram_sase->options & MM_PARTITION_OPT_SASE) != MM_PARTITION_OPT_SASE)
      {
         MM_FATAL_ERROR2(("mandatory log_mem '" RAM_SASE_NAME
                          "' must have mapping type SASE\n"),
                         OSE_MM_ECONFIGURATION_ERROR, (OSERRCODE)
                         "Mandatory log_mem 'RAM_SASE' must have mapping type SASE.");
      }
      
      /* log_mem 'RAM_SASE' must be located inside the physical RAM. */
      if (!ram_sase->is_subset_of_phys_mem(ram_sase))
      {
         MM_FATAL_ERROR2(("log_mem entry '" RAM_SASE_NAME
                          " must be located inside the physical RAM\n"),
                         OSE_MM_ECONFIGURATION_ERROR, (OSERRCODE)
                         "Mandatory log_mem RAM_SASE outside phys_mem.");
      }

#ifdef OSMIPS
      if (!ram_sase->is_subset_of(ram_sase, KUSEG_BASE, KUSEG_SIZE))
      {
         MM_FATAL_ERROR2(("%s must be located in KUSEG [%#10lx, %#10lx)\n",
                          RAM_SASE_NAME,
                          KUSEG_BASE, KUSEG_BASE + KUSEG_SIZE),
                         OSE_MM_ECONFIGURATION_ERROR, (OSERRCODE)
                         "Mandatory log_mem RAM_SASE outside KUSEG.");
      }
#endif
   }

   /* Check log_mem HAL */
   {
      struct partition_class *part_hal_class =
         mm_find_partition_by_name(HAL_LOG_MEM_NAME);
      
      if (part_hal_class == NULL)
      {
         /* Try to create log_mem/HAL somewhere. */
         part_hal_class =
	    mm_create_hal_partition(HAL_LOG_MEM_NAME, HAL_LOG_MEM_SIZE,
				    partition_is_implicit(ram_sase));
         if (part_hal_class == NULL)
         {
            MM_FATAL_ERROR2(("log_mem " HAL_LOG_MEM_NAME " missing, and could"
                             " not be created automatically.\n"),
                            OSE_MM_ECONFIGURATION_ERROR,
                            (OSERRCODE)("log_mem " HAL_LOG_MEM_NAME " missing,"
                                        " and could not be created"
                                        " automatically."));
         }
         
         {
            /* The HAL partition is not implicit => the get_base function is
             * available. */
            int status;
            log_addr_t base;
            log_size_t size;

            status = part_hal_class->get_bank_info(part_hal_class, 0, &base, &size);
            if (status == 0)
            {
               MM_FATAL_ERROR2(("log_mem " HAL_LOG_MEM_NAME " missing, and could"
                                " not be created automatically.\n"),
                               OSE_MM_ECONFIGURATION_ERROR, (OSERRCODE)0);
            }
            
            ramlog_printf("mm: log_mem  [0x%08lx-0x%08lx]     %4s %-32s\n",
                          base, base+size-1, "SAS", part_hal_class->name);
         }

      }
      else
      {
         log_addr_t base;
         log_size_t size;
         struct allocator *pa;
         int status;

         status = part_hal_class->get_bank_info(part_hal_class, 0, &base, &size);
         if (status == 0)
         {
            MM_FATAL_ERROR2(("log_mem " HAL_LOG_MEM_NAME " has no configured "
                             "logical range (bank).\n"),
                            OSE_MM_ECONFIGURATION_ERROR, (OSERRCODE)0);
         }
         
         if ((part_hal_class->options & MM_PARTITION_MAPTYPE_MASK) !=
             MM_PARTITION_OPT_SAS)
         {
            MM_FATAL_ERROR2(("Mandatory log_mem '" HAL_LOG_MEM_NAME "' must have"
                             " mapping_type SAS.\n"),
                            OSE_MM_ECONFIGURATION_ERROR, (OSERRCODE)
                            ("Mandatory log_mem '" HAL_LOG_MEM_NAME "' must have"
                             " mapping_type SAS."));
         }

         if (size < 16 * HAL_SIZE_1M)
         {
            MM_FATAL_ERROR2(("Size of log_mem '" HAL_LOG_MEM_NAME "' must be"
                             " at least 16 MB.\n"),
                            OSE_MM_EHAL_LACKING_MEMORY, (OSERRCODE)
                            ("log_mem 'HAL' must be at least 16 MB."));
         }         

         pa = mm_find_allocator_by_range(base, size);
         if (NULL != pa && partition_is_implicit(ram_sase))
         {
            MM_FATAL_ERROR2(("log_mem entry '" HAL_LOG_MEM_NAME "' is overlaps"
                             " physical memory %s.\n", pa->name),
                            OSE_MM_ECONFIGURATION_ERROR,
                            (OSERRCODE)
                            ("log_mem entry 'HAL' is not allowed to overlap "
                             "the boundaries of phys_mem."));
         }
      }
      gmm_state.metadata_log_mem = part_hal_class;
   }
}

/*****************************************************************************/
/*
 * Check:
 *   - log_mem 'RAM' with mapping type SASE, covered by RAM memory.
 */

static void
check_logmem_entries_nommu(void)
{
   struct partition_class *log_ram  = mm_find_partition_by_name(RAM_NAME);

   if (log_ram == NULL)
   {
      MM_FATAL_ERROR2(("Missing mandatory log_mem 'RAM'\n"),\
                      OSE_MM_ECONFIGURATION_ERROR,
                      (OSERRCODE)("Missing mandatory log_mem 'RAM'."));
   }

   if  ((log_ram->options & MM_PARTITION_OPT_SASE) != MM_PARTITION_OPT_SASE)
   {
      MM_FATAL_ERROR2(("Mandatory log_mem 'RAM' must have mapping " \
                       "type SASE\n"),
                      OSE_MM_ECONFIGURATION_ERROR, (OSERRCODE)
                      "Mandatory log_mem 'RAM' must have mapping type SASE.");
   }
   
#ifdef OSMIPS
   /* !MMU -> log_mem "RAM" must be SASE and inside KSEG0 or KSEG1. */
   if (!partition_is_implicit(log_ram) && 
       !(log_ram->is_subset_of(log_ram, KSEG0_BASE, KSEG0_SIZE) ||
         log_ram->is_subset_of(log_ram, KSEG1_BASE, KSEG1_SIZE)))
   {
      MM_FATAL_ERROR2(("%s must be located in KSEG0 [%#10lx, %#10lx] "
                       "or KSEG1 [%#10lx, %#10lx]\n",
                       LOGMEM_PREFIX RAM_NAME,
                       KSEG0_BASE, KSEG0_BASE + KSEG0_SIZE - 1,
                       KSEG1_BASE, KSEG1_BASE + KSEG1_SIZE - 1),
                      OSE_MM_ECONFIGURATION_ERROR, (OSERRCODE)
                      "Mandatory log_mem RAM outside KSEG0/KSEG1.");
   }
#endif
   
   if (!log_ram->is_subset_of_phys_mem(log_ram))
   {
      MM_FATAL_ERROR2(("log_mem entry '" RAM_NAME "' must be located inside "
                       "the physical RAM\n"),
                      OSE_MM_ECONFIGURATION_ERROR, (OSERRCODE)
                      "Mandatory log_mem RAM outside phys_mem.");
   }
}

/*****************************************************************************
 *
 * Check that all required log_mem entries exist with correct
 * properties:
  */
static void
check_logmem_entries(void) {
   if (HAS_MMU) {
      check_logmem_entries_mmu();
   }
   else {
      check_logmem_entries_nommu();
   }
}


/* Map all regions in kernel domain. */
static void
map_regions(void)
{
   struct region *region;
   core_rbtree_t *tree;

   VERBOZE(("mm: map_regions()\n"));
   tree = &gmm_state.domains.domain_list[MM_KERNEL_DOMAIN]->regions;
   region = (struct region*)core_rbtree_first_no_lock(tree);
   while (region != NULL)
   {
      mm_status_t status = mm_map_region(MM_KERNEL_DOMAIN, region);

      if (status != MM_SUCCESS)
      {
         MM_FATAL_ERROR2(("failed to map region %s\n",
                          region->name),
                         OSE_MM_ECONFIGURATION_ERROR,
                         (OSERRCODE)("Failed to map region."));
      }
      region = (struct region*)core_rbtree_next_no_lock(tree, &region->node);

      mm_kick_wdog();
   }
}

static long decode_from_permissions(long perm)
{
   if(perm & MM_SU_EXECUTE) 
   {
      return (HAL_CONTENTS_TYPE_TEXT|HAL_CONTENTS_TYPE_DATA);
   }
   return HAL_CONTENTS_TYPE_DATA;
}

static void
configure_resident(const char *param)
{
   struct partition_class *p_class;
   U32 perm, cache_mode;
   const char *name;
   hal_status_t status;
   
   name = param + strlen(RES_MEM_PREFIX);

   REQUIRE_ATTR_ENUM(perm, param,
                     PERMISSION_ATTRIBUTE_KEY,
                     permission_encoding);

   {
      char cache_type[MM_NAME_LENGTH+1];
      char mem_type[MM_NAME_LENGTH+1];

      REQUIRE_ATTR_STRING(cache_type, MM_NAME_LENGTH,
                          param, CACHE_ATTRIBUTE_KEY);
      REQUIRE_ATTR_STRING(mem_type, MM_NAME_LENGTH,
                          param, MEMORY_ATTRIBUTE_KEY);
      
      status = hal_get_cache_mode(cache_type, mem_type, &cache_mode);
      if (status != HAL_SUCCESS)
      {
         MM_FATAL_ERROR2(("Illegal cache and memory type combination detected "
                          "for resident log_mem type %s\n", param),
                         OSE_MM_ECONFIGURATION_ERROR, (OSERRCODE)
                         "Illegal cache and memory type combination.");
      }
   }

   p_class = mm_find_partition_by_name(name);
   if (p_class == NULL)
   {
      MM_FATAL_ERROR2(("Could not find resident log_mem '%s'\n", name),
                      OSE_MM_ECONFIGURATION_ERROR,
                      (OSERRCODE)("Could not find resident log_mem."));
   }

   if  ((p_class->options & MM_PARTITION_OPT_SASE) != MM_PARTITION_OPT_SASE)
   {
      MM_FATAL_ERROR2(("resident log_mem must have mapping type SASE\n"),
         OSE_MM_ECONFIGURATION_ERROR, (OSERRCODE)
         "Resident log_mem must have mapping type SASE.");
   }

   p_class->res_perm = perm;
   status = hal_actual_permissions(
      (U32*)&(p_class->res_perm), perm, p_class->banks->base, HAL_RESIDENT);
   if (status != HAL_SUCCESS)
   {
      MM_FATAL_ERROR2(("Illegal access permissions specified "
                       "for resident log_mem '%s'\n", name),
                      OSE_MM_ECONFIGURATION_ERROR, (OSERRCODE)
                      ("Illegal access permissions specified "
                       "for resident log_mem."));
   }

   status = hal_actual_cache_mode(
      (U32*)&(p_class->res_cache), cache_mode, p_class->banks->base);

   if (status != HAL_SUCCESS)
   {
      MM_FATAL_ERROR2(("Illegal cache mode attributes specified "
                       "for resident log_mem '%s'\n", name),
                      OSE_MM_ECONFIGURATION_ERROR, (OSERRCODE)
                      ("Illegal cache mode attributes specified "
                       "for resident log_mem."));
   }

   p_class->options |= MM_PARTITION_OPT_RESIDENT;

   fail_if_overlap2(LOGMEM_PREFIX, include_resident,
                    PHYS_MEM_PREFIX, include_phys_ram,
                    "Resident log_mem is not allowed "
                    "to overlap physical " PHYS_RAM_PREFIX  " memory.");
   
   p_class->res_options = decode_from_permissions(p_class->res_perm);
}

static void
create_boot_region(void)
{
   /* Free the memory allocated for the boot heap. It will soon be allocated
    * again by add_static_region. */
   {
      struct allocator *ram = gmm_state.ram_allocator;
      struct phys_object phys_obj;
      
      phys_obj.data_representation = PHYS_OBJ_FRAG_LIST;
      phys_obj.data.frag_list = boot_heap_data.frag_list;
      ram->free(ram, &phys_obj);
   }
   
   {
      struct region *region;
      log_addr_t log_base;
      log_size_t size;
      
      log_base = ALIGN_DOWN(boot_heap_data.curr_base,
                            hal_get_info()->page_sizes[0]);
      
      /* Both base and initial_top+1 should be properly aligned here. */
      size = boot_heap_data.init_top+1-log_base;
      
#ifdef OSMIPS
      log_base &= ~KSEG0_BASE;
#endif

      region = add_static_region(BOOT_REGION_NAME,
                                 BOOT_REGION_NAME,
                                 log_base, size,
                                 DEFAULT_CACHE_MODE,
                                 gmm_state.internal_access_mode,

                                 /* Phys base == log base */
                                 (phys_addr_t)log_base,

                                 /* Reserve the physical memory.*/
                                 TRUE);
      if (region == NULL)
      {
         MM_FATAL_ERROR2(("failed to add boot region (base:%#x, size:%#x)\n",
                          log_base, size),
                         OSE_MM_ECONFIGURATION_ERROR,
                         (OSERRCODE)("Failed to add boot region."));
      }
      
      if (mm_map_region(MM_KERNEL_DOMAIN, region) != MM_SUCCESS)
      {
         MM_FATAL_ERROR2(("failed to map region %s\n",
                          region->name),
                         OSE_MM_ECONFIGURATION_ERROR,
                         (OSERRCODE)("Failed to map region."));
      }
   }
}

/*****************************************************************************/
static void
find_sase_log_mem(phys_addr_t base, phys_addr_t top,
                  log_addr_t *sase_base, log_size_t *sase_size,
                  uint8_t *is_ram_sase)
{
   param_t param;
   int found_sase = 0;

   *sase_size = 0;
   *sase_base = 0;
   *is_ram_sase = 0;
   MAP_SYSPARAM(LOGMEM_PREFIX, param) {
      mapping_type_t mapping_type;

      REQUIRE_ATTR_ENUM_MAP_TYPE(mapping_type, param, MAPPING_ATTRIBUTE_KEY,
                                 mapping_type_encoding);
      if (mapping_type == SASE)
      {
         char *name = param + strlen(LOGMEM_PREFIX);
         char *end_of_name = strchr(name, '/');
         const char *ram_sase_name = ram_sase_param_name() +
            strlen(LOGMEM_PREFIX);
         log_addr_t found_base;
         log_size_t found_size;
         log_size_t size = top - base;

         readBaseSizeAndTop32(param, &found_base, &found_size, NULL);
         if (!((base >= found_base && base < found_base + found_size) ||
               (found_base >= base && found_base < base + size)))
         {
            /* SASE region does not overlap with phys_mem region. */
            continue;
         }

         if (end_of_name)
         {
            *end_of_name = '\0';
         }
         *sase_base = found_base;
         *sase_size = found_size;
         if (strcmp(name, ram_sase_name) == 0)
         {
            *is_ram_sase = 1;
         }

         if (found_sase)
         {
            MM_FATAL_ERROR2(("Found multiple SASE logical memory regions in "
                             "physical memory space ([0x%08llx-0x%08llx])\n",
                             base, top),
                            OSE_MM_ECONFIGURATION_ERROR,
                            (OSERRCODE)
                            ("Found multiple SASE logical memory regions "
                             "in one physical memory space"));
         }
         found_sase = 1;
      }
   }

}

/*****************************************************************************/

static void
add_phys_mem_explicit_ram_sase(void)
{
   U32 ram_sase_ok = 0;
   param_t param;
   log_size_t sase_size;
   log_addr_t sase_base;

   /* PHYS_RAM_PREFIX will match all parameters that begins with
    * that string => All phys types. */
   MAP_SYSPARAM(PHYS_RAM_PREFIX, param) {
      char *name = param+strlen(PHYS_MEM_PREFIX);
      char *end_of_name = strchr(name, '/');
      phys_addr_t base, top;
      phys_size_t size;
      struct allocator *phys_mem;
      uint8_t is_ram_sase;

      assert(end_of_name);

      readBaseSizeAndTop64(param, &base, &size, &top);

      if (NULL != hal_validate_phys_range)
      {
         hal_status_t hal_status = hal_validate_phys_range(base, size);

         if (HAL_SUCCESS != hal_status)
         {
            MM_FATAL_ERROR2(("Range [0x%08llx-0x%08llx] for parameter %s is "
                             "not supported by the HAL.\n",
                             base, base+size-1ULL, param),
                            OSE_MM_ECONFIGURATION_ERROR,
                            (OSERRCODE)
                            ("Unsupported physical range (HAL)."));
         }
      }

      /* Null terminate the name of the phys mem:
       * "krn/phys_mem/XXX/..." -> "krn/phys_mem/XXX\0..." */
      *end_of_name = '\0';

      phys_mem = mm_find_allocator_by_name(name);
      if (!phys_mem) {
         phys_mem = mm_create_ram_allocator(name);
      }

      find_sase_log_mem(base, top, &sase_base, &sase_size, &is_ram_sase);
      if (sase_size == 0)
      {
         phys_mem->add_sas_bank(phys_mem, base, top + 1 - base);
         ramlog_printf("mm: phys_mem [0x%010llx-0x%010llx]  SAS %-32s\n",
                       (phys_addr_t)base, (phys_addr_t)top, name);
      }
      else
      {
         log_addr_t sase_top = sase_base + sase_size - 1;

         if (sase_base > base) {
            phys_mem->add_sas_bank(phys_mem, base, sase_base - base);
            ramlog_printf("mm: phys_mem [0x%010llx-0x%010llx] SASE %-32s\n",
                          (phys_addr_t)base, (phys_addr_t)sase_base - 1, name);
         }
         phys_mem->add_sase_bank(phys_mem, sase_base, sase_size);
         ramlog_printf("mm: phys_mem [0x%010llx-0x%010llx] SASE %-32s\n",
                       (phys_addr_t)sase_base, (phys_addr_t)sase_top, name);
         if (sase_top < top) {
            phys_mem->add_sas_bank(phys_mem, (phys_addr_t)sase_top + 1,
                                   top - sase_top);
            ramlog_printf("mm: phys_mem [0x%010llx-0x%010llx]  SAS %-32s\n",
                          (phys_addr_t)sase_top+1, (phys_addr_t)top, name);
         }
         if (is_ram_sase)
         {
            ram_sase_ok = 1;
         }
      }

      /* Must be restored since param is used to find next. */
      *end_of_name = '/';
   }

   if (!ram_sase_ok) {
      MM_FATAL_ERROR2(("Found no physical memory space that matches "
                       "RAM_SASE\n"),
                      OSE_MM_ECONFIGURATION_ERROR,
                      (OSERRCODE)
                      ("Found no physical memory space that matches "
                       "RAM_SASE"));
   }
}

/*****************************************************************************/

static void
add_phys_mem_implicit_ram_sase(void)
{
   /* Read memory banks */
   
   param_t param;
   
   /* PHYS_MEM_PREFIX "RAM" will match all parameters that begins with
    * that string. */
   MAP_SYSPARAM(PHYS_RAM_PREFIX, param) {
      char *name = param+strlen(PHYS_MEM_PREFIX);
      char *end_of_name = strchr(name, '/');
      phys_addr_t base, top;
      phys_size_t size;
      struct allocator *phys_mem;
      phys_addr_t upper_limit;

#ifdef OSMIPS
      upper_limit = 0x80000000;
#else
      upper_limit = 0x100000000ULL;
#endif

      assert(end_of_name);
      readBaseSizeAndTop64(param, &base, &size, &top);
      
      if (NULL != hal_validate_phys_range)
      {
         hal_status_t hal_status = hal_validate_phys_range(base, size);

         if (HAL_SUCCESS != hal_status)
         {
            MM_FATAL_ERROR2(("Range [0x%08llx-0x%08llx] for parameter %s is "
                             "not supported by the HAL.\n",
                             base, base+size-1ULL, param),
                            OSE_MM_ECONFIGURATION_ERROR,
                            (OSERRCODE)
                            ("Unsupported physical range (HAL)."));
         }
      }
      
      /* Null terminate the name of the phys mem:
       * "krn/phys_mem/XXX/..." -> "krn/phys_mem/XXX\0..." */
      *end_of_name = '\0';
      
      phys_mem = mm_find_allocator_by_name(name);
      if (!phys_mem) {
         phys_mem = mm_create_ram_allocator(name);
      }
      
      if (base >= upper_limit)
      {
         /* Entirely above upper_limit. */
         phys_mem->add_sas_bank(phys_mem, base, top+1-base);
         ramlog_printf("mm: phys_mem [0x%010llx-0x%010llx]  SAS %-32s\n",
                       (phys_addr_t)base, (phys_addr_t)top, name);
      }
      else if (top < upper_limit)
      {
         /* Entirely below upper_limit. */
         phys_mem->add_sase_bank(phys_mem, base, top+1-base);
         ramlog_printf("mm: phys_mem [0x%010llx-0x%010llx] SASE %-32s\n",
                       (phys_addr_t)base, (phys_addr_t)top, name);
      }
      else {
         /* Both below and above upper_limit. */

         phys_addr_t base0 = base;
         phys_addr_t base1 = upper_limit;
         phys_addr_t top0  = upper_limit - 1;
         phys_addr_t top1  = top;
         
         phys_mem->add_sase_bank(phys_mem, base0, top0+1-base0);
         ramlog_printf("mm: phys_mem [0x%010llx-0x%010llx] SASE %-32s\n",
                       (phys_addr_t)base0, (phys_addr_t)top0, name);
         phys_mem->add_sas_bank(phys_mem, base1, top1+1-base1);
         ramlog_printf("mm: phys_mem [0x%010llx-0x%010llx]  SAS %-32s\n",
                       (phys_addr_t)base1, (phys_addr_t)top1, name);
      }

      /* Must be restored since param is used to find next. */
      *end_of_name = '/';
   }
}

/**************************************************************************/

#ifdef OSX86

static U32
read_cache_type_x86(param_t param)
{
   static const struct enumeration enum_policy[] = {
      {"disabled",      HAL_X86_CACHE_PCD},
      {"inhibited",     HAL_X86_CACHE_PCD},
      {"inhibit",       HAL_X86_CACHE_PCD},
      {"uncached",      HAL_X86_CACHE_PCD},
      {"off",           HAL_X86_CACHE_PCD},
      {"false",         HAL_X86_CACHE_PCD},
      {"0",             HAL_X86_CACHE_PCD},
      {"write_through", HAL_X86_CACHE_PWT},
      {"copy_back",     0},
      {"write_back",    0},
      {NULL,            0}
   };
   
   U32 bits = 0;
   long val;
   int rc;

   rc = sysparam_readattr_enum(param, "policy", enum_policy, &val);
   if (rc == EOK)
   {
      bits |= val;
   }
   else if (rc == EINVAL)
   {
      ATTRIBUTE_ERROR(param, "policy", rc);
   }
   else
   {
      /* Default: write_back = 0. */
   }

   return bits;
}

#endif

/**************************************************************************/

#ifdef OSMIPS

static U32
read_cache_type_mips(param_t param)
{
   static const struct enumeration enum_policy[] = {
      {"inhibited",    0},
      {"inhibit",      0},
      {"uncached",     0},
      {"write_back",   HAL_MIPS_CACHE_ENABLE},
      {"copy_back",    HAL_MIPS_CACHE_ENABLE},
      {"cacheable",    HAL_MIPS_CACHE_ENABLE},
      {NULL,           0}
   };
   
   U32 bits = 0;
   long val;
   int rc;

   rc = sysparam_readattr_enum(param, "policy", enum_policy, &val);
   if (rc == EOK)
   {
      bits |= val;
   }
   else if (rc == EINVAL)
   {
      ATTRIBUTE_ERROR(param, "policy", rc);
   }
   else
   {
      /* Default: cacheable. */
      bits |= HAL_MIPS_CACHE_ENABLE;
   }

   return bits;
}

#endif

/**************************************************************************/
#ifdef OSPP

static U32
read_cache_type_ppc(param_t param)
{
   static const struct enumeration enum_policy[] = {
      {"write_back",    0},
      {"copy_back",     0},
      {"write_through", HAL_PPC_CACHE_WT},
      {"inhibited",     HAL_PPC_CACHE_CI},
      {"inhibit",       HAL_PPC_CACHE_CI},
      {"uncached",      HAL_PPC_CACHE_CI},
      {NULL,            0}
   };
   
   U32 bits = 0;
   long val;
   int rc;

   rc = sysparam_readattr_enum(param, "policy", enum_policy, &val);
   if (rc == EOK)
   {
      bits |= val;
   }
   else if (rc == EINVAL)
   {
      ATTRIBUTE_ERROR(param, "policy", rc);
   }
   else
   {
      /* Default: write_back = 0. */
   }

   return bits;
}

/**************************************************************************/

static U32
read_mem_type_ppc(param_t param)
{
   static const struct enumeration enum_endian[] = {
      {"big",    0},
      {"little", HAL_PPC_MEM_ENDIAN},
      {NULL,     0}
   };
   
   U32 bits = 0;
   Boolean b;
   long val;
   int rc;

   rc = sysparam_readattr_enum(param, "endian", enum_endian, &val);
   if (rc == EOK)
   {
      bits |= val;
   }
   else if (rc == EINVAL)
   {
      ATTRIBUTE_ERROR(param, "endian", rc);
   }
   else
   {
      /* Default: big = 0. */
   }

   rc = sysparam_readattr_bool(param, "guarded", &b);
   if (rc == EOK && b == True)
   {
      bits |= HAL_PPC_MEM_GUARDED;
   }
   else if (rc == EINVAL)
   {
      ATTRIBUTE_ERROR(param, "guarded", rc);
   }
   else
   {
      /* Default: non-guarded = 0. */
   }

   rc = sysparam_readattr_bool(param, "coherency", &b);
   if (rc == EOK && b == True)
   {
      bits |= HAL_PPC_MEM_COHERENT;
   }
   else if (rc == EOK && b == False)
   {
      /* Non coherent choosen. */
   }
   else if (rc == EINVAL)
   {
      ATTRIBUTE_ERROR(param, "coherency", rc);
   }
   else
   {
#ifdef OSE_SMP
      bits |= HAL_PPC_MEM_COHERENT;
#endif
   }

   return bits;
}

#endif

/**************************************************************************/
#ifdef OSARM

static U32
read_cache_type_arm(param_t param)
{
   static const struct enumeration enum_policy[] = {
      {"write_back",    HAL_ARM_CACHE_WB},
      {"copy_back",     HAL_ARM_CACHE_WB},
      {"write_through", HAL_ARM_CACHE_WT},
      {"inhibited",     HAL_ARM_CACHE_CI},
      {"inhibit",       HAL_ARM_CACHE_CI},
      {"uncached",      HAL_ARM_CACHE_CI},
      {NULL,            0}
   };

   static const struct enumeration enum_allocate[] = {
      {"read",          0},
      {"write",         HAL_ARM_CACHE_WA},
      {"disabled",      0},
      {"enabled",       HAL_ARM_CACHE_WA},
      {"false",         0},
      {"true",          HAL_ARM_CACHE_WA},
      {"off",           0},
      {"on",            HAL_ARM_CACHE_WA},
      {"no",            0},
      {"yes",           HAL_ARM_CACHE_WA},
      {"0",             0},
      {"1",             HAL_ARM_CACHE_WA},
      {NULL,            0}
   };
   
   U32 bits = 0;
   long val;
   int rc;

   rc = sysparam_readattr_enum(param, "policy", enum_policy, &val);
   if (rc == EOK)
   {
      bits |= val;
   }
   else if (rc == EINVAL)
   {
      ATTRIBUTE_ERROR(param, "policy", rc);
   }
   else
   {
      /* Default: write_back */
      bits |= HAL_ARM_CACHE_WB;
   }

   rc = sysparam_readattr_enum(param, "allocate", enum_allocate, &val);
   if (rc == EOK)
   {
      bits |= val;
   }
   else if (rc == EINVAL)
   {
      ATTRIBUTE_ERROR(param, "allocate", rc);
   }
   else
   {
      /* Default: read allocation = 0 */
   }

   return bits;
}

/**************************************************************************/

static U32
read_mem_type_arm(param_t param)
{
   static const struct enumeration enum_type[] = {
      {"strongly_ordered",  HAL_ARM_MEM_ORDERED},
      {"ordered",           HAL_ARM_MEM_ORDERED},
      {"device",            HAL_ARM_MEM_DEVICE },
      {"normal",            HAL_ARM_MEM_NORMAL },
      {NULL, 0}
   };
   
   U32 bits = 0;
   Boolean b;
   long val;
   int rc;

   rc = sysparam_readattr_enum(param, "type", enum_type, &val);
   if (rc == EOK)
   {
      bits |= val;
   }
   else if (rc == EINVAL)
   {
      ATTRIBUTE_ERROR(param, "type", rc);
   }
   else
   {
      /* Default: normal. */
      bits |= HAL_ARM_MEM_NORMAL;
   }

   rc = sysparam_readattr_bool(param, "shared", &b);
   if (rc == EOK && b == True)
   {
      bits |= HAL_ARM_MEM_SHARED;
   }
   else if (rc == EINVAL)
   {
      ATTRIBUTE_ERROR(param, "shared", rc);
   }
   else
   {
#ifdef OSE_SMP
      if (bits & HAL_ARM_MEM_NORMAL)
      {
         bits |= HAL_ARM_MEM_SHARED;
      }
#endif
   }


   return bits;
}

#endif

/**************************************************************************/

static void
mm_cache_mem_type_init(void)
{
   hal_status_t status;
   param_t param;
   
   /* Add default types. */
#if defined(OSPP)
   hal_add_ctype("inhibited",            HAL_PPC_CACHE_CI, 0);
   hal_add_ctype("write_through",        HAL_PPC_CACHE_WT, 0);
   hal_add_ctype("copy_back",            0, 0);
#  ifdef OSE_SMP
   hal_add_mtype("ordered_access",       HAL_PPC_MEM_COHERENT|HAL_PPC_MEM_GUARDED);
   hal_add_mtype("speculative_access",   HAL_PPC_MEM_COHERENT);
#  else
   hal_add_mtype("ordered_access",       HAL_PPC_MEM_GUARDED);
   hal_add_mtype("speculative_access",   0);
#  endif
   /* This type is for internal use only (htab).
    * Not exposed in external interfaces. */
   hal_add_mtype("speculative_coherent", HAL_PPC_MEM_COHERENT);
#elif defined(OSARM)
   hal_add_ctype("inhibited",            HAL_ARM_CACHE_CI,0);
   hal_add_ctype("write_through",        HAL_ARM_CACHE_WT,0);
   hal_add_ctype("copy_back",            HAL_ARM_CACHE_WB | HAL_ARM_CACHE_WA,0);
   if (ARM_HAL_IS_V6)
   {
      hal_add_mtype("ordered_access",    HAL_ARM_MEM_DEVICE|HAL_ARM_MEM_SHARED);
   }
   else
   {
      hal_add_mtype("ordered_access",    HAL_ARM_MEM_ORDERED|HAL_ARM_MEM_SHARED);
   }
#  ifdef OSE_SMP
   hal_add_mtype("speculative_access",   HAL_ARM_MEM_NORMAL|HAL_ARM_MEM_SHARED);
#  else
   hal_add_mtype("speculative_access",   HAL_ARM_MEM_NORMAL);
#  endif
   hal_add_mtype("speculative_coherent", HAL_ARM_MEM_NORMAL|HAL_ARM_MEM_SHARED);
#elif defined(OSMIPS)
   hal_add_ctype("inhibited",            0,0);
   hal_add_ctype("write_through",        HAL_MIPS_CACHE_ENABLE,0);
   hal_add_ctype("copy_back",            HAL_MIPS_CACHE_ENABLE,0);
   hal_add_mtype("ordered_access",       0);
   hal_add_mtype("speculative_access",   0);
   hal_add_mtype("speculative_coherent", 0);
#elif defined(OSX86)
   hal_add_ctype("inhibited",            HAL_X86_CACHE_PCD,0);
   hal_add_ctype("write_through",        HAL_X86_CACHE_PWT,0);
   hal_add_ctype("copy_back",            0,0);
   hal_add_mtype("ordered_access",       0);
   hal_add_mtype("speculative_access",   0);
   hal_add_mtype("speculative_coherent", 0);
#else
   hal_add_ctype("inhibited",            0,0);
   hal_add_ctype("write_through",        0,0);
   hal_add_ctype("copy_back",            0,0);
   hal_add_mtype("ordered_access",       0);
   hal_add_mtype("speculative_access",   0);
   hal_add_mtype("speculative_coherent", 0);
#endif
   
   
   MAP_SYSPARAM_FILTER(CACHETYPE_PREFIX, param, exclude_outer) {
      char *name = param + strlen(CACHETYPE_PREFIX);
      char *level_name = strchr(name, '/');
      U32 bits;

#ifndef OSARM
      /* Per-level configuration only supported on ARM. */
      if (level_name != NULL)
      {
         MM_FATAL_ERROR2(
            ("krn/cache_type per-level configuration not supported.\n"),
            OSE_MM_ECONFIGURATION_ERROR,
            (OSERRCODE)("krn/cache_type per-level configuration not supported."));
      }
#else
      /* Verify that level is 1/inner, 2/outer, or all. */
      if (level_name != NULL)
      {
         if (strcmp(level_name+1, "inner") != 0 &&
             strcmp(level_name+1, "1")     != 0 &&
             strcmp(level_name+1, "all")   != 0)
         {
            MM_FATAL_ERROR2(
               ("krn/cache_type per-level configuration is illegal.\n"),
               OSE_MM_ECONFIGURATION_ERROR,
               (OSERRCODE)("krn/cache_type per-level configuration is illegal."));
         }
      }
#endif
      
#if   defined(OSPP)
      bits = read_cache_type_ppc(param);
#elif defined(OSARM)
      bits = read_cache_type_arm(param);
      bits = HAL_ARM_CACHE(bits, 1);
#elif defined(OSMIPS)
      bits = read_cache_type_mips(param);
#elif defined(OSX86)
      bits = read_cache_type_x86(param);
#else
      bits = 0;
      MM_FATAL_ERROR2(
         ("krn/cache_type configuration not supported in soft kernel.\n"),
         OSE_MM_ECONFIGURATION_ERROR,
         (OSERRCODE)("krn/cache_type configuration not supported in soft kernel."));
#endif

      if (level_name != NULL)
         *level_name = 0;

      status = hal_add_ctype(name, bits, 0);
      if (status != HAL_SUCCESS)
      {
         MM_FATAL_ERROR2(
            ("Failed to register cache type '%s': status=%d\n", name, status),
            OSE_MM_ECONFIGURATION_ERROR,
            (OSERRCODE)("Failed to register cache type configuration."));
      }

      /* Restore parameter string to allow iteration to the next type. */
      if (level_name != NULL)
         *level_name = '/';
   }


   MAP_SYSPARAM_FILTER(CACHETYPE_PREFIX, param, include_outer) {
#ifndef OSARM
      MM_FATAL_ERROR2(
         ("krn/cache_type per-level configuration is not supported.\n"),
         OSE_MM_ECONFIGURATION_ERROR,
         (OSERRCODE)("krn/cache_type per-level configuration not supported."));
#else
      char *name = param + strlen(CACHETYPE_PREFIX);
      char *level_name = strchr(name, '/');
      U32 bits;

      bits = read_cache_type_arm(param);
      bits = HAL_ARM_CACHE(bits, 2);

      *level_name = 0;
      status = hal_add_ctype(name, bits, HAL_APPEND);
      if (status != HAL_SUCCESS)
      {
         MM_FATAL_ERROR2(
            ("Failed to register cache type '%s': status=%d\n", name, status),
            OSE_MM_ECONFIGURATION_ERROR,
            (OSERRCODE)("Failed to register cache type configuration."));
      }
      *level_name = '/';
#endif
   }

   
   MAP_SYSPARAM(MEMTYPE_PREFIX, param) {
      const char *name = param + strlen(MEMTYPE_PREFIX);
      char *level_name = strchr(name, '/');
      U32 bits;

      if (level_name != NULL)
      {
         MM_FATAL_ERROR2(
            ("krn/mem_type per-level configuration is illegal.\n"),
            OSE_MM_ECONFIGURATION_ERROR,
            (OSERRCODE)("krn/mem_type per-level configuration is illegal."));
      }

#if   defined(OSPP)
      bits = read_mem_type_ppc(param);
#elif defined(OSARM)
      bits = read_mem_type_arm(param);
#else
      bits = 0xDEADBEEF; /* Kill warning. */
      MM_FATAL_ERROR2(
         ("krn/mem_type configuration not supported for this architecture.\n"),
         OSE_MM_ECONFIGURATION_ERROR,
         (OSERRCODE)("krn/mem_type configuration not supported for this architecture."));
#endif

      status = hal_add_mtype(name, bits);
      if (status != HAL_SUCCESS)
      {
         MM_FATAL_ERROR2(
            ("Failed to register memory type '%s': status=%d\n", name, status),
            OSE_MM_ECONFIGURATION_ERROR,
            (OSERRCODE)("Failed to register memory type configuration."));
      }
   }
}


/**************************************************************************/



/******************************************************************************
 * This is the main init function.
 *
 * The order of events is determined by dependencies between the
 * different modules.
 *
 * A coarse outline:
 * 1) Determine an area of free SASE mapped memory in RAM. Internal allocators
 *    break memory from this area during the boot phase.
 *
 *    If a region with name MM-meta-data exists in system parameters,
 *    then this region is used. Otherwise free space is located and a
 *    region ic created (with name = MM-data-data).
 *
 * 2) create data structures for static regions. When this is done
 *    memory may be allocated without "stepping on someones toe".
 *    mm_page_alloc() is ready after this.
 * 3) Initialize HAL
 * 4) map all static regions.
 * 5) enable MMU.
 *
 */

void
mm_config_init(void)
{
   Boolean implicit_ram_sase = 0;

   assert(hal_get_info() != NULL);

   verify_configuration();

   {
      long max_domains = 0;
      int rc = sysparam_readattr_long("krn/gmm", "max_domains", &max_domains);

      if(EOK == rc && max_domains > 0 && max_domains <= MM_MAX_DOMAINS)
      {
	 gmm_state.max_domains = (U32)max_domains;
      }
      else if (ENOENT == rc)
      {
	 gmm_state.max_domains = MM_MAX_DOMAINS;
      }
      else
      {
         MM_FATAL_ERROR2(("krn/gmm=max_domains:%ld is out of range [1..%u]\n",
                          max_domains, MM_MAX_DOMAINS),
                         OSE_MM_ECONFIGURATION_ERROR,
                         (OSERRCODE)"Failed to set max_domains (value out of range).");
      }
      ramlog_printf("mm: max_domains:%u @ %p\n",
		    gmm_state.max_domains,
		    &gmm_state.max_domains);
   }

   if (HAL_SUCCESS !=
       hal_actual_permissions(&(gmm_state.internal_access_mode),
                              DEFAULT_ACCESS_ATTR, /*ram_sase_base*/0, 0))
   {
      MM_FATAL_ERROR2(("Default access attributes "
                       "incompatible with location of log_mem RAM.\n"),
                      OSE_MM_ECONFIGURATION_ERROR,
                      (OSERRCODE)
                      ("Default access attributes "
                       "incompatible with location of log_mem RAM."));
   }
   
   if (EOK != sysparam_readattr_bool(ram_sase_param_name(),
                                     IMPLICIT_ATTRIBUTE_KEY,
                                     &implicit_ram_sase))
   {
      implicit_ram_sase = False;
   }
   
   {
      /* Create boot heap */
      param_t param;

      if (sysparam_read(REGION_PREFIX BOOT_REGION_NAME, param, sizeof param))
      {
         {
            U32 configured_perm;
            U32 min_perm = MM_SU_WRITE | MM_SU_READ;
            int rc;
            
            rc = sysparam_readattr_enum(REGION_PREFIX BOOT_REGION_NAME,
                                        PERMISSION_ATTRIBUTE_KEY,
                                        permission_encoding,
                                        (long *) &configured_perm);

            if (rc != EOK)
            {
               MM_FATAL_ERROR2(
                  ("Attribute " PERMISSION_ATTRIBUTE_KEY " for parameter "
                   REGION_PREFIX BOOT_REGION_NAME " is missing/invalid.\n"),
                  OSE_MM_ECONFIGURATION_ERROR,
                  (OSERRCODE)
                  ("Attribute " PERMISSION_ATTRIBUTE_KEY " for parameter "
                   REGION_PREFIX BOOT_REGION_NAME " is missing/invalid."));
            }
            
            if ((configured_perm & min_perm) != min_perm) {
               const char *msg =
                  "Boot heap requires read and write permission for supervisor"
                  " mode. Check attribute " PERMISSION_ATTRIBUTE_KEY " for"
                  " configuration parameter " REGION_PREFIX BOOT_REGION_NAME
                  ".";
               
               MM_FATAL_ERROR2(("%s\n", msg),
                               OSE_MM_ECONFIGURATION_ERROR,
                               (OSERRCODE)msg);
            }
         }
         
         {
            log_addr_t base, top;
            
            readBaseSizeAndTop32(REGION_PREFIX BOOT_REGION_NAME,
                                 &base, NULL, &top);
            init_boot_heap(base, top);
            boot_heap_data.auto_configured = 0;
            VERBOZE(("Boot heap explicitly configured [0x%08lx-0x%08lx]\n",
                     base, top));
         }
      }
      else if (implicit_ram_sase) {
         auto_configure_boot_heap_implicit_ram_sase();
      }
      else {
         auto_configure_boot_heap_explicit_ram_sase();
      }
   }
   
   mm_allocator_init();
   
   if (implicit_ram_sase)
   {
      add_phys_mem_implicit_ram_sase();
   }
   else
   {
      add_phys_mem_explicit_ram_sase();
   }
   
   reserve_boot_heap_mem();
   
   gmm_state.state = zone_ok;

#ifdef HOST_TEST
   return;
#endif

   /* Init sub-systems and keep watchdog alive. */
   mm_kick_wdog();
   mm_partition_init();
   mm_kick_wdog();
   mm_domain_init();
   mm_kick_wdog();
   mm_phys_object_init();
   mm_log_object_init();
   mm_kick_wdog();

   {
      /* Create log_mem entries from system parameters. */
      param_t param;

      /* Implicit log_mem:s must be created first of all, otherwise
       * there is a risk that other log_mem:s are allowed to overlap
       * phys_mem. Sysparam doesn't read configuration entries
       * in any sane order. */
      MAP_SYSPARAM_FILTER(LOGMEM_PREFIX, param, include_implicit_only) {
         struct log_mem_param log_mem_param;

         read_logmem_entry(param, &log_mem_param);
         create_logmem_entry(&log_mem_param, implicit_ram_sase);
      }

      mm_kick_wdog();

      /* Next.. create regular log_mem:s that aren't aliases. */
      MAP_SYSPARAM_FILTER(LOGMEM_PREFIX, param, exclude_aliases_and_implicit) {
         struct log_mem_param log_mem_param;

         read_logmem_entry(param, &log_mem_param);
         create_logmem_entry(&log_mem_param, implicit_ram_sase);
      }

      mm_kick_wdog();

      /* And take aliases last when it is "guaranteed" that the
       * derived log_mem exists (unless there is a configuration error).  */
      MAP_SYSPARAM_FILTER(LOGMEM_PREFIX, param, include_aliases_only) {
         struct log_mem_param log_mem_param;

         read_logmem_entry(param, &log_mem_param);
         create_logmem_entry(&log_mem_param, implicit_ram_sase);
      }

      check_logmem_entries();
      gmm_state.state = logmem_entries_created;
   }

   mm_kick_wdog();
   mm_cache_mem_type_init();
   mm_kick_wdog();

   
   /* Create GMM part of system domain. */
   {
      mm_status_t status = mm_add_domain(&gmm_state.domains,
                                         KERNEL_DOMAIN_NAME,
                                         &gmm_state.system_domain);
      
      if (status != MM_SUCCESS)
      {
         MM_FATAL_ERROR2(("Failed to create system domain (status=0x%lx).\n",
                          status),
                         OSE_MM_ECONFIGURATION_ERROR,
                         (OSERRCODE)("Failed to create system domain."));
      }
   }


   if (!boot_heap_data.auto_configured) {
      
      /* Free the memory allocated for the boot heap. It will soon be
       * allocated again by configure_region.
       */
      struct allocator *ram = gmm_state.ram_allocator;
      struct phys_object phys_obj;
      
      phys_obj.data_representation = PHYS_OBJ_FRAG_LIST;
      phys_obj.data.frag_list = boot_heap_data.frag_list;
      ram->free(ram, &phys_obj);
   }

   {
      param_t param;

      MAP_SYSPARAM(RES_MEM_PREFIX, param) {
         configure_resident(param);
         mm_kick_wdog();
      }
   }
   
   {
      /*
       * Create automatic regions with fixed address.
       * No mapping is done.
       * After this we know which space is free and which is not.
       */
      param_t param;

      MAP_SYSPARAM(REGION_PREFIX, param) {
         configure_region(param);
         mm_kick_wdog();
      }
      report_static_phys_mem_reserved();
      gmm_state.state = user_regions_created;
   }

   
   /*
    * Give the HAL module a change to allocate memory for hash
    * tables (mm_alloc_log_mem).
    */
   if (HAL_SUCCESS != hal_init_2())
   {
      MM_FATAL_ERROR2(("hal_init_2() failed.\n"),
                      MM_EINTERNAL_ERROR, 0);
   }

   mm_kick_wdog();

   /* After this call the HAL is functional. */
   {
      struct partition *hal = mm_find_domain_partition_by_name(
         gmm_state.system_domain, HAL_LOG_MEM_NAME);

      /* For the nommu case, hal_init_3 does not use the parameter. */
      if (HAS_MMU && (hal == NULL)) {
         MM_FATAL_ERROR2(("Partition " HAL_LOG_MEM_NAME " not found."),
                         MM_EINTERNAL_ERROR, 0);
      }
      if (HAL_SUCCESS != hal_init_3(hal)) {
         MM_FATAL_ERROR2(("hal_init_3() failed.\n"),
                         MM_EINTERNAL_ERROR, 0);
      }
   }
   gmm_state.state = hal_functional;

   mm_kick_wdog();

   /* Create HAL part of system domain. */
   assert(gmm_state.system_domain != NULL);
   {
      hal_status_t status;
      status = hal_create_domain(MM_KERNEL_DOMAIN, TRUE,
                                 &gmm_state.system_domain->hal_domain);
      if (status != HAL_SUCCESS) {
         MM_FATAL_ERROR2(("hal_create_domain failed (%d)\n", status),
                         MM_EINTERNAL_ERROR, 0);
      }
   }

   /* Notify zone allocator that HAL is functional. */
   mm_zone_late_init(HAL_LOG_MEM_NAME);

   mm_kick_wdog();

   /* Map all user supplied regions. */
   map_regions();
   gmm_state.state = regions_mapped;

   mm_kick_wdog();

   boot_heap_data.active = 0;
   
   /* Make a region of the boot area unless it was explicitly configured. */
   if (boot_heap_data.auto_configured) {
      /* Now, reserve the amount of memory that we've actually used
       * for the boot area.
       */
      create_boot_region();
   }

   /* Initialize multithreading state */
   gmm_state.multithread_enable = 0;

   /* The HAL module may now enable the MMU. */
   hal_init_4();

   mm_kick_wdog();

   ramlog_printf("MM init completed\n");
}
