/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2006 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

#include "core_hash.h"
#include "ose_err/mm_err.h"
#include "ramlog.h"
#include "ose.h"
#include "heapapi.h"
#ifdef STANDALONE
#include "standalone.h"
#endif

/************************** DEFINES *****************************/
#define HASH_ITERATOR_INVALID     0xfffffffful
#define HASH_ITERATOR_END         0xfffffffeul

/************************** PROTOTYPES **************************/

static uint32_t
hash_function(core_hash_table_t *hash, uint32_t value);

/************************** STATIC VARIABLES **************************/

/* A reference to the system heap that we need to allocate everything from. */
static OSHEAP system_heap = 0;


/*
 *===========================================================================
 *                         core_hash_table_init
 *===========================================================================
 *
 * Description
 * -----------
 * Initializes a hash table with a specific compare function and
 * hash value generator.
 * Needs to be called before any other operation on the hash table.
 *
 * Parameters
 * ----------
 * no_of_buckets  The number of buckets in this hash table.
 *                Statically defined at initialization.
 * funcs          Set of function callbacks required by the hash table.
 */
void
core_hash_table_init(core_hash_table_t *hash,
                     uint32_t           no_of_buckets,
                     core_hash_funcs_t *funcs)
{
   uint32_t i;

   if (no_of_buckets == 0)
   {
      ramlog_printf("Bucket size in hash cannot be zero.\n");
      error2(OSE_PAGING_INTERNAL_ERROR, 0);
   }

   if (!system_heap)
   {
      PROCESS pid;

      hunt("osemain", 0, &pid, NULL);
      if (pid == 0)
      {
         ramlog_printf("ERROR: Could not find the osemain process.\n");
         error2(OSE_PAGING_INTERNAL_ERROR, 0);
      }
      system_heap = ose_heap_ref(pid);
      if (system_heap == 0)
      {
         ramlog_printf("ERROR: Could not find the system heap.\n");
         error2(OSE_PAGING_INTERNAL_ERROR, 0);
      }
   }

   hash->funcs = funcs;
   hash->no_of_buckets = no_of_buckets;
   hash->bucket = (core_list_t *)
      ose_heap_malloc(system_heap,
                      no_of_buckets * sizeof(core_list_t),
                      __FILE__, __LINE__);
   
   if (!hash->bucket)
   {
      ramlog_printf("Ran out of memory initializing hash.\n");
      error2(OSE_PAGING_INTERNAL_ERROR, 0);
   }

   for (i = 0; i < no_of_buckets; i++)
   {
      core_list_init(&hash->bucket[i], (core_list_funcs_t*)&funcs->free_link);
   }
}

/*
 *===========================================================================
 *                         core_hash_table_delete
 *===========================================================================
 *
 * Description
 * -----------
 * Deletes a hash table. It does not delete any of the entries in
 * the hash table.
 *
 */
void
core_hash_table_delete(core_hash_table_t *hash)
{
   hash->no_of_buckets = 0;
   ose_heap_free(system_heap, hash->bucket);
}

/*
 *===========================================================================
 *                        hash_function
 *===========================================================================
 *
 * Description
 * -----------
 * Returns the correct bucket to search in.
 *
 * Returns
 * -------
 * uint32_t   Returns the correct bucket to search in.
 *
 */
static uint32_t
hash_function(core_hash_table_t *hash, uint32_t value)
{
   value += (value << 3);
   value ^= (value >> 11);
   value += (value << 15);
   return value % hash->no_of_buckets;
}

/*
 *===========================================================================
 *                        core_hash_find
 *===========================================================================
 *
 * Description
 * -----------
 * Searches for a matching entry in the hash table *hash.
 *
 * Returns
 * -------
 * struct core_list_link_t *   Returns the found entry or NULL if it is not found.
 *
 */
core_hash_link_t *
core_hash_find(core_hash_table_t    *hash,
               core_hash_iterator_t *iter,
               const void *key)
{
   core_list_link_t *entry;
   uint32_t what_bucket = hash_function(hash, hash->funcs->value(key));

   iter->bucket = what_bucket;

   entry = core_list_head(&hash->bucket[what_bucket], &iter->iterator);
   while (entry)
   {
      if (hash->funcs->compare(entry, key))
      {
         return entry;
      }
      entry = core_list_next(&iter->iterator);
   }
   
   core_hash_iterator_release(iter);
   
   return NULL;
}


/*
 *===========================================================================
 *                        core_hash_iterator_init
 *===========================================================================
 *
 * Description
 * -----------
 * Initializes a hash table iterator used to reference entries. 
 *
 * The iterator must always be released when it is no longer needed. Failure
 * to do so could result in hash entries being incorrectly referenced and
 * impossible to remove.
 *
 */
void
core_hash_iterator_init(core_hash_iterator_t *iterator,
                        core_hash_compare_func_t compare)
{
   iterator->hash    = NULL;
   iterator->key     = NULL;
   iterator->compare = compare;
   iterator->bucket  = HASH_ITERATOR_INVALID;
   core_list_iterator_init(&iterator->iterator, 0);
}


/*
 *===========================================================================
 *                        core_hash_reference
 *===========================================================================
 *
 * Description
 * -----------
 * Used to reference an explicit entry in the hash when only a pointer to
 * the object is available. Useful for situations where a node exists in
 * multiple storage containers and it would be inefficient to search for
 * the node in each container to get hold of an iterator. 
 */
void
core_hash_reference(core_hash_table_t *hash,
                    core_hash_iterator_t *iterator,
                    core_hash_link_t *link)
{
   uint32_t what_bucket;
   
   if (link->next == NULL || link->prev == NULL)
   {
      return ;
   }

   what_bucket = hash_function(hash, hash->funcs->value(link));
   
   core_list_reference(&hash->bucket[what_bucket], &iterator->iterator, link);
   iterator->hash    = hash;
   iterator->bucket  = what_bucket;
   iterator->key     = NULL;
   iterator->compare = NULL;
}


/*
 *===========================================================================
 *                        core_hash_first
 *===========================================================================
 *
 * Description
 * -----------
 * Retrieves the first entry in the hash that is considered equal to the key
 * by the compare function implemented by the chosen iterator. If the
 * iterator was initialized without compare function, the first entry in the
 * hash is returned.
 * 
 * Returns
 * -------
 * struct core_hash_link_t* Returns the found entry or NULL if it is not found.
 *
 */
core_hash_link_t *
core_hash_first(core_hash_table_t    *hash, 
                core_hash_iterator_t *iterator,
                const void *key)
{
   core_list_link_t *entry;

   iterator->hash   = hash;
   iterator->key    = key;
   iterator->bucket = 0;

   entry = core_list_head(&hash->bucket[iterator->bucket],
                          &iterator->iterator);

   if (entry == NULL ||
       (iterator->compare && !iterator->compare(entry, key)))
   {
      return core_hash_next(iterator);
   }

   return entry;
}

/*
 *===========================================================================
 *                        core_hash_next
 *===========================================================================
 *
 * Description
 * -----------
 * Iterates over the hash by advancing the iterator. The next entry chosen
 * is considered equal by the compare function implemented by the iterator.
 * If the iterator was initialized without compare function, the next entry
 * in the hash is returned.
 *
 * If the entry currently pointed to has been flagged for deletion and no
 * longer is referenced by any iterator, it is removed permanently.
 *
 * Returns
 * -------
 * struct core_hash_link_t* Returns the found entry or NULL if it is not found.
 *
 */
core_hash_link_t *
core_hash_next(core_hash_iterator_t *iterator)
{
   core_list_link_t *entry;
   
   if (iterator->bucket == HASH_ITERATOR_INVALID ||
       iterator->bucket == HASH_ITERATOR_END)
   {
      return NULL;
   }

   entry = core_list_next(&iterator->iterator);
   
   if (entry == NULL)
   {
      iterator->bucket++;
   }

   while (iterator->bucket < iterator->hash->no_of_buckets)
   {
      if (entry == NULL)
      {
         core_list_iterator_release(&iterator->iterator);
         entry = core_list_head(&iterator->hash->bucket[iterator->bucket],
                                &iterator->iterator);
      }

      while (entry != NULL)
      {
         if (!iterator->compare || iterator->compare(entry, iterator->key))
         {
            return entry;
         }

         entry = core_list_next(&iterator->iterator);
      }

      iterator->bucket++;
   }

   if (iterator->bucket < iterator->hash->no_of_buckets)
   {
      iterator->bucket = HASH_ITERATOR_END;
   }
   
   return NULL;
}

/*
 *===========================================================================
 *                        core_hash_insert
 *===========================================================================
 *
 * Description
 * -----------
 * Inserts an entry in the hash_table *hash unless it is already there.
 *
 * Returns
 * -------
 * int      Returns one if the entry was inserted,
 *          or zero is it is already in a hash table.
 *
 */
int
core_hash_insert(core_hash_table_t *hash, core_hash_link_t *entry)
{
   uint32_t what_bucket;
   
   if (entry->next != NULL || entry->prev != NULL)
   {
      return 0;
   }

   what_bucket = hash_function(hash, hash->funcs->value(entry));
   core_list_insert_head(&hash->bucket[what_bucket], entry);
   return 1;
}

/*
 *===========================================================================
 *                        core_hash_iterator_release
 *===========================================================================
 *
 * Description
 * -----------
 * Invalidates an iterator. The iterator dereferences the entry currently
 * pointed to and removes it permanently if it has been flagged for deletion.
 *
 * This function MUST ALWAYS be called when an iterator is no longer needed.
 *
 */
void
core_hash_iterator_release(core_hash_iterator_t *iterator)
{
   iterator->hash    = NULL;
   iterator->key     = NULL;
   iterator->bucket  = HASH_ITERATOR_INVALID;
   core_list_iterator_release(&iterator->iterator);
}


/*
 *===========================================================================
 *                        core_hash_reserve
 *===========================================================================
 *
 * Description
 * -----------
 * Same functionality as core_list_reserve. 
 */
int
core_hash_reserve(core_hash_iterator_t *it, int enable)
{
   return core_list_reserve(&it->iterator, enable);
}


/*
 *===========================================================================
 *                        core_hash_delete
 *===========================================================================
 *
 * Description
 * -----------
 * Deletes an entry in the hash table *hash.
 *
 * Returns
 * -------
 * int     Returns one if the entry was deleted,
 *         or zero if it could not be found.
 *
 */
void
core_hash_delete(core_hash_iterator_t *iter)
{
   core_list_delete(&iter->iterator);
}


/*
 *===========================================================================
 *                        core_hash_delete_direct
 *===========================================================================
 *
 * Description
 * -----------
 * Deletes an entry in the hash table immediately, ignoring the reference
 * counter of the entry. Use with caution. 
 *
 */
void
core_hash_delete_direct(core_hash_table_t *hash, core_hash_link_t *entry)
{
   uint32_t what_bucket;
   
   what_bucket = hash_function(hash, hash->funcs->value(entry));
   core_list_delete_direct(&hash->bucket[what_bucket], entry);
}


#if 0
/* For debug purposes only. */
static int
validate_hash(struct core_hash_table_t *hash)
{
   uint32_t i;
   Msr sr;

   spin_lock_irqsave(&hash->bucket[0].lock, sr);
   for (i = 0; i < hash->no_of_buckets; i++)
   {
      core_hash_link_t *entry = list_head(&hash->bucket[i]);

      while (entry)
      {
         struct page_metadata_t *pgm = (struct page_metadata_t *)entry;

         if (pgm->valid == 0)
         {
            ramlog_printf("Validation of hash failed.\n");
            error2(OSE_PAGING_INTERNAL_ERROR, 0);
            return 0;
         }
         entry = list_next(&hash->bucket[i], entry);
      }
   }
   spin_unlock_irqrestore(&hash->bucket[0].lock, sr);

   return 1;
}
#endif
