/* COPYRIGHT-ENEA-SRC-R2 *
**************************************************************************
* Copyright (C) 2004-2006 by Enea Software AB.
* All rights reserved.
*
* This Software is furnished under a software license agreement and
* may be used only in accordance with the terms of such agreement.
* Any other use or reproduction is prohibited. No title to and
* ownership of the Software is hereby transferred.
*
* PROPRIETARY NOTICE
* This Software consists of confidential information.
* Trade secret law and copyright law protect this Software.
* The above notice of copyright on this Software does not indicate
* any actual or intended publication of such Software.
**************************************************************************
* COPYRIGHT-END */

/* OSE doubly linked list package, main functions.
 */

#include "core_list.h"
#include "assert.h"


#define LINK_DELETED      0x80000000   /* Flag for node deletion */
#define LINK_RESERVED     0x40000000   /* Flag indicating a reserved node */
#define LINK_REFCOUNT     0x3FFFFFFF   /* Mask to get reference count */
#define ITERATOR_INVALID  0x80000000   /* Flag to indicate deleted iterator */


/*
 *===========================================================================
 *                         Private static declarations
 *===========================================================================
 */
static core_list_link_t *
list_next(core_list_t *list, core_list_link_t *link);

static core_list_link_t *
list_prev(core_list_t *list, core_list_link_t *link);

static void
list_delete(core_list_link_t *link);

void
core_list_assert(const core_list_t *list, core_list_assert_func_t check);

/*
 *===========================================================================
 *                         core_list_init
 *===========================================================================
 *
 * Description
 * -----------
 * Initializes a doubly linked list.
 * Needs to be called before any other operation on the list.
 *
 */
void
core_list_init(core_list_t *list, core_list_funcs_t *funcs)
{
   if (list != NULL)
   {
      list->next  = (core_list_link_t *) list;
      list->prev  = (core_list_link_t *) list;
      spin_lock_init(&list->lock);
      list->funcs = funcs;
   }
}

/*
 *===========================================================================
 *                         core_list_link_init
 *===========================================================================
 *
 * Description
 * -----------
 * Initializes a list entry.
 * Needs to be called before the entry is inserted into a list.
 *
 */
void
core_list_link_init(core_list_link_t *lnk)
{
   if (lnk != NULL)
   {
      lnk->next = NULL;
      lnk->prev = NULL;
      lnk->no_refs = 0;
   }
}

/*
 *===========================================================================
 *                         core_list_assert
 *===========================================================================
 *
 * Description
 * -----------
 * Checks that each entry in the list is intact and that the compare function
 * is valid for each entry.
 *
 */
#ifdef ASSERT
void
core_list_assert(const core_list_t *list, core_list_assert_func_t check)
{
   const core_list_link_t *prev;
   const core_list_link_t *next;
   const core_list_link_t *link;

   assert(list != NULL);
   /* Check list head. */
   assert(list->next != NULL);
   assert(list->prev != NULL);

   /* Backward pass. */
   link = list->prev;
   next = (const core_list_link_t *) list;
   while (link != (const core_list_link_t *) list)
   {
      assert(link->next == next);
      next = link;
      link = link->prev;
   }

   /* Forward pass, use user supplied function to check nodes. */
   link = link->next;
   prev = (const core_list_link_t *) list;
   while (link != (const core_list_link_t *) list)
   {
      assert(link->prev == prev);
      if (check)
      {
         check(link);
      }
      prev = link;
      link = link->next;
   }
}
#else
#define core_list_assert(l,f)
#endif

/*
 *===========================================================================
 *                         list_delete
 *===========================================================================
 *
 * Description
 * -----------
 * Removes the specified entry from the list it is in without taking a
 * spin lock or disabling interrupts.
 *
 */
static void
list_delete(core_list_link_t *link)
{
   if (link != NULL)
   {
      assert(link->next != NULL && link->prev != NULL);

      link->next->prev = link->prev;
      link->prev->next = link->next;
      link->next = NULL;
      link->prev = NULL;
   }
}

/*
 *===========================================================================
 *                         core_list_delete_no_lock
 *===========================================================================
 *
 * Description
 * -----------
 * Removes the specified entry from the list it is in without taking a
 * spin lock or disabling interrupts.
 *
 */
void
core_list_delete_no_lock(core_list_iterator_t *iter)
{
   core_list_link_t *successor;
   
   if (iter->current == NULL || (iter->options & ITERATOR_INVALID))
   {
      return;
   }

   successor = iter->current;
   
   if (iter->options & CORE_LIST_ITERATE_REVERSE)
   {
      /* Find the next node the iterator would visit (reverse order) */
      do
      {
         successor = list_prev(iter->list, successor);
      }
      while (successor != NULL && (successor->no_refs & LINK_DELETED));
   }
   else
   {
      /* Find the next node the iterator would visit (in-order) */
      do
      {
         successor = list_next(iter->list, successor);
      }
      while (successor != NULL && (successor->no_refs & LINK_DELETED));
   }

   /* Got the successor node in the
    * iteration series, marked it used. */
   if (successor != NULL)
   {
      successor->no_refs++;
   }
   
   /* The iterator is invalidated and the current
    * node reference count is decremented. */
   iter->current->no_refs--;
   iter->current->no_refs |= LINK_DELETED;

   /* If the resulting reference count is 0, the node
    * is deleted from the list permanently */
   if (!(iter->current->no_refs & LINK_REFCOUNT))
   {
      list_delete(iter->current);
      iter->current->no_refs = 0;
      if (iter->list->funcs && iter->list->funcs->free_link)
      {
         iter->list->funcs->free_link(iter->current);
      }
   }


   /* FIXME: Call node_deleted-event */
   
   /* Update iterator with successor node.
    * User must call next/prev to clear the invalid
    * flag before using the iterator again */
   iter->current  = successor;
   iter->options |= ITERATOR_INVALID;
}

/*
 *===========================================================================
 *                         core_list_delete
 *===========================================================================
 *
 * Description
 * -----------
 * Removes the specified entry from the list it is in.
 *
 */
void
core_list_delete(core_list_iterator_t *iter)
{
   Msr sr;
   core_list_link_t *successor;
   
   if (iter->current == NULL || (iter->options & ITERATOR_INVALID))
   {
      return;
   }

   successor = iter->current;
   
   spin_lock_irqsave(&iter->list->lock, sr);

   if (iter->options & CORE_LIST_ITERATE_REVERSE)
   {
      /* Find the next node the iterator would visit (reverse order) */
      do
      {
         successor = list_prev(iter->list, successor);
      }
      while (successor != NULL && (successor->no_refs & LINK_DELETED));
   }
   else
   {
      /* Find the next node the iterator would visit (in-order) */
      do
      {
         successor = list_next(iter->list, successor);
      }
      while (successor != NULL && (successor->no_refs & LINK_DELETED));
   }

   /* Got the successor node in the
    * iteration series, marked it used. */
   if (successor != NULL)
   {
      successor->no_refs++;
   }
   
   /* The iterator is invalidated and the current
    * node reference count is decremented. */
   iter->current->no_refs--;
   iter->current->no_refs |= LINK_DELETED;

   /* If the resulting reference count is 0, the node
    * is deleted from the list permanently */
   if (!(iter->current->no_refs & LINK_REFCOUNT))
   {
      list_delete(iter->current);
      iter->current->no_refs = 0;
      if (iter->list->funcs && iter->list->funcs->free_link)
      {
         iter->list->funcs->free_link(iter->current);
      }
   }

   spin_unlock_irqrestore(&iter->list->lock, sr);

   /* FIXME: Call node_deleted-event */
   
   /* Update iterator with successor node.
    * User must call next/prev to clear the invalid
    * flag before using the iterator again */
   iter->current  = successor;
   iter->options |= ITERATOR_INVALID;
}

/*
 *===========================================================================
 *                        core_list_reserved_delete
 *===========================================================================
 *
 * Description
 * -----------
 * This function only deletes a node if it is referenced by a single client.
 * If the reference count is higher than one the call aborts and no delayed
 * removal will take place. Failure does not invalidate the iterator.
 *
 * Returns
 * -------
 * Non-zero on success.
 */
int
core_list_reserved_delete(core_list_iterator_t *it)
{
   Msr sr;
   
   if (it->current == NULL || (it->options & ITERATOR_INVALID))
   {
      return 0;
   }
   
   spin_lock_irqsave(&it->list->lock, sr);
   if (it->current->no_refs != 1)
   {
      spin_unlock_irqrestore(&it->list->lock, sr);
      return 0;
   }

   core_list_delete_no_lock(it);

   spin_unlock_irqrestore(&it->list->lock, sr);

   return 1;
}

/*
 *===========================================================================
 *                         core_list_delete_direct
 *===========================================================================
 *
 * Description
 * -----------
 * Removes the specified entry from the list it is in.
 *
 */
void core_list_delete_direct(core_list_t *list, core_list_link_t *link_to_remove)
{
   Msr sr;

   spin_lock_irqsave(&list->lock, sr);
   list_delete(link_to_remove);
   spin_unlock_irqrestore(&list->lock, sr);

   if (list->funcs && list->funcs->free_link)
   {
      list->funcs->free_link(link_to_remove);
   }
}

/*
 *===========================================================================
 *                         core_list_reserve
 *===========================================================================
 *
 * Description
 * -----------
 * Used by a client to reserve the node for exclusive use, for example when
 * the client intends to remove the node from the list. No other iterator
 * will traverse a reserved node. The reservation will fail and 0 is returned
 * if more than one client is referencing the node.
 */
int
core_list_reserve(core_list_iterator_t *it, uint32_t enable)
{
   Msr sr;

   if (it == NULL || it->current == NULL)
   {
      return 0;
   }
   
   spin_lock_irqsave(&it->list->lock, sr);

   /* Node must be referenced by one iterator alone */
   if ((it->current->no_refs & LINK_REFCOUNT) != 1)
   {
      spin_unlock_irqrestore(&it->list->lock, sr);
      return 0;
   }

   /* Everything alright, reserve */
   if (enable)
   {
      it->current->no_refs |= LINK_RESERVED;
   }
   else
   {
      it->current->no_refs &= ~LINK_RESERVED;
   }
   spin_unlock_irqrestore(&it->list->lock, sr);

   return 1;
}

/*
 *===========================================================================
 *                         core_list_insert_head_no_lock
 *===========================================================================
 *
 * Description
 * -----------
 * Inserts the entry first into the list pointed to by 'list' without
 * aquiring any spin lock or disabling interrupts.
 *
 */
void
core_list_insert_head_no_lock(core_list_t *list, core_list_link_t *link_to_insert)
{
   assert(link_to_insert != NULL);

   link_to_insert->next = list->next;
   link_to_insert->prev = (core_list_link_t *)list;
   list->next->prev = link_to_insert;
   list->next = link_to_insert;
}

/*
 *===========================================================================
 *                         core_list_insert_head
 *===========================================================================
 *
 * Description
 * -----------
 * Inserts the entry first into the list pointed to by 'list'.
 *
 */
void
core_list_insert_head(core_list_t *list, core_list_link_t *link_to_insert)
{
   Msr sr;

   spin_lock_irqsave(&list->lock, sr);
   link_to_insert->next = list->next;
   link_to_insert->prev = (core_list_link_t *)list;
   list->next->prev = link_to_insert;
   list->next = link_to_insert;
   spin_unlock_irqrestore(&list->lock, sr);
}

/*
 *===========================================================================
 *                         core_list_insert_tail_no_lock
 *===========================================================================
 *
 * Description
 * -----------
 * Inserts the entry last into the list pointed to by 'list' without
 * aquiring any spin lock or disabling interrupts.
 *
 */
void
core_list_insert_tail_no_lock(core_list_t *list, core_list_link_t *link_to_insert)
{
   assert(link_to_insert != NULL);

   link_to_insert->next = (core_list_link_t *)list;
   link_to_insert->prev = list->prev;
   list->prev->next = link_to_insert;
   list->prev = link_to_insert;
}

/*
 *===========================================================================
 *                         core_list_insert_tail
 *===========================================================================
 *
 * Description
 * -----------
 * Inserts the entry last into the list pointed to by 'list'.
 *
 */
void
core_list_insert_tail(core_list_t *list, core_list_link_t *link_to_insert)
{
   Msr sr;

   spin_lock_irqsave(&list->lock, sr);
   link_to_insert->next = (core_list_link_t *)list;
   link_to_insert->prev = list->prev;
   list->prev->next = link_to_insert;
   list->prev = link_to_insert;
   spin_unlock_irqrestore(&list->lock, sr);
}

/*
 *===========================================================================
 *                         core_list_insert_after_no_lock
 *===========================================================================
 *
 * Description
 * -----------
 * Inserts the entry link_to_insert after the entry link_in_list
 * without aquiring any spin lock or disabling interrupts.
 *
 */
void
core_list_insert_after_no_lock(core_list_iterator_t *iter,
                               core_list_link_t *link_to_insert)
{
   assert(iter != NULL);
   assert(iter->current != NULL);
   assert(!(iter->options & ITERATOR_INVALID));
   assert(iter->current->next != NULL && iter->current->prev != NULL);
   assert(link_to_insert != NULL);

   link_to_insert->next = iter->current->next;
   link_to_insert->prev = iter->current;
   iter->current->next->prev = link_to_insert;
   iter->current->next = link_to_insert;
}

/*
 *===========================================================================
 *                         core_list_insert_after
 *===========================================================================
 *
 * Description
 * -----------
 * Inserts the entry link_to_insert after the entry link_in_list.
 *
 */
void
core_list_insert_after(core_list_t *list,
                       core_list_iterator_t *iter,
                       core_list_link_t *link_to_insert)
{
   Msr sr;

#ifndef OSE_SMP
   (void)list;
#endif
   
   assert(iter != NULL);
   assert(iter->current != NULL);
   assert(!(iter->options & ITERATOR_INVALID));
   assert(iter->current->next != NULL && iter->current->prev != NULL);
   assert(link_to_insert != NULL);

   spin_lock_irqsave(&list->lock, sr);
   link_to_insert->next = iter->current->next;
   link_to_insert->prev = iter->current;
   iter->current->next->prev = link_to_insert;
   iter->current->next = link_to_insert;
   spin_unlock_irqrestore(&list->lock, sr);
}

/*
 *===========================================================================
 *                         core_list_insert_before_no_lock
 *===========================================================================
 *
 * Description
 * -----------
 * Inserts the entry link_to_insert before the entry link_in_list
 * without aquiring any spin lock or disabling interrupts.
 *
 */
void
core_list_insert_before_no_lock(core_list_iterator_t *iter,
                                core_list_link_t *link_to_insert)
{
   assert(iter != NULL);
   assert(iter->current != NULL);
   assert(!(iter->options & ITERATOR_INVALID));
   assert(iter->current->next != NULL && iter->current->prev != NULL);
   assert(link_to_insert != NULL);

   link_to_insert->prev = iter->current->prev;
   link_to_insert->next = iter->current;
   iter->current->prev->next = link_to_insert;
   iter->current->prev = link_to_insert;
}

/*
 *===========================================================================
 *                         core_list_insert_before
 *===========================================================================
 *
 * Description
 * -----------
 * Inserts the entry link_to_insert before the entry link_in_list.
 *
 */
void
core_list_insert_before(core_list_t *list,
                        core_list_iterator_t *iter,
                        core_list_link_t *link_to_insert)
{
   Msr sr;

#ifndef OSE_SMP
   (void)list;
#endif

   assert(iter != NULL);
   assert(iter->current != NULL);
   assert(!(iter->options & ITERATOR_INVALID));
   assert(iter->current->next != NULL && iter->current->prev != NULL);
   assert(link_to_insert != NULL);

   spin_lock_irqsave(&list->lock, sr);
   link_to_insert->prev = iter->current->prev;
   link_to_insert->next = iter->current;
   iter->current->prev->next = link_to_insert;
   iter->current->prev = link_to_insert;
   spin_unlock_irqrestore(&list->lock, sr);
}


/*
 *===========================================================================
 *                         core_list_length
 *===========================================================================
 *
 * Description
 * -----------
 * Returns the number of entries in the list.
 *
 */
uint32_t
core_list_length(core_list_t *list)
{
   uint32_t count = 0;
   core_list_link_t *link;
   core_list_iterator_t iter;

   core_list_assert(list, NULL);

   core_list_iterator_init(&iter, 0);

   link = core_list_head(list, &iter);
   while (link != NULL)
   {
      count++;
      link = core_list_next(&iter);
   }
   core_list_iterator_release(&iter);
   
   return count;
}


/*
 *===========================================================================
 *                         core_list_insert_ordered
 *===========================================================================
 *
 * Description
 * -----------
 * Inserts the entry ll before the entry for which the compare function is
 * positive.
 *
 */
void
core_list_insert_ordered(core_list_t *list,
                         core_list_link_t *link_to_insert,
                         core_list_insert_ordered_func_t compare)
{
   core_list_link_t *link;
   core_list_iterator_t iter;

   assert(link_to_insert != NULL);

   core_list_iterator_init(&iter, 0);

   link = core_list_head(list, &iter);
   while (link != NULL)
   {
      if (compare(link_to_insert, link) < 0)
      {
         break; /* don't break if equal value, skip past */
      }
      link = core_list_next(&iter);
   }
   
   /* new link is always inserted AFTER equally valued links */
   if (link != NULL)
   {
      core_list_insert_before(list, &iter, link_to_insert);
   }
   else
   {
      core_list_insert_tail(list, link_to_insert);
   }

   core_list_iterator_release(&iter);
   
   assert((link_to_insert->next != NULL) && (link_to_insert->prev != NULL));
}


/*
 *===========================================================================
 *                         core_list_traverse
 *===========================================================================
 *
 * Description
 * -----------
 * Traverses the list and calls the apply function for each entry.
 *
 */
void *
core_list_traverse(core_list_t *list, core_list_traverse_func_t apply, void *arg)
{
   core_list_link_t *current;
   core_list_iterator_t iter;

   core_list_iterator_init(&iter, 0);
   
   current = core_list_head(list, &iter);
   while (current != NULL)
   {
      core_list_link_t *next = core_list_next(&iter);
      void *stp = apply(current, arg);

      if (stp != NULL)
      {
         return stp;
      }
      current = next;
   }

   core_list_iterator_release(&iter);
   return NULL;
}


/*
 *===========================================================================
 *                         core_list_head
 *===========================================================================
 *
 * Description
 * -----------
 * Returns the first entry in the list.
 *
 */
core_list_link_t *
core_list_head(core_list_t *list, core_list_iterator_t *iter)
{
   Msr sr;
   core_list_link_t *next;
   
   iter->list = list;

   spin_lock_irqsave(&list->lock, sr);

   next = (core_list_link_t*)list;

   /* Traverse list from end until a node is found which
    * is not flagged for deletion, or end-of-list is reached. */
   do
   {
      next = list_next(list, next);
   }
   while (next != NULL && (next->no_refs & LINK_DELETED));

   if (next != NULL)
   {
      next->no_refs++;
   }

   spin_unlock_irqrestore(&list->lock, sr);

   iter->options &= ~ITERATOR_INVALID;
   iter->current  = next;
   return next;
}

/*
 *===========================================================================
 *                         core_list_tail
 *===========================================================================
 *
 * Description
 * -----------
 * Returns the last entry in the list.
 *
 */
core_list_link_t *
core_list_tail(core_list_t *list, core_list_iterator_t *iter)
{
   Msr sr;
   core_list_link_t *prev;

   iter->list = list;
   
   spin_lock_irqsave(&list->lock, sr);

   prev = (core_list_link_t*)list;

   /* Traverse list from end until a node is found which
    * is not flagged for deletion, or end-of-list is reached. */
   do
   {
      prev = list_prev(list, prev);
   }
   while (prev != NULL && (prev->no_refs & LINK_DELETED));

   if (prev != NULL)
   {
      prev->no_refs++;
   }

   spin_unlock_irqrestore(&list->lock, sr);

   iter->options &= ~ITERATOR_INVALID;
   iter->current  = prev;
   return prev;
}


/*
 *===========================================================================
 *                         list_next
 *===========================================================================
 *
 * Description
 * -----------
 * Returns the next entry in the list.
 *
 */
static core_list_link_t *
list_next(core_list_t *list, core_list_link_t *link)
{
   if (link->next != (core_list_link_t*)list)
   {
      return link->next;
   }
   return NULL;
}



/*
 *===========================================================================
 *                         list_prev
 *===========================================================================
 *
 * Description
 * -----------
 * Returns the previous entry in the list.
 *
 */
static core_list_link_t *
list_prev(core_list_t *list, core_list_link_t *link)
{
   if (link->prev != (core_list_link_t*)list)
   {
      return link->prev;
   }
   return NULL;
}

/*
 *===========================================================================
 *                         core_list_iterator_init
 *===========================================================================
 *
 * Description
 * -----------
 * Initializes a list iterator, pointing to the node passed as argument.
 *
 * A node pointed to by an iterator is guaranteed to exist as long it is
 * referenced by any iterator. Deleted nodes will not be removed from the
 * list until all references to the node have disappeared. New iterators will
 * not discover deleted nodes which have not yet been removed. 
 *
 * The iterator must be released, by calling core_list_iterator_release,
 * when it is no longer needed. The third parameter specifies if the
 * iterator traverses the list forwards or backwards.
 */
void
core_list_iterator_init(core_list_iterator_t *iter, uint32_t options)
{
   iter->list    = NULL;
   iter->current = NULL;
   iter->options = options | ITERATOR_INVALID;
}

/*
 *===========================================================================
 *                        core_list_reference
 *===========================================================================
 *
 * Description
 * -----------
 * Used to reference an explicit entry in the hash when only a pointer to
 * the object is available. Useful for situations where a node exists in
 * multiple storage containers and it would be inefficient to search for
 * the node in each container to get hold of an iterator. 
 */
void
core_list_reference(core_list_t *list,
                    core_list_iterator_t *iter,
                    core_list_link_t *link)
{
   Msr sr;
   
   if (link->next == NULL || link->prev == NULL)
   {
      return;
   }

   spin_lock_irqsave(&list->lock, sr);

   link->no_refs++;

   spin_unlock_irqrestore(&list->lock, sr);

   iter->list = list;
   iter->options &= ~ITERATOR_INVALID;
   iter->current  = link;
}

/*
 *===========================================================================
 *                         core_list_iterator_node
 *===========================================================================
 *
 * Description
 * -----------
 * Used to retrieve the node currently pointed to by the iterator. It is not
 * legal to reference the node without the iterator object. 
 */
core_list_link_t *
core_list_iterator_node(core_list_iterator_t *iter)
{
   if (iter == NULL || (iter->options & ITERATOR_INVALID))
   {
      return NULL;
   }
   return iter->current;
}

/*
 *===========================================================================
 *                         core_list_next
 *===========================================================================
 *
 * Description
 * -----------
 * Advances the iterator to the next node in the list. The node is returned
 * unless end-of-list is reached (NULL). Using iterator_next on reverse-order
 * iterators is not legal and NULL is returned.
 */
core_list_link_t *
core_list_next(core_list_iterator_t *iter)
{
   Msr sr;
   core_list_link_t *next;

   if (iter->options & CORE_LIST_ITERATE_REVERSE)
   {
      return NULL;
   }

   if (iter->current == NULL)
   {
      return NULL;
   }

   spin_lock_irqsave(&iter->list->lock, sr);

   /* "Next" on an invalid iterator simply means
    * that it is validated. However, if the node it
    * refers to was deleted we need to go find a new. */
   if ((iter->options & ITERATOR_INVALID))
   {
      iter->options &= ~ITERATOR_INVALID;

      if (!(iter->current->no_refs & LINK_DELETED))
      {
         spin_unlock_irqrestore(&iter->list->lock, sr);
         return iter->current;
      }
   }

   iter->current->no_refs--;
   iter->current->no_refs &= ~LINK_RESERVED;
   
   /* Locate the next node in the list which has not been
    * flagged for deletion */
   next = iter->current;
   do
   {
      next = list_next(iter->list, next);
   }
   while (next != NULL && (next->no_refs & (LINK_DELETED|LINK_RESERVED)));

   /* If current node is flagged for deletion and
    * has no more references, delete it for good. */
   if ((iter->current->no_refs & (LINK_DELETED|LINK_REFCOUNT)) == LINK_DELETED)
   {
      list_delete(iter->current);
      if (iter->list->funcs && iter->list->funcs->free_link)
      {
         iter->list->funcs->free_link(iter->current);
      }
   }

   if (next != NULL)
   {
      next->no_refs++;
   }
       
   spin_unlock_irqrestore(&iter->list->lock, sr);

   iter->current = next;
   return next;   
}

/*
 *===========================================================================
 *                         core_list_prev
 *===========================================================================
 *
 * Description
 * -----------
 * Advances the iterator to the previous node in the list. The node is returned
 * unless end-of-list is reached (NULL). Using iterator_prev on in-order
 * iterators is not legal and NULL is returned.
 */
core_list_link_t *
core_list_prev(core_list_iterator_t *iter)
{
   Msr sr;
   core_list_link_t *prev;

   if (!(iter->options & CORE_LIST_ITERATE_REVERSE))
   {
      return NULL;
   }

   if (iter->current == NULL)
   {
      return NULL;
   }

   spin_lock_irqsave(&iter->list->lock, sr);

   /* "Next" on an invalid iterator simply means
    * that it is validated. However, if the node it
    * refers to was deleted we need to go find a new. */
   if ((iter->options & ITERATOR_INVALID))
   {
      iter->options &= ~ITERATOR_INVALID;

      if (!(iter->current->no_refs & LINK_DELETED))
      {
         spin_unlock_irqrestore(&iter->list->lock, sr);
         return iter->current;
      }
   }

   iter->current->no_refs--;
   iter->current->no_refs &= ~LINK_RESERVED;

   /* Locate the next node in the list which has not been
    * flagged for deletion */
   prev = iter->current;
   do
   {
      prev = list_prev(iter->list, prev);
   }
   while (prev != NULL && !(prev->no_refs & (LINK_DELETED|LINK_RESERVED)));
          
   /* If the current node is flagged for deletion and has
    * no more references, delete it for good. */
   if ((iter->current->no_refs & (LINK_DELETED|LINK_RESERVED)) == LINK_DELETED)
   {
      list_delete(iter->current);
      if (iter->list->funcs && iter->list->funcs->free_link)
      {
         iter->list->funcs->free_link(iter->current);
      }
   }

   if (prev != NULL)
   {
      prev->no_refs++;
   }
       
   spin_unlock_irqrestore(&iter->list->lock, sr);

   iter->current = prev;
   return prev;   
}

/*
 *===========================================================================
 *                         core_list_iterator_release
 *===========================================================================
 *
 * Description
 * -----------
 * Destroys an iterator. This function must always be called when an iterator
 * is no longer needed. If the iterator references a deleted node, it will
 * remove the node from the list if it is the last refering iterator.
 */
void
core_list_iterator_release(core_list_iterator_t *iter)
{
   Msr sr;

   if (iter->list == NULL || iter->current == NULL)
   {
      iter->list     = NULL;
      iter->current  = NULL;
      iter->options |= ITERATOR_INVALID;
      return;
   }
   
   spin_lock_irqsave(&iter->list->lock, sr);
   
   iter->current->no_refs--;
   iter->current->no_refs &= ~LINK_RESERVED;

   /* If the current node is flagged for deletion and has
    * no more references, delete it for good. */
   if ((iter->current->no_refs & LINK_DELETED) == LINK_DELETED)
   {
      list_delete(iter->current);
      if (iter->list->funcs && iter->list->funcs->free_link)
      {
         iter->list->funcs->free_link(iter->current);
      }
   }

   spin_unlock_irqrestore(&iter->list->lock, sr);

   iter->list     = NULL;
   iter->current  = NULL;
   iter->options |= ITERATOR_INVALID;
}
