/* COPYRIGHT-ENEA-SRC-R2 *
**************************************************************************
* Copyright (C) 2004-2008 by Enea Software AB.
* All rights reserved.
*
* This Software is furnished under a software license agreement and
* may be used only in accordance with the terms of such agreement.
* Any other use or reproduction is prohibited. No title to and
* ownership of the Software is hereby transferred.
*
* PROPRIETARY NOTICE
* This Software consists of confidential information.
* Trade secret law and copyright law protect this Software.
* The above notice of copyright on this Software does not indicate
* any actual or intended publication of such Software.
**************************************************************************
* COPYRIGHT-END */

#include "core_rbtree.h"
#include <assert.h>
#include <stdlib.h>


/*
 *			    ASSERT MACROS
 */
#ifdef NDEBUG
#define assert_tree(t,n)  do {} while(0)
#else
static void assert_tree(core_rbtree_t *tree,
                        const core_rbtree_node_t *not_in_tree);
#endif

#define MAX(a,b) (a>b?a:b)

/*
 *			 TREE/NODE SELECTORS
 */
#define ToL(x)            ((unsigned long)(x))    /* ToLong */
#define ToN(x)            ((core_rbtree_node_t*)(x))      /* ToNode */
#define ToPointer(x)      (ToL(x) & ~0x1ul)
#define ToColor(x)        (ToL(x) & 0x1ul)
#define IsRed(x)          (ToColor((x)->left) == 0x1ul)
#define IsBlack(x)        (ToColor((x)->left) == 0x0ul)
#define ColorRed(x)       do { (x)->left = ToN(ToPointer((x)->left) | 0x1ul); \
                          } while(0)
#define ColorBlack(x)     do { (x)->left = ToN(ToPointer((x)->left)); \
                          } while(0)
#define ColorSame(x,y)    do { (x)->left = ToN(ToPointer((x)->left) | \
                                               ToColor((y)->left)); \
                          } while(0)
#define IsRoot(t,x)       ((x)->parent == NIL(tree))
#define Parent(x)         ((x)->parent)
#define AssignLeft(x,y)   do { (x)->left = ToN(ToColor((x)->left) | ToL(y)); \
                          } while(0)
#define AssignRight(x,y)  do { (x)->right  = (y); } while(0)
#define AssignParent(x,y) do { (x)->parent = (y); } while(0)
#define Left(x)           (ToN(ToPointer((x)->left))) /* rvalue only */
#define Right(x)          ((x)->right)
#define IsLeftChild(x)    (ToN(ToPointer((x)->parent->left)) == (x))
#define IsRightChild(x)   ((x)->parent->right == (x))

#undef NIL
#define NIL(t)            (&(t)->nil)


/*
 *		      LOCAL FUNCTION PROTOTYPES
 */
static void rotate_left(core_rbtree_t *tree, core_rbtree_node_t *x);
static void rotate_right(core_rbtree_t *tree, core_rbtree_node_t *x);
static void delete_fixup(core_rbtree_t *tree, core_rbtree_node_t *x);


/*
 *---------------------------------------------------------------------------
 *				rotate_left
 *---------------------------------------------------------------------------
 */
static void
rotate_left(core_rbtree_t *tree, core_rbtree_node_t *x)
{
   core_rbtree_node_t *y;

   assert(x != NIL(tree));
   assert(Right(x) != NIL(tree));

   y = Right(x);

   /* establish x's right link */
   AssignRight(x, Left(y));

   if (Left(y) != NIL(tree))
      AssignParent(Left(y), x);

   AssignParent(y, Parent(x));
   if (IsRoot(tree, x))
   {
      tree->root = y;
   }
   else if (IsLeftChild(x))
   {
      AssignLeft(Parent(x), y);
   }
   else /* IsRightChild(x) */
   {
      AssignRight(Parent(x), y);
   }

   /* link x and y */
   AssignLeft(y, x);
   AssignParent(x, y);
}


/*
 *---------------------------------------------------------------------------
 *			       rotate_right
 *---------------------------------------------------------------------------
 */
static void
rotate_right(core_rbtree_t *tree, core_rbtree_node_t *x)
{
   core_rbtree_node_t *y;

   assert(x != NIL(tree));
   assert(Left(x) != NIL(tree));

   y = Left(x);

   /* establish x's left link */
   AssignLeft(x, Right(y));

   if (Right(y) != NIL(tree))
      AssignParent(Right(y), x);

   AssignParent(y, Parent(x));
   if (IsRoot(tree, x))
   {
      tree->root = y;
   }
   else if (IsLeftChild(x))
   {
      AssignLeft(Parent(x), y);
   }
   else /* IsRightChild(x) */
   {
      AssignRight(Parent(x), y);
   }

   /* link x and y */
   AssignRight(y, x);
   AssignParent(x, y);
}



void
core_rbtree_init(core_rbtree_t *tree, rbtree_compare_func_t compare)
{
   /* Init special NIL node */
   tree->nil.parent = NIL(tree);
   tree->nil.left = NIL(tree);
   tree->nil.right = NIL(tree);

   tree->root = NIL(tree);
   tree->compare = compare;
   spin_lock_init(&tree->lock);
}


/******************************************************************************
 * LOCKED RBTREE FUNCTIONS
 *****************************************************************************/

int
core_rbtree_insert(core_rbtree_t *tree, void *node_ptr)
{
   core_rbtree_node_t *current, *parent;
   core_rbtree_node_t *node = (core_rbtree_node_t*)node_ptr;
   int cmp;
   Msr msr = 0;
   
   (void)msr;

   AssignLeft(node,  NIL(tree));
   AssignRight(node, NIL(tree));

   /* Locate node to insert below (the new parent) */
   parent = NIL(tree);
   current = tree->root;

   spin_lock_irqsave(&tree->lock, msr);
   
   while (current != NIL(tree))
   {
      parent = current;

      cmp = tree->compare(node, current);

      if (cmp == 0)
      {
	 /* Nodes equal => Reject duplicate! */
         spin_unlock_irqrestore(&tree->lock, msr);
	 return -1;
      }
      else if (cmp < 0)
      {
	 /* (node < current) => Go Left */
	 current = Left(current);
      }
      else
      {
	 /* (node > current) => Go right */
	 current = Right(current);
      }
   }

   AssignParent(node, parent);
   ColorRed(node);

   if (parent == NIL(tree))
   {
      tree->root = node;
   }
   else if (tree->compare(node, parent) < 0)
   {
      AssignLeft(parent, node);
   }
   else
   {
      AssignRight(parent, node);
   }

   
   /* Check Red-Black properties */
   while (!IsRoot(tree, node) && IsRed(Parent(node)))
   {
      if (IsLeftChild(Parent(node)))
      {
	 core_rbtree_node_t *y;

	 y = Right(Parent(Parent(node)));

	 if (IsRed(y))
	 {
	    ColorBlack(Parent(node));
	    ColorBlack(y);
	    node = Parent(Parent(node));
	    ColorRed(node);
	 }
	 else
	 {
	    if (IsRightChild(node))
	    {
	       node = Parent(node);
	       rotate_left(tree, node);
	    }

	    ColorBlack(Parent(node));
	    ColorRed(Parent(Parent(node)));
	    rotate_right(tree, Parent(Parent(node)));
	 }
      }
      else /* IsRightChild(Parent(node)) */
      {
	 core_rbtree_node_t *y;

	 y = Left(Parent(Parent(node)));

	 if (IsRed(y))
	 {
	    ColorBlack(Parent(node));
	    ColorBlack(y);
	    node = Parent(Parent(node));
	    ColorRed(node);
	 }
	 else
	 {
	    if (IsLeftChild(node))
	    {
	       node = Parent(node);
               rotate_right(tree, node);
	    }

	    ColorBlack(Parent(node));
	    ColorRed(Parent(Parent(node)));
	    rotate_left(tree, Parent(Parent(node)));
	 }
      }
   }
   ColorBlack(tree->root);

   spin_unlock_irqrestore(&tree->lock, msr);

   assert_tree(tree, NULL);
   return 0;
}


core_rbtree_node_t *
core_rbtree_remove(core_rbtree_t *tree, void *node_ptr)
{
   core_rbtree_node_t *x, *y;
   core_rbtree_node_t *node = (core_rbtree_node_t*)node_ptr;
   int do_fixup;
   Msr msr = 0;

   (void)msr;

   spin_lock_irqsave(&tree->lock, msr);
   
   if (Left(node) == NIL(tree) || Right(node) == NIL(tree))
   {
      y = node;
   }
   else
   {
      y = core_rbtree_next_no_lock(tree, node);
      assert(y != node);
      assert(y != NIL(tree));
   }
   
   assert(Left(y) == NIL(tree) || Right(y) == NIL(tree));

   /* x is y's only child */
   if ((x = Left(y)) == NIL(tree))
   {
      x = Right(y);
   }

   /*
    * Unlink y
    */
   AssignParent(x, Parent(y));

   if (IsRoot(tree, y))
   {
      tree->root = x;
   }
   else if (IsLeftChild(y))
   {
      AssignLeft(Parent(y), x);
   }
   else
   {
      AssignRight(Parent(y), x);
   }

   if (tree->root != NIL(tree))
   {
      do_fixup = IsBlack(y);
   }
   else
   {
      do_fixup = 0;
   }

   if (y != node)
   {
      assert(y != NIL(tree));
      assert(node != NIL(tree));

      /*
       * Swap 'node' and 'y'
       */
      AssignParent(y, Parent(node));
      if (IsRoot(tree, node))
      {
	 tree->root = y;
      }
      else if (IsLeftChild(node))
      {
	 AssignLeft(Parent(node), y);
      }
      else
      {
	 AssignRight(Parent(node), y);
      }

      AssignParent(Left(node), y);
      AssignLeft(y, Left(node));

      AssignParent(Right(node), y);
      AssignRight(y, Right(node));

      ColorSame(y, node);
   }

   if (do_fixup)
   {
      delete_fixup(tree, x);
   }

   spin_unlock_irqrestore(&tree->lock, msr);

   assert_tree(tree, node);

   node->parent = NULL;
   node->left = NULL;
   node->right = NULL;
   
   return node;
}


/*
 *---------------------------------------------------------------------------
 *			   delete_fixup
 *---------------------------------------------------------------------------
 * Maintain Red-Black tree properties after delete operation.
 *
 */
static void
delete_fixup(core_rbtree_t *tree, core_rbtree_node_t *x)
{
   core_rbtree_node_t *w;
   Msr msr = 0;

   (void)msr;

   while (x != tree->root && IsBlack(x))
   {
      assert(Parent(x) != NIL(tree));

      if (IsLeftChild(x))
      {
	 /* w is x's sibling */
	 w = Right(Parent(x));

	 if (IsRed(w))
	 {
	    ColorBlack(w);
	    ColorRed(Parent(x));
	    rotate_left (tree, Parent(x));
	    w = Right(Parent(x));
	 }

	 if (IsBlack(Left(w)) && IsBlack(Right(w)))
	 {
	    ColorRed(w);
	    x = Parent(x);
	 }
	 else
	 {
	    if (IsBlack(Right(w)))
	    {
	       ColorBlack(Left(w));
	       ColorRed(w);
	       rotate_right (tree, w);
	       w = Right(Parent(x));
	    }

	    ColorSame(w, Parent(x));
	    ColorBlack(Parent(x));
	    ColorBlack(Right(w));
	    rotate_left (tree, Parent(x));
	    break;
	 }
      }
      else /* IsRightChild(x) */
      {
	 w = Left(Parent(x));

	 if (IsRed(w))
	 {
	    ColorBlack(w);
	    ColorRed(Parent(x));
	    rotate_right (tree, Parent(x));
	    w = Left(Parent(x));
	 }

	 if (IsBlack(Left(w)) && IsBlack(Right(w)))
	 {
	    ColorRed(w);
	    x = Parent(x);
	 }
	 else
	 {
	    if (IsBlack(Left(w)))
	    {
	       ColorBlack(Right(w));
	       ColorRed(w);
	       rotate_left (tree, w);
	       w = Left(Parent(x));
	    }

	    ColorSame(w, Parent(x));
	    ColorBlack(Parent(x));
	    ColorBlack(Left(w));
	    rotate_right (tree, Parent(x));
	    break;
	 }
      }
   }
   ColorBlack(x);
}


core_rbtree_node_t *
core_rbtree_find(core_rbtree_t *tree,
                   rbtree_compare_func_t compare,
                   const void *key)
{
   core_rbtree_node_t *current = tree->root;
   int cmp;
   Msr msr = 0;

   (void)msr;

   spin_lock_irqsave(&tree->lock, msr);

   while (current != NIL(tree))
   {
      cmp = compare(key, current);
      if (cmp == 0)
      {
	 /* Match */
         spin_unlock_irqrestore(&tree->lock, msr);
	 return current;
      }
      else if (cmp < 0)
      {
	 current = Left(current);
      }
      else
      {
	 current = Right(current);
      }
   }

   spin_unlock_irqrestore(&tree->lock, msr);

   return NULL;
}

core_rbtree_node_t *
core_rbtree_find_best(core_rbtree_t *tree,
                      rbtree_compare_best_func_t compare,
                      const void *key)
{
   core_rbtree_node_t *current = tree->root;
   core_rbtree_node_t *best = NULL;
   int cmp;
   Msr msr = 0;

   spin_lock_irqsave(&tree->lock, msr);

   while (current != NIL(tree))
   {
      cmp = compare(key, current, best);
      if (cmp == 0)
      {
	 /* Exact match */
         spin_unlock_irqrestore(&tree->lock, msr);
         return current;
      }
      else if (cmp == -1)
      {
	 current = Left(current);
      }
      else if (cmp == 1)
      {
	 current = Right(current);
      }
      else if (cmp < -1)
      {
         /* Callback has found a possible match, but
          * wishes to continue search in the left subtree. */
         best = current;
         current = Left(current);
      }
      else
      {
         /* Callback has found a possible match, but
          * wishes to continue search in the right subtree. */
         best = current;
         current = Right(current);
      }
   }

   spin_unlock_irqrestore(&tree->lock, msr);

   return best;
}

core_rbtree_node_t *
core_rbtree_first(core_rbtree_t *tree)
{
   core_rbtree_node_t *node;
   Msr msr = 0;

   (void)msr;

   spin_lock_irqsave(&tree->lock, msr);

   if ((node = tree->root) == NIL(tree))
   {
      spin_unlock_irqrestore(&tree->lock, msr);
      return NULL;
   }
   else
   {
      while (Left(node) != NIL(tree))
      {
	 node = Left(node);
      }
   }

   spin_unlock_irqrestore(&tree->lock, msr);

   return node;
}


core_rbtree_node_t *
core_rbtree_next(core_rbtree_t *tree, void *node_ptr)
{
   core_rbtree_node_t *node = (core_rbtree_node_t*)node_ptr;
   Msr msr = 0;

   (void)msr;

   spin_lock_irqsave(&tree->lock, msr);

   if (Right(node) != NIL(tree))
   {
      /* go right 1... */
      node = Right(node);

      /* ...then left to the end */
      while (Left(node) != NIL(tree))
      {
	 node = Left(node);
      }
   }
   else
   {
      /* while you're the right child, chain up parent link */
      core_rbtree_node_t *p;

      p = Parent(node);

      while (p != NIL(tree) && node == Right(p))
      {
	 node = p;
	 p = Parent(p);
      }

      if (p == NIL(tree))
      {
	 node = NULL;
      }
      else
      {
	 node = p;
      }
   }
   
   spin_unlock_irqrestore(&tree->lock, msr);

   return (core_rbtree_node_t *)node;
}

core_rbtree_node_t *
core_rbtree_last(core_rbtree_t *tree)
{
   core_rbtree_node_t *node;
   Msr msr = 0;

   (void)msr;

   spin_lock_irqsave(&tree->lock, msr);

   if ((node = tree->root) == NIL(tree))
   {
      return NULL;
   }
   else
   {
      while (Right(node) != NIL(tree))
      {
	 node = Right(node);
      }
   }

   spin_unlock_irqrestore(&tree->lock, msr);

   return node;
}


core_rbtree_node_t *
core_rbtree_prev(core_rbtree_t *tree, void *node_ptr)
{
   core_rbtree_node_t *node = (core_rbtree_node_t*)node_ptr;
   Msr msr = 0;

   (void)msr;

   spin_lock_irqsave(&tree->lock, msr);

   if (Left(node) != NIL(tree))
   {
      /* go right 1... */
      node = Left(node);

      /* ...then right to the end */
      while (Right(node) != NIL(tree))
      {
	 node = Right(node);
      }
   }
   else
   {
      /* while you're the right child, chain up parent link */
      core_rbtree_node_t *p;

      p = Parent(node);

      while (p != NIL(tree) && node == Left(p))
      {
	 node = p;
	 p = Parent(p);
      }

      if (p == NIL(tree))
      {
	 node = NULL;
      }
      else
      {
	 node = p;
      }
   }
   
   spin_unlock_irqrestore(&tree->lock, msr);

   return (core_rbtree_node_t *)node;
}

void
core_rbtree_traverse(core_rbtree_t *tree,
                       rbtree_traverse_func_t func)
{
   core_rbtree_node_t *current;
   Msr msr = 0;

   (void)msr;

   spin_lock_irqsave(&tree->lock, msr);

   for (current = core_rbtree_first_no_lock(tree);
        current != (void*)0;
        current = core_rbtree_next_no_lock(tree, current))
   {
      func(tree, current);
   }

   spin_unlock_irqrestore(&tree->lock, msr);   
}


/******************************************************************************
 * UNLOCKED RBTREE FUNCTIONS
 *****************************************************************************/

int
core_rbtree_insert_no_lock(core_rbtree_t *tree, void *node_ptr)
{
   core_rbtree_node_t *current, *parent;
   core_rbtree_node_t *node = (core_rbtree_node_t*)node_ptr;
   int cmp;

   AssignLeft(node,  NIL(tree));
   AssignRight(node, NIL(tree));

   /* Locate node to insert below (the new parent) */
   parent = NIL(tree);
   current = tree->root;

   while (current != NIL(tree))
   {
      parent = current;

      cmp = tree->compare(node, current);

      if (cmp == 0)
      {
	 /* Nodes equal => Reject duplicate! */
	 return -1;
      }
      else if (cmp < 0)
      {
	 /* (node < current) => Go Left */
	 current = Left(current);
      }
      else
      {
	 /* (node > current) => Go right */
	 current = Right(current);
      }
   }

   AssignParent(node, parent);
   ColorRed(node);

   if (parent == NIL(tree))
   {
      tree->root = node;
   }
   else if (tree->compare(node, parent) < 0)
   {
      AssignLeft(parent, node);
   }
   else
   {
      AssignRight(parent, node);
   }

   /* Check Red-Black properties */
   while (!IsRoot(tree, node) && IsRed(Parent(node)))
   {
      if (IsLeftChild(Parent(node)))
      {
	 core_rbtree_node_t *y;

	 y = Right(Parent(Parent(node)));

	 if (IsRed(y))
	 {
	    ColorBlack(Parent(node));
	    ColorBlack(y);
	    node = Parent(Parent(node));
	    ColorRed(node);
	 }
	 else
	 {
	    if (IsRightChild(node))
	    {
	       node = Parent(node);
	       rotate_left(tree, node);
	    }

	    ColorBlack(Parent(node));
	    ColorRed(Parent(Parent(node)));

	    rotate_right(tree, Parent(Parent(node)));
	 }
      }
      else /* IsRightChild(Parent(node)) */
      {
	 core_rbtree_node_t *y;

	 y = Left(Parent(Parent(node)));

	 if (IsRed(y))
	 {
	    ColorBlack(Parent(node));
	    ColorBlack(y);
	    node = Parent(Parent(node));
	    ColorRed(node);
	 }
	 else
	 {
	    if (IsLeftChild(node))
	    {
	       node = Parent(node);

               rotate_right(tree, node);
	    }

	    ColorBlack(Parent(node));
	    ColorRed(Parent(Parent(node)));
            
	    rotate_left(tree, Parent(Parent(node)));
	 }
      }
   }
   ColorBlack(tree->root);

   assert_tree(tree, NULL);
   return 0;
}


core_rbtree_node_t *
core_rbtree_remove_no_lock(core_rbtree_t *tree, void *node_ptr)
{
   core_rbtree_node_t *x, *y;
   core_rbtree_node_t *node = (core_rbtree_node_t*)node_ptr;
   int do_fixup;

   if (Left(node) == NIL(tree) || Right(node) == NIL(tree))
   {
      y = node;
   }
   else
   {
      y = core_rbtree_next_no_lock(tree, node);
      assert(y != node);
      assert(y != NIL(tree));
   }
   
   assert(Left(y) == NIL(tree) || Right(y) == NIL(tree));

   /* x is y's only child */
   if ((x = Left(y)) == NIL(tree))
   {
      x = Right(y);
   }

   /*
    * Unlink y
    */
   AssignParent(x, Parent(y));

   if (IsRoot(tree, y))
   {
      tree->root = x;
   }
   else if (IsLeftChild(y))
   {
      AssignLeft(Parent(y), x);
   }
   else
   {
      AssignRight(Parent(y), x);
   }

   if (tree->root != NIL(tree))
   {
      do_fixup = IsBlack(y);
   }
   else
   {
      do_fixup = 0;
   }

   if (y != node)
   {
      assert(y != NIL(tree));
      assert(node != NIL(tree));

      /*
       * Swap 'node' and 'y'
       */
      AssignParent(y, Parent(node));
      if (IsRoot(tree, node))
      {
	 tree->root = y;
      }
      else if (IsLeftChild(node))
      {
	 AssignLeft(Parent(node), y);
      }
      else
      {
	 AssignRight(Parent(node), y);
      }

      AssignParent(Left(node), y);
      AssignLeft(y, Left(node));

      AssignParent(Right(node), y);
      AssignRight(y, Right(node));

      ColorSame(y, node);
   }

   if (do_fixup)
   {
      delete_fixup(tree, x);
   }

   assert_tree(tree, node);

   node->parent = NULL;
   node->left = NULL;
   node->right = NULL;

   return node;
}


core_rbtree_node_t *
core_rbtree_find_no_lock(core_rbtree_t *tree,
                   rbtree_compare_func_t compare,
                   const void *key)
{
   core_rbtree_node_t *current = tree->root;
   int cmp;

   while (current != NIL(tree))
   {
      cmp = compare(key, current);
      if (cmp == 0)
      {
	 /* Match */
	 return current;
      }
      else if (cmp < 0)
      {
	 current = Left(current);
      }
      else
      {
	 current = Right(current);
      }
   }

   return NULL;
}


core_rbtree_node_t *
core_rbtree_find_best_no_lock(core_rbtree_t *tree,
                              rbtree_compare_best_func_t compare,
                              const void *key)
{
   core_rbtree_node_t *current = tree->root;
   core_rbtree_node_t *best = NULL;
   int cmp;

   while (current != NIL(tree))
   {
      cmp = compare(key, current, best);
      if (cmp == 0)
      {
	 /* Exact match */
	 return current;
      }
      else if (cmp == -1)
      {
	 current = Left(current);
      }
      else if (cmp == 1)
      {
	 current = Right(current);
      }
      else if (cmp < -1)
      {
         /* Callback has found a possible match, but
          * wishes to continue search in the left subtree. */
         best = current;
         current = Left(current);
      }
      else
      {
         /* Callback has found a possible match, but
          * wishes to continue search in the right subtree. */
         best = current;
         current = Right(current);
      }
   }

   return best;
}


core_rbtree_node_t *
core_rbtree_first_no_lock(core_rbtree_t *tree)
{
   core_rbtree_node_t *node;

   if ((node = tree->root) == NIL(tree))
   {
      return NULL;
   }
   else
   {
      while (Left(node) != NIL(tree))
      {
	 node = Left(node);
      }
   }

   return node;
}


core_rbtree_node_t *
core_rbtree_next_no_lock(core_rbtree_t *tree, void *node_ptr)
{
   core_rbtree_node_t *node = (core_rbtree_node_t*)node_ptr;
   
   if (Right(node) != NIL(tree))
   {
      /* go right 1... */
      node = Right(node);

      /* ...then left to the end */
      while (Left(node) != NIL(tree))
      {
	 node = Left(node);
      }
   }
   else
   {
      /* while you're the right child, chain up parent link */
      core_rbtree_node_t *p;

      p = Parent(node);

      while (p != NIL(tree) && node == Right(p))
      {
	 node = p;
	 p = Parent(p);
      }

      if (p == NIL(tree))
      {
	 node = NULL;
      }
      else
      {
	 node = p;
      }
   }

   return (core_rbtree_node_t *)node;
}

core_rbtree_node_t *
core_rbtree_last_no_lock(core_rbtree_t *tree)
{
   core_rbtree_node_t *node;

   if ((node = tree->root) == NIL(tree))
   {
      return NULL;
   }
   else
   {
      while (Right(node) != NIL(tree))
      {
	 node = Right(node);
      }
   }

   return node;
}


core_rbtree_node_t *
core_rbtree_prev_no_lock(core_rbtree_t *tree, void *node_ptr)
{
   core_rbtree_node_t *node = (core_rbtree_node_t*)node_ptr;

   if (Left(node) != NIL(tree))
   {
      /* go right 1... */
      node = Left(node);

      /* ...then right to the end */
      while (Right(node) != NIL(tree))
      {
	 node = Right(node);
      }
   }
   else
   {
      /* while you're the right child, chain up parent link */
      core_rbtree_node_t *p;

      p = Parent(node);

      while (p != NIL(tree) && node == Left(p))
      {
	 node = p;
	 p = Parent(p);
      }

      if (p == NIL(tree))
      {
	 node = NULL;
      }
      else
      {
	 node = p;
      }
   }
   
   return (core_rbtree_node_t *)node;
}


void
core_rbtree_traverse_no_lock(core_rbtree_t *tree, rbtree_traverse_func_t func)
{
   core_rbtree_node_t *current;

   for (current = core_rbtree_first_no_lock(tree);
        current != (void*)0;
        current = core_rbtree_next_no_lock(tree, current))
   {
      func(tree, current);
   }
}


/*
 *===========================================================================
 *			       assert_tree
 *===========================================================================
 *
 */
#ifndef NDEBUG
void
assert_tree(core_rbtree_t *tree, const core_rbtree_node_t *not_in_tree)
{
   (void)not_in_tree; 
   
   /* The root node must be colored BLACK */
   assert(tree->root == NIL(tree) || IsBlack(tree->root));

   /* The null_node is invariant (except parent) */
   assert(Left(NIL(tree)) == NIL(tree));
   assert(Right(NIL(tree)) == NIL(tree));
   assert(IsBlack(NIL(tree)));

}
#endif /* !NDEBUG */
