/* COPYRIGHT-ENEA-SRC-R2 *
**************************************************************************
* Copyright (C) 2004-2006 by Enea Software AB.
* All rights reserved.
*
* This Software is furnished under a software license agreement and
* may be used only in accordance with the terms of such agreement.
* Any other use or reproduction is prohibited. No title to and
* ownership of the Software is hereby transferred.
*
* PROPRIETARY NOTICE
* This Software consists of confidential information.
* Trade secret law and copyright law protect this Software.
* The above notice of copyright on this Software does not indicate
* any actual or intended publication of such Software.
**************************************************************************
* COPYRIGHT-END */


#include "core_segtree.h"
#include "ramlog.h"



/*
 *===========================================================================
 *                         Private helper functions. 
 *===========================================================================
 */

static int
segment_add(core_segtree_t *tree, core_segtree_node_t *new_segment,
            uint64_t start, uint64_t end);

static int
segment_remove(core_segtree_t *tree, core_segtree_node_t *old_segment,
               uint64_t start, uint64_t end);

static int
segment_split(core_segtree_t *tree,
              core_segtree_node_t *left_segment,
              uint64_t position,
              core_segtree_node_t **out_segment);

static int
segment_merge(core_segtree_t *tree,
              core_segtree_node_t *left,
              core_segtree_node_t *right);



/*
 *===========================================================================
 *                         core_segtree_init
 *===========================================================================
 *
 * Description
 * -----------
 * See core_segtree.h.
 * 
 */
int
core_segtree_init(core_segtree_t *tree,
                  core_segtree_funcs_t *funcs,
                  uint32_t node_size)
{
   if (funcs == NULL ||
       funcs->alloc_node == NULL ||
       funcs->free_node  == NULL ||
       funcs->compare    == NULL)
   {
      return -1;
   }

   tree->funcs = funcs;
   tree->node_size = node_size;
   
   core_rbtree_init(&tree->tree, (rbtree_compare_func_t)funcs->compare);

   return 0;
}


/*
 *===========================================================================
 *                         core_segtree_node_init
 *===========================================================================
 *
 * Description
 * -----------
 * See core_segtree.h.
 * 
 */
void
core_segtree_node_init(core_segtree_t *tree, core_segtree_node_t *segment,
                       uint64_t start, uint64_t end)
{

   (void)tree;
   
   segment->start = start;
   segment->end   = end;
   segment->ref_count = 1;
}


/*
 *===========================================================================
 *                         core_segtree_find
 *===========================================================================
 *
 * Description
 * -----------
 * See core_segtree.h.
 * 
 */
core_segtree_node_t *
core_segtree_find(core_segtree_t *tree,
                  core_segtree_compare_t find_func,
                  void *key)
{
   return (core_segtree_node_t *)
      core_rbtree_find_no_lock(&tree->tree,
                               (rbtree_compare_func_t)find_func, key);
}


/*
 *===========================================================================
 *                         core_segtree_add
 *===========================================================================
 *
 * Description
 * -----------
 * See core_segtree.h.
 * 
 */
int
core_segtree_add(core_segtree_t *tree, core_segtree_node_t *node)
{
   if (tree->funcs->add)
   {
      /* Make sure the user allows adding this type of interval */

      core_rbtree_node_t *old = 
         core_rbtree_find_no_lock(&tree->tree,
                                  (rbtree_compare_func_t)tree->funcs->add,
                                  node);
      if (old != NULL)
      {
         return -1;
      }
   }

   return segment_add(tree, node, node->start, node->end);
}


/*
 *===========================================================================
 *                         core_segtree_remove
 *===========================================================================
 *
 * Description
 * -----------
 * See core_segtree.h.
 * 
 */
int
core_segtree_remove(core_segtree_t *tree, core_segtree_node_t *node)
{
   uint64_t start = node->start, end = node->end;
   core_segtree_node_t *segment;
   int status;

   /* First start by splitting the ends of the new interval to avoid
    * a problematic rollback if memory gets exhausted in the
    * middle of the operation. */


   /* Left end probed first */
   core_segtree_node_init(tree, node, start, start+1);
   
   segment = (core_segtree_node_t*)
      core_rbtree_find_no_lock(&tree->tree, tree->funcs->compare, node);

   if (segment == NULL)
   {
      ramlog_printf("mman: segment_remove failed to locate a range\n");
      return 0;
   }

   /* Check if unmap starts in the middle of a segment.
    *  To remove:     |----------|
    *  Found:     |-----|
    *  Result:    |---|-|
    */
   if (segment->start < start)
   {
      core_segtree_node_t *right = NULL;

      status = segment_split(tree, segment, start, &right);
      if (status < 0)
      {
         ramlog_printf("segment_remove failed to split existing segment\n");
         return status;
      }
   }


   /* Right end probed last */
   core_segtree_node_init(tree, node, end-1, end);
   
   segment = (core_segtree_node_t*)
      core_rbtree_find_no_lock(&tree->tree, tree->funcs->compare, node);
   
   if (segment == NULL)
   {
      ramlog_printf("mman: segment_remove failed to locate a range\n");
      return 0;
   }

   /* Check if unmap ends in the middle of a segment. 
    *  To remove:   |---|
    *  Found:       |----------|
    *  Result:      |---|------|
    */
   if (segment->end > end)
   {
      core_segtree_node_t *right = NULL;

      status = segment_split(tree, segment, end, &right);
      if (status < 0)
      {
         ramlog_printf("mman: segment_remove failed to split existing range\n");
         return status;
      }
   }

   /* It should now be safe to deference individual intervals, no more
    * memory will be consumed. */
   
   return segment_remove(tree, node, start, end);
}


/*
 *===========================================================================
 *                         core_segtree_empty
 *===========================================================================
 *
 * Description
 * -----------
 * Determines if the tree is empty or not.
 *
 * Returns
 * -------
 * Non-zero if empty.
 */
int
core_segtree_empty(core_segtree_t *tree)
{
   return tree->tree.root == &tree->tree.nil;
}



/*
 *===========================================================================
 *                         segment_split
 *===========================================================================
 *
 * Description
 * -----------
 * Private helper function used to split an interval.
 * 
 */
static int
segment_split(core_segtree_t *tree,
              core_segtree_node_t *left_segment,
              uint64_t position,
              core_segtree_node_t **out_segment)
{
   core_segtree_node_t *right_segment;

   right_segment = tree->funcs->alloc_node(tree->node_size);
   if (right_segment == NULL)
   {
      return -1;
   }

   right_segment->start     = position;
   right_segment->end       = left_segment->end;
   right_segment->ref_count = left_segment->ref_count;
   left_segment->end        = position;

   *out_segment = right_segment;

   if (core_rbtree_insert_no_lock(&tree->tree,
                                  &right_segment->node) < 0)
   {
      ramlog_printf("Failed to split a segment\n");
      error2(0, 0);
   }

   if (tree->funcs->ctor)
   {
      /* Call constructor, global representation not altered */
      tree->funcs->ctor(tree, left_segment, right_segment, 0);
   }
   
   return 0;
}


/*
 *===========================================================================
 *                         segment_merge
 *===========================================================================
 *
 * Description
 * -----------
 * Private helper function used to merge two intervals.
 * 
 */
static int
segment_merge(core_segtree_t *tree,
              core_segtree_node_t *left_segment,
              core_segtree_node_t *right_segment)
{
   if (left_segment->ref_count != right_segment->ref_count ||
       (left_segment->end  != right_segment->start &&
        right_segment->end != left_segment->start))
   {
      ramlog_printf("Failed to merge two segments\n");
      error2(0, 0);
   }
   

   if (left_segment->end == right_segment->start)
   {
      core_rbtree_remove_no_lock(&tree->tree, &right_segment->node);
      if (tree->funcs->dtor)
      {
         /* Call destructor, global representation not altered */
         tree->funcs->dtor(tree, left_segment, right_segment, 0);
      }
      left_segment->end = right_segment->end;
      tree->funcs->free_node(right_segment);
   }
   else
   {
      core_rbtree_remove_no_lock(&tree->tree, &right_segment->node);
      if (tree->funcs->dtor)
      {
         /* Call destructor, global representation not altered */
         tree->funcs->dtor(tree, left_segment, right_segment, 0);
      }
      left_segment->start = right_segment->start;
      tree->funcs->free_node(right_segment);
   }

   return 0;
}


/*
 *===========================================================================
 *                         segment_add
 *===========================================================================
 *
 * Description
 * -----------
 * Private helper function used to add a new interval to the tree.
 *
 * Adding a new interval is performed recursively. This is because of
 * performance reasons and the search for existing overlapping intervals
 * already present in the tree. 
 *
 * The function will do a complete rollback if the system runs out
 * of memory while adding new pieces of the interval.
 */
static int
segment_add(core_segtree_t *tree,
            core_segtree_node_t *new_segment,
            uint64_t start, uint64_t end)
{
   core_segtree_node_t *segment;
   int status;

   core_segtree_node_init(tree, new_segment, start, end);
   
   segment = (core_segtree_node_t*)
      core_rbtree_find_no_lock(&tree->tree, tree->funcs->compare, new_segment);

   if (segment != NULL)
   {
      /* New range covers the existing range completely.
       *  
       *  New range:   |----------|
       *  Found range:    |----|
       *  Result:      |--|----|--|
       */
      if (segment->start >= start && 
          segment->end   <= end)
      {
         /* Add left side recursively */
         if (segment->start > start)
         {
            status = segment_add(tree, new_segment,
                                 start, segment->start);
            if (status < 0)
            {
               return status;
            }
         }

         /* Add right side recursively */
         if (segment->end < end)
         {
            status = segment_add(tree, new_segment,
                                 segment->end, end);
            if (status < 0)
            {
               /* Rollback left segment */
               segment->end = segment->start;
               segment->start = start;
               segment_remove(tree, segment, start, segment->start);
               return status;
            }
         }

         segment->ref_count++;
      }

      /* Found range covers new range completely.
       *  
       *  New range:      |----|
       *  Found range: |----------|
       *  Result:      |--|----|--|
       */
      else if (segment->start <= start &&
               segment->end   >= end)
      {
         core_segtree_node_t *left   = NULL;
         core_segtree_node_t *middle = segment;
         core_segtree_node_t *right  = NULL;
         
        /* Add left side recursively */
         if (segment->start < start)
         {
            status = segment_split(tree, segment, start, &middle);
            if (status < 0)
            {
               return status;
            }
            left = segment;
         }

         /* Add right side recursively */
         if (segment->end > end)
         {
            status = segment_split(tree, middle, end, &right);
            if (status < 0)
            {
               if (left != NULL)
               {
                  /* Do rollback of left segment */
                  segment_merge(tree, left, middle);
               }
               return status;
            }
         }

         middle->ref_count++;
      }

      /* New range covers left part of existing range.
       *  New range:   |-----|
       *  Found range:    |-----|
       *  Result:      |--|--|--|
       */
      else if (segment->start > start)
      {
         core_segtree_node_t *right = NULL;

         status = segment_split(tree, segment, end, &right);
         if (status < 0)
         {
            return status;
         }

         status = segment_add(tree, new_segment,
                              start, segment->start);
         if (status < 0)
         {
            /* Do rollback of split */
            segment_merge(tree, segment, right);
            return status;
         }

         segment->ref_count++;
      }

      /* New range covers right part of existing range.
       *  New range:      |-----|
       *  Found range: |-----|
       *  Result:      |--|--|--|
       */
      else if (segment->start < start)
      {
         core_segtree_node_t *right = NULL;

         status = segment_split(tree, segment, start, &right);
         if (status < 0)
         {
            return status;
         }

         status = segment_add(tree, new_segment,
                              segment->end, end);
         if (status < 0)
         {
            segment_merge(tree, segment, right);
            return status;
         }

         /* segment->ref_count += 1; */
      }
      else
      {
         ramlog_printf("Unknown condition detected by segment_add\n");
         error2(0, 0);
      }
   }
   else
   {
      segment = (core_segtree_node_t*)
         tree->funcs->alloc_node(tree->node_size);

      if (segment == NULL)
      {
         ramlog_printf("Failed to allocate a new segment node\n");
         return -1;
      }

      segment->start     = start;
      segment->end       = end;
      segment->ref_count = 1;

      status = core_rbtree_insert_no_lock(&tree->tree, &segment->node);
      if (status < 0)
      {
         ramlog_printf("mman: Added page range overlaps existing range\n");
         error2(0, 0);
      }

      if (tree->funcs->ctor)
      {
         /* Call constructor, global representation altered */
         tree->funcs->ctor(tree, new_segment, segment, 1);
      }
   }

   return 0;
}


/*
 *===========================================================================
 *                         segment_remove
 *===========================================================================
 *
 * Description
 * -----------
 * Private helper function used to remove an interval from the tree. The
 * interval must be complete, i.e. removing an interval with pieces missing
 * is not permitted. 
 *
 * The function will do a complete rollback if the system runs out
 * of memory while removing pieces of the interval. New allocations will
 * take place if only part of an existing interval is removed.
 * 
 */
static int
segment_remove(core_segtree_t *tree, core_segtree_node_t *old_segment,
               uint64_t start, uint64_t end)
{
   core_segtree_node_t *segment;
   int status = 0;

   core_segtree_node_init(tree, old_segment, start, end);
   
   segment = (core_segtree_node_t*)
      core_rbtree_find_no_lock(&tree->tree, tree->funcs->compare, old_segment);

   if (segment == NULL)
   {
      ramlog_printf("mman: segment_remove failed to locate a range\n");
      return 0;
   }

   /* Check if unmap starts in the middle of a segment.
    *  To remove:     |----------|
    *  Found:     |-----|
    *  Result:    |---|-|
    */
   if (segment->start < start)
   {
      core_segtree_node_t *right = NULL;

      status = segment_split(tree, segment, start, &right);
      if (status < 0)
      {
         ramlog_printf("segment_remove failed to split existing segment\n");
         return status;
      }
      segment = right;
   }

   /* Check if unmap ends in the middle of a segment. 
    *  To remove:   |---|
    *  Found:       |----------|
    *  Result:      |---|------|
    */
   if (segment->end > end)
   {
      core_segtree_node_t *right = NULL;

      status = segment_split(tree, segment, end, &right);
      if (status < 0)
      {
         ramlog_printf("mman: segment_remove failed to split existing range\n");
         return status;
      }
   }

   segment->ref_count--;

   /* Recursively remove segments to the right of the found interval.
    *
    * To remove: |---------|
    * Found:         |-|
    * Recursion:       |---|
    */
   if (segment->end < end)
   {
      status = segment_remove(tree, old_segment, segment->end, end);
      if (status < 0)
      {
         ramlog_printf("mman: segment_remove failed to remove existing range\n");
         return status;
      }
   }

   /* Recursively remove segments to the left of the found interval.
    *
    * To remove: |---------|
    * Found:         |-|
    * Recursion: |---|
    */
   if (start < segment->start)
   {
      segment_remove(tree, old_segment, start, segment->start);
      if (status < 0)
      {
         ramlog_printf("mman: segment_remove failed to remove existing range\n");
         return status;
      }
   }
   
   if (segment->ref_count == 0)
   {
      core_rbtree_remove_no_lock(&tree->tree, &segment->node);
      if (tree->funcs->dtor)
      {
         tree->funcs->dtor(tree, old_segment, segment, 1);
      }
      tree->funcs->free_node(segment);
   }
   else
   {
      core_segtree_node_t *other = NULL;
      
      /* Try to merge the current segment with its left sibling */
      core_segtree_node_init(tree, old_segment, segment->start - 1, segment->start);
      other = (core_segtree_node_t*)
         core_rbtree_find_no_lock(&tree->tree, tree->funcs->compare, old_segment);
      if (other != NULL && other->ref_count == segment->ref_count)
      {
         segment_merge(tree, other, segment);
         segment = other;
      }

      /* Try to merge the current segment with its right sibling */
      core_segtree_node_init(tree, old_segment, segment->end, segment->end + 1);
      other = (core_segtree_node_t*)
         core_rbtree_find(&tree->tree, tree->funcs->compare, old_segment);
      if (other != NULL && other->ref_count == segment->ref_count)
      {
         segment_merge(tree, other, segment);
      }
      
   }

   return 0;
}

