/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2015 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

/* C runtime include files. */
#include <string.h>

/* OSE public include files. */
#include "assert.h"
#include "cache.h"
#include "mm.sig"
#include "mmtypes.h"

/* MM internal include files. */
#include "common.h"
#include "compat.h"
#include "domain.h"
#include "gmm.h"
#include "hal.h"
#include "list_ext.h"
#include "zone.h"
#include "mm_log.h"
#include "mm_profile.h"
#include "mm_util_internal.h"
#include "log_object.h"


#define PRINT(s) log(CATEGORY_DOMAIN, s)

#define DOMAIN_MAGIC 0x12121212
#define valid_segnum(segnum) ((segnum) < gmm_state.max_domains)

/*
 * Physical resource container.
 */
struct phys_res
{
   list_link_t link;
   struct phys_object *pobj;
};


struct phys_mem_chunk
{
   phys_size_t base;
   phys_size_t size;
};

/*
 * Initial region mem chunk container.
 */
struct init_region_chunk
{
   list_link_t link;
   struct phys_mem_chunk mem_chunk;
};

struct zone region_zone;
static struct zone domain_zone;
static struct zone phys_res_zone;
static struct zone init_region_chunk_zone;


#if defined(ASSERT)
static void
region_assert(const struct region *region)
{
   assert(region->log_object != NULL);
}

static void 
region_assert_tree(const void *tree, void *region_ptr)
{
   region_assert((struct region*)region_ptr);
}

static void
domain_assert(struct domain *domain)
{
   assert(domain != NULL);
   assert(DOMAIN_MAGIC == domain->magic);
   core_rbtree_traverse_no_lock(&domain->regions, region_assert_tree);
}

static void
domain_assert_all(struct domains *domains)
{
   int i;

   for (i = 0; i < MM_MAX_DOMAINS; ++i)
   {
      struct domain *domain = domains->domain_list[i];

      if (NULL != domain)
      {
         domain_assert(domain);
      }
   }
}
#else
#define region_assert(a)
#define domain_assert(a)
#define domain_assert_all(a)   (void)a
#endif

/*
 * Returns the given region if its base address equals the
 * given address. Otherwise, NULL is returned.
 */
static int
compare_exactly_func(const void *base_ptr,
                     const void *region_ptr)
{
   const log_addr_t *base = (const log_addr_t*)base_ptr;
   const struct region *region = (const struct region *)region_ptr;
   
   assert(NULL != region);
   region_assert(region);

   if (*base < region->log_object->base)
   {
      return -1;
   }
   else if (*base > region->log_object->base)
   {
      return 1;
   }

   return 0;
}

/*
 * Returns the given region if the given address is within
 * range of the given region.
 */
static int
compare_within_func(const void *base_ptr,
                    const void *region_ptr)
{
   const log_addr_t *base = (const log_addr_t*)base_ptr;
   const struct region *region =
      (const struct region *)region_ptr;

   assert(NULL != region);
   region_assert(region);

   if (*base < region->log_object->base)
   {
      return -1;
   }
   else if (*base >= (region->log_object->base +
                      region->log_object->size))
   {
      return 1;
   }

   return 0;
}


/*
 * Function to compare two regions. Used by red/black-tree
 * to sort its nodes.
 */
static int
compare_regions_func(const void *a_ptr, const void *b_ptr)
{
   const struct region *a = (const struct region*)a_ptr;
   const struct region *b = (const struct region*)b_ptr;
   
   assert(NULL != a);
   assert(NULL != b);
   region_assert(a);
   region_assert(b);

   if (a->log_object->base < b->log_object->base)
   {
      return -1;
   }
   else if (a->log_object->base > b->log_object->base)
   {
      return 1;
   }

   return 0;
}


/*
 * Returns:
 *  1 - If the "toinsert" region precede the region "compare" with
 *        respect its the base address.
 *  0 - If bases are equal.
 * -1 - If "toinsert" base >  "compare" base.
 */
static int
preceding_region(const list_link_t *link_toinsert,
		 const list_link_t *link_compare)
{
   const struct init_region_chunk *reg_toinsert =
      (const struct init_region_chunk *) link_toinsert;

   const struct init_region_chunk *reg_compare =
      (const struct init_region_chunk *) link_compare;
   int result;
   
   
   result =
      (reg_toinsert->mem_chunk.base < reg_compare->mem_chunk.base) ? 1 :
      (reg_toinsert->mem_chunk.base > reg_compare->mem_chunk.base) ? -1 : 0;

   return result;
}


/*
 * Save memory areas freed from initial dynamic regions.
 * Such memory areas must not be used subsequent persistent regions
 * or there will be a conflict at next warmstart.
 *
 * Initial dynamic regions are created in start up and made of a continues
 * memory chunk. Usually such a region is SASE but its physical address may be
 * different from the logical, thus we have to extract the memory chunk from
 * pobj.
 *
 */ 

static void
save_initial_dynamic_region(struct phys_object *pobj)
{
   struct init_region_chunk *init_reg;
   struct phys_mem_cursor cursor;


   init_reg = mm_zalloc(&init_region_chunk_zone);

   if (init_reg == NULL) {
      /* We are out of phys mem but still proceed since
         a) don't want to deny delete_region due to system memory running low.
         b) this will not give any problem until next warm start. 
      */
      return; 
   }
   
   mm_phys_mem_get_first(pobj, &cursor);

   init_reg->mem_chunk.base = cursor.phys_page.pa;
   init_reg->mem_chunk.size = cursor.phys_page.size;
   
   
   /* Initial regions are continues memory, should only be one frag. */
   if (mm_phys_mem_get_next(&cursor)) {
      INTERNAL_ERROR();
   }

   init_list_link(&(init_reg->link));

   /* Insert sorted by addresses in ascending order.*/
   insert_list_ordered(&(gmm_state.deleted_init_dynamic_regions),
                       &(init_reg->link),
                       preceding_region);
}


/*
 * Returns NULL if no overlap, else link is returned.
 *
 */
static void *
mem_chunk_overlap(list_link_t *link_compare, void *mem_chunk)
{
   const struct init_region_chunk *reg_compare =
      (const struct init_region_chunk *) link_compare;
   
   const struct phys_mem_chunk *chunk_to_check =
      (const struct phys_mem_chunk *) mem_chunk;
      
   return mm_phys_overlap(chunk_to_check->base, chunk_to_check->size,
                          reg_compare->mem_chunk.base,
                          reg_compare->mem_chunk.size) ?
      link_compare : NULL;
}

/* Checks given range for overlap with any freed memory that originated
 * from a deleted initial region.
 */
mm_bool_t
mm_initial_region_overlap(phys_addr_t base,  phys_size_t size)
{

   struct phys_mem_chunk mem_chunk;
   struct init_region_chunk *overlap_init_reg;
   mem_chunk.base = base;
   mem_chunk.size = size;

   overlap_init_reg = traverse_list(&(gmm_state.deleted_init_dynamic_regions),
                       mem_chunk_overlap,
                       (void *) &mem_chunk);

   if (overlap_init_reg == NULL)
   {
      return FALSE;
   }
   else
   {
      /* FIXME, change to PRINT() */
      ramlog_printf("mm_initial_region_overlap(%#llx, %#llx) overlap with\n"
                    " memory freed originated from inital region base: "
                    "%#llx size: %#llx\n", base, size,
                    overlap_init_reg->mem_chunk.base,
                    overlap_init_reg->mem_chunk.base);
      return TRUE;
   }
}


int ose_mm_segnum_cnt = 0;

/* Allocate a segnum, and return true if operation succeeded.
 * The domains struct is updated if the segnum was successfully
 * updated.
 */
static mm_bool_t
alloc_segnum(struct domains  *domains,
	     unsigned        *result)
{
   const U32 max_domains = gmm_state.max_domains;
   assert(NULL != domains);
   assert(NULL != result);

   if (domains->no_free > 0)
   {
      struct domain **array = domains->domain_list;

      /* Scan for next free entry in domain vector. */
      while (array[domains->next_ix] != NULL)
      {
         domains->next_ix =
            (domains->next_ix + 1) % max_domains;
         assert(domains->next_ix < max_domains);
      }

      *result = domains->next_ix;
      --(domains->no_free);

      /* Bump the index in order to not reuse the current segnum
       * in case the domain is deleted before another is created */

      /* Note that Powerpc PEM architecture relies on the fact that
       * segnum index is bumped. The reason is that bumping segnum
       * increases the time before any segnum is re-cycled.
       * This minimizes the risc of hitting old uncleaned PTE's
       * in the hash_table. See invalidate_tlb_range() family_pem.c
       * for details.
       *
       * An improvement in order to maximize the time of recycling a segnum
       * a list should be used and allocating fresh segnumns from the
       * beginning and return used in the end.
       * Drawback: the list would require extra footprint and will
       * affect all architectures but would only be needed by PEM.
       */
      domains->next_ix =
         (domains->next_ix + 1) % max_domains;
      assert(domains->next_ix < max_domains);

      ose_mm_segnum_cnt++; /* Keep track on wrap situations primarily for
                              pem. */
      return TRUE;
   }

   return FALSE;
}

/*
 * Frees the segnum allocated by alloc_segnum.
 */
static void
free_segnum(struct domains *domains, unsigned segnum)
{
   assert(NULL != domains);
   assert(valid_segnum(segnum));
   assert(NULL != domains->domain_list[segnum]);
   domain_assert_all(domains);
   domains->domain_list[segnum] = NULL;
   ++(domains->no_free);
   domain_assert_all(domains);
}

/*
 * The share number sent to the HAL should be one, if the operation
 * should affect the mappings in all domains. If only the specified
 * domain should be affected, share should be greater than one.
 *
 * As share variable, the ref_count value in logical region could
 * be used for most cases. The only exception is for when a logical
 * region is re-created in a new domain. When this happens, the
 * ref_count variable is greater than one, since two regions
 * exists for that logical region. One of them is the core domain
 * region with phys_object that is NULL.
 * Hence, the share variable will be (log_object->ref_count - 1).
 */
static U32
get_share_value(struct log_object  *log_object)
{
   U32             share;
   struct region  *core_instance;
   struct domains *domains;

   domains = &gmm_state.domains; /* FIXME should be passed as parameter */
   assert(NULL != log_object);

   share = log_object->ref_count;

   /* Optimize performance by only do check if neccessary. */
   if (share > 1)
   {
      core_instance = mm_find_region_by_addr(
         domains, MM_KERNEL_DOMAIN, log_object->base, TRUE);

      /* clli */
      if (core_instance != NULL && core_instance->phys_object == NULL)
      {
         --share;
      }
   }
   return share;
}

/*
 *  See domain.h
 *
 */
int
mm_set_hal_options(struct domain *domain, struct region *region,
                   mm_bool_t change)
{
   int mas;
   int share_count;
   int manager_mode;
   int options = 0;
   struct partition *partition;
   struct log_object *lobj = region->log_object;

   partition = mm_log_object_partition(lobj, domain);
   assert(partition != NULL);

   share_count = get_share_value(lobj);
   mas = (lobj->options & MM_PARTITION_OPT_EXCLUSIVE) == 0;
   manager_mode = partition_is_manager_mode(partition->class);
   
   if ((share_count > 1) && !mas)
   {
      options |= HAL_SHARE;
   }

   if (change)
   {
      options |= HAL_CHANGE;
   }

   if (mas)
   {
      options |= HAL_MAS;
   }

   if (manager_mode)
   {
      options |= HAL_MANAGER_MODE;
   }

   if (region_is_tlb_global(region))
   {         
      options |= HAL_GLOBAL_MAPPING;
   }

   if (region_is_tlb_global_transient(region))
   {         
      options |= HAL_GLOBAL_MAPPING_TRANSIENT;
   }
   
   return options;
}


/*
 ***********************************************************************
 * public functions
 ***********************************************************************
 */
void
mm_domain_init(void)
{
   memset(gmm_state.domains.domain_list, 0,
          sizeof(gmm_state.domains.domain_list));
   if (gmm_state.max_domains == 0)
      gmm_state.max_domains = MM_MAX_DOMAINS;
   gmm_state.domains.no_free = gmm_state.max_domains;
   gmm_state.domains.next_ix = MM_KERNEL_DOMAIN;

   mm_zinit(&region_zone, "region",
            sizeof(struct region), 3, 10);
   mm_zinit(&domain_zone, "domain",
            sizeof(struct domain), 2, 3);
   mm_zinit(&phys_res_zone, "phys_res",
            sizeof(struct phys_res), 2, 3);
   mm_zinit(&init_region_chunk_zone, "init_region_chunk_zone",
            sizeof(struct init_region_chunk), 2, 3);

   init_list(&(gmm_state.deleted_init_dynamic_regions));
}

/*
 * Creates a new domain, and adds it to the domain list.
 * It also creates a new HAL domain.
 */
mm_status_t
mm_add_domain(struct domains  *domains,
	      const char      *name,
              struct domain  **out_domain)
{
   struct domain *domain = NULL;
   hal_status_t hal_status;
   mm_status_t status;
   unsigned seg_num;

   assert(NULL != domains);
   assert(NULL != name);
   domain_assert_all(domains);

   /* Try to allocate a new domain id. */
   if (!alloc_segnum(domains,  &seg_num))
   {
      return MM_EOUT_OF_DOMAINS;
   }
   

   PRINT(("mm: mm_add_domain entry segnum %ld "\
          "domains->domain_list[seg_num]: %#lx\n",
          seg_num, domains->domain_list[seg_num]));

   domain = mm_zalloc(&domain_zone);
   if (!domain)
   {
      free_segnum(domains, seg_num);
      return MM_EOUT_OF_PHYS_MEM;
   }

   memset(domain, 0, sizeof(struct domain));
   
#ifdef ASSERT
   domain->magic = DOMAIN_MAGIC;
#endif

   /* Initialize domain descriptor. */
   domain->segnum = seg_num;
   core_rbtree_init(&domain->regions, compare_regions_func);
   init_list(&(domain->phys_resources));
   strncpy(&domain->name[0], name, sizeof(domain->name));
   domain_assert(domain);
   domain->mas_cnt = 0;

   status = mm_create_domain_partitions(domain);
   if (status != MM_SUCCESS)
   {
      free_segnum(domains, seg_num);
      mm_zfree(&domain_zone, domain);
      return status;
   }
   
   /* Hook in the new domain. */
   domains->domain_list[seg_num] = domain;

   if (seg_num != MM_KERNEL_DOMAIN)
   {
      hal_status = hal_create_domain(seg_num, FALSE, &domain->hal_domain);
      if (hal_status != HAL_SUCCESS)
      {
         /*
          * HAL could not create domain, cleanup and return error.
          * The hal_status parameter will be set to the status of the
          * hal_create_domain operation.
          * The domain is freed, and the seg_num is returned.
          */
         mm_destroy_domain_partitions(domain);
         free_segnum(domains, seg_num);
         mm_zfree(&domain_zone, domain);

         return mm_hal_status_to_mm_status(hal_status);
      }
   }
   else
   {
      /* Core domain created once in a special way in start up phase. */
#ifdef ASSERT
      static int cnt = 0;
      cnt++;
      assert(cnt == 1);
#endif
   }

   *out_domain = domain;

   PRINT(("mm: mm_add_domain exits segnum %ld "\
          "domains->domain_list[segnum]: %#lx\n",
          seg_num, domains->domain_list[seg_num]));

   domain_assert_all(domains);

   return MM_SUCCESS;
}

/*
 * Returns the domain that has the given segnum.
 */
struct domain *
mm_find_domain_by_domid(struct domains *domains,
			OSDOMAIN        segnum)
{
   struct domain *res = NULL;
   assert(NULL != domains);

   /* First validate the index (segnum). */
   if (valid_segnum(segnum))
   {
      res = domains->domain_list[segnum];
   }

   domain_assert_all (domains);

   return res;
}

mm_status_t
mm_unmap_region(struct domain *domain, struct region *region)
{
   struct log_object *lobj;
   int options;
   int share;

   assert(region != NULL);
   assert(region->phys_object != NULL);
   assert(region->log_object  != NULL);

   lobj    = region->log_object;
   options = mm_set_hal_options(domain, region, FALSE);
   share   = get_share_value(lobj);
   
#if defined OSARM || defined OSMIPS
   /* Needed to prevent new mappings from inheriting dirty
    * cache lines previously used by the unmapped region.
    * FIXME: This will not work with MAS regions. */
   if (region->permissions != SU_NA_USR_NA && share <= 1)
   {
      ose_cache(CACHE_INVALIDATE_D, CACHE_L1, lobj->base, lobj->size);
      ose_cache(CACHE_INVALIDATE_I, CACHE_L1, lobj->base, lobj->size);
   }
#endif

   if (!region_is_resident(region))
   {
      hal_unmap_range(domain->hal_domain,
                      lobj->base, lobj->size,
                      options);
   }

   region->meta_options &= ~MM_REGION_OPT_MAPPED;

   return MM_SUCCESS;
}

static void
tlb_global_extra_handling(U32 share, log_addr_t base)
{
   /* Is the region to remove:
    * a) from an app domain.
    * b) the last shared and the original region belongs the system domain?
    * Note that the shared counter is not decremented yet.  */
   if (share == 2) {
      struct region  *sys_region = mm_find_region_by_addr(
         &gmm_state.domains, gmm_state.system_domain->segnum, base, TRUE);

      /* Avoid logical system regions, for example LM_DATA of an LM. */  
      if (sys_region != NULL && !region_is_logical(sys_region)) {
          hal_status_t rc;
          int sys_region_options;

          PRINT(("mm: mm_remove_region  base: %#lx "\
                 "size: %#lx name: %s sys region with share:%d\n",
                 base, sys_region->name, share));
          
         /* If a region in the system domain is shared (thereby loosing its
          * TLB-global attribute) and later is unshared,
          * re-institute the system region as  TLB-global
          */
         sys_region->meta_options |= MM_REGION_OPT_TLB_GLOBAL;


         /* Call HAL to change the MMU mappings to be TLB global
          */
         sys_region_options = mm_set_hal_options(gmm_state.system_domain,
                                                 sys_region, TRUE /*Change*/);

         /* We know that ref_count==2 will be decremented right after
          * mm_remove_region, therefore mask out HAL_SHARE option.
          * */
         sys_region_options &= ~HAL_SHARE;
         
         rc = mm_map_area(gmm_state.system_domain->hal_domain,
                                  sys_region->phys_object,
                                  sys_region->log_object->base,
                                  sys_region_options,
                                  sys_region->permissions,
                                  sys_region->cache_mode);

         if (rc != HAL_SUCCESS) {
            INTERNAL_ERROR();
         }

      }
      
   }
}

/*
 * adapt_global_page_table(struct region  *sys_region, mm_bool_t add_region)
 *
 * This function was added to solve  OSECR-19795: Razor Load case:
 * 1 Create region P in sys domain su_rw_usr_rw.
 * 2 Share  region P to app domain A (and potentially to other app domains).
 * 3 Delete region P in sys domain.
 */
static void 
adapt_global_page_table(struct region  *sys_region, mm_bool_t add_region)
{
   hal_status_t rc;
   int sys_region_options;
   U32 global_page_table_perm;

   /* 
    * Call HAL to change the MMU mappings.
    */
   sys_region_options = mm_set_hal_options(gmm_state.system_domain,
                                           sys_region, TRUE /*Change*/);
   /*
    * Mask out HAL_SHARE option in order to affect global page table.
    */
   sys_region_options &= ~HAL_SHARE;

   if (add_region)
   {
      /* A region is about to be shared/added to sys domain.
       * Regions in sys domain represent user mode global access.
       * Adapt user-mode access to global page tree.
       */
      global_page_table_perm = sys_region->permissions;
   }
   else
   {
      /* Region is about to be removed from sys domain.
       * Clear user-mode access from global page tree.
       */
      rc = hal_actual_permissions(&global_page_table_perm,
                                  sys_region->permissions & 0xf00,
                                  sys_region->log_object->base, 0);
      if (rc != HAL_SUCCESS) {
         INTERNAL_ERROR();
      }      
   }

   rc = mm_map_area(gmm_state.system_domain->hal_domain,
                    sys_region->phys_object,
                    sys_region->log_object->base,
                    sys_region_options,
                    global_page_table_perm,
                    sys_region->cache_mode);

   if (rc != HAL_SUCCESS) {
      INTERNAL_ERROR();
   }
}

/*
 * Removes the given region from the domain, and frees the
 * region struct. It also unmaps the region from HAL.
 */
void
mm_remove_region(struct domains *domains,
                 struct domain  *domain,
                 struct region  *region)
{
   struct log_object  *lobj;
   struct phys_object *pobj;
   U32 share;

   (void)domains; /* Only used in assert(). */
   
   assert(domains != NULL);
   assert(domain != NULL);
   assert(region != NULL);
   assert(region->log_object != NULL);
   assert(!region_is_static(region));
   region_assert(region);
   domain_assert_all(domains);

   /* Get base and size from region. */
   lobj = region->log_object;

   PRINT(("mm: mm_remove_region  base: %#lx "\
          "size: %#lx name: %s \n",
          lobj->base, lobj->size, region->name));

   MM_PROFILE_CHECKPOINT("remove_region", __FILE__, __LINE__);

   /* Get the share value to be used in hal_unmap_area before
    * the tree of regions has been changed.
    */
   share = get_share_value(lobj);

   /* Remove region domains list. */
   core_rbtree_remove_no_lock(&domain->regions, &region->node);
   
   domain_assert_all(domains);
   pobj = region->phys_object;

   /* the phys_container caches the persistency flag, set by the ram
    * allocator in its sase/sas allocation phase.
    */
   if(region_is_persistent(region))
   {
      /* ------------------------------------------------------------ 
       * Persistency is only allowed in the kernel domain.
       */
      if (domain->segnum == MM_KERNEL_DOMAIN)
      {
         mm_pmm_free(lobj->base);
      }
   }


   /* Detach from any physical memory. */
   if (NULL != pobj)
   {
      int options;

      options = mm_set_hal_options(domain, region, FALSE);

#if defined OSARM || defined OSMIPS
      /* Needed to prevent new mappings from inheriting dirty
       * cache lines previously used by the unmapped region.
       * FIXME: This will not work with MAS. */
      if (region->permissions != SU_NA_USR_NA && share <= 1)
      {
	 ose_cache(CACHE_INVALIDATE_D, CACHE_L1, lobj->base, lobj->size);
	 ose_cache(CACHE_INVALIDATE_I, CACHE_L1, lobj->base, lobj->size);
      }
#endif

      if (!region_is_resident(region))
      {
         hal_unmap_range(domain->hal_domain,
                         lobj->base, lobj->size,
                         options);


         if(region_is_initial_dynamic(region))
         {
            /* ------------------------------------------------------------ 
             * Must remember memory areas freed from initial dynamic regions.
             * Such memory must not be used subsequent persistent regions
             * or there will be a conflict at next warmstart.
             */
            save_initial_dynamic_region(pobj);
         }
         tlb_global_extra_handling(share, lobj->base);

         /* Region is removed from system domain (but shared by app domains),  
          * clear user-mode access in global page table.. */
         if (share > 1 &&  domain->segnum == MM_KERNEL_DOMAIN)
            adapt_global_page_table(region, FALSE);
      }
   }
   else if (region_is_faultable(region))
   {
      hal_uncommit_range(domain->hal_domain, lobj->base, lobj->size, 0);
   }
      

   if (!region_is_exclusive(region))
   {
      gmm_state.mas_cnt--;
      domain->mas_cnt--;
      if (domain->mas_cnt == 0)
      {
         /* Notify kernel that the last region
          * located in a MAS partition has been deleted from the domain. */
         get_sysconf(80, domain->segnum, domain->mas_cnt, 0);
      }

      assert((int)gmm_state.mas_cnt >= 0);
      assert((int)domain->mas_cnt >= 0);
   }

   call_region_deleted_event(domain->segnum, lobj->base);
   
   domain_assert_all(domains);
#ifdef ASSERT
   bzero(region, sizeof(region));
#endif
   domain_assert_all(domains);
   mm_zfree(&region_zone, region);
}

static void
free_phys_res(struct phys_res *phys_res)
{
   if (NULL != phys_res->pobj)
   {
      if (NULL != phys_res->pobj->allocator)
      {
         phys_res->pobj->allocator->free(phys_res->pobj->allocator,
                                         phys_res->pobj);
      }
      mm_detach_phys_object(phys_res->pobj);
   }
   mm_zfree(&phys_res_zone, phys_res);
}

/*
 * Removes the given domain from the list of domains, frees the
 * struct and deletes the corresponding HAL domain.
 * Also, the segnum is freed,
 * as well as any physical resources owned by the domain.
 */
mm_bool_t
mm_remove_domain(struct domains *domains, OSDOMAIN segnum)
{
   mm_bool_t      status = FALSE;
   struct domain *domain;

   assert(NULL != domains);
   assert(valid_segnum(segnum));
   assert(MM_KERNEL_DOMAIN != segnum); /* Kernel domain can't be removed. */

   PRINT(("mm: mm_remove_domain entry segnum %ld "\
          "domains->domain_list[seg_num]: %#lx\n",
          segnum, domains->domain_list[segnum]));

   domain = domains->domain_list[segnum];
   domain_assert(domain);
   free_segnum(domains, segnum);

   {
      struct region *curr_region;
      
      /* Remove all active regions of the domain. */
      curr_region = (struct region*)
         core_rbtree_first_no_lock(&domain->regions);
      while (curr_region != NULL)
      {
         struct log_object  *lobj = curr_region->log_object;
         struct phys_object *pobj = curr_region->phys_object;
         struct region *next_region = (struct region *)
            core_rbtree_next_no_lock(&domain->regions, &curr_region->node);
         region_assert(curr_region);

         mm_remove_region(domains, domain, curr_region);

         /* It is important that the physical memory always
          * is freed first since the logical allocator might
          * need it while freeing the logical resources. */
         if (pobj != NULL)
         {
            mm_detach_phys_object(pobj);
         }
         
         if (lobj != NULL)
         {
            mm_detach_log_object(lobj, domain);
         }

         curr_region = next_region;
      }
   }

   {
      /* Remove all physical resources of the domain. */
      struct phys_res *phys_res =
         (struct phys_res *)list_head(&(domain->phys_resources));
      while (NULL != phys_res)
      {
         delete_from_list_no_lock((list_link_t *)phys_res);
         free_phys_res(phys_res);
         phys_res = (struct phys_res *)list_head(&(domain->phys_resources));
      }
   }

   hal_delete_domain(domain->segnum, domain->hal_domain);
   mm_destroy_domain_partitions(domain);
   mm_zfree(&domain_zone, domain);

   PRINT(("mm: mm_remove_domain exit segnum %ld "\
          "domains->domain_list[seg_num]: %#lx\n",
          segnum, domains->domain_list[segnum]));
   status = TRUE;

   domain_assert_all(domains);

   return status;
}

/*
 * See domain.h
 */
mm_status_t
mm_add_region(OSDOMAIN segnum,
              struct log_object  *lobj,
              struct phys_object *pobj,
              U32 permissions,
              U32 cache_mode,
              U32 options,
              U32 meta_options,
              const char *name,
              const char *log_mem,
              struct region **reg_out)
{
   struct domain *domain;
   struct region *region;
   mm_status_t status;
   hal_status_t hal_status;
   int share;

   assert(lobj != NULL);
   assert(name != NULL);
   assert(valid_segnum(segnum));
   domain = gmm_state.domains.domain_list[segnum];
   assert(domain != NULL);
   domain_assert(domain);
   share = get_share_value(lobj);

   if (reg_out != NULL)
   {
      *reg_out = NULL;
   }

   /* Optimize by making all regions *created in system domain* as
    * TLB-global (TID=0 etc.). 
    */   
   if (segnum == gmm_state.system_domain->segnum && share == 1)
   {
      meta_options |= MM_REGION_OPT_TLB_GLOBAL;
   }
   
   if (log_object_resident(lobj) && permissions == SU_NA_USR_NA)
   {
      /*
       * Do not round NA_NA permissions if the region is created
       * in a resident log_mem. The region's permissions would be
       * rounded to match those of the log_mem which would be
       * problematic if the region is to be used for duplicate
       * regions. Duplicate regions are only allowed with a NA_NA
       * region as source.
       */
   }
   else
   {
      hal_status = hal_actual_permissions(&permissions, permissions, lobj->base, 0);
      if (hal_status != HAL_SUCCESS)
      {
         return MM_EILLEGAL_PERMISSION;
      }
   }

   hal_status = hal_actual_cache_mode(&cache_mode, cache_mode, lobj->base);
   if (hal_status != HAL_SUCCESS)
   {
      return MM_EILLEGAL_CACHE_MODE;
   }

   region = mm_zalloc(&region_zone);
   if (region == NULL)
   {
      *reg_out = NULL;
      return MM_EOUT_OF_PHYS_MEM;
   }

   region->domain       = domain;
   region->log_object   = lobj;
   region->phys_object  = pobj;
   region->permissions  = permissions;
   region->cache_mode   = cache_mode;
   region->options      = options;
   region->meta_options = meta_options;
   strlcpy(region->name, name, MM_NAME_LENGTH);
   strlcpy(region->log_mem, log_mem, MM_NAME_LENGTH);

   region_assert(region);

   if (region_is_faultable(region))
   {
      hal_status = hal_commit_range(
         domain->hal_domain, lobj->base, lobj->size,
         mm_set_hal_options(domain, region, FALSE));
      if (hal_status != HAL_SUCCESS)
      {
         mm_zfree(&region_zone, region);
         return mm_hal_status_to_mm_status(hal_status);
      }
   }
   else if (region_is_mapped(region))
   {
      status = mm_map_region(segnum, region);
      if (status != MM_SUCCESS)
      {
         /* Caller must detach logical and physical resources */
         mm_zfree(&region_zone, region);
         return status;
      }

      /* Reinstitute global usr permissions if share to sys domain. */
      if (share > 1 &&  domain->segnum == MM_KERNEL_DOMAIN)
         adapt_global_page_table(region, TRUE);
   }

   /* Insert the new region in the domain. */
   {
      int status = core_rbtree_insert_no_lock(&domain->regions, &region->node);
      if (status < 0)
      {
         /* This should normally never happen. If it does happen
          * the logical allocator has distributed the same logical
          * memory twice, even though it was already being used. */
         if (region_is_faultable(region))
         {
            hal_uncommit_range(domain->hal_domain,
                               lobj->base, lobj->size, 0);
         }
         else if (region_is_mapped(region))
         {
            mm_unmap_region(domain, region);
         }

         mm_zfree(&region_zone, region);

         return MM_EOVERLAPPING_ADDRESS;
      }
   }

   if (reg_out != NULL)
   {
      *reg_out = region;
   }

   if (!region_is_exclusive(region))
   {
      gmm_state.mas_cnt++;
      domain->mas_cnt++;
      if (domain->mas_cnt >= 1)
      {
         /* Notify kernel that the domain now contains
          * at least one region located in a MAS partition. */
         get_sysconf(80, domain->segnum, domain->mas_cnt, 0);
      }
   }

   domain_assert(domain);

   return MM_SUCCESS;
}

/*
 * See domain.h
 */
mm_status_t
mm_map_region(OSDOMAIN segnum, struct region *region)
{
   struct log_object  *lobj;
   struct phys_object *pobj;
   struct domain *domain = NULL;
   int options;
   hal_status_t hal_rc;

   assert(region != NULL);

   domain = gmm_state.domains.domain_list[segnum];
   domain_assert(domain);

   pobj = region->phys_object;
   lobj = region->log_object;

   if (pobj == NULL)
   {
      region->meta_options &= ~MM_REGION_OPT_MAPPED;
      return MM_SUCCESS;
   }

   if (region_is_resident(region))
   {
      region->meta_options |= MM_REGION_OPT_MAPPED;
      return MM_SUCCESS;
   }
   
   options = mm_set_hal_options(domain, region, FALSE);
   

   /* Now call HAL the enter the mapping into the MMU. Observe that the
    * return type is hal_status_t, which is different from mm_status_t.
    */
   hal_rc = mm_map_area(domain->hal_domain,
                        pobj,
                        lobj->base,
                        options,
                        region->permissions,
                        region->cache_mode);
      
   if (hal_rc != HAL_SUCCESS)
   {
      return mm_hal_status_to_mm_status(hal_rc);
   }
      
   region->meta_options |= MM_REGION_OPT_MAPPED;

   MM_PROFILE_CHECKPOINT("map_region", __FILE__, __LINE__);

   return MM_SUCCESS;
}



/*
 * Returns the region that matches the given base address.
 * If "exact_match" is true, the given base address must equal
 * the base address of the region.
 * Otherwise, if "exact_match" is false, it is enough that the
 * given address is inside the boundaries of the region.
 */

struct region *
mm_find_region_by_addr(struct domains *domains,
                       OSDOMAIN segnum,
                       log_addr_t base,
                       mm_bool_t exact_match)
{
   struct domain *domain;
   rbtree_compare_func_t compare_func;

   assert(domains != NULL);
   assert(valid_segnum(segnum));

   domain = domains->domain_list[segnum];

   if (domain == NULL)
   {
      return NULL;
   }
   
   assert(domain != NULL);
   domain_assert(domain);

   /* Choose matching function. */
   compare_func = (exact_match ? compare_exactly_func : compare_within_func);

   return (struct region*)core_rbtree_find_no_lock(
      &domain->regions, compare_func, (void *)&base);
}

void
mm_select_domain(U32 segnum)
{
#if 0
   ramlog_printf("mm: mm_select_domain ENTRY segnum: %#ld "\
                 "gmm_state.domains.domain_list[segnum]->hal_domain: %#lx\n",
                 segnum, gmm_state.domains.domain_list[segnum]->hal_domain);
#endif

   assert(gmm_state.domains.domain_list[segnum] != 0);

   hal_select_domain(gmm_state.domains.domain_list[segnum]->hal_domain);

#if 0
   ramlog_printf("mm: mm_select_domain EXIT segnum: %#ld "\
                 "gmm_state.domains.domain_list[segnum]->hal_domain: %#lx\n",
                 segnum, gmm_state.domains.domain_list[segnum]->hal_domain);
#endif
}


const char *
mm_get_phys_mem(struct region *region)
{
   
   if (region->phys_object && region->phys_object->allocator)
   {
      return region->phys_object->allocator->name;
   }
   
   /* There is no way to distinguish the case with a region without physical
    * allocator (e.g. a logical region) from the case with a region whose
    * allocator is named "". An alternative would be to return NULL. One user
    * of this function is the GET_PHYS_MEM_INFO function and in that interface,
    * a string must be returned. It would be possible to use a default value in
    * that specific case, but I think it's better to define that default value
    * ("") here.
    */
   return "";
}

mm_status_t
domain_alloc_phys_resource(struct domains *domains,
                           OSDOMAIN segnum,
                           phys_addr_t *base,
                           phys_size_t size,
                           phys_addr_t align,
                           char *name)
{
   struct domain *domain;
   struct allocator *allocator;
   mm_status_t status;
   struct phys_object *pobj;

   if (align & (align - 1))
   {
      ramlog_printf("Alignment (%#llx) is not a power of 2.\n", align);
      return MM_EILLEGAL_AREA;
   }

   if (NULL == (domain = mm_find_domain_by_domid(domains, segnum)))
   {
      ramlog_printf("Domain %#x not found.\n", segnum);
      return MM_ENON_EXISTING_DOMAIN;
   }

   if (NULL == (allocator = mm_find_allocator_by_name(name)))
   {
      ramlog_printf("Allocator for '%s' not found.\n", name);
      return MM_ENON_EXISTING_PHYS_MEM;
   }

   if (NULL == (pobj = mm_create_phys_object(allocator, 0)))
   {
      ramlog_printf("Failed to create physical object.\n");
      return MM_EOUT_OF_PHYS_MEM;
   }

   mm_attach_phys_object(pobj);

   /* Preferrably allocate physical memory located above 4GiB
    * (leaving as much physical memory as possible for SASE regions,
    *  who require physical memory below 4GiB
    *  due to 32 bits logical memory space). */
   if (MM_SUCCESS !=
       (status = allocator->alloc_c(allocator, pobj, size, align,
                                    MM_PHYS_SIZE_4GiB, 0 - MM_PHYS_SIZE_4GiB)))
   {
      /* If no success above 4GiB, try without constraints. */
      if (MM_SUCCESS !=
          (status = allocator->alloc_c(allocator, pobj, size, align,
                                       0, 0)))
      {
         mm_detach_phys_object(pobj);
         return status;
      }
   }

   {
      struct phys_res *phys_resource = mm_zalloc(&phys_res_zone);
      if (NULL == phys_resource)
      {
         allocator->free(allocator, pobj);
         mm_detach_phys_object(pobj);
         ramlog_printf("Failed to allocate physical resource descriptor\n");
         return MM_EOUT_OF_PHYS_MEM;
      }

      phys_resource->pobj = pobj;
      init_list_link(&(phys_resource->link));
      insert_list_tail(&(domain->phys_resources), &(phys_resource->link));
   }

   {
      /* Extract the allocated physical address */
      phys_addr_t frag_start;
      phys_addr_t frag_len;
      mm_extract_phys_addr(pobj, 0, base, &frag_start, &frag_len);
   }

   return MM_SUCCESS;
}

mm_status_t
domain_free_phys_resource(struct domains *domains,
                          OSDOMAIN segnum,
                          phys_addr_t base)
{
   struct domain *domain;

   if (NULL == (domain = mm_find_domain_by_domid(domains, segnum)))
   {
      return MM_ENON_EXISTING_DOMAIN;
   }

   {
      /* Search for phys object in domain */
      struct phys_res *phys_res =
         (struct phys_res *)list_head(&(domain->phys_resources));
      while (NULL != phys_res)
      {
         phys_addr_t phys_base;
         phys_addr_t frag_start;
         phys_size_t frag_len;
         mm_extract_phys_addr(phys_res->pobj, 0, &phys_base,
                              &frag_start, &frag_len);
         if (phys_base == base)
         {
            /* Done */
            break;
         }
         phys_res =
            (struct phys_res *)list_next(&(domain->phys_resources), phys_res);
      }
      if (NULL == phys_res)
      {
         ramlog_printf("Physical resource with base-address %#llx not found "
                       "in domain %s.\n", base, domain->name);
         return MM_ENON_EXISTING_PHYS_MEM;
      }

      delete_from_list_no_lock((list_link_t *)phys_res);
      free_phys_res(phys_res);
   }

   return MM_SUCCESS;
}

mm_status_t
in_domain_phys_resource(struct domains *domains,
                        OSDOMAIN segnum,
                        phys_addr_t base,
                        phys_addr_t size)
{
   struct domain *domain;

   if (NULL == (domain = mm_find_domain_by_domid(domains, segnum)))
   {
      return MM_ENON_EXISTING_DOMAIN;
   }

   {
      /* Search for phys object in domain */
      struct phys_res *phys_res =
         (struct phys_res *)list_head(&(domain->phys_resources));
      while (NULL != phys_res)
      {
         phys_addr_t phys_base;
         phys_addr_t frag_start;
         phys_size_t frag_len;
         mm_extract_phys_addr(phys_res->pobj, 0, &phys_base,
                              &frag_start, &frag_len);
         if ((phys_base <= base) &&
             ((frag_start - 1 + frag_len) >= (base - 1 + size)))
         {
            /* The supplied physical fragment is entirely contained
             * in a physical resource allocated in the domain - Done */
            break;
         }
         phys_res =
            (struct phys_res *)list_next(&(domain->phys_resources), phys_res);

      }

      if (NULL == phys_res)
      {
         ramlog_printf("Physical resource containing fragment %#llx:%#llx "
                       "not found in domain %s.\n",
                       base, base - 1 + size, domain->name);
         return MM_ENON_EXISTING_PHYS_MEM;
      }
   }

   return MM_SUCCESS;
}
