/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2006 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

#ifndef _DOMAIN_H
#define _DOMAIN_H

/* OSE public include files. */
#include "ose.h"
#include "mm.sig"
#include "core_rbtree.h"
#include "mm_util.h"

/* Core internal include files. */
#include "list_ext.h"

/* MM internal include files. */
#include "phys_object.h"
#include "log_object.h"
#include "partition.h"
#include "hal.h"


/*
 * Different internal meta-properies of regions.
 *
 *  * static     - Sys confed region of type non-removable.
 *  * persistent - Set if region will survive a warm restart.
 *  * removable  - Set if region is dynamically created (not static)
 *  * mapped     - Region has a mapping between logical and physical memory.
 *  * logical    - Region has no physical resource
 *  * faultable  - Region is logical, with pre-allocated page tables.
 *  * logical_sase SASE Region has no physical resource, default not allowed.
 *  * tlb_global   Set if region can be mapped TLB-global, default not allowed.
 *  * initial_dynamic
 *                 Sys confed region of type removable.
 *  * tlb_global_transient
 *                 Set if region is in a transient state, changing
 *                 from TLB-global to non TLB-global.
 *                 
 *
 */
#define MM_REGION_OPT_PERSISTENT           0x002
#define MM_REGION_OPT_INITIAL_STATIC       0x004
#define MM_REGION_OPT_MAPPED               0x008
#define MM_REGION_OPT_LOGICAL              0x010
#define MM_REGION_OPT_FAULTABLE            0x020
#define MM_REGION_OPT_LOGICAL_SASE         0x040
#define MM_REGION_OPT_TLB_GLOBAL           0x080
#define MM_REGION_OPT_INITIAL_DYNAMIC      0x100
#define MM_REGION_OPT_TLB_GLOBAL_TRANSIENT 0x200

/* Helper macros to examine meta-properties
 */
#define region_is_persistent(region) \
   (((region)->meta_options & MM_REGION_OPT_PERSISTENT) != 0)

#define region_is_static(region)  \
   (((region)->meta_options & MM_REGION_OPT_INITIAL_STATIC) != 0)

#define region_is_mapped(region)  \
   (((region)->meta_options & MM_REGION_OPT_MAPPED) != 0)

#define region_is_exclusive(region)  \
   log_object_exclusive(region->log_object)

#define region_is_logical(region)  \
   (((region)->meta_options & MM_REGION_OPT_LOGICAL) != 0)

#define region_is_faultable(region)  \
   (((region)->meta_options & MM_REGION_OPT_FAULTABLE) != 0)

#define region_is_physical(region)  \
   log_object_physical(region->log_object)

#define region_is_logical_sase(region)  \
   (((region)->meta_options & MM_REGION_OPT_LOGICAL_SASE) != 0)

#define region_is_resident(region)   \
   log_object_resident(region->log_object)

#define region_is_tlb_global(region)  \
   (((region)->meta_options & MM_REGION_OPT_TLB_GLOBAL) != 0)

#define region_is_initial_dynamic(region)  \
   (((region)->meta_options & MM_REGION_OPT_INITIAL_DYNAMIC) != 0)

#define region_is_tlb_global_transient(region)  \
   (((region)->meta_options & MM_REGION_OPT_TLB_GLOBAL_TRANSIENT) != 0)

/*
 * A region is connected to and owned by a domain. It
 * manages the relation between logical and physical memory. 
 */
struct region
{
   core_rbtree_node_t  node;         /* Linked into domain region tree */
   uint32_t            permissions;
   uint32_t            cache_mode;
   uint32_t            options;      /* Options from mm_create_region */
   uint32_t            meta_options; /* persistent, removable, active, ... */
   struct log_object  *log_object;   /* Logical memory */
   struct phys_object *phys_object;  /* Physical memory */
   char                name[MM_NAME_LENGTH + 1];
   char                log_mem[MM_NAME_LENGTH + 1];
   struct domain      *domain;
};

/*
 * Domain container.
 * Contains all information for a domain.
 */
struct domain
{
   struct partition  *partition_instances; /* List of private (MAS) partitions */
   struct hal_domain *hal_domain; /* HAL specific data. */
   core_rbtree_t      regions;    /* Tree of owned/shared regions. */
   list_t             phys_resources;  /* List of owned physical resources. */
   OSDOMAIN           segnum;     /* Domain identity. */
   U32                mas_cnt;    /* Number of MAS regions in domain */
   char name[MM_NAME_LENGTH + 1]; /* Name. */

#ifdef ASSERT
   U32 magic;
#endif
};



/*
 * Vector with all domains.
 * The domain identity (segnum) is used as a key.
 */
struct domains
{
      struct domain *domain_list[MM_MAX_DOMAINS];
      unsigned int   no_free;
      unsigned int   next_ix;
};

/*
 * Initialize sub-system.
 */
void
mm_domain_init(void);

/*
 * Create a domain descriptor and assign a segment number to it.
 * The segment number is contained in the domain descriptor.
 * A status value is returned.
 * If successful, the new domain is returned in the domain parameter
 * and the status returned is MM_SUCCESS.
 */
mm_status_t
mm_add_domain(struct domains  *domains,
              const char      *name,
              struct domain  **domain);

/*
 * Find a domain descriptor with a segment number as a key.
 * If
 *   1 <segnum> is invalid
 * or
 *   2 no domain descriptor is apssociated with the segment number
 * NULL is returned.
 */
struct domain *
mm_find_domain_by_domid(struct domains *domains,
			OSDOMAIN        segnum);

/*
 * Delete a domain and all associated regions.
 * The kernel domain can't be removed and this function will return
 * FALSE when operation on the kernel domain.
 * Otherwise TRUE is returned.
 *
 * The corresponding HAL domain will be deleted by a call to
 * hal_delete_domain.
 */
mm_bool_t
mm_remove_domain(struct domains *domains, OSDOMAIN segnum);

/*
 * mm_add_region()
 *
 * Creates a region and registers it with the selected domain.
 *
 * If <meta_options> has the MM_REGION_OPT_MAPPED bit set
 * the region is also mapped using mm_map_region().
 */
mm_status_t
mm_add_region(OSDOMAIN segnum,
              struct log_object *log_obj,
              struct phys_object *phys_obj,
              U32  permissions,
              U32  cache_mode,
              U32  options,
              U32  meta_options,
              const char *name,
              const char *log_mem,
              struct region **reg_out);

/*
 * Map region <region> in domain <segnum>.
 */
mm_status_t
mm_map_region(OSDOMAIN segnum,
              struct region *reg);

/*
 * Unmap region <ireg> in its domain.
 */
mm_status_t
mm_unmap_region(struct domain *domain, struct region *region);


/*
 * Look for a region.
 * Two types of search methods exist:
 *   1 exact match. <base> must match the base address of the region.
 *   2 inexact match. <base> must reside inside the region.
 * Returns the region with the given base address.
 * NULL is returned if no matching region is found.
 */
struct region *
mm_find_region_by_addr(struct domains *domains,
                       OSDOMAIN segnum,
                       log_addr_t base,
                       mm_bool_t exact_match);

/*
 * Removes the given region.
 *
 * Detach is performed on the logical object, and on the
 * physical object if there is one.
 *
 * If the parameter phys_object in "region" is not zero, the region will be
 * unmapped in HAL.
 */
void
mm_remove_region(struct domains *domains,
                 struct domain  *domain,
                 struct region  *region);


/*
 * Returns a list with struct domain_id with all domains
 * that has regions at the specified logical base address.
 * Only regions with non-NULL phys_object are considered.
 */
list_t *
mm_create_share_domains_list(struct domains *domains,
                             log_addr_t      base);

/*
 * Frees the list of struct domain_id:s as well as the list itself.
 */
void
mm_delete_share_domains_list(list_t *shared_list);


/*
 * Change active domain.
 */
void
mm_select_domain(U32 seg_id);

/*
 * Set option parameter for hal map/unmap calls.
 */
int
mm_set_hal_options(struct domain *domain,
                   struct region *region,
                   mm_bool_t change);

/*
 * Returns the name of the physical memory used for the region. If that can't
 * be extracteg (e.g. if region is a logical region), "" is returned.
 */
const char *
mm_get_phys_mem(struct region *);

/*
 * Allocate a physical resource to the given domain
 * using the allocator specified by name.
 * Returns MM_SUCCESS on successful operation.
 */
mm_status_t
domain_alloc_phys_resource(struct domains *domains,
                           OSDOMAIN segnum,
                           phys_addr_t *base,
                           phys_size_t size,
                           phys_addr_t align,
                           char *name);

/*
 * Free a physical resource from the given domain, making it available for
 * allocation from the associated allocator again.
 * Returns MM_SUCCESS on successful operation.
 * */
mm_status_t
domain_free_phys_resource(struct domains *domains,
                          OSDOMAIN segnum,
                          phys_addr_t base);

/* Check whether a physical fragment is (entirely) within a physical resource
 * which is allocated from the given domain.
 * Returns MM_SUCCESS on successful operation.
 */
mm_status_t
in_domain_phys_resource(struct domains *domains,
                        OSDOMAIN segnum,
                        phys_addr_t base,
                        phys_addr_t size);

/* Checks given range for overlap with any freed memory that originated
 * from a deleted initial region.
 * Returns TRUE if overlap found, otherwise FALSE.
 */
mm_bool_t
mm_initial_region_overlap(phys_addr_t base,  phys_size_t size);

#endif /* _DOMAIN_H */
