/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2006 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

#ifndef GMM_H_
#define GMM_H_

/* C runtime include files. */

/* OSE public include files. */
#include "list_ext.h"
#include "mm.sig"
#include "mm_internal.sig"
#include "pmm.sig"
#include "ose.h"
#include "osetypes.h"
#include "cpu.h"

/* OSE internal header files. */

/* Block internal include files. */
#include "partition.h"
#include "domain.h"
#include "hal.h"
#include "mm_access_internal.h"

#define MM_ULLONG_MAX      0xFFFFFFFFFFFFFFFFULL
#define MM_PHYS_SIZE_4GiB (0x0000000100000000ULL)

/* Name of always existing kernel domain. */
#define KERNEL_DOMAIN_NAME "System"

/* Default values used by GMM when mapping RAM for internal needs. */
#define DEFAULT_ACCESS_ATTR SU_RW_USR_NA
#define DEFAULT_CACHE_MODE  (MM_CACHE_COPY_BACK | MM_SPECULATIVE_ACCESS)

/*
 * Multithread locking. Uses the internal state of GMM to determine if
 * locks should be taken or not.
 */
#define MM_MULTITHREAD_LOCK(sr) \
           if (gmm_state.multithread_enable) { \
             LOCK_SAVE(sr); \
             (void)(sr); \
           } 

#define MM_MULTITHREAD_UNLOCK(sr) \
           if (gmm_state.multithread_enable) { \
             LOCK_RESTORE(sr); \
           } 

/* A pointer to the sempahore used to hinder more than one process to
 * enter the MM reclamation loop. */
extern MUTEX reclaim_region_mutex;

/*
 * MM states used from start to end.
 * The state transitions are monotonic from the initial value 'reset' to
 * the final value 'mmu_enabled'.
 */
enum mm_state {reset,                 /* initial state. */
               zone_ok,               /* mm_zalloc() operational .*/
               page_ok,               /* page_alloc() operational .*/
               logmem_entries_created,/* list with logmem entries created. */
               user_regions_created,  /* static regions created .*/
               hal_functional,        /* HAL interface fully functional. */
               regions_mapped,        /* static regions mapped .*/
               arch_init_mmu,         /* Optional per architecture state
                                       * that indicates a delicate
                                       * state when MM is taking over the
                                       * MMU from the boot. For example
                                       * polled_write in ramlog_printf are
                                       * stopped in this state. */
               arch_mmu_enabled,      /* Optional per architecture state. */
               mmu_enabled};          /* Generic MMU enabled (if applicable).*/


typedef enum mm_state mm_state_t;


/*
 * GMM 'world' struct.
 * Contains all state in GMM.
 */
struct gmm_vars
{
      /* List with all partition classes (log_mems). */
      struct partition_class *partition_classes;
      struct partition       *partition_instances;

      /* Array with all domains. */
      struct domains domains;

      /* List with all allocators. */
      struct allocator *allocators;

      /* HAL-specific information. Retrieved from HAL during boot. */
      const struct hal_info *hal_info;

      /* D-cache enable attribute. Retrieved from system parameters. */
      U32 d_cache;

      /* I-cache enable attribute. Retrieved from system parameters. */
      U32 i_cache;

      struct domain *system_domain;

      /* Rounded access mode for internal mappings. */
      U32 internal_access_mode;

      /* Current MM state. */
      mm_state_t state;

      /* Indicates that when data is accessable from freeze mode */
      OSBOOLEAN data_accessable;

      /* Global MAS counter. */
      U32 mas_cnt;

      PROCESS pmm;
   
      struct allocator *ram_allocator;

      /* Indicates whether multithreading is enabled and locks needed */
      U32 multithread_enable;

      /* The partition where logical resources for metadata is located. */
      struct partition_class *metadata_log_mem;

      U32  metadata_size;

      U32  metadata_size_high;

      PROCESS gmm_pid;

      /* Ordered list of deleted initial dynamic regions. */
      list_t deleted_init_dynamic_regions; 

      U32 max_domains;

      /* Disables mm_make_resident signals if a super page
       * threshold has been configured to avoid clashes in TLB1
       * on e500mv/v2. */
      U32 make_resident_disabled;
};

/*
 * This struct should contain the global states of GMM
 * The definition is located in gmm_common.c.
 */
extern struct gmm_vars gmm_state;

struct timepair
{
   long long start;
   long long end;
};

struct gmm_stats
{
   struct timepair open_exception_area;
   struct timepair install_exception_handlers;
   struct timepair connect_pmm;
};

extern struct gmm_stats gmm_stats;

int
mm_stat(int argc, char *argv[]);

void mm_pmm_free(log_addr_t base);

union SIGNAL *
mm_get_phys_mem_info(struct gmm_vars *, union SIGNAL *);

size_t
ose_mm_reclaim_component_memory(struct gmm_vars *gmm_state,
                                U32 required_size,
                                const char *phys_mem);

/*
 * Verifies that the permission bits are valid.
 */
mm_bool_t
mm_check_permission(U32 perm);

/*
 * Verifies that the cache mode bits are valid.
 */
mm_bool_t
mm_check_cache_mode(U32 cache_mode);

/*
 * Translate a HAL status code to MM status code.
 */
mm_status_t
mm_hal_status_to_mm_status(hal_status_t status);


/*
 * Function calls below corresponds to the functionality in the  mm.sig
 * signalling interface
 * It is needed for code executing in the GMM process which is located
 * in other files than gmm.c, since in this situation GMM cannot send
 * a signal to itself to resolve the functionality.
 * 
 * Prefix these functions gmm_ both for namespace reasons and to distinguish
 * from other mm function calls.
 */

/*
 * Create a region.
 * Parameters are described in mm.sig.
 * Only available from gmm process.
 *
 */

/* This define is needed when creating a bat mapping for the hash table.
 * 
 */
#define MM_CACHE_MEMORY_COHERENCE 0x020

/*
 * Overrides internal sanity checks in the MM and allows a
 * logical SASE region to be created.
 *
 * One use case is when PPC PEM Hash Table is mapped with DBAT and
 * doesn't need to be mapped in page tables.
 *
 * Use this option with care as it may prohibit all further allocations
 * of SASE memory from the covered phys_mem. To avoid that, make sure
 * that the covered physical memory is also allocated.
 */
#define MM_ALLOW_LOGICAL_SASE     0x040

/*
 * This option indicates that the extended cache mode bit format
 * is used. If set, the 8 bits located at MM_CACHE_SHIFT are set
 * to the internal index of the requested cache type configuration. 
 */
#define MM_CACHE_EXTENDED         0x080

/*
 * This option indicates that the extended cache mode bit format
 * is used. If set, the 8 bits located at MM_ACCESS_SHIFT are set
 * to the internal index of the requested memory type configuration. 
 */
#define MM_ACCESS_EXTENDED        0x100

/*
 * The cache type index is shifted this number of bits into the
 * cache_mode field, if MM_CACHE_EXTENDED is set.
 */
#define MM_CACHE_SHIFT            16

/*
 * The memory type index is shifted this number of bits into the
 * cache_mode field, if MM_ACCESS_EXTENDED is set.
 */
#define MM_ACCESS_SHIFT           24


mm_status_t
gmm_create_region(  
   U32                  *size,
   const OSDOMAIN       domain,
   const char           name[MM_NAME_LENGTH + 1],
   const char           phys_mem[MM_NAME_LENGTH + 1],
   const char           log_mem[MM_NAME_LENGTH + 1],
   U32                  *permissions,
   U32                  *cache_mode,
   OSADDRESS            *base,            
   U32                  *align,           /* optional, can be NULL */
   const U32             options,         /* optional, can be zero */
   const U32             low_limit,       /* optional, can be zero */
   U32                  *phys_count,      /* optional, can be NULL */
   struct phys_resource *phys_resources   /* optional, can be NULL */
   );

#if defined(OSE_SMP) && defined(OSPP)
hal_status_t  
get_mmu_stat_other_core(struct mm_tag_ext *tag);
#endif

#endif /* GMM_H_ */
