/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2005-2006 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */


/* OSE public include files. */
#include "assert.h"
#include "ose.h"
#include "hal.h"
#include "mm_err.h"
#include "mmtypes.h"
#include "common.h"
#include "gmm.h"

#include <string.h>

typedef hal_status_t (*INIT_FPTR)(const struct cpu_descriptor *cpu_descriptor,
                                  log_addr_t vector_base,
                                  log_size_t vector_size,
                                  mm_bool_t enable_d_cache,
                                  mm_bool_t enable_i_cache);

/*
 * see hal.h
 */
hal_status_t hal_validate_phys_range_32bit(phys_addr_t base,
                                           phys_size_t size)
{
   phys_addr_t top = base + size - 1ULL;

   if (size == 0)
   {
      return HAL_SUCCESS;
   }
   
   /* base > top => base + size has wrapped */
   if (0 != (top >> 32) || base > top)
   {
      ramlog_printf("Unsupported range [0x%08llx-0x%08llx]. "
                    "Supported range is [0x00000000-0xffffffff]\n",
                    base, top);
      return HAL_EUNSUPPORTED_PHYSICAL_RANGE;
   }
   
   return HAL_SUCCESS;
}

struct arch_cpu_descriptor *hal_arch_cpu_descriptor;

hal_status_t (*hal_init_2)(void);

hal_status_t (*hal_init_3)(struct partition *partition);

void (*hal_init_4)(void);

void (*hal_init_5)(void);

hal_status_t (*hal_validate_phys_range)(phys_addr_t base,
                                        phys_size_t size)
   = hal_validate_phys_range_32bit;
   
const struct hal_info * (*hal_get_info)(void);

hal_status_t (*hal_actual_permissions)(U32 *actual_permissions,
                                       U32 requested_permissions,
                                       log_addr_t base, U32 options);

/**************************************************************************/

static struct hal_cache_type hal_ctype[HAL_CTYPE_COUNT];
static struct hal_memory_type hal_mtype[HAL_MTYPE_COUNT];

/**************************************************************************/

static struct hal_cache_type *
get_cache_type(U32 rcache)
{
   struct hal_cache_type *c_type;

   if (!(rcache & MM_CACHE_EXTENDED))
   {
      switch (rcache & (MM_CACHE_INHIBITED|MM_CACHE_WRITE_THROUGH|MM_CACHE_COPY_BACK))
      {
         case MM_CACHE_INHIBITED:
            c_type = &hal_ctype[0];
            break;
         case MM_CACHE_WRITE_THROUGH:
            c_type = &hal_ctype[1];
            break;
         case MM_CACHE_COPY_BACK:
            c_type = &hal_ctype[2];
            break;
         default:
            return NULL;
      }
   }
   else
   {
      U32 i = (rcache >> MM_CACHE_SHIFT) & 0xFF;

      if (i>=HAL_CTYPE_COUNT)
         return NULL;

      c_type = &hal_ctype[i];

      if (!*c_type->name)
         return NULL;
   }

   return c_type;
}

/**************************************************************************/

static struct hal_memory_type *
get_memory_type(U32 rcache)
{
   struct hal_memory_type *m_type;

   if (!(rcache & MM_ACCESS_EXTENDED))
   {
      switch (rcache & (MM_CACHE_MEMORY_COHERENCE|
                        MM_SPECULATIVE_ACCESS|
                        MM_ORDERED_ACCESS))
      {
         case MM_ORDERED_ACCESS:
            m_type = &hal_mtype[0];
            break;
         case MM_SPECULATIVE_ACCESS:
            m_type = &hal_mtype[1];
            break;
         case MM_CACHE_MEMORY_COHERENCE|MM_SPECULATIVE_ACCESS:
            m_type = &hal_mtype[2];
            break;
         default:
            return NULL;
      }
   }
   else
   {
      U32 i = (rcache >> MM_ACCESS_SHIFT) & 0xFF;

      if (i>=HAL_MTYPE_COUNT)
         return NULL;

      m_type = &hal_mtype[i];

      if (!*m_type->name)
         return NULL;
   }

   return m_type;
}

/**************************************************************************/

static hal_status_t 
mmu_hal_get_cmbits(U32 cache_mode, U32 *bits)
{
   struct hal_cache_type *c_type = get_cache_type(cache_mode);
   struct hal_memory_type *m_type = get_memory_type(cache_mode);

   if (c_type == NULL || m_type == NULL)
      return HAL_EILLEGAL_CACHE_MODE;

   *bits = c_type->cache | m_type->memory;

   return HAL_SUCCESS;
}

/**************************************************************************/

#define old_cache_mode(cache) \
  ((cache) & (MM_SPECULATIVE_ACCESS|MM_ORDERED_ACCESS \
  MM_CACHE_INHIBITED|MM_CACHE_WRITE_THROUGH|MM_CACHE_COPY_BACK))

/**************************************************************************/

static hal_status_t
mmu_hal_add_ctype(const char *name, U32 cache, U32 options)
{
   size_t i;

   if (!strcmp(name, "inhibited"))
   {
      struct hal_cache_type *type = &hal_ctype[0];
      strlcpy(type->name, name, MM_NAME_LENGTH);
      type->cache = (options & HAL_APPEND) ? (type->cache | cache) : cache;
      type->index = 0;
      return hal_valid_cmbits(&type->cache, NULL, NULL, 0);
   }

   if (!strcmp(name, "write_through"))
   {
      struct hal_cache_type *type = &hal_ctype[1];
      strlcpy(type->name, name, MM_NAME_LENGTH);
      type->cache = (options & HAL_APPEND) ? (type->cache | cache) : cache;
      type->index = 1;
      return hal_valid_cmbits(&type->cache, NULL, NULL, 0);
   }

   if (!strcmp(name, "copy_back"))
   {
      struct hal_cache_type *type = &hal_ctype[2];
      strlcpy(type->name, name, MM_NAME_LENGTH);
      type->cache = (options & HAL_APPEND) ? (type->cache | cache) : cache;
      type->index = 2;
      return hal_valid_cmbits(&type->cache, NULL, NULL, 0);
   }

   for (i=0; i<HAL_CTYPE_COUNT; i++)
   {
      struct hal_cache_type *type = &hal_ctype[i];

      if (!strcmp(type->name, name) && (options & HAL_APPEND))
      {
         /* Appending a new level to an existing cache type. */
         type->cache |= cache;
         return hal_valid_cmbits(&type->cache, NULL, NULL, 0);
      }
      else if (!*type->name && !(options & HAL_APPEND))
      {
         /* Adding a new cache type. */
         strlcpy(type->name, name, MM_NAME_LENGTH);
         type->cache = cache;
         type->index = i;
         return hal_valid_cmbits(&type->cache, NULL, NULL, 0);
      }
   }

   return HAL_EOUT_OF_RESOURCES;
}

/**************************************************************************/

static hal_status_t
mmu_hal_add_mtype(const char *name, U32 memory)
{
   size_t i;

   if (!strcmp(name, "ordered_access"))
   {
      struct hal_memory_type *type = &hal_mtype[0];
      strlcpy(type->name, name, MM_NAME_LENGTH);
      type->memory = memory;
      type->index = 0;
      return hal_valid_cmbits(NULL, &type->memory, NULL, 0);
   }

   if (!strcmp(name, "speculative_access"))
   {
      struct hal_memory_type *type = &hal_mtype[1];
      strlcpy(type->name, name, MM_NAME_LENGTH);
      type->memory = memory;
      type->index = 1;
      return hal_valid_cmbits(NULL, &type->memory, NULL, 0);
   }

   if (!strcmp(name, "speculative_coherent"))
   {
      struct hal_memory_type *type = &hal_mtype[2];
      strlcpy(type->name, name, MM_NAME_LENGTH);
      type->memory = memory;
      type->index = 2;
      return hal_valid_cmbits(NULL, &type->memory, NULL, 0);
   }

   for (i=0; i<HAL_MTYPE_COUNT; i++)
   {
      struct hal_memory_type *type = &hal_mtype[i];

      if (!*type->name)
      {
         strlcpy(type->name, name, MM_NAME_LENGTH);
         type->memory = memory;
         type->index = i;
         return hal_valid_cmbits(NULL, &type->memory, NULL, 0);
      }
   }

   return HAL_EOUT_OF_RESOURCES;
}

/**************************************************************************/

static hal_status_t
mmu_hal_get_cache_mode(const char *cache, const char *memory, U32 *cache_mode)
{
   struct hal_cache_type *c_type = hal_ctype;
   struct hal_memory_type *m_type = hal_mtype;
   size_t i;

   if (cache == NULL || memory == NULL || cache_mode == NULL)
      return HAL_ENOT_FOUND;

   /* Translate legacy keywords (remove optional cache_-prefix). */
   if (strncmp("cache_", cache, sizeof("cache_") - 1) == 0)
   {
      if (strcmp("cache_inhibited", cache) == 0 ||
          strcmp("cache_write_through", cache) == 0 ||
          strcmp("cache_copy_back", cache) == 0)
      {
         cache += sizeof("cache_") - 1;
      }
   }
   
   *cache_mode = 0;

   for (i=0; *c_type->name && i<HAL_CTYPE_COUNT; i++, c_type++)
   {
      if(!strcmp(c_type->name, cache))
      {
         *cache_mode |= (i << MM_CACHE_SHIFT);
         *cache_mode |= MM_CACHE_EXTENDED;
         break;
      }
   }

   if (!(*c_type->name) || i == HAL_CTYPE_COUNT) return HAL_ENOT_FOUND;

   for (i=0; *m_type->name && i<HAL_MTYPE_COUNT; i++, m_type++)
   {
      if(!strcmp(m_type->name, memory))
      {
         *cache_mode |= (i << MM_ACCESS_SHIFT);
         *cache_mode |= MM_ACCESS_EXTENDED;
         break;
      }
   }
   
   if (!(*m_type->name) || i == HAL_CTYPE_COUNT) return HAL_ENOT_FOUND;

   /* Add legacy bits. */
   return hal_legacy_cmbits(c_type->cache|m_type->memory, cache_mode);
}

/**************************************************************************/

static hal_status_t
mmu_hal_get_cache_type(U32 cache_mode, char *cache, char *memory)
{
   struct hal_cache_type *c_type = get_cache_type(cache_mode);
   struct hal_memory_type *m_type = get_memory_type(cache_mode);

   if (c_type == NULL || m_type == NULL)
      return HAL_EILLEGAL_CACHE_MODE;

   if (cache != NULL)  strlcpy(cache,  c_type->name, MM_NAME_LENGTH);
   if (memory != NULL) strlcpy(memory, m_type->name, MM_NAME_LENGTH);
   
   return HAL_SUCCESS;   
}

/**************************************************************************/

static hal_status_t
mmu_hal_actual_cache_mode(U32 *acache, U32 rcache, log_addr_t base)
{
   struct hal_cache_type *c_type = get_cache_type(rcache);
   struct hal_memory_type *m_type = get_memory_type(rcache);
   hal_status_t status;
   U32 bits;
   
   (void)base;

   if (c_type == NULL || m_type == NULL)
      return HAL_EILLEGAL_CACHE_MODE;
   
   bits = c_type->cache | m_type->memory;

   status = hal_valid_cmbits(NULL, NULL, &bits, base);
   if (status != HAL_SUCCESS)
      return status;

   /* Create new cache mode ... */
   *acache = 0;
   *acache |= MM_ACCESS_EXTENDED;
   *acache |= MM_CACHE_EXTENDED;
   *acache |= c_type->index << MM_CACHE_SHIFT;
   *acache |= m_type->index << MM_ACCESS_SHIFT;

   /* Add legacy bits. */
   return hal_legacy_cmbits(bits, acache);
}

/**************************************************************************/

hal_status_t
(*hal_legacy_cmbits)(U32 cmbits, U32 *cache_mode) = NULL;

/**************************************************************************/

hal_status_t
(*hal_get_cmbits)(U32 cmbits, U32 *cache_mode)
   = mmu_hal_get_cmbits;


/**************************************************************************/

hal_status_t
(*hal_actual_cache_mode)(U32 *actual_cache_mode,
                         U32 requested_cache_mode,
                         log_addr_t base)
   = mmu_hal_actual_cache_mode;

/**************************************************************************/

hal_status_t
(*hal_add_ctype)(const char *name, U32 cache, U32 options)
   = mmu_hal_add_ctype;

/**************************************************************************/

hal_status_t
(*hal_add_mtype)(const char *name, U32 memory)
   = mmu_hal_add_mtype;

/**************************************************************************/
   
hal_status_t
(*hal_get_cache_mode)(const char *cache, const char *mem, U32 *cache_mode)
   = mmu_hal_get_cache_mode;

/**************************************************************************/

hal_status_t
(*hal_get_cache_type)(U32 cache_mode, char *cache, char *mem)
   = mmu_hal_get_cache_type;

/**************************************************************************/
hal_status_t
(*hal_valid_cmbits)(U32 *cbits, U32 *mbits, U32 *cmbits, log_addr_t base)
   = NULL;

/**************************************************************************/

hal_status_t (*hal_create_domain)(unsigned int seg_id,
                                  mm_bool_t system_domain,
                                  struct hal_domain **domain);

void (*hal_delete_domain)(unsigned int seg_id,
                          struct hal_domain *domain);

void (*hal_select_domain)(struct hal_domain *hal_domain);

hal_status_t (*hal_map_page)(struct hal_domain *domain,
                             log_addr_t la,
                             phys_addr_t pa,
                             log_size_t size,
                             int options,
                             U32 access_attr,
                             U32 mem_attr);


void (*hal_unmap_range)(struct hal_domain *domain,
                        log_addr_t         la,
                        log_size_t         size,
                        int                options);


hal_status_t (*hal_lookup)(OSDOMAIN          domain,
			   log_addr_t        log_addr,
			   phys_addr_t       *page_base,
			   log_size_t        *page_size,
			   U32               *perm,
                           U32                options);


hal_status_t (*hal_get_mmu_statistics)(struct mm_tag_ext *tag);
   
hal_status_t
(*hal_commit_range)(struct hal_domain *domain,
                    log_addr_t base, log_size_t size, U32 options);

hal_status_t
(*hal_uncommit_range)(struct hal_domain *domain,
                      log_addr_t base, log_size_t size, U32 options);


#ifdef OSPP
void (*hal_invalidate_tlb_range)(log_addr_t base,
                                 log_size_t size,
                                 struct hal_domain *domain,
                                 mm_bool_t domain_invariant);
#endif

void (*hal_invalidate_tlb_local)(log_addr_t base, log_size_t size) = NULL;

void (*hal_invalidate_tlb)(void) = NULL;

hal_status_t
(*hal_create_resident_mapping)(struct hal_domain *domain,
                               log_addr_t la,
                               phys_addr_t pa,
                               log_size_t size,
                               U32 options,
                               U32 access_attr,
                               U32 mem_attr) = NULL;

hal_status_t
(*hal_delete_resident_mapping)(struct hal_domain *domain,
                               log_addr_t la,
                               log_size_t size,
                               U32 options) = NULL;

hal_status_t
(*hal_delete_resident_mappings)(log_addr_t la,
                                log_size_t size) = NULL;

#ifdef OSARM

void (*armhal_enable_static_page_tables)(void);

void (*armhal_disable_static_page_tables)(void);

#endif

extern void zzkrn_halted_hook(void);

#ifdef _BullseyeCoverage
#pragma BullseyeCoverage off
#endif

static hal_status_t
illegal_hal_init_1(const struct cpu_descriptor *cpu_descriptor,
                   log_addr_t vector_base,
                   log_size_t vector_size,
                   mm_bool_t enable_d_cache,
                   mm_bool_t enable_i_cache)
{
   (void)cpu_descriptor;
   (void)vector_base;
   (void)vector_size;
   (void)enable_d_cache;
   (void)enable_i_cache;

   MM_FATAL_ERROR2(("No HAL module installed, probably too old krncon.c!\n"),
                   OSE_MM_EHAL_SUBHAL_NO_MATCH,
                   (OSERRCODE) 0);
   zzkrn_halted_hook();
   return HAL_SUCCESS;
}

#ifdef _BullseyeCoverage
#pragma BullseyeCoverage on
#endif

INIT_FPTR hal_init_1 = illegal_hal_init_1;
