/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2006 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

#include "hash.h"
#include "ose_err/mm_err.h"
#include "ramlog.h"
#include "ose.h"
#include "heapapi.h"
#ifdef STANDALONE
#include "standalone.h"
#endif


/************************** PROTOTYPES **************************/

static uint32_t
hash_function(struct hash_table_t *hash, uint32_t value);

/************************** STATIC VARIABLES **************************/

/* A reference to the system heap that we need to allocate everything from. */
static OSHEAP system_heap = 0;


/*
 *===========================================================================
 *                         init_hash
 *===========================================================================
 *
 * Description
 * -----------
 * Initializes a hash table with a specific compare function and
 * hash value generator.
 * Needs to be called before any other operation on the hash table.
 *
 * Parameters
 * ----------
 * no_of_buckets  The number of buckets in this hash table.
 *                Statically defined at initialization.
 * compare        Callback function for comparing two nodes. This
 *                function returns 0 if key_a does not match key_b,
 *                and 1 if key_a equals key_b.
 * get_hash_value Returns a 32 bit value given an entry. Try to make this
 *                function return a value that spreads the entries over
 *                all the hash table buckets.
 *
 */
void
init_hash(struct hash_table_t *hash, uint32_t no_of_buckets,
          get_hash_value_func *get_hash_value, compare_func *compare)
{
   uint32_t i;

   if (no_of_buckets == 0)
   {
      ramlog_printf("Bucket size in hash cannot be zero.\n");
      error2(OSE_PAGING_INTERNAL_ERROR, 0);
   }

   if (!system_heap)
   {
      PROCESS pid;

      hunt("osemain", 0, &pid, NULL);
      if (pid == 0)
      {
         hunt("main", 0, &pid, NULL);
         if (pid == 0)
         {
            ramlog_printf("ERROR: Could not find the osemain process.\n");
            error2(OSE_PAGING_INTERNAL_ERROR, 0);
         }
      }
      system_heap = ose_heap_ref(pid);
      if (system_heap == 0)
      {
         ramlog_printf("ERROR: Could not find the system heap.\n");
         error2(OSE_PAGING_INTERNAL_ERROR, 0);
      }
   }

   hash->no_of_buckets = no_of_buckets;
   hash->get_hash_value = get_hash_value;
   hash->compare = compare;
   hash->bucket = (list_t *)ose_heap_malloc(system_heap,
                                            no_of_buckets * sizeof(list_t),
                                            __FILE__, __LINE__);
   if (!hash->bucket)
   {
      ramlog_printf("Ran out of memory initializing hash.\n");
      error2(OSE_PAGING_INTERNAL_ERROR, 0);
   }
   for (i = 0; i < no_of_buckets; i++)
   {
      init_list(&hash->bucket[i]);
   }
}

/*
 *===========================================================================
 *                         delete_hash
 *===========================================================================
 *
 * Description
 * -----------
 * Deletes a hash table. It does not delete any of the entries in
 * the hash table.
 *
 */
void
delete_hash(struct hash_table_t *hash)
{
   hash->no_of_buckets = 0;
   ose_heap_free(system_heap, hash->bucket);
}

/*
 *===========================================================================
 *                        hash_function
 *===========================================================================
 *
 * Description
 * -----------
 * Returns the correct bucket to search in.
 *
 * Returns
 * -------
 * uint32_t   Returns the correct bucket to search in.
 *
 */
static uint32_t
hash_function(struct hash_table_t *hash, uint32_t value)
{
   value += (value << 3);
   value ^= (value >> 11);
   value += (value << 15);
   return value % hash->no_of_buckets;
}

/*
 *===========================================================================
 *                        find_in_hash
 *===========================================================================
 *
 * Description
 * -----------
 * Searches for a matching entry in the hash table *hash.
 *
 * Returns
 * -------
 * struct list_link_t *   Returns the found entry or NULL if it is not found.
 *
 */
list_link_t *
find_in_hash(struct hash_table_t *hash, const void *key)
{
   uint32_t what_bucket = hash_function(hash, hash->get_hash_value(key));
   list_link_t *entry = list_head(&hash->bucket[what_bucket]);

   while (entry)
   {
      if (hash->compare(entry, key))
      {
         return entry;
      }
      entry = list_next(&hash->bucket[what_bucket], entry);
   }
   return NULL;
}

/*
 *===========================================================================
 *                        iterate_over_hash
 *===========================================================================
 *
 * Description
 * -----------
 * Iterates over the hash. If the iterator is initialized with the
 * INIT_ITERATOR() macro, the search starts at
 * the beginning of the hash table, otherwise at the entry after the
 * point the iterator points to. The iterator is modified to point to
 * the position of the mathching entry by this function.
 * It returns an entry matching the key and uses the provided compare function.
 * By repeatedly calling this function
 * and supplying the iterator every time, you will iterate over all matching
 * entries in the hash table.
 * If the compare function is NULL, any entry is considered to match.
 *
 * Returns
 * -------
 * struct list_link_t *   Returns the found entry or NULL if it is not found.
 *
 */
list_link_t *
iterate_over_hash(struct hash_table_t *hash, const void *key,
                  hash_iterator *iterator, compare_func *compare)
{
   if (iterator->bucket == NOT_INITIALIZED)
   {
      iterator->bucket = 0;
      iterator->entry = NULL;
   }
   else
   {
      iterator->entry = list_next(&hash->bucket[iterator->bucket],
                                  iterator->entry);
      if (!iterator->entry)
      {
         iterator->bucket++;
         iterator->entry = NULL; 
      }
   }

   while (iterator->bucket < hash->no_of_buckets)
   {
      if (!iterator->entry)
      {
         iterator->entry = list_head(&hash->bucket[iterator->bucket]);
      }

      while (iterator->entry)
      {
         if (!compare || compare(iterator->entry, key))
         {
            return iterator->entry;
         }
         iterator->entry = list_next(&hash->bucket[iterator->bucket],
                                     iterator->entry);
      }

      iterator->bucket++;
      iterator->entry = NULL;
   }

   return NULL;
}

/*
 *===========================================================================
 *                        insert_in_hash
 *===========================================================================
 *
 * Description
 * -----------
 * Inserts an entry in the hash_table *hash unless it is already there.
 *
 * Returns
 * -------
 * int      Returns one if the entry was inserted,
 *          or zero is it is already in a hash table.
 *
 */
int
insert_in_hash(struct hash_table_t *hash, list_link_t *entry)
{
   uint32_t what_bucket;
   
   if (entry->next != NULL || entry->prev != NULL)
   {
      return 0;
   }

   what_bucket = hash_function(hash, hash->get_hash_value(entry));
   insert_list_head(&hash->bucket[what_bucket], entry);
   return 1;
}

/*
 *===========================================================================
 *                        delete_from_hash
 *===========================================================================
 *
 * Description
 * -----------
 * Deletes an entry in the hash table *hash.
 *
 * Returns
 * -------
 * int     Returns one if the entry was deleted,
 *         or zero if it could not be found.
 *
 */
int
delete_from_hash(struct hash_table_t *hash, list_link_t *entry)
{
   uint32_t what_bucket;
   
   if (entry->next == NULL || entry->prev == NULL)
   {
      return 0;
   }

   what_bucket = hash_function(hash, hash->get_hash_value(entry));
   delete_from_list(&hash->bucket[what_bucket], entry);
   return 1;
}

#if 0
/* For debug purposes only. */
static int
validate_hash(struct hash_table_t *hash)
{
   uint32_t i;
   Msr sr;

   spin_lock_irqsave(&hash->bucket[0].lock, sr);
   for (i = 0; i < hash->no_of_buckets; i++)
   {
      list_link_t *entry = list_head(&hash->bucket[i]);

      while (entry)
      {
         struct page_metadata_t *pgm = (struct page_metadata_t *)entry;

         if (pgm->valid == 0)
         {
            ramlog_printf("Validation of hash failed.\n");
            error2(OSE_PAGING_INTERNAL_ERROR, 0);
            return 0;
         }
         entry = list_next(&hash->bucket[i], entry);
      }
   }
   spin_unlock_irqrestore(&hash->bucket[0].lock, sr);

   return 1;
}
#endif
