/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2008 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

#include "partition.h"
#include "implicit_partition.h"

#include "assert.h"
#include "configure.h"
#include "common.h"
#include "string.h"
#include "zone.h"

#ifdef OSMIPS
#  include "cpu_access.h"
#endif

/*****************************************************************************/

struct implicit_partition
{
   struct partition root;
};

static struct zone implicit_zone;  /* Implicit partition descriptor zone  */

/*****************************************************************************/

void
mm_implicit_partition_init()
{
   mm_zinit(&implicit_zone, "struct implicit_partition",
            sizeof(struct implicit_partition), 10, 20);
}

/*****************************************************************************/

static int
implicit_is_superset_of(const struct partition_class *p_class,
                        log_addr_t base, log_size_t size)
{
   struct allocator *pa = mm_find_allocator_by_range(base, size);
   const struct partition_class *part;
   
   (void)p_class;

   /* First verify that it does not overlap any explicit log_mem. */
   for (part = gmm_state.partition_classes; part != NULL; part = part->next)
   {
      if (!partition_is_implicit(part) && part->overlap(part, base, size))
      {
         return 0;
      }
   }
   
   /* This might not be entirely accurate. If the range is completely covered
    * by physical memory, but not of the same type, pa->sub_area will return 0.
    */
   if (NULL == pa)
   {
      return 0;
   }
   if (pa->sub_area(pa, base, size))
   {
      return 1;
   }

   return 0;
}

/*****************************************************************************/

static int
implicit_is_subset_of(const struct partition_class *p_class,
                      log_addr_t base, log_size_t size)
{
   /* FIXME
    *  This function is only used for MIPS and then only to determine if a
    * partition is inside KUSEG, KSEG0 or KSEG1. In the case of an implicit
    * partition in a MIPS system, the partition is always a subset of KSEG0.
    * One day a proper implementation ight be needed.
    */
   
   
#ifdef OSMIPS
   if (base == 0 && size >= KSEG0_SIZE)
   {
      return 1;
   }
   
#else

   (void)p_class;
   (void)base;
   (void)size;

   INTERNAL_ERROR();   
#endif
   /*lint -save -e527 "Unreachable code". */
   return 0;
   /*lint -restore */
}

/*****************************************************************************/

static int
implicit_overlap(const struct partition_class *p_class,
                 log_addr_t base, log_size_t size)
{
   /* is_superset_of() is not fully equivalent with overlap()
    * except if size == 1, which current is the only use-case for overlap().
    *
    * If another usage should arise, trigger an error and let
    * the developer implement a proper overlap() function. */
   if (size != 1)
   {
      INTERNAL_ERROR();
   }
   return implicit_is_superset_of(p_class, base, size);
}

/*****************************************************************************/

static int
implicit_is_subset_of_phys_mem(const struct partition_class *p_class)
{
   /* This log_mem is defined as all log_mem that is covered by physical
    * memory and that does not belong to any other log_mem => This log_mem's
    * memory is always a subarea of phys_mem.
    */

   (void)p_class;
   return 1;
}


/*****************************************************************************
 * See partition.h
 */
static struct partition *
implicit_create(struct partition_class *p_class)
{
   struct implicit_partition *partition;

   partition = mm_zalloc(&implicit_zone);
   if (partition == NULL)
   {
      return NULL;
   }

   partition->root.next = NULL;
   partition->root.class = p_class;
   
   return &partition->root;
}


/*****************************************************************************
 * See partition.h
 */
static void
implicit_destroy(struct partition *partition)
{
   mm_zfree(&implicit_zone, partition);
}

/*****************************************************************************
 * See partition.h
 */
static int
implicit_get_bank_info(const struct partition_class *p_class, int index,
                       log_addr_t *base, log_size_t *size)
{
   (void)p_class;
   (void)index;
   (void)base;
   (void)size;
   /* Implicit partitions don't have banks yet. */
   return 0;
}

/*****************************************************************************
 * See partition.h
 */
static mm_status_t
dummy_alloc_regular(struct partition  *partition,
                    struct log_object *log_obj,
                    log_size_t size, log_size_t alignment, uint32_t options)
{
   (void)partition;
   (void)log_obj;
   (void)size;
   (void)alignment;
   (void)options;   
   
   MM_FATAL_ERROR2(("dummy_alloc_regular.\n"), OSE_MM_EINTERNAL_ERROR, 0);
   
   return MM_SUCCESS;
}


/*****************************************************************************
 * See partition.h
 */
static mm_status_t
implicit_alloc_fixed(struct partition  *partition,
                     struct log_object *log_obj,
                     log_size_t size, log_addr_t fixed, uint32_t options)
{
   assert(log_obj != NULL);

   (void)partition;
   (void)options;
   
   log_obj->base = fixed;
   log_obj->size = size;
   log_obj->options = MM_PARTITION_OPT_SASE;
   log_obj->freemem = (struct freemem *)~0UL;
      
   return MM_SUCCESS;
}

/*****************************************************************************
 * See partition.h
 */
static void
implicit_free(struct partition *partition,
             struct log_object *log_obj, uint32_t options)
{
   (void)partition;
   (void)options;

   log_obj->base = 0;
   log_obj->size = 0;
   log_obj->options = 0;
   log_obj->freemem = NULL;
}


/*****************************************************************************
 * See partition.h
 */
static struct partition_class *
implicit_init(uint32_t options)
{
   struct partition_class *p_class = NULL;

   assert(options == (MM_PARTITION_OPT_IMPLICIT  |
                      MM_PARTITION_OPT_EXCLUSIVE |
                      MM_PARTITION_OPT_PHYSICAL));
   
   if (!mm_alloc_boot_space(sizeof *p_class, (log_addr_t*)&p_class)) /*lint !e740 */
   {
      MM_FATAL_ERROR2(("Boot heap exhausted.\n"), 0, 0);
   }

   if (p_class == NULL)
      return NULL;
   
   memset(p_class, 0, sizeof *p_class);

   /* Initialise it. */
   p_class->options = options;

   /* Initialize functions */
   p_class->init                  = implicit_init;
   p_class->init_bank             = NULL;
   p_class->create                = implicit_create;
   p_class->destroy               = implicit_destroy;
   p_class->is_subset_of          = implicit_is_subset_of;
   p_class->is_superset_of        = implicit_is_superset_of;
   p_class->is_subset_of_phys_mem = implicit_is_subset_of_phys_mem;
   p_class->overlap               = implicit_overlap;
   p_class->get_bank_info         = implicit_get_bank_info;
   p_class->alloc                 = dummy_alloc_regular;
   p_class->alloc_f               = implicit_alloc_fixed;
   p_class->free                  = implicit_free;

   return p_class;
}

/*****************************************************************************/


struct partition_class *
mm_add_implicit_sase_partition(const char *name)
{
   struct partition_class *part;
   
   assert(name != NULL);
   
   part = implicit_init(MM_PARTITION_OPT_IMPLICIT  |
                        MM_PARTITION_OPT_SASE);

   if (part == NULL)
      return NULL;

   strlcpy(part->name, name, MM_NAME_LENGTH);
   
   /* Insert new partition into list with partitions. */
   part->next = gmm_state.partition_classes;
   gmm_state.partition_classes = part;

   return part;
}
