/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2007 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

/**
 * @private
 * 
 * @file mfmi.c
 *
 * @brief Memory manager Freeze Mode Interface
 *
 * @long
 * This file implements the Memory manager Freeze Mode Interface.
 * It may be used to fetch information from the executive in a freeze mode
 * environment. The purpose is to make Memory concepts accessable
 * for freeze mode debug monitors that can't use ordinary system calls.
 * @long
 * WARNING: Use of mfmi functionality for any other purpose will severely
 * damage the interrupt latency!
 *
 */

/*
*****************************************************************************
** INCLUDE FILES
*****************************************************************************
*/

/* Make sure we don't use debug versions of system calls internally. */

#ifdef OS_DEBUG
#undef OS_DEBUG
#endif

#include "gmm.h"
#include "domain.h"
#include "allocator.h"
#include "mfmi.h"		/* Public interface */
#include "string.h"		/* str* and mem* functions */
#include "cpu_intlock.h"	/* Interrupt lock save and restore */

extern struct gmm_vars gmm_state;

/*
*****************************************************************************
** MACROS
*****************************************************************************
*/

/*
*****************************************************************************
** TYPES
*****************************************************************************
*/

static MfmiStatus get_next_region(struct MfmiRegionCursor *rc,
				  struct MfmiRegionInfo   *ri);

/*
 *---------------------------------------------------------------------------
 * Size of the local cursor structs must not exeed the size of the public
 * view of the corresponding struct, found in pfmi.h!
 *---------------------------------------------------------------------------
 */

struct DomainCursor
{
  OSDOMAIN domain;
};

struct RegionCursor
{
  struct domain *domain_desc;
  struct region *region_desc;
};

struct PhysMemTypeCursor
{
   struct allocator *allocator;
};

/*
*****************************************************************************
** LOCAL FUNCTIONS
*****************************************************************************
*/

static MfmiStatus
get_domain_info(OSDOMAIN domain,
		struct MfmiDomainInfo *di)
{
  struct domain *domain_desc;

  /* Check if data is locked by MM process */
  if (!gmm_state.data_accessable)
    return MFMI_DATA_LOCKED;

  /* Search for domain */
  domain_desc = mm_find_domain_by_domid(&gmm_state.domains, domain);

  if (domain_desc == NULL)
    return MFMI_EDOMAIN;

  /* Copy data */
  di->domain = domain;
  strncpy(di->name, domain_desc->name, sizeof(domain_desc->name));

  return MFMI_OK;
}


static MfmiStatus
get_region_info(OSDOMAIN domain,
		OSADDRESS address,
		struct MfmiRegionInfo *ri)
{
  struct domain *domain_desc;
  struct region *region_desc;
  struct log_object *lobj;
  struct partition *partition;

  /* Check if data is locked by MM process */
  if (!gmm_state.data_accessable)
    return MFMI_DATA_LOCKED;

  /* Search for domain */
  domain_desc = mm_find_domain_by_domid(&gmm_state.domains, domain);

  if (domain_desc == NULL)
    return MFMI_EDOMAIN;

  /* Search for region */
  region_desc = mm_find_region_by_addr(
     &gmm_state.domains, domain, address, TRUE);
  if (region_desc == NULL)
    return MFMI_EREGION;

  lobj = region_desc->log_object;

  partition = mm_log_object_partition(lobj, domain_desc);

  /* Copy data */
  ri->base        = lobj->base;
  ri->size        = lobj->size;;
  ri->permissions = region_desc->permissions;;
  ri->cache_mode  = region_desc->cache_mode;

  strncpy(ri->name,     region_desc->name,            sizeof ri->name);
  strncpy(ri->phys_mem, mm_get_phys_mem(region_desc), sizeof ri->phys_mem);
  strncpy(ri->log_mem,  partition->class->name,       sizeof ri->log_mem);
  
  ri->options = 0;
  if (ri->name[0] != '\0')
    ri->options |= MM_CREATE_REGION_SPECIFY_NAME;
  if (PHYS_OBJ_CLIENT_PAGES == region_desc->phys_object->data_representation)
    ri->options |= MM_CREATE_REGION_SPECIFY_PHYS;

  return MFMI_OK;
}

static MfmiStatus
get_phys_mem_type_info(struct MfmiPhysMemTypeInfo *pmti,
                       struct allocator *allocator)
{

   /* Check if data is locked by MM process */
   if (!gmm_state.data_accessable)
   {
      return MFMI_DATA_LOCKED;
   }

   if (NULL == allocator)
   {
      return MFMI_UNAVAILABLE;
   }

   strncpy(pmti->name, allocator->name, sizeof pmti->name);
   pmti->total_size = allocator->get_total_mem(allocator);
   pmti->free_size = allocator->get_free_mem(allocator);
   pmti->max_frag_size = allocator->get_biggest_frag(allocator);
   pmti->inactive_size = 0;
   pmti->free_watermark = allocator->get_free_watermark(allocator);

   return MFMI_OK;
} /* get_phys_mem_type_info(...) */

static MfmiStatus
get_first_phys_mem_type(struct MfmiPhysMemTypeCursor *mpmtc,
                        struct MfmiPhysMemTypeInfo *mpmti)
{
   struct PhysMemTypeCursor *pmtc = (struct PhysMemTypeCursor *)mpmtc;

   /* Check if data is locked by MM process */
   if (!gmm_state.data_accessable)
   {
      return MFMI_DATA_LOCKED;
   }

   pmtc->allocator = gmm_state.allocators;

   if (NULL == pmtc->allocator)
   {
      return MFMI_END_OF_LIST;
   }

   return get_phys_mem_type_info(mpmti, pmtc->allocator);
} /* get_first_phys_mem_type(...) */
                        

static MfmiStatus
get_next_phys_mem_type(struct MfmiPhysMemTypeCursor *mpmtc,
                       struct MfmiPhysMemTypeInfo *mpmti)
{
   struct PhysMemTypeCursor *pmtc = (struct PhysMemTypeCursor *)mpmtc;

   /* Check if data is locked by MM process */
   if (!gmm_state.data_accessable)
   {
      return MFMI_DATA_LOCKED;
   }

   if (NULL != pmtc)
   {
      pmtc->allocator = pmtc->allocator->next;
   }

   if (NULL == pmtc->allocator)
   {
      return MFMI_END_OF_LIST;
   }
   
   return get_phys_mem_type_info(mpmti, pmtc->allocator);
} /* get_next_phys_mem_type_info(...) */

/* FIXME Now there can be more than one ram allocator. How shall
 * get_memory_info handle that? What shall be returned? */
static MfmiStatus
get_memory_info(enum MfmiInfoTag tag,
		mm_size_t *info)
{
  struct allocator *allocator;
  struct mm_get_phys_mem_info mem_info;
  struct mm_get_phys_mem_info *mem_info_p = &mem_info;

  /* Check if data is locked by MM process */
  if (!gmm_state.data_accessable)
    return MFMI_DATA_LOCKED;

  /* Search for allocator */
  allocator = mm_find_allocator_by_name("RAM");

  if (allocator == NULL)
    return MFMI_UNAVAILABLE;

  /* Translate mfmi-tag to mm-tag*/
  switch (tag)
    {
    case MFMI_INFO_PHYS_MEM_TOTAL:
      mem_info.tags[0].tag_id = MM_TAG_INFO_PHYS_MEM_TOTAL_MEMORY;
      break;
    case MFMI_INFO_PHYS_MEM_FREE:
      mem_info.tags[0].tag_id = MM_TAG_INFO_PHYS_MEM_FREE_MEMORY;
      break;
    case MFMI_INFO_PHYS_MAX_FRAG_SIZE:
      mem_info.tags[0].tag_id = MM_TAG_INFO_PHYS_MEM_BIGGEST_FRAG;
      break;
    case MFMI_INFO_PHYS_MEM_INACTIVE: /* Deprecated */
      mem_info.tags[0].tag_id = MM_TAG_INFO_PHYS_MEM_INACTIVE_MEMORY;
      break;
    case MFMI_INFO_PHYS_MEM_WATERMARK:
      mem_info.tags[0].tag_id = MM_TAG_INFO_PHYS_MEM_FREE_WATERMARK;
      break;
    default:
      return MFMI_UNAVAILABLE;
    }

  /* Get data for tag */
  mem_info.count = 1;
  allocator->info(allocator,
		  (union SIGNAL*)mem_info_p,
		  (union SIGNAL **)&mem_info_p);

  /* Copy data */
  *info = (mm_size_t)mem_info.tags[0].tag_value;

  return MFMI_OK;
}

static MfmiStatus
get_next_domain(struct MfmiDomainCursor *dc,
		struct MfmiDomainInfo   *di)
{
  struct DomainCursor *cursor = (struct DomainCursor*)dc;
  struct domains *domains = &gmm_state.domains;
  MfmiStatus status = MFMI_END_OF_LIST;

  /* Check if data is locked by MM process */
  if (!gmm_state.data_accessable)
    return MFMI_DATA_LOCKED;

  /* Find domain */
  while (cursor->domain < gmm_state.max_domains)
    {
      if (domains->domain_list[cursor->domain] != NULL)
	{
	  /* Copy domain data */
	  di->domain = domains->domain_list[cursor->domain]->segnum;
	  strcpy(di->name, domains->domain_list[cursor->domain]->name);
	  cursor->domain++;
	  status = MFMI_OK;
	  break;
	}
      cursor->domain++;
    }

  return status;
}

static MfmiStatus
get_first_region(OSDOMAIN domain,
		 struct MfmiRegionCursor *rc,
		 struct MfmiRegionInfo   *ri)
{
  struct RegionCursor *cursor = (struct RegionCursor*)rc;

  /* Check if data is locked by MM process */
  if (!gmm_state.data_accessable)
    return MFMI_DATA_LOCKED;

  /* Search for domain */
  cursor->domain_desc = mm_find_domain_by_domid(&gmm_state.domains, domain);

  if (cursor->domain_desc == NULL)
    return MFMI_EDOMAIN;

  /* Init region cursor to first region in list */
  cursor->region_desc = (struct region*)
     core_rbtree_first_no_lock(&cursor->domain_desc->regions);

  /* Get data for first region */
  return get_next_region(rc, ri);
}

static MfmiStatus
get_next_region(struct MfmiRegionCursor *rc,
		struct MfmiRegionInfo   *ri)
{
  struct RegionCursor *cursor = (struct RegionCursor*)rc;
  struct domain *domain_desc = cursor->domain_desc;
  struct region *region_desc = cursor->region_desc;
  struct log_object *lobj;
  struct partition *partition;

  /* Check if data is locked by MM process */
  if (!gmm_state.data_accessable)
    return MFMI_DATA_LOCKED;

  /* No more regions in this domain */
  if (cursor->region_desc == NULL)
    return MFMI_END_OF_LIST;

  lobj = region_desc->log_object;
  partition = mm_log_object_partition(lobj, region_desc->domain);
  
  /* Copy region data */
  ri->base        = lobj->base;
  ri->size        = lobj->size;;
  ri->permissions = region_desc->permissions;;
  ri->cache_mode  = region_desc->cache_mode;

  strncpy(ri->name,     region_desc->name,            sizeof ri->name);
  strncpy(ri->phys_mem, mm_get_phys_mem(region_desc), sizeof ri->phys_mem);
  strncpy(ri->log_mem,  partition->class->name,       sizeof ri->log_mem);

  ri->options = 0;
  if (ri->name[0] != '\0')
    ri->options |= MM_CREATE_REGION_SPECIFY_NAME;
  if (NULL != region_desc->phys_object &&
      PHYS_OBJ_CLIENT_PAGES == region_desc->phys_object->data_representation)
    ri->options |= MM_CREATE_REGION_SPECIFY_PHYS;

  /* Advance region cursor to next region */
  cursor->region_desc = (struct region*)
     core_rbtree_next_no_lock(&domain_desc->regions, cursor->region_desc);
  return MFMI_OK;
}

static MfmiStatus
get_address_info(OSDOMAIN domain,
		 OSADDRESS log_address,
		 struct MfmiAddressInfo *phys)
{
  struct domain      *domain_desc;
  struct region      *region_desc;
  struct phys_object *phys_object;
  mm_bool_t           translate_success;

  phys_addr_t phys_address;
  phys_addr_t fragment_base;
  phys_size_t fragment_size;
  phys_size_t offset;
  
  /* Check if data is locked by MM process */
  if (!gmm_state.data_accessable)
    return MFMI_DATA_LOCKED;

  /* Search for domain */
  domain_desc = mm_find_domain_by_domid(&gmm_state.domains, domain);

  if (domain_desc == NULL)
    return MFMI_EDOMAIN;

  /* Search for region based on logical address */
  region_desc = mm_find_region_by_addr(
     &gmm_state.domains, domain, log_address, FALSE);
  if (region_desc == NULL)
    return MFMI_UNAVAILABLE;

  offset = log_address - region_desc->log_object->base;
  phys_object = region_desc->phys_object;
  
  if (phys_object == NULL)
    return MFMI_UNAVAILABLE;

  /* Translate logical address to physical */
  translate_success = mm_extract_phys_addr(phys_object,
                                           offset,
                                           &phys_address,
                                           &fragment_base,
                                           &fragment_size);
  if (!translate_success)
    return MFMI_UNAVAILABLE;

  /* Copy data */
  phys->address = phys_address;
  phys->fragment_base = fragment_base;
  phys->fragment_size = fragment_size;

  return MFMI_OK;
}

/*
*****************************************************************************
** PUBLIC FUNCTIONS
*****************************************************************************
*/


MfmiStatus
ose_mfmi_get_domain_info(OSDOMAIN domain,
			 struct MfmiDomainInfo *di)
{
  U32 lock;
  MfmiStatus status;

  LOCK_SAVE(lock);
  status = get_domain_info(domain, di);
  LOCK_RESTORE(lock);

  return status;
}


MfmiStatus
ose_mfmi_get_region_info(OSDOMAIN domain,
			 OSADDRESS address,
			 struct MfmiRegionInfo *ri)
{
  U32 lock;
  MfmiStatus status;

  LOCK_SAVE(lock);
  status = get_region_info(domain, address, ri);
  LOCK_RESTORE(lock);

  return status;
}

MfmiStatus
ose_mfmi_get_memory_info(enum MfmiInfoTag tag,
			 mm_size_t *info)
{
  U32 lock;
  MfmiStatus status;

  LOCK_SAVE(lock);
  status = get_memory_info(tag, info);
  LOCK_RESTORE(lock);

  return status;
}


MfmiStatus
ose_mfmi_get_first_domain(struct MfmiDomainCursor *dc,
			  struct MfmiDomainInfo   *di)
{
  struct DomainCursor *cursor = (struct DomainCursor*)dc;
  MfmiStatus status;
  U32 lock;

  LOCK_SAVE(lock);
  cursor->domain = 0;
  status = get_next_domain(dc, di);
  LOCK_RESTORE(lock);

  return status;
}

MfmiStatus
ose_mfmi_get_next_domain(struct MfmiDomainCursor *dc,
			 struct MfmiDomainInfo   *di)
{
  MfmiStatus status;
  U32 lock;

  LOCK_SAVE(lock);
  status = get_next_domain(dc, di);
  LOCK_RESTORE(lock);

  return status;
}


MfmiStatus
ose_mfmi_get_first_region(OSDOMAIN domain,
			  struct MfmiRegionCursor *rc,
			  struct MfmiRegionInfo   *ri)
{
  MfmiStatus status;
  U32 lock;

  LOCK_SAVE(lock);
  status = get_first_region(domain, rc, ri);
  LOCK_RESTORE(lock);

  return status;
}

MfmiStatus
ose_mfmi_get_next_region(struct MfmiRegionCursor *rc,
			 struct MfmiRegionInfo   *ri)
{
  MfmiStatus status;
  U32 lock;

  LOCK_SAVE(lock);
  status = get_next_region(rc, ri);
  LOCK_RESTORE(lock);

  return status;
}


MfmiStatus
ose_mfmi_get_first_phys_mem_type(struct MfmiPhysMemTypeCursor *pmtc,
                                 struct MfmiPhysMemTypeInfo   *pmti)
{
  MfmiStatus status;
  U32 lock;

  LOCK_SAVE(lock);
  status = get_first_phys_mem_type(pmtc, pmti);
  LOCK_RESTORE(lock);

  return status;
}

MfmiStatus
ose_mfmi_get_next_phys_mem_type(struct MfmiPhysMemTypeCursor *pmtc,
                                struct MfmiPhysMemTypeInfo   *pmti)
{
  MfmiStatus status;
  U32 lock;

  LOCK_SAVE(lock);
  status = get_next_phys_mem_type(pmtc, pmti);
  LOCK_RESTORE(lock);

  return status;
}

MfmiStatus
ose_mfmi_get_address_info(OSDOMAIN domain,
			  OSADDRESS log_address,
			  struct MfmiAddressInfo *phys_info)
{
  MfmiStatus status;
  U32 lock;

  LOCK_SAVE(lock);
  status = get_address_info(domain, log_address, phys_info);
  LOCK_RESTORE(lock);

  return status;
}

MfmiStatus
ose_mfmi_get_cache_type(uint32_t cache_mode,
                        char *cache_type,
                        char *memory_type)
{
   hal_status_t status;
   Msr msr;

   LOCK_SAVE(msr);
   status = hal_get_cache_type((U32)cache_mode, cache_type, memory_type);
   LOCK_RESTORE(msr);

   return status == HAL_SUCCESS ? MFMI_OK : MFMI_END_OF_LIST;
}

/* ?????????

-What about assert() in called functions???
-Data lock handling? Required if accessed in run mode
-Interrupt locking required? (not in hfmi)

????????? */
