/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2007 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

/**
 * @public
 *
 * @file mm_access.c
 *
 * @brief Implementation of MM safe copy and forced write.
 */

#include "assert.h"
#include "cache.h"
#include "cpu_intlock.h"
#include "efs.h"
#include "mm_access.h"
#include "mm_err.h"
#include "osarch.h"
#include "string.h"
#include "ose.h"
#include "mm.h"

#include "hal.h"
#include "partition.h"

#include "common.h"
#include "gmm.h"
#include "mm_access_internal.h"
#include "mm_trap_internal.h"
#include "mmtypes.h"
#include "cpu_access.h"
#include "mm_log.h"
#include "mm_private.h"
#include "hal.h"

#ifndef softOSE
#include "cpu_hal_arch.h"
#endif

#define VERBOZE(s) log(CATEGORY_MM_ACCESS, s)
#define PRINT_MAS(s) log(CATEGORY_MAS, s)

union SIGNAL
{
   struct mm_forced_write mm_forced_write;
};

mm_bool_t
mm_mas_mode_active(void);

int
mm_access_safecpy_internal(OSDOMAIN dst_domain,
                           void *dst_addr,
                           OSDOMAIN src_domain,
                           const void *src_addr,
                           size_t size,
                           int freeze_mode);

static log_addr_t freeze_mode_la;
static log_addr_t run_mode_la;

/* Forced write uses copy-back and a flush instead of inhibited mapping.
 * The main reason is that we currently lack support for some L2 cache
 * contollers and are unable to invalidate all cache levels as required.
 * With copy-back mapping, L1 maintenance operations are sufficient as
 * long as the L2 cache is unified and non-aliasing.  
 */
static U32 mem_attr = (MM_SPECULATIVE_ACCESS | MM_CACHE_COPY_BACK);

extern mm_status_t mm_l2p_by_domain(OSDOMAIN    domain,
                                 log_addr_t  la,
                                 phys_addr_t *pa);

extern size_t
mm_validate_buffer_by_domain(log_addr_t bbase,	          /* buffer base */
                          size_t bsize,	          /* buffer size */
                          OSDOMAIN domain,               /* validation domain */
                          unsigned int bperm,  	  /* buffer permissions */
                          unsigned int options,	  /* validation options */
                          unsigned int *result,	  /* option results */
                          OSADDRESS supervisor_mode,    /* supervisior mode*/
                          U32 hal_options);


/* Safe memcpy() provided by EXEC. */
extern int
zzguarded_memcpy(void *dest,
                 const char *source,
                 size_t size,
                 int freeze_mode_flag);


static enum MmStatus
sanity_check_source(log_addr_t addr, log_size_t size)
{
   U32 result;
   U32 perm = MM_PERM_READ;
   
#ifdef OSMIPS
   if (mm_log_inside(addr, KSEG0_BASE, KSEG0_SIZE) || 
       mm_log_inside(addr, KSEG1_BASE, KSEG1_SIZE))
   {
      /* Access is ok */
      return MM_SUCCESS;
   }
   else
#endif
   {
      if (mm_space_is_mas(addr, 1))
      {
         return MM_EMAS_NOT_SUPPORTED;
      }

      result = mm_validate_buffer_by_domain(addr, size, MM_KERNEL_DOMAIN, perm, 0, NULL, 1, 0);

      if (result < size)
      {
         return MM_ENO_READ_PERMISSION;
      }

   }

   return MM_SUCCESS;
}

/*
 *
 */
static enum MmStatus
sanity_check_addr_and_domain(log_addr_t addr, OSDOMAIN domid)
{
   struct domain *domain;
   struct region *region;


#ifdef OSMIPS
   
   if (domid == 0) {
      if ((addr & KSEG_MASK) == KSEG0_BASE) {
         return MM_SUCCESS;
      }

      if ((addr & KSEG_MASK) == KSEG1_BASE) { 
         return MM_SUCCESS;
      }
   }
   
#endif
   
   domain = mm_find_domain_by_domid(&gmm_state.domains, domid);
   if (domain == NULL)
   {
      return MM_ENON_EXISTING_DOMAIN;
   }

   region = mm_find_region_by_addr(&gmm_state.domains, domain->segnum, addr, 0);
   if (region == NULL)
   {
      return MM_ENON_EXISTING_REGION;
   }

   if (!region_is_exclusive(region))
   {
      return MM_EMAS_NOT_SUPPORTED;
   }
   
   return MM_SUCCESS;
}


mm_bool_t
mm_mas_mode_active(void)
{
   mm_bool_t rc =  FALSE;
   assert((int)gmm_state.mas_cnt >= 0);

   if(gmm_state.mas_cnt > 0)
   {
      rc = TRUE;
   }

   PRINT_MAS(("gmm:  mm_mas_mode_active returns (%d)\n", rc));

   return rc;
}

/* See mm_access.h */
int
mm_access_safecpy_internal(OSDOMAIN dst_domain,
                 void *dst_addr,
                 OSDOMAIN src_domain,
                 const void *src_addr,
                 size_t size,
                 int freeze_mode)
{
   int rc;
   OSBOOLEAN set_seg_status;
   OSBOOLEAN need_set_seg =
      (dst_domain != src_domain) && mm_mas_mode_active();

   VERBOZE(("mm_access_safecpy(dst_domain=%#lx, dst_addr=%#lx, "\
            "src_domain=%#lx, src_addr=%#lx, size=%#lx, freeze_mode=%#lx) "\
            "need_set_seg=%d \n", dst_domain, dst_addr, src_domain, src_addr,
            size, freeze_mode, need_set_seg));

   /*
    * In MAS case we need to operate in correct page table.
    * Note that if a copy between two non System domains with MAS
    * an extra memcpy is needed since both pages tables cannot
    * be set simultaneously. I.e. an tmp buffer is needed to first copy
    * to from src_domain and then secocondly to dst_domain.
    */
   if (need_set_seg)
   {
      /* FIXME. security should check that caller is supervisor
       * or we will allow to read other domains secure data.
       */
      if (dst_domain == MM_KERNEL_DOMAIN)
      {
         set_seg_status = set_segment(0, src_domain);
      }
      else if (src_domain == MM_KERNEL_DOMAIN)
      {
         set_seg_status = set_segment(0, dst_domain);
         if (set_seg_status != 0)
         {
            rc = MM_EDOMAIN_NOT_MAS_ACTIVE;
            goto exit_error;
         }
      }
      else
      {
         /* One of the domains must be the system domain.
          * Copy between two non system domains not allowed
          * for security reasons.
          */
         rc = MM_EILLEGAL_OPERATION;
         goto exit_error;
      }
   }

   if (freeze_mode)
   {
      Msr sr;

      LOCK_SAVE(sr);
      rc = zzguarded_memcpy(dst_addr, src_addr, size, freeze_mode);
      LOCK_RESTORE(sr);
   }
   else
   {
      rc = zzguarded_memcpy(dst_addr, src_addr, size, freeze_mode);
   }


   if (need_set_seg)
   {
      set_seg_status = set_segment(1,0);   /* Go back. */
      assert(set_seg_status == 0);
   }

   return rc;


   /* Use only one ramlog_printf for all faults to keep footprint down. */
  exit_error:
   ramlog_printf("mm: mm_access_safecpy fail with status:%#x see "
                 "enum MmStatus in mm.sig. \n", rc);
   return rc;
}



/* This function should be called as soon as HAL can perform
 * hal_map_area. After this call, HAL must have allocated all required
 * resources from GMM in order to perform the mapping.
 *
 * If the partition is NULL, this is taken as an indication that alias
 * mappings are not necesseray, and only normal mm_access_safecpy
 * is performed.
 */
void
mm_access_init(struct partition *partition)
{
   if (hal_get_info()->has_mmu)
   {
      U32 freeze_mode_access;
      U32 run_mode_access;
      U32 options = HAL_GLOBAL_MAPPING;


      /* Only check partition in MMU case otherwise to avoid
       * HAL log mem for No MMU.
       */
      assert(partition != NULL);

      /* Allocate logical space. */
      if ((TRUE !=
           mm_alloc_log_mem2(partition,
                             gmm_state.hal_info->page_sizes[0],
                             &freeze_mode_la)) ||
          (TRUE !=
           mm_alloc_log_mem2(partition,
                             gmm_state.hal_info->page_sizes[0],
                             &run_mode_la)))
      {
         MM_FATAL_ERROR2(("out of memory!!.\n"),
                         OSE_MM_ECONFIGURATION_ERROR,
                         (OSERRCODE)("Out of memory!"));
      }

      if ((HAL_SUCCESS !=
           hal_actual_permissions(&freeze_mode_access,
                                  SU_NA_USR_NA,
                                  freeze_mode_la, 0)) ||
          (HAL_SUCCESS !=
           hal_actual_permissions(&run_mode_access,
                                  SU_NA_USR_NA,
                                  run_mode_la, 0)))
      {
         INTERNAL_ERROR();
      }

#ifndef softOSE
      if (hal_arch_cpu_descriptor) {
         U32 mm_attributes = hal_arch_cpu_descriptor->mm_attributes;
         if (mm_attributes & ATTRIBUTE_HW_CACHE_COHERENCY) {
            /* With HW cache coherency protocol cache inhibit in the alias
             * mapping should be avoided, otherwise the coherency protocol
             * will be corrupted.
             */
            mem_attr = (MM_SPECULATIVE_ACCESS | MM_CACHE_COPY_BACK);
            VERBOZE(("mm_access_init: HW_CACHE_COEHERENCY detected "
                     "mm_attributes:  %#lx, updated mem_attr: %#lx\n",
                     mm_attributes, mem_attr));
         }
      }
#endif

     /* Make sure page tables are allocated in HAL. */
      if (hal_map_page(gmm_state.system_domain->hal_domain,
                       freeze_mode_la, 0, gmm_state.hal_info->page_sizes[0],
                       options, freeze_mode_access, mem_attr) != HAL_SUCCESS ||
          hal_map_page(gmm_state.system_domain->hal_domain,
                       run_mode_la, 0, gmm_state.hal_info->page_sizes[0],
                       options, run_mode_access, mem_attr) != HAL_SUCCESS)
      {
         INTERNAL_ERROR();
      }
   }
}


/*
 * In case of MIPS architecture check if range is inside KSEG0/KSEG1.
 */
static mm_bool_t
check_mips_fixed_mappings(U32 base, size_t size)
{
   mm_bool_t rc = FALSE;

#ifdef OSMIPS
   assert((base + size - 1) > 0);
   if ((mm_log_inside(base, KSEG0_BASE, KSEG0_SIZE)) ||
       (mm_log_inside(base, KSEG1_BASE, KSEG1_SIZE)) ||
       (mm_log_inside((base + size - 1), KSEG0_BASE, KSEG0_SIZE)) ||
       (mm_log_inside((base + size - 1), KSEG1_BASE, KSEG1_SIZE)))
   {
      rc = TRUE;
   }
#else
   (void)base;
   (void)size;
#endif
   
   return rc;
}

#ifdef ASSERT
static void
verifyCopy(const unsigned char *s, const unsigned char *d,
           size_t size, int line)
{
   size_t i;

   for (i = 0; i < size; i++) {
      if (s[i] != d[i]) {
         ramlog_printf("%s:%d. forced_write failed\n", __FILE__, line);
         ramlog_printf("0x%08lx[%3ld] != 0x%08lx[%3ld] (0x%02x, 0x%02x)\n",
                       s, i, d, i, s[i], d[i]);
         assert(0);
      }
   }
}
#endif

static int
do_forced_write(OSDOMAIN dst_domain,
                void *dst,
                const void *src,
                size_t size,
                log_addr_t la,
                int freeze_mode)
{
   mm_bool_t mips_fixed_mapping;

   assert(dst != NULL);
   assert(src != NULL);

   /* For MIPS architecture:
    * In case the write is located in the fixed KSEG0/KSEG1 mappings
    * write is already allowed from supervisormode and hence no other
    * MMU mapping is needed. Furthermore is it not only unnecessary but
    * also incorrect to map towards KSEG0 risking undocumented behaviour.
    * (Malta handles it OK but broadcom gets corrupted.)
    */
   mips_fixed_mapping = check_mips_fixed_mappings((U32)dst, size);

   VERBOZE(("do_forced_write(dst_domain=%#lx, " \
            "dst=%#lx, src=%#lx, size=%#lx, "\
            "la=%#lx freeze_mode=%#lx) mips_fixed_mapping: %d\n",
            dst_domain, dst, src, size, la ,freeze_mode, mips_fixed_mapping));

#ifndef softOSE
   {
     int func_rc;
     if (hal_get_info()->has_mmu &&
         hal_get_info()->enabled_mmu &&
         !mips_fixed_mapping)
     {
       phys_addr_t pa;
       U32 act_access;
       U32 offset;
       hal_status_t hal_rc;
       mm_status_t  rc;
       log_size_t   t_size;
       void * t_dst = (void*)dst;
       const void * t_src = (const void*)src;
          
       hal_rc = hal_actual_permissions(&act_access, SU_RW_USR_NA, la, 0);
       if (HAL_SUCCESS != hal_rc)
       {
         ramlog_printf("mm: do_forced_write, hal_actual_permissions failed "\
                       "with status:%#x see enum hal_status in hal.h \n",
                       hal_rc);
         func_rc = MM_EOUT_OF_PHYS_MEM;
         goto exit_error;
       }

       t_size = size;
       
       do
       {

          offset = (U32)t_dst - ALIGN_DOWN((U32) t_dst,
                                         gmm_state.hal_info->page_sizes[0]);
          size   = t_size < gmm_state.hal_info->page_sizes[0]
                   ? t_size
                   : gmm_state.hal_info->page_sizes[0];

          if((size + offset) > gmm_state.hal_info->page_sizes[0])
          {
             size = gmm_state.hal_info->page_sizes[0] - offset;
          }
          rc = mm_l2p_by_domain(dst_domain, (log_addr_t)t_dst, &pa);
          if (rc == MM_SUCCESS)
          {
             U32 options = HAL_GLOBAL_MAPPING | HAL_CHANGE;
             /*
              * Map target physical memory into alias mapping virtual page.
              * IMPORTANT: This mapping is cache inhibited except for
              * BCM91250/91280.
              */
             hal_rc = hal_map_page(gmm_state.system_domain->hal_domain, la, pa,
                                   gmm_state.hal_info->page_sizes[0],
                                   options, act_access, mem_attr);

             if (HAL_SUCCESS == hal_rc)
             {
                /*
                 * Invalidate target range in caches.
                 * REQUIREMENT: both operations must be treated as load by
                 * the MMU since the target range may be write protected.
                 *
                 * Note that a cache_sync() is not suffiecent since the
                 * write is performed using an alias mapping and therefore
                 * the range dst->dst+size must also be invalidated in dcache.
                 *
                 */
                rc = mm_validate_buffer_by_domain((log_addr_t)t_dst, size, dst_domain, MM_PERM_READ, 0, NULL, 1, 0);
                if(rc)
                {
#ifdef OSPP		   
		   /*
		     On PPC we don't need to flush/invalidate the data
		     cache as the data cache is always physical.

		     To use cache_flush_invalidate_d here for PPC
		     might seem safe and harmless. However, on
		     architectures that overrides the cache method in
		     the cpu-hal, cache_flush_invalidate_d will
		     execute as a single flush and a single
		     invalidate. A flush (dcbst) is treated as a
		     load. Invalidate (dcbi) is treated as a store. A
		     store in write protected memory will fail on some
		     architectures.

		     The native cache_flush_invalidate_d
		     implementation use the atomic dcbf to both flush
		     and invalidate. Dcbf is treated as a load by the
		     CPU which should be OK to use as long as it is
		     not overridden.
		   */
#else
                   cache_flush_invalidate_d((OSADDRESS) t_dst, size);
#endif                   
                }

                /* Perform memcpy().*/
                VERBOZE(("do_forced_write: Do zzguarded_memcpy"
                         "(la + offset: %#010lx, src: %#010lx, size: %#010lx, "\
                         "freeze_mode: %d) act_access: %#lx "\
                         "mem_attr: %#lx\n",
                         (la + offset), t_src, size, freeze_mode,
                         act_access, mem_attr));

                func_rc = zzguarded_memcpy((void *)(la + offset), t_src, size,
                                           freeze_mode);

                cache_flush_invalidate_d((OSADDRESS) (la + offset), size);
                
                if(rc)
                {
                   cache_invalidate_i((OSADDRESS) t_dst, size);
                }
                
                VERBOZE(("do_forced_write: Did zzguarded_memcpy"
                         "func_rc:%d\n",  func_rc));

#ifndef softOSE
                if (hal_arch_cpu_descriptor) {
                   U32 mm_attributes = hal_arch_cpu_descriptor->mm_attributes;
                   if (mm_attributes & ATTRIBUTE_I_CACHE_VIRTUAL_TAGGED) {
                      VERBOZE(("do_forced_write: I_CACHE_VIRTUAL_TAGGED detected.\n"));
                      /* An alternative to flush all Icache would be to flush
                       * only the range (dst->dst+size) in the correct domain
                       * however that will not work for MIPS since ASID 0 is used for
                       * all supervisor domains.
                       */
                      cache_invalidate_i_complete(0);
                   }
                }
#endif

#ifdef ASSERT
                if(rc)
                {
                   verifyCopy(t_dst, t_src, size, __LINE__);
                }
                verifyCopy((unsigned char*)(la + offset), t_src, size, __LINE__);
#endif

                /* Unmap target physical memory. */
                hal_rc = hal_map_page(gmm_state.system_domain->hal_domain, la, 0,
                                      gmm_state.hal_info->page_sizes[0],
                                      options, act_access, mem_attr);
 
                VERBOZE(("do_forced_write: Did hal_map_page"
                         "hal_rc:%d\n",  hal_rc));

                if (func_rc != 0)
                {
                   func_rc = MM_EWRITE_FAILED; /* zzguarded_memcpy failed. */
                   goto exit_error;
                }
             }
             else /* hal_map_page failed. */
             {
                ramlog_printf("mm: do_forced_write, hal_map_page failed with "\
                              "status:%#x see enum hal_status in hal.h \n",
                              hal_rc);
                func_rc = MM_EOUT_OF_PHYS_MEM;
                goto exit_error;
             }
          }
          else /* mm_l2p_by_domain failed. The region might be paged, so ignore this. */
          {
             return MM_ENO_PHYSICAL_MAPPING;
          }
          
          t_size -= size;
          t_dst   =(void *) ((log_addr_t)t_dst + size);
          t_src   =(void *) ((log_addr_t)t_src + size);
          
       }while(t_size > 0);
       
       return 0;
       
     }      /* if (hal_get_info()->has_mmu && hal_get_info()->enabled_mmu) */
     else
     {
       if(0 != zzguarded_memcpy(dst, src, size, freeze_mode))
       {
          func_rc = MM_EWRITE_FAILED;   /* zzguarded_memcpy failed. */
          goto exit_error;
       }
       cache_sync((OSADDRESS) dst, size);
       return 0; /* Success (MMU disabled). */
     }

   exit_error:
     ramlog_printf("mm: do_forced_write fail with status:%#x "\
                   "see enum MmStatus in mm.sig. \n", func_rc);
     return func_rc;
   }
#else
   memcpy(dst, src, size);
   return 0;
#endif
}



/*
 * Called from GMM when receiving MM_FORCED_WRITE_REQUEST.
 * I.e. exceutes in gmm context  = su.
 */
union SIGNAL *
mm_access_forced_write_runmode(union SIGNAL *request)
{
   int rc;
   struct mm_forced_write *req;

   assert(request != NULL);
   req = (struct mm_forced_write *)request;
   req->sig_no = MM_FORCED_WRITE_REPLY;

   VERBOZE(("mm_access_forced_write_runmode(req->dst_domain=%#lx, "\
            "req->dst=%#lx, req->src=%#lx, req->size=%#lx) *req->src=%#lx\n",
            req->dst_domain, req->dst, req->src, req->size,
            *(U32*)(U32)req->src));


   /* As long as MAS write's not supported deny overlapping address */
   if (mm_log_inside((log_addr_t)req->dst, (log_addr_t)req->src, req->size))
   {
      rc = MM_EILLEGAL_OPERATION;
      goto exit_error;
   }

   /* Validate source region accessibility */
   rc = sanity_check_source((log_addr_t)req->src, req->size);
   if (rc != MM_SUCCESS)
   {
      goto exit_error;
   }

   /* Validate destination region */
   rc = sanity_check_addr_and_domain((log_addr_t)req->dst, req->dst_domain);
   if (rc != MM_SUCCESS)
   {
      goto exit_error;
   }

   rc = do_forced_write(req->dst_domain,
                        req->dst,
                        req->src,
                        req->size,
                        run_mode_la,
                        0);
   req->status = rc;
   return request;

  exit_error:
   /* Use only one ramlog_printf for all faults to keep footprint down. */
   ramlog_printf("mm: mm_access_forced_write_runmode fail with status:%#x "\
                 "see enum MmStatus in mm.sig. \n", rc);

   req->status = rc;
   return request;
}


int
mm_access_forced_write(OSDOMAIN dst_domain,
                       void *dst,
                       const void *src,
                       size_t size,
                       int freeze_mode)
{
   mm_status_t rc;

   assert(dst != NULL);
   assert(src != NULL);

   VERBOZE(("mm_access_forced_write(dst_domain=%#lx, dst=%#lx, "\
            "src=%#lx, size=%#lx, freeze_mode=%#lx) *src=%#lx\n",
            dst_domain, dst, src, size, freeze_mode, *(U32*)(U32)src));

   /* As long as MAS write's not supported deny overlapping address */
   if (mm_log_inside((log_addr_t)dst, (log_addr_t)src, size))
   {
      rc = MM_EILLEGAL_OPERATION;
      goto exit_error;
   }

   /* Validate source region accessibility */
   rc = sanity_check_source((log_addr_t)src, size);
   if (rc != MM_SUCCESS)
   {
      goto exit_error;
   }

   rc = sanity_check_addr_and_domain((log_addr_t)dst, dst_domain);
   if (rc != MM_SUCCESS)
   {
      goto exit_error;
   }

   if (freeze_mode)
   {
      Msr sr;

      LOCK_SAVE(sr);
      rc = do_forced_write(dst_domain, dst, src, size, freeze_mode_la, 1);
      LOCK_RESTORE(sr);
   }
   else
   {
      PROCESS gmm_;
      union SIGNAL *request;
      static const SIGSELECT sigsel[] = {1, MM_FORCED_WRITE_REPLY};

      /* FIXME. Check that kernel is started. */
      gmm_ = ose_hunt_w_tmo("ose_memory_mgr", 0, 0);
      if (0 == gmm_)
      {
         INTERNAL_ERROR();
      }
      request = alloc(sizeof(struct mm_forced_write), MM_FORCED_WRITE_REQUEST);
      request->mm_forced_write.dst_domain = dst_domain;
      request->mm_forced_write.dst = dst;
      request->mm_forced_write.src = src;
      request->mm_forced_write.size = size;
      send(&request, gmm_);
      request = receive(sigsel);
      rc = request->mm_forced_write.status;
      free_buf(&request);
   }

   if (rc != MM_SUCCESS)
   {
      goto exit_error;
   }
   VERBOZE(("mm_access_forced_write exit. dst=%#lx *dst=%#lx "
            "src=%#lx *src=%#lx\n",
            dst, *(U32*)(U32)dst, src, *(U32*)(U32)src));
   return rc;

  exit_error:
   ramlog_printf("mm: mm_access_forced_write fail with status:%#x "\
                 "see enum MmStatus in mm.sig. \n", rc);
   return rc;
}
