/* COPYRIGHT-ENEA-SRC-R1 *
 **************************************************************************
 * Copyright (C) 2005 by Enea Embedded Technology AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

#include "gmm.h"
#include "configure.h"
#include "compiler_macros.h"
#include "cpu_hal_arch.h"
#include "common.h"
#include "errno.h"
#include "string.h"
#include "sysparam.h"
#include "assert.h"
#include "mm_log.h"

#define CACHE_PREFIX                "krn/cache"
#define LOG_PREFIX                  "krn/mm/log_level"
#define DCACHE_ATTRIBUTE_KEY        "d-cache"
#define ICACHE_ATTRIBUTE_KEY        "i-cache"

static const struct enumeration boolean_encoding[] =
{
   {"true", 1},
   {"false", 0},
   {"enabled", 1},
   {"disabled", 0},
   {"1", 1},
   {"0", 0},
   { NULL, 0}
};

static void dummy_select(U32 dm)
{
   (void)dm;
}

static struct mm_handlers handlers =
{
   &dummy_select
};

/*
 * weak reference to functions in gmm.c, to resolve in case
 * of gmm exist.
 */

WEAK_REF(void
mm_install_exception_handlers_internal(void));

WEAK_REF(struct mm_handlers *
mm_open_exception_area_internal(const struct cpu_descriptor *cpu_descriptor,
                                char *vector_base,
                                size_t vector_size));

/* weak reference to function in mm_access.c.
 * if not present, we don't have an MMU.
 */

WEAK_REF(int
mm_access_safecpy_internal(OSDOMAIN dst_domain,
                 void *dst_addr,
                 OSDOMAIN src_domain,
                 const void *src_addr,
                 size_t size,
                  int freeze_mode));

/*
 * Trampoline to mm_access_safecpy_internal if GMM is present. 
 */
int
mm_access_safecpy(OSDOMAIN dst_domain,
                  void *dst_addr,
                  OSDOMAIN src_domain,
                  const void *src_addr,
                  size_t size,
                  int freeze_mode);

/*
 * global variable used in many MM parts.
 */
struct gmm_vars  gmm_state;
struct gmm_stats gmm_stats;

/*
 * Log configuring moved from configure.c to here.
 */

static U32 log_mask = 0;

U32
mm_get_log_mask(void)
{
   return log_mask;
}

U32
mm_set_log_mask(U32 mask)
{
   U32 old = log_mask;

   log_mask = mask;
   return old;
}

const char *
mm_get_log_mask_name(U32 bit)
{
   switch (bit)
   {
      case CATEGORY_START:		return "start";
      case CATEGORY_HAL:		return "hal";
      case CATEGORY_MALLOC:		return "malloc";
      case CATEGORY_PAGE:		return "page";
      case CATEGORY_SIGNAL:		return "signal";
      case CATEGORY_MISC:		return "misc";
      case CATEGORY_MAS:		return "mas";
      case CATEGORY_PARTITION:		return "partition";
      case CATEGORY_DOMAIN:		return "domain";
      case CATEGORY_HAL_DETAIL:		return "hal_detail";
      case CATEGORY_MM_ACCESS:		return "mm_access";
      case CATEGORY_START_DETAIL:	return "start_detail";
      case CATEGORY_MM_TRAP:		return "mm_trap";
      case CATEGORY_HAL_GLOBAL:		return "hal_global";
      case CATEGORY_MMU_STAT:		return "mmu_stat";         
      case CATEGORY_FAILED_REQUEST:	return "failed request";         
   }
   return NULL;
}

void
mm_configure_log_level(void)
{
   (void)sysparam_readattr_long(LOG_PREFIX, "level", (long *) &log_mask);
}

/* global function mm_configure_cache_mode
 * moved here from configure.c
 */

void
mm_configure_cache_mode(void)
{
   int rc;
   const char *attribute;

   /* Default values. */
   gmm_state.d_cache = 0;
   gmm_state.i_cache = 0;

   attribute = DCACHE_ATTRIBUTE_KEY;
   rc = sysparam_readattr_enum(CACHE_PREFIX,
                               attribute,
                               boolean_encoding,
                               (long *)(&gmm_state.d_cache));
   if (EOK == rc)
   {
      attribute = ICACHE_ATTRIBUTE_KEY;
      rc = sysparam_readattr_enum(CACHE_PREFIX,
                                  attribute,
                                  boolean_encoding,
                                  (long *)(&gmm_state.i_cache));
   }
   if (rc != EOK)
   {
      const char *etype;

      etype = ENOENT == rc ? "missing": "invalid";
      MM_FATAL_ERROR2(("%s attribute '%s' in 'krn/cache'.\n", etype,
                       attribute),
                      OSE_MM_ECONFIGURATION_ERROR,
                      (OSERRCODE) "/krn/cache/");
   }
}

/*
 * function mm_access_safecpy have trampolin body here, that resolves to
 * mm_access.c only if gmm is present.
 */

int
mm_access_safecpy(OSDOMAIN dst_domain,
                 void *dst_addr,
                 OSDOMAIN src_domain,
                 const void *src_addr,
                 size_t size,
                 int freeze_mode)
{
   if(mm_access_safecpy_internal != 0)
   {
      return mm_access_safecpy_internal(dst_domain,dst_addr,src_domain,src_addr,size,freeze_mode);
   }
   else
   {
      memcpy(dst_addr,src_addr,size);
      return True;
   }
}

/*
 * First function to be called in MM.
 * If gmm.o is not present, perform CPU HAL init as gmm.c would do.
 */

struct mm_handlers *
mm_open_exception_area(const struct cpu_descriptor *cpu_descriptor,
                       char *vector_base,
                       size_t vector_size)
{
   if(mm_open_exception_area_internal != 0)
   {
      return mm_open_exception_area_internal(cpu_descriptor,
                                             vector_base, vector_size);
   }
   else
   {
      mm_configure_cache_mode();

      hal_init_1(cpu_descriptor,
                 (log_addr_t)vector_base,
                 vector_size,
                 gmm_state.d_cache,
                 gmm_state.i_cache);
      
      hal_init_2();
   }
   return &handlers;
}

/*
 * If gmm.o is present, perform exception install.
 */

void
mm_install_exception_handlers(void)
{
   if(mm_install_exception_handlers_internal != 0)
   {
      mm_install_exception_handlers_internal();
   }
}
   

