/* COPYRIGHT-ENEA-SRC-R1 *
 **************************************************************************
 * Copyright (C) 2007 by Enea Embedded Technology AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

#include "ose.h"
#include "bios.h"
#include "mm_trap.h"
#include "mm.h"
#include "mm.sig"
#include "mm_err.h"
#include "mm_log.h"
#include "hal.h"
#include "pose.h"
#include "partition.h"
#include "gmm.h"
#include "demand_pager.h"
#include "mm_util_internal.h"
#include "mm_trap_internal.h"
#include "mm_private.h"
#include "compiler_macros.h"
#include "common.h"
#include "cpu_access.h"
#include "mm_access.h"


extern void mm_init_bios_interface(void);
extern OSBOOLEAN ose_started(void);

/* 
** Don't use the current_process() in crt but call pose directly.
** A little faster, but the main reason is that epm trace exceptions
** may require L2 D/I cache sync when setting/restoring breakpoints.
** And the L2 cache sync may require l2p address translation services
** provided by this file, and l2p should not get stuck in a breakpoint
** on current_process in the crt of the monolith. 
*/
extern PROCESS zzkrn_current_process(void);
#define current_process() zzkrn_current_process()

#define VERBOZE(s) log(CATEGORY_MM_TRAP, s)



/*****************************************************************************
** Validate the access of a buffer without accessing it.
** See mm.h for more information.
**
** This function has been superseeded by its equivalent in mman.
** ose_mm_validate_buffer now traps to mman_validate_buffer which uses
** this function in part. mm_validate_buffer is however still exported
** through bios for backwards compatibility. 
*/

size_t
mm_validate_buffer_by_domain(log_addr_t bbase,	          /* buffer base */
                             size_t bsize,	          /* buffer size */
                             OSDOMAIN domain,             /* validation domain */
                             unsigned int bperm,  	  /* buffer permissions */
                             unsigned int options,	  /* validation options */
                             unsigned int *result,	  /* option results */
                             OSADDRESS supervisor_mode,   /* supervisior mode*/
                             U32 hal_options)
                     
{

   log_size_t psize;                              /* page size */
   U32 pperm;   				  /* page permissions */
   log_size_t asize = 0;                          /* accessable size */
   log_addr_t tbase = bbase;			  /* temporary base */

   struct partition_class *p;
   
   if (domain > gmm_state.max_domains)
   {
      return MM_ENON_EXISTING_DOMAIN;
   }
   
#ifdef softOSE
   /* Perform some form of very rudimentary validation on sfk. */
   if (bbase < 0x1000)
   {
      return 0;
   }
#endif

   /* Adjust permissions. */
   bperm &= (MM_PERM_READ|MM_PERM_WRITE);
   
   if (supervisor_mode)
   {
      bperm <<= MM_SU_SHIFT;
   }
   
#ifdef OSMIPS
   /*
    * The monolith resides in KSEG0/KSEG1 on MIPS. These areas are not mapped
    * in the MMU HAL and are therefore considered unaccessible without this
    * this piece of code.
    */
   if (mm_log_inside(bbase, KSEG0_BASE, KSEG0_SIZE) && supervisor_mode)
   {
      size_t kseg = KSEG0_SIZE - (bbase - KSEG0_BASE);
      if (result != NULL)
         *result = 0;
      return kseg > bsize ? bsize : kseg;
   }
   if (mm_log_inside(bbase, KSEG1_BASE, KSEG1_SIZE) && supervisor_mode)
   {
      size_t kseg = KSEG1_SIZE - (bbase - KSEG1_BASE);
      if (result != NULL)
         *result = 0;
      return kseg > bsize ? bsize : kseg;
   }
#endif
   
   while(asize <= bsize)
   {
      /* Check permissions for all pages used to map buffer */
      if(hal_lookup(domain, tbase, NULL, &psize, &pperm, hal_options) == HAL_SUCCESS)
      {
         /* Abort further test when permissions gets insufficient */
         if ((pperm & bperm) != bperm)
         {
            break;
         }

         /* Align to page size and increment */
         tbase &= ~(psize-1);
         tbase += psize;
      }
      else if((hal_create_resident_mapping != NULL) &&
              ((p = mm_space_is_resident(tbase, 1)) != NULL))
      {
         /* Multiple banks is not supported. Currently,
          * a SASE log_mem cannot have multiple banks. */
         if (p->banks->next != NULL)
         {
            INTERNAL_ERROR();
         }

         if((bperm & p->res_perm) != bperm)
         {  
            break;
         }

         tbase = p->banks->top + 1;            
      }
      else
      {
         /* No lookup succeeded. Abort. The address may point to a
          * SU_NA_USR_NA region for which access always is denied. */
         break;
      }

      /* Update resulting accessible size so far */
      asize = tbase - bbase;

      /* Abort further testing when the current
       * address wraps at the top of the address space. */
      if (tbase == 0)
      {
         break;
      }
   }

   if(asize > bsize)
   {
      asize = bsize;
   }

   /* Check the result pointer */
   if (result != NULL)
   {
      /* If requested - check if this buffer resides in MAS mapped memory */
      if (options & MM_VALIDATE_IN_MAS)
      {
         if (mm_space_is_mas(bbase, asize))
         {
            *result |= MM_VALIDATE_IN_MAS;	  /* Reset IN_MAS flag */
         }
         else
         {
            *result &= ~MM_VALIDATE_IN_MAS;	  /* Reset IN_MAS flag */
         }
      }
   }
   return asize;
}

/*****************************************************************************/

size_t
mm_validate_buffer(log_addr_t bbase,	          /* buffer base */
		   size_t bsize,	          /* buffer size */
		   PROCESS pid,		          /* validation context */
		   unsigned int bperm,  	  /* buffer permissions */
		   unsigned int options,	  /* validation options */
		   unsigned int *result)	  /* option results */
{
   OSADDRESS supervisor_mode = 1;
   OSDOMAIN domain = MM_KERNEL_DOMAIN;
   U32 hal_options = 0;                           /* Options passed to hal */
   OSBOOLEAN manager_mode = 0;
   
   VERBOZE(("mm_validate_buffer(base=0x%08x size=0x%08x "
	    "pid=0x%08x perm=0x%08x options=0x%08x)\n",
	    bbase, bsize, pid, bperm, options));
   
   /* Check that the kernel has started and that we can use system calls */
   if (ose_started())
   {
      /* Use current process if pid is 0 */
      if (pid == 0)
      {
	 pid = current_process();
      }
      else
      {
	 /* Check pid. Only allow process and block identifiers */
	 if (zzkrn_get_pid_info(pid, 9) > OS_BLOCK)
	 {
	    return 0;
	 }
      }
      
      /* Get execution mode for access process */
      supervisor_mode = zzkrn_get_pid_info(pid, 10);
   
      /* Get domain for process */
      domain = zzkrn_get_pid_info(pid, 16);
      
      manager_mode = zzkrn_get_pid_info(pid, 20);
      if (manager_mode)
      {
	 hal_options |= HAL_MANAGER_MODE;
      }      
   }

   return mm_validate_buffer_by_domain(
      bbase, bsize, domain, bperm, options,
      result, supervisor_mode, hal_options);
}


/*****************************************************************************
** Lookup the physical address for a logical pointer.
** See mm.h for more information.
*/
mm_status_t
mm_l2p_by_domain(OSDOMAIN domain, log_addr_t  la, phys_addr_t *pa)
{
   log_size_t size = 0;

   struct partition_class *p;
   
   if (domain > gmm_state.max_domains)
   {
      return MM_ENON_EXISTING_DOMAIN;
   }

   if (hal_lookup(domain, la, pa, &size, NULL, 0) == HAL_SUCCESS)
   {
      /* Adjust result for non page-aligned requests */
      *pa |= (phys_addr_t)(la & ((log_addr_t)size-1));
   }
   else if((hal_create_resident_mapping != NULL) &&
           (p = mm_space_is_resident(la, 1)) != NULL)
   {
      /* l2p() in a resident log_mem will always return
       * a physical address that is identical to the logical.
       * This works under the assumption that the log_mem has
       * SASE mapping type. Check that this is true. */
      if (partition_is_virtual(p) || partition_is_non_exclusive(p))
      {
         INTERNAL_ERROR();
      }
      
      /* Multiple banks is not supported. Currently,
       * a SASE log_mem cannot have multiple banks. */
      if (p->banks->next != NULL)
      {
         INTERNAL_ERROR();
      }

      *pa = (phys_addr_t)la;
   }
   else
   {
      log_size_t offset;
      phys_addr_t pa_start;
      phys_size_t size;
      
      struct region *region;

      /* This part of the lookup is necessary in order to support translations
       * of SU_NA_USR_NA regions that are located in the same domain.
       *
       * The lookup is used by forced_write() and is needed to support
       * writing of vectors to the exception area when it is configured
       * with no access. Note that the translation will fail for globally
       * visible regions (in system domain or other application domain).
       * There is currenly no use-case to motivate further tree searches. */

      region = mm_find_region_by_addr(
         &gmm_state.domains, domain, la, FALSE);
      if(region != NULL && region->phys_object != NULL)
      {
         offset = la - region->log_object->base;
         mm_extract_phys_addr(region->phys_object, offset, pa, &pa_start, &size);
         return MM_SUCCESS;
      }

      return MM_ENO_PHYSICAL_MAPPING;
   }

   return MM_SUCCESS;
}

/*****************************************************************************/

mm_status_t
mm_l2p(PROCESS pid, log_addr_t la, phys_addr_t *pa)
{
   OSDOMAIN domain = MM_KERNEL_DOMAIN;

   if (ose_started())
   {
      if (pid == 0)
      {
	 pid = current_process();
      }
      
      domain = (OSDOMAIN)zzkrn_get_pid_info(pid, 16);
   }
   return mm_l2p_by_domain(domain, la, pa);
}


/*****************************************************************************
** Lookup the physical address for a logical pointer.
** See mm.h for more information.
*/

static mm_status_t
mm_l2p_range(PROCESS      pid,
             log_addr_t   la,
             phys_addr_t *pa,
             size_t      *range)
{
   OSDOMAIN domain = MM_KERNEL_DOMAIN;
   log_addr_t  t_la = la;
   phys_addr_t t_pa;
   log_size_t  t_size = 0;
   struct partition_class *p = NULL;
   struct region *region;

   if (ose_started())
   {
      if (pid == 0)
      {
	 pid = current_process();
      }
      
      domain = (OSDOMAIN)zzkrn_get_pid_info(pid, 16);
      
      if (domain > gmm_state.max_domains)
      {
	 return MM_ENON_EXISTING_DOMAIN;
      }
      
   }

   if (hal_lookup(domain, t_la, &t_pa, &t_size, NULL, 0) == HAL_SUCCESS)
   {
      /* Adjust result for non page-aligned requests */
      if (pa != NULL)
      {
         *pa = t_pa | (phys_addr_t)(la & ((log_addr_t)t_size-1));
      }

      t_pa += t_size;
      t_size -= (phys_addr_t)(la & ((log_addr_t)t_size-1));
      t_la += t_size;

   }
   else if((hal_create_resident_mapping != NULL) &&
           (p = mm_space_is_resident(la, 1)) != NULL)
   {
      /* l2p() in a resident log_mem will always return
       * a physical address that is identical to the logical.
       * This works under the assumption that the log_mem has
       * SASE mapping type. Check that this is true. */
      if (partition_is_virtual(p) || partition_is_non_exclusive(p))
      {
         INTERNAL_ERROR();
      }
      
      /* Multiple banks is not supported. Currently,
       * a SASE log_mem cannot have multiple banks. */
      if (p->banks->next != NULL)
      {
         INTERNAL_ERROR();
      }

      *pa  = la;
      t_size = p->banks->top - la + 1;         
      t_pa = p->banks->top + 1;
      t_la = p->banks->top + 1;         
   }
   else
   {
      log_size_t offset;
      phys_addr_t pa_start;
      phys_size_t size;

      /* This part of the lookup is necessary in order to support translations
       * of SU_NA_USR_NA regions that are located in the same domain.
       *
       * The lookup is used by forced_write() and is needed to support
       * writing of vectors to the exception area when it is configured
       * with no access. Note that the translation will fail for globally
       * visible regions (in system domain or other application domain).
       * There is currenly no use-case to motivate further tree searches. */

      region = mm_find_region_by_addr(
         &gmm_state.domains, domain, la, FALSE);
      if((region == NULL) || (region->phys_object == NULL))
      {
         return MM_ENO_PHYSICAL_MAPPING;
      }

      offset = la - region->log_object->base;
      mm_extract_phys_addr(region->phys_object, offset, pa, &pa_start, &size);
      t_size += size - offset;
      t_la += size - offset;
      t_pa = pa_start + size;
   }

   while((range != NULL) && ((*range == 0) || (*range > t_size)) && (t_la != 0))
   {
      log_size_t  page_size;
      phys_addr_t phys;
      
      if (hal_lookup(domain, t_la, &phys, &page_size, NULL, 0) == HAL_SUCCESS)
      {   
         /* Determine how far the physical memory is contiguous. */
         if (phys != t_pa)
         {
            break;
         }

         t_pa   += page_size;
         t_size += page_size;
         t_la   += page_size;         
      }
      else if ((hal_create_resident_mapping != NULL) &&
               (p = mm_space_is_resident(t_la, 1)) != NULL)
      {
         /* l2p() in a resident log_mem will always return
          * a physical address that is identical to the logical.
          * This works under the assumption that the log_mem has
          * SASE mapping type. Check that this is true. */
         if (partition_is_virtual(p) || partition_is_non_exclusive(p))
         {
            INTERNAL_ERROR();
         }
      
         /* Multiple banks is not supported. Currently,
          * a SASE log_mem cannot have multiple banks. */
         if (p->banks->next != NULL)
         {
            INTERNAL_ERROR();
         }

         if(p->banks->base != t_pa)
         {
            break;
         }
            
         t_size += p->banks->top - t_la + 1;
         t_la = p->banks->top + 1;
         t_pa = p->banks->top + 1;
      }
      else
      {
         phys_addr_t pa_start;
         phys_size_t size;

         region = mm_find_region_by_addr(
            &gmm_state.domains, domain, t_la, FALSE);
                  
         if(region == NULL ||
            region->phys_object == NULL ||
            region->permissions != SU_NA_USR_NA)
         {
            break;
         }

         mm_extract_phys_addr(region->phys_object, 0, &phys, &pa_start, &size);
            
         if(phys != t_pa)
         {
            break;
         }               

         t_size += size + pa_start - phys;
         t_la += size + pa_start - phys;
         t_pa = pa_start + size;
      }
   }

   if(range != NULL)
   {
      if (*range == 0)
      {
         *range = (size_t)t_size;
      }
      else
      {
         *range = (t_size < *range) ? (size_t)t_size : *range;
      }
   }
   return MM_SUCCESS;
}

/* Tries to free pages in the pager page cache */
static mm_status_t
mm_reclaim_region(OSDOMAIN *domain, log_addr_t *addr, size_t *size)
{
   struct mm_reclaim_physical_event_t event;

   event.domain = MM_EVENT_ANY_DOMAIN;
   event.phys_mem = MM_INTERNAL_RAM;
   event.internal_identifier = RECLAIM_ALL_INTERNAL;
   event.size = *size;
   event.inquiry = 0;
   event.event_no = MM_EVENT_RECLAIM_REGION_PHYSICAL;
   call_reclaim_region_event(&event);

   *domain = event.domain;
   *addr = event.base;
   *size = event.size;
   return MM_SUCCESS;
}

/**************************************************************************/

mm_status_t
mm_get_cache_mode(const char *ctype, const char *mtype, U32 *cache_mode);

mm_status_t
mm_get_cache_mode(const char *ctype, const char *mtype, U32 *cache_mode)
{
   hal_status_t status;
   
   status = hal_get_cache_mode(ctype, mtype, cache_mode);
   return status == HAL_SUCCESS ? MM_SUCCESS : MM_EILLEGAL_CACHE_MODE;
}

/**************************************************************************/

mm_status_t
mm_get_cache_type(U32 cache_mode, char *cache_type, char *memory_type);

mm_status_t
mm_get_cache_type(U32 cache_mode, char *cache_type, char *memory_type)
{
   hal_status_t status;

   /* With both 'cache_type' and 'memory_type' == NULL,
    * the funciton may be used to validate the cache mode. */

   status = hal_get_cache_type(cache_mode, cache_type, memory_type);
   return status == HAL_SUCCESS ? MM_SUCCESS : MM_EILLEGAL_CACHE_MODE;
}

/**************************************************************************/


/*
 *===========================================================================
 *                        mm_trap_interface
 *===========================================================================
 */

static
long mm_trap_interface(long fc,
		       long arg1,
		       long arg2,
		       long arg3,
		       long arg4,
		       long arg5,
		       long arg6)
{
   long result;

   switch (fc)
   {
      case MM_TRAP_GET_INTERFACE_VERSION:
      {
	 result = (long)MM_TRAP_INTERFACE_VERSION;
	 break;
      }
      case MM_TRAP_GET_LAST_FCODE:
      {
	 /* Update when adding function codes to the MM trap interface */
	 result = (long)MM_TRAP_GET_CACHE_TYPE;
	 break;
      }
      case MM_TRAP_VALIDATE_BUFFER:
      {
	 result = (long)mm_validate_buffer((log_addr_t)   arg1,
					   (size_t)       arg2,
					   (PROCESS)      arg3,
					   (unsigned int) arg4,
					   (unsigned int) arg5,
					   (unsigned int*)arg6);
	 break;
      }
      case MM_TRAP_L2P:
      {
	 result = (long)mm_l2p((PROCESS)      arg1,
			       (log_addr_t)   arg2,
			       (phys_addr_t*) arg3);
	 break;
      }
      case MM_TRAP_L2P_RANGE:
      {
         result = mm_l2p_range((PROCESS)     arg1,
                               (log_addr_t)  arg2,
                               (phys_addr_t*)arg3,
                               (size_t*)     arg4);
         break;
      }
      case MM_TRAP_RECLAIM_REGION:
      {
         result = mm_reclaim_region((OSDOMAIN*)arg1,
                                    (log_addr_t*)arg2,
                                    (size_t*)arg3);
         break;
      }
      case MM_TRAP_ACCESS_FORCED_WRITE:
      {
         result = mm_access_forced_write((OSDOMAIN)arg1,
                                         (void *)arg2,
                                         (const void *)arg3,
                                         (size_t)arg4,
                                         (int)arg5);
         break;
      }
      case MM_TRAP_GET_CACHE_MODE:
      {
         result = mm_get_cache_mode(
            (const char *) arg1,
            (const char *) arg2,
            (U32 *) arg3);
         break;
      }
      case MM_TRAP_GET_CACHE_TYPE:
      {
         result = mm_get_cache_type(
            (U32) arg1,
            (char *) arg2,
            (char *) arg3);
         break;
      }
      default:
      {
	 result = -1;
	 break;
      }
   }
   return result;
}


/*
 *===========================================================================
 *                        mm_init_bios_interface
 *===========================================================================
 */

void
mm_init_bios_interface(void)
{
   long status;

   status = biosInstall(MM_BIOS_NAME, mm_trap_interface, BIOS_USERMODE_ALLOWED);

   if(status != BIOS_SUCCESS)
   {
      /* The install failed. */
      error(OSE_MM_EBIOS_INSTALL_FAILED);
      stop(current_process());
   }
}
