/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2006 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

#include "demand_pager.h"
#include "mm_util.h"
#include "mmtypes.h"
#include "mm.sig"
#include "mm_internal.sig"
#include "mm.h"
#include "pose.h"
#include "slab.h"
#include "compiler.h"
#include "gmm.h"
#include "mm_util_internal.h"
#include "common.h"

#define ANY_COMPONENT  (USER_DEFINED_COMPONENT_START - 1)

#define	DECONST(type, var) \
  ((type)(ptrdiff_t)(const void *)(var))

struct _mm_event_t
{
   list_link_t link;
   union mm_event_t mm_event;
};

/* A slab cache that we need for the event allocations we perform. */
static struct slab_cache *mm_event_slab;

static list_t event_list[MM_NO_OF_EVENTS];
static mm_event_policy_t *event_policy[MM_NO_OF_EVENTS];

MUTEX reclaim_region_mutex;


mm_status_t
ose_mm_map_range(OSDOMAIN    domain,
                 log_addr_t  la,
                 phys_addr_t pa,
                 size_t      size,
                 U32         permission,
                 U32         cache_mode,
                 int         options)
{
   struct hal_domain *hal_domain;
   hal_status_t hal_status = HAL_SUCCESS;
   size_t page_size, mapped_len;
   U32 aperm, acache_mode;

#if 0
   ramlog_printf("MAP:   domain=%d la=%#lx, pa=%#llx, size=%d, "
		 "permission=%#lx, cache_mode=%#lx, options=%#lx\n",
		 domain, la, pa, size, permission, cache_mode, options);
#endif

   if (!gmm_state.multithread_enable)
   {
      return MM_EILLEGAL_OPERATION;
   }

   /* Parameter sanity checks */

   /* Valid combination of permission flags ? */
   if (mm_check_permission(permission) == FALSE)
   {
      return MM_EILLEGAL_PERMISSION;
   }

   /* Permissions supported by HAL ? */
   if (hal_actual_permissions(&aperm, permission, la, 0) != HAL_SUCCESS)
   {
      return MM_EILLEGAL_PERMISSION;
   }

   /* Valid combination of cache modes ? */
   if (mm_check_cache_mode(cache_mode) == FALSE)
   {
      return MM_EILLEGAL_CACHE_MODE;
   }

   if (hal_actual_cache_mode(&acache_mode, cache_mode, la) != HAL_SUCCESS)
   {
      return MM_EILLEGAL_CACHE_MODE;
   }

   if (domain >= gmm_state.max_domains ||
       gmm_state.domains.domain_list[domain] == NULL ||
       gmm_state.domains.domain_list[domain]->hal_domain == NULL)
   {
      return MM_ENON_EXISTING_DOMAIN;
   }

   /* FIXME: Check that we only map in our own domain. */


   hal_domain = gmm_state.domains.domain_list[domain]->hal_domain;

   /* FIXME: Do we always want to map the smallest page size ? */
   page_size = hal_get_info()->page_sizes[0];

   /* Addresses correctly aligned ? */
   if (!mm_log_aligned(la, page_size) ||
       !mm_log_aligned(pa, page_size))
   {
      return MM_EILLEGAL_BASE_ADDRESS;
   }

   /* Valid options ? */
   if (options & ~(HAL_CHANGE|HAL_MANAGER_MODE))
   {
      return MM_EOPTION_NOT_SUPPORTED;
   }

   /* Optimize by making all regions *created in system domain* as
    * TLB-global (TID=0 etc.). 
    */ 
   if (MM_KERNEL_DOMAIN == domain)
   {
      options |= HAL_GLOBAL_MAPPING;
   }

   if (size == 0)
   {
      return MM_ESIZE_IS_ZERO;
   }

   if (!mm_log_aligned(size, page_size))
   {
      return MM_EILLEGAL_OPERATION;
   }

#if defined(OSPP) || defined (OSMIPS)
   /*
    * For PPC and MIPS, it is not possible to create CI regions with
    * write disallowed, since we are invalidating the region.
    * Therefore we must set the su-write permission, create the region
    * and then perform a change permission back to the intended
    * permission.
    */
   if (acache_mode & MM_CACHE_INHIBITED)
   {
      U32 tperm = aperm | MM_SU_READ | MM_SU_WRITE;
      OSBOOLEAN rc;

      mapped_len = 0;

      while (mapped_len < size)
      {
         hal_status = hal_map_page(hal_domain, la+mapped_len, pa+mapped_len,
                                   page_size, options, tperm, acache_mode);

         if (hal_status != HAL_SUCCESS)
         {
            if (mapped_len > 0)
            {
               hal_unmap_range(hal_domain, la, mapped_len, options);
            }
            return mm_hal_status_to_mm_status(hal_status);
         }

         mapped_len += page_size;
      }

      
      if (options & HAL_MAS)
      {
         /* Temporarily go to mas domain to be able to
          * do cache invalidate.
          */
         rc = set_segment(0, domain);

         if (rc != 0)
         {
            hal_unmap_range(hal_domain, la, size, options);
            ramlog_printf("mm: Invalidation of cache inhibit MAS region "\
                          "failed since the domain is not active. base: "\
                          "%#lx size: %#lx domain: %lx \n",
                          la, size, domain);
            return MM_EDOMAIN_NOT_MAS_ACTIVE;
         }

         cache_invalidate_d(la, size);
         rc = set_segment(1,0);                         /* Go back. */
      }
      else
      {
         cache_invalidate_d(la, size);
      }

      options |= HAL_CHANGE;
   }
#endif

   mapped_len = 0;

   while (mapped_len < size)
   {
      hal_status = hal_map_page(hal_domain, la+mapped_len, pa+mapped_len,
                                page_size, options, aperm, acache_mode);

      if ((hal_status == HAL_EEND_OF_PHYS_MEMORY ||
           hal_status == HAL_EEND_OF_LOG_MEMORY) &&
          ose_mm_reclaim_component_memory(&gmm_state, PAGE_SIZE,
                                          MM_INTERNAL_RAM))
      {
         continue;
      }

      if (hal_status != HAL_SUCCESS)
      {
         if (mapped_len > 0)
         {
            hal_unmap_range(hal_domain, la, mapped_len, options);
         }
         break;
      }

      mapped_len += page_size;
   }

   return mm_hal_status_to_mm_status(hal_status);
}

mm_status_t
ose_mm_unmap_range(OSDOMAIN      domain,
                   log_addr_t    la,
                   size_t        size,
                   int           options)
{
   struct hal_domain *hal_domain;
   size_t page_size;

#if 0
   ramlog_printf("UNMAP: domain=%d la=%#lx, size=%d, options=%#lx\n",
		 domain, la, size, options);
#endif

   if (!gmm_state.multithread_enable)
   {
      return MM_EILLEGAL_OPERATION;
   }

   if (domain >= gmm_state.max_domains ||
       gmm_state.domains.domain_list[domain] == NULL ||
       gmm_state.domains.domain_list[domain]->hal_domain == NULL)
   {
      return MM_ENON_EXISTING_DOMAIN;
   }

   /* FIXME: Do we always want to map the smallest page size ? */
   page_size = hal_get_info()->page_sizes[0];

   if (!mm_log_aligned(la, page_size))
   {
      return MM_EILLEGAL_BASE_ADDRESS;
   }

   if (options != 0)
   {
      return MM_EOPTION_NOT_SUPPORTED;
   }

   if (size == 0)
   {
      return MM_ESIZE_IS_ZERO;
   }

   if (!mm_log_aligned(size, page_size))
   {
      return MM_EILLEGAL_OPERATION;
   }

   hal_domain = gmm_state.domains.domain_list[domain]->hal_domain;
   hal_unmap_range(hal_domain, la, size, options);

   return MM_SUCCESS;
}

/*
 *===========================================================================
 *                     mm_get_free_mem
 *===========================================================================
 *
 * Description
 * -----------
 * Get total amount free MM memory that is not used in the supplied
 * physical memory type.
 *
 * Parameters
 * ----------
 * char*              The name of the physical memory type.
 *
 * Returns
 * -------
 * size_t             Returns the amount of available memory in the provided
 *                    physical memory type. If the type does not exist,
 *                    0 is returned.
 *
 */
static size_t
mm_get_free_mem(const char *phys_mem)
{
   struct mm_get_phys_mem_info mem_info;
   struct mm_get_phys_mem_info *mem_info_p = &mem_info;

   strncpy(mem_info.name, phys_mem, MM_NAME_LENGTH + 1);
   mem_info.count = 1;
   mem_info.tags[0].tag_id = MM_TAG_INFO_PHYS_MEM_FREE_MEMORY;

   mem_info_p = (struct mm_get_phys_mem_info *)
      mm_get_phys_mem_info(&gmm_state, (union SIGNAL*)mem_info_p);

   if (mem_info_p->status == MM_ENON_EXISTING_PHYS_MEM)
   {
      return (size_t)0;
   }
   else
   {
      return (size_t)mem_info_p->tags[0].tag_value;
   }
}

/*
 *===========================================================================
 *                     mm_get_base_address
 *===========================================================================
 *
 * Description
 * -----------
 * Get the base address of the supplied logical partition.
 *
 * Parameters
 * ----------
 * char*              The name of the logical_region.
 *
 * Returns
 * -------
 * log_addr_t         Returns the base address of the logical region.
 *                    If the type does not exist, 0 is returned.
 *
 */
static log_addr_t
mm_get_base_address(const char *log_mem)
{
   struct partition_class *p_class;
   log_addr_t base;
   int status;

   /* FIXME(rand) - Assuming only one logical bank in the partition */
   p_class = mm_find_partition_by_name(log_mem);
   if (p_class == NULL)
   {
      return 0;
   }

   if (p_class->get_bank_info == NULL)
   {
      return 0;
   }
   
   status = p_class->get_bank_info(p_class, 0, &base, NULL);
   if (!status)
   {
      return 0;
   }
   
   return base;
}

/*
 *===========================================================================
 *                     ose_mm_get_freeable_memory
 *===========================================================================
 *
 * Description
 * -----------
 * Get total amount free MM memory that is available to a
 * specified component in the supplied physical memory type.
 *
 * Parameters
 * ----------
 * char*              The name of the physical memory type
 * size               The max amount of memory the caller is interested
 *                    in. Whenever the reclaim framework finds this much
 *                    free memory, it returns without asking any more
 *                    componenets, thus improving performance.
 * identifier         The identifier of the component that asks.
 *
 * Returns
 * -------
 * size_t             Returns the amount of available memory in the provided
 *                    physical memory type. If the region does not exist,
 *                    0 is returned.
 *
 */
size_t
ose_mm_get_freeable_memory(const char *phys_mem,
                           size_t size,
                           ose_component_identifier_t identifier)
{
   struct mm_reclaim_physical_event_t event;
   size_t free_mem = mm_get_free_mem(phys_mem);

   if (free_mem >= size)
   {
      return free_mem;
   }
   /* We assume that the log_mem has the same name as the phys_mem that it
    * mostly uses. */
   event.base = mm_get_base_address(phys_mem);
   event.domain = MM_EVENT_ANY_DOMAIN;
   event.phys_mem = phys_mem;
   event.identifier = identifier;
   event.inquiry = 1;
   event.size = size - free_mem;
   event.event_no = MM_EVENT_RECLAIM_REGION_PHYSICAL;
   event.internal_identifier = RECLAIM_ALL_INTERNAL;
   /* Check how much memory that could be reclaimed from the components. */
   call_reclaim_region_event(&event);

   return event.size + free_mem;
}

/*
 *===========================================================================
 *                        ose_mm_get_free_mem
 *===========================================================================
 *
 * Description
 * -----------
 * Returns the number of bytes that are free to use in the "RAM" physical
 * memory type.
 *
 * Returns
 * -------
 * size_t             Returns the amount of available memory
 *
 */
size_t
ose_mm_get_free_mem(void)
{
   return mm_get_free_mem(MM_INTERNAL_RAM);
}

/*
 *===========================================================================
 *                        find_event_handler
 *===========================================================================
 *
 * Description
 * -----------
 * Finds an event handler with the provided identifiers.
 *
 * Parameters
 * ----------
 * event_no               The event number.
 * identifier             The identifier of the registered callback I wish
 *                        to retrieve.
 * union mm_event_t       A pointer to the event.
 *
 * Returns
 * -------
 * struct _mm_event_t     Returns the event or NULL if none could be found.
 *
 */
static struct _mm_event_t *
find_event_handler(mm_event_no_t event_no,
                   ose_component_identifier_t identifier,
                   union mm_event_t *event)
{
   struct _mm_event_t *_event;

   if (event_no >= MM_NO_OF_EVENTS)
   {
      return NULL;
   }

   _event = (struct _mm_event_t *)list_head(&event_list[event_no]);
   while (_event)
   {
      if (identifier == ANY_COMPONENT)
      {
         return _event;
      }

      switch(event_no)
      {
         struct mm_reclaim_physical_event_t *event_list;
         struct mm_reclaim_physical_event_t *event_in;

         case MM_EVENT_RECLAIM_REGION_PHYSICAL:
            if (event == NULL)
            {
               ramlog_printf("Error: You need to supply an union mm_event_t for MM_EVENT_RECLAIM_REGION_PHYSICAL.\n");
               error2(OSE_MM_EINTERNAL_ERROR, 0);
            }
            event_list = &_event->mm_event.reclaim;
            event_in = &event->reclaim;
            if ((event_list->base >= event_in->base ||
                 event_list->base + event_list->size < event_in->base) &&
                event_list->identifier == identifier &&
                (event_in->domain == MM_EVENT_ANY_DOMAIN ||
                 event_list->domain == event_in->domain))
            {
               return _event;
            }
            break;

         case MM_EVENT_REGION_DELETED:
            if (_event->mm_event.delete.identifier == identifier)
            {
               return _event;
            }
            break;
      }

      _event = (struct _mm_event_t *)list_next(&event_list[event_no],
                                              &_event->link);
   }
   return NULL;
}

/*
 *===========================================================================
 *                        ose_mm_install_handler
 *===========================================================================
 *
 * Description
 * -----------
 * Installs a callback handler in the MM.
 * For all events you need to fill in event_no, event_handler and
 * identifier in the mm_event_t union. Fill in the event you
 * would like to register for in event_no, your handler in
 * event_handler and the identity of your component in the identity
 * field.
 * See each event for event specific fields that need to be filled in
 * at instal time.
 *
 * Parameters
 * ----------
 * event        A pointer to the event to register for.
 * beginning    Insert the event in the beginning of the event queue,
 *              instead of the end, which is the default.
 *
 * Returns
 * -------
 * mm_status_t        Returns MM_SUCCESS for success.
 *
 */
mm_status_t
ose_mm_install_handler(union mm_event_t *event_raw,
                       OSBOOLEAN beginning)
{
   struct _mm_event_t *_event;
   struct mm_reclaim_physical_event_t *event = &event_raw->reclaim;
   size_t event_struct_size;

   if (event->event_no >= MM_NO_OF_EVENTS)
   {
      return MM_ENON_EXISTING_EVENT;
   }
   else if (!event->event_handler)
   {
      return MM_ENON_EXISTING_EVENT_HANDLER;
   }

   if (event->event_no != MM_EVENT_REGION_DELETED)
   {
      event_struct_size = sizeof(struct mm_reclaim_physical_event_t);
   }
   else
   {
      event_struct_size = sizeof(struct mm_region_deleted_event_t);
   }
   _event = (struct _mm_event_t *)slab_alloc(mm_event_slab, SLAB_SLEEP);
   if (!_event)
   {
      return MM_EOUT_OF_PHYS_MEM;
   }

   init_list_link(&_event->link);
   memcpy(&_event->mm_event, event, event_struct_size);
   if (event->event_no == MM_EVENT_RECLAIM_REGION_PHYSICAL)
   {
      struct mm_reclaim_physical_event_t *reclaim =
         &_event->mm_event.reclaim;

      reclaim->phys_mem = (char *)slab_alloc(mm_event_slab, SLAB_SLEEP);
      if (!reclaim->phys_mem)
      {
         slab_free(mm_event_slab, _event);
         return MM_EOUT_OF_PHYS_MEM;
      }
      strlcpy(DECONST(char *, reclaim->phys_mem), event->phys_mem,
              MM_NAME_LENGTH + 1);
   }

   if (beginning)
   {
      insert_list_head(&event_list[event->event_no], &_event->link);
   }
   else
   {
      insert_list_tail(&event_list[event->event_no], &_event->link);
   }

   return MM_SUCCESS;
}

/*
 *===========================================================================
 *                        ose_mm_uninstall_handler
 *===========================================================================
 *
 * Description
 * -----------
 * Uninstalls an event handler in the MM.
 * The MM event is identified by the supplied mm_event_t union.
 * See each event for event specific fields that need to be filled in
 * for successful identification.
 *
 * Parameters
 * ----------
 * event          The event this handler should be removed from.
 *                The handler is identified by event_no, event_handler
 *                and identifier plus any event specific fields.
 *
 * Returns
 * -------
 * mm_status_t    Returns MM_SUCCESS for success.
 *
 */
mm_status_t
ose_mm_uninstall_handler(union mm_event_t *event_raw)
{
   struct _mm_event_t *_event;
   struct mm_reclaim_physical_event_t *event = &event_raw->reclaim;

   if (event->event_no >= MM_NO_OF_EVENTS)
   {
      return MM_ENON_EXISTING_EVENT;
   }
   else if (!event->event_handler)
   {
      return MM_ENON_EXISTING_EVENT_HANDLER;
   }

   _event = find_event_handler(event->event_no, event->identifier, event_raw);
   if (_event)
   {
      delete_from_list(&event_list[event->event_no], &_event->link);
      slab_free(mm_event_slab, _event);
      return MM_SUCCESS;
   }

   return MM_ENON_EXISTING_EVENT_HANDLER;
}

/*
 *===========================================================================
 *                        ose_mm_install_policy
 *===========================================================================
 *
 * Description
 * -----------
 * Install a policy for an MM event.
 * The policy installed should pick what registered component to
 * call for this event and call it. Once it has gotten a satisfactory
 * result back (for example, a callback has indicated a region to
 * reclaim), it should return.
 * If no policy is installed for an event, FIFO is used.
 *
 * Parameters
 * ----------
 * event_no        The id of the event this policy should be called for.
 * policy_handler  The policy function called every time an event is
 *                 triggered.
 *
 * Returns
 * -------
 * mm_status_t        Returns MM_SUCCESS for success.
 *
 */
mm_status_t
ose_mm_install_policy(mm_event_no_t event_no,
                      mm_event_policy_t *policy_handler)
{
   if (event_no >= MM_NO_OF_EVENTS)
   {
      return MM_ENON_EXISTING_EVENT;
   }
   else if (event_policy[event_no] != NULL)
   {
      return MM_EVENT_POLICY_ALREADY_INSTALLED;
   }
   else if (policy_handler == NULL)
   {
      return MM_EILLEGAL_OPERATION;
   }

   event_policy[event_no] = policy_handler;
   return MM_SUCCESS;
}

/*
 *===========================================================================
 *                        ose_mm_uninstall_policy
 *===========================================================================
 *
 * Description
 * -----------
 * Uninstall a policy for an MM event.
 * Will revert the policy to the default FIFO.
 *
 * Parameters
 * ----------
 * event_no        The id of the event this policy should be uninstalled for.
 *
 * Returns
 * -------
 * mm_status_t     Returns MM_SUCCESS for success.
 *
 */
mm_status_t
ose_mm_uninstall_policy(mm_event_no_t event_no)
{
   if (event_no >= MM_NO_OF_EVENTS)
   {
      return MM_ENON_EXISTING_EVENT;
   }
   else if (event_policy[event_no] == NULL)
   {
      return MM_EVENT_POLICY_NOT_INSTALLED;
   }

   event_policy[event_no] = NULL;
   return MM_SUCCESS;
}

/*
 *===========================================================================
 *                        ose_mm_get_event_handler
 *===========================================================================
 *
 * Description
 * -----------
 * Gets the event handler identified by the supplied name.
 * To be used in an MM event policy handler. If the handler is not
 * registered in the supplied logical memory partition, NULL is
 * returned.
 *
 * Parameters
 * ----------
 * event_no     The id of the event this handler should be called for.
 * identifier   The identifier of the registered callback I wish
 *              to retrieve.
 * event        Certain events might require that event specific fields
 *              are also filled in for successful
 *              identification.
 *
 * Returns
 * -------
 * mm_event_callback_t   The handler corresponding to the supplied name, or
 *                       NULL if none was found.
 *
 */
mm_event_callback_t *
ose_mm_get_event_handler(mm_event_no_t event_no,
                         ose_component_identifier_t identifier,
                         union mm_event_t *event)
{
   struct _mm_event_t *_event;

   _event = find_event_handler(event_no, identifier, event);
   if (_event)
   {
      return _event->mm_event.reclaim.event_handler;
   }
   else
   {
      return NULL;
   }
}

/*
 *===========================================================================
 *                     ose_mm_get_next_event_handler
 *===========================================================================
 *
 * Description
 * -----------
 * Gets the next event handler of the supplied event.
 * To be used in an MM event policy handler.
 * If there are no more handlers registered in the supplied domain and
 * logical memory address, NULL is returned.
 *
 * Parameters
 * ----------
 * event_no     The id of the event this handler should be called for.
 * handler      The current event handler, or NULL if we should get
 *              the first registered handler.
 *
 * Returns
 * -------
 * mm_event_callback_t  The next handler, or NULL if it does not exist.
 *
 */
mm_event_callback_t *
ose_mm_get_next_event_handler(mm_event_no_t event_no,
                              mm_event_callback_t *handler)
{
   struct _mm_event_t *_event;

   if (handler == NULL)
   {
      _event = (struct _mm_event_t *)list_head(&event_list[event_no]);
   }
   else
   {
      _event = find_event_handler(event_no, ANY_COMPONENT, NULL);
      while (!_event && _event->mm_event.reclaim.event_handler != handler)
      {
         _event = (struct _mm_event_t *)list_next(&event_list[event_no],
                                                  &_event->link);
      }
   }

   if (_event)
   {
      _event = (struct _mm_event_t *)list_next(&event_list[event_no],
                                               &_event->link);
      if (_event)
      {
         return _event->mm_event.reclaim.event_handler;
      }
   }
   return NULL;
}

/*
 *===========================================================================
 *                        call_region_deleted_event
 *===========================================================================
 *
 * Description
 * -----------
 * Calls all the registered region deleted events.
 *
 * Parameters
 * ----------
 * domain             The domain that got deleted.
 * base               The start address of the deleted region.
 *
 */
void
call_region_deleted_event(OSDOMAIN domain, log_addr_t base)
{
   struct _mm_event_t *event;
   struct mm_region_deleted_event_t event_parameters;

   event_parameters.event_no = MM_EVENT_REGION_DELETED;
   event_parameters.domain = domain;
   event_parameters.base = base;

   if (event_policy[MM_EVENT_REGION_DELETED] == NULL)
   {
      event = (struct _mm_event_t *)
         list_head(&event_list[MM_EVENT_REGION_DELETED]);
      while (event)
      {
         (event->mm_event.reclaim.event_handler)((union mm_event_t *)&event_parameters);

         event = (struct _mm_event_t *)
            list_next(&event_list[MM_EVENT_REGION_DELETED], &event->link);
      }
   }
   else
   {
      (event_policy[MM_EVENT_REGION_DELETED])((union mm_event_t *)&event_parameters);
   }
}

/*
 *===========================================================================
 *                        call_reclaim_region_event
 *===========================================================================
 *
 * Description
 * -----------
 * Calls all the registered reclaim region events and stops as soon as
 * it has found one that can be deleted.
 *
 */
void
call_reclaim_region_event(struct mm_reclaim_physical_event_t *reclaim)
{
   OSBOOLEAN kill_lock_status;

   kill_lock_status = zzkrn_kill_lock(0);
   if (unlikely(kill_lock_status == 0))
   {
      ramlog_printf("mm_util: kill_lock failed.\n");
      error2(OSE_MM_EINTERNAL_ERROR, 0);
   }

   if (ose_mutex_trylock(&reclaim_region_mutex))
   {
      struct _mm_event_t *_event;
      size_t requested_size = reclaim->size;
      size_t reclaimed_size = 0;

      reclaim->event_no = MM_EVENT_RECLAIM_REGION_PHYSICAL;

      if (event_policy[MM_EVENT_RECLAIM_REGION_PHYSICAL] == NULL)
      {
         _event = (struct _mm_event_t *)
            list_head(&event_list[MM_EVENT_RECLAIM_REGION_PHYSICAL]);
         while (_event)
         {
            if (reclaim->inquiry &&
                reclaim->identifier == _event->mm_event.reclaim.identifier)
            {
               /* We have reached the component that is asking. Break. */
               break;
            }

            reclaim->size = requested_size;
            (_event->mm_event.reclaim.event_handler)((union mm_event_t *)reclaim);
            reclaimed_size += reclaim->size;
            if ((!reclaim->inquiry && reclaim->size != 0) ||
                (reclaim->inquiry && reclaim->size > requested_size))
            {
               reclaim->identifier = _event->mm_event.reclaim.identifier;
               break;
            }
            requested_size -= reclaim->size;

            _event = (struct _mm_event_t *)
               list_next(&event_list[MM_EVENT_RECLAIM_REGION_PHYSICAL],
                         &_event->link);
         }

         reclaim->size = reclaimed_size;
      }
      else
      {
         (event_policy[MM_EVENT_RECLAIM_REGION_PHYSICAL])((union mm_event_t *)reclaim);
         reclaimed_size = reclaim->size;
      }

      ose_mutex_unlock(&reclaim_region_mutex);
   }
   else
   {
      /* Some memory has been reclaimed by another thread. If it is not enough
       * this reclaim will be called again. */
      ose_mutex_lock(&reclaim_region_mutex);
      ose_mutex_unlock(&reclaim_region_mutex);
      reclaim->size = 1;
   }

   zzkrn_kill_unlock(0);
}


/*
 *===========================================================================
 *                        mm_util_init
 *===========================================================================
 *
 * Description
 * -----------
 * Initializes the MM util functions.
 *
 */
void
mm_util_init(void)
{
   int i;
   size_t size = MM_NAME_LENGTH + 1;

   /* The slab size is the maximum of all event sizes with the MM name string
    * size included as it is allocated from here. */
   if (size < sizeof(struct mm_reclaim_physical_event_t) + sizeof(list_link_t))
   {
      size = sizeof(struct mm_reclaim_physical_event_t) + sizeof(list_link_t);
   }
   if (size < sizeof(struct mm_region_deleted_event_t) + sizeof(list_link_t))
   {
      size = sizeof(struct mm_region_deleted_event_t) + sizeof(list_link_t);
   }
   mm_event_slab = create_slab_allocator("mm_events",
                                         size, 0, 0, NULL, NULL, 0);

   for (i = 0; i < MM_NO_OF_EVENTS; i++)
   {
      init_list(&event_list[i]);
      event_policy[i] = NULL;
   }
}

/*
 *===========================================================================
 *                        ose_mm_multithread
 *===========================================================================
 *
 * Description
 * -----------
 * Enables or disables multithreading support in MM. Multithreading
 * support must be enabled before any use of the mm_util.h interface.
 * The function may be called several times by different clients.
 *
 * Parameters
 * ----------
 * enable        Non-zero to enable, zero to disable
 *
 * Returns
 * -------
 * mm_status_t   Returns MM_SUCCESS for success.
 *
 */
mm_status_t
ose_mm_multithread(mm_bool_t enable)
{
   struct mm_multithread *request_sig, *reply_sig;
   mm_status_t status;
   SIGSELECT reply_select[] = {1, MM_MULTITHREAD_REPLY};
   PROCESS mm_pid;


   /* Obtain GMM pid */
   hunt(GMM_NAME_STRING, 0, &mm_pid, NULL);
   if (mm_pid == 0)
   {
      return MM_EINTERNAL_ERROR;
   }

   request_sig = (struct mm_multithread*)
      alloc(sizeof(struct mm_multithread), MM_MULTITHREAD_REQUEST);

   request_sig->enable = enable;

   /* Make request */
   send((union SIGNAL**)&request_sig, mm_pid);
   reply_sig = (struct mm_multithread*)receive(reply_select);

   /* Return status to client */
   status = reply_sig->status;
   free_buf((union SIGNAL**)&reply_sig);

   return status;
}

