/* COPYRIGHT-ENEA-SRC-R2 *
**************************************************************************
 * Copyright (C) 2004-2013 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

/* C runtime include files. */
#include <stddef.h>
#include <string.h>

/* OSE public include files. */
#include "assert.h"
#include "list_ext.h"

/* OSE internal include files. */

/* MM internal include files. */
#include "common.h"
#include "compat.h"
#include "hal.h"
#include "gmm.h"
#include "mm_private.h"
#include "partition.h"
#include "implicit_partition.h"
#include "zone.h"
#include "mm_log.h"
#include "configure.h"
#include "log_object.h"
#include "domain.h"
#include "cpu_smp.h"

#ifdef OSMIPS
#include "cpu_access.h"
#endif

#define PRINT(s) log(CATEGORY_PARTITION, s)
#define PRINT_START(s) log(CATEGORY_START, s)


static struct zone freemem_zone;    /* Free logical memory descriptor zone */
static struct zone partition_zone;  /* First-fit partition descriptor zone  */


/*****************************************************************************/

struct freemem
{
   core_rbtree_node_t base_indexed;
   log_addr_t         base;
   log_addr_t         top;
};

union freemem_prealloc
{
   list_link_t     link;
   struct freemem  freemem;
};

struct firstfit_partition_class
{
   struct partition_class root;
   core_rbtree_t      base_indexed;  /* Global SAS/SASE tree */
   list_t             combined_list; /* List of instances of the class
                                      * with combined mapping type. */
   uint64_t           global_sas_top;
   uint64_t           global_mas_top;
};

struct firstfit_partition
{
   struct partition   root;

   /* These fields must only be used if the partition is non-exclusive. */
   core_rbtree_t      base_indexed;  /* Local MAS tree */
   list_link_t        combined_link; /* Link for combined instance list */
   uint64_t           mas_top;
};


/*
 * 
 * 
 */
#define NR_OF_PREALLOCATED_FREEMEM_ELEMENTS 2
static list_t freemem_prealloc_list;

static void
firstfit_partition_init(void);

void
mm_partition_init(void)
{
   U32 i;
   
   gmm_state.partition_classes = NULL;
   gmm_state.partition_instances = NULL;

   firstfit_partition_init();
   mm_implicit_partition_init();

   /* Set a higher min value on zone for logical_region
    * since this z-allocator is needed when extending all other
    * zone allocators. And in order for another z-allocator to extend
    * one struct freemem is needed.
    * To study this problem set low min values boot a system
    * and execute mm_test domain as first test case.
    */
   mm_zinit(&freemem_zone, "struct freemem",
            sizeof(union freemem_prealloc), 20, 40);

   init_list(&freemem_prealloc_list);
   for (i = 0; i < NR_OF_PREALLOCATED_FREEMEM_ELEMENTS; i++) {
      union freemem_prealloc *freemem;
      
      freemem = mm_zalloc(&freemem_zone);
      init_list_link(&freemem->link);
      insert_list_head_no_lock(&freemem_prealloc_list, &freemem->link);
   }
}

/*****************************************************************************
 *
 * Ensure that at least cnt elements can be allocated without causing zone
 * extension. If some day, the zone allocator frees unused pages, the current
 * algorithm might not work.
 */
static void pre_extend_zone(struct zone *z, U32 cnt) {
#ifdef ASSERT
  extern int mm_pending_extend;
#endif

  /* It is assumed that the element allocated by mm_zalloc(.) has room for
   * a (void*). Although we get an unusual cast to void***, i think that
   * void** is the best type for the list variable. */
  void **list = NULL;
  
  assert(mm_pending_extend == 0);

  while (cnt--) {
     void *tmp = mm_zalloc(z);
     *(void***)tmp = list;
     list = tmp;
  }
  
  while (list) {
     void *tmp = (void*)list;
     list = *list;
     mm_zfree(z, tmp);
  }
}

/*****************************************************************************
 *
 * Ensure that at least regCnt logical regions can be allocated without causing
 * zone extention. If a zone extension occurs during the restore phase it could
 * cause the restoration to fail. The reason is that there will be race for
 * breaking memory in MM between new mm heap fragments and the persistent
 * regions that should be restored.
 */

#ifdef HOST_TEST
struct zone region_zone;
struct zone logobj_zone;
#else
extern struct zone region_zone;
extern struct zone logobj_zone;
#endif

void
mm_partition_prepare_for_warm_start(U32 regCnt)
{
  pre_extend_zone(&region_zone, regCnt);
  pre_extend_zone(&logobj_zone, regCnt);
  pre_extend_zone(&freemem_zone, regCnt);
}


/*
 *===========================================================================
 * Functions for partitions.
 *===========================================================================
 */
static int
common_is_subset_of(const struct partition_class *partition,
                    log_addr_t base, log_size_t size);

static int
common_is_superset_of(const struct partition_class *partition,
                      log_addr_t base, log_size_t size);

static int
common_is_subset_of_phys_mem(const struct partition_class *partition);

static int
common_overlap(const struct partition_class *partition,
               log_addr_t base, log_size_t size);

static int
common_get_bank_info(const struct partition_class *p_class,
                     int index, log_addr_t *base, log_size_t *size);

static struct partition_class *
firstfit_init(uint32_t options);

static struct partition_bank *
firstfit_init_bank(struct partition_class *p_class,
                   log_addr_t base, log_size_t size);

static struct partition *
firstfit_create(struct partition_class *p_class);

static void
firstfit_destroy(struct partition *partition);

static mm_status_t
firstfit_alloc_regular(struct partition  *partition,
                      struct log_object *log_obj,
                      log_size_t size, log_size_t alignment, uint32_t options);

static mm_status_t
firstfit_alloc_fixed(struct partition  *partition,
                    struct log_object *log_obj,
                    log_size_t size, log_addr_t fixed, uint32_t options);

static void
firstfit_free(struct partition *partition,
             struct log_object *log_obj, uint32_t options);

static uint64_t 
firstfit_global_mas_top(const struct firstfit_partition_class *part_class);

static uint64_t 
firstfit_local_mas_top(const struct firstfit_partition *part,
                       core_rbtree_t *tree,
                       struct freemem   *free);

static uint64_t 
firstfit_local_sas_top(const struct firstfit_partition *part,
                       core_rbtree_t *tree,
                       struct freemem *free);

/*****************************************************************************/


/*****************************************************************************
 * See partition.h
 */
static int
common_is_superset_of(const struct partition_class *partition,
                      log_addr_t base, log_size_t size)
{
   struct partition_bank *bank;
   log_addr_t top = base + (size - 1);

   assert(partition != NULL);
   assert(size > 0);
   
   /* Check wrap */
   if (base > top)
      return 0;

   bank = partition->banks;
   while (bank != NULL)
   {
      /* Check if base is inside the bank. */
      if (base >= bank->base && base <= bank->top)
      {
         struct partition_bank *bank_next = bank->next;
         struct partition_bank *bank_prev = bank;
                                                          
         /* If top is less than bank top,
            the range is a subset of a single bank. */
         if (top <= bank->top)
            return 1;

         /* The range might also be a subset of several
          * consecutive banks. Check if top is part of the next bank. */
         while (bank_next != NULL && bank_prev->top + 1 == bank_next->base)
         {
            if (top <= bank_next->top)
               return 1;

            /* Still not enough... */
            bank_prev = bank_next;
            bank_next = bank_next->next;
         }

         return 0;
      }

      bank = bank->next;
   }

   return 0;
}

/*****************************************************************************
 * See partition.h
 */
static int
common_is_subset_of(const struct partition_class *partition,
                    log_addr_t base, log_size_t size)
{
   struct partition_bank *first_bank, *last_bank;
   log_addr_t top = base + (size - 1);

   assert(partition != NULL);
   assert(size > 0);

   /* Check wrap */
   if (top < base)
      return 0;

   first_bank = partition->banks;
   if (first_bank == NULL)
      return 0;

   last_bank  = first_bank;
   while (last_bank->next != NULL)
      last_bank = last_bank->next;
   
   if (base <= first_bank->base && top >= last_bank->top)
      return 1;
      
   return 0;
}

/*****************************************************************************
 * See partition.h
 */
static int
common_is_subset_of_phys_mem(const struct partition_class *partition)
{
   struct partition_bank *bank;
   struct allocator *pa;

   assert(partition != NULL);
   
   bank = partition->banks;

   while (bank != NULL)
   {
      phys_addr_t base = bank->base;
      phys_size_t size = (bank->top + 1) - bank->base;
      int found = 0;

      for (pa = gmm_state.allocators; NULL != pa; pa = pa->next)
      {
         if (pa->overlap2(pa, &base, &size))
         {
            /* Physical allocator pa has a bank that overlaps the partition's
             * range. Note that base & size is both in and out parameters, see
             * partition.h. The variables base & size have been updated with
             * values for the overlapping bank.
             */

            if (base > bank->base)
            {
               /* Partition's base is not covered by the bank. */
               return 0;
            }

            if (base + size <= bank->top)
            {
               /* Partition's top is not covered by the bank. */
               return 0;
            }

            found = 1;
            break;
         }
      }


      if (!found)
      {
         return 0;
      }

      bank = bank->next;
   }

   /* Entire log_mem overlaps phys_mem */
   return 1;
}


/*****************************************************************************
 * See partition.h
 */
static int
common_overlap(const struct partition_class *partition,
               log_addr_t base, log_size_t size)
{
   struct partition_bank *bank;

   assert(partition != NULL);
   assert(size > 0);

   bank = partition->banks;
   while (bank != NULL)
   {
      log_addr_t top = base + (size - 1);

      if (bank->base <= top && base <= bank->top)
      {
         return 1;
      }

      bank = bank->next;
   }

   return 0;
}

/*****************************************************************************
 * See partition.h
 */
static int
find_gap_outside_phys_mem(log_addr_t *base_,
                          log_size_t size,
                          log_addr_t max_top)
{

   log_addr_t base = *base_;
   log_addr_t top;
   
  scan_start:
   top = base + size-1;

   if (top > max_top || top < base) /* wrap */
   {
      return 0;
   }
   
   {
      struct allocator *pa;
      for (pa = gmm_state.allocators; NULL != pa; pa = pa->next)
      {
         phys_addr_t bank_base = base;
         phys_size_t bank_size = size;
         
         if (pa->overlap2(pa, &bank_base, &bank_size))
         {
            base = (log_addr_t)(bank_base+bank_size);

            if (base < bank_base) /*wrap */
            {
               /* We have hit a bank that contains memory above 4G. */
               return 0;
            }
            goto scan_start; /*lint !e801 */
         }
      }
   }
   
   *base_ = base;
   return 1;
}

/*****************************************************************************
 * See partition.h
 */
static int
common_get_bank_info(const struct partition_class *p_class,
                     int idx, log_addr_t *base, log_size_t *size)
{
   struct partition_bank *bank;
   unsigned int i;

   assert(p_class != NULL);

   for (i = 0, bank = p_class->banks;
        i < (unsigned int)idx && bank != NULL;
        i++, bank = bank->next)
   {
   }

   if (bank == NULL)
   {
      return 0;
   }

   if (base != NULL)
   {
      *base = bank->base;
   }

   if (size != NULL)
   {
      *size = (bank->top + 1) - bank->base;
   }

   return 1;
}

/****************************************************************************/

static int
firstfit_base_compare(const void *_a, const void *_b)
{
   const struct freemem *left = (const struct freemem *)_a;
   const struct freemem *right = (const struct freemem *)_b;

   /* Index by base address only */
   if (left->base < right->base)
   {
      return -1;
   }
   else if (left->base > right->base)
   {
      return 1;
   }
   
   return 0;
}

/****************************************************************************/

static int
firstfit_base_find(const void *_key, const void *_current)
{
   const log_addr_t *base = (const log_addr_t *)_key;
   const struct freemem *current = (const struct freemem*)_current;

   if (*base < current->base)
   {
      return -1;
   }
   else if (*base > current->top)
   {
      return 1;
   }

   return 0;
}

/****************************************************************************/

/* Compare function to find greater or equal */
static int
firstfit_base_find_ge(const void *_key, const void *_current, const void *best)
{
   const log_addr_t *base = (const log_addr_t *)_key;
   const struct freemem *current = (const struct freemem*)_current;

   (void)best;    /* Avoid unused parameter warning */

   if (*base < current->base)
   {
      /* Current is greater, acceptable but go left to try find better fit. */
      return -2;
   }
   else if (*base > current->top)
   {
      /* current is less, not acceptable, go right */
      return 1;
   }

   return 0;      /* match. */
}

/****************************************************************************/

static uint64_t 
firstfit_global_mas_top(const struct firstfit_partition_class *part_class)
{
   list_link_t *link;
   uint64_t top = 0;

   for (link = list_head(&part_class->combined_list);
        link != NULL;
        link = list_next(&part_class->combined_list, link))
   {
      struct firstfit_partition *part;

      part = (struct firstfit_partition*)
         ((log_addr_t)link-offsetof(struct firstfit_partition,combined_link));

      if (part->mas_top > top)
         top = part->mas_top;
   }
   
   return top;
}

/****************************************************************************/

static uint64_t 
firstfit_local_mas_top(const struct firstfit_partition *part,
                       core_rbtree_t *tree,
                       struct freemem *free)
{
   struct partition_bank *bank;
   struct freemem *prev_free = (struct freemem *)
      core_rbtree_prev_no_lock(tree, &free->base_indexed);

   for (bank = part->root.class->banks;
        bank != NULL;
        bank = bank->next)
   {
      if (bank->base == free->base)
      {
         if (prev_free != NULL)
         {
            free = prev_free;
         }
         break;
      }
   }

   return free->base;
}

/****************************************************************************/

static uint64_t 
firstfit_local_sas_top(const struct firstfit_partition *part,
                       core_rbtree_t *tree,
                       struct freemem *free)
{
   struct partition_bank *bank;
   struct freemem *next_free = (struct freemem *)
      core_rbtree_next_no_lock(tree, &free->base_indexed);

   for (bank = part->root.class->banks;
        bank != NULL;
        bank = bank->next)
   {
      if (bank->top == free->top)
      {
         if (next_free != NULL)
         {
            free = next_free;
         }
         break;
      }
   }

   return (uint64_t)free->top + 1;
}

/*****************************************************************************/

static void
firstfit_update_top_limits(struct partition *p,
                           const struct log_object *mem)
{
   struct firstfit_partition *part = (struct firstfit_partition *)p;

   assert(part != NULL);
   assert(mem  != NULL);
   
   if (partition_is_combined(p->class))
   {
      struct firstfit_partition_class *part_class =
         (struct firstfit_partition_class*)p->class;

      if (mem->options & MM_PARTITION_OPT_EXCLUSIVE)
      {
         if (mem->base < part_class->global_sas_top)
         {
            /* Breaking new grounds. Update top */
            part_class->global_sas_top = mem->base;
         }
      }
      else
      {
         if (mem->base + mem->size > part->mas_top)
         {
            /* Breaking new grounds. Update top */
            part->mas_top = mem->base + mem->size;
         }
         
         if (part->mas_top > part_class->global_mas_top)
         {
            /* Breaking new grounds. Update top */
            part_class->global_mas_top = part->mas_top;
         }
      }
   }
}

/*****************************************************************************/

static struct freemem *
freemem_get(list_t *list)
{
   union freemem_prealloc *pre = NULL;

   if (list == NULL)
   {
      pre = mm_zalloc(&freemem_zone);
   }
   else
   {
      pre = (union freemem_prealloc *)list_head(list);
      if (pre != NULL)
      {
         delete_from_list_no_lock(&pre->link);
      }
   }

   return pre != NULL ? &pre->freemem : NULL;
}

static int
freemem_put(struct freemem *free, list_t *list)
{
   union freemem_prealloc *pre = (union freemem_prealloc*)free;

   if (free == NULL)
      return 0;

   if (list == NULL)
   {
      mm_zfree(&freemem_zone, &pre->link);
   }
   else
   {
      insert_list_head_no_lock(list, &pre->link);
   }
   
   return 1;
}


/*****************************************************************************
 * See partition.h
 */
static struct partition_class *
firstfit_init(uint32_t options)
{
   struct firstfit_partition_class *p_class = NULL;

   /* Extra caution. mm_alloc_boot_space might not update <p_class>. */
   if (!mm_alloc_boot_space(sizeof(struct firstfit_partition_class), (void*)&p_class))
   {
      MM_FATAL_ERROR2(("Boot heap exhausted.\n"), 0, 0);
   }

   /* Initialise it. */
   (void)memset(p_class, 0, sizeof(struct firstfit_partition_class));

   /* Choose allocator implementation */
   p_class->root.init                  = firstfit_init;
   p_class->root.init_bank             = firstfit_init_bank;
   p_class->root.create                = firstfit_create;
   p_class->root.destroy               = firstfit_destroy;
   p_class->root.is_subset_of          = common_is_subset_of;
   p_class->root.is_superset_of        = common_is_superset_of;
   p_class->root.is_subset_of_phys_mem = common_is_subset_of_phys_mem;
   p_class->root.overlap               = common_overlap;
   p_class->root.get_bank_info         = common_get_bank_info;
   p_class->root.alloc                 = firstfit_alloc_regular;
   p_class->root.alloc_f               = firstfit_alloc_fixed;
   p_class->root.free                  = firstfit_free;
   
   p_class->root.options = options;
   p_class->global_mas_top = ~0;
   p_class->global_sas_top =  0;

   core_rbtree_init(&p_class->base_indexed, firstfit_base_compare);
   init_list(&p_class->combined_list);

   return &p_class->root;
}

/*****************************************************************************
 * See partition.h
 */
static struct partition_bank *
firstfit_init_bank(struct partition_class *p,
                   log_addr_t base, log_size_t size)
{
   struct firstfit_partition_class *part_class =
      (struct firstfit_partition_class*)p;
   struct partition_bank *bank = NULL;

   /* Extra caution. mm_alloc_boot_space might not update <bank>. */
   if (!mm_alloc_boot_space(sizeof(struct partition_bank), (void*)&bank))
   {
      MM_FATAL_ERROR2(("Boot heap exhausted.\n"), 0, 0);
   }

   bank->base = base;
   bank->top  = base + (size - 1);

   if (partition_is_exclusive(p))
   {
      struct log_object lobj;
      struct firstfit_partition part;

      part.root.class = &part_class->root;

      /* New memory is added to the global tree by freeing each bank's
       * memory into the allocator. That way, consecutive banks will
       * have their memory merged into a single larger fragment
       * (as opposed to several smaller ones). */
      
      lobj.freemem = freemem_get(NULL);
      if (lobj.freemem == NULL)
      {
         MM_FATAL_ERROR2(("Failed to attach logical bank to partition.\n"),
                         0, 0);
      }

      lobj.base    = base;
      lobj.size    = size;
      lobj.options = MM_PARTITION_OPT_EXCLUSIVE|MM_PARTITION_OPT_VIRTUAL;

      part_class->root.free(&part.root, &lobj, lobj.options);
   }
   
   if (bank->top > part_class->global_sas_top)
   {
      part_class->global_sas_top = bank->top + 1;
   }

   if (bank->base < part_class->global_mas_top)
   {
      part_class->global_mas_top = bank->base;
   }

   return bank;
}

/*****************************************************************************
 * See partition.h
 *
 * Instantiates a partition class.
 */
static struct partition *
firstfit_create(struct partition_class *p)
{
   struct firstfit_partition *part;
   struct firstfit_partition_class *part_class =
      (struct firstfit_partition_class*)p;

   /* A partition class must have at least one bank and one alias
    * when it is instantiated. There is one exception though: log_mem:s
    * configured automatically due to static regions. Such partition classes
    * are created after the system domain, which means that they have missed
    * the only opportunity to get instantiated (global partitions are only
    * instantiated once). We therefore instantiate all classes directly when
    * they are created if the system domain is present. Autoconfigured classes
    * are never non-exclusive which means that we can assume that it is safe
    * to ignore initialization of free local memory. Global free memory is
    * added to the partition instance when banks are attached.
    *
    * The assert:s below should detect most illegal attempts to instantiate
    * bank-less non-exclusive partition classes. 
    */
   assert(p != NULL);
   assert(p->banks   != NULL || !partition_is_non_exclusive(p));
   assert(p->aliases != NULL || !partition_is_non_exclusive(p));


   part = mm_zalloc(&partition_zone);
   if (part == NULL)
   {
      return NULL;
   }

   part->root.next  = NULL;
   part->root.class = p;

   /* If the partition supports non-exclusive allocations add free
    * memory to the local trees of the new partition instance. */ 
   if (partition_is_non_exclusive(p))
   {
      struct partition_bank *bank;

      core_rbtree_init(&part->base_indexed, firstfit_base_compare);

      for (bank = p->banks; bank != NULL; bank = bank->next)
      {
         struct log_object lobj;

         /* New memory is added to the local tree by freeing each bank's
          * memory into the allocator. That way, consecutive banks will
          * have their memory merged into a single larger fragment
          * (as opposed to several smaller ones). */
         lobj.freemem = freemem_get(NULL);
         if (lobj.freemem == NULL)
         {
            struct freemem *freemem = (struct freemem*)
               core_rbtree_first_no_lock(&part->base_indexed);
            while (freemem != NULL)
            {
               (void)core_rbtree_remove_no_lock(
                  &part->base_indexed, &freemem->base_indexed);
               freemem_put(freemem, NULL);
               freemem = (struct freemem*)
                  core_rbtree_first_no_lock(&part->base_indexed);
            }

            mm_zfree(&partition_zone, part);
            return NULL;
         }

         lobj.base    = bank->base;
         lobj.size    = (bank->top + 1) - bank->base;
         lobj.options = MM_PARTITION_OPT_MAS;

         part_class->root.free(&part->root, &lobj, lobj.options);
      }
   }

   if (partition_is_combined(p))
   {
      init_list_link(&part->combined_link);
      insert_list_head_no_lock(&part_class->combined_list,
                               &part->combined_link);
      part->mas_top = p->banks->base;
   }
   
   return &part->root;
}


/*****************************************************************************
 * See partition.h
 */
static void
firstfit_destroy(struct partition *p)
{
   struct firstfit_partition *part = (struct firstfit_partition *)p;
   core_rbtree_t *base_indexed_tree;
   struct freemem *free;

   /* Exclusive-only partitions should never be destroyed. */
   assert(part != NULL);
   assert(part->root.class != NULL);
   assert(partition_is_non_exclusive(part->root.class));
   
   /*
    * Start with a cleanup of the memory in the local,
    * non-exclusive tree. 
    * 
    * If the partition is both exclusive and non-exclusive
    * we don't want to destroy the data representing the
    * exclusive address space. This is ok since the global
    * tree data always is referenced by the system domain and will
    * be reused if another domain is created. 
    */
   base_indexed_tree = &part->base_indexed;

   free = (struct freemem*)
      core_rbtree_first_no_lock(base_indexed_tree);
   while (free != NULL)
   {
      (void)core_rbtree_remove_no_lock(
         base_indexed_tree, &free->base_indexed);

      mm_zfree(&freemem_zone, free);

      free = (struct freemem*)
         core_rbtree_first_no_lock(base_indexed_tree);
   }

   if (partition_is_combined(p->class))
   {
      struct firstfit_partition_class *part_class =
         (struct firstfit_partition_class *)p->class;

      delete_from_list_no_lock(&part->combined_link); 

      if (part->mas_top == part_class->global_mas_top)
      {
         part_class->global_mas_top = firstfit_global_mas_top(part_class);
      }
   }

   mm_zfree(&partition_zone, part);
}

/*****************************************************************************/
static mm_status_t
firstfit_extract(struct partition  *partition,
                 core_rbtree_t     *base_indexed_tree,
                 struct freemem    *free,
                 struct log_object *log_obj,
                 log_addr_t fixed, log_size_t size, uint32_t options)
{
   struct firstfit_partition_class *part_class;

   assert(partition != NULL);
   assert(base_indexed_tree != NULL);
   assert(free != NULL);
   assert(log_obj != NULL);
   assert(log_obj->freemem == NULL);

   part_class = (struct firstfit_partition_class*)partition->class;
   assert(part_class != NULL);

   
   /* Memory found and ready for allocation. If the partition is a combined
    * SAS and MAS partition, make sure all top borders are enforced, i.e.
    * don't allow a MAS allocation above the SAS top limit or a SAS allocation
    * below the MAS top limit.
    *
    * The limits are updated first when the allocation has been successfully
    * carried out. 
    */
   if (partition_is_combined(partition->class))
   {
      if (options & MM_PARTITION_OPT_EXCLUSIVE)
      {
         if (fixed < part_class->global_mas_top)
         {
            log_obj->base = 0;
            log_obj->size = 0;
            return MM_EOUT_OF_LOG_MEM;
         }
      }
      else
      {
         if (fixed + size > part_class->global_sas_top)
         {
            log_obj->base = 0;
            log_obj->size = 0;
            return MM_EOUT_OF_LOG_MEM;
         }
      }
   }

   /* Mark logical object as resident if partition is.
    * Flag is used to determin resident status of a region. */
   if (partition_is_resident(partition->class))
   {
      options |= MM_PARTITION_OPT_RESIDENT;
   }

   /* Check if free area is exact match */
   if (free->base == fixed && ((free->top + 1) - free->base) == size)
   {
      log_obj->base = free->base;
      log_obj->size = (free->top + 1) - free->base;
      log_obj->options = options;
      log_obj->freemem = free;

      (void)core_rbtree_remove_no_lock(
         base_indexed_tree, &free->base_indexed);

      firstfit_update_top_limits(partition, log_obj);

      return MM_SUCCESS;
   }
   
   /* Check if free area needs splitting */
   if (free->base < fixed && (fixed + (size - 1)) < free->top)
   {
      struct freemem *split;

      log_obj->base = fixed;
      log_obj->size = size;
      log_obj->options = options;

      /* Use a spare freemem-object from the list of pre-allocated objects
       * to describe the new part of the area to be split. The existing area
       * is temporarily removed from the tree and if it was the only one
       * available a zone extension would fail. */
      split = freemem_get(&freemem_prealloc_list);
      assert(split != NULL);
      
      split->base = fixed + size;
      split->top  = free->top;
      free->top   = fixed - 1;

      {
         int status = core_rbtree_insert_no_lock(
            base_indexed_tree, &split->base_indexed);
         (void)status;
         assert(status == 0);
      }

      /* At least two logical areas in the tree. It is safe to restore
       * the list of pre-allocated freemem-objects with a zone allocation. */
      if (!freemem_put(freemem_get(NULL), &freemem_prealloc_list))
      {
         /* Out of physical memory. Do a complete rollback. */
         (void)core_rbtree_remove_no_lock(
            base_indexed_tree, &split->base_indexed);
         free->top = split->top;
         (void)freemem_put(split, &freemem_prealloc_list);
         log_obj->base = 0;
         log_obj->size = 0;
         return MM_EOUT_OF_PHYS_MEM;
      }

      log_obj->freemem = freemem_get(NULL);
      if (log_obj->freemem == NULL)
      {
         /* Out of physical memory. Do a complete rollback. */
         (void)core_rbtree_remove_no_lock(
            base_indexed_tree, &split->base_indexed);
         free->top = split->top;
         (void)freemem_put(split, NULL);
         log_obj->base = 0;
         log_obj->size = 0;
         return MM_EOUT_OF_PHYS_MEM;
      }

      firstfit_update_top_limits(partition, log_obj);
      
      return MM_SUCCESS;
   }

   /* The free area only requires adjustments in either of its ends */
   log_obj->base = fixed;
   log_obj->size = size;
   log_obj->options = options;

   if (free->base == fixed)
      free->base += size;
   else /* free->base + free->size == fixed + size */
      free->top -= size;

   log_obj->freemem = freemem_get(NULL);
   if (log_obj->freemem == NULL)
   {
      if (fixed + size == free->base)
         free->base -= size;
      else
         free->top += size;

      log_obj->base = 0;
      log_obj->size = 0;
      return MM_EOUT_OF_PHYS_MEM;
   }

   firstfit_update_top_limits(partition, log_obj);

   return MM_SUCCESS;
}


/*****************************************************************************
 * See partition.h
 */
static mm_status_t
firstfit_alloc_regular(struct partition  *p,
                       struct log_object *log_obj,
                       log_size_t size, log_size_t alignment, uint32_t options)
{
   struct firstfit_partition *part = (struct firstfit_partition *)p;
   core_rbtree_t *base_indexed_tree;
   struct freemem *free;
   log_addr_t base = 0;
   mm_status_t status;
   int direction = 0; /* From bottom */
   Msr sr = 0;

   assert(part != NULL);
   assert(log_obj != NULL);
   assert(size > 0);
   assert(mm_log_aligned(size,      hal_get_info()->page_sizes[0]));
   assert(mm_log_aligned(alignment, hal_get_info()->page_sizes[0]));

   if (options & MM_PARTITION_OPT_EXCLUSIVE)
   {
      struct firstfit_partition_class *part_class =
         (struct firstfit_partition_class*)p->class;
      base_indexed_tree = &part_class->base_indexed;

      if (partition_is_combined(p->class))
      {
         direction = 1; /* From top */
      }
   }
   else
   {
      base_indexed_tree = &part->base_indexed;
   }

   
   if (alignment == 0)
   {
      alignment = hal_get_info()->page_sizes[0];
   }
   
   MM_MULTITHREAD_LOCK(sr);

   /* Locate a free chunk of logical address space in a
    * firstfit manner, either from the bottom or the top
    * depending on type of memory allocated.  */
   if (direction == 0) /* Direction = From bottom */
   {
      free = (struct freemem *)
         core_rbtree_first_no_lock(base_indexed_tree);

      /* A logical area large enough was found. Make sure it fulfills
       * the alignment requirements. If not, try the next free fragment. */
      while (free)
      {
         base = free->base;
         base = mm_log_align(base, alignment);

         if (base >= free->base &&
             (base + size - 1) <= free->top &&
             base < (base + size - 1))
         {
            break;
         }

         free = (struct freemem *)
            core_rbtree_next_no_lock(base_indexed_tree, &free->base_indexed);
      }
   }
   else /* Direction = From top */
   {
      free = (struct freemem *)
         core_rbtree_last_no_lock(base_indexed_tree);

      /* A logical area large enough was found. Make sure it fulfills
       * the alignment requirements. If not, move to the next larger size
       * and try again. */
      while (free)
      {
         base = (free->top + 1) - size;

         if (!mm_log_aligned(base, alignment))
         {
            base = mm_log_align(base, alignment);
            base -= alignment;
         }


         if (base >= free->base &&
             (base + size - 1) <= free->top &&
             base < (base + size - 1))
         {
            break;
         }

         free = (struct freemem *)
            core_rbtree_prev_no_lock(base_indexed_tree, &free->base_indexed);
      }
   }
      

   if (free == NULL)
   {
      MM_MULTITHREAD_UNLOCK(sr);
      log_obj->base = 0;
      log_obj->size = 0;
      return MM_EOUT_OF_LOG_MEM;
   }

   /* We now know that there is a piece of memory available at <base>
    * that suits the request. Now extract it and hand it over to the caller. */
   status = firstfit_extract(p, base_indexed_tree,
                             free, log_obj, base, size, options);
   MM_MULTITHREAD_UNLOCK(sr);

   return status;
}


/*****************************************************************************
 * See partition.h
 */
static mm_status_t
firstfit_alloc_fixed(struct partition  *p,
                     struct log_object *log_obj,
                     log_size_t size,
                     log_addr_t request_addr,
                     uint32_t options)
{
   struct firstfit_partition *part = (struct firstfit_partition *)p;
   core_rbtree_t *base_indexed_tree;
   struct freemem *free_right;
   struct freemem *free_left;
   mm_status_t status;
   Msr sr = 0;

   assert(part != NULL);
   assert(log_obj != NULL);
   assert(size > 0);
   assert(mm_log_aligned(size,  hal_get_info()->page_sizes[0]));
   assert(mm_log_aligned(request_addr, hal_get_info()->page_sizes[0]));

   if (options & MM_PARTITION_OPT_EXCLUSIVE)
   {
      struct firstfit_partition_class *part_class =
         (struct firstfit_partition_class*)p->class;
      base_indexed_tree = &part_class->base_indexed;
   }
   else
   {
      base_indexed_tree = &part->base_indexed;
   }

   if (!common_is_superset_of(p->class, request_addr, size))
   {
      log_obj->base = 0;
      log_obj->size = 0;
      return MM_EOUTSIDE_LOG_MEM;
   }

   MM_MULTITHREAD_LOCK(sr);
   
   free_right = free_left = (struct freemem *)core_rbtree_find_best_no_lock(
      base_indexed_tree, firstfit_base_find_ge, &request_addr);

   /* Check if the found free_right range overlaps the request address */
   if ((free_right != NULL) && (free_right->base <= request_addr))
   {
      if ((request_addr + size - 1) <= free_right->top)
      {
         goto found_free;  /* Exact match */
      }
      if ((options & MM_PARTITION_OPT_SEARCH_REVERSE) &&
          ((free_right->top - free_right->base + 1) >= size))
      {
         request_addr = free_right->top - size + 1;
         goto found_free;  /* Overlaping range that extends to left. */
      }
      if (options & MM_PARTITION_OPT_SEARCH_FORWARD)
      {
         /* We are about to search right. This element already examined. */
         goto next_forward;
      }
   }

   /* Search to the right until we find a range that is large enough. */
   if (!(options & MM_PARTITION_OPT_SEARCH_FORWARD))
   {
      free_right = NULL;
   }
   else
   {
      while (free_right != NULL)
      {
         if ((free_right->top - free_right->base + 1) >= size)
         {
            /* Found range that is big enough */
            break;
         }
       next_forward:
         free_right = (struct freemem *)
            core_rbtree_next_no_lock(base_indexed_tree,
                                     &free_right->base_indexed);
      }
   }

   /* Search to the left until we find a range that is large enough. */
   if (!(options & MM_PARTITION_OPT_SEARCH_REVERSE))
   {
      free_left = NULL;
   }
   else
   {
      if (free_left == NULL)
      {
         /* There is nothing right of request_addr; start at right edge. */
         free_left = (struct freemem *)
                  core_rbtree_last_no_lock(base_indexed_tree);
      }
      else
      {
         /* free_left is to the right or overlapping. Start by going left. */
         goto next_reverse;
      }

      while (free_left != NULL)
      {
         if ((free_left->top - free_left->base + 1) >= size)
         {
            /* Found range that is big enough */
            break;
         }
       next_reverse:
         free_left = (struct freemem *)
            core_rbtree_prev_no_lock(base_indexed_tree,
                                     &free_left->base_indexed);
      }
   }

   if ((free_left != NULL) && (free_right != NULL))
   {
      /* Both free_left and free_right are valid. Select the closest */
      if ((request_addr - free_left->top) < (free_right->base - request_addr))
      {
         free_right = NULL;   /* free_left is closest */
      }
      else
      {
         free_left = NULL;    /* free_right is closest */
      }
   }

   if (free_left != NULL)
   {
      /* We use free_left. Adjust request_addr to the right part of it. */
      request_addr = free_left->top - size + 1;
      free_right = free_left;
   }
   else if (free_right != NULL)
   {
      /* We use free_right Adjust request_addr to the left part of it. */
      request_addr = free_right->base;
   }
   else
   {
      /* Nothing found. */
      log_obj->base = 0;
      log_obj->size = 0;
      status = MM_EOUT_OF_LOG_MEM;
      goto done;
   }

 found_free:
   status = firstfit_extract(p, base_indexed_tree,
                             free_right, log_obj, request_addr, size,
                             options & ~MM_PARTITION_OPT_SEARCH);

 done:
   if ((status == MM_EOUT_OF_LOG_MEM) &&
       ((options & MM_PARTITION_OPT_SEARCH) == 0))
   {
      /* We are not allowed to scan - convert the status code. */
      status = MM_EOVERLAPPING_ADDRESS;
   }

   MM_MULTITHREAD_UNLOCK(sr);
   return status;
}


/*****************************************************************************
 * See partition.h
 */
static void
firstfit_free(struct partition *p,
              struct log_object *log_obj, uint32_t options)
{
   struct firstfit_partition *part = (struct firstfit_partition *)p;
   core_rbtree_t *base_indexed_tree;
   struct freemem *node = NULL;
   log_addr_t base;
   Msr sr = 0;

   assert(part    != NULL);
   assert(log_obj != NULL);
   assert(log_obj->freemem != NULL);

   (void)options;
   
   if (log_obj->options & MM_PARTITION_OPT_EXCLUSIVE)
   {
      struct firstfit_partition_class *part_class =
         (struct firstfit_partition_class*)p->class;
      base_indexed_tree = &part_class->base_indexed;
   }
   else
   {
      base_indexed_tree = &part->base_indexed;
   }

   MM_MULTITHREAD_LOCK(sr);

   /* Check if it is possible to merge with left adjacent free area */
   if (log_obj->base != 0)
   {
      struct freemem *temp;

      base = log_obj->base - 1;

      temp = (struct freemem *)core_rbtree_find_no_lock(
         base_indexed_tree, firstfit_base_find, &base);
      if (temp != NULL)
      {
         temp->top += log_obj->size;
         node = temp;
      }
   }

   /* Check if it is possible to merge with right adjacent free area */
   if (log_obj->base + log_obj->size != 0)
   {
      struct freemem *temp;

      base = log_obj->base + log_obj->size;

      temp = (struct freemem *)core_rbtree_find_no_lock(
         base_indexed_tree, firstfit_base_find, &base);
      if (node != NULL && temp != NULL)
      {
         /* We're filling a hole completely. */
         (void)core_rbtree_remove_no_lock(
            base_indexed_tree, &temp->base_indexed);
         node->top = temp->top;
         (void)freemem_put(temp, NULL);
      }
      else if (temp != NULL)
      { 
         temp->base -= log_obj->size;
         node = temp;
      }
   }

   /* No merge possible, add a completely new fragment. */
   if (node == NULL)
   {
      int status = 0;
      (void)status;
      
      /* Using zalloc here to allocate a freemem-object could cause a
       * zone extension. We use a spare instead to free the logical
       * memory first, and restore the spare list when done. */
      node = log_obj->freemem;
      log_obj->freemem = NULL;

      node->base = log_obj->base;
      node->top  = log_obj->base + (log_obj->size - 1);

      status = core_rbtree_insert_no_lock(
         base_indexed_tree, &node->base_indexed);
      assert(status == 0);
   }
   else
   {
      int status = 0;
      (void)status;
      
      /* The freed logical memory did not require a new freemem-object
       * since it was merged with an adjacent piece of memory. The spare
       * freemem-object created earlier during the allocation is no
       * longer needed. */
      status = freemem_put(log_obj->freemem, NULL);
      assert(status != 0);
      log_obj->freemem = NULL;
   }

   /* Memory has been freed. If the partition is a combined
    * SAS and MAS partition, make sure it all top borders are updated
    * accordingly to re-enable allocation of the opposite mapping type. */
   if (partition_is_combined(p->class))
   {
      struct firstfit_partition_class *part_class =
         (struct firstfit_partition_class*)p->class;

      if (log_obj->options & MM_PARTITION_OPT_EXCLUSIVE)
      {
         if (log_obj->base == part_class->global_sas_top)
         {
            part_class->global_sas_top =
               firstfit_local_sas_top(part, base_indexed_tree, node);
         }
      }
      else
      {
         if (log_obj->base + log_obj->size == part->mas_top)
         {
            part->mas_top =
               firstfit_local_mas_top(part, base_indexed_tree, node);
         }

         if (log_obj->base + log_obj->size == part_class->global_mas_top)
         {
            part_class->global_mas_top = firstfit_global_mas_top(part_class);
         }
      }
   }

   log_obj->base = 0;
   log_obj->size = 0;

   MM_MULTITHREAD_UNLOCK(sr);
}

/*****************************************************************************/

static void
firstfit_partition_init(void)
{
   mm_zinit(&partition_zone, "struct firstfit_partition",
            sizeof(struct firstfit_partition), 10, 20);
}

/*****************************************************************************/

static struct partition_bank *
get_next_bank_above(log_addr_t base)
{
   struct partition_class *part = gmm_state.partition_classes;
   struct partition_bank  *bank = NULL, *best_bank = NULL;

   if (part == NULL)
   {
      return NULL;
   }

   /*
    * All partitions must have at least one bank for this to work.
    * (Except implicit partitions)
    */
   
   while (part != NULL && partition_is_implicit(part))
   {
      /* Skip implicit partitions without banks */
      part = part->next;
   }

   if (part == NULL)
      return NULL;
   
   bank = part->banks;
   while (bank != NULL)
   {
      if (bank->base >= base)
      {
         if (best_bank == NULL || best_bank->base > bank->base)
         {
            best_bank = bank;
         }
      }

      bank = bank->next;

      if (bank == NULL)
      {
         part = part->next;
         while (part != NULL && partition_is_implicit(part))
         {
            /* Skip implicit partitions without banks */
            part = part->next;
         }

         if (part == NULL)
         {
            break;
         }

         bank = part->banks;
      }
   }

   return best_bank;
}

struct partition_class *
mm_create_hal_partition(const char *name,
                        const log_size_t size,
			int   implicit)
{
   log_addr_t min_base = 0;
   log_addr_t max_top  = (log_addr_t)-1;
   log_addr_t prev_top;
   struct partition_class *part = gmm_state.partition_classes;
   struct partition_bank  *bank = NULL;
   log_addr_t base;
   
   assert(name != NULL);

#if defined OSMIPS
   min_base = KSEG3_BASE;
   max_top  = KSEG3_BASE + KSEG3_SIZE-1;
#else
   min_base = 0;
   max_top  = (log_addr_t)-1;
#endif

   prev_top = min_base - 1;

   if (part == NULL)
   {
      goto skip_phys_mem;
   }
   
   bank = get_next_bank_above((log_addr_t)0);
  
   while (bank)
   {
      log_addr_t bank_base;
      log_addr_t bank_top;

      bank_base = bank->base;
      bank_top  = bank->top;

      if (bank_top < min_base) {
         bank = get_next_bank_above(bank->base + 1);
         continue;
      }

      if (bank_base > min_base) {
         prev_top = min_base - 1;
      }
      else
      {
         bank = get_next_bank_above(bank->base + 1);
         prev_top = bank_top;
      }
      break;
   }
   
   while (bank) {
      /* Current partition. */
      log_addr_t bank_base;
      log_addr_t bank_top;
      
      /* Gap before current partition. */
      log_addr_t gap_base;
      log_size_t gap_size;
      
      bank_base = bank->base;
      bank_top  = bank->top;
      
      if (bank_base > max_top)
      {
         break;
      }

      if (bank_base <= min_base)
      {
         prev_top = bank_top;
         continue;
      }
      
      gap_base = prev_top + 1;
      gap_size = bank_base - gap_base;
      base = gap_base;
      
      if ((implicit && gap_size > 0 &&
	   find_gap_outside_phys_mem(&base, size, bank_base-1)) ||
	  (!implicit && gap_size >= size))
      {
         struct partition_class *hal_class;
         mm_status_t status;

         hal_class = mm_add_partition(MM_PARTITION_OPT_SAS);
         if (hal_class == NULL)
         {
            return NULL;
         }

         status = mm_add_partition_alias(hal_class, name, MM_PARTITION_OPT_SAS);
         if (status != MM_SUCCESS)
         {
            return NULL;
         }

         /* FIXME, with many small regions on a board equipped with large
          * RAM mempory. One can run out of HAL_LOG_MEM due to zone allocator
          * extensions. (mpc8641-8G run mm_test -cb and thereafter 
          * mm_test check_create_region_perf_many_domains.)
          * One improvement could be to to allocate largest gap for HAL_LOG_MEM
          * with size >= 32 MB, instead always limit to 32 MB.
          */ 
         status = mm_add_partition_bank(hal_class, base, size);
         if (status != MM_SUCCESS)
         {
            return NULL;
         }

         return hal_class;
      }

      prev_top = bank_top;
      bank = get_next_bank_above(bank->base + 1);
   }
   
   /* Last chance. The area above the highest partition. */
  skip_phys_mem:
   base = prev_top + 1;
   if ((implicit && find_gap_outside_phys_mem(&base, size, max_top)) ||
       (!implicit && (uint64_t)max_top-base+1 >= size))
   {
      struct partition_class *hal_class;
      mm_status_t status;

      hal_class = mm_add_partition(MM_PARTITION_OPT_SAS);
      if (hal_class == NULL)
      {
         return NULL;
      }

      status = mm_add_partition_alias(hal_class, name, MM_PARTITION_OPT_SAS);
      if (status != MM_SUCCESS)
      {
         return NULL;
      }

      status = mm_add_partition_bank(hal_class, base, size);
      if (status != MM_SUCCESS)
      {
         return NULL;
      }

      return hal_class;
   }
   
   return NULL;
}

/******************************************************************************
 * See partition.h
 */
struct partition_class *
mm_find_partition_by_name(const char *name)
{
   struct partition_class *p_class;
   struct partition_alias *alias;

   assert(NULL != name);

   for (p_class = gmm_state.partition_classes;
        p_class != NULL;
        p_class = p_class->next)
   {
      for (alias = p_class->aliases; alias != NULL; alias = alias->next)
      {
         if (strcmp(alias->name, name) == 0)
         {
            return p_class;
         }
      }
   }

   return NULL;
}

/******************************************************************************
 * See partition.h
 */
struct partition_class *
mm_find_partition_by_address(log_addr_t addr)
{
   struct partition_class *p;

   for (p = gmm_state.partition_classes; p != NULL; p = p->next)
   {
      if (p->is_superset_of(p, addr, (log_size_t)1))
      {
         return p;
      }
   }
   
   return NULL;
}

/******************************************************************************
 * See partition.h
 */
struct partition_class *
mm_find_sase_partition_by_range(log_addr_t base, log_size_t size)
{
   phys_addr_t top = (phys_addr_t)base + size - 1;

   struct partition_class *p;
   for (p = gmm_state.partition_classes; p != NULL; p = p->next)
   {
      struct partition_bank *bank;
      for (bank = p->banks; bank != NULL; bank = bank->next)
      {
         if (!partition_is_sase(p)) continue;
         if (top < bank->base)      continue;
         if (bank->top < base)      continue;

         return p;
      }
   }
   
   return NULL;
}


/****************************************************************************
 * See mm_private.h
 */
mm_bool_t
mm_space_is_mas(log_addr_t base, log_size_t size)
{
   struct partition_class *p;

   /* It is important that this function is extremely fast as it is used from
    * the validate_buffer interface used in each and every read or write file
    * system access. For that reason, this loop it removed. */
   
   /* Previous version of this function handled the case size=0 as size=1.
    * Don't change that behaviour.
    */
   if (size == 0)
   {
      size = 1;
   }
   
   for (p = gmm_state.partition_classes; p != NULL; p = p->next)
   {
      if (partition_is_non_exclusive(p) && p->overlap(p, base, size))
      {
         /* The partition holding the base address was found and supports
          * MAS mapping. If it also supports SAS, check if the address resides
          * below the current global MAS allocation top. */
         if (partition_is_combined(p))
         {
            struct firstfit_partition_class *fp =
               (struct firstfit_partition_class *)p;
            if (base < fp->global_mas_top)
               return TRUE;
            else
               return FALSE;
         }

         return TRUE;
      }
   }

   return FALSE;
}

/****************************************************************************
 * See mm_private.h
 */
struct partition_class *
mm_space_is_resident(log_addr_t base, log_size_t size)
{
   struct partition_class *p;

   /* It is important that this function is extremely fast as it is used from
    * the validate_buffer interface used in each and every read or write file
    * system access. */
   if (size == 0)
   {
      size = 1;
   }
   
   for (p = gmm_state.partition_classes; p != NULL; p = p->next)
   {
      if (partition_is_resident(p) && p->overlap(p, base, size))
      {
          return p;
      }
   }

   return NULL;
}

/*****************************************************************************
 * See partition.h
 *
 */

/*
 * See partition.h
 */
struct partition_class *
mm_add_partition(uint32_t options)
{
   struct partition_class *p_class;
   struct partition *partition;

   /* Assuming only one logical allocator implementation.
    * firstfit_init will never return an invalid pointer. */
   p_class = firstfit_init(options);

   /* Register the partition class */
   p_class->next = gmm_state.partition_classes;
   gmm_state.partition_classes = p_class;

   
   /* If the system domain already has been created,
    * all global partition classes are instantiated.
    * We must therefore manually instantiate all new classes that
    * appear beyond that point. It is needed to support
    * automatically configured log_mem:s, e.g. static regions
    * configured in no mans land. 
    * 
    * Only exclusive partitions are automatically configured.
    * That assumption is asserted in firstfit_create().
    */
   if (gmm_state.system_domain != NULL)
   {
      partition = p_class->create(p_class);
      if (partition == NULL)
      {
         return NULL;
      }

      partition->next = gmm_state.partition_instances;
      gmm_state.partition_instances = partition;
   }

   return p_class;
}


/*************************************************************************
 * See partition.h
 */
mm_status_t
mm_add_partition_bank(struct partition_class *p_class,
                      log_addr_t base, log_size_t size)
{
   struct partition_bank  *bank, *old_bank, *prev_old_bank = NULL;
   struct partition_class *part;

   /* First check if the new partition bank overlaps with any existing
    * partition bank. */
   for (part = gmm_state.partition_classes; part != NULL; part = part->next)
   {
      if (!partition_is_implicit(part) && part->overlap(part, base, size))
      {
         ramlog_printf("mm: mm_add_partion_bank detected overlap between new\n"
                       "    log_mem:%s  [%#010lx - %#010lx] and existing\n"
                       "    log_mem:%s\n",
                       p_class->name, base, base + size -1, part->name);

         return MM_EOVERLAPPING_ADDRESS;
      }
   }

   /* Create and initialize a new bank.
    * The function will never return an invalid pointer. */
   bank = p_class->init_bank(p_class, base, size);

   /* Insert the new bank into ordered partition bank list */
   for (old_bank = p_class->banks; old_bank != NULL; old_bank = old_bank->next)
   {
      if (bank->base < old_bank->base)
      {
         break;
      }
      prev_old_bank = old_bank;
   }

   if (prev_old_bank == NULL)
   {
      bank->next = p_class->banks;
      p_class->banks = bank;
   }
   else
   {
      bank->next = prev_old_bank->next;
      prev_old_bank->next = bank;
   }

   return MM_SUCCESS;
}


/*************************************************************************
 * See partition.h
 */
mm_status_t
mm_add_partition_alias(struct partition_class *p_class,
                       const char *name, uint32_t options)
{
   struct partition_alias *alias = NULL;
   struct partition_class *part;

   /* First check that no other alias has the same name */
   for (part = gmm_state.partition_classes; part != NULL; part = part->next)
   {
      for (alias = part->aliases; alias != NULL; alias = alias->next)
      {
         if (strcmp(alias->name, name) == 0)
         {
            ramlog_printf("mm: mm_add_partion_alias detected an existing\n"
                          "log_mem alias with identical name.\n"
                          "log_mem:%s\n", part->name);
            
            return MM_EILLEGAL_OPERATION;
         }
      }
   }

   alias = NULL;
   
   /* Extra caution. mm_alloc_boot_space might not update <alias>. */
   if (!mm_alloc_boot_space(sizeof(struct partition_alias), (void*)&alias))
   {
      MM_FATAL_ERROR2(("Boot heap exhausted.\n"), 0, 0);
   }

   alias->options = options;
   (void)strlcpy(alias->name, name, MM_NAME_LENGTH);

   /* The first alias always becomes the default */
   if (p_class->aliases == NULL)
   {
      alias->options |= MM_PARTITION_OPT_DEFAULT;
      (void)strlcpy(p_class->name, name, MM_NAME_LENGTH);
   }

   p_class->options |= options & MM_PARTITION_MAPTYPE_MASK;
   
   /* Insert into list of aliases */
   alias->next = p_class->aliases;
   p_class->aliases = alias;

   return MM_SUCCESS;
}


/*
 *************************************************************************
 * Functions needed by HAL.
 *************************************************************************
 */

/*
 * See partition.h
 *
 * Note that is important that this functions don't need make any more
 * zalloc than struct logical region. Otherwise page allocator extensions
 * relys on this function and no zone allocators can extend during
 * page allocator extension.
 *
 */

mm_bool_t
mm_alloc_log_mem2(struct partition *partition,
                  log_size_t size, log_addr_t *base)
{
   return mm_alloc_log_mem(partition, size, size, base);
}

mm_bool_t
mm_alloc_log_mem(struct partition *partition,
                 log_size_t size, log_size_t alignment, log_addr_t *base)
{
   struct log_object lobj;
   mm_status_t status;
   Msr sr = 0;

   assert(partition != NULL);

   MM_MULTITHREAD_LOCK(sr);

   /* First try to find some free space. */
   if (partition->class->alloc == NULL)
   {
      MM_MULTITHREAD_UNLOCK(sr);
      goto exit_false;
   }

   /* For regular regions, lobj is linked to the region descriptor
    * Mappings in the HAL are light weight and does not have a region
    * descriptor. Upon success, the freemem object must therefor be
    * moved to the freemem_prealloc_list before leaving this stack
    * context (see below). */
   lobj.base = 0;
   lobj.size = 0;
   lobj.freemem = NULL;

   /* Try to allocate a piece of logical memory */
   while (1)
   {
      size_t reclaimed_size;;

      status = partition->class->alloc(
         partition, &lobj, size, alignment,
         MM_PARTITION_OPT_SAS);

      if (status == MM_EOUT_OF_LOG_MEM || status == MM_EOUT_OF_PHYS_MEM)
      {
         reclaimed_size = ose_mm_reclaim_component_memory(
            &gmm_state, (U32)size, MM_INTERNAL_RAM);
         if (reclaimed_size == 0)
            break;
      }
      else
      {
         break;
      }
   }

   if (status != MM_SUCCESS)
   {
      MM_MULTITHREAD_UNLOCK(sr);
      goto exit_false;
   }

   /* As allocation succeeded, move the freemem object to the list.*/
   (void)freemem_put(lobj.freemem, &freemem_prealloc_list);

   *base = lobj.base;
   gmm_state.metadata_size += lobj.size;

   if (gmm_state.metadata_size > gmm_state.metadata_size_high)
   {
      gmm_state.metadata_size_high = gmm_state.metadata_size;
   }
   
   MM_MULTITHREAD_UNLOCK(sr);

   
   PRINT(("mm_alloc_log_mem2: base: %#08lx size %#08lx found in"\
          "partition %s\n", *base, size, partition->class->name));

   return TRUE;

  exit_false:
   /* FIXME, would be nice to distinguish when out of HAL_LOG_MEM */
   ramlog_printf("MM: %s : %ld \n"
                 "No logical space of size %#lx found in partition %s\n",
                 __FILE__, __LINE__,size, partition->class->name);

   return FALSE;

}


void
mm_query_log_mem2(U32 *usage, U32 *usage_high)
{
   *usage = gmm_state.metadata_size;
   *usage_high = gmm_state.metadata_size_high;
}


/*
 * See hal.h
 */
void
mm_free_log_mem(struct partition *partition,
		log_addr_t base, log_size_t size)
{
   struct log_object lobj;
   Msr sr = 0;

   MM_MULTITHREAD_LOCK(sr);
   lobj.base = base;
   lobj.size = size;
   lobj.options = MM_PARTITION_OPT_SAS;

   /* When freeing the memory a new freemem container might be
    * needed. It was pre-allocated during the allocation and
    * stored in the prealloc list. Retrieve it and put it in the
    * fake log_object so that the allocator thinks everything
    * is normal. */
   lobj.freemem = freemem_get(&freemem_prealloc_list);

   gmm_state.metadata_size -= size;
   (void)partition->class->free(partition, &lobj, 0);

   MM_MULTITHREAD_UNLOCK(sr);
}

/*****************************************************************************/

/*
 * Instantiates all partition classes in a newly created domain.
 */
mm_status_t
mm_create_domain_partitions(struct domain *domain)
{
   struct partition *partition;
   struct partition_class *p_class;
   int global = (gmm_state.partition_instances == NULL);

   assert(domain != NULL);
   assert(domain->partition_instances == NULL);

   domain->partition_instances = NULL;
   
   p_class = gmm_state.partition_classes;
   while (p_class != NULL)
   {
      /* Skip global partitions if they already have been created */
      if (!global && !partition_is_non_exclusive(p_class))
      {
         p_class = p_class->next;
         continue;
      }
      
      partition = p_class->create(p_class);
      if (partition == NULL)
      {
         partition = domain->partition_instances;
         while (partition)
         {
            struct partition *next = partition->next;

            partition->class->destroy(partition);
            partition = next;
         }

         domain->partition_instances = NULL;

         return MM_EOUT_OF_PHYS_MEM;
      }

      /* Add new partition to domain list or global list. */
      if (partition_is_non_exclusive(p_class))
      {
         partition->next = domain->partition_instances;
         domain->partition_instances = partition;
      }
      else
      {
         partition->next = gmm_state.partition_instances;
         gmm_state.partition_instances = partition;
      }

      p_class = p_class->next;
   }

   return MM_SUCCESS;
}

/*
 * Destroys all instantiated partitions in all domain.  
 */
mm_status_t
mm_destroy_domain_partitions(struct domain *domain)
{
   struct partition *partition;

   assert(domain != NULL);
 
   partition = domain->partition_instances;
   while (partition != NULL)
   {
      struct partition *next = partition->next;
      partition->class->destroy(partition);
      partition = next;
   }

   domain->partition_instances = NULL;

   /* Global partitions are never destroyed */

   return MM_SUCCESS;
}

/*
 * Finds a domain's private logical allocator (partition) by name.
 *
 * Use domain=NULL for system domain.
 */
struct partition *
mm_find_domain_partition_by_name(const struct domain *domain, const char *name)
{
   struct partition *partition;
   struct partition_alias *alias;

   if (domain == NULL)
   {
      domain = gmm_state.system_domain;
   }
   
   assert(domain != NULL);
   assert(name != NULL);

   partition = domain->partition_instances;
   while (partition != NULL)
   {
      for (alias = partition->class->aliases; alias != NULL; alias = alias->next)
      {
         if (strncmp(alias->name, name, MM_NAME_LENGTH) == 0)
         {
            return partition;
         }
      }

      partition = partition->next;
   }

   partition = gmm_state.partition_instances;
   while (partition != NULL)
   {
      for (alias = partition->class->aliases; alias != NULL; alias = alias->next)
      {
         if (strncmp(alias->name, name, MM_NAME_LENGTH) == 0)
         {
            return partition;
         }
      }

      partition = partition->next;
   }

   return partition;
}

/*
 * Finds a domain's private logical allocator (partition) by address. 
 */
struct partition *
mm_find_domain_partition_by_addr(const struct domain *domain, log_addr_t base)
{
   struct partition *partition;

   if (domain == NULL)
   {
      domain = gmm_state.system_domain;
   }

   assert(domain != NULL);

   partition = domain->partition_instances;
   while (partition != NULL)
   {
      if (partition->class->overlap(partition->class, base, (log_size_t)1))
      {
         return partition;
      }

      partition = partition->next;
   }

   partition = gmm_state.partition_instances;
   while (partition != NULL)
   {
      if (partition->class->overlap(partition->class, base, (log_size_t)1))
      {
         return partition;
      }

      partition = partition->next;
   }

   return partition;   
}


/*
 * Look up the mapping type of a partition given an alias name. 
 */
mm_status_t
mm_find_partition_mapping_type(const struct partition_class *p_class,
                               char *alias_name,
                               mapping_type_t *type)
{
   struct partition_alias *alias = NULL;

   assert(type != NULL);

   alias = mm_find_partition_alias(p_class, alias_name);
   if (alias == NULL)
   {
      return MM_ENON_EXISTING_LOG_MEM;
   }

   switch (alias->options & MM_PARTITION_MAPTYPE_MASK)
   {
      case MM_PARTITION_OPT_SASE:
         *type = SASE;
         break;
      case MM_PARTITION_OPT_SAS:
         *type = SAS;
         break;
      case MM_PARTITION_OPT_MAS:
         *type = MAS;
         break;

      default:
         INTERNAL_ERROR();
   }

   return MM_SUCCESS;
}

/*
 * Look up a partition alias. 
 */
struct partition_alias *
mm_find_partition_alias(const struct partition_class *p_class,
                        char *alias_name)
{
   struct partition_alias *alias = NULL;

   if (p_class == NULL && alias_name != NULL && *alias_name != '\0')
   {
      for (p_class = gmm_state.partition_classes;
           p_class != NULL;
           p_class = p_class->next)
      {
         for (alias = p_class->aliases; alias != NULL; alias = alias->next)
         {
            if (strcmp(alias->name, alias_name) == 0)
            {
               return alias;
            }
         }
      }
   }
   else if (p_class != NULL)
   {
      if (alias_name == NULL || *alias_name == '\0')
      {
         for (alias = p_class->aliases; alias != NULL; alias = alias->next)
         {
            if (alias->options & MM_PARTITION_OPT_DEFAULT)
            {
               if (alias_name != NULL && *alias_name == '\0')
               {
                  /* Report alias name to caller. */
                  strlcpy(alias_name, alias->name, MM_NAME_LENGTH);
               }
               return alias;
            }
         }
         return NULL;
      }
      else
      {
         for (alias = p_class->aliases; alias != NULL; alias = alias->next)
         {
            if (strcmp(alias->name, alias_name) == 0)
            {
               return alias;
            }
         }
      }
   }

   return NULL;
}


#if defined(OSPP) || defined(OSARM)
/**************************************************************************/
/*
 * Resident log_mem mapping supported on powerpc and arm.
 * */

static void
map_resident_range(log_addr_t base, log_addr_t top,
                   U32 perm, U32 cache, U32 options)
{
   /* Sizes hardcoded for PEM */
   const log_size_t *page_sizes;
   unsigned num_page_sizes;

   top += 1;

   options |= HAL_GLOBAL_MAPPING;
   
   /* If no resident mapping support is available,
    * the range is mapped with regular dynamic pages
    * in the global translation tree. */
   if (hal_get_info()->no_resident_page_sizes && hal_create_resident_mapping)
   {
      page_sizes = hal_get_info()->resident_page_sizes;
      num_page_sizes = hal_get_info()->no_resident_page_sizes;
   }
   else
   {
      page_sizes = hal_get_info()->page_sizes;
      num_page_sizes = hal_get_info()->no_page_sizes;
   }

   while (base < top)
   {
      hal_status_t status = HAL_SUCCESS;
      int ix;

      for (ix = num_page_sizes - 1; ix >= 0; ix--)
      {
         if (page_sizes[ix] <= (top-base) &&
             mm_log_aligned(base, page_sizes[ix]))
         {
            break;
         }
      }
      
      if (ix < 0)
      {
         MM_FATAL_ERROR2(("Failed to make log_mem resident. "
                          "Illegal alignment at 0x%08x.\n", base),
                         OSE_MM_ECONFIGURATION_ERROR, (OSERRCODE)
                         "Failed to log_mem resident. Illegal alignment.");
      }

      if (hal_get_info()->no_resident_page_sizes && hal_create_resident_mapping)
      {
         status = hal_create_resident_mapping(
            gmm_state.system_domain->hal_domain,
            base, (phys_addr_t)base,
            page_sizes[ix],
            options, perm, cache);

         ramlog_printf("mm: res_mem  [0x%08x-0x%08x]  0x%08x  [",
                       base, base + page_sizes[ix] - 1, page_sizes[ix]);
         if (options & HAL_CONTENTS_TYPE_TEXT)
            ramlog_printf("I");
         if (options & HAL_CONTENTS_TYPE_DATA)
            ramlog_printf("D");
         ramlog_printf("]\n");
      }
      else
      {
	 
#ifdef OSE_SMP         
         if(zzkrn_cpu_id() == 0)
#endif
            {
               status = hal_map_page(
                  gmm_state.system_domain->hal_domain,
                  base, (phys_addr_t)base,
                  page_sizes[ix],
                  options, perm, cache);
               
               ramlog_printf("mm: page table res_mem [0x%08x-0x%08x] 0x%08x [",
                             base, base + page_sizes[ix] - 1, page_sizes[ix]);
               if (options & HAL_CONTENTS_TYPE_TEXT)
                  ramlog_printf("I");
               if (options & HAL_CONTENTS_TYPE_DATA)
                  ramlog_printf("D");
               ramlog_printf("]\n");
            }
      }

      if (status != HAL_SUCCESS)
      {
         MM_FATAL_ERROR2(("Failed to make area resident: status = 0x%x\n", status),
                         OSE_MM_ECONFIGURATION_ERROR, (OSERRCODE)
                         "Failed to make area resident.");
      }

      base += page_sizes[ix];
   }
}


/*
 *
 */
void
mm_map_resident_partitions(void)
{
   struct partition_class *p_class;
   struct partition_bank  *p_bank;

   for (p_class = gmm_state.partition_classes;
        p_class != NULL;
        p_class = p_class->next)
   {
      if (!partition_is_resident(p_class))
         continue;

      for (p_bank = p_class->banks;
           p_bank != NULL;
           p_bank = p_bank->next)
      {
         map_resident_range(p_bank->base,
                            p_bank->top,
                            p_class->res_perm,
                            p_class->res_cache,
                            p_class->res_options);

         mm_kick_wdog();
      }
   }
}
#endif
