/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2007 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

/* C runtime include files. */
#include <stddef.h>

/* OSE public include files. */
#include "assert.h"

/* OSE internal include files. */
#include "page.h"

/* MM internal include files. */
#include "allocator.h"
#include "common.h"
#include "zone.h"
#include "phys_object.h"
#include "gmm.h"

/*
 * Avoid free macro from OSE Heap. Observe that the heap
 * is not used in MM, but stddef.h includes malloc.h that
 * causes the problem.
 */
#ifdef free
#undef free
#endif

/*
 * Used for phys_object.
 */
static struct zone mix_zone;

void
mm_phys_object_init(void)
{
   mm_zinit(&mix_zone, "struct phys_object",
            sizeof(struct phys_object), 5, 20);
}

/*
 * See phys_object.h
 */
struct phys_object *
mm_create_phys_object(struct allocator *allocator, U32 meta_options)
{
   struct phys_object *object;

   object = mm_zalloc(&mix_zone);
   if (object)
   {
      object->meta_options = meta_options;
      object->allocator = allocator;
      
      object->data_representation = PHYS_OBJ_NULL;

      /* Start with one instance region reference. */
      object->ref_count = 0;
   }
   return object;
}

/*
 * Attach to an object.
 * Increases the reference counter of an object to
 * prevent the physical resources from being freed. 
 */
void
mm_attach_phys_object(struct phys_object *object)
{
   assert(object != NULL);

   object->ref_count++;
}

/*
 * See phys_object.h
 */
void
mm_detach_phys_object(struct phys_object *object)
{
   assert(object != NULL);

   if (object->ref_count > 0)
   {
      /* Decrease use count of this phys_object.
       * If nobody references it, it is destroyed and
       * the physical memory is freed.. */
      object->ref_count--;
   }

   /* First check the reference counter in physical object. */
   if (object->ref_count == 0)
   {
      /* Last user of this object is gone. */

      /* Start with clearing attached resources. */

      if (object->allocator) {
         struct allocator *allocator = object->allocator;
         
         allocator->free(allocator, object);
      }
      else {
         switch (object->data_representation) {
            case PHYS_OBJ_CLIENT_PAGES:
               free_buf((union SIGNAL**)(void*)&object->data.client_pages);
               break;
               
            case PHYS_OBJ_FRAG_LIST:
               INTERNAL_ERROR();
               /*lint -unreachable */
               break;
               
            case PHYS_OBJ_NULL:
            case PHYS_OBJ_DIRECT_PHYS_FRAG:
               break;
               
            default:
               INTERNAL_ERROR();
               /*lint -unreachable */
         }
      }

      /* Then destroy <object> itself. */
      mm_zfree(&mix_zone, object);
   }
}


static mm_bool_t
search_phys_container(const struct phys_object *object,
                      phys_size_t offset,
                      phys_addr_t *pa,
                      phys_addr_t *pa_start,
                      phys_size_t *size)
{
   struct phys_mem_cursor cursor;

   assert(object != NULL);
   assert(pa != NULL);
   assert(pa_start != NULL);
   assert(size != NULL);

   mm_phys_mem_get_first(object, &cursor);
   do
   {
      /* Calculate remaining offset after this fragment. */
      if (offset < cursor.phys_page.size)
      {
         /* The logical address is translated to this fragment, */
         *pa = cursor.phys_page.pa + offset;
         *pa_start = cursor.phys_page.pa;
         *size = cursor.phys_page.size;
         return TRUE;
      }

      /* Not within this fragment. Move on to next fragment. */
      offset -= cursor.phys_page.size;
   } while (mm_phys_mem_get_next(&cursor));
   return FALSE;
}


/*
 * See phys_object.h.
 */
mm_bool_t
mm_extract_phys_addr(const struct phys_object *object,
                     phys_size_t offset,
                     phys_addr_t *pa,
                     phys_addr_t *pa_start,
                     phys_size_t *size)
{
   assert(object != NULL);
   assert(pa != NULL);
   assert(pa_start != NULL);
   assert(size != NULL);

   return search_phys_container(object, offset, pa, pa_start, size);
}

/*****************************************************************************
 *
 * See phys_object.h
 */

void
mm_phys_mem_get_first(const struct phys_object *phys_obj,
                      struct phys_mem_cursor *cursor)
{
   assert(phys_obj != NULL);
   assert(cursor != NULL);
   cursor->phys_obj = phys_obj;
   switch (phys_obj->data_representation)
   {
      case PHYS_OBJ_NULL:
         INTERNAL_ERROR();
         /*lint -unreachable */
         break;

      case PHYS_OBJ_FRAG_LIST:
      {
         phys_addr_t base = FC_GET_BASE(phys_obj->data.frag_list);
         phys_addr_t top  = FC_GET_TOP(phys_obj->data.frag_list);
         phys_size_t size = top+(phys_addr_t)1-base;

         assert(phys_obj->data.frag_list);
         cursor->data.current_frag = phys_obj->data.frag_list;
         cursor->phys_page.pa   = base;
         cursor->phys_page.size = size;
         break;
      }
      
      case PHYS_OBJ_DIRECT_PHYS_FRAG:
         cursor->phys_page.pa   = phys_obj->data.direct_phys_frag.pa;
         cursor->phys_page.size = phys_obj->data.direct_phys_frag.size;
         break;
         
      case PHYS_OBJ_CLIENT_PAGES:
      {
         struct mm_create_region *cp = phys_obj->data.client_pages;
         assert(cp->phys_count > 0);
         cursor->data.index = 0;
         cursor->phys_page.pa   = cp->phys_resources[0].base;
         cursor->phys_page.size = cp->phys_resources[0].size;
         break;
      }
         
      default:
         INTERNAL_ERROR();
         /*lint -unreachable */
         break;
   }
}


/*****************************************************************************
 *
 * See phys_object.h
 */
int
mm_phys_mem_get_next(struct phys_mem_cursor *cursor)
{
   assert(cursor != NULL);
   switch (cursor->phys_obj->data_representation) {

      case PHYS_OBJ_NULL:
         INTERNAL_ERROR();
         /*lint -unreachable */
         break;
         
      case PHYS_OBJ_FRAG_LIST:
      {
         struct fragment_container *f = FC_GET_NEXT(cursor->data.current_frag);

         if (f == NULL) return 0;

         cursor->data.current_frag = f;
         cursor->phys_page.pa   = FC_GET_BASE(f);
         cursor->phys_page.size = FC_GET_SIZE(f);
      }
      break;
      
      case PHYS_OBJ_DIRECT_PHYS_FRAG:
         /* Direct SASE always contains one element. */
         return 0;
         
      case PHYS_OBJ_CLIENT_PAGES:
      {
         struct mm_create_region *cp = cursor->phys_obj->data.client_pages;

         cursor->data.index++;
         if (cursor->data.index >= cp->phys_count) {
            return 0;
         }
         
         cursor->phys_page.pa   = cp->phys_resources[cursor->data.index].base;
         cursor->phys_page.size = cp->phys_resources[cursor->data.index].size;
         break;
      }
         
      default:
         INTERNAL_ERROR();
         /*lint -unreachable */
         break;
         
   }
   return 1;
}

/******************************************************************************
 *
 */
void print_phys_object(struct phys_object *p) {
   struct phys_mem_cursor cursor;
   ramlog_printf("phys_obj(0x%08lx)\n", p);
   if (p == NULL) return;
   if (p->data_representation == PHYS_OBJ_NULL) return;
   
   mm_phys_mem_get_first(p, &cursor);
   do {
      ramlog_printf("pa: 0x%08llx, sz:0x%08llx\n",
                    cursor.phys_page.pa,
                    cursor.phys_page.size);
   } while (mm_phys_mem_get_next(&cursor));
   ramlog_printf("-\n");
}

