/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2007 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

/* C runtime include files. */
#include <string.h>

/* OSE public include files. */
#include "assert.h"
#include "mm.sig"
#include "mm_err.h"
#include "mmtypes.h"
#include "ose.h"
#include "osetypes.h"
#include "spinlock.h"


/* OSE internal include files. */
#include "hal.h"


/* MM internal include files. */
#include "allocator.h"
#include "gmm.h"
#include "hal.h"
#include "page.h"
#include "ram_allocator.h"
#include "zone.h"
#include "ramlog.h"
#include "configure.h"
#include "mm_log.h"

union SIGNAL {
   struct mm_get_phys_mem_info mm_get_phys_mem_info;
};


#define VERBOZE(s)  log(CATEGORY_START, s)

#define DETECT_MULTIPLE_PAGE_ALLOCATIONS 0
#if DETECT_MULTIPLE_PAGE_ALLOCATIONS
/* If this code is included, a parallel memory representation is used to
 * detect if a page is allocated more than once, which would be an error.
 */
static U32 *bits = NULL;

#  ifdef OSMIPS
#    define PAGE_SHIFT 13
#  else
#    define PAGE_SHIFT 12
#  endif

#  define MAX_NR_OF_PAGES (1UL << (32-PAGE_SHIFT))
#  define BITS_SIZE       (MAX_NR_OF_PAGES/8) /* 8 bits per byte. */

#  define INIT_BITS \
   do {\
      if (bits == NULL) {\
         bits = boot_alloc(BITS_SIZE);\
         bzero(bits, BITS_SIZE);\
      }\
   } while (0)

#  define ALLOC_BITS(b, s) alloc_bits(__FILE__, __LINE__, (b), (s))

static void alloc_bits(const char *file, int line,
                       log_addr_t base, log_size_t size)
{
   U32 page_nr  = base >> PAGE_SHIFT;
   U32 page_cnt = size >> PAGE_SHIFT;

   ramlog_printf("alloc_bits(0x%08lx-0x%08lx)\n",
                 base, base+size-1);

   while (page_cnt) {
      U32 ix = page_nr >> 5;
      U32 mask = 1UL << (page_nr & 0x1fUL);
      if (!(bits[ix] & mask)) {
         ramlog_printf("%s:%d. ERROR(alloc_bits): [0x%08lx-0x%08lx] "
                       "base: 0x%08lx bits[0x%lx]: 0x%08lx, mask: 0x%08lx\n",
                       file, line, base, base+size-1,
                       page_nr<<PAGE_SHIFT, ix, bits[ix], mask);
         INTERNAL_ERROR();
      }
      bits[ix] &= ~mask;
      page_cnt--;
      page_nr++;
   }
}

#  define FREE_BITS(b, s) free_bits(__FILE__, __LINE__, (b), (s))
static void free_bits(const char *file, int line,
                      log_addr_t base, log_size_t size)
{
   U32 page_nr = base >> PAGE_SHIFT;
   U32 page_cnt   = size >> PAGE_SHIFT;

   ramlog_printf("free_bits(0x%08lx-0x%08lx)\n",
                 base, base+size-1);

   while (page_cnt) {
      U32 ix = page_nr >> 5;
      U32 mask = 1UL << (page_nr & 0x1fUL);
      if (bits[ix] & mask) {
         ramlog_printf("%s:%d. ERROR(free_bits): [0x%08lx-0x%08lx] "
                       "base: 0x%08lx bits[0x%lx]: 0x%08lx, mask: 0x%08lx\n",
                       file, line, base, base+size-1,
                       page_nr<<PAGE_SHIFT, ix, bits[ix], mask);
         INTERNAL_ERROR();
      }
      bits[ix] |= mask;
      page_cnt--;
      page_nr++;
   }
}

#else

#  define ALLOC_BITS(b, s)
#  define FREE_BITS(b, s)
#  define INIT_BITS

#endif

/*****************************************************************************/

#ifdef ASSERT

/* To catch errors in linked lists, a time-to-live (TTL) counter initiated to
 * this value is used when traversing loops.
 * It is used as:
 *   {
 *     struct fragment_container *f = <list_head>;
 *     INIT_TTL;
 *     ...
 *     while (f) {
 *       ...
 *       CHECK_TTL;
 *       f = f->next;
 *     }
 *   }
 */

static unsigned long long max_frag_cnt = 0;
static unsigned long long max_bank_cnt = 0;

/*lint -save -e773 */
#  define INIT_FRAG_TTL unsigned long long frag_ttl_ = max_frag_cnt
#  define INIT_BANK_TTL unsigned long long bank_ttl_ = max_bank_cnt
/*lint -restore */
#  define CHECK_FRAG_TTL \
do {\
   if (!frag_ttl_--) {\
      INTERNAL_ERROR();\
   }\
} while (0)\

#  define CHECK_BANK_TTL \
do {\
   if (!bank_ttl_--) {\
      INTERNAL_ERROR();\
   }\
} while (0)\


#else

#  define INIT_FRAG_TTL
#  define INIT_BANK_TTL
#  define CHECK_FRAG_TTL
#  define CHECK_BANK_TTL

#endif

/*
 * When allocating scattered memory, it is necessary to know the fragments'
 * algnment, but the struct fragment_container is not designed to contain that
 * information.
 * One solution would be to increase the struct, but we use a lot of them, so
 * this would cause an unnecessary growth of metadata.
 * Instead, we use the fact that the fragment base is aligned at smallest page
 * size, leaving a number of bits that are always 0 at the end. If we assume
 * that there are no more than 256 different page sizes and that the smallest
 * page_size is 256 bytes or larger, then we can use the last 8 bits. They are
 * used to temporarily store the alignment as an index to the page size array.
 * The use is only internal, and all fragments that leaves the scope of this
 * file will have these bits set to 0 <=> The fragment base is the actual
 * fragment base.
 *
 * Be careful then using ~PAGE_INDEX_MASK together with a 64 bit valiable.
 *  ~PAGE_INDEX_MASK              == 0xffffff00
 *  ~(phys_addr_t)PAGE_INDEX_MASK == 0xffffffffffffff00
 */

#define PAGE_INDEX_MASK 0xffUL

/******************************************************************************
 * CONTAINER_EXTEND_START is the threshold where new containers are allocated.
 * The exact condition is (free_cnt <= CONTAINER_EXTEND_START). This means
 * that the number of containers during the extension is
 *   (CONTAINER_EXTEND_START-1).
 *
 * The largest number of required containers occurs in a system with MMU and
 * when the MMU is enabled. That number is estimated below. (Compare with the
 * extension code in alloc_fragment_container).
 *
 * + mm_alloc_log_mem2:
 *     This call can cause a zone extension -> mm_page_alloc, which requires
 *     max 2 + 1 temporary container.
 *
 * +  mm_page_alloc:
 *     max2 + 1 temporary container.
 *
 * + hal_map_page:
 *     The actual number of containers depends on the actual mmu, but an upper
 *     estimate is:
 *       1 page for a new L1 table (mapping of regin in a new domain)
 *       2 pages for a new local L2 node. (The PTE might be 64 bits.)
 *       2 pages for a new global L2 node.
 *     Each page requires max 2 containers + 1 temporary container.
 *
 * This results in an upper estimate 15 containers (14 + 1 temporary)
 *
 * CONTAINER_EXTEND_STOP is the min number of containers before extension is
 * stopped. The actual value is not critical as long as it is larger than
 * CONTAINER_EXTEND_START.
 */

#define CONTAINER_EXTEND_START 15UL
#define CONTAINER_EXTEND_STOP  25UL

/*****************************************************************************/

/*
 * The mem_bank struct contains the data for one memory bank. If there are
 * more banks, they are linked to a list through the next-field. The fragments
 * array contains pre allocated containers, and its actual size depends on the
 * bank size.
 */
struct mem_bank {
   struct mem_bank *next;

   /* FIXME, optimize for footprint. */
   phys_addr_t base;
   phys_addr_t top;
   
   /* The size of non allocated physical memory in this memory bank. */
   phys_size_t free_size;

   struct fragment_container *free_mem;
   struct fragment_container *unused_containers;
   U32 unlinked_containers_cnt;
   U32 frag_cnt;
   struct fragment_container fragments[1];
};

struct allocator_internal_data {
   struct mem_bank *sas_banks;

   struct mem_bank *sase_banks;

   /* The size of non allocated physical memory in all memory banks. */
   phys_size_t free_size;

   phys_size_t watermark;
};


/* A list of memory allocated through mm_page_alloc. */
static struct fragment_container *page_list = NULL;

/*****************************************************************************/

static void *
boot_alloc(size_t sz)
{
   log_addr_t tmp;

   if (!mm_alloc_boot_space(sz, &tmp)) {
      MM_FATAL_ERROR2(("Boot heap exhausted.\n"), 0, 0);
   }
   return (void*)tmp;
}

/*****************************************************************************/
#ifdef TEST_ALLOC_FRAGMENT_CONTAINER

static void
publish_internal_state(U32* _extending_p, U32* _init_done_p);

static void
test_before_extension(void);

#endif

static struct fragment_container *
alloc_fragment_container(struct mem_bank *b)
{
   struct fragment_container *tmp = NULL;
   Msr sr = 0;


   /* Currently, all containers that can be needed is pre allocated in
    * add_bank() for reasons described in that function. If things change and
    * pre allocation is no longer needed, watch out for recursion:
    *   - alloc_fragment_container needs to allocate a new page and calls
    *     mm_page_alloc
    *   - mm_page_alloc calls alloc_fragment
    *   - alloc_fragment calls alloc_fragment_container
    */
   MM_MULTITHREAD_LOCK(sr);

   if (b->unused_containers) {
      tmp = b->unused_containers;
      b->unused_containers = FC_GET_NEXT(b->unused_containers);
   }
   else if (b->unlinked_containers_cnt) {
      b->unlinked_containers_cnt--;
      tmp = &b->fragments[b->unlinked_containers_cnt];
   }
   else {
      MM_FATAL_ERROR2(
         ("Failed to allocate fragment container"),
         MM_EINTERNAL_ERROR, 0);
   }
   MM_MULTITHREAD_UNLOCK(sr);

   memset(tmp, 0, sizeof(struct fragment_container));
   return tmp;
}


/*****************************************************************************/

static void
free_fragment_container(struct mem_bank *b,
                        struct fragment_container *f)
{
   Msr sr = 0;
   
   MM_MULTITHREAD_LOCK(sr);
   memset(f, 0, sizeof(struct fragment_container));
   FC_SET_NEXT(f, b->unused_containers);
   b->unused_containers = f;
   MM_MULTITHREAD_UNLOCK(sr);
}

/*****************************************************************************/

static void
print_range(int (*print)(const char*, ...),
            const char *prefix,
            phys_addr_t base,
            phys_addr_t top,
            const char *suffix)
{
   phys_size_t size = top + 1 - base;

   /*lint -save -e522 "Expected void type, assignment, increment or decrement"
    */
   print("%s[0x%08llx - 0x%08llx] (0x%llx)%s",
         prefix, base, top, size, suffix);
   /*lint -restore */
}

/*****************************************************************************/
static void
print_bank_list(int (*print)(const char*, ...),
                const char *name,
                const struct mem_bank *bank)
{
   /*lint -save -e522 "Expected void type, assignment, increment or decrement"
    */
   U32 n,bankNr = 0;
   struct fragment_container *tmp;
   INIT_BANK_TTL;

   print("  %s banks:\n", name);
   while (bank) {
      print("    Bank %lu\n", bankNr++);
      print("      Bank range:\n");
      print_range(print, "        ", bank->base, bank->top, "\n");
      print("\n");
      print("      Free mem:\n");

      {
         phys_size_t tot_free_size = 0;

         for (tmp = bank->free_mem; tmp; tmp = FC_GET_NEXT(tmp)) {
            print_range(print, "        ", FC_GET_BASE(tmp),
                        FC_GET_TOP(tmp), "\n");
            assert(FC_GET_TOP(tmp));
            tot_free_size += FC_GET_SIZE(tmp);
         }
         print("\n");
         print("      Total free mem : 0x%08llx\n", tot_free_size);
      }
      for (n = 0, tmp = bank->unused_containers; tmp; tmp = FC_GET_NEXT(tmp)) n++;

      print("\n");
      print("      Unused fragments:\n");
      print("        linked     : %lu\n", n);
      print("        not linked : %lu\n",
                    bank->unlinked_containers_cnt);
      
      print("\n");
      bank = bank->next;
      CHECK_BANK_TTL;
   }
   /*lint -restore */
}
/*****************************************************************************/

static void
dump_state(const struct allocator *a, int (*print)(const char*, ...))
{
   /*lint -save -e522 "Expected void type, assignment, increment or decrement".
    * Refers to 'print' 
    */
   print("Phys allocator \"%s\":\n", a->name);
   print_bank_list(print, "SAS",  a->state->sas_banks);
   print_bank_list(print, "SASE", a->state->sase_banks);
   if (a == gmm_state.ram_allocator) {
      struct fragment_container *f = page_list;
      INIT_FRAG_TTL;

      print("Memory allocated through mm_page_alloc:\n");
      while (f) {
         print_range(print, "  ",  FC_GET_BASE(f), FC_GET_TOP(f), "\n");
         f = FC_GET_NEXT(f);
         CHECK_FRAG_TTL;
      }
   }

   print("Free memory: %#llx\n", a->state->free_size);
   print("Free watermark: %#llx\n", a->state->watermark);
   /*lint -restore */
}

/*****************************************************************************/

static struct
mem_bank *find_bank(const struct allocator_internal_data *id, phys_addr_t addr)
{
   struct mem_bank *tmp;
   for (tmp = id->sas_banks; tmp; tmp = tmp->next) {
      if (addr >= tmp->base && addr <= tmp->top) {
         return tmp;
      }
   }
   for (tmp = id->sase_banks; tmp; tmp = tmp->next) {
      if (addr >= tmp->base && addr <= tmp->top) {
         return tmp;
      }
   }
   return NULL;
}

/*****************************************************************************/

static struct fragment_container *
extract_from_fragment(struct allocator_internal_data *id,
                      struct mem_bank *b,
                      struct fragment_container **f,
                      phys_addr_t base,
                      phys_addr_t top)
{
   struct fragment_container *low, *high;
   struct fragment_container *tmp;
   struct fragment_container *prev = NULL;
   Msr sr = 0;

   MM_MULTITHREAD_LOCK(sr);
   
   assert(base < top);
   assert(FC_GET_BASE(*f) <= base);
   assert(FC_GET_TOP(*f)  >= top );

   if (FC_GET_BASE(*f) != base) {
      /* Split the fragment at base. f will refer to the fragment
       * starting at base. */

      low = alloc_fragment_container(b);
      FC_SET_BASE(low, base);
      FC_SET_TOP(low, FC_GET_TOP(*f));
      FC_SET_NEXT(low, FC_GET_NEXT(*f));
      FC_SET_TOP(*f, base-1);
      FC_SET_NEXT(*f, low);

      prev = *f;
      f = &low;
   }


   if (FC_GET_TOP(*f) != top) {
      /* Split the fragment at top. f will refer to the fragment
       * starting at base. */

      high = alloc_fragment_container(b);
      FC_SET_BASE(high, top+1);
      FC_SET_TOP(high, FC_GET_TOP(*f));
      FC_SET_NEXT(high, FC_GET_NEXT(*f));
      FC_SET_TOP(*f, top);
      FC_SET_NEXT(*f, high);
   }

   /* Unlink f. After this the extracted fragment is refered to by tmp. */
   tmp = *f;
   if (prev)
   {
      FC_SET_NEXT(prev, FC_GET_NEXT(*f));
   }
   else
   {
      *f = FC_GET_NEXT(*f);
   }
   FC_SET_NEXT(tmp, NULL);

   assert(FC_GET_BASE(tmp) == base);
   assert(FC_GET_TOP(tmp) == top);

   id->free_size -= top + 1 - base;
   b->free_size -= top + 1 - base;

   if (id->watermark > id->free_size)
   {
      id->watermark = id->free_size;
   }

   
   assert(tmp);
   ALLOC_BITS(base, top+1-base);

   MM_MULTITHREAD_UNLOCK(sr);
   return tmp;
}

/*****************************************************************************/

static struct fragment_container *
alloc_fragment(struct allocator_internal_data *id,
               struct mem_bank *b,
               phys_size_t size,
               phys_size_t alignment,
               phys_addr_t range_base,
               phys_size_t range_size,
               U32 meta_options)
{
   phys_addr_t range_top = range_base + range_size - 1;
   Msr sr = 0;

   if (size == 0) {
      ramlog_printf("Size is 0.\n");
      return NULL;
   }

   if (range_top < range_base) {
      ramlog_printf("Request to allocate memory in invalid range "
                    "[0x%08lx-0x%08lx].\n", range_base, range_top);
      return NULL;
   }
   
   if (alignment == 0) {
      /* Setting alignment to 0 is interpreted as no constraints on the
       * alignment. */
      alignment = 1;
   }
   else if (alignment & (alignment-1)) {
      MM_FATAL_ERROR2(
         ("Alignment (0x%lx) is not a power of 2.\n", alignment),
         0, 0);
   }                 
   
   MM_MULTITHREAD_LOCK(sr);
   for (; b; b = b->next) {
      struct fragment_container *f;
      struct fragment_container *prev = NULL;

      for (f = b->free_mem; f; prev = f, f = FC_GET_NEXT(f)) {
         phys_addr_t base = FC_GET_BASE(f);

         if (base < range_base) {
            base = range_base;
         }
         base = ALIGN_UP(base, (phys_addr_t)alignment);

         {
            phys_addr_t top = base + size - 1;
            struct fragment_container *tmp;
            
            /* Check that [base..top] is within the fragment *f. */
            if ((top > range_top) || (top > FC_GET_TOP(f))) {
               continue;
            }
            
            if (meta_options & MM_REGION_OPT_PERSISTENT) {
               /* Avoid memory that originated from any initial region. */
               if (mm_initial_region_overlap(base, size)) {
                  continue;
               }
            }

            
            /* Now, we have a fragment that can be used. */
            tmp = extract_from_fragment(id, b, &f, base, top);

            /* Have to link in the updated fragment. */
            if (prev)
            {
               FC_SET_NEXT(prev, f);
            }
            else
            {
               b->free_mem = f;
            }
            MM_MULTITHREAD_UNLOCK(sr);
            return tmp;
         }
      }
   }
   MM_MULTITHREAD_UNLOCK(sr);

   return NULL;
}

/******************************************************************************
 * Frees a list of containers.
 *
 * At first, the function was implemented as described below:
 *
 *  while (list) {
 *    find_bank
 *    find position in free_list
 *    insert
 *  }
 *
 * This is functionally correct, but a little slow.
 * The only function that generates a fragment list is alloc_f, and these lists
 * have the following useful properties:
 * 1) All fragments from one bank is grouped together.
 * 2) All fragments from one bank is sorted on ascending address.
 *
 * Two optimizations are made based on these properties:
 * - A check if the current fragment is located in the same bank as the
 *   previous is made before making a new bank search.
 * - The position where the previuos fragment was inserted is kept and used as
 *   starting point when the position of the current fragment is searched for.
 */

static void
free_fragment_list(struct allocator_internal_data *id,
                   struct fragment_container **head)
{
   struct fragment_container *list, *f;
   Msr sr = 0;

   assert(head);
   if (*head == NULL) { return; }

   MM_MULTITHREAD_LOCK(sr);
   list = *head;
   *head = NULL;
   f = list;
   list = FC_GET_NEXT(list);

   assert(f);
#if FC_PACK_LEVEL != 2   
   assert(((U32)f & 3) == 0);
#endif   
   assert(FC_GET_BASE(f) < FC_GET_TOP(f));

   for (;;) {
      struct mem_bank *curr_bank = find_bank(id, FC_GET_BASE(f));
      struct fragment_container *search_start;

      if (curr_bank == NULL) {
         MM_MULTITHREAD_UNLOCK(sr);
         MM_FATAL_ERROR2(
            ("Base address 0x%08llx not located in any memory bank.\n",
             FC_GET_BASE(f)),
            0, 0);
      }

      search_start = curr_bank->free_mem;

      for (;;) {
         struct fragment_container *prev = NULL;

         /* Find sorted position in free list */
         {
            struct fragment_container *first_after = search_start;
            
            INIT_FRAG_TTL;
            
            while (first_after &&
                   (FC_GET_BASE(first_after) < FC_GET_BASE(f))) {
               prev = first_after;
               first_after = FC_GET_NEXT(first_after);
               CHECK_FRAG_TTL;
            }
         }

         id->free_size += FC_GET_SIZE(f);
         curr_bank->free_size += FC_GET_SIZE(f);
         FREE_BITS(FC_GET_BASE(f), FC_GET_SIZE(f));

         /* Insert in free list */
         if (prev == NULL) {

            /* Insert the fragment at the head of the list. */

            FC_SET_NEXT(f, curr_bank->free_mem);
            curr_bank->free_mem = f;
         }
         else {

            assert(FC_GET_TOP(prev) < FC_GET_BASE(f));

            /* Insert the fragment after prev. */

            FC_SET_NEXT(f, FC_GET_NEXT(prev));
            FC_SET_NEXT(prev, f);

            /* Join with preceeding fragment if possible. There is obviously
             * no idea to try this if the fragment was inserted first in the
             * list. */
            if (FC_GET_TOP(prev)+1 == FC_GET_BASE(f)) {
               FC_SET_TOP(prev, FC_GET_TOP(f));
               FC_SET_NEXT(prev, FC_GET_NEXT(f));
               free_fragment_container(curr_bank, f);
               f = prev;
            }
         }

         /* Join with succeeding fragment if possible */
         if (FC_GET_NEXT(f) && (FC_GET_TOP(f)+1 == FC_GET_BASE(FC_GET_NEXT(f)))) {
            struct fragment_container *tmp = FC_GET_NEXT(f);
            FC_SET_TOP(f, FC_GET_TOP(tmp));
            FC_SET_NEXT(f, FC_GET_NEXT(tmp));
            free_fragment_container(curr_bank, tmp);
         }

         search_start = f;
         
         f = list;
         if (f == NULL) {
            MM_MULTITHREAD_UNLOCK(sr);
            return; /* End of list was reached. */
         }

         list = FC_GET_NEXT(list);

#if FC_PACK_LEVEL != 2
         assert(((U32)f & 3) == 0);
#endif
         assert(FC_GET_BASE(f) < FC_GET_TOP(f));

         /* (search_start->base > f->top) was earlier
          * (curr_bank->base > f->top), but this version allows the function
          * to handle an unordered list within one bank.*/
         if (FC_GET_BASE(search_start) > FC_GET_TOP(f) || curr_bank->top < FC_GET_BASE(f)) {
            break; /* Search for new bank. */
         }
      }
   }
}


/*****************************************************************************
 *
 * Initially, the following was used:
 *
    struct mem_bank *bank_lists[] = {id->sas_banks, id->sase_banks};
    U32 i;

    for (i = 0; i < (sizeof bank_lists)/(sizeof bank_lists[0]); i++) {
      ...
    }

 * This was replaced since some compilers require that initializers are
 * constant.
 * Instead the function next_bank() is used:
 *

    struct mem_bank *bank_list;

    for (bank_list = next_bank_list(id, NULL);
         bank_list != NULL;
         bank_list = next_bank_list(id, bank_list))
    {
      ...
    }
*/

static struct mem_bank *
next_bank_list(const struct allocator_internal_data *id,
               const struct mem_bank *curr)
{
   if (curr == NULL) {
      if (id->sas_banks != NULL) {
         return id->sas_banks;
      }
      return id->sase_banks;
   }
   
   if (curr == id->sas_banks) {
      return id->sase_banks;
   }
   return NULL;
}

/*****************************************************************************/

static struct fragment_container *
alloc_fragment_first(struct allocator_internal_data *id,
                     struct mem_bank *bank_list,
                     phys_size_t size,
                     phys_size_t alignment,
                     U32 meta_options)

{
   struct mem_bank *b;
   Msr sr = 0;

   MM_MULTITHREAD_LOCK(sr);
   for (b = bank_list; b; b = b->next) {

      struct fragment_container *f;
      struct fragment_container *prev = NULL;

      for (f = b->free_mem; f; prev = f, f = FC_GET_NEXT(f)) {
         phys_addr_t base;

         base = ALIGN_UP(FC_GET_BASE(f), (phys_addr_t)alignment);

         /* This condition is necessary for the case where the
          * alignment pushes base above the top. */
         if (base < FC_GET_TOP(f)) {
            phys_addr_t top = base+size-1;
            struct fragment_container *fragment;

            if (top > FC_GET_TOP(f)) {
               top = FC_GET_TOP(f);
            }

            assert(top > base);


            if (meta_options & MM_REGION_OPT_PERSISTENT) {
               /* Avoid memory that originated from any initial region. */
               if (mm_initial_region_overlap(base, size)) {
                  continue;
               }
            }

            /* Now, we have a fragment that can be used. */
            fragment = extract_from_fragment(id, b, &f, base, top);

            /* Have to link in the updated fragment. */
            if (prev)
            {
               FC_SET_NEXT(prev, f);
            }
            else
            {
               b->free_mem = f;
            }
            
            MM_MULTITHREAD_UNLOCK(sr);
            return fragment;
         }
      }
   }
   MM_MULTITHREAD_UNLOCK(sr);
   return NULL;
}

/*****************************************************************************/

static struct fragment_container *
alloc_fragment_fixed(struct allocator_internal_data *id,
                     phys_addr_t base,
                     phys_size_t size)
{
   phys_addr_t top = base + size - 1;

   struct mem_bank *b = find_bank(id, base);


   /* Fixed allocations are used in start up phase. Must check
    * that list is initialized. */
   if (list_head(&(gmm_state.deleted_init_dynamic_regions)) != NULL) {
      /* Avoid memory that originated from any initial region. */
      if (mm_initial_region_overlap(base, size)) {
         return NULL;
      }
   }
   
   if (b) {
      struct fragment_container *f;
      struct fragment_container *prev = NULL;

      for (f = b->free_mem; f; prev = f, f = FC_GET_NEXT(f)) {
         struct fragment_container *tmp;
         
         if (base < FC_GET_BASE(f)) { continue; }
         if (top  > FC_GET_TOP(f))  { continue; }

         /* Now, we have a fragment that can be used. */
         tmp = extract_from_fragment(id, b, &f, base, top);

         /* Have to link in the updated fragment. */
         if (prev)
         {
            FC_SET_NEXT(prev, f);
         }
         else
         {
            b->free_mem = f;
         }
         
         return tmp;
      }
   }
   return NULL;
}

/*****************************************************************************/

static struct mem_bank *
bank_overlap(const struct allocator_internal_data *id,
             phys_addr_t base,
             phys_addr_t size)
{
   phys_addr_t top = base + size - (phys_addr_t)1;
   struct mem_bank *bank;

   assert(size == 0 || top >= base);
   for (bank = id->sas_banks; bank; bank = bank->next) {
      if (bank->base > top)  { continue; }
      if (bank->top  < base) { continue; }

      return bank;
   }

   for (bank = id->sase_banks; bank; bank = bank->next) {
      if (bank->base > top)  { continue; }
      if (bank->top  < base) { continue; }

      return bank;
   }

   return NULL;
}

/*****************************************************************************/

static struct mem_bank *
bank_subarea(const struct allocator_internal_data *id,
             phys_addr_t base,
             phys_addr_t size)
{
   phys_addr_t top = base+size-1;
   struct mem_bank *bank;

   assert(top >= base);
   for (bank = id->sas_banks; bank; bank = bank->next) {
      if (bank->base > base) { continue; }
      if (bank->top  < top)  { continue; }
      return bank;
   }

   for (bank = id->sase_banks; bank; bank = bank->next) {
      if (bank->base > base) { continue; }
      if (bank->top  < top)  { continue; }
      return bank;
   }

   return NULL;
}

/*****************************************************************************/

#ifdef ASSERT
static phys_size_t
calc_free_mem(const struct allocator_internal_data *id)
{
   phys_size_t sz = 0;
   struct mem_bank *b;
   Msr sr = 0;
   INIT_BANK_TTL;

   MM_MULTITHREAD_LOCK(sr);
   b = id->sas_banks;
   while (b) {
      struct fragment_container *f = b->free_mem;
      while (f) {
         INIT_FRAG_TTL;

         sz += FC_GET_SIZE(f);
         f = FC_GET_NEXT(f);
         CHECK_FRAG_TTL;
      }
      b = b->next;
      CHECK_BANK_TTL;
   }

   b = id->sase_banks;
   while (b) {
      struct fragment_container *f = b->free_mem;
      while (f) {
         INIT_FRAG_TTL;

         sz += FC_GET_SIZE(f);
         f = FC_GET_NEXT(f);
         CHECK_FRAG_TTL;
      }
      b = b->next;
      CHECK_BANK_TTL;
   }
   MM_MULTITHREAD_UNLOCK(sr);
   return sz;
}
#endif

/*****************************************************************************/

static phys_size_t
size_of_free_mem(const struct allocator *allocator, mm_bool_t sas) {
   struct allocator_internal_data *id = allocator->state;

   if (sas)
   {
      /* For SAS,
       * both memory not allocated to page-allocator
       * and that available for SASE is of interest. */
      return id->free_size;
   }
   else
   {

      /* FIXME, OSECR-9506, calculation of sase_banks->free_size
       * incorrect. It shows only in NO MMU case, for example cortex-r4. 
       * The gmm remake of MM_CHECK_LOW_LIMIT (the region is now created
       * before check) eliminated the need for free_mem function separating 
       * on SAS or SASE. Even though this function is not needed anymore
       * it would be interesting to find the root cause of the incorrect value
       * in  id->sase_banks->free_size
       * Note that on ppc this problem does not seem to occur. */
#ifndef HOST_TEST 
      assert(0);
#endif
      return id->sase_banks->free_size;
      
   }
}

/*****************************************************************************
 *
 * Free memory allocated by this allocator.
 */
static void
mem_free(const struct allocator   *allocator,
         struct phys_object *phys_object)
{
   Msr sr = 0;
   assert(allocator   != NULL);
   assert(phys_object != NULL);

   switch (phys_object->data_representation) {
      case PHYS_OBJ_NULL:
         break;

      case PHYS_OBJ_FRAG_LIST:
      {
         MM_MULTITHREAD_LOCK(sr);
         assert(allocator->state->free_size == calc_free_mem(allocator->state));
         free_fragment_list(allocator->state,
                            &phys_object->data.frag_list);
         assert(allocator->state->free_size == calc_free_mem(allocator->state));
         MM_MULTITHREAD_UNLOCK(sr);
         break;
      }
      default:
         ramlog_printf("data_representation = 0x%x\n",
                       phys_object->data_representation);
         INTERNAL_ERROR();
   }
   phys_object->data_representation = PHYS_OBJ_NULL;
}

/******************************************************************************
 * Allocate continuous phys mem at fixed address.
 */
static mm_status_t
alloc_f(const struct allocator *allocator,
        struct fragment_container **fragment_container,
        phys_addr_t base,
        phys_size_t size)
{
   struct fragment_container *f;
   Msr sr = 0;

   log(CATEGORY_PAGE, ("{%s}->alloc_f(0x%08llx-0x%08llx)\n",
                       allocator->name, base, base+size-1));   
   
   MM_MULTITHREAD_LOCK(sr);
   assert(allocator != NULL);
   assert(allocator->state != NULL);
   assert(fragment_container != NULL);

   assert(allocator->state->free_size == calc_free_mem(allocator->state));
   f = alloc_fragment_fixed(allocator->state, base, size);
   assert(allocator->state->free_size == calc_free_mem(allocator->state));

   if (f == NULL) {
      *fragment_container = NULL;
      MM_MULTITHREAD_UNLOCK(sr);

      /* Reclaim is done at a higher level in gmm.c. */
      return MM_EOUT_OF_PHYS_MEM;
   }
   *fragment_container = f;

   MM_MULTITHREAD_UNLOCK(sr);
   return MM_SUCCESS;
}

/******************************************************************************
 * Allocate continuous memory.
 * Only called through an allocator struct.
 */
static mm_status_t
alloc_c(const struct allocator *allocator,
        struct phys_object *phys_object,
        phys_size_t size,
        phys_size_t alignend,
        phys_addr_t range_base,
        phys_size_t range_size)
{
   struct fragment_container *f;
   mm_status_t status = MM_SUCCESS;
   Msr sr = 0;

   log(CATEGORY_PAGE, ("{%s}->alloc_e(size:0x%08lx)\n",
                       allocator->name, size));

   MM_MULTITHREAD_LOCK(sr);
   assert(allocator != NULL);
   assert(allocator->state != NULL);
   assert(phys_object != NULL);

   assert(allocator->state->free_size == calc_free_mem(allocator->state));

   /* First try allocating from SAS banks. */
   f = alloc_fragment(allocator->state, allocator->state->sas_banks,
                      size, alignend, range_base, range_size,
                      phys_object->meta_options);
   assert(allocator->state->free_size == calc_free_mem(allocator->state));

   if (NULL == f)
   {
      /* Then try allocating from SASE banks */
      f = alloc_fragment(allocator->state, allocator->state->sase_banks,
                         size, alignend, range_base, range_size,
                         phys_object->meta_options);
      assert(allocator->state->free_size == calc_free_mem(allocator->state));
   }

   if (f == NULL) {
      phys_object->data_representation = PHYS_OBJ_NULL;
      MM_MULTITHREAD_UNLOCK(sr);

      /* Reclaim is done at a higher level in gmm.c. */
      return MM_EOUT_OF_PHYS_MEM;
   }

   assert(FC_GET_SIZE(f) == (phys_size_t)size);
   
   phys_object->data_representation = PHYS_OBJ_FRAG_LIST;
   phys_object->data.frag_list = f;

   MM_MULTITHREAD_UNLOCK(sr);
   return status;
}

/******************************************************************************
 * Allocate continuous memory where logical addresses equals physical
 * (i.e. from sase bank).
 * Only called through an allocator struct.
 */
static mm_status_t
alloc_e(const struct allocator *allocator,
        struct phys_object *phys_object,
        phys_size_t size,
        phys_size_t alignend,
        phys_addr_t range_base,
        phys_size_t range_size)
{
   struct fragment_container *f;
   mm_status_t status = MM_SUCCESS;
   Msr sr = 0;

   log(CATEGORY_PAGE, ("{%s}->alloc_e(size:0x%08lx)\n",
                       allocator->name, size));

   MM_MULTITHREAD_LOCK(sr);
   assert(allocator != NULL);
   assert(allocator->state != NULL);
   assert(phys_object != NULL);

   assert(allocator->state->free_size == calc_free_mem(allocator->state));
   f = alloc_fragment(allocator->state, allocator->state->sase_banks,
                      size, alignend, range_base, range_size,
                      phys_object->meta_options);
   assert(allocator->state->free_size == calc_free_mem(allocator->state));

   if (f == NULL) {
      phys_object->data_representation = PHYS_OBJ_NULL;
      MM_MULTITHREAD_UNLOCK(sr);

      /* Reclaim is done at a higher level in gmm.c. */
      return MM_EOUT_OF_PHYS_MEM;
   }

   assert(FC_GET_SIZE(f) == (phys_size_t)size);
   
   phys_object->data_representation = PHYS_OBJ_FRAG_LIST;
   phys_object->data.frag_list = f;

   MM_MULTITHREAD_UNLOCK(sr);
   return status;
}

/******************************************************************************
 * Allocate SAS memory.
 * Only called through an allocator struct.
 */

static log_addr_t
calc_log_frag_base(log_addr_t log_base, const struct fragment_container *f) {
   U32 page_index = FC_GET_PAGE_INDEX(f);
   log_addr_t alignment = gmm_state.hal_info->page_sizes[page_index]-1;
   return ((log_base+alignment) & ~alignment);
}

/**********
 * add_pages tries to fill the range [log_range_base-log_range_top] with
 * fragments with size and alignment = page_sizes[page_index]. 
 */

static void
add_pages(struct allocator_internal_data *id,
          struct mem_bank *bank_list,
          log_addr_t log_range_base,
          log_addr_t log_range_top,
          log_size_t page_index,
          struct fragment_container **insert_pos,
          U32 meta_options)
{
   log_addr_t log_frag_base, log_frag_top;
   log_size_t page_size;
   struct fragment_container *prev = NULL;
   struct fragment_container *last_insert = NULL;
   Msr sr = 0;

   MM_MULTITHREAD_LOCK(sr);
   if (log_range_top < log_range_base) { /* Detect wrap. */
      MM_MULTITHREAD_UNLOCK(sr);
      return;
   }

   if (log_range_base == log_range_top + 1) { /* <=> size == 0. */
      MM_MULTITHREAD_UNLOCK(sr);
      return;
   }

   if (page_index == 0) {
      log_size_t size = log_range_top+1-log_range_base;
      log_size_t alignment = gmm_state.hal_info->page_sizes[0];

      while (size) {

         struct fragment_container *f
            = alloc_fragment_first(id, bank_list, size, alignment,
                                   meta_options);
        if (f == NULL) {
            break;
         }

         /*
         ** Splice the new element into the list, pushing the original
         ** entry pointed to by insert_pos ahead..
         */
         if (!last_insert)
         {
            FC_SET_NEXT(f, *insert_pos);
            *insert_pos = f;
            last_insert = f;
         }
         else
         {
            FC_SET_NEXT(f, FC_GET_NEXT(last_insert));
            FC_SET_NEXT(last_insert, f);
            last_insert = f;
         }

         assert((log_size_t)FC_GET_SIZE(f) == FC_GET_SIZE(f));
         size -= (log_size_t)FC_GET_SIZE(f);
      }
      MM_MULTITHREAD_UNLOCK(sr);
      return;
   }

   page_size = gmm_state.hal_info->page_sizes[page_index];

   log_frag_base = ALIGN_UP(log_range_base, page_size);
   if (log_frag_base < log_range_base) { /* Detect wrap. */
      MM_MULTITHREAD_UNLOCK(sr);
      return;
   }
   
   /* log_frag_top is the top of a fragment starting at log_range_base. */
   log_frag_top = log_frag_base + page_size-1;

   while (log_range_top >= log_frag_top && log_frag_top > log_range_base) {
      struct fragment_container *f = alloc_fragment(id,
                                                    bank_list,
                                                    page_size,
                                                    page_size, 
                                                    0ULL, 0ULL, meta_options);
      if (f == NULL) {
         break;
      }

      if (prev != NULL && FC_GET_TOP(prev)+1 == FC_GET_BASE(f)) {
         /* The new fragment can be joined with the previous. */
         FC_SET_TOP(prev, FC_GET_TOP(f));
         free_fragment_container(find_bank(id, FC_GET_BASE(f)), f);
      }
      else {

         if (!last_insert)
         {
            FC_SET_NEXT(f, *insert_pos);
            *insert_pos = f;
            last_insert = f;
         }
         else
         {
            FC_SET_NEXT(f, FC_GET_NEXT(last_insert));
            FC_SET_NEXT(last_insert, f);
            last_insert = f;
         }
         FC_SET_PAGE_INDEX(f, page_index);
         prev = f;
      }
      
      log_frag_base += page_size;
      log_frag_top  += page_size;
   }
   MM_MULTITHREAD_UNLOCK(sr);
}

/**********
 * Helper to alloc_s. 
 */

static void
alloc_s_(struct allocator_internal_data *id,
         struct mem_bank *bank_list,
         log_addr_t log_base,
         log_addr_t log_top,
         struct fragment_container **frag_list,
         U32 meta_options)
{
   U32 page_index;
   Msr sr = 0;

   /* The exit criteria (page_index < ~PAGE_INDEX_MASK) is
    * (page_index >= 0) modified to handle the fact that
    * page_index is unsigned (from struct hal_info). */
   MM_MULTITHREAD_LOCK(sr);
   for (page_index = gmm_state.hal_info->no_page_sizes-1;
        page_index < ~PAGE_INDEX_MASK;
        page_index--)
   {
      log_addr_t log_range_base = log_base;
      struct fragment_container *curr = *frag_list;
      struct fragment_container *prev = NULL;
      while (curr != NULL)
      {
         struct fragment_container *next = FC_GET_NEXT(curr);
         log_size_t size = FC_GET_TOP(curr) + 1ULL - (FC_GET_BASE(curr) & ~(phys_addr_t)PAGE_INDEX_MASK);
         log_addr_t log_range_top =
            calc_log_frag_base(log_range_base, curr) - 1;
         add_pages(id, bank_list,
                   log_range_base, log_range_top,
                   page_index, &curr, meta_options);         
         log_range_base = log_range_top + 1 + size;

         /* curr may now have been updated, link it into the
          * original list. */
         if (!prev)
         {
             prev = *frag_list;
            *frag_list = curr;
         }
         else
         {
            FC_SET_NEXT(prev, curr);
            prev = curr;
         }
         curr = next;

         /* If one or more fragments are inserted by add_page immediately after prev,
          * prev must be stepped past these new fragments so that it points at the
          * correct fragment previous to curr in the next lap.
          */
         while (FC_GET_NEXT(prev) != curr)
         {
            prev = FC_GET_NEXT(prev);
         }
      }

      add_pages(id, bank_list, log_range_base, log_top, page_index, &curr,
                meta_options);
      if (!prev)
      {
         *frag_list = curr;
      }
      else
      {
         FC_SET_NEXT(prev, curr);
      }
   }
   MM_MULTITHREAD_UNLOCK(sr);
}

/**********/

static mm_status_t
alloc_s(const struct allocator *allocator,
        struct phys_object *phys_object,
        phys_size_t req_size,
        phys_addr_t log_base)
{
   log(CATEGORY_PAGE, ("{%s}->alloc_s(size:0x%08lx, log base:0x%lx)\n",
                       allocator->name, req_size, log_base));
   
   assert(allocator != NULL);
   assert(phys_object != NULL);
   
   /* The fragments returned by alloc_fragment_first have .next = NULL, but
    * for the case where no fragments at all could be allocated, this
    * initiation is necessary. */
   phys_object->data.frag_list = NULL;
   
   assert(allocator->state->free_size == calc_free_mem(allocator->state));

   alloc_s_(allocator->state, allocator->state->sas_banks,
            log_base, log_base+req_size-1, &phys_object->data.frag_list,
            phys_object->meta_options);
   
   assert(allocator->state->free_size == calc_free_mem(allocator->state));

   alloc_s_(allocator->state, allocator->state->sase_banks,
            log_base, log_base+req_size-1, &phys_object->data.frag_list,
            phys_object->meta_options);
   
   assert(allocator->state->free_size == calc_free_mem(allocator->state));
   
   {
      phys_size_t actual_size = 0;
      struct fragment_container *f;
      
       for (f = phys_object->data.frag_list; f; f = FC_GET_NEXT(f)) {
         FC_SET_PAGE_INDEX(f, 0);
         actual_size += (phys_size_t)FC_GET_SIZE(f);
      }
      
      if (req_size != actual_size) {
         free_fragment_list(allocator->state, &phys_object->data.frag_list);
         phys_object->data_representation = PHYS_OBJ_NULL;

         /* Reclaim is done at a higher level in gmm.c. */
         return MM_EOUT_OF_PHYS_MEM;
      }
   }
   
   phys_object->data_representation = PHYS_OBJ_FRAG_LIST;
   
   return MM_SUCCESS;
}

/******************************************************************************
 * Write to RAM memory.
 * Only called through an allocator struct.
 */
static mm_status_t
mem_write(struct allocator *allocator,
          log_addr_t source,
          log_addr_t destination,
          log_size_t size)
{

   (void) allocator;  /* Not needed in this allocator. */

   memcpy((void *) destination, (const void *) source, size);
   return MM_SUCCESS;
}


/*****************************************************************************
 */
static phys_size_t
get_tot_phys_mem(const struct allocator *a) {
   struct allocator_internal_data *id = a->state;
   phys_size_t ack_size = 0;

   struct mem_bank *bank_list;

   assert(a);
   for (bank_list = next_bank_list(id, NULL);
        bank_list;
        bank_list = next_bank_list(id, bank_list))
   {

      struct mem_bank *b;

      for (b = bank_list; b; b = b->next) {
         ack_size += b->top+1;
         ack_size -= b->base;
      }
   }
   return ack_size;
}

/*****************************************************************************/

static phys_size_t
get_free_phys_mem(const struct allocator *a) {
   struct allocator_internal_data *id = a->state;

   return id->free_size;
}

/*****************************************************************************
 *
 * The physical address space might be 64 bits, which (with a page size of 4K)
 * vould require 2^54 fragments. The fragments are however be allocated before
 * the MMU is enabled, and at that time, the size of the addressable space
 * is 32 bits. The number of fragments can therefore not exceed the range of
 * an U32.
 */

static U32
get_phys_frag_count(const struct allocator *a) {
   struct allocator_internal_data *id = a->state;
   U32 cnt = 0;

   struct mem_bank *bank_list;

   assert(a);
   for (bank_list = next_bank_list(id, NULL);
        bank_list;
        bank_list = next_bank_list(id, bank_list))
   {

      struct mem_bank *b;

      for (b = bank_list; b; b = b->next) {
         struct fragment_container *f;
         for (f = b->free_mem; f; f = FC_GET_NEXT(f))
         {
            cnt++;
         }
      }
   }
   return cnt;
}

/*****************************************************************************/

static phys_size_t
get_biggest_phys_frag(const struct allocator *a) {
   struct allocator_internal_data *id = a->state;
   phys_size_t max_size = 0;

   struct mem_bank *bank_list;

   assert(a);
   for (bank_list = next_bank_list(id, NULL);
        bank_list;
        bank_list = next_bank_list(id, bank_list))
   {

      struct mem_bank *b;

      for (b = bank_list; b; b = b->next) {

         struct fragment_container *f;

         for (f = b->free_mem; f; f = FC_GET_NEXT(f)) {
            phys_addr_t size = FC_GET_SIZE(f);
            if (size > max_size) {
               max_size = size;
            }
         }
      }
   }
   return max_size;
}

/*****************************************************************************/

static phys_size_t
get_free_watermark(const struct allocator *allocator)
{
   return allocator->state->watermark;
} /* get_free_watermark(...) */

/*****************************************************************************/

static phys_size_t
get_metadata_size(const struct allocator *a)
{
   struct allocator_internal_data *id = a->state;
   U32 cnt = 0;

   struct mem_bank *bank_list;

   assert(a);
   for (bank_list = next_bank_list(id, NULL);
        bank_list;
        bank_list = next_bank_list(id, bank_list))
   {

      struct mem_bank *b;

      for (b = bank_list; b; b = b->next)
      {
         cnt += b->frag_cnt;
      }
   }

   return (phys_size_t)cnt * sizeof(struct fragment_container);
}

/*****************************************************************************
 * Process a TAG signal.
 * Only called through an allocator struct.
 */
static void
mem_info(struct allocator *allocator,
         union SIGNAL     *tag_sig,
         union SIGNAL     **result)
{
   unsigned tag_index;
   struct mm_get_phys_mem_info *request;
   struct mm_tag *tags;
   /*
     struct container container;
   */
   assert(NULL != allocator);
   assert(NULL != tag_sig);
   assert(NULL != result);

   request = &tag_sig->mm_get_phys_mem_info;
   tags = &request->tags[0];

   /* Always return <tag_sig> as out-parameter. */
   *result = tag_sig;

   /* Traverse tag list. */
   for (tag_index = 0; tag_index < request->count; tag_index++)
   {
      enum MmTagId tag_id = (enum MmTagId) tags[tag_index].tag_id;

      switch (tag_id)
      {
         /* FIXME.
          *   Introduction of support for 64 bit addresses makes the 32 bit
          *   answer field too small. One solution is to change the type of
          *   (struct mm_tag).tag_value to unsigned long long.
          */
         case MM_TAG_INFO_PHYS_MEM_TOTAL_MEMORY:
         {
            phys_size_t sz = get_tot_phys_mem(allocator);
            if ((U32)sz != sz) {
               sz = (U32)-1;
            }
            tags[tag_index].tag_value = (U32)sz;
            break;
         }
         case MM_TAG_INFO_PHYS_MEM_FREE_MEMORY:
         {
            phys_size_t sz = get_free_phys_mem(allocator);
            if ((U32)sz != sz) {
               sz = (U32)-1;
            }
            tags[tag_index].tag_value = (U32)sz;
            break;
         }
         case MM_TAG_INFO_PHYS_MEM_FRAG_COUNT:
         {
            tags[tag_index].tag_value = get_phys_frag_count(allocator);
            break;
         }
         case MM_TAG_INFO_PHYS_MEM_BIGGEST_FRAG:
         {
            phys_size_t sz = get_biggest_phys_frag(allocator);
            if ((U32)sz != sz) {
               sz = (U32)-1;
            }
            tags[tag_index].tag_value = (U32)sz;
            break;
         }
         case MM_TAG_INFO_PHYS_MEM_INACTIVE_MEMORY: /* Deprecated */
         {
            tags[tag_index].tag_value = 0;
            break;
         }
         default:
         {
            request->err_index = tag_index;
            request->status = MM_EILLEGAL_TAG;
            return;
         }
      }

      request->status = MM_SUCCESS;
   }
}


/******************************************************************************
 * Check for overlap.
 *   Returns a nonzero value if least one address in [base .. base+size-1] is
 *   located in one of the allocator's banks. Otherwise 0 is returned.
 * Only called through an allocator struct.
 */
static mm_bool_t
mem_overlap(const struct allocator *allocator,
            phys_addr_t base,
            phys_size_t size)
{
   assert(allocator != NULL);

   return NULL != bank_overlap(allocator->state, base, size);
}

/******************************************************************************
 * Check for overlap.
 *   Like mem_overlap, except that base and size is updated with the
 *   overlapping bank's base and size.
 * Only called through an allocator struct.
 */
static mm_bool_t
mem_overlap2(const struct allocator *allocator,
             phys_addr_t *base,
             phys_size_t *size)
{
   struct mem_bank *b;

   assert(allocator != NULL);

   b = bank_overlap(allocator->state, *base, *size);

   if (NULL == b)
   {
      return FALSE;
   }

   *base = b->base;
   *size = b->top+1-b->base;

   return TRUE;
}


/******************************************************************************
 * Find allocator super set.
 * Only called through an allocator struct.
 */
static void
find_super_set(const struct allocator *allocator,
               phys_addr_t *base,
               phys_size_t *top)
{

   struct mem_bank *bank_list;
   const struct allocator_internal_data *id = allocator->state;
   phys_addr_t super_base =  MM_ULLONG_MAX;
   phys_addr_t super_top  = 0;

   assert(allocator != NULL);

   for (bank_list = next_bank_list(id, NULL);
        bank_list;
        bank_list = next_bank_list(id, bank_list))
   {
      struct mem_bank *bank;
         
      for (bank = bank_list; bank; bank = bank->next) {
         if (bank->base < super_base) { super_base = bank->base; }
         if (bank->top  > super_top)  { super_top  = bank->top;  }
      }
   }
   
   assert(super_base != MM_ULLONG_MAX && super_top != 0);

   *base = super_base;
   *top  = super_top;
}

/******************************************************************************
 * Check for sub area
 *   Returns a nonzero value if the entire range [base .. base+size-1] is
 *   located in one of the allocator's banks. Otherwise 0 is returned.
 * Only called through an allocator struct.
 */
static mm_bool_t
mem_sub_area(const struct allocator *allocator,
             phys_addr_t base,
             phys_size_t size)
{
   assert(allocator != NULL);

   return NULL != bank_subarea(allocator->state, base, size);
}


/******************************************************************************
 * The common part to add_sas_bank & add_sase_bank.
 */
static void
add_bank(struct allocator *allocator,
         phys_addr_t base,
         phys_size_t size,
         struct mem_bank **list)
{
   struct fragment_container *f;
   struct mem_bank *bank;
   log_size_t align = allocator->alignment;

   assert(allocator->state->free_size == calc_free_mem(allocator->state));

   bank = bank_overlap(allocator->state, base, size);
   if (bank) {
      phys_addr_t top = base + size - 1;
      MM_FATAL_ERROR2(
         ("Allocator \"%s\": Attempt to add a bank [0x%08llx-0x%08llx] that"
          " overlaps an existing bank [0x%08llx-0x%08llx].\n",
          allocator->name, base, top, bank->base, bank->top),
         OSE_MM_ECONFIGURATION_ERROR,
         (OSERRCODE)("Attempt to add a bank that overlaps an existing bank."));
   }

   {
      unsigned long long  frag_cnt;
      phys_size_t bank_size;

      /* frag_cnt is the number of containers that shall be pre allocated.
       * (container is here used as short name for struct fragment container.)
       * Currently set to the teoretical max value of memory fragments
       * (= total memory/smallest handled memory size).
       * The reason for pre allocating so many containers is to ensure that a
       * warm start can be performed without allocating new containers. If
       * warm start is not used, allocating on demand could be considered.
       */
      frag_cnt = (size+align-1)/align;
      bank_size = sizeof *bank;

      /* One container is located in the bank struct. */
      bank_size += (frag_cnt-1) * (sizeof *f);
      VERBOZE(("MM:add_bank: name %s [%#08llx-%#08llx]\n"
               "   bank_size %#08llx, frag_cnt%#08llx,"
               "sizeof *bank,%#08lx (sizeof *frag)%#08lx \n",
               allocator->name, base, size, bank_size, frag_cnt,
               sizeof *bank,(sizeof *f)));
      
      /* There is an upper limit of 4G during the startup before the mmu
       * is enabled. */
      if (bank_size != (log_size_t)bank_size) {
         MM_FATAL_ERROR2(
            ("Allocator \"%s\": Representation of [0x%08llx-0x%08llx] requires"
             " too much memory (%llu bytes).\n",
             allocator->name, base, base+size-1, bank_size),
            OSE_MM_ECONFIGURATION_ERROR, (OSERRCODE) 0);
      }
      
      bank = boot_alloc((size_t)bank_size);
      bank->unlinked_containers_cnt = (U32)frag_cnt;
      bank->next = NULL;
      bank->free_mem = NULL;
      bank->unused_containers = NULL;
      bank->frag_cnt = frag_cnt;

#ifdef ASSERT
      max_frag_cnt += bank->unlinked_containers_cnt;
      max_bank_cnt++;
#endif
   }

   if (base & (align-1)) {
      MM_FATAL_ERROR2(
         ("Allocator \"%s\": Bank base (0x%08llx) is not properly aligned."
          " Required alignment: 0x%lx\n", allocator->name, base, align),
         OSE_MM_ECONFIGURATION_ERROR,
         (OSERRCODE)("Bank base is not properly aligned."));
   }
   if (size & (align-1)) {
      MM_FATAL_ERROR2(
         ("Allocator \"%s\": Bank size (0x%08llx) is not properly aligned."
          " Required alignment: 0x%lx\n", allocator->name, size, align),
         OSE_MM_ECONFIGURATION_ERROR,
         (OSERRCODE)("Bank size is not properly aligned."));
   }

   bank->base = base;
   bank->top = base+size-1;

   f = alloc_fragment_container(bank);
   FC_SET_BASE(f, base);
   FC_SET_TOP(f, base + size - 1);
   bank->free_mem = f;

   allocator->state->free_size += size;
   bank->free_size += size;
   allocator->state->watermark += size;

   {
      /* Insert bank in bank list ordered by base address. */
      INIT_FRAG_TTL;
      while (*list && (*list)->base < base) {
         list = &(*list)->next;
         CHECK_FRAG_TTL;
      }
      bank->next = *list;
      *list = bank;
   }
   if (mm_get_log_mask() & CATEGORY_PAGE) {
      dump_state(allocator, (int (*)(const char *,...))ramlog_printf);
   }

   FREE_BITS(base, size);
   assert(allocator->state->free_size == calc_free_mem(allocator->state));
}


/*****************************************************************************/
static void
add_sas_bank(struct allocator *allocator,
                         phys_addr_t base, phys_size_t size)
{
   add_bank(allocator, base, size, &allocator->state->sas_banks);
}

/*****************************************************************************/

static void
add_sase_bank(struct allocator *allocator,
                          log_addr_t base, log_size_t size)
{
   phys_addr_t top = (phys_addr_t)base + (phys_addr_t)size - 1;

   if (top != (log_addr_t)top) {
      MM_FATAL_ERROR2(
         ("Attempt to add SASE bank ([0x%08lx-0x%08llx]) that exceeds the"
          " range of the logical memory space.\n", base, top),
         OSE_MM_ECONFIGURATION_ERROR,
         (OSERRCODE)("Attempt to add SASE bank that exceeds the range "
                     "of the logical memory space"));
   }

   add_bank(allocator, base, size, &allocator->state->sase_banks);
}

/*****************************************************************************/

static int
get_bank_info(const struct allocator *allocator,
              int index, phys_addr_t *base, phys_size_t *size)
{
   struct mem_bank *bank_list;
   const struct allocator_internal_data *id = allocator->state;
   struct mem_bank *bank;
   int count = 0;

   assert(allocator != NULL);

   for (bank_list = next_bank_list(id, NULL);
        bank_list;
        bank_list = next_bank_list(id, bank_list))
   {
      for (bank = bank_list; bank; bank = bank->next, count++) {
         if (count == index)
         {
            *base = bank->base;
            *size = bank->top - bank->base + 1;
            return 1;
         }
      }
   }
   
   return 0;   
}

/**************************************************************************/

struct allocator *
mm_create_ram_allocator(const char *name) {
   struct allocator *ram;
   struct allocator_internal_data *id;
   log_size_t min_page_size = gmm_state.hal_info->page_sizes[0];

   if (~PAGE_INDEX_MASK & gmm_state.hal_info->no_page_sizes) {
      INTERNAL_ERROR();
   }
   
   if (PAGE_INDEX_MASK & min_page_size) {
      INTERNAL_ERROR();
   }
   
   if (min_page_size == 0) {
      MM_FATAL_ERROR2(
         ("Smallest page size is 0.\n"),
         OSE_MM_ECONFIGURATION_ERROR, (OSERRCODE)("Smallest page size is 0."));
   }
   if (min_page_size & (min_page_size-1)) {
      MM_FATAL_ERROR2(
         ("Smallest page size (0x%lx) is not a power of 2.\n", min_page_size),
         OSE_MM_ECONFIGURATION_ERROR,
         (OSERRCODE)("Smallest page size is not a power of 2."));
   }

   ram = boot_alloc(sizeof *ram);
   id  = boot_alloc(sizeof *id);


   log(CATEGORY_PAGE, (" addr of ram: %#08lx sizeof *ram: %#08lx\n",
                       ram, sizeof *ram));       
   log(CATEGORY_PAGE, (" addr of id: %#08lx sizeof *id: %#08lx\n",
                       id, sizeof *id));
   
   id->sas_banks  = NULL;
   id->sase_banks = NULL;

   strcpy(ram->name, name);
   ram->add_sas_bank     = add_sas_bank;
   ram->add_sase_bank    = add_sase_bank;
   ram->alignment        = min_page_size;
   ram->alloc_c          = alloc_c;
   ram->alloc_e          = alloc_e;
   ram->alloc_f          = alloc_f;
   ram->alloc_s          = alloc_s;
   ram->free             = mem_free;
   ram->info             = mem_info;
   ram->get_total_mem    = get_tot_phys_mem;
   ram->get_free_mem     = get_free_phys_mem;
   ram->get_frag_count   = get_phys_frag_count;
   ram->get_biggest_frag = get_biggest_phys_frag;
   ram->get_free_watermark = get_free_watermark;
   ram->overlap          = mem_overlap;
   ram->overlap2         = mem_overlap2;
   ram->sub_area         = mem_sub_area;
   ram->find_super_set   = find_super_set;
   ram->size_of_free_mem = size_of_free_mem;
   ram->state            = id;
   ram->write            = mem_write;
   ram->dump_state       = dump_state;
   ram->get_metadata_size= get_metadata_size;
   ram->get_bank_info    = get_bank_info;

   /* Insert ram allocator into global list of allocators. */
   mm_register_allocator(ram);

   /* Cache the default allocator, since this is used for all MM meta-data. */
   if (!strcmp(name, "RAM"))
   {
      gmm_state.ram_allocator = ram;
   }

   INIT_BITS;
   
   id->watermark = 0;
   id->free_size = 0;

   return ram;
}

/*****************************************************************************
 * Range 1 is the area below the boot heap
 * Range 2 is the boot heap
 * Range 3 is the area above the boot heap
 * The idea is to save the boot heap area as long as possible then allocating
 * pages through mm_page_alloc.
 */

static phys_addr_t page_range_1_base;
static phys_size_t page_range_1_size;
static phys_addr_t page_range_2_base;
static phys_size_t page_range_2_size;
static phys_addr_t page_range_3_base;
static phys_size_t page_range_3_size;
static U32        page_alloc_allowed = 0;

void
set_boot_heap_range(log_addr_t base, log_size_t size) {
   if (size == 0) {
      ramlog_printf("Attempt to set boot heap range with size == 0.\n");
      INTERNAL_ERROR();
   }
   
   page_range_1_base = 0;
   page_range_1_size = base;
   page_range_2_base = base;
   page_range_2_size = size;
   page_range_3_base = base + size;
   page_range_3_size = (1ULL<<32) - page_range_3_base;
   
   page_alloc_allowed = 1;
}

/*****************************************************************************/

static int static_phys_mem_reserved = 0;
void report_static_phys_mem_reserved(void) {
   static_phys_mem_reserved = 1;
}

/******************************************************************************
 * Free a page allocated by mm_page_alloc.
 */
void
mm_page_free(log_addr_t base, size_t size) {
   struct fragment_container *f;
   struct fragment_container *prev = NULL;
   Msr sr = 0;

   log(CATEGORY_PAGE, ("mm_page_free(base:0x%08lx, size:0x%08lx)\n",
                       base, size));

   MM_MULTITHREAD_LOCK(sr);
   for (f = page_list; f; prev = f, f = FC_GET_NEXT(f)) {
      if (FC_GET_BASE(f) == base) {
         struct fragment_container *tmp = f;

         assert(FC_GET_SIZE(f) == (phys_size_t)size);
         if (prev)
         {
            FC_SET_NEXT(prev, FC_GET_NEXT(f));
         }
         else
         {
            page_list = FC_GET_NEXT(f);
         }
#if FC_PACK_LEVEL < 2
         assert(((U32)tmp & 3) == 0);
#endif

         /* Allow f to be interpreted as a one element list. */
         FC_SET_NEXT(tmp, NULL);

         free_fragment_list(gmm_state.ram_allocator->state, &tmp);

         MM_MULTITHREAD_UNLOCK(sr);
         return;
      }
   }
   MM_MULTITHREAD_UNLOCK(sr);
   INTERNAL_ERROR();
}

/******************************************************************************
 * Allocate a page of size <size>.
 * The fragment containers (returned from alloc_fragment) are not passed to
 * the caller, and to prevent them from being lost, they are saved in a list.
 */
mm_bool_t
mm_page_alloc(size_t size, log_addr_t *base_)
{
   struct allocator_internal_data *id;
   struct fragment_container *f;
   Msr sr;
   size_t reclaimed_size;

restart_page_alloc:
   f = NULL;
   sr = 0;

   if (!page_alloc_allowed) {
      return FALSE;
   }
   
   /* Don't do this if page_alloc_allowed is false. gmm_state.ram_allocator
    * might be NULL.
    */
   assert(gmm_state.ram_allocator);
   id = gmm_state.ram_allocator->state;
   
#if 0
   /* FIXME. Currently some ARM targets relies on the assumtion that page_alloc
    * uses only memory from RAM_SASE. Seting start of RAM_SASE to 4K is assumed
    * to prevent page_alloc from using that memory.
    *
    * Extract fom the conf file:
    *
    *    krn/phys_mem/RAM/1=base:0x0 size:56M
    *    krn/region/exceptions_high=base:0xFFFF0000 phys_base:0 size:4K ...
    *    krn/log_mem/RAM_SASE=base:0x00001000 size:31M mapping_type:SASE
    */
   
   MM_MULTITHREAD_LOCK(sr);
   f = alloc_fragment(id, id->sas_banks, size, size,
                      page_range_1_base, page_range_1_size, 0);
   if (f != NULL) {
      MM_MULTITHREAD_UNLOCK(sr);
      goto done_ok; /*lint !e801 */
   }
   
   f = alloc_fragment(id, id->sas_banks, size, size,
                      page_range_2_base, page_range_2_size, 0);
   
   if (f != NULL) {
      MM_MULTITHREAD_UNLOCK(sr);
      goto done_ok; /*lint !e801 */
   }

   MM_MULTITHREAD_UNLOCK(sr);
#else
   (void)page_range_1_base;
   (void)page_range_1_size;
   (void)page_range_3_base;
   (void)page_range_3_size;
#endif
   
   /* Dont use the memory outside the boot heap before the physical resources
    * for the static regions have been allocated. */
   MM_MULTITHREAD_LOCK(sr);
   if (static_phys_mem_reserved) {
      if (size & (size-1))
      {
         /* If we want to allocate a size that is not a power of two,
          * only demand page alignment. */
         f = alloc_fragment(id, id->sase_banks, (phys_size_t)size,
                            gmm_state.hal_info->page_sizes[0],
                            0ULL, 0x100000000ULL, 0);
      }
      else
      {
         f = alloc_fragment(id, id->sase_banks, (phys_size_t)size, size,
                            0ULL, 0x100000000ULL, 0);
      }
      if (f != NULL) {
         MM_MULTITHREAD_UNLOCK(sr);
         goto done_ok; /*lint !e801 */
      }
   }
   else {
      if (size & (size-1))
      {
         /* If we want to allocate a size that is not a power of two,
          * only demand page alignment. */
         f = alloc_fragment(id, id->sase_banks, (phys_size_t)size,
                            gmm_state.hal_info->page_sizes[0],
                            page_range_2_base, page_range_2_size, 0);
      }
      else
      {
         f = alloc_fragment(id, id->sase_banks, (phys_size_t)size, size,
                            page_range_2_base, page_range_2_size, 0);
      }
      if (f != NULL) {
         MM_MULTITHREAD_UNLOCK(sr);
         goto done_ok; /*lint !e801 */
      }
   }


   MM_MULTITHREAD_UNLOCK(sr);


    /* By repeated reclaims we try to find a large enough memory chunk.
    */
   while (1)
   {
      reclaimed_size = ose_mm_reclaim_component_memory(&gmm_state,
                                                       size,
                                                       MM_INTERNAL_RAM);
      if (reclaimed_size == 0) {
         log(CATEGORY_PAGE, ("mm: page_alloc() could not find size %#08lx\n",
                             size));
         break;
      }
      else if (reclaimed_size >= size){
         goto restart_page_alloc;
      }
   }
   
   return FALSE;
      
  done_ok:
   
   {
      Msr sr = 0;
      
      MM_MULTITHREAD_LOCK(sr);

      /* Store the fragment container in the list page_list. */
      FC_SET_NEXT(f, page_list);
      page_list = f;
      MM_MULTITHREAD_UNLOCK(sr);
   }
      
   assert((log_addr_t)FC_GET_BASE(f) == FC_GET_BASE(f));

   *base_ = (log_addr_t)FC_GET_BASE(f);

   return TRUE;
}

/*****************************************************************************/
