/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2007 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

#include <string.h>
#include "ose.h"
#include "mm_err.h"
#include "pmm.sig"
#include "ramlog.h"
#include "shell.h"
#include "stdio.h"
#include "stdlib.h"
#include "sysparam.h"
#include "mm.sig"
#include "cache.h"

#define RPMM_NAME         ose_ram_pmm
#define RPMM_REGION_NAME "rpmm"
#define RPMM_MAGIC        0x41514513
#define BUILD_MAX         80

#define LOG       ramlog_printf("PMM: "), ramlog_printf
#define LOG_ERROR ramlog_printf("PMM ERROR: "),\
                  ramlog_printf("%s:%d\n  ", __FILE__, __LINE__),\
                  ramlog_printf

/* Used in the PMM_ASSIGN signal. */
#define PMM_MEMORY_TYPE "RAM"

extern OSENTRYPOINT RPMM_NAME;

/****************************************/

union SIGNAL
{
  SIGSELECT sigNo;
  struct pmm_alloc_sas_request       alloc_sas_request;
  struct pmm_alloc_sase_request      alloc_sase_request;
  struct pmm_assign                  assign;
  struct pmm_change                  change;
  struct pmm_corrupt_metadata        corrupt_metadata;
  struct pmm_free                    free;
  struct pmm_restore_binding_request restore_binding_request;
  struct pmm_restore_region_reply    restore_region_reply;
  struct pmm_restore_region_request  restore_region_request;
  struct pmm_start_reply             start_reply;
  struct pmm_start_request           start_request;
  struct pmm_version                 version;
  struct pmm_write                   write;
};

/****************************************/

struct regionListElement
{
  struct regionListElement *next;
  log_addr_t               base;
  int                      no_fragments;
  char                     metadata[PMM_MAX_METALEN];
  struct phys_resource     fragments[1];
};

/****************************************/

struct restartData
{
  U32  magic;
  char build_string[BUILD_MAX];
  int  reg_cnt;
  char data[1];
};

/****************************************/

static struct restartData       *rpmm;
static U32                      max_nr_of_regions;
static char                     *rpmmMax;
static U32                      status;
static char                     last_build[BUILD_MAX];
static PROCESS                  rpmm_pid_;
static struct regionListElement *regionList = 0;
static size_t                    usedSize;
static size_t                    usedSizePeak;

/****************************************/

/* Some macros to create string literals from macros. */
#define STR(x) #x
#define STRSTR(x) STR(x)

#ifdef BUILD_INFO
static const char build_info[] = STRSTR(BUILD_INFO);
#else
static const char build_info[] = "no BUILD_INFO set.";
#endif

/****************************************/

static void insertInRegionList(struct regionListElement *new) {
  struct regionListElement **tmp = &regionList;

  while (*tmp && (*tmp)->base < new->base) {
    tmp = &(*tmp)->next;
  }
  new->next = *tmp;
  *tmp = new;
}

/****************************************/

static struct regionListElement* mallocListElement(U32 nFrag) {
  size_t size = sizeof(struct regionListElement);

  /* Since fragments is an array of length 1, it is present
   * even if nFrag == 0. */
  if (nFrag) size += sizeof(struct phys_resource) * (nFrag - 1);

  /* FIXME.
   * Malloc here can cause deadlock if the system heap
   * is extended. Use alloc is one way to solve this problem.
   *
   */
  return malloc(size);
}

/****************************************/

static U32 percent(U32 x, U32 tot) {
  /* (100*x)/tot would be rounded down towards 0. For memory usage, rounding
   * up is preferred. */
  x = 100*(tot-x);
  x /= tot;
  return 100-x;
}

/****************************************/

static size_t regionDataSize(struct regionListElement *reg) {
  size_t size = strlen(reg->metadata) + 1;
  size += sizeof reg->no_fragments;
  size += sizeof reg->base;
  size += sizeof(U32); /* reg->metadata; */

  size += reg->no_fragments*(sizeof reg->fragments[0].base +
                             sizeof reg->fragments[0].size);
  return size;
}

/****************************************/

#define SAVE2(src, size)\
  if (data+size > rpmmMax) {\
    LOG_ERROR("RAM PMM region exhausted.\n");\
    error(PMM_EOUT_OF_MEMORY);\
  }\
  memcpy(data, &(src), size);\
  data += size

#define SAVE(src) SAVE2(src, sizeof src)

/*****/

static void rpmm_save(void)
{
  U32 elemCnt = 0;
  char *data = rpmm->data;

  while (regionList)
  {
    int i;
    struct regionListElement *srcReg = regionList;
    U32 metaLen = strlen(srcReg->metadata) + 1;
    regionList = regionList->next;

    /* - Save no_fragments first since it determines the size of the list
     *   element (1).
     * - Note that metaLen does not affect the size of the list element(1)
     *   since a fixed length (PMM_MAX_METALEN) buffer is used.
     *
     * (1) See struct regionListElement
     */
    SAVE(srcReg->no_fragments);
    SAVE(srcReg->base);
    SAVE(metaLen);
    SAVE2(srcReg->metadata, metaLen);

    for (i = 0; i < srcReg->no_fragments; i++) {
      /* SAVE(srcReg->fragments[i]) is not used since it would use 16 bytes
       * instead of 12.
       */
      SAVE(srcReg->fragments[i].base);
      SAVE(srcReg->fragments[i].size);
    }

    elemCnt++;
    free(srcReg);
  }

  rpmm->reg_cnt = elemCnt;

  cache_sync((OSADDRESS)rpmm, data-(char*)rpmm);
  /* This might seem like a good place to update usedSize & usedSizePeak.
   * However, putting it in rpmm_load is better because:
   * + At (warm) start, only rpmm_load is called.
   * + rpmm_save is only called from rpmm_sync, which also calls rpmm_load
   */
}

#undef SAVE

/****************************************/

#define LOAD2(dst, size)\
  memcpy(&(dst), data, size);\
  data += (size)

#define LOAD(dst) LOAD2(dst, sizeof dst)

/*****/

static void rpmm_load(void)
{
  int i;
  char *data = rpmm->data;

  for (i=0; i < rpmm->reg_cnt; i++)
  {
    int j;
    struct regionListElement *dstReg;
    U32 no_fragments;
    U32 metaLen;

    LOAD(no_fragments);
    dstReg = mallocListElement(no_fragments);
    dstReg->no_fragments = no_fragments;

    LOAD(dstReg->base);
    LOAD(metaLen);
    LOAD2(dstReg->metadata, metaLen);

    for (j=0; j < dstReg->no_fragments; j++) {
      LOAD(dstReg->fragments[j].base);
      LOAD(dstReg->fragments[j].size);
    }

    insertInRegionList(dstReg);
  }

  usedSize = data-(char*)rpmm;

  if (usedSize > usedSizePeak) {
    size_t totSize = rpmmMax - (char*)rpmm;

    LOG("New peak value for memory usage: %lu of %lu, (%lu%%)\n",
        usedSize, totSize, percent(usedSize, totSize));
    usedSizePeak = usedSize;
  }
}

#undef LOAD

/****************************************/

static void rpmm_sync(void)
{
  rpmm_save();
  rpmm_load();
}

/****************************************/

static int rpmm_list(void)
{
  struct regionListElement * reg = regionList;

  while (reg)
  {
    int i;

    printf("Persistent region: %s\n",reg->metadata);

    for (i = 0; i < reg->no_fragments; i++)
    {
      printf("    [0x%llx - 0x%llx] size: 0x%x\n",
             reg->fragments[i].base,
             reg->fragments[i].base+reg->fragments[i].size-1,
             reg->fragments[i].size);
    }
    reg = reg->next;
  }
  return RET_SUCCESS;
}

/****************************************/

static void rpmm_reset(void)
{
  while (regionList)
  {
    struct regionListElement *reg = regionList;
    regionList = regionList->next;
    free(reg);
  }

  strncpy(last_build, rpmm->build_string, BUILD_MAX-1);
  last_build[BUILD_MAX-1] = 0;

  rpmm->magic = RPMM_MAGIC;

  strncpy(rpmm->build_string,build_info,BUILD_MAX-1);
  rpmm->build_string[BUILD_MAX-1]=0;

  rpmm->reg_cnt = 0;
  usedSize = (size_t)&((struct restartData*)0)->data;
}

/****************************************/

static int print_status(void)
{
   if(status == PMM_START_WARM)
   {
      printf("The system recognized a warm start, regions were recovered.\n");
   }
   else
   {
      printf("The system was cold started, no recovery attempted.\n");
   }
#ifdef BUILD_INFO
   printf("Current image  : %s\n",build_info);
   if(strlen(last_build))
   {
      printf("Previous image : %s\n",last_build);
   }
#endif

   {
     size_t  totSize = rpmmMax - (char*)rpmm;

     printf("Memory status:\n"
            "  Current usage : %lu of %lu (%lu%%)\n"
            "  Peak usage    : %lu of %lu (%lu%%)\n",
            usedSize, totSize, percent(usedSize, totSize),
            usedSizePeak, totSize, percent(usedSizePeak, totSize));
   }

   printf("Max number of persistent regions : %lu\n", max_nr_of_regions);
   return RET_SUCCESS;
}

/****************************************/

/* About findElement():
 * 1) The return type is (X**) and not (X*) to make it possible to remove
 *    the element from the list using the return value. See ram_pmm_free().
 * 2) If the function returns, then the returned value is always a valid
 *    pointer. This makes things like x = *findElement(y); safe.
 */

static struct regionListElement
**findElement(struct regionListElement **list, U32 base) {
  while (*list) {

    /* There should be only one matching element in the list, so we can
     * leave at first match */
    if(base == (*list)->base) return list;

    list = &(*list)->next;
  }
  error2(PMM_EBAD_REGION_ADDRESS, base);
  return 0;
}

/****************************************/

static void rpmm_init(PROCESS *gmm, U32 *status) {
  const char *region = "krn/region/" RPMM_REGION_NAME;
  long base,size;

  if (hunt(GMM_NAME_STRING, 0, gmm, NULL) == 0) {
    LOG_ERROR("GMM process \"%s\" not found.\n", GMM_NAME_STRING);
    error(PMM_EGMM_NOT_FOUND);
  }

  if(sysparam_readattr_long(region, "base", &base)) {
    LOG_ERROR("Failed to get attribute base of configuration "
        "parameter %s.\n",  region);
    error(PMM_ECONFIGURATION);
  }

  if (sysparam_readattr_long(region, "size", &size)) {
    LOG_ERROR("Failed to get attribute size of configuration "
        "parameter %s.\n",  region);
    error(PMM_ECONFIGURATION);
  }

  rpmm = (struct restartData *) base;
  rpmmMax = (char*)(base+size);
  usedSizePeak = usedSize = (size_t)&((struct restartData*)0)->data;
  *status = PMM_START_WARM;

  if(rpmm->magic != RPMM_MAGIC) {
    LOG("Magic not found.\n");
    *status = PMM_START_COLD;
    return;
  }

  if (strncmp(rpmm->build_string, build_info, BUILD_MAX)) {
    LOG("ID is different.\n");
    *status = PMM_START_COLD;
    return;
  }
}

/****************************************/

static union SIGNAL*
ram_pmm_alloc_sase(struct pmm_alloc_sase_request *request)
{
   U8 *metadata = request->metadata;

   struct pmm_alloc_sase_reply * reply = (struct pmm_alloc_sase_reply * )
     alloc(sizeof(struct pmm_alloc_sase_reply), PMM_ALLOC_SASE_REPLY);

   struct regionListElement *newReg;

   /* LOG("ram pmm alloc sase base:0x%x\n", request->base); */

   if (rpmm->reg_cnt >= (int)max_nr_of_regions) {
     LOG_ERROR("Max number of persistent regions (%d) exceeded.\n",
               max_nr_of_regions);
     reply->ecode = PMM_ETOO_MANY_REGIONS;
     return (union SIGNAL*)reply;
   }

   if(request->metadata_size > PMM_MAX_METALEN) {
     LOG_ERROR("Length of metadata (\"%s\") is larger than PMM_MAX_METALEN"
               " (%lu)\n", metadata, PMM_MAX_METALEN);
     reply->ecode = OSE_PMM_ECORRUPT_METADATA;
     return (union SIGNAL*)reply;
   }

   newReg = mallocListElement(1);
   strcpy(newReg->metadata, (char *)metadata);

   newReg->no_fragments      = 1;
   newReg->base              = request->base;
   newReg->fragments[0].base = (phys_addr_t)request->base;
   newReg->fragments[0].size = request->region_size;

   if (regionDataSize(newReg) + usedSize <= (U32)(rpmmMax-(char*)rpmm)) {
     insertInRegionList(newReg);
     reply->ecode = PMM_EOK;
   }
   else {
     free(newReg);
     reply->ecode = PMM_EOUT_OF_MEMORY;
   }

   return (union SIGNAL*)reply;
}

/****************************************/

static union SIGNAL*
ram_pmm_alloc_sas(struct pmm_alloc_sas_request *request)
{
   U32 i;

   struct pmm_alloc_sas_reply * reply = (struct pmm_alloc_sas_reply *)
     alloc(sizeof(struct pmm_alloc_sas_reply) +
           sizeof(struct phys_resource) * (request->no_fragments - 1),
           PMM_ALLOC_SAS_REPLY);

   struct regionListElement *newReg;

   /* LOG("ram pmm alloc sas:base : 0x%08lx\n", request->base); */

   if (rpmm->reg_cnt >= (int)max_nr_of_regions) {
     LOG_ERROR("Max number of persistent regions (%d) exceeded.\n",
               max_nr_of_regions);
     reply->ecode = PMM_ETOO_MANY_REGIONS;
     return (union SIGNAL*)reply;
   }

   if(request->metadata_size > PMM_MAX_METALEN) {
     LOG_ERROR("Length of metadata (\"%s\") is larger than PMM_MAX_METALEN"
               " (%lu)\n", request->metadata, PMM_MAX_METALEN);
     reply->ecode = OSE_PMM_ECORRUPT_METADATA;
     return (union SIGNAL*)reply;
   }

   newReg = mallocListElement(request->no_fragments);

   strcpy(newReg->metadata, (char *)request->metadata);
   newReg->no_fragments = request->no_fragments;
   newReg->base = request->base;
   for (i=0; i < request->no_fragments; i++)
   {
      newReg->fragments[i] = request->fragments[i];
   }

   if (regionDataSize(newReg) + usedSize <= (U32)(rpmmMax-(char*)rpmm)) {
     insertInRegionList(newReg);
     reply->ecode = PMM_EOK;
   }
   else {
     free(newReg);
     reply->ecode = PMM_EOUT_OF_MEMORY;
   }

   return (union SIGNAL*)reply;
}

/****************************************/

static U32
ram_pmm_change(struct pmm_change *request)
{
   struct regionListElement *reg = *findElement(&regionList, request->base);

   if(request->metadata_size > PMM_MAX_METALEN) {
     LOG_ERROR("Length of metadata (\"%s\") is larger than PMM_MAX_METALEN"
               " (%lu)\n", request->metadata, PMM_MAX_METALEN);
     return OSE_PMM_ECORRUPT_METADATA;
   }

   strcpy(reg->metadata, (char *)request->metadata);

   return PMM_EOK;
}

/****************************************/

static U32
ram_pmm_free(log_addr_t base)
{
  struct regionListElement **reg, *tmp;

  /* LOG("ram pmm free: 0x%x.\n",base); */

  reg = findElement(&regionList, base);
  tmp = *reg;

  *reg = (*reg)->next;
  free(tmp);

  return PMM_EOK;
}

/****************************************/

static union SIGNAL *
ram_pmm_start_restore_phase(void)
{
  U32 *data;
  U32 reg_cnt = 0;
  U32 replySize;
  struct pmm_restore_start_reply *reply;
  struct regionListElement       *reg;

  for (reg = regionList; reg; reg = reg->next) reg_cnt++;

  replySize = sizeof(struct pmm_restore_start_reply) +
    sizeof(log_addr_t) * reg_cnt;

  reply = (struct pmm_restore_start_reply*)alloc(replySize,
                                                 PMM_RESTORE_START_REPLY);

  /*
   *  fill in regions to restore in reply sig.
   */

  data = reply->regions;

  reg = regionList;
  while (reg) {
    *data++ = reg->base;
    reg = reg->next;
  }

  /* Fill in last part of sig */
  reply->no_regions = reg_cnt;

  return (union SIGNAL*)reply;
}

/****************************************/

static union SIGNAL *
ram_pmm_restore_region(log_addr_t base)
{
  struct  pmm_restore_region_reply *reply;
  struct regionListElement *reg = *findElement(&regionList, base);
  U32 replySize, metaSize;

  metaSize = strlen(reg->metadata);
  replySize = sizeof(struct  pmm_restore_region_reply) + metaSize;
  reply = (struct  pmm_restore_region_reply*)alloc(replySize,
                                                   PMM_RESTORE_REGION_REPLY);

  strcpy((char *)reply->metadata, reg->metadata);
  reply->base          = base;
  reply->region_type   = PMM_BOUND;
  reply->metadata_size = metaSize;

  return (union SIGNAL*)reply;
}

/****************************************/

static union SIGNAL *
ram_pmm_restore_binding(log_addr_t base)
{
  int i;
  struct pmm_restore_binding_reply *reply;
  struct regionListElement *reg = *findElement(&regionList, base);
  U32 replySize;

  replySize = sizeof(struct pmm_restore_binding_reply) +
    sizeof(struct phys_resource) * reg->no_fragments;

  reply = (struct pmm_restore_binding_reply*)alloc(replySize,
                                                   PMM_RESTORE_BINDING_REPLY);

  reply->no_fragments = reg->no_fragments;
  reply->base = base;

  for (i=0; i < reg->no_fragments; i++)
  {
    reply->fragments[i] = reg->fragments[i];
  }
  return (union SIGNAL*)reply;
}


/*
 ****************************************************************************
 * PROCESS.
 ****************************************************************************
 */

OS_PROCESS(RPMM_NAME)
{
  static const SIGSELECT sel_start[] = {1, PMM_START_REQUEST};
  static const SIGSELECT sel_all[]   = {0};
  static const SIGSELECT sel_init[]  = {7,PMM_INIT_REQUEST,
                                        PMM_RESTORE_START_REQUEST,
                                        PMM_RESTORE_REGION_REQUEST,
                                        PMM_RESTORE_BINDING_REQUEST,
                                        PMM_CHANGE_REQUEST,
                                        PMM_RESTORE_COMPLETE_REQUEST,
                                        PMM_CORRUPT_METADATA};
  union SIGNAL *sig, *reply_sig;
  Boolean restore_complete = False;
  PROCESS gmm_pid_;

  rpmm_init(&gmm_pid_, &status);

  sig = receive(sel_start);

  if (sig->start_request.options & PMM_START_OPTION_FORCE_COLD_START) {
    LOG("Cold start (forced)\n");
    rpmm_reset();
  }
  else if (status == PMM_START_COLD) {
    LOG("Cold start\n");
    rpmm_reset();
  }
  else if (status == PMM_START_WARM) {
    LOG("Warm start\n");
    rpmm_load();
  }

  max_nr_of_regions = sig->start_request.max_nr_of_regions;
  {
    size_t totSize = rpmmMax - (char*)rpmm;
    size_t typicalSize = sizeof(struct regionListElement)
      + 19*sizeof(struct phys_resource);
    if (typicalSize*max_nr_of_regions > totSize) {

      /* To documentation:
       * "The amount of memory needed by PMM can not be calculated in advance,
       * but an upper estimate is 500 bytes per region. The warning appears
       * when that estimated size * the configured value of
       * <max number of regions> exceeds the size of region "rpmm"
       * It does not necessarily indicate a problem, it is just a hint that
       * the system is close to the limit where creation of persistent regions
       * fails due to memory shortage."
       */

      LOG("Warning. Max number of persistent regions (=%#d) is high compared"\
          "\nwith configured " RPMM_REGION_NAME " region size (=%#x) "\
          "anticpating a average meta data size (=%#x) bytes. \n",
          max_nr_of_regions, totSize, typicalSize);
    }
  }

  rpmm_pid_ = current_process();

  sig->sigNo = PMM_START_REPLY;
  sig->start_reply.status = status;
  send(&sig, sender(&sig));

  /*
   * Notify the GMM that a PMM that handles ram exists,
   * but only IF valid persistent meta data area found!
   * If initialisation fails, we will stop there.
   */

  sig = alloc(sizeof(struct pmm_assign) + strlen(PMM_MEMORY_TYPE),
              PMM_ASSIGN);
  sig->assign.version = PMM_SIG_VERSION;
  sig->assign.max_nr_of_regions= max_nr_of_regions;
  strcpy(sig->assign.memory_type, PMM_MEMORY_TYPE);
  send(&sig, gmm_pid_);

  /*
   * Handle the init phase.
   */
  {
    sig = receive(sel_init);
    if (sig->sigNo != PMM_INIT_REQUEST)
    {
      error2(OSE_PMM_EBAD_SIGNAL_SEQUENCE, (OSERRCODE) sig);
      free_buf(&sig);
    }

    /* Reuse signal for reply. */
    sig->sigNo = PMM_INIT_REPLY;
    send(&sig, gmm_pid_);
  }

  /*
   * Handle the restore phase. First signal must be
   * PMM_RESTORE_START_REQUEST.
   */
  {
    sig = receive(sel_init);
    if (sig->sigNo != PMM_RESTORE_START_REQUEST)
    {
      error2(OSE_PMM_EBAD_SIGNAL_SEQUENCE, (OSERRCODE) sig);
      free_buf(&sig);
    }

    free_buf(&sig);
    reply_sig = ram_pmm_start_restore_phase();
    send(&reply_sig, gmm_pid_);
  }

  /* Receive other restore signals. */
  while (restore_complete == False)
  {
    sig = receive(sel_init);
    switch (sig->sigNo)
    {
      case PMM_RESTORE_REGION_REQUEST:
        {
          U32 region_type;
          log_addr_t base;

          reply_sig =
            ram_pmm_restore_region(sig->restore_region_request.base);
          region_type = reply_sig->restore_region_reply.region_type;
          base = reply_sig->restore_region_reply.base;
          free_buf(&sig);
          send(&reply_sig, gmm_pid_);

          if (region_type == PMM_BOUND)
          {
            /* Receive bind signal for the region. */
            sig = receive(sel_init);
            if ((sig->sigNo == PMM_RESTORE_BINDING_REQUEST) &&
                (sig->restore_binding_request.base == base))
            {
              free_buf(&sig);
              reply_sig = ram_pmm_restore_binding(base);
              send(&reply_sig, gmm_pid_);
            }
            else if ((sig->sigNo == PMM_CORRUPT_METADATA))
            {
              free_buf(&sig);
            }
            else
            {
              error2(OSE_PMM_EBAD_SIGNAL_SEQUENCE, (OSERRCODE) sig);
              free_buf(&sig);
            }
          }
          break;
        }

      case PMM_RESTORE_COMPLETE_REQUEST:

        /* Reuse signal for reply. */
        sig->sigNo = PMM_RESTORE_COMPLETE_REPLY;
        send(&sig, gmm_pid_);
        LOG("Restore phase completed\n");
        restore_complete = True;
        break;

      case PMM_CORRUPT_METADATA:
        {
          log_addr_t base;

          base = sig->corrupt_metadata.base;
          error2(OSE_PMM_ECORRUPT_GMM_METADATA, (OSERRCODE) base);
          /* Error handler should not return. */
          break;
        }

      default:
        error2(OSE_PMM_EBAD_SIGNAL_SEQUENCE, (OSERRCODE) sig);
        free_buf(&sig);
        break;
    }
  }

  /*
   * Enter normal signal handling.
   */
  for (;;)
  {
    sig = receive(sel_all);
    switch (sig->sigNo)
    {
      case PMM_VERSION_REQUEST:
        /* Reuse signal for reply. */
        sig->sigNo = PMM_VERSION_REPLY;
        sig->version.sig_version = PMM_SIG_VERSION;
        send(&sig, sender(&sig));

        break;

      case PMM_ALLOC_SAS_REQUEST:
        /* Alloc reply signal. */
        reply_sig = ram_pmm_alloc_sas(&sig->alloc_sas_request);
        free_buf(&sig);
        send(&reply_sig, gmm_pid_);
        rpmm_sync();
        break;

      case PMM_ALLOC_SASE_REQUEST:

        /* Alloc reply signal. */
        reply_sig = ram_pmm_alloc_sase(&sig->alloc_sase_request);
        free_buf(&sig);
        send(&reply_sig, gmm_pid_);
        rpmm_sync();
        break;

      case PMM_CHANGE_REQUEST:

        /* Alloc reply signal. */
        sig->sigNo = PMM_CHANGE_REPLY;
        sig->change.ecode = ram_pmm_change(&sig->change);
        send(&sig, gmm_pid_);
        rpmm_sync();
        break;

      case PMM_FREE_REQUEST:
        /* Reuse signal for reply. */
        sig->free.ecode = ram_pmm_free(sig->free.base);
        sig->sigNo = PMM_FREE_REPLY;
        send(&sig, gmm_pid_);
        rpmm_sync();
        break;

      case PMM_WRITE_REQUEST:
        LOG("ram_pmm ILLEGAL write dest: 0x%x\n",
            sig->write.dest);
        error2(OSE_PMM_EUNKNOWN_SIGNAL, (OSERRCODE) sig);
        break;

      case PMM_RESET:
        rpmm_reset();
        break;

      default:
        error2(OSE_PMM_EUNKNOWN_SIGNAL, (OSERRCODE) sig);
        free_buf(&sig);
        break;
    }
  }
}

static int reset_sig(void)
{
  if(rpmm_pid_)
  {
    union SIGNAL *sig = alloc(sizeof(struct pmm_reset), PMM_RESET);
    send(&sig,rpmm_pid_);
  }
  return RET_SUCCESS;
}

static int
cmd_rpmm(int argc, char **argv)
{
  if(argc == 2) {
    if (!strcmp(argv[1], "-l")) return rpmm_list();

    if (!strcmp(argv[1], "-s")) return print_status();

    if (!strcmp(argv[1], "-r")) return reset_sig();
  }

  printf("Usage:\n");
  printf("  %s -s\n"
         "    Print status (warm, cold, build_info).\n"
         "\n", argv[0]);
  printf("  %s -l\n"
         "    List regions.\n"
         "\n", argv[0]);
  printf("  %s -r\n"
         "    Reset restart data. (Current set of persistent regions will\n"
         "    not be recreated after reboot.)\n"
         "\n", argv[0]);

  return RET_SUCCESS;
}


void add_ram_pmm_cmds(void);

void
add_ram_pmm_cmds(void)
{
  shell_add_cmd("rpmm", "rpmm",
                "ram pmm commands", cmd_rpmm);
}
