/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2006 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

#include <assert.h>
#include <string.h>
#include <limits.h>
#include <stdio.h>
#include "configure.h"
#include "gmm.h"
#include "hal.h"
#include "page.h"
#include "slab.h"
#include "mm_util.h"
#include "mm.h"
#include "mm.sig"
#include "partition.h"
#include "ramlog.h"
#include "ose_err/mm_err.h"
#include "mm_internal.sig"
#ifdef STANDALONE
#include "standalone.h"
#endif

/**
 * A SLAB allocator for OSE's demand pager.
 */

/****************************** DEFINES *******************************/

#define ALIGN_DOWN(a,alignment)   ((a) - ((a) & (alignment - 1)))
#define ALIGN_UP(a,alignment) ALIGN_DOWN(((a) + (alignment) - 1),alignment)


/************************** PROTOTYPES ********************************/

/*static int
  slab_print_statistics(int argc, char **argv);*/

static void
slab_garbage_collect(union mm_event_t *event);

static uint8_t *
get_beginning_of_slab(struct slab_ctrl *slab);

static uint32_t
generate_buf_ctrl_key(const void *a_raw);

static int
compare_buf_ctrl(const void *a_raw, const void *b_raw);

static uint32_t
size_to_static_index(size_t size);

#ifdef ASSERT
static void
slab_debug_ctor(void *ptr, size_t size);

static void
slab_debug_dtor(void *ptr, size_t size);
#endif /* ASSERT */


/************************** STATIC VARIABLES **************************/

/* The list of slab caches. */
static list_t slab_cache_list;

/* A lock for the above list. */
static spinlock_t slab_cache_list_lock;

/* The initial slab cache entry that is used to allocate other slab caches.
 * Statically allocated in the .data section. */
static struct slab_cache slab_cache_slab;

/* The number of bytes that can be freed in all slab caches. */
static uint32_t no_of_freeable_bytes;

/* The slab from which the slab control structures are allocated.
 * Only for slabs with large entries. */
static struct slab_cache *slab_and_buf_ctrl_slab = NULL;

/* The log mem partition that the slab uses for all its allocations. */
static struct partition *slab_log_partition = NULL;

/* A list of sizes used by the static slab allocator */
static uint32_t slab_static_sizes[SLAB_STATIC_NUM_SIZES] = SLAB_STATIC_SIZES;

/* A list of slab caches used by the static slab allocator */
static struct slab_cache *slab_static_caches[SLAB_NUM_OPTIONS][SLAB_STATIC_NUM_SIZES];

/***************************************************************************
 * Local SLAB functions.
 *   init_slab_control_block
 *   delete_slab
 *   delete_slab_from_mm
 *   create_new_slab_from_mm
 *   create_new_slab_async
 *   in_use
 *   is_full
 *   mark_used
 *   mark_unused
 *   get_free_entry_in_slab
 *   get_slab_to_allocate_from
 *   get_slab_control_block
 *   get_beginning_of_slab
 *   create_region
 *   delete_region
 *
 ***************************************************************************/

/*
 *===========================================================================
 *                        init_slab_control_block
 *===========================================================================
 *
 * Description
 * -----------
 * Initializes the control block of a new slab.
 *
 * Parameter description
 * ---------------------
 * page    Pointer to the raw memory that is going to make up the slab.
 *
 * Returns
 * -------
 * struct slab_ctrl*   Returns a pointer to the created slab control block.
 *
 */
static struct slab_ctrl *
init_slab_control_block(struct slab_cache *slab_cache, uint8_t *page)
{
   struct slab_ctrl *metadata;
   uint32_t i;

#ifdef ASSERT
   memset(page, 0x0f, slab_cache->mm_alloc_size);
#endif

   if (slab_cache->size < EMBED_SLAB_CTRL_THRESHOLD)
   {
      uint8_t **p = (uint8_t **)page;

      metadata = (struct slab_ctrl *)(page + PAGE_SIZE -
                                      sizeof(struct slab_ctrl));
      metadata->free_list = NULL;
      metadata->no_of_used_entries = 0;
      metadata->slab_cache = slab_cache;
      for (i = 0; i < slab_cache->entries_per_slab; i++)
      {
         *p = (uint8_t *)metadata->free_list;
         metadata->free_list = (struct buf_ctrl *)p;
         p += slab_cache->size >> 2;
      }
   }
   else
   {
      metadata = (struct slab_ctrl *)slab_alloc(slab_and_buf_ctrl_slab,
                                                SLAB_SLEEP);
      if (unlikely(!metadata))
      {
         ramlog_printf("Error: Failed to alloc slab control block.\n");
         error(OSE_MM_EINTERNAL_ERROR);
      }

      metadata->free_list = NULL;
      metadata->no_of_used_entries = 0;
      metadata->slab_cache = slab_cache;
      init_list(&metadata->buf_ctrl_list);
      for (i = 0; i < slab_cache->entries_per_slab; i++)
      {
         struct buf_ctrl *buf_ctrl;
         int hash_status;

         buf_ctrl = (struct buf_ctrl *)slab_alloc(slab_and_buf_ctrl_slab,
                                                  SLAB_SLEEP);
         if (unlikely(!buf_ctrl))
         {
            ramlog_printf("Error: Failed to alloc buffer control block.\n");
            error(OSE_MM_EINTERNAL_ERROR);
         }

         buf_ctrl->free_list = metadata->free_list;
         metadata->free_list = buf_ctrl;
         buf_ctrl->slab = metadata;
         buf_ctrl->buf = page + i * slab_cache->size;
         init_list_link(&buf_ctrl->buf_ctrl_link);
         init_list_link(&buf_ctrl->hash_link);
         insert_list_tail_no_lock(&metadata->buf_ctrl_list,
                                  &buf_ctrl->buf_ctrl_link);
         hash_status = insert_in_hash(&slab_cache->buf_ctrl_hash,
                                      &buf_ctrl->hash_link);
         if (unlikely(!hash_status))
         {
            ramlog_printf("Error: Hash entry is already in hash in init_slab_control_block.\n");
            error(OSE_MM_EINTERNAL_ERROR);
         }
      }
   }
   init_list_link(&metadata->link);

   return metadata;
}

/*
 *===========================================================================
 *                        delete_slab
 *===========================================================================
 *
 * Description
 * -----------
 * Deletes a slab from the slab cache, but does not deallocate it from MM.
 *
 */
static void
delete_slab(struct slab_ctrl *entry)
{
   struct slab_cache *slab_cache = entry->slab_cache;

   delete_from_list(&slab_cache->slab_list, &entry->link);

   if (slab_cache->size >= EMBED_SLAB_CTRL_THRESHOLD)
   {
      struct buf_ctrl *buf_ctrl =
         (struct buf_ctrl *)list_head(&entry->buf_ctrl_list);

      if (buf_ctrl)
      {
         buf_ctrl = (struct buf_ctrl *)((ptrdiff_t)buf_ctrl -
                                        sizeof(struct buf_ctrl *));
      }
      while (buf_ctrl)
      {
         struct buf_ctrl *next_buf_ctrl;
         int hash_status;

         hash_status = delete_from_hash(&slab_cache->buf_ctrl_hash,
                                        &buf_ctrl->hash_link);
         if (unlikely(!hash_status))
         {
            ramlog_printf("Error: Hash entry already removed in delete_slab.\n");
            error(OSE_MM_EINTERNAL_ERROR);
         }
         next_buf_ctrl = (struct buf_ctrl *)list_next(&entry->buf_ctrl_list,
                                                     &buf_ctrl->buf_ctrl_link);
         if (next_buf_ctrl)
         {
            next_buf_ctrl = (struct buf_ctrl *)((ptrdiff_t)next_buf_ctrl -
                                                sizeof(struct buf_ctrl *));
         }
         delete_from_list_no_lock(&buf_ctrl->buf_ctrl_link);
         slab_free(slab_and_buf_ctrl_slab, buf_ctrl);
         buf_ctrl = next_buf_ctrl;
      }

      slab_free(slab_and_buf_ctrl_slab, entry);
   }

   slab_cache->no_of_slabs--;
   slab_cache->max_no_of_used_entries -= slab_cache->entries_per_slab;
   no_of_freeable_bytes -= slab_cache->mm_alloc_size;
}

/*
 *===========================================================================
 *                        create_region
 *===========================================================================
 *
 * Description
 * -----------
 * Allocates region of metadata from MM, either synchonously or asyncronously.
 *
 * Returns
 * -------
 * Non-zero on success.
 */
static int
create_region(struct slab_cache *slab_cache, size_t size, log_addr_t *la_out)
{
   log_addr_t la, pa, tmp_pa, page;
   static PROCESS mm_pid;

   if (slab_cache->options & SLAB_SLEEP)
   {
      static SIGSELECT reply_sig[] = {1, MM_ALLOC_METADATA_REPLY};
      struct mm_alloc_metadata *sig = (struct mm_alloc_metadata*)
         alloc(sizeof(struct mm_alloc_metadata), MM_ALLOC_METADATA_REQUEST);

      sig->base = 0;
      sig->size = size;

      if (!mm_pid)
      {
         /* Non-blocking hunt for the MM server. */
         hunt("ose_memory_mgr", 0, &mm_pid, NULL);
      }

      send((union SIGNAL**)&sig, mm_pid);
      sig = (struct mm_alloc_metadata*)receive(reply_sig);

      if (sig->status != 0)
      {
         free_buf((union SIGNAL**)&sig);
         return 0;
      }

      la = sig->base;
      
      free_buf((union SIGNAL**)&sig);
   }
   else if (gmm_state.hal_info->has_mmu)
   {
      if (!mm_alloc_log_mem2(slab_log_partition, size, &la))
      {
         return 0;
      }

      if (!mm_page_alloc(size, &tmp_pa))
      {
         return 0;
      }
      
      pa = tmp_pa;

      for (page = 0; page < size; page += PAGE_SIZE)
      {
         hal_status_t status;
         U32 options = HAL_GLOBAL_MAPPING;


         status = hal_map_page(gmm_state.system_domain->hal_domain,
                               la + page, pa + page, PAGE_SIZE, options,
                               gmm_state.internal_access_mode,
                                DEFAULT_CACHE_MODE);

         if (status != HAL_SUCCESS)
         {
            hal_unmap_range(gmm_state.system_domain->hal_domain,
                            la, slab_cache->size, 0);
            mm_page_free(pa, slab_cache->mm_alloc_size);
            mm_free_log_mem(slab_log_partition, la, size);
            return 0;
         }
      }
   }
   else
   {
      if (!mm_page_alloc(size, &la))
      {
         return 0;
      }
   }

   *la_out = la;

   return 1;
}

/*
 *===========================================================================
 *                        delete_region
 *===========================================================================
 *
 * Description
 * -----------
 * Deletes a region of metadata from MM, either synchonously or asyncronously.
 *
 */
static int
delete_region(uint32_t options, log_addr_t base, size_t size,
              int force_synchronous)
{
   phys_addr_t phys;

   if (!force_synchronous && options & SLAB_SLEEP)
   {
      static PROCESS mm_pid;
      static SIGSELECT reply_sig[] = {1, MM_FREE_METADATA_REPLY};
      struct mm_free_metadata *sig = (struct mm_free_metadata*)
         alloc(sizeof(struct mm_free_metadata), MM_FREE_METADATA_REQUEST);
      sig->base = base;
      sig->size = size;

      if (!mm_pid)
      {
         /* Non-blocking hunt for the MM server. */
         hunt("ose_memory_mgr", 0, &mm_pid, NULL);
      }

      send((union SIGNAL**)&sig, mm_pid);
      sig = (struct mm_free_metadata*)receive(reply_sig);

      if (sig->status != 0)
      {
         free_buf((union SIGNAL**)&sig);
         return 0;
      }

      free_buf((union SIGNAL**)&sig);
   }
   else
   {
      if (hal_lookup(gmm_state.system_domain->segnum, base, &phys, NULL, NULL, 0)
          != HAL_SUCCESS)
      {
         return 0;
      }

      hal_unmap_range(gmm_state.system_domain->hal_domain, base, size, 0);
      mm_page_free((log_addr_t)phys, size);

      if (gmm_state.hal_info->has_mmu)
      {
         mm_free_log_mem(slab_log_partition, base, size);
      }
   }

   return 1;
}


/*
 *===========================================================================
 *                        delete_slab_from_mm
 *===========================================================================
 *
 * Description
 * -----------
 * Deallocates a slab from MM and removes it from the slab cache.
 *
 */
static void
delete_slab_from_mm(struct slab_ctrl *slab)
{
   log_addr_t page = (log_addr_t)get_beginning_of_slab(slab);
   size_t size = slab->slab_cache->mm_alloc_size;
   uint32_t options = slab->slab_cache->options;

   delete_slab(slab);
   if (!delete_region(options, page, size, 0))
   {
      ramlog_printf("Slab: Failed to delete a slab\n");
      error2(OSE_MM_EINTERNAL_ERROR, 0);
   }
}

/*
 *===========================================================================
 *                      create_new_slab_from_mm
 *===========================================================================
 *
 * Description
 * -----------
 * Allocates and creates a new slab and puts it in the list provided.
 * The slab is allocated from the MM.
 *
 * Restriction
 * -----------
 * Cannot be called from MM context. Returns with lock acquired.
 *
 * Returns
 * -------
 * struct slab_ctrl *  A pointer to the created slab or NULL if it
 *                     was not possible to create a new slab.
 *
 */
static struct slab_ctrl *
create_new_slab_from_mm(struct slab_cache *slab_cache, Msr *sr)
{
   log_addr_t la;
   struct slab_ctrl *slab;

   if (!create_region(slab_cache, slab_cache->mm_alloc_size, &la))
   {
      spin_lock_irqsave(&slab_cache->lock, *sr);
      return NULL;
   }

   spin_lock_irqsave(&slab_cache->lock, *sr);

   slab = init_slab_control_block(slab_cache, (uint8_t *)la);
   slab_cache->no_of_slabs++;
   slab_cache->max_no_of_used_entries += slab_cache->entries_per_slab;
   no_of_freeable_bytes += slab_cache->mm_alloc_size;
   insert_list_head(&slab_cache->slab_list, &slab->link);

   return slab;
}


/*
 *===========================================================================
 *                     create_new_slab_async
 *===========================================================================
 *
 * Description
 * -----------
 * Allocates and creates a new slab and puts it in the list provided.
 * The page is allocated asynchronously, and can thus be called from
 * any context. This call will return immediately, and the new
 * slab will be put into the slab list later by the slab_helper process.
 *
 * Returns
 * -------
 * struct slab_ctrl *  A pointer to the created slab or NULL if it
 *                     was not possible to create a new slab.
 *
 */
static struct slab_ctrl *
create_new_slab_async(struct slab_cache *slab_cache, Msr *sr)
{
   /* FIXME: Implement this function. */
   return create_new_slab_from_mm(slab_cache, sr);
}

/*
 *===========================================================================
 *                             in_use
 *===========================================================================
 *
 * Description
 * -----------
 * Is the slab used by anyone? Or, has any memory been allocated from this
 * slab?
 *
 * Returns
 * -------
 * int   One if the slab is in use, zero otherwise.
 *
 */
static inline int
in_use(struct slab_ctrl *slab)
{
   if (slab->no_of_used_entries)
   {
      return 1;
   }
   else
   {
      return 0;
   }
}

/*
 *===========================================================================
 *                             is_full
 *===========================================================================
 *
 * Description
 * -----------
 * Is the slab fully used? Or, is there no more free space in it?
 *
 * Returns
 * -------
 * int   One if the slab is full, zero otherwise.
 *
 */
static inline int
is_full(struct slab_ctrl *slab)
{
   if (slab->free_list)
   {
      return 0;
   }
   else
   {
      return 1;
   }
}

/*
 *===========================================================================
 *                             mark_used
 *===========================================================================
 *
 * Description
 * -----------
 * Mark the entry in the slab pointed to by buf as used.
 *
 */
static inline void
mark_used(struct slab_ctrl *slab, struct buf_ctrl *buf)
{
   assert(slab->free_list == buf);
   slab->free_list = buf->free_list;
   slab->no_of_used_entries++;
   slab->slab_cache->no_of_used_entries++;
}

/*
 *===========================================================================
 *                             mark_unused
 *===========================================================================
 *
 * Description
 * -----------
 * Mark the entry in the slab pointed to by buf as not used.
 *
 */
static inline void
mark_unused(struct slab_ctrl *slab, struct buf_ctrl *buf)
{
   buf->free_list = slab->free_list;
   slab->free_list = buf;
   slab->no_of_used_entries--;
   slab->slab_cache->no_of_used_entries--;
}

/*
 *===========================================================================
 *                          get_free_entry_in_slab
 *===========================================================================
 *
 * Description
 * -----------
 * Produces a pointer to a free entry in the provided slab.
 *
 * Returns
 * -------
 * uint8_t *  A pointer to a free entry in the slab, or NULL if none
 *            could be found.
 *
 */
static uint8_t *
get_free_entry_in_slab(struct slab_ctrl *slab)
{
   struct buf_ctrl *buf_ctrl;
   struct slab_cache *slab_cache = slab->slab_cache;

   /* Get the address of the free buffer. */
   buf_ctrl = slab->free_list;
   if (buf_ctrl)
   {
      if (!in_use(slab))
      {
         no_of_freeable_bytes -= slab_cache->mm_alloc_size;
      }
      mark_used(slab, buf_ctrl);

      /* If slab is fully used, move the first_full slab marker. */
      if (is_full(slab))
      {
         slab_cache->partially_empty =
            (struct slab_ctrl *)slab_cache->partially_empty->link.prev;
      }

      if (slab_cache->size < EMBED_SLAB_CTRL_THRESHOLD)
      {
         return (uint8_t *)buf_ctrl;
      }
      else
      {
         return (uint8_t *)buf_ctrl->buf;
      }
   }

   return NULL;
}

/*
 *===========================================================================
 *                          get_slab_control_block
 *===========================================================================
 *
 * Description
 * -----------
 * Gets the slab control block that corresponds to the provided entry.
 *
 * Returns
 * -------
 * struct slab_ctrl *  A pointer to the corresponding slab control block.
 *
 */
static struct slab_ctrl *
get_slab_control_block(struct slab_cache *slab_cache, struct buf_ctrl **buf)
{
   if (slab_cache->size < EMBED_SLAB_CTRL_THRESHOLD)
   {
      return (struct slab_ctrl *)(((ptrdiff_t)*buf & ~(PAGE_SIZE-1)) +
                                  PAGE_SIZE - sizeof(struct slab_ctrl));
   }
   else
   {
      struct buf_ctrl search_buf_ctrl;

      search_buf_ctrl.buf = *buf;
      *buf = (struct buf_ctrl *)find_in_hash(&slab_cache->buf_ctrl_hash,
                                             &search_buf_ctrl.hash_link);
      if (unlikely(!*buf))
      {
         ramlog_printf("Error: Trying to free a non-allocated address.\n");
         error(OSE_MM_EINTERNAL_ERROR);
      }
      /* Buffer control blocks are members of two lists and the hash
       * happens to be the one that is second in the struct. */
      *buf = (struct buf_ctrl *)((ptrdiff_t)*buf -
                                 sizeof(list_link_t) -
                                 sizeof(struct buf_ctrl *));
      return (*buf)->slab;
   }
}

/*
 *===========================================================================
 *                        get_slab_to_allocate_from
 *===========================================================================
 *
 * Description
 * -----------
 * Gets a slab to allocate from. It will return a partially empty one,
 * if one exists. If not, an empty slab will be returned.
 * If there are only full slabs, a full one will be returned.
 *
 * Returns
 * -------
 * struct slab_ctrl*   A pointer to the slab to allocate from.
 *
 */
static inline struct slab_ctrl *
get_slab_to_allocate_from(struct slab_cache *slab_cache)
{
   struct slab_ctrl *slab =
      (struct slab_ctrl *)slab_cache->partially_empty->link.prev;

   if (slab != (struct slab_ctrl *)&slab_cache->slab_list)
   {
      return slab;
   }
   else
   {
      return NULL;
   }
}

/*
 *===========================================================================
 *                          get_beginning_of_slab
 *===========================================================================
 *
 * Description
 * -----------
 * Gets a pointer to the first entry in the slab that belongs to the
 * provided slab control structure.
 *
 * Returns
 * -------
 * uint8_t *  A pointer to the first entry in the slab.
 *
 */
static uint8_t *
get_beginning_of_slab(struct slab_ctrl *slab)
{
   if (slab->slab_cache->size < EMBED_SLAB_CTRL_THRESHOLD)
   {
      return (uint8_t *)((ptrdiff_t)slab & ~(PAGE_SIZE-1));
   }
   else
   {
      struct buf_ctrl *buf_ctrl;

      buf_ctrl = (struct buf_ctrl *)list_head(&slab->buf_ctrl_list);
      buf_ctrl = (struct buf_ctrl *)((ptrdiff_t)buf_ctrl -
                                     sizeof(struct buf_ctrl *));
      return buf_ctrl->buf;
   }
}



/***************************************************************************
 * SLAB ALLOCATOR FUNCTIONS
 *
 * You can call:
 *   create_slab_allocator
 *   delete_slab_allocator
 *   slab_alloc
 *   slab_free
 *
 * MM can call:
 *   slab_alloc (with the SLAB_NOSLEEP option)
 *   slab_free
 *
 ***************************************************************************/

/*
 *===========================================================================
 *                        create_slab_allocator
 *===========================================================================
 *
 * Description
 * -----------
 * Creates a slab allocator for a specific type of data.
 *
 * Parameter description
 * ---------------------
 * name         Name of the entries in the slab, or NULL to obtain a shared
 *              cache suitable for the request size.
 * size         The size of each entry in the slab. Cannot be changed
 *              after creation.
 * align        The alignment of the data. If no alignment is desired,
 *              enter a zero.
 * spares       The number of spare entries that are allocated.
 *              The slab cannot ask for more pages from the MM when it is
 *              called by the MM. Instead, it will use memory from one or
 *              more of these spare entries, and as soon the slab is called
 *              from some other process, it allocates more entries so that
 *              it have the configured amount of spare entries again.
 * constructor  A pointer to a function that is called every time an
 *              object is created. Usefull for setting the initial
 *              state of the entry.
 * destructor   A pointer to a function that is called every time an
 *              object is freed. Usefull for deallocating stuff.
 *
 * options      Option mask. Supports the mutually exclusive options
 *              SLAB_SLEEP and SLAB_NOSLEEP. With SLAB_SLEEP the slab
 *              will allocate its memory using signalling. SLAB_NOSLEEP
 *              does not use signalling, but can only be used when MM
 *              is re-entrancy protected. (See mm.sig, mm_multithread). 
 *
 * Returns
 * -------
 * struct slab_cache*   Returns a pointer to the created slab.
 *
 * Restrictions
 * ------------
 * You cannot create a slab with an entry size larger than
 * EMBED_SLAB_CTRL_THRESHOLD before the system heap has been initialized.
 * This is usually 256 bytes.
 *
 */
struct slab_cache *
create_slab_allocator(const char *name, size_t size, uint32_t align,
                      uint32_t spares,
                      void (*constructor)(void *, size_t),
                      void (*destructor)(void *, size_t),
                      uint32_t options)
{
   struct slab_cache *slab_cache;
   struct slab_ctrl *slab;
   static int initialized = 0;
   uint32_t size_index   = size_to_static_index(size);
   uint32_t option_index = options & (SLAB_SLEEP|SLAB_NOSLEEP);

   if (unlikely(size == 0))
   {
      ramlog_printf("Size cannot be zero for a slab allocator.\n");
      error2(OSE_MM_EINTERNAL_ERROR, 0);
   }

   /* Create the new slab allocator. */
   if (unlikely(!initialized))
   {
      slab_cache = &slab_cache_slab;
      initialized = 1;
   }
   else
   {
      /* Anonymous caches are static and shared between clients. A cache for
       * this entry size is only created if no previous cache exists. */
      if (!name)
      {
         if (size_index != SLAB_STATIC_INVALID_SIZE &&
             slab_static_caches[option_index][size_index] != NULL)

         {
            return slab_static_caches[option_index][size_index];
         }
         else if (size_index != SLAB_STATIC_INVALID_SIZE)
         {
            /* Override parameters to make sure they
             * are suitable to all possible clients */
            size  = slab_static_sizes[size_index];
            align = SLAB_STATIC_ALIGN;
         }
      }

      slab_cache = (struct slab_cache *)slab_alloc(&slab_cache_slab, SLAB_SLEEP);
      if (unlikely(!slab_cache))
      {
         ramlog_printf("Could not allocate memory for slab allocator %s.\n",
                       name);
         error2(OSE_MM_EINTERNAL_ERROR, 0);
      }
   }

   if (name)
   {
      strlcpy(slab_cache->name, name, SLAB_NAME_MAX);
   }
   else if (size_index != SLAB_STATIC_INVALID_SIZE)
   {
      snprintf(slab_cache->name, SLAB_NAME_MAX,
               SLAB_STATIC_NAME_PREFIX "%d", size);
   }
   else
   {
      strlcpy(slab_cache->name, "slab_anonymous", SLAB_NAME_MAX);
   }

   slab_cache->options = options & (SLAB_SLEEP|SLAB_NOSLEEP);

   /* Word align the size. */
   slab_cache->size = ALIGN_UP(size, 4);
   slab_cache->align = align;
   if (slab_cache->align != 0)
   {
      if ((slab_cache->size & (slab_cache->align - 1)) != 0)
      {
         slab_cache->size += slab_cache->align -
            (slab_cache->size & (slab_cache->align - 1));
      }
   }

   if (slab_cache->size < EMBED_SLAB_CTRL_THRESHOLD)
   {
      slab_cache->mm_alloc_size = PAGE_SIZE;
      slab_cache->entries_per_slab = (PAGE_SIZE - sizeof(struct slab_ctrl)) /
         slab_cache->size;
   }
   else if (slab_cache->size % PAGE_SIZE == 0)
   {
      slab_cache->mm_alloc_size = size;
      slab_cache->entries_per_slab = 1;
   }
   else if (slab_cache->size > MAX_MM_ALLOC_SIZE)
   {
      slab_cache->mm_alloc_size = ALIGN_UP(slab_cache->size, PAGE_SIZE);
      slab_cache->entries_per_slab = slab_cache->size /
                                     slab_cache->mm_alloc_size;
   }
   else
   {
      uint32_t i, max_pages;
      uint32_t min_waste_page_size = 0;
      uint32_t min_waste_size = UINT_MAX;
      uint32_t no_of_entries = 0;

      /* Calculate the mm_alloc_size that provides the least amount of wasted
       * space. Though, the number of pages has to be less than
       * MAX_MM_ALLOC_SIZE and the number of entries cannot be larger
       * than 32. */
      max_pages = MAX_MM_ALLOC_SIZE / PAGE_SIZE;
      if ((max_pages * PAGE_SIZE) / slab_cache->size > 32)
      {
         max_pages = (32 * slab_cache->size) / PAGE_SIZE;
      }
      for (i = 1; i <= max_pages; i++)
      {
         uint32_t waste;

         while ((no_of_entries + 1) * slab_cache->size <= PAGE_SIZE * i)
         {
            no_of_entries++;
         }
         if (no_of_entries == 0)
         {
            continue;
         }
         waste = PAGE_SIZE * i - no_of_entries * slab_cache->size;

         if (min_waste_page_size == 0 || min_waste_size > waste)
         {
            min_waste_size = waste;
            min_waste_page_size = i * PAGE_SIZE;
         }
      }

      slab_cache->mm_alloc_size = min_waste_page_size;
      slab_cache->entries_per_slab = slab_cache->mm_alloc_size /
                                     slab_cache->size;
   }

   slab_cache->constructor = constructor;
   slab_cache->destructor = destructor;

#ifdef ASSERT
   if (constructor == NULL)
   {
      slab_cache->constructor = slab_debug_ctor;
   }

   if (destructor == NULL)
   {
      slab_cache->destructor = slab_debug_dtor;
   }
#endif

   slab_cache->spare_entries = spares;
   spin_lock_init(&slab_cache->lock);
   slab_cache->no_of_slabs = 0;
   slab_cache->no_of_used_entries = 0;
   slab_cache->max_used_spare_entries = 0;
   init_list(&slab_cache->slab_list);
   slab_cache->partially_empty = (struct slab_ctrl *)&slab_cache->slab_list;
   init_list_link(&slab_cache->link);
   insert_list_tail(&slab_cache_list, &slab_cache->link);

   /* If the cache is anonymous and has a valid static size
    * it is registered as a shared static cache. */
   if (name == NULL && size_index != SLAB_STATIC_INVALID_SIZE &&
       slab_static_caches[option_index][size_index] == NULL)
   {
      slab_static_caches[option_index][size_index] = slab_cache;
   }

   if (slab_cache->size >= EMBED_SLAB_CTRL_THRESHOLD)
   {
      if (!slab_and_buf_ctrl_slab)
      {
         uint32_t size;

         if (sizeof(struct slab_ctrl) > sizeof(struct buf_ctrl))
         {
            size = sizeof(struct slab_ctrl);
         }
         else
         {
            size = sizeof(struct buf_ctrl);
         }
         /* Allocate the slab for the slab and buffer control structures. */
         slab_and_buf_ctrl_slab = create_slab_allocator("slab_and_buf_ctrl",
                                                size, 0, 0, NULL, NULL, 0);
      }
      init_hash(&slab_cache->buf_ctrl_hash, 32, generate_buf_ctrl_key,
                compare_buf_ctrl);
   }

   /* Create the initial slab and any spare entries, if necessary. */
   while (slab_cache->no_of_slabs * slab_cache->entries_per_slab <
          slab_cache->spare_entries + 1)
   {
      Msr sr;
      slab = create_new_slab_from_mm(slab_cache, &sr);
      spin_unlock_irqrestore(&slab_cache->lock, sr);
      if (unlikely(!slab))
      {
         ramlog_printf("Could not allocate memory for slab allocator %s.\n",
                       name);
         error2(OSE_MM_EINTERNAL_ERROR, 0);
      }
   }
   slab_cache->max_no_of_used_entries = slab_cache->no_of_slabs *
      slab_cache->entries_per_slab - slab_cache->spare_entries;

   return slab_cache;
}

/*
 *===========================================================================
 *                        delete_slab_allocator
 *===========================================================================
 *
 * Description
 * -----------
 * Deletes the slab allocator pointed to by *slab. Deallocated all the
 * memory associated with this slab.
 *
 * Restrictions
 * ------------
 * Cannot be called from MM context.
 *
 */
void
delete_slab_allocator(struct slab_cache *slab_cache)
{
   struct slab_ctrl *slab;

   /* Static slab caches are never destroyed */
   if (strstr(slab_cache->name, SLAB_STATIC_NAME_PREFIX) != NULL)
   {
      return;
   }

   slab = (struct slab_ctrl *)list_head(&slab_cache->slab_list);
   while (slab != NULL)
   {
      struct slab_ctrl *current_slab = slab;

      slab = (struct slab_ctrl *)list_next(&slab_cache->slab_list, slab);
      delete_slab_from_mm(current_slab);
   }

   if (slab_cache->size >= EMBED_SLAB_CTRL_THRESHOLD)
   {
      delete_hash(&slab_cache->buf_ctrl_hash);
   }
   delete_from_list(&slab_cache_list, &slab_cache->link);
   slab_free(&slab_cache_slab, slab_cache);
}

/*
 *===========================================================================
 *                        slab_alloc
 *===========================================================================
 *
 * Description
 * -----------
 * Allocate an entry from the slab pointed to by *slab
 *
 * Parameter description
 * ---------------------
 * slab       The slab to allocate from.
 * flags      Either SLAB_SLEEP or SLAB_NOSLEEP, indicating if it is
 *            acceptable to wait for memory if none is currently available.
 *
 * Returns
 * -------
 * void*   Returns a pointer to the newly allocated slab entry.
 *
 */
void *
slab_alloc(struct slab_cache *slab_cache, int flags)
{
   uint8_t *entry = NULL;
   struct slab_ctrl *slab;
   Msr sr;

   spin_lock_irqsave(&slab_cache->lock, sr);

   slab = get_slab_to_allocate_from(slab_cache);
   if (likely(slab))
   {
      entry = get_free_entry_in_slab(slab);
   }
   else
   {
      /* Allocate a new slab. */
      if (flags == SLAB_SLEEP)
      {
         /* If not, allocate a new slab. */
         spin_unlock_irqrestore(&slab_cache->lock, sr);

         slab = create_new_slab_from_mm(slab_cache, &sr);
         if (!slab)
         {
            spin_unlock_irqrestore(&slab_cache->lock, sr);
            return NULL;
         }

         entry = get_free_entry_in_slab(slab);
      }
      else
      {
         spin_unlock_irqrestore(&slab_cache->lock, sr);

         /* This alloc cannot sleep, thus we cannot ask the MM for more
          * memory. Print an error message and die. */
         ramlog_printf("Slab %s has run out of spare entries due to "
                       "SLAB_NOSLEEP allocations.\n", slab_cache->name);
         error2(OSE_MM_EINTERNAL_ERROR, 0);
      }
   }

   if (unlikely(slab_cache->no_of_used_entries >
                slab_cache->max_no_of_used_entries && flags == SLAB_SLEEP))
   {
      if (slab_cache->max_used_spare_entries < slab_cache->no_of_used_entries -
          slab_cache->max_no_of_used_entries)
      {
         slab_cache->max_used_spare_entries = slab_cache->no_of_used_entries -
            slab_cache->max_no_of_used_entries;
      }

      spin_unlock_irqrestore(&slab_cache->lock, sr);

      create_new_slab_async(slab_cache, &sr);
   }

   spin_unlock_irqrestore(&slab_cache->lock, sr);

   if (slab_cache->constructor)
   {
      slab_cache->constructor(entry, slab_cache->size);
   }

   return entry;
}

/*
 *===========================================================================
 *                        slab_free
 *===========================================================================
 *
 * Description
 * -----------
 * Free the entry *buf from the slab pointed to by *slab
 *
 */
void
slab_free(struct slab_cache *slab_cache, void *buf)
{
   struct slab_ctrl *slab;
   struct buf_ctrl *buf_ctrl = buf;
   int was_full;
   Msr sr;

   slab = get_slab_control_block(slab_cache, &buf_ctrl);

   if (slab_cache->destructor)
   {
      slab_cache->destructor(buf, slab_cache->size);
   }

   spin_lock_irqsave(&slab_cache->lock, sr);
   was_full = is_full(slab);
   mark_unused(slab, buf_ctrl);

   /* If slab is empty now, move it to the top of the list if it is
    * not already there. */
   if (!in_use(slab))
   {
      delete_from_list_no_lock(&slab->link);
      insert_list_head_no_lock(&slab_cache->slab_list, &slab->link);
      no_of_freeable_bytes += slab_cache->mm_alloc_size;
   }
   /* If slab is partially used, move it to the partially_empty list
    * marker. */
   else if (was_full)
   {
      if (slab_cache->partially_empty == slab)
      {
         slab_cache->partially_empty =
            (struct slab_ctrl *)slab_cache->partially_empty->link.next;
      }
      delete_from_list_no_lock(&slab->link);
      insert_list_before_no_lock(&slab_cache->partially_empty->link,
                                 &slab->link);
   }

   spin_unlock_irqrestore(&slab_cache->lock, sr);
}


/***************************************************************************
 * VARIOUS SLAB FUNCTIONS
 *
 * You can call:
 *   init_slab
 *   slab_print_statistics
 *
 * Used for the slab control hash table:
 *   generate_buf_ctrl_key
 *   compare_buf_ctrl
 *
 * MM can call:
 *   slab_garbage_collect
 *
 ***************************************************************************/

/*
 *===========================================================================
 *                        init_slab
 *===========================================================================
 *
 * Description
 * -----------
 * Initializes the slab to allocate memory from the provided partition.
 * Needs to be called once before the slab is used.
 *
 */
void
init_slab(const char *slab_log_mem_name)
{
   if (MAX_MM_ALLOC_SIZE < EMBED_SLAB_CTRL_THRESHOLD)
   {
      ramlog_printf("ERROR: MAX_MM_ALLOC_SIZE cannot be less than EMBED_SLAB_CTRL_THRESHOLD.\n");
      error2(OSE_MM_EINTERNAL_ERROR, 0);
   }
   if (EMBED_SLAB_CTRL_THRESHOLD > PAGE_SIZE - sizeof(struct slab_ctrl))
   {
      ramlog_printf("ERROR: EMBED_SLAB_CTRL_THRESHOLD too large.\n");
      error2(OSE_MM_EINTERNAL_ERROR, 0);
   }

   if (gmm_state.hal_info->has_mmu)
   {
      slab_log_partition = mm_find_domain_partition_by_name(
         gmm_state.system_domain, slab_log_mem_name);
      if (slab_log_partition == NULL)
      {
         ramlog_printf("ERROR: Slab could not find logical memory partition.\n");
         error2(OSE_MM_EINTERNAL_ERROR, 0);
      }
   }

   /* Initialize the slab cache list. */
   init_list(&slab_cache_list);
   spin_lock_init(&slab_cache_list_lock);
   no_of_freeable_bytes = 0;

   /* Just to get rid of a compiler warning with rvct on unipro. */
   (void)slab_cache_list_lock;

   /* Disabled for now. Maybe put this in the test lib somewhere. */
   /*shell_add_cmd("slab_stats", "slab_stats", "List slab statistics.\n",
     slab_print_statistics);*/

   /* Create the slab from which we allocate other slab caches. */
   create_slab_allocator("slab_caches", sizeof(struct slab_cache),
                         0, 0, NULL, NULL, 0);

   /* Initialize static slab metadata */
   bzero(slab_static_caches, sizeof(struct slab_cache*) *
         SLAB_STATIC_NUM_SIZES * SLAB_NUM_OPTIONS);
}

/*
 *===========================================================================
 *                        init_slab2
 *===========================================================================
 *
 * Description
 * -----------
 * The last initialization of the slab. Registers the MM events.
 * The MM event mechanism uses the slab, therefor it cannot be called
 * in init_slab.
 *
 */
void
init_slab2(void)
{
   struct mm_reclaim_physical_event_t mm_event;
   mm_status_t status;

   mm_event.base = 0;
   mm_event.domain = MM_EVENT_ANY_DOMAIN;
   mm_event.identifier = COMMON_MM_COMPONENT;
   mm_event.event_handler = slab_garbage_collect;
   mm_event.event_no = MM_EVENT_RECLAIM_REGION_PHYSICAL;
   mm_event.phys_mem = "RAM";
   status = ose_mm_install_handler((union mm_event_t *)&mm_event, 1);
   if (status != MM_SUCCESS)
   {
      ramlog_printf("Slab: ose_mm_install_handler failed. Status: %d\n",
                    status);
      error2(OSE_MM_EINTERNAL_ERROR, 0);
   }
}

/*
 *===========================================================================
 *                        slab_garbage_collect
 *===========================================================================
 *
 * Description
 * -----------
 * Ask the slab to garbage collect some of its unused slabs.
 *
 * Returns
 * -------
 * Inside the event union:
 * domain  The domain of the returned data.
 * base    The base address of the returned data.
 * size    The size of the returned data or zero if it could not garbage
 *         collect any data.
 * inquiry Just check how much data we could potentially remove.
 *
 */
static void
slab_garbage_collect(union mm_event_t *event_raw)
{
   struct slab_cache *slab_cache;
   struct mm_reclaim_physical_event_t *event = &event_raw->reclaim;
   size_t requested_size = event->size;

   /* The default algorihtm is to find the first best slab cache and
    * free one unused slab in it. */

   event->domain = 0;
   event->base = 0;
   event->size = 0;
   if (strncmp(event->phys_mem, MM_INTERNAL_RAM, MM_NAME_LENGTH + 1) != 0)
   {
      /* Skip everything not in phys mem "RAM". */
      return;
   }

   if (event->inquiry || no_of_freeable_bytes == 0)
   {
      event->size = no_of_freeable_bytes;
      return;
   }

   slab_cache = (struct slab_cache *)list_head(&slab_cache_list);
   while (slab_cache)
   {
      struct slab_ctrl *slab;
      Msr sr;

      spin_lock_irqsave(&slab_cache->lock, sr);
      slab = (struct slab_ctrl *)list_head(&slab_cache->slab_list);
      while (slab)
      {
         if (!in_use(slab) && slab_cache->no_of_used_entries +
             slab_cache->entries_per_slab < slab_cache->max_no_of_used_entries)
         {
            if (event->inquiry == 0)
            {
               event->base = (log_addr_t)get_beginning_of_slab(slab);
               event->size = slab_cache->mm_alloc_size;
               delete_slab(slab);
               spin_unlock_irqrestore(&slab_cache->lock, sr);
               if (!delete_region(slab_cache->options, event->base, event->size, 1))
               {
                  ramlog_printf("Slab: Failed to garbage collect a slab\n");
                  error2(OSE_MM_EINTERNAL_ERROR, 0);
               }
               return;
            }
            else
            {
               event->size += slab_cache->mm_alloc_size;
               if (event->size >= requested_size)
               {
                  spin_unlock_irqrestore(&slab_cache->lock, sr);
                  return;
               }
            }
         }
         else
         {
            /* Empty entries are at the head of the queue, so we can
             * safely skip the rest. */
            break;
         }

         slab = (struct slab_ctrl *)list_next(&slab_cache->slab_list,
                                              &slab->link);
      }
      spin_unlock_irqrestore(&slab_cache->lock, sr);

      slab_cache = (struct slab_cache *)list_next(&slab_cache_list,
                                                  &slab_cache->link);
   }
}

/*
 *===========================================================================
 *                        slab_print_statistics
 *===========================================================================
 *
 * Description
 * -----------
 * Prints some configuration and run-time statistics associated with the
 * slab cache.
 *
 */
#if 0
static int
slab_print_statistics(int argc, char **argv)
{
   struct slab_cache *slab_cache;

   printf("Slab statistics:\n\n");

   slab_cache = (struct slab_cache *)list_head(&slab_cache_list);
   while (slab_cache)
   {
      printf("Slab name: %s\n", slab_cache->name);
      printf("\tSize: %u\n", slab_cache->size);
      printf("\tEntries per slab: %u\n", slab_cache->entries_per_slab);
      printf("\tMM allocation size: 0x%x\n", slab_cache->mm_alloc_size);
      printf("\tAlignment: %x\n", slab_cache->align);
      printf("\tSpare entries: %u\n", slab_cache->spare_entries);
      if (slab_cache->spare_entries)
      {
         printf("\tMax used spare entries: %u\n",
                slab_cache->max_used_spare_entries);
      }
      printf("\tNo of slabs allocated: %u (%u KiB)\n", slab_cache->no_of_slabs,
             (slab_cache->no_of_slabs * slab_cache->mm_alloc_size) / 1024);
      if (slab_cache->no_of_used_entries * slab_cache->size < 1024)
      {
         printf("\tAllocated entries: %u (%u Bytes)\n",
                slab_cache->no_of_used_entries,
                slab_cache->no_of_used_entries * slab_cache->size);
      }
      else
      {
         printf("\tAllocated entries: %u (%u KiB)\n",
                slab_cache->no_of_used_entries,
                (slab_cache->no_of_used_entries * slab_cache->size) / 1024);
      }
      printf("\tPercentage used: %u%%\n", (100 * slab_cache->no_of_used_entries
                * slab_cache->size) /
             (slab_cache->no_of_slabs * slab_cache->mm_alloc_size));
      printf("\n");

      slab_cache = (struct slab_cache *)list_next(&slab_cache_list,
                                                  &slab_cache->link);
   }

   return 1;
}
#endif

/*
 *===========================================================================
 *                        compare_buf_ctrl
 *===========================================================================
 *
 * Description
 * -----------
 * Local function for comparing two buffer control entries.
 *
 */
static int
compare_buf_ctrl(const void *a_raw, const void *b_raw)
{
   const struct buf_ctrl *a = (const struct buf_ctrl *)
      ((ptrdiff_t)a_raw - sizeof(list_link_t) - sizeof(struct buf_ctrl *));
   const struct buf_ctrl *b = (const struct buf_ctrl *)
      ((ptrdiff_t)b_raw - sizeof(list_link_t) - sizeof(struct buf_ctrl *));

   if (a->buf == b->buf)
   {
      return 1;
   }
   else
   {
      return 0;
   }
}

/*
 *===========================================================================
 *                     generate_buf_ctrl_key
 *===========================================================================
 *
 * Description
 * -----------
 * Local function for generating a hash table key from a
 * buffer pointer.
 *
 */
static uint32_t
generate_buf_ctrl_key(const void *a_raw)
{
   const struct buf_ctrl *a = (const struct buf_ctrl *)
      ((ptrdiff_t)a_raw - sizeof(list_link_t) - sizeof(struct buf_ctrl *));

   return (uint32_t)a->buf;
}


/*
 *===========================================================================
 *                     size_to_static_index
 *===========================================================================
 *
 * Description
 * -----------
 * Utility function for size to index conversion, used by the
 * static allocator to locate the proper slab cache.
 *
 * Returns
 * -------
 * The index, or SLAB_STATIC_INVALID_SIZE if the size cannot be accomodated.
 *
 */
static uint32_t
size_to_static_index(size_t size)
{
   uint32_t index;

   for (index = 0; index < SLAB_STATIC_NUM_SIZES; index++)
   {
      if (size < slab_static_sizes[index])
      {
         return index;
      }
   }

   return SLAB_STATIC_INVALID_SIZE;
}

#ifdef ASSERT
/*
 *===========================================================================
 *                     slab_debug_ctor
 *===========================================================================
 *
 * Description
 * -----------
 * Local function verifying that slab buffers aren't written to after free.
 * 
 */
static void
slab_debug_ctor(void *ptr, size_t size)
{
   uint32_t index;
   char *buf = (char*)ptr;

   /* Verify that the debug pattern written during free is still intact.
    * Starts check at offset 4 since first word is used as link pointer
    * in the singly linked list of free buffers.*/
   for (index = 4; index < size; index++)
   {
      if (buf[index] != 0x0f)
      {
         ramlog_printf("Error: Slab buffer was written to after free.\n");
         error(OSE_MM_EINTERNAL_ERROR);
      }
   }

   /* Write an evil pattern likely to trigger errors if applications
    * don't initialize data correctly. */
   memset(ptr, 0xff, size);
}

/*
 *===========================================================================
 *                     slab_debug_dtor
 *===========================================================================
 *
 * Description
 * -----------
 * Local function verifying that slab buffers aren't freed twice.
 *
 */
static void
slab_debug_dtor(void *ptr, size_t size)
{
   uint32_t index;
   char *buf = (char*)ptr;
   
   /* Verify that the free debug pattern is not present. If so, the buffer
    * has been freed twice. */
   for (index = 0; index < size; index++)
   {
      if (buf[index] != 0x0f)
      {
         break;
      }
      else if (index == size-1)
      {
         ramlog_printf("Error: Slab buffer was freed twice.\n");
         error(OSE_MM_EINTERNAL_ERROR);
      }
   }

   /* Fill buffer with a debug pattern identifying the buffer as free. */
   memset(ptr, 0x0f, size);
}
#endif /* ASSERT */
