/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2006 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

#include "assert.h"
#include "common.h"
#include "configure.h"
#include "gmm.h"
#include "hal.h"
#include "mm_cmd.h"
#include "mm_log.h"
#include "page.h"
#include "partition.h"
#include "stdio.h"
#include "string.h"
#include "zone.h"

#define PRINT(s) log(CATEGORY_MALLOC, s)

/*
  #define ZONE_ERROR_INVALID 0
  #define ZONE_ERROR_NOTFREE 1
  #define ZONE_ERROR_ALREADYFREE 2
  
  #define ZENTRY_FREE	0x12342378
  
  #define MIN_ITEM_SIZE      8
*/

/* public to mm_internal_cmds.c
 */
OSE_SLIST_HEAD(all_zones, struct zone, all_next); 

/* Static variables. */
static OSE_TAILQ_HEAD(extend_zones, struct zone, extend_next);

int mm_pending_extend;
static char *page_curr, *extra_chance = NULL;
static unsigned page_cnt = 0;
static struct partition *zone_log_mem;


/*
 * Insert new items in zone <z> by chopping the memory buffer
 * <new_mem>.
 */
static void
insert_new_items(zone_t z, void *new_mem, size_t nbytes)
{
   unsigned nitems, i;
   void **p;

   PRINT(("insert_new_items(z=%p, new_mem=%p, nbytes=%d).\n",
          z, new_mem, nbytes));
   PRINT(("   name=%s, item size=%d\n", z->zname, z->zsize));

   assert(z != NULL);

   if (0 == nbytes)
   {
      return;
   }
   assert(nbytes >= z->zsize);
   assert(new_mem != NULL);

   nitems = nbytes / z->zsize;

   p = new_mem;
   for (i = 0; i < nitems; i++)
   {
      p[0] = z->zitems;
      z->zitems = p;
      p += (z->zsize >> 2);
   }
   z->zfreecnt += nitems;
   z->ztotal += nitems;
   z->znalloc++;
}

#define MM_ZONE_ALLOCATOR_NAME "MM-zone-allocator"
static int
break_sas_page(char **base)
{
   log_addr_t pa, la;
   hal_status_t status;
   mm_bool_t rc;
   U32  options = HAL_GLOBAL_MAPPING;

   assert(base != NULL);
   assert(gmm_state.state >= mmu_enabled);

   rc = mm_alloc_log_mem2(
      zone_log_mem, gmm_state.hal_info->page_sizes[0], &la);
   if (rc != TRUE)
   {
      return 0;
   }
   
   rc = mm_page_alloc(gmm_state.hal_info->page_sizes[0], &pa);
   if (rc != TRUE)
   {
      mm_free_log_mem(zone_log_mem, la, gmm_state.hal_info->page_sizes[0]);
      return 0;
   }
   
   /*
    * FIXME: Copy-back is probably not good for
    * ARM1176. See page table cache mode comment
    * in arm/krn/hal_mmu/ptables.h
    */
   status = hal_map_page(
      gmm_state.system_domain->hal_domain,
      la, pa,
      gmm_state.hal_info->page_sizes[0],
      options,
      gmm_state.internal_access_mode,
      DEFAULT_CACHE_MODE);
   if (status != HAL_SUCCESS)
   {
      mm_free_log_mem(zone_log_mem, la, gmm_state.hal_info->page_sizes[0]);
      mm_page_free(pa, gmm_state.hal_info->page_sizes[0]);
   }

   *base = (char *)la;

   PRINT(("break_sas_page(base=%p).\n", *base));

   return 1;
}

/*
 * Get <size> bytes of memory from page allocator.
 */
static int
get_page(char **base)
{
   int rc;
   Msr sr = 0;

   mm_pending_extend = 1;

   /*
    * First try boot region. Since this region has slack there
    * probably are pages left.
    */
   rc = mm_alloc_boot_space(gmm_state.hal_info->page_sizes[0],
                            (log_addr_t *)base);
   if (!rc)
   {
      if (gmm_state.hal_info->has_mmu)
      {
         if (gmm_state.state >= mmu_enabled)
         {
            rc = break_sas_page(base);
         }
         else if (extra_chance) {
            rc = 1;
            *base = extra_chance;
            extra_chance = NULL;
         }
         else {
            ramlog_printf("mm: Not enough MM boot heap\n");
            ramlog_printf("mm: Please add a region with the name " \
                          "\"MM-meta-data\" inside \"log_mem/RAM_SASE\" with "\
                          "size set to for example 0x10000.\n");
            ramlog_printf("mm: If this still fails, increase the size.\n");
            ramlog_printf("mm: Example of configuration "\
                          "(assume log_mem/RAM-SASE located at address 0):\n"
                          "mm: krn/region/MM-meta-data=base:0x01ff0000 " \
                          "size:0x10000 perm:su_rw_usr_na cache:copy_back " \
                          "mem:speculative_access.\n");
            MM_FATAL_ERROR2(("Not enough MM internal Heap\n"),
                          OSE_MM_ECONFIGURATION_ERROR,
                          (OSERRCODE)("Not enough MM internal heap."));
         }
      }
      else
      {
         rc = mm_page_alloc(gmm_state.hal_info->page_sizes[0],
                            (log_addr_t *)base);
      }
   }
   MM_MULTITHREAD_LOCK(sr);
   mm_pending_extend = 0;
   page_cnt++;
   MM_MULTITHREAD_UNLOCK(sr);

   PRINT(("get_page(base=%p).\n", *base));

   return rc;
}


/*
 * Allocate more items for zone <z>.
 */
static int
extend_zone(zone_t z)
{
   static size_t bytes_left = 0;
   size_t extend_size;
   Msr sr = 0;

   assert(z != NULL);

   MM_MULTITHREAD_LOCK(sr);
   extend_size = z->zsize * (z->zfreemax - z->zfreecnt);

   PRINT(("extend_zone(z=%p). extend_size=%#x zone_name=%s bytes_left=%lu\n",
          z, extend_size, z->zname, bytes_left));
   
   while (extend_size)
   {
      size_t size;

      if (extend_size <= bytes_left)
      {
         size = extend_size;
      }
      else if (bytes_left >= z->zsize)
      {
         size = (bytes_left / z->zsize) * z->zsize;
      }
      else
      {
         size = 0;
         bytes_left= 0;
      }

      insert_new_items(z, page_curr, size);
      
      extend_size -= size;
      bytes_left -= size;
      if (0 == bytes_left)
      {
         MM_MULTITHREAD_UNLOCK(sr);
         if (!get_page(&page_curr))
         {
            return 0;
         }
         MM_MULTITHREAD_LOCK(sr);
         bytes_left = gmm_state.hal_info->page_sizes[0];
      }
      else
      {
         page_curr += size;
      }
   }

   MM_MULTITHREAD_UNLOCK(sr);
   return 1;
}


/*
 * Try to allocate more items for zone <z>.
 * If we have recursed, put the zone on the waiting list and
 * return. Since every zone has a low watermark with enough free
 * items, this recursive behaviour is OK.
 */
static int
get_more(zone_t z)
{
   Msr sr = 0;
   
   assert(z != NULL);

   MM_MULTITHREAD_LOCK(sr);
   if (mm_pending_extend)
   {
      /*
       * We have recursed. Put zone into waiting queue and return.
       */
      if (!(z->flags & ZONE_QUEUED_FOR_EXTEND))
      {
         OSE_TAILQ_INSERT_TAIL(&extend_zones, z);
         z->flags |= ZONE_QUEUED_FOR_EXTEND;
         assert(z->zfreecnt != 0);
      }
      MM_MULTITHREAD_UNLOCK(sr);
      return 1;
   }
   MM_MULTITHREAD_UNLOCK(sr);

   /* Extend primary zone. */
   if (!extend_zone(z))
   {
      return 0;
   }

   /* Extend zones that have recursed. */
   while (1)
   {
      struct zone *zone;

      MM_MULTITHREAD_LOCK(sr);
      if (OSE_TAILQ_EMPTY(&extend_zones))
      {
         MM_MULTITHREAD_UNLOCK(sr);
         break;
      }
      
      zone = OSE_TAILQ_FIRST(&extend_zones);
      OSE_TAILQ_REMOVE(&extend_zones, zone);
      zone->flags &= ~ZONE_QUEUED_FOR_EXTEND;
      MM_MULTITHREAD_UNLOCK(sr);
         
      /* Current (primary) zone has already been extended. */
      if (zone != z)
      {
         if (!extend_zone(zone))
         {
            return 0;
         }
      }
   }

   return 1;
}


/*
 * See zone.h
 */
void
mm_zfree(zone_t z, void *item)
{
   Msr sr = 0;
   
   assert(z != NULL);
   assert(item != NULL);
   assert(gmm_state.state >= zone_ok);

#ifdef ASSERT
   memset(item, 0xff, z->zsize);
#endif

   PRINT(("mm: mm_zfree(z): zfreecnt: %#x zfreemin: %#x name: %s\n",
          z->zfreecnt,  z->zfreemin, z->zname));

   MM_MULTITHREAD_LOCK(sr);
   ((void **) item)[0] = z->zitems;
   z->zitems = item;
   z->zfreecnt++;
   MM_MULTITHREAD_UNLOCK(sr);
}


/*
 * See zone.h
 */
void *
mm_zalloc(zone_t z)
{
   void *item;
   Msr sr = 0;

   assert(z != NULL);
   assert(gmm_state.state >= zone_ok);

   PRINT(("mm: mm_zalloc(z): zfreecnt: %#x zfreemin: %#x name: %s\n",
          z->zfreecnt,  z->zfreemin, z->zname));

   if (0 == z->zfreecnt)
   {
      return NULL;
   }

   if (z->zfreecnt <= z->zfreemin)
   {
      if (!get_more(z))
      {
         return NULL;
      }
   }

   MM_MULTITHREAD_LOCK(sr);
   if (0 == z->zfreecnt)
   {
      MM_MULTITHREAD_UNLOCK(sr);
      return NULL;
   }

   item = z->zitems;
   z->zitems = ((void **) item)[0];
   assert(item != NULL);
   z->zfreecnt--;
   z->znalloc++;
   MM_MULTITHREAD_UNLOCK(sr);

   return item;
}

/*
 * See zone.h
 */
void
mm_zinit(zone_t z,
         const char *name,
         size_t item_size,
         unsigned min_item_count,
         unsigned max_item_count)
{
   static U32 extra_chance_initiated = 0;

   assert(z != NULL);
   assert(name != NULL);

   bzero((void *)z, sizeof *z);
   z->zname = name;
#ifdef OSMIPS
   z->zsize = ALIGN(item_size, 8); /* Maybe redundant for boot init... */
#else
   z->zsize = ALIGN(item_size, 4); /* Maybe redundant for boot init... */
#endif
   z->zitems = NULL;
   z->zfreemin = min_item_count;
   z->zfreemax = max_item_count;

   OSE_SLIST_INSERT_HEAD(&all_zones, z);

   PRINT(("mm: mm_zinit(name: %s, item_size %#x"
          "min_item_count: %#x,  max_item_count %#x)\n",
          name, item_size, min_item_count, max_item_count));


   /* Fill zone with items. */
   if (!get_more(z))
   {
      INTERNAL_ERROR();
   }
   
   if (!extra_chance_initiated) {
      int rc = mm_alloc_boot_space(gmm_state.hal_info->page_sizes[0],
                                   (log_addr_t *)&extra_chance);
      if (!rc) {
         INTERNAL_ERROR();
      }
      if (extra_chance == NULL) {
         INTERNAL_ERROR();
      }
      extra_chance_initiated = 1;
   }
}


/*
 * See zone.h
 */
void
mm_zone_late_init(const char *zone_log_mem_name)
{
   if (hal_get_info()->has_mmu)
   {
      zone_log_mem = mm_find_domain_partition_by_name(
         gmm_state.system_domain, zone_log_mem_name);

      if (NULL == zone_log_mem)
      {
         INTERNAL_ERROR();
      }
   }
}

unsigned
mm_zone_get_page_cnt(void)
{
   return page_cnt;
}


