/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2012 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

#ifndef _PARTITION_H
#define _PARTITION_H

/* C runtime include files. */
#include <stddef.h>

/* OSE public include files. */
#include "list_ext.h"
#include "mm.sig"
#include "osetypes.h"



/*
 * partition.h - The logical memory allocator interface.
 *
 * A partition is the internal representation of a log_mem configuration
 * item. Partitions can manage several distinct ranges of logical_memory
 * called banks in a user transparent manner. 
 *
 * 
 */
struct partition_bank;
struct partition_class;
struct partition_alias;
struct partition; /* Instance of partition_class */
struct log_object;
struct domain;


/*
 * Bits below are combined to form the configuration of a partition. 
 *
 * Explanation:
 * 
 * manager_mode  - ARM specific feature. Allows system software to
 *                 relax write permissions of regions mapped within
 *                 the partition.
 * non-exclusive - Allocations from the partition does only affect the
 *                 address space of the domain requesting the allocation.
 *                 Another domain can still allocate the same logical memory.
 * exclusive     - Allocations from the partition affects the address
 *                 space all domains. Occupied logical memory cannot
 *                 be allocated by any domain in the system. 
 * physical      - The partition will only distribute logical memory where
 *                 there is free physical memory at the same address.
 * virtual       - The partition can distribute logical memory independently
 *                 of where there is free physical memory. 
 * implicit      - The partition has implicitly configured banks. Currenlty
 *                 only supported for exlusive, physical partitions.
 *                 The partition will distribute logical memory where no
 *                 other partition is configured and there is physical memory
 *                 beneath.
 * default       - Flag used to indicate the default alias/mapping type.
 *                 The default mapping type is required in combined
 *                 partitions, where there might be multiple mapping types,
 *                 if an alias name is unavailable during an
 *                 operation, for example due to the use of
 *                 LOG_MEM_FROM_ADDRESS.
 * resident      - The partition is statically mapped with resident
 *                 MMU pages. It will emulate the nommu memory model.
 *                 
 *
 ******************************************************************************
 *
 * Flags can be combined if a partition supports more than one
 * traditional mapping type, i.e. a partition can be both exclusive and
 * non-exclusive, as well as both virtual and physical. See
 * MM_PARTITION_MAPTYPE_MASK for a list of flags that can be combined. 
 * 
 */
#define MM_PARTITION_OPT_MANAGER_MODE    0x0001
#define MM_PARTITION_OPT_NON_EXCLUSIVE   0x0002
#define MM_PARTITION_OPT_EXCLUSIVE       0x0004
#define MM_PARTITION_OPT_PHYSICAL        0x0008
#define MM_PARTITION_OPT_VIRTUAL         0x0010
#define MM_PARTITION_OPT_IMPLICIT        0x0040
#define MM_PARTITION_OPT_DEFAULT         0x0080
#define MM_PARTITION_OPT_RESIDENT        0x0100
#define MM_PARTITION_OPT_INITIALIZED     0x0200
#define MM_PARTITION_OPT_COMBINED        0x0006

/*
 * Option bits for the alloc_f allocation algorithm:
 * None of the below flags:         Return only an exact match.
 * MM_PARTITION_OPT_SEARCH_FORWARD  Return the requested address or the first
 *                                  free address after of the requested.
 * MM_PARTITION_OPT_SEARCH_REVERSE  Return the requested address or the first
 *                                  free address before of the requested.
 * Both of the above flags:         Return the free address that is nearest the
 *                                  requested address.
 */
#define MM_PARTITION_OPT_SEARCH_FORWARD  0x0400
#define MM_PARTITION_OPT_SEARCH_REVERSE  0x0800
#define MM_PARTITION_OPT_SEARCH          (MM_PARTITION_OPT_SEARCH_FORWARD | \
                                          MM_PARTITION_OPT_SEARCH_REVERSE)


#define MM_PARTITION_MAPTYPE_MASK        (MM_PARTITION_OPT_NON_EXCLUSIVE | \
                                          MM_PARTITION_OPT_EXCLUSIVE     | \
                                          MM_PARTITION_OPT_PHYSICAL      | \
                                          MM_PARTITION_OPT_VIRTUAL)

#define MM_PARTITION_OPT_SASE   (MM_PARTITION_OPT_EXCLUSIVE| \
                                 MM_PARTITION_OPT_PHYSICAL)
#define MM_PARTITION_OPT_SAS    (MM_PARTITION_OPT_EXCLUSIVE| \
                                 MM_PARTITION_OPT_VIRTUAL)
#define MM_PARTITION_OPT_MAS    (MM_PARTITION_OPT_NON_EXCLUSIVE| \
                                 MM_PARTITION_OPT_VIRTUAL)

#define partition_is_manager_mode(partition)   \
   (((partition)->options & MM_PARTITION_OPT_MANAGER_MODE) != 0)
#define partition_is_exclusive(partition)      \
   (((partition)->options & MM_PARTITION_OPT_EXCLUSIVE) != 0)
#define partition_is_non_exclusive(partition)  \
   (((partition)->options & MM_PARTITION_OPT_NON_EXCLUSIVE) != 0)
#define partition_is_physical(partition)       \
   (((partition)->options & MM_PARTITION_OPT_PHYSICAL) != 0)
#define partition_is_virtual(partition)        \
   (((partition)->options & MM_PARTITION_OPT_VIRTUAL) != 0)
#define partition_is_implicit(partition)       \
   (((partition)->options & MM_PARTITION_OPT_IMPLICIT) != 0)
#define partition_is_combined(partition)       \
   (((partition)->options & MM_PARTITION_OPT_COMBINED) == MM_PARTITION_OPT_COMBINED)
#define partition_is_resident(partition)       \
   (((partition)->options & MM_PARTITION_OPT_RESIDENT) != 0)
#define partition_is_initialized(partition)       \
   (((partition)->options & MM_PARTITION_OPT_INITIALIZED) != 0)
#define partition_is_sase(partition)       \
   (((partition)->options & MM_PARTITION_OPT_SASE) == MM_PARTITION_OPT_SASE)

/*
 * How the memory is intended to be mapped, e.g. SASE/SAS/MAS.
 */
typedef enum MappingType mapping_type_t;



/*
 * A partition bank is a range of logical memory managed by a single partition.
 */
struct partition_bank
{
   struct partition_bank *next;         /* Linked into list of banks */
   log_addr_t      base;
   log_addr_t      top;
};


/* A partition class may consist of one or several aliases. An alias
 * is used to identify a partition with a name and connect that name
 * with a mapping type supported by the partition. */
struct partition_alias
{
   struct partition_alias *next;
   uint32_t        options;              /* Supported features          */
   char            name[MM_NAME_LENGTH]; /* The name                    */
};


/*
 * A partition class is the direct representation
 * of the system configuration. <partition_class>es
 * are instantiated into <partition>s when new domains
 * are created.
 *
 */
struct partition_class
{
   struct partition_class  *next;        /* Linked into list of configs. */
   struct partition_bank   *banks;       /* List of partition banks.     */
   struct partition_alias  *aliases;     /* List of partition aliases.   */
   uint32_t                 options;     /* See option attributes above. */
   char                     name[MM_NAME_LENGTH]; /* The name            */

   /* These attributes are not required if not mapped resident. */
   uint32_t                 res_perm;    /* Resident map permissions     */
   uint32_t                 res_cache;   /* Resident map cache mode      */
   uint32_t                 res_options; /* Resident map (BAT reg) type. */

   /* Allocates and initializes a new partition configuration.
    *
    * The function is called from mm_add_partition to allow the
    * allocator to create and initialize global allocator data. */
   struct partition_class *
   (*init)(uint32_t options);

   /* Allocates and initalizes a new partition bank.
    *
    * The function is called from mm_add_partition_bank to allow the
    * allocator to update internal data structures with the added logical
    * space. 
    */
   struct partition_bank *
   (*init_bank)(struct partition_class *config,
                log_addr_t base, log_size_t size);

   /* Instantiate a new partition from the configuration. Must be called
    * once for each partition config and domain. Allocates and
    * initializes metadata that is private to each domain. */
   struct partition *
   (*create)(struct partition_class *config);


   /* Called when a domain is destroyed for each instantiated
    * partition template. */
   void
   (*destroy)(struct partition *);


   /* Returns TRUE if the memory handled by the log_mem is a subset of the
    * range specified by base & size. Otherwise FALSE is returned. The partition
    * is considered a subset even if it is exactly the range specified.
    */
   int
   (*is_subset_of)(const struct partition_class *,
                   log_addr_t base, log_size_t size);

   /* Returns TRUE if the memory handled by the log_mem is a superset of the
    * range specified by base & size. Otherwise FALSE is returned. The partition
    * is considered a superset even if it is exactly the range specified.
    */
   int
   (*is_superset_of)(const struct partition_class *,
                     log_addr_t base, log_size_t size);

   /* Returns TRUE if the memory handled by the log_mem is entirely covered by
    * physical memory. Used during system start to check SASE mapped log_mem
    * entries. */
   int
   (*is_subset_of_phys_mem)(const struct partition_class *);

   /* Returns TRUE if any address in the range specified by base & size is
    * handled by partition. Otherwise FALSE is returned.
    */
   int
   (*overlap)(const struct partition_class *, log_addr_t base, log_size_t size);

   /*
    * Retrieves information about the location of a partition bank. Returns
    * 0 when the bank index number is out of bounds. Indexes start at 0.
    */
   int
   (*get_bank_info)(const struct partition_class *, int index,
                    log_addr_t *base, log_size_t *size);
   /*
    * Allocation methods follow below. 
    *
    * Requirements:
    *   - Regular allocation
    *   - Explicit allocation with base address 
    *   - Aligned allocation
    *   - Allocation from left/right/top/bottom
    *   - Exclusive allocation
    *
    * Options:
    *   - Exclusive 
    *   - Left vs Right
    *
    * Dependencies:
    *   - Zone allocator (preferrably distributing SASE-mapped memory)
    *
    * Notes:
    *   log_obj should be pre-allocated by GMM and is filled in by
    *   the logical allocator. It is the responsibility of GMM to
    *   manage the shared logical resource properly through reference
    *   counters or similar when a region has been shared/duplicated.
    */

   /* Allocate aligned */
   mm_status_t
   (*alloc)(struct partition  *partition,
            struct log_object *log_obj,      /* Result */
            log_size_t size, log_size_t alignment, uint32_t options);
   
   /* Allocate near fixed address. */
   mm_status_t
   (*alloc_f)(struct partition *partition,
              struct log_object *log_obj,      /* Result */
              log_size_t size, log_addr_t fixed, uint32_t options);

   /*
    * Free a used logical range
    *
    * Notes:
    *    GMM should free the logical range only when no region
    *    no longer use it. Once freed, the log_obj should be destroyed.
    */
   void
   (*free)(struct partition *partition,
           struct log_object *log_obj, uint32_t options);
};



/*
 * Partition - The instantiated partition configuration.
 *
 * Each configuration item is instantiated in all new domains.
 * This allows the allocator to keep private data for each domain.
 *
 * Allocators should embed this structure in their own specific data
 * structure, e.g:
 *
 * struct mas_partition {
 *    struct partition   partition;    Common  allocator data  
 *    struct list_t      free_list;    Private allocator data 
 * }
 * 
 */
struct partition
{
   struct partition        *next;
   struct partition_class  *class;
};




/*
 * Must be called before FIXME.
 */
void
mm_partition_init(void);

/*
 * Prepare for warm start.
 */

void
mm_partition_prepare_for_warm_start(U32 regCnt);

/*
 * Insert a partition configuration descriptor into partition list.
 *
 * To be fully functional (and usable), at least one alias and one
 * bank must be added to new partition. The alias is needed to
 * identify the partition. Without a bank the partition will never
 * be able to distribute any logical address space. 
 */
struct partition_class *
mm_add_partition(uint32_t options);

/*
 * Adds a new bank to a partition class. 
 */
mm_status_t 
mm_add_partition_bank(struct partition_class *config,
                      log_addr_t base, log_size_t size);

/*
 * Adds a new bank to a partition class. 
 */
mm_status_t 
mm_add_partition_alias(struct partition_class *config,
                       const char *name, uint32_t options);

/*
 * Allocate a sas partition in the space between all existing
 * partitions. True is returned if the operation was successful.  If
 * implicit_ram_sase is true the HAL log_mem can not be placed over
 * phys_mem
 */
struct partition_class *
mm_create_hal_partition(const char *name,
                        log_size_t size,
			int implicit_ram_sase);

/*
 * Search for a partition descriptor with name <name> in a list <list>
 * with partition descriptors.
 * Return NULL if not found.
 */
struct partition_class *
mm_find_partition_by_name(const char *name);

/*
 * Search for a partition descriptor that contains address <addr>.
 * Return NULL if not found.
 */
struct partition_class *
mm_find_partition_by_address(log_addr_t addr);


/*
 * Search for a SASE partition that overlaps with the given range.
 * Return NULL if not found.
 */
struct partition_class *
mm_find_sase_partition_by_range(log_addr_t addr, log_size_t size);


/*
 * Look up the mapping type of a partition given an alias name.
 *
 * If 'alias_name' is NULL, the mapping type of the
 * default alias is returned.
 *
 * If 'alias_name' is an empty string, the mapping type of the
 * default alias is returned and the string updated with its name.
 *
 * Returns !MM_SUCCESS if the alias is unknown or no default alias exist.
 */
mm_status_t
mm_find_partition_mapping_type(const struct partition_class *p_class,
                               char *alias_name, mapping_type_t *type);

/*
 * Look up a partition alias given an alias name. 
 * 
 * If 'alias_name' is NULL, the default alias is returned.
 *
 * If 'alias_name' is an empty string, the default alias is
 * returned and the string updated with its name.
 *
 * Returns NULL if not found.
 */
struct partition_alias *
mm_find_partition_alias(const struct partition_class *p_class,
                        char *alias_name);

/*
 * Look up a logical region at address <base> in partition <partition>.
 * Return it when match, otherwise NULL.
 * If exact_match != 0, <base> must == base address of
 * region. Otherwise it is enough to reside in the region.
 */
struct log_object *
mm_lookup_logreg_by_addr(const struct partition *partition,
                         log_addr_t base,
                         int exact_match);

/*
 * Return the logical region with name <name>, otherwise NULL.
 */
struct log_object *
mm_lookup_logreg_by_name(const struct partition *partition,
                         const char *name);

/*
 * Allocate a chunk of logical space.
 */
mm_bool_t
mm_alloc_log_mem2(struct partition *partition,
                  log_size_t size, log_addr_t *base);

/*
 * Allocate a chunk of logical space with specified alignment.
 */
mm_bool_t
mm_alloc_log_mem(struct partition *partition,
                 log_size_t size, log_size_t alignment, log_addr_t *base);

/*
 * Free a piece of logical space.
 */
void
mm_free_log_mem(struct partition *partition,
		log_addr_t base, log_size_t size);

/*
 * Instantiates all partition configurations in a newly created domain.
 */
mm_status_t
mm_create_domain_partitions(struct domain *domain);

/*
 * Destroys all instantiated partitions in the domain.  
 */
mm_status_t
mm_destroy_domain_partitions(struct domain *domain);

/*
 * Finds a domain's private logical allocator (partition) by name.
 *
 * Global SAS and SASE capable partitions are searched as well. 
 */
struct partition *
mm_find_domain_partition_by_name(const struct domain *domain,
                                 const char *name);

/*
 * Finds a domain's private logical allocator (partition) by address. 
 *
 * Global SAS and SASE capable partitions are searched as well. 
 */
struct partition *
mm_find_domain_partition_by_addr(const struct domain *domain,
                                 log_addr_t base);

/*
 * Called during system startup to map all resident partitions.
 * If resident pages are not supported, partitions are mapped
 * globally in the system domain using regular dynamic pages.
 */
void
mm_map_resident_partitions(void);

#endif /* _PARTITION_H */
