/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2006 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

#include "stdint.h"

#define SYS_TRACE_NAME "OSe_trace"

#define LAST_TRACE_ENTRY_NBR (24)
#define HIDDEN_TRACE_ENTRY_OFFSET 100

/*
** These are the trace entries recognised by the rmm.
** Other trace entries are ignored.
*/

/*
** This is a trace record.
** It contains a logged signal send or receive event.
*/

#define TRACE_SEND (4)
#define TRACE_RECEIVE (5)

struct trace_send_receive
{
   uint16_t        event_no;
   uint16_t        sigsize;
   struct abstime *abstime;
   struct trace_procname *from;
   struct trace_procname *to;
   uint16_t        cpu;
   uint16_t        dummy;
   uint8_t         data[1];     /* variable size. */
};

/*
** This is a trace record.
** It contains a logged user error event.
*/

#define TRACE_USER_ERROR (11)

struct trace_user_error
{
   uint16_t        event_no;
   uint16_t        cpu;
   struct abstime *abstime;
   struct trace_procname *process;
   uint32_t        ecode;
   uint32_t        extra;
};

/*
** This is a trace record.
** It contains a logged OSE error event.
*/

#define TRACE_OSERROR (6)

struct trace_ose_error
{
   uint16_t        event_no;
   uint16_t        cpu;
   struct abstime *abstime;
   struct trace_procname *process;
   uint32_t        ecode;
   uint32_t        extra;
};

/*
** This is a trace record.
** It contains a logged create or kill process event.
*/

#define TRACE_CREATE (1)
#define TRACE_KILL (2)

struct trace_create_kill
{
   uint16_t        event_no;
   uint16_t        cpu;
   struct abstime *abstime;
   struct trace_procname *process;
};

/*
** This is a trace record.
** It contains a logged swap event.
*/

#define TRACE_SWAP (3)

struct trace_swap
{
   uint16_t        event_no;
   uint16_t        cpu;
   struct abstime *abstime;
   struct trace_procname *to_process;
   struct trace_procname *from_process;
};

/*
** This is a trace record
** It contains a logged user event.
*/

#define TRACE_USER_EVENT (13)

struct trace_user_event
{
   uint16_t        event_no;
   uint16_t        event_size;
   struct abstime *abstime;
   struct trace_procname *process;
   uint32_t        requested_size;
   uint32_t        event_id;
   uint16_t        cpu;
   uint16_t        dummy;
   uint8_t         data[1];     /* variable size. */
};

#define TRACE_ALLOC (15)
#define TRACE_FREE_BUF (16)

struct trace_alloc_free_buf
{
   uint16_t        event_no;
   uint16_t        sigsize;
   struct abstime *abstime;
   struct trace_procname *process;
   uint32_t        address;
   uint32_t        requested_size;
   uint16_t        cpu;
   uint16_t        dummy;
   uint8_t         data[1];
};
  
/*
** This is a trace record.
** It contains a logged bind event.
*/

#define TRACE_BIND (14)

struct trace_bind
{
   uint16_t        event_no;
   uint16_t        dummy;       /* To force equal alignment on all compilers. */
   struct abstime *abstime;
   struct trace_procname *process;
   uint16_t        from_eu;
   uint16_t        to_eu;
};

/*
** This is a trace record.
** It contans a logged timeout event.
*/
#define TRACE_RECEIVE_TMO (24)

struct trace_receive_tmo
{
   uint16_t        event_no;
   uint16_t        cpu;
   struct abstime *abstime;
   struct trace_procname *process;
   uint32_t        timeout;   
};


/*
** This is a trace record.
** It contains a logged system start event.
**
** This entry is the highest visible trace entry.
** This fact is used in the code.
*/

#define TRACE_SYSTEM_START (19)

struct trace_system_start
{
   uint16_t        event_no;
   uint8_t         warm_flag;
   uint8_t         reserved;
   struct abstime *abstime;
};

/*
** This is a subtrace record.
** It contains a process name record.
**
** This record is not directly visible in the trace. It is referenced from
** other trace records.
*/

#define TRACE_PROCNAME (HIDDEN_TRACE_ENTRY_OFFSET + 20)

struct trace_procname
{
   uint16_t        event_no;
   uint8_t         namelen;
   uint8_t         priority;    /* Copy of the priority field in the pcb. */
   uint32_t        pid;
   char            name[2];     /* variable size for two names. */
};

/*
** This is a subtrace record.
** It contains an absolute time stamp.
** If abs_sec = 0, then the time was never set.
**
** This record is not directly visible in the trace. It is referenced from
** other trace records.
*/

#define TRACE_ABSTIME (HIDDEN_TRACE_ENTRY_OFFSET + 21)

struct abstime
{
   uint16_t        event_no;
   uint16_t        dummy;       /* To force equal alignment on all compilers. */
   uint32_t        abs_sec;     /* Seconds since 0:0:0 1 jan-1970 GMT. */
   uint32_t        abs_tick;    /* The corresponding system tick value. */
   uint32_t        abs_usec;    /* The corresponding microseconds subtick value. */
};

/*
** This is a trace record.
** This entry is special.
**
** it is not used to store events. Instead, it is used when
** allocating dynamic space that should survive warm restarts.
** Such data is used for instance when storing action lists.
**
** The trace_hidden event is only used when hidden space has been
** freed. Up to that point, the data is invisible anyway since
** the display trace operations do not recognize buffers that are
** still allocated.
**
*/

#define TRACE_HIDDEN (HIDDEN_TRACE_ENTRY_OFFSET + 22)

struct trace_hidden
{
   uint16_t        event_no;
   uint8_t         data[1];     /* variable size. */
};

#define TRACE_NOT_USED (0xFFFF)

struct trace_not_used
{
   uint16_t        event_no;
   uint8_t         data[1];     /* variable size. */
};



/*
** The trace union defines all records that can occur in the trace buffer.
*/

union TRACE
{
   uint16_t        event_no;
   struct trace_procname trace_procname;
   struct trace_send_receive trace_send_receive;
   struct trace_swap trace_swap;
   struct trace_user_error trace_user_error;
   struct trace_ose_error trace_ose_error;
   struct trace_create_kill trace_create_kill;
   struct trace_system_start trace_system_start;
   struct trace_user_event trace_user_event;
   struct trace_bind trace_bind;
   struct trace_alloc_free_buf trace_alloc_free_buf;
   struct abstime  abstime;
   struct trace_hidden trace_hidden;
   struct trace_not_used trace_not_used;
   struct trace_receive_tmo trace_receive_tmo;   
};

