/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2005-2007 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

/**
 * @file _zzmain.c
 *
 * @brief This file contains the _zzmain process, which
 * replaces the zzmain from osemain.c as the main entrypoint
 * scheduled as the first process from POSE.
 *
 * The process acts in three alternative ways, depending on if
 * a valid sysparam configuration is found.
 *
 * 1: No system parameters for any options are found. The process just continues into
 *    the zzmain entrypoint in osemain.c WITHOUT starting the GMM process.
 *    => the "legacy" mode that works for old reference systems.
 *
 * 2. If a syspool extender or gmm configuration is found, create GMM in the
 *    system pool and initializes the syspool extender if that is configured.
 *    Then continue into zzmain in osemain.c.
 *
 * 3. A mainpool config is found. The process does 2, allocates and creates a osemain pool,
 *    attaches it to a new osemain block, and creates the osemain zzmain function
 *    in that block. It then initializes the mainpool extender in that block and
 *    starts the new osemain process (using entry zzmain). 
 *    This procedure causes the zzmain function is osemain.con to
 *    be executed in the context of the main pool.
 */

#include "ose.h"    /* OSE system calls. */
#include "osetypes.h"
#include "pose.h"
#include "mm.h"
#include "ramlog.h"
#include "sysparam.h"
#ifdef softOSE
#include "purify_param_conf.h"
#endif
#include "efs.h"
#include "thinlib.h"
#include "ose_smp.h"

/* Core internal kernel services. */
#include "core_internal.h"

/* default stack size and priority for GMM if that shall be started.
 */

#define GMM_STACK 2000
#define GMM_PRIO  20

union SIGNAL
{
   SIGSELECT sig_no;
   struct mm_create_region mmap;
};

/*
 * function that initializes pthread biosmodule
 */

void ose_init_pthreads(void);

/*
 * function that can be used to determine of kernel is up.
 */

OSBOOLEAN ose_started(void);

/*
 * the zzmain (osemain.c) and this entrypoint.
 */

OSENTRYPOINT zzmain;
OSENTRYPOINT _zzmain;

/*
 * the "secret" call to Target Interface is needed.
 */

void zzinit_OSE(void);

/*
 * These functions initializes the pool extender.
 */

void 
init_syspool_extender(OSPRIORITY frag_prio,
                       Boolean user_allowed,
                       OSADDRESS sysfrag_size);

void 
init_mainpool_extender(PROCESS mainpool_pid,
                       PROCESS main_bid,
                       OSPRIORITY frag_prio,
                       Boolean user_allowed,
                       OSADDRESS mainfrag_size);

/*
 * configuration vales, of which the sizes also denotes "enabled".
 */

static  OSADDRESS  mainpool_size = 0x0;
static  OSADDRESS  sysfrag_size = 0x0;
static  OSADDRESS  mainfrag_size = 0x0;
static  Boolean    sys_user_allowed = False;
static  Boolean    main_user_allowed = False;
static  OSADDRESS  gmm_stacksize = GMM_STACK;
static  OSPRIORITY gmm_prio = GMM_PRIO;
static  Boolean    mainpool_fill = False;
static  U32        mainpool_cache;

/* if this flag is set, we will create GMM in system block.*/

static OSBOOLEAN   start_gmm = False;

static OSBOOLEAN ose_running = False;

OSBOOLEAN ose_started(void)
{
   return ose_running;
}

static void set_ose_started(void)
{
   ramlog_printf("kernel is up.\n");
   ose_running = True;
}

/*
 * Function that tries to extract the configuration values above.
 * It also sets the start_gmm flag if configuration is detected.
 */

static void read_config(void)
{
   OSADDRESS fragsize=0;
   unsigned long int prio = (unsigned long int)gmm_prio;

   /*
    * look for gmm params.
    */
   if(!sysparam_readattr_long("krn/gmm", "stack_size", (long *) &gmm_stacksize) &&
      !sysparam_readattr_long("krn/gmm", "prio", (long *) &prio))
   {
      start_gmm = True;
      gmm_prio = (OSPRIORITY)prio;
   }

   /*
    * syspool extend enabled?
    */
   if (!sysparam_readattr_long("krn/syspool", "frag_size", (long *) &sysfrag_size) &&
       (sysfrag_size > 0))
   {
      start_gmm = True;
   }
   
   /* user mode parameter optional for system pool. */
   sysparam_readattr_bool("krn/syspool", "usermode", &sys_user_allowed);

   /*
    * all attributes must be there to enable mainpool.
    */
   if(!sysparam_readattr_long("krn/mainpool", "init_size", (long *) &mainpool_size) &&
      !sysparam_readattr_long("krn/mainpool", "frag_size", (long *) &fragsize) &&
      !sysparam_readattr_bool("krn/mainpool", "usermode", &main_user_allowed))
   {
      char ctype[MM_NAME_LENGTH+1];
      char mtype[MM_NAME_LENGTH+1];

      /* Should we fill the mainpool with a debug pattern? */
      Boolean fill;
      if(!sysparam_readattr_bool("krn/mainpool", "pattern_fill", &fill))
      {
         mainpool_fill = fill;
      }
      start_gmm = True;
      mainfrag_size = fragsize;

      /* Configure mainpool cache and memory attributes. */
      strlcpy(ctype, "copy_back", MM_NAME_LENGTH);
      strlcpy(mtype, "speculative_access", MM_NAME_LENGTH);
      sysparam_readattr_string("krn/mainpool", "cache", ctype, MM_NAME_LENGTH);
      sysparam_readattr_string("krn/mainpool", "mem",   mtype, MM_NAME_LENGTH);
      if(ose_mm_get_cache_mode(ctype, mtype, &mainpool_cache))
      {
         mainpool_cache = MM_CACHE_COPY_BACK | MM_SPECULATIVE_ACCESS;
      }
   }   
}

static void start_mm(void)
{
   static const SIGSELECT sigsel[] = { 1, MM_INTERFACE_INFO_REPLY };
   union SIGNAL *sig;
   extern OSENTRYPOINT ose_memory_mgr;
   PROCESS ose_mm_pid;
   
   ose_mm_pid = create_process(OS_PRI_PROC,
                               "ose_memory_mgr",
                               ose_memory_mgr,
                               (OSADDRESS) gmm_stacksize,
                               (OSPRIORITY) gmm_prio,
                               (OSTIME)0,
                               (PROCESS) 0,
                               (struct OS_redir_entry *) 0,
                               (OSVECTOR) 0,
                               (OSUSER) 0);
   
   start(ose_mm_pid);
   
   /* Wait for MM to start servicing requests. */
   sig = alloc(sizeof(struct mm_interface_info_request),
               MM_INTERFACE_INFO_REQUEST);
   send(&sig, ose_mm_pid);
   sig = receive(sigsel);
   free_buf(&sig);
   
   /* Make sure we get an error if the system dies. */
   attach(0, ose_mm_pid);
}

/*
** Retrieves information about the mainpool
** Returnes True if a mainpool exists and 
** has the supplied access permissions.
**
** Only exported and used internally within core.
*/
Boolean ose_has_mainpool(Boolean user_mode);

Boolean ose_has_mainpool(Boolean user_mode)
{
  if (mainpool_size > 0)
    {
      if (main_user_allowed == False && user_mode == True)
	{
	  return False;
	}
      return True;
    }
  return False;
}

/*
** A new _zzmain entrypoint that now pose refers to.
** This process is used to create the globally available
** main pool before running the actual osemain code.
**
** The entrypoint is also the place to add core specific start
** issues such as synchronization, TI calls, etc.
*/

OS_PROCESS(_zzmain)
{
   PROCESS block_;
   PROCESS osemain_pid_;
   static const SIGSELECT sigsel[] = {0};
   union SIGNAL *sig;
   static void (*thinlib_ref)(void) = NULL;

   /*
    * Perform TI initialization that was previously done via ZZOS_MAIN macro.
    */  
#ifdef softOSE
   /* Reads the Purify Config parameters from the .conf file */
   read_purify_param_conf();
#endif   
   zzinit_OSE();
   read_config();

   /*
    * Just to make sure we link in the shared block functionality in
    * the monolith.
    */
   (void)efs_get_ose_shared_block();

   /* start GMM in system block if "new" configuration policy. */
   
   if(start_gmm)
   {
      start_mm();
   }

   /* if syspool frag size found, enable it.*/
   
   if(sysfrag_size)
   {
      init_syspool_extender(gmm_prio, sys_user_allowed, sysfrag_size);
   }

   /* If mainpool enabled, create new application block/process
    * and let zzmain start there.
    */
   
   if(mainpool_size)
   {
      struct sysctlblk *sysctlblk = (struct sysctlblk *)get_sysconf(81,1,0,0);
      static const SIGSELECT create_rep[] = {1, MM_CREATE_REGION_REPLY};
      PROCESS mm_pid;
      OSADDRESS sigsizes[9];
      OSADDRESS stacksizes[9];
      OSADDRESS mainpool_start;
      
      /* Create the osemain block in supervisor mode. */  
      block_ = create_block("osemain",0,0,0,1);    

      /* A (virtually) non-blocking hunt. */
      (void)hunt("ose_memory_mgr", 0, &mm_pid, NULL);
      if (mm_pid == 0)
      {
         ramlog_printf("Process ose_memory_mgr not found, MM not available.\n");
         return;
      }

      sig = alloc(sizeof(struct mm_create_region), MM_CREATE_REGION_REQUEST);
      sig->mmap.size = mainpool_size;
      sig->mmap.domain = 0;
      sig->mmap.permissions = main_user_allowed ? SU_RW_USR_RW : SU_RW_USR_RO;
      sig->mmap.cache_mode = mainpool_cache;
      sig->mmap.options = MM_CREATE_REGION_SPECIFY_NAME;
      sig->mmap.base = 0;
      strcpy(sig->mmap.phys_mem, MM_INTERNAL_RAM);
      strcpy(sig->mmap.log_mem, ose_mm_sysram_name());
      strcpy(sig->mmap.name, "mainpool");

      send(&sig, mm_pid);
      sig = receive(create_rep);

      if (sig->mmap.status != MM_SUCCESS)
      {
         /* FIXME: What to do here? */
      }

      mainpool_start = sig->mmap.base;
      free_buf(&sig);


      /* inherit sig/stack sizes from syspool, needs to do a ugly
       * copy.
       */
      sigsizes[0]=8;
      stacksizes[0]=8;
      memcpy(&sigsizes[1],
             &sysctlblk->sigsize_tab[0],
             sizeof(OSADDRESS) * 8);
      memcpy(&stacksizes[1],
             &sysctlblk->stacksize_tab[0],
             sizeof(OSADDRESS) * 8);
      
      /* Create the main pool. */
      zzkrn_create_program_pool(
                             block_,         /* Attach pool to this block */
                             mainpool_start, /* Start address of pool memory */
                             mainpool_size,  /* Size of pool memory */
                             sigsizes,       /* Def signal buffer sizes. */
                             stacksizes,     /* Def stack buffer sizes. */
                             mainpool_fill,  /* Fill with debug pattern? */
                             NULL, 
                             0, 
                             0, 
                             NULL);

      /*
       * If user mode access allowed, mark the pool as static to permit
       * s_alloc_nil from user programs with a private signal pool.
       */
      if(main_user_allowed)
	{
	  static const OSADDRESS taglist[] = {OSE_POOL_ATTR_IS_STATIC,
					      OSE_POOL_ATTR_TAGEND};

	  ose_set_pool_attr(get_sig_poolid(block_), taglist);
	}
      
      /* Create the zzmain process in osemain.c */
      osemain_pid_ = create_process(OS_PRI_PROC,
                                    "osemain",
                                    zzmain,
                                    (OSADDRESS) sysctlblk->zzmain_stacksize,
                                    (OSPRIORITY) sysctlblk->main_priority,
                                    (OSTIME)0,
                                    (PROCESS) block_,
                                    (struct OS_redir_entry *) 0,
                                    (OSVECTOR) 0,
                                    (OSUSER) 0);
      
      /* Make sure we get an error if the system dies. */
      attach(0, block_);

      /* Initialize the extender if mainpool frag size is specified. */
      if (mainfrag_size)
      {
         init_mainpool_extender(get_sig_poolid(osemain_pid_),
                                block_,
                                gmm_prio,
                                main_user_allowed,
                                mainfrag_size);
      }

      /* install pthread bios module, should be designed out! */
      ose_init_pthreads();
      set_ose_started();
      start(osemain_pid_);
      sig = receive(sigsel);
      free_buf(&sig);   
      kill_proc(current_process());
   }
   else
   {
      /* install pthread bios module, should be designed out! */
      ose_init_pthreads();
      /* continue into the zzmain entry in osemain.c. */
      set_ose_started();
      zzmain();
   }
   /* not reached */

   /* The following code is for link purposes only, it does not need to be
    * executed. It ensures that the thin library functionality is available for
    * weak linking against in the core module. */   
   if (NULL == thinlib_ref)
   {
      thinlib_ref = ose_thinlib_init;
   }   
}

