/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2005-2013 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

/**
 * @file pool_extend.c
 *
 * @brief This file contains the core pool extender functionality.
 *
 * This file contains the pool extender implementation for systems
 * that should support automatic runtime extension of the system pool and
 * a single shared user pool from the MM (OSE Memory Manager).
 *
 */

#include <stdio.h>
#include "ose.h"
#include "cpu_intlock.h"
#include "osetypes.h"
#include "ose_err.h"
#include "ramlog.h"
#include "mm.sig"
#include "mm.h"
#include "mm_util.h"
#include "sysparam.h"
#include "spinlock.h"
#include <string.h>
#include <stdlib.h>

#ifdef OSMIPS       
#include "cpu_access.h"
#endif

#define FRAG_STACK_SIZE 1000

union SIGNAL
{
    SIGSELECT                sig_no;
    struct mm_create_region  mmap;
};

OSENTRYPOINT ose_syspool_frag;
OSENTRYPOINT ose_mainpool_frag;

void 
init_syspool_extender(OSPRIORITY frag_prio,
                      OSBOOLEAN user_perm,
                      OSADDRESS sysfrag_size);

void 
init_mainpool_extender(PROCESS mainpool_pid,
                       PROCESS main_bid,
                       OSPRIORITY frag_prio,
                       OSBOOLEAN user_perm,
                       OSADDRESS mainfrag_size);

/* The system pool fragment allocator. */
static PROCESS ose_syspool_frag_;

/* The main pool fragment allocator. */
static PROCESS ose_mainpool_frag_;

/*
** The out of memory flags indicate that we can not obtain more memory
** from the memory manager.
*/

static int out_of_sysmemory = 0;
static int out_of_mainmemory = 0;

/* The main pool is the global pool defined in the kernel monolith. */
static PROCESS main_poolid;

/*
** Pointer to the pending system pool fragment list and a configurable
** definition of the fixed fragment size.
** The fragment size should be selected >> 64k to support the largest
** possible stack or signal size.
** On the other hand, it should be selected << the maximum expected
** pool size. 256 kbytes is a reasonable value for a system pool
** when it is not in heavy use as a signal or user stack pool.
*/

static void *sysfrag = (void *) 0;

/*
** Pointer to the pending main pool fragment list, and a configurable
** definition of the fixed fragment size.
** The fragment size should be selected as for the system pool.
** 256 kbytes is a reasonable default value also in this case.
*/

static void *mainfrag = (void *) 0;

#ifdef OSE_SMP
/*
** Spinlocks to protect pool extension from more than one cpu at a time.
*/

spinlock_t sysfrag_spin_lock;
spinlock_t mainfrag_spin_lock;
#endif

/*
** The pool extension sequence number.
** This value is provided by the kernel to the error handler.
** The first error has sequence 0.
**/
static OSERRCODE global_main_sequence = 0x7f;

static U32 sysfragsize;
static U32 mainfragsize;
static OSBOOLEAN sys_user_allowed;
static OSBOOLEAN main_user_allowed;
static int init_done;
static PROCESS mm_pid;
static const char *logmem;

/* callback function - may be installed by others, such that they know
** when the pool extender is busy.
**
** (basically an ugly workaround for traffic overload in the interrupt
**  context)
*/
static void (*callback)(int); /* mainpool hook */
static void (*syspoolext_callback)(int); /* syspool hook */ 


static OSADDRESS
the_extender(OSERRCODE ecode, OSERRCODE extra)
{
   Msr msr;
   static OSERRCODE global_sys_sequence = 0x7f; /* The first error
						   has sequence 0. */
   OSERRCODE current_sequence;

#ifdef DEBUG_EXTENDER
   ramlog_printf("extender: %x %x (%s in_q:%u)\n",
		 ecode, extra, get_pid_info(current_process(), 14),
		 get_pid_info(current_process(), 18));
#endif
   
   if( (ecode & (OSE_EERROR_MASK | OSE_EFATAL_MASK)) == OSE_ENO_KERN_SPACE)
   {
      /* Disable interrupts to avoid race conditions
	 towards interrupt processes. */
      spin_lock_irqsave(&sysfrag_spin_lock, msr);

      /*
      ** Extract the pool extend sequence counter from the extra parameter.
      ** Compare it to the extend_pool operation last executed.
      */
      current_sequence = (global_sys_sequence - (extra >> 24)) & 0x7f; 

#ifdef DEBUG_EXTENDER
      ramlog_printf("extender: KERN_SPACE, gseq: %u\n", current_sequence);
#endif

      /*
       * Check if the current error was already handled by a higher priority
       * process.
       */
      if(current_sequence < 63) 
      {
	 /* Exit succesfully if so. */
	 spin_unlock_irqrestore(&sysfrag_spin_lock, msr);
	 return (OSADDRESS) 1;
      }
         
      /*
       * Check that we have a new pool fragment.
       * If we don't have it, then we are either out of virtual memory, or a
       * new fragment could not be created in time because of very high load
       * We treat both these situations as a fatal error.
       */
      if(sysfrag)
      {
	 OSADDRESS f;
            
	 /* unlink the first fragment from the fragment list. */
	 f = (OSADDRESS) sysfrag;
	 sysfrag = *((void **)sysfrag);

	 /* Add the new fragment to the pool. This is an ose call. */
	 extend_pool(extra, (OSADDRESS) f, (OSADDRESS) sysfragsize);
            
	 /* Save the currently handled sequence number. */
	 global_sys_sequence = (extra >> 24) & 0x7f;
            
	 /*
	 ** Invoke the ose_syspool_frag helper process to create a
	 ** new fragment.
	 ** Note that we signal the semaphore with interrupts still
	 ** disabled, and with a locked dispatcher. This is allowed
	 ** because the signal_fsem() call preserves interrupt status.
	 ** Also, since the dispatcher is locked, no process switch will
	 ** take place until we return from the error handler.
	 */
	 if(!sysfrag)
	 {
	    /* if syspoolext_callback function is installed,
	     * indicate that we are now busy */
	    if (syspoolext_callback)
	    {
	       syspoolext_callback((int)sysfragsize);
	    }
	    signal_fsem(ose_syspool_frag_);
	 }
            
	 /* Restore interrupt status and exit succesfully. */
	 spin_unlock_irqrestore(&sysfrag_spin_lock, msr);
	 return (OSADDRESS) 1;
      }
      else
      {
	 /* Restore interrupt status. */
	 spin_unlock_irqrestore(&sysfrag_spin_lock, msr);

#ifdef DEBUG_EXTENDER
	 ramlog_printf("extender: no sysfrag, out_of_sysmemory: %s\n",
		       out_of_sysmemory ? "yes" : "no");
#endif	 
	 /* Report a recursive error if we are out of memory due
	    to overload. */
	 if(!out_of_sysmemory)
	 {
	    return 0;
	 }
      }
   }
   if(((ecode &
	(OSE_EERROR_MASK | OSE_EFATAL_MASK)) == OSE_ENO_USER_SIGSPACE) ||
      ((ecode &
	(OSE_EERROR_MASK | OSE_EFATAL_MASK)) == OSE_ENO_USER_STACKSPACE))
   {
      /* Disable interrupts to avoid race conditions towards interrupt
	 processes. */
      spin_lock_irqsave(&mainfrag_spin_lock, msr);

      /*
      ** Extract the pool extend sequence counter from the extra parameter.
      ** Compare it to the extend_pool operation last executed.
      */
      current_sequence = (global_main_sequence - (extra >> 24)) & 0x7f; 

#ifdef DEBUG_EXTENDER
      ramlog_printf("extender: USR_SPACE, gseq: %u\n", current_sequence);
#endif

      /*
      ** Check if the current error was already handled by a higher priority
      ** process.
      */
      if(current_sequence < 63)
      {
	 /* Exit succesfully if so. */
	 spin_unlock_irqrestore(&mainfrag_spin_lock, msr);
	 return (OSADDRESS) 1;
      }
         
      /*
      ** Check that we have a new pool fragment.
      ** If we don't have it, then we are either out of virtual memory, or a
      ** new fragment could not be created in time because of very high load.
      ** We treat both these situations as a fatal error.
      */
      if(mainfrag)
      {
	 OSADDRESS f;
            
	 /* unlink the first fragment from the fragment list. */
	 f = (OSADDRESS) mainfrag;
	 mainfrag = *((void **)mainfrag);
            
	 /* Add the new fragment to the pool. This is an ose call. */
	 extend_pool(extra, (OSADDRESS) f, (OSADDRESS) mainfragsize);
            
	 /* Save the currently handled sequence number. */
	 global_main_sequence = (extra >> 24) & 0x7f;
            
	 /*
	 ** Invoke the ose_mainpool_frag helper process to create a new fragment.
	 ** Note that we signal the semaphore with interrupts still
	 ** disabled, and with a locked dispatcher. This is allowed
	 ** because the signal_fsem() call preserves interrupt status.
	 ** Also, since the dispatcher is locked, no process switch will
	 ** take place until we return from the error handler.
	 */
	 if(!mainfrag)
	 {
	    /* if callback function is installed,
	     * indicate that we are now busy */
	    if (callback)
	    {
	       callback((int)mainfragsize);
	    }
	    signal_fsem(ose_mainpool_frag_);	       
	 }
            
	 /* Restore interrupt status and exit succesfully. */
	 spin_unlock_irqrestore(&mainfrag_spin_lock, msr);
	 return (OSADDRESS) 1;
      }
      else
      {
	 /* Restore interrupt status. */
	 spin_unlock_irqrestore(&mainfrag_spin_lock, msr);

#ifdef DEBUG_EXTENDER
	 ramlog_printf("extender: no mainfrag, out_of_mainmemory: %s\n",
		       out_of_mainmemory ? "yes" : "no");
#endif

	 /* Report a recursive error if we are out of memory due to overload. */
	 if(!out_of_mainmemory)
	 {
	    return 0;
	 }
      }
   }
   /*
   ** Allow the error handler to fail if no memory fragment is available,
   ** or if we don't recognize the error code.
   */   
   return (OSADDRESS) 0;   
}

/*
** Entrypoint for system pool extension callback.
*/
static OSADDRESS
syspool_extender(PROCESS pool, OSERRCODE extra)
{
  (void)pool;
  return the_extender(OSE_ENO_KERN_SPACE, extra);
}

/*
** Entrypoint for main pool extension callback.
*/

static OSADDRESS
mainpool_extender(PROCESS pool, OSERRCODE extra)
{
  (void)pool;
  return the_extender(OSE_ENO_USER_SIGSPACE, extra);
}

/*
** Intialization of the system pool extender 
**
*/

static void
init_pool_extender(void)
{
   if (init_done)
      return;
   init_done = True;
   logmem = ose_mm_sysram_name();

   /* A (virtually) non-blocking hunt. */
   (void)hunt("ose_memory_mgr", 0, &mm_pid, NULL);
   if (mm_pid == 0)
   {
      ramlog_printf("exec: init_pool_extender. Process ose_memory_mgr not found, MM not available.\n");
      return;
   }
}

static void
install_mm_event( void(*fp)(union mm_event_t *event) )
{
	mm_status_t status;
	struct mm_region_deleted_event_t mm_event;
	
	mm_event.event_no      = MM_EVENT_REGION_DELETED;
	mm_event.event_handler = fp;
	mm_event.identifier    = 0;
	mm_event.domain        = MM_EVENT_ANY_DOMAIN;

	status = ose_mm_install_handler((union mm_event_t *)&mm_event, True);
    if(status != MM_SUCCESS)
    { 
       /* Just notify that the event handler isn't available */
       ramlog_printf("exec: %s fails to install the MM event handler, status = 0x%08x\n",
	                  __func__, status);
   }	
}

/* These functions are called by MM when it has deleted a region. I.e it may be
 * enough memory available for a new fragment. Note that extension of mainpool
 * and syspool can be enabled or disabled independently of each other. */ 
static void
mm_event_mainpool(union mm_event_t *event)
{
	(void)event;

   /* Wake-up ose_mainpool_frag only if there isn't any fragment in spare */
   if (!mainfrag) 
   {
      /* if callback function is installed, indicate that we are now busy */
      if (callback)
      {
         callback((int)mainfragsize);
      }
      signal_fsem(ose_mainpool_frag_);
   }
}

static void
mm_event_syspool(union mm_event_t *event)
{
	(void)event;

   /* Wake-up ose_syspool_frag only if there isn't any fragment in spare */
   if (!sysfrag)
   {
      /* if syspoolext_callback function is installed, indicate that we are
       * now busy */           
      if (syspoolext_callback)
      {
         syspoolext_callback((int)sysfragsize);
      }
      signal_fsem(ose_syspool_frag_);
   }
}

void 
init_syspool_extender(OSPRIORITY frag_prio,
                      OSBOOLEAN user_perm,
                      OSADDRESS sysfrag_size)
{
   extern OSENTRYPOINT ose_syspool_frag;

   init_pool_extender();
   sysfragsize = sysfrag_size;
   
#ifdef OSE_SMP
   spin_lock_init(&sysfrag_spin_lock);
#endif

   ramlog_printf("exec: Starting syspool extender.\n");
   
   /* set the user access permission. */
   sys_user_allowed = user_perm;

   ose_syspool_frag_ = create_process(OS_PRI_PROC,
                                      "ose_syspool_frag",
                                      ose_syspool_frag,
                                      (OSADDRESS) FRAG_STACK_SIZE,
                                      (OSPRIORITY) frag_prio,
                                      (OSTIME)0,
                                      (PROCESS) 0,
                                      (struct OS_redir_entry *) 0,
                                      (OSVECTOR) 0,
                                      (OSUSER) 0);
   
   start(ose_syspool_frag_);
   attach(0, ose_syspool_frag_);
   
   /* Install the MM event. */          
   install_mm_event(mm_event_syspool);
}

/*
** Intialization of the main pool extender 
**
** The identity of the main pool identity
** is passed.
*/

void
init_mainpool_extender(PROCESS mainpool_pid,
                       PROCESS main_bid,
                       OSPRIORITY frag_prio,
                       OSBOOLEAN user_perm,
                       OSADDRESS mainfrag_size)
{
   extern OSENTRYPOINT ose_mainpool_frag;
   
    init_pool_extender();

  /* Set the main pool id. */
   main_poolid = mainpool_pid;
   
   mainfragsize = mainfrag_size;
   
#ifdef OSE_SMP
   spin_lock_init(&mainfrag_spin_lock);
#endif

   /* set the user access permission. */
   main_user_allowed = user_perm;
   
   ramlog_printf("exec: Starting mainpool extender.\n");
   
   /* create fragment process for mainpool, but do not start it. */
   ose_mainpool_frag_ = create_process(OS_PRI_PROC,
                                       "ose_mainpool_frag",
                                       ose_mainpool_frag,
                                       (OSADDRESS) FRAG_STACK_SIZE,
                                       (OSPRIORITY) frag_prio,
                                       (OSTIME)0,
                                       (PROCESS) main_bid,
                                       (struct OS_redir_entry *) 0,
                                       (OSVECTOR) 0,
                                       (OSUSER) 0);
   
   start(ose_mainpool_frag_);
   attach(0, ose_mainpool_frag_);

   /* Install the MM event. */          
   install_mm_event(mm_event_mainpool);
}

/*
 * Function to install one callback that indicates the state of the mainpool
 * extender. Used only by cello?
 */
int 
ose_install_poolext_callback(void (*func)(int));

int 
ose_install_poolext_callback(void (*func)(int))
{
   if (!callback)
      callback = func;
   else
      return -1;

   return 0;
}

/*
 * Function to install one callback that indicates the state of the syspool
 * extender. Used only by cello?
 */
int 
ose_install_syspoolext_callback(void (*func)(int));

int 
ose_install_syspoolext_callback(void (*func)(int))
{
   if (!syspoolext_callback)
      syspoolext_callback = func;
   else
      return -1;

   return 0;
}

/*
** The fragment processes may not do alloc or free_buf dynamically,
** so ose_mmap() can not be used. Pre-allocate one signal and keep it.
*/
static OSADDRESS 
frag_mmap(union SIGNAL **sig, mm_status_t *status)
{
  OSADDRESS addr;
  static const SIGSELECT reply[] = {1, MM_CREATE_REGION_REPLY};
   
  if (logmem == NULL || !mm_pid)
    {
      return 0; /* MM not available. */
    }

  (*sig)->sig_no = MM_CREATE_REGION_REQUEST;
  send(sig, mm_pid);
  do
    {
      *sig = receive_from(~0, reply, mm_pid);
    }
  while(NIL == *sig);
  addr = (*sig)->mmap.base;
  *status = (*sig)->mmap.status;
  if ((*sig)->mmap.status != MM_SUCCESS)
    {
      addr = 0;
    }
  return addr;
}

/*
** These process as well as MM should be on a priority
** higher than the ose_sysd process or any typical application process.
** This increases the chance of pool fragments being created at a rate high
** enough to avoid a second out of pool error in a situation where no
** new fragment is yet available.
**
** If this rule can not be obeyed, then it is possible that the system
** will at rare occasions encounter a fatal memory shortage error.
** This error occurs when signals are allocated or processes created at
** a rate that causes an entire pool fragment to be exhausted before
** there is time to create another one.
**
*/

OS_PROCESS(ose_syspool_frag)
{
   char ctype[MM_NAME_LENGTH+1];
   char mtype[MM_NAME_LENGTH+1];
   mm_status_t status = 0;
   U32 num_created_fragments = 0;
   U32 cache_mode;
   Msr msr;
   Boolean pattern_fill;
   union SIGNAL *sig;
   
   /* Fast semaphores should be initialized before they can be used. */
   set_fsem(0, current_process());

   /* Initialize syspool fragments with a debug pattern? Default: no. */
   if(sysparam_readattr_bool("krn/syspool", "pattern_fill", &pattern_fill))
   {
      /* Also check legacy/sfk version of the parameter */
      if(sysparam_readattr_bool("krn/system_pool", "pattern_fill", 
                                &pattern_fill))
      {
         pattern_fill = False;
      }
   }

   strlcpy(ctype, "copy_back", MM_NAME_LENGTH);
   strlcpy(mtype, "speculative_access", MM_NAME_LENGTH);
   sysparam_readattr_string("krn/syspool", "cache", ctype, MM_NAME_LENGTH);
   sysparam_readattr_string("krn/syspool", "mem",   mtype, MM_NAME_LENGTH);
   if(ose_mm_get_cache_mode(ctype, mtype, &cache_mode))
   {
      cache_mode = MM_CACHE_COPY_BACK | MM_SPECULATIVE_ACCESS;
   }

   sig = alloc(sizeof(struct mm_create_region), MM_CREATE_REGION_REQUEST);
   sig->mmap.size = sysfragsize;
   sig->mmap.domain = 0;
   strcpy(sig->mmap.phys_mem, MM_INTERNAL_RAM);
   strcpy(sig->mmap.log_mem, logmem);
   sig->mmap.permissions = sys_user_allowed ? SU_RW_USR_RW : SU_RW_USR_RO;
   sig->mmap.cache_mode = cache_mode;
   sig->mmap.options = MM_CREATE_REGION_SPECIFY_NAME;
   sig->mmap.base = 0;
   sig->mmap.phys_count = 0;

   for(;;)
   {
      /*
      ** Create a system pool memory region with size = sysfragsize, and
      ** store a pointer to it in sysfrag. Leave sysfrag set to 0
      ** if allocation fails.
      ** Skip allocation if the old fragment is not yet taken.
      */

      if(!sysfrag)
      {
         OSADDRESS addr;
         
	 sprintf(sig->mmap.name, "syspool_chunk%u", num_created_fragments++);

         /* Allocate physical, SASE, cached memory acc user mode.*/
         addr = frag_mmap(&sig, &status);
	 
         if (addr)
         {
#ifdef OSMIPS
            /* Ensure that the sys_pool_extension is in KSEG0, as this is
            ** required by the targent interface for MIPS. E.g. the interrupt
            ** stack must be in KSEG0.
            ** As sys_pool_extension is in KSEG0, user mode can not access it.
            ** The ose_mmap() above can set USR access to RW or RO but this is
            ** thus not valid for MIPS, there will be no USR rights.
            */
            addr += KSEG0_BASE;
#endif
            /* Yes we got sysmemory */
            out_of_sysmemory = 0;

            if(pattern_fill)
            {
               /* Fill extend fragment with debug pattern */
               memset((char *) addr, 0xcc, (size_t) sysfragsize);
            }

            /* Disable interrupts to avoid race conditions towards
             * add_frag_list(). */
            spin_lock_irqsave(&sysfrag_spin_lock, msr);

            *((void **)addr) = (void *) sysfrag; /* Clear forward pointer. */
            sysfrag = (void*) addr;

            spin_unlock_irqrestore(&sysfrag_spin_lock, msr);
         }
         else
         {
            /* Decrease the counter since the frag_mmap failed */
            num_created_fragments--;
            /* Also set the out of memory indicator. */
            out_of_sysmemory = 1;
            ramlog_printf("exec: Pool extender:out of system pool memory, "
            "MM err: 0x%x\n",status);
	 }
      }

      /* if callback function is installed, indicate that we are now idle */
      if (syspoolext_callback)
      {
         syspoolext_callback(0);
      }

      if(num_created_fragments == 1)
      {
         /*
          * install the kernel extender hook to be called before system error
          * handler.
          */
          ose_set_pool_extender(get_sig_poolid(current_process()),
                                syspool_extender,
                                &sysfrag,
                                get_pri(current_process()));
      }

      /* Wait for any of the fragments to be taken by the error handler. */
      wait_fsem(1);
   }
}

OS_PROCESS(ose_mainpool_frag)
{
   char ctype[MM_NAME_LENGTH+1];
   char mtype[MM_NAME_LENGTH+1];
   mm_status_t status = 0;
   U32 num_created_fragments = 0;
   U32 cache_mode;
   Msr msr;
   Boolean pattern_fill;
   union SIGNAL *sig;

   /* Fast semaphores should be initialized before they can be used. */
   set_fsem(0, current_process());

   /* Initialize mainpool fragments with a debug pattern? Default: no. */
   if(sysparam_readattr_bool("krn/mainpool", "pattern_fill", &pattern_fill))
      pattern_fill = False;

   strlcpy(ctype, "copy_back", MM_NAME_LENGTH);
   strlcpy(mtype, "speculative_access", MM_NAME_LENGTH);
   sysparam_readattr_string("krn/mainpool", "cache", ctype, MM_NAME_LENGTH);
   sysparam_readattr_string("krn/mainpool", "mem",   mtype, MM_NAME_LENGTH);
   if(ose_mm_get_cache_mode(ctype, mtype, &cache_mode))
   {
      cache_mode = MM_CACHE_COPY_BACK | MM_SPECULATIVE_ACCESS;
   }
   
   sig = alloc(sizeof(struct mm_create_region), MM_CREATE_REGION_REQUEST);
   sig->mmap.size = mainfragsize;
   sig->mmap.domain = 0;
   strcpy(sig->mmap.phys_mem, MM_INTERNAL_RAM);
   strcpy(sig->mmap.log_mem, logmem);
   sig->mmap.permissions = main_user_allowed ? SU_RW_USR_RW : SU_RW_USR_RO;
   sig->mmap.cache_mode = cache_mode;
   sig->mmap.options = MM_CREATE_REGION_SPECIFY_NAME;
   sig->mmap.base = 0;
   sig->mmap.phys_count = 0;

   for(;;)
   {
      /*
      ** Create a main pool memory region with size = mainfragsize, and
      ** store a pointer to it in mainfrag. Leave mainfrag set to 0
      ** if allocation fails.
      ** Skip allocation if the old fragment is not yet taken or if
      ** we are out of memory.
      */
      
      if(!mainfrag)
      {	     
	 OSADDRESS addr;
         
	 sprintf(sig->mmap.name, "mainpool_chunk%u", num_created_fragments++);
	 
         /* Allocate physical, SASE, cached memory acc user mode.*/
         addr = frag_mmap(&sig, &status);
   
	 if (addr)
	 {
            /* Yes we got main memory */
            out_of_mainmemory = 0;
            if(pattern_fill)
            {
               /* Fill extend fragment with debug pattern */
               memset((char *) addr, 0xcc, (size_t) mainfragsize);
            }

	    /* Disable interrupts to avoid race conditions towards
               add_frag_list(). */	
	    spin_lock_irqsave(&mainfrag_spin_lock, msr);

	    *((void **)addr) = (void *) mainfrag; /* Clear forward pointer. */
	    mainfrag = (void*) addr;

	    spin_unlock_irqrestore(&mainfrag_spin_lock, msr);
	 }
	 else
	 {
	    /* Decrease the counter since the frag_mmap failed */
	    num_created_fragments--;
	    /* Also set the out of memory indicator. */
	    out_of_mainmemory = 1;
	    ramlog_printf("exec: Pool extender: out of main pool memory, "
			  "MM err: 0x%x\n",status);
	 }
      }

      /* if callback function is installed, indicate that we are now idle */
      if (callback)
      {
	 callback(0);
      }

      if(num_created_fragments == 1)
      {
	 /*
	  * install the kernel extender hook to be called before system error
	  * handler.
	  */
	ose_set_pool_extender(main_poolid,
			      mainpool_extender,
			      &mainfrag,
			      get_pri(current_process()));
      }

      /* Wait for any of the fragments to be taken by the error handler. */
      wait_fsem(1);
   }
}
