/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2002-2007 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

/**
 * @private
 * @file board_parameter_conf
 * @toc api/core/init
 * @version 
 * @copyright Copyright (C) 2002-2007 Enea Software . All rights reserved.
 * @brief Implementation for board parameter storage.
 *
 * @long
 *	This file implements a backend to the boardparam.h interface,
 *	to read and write/modify the board parameters, that are
 *	contributors of the system parameters.
 *
 *	The implementation allows for one set of Board Parameters,
 *	stored in either of two flash blocks. A BP set consists of
 *	one or more sequential BP entries, which modifies the BP set
 *	in increments by adding, modifying or deleting board parameters.
 *	Each entry has a version number which is the version of the prior
 *	entry incremented by one, and the version of the whole
 *	BP set is the version of its last entry. When a flash block is about
 *	to become full, the result of its entries are compiled into a single
 *	entry in the other flash block, and that block is considered now to
 *	be active. At startup two BP sets might be found, one in each block,
 *	and naturally the one with the highest version number is chosen as
 *	the active one.
 *   
 *	This implementation always maintains a cache, i.e. a set of pointers
 *	to the active board parameters in the active BP set. For example if
 *	one board parameter is added by one entry, its address will be
 *	cached. But if that same board parameter is deleted by a later entry
 *	in the same BP set, the corresponding cached address will be removed.
 *
 *	A dynamic prioritized process, "ose_boardparam", is used to process
 *	and serialize the read/write operations. ose_boardparam is created and
 *	started at the first call to this module which happens to take place
 *	after the Flash Access Manager has come up, indicating writes to flash
 *	are  possible.
 */

/*
 **********************************************************
 * INCLUDE FILES.
 **********************************************************
 */

#include <ctype.h>

#include "stddef.h"
#include "string.h"
#include "stdio.h"
#include "flash_api.h"
#include "flash_conf.h"
#include "assert.h"
#include "boardparam.h"
#include "sysparam_err.h"
#include "ose.h"
#include "cpu.h"
#include "ramlog.h"
#include "sysparam.h"

/*
 **********************************************************
 *   MANIFEST CONSTANTS.
 **********************************************************
 */

/**
 * @macro BP_PARAM_NAME
 * @brief Name of the system parameter that defines the region for board parameters
 * @long  This configuration defines the base address and size of the region
 *        reserved for board parameters.
 *        Syntax of value: base:<base address> size:<size in bytes>
 * @seealso BP_MIN_SIZE
 * @seealso BP_MAX_SIZE
 */
#define BP_PARAM_NAME "krn/boardparam"
#define BP_PARAM_NAME_OLD "krn/region/boardparam"

/**
 * @macro BP_MIN_SIZE
 * @brief Minimum size for the board parameter region.
 * @seealso BP_PARAM_NAME
 */
#define BP_MIN_SIZE 1000 /* how small can it be? */

/**
 * @macro BP_MAX_SIZE
 * @brief Maximum size for the board parameter region.
 * @seealso BP_PARAM_NAME
 */
#define BP_MAX_SIZE 0x00800000


/* --- MAX_NR_BP controls the number of statically allocated pointers to
 * BP names in the read-cache. */
#define MAX_NR_BP		100
#define BP_ENTRY_MAGIC		0xA502

/* The time in milliseconds to wait while the ose_boardparam is being created*/
#define HUNT_TMO 1000

/* value to return when read intialization of BP is done. */
#define INIT_DONE        0

/* value to return when read intialization of BP is not yet done. */
#define INIT_NOT_DONE    -1

/*
 ***********************************************************
 *   MACROS.
 **********************************************************
 */
#define ALIGN(adr, with)  ((((U32)(adr)) + (with) - 1) & ~((with) - 1))


/* For debugging your system, define ASSERT_IN_PARAMETERS */
/* #define ASSERT_IN_PARAMETERS */
#ifdef ASSERT_IN_PARAMETERS
#define inparam_assert assert
#else
#define inparam_assert(x) ((void)0)
#endif

#ifndef STRSTR
#define STR(x) #x
#define STRSTR(x) STR(x)
#endif

#define WARNING(fmt)      \
(void)ramlog_printf(__FILE__" "STRSTR(__LINE__)": "fmt"\n")
#define WARNING2(fmt,val) \
(void)ramlog_printf(__FILE__" "STRSTR(__LINE__)": "fmt"\n", val)
#define WARNING3(fmt,v1,v2) \
(void)ramlog_printf(__FILE__" "STRSTR(__LINE__)": "fmt"\n", v1, v2)
#define WARNING4(fmt,v1,v2,v3) \
(void)ramlog_printf(__FILE__" "STRSTR(__LINE__)": "fmt"\n", v1, v2, v3)

#define ERROR(msg) do { /*lint -save -e717 */\
   (void)ramlog_printf(__FILE__" "STRSTR(__LINE__)" ERROR: %s\n", msg);\
   error2(OSE_SYSPARAM_EILLEGAL_STATE, (OSERRCODE)msg);\
} while (0) /*lint -restore */

/*
 **********************************************************
 *   TYPES.
 **********************************************************
 */

/*
 ****************************************************************************
 * SIGNALS
 ****************************************************************************
 */

#define BP_SIGBASE  0xe30700

/*===========================================================================
 * BOARDPARAM_READ_REQUEST
 *===========================================================================
 *
 * Description:
 * This signal requests a read of the board parameter value associated with
 * the supplied parameter name.
 *
 * Request parameters:
 *       name: The name of the wanted parameter.
 *
 * Reply parameters:
 *	value: The ASCII-string value of the requested parameter. NULL-string
 *		if no matching parameter was found.
 */

#define BOARDPARAM_READ_REQUEST (BP_SIGBASE + 0)
/* !-SIGNO(struct BoardparamReadRequest)-! */
#define BOARDPARAM_READ_REPLY (BP_SIGBASE + 1)
/* !-SIGNO(struct BoardparamReadReply)-! */

struct BoardparamReadRequest
{
      SIGSELECT sig_no;
      char 	name[1];
};

struct BoardparamReadReply
{
      SIGSELECT sig_no;
      char value[1];
};


/*===========================================================================
 * BOARDPARAM_WRITE_REQUEST
 *===========================================================================
 *
 * Description:
 *	This signal is used to add, modify or delete a board parameter.
 *
 * Request parameters:
 *	name_value: two consequent null-terminated ASCII strings containing
 *		first the name of the parameter to add, modify or delete, and
 *		second the value to give it (null if the parameter is to be
 *		deleted).
 */

#define BOARDPARAM_WRITE_REQUEST (BP_SIGBASE + 2)
/* !-SIGNO(struct BoardparamWriteRequest)-! */

#define BOARDPARAM_WRITE_REPLY   (BP_SIGBASE + 3)
/* !-SIGNO(struct BoardparamWriteReply)-! */

struct BoardparamWriteRequest
{
      SIGSELECT sig_no;
      char name_value_pair[2]; /* Hold two bytes for the two null chararcters*/
};

struct BoardparamWriteReply
{
      SIGSELECT sig_no;
      S32 result;
};

/*===========================================================================
 * BOARDPARAM_LIST_NEXT
 *===========================================================================
 *
 * Description:
 *	This signal is used to iteratively traverse the board parameter name
 *	space to list the available parameter names.
 *
 * Request parameters:
 *	str: the name to start from.
 *
 */

#define BOARDPARAM_LIST_NEXT_REQUEST (BP_SIGBASE + 4) /* !-SIGNO(struct BoardparamListNextRequest)-! */

struct BoardparamListNextRequest
{
      SIGSELECT sig_no;
      char str[1];
};

#define BOARDPARAM_LIST_NEXT_REPLY (BP_SIGBASE + 5) /* !-SIGNO(struct BoardparamListNextReply)-! */

struct BoardparamListNextReply
{
      SIGSELECT sig_no;
      char name[1];
};

/* Signal used to hunt for the ose_boardparam process */
#define BOARDPARAM_HUNT		   (BP_SIGBASE + 6) /* !-SIGNO(SIGSELECT)-! */


union SIGNAL
{
      SIGSELECT sig_no;
      struct BoardparamReadRequest read_req;
      struct BoardparamReadReply read_reply;
      struct BoardparamWriteRequest write_req;
      struct BoardparamWriteReply write_reply;
      struct BoardparamListNextRequest list_req;
      struct BoardparamListNextReply list_reply;
};

/* The following defines the structure of the header which preceeds each
 * board parameter entry in flash memory.
 */
struct BpEntryHeader
{
      /*Marks the beginning of an entry. */
      unsigned short magic;
      /* 16 bit ones complement checksum, applied to both header and data */
      unsigned short checksum;
      /* The size, in bytes, of the of the whole entry (the sum of the
       * entry header and the data that follows) */
      unsigned short size;
      /* The version of the last entry in a set of BPs uniquely identifies it.
       */
      unsigned short version;  
};

/* The cached version of BP entries contains a RAM image of
 * the BP table and a sorted array of pointers to the list
 * of names identifying each BP entry in the table.
 */

struct BpState
{
      const char *active_block;	  /* Flash sector 1 */
      const char *inactive_block; /* Flash sector 2 */
      size_t block_size;	  /* Flash sector size (same for both) */
      PROCESS ose_bp_pid;
      U16 version;
      char *write_ptr;
      int nr_names;
      const char *name_array[MAX_NR_BP];
};

/*
 **********************************************************
 *   VARIABLES
 **********************************************************
 */


/* The bp_state reserves all of the RAM needed to manage a BP
 * table of the maximum size.
 * FIXME: bp_state is initialized and is put into .data section.
 */
static struct BpState bp_state =
{
   NULL,/* active */
   NULL,/* inactive */
   0, /* blsize */
   0, /* ose_bp_pid */
   0, /* version */
   NULL, /* write_ptr, "NULL" means not initalized for writes. */
   -1,   /* nr_names, "-1" means not initialised for reads. */
   {NULL}, /* name_array */
};

static Boolean fam_is_up; /* Initialized to False by crt0. */

/*
 **********************************************************
 *   FUNCTIONS.
 **********************************************************
 */

/*
 *=========================================================
 *        remove_start_and_end_spaces_from_string
 *=========================================================
 * Description: Removes all spaces in front and from the 
 * end of a string.
 *
 * Argument: Pointer to a char pointer. The char pointer will
 *           be updated with the new start of the string.
 *           Termination will be added after the last non-
 *           space character in the string.
 *
 * Returns: Nothing
 *   
 */
static void
remove_start_and_end_spaces_from_string(char **str_ptr)
{
   int idx;
   size_t stringlen;
   char *text;

   /* Remove spaces in front of string */
   while (isspace((unsigned char)**str_ptr))
   {
      (*str_ptr)++;
   }

   /* Remove spaces in the end of the string */
   idx = 1;
   stringlen = strlen(*str_ptr);
   text = ((*str_ptr) + stringlen - 1);
   while (isspace((unsigned char)*text))
   {
      idx++;
      text = ((*str_ptr) + stringlen - idx);
   }

   /* Add null termination after (+1) the last non-space */
   *(text + 1) = '\0';
}

static void
read_from_flash(void *dest, const char *src, size_t len)
{
   if (fam_is_up)
   {
      enum FlashResult read_result = flash_read(dest, (FlashAddress)src, len);
      if (read_result != FLASH_EOK)
      {
	 ERROR("flash_read() failed.");
      }
   }
   else
   {
      memcpy(dest, src, len);
   }
}

static int
write_from_flash_to_flash(char *dest, const char *src, size_t len)
{
   int retval;
   unsigned int offs = 0;
   char buffer[64];
   size_t size_left = len;
   if (size_left > 0)
   {
      retval = 0;
      do
      {
	 size_t cpy_size = ((size_left >= sizeof(buffer))
			    ? sizeof(buffer)
			    : size_left);
	 enum FlashResult result;
	 read_from_flash(buffer, src + offs, cpy_size);
	 result = flash_write(dest + offs, buffer, cpy_size);
	 if (result == FLASH_EOK)
	 {
	    offs += cpy_size;
	    size_left -= cpy_size;
	    /* then loop */
	 }
	 else
	 {
	    WARNING2("flash_write failed with result %d", result);
	    retval = -1;
	    break;
	 }
      } while (size_left > 0);
   }
   else
   {
      assert(0);
      WARNING("write_from_flash_to_flash() called with destination"
              " buffer size == 0");
      retval = -1;
   }
   return retval;
}



/*
 *=========================================================
 *                      calc_checksum
 *=========================================================
 * Description: The integrity of a set of board parameters
 * is verified by maintaining redundancy information in the
 * form of a 16 bit ones complement checksum.
 *
 * Returns:  The 16 bit ones complement checksum.
 *   
 */
static U32
calc_checksum(U32 start_sum, const void *whence, size_t lenght)
{
   const U16 *addr = whence;
   U32 sum   = start_sum;
   
   assert(whence != NULL);
   assert(lenght > 0);
   
   /* Sum the 16-bit words */
   for ( ; lenght > 1; lenght -= 2)
   {
      sum += *addr++;
      
      /* Add in the carry amount */
      if (sum & 0xffff0000)
      {
	 sum += ((sum >> 16) & 0xffff);
	 sum &= 0xffff;
      }
   }
   
   /* If there is a leftover byte */
   if (lenght == 1)
   {
      sum += (((*(const char *)addr) << 8) & 0xff00);
      
      /* Add in the carry amount */
      if (sum & 0xffff0000)
      {
	 sum += ((sum >> 16) & 0xffff);
	 sum &= 0xffff;
      }
   }
   return sum;
}

static U32
calc_checksum_from_flash(U32 start_sum, const char *whence, size_t len)
{
   U32 sum;
   if (fam_is_up)
   {
      unsigned int offs = 0;
      char buffer[64];
      size_t size_left = len;

      sum = start_sum;
      while (size_left > 0)
      {
	 size_t batch_size = ((size_left >= sizeof(buffer))
			      ? sizeof(buffer)
			      : size_left);
	 assert(size_left <= len);
	 assert((offs + size_left) == len);
	 read_from_flash(buffer, whence + offs, batch_size);
	 sum = calc_checksum(sum, buffer, batch_size);
	 offs += batch_size;
	 size_left -= batch_size;
	 /* loop */
      }
   }
   else
   {
      sum = calc_checksum(start_sum, whence, len);
   }
   return sum;   
}

/*
 *=========================================================
 *                      read_entry_header
 *=========================================================
 * Description: Verifies the integrity of a BP entry using
 * its checksum information.
 *
 * Returns:  TRUE is the checksum is valid, FALSE otherwise
 */
static int
read_entry_header(const char *entry_addr, unsigned short *version,
		  unsigned short *size)
{
   struct BpEntryHeader entry;
   assert(entry_addr != NULL);
   read_from_flash(&entry, entry_addr, sizeof(struct BpEntryHeader));
   if (entry.magic == BP_ENTRY_MAGIC)
   {
      /* FIXME: verify that 'entry_addr' and/plus 'entry,size' are within
       * flash boundaries before calling the checksum routine.
      */
      U32 cksum = calc_checksum_from_flash(0, entry_addr, entry.size);
      if ((0xffff & ~cksum) == 0)
      {
	 if (version != NULL)
	 {
	    *version = entry.version;
	 }
	 if (size != NULL)
	 {
	    *size = entry.size;
	 }
	 return 1;
      }
   }
   return 0;
}

static int
is_erased_flash_area(const char *whence, size_t size)
{
#define CMP_BUFSIZE 0x1000
   int retval = 1;
   const char *limit = whence + size;
   size_t size_left = size;
   size_t bufsize = size_left < CMP_BUFSIZE ? size_left : CMP_BUFSIZE;
   char *buffer = (char *)alloc(bufsize, 0);
   
   assert(whence != NULL);
   assert(whence < limit);
   
   do
   {
      enum FlashResult read_result = flash_read(buffer,
                                                (char *)limit - size_left,
						bufsize);
      if (read_result == FLASH_EOK)
      {
	 U32 offs = 0;
	 
	 while (offs < bufsize)
	 {
	    static const U32 all_Fs[] = {0xffffffff, 0xffffffff, 0xffffffff};
	    size_t cmp_sz = ((bufsize - offs) < sizeof all_Fs
			     ? (bufsize - offs) : sizeof all_Fs);
	    if (memcmp(buffer + offs, all_Fs, cmp_sz)  == 0)
	    {
	       offs += cmp_sz;
	       /* loop */	
	    }
	    else
	    {
	       retval = 0;
	       goto comp_done;
	    }
	 }
      }
      else
      {
	 ERROR("flash_read() failed.");
	 retval = 0;
	 break;
      }
      size_left -= bufsize;
      bufsize = size_left < CMP_BUFSIZE ? size_left : CMP_BUFSIZE;
   } while (size_left > 0);

  comp_done:   
   free_buf((union SIGNAL **)&buffer);

   /* flash area OK */
   return retval;
}


static size_t
strlen_flash(const char *str)
{
   size_t len;
   if (fam_is_up)
   {
      len = 0;
      for (;;)
      {
	 unsigned i;
	 char buffer[32];
	 read_from_flash(buffer, str + len, sizeof(buffer));
	 for (i = 0; i < sizeof(buffer); i++)
	 {
	    if (buffer[i] == '\0')
	    {
	       return len + i;
	    }
	 }
	 len += sizeof(buffer);
      }
   }
   else
   {
      len = strlen(str);
   }
   return len;
}


static int
strcmp_ram_flash(const char *s1, const char *s2)
{
   int strcmp_result;
   if (fam_is_up)
   {
      char buffer[32];
      unsigned offs = 0;
      unsigned length_left = strlen(s1) + 1;
      
      for (;;)
      {
	 read_from_flash(buffer, s2 + offs, sizeof(buffer));
             
	 strcmp_result = strncmp(s1 + offs, buffer, sizeof(buffer));
	 if ((strcmp_result != 0) || (length_left <= sizeof(buffer)))
	 {
	    /* no use in comparing any more, so break */
	    break;
	 }
	 else
	 {
	    length_left	-= sizeof(buffer);
	    offs		+= sizeof(buffer);
	 }
      }
   }
   else
   {
      strcmp_result = strcmp(s1, s2);
   }
   
   return strcmp_result;
}


static int
strcmp_flash_flash(const char *s1, const char *s2)
{
   int strcmp_result;
   if (fam_is_up)
   {
      char s1_buf[32];
      char s2_buf[32];
      U32 offs = 0;
      
      for (;;)
      {
	 read_from_flash(s1_buf, s1 + offs, sizeof(s1_buf));
	 read_from_flash(s2_buf, s2 + offs, sizeof(s2_buf));
	 strcmp_result = strncmp(s1_buf, s2_buf, sizeof(s1_buf));

	 if ((strcmp_result != 0) || (strnlen(s1_buf, sizeof(s1_buf)) < sizeof(s1_buf)))
	 {
	    break;
	 }
	 else
	 {
	    offs += sizeof(s1_buf);
	 }
      }
   }
   else
   {
      strcmp_result = strcmp(s1, s2);
   }
   return strcmp_result;
}

/*
 *=========================================================
 *                      find_bp_name
 *=========================================================
 * Description: 
 *
 * The following routine searches the BP name list, looking
 * for the specified name/value pair. If successful, a
 * pointer to that name/value pair is returned; otherwise
 * NULL is returned.
 *
 * If an index pointer is supplied the routine will return
 * an index into the search array. For successful
 * operations the index returned corresponds to the index
 * of the matching entry. For unsuccessful operations the
 * index returned is the index of the entry the would
 * lexicographically follow the search name if it were
 * present.   
 */

static const char *
find_bp_name(const char *name, const char **name_array, int nr_names,
	     int *index, Boolean name_is_in_flash)
{
   static int latest_index; /* Initialized to 0 by crt0. */
   int min = 0;
   int max = nr_names - 1;
   int i = latest_index <= max ? latest_index : (max / 2);

   assert(name != NULL);
   assert(name_array != NULL);

   while(max >= min)
   {
      int compare;
      if (name_is_in_flash)
      {
	 compare = strcmp_flash_flash(name, name_array[i]);
      }
      else
      {
	 compare = strcmp_ram_flash(name, name_array[i]);
      }
      if (compare == 0)
      {
	 /* Name found */
	 if (index)
	 {
	    *index = i;
	 }
	 latest_index = i;
	 return name_array[i];
      }
      else if (compare > 0)
      {
	 min = i + 1;
      }
      else
      {
	 max = i - 1;
      }
      i = (unsigned)(min + max) >> 1;
   }
   
   /* If the name was not found here, return to the index
    * and the name that will follow alphabetically.
    */
   if (index)
   {
      *index = min;
   }
   if (min < nr_names)
   {
      latest_index = min;
   }
   return NULL;
}


static unsigned short
get_version(const char *flash_base, char **write_ptr)
{
   unsigned short version;
   unsigned short entry_size;
   const char *entry_addr = flash_base;

   assert(flash_base != NULL);

   if (read_entry_header(entry_addr, &version, &entry_size))
   {
      unsigned short prev_version = version;

      entry_addr += entry_size;
      while (read_entry_header(entry_addr, &version, &entry_size))
      {
	 if (version == (U16)(prev_version + 1))
	 {	     
	    prev_version = version;
	    /* Update "entry" to point to the next potential entry position. */
	    entry_addr += entry_size;
	 }
	 else
	 {
 	    WARNING("Stale data in BP area:(version != prev_version + 1)");
	    /* Found non-contiguous versions. Return the last valid version. */
	    version = prev_version;
	    break;
	 }
      }
   }
   else
   {
      version = 0;
   }
   *write_ptr = (char *)entry_addr;
   return version;
}

static unsigned short
get_first_version(const char *flash_base)
{
   unsigned short version;
   unsigned short entry_size;
   const char *entry_addr = flash_base;

   assert(flash_base != NULL);

   if (!read_entry_header(entry_addr, &version, &entry_size))
   {
      version = 0;
   }

   return version;
}

/*
 *=========================================================
 *                      update_name_array
 *=========================================================
 * Description: 
 *	The following routine is used to update the
 *	BP Cache with a new board parameter entry.
 *	The parameters specified in the new entry modify the active set of
 *	board parameters by adding new parameters to it or by modifying or
 *	annulating existing parameters. All of the entries in a given "set" of
 *	board parameters are assumed to be unique. 
 *
 * Returns:
 *	The number of names after the update, or -1 on failure.
 */

static int
update_name_array(const char **name_array, int nr_names_in,
		  const char *entry_addr)
{
   unsigned short entry_size;
   int nr_names;
   
   if (read_entry_header(entry_addr, NULL, &entry_size))
   {
      const char *curr_name = entry_addr + sizeof(struct BpEntryHeader);
      const char *limit = entry_addr + entry_size;
      nr_names = nr_names_in;
      while (curr_name < limit)
      {
	 size_t namelen = strlen_flash(curr_name);
	 /* valuelen is the length of the string following the name string */
	 size_t valuelen = strlen_flash(curr_name + namelen + 1);
	 int index;
	 assert(namelen > 0);
	 
	 if (find_bp_name(curr_name, name_array, nr_names, &index, True))
	 {
	    /* A name equal to curr_name already exists in the list */
	    if (valuelen == 0)
	    {
	       /* The board parameter is being deleted, so remove it. */
	       int j;
	       for(j = index + 1; j < nr_names; j++)
	       {
		  name_array[j-1] = name_array[j];
	       }
	       nr_names--;
	    }
	    else
	    {
	       /* The board parameter is being modified, update the cache. */
	       name_array[index] = curr_name;
	    }
	 }
	 else
	 {
	    /* Else curr_name not found in list, 'index' corresponds
	     * now to the name lexicographically following curr_name.
	     */
	    if (valuelen == 0)
	    {
	       /*  a new BP had no value string, corrupt table? */
	       ERROR("a new BP had no value string, corrupt table?");
	       nr_names = -1;
	       break;
	    }
	    else if (nr_names >= MAX_NR_BP)
	    {
	       /* The entry is being added, but there is no space. */
	       WARNING("update_name_array() out of entries in name_array.");
	       nr_names = -1;
	       break;
	    }
	    else
	    {
	       /* The entry is new, add it to the cache. */
	       int j;
	       /* Make room for inserting the new name by moving all
		* posterior names one step in the list.
		*/
	       for (j = nr_names; j > index; j--)
	       {
		  name_array[j] = name_array[j-1];
	       }
	       /* Insert new entry */
	       name_array[index] = curr_name;
	       nr_names++;
	    }
	 }
	 
	 /* Update curr_name for next loop. */
	 curr_name = curr_name + namelen + valuelen + 2;
	 curr_name = (const char *)ALIGN(curr_name, 4);
      } /* while (curr_name < limit) */
   }
   else
   {
      ERROR("update_name_array() received invalid entry");
      nr_names = -1;
   }
   return nr_names;
}

/*
 *=========================================================
 *                      cache_pointers_from_flash
 *=========================================================
 * Description: 
 *
 * The following routine is used to...
 *   1) Read a new set of board parameters from flash
 *   2) Validate the new flash data using its checksum
 *   3) Build a cache entry for the new flash data  
 *
 * Returns:
 *
 * Zero (0) on success, non-zero otherwise.
 *  
 */
static int
cache_pointers_to_flash(const char *flash_base,
		       const char **name_array)
{
   const char *entry_addr = flash_base;
   unsigned short entry_size;
   int nr_names = 0;

   assert(flash_base != NULL);
   assert(name_array != NULL);
   while (read_entry_header(entry_addr, NULL, &entry_size))
   {
      nr_names = update_name_array(name_array, nr_names, entry_addr);
      assert(nr_names <= MAX_NR_BP);
      if (nr_names < 0)
      {
	 assert(!"cache_pointers_to_flash(): call to "
		"update_name_array_failed!");
	 break; /* No more room for names? */
      }
      /* Update "entry" to point to the next potential entry position. */
      entry_addr += entry_size;
   }
   return nr_names;
}



/*
 *=========================================================
 *                      prepend_entry_header
 *=========================================================
 * Description: 
 *	The following routine is used to write a BP entry header
 *	in front of a new entry of board parameters in flash.
 *
 * Returns:
 *	The result from the flash_write() call.
 */
static int
prepend_entry_header(const char *data, size_t data_size,
		     unsigned short version)
{
   char *entry_hdr_addr = (char *)data - sizeof(struct BpEntryHeader);
   struct BpEntryHeader header;
   U32 cksum;
   
   assert(data != NULL);
   assert(data_size > 0);
   
   header.magic    = BP_ENTRY_MAGIC;
   header.size     = (U16)(sizeof(struct BpEntryHeader) + data_size);
   header.checksum = 0;
   header.version  = version;
   
   cksum = calc_checksum(0, &header, sizeof(struct BpEntryHeader));
   cksum = calc_checksum_from_flash(cksum, data, data_size);
   header.checksum = (unsigned short) (0xffff & ~cksum);
   /* Copy the entry header to flash memory. */
   return flash_write(entry_hdr_addr, &header, sizeof(struct BpEntryHeader));
}


/*
 *=========================================================
 *     mirror_flash_entries
 *=========================================================
 * Description: 
 *	The following routine is used to copy the 'current'
 *	contents of one flash block to another (backup) flash
 *	block. The source of the copy operation is
 *	specified as a pointer to a cache structure and the
 *	destination is specified as the start address of a flash
 *	block. The routine will create a single BP entry in the
 *	destination flash block. This entry will contain the
 *	current value of each board parameter, as designated by
 *	the input cache structure.
 *
 * Returns:
 *	The address of the byte in flash immediately following
 *	the last written content, i.e. the write pointer, and NULL
 *	on failure.
 */
static char *
mirror_flash_entries(const char *dest_block,
		     size_t blsize,
		     const char **name_array,
		     int nr_names, unsigned short new_version)
{
   assert(dest_block);
   
   if (nr_names > 0)
   {
      int i;
      const char *limit = dest_block + blsize;
      const char *data_start = dest_block + sizeof(struct BpEntryHeader);
      char *write_ptr = (char *)data_start;
      
      for(i = 0; i < nr_names; i++)
      {
	 const char *curr_name = name_array[i];
	 size_t namelen = strlen_flash(curr_name);
	 size_t valuelen = strlen_flash(curr_name + namelen + 1);
	 /* Determine the length of the current name/value pair. */
	 size_t lenght = ALIGN(namelen + valuelen + 2, 4);
	 
	 assert(namelen > 0);
	 assert(valuelen > 0);
	 if (write_ptr + lenght <= limit)
	 {
	    /* Copy the name/value pair to flash memory. */
	    if (write_from_flash_to_flash(write_ptr, curr_name, lenght) == 0)
	    {
	       write_ptr  += lenght;
	    }
	    else
	    {
	       WARNING("write_from_flash_to_flash() failed!");
	       return NULL; /* Unsuccessful EXIT */
	    }
	 }
	 else
	 {
	    /* Else no space left on flash bank. This should be impossible
	     * since the flash blocks must be equally sized => illegal state.*/
	    ERROR("Run out of space in flash while duplicating entries from "
		  "active flash");
	    /* error() shall not return */
	    /* Make compiler happy by putting a return point here */
	    return NULL;
	 }
      } /* for(i = 0; i < nr_names; i++) */
      
      /* Now we have successfully copied all entry data to the new flash
       * bank. Create an entry header and prepend it to the flash memory
       * containing the data written above.
       */
      if (prepend_entry_header(data_start, (size_t)(write_ptr - data_start),
			       new_version) == 0)
      {
	 /* DONE successfully! */
	 return write_ptr;
      }
      else
      {
	 WARNING("prepend_entry_header failed!");
	 return NULL;
      }
   } /* if (nr_names > 0) */
   else
   {
      /* No names to flush, write pointer is at the beginning of flash. */
      return (char *)dest_block;
   }
}

/*
 *=========================================================
 *     switch_active_flash
 *=========================================================
 * Description: 
 *	switch_active_flash() determines which flash block is currently
 *	active, and then condenses its entries to a single entry
 *	in the other flash block. Finally, the pointers in the
 *	name cache are redirected to point at the names in the new
 *	flash block.
 *
 * Returns:
 *	The address of the byte in flash immediately following
 *	the last written content, i.e. the write pointer, and NULL
 *	on failure.
 */
static int
switch_active_flash(struct BpState *bs)
{
   int retval;
   const char *new_flash_base = bs->inactive_block;
   /* FIXME: new_version should handle version numbers wrapping around
    * and zero as magic number. */
   unsigned short new_version = bs->version + 1;
   
   if (flash_erase((char *)new_flash_base, bs->block_size) == FLASH_EOK)
   {
      char *new_write_ptr = mirror_flash_entries(new_flash_base,
						 bs->block_size,
						 bs->name_array,
						 bs->nr_names,
						 new_version);
      if (new_write_ptr != NULL)
      {
	 /* Update name list to point at names in the new active flash */
	 int new_nr_names = cache_pointers_to_flash(new_flash_base,
						    bs->name_array);
	 if (new_nr_names != bs->nr_names)
	 {
	    /* Internal error */
	    ERROR("Nr names differ in BP set in new flash!");
	    retval = -1;
	 }
	 else
	 {
	    /* Success */
	    const char *tmp = bs->inactive_block;
	    bs->inactive_block = bs->active_block;
	    bs->active_block = tmp;
	    bs->write_ptr = new_write_ptr;
            bs->version++;
	    retval = 0; /* succeeded */

	 }
      }
      else
      {
	 WARNING("mirror_flash_entries() failed!");
	 retval = -1;
      }
   }
   else
   {
      WARNING("switch_active_flash(): could not erase flash block"); 
      retval = -1;
   }
   return retval;
}

/*
 *=========================================================
 *                     probe_for_old_bp
 *=========================================================
 * Description:
 *
 */
static const char *
probe_for_old_bp(const char *block1, const char *block2, size_t block_size,
		 char **write_ptr, unsigned short *version)
{
   const char *active_flash_base;
   char *wp1;
   char *wp2;
   unsigned short v1  = get_version(block1, &wp1);
   unsigned short v2  = get_version(block2, &wp2);
   unsigned short vf1 = get_first_version(block1);
   unsigned short vf2 = get_first_version(block2);
   Boolean block1_empty = (block1 == wp1);
   Boolean block2_empty = (block2 == wp2); 

   /* Make the flash image with the newest version active. */

   /* FIXME: wrapping version numbers are not handled! Should check the version
    * of the last entry in one flash block against the first one in the other, 
    * to see if they are consecutive. Version "zero" has a magic meaning here
    * that should be handled when setting versions to entry headers.
    */

#if 0

   /* If both headers are valid, see which one has the highest version. */
   if (v1 > v2)
   {
      active_flash_base   = block1;
      *write_ptr = wp1;
      *version = v1;
   }
   else if (v2 > v1)
   {
      active_flash_base   = block2;
      *write_ptr = wp2;
      *version = v2;
   }
   else
   {
      /* Else both flash blocks have the same version, which is illegal
       * unless v1 == v2 == zero, i.e. both are empty from board parameters.
       * Let's set either of the blocks as active, for instance block 1.
       */
      active_flash_base = block1;
      *write_ptr = wp1;
      *version = v1;
   }
#else
   if(block2_empty)
   {
      active_flash_base   = block1;
      *write_ptr = wp1;
      *version = v1;
   }
   else if(block1_empty)
   {
      active_flash_base   = block2;
      *write_ptr = wp2;
      *version = v2;
   }   
   else if ((v1+1) == vf2)
   {
      active_flash_base   = block2;
      *write_ptr = wp2;
      *version = v2;
   }
   else if((v2+1) == vf1)
   {
      active_flash_base   = block1;
      *write_ptr = wp1;
      *version = v1;
   }
   else
   {
      /* This should NEVER happen */
      active_flash_base   = block1;
      *write_ptr = wp1;
      *version = v1;
   }
#endif

   return active_flash_base;
}

/*
 * The check will be done for every read until the fam is up to ensure the initialization
 * of the parameter state "bs" variable for different type of flash interfaces
 * like i.e. QSPI Serial NOR Flash
 */
static int
check_initialise_for_reads(struct BpState *bs)
{
   static Boolean ok; /* Initialized to False by crt0. */
   int retval;
   
   if (ok)
   {
      /* BP area is already initialized. */
      retval = INIT_DONE;
   }
   else
   {
      static Boolean early_tried; /* Initialized to False by crt0. */
      static Boolean try_after_fam; /* Initialized to False by crt0. */

      /* check if flash api is available (if FAM is up). */
      fam_is_up = flash_is_api_available();
     
      if ( (False == early_tried) || (fam_is_up) )
      {
         /* this "if" case should be entered maximum 2 times depending on "direct_access" value.
          * First time when "early_tried" is False. If "direct_access" is yes or not 
          * specified, flash boardparam blocks are initialized and "ok" variable becomes true. 
          * This way current "if" case will never be reached again. 
          * 
          * If "direct_access" is "no", then "try_after_fam" becomes true and current if case 
          * will be entered once more when "fam_is_up", flash boardparam blocks will be initialized 
          * and "ok" will be set to true.
          */
         long bp_base;
         long bp_size;
         Boolean old_mode;
         early_tried = True;

         /* Set safe mode to avoid calling boardparam_read recursively. */
         old_mode = sysparam_set_safe_mode(True);

         if (((sysparam_readattr_long(BP_PARAM_NAME, "base", &bp_base) == 0)
              && (sysparam_readattr_long(BP_PARAM_NAME, "size", &bp_size) == 0)
              && bp_size >= BP_MIN_SIZE
              && bp_size <= BP_MAX_SIZE)
             ||
             ((sysparam_readattr_long(BP_PARAM_NAME_OLD, "base", &bp_base) == 0)
              && (sysparam_readattr_long(BP_PARAM_NAME_OLD, "size", &bp_size) == 0)
              && bp_size >= BP_MIN_SIZE
              && bp_size <= BP_MAX_SIZE
                ))
         {
            /* check if we should enable early flash access before FAM is up. */
            if( False == try_after_fam )
            {
               /* By default, early flash access should be enabled. */
               char direct_access_str[4] = "yes";
           
               if( (sysparam_readattr_string(BP_PARAM_NAME, "direct_access",
                    direct_access_str, sizeof(direct_access_str))== 0) ||
                   (sysparam_readattr_string(BP_PARAM_NAME_OLD, "direct_access",
                    direct_access_str, sizeof(direct_access_str))== 0)
                 )
               {
                  if(0 == strcmp(direct_access_str, "no"))
                  {
                     /* start boardparam only after FAM is up. */
                     /* BP area is not yet initialized. The initialization will be done
                      * when FAM is up because early flash access is disabled.  */
                     try_after_fam = True;
                  }
               }
            }
           
            if( (False == try_after_fam) || (fam_is_up) )
            {
               size_t blsz = (size_t)bp_size / 2;
               const char *block1 = (const char *)bp_base;
               const char *block2 = (const char *)bp_base + blsz;
               const char *act = probe_for_old_bp(block1, block2, blsz,
                                               &bs->write_ptr, &bs->version);
            
               assert(act == block1 || act == block2);
	    
               /* BP area found in flash which we can read. Prepare read-cache. */
               bs->nr_names = cache_pointers_to_flash(act, bs->name_array);
               bs->active_block = act;
               bs->inactive_block = (act == block1) ? block2 : block1;
               bs->block_size = blsz;
               /* Done successfully */
               ok = True;

               /* Bp area is now initialized. */
               retval = INIT_DONE;
            }
            else
            {
               /* boardparam area was not intialized because FAM is not yet up. */
               retval = INIT_NOT_DONE;
            }
         }
         else
         {
            WARNING("Parameter " BP_PARAM_NAME " missing or invalid.");
            /* BP area is not intialized for reads because boardparams configuration 
             * values are missing from BSP configuration file. */
            retval = INIT_NOT_DONE; /* Failed to read configuration */
         }
         
         sysparam_set_safe_mode(old_mode);
      }
      else
      {            
         /* this case should be entered every time until FAM is up, but only when 
          * early flash access is disabled. */
         /* the flash parameters are not yet available because FAM is not yet up
          * and early flash access is disabled. */            
         retval = INIT_NOT_DONE;
      }
   }
   return retval;
}


/*
 *=========================================================
 *             initialise_for_write
 *=========================================================
 * Description:
 *	This function prepares the boardparam module's control
 *	data and the active flash for writes. This implies
 *	finding which flash block has the newest entries, controlling
 *	that the remaining area in the flash block is writeable
 *	(not corrupted) and switch active flash blocks otherwise.
 *	The function assumes single-threaded execution.
 *
 * Returns:
 *	The resulting write pointer, NULL for failure.
 */
static int
prepare_to_write(struct BpState *bs, size_t data_size)
{
   static Boolean hw_error; /* Initialized to False by crt0. */
   if (hw_error == False)
   {
      Boolean switched_once = False;
      size_t space_left = (size_t)(bs->active_block + bs->block_size
				   - bs->write_ptr);
      size_t needed_size = data_size + sizeof(struct BpEntryHeader);
      
      assert(bs);
      if (bs->nr_names >= 0)
      { 
         /* Determine the amount of storage available in the active flash.*/
         /* If the amount of space available in the active flash is
          * insufficient, switch to the inactive flash and calculate the
          * amount of storage available in the other block of flash.
          */
         assert(needed_size < bs->block_size);
         if (space_left < needed_size)
         {
	    /* FIXME: we could beforehand calculate if the switch of flashes
	     * would really satisfy the space requirements. */
	    /* Switch to the inactive_flash image */
	    WARNING("Info: switching active flash erase-block because current is "
	        	 "full.");
	    if (switch_active_flash(bs) == 0)
	    {
	       switched_once = True;
	       assert(bs->write_ptr != NULL);
	       space_left = (size_t)(bs->active_block +
				     bs->block_size - bs->write_ptr);
	    }
	    else
	    {
	       ERROR("switch_active_flash() failed!");
	       hw_error = True;
	    }
         }
      
         /* Check again */
         if (space_left >= needed_size)
         {
	    if (!is_erased_flash_area(bs->write_ptr, needed_size))
	    {
	       /* Found unerased (corrupted) flash beyond the write ptr,
	        * switch active blocks to get a fresh flash to continue with.
	        */
	       if (switched_once == False)
	       {
	          WARNING("Switching active flash erase-block because current is "
		          "not erased at write point.");
	          if (switch_active_flash(bs) != 0)
	          {
		     WARNING("initialise_for_writes(): call to "
			     "swich_active_flash() failed!");
		     hw_error = True;
	          }
	          /* Else succeded! */
	       }
	       else
	       {
	          /* unerased flash in what we just erased!! */
	          hw_error = True;
	       }
	    }
	    /* Else OK! */
         }
         else
         {
	    /* Could be that the block really is too full (condensed into one
	     * entry) to hold this new parameter.
	     */
	    WARNING("Could not get enough free flash despite having switched "
	   	    "active flash eraseblock!");
	    hw_error = True;
         }
      }
      else
      {
         hw_error = True;
      }
   }
   return hw_error ? -1 : 0;
}

/*
 *=========================================================
 *                      handle_boardparam_read
 *=========================================================
 * Description:
 *
 * The following routine will search the active cache
 * for the name/value pair specified.
 *
 */

static size_t
handle_boardparam_read(struct BpState *bs, const char *req_name,
		       char *buffer, size_t buffer_size)
{
   size_t value_len;
   
   assert(req_name != NULL);
   assert(*req_name != '\0');
   /* if buffer is non-null, size must be > 0 */ 
   assert((buffer_size == 0) || (buffer != NULL));

   if (bs->nr_names >= 0)
   {
      const char *name_in_flash = find_bp_name(req_name, bs->name_array,
					       bs->nr_names, NULL, False);
      
      if (name_in_flash != NULL)
      {
	 size_t name_len = strlen_flash(name_in_flash);
	 /* The value string immediately follows the name's terminating
	  * null character.
	  */
	 const char *value_addr = name_in_flash + name_len + 1;
	 value_len = strlen_flash(value_addr);
	 assert(name_len > 0);
	 /* Parameter found, copy the results if there's a buffer.*/
	 if ((buffer != NULL) && (buffer_size > 0))
	 {
	    size_t cpy_sz = value_len < buffer_size
	       ? value_len : buffer_size - 1;
	    read_from_flash(buffer, value_addr, cpy_sz);
	    buffer[cpy_sz] = '\0'; /* null terminate even if truncated */
	 }
	 /* SUCCESS */
      }
      else
      {
	 value_len = 0;
      }
   }
   else
   {
      value_len = 0;
   }
   return value_len;
}
   


/*
 *=========================================================
 *                      boardparam_write
 *=========================================================
 * Description:
 *	The following routine will add or modify one board parameter.
 */

static int
handle_boardparam_write(struct BpState *bs, const char *name,
			const char *value)
{
   int retval;
   size_t namelen;
   size_t valuelen;
   size_t data_size = 0;
   char *name_buffer = (char *)alloc(SYSPARAM_MAX_NAME_LENGTH + 1, 0);
   char *value_buffer = (char *)alloc(SYSPARAM_MAX_VALUE_LENGTH + 1, 0);
   char *parsed_name = name_buffer;
   char *parsed_value = value_buffer; 

   assert(name_buffer != NULL);
   assert(value_buffer != NULL);
   assert(name != NULL);
   assert(value != NULL);
   assert(bs);
   if (bs->nr_names < 0)
   {
     WARNING("Parameters must be initialised for wtite");
     retval = -1;
     goto cleanup;
   }

   /* Check that the name size are within range */
   namelen = strlen(name);
   if (namelen > SYSPARAM_MAX_NAME_LENGTH)
   {
     WARNING2("Parameter name is to long, maximum size is %u characters", SYSPARAM_MAX_NAME_LENGTH);
     retval = -1;
     goto cleanup;
   }

   /* Check that the value size are within range */
   valuelen = strlen(value);
   if (valuelen > SYSPARAM_MAX_VALUE_LENGTH)
   {
     WARNING2("Parameter value is to long, maximum size is %u characters", SYSPARAM_MAX_VALUE_LENGTH);
     retval = -1;
     goto cleanup;
   }

   /* Create working copies of the strings */
   strcpy(name_buffer, name);
   strcpy(value_buffer, value);

   remove_start_and_end_spaces_from_string(&parsed_name);
   remove_start_and_end_spaces_from_string(&parsed_value);
   namelen = strlen(parsed_name);
   valuelen = strlen(parsed_value);

   assert(namelen > 0);

   /* "data_size" must be 32-bit aligned due to flash drv restrictions */
   data_size = ALIGN(namelen + valuelen + 2, 4);

   /* Check that the active flash is OK for writes (we may have an
    * active_flash with valid old BPs followed by corrupted data).
    */

   if (prepare_to_write(bs, data_size) == 0)
   {
      const char *name_in_flash = find_bp_name(name, bs->name_array,
					       bs->nr_names, NULL, False);
      /* If the parameter is to be deleted, check that it is present in
       * the current BP set. */
      if ((valuelen != 0) || (name_in_flash != NULL))
      {
	 /* FIXME: If the parameter is to be modified, check that the
	  * new value is different from the previous value. */
 
	 /* Copy the requested name + value to flash memory. */
	 /* The buffering into a signal before the flash-write is a
	  * patch due to the intelx16.c driver not being able to take
	  * non-word-aligned data for writes.
	  */
	 enum FlashResult flash_result;
	 char *write_buffer = (char *) alloc(data_size, 0);
	 char *data_addr = (bs->write_ptr
			    + sizeof(struct BpEntryHeader));
         memset(write_buffer, 0, data_size);
	 strcpy(write_buffer, parsed_name);
	 strcpy(write_buffer + namelen + 1, parsed_value);

	 flash_result = flash_write(data_addr, (void *)write_buffer,data_size);
	 if (flash_result == FLASH_EOK)
	 {
	    /* FIXME: the assignment of 'new_version should handle
	     * the version number wrapping around, and zero as magic
	     * value.
	     */
	    unsigned short new_version = bs->version + 1;
	    
	    /* Create an entry header and prepend it to the
	     * flash memory containing the data written
	     * above.
	     */
	    if (prepend_entry_header(data_addr, data_size,
				     new_version) == 0)
	    {
	       const char *entry_addr = bs->write_ptr;

	       int new_nr_names = update_name_array(bs->name_array,
						    bs->nr_names,
						    entry_addr);
	       if (new_nr_names >= 0)
	       {
		  size_t needed_size = data_size
		     + sizeof(struct BpEntryHeader);
		  assert(new_nr_names <= MAX_NR_BP);
		  bs->nr_names = new_nr_names;
		  bs->write_ptr += needed_size;
		  bs->version++;
		  retval = 0; /* Done successfully! */
	       }
	       else
	       {
		  WARNING("update_name_array() failed!");
		  retval = -1;
	       }
	    }
	    else
	    {
	       WARNING("prepend_entry_header() failed!");
	       retval = -1;
	    }
	 }
	 else
	 {
	    WARNING4("flash_write() failed with code 0x%x when trying to "
		     "write %d bytes to 0x%08lx", flash_result, data_size,
		     data_addr);
	    retval = -1;
	 }
	 free_buf((union SIGNAL **)&write_buffer);
      }
      else
      {
	 /* Trying to delete a non-existing parameter. */
	 WARNING("Trying to delete a non-existing parameter.");
	 inparam_assert(!"Trying to delete a non-existing parameter.");
	 retval = -1;
      }
   } /* if (prepare_to_write(bs, data_size) == 0) */
   else
   {
      /* Else, could not initialize for writes */
      WARNING("boardparam_write() could not initialize for writes");
      /* FIXME: call error()? */
      retval = -1;
   }

 cleanup:
   
   free_buf((union SIGNAL **)&name_buffer);
   free_buf((union SIGNAL **)&value_buffer);

   return retval;
}

   
/*
 *=========================================================
 *	boardparam_list_next                     
 *=========================================================
 * Description:
 *
 * The following routine will iterate through the
 * board parameter namesapce.  
 *
 */

static size_t
handle_boardparam_list_next(struct BpState *bs, const char *prev_name,
			    char *buffer, size_t buffer_size)
{
   size_t len;
   
   assert(prev_name != NULL);
   assert((buffer_size == 0) || (buffer != NULL));
   assert(bs);

   /* The active cache cannot be allowed to change
    * during an individual search/read operation.
    */
   
   if (bs->nr_names >= 0)
   {
      int i;
      /* If the prev_name is not NULL, then find the name
       * that follows (or would follow) it in the list of
       * board parameter names.
       */
      if (prev_name[0] != '\0')
      {
	 if ((find_bp_name(prev_name, bs->name_array, bs->nr_names, &i, False))
	     != NULL)
	 {
	    i++;
	 }
	 /* Else "i" gets updated by find_bp_name with the index of the
	  * name that would follow alphabetically.
	  */
      }
      else /* Else start with the first name in the list.*/
      {
	 i = 0;
      }
      
      /* The index specifies a valid BP entry, try and
       * return the value of that nr_entries name.
       */
      if (i < bs->nr_names)
      {
	 const char *name_in_flash = bs->name_array[i];
	 len  = strlen_flash(name_in_flash);
	 
	 /* Parameter found, copy the results.  */
	 if ((buffer != NULL) && (buffer_size > 0))
	 {
	    size_t cpy_sz = len < buffer_size ? len : buffer_size - 1;
	    read_from_flash(buffer, name_in_flash, cpy_sz);
	    buffer[cpy_sz] = '\0';
	 }
	 /* Done successfully. */
      }
      else /* Else end of table reached. */
      {
	 len = 0;
      }
   } /* Else no board parameters cached */
   else
   {
      len = 0;
   }
   return len;
}


/*
 *=========================================================
 *      ose_boardparam
 *=========================================================
 * Description:
 *
 * This is the proxy process for the board parameter handling whose
 * main purpose is to synchronize reads and writes.
 */

static void
ose_boardparam(void)
{
   for (;;)
   {
      struct BpState *bs = &bp_state;
      static const SIGSELECT sig_any[] = {0};
      union SIGNAL *req = receive(sig_any);
      union SIGNAL *reply;
      
      switch(req->sig_no)
      {
	 case BOARDPARAM_READ_REQUEST:
	 {
	    size_t len = handle_boardparam_read(bs, req->read_req.name,NULL,0);
	    reply = alloc(sizeof(struct BoardparamReadReply) + len,
			  BOARDPARAM_READ_REPLY);
	    if (len > 0)
	    {
	       (void)handle_boardparam_read(bs, req->read_req.name,
					    reply->read_reply.value,
					    len + 1);
	    }
	    else
	    {
	       reply->read_reply.value[0] = '\0';
	    }
	    break;
	 }
	 case BOARDPARAM_LIST_NEXT_REQUEST:
	 {
	    size_t len = handle_boardparam_list_next(bs, req->list_req.str,
						     NULL, 0);
	    reply = alloc(sizeof(struct BoardparamListNextReply) + len,
			  BOARDPARAM_LIST_NEXT_REPLY);
	    if (len > 0)
	    {
	       (void)handle_boardparam_list_next(bs, req->list_req.str,
						 reply->list_reply.name,
						 len + 1);
	    }
	    else
	    {
	       reply->list_reply.name[0] = '\0';
	    }
	    break;
	 }
	 case BOARDPARAM_WRITE_REQUEST:
	 {
	    const char *name = req->write_req.name_value_pair;
	    const char *value = name + strlen(name) + 1;
	    int result = handle_boardparam_write(bs, name, value);
	    reply = alloc(sizeof(struct BoardparamWriteReply),
			  BOARDPARAM_WRITE_REPLY);
	    reply->write_reply.result = result;
	    break;
	 }
	 default:
	    ERROR("Unknown signal arrived at ose_boardparam");
      } /* switch(req->sig_no) */
      send(&reply, sender(&req));
      free_buf(&req);
   } /* for (;;) */
}

/*
 *=========================================================
 *                      signal_boardparam_read()
 *=========================================================
 * Description:
 *	This function transmits a boardparam_read-request to the
 *	ose_boardparam process.
 *
 */
static size_t
signal_boardparam_read(PROCESS ose_boardparam_pid, const char *req_name,
		       char *buffer, size_t buffer_size)
{
   size_t value_len;
   static const SIGSELECT sigsel_reply[] = { 1, BOARDPARAM_READ_REPLY };
   union SIGNAL *sig = alloc(sizeof(struct BoardparamReadRequest)
			     + strlen(req_name), BOARDPARAM_READ_REQUEST);
   strcpy(sig->read_req.name, req_name);
   send(&sig, ose_boardparam_pid);
   sig = receive(sigsel_reply);
   value_len = strlen(sig->read_reply.value);
   if ((buffer != NULL) && (buffer_size > 0))
   {
      size_t cpy_sz = value_len < buffer_size ? value_len : buffer_size - 1;
      memcpy(buffer, sig->read_reply.value, cpy_sz);
      buffer[cpy_sz] = '\0';
   }
   free_buf(&sig);
   return value_len;
}


/*
 *=========================================================
 *                      signal_boardparam_write()
 *=========================================================
 * Description:
 *	This function transmits a boardparam_write-request to the
 *	ose_boardparam process.
 */
static int
signal_boardparam_write(PROCESS ose_boardparam_pid, const char *name,
			const char *value)
{
   int retval;
   size_t namelen = strlen(name);
   size_t valuelen = strlen(value);
   static const SIGSELECT sigsel_reply[] = { 1, BOARDPARAM_WRITE_REPLY };
   union SIGNAL *sig = alloc(sizeof(struct BoardparamWriteRequest)
			     + namelen + valuelen, BOARDPARAM_WRITE_REQUEST);
   
   strcpy(sig->write_req.name_value_pair, name);
   strcpy(sig->write_req.name_value_pair + namelen + 1, value);
   send(&sig, ose_boardparam_pid);
   sig = receive(sigsel_reply);
   retval = sig->write_reply.result;
   free_buf(&sig);
   return retval;
}


/*
 *=========================================================
 *	signal_boardparam_list_next                     
 *=========================================================
 * Description:
 *
 * The following routine will transmit a boardparam_list_next-request
 * to the ose_boardparam_process using signals.
 *
 */

static size_t
signal_boardparam_list_next(PROCESS ose_boardparam_pid, const char *str,
			    char *buffer, size_t buffer_size)
{
   size_t len;
   static const SIGSELECT sigsel_reply[] ={1,BOARDPARAM_LIST_NEXT_REPLY};
   union SIGNAL *sig = alloc(sizeof(struct BoardparamListNextRequest)
			     + strlen(str), BOARDPARAM_LIST_NEXT_REQUEST);
   strcpy(sig->list_req.str, str);
   send(&sig, ose_boardparam_pid);
   sig = receive(sigsel_reply);
   len = strlen(sig->list_reply.name);
   if ((buffer != NULL) && (buffer_size > 0))
   {
      size_t cpy_sz = len < buffer_size ? len : buffer_size - 1;
      memcpy(buffer, sig->list_reply.name, cpy_sz);
      buffer[cpy_sz] = '\0';
   }
   free_buf(&sig);
   return len;
}
 

static PROCESS
fork_ose_boardparam(void)
{
   static PROCESS ose_boardparam_being_created_by; /* Initialized to 0 by crt0. */
   PROCESS curr_proc = current_process();
   PROCESS bp_pid;
   U32 mask;
   
   LOCK_SAVE(mask);
   if (ose_boardparam_being_created_by == 0)
   {
      ose_boardparam_being_created_by = curr_proc;
      LOCK_RESTORE(mask);
      bp_pid = create_process(OS_PRI_PROC, "ose_boardparam",
			      ose_boardparam,
			      512, 10, 0,0,0,0,0);
      start(bp_pid);
   }
   else
   {
      /* FIXME: attach on creator's pid in case it dies? */
      static const SIGSELECT sigsel[] = { 1, BOARDPARAM_HUNT };
      union SIGNAL *sig;
      
      LOCK_RESTORE(mask);
      sig = alloc(sizeof(SIGSELECT), BOARDPARAM_HUNT);
      (void)hunt("ose_boardparam", 0, 0, &sig);
      sig = receive_w_tmo(HUNT_TMO, sigsel);
      if (sig != NIL)
      {
	 bp_pid = sender(&sig);
	 free_buf(&sig);
      }
      else
      {
	 ERROR("hunt for ose_boardparam process timed out!");
	 bp_pid = 0; /* error() should not return */
      }
   }
   return bp_pid;
}

/*
 **********************************************************
 *   PUBLIC FUNCTIONS
 **********************************************************
 */

/*
 *=========================================================
 *                      boardparam_read
 *=========================================================
 * Description:
 *
 * The following routine will search the active cache
 * for the name/value pair specified.
 *
 */

size_t
boardparam_read(const char *req_name, char *buffer, size_t buffer_size)
{
   size_t value_len;
   
   if ((req_name != NULL)
       && (*req_name != '\0')
       && ((buffer_size == 0) || (buffer != NULL))) /* if buffer is non-null,
						     * size must be > 0 */ 
   {
      struct BpState *bs = &bp_state;
      if (check_initialise_for_reads(bs) == 0)
      {
	 unsigned long cpumask;
      
	 LOCK_SAVE(cpumask);
	 if (bs->ose_bp_pid != 0)
	 {
	    LOCK_RESTORE(cpumask);
	    value_len = signal_boardparam_read(bs->ose_bp_pid, req_name,
					       buffer, buffer_size);
	 }
	 else if (flash_is_api_available())
	 {
	    fam_is_up = True;
	    LOCK_RESTORE(cpumask);
	    bs->ose_bp_pid = fork_ose_boardparam();
	    value_len = signal_boardparam_read(bs->ose_bp_pid, req_name,
					       buffer, buffer_size);
	 }
	 else
	 {
	    value_len = handle_boardparam_read(bs, req_name, buffer,
					       buffer_size);
	    LOCK_RESTORE(cpumask);
	 }
      }
      else
      {
	 /* Failed to initialise for reads (perhaps initialisation ongoing).*/
	 value_len = 0;
      }
   }
   else /* else in-parameter check failed */
   {
      inparam_assert(!"Illegal in-parameter to boardparam_read()");
      value_len = 0;
   }
   return value_len;
}



/*
 *=========================================================
 *                      boardparam_write
 *=========================================================
 * Description:
 * (public function) The following routine will add or
 * modify one board parameter.
 */

int
boardparam_write(const char *name, const char *value)
{
   int retval;
   
   if ((name != NULL)
       && (name[0] != '\0')
       && (value != NULL))
   {
      struct BpState *bs = &bp_state;

      if (check_initialise_for_reads(bs) == 0)
      {
	 unsigned long cpumask;
	 LOCK_SAVE(cpumask);
	 if (bs->ose_bp_pid != 0)
	 {
	    LOCK_RESTORE(cpumask);
	    retval = signal_boardparam_write(bs->ose_bp_pid, name, value);
	 }
	 else if (flash_is_api_available())
	 {
	    fam_is_up = True;
	    LOCK_RESTORE(cpumask);
	    bs->ose_bp_pid = fork_ose_boardparam();
	    retval = signal_boardparam_write(bs->ose_bp_pid, name, value);
	 }
	 else
	 {
	    LOCK_RESTORE(cpumask);
	    assert(!"called boardparam_write before FAM was up!!");
	    retval = -1;
	 }
      }
      else
      {
        WARNING("Init failed\n");
        
	 /* could not initialise for reads (config missing?) */
	 retval = -1;
      }
   }
   else /* Illegal in-parameter to boardparam_write */
   {
      WARNING("Illegal argument to boardparam_write()");
      inparam_assert(!"Illegal argument to boardparam_write()");
      retval = -1;
   }
   return retval;
}

   
/*
 *=========================================================
 *	boardparam_list_next                     
 *=========================================================
 * Description:
 *
 * The following routine will iterate through the
 * board parameter namesapce.  
 *
 */

size_t
boardparam_list_next(const char *str, char *buffer, size_t buffer_size)
{
   size_t len;
   if ((str != NULL) && ((buffer_size == 0) || (buffer != NULL)))
   {
      struct BpState *bs = &bp_state;

      if (check_initialise_for_reads(bs) == 0)
      {
	 unsigned long cpumask;
	 
	 LOCK_SAVE(cpumask);
	 if (bs->ose_bp_pid != 0)
	 {
	    LOCK_RESTORE(cpumask);
	    len = signal_boardparam_list_next(bs->ose_bp_pid, str, buffer,
					      buffer_size);
	 }
	 else if (flash_is_api_available())
	 {
	    fam_is_up = True;
	    LOCK_RESTORE(cpumask);
	    bs->ose_bp_pid = fork_ose_boardparam();
	    len = signal_boardparam_list_next(bs->ose_bp_pid, str, buffer,
					      buffer_size);
	 }
	 else
	 {
	    len = handle_boardparam_list_next(&bp_state, str, buffer,
					      buffer_size);
	    LOCK_RESTORE(cpumask);
	 }
      }
      else
      {
	 /* Could not initialise for reads */
	 len = 0;
      }
   }
   else /* In-parameter check failed. */
   {
      WARNING("Illegal parameter to boardparam_list_next()!");
      inparam_assert(!"Illegal parameter to boardparam_list_next()!");
      len = 0;
   }
   return len;
}
