/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2006 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */
/****************************************************************
 *   Description:
 *	This file defines the interface to functions to
 *	list and read load-module parameters in the core module (OSE-monolith),
 *	to be used by the System Parameter API.
 */

#ifndef OSE_LMPARAM_H_
#define OSE_LMPARAM_H_

#include "stddef.h"

#ifdef __cplusplus
extern "C" {
#endif

   
   /****************************************************************
    *  Function:
    *	lmparam_read
    *
    *  Purpose:
    *   This function is used to retrieve the string that is the value of a
    *	configuration parameter. The value is written to the submitted buffer,
    *	NULL-terminated except when the buffer size is equal or less than the
    *	parameter-value length, in which case the value string will be
    *	truncated.
    *
    *  Parameters:
    *	name:		Specifies the name of the parameter to be read.
    *   buffer:		Where the value of the parameter should be written.
    *	buffer_size:	Self-explaining.
    *
    *  Returns:
    *	The actual length of the parameter value, (regardless if the buffer
    *	was too small to hold it and the value had to be truncated), without
    *	trailing NULL. If the parameter was not found, the function returns 0.
    *
    *  Side effects:
    *	If a parameter with matching name was found, its value is written to
    *	the memory pointed to by 'buffer'. If buffer_size is bigger than the
    *	actual length, a terminating NULL character is written after the
    *	parameter value.
    ****************************************************************
    */
   size_t lmparam_read(const char *name, char *buffer, size_t buffer_size);
   
   
   /****************************************************************
    *  Function:
    *	lmparam_list_next
    *
    * Purpose:
    *   This function supports iterative traversal of the configuration
    *	parameter namespace. The function takes the string submitted as
    *	"str" and searchs for the parameter name that would
    *	lexicographically follow.
    *
    * Example:
    *	This example would print all parameters, except those larger than
    *	NAME/VAL_BUF_SIZE. If too long names are encountered, the loop is
    *	left because we would probably not be able to iterate further.
    *
    *		char name[NAME_BUF_SIZE];
    *		size_t len = lmparam_list_next("", name, NAME_BUF_SIZE);
    *		while (len > 0)
    *		{
    *			if (len < NAME_BUF_SIZE)
    *			{
    *				char value[VAL_BUF_SIZE];
    *				size_t val_len = lmparam_read(name,value,
    *								VAL_BUF_SIZE);
    *				if (val_len > 0 && val_len < VAL_BUF_SIZE)
    *				{
    *					printf("%s=\"%s\"\n", name, value);
    *				}
    *			}
    *			else
    *			{
    *				break;
    *			}
    *			len = lmparam_list_next(name, name, NAME_BUF_SIZE);
    *		}
    *
    *	Parameters:
    *		str:		Specifies the starting point for the search
    *				as described  above. It may be a be a pointer
    *				to NULL, in which case the (lexicographically)
    *				first configuration parameter is searched for.
    *		buffer:		Pointer to the buffer where this function will
    *				write the found parameter name, or its first
    *				bytes up to "buffer_size" if "buffer_size" is
    *				smaller	than the found parameter name length.
    *				May be NULL, in which case the copy will be
    *				ignored (but the found parameter name lenght
    *				will still be returned).
    *		buffer_size:	Self-explaining. Must be 0 if "buffer" is NULL.
    *
    *  Returns:
    *		The function returns the lenght of the found parameter name,
    *		regardless if the string copied to "buffer" was truncated
    *		(because it was too small to hold it) or "buffer" was NULL.
    *		If no parameter name lexicographically following "str",
    *		could be found, the function returns zero.
    ****************************************************************
    */
   size_t lmparam_list_next(const char *str, char *buffer,size_t buffer_size);
   
#ifdef __cplusplus
}
#endif

#endif /* #ifndef OSE_LMPARAM_H_ */
