/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2006 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

/**
 * @private
 * @file sysparam_command
 * @toc api/core/init/sysparam
 * @brief System parameter shell commands
 * @long  file uses the sysparam.h API to implement OSE Shell commands
 *	  to read and write/modify the system parameters.
 */

#define MALLOC_PRIVATE
#include <malloc.h>
#include "stdio.h"
#include "string.h"
#include "shell.h"
#include "efs.h" /* for eprintf() */
#include "sysparam.h"
#include <ctype.h>
#include "lmparam.h"
#include "boardparam.h"

#define STR(x) #x
#define STRSTR(x) STR(x)

static int
cmd_sp_read(int argc, char **argv)
{
   const char *prefix;
   char filter;
   int retval;
   
   switch (argc)
   {
      case 1:
	 prefix = "";
	 filter = 0;
	 retval = RET_SUCCESS;
	 break;
      case 2:
	 if ((strcmp(argv[1], "-b") == 0) || (strcmp(argv[1], "-l") == 0))
	 {
	    prefix = "";
	    filter = argv[1][1];
	 }
	 else
	 {
	    prefix = argv[1];
	    filter = 0;
	 }
	 retval = RET_SUCCESS;
	 break;
      case 3:
	 if ((strcmp(argv[1], "-b") == 0) || (strcmp(argv[1], "-l") == 0))
	 {
	    prefix = argv[2];	
	    filter = argv[1][1];
	    retval = RET_SUCCESS;
	    break;
	 }
	 /* Else syntax error, fall through to 'default' label */
      default:
	 (void)eprintf("Syntax error. Usage: spr [-b | -l] [<prefix>]\n");
	 retval = RET_ERROR;
   }

   if (retval == RET_SUCCESS)
   {
      size_t prefix_len = strlen(prefix);
      if (prefix_len < SYSPARAM_MAX_NAME_LENGTH)
      {
	 char name[SYSPARAM_MAX_NAME_LENGTH + 1];
	 char value[SYSPARAM_MAX_VALUE_LENGTH + 1];
	 
	 strcpy(name, prefix);
	 if (filter == 0)
	 {
	    do
	    {
	       if (sysparam_read(name, value, sizeof(value)) > 0)
	       {
		  printf("%s=\"%s\"\n", name, value);
	       }
	    } while ((sysparam_list_next(name, name, sizeof(name)) > 0)
		     && (strncmp(name, prefix, prefix_len) == 0));
	 }
	 else if (filter == 'l')
	 {
	    do
	    {
	       if (lmparam_read(name, value, sizeof(value)) > 0)
	       {
		  printf("%s=\"%s\"\n", name, value);
	       }
	    } while ((lmparam_list_next(name, name, sizeof(name)) > 0)
		     && (strncmp(name, prefix, prefix_len) == 0));
	 }
	 else if (filter == 'b')
	 {
	    do
	    {
	       if (boardparam_read(name, value, sizeof(value)) > 0)
	       {
		  printf("%s=\"%s\"\n", name, value);
	       }
	    } while ((boardparam_list_next(name, name, sizeof(name)) > 0)
		     && (strncmp(name, prefix, prefix_len) == 0));
	 }
	 else
	 {
	    (void)eprintf("Syntax error. Usage: spr [-b | -l] [<prefix>]\n");
	    retval = RET_ERROR;
	 }
      }
      else
      {
	 eprintf("Parameter name too long: %s.\n", argv[1]);
	 retval = RET_ERROR;
      }
   }
   return retval;
}

static int
cmd_sp_write(int argc, char **argv)
{
   if (argc == 3)
   {
      Boolean found_equal_sign = False;
      unsigned int len;
      char *name = argv[1];
      char *value = argv[2];

      len = strlen(name);
      if (name[len - 1] == '=')
      {
         /* Remove trailing equal sign parameter name */
         name[len - 1] = '\0';

         found_equal_sign = True;
      }
      
      if (value[0] == '=')
      {
         /* Remove beginning equal sign from parameter value */
         value++;

         found_equal_sign = True;
      }

      if (found_equal_sign)
      {
         if (sysparam_write(name, value) != 0)
         {
            printf("System parameter write failed.\n");
         }
      }
      else
      {
	 goto syntax_error;
      }
   }
   else if (argc == 2)
   {
      const char *name = argv[1];
      char *value = strchr(name, '=');
      /* Check for embedded assignment syntax. */
      if (value != NULL)
      {
	 unsigned len = strlen(value + 1);
	 *value++ = '\0';
	 /* Check for surrounding quotes. */
	 if (*value == '"' && value[len - 1] == '"')
	 {
	    /* Remove surrounding quotes. */
	    value[len - 1] = '\0';
	    value++;
	 }
	 if (sysparam_write(name, value) != 0)
	 {
	    printf("System parameter write failed.\n");
	 }
      }
      else
      {
	 goto syntax_error;
      }
   }
   else
   {
syntax_error:
      (void)eprintf("Syntax error. Usage: spw <name>=<value>\n");
      return RET_ERROR;
   }
   return RET_SUCCESS;
}


static int
cmd_sp_delete(int argc, char **argv)
{
   if (argc == 2)
   {
      const char *name = argv[1];
      if (sysparam_write(name, "") != 0)
      {
	 printf("System parameter delete failed.\n");
      }
   }
   else
   {
      (void)eprintf("Syntax error. Usage: spdel <name>\n");
      return RET_ERROR;
   }
   return RET_SUCCESS;
}


/*===========================================================================
 *                    getln
 *===========================================================================
 * Description:  Gets a string of max <size>, se below, form stdin.
 *
 * Parameters:
 *              buffer:   A buffer to store the string into.
 *              size:     Max size to read from stdin.
 * Returns:     The length of the line.
 * 
 */ 

static int
getln(char* buffer, int size)
{
   int i;
   for (i = 0; i < size; i++)
   {
      char ch = (char)fgetc(stdin);
      if (ch == '\n')
      {
	 buffer[i] = '\0';
	 /* Done, return length */
	 return i;
      }
      else
      {
	 buffer[i] = ch;
      }
   }
   
   /* If here, buffer was too small, throw away the rest of the line.*/
   eprintf("\n Error: Too long input ignored.\n");
   while(fgetc(stdin) != '\n')
   {
      /* Skip char */
   }
   return 0;
}

/*
 *===========================================================================
 *                    display_and_modify_sp
 *===========================================================================
 * Description:  
 *
 * Parameters: index, which boot method to ask for.  
 * Returns:      
 *
 */
static void
display_and_modify_sp(const char *name)
{
   char value[SYSPARAM_MAX_VALUE_LENGTH + 1];
   size_t value_len = sysparam_read(name, value, SYSPARAM_MAX_VALUE_LENGTH);
   if (value_len > 0)
   {
      int line_len;
      if (value_len >= SYSPARAM_MAX_VALUE_LENGTH) /* Value truncated */
      {
	 value[SYSPARAM_MAX_VALUE_LENGTH] = '\0';
	 printf("%s=\"%s\"... ? ",name, value); 
      }
      else
      {
	 printf("%s=\"%s\" ? ",name, value);
      }
      fflush(stdout);
      line_len = getln(value, SYSPARAM_MAX_VALUE_LENGTH);
      if (line_len > 0)
      {
	 int result = sysparam_write(name, value);
	 if (result != 0)
	 {
	    printf("System parameter write failed\n");
	 }
      }
   }
}

/*
 *===========================================================================
 *                    cmd_sp_modify
 *===========================================================================
 * Description:  Display and modify system parameters.
 *		The interactive user has to press <ENTER> to accept the
 *		existing value, and "value<ENTER>" to change the actual value
 *
 * Parameters:   
 * Returns:      RET_SUCCESS for success else RET_ERROR.
 */

static int
cmd_sp_modify(int argc, char **argv)
{
   int retval;
   if ((argc == 1) || (argc == 2))
   {
      const char *prefix = (argc == 1) ? "" : argv[1];
      size_t prefix_len = strlen(prefix);

      if (prefix_len <= SYSPARAM_MAX_NAME_LENGTH)
      {
	 char name[SYSPARAM_MAX_NAME_LENGTH + 1];
	 strcpy(name, prefix);
	 do
	 {
	    display_and_modify_sp(name);
	 } while ((sysparam_list_next(name, name, sizeof name) > 0)
		  && (strncmp(name, prefix, prefix_len) == 0));
	 retval = RET_SUCCESS; /* Done */
      }
      else
      {
	 eprintf("Error: parameter name \"%s\" too long (max size "
		 STRSTR(SYSPARAM_MAX_NAME_LENGTH)").\n", argv[1]);
	 retval = RET_ERROR;
      }
   }
   else
   {
      eprintf("Syntax error. Usage: spmod [<prefix>]\n");
      retval = RET_ERROR;
   }
   return retval;
}

static int
next_attr(char **pp, char **name, char **value)
{
   char *sep, *eop;

   while (**pp && isspace((unsigned char)**pp))
   {
      (*pp)++;
   }

   if ((sep = strchr(*pp, ':')) == NULL)
   {
      return -1;
   }

   *name = *pp;
   *sep = '\0';

   if ((eop = strchr(sep+1, ' ')) != NULL)
   {
      *eop = '\0';
      *pp = eop+1;
   }
   else
   {
      eop = strchr(sep+1, '\0');
      *pp = eop;
   }

   *value = sep+1;

   return 0;
}

static int
shellcmd_spedit(int argc, char **argv)
{
   const char *spname = argv[1];
   char *buf0;
   char *buf1 = NULL;
   char *buf2 = NULL;
   char *attr   = 0;
   char *value0 = 0;
   char *use_value = 0;
   char value1[64];
   char *p0;
   char *curr_buf = NULL;
   size_t n;
   int need_write = 0;
   int c;
   size_t total_size = 0;

   if (argc < 2)
   {
      printf("usage: spedit <name>\n");
      return 1;
   }

   n = sysparam_read(spname, NULL, 0);

   buf0 = malloc(n+1);

   n = sysparam_read(spname, buf0, n+1);
   if (n == 0)
   {
      printf("%s: No such parameter.\n", spname);
      return 1;
   } 

   if (strchr(buf0, ':') == NULL)
   {
     printf("Parameter not of the form 'name=attribute:value'. "
            "Please use spw instead.\n");
     return 1;
   }

   p0 = buf0;

   while (next_attr(&p0, &attr, &value0) != -1)
   {
      size_t valuelen;
      size_t attrlen;
      printf("%s [%s]: ", attr, value0);
      fflush(stdout);
      fgets(value1, sizeof value1, stdin);
      valuelen = strlen(value1);
      attrlen = strlen(attr);

      /* Check if new or old attribut shall be used */
      if (valuelen > 1)
      {
	 value1[valuelen - 1] = '\0'; /* Remove \n from buffer */
         valuelen--;
         use_value = value1;
         need_write = 1;
      }
      else
      {
         use_value = value0;
	 valuelen = strlen(value0);
      }

      /* Reallocate buffers */
      if (buf1 == NULL)
      {
         /* Allocate new buffer for total value size, +3 is one ':', one ' ' and one '\0' */
         buf1 = (char *)malloc(total_size + valuelen + attrlen + 3);
         
         if (buf2 != NULL)
         {
            memcpy(buf1, buf2, total_size);
            free((void *)buf2);
            buf2 = NULL;
         }

         curr_buf = buf1;
      }
      else
      {
         /* Allocate new buffer for total value size, +2 is one ':', one ' ' and one '\0' */
         buf2 = (char *)malloc(total_size + valuelen + attrlen + 3);
         memcpy(buf2, buf1, total_size);
         free((void *)buf1);
         buf1 = NULL;
         
         curr_buf = buf2;
      }

      /* Add the modified attribute to the list of attributes */
      sprintf(curr_buf + total_size, "%s:%s ", attr, use_value);

      /* +2 is one ':' and one ' ' */
      total_size += (valuelen + attrlen + 2);
   }
   
   if (need_write)
   {
      if (curr_buf[strlen(curr_buf)-1] == ' ')
      {
	 curr_buf[strlen(curr_buf)-1] = '\0';
      }
      printf("\n%s = %s\n", spname, curr_buf);
      printf("OK? [Y/n]: ");
      fflush(stdout);
      c = fgetc(stdin);
      if (c == '\n' || c == 'Y' || c == 'y')
      {
	 int err;
	
	 if ((err = sysparam_write(spname, curr_buf)) != 0)
	 {
	    printf("spedit: Failed to update parameter, error %d.\n", err);
	 }
      }
   }

   if (curr_buf != NULL)
   {
      free(curr_buf);
   }

   free(buf0);

   return 0;
}

void
sysparam_add_cmds(void)
{
/**
 * @command spr
 * @syntax  spr [<prefix>]
 * @brief   Reads a system parameter.
 * @long    This command reads the system parameters stored in flash.
 * @option  -b Show only parameters stored as board parameters.
 * @option  -l Show only parameters stored as Load Module-parameters.
 * @input   prefix  If given, it will display all parameters starting with
 *          'prefix'. If not given, all parameters will be displayed.
 * @return  Zero if successful, non-zero if failed.
 */
   shell_add_cmd("spr", "spr [-b | -l] [<prefix>]", "Read a system parameter.",
		 cmd_sp_read);

/**
 * @command spw
 * @syntax  spw <name> <value>
 * @brief   Writes a system parameter.
 * @long    This command writes a new value into a system parameter
 *          and stores this in flash.
 * @input   name   Name of parameter to change.
 * @input   value  Value to give the parameter.
 * @return  Zero if successful, non-zero if failed.
 */
   shell_add_cmd("spw", "spw <name>=<value>", "Write a system parameter.",
		 cmd_sp_write);
/**
 * @command spdel
 * @syntax  spdel <name>
 * @brief   Deletes a system parameter.
 * @input   name  Name of parameter to delete.
 * @return  Zero if successful, non-zero if failed.
 */
   shell_add_cmd("spdel", "spdel <name>", "Delete a system parameter.",
		 cmd_sp_delete);

/**
 * @command spmod
 * @syntax  spmod [<prefix>]
 * @brief   Stepwise view and modify system parameters.
 * @long    This command will walk you through all system parameters
 *          currently defined and give you a chance to change them.
 * @input   prefix  If given, will match all parameters starting with
 *                'prefix'.
 * @return  Zero if successful, non-zero if failed.
 */
   shell_add_cmd("spmod", "spmod [<prefix>]","Stepwise view and modify system "
		 "parameters.",
		 cmd_sp_modify);

/**
 * @command spedit
 * @syntax  spedit <name>
 * @brief   Interactive edit of system parameter attributes
 * @input   name  Name of the system parameter to edit.
 * @return  Zero if successful, non-zero if failed.
 */
   shell_add_cmd("spedit", "spedit <name>", "Edit system parameters interactively",
		  shellcmd_spedit);
}
