/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2007 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

#include "pthread_internal.h"
#include "cpu.h"
#include "bios.h"
#include "spinlock.h"

#define MILLI2MICRO 1000
#define MILLI2NANO 1000000
#define NONE2MILLI 1000

#ifdef OSE_SMP
/* Used on SMP when locking interrupts. */
static spinlock_t bpl_spinlock = SPIN_LOCK_INIT;
#endif

#if !defined(softOSE)
#define SET_INTERRUPT_LOCK(msr)                                 \
   do                                                           \
   {                                                            \
      PTHREAD_SMP_LOG(("Pthread %x attempting to grab BPL spinlock %s %d\n", current_process(), __FILE__, __LINE__));  \
      spin_lock_irqsave(&bpl_spinlock, msr);			\
      PTHREAD_SMP_LOG(("Pthread %x got BPL spinlock %s %d\n", current_process(), __FILE__, __LINE__));  \
   } while (0)

#define REMOVE_INTERRUPT_LOCK(msr)                              \
   do                                                           \
   {                                                            \
      PTHREAD_SMP_LOG(("Pthread %x unlocking BPL spinlock %s %d\n", current_process(), __FILE__, __LINE__));  \
      spin_unlock_irqrestore(&bpl_spinlock, msr);		\
   } while (0)

#else

/* No interrupt locking necessary on SFK. */
#define SET_INTERRUPT_LOCK(msr)					\
   do								\
   {								\
      msr = 0; /* Removes compiler warning. */			\
   } while (0)

#define REMOVE_INTERRUPT_LOCK(msr)				\
   do								\
   {								\
      (void)msr;						\
   } while (0)

#endif

struct single_link
{
   pthread_t tid;
   U32 next;
};

struct double_link
{
   pthread_t tid;
   U32 thread_state;
   U32 next;
   U32 prev;
};

/* xxose_pthread_cond_signal - signals the given condition variable
 * Signaling the mutex is done by unlinking the first element
 * in the list of waiting threads, and sending a signal to the
 * first waing thread.
 * Interrupts must be locked before entering this scope.
 */
static void
signal_condition_variable(pthread_cond_t *cond, U32 msr)
{
   struct double_link *first;
   struct thread_state *thread_state;
   struct double_link *newfirst;
   union SIGNAL *sig;
   PROCESS to_signal;

   first = (struct double_link *) cond->first;
   thread_state = (struct thread_state *) first->thread_state;
   newfirst = (struct double_link *) first->next;

   assert(thread_state != NULL);
   thread_state->flags &= ~THREAD_STATE_CONDWAITING;
   cond->nwaiters--;

   /* unlink the first waiter */
   cond->first = (U32) newfirst;

   if (newfirst == NULL) /* last waiter */
   {
      cond->mutex = 0;
      cond->last = 0;
   }
   else
   {
      newfirst->prev = 0;
      assert(cond->last != 0);
   }

   to_signal = (PROCESS) (first->tid);
   REMOVE_INTERRUPT_LOCK(msr);
   sig = alloc(sizeof (struct signal_cond_var), SIGNAL_COND_VAR);
   sig->signal_cond_var.cancel = False;
   send(&sig, to_signal);
}

/* lock_mutex - locks the given mutex. This is simply a wrapper for the
 * non-error calling version of the regular OSE mutex lock function. */
static int
lock_mutex(pthread_mutex_t *mutex, pthread_t thispth)
{
   (void)thispth;
   
   PTHREAD_SMP_LOG(("Pthread %x waiting for mutex %s %d\n", current_process(), __FILE__, __LINE__));
   {
      OSERRCODE rc = 0;
      OSADDRESS extra = 0;
      ose_mutex_lock_unsafe((MUTEX *)mutex, &rc, &extra);
      if (rc > 1)
      {
         return EINVAL;
      }
   }
   PTHREAD_SMP_LOG(("Pthread %x got the mutex %s %d\n", current_process(), __FILE__, __LINE__));

   return OSEPTH_STATUS_OK;   
}

/* unlock_mutex - unlocks the given mutex
 * The next mutex owner (if any) is waken up.
 */
static int
unlock_mutex(pthread_mutex_t *mutex)
{
   PTHREAD_SMP_LOG(("Pthread %x released mutex %s %d\n", current_process(), __FILE__, __LINE__));
   {
      OSERRCODE rc = 0;
      OSADDRESS extra = 0;
      ose_mutex_unlock_unsafe((MUTEX *)mutex, &rc, &extra);
      if (rc > 1)
      {
         return EINVAL;
      }
   }

   return OSEPTH_STATUS_OK;      
}

/* osepth_cond_broadcast - signals all the waiting threads of the condition
 * variable.
 * If the given condition variable is not valid, a non-zero error code
 * is returned. Otherwise, zero is returned.
 *
 */
static int
osepth_cond_broadcast(pthread_cond_t * cond)
{
   if (cond == NULL ||
       (cond->first == 0xFFFF &&
	cond->last == 0xFFFF &&
	cond->mutex == 0xFFFF))
   {
      return EINVAL;
   }
   else
   {
      U32 ntimes = cond->nwaiters;
      while (ntimes-- != 0)
      {
	 U32 msr;
	 SET_INTERRUPT_LOCK(msr);

	 if(cond->first != 0)
	 {
	   /* The first waiterer is still there. */
	    signal_condition_variable(cond, msr);
	 }
	 else
	 {
	   /* The first waiterer disappeared. */
	   REMOVE_INTERRUPT_LOCK(msr);
	   break;
	 }
      }

      return OSEPTH_STATUS_OK;
   }
}

/* osepth_cond_signal - signals the given condition variable
 * If signalling is possible, xxose_pthread_cond_signal is called, and
 * zero is returned.
 * Otherwise, a non-zero error code is returned.
 */
static int
osepth_cond_signal(pthread_cond_t * cond)
{
   if (cond == NULL ||
       (cond->first == 0xFFFF &&
	cond->last == 0xFFFF &&
	cond->mutex == 0xFFFF))
   {
      return EINVAL;
   }
   else /* cond is ok */
   {
      U32 msr;

      SET_INTERRUPT_LOCK(msr);

      if (cond->first != 0)
      {
	 signal_condition_variable(cond, msr);
      }
      else
      {
	 REMOVE_INTERRUPT_LOCK(msr);
      }

      return OSEPTH_STATUS_OK;
   }
}

/* cond_waitrel - waits for the given condition variable with a deadline
 * If waiting for the given condition variable is not possible, a non-zero
 * error code is returned.
 * If the given mutex is used when locking the condition varible.
 * If the current pthread was in cancelled state, it exits by calling
 * osepth_end_thread.
 * If waiting was possible, the thread starts receiving a signal.
 * When a signal arrives, it can indicate either that the thread was
 * cancelled, or that the condition variable was signalled.
 * If no signal is recieved because of a receive_w_tmo timout, a non-zero
 * error code is returned.
 * Observe that this implementation is essentially the same as for
 * osepth_cond_wait.
 */
static int
cond_waitrel(pthread_cond_t * cond,
	     pthread_mutex_t * mutex,
	     unsigned long reltime)
{
   struct thread_state *thread_state =  osepth_private_thread_state();
   PROCESS thispth = current_process();

   if (cond == NULL ||
       (cond->first == 0xFFFF &&
	cond->last == 0xFFFF &&
	cond->mutex == 0xFFFF))
   {
      return EINVAL;
   }
   else if (cond->mutex != 0 &&	/* mutex error */
	    (pthread_mutex_t *) cond->mutex != mutex)
   {
      return EINVAL;
   }
   else if (cond->mutex != 0 &&	/* not mutex owner */
	    ((((MUTEX_INT_TYPE *) cond->mutex)->owner ^ thispth) << 1) != 0)
   {
      return EINVAL;
   }
   else if (cond->mutex == 0 && /* check if owner */
	    ((((MUTEX_INT_TYPE *) mutex)->owner ^ thispth) << 1) != 0)
   {
      return EINVAL;
   }
   else /* possible to wait */
   {
      BPL_LOCK();

      if (thread_state->flags & THREAD_STATE_CANCELED) /* we were canceled */
      {
         BPL_UNLOCK();
	 return OSEPTH_END_THREAD;
      }
      else /* add to waiterers list */
      {
	 union SIGNAL *sig;
	 struct double_link newlink;
	 static const SIGSELECT condsig[] = {1, SIGNAL_COND_VAR};
         int rc;

	 cond->mutex = (U32) mutex; /* always correct */

	 /* add this thread as waiterer */
	 cond->nwaiters++;
	 newlink.tid = thispth;
	 newlink.thread_state = (U32) thread_state;

	 if (cond->first == 0)	/* no condwaiters */
	 {
	    assert(cond->last == 0);
	    cond->first = (U32) & newlink;
	    cond->last = (U32) & newlink;
	    newlink.prev = 0;
	    newlink.next = 0;
	 }
	 else /* add as last */
	 {
	    struct double_link *oldlast = (struct double_link *) cond->last;
	    assert(cond->last != 0);
	    newlink.prev = (U32) oldlast;
	    newlink.next = 0;
	    oldlast->next = (U32) & newlink;
	    cond->last = (U32) & newlink;
	 }

	 /* add cond.var info to thread_state */
	 thread_state->cond = (U32) cond;
	 thread_state->condlink = (U32) & newlink;
	 thread_state->flags |= THREAD_STATE_CONDWAITING;

	 /* unlock mutex */
	 rc = unlock_mutex(mutex);
         if (rc != OSEPTH_STATUS_OK)
         {
	    BPL_UNLOCK();
            return rc;
         }
         
         /* Release the lock while we wait for signals. */
         BPL_UNLOCK();
         
	 sig = receive_w_tmo((OSTIME) reltime, (const SIGSELECT *) condsig);
	 assert(sig == NULL || \
		!(thread_state->flags & THREAD_STATE_CONDWAITING));

        /* If someone has signaled the condition variable, or canceled
         * this thread, they must have zeroed the PTHCB_CONDWAITIN flag
         * and are about to send a signal. However, while they're sending
         * the signal, this process has timed out in its receive_w_tmo.
         *
         * Since we "know" that an  OSE signal is on its way, we change to
         * a normal receive and recieve the signal on its way to us.
         */
         if (sig == NIL && !(thread_state->flags & THREAD_STATE_CONDWAITING))
         {
            sig = receive(condsig); 
         }
         
	 /* If sig is NULL, we have to unlink ourselves from the
	  * waiterers list in the condition variable.
	  * Since this is done with process priority xxpth_prio_max,
	  * we will not be interrupted.
	  */
	 if (sig == NIL) /* have to unlink ourselves from the list */
	 {
            /* Grab the BPL again. */
            BPL_LOCK();
            
            assert(cond->nwaiters != 0);
            thread_state->flags &= ~THREAD_STATE_CONDWAITING;            
	    cond->nwaiters--;

	    if (newlink.prev == 0) /* newlink is first */
	    {
	       struct double_link *newfirst = \
		  (struct double_link *) newlink.next;

	       cond->first = (U32) newfirst;

	       if (newfirst == NULL) /* last waiter */
	       {
		  cond->mutex = 0;
		  cond->last = 0;
	       }
	       else /* more waiterers */
	       {
		  newfirst->prev = 0;
		  assert(cond->last != 0);
	       }
	    }
	    else /* not first */
	    {
	       struct double_link *next = (struct double_link *) newlink.next;
	       struct double_link *prev = (struct double_link *) newlink.prev;

	       assert(cond->first != 0);
	       assert(prev != NULL);
	       prev->next = (U32) next;

	       if (next == NULL)
	       {
		  assert(cond->last == (U32) & newlink);
		  cond->last = (U32) prev;
	       }
	       else
	       {
		  next->prev = (U32) prev;
	       }
	    }

            /* We don't want to keep the BPL while waiting for a mutex. */
	    BPL_UNLOCK();
	    rc = lock_mutex(mutex, thispth);
            if (rc != OSEPTH_STATUS_OK)
            {
               return rc;
            }
            
	    return ETIMEDOUT;
	 }

	 lock_mutex(mutex, thispth);
         if (rc != OSEPTH_STATUS_OK)
         {
            return rc;
         }

	 if (sig->signal_cond_var.cancel) /* we were canceled */
	 {
	    free_buf(&sig);

	    return OSEPTH_END_THREAD;
	 }
	 else /* simple signaling */
	 {
	    free_buf(&sig);

	    return OSEPTH_STATUS_OK;
	 }
      }
   }
}

/* osepth_cond_timedwait - wait on the given condition varible with a timeout
 * If the timeout has already occured, a non-zero error code is returned.
 * Otherwise, cond_waitrel is called, and its returnvalue is returned.
 */
static int
osepth_cond_timedwait(pthread_cond_t * cond,
		 pthread_mutex_t * mutex,
		 const struct timespec *abstime)
{
   if (abstime == NULL)
   {
      return EFAULT;
   }
   else
   {
      long abs_millis;
      time_t now_sec;
      time_t abs_sec;
      long now_millis;
      long secdiff;
      struct TimePair now;

      rtc_get_time(&now);
      now_sec = now.seconds;
      now_millis = now.micros / MILLI2MICRO; /* convert to milliseconds */
      abs_sec = abstime->tv_sec;
      abs_millis = abstime->tv_nsec / MILLI2NANO; /* convert to milliseconds */
      secdiff = abs_sec - now_sec; /* possitive or zero */

      if (secdiff < 0 || (secdiff == 0 && abs_millis <= now_millis))
      {
	 return ETIMEDOUT;
      }
      else /* abs_sec >= now_sec */
      {
	 unsigned long reltime;	/* resulting timeout time in milli */

	 if (secdiff == 0) /* know that abs_millis > now_millis */
	 {
	    reltime = abs_millis - now_millis;
	 }
	 else
	 {
	    reltime = NONE2MILLI * secdiff + abs_millis - now_millis;
	 }

	 return cond_waitrel(cond, mutex, reltime);
      }
   }
}

/* osepth_cond_wait - waits for the given condition variable
 * If waiting for the given condition variable is not possible, a non-zero
 * error code is returned.
 * If the given mutex is used when locking the condition varible.
 * If the current pthread was in cancelled state, it exits by calling
 * osepth_end_thread.
 * If waiting was possible, the thread starts receiving a signal.
 * When a signal arrives, it can indicate either that the thread was
 * cancelled, or that the condition variable was signalled.
 */
static int
osepth_cond_wait(pthread_cond_t * cond,
	    pthread_mutex_t * mutex)
{
   PROCESS thispth = current_process();
   struct thread_state *thread_state =  osepth_private_thread_state();

   if (cond == NULL ||		/* not initialized */
       (cond->first == 0xFFFF &&
	cond->last == 0xFFFF &&
	cond->mutex == 0xFFFF))
   {
      return EINVAL;
   }
   else if (mutex == NULL)
   {
      return EINVAL;
   }
   else if (cond->mutex != 0 &&	/* mutex error */
	    (pthread_mutex_t *) cond->mutex != mutex)
   {
      return EINVAL;
   }
   else if (cond->mutex != 0 && /* check if owner */
	    ((((MUTEX_INT_TYPE *) cond->mutex)->owner ^ thispth) << 1) != 0)
   {
      return EINVAL;
   }
   else if (cond->mutex == 0 && /* check if owner */
	    ((((MUTEX_INT_TYPE *) mutex)->owner ^ thispth) << 1) != 0)
   {
      return EINVAL;
   }
   else /* possible to wait */
   {
      BPL_LOCK();

      if (thread_state->flags & THREAD_STATE_CANCELED)	/* we were canceled */
      {
         BPL_UNLOCK();
	 return OSEPTH_END_THREAD;
      }
      else /* add to waiterers list */
      {
	 union SIGNAL *sig;
	 struct double_link newlink;
	 static const SIGSELECT condsig[] = {1, SIGNAL_COND_VAR};
         int rc;

	 cond->mutex = (U32) mutex;	/* always correct */

	 /* add this as waiter */
	 cond->nwaiters++;
	 newlink.tid = thispth;
	 newlink.thread_state = (U32) thread_state;

	 if (cond->first == 0)	/* no condwaiters */
	 {
	    assert(cond->last == 0);
	    cond->first = (U32) & newlink;
	    cond->last = (U32) & newlink;
	    newlink.prev = 0;
	    newlink.next = 0;
	 }
	 else /* add as last */
	 {
	    struct double_link *oldlast = (struct double_link *) cond->last;
	    assert(cond->last != 0);
	    newlink.prev = (U32) oldlast;
	    newlink.next = 0;
	    oldlast->next = (U32) & newlink;
	    cond->last = (U32) & newlink;
	 }

	 /* add cond.var info to thread_state */
	 thread_state->cond = (U32) cond;
	 thread_state->condlink = (U32) & newlink;
	 thread_state->flags |= THREAD_STATE_CONDWAITING;

	 /* unlock mutex */
	 rc = unlock_mutex(mutex);
         if (rc != OSEPTH_STATUS_OK)
         {
	    BPL_UNLOCK();
            return rc;
         }

         /* Release the BPL while in receive. */
         BPL_UNLOCK();

	 /* Wait until signaled or canceled */
	 sig = receive((const SIGSELECT *) condsig);
         
	 /* restore thread_state */
	 assert(!(thread_state->flags & THREAD_STATE_CONDWAITING));
	 rc = lock_mutex(mutex, thispth);
         if (rc != OSEPTH_STATUS_OK)
         {
            return rc;
         }
         
	 if (sig->signal_cond_var.cancel)	/* we were canceled */
	 {
	    free_buf(&sig);

	    return OSEPTH_END_THREAD;
	 }
	 else /* simple signaling */
	 {
	    free_buf(&sig);

	    return OSEPTH_STATUS_OK;
	 }
      }
   }
}

/* osepth_mutex_lock - locks the given mutex for the current thread
 * If possible, the mutex is locked by calling xxose_pthread_mutex_lock,
 * and returning zero.
 * Otherwise, an error code is returned.
 */
static int
osepth_mutex_lock(pthread_mutex_t *m)
{
   pthread_t thispth = (pthread_t) current_process();
   MUTEX_INT_TYPE *mutex = (MUTEX_INT_TYPE *)m;

   if (mutex == NULL ||
       mutex->magic == MAGIC_MUTEX_DESTROYED)
   {
      return EINVAL;
   }
   else if (((mutex->owner ^ (PROCESS) thispth) << 1) == 0)
   {
      return EDEADLK;
   }
   else
   {
      OSERRCODE rc = 0;
      OSADDRESS extra = 0;
      ose_mutex_lock_unsafe((MUTEX *)m, &rc, &extra);
      if (rc > 1)
      {
         return EINVAL;
      }
      return OSEPTH_STATUS_OK;
   }   
}

/* osepth_mutex_unlock - unlocks the given mutex
 * If the unlocking could not be performed, an error code
 * is returned. On successful unlocking, zero is returned.
 */
static int
osepth_mutex_unlock(pthread_mutex_t *m)
{
   MUTEX_INT_TYPE *mutex = (MUTEX_INT_TYPE *)m;

   if (mutex == NULL ||
       mutex->magic == MAGIC_MUTEX_DESTROYED)
   {
      return EINVAL;
   }
   else if (((mutex->owner ^ current_process()) << 1) != 0)
   {
      return EPERM;
   }
   else
   {
      OSERRCODE rc = 0;
      OSADDRESS extra = 0;
      ose_mutex_unlock_unsafe((MUTEX *)mutex, &rc, &extra);
      if (rc > 1)
      {
         return EINVAL;
      }
      return OSEPTH_STATUS_OK;
   }
}

static int
osepth_mutex_trylock(pthread_mutex_t *m)
{
   pthread_t thispth = (pthread_t) current_process();
   MUTEX_INT_TYPE *mutex = (MUTEX_INT_TYPE *)m;

   if (mutex == NULL ||
       mutex->magic == MAGIC_MUTEX_DESTROYED)
   {
      return EINVAL;
   }
   if (((mutex->owner ^ (PROCESS) thispth) << 1) == 0)
   {
      return EDEADLK;
   }
   else
   {
      OSERRCODE rc = 0;
      OSADDRESS extra = 0;
      int ret = ose_mutex_trylock_unsafe((MUTEX *)m, &rc, &extra);
      
      if (ret == 0)
      {
         return EBUSY;
      }
      else
      {
         if (rc > 1)
         {
            return rc;
         }
         else
         {
            return OSEPTH_STATUS_OK;
         }
      }
   }
}

/* osepth_pthread_cancel - cancal the given thread
 * If the thread to cancel is not cancelable, zero is returned. For the
 * following cases, assume that the thread to cancel is cancelable.
 * If the calling thread is the same as the given thread, the canceled flag
 * is set in the thread_state.
 * If the thread thread to cancel is running, but is executing a cancellation
 * point (joining another thread or waiting for a condition variable), the
 * canceled flag in the thread_state is set, and zero is returned.
 * If the thread to cancel is joining another thread, the canceled flag of
 * its thread_state is set, and its fast semaphore is signaled.
 * If the thread to cancel is waiting on a condition variable, the thread to
 * cancel is removed from the condition variable, and sent a signal
 * with information that it has been canceled (as opposed to waken up by
 * signaling of the condition variable).
 * Observe that only deferred cancellation is implemented.
 */
static int
osepth_pthread_cancel(pthread_t tocancel)
{
   struct thread_state *thread_state = osepth_other_thread_state(tocancel);
   PROCESS thisproc = current_process();

   BPL_LOCK();

   if (thread_state == NULL)
   {
      BPL_UNLOCK();

      return ESRCH;
   }
   else if ((((PROCESS) tocancel ^ thisproc) << 1) == 0)	/* self cancel */
   {
      if (thread_state->flags & THREAD_STATE_CANCELABLE)
      {
	 thread_state->flags |= THREAD_STATE_CANCELED;
      }

      BPL_UNLOCK();

      return OSEPTH_STATUS_OK;
   }
   else
   {
      U32 flags = thread_state->flags;

      if (flags & THREAD_STATE_CANCELABLE && /* cancelation possible */
	  flags & THREAD_STATE_RUNNING)
      {
	 thread_state->flags |= THREAD_STATE_CANCELED;	/* cancel marked */

	 if (flags & THREAD_STATE_JOINING) /* tocancel is joining another */
	 {
	    thread_state->flags &= ~THREAD_STATE_JOINING;
	    signal_fsem((PROCESS) tocancel);
	    BPL_UNLOCK();
	 }
	 else if (flags & THREAD_STATE_CONDWAITING) /* waiting on cond var */
	 {
	    union SIGNAL *sig;
	    pthread_cond_t *cond = (pthread_cond_t *) thread_state->cond;
	    struct double_link *condlink = \
	       (struct double_link *) thread_state->condlink;

	    assert(cond != NULL);
	    assert(condlink != NULL);
	    assert(thread_state != NULL);
	    thread_state->flags &= ~THREAD_STATE_CONDWAITING;
	    thread_state->cond = 0;
	    thread_state->condlink = 0;
	    cond->nwaiters--;

	    /* unlink condlink from list */
	    if (condlink->prev == 0) /* condlink is first */
	    {
	       struct double_link *newfirst = \
		  (struct double_link *) condlink->next;

	       cond->first = (U32) newfirst;
               
	       if (newfirst == NULL)	/* last waiter */
	       {
		  cond->mutex = 0;
		  cond->last = 0;
	       }
	       else /* more waiterers */
	       {
		  newfirst->prev = 0;
		  assert(cond->last != 0);
	       }
	    }
	    else /* not first */
	    {
	       struct double_link *next = \
		  (struct double_link *) condlink->next;
	       struct double_link *prev = \
		  (struct double_link *) condlink->prev;

	       assert(cond->first != 0);
	       assert(prev != NULL);
	       prev->next = (U32) next;

	       if (next == NULL)
	       {
		  assert(cond->last == (U32) condlink);
		  cond->last = (U32) prev;
	       }
	       else
	       {
		  next->prev = (U32) prev;
	       }
	    }

	    BPL_UNLOCK();
	    sig = alloc(sizeof (struct signal_cond_var), SIGNAL_COND_VAR);
	    sig->signal_cond_var.cancel = True;
	    send(&sig, tocancel);
	 }
	 else /* just set the THREAD_STATE_CANCELED, which is already done */
	 {
	    BPL_UNLOCK();
	 }
      }
      else /* not cancelable */
      {
	 BPL_UNLOCK();
	 assert(!(flags & THREAD_STATE_CANCELED));
      }

      return OSEPTH_STATUS_OK;
   }
}

/* osepth_setcancelstate - sets cancellation state of current thread
 * The second parameter oldstat will contain the old state.
 */
static int
osepth_setcancelstate(int state,
		 int *oldstate)
{
   struct thread_state *thread_state =  osepth_private_thread_state();

   if (thread_state == NULL)
   {
      OSEPTH_ERROR2(OSE_PTH_NOPTHREAD, current_process());

      return 0; /* NOTREACHED */
   }
   else
   {
      U32 flags = thread_state->flags;

      if (state == PTHREAD_CANCEL_ENABLE)
      {
	 U32 msr;

	 SET_INTERRUPT_LOCK(msr);

	 if (oldstate != NULL)
	 {
	    *oldstate = (int) (flags & THREAD_STATE_CANCELABLE) ?
	       PTHREAD_CANCEL_ENABLE : PTHREAD_CANCEL_DISABLE;
	 }

	 if (!(flags & THREAD_STATE_CANCELED))
	 {
	    thread_state->flags |= THREAD_STATE_CANCELABLE;
	 }

	 REMOVE_INTERRUPT_LOCK(msr);

	 return OSEPTH_STATUS_OK;
      }
      else if (state == PTHREAD_CANCEL_DISABLE)
      {
	 U32 msr;

	 SET_INTERRUPT_LOCK(msr);

	 if (oldstate != NULL)
	 {
	    *oldstate = (flags & THREAD_STATE_CANCELABLE) ?
	       PTHREAD_CANCEL_ENABLE : PTHREAD_CANCEL_DISABLE;
	 }

	 if (!(flags & THREAD_STATE_CANCELED))
	 {
	    thread_state->flags &= ~THREAD_STATE_CANCELABLE;
	 }

	 REMOVE_INTERRUPT_LOCK(msr);

	 return OSEPTH_STATUS_OK;
      }
      else
      {
	 return EINVAL;
      }
   }
}

long
osepth_trap_handler(long fc,
                    long ecode,
                    long arg1,
                    long arg2,
                    long arg3)
{
   long *p = (long *) ecode;

   switch (fc)
   {
      case OSEPTH_COND_BROADCAST:
      {
	 *p = osepth_cond_broadcast((pthread_cond_t *) arg1);

	 return 0;
      }
      case OSEPTH_COND_SIGNAL:
      {
	 *p = osepth_cond_signal((pthread_cond_t *) arg1);

	 return 0;
      }
      case OSEPTH_COND_TIMEDWAIT:
      {
	 *p = osepth_cond_timedwait((pthread_cond_t *) arg1,
				    (pthread_mutex_t *) arg2,
				    (const struct timespec *) arg3);

	 return *p; /* used by the flib */
      }
      case OSEPTH_COND_WAIT:
      {
	 *p = osepth_cond_wait((pthread_cond_t *) arg1,
			       (pthread_mutex_t *) arg2);

	 return *p; /* used by the flib */
      }
      case OSEPTH_MUTEX_LOCK:
      {
	 *p = osepth_mutex_lock((pthread_mutex_t *) arg1);

	 return 0;
      }
      case OSEPTH_MUTEX_TRYLOCK:
      {
	 *p = osepth_mutex_trylock((pthread_mutex_t *) arg1);

	 return 0;
      }
      case OSEPTH_MUTEX_UNLOCK:
      {
	 *p = osepth_mutex_unlock((pthread_mutex_t *) arg1);

	 return 0;
      }
      case OSEPTH_SETCANCELSTATE:
      {
	 *p = osepth_setcancelstate((int) arg1, (int *) arg2);

	 return 0;
      }
      case OSEPTH_PTHREAD_CANCEL:
      {
	 *p = osepth_pthread_cancel((pthread_t) arg1);

	 return 0;
      }
      default:
      {
	 OSEPTH_ERROR2(OSE_PTH_EUNKNOWN_FCODE,
                             (OSERRCODE) fc);
	 kill_proc(current_process());
      }
   }

   return -1;
}

/* Called from _zzmain */
void ose_init_pthreads(void);

void ose_init_pthreads(void)
{
   long err;
   
   err = biosInstall(OSE_PTHREAD_BIOS_FLIB,
                     *(BiosFunction *) osepth_trap_handler,
                     BIOS_USERMODE_ALLOWED);
   
   if (err != BIOS_SUCCESS)
   {
      OSEPTH_ERROR2(OSE_PTH_EBIOS_INSTALL_FAILED,
                    (OSERRCODE) err);
   }
}
