/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2009 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */




/**
 * @file ramdump.c
 *
 * @brief This file contains the ramdump utility that is used to store
 * information in RAM while in freeze mode.
 *
 */

#include "string.h"

#include "ose.h"
#include "osetypes.h"
#include "stdint.h"

#if defined(DEBUG)
#  define ASSERT
#endif

#include "assert.h"
#include "ose_err.h"
#include "bios.h"
#include "efmi.h"
#include "sysparam.h"

#ifdef RAMDUMP_USE_ZLIB
/* Compression utils */
#include "zlib.h"
#else
#define Z_OK 0
typedef int z_stream;
#endif

#include "ramlog.h"

#include "ramdump.h"
#include "ramdump_format.h"
#include "ramdump_err.h"
#include "ramdump_internal.h"
#include "ramdump_heap.h"
#include "ramdump_private.h"

/* LOCK_SAVE & LOCK_RESTORE macros */
#include "cpu.h"

/* Need ntohl and htonl */
#include "netinet/in.h"

#ifdef softOSE
#define RAMDUMP_DO_NOT_SWITCH_STACK
#endif

#ifndef RAMDUMP_DO_NOT_SWITCH_STACK
/**
 * @macro RAMDUMP_STACK_SIZE
 *
 * @brief Size of the internal stack.
 */
#define RAMDUMP_STACK_SIZE 2048
#endif

/**
 * @macro RAMDUMP_PERSISTENT_DATA_SIZE
 *
 * @brief Amount of persistent data in dump area, cannot be changed, ever.
 *
 * @seealso
 */
#define RAMDUMP_PERSISTENT_DATA_SIZE	((U32) 0x500)

/**
 * @macro RAMDUMP_MIN_WORK_AREA_LENGTH
 *
 * @brief Minimum size of the ramdump work area.
 *
 * @seealso
 */
#define RAMDUMP_MIN_WORK_AREA_LENGTH	((U32) 0x10000) /* 64 KB */

/* Number of entries in the dump table */
/**
 * @macro DUMP_TABLE_SIZE
 *
 * @brief Number of entries in the dump table.
 */
#define DUMP_TABLE_SIZE         (64)

/**
 * @macro MAX_IFF_DEPTH
 *
 * @brief Max supported depth of the dump IFF structure.
 */
#define MAX_IFF_DEPTH           (9)

/**
 * @macro NO_DUMP_POS
 *
 * @brief Definition of an invalid dump position.
 */
#define NO_DUMP_POS             (0xffffffffUL)

/**
 * @macro DUMP_ALIGNMENT
 *
 * @brief Memory alignment of the dump area.
 */
#define DUMP_ALIGNMENT          (64UL)

#if !(defined(OSPP) || defined(OSARM) || defined(OSMIPS) || defined(OSX86) || defined(softOSE))
#  error "undefined or unrecognized architecture"
#endif

/* ZLIB compression */
#ifdef RAMDUMP_USE_ZLIB
/**
 * @macro ZLIB_COMPR_OUT_SIZE
 *
 * @brief Size of zlib compression output buffer.
 */
#  define ZLIB_COMPR_OUT_SIZE   (0x1000UL) /* 4K */

/**
 * @macro ZLIB_COMPR_IN_SIZE
 *
 * @brief Size of zlib compression input buffer
 */
#  define ZLIB_COMPR_IN_SIZE    (0x4000UL) /* 16K */
#endif


#ifdef OSE_SMP
#include "ose_smp.h"
#include "cpu_smp.h"		/* FIXME need OSE_SMP_MAX_CPUS */
#include "spinlock.h"

#define SMP_ARRAY [OSE_SMP_MAX_CPUS]
#define SMP_INDEX [zzkrn_cpu_id()]
#define SMP_INDEX_X(x) [x]
#define NOT_SMP_CODE(x)
#define SMP_CODE(x) x
#define SMP_CODE_YES_NO(x,y) x
#else
#define SMP_ARRAY
#define SMP_INDEX
#define SMP_INDEX_X(x)
#define NOT_SMP_CODE(x) x
#define SMP_CODE(x)
#define SMP_CODE_YES_NO(x,y) y
#endif

/*
 * Locks, we must disable interrupts while performing dumps.
 */

 /**
 * @data ramdump_msr
 *
 * @brief Machine status register
 *
 * @long
 *
 * @seealso
 */
static U32 ramdump_msr SMP_ARRAY;
SMP_CODE(static spinlock_t ramdump_spinlock = SPIN_LOCK_INIT;)

/**
 * @macro RAMDUMP_LOCK_PUSH
 *
 * @brief

 * @return
 */
#define RAMDUMP_LOCK_PUSH() do { \
	LOCK_SAVE(ramdump_msr SMP_INDEX); \
	SMP_CODE(spin_lock(&ramdump_spinlock);) \
} while(0)

/**
 * @macro RAMDUMP_LOCK_POP
 *
 * @brief
 *
 * @return
 */
#define RAMDUMP_LOCK_POP() do { \
	SMP_CODE(spin_unlock(&ramdump_spinlock);) \
	LOCK_RESTORE(ramdump_msr SMP_INDEX); } while(0)

struct FmmSignal
{
   U32 size;
   union
   {
      U32 signo;
   } data;
};

/**
 * @data dump_locked
 *
 * @brief Indicates if the dump is locked or not.
 *
 * @long Indicates if the dump is locked or not. Forward declaration.
 *
 * @seealso dump_lock dump_unlock
 */
static int dump_locked;

/**
 * @function dump_lock
 *
 * @brief Lock ramdump to prevent recursive calls to dump functions.
 *
 * @input block_interrupts Non-zero indicates that interrupts should
 * be blocked.
 *
 * @return Returns non-zero on success.
 *
 * @seealso dump_unlock
 */
static int
dump_lock(int block_interrupts)
{
   RAMDUMP_LOCK_PUSH();

   assert(block_interrupts == 0 || dump_locked == 0);

   /* Recursive locking? */
   if (dump_locked == 0)
   {

      /* Get dump lock */
      dump_locked++;

      /* Release interrupt block, if not handled */
      if (!block_interrupts)
      {
         RAMDUMP_LOCK_POP();
      }
      return 1;
   }
   else
   {
      /* Failed to get dump lock */
      RAMDUMP_LOCK_POP();
      return 0;
   }
}

/**
 * @function dump_unlock
 *
 * @brief Unlock ramdump, assuming it was previously locked by dump_lock.
 *
 * @input block_interrupts Non-zero indicates that interrupts should
 * be unblocked.
 *
 * @seealso dump_lock
 */
static void
dump_unlock(int block_interrupts)
{
   /* Block interrupts when checking lock */
   if (!block_interrupts)
   {
      RAMDUMP_LOCK_PUSH();
   }

   assert(dump_locked == 1);
   dump_locked--;

   RAMDUMP_LOCK_POP();
}

/**
 * @macro INTERRUPTS_BLOCKED
 *
 * @brief Indicates that interrupts should be blocked.
 */
#define INTERRUPTS_BLOCKED (1)

/**
 * @macro INTERRUPS_ENABLED
 *
 * @brief Indicates that interrupts should be enabled.
 */
#define INTERRUPTS_ENABLED (0)

/**
 * @macro DUMP_LOCK
 *
 * @brief
 *
 * @long
 *
 * @input block_interrupts Indicates that interrupts should be blocked.
 *
 * @return
 *
 * @seealso DUMP_UNLOCK
 */
#define DUMP_LOCK(block_interrupts) dump_lock(block_interrupts)

/**
 * @macro DUMP_UNLOCK
 *
 * @brief
 *
 * @long
 *
 * @input block_interrupts Indicates that interrupts should be unblocked.
 *
 * @return
 *
 * @seealso DUMP_LOCK
 */
#define DUMP_UNLOCK(block_interrupts) dump_unlock(block_interrupts)

/**
 * @macro ALIGNED
 *
 * @brief Checks if an address is aligned to a given size.
 *
 * @long Checks if an address is aligned to a given size; size being an
 * even power of two.
 *
 * @input addr Address to check alignment for.
 * @input size Alignment size to check address for.
 *
 * @return Returns non-zero if address is aligned to size.
 *
 * @seealso
 *
 * @example
 *
 */
#define ALIGNED(addr, size) (((unsigned long) (addr) &\
			                    (unsigned long) ((size)-1)) == 0)

/**
 * @macro U32_ALIGNED
 *
 * @brief Checks if an address is U32-aligned.
 *
 * @input addr Address to check for U32 alignment.
 *
 * @return Non-zero if address is aligned.
 */
#define U32_ALIGNED(addr)   ALIGNED(addr, sizeof(U32))

/**
 * @macro ALIGN_TO
 *
 * @brief Aligns value to the specified size.
 *
 * @long Aligns value to the specified size.
 *
 * @input value The value to align.
 * @input size The size to align value to.
 *
 * @return Returns the size aligned value.
 */
#define ALIGN_TO(value, size) (((value) + (size) - 1UL) & ~((size) - 1UL))

/**
 * @macro NEXT_ALIGNED_DUMP_POS
 *
 * @brief Aligns specified position to the next properly aligned
 * dump position.
 *
 * @long Aligns the specified position to the next properly aligned
 * dump position.
 *
 * @input pos The dump position to align.
 *
 * @return Returns the next properly aligned dump position.
 */
#define NEXT_ALIGNED_DUMP_POS(pos) DUMP_POS(ALIGN_TO((pos), DUMP_ALIGNMENT), 0)

/*
 * Macros that concern variables in the dump area
 */

/**
 * @macro DUMP_AREA_P
 *
 * @brief Pointer to the dump area.
 */
#define DUMP_AREA_P		(ramdump_dump_area)

/**
 * @macro DUMP_AREA_SIZE
 *
 * @brief Length of the dump area.
 */
#define DUMP_AREA_SIZE          (ramdump_dump_area_length)

/**
 * @macro RAMDUMP_PERS_DATA
 *
 * @brief Pointer to the persistent data area.
 */
#define RAMDUMP_PERS_DATA		(ramdump_area->pers_data)

/**
 * @macro CURR_DUMP_ID
 *
 * @brief Current dump id.
 */
#define CURR_DUMP_ID            (RAMDUMP_PERS_DATA.vars.curr_dump.seq_id)

/**
 * @macro CURR_DUMP_BLOCK_COUNT
 *
 * @brief Number of blocks in current dump.
 */
#define CURR_DUMP_BLOCK_COUNT   (RAMDUMP_PERS_DATA.vars.curr_dump.block_count)

/**
 * @macro DUMP_TABLE
 *
 * @brief Table of dump pointers.
 */
#define DUMP_TABLE              (RAMDUMP_PERS_DATA.vars.dumps)

/**
 * @macro HASH_DUMP_ID
 *
 * @brief Hash function for dump table index based on dump id.
 *
 * @input dump_id Dump identifier to compute index from.
 *
 * @return Returns dump table index for a particular dump.
 */
#define HASH_DUMP_ID(dump_id)	((dump_id) % DUMP_TABLE_SIZE)

/*
 * Macros for handling access to dump area
 */

/**
 * @macro RAMDUMP_FALSE
 *
 * @brief Represents boolean false in some functions.
 */
#define RAMDUMP_FALSE 0

/**
 * @macro RAMDUMP_TRUE
 *
 * @brief Represents boolean true in some functions.
 */
#define RAMDUMP_TRUE 1

/**
 * @macro VALID_DUMP_POS
 *
 * @brief Checks if dump position is valid.
 *
 * @long Checks if dump position is valid, i.e. in the
 * interval [0, DUMP_AREA_SIZE), need not test (U32) pos >= 0
 *
 * @input pos The dump position to check for validity.
 *
 * @return Non-zero if the dump position is valid.
 *
 * @seealso DUMP_POS DUMP_OFFSET
 */
#define VALID_DUMP_POS(pos) ((pos) < DUMP_AREA_SIZE)


/**
 * @macro DUMP_POS
 *
 * @brief
 *
 * @input dump_pos
 * @input offset
 *
 * @return
 *
 * @seealso VALID_DUMP_POS DUMP_OFFSET
 */
#  define DUMP_POS(dump_pos, offset) (((dump_pos) + (offset)) %\
                                      DUMP_AREA_SIZE)

/**
 * @macro DUMP_OFFSET
 *
 * @brief
 *
 * @input offset_pos
 * @input start_pos
 *
 * @return
 *
 * @seealso DUMP_POS VALID_DUMP_POS
 */
#  define DUMP_OFFSET(offset_pos, start_pos) \
    (((offset_pos) + DUMP_AREA_SIZE - (start_pos)) % DUMP_AREA_SIZE)

/**
 * @macro DUMP_VAL_U32
 *
 * @brief
 *
 * @input dump_pos
 * @input offset
 *
 * @return
 */
#define DUMP_VAL_U32(dump_pos, offset) \
  (*(U32 *) ((unsigned char *) DUMP_AREA_P +\
             DUMP_POS((dump_pos), (offset))))


#define RAMDUMP_OK 0
#define RAMDUMP_FAIL -1

typedef int RamdumpResult;


/**
 * @type DumpOffset
 *
 * @brief Offset in the dump.
 */
typedef U32 DumpOffset;

/**
 * @type DumpPosList
 *
 * @brief This array holds positions for parent chunks.
 *
 * @long This array holds positions for parent chunks, in order
 * to speed up the process of extending a dump and update
 * chunk sizes.
 */
typedef DumpPos DumpPosList[ MAX_IFF_DEPTH ];

/**
 * @type DumpPosSize
 *
 * @brief Holds a dump position and a chunk size
 */
typedef struct DumpPosSize
{
   DumpPos pos;
   U32 size;
} DumpPosSize;

/**
 * @type DumpPosSizeList
 *
 * @brief This array holds positions and sizes for parent chunks.
 *
 * @long This array holds positions and sizes for parent chunks, in order
 * to "undo" a FORM/BLOC that was not completely dumped
 */
typedef DumpPosSize DumpPosSizeList[ MAX_IFF_DEPTH ];


/**
 * @type RamdumpPersistentData
 *
 * @brief Persistent data in dump area.
 */
union RamdumpPersistentData
{
   struct
   {
      /*
       * Current dump and block count
       */
      struct
      {
         U32 seq_id;
         U32 block_count;
      } curr_dump;

      /*
       * Pointers to first block in each dump.
       * Indexed by seqId % DUMP_TABLE_SIZE
       */
      DumpPos dumps[ DUMP_TABLE_SIZE ];

   } vars;
   U8 rawBlock[ RAMDUMP_PERSISTENT_DATA_SIZE ];
};


#ifdef RAMDUMP_USE_ZLIB
/**
 * @type ZlibBuffers
 *
 * @brief Compression buffers for use by zlib.
 *
 * @field out_buf Buffer for zlib output.
 * @field in_buf Buffer for zlib input.
 */
struct ZlibBuffers
{
   U8 out_buf[ZLIB_COMPR_OUT_SIZE];
   U8 in_buf[ZLIB_COMPR_IN_SIZE];
};
#endif

/**
 * @type Dumparea
 *
 * @brief Actual dump area for post-mortem dumps.
 *
 * @field rawBlock
 */
struct DumpArea
{
   U32 rawBlock[ DUMP_ALIGNMENT >> 2];
};

/**
 * @type RamdumpArea
 *
 * @brief The ramdump memory area.
 *
 * @field pers_data Persistent ramdump data.
 */
struct RamdumpArea
{
   /* This comes always first */
   union RamdumpPersistentData pers_data;

   /* The rest is taken by dump area, aligned to DUMP_ALIGNMENT */
   struct DumpArea dump_area;
};

/* Access functions to dump area */

static void dump_area_read(void *buffer, DumpPos pos,
                           DumpOffset offset, unsigned long bytes);

static void dump_area_write(const void *buffer, DumpPos pos,
                            DumpOffset offset, unsigned long bytes);



/* IFF structure */

static U32 get_chunk_id(DumpPos pos);

static U32 get_chunk_data_size(DumpPos pos);

static U32 get_chunk_size(DumpPos pos);

static U32 get_chunk_data_U32(DumpPos pos, DumpOffset offset);

static int is_group_chunk(DumpPos pos);

static IffId get_group_contents_id(DumpPos pos);

static DumpPos get_first_group_chunk(DumpPos pos);

static DumpPos skip_chunk(DumpPos pos, U32 size_left);

static U32 is_valid_chunk(DumpPos pos, U32 max_chunk_size,
                          unsigned int depth, unsigned int maxdepth);

static U32 is_valid_iff_structure_depth(DumpPos pos, unsigned int maxdepth);

static DumpPos locate_chunk(DumpPos top_chunk, DumpOffset offset,
                            const IffId *cs,
                            DumpOffset last_located_offset,
                            unsigned long max_size);

static DumpPos find_chunk_in_valid_iff(DumpPos dump_pos, const IffId *cs);

static DumpPos find_next_chunk_in_valid_iff(DumpPos dump_pos,
                                            const IffId *cs,
                                            DumpPos last_found_chunk);


/* Dump validation, dump traversal */

static int is_dump_header(DumpPos pos);

static int is_valid_dump(DumpPos pos);

static RamdumpDumpId find_dump_id(DumpPos pos);

static DumpPos skip_dump(DumpPos pos);

/* Dump table */

static int is_dump_index_consistent(void);

static void build_dump_index(void);

static void update_dump_index(RamdumpDumpId dump_id, DumpPos new_dump);

static void write_chunk_header(DumpPos pos, IffId id, U32 data_size);

static void write_group_header(DumpPos pos, IffId id, U32 data_size,
                               IffId form_type);

static DumpPos extend_dump(DumpPos dump_pos, const DumpPosList plist,
                           int depth, unsigned long size);

static DumpPos add_text_string(DumpPos dump, const DumpPosList plist,
                               int depth, IffId id, const char *text);

static DumpPos add_description(DumpPos dump, const DumpPosList plist,
                               int depth, const char *descr);

/* Initializes a new dump */
static void write_new_dump(DumpPos pos);

/* Retrieves the most recent dump up to a given dump id */
static DumpPos get_most_recent_dump(RamdumpDumpId dump_id);

static DumpPos get_dump_to_extend(void);

static int store_dump_sizes(DumpPos pos);

static int restore_dump_sizes(void);

static DumpPos write_new_block_header(DumpPos pos, DumpPosList plist,
                                      int *depth);

static long advance_dump_id(void);

/* Implementation of trap functions */

static long dump_error_info(const char *description,
                            OSBOOLEAN user_called, OSERRCODE error_code,
                            OSERRCODE extra, RamdumpStatus *ret_code);

static long dump_text_info(const char *description, RamdumpStatus *ret_code);

static long dump_memory_image(const char *description, OSADDRESS start_address,
                              U32 length, RamdumpStatus *ret_code);

static long get_dump_id(U32 nth_last, RamdumpDumpId *dump_idp);

static long get_dump_info(RamdumpDumpId dump_id,
                          struct RamdumpDumpInfo *result,
                          RamdumpStatus *ret_code);

static long read_dump(RamdumpDumpId dump_id, U32 pos, void *buffer, U32 bytes,
                      S32 *bytes_read);

static void ramdump_initialize_area(void);
static RamdumpStatus ramdump_configure_ramdump_area(OSADDRESS base, U32 size,
						    U32 work_area_size,
                                                    U32 pmd_work_area_size);

/* Misc */

/**
 * @data watchdog_cb
 *
 * @brief Watchdog callback function.
 *
 * @seealso
 */
static RamdumpWatchdogCb *watchdog_cb = NULL;

/**
 * @data dump_locked
 *
 * @brief Indicates if the dump is locked or not.
 *
 * @long Indicates if the dump is locked or not.
 *
 * @seealso dump_lock dump_unlock
 */
static int dump_locked = 0;

/**
 * @data ramdump_initialized
 *
 * @brief Non-zero if ramdump has been initialized.
 */
static volatile int ramdump_initialized = 0;

/**
 * @data ramdump_area
 *
 * @brief Pointer to the ramdump area.
 */
struct RamdumpArea *ramdump_area = NULL;

/**
 * @data ramdump_area_length
 *
 * @brief Size of the ramdump area.
 */
U32 ramdump_area_length = 0;

/**
 * @data ramdump_work_area_length
 *
 * @brief Size of the ramdump work area.
 */
U32 ramdump_work_area_length = 0;

/**
 * @data ramdump_dump_area
 *
 * @brief Pointer to the actual dump area.
 */
struct DumpArea *ramdump_dump_area = NULL;

/**
 * @data ramdump_dump_area_length
 *
 * @brief Size of the dump area.
 */
U32 ramdump_dump_area_length = 0;

#ifdef RAMDUMP_USE_ZLIB
/**
 * @data ramdump_zlib_buffers
 *
 * @brief Zlib compression buffers.
 */
static struct ZlibBuffers *ramdump_zlib_buffers;
#endif

/**
 * @data compr_stream
 *
 * @brief zlib compression stream
 */
 static z_stream compr_stream;

/**
 * @data last_chunk_depth
 *
 * @brief depth of last chunk written to dump
 */
int last_chunk_depth = 0;

/**
 * @data last_chunk_plist
 *
 * @brief list of dump positions for the last chunk written and its parents
 */
static DumpPosList last_chunk_plist;

/**
 * @data stored_dump_sizes
 *
 * @brief list of stored dump positions and chunk sizes for the purpose of
 * reverting the write operation for an unfinished FORM/BLOC
 */
static DumpPosSizeList stored_dump_sizes;


#ifndef RAMDUMP_DO_NOT_SWITCH_STACK
/**
 * @data ramdump_stack
 *
 * @brief Internal stack to use in system error handler.
 */
static uint64_t ramdump_stack[ RAMDUMP_STACK_SIZE / sizeof(uint64_t) ];

/**
 * @macro RAMDUMP_STACK
 *
 * @brief Macro for pointer to stack base.
 */
#define RAMDUMP_STACK ramdump_stack

/**
 * @macro RAMDUMP_STACK_TOP
 *
 * @brief Macro for pointer to stack top.
 */
#define RAMDUMP_STACK_TOP ((U32 *) ((char *) RAMDUMP_STACK + \
                          sizeof(RAMDUMP_STACK)-16))
#endif

/**
 * @function dump_area_read
 *
 * @brief Copies a number of bytes from dump area into a buffer.
 *
 * @long Copies a number of bytes from dump area into a buffer.
 * Wrap is handled.
 *
 * @output buffer Buffer to copy bytes from dump to.
 * @input pos Dump position to read from, with a specified offset
 * @input offset Offset from the dump position to read from.
 * @input bytes Number of bytes to copy
 *
 * @seealso dump_area_write
 */
static void
dump_area_read(void *buffer, DumpPos pos,
	       DumpOffset offset, unsigned long bytes)
{
   DumpPos src_pos;
   unsigned long left_to_wrap;

   assert(VALID_DUMP_POS(pos));
   assert(offset <= DUMP_AREA_SIZE);
   assert(bytes <= DUMP_AREA_SIZE);
   assert(offset + bytes <= DUMP_AREA_SIZE);

   src_pos = DUMP_POS(pos, offset);

   assert(VALID_DUMP_POS(src_pos));
   assert(VALID_DUMP_POS(DUMP_POS(src_pos, bytes - 1)));

   /* Will wrap occur? */
   left_to_wrap = DUMP_AREA_SIZE - src_pos;

   if (bytes > left_to_wrap)
   {
      /* Yes, copy in two steps */
      memcpy(buffer, (const void *) ((char *) DUMP_AREA_P + src_pos),
             left_to_wrap);

      /* Wrap */
      memcpy((void *) ((char *) buffer + left_to_wrap),
             (const void *) DUMP_AREA_P,
             bytes - left_to_wrap);
   }
   else
   {
      memcpy(buffer, (const void *) ((char *) DUMP_AREA_P + src_pos), bytes);
   }
}


/**
 * @function dump_area_write
 *
 * @brief Copies a number of bytes from a buffer into the dump area.
 *
 * @long Copies a number of bytes from a buffer into the dump area.
 * Wrap is handled.
 *
 * @input buffer Buffer to copy bytes from.
 * @input pos Dump position to write bytes to, given an offset
 * @input offset Offset from the dump position to write bytes to.
 * @input bytes Number of bytes to copy.
 *
 * @seealso dump_area_read
 */
static void
dump_area_write(const void *buffer, DumpPos pos,
      DumpOffset offset, unsigned long bytes)
{
   DumpPos dest_pos;
   unsigned long left_to_wrap;

   assert(VALID_DUMP_POS(pos));
   assert(offset <= DUMP_AREA_SIZE);
   assert(bytes <= DUMP_AREA_SIZE);
   assert(offset + bytes <= DUMP_AREA_SIZE);

   dest_pos = DUMP_POS(pos, offset);

   assert(VALID_DUMP_POS(dest_pos));
   assert(VALID_DUMP_POS(DUMP_POS(dest_pos, bytes - 1)));

   /* Will wrap occur? */
   left_to_wrap = DUMP_AREA_SIZE - dest_pos;

   if (bytes > left_to_wrap)
   {
      /* Yes, copy in two steps */
      memcpy((void *) ((char *) DUMP_AREA_P + dest_pos), buffer, left_to_wrap);

      /* Wrap */
      memcpy((void *) DUMP_AREA_P,
             /* (The const char *) cast was originally a (char *) cast
                but it caused a warning. */
             (const void *) ((const char *) buffer + left_to_wrap),
             bytes - left_to_wrap);
   }
   else
   {
      memcpy((void *) ((char *) DUMP_AREA_P + dest_pos), buffer, bytes);
   }
}

/**
 * @function dump_area_clear
 *
 * @brief Clears a number of bytes in the dump area.
 *
 * @long Clears a number of bytes in the dump area. Wrap is handled.
 *
 * @input pos Dump position to write bytes to, given an offset
 * @input offset Offset from the dump position to write bytes to.
 * @input bytes Number of bytes to clear.
 * @input value Clear byte value.
 *
 * @seealso dump_area_write
 */
static void
dump_area_clear(DumpPos pos, DumpOffset offset, unsigned long bytes,
                unsigned char value)
{
   DumpPos dest_pos;
   unsigned long left_to_wrap;

   assert(VALID_DUMP_POS(pos));
   assert(offset <= DUMP_AREA_SIZE);
   assert(bytes <= DUMP_AREA_SIZE);
   assert(offset + bytes <= DUMP_AREA_SIZE);

   dest_pos = DUMP_POS(pos, offset);

   assert(VALID_DUMP_POS(dest_pos));
   assert(VALID_DUMP_POS(DUMP_POS(dest_pos, bytes - 1)));

   /* Will wrap occur? */
   left_to_wrap = DUMP_AREA_SIZE - dest_pos;

   if (bytes > left_to_wrap)
   {
      /* Yes, clear in two steps */
      memset((void *) ((char *) DUMP_AREA_P + dest_pos), value, left_to_wrap);

      /* Wrap */
      memset((void *) DUMP_AREA_P, value, bytes - left_to_wrap);
   }
   else
   {
      memset((void *) ((char *) DUMP_AREA_P + dest_pos), value, bytes);
   }
}

/*
 * ==========================================================================
 *
 * Iff structure functionality.
 *   - Access functions for various data fields of chunks.
 *   - Get chunk id, chunk size, chunk data size, etc.
 *   -
 *
 * ==========================================================================
 */

/**
 * @function get_chunk_id
 *
 * @brief Get the chunk id for the chunk at the given dump position.
 *
 * @input pos Dump position pointing at the start of a chunk.
 *
 * @return Returns the chunk id for the chunk.
 */
static U32
get_chunk_id(DumpPos pos)
{
   assert(VALID_DUMP_POS(pos));
   assert(U32_ALIGNED(pos));

   /* Read ID */
   return ntohl(DUMP_VAL_U32(pos, OFFSET(struct RamdumpChunkHeader, id)));
}

/**
 * @function get_chunk_data_size
 *
 * @brief Retrieve the data size of a chunk at the given dump position.
 *
 * @input pos Dump position pointing at the start of a chunk.
 *
 * @return Returns the data size of the chunk.
 */
static U32
get_chunk_data_size(DumpPos pos)
{
   assert(VALID_DUMP_POS(pos));
   assert(U32_ALIGNED(pos));

   /* Read size */
   return ntohl(DUMP_VAL_U32(pos, OFFSET(struct RamdumpChunkHeader, size)));
}

/**
 * @function get_chunk_size
 *
 * @brief Get the total size of a chunk, header included.
 *
 * @input pos Dump position pointing to the start of the chunk.
 *
 * @return Returns the total size of the chunk, header included.
 */
static U32
get_chunk_size(DumpPos pos)
{
   assert(VALID_DUMP_POS(pos));
   assert(U32_ALIGNED(pos));

   return get_chunk_data_size(pos) + sizeof(struct RamdumpChunkHeader);
}

/**
 * @function get_chunk_data_U32
 *
 * @brief Get a U32 value from chunk data.
 *
 * @input pos Dump position pointing at the start of a chunk.
 * @input offset Offset from the start of the chunk to read the value from.
 *
 * @return Returns the U32 value at the given offset of the chunk.
 */
static U32
get_chunk_data_U32(DumpPos pos, DumpOffset offset)
{
   assert(VALID_DUMP_POS(pos));
   assert(U32_ALIGNED(pos));
   assert(U32_ALIGNED(offset));

   assert(offset < get_chunk_size(pos));

   return DUMP_VAL_U32(pos, offset);
}

/**
 * @function is_group_id
 *
 * @brief Checks if a chunk id is a FORM, LIST, CAT_ or PROP.
 *
 * @input id Chunk id to check.
 *
 * @return Returns non-zero if id is a FORM, LIST, CAT_ or PROP.
 *
 * @seealso is_group_chunk
 */
static int
is_group_id(IffId id)
{
   return (id == RAMDUMP_TAG_FORM || id == RAMDUMP_TAG_LIST ||
           id == RAMDUMP_TAG_PROP || id == RAMDUMP_TAG_CAT_);
}

/**
 * @function is_group_chunk
 *
 * @brief Checks if the id of a chunk is a FORM, LIST, CAT_ or PROP.
 *
 * @input pos Dump position pointing at the start of a chunk.
 *
 * @return Returns non-zero if id of chunk is a FORM, LIST, CAT_ or PROP.
 *
 * @seealso is_group_id
 */
static int
is_group_chunk(DumpPos pos)
{
   IffId id;

   assert(VALID_DUMP_POS(pos));
   assert(U32_ALIGNED(pos));

   /* Read ID */
   id = get_chunk_id(pos);

   return is_group_id(id);
}

/**
 * @function get_group_contents_id
 *
 * @brief Get the form type or contents type of a FORM, LIST, CAT_ or
 * PROP chunk.
 *
 * @input pos Dump position pointing at the start of a chunk.
 *
 * @return Returns the form type or contents type of a group chunk.
 */
static IffId
get_group_contents_id(DumpPos pos)
{
   DumpOffset offset = OFFSET(struct RamdumpHeaderLIST, cont_type);

   assert(is_group_chunk(pos));

   return ntohl(DUMP_VAL_U32(pos, offset));
}


/**
 * @function get_first_group_chunk
 *
 * @brief Get first chunk in a group.
 *
 * @long Gets the first subchunk of a group chunk.
 *
 * @input pos Dump position pointing at the start of a group chunk.
 *
 * @return Returns the position of the first subchunk of a group chunk,
 * or NO_DUMP_POS if no subchunk was found.
 */
static DumpPos
get_first_group_chunk(DumpPos pos)
{
   U32 size;

   assert(VALID_DUMP_POS(pos));
   assert(U32_ALIGNED(pos));

   /* Check if group chunk */
   if (!is_group_chunk(pos))
   {
      return NO_DUMP_POS;
   }

   /* Get size */
   size = get_chunk_size(pos);

   assert(U32_ALIGNED(size));
   assert(size >= sizeof(struct RamdumpHeaderFORM));

   size -= sizeof(struct RamdumpHeaderFORM);

   if (size > 0)
   {
      assert(size >= sizeof(struct RamdumpChunkHeader));

      return DUMP_POS(pos, sizeof(struct RamdumpHeaderFORM));
   }
   else
   {
      return NO_DUMP_POS;
   }
}

/**
 * @function skip_chunk
 *
 * @brief Skips to the next chunk.
 *
 * @input pos Dump position pointing at the start of a chunk.
 * @input size_left Size left at chunk level.
 *
 * @return Position of the next chunk.
 */
static DumpPos
skip_chunk(DumpPos pos, U32 size_left)
{
   DumpPos next_chunk_pos = NO_DUMP_POS;
   U32 ch_size;

   assert(U32_ALIGNED(size_left));
   assert(size_left >= sizeof(struct RamdumpChunkHeader));

   /* Chunk size */
   ch_size = get_chunk_size(pos);
   assert(U32_ALIGNED(ch_size));

   if (ch_size <= size_left)
   {
      next_chunk_pos = DUMP_POS(pos, ch_size);

      assert(VALID_DUMP_POS(next_chunk_pos));
   }

   return next_chunk_pos;
}

/**
 * @function is_valid_chunk
 *
 * @brief Checks if the specified iff chunk is valid.
 *
 * @long Checks if the iff chunk is valid. The checking is aware
 * of depth and goes no deeper than maxdepth levels. To check only
 * the topmost level, use depth = 0, maxdepth = 1.
 *
 * @input pos Dump position pointing at start of chunk.
 * @input max_chunk_size Size of chunk including subchunks.
 * @input depth Current depth in chunk tree.
 * @input maxdepth Maximum depth to check.
 *
 * @return Returns the number of bytes in the chunk, or zero if the
 * chunk is invalid.
 */
static U32
is_valid_chunk(DumpPos pos, U32 max_chunk_size,
	       unsigned int depth, unsigned int maxdepth)
{
   U32 size_left = max_chunk_size;
   unsigned long chunk_size;

   assert(VALID_DUMP_POS(pos));
   assert(U32_ALIGNED(pos));

   if (size_left < sizeof(struct RamdumpChunkHeader))
   {
      return 0;
   }

   /* Read chunk size */
   chunk_size = get_chunk_size(pos);
   if (!U32_ALIGNED(chunk_size) || chunk_size > size_left)
   {
      return 0;
   }

   assert(U32_ALIGNED(chunk_size));
   assert(chunk_size <= size_left);

   /* Dump structure may contain a vast number of entries at level 2 */
   if (depth == 2)
   {
      ramdump_kick_watchdog();
   }

   /* Only check maxdepth levels */
   if ((depth + 1) < maxdepth && is_group_chunk(pos))
   {
      /* Check that chunk is big enough for header */
      if (size_left < sizeof(struct RamdumpHeaderFORM) ||
          chunk_size < sizeof(struct RamdumpHeaderFORM))
      {
         return 0;
      }
      else
      {
         DumpPos tmp_pos;

         /* Skip group header */
         U32 chunk_size_left = chunk_size - sizeof(struct RamdumpHeaderFORM);

         /* We may now have a sequence of chunks */
         if (chunk_size_left > 0)
         {

            tmp_pos = get_first_group_chunk(pos);

            while (chunk_size_left > 0)
            {
               U32 tmp_chunk_size;

               /* No more chunks, although size says so? */
               if (tmp_pos == NO_DUMP_POS)
               {
                  return 0;
               }

               /* Check chunk */
               tmp_chunk_size = is_valid_chunk(tmp_pos, chunk_size_left,
                                               depth + 1, maxdepth);

               /* Valid chunk? */
               if (tmp_chunk_size == 0)
               {
                  return 0;
               }

               assert(tmp_chunk_size <= chunk_size_left);

               /* Skip valid chunk */
               tmp_pos = skip_chunk(tmp_pos, chunk_size_left);
               chunk_size_left -= tmp_chunk_size;
            }
         }

         assert(chunk_size_left == 0);

         /* Thru sequence of chunks */
      }
   }

   return chunk_size; /* size of chunk */
}

/**
 * @macro is_valid_iff_structure
 *
 * @brief Check if the specified iff structure is valid to the maximum depth.
 *
 * @input pos Dump position pointing at the start of a chunk,
 *
 * @return Returns number of bytes in the structure, or zero if the
 * structure is invalid.
 */
#define is_valid_iff_structure(pos) is_valid_iff_structure_depth(pos, \
                                                                 MAX_IFF_DEPTH)


/**
 * @function is_valid_iff_structure_depth
 *
 * @brief Checks if the specified iff structure is valid.
 *
 * @long Checks if the specified iff structure is valid. The check
 * is only performed to the indicated number of levels.
 *
 * @input pos Dump position pointing at the start of a chunk.
 * @input maxdepth Maximum depth to check.
 *
 * @return Returns number of bytes in the structure, or zero if the
 * structure is invalid.
 */
static U32
is_valid_iff_structure_depth(DumpPos pos, unsigned int maxdepth)
{
  IffId id;
  U32 size;

  assert(VALID_DUMP_POS(pos));
  assert(U32_ALIGNED(pos));

  /* Read ID and size */
  id = get_chunk_id(pos);
  size = get_chunk_size(pos);

  /* Check size constraints */
  if (size > DUMP_AREA_SIZE)
  {
     return 0;
  }

  assert(size <= DUMP_AREA_SIZE);

  /* Top level tags */
  if (id == RAMDUMP_TAG_FORM || id == RAMDUMP_TAG_LIST ||
      id == RAMDUMP_TAG_CAT_ || id == RAMDUMP_TAG_FILL)
  {
     return is_valid_chunk(pos, size, 0, maxdepth);
  }

  return 0;
}

/**
 * @function locate_chunk
 *
 * @brief Finds the next matching chunk according to a specification.
 *
 * @long Finds the next matching chunk according to a specification,
 * e.g. {'FORM', 'PMD_', 'DPID'}. Chunks matching the specification
 * are ignored until the algorithm has passed the last found chunk.
 * No more than max_size bytes are scanned.
 *
 * @input top_chunk Dump position pointing at the top chunk.
 * @input offset Offset to start looking at.
 * @input cs Chunk specification array.
 * @input last_located_offset Offset of the last located chunk.
 * @input max_size Scan at most max_size bytes.
 *
 * @return Returns the position of the chunk, if one was found. Returns
 * NO_DUMP_POS on failure.
 *
 * @seealso find_chunk_in_valid_iff
 */
static DumpPos
locate_chunk(DumpPos top_chunk, DumpOffset offset, const IffId *cs,
             DumpOffset last_located_offset, unsigned long max_size)
{
   DumpOffset curr_offset = offset;
   unsigned long size_left = max_size;
   DumpPos tmp_pos;
   DumpPos matching_pos;
   IffId id;

   assert(*cs);
   if (!*cs)
   {
      return NO_DUMP_POS;
   }

   /*
    * If last located chunk is not within size_left bytes, we don't
    * bother to check this chunk.
    */
   if (last_located_offset >= (curr_offset + size_left))
   {
      /* No */
      return NO_DUMP_POS;
   }

   /* Scan chunks */
   while (size_left > 0)
   {
      unsigned long tmp_size;
      DumpPos next_pos;

      tmp_pos = DUMP_POS(top_chunk, curr_offset);
      id = get_chunk_id(tmp_pos);

      /* Prepare to skip chunk */
      tmp_size = get_chunk_size(tmp_pos);
      next_pos = skip_chunk(tmp_pos, size_left);

      matching_pos = NO_DUMP_POS;

      /*
       * If no previous match, or if we are checking a chunk that
       * may contain an offset later than the last found, check for match
       */
      if (last_located_offset == 0 ||
          (curr_offset + tmp_size) > last_located_offset)
      {
         /* Check if match */
         if (id == *cs || *cs == RAMDUMP_TAG_WILD)
         {
            unsigned long header_size;

            if (is_group_id(id))
            {
               /* Get wanted contents type */
               IffId ctype_to_match = *(cs + 1);

               header_size = sizeof(struct RamdumpHeaderLIST);

               /* No malformed id */
               assert(ctype_to_match);
               if (!ctype_to_match)
               {
                  return NO_DUMP_POS;
               }

               /* Check if match */
               if (ctype_to_match == RAMDUMP_TAG_WILD ||
                   get_group_contents_id(tmp_pos) == ctype_to_match)
               {
                  assert(size_left >= header_size);

                  if (*(cs + 2))
                  {
                     /* Find chunk in substructure */
                     matching_pos = locate_chunk(top_chunk,
                                                 curr_offset + header_size,
                                                 cs + 2, last_located_offset,
                                                 tmp_size - header_size);
		            }
                  else
                  {
                     /* Last part in specification matched */

                     /* A new match? */
                     if (curr_offset > last_located_offset)
                     {
                        matching_pos = tmp_pos;
                     }
                  }
               }
            }
            else
            {
               header_size = sizeof(struct RamdumpChunkHeader);

               assert(size_left >= header_size);

               if (*(cs + 1))
               {
                  /* Find chunk in substructure */
                  matching_pos = locate_chunk(top_chunk,
                                              curr_offset + header_size,
                                              cs + 1, last_located_offset,
                                              tmp_size - header_size);
               }
               else
               {
                  /* Last chunk in specification matched */

                  /* A new match? */
                  if (curr_offset > last_located_offset)
                  {
                     matching_pos = tmp_pos;
                  }
               }
            }

            if (matching_pos != NO_DUMP_POS)
            {
               return matching_pos;
            }
         }
      }

      /* Skip to next chunk */
      assert(size_left >= tmp_size);
      size_left -= tmp_size;
      curr_offset += tmp_size;

      (void) next_pos; /* Lint */
      assert(DUMP_POS(top_chunk, curr_offset) == next_pos);
   }

   /* No match found, and we should have checked every byte */
   assert(size_left == 0);

   return NO_DUMP_POS;
}

/**
 * @function func_chunk_in_valid_iff
 *
 * @brief Finds a chunk according to a specification.
 *
 * @long Finds a chunk according to a specification, in a valid iff
 * structure, e.g. {'FORM', 'PMD_', 'DPID'}
 *
 * @input dump_pos Dump position where the scan will start.
 * @input cs Chunk specification array.
 *
 * @return Returns the position of the chunk, if successful. Returns
 * NO_DUMP_POS on failure.
 *
 * @seealso locate_chunk
 */
static DumpPos
find_chunk_in_valid_iff(DumpPos dump_pos, const IffId *cs)
{
   /*
    * assert(VALID_DUMP_POS(dump_pos));
    * assert(is_valid_iff_structure(dump_pos));
    */

   return locate_chunk(dump_pos, 0, cs, 0, get_chunk_size(dump_pos));
}

/**
 * @function find_next_chunk_in_valid_iff
 *
 * @brief Finds the next matching chunk according to a specification.
 *
 * @long Finds the next matching chunk according to a specification
 * in a valid iff structure, e.g. {'FORM', 'PMD_', 'DPID'}. Chunks
 * matching the specification are ignored until the algorithm has
 * passed the last found chunk.
 *
 * @input dump_pos Dump position to start scanning at.
 * @input cs Chunk specification array.
 * @input last_found_chunk
 *
 * @return Returns position of next matching chunk or NO_DUMP_POS on
 * failure.
 *
 * @seealso locate_chunk
 */
static DumpPos
find_next_chunk_in_valid_iff(DumpPos dump_pos, const IffId *cs,
			     DumpPos last_found_chunk)
{
   /*
    * assert(VALID_DUMP_POS(dump_pos));
    * assert(is_valid_iff_structure(dump_pos));
    */

   return locate_chunk(dump_pos, 0, cs,
                       DUMP_OFFSET(last_found_chunk, dump_pos),
                       get_chunk_size(dump_pos));
}

/*
 * ==========================================================================
 *
 * Dump functionality.
 *   - Validating a dump
 *   - Handles a dump table for fast access
 *   - Traversing the dump area to find dumps
 *
 * ==========================================================================
 */

/**
 * @function is_dump_header
 *
 * @brief Checks ('FORM' size 'PMD ') in header.
 *
 * @input pos Dump position pointing at the first chunk of the dump.
 *
 * @return Returns non-zero if header is valid.
 */
static int
is_dump_header(DumpPos pos)
{
   struct RamdumpHeaderFORM ramdump_FORM_header;

   assert(VALID_DUMP_POS(pos));
   assert(U32_ALIGNED(pos));

   dump_area_read(&ramdump_FORM_header, pos, 0,
                  sizeof(struct RamdumpHeaderFORM));

   ramdump_FORM_header.chd.id = ntohl(ramdump_FORM_header.chd.id);
   ramdump_FORM_header.chd.size = ntohl(ramdump_FORM_header.chd.size);
   ramdump_FORM_header.form_type = ntohl(ramdump_FORM_header.form_type);

   if (ramdump_FORM_header.chd.id == RAMDUMP_TAG_FORM &&
       ramdump_FORM_header.form_type == RAMDUMP_TAG_PMD_)
   {
      return RAMDUMP_TRUE;
   }

   return RAMDUMP_FALSE;
}

/* Tag specification paths to various common information */

/**
 * @data dpid
 *
 * @brief Tag specification array for locating a dump_id
 */
static const IffId dpid[] = { RAMDUMP_TAG_FORM, RAMDUMP_TAG_PMD_,
                              RAMDUMP_TAG_DPID, 0 };

/**
 * @data block_hd
 *
 * @brief Tag specification array for locating a block header.
 */
static const IffId block_hd[] = { RAMDUMP_TAG_FORM, RAMDUMP_TAG_PMD_,
                                  RAMDUMP_TAG_LIST, RAMDUMP_TAG_BLOC,
                                  RAMDUMP_TAG_FORM, RAMDUMP_TAG_BLOC,
                                  RAMDUMP_TAG_BLHD, 0 };

/**
 * @function is_valid_dump
 *
 * @brief Validates a dump.
 *
 * @long Validates a dump based on:
 * 1) valid dump header,
 * 2) valid iff structure,
 * 3) FORM BLOCs contain valid dump_id and enumeration.
 *
 * @input pos Dump position pointing to the beginning of a dump.
 *
 * @return Returns non-zero if dump is valid.
 */
static int
is_valid_dump(DumpPos pos)
{
   DumpPos tmp_pos, last_block_hd_pos;
   RamdumpDumpId dump_id;
   U32 block_no = 0;

   assert(VALID_DUMP_POS(pos));
   assert(U32_ALIGNED(pos));

   /* 1. Header */
   if (!is_dump_header(pos))
   {
      return RAMDUMP_FALSE;
   }

   /* 2. IFF structure */
   if (!is_valid_iff_structure(pos))
   {
      return RAMDUMP_FALSE;
   }

   /* 3. Retreive dump id */
   tmp_pos = find_chunk_in_valid_iff(pos, dpid);

   if (tmp_pos == NO_DUMP_POS)
   {
      return RAMDUMP_FALSE;
   }

   dump_id = get_chunk_data_U32(tmp_pos,
                                OFFSET(struct RamdumpChunkDPID, dump_id));

   /* 4. Check all blocks for existing dump id and correct block enumeration */
   /* IFF structure ok above */
   last_block_hd_pos = find_chunk_in_valid_iff(pos, block_hd);

   /*
    * Since there may be many blocks, we cannot afford to check them all.
    * However, block enumeration is guaranteed by the dump lock in
    * ramdump_trap_handler.
    */
   if (last_block_hd_pos != NO_DUMP_POS)
   {
      /* Check dump_id and block no */
      if (get_chunk_data_U32(last_block_hd_pos,
                             OFFSET(struct RamdumpChunkBLHD, dump_id)) ==
          dump_id)
      {
         U32 block = get_chunk_data_U32(last_block_hd_pos,
                                        OFFSET(struct RamdumpChunkBLHD,
                                        block_no));

         if (block != block_no)
         {
            return RAMDUMP_FALSE;
         }
      }
      else
      {
         /* Another dump_id found, abort */
         return RAMDUMP_FALSE;
      }

      block_no++;
   }

   /* At least one valid block */
   if (block_no == 0)
   {
      return RAMDUMP_FALSE;
   }

   return RAMDUMP_TRUE;
}

/**
 * @function find_dump_id
 *
 * @brief Finds the DPID tag for the dump id in the dump and returns the
 * dump id.
 *
 * @input pos Dump position pointing at the start of the dump.
 *
 * @return Returns the dump id of the DPID chunk or 0 on failure.
 */
static RamdumpDumpId
find_dump_id(DumpPos pos)
{
   DumpPos chpos;
   RamdumpDumpId dump_id = 0;

   assert(VALID_DUMP_POS(pos));
   assert(is_valid_iff_structure_depth(pos, 2));

   chpos = find_chunk_in_valid_iff(pos, dpid);

   if (chpos != NO_DUMP_POS)
   {
      dump_id = get_chunk_data_U32(chpos,
                                   OFFSET(struct RamdumpChunkDPID, dump_id));
   }

   return dump_id;
}


/**
 * @function skip_dump
 *
 * @brief Skips the dump that starts from a given position and gives
 * the next dump aligned address.
 *
 * @input pos Dump position pointing to start of dump to skip.
 *
 * @return Returns dump position of next dump.
 */
static DumpPos
skip_dump(DumpPos pos)
{
   DumpPos aligned_pos, next_pos;

   assert(is_valid_dump(pos));

   next_pos = skip_chunk(pos, DUMP_AREA_SIZE);

   if (!VALID_DUMP_POS(next_pos))
   {
      return NO_DUMP_POS;
   }

   aligned_pos = NEXT_ALIGNED_DUMP_POS(next_pos);

   if (!VALID_DUMP_POS(aligned_pos))
   {
      return NO_DUMP_POS;
   }

   assert(ALIGNED(aligned_pos, DUMP_ALIGNMENT));

   return aligned_pos;
}

/**
 * @function is_dump_index_consistent
 *
 * @brief Checks the consistency of the dump index.
 *
 * @long Check the consistency of the dump index. The dump index
 * is a global array. Each entry in the array is a dump position.
 *
 * @return Returns non-zero if the dump index is consistent.
 *
 * @seealso DUMP_TABLE
 */
static int
is_dump_index_consistent(void)
{
   /* This was orignially a signed int, but it
      caused a warning in the comparison later
      in this function. */
   unsigned int i;

   /* Kick watchdog, table contains possibly 64 entries */
   ramdump_kick_watchdog();

   /*
    * Scan the table to see if any entry is not consistent
    */
   for (i = 0; i < DUMP_TABLE_SIZE; i++)
   {
      DumpPos pos = DUMP_TABLE[ i ];

      if (pos == NO_DUMP_POS)
      {
         /* Ok */
      }
      else if (VALID_DUMP_POS(pos) && ALIGNED(pos, DUMP_ALIGNMENT) &&
	            is_dump_header(pos) && is_valid_iff_structure_depth(pos, 2))
      {
         RamdumpDumpId dump_id = find_dump_id(pos);

         /* Kick watchdog, table contains possibly 64 entries */
         ramdump_kick_watchdog();

         if (!dump_id || HASH_DUMP_ID(dump_id) != i)
         {
            return RAMDUMP_FALSE;
         }

         /* Ok */
      }
      else
      {
         /* Here ? Discrepancy found */
         return RAMDUMP_FALSE;
      }
   }

   return RAMDUMP_TRUE;
}

/**
 * @function clear_dumps
 *
 * @brief Clears the dump area and dump index.
 *
 * @seealso DUMP_TABLE
 */
static long
clear_dumps(void)
{
   U32 i;
   U32 step_size = 16384;
   U32 steps;
   U32 last_step_size;

   /* Initialize table to empty */
   for (i = 0; i < DUMP_TABLE_SIZE; i++)
   {
      DUMP_TABLE[i] = NO_DUMP_POS;
   }

   steps = DUMP_AREA_SIZE / step_size;
   last_step_size = DUMP_AREA_SIZE - steps * step_size;

   ramdump_kick_watchdog();

   for (i = 0; i < steps; i++)
   {
      memset(((char *)DUMP_AREA_P) + i * step_size, 0xDEADBEEFUL, step_size);
      ramdump_kick_watchdog();
   }

   if (last_step_size > 0)
   {
      memset(((char *)DUMP_AREA_P) + steps * step_size, 0xDEADBEEFUL,
             last_step_size);
      ramdump_kick_watchdog();
   }

   return 0;
}

/**
 * @function build_dump_index
 *
 * @include
 *
 * @brief Builds the dump index from scratch.
 *
 * @long This function builds the dump index from scratch, i.e. it does
 * not assume anything about the state being valid in the dump area.
 *
 * @seealso DUMP_TABLE
 */
static void
build_dump_index(void)
{
   DumpPos pos, next_pos = 0;
   int i;

   /* Initialize table to empty */
   for (i = 0; i < DUMP_TABLE_SIZE; i++)
   {
      DUMP_TABLE[ i ] = NO_DUMP_POS;
   }

   assert(ALIGNED(next_pos, DUMP_ALIGNMENT));
   assert(VALID_DUMP_POS(next_pos));

   /* Scan to next valid dump */
   i = 0;
   do
   {
      pos = next_pos;

      /* Kick watchdog, scanning the whole pmd area might take some time */
      if ((i++ & 0xff) == 0)
      {
         ramdump_kick_watchdog();
      }

      if (is_valid_dump(pos))
      {
         /* Found valid */
         RamdumpDumpId dump_id = find_dump_id(pos);
         DumpPos old_pos = DUMP_TABLE[ HASH_DUMP_ID(dump_id) ];

         /* Kick watchdog, scanning the whole dump area might take some time */
         ramdump_kick_watchdog();

         if (old_pos == NO_DUMP_POS || !is_valid_dump(old_pos))
         {
            DUMP_TABLE[ HASH_DUMP_ID(dump_id)] = pos;
         }
         else
         {
            RamdumpDumpId old_dump_id = find_dump_id(old_pos);

            /* Handles wrap nicely */
            if (((dump_id - old_dump_id) & 0x80000000UL) == 0)
            {
               /* Update with later dump */
               DUMP_TABLE[ HASH_DUMP_ID(dump_id) ] = pos;
            }
         }
         /* Now skip blocks within the same dump */
         next_pos = skip_dump(pos);
      }
      else
      {
         next_pos = NEXT_ALIGNED_DUMP_POS(pos + 1);
      }
   } while (VALID_DUMP_POS(next_pos) && next_pos > pos);

   /*
    * We should now have an index for each dump id that was found
    */
   assert(is_dump_index_consistent());
}

/**
 * @function update_dump_index
 *
 * @brief Updates the dump index table with a new dump.
 *
 * @input dump_id Dump identifier of the new dump.
 * @input new_dump Dump position of the new dump.
 *
 * @seealso DUMP_TABLE
 */
static void
update_dump_index(RamdumpDumpId dump_id, DumpPos new_dump)
{
   int i;

   assert(VALID_DUMP_POS(new_dump));
   assert(is_dump_header(new_dump));

   DUMP_TABLE[ HASH_DUMP_ID(dump_id) ] = new_dump;

   /*
    * Loop through the table and remove any entries that doesn't point to
    * a valid dump
    */
   for (i = 0; i < DUMP_TABLE_SIZE; i++)
   {
      DumpPos pos = DUMP_TABLE[ i ];

      /* Still valid? */
      if (VALID_DUMP_POS(pos))
      {

         /* Reset entry if not valid */
         if (!is_dump_header(pos))
         {
            /*
             * By doing this we might lose a valid, older dump that
             * hashes on the same entry in the table, but that dump
             * is DUMP_TABLE_SIZE dumps old. It should be enough to be
             * able to see the DUMP_TABLE_SIZE latest entries
             */
            DUMP_TABLE[ i ] = NO_DUMP_POS;
         }
      }
   }
}

/**
 * @function write_chunk_header
 *
 * @brief Writes a chunk header to the dump.
 *
 * @input pos Dump position where the chunk will be written
 * @input id Chunk id tag.
 * @input data_size Data size of the chunk.
 */
static void
write_chunk_header(DumpPos pos, IffId id, U32 data_size)
{
   DUMP_VAL_U32(pos, OFFSET(struct RamdumpChunkHeader, id)) = htonl(id);
   DUMP_VAL_U32(pos, OFFSET(struct RamdumpChunkHeader,
                            size)) = htonl(data_size);
}

/**
 * @function write_group_header
 *
 * @brief Writes a group header (FORM, LIST, CAT_ or PROP).
 *
 * @input pos Dump position where the group header will be written.
 * @input id Group chunk id.
 * @input data_size Data size of the group chunk.
 * @input form_type Group subchunk id.
 */
static void
write_group_header(DumpPos pos, IffId id,
		   U32 data_size, IffId form_type)
{
   struct RamdumpHeaderFORM hd_FORM;

   hd_FORM.chd.id = htonl(id);
   hd_FORM.chd.size = htonl(data_size);
   hd_FORM.form_type = htonl(form_type);

   dump_area_write((const void *) &hd_FORM, pos, 0,
                   sizeof(struct RamdumpHeaderFORM));
}

/**
 * @function add_child_location
 *
 * @brief Set the chunk position at given depth to given position.
 *
 * @input plist Array of parent chunks.
 * @input depth Depth in parent array to write position.
 * @input pos Dump position to put in the parent array at given depth.
 */
static void
add_child_location(DumpPosList plist, int depth, DumpPos pos)
{
	int i;

	assert(depth < MAX_IFF_DEPTH);

	plist[ depth ] = pos;

	depth++;

	/* Keep track of latest extended chunk */
	for (i = 0; i < depth; i++)
	{
		last_chunk_plist[i] = plist[i];
	}
	last_chunk_depth = depth;
}

/**
 * @function extend_dump
 *
 * @brief Extends a dump with a number of bytes and returns the position
 * of the added chunk.
 *
 * @long Extends a dump with a number of bytes and returns the position
 * of the added chunk. A parent chunk position list must be specified
 * in order to update the chunk sizes in the dump.
 *
 * @input dump_pos Dump position where the dump will be extended.
 * @input plist Contains the parent positions.
 * @input depth Indicates the number of positions in the list.
 * @input size Chunk size.
 *
 * @return
 */
static DumpPos
extend_dump(DumpPos dump_pos, const DumpPosList plist,
	    int depth, unsigned long size)
{
   /* Check that the parent chunk is the last chunk in the dump */
   DumpOffset parent_chunk_offset;
   U32 p_ch_size;
   U32 old_dump_size = get_chunk_size(dump_pos);
   DumpPos new_chunk_pos;
   int i;

   assert(plist);
   assert(depth > 0);
   assert(depth < MAX_IFF_DEPTH);

   parent_chunk_offset = DUMP_OFFSET(plist[ depth - 1 ], dump_pos);
   p_ch_size = get_chunk_size(plist[ depth - 1 ]);

   if (old_dump_size != (parent_chunk_offset + p_ch_size))
   {
      return NO_DUMP_POS;
   }

   assert(ALIGNED(size, 4));

   if ((old_dump_size + size) > DUMP_AREA_SIZE)
   {
      return NO_DUMP_POS;
   }

   /*
    * New chunk size must be at least a minimal chunk,
    * or a minimal extension of the parent chunk,
    * This corrects bug #3193, incomplete dumps.
    */

   new_chunk_pos = DUMP_POS(dump_pos, old_dump_size);

   if (size >= sizeof(struct RamdumpChunkHeader))
   {
      /* First write a filler chunk */
      write_chunk_header(new_chunk_pos, RAMDUMP_TAG_FILL,
                         size - sizeof(struct RamdumpChunkHeader));
   }
   else
   {
      /* Fill with zeros */
      struct RamdumpChunkHeader phony_header;
      memset((void *) &phony_header, 0x00, sizeof(struct RamdumpChunkHeader));
      dump_area_write((const void *) &phony_header, new_chunk_pos, 0, size);
   }

   /*
    * Update chunk sizes from top chunk down to the parent chunk
    * of the new chunk
    */
   for (i = 0; i < depth; i++)
   {
      DumpOffset dump_offset = DUMP_OFFSET(plist[ i ], dump_pos);
      DumpPos tmp_pos;
      DumpOffset sz_off;
      U32 ch_size;

      assert(dump_offset <= parent_chunk_offset);

      tmp_pos = DUMP_POS(dump_pos, dump_offset);
      sz_off = dump_offset + OFFSET(struct RamdumpChunkHeader, size);
      ch_size = get_chunk_size(tmp_pos);

      /* Make rvct stop warning about ch_size */
      (void)ch_size;

      assert(ch_size <= old_dump_size);
      assert(U32_ALIGNED(ch_size));

      /* Update size */
      DUMP_VAL_U32(dump_pos, sz_off) = htonl(
            ntohl(DUMP_VAL_U32(dump_pos, sz_off)) + size);
   }

   /*
    * If we've updated the parent_chunk_size, all grandparent's must
    * have updated size as well, check the top chunk for example
    */
   assert(get_chunk_size(dump_pos) == (parent_chunk_offset +
                  get_chunk_size(DUMP_POS(dump_pos, parent_chunk_offset))));

   /* Return position of new chunk */
   return new_chunk_pos;
}

/**
 * @function add_text_string
 *
 * @brief Extends a dump with text string.
 *
 * @long Extends a dump with text data. Pads with zeros for proper alignment.
 *
 * @input dump Dump position where the text chunk will be added.
 * @input plist List of parent chunks.
 * @input depth Depth in chunk tree where the text chunk will be added.
 * @input id Chunk id tag for the text chunk.
 * @input text Text string to extend the dump with.
 *
 * @return Returns position of the new chunk or NO_DUMP_POS on failure.
 */
static DumpPos
add_text_string(DumpPos dump, const DumpPosList plist,
		int depth, IffId id, const char *text)
{
   const char *str = (text == NULL) ? "" : text;
   unsigned int text_len = strlen(str) + 1;
   unsigned int text_len_aligned = ALIGN_TO(text_len, sizeof(U32));
   unsigned int extend_size = sizeof(struct RamdumpChunkHeader) +
                              text_len_aligned;
   const U32 zero = 0;

   DumpOffset offset = 0;
   DumpPos tmp_pos = extend_dump(dump, plist, depth, extend_size);

   if (tmp_pos != NO_DUMP_POS) {
      write_chunk_header(tmp_pos, id, text_len_aligned);

      offset += sizeof(struct RamdumpChunkHeader);

      dump_area_write((const void *)str, tmp_pos, offset,
		      text_len); /* padding not copied */

      /* Copy zeros as padding */
      if (text_len_aligned - text_len > 0)
      {
         dump_area_write((const void *) &zero, tmp_pos,
                         offset + text_len,
                         text_len_aligned - text_len);
      }
   }

   return tmp_pos;
}


/**
 * @function add_description
 *
 * @brief Extends a dump with description info.
 *
 * @input dump Dump position where the dump will be extended.
 * @input plist List of parent chunks.
 * @input depth Depth in the chunk tree where the dump will be extended.
 * @input descr Text description to write to the dump.
 *
 * @return Returns the dump pos of the chunk or NO_DUMP_POS on failure.
 */
static DumpPos
add_description(DumpPos dump, const DumpPosList plist,
                int depth, const char *descr)
{
   return add_text_string(dump, plist, depth, RAMDUMP_TAG_DESC, descr);
}

/**
 * @function write_new_dump
 *
 * @brief Initializes a new dump.
 *
 * @input pos Dump position where the new dump starts.
 */
static void
write_new_dump(DumpPos pos)
{
   union
   {
      struct RamdumpChunkVERS ch_VERS;
      struct RamdumpChunkDPID ch_DPID;
      struct RamdumpHeaderLIST hd_LIST;
   } u;
   DumpPos tmp_pos;
   DumpOffset offset = 0;
   unsigned int extend_size;
   DumpPosList plist;
   int depth = 0;

   assert(VALID_DUMP_POS(pos));

   extend_size = sizeof(struct RamdumpChunkVERS) +
                 sizeof(struct RamdumpChunkDPID) +
                 sizeof(struct RamdumpHeaderLIST);

   /* Dump header */
   write_group_header(DUMP_POS(pos, offset), RAMDUMP_TAG_FORM,
                      sizeof(struct RamdumpHeaderFORM) -
                      sizeof(struct RamdumpChunkHeader),
                      RAMDUMP_TAG_PMD_);
   add_child_location(plist, depth++, pos);

   tmp_pos = extend_dump(pos, plist, depth, extend_size);

   assert(DUMP_OFFSET(tmp_pos, pos) == sizeof(struct RamdumpHeaderFORM));

   /* PMD version */
   u.ch_VERS.chd.id = htonl(RAMDUMP_TAG_VERS);
   u.ch_VERS.chd.size = htonl(sizeof(struct RamdumpChunkVERS) -
                              sizeof(struct RamdumpChunkHeader));
   u.ch_VERS.version = RAMDUMP_VERSION;

   dump_area_write((const void *) &u.ch_VERS, tmp_pos, offset,
                   sizeof(struct RamdumpChunkVERS));

   offset += sizeof(struct RamdumpChunkVERS);

   /* Dump id */
   u.ch_DPID.chd.id = htonl(RAMDUMP_TAG_DPID);
   u.ch_DPID.chd.size = htonl(sizeof(struct RamdumpChunkDPID) -
                              sizeof(struct RamdumpChunkHeader));
   u.ch_DPID.dump_id = CURR_DUMP_ID;

   dump_area_write((const void *) &u.ch_DPID, tmp_pos, offset,
                   sizeof(struct RamdumpChunkDPID));

   offset += sizeof(struct RamdumpChunkDPID);

   /* Empty block list */
   u.hd_LIST.chd.id = htonl(RAMDUMP_TAG_LIST);
   u.hd_LIST.chd.size = htonl(sizeof(struct RamdumpHeaderLIST) -
                              sizeof(struct RamdumpChunkHeader));
   u.hd_LIST.cont_type = htonl(RAMDUMP_TAG_BLOC);

   dump_area_write((const void *)&u.hd_LIST, tmp_pos, offset,
                   sizeof(struct RamdumpHeaderLIST));

   update_dump_index(find_dump_id(pos), pos);
}

/**
 * @function get_most_recent_dump
 *
 * @brief Gets the start position of the most recent dump.
 *
 * @long Gets the start position of the most recent dump, even current
 * if it contains any information.
 *
 * @input dump_id Dump id of the current dump.
 *
 * @return Returns the dump position of the most recent dump, or
 * NO_DUMP_POS on failure.
 */
static DumpPos
get_most_recent_dump(RamdumpDumpId dump_id)
{
   unsigned int i;
   U32 min_dist = 0xffffffff;
   DumpPos most_recent_dump_entry = NO_DUMP_POS;

   /*
    * The dump table is updated every time a block is written,
    * but when the dump area is completely fresh, the table is not consistent
    * Scan the table to see if any entry is not consistent
    */
   if (!is_dump_index_consistent())
   {
      build_dump_index();
   }

   /* Back up sequence id to find the latest valid dump */
   for (i = 0; i < DUMP_TABLE_SIZE; i++)
   {
      DumpPos pos = DUMP_TABLE[ HASH_DUMP_ID(dump_id - i) ];

      if (VALID_DUMP_POS(pos) && is_dump_header(pos) && is_valid_dump(pos))
      {
         RamdumpDumpId this_dump_id = find_dump_id(pos);
         U32 dist = dump_id - this_dump_id;

         /* Update minimum distance */
         if (dist < min_dist)
         {
            min_dist = dist;
            most_recent_dump_entry = pos;

            /* Stop if distance optimal, a very common case */
            if (dist == i)
            {
               break;
            }
         }
      }
   }

   if (min_dist < 0x80000000UL)
   {
      /* Return a valid dump */
      assert(is_valid_dump(most_recent_dump_entry));
      return most_recent_dump_entry;
   }
   else
   {
      return NO_DUMP_POS;
   }
}

/**
 * @function get_dump_pos
 *
 * @brief Locates and returns the start position of a dump.
 *
 * @long Locates and returns the start position of a dump, even current
 * if it contains any information.
 *
 * @input dump_id Dump id of the dump to find the start of.
 *
 * @return Returns the start position of the dump, or NO_DUMP_POS
 * on failure.
 */
static DumpPos
get_dump_pos(RamdumpDumpId dump_id)
{
   DumpPos pos;

   /*
    * The dump table is updated every time a block is written,
    * but when the dump area is completely fresh, the table is not consistent
    * Scan the table to see if any entry is not consistent
    */
   if (!is_dump_index_consistent())
   {
      build_dump_index();
   }

   pos = DUMP_TABLE[ HASH_DUMP_ID(dump_id) ];

   if (VALID_DUMP_POS(pos) && is_dump_header(pos) &&
       is_valid_iff_structure_depth(pos, 2) && find_dump_id(pos) == dump_id)
   {
      return pos;
   }

   return NO_DUMP_POS;
}

/**
 * @function get_dump_to_extend
 *
 * @brief Finds the dump to extend with new information.
 *
 * @return Returns dump position for dump that should be extended.
 */
static DumpPos
get_dump_to_extend(void)
{
   DumpPos dump_pos;
   RamdumpDumpId curr_dump_id = CURR_DUMP_ID;

   /* Find most recent dump */
   dump_pos = get_most_recent_dump(curr_dump_id);

   /* assert(!VALID_DUMP_POS(dump_pos) || is_valid_dump(dump_pos)) */

   if (VALID_DUMP_POS(dump_pos))
   {
      RamdumpDumpId dump_id = find_dump_id(dump_pos);

      if (dump_id == curr_dump_id)
      {
         /* No more space for this dump? */
         if (get_chunk_size(dump_pos) >= DUMP_AREA_SIZE)
         {
            return NO_DUMP_POS;
         }
      }
      else
      {
         /* Not current dump, we have to start a new one */
         dump_pos = skip_dump(dump_pos);
         write_new_dump(dump_pos);
      }
   }
   else
   {
      /* Start dump from beginning of dump area */
      dump_pos = 0;
      write_new_dump(dump_pos);
   }

   assert(VALID_DUMP_POS(dump_pos));

   return dump_pos;
}

/*
 * ==========================================================================
 *
 * Block functionality.
 *   - Adding a new block header.
 * ==========================================================================
 */

static int
store_dump_sizes(DumpPos dump)
{
   const IffId bloc_list[] = {
      RAMDUMP_TAG_FORM, RAMDUMP_TAG_PMD_, RAMDUMP_TAG_LIST, RAMDUMP_TAG_BLOC, 0
   };

   DumpPos pos;

   /* Store size of FORM/PMD chunk */
   if (get_chunk_id(dump) == RAMDUMP_TAG_FORM
       && get_group_contents_id(dump) == RAMDUMP_TAG_PMD_)
   {
      stored_dump_sizes[0].pos = dump;
      stored_dump_sizes[0].size = DUMP_VAL_U32(dump,
                                               OFFSET(struct RamdumpChunkHeader,
                                                      size));
   }
   else
   {
      return 0;
   }

   /* Store size of LIST/BLOC chunk. */
   pos = find_chunk_in_valid_iff(dump, bloc_list);
   if (pos != NO_DUMP_POS)
   {
      stored_dump_sizes[1].pos = pos;
      stored_dump_sizes[1].size = DUMP_VAL_U32(pos,
                                               OFFSET(struct RamdumpChunkHeader,
                                               size));
   }
   else
   {
      return 0;
   }

   return 1;
}

static int
restore_dump_sizes(void)
{
   if (get_chunk_id(stored_dump_sizes[0].pos) == RAMDUMP_TAG_FORM
       && get_group_contents_id(stored_dump_sizes[0].pos) == RAMDUMP_TAG_PMD_)
   {
      /* Restore size of FORM/PMD chunk */
      DUMP_VAL_U32(stored_dump_sizes[0].pos,
                   OFFSET(struct RamdumpChunkHeader, size))
         = stored_dump_sizes[0].size;
   }
   else
   {
      return 0;
   }


   if (get_chunk_id(stored_dump_sizes[1].pos) == RAMDUMP_TAG_LIST
       && get_group_contents_id(stored_dump_sizes[1].pos) == RAMDUMP_TAG_BLOC)
   {
      /* Restore size of LIST/BLOCK chunk */
      DUMP_VAL_U32(stored_dump_sizes[1].pos,
                   OFFSET(struct RamdumpChunkHeader, size))
         = stored_dump_sizes[1].size;
   }
   else
   {
      return 0;
   }

   return 1;
}

/**
 * @function write_new_block_header
 *
 * @brief Prepares a new block in the dump area.
 *
 * @long Prepares a new block in the dump area. Fills in the parent
 * chunk positions and length of parent list, so that the new block
 * may be extended using this information.
 *
 * @input pos Position to write the new block at.
 * @input plist List of parent chunks.
 * @output depth The depth where the chunk was written.
 *
 * @return Returns the dump position where the chunk was written,
 * or NO_DUMP_POS on failure.
 */
static DumpPos
write_new_block_header(DumpPos pos, DumpPosList plist,
                       int *depth)
{
   const IffId bloc_list[] = {
      RAMDUMP_TAG_FORM, RAMDUMP_TAG_PMD_, RAMDUMP_TAG_LIST, RAMDUMP_TAG_BLOC, 0
   };

   union
   {
      struct RamdumpChunkBLHD ch_BLHD;
   } u;

   DumpPos tmp_pos;
   DumpOffset offset = 0;
   unsigned int extend_size;

   OSTICK ticks, uticks;
   OSTICK abs_time, ref_ticks, ref_uticks;
   OSTICK secs, microseconds;
   OSTICK tick_len = system_tick();

/* Remove the 'if' */
   assert(VALID_DUMP_POS(pos));

   extend_size = sizeof(struct RamdumpHeaderFORM) +
                 sizeof(struct RamdumpChunkBLHD);

   if (tick_len == 0) tick_len = 1000;

   /* Get tick */
   ticks = get_systime(&uticks);

   /* Get abs. time */
   get_abstime(&abs_time, &ref_ticks, &ref_uticks);

   if (uticks < ref_uticks)
   {
      uticks += tick_len;
      ticks--;
   }

   {
      OSTICK delta = (ticks - ref_ticks);
      OSTICK mdelta = delta / 1000000;
      OSTICK rdelta = delta % 1000000;
      OSTICK tmp = rdelta * tick_len + uticks - ref_uticks;
      secs = abs_time + mdelta * tick_len + tmp / 1000000;
      microseconds = tmp % 1000000;
   }

   do
   {
      *depth = 0;

      /* Extend dump */
      add_child_location(plist, (*depth)++, pos);

      add_child_location(plist, (*depth)++,
                         find_chunk_in_valid_iff(pos, bloc_list));

      tmp_pos = extend_dump(pos, plist, *depth, extend_size);

      /* No more space in dump for block ? */
      if (!VALID_DUMP_POS(tmp_pos))
      {
         break;
      }

      /* Dump header */
      write_group_header(tmp_pos, RAMDUMP_TAG_FORM,
                         extend_size - sizeof(struct RamdumpChunkHeader),
                         RAMDUMP_TAG_BLOC);
      add_child_location(plist, (*depth)++, tmp_pos);

      offset += sizeof(struct RamdumpHeaderFORM);

      /* Block header */
      u.ch_BLHD.chd.id = htonl(RAMDUMP_TAG_BLHD);
      u.ch_BLHD.chd.size = htonl(sizeof(struct RamdumpChunkBLHD) -
                                 sizeof(struct RamdumpChunkHeader));
      u.ch_BLHD.dump_id = CURR_DUMP_ID;
      u.ch_BLHD.block_no = CURR_DUMP_BLOCK_COUNT++;
      u.ch_BLHD.clock = secs;
      u.ch_BLHD.microseconds = microseconds;

      dump_area_write((const void *)&u.ch_BLHD, tmp_pos, offset,
                      sizeof(struct RamdumpChunkBLHD));
   }
   while (0);

   return tmp_pos;
}

/**
 * @function advance_dump_id
 *
 * @brief Starts a new dump.
 *
 * @long Starts a new dump. Successive dumped blocks will start a new dump.
 *
 * @return Returns zero
 */
static long
advance_dump_id(void)
{
   int i;
   U32 curr_dump_id_range = CURR_DUMP_ID;

   /* TOOLSCR-2491: The ramdump id limits in [0x00000000 to 0x7FFFFFFF] */
   CURR_DUMP_ID = (curr_dump_id_range + 1) & 0x7FFFFFFFUL;

   /* Skip sequence identifier zero */
   if (CURR_DUMP_ID == 0)
   {
      CURR_DUMP_ID++;
   }

   CURR_DUMP_BLOCK_COUNT = 0;

   assert(CURR_DUMP_ID != 0 && CURR_DUMP_ID < 0x80000000UL);

   /* Clear last chunk tracker */
   last_chunk_depth = 0;
   for (i = 0; i < MAX_IFF_DEPTH; i++)
   {
      last_chunk_plist[i] = 0;
   }

   return 0;
}


/**
 * @function dump_error_info
 *
 * @brief Writes a block with information about an error event.
 *
 * @input description Text description of error.
 * @input user_called Value of user_called flag.
 * @input error_code Error code.
 * @input extra Extra parameter.
 *
 * @output ret_code Return code is filled in with one of the following:
 * RAMDUMP_SUCCESS, RAMDUMP_ECURR_DUMP_TOO_LARGE,
 * RAMDUMP_EMEMORY_ACCESS_NOK or RAMDUMP_EUNABLE_TO_GET_CURRENT_PROCESS_INFO.
 *
 * @return Returns zero.
 */
static long
dump_error_info(const char *description,
                OSBOOLEAN user_called, OSERRCODE error_code,
                OSERRCODE extra, RamdumpStatus *ret_code)
{
   RamdumpStatus return_code = RAMDUMP_SUCCESS;
   DumpPos dump = NO_DUMP_POS;
   DumpPosList plist;
   int depth;
   struct EfmiCurrent current;
   EfmiStatus efmi_status;

   do
   {
      struct RamdumpChunkERIN ch_ERIN;
      DumpPos block_pos, erbl_pos, descr_pos;
      DumpOffset offset = 0;
      U32 extend_size;

      /* Interrupts MUST be disabled when getting the current process. */
      efmi_status = ose_efmi_get_current(&current, sizeof(struct EfmiCurrent));

      if (efmi_status != EFMI_OK)
      {
         return_code = RAMDUMP_EUNABLE_TO_GET_CURRENT_PROCESS_INFO;
         current.type = 0;
         current.ni_process = 0;
         current.segid = 0;
         current.segment = 0;
         current.block = 0;
         current.process = 0;
      }

      dump = get_dump_to_extend();

      if (dump == NO_DUMP_POS)
      {
          return_code = RAMDUMP_ECURR_DUMP_TOO_LARGE;
          break;
      }

      if (!store_dump_sizes(dump))
      {
         return_code = RAMDUMP_EPOS_OR_DUMP_NOT_AVAILABLE;
         break;
      }

      /* Write header */
      block_pos = write_new_block_header(dump, plist, &depth);

      if (block_pos == NO_DUMP_POS)
      {
          restore_dump_sizes();
          return_code = RAMDUMP_ECURR_DUMP_TOO_LARGE;
          break;
      }

      /* Is there any easier way to find out the size? */
      extend_size = sizeof(struct RamdumpHeaderFORM) +
                    sizeof(struct RamdumpChunkERIN);

      erbl_pos = extend_dump(dump, plist, depth, extend_size);

      if (erbl_pos == NO_DUMP_POS)
      {
          restore_dump_sizes();
          return_code = RAMDUMP_ECURR_DUMP_TOO_LARGE;
          break;
      }

      /* Form header */
      write_group_header(erbl_pos, RAMDUMP_TAG_FORM,
                         extend_size - sizeof(struct RamdumpChunkHeader),
                         RAMDUMP_TAG_ERBL);
      add_child_location(plist, depth++, erbl_pos);

      offset += sizeof(struct RamdumpHeaderFORM);

      /* Error info */
      ch_ERIN.chd.id = htonl(RAMDUMP_TAG_ERIN);
      ch_ERIN.chd.size = htonl(sizeof(struct RamdumpChunkERIN) -
                               sizeof(struct RamdumpChunkHeader));
      ch_ERIN.user_called = user_called;
      ch_ERIN.error_code = error_code;
      ch_ERIN.extra = extra;
      ch_ERIN.curr_proc = (PROCESS)current.process;

      dump_area_write((const void *) &ch_ERIN, erbl_pos, offset,
                      sizeof(struct RamdumpChunkERIN));

      /* Description */
      descr_pos = add_description(dump, plist, depth, description);

      if (descr_pos == NO_DUMP_POS)
      {
          restore_dump_sizes();
          return_code = RAMDUMP_ECURR_DUMP_TOO_LARGE;
          break;
      }
   } while (0); /* Just to use break; to avoid ugly nesting */

   /* Update dump index */
   if (dump != NO_DUMP_POS)
   {
      update_dump_index(find_dump_id(dump), dump);
   }

   if (ret_code)
   {
      *ret_code = return_code;
   }

   return 0;
}

/**
 * @function dump_text_info
 *
 * @brief Writes a block with a user-provided textual description.
 *
 * @input description Text description.
 *
 * @output ret_code Return code is filled in with one of the following:
 * RAMDUMP_SUCCESS, RAMDUMP_ECURR_DUMP_TOO_LARGE, or
 * RAMDUMP_EMEMORY_ACCESS_NOK.
 *
 * @return Returns zero.
 */
static long
dump_text_info(const char *description, RamdumpStatus *ret_code)
{
   RamdumpStatus return_code = RAMDUMP_ECURR_DUMP_TOO_LARGE;
   DumpPos dump;
   DumpPosList plist;
   int depth;

   dump = get_dump_to_extend();

   (void) return_code; /* Lint */

   do
   {
      DumpPos block_pos, txbl_pos, pos;
      U32 extend_size;

      if (dump == NO_DUMP_POS)
      {
         return_code = RAMDUMP_ECURR_DUMP_TOO_LARGE;
         break;
      }

      if (!store_dump_sizes(dump))
      {
         return_code = RAMDUMP_EPOS_OR_DUMP_NOT_AVAILABLE;
         break;
      }

      /* Write header */
      block_pos = write_new_block_header(dump, plist, &depth);

      if (block_pos == NO_DUMP_POS)
      {
         restore_dump_sizes();
         return_code = RAMDUMP_ECURR_DUMP_TOO_LARGE;
         break;
      }

      /* Is there any easier way to find out the size? */
      extend_size = sizeof(struct RamdumpHeaderFORM);

      txbl_pos = extend_dump(dump, plist, depth, extend_size);

      if (txbl_pos == NO_DUMP_POS)
      {
         restore_dump_sizes();
         return_code = RAMDUMP_ECURR_DUMP_TOO_LARGE;
         break;
      }

      /* Form header */
      write_group_header(txbl_pos, RAMDUMP_TAG_FORM,
                         extend_size - sizeof(struct RamdumpChunkHeader),
                         RAMDUMP_TAG_TXBL);

      add_child_location(plist, depth++, txbl_pos);

      /* Description */
      pos = add_description(dump, plist, depth, description);

      if (pos == NO_DUMP_POS)
      {
         restore_dump_sizes();
         return_code = RAMDUMP_ECURR_DUMP_TOO_LARGE;
         break;
      }

      return_code = RAMDUMP_SUCCESS;
   } while (0); /* Just to use break; to avoid ugly nesting */

   /* Update dump index */
   if (dump != NO_DUMP_POS)
   {
      update_dump_index(find_dump_id(dump), dump);
   }

   if (ret_code)
   {
      *ret_code = return_code;
   }

   return 0;
}

/**
 * @function ramdump_kick_watchdog
 *
 * @brief Calls the currently set ramdump watchdog callback function.
 */
void
ramdump_kick_watchdog(void)
{
   if (watchdog_cb != NULL)
   {
      (*watchdog_cb)();
   }
}

#ifdef RAMDUMP_USE_ZLIB
/**
 * @function ramdump_zlib_alloc
 *
 * @brief Wrapper for ramdump_heap_malloc in order to provide zlib with a
 * memory allocation function that conforms to the zlib function prototype.
 *
 * @input opaque Unusued parameter.
 * @input items Number of items to allocate.
 * @size Size of each item.
 *
 * @return Returns a pointer to allocated memory or NULL on failure.
 */
static voidpf
ramdump_zlib_alloc(voidpf opaque, uInt items, uInt size)
{
   voidpf buffer;
   unsigned long buffer_size = items * size;

   (void)opaque;
   assert(buffer_size <= 0x10000); /* 64 K */

   buffer = (voidpf)ramdump_heap_malloc(buffer_size);

   return buffer;
}
#endif

#ifdef RAMDUMP_USE_ZLIB
/**
 * @function ramdump_zlib_free
 *
 * @brief Wrapper for ramdump_heap_free in order to provide zlib with a memory
 * deallocation function that conforms to the zlib function prototype.
 *
 * @long Wrapper for ramdump_free in order to provide zlib with a memory
 * deallocation function that conforms to the zlib function prototype.
 *
 * @input opaque Unused parameter.
 * @input address Pointer to memory area to be freed.
 */
static void
ramdump_zlib_free(voidpf opaque, voidpf address)
{
   (void)opaque;
   ramdump_heap_free(address);
}
#endif

/**
 * @function add_memory_image_data
 *
 * @brief Extends a dump with memory image data.
 *
 * @input dump Dump position where the dump will be extended.
 * @input plist List of parent chunks.
 * @input depth Depth in chunk tree of chunk.
 * @input start_address Start address of memory area to dump.
 * @input length Size of memory area to dump in bytes.
 *
 * @return Returns dump position of chunk or NO_DUMP_POS on failure.
 */
static DumpPos
add_memory_image_data(DumpPos dump, DumpPosList plist, int depth,
		      OSADDRESS start_address, U32 length)
{
   DumpPos pos;
   unsigned int extend_size;

#ifdef RAMDUMP_USE_ZLIB

   /* Compress with zlib utilities */
   do
   {
      struct RamdumpChunkZDEF ch_ZDEF;
      DumpPos zdat_pos, image_pos;
      DumpOffset zdef_offset;
      int i;
      extend_size = sizeof(struct RamdumpHeaderFORM) +
                    sizeof(struct RamdumpChunkZDEF);

      /* ZData form */
      pos = extend_dump(dump, plist, depth, extend_size);

      if (pos == NO_DUMP_POS)
      {
         break;
      }

      /* Remember position of this chunk */
      zdat_pos = pos;

      write_group_header(zdat_pos, RAMDUMP_TAG_FORM,
                         extend_size - sizeof(struct RamdumpChunkHeader),
                         RAMDUMP_TAG_ZDAT);

      zdef_offset = sizeof(struct RamdumpHeaderFORM);

      /* Compression header */
      ch_ZDEF.chd.id = htonl(RAMDUMP_TAG_ZDEF);
      ch_ZDEF.chd.size = htonl(sizeof(struct RamdumpChunkZDEF) -
                               sizeof(struct RamdumpChunkHeader));

      ch_ZDEF.compression = RAMDUMP_COMPRESS_ZLIB_Z77;
      for (i = 0; i < 4; i++)
      {
         ch_ZDEF.compr_parameters[ i ] = 0;
      }
      ch_ZDEF.image_length = 0; /* Will be updated later */
      ch_ZDEF._reserved = 0;

      /* Write compression header */
      dump_area_write((const void *) &ch_ZDEF, zdat_pos, zdef_offset,
                      sizeof(struct RamdumpChunkZDEF));
      add_child_location(plist, depth++, zdat_pos);

      /* Write empty data chunk */
      extend_size = sizeof(struct RamdumpChunkHeader);

      pos = extend_dump(dump, plist, depth, extend_size);

      if (pos == NO_DUMP_POS)
      {
         break;
      }

      /* Remember position of image data */
      image_pos = pos;

      write_chunk_header(pos, RAMDUMP_TAG_DATA, 0);
      add_child_location(plist, depth++, image_pos);

      /* Compress memory image */
      {
         /* ZLIB stream object */
         z_stream zstrm;
         U32 mem_count = 0;
         U32 z_count = 0;
         int ret;
         int windowBits, memLevel;

         /* Provide post-mortem safe memory allocation routines for ZLIB */
         zstrm.zalloc = &ramdump_zlib_alloc;
         zstrm.zfree = &ramdump_zlib_free;
         zstrm.opaque = (voidpf) 0;

      	/* Initialize compression (Commented out in old PMD too) */
      	/* ret = deflateInit(&zstrm, Z_BEST_SPEED); */
      	/* ret = deflateInit(&zstrm, Z_DEFAULT_COMPRESSION); */
      	/* ret = deflateInit(&zstrm, Z_BEST_COMPRESSION); */

         /* Memory requirements for deflate:
          * (1 << (windowBits+2)) + (1 << (memLevel+9)) + a few KB for small
          * objects
          */
         windowBits = 15;
         memLevel = 8;
         /* Assert loop stop */
         while (windowBits > 0 && memLevel > 0 &&
                (1 << (windowBits + 2)) + (1 << (memLevel + 9)) + 0x2000 >=
                (int)(ramdump_work_area_length -
                sizeof(struct ZlibBuffers))/2)
         {
            windowBits--;
            memLevel--;
         }

	      assert(windowBits >= 8 && windowBits <= 15);
         assert(memLevel >= 1 && memLevel <= 9);

         /* Kick watchdog */
         ramdump_kick_watchdog();

         ret = deflateInit2(&zstrm,
                            Z_BEST_SPEED,
                            Z_DEFLATED,
                            windowBits,
                            memLevel,
                            Z_DEFAULT_STRATEGY);

         /* Init OK? */
         if (ret != Z_OK)
         {
            assert(0);
            break;
         }

         /* Initialize stream input and output data */
         if ((length - mem_count) > ZLIB_COMPR_IN_SIZE)
         {
            zstrm.avail_in = ZLIB_COMPR_IN_SIZE;
         }
         else
         {
            zstrm.avail_in = length - mem_count;
         }
         (void)memmove((void *) ramdump_zlib_buffers->in_buf,
                       (const void *) (start_address + mem_count),
                       zstrm.avail_in);
         zstrm.next_in = (Bytef *) ramdump_zlib_buffers->in_buf;
         mem_count += zstrm.avail_in;
         assert(mem_count <= length);

         /* Consider pm_alloc() in the future */
         zstrm.next_out = (Bytef *) ramdump_zlib_buffers->out_buf;
         zstrm.avail_out = ZLIB_COMPR_OUT_SIZE;

         /* Compress input iteratively */
         while ((length - mem_count) > 0)
         {
            /* Kick watchdog */
            ramdump_kick_watchdog();

            /* Compress some input */
            ret = deflate(&zstrm, Z_NO_FLUSH);

            /* Something wrong? */
            if (ret != Z_OK)
            {
               assert(ret != Z_STREAM_END);
               break;
            }

            /* Check status */
            if (zstrm.avail_in == 0)
            {
               U32 avail_in;
               /* Need more input */

               /* Update input data */
               if ((length - mem_count) > ZLIB_COMPR_IN_SIZE)
               {
                  avail_in = ZLIB_COMPR_IN_SIZE;
               }
               else
               {
                  avail_in = length - mem_count;
               }
               (void) memmove((void *) ramdump_zlib_buffers->in_buf,
			                     (const void *)(start_address + mem_count),
			                     avail_in);
	            zstrm.next_in = (Bytef *)ramdump_zlib_buffers->in_buf;
               zstrm.avail_in = avail_in;
               mem_count += avail_in;
               assert(mem_count <= length);
            }

            /* Check if output buffer has been filled */
            if (zstrm.avail_out == 0)
            {

               /* Write this piece of compressed data to dump */
               extend_size = ZLIB_COMPR_OUT_SIZE;

               assert(U32_ALIGNED(extend_size));
               pos = extend_dump(dump, plist, depth, extend_size);

               if (pos == NO_DUMP_POS)
               {
                  break;
               }

               /* Copy memory to dump */
               dump_area_write((const void *)ramdump_zlib_buffers->out_buf,
                               image_pos,
                               sizeof(struct RamdumpChunkHeader) + z_count,
                               ZLIB_COMPR_OUT_SIZE);

               /* Update size of image */
               z_count += ZLIB_COMPR_OUT_SIZE;

               /* Update output data */
               zstrm.next_out = (Bytef *) ramdump_zlib_buffers->out_buf;
               zstrm.avail_out = ZLIB_COMPR_OUT_SIZE;
            }

            assert(zstrm.avail_in > 0 && zstrm.avail_out > 0);
         }


         /* We are here if */
         /* 1. we have provided all input to deflate, length == mem_count */
         /* 2. deflate returned != Z_OK or Z_STREAM_END (error) */
         /* 3. dump could not be extended. (pos == NO_DUMP_POS) */
         if (length == mem_count)
         {

            assert((zstrm.total_in + zstrm.avail_in) == mem_count);

            do
            {
               /* Kick watchdog */
               ramdump_kick_watchdog();

               /* Tell deflate to flush any data it has left */
               ret = deflate(&zstrm, Z_FINISH);

               if (ret == Z_OK || ret == Z_STREAM_END)
               {

                  /* Write this piece of compressed data to dump */
                  extend_size = ALIGN_TO(ZLIB_COMPR_OUT_SIZE - zstrm.avail_out,
                                         sizeof(U32));

                  assert(U32_ALIGNED(extend_size));
                  pos = extend_dump(dump, plist, depth, extend_size);

                  if (pos == NO_DUMP_POS)
                  {
                     break;
                  }

                  /* Copy memory to dump */
                  dump_area_write((const void *) ramdump_zlib_buffers->out_buf,
                                  image_pos,
                                  sizeof(struct RamdumpChunkHeader) + z_count,
                                  ZLIB_COMPR_OUT_SIZE - zstrm.avail_out);

                  /* Update size of image */
                  z_count += ZLIB_COMPR_OUT_SIZE - zstrm.avail_out;

                  /* Update output data */
                  zstrm.next_out = (Bytef *) ramdump_zlib_buffers->out_buf;
                  zstrm.avail_out = ZLIB_COMPR_OUT_SIZE;
               }
            } while (ret == Z_OK);

            /* Compressed count should be same as deflate count */
            assert(!(pos != NO_DUMP_POS && ret == Z_STREAM_END &&
                   zstrm.total_out != z_count));
         }

         /* 2. deflate returned != Z_OK or Z_STREAM_END (error) */
         if (!(ret == Z_OK || ret == Z_STREAM_END))
         {
            assert(0);
         }

         /* 3. dump could not be extended. (pos == NO_DUMP_POS) */

         /* Deflate has some internal memory storage that must be freed */
         (void) deflateEnd(&zstrm);

         /* Update written image length */
         DUMP_VAL_U32(zdat_pos, zdef_offset +
                      OFFSET(struct RamdumpChunkZDEF, image_length)) = z_count;
      }

      /* Return position of this data block, compressed or not */
      if (pos != NO_DUMP_POS)
      {
         pos = zdat_pos;
      }
   } while (0);

#else

   /* No compression */
   do
   {
      U32 data_len = ALIGN_TO(length, sizeof(U32));
      extend_size = sizeof(struct RamdumpChunkHeader) + data_len;

      /* Data image */
      pos = extend_dump(dump, plist, depth, extend_size);

      if (pos == NO_DUMP_POS)
      {
         break;
      }

      write_chunk_header(pos, RAMDUMP_TAG_DATA, data_len);

      /* Kick watchdog */
      ramdump_kick_watchdog();

      /* Copy memory to dump */
      dump_area_write((const void *) start_address, pos,
                      sizeof(struct RamdumpChunkHeader), length);

   } while (0);

#endif

   return pos;
}

/**
 * @function ramdump_dump_memory_image2
 *
 * @long Writes a block containing a memory image of specified length,
 * starting from the specified memory address.
 *
 * @input dump Dump position to write to.
 * @input description Text description of memory area.
 * @input start_address Start address of memory area to dump.
 * @input length Size of memory area.
 *
 * @output ret_code Return code is filled in with RAMDUMP_SUCCESS,
 * RAMDUMP_ECURR_DUMP_TOO_LARGE or RAMDUMP_EMEMORY_ACCESS_NOK.
 */
void
ramdump_dump_memory_image2(DumpPos dump, const char *description,
                           OSADDRESS start_address, U32 length,
                           RamdumpStatus *ret_code)
{
   RamdumpStatus return_code = RAMDUMP_ECURR_DUMP_TOO_LARGE;
   DumpPosList plist;
   int depth;

   (void) return_code; /* Lint */

   do
   {
      struct RamdumpChunkMHD_ ch_MHD_;
      DumpPos block_pos, mbl_pos, mhd_pos, pos;
      U32 extend_size;

      if (dump == NO_DUMP_POS)
      {
         return_code = RAMDUMP_ECURR_DUMP_TOO_LARGE;
         break;
      }

      if (!store_dump_sizes(dump))
      {
         return_code = RAMDUMP_EPOS_OR_DUMP_NOT_AVAILABLE;
         break;
      }

      /* Write header */
      block_pos = write_new_block_header(dump, plist, &depth);

      if (block_pos == NO_DUMP_POS)
      {
         restore_dump_sizes();
         return_code = RAMDUMP_ECURR_DUMP_TOO_LARGE;
         break;
      }

      /* Is there any easier way to find out the size? */
      extend_size = sizeof(struct RamdumpHeaderFORM);

      mbl_pos = extend_dump(dump, plist, depth, extend_size);

      if (mbl_pos == NO_DUMP_POS)
      {
         restore_dump_sizes();
         return_code = RAMDUMP_ECURR_DUMP_TOO_LARGE;
         break;
      }

      /* Form header */
      write_group_header(mbl_pos, RAMDUMP_TAG_FORM,
                         sizeof(struct RamdumpHeaderFORM) -
                         sizeof(struct RamdumpChunkHeader),
                         RAMDUMP_TAG_MBL_);
      add_child_location(plist, depth++, mbl_pos);

      /* Description */
      pos = add_description(dump, plist, depth, description);

      if (pos == NO_DUMP_POS)
      {
         restore_dump_sizes();
         return_code = RAMDUMP_ECURR_DUMP_TOO_LARGE;
         break;
      }

      /* Memory image header */
      extend_size = sizeof(struct RamdumpChunkMHD_);
      mhd_pos = extend_dump(dump, plist, depth, extend_size);

      if (mhd_pos == NO_DUMP_POS)
      {
         return_code = RAMDUMP_ECURR_DUMP_TOO_LARGE;
         break;
      }

      ch_MHD_.chd.id = htonl(RAMDUMP_TAG_MHD_);
      ch_MHD_.chd.size = htonl(sizeof(struct RamdumpChunkMHD_) -
                               sizeof(struct RamdumpChunkHeader));
      ch_MHD_.start_address = start_address;
      /* Delay update of memory area length to avoid dump inconsistency */
      ch_MHD_.length = 0;

      dump_area_write((const void *) &ch_MHD_, mhd_pos, 0,
                      sizeof(struct RamdumpChunkMHD_));

      pos = add_memory_image_data(dump, plist, depth, start_address, length);

      if (pos == NO_DUMP_POS)
      {
         restore_dump_sizes();
         return_code = RAMDUMP_ECURR_DUMP_TOO_LARGE;
         break;
      }

      /*
       * Delay update of memory area length to avoid dump inconsistency,
       * with compressed image length. 990824, bjjo
       */
      DUMP_VAL_U32(mhd_pos, OFFSET(struct RamdumpChunkMHD_, length)) = length;

      return_code = RAMDUMP_SUCCESS;
   } while (0); /* Just to use break; to avoid uggly nesting */

   /* Update dump index */
   if (dump != NO_DUMP_POS)
   {
      update_dump_index(find_dump_id(dump), dump);
   }

   if (ret_code)
   {
      *ret_code = return_code;
   }
}

/**
 * @function dump_memory_image
 *
 * @brief Writes a block containing a memory image of specified length
 * starting from the specified memory address.
 *
 * @input description Text description of memory area.
 * @input start_address Start address of memory area to dump.
 * @input length Size of memory area to dump.
 *
 * @output ret_code The return code is filled in with one of the following:
 * RAMDUMP_SUCCESS, RAMDUMP_ECURR_DUMP_TOO_LARGE or
 * RAMDUMP_EMEMORY_ACCESS_NOK
 *
 * @return Returns zero.
 */
static long
dump_memory_image(const char *description, OSADDRESS start_address, U32 length,
                  RamdumpStatus *ret_code)
{
   DumpPos dump;

   dump = get_dump_to_extend();

   if (dump == NO_DUMP_POS)
   {
      if (ret_code)
      {
         *ret_code = RAMDUMP_ECURR_DUMP_TOO_LARGE;
      }
   }
   else
   {
      ramdump_dump_memory_image2(dump, description, start_address,
                                 length, ret_code);
   }

   return 0;
}

#ifdef RAMDUMP_USE_ZLIB

static int
open_compression_stream(z_stream *zstrm)
{
	int ret;
	int windowBits, memLevel;

	assert(zstrm != NULL);

	/* Provide post-mortem safe memory allocation routines for ZLIB */
	zstrm->zalloc = &ramdump_zlib_alloc;
	zstrm->zfree = &ramdump_zlib_free;
	zstrm->opaque = (voidpf) 0;

	/* Memory requirements for deflate:
	 * (1 << (windowBits+2)) + (1 << (memLevel+9)) + a few KB for small
	 * objects
	 */
	windowBits = 15;
	memLevel = 8;
	/* Assert loop stop */
	while (windowBits > 0 && memLevel > 0 &&
		   (1 << (windowBits + 2)) + (1 << (memLevel + 9)) + 0x2000 >=
		   (int)(ramdump_work_area_length -
		   sizeof(struct ZlibBuffers))/2)
	{
		windowBits--;
		memLevel--;
	}

   assert(windowBits >= 8 && windowBits <= 15);
   assert(memLevel >= 1 && memLevel <= 9);

	/* Open compression stream. */
	ret = deflateInit2(zstrm, Z_BEST_SPEED, Z_DEFLATED, windowBits,
					   memLevel, Z_DEFAULT_STRATEGY);

	/* Tell zlib where to find the cleared output buffer
	   and what size it is. */
	zstrm->avail_out = ZLIB_COMPR_OUT_SIZE;
	zstrm->next_out = (Bytef *)ramdump_zlib_buffers->out_buf;

	return ret;
}

static void
close_compression_stream(z_stream *zstrm)
{
	assert(zstrm != NULL);

	if (zstrm != NULL)
	{
		/* Close compression stream. */
		(void) deflateEnd(zstrm);
	}
}

static U32
refill_input_buffer(z_stream *zstrm, OSADDRESS start_address,
                    U32 length, U32 left_to_proc)
{
   U32 pad = ALIGN_TO(length, sizeof(U32)) - length;
   U32 bytes_left = left_to_proc;
   U32 non_pad_left = 0;

   assert(zstrm != NULL);

   /* Is there more data left to process than can fit in input buffer? */
   if (bytes_left > ZLIB_COMPR_IN_SIZE)
   {
      /* If so, limit the amount of data to process
         in this pass to size of the input buffer. */
      zstrm->avail_in = ZLIB_COMPR_IN_SIZE;
   }
   else
   {
      /* If not, tell zlib how much data there is left to process. */
      zstrm->avail_in = bytes_left;
   }

   if (zstrm->avail_in > 0)
   {
      /* Is there only padding left? */
      if (bytes_left <= pad)
      {
         /* If so, clear the bytes that represent
            the padding in the zlib input buffer. */
         (void)memset((void *)ramdump_zlib_buffers->in_buf, 0, zstrm->avail_in);
      }
      else
      {
         /* There is at least some non-padding left. */
         non_pad_left = bytes_left - pad;

         /* Is some of the data to copy in this pass actually padding? */
         if (zstrm->avail_in > non_pad_left)
         {
            /* Copy data from memory, up to the padding,
               to the zlib input buffer. */
            (void)memmove((void *)ramdump_zlib_buffers->in_buf,
                       (const void *)(start_address + (length - non_pad_left)),
                       non_pad_left);
            /* The rest is padding, so clear the bytes that represent
               the padding in the zlib input buffer. */
            (void)memset((void *)(ramdump_zlib_buffers->in_buf + non_pad_left),
                         0, zstrm->avail_in - non_pad_left);
         }
         else
         {
            /* If not, copy data to compress to ramdump zlib input buffer. */
            (void)memmove((void *)ramdump_zlib_buffers->in_buf,
                       (const void *)(start_address + (length - non_pad_left)),
                       zstrm->avail_in);
         }
      }

      /* Tell zlib where the input buffer is located. */
      zstrm->next_in = (Bytef *)ramdump_zlib_buffers->in_buf;

      /* Update byte count left to process. */
      bytes_left -= zstrm->avail_in;
   }
   return bytes_left;
}


static RamdumpResult
empty_output_buffer(z_stream *zstrm, DumpPos dump, DumpPosList plist,
					int depth)
{
	RamdumpResult result = RAMDUMP_FAIL;
	DumpPos pos = NO_DUMP_POS;
	unsigned int size;
	unsigned int extend_size;
	DumpPos zdat_pos;
	DumpPos image_pos;
	DumpOffset zdef_offset;
	U32 pad_compensate = 0;
	U32 prev_image_size;

	assert(zstrm != NULL);

	/* Number of bytes in the output buffer. */
	size = ZLIB_COMPR_OUT_SIZE - zstrm->avail_out;

	if (size > 0)
	{
		/* Remember position of zdat chunk */
		zdat_pos = plist[depth - 2];

		/* Offset of ZDEF chunk relative to ZDAT chunk. */
		zdef_offset = sizeof(struct RamdumpHeaderFORM);

		/* Remember position of image data */
		image_pos = plist[depth - 1];

		/* Get size of the compressed image already written. */
		prev_image_size = DUMP_VAL_U32(zdat_pos,
									   zdef_offset +
									   OFFSET(struct RamdumpChunkZDEF,
									   		  image_length));

		/* Figure out by how many bytes the image is padded. */
		pad_compensate = get_chunk_data_size(plist[depth - 1]) - prev_image_size;

		/* Calculate the size to extend by, aligned to 4 bytes,
		   compensating for the previous image padding. */
		extend_size = ALIGN_TO(size - pad_compensate, sizeof(U32));

		/* Extend dump to make room for the output data. */
		pos = extend_dump(dump, plist, depth, extend_size);

		/* If dump could not be extended, stop processing data. */
		if (pos != NO_DUMP_POS)
		{
			/* Write compressed data from output buffer to dump. */
			dump_area_write((const void *)ramdump_zlib_buffers->out_buf,
							image_pos,
							sizeof(struct RamdumpChunkHeader) + prev_image_size,
							size);

			/* Update written image length. */
			DUMP_VAL_U32(zdat_pos,
						 zdef_offset + OFFSET(struct RamdumpChunkZDEF,
						 image_length)) += size;

			/* Tell zlib where to find the cleared output buffer. */
			zstrm->next_out = (Bytef *)ramdump_zlib_buffers->out_buf;
			zstrm->avail_out = ZLIB_COMPR_OUT_SIZE;

			result = RAMDUMP_OK;
		}
	}
	else
	{
		result = RAMDUMP_OK;
	}

	return result;
}

static RamdumpResult
compress_append_data(z_stream *zstrm, DumpPos dump, DumpPosList plist,
					 int depth, OSADDRESS start_address, U32 length,
					 int *zlib_status)
{
	U32 left_to_proc;
	RamdumpResult result = RAMDUMP_OK;

	assert(zstrm != NULL);
	assert(zlib_status != NULL);

	*zlib_status = Z_OK;

   /* Set the number of bytes to compress,
      aligned up to multiple of 4 bytes. */
   left_to_proc = ALIGN_TO(length, sizeof(U32));

	/* (Kick the watchdog.) */
	ramdump_kick_watchdog();

	/* Tell zlib where to find cleared input buffer. */
	zstrm->avail_in = 0;
	zstrm->next_in = (Bytef *)ramdump_zlib_buffers->in_buf;

	/* Loop until there is no data left to hand over
	   to zlib and zlib has no data left to process
	   and zlib output buffer is not full */
	while (left_to_proc > 0 ||
		   zstrm->avail_in > 0 ||
		   zstrm->avail_out == 0)
	{
		/* Is the zlib input buffer empty and
		   is there any data left to process? */
		if (zstrm->avail_in == 0 && left_to_proc > 0)
		{
			/* If so, the input buffer needs a refill. */
			left_to_proc = refill_input_buffer(zstrm, start_address,
											   length, left_to_proc);
		}

		/* Is the output buffer full? */
		if (zstrm->avail_out == 0)
		{
			/* If so, the output buffer needs to be emptied
			   by writing pending output to dump. */
			result = empty_output_buffer(zstrm, dump, plist, depth);

			/* Did writing of data fail? */
			if (result != RAMDUMP_OK)
			{
				/* If so, stop processing data. */
				break;
			}
		}

		/* Compress data in input buffer. */
		*zlib_status = deflate(zstrm, Z_NO_FLUSH);

		/* Did something go wrong during compression? */
		if (*zlib_status != Z_OK)
		{
			/* If so, stop processing data. */
			result = RAMDUMP_FAIL;
			break;
		}
	}

	if (result == RAMDUMP_OK)
	{
		/* Does the output buffer contain data? */
		if (zstrm->avail_out < ZLIB_COMPR_OUT_SIZE)
		{
			/* If so, the output buffer needs to be emptied
			   by writing pending output to dump. */
			result = empty_output_buffer(zstrm, dump, plist, depth);
		}
	}

	return result;
}

static RamdumpResult
compress_append_close_data(z_stream *zstrm, DumpPos dump, DumpPosList plist,
					  	   int depth, int *zlib_status)
{
	RamdumpResult result = RAMDUMP_OK;

	assert(zstrm != NULL);
	assert(zlib_status != NULL);

	/* (Kick the watchdog.) */
	ramdump_kick_watchdog();

	*zlib_status = Z_OK;

	/* Tell zlib where to find the cleared input buffer. */
	zstrm->avail_in = 0;
	zstrm->next_in = (Bytef *)ramdump_zlib_buffers->in_buf;

	do
	{
		*zlib_status = deflate(zstrm, Z_FINISH);

		/* Did something go wrong during compression? */
		if (*zlib_status != Z_OK && *zlib_status != Z_STREAM_END)
		{
			/* If so, stop processing data. */
			result = RAMDUMP_FAIL;
			break;
		}

		/* Does the output buffer contain any data? */
		if (zstrm->avail_out < ZLIB_COMPR_OUT_SIZE)
		{
			/* If so, the output buffer needs to be emptied
			   by writing pending output to dump. */
			result = empty_output_buffer(zstrm, dump, plist, depth);

			/* Did writing of data fail? */
			if (result != RAMDUMP_OK)
			{
				/* If so, stop processing data. */
				break;
			}
		}

	}
	while(*zlib_status == Z_OK);

	return result;
}
#endif /* RAMDUMP_USE_ZLIB */

static RamdumpResult
append_data(DumpPos dump, DumpPosList plist, int depth,
            OSADDRESS start_address, U32 length)
{
   RamdumpResult result = RAMDUMP_FAIL;
   DumpPos pos;
   U32 data_len = ALIGN_TO(length, sizeof(U32));

   /* (Kick the watchdog.) */
   ramdump_kick_watchdog();

   pos = extend_dump(dump, plist, depth, data_len);

   if (pos != NO_DUMP_POS)
   {
      dump_area_clear(pos, 0, data_len, 0);
      dump_area_write((const void *)start_address, pos, 0, length);
      result = RAMDUMP_OK;
   }

   return result;
}

static RamdumpResult
add_request_data(DumpPos dump, DumpPosList plist, int depth,
				 z_stream *zstrm, const FmmSignal *signal, enum RamdumpCompressValue compression,
				 int *zlib_status)
{
	RamdumpResult result = RAMDUMP_OK;
	DumpPos pos = NO_DUMP_POS;
	unsigned int extend_size;
	OSADDRESS start_address;
	U32 length;

	assert(zstrm != NULL);
	assert(zlib_status != NULL);

	start_address = (OSADDRESS)signal;
   /* Size of signal field + size of signal data. */
	length = sizeof(U32) + signal->size;

	*zlib_status = Z_OK;

#ifdef RAMDUMP_USE_ZLIB
	/* Should the data be compressed? */
	if (compression == RAMDUMP_COMPRESSION_ZLIB)
	{
		/* If so, create compressed data chunk structure. */
		do
		{

			struct RamdumpChunkZDEF ch_ZDEF;
			DumpPos zdat_pos, image_pos;
			DumpOffset zdef_offset;
			RamdumpResult result;
			int i;

			extend_size = sizeof(struct RamdumpHeaderFORM) +
					      sizeof(struct RamdumpChunkZDEF);

			/* ZData form */
			pos = extend_dump(dump, plist, depth, extend_size);

			if (pos == NO_DUMP_POS)
			{
				result = RAMDUMP_FAIL;
				break;
			}

			/* Remember position of this chunk */
			zdat_pos = pos;

			write_group_header(zdat_pos, RAMDUMP_TAG_FORM,
							   extend_size - sizeof(struct RamdumpChunkHeader),
							   RAMDUMP_TAG_ZDAT);

			zdef_offset = sizeof(struct RamdumpHeaderFORM);

			/* Compression header */
			ch_ZDEF.chd.id = htonl(RAMDUMP_TAG_ZDEF);
			ch_ZDEF.chd.size = htonl(sizeof(struct RamdumpChunkZDEF) -
									 sizeof(struct RamdumpChunkHeader));

			ch_ZDEF.compression = RAMDUMP_COMPRESS_ZLIB_Z77;
			for (i = 0; i < 4; i++)
			{
				ch_ZDEF.compr_parameters[ i ] = 0;
			}
			ch_ZDEF.image_length = 0; /* Will be updated later */
			ch_ZDEF._reserved = 0;

			/* Write compression header */
			dump_area_write((const void *) &ch_ZDEF, zdat_pos, zdef_offset,
							sizeof(struct RamdumpChunkZDEF));
			add_child_location(plist, depth++, zdat_pos);

			/* Write empty data chunk */
			extend_size = sizeof(struct RamdumpChunkHeader);

			pos = extend_dump(dump, plist, depth, extend_size);

			if (pos == NO_DUMP_POS)
			{
				result = RAMDUMP_FAIL;
				break;
			}

			/* Remember position of image data */
			image_pos = pos;

			write_chunk_header(pos, RAMDUMP_TAG_DATA, 0);
			add_child_location(plist, depth++, image_pos);

			result = compress_append_data(zstrm, dump, plist, depth,
								     	  start_address, length,
								     	  zlib_status);

			if (result != RAMDUMP_OK)
			{
				break;
			}
		}
		while(0);
	}
	else
#endif
	{
		/* If not, create non compressed data chunk structure. */

		do
		{
			extend_size = sizeof(struct RamdumpChunkHeader);

			pos = extend_dump(dump, plist, depth, extend_size);

			if (pos == NO_DUMP_POS)
			{
				result = RAMDUMP_FAIL;
				break;
			}

			write_chunk_header(pos, RAMDUMP_TAG_DATA, 0);
			add_child_location(plist, depth++, pos);

			result = append_data(dump, plist, depth, start_address, length);
		}
		while(0);
	}

	return result;
}

static RamdumpResult
add_reply_data(DumpPos dump, DumpPosList plist, int depth,
			   z_stream *zstrm, const FmmSignal *signal,
			   int *zlib_status)
{
	RamdumpResult result;
	OSADDRESS start_address;
	U32 length;
   DumpOffset sghd_offset;

	assert(zstrm != NULL);
	assert(signal != NULL);
	assert(zlib_status != NULL);

	start_address = (OSADDRESS)signal;
   /* Size of signal field + size of signal data. */
	length = sizeof(uint32_t) + signal->size;

	*zlib_status = Z_OK;

   /* Is data chunk in a non compressed memory block? */
   if (get_chunk_id(plist[depth - 2]) == RAMDUMP_TAG_FORM &&
      get_group_contents_id(plist[depth - 2]) == RAMDUMP_TAG_SGBL &&
      get_chunk_id(plist[depth - 1]) == RAMDUMP_TAG_DATA)
   {
      /* If so, add non compressed data to chunk. */
      result = append_data(dump, plist, depth, start_address, length);

      /* Offset of SGHD chunk relative to FORM/SGBL chunk. */
      sghd_offset = sizeof(struct RamdumpHeaderFORM);

      /* Update length field of SGHD chunk */
      DUMP_VAL_U32(plist[depth - 2], /* Position of FORM/SGBL chunk */
                   sghd_offset + OFFSET(struct RamdumpChunkSGHD, length))
                   += length;
   }
#ifdef RAMDUMP_USE_ZLIB
   /* Or is data chunk in a compressed memory block? */
	else if (get_chunk_id(plist[depth - 3]) == RAMDUMP_TAG_FORM &&
		get_group_contents_id(plist[depth - 3]) == RAMDUMP_TAG_SGBL &&
		get_chunk_id(plist[depth - 2]) == RAMDUMP_TAG_FORM &&
		get_group_contents_id(plist[depth - 2]) == RAMDUMP_TAG_ZDAT &&
		get_chunk_id(plist[depth - 1]) == RAMDUMP_TAG_DATA)
	{
		/* If so, add compressed data to chunk. */
		result = compress_append_data(zstrm, dump, plist, depth,
								 	  start_address, length,
								 	  zlib_status);

      /* Offset of SGHD chunk relative to FORM/SGBL chunk. */
      sghd_offset = sizeof(struct RamdumpHeaderFORM);

      /* Update length field of SGHD chunk */
      DUMP_VAL_U32(plist[depth - 3], /* Position of FORM/SGBL chunk */
                   sghd_offset + OFFSET(struct RamdumpChunkSGHD, length))
                   += ALIGN_TO(length, sizeof(U32));
	}
#endif
   else
   {
      result = RAMDUMP_FAIL;
   }

	return result;
}

static RamdumpResult
add_endmark_data(DumpPos dump, DumpPosList plist, int depth,
				 z_stream *zstrm, const FmmSignal *signal,
				 int *zlib_status)
{
	RamdumpResult result = RAMDUMP_FAIL;
	DumpPos pos = NO_DUMP_POS;
	OSADDRESS start_address;
	U32 length = 0;
   DumpOffset sghd_offset;

	assert(zstrm != NULL);
	assert(signal != NULL);
	assert(zlib_status != NULL);

   (void)pos;

	*zlib_status = Z_OK;

	start_address = (OSADDRESS)signal;

	if (signal != NULL)
	{
      /* Size of signal field + size of signal data. */
		length = sizeof(uint32_t) + signal->size;
	}


   /* Is data chunk in a non compressed memory block? */
   if (get_chunk_id(plist[depth - 2]) == RAMDUMP_TAG_FORM &&
      get_group_contents_id(plist[depth - 2]) == RAMDUMP_TAG_SGBL &&
      get_chunk_id(plist[depth - 1]) == RAMDUMP_TAG_DATA)
   {
      /* If so, add non compressed data to chunk. */
      result = append_data(dump, plist, depth, start_address, length);

      /* Offset of SGHD chunk relative to FORM/SGBL chunk. */
      sghd_offset = sizeof(struct RamdumpHeaderFORM);

      /* Update length field of SGHD chunk */
      DUMP_VAL_U32(plist[depth - 2], /* Position of FORM/SGBL chunk */
                   sghd_offset + OFFSET(struct RamdumpChunkSGHD, length))
                   += length;
   }
#ifdef RAMDUMP_USE_ZLIB
	/* Or is data chunk in a compressed memory block? */
	else if (get_chunk_id(plist[depth - 3]) == RAMDUMP_TAG_FORM &&
		get_group_contents_id(plist[depth - 3]) == RAMDUMP_TAG_SGBL &&
		get_chunk_id(plist[depth - 2]) == RAMDUMP_TAG_FORM &&
		get_group_contents_id(plist[depth - 2]) == RAMDUMP_TAG_ZDAT &&
		get_chunk_id(plist[depth - 1]) == RAMDUMP_TAG_DATA)
	{
		/* If so, add compressed data to chunk. */
		do
		{
			/* Is there an endmark signal to write? */
			if (signal != NULL)
			{
				/* If so, add it to the current data chunk. */

				result = compress_append_data(zstrm, dump, plist, depth,
										 	  start_address, length,
											  zlib_status);
				if (result != RAMDUMP_OK)
				{
					break;
				}
			}

			pos = compress_append_close_data(zstrm, dump, plist,
											 depth, zlib_status);

			if (pos == NO_DUMP_POS)
			{
				result = RAMDUMP_FAIL;
				break;
			}
			else
			{
				result = RAMDUMP_OK;
			}

         /* Offset of SGHD chunk relative to FORM/SGBL chunk. */
         sghd_offset = sizeof(struct RamdumpHeaderFORM);

         /* Update length field of SGHD chunk */
         DUMP_VAL_U32(plist[depth - 3], /* Position of FORM/SGBL chunk */
                      sghd_offset + OFFSET(struct RamdumpChunkSGHD, length))
                      += ALIGN_TO(length, sizeof(U32));
		} while(0);
	}
#endif

	return result;
}

static long
dump_request(z_stream *zstrm, const FmmSignal *request, enum RamdumpCompressValue compression, RamdumpStatus *ret_code)
{
	RamdumpStatus return_code = RAMDUMP_ECURR_DUMP_TOO_LARGE;
	DumpPosList plist;
	DumpPos dump;
	int depth;
	int zlib_status = Z_OK;

	assert(zstrm != NULL);
	assert(request != NULL);

	dump = get_dump_to_extend();

	(void)return_code; /* Lint */

	do
	{
		struct RamdumpChunkSGHD ch_SGHD;
		DumpPos block_pos;
		DumpPos sgbl_pos;
		DumpPos sghd_pos;
		U32 extend_size;
		RamdumpResult result = RAMDUMP_FAIL;

		if (dump == NO_DUMP_POS)
		{
			return_code = RAMDUMP_ECURR_DUMP_TOO_LARGE;
			break;
		}

      if (!store_dump_sizes(dump))
      {
         return_code = RAMDUMP_EPOS_OR_DUMP_NOT_AVAILABLE;
         break;
      }

		/* Write header */
		block_pos = write_new_block_header(dump, plist, &depth);

		if (block_pos == NO_DUMP_POS)
		{
         restore_dump_sizes();
			return_code = RAMDUMP_ECURR_DUMP_TOO_LARGE;
			break;
		}

		/* Is there any easier way to find out the size? */
		extend_size = sizeof(struct RamdumpHeaderFORM);

		sgbl_pos = extend_dump(dump, plist, depth, extend_size);

		if (sgbl_pos == NO_DUMP_POS)
		{
         restore_dump_sizes();
			return_code = RAMDUMP_ECURR_DUMP_TOO_LARGE;
			break;
		}

		/* Form header */
		write_group_header(sgbl_pos, RAMDUMP_TAG_FORM,
						   sizeof(struct RamdumpHeaderFORM) -
						   sizeof(struct RamdumpChunkHeader),
						   RAMDUMP_TAG_SGBL);
		add_child_location(plist, depth++, sgbl_pos);

		/* Request signal number */
		extend_size = sizeof(struct RamdumpChunkSGHD);
		sghd_pos = extend_dump(dump, plist, depth, extend_size);

		if (sghd_pos == NO_DUMP_POS)
		{
         restore_dump_sizes();
			return_code = RAMDUMP_ECURR_DUMP_TOO_LARGE;
			break;
		}

		ch_SGHD.chd.id = htonl(RAMDUMP_TAG_SGHD);
		ch_SGHD.chd.size = htonl(sizeof(struct RamdumpChunkSGHD) -
								 sizeof(struct RamdumpChunkHeader));
		ch_SGHD.signo = request->data.signo;
      /* Default value. Will be overwritten when endmark is dumped. */
      ch_SGHD.status = 0xBADC0DE;
      /* Size of signal field + size of signal data, aligned to 32 bits. */
      ch_SGHD.length = ALIGN_TO(sizeof(uint32_t) + request->size, sizeof(U32));

		dump_area_write((const void *)&ch_SGHD, sghd_pos, 0,
						sizeof(struct RamdumpChunkSGHD));

		/* Dump request signal data */
		result = add_request_data(dump, plist, depth, zstrm, request,
					     		  compression, &zlib_status);

		if (result != RAMDUMP_OK)
		{
         restore_dump_sizes();
			if (zlib_status != Z_OK)
			{
				return_code = RAMDUMP_ECOMPRESSION_FAILED;
				break;
			}
			else
			{
				return_code = RAMDUMP_ECURR_DUMP_TOO_LARGE;
				break;
			}
		}

		return_code = RAMDUMP_SUCCESS;
	} while (0); /* Just to avoid ugly nesting. */

	/* Update dump index */
	if (dump != NO_DUMP_POS)
	{
		update_dump_index(find_dump_id(dump), dump);
	}

	if (ret_code)
	{
		*ret_code = return_code;
	}

	return 0;
}



static long
dump_reply(z_stream *zstrm, const FmmSignal *reply, RamdumpStatus *ret_code)
{
	RamdumpStatus return_code = RAMDUMP_ECURR_DUMP_TOO_LARGE;
	RamdumpResult result = RAMDUMP_FAIL;
	DumpPos dump;
	int zlib_status = Z_OK;

	assert(zstrm != NULL);
	assert(reply != NULL);

	dump = get_dump_to_extend();

	(void)return_code; /* Lint */

	/* add reply data */
	result = add_reply_data(dump, last_chunk_plist, last_chunk_depth,
						 	zstrm, reply, &zlib_status);

	if (result != RAMDUMP_OK)
	{
      restore_dump_sizes();
		if (zlib_status != Z_OK)
		{
			return_code = RAMDUMP_ECOMPRESSION_FAILED;
		}
		else
		{
			return_code = RAMDUMP_ECURR_DUMP_TOO_LARGE;
		}
	}
	else
	{
		return_code = RAMDUMP_SUCCESS;
	}

	if (ret_code)
	{
		*ret_code = return_code;
	}

	return 0;
}

static long
dump_endmark(z_stream *zstrm, const FmmSignal *endmark, int32_t status,
             RamdumpStatus *ret_code)
{
	RamdumpStatus return_code = RAMDUMP_ECURR_DUMP_TOO_LARGE;
	RamdumpResult result = RAMDUMP_FAIL;
	DumpPos dump;
   DumpPos *plist = last_chunk_plist;
   int depth = last_chunk_depth;
	int zlib_status = Z_OK;

	assert(zstrm != NULL);
	assert(endmark != NULL);

	dump = get_dump_to_extend();

	(void)return_code; /* Lint */

	/* add endmark data */
	result = add_endmark_data(dump, plist, depth, zstrm, endmark, &zlib_status);

	if (result != RAMDUMP_OK)
	{
      restore_dump_sizes();
		if (zlib_status != Z_OK)
		{
			return_code = RAMDUMP_ECOMPRESSION_FAILED;
		}
		else
		{
			return_code = RAMDUMP_ECURR_DUMP_TOO_LARGE;
		}
	}
	else
	{
      if (get_chunk_id(plist[depth - 3]) == RAMDUMP_TAG_FORM
          && get_group_contents_id(plist[depth - 3]) == RAMDUMP_TAG_SGBL)
      {
         /* Offset of SGHD chunk relative to FORM/SGBL chunk. */
         DumpOffset sghd_offset = sizeof(struct RamdumpHeaderFORM);

         /* Update status field of SGHD chunk */
         DUMP_VAL_U32(plist[depth - 3], /* Position of FORM/SGBL chunk */
                      sghd_offset + OFFSET(struct RamdumpChunkSGHD, status))
                      = status;
      }
      else if (get_chunk_id(plist[depth - 2]) == RAMDUMP_TAG_FORM
          && get_group_contents_id(plist[depth - 2]) == RAMDUMP_TAG_SGBL)
      {
         /* Offset of SGHD chunk relative to FORM/SGBL chunk. */
         DumpOffset sghd_offset = sizeof(struct RamdumpHeaderFORM);

         /* Update status field of SGHD chunk */
         DUMP_VAL_U32(plist[depth - 2], /* Position of FORM/SGBL chunk */
                      sghd_offset + OFFSET(struct RamdumpChunkSGHD, status))
                      = status;
      }

		return_code = RAMDUMP_SUCCESS;
	}

	if (ret_code)
	{
		*ret_code = return_code;
	}

	return 0;
}

/**
 * @function get_dump_id
 *
 * @brief Retrieves the nth last dump identifier in the dump history.
 *
 * @input nth_last Nth last dump identifier will be retrieved. Zero for
 * most recent.
 *
 * @output dump_idp Dump identifier or zero if the dump could not be found.
 *
 * @return Returns zero.
 */
static long
get_dump_id(U32 nth_last, RamdumpDumpId *dump_idp)
{
   int count = nth_last;
   RamdumpDumpId dump_id;
   RamdumpDumpId curr_dump_id = CURR_DUMP_ID;

   /* Decrease count and skip dumps until we have found the desired one */
   do
   {
      DumpPos dump = get_most_recent_dump(curr_dump_id);
      dump_id = 0;

      if (VALID_DUMP_POS(dump))
      {
      	 /* This will be returned if counter reaches zero */
      	 curr_dump_id = dump_id = find_dump_id(dump);
      }
      else
      {
          /* None found in table, none will be found the next lap either */
          break;
      }

      curr_dump_id--;
      if (curr_dump_id & 0x80000000UL)
      {
         curr_dump_id = 0x7fffffffUL;
      }
   } while (count-- > 0);

   /* Return any found dump id in parameter */
   *dump_idp = dump_id;

   return 0;
}

/**
 * @function get_dump_length
 *
 * @brief Returns the number of bytes found in a valid dump
 * that starts at the specified block.
 *
 * @input dump Dump position pointing at the start of a dump.
 *
 * @return Returns the size of the dump.
 */
static U32
get_dump_length(DumpPos dump)
{
   assert(is_valid_iff_structure(dump));

   return get_chunk_size(dump);
}

/**
 * @function get_dump_info
 *
 * @brief Retrieves information about a dump in the dump area.
 *
 * @input dump_id Dump id of the dump to get info for.
 * @output dump_info Pointer to struct to fill out with info.
 * @output ret_code RAMDUMP_SUCCESS on success,
 * RAMDUMP_EPOS_OR_DUMP_NOT_AVAILABLE or
 * RAMDUMP_EINVALID_ARGUMENT on failure.
 *
 * @return Returns zero.
 */
static long
get_dump_info(RamdumpDumpId dump_id, struct RamdumpDumpInfo *dump_info,
	      RamdumpStatus *ret_code)
{
   if (dump_info)
   {
      DumpPos dump;

      /*
       * The dump table is updated every time a block is written,
       * but when the dump area is completely fresh, the table
       * is not consistent.
       * Scan the table to see if any entry is not consistent
       */
      if (!is_dump_index_consistent())
      {
         build_dump_index();
      }

      /* Get dump position */
      dump = get_dump_pos(dump_id);

      if (VALID_DUMP_POS(dump) && is_valid_dump(dump))
      {
         DumpPos tmp_pos = find_chunk_in_valid_iff(dump, block_hd);

         /* Fill in dump info */
         dump_info->dump_id = dump_id;
         dump_info->size = get_dump_length(dump);

         assert(tmp_pos != NO_DUMP_POS);

         dump_info->clock = get_chunk_data_U32(tmp_pos,
                                               OFFSET(struct RamdumpChunkBLHD,
                                               clock));
         dump_info->microseconds = get_chunk_data_U32(tmp_pos,
                                                OFFSET(struct RamdumpChunkBLHD,
                                                microseconds));

         *ret_code = RAMDUMP_SUCCESS;
      }
      else
      {
         *ret_code = RAMDUMP_EPOS_OR_DUMP_NOT_AVAILABLE;
      }
   }
   else
   {
      *ret_code = RAMDUMP_EINVALID_ARGUMENT;
   }

   return 0;
}

/**
 * @function read_dump
 *
 * @brief Reads a number of bytes from a dump to a provided buffer.
 *
 * @long Reads a number of bytes from a certain dump that starts at
 * the given offset in the dump, to a buffer given by the caller.
 *
 * @input dump_id Dump id of dump to read from.
 * @input pos Dump position to read from.
 * @output buffer Buffer to copy the bytes to.
 * @input bytes Number of bytes to read.
 * @output bytes_read Number of copied bytes. Contains the error code
 * RAMDUMP_EPOS_OR_DUMP_NOT_AVAILABLE if the dump id was not found
 * or if the position in the dump is invalid.
 *
 * @return Returns zero.
 */
static long
read_dump(RamdumpDumpId dump_id, U32 pos, void *buffer, U32 bytes,
          S32 *bytes_read)
{
   /* default "return" value */
   S32 b_read = RAMDUMP_EPOS_OR_DUMP_NOT_AVAILABLE;
   DumpPos dump = get_dump_pos(dump_id);

   /* assert(!VALID_DUMP_POS(dump) || is_valid_dump(dump)) */

   if (VALID_DUMP_POS(dump))
   {
      /* Kick watchdog */
      ramdump_kick_watchdog();

      /* Now scan thru the dump and copy the number of bytes requested */
      dump_area_read(buffer, dump, pos, bytes);
      b_read = bytes;
   }

   if (bytes_read)
   {
      *bytes_read = b_read;
   }

   return 0;
}

/**
 * @function find_dump_chunk
 *
 * @brief Finds a chunk in a dump according to a specification,
 * e.g. {'FORM', 'PMD ', 'DPID'}.
 *
 * @long Finds a chunk in a dump according to a specification,
 * e.g. {'FORM', 'PMD ', 'DPID'}.
 *
 * @long The search begins at the indicated offset in the dump.
 * This offset should point at a valid chunk.
 *
 * @long An optional last location offset offers the ability to search
 * for other matching chunks. A last location offset of zero will give
 * the first matching chunk.
 *
 * @input dump_id Dump id of the dump to search.
 * @input start_offset Dump position offset where the search will start.
 * @input cs Chunk match specification.
 * @input last_located_offset Offset of the last location where matching
 * chunk was found.
 * @output found_offset The offset where the chunk was found.
 *
 * @output found_offset The offset of the matching chunk, or the error code
 * RAMDUMP_EPOS_OR_DUMP_NOT_AVAILABLE if the start offset is invalid or
 * no chunk was found.
 *
 * @return Returns zero.
 */
static long
find_dump_chunk(RamdumpDumpId dump_id, DumpOffset start_offset, const IffId *cs,
                DumpOffset last_located_offset, S32 *found_offset)
{
   /* default "return" value */
   S32 f_offset = RAMDUMP_EPOS_OR_DUMP_NOT_AVAILABLE;

   /* Get start position of dump */
   DumpPos dump = get_dump_pos(dump_id);

   /* assert(!VALID_DUMP_POS(dump) || is_valid_dump(dump)) */

   /* Check if dump found */
   if (VALID_DUMP_POS(dump))
   {
      DumpPos found_pos;

      /* Find next matching chunk */
      found_pos = find_next_chunk_in_valid_iff(DUMP_POS(dump, start_offset),
				                                   cs,
                                               DUMP_POS(dump,
                                                        last_located_offset));

      if (VALID_DUMP_POS(found_pos))
      {
         f_offset = DUMP_OFFSET(found_pos, dump);
      }
   }

   if (found_offset)
   {
      *found_offset = f_offset;
   }

   return 0;
}

/**
 * @function configure_ramdump_area
 *
 * @brief Configures the location and size of the ramdump area.
 *
 * @long Configures the location and size of the ramdump area, as well
 * as the work area size. Until this function has been called, the Ramdump
 * is considered uninitialized and the API functions (except this one)
 * will return RAMDUMP_ERAMDUMP_NOT_INITIALIZED.
 *
 * @input base Base address of the ramdump area
 * @input size Size of the ramdump area.
 * @input work_area_size Size of the ramdump work area.
 * @input pmd_work_area_size Size of space to inject between book keeping
 * data and dump storage area for storage compatibility with old pmd.
 *
 * @return Returns zero. Triggers error handler on failure.
 */
static long
configure_ramdump_area(OSADDRESS base, U32 size, U32 heap_size,
                       U32 pmd_work_area_size)
{
   struct TmpRamdumpArea
   {
      union RamdumpPersistentData pers_data;
      struct DumpArea dump_area;
   };

   if (base == 0)
   {
      error2((OSERRCODE)(OSE_RAMDUMP_EBAD_RAMDUMP_AREA_CONFIGURATION),
             (OSERRCODE)base);
   }
   else if (heap_size < RAMDUMP_MIN_WORK_AREA_LENGTH)
   {
      error2((OSERRCODE)(OSE_RAMDUMP_EBAD_RAMDUMP_HEAP_CONFIGURATION),
             (OSERRCODE) heap_size);
   }
   else if (size < (sizeof(struct TmpRamdumpArea)))
   {
      error2((OSERRCODE)(OSE_RAMDUMP_EBAD_RAMDUMP_AREA_CONFIGURATION),
             (OSERRCODE)size);
   }
   else
   {
      ramdump_initialized = 0;

      /* Initialize dump area pointers */
      ramdump_area = (struct RamdumpArea *)base;
      ramdump_area_length = size;
      ramdump_work_area_length = heap_size;

      ramdump_dump_area =
         (struct DumpArea *)((char *)base +
                             OFFSET(struct TmpRamdumpArea, dump_area) +
                             pmd_work_area_size);

      ramdump_dump_area_length =
         size - (U32)(OFFSET(struct TmpRamdumpArea, dump_area) +
                      pmd_work_area_size);

      assert(sizeof(ramdump_area->pers_data) == RAMDUMP_PERSISTENT_DATA_SIZE);

#ifdef RAMDUMP_USE_ZLIB
      /* Allocate memory for the zlib buffers from ramdump heap.
         This memory will not be freed, as the life span
         of the ramdump is the same as that of the system. */
      ramdump_zlib_buffers = (struct ZlibBuffers *)ramdump_heap_malloc(
                                       sizeof(struct ZlibBuffers));

      if (ramdump_zlib_buffers == NULL)
      {
         error2((OSERRCODE)(OSE_RAMDUMP_ENOT_ENOUGH_HEAP_MEMORY), 0);
      }
#endif

      ramdump_initialized = 1;
   }

   return 0;
}

/**
 * @function set_watchdog_func
 *
 * @brief Sets the watchdog callback function to be invoked during
 * long operations.
 *
 * @input f Callback function
 *
 * @return Returns zero.
 */
static long
set_watchdog_func(RamdumpWatchdogCb *f, RamdumpStatus *return_code)
{
   struct EfmiCurrent current;
   struct EfmiBlockInfo block_info;
   EfmiStatus status;

   /* Interrupts MUST be disabled when getting the current process. */
   status = ose_efmi_get_current(&current, sizeof(struct EfmiCurrent));

   if (status != EFMI_OK)
   {
      *return_code = RAMDUMP_EUNABLE_TO_GET_CURRENT_PROCESS_INFO;
   }
   else
   {
      status = ose_efmi_get_block_info((PROCESS)current.block,
                                       &block_info,
                                       sizeof(struct EfmiBlockInfo));

      if (status != EFMI_OK)
      {
         *return_code = RAMDUMP_EUNABLE_TO_GET_CURRENT_BLOCK_INFO;
      }
      else
      {
         if (block_info.supervisor_mode)
         {
            /* In supervisor mode, set watchdog. */
            watchdog_cb = f;
            *return_code = RAMDUMP_SUCCESS;
         }
         else
         {
            /* Not in supervisor mode, do not set watchdog. */
            *return_code = RAMDUMP_EMUST_BE_SUPERVISOR;
         }
      }
   }

   return 0;
}

/* ======================================================== */


/**
 * @data ramdump_stack_switch_args
 *
 * @brief Struct for passing arguments to a function that needs
 * to be called with the ramdump stack.
 *
 * @long Struct for passing arguments to a function that needs
 * to be called with the ramdump stack. The variable is global
 * in order to minimize the stack usage.
 */
static StackSwitchArgs ramdump_stack_switch_args;

/**
 * @function ramdump_start_handler1
 *
 * @brief Start handler that initialized the ramdump.
 *
 * @long Start handler that initializes the ramdump, specified
 * in the kernel configuration file.
 *
 * @example
 * START_HANDLER1(ramdump_start_handler1)
 */
void
ramdump_start_handler1(void)
{
#ifdef LIBWHAT
  LIBWHAT
#endif

  ASSERT_IFF_STRUCTURE(); /* No memory accesses */

#ifndef RAMDUMP_DO_NOT_SWITCH_STACK
  /* Fill stack with pattern */
  ASSERT_CODE(memset((void *) ramdump_stack, 0xEE, sizeof(ramdump_stack)));
#endif

  /* Run the ramdump heap start handler too. It will be run
     as a normal start handler once the freeze mode heap
     becomes its own separate module, but for now it will be
     called from here. */
  ramdump_heap_start_handler1();

  /*
   * Initialize configuration
   */
  ramdump_initialize_area();

  /*
   * Simulate a call to trap handler
   */
  ramdump_start_new_dump();
}

/**
 * @function ramdump_initialize_area
 *
 * @brief Reads the ramdump system parameters and initializes the ramdump
 * accordingly.
 */
static void
ramdump_initialize_area(void)
{
   int result;
   long dump_area_base;
   long dump_area_size;
   long heap_size;
   long pmd_work_area_size;

   result = sysparam_readattr_long("krn/ramdump/dumps",
				   "base", &dump_area_base);
   if (result != 0)
   {
      ramlog_printf("ramdump: Unable to read parameter krn/ramdump/dumps, "
		    "attribute base.\n");
      error2((OSERRCODE)(OSE_RAMDUMP_EERROR_READING_DUMPS_SYSPARAM),
	     (OSERRCODE)NULL);
   }

   result = sysparam_readattr_long("krn/ramdump/dumps",
				   "size", &dump_area_size);
   if (result != 0)
   {
      ramlog_printf("ramdump: Unable to read parameter krn/ramdump/dumps, "
		    "attribute size.\n");
      error2((OSERRCODE)(OSE_RAMDUMP_EERROR_READING_DUMPS_SYSPARAM),
	     (OSERRCODE)NULL);
   }

   result = sysparam_readattr_long("krn/ramdump/heap",
				   "size", &heap_size);
   if (result != 0)
   {
      ramlog_printf("ramdump: Unable to read parameter krn/ramdump/heap, "
		    "attribute size.\n");
      error2((OSERRCODE)(OSE_RAMDUMP_EERROR_READING_HEAP_SYSPARAM),
	     (OSERRCODE)NULL);
   }

   heap_size = ALIGN_TO(heap_size, sizeof(uint64_t));

   if (((U32)heap_size) < RAMDUMP_MIN_WORK_AREA_LENGTH)
   {
      ramlog_printf("ramdump: Ramdump heap size is too small.");
      error2((OSERRCODE)(OSE_RAMDUMP_EBAD_RAMDUMP_HEAP_CONFIGURATION),
	     (OSERRCODE)heap_size);
   }

   result = sysparam_readattr_long("krn/ramdump/pmd_work_area",
               "size", &pmd_work_area_size);
   if (result != 0)
   {
      /* Unable to find pmd_work_area parameter, assume size 0 */
      pmd_work_area_size = 0;
   }

   ramdump_configure_ramdump_area((OSADDRESS)dump_area_base,
				  (U32)dump_area_size,
				  (U32)heap_size,
                                  (U32)pmd_work_area_size);

}




/* -------------------------------------------------------------
 * Public interface functions
 * -------------------------------------------------------------
 */


/**
 * @function ramdump_start_new_dump
 *
 * @brief Starts a new dump.
 *
 * @long This function starts a new dump after the last one in
 * the post-mortem dump area. Subsequent calls to ramdump_dump_xxx
 * will write information in the same dump.
 *
 * @long If several dumps are preferred, call this function to start a
 * new dump.
 *
 * @return If the function succeeds, RAMDUMP_SUCCESS is returned.
 * If the dump function was called recursively, the error code
 * RAMDUMP_ERECURSIVE_DUMP_CALL is returned. If the ramdump area is
 * uninitialized, the error code RAMDUMP_ERAMDUMP_NOT_INITIALIZED
 * is returned.
 *
 */
RamdumpStatus
ramdump_start_new_dump(void)
{
   RamdumpStatus return_code = RAMDUMP_SUCCESS;

   /* Try to get dump lock and disabled interrupts */
   if (DUMP_LOCK( INTERRUPTS_BLOCKED ))
   {
      if (ramdump_initialized != 0)
      {
#ifdef RAMDUMP_DO_NOT_SWITCH_STACK
         advance_dump_id();
#else
         ramdump_switch_stack((U32)RAMDUMP_STACK_TOP,
                              (StackSwitchFunc)advance_dump_id,
                               NULL);
#endif
      }
      else
      {
         return_code = RAMDUMP_ERAMDUMP_NOT_INITIALIZED;
      }

      /* Release lock and restore interrupts */
      DUMP_UNLOCK( INTERRUPTS_BLOCKED );
   }
   else
   {
      /* Return without lock and enabled interrupts */
      return_code = RAMDUMP_ERECURSIVE_DUMP_CALL;
   }

   /* Return without lock and enabled interrupts */
   return return_code;
}

/**
 * @function ramdump_private_clear_dumps
 *
 * @brief Clears dump area.
 *
 * @long This function clears the dump area.
 *
 * @return If the function succeeds, RAMDUMP_SUCCESS is returned.
 * If the dump function was called recursively, the error code
 * RAMDUMP_ERECURSIVE_DUMP_CALL is returned. If the ramdump area is
 * uninitialized, the error code RAMDUMP_ERAMDUMP_NOT_INITIALIZED
 * is returned.
 *
 */
RamdumpStatus
ramdump_clear_dumps(void)
{
   RamdumpStatus return_code = RAMDUMP_SUCCESS;

   /* Try to get dump lock */
   if (DUMP_LOCK( INTERRUPTS_ENABLED ))
   {
      if (ramdump_initialized != 0)
      {
#ifdef RAMDUMP_DO_NOT_SWITCH_STACK
         clear_dumps();
#else
         ramdump_switch_stack((U32)RAMDUMP_STACK_TOP,
                              (StackSwitchFunc)clear_dumps,
                               NULL);
#endif
      }
      else
      {
         return_code = RAMDUMP_ERAMDUMP_NOT_INITIALIZED;
      }

      /* Release lock */
      DUMP_UNLOCK( INTERRUPTS_ENABLED );
   }
   else
   {
      /* Return without lock and enabled interrupts */
      return_code = RAMDUMP_ERECURSIVE_DUMP_CALL;
   }

   /* Return without lock and enabled interrupts */
   return return_code;
}


/**
 * @function ramdump_dump_error_info
 *
 * @brief Writes information about an error event into the current dump.
 *
 * @long NOTE: Interrupts must be disabled when calling this function.
 *
 * @input description User-provided optional text. The string is only
 * copied up to the first terminator character.
 * @input user_called Pass value of user_called flag here.
 * @input error_code Pass error code here.
 * @input extra Pass extra parameter here.
 *
 * @return If the function succeeds, RAMDUMP_SUCCESS is returned.
 * If the current dump exceeds its size limit, the error code
 * RAMDUMP_ECURR_DUMP_TOO_LARGE is returned. If the description pointer
 * is invalid, the error code RAMDUMP_EMEMORY_ACCESS_NOK is returned.
 * If the dump function was called recursively, the error code
 * RAMDUMP_ERECURSIVE_DUMP_CALL is returned. If the ramdump area is
 * uninitialized, the error code RAMDUMP_ERAMDUMP_NOT_INITIALIZED
 * is returned. RAMDUMP_EUNABLE_TO_GET_CURRENT_PROCESS_INFO is returned
 * if the function is unable to get information for the current process.
 */
RamdumpStatus
ramdump_dump_error_info(const char *description,
			OSBOOLEAN user_called, OSERRCODE error_code,
			OSERRCODE extra)
{
   RamdumpStatus ret_code = RAMDUMP_SUCCESS;

   if (DUMP_LOCK( INTERRUPTS_BLOCKED ))
   {
      if (ramdump_initialized != 0)
      {
#ifdef RAMDUMP_DO_NOT_SWITCH_STACK
         dump_error_info(description, user_called, error_code,
                         extra, &ret_code);
#else
         ramdump_stack_switch_args.arg1 = (long)description;
         ramdump_stack_switch_args.arg2 = (long)user_called;
         ramdump_stack_switch_args.arg3 = (long)error_code;
         ramdump_stack_switch_args.arg4 = (long)extra;
         ramdump_stack_switch_args.arg5 = (long)&ret_code;

         ramdump_switch_stack((U32)RAMDUMP_STACK_TOP,
                              (StackSwitchFunc)dump_error_info,
                              &ramdump_stack_switch_args);
#endif
      }
      else
      {
         /* Ramdump area has not been initialized */
         ret_code = RAMDUMP_ERAMDUMP_NOT_INITIALIZED;
      }

      DUMP_UNLOCK( INTERRUPTS_BLOCKED );
   }
   else
   {
      ret_code = RAMDUMP_ERECURSIVE_DUMP_CALL;
   }

   return ret_code;
}

/**
 * @function ramdump_dump_text_info
 *
 * @brief Writes an arbitrary textual description into the current dump.
 *
 * @long Writes an arbitrary textual description into the current dump.
 *
 * @input description User-provided optional text. The string
 * is only copied up to the first terminator character.
 *
 * @return If the function succeeds, RAMDUMP_SUCCESS is returned.
 * If the current dump exceeds its size limit, the error code
 * RAMDUMP_ECURR_DUMP_TOO_LARGE is returned. If the description pointer
 * is invalid, the error code RAMDUMP_EMEMORY_ACCESS_NOK is returned.
 * If the dump function was called recursively, the error code
 * RAMDUMP_ERECURSIVE_DUMP_CALL is returned. If the ramdump area is
 * uninitialized, the error code RAMDUMP_ERAMDUMP_NOT_INITIALIZED
 * is returned.
 */
RamdumpStatus
ramdump_dump_text_info(const char *description)
{
   RamdumpStatus ret_code = RAMDUMP_SUCCESS;

   if (DUMP_LOCK( INTERRUPTS_ENABLED ))
   {
      if (ramdump_initialized != 0)
      {
#ifdef RAMDUMP_DO_NOT_SWITCH_STACK
         dump_text_info(description, &ret_code);
#else
         ramdump_stack_switch_args.arg1 = (long)description;
         ramdump_stack_switch_args.arg2 = (long)&ret_code;

         ramdump_switch_stack((U32)RAMDUMP_STACK_TOP,
                              (StackSwitchFunc)dump_text_info,
                              &ramdump_stack_switch_args);
#endif
      }
      else
      {
         /* Ramdump area has not been initialized */
         ret_code = RAMDUMP_ERAMDUMP_NOT_INITIALIZED;
      }
      DUMP_UNLOCK( INTERRUPTS_ENABLED );
   }
   else
   {
      ret_code = RAMDUMP_ERECURSIVE_DUMP_CALL;
   }

   return ret_code;
}

/**
 * @function ramdump_dump_memory_image
 *
 * @brief Dumps a section of memory to the current dump.
 *
 * @long Dumps a section of memory to the current dump.
 *
 * @input description User-provided optional text. The string is only
 * copied up to the first terminator character.
 * @input start_address Start address of memory to dump.
 * @input length Number of bytes to dump.
 *
 * @return If the function succeeds, RAMDUMP_SUCCESS is returned.
 * If the current dump exceeds its size limit, the function fails
 * and returns the error code RAMDUMP_ECURR_DUMP_TOO_LARGE. If the memory
 * access fails or the description pointer is invalid, the error code
 * RAMDUMP_EMEMROY_ACCESS_NOK is returned. If the dump function was called
 * recursively, the error code RAMDUMP_ERECURSIVE_DUMP_CALL is returned.
 * If the ramdump area is unitialized, the error code
 * RAMDUMP_ERAMDUMP_NOT_INITIALIZED is returned.
 */
RamdumpStatus
ramdump_dump_memory_image(const char *description,
			  OSADDRESS start_address, U32 length)
{
   RamdumpStatus ret_code = RAMDUMP_SUCCESS;

   if (DUMP_LOCK( INTERRUPTS_ENABLED ))
   {

      if (ramdump_initialized != 0)
      {
#ifdef RAMDUMP_DO_NOT_SWITCH_STACK
         dump_memory_image(description, start_address, length, &ret_code);
#else
         ramdump_stack_switch_args.arg1 = (long)description;
         ramdump_stack_switch_args.arg2 = (long)start_address;
         ramdump_stack_switch_args.arg3 = (long)length;
         ramdump_stack_switch_args.arg4 = (long)&ret_code;

         ramdump_switch_stack((U32)RAMDUMP_STACK_TOP,
                              (StackSwitchFunc)dump_memory_image,
                              &ramdump_stack_switch_args);
#endif
      }
      else
      {
         /* Dump area has not been initialized */
         ret_code = RAMDUMP_ERAMDUMP_NOT_INITIALIZED;
      }

      DUMP_UNLOCK( INTERRUPTS_ENABLED );
   }
   else
   {
      ret_code = RAMDUMP_ERECURSIVE_DUMP_CALL;
   }

   return ret_code;
}

/**
 * @function ramdump_get_dump_id
 *
 * @brief Retrieves a dump identifier in the dump history.
 *
 * @long This function retrieves a dump identifier in the dump history.
 * The function retrieves the latest dump if nthLast == 0,
 * the second latest dump if nthLast == 1 and so on.
 *
 * @input nth_last An index to the nth last dump id to retrieve.
 *
 * @return If the function succeeds, a dump identifier is returned. If
 * no nth_last dump was found, the function returns zero.
 */
RamdumpDumpId
ramdump_get_dump_id(U32 nth_last)
{
   RamdumpDumpId dump_id = 0;

   if (ramdump_initialized != 0)
   {
      get_dump_id(nth_last, &dump_id);
   }

   return dump_id;
}

/**
 * @function ramdump_get_dump_info
 *
 * @brief Retrieves basic information about a dump.
 *
 * @long This function retrieves some basic information about a dump,
 * such as size and timestamp.
 *
 * @input dump_id Dump identifier.
 * @output dump_info User-provided dump information struct for the
 * function to fill out.
 *
 * @return If the function succeeds, RAMDUMP_SUCCESS is returned.
 * If the specified dump id is not valid, the error code
 * RAMDUMP_EPOS_OR_DUMP_NOT_AVAILABLE is returned. If dump_info
 * is a NULL pointer, RAMDUMP_ERAMDUMP_INVALID_ARGUMENT is returned.
 * If the ramdump area is uninitialized, the error code
 * RAMDUMP_ERAMDUMP_NOT_INITIALIZED is returned.
 */
RamdumpStatus
ramdump_get_dump_info(RamdumpDumpId dump_id, struct RamdumpDumpInfo *dump_info)
{
   RamdumpStatus ret_code = RAMDUMP_SUCCESS;

   if (ramdump_initialized != 0)
   {
      get_dump_info(dump_id, dump_info, &ret_code);
   }
   else
   {
      ret_code = RAMDUMP_ERAMDUMP_NOT_INITIALIZED;
   }

   return ret_code;
}

/**
 * @function ramdump_read_dump
 *
 * @brief Copies a number of bytes of a dump into a user-provided buffer.
 *
 * @long Copies a number of bytes of a dump into a user-provided buffer.
 *
 * @input dump_id Dump identifier.
 * @input pos Byte offset in the dump.
 * @output buffer User-provided buffer into which data is copied.
 * @input bytes Number of bytes to copy.
 *
 * @return If the function succeeds, it returns the number of
 * bytes copied. If memory access to the buffer fails, the error code
 * RAMDUMP_EMEMORY_ACCESS_NOK is returned. If the position does not
 * exist in the dump, or if the dump id is not a valid active dump id,
 * the error code RAMDUMP_EPOS_OR_DUMP_NOT_AVAILABLE is returned.
 * If the ramdump area is uninitialized, the error code
 * RAMDUMP_ERAMDUMP_NOT_INITIALIZED is returned.
 */
S32
ramdump_read_dump(RamdumpDumpId dump_id, U32 pos, void *buffer, U32 bytes)
{
   S32 bytes_read = 0;

   if (ramdump_initialized != 0)
   {
      read_dump(dump_id, pos, buffer, bytes, &bytes_read);
   }
   else
   {
      bytes_read = RAMDUMP_ERAMDUMP_NOT_INITIALIZED;
   }

   return bytes_read;
}

/**
 * @function ramdump_find_dump_c
 *
 * @include
 *
 * @brief Locates chunk in dump according to a matcthing specification,
 * e.g. {'FORM', 'PMD ', 'DPID'}
 *
 * @input dump_id Dump id of the dump to scan.
 * @input start_offset Start offset in the dump.
 * @input cs Chunk specification.
 * @last_located_offset Location of the last matching chunk.
 *
 * @return Returns the found offset.
 */
S32
ramdump_find_dump_c(RamdumpDumpId dump_id, U32 start_offset, const IffId *cs,
                    U32 last_located_offset)
{
   S32 found_offset = -1;

   if (ramdump_initialized != 0)
   {
      find_dump_chunk(dump_id, start_offset, cs, last_located_offset, &found_offset);
   }
   else
   {
      found_offset = RAMDUMP_ERAMDUMP_NOT_INITIALIZED;
   }

   return found_offset;
}

/**
 * @function ramdump_configure_ramdump_area
 *
 * @brief Configures the ramdump area.
 *
 * @input base Base of the dump area.
 * @input size Size of the dump area.
 * @input work_area_size Size of the work space in the heap.
 * @input pmd_work_area_size Size of space to inject between book keeping
 * data and dump storage area for storage compatibility with old pmd.
 *
 * @return Returns RAMDUMP_SUCCESS on success. Returns error code
 * on failure.
 */
static RamdumpStatus
ramdump_configure_ramdump_area(OSADDRESS base, U32 size,
			       U32 work_area_size, U32 pmd_work_area_size)
{
   RamdumpStatus ret_code = RAMDUMP_SUCCESS;

   if (DUMP_LOCK( INTERRUPTS_BLOCKED ))
   {
#ifdef RAMDUMP_DO_NOT_SWITCH_STACK
      configure_ramdump_area(base, size, work_area_size, pmd_work_area_size);
#else
      ramdump_stack_switch_args.arg1 = (long)base;
      ramdump_stack_switch_args.arg2 = (long)size;
      ramdump_stack_switch_args.arg3 = (long)work_area_size;
      ramdump_stack_switch_args.arg4 = (long)pmd_work_area_size;

      ramdump_switch_stack((U32)RAMDUMP_STACK_TOP,
                           (StackSwitchFunc)configure_ramdump_area,
                           &ramdump_stack_switch_args);
#endif
      ret_code = RAMDUMP_SUCCESS;

      DUMP_UNLOCK( INTERRUPTS_BLOCKED );
   }
   else
   {
      ret_code = RAMDUMP_ERECURSIVE_DUMP_CALL;
   }

   return ret_code;
}

/**
 * @function ramdump_set_watchdog_function
 *
 * @brief Sets a watchdog callback function to be called during dumping.
 *
 * @long Watchdog support is provided if dumps are time consuming.
 * A callback function can be installed in supervisor mode,
 * to be called at regular intervals during the dump phase.
 * It is an error to call this function in non-supervisor mode.
 * Installing a NULL pointer disables watchdog support.
 *
 * @input f Pointer to the watchdog callback function to install.
 *
 * @return If the function succeeds in setting the watchdog callback
 * function pointer, the return value is RAMDUMP_SUCCESS. If the ramdump
 * area is uninitialized, the function fails and returns the error code
 * RAMDUMP_ERAMDUMP_NOT_INITIALIZED.
 *
 * @seealso
 *
 * @example
 *
 */
RamdumpStatus
ramdump_set_watchdog_function(RamdumpWatchdogCb *f)
{
   RamdumpStatus ret_code = RAMDUMP_SUCCESS;

   if (DUMP_LOCK( INTERRUPTS_BLOCKED ))
   {
      if (ramdump_initialized != 0)
      {
#ifdef RAMDUMP_DO_NOT_SWITCH_STACK
         set_watchdog_func(f, &ret_code);
#else
         ramdump_stack_switch_args.arg1 = (long)f;
         ramdump_stack_switch_args.arg2 = (long)&ret_code;

         ramdump_switch_stack((U32)RAMDUMP_STACK_TOP,
                              (StackSwitchFunc)set_watchdog_func,
                              &ramdump_stack_switch_args);
#endif
         DUMP_UNLOCK( INTERRUPTS_BLOCKED );
      }
      else
      {
         /* Ramdump area has not been initialized */
         ret_code = RAMDUMP_ERAMDUMP_NOT_INITIALIZED;
      }
   }
   else
   {
      ret_code = RAMDUMP_ERECURSIVE_DUMP_CALL;
   }

   return ret_code;
}

/* -------------------------------------------------------------
 * Private interface functions, for use by FMD
 * -------------------------------------------------------------
 */

/**
 * @function ramdump_private_start_new_dump
 *
 * @brief Starts a new dump.
 *
 * @long This function starts a new dump after the last one in
 * the post-mortem dump area. Subsequent calls to ramdump_dump_xxx
 * will write information in the same dump.
 *
 * @long If several dumps are preferred, call this function to start a
 * new dump.
 *
 * @long NOTE: This function does not switch the stack.
 *
 * @return If the function succeeds, RAMDUMP_SUCCESS is returned.
 * If the dump function was called recursively, the error code
 * RAMDUMP_ERECURSIVE_DUMP_CALL is returned. If the ramdump area is
 * uninitialized, the error code RAMDUMP_ERAMDUMP_NOT_INITIALIZED
 * is returned.
 *
 */
RamdumpStatus
ramdump_private_start_new_dump(void)
{
	RamdumpStatus return_code = RAMDUMP_SUCCESS;


	if (DUMP_LOCK( INTERRUPTS_BLOCKED ))
	{
		if (ramdump_initialized != 0)
		{
			advance_dump_id();
		}
		else
		{
			return_code = RAMDUMP_ERAMDUMP_NOT_INITIALIZED;
		}

		DUMP_UNLOCK( INTERRUPTS_BLOCKED );
	}
	else
	{
		/* Return without lock and enabled interrupts */
		return_code = RAMDUMP_ERECURSIVE_DUMP_CALL;
	}

	/* Return without lock and enabled interrupts */
	return return_code;
}

/**
 * @function ramdump_private_clear_dumps
 *
 * @brief Clears dump area.
 *
 * @long This function clears the dump area.
 *
 * @long NOTE: This function does not switch the stack.
 *
 * @return If the function succeeds, RAMDUMP_SUCCESS is returned.
 * If the dump function was called recursively, the error code
 * RAMDUMP_ERECURSIVE_DUMP_CALL is returned. If the ramdump area is
 * uninitialized, the error code RAMDUMP_ERAMDUMP_NOT_INITIALIZED
 * is returned.
 *
 */
RamdumpStatus
ramdump_private_clear_dumps(void)
{
    RamdumpStatus return_code = RAMDUMP_SUCCESS;

    if (DUMP_LOCK( INTERRUPTS_ENABLED ))
    {
        if (ramdump_initialized != 0)
        {
            clear_dumps();
        }
        else
        {
            return_code = RAMDUMP_ERAMDUMP_NOT_INITIALIZED;
        }

        DUMP_UNLOCK( INTERRUPTS_ENABLED );
    }
    else
    {
        /* Return without lock and enabled interrupts */
        return_code = RAMDUMP_ERECURSIVE_DUMP_CALL;
    }

    /* Return without lock and enabled interrupts */
    return return_code;
}

/**
 * @function ramdump_private_dump_error_info
 *
 * @brief Writes information about an error event into the current dump.
 *
 * @long Writes information about an error event into the current dump.
 *
 * @long NOTE: This function does not switch the stack.
 *
 * @input description User-provided optional text. The string is only
 * copied up to the first terminator character.
 * @input user_called A parameter to the error handler.
 * @input error_code A parameter to the error handler.
 * @input extra A parameter to the error handler
 *
 * @return If the function succeeds, RAMDUMP_SUCCESS is returned.
 * If the current dump exceeds its size limit, the error code
 * RAMDUMP_ECURR_DUMP_TOO_LARGE is returned. If the description pointer
 * is invalid, the error code RAMDUMP_EMEMORY_ACCESS_NOK is returned.
 * If the dump function was called recursively, the error code
 * RAMDUMP_ERECURSIVE_DUMP_CALL is returned. If the ramdump area is
 * uninitialized, the error code RAMDUMP_ERAMDUMP_NOT_INITIALIZED
 * is returned. RAMDUMP_EUNABLE_TO_GET_CURRENT_PROCESS_INFO is returned
 * if the function is unable to get information for the current process.
 */
RamdumpStatus
ramdump_private_dump_error_info(const char *description,
         OSBOOLEAN user_called, OSERRCODE error_code,
         OSERRCODE extra)
{
	RamdumpStatus ret_code = RAMDUMP_SUCCESS;

	if (DUMP_LOCK( INTERRUPTS_BLOCKED ))
	{
		if (ramdump_initialized != 0)
		{
			dump_error_info(description, user_called, error_code,
							extra, &ret_code);
		}
		else
		{
			/* Ramdump area has not been initialized */
			ret_code = RAMDUMP_ERAMDUMP_NOT_INITIALIZED;
		}

		DUMP_UNLOCK( INTERRUPTS_BLOCKED );
	}
	else
	{
		ret_code = RAMDUMP_ERECURSIVE_DUMP_CALL;
	}

	return ret_code;
}

/**
 * @function ramdump_private_dump_text_info
 *
 * @brief Writes an arbitrary textual description into the current dump.
 *
 * @long Writes an arbitrary textual description into the current dump.
 *
 * @long NOTE: This function does not switch the stack.
 *
 * @input description User-provided optional text. The string
 * is only copied up to the first terminator character.
 *
 * @return If the function succeeds, RAMDUMP_SUCCESS is returned.
 * If the current dump exceeds its size limit, the error code
 * RAMDUMP_ECURR_DUMP_TOO_LARGE is returned. If the description pointer
 * is invalid, the error code RAMDUMP_EMEMORY_ACCESS_NOK is returned.
 * If the dump function was called recursively, the error code
 * RAMDUMP_ERECURSIVE_DUMP_CALL is returned. If the ramdump area is
 * uninitialized, the error code RAMDUMP_ERAMDUMP_NOT_INITIALIZED
 * is returned.
 */
RamdumpStatus
ramdump_private_dump_text_info(const char *description)
{
	RamdumpStatus ret_code = RAMDUMP_SUCCESS;

	if (DUMP_LOCK( INTERRUPTS_ENABLED ))
	{
		if (ramdump_initialized != 0)
		{
			dump_text_info(description, &ret_code);
		}
		else
		{
			/* Ramdump area has not been initialized */
			ret_code = RAMDUMP_ERAMDUMP_NOT_INITIALIZED;
		}
		DUMP_UNLOCK( INTERRUPTS_ENABLED );
	}
	else
	{
		ret_code = RAMDUMP_ERECURSIVE_DUMP_CALL;
	}

	return ret_code;
}

/**
 * @function ramdump_private_dump_memory_image
 *
 * @brief Dumps a section of memory to the current dump.
 *
 * @long Dumps a section of memory to the current dump.
 *
 * @long NOTE: This function does not switch the stack.
 *
 * @input description User-provided optional text. The string is only
 * copied up to the first terminator character.
 * @input start_address Start address of memory to dump.
 * @input length Number of bytes to dump.
 *
 * @return If the function succeeds, RAMDUMP_SUCCESS is returned.
 * If the current dump exceeds its size limit, the function fails
 * and returns the error code RAMDUMP_ECURR_DUMP_TOO_LARGE. If the memory
 * access fails or the description pointer is invalid, the error code
 * RAMDUMP_EMEMROY_ACCESS_NOK is returned. If the dump function was called
 * recursively, the error code RAMDUMP_ERECURSIVE_DUMP_CALL is returned.
 * If the ramdump area is unitialized, the error code
 * RAMDUMP_ERAMDUMP_NOT_INITIALIZED is returned.
 */
RamdumpStatus
ramdump_private_dump_memory_image(const char *description,
           OSADDRESS start_address, U32 length)
{
	RamdumpStatus ret_code = RAMDUMP_SUCCESS;

	if (DUMP_LOCK( INTERRUPTS_ENABLED ))
	{

		if (ramdump_initialized != 0)
		{
			dump_memory_image(description, start_address, length, &ret_code);
		}
		else
		{
			/* Dump area has not been initialized */
			ret_code = RAMDUMP_ERAMDUMP_NOT_INITIALIZED;
		}

		DUMP_UNLOCK( INTERRUPTS_ENABLED );
	}
	else
	{
		ret_code = RAMDUMP_ERECURSIVE_DUMP_CALL;
	}

	return ret_code;
}

/**
 * @function ramdump_private_set_watchdog_function
 *
 * @brief Sets a watchdog callback function to be called during dumping.
 *
 * @long Watchdog support is provided if dumps are time consuming.
 * A callback function can be installed in supervisor mode,
 * to be called at regular intervals during the dump phase.
 * It is an error to call this function in non-supervisor mode.
 * Installing a NULL pointer disables watchdog support.
 *
 * @long NOTE: This function does not switch the stack.
 *
 * @input f Pointer to the watchdog callback function to install.
 *
 * @return If the function succeeds in setting the watchdog callback
 * function pointer, the return value is RAMDUMP_SUCCESS. If the ramdump
 * area is uninitialized, the function fails and returns the error code
 * RAMDUMP_ERAMDUMP_NOT_INITIALIZED.
 *
 * @seealso
 *
 * @example
 *
 */
RamdumpStatus
ramdump_private_set_watchdog_function(RamdumpWatchdogCb *f)
{
	RamdumpStatus ret_code = RAMDUMP_SUCCESS;

	if (DUMP_LOCK( INTERRUPTS_BLOCKED ))
	{
		if (ramdump_initialized != 0)
		{
			set_watchdog_func(f, &ret_code);
			DUMP_UNLOCK( INTERRUPTS_BLOCKED );
		}
		else
		{
			/* Ramdump area has not been initialized */
			ret_code = RAMDUMP_ERAMDUMP_NOT_INITIALIZED;
		}
	}
	else
	{
		ret_code = RAMDUMP_ERECURSIVE_DUMP_CALL;
	}

	return ret_code;
}

/**
 * @function ramdump_private_dump_request
 *
 * @brief Writes the request signal of a transaction to the current dump.
 *
 * @long This function is meant for opening the dump of a signal
 * transaction (request reply1 [... replyN endmark]) by writing the request
 * signal to the dump. Any subsequent replies should be dumped using the
 * ramdump_private_dump_transaction_reply function. The endmark, if any,
 * should be dumped using the ramdump_private_transaction_end function.
 * Note that the ramdump_private_transaction_end function should be called
 * in order to close the transaction dump block, even if the transaction
 * has no endmark signal.
 *
 * @long NOTE: This function does not switch the stack before proceeding.
 *
 * @input request Pointer to a request signal in FmmSignal format. Signal
 * will be written in its entirety to the dump.
 * @input compress Request signal (and subsequent reply signals, if any)
 * will be written to the dump using compression if this parameter is non-zero.
 *
 * @return If the function succeeds, RAMDUMP_SUCCESS is returned.
 * If the current dump exceeds its size limit, the function fails
 * and returns the error code RAMDUMP_ECURR_DUMP_TOO_LARGE.
 * If the compression operation fails, RAMDUMP_ECOMPRESSION_FAILED is returned.
 * If the dump function was called recursively, the error code
 * RAMDUMP_ERECURSIVE_DUMP_CALL is returned.
 * If the ramdump area is unitialized, the error code
 * RAMDUMP_ERAMDUMP_NOT_INITIALIZED is returned.
 */
RamdumpStatus
ramdump_private_dump_request(const FmmSignal *request,
                             enum RamdumpCompressValue compression)
{
	RamdumpStatus ret_code = RAMDUMP_SUCCESS;

#ifndef RAMDUMP_USE_ZLIB
   if (compression == RAMDUMP_COMPRESSION_ZLIB)
   {
      return RAMDUMP_EOPERATION_REQUIRES_ZLIB;;
   }
#endif

	if (request != NULL)
	{
		if (DUMP_LOCK( INTERRUPTS_ENABLED ))
		{

			if (ramdump_initialized != 0)
			{
#ifdef RAMDUMP_USE_ZLIB
				if (compression == RAMDUMP_COMPRESSION_ZLIB)
				{
					int compr_status = open_compression_stream(&compr_stream);
					if (compr_status != Z_OK)
					{
						ret_code = RAMDUMP_ENO_COMPRESSION_STREAM;
					}
					else
					{
						dump_request(&compr_stream, request, compression, &ret_code);
					}
				}
				else
#endif
				{
					dump_request(&compr_stream, request, 0, &ret_code);
				}
			}
			else
			{
				/* Dump area has not been initialized */
				ret_code = RAMDUMP_ERAMDUMP_NOT_INITIALIZED;
			}

			DUMP_UNLOCK( INTERRUPTS_ENABLED );
		}
		else
		{
			ret_code = RAMDUMP_ERECURSIVE_DUMP_CALL;
		}
	}
	else
	{
		ret_code = RAMDUMP_EINVALID_ARGUMENT;
	}

	return ret_code;
}

/**
 * @function ramdump_private_dump_reply
 *
 * @brief Writes a reply signal of a transaction to the current dump.
 *
 * @long This function is meant for writing a reply of a transaction
 * (request reply1 [... replyN endmark]). A requirement for using this function
 * is that the last thing written to the current dump was either a request
 * signal (using ramdump_private_dump_transaction_begin) or a reply signal
 * (using ramdump_private_dump_transaction_reply).
 *
 * @long NOTE: This function does not switch the stack before proceeding.
 *
 * @input reply Pointer to a reply or endmark signal in FmmSignal format.
 * The signal will be written in its entirety to the dump.
 *
 * @return If the function succeeds, RAMDUMP_SUCCESS is returned.
 * If the current dump exceeds its size limit, the function fails
 * and returns the error code RAMDUMP_ECURR_DUMP_TOO_LARGE.
 * If the compression operation fails, RAMDUMP_ECOMPRESSION_FAILED is returned.
 * If the dump function was called recursively, the error code
 * RAMDUMP_ERECURSIVE_DUMP_CALL is returned.
 * If the ramdump area is unitialized, the error code
 * RAMDUMP_ERAMDUMP_NOT_INITIALIZED is returned.
 */
RamdumpStatus
ramdump_private_dump_reply(const FmmSignal *reply)
{
	RamdumpStatus ret_code = RAMDUMP_SUCCESS;

	if (reply != NULL)
	{
		if (DUMP_LOCK( INTERRUPTS_ENABLED ))
		{

			if (ramdump_initialized != 0)
			{
				dump_reply(&compr_stream, reply, &ret_code);
			}
			else
			{
				/* Dump area has not been initialized */
				ret_code = RAMDUMP_ERAMDUMP_NOT_INITIALIZED;
			}

			DUMP_UNLOCK( INTERRUPTS_ENABLED );
		}
		else
		{
			ret_code = RAMDUMP_ERECURSIVE_DUMP_CALL;
		}
	}
	else
	{
		ret_code = RAMDUMP_EINVALID_ARGUMENT;
	}

	return ret_code;
}

/**
 * @function ramdump_private_dump_endmark
 *
 * @brief Writes an endmark signal of a transaction to the current dump.
 *
 * @long This function is meant for writing an endmark signal, if any, of
 * a transaction (request reply1 [... replyN endmark]) and also for closing
 * the transaction dump block. A requirement for using this function is that
 * the last thing written to the current dump was either a request signal
 * (using ramdump_private_dump_transaction_begin) or a reply signal
 * (using ramdump_private_dump_transaction_reply).
 *
 * @long NOTE: This function does not switch the stack before proceeding.
 *
 * @input endmark Pointer to a reply or endmark signal in FmmSignal format.
 * The signal will be written in its entirety to the dump. If the transaction
 * has no endmark, this parameter should be NULL.
 * @input status A status value indicating the status of the dumped signal
 * transaction. Zero is reserved for a success status value. Any non-zero
 * value indicates a failure status value. Until this function has been called,
 * the status value present in the dump for the unfinished transaction will
 * contain the default failure status value 0xBADCODE. Any further
 * differentiation of the non-zero status value space is up to the user of
 * the function.
 *
 * @return If the function succeeds, RAMDUMP_SUCCESS is returned.
 * If the current dump exceeds its size limit, the function fails
 * and returns the error code RAMDUMP_ECURR_DUMP_TOO_LARGE.
 * If the compression operation fails, RAMDUMP_ECOMPRESSION_FAILED is returned.
 * If the dump function was called recursively, the error code
 * RAMDUMP_ERECURSIVE_DUMP_CALL is returned.
 * If the ramdump area is unitialized, the error code
 * RAMDUMP_ERAMDUMP_NOT_INITIALIZED is returned.
 */
RamdumpStatus
ramdump_private_dump_endmark(const FmmSignal *endmark, int32_t status)
{
	RamdumpStatus ret_code = RAMDUMP_SUCCESS;

	if (DUMP_LOCK( INTERRUPTS_ENABLED ))
	{

		if (ramdump_initialized != 0)
		{
			dump_endmark(&compr_stream, endmark, status, &ret_code);
#ifdef RAMDUMP_USE_ZLIB
			close_compression_stream(&compr_stream);
#endif
		}
		else
		{
			/* Dump area has not been initialized */
			ret_code = RAMDUMP_ERAMDUMP_NOT_INITIALIZED;
		}

		DUMP_UNLOCK( INTERRUPTS_ENABLED );
	}
	else
	{
		ret_code = RAMDUMP_ERECURSIVE_DUMP_CALL;
	}

	return ret_code;
}

