/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2005-2006 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

/**
 * @file ramdump_cmds.c
 *
 * @brief Shell commands for the ramdump.
 */

#include "shell.h"
#include "stdio.h"
#include "ramdump.h"
#include "ramdump_heap.h"
#include "ramlog.h"
#include "stdlib.h"
#include "errno.h"
#include "time.h"

#define SAVE_BUFFER_SIZE 1024
#define MIN_SIZE(a, b) ((a) < (b))?(a):(b)
#define RAMDUMPSAVE_USAGE "Usage: ramdumpsave <dump id> <file>\n"

/**
 * @function parse_UL
 *
 * @brief Parses a number, hexadecimal or decimal
 *
 * @return Non-zero if OK, zero if error.
 */
static int
parse_UL(const char *str, unsigned long *value)
{
   unsigned long val = 0;
   const char *p;
   int c;
   unsigned d;
      
   if (strncmp(str, "0x", 2) == 0 ||
       strncmp(str, "0X", 2) == 0)
   {
      /* Hex */
      if (str[ 2 ] == '\0') /* only "0x" is bad */
	 return 0; /* Bad format */
    
      for (p = &str[ 2 ]; *p != '\0'; p++) {
	 c = *p;

	 if (c >= '0' && c <= '9') {
	    d = c - '0';
	 } else if (c >= 'a' && c <= 'f') {
	    d = c - 'a' + 10;
	 } else if (c >= 'A' && c <= 'F') {
	    d = c - 'A' + 10;
	 } else {
	    return 0; /* Bad format */
	 }

	 val = (val << (unsigned int) 4) | (d & 0xf);
      }
   } else {
      /* Decimal */
      for (p = &str[ 0 ]; *p != '\0'; p++) {
	 c = *p;
	 
	 if (c >= '0' && c <= '9') {
	    d = c - '0';
	 } else {
	    return 0; /* Bad format */
	 }

	 val = val * 10 + d;
      }
   }

   *value = val;
   return 1;
}

/**
 * @function cmd_ramdumps
 *
 * @brief Lists dumps in memory.
 *
 * @input argc Number of command line arguments
 * @input argv Array of command line arguments
 *
 * @returns Shell command return code.
 */
static int
cmd_ramdumps(int argc, char *argv[])
{
   unsigned int i;
   unsigned int j;
   const char *column_header = "%8s %15s  %15s %s\n";
   const char *column_format = "%8u %15s  %15lu %s, %u ms\n";

   RamdumpDumpId id;
   char time_stamp[32];
   char dump_id[16]; 
   struct RamdumpDumpInfo dump_info;
   RamdumpStatus status;
   struct tm tm_time;
   time_t time;

   id = ramdump_get_dump_id(0);
   if (id == 0)
   {
      fprintf(stdout, "No dumps found.\n");
   }
   else
   {   
      printf(column_header,
	     "Nth Last", "Dump id", "Size (bytes)", "Time stamp");
   
      for (i = 0; i < RAMDUMP_MAX_DUMPS_INFO; i++)
      {
	 id = ramdump_get_dump_id(i);
	 if (id == 0)
	    break;

	 status = ramdump_get_dump_info(id, &dump_info);
	 if (status != RAMDUMP_SUCCESS)
	    break;

#if defined(softOSE) && defined(clock)
#undef clock
#endif

    time = dump_info.clock;

	 /* Get user-friendly time */
	 (void) localtime_r(&time, &tm_time);
	 (void) asctime_r(&tm_time, time_stamp, 31);

	 for (j = 0; j < 31; j++)
	 {
	    if (time_stamp[j] == '\n')
	    {
	       time_stamp[j] = '\0';
	       break;
	    }
	 }
	 
	 snprintf(dump_id, 15, "0x%lx", dump_info.dump_id);
	 dump_id[15] = '\0';
	 
	 fprintf(stdout, column_format, i,
		dump_id, (unsigned long)dump_info.size,
		time_stamp,
		(unsigned int)(dump_info.microseconds/1000));
      
      }
   }

   return 0;
}

/**
 * @function cmd_ramdumpsave
 *
 * @brief Saves specified dump to a file.
 *
 * @input argc Number of command line arguments
 * @input argv Array of command line arguments
 *
 * @returns Shell command return code.
 */
static int
cmd_ramdumpsave(int argc, char *argv[])
{
   RamdumpDumpId dump_id = 0;
   int result = 0;
   S32 bytes_read;
   S32 bytes_written;
   char *buffer;
   struct RamdumpDumpInfo info;
   RamdumpStatus status;
   S32 pos;
   FILE *file;

   if (argc != 3)
   {
      fprintf(stderr, RAMDUMPSAVE_USAGE);
      return RET_ERROR;
   }

   result = parse_UL(argv[1], (unsigned long *)&dump_id);
   if (result == 0)
   {
      fprintf(stderr, "Bad dump id. Use decimal (e.g. 17) or "
	      "hexadecimal (e.g. 0x11)\n");
      fprintf(stderr, RAMDUMPSAVE_USAGE);
      return RET_ERROR;
   }

   errno = 0;
   file = fopen(argv[2], "w");
   if (file == NULL)
   {
      perror("Error: Unable to open file");
      return RET_ERROR;
   }
   
   status = ramdump_get_dump_info(dump_id, &info);
   if (status != RAMDUMP_SUCCESS)
   {
      fprintf(stderr, "Error: Unable to retrieve dump info.\n");
      return RET_ERROR;
   }
   
   buffer = (char *)malloc(SAVE_BUFFER_SIZE * sizeof(char));
   
   for (pos = 0; pos < (S32)info.size; pos += bytes_read)
   {
      bytes_read = ramdump_read_dump(dump_id, pos,
				     (void *)buffer,
				     MIN_SIZE(SAVE_BUFFER_SIZE,
					      info.size - pos));
      
      bytes_written = fwrite(buffer, 1, bytes_read, file);
      if (bytes_read != bytes_written)
      {
	 fprintf(stderr, "Error: Unable to save dump to file.\n");
	 result = RET_ERROR;
	 break;
      }
   }

   fprintf(stdout, "Saved.\n");

   free(buffer);

   fclose(file);
   
   return result;
}

/**
 * @function cmd_ramdumpclear
 *
 * @brief Clears all dumps.
 *
 * @input argc Number of command line arguments
 * @input argv Array of command line arguments
 *
 * @returns Shell command return code.
 */
static int
cmd_ramdumpclear(int argc, char *argv[])
{
   int result = 0;
   RamdumpStatus status;
    
   status = ramdump_clear_dumps();
   if (status != RAMDUMP_SUCCESS)
   {
      fprintf(stderr, "Error: Unable to clear dumps.\n");
      return RET_ERROR;
   }
    
   status = ramdump_start_new_dump();
   if (status != RAMDUMP_SUCCESS)
   {
      fprintf(stderr, "Error: Unable to start new dump.\n");
      return RET_ERROR;
   }

   fprintf(stdout, "Cleared all dumps.\n");   
    
   return result;
}

/**
 * @function add_ramdump_cmds
 *
 * @brief Adds ramdump shell commands.
 */
void
add_ramdump_cmds(void)
{
   shell_add_cmd("ramdumps", "ramdumps", "List most recent dumps",
		 cmd_ramdumps);
   shell_add_cmd("ramdumpsave", "ramdumpsave <dump id> <file>",
		 "Save the contents of a dump to a file.",
		 cmd_ramdumpsave);
   shell_add_cmd("ramdumpclear", "ramdumpclear", "Clear all dumps.",
         cmd_ramdumpclear);
}
