/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2005-2006 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

/**
 * @file ramdump_format.h
 *
 * @brief Internal header file that contains the ramdump file format.
 */


#ifndef _RAMDUMP_FORMAT_H
#define _RAMDUMP_FORMAT_H

/*
  The following grammar describes chunks in the dump format.
  Chunks are aligned to 32 bits, but the dump can be saved in either
  big endian or little endian format.

  Chunk		::=	ID Size U32*
  ID		::=	U32
  Size		::=	U32
  Dump		::=	'FORM' Size 'PMD ' (Vers | DumpId | BlockList)*
  Vers		::=	'VERS' Size(4) U32(<version>)
  DumpId	::=	'DPID' Size(4) U32(<dump id>)
  BlockList	::=	'LIST' Size 'BLOC' (BlockForm)*
  BlockForm	::=	'FORM' Size 'BLOC' (BlockChunk)*
  BlockChunk	::=	BlockHead | ErrorForm | TextForm | MemoryForm
  BlockHead	::=	'BLHD' Size(16) U32(<dump id>) U32(<block no>)
			U32(<seconds since Jan 1 1970, 00:00>)
			U32(<additional microseconds>)
  ErrorForm	::=	'FORM' Size 'ERBL' (Descr | ErrInfo)*
  Descr		::=	'DESC' Size(<size of text>) <text>
  ErrInfo	::=	'ERIN' Size(16) U32(<user_called>) U32(<error_code>)
  			U32(<extra>) U32(<curr_proc>)
  TextForm	::=	'FORM' Size 'TXBL' (Descr)*
  MemoryForm	::=	'FORM' Size 'MBL ' (Descr | (MemDef (Data | Zdata)))*
  MemDef	::=	'MHD' Size(8) U32(<start>) U32(<length>)
  Data		::=	'DATA' Size(<length of data>) <data>
  Zdata		::=	'FORM' Size 'ZDAT' Zdef Data
  Zdef		::=	'ZDEF' Size(4 + <size of parameters>)
  			U32(<method>) <parameters>

*/


/*
 * C++ needs to know that this is C code.
 */

#ifdef __cplusplus
extern "C" {
#endif

#include "osetypes.h"

/**
 * @macro RAMDUMP_COMPRESS_NONE
 *
 * @brief Flag that specifies that no compression should be used.
 */
#define RAMDUMP_COMPRESS_NONE       ((U32) 0x0)

/**
 * @macro RAMDUMP_COMPRESS_ZLIB_Z77
 *
 * @brief Flag that specifies that Z77 should be used for compression.
 */
#define RAMDUMP_COMPRESS_ZLIB_Z77   ((U32) 0x1)
   
/**
 * @macro IFF_ID
 *
 * @brief Turns four characters into an Iff Id
 *
 * @input a The first letter of the Iff Id
 * @input b The second letter of the Iff id
 * @input c The third letter of the Iff id
 * @input d The fourth letter of the Iff id
 *
 * @return Returns the letters as an Iff id.
 */
#define IFF_ID(a, b, c, d) ( ((U32) (a) << 24) | ((U32) (b) << 16) |\
                             ((U32) (c) << 8) | (U32) (d) )

/**
 * @macro RAMDUMP_TAG_FORM
 *
 * @brief The FORM iff id.
 */
#define RAMDUMP_TAG_FORM IFF_ID('F', 'O', 'R', 'M')

/**
 * @macro RAMDUMP_TAG_LIST
 *
 * @brief The LIST iff id.
 */
#define RAMDUMP_TAG_LIST IFF_ID('L', 'I', 'S', 'T')

/**
 * @macro RAMDUMP_TAG_PROP
 *
 * @brief The PROP iff id.
 */
#define RAMDUMP_TAG_PROP IFF_ID('P', 'R', 'O', 'P')

/**
 * @macro RAMDUMP_TAG_CAT_
 *
 * @brief The CAT_ iff id
 */
#define RAMDUMP_TAG_CAT_ IFF_ID('C', 'A', 'T', ' ')

/**
 * @macro RAMDUMP_TAG_FILL
 *
 * @brief The FILL iff id
 */
#define RAMDUMP_TAG_FILL IFF_ID(' ', ' ', ' ', ' ')

/**
 * @macro RAMDUMP_TAG_WILD
 *
 * @brief Used internally for matching chunks with wildcards
 */
#define RAMDUMP_TAG_WILD IFF_ID('*', '*', '*', '*')

/**
 * @macro RAMDUMP_TAG_PMD_
 *
 * @brief The PMD_ iff id.
 */
#define RAMDUMP_TAG_PMD_ IFF_ID('P', 'M', 'D', ' ')

/**
 * @macro RAMDUMP_TAG_BLOC
 *
 * @brief The BLOC iff id.
 */
#define RAMDUMP_TAG_BLOC IFF_ID('B', 'L', 'O', 'C')

/**
 * @macro RAMDUMP_TAG_DPID
 *
 * @brief The DPID iff id.
 */
#define RAMDUMP_TAG_DPID IFF_ID('D', 'P', 'I', 'D')

/**
 * @macro RAMDUMP_TAG_VERS
 *
 * @brief The VERS iff id.
 */
#define RAMDUMP_TAG_VERS IFF_ID('V', 'E', 'R', 'S')

/**
 * @macro RAMDUMP_TAG_BLHD
 *
 * @brief The BLHD iff id.
 */
#define RAMDUMP_TAG_BLHD IFF_ID('B', 'L', 'H', 'D')

/**
 * @macro RAMDUMP_TAG_DESC
 *
 * @brief The DESC iff id.
 */
#define RAMDUMP_TAG_DESC IFF_ID('D', 'E', 'S', 'C')

/**
 * @macro RAMDUMP_TAG_ERBL
 *
 * @brief The ERBL iff id.
 */
#define RAMDUMP_TAG_ERBL IFF_ID('E', 'R', 'B', 'L')

/**
 * @macro RAMDUMP_TAG_ERIN
 *
 * @brief The ERIN iff id.
 */
#define RAMDUMP_TAG_ERIN IFF_ID('E', 'R', 'I', 'N')

/**
 * @macro RAMDUMP_TAG_TXBL
 *
 * @brief The TXBL iff id.
 */
#define RAMDUMP_TAG_TXBL IFF_ID('T', 'X', 'B', 'L')


/**
 * @macro RAMDUMP_TAG_MBL_
 *
 * @brief The MBL_ iff id.
 */
#define RAMDUMP_TAG_MBL_ IFF_ID('M', 'B', 'L', ' ')

/**
 * @macro RAMDUMP_TAG_MHD_
 *
 * @brief The MHD_ iff id.
 */
#define RAMDUMP_TAG_MHD_ IFF_ID('M', 'H', 'D', ' ')

/**
 * @macro RAMDUMP_TAG_DATA
 *
 * @brief The DATA iff id.
 */
#define RAMDUMP_TAG_DATA IFF_ID('D', 'A', 'T', 'A')

/**
 * @macro RAMDUMP_TAG_ZDAT
 *
 * @brief The ZDAT iff id.
 */
#define RAMDUMP_TAG_ZDAT IFF_ID('Z', 'D', 'A', 'T')

/**
 * @macro RAMDUMP_TAG_ZDEF
 *
 * @brief The ZDEF iff id
 */
#define RAMDUMP_TAG_ZDEF IFF_ID('Z', 'D', 'E', 'F')

/**
 * @macro RAMDUMP_TAG_SGBL
 *
 * @brief The SGBL iff id
 */
#define RAMDUMP_TAG_SGBL IFF_ID('S', 'G', 'B', 'L')

/**
 * @macro RAMDUMP_TAG_SGHD
 *
 * @brief The SGHD iff id
 */
#define RAMDUMP_TAG_SGHD IFF_ID('S', 'G', 'H', 'D')

/**
 * @macro ASSERT_IFF_STRUCTURE
 *
 * @brief Makes sure the iff chunk structs are of the correct size.
 */
#define ASSERT_IFF_STRUCTURE() do {\
  assert(sizeof(struct RamdumpChunkHeader) == 8);\
  assert(sizeof(struct RamdumpHeaderFORM) == 12);\
  assert(sizeof(struct RamdumpHeaderLIST) == 12);\
  assert(sizeof(struct RamdumpHeaderCAT_) == 12);\
  assert(sizeof(struct RamdumpHeaderPROP) == 12);\
  assert(sizeof(struct RamdumpChunkVERS) == 12);\
  assert(sizeof(struct RamdumpChunkDPID) == 12);\
  assert(sizeof(struct RamdumpChunkBLHD) == 24);\
  assert(sizeof(struct RamdumpChunkERIN) == 24);\
  assert(sizeof(struct RamdumpChunkMHD_) == 16);\
  assert(sizeof(struct RamdumpChunkZDEF) == 36);\
  assert(sizeof(struct RamdumpChunkSGHD) == 20);\
} while (0);


/**
 * @type IffId
 *
 * @brief Represents an iff chunk id.
 */
typedef U32 IffId;

/**
 * @type RamdumpChunkHeader
 *
 * @brief An IFF chunk header structure.
 *
 * @field id Iff id of the chunk.
 * @field size Size of the chunk minus the header.
 */
struct RamdumpChunkHeader
{
  IffId id;
  U32 size;
};

/**
 * @type RamdumpHeaderFORM
 *
 * @brief Represents a FORM chunk.
 */
struct RamdumpHeaderFORM
{
  struct RamdumpChunkHeader chd;
  IffId form_type;
};

/**
 * @type RamdumpHeaderLIST
 *
 * @brief Represents a LIST chunk
 */
struct RamdumpHeaderLIST
{
  struct RamdumpChunkHeader chd;
  IffId cont_type;
};

/**
 * @type RamdumpHeaderCAT_
 *
 * @brief Represents a CAT_ chunk
 */
struct RamdumpHeaderCAT_
{
  struct RamdumpChunkHeader chd;
  IffId cont_type;
};

/**
 * @type RamdumpHeaderPROP
 *
 * @brief Represents a PROP chunk.
 */
struct RamdumpHeaderPROP
{
  struct RamdumpChunkHeader chd;
  IffId form_type;
};

/**
 * @type RamdumpChunkVERS
 *
 * @brief Represents a VERS chunk.
 */
struct RamdumpChunkVERS
{
  struct RamdumpChunkHeader chd;
  U32 version;
};

/**
 * @type RamdumpChunkDESC
 *
 * @brief Represents a DESC chunk.
 */
struct RamdumpChunkDESC
{
  struct RamdumpChunkHeader chd;
  char descr[ 4 ];
};

/**
 * @type RamdumpChunkDPID
 *
 * @brief Represents a DPID chunk
 */
struct RamdumpChunkDPID
{
  struct RamdumpChunkHeader chd;
  U32 dump_id;
};

/**
 * @type RamdumpChunkBLHD
 *
 * @brief Represents a BLHD chunk.
 */
struct RamdumpChunkBLHD
{
  struct RamdumpChunkHeader chd;
  U32 dump_id;
  U32 block_no;
  U32 clock;   /* Seconds since Jan 1, 1970, 00:00:00 */
  U32 microseconds;
};

/**
 * @type RamdumpChunkERIN
 *
 * @brief Represents an ERIN chunk.
 */
struct RamdumpChunkERIN
{
  struct RamdumpChunkHeader chd;
  U32 user_called;
  U32 error_code;
  U32 extra;
  U32 curr_proc;
};

/**
 * @type RamdumpChunkMHD_
 *
 * @brief Represents an MHD_ chunk.
 */
struct RamdumpChunkMHD_
{
  struct RamdumpChunkHeader chd;
  U32 start_address;
  U32 length;
};

/**
 * @type RamdumpChunkZDEF
 *
 * @brief Represents an ZDEF chunk.
 */
struct RamdumpChunkZDEF
{
  struct RamdumpChunkHeader chd;
  U32 compression;
  U32 compr_parameters[ 4 ];
  U32 image_length;       /* Length of compressed image */
  U32 _reserved;
};

/**
 * @type RamdumpChunkSGHD
 * 
 * @brief Represents an SGHD chunk.
 */
struct RamdumpChunkSGHD
{
	struct RamdumpChunkHeader chd;
	U32 signo;
   U32 status;
   U32 length;
};

#ifdef __cplusplus
}
#endif

#endif /* # ifndef _RAMDUMP_FORMAT_H */
