/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2005-2006 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

#include "ose_err/ramdump_err.h"
#include "ramdump_heap.h"
#include "stdint.h"
#include "ramlog.h"
#include "sysparam.h"
#include "ose.h"

#define ALIGN_TO(value, size) ((((uint32_t)(value)) + \
                              ((uint32_t)(size)) - 1UL) & \
                              ~(((uint32_t)(size)) - 1UL))
#define BLK2PTR(x) (((BlockHeader *)x) + 1)
#define PTR2BLK(x) (((BlockHeader *)x) - 1)

typedef uint64_t HeapAlignment;

typedef union aligned_block_header
{
   struct
   {
      union aligned_block_header *next_block; /* Next block if in free list. */
      size_t size; /* Size of block expressed in number of
		      header sized blocks, including header. */
   } header;

   HeapAlignment x;
} BlockHeader;

/* Head of list of free blocks. */
static BlockHeader free_list_head;

/* List of free blocks. */
static BlockHeader *free_list = NULL;

/* Total heap size expressed in multiple
   of block header sized units. */
static size_t heap_size = 0;

/* Non-zero if heap has been initialized. */
static int heap_initialized = 0;

/**
 * @function expand_block
 *
 * @brief Attempts to expand block with adjacent blocks from free list.
 *
 * @long The function expand_block attempts to expand the specified block
 * by merging it with directly adjacent free blocks, if the size of the
 * expanded block exceeds or equals the minimum requested size.
 *
 * @input original_block Pointer to the block to expand.
 * @input min_units Minimum size of expanded block, as multiple
 * of block header sized units.
 *
 * @return Pointer to the expanded block, or a pointer to the original
 * unmodified block if there were no directly adjacent free blocks.
 * Returns NULL if request cannot be satisfied.
 */
static BlockHeader *
expand_block(BlockHeader *original_block, size_t min_units)
{
   BlockHeader *prev_block;
   BlockHeader *prev_prev_block;
   BlockHeader *block;
   size_t expanded_size;
   int adj_to_prev = 0;
   int adj_to_current = 0;
   
   if (!heap_initialized)
   {
      ramlog_printf("ramdump_heap: Call to ramdump heap before heap was"
		    " initialized.\n");
      return NULL;
   }
   
   if (original_block == NULL)
   {
      return NULL;
   }

   /* Is block header pointer out of bounds? */
   if (original_block < free_list
       || original_block + original_block->header.size > free_list + heap_size)
   {
      return NULL;
   }

   /* Find place in free list for block. */
   prev_prev_block = NULL;
   prev_block = &free_list_head;

   /* Iterate over free list. Upon exit from the loop,
      the expand block should be checked for adjacency to
      the current block and the previous block. */
   for (block = prev_block->header.next_block;
	block != NULL;
	prev_prev_block = prev_block, prev_block = block,
	   block = block->header.next_block)
   {
      /* Does freed block belong before current block? */
      if (original_block < block)
      {
	 break;
      }
   }

   /* Initiate calculation of expanded block size. */
   expanded_size = original_block->header.size;
   
   /* Does original block follow immediately after the previous block? */
   if (prev_block + prev_block->header.size == original_block)
   {
      /* Add size of previous block to total expanded block size. */
      expanded_size += prev_block->header.size;
      adj_to_prev = 1;
   }

   /* Does current block follow immediately after the original block? */
   if (original_block + original_block->header.size == block)
   {
      /* Add size of current block to total expanded block size. */
      expanded_size += block->header.size;
      adj_to_current = 1;
   }

   /* Is the expanded size too small? */
   if (expanded_size < min_units)
   {
      return NULL;
   }
   
   /* Does original block follow immediately after the previous block? */
   if (adj_to_prev)
   {
      /* Merge original block with previous block. */
      prev_prev_block->header.next_block = block;
      prev_block->header.next_block = NULL;
      prev_block->header.size += original_block->header.size;
      original_block = prev_block;
      prev_block = prev_prev_block;
   }

   /* Does current block follow immediately after the original block? */
   if (adj_to_current)
   {
      /* Merge original block with current block. */
      prev_block->header.next_block = block->header.next_block;
      block->header.next_block = NULL;
      original_block->header.size += block->header.size;
   }
   
   return original_block;
}

/**
 * @function ramdump_heap_init
 *
 * @brief Initializes heap.
 *
 * @long The function ramdump_heap_init initializes the heap by
 * creating a list of free memory blocks starting at the specified
 * base address. The amount of free memory initially available on
 * the heap is determined by the size parameter. The size of the
 * heap is fixed.
 *
 * @input base The start address of the heap memory area.
 * @input size The size of the heap memory area.
 *
 * @return Returns zero on success or one of
 * RAMDUMP_HEAP_EHEAP_AREA_SIZE_TOO_SMALL and
 * RAMDUMP_HEAP_ENO_BASE_ADDRESS_CONFIGURED on failure.
 *
 * @example
 * RamdumpHeapStatus status = ramdump_heap_init(0x10000, 1000);
 */
RamdumpHeapStatus
ramdump_heap_init(char *base, size_t size)
{
   char *aligned_base;
   size_t aligned_size;
   ptrdiff_t alignment_padding;
   
   if (base == NULL)
   {
      ramlog_printf("ramdump_heap: No base address configured.\n");
      return RAMDUMP_HEAP_ENO_BASE_ADDRESS_CONFIGURED;
   }

   /* Adjust base pointer to desired block alignment. */
   aligned_base = (unsigned char *)ALIGN_TO(base, sizeof(HeapAlignment));
   alignment_padding = aligned_base - base;
   aligned_size = size - alignment_padding;

   /* Calculate space in header sized units. */
   heap_size = (size + sizeof(BlockHeader) - 1) /
      sizeof(BlockHeader) - 1;

   /* Make sure heap area has room for at least
      one correctly aligned block header. */   
   if (heap_size < 1)
   {
      /* TODO: This should not just return. Or should it? */
      ramlog_printf("ramdump_heap: Heap area size too small.\n");
      return RAMDUMP_HEAP_ENO_BASE_ADDRESS_CONFIGURED;
   }

   /* Initialize list of free blocks at aligned base address. */
   free_list = (BlockHeader *)aligned_base;
   free_list_head.header.next_block = free_list;
   free_list_head.header.size = 0;

   /* Initially, there will be one free block that
      encompasses the entire heap memory area. */
   free_list->header.next_block = NULL;
   free_list->header.size = heap_size;

   /* Heap is now officially initialized. */
   heap_initialized = 1;

   return 0;
}

/**
 * @function ramdump_heap_malloc
 *
 * @include stddef.h
 *
 * @brief Allocates a region large enough to hold an object
 * of specified size.
 *
 * @long The function ramdump_heap_malloc allocates a region of memory
 * large enough to hold an object whose size (as measured by the
 * sizeof operator) equals the parameter size. A pointer to the
 * first element of the region is returned. The caller may then use
 * a cast operator to convert this pointer to another pointer type.
 * If it is impossible for some reason to perform the requested
 * allocation, then a NULL pointer is returned. If the requested
 * size is 0, then the function will return either a NULL pointer
 * or a non-NULL pointer that nonetheless must not be used to access
 * an object. The allocated memory is not initialized in any way,
 * so the caller cannot depend on its contents.
 *
 * @input size Size of allocated memory region.
 *
 * @return Pointer to the allocated memory region, or NULL if
 * the allocation can not be performed.
 *
 * @example
 * int *ptr = (int *)ramdump_heap_malloc(sizeof(int));
 */
void *
ramdump_heap_malloc(size_t size)
{
   size_t unit_count;
   BlockHeader *block;
   BlockHeader *prev_block;
   BlockHeader *new_block;

   if (!heap_initialized)
   {
      ramlog_printf("ramdump_heap: Call to ramdump heap before heap was"
		    " initialized.\n");
      return NULL;
   }
   
   if (size == 0)
   {
      return NULL;
   }

   /* Calculate requested size in header sized units, as the memory
      will be allocated in multiples of header sized units. */
   unit_count = (size + sizeof(BlockHeader) - 1) /
      sizeof(BlockHeader);

   /* Add one unit to the size for the header. */
   unit_count++;

   /* Iterate over free list in search of a suitable block. */
   new_block = NULL;
   prev_block = &free_list_head;

   for (block = prev_block->header.next_block;
	block != NULL;
	prev_block = block, block = block->header.next_block)
   {
      /* Is block big enough? */
      if (block->header.size >= unit_count)
      {
	 /* Is block the exact right size? */
	 if (block->header.size == unit_count)
	 {
	    /* Block is the exact right size.
	       Remove it from the free list. */
	    prev_block->header.next_block = block->header.next_block;
	    new_block = block;
	    break;
	 }
	 else
	 {
	    /* Block is bigger than requested block size. */

#ifdef RAMDUMP_HEAP_DIVIDE_FROM_BACK	    
	    /* Create a new block, of the requested size,
	       from the end of the old block. Change the
	       size field of the old block accordingly. */
	    block->header.size -= unit_count;
	    new_block = block + block->header.size;
	    new_block->header.next_block = NULL;
	    new_block->header.size = unit_count;
#else
	    /* Create a new block, of the requested size,
	       from the beginning of the old block and remove
	       it from the free list. */
	    new_block = block;
	    block = block + unit_count;
	    block->header.size = new_block->header.size - unit_count;
	    block->header.next_block = new_block->header.next_block;
	    new_block->header.size = unit_count;
	    new_block->header.next_block = NULL;
	    prev_block->header.next_block = block;
#endif
	    
	    break;
	 }
      }
   }

   /* Was a suitable free block found? */
   if (new_block != NULL)
   {
      /* Return pointer to free memory after block header. */
      return (void *)BLK2PTR(new_block);
   }

   /* No suitable free block was found. */
   return NULL;
}

/**
 * @function ramdump_heap_calloc
 *
 * @include stddef.h
 *
 * @brief Allocates a region of memory large enough to hold
 * an array of element_count elements, each of size element_size.
 *
 * @long The function ramdump_heap_calloc allocates a region of memory
 * large enough to hold an array of element_count elements, each
 * of size element_size (typically given by the sizeof operator).
 * The region of memory is cleared bitwise to zero, and a pointer
 * to the first element of the region is returned. If for some
 * reason it is impossible to perform the requested allocation,
 * or if element_count or element_size is zero, then the return
 * value is NULL.
 *
 * @input element_count Number of elements to allocate space for.
 * @input element_size The size of each element.
 *
 * @return Pointer to the first element in the allocated memory area.
 *
 * @example
 * int *int_array = (int *)ramdump_heap_calloc(100, sizeof(int));
 */
void *
ramdump_heap_calloc(size_t element_count, size_t element_size)
{
   unsigned char *pointer;
   size_t i;
   
   if (!heap_initialized)
   {
      ramlog_printf("ramdump_heap: Call to ramdump heap before heap was"
		    " initialized.\n");
      return NULL;
   }

   if (element_count == 0 || element_size == 0)
   {
      return NULL;
   }

   /* Allocate memory by calling ramdump_heap_malloc. */
   pointer = (unsigned char *)ramdump_heap_malloc(element_count * element_size);

   if (pointer != NULL)
   {
      /* Set allocated memory to zero. (Is memset() allowed?) */
      for (i = 0; i < element_count * element_size; i++)
      {
	 pointer[i] = 0;
      }
   }

   /* Return pointer to allocated memory. */
   return (void *)pointer;
}

/**
 * @function ramdump_heap_realloc
 *
 * @include stddef.h
 *
 * @brief Takes a pointer to a previously allocated memory
 * region and changes its size while preserving its contents.
 *
 * @long The function ramdump_heap_realloc takes a pointer to a memory
 * region previously allocated by one of the standard functions
 * and changes its size while preserving its contents. If necessary,
 * the contents are copied to a new memory region. A pointer to the
 * (possibly new) memory region is returned. If the request cannot
 * be satisfied, a null pointer is returned and the old region is
 * not disturbed. If the first argument to ramdump_heap_realloc is a null
 * pointer, then the function behaves like ramdump_heap_malloc. If the
 * pointer parameter is not NULL and the size parameter is zero, then
 * ramdump_heap_realloc returns either a NULL pointer or a pointer that
 * must not be used, and the old region is deallocated. If the new
 * size is smaller than the old size, then some of the old contents
 * at the end of the old region will be discarded. If the new size
 * is larger than the old size, then all of the old contents are
 * preserved and new space is added at the end. The new space is not
 * initialized in any way, and the caller must assume that it contains
 * garbage information. Whenever ramdump_heap_realloc returns a pointer
 * that is different from its first argument, the programmer should
 * assume that the old region of memory was free and should not be used.
 *
 * @input pointer Pointer to the memory region to reallocate.
 * @input size Size of the reallocated region.
 *
 * @return Pointer to the reallocated memory region or NULL if
 * the reallocation could not be performed.
 *
 * @example
 * char *ptr = (char *)ramdump_heap_malloc(100 * sizeof(char));
 * ptr = (char *)ramdump_heap_realloc(ptr, 200 * sizeof(char));
 */
void *
ramdump_heap_realloc(void *pointer, size_t size)
{
   BlockHeader *block;
   void *new_pointer = NULL;
   size_t unit_count;
   
   if (!heap_initialized)
   {
      ramlog_printf("ramdump_heap: Call to ramdump heap before heap was"
		    " initialized.\n");
      return NULL;
   }

   if (size == 0)
   {
      return NULL;
   }
   
   /* Is pointer NULL? */
   if (pointer == NULL)
   {
      /* Then realloc behaves like malloc. */
      return (void *)ramdump_heap_malloc(size);
   }
   
   /* Get pointer to block header. */
   block = PTR2BLK(pointer);

   /* Is block header pointer out of bounds? */
   if (block < free_list
       || block + block->header.size > free_list + heap_size)
   {
      return NULL;
   }

   /* Calculate requested size in header sized units. */
   unit_count = (size + sizeof(BlockHeader) - 1) /
      sizeof(BlockHeader);

   /* Add one unit for the header. */
   unit_count++;

   /* Compare requested size to original size. */
   if (unit_count == block->header.size)
   {
      /* The requested size is the SAME as the original size.
	 Modify nothing and return the original block. */
      new_pointer = pointer;
   }
   else if (unit_count < block->header.size)
   {
      /* The requested size is SMALLER than the original size.
	 The block needs to be cropped. */

      BlockHeader *freed_block;

      /* Free the unwanted space at the end of the block. */
      freed_block = block + unit_count;
      freed_block->header.next_block = NULL;
      freed_block->header.size = block->header.size - unit_count;
      ramdump_heap_free(BLK2PTR(freed_block));

      /* The original pointer to the start of the allocated area
	 has not changed, so return it. */
      new_pointer = pointer;
   }
   else
   {
      /* The requested size is LARGER than the original size.
	 The block needs to be expanded or replaced with a larger block. */

      BlockHeader *expanded_block;
      
      /* Try to expand block. */
      expanded_block = expand_block(block, unit_count);

      /* Was block expanded sufficiently? */
      if (expanded_block != NULL && expanded_block->header.size >= unit_count)
      {
	 size_t i;
	 
	 /* Was block expanded at the front? */
	 if (expanded_block < block)
	 {
	    /* Move block contents to front of expanded block.
	       (Is memmove() allowed?) */
	    for (i = 0;
		 i < block->header.size * sizeof(BlockHeader);
		 i++)
	    {
	       ((char *)expanded_block)[i] = ((char *)block)[i];
	    }
	 }

	 /* Is the expanded block larger than required? */
	 if (expanded_block->header.size > unit_count)
	 {
	    BlockHeader *freed_block;
	    
	    /* Free the superfluous end of the block. */
	    freed_block = expanded_block + unit_count;
	    freed_block->header.next_block = NULL;
	    freed_block->header.size = expanded_block->header.size
	       - unit_count;
	    ramdump_heap_free(BLK2PTR(freed_block));
	    expanded_block->header.size = unit_count;
	 }

	 new_pointer = BLK2PTR(expanded_block);
      }
      else
      {
	 size_t i;
	 size_t byte_count;
	 
	 /* Allocate a new block instead. */
	 new_pointer = (void *)ramdump_heap_malloc(size);
   
	 if (new_pointer == NULL)
	 {
	    return NULL;
	 }
      
	 /* Copy memory from old area to new area. (Is memcpy() allowed?) */
	 byte_count = (block->header.size - 1) * sizeof(BlockHeader);
	 for (i = 0; i < byte_count; i++)
	 {
	    ((char *)new_pointer)[i] = ((char *)pointer)[i];
	 }
	 
	 /* Free old memory block. */
	 ramdump_heap_free(pointer);
      }
   }
   
   return new_pointer;
}

/**
 * @function ramdump_heap_free
 *
 * @include stddef.h
 *
 * @brief Deallocates a previously allocated memory region.
 *
 * @long The function ramdump_heap_free deallocates a region of memory
 * previously allocated by ramdump_heap_malloc, ramdump_heap_calloc or
 * ramdump_heap_realloc. The argument of ramdump_heap_free must be a pointer
 * that is the same as a pointer previously returned by one of the
 * allocation functions. If the argument is a NULL pointer, then
 * the call has no effect. Once a region of memory has been freed,
 * it must not be used for any other purpose. The use of any pointer
 * into the region (a "dangling pointer") will have unpredictable
 * effects. Likewise, allocating a region of storange once, but
 * freeing it more than once, has unpredictable effects.
 *
 * @input pointer Pointer to a previously allocated region of memory
 * that will be freed.
 *
 * @example
 * char *ptr = (char *)ramdump_heap_malloc(100 * sizeof(char));
 * ramdump_heap_free(ptr);
 */
void
ramdump_heap_free(void *pointer)
{
   BlockHeader *freed_block;
   BlockHeader *block;
   BlockHeader *prev_block;

   if (!heap_initialized)
   {
      ramlog_printf("ramdump_heap: Call to ramdump heap before heap was"
		    " initialized.\n");
   }
   
   if (pointer == NULL)
   {
      return;
   }

   /* Get pointer to block header. */
   freed_block = PTR2BLK(pointer);

   /* Is block size invalid? */
   if (freed_block->header.size == 0)
   {
      return;
   }

   /* Is block header pointer out of bounds? */
   if (freed_block < free_list
       || freed_block + freed_block->header.size > free_list + heap_size)
   {
      return;
   }

   /* Find place in free list for block.
      Iterate over free list. Upon exit from the loop,
      the freed block should be inserted between the
      current block and the previous block. */
   prev_block = &free_list_head;
   for (block = prev_block->header.next_block;
	block != NULL;
	prev_block = block, block = block->header.next_block)
   {
      /* Does freed block belong before current block? */
      if (freed_block < block)
      {
	 break;
      }
   }

   /* Insert block between previous block and current block. */
   prev_block->header.next_block = freed_block;
   freed_block->header.next_block = block;

   /* Does freed block follow immediately after the previous block? */
   if (prev_block + prev_block->header.size == freed_block)
   {
      /* Merge freed block with previous block. */
      prev_block->header.next_block = block;
      prev_block->header.size += freed_block->header.size;
      freed_block = prev_block;
   }

   /* Does current block follow immediately after the freed block? */
   if (freed_block + freed_block->header.size == block)
   {
      /* Merge freed block with current block. */
      freed_block->header.next_block = block->header.next_block;
      freed_block->header.size += block->header.size;
   }
}

void
ramdump_heap_start_handler1(void)
{
   int result;
   long base;
   long size;

   result = sysparam_readattr_long("krn/ramdump/heap", "base", &base);
   if (result != 0)
   {
      ramlog_printf("ramdump: Unable to read parameter krn/ramdump/heap, "
		    "attribute base.\n");
      error2((OSERRCODE)(OSE_RAMDUMP_EERROR_READING_HEAP_SYSPARAM),
	     (OSERRCODE)NULL);
   }

   result = sysparam_readattr_long("krn/ramdump/heap", "size", &size);

   if (result != 0)
   {
      ramlog_printf("ramdump: Unable to read parameter krn/ramdump/heap, "
		 "attribute size.\n");
      error2((OSERRCODE)(OSE_RAMDUMP_EERROR_READING_HEAP_SYSPARAM),
	     (OSERRCODE)NULL);
   }

   result = ramdump_heap_init((char *)base, size);

   if (result != 0)
   {
      ramlog_printf("ramdump: Unable to init ramdump heap.\n");
      error2(OSE_RAMDUMP_EUNABLE_TO_INIT_RAMDUMP_HEAP, result);
   }
}



/* ------------ Test functions -------------- */

void ramdump_heap_get_free_list_stats(uint32_t *start, uint32_t *size);
uint32_t ramdump_heap_get_next_block(uint32_t prev_block_ptr,
				     uint32_t *is_free,
				     uint32_t *offset,
				     uint32_t *size);
uint32_t ramdump_heap_get_next_free_block(uint32_t prev_block_ptr,
					  uint32_t *offset,
					  uint32_t *size);
uint32_t ramdump_heap_is_free(uint32_t block_ptr);

void
ramdump_heap_get_free_list_stats(uint32_t *start, uint32_t *size)
{
   if (start != NULL)
      *start = (uint32_t)free_list;
   if (size != NULL)
      *size = heap_size * sizeof(BlockHeader);
}

uint32_t
ramdump_heap_get_next_block(uint32_t prev_block_ptr, uint32_t *is_free,
			    uint32_t *offset, uint32_t *size)
{
   BlockHeader *prev_block = (BlockHeader *)prev_block_ptr;
   BlockHeader *block;

   if (prev_block == NULL)
   {
      block = free_list;
   }
   else
   {
      block = prev_block + prev_block->header.size;
   }

   if (block < free_list + heap_size)
   {
      if (offset != NULL)
	 *offset = (block - free_list) * sizeof(BlockHeader);
      if (size != NULL)
	 *size = block->header.size * sizeof(BlockHeader);
      if (is_free != NULL)
	 *is_free = ramdump_heap_is_free((uint32_t)block);
      return (uint32_t)block;
   }

   return 0;
}

uint32_t
ramdump_heap_get_next_free_block(uint32_t prev_block_ptr, uint32_t *offset,
				 uint32_t *size)
{
   BlockHeader *prev_block = (BlockHeader *)prev_block_ptr;
   BlockHeader *block;

   if (prev_block == NULL)
   {
      block = free_list_head.header.next_block;
   }
   else
   {
      block = prev_block->header.next_block;
   }

   if (block != NULL)
   {
      if (offset != NULL)
	 *offset = (block - free_list) * sizeof(BlockHeader);
      if (size != NULL)
	 *size = block->header.size * sizeof(BlockHeader);
      return (uint32_t)block;
   }

   return 0;
}

uint32_t
ramdump_heap_is_free(uint32_t block_ptr)
{
   BlockHeader *prev_block;
   BlockHeader *block;
   
   prev_block = &free_list_head;

   for (block = prev_block->header.next_block;
	block != NULL;
	prev_block = block, block = block->header.next_block)
   {
      if (((uint32_t)block) == ((uint32_t)block_ptr))
      {
	 return 1;
      }
   }
   return 0;
}

