/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2005-2006 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

/**
 * @file ramdump_internal.h
 *
 * @brief This file is an internal header file.
 *
 */

#ifndef _RAMDUMP_INTERNAL_H
#define _RAMDUMP_INTERNAL_H

#ifdef __cplusplus
extern "C" {
#endif

#include "ramdump.h"
#include "ramdump_format.h"

#define RAMDUMP_VERSION             ((U32) 0x0100)

/*
 * BIOS function codes for trapping into dump functions
 */
#define RAMDUMP_START_NEW_DUMP          0x1
#define RAMDUMP_DUMP_ERROR_INFO         0x2
#define RAMDUMP_DUMP_TEXT_INFO          0x3
#define RAMDUMP_DUMP_MEMORY_IMAGE       0x4
#define RAMDUMP_GET_DUMP_ID             0x6
#define RAMDUMP_GET_DUMP_INFO           0x7
#define RAMDUMP_GET_DUMPS               0x8
#define RAMDUMP_READ_DUMP               0x9
#define RAMDUMP_FIND_CHUNK              0xa
#define RAMDUMP_CONFIGURE_RAMDUMP_AREA  0xd
#define RAMDUMP_SET_WATCHDOG_FUNC       0xe

/**
 * @macro OFFSET
 * @brief Determine offset within a structure of union.
 */
#define OFFSET(type, field) ((size_t) ((char *) (&((type *) 0)->field)))

/**
 * @type DumpPos
 * @brief Dump position.
 */
typedef U32 DumpPos;

/**
 * @type StackSwitchArgs
 * @brief Arguments passed to a function after changing stack
 * in order to avoid problems.
 */
typedef volatile struct StackSwitchArgs
{
   long arg1;
   long arg2;
   long arg3;
   long arg4;
   long arg5;
   long arg6;
} StackSwitchArgs;

/**
 * @type StackSwitchFunc
 * @brief Function to call after switching stack
 */
typedef long (*StackSwitchFunc)(U32 arg1, U32 arg2, U32 arg3,
                                U32 arg4, U32 arg5, U32 arg6);

/**
 * @function ramdump_dump_memory_image2
 *
 * @long Writes a block containing a memory image of specified length,
 * starting from the specified memory address.
 *
 * @input dump Dump position to write to.
 * @input description Text description of memory area.
 * @input start_address Start address of memory area to dump.
 * @input length Size of memory area.
 *
 * @output ret_code Return code is filled in with RAMDUMP_SUCCESS,
 * RAMDUMP_ECURR_DUMP_TOO_LARGE or RAMDUMP_EMEMORY_ACCESS_NOK.
 */   
void ramdump_dump_memory_image2(DumpPos dump, const char *description,
                                OSADDRESS start_address, U32 length,
                                RamdumpStatus *ret_code);

/**
 * @function ramdump_kick_watchdog
 *
 * @brief Calls the currently set ramdump watchdog callback function.
 */
void ramdump_kick_watchdog(void);
   
/**
 * @function ramdump_find_dump_c
 *
 * @include 
 *
 * @brief Locates chunk in dump according to a matcthing specification,
 * e.g. {'FORM', 'PMD ', 'DPID'}
 *
 * @input dump_id Dump id of the dump to scan.
 * @input start_offset Start offset in the dump.
 * @input cs Chunk specification.
 * @last_located_offset Location of the last matching chunk.
 *
 * @return Returns the found offset.
 */
S32 ramdump_find_dump_c(RamdumpDumpId dump_id, U32 start_offset,
                        const IffId *cs, U32 last_located_offset);

/**
 * @function ramdump_switch_stack
 *
 * @brief Switches the stack pointer, calls the function and then
 * switches the stack back.
 *
 * @input new_sp New stack pointer.
 * @input func Function to call after switching stack.
 * @input args Arguments to pass on to function.
 *
 * @return Return code from the called function.
 */
U32 ramdump_switch_stack(U32 new_sp, StackSwitchFunc func,
                         StackSwitchArgs *args);

#ifdef __cplusplus
}
#endif

#endif /* # ifndef _RAMDUMP_INTERNAL_H */
