/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2005-2006 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

#include "stdarg.h"
#include "sysparam.h"
#include "errno.h"
#include "assert.h"
#include "ramlog.h"
#include "string.h"
#include "outfmt.h"
#include "cache.h"
#include "cpu.h"
#include "bios.h"
#include "ramlog_trap.h"
#include "spinlock.h"
#include "ose.h"
#include "stdint.h"
#include "compiler_macros.h"
#include "stdlib.h"
#include "cpu_smp.h"
#ifndef softOSE
#include "gmm.h"
#endif

/* Configurable */
#define RAMLOG_MIN_SIZE 0x00000100
#define RAMLOG_MAX_SIZE 0x01000000
#define RAMLOG_PARAM_NAME "krn/ramlog"
#define RAMLOG_PARAM_NAME_OLD "krn/region/ramlog"

/* Macros */
#define IS_IN_RANGE(x, beg, end) \
 ((beg) <= (end) ? (x) >= (beg) && (x) <= (end) : (x) >= (beg) || (x) <= (end))

/*
 * If this function is present, call it during kernel startup to enable
 * polled write on console for debug purposes.
 */

#ifndef WIN32
WEAK_REF(void ose_polled_write(const char *buf, size_t size));
#endif

struct Ramlog
{
   /* Current virtual write pointer. */
   RamlogVirtAddr newest;
   /* Current virtual oldest-data pointer. */
   RamlogVirtAddr oldest;
   /* Size of the complete ramlog (from system parameters), the actual
    * ramlog buffer is slightly smaller, 3 words of metadata must be
    * subtracted. */
   size_t size;
   /* Token declaration of the actual ramlog buffer (variable size). */
   char base[1];
};

#ifdef OSE_SMP
static spinlock_t smp_lock = SPIN_LOCK_INIT;
#endif

static long log_poll; /* Initialized to False by crt0. */

static Boolean init_begun; /* Initialized to False by crt0. */

static long flush_dcache; /* Initialized to False by crt0. */

static cpuid_t g_num_cpus;
static long g_cpustamp; /* Initialized to False by crt0. */
static long g_timestamp; /* Initialized to False by crt0. */

static unsigned long g_leading_zeros_term;
static OSTICK g_last_tick;
static OSTICK g_usec;

static size_t do_ramlog_write(const char *buf, size_t size);

OSTIME zzkrn_system_tick(void);
OSTICK zzkrn_get_systime(OSTICK *microsecs);
OSTICK zzkrn_get_ticks(void);
OSBOOLEAN ose_started(void);
    
void ose_ramlog_restart(void);

void ose_ramlog_restart(void)
{
   init_begun = False;
}

/*
 *===========================================================================
 *          get_ramlog_ptr
 *===========================================================================
 * Description:	Initializes ramlog.  If control data is invalid, it resets it.
 *              Saves startup newest so ramlog entries can be retrieved "since
 *		last reset".
 *
 * Returns:  pointer to the ramlog, NULL if initialisation failed.
 */

static struct Ramlog *
get_ramlog_ptr(void)
{
   static struct Ramlog *ramlog; /* Initialized to NULL by crt0. */
   U32 mask;

   LOCK_SAVE(mask);
   if (ramlog == NULL)
   {
      if (!init_begun)
      {
	 /* Have a non-static copy of the pointer to the ramlog while we
	  * initilise. */
	 struct Ramlog *rl;
	 long size;

#ifdef LIBWHAT
	 LIBWHAT
#endif
	 init_begun = True;
	 LOCK_RESTORE(mask);
	 if (((sysparam_readattr_long(RAMLOG_PARAM_NAME, "base",
				     (long *)&rl) == EOK)
	     && (sysparam_readattr_long(RAMLOG_PARAM_NAME, "size",
					&size)  == EOK)
	     && (size >= RAMLOG_MIN_SIZE)
              && (size <= RAMLOG_MAX_SIZE))
	     || /* Backwards compatibility. */
	     ((sysparam_readattr_long(RAMLOG_PARAM_NAME_OLD, "base",
				     (long *)&rl) == EOK)
	     && (sysparam_readattr_long(RAMLOG_PARAM_NAME_OLD, "size",
					&size)  == EOK)
	     && (size >= RAMLOG_MIN_SIZE)
              && (size <= RAMLOG_MAX_SIZE))
	    )
	 {
#define ONLY_ON_SMP 2
	    static const struct enumeration enumeration[6] =
	    {
	       { "true", True},
	       { "yes", True},
	       { "false", False},
	       { "no", False},
	       { "smp", ONLY_ON_SMP},
	       { NULL, 0}
	    };
	    size_t actual_size = size - offsetof(struct Ramlog, base);

	    /* Determine whether ramlog should flush D-cache or not. */
	    flush_dcache = True;
	    (void) sysparam_readattr_enum(RAMLOG_PARAM_NAME, "flush",
					  enumeration, &flush_dcache);

	    /* Determine whether ramlog should call out to log or not. */
	    (void) sysparam_readattr_enum(RAMLOG_PARAM_NAME, "log",
					  enumeration, &log_poll);

	    /* Determine whether ramlog should add timestamps or not */
            g_timestamp = False;
	    (void) sysparam_readattr_enum(RAMLOG_PARAM_NAME, "timestamp",
					  enumeration, &g_timestamp);
            
	    /* Determine whether ramlog should add cpu_id or not */
	    g_num_cpus = zzkrn_num_cpus();
            g_cpustamp = False;
	    (void) sysparam_readattr_enum(RAMLOG_PARAM_NAME, "cpustamp",
					  enumeration, &g_cpustamp);
	    if (g_cpustamp == ONLY_ON_SMP && g_num_cpus < 2)
	       g_cpustamp = False;
#undef ONLY_ON_SMP

	    /* Check if there's an old ramlog to continue with. */
	    if (rl->size != actual_size ||
		rl->newest - rl->oldest > actual_size)
	    {
	       /* No valid metadata to continue with, reset ramlog. */
	       rl->size = actual_size;
	       rl->newest = 0;
	       rl->oldest = 0;
	       if (flush_dcache && cache_initialized())
	       {
		  cache_flush_d((OSADDRESS)rl, sizeof *rl);
	       }
	    }
	    /* Else ramlog metadata was OK from previous session */

	    /* By initialising the static variable 'ramlog', we're
	     * enabling the ramlog functionality.
	     */
	    LOCK_SAVE(mask);
	    ramlog = rl;
	    (void) do_ramlog_write(RAMLOG_STR_SESSION_START,
				sizeof(RAMLOG_STR_SESSION_START) - 1);
	    LOCK_RESTORE(mask);
	 } /* if (<system parameters ok>) */
	 /* Else the configuration was missing or wrong, leave 'init_begun'
	  * as 'True' to prevent from further attempts to initialise.
	  */
      }
      else /* else initialisation either on-going or failed. */
      {
	 LOCK_RESTORE(mask);
      }
   } /* if (ramlog == NULL) */
   else
   {
      LOCK_RESTORE(mask);
   }
   return ramlog;
}


/*
 *===========================================================================
 *          write_and_flush()
 *===========================================================================
 * Description:	 Copy userdata to ramlog and flush it from D-cache.
 *
 * Returns:	nothing.
 *
 */
static void
write_and_flush(char *dest, const char *src, size_t size)
{
   memcpy(dest, src, size);
   if (flush_dcache && cache_initialized())
   {
      cache_flush_d((OSADDRESS)dest, size);
   }
}


/*
 *===========================================================================
 *          do_ramlog_write()
 *===========================================================================
 * Description:	 Writes size bytes from buf to ramlog.
 *               Must always be called with taken spin lock 
 *
 * Returns:	Number of bytes written.
 *
 */
static size_t
do_ramlog_write(const char *buf, size_t size)
{
   struct Ramlog *ramlog = get_ramlog_ptr();

#ifndef WIN32   
   if(ose_polled_write != NULL && log_poll)
   {
      Boolean mmu_setup_phase = False;
      
#ifndef softOSE
      mmu_setup_phase = (gmm_state.state == arch_init_mmu) ? True : False;
#endif         

      /* Avoid polled writing when in the sensitive
         state of setting up MMU. OSECR-8575 */
      if (!mmu_setup_phase)
      {
         ose_polled_write(buf,size);
      }
   }
#endif
   
   if (ramlog != NULL)
   {
      size_t offs;
      size_t size_before_wrap;
      size_t ramlog_actual_size = ramlog->size - offsetof(struct Ramlog, base);
      
      /* Compute offset to start writing to. */
      offs = ramlog->newest % ramlog_actual_size;

      /* Clip write size to ramlog buffer size. */
      if (size > ramlog_actual_size)
      {
	 size = ramlog_actual_size;
      }

      /* Update newest pointer. */
      ramlog->newest += size;

      /* Update oldest pointer if needed. */
      if ((ramlog->newest - ramlog->oldest) > ramlog_actual_size)
      {
	 /* If here, the ramlog buffer is full, so each update of the newest
	  * pointer must be followed by an equivalent update of the oldest
	  * pointer.
	  */
	 ramlog->oldest = ramlog->newest - ramlog_actual_size;
      }

      size_before_wrap = ramlog_actual_size - offs;

      /* Check if targeted range span over the wrap. */
      if (size > size_before_wrap)
      {
	 /* Yes, we have to copy the data in two batches over the wrap. */
	 size_t size_after_wrap = size - size_before_wrap;
	 write_and_flush(ramlog->base+offs, buf, size_before_wrap);
	 write_and_flush(ramlog->base, buf+size_before_wrap, size_after_wrap);
      }
      else
      {
	 /* Write to the log in one batch. */
	 write_and_flush(&ramlog->base[offs], buf, size);
      }
      
      /* FIXME: is it nessesary to flush cache with taken spin lock ? */
      /* Finally flush ramlog metadata from D-cache. */
      if (flush_dcache && cache_initialized())
      {
	 cache_flush_d((OSADDRESS)ramlog, sizeof *ramlog);
      }
   } /* if (ramlog != NULL) */
   else
   {
#ifdef softOSE /* since it may occur in soft, put text on console. */
      extern void write(int, const char*, int);
      write(1,buf,size);
#endif
      /* Ramlog not configured/enabled, ignore call. */
      size = 0;
   }
   return size;
}

/*
 *===========================================================================
 *          u32todecstr()
 *===========================================================================
 * Description:	Converts an uint32_t value to an ascii string in decimal
 *              format. The conversion starts with the least significant
 *              digit and put it at the end of the string.
 *
 * Input      : ptr, shall point at the highest address of the string.
 *
 * Input      : val, the value to convert.
 *
 * Returns    : A pointer that points at the most significant digit in the 
 *              string after conversion
 *  
 * Restriction:  The string must be large enough to hold all digits.
 *               unsigned long must be of the same size as uint32_t.
 *               
 *
 */
static char *
u32todecstr(char * ptr, uint32_t val)
{
  /* ldiv_t has the members quot and rem */
  ldiv_t res;
  
  /* must remember the start pointer in case of division by ten  */
  char *sp = ptr;
  
  /* rem initiates to 10 to make it possible to know if the */
  /* division by ten is done or not */
  unsigned long rem = 10;

  /* since ldiv only support signed long and not unsigned */
  /* the value must be divided if the higest bit is set */
  if(val & 0x80000000)
  {
    res.quot = val / 10;
    /* must remember the reminder so it can be added when the */
    /* conversion is done, after this modulo rem can have values 0-9 */
    rem = val % 10;
    /* make space for the last digit */
    ptr--;
  }
  else
  {
    res.quot = val;
  }
  
  /* By having the comparison last in the loop there is no need to have */
  /* extra code in the case that val is zero */
  do
  {
    res = ldiv(res.quot ,10);
    /* write the reminder to the decimal string */
    *ptr = res.rem + '0';
    ptr--;
  }
  while(res.quot > 0);
  
  /* let the pointer point at the first digit */
  ptr++;
  
  /* add the reminder to the first digit */
  if(rem != 10)
  {
    *sp= rem +'0';
  }
  return(ptr);
}   


/*
 *===========================================================================
 *          format_timestamp()
 *===========================================================================
 * Description:	 Formats the timestamp string according to tick and usec to
 *               the ramlog format 123.012:
 *               
 * Input      : ep, end pointer shall point at the highest address of the string.
 *
 * Returns    : A pointer that points at the first digit to be written.
 *
 */
static char *
format_timestamp(char * ep)
{
  char * cp;
          
  /* convert usec */
  cp = u32todecstr(ep, g_usec);
  
  /* insert leading zeros */
  while( (ep - g_leading_zeros_term) < cp)
  {
    cp--;
    *cp = '0';
  }
    
  /* insert the dot */
  cp--;
  *cp = '.';
  cp--;
  
  /* convert tick */
  cp = u32todecstr(cp, g_last_tick);
  return(cp);
}
  
size_t
ramlog_handle_write(const char *buf, size_t size)
{
  
  /* The timestamp and the message must be written to the ramlog with no */
  /* interruption, otherwise can a later timestamp be written before an  */
  /* earlier in the ramlog. */
  
  /* global variables is used to minimize the stack consumption, this is */
  /* possible since no interruption can occur */
 
  /* The formattet timestamp shall look like this in the ramlog */
  /* 123.012: */
  /* 123 is ticks with no leading zeros */
  /* 012 is micro seconds with leading zeros. */
  /* Note that a tick can be one of 1,2,4,8,16,32,64 milli seconds  */
    
  static const char g_zero_tick[] = "0.000:";      
  static const char timestamp_format[] = "Timestamp format tick.usec: (1 tick = ";
  static const char micro_seconds[] = " micro seconds)\n";
  static char g_last_char = '\n';
  static OSTIME g_usec_per_tick; 
  
  /* let crt0 initiate the colon and terminating null character */
  static char g_timestamp_buf[18] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,':','\0'};

  U32 mask;
  char * print_p; /* printing pointer */

  static int kernel_started = 0;
    
  /* Lock to ensure consistency between the timestamp and the ramlog message.*/
  if(likely(kernel_started))
  {
     spin_lock_irqsave(&smp_lock, mask);
  }
  else
  {
     /* We need this call to init g_cpustamp etc. which are used below. */
     (void)get_ramlog_ptr();
     /*
     ** Keep sampling zzkrn_num_cpus() during system start since it's
     ** return value can change. This can happen if someone is using
     ** the ramlog before the board setup has determined the correct
     ** return value of zzkrn_num_cpus().
     */
     g_num_cpus = zzkrn_num_cpus();
     LOCK_SAVE(mask);
  }

  /* write the timestamp only if the previous write */
  /* to the ramlog was terminated with line feed */ 
  if(likely(g_last_char == '\n'))
  {
     if(unlikely(g_cpustamp))
     {
#ifdef OSE_SMP
	static const char cpu_str[] =
	   "[0] [1] [2] [3] [4] "
	   "[5] [6] [7] [8] [9] ";

	if(g_num_cpus <= 10)
	{
	   (void)do_ramlog_write(&cpu_str[zzkrn_cpu_id() * 4], 4);
	}
	else
	{
	   cpuid_t cpu_id = zzkrn_cpu_id();
	   char str[5];

	   str[0] = '[';
	   str[1] = '0' + cpu_id / 10;
	   str[2] = '0' + cpu_id % 10;
	   str[3] = ']';
	   str[4] = ' ';
	   (void)do_ramlog_write(&str[0], sizeof(str));
	}
#else
	(void)do_ramlog_write("[0] ", 4);
#endif
     }
     
     /* Do the timestamp only if it is enabled by the system parameter */  
     if(unlikely(g_timestamp))
     {
        /* the zzkrn_get_systime() and zzkrn_system_tick() does not work as long as
         * zzkrn_get_ticks() returns zero FIXME: is this a reliable solution */
        if(likely(g_last_tick != 0))
        {
	   g_last_tick = zzkrn_get_systime(&g_usec);
	   print_p = format_timestamp(&g_timestamp_buf[sizeof(g_timestamp_buf)-3]);
	   /* write the timestamp to ramlog */
	   (void)do_ramlog_write(print_p, strlen(print_p));
        }
        else /* g_last_tick = 0 */
        {  
	   g_last_tick = zzkrn_get_ticks();

	   if(g_last_tick == 0)
	   {
	      (void)do_ramlog_write(g_zero_tick, strlen(g_zero_tick));
	   }
	   else
	   {                 
	      /* last_tick is not zero -> zzkrn_system_tick() and zzkrn_get_systime
	       * works  */
	      g_last_tick = zzkrn_get_systime(&g_usec);
	      g_usec_per_tick = zzkrn_system_tick();
	      g_leading_zeros_term = 2;          /* 1000 usec/tick */
	      if(g_usec_per_tick > 1000)  /* 2000 - 8000 usec/tick */
		 g_leading_zeros_term = 3;
	      if(g_usec_per_tick > 8000)  /* 16000 - 64000 usec/tick */
		 g_leading_zeros_term = 4;

	      /* write the timestamp message to ramlog */
	      print_p = format_timestamp(&g_timestamp_buf[sizeof(g_timestamp_buf)-3]);
	      /* write the timestamp to ramlog */
	      (void)do_ramlog_write(print_p, strlen(print_p));
	      /* write timestamp message */
	      (void)do_ramlog_write(timestamp_format, strlen(timestamp_format));
            
	      /* temporarily replace the colon with \0 since the colon is */
	      /* not wanted in this particular print */ 
	      g_timestamp_buf[sizeof(g_timestamp_buf)-2] = '\0';
            
	      /* write the number of usec_per_tick to ramlog */ 
	      print_p = u32todecstr(&g_timestamp_buf[sizeof(g_timestamp_buf)-3],
				    g_usec_per_tick);
	      (void)do_ramlog_write(print_p, strlen(print_p));
	    
	      /* put back the colon */ 
	      g_timestamp_buf[sizeof(g_timestamp_buf)-2] = ':';
	    
	      (void)do_ramlog_write(micro_seconds, strlen(micro_seconds));
	    
	      /* Make the timestamp for the original message */
	      print_p = format_timestamp(&g_timestamp_buf[sizeof(g_timestamp_buf)-3]);
	      (void)do_ramlog_write(print_p, strlen(print_p));
	   }
        }  
     } /* g_timestamp */
  } /* g_last_char == '\n' */
    
  /* write the original message to the ramlog */
  size = do_ramlog_write(buf, size);

  /* store the last character in the buffer */
  g_last_char = *(buf + size-1);
    
  /* All writes to the ramlog is done so it is safe to unlock. */
  if(likely(kernel_started))
  {
     spin_unlock_irqrestore(&smp_lock, mask);
  }
  else
  {
     kernel_started = ose_started();
     LOCK_RESTORE(mask);
  }
  return(size);
}
/*
 *===========================================================================
 *
 *===========================================================================
 * Description:
 *
 * Returns:
 *
 */
size_t
ramlog_handle_read(char *buf, size_t size, RamlogVirtAddr addr,
	    RamlogVirtAddr *actual_addr)
{
   struct Ramlog *ramlog = get_ramlog_ptr();
   if (ramlog != NULL)
   {
      size_t offs;
      size_t size_before_wrap;
      size_t ramlog_actual_size = ramlog->size - offsetof(struct Ramlog, base);
      U32 mask;

      /* Sample values of newest and oldest atomically. */
      LOCK_SAVE(mask);

      /* clip addr to be within Oldest and Newest. */
      if (!IS_IN_RANGE(addr, ramlog->oldest, ramlog->newest))
      {
	 addr = ramlog->oldest;
      }
      /* clip size to physical size */
      if (size > ramlog_actual_size)
      {
	 size = ramlog_actual_size;
      }
      /* clip size to actual size */
      if (size > (ramlog->newest - addr))
      {
	 size = ramlog->newest - addr;
      }

      /* Calculate start offset within the ramlog buffer. */
      offs = addr % ramlog_actual_size;

      size_before_wrap = ramlog_actual_size - offs;
      /* See if requested log contents span over a wrap boundary. */
      if (size > size_before_wrap)
      {
	 /* Log spans over wrap boundary, so we have to copy the data in two
	  * steps.*/
	 size_t size_after_wrap = size - size_before_wrap;
	 memcpy(buf,                    &ramlog->base[offs], size_before_wrap);
	 memcpy(buf + size_before_wrap, &ramlog->base[0],    size_after_wrap);
      }
      else
      {
	 /* Copy the log to the buffer. */
	 memcpy(buf, &ramlog->base[offs], size);
      }

      LOCK_RESTORE(mask);

      if (actual_addr != NULL)
      {
         *actual_addr = addr;
      }      
   } /* if (ramlog != NULL) */
   else
   {
      /* Ignore call. */
      size = 0;
   }
   return size;
}


RamlogVirtAddr
ramlog_handle_addr_of_newest(void)
{
   RamlogVirtAddr newest;
   struct Ramlog *ramlog = get_ramlog_ptr();
   if (ramlog != NULL)
   {
      newest = ramlog->newest;
   }
   else
   {
      newest = 0;
   }
   return newest;
}

RamlogVirtAddr
ramlog_handle_addr_of_oldest(void)
{
   RamlogVirtAddr oldest;
   struct Ramlog *ramlog = get_ramlog_ptr();
   if (ramlog != NULL)
   {
      oldest = ramlog->oldest;
   }
   else
   {
      oldest = 0;
   }
   return oldest;
}

RamlogVirtAddr
ramlog_handle_set_oldest(RamlogVirtAddr new_oldest)
{
   struct Ramlog *ramlog = get_ramlog_ptr();
   if (ramlog != NULL)
   {
      U32 mask;

      spin_lock_irqsave(&smp_lock, mask);
      /* Check whether the new oldest pointer is within the current range. */
      if (IS_IN_RANGE(new_oldest, ramlog->oldest, ramlog->newest))
      {
	 /* Yes, update oldest pointer. */
	 ramlog->oldest = new_oldest;
      }
      else
      {
	 /* No, new_oldest is invalid or already obsolete.
	    Return the actual oldest pointer. */
	 new_oldest = ramlog->oldest;
      }
      spin_unlock_irqrestore(&smp_lock, mask);
   }
   /* Else ramlog not available, return the specified oldest pointer. */
   return new_oldest;
}

RamlogVirtAddr
ramlog_handle_find_next(RamlogVirtAddr req_start, const char *pattern)
{
   size_t patlen = strlen(pattern);
   struct Ramlog *ramlog = get_ramlog_ptr();

   if ((ramlog != NULL)
       && (patlen > 0)
       && (patlen <= ramlog->size - offsetof(struct Ramlog, base))
       && (ramlog->newest != ramlog->oldest)
       && (req_start != ramlog->newest))
   {
      size_t ramlog_actual_size = ramlog->size - offsetof(struct Ramlog, base);
      RamlogVirtAddr actual_start = (IS_IN_RANGE(req_start,
						 ramlog->oldest,
						 ramlog->newest)
				     ? req_start
				     : ramlog->oldest);
      unsigned searched = 1;
      size_t curr_offs = (actual_start + 1) % ramlog_actual_size;
      size_t offs_of_newest = ramlog->newest % ramlog_actual_size;

      /* Check if we have to search beyond 'newest' up to the wrap boundary.
       * If curr_offs == offs_of_newest, we have the whole ramlog circular
       * buffer to search through.
       */
      if (curr_offs >= offs_of_newest)
      {
	 /* 1st: search until RAMLOG buffer limit - patlen + 1. */
	 for ( ; (curr_offs + patlen - 1) < ramlog_actual_size;
	       curr_offs++, searched++)
	 {
	    const char *tg = &ramlog->base[curr_offs];
	    if (strncmp(tg, pattern, patlen) == 0)
	    {
	       /* Successful EXIT */
	       assert(searched < ramlog_actual_size);
	       return actual_start + (RamlogVirtAddr)searched;
	    }
	 }

	 /* 2nd: see of pattern is split over wrap boundary. */
	 for ( ; curr_offs < ramlog_actual_size; curr_offs++, searched++)
	 {
	    size_t chars_before_wrap = ramlog_actual_size - curr_offs;
	    /* See if first part of pattern is at end of circular ramlog
	     * buffer and second part at beginning.
	     */
	    const char *first_part = &ramlog->base[curr_offs];
	    const char *second_part= &ramlog->base[0];
	    if ((strncmp(first_part, pattern, chars_before_wrap) == 0)
		&& (strncmp(second_part, pattern + chars_before_wrap,
			    patlen - chars_before_wrap) == 0))
	    {
	       /* Successful EXIT */
	       assert(searched < ramlog_actual_size);
	       return actual_start + (RamlogVirtAddr)searched;
	    }
	 }

	 /* Now curr_offs == ramlog_actual_size, and the search has to continue
	  * from beginning of circular buffer up to offset of newest, so
	  * set curr_offs = 0
	  */
	 curr_offs = 0;
      }

      /* From here, offs_of_newest is ahead of curr_offs in the ramlog
       * buffer, see if we find pattern!
       */
      for ( ; (curr_offs + patlen) < offs_of_newest; curr_offs++, searched++)
      {
	 const char *tg = &ramlog->base[curr_offs];
	 if (strncmp(tg, pattern, patlen) == 0)
	 {
	    /* Successful EXIT. */
	    assert(searched < ramlog_actual_size);
	    return actual_start + (RamlogVirtAddr)searched;
	 }
      }

      /* If here, curr_offs + patlen >= offs_of_newest, so 'pattern' would
       * not have room enough to fit. So search failed, return "req_start"
       * for "not found".
       */
      assert(searched < ramlog_actual_size);

   } /* if ((patlen > 0) && (patlen <= ramlog_actual_size)
	&& (ramlog->newest != ramlog->oldest)) */
   return req_start; /* Unsuccessful EXIT. */
}


RamlogVirtAddr
ramlog_handle_find_previous(RamlogVirtAddr req_start, const char *pattern)
{
   size_t patlen = strlen(pattern);
   struct Ramlog *ramlog = get_ramlog_ptr();
   size_t ramlog_actual_size = ramlog ?
			       ramlog->size - offsetof(struct Ramlog, base) :
			       0 ;

   if ((ramlog != NULL)
       && (patlen > 0)
       && (patlen <= ramlog_actual_size)
       && (ramlog->newest != ramlog->oldest)
       && (req_start != ramlog->oldest))
   {
      RamlogVirtAddr actual_start = (IS_IN_RANGE(req_start,
						 ramlog->oldest,
						 ramlog->newest)
				     ? req_start
				     : ramlog->newest);
      int searched = 1;
      int curr_offs = (int)((actual_start - 1) % ramlog_actual_size);
      int offs_of_oldest = (int)(ramlog->oldest % ramlog_actual_size);

      /* Check if we have to search beyond 'oldest' down to the wrap
       * boundary.
       */
      if (curr_offs < offs_of_oldest)
      {
	 /* 1st: search until start of RAMLOG buffer. */
	 for ( ; curr_offs >= 0; curr_offs--, searched++)
	 {
	    const char *tg = &ramlog->base[curr_offs];
	    if (strncmp(tg, pattern, patlen) == 0)
	    {
	       /* Successful EXIT */
	       return actual_start - (RamlogVirtAddr)searched;
	    }
	 }

	 /* Now curr_offs == -1. Make it point to last char of buffer. */
	 curr_offs = (int)ramlog_actual_size - 1;

	 /* 2nd: see if pattern is split over wrap boundary. */
	 for ( ; ((size_t)curr_offs + patlen) > ramlog_actual_size;
	       curr_offs--, searched++)
	 {
	    size_t chars_before_wrap = ramlog_actual_size - (unsigned)curr_offs;
	    const char *first_part = &ramlog->base[curr_offs];
	    const char *second_part = &ramlog->base[0];
	    /* See if first part of pattern is at end of circular ramlog
	     * buffer and second part at beginning. */
	    if ((strncmp(first_part, pattern, chars_before_wrap) == 0)
		&& (strncmp(second_part, pattern + chars_before_wrap,
			    patlen - chars_before_wrap) == 0))
	    {
	       /* Successful EXIT */
	       return actual_start - (RamlogVirtAddr)searched;
	    }
	 }

	 /* Now curr_offs == (ramlog_actual_size - 1 - patlen), and the search
	  * has to continue down to offset of oldest.
	  */
      }

      /* From here, offs_of_oldest is below of curr_offs in the ramlog
       * buffer. Continue seeking backwards for pattern.
       */
      for (;;)
      {
	 const char *tg = &ramlog->base[curr_offs];
	 if (strncmp(tg, pattern, patlen) == 0)
	 {
	    /* Successful EXIT. */
	    return actual_start - (RamlogVirtAddr)searched;
	 }
	 else if (curr_offs <= offs_of_oldest)
	 {
	    break; /* done, found nothing */
	 }
	 else
	 {
	    /* Loop */
	    curr_offs--;
	    searched++;
	 }
      }

      /* If here, the search failed; return "req_start" for "not found".  */

   } /* if ((patlen > 0) && (ramlog->newest != ramlog->oldest)) */
   /* Return "req_start" for unsuccess. */
   return req_start; /* Unsuccessful EXIT. */
}



static long
ramlog_traphandler(long fc, long arg1, long arg2, long arg3, long arg4,
                   long arg5, long arg6)
{
   (void) arg5;
   (void) arg6;
   
    switch (fc)
    {
	case FC_ADDR_OF_OLDEST :
	    return ramlog_handle_addr_of_oldest();
	case FC_ADDR_OF_NEWEST :
	    return ramlog_handle_addr_of_newest();
	case FC_FIND_NEXT :
	    return ramlog_handle_find_next(arg1, (const char *)arg2);
	case FC_FIND_PREV :
	    return ramlog_handle_find_previous(arg1, (const char *)arg2);
	case FC_READ :
	    return ramlog_handle_read((char *)arg1, arg2, arg3,
                                      (RamlogVirtAddr *)arg4);
	case FC_WRITE :
	    return ramlog_handle_write((const char *)arg1, arg2);
	case FC_SET_OLDEST :
	    return ramlog_handle_set_oldest(arg1);
	default :
	    return -1;
    }
}

void
ramlog_bios_install(void)
{
      
  (void)biosInstall(BIOSNAME_RAMLOG,
                     ramlog_traphandler,
                     BIOS_USERMODE_ALLOWED);
}

size_t kprintf(const char *fmt, ...);

size_t
kprintf(const char *fmt, ...)
{
/*
 * use static variables, since we serialize this...
 */
   static int fmt_len;
   static char buffer[RAMLOG_PRINTF_MAXSIZE];
   static va_list ap;
#ifdef OSE_SMP
   static spinlock_t klock = SPIN_LOCK_INIT;
#endif
   static int kernel_started = 0;

   U32 mask;
   size_t ret_len;
   /* Lock to ensure serialization of buffer access */
   if(likely(kernel_started))
   {
      spin_lock_irqsave(&klock, mask);
   }
   else
   {
      LOCK_SAVE(mask);
   }


   /* Now we have exclusive use of the buffer */
   va_start(ap, fmt);
   fmt_len = vsnprintf(buffer, sizeof(buffer), fmt, ap);
   va_end(ap);
   if (fmt_len > 0)
   {
      ret_len = ramlog_handle_write(buffer, (size_t)fmt_len);
   }
   else
   {
      if (fmt_len != 0)
      {
         /* Try to print out whatever is possible.
          * vsnprintf() sees so that 'buffer' contains a printable and
          * null-terminated string.
          */
         ret_len = strlen(buffer);
         if (ret_len > 0)
         {
            ret_len = ramlog_handle_write(buffer, ret_len);
         }
      }
      else
      {
         /* Else fmt_len == 0, do nothing more. */
         ret_len = 0;
      }
   }
   
   if(likely(kernel_started))
   {
      spin_unlock_irqrestore(&klock, mask);
   }
   else
   {
      kernel_started = ose_started();
      LOCK_RESTORE(mask);
   }
   return ret_len;
}
