/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2009 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */


/**
 *  @file trace.c
 *  @brief This is the implementation of the system trace api.
 */
#include "stdlib.h"
#include "string.h"
#include "stdint.h"
#include "ramlog.h"
#include "ramtrace.h"
#include "assert.h"
#include "ose.h" /* FIXME: Needed by inline.h (indirectly included) */
#include "cpu_access.h"
#include "pose.con"
#include "pose.h"
#include "spinlock.h"
#include "compiler.h"

/* Private flags for trace entry (public flags are in ramtrace.h) */
#define T_FREE     0x00
#define T_BUSY     0x01  /* Entry is beeing updated by client. */
#define T_LOCKED   0x02  /* Entry explicitly locked */
#define T_VALID    0x08  /* Trace entry may contain valid data. */
/*      T_FORCE    0x10     defined in ramtrace.h */
/*      T_NOTIME   0x20     defined in ramtrace.h */
/*      T_HIDDEN   0x40     defined in ramtrace.h */
/*      T_REFERRED 0x80     defined in ramtrace.h */
#define T_ALLFLAGS (T_BUSY | T_LOCKED | T_VALID | T_FORCE | T_NOTIME | T_HIDDEN | T_REFERRED)

struct trace_entry
{
   /* Set if valid. zero if invalid. */
   uint8_t   flags;
   /* Trace sequence number. */
   uint8_t   seq;
   /* Actual entry size (in trace_entry blocks) */
   EntrySize sz;
   /* No of bytes caller wants to store */
   EntrySize size;
   /* Counter for number of times a locked entry has bean passed by tail */
   uint16_t  ow_count;
   /* Timestamp */
   uint32_t  time_hi;
   uint32_t  time_lo;
};

/* Return pointer to trace_entry header given trace data pointer */
#define te_header(p) ((struct trace_entry *)(p) - 1)

/* Return pointer to trace data given the trace entry header */
#define te_data(t)   ((struct trace_entry *)(t) + 1)

/* This is_readable() macro returns true if the specified trace entry is
 * OK to pass to the client in a ose_trace_read().
 */
#define is_readable(p,e) \
   ((e)->seq == (p)->seq && ((e)->flags & (T_VALID | T_BUSY | T_HIDDEN)) == T_VALID)

/* This will_be_readable() macro returns true if the specified trace entry is
 * OK to pass to the client in a ose_trace_read() when is_readable() is true.
 */
#define will_be_readable(p,e) \
   ((e)->seq == (p)->seq && ((e)->flags & (T_VALID | T_HIDDEN)) == T_VALID)

/* The size encoded in the 'sz' field in the total size of the entry (including
 * the trace entry header) in units of trace entry headers
 */
#define trace_entry_sz(s) ((((((s) + 15) & (~0 << 4)) / sizeof(struct trace_entry))) + 1)

/* Magic value for trace_head */
#define OSE_TRACE_MAGIC_HEAD 0xf01f2ac5

/* This structure defines the trace buffer parameters and is located
 * at the lowest address of the trace buffer.
 */
struct trace_head
{
   /* [0,31] Frequently accessed variables */
   uint32_t magic_head;         /* Magic number (trace format id). */
   spinlock_t lock;             /* Spinlock for this trace. */
   GetTimeFuncPointer time_fn;  /* Timebase function */
   uint8_t enabled;             /* Set if tracing is allowed. */
   uint8_t seq;                 /* Trace sequence number. */
   uint8_t stamp_valid;         /* Nonzero when time_stamp is present.  */
   uint8_t pad;                 /* Just for alignment */
   struct trace_entry *head;    /* Points at next free byte. */
   struct trace_entry *tail;    /* Points at start of oldest block. */
   struct trace_entry *beg;     /* First byte of free space. */
   struct trace_entry *end;     /* One byte beyond free space. */
   /* [32,63] Less frequently accessed variables */
   uint32_t time_freq;          /* Frequency of time counter. */
   uint32_t unused1;            /* Unused */
   uint32_t unused2;            /* Unused */
   uint32_t unused3;            /* Unused */
   int32_t checksum;             /* For ptrs, time_fn, freq, and name fields. */
   char    magic[8];             /* Contains the magic ID "IsAtRaCe". */
   char    trace_name[TRACE_MAX_NAME_LEN + 4]; /* Ascii trace name. */
};

static int32_t
compute_checksum(const volatile struct trace_head *header)
{
   int32_t sum;
   int i;

   sum = ((int)header->beg +
          (int)header->end +
          (int)header->time_fn +
          (int)header->time_freq);

   for (i=0; i<=TRACE_MAX_NAME_LEN; i++)
   {
      sum += header->trace_name[i] + i;
   }

   return sum;
}

static void
init_magic(volatile struct trace_head *header)
{
   header->checksum = compute_checksum(header);
   header->magic[5] = 'a' - 1;
   header->magic[1] = 's';
   header->magic[3] = 't';
   header->magic[7] = 'e';
   header->magic[0] = 'I';
   header->magic[2] = 'A';
   header->magic[4] = 'R';
   header->magic[6] = 'C';
   header->magic_head = OSE_TRACE_MAGIC_HEAD;
   header->magic[5] += 1;
}

static int
check_magic(const volatile struct trace_head *header)
{
   int magic_ok;

   magic_ok =
      compute_checksum(header) == header->checksum &&
      header->magic[5] == 'a' &&
      header->magic[1] == 's' &&
      header->magic[3] == 't' &&
      header->magic[7] == 'e' &&
      header->magic[0] == 'I' &&
      header->magic[2] == 'A' &&
      header->magic[4] == 'R' &&
      header->magic[6] == 'C' &&
      header->magic_head == OSE_TRACE_MAGIC_HEAD;
   return magic_ok;
}

static int
check_entry(const struct trace_head *p, struct trace_entry *t)
{
   /* Check that the entry ends within the trace buffer. */
   if (t + t->sz > p->end)
   {
      return 101;
   }

   /* Check that the size field is non-zero. */
   if (t->sz == 0)
   {
      return 102;
   }

   /* Check that the two size fields match if entry is free. */
   if (t->flags & T_VALID)
   {
      if (t->sz != trace_entry_sz(t->size))
      {
         return 103;
      }
   }

   /* Check value of flags. */
   if ((t->flags & ~T_ALLFLAGS) != 0)
   {
      return 104;
   }

   /* Check that the entry does not overwrite the header. */
   if (t < p->head && t + t->sz > p->head)
   {
      return 105;
   }

   /* Consider any taken trace entry invalid. */
   if (t->flags & T_BUSY)
   {
      /* Keep the T_REFERRED flag and reset overwrite count*/
      if (t->flags & T_REFERRED)
      {
         t->flags = T_FREE | T_REFERRED;
         t->ow_count = 0;
      }
      else
      {
         t->flags = T_FREE;
      }
   }
   else
   {
      /* Make sure the element is not locked. */
      t->flags &= ~T_LOCKED;
   }
   return 0;
}

static int
check_trace(const struct trace_head *p, uint32_t size, const char *name,
            uint32_t time_freq)
{
   int result = 0;
   struct trace_entry *t;

   /* Size in units of trace entries */
   size -= sizeof(struct trace_head);
   size /= sizeof(struct trace_entry);

   /* Check that magic ID is OK. */
   if (!check_magic(p))
   {
      return 1;
   }

   /* Check that we still use the same time base. */
   if(time_freq != p->time_freq)
   {
      return 2;
   }

   /* Check that base limit is correctly selected. */
   if (p->beg != (const struct trace_entry *) (p + 1))
   {
      return 3;
   }

   /* Check that end limit is correctly selected. */
   if (p->end != p->beg + size)
   {
      return 4;
   }

   /* Check that head and tail are within limits. */
   if (p->head < p->beg || p->head > p->end)
   {
      return 5;
   }

   if (p->tail < p->beg || p->tail > p->end)
   {
      return 6;
   }

   /* Check that head and tail are aligned. */
   if (((char *)p->head - (char *)p->beg) % sizeof(struct trace_entry))
   {
      return 7;
   }

   if (((char *)p->tail - (char *)p->beg) % sizeof(struct trace_entry))
   {
      return 8;
   }

   /* Check that time stamp flag is the same. */
   if (p->stamp_valid != 1)
   {
      return 9;
   }

   /* Check that trace name is the same. */
   if (strcmp(name, p->trace_name) != 0)
   {
      return 10;
   }

   /* Check all entries from start of buffer to head. */
   for (t = p->beg; t < p->head && result == 0; t += t->sz)
   {
      result = check_entry(p, t);
   }

   /* Check all entries from tail to end of buffer. */
   if (result == 0 && p->tail != p->beg)
   {
      for (t = p->tail; t < p->end && result == 0; t += t->sz)
      {
         result = check_entry(p, t);
      }
   }

   return result;
}

static struct trace_entry *
next_entry(const volatile struct trace_head *trace, struct trace_entry *entry)
{
   entry += entry->sz;
   if (entry == trace->end)
   {
      if (trace->beg == trace->head)
      {
         return NULL;
      }
      else
      {
         return trace->beg;
      }
   }
   else if (entry == trace->head)
   {
      return NULL;
   }
   return entry;
}

/*
 *===========================================================================
 *                       RAMTRACE API FUNCTIONS
 *===========================================================================
 */

int
ose_trace_create(void *buffer, uint32_t size, const char *name,
                 GetTimeFuncPointer time_fn, uint32_t time_freq)
{
   struct trace_head *p = buffer;

   if (size < TRACE_MIN_SIZE)
   {
      /* Requested trace is too small. */
      return -1;
   }

   /* Make size into units of trace entries */
   size -= sizeof(struct trace_head);
   size /= sizeof(struct trace_entry);

   /*
    * Initialize trace header
    */
   memset(p, 0, sizeof(struct trace_head));
   spin_lock_init(&p->lock);
   p->beg         = (struct trace_entry *) (p + 1);
   p->end         = p->beg + size;
   p->head        = p->beg + 1;
   p->tail        = p->beg;
   p->time_fn     = time_fn;
   p->time_freq   = time_freq;
   p->stamp_valid = 1;
   p->enabled     = 0;
   p->seq         = 0;
   strncpy(p->trace_name, name, TRACE_MAX_NAME_LEN);
   p->trace_name[TRACE_MAX_NAME_LEN] = 0;

   /* Initialize tail trace entry */
   p->tail->sz    = 1;
   p->tail->flags = T_FREE; /* Not valid */

   p->unused1 = 0;
   p->unused2 = 0;
   p->unused3 = 0;

   init_magic(p);

   return 0;
}

int
ose_trace_recover(void *buffer, uint32_t size, const char *name,
                  GetTimeFuncPointer time_fn, uint32_t time_freq)
{
   struct trace_head *p = buffer;
   int result;

   result = check_trace(p, size, name, time_freq);

   if (result != 0)
   {
      /* The trace could not be recovered. */
      int rc = ose_trace_create(buffer, size, name, time_fn, time_freq);
      if (rc < 0)
      {
         result = rc;
      }
   }
   else
   {
      /* The trace is recoverable. */

      /* Reinitialize timer function pointer if that has changed. */
      p->time_fn = time_fn;
      p->time_freq = time_freq;

      /* The above code may have altered the checksum so recalculate. */
      p->checksum = compute_checksum(p);

      ose_trace_disable(p);
   }

   return result;
}

void
ose_trace_clear(void *buffer)
{
   struct trace_head *p = buffer;
   LOCKSTATE msr;

   SPIN_LOCK_SAVE(&p->lock, msr);
   p->seq++;
   SPIN_LOCK_RESTORE(&p->lock, msr);
}

void
ose_trace_enable(void *buffer)
{
   struct trace_head *p = buffer;
   p->enabled = 1;
}


void
ose_trace_disable(void *buffer)
{
   struct trace_head *p = buffer;
   p->enabled = 0;
}

void *
ose_trace_alloc(void *buffer, EntrySize size, uint32_t flags)
{
   LOCKSTATE msr;
   volatile struct trace_head *p = (struct trace_head *) buffer;
   struct trace_entry *tl;
   struct trace_entry *hd;
   int head_wrapped = 0;
   int tail_wrapped = 0;
   EntrySize sz;

   if (!p->enabled && !(flags & T_FORCE))
   {
      return NULL;
   }

   sz = trace_entry_sz(size);

   /* Disable interrupts in case we are called concurrently. */
   SPIN_LOCK_SAVE(&p->lock, msr);

   tl = p->tail;
   hd = p->head;

   /* Allocate a new trace entry if one can be obtained. */
   for(;;)
   {
      if (likely(tl >= hd))
      {
         if (likely(hd + sz < tl))
         {
            /* Space available at 'hd' pointer */
            break;
         }
         else if (likely((tl->flags & (T_BUSY | T_LOCKED | T_REFERRED)) == 0))
         {
            /* Advance tail, dropping last entry to create some free space. */
            tl += tl->sz;
            if (tl == p->end)
            {
               if (hd + sz < tl)
               {
                  p->tail = p->beg;
                  /* Space available at 'hd' pointer */
                  break;
               }
               /* Check that we have not wrapped before. */
               if (unlikely(tail_wrapped))
               {
		  uint32_t head_d = hd - p->beg;
		  uint32_t tail_d = p->tail - p->beg;
                  /* Prevent tl from being at end and hd at start next run. */
                  p->head   = hd;
		  MEMBAR();
                  SPIN_LOCK_RESTORE(&p->lock, msr);
                  return NULL;
               }
               tl = p->beg;
               tail_wrapped = 1;
            }
            p->tail = tl;
         }
         else /* flags has T_BUSY, T_LOCKED or T_REFERRED */
         {
            /* Skip and invalidate busy entry at the tail. */

            if (hd != tl)
            {
               /* Throw away free space at 'hd' since it's too small to use. */
               hd->sz    = (uint16_t)(tl - hd);
               hd->flags = T_FREE;
            }
            if (unlikely(tl->flags & T_LOCKED))
            {
                  tl->ow_count++;
            }
            else if (tl->flags & T_REFERRED)
            {
               if (!(tl->flags & T_BUSY))
               {
                  if (tl->ow_count)
                  {
                     tl->flags &= ~T_REFERRED;
                  }
                  else
                  {
                     tl->ow_count++;
                  }
               }
            }
            tl->flags &= ~T_VALID;
            tl = tl + tl->sz;
            if (unlikely(tl == p->end))
            {
               /* Check that we have not wrapped before. */
               if (unlikely(tail_wrapped))
               {
		  uint32_t head_d = hd - p->beg;
		  uint32_t tail_d = p->tail - p->beg;
                  /* Prevent tl from being at end and hd at start next run. */
                  p->head   = hd;
		  MEMBAR();
                  SPIN_LOCK_RESTORE(&p->lock, msr);
                  return NULL;
               }
               tl = p->beg;
               tail_wrapped = 1;
            }
            p->tail = hd = tl;
         }
      }
      else /* tl < hd */
      {
         if (hd + sz < p->end)
         {
            /* Space available at 'hd' pointer */
            break;
         }
         else if (unlikely(!head_wrapped))
         {
            /*
             * At end of trace buffer and the space left to p->end is too small
             * to hold the requested trace. Wrap head pointer and throw away
             * end bit too small to use (but only if it exists).
             */
            if(likely(hd < p->end))
            {
               hd->sz    = (uint16_t)(p->end - hd);
               hd->flags = T_FREE;
            }

            hd = p->beg;
            head_wrapped = 1; /* Note that we have tried this. */
         }
         else
         {
	    uint32_t head_d = p->head - p->beg;
	    uint32_t tail_d = p->tail - p->beg;
            /* We did not find a hole big enough so bail out. */
	    MEMBAR();
            SPIN_LOCK_RESTORE(&p->lock, msr);
            return NULL;
         }
      }
   }

   /* Advance head */
   p->head   = hd + sz;

   assert(p->head != p->tail);
   assert(p->head != p->end);

   hd->flags = flags | T_VALID | T_BUSY;
   hd->sz    = sz;

   if (p->time_fn && !(flags & T_NOTIME))
   {
      hd->time_hi = p->time_fn(&hd->time_lo);
   }
   else
   {
      hd->time_hi = hd->time_lo = 0;
   }
   MEMBAR();
   SPIN_LOCK_RESTORE(&p->lock, msr);

   hd->size  = size;
   hd->seq   = p->seq;

   /* Return found block. */
   return hd+1;
}

void
ose_trace_commit(void *entry)
{
   struct trace_entry *e = entry;
   (e-1)->flags &= ~T_BUSY;
}

void
ose_trace_reclaim(void *entry)
{
   struct trace_entry *e = entry;
   (e-1)->flags = T_VALID | T_BUSY | T_HIDDEN;
}

void *
ose_trace_find(void *first_mem, uint32_t mem_size, const char *name)
{
   struct trace_head *head = NULL;
   char              *pos;

   /* Align up first_mem to an even 4-byte pointer */
   pos = (char *) ((uint32_t)first_mem & ~3);

   /* Search for trace. Return 0 if not found. */
   if (mem_size < sizeof(struct trace_head))
   {
      /* Too small range. */
      return 0;
   }

   mem_size -= sizeof(struct trace_head);
   mem_size >>= 2;

   while (mem_size-- > 0)
   {
      struct trace_head *p = (struct trace_head *)pos;

      /* If magic identity is OK. */
      if (check_magic(p))
      {
         /* If name is OK. */
         if((!*name) || (!strcmp(name, (const char *)p->trace_name)))
         {
            head = p;
            break;  /* Found it! */
         }
      }
      pos += 4;
   }

   return head;
}

int
ose_trace_open(void *buffer, TraceEntryState *cursor)
{
   volatile struct trace_head *trace = buffer;

   if (!check_magic(trace))
   {
      return -1;
   }

   /* Initialize return structure. */
   cursor->trace_entry = NULL;
   cursor->stamp_valid = trace->stamp_valid;
   cursor->time_freq   = trace->time_freq;
   cursor->base        = trace;
   strncpy(cursor->trace_name, (const char *)trace->trace_name, TRACE_MAX_NAME_LEN);
   cursor->trace_name[TRACE_MAX_NAME_LEN] = 0;

   return 0;
}

int
ose_trace_close(TraceEntryState *cursor)
{
   struct trace_entry *here;

   if (cursor->trace_entry != NULL)
   {
      /* Continue from trace entry at cursor */
      here = (struct trace_entry *)cursor->trace_entry - 1;

      /* Unlock this entry */
      assert(here->flags & T_LOCKED);
      here->flags &= ~T_LOCKED;
      MEMBAR();

      cursor->trace_entry = NULL;
   }

   return 0;
}

/*
 * FIXME: When reading the trace continuously we need to wait for busy entries
 * to be completed, but when reading the traditional way we need to skip them.
 * With the existing public interface there seems to be no way for us to know
 * what to do. For now use this non public function from the RMM for contiuous
 * reading, and let the public function support the traditional use case.
 */
int zzose_trace_read(TraceEntryState *cursor, int wait);

int
zzose_trace_read(TraceEntryState *cursor, int wait)
{
   volatile struct trace_head  *trace = cursor->base;
   struct trace_entry *here;
   struct trace_entry *start = NULL;
   int                 loss = 0;
   LOCKSTATE           msr;

   if (trace == NULL || trace->magic_head != OSE_TRACE_MAGIC_HEAD)
   {
      /* Something is broken, we have an invalid trace or cursor */
      return 0;
   }

   if (cursor->trace_entry == NULL)
   {
	  SPIN_LOCK_SAVE(&trace->lock, msr);
      here = trace->tail;
   }
   else
   {
      /* Continue from trace entry at cursor */
      here = te_header(cursor->trace_entry);

      if (0 != (((unsigned long)here) & 0xf)  /* pointer must be aligned */
          || !(here >= trace->beg && here < trace->end && (here + here->sz) <= trace->end)
          || (0 == (here->flags & T_LOCKED)) /* entry must be locked */
          || (0 != (here->flags & (T_BUSY | T_HIDDEN)))) /* entry must not be busy or hidden */
      {
	    ramlog_printf("ramtrace: Invalid cursor. here = %x, beg = %x, end = %x",
      	               here, trace->beg, trace->end);
	    if (0 == (((unsigned long)here) & 0xf)
	        && (here >= trace->beg && here < trace->end && (here + here->sz) <= trace->end))
	    {
	       ramlog_printf(" flags = %u sz = %u size = %u\n",
			             here->flags, here->sz, here->size);
	    }
	    else
	    {
	       ramlog_printf("\n");
	    }
        /* Something has gone wrong with this entry or cursor */
        return 0;
      }
      start = here;

      SPIN_LOCK_SAVE(&trace->lock, msr);

      /* If tail has passed the cursor, start over from current tail */
      if (!(here->flags & T_VALID))
      {
         loss = (trace->end - trace->beg) * (here->ow_count - 1);
         if (trace->tail > here)
            loss += trace->tail - here - here->sz;
         else
         {
            loss += trace->tail - trace->beg;
            loss += trace->end - here + here->sz;
         }
         /* Calculate loss in bytes */
         loss *= sizeof(struct trace_entry);
         here = trace->tail;
      }
      else
      {
         /* Returns NULL if no more entries */
         here = next_entry(trace, here);
      }
   }

   /* Skip entries that are not valid. */
   while (here != NULL
          && ((wait) ? !will_be_readable(trace, here)
                     : !is_readable(trace, here)))
   {
      here = next_entry(trace, here);
   }

   if (here != NULL && is_readable(trace, here))
   {
      if (start)
      {
         /* Unlock the previous entry */
         start->flags &= ~T_LOCKED;
      }
      here->flags |= T_LOCKED;
      here->ow_count = 0;

      /* We have now locked the new entry and can unlock interrupts. */
      SPIN_LOCK_RESTORE(&trace->lock, msr);

      /* Read next trace entry into return structure. */
      cursor->trace_entry = te_data(here);
      cursor->tick        = here->time_hi;
      cursor->time_count  = here->time_lo;
      cursor->entry_size  = here->size;
      if (0 == loss)
      {
         loss = 1;
      }
   }
   else
   {
      if (cursor->trace_entry)
      {
         (te_header(cursor->trace_entry))->flags |= T_LOCKED;
      }
      SPIN_LOCK_RESTORE(&trace->lock, msr);
   }

   return loss;
}

int
ose_trace_read(TraceEntryState *cursor)
{
   return zzose_trace_read(cursor, 0);
}

#ifdef USE_TRACE_LEGACY

/*
 *===========================================================================
 *                          LEGACY FUNCTIONS
 *===========================================================================
 */

void *
ose_trace_alloc_subentry(void *buffer, EntrySize size)
{
   return ose_trace_alloc(buffer, size, T_HIDDEN | T_NOTIME);
}

void
ose_trace_free_subentry(void *entry)
{
   struct trace_entry *buf = entry;
   (buf - 1)->flags = T_FREE;
}

static int
read_next(struct trace_head *trace, struct trace_entry *here, TraceEntryState *cursor)
{
   int result = 0;

   /* Find next valid trace entry. Return 0 if empty. */
   while ((here = next_entry(trace, here)) != NULL)
   {
      /* Skip next entry if erased or invalid. */
      if (is_readable(trace, here))
      {
         break;
      }
   }

   if (here != NULL && is_readable(trace, here))
   {
      /* Read next trace entry into return structure. */
      cursor->trace_entry = te_data(here);
      cursor->tick        = here->time_hi;
      cursor->time_count  = here->time_lo;
      cursor->entry_size  = here->size;
      here->flags |= T_LOCKED;
      result = 1;
   }
   return result;
}

int
ose_trace_get_first(void *buffer, TraceEntryState *cursor)
{
   struct trace_head *p = buffer;
   struct trace_entry *tl;
   int result = 0;
   LOCKSTATE msr;

   /* Initialize return structure. */
   cursor->base        = (void *) p;
   cursor->stamp_valid = p->stamp_valid;
   cursor->time_freq   = p->time_freq;
   strncpy(cursor->trace_name, p->trace_name, TRACE_MAX_NAME_LEN);
   cursor->trace_name[TRACE_MAX_NAME_LEN] = 0;

   SPIN_LOCK_SAVE(&p->lock, msr);

   /* Trace is empty if head == tail. */
   if ((tl = p->tail) != p->head)
   {
      result = 1;

      tl->flags |= T_LOCKED;

      cursor->trace_entry = te_data(p->tail);
      cursor->tick        = tl->time_hi;
      cursor->time_count  = tl->time_lo;
      cursor->entry_size  = tl->size;

      /* If tail entry is not valid, scan for oldest valid trace entry. */
      if (!is_readable(p,tl))
      {
         result = read_next(p, tl, cursor);
      }
   }

   SPIN_LOCK_RESTORE(&p->lock, msr);

   return result;
}

int
ose_trace_get_next(TraceEntryState *cursor)
{
   struct trace_head  *trace = (struct trace_head *)cursor->base;
   struct trace_entry *here  = te_header(cursor->trace_entry);
   int result = 0;
   LOCKSTATE msr;

   assert(here->flags & T_LOCKED);

   here->flags &= ~T_LOCKED;

   if (!(here->flags & T_VALID))
   {
      /* Tail has passed the cursor, start over from current tail */
      return ose_trace_get_first(trace, cursor);
   }

   SPIN_LOCK_SAVE(&trace->lock, msr);

   /* More entries? */
   if (here != trace->head)
   {
      result = read_next(trace, here, cursor);
   }

   SPIN_LOCK_RESTORE(&trace->lock, msr);

   return result;
}
#endif
