/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2009 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

/**
 *  @File prof_hash.c
 *  @brief This is the implementation profiling processes hash table used in prof.c
 *
 */

#include "prof_hash.h"
/* #include "ramtrace.h" */

/* static functions */
static uint32_t prof_hash_get_pid_hash(uint32_t pid, struct prof_hash *data);
static OSBOOLEAN prof_hash_search_overflow_list(uint32_t pid, uint32_t *index, struct prof_hash *data);
static OSBOOLEAN prof_hash_add_pid_to_overflow_list(const uint32_t pid, uint32_t *index, struct prof_hash *data);
static OSBOOLEAN prof_add_pid_to_hash_list(const uint32_t pid, uint32_t *index, struct prof_hash *data);

extern union SIGNAL *alloc_sample_pool(uint32_t size);
extern void free_sample_pool(union SIGNAL ** pool);

void prof_hash_init(struct prof_hash *data)
{
	  data->index_cnt = 0;
	  data->pid_hash_table = NULL;
	  data->hash_list_size = 0;
	  data->hash_list_mask = 0;
	  data->pid_overflow_list = NULL;
	  data->overflow_list_size = 0;
	  data->overflow_list_pos_cnt = 0;
	  data->sample_counter = 0;
	  data->hash_maxpid = 0;
	  SMP_CODE(spin_lock_init(&data->hash_lock));
}
/*
** Adds a PID entry to the sample array and to the indexing look-up lists.
** If both the hash position and the overflow lists are full the "other"
** processes position is returned.
*/
void
prof_hash_add_pid(const uint32_t pid, uint32_t *index, struct prof_hash *data)
{

   /* unsafe make sure that the central_pid_lock[block] is taken */
   /* If the process list is full add process to "other" */
   if(data->index_cnt == data->hash_maxpid)
   {
      /* The process list is full for this sample so we return "other" */
      *index = data->hash_maxpid;
   }
   else
   {
      if(!prof_add_pid_to_hash_list(pid, index, data))
      {
         if(!prof_hash_add_pid_to_overflow_list(pid, index, data))
         {
            /* If we fail to add the pid to the overflow list (list full) we
             ** add it to the "other" processes slot */
            *index = data->hash_maxpid;
         }
      }
   }
}

/*
** Looks for the specified PID at it's hash list position. If that position
** already is occupied by another PID for this sample, the overflow list
** is searched.
*/
OSBOOLEAN
prof_hash_get_pid(uint32_t pid, uint32_t *index, struct prof_hash *data)
{
  OSBOOLEAN retval = (OSBOOLEAN) 0;
  uint32_t hash_index;

  hash_index = prof_hash_get_pid_hash(pid, data);

  if(data->pid_hash_table[hash_index].sample == data->sample_counter)
  {
    if( data->pid_hash_table[hash_index].pid == pid )
    {
      *index = data->pid_hash_table[hash_index].index;
      retval = (OSBOOLEAN) 1;
    }
    else
    {
      if(prof_hash_search_overflow_list(pid, index, data))
      {
        retval = (OSBOOLEAN) 1;
      }
      else
      {
        retval = (OSBOOLEAN) 0;
      }
    }
  }
  return retval;
}

void prof_hash_reset(struct prof_hash *data)
{
   data->overflow_list_pos_cnt = 0;
   data->index_cnt = 0;
   data->sample_counter++;
}

/*
** Frees the index buffers for per process profiling
*/
void
prof_hash_free_proc_index_bufs(struct prof_hash *data)
{
   free_sample_pool( (union SIGNAL **)&data->pid_hash_table );
   free_sample_pool( (union SIGNAL **)&data->pid_overflow_list);
}



/*
** Sets the global hash size and overflow list sizes so that the selected
** number of PIDs will fit into them and performance on hashing will be
** good
*/
static void
prof_hash_set_indexing_sizes(uint32_t max_nr_pids, struct prof_hash *data)
{
   uint32_t min_pids_size = max_nr_pids;
  /* hash_maxpid is the highest value we may return as index */
  data->hash_maxpid = max_nr_pids;

  /* Minimum hash list size is set to 32 for performance reasons */
  if (min_pids_size < 32)
  {
    min_pids_size = 32;
  }


  data->hash_list_mask = 1;
  /* maxpid is chosen as the power of 2 on or closest above max_nr_pids */
  while(data->hash_list_mask < min_pids_size)
  {
    data->hash_list_mask = (data->hash_list_mask << 1) | 1 ;
  }
  data->hash_list_size = data->hash_list_mask + 1;

  /* The overflow list only need to handle max_nr_pids for it's worst case */
  data->overflow_list_size = max_nr_pids;
}


/*
** Sets up and intializes (or re-initializes existing) indexing
** buffers for the per process monitoring
*/
OSBOOLEAN
prof_hash_create_proc_index_bufs(uint32_t procs, struct prof_hash *data)
{
  OSBOOLEAN retval = (OSBOOLEAN) 0;
  uint32_t i;

  /* Calculate and set indexing function sizes */
  prof_hash_set_indexing_sizes(procs, data);

  /* Free any existing indexing buffers */
  prof_hash_free_proc_index_bufs(data);

  /* Allocate the hash list search structure and reset it */
  data->pid_hash_table =
        (void *) alloc_sample_pool(sizeof(PidHashItem) * data->hash_list_size);

  /* Allocate the hash overflow list and reset it*/
  data->pid_overflow_list =
        (void *) alloc_sample_pool(sizeof(PidHashItem) * data->overflow_list_size);

  if( (data->pid_hash_table != NULL) &&
      (data->pid_overflow_list != NULL))
  {
    /* Else Initialize buffers and variables */
    for(i = 0; i < data->hash_list_size; i++)
    {
      data->pid_hash_table[i].pid = 0;
    }
    for(i = 0; i < data->overflow_list_size; i++)
    {
      data->pid_overflow_list[i].pid = 0;
    }
    retval = (OSBOOLEAN) 1;
  }
  else
  {
    /* Failed to allocate */
    prof_hash_free_proc_index_bufs(data);
    retval = (OSBOOLEAN) 0;
  }

  /* Reset other indexing variables */
  data->sample_counter = 0;
  data->index_cnt = 0;
  data->overflow_list_pos_cnt = 0;

  return retval;
}


/* static functions after this line */
/*
** Tries to add a PID to it's corresponding position in the hash list.
** If the position already is filled for this sample it returns false.
*/
static OSBOOLEAN
prof_add_pid_to_hash_list(const uint32_t pid, uint32_t *index, struct prof_hash *data /*, struct prof_proc_sample *sample*/)
{
  uint32_t hash_index;

  /* If hash position is vacant, use it for the pid */
  hash_index = prof_hash_get_pid_hash(pid, data);

  if( data->pid_hash_table[hash_index].sample != data->sample_counter )
  {
    /* Set data in hash list */
    data->pid_hash_table[hash_index].sample = data->sample_counter;
    *index = data->pid_hash_table[hash_index].index = data->index_cnt++;
    data->pid_hash_table[hash_index].pid = pid;
    /* Set data in process list */
   /* sample->pids[data->index_cnt] = pid;
    sample->samples[data->index_cnt] = 0; */
    return (OSBOOLEAN) 1;
  }
  else
  {
    return (OSBOOLEAN) 0;
  }
}


/* Do a linear search of the overflow list for index of supplied PID nr */
static OSBOOLEAN
prof_hash_search_overflow_list(uint32_t pid, uint32_t *index, struct prof_hash *data)
{
  uint32_t i;

  for( i = 0; i < data->overflow_list_pos_cnt; i++ )
  {
    if(data->pid_overflow_list[i].pid == pid)
    {
      if(data->pid_overflow_list[i].sample == data->sample_counter)
      {
        *index = data->pid_overflow_list[i].index;
        return True;
      }
    }
  }
  return False;
}


/*
** Returns the hash table position for a specific PID
*/
static uint32_t
prof_hash_get_pid_hash(uint32_t pid, struct prof_hash *data)
{
   return pid & data->hash_list_mask;
}

/*
**  Add item to the next vacant list position in the overflow list
*/

static OSBOOLEAN
prof_hash_add_pid_to_overflow_list(const uint32_t pid, uint32_t *index, struct prof_hash *data)
{
  OSBOOLEAN retval;

  if(data->overflow_list_pos_cnt < data->overflow_list_size)
  {
    /* Set data in overflow list */
    data->pid_overflow_list[data->overflow_list_pos_cnt].pid = pid;
    *index = data->pid_overflow_list[data->overflow_list_pos_cnt].index = data->index_cnt++;
    data->pid_overflow_list[data->overflow_list_pos_cnt].sample = data->sample_counter;

  /*  sample->pids[data->index_cnt] = pid;
    sample->samples[data->index_cnt] = 0; */

    /* Update overflow list counter and return success */
    data->overflow_list_pos_cnt++;
    retval = (OSBOOLEAN) 1;
  }
  else
  {
    /* The overflow list is full !*/
    retval = (OSBOOLEAN) 0;
  }
  return retval;
}


