/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2012 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */



/*
 * bind_proc(cpuid_t cpu)
 *
 * Migrates ose_monitor process to profiling core.
 *
 * cpu - The source profiling cpu core. 
 *  
 * returns 0 on sucess and MONITOR_STATUS_BIND_FAILED on fail.
 */
#ifdef OSE_SMP
static MonitorStatus
bind_proc(cpuid_t cpu)
{
   MonitorStatus reply_status = MONITOR_STATUS_OK;

   if (!ose_bind_process(rw.counter_state.rmm_pid, cpu))
   {
      ramlog_printf("\nrmm: could not bind the ose_monitor process to cpu %d "
                    "for source profiling.\n", cpu);

      reply_status = MONITOR_STATUS_BIND_FAILED;
   }
   return reply_status;
}
#endif

/*
 * allocate_sample_buffers(uint32_t count, uint32_t size)
 *
 * Allocate sample buffers.
 *
 * nbr - number of buffers.
 */
static void
allocate_sample_buffers(uint32_t count, uint32_t size)
{
   union SIGNAL *buffer;
   uint32_t i;
   for (i = 0; i < count; i++) 
   {
      buffer = alloc(sizeof(MonitorCounterReportsNotify) +
                           (size - 1) * sizeof(uint64_t),
                           MONITOR_COUNTER_REPORTS_NOTIFY);

      profiling_add_sample_buffer((uint64_t *)&buffer->counter_reports_notify.reports,
                                    size);
   }
}


/**************************************************
 * Callbacks
 **************************************************/

/*
 * counter_buffer_full_callback(uint32_t event_type,
 *                               uint64_t *sample_buffer,
 *                               uint32_t missed_events)
 *
 * Callback used by the counter profiling API.
 *
 * event_type    - type of counter.
 * sample_buffer - buffer with sampled data.
 * missed_events - number of missed events. 
 */
static void
counter_buffer_full_callback(uint32_t event_type,
                             uint64_t *sample_buffer,
                             uint32_t missed_events)
{
   union SIGNAL *notify;
   union SIGNAL *buffer = NULL;

   if (missed_events == 0) 
   {
      buffer = (union SIGNAL *)(((char *)sample_buffer)
                  - offsetof(MonitorCounterReportsNotify, reports));
      buffer->counter_reports_notify.euId = ose_cpu_id(); 
      buffer->counter_reports_notify.reserved0 = 0;
      buffer->counter_reports_notify.counterType = event_type;
      buffer->counter_reports_notify.reserved1 = 0;
      buffer->counter_reports_notify.counterReportType = MONITOR_COUNTER_REPORT;
      buffer->counter_reports_notify.reportsCount = 
                                     rw.counter_state.max_reports;
   }

   notify = alloc(sizeof(BufferFullNotify), BUFFER_FULL_NOTIFY);
   notify->buffer_full_notify.buffer = buffer;
   notify->buffer_full_notify.euId = ose_cpu_id();

   send(&notify, rw.counter_state.rmm_pid); 
}


/*
 * counter_free_buffer_callback(uint64_t *sample_buffer)
 *
 * Callback used by the counter profiling API.
 *
 * sample_buffer - buffer with sampled data. 
 */
static void
counter_free_buffer_callback(uint64_t *sample_buffer)
{
   union SIGNAL *buffer;

   buffer = (union SIGNAL *)(((char *)sample_buffer)
		      - offsetof(MonitorCounterReportsNotify, reports));
   free_buf(&buffer); 
}

/*
 * get_counter_type_enable(cpuid_t cpu, uint32_t type,
 *                         MonitorBoolean *enabled, uint64_t *counterValue)
 *
 * Helper function for handle_get_counter_type_enabled_request().
 *
 * cpu - The cpu id 
 * type - Event type 
 * enabled - sets enabled value of reply signal
 * counterValue - sets counter value of reply signal
 * Returns MonitorStatus
 */
static MonitorStatus
get_counter_type_enable(cpuid_t cpu, uint32_t type, 
                        MonitorBoolean *enabled, uint64_t *counterValue)
{
   MonitorStatus reply_status = MONITOR_STATUS_OK;
   ProfilingPerfCounterStatus counter_status = PROFILING_PC_STATUS_OK;
   uint64_t trigger_count = 0;
   uint32_t use_trigger = 0;

#ifdef OSE_SMP
   reply_status = bind_proc(cpu); 
   if (reply_status != MONITOR_STATUS_OK)
   {
      return reply_status;
   }
#endif

   if (profiling_is_counting(type))
   {
      *enabled = MONITOR_BOOLEAN_TRUE;
      counter_status = profiling_get_trigger_count(type,
                                                   &trigger_count, 
                                                   &use_trigger);

      if (counter_status != PROFILING_PC_STATUS_OK) 
      {
         *counterValue = 0;
         return MONITOR_STATUS_REQUEST_INVALID;
      }
      else 
      {
         *counterValue = trigger_count;
      }
   }
   else 
   {
      uint32_t type_count;
      const ProfilingPerfCounterEventType *types;
      uint32_t i;
      enabled = MONITOR_BOOLEAN_FALSE;
      types = profiling_get_supported_counter_event_types(&type_count);
      *counterValue = 0;

      for (i = 0; i < type_count; i++) 
      {
         if (types[i].type == type) 
         {
             *counterValue = 2000000;
         }
      }
   }
#ifdef OSE_SMP
   reply_status = bind_proc(0);
#endif

   return reply_status;
}

/*
 * handle_get_counter_type_enabled_request(Offspring *offspring, 
 *                                                   union SIGNAL *sig)
 * 
 * Signal handler, get information about the state 
 * of a specific counter type.
 *
 * offspring - Offspring structure for offspring and client pid.
 * sig - request signal. 
 */
static int 
handle_get_counter_type_enabled_request(Offspring *offspring, union SIGNAL *sig)
{
   MonitorStatus reply_status = MONITOR_STATUS_OK;
   cpuid_t cpu = sig->get_counter_type_enabled_request.euId ;
   MonitorBoolean enabled = 0; 
   uint64_t counterValue = 0; 
   uint32_t type = sig->get_counter_type_enabled_request.counterType;
   union SIGNAL *reply = alloc(sizeof(MonitorGetCounterTypeEnabledReply),
                               MONITOR_GET_COUNTER_TYPE_ENABLED_REPLY);

#ifdef OSE_SMP
   cpuid_t cpu_id;
   if (cpu == MONITOR_ALL_EXECUTION_UNITS)
   {
      for (cpu_id = 0; cpu_id < ose_num_cpus(); cpu_id++)
      {
         reply_status = get_counter_type_enable(cpu_id, type, 
                                                &enabled, &counterValue);
         if (reply_status != MONITOR_STATUS_OK)
         {
            break;
         }
      }
   }
   else
   {
      reply_status = get_counter_type_enable(cpu, type, 
                                             &enabled, &counterValue);
   } 
#else
   reply_status = get_counter_type_enable(cpu, type, &enabled, &counterValue);
#endif

    reply->get_counter_type_enabled_reply.counterType =
            sig->get_counter_type_enabled_request.counterType;
    reply->get_counter_type_enabled_reply.maxReports = 
                                          rw.counter_state.max_reports;
    reply->get_counter_type_enabled_reply.euId = cpu;
    reply->get_counter_type_enabled_reply.reserved0 = 0;
    reply->get_counter_type_enabled_reply.reserved1 = 0;
    reply->get_counter_type_enabled_reply.reserved2 = 0;

    reply->get_counter_type_enabled_reply.counterValue = counterValue;
    reply->get_counter_type_enabled_reply.enabled = enabled;

    reply->get_counter_type_enabled_reply.status = reply_status;
    send_w_s(&reply, offspring->pid, offspring->client_);
    free_buf(&sig);
    return 1;
}


/*
 * any_active_counter() 
 * 
 * Check if there are any active counters running.
 *
 * return 0 if no counters running, otherwise 1
 */
static uint8_t 
any_active_counter() 
{
   uint32_t i; 

   /* Get the map entry */
   for (i = 0; i < MAP_SIZE; i++) 
   {
      MapEntry *tmp = rw.counter_state.map[i];
      if (tmp == 0) 
      {
         continue;
      }
      
      if (profiling_is_counting(tmp->type)) 
      {
         return 1;
      } 
   } 
   return 0;
}

/*
 * remove_sample_buffers() 
 * 
 * Remove the sample buffers if no counter are running.
 *
 * return 0 success, 1 fail
 */
static uint8_t 
remove_sample_buffers() 
{
   ProfilingPerfCounterStatus counter_status = PROFILING_PC_STATUS_OK;

   /* Flush the sample buffer if no counter is running */
   if ((profiling_flush_sample_buffers != NULL) && !any_active_counter()) 
   {
      counter_status = 
                   profiling_flush_sample_buffers(counter_free_buffer_callback);
      if (counter_status != PROFILING_PC_STATUS_OK) 
      {
         return 1; 
      }
   }
   return 0;
}

/*
 * add_map_entry(Offspring *offspring, uint32_t type, cpuid_t cpu) 
 * 
 * Add en entry in the map containing client and counter types.
 * If a counter is running and a counter of the same type is 
 * being added, the operation will fail with return value 1.
 *
 * offspring - Offspring structure for offspring and client pid.
 * type   - Counter type.
 * cpu - Cpu id.
 *
 * return 0 succes, 1 fail
 */
static uint8_t 
add_map_entry(Offspring *offspring, uint32_t type, cpuid_t cpu) 
{
   uint32_t i = 0;
   uint16_t filled_map = 1;

   if (profiling_is_counting(type)) 
   {
      return 1; /* The counter is in use */   
   }

   if (!offspring->pid) 
   {
      return 1;
   }
   else 
   { 
      /* Add client to map when enabling the profiling */
      for (i = 0; i < MAP_SIZE; i++) 
      {
         if (rw.counter_state.map[i] == 0) 
         {
             MapEntry *entry = (MapEntry *) alloc(sizeof(MapEntry ), 0);
             memset (entry, 0, sizeof(MapEntry));
             entry->offspring = offspring;
             entry->type = type;
             entry->enabled = MONITOR_BOOLEAN_FALSE;
             entry->loss = 0; 
             entry->cpu = cpu; 
             rw.counter_state.map[i] = entry;
             filled_map = 0;
             break;
         }
      }
   }
   if (i == MAP_SIZE && filled_map) 
   {
      return 1;  /* Map is filled */
   }
   else 
   {
      return 0;
   }
}

/*
 * remove_map_entry(Offspring *offspring, uint32_t type, cpuid_t cpu) 
 * 
 * Remove one entry in the map containing client and counter types.
 *
 * offspring - pointer to a Offspring struct.
 * type   - counter type.
 * cpu - cpu core.
 */
static uint8_t
remove_map_entry(Offspring *offspring, uint32_t type, cpuid_t cpu)
{
   uint32_t i = 0;
   uint8_t ret = 0;

#ifdef OSE_SMP
   if ((ret = bind_proc(cpu)) != MONITOR_STATUS_OK)
   {
       return ret;
   }
#endif

   /* Remove client from map when disabling the profiling */
   for (i = 0; i < MAP_SIZE; i++) 
   {
      MapEntry *entry = rw.counter_state.map[i];
      if (entry == 0) 
      {
         continue;
      }

      if (entry->offspring->pid == offspring->pid &&
            entry->type == type && entry->cpu == cpu) 
      { 
          /* The counter is active, stop it */   
          if (profiling_stop_counting(entry->type) != PROFILING_PC_STATUS_OK) 
          {
             ret = 1;
	  }
          rw.counter_state.map[i] = 0;
          free_buf((union SIGNAL **)&entry);
          break;
      }
   }
 
   if (remove_sample_buffers()) 
   {
        ret = 1;       
   }

#ifdef OSE_SMP
    ret = bind_proc(0);
#endif

   return ret;
}

/*
 * remove_client_map_entry(Offspring *offspring) 
 * 
 * Remove all entries belongin to the client
 * and if there are any active counters,
 * stop them now.
 *
 * offspring - Offspring structure for offspring pid.
 */
static void
remove_client_map_entry(Offspring *offspring)
{
   uint32_t i = 0;

   /* Remove all entries for this client */
   for (i = 0; i < MAP_SIZE; i++) 
   {
      MapEntry *entry = rw.counter_state.map[i];
      if (entry == 0) 
      {
         continue;
      }

      if (entry->offspring->pid == offspring->pid) 
      {
         if (profiling_is_counting(entry->type)) 
         {
            /* The counter is active, stop it.
             * We don't bother about the return value from
             * profiling_stop_counting() since we
             * don't have any clients to send it to.  
             */   
            profiling_stop_counting(entry->type); 
         }          
         rw.counter_state.map[i] = 0;
         free_buf((union SIGNAL **)&entry);
      }
   }

   remove_sample_buffers();
}

/*
 * get_map_entry(uint32_t type, cpuid_t cpu, MapEntry **entry) 
 * 
 * Get en entry in the map containing client, counter types, enable/disable.
 * If no matching entry is found the operation will fail with return value 1.
 *
 * type   - counter type.
 * cpu    - cpu id.
 * entry  - pointer to a MapEntry struct.
 *
 * return 0 if entry is found, otherwise 1
 */
static uint8_t 
get_map_entry(uint32_t type, cpuid_t cpu, MapEntry **entry) 
{
   uint32_t i; 

   /* Get the map entry */
   for (i = 0; i < MAP_SIZE; i++) 
   {
      MapEntry *tmp = rw.counter_state.map[i];
      if (tmp == 0) 
      {
         continue;
      }
      
      if (type == tmp->type && tmp->offspring->pid && cpu == tmp->cpu) 
      {
         *entry = tmp;
         return 0;
      } 
   } 
   return 1;
}

/*
 * set_counter_type_enable(Offspring *offspring, 
 *                         uint32_t type, uint32_t max_reports, 
 *                         uint64_t trigger_count, cpuid_t cpu)
 * 
 * Helper function for handle_set_counter_type_enabled_request().
 *
 * offspring  - offspring pid to be added to MapEntry.
 * type       - counter type. 
 * max_reports - number of reports to be sent.
 * trigger_count - The trigger count to start counter profier.
 * cpu - The cpu id. 
 */
static MonitorStatus
set_counter_type_enable(Offspring *offspring, 
                        uint32_t type, uint32_t max_reports, 
                        uint64_t trigger_count, cpuid_t cpu)
{
   ProfilingPerfCounterStatus counter_status = PROFILING_PC_STATUS_OK;
   MonitorStatus reply_status = MONITOR_STATUS_OK;
   uint32_t pre_alloc_buf = 1;

#ifdef OSE_SMP
   reply_status = bind_proc(cpu);
   if (reply_status != MONITOR_STATUS_OK)
   {
      return reply_status;
   }
#endif

   /* Add client to map when enable the profiling */
   if (add_map_entry(offspring, type, cpu))
   {
      return MONITOR_STATUS_RESOURCE_EXHAUSTED;
   }
   else
   {
      if (any_active_counter())
      {
         /* We have active counters, reuse the max_reports */
          max_reports = rw.counter_state.max_reports;
      }
      else
      {
          pre_alloc_buf = 5; /* First session, pre allocate buffers */
           /* Reuse the max_reports for the following sessions */
          rw.counter_state.max_reports = max_reports;
      }

      allocate_sample_buffers(pre_alloc_buf, max_reports);
      counter_status = profiling_start_counting_with_sampling(type,
                                                  trigger_count,
                                                  counter_buffer_full_callback);

      if (counter_status != PROFILING_PC_STATUS_OK)
      {
         /* Remove an entry from Map, as starting profiling failed. */
         uint8_t remove_entry_status = remove_map_entry(offspring, type, cpu);
         assert(MONITOR_STATUS_OK == remove_entry_status);
         return MONITOR_STATUS_RESOURCE_EXHAUSTED;
      }
   }

#ifdef OSE_SMP 
   reply_status = bind_proc(0);
#endif

   return reply_status;
}

/*
 * handle_set_counter_type_enabled_request(Offspring *offspring, 
 *                                         union SIGNAL *sig)
 * 
 * Signal handler, set a counter profiling session.
 *
 * offspring - an offspring structure for offspring and client pid.
 * sig - request signal. 
 */
static int 
handle_set_counter_type_enabled_request(Offspring *offspring, union SIGNAL *sig)
{
   MonitorStatus reply_status = MONITOR_STATUS_OK;
   union SIGNAL *reply = alloc(sizeof(MonitorSetCounterTypeEnabledReply),
                                MONITOR_SET_COUNTER_TYPE_ENABLED_REPLY);
   cpuid_t cpu = sig->set_counter_type_enabled_request.euId;
   uint32_t type = sig->set_counter_type_enabled_request.counterType;

   if (sig->set_counter_type_enabled_request.counterValue == 0) 
   {
       reply_status = MONITOR_STATUS_REQUEST_INVALID;
   }
   else if (sig->set_counter_type_enabled_request.enabled) 
   {
      /* Check the buffer size */ 
      uint32_t max_reports = sig->set_counter_type_enabled_request.maxReports;
      uint32_t sigsize = sizeof(MonitorCounterReportsNotify) + 
                               (max_reports - 1) * sizeof(uint64_t);
      if (!get_max_sigsize(sigsize))
      { 
         ramlog_printf("\nrmm: set_counter_type_enabled failed to allocate "
                       "%d bytes:  No signal large enough.\n", max_reports);
         reply_status = MONITOR_STATUS_MEMORY_EXHAUSTED;
      }
      else
      {
         uint64_t trigger_count = 
                             sig->set_counter_type_enabled_request.counterValue;
#ifdef OSE_SMP 
         cpuid_t cpu_id, remove_cpu_id;
         if (cpu == MONITOR_ALL_EXECUTION_UNITS)
         {
            for ( cpu_id = 0; cpu_id < ose_num_cpus(); cpu_id++)
            {
               reply_status = set_counter_type_enable(offspring, 
                                                      type, max_reports, 
                                                      trigger_count, cpu_id);
               if (reply_status != MONITOR_STATUS_OK)
               {
                   /* Failed to start for ALL execution units because 
                    * already profiling for cpu/core is running.
                    * Stop profiling for execution units which were 
                    * started for ALL.
                    */
                   for (remove_cpu_id = 0; remove_cpu_id < cpu_id; 
                        remove_cpu_id++)
                   {
                       uint8_t remove_entry_staus = remove_map_entry(offspring,
                                                           type, remove_cpu_id);
                       assert(MONITOR_STATUS_OK == remove_entry_staus);
                   }
                   break;
               }
            }
         }
         else
         {
            reply_status = set_counter_type_enable(offspring, 
                                                   type, max_reports, 
                                                   trigger_count, cpu);
         }

#else 
         reply_status = set_counter_type_enable(offspring, type, max_reports, 
                                                trigger_count, cpu);
#endif
      }
   }
   else
   {
      uint8_t remove_entry_staus = 0;
       /* Remove an entry in the client - counter type map */
#ifdef OSE_SMP
      cpuid_t cpu1;
      if (cpu ==  MONITOR_ALL_EXECUTION_UNITS)
      {
         for (cpu1 = 0; cpu1 < ose_num_cpus(); cpu1++)
         {
            remove_entry_staus = remove_map_entry(offspring, type, cpu1);
            assert(MONITOR_STATUS_OK == remove_entry_staus);
         }    
      }
      else
      {
         remove_entry_staus = remove_map_entry(offspring, type, cpu);
         assert(MONITOR_STATUS_OK == remove_entry_staus);
      }
#else 
      remove_entry_staus = remove_map_entry(offspring, type, cpu);
      assert(MONITOR_STATUS_OK == remove_entry_staus);
#endif
      (void)remove_entry_staus;	/* Kill compiler warning */
   }
   reply->set_counter_type_enabled_reply.status = reply_status;
   send_w_s(&reply, offspring->pid, offspring->client_);
   free_buf(&sig);
   return 1;
}


/*
 * handle_get_counter_reports_notify_enabled_request(Offspring *offspring, 
 *                                                        union SIGNAL *sig)
 * 
 * Signal handler, get the enable/disable state of the 
 * counter profiling session.
 *
 * offspring - Offspring structure for offspring and client pid.
 * sig - request signal. 
 */
static int
handle_get_counter_reports_notify_enabled_request(Offspring *offspring,
                                                  union SIGNAL *sig)
{
   MonitorStatus reply_status = MONITOR_STATUS_OK;
   union SIGNAL *reply = 
                       alloc(sizeof(MonitorGetCounterReportsNotifyEnabledReply),
                              MONITOR_GET_COUNTER_REPORTS_NOTIFY_ENABLED_REPLY);
   uint32_t type = sig->get_counter_reports_notify_enabled_request.counterType;
   cpuid_t cpu = sig->get_counter_reports_notify_enabled_request.euId;
   MapEntry *entry = NULL;

#ifdef OSE_SMP
   cpuid_t cpu_id;
   if (cpu == MONITOR_ALL_EXECUTION_UNITS)
   {
      for (cpu_id = 0; cpu_id < ose_num_cpus(); cpu_id++)
      {
         if (get_map_entry(type, cpu_id, &entry)) 
         {
            reply->get_counter_reports_notify_enabled_reply.status = 
                                              MONITOR_STATUS_REQUEST_INVALID;
            reply->get_counter_reports_notify_enabled_reply.enabled = 
                                              MONITOR_BOOLEAN_FALSE;
            break;
         }
         else 
         {
             reply->get_counter_reports_notify_enabled_reply.status = 
                                                             reply_status;
             reply->get_counter_reports_notify_enabled_reply.enabled = 
                                                             entry->enabled;
         }
      }
   }
   else
   {
      if (get_map_entry(type, cpu, &entry))
      {
         reply->get_counter_reports_notify_enabled_reply.status = 
                                                 MONITOR_STATUS_REQUEST_INVALID;
         reply->get_counter_reports_notify_enabled_reply.enabled = 
                                                 MONITOR_BOOLEAN_FALSE;
      }
      else
      {
         reply->get_counter_reports_notify_enabled_reply.status = reply_status;
         reply->get_counter_reports_notify_enabled_reply.enabled = 
                                                              entry->enabled;
      }

   } 
#else
   if (get_map_entry(type, cpu, &entry))
   {
      reply->get_counter_reports_notify_enabled_reply.status = 
                                                MONITOR_STATUS_REQUEST_INVALID;
      reply->get_counter_reports_notify_enabled_reply.enabled = 
                                                MONITOR_BOOLEAN_FALSE;
   }
   else
   {
      reply->get_counter_reports_notify_enabled_reply.status = reply_status;
      reply->get_counter_reports_notify_enabled_reply.enabled = entry->enabled;
   }
#endif  

   reply->get_counter_reports_notify_enabled_reply.counterType = type;
   reply->get_counter_reports_notify_enabled_reply.euId = cpu;
   reply->get_counter_reports_notify_enabled_reply.reserved0 = 0;

   send_w_s(&reply, offspring->pid, offspring->client_);
   free_buf(&sig);
   return 1;
}

/*
 * set_counter_reports_notify(uint32_t type, cpuid_t cpu, 
 *                                           MonitorBoolean enabled)
 * 
 * Helper function for handle_set_counter_reports_notify_enabled_request.
 * 
 * type - Event type.
 * cpu - Cpu id.
 * enabled - enabled/disabled status 
 * returns MonitorStatus 
 */
static MonitorStatus
set_counter_reports_notify(uint32_t type, cpuid_t cpu, MonitorBoolean enabled)
{
   MapEntry *entry = NULL;
   MonitorStatus reply_status = MONITOR_STATUS_OK;

   if (get_map_entry(type, cpu, &entry)) 
   {
      reply_status = MONITOR_STATUS_REQUEST_INVALID;
   }
   else 
   {
      if (enabled) 
      {
         /* Enable notifications here */
         entry->enabled = MONITOR_BOOLEAN_TRUE;
      }
      else 
      {
         /* Disable notifications here */
         entry->enabled = MONITOR_BOOLEAN_FALSE;
      }
   }
   return reply_status;
}

/*
 * handle_set_counter_reports_notify_enabled_request(Offspring *offspring, 
 *                                                            union SIGNAL *sig)
 * 
 * Signal handler, set the enable/disable state of the 
 * counter profiling session.
 *
 * offspring - offspring structure for offspring and client pid.
 * sig - request signal. 
 */
static int
handle_set_counter_reports_notify_enabled_request(Offspring *offspring, 
                                                            union SIGNAL *sig)
{
   MonitorStatus reply_status = MONITOR_STATUS_OK;
   union SIGNAL *reply = 
                       alloc(sizeof(MonitorSetCounterReportsNotifyEnabledReply),
                              MONITOR_SET_COUNTER_REPORTS_NOTIFY_ENABLED_REPLY);

   uint32_t type = sig->set_counter_reports_notify_enabled_request.counterType;
   cpuid_t cpu = sig->set_counter_reports_notify_enabled_request.euId;

   MonitorBoolean enabled = 
                        sig->set_counter_reports_notify_enabled_request.enabled;

#ifdef OSE_SMP
   cpuid_t cpu_id;
   if (cpu == MONITOR_ALL_EXECUTION_UNITS)
   {
      for (cpu_id = 0; cpu_id < ose_num_cpus(); cpu_id++)
      {  
         reply_status = set_counter_reports_notify(type, cpu_id, enabled);
         if (reply_status != MONITOR_STATUS_OK)
         {
            break;
         }
      } 
   }
   else
   {
      reply_status = set_counter_reports_notify(type, cpu, enabled);
   }
#else
   reply_status = set_counter_reports_notify(type, cpu, enabled);
#endif    

   reply->set_counter_reports_notify_enabled_reply.status = reply_status;
   send_w_s(&reply, offspring->pid, offspring->client_);
   free_buf(&sig);
   return 1;
}

/*
 * handle_disconnect_request_counter(union SIGNAL *sig)
 * 
 * Signal handler, to handle the disconnect request.
 *
 * offspring - Offspring for offspring_pid.
 * sig - request signal. 
 */
static void
handle_disconnect_request_counter(Offspring *offspring, union SIGNAL *sig)
{
   /* rmm's disconnect will remove the data for the client
    * thus the Offspring pointed to by the offspring->pid 
    * is 0 and not sender(&sig) if the rmm's signal handler 
    * is called before this one.
    */

#ifdef OSE_SMP
   cpuid_t cpu;
   for (cpu = 0; cpu < ose_num_cpus(); cpu++)
   { 
      if (bind_proc(cpu) == MONITOR_STATUS_OK)
      {
#endif

         if (offspring && ((offspring->pid == 0) || 
            (offspring->pid == sender(&sig))))
         {
            remove_client_map_entry(offspring);
         }
#ifdef OSE_SMP
         bind_proc(0);
      }
   }
#endif
}    

/*
 * send_reports(MapEntry *entry, union SIGNAL *buffer)
 * 
 * Send the reports to the client, also sends
 * the flow control request when needed.
 *
 * entry  - pointer to a MapEntry
 * buffer - pointer to the signal containing the reports 
 */
static void
send_reports(MapEntry *entry, union SIGNAL *buffer)
{
   Offspring *offspring = NULL;
   union SIGNAL *flowsig, *lossnotify;

   offspring = entry->offspring; 

   if (offspring->flow_ctrl_nblock > 0) 
   {
      /* If we have loss send them */
      if (entry->loss) 
      {
         lossnotify = alloc(sizeof(MonitorCounterReportsLossNotify),
                                       MONITOR_COUNTER_REPORTS_LOSS_NOTIFY);
         lossnotify->counter_reports_loss_notify.euId = entry->cpu;
         lossnotify->counter_reports_loss_notify.reserved0 = 0;
         lossnotify->counter_reports_loss_notify.counterType = entry->type; 
         lossnotify->counter_reports_loss_notify.reportsLost = entry->loss;
         lossnotify->counter_reports_loss_notify.reserved1 = 0;
         entry->loss = 0;
         send_w_s(&lossnotify, offspring->pid,  offspring->client_);
      } 

      offspring->flow_ctrl_nblock--;

      if (offspring->flow_ctrl_nblock == (rw.flow_ctrl_nblock / 2)) 
      {
		   
         flowsig = alloc(sizeof (MonitorFlowControl),
                         MONITOR_NON_BLOCKING_FLOW_CONTROL_REQUEST);
 
         send_w_s(&flowsig, offspring->pid, offspring->client_);
      }
        
      send_w_s(&buffer, offspring->pid, offspring->client_);
#ifdef OSE_SMP
      if (bind_proc(entry->cpu) == MONITOR_STATUS_OK)
      {
         allocate_sample_buffers(1, rw.counter_state.max_reports);
         bind_proc(0);
      }
#else
      allocate_sample_buffers(1, rw.counter_state.max_reports);
#endif
   }
   else 
   { 
      /* We have not received the nonblocking flow control reply, so add loss */
      entry->loss += buffer->counter_reports_notify.reportsCount;

#ifdef OSE_SMP
      if (bind_proc(entry->cpu) == MONITOR_STATUS_OK)
      {
         profiling_add_sample_buffer(
                            (uint64_t *)&buffer->counter_reports_notify.reports,
                            buffer->counter_reports_notify.reportsCount);
         bind_proc(0);
      }
#else
      profiling_add_sample_buffer(
                            (uint64_t *)&buffer->counter_reports_notify.reports,
                            buffer->counter_reports_notify.reportsCount);

#endif
   }
}

/*
 * handle_buffer_full_notify(union SIGNAL *notify)
 * 
 * Handle the buffer full signal from the callback 
 * that contain the sampled data.
 *
 * notify - request signal. 
 */
static void
handle_buffer_full_notify(struct BufferFullNotify *notify)
{
   union SIGNAL *buffer;
   MapEntry *entry = NULL;
   uint32_t type;
   cpuid_t cpu;

   if (notify->buffer == NULL) 
   {
#ifdef OSE_SMP
      if (bind_proc(notify->euId) == MONITOR_STATUS_OK)
      {
         allocate_sample_buffers(2, rw.counter_state.max_reports);
         bind_proc(0);
      }
#else
      allocate_sample_buffers(2, rw.counter_state.max_reports);
#endif
   }
   else 
   {
      buffer = notify->buffer;
      type = buffer->counter_reports_notify.counterType;
      cpu = buffer->counter_reports_notify.euId;
      if (!get_map_entry(type, cpu, &entry)) 
      {        
         if (entry->enabled) 
         {
            send_reports(entry, buffer);
         }    
         else 
         {
#ifdef OSE_SMP
            if (bind_proc(notify->euId) == MONITOR_STATUS_OK)
            { 

               profiling_add_sample_buffer(
                         (uint64_t *)&buffer->counter_reports_notify.reports, 
                         buffer->counter_reports_notify.reportsCount);
               bind_proc(0);
            }
#else
            profiling_add_sample_buffer(
                          (uint64_t *)&buffer->counter_reports_notify.reports, 
                          buffer->counter_reports_notify.reportsCount);
#endif 
         }
      }
   }
}


/*
 * rmm_counterprof_int(void)
 *
 * Initialization function for RMM counter profiler.
 */

static void rmm_counter_prof_init(void)
{
   uint32_t i = 0;

   rw.counter_state.rmm_pid = current_process();
   rw.counter_state.max_reports = 1000;

   for (i = 0; i <  MAP_SIZE; i++) 
   {
      rw.counter_state.map[i] = 0;
   }
}
 
