/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2010 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

/*
 * This file contains internal signals for the RMM.
 */

#ifndef RMM_INTERNAL_SIG_
#define RMM_INTERNAL_SIG_

#include "stdint.h"
#include "ose.h"
#include "monitor.sig"
#include "pose.con"

/*
** The complexpid structure contains a single process name
** expression expanded for quick compare against a
** debug_pcb structure.
**
** We support any flavor of the "segmentpid:block:process" form, including
** the special substring formats "#0x...." and "$0x...." for
** pids and entrypoints.
**
** See also: match_proc(), struct procnamelist
*/

struct complexpid
{
   uint32_t           sid;
   union
   {
      uint32_t        bid;
      char           *name;
   } blockptr;
   union
   {
      uint32_t        pid;
      char           *name;
      void           *entrypoint;
   } procptr;
   uint8_t         segmenttype;         /* 0=match anything, 1=sid */
   uint8_t         blocknametype;       /* 0=match anything, 1=pid, 2=string. */
   uint8_t         procnametype;        /* 0=match anything, 1=pid, 2=string, 3=entrypoint. */
};

/*
** The procnamelist structure contains a list of process name
** expressions expanded for quick compare against a
** debug_pcb structure.
**
** We treat simple pid expressions in a special way by storing them
** in a sorted array. This is because we expect large lists of such pids
** from graphical system debugger tools, and we want to provide
** reasonable efficiency for this case.
**
** Note that this structure has three variable size arrays. We actually use
** only the first one, the rest are declared only for clarity. Singlepids and
** strings are instead appended in the buffer space remaining after the
** complexpidlist[] array.
**
** See also: match_proc(), struct complexpid
*/

struct procnamelist
{
   uint32_t        no_of_singlepids;
   uint32_t        no_of_complexpids;
   PROCESS        *singlepidlist;
   struct complexpid complexpidlist[1]; /* Variable size. */
   uint32_t        singlepids[1];       /* Variable size. */
   char            names[1];    /* Variable size. */
};

#define START_MODEFLAG (31906)  /* !-SIGNO(struct start_modeflag)-! */

struct start_modeflag
{
   SIGSELECT       sig_no;
   uint8_t         mode;
};

/* 35950 Reserved for prof to rmm communication see prof.h */

#ifdef USE_EPM

/**
 * @signal RmmScopeAttachSig
 *
 * @brief Used to attach to the source code scope.
 *
 * @member RMM_SCOPE_ATTACH_SIG
 * @member RmmScopeAttachSig
 */

/** @macro RMM_SCOPE_ATTACH_SIG */
#   define RMM_SCOPE_ATTACH_SIG (35952)

/**
 * @type RmmScopeAttachSig
 *
 * @input client_pid Pid of the client that the scope belongs to.
 */
typedef struct RmmScopeAttachSig
{
   SIGSELECT       sig_no;
   PROCESS         offspring_pid;
} RmmScopeAttachSig;

#endif

typedef struct
{
   SIGSELECT       sigNo;
   uint32_t        traceEntry;
   uint32_t        tick;
   uint32_t        ntick;
   uint32_t        abstime;
   uint32_t        abstick;
} ReadTraceReplyHeader;

/*
** Monitor signal sent from debugger hooks to the rmm.
*/

#define MONITOR_REPORT (35953)

/*
** Catch signal sent from debugger hooks to the rmm.
*/

#define CATCH_REPORT (35954)


/*
** Attach signal sent from the "dead" block.
** This will remove posts from user event record
** list! (if block is present in list)
*/

#define USER_EVENT_ATTACH_SIG (35955)

typedef struct UserEventAttachSig
{
   SIGSELECT signo;
   PROCESS sid;
} UserEventAttachSig;

/**
 * @signal MonitorCreateEventStructRequest
 *
 * @brief Create struct with user posts
 *
 * @member MONITOR_CREATE_EVENT_STRUCT_REQUEST
 * @member MonitorCreateEventStructRequest
 */

/** @macro MONITOR_CREATE_EVENT_STRUCT_REQUEST */
#define MONITOR_CREATE_EVENT_STRUCT_REQUEST (35956)

/**
 * @type MonitorCreateEventStructRequest
 *
 * @output sigNo      MONITOR_CREATE_EVENT_STRUCT_REQUEST
 * @output sid        Segment id of the caller
 * @output event id   ID of the event number
 * @output trig       pointer to the trig variable
 * @output trig_value the value of trig
 * @output sid_exist  if not exist attach segment
*/
typedef struct MonitorCreateEventStructRequest
{
   uint32_t sigNo;
   PROCESS sid;
   uint32_t event_id;
   OSADDRESS *trig;
   uint8_t trig_value;
   OSBOOLEAN sid_exist;
} MonitorCreateEventStructRequest;

/**
 * @signal MonitorCreateEventStructReply
 *
 * @brief Monitor create event struct reply
 *
 * @member MONITOR_CREATE_EVENT_STRUCT_REPLY
 */

/** @macro MONITOR_CREATE_EVENT_STRUCT_REPLY */
#define MONITOR_CREATE_EVENT_STRUCT_REPLY (35957)




/*
** Attach signal sent from the "dead" segment.
** This will remove all references to the "dead" segment
*/

#define USER_REPORTS_ATTACH_SIG (35958)

typedef struct UserReportsAttachSig
{
   SIGSELECT signo;
   PROCESS sid;
} UserReportsAttachSig;



#define MONITOR_CREATE_USER_REPORTS_STRUCT_REQUEST 35959
/**
 * @type MonitorCreateReportStructRequest
 *
 * @input sigNo            MONITOR_CREATE_USER_REPORTS_STRUCT_REQUEST
 * @input reportNo         Report class identification number
 * @input trig             Pointer to a non local boolean variable
 * @input multiple         Set to one to create a report class with multiple values
 * @input continuous       Set to one to have the previous value copied to the next
 *                         interval start. Set to zero to have the value cleared at
 *                         interval start.
 * @input maxmin           Set to one to have ose_set_report_val track the maximum and
 *                         minimum during each interval. Set to zero to track the current
 *                         value only.
 * @input sid              The calling application segment id
 */
typedef struct MonitorCreateUserReportsStructRequest
{
   uint32_t sigNo;
   uint32_t reportNo;
   uint32_t *trig;
   MonitorBoolean multiple;
   MonitorBoolean continuous;
   MonitorBoolean maxmin;
   uint32_t sid;
} MonitorCreateUserReportsStructRequest;


#define MONITOR_CREATE_USER_REPORTS_STRUCT_REPLY 35960
/**
 * @type MonitorCreateReportStructReply
 *
 * @output sigNo             MONITOR_CREATE_REPORT_STRUCT_REPLY
 *
 */
typedef struct MonitorCreateUserReportsStructReply
{
   uint32_t sigNo;
} MonitorCreateUserReportsStructReply;

/* 35961 Reserved for prof to rmm communication see prof.h */


/*
** Extended catch or monitor event sent to the client.
** (The extended information consists of pid_from, pid_to,
** ptype_from, ptype_to, file, line, event_absutick and event_seq_no.)
**
** This signal is sent by the LDM instead of the normal event_report
** whenever the current client logged in with the D1[1] bit set to zero
** in the supported_features field of the connect signal.
**
**
** event_tick:
**        The number of ticks after the last system restart at which
**        time the event occurred.
**
** event_utick:
**        The number of microseconds after the last system tick at which
**        time the event occurred.
**
** event_abssec:
**        The most recent absolute time given to the LDM, measured in seconds
**        since 00:00:00 1 jan 1970 GMT.
**        The field conatins zero if no absolute time is known.
**
** event_abstick:
**        The system tick value that was current at the time of event_abssec.
**        The absolute time at which the event occurred can be calculated as
**        event_abssec + (event_tick - event_abstick) / ticks per second.
**
** sig_from:
**        Valid only if event_type = 1, 2, 3, 4, 5, 6, 11 or 13.
**        Index into the data[] array.
**        For send or receive events, this field represents the name
**        of the sending process.
**        For a swap event, this field represents the name of the
**        process swapped out.
**        For other events, it represents the name of the process
**        causing the event.
**        Contains the index of the first character in the
**        process name.
**
** sig_to:
**        Valid only if event_type = 3, 4, 5, 13 or 14.
**        Index into the data[] array.
**        For send or receive events, this field represents the name
**        of the destination process.
**        For a swap event, this field represents the name of the
**        process swapped in.
**        For user reported events, this field represents a format
**        string similar to a printf()-format string.
**        For a bind event, this field represents the name of the
**        process being rebound.
**        Contains the index of the first character in the string.
**
** ecode:
**        Valid only if event_type = 1, 6, 11, 13 or 14.
**        For error events, this field contains the reported error code.
**        For user reported events, this field represents the event name,
**        in which case it contains the index of the first character in
**        the string.
**        For create events, this field contains the block id in which the
**        new process was created.
**        For bind events, this field contains the cpuid of the cpu the
**        re-bound process was moved from.
**
** signo:
**        Valid only if event_type = 4, 5, 6, 11, 13 or 14.
**        This field contains the signal number of the signal sent
**        or received.
**        For error events, this field contains the extra error parameter.
**        For user reported events, this field represents five consecutive
**        32 bit values to be applied to the format string. In this case
**        the field contains the index of the 32 bit word in the array.
**        For bind events, this field contains the cpuid of the cpu the
**        re-bound process was moved to.
**
** cpu:
**        ID of the cpu on which the event occurred.
**
** sig_data:
**        Valid only if event_type = 4 or 5 and if action = 1 or 2 or 10.
**        Index into the data[] array.
**        This field represents all signal data including the signal number.
**        Contains the index of the first byte in the
**        signal data field.
**
** sig_datsize:
**        Valid only if event_type = 4 or 5 and if action = 1 or 2 or 10.
**        Contains the number of data bytes present in the sig_data field.
**
** event_type:
**        This field defines the event type.
**        1  => A process was created.
**        2  => A process was killed.
**        3  => A process was swapped in or out.
**        4  => A signal was sent.
**        5  => A signal was received.
**        6  => The OSE kernel reports an error.
**        7  => An alarm was raised.
**        8  => An alarm was ceased.
**        9  => A system cold start occured.
**        10 => A system warm start occurred.
**        11 => A process reports an error.
**        12 => Montor data was lost. An unknown number of events are missing.
**        13 => A user reported event occurred.
**        14 => A process was rebound.
**
** action:
**        1 => The event was reported by the catcher.
**        2 => The event is a short monitor event.
**        10 => The event is a long monitor event.
**
** sig_address:
**        Valid only if event_type = 4 or 5 and if action = 1 or 2 or 10.
**        Contains a pointer to the signal buffer at the sig_no field.
**        For action = 1, this data must be fetched from the target for
**        display using display_memory signals when the entire signal is
**        needed.
**        Note that this is a slightly unsafe operation since data may no longer
**        be present at the time of the get_memory operation.
**
** sig_size:
**        Valid only if event_type = 4 or 5 and if action = 1 or 2 or 10.
**        Contains the number of data bytes present in the signal.
**        (sigsize(&sig_address))
**
** pid_from:
**        This field contains the process id (pid) of the process
**        indicated in the sig_from field.
**
** pid_to:
**        This field contains the process id (pid) of the process
**        indicated in the sig_to field.
**
** ptype_from:
**        This field contains the type of the process indicated
**        in the sig_from field. The following types are available:
**        0     => A prioritized process.
**        64    => A background process.
**        128   => An interrupt process.
**        256   => A timer process.
**        512   => A phantom process.
**        1024  => A block.
**
** ptype_to:
**        This field contains the type of the process indicated
**        in the sig_to field.
**
** line:
**        For some events, the kernel may be able to indicate the
**        file and line at which the corresponding system call occurred.
**        In these cases, the line field contains the line number
**        and the file field contain the file name.
**        The line field is zero if no information is available.
**
** file:
**        Index into the data[] array.
**        This field is valid only if the line field is nonzero,
**        in which case it contains the index of the first character in
**        the file name string.
**
** action_id:
**        The id of an actionpoint that caused this report.
**
** event_absutick:
**        The microsecond value that was current at the time of event_abssec.
**        The microsend part of the absolute time at which the event
**        occurred can be calculated as event_utick - event_absutick.
**        Please note that the result may be negative, in which case
**        it is necessary to readjust the event_tick and event_utick values.
**
** data[]:
**        Variable size.
**        Contains process names and a binary image of caught or monitored
**        signal contents for send and receive events.
**        Contains printable binary values, format string, event and process
**        names for user reported events.
*/

struct event_report_verbose
{
   SIGSELECT       sig_no;      /* EVENT */
   uint32_t        event_tick;
   uint32_t        event_utick;
   uint32_t        event_abssec;
   uint32_t        event_abstick;
   uint32_t        sig_from;
   uint32_t        sig_to;
   uint32_t        ecode;
   uint32_t        signo;
   uint16_t        cpu;
   uint16_t        sig_data;
   uint16_t        sig_datasize;
   uint8_t         event_type;
   uint8_t         action;
   uint32_t        sig_address;
   uint32_t        sig_size;
   uint32_t        pid_from;
   uint32_t        pid_to;
   uint32_t        ptype_from;
   uint32_t        ptype_to;
   uint32_t        line;
   uint32_t        file;
   uint32_t        action_id;
   uint32_t        event_absutick;
   uint8_t         data[1];
};

/*
 * Sent from the main RMM process to the offspring handling a client.
 */
#define RMM_OFFSPRING_INFO 35961 /* !-SIGNO(RmmOffspringInfo)-! */

typedef struct
{
   SIGSELECT  sig_no;
   PROCESS    client_;
   MonitorScope scope;
   int32_t    flow_ctrl_blocking;
   int32_t    flow_ctrl_non_blocking;
   uint32_t   req_disabled_bitmask[64];
} RmmOffspringInfo;

typedef struct RmmAttSig_s
{
   SIGSELECT sig_no;
   struct Client *client;
}RmmAttSig;

/*
 * Sent from rmm to the correct scope.
 * The receiver will send all its tracepoints to the sender of the request
 */
#define RMM_GET_TRACEPOINTS_REQUEST 35962

typedef struct
{
   SIGSELECT sig_no;
   PROCESS receiver;
} RmmGetTracepointsRequest;

/*
 * Sent from rmm to the correct scope.
 * The receiver will send all its actions for a specified tracepoint to the
 * sender of the request.
 */
#define RMM_GET_TRACEPOINT_ACTIONS_REQUEST 35963

typedef struct
{
   SIGSELECT sig_no;
   PROCESS receiver;
   uint32_t tracepoint_id;
} RmmGetTracepointActionsRequest;

/*
 * Sent from rmm to the correct scope.
 * The receiver will send all its tracepoint variables to the sender of the request
 */
#define RMM_GET_TRACEPOINT_VARIABLES_REQUEST 35964

typedef struct
{
   SIGSELECT sig_no;
   PROCESS receiver;
} RmmGetTracepointVariablesRequest;

/*
 * Sent by an offspring to the monitor to request information about all other
 * offsprings
 */
#define RMM_GET_OFFSPRINGS_REQUEST 35965

#define RMM_GET_OFFSPRINGS_REPLY 35966

typedef struct
{
   SIGSELECT sig_no;
   uint32_t offspringsLeft;
   uint32_t clientPid;
   uint32_t offspringPid;
   MonitorScope scope;
   MonitorScope interceptScope;
   uint32_t nbrOfBreakpoints;
   uint32_t nameLength;
   uint8_t clientName[1];
} RmmGetOffspringsReply;

#define RMM_GET_OFFSPRINGS_ENDMARK 35967

/*
 * Sent by an offspring to an other offspring to request overall tracepoint
 * information.
 */
#define RMM_GET_TP_CONF_INFO_REQUEST 35968

typedef struct
{
   SIGSELECT sig_no;
   uint32_t offspringPid;
} RmmGetTPConfInfoRequest;


#define RMM_GET_TP_CONF_INFO_REPLY 35969

typedef struct
{
   SIGSELECT sig_no;
   uint32_t nbrOfTracepoints;
   uint32_t nbrOfTracepointActions;
   uint32_t nbrOfTracepointVariables;
   uint64_t tracepointBufferSize;
   uint32_t tracepointBufferType;
} RmmGetTPConfInfoReply;

/*
 * Send from rmm to the correct scope.
 * The receiver will send its annotations for a specified tracepoint to the
 * sender of the request.
 */
#define RMM_GET_TRACEPOINT_ANNOTATIONS_REQUEST 35970

typedef struct
{
   SIGSELECT sig_no;
   PROCESS receiver;
   uint32_t tracepoint_id;
} RmmGetTracepointAnnotationsRequest;

/* BUFFER_FULL_NOTIFY:
 *   Buffer full notification for Source profiling.
 */

/* Internal signal. */
#define BUFFER_FULL_NOTIFY (22)

typedef struct BufferFullNotify
{
    uint32_t sigid;
    cpuid_t euId;
    union SIGNAL *buffer;
} BufferFullNotify;

/*
 * Macros
 */

/*
** Lock and unlock the dispatcher. No process switches
** will occur between the LOCK_SWAP() and UNLOCK_SWAP() macros.
**
** There is an important feature in this mechanism.
** If a process uses LOCK_SWAP(), then makes itself not ready and
** attempts a process switch, then the swap is postponed
** until the UNLOCK_SWAP macro is used.
*/

#define LOCK_SWAP() zzlock_dispatcher()

#define UNLOCK_SWAP() {\
                         LOCK();\
                         if (zzunlock_dispatcher()){\
                           zzquit_int();} /* Swap. */\
                         UNLOCK();\
                      }

/*
 * Functions
 * */

OSADDRESS
zzrmm_get_entrypoint(PROCESS pid);

struct procnamelist *
zzrmm_scope_to_plist(MonitorScope * scope, MonitorTag * tags, uint32_t tags_count);

int
zzrmm_is_sysproc(PROCESS pid);

/*
** utility to match a process to a set of wildcard names.
** The debug_pcb structure represents the process to check
** The procnamelist contains a verbose view of the wildcard list
** to check.
** Returns 0 if no match was found.
** Returns 1 if a match was found.
*/

#include "pose.h"

int
zzrmm_match_proc(struct debug_pcb *proc, struct procnamelist *wildcard);

#endif /* RMM_INTERNAL_SIG_ */
