/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2010-2011 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

/* Dear maintainer,
 *
 * Before you go on and change the code there is a few things that I want you to know
 * The first is that the code is written in a way that makes it easy to handle the
 * callback from the EPM, for performance reason there is no error check here. We blindly
 * believe in all the things we see there. Instead there are test of the input values in
 * the interface functions. So don't forget to check input data, at the gate, ones bad
 * input reach the core it will crash.
 * Another thing, still related, is that the layout is made so that the callback is easy
 * to write and also fast. Some fields in the data structures are redundant but put there
 * so that the callback might get a little bit faster.
 *
 * A good thing to know about the code is also that you can do direct struct copy in C,
 * it makes the code a bit more clean and we have used it as often it was possible.
 * That is:
 * struct foo_s foo, bar;
 * foo.a = 5;
 * foo.b = 8;
 * now
 * bar = foo will assign booth a and b in bar;
 *
 * Files that handle tracepoints is:
 * rmm_tracepoint.c: The core. It is a sub part of rmm_offspring.c and is the external
 *                   interface handler.
 *
 * tracepoint_buffer.c: Handle the tracepoint buffer. Allocates and maintains the trace.
 *
 * tracepoint_bytecode.c: Host the tracepoint_bytecode_eval function that is used in
 *                        conditional tracepoints and in some actions.
 *
 * core/arch/<arch>/rmm/tracepoint*: Host exception handling, tracepoint_memcpy
 *                                   and register id translation that is used
 *                                   to copy memory to trace. Only implemented
 *                                   for PowerPC at the time of this note.
 *
 * Have a lot of fun,
 * /The Authors
 */

typedef struct Client Client;

#include "rmm_tracepoint.h"
#include "monitor_execution.sig"
#include "ose.h"
#include "ose_err.h"
#include "assert.h"
#include "stdlib.h"
#include "string.h"
#include "stddef.h"

#include "ramlog.h"
#include "tracepoint_buffer.h"
#include "tracepoint_bytecode.h"
#include "epm.h"
#include "mm.sig"
#define LOCK_FUNCTIONS_CONFIGURED
#include "rmm_internal.sig"
#include "rmm_utility.h"
#include "cpu.h"
#include "cpu_regdump.h"
#include "cpu_access.h"
#include "epm_private.h"
#ifdef PPC
#include "cpu_regs.h"
#include "tracepoint_jitcompiler.h"
#endif
#include "compiler_macros.h"
#include "atomic.h"
#include "rmm.h"


#ifdef PPC
/* Hidden function in core to save timestamp*/
long long
ose_read_tb(void);

/* Register lookup functions. */
extern int
read_register_for_id(struct cpu_core *dump, OSADDRESS gdb_reg_id, OSADDRESS *value);
extern int
is_register_id_valid(OSADDRESS gdb_reg_id);

#else
WEAK_REF(long long ose_read_tb(void));

/* Dummy register lookup functions */
static int
read_register_for_id(void *dump, OSADDRESS gdb_reg_id, OSADDRESS *value)
{
   TOUCH(dump);
   TOUCH(gdb_reg_id);
   TOUCH(value);
   return 0;
}
static int
is_register_id_valid(OSADDRESS gdb_reg_id)
{
   TOUCH(gdb_reg_id);
   return 0;
}

#endif

#define TOUCH(foo) foo=foo

static const char *buffer_sysparam = "krn/monitor/tracepoint";

typedef struct TraceData
{
   ListLink link;
   uint32_t length;
   uint8_t *data;
} TraceData;

typedef struct MonitorTracepointGeneralTracepointSig
{
   SIGSELECT sig_no;
   MonitorStatus status;
} MonitorTracepointGeneralTracepointSig;

union SIGNAL
{
   SIGSELECT sig_no;
   MonitorTracepointGeneralTracepointSig general_sig;
   MonitorTracepointCreateTracepointRequest create_tracepoint_request;
   MonitorTracepointCreateTracepointReply create_tracepoint_reply;
   MonitorTracepointRemoveTracepointRequest remove_tracepoint_request;
   MonitorTracepointRemoveTracepointReply remove_tracepoint_reply;
   MonitorTracepointConfigureTracepointRequest set_tracepoint_enabled_request;
   MonitorTracepointConfigureTracepointReply set_tracepoint_enabled_reply;
   MonitorTracepointCreateActionRequest create_action_request;
   MonitorTracepointCreateActionReply create_action_reply;
   MonitorTracepointSetAnnotationsRequest set_annotations_request;
   MonitorTracepointSetAnnotationsReply set_annotations_reply;
   MonitorTracepointCreateVariableRequest create_variable_request;
   MonitorTracepointCreateVariableReply create_variable_reply;
   MonitorTracepointGetVariableValueRequest get_variable_value_request;
   MonitorTracepointGetVariableValueReply get_variable_value_reply;
   MonitorTracepointRemoveVariableRequest remove_variable_request;
   MonitorTracepointRemoveVariableReply remove_variable_reply;
   MonitorTracepointClearBufferRequest clear_buffer_request;
   MonitorTracepointClearBufferReply clear_buffer_reply;
   MonitorTracepointStartReply start_reply;
   MonitorTracepointStopReply stop_reply;
   MonitorTracepointGetStatusReply get_status_reply;
   MonitorTracepointConfigureTracebufferRequest configure_request;
   MonitorTracepointConfigureTracebufferReply configure_reply;
   struct mm_create_region mm_create_region;
   struct mm_delete_region mm_delete_region;
   RmmGetTracepointsRequest rmm_get_tracepoints_request;
   MonitorTracepointGetTracepointsRequest get_tracepoints_request;
   MonitorTracepointGetTracepointsReply get_tracepoints_reply;
   MonitorTracepointGetTracepointsEndmark get_tracepoints_endmark;
   RmmGetTracepointActionsRequest rmm_get_actions_request;
   MonitorTracepointGetActionsRequest get_actions_request;
   MonitorTracepointGetActionsReply get_actions_reply;
   MonitorTracepointGetActionsEndmark get_actions_endmark;
   RmmGetTracepointAnnotationsRequest rmm_get_annotations_request;
   MonitorTracepointGetAnnotationsRequest get_annotations_request;
   MonitorTracepointGetAnnotationsReply get_annotations_reply;
   RmmGetTracepointVariablesRequest rmm_get_variables_request;
   MonitorTracepointGetVariablesRequest get_variables_request;
   MonitorTracepointGetVariablesReply get_variables_reply;
   MonitorTracepointGetVariablesEndmark get_variables_endmark;
   MonitorTracepointGetTracebufferReply get_tracebuffer_reply;
   MonitorTracepointGetTracebufferEndmark get_tracebuffer_endmark;
   MonitorGetConnectedClientsReply get_connected;
   RmmGetTPConfInfoReply get_tp_conf_info_reply;
   RmmGetOffspringsReply rmm_get_offsprings_reply;
};

/**
 * @function rmm_offspring_send2client
 * @brief send a reply signal that is under flow control.
 * @long returns if successful and kill the current process if the client was lost.
 */
void
zzrmm_send2client(union SIGNAL **signal,
                  Client *client);

typedef void
(*ActionHandler)(TracepointManagerInfo *tpm_info,
                 TracepointAction *actionTracepointManagerInfo,
                 struct cpu_regdump *dump,
                 uint32_t frameid);

#if defined(softOSE) || defined(__i386)

static int
tracepoint_callback(const struct EpmExecutionPoint *ep,
                    struct cpu_regdump *rdump,
                    struct EfmiCurrent *current)
{
   (void)ep;
   (void)rdump;
   (void)current;
   return 0;
}
#else

static int tracepoint_memcpy_generic(void * to, const void *from, size_t size)
{
   memcpy(to, from, size);
   return 0;
}

int
(*tracepoint_memcpy)(void * to, const void *from, size_t size) = tracepoint_memcpy_generic;


static void
log_error(TracepointManagerInfo *tpm_info,
          uint32_t frameid,
          uint32_t ecode,
          uint32_t extra)
{
   enum MonitorTraceStatus status;
   TraceEntry *TraceEntryTp;

   status = tracepoint_buffer_alloc(tpm_info->trace_buffer_handle,
                                    0,
                                    sizeof(Collect),
                                    &TraceEntryTp);
   if (status != TRACEPOINT_RUNNING)
   {
      tpm_info->status = status;
      return;
   }
   TraceEntryTp->type = COLLECT;
   TraceEntryTp->frameid = frameid;
   TraceEntryTp->u.collect.type = COLLECT_ERROR;
   TraceEntryTp->u.collect.u.error.errorcode = ecode;
   TraceEntryTp->u.collect.u.error.extra = extra;
   TraceEntryTp->commit = TRACE_ENTRY_COMMITED;
}

static ActionHandler action_handler[4];

static void
save_reg(TracepointManagerInfo *tpm_info,
         TracepointAction *action,
         struct cpu_regdump *rdump,
         uint32_t frameid)
{
   TraceEntry *trace_entry;
   struct cpu_core *dump = (struct cpu_core *) rdump->sect[0].address;
   uint32_t status;
   uint32_t entrysize;

   TOUCH(action);
   entrysize = sizeof(Collect) + sizeof(struct cpu_core);
   status = tracepoint_buffer_alloc(tpm_info->trace_buffer_handle,
                                    0,
                                    entrysize,
                                    &trace_entry);
   if (status != TRACEPOINT_RUNNING)
   {
      tpm_info->status = status;
      return;
   }
   trace_entry->type = COLLECT;
   trace_entry->frameid = frameid;
   trace_entry->u.collect.type = COLLECT_REGISTER;

   status = tracepoint_memcpy(&trace_entry->u.collect.u,
                              dump,
                              sizeof(struct cpu_core));
   if (status)
   {
      log_error(tpm_info,frameid, TRACEPOINT_ERROR_MEMORY_READ, (uint32_t) dump);
   }
   trace_entry->commit = TRACE_ENTRY_COMMITED;
}

static void
expression_handler(TracepointManagerInfo *tpm_info,
                   TracepointAction *action,
                   struct cpu_regdump *rdump,
                   uint32_t frameid)
{
   unsigned long ecode;
   uint64_t ret = 0;
   int64_t *tvar = tpm_info->variable_table;
   MonitorTracepointActionExpression *expression =
      &action->tracepointAction.expression;
   struct cpu_core *dump = (struct cpu_core *) rdump->sect[0].address;

   ecode = tracepoint_bytecode_eval(tpm_info,
                                    frameid,
                                    expression->byteCode,
                                    expression->byteCodeLength,
                                    dump,
                                    &ret,
                                    tvar,
                                    tpm_info->variable_id_counter);

   if (TRACEPOINT_OK != ecode)
   {
      log_error(tpm_info, frameid, ecode, ret);
   }
}

static void
save_mem(TracepointManagerInfo *tpm_info,
         TracepointAction *action,
         struct cpu_regdump *rdump,
         uint32_t frameid)
{
   uint32_t entrysize;
   TraceEntry *trace_entry;
   enum MonitorTraceStatus status;
   MonitorTracepointActionSaveMemory *save_memory =
   &action->tracepointAction.saveMemory;
   OSADDRESS base = save_memory->base;

   if (save_memory->isRegister)
   {
      struct cpu_core *dump = (struct cpu_core *) rdump->sect[0].address;
      OSADDRESS address = 0;
      int error = read_register_for_id(dump, base, &address);
      if (error)
      {
         log_error(tpm_info, frameid, TRACEPOINT_ERROR_REGISTER, base);
         return;
      }
      base = address;
   }

   entrysize = sizeof(Collect) + save_memory->length;
   status = tracepoint_buffer_alloc(tpm_info->trace_buffer_handle,
                                    0,
                                    entrysize,
                                    &trace_entry);
   if (status != TRACEPOINT_RUNNING)
   {
      tpm_info->status = status;
      return;
   }

   trace_entry->type = COLLECT;
   trace_entry->frameid = frameid;
   trace_entry->u.collect.type = COLLECT_MEMORY;
   trace_entry->u.collect.u.mem.address = base;
   trace_entry->u.collect.u.mem.length = save_memory->length;

   status = tracepoint_memcpy(trace_entry->u.collect.u.mem.data,
                              (void *) base,
                              save_memory->length);
   if (status)
   {
      log_error(tpm_info,frameid, TRACEPOINT_ERROR_MEMORY_READ, base);
   }
   trace_entry->commit = TRACE_ENTRY_COMMITED;
}

/* whilestepping n should step n times and collect data after each step
 * The first call will only set the next ep.
 */
static int
whilestepping_callback(struct EpmExecutionPoint *ep,
                       struct cpu_regdump *rdump,
                       struct EfmiCurrent *current)
{
   TracepointAction *action;
   struct whilestepping_s *stepstone = (struct whilestepping_s *) ep;
   TracepointManagerInfo *tpm_info = stepstone->tpmi;
   EpmStatus epm_status;
   enum MonitorTraceStatus status;
   TraceEntry *TraceEntryTp;
   long long timestamp = 0;
   cpuid_t current_cpu = ose_cpu_id();

   /* Unless the call is made from the tracepoint location, 
    * remove the eps behind us, create a new frame and perform 
    * the actions associated with the while-stepping
    * The test lock strange but the call from whilstepping_handler
    * does not set the current pointer.
    */
   tpm_info->active_callbacks[current_cpu]++;

   if (current)
   {
      stepstone->stepcount--;
      ose_epm_remove_execution_point(ep);

      /* Create TP entry for the new frameid */
      status = tracepoint_buffer_alloc(tpm_info->trace_buffer_handle,
                                       stepstone->tracepointId,
                                       sizeof(Tp),
                                       &TraceEntryTp);
   
      if (status != TRACEPOINT_RUNNING)
      {
         tpm_info->status = status;
         goto exit;
      }

#ifndef softOSE
      if ((void*)ose_read_tb != 0)
      {
         timestamp = ose_read_tb();
      }
#endif

      TraceEntryTp->type = TP;
      TraceEntryTp->u.tp.process = current->process;
      TraceEntryTp->u.tp.timestamp_hi = (0xFFFFFFFF00000000ll & timestamp) >> 32;
      TraceEntryTp->u.tp.timestamp_lo = (0x00000000FFFFFFFFll & timestamp);
      TraceEntryTp->u.tp.tpid = stepstone->tracepointId;
      TraceEntryTp->u.tp.error.errorcode = TRACEPOINT_OK;
      TraceEntryTp->u.tp.error.extra = 0;
      TraceEntryTp->commit = TRACE_ENTRY_COMMITED;

      action = list_head(stepstone->action_list);
      while (action)
      {
         action_handler[action->actionType](tpm_info, action, rdump, TraceEntryTp->frameid);
         action = list_next(stepstone->action_list, action);
      }
   }
   if (stepstone->stepcount > 0)
   {
      epm_status = ose_epm_next_instruction(rdump,
               0, /* Follow function call. */
               &stepstone->ep.address);
      if (epm_status != EPM_OK)
      {
         log_error(tpm_info, 0, TRACEPOINT_ERROR_EPM_ERROR, epm_status);
         goto exit;
      }
      epm_status = ose_epm_add_execution_point(ep);
      if (epm_status != EPM_OK)
      {
         log_error(tpm_info, 0, TRACEPOINT_ERROR_EPM_ERROR, epm_status);
         goto exit;
      }
   }
   else
   {
      /* Ups!! FIXME, we drop a signal here but the idea is that we will fix this when the
       * epm is ready, there is two way to fix, install a free handler or make sure that
       * the epm don't need the data after call to the callback. But now we just drop it,
       */
      free_buf((union SIGNAL **)&stepstone);
   }

  exit:
   tpm_info->active_callbacks[current_cpu]--;
   return 0;
}

static void
whilestepping_handler(TracepointManagerInfo *tpm_info,
                      TracepointAction *action,
                      struct cpu_regdump *rdump,
                      uint32_t frameid)
{
   struct whilestepping_s *stepstone;
   int stepcount = action->tracepointAction.whileStepping.stepCount;
   TOUCH(frameid);

   if (stepcount > 0)
   {
      stepstone = (struct whilestepping_s *) alloc(sizeof(struct whilestepping_s),
                                                   0);
      stepstone->tracepointId = action->tracepointId;
      stepstone->stepcount = stepcount;
      stepstone->tpmi = tpm_info;
      stepstone->ep.flags = 0;
      stepstone->ep.pid = current_process();
      stepstone->ep.cb = (int
                          (*)(const struct EpmExecutionPoint*,
                              struct cpu_regdump*,
                              struct EfmiCurrent*)) whilestepping_callback;
      stepstone->action_list = action->action_list;

      /* Make the first call to while_stepping that way we don't
       * need the first set of a local ep.
       */
      whilestepping_callback(&stepstone->ep,
                             rdump,
                             NULL);
   }
}

static int
tracepoint_callback(struct EpmExecutionPoint *ep,
         struct cpu_regdump *rdump,
         struct EfmiCurrent *current)
{
   ExecutionPointTraceInfo *ept =
      (ExecutionPointTraceInfo *) ((OSADDRESS) ep -
                                   offsetof(ExecutionPointTraceInfo, ep));
   Tracepoint *tp;
   TracepointAction *action;
   TracepointManagerInfo *tpm_info = ept->tpmi;
   struct cpu_core *dump = (struct cpu_core *) rdump->sect[0].address;
#ifdef PPC
   Msr msr = get_msr();
#else
   /* Add support for other arch as tracepoints are fully supported. */
#endif
   cpuid_t current_cpu = ose_cpu_id();

   tpm_info->active_callbacks[current_cpu]++;
   tp = &ept->tp;

   if (tpm_info->status == TRACEPOINT_RUNNING)
   {
      int64_t *tvar = tpm_info->variable_table;
      enum MonitorTraceStatus status;
      unsigned long ecode = TRACEPOINT_OK;
      TraceEntry *TraceEntryTp;
      long long timestamp = 0;
      uint64_t ret = 0;

      if (tp->byteCodeLength)
      {
         /* Enable interrupts only if enabled at tp location. */
#ifdef PPC
         if (dump->msr & MSR_EE)
         {
            set_msr(msr | MSR_EE);
         }
#else
         /* Add support for other arch as tracepoints are fully supported. */
#endif

         ecode = tracepoint_bytecode_eval(tpm_info,
                                          0, /* There is no frameid.*/
                                          tp->byteCode,
                                          tp->byteCodeLength,
                                          dump,
                                          &ret,
                                          tvar,
                                          tpm_info->variable_id_counter);
#ifdef PPC
         /* Restore msr */
         set_msr(msr);
#else
         /* Add support for other arch as tracepoints are fully supported. */
#endif
         if (!ret && ecode == TRACEPOINT_OK)
         {
            /* Nothing more to do byte code
             * eval to False and no error.
             */
            goto exit;
         }
      }

      if (tp->passCount)
      {
         uint32_t passCount = atomic_dec_return((atomic_t *)&tp->passCountCurr);

         /* Check for zero */
         if (passCount == 0)
         {
            tpm_info->status = TRACEPOINT_PASSCOUNT;
         }
      }

      status = tracepoint_buffer_alloc(tpm_info->trace_buffer_handle,
                                       tp->id,
                                       sizeof(Tp),
                                       &TraceEntryTp);
      if (status != TRACEPOINT_RUNNING)
      {
         tpm_info->status = status;
         /* Restore msr
          * NOTE: restoring msr could be skipped, since the epm will disable interrupts
          * when we return from the callback */
         goto exit;
      }
#ifndef softOSE
      if ((void*)ose_read_tb != 0)
      {
         timestamp = ose_read_tb();
      }
#endif
      TraceEntryTp->type = TP;
      TraceEntryTp->u.tp.process = current->process;
      TraceEntryTp->u.tp.timestamp_hi = (0xFFFFFFFF00000000ll & timestamp) >> 32;
      TraceEntryTp->u.tp.timestamp_lo = (0x00000000FFFFFFFFll & timestamp);
      TraceEntryTp->u.tp.tpid = tp->id;
      TraceEntryTp->u.tp.error.errorcode = ecode;
      TraceEntryTp->u.tp.error.extra = ret;
      TraceEntryTp->commit = TRACE_ENTRY_COMMITED;

      if (ecode == TRACEPOINT_OK)
      {
         action = list_head(tp->action_list);
         while (action)
         {
            /* Enable interrupts only if enabled at tp location. */
#ifdef PPC
            if (dump->msr & MSR_EE)
            {
               set_msr(msr | MSR_EE);
               /* disable interrupts again. */
               set_msr(msr);
            }
#else
            /* Add support for other arch as tracepoints are fully supported. */
#endif
            if (!tpm_info->status == TRACEPOINT_RUNNING)
            {
               /* Do not continue if the tracepoints was disabled */
               goto exit;
            }

            /*Save the data in trace*/
            action_handler[action->actionType](tpm_info, action, rdump, TraceEntryTp->frameid);
            action = list_next(tp->action_list,action);
         }
      }
   }

  exit:
   tpm_info->active_callbacks[current_cpu]--;
   return 0;
}

#endif   /* Soft OSE and x86 */

static Tracepoint*
get_tracepoint(TracepointManagerInfo *tpm_info,
               MonitorTracepoint id,
               ExecutionPointTraceInfo **ept_p)
{
   ExecutionPointTraceInfo *ept;
   Tracepoint *tp = NULL;

   /* A linear search through all the trace points.
    * The mess is that tracepoints are in two list.
    * one for enabled and one for disabled tracepoints.
    */
   ept = list_head(tpm_info->enabled_tracepoint_list);
   while (ept)
   {
      if (id == ept->tp.id)
      {
         tp = &ept->tp;
         goto done;
      }
      ept = list_next(tpm_info->enabled_tracepoint_list, &ept->link);
   }

   ept = list_head(tpm_info->disabled_tracepoint_list);
   while (ept)
   {
      if (id == ept->tp.id)
      {
         tp = &ept->tp;
         goto done;
      }
      ept = list_next(tpm_info->disabled_tracepoint_list, &ept->link);
   }

  done:
   /* Return ept pointer if wanted */
   if (ept_p != NULL)
   {
      *ept_p = ept;
   }

   return tp;
}

static void
enable_tracepoint(TracepointManagerInfo *tpm_info, Tracepoint* tp)
{
   ExecutionPointTraceInfo *ept;

   ept = list_head(tpm_info->disabled_tracepoint_list);
   while (ept)
   {
      if (tp->id == ept->tp.id)
      {
         list_unlink(ept);
         list_insert_tail(tpm_info->enabled_tracepoint_list, ept);
         return;
      }
      ept = list_next(tpm_info->disabled_tracepoint_list, ept);
   }
}

static void
disable_tracepoint(TracepointManagerInfo *tpm_info, Tracepoint* tp)
{
   ExecutionPointTraceInfo *ept;

   ept = list_head(tpm_info->enabled_tracepoint_list);
   while (ept)
   {
      if (tp->id == ept->tp.id)
      {
         list_unlink(ept);
         list_insert_tail(tpm_info->disabled_tracepoint_list, ept);
         return;
      }
      ept = list_next(tpm_info->enabled_tracepoint_list, ept);
   }
}


static TracepointAction*
get_while_stepping_action(Tracepoint *tp)
{
   TracepointAction *action = list_head(tp->action_list);
   while (action)
   {
      if (action->actionType == MONITOR_TRACEPOINT_ACTION_WHILE_STEPPING)
      {
         return action;
      }
      action = list_next(tp->action_list, &action->link);
   }
   return NULL;
}

static TracepointVariable*
get_variable(TracepointManagerInfo *tpm_info, MonitorTracepointVariable id)
{
   TracepointVariable *variable = list_head(tpm_info->variables_list);
   while (variable)
   {
      if (variable->id == id)
      {
         return variable;
      }
      variable = list_next(tpm_info->variables_list, &variable->link);
   }
   return NULL;
}

static void
destroy_list(List *list)
{
   void *link;

   while (link = list_head(list), link != NULL)
   {
      list_unlink(link);
      free_buf((union SIGNAL **) &link);
      link = NULL;
   }

   free_buf((union SIGNAL **) &list);
}

static char *
create_buffer_region(uint32_t size, uint32_t *actual_size)
{
   const SIGSELECT all_signals[] =
   { 0 };
   char *memory = NULL;
   union SIGNAL *signal;
   PROCESS ose_mm_;

   /* Make sure memory manager is available */
   if (!hunt(GMM_NAME_STRING, 0, &ose_mm_, NULL))
   {
      ramlog_printf("rmm_tracepoint: ose_memory_mgr not found!\n");
      return NULL;
   }

   signal = alloc(sizeof(struct mm_create_region), MM_CREATE_REGION_REQUEST);

   signal->mm_create_region.size = size;
   signal->mm_create_region.domain = MM_KERNEL_DOMAIN;
   strncpy(signal->mm_create_region.name, "trace_buffer", MM_NAME_LENGTH);
   strncpy(signal->mm_create_region.phys_mem, MM_INTERNAL_RAM, MM_NAME_LENGTH);
   strncpy(signal->mm_create_region.log_mem, "RAM", MM_NAME_LENGTH);
   signal->mm_create_region.permissions = SU_RW_USR_NA;
   signal->mm_create_region.cache_mode = MM_CACHE_COPY_BACK
            | MM_SPECULATIVE_ACCESS;
   signal->mm_create_region.base = 0;
   signal->mm_create_region.options = MM_CREATE_REGION_SPECIFY_NAME
            | MM_CREATE_REGION_SPECIFY_ALIGN;
   signal->mm_create_region.align = 32;
   signal->mm_create_region.phys_count = 0;
   signal->mm_create_region.low_limit = 0;
   send(&signal, ose_mm_);

   signal = receive_from(~0UL, all_signals, ose_mm_);

   /* If mm did not reply in time then... well we don't
    * wait more here at least. The system is probably going
    * down.
    */
   if (signal == NIL)
   {
      ramlog_printf("rmm_tracepoint: Didn't get reply from mm"
                    ", no tracebuffer created.\n");
      return NULL;
   }
   if (signal->sig_no == MM_CREATE_REGION_REPLY)
   {
      if (signal->mm_create_region.status == MM_SUCCESS)
      {
         *actual_size = signal->mm_create_region.size;
         memory = (char *) signal->mm_create_region.base;
      }
      else
      {
         ramlog_printf("rmm_tracepoint: create_region of type \"RAM\" and size %u failed "
                       "with status %u\n",
                       size,
                       signal->mm_create_region.status);
      }
   }
   free_buf(&signal);
   signal = NULL;
   return memory;
}

static void
delete_buffer_region(OSADDRESS base)
{
   const SIGSELECT all_signals[] =
   { 0 };
   union SIGNAL *signal;
   PROCESS ose_mm_;

   /* Make sure memory manager is available */
   if (!hunt(GMM_NAME_STRING, 0, &ose_mm_, NULL))
   {
      ramlog_printf("rmm_tracepoint: ose_memory_mgr not found!\n");
      return;
   }

   signal = alloc(sizeof(struct mm_delete_region), MM_DELETE_REGION_REQUEST);
   signal->mm_delete_region.domain = MM_KERNEL_DOMAIN;
   signal->mm_delete_region.base = base;
   send(&signal, ose_mm_);

   signal = receive_from(~0UL, all_signals, ose_mm_);
   /* If mm did not reply in time then... well we don't
    * wait more here at least. The system is probably going
    * down.
    */
   if (signal == NIL)
   {
      return;
   }
   if (signal->sig_no == MM_DELETE_REGION_REPLY)
   {
      if (signal->mm_delete_region.status != MM_SUCCESS)
      {
         ramlog_printf("rmm_tracepoint: delete_region with base %lu failed with status %u\n",
                       base,
                       signal->mm_delete_region.status);
      }
   }

   free_buf(&signal);
   signal = NULL;
}

static MonitorStatus
check_preconditions_start(TracepointManagerInfo *tpm_info)
{
   /* Check that at least one tracepoint has been created */
   if (!list_length(tpm_info->enabled_tracepoint_list))
   {
      ramlog_printf("rmm_tracepoint: cannot start trace "
                    "experiment without at least one tracepoint\n");
      return MONITOR_STATUS_REQUEST_INVALID;
   }
   /* Check if tracepoint experiment status is valid */
   return MONITOR_STATUS_OK;
}

#define VERIFY_ADDRESS(addr) do {if (addr == 0)\
         return MONITOR_STATUS_REQUEST_INVALID;}while (0)

#define VERIFY_RANGE(element,low,high) do {if ((int)element < low || (int)element > high) \
         return MONITOR_STATUS_REQUEST_INVALID;} while (0)

#define VERIFY_BUFFER_SIZE(size) do {if (size > (uint64_t)0xFFFFFFFF) \
         return MONITOR_STATUS_RESOURCE_EXHAUSTED;} while (0)

#define VERIFY_SCOPE(offspring_scope,tp_scope) do {MonitorScope scope;  \
      scope = zzrmm_get_scope_intersection(offspring_scope,tp_scope);   \
      if (scope.type == MONITOR_SCOPE_VOID)                             \
         return MONITOR_STATUS_REQUEST_INVALID;                         \
   } while(0)

#define VERIFY_BYTECODE_LENGTH(sig, type, field, length) do {  \
      if (sigsize(&sig) < offsetof(type, field) + length - 1)  \
         return MONITOR_STATUS_REQUEST_INVALID;                \
   } while(0)

static int
check_input_param_create_tracepoint(union SIGNAL *sig,
                                    TracepointManagerInfo *tpm_info)
{
   MonitorTracepointCreateTracepointRequest *s;
   s = &sig->create_tracepoint_request;

   VERIFY_ADDRESS(s->address);
   VERIFY_SCOPE(&tpm_info->scope, &s->scope);
   VERIFY_RANGE(s->enabled, 0, 1);

   /* TODO update when fast tp is supported for an ARCH Powerpc currently supported*/
#ifdef PPC
   VERIFY_RANGE(s->tracepointType,
                MONITOR_TRACEPOINT_TRAP_BASED_TRACEPOINT,
                MONITOR_TRACEPOINT_FAST_TRACEPOINT);
#else
   VERIFY_RANGE(s->tracepointType,
                MONITOR_TRACEPOINT_TRAP_BASED_TRACEPOINT,
                MONITOR_TRACEPOINT_TRAP_BASED_TRACEPOINT);
#endif

   VERIFY_BYTECODE_LENGTH(sig, MonitorTracepointCreateTracepointRequest,
                          byteCode,
                          s->byteCodeLength);
   return MONITOR_STATUS_OK;
}

#if defined(softOSE) || defined(__i386)

#else
static void 
update_bytecode_tvar_ids(TracepointManagerInfo *tpm_info)
{
   ExecutionPointTraceInfo *ept;   
   TracepointAction *action;
   TracepointAction *while_stepping_action;
   unsigned long ecode;
   uint64_t ret_val;

   ept = list_head(tpm_info->enabled_tracepoint_list);
   while (ept)
   {
      if(ept->tp.tracepointType != MONITOR_TRACEPOINT_FAST_TRACEPOINT)
      {
         if (!ept->tp.byteCodeUpdated)
         {
            ecode = tracepoint_bytecode_update_variable_ids(tpm_info,
                                                       ept->tp.byteCode,
                                                       ept->tp.byteCodeLength,
                                                       &ret_val); 
            if (ecode == TRACEPOINT_OK) 
            {
               ept->tp.byteCodeUpdated = True;
            }
            else
            {
               log_error(tpm_info, 0, ecode, ret_val); 
            }
         }
      }
      action = list_head(ept->tp.action_list);
      while (action)
      {
         if ((!action->byteCodeUpdated) && 
             (action->actionType == MONITOR_TRACEPOINT_ACTION_EXPRESSION))
         {
            /*UPDATE THE BYTECODE*/ 
            ecode = tracepoint_bytecode_update_variable_ids(tpm_info,
                         action->tracepointAction.expression.byteCode,
                         action->tracepointAction.expression.byteCodeLength,
                         &ret_val);
            if (ecode == TRACEPOINT_OK)
            {
               action->byteCodeUpdated = True;
            }
            else
            {
               log_error(tpm_info, 0, ecode, ret_val);
            }
         }
         else if (action->actionType == MONITOR_TRACEPOINT_ACTION_WHILE_STEPPING)
         {
            while_stepping_action = list_head(action->action_list);
            while (while_stepping_action)
            {
               if ((!while_stepping_action->byteCodeUpdated) && 
                   (while_stepping_action->actionType == MONITOR_TRACEPOINT_ACTION_EXPRESSION))
               {
                  /*UPDATE THE BYTECODE*/ 
                  ecode = tracepoint_bytecode_update_variable_ids(tpm_info,
                               while_stepping_action->tracepointAction.expression.byteCode,
                               while_stepping_action->tracepointAction.expression.byteCodeLength,
                               &ret_val);
                  if (ecode == TRACEPOINT_OK)
                  {
                     while_stepping_action->byteCodeUpdated = True;
                  }
                  else
                  {
                     log_error(tpm_info, 0, ecode, ret_val);
                  }
               }
               while_stepping_action = list_next(action->action_list,
                                           &while_stepping_action->link);
            }
         }
         action = list_next(ept->tp.action_list, action);
      }
      ept = list_next(tpm_info->enabled_tracepoint_list, &ept->link); 
   }
}
#endif /* Soft OSE and x86 */

static int
handle_tracepoint_create_tracepoint_request(union SIGNAL *sig,
                                            TracepointManagerInfo *tpm_info)
{
   ExecutionPointTraceInfo *ept;
   MonitorScope new_scope;
   MonitorStatus monitor_status;
#ifdef PPC
   union SIGNAL *tmp;
   void * code_area = NULL;
   size_t code_size = 0;

   int tp_sp = TP_STACK_SIZE;
   int longcode_ret = 0;
   int ret; 
#endif
   union SIGNAL *reply;
   reply = alloc(sizeof(struct MonitorTracepointCreateTracepointReply),
                 MONITOR_TRACEPOINT_CREATE_TRACEPOINT_REPLY);
   reply->create_tracepoint_reply.byteCodeError = TRACEPOINT_OK;

   monitor_status = check_input_param_create_tracepoint(sig, tpm_info);
   if (monitor_status)
   {
      goto exit;
   }

   /* Create a new trace point object */
   ept = list_init_link(alloc(sizeof(ExecutionPointTraceInfo)
            + sig->create_tracepoint_request.byteCodeLength, 0));
   ept->tp.id = tpm_info->tracepoint_id_counter++;
   ept->tp.address = sig->create_tracepoint_request.address;
   ept->tp.scope = sig->create_tracepoint_request.scope;
   ept->tp.enabled = sig->create_tracepoint_request.enabled;
   ept->tp.passCount = sig->create_tracepoint_request.passCount;
   ept->tp.tracepointType = sig->create_tracepoint_request.tracepointType;
   ept->tp.byteCodeLength = sig->create_tracepoint_request.byteCodeLength;
   ept->tp.byteCodeUpdated = False;
   ept->tp.action_list = list_init((List *) alloc(sizeof(List), 0));
   ept->tp.annotations = NULL;
   memcpy(&ept->tp.byteCode,
          &sig->create_tracepoint_request.byteCode,
          sig->create_tracepoint_request.byteCodeLength);

   ept->address = sig->create_tracepoint_request.address;
   ept->tpmi = tpm_info;
   ept->ep.address = sig->create_tracepoint_request.address;
   ept->ep.cb = (int (*)(const struct EpmExecutionPoint*,
                         struct cpu_regdump*,
                         struct EfmiCurrent*))tracepoint_callback;

   new_scope = zzrmm_get_scope_intersection(&tpm_info->scope, &sig->create_tracepoint_request.scope);
   /* If trig scope is system, we need to set an ep flag to
    * indicate that all processes should be stopped. */
   if (new_scope.type == MONITOR_SCOPE_SYSTEM)
   {
      ept->ep.flags = EPM_EP_ANY_PROCESS;
   }
   else
   {
      ept->ep.flags = 0;
   }
   ept->ep.pid = new_scope.id;
#ifdef PPC
   if(ept->tp.tracepointType == MONITOR_TRACEPOINT_FAST_TRACEPOINT)
   { 
      if(ept->tp.byteCodeLength > 499) /* Max size allow by jit compiler */
      {
        monitor_status = MONITOR_STATUS_REQUEST_INVALID;
        reply->create_tracepoint_reply.byteCodeError  = TRACEPOINT_ERROR_PROGRAM_TOO_LARGE;
        goto exit;  
      }
      if(ept->tp.byteCodeLength == 0)
      {
      monitor_status = MONITOR_STATUS_REQUEST_INVALID;
      reply->create_tracepoint_reply.byteCodeError = TRACEPOINT_ERROR_BYTECODE_MISSING;
      goto exit; 
      }
      longcode_ret= tracepoint_byte2longcode(ept->tp.byteCode,
                                             ept->tp.byteCodeLength,
                                             &tp_sp,&longcode[0]);
      if(longcode_ret != TRACEPOINT_OK )
      {
        monitor_status = MONITOR_STATUS_REQUEST_INVALID;
        reply->create_tracepoint_reply.byteCodeError = longcode_ret;
        goto exit;  
      }
      tmp = alloc(3500,0);
      ret = tracepoint_generate_asmcode(&longcode[0],ept->tp.byteCodeLength,
                                        (uint32_t *)tmp,3500);
      if(ret != TRACEPOINT_OK )
      {
        monitor_status = MONITOR_STATUS_REQUEST_INVALID;
        reply->create_tracepoint_reply.byteCodeError = ret;
        free_buf(&tmp);
        goto exit;  
      }
      code_size = jit_compiler.asmcode_len;
      code_area = ose_epm_alloc_fast_execution_point(&ept->ep, code_size);
      if(code_area == NULL )
      {
        monitor_status = MONITOR_STATUS_EPM_ERROR;
        reply->create_tracepoint_reply.byteCodeError = TRACEPOINT_ERROR_EPM_ERROR;
        free_buf(&tmp);
        goto exit;  
      }
      memcpy(code_area,tmp,code_size);

      ose_cache(CACHE_FLUSH_D, CACHE_PGM_LOAD,  (OSADDRESS) code_area, 
                                                         code_size);
      ose_cache(CACHE_INVALIDATE_I, CACHE_PGM_LOAD, (OSADDRESS) code_area, 
                                                             code_size);
      free_buf(&tmp);
   }
#endif
   if (ept->tp.enabled)
   {
      list_insert_tail(tpm_info->enabled_tracepoint_list, ept);
   }
   else
   {
      list_insert_tail(tpm_info->disabled_tracepoint_list, ept);
   }

   reply->create_tracepoint_reply.id = ept->tp.id;
  exit:

   reply->create_tracepoint_reply.status = monitor_status;
   reply->create_tracepoint_reply.reserved0 = 0;
   reply->create_tracepoint_reply.reserved1 = 0;
   send(&reply, sender(&sig));
   return monitor_status;
}

static int
handle_tracepoint_remove_tracepoint_request(union SIGNAL *sig,
                                            TracepointManagerInfo *tpm_info)
{
   ExecutionPointTraceInfo *ept;
   Tracepoint *tp;
   union SIGNAL *reply;
   MonitorStatus monitor_status = MONITOR_STATUS_OK;

   /* Get the tp from list*/
   tp = get_tracepoint(tpm_info, sig->remove_tracepoint_request.id, &ept);
   if (tp)
   {
      /* Free all actions */
      if (tp->action_list != NULL)
      {
         /*Remove while stepping list, if created*/
         TracepointAction *action = get_while_stepping_action(tp);
         if (action != NULL && action->action_list != NULL)
         {
            tpm_info->nr_of_actions -= list_length(action->action_list);
            destroy_list(action->action_list);
         }

         /* Remove the action list */
         tpm_info->nr_of_actions -= list_length(tp->action_list);
         destroy_list(tp->action_list);
      }
      if (tp->annotations != NULL)
      {
         free_buf((union SIGNAL **) &tp->annotations);
      }

#ifdef PPC
      if(ept->tp.tracepointType == MONITOR_TRACEPOINT_FAST_TRACEPOINT)
      {
        ose_epm_free_fast_execution_point(&ept->ep);
      }
#endif
      list_unlink(ept);
      free_buf((union SIGNAL **) &ept);
      ept = NULL;
   }
   else
   {
      monitor_status = MONITOR_STATUS_ID_INVALID;
   }
   reply = alloc(sizeof(struct MonitorTracepointRemoveTracepointReply),
                 MONITOR_TRACEPOINT_REMOVE_TRACEPOINT_REPLY);
   reply->remove_tracepoint_reply.status = monitor_status;
   reply->remove_tracepoint_reply.reserved0 = 0;
   reply->remove_tracepoint_reply.reserved1 = 0;
   send(&reply, sender(&sig));
   return monitor_status;
}

static int
check_input_param_configure_tracepoint(union SIGNAL *sig)
{
   MonitorTracepointConfigureTracepointRequest *s;
   s = &sig->set_tracepoint_enabled_request;

   VERIFY_RANGE(s->enabled, 0, 1);
   return 0;
}


static int
handle_tracepoint_configure_tracepoint_request(union SIGNAL *sig,
                                                 TracepointManagerInfo *tpm_info)
{
   Tracepoint *tp;
   union SIGNAL *reply;
   MonitorStatus monitor_status;

   monitor_status = check_input_param_configure_tracepoint(sig);
   if (monitor_status != MONITOR_STATUS_OK)
   {
      goto exit;
   }

   /* Get the tp from list*/
   tp = get_tracepoint(tpm_info, sig->set_tracepoint_enabled_request.id, NULL);

   if (tp)
   {
      /* Check if tp need to be moved from enabled or disabled list*/
      if (tp->enabled != sig->set_tracepoint_enabled_request.enabled)
      {
         if (tp->enabled)
         {
            disable_tracepoint(tpm_info, tp);
         }
         else
         {
            enable_tracepoint(tpm_info, tp);
         }
         tp->enabled = sig->set_tracepoint_enabled_request.enabled;
      }
   }
   else
   {
      monitor_status = MONITOR_STATUS_ID_INVALID;
      goto exit;
   }

   exit:
   reply = alloc(sizeof(struct MonitorTracepointConfigureTracepointReply),
                 MONITOR_TRACEPOINT_CONFIGURE_TRACEPOINT_REPLY);
   reply->set_tracepoint_enabled_reply.status = monitor_status;
   reply->set_tracepoint_enabled_reply.reserved0 = 0;
   reply->set_tracepoint_enabled_reply.reserved1 = 0;
   send(&reply, sender(&sig));
   return monitor_status;
}

static int
verify_action_request(union SIGNAL *sig, Tracepoint *tp)
{
   MonitorTracepointActionData * action =
            &(sig->create_action_request.tracepointAction);

   if (tp == NULL)
   {
      return MONITOR_STATUS_ID_INVALID;
   }

   switch (sig->create_action_request.actionType)
   {
      case MONITOR_TRACEPOINT_ACTION_EXPRESSION:
         /* The byte code length must be allocated */
         VERIFY_BYTECODE_LENGTH(sig, MonitorTracepointCreateActionRequest,
                  tracepointAction.expression.byteCode,
                  action->expression.byteCodeLength);
         break;

      case MONITOR_TRACEPOINT_ACTION_SAVE_MEMORY:

         /* If isRegister set the base will be interpret as a register id that
          * holds the base address.
          */
         if (action->saveMemory.isRegister)
         {
            /* Verify register id */
            if (!is_register_id_valid(action->saveMemory.base))
            {
               return MONITOR_STATUS_REQUEST_INVALID;
            }
         }
         else
         {
            /* Base must be separated from NULL, if it is an address */
            VERIFY_ADDRESS(action->saveMemory.base);
         }
         break;

      case MONITOR_TRACEPOINT_ACTION_WHILE_STEPPING:
         /* It is not allowed to create multiple while stepping actions for a
          * tracepoint.
          */
         if (get_while_stepping_action(tp) != NULL)
         {
            return MONITOR_STATUS_REQUEST_INVALID;
         }
         break;

      case MONITOR_TRACEPOINT_ACTION_SAVE_REGISTER:
         /* We do nothing here, since the fields in save
          * registers not can be specified wrongly.
          */
         break;

      default:
         /* Non supported action type. Error is reported */
         return MONITOR_STATUS_REQUEST_INVALID;
   }
   return MONITOR_STATUS_OK;
}

static int
handle_tracepoint_create_action_request(union SIGNAL *sig,
                                        TracepointManagerInfo *tpm_info)
{
   Tracepoint *tp;
   TracepointAction *action = NULL;
   union SIGNAL *reply;
   MonitorStatus monitor_status = MONITOR_STATUS_OK;
   MonitorTracepointActionExpression *exp;

   reply = alloc(sizeof(struct MonitorTracepointCreateActionReply),
                 MONITOR_TRACEPOINT_CREATE_ACTION_REPLY);

   /* Get the tp from list*/
   tp = get_tracepoint(tpm_info, sig->create_action_request.tracepointId, NULL);

   /* Verify that the request signal is valid */
   monitor_status = verify_action_request(sig, tp);
   if (monitor_status != MONITOR_STATUS_OK)
   {
      goto exit;
   }
   /* Create action */
   switch (sig->create_action_request.actionType)
   {
      case MONITOR_TRACEPOINT_ACTION_EXPRESSION:
         exp = &sig->create_action_request.tracepointAction.expression;
         action = list_init_link(alloc(sizeof(TracepointAction)
                  + exp->byteCodeLength, 0));
         action->tracepointAction.expression.byteCodeLength
                  = exp->byteCodeLength;
         memcpy(action->tracepointAction.expression.byteCode,
                exp->byteCode,
                exp->byteCodeLength);
         break;

      case MONITOR_TRACEPOINT_ACTION_WHILE_STEPPING:
         action = list_init_link(alloc(sizeof(TracepointAction), 0));
         action->tracepointAction = sig->create_action_request.tracepointAction;
         action->action_list = list_init((List *) alloc(sizeof(List), 0));
         break;

         /* Intend to fall trough */
      case MONITOR_TRACEPOINT_ACTION_SAVE_REGISTER:
      case MONITOR_TRACEPOINT_ACTION_SAVE_MEMORY:
         action = list_init_link(alloc(sizeof(TracepointAction), 0));
         action->tracepointAction = sig->create_action_request.tracepointAction;
         break;
   }

   if (sig->create_action_request.actionId)
   {

      TracepointAction *a = list_head(tp->action_list);
      while (a)
      {
         if (a->actionId == sig->create_action_request.actionId
                  && a->actionType == MONITOR_TRACEPOINT_ACTION_WHILE_STEPPING)
         {
            action->whileSteppingId = sig->create_action_request.actionId;
            list_insert_tail(a->action_list, action);
            break;
         }
         a = list_next(tp->action_list, a);
      }
      if (a == NULL)
      {
         action->whileSteppingId = 0;
         monitor_status = MONITOR_STATUS_ID_INVALID;
         goto exit;
      }
   }
   else
   {
      action->whileSteppingId = 0;
      list_insert_tail(tp->action_list, action);
   }

   action->actionId = tpm_info->action_id_counter++;
   tpm_info->nr_of_actions++;
   action->tracepointId = tp->id;
   action->actionType = sig->create_action_request.actionType;
   action->byteCodeUpdated = False;
   reply->create_action_reply.actionId = action->actionId;

   exit:
   reply->create_action_reply.status = monitor_status;
   reply->create_action_reply.reserved0 = 0;
   reply->create_action_reply.reserved1 = 0;

   send(&reply, sender(&sig));
   return monitor_status;
}

static int
verify_set_annotations_request(union SIGNAL *sig, Tracepoint *tp)
{
   size_t accumulated_size = 0;
   uint32_t i = 0;
   char *cursor;

   if (tp == NULL)
   {
      return MONITOR_STATUS_ID_INVALID;
   }

   /* We have the right number of strings and the sizes add up */
   cursor = (char *)(&sig->set_annotations_request.annotations[0]);
   for (i = 0; (i < sig->set_annotations_request.no_of_annotations) &&
               (accumulated_size < sig->set_annotations_request.annotations_size);
        i++)
   {
      accumulated_size += strlen(cursor) + 1;
      cursor += strlen(cursor) + 1;
   }

   if (((uint32_t)accumulated_size != sig->set_annotations_request.annotations_size) ||
       (i != sig->set_annotations_request.no_of_annotations))
   {
      return MONITOR_STATUS_REQUEST_INVALID;
   }

   return MONITOR_STATUS_OK;
}

static int
handle_tracepoint_set_annotations_request(union SIGNAL *sig,
                                          TracepointManagerInfo *tpm_info)
{
   Tracepoint *tp;
   union SIGNAL *reply;
   MonitorStatus monitor_status = MONITOR_STATUS_OK;

   reply = alloc(sizeof(struct MonitorTracepointSetAnnotationsReply),
                 MONITOR_TRACEPOINT_SET_ANNOTATIONS_REPLY);

   /* Get the tp from list*/
   tp  = get_tracepoint(tpm_info, sig->set_annotations_request.tracepointId,
                        NULL);

   /* Verify that the request signal is valid */
   monitor_status = verify_set_annotations_request(sig, tp);
   if (monitor_status != MONITOR_STATUS_OK)
   {
      goto exit;
   }

   /* If there are already annotations they need to be replaced */
   if (tp->annotations != NULL)
   {
      free_buf((union SIGNAL **) &tp->annotations);
   }

   tp->annotations = (TracepointAnnotations *)
                         alloc(sizeof(TracepointAnnotations) +
                         (size_t)sig->set_annotations_request.annotations_size,
                         0);
   if (tp->annotations == NULL)
   {
      monitor_status = MONITOR_STATUS_MEMORY_EXHAUSTED;
      goto exit;
   }
   tp->annotations->no_of_annotations = sig->set_annotations_request.no_of_annotations;
   tp->annotations->annotations_size = sig->set_annotations_request.annotations_size;
   if (sig->set_annotations_request.annotations_size)
   {
      memcpy(&tp->annotations->annotations[0],
             &sig->set_annotations_request.annotations[0],
             sig->set_annotations_request.annotations_size);
   }

   exit:
   reply->set_annotations_reply.status = monitor_status;
   reply->set_annotations_reply.reserved0 = 0;
   reply->set_annotations_reply.reserved1 = 0;

   send(&reply, sender(&sig));
   return monitor_status;
}

static MonitorStatus
get_offspring(MonitorScope scope,
              TracepointManagerInfo *tpm_info,
              PROCESS *offspring_pid)
{
   union SIGNAL *request;
   union SIGNAL *reply;
   MonitorBoolean gotEndmark = False;
   SIGSELECT get_offsprings_filter[] =
   { 2,
     RMM_GET_OFFSPRINGS_REPLY, RMM_GET_OFFSPRINGS_ENDMARK };

   /* Find the offspring that are attached to correct scope */
   request = alloc(sizeof(SIGSELECT), RMM_GET_OFFSPRINGS_REQUEST);
   send(&request, tpm_info->rmm_pid);
   while (!gotEndmark)
   {
      reply = receive_from(10000, get_offsprings_filter, tpm_info->rmm_pid);
      if (reply == NIL)
      {
         ramlog_printf("rmm_tracepoint: Monitor not responding when "
                       "retrieving offspring information.\n");
         return  MONITOR_STATUS_INTERNAL_ERROR;
      }
      switch (reply->sig_no)
      {
         case RMM_GET_OFFSPRINGS_REPLY:
            if (reply->rmm_get_offsprings_reply.scope.id == scope.id
                     && reply->rmm_get_offsprings_reply.scope.type
                              == scope.type)
            {
               *offspring_pid = reply->rmm_get_offsprings_reply.offspringPid;
            }
            break;
         case RMM_GET_OFFSPRINGS_ENDMARK:
            gotEndmark = True;
            break;
         default:
            ramlog_printf("rmm_tracepoint: Got unexpected signal %d, when "
                          "retrieving offspring information \n",
                          reply->sig_no);
            assert(0);
            break;
      }
      free_buf(&reply);
   }
   return MONITOR_STATUS_OK;
}

static void *
send_tracepoint_settings(ExecutionPointTraceInfo *ept,
                         TracepointManagerInfo *tpm_info,
                         PROCESS receiver)
{
   union SIGNAL *reply;
   Tracepoint *tp = &ept->tp;

   reply = alloc(sizeof(struct MonitorTracepointGetTracepointsReply)
                          + (tp->byteCodeLength - sizeof(uint8_t)),
                 MONITOR_TRACEPOINT_GET_TRACEPOINTS_REPLY);

   reply->get_tracepoints_reply.address = tp->address;
   reply->get_tracepoints_reply.enabled = tp->enabled;
   reply->get_tracepoints_reply.id = tp->id;
   reply->get_tracepoints_reply.passCount = tp->passCount;
   reply->get_tracepoints_reply.tracepointType = tp->tracepointType;
   reply->get_tracepoints_reply.scope.id = tp->scope.id;
   reply->get_tracepoints_reply.scope.type = tp->scope.type;
   reply->get_tracepoints_reply.byteCodeLength = tp->byteCodeLength;
   reply->get_tracepoints_reply.reserved0 = 0;
   reply->get_tracepoints_reply.reserved1 = 0;
   memcpy(reply->get_tracepoints_reply.byteCode,
          tp->byteCode,
          tp->byteCodeLength);

   if (receiver == tpm_info->client_pid)
   {
      zzrmm_send2client(&reply, tpm_info->client);
   }
   else
   {
      send(&reply, receiver);
   }
   return NULL;
}

static void
send_tracepoint_replies(TracepointManagerInfo *tpm_info, PROCESS receiver)
{
   ExecutionPointTraceInfo *ept;

   /* Traverse the lists of tracepoints and send back the settings
    */
   ept = list_head(tpm_info->enabled_tracepoint_list);
   while (ept)
   {
      send_tracepoint_settings(ept, tpm_info, receiver);
      ept = list_next(tpm_info->enabled_tracepoint_list, &ept->link);
   }

   ept = list_head(tpm_info->disabled_tracepoint_list);
   while (ept)
   {
      send_tracepoint_settings(ept, tpm_info, receiver);
      ept = list_next(tpm_info->disabled_tracepoint_list, &ept->link);
   }
}

static int
handle_tracepoint_get_tracepoints_request(union SIGNAL *sig,
                                          TracepointManagerInfo *tpm_info)
{
   union SIGNAL *request;
   union SIGNAL *reply;
   union SIGNAL *endmark;
   MonitorStatus status = MONITOR_STATUS_OK;
   MonitorBoolean gotEndmark = False;
   PROCESS offspring_pid = 0;
   OSATTREF attref;
   SIGSELECT get_tps_filter[] = { 3, MONITOR_TRACEPOINT_GET_TRACEPOINTS_REPLY,
                                  MONITOR_TRACEPOINT_GET_TRACEPOINTS_ENDMARK,
                                  OS_ATTACH_SIG };

   if (sig->get_tracepoints_request.scope.id != tpm_info->scope.id
            || sig->get_tracepoints_request.scope.type != tpm_info->scope.type)
   {
      /* Find the offspring that are attached to correct scope */
      status = get_offspring(sig->get_tracepoints_request.scope, tpm_info,
                             &offspring_pid);
      if (status != MONITOR_STATUS_OK)
      {
         goto done;
      }
      if (offspring_pid == 0)
      {
         status = MONITOR_STATUS_REQUEST_INVALID;
         goto done;
      }

      /* Request the tracepoints from the offspring */
      request = alloc(sizeof(RmmGetTracepointsRequest),
                      RMM_GET_TRACEPOINTS_REQUEST);
      attref = attach(NULL, offspring_pid);
      request->rmm_get_tracepoints_request.receiver = current_process();
      send_w_s(&request, tpm_info->rmm_pid, offspring_pid);

      while (!gotEndmark)
      {
         reply = receive_from(10000, get_tps_filter, offspring_pid);
         if (reply == NIL)
         {
            status = MONITOR_STATUS_INTERNAL_ERROR;
            ramlog_printf("rmm_tracepoint: Monitor not responding when "
                                   "retrieving tracepoint information.\n");
            goto done;
         }
         switch (reply->sig_no)
         {
            case MONITOR_TRACEPOINT_GET_TRACEPOINTS_REPLY:
               zzrmm_send2client(&reply, tpm_info->client);
               break;
            case MONITOR_TRACEPOINT_GET_TRACEPOINTS_ENDMARK:
               gotEndmark = True;
               break;
            case OS_ATTACH_SIG:
               status = MONITOR_STATUS_CLIENT_KILLED;
               free_buf(&reply);
               goto done;
            default:
               ramlog_printf("rmm_tracepoint: Got unexpected signal %d, when "
                             "retrieving tracepoint information.\n",
                             reply->sig_no);
               assert(0);
               break;
         }
         if (reply != NIL)
         {
            free_buf(&reply);
         }
      }
      detach(&attref);
   }
   else
   {
      /* The requested tracepoints exist in this offspring, send replies to the
       * client.
       */
      send_tracepoint_replies(tpm_info, tpm_info->client_pid);
   }

   /* Send an endmark to give the client status and inform that all tracepoints
    * has been sent.
    */
   done:
   endmark = alloc(sizeof(MonitorTracepointGetTracepointsEndmark),
                         MONITOR_TRACEPOINT_GET_TRACEPOINTS_ENDMARK);
   endmark->get_tracepoints_endmark.status = status;
   send(&endmark, sender(&sig));
   return 0;
}


static void
handle_rmm_get_tracepoints_request(union SIGNAL *sig,
                                    TracepointManagerInfo *tpm_info)
{
   union SIGNAL *endmark;
   PROCESS receiver = sig->rmm_get_tracepoints_request.receiver;

   send_tracepoint_replies(tpm_info, receiver);
   endmark = alloc(sizeof(MonitorTracepointGetTracepointsEndmark),
                   MONITOR_TRACEPOINT_GET_TRACEPOINTS_ENDMARK);
   endmark->get_tracepoints_endmark.status = MONITOR_STATUS_OK;
   send(&endmark, receiver);
}

static void *
send_action_settings(TracepointAction *action, TracepointManagerInfo *tpm_info, PROCESS receiver)
{
   union SIGNAL *reply;
   uint32_t reply_size = sizeof(MonitorTracepointGetActionsReply);
   uint32_t action_size = sizeof(MonitorTracepointActionData);

   /* Extra memory needs to be allocated and copied,
    * if the action contain byte code.
    */
   if (action->actionType == MONITOR_TRACEPOINT_ACTION_EXPRESSION)
   {
      action_size += action->tracepointAction.expression.byteCodeLength
               - sizeof(uint8_t);
      reply_size += action->tracepointAction.expression.byteCodeLength
               - sizeof(uint8_t);
   }

   reply = alloc(reply_size, MONITOR_TRACEPOINT_GET_ACTIONS_REPLY);
   reply->get_actions_reply.actionId = action->actionId;
   reply->get_actions_reply.actionType = action->actionType;
   reply->get_actions_reply.tracepointId = action->tracepointId;
   reply->get_actions_reply.whileSteppingId = action->whileSteppingId;
   reply->get_actions_reply.reserved0 = 0;
   reply->get_actions_reply.reserved1 = 0;
   memcpy(&reply->get_actions_reply.tracepointAction,
          &action->tracepointAction,
          action_size);

   if (receiver == tpm_info->client_pid)
   {
      zzrmm_send2client(&reply, tpm_info->client);
   }
   else
   {
      send(&reply, receiver);
   }

   return NULL;
}

static MonitorStatus
send_action_replies(TracepointManagerInfo *tpm_info, MonitorTracepoint tracepoint_id, PROCESS receiver)
{
   Tracepoint *tp;

   tp = get_tracepoint(tpm_info, tracepoint_id, NULL);
   if (tp != NULL)
   {
      /*Traverse the list of actions and send back the settings,
       under blocking flow control
       */
      if (tp->action_list != NULL)
      {
         TracepointAction *action = list_head(tp->action_list);
         while (action)
         {
            send_action_settings(action, tpm_info, receiver);

            /* If the action is a while stepping action, shall also the actions
             * in the its action list be sent.
             */
            if (action->actionType == MONITOR_TRACEPOINT_ACTION_WHILE_STEPPING
                     && action->action_list != NULL)
            {
               TracepointAction *while_stepping_action =
                        list_head(action->action_list);
               while (while_stepping_action)
               {
                  send_action_settings(while_stepping_action, tpm_info, receiver);
                  while_stepping_action
                           = list_next(action->action_list, &while_stepping_action->link);
               }
            }
            action = list_next(tp->action_list, &action->link);
         }
      }
   }

   else
   {
      return MONITOR_STATUS_ID_INVALID;
   }

   return MONITOR_STATUS_OK;
}


static int
handle_tracepoint_get_actions_request(union SIGNAL *sig,
                                      TracepointManagerInfo *tpm_info)
{
   union SIGNAL *request;
   union SIGNAL *reply;
   union SIGNAL *endmark;
   MonitorStatus status = MONITOR_STATUS_OK;
   MonitorBoolean gotEndmark = False;
   PROCESS offspring_pid = 0;
   OSATTREF attref;
   SIGSELECT get_action_filter[] = { 3, MONITOR_TRACEPOINT_GET_ACTIONS_REPLY,
                                  MONITOR_TRACEPOINT_GET_ACTIONS_ENDMARK,
                                  OS_ATTACH_SIG };

   if (sig->get_actions_request.scope.id != tpm_info->scope.id
            || sig->get_actions_request.scope.type != tpm_info->scope.type)
   {
      /* Find the offspring that are attached to correct scope */
      status = get_offspring(sig->get_actions_request.scope, tpm_info,
                             &offspring_pid);
      if (status != MONITOR_STATUS_OK)
      {
         goto done;
      }
      if (offspring_pid == 0)
      {
         status = MONITOR_STATUS_REQUEST_INVALID;
         goto done;
      }

      /* Request the actions from the offspring */
      request = alloc(sizeof(RmmGetTracepointActionsRequest),
                      RMM_GET_TRACEPOINT_ACTIONS_REQUEST);
      attref = attach(NULL, offspring_pid);
      request->rmm_get_actions_request.receiver = current_process();
      request->rmm_get_actions_request.tracepoint_id = sig->get_actions_request.tracepointId;
      send_w_s(&request, tpm_info->rmm_pid, offspring_pid);

      while (!gotEndmark)
      {
         reply = receive_from(10000, get_action_filter, offspring_pid);
         if (reply == NIL)
         {
            status = MONITOR_STATUS_INTERNAL_ERROR;
            ramlog_printf("rmm_tracepoint: Monitor not responding when "
                          "retrieving tracepoint action information.\n");
            goto done;
         }
         switch (reply->sig_no)
         {
            case MONITOR_TRACEPOINT_GET_ACTIONS_REPLY:
               zzrmm_send2client(&reply, tpm_info->client);
               break;
            case MONITOR_TRACEPOINT_GET_ACTIONS_ENDMARK:
               status = reply->get_actions_endmark.status;
               gotEndmark = True;
               break;
            case OS_ATTACH_SIG:
               status = MONITOR_STATUS_CLIENT_KILLED;
               free_buf(&reply);
               goto done;
            default:
               ramlog_printf("rmm_tracepoint: Got unexpected signal %d, when"
                             "retrieving tracepoint action information.\n",
                             reply->sig_no);
               assert(0);
               break;
         }
         if (reply != NIL)
         {
            free_buf(&reply);
         }
      }
      detach(&attref);
   }
   else
   {
      /* The requested actions exist in this offspring, send replies to the
       * client.
       */
      send_action_replies(tpm_info, sig->get_actions_request.tracepointId ,tpm_info->client_pid);
   }

   /* Send an endmark to give the client status and inform that all actions
    * has been sent.
    */
   done:
   endmark = alloc(sizeof(MonitorTracepointGetActionsEndmark),
                         MONITOR_TRACEPOINT_GET_ACTIONS_ENDMARK);
   endmark->get_actions_endmark.status = status;
   send(&endmark, sender(&sig));
   return 0;
}

static void
handle_rmm_get_actions_request(union SIGNAL *sig,
                                    TracepointManagerInfo *tpm_info)
{
   union SIGNAL *endmark;
   MonitorStatus status;
   PROCESS receiver = sig->rmm_get_actions_request.receiver;
   MonitorTracepoint tracepoint_id = sig->rmm_get_actions_request.tracepoint_id;

   status = send_action_replies(tpm_info, tracepoint_id, receiver);
   endmark = alloc(sizeof(MonitorTracepointGetActionsEndmark),
                   MONITOR_TRACEPOINT_GET_ACTIONS_ENDMARK);
   endmark->get_tracepoints_endmark.status = status;
   send(&endmark, receiver);
}

static MonitorStatus
send_annotations(TracepointManagerInfo *tpm_info, MonitorTracepoint tracepoint_id, PROCESS receiver)
{
   Tracepoint *tp;
   union SIGNAL *reply;

   tp = get_tracepoint(tpm_info, tracepoint_id, NULL);
   if (tp != NULL)
   {
      if (tp->annotations != NULL)
      {
         reply = alloc(sizeof(MonitorTracepointGetAnnotationsReply) +
                       (size_t)tp->annotations->annotations_size,
                       MONITOR_TRACEPOINT_GET_ANNOTATIONS_REPLY);
         reply->get_annotations_reply.status = MONITOR_STATUS_OK;
         reply->get_annotations_reply.tracepointId = tracepoint_id;
         reply->get_annotations_reply.no_of_annotations = tp->annotations->no_of_annotations;
         reply->get_annotations_reply.annotations_size = tp->annotations->annotations_size;
         reply->get_annotations_reply.reserved0 = 0;
         reply->get_annotations_reply.reserved1 = 0;
         if (tp->annotations->annotations_size)
         {
            memcpy(&reply->get_annotations_reply.annotations[0],
                   &tp->annotations->annotations[0],
                   (size_t)tp->annotations->annotations_size);
         }
      }
      /* There are no annotations */
      else
      {
         reply = alloc(sizeof(MonitorTracepointGetAnnotationsReply),
                       MONITOR_TRACEPOINT_GET_ANNOTATIONS_REPLY);
         reply->get_annotations_reply.status = MONITOR_STATUS_OK;
         reply->get_annotations_reply.tracepointId = tracepoint_id;
         reply->get_annotations_reply.no_of_annotations = 0;
         reply->get_annotations_reply.annotations_size = 0;
         reply->get_annotations_reply.reserved0 = 0;
         reply->get_annotations_reply.reserved1 = 0;
      }

      send(&reply, receiver);
   }
   else
   {
      return MONITOR_STATUS_ID_INVALID;
   }

   return MONITOR_STATUS_OK;
}

static int
handle_tracepoint_get_annotations_request(union SIGNAL *sig,
                                          TracepointManagerInfo *tpm_info)
{
   union SIGNAL *request;
   union SIGNAL *reply;
   union SIGNAL *reply_error;
   MonitorStatus status = MONITOR_STATUS_OK;
   PROCESS offspring_pid = 0;
   OSATTREF attref;
   SIGSELECT get_annotations_filter[] = { 2,
                                     MONITOR_TRACEPOINT_GET_ANNOTATIONS_REPLY,
                                     OS_ATTACH_SIG };

   if (sig->get_annotations_request.scope.id != tpm_info->scope.id ||
       sig->get_annotations_request.scope.type != tpm_info->scope.type)
   {
      /* Find the offspring that is attached to correct scope */
      status = get_offspring(sig->get_annotations_request.scope, tpm_info,
                            &offspring_pid);
      if (status != MONITOR_STATUS_OK)
      {
         goto done;
      }
      if (offspring_pid == 0)
      {
         status = MONITOR_STATUS_REQUEST_INVALID;
         goto done;
      }

      /* Request the annotations from the offspring */
      request = alloc(sizeof(RmmGetTracepointAnnotationsRequest),
                      RMM_GET_TRACEPOINT_ANNOTATIONS_REQUEST);
      attref = attach(NULL, offspring_pid);
      request->rmm_get_annotations_request.receiver = current_process();
      request->rmm_get_annotations_request.tracepoint_id = sig->get_annotations_request.tracepointId;
      send_w_s(&request, tpm_info->rmm_pid, offspring_pid);

      reply = receive_from(10000, get_annotations_filter, offspring_pid);
      if (reply == NIL)
      {
         status = MONITOR_STATUS_INTERNAL_ERROR;
         ramlog_printf("rmm_tracepoint: Monitor not responding when "
                       "retrieving tracepoint annotations information.\n");
         goto done;
      }
      switch (reply->sig_no)
      {
         case MONITOR_TRACEPOINT_GET_ANNOTATIONS_REPLY:
            send(&reply, tpm_info->client_pid);
            break;
         case OS_ATTACH_SIG:
            status = MONITOR_STATUS_CLIENT_KILLED;
            free_buf(&reply);
            goto done;
         default:
            ramlog_printf("rmm_tracepoint: Got unexpected signal %d, when"
                          "retrieving tracepoint annotations information.\n",
                          reply->sig_no);
            assert(0);
            break;
      }
      if (reply != NIL)
      {
         free_buf(&reply);
      }
      detach(&attref);
   }
   else
   {
      /* The requested annotations exist in this offspring, send the reply
       * to the client.
       */
      status = send_annotations(tpm_info, sig->get_annotations_request.tracepointId, tpm_info->client_pid);
   }

   done:
   /* If thre is a fail send a dummy signal to the sender */
   if (status != MONITOR_STATUS_OK)
   {
      reply_error = alloc(sizeof(MonitorTracepointGetAnnotationsReply),
                          MONITOR_TRACEPOINT_GET_ANNOTATIONS_REPLY);
      reply_error->get_annotations_reply.status = status;
      reply_error->get_annotations_reply.no_of_annotations = 0;
      reply_error->get_annotations_reply.annotations_size = 0;
      reply_error->get_annotations_reply.reserved0 = 0;
      reply_error->get_annotations_reply.reserved1 = 0;
      send(&reply_error, sender(&sig));
   }
   return 0;
}

static void
handle_rmm_get_annotations_request(union SIGNAL *sig,
                                   TracepointManagerInfo *tpm_info)
{
   PROCESS receiver = sig->rmm_get_annotations_request.receiver;
   MonitorTracepoint tracepoint_id = sig->rmm_get_annotations_request.tracepoint_id;

   send_annotations(tpm_info, tracepoint_id, receiver);
}


static void *
send_variables(TracepointVariable *variable, TracepointManagerInfo *tpm_info, PROCESS receiver)
{
   union SIGNAL *reply;

   reply = alloc(sizeof(struct MonitorTracepointGetVariablesReply),
                 MONITOR_TRACEPOINT_GET_VARIABLES_REPLY);
   reply->get_variables_reply.id = variable->id;
   reply->get_variables_reply.initialValue = variable->initial_value;
   reply->get_variables_reply.reserved0 = 0;
   reply->get_variables_reply.reserved1 = 0;
   if (tpm_info->state != INIT && tpm_info->state != CONFIGURED)
   {
      reply->get_variables_reply.variableStatus = 0;
   }
   else
   {
      reply->get_variables_reply.variableStatus = 1;
   }

   if (receiver == tpm_info->client_pid)
   {
      zzrmm_send2client(&reply, tpm_info->client);
   }
   else
   {
      send(&reply, receiver);
   }

   return NULL;
}


static void
send_variable_replies(TracepointManagerInfo *tpm_info, PROCESS receiver)
{
   TracepointVariable *variable;

   /* Traverse the lists of variables and send the settings to the receiver */
   variable = list_head(tpm_info->variables_list);
   while (variable)
   {
      send_variables(variable, tpm_info, receiver);
      variable = list_next(tpm_info->variables_list, &variable->link);
   }
}

static int
handle_tracepoint_get_variables_request(union SIGNAL *sig,
                                        TracepointManagerInfo *tpm_info)
{
   union SIGNAL *request;
   union SIGNAL *reply;
   union SIGNAL *endmark;
   MonitorStatus status = MONITOR_STATUS_OK;
   MonitorBoolean gotEndmark = False;
   PROCESS offspring_pid = 0;
   OSATTREF attref;
   SIGSELECT get_variable_filter[] = { 3, MONITOR_TRACEPOINT_GET_VARIABLES_REPLY,
                                  MONITOR_TRACEPOINT_GET_VARIABLES_ENDMARK,
                                  OS_ATTACH_SIG };

   if (sig->get_variables_request.scope.id != tpm_info->scope.id
            || sig->get_variables_request.scope.type != tpm_info->scope.type)
   {
      /* Find the offspring that are attached to correct scope */
      status = get_offspring(sig->get_variables_request.scope, tpm_info,
                             &offspring_pid);
      if (status != MONITOR_STATUS_OK)
      {
         goto done;
      }
      if (offspring_pid == 0)
      {
         status = MONITOR_STATUS_REQUEST_INVALID;
         goto done;
      }

      /* Request the variables from the offspring */
      request = alloc(sizeof(RmmGetTracepointVariablesRequest),
                      RMM_GET_TRACEPOINT_VARIABLES_REQUEST);
      attref = attach(NULL, offspring_pid);
      request->rmm_get_variables_request.receiver = current_process();
      send_w_s(&request, tpm_info->rmm_pid, offspring_pid);

      while (!gotEndmark)
      {
         reply = receive_from(10000, get_variable_filter, offspring_pid);
         if (reply == NIL)
         {
            status = MONITOR_STATUS_INTERNAL_ERROR;
            ramlog_printf("rmm_tracepoint: Monitor not responding when "
                          "retrieving tracepoint variable information.\n");
            goto done;
         }
         switch (reply->sig_no)
         {
            case MONITOR_TRACEPOINT_GET_VARIABLES_REPLY:
               zzrmm_send2client(&reply, tpm_info->client);
               break;
            case MONITOR_TRACEPOINT_GET_VARIABLES_ENDMARK:
               gotEndmark = True;
               break;
            case OS_ATTACH_SIG:
               status = MONITOR_STATUS_CLIENT_KILLED;
               free_buf(&reply);
               goto done;
            default:
               ramlog_printf("rmm_tracepoint: Got unexpected signal %d, when"
                             "retrieving tracepoint variable information.\n",
                             reply->sig_no);
               assert(0);
               break;
         }
         if (reply != NIL)
         {
            free_buf(&reply);
         }
      }
      detach(&attref);
   }
   else
   {
      /* The requested variables exist in this offspring, send replies to the
       * client.
       */
      send_variable_replies(tpm_info, tpm_info->client_pid);
   }

   /* Send an endmark to give the client status and inform that all tracepoints
    * has been sent.
    */
   done:
   endmark = alloc(sizeof(MonitorTracepointGetVariablesEndmark),
                         MONITOR_TRACEPOINT_GET_VARIABLES_ENDMARK);
   endmark->get_tracepoints_endmark.status = status;
   send(&endmark, sender(&sig));
   return 0;
}

static void
handle_rmm_get_variables_request(union SIGNAL *sig,
                                    TracepointManagerInfo *tpm_info)
{
   union SIGNAL *endmark;
   PROCESS receiver = sig->rmm_get_variables_request.receiver;

   send_variable_replies(tpm_info, receiver);
   endmark = alloc(sizeof(MonitorTracepointGetVariablesEndmark),
                   MONITOR_TRACEPOINT_GET_VARIABLES_ENDMARK);
   endmark->get_variables_endmark.status = MONITOR_STATUS_OK;
   send(&endmark, receiver);
}

static int
handle_tracepoint_create_variable_request(union SIGNAL *sig,
                                          TracepointManagerInfo *tpm_info)
{
   union SIGNAL *reply;
   TracepointVariable *variable;

   /* Create variable */
   variable = list_init_link(alloc(sizeof(TracepointVariable), 0));
   variable->initial_value = sig->create_variable_request.value;
   variable->value = sig->create_variable_request.value;
   variable->id = tpm_info->variable_id_counter++;
   variable->client_side_id = sig->create_variable_request.reserved0;

   /* Add variable to list */
   list_insert_tail(tpm_info->variables_list, variable);

   if (variable->client_side_id)
   {
      /* inform TracepointManager to convert ids inside the bytecode */
      tpm_info->convert_bytecode_variable_ids = True; 
   }

   reply = alloc(sizeof(struct MonitorTracepointCreateVariableReply),
                 MONITOR_TRACEPOINT_CREATE_VARIABLE_REPLY);
   reply->create_variable_reply.status = MONITOR_STATUS_OK;
   reply->create_variable_reply.id = variable->id;
   reply->create_variable_reply.reserved0 = 0;
   reply->create_variable_reply.reserved1 = 0;
   send(&reply, sender(&sig));
   return 0;
}

static int
handle_tracepoint_get_variable_value_request(union SIGNAL *sig,
                                             TracepointManagerInfo *tpm_info)
{
   union SIGNAL *reply;
   MonitorTracepointGetVariableValueReply *ret_var;
   TracepointVariable *variable;
   MonitorStatus monitor_status = MONITOR_STATUS_OK;

   reply = alloc(sizeof(struct MonitorTracepointGetVariableValueReply),
                 MONITOR_TRACEPOINT_GET_VARIABLE_VALUE_REPLY);
   ret_var = &reply->get_variable_value_reply;

   if (tpm_info->state != INIT && tpm_info->state != CONFIGURED)
   {
      ret_var->variableStatus = 0;
   }
   else
   {
      ret_var->variableStatus = 1;
   }

   variable = get_variable(tpm_info, sig->get_variable_value_request.id);
   if (variable)
   {
      ret_var->value
               = tpm_info->variable_table ? tpm_info->variable_table[variable->id]
                                          : variable->value;
      ret_var->initialValue = variable->initial_value;
   }
   else
   {
      monitor_status = MONITOR_STATUS_ID_INVALID;
   }

   ret_var->reserved0 = 0;
   ret_var->reserved1 = 0;
   ret_var->status = monitor_status;
   send(&reply, sender(&sig));
   return monitor_status;
}

static int
handle_tracepoint_remove_variable_request(union SIGNAL *sig,
                                          TracepointManagerInfo *tpm_info)
{
   union SIGNAL *reply;
   TracepointVariable *variable;
   MonitorStatus monitor_status = MONITOR_STATUS_OK;

   variable = get_variable(tpm_info, sig->remove_variable_request.id);
   if (variable)
   {
      list_unlink(variable);
      free_buf((union SIGNAL **) &variable);
      variable = NULL;
   }
   else
   {
      monitor_status = MONITOR_STATUS_ID_INVALID;
   }

   reply = alloc(sizeof(struct MonitorTracepointRemoveVariableReply),
                 MONITOR_TRACEPOINT_REMOVE_VARIABLE_REPLY);
   reply->remove_variable_reply.status = monitor_status;
   reply->remove_variable_reply.reserved0 = 0;
   reply->remove_variable_reply.reserved1 = 0;
   send(&reply, sender(&sig));
   return monitor_status;
}

static int
check_input_configure(union SIGNAL *sig)
{
   MonitorTracepointConfigureTracebufferRequest *s;
   s = &sig->configure_request;

   /* TODO update when fast tp is supported*/
   VERIFY_RANGE(s->bufferType,
            MONITOR_TRACEPOINT_BUFFER_LINEAR,
            MONITOR_TRACEPOINT_BUFFER_CIRCULAR);

   VERIFY_BUFFER_SIZE(s->bufferSize);

   return MONITOR_STATUS_OK;
}


static int
handle_tracepoint_configure_request(union SIGNAL *sig,
                                    TracepointManagerInfo *tpm_info)
{
   uint32_t actual_size = 0;
   union SIGNAL *reply;
   MonitorStatus monitor_status;
   long conf_buffer_size;

   monitor_status = check_input_configure(sig);
   if (monitor_status != MONITOR_STATUS_OK)
   {
      goto exit;
   }

   /* Determine which buffer_size to use */
   if (tpm_info->state != RUNNING)
   {
      uint32_t buffer_size;

      if (sysparam_readattr_long(buffer_sysparam, "core_buffer_size", &conf_buffer_size))
      {
         buffer_size = DEFAULT_TRACE_BUFFER_SIZE * ose_num_cpus();
      }
      else
      {
         buffer_size = conf_buffer_size * ose_num_cpus();
      }
       
      if (sig->configure_request.bufferSize != 0)
      {
         buffer_size = (uint32_t)(sig->configure_request.bufferSize);
      }

      /* Check buffer status */
      if (tpm_info->trace_buffer_handle &&
          tpm_info->configured_size != buffer_size)
      {
         /* New size of buffer, delete old one */
         (void) tracepoint_buffer_destroy(tpm_info->trace_buffer_handle);
         delete_buffer_region((OSADDRESS) tpm_info->buffer);
         tpm_info->trace_buffer_handle = NULL;
         tpm_info->buffer = NULL;
      }

      if (!tpm_info->buffer)
      {
         /* Buffer region is not created, so create it and init buffer manager */
         tpm_info->buffer = create_buffer_region(buffer_size, &actual_size);
         if (!tpm_info->buffer)
         {
            ramlog_printf("rmm_tracepoint: unable to create region!\n");
            monitor_status = MONITOR_STATUS_MEMORY_EXHAUSTED;
            tpm_info->status = TRACEPOINT_ERROR;
            goto exit;
         }
         tpm_info->configured_size = buffer_size;
         tpm_info->trace_buffer_handle
                  = tracepoint_buffer_init(tpm_info->buffer, actual_size);
      }
      else
      {
         /* Old buffer can be reused, only clear it */
         tracepoint_buffer_clear(tpm_info->trace_buffer_handle);
      }
   }
   tracepoint_buffer_config(tpm_info->trace_buffer_handle,
                            TRACEBUFFER_CONFIG_BUFFER_TYPE,
                            sig->configure_request.bufferType);

   if (tpm_info->state == INIT)
   {
      tpm_info->status = TRACEPOINT_NOTRUN;
   }
   exit:

   reply = alloc(sizeof(struct MonitorTracepointConfigureTracebufferReply),
                 MONITOR_TRACEPOINT_CONFIGURE_TRACEBUFFER_REPLY);
   reply->configure_reply.status = monitor_status;
   reply->configure_reply.reserved0 = 0;
   reply->configure_reply.reserved1 = 0;
   send(&reply, sender(&sig));
   return monitor_status;
}

static void
remove_eps(TracepointManagerInfo *tpm_info)
{
   ExecutionPointTraceInfo *ept;
   EpmStatus epm_status;

   ept = list_head(tpm_info->enabled_tracepoint_list);
   while (ept)
   {
#ifdef PPC
      if(ept->tp.tracepointType == MONITOR_TRACEPOINT_FAST_TRACEPOINT)
      {
        epm_status = ose_epm_disable_fast_execution_point(&ept->ep); 
      }
      else
#endif
      {
      epm_status = ose_epm_remove_execution_point(&ept->ep);
      }
      if (epm_status != EPM_OK)
      {
         ramlog_printf("rmm_tracepoint: error received while removing EP: %d\n",
                       epm_status);
         tpm_info->status = TRACEPOINT_ERROR;
      }
      ept = list_next(tpm_info->enabled_tracepoint_list, &ept->link);
   }
}

#ifdef PPC
/* The function remove_eps needs to have been run before remove_fast_tracepoint */
/* Make sure that the function wait_on all cores has been run */
static void
remove_fast_tracepoints(TracepointManagerInfo *tpm_info)
{

   ExecutionPointTraceInfo *ept;
   
   /* Handle tracepoints in enable list */
   ept = list_head(tpm_info->enabled_tracepoint_list);
   while (ept)
   {
      if(ept->tp.tracepointType == MONITOR_TRACEPOINT_FAST_TRACEPOINT)
      {
         ose_epm_free_fast_execution_point(&ept->ep);
      }
      ept = list_next(tpm_info->enabled_tracepoint_list, &ept->link);
   }
   
   /* Handle tracepoints in disable list */
   ept = list_head(tpm_info->disabled_tracepoint_list);
   while (ept)
   {
      if(ept->tp.tracepointType == MONITOR_TRACEPOINT_FAST_TRACEPOINT)
      {
         ose_epm_free_fast_execution_point(&ept->ep);
      }
      ept = list_next(tpm_info->disabled_tracepoint_list, &ept->link);
   }
   
}
#endif


static int
handle_tracepoint_start_request(union SIGNAL *sig,
                                TracepointManagerInfo *tpm_info)
{
   ExecutionPointTraceInfo *ept;
   EpmStatus epm_status;
   union SIGNAL *reply;
   MonitorStatus monitor_status;

   /* Verify that the trace exeriment can be started */
   monitor_status = check_preconditions_start(tpm_info);

   if (monitor_status == MONITOR_STATUS_OK)
   {
      TracepointVariable *tvar;
      size_t var_table_size = sizeof(int64_t) * tpm_info->variable_id_counter;

#if defined(softOSE) || defined(__i386)

#else
      /* If needed, update the IDs inside the bytecode */
      if (tpm_info->convert_bytecode_variable_ids)
      {
         update_bytecode_tvar_ids(tpm_info);
      }
#endif

      tpm_info->variable_table = (int64_t *) alloc(var_table_size, 0);
      memset(tpm_info->variable_table, 0, var_table_size);

      tvar = list_head(tpm_info->variables_list);
      while (tvar)
      {
         tpm_info->variable_table[tvar->id] = tvar->initial_value;
         tvar = list_next(tpm_info->variables_list, tvar);
      }

      /* Add all EPs to EPM */
      ept = list_head(tpm_info->enabled_tracepoint_list);
      while (ept)
      {
         atomic_set(&ept->tp.passCountCurr, ept->tp.passCount);
#ifdef PPC
         if(ept->tp.tracepointType == MONITOR_TRACEPOINT_FAST_TRACEPOINT)
         {
            epm_status = ose_epm_enable_fast_execution_point(&ept->ep); 
         }
         else
#endif
         {
         epm_status = ose_epm_add_execution_point(&ept->ep);
         }
         if (epm_status != EPM_OK)
         {
            ramlog_printf("rmm_tracepoint: error received while adding EP: %d\n",
                          epm_status);
            /* Remove already added eps. Will probably generate an epm error,
             * but since start has failed that doesn't matter */
            remove_eps(tpm_info);
            monitor_status = MONITOR_STATUS_EPM_ERROR;
            tpm_info->status = TRACEPOINT_ERROR;
            break;
         }
         ept = list_next(tpm_info->enabled_tracepoint_list, ept);
      }

      if (monitor_status == MONITOR_STATUS_OK)
      {
         tpm_info->status = TRACEPOINT_RUNNING;
      }
   }

   reply = alloc(sizeof(struct MonitorTracepointStartReply),
                 MONITOR_TRACEPOINT_START_REPLY);
   reply->start_reply.status = monitor_status;
   reply->start_reply.reserved0 = 0;
   reply->start_reply.reserved1 = 0;
   send(&reply, sender(&sig));
   return monitor_status;
}

static int
handle_tracepoint_stop_request(union SIGNAL *sig,
                               TracepointManagerInfo *tpm_info)
{
   ExecutionPointTraceInfo *ept;
   EpmStatus epm_status;
   union SIGNAL *reply;
   TracepointVariable *tvar;
   MonitorStatus monitor_status = MONITOR_STATUS_OK;
   int done = 0;
   cpuid_t cpu;

   if (tpm_info->status == TRACEPOINT_RUNNING)
   {
      tpm_info->status = TRACEPOINT_STOP;
   }

   /* Remove all EPs from EPM */
   ept = list_head(tpm_info->enabled_tracepoint_list);
   while (ept)
   {
#ifdef PPC
     if(ept->tp.tracepointType == MONITOR_TRACEPOINT_FAST_TRACEPOINT)
     {
       epm_status = ose_epm_disable_fast_execution_point(&ept->ep); 
     }
     else
#endif
      {
      epm_status = ose_epm_remove_execution_point(&ept->ep);
      }
      if (epm_status != EPM_OK)
      {
         ramlog_printf("rmm_tracepoint: error received while removing EP: %d\n",
                       epm_status);
         monitor_status = MONITOR_STATUS_EPM_ERROR;
         tpm_info->status = TRACEPOINT_ERROR;
      }

      ept = list_next(tpm_info->enabled_tracepoint_list, ept);
   }
   /* Copy back the tvar to the linked list. and delete the table */
   tvar = list_head(tpm_info->variables_list);
   while (tvar)
   {
      tvar->value = tpm_info->variable_table[tvar->id];
      tvar = list_next(tpm_info->variables_list, tvar);
   }

   /* Wait for possible running callbacks to complete before free to be sure*/
   do
   {
      done = 0;
      for (cpu = 0; cpu < ose_num_cpus(); cpu++)
      {
         if (tpm_info->active_callbacks[cpu] != 0)
            done++;
      }
      if (done)
         delay(1);
    }
    while (done);

   free_buf((union SIGNAL **) &tpm_info->variable_table);
   tpm_info->variable_table = NULL;

   reply = alloc(sizeof(struct MonitorTracepointStopReply),
                 MONITOR_TRACEPOINT_STOP_REPLY);
   reply->stop_reply.status = monitor_status;
   reply->stop_reply.reserved0 = 0;
   reply->stop_reply.reserved1 = 0;
   send(&reply, sender(&sig));
   return monitor_status;
}

static int
handle_tracepoint_clear_buffer_request(union SIGNAL *sig,
                                       TracepointManagerInfo *tpm_info)
{
   union SIGNAL *reply;
   reply = alloc(sizeof(struct MonitorTracepointClearBufferReply),
                 MONITOR_TRACEPOINT_CLEAR_BUFFER_REPLY);

   tracepoint_buffer_clear(tpm_info->trace_buffer_handle);
   tpm_info->status = TRACEPOINT_NOTRUN;
   reply->clear_buffer_reply.status = MONITOR_STATUS_OK;
   reply->clear_buffer_reply.reserved0 = 0;
   reply->clear_buffer_reply.reserved1 = 0;

   send(&reply, sender(&sig));
   return 0;
}

static int
handle_tracepoint_status_request(union SIGNAL *sig,
                                 TracepointManagerInfo *tpm_info)
{
   union SIGNAL *reply;
   cpuid_t i;
   cpuid_t cpus = ose_num_cpus();

   reply = alloc(sizeof(MonitorTracepointGetStatusReply) +
                 (cpus - 1) * sizeof(MonitorTracepointBufferStatus),
                 MONITOR_TRACEPOINT_GET_STATUS_REPLY);

   reply->get_status_reply.traceStatus = tpm_info->status;
   reply->get_status_reply.cores = cpus;

   if (tpm_info->trace_buffer_handle)
   {
      MonitorTracepointBufferStatus *status_table =  reply->get_status_reply.perCoreStatus;
      TraceBuffer *trace = tpm_info->trace_buffer_handle;
      for (i = 0; i < cpus; i++)
      {
         status_table[i].frames = trace[i].frames;
         status_table[i].created = trace[i].created;
         status_table[i].size = (uint64_t)trace[i].size;
         status_table[i].free = trace[i].freespace;
         status_table[i].bufferType = trace[i].buffer_type;
      }
   }
   else
   {
      MonitorTracepointBufferStatus *table =  reply->get_status_reply.perCoreStatus;
      for (i = 0; i < cpus; i++)
      {
         table[i].frames = 0;
         table[i].created = 0;
         table[i].size = 0;
         table[i].free = 0;
         table[i].bufferType = MONITOR_TRACEPOINT_BUFFER_NONE;
      }
   }

   reply->get_status_reply.status = MONITOR_STATUS_OK;
   reply->get_status_reply.reserved0 = 0;
   reply->get_status_reply.reserved1 = 0;
   send(&reply, sender(&sig));
   return 0;
}

static void
send_tracebuffer(MonitorTracepointGetTracebufferReply *tp_data,
                 TracepointManagerInfo *tpm_info,
                 List *action_list,
                 OSADDRESS max_sig_size)
{
   union SIGNAL *reply;
   TraceData *trace_data;
   uint8_t *data;
   size_t wanted_signal_size = sizeof(MonitorTracepointGetTracebufferReply)
            + tp_data->dataLength;
   if (wanted_signal_size > max_sig_size)
   {
      /*
       * Cannot fit trace data in one signal.
       * Copy all info except trace data and set error code
       * */
      reply = alloc(sizeof(MonitorTracepointGetTracebufferReply),
                    MONITOR_TRACEPOINT_GET_TRACEBUFFER_REPLY);
      memcpy(reply, tp_data, sizeof(MonitorTracepointGetTracebufferReply));
      reply->get_tracebuffer_reply.errorCode = TRACEPOINT_ERROR_TRACE_TOO_LARGE;
      reply->get_tracebuffer_reply.eCodeExtra = wanted_signal_size;
      reply->get_tracebuffer_reply.dataLength = 0;

      /* We still have to loop through and free all trace data */
      trace_data = list_head(action_list);
      while (trace_data)
      {
         if(trace_data->data != NULL)
         {
            /*Free data allocated in append_tracebuffer_data*/
            free_buf((union SIGNAL **) &trace_data->data);
            trace_data->data = NULL;
         }
         trace_data = list_next(action_list, trace_data);
      }
   }
   else
   {
      reply = alloc(sizeof(MonitorTracepointGetTracebufferReply)
               + tp_data->dataLength, MONITOR_TRACEPOINT_GET_TRACEBUFFER_REPLY);
      memcpy(reply, tp_data, sizeof(MonitorTracepointGetTracebufferReply));

      data = &reply->get_tracebuffer_reply.traceData[0];
      trace_data = list_head(action_list);
      while (trace_data)
      {
         if(trace_data->data != NULL)
         {
            memcpy(data, trace_data->data, trace_data->length);
            data += trace_data->length;

            /*Free data allocated in append_tracebuffer_data*/
            free_buf((union SIGNAL **) &trace_data->data);
            trace_data->data = NULL;
         }
         trace_data = list_next(action_list, trace_data);
      }
   }
   zzrmm_send2client(&reply, tpm_info->client);
}

#if defined(softOSE) || defined(__i386)
static uint32_t
append_tracebuffer_data(Collect *collect, uint8_t **data)
{
   /*Dummy*/
   return 0;
}
#else
static uint32_t
append_tracebuffer_data(Collect *collect, uint8_t **data)
{
   uint32_t data_size = 1; /*1 byte R, M, V or E */
   uint8_t *buff;
   switch (collect->type)
   {
      case COLLECT_REGISTER:
         data_size += sizeof(struct cpu_core);
         buff = (uint8_t *) alloc(data_size, 0);

         *buff = 'R';
         memcpy(&buff[1], &collect->u, sizeof(struct cpu_core));
         break;
      case COLLECT_MEMORY:
         data_size += save_mem_header + collect->u.mem.length;
         buff = (uint8_t *) alloc(data_size, 0);

         *buff = 'M';
         memcpy(&buff[1], &collect->u.mem.address, sizeof(uint64_t));
         memcpy(&buff[1 + sizeof(uint64_t)],
                &collect->u.mem.length,
                sizeof(uint16_t));
         memcpy(&buff[1 + sizeof(uint64_t) + sizeof(uint16_t)],
                collect->u.mem.data,
                collect->u.mem.length);
         break;
      case COLLECT_VARIABLE:
         data_size += sizeof(uint32_t) + sizeof(int64_t);
         buff = (uint8_t *) alloc(data_size, 0);

         *buff = 'V';
         memcpy(&buff[1], &collect->u.var.id, sizeof(uint32_t));
         memcpy(&buff[1 + sizeof(uint32_t)],
                &collect->u.var.value,
                sizeof(int64_t));
         break;
      case COLLECT_ERROR:
         data_size += 2 * sizeof(uint32_t);
         buff = (uint8_t *) alloc(data_size, 0);

         *buff = 'E';
         memcpy(&buff[1], &collect->u.error.errorcode, sizeof(uint32_t));
         memcpy(&buff[1 + sizeof(uint32_t)],
                &collect->u.error.extra,
                sizeof(uint32_t));
         break;
      default:
         data_size = 0;
         buff = NULL;
         break;
   }

   *data = buff;
   return data_size;
}
#endif


static MonitorStatus
send_tracebuffer_replies(TracepointManagerInfo *tpm_info)
{
   cpuid_t cpu;
   TraceEntry *entry = NULL;
   TraceEntry ** list;
   TraceData *trace_data;
   List *action_list = NULL;
   uint8_t *data = 0;
   uint32_t extra_size = 0;
   TraceBuffer *trace_bh;
   cpuid_t cpus;
   OSADDRESS max_sig_size;
   MonitorTracepointGetTracebufferReply *tracepoint_data = NULL;
   uint64_t min_timeStamp = 0xffffffffffffffffULL; /* As high as possible */
   uint64_t cur_timeStamp;
   uint32_t active_core = 0;
   struct OS_pcb *pcbbuf;
   cpus = ose_num_cpus();
   pcbbuf = get_pcb(current_process());
   list = (TraceEntry **)alloc(sizeof(TraceEntry *)* cpus, 0);

   if (pcbbuf)
   {
      switch ((enum PROCESS_TYPE) pcbbuf->type)
      {
         case OS_PRI_PROC:
            max_sig_size = pcbbuf->max_sigsize;
            break;

         default:
            return MONITOR_STATUS_INTERNAL_ERROR;
      }
      free_buf((union SIGNAL **) &pcbbuf);
   }
   else
   {
      return MONITOR_STATUS_INTERNAL_ERROR;
   }

   /* Update array and find first entry */ 
   for ( cpu = 0;cpu < cpus; cpu++)
   {
      trace_bh = &tpm_info->trace_buffer_handle[cpu];
      list[cpu] = list_tail(trace_bh->dlink);
      if(list[cpu])
      {
         cur_timeStamp = list[cpu]->u.tp.timestamp_hi;
         cur_timeStamp <<= 32;
         cur_timeStamp += list[cpu]->u.tp.timestamp_lo;
         if(cur_timeStamp < min_timeStamp)
         {
            min_timeStamp = cur_timeStamp;
            active_core = cpu;
            entry = list[cpu];

         }
      }
   }

   while (entry)
   {
      if (entry->type == TP /*TRACEPOINT_TYPE*/)
      {
         if (!tracepoint_data)
         {
            tracepoint_data = (MonitorTracepointGetTracebufferReply *)
               alloc(sizeof(MonitorTracepointGetTracebufferReply),
                     MONITOR_TRACEPOINT_GET_TRACEBUFFER_REPLY);
         }
         else
         {
            tracepoint_data->sigLeft = 0; /*TODO handle this if the signal get too big!*/
            tracepoint_data->dataLength = extra_size;
            send_tracebuffer(tracepoint_data, tpm_info, action_list, max_sig_size);

            /*Free resources*/
            destroy_list(action_list);
            
            /* Move the entry to the list */
            list[active_core] = entry; /* Move the newly found entry to the list */
            min_timeStamp = 0xffffffffffffffffULL; /* As high as possible */
            for ( cpu = 0;cpu < cpus; cpu++)
            {
               if(list[cpu])
               {
                  cur_timeStamp = list[cpu]->u.tp.timestamp_hi;
                  cur_timeStamp <<= 32;
                  cur_timeStamp += list[cpu]->u.tp.timestamp_lo;
                  if(cur_timeStamp < min_timeStamp)
                  {
                     min_timeStamp = cur_timeStamp;
                     active_core = cpu;
                     entry = list[cpu];

                  }
               }
            }
            /* Find next Entry atleast one left */
         }

         /*Copy tracepoint data from entry*/
         tracepoint_data->snapshotId = entry->frameid;
         tracepoint_data->pid = entry->u.tp.process;
         tracepoint_data->core = active_core;
         tracepoint_data->timeStamp = entry->u.tp.timestamp_hi;
         tracepoint_data->timeStamp <<= 32;
         tracepoint_data->timeStamp += entry->u.tp.timestamp_lo;
         tracepoint_data->tracepointId = entry->u.tp.tpid;
         tracepoint_data->errorCode = entry->u.tp.error.errorcode;
         tracepoint_data->eCodeExtra = entry->u.tp.error.extra;
         tracepoint_data->dataLength = 0; /*Set later*/
         extra_size = 0;

         /*Reset action_list for new tracepoint*/
         action_list = list_init((List *) alloc(sizeof(List), 0));
      }
      else if (entry->type == COLLECT/*ACTION_TYPE*/)
      {
         if (action_list)
         {
            trace_data = list_init_link(alloc(sizeof(TraceData), 0));
            trace_data->length = append_tracebuffer_data(&entry->u.collect, &data);
            trace_data->data = data;
            extra_size += trace_data->length;
            list_insert_tail(action_list, trace_data);
         }
      }
      trace_bh = &tpm_info->trace_buffer_handle[active_core];
      entry = list_prev(trace_bh->dlink, entry);
      if(entry == NULL) /* No more entries on this core */
      {
         if (tracepoint_data)
         {
            tracepoint_data->sigLeft = 0; /*TODO handle this if the signal get too big!*/
            tracepoint_data->dataLength = extra_size;
            send_tracebuffer(tracepoint_data, tpm_info, action_list, max_sig_size);

            /*Free all resources*/
            free_buf((union SIGNAL **) &tracepoint_data); /* Unnessesary ??*/
            tracepoint_data = NULL;
            destroy_list(action_list);
         }
         list[active_core] = NULL;
         min_timeStamp = 0xffffffffffffffffLL; /* As high as possible */
         /* Find next entry */
         for ( cpu = 0;cpu < cpus; cpu++)
         {
            if(list[cpu])
            {
               cur_timeStamp = list[cpu]->u.tp.timestamp_hi;
               cur_timeStamp <<= 32;
               cur_timeStamp += list[cpu]->u.tp.timestamp_lo;
               if(cur_timeStamp < min_timeStamp)
               {
                  min_timeStamp = cur_timeStamp;
                  active_core = cpu;
                  entry = list[cpu];

               }
            }
         }
         /* Entry may be NULL, this will terminate the while loop */ 

         
      }
   }

   /*Send last tracepoint if applicable*/
   free_buf((union SIGNAL **) &list); /* Unnessesary ??*/
   return MONITOR_STATUS_OK;
}

static int
handle_tracepoint_tracebuffer_request(union SIGNAL *sig,
                                      TracepointManagerInfo *tpm_info)
{
   union SIGNAL *endmark;
   MonitorStatus monitor_status = MONITOR_STATUS_OK;
   TOUCH(sig);

   /* Create tracebuffer replies to be sent to client */
   if (tpm_info->trace_buffer_handle->sorted == False)
   {
      tracepoint_buffer_sort(tpm_info->trace_buffer_handle);
      tpm_info->trace_buffer_handle->sorted = True;
   }
   monitor_status = send_tracebuffer_replies(tpm_info);

   /* Send an endmark to notify the client that all traces
    * are sent */
   endmark = alloc(sizeof(MonitorTracepointGetTracebufferEndmark),
                   MONITOR_TRACEPOINT_GET_TRACEBUFFER_ENDMARK);
   endmark->get_tracebuffer_endmark.status = monitor_status;
   send(&endmark, tpm_info->client_pid);
   return monitor_status;
}

enum action {
   CREATE_TRACEPOINT,
   REMOVE_TRACEPOINT,
   CONFIGURE_TRACEPOINT,
   CREATE_ACTION,
   SET_ANNOTATIONS,
   CREATE_VARIABLE,
   REMOVE_VARIABLE,
   TSTART,
   TSTOP,
   CLEAR_BUFFER,
   GET_STATUS,
   GET_VARIABLE_VALUE,
   GET_TRACEBUFFER,
   CONFIGURE_TRACEBUFFER,
};


static SIGSELECT req2stateindex[] = {
   MONITOR_TRACEPOINT_CREATE_TRACEPOINT_REQUEST,
   MONITOR_TRACEPOINT_REMOVE_TRACEPOINT_REQUEST,
   MONITOR_TRACEPOINT_CONFIGURE_TRACEPOINT_REQUEST,
   MONITOR_TRACEPOINT_CREATE_ACTION_REQUEST,
   MONITOR_TRACEPOINT_SET_ANNOTATIONS_REQUEST,
   MONITOR_TRACEPOINT_CREATE_VARIABLE_REQUEST,
   MONITOR_TRACEPOINT_REMOVE_VARIABLE_REQUEST,
   MONITOR_TRACEPOINT_START_REQUEST,
   MONITOR_TRACEPOINT_STOP_REQUEST,
   MONITOR_TRACEPOINT_CLEAR_BUFFER_REQUEST,
   MONITOR_TRACEPOINT_GET_STATUS_REQUEST,
   MONITOR_TRACEPOINT_GET_VARIABLE_VALUE_REQUEST,
   MONITOR_TRACEPOINT_GET_TRACEBUFFER_REQUEST,
   MONITOR_TRACEPOINT_CONFIGURE_TRACEBUFFER_REQUEST,
};


static enum state s[14][5] = {
  /*INIT,       CONFIGURED, RUNNING, FULL_PASSCNT, STOP,       */
   {INIT,       CONFIGURED, RUNNING, FULL_PASSCNT, STOP,       },       /* CREATE_TRACEPOINT */
   {INIT,       CONFIGURED, RUNNING, FULL_PASSCNT, STOP,       },       /* REMOVE_TRACEPOINT */
   {INIT,       CONFIGURED, RUNNING, FULL_PASSCNT, STOP,       },       /* CONFIGURE_TRACEPOINT */
   {INIT,       CONFIGURED, RUNNING, FULL_PASSCNT, STOP,       },       /* CREATE_ACTION */
   {INIT,       CONFIGURED, RUNNING, FULL_PASSCNT, STOP,       },       /* SET_ANNOTATIONS */
   {INIT,       CONFIGURED, RUNNING, FULL_PASSCNT, STOP,       },       /* CREATE_VARIABLE */
   {INIT,       CONFIGURED, RUNNING, FULL_PASSCNT, STOP,       },       /* REMOVE_VARIABLE */
   {INIT,       RUNNING,    RUNNING, FULL_PASSCNT, STOP,       },       /* START */
   {INIT,       CONFIGURED, STOP,    STOP,         STOP,       },       /* STOP */
   {INIT,       CONFIGURED, RUNNING, FULL_PASSCNT, CONFIGURED, },       /* CLEAR_BUFFER */
   {INIT,       CONFIGURED, RUNNING, FULL_PASSCNT, STOP,       },       /* GET_STATUS */
   {INIT,       CONFIGURED, RUNNING, FULL_PASSCNT, STOP,       },       /* GET_VARIABLE_VALUE */
   {INIT,       CONFIGURED, RUNNING, FULL_PASSCNT, STOP,       },       /* GET_TRACEBUFFER */
   {CONFIGURED, CONFIGURED, RUNNING, FULL_PASSCNT, STOP,       },       /* CONFIGURE_TRACEBUFFER */
};


typedef int (*tracepoint_handler)(union SIGNAL *, TracepointManagerInfo *);

static tracepoint_handler t[5][17] = {
   /* INIT */
   {
      handle_tracepoint_create_tracepoint_request,
      handle_tracepoint_remove_tracepoint_request,
      handle_tracepoint_configure_tracepoint_request,
      handle_tracepoint_create_action_request,
      handle_tracepoint_set_annotations_request,
      handle_tracepoint_create_variable_request,
      handle_tracepoint_remove_variable_request,
      NULL,
      NULL,
      NULL,
      handle_tracepoint_status_request,
      handle_tracepoint_get_variable_value_request,
      NULL,
      handle_tracepoint_configure_request,
   },
   /* CONFIGURED */
   {
      handle_tracepoint_create_tracepoint_request,
      handle_tracepoint_remove_tracepoint_request,
      handle_tracepoint_configure_tracepoint_request,
      handle_tracepoint_create_action_request,
      handle_tracepoint_set_annotations_request,
      handle_tracepoint_create_variable_request,
      handle_tracepoint_remove_variable_request,
      handle_tracepoint_start_request,
      NULL,
      NULL,
      handle_tracepoint_status_request,
      handle_tracepoint_get_variable_value_request,
      NULL,
      handle_tracepoint_configure_request,
   },
   /* RUNNING */
   {
      NULL,
      NULL,
      NULL,
      NULL,
      NULL,
      NULL,
      NULL,
      NULL,
      handle_tracepoint_stop_request,
      NULL,
      handle_tracepoint_status_request,
      handle_tracepoint_get_variable_value_request,
      NULL,
      handle_tracepoint_configure_request,  /* Ignoring size */
   },
   /* FULL_PASSCNT */
   {
      NULL,
      NULL,
      NULL,
      NULL,
      NULL,
      NULL,
      NULL,
      NULL,
      handle_tracepoint_stop_request,
      NULL,
      handle_tracepoint_status_request,
      handle_tracepoint_get_variable_value_request,
      NULL,
      NULL,
   },
   /* STOP */
   {
      handle_tracepoint_create_tracepoint_request,
      handle_tracepoint_remove_tracepoint_request,
      handle_tracepoint_configure_tracepoint_request,
      handle_tracepoint_create_action_request,
      handle_tracepoint_set_annotations_request,
      handle_tracepoint_create_variable_request,
      handle_tracepoint_remove_variable_request,
      NULL,
      NULL,
      handle_tracepoint_clear_buffer_request,
      handle_tracepoint_status_request,
      handle_tracepoint_get_variable_value_request,
      handle_tracepoint_tracebuffer_request,
      NULL,
   },
};

#define SIZEOF_TABLE(t) sizeof(t)/sizeof(t[0])

static void
handle_tracepoint_signal(union SIGNAL *sig, TracepointManagerInfo **tpm_info)
{


   tracepoint_handler request_handler;
   unsigned int handler_index = 0;
   int found_handler = 0;
   enum state curr_state = (*tpm_info)->state;


   if ((*tpm_info)->state == RUNNING && ((*tpm_info)->status == TRACEPOINT_FULL ||
                                         (*tpm_info)->status == TRACEPOINT_PASSCOUNT))
   {
      /* Update the state before try to call some handler. */
      curr_state = FULL_PASSCNT;
   }

   /* Signals that will read tracepoint configurations. Theese signals will not
    * modify data or change the state*/
   switch (sig->sig_no)
   {
      case MONITOR_TRACEPOINT_GET_TRACEPOINTS_REQUEST:
         handle_tracepoint_get_tracepoints_request(sig, *tpm_info);
         return;

      case RMM_GET_TRACEPOINTS_REQUEST:
         handle_rmm_get_tracepoints_request(sig, *tpm_info);
         return;

      case MONITOR_TRACEPOINT_GET_ACTIONS_REQUEST:
         handle_tracepoint_get_actions_request(sig, *tpm_info);
         return;

      case RMM_GET_TRACEPOINT_ACTIONS_REQUEST:
         handle_rmm_get_actions_request(sig, *tpm_info);
         return;

      case MONITOR_TRACEPOINT_GET_ANNOTATIONS_REQUEST:
         handle_tracepoint_get_annotations_request(sig, *tpm_info);
         return;

      case RMM_GET_TRACEPOINT_ANNOTATIONS_REQUEST:
         handle_rmm_get_annotations_request(sig, *tpm_info);
         return;

      case MONITOR_TRACEPOINT_GET_VARIABLES_REQUEST:
          handle_tracepoint_get_variables_request(sig, *tpm_info);
          return;

      case RMM_GET_TRACEPOINT_VARIABLES_REQUEST:
         handle_rmm_get_variables_request(sig, *tpm_info);
         return;
   }

   for (handler_index = 0 ; handler_index <  SIZEOF_TABLE(req2stateindex) ; handler_index++)
   {
      if (sig->sig_no == req2stateindex[handler_index])
      {
         found_handler = 1;
         break;
      }
   }

   if (!found_handler)
   {
      /* Wow we got a req that we don't expect... should we kill the
       * system?
       */
      error2(0xa7badf0d, handler_index);
   }

   request_handler = t[curr_state][handler_index];

   if (request_handler)
   {
      int error;
      error = request_handler(sig, *tpm_info);
      /* Call the propriet handler, and update state */
      if (error == 0)
      {
         (*tpm_info)->state = s[handler_index][curr_state];
      }
      else
      {
         ramlog_printf("rmm_tracepoint: signal handler for signal:"
                       " %d returned error: %d\n", sig->sig_no, error);
      }
   }
   else
   {
      /* ... or send a error to the user */
      union SIGNAL *rep = alloc(sizeof(MonitorTracepointGeneralTracepointSig),
                                sig->sig_no + 1);

      rep->general_sig.status = MONITOR_STATUS_SCOPE_STATE_INVALID;
      send(&rep, sender(&sig));
      return;
   }
}

static void
create_tp_manager(TracepointManagerInfo **tpm_info,
                  Client *client,
                  int *flow_ctrl_block,
                  int32_t flow_ctrl_blocking,
                  PROCESS client_pid,
                  PROCESS rmm_pid)
{
   /* Create a TracepointManagerInfo object*/
   int cpu;
   *tpm_info
            = (struct TracepointManagerInfo*) alloc(sizeof(TracepointManagerInfo),
                                                    0);
   assert(*tpm_info);
   memset(*tpm_info, 0, sizeof(TracepointManagerInfo));
   (*tpm_info)->client = client;
   (*tpm_info)->tracepoint_id_counter = 1;
   (*tpm_info)->action_id_counter = 1;
   (*tpm_info)->variable_id_counter = 1;
   (*tpm_info)->block_flow_ctrl_counter = flow_ctrl_block;
   (*tpm_info)->block_flow_ctrl_limit = flow_ctrl_blocking;
   (*tpm_info)->client_pid = client_pid;
   (*tpm_info)->rmm_pid = rmm_pid;
   (*tpm_info)->status = TRACEPOINT_NOTRUN;
   (*tpm_info)->state = INIT;
   (*tpm_info)->enabled_tracepoint_list = list_init((List *) alloc(sizeof(List), 0));
   (*tpm_info)->disabled_tracepoint_list = list_init((List *) alloc(sizeof(List), 0));
   (*tpm_info)->variables_list = list_init((List *) alloc(sizeof(List), 0));
   (*tpm_info)->scope.type = MONITOR_SCOPE_VOID;
   (*tpm_info)->convert_bytecode_variable_ids = 0;
   (*tpm_info)->active_callbacks = (uint32_t *) alloc(sizeof(uint32_t) * ose_num_cpus(), 0);
   for (cpu = 0; cpu < ose_num_cpus(); cpu++)
   {
      (*tpm_info)->active_callbacks[cpu] = 0;
   }
}

static void
tp_clean(TracepointManagerInfo **tpm_info_pp)
{
   int done = 0;
   cpuid_t cpu;
   if (tpm_info_pp)
   {
      TracepointManagerInfo *tpm_info = *tpm_info_pp;

      /* Remove eps if needed */
      if (tpm_info->status == TRACEPOINT_RUNNING ||
          tpm_info->status == TRACEPOINT_FULL ||
          tpm_info->status == TRACEPOINT_PASSCOUNT ||
          tpm_info->status == TRACEPOINT_ERROR)
      {
         remove_eps(tpm_info);

         do 
         {
            done = 0;
            for (cpu = 0; cpu < ose_num_cpus(); cpu++) 
            {
               if (tpm_info->active_callbacks[cpu] != 0)
                  done++;
            }
            if (done)
               delay(1);
         }
         while (done);

#ifdef PPC         
         /* Remove data allocated for Fast Tracepoints */
         remove_fast_tracepoints(tpm_info);
#endif
      }
#ifdef PPC         
      else /* Everything is already stoped, remove the tp saftely */
      /* Remove data allocated for Fast Tracepoints */
      {
         remove_fast_tracepoints(tpm_info);
      }
#endif
      

      if (tpm_info->buffer)
      {
         delete_buffer_region((OSADDRESS) tpm_info->buffer);
      }
      free_buf ((union SIGNAL **) &tpm_info->active_callbacks);
   }
}

static void
tp_get_settings(union SIGNAL **sig, TracepointManagerInfo **tpm_info_pp)
{
   TraceBuffer *buffer;
   cpuid_t i;
   TracepointManagerInfo *tpm_info = *tpm_info_pp;
   RmmGetTPConfInfoReply *get_tp_conf = &(*sig)->get_tp_conf_info_reply;

   get_tp_conf->nbrOfTracepoints
            = list_length(tpm_info->enabled_tracepoint_list)
                     + list_length(tpm_info->disabled_tracepoint_list);
   get_tp_conf->nbrOfTracepointActions = tpm_info->nr_of_actions;
   get_tp_conf->nbrOfTracepointVariables
            = list_length(tpm_info->variables_list);
   get_tp_conf->tracepointBufferSize = 0;

   if (tpm_info->trace_buffer_handle)
   {
      for (i = 0, buffer = tpm_info->trace_buffer_handle;
           i < ose_num_cpus();
           i++, buffer = &(tpm_info->trace_buffer_handle)[i])
      {
         get_tp_conf->tracepointBufferSize += (uint64_t)(buffer->size);
      }
      get_tp_conf->tracepointBufferType
         = tpm_info->trace_buffer_handle->buffer_type;
   }
   else
   {
      get_tp_conf->tracepointBufferType = MONITOR_TRACEPOINT_BUFFER_NONE;
   }
}

/*
 * Function pointer to enable that tracepoint code only get linked
 *  when USE_TRACEPOINT flag is set
 */
extern void
(*tp_handle)(union SIGNAL *sig, TracepointManagerInfo **tpm_info);

extern void
(*tp_settings)(union SIGNAL **sig, TracepointManagerInfo **tpm_info);

extern void
(*tp_cleanup)(TracepointManagerInfo **tpm_info);

extern void
(*init_tracepoint_manager)(TracepointManagerInfo **tpm_info,
                           Client *client,
                           int *flow_ctrl_block,
                           int32_t flow_ctrl_blocking,
                           PROCESS client_pid,
                           PROCESS rmm_pid);

/*
 * Called to set the function pointer that handles all tracepoint signals
 * The default value is set in rmm_offspring.c
 */
void
tracepoint_install_exception(void);

void
ose_rmm_tracepoint_init(void)
{
   tp_handle = handle_tracepoint_signal;
   tp_cleanup = tp_clean;
   tp_settings = tp_get_settings;
   init_tracepoint_manager = create_tp_manager;

#ifdef PPC
   /* The installed handler is defined in a architecture specific directory.
    */
   tracepoint_install_exception();
#else
   /* Add support for other arch as tracepoints are fully supported. */
#endif


#if defined(softOSE) || defined(__i386)
#else
   action_handler[MONITOR_TRACEPOINT_ACTION_SAVE_REGISTER] = save_reg;
   action_handler[MONITOR_TRACEPOINT_ACTION_EXPRESSION] = expression_handler;
   action_handler[MONITOR_TRACEPOINT_ACTION_SAVE_MEMORY] = save_mem;
   action_handler[MONITOR_TRACEPOINT_ACTION_WHILE_STEPPING] = whilestepping_handler;
#endif
}
