/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2010-2011 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#include "cpu_smp.h"
#include "ose.h"
#include "monitor.sig"
#include "monitor_tracepoint.sig"
#include "tracepoint_buffer.h"
#include "list.h"
#include "assert.h"
#include "cpu.h"
#define TOUCH(x) x=x

TraceBuffer *
tracepoint_buffer_init(void *buffer, uint32_t trace_size)
{
   TraceBuffer *tracearray;
   TraceBuffer *trace;
   uint32_t trace_size_per_core;
   cpuid_t cpu;
   char *b = buffer;
   cpuid_t nr_cpus = ose_num_cpus();

   tracearray = (TraceBuffer *) alloc(sizeof(TraceBuffer) * nr_cpus, 0);

   trace_size_per_core = trace_size / nr_cpus;

   memset(tracearray, 0, sizeof(TraceBuffer) * nr_cpus);
   for (cpu = 0, trace = tracearray;
        cpu < nr_cpus;
        cpu++, trace = &tracearray[cpu])
   {
      trace->dlink = list_init((List *)alloc(sizeof(List),0));
      trace->size = trace_size_per_core;
      trace->freespace = trace_size_per_core;
      trace->start = (OSADDRESS) &b[cpu * trace_size_per_core];
      trace->end = (OSADDRESS) &b[(cpu+1) * trace_size_per_core];
      trace->buffer_type = MONITOR_TRACEPOINT_BUFFER_NONE;
      trace->next_free = trace->start;
      trace->sorted = False;
   }
   return tracearray;
}

void *
tracepoint_buffer_destroy(TraceBuffer *tracearray)
{
   cpuid_t cpu;
   TraceBuffer *trace = tracearray;
   OSADDRESS buffer = trace->start;
   cpuid_t nr_cpus = ose_num_cpus();

   for (cpu = 0, trace = tracearray;
        cpu < nr_cpus;
        cpu++, trace = &tracearray[cpu])
   {
      free_buf((union SIGNAL **)&trace->dlink);
   }

   memset(tracearray, 0x67, sizeof(TraceBuffer) *  ose_num_cpus());
   free_buf((union SIGNAL **) &tracearray);
   return (void *) buffer;
}

void
tracepoint_buffer_clear(TraceBuffer *tracearray)
{
   TraceBuffer *trace;
   cpuid_t cpu;
   cpuid_t nr_cpus = ose_num_cpus();

   for (cpu = 0, trace = tracearray ; cpu < nr_cpus; cpu++, trace = &tracearray[cpu])
   {
      list_init(trace->dlink);
      trace->freespace = trace->end - trace->start;
      trace->next_free = trace->start;
      trace->frames = 0;
      trace->created = 0;
      trace->add_junkpoint = NULL;
      trace->junk = 0;
      trace->sorted = False;
   }
}

#define ALIGNMENT sizeof(long long)
#define ALIGN_SIZE(x) (((x)+ALIGNMENT-1)&~(ALIGNMENT-1))

enum MonitorTraceStatus
tracepoint_buffer_alloc(TraceBuffer *tracep,
                        uint32_t tracepoint_id,
                        size_t size,
                        TraceEntry **buff)
{
   OSADDRESS next_entry;
   TraceEntry *entry;
   uint32_t real_size;
   uint32_t uint_free_space;
   uint32_t junk = 0;
   TraceBuffer *trace = &tracep[zzkrn_cpu_id()];
   Msr msr;

   real_size = ALIGN_SIZE(size) + sizeof(TraceEntryHead);
   LOCK_SAVE(msr);

   entry = (TraceEntry *) trace->next_free;

   /* If the trace is set to linjar mode and it is full
    * then return that it is not possible to prosead.
    * Here the free space should always be positive so use a
    * unsined variable to silent the compiler.
    */
   uint_free_space = (uint32_t) trace->freespace;

   if ((trace->buffer_type == MONITOR_TRACEPOINT_BUFFER_LINEAR) &&
       (uint_free_space < real_size ))
   {
      trace->freespace = 0;
      LOCK_RESTORE(msr);
      return TRACEPOINT_FULL;
   }

   next_entry = (OSADDRESS) entry;
   next_entry += real_size;

   if (next_entry > trace->end)
   {
      /* Running over the edge in this buffer. Move pointer to beginning.
       * And count the last byte in the log as lost bytes. Re-take them when we
       * pass the area for freeing. We don't have a pointer to the buffer that
       * is in the end of the buffer so get that from tail.
       */
      junk = trace->end - trace->next_free;
      trace->next_free = trace->start + real_size;
      trace->junk = junk;
      entry = (TraceEntry *) trace->start;
      trace->add_junkpoint = list_tail(trace->dlink);
   }
   else
   {
      trace->next_free = next_entry;
   }

   /* Update the head to the next buffer in front of the tail.
    * There can be more then one old buffer that get lost.
    */
   trace->freespace -= real_size + junk;

   while (trace->freespace < 0)
   {
      TraceEntry * head = list_head(trace->dlink);
      trace->freespace += head->real_size;
      /* If head point to the end of the buffer then add the garbage at the
       * end of buffer to the free space.
       */
      if(head->commit != TRACE_ENTRY_COMMITED)
      {
         LOCK_RESTORE(msr);
         return TRACEPOINT_ERROR;
      }
      if (head == trace->add_junkpoint)
      {
         trace->freespace+= trace->junk;
      }
      if (TP == head->type)
      {
         trace->frames--;
      }

      list_unlink(head);
   }

   /* Fill entry header after updating the last part of the trace
    * header it is time to let go of the lock.
    */
   list_init_link(entry);
   list_insert_tail(trace->dlink, entry);
   entry->data_size = size;
   entry->real_size = real_size;
   entry->commit = TRACE_ENTRY_NOT_COMMITED;

   if (tracepoint_id)
   {
      entry->frameid = trace->created++;
      trace->frames++;
   }

   LOCK_RESTORE(msr);
   *buff = entry;
   return TRACEPOINT_RUNNING;
}

void
tracepoint_buffer_config(TraceBuffer *trace, uint32_t tag, uint32_t val)
{
   cpuid_t cpu;
   cpuid_t cpus = ose_num_cpus();

   for (cpu = 0 ; cpu < cpus ; cpu++ )
   {
      switch (tag)
      {
         case TRACEBUFFER_CONFIG_BUFFER_TYPE:
            trace[cpu].buffer_type = val;
            break;
         default:
            break;
      }
   }
}

static int
compare(const ListLink *new_elem, const ListLink *current)
{
   const TraceEntry *new_entry = (const TraceEntry *)new_elem;
   const TraceEntry *current_element = (const TraceEntry *)current;

   /* To you who think that this test looks strange,
    * read this before editing the code!!!
    *
    * Acording to list.h compare should return positive if
    * the new element should be inserted before the existing
    * element, but due to hard work and lots of debugging I
    * found out that before meens list_insert_tail...
    * I want the sorting to go as fast as possible, and because
    * list is almost sorted when it comes in I think that it is
    * best to test all new elements towards the last insorted
    * and hope that it will replace that as the new head.
    * And to achieve that I need to invert the test. So...
    * return -1 forces list_insert_ordered to use the function
    * list_insert_before.
    */
   if (new_entry->frameid >= current_element->frameid)
   {
      return -1;
   }
   return 1;
}

static void*
find_frame_id(ListLink *current_elem, void *new_elem)
{
   TraceEntry *current = (TraceEntry *)current_elem;
   TraceEntry *new = (TraceEntry *)new_elem;
   if (current->frameid == new->frameid)
   {
      return current;
   }
   return NULL;
}

void
tracepoint_buffer_sort(TraceBuffer *tracearray)
{
   cpuid_t cpu;
   cpuid_t nr_cpus = ose_num_cpus();
   TraceBuffer *trace = tracearray;

   for (cpu = 0, trace = tracearray ; cpu < nr_cpus; cpu++, trace = &tracearray[cpu])
   {
      TraceEntry *entry = list_head(trace->dlink);
      TraceEntry *matching_entry;
      List *strace = list_init((List *)alloc(sizeof(List),0));
      List *wrapped_collects = list_init((List *)alloc(sizeof(List), 0));

      /*
       * If buffer starts with COLLECT, buffer has wrapped and the first entries are
       * the latest info from the experiment and should be saved. Put all COLLECTS
       * in a separate list.
       */
      while (entry && entry->type == COLLECT)
      {
         list_unlink(entry);
         list_insert_head(wrapped_collects, entry);
         entry = list_head(trace->dlink);
      }

      /* Now the rest should go sorted into the new list. */
      while (entry)
      {
         list_unlink(entry);
         list_insert_ordered(strace, entry, compare);
         entry = list_head(trace->dlink);
      }

      /*
       * When first sorting is done, put the first COLLECTS in the list where they belong,
       * based on frame id. COLLECTS without TP are removed.
       */
      entry = list_head(wrapped_collects);
      while (entry)
      {
         list_unlink(entry);
         /*
          * Check if COLLECT has a matching entry in list.
          * If there is a hit, it will be the COLLECT before our current entry,
          * since the list is ordered with the earliest item at the bottom.
          */
         matching_entry = list_traverse(strace, find_frame_id, entry);
         if (matching_entry)
         {
            list_insert_before(matching_entry, entry);
         }
         entry = list_head(wrapped_collects);
      }

      /*
       * The list strace is sorted let it replace the old
       * one in the TraceBuffer.  By now the old trace list
       * is empty so free it and let the new one replace it.
       */
      free_buf((union SIGNAL **)&trace->dlink);
      free_buf((union SIGNAL **)&wrapped_collects);
      trace->dlink = strace;
   }
}
