/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2011 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */
/**
 * @toc File_System:
 *
 * @file ose_spi/fm.sig
 *
 * @brief Signal interface for format and volume managers.
 *
 * @restriction User/group ownership are not implemented.
 *
 *
 */

#ifndef _FM_SIG
#define _FM_SIG

#include "ose.h"
#include "osetypes.h"
#include "ose_spi/efstypes.h"
#include "stddef.h"

#ifdef __cplusplus
extern "C" {
#endif

/*
 ******************************************************************************
 * Declarations for signals.
 ******************************************************************************
 */

/** @group FM_MODE_XXXXXX @brief See enum @link "" "FmModeEnum" */

/**
 * @type FmModeEnum
 *
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * enum FmModeEnum
 * {
 *   FM_MODE_READ       = 0x0001,
 *   FM_MODE_WRITE      = 0x0002,
 *   FM_MODE_CREATE     = 0x0004,
 *   FM_MODE_EXCLUSIVE  = 0x0008,
 *   FM_MODE_TRUNCATE   = 0x0010,
 *   FM_MODE_TEXT       = 0x0020,
 *   FM_MODE_DIRECTORY  = 0x0040,
 *   FM_MODE_SLINK      = 0x0080,
 *   FM_MODE_HLINK      = 0x0100,
 *   FM_MODE_APPEND     = 0x0200,
 *   FM_MODE_TRANS      = 0x0400
 * };
 *
 * @brief Bits indicating access modes when the file is open.
 *
 * @long These values are bits in a bitmask and may be combined together to form
 *       the value for the field. The fixed size type @link "ose_spi/efstypes.h"
 *       "EfsBitmask" is used to store the resulting value, because an ENUM
 *       would have a compiler dependent size.
 * @long Which bits are supported depends on the volume manager.
 *
 * @field FM_MODE_READ      Allow reading from the open object.
 * @field FM_MODE_WRITE     Allow writing to the open object.
 * @field FM_MODE_CREATE    Create object if it does not exists.
 * @field FM_MODE_EXCLUSIVE Fail if the object already exists and already is
 *                          open.
 * @field FM_MODE_TRUNCATE  Truncate, if object already exists.
 * @field FM_MODE_TEXT      Open object in text mode (raw binary mode is
 *                          default).
 * @field FM_MODE_DIRECTORY Open/create directory.
 * @field FM_MODE_SLINK     Open/create soft (symbolic) link.
 * @field FM_MODE_HLINK     Open/create hard link.
 * @field FM_MODE_APPEND    Append on write.
 * @field FM_MODE_TRANS     Transactional file writes; crash or power outage
 *                          before fsync or close reverts file contents to the
 *                          state at open or previous fsync time. This flag is
 *                          silently ignored if the volume manager does
 *                          not support transactional file writes.
 */
enum FmModeEnum
{
   FM_MODE_READ         = 0x0001,
   FM_MODE_WRITE        = 0x0002,
   FM_MODE_CREATE       = 0x0004,
   FM_MODE_EXCLUSIVE    = 0x0008,
   FM_MODE_TRUNCATE     = 0x0010,
   FM_MODE_TEXT         = 0x0020,
   FM_MODE_DIRECTORY    = 0x0040,
   FM_MODE_SLINK        = 0x0080,
   FM_MODE_HLINK        = 0x0100,
   FM_MODE_APPEND       = 0x0200,
   FM_MODE_TRANS        = 0x0400
};

/**
 * @type FmTypeEnum
 *
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * enum FmTypeEnum
 * {
 *   FM_TYPE_FILE       = 0x0,
 *   FM_TYPE_STREAM     = 0x1,
 *   FM_TYPE_DIR        = 0x2,
 *   FM_TYPE_SLINK      = 0x3,
 * };
 *
 * @brief Object types.
 *
 * @long  Which bits are supported depends on the volume manager.
 *
 * @field FM_TYPE_FILE   Plain (seekable) file.
 * @field FM_TYPE_STREAM Byte stream (not seekable); terminals, pipes etc.
 * @field FM_TYPE_DIR    Directory.
 * @field FM_TYPE_SLINK  Soft (symbolic) link.
 */
enum FmTypeEnum
{
   FM_TYPE_FILE         = 0x0,
   FM_TYPE_STREAM       = 0x1,
   FM_TYPE_DIR          = 0x2,
   FM_TYPE_SLINK        = 0x3
};

/**
 * @type FmCancelEnum
 *
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * enum FmCancelEnum
 * {
 *   FM_CANCEL_PRIVATE   = 0x0,
 *   FM_CANCEL_GLOBAL    = 0x1,
 *   FM_CANCEL_BYBUFFER  = 0x2,
 * };
 *
 * @brief Cancellation flags
 *
 * @field FM_CANCEL_PRIVATE Cancel matching operations for sender process.
 * @field FM_CANCEL_GLOBAL  Cancel matching operations for all processes.
 * @field FM_CANCEL_BYBUFFER Cancel matching (including buffer ptr) operations.
 */
enum FmCancelEnum
{
   FM_CANCEL_PRIVATE      = 0x0,
   FM_CANCEL_GLOBAL       = 0x1,
   FM_CANCEL_BYBUFFER     = 0x2
};

/**
 * @type FmAccessEnum
 *
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * enum FmAccessEnum
 * {
 *   FM_ACCESS_READ      = 0x01,
 *   FM_ACCESS_WRITE     = 0x02,
 *   FM_ACCESS_RANDOM    = 0x04,
 *   FM_ACCESS_REMOVABLE = 0x08,
 *   FM_ACCESS_FORMAT    = 0x10
 * };
 *
 * @brief Bits indicating volume (disk) accessibility.
 *
 * @long These values are bits in a bitmask and may be OR'ed together to form
 *       the value for the field. The fixed size type @link "ose_spi/efstypes.h"
 *       "EfsBitmask" is used to store the resulting value, because an ENUM
 *       would have a compiler dependent size.
 *
 * @field FM_ACCESS_READ      Volume is readable.
 * @field FM_ACCESS_WRITE     Volume is writable.
 * @field FM_ACCESS_RANDOM    Volume is randomly accessible.
 * @field FM_ACCESS_REMOVABLE Volume is removable.
 * @field FM_ACCESS_FORMAT    Volume is formatted.
 */
enum FmAccessEnum
{
   FM_ACCESS_READ       = 0x01,
   FM_ACCESS_WRITE      = 0x02,
   FM_ACCESS_RANDOM     = 0x04,
   FM_ACCESS_REMOVABLE  = 0x08,
   FM_ACCESS_FORMAT     = 0x10
};

/**
 * @type FmWhenceEnum
 *
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * enum FmWhenceEnum
 * {
 *   FM_WHENCE_BEGIN    = 0x0,
 *   FM_WHENCE_CURRENT  = 0x1,
 *   FM_WHENCE_END      = 0x2
 * };
 *
 * @brief Position origins.
 *
 * @long For positioning, the following origins are defined from which the
 *       offset is calculated.
 *
 * @field FM_WHENCE_BEGIN   From the beginning of the file.
 * @field FM_WHENCE_CURRENT From the current file position.
 * @field FM_WHENCE_END     From the end of the file, normally with a negative
 *                          offset.
 */
enum FmWhenceEnum
{
   FM_WHENCE_BEGIN      = 0x0,
   FM_WHENCE_CURRENT    = 0x1,
   FM_WHENCE_END        = 0x2
};

/**
 * @type FmTimeEnum
 *
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 *
 * @brief Bits selecting object timestamps.
 *
 * @long These values are bits in a bitmask and may be combined together to form
 *       the value for the field. The fixed size type @link "ose_spi/efstypes.h"
 *       "EfsBitmask" is used to store the resulting value, because an ENUM
 *       would have a compiler dependent size.
 *
 * @field FM_TIME_CREATION Implementation defined: either time when the object
 *                         was created or POSIX ctime.
 * @field FM_TIME_MODDATA  Time of last modification of the object content.
 * @field FM_TIME_MODHEAD  Time of last modification of the object metadata.
 * @field FM_TIME_ACCESS   Time of last access (last read or write).
 */
enum FmTimeEnum
{
   FM_TIME_CREATION     = 0x1,
   FM_TIME_MODDATA      = 0x2,
   FM_TIME_MODHEAD      = 0x4,
   FM_TIME_ACCESS       = 0x8
};

/**
 * @type FmAttrEnum
 *
 * @include "ose_spi/fm.sig"
 *
 * @brief Bits indicating permission attributes for files or directories.
 *
 * @long These bits indicate read, write, execute permissions for the classes
 *       owner, group and others. And the additional archive attribute, which
 *       optionally may be set and tested for by an archiving application.
 * @long These values are bits in a bitmask and may be combined together to
 *       form the value for the field. The fixed size
 *       type @link "ose_spi/efstypes.h" "EfsBitmask" is used to store the
 *       resulting value, because an ENUM would have a compiler dependent size.
 *
 * @field FM_ATTR_OTH_EXECUTE Others may execute.
 * @field FM_ATTR_OTH_READ    Others may read.
 * @field FM_ATTR_OTH_WRITE   Others may write.
 * @field FM_ATTR_GRP_EXECUTE Group may execute.
 * @field FM_ATTR_GRP_READ    Group may read.
 * @field FM_ATTR_GRP_WRITE   Group may write.
 * @field FM_ATTR_OWN_EXECUTE Owner may execute.
 * @field FM_ATTR_OWN_READ    Owner may read.
 * @field FM_ATTR_OWN_WRITE   Owner may write.
 * @field FM_ATTR_ARCHIVED    Set when file is changed.
 * @field FM_ATTR_NAMEOK      Internal use.
 * @field FM_ATTR_XOPEN       Only supports open for write if no previous open
 *                            streams.
 * @field FM_ATTR_READ        All classes may read.
 * @field FM_ATTR_WRITE       All classes may write.
 * @field FM_ATTR_EXECUTE     All classes may execute.
 * @field FM_ATTR_HIDDEN      Hidden file.
 * @field FM_ATTR_USER0       User defined attribute bit 0.
 * @field FM_ATTR_USER1       User defined attribute bit 1.
 * @field FM_ATTR_USER2       User defined attribute bit 2.
 * @field FM_ATTR_USER3       User defined attribute bit 3.
 * @field FM_ATTR_ISVTX       Sticky bit, semantics as in UNIX.
 * @field FM_ATTR_SUID        Program will run as if started by owner.
 * @field FM_ATTR_SGID        Program will run as if started by group.
 * @field FM_ATTR_STARTUP     File is marked as a startup file.
 */
enum FmAttrEnum
{
   FM_ATTR_OTH_EXECUTE  = 0x001,
   FM_ATTR_OTH_READ     = 0x002,
   FM_ATTR_OTH_WRITE    = 0x004,
   FM_ATTR_GRP_EXECUTE  = 0x008,
   FM_ATTR_GRP_READ     = 0x010,
   FM_ATTR_GRP_WRITE    = 0x020,
   FM_ATTR_OWN_EXECUTE  = 0x040,
   FM_ATTR_OWN_READ     = 0x080,
   FM_ATTR_OWN_WRITE    = 0x100,
   FM_ATTR_ARCHIVED     = 0x200,
   FM_ATTR_NAMEOK       = 0x400,
   FM_ATTR_XOPEN        = 0x800,
   FM_ATTR_HIDDEN       = 0x1000,
   FM_ATTR_USER0        = 0x10000,
   FM_ATTR_USER1        = 0x20000,
   FM_ATTR_USER2        = 0x40000,
   FM_ATTR_USER3        = 0x80000,
   FM_ATTR_ISVTX        = 0x100000,
   FM_ATTR_SUID         = 0x200000,
   FM_ATTR_SGID         = 0x400000,
   FM_ATTR_STARTUP      = 0x800000,
   FM_ATTR_READ         = (FM_ATTR_OTH_READ |
                           FM_ATTR_GRP_READ |
                           FM_ATTR_OWN_READ),
   FM_ATTR_WRITE        = (FM_ATTR_OTH_WRITE |
                           FM_ATTR_GRP_WRITE |
                           FM_ATTR_OWN_WRITE),
   FM_ATTR_EXECUTE      = (FM_ATTR_OTH_EXECUTE |
                           FM_ATTR_GRP_EXECUTE |
                           FM_ATTR_OWN_EXECUTE)
};

/**
 * @type FmCondEnum
 *
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * enum FmCondEnum
 * {
 *   FM_COND_READ       = 0x1,
 *   FM_COND_WRITE      = 0x2,
 *   FM_COND_ERROR      = 0x4,
 *   FM_COND_HANGUP     = 0x8
 * };
 *
 * @brief Bits indicating waiting event conditions.
 *
 * @long These values are bits in a bitmask and may be combined together to
 *       form the value for the field. The fixed size
 *       type @link "ose_spi/efstypes.h" "EfsBitmask" is used to store the
 *       resulting value, because an ENUM would have a compiler dependent size.
 *
 * @field FM_COND_READ   Wait for possibility to read.
 * @field FM_COND_WRITE  Wait for possibility to write.
 * @field FM_COND_ERROR  Wait for error condition.
 * @field FM_COND_HANGUP Wait for hangup condition.
 */
enum FmCondEnum
{
   FM_COND_READ         = 0x1,
   FM_COND_WRITE        = 0x2,
   FM_COND_ERROR        = 0x4,
   FM_COND_HANGUP       = 0x8
};

/**
 * @type FmOperEnum
 *
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * enum FmOperEnum
 * {
 *   FM_OPER_READ       = 0x1,
 *   FM_OPER_WRITE      = 0x2,
 *   FM_OPER_WAIT       = 0x4,
 *   FM_OPER_DRAIN      = 0x8,
 *   FM_OPER_UPDATES    = 0x10
 * };
 *
 * @brief Bits indicating cancel pending operations on an open object.
 *
 * @long These values are bits in a bitmask and may be combined together to
 *       form the value for the field. The fixed size type
 *       @link "ose_spi/efstypes.h" "EfsBitmask" is used to store the
 *       resulting value, because an ENUM would have a compiler dependent size.
 *
 * @field FM_OPER_READ    Cancel all pending read operations.
 * @field FM_OPER_WRITE   Cancel all pending write operations.
 * @field FM_OPER_WAIT    Cancel all pending wait operations.
 * @field FM_OPER_DRAIN   Cancel all pending drain operations.
 * @field FM_OPER_UPDATES Cancel all pending (unsynched) writes.
 */
enum FmOperEnum
{
   FM_OPER_READ         = 0x1,
   FM_OPER_WRITE        = 0x2,
   FM_OPER_WAIT         = 0x4,
   FM_OPER_DRAIN        = 0x8,
   FM_OPER_UPDATES      = 0x10
};

/**
 * @type FmPositionEnum
 *
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * enum FmPositionEnum
 * {
 *   FM_POSITION_AT_END  = -1,
 *   FM_POSITION_CURRENT = -2
 * };
 *
 * @brief Special values for the position field for reads and writes.
 *
 * @long The position field in the read and write signals give the option
 *       to update the file offset before starting the transfer. The options
 *       include keeping the offset unchanged, seeeking to end-of-file and
 *       seeking a specific file position.
 *
 * @long WARNING: These constants are enums, which are signed, while the
 *                "position" fields in the signals are unsigned. These
 *                constants should be cast to @link "" "EfsEnum" when
 *                comparing, e.g.
 *                (sig->position == (EfsEnum)FM_POSITION_CURRENT)
 *
 * @long The file offset will be set to a the value of the "position" field
 *       if the value is in the range 0x00000000 to 0xFFFFFFFD.
 *
 * @field FM_POSITION_AT_END  Adjust the file offset to end-of-file before
 *                            starting the transfer.
 * @field FM_POSITION_CURRENT The file offset of the object shall not be
 *                            updated before starting the transfer.
 *
 * @seealso FmRead
 * @seealso FmWrite
 * @seealso FmReadPtr
 * @seealso FmWritePtr
 */
enum FmPositionEnum
{
   FM_POSITION_AT_END   = -1,
   FM_POSITION_CURRENT  = -2
};

/**
 * @type FmBuffEnum
 *
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * enum FmBuffEnum
 * {
 *   FM_BUFF_NONE               = 0x0,
 *   FM_BUFF_LINE               = 0x1,
 *   FM_BUFF_FULL               = 0x2
 * };
 *
 * @brief Buffering modes of an object.
 *
 * @field FM_BUFF_NONE No buffering.
 * @field FM_BUFF_LINE Line buffering (until new line).
 * @field FM_BUFF_FULL Full buffering.
 */
enum FmBuffEnum
{
   FM_BUFF_NONE         = 0x0,
   FM_BUFF_LINE         = 0x1,
   FM_BUFF_FULL         = 0x2
};

/**
 * @type FmLkopEnum
 *
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * enum FmLkopEnum
 * {
 *   FM_LKOP_INFO       = 0x0,
 *   FM_LKOP_OBTAIN     = 0x1,
 *   FM_LKOP_OBTAINW    = 0x2,
 *   FM_LKOP_RELEASE    = 0x3
 * };
 *
 * @brief Types of lock operations on an object.
 *
 * @field FM_LKOP_INFO    Get lock information.
 * @field FM_LKOP_OBTAIN  Obtain lock non-blocking.
 * @field FM_LKOP_OBTAINW Obtain lock blocking.
 * @field FM_LKOP_RELEASE Release lock.
 */
enum FmLkopEnum
{
   FM_LKOP_INFO         = 0x0,
   FM_LKOP_OBTAIN       = 0x1,
   FM_LKOP_OBTAINW      = 0x2,
   FM_LKOP_RELEASE      = 0x3
};

/**
 * @type FmLktpEnum
 *
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * enum FmLktpEnum
 * {
 *   FM_LKTP_SHARED     = 0x0,
 *   FM_LKTP_EXCLUSIVE  = 0x1
 * };
 *
 * @brief Types of locks on an object.
 *
 * @field FM_LKTP_SHARED    Shared (read) lock.
 * @field FM_LKTP_EXCLUSIVE Exclusive (write) lock.
 */
enum FmLktpEnum
{
   FM_LKTP_SHARED       = 0x0,
   FM_LKTP_EXCLUSIVE    = 0x1
};

/**
 * @type FmQueueEnum
 *
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * enum FmQueueEnum
 * {
 *   FM_QUEUE_RECEIVE   = 0x1,
 *   FM_QUEUE_TRANSMIT  = 0x2
 * };
 *
 * @brief Terminal queue(s).
 *
 * @long These values are bits in a bitmask and may be combined together to
 *       form the
 *       value for the field. The fixed size type @link "ose_spi/efstypes.h"
 *       "EfsBitmask" is used to store the resulting value, because an ENUM
 *       would have a compiler dependent size.
 *
 * @field FM_QUEUE_RECEIVE  Receive (input) queue.
 * @field FM_QUEUE_TRANSMIT Transmit (output) queue.
 */
enum FmQueueEnum
{
   FM_QUEUE_RECEIVE     = 0x1,
   FM_QUEUE_TRANSMIT    = 0x2
};

/**
 * @type FmTiflEnum
 *
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * enum FmTiflEnum
 * {
 *   FM_TIFL_LINE_EDIT   = 0x001,
 *   FM_TIFL_SAVE_LINE   = 0x002,
 *   FM_TIFL_MAP_NL_CR   = 0x004,
 *   FM_TIFL_MAP_CR_NL   = 0x008,
 *   FM_TIFL_IGNORE_CR   = 0x010,
 *   FM_TIFL_FULL_BELL   = 0x020,
 *   FM_TIFL_ECHO        = 0x040,
 *   FM_TIFL_ECHO_CTRL   = 0x080,
 *   FM_TIFL_FLOW_CTRL   = 0x100,
 *   FM_TIFL_ASYNC_SIGS  = 0x200,
 *   FM_TIFL_MAP_NL_CRNL = 0x400,
 *   FM_TIFL_HUP_CLOSE   = 0x800,
 *   FM_TIFL_RAW_INPUT   = 0x1000
 * };
 *
 * @brief FmTerminalInfo flags.
 *
 * @long These values are bits in a bitmask and may be combined together to
 *       form the
 *       value for the field. The fixed size type @link "ose_spi/efstypes.h"
 *       "EfsBitmask" is used to store the resulting value, because an ENUM
 *       would have a compiler dependent size.
 *
 * @field FM_TIFL_LINE_EDIT   Input, line edit, using the line editing terminal
 *                            control characters defined in the ctrlChars field.
 * @field FM_TIFL_SAVE_LINE   Input, save lines in history.
 * @field FM_TIFL_MAP_NL_CR   Input, map NL to CR.
 * @field FM_TIFL_MAP_CR_NL   Input, map CR to NL.
 * @field FM_TIFL_IGNORE_CR   Input, ignore CR.
 * @field FM_TIFL_FULL_BELL   Input, output bell on full input buffer.
 * @field FM_TIFL_ECHO        Input, echo input characters.
 * @field FM_TIFL_ECHO_CTRL   Input, echo control characters as ^char.
 * @field FM_TIFL_FLOW_CTRL   Input, enable START/STOP, using the start and stop
 *                            flow control characters in the ctrlChars field
 *                            (usually XON/XOFF).
 * @field FM_TIFL_ASYNC_SIGS  Input, enable INTR/SUSP, using the asynchronous
 *                            signaling control characters in the ctrlChars
 *                            field.
 * @field FM_TIFL_MAP_NL_CRNL Output, map NL to CR+NL.
 * @field FM_TIFL_HUP_CLOSE   Hardware control, hangup on last close.
 * @field FM_TIFL_RAW_INPUT   Input, 100 percent raw mode.

 */
enum FmTiflEnum
{
   FM_TIFL_LINE_EDIT    = 0x001,
   FM_TIFL_SAVE_LINE    = 0x002,
   FM_TIFL_MAP_NL_CR    = 0x004,
   FM_TIFL_MAP_CR_NL    = 0x008,
   FM_TIFL_IGNORE_CR    = 0x010,
   FM_TIFL_FULL_BELL    = 0x020,
   FM_TIFL_ECHO         = 0x040,
   FM_TIFL_ECHO_CTRL    = 0x080,
   FM_TIFL_FLOW_CTRL    = 0x100,
   FM_TIFL_ASYNC_SIGS   = 0x200,
   FM_TIFL_MAP_NL_CRNL  = 0x400,
   FM_TIFL_HUP_CLOSE    = 0x800,
   FM_TIFL_RAW_INPUT    = 0x1000
};

/**
 * @type FmTccEnum
 *
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * enum FmTccEnum
 * {
 *   FM_TCC_EOL         = 0x00,
 *   FM_TCC_EOF         = 0x01,
 *   FM_TCC_ERASELEFT   = 0x02,
 *   FM_TCC_ERASERIGHT  = 0x03,
 *   FM_TCC_WERASE      = 0x04,
 *   FM_TCC_KILL        = 0x05,
 *   FM_TCC_REPRINT     = 0x06,
 *   FM_TCC_INTR        = 0x07,
 *   FM_TCC_SUSP        = 0x08,
 *   FM_TCC_START       = 0x09,
 *   FM_TCC_STOP        = 0x0A,
 *   FM_TCC_DISCARD     = 0x0B,
 *   FM_TCC_VERBATIM    = 0x0C,
 *   FM_TCC_STATUS      = 0x0D,
 *   FM_TCC_CURLEFT     = 0x0E,
 *   FM_TCC_CURRIGHT    = 0x0F,
 *   FM_TCC_CURBEGIN    = 0x10,
 *   FM_TCC_CUREND      = 0x11,
 *   FM_TCC_SEARCH      = 0x12,
 *   FM_TCC_PREVLINE    = 0x13,
 *   FM_TCC_NEXTLINE    = 0x14,
 *   FM_TCC_ERASEEOL    = 0x15,
 *   FM_TCC_PASTE       = 0x16,
 *   FM_TCC_COMPLETE    = 0x17
 * };
 *
 * @brief Terminal control characters in the FmTerminalInfo ctrlChars field.
 *
 * @field FM_TCC_EOL        End of line (FM_TIFL_LINE_EDIT).
 * @field FM_TCC_EOF        End of file (FM_TIFL_LINE_EDIT).
 * @field FM_TCC_ERASELEFT  Erase char to left (FM_TIFL_LINE_EDIT).
 * @field FM_TCC_ERASERIGHT Erase char to right (FM_TIFL_LINE_EDIT).
 * @field FM_TCC_WERASE     Erase word to left (FM_TIFL_LINE_EDIT).
 * @field FM_TCC_KILL       Kill line (FM_TIFL_LINE_EDIT).
 * @field FM_TCC_REPRINT    Reprint input (FM_TIFL_LINE_EDIT).
 * @field FM_TCC_INTR       Interrupt process (FM_TIFL_ASYNC_SIGS).
 * @field FM_TCC_SUSP       Suspend process (FM_TIFL_ASYNC_SIGS).
 * @field FM_TCC_START      Start output flow (XON) (FM_TIFL_FLOW_CTRL).
 * @field FM_TCC_STOP       Stop output flow (XOFF) (FM_TIFL_FLOW_CTRL).
 * @field FM_TCC_DISCARD    Discard queued output (FM_TIFL_LINE_EDIT).
 * @field FM_TCC_VERBATIM   Verbatim input (FM_TIFL_LINE_EDIT).
 * @field FM_TCC_STATUS     Print status (FM_TIFL_LINE_EDIT).
 * @field FM_TCC_CURLEFT    Cursor left (FM_TIFL_LINE_EDIT).
 * @field FM_TCC_CURRIGHT   Cursor right (FM_TIFL_LINE_EDIT).
 * @field FM_TCC_CURBEGIN   Cursor to beginning of line (FM_TIFL_LINE_EDIT).
 * @field FM_TCC_CUREND     Cursor to end of line (FM_TIFL_LINE_EDIT).
 * @field FM_TCC_SEARCH     Search history (FM_TIFL_LINE_EDIT).
 * @field FM_TCC_PREVLINE   Previous line in history (FM_TIFL_LINE_EDIT).
 * @field FM_TCC_NEXTLINE   Next line in history (FM_TIFL_LINE_EDIT).
 * @field FM_TCC_ERASEEOL   Erase to end of line (FM_TIFL_LINE_EDIT).
 * @field FM_TCC_PASTE      Paste clipboard (FM_TIFL_LINE_EDIT).
 * @field FM_TCC_COMPLETE   Complete word/filename (FM_TIFL_LINE_EDIT).
 *
 * @seealso FmTiflEnum
 */
enum FmTccEnum
{
   FM_TCC_EOL           = 0x00,
   FM_TCC_EOF           = 0x01,
   FM_TCC_ERASELEFT     = 0x02,
   FM_TCC_ERASERIGHT    = 0x03,
   FM_TCC_WERASE        = 0x04,
   FM_TCC_KILL          = 0x05,
   FM_TCC_REPRINT       = 0x06,
   FM_TCC_INTR          = 0x07,
   FM_TCC_SUSP          = 0x08,
   FM_TCC_START         = 0x09,
   FM_TCC_STOP          = 0x0A,
   FM_TCC_DISCARD       = 0x0B,
   FM_TCC_VERBATIM      = 0x0C,
   FM_TCC_STATUS        = 0x0D,
   FM_TCC_CURLEFT       = 0x0E,
   FM_TCC_CURRIGHT      = 0x0F,
   FM_TCC_CURBEGIN      = 0x10,
   FM_TCC_CUREND        = 0x11,
   FM_TCC_SEARCH        = 0x12,
   FM_TCC_PREVLINE      = 0x13,
   FM_TCC_NEXTLINE      = 0x14,
   FM_TCC_ERASEEOL      = 0x15,
   FM_TCC_PASTE         = 0x16,
   FM_TCC_COMPLETE      = 0x17
};

/**
 * @macro FM_TCC_DISABLED
 *
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_TCC_DISABLED              ((unsigned char)0xFF)
 *
 * @brief Special value used when disabling a terminal control character.
 *
 * @long Represents "No-Control-Character assigned".
 */
#define FM_TCC_DISABLED         ((unsigned char)0xFF)

/**
 * @macro FM_TCC_SIZE
 *
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_TCC_SIZE          0x20
 *
 * @brief Buffer size for FmTerminalInfo ctrlChars buffer.
 */
#define FM_TCC_SIZE             0x20

/**
 * @type FmTermEnum
 *
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * enum FmTermEnum
 * {
 *   FM_TERM_READMIN    = 0x00001,
 *   FM_TERM_READTMO    = 0x00002,
 *   FM_TERM_RXBUFFER   = 0x00004,
 *   FM_TERM_RXSPEED    = 0x00008,
 *   FM_TERM_TXSPEED    = 0x00010,
 *   FM_TERM_RXFLOW     = 0x00020,
 *   FM_TERM_TXFLOW     = 0x00040,
 *   FM_TERM_CHARBITS   = 0x00080,
 *   FM_TERM_STOPBITS   = 0x00100,
 *   FM_TERM_PARITY     = 0x00200,
 *   FM_TERM_FLOWCTRL   = 0x00400,
 *   FM_TERM_FLAGS      = 0x00800,
 *   FM_TERM_WIDTH      = 0x01000,
 *   FM_TERM_HEIGHT     = 0x02000,
 *   FM_TERM_CTRLCHARS  = 0x04000,
 *   FM_TERM_VOLNAME    = 0x08000,
 *   FM_TERM_DEVNAME    = 0x10000,
 *   FM_TERM_TERMTYPE   = 0x20000,
 *   FM_TERM_CLEARHIST  = 0x40000,
 *   FM_TERM_IDLETMO    = 0x80000
 * };
 *
 * @brief Bits indicating valid fields in the FmTerminalInfo structure.
 *
 * @long These bits are set in the FmTerminalInfo valid field to indicate
 *       which fields
 *       are present in the structure. Fields for which the corresponding
 *       bits are not set should be ignored.
 *
 * @long These values are bits in a bitmask and may be combined together to
 *       form the
 *       value for the field. The fixed size type @link "ose_spi/efstypes.h"
 *       "EfsBitmask" is used to store the resulting value, because an ENUM
 *       would have a compiler dependent size.
 *
 * @field FM_TERM_READMIN   Specifies that the readSizeMin field is valid.
 * @field FM_TERM_READTMO   Specifies that the readTimeOut field is valid.
 * @field FM_TERM_RXBUFFER  Specifies that the rxBuffer field is valid.
 * @field FM_TERM_RXSPEED   Specifies that the rxSpeed field is valid.
 * @field FM_TERM_TXSPEED   Specifies that the txSpeed field is valid.
 * @field FM_TERM_RXFLOW    Specifies that the rxFlow field is valid.
 * @field FM_TERM_TXFLOW    Specifies that the txFlow field is valid.
 * @field FM_TERM_CHARBITS  Specifies that the charBits field is valid.
 * @field FM_TERM_STOPBITS  Specifies that the stopBits field is valid.
 * @field FM_TERM_PARITY    Specifies that the parity field is valid.
 * @field FM_TERM_FLOWCTRL  Specifies that the flowCtrl field is valid.
 * @field FM_TERM_FLAGS     Specifies that the flags field is valid.
 * @field FM_TERM_WIDTH     Specifies that the width field is valid.
 * @field FM_TERM_HEIGHT    Specifies that the height field is valid.
 * @field FM_TERM_CTRLCHARS Specifies that the ctrlChars field is valid.
 * @field FM_TERM_VOLNAME   Specifies that the volumeName field is valid.
 * @field FM_TERM_DEVNAME   Specifies that the deviceName field is valid.
 * @field FM_TERM_TERMTYPE  Specifies that the termType field is valid.
 * @field FM_TERM_CLEARHIST When this bit is set, the history of shell
 *                          commands will be cleared.
 * @field FM_TERM_IDLETMO   Specifies that the idleTmo field is valid.
 */
enum FmTermEnum
{
   FM_TERM_READMIN      = 0x00001,
   FM_TERM_READTMO      = 0x00002,
   FM_TERM_RXBUFFER     = 0x00004,
   FM_TERM_RXSPEED      = 0x00008,
   FM_TERM_TXSPEED      = 0x00010,
   FM_TERM_RXFLOW       = 0x00020,
   FM_TERM_TXFLOW       = 0x00040,
   FM_TERM_CHARBITS     = 0x00080,
   FM_TERM_STOPBITS     = 0x00100,
   FM_TERM_PARITY       = 0x00200,
   FM_TERM_FLOWCTRL     = 0x00400,
   FM_TERM_FLAGS        = 0x00800,
   FM_TERM_WIDTH        = 0x01000,
   FM_TERM_HEIGHT       = 0x02000,
   FM_TERM_CTRLCHARS    = 0x04000,
   FM_TERM_VOLNAME      = 0x08000,
   FM_TERM_DEVNAME      = 0x10000,
   FM_TERM_TERMTYPE     = 0x20000,
   FM_TERM_CLEARHIST    = 0x40000,
   FM_TERM_IDLETMO      = 0x80000
};

/**
 * @type FmNotificationEnum
 *
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * enum FmNotificationEnum
 * {
 *   FMN_CREATED          = 0,
 *   FMN_FORMAT           = 1,
 *   FMN_MODIFIED         = 2,
 *   FMN_REMOVED          = 3,
 *   FMN_RENAMED          = 4,
 *   FMN_UNMOUNT          = 5
 * };
 *
 * @brief Type of notification event.
 *
 * @field FMN_CREATED  File or directory created.
 * @field FMN_FORMAT   Volume has been formatted.
 * @field FMN_MODIFIED File or directory has been modified.
 * @field FMN_REMOVED  File or directory has been removed.
 * @field FMN_RENAMED  File or directory has been renamed.
 * @field FMN_UNMOUNT  Volume has been unmounted.
 */
enum FmNotificationEnum
{
   FMN_CREATED          = 0,
   FMN_FORMAT           = 1,
   FMN_MODIFIED         = 2,
   FMN_REMOVED          = 3,
   FMN_RENAMED          = 4,
   FMN_UNMOUNT          = 5
};


/* Invalid parameters (file names, handles, seek positions, locks etc.) */

/**
 * @macro FM_ERRBASE
 *
 * @include "ose_spi/fm.sig"
 *
 * @declaration FM_ERRBASE #define FM_ERRBASE 200
 *
 * @brief Start of FM/VM status codes.
 * @long The FM_E* status codes are applicable to the status field 
 *       of the FM_*_REPLY signals. These codes are all in the range 200-233.
 *       The following are possible values that a returned status for a
 *       request can have. A client should be able to handle both OSE 
 *       @link "errno.h" "errno" codes, which are preferred for new designs, 
 *       and old FM_E* status codes. In order to do so 
 *       @link "efs.h" "efs_status_to_errno", which provides FM_E* codes 
 *       conversion, may be used.
 */
#define FM_ERRBASE          (200)

/**
 * @macro FM_EINVALID_NAME
 *
 * @brief Given object (file or directory) name could not be found.
 * @long This is a status code applicable to the status field of 
 *       the FM_*_REPLY signals.
 *
 * @declaration #define FM_EINVALID_NAME 200
 */
#define FM_EINVALID_NAME             (200)

/**
 * @macro FM_EINVALID_HANDLE
 *
 * @brief Given handle is unknown, either because it has not been
 *        opened, or because it has been closed.
 * @long This is a status code applicable to the status field of 
 *       the FM_*_REPLY signals.
 *
 * @declaration #define FM_EINVALID_HANDLE 201
 */
#define FM_EINVALID_HANDLE           (201)

/**
 * @macro FM_EINVALID_POSITION
 *
 * @brief Requested file position does not exist.
 *        E.g. when trying to seek past the end of the file.
 * @long This is a status code applicable to the status field of 
 *       the FM_*_REPLY signals.
 *
 * @declaration #define FM_EINVALID_POSITION 202
 */
#define FM_EINVALID_POSITION         (202)

/**
 * @macro FM_EINVALID_DEVICE
 *
 * @brief Given device is not registered to FSS.
 * @long This is a status code applicable to the status field of 
 *       the FM_*_REPLY signals.
 *
 * @declaration #define FM_EINVALID_DEVICE 203
 */
#define FM_EINVALID_DEVICE           (203)

/**
 * @macro FM_EINVALID_BLOCK
 *
 * @brief Given block handle does not exist.
 * @long This is a status code applicable to the status field of 
 *       the FM_*_REPLY signals.
 *
 * @declaration #define FM_EINVALID_BLOCK 204
 */
#define FM_EINVALID_BLOCK            (204)

/**
 * @macro FM_EINVALID_LOCK
 *
 * @brief Given lock handle does not exist.
 * @long This is a status code applicable to the status field of 
 *       the FM_*_REPLY signals.
 *
 * @declaration #define FM_EINVALID_LOCK 205
 */
#define FM_EINVALID_LOCK             (205)

/**
 * @macro FM_EINVALID_PARAMETERS
 *
 * @brief At least one parameter in signal is invalid.
 * @long This is a status code applicable to the status field of 
 *       the FM_*_REPLY signals.
 *
 * @declaration #define FM_EINVALID_PARAMETERS 206
 */
#define FM_EINVALID_PARAMETERS       (206)

/**
 * @macro FM_EINVALID_CLIENT
 *
 * @brief Access from the client is not permitted.
 * @long This is a status code applicable to the status field of 
 *       the FM_*_REPLY signals.
 *
 * @declaration #define FM_EINVALID_CLIENT 207
 */
#define FM_EINVALID_CLIENT           (207)

/* Reason why operation is not permitted. */
/**
 * @macro FM_EOPER_NOT_IMPLEMENTED
 *
 * @brief Requested operation is not implemented.
 * @long This is a status code applicable to the status field of 
 *       the FM_*_REPLY signals.
 *
 * @declaration #define FM_EOPER_NOT_IMPLEMENTED 208
 */
#define FM_EOPER_NOT_IMPLEMENTED     (208)

/**
 * @macro FM_EOPER_NOT_SUPPORTED
 *
 * @brief Requested operation is not supported.
 * @long This is a status code applicable to the status field of 
 *       the FM_*_REPLY signals.
 *
 * @declaration #define FM_EOPER_NOT_SUPPORTED 209
 */
#define FM_EOPER_NOT_SUPPORTED       (209)

/**
 * @macro FM_EOPER_CANCELLED
 *
 * @brief The operation has been canceled.
 * @long This is a status code applicable to the status field of 
 *       the FM_*_REPLY signals.
 *
 * @declaration #define FM_EOPER_CANCELLED 210
 */
#define FM_EOPER_CANCELLED           (210)

/**
 * @macro FM_EALREADY_MOUNTED
 *
 * @brief Volume is already mounted. Volume must be unmounted first,
 *        but make sure all open files are closed before unmounting.
 * @long This is a status code applicable to the status field of 
 *       the FM_*_REPLY signals.
 *
 * @declaration #define FM_EALREADY_MOUNTED 211
 */
#define FM_EALREADY_MOUNTED          (211)

/**
 * @macro FM_EDISK_NOT_FORMATTED
 *
 * @brief The disk/volume is not formatted.
 * @long This is a status code applicable to the status field of 
 *       the FM_*_REPLY signals.
 *
 * @declaration #define FM_EDISK_NOT_FORMATTED 212
 */
#define FM_EDISK_NOT_FORMATTED       (212)

/**
 * @macro FM_EDISK_NOT_VALIDATED
 *
 * @brief The disk/volume validation failed. May also indicate a volume that 
 *        is mounted read-only due to validation failure.
 * @long This is a status code applicable to the status field of 
 *       the FM_*_REPLY signals.
 *
 * @declaration #define FM_EDISK_NOT_VALIDATED 213
 */
#define FM_EDISK_NOT_VALIDATED       (213)

/**
 * @macro FM_EDISK_WRITE_PROTECT
 *
 * @brief The disk/volume is write protected, i.e. it has been mounted as 
 *        read-only and an operation that needs to write to disk was tried.
 * @long This is a status code applicable to the status field of 
 *       the FM_*_REPLY signals.
 *
 * @declaration #define FM_EDISK_WRITE_PROTECT 214
 */
#define FM_EDISK_WRITE_PROTECT       (214)

/**
 * @macro FM_EDISK_BUSY
 *
 * @brief Operation can not be completed because it is busy, probably because
 *        there are open files.
 * @long This is a status code applicable to the status field of 
 *       the FM_*_REPLY signals.
 *
 * @declaration #define FM_EDISK_BUSY 215
 */
#define FM_EDISK_BUSY                (215)

/**
 * @macro FM_EDISK_FULL
 *
 * @brief The disk is full.
 * @long This is a status code applicable to the status field of 
 *       the FM_*_REPLY signals.
 *
 * @declaration #define FM_EDISK_FULL 216
 */
#define FM_EDISK_FULL                (216)

/**
 * @macro FM_EOBJECT_DELETE_PROTECT
 *
 * @brief It is not allowed to remove the object.
 * @long This is a status code applicable to the status field of 
 *       the FM_*_REPLY signals.
 *
 * @declaration #define FM_EOBJECT_DELETE_PROTECT 217
 */
#define FM_EOBJECT_DELETE_PROTECT    (217)

/**
 * @macro FM_EOBJECT_WRITE_PROTECT
 *
 * @brief The object is write protected, e.g. opened as read-only.
 * @long This is a status code applicable to the status field of 
 *       the FM_*_REPLY signals.
 *
 * @declaration #define FM_EOBJECT_WRITE_PROTECT 218
 */ 
#define FM_EOBJECT_WRITE_PROTECT     (218)

/**
 * @macro FM_EOBJECT_READ_PROTECT
 *
 * @brief The object is read protected.
 * @long This is a status code applicable to the status field of 
 *       the FM_*_REPLY signals.
 *
 * @declaration #define FM_EOBJECT_READ_PROTECT 219
 */
#define FM_EOBJECT_READ_PROTECT      (219)

/**
 * @macro FM_EOBJECT_EXISTS
 *
 * @brief The object already exists. E.g. when trying to create an object that
 *        already exists.
 * @long This is a status code applicable to the status field of 
 *       the FM_*_REPLY signals.
 *
 * @declaration #define FM_EOBJECT_EXISTS 220
 */
#define FM_EOBJECT_EXISTS            (220)

/**
 * @macro FM_EOBJECT_NOT_FOUND
 *
 * @brief The object is not found.
 * @long This is a status code applicable to the status field of 
 *       the FM_*_REPLY signals.
 *
 * @declaration #define FM_EOBJECT_NOT_FOUND 221
 */
#define FM_EOBJECT_NOT_FOUND         (221)

/**
 * @macro FM_EOBJECT_IN_USE
 *
 * @brief The object is in use. E.g. when trying to remove a file that has
 *        been opened.
 * @long This is a status code applicable to the status field of 
 *       the FM_*_REPLY signals.
 *
 * @declaration #define FM_EOBJECT_IN_USE 222
 */
#define FM_EOBJECT_IN_USE            (222)

/**
 * @macro FM_EOBJECT_TOO_LARGE
 *
 * @brief The object is too large.
 * @long This is a status code applicable to the status field of 
 *       the FM_*_REPLY signals.
 *
 * @declaration #define FM_EOBJECT_TOO_LARGE 223
 */
#define FM_EOBJECT_TOO_LARGE         (223)

/**
 * @macro FM_EOBJECT_WRONG_TYPE
 *
 * @brief The object is of wrong type. E.g. when trying to do a file 
 *        operation on a directory or vice versa.
 * @long This is a status code applicable to the status field of 
 *       the FM_*_REPLY signals.
 *
 * @declaration #define FM_EOBJECT_WRONG_TYPE 224
 */
#define FM_EOBJECT_WRONG_TYPE        (224)

/**
 * @macro FM_EDIR_NOT_EMPTY
 *
 * @brief The directory is not empty. E.g. when trying to remove a directory 
 *        and there are still files in it.
 * @long This is a status code applicable to the status field of 
 *       the FM_*_REPLY signals.
 *
 * @declaration #define FM_EDIR_NOT_EMPTY 225
 */
#define FM_EDIR_NOT_EMPTY            (225)

/**
 * @macro FM_ELOCK_COLLISION
 *
 * @brief File locking collision has occurred.
 * @long This is a status code applicable to the status field of 
 *       the FM_*_REPLY signals.
 *
 * @declaration FM_ELOCK_COLLISION 226
 */
#define FM_ELOCK_COLLISION           (226)

/**
 * @macro FM_EHEAP_EXHAUSTED
 *
 * @brief Not enough space on heap to complete the requested operation. 
 *        Configure more heap.
 * @long This is a status code applicable to the status field of 
 *       the FM_*_REPLY signals.
 *
 * @declaration FM_EHEAP_EXHAUSTED 227
 */
#define FM_EHEAP_EXHAUSTED           (227)

/**
 * @macro FM_EDEVICE_LOST
 *
 * @brief The device driver process and the corresponding volume
 *        manager has been terminated. Until it has been restarted, no more
 *        operations can be done on the media.
 * @long This is a status code applicable to the status field of 
 *       the FM_*_REPLY signals.
 *
 * @declaration FM_EDEVICE_LOST 228
 */
#define FM_EDEVICE_LOST              (228)

/**
 * @macro FM_EFSS_LOST
 *
 * @brief The file system server (FSS) process has been terminated, making 
 *        name resolving impossible. Restart the FSS process to resolve 
 *        the situation.
 * @long This is a status code applicable to the status field of 
 *       the FM_*_REPLY signals.
 *
 * @declaration FM_EFSS_LOST 229
 */
#define FM_EFSS_LOST                 (229)

/**
 * @macro FM_EVM_LOST
 *
 * @brief The volume manager process has been terminated. A new mount may be 
 *        tried, but the file system could be in an unknown state depending 
 *        on the reason for the process to be terminated.
 * @long This is a status code applicable to the status field of 
 *       the FM_*_REPLY signals.
 *
 * @declaration #define FM_EVM_LOST 230
 */
#define FM_EVM_LOST                  (230)

/**
 * @macro FM_EOBJECT_EXECUTE_PROTECT
 *
 * @brief The object can not be executed. For directories this means that it 
 *        is not possible to do a change directory to the directory.
 * @long This is a status code applicable to the status field of 
 *       the FM_*_REPLY signals.
 *
 * @declaration #define FM_EOBJECT_EXECUTE_PROTECT 231
 */
#define FM_EOBJECT_EXECUTE_PROTECT   (231)

/**
 * @macro FM_EOPER_NOT_ALLOWED
 *
 * @brief The operation was not allowed.
 * @long This is a status code applicable to the status field of 
 *       the FM_*_REPLY signals.
 *
 * @declaration #define FM_EOPER_NOT_ALLOWED 232
 */
#define FM_EOPER_NOT_ALLOWED         (232)

/**
 * @macro FM_ENOT_ABSOLUTE_PATH
 *
 * @brief Not an absolute path. E.g. when a volume name was expected, which 
 *        must be specified as an absolute path.
 * @long This is a status code applicable to the status field of 
 *       the FM_*_REPLY signals.
 *
 * @declaration #define FM_ENOT_ABSOLUTE_PATH 233
 */
#define FM_ENOT_ABSOLUTE_PATH        (233)

/**
 * @macro FM_EBAD_ADDRESS
 *
 * @brief Address used in call to read or write is invalid.
 * @long This is a status code applicable to the status field of 
 *       the FM_*_REPLY signals.
 *
 * @declaration #define FM_EBAD_ADDRESS 14
 */
#define FM_EBAD_ADDRESS (14)
/* avoid recompilation of CRT by setting to EFAULT which is 14. */


/**
 * @type FmFadvEnum
 *
 * @include "ose_spi/fm.sig"
 *
 * @brief Values indicating access patterns on an open file handle.
 *
 * @long Which values are supported depends on the volume manager.
 *
 * @field FM_FADV_NORMAL     No specific advice can be given.
 * @field FM_FADV_SEQUENTIAL Sequential access with increasing offsets.
 * @field FM_FADV_RANDOM     Random order access.
 * @field FM_FADV_WILLNEED   The data will soon be accessed.
 * @field FM_FADV_DONTNEED   The data will not be accessed anytime soon.
 * @field FM_FADV_NOREUSE    The data will only be accessed once.
 *
 * @seealso FmFadvise
 */
enum FmFadvEnum
{
   FM_FADV_NORMAL     = 0,
   FM_FADV_SEQUENTIAL = 1,
   FM_FADV_RANDOM     = 2,
   FM_FADV_WILLNEED   = 3,
   FM_FADV_DONTNEED   = 4,
   FM_FADV_NOREUSE    = 5
};


/**
 * @type FmHandle
 *
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * typedef OSADDRESS FmHandle;
 *
 * @brief A type defining a handle (reference) to an open object.
 * @long  All handles managed by a volume manager must be unique, and
 *        no other information than the handle shall be used for
 *        identifying an open object.
 * @long  A volume manager is recommended not to re-use a recently closed
 *        FmHandle number, to avoid any risk of confusing a new handle
 *        with a recently closed handle.
 */
typedef OSADDRESS FmHandle;


/**
 * @type FmVolumeInfo
 *
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * struct FmVolumeInfo
 * {
 *   EfsBoolean bspDevice;
 *   U32        unit;
 *   U32        loBlock;
 *   U32        hiBlock;
 *   U32        freeBlocks;
 *   U32        blockSize;
 *   U32        creationTime;
 *   U32        modificationTime;
 *   EfsBitmask access;
 *   U32        volumeName;
 *   U32        deviceName;
 *   char       s[4];
 *   U32        extensionVersion;
 *   U32        formattedSizeBlocks;
 *   char       actualStringPool[2];
 * };
 *
 * @brief Information about a volume.
 * @long  NOTE: The fields extensionVersion and formattedSizeBlocks only exist
 *        when field volumeName is not 0.
 *
 * @field bspDevice        True if volume is using a BSP device driver,
 *                         False if using an EFS device driver.
 * @field unit             Unit of mounted device, depending on the device
 *                         driver.
 * @field loBlock          Lowest block in mounted range.
 * @field hiBlock          Highest block in mounted range.
 * @field freeBlocks       Number of free blocks. This is 0 on an unformatted
 *                         mounted volume.
 * @field blockSize        Size of block in bytes, depending on the device
 *                         driver.
 * @field creationTime     time() of creation of the file system (when it was
 *                         formatted).
 * @field modificationTime time() of latest modification of any object on
 *                         volume.
 * @field access           Is device readable, writable etc. (@link ""
 *                         "FmAccessEnum"). The access field is 0 on an
 *                         unformatted mounted volume.
 * @field volumeName       Offset in the string pool (field s) to the name
 *                         string for the mounted volume.
 * @field deviceName       Offset in the string pool (field s) to the name
 *                         string for the lower level device driver.
 * @field s                String pool offset base. Storage for strings starts
 *                         at actualStringPool, but the offsets of the strings
 *                         are calculated relative this field. (The purpose 
 *                         of this field is to provide backwards compatibility
 *                         with an earlier version of the signal definition.
 *                         There is no actual data stored in this field.)
 * @field extensionVersion This field is only valid if volumeName != 0. This
 *                         field indicates how much extension information
 *                         exists. Currently, the version number is 1. If this
 *                         field does not exist (volumeName == 0), then
 *                         extension version is 0, meaning no extensions are
 *                         available.
 * @field formattedSizeBlocks This field is only valid if extensionVersion >= 1.
 *                         This field indicates how many blocks are available
 *                         when volume is newly formatted.
 * @field actualStringPool Variable size. Even though string offsets are given
 *                         relative to field s, this field marks where the
 *                         strings will actually be stored. The area between
 *                         field s and this field contains extensions.
 *                         @break This field is of variable size. The size of
 *                         the string(s) must be added when calculating the
 *                         required signal buffer size. Excess space should
 *                         be ignored if the buffer is larger than needed.
 */
struct FmVolumeInfo
{
   EfsBoolean bspDevice;
   U32        unit;
   U32        loBlock;
   U32        hiBlock;
   U32        freeBlocks;
   U32        blockSize;
   U32        creationTime;
   U32        modificationTime;
   EfsBitmask access;
   U32        volumeName;
   U32        deviceName;
   char       s[4];              /* Is 4 to be aligned. */
   /* Following fields belong to the extended version of FmVolumeInfo.
    * These fields only exists if volumeName != 0. */
   U32        extensionVersion;
   /* Fields belonging to extension version >= 1. */
   U32        formattedSizeBlocks;
   /* Reserve space for string terminators and also have a starting point for
    * string storage. */
   char       actualStringPool[2]; /* Must be last! */
};

/**
 * @type FmObjectInfo
 *
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * struct FmObjectInfo
 * {
 *   EfsEnum    type;
 *   EfsBitmask attrs;
 *   U32        diskId;
 *   U32        links;
 *   U32        creationTime;
 *   U32        dataModTime;
 *   U32        headModTime;
 *   U32        accessTime;
 *   U32        owner;
 *   U32        group;
 *   U32        sizeBytes;
 *   U32        sizeBlocks;
 *   U32        bufferMode;
 *   U32        bufferSize;
 *   U32        openCount;
 *   U32        name;
 *   char       s[4];
 *   U32        extensionVersion;
 *   U64        ino;
 *   char       actualStringPool[1];
 * };
 *
 * @brief Information about an object.
 *
 * @long The time fields are used differently by different volume managers.
 *       See the "OSE Core Extensions User's Guide" for information about the
 *       different volume managers.
 *       Examples of dependencies:
 *       JEFF volumes can be configured via mount parameters to use
 *       different ways to treat the access and modification time fields.
 *       VFAT volumes have limited support for the time fields.
 *
 * @field type             Type of object (@link "" "FmTypeEnum").
 * @field attrs            Attributes (@link "" "FmAttrEnum").
 * @field diskId           Uniquely identifying value. If no such value exists,
 *                         this field is 0. If such a value exists, but does not
 *                         fit into this field, the value is instead stored in
 *                         ino, and this field is 0.
 * @field links            Number of hard links to object.
 * @field creationTime     Time of creation/latest status change, i.e.
 *                         POSIX ctime (if implemented in the volume manager).
 *                         Example: VFAT only updates it at creation time.
 * @field dataModTime      Time of latest data modification. Implementation
 *                         defined. Example: With JEFF, a mount option can
 *                         select additional reasons to update this field.
 * @field headModTime      Time of latest header modification. Implementation
 *                         defined, often not used.
 * @field accessTime       Time of latest access (if implemented and enabled
 *                         in the volume manager).
 *                         Example: With JEFF, it is default disabled, but can
 *                         be enabled as a mount option.
 * @field owner            User identity of owner.
 * @field group            Owner group.
 * @field sizeBytes        Size in bytes, max 4G-2 per file.
 * @field sizeBlocks       Size in blocks.
 * @field bufferMode       Preferred setvbuf() buffer mode
 *                         (@link "" "FmBuffEnum").
 * @field bufferSize       Preferred setvbuf() buffer size.
 * @field openCount        Number of opens/dups for this object.
 * @field name             Offset in the string pool (field s) to the name
 *                         string for the object.
 *                         @break The full path of the object is given. This
 *                         starts with the mount point that is registered in
 *                         the FSS, including the leading slash "/".
 * @field s                String pool. This area stores both extensions to this
 *                         signal as well as strings. Storage for strings starts
 *                         at actualStringPool. All offsets for strings are
 *                         calculated from this field.
 * @field extensionVersion This field is only valid if name != 0. This field
 *                         indicates how much extension information that exists.
 *                         Currently, the version number is 1. If this field
 *                         does not exist (name == 0), then extension
 *                         version is 0, meaning no extensions are available.
 * @field ino              This field is only valid if extensionVersion >= 1,
 *                         and diskId is 0.
 *                         Inode number of object (64 bits). If the object does
 *                         not have a unique inode number, this field is 0.
 * @field actualStringPool Variable size. Even though string offsets are given
 *                         relative to field s, this field marks where the
 *                         strings will actually be stored. The area between
 *                         field s and this field contains extensions.
 *                         @break This field is of variable size. The size of
 *                         the string(s) must be added when calculating the
 *                         required signal buffer size. Excess space should
 *                         be ignored if the buffer is larger than needed.
 */
struct FmObjectInfo
{
   EfsEnum    type;
   EfsBitmask attrs;
   U32        diskId;
   U32        links;
   U32        creationTime;
   U32        dataModTime;
   U32        headModTime;
   U32        accessTime;
   U32        owner;
   U32        group;
   U32        sizeBytes;
   U32        sizeBlocks;
   U32        bufferMode;
   U32        bufferSize;
   U32        openCount;
   U32        name;
   char       s[4];              /* Is 4 to be aligned. */
   /* Following fields belong to the extended version of FmObjectInfo.
    * These fields only exists if name != 0. */
   U32        extensionVersion;
   /* Fields belonging to extension version >= 1. */
   U64        ino;
   /* Reserve space for string terminators and also have a starting point for
    * string storage. */
   char       actualStringPool[1]; /* Must be last! */
};

/**
 * @type FmTerminalInfo
 *
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * struct FmTerminalInfo
 * {
 *   EfsBitmask    valid;
 *   U32           readSizeMin;
 *   U32           readTimeOut;
 *   U32           rxBuffer;
 *   U32           rxSpeed;
 *   U32           txSpeed;
 *   EfsBoolean    rxFlow;
 *   EfsBoolean    txFlow;
 *   U8            charBits;
 *   U8            stopBits;
 *   U8            parity;
 *   U8            flowCtrl;
 *   EfsBitmask    flags;
 *   U16           width;
 *   U16           height;
 *   unsigned char ctrlChars[FM_TCC_SIZE];
 *   U16           volumeName;
 *   U16           deviceName;
 *   U16           termType;
 *   char          s[4];
 *   U32           idleTmo;
 *   char          actualStringPool[3];
 * };
 *
 * @brief Information about a terminal.
 *
 * @field valid           Which fields are valid in this structure
 *                        (@link "" "FmTermEnum").
 * @field readSizeMin     Read minimum size, disabled by FM_TIFL_LINE_EDIT
 *                        (@link "" "FmTiflEnum").
 * @field readTimeOut     Read time out (in milliseconds),
 *                        disabled by FM_TIFL_LINE_EDIT
 *                        (@link "" "FmTiflEnum").
 * @field rxBuffer        Size of input buffer.
 * @field rxSpeed         Input baud rate.
 * @field txSpeed         Output baud rate.
 * @field rxFlow          Flow of receive queue. False (0) if stopped.
 * @field txFlow          Flow of transmit queue. False (0) if stopped.
 * @field charBits        5, 6, 7 or 8 bits per character.
 * @field stopBits        1 or 2 stop bits.
 * @field parity          'O'dd, 'E'ven or 'N'o parity.
 * @field flowCtrl        'N'one or 'H'ardware.
 * @field flags           Diverse boolean flags (@link "" "FmTiflEnum").
 * @field width           Width of terminal in characters.
 * @field height          Height of terminal in characters.
 * @field ctrlChars       Control character
 *                        (@link "" "FmTccEnum",@link "" "FM_TCC_DISABLED").
 * @field volumeName      Offset in the string pool to the name string for the
 *                        mounted volume.
 * @field deviceName      Offset in the string pool to the name string for the
 *                        lower level device driver.
 * @field termType        Offset in the string pool to a terminal type string.
 * @field s               String pool offset base. Storage for strings starts
 *                        at actualStringPool, but the offsets of the strings
 *                        are calculated relative this field. (The purpose 
 *                        of this field is to provide backwards compatibility
 *                        with an earlier version of the signal definition.
 *                        There is no actual data stored in this field.)
 * @field idleTmo         Value for the console idle timeout interval.
 * @field actualStringPool  String pool for variable sized data. The strings 
 *                        referred by volumeName, deviceName and termType are 
 *                        stored here. Even though string offsets are given
 *                        relative to field s, this field marks where the
 *                        strings will actually be stored. The area between
 *                        field s and this field contains extensions. This 
 *                        field is of variable size. The size of the string(s) 
 *                        must be added when calculating the required signal 
 *                        buffer size. Excess space should be ignored if the 
 *                        buffer is larger than needed.
 */
struct FmTerminalInfo
{
   EfsBitmask    valid;
   U32           readSizeMin;
   U32           readTimeOut;
   U32           rxBuffer;
   U32           rxSpeed;
   U32           txSpeed;
   EfsBoolean    rxFlow;
   EfsBoolean    txFlow;
   U8            charBits;
   U8            stopBits;
   U8            parity;
   U8            flowCtrl;
   EfsBitmask    flags;
   U16           width;
   U16           height;
   unsigned char ctrlChars[FM_TCC_SIZE];
   U16           volumeName;
   U16           deviceName;
   U16           termType;
   char          s[4];      /*Aligned to 4 bytes*/
   U32           idleTmo;
   char          actualStringPool[3];
};


/*
 *****************************************************************************
 * Signals to a Format Manager process.
 *****************************************************************************
 */

/**
 * @signal FmInterface
 *
 * @include "ose_spi/fm.sig"
 *
 * @brief Get interface information.
 *
 * @long Obtains information about the Format Manager signal interface. Unless
 *       an error reply status is returned, the reply signal includes a list of
 *       supported signals, optionally a BIOS trap handle and a descriptive
 *       string for this type of format manager. A typical whatStr could be
 *       "EFS 3.0.0 ExtFAT Format Manager".
 *
 * @member FmInterfaceRequest
 * @member FmInterfaceReply
 *
 * @seealso sendFmInterfaceRequest
 * @seealso sendFmInterfaceReply
 */

/**
 * @macro FM_INTERFACE_REQUEST
 * @brief Get interface information.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_INTERFACE_REQUEST           30100
 *
 * @seealso FmInterface
 */
#define FM_INTERFACE_REQUEST  (30100) /* !-SIGNO(struct FmInterfaceRequest)-! */

/**
 * @macro FM_INTERFACE_REPLY
 * @brief Get interface information.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_INTERFACE_REPLY             30101
 *
 * @seealso FmInterface
 */
#define FM_INTERFACE_REPLY    (30101)    /* !-SIGNO(struct FmInterfaceReply)-! */

/**
 * @type FmInterfaceRequest
 *
 * @declaration
 * struct FmInterfaceRequest
 * {
 *   SIGSELECT sigNo;
 * };
 *
 * @input sigNo FM_INTERFACE_REQUEST
 */
struct FmInterfaceRequest
{
   SIGSELECT sigNo;
};

/**
 * @type FmInterfaceReply
 *
 * @declaration
 * struct FmInterfaceReply
 * {
 *   SIGSELECT sigNo;
 *   EfsStatus status;
 *   char      whatStr[32];
 *   U32       biosHandle;
 *   SIGSELECT sigs[1];
 * };
 *
 * @output sigNo      FM_INTERFACE_REPLY
 * @output status     EFS_SUCCESS if success, which is required for other
 *                    reply data
 *                    to be valid. Other status codes might appear.
 * @output whatStr    A general text string describing this format manager.
 *                    Maximum
 *                    31 bytes plus a terminating null byte. This is not the
 *                    type string registered in FSS.
 * @output biosHandle Optional BIOS handle for a trap interface, only useable in
 *                    the same system.
 * @output sigs       An array containing supported signals.
 *                    First entry is the number of signals in the array.
 *                         @break This field is of variable size. The size of
 *                         the data must be added when calculating the
 *                         required signal buffer size. Excess space should
 *                         be ignored if the buffer is larger than needed.
 */
struct FmInterfaceReply
{
   SIGSELECT sigNo;
   EfsStatus status;
   char      whatStr[32];
   U32       biosHandle;
   SIGSELECT sigs[1];
};

/**
 * @macro sendFmInterfaceRequest
 *
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define sendFmInterfaceRequest(fmPidMA)
 *
 * @brief Send signal @link "" "FmInterface" request.
 *
 * @input fmPidMA PID of the format manager to send the interface request to.
 *
 * @seealso FmInterface
 * @seealso sendFmInterfaceReply
 */
#define sendFmInterfaceRequest(fmPidMA) \
do { \
	union SIGNAL *sigMA = alloc(sizeof(struct FmInterfaceRequest), \
                                    FM_INTERFACE_REQUEST); \
	send(&sigMA, (fmPidMA)); \
} while (0)

/**
 * @macro sendFmInterfaceReply
 *
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define sendFmInterfaceReply(statusMA, whatMA, handleMA, sigLoMA,
 *                              sigHiMA, fmPidMA)
 *
 * @brief Send signal @link "" "FmInterface" reply.
 *
 * @input statusMA EFS_SUCCESS if success, which is required for other reply
 *                 data to be valid. Other status codes might appear.
 * @input whatMA   A general text string describing this format manager. Maximum
 *                 31 bytes plus a terminating null byte. This is not the type
 *                 string registered in FSS.
 *                 The sent signal's whatStr will not be null terminated if the
 *                 text string has to be truncated.
 * @input handleMA Optional BIOS handle for a trap interface, only useable in
 *                 the same system.
 * @input sigLoMA  Lowest supported signal.
 * @input sigHiMA  Highest supported signal.
 * @input fmPidMA  PID of process to send signal to.
 *
 * @seealso FmInterface
 * @seealso sendFmInterfaceRequest
 */
#define sendFmInterfaceReply(statusMA, whatMA, handleMA, sigLoMA, \
                             sigHiMA, fmPidMA) \
do { \
	/* The union is a trick to avoid breaking the strict-aliasing rules*/\
	union U \
	{ \
		union SIGNAL *s;\
		struct FmInterfaceReply *x;\
	} u;\
	unsigned i; \
	U32 numSigs = ((sigHiMA) - (sigLoMA)) / 2 + 1; \
	struct FmInterfaceReply *sigMA; \
	u.s = alloc(sizeof(struct FmInterfaceReply) + \
                      numSigs * sizeof(SIGSELECT), FM_INTERFACE_REPLY); \
	sigMA = u.x; \
	sigMA->status = (statusMA); \
	strncpy(sigMA->whatStr, (whatMA), sizeof sigMA->whatStr); \
	sigMA->biosHandle = (handleMA); \
	sigMA->sigs[0] = numSigs; \
	for (i = 0; i < numSigs; i++) \
	{ \
		sigMA->sigs[i + 1] = (sigLoMA) + 2 * i; \
	} \
	send(&u.s, (fmPidMA)); \
} while (0)


/**
 * @signal FmMount
 *
 * @include "ose_spi/fm.sig"
 *
 * @brief Mount a new volume.
 *
 * @long This signal is sent to a format manager, which creates a volume
 *       manager and mounts a new volume with given volume name on given
 *       device.
 * @long Device and format manager dependent parameters are given in
 *       a parameter text string with the format:
 * @long "<param name 1>[=<param value 1>][,<param name 2>
 *          [=<param value 2>]]...".
 * @long Note that boolean parameters, i.e. 'param name' without
 *       '=<param value>", may be given.
 * @long A volume might be unformatted when mounted, which means that the volume
 *       will be in unformatted state. When in this state the only things that
 *       can be done on the volume is unmount and format.
 * @long The sender of the reply signal will be the new volume manager process.
 *       All request for this volume must be sent to this process, i.e. make
 *       sure to do "sender(&sig)" to get the PID to be used in future signals.
 *
 * @member FmMount
 *
 * @seealso sendFmMountRequest
 * @seealso FmUnmount
 * @seealso FmFormatVol
 */

/**
 * @macro FM_MOUNT_REQUEST
 * @brief Mount a new volume.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_MOUNT_REQUEST               30102
 *
 * @seealso FmMount
 */
#define FM_MOUNT_REQUEST               (30102)    /* !-SIGNO(struct FmMount)-! */

/**
 * @macro FM_MOUNT_REPLY
 * @brief Mount a new volume.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_MOUNT_REPLY                      30103
 *
 * @seealso FmMount
 */
#define FM_MOUNT_REPLY                 (30103)    /* !-SIGNO(struct FmMount)-! */

/**
 * @type FmMount
 *
 * @declaration
 * struct FmMount
 * {
 *   SIGSELECT sigNo;
 *   EfsStatus status;
 *   U32       volumeName;
 *   U32       deviceName;
 *   U32       parameters;
 *   char      s[3];
 * };
 *
 * @input sigNo      FM_MOUNT_REQUEST
 * @input volumeName Byte offset into string pool where volume name string is
 *                   stored. The volume name is also refered to as the "mount
 *                   point", which must be registered in the FSS during the
 *                   mounting process. The mount point is the absolute path
 *                   name of the root directory of the volume, which must
 *                   begin with a slash "/". All name strings that refer to
 *                   objects in the volume are absolute paths, which begin
 *                   with the mount point.
 * @input deviceName Byte offset into string pool where device name string is
 *                   stored. This must be an existing device driver that the
 *                   volume manager will interface. The device name must be
 *                   registered in the file system server, either as a DDC
 *                   or DDB device depending on which volume manager is
 *                   mounted.
 * @input parameters Byte offset into string pool where extra parameter
 *                   string is stored.
 *                   The parameter string has the following syntax:
 *                   "<param name 1>[=<param value 1>]
 *                   [,<param name 2>[=<param value 2>]]...".
 *                   E.g. "unit=0,lo=0,hi=1024".
 * @input s          String pool for variable sized data. The strings refered
 *                   by volumeName, deviceName and parameters are stored here.
 *                         @break This field is of variable size. The size of
 *                         the string(s) must be added when calculating the
 *                         required signal buffer size. Excess space should
 *                         be ignored if the buffer is larger than needed.
 *
 * @output sigNo     FM_MOUNT_REPLY
 * @output status    EFS_SUCCESS if success, which means that the sender of the
 *                   reply signal is the new volume manager process.
 *                   Preferred @link "errno.h" "EREMOUNT" or old
 *                   @link "" "FM_EALREADY_MOUNTED" if it is
 *                   already mounted.
 *                   Other status codes might appear, e.g. from lower
 *                   levels like the device driver.
 */
struct FmMount
{
   SIGSELECT sigNo;
   EfsStatus status;
   U32       volumeName;
   U32       deviceName;
   U32       parameters;
   char      s[3];
};

/**
 * @macro sendFmMountRequest
 *
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define sendFmMountRequest(_volMA, _devMA, _paramsMA, fmPidMA)
 *
 * @brief Send signal @link "" "FmMount" request.
 *
 * @input _volMA    Volume name string. This is the root directory where the
 *                  volume root is. The volume name string must begin with "/".
 * @input _devMA    Device name string. This must be an existing device driver
 *                  that the volume manager will interface. The device name
 *                  must be registered in the file system server, either as a
 *                  DDC or DDB device depending on which volume manager is
 *                  mounted.
 * @input _paramsMA Extra parameters string.
 *                  The parameter string has the following syntax:
 *                  "<param name 1>[=<param value 1>]
 *                   [,<param name 2>[=<param value 2>]]...".
 *                  E.g. "unit=0,lo=0,hi=1024".
 * @input fmPidMA   PID of format manager to send the mount request to.
 *
 *
 * @seealso FmMount
 * @seealso sendFmUnmountRequest
 *
 * @example
 * // An example mounting a partition (block 0..199) of the device named
 * // 'ramdisk' with 9 extended FAT buffers and 20 directory buffers and
 * // the resulting volume called '/ram'.
 * // This signal is sent directly to the "extfat" format manager, with process
 * // ID fmExtFatPid (in the example) was acquired from file system server.
 * sendFmMountRequest("/ram",
 *                    "ramdisk",
 *                    "unit=0,lo=0,hi=199,fatbuffers=9,dirbuffers=20",
 *                    fmExtFatPid)
 */
#define sendFmMountRequest(_volMA, _devMA, _paramsMA, fmPidMA) \
do { \
	/* The union is a trick to avoid breaking the strict-aliasing rules*/\
	union U \
	{ \
		union SIGNAL *s;\
		struct FmMount *x;\
	} u;\
	struct FmMount *sigMA; \
	const char *volMA = (_volMA); \
	const char *devMA = (_devMA); \
	const char *paramsMA = (_paramsMA); \
	U32 volLenMA = strlen(volMA); \
	U32 devLenMA = strlen(devMA); \
	U32 paramsLenMA = strlen(paramsMA); \
	u.s = alloc(sizeof(struct FmMount) + volLenMA + devLenMA + paramsLenMA,\
                    FM_MOUNT_REQUEST); \
	sigMA = u.x; \
	sigMA->volumeName = 0; \
	strcpy(&sigMA->s[sigMA->volumeName], volMA); \
	sigMA->deviceName = volLenMA + 1; \
	strcpy(&sigMA->s[sigMA->deviceName], devMA); \
	sigMA->parameters = volLenMA + 1 + devLenMA + 1; \
	strcpy(&sigMA->s[sigMA->parameters], paramsMA); \
	send(&u.s, (fmPidMA)); \
} while (0)

/*
 *****************************************************************************
 * Signals to a mounted Volume Manager.
 *****************************************************************************
 */

/**
 * @signal FmUnmount
 *
 * @include "ose_spi/fm.sig"
 *
 * @brief Unmount a mounted volume.
 *
 * @long This signal is sent to the volume manager process. The volume manager
 *       PID can be retrieved by taking the sender of the mount reply signal.
 * @long The volume manager will terminate after having written any pending
 *       volume buffers to the device driver, closed the volume and sent the
 *       reply signal. When the volume manager terminates, it is unregistered
 *       by the file system server.
 * @long If force is False (0) and any files are still open on the volume,
 *       the volume manager refuses to be unmounted and returns status
 *       preferred @link "errno.h" "EBUSY" or old
 *       @link "" "FM_EDISK_BUSY".
 *       If force is True (1), open files are discarded.
 *
 * @member FmUnmount
 *
 * @seealso sendFmUnmountRequest
 * @seealso FmMount
 */

/**
 * @macro FM_UNMOUNT_REQUEST
 * @brief Unmount a mounted volume.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_UNMOUNT_REQUEST             30104
 *
 * @seealso FmUnmount
 */
#define FM_UNMOUNT_REQUEST       (30104)    /* !-SIGNO(struct FmUnmount)-! */

/**
 * @macro FM_UNMOUNT_REPLY
 * @brief Unmount a mounted volume.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_UNMOUNT_REPLY               30105
 *
 * @seealso FmUnmount
 */
#define FM_UNMOUNT_REPLY         (30105)    /* !-SIGNO(struct FmUnmount)-! */

/**
 * @type FmUnmount
 *
 * @declaration
 * struct FmUnmount
 * {
 *   SIGSELECT  sigNo;
 *   EfsStatus  status;
 *   EfsBoolean force;
 *   U32        volumeName;
 *   char       s[1];
 * };
 *
 * @input sigNo      FM_UNMOUNT_REQUEST
 * @input force      If true (1) unmount will be done even if there are open
 *                   files.
 * @input volumeName Offset (always 0) in the string pool to a volume name
 *                   string identifying the volume.
 * @input s          String pool for variable sized data. The string refered
 *                   by volumeName is stored here.
 *                         @break This field is of variable size. The size of
 *                         the string(s) must be added when calculating the
 *                         required signal buffer size. Excess space should
 *                         be ignored if the buffer is larger than needed.
 *
 * @output sigNo     FM_UNMOUNT_REPLY
 * @output status    EFS_SUCCESS if success.
 *                   Preferred @link "errno.h" "EBUSY" or old
 *                   @link "" "FM_EDISK_BUSY" if there are files
 *                   open on the volume and the force argument is false.
 *                   Other status codes might appear.
 */
struct FmUnmount
{
   SIGSELECT  sigNo;
   EfsStatus  status;
   EfsBoolean force;
   U32        volumeName;
   char       s[1];
};

/**
 * @macro sendFmUnmountRequest
 *
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define sendFmUnmountRequest(forceMA, _volumeNameMA, fmPidMA)
 *
 * @brief Send signal @link "" "FmUnmount" request.
 *
 * @input forceMA       If true (1) unmount will be done even if there are
 *                      open files.
 * @input _volumeNameMA Volume name string.
 * @input fmPidMA       PID of the volume manager process to send signal to.
 *
 * @seealso FmUnmount
 * @seealso FmMount
 * @seealso sendFmMountRequest
 *
 * @example sendUnmountRequest(False, "/ram", vmPID);
 */

#define sendFmUnmountRequest(forceMA, _volumeNameMA, fmPidMA) \
do { \
	/* The union is a trick to avoid breaking the strict-aliasing rules*/\
	union U \
	{ \
		union SIGNAL *s;\
		struct FmUnmount *x;\
	} u;\
	const char *volumeNameMA = (_volumeNameMA); \
	struct FmUnmount *sigMA; \
	u.s = alloc(sizeof(struct FmUnmount) + strlen(volumeNameMA), \
                    FM_UNMOUNT_REQUEST); \
	sigMA = u.x; \
	sigMA->force = (forceMA); \
	sigMA->volumeName = 0; \
	strcpy(&sigMA->s[0], volumeNameMA); \
	send(&u.s, (fmPidMA)); \
} while (0)


/**
 * @signal FmFormatVol
 *
 * @include "ose_spi/fm.sig"
 *
 * @brief Format a mounted volume.
 *
 * @long This signal is sent to a volume manager process. The volume manager
 *       PID can be retrieved by taking the sender of the mount reply signal.
 * @long Device and volume manager dependent parameters are given in an extra
 *       parameter text string.
 * @long If there are open files on the volume and force is False (0), the
 *       volume manager refuses and replies with status preferred
 *       @link "errno.h" "EBUSY" or old @link "" "FM_EDISK_BUSY".
 *
 * @member FmFormatVol
 *
 * @seealso sendFmFormatVolRequest
 * @seealso FmMount
 * @seealso FmUnmount
 */

/**
 * @macro FM_FORMAT_VOL_REQUEST
 * @brief Format a mounted volume.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_FORMAT_VOL_REQUEST          (30106)
 *
 * @seealso FmFormatVol
 */
#define FM_FORMAT_VOL_REQUEST    (30106)    /* !-SIGNO(struct FmFormatVol)-! */

/**
 * @macro FM_FORMAT_VOL_REPLY
 * @brief Format a mounted volume.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_FORMAT_VOL_REPLY            30107
 *
 * @seealso FmFormatVol
 */
#define FM_FORMAT_VOL_REPLY      (30107)    /* !-SIGNO(struct FmFormatVol)-! */

/**
 * @type FmFormatVol
 *
 * @declaration
 * struct FmFormatVol
 * {
 *   SIGSELECT  sigNo;
 *   EfsStatus  status;
 *   EfsBoolean verify;
 *   EfsBoolean force;
 *   U32        parameters;
 *   char       s[1];
 * };
 *
 * @input sigNo      FM_FORMAT_VOL_REQUEST
 * @input verify     If True (1) write will be verified.
 *                   How the verification is implemented depends on the
 *                   volume manager.
 * @input force      If True (1) format is done even if there are open files.
 * @input parameters Byte offset (always 0) into the string pool where extra
 *                   parameters string is stored. What parameters to give
 *                   depends on volume manager and device driver.
 * @input s          String pool for variable sized data. The string refered
 *                   by parameters is stored here.
 *                         @break This field is of variable size. The size of
 *                         the string(s) must be added when calculating the
 *                         required signal buffer size. Excess space should
 *                         be ignored if the buffer is larger than needed.
 *
 * @output sigNo     FM_FORMAT_VOL_REPLY
 * @output status    EFS_SUCCESS if success.
 *                   Preferred @link "errno.h" "EBUSY" or old
 *                   @link "" "FM_EDISK_BUSY" if there are open
 *                   files and force is False (0).
 *                   Other status codes might appear, e.g. coming from
 *                   device driver.
 */
struct FmFormatVol
{
   SIGSELECT  sigNo;
   EfsStatus  status;
   EfsBoolean verify;
   EfsBoolean force;
   U32        parameters;
   char       s[1];
};

/**
 * @macro sendFmFormatVolRequest
 *
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define sendFmFormatVolRequest(verifyMA, forceMA, parametersMA,fmPidMA)
 *
 * @brief Send signal @link "" "FmFormatVol" request.
 *
 * @input verifyMA     If True (1) write will be verified.
 *                     How the verification is implemented depends on the
 *                     volume manager.
 * @input forceMA      If True (1) format is done even if there are open files.
 * @input parametersMA Extra parameters string. What parameters to give
 *                     depends on volume manager and device driver.
 * @input fmPidMA      PID of volume manager process to send the request to.
 *
 * @seealso FmFormat
 * @seealso FmUnmount
 *
 * @example sendFmFormatVolRequest(False, False, "unit=0,partition=1", vmPID);
 */
#define sendFmFormatVolRequest(verifyMA, forceMA, parametersMA,fmPidMA) \
do { \
	/* The union is a trick to avoid breaking the strict-aliasing rules*/\
	union U \
	{ \
		union SIGNAL *s;\
		struct FmFormatVol *x;\
	} u;\
	const char *_parameters = (parametersMA); \
	struct FmFormatVol *sigMA; \
	u.s = alloc(sizeof(struct FmFormatVol) + strlen(_parameters), \
                    FM_FORMAT_VOL_REQUEST); \
	sigMA = u.x; \
	sigMA->verify = (verifyMA); \
	sigMA->force = (forceMA); \
	sigMA->parameters = 0; \
	strcpy(&sigMA->s[0], _parameters); \
	send(&u.s, (fmPidMA)); \
} while (0)


/**
 * @signal FmValidateVol
 *
 * @include "ose_spi/fm.sig"
 *
 * @brief Validate and optionally repair the file structure on a mounted volume.
 *
 * @long This signal is sent to a volume manager process. The volume manager
 *       PID can be retrieved by taking the sender of the mount reply signal.
 * @long If repair is set to False (0), the validation is performed without
 *       changing
 *       the volume and the number of problems is reported in the
 *       remainProblems field.
 * @long If repair is set to True (1), attempts are done to repair any problems
 *       and the
 *       number of repaired problems as well as any number of problems that
 *       could not
 *       be corrected are reported.
 *
 * @member FmValidateVol
 *
 * @seealso sendValidateVol
 * @seealso FmFormat
 * @seealso FmMount
 */

/**
 * @macro FM_VALIDATE_VOL_REQUEST
 * @brief Validate and optionally repair the file structure on a mounted volume.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_VALIDATE_VOL_REQUEST        30108
 *
 * @seealso FmValidateVol
 */
#define FM_VALIDATE_VOL_REQUEST  (30108)    /* !-SIGNO(struct FmValidateVol)-! */

/**
 * @macro FM_VALIDATE_VOL_REPLY
 * @brief Validate and optionally repair the file structure on a mounted volume.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_VALIDATE_VOL_REPLY          30109
 *
 * @seealso FmValidateVol
 */
#define FM_VALIDATE_VOL_REPLY    (30109)    /* !-SIGNO(struct FmValidateVol)-! */

/**
 * @type FmValidateVolRepairBits
 *
 * @brief Bit values that can be OR'ed together for @link "" "FmValidateVol"
 *        field repair.
 *
 * @field FM_VALIDATE_VOL_REPAIR  Request that validate tries to repair found
 *                                problems.
 * @field FM_VALIDATE_VOL_REPORT  Request that validate sends
 *                                @link "" "FmValidateVolMsg" to report
 *                                progress.
 */
enum FmValidateVolRepairBits
{
   FM_VALIDATE_VOL_REPAIR = 1,
   FM_VALIDATE_VOL_REPORT = 2
};

/**
 * @type FmValidateVol
 *
 * @declaration
 * struct FmValidateVol
 * {
 *   SIGSELECT  sigNo;
 *   EfsStatus  status;
 *   EfsBitmask repair;
 *   U32        fixedProblems;
 *   U32        remainProblems;
 * };
 *
 * @input sigNo           FM_VALIDATE_VOL_REQUEST
 * @input repair          See @link "" "FmValidateVolRepairBits" for valid
 *                        values.
 *                        If no bit is set, validate will perform validation
 *                        only.
 *
 * @output sigNo          FM_VALIDATE_VOL_REPLY
 * @output status         EFS_SUCCESS if success (other reply data valid).
 *                        Preferred @link "errno.h" "EROFS" or old
 *                        @link "" "FM_EDISK_WRITE_PROTECT" if
 *                        the volume is read-only.
 *                        Preferred @link "errno.h" "EACCES" or old
 *                        @link "" "FM_EDISK_NOT_FORMATTED" if
 *                        the volume is not recognized as formatted.
 *                        Other status codes might appear, e.g. coming
 *                        from device driver.
 * @output fixedProblems  Number of corrected volume problems.
 * @output remainProblems Number of remaining volume problems.
 */
struct FmValidateVol
{
   SIGSELECT  sigNo;
   EfsStatus  status;
   EfsBitmask repair;
   U32        fixedProblems;
   U32        remainProblems;
};

/**
 * @macro sendFmValidateVolRequest
 *
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define sendFmValidateVolRequest(repairMA, fmPidMA)
 *
 * @brief Send signal @link "" "FmValidateVol" request.
 *
 * @input repairMA If True (1), it also attempts to repair any damaged
 *                 structure. Otherwise it only checks the volume.
 * @input fmPidMA  PID of volume manager process to send the request to.
 *
 * @seealso FmValidateVol
 *
 * @example sendFmValidateVolRequest(True, vmPID);
 */
#define sendFmValidateVolRequest(repairMA, fmPidMA) \
do { \
	/* The union is a trick to avoid breaking the strict-aliasing rules*/\
	union U \
	{ \
		union SIGNAL *s;\
		struct FmValidateVol *x;\
	} u;\
	struct FmValidateVol *sigMA; \
	u.s = alloc(sizeof(struct FmValidateVol), FM_VALIDATE_VOL_REQUEST); \
	sigMA = u.x; \
	sigMA->repair = (repairMA); \
	send(&u.s, (fmPidMA)); \
} while (0)


/**
 * @signal FmRelabelVol
 *
 * @include "ose_spi/fm.sig"
 *
 * @brief Set a new persistent volume label (name string).
 *
 * @long This signal is sent to a volume manager process. The volume manager
 *       PID can be retrieved by taking the sender of the mount reply signal.
 * @long For easy administration it is recommended to use the volume name as the
 *       label when possible.
 *
 * @member FmRelabelVol
 *
 * @seealso FmMount
 * @seealso sendFmRelabelVolRequest
 */

/**
 * @macro FM_RELABEL_VOL_REQUEST
 * @brief Set a new persistent volume label (name string).
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_RELABEL_VOL_REQUEST         30110
 *
 * @seealso FmRelabelVol
 */
#define FM_RELABEL_VOL_REQUEST   (30110)    /* !-SIGNO(struct FmRelabelVol)-! */

/**
 * @macro FM_RELABEL_VOL_REPLY
 * @brief Set a new persistent volume label (name string).
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_RELABEL_VOL_REPLY           30111
 *
 * @seealso FmRelabelVol
 */
#define FM_RELABEL_VOL_REPLY     (30111)    /* !-SIGNO(struct FmRelabelVol)-! */

/**
 * @type FmRelabelVol
 *
 * @declaration
 * struct FmRelabelVol
 * {
 *   SIGSELECT sigNo;
 *   EfsStatus status;
 *   U32       newLabel;
 *   char      s[1];
 * };
 *
 * @input sigNo    FM_RELABEL_VOL_REQUEST
 * @input newLabel Byte offset into string pool (always 0) where the new label
 *                 string is stored.
 * @input s        String pool for variable sized data. The string
 *                 refered by newLabel is stored here.
 *                         @break This field is of variable size. The size of
 *                         the string(s) must be added when calculating the
 *                         required signal buffer size. Excess space should
 *                         be ignored if the buffer is larger than needed.
 *
 * @output sigNo   FM_RELABEL_VOL_REPLY
 * @output status  EFS_SUCCESS if success.
 *                 Preferred @link "errno.h" "ENOTSUP" or old
 *                 @link "" "FM_EOPER_NOT_SUPPORTED" if the
 *                 volume does not support a persistent label.
 *                 Other status codes might appear, e.g. coming from
 *                 device driver.
 */
struct FmRelabelVol
{
   SIGSELECT sigNo;
   EfsStatus status;
   U32       newLabel;
   char      s[1];
};

/**
 * @macro sendFmRelabelVolRequest
 *
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define sendFmRelabelVolRequest(_newLabelMA, fmPidMA)
 *
 * @brief Send signal @link "" "FmRelabelVol" request.
 *
 * @input _newLabelMA New label string to be set on volume.
 * @input fmPidMA     PID of volume manager to send the request to.
 *
 * @seealso FmRelabelVol
 *
 * @example sendFmRelabelVolRequest("/ram", vmPID);
 */
#define sendFmRelabelVolRequest(_newLabelMA, fmPidMA) \
do { \
	/* The union is a trick to avoid breaking the strict-aliasing rules*/\
	union U \
	{ \
		union SIGNAL *s;\
		struct FmRelabelVol *x;\
	} u;\
	struct FmRelabelVol *sigMA; \
        const char *newLabelMA = (_newLabelMA); \
        u.s = alloc(sizeof(struct FmRelabelVol) + strlen(newLabelMA), \
                    FM_RELABEL_VOL_REQUEST); \
	sigMA = u.x; \
	sigMA->newLabel = 0; \
	strcpy(&sigMA->s[0], newLabelMA); \
	send(&u.s, (fmPidMA)); \
} while (0)


/**
 * @signal FmSyncVol
 *
 * @include "ose_spi/fm.sig"
 *
 * @brief Write pending (not-yet-written) data buffers for the volume to the
 *        physical storage.
 *
 * @long This signal is sent to a volume manager process. The volume manager
 *       PID can be retrieved by taking the sender of the mount reply signal.
 * @long The reply signal is not sent until all data pending for write has been
 *       written to the physical storage.
 * @long NOTE: To really force all data to be written, first force any process
 *       space cache out-buffers to the volume.
 *
 * @member FmSyncVol
 *
 * @seealso FmMount
 * @seealso sendFmSyncVol
 */

/**
 * @macro FM_SYNC_VOL_REQUEST
 * @brief Write pending (not-yet-written) data buffers for the volume to the
 *        physical storage.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_SYNC_VOL_REQUEST            30112
 *
 * @seealso FmSyncVol
 */
#define FM_SYNC_VOL_REQUEST      (30112)    /* !-SIGNO(struct FmSyncVol)-! */

/**
 * @macro FM_SYNC_VOL_REPLY
 * @brief Write pending (not-yet-written) data buffers for the volume to the
 *        physical storage.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_SYNC_VOL_REPLY              30113
 *
 * @seealso FmSyncVol
 */
#define FM_SYNC_VOL_REPLY        (30113)    /* !-SIGNO(struct FmSyncVol)-! */

/**
 * @type FmSyncVol
 *
 * @declaration
 * struct FmSyncVol
 * {
 *   SIGSELECT sigNo;
 *   EfsStatus status;
 * };
 *
 * @input sigNo   FM_SYNC_VOL_REQUEST
 *
 * @output sigNo  FM_SYNC_VOL_REPLY
 * @output status EFS_SUCCESS if success.
 *                Preferred @link "errno.h" "ENOSPC" or old
 *                @link "" "FM_EDISK_FULL" if there is not
 *                enough space on the volume.
 *                Other status codes might appear, e.g. coming from
 *                device driver.
 */
struct FmSyncVol
{
   SIGSELECT sigNo;
   EfsStatus status;
};

/**
 * @macro sendFmSyncVolRequest
 *
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define sendFmSyncVolRequest(fmPidMA)
 *
 * @brief Send signal @link "" "FmSyncVol" request.
 *
 * @input fmPidMA PID of volume manager process to send the request to.
 *
 * @seealso FmSyncVol
 */
#define sendFmSyncVolRequest(fmPidMA) \
do { \
        union SIGNAL *sigMA = alloc(sizeof(struct FmSyncVol), \
                                    FM_SYNC_VOL_REQUEST); \
        send(&sigMA, (fmPidMA)); \
} while (0)


/**
 * @signal FmExamineVol
 *
 * @include "ose_spi/fm.sig"
 *
 * @brief Request information about a volume.
 *
 * @long This signal is sent to a volume manager process. The volume manager
 *       PID can be retrieved by taking the sender of the mount reply signal.
 * @long The reply contains an @link "" "FmVolumeInfo" structure.
 * @long Also an unformatted volume will return an @link "" "FmVolumeInfo"
 *       structure but the access field will be 0. Other fields in the structure
 *       are irrelevant in this case.
 *
 * @member FmExamineVolRequest
 * @member FmExamineVolReply
 *
 * @seealso sendFmExamineVolRequest
 * @seealso FmVolumeInfo
 * @seealso FmMount
 */

/**
 * @macro FM_EXAMINE_VOL_REQUEST
 * @brief Request information about a volume.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_EXAMINE_VOL_REQUEST         30114
 *
 * @seealso FmExamineVol
 */
#define FM_EXAMINE_VOL_REQUEST (30114) /* !-SIGNO(struct FmExamineVolRequest)-! */

/**
 * @macro FM_EXAMINE_VOL_REPLY
 * @brief Request information about a volume.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_EXAMINE_VOL_REPLY           30115
 *
 * @seealso FmExamineVol
 */
#define FM_EXAMINE_VOL_REPLY   (30115)  /* !-SIGNO(struct FmExamineVolReply)-! */

/**
 * @type FmExamineVolRequest
 *
 * @declaration
 * struct FmExamineVolRequest
 * {
 *   SIGSELECT sigNo;
 * };
 *
 * @input sigNo FM_EXAMINE_VOL_REQUEST
 */
struct FmExamineVolRequest
{
   SIGSELECT sigNo;
};

/**
 * @type FmExamineVolReply
 *
 * @declaration
 * struct FmExamineVolReply
 * {
 *   SIGSELECT           sigNo;
 *   EfsStatus           status;
 *   struct FmVolumeInfo info;
 * };
 *
 * @output sigNo  FM_EXAMINE_VOL_REPLY
 * @output status EFS_SUCCESS if success (other reply data valid).
 *                Other status codes might appear, e.g. coming from
 *                device driver.
 * @output info   An @link "" "FmVolumeInfo" structure describing the volume.
 *                Valid only if returned status is EFS_SUCCESS.
 */
struct FmExamineVolReply
{
   SIGSELECT           sigNo;
   EfsStatus           status;
   struct FmVolumeInfo info;
};

/**
 * @macro sendFmExamineVolRequest
 *
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define sendFmExamineVolRequest(fmPidMA)
 *
 * @brief Send signal @link "" "FmExamineVol" request.
 *
 * @input fmPidMA PID of volume manager to send the request to.
 *
 * @seealso FmExamineVol
 */
#define sendFmExamineVolRequest(fmPidMA) \
do { \
        union SIGNAL *sigMA = alloc(sizeof(struct FmExamineVolRequest), \
                                    FM_EXAMINE_VOL_REQUEST); \
        send(&sigMA, (fmPidMA)); \
} while (0)

/**
 * @macro FM_VALIDATE_VOL_MSG_REQUEST
 * @brief Request to handle a validation report message.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_VALIDATE_VOL_MSG_REQUEST        30116
 *
 * @seealso FmValidateVolMsg
 */
#define FM_VALIDATE_VOL_MSG_REQUEST          (30116)

/**
 * @macro FM_VALIDATE_VOL_MSG_CFM
 * @brief Acknowledge of handled validation report message.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_VALIDATE_VOL_MSG_CFM            30117
 *
 * @seealso FmValidateVolMsg
 */
#define FM_VALIDATE_VOL_MSG_CFM              (30117)

/**
 * @type FmValidateVolMsg
 *
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * struct FmValidateVolMsg
 * {
 *     SIGSELECT                      sigNo;
 *     U32                            reportSeverity;
 *     U32                            count;
 *     U8                             symbolData[1];
 * };
 *
 * @brief Message with information from validate volume
 *
 * @long This signal interface is used for sending reports with additional
 *       information generated during a "validate volume" operation.
 *
 * @long The client requests generation of reports by setting the 
 *       FM_VALIDATE_VOL_REPORT flag in @link "" "FmValidateVol". The
 *       volume manager will then send back FM_VALIDATE_VOL_MSG_REQUEST
 *       signals to the client during the "validate volume" operation. The
 *       client must acknowledge each signal with FM_VALIDATE_VOL_MSG_CFM. 
 *       The sending of reports may continue until the "validate volume"
 *       operation is concluded with @link "" "FM_VALIDATE_VOL_REPLY".
 *
 * @long Each of these signals contain a "report", consisting of a report
 *       name followed by a list of records with additional information.
 *
 * @long symbolData contains a report name as a null terminated UTF-8
 *       string, immediately followed by a packed list of records. The
 *       number of records is described in the "count" field.
 *
 * @long Each record is a packed triplet of type, name and value, formatted
 *       as follows:@break
 * @tab type @tab A byte (U8), describing how to decode the "value" field.
 *                The code values are described in to @link ""
 *                "FmValidateSymbolType". @break
 * @tab name @tab A null terminated string of UTF-8 characters, containing
 *                the name of the record.@break
 * @tab value @tab The value of the record, packed and encoded as an array
 *                of bytes (U8). The encoding and the length of this field
 *                is described by the "type" field. See @link ""
 *                "FmValidateSymbolType" for the format of each type. Binary
 *                values are ordered according to the endian convention of
 *                the sender. Note that the value does not have natural
 *                alignment since it is packed immediately after the previous
 *                field. @break
 *
 * @output sigNo           FM_VALIDATE_VOL_MSG_REQUEST
 * @output reportSeverity  See @link "" "FmValidateSeverity" for valid values.
 * @output count           No. of records in the string symbolData
 * @output symbolData      The information from the file system is packed in
 *                         this Array. See more information above about the
 *                         contents of symbolData.
 *                         @break This field is of variable size. The size of
 *                         the data must be added when calculating the
 *                         required signal buffer size. Excess space should
 *                         be ignored if the buffer is larger than needed.
 *
 * @input sigNo FM_VALIDATE_VOL_MSG_CFM. This signal is
 *              used to confirm retrieval of FM_VALIDATE_VOL_MSG_REQUEST
 *
 * @example
 * The raw encoding of symbolData for:
 * - one record (count=1)
 * - symbol type FM_VALIDATE_SYMBOL_U64 (type=8)
 *
 * "File system summary"            report name, followed by null-terminator
 *                                  byte
 * 0x8                              U8 type FM_VALIDATE_SYMBOL_U64
 *                                  (unsigned 64-bit integer)
 * "TotalDirectoryCount"            name, followed by null-terminator byte
 * 0x1,0x0,0x0,0x0,0x0,0x0,0x0,0x0  value 1 (64 bit, this example is in
 *                                  little-endian order)
 */
struct FmValidateVolMsg
{
      SIGSELECT                      sigNo;
      U32                            reportSeverity;
      U32                            count;
      U8                             symbolData[1];
};

/**
 * @type FmValidateSeverity
 * @brief Valid severity level values for field reportSeverity in
 *        @link "" "FmValidateVolMsg"
 *
 * @field FM_VALIDATE_INFORMATIONAL Informational message
 * @field FM_VALIDATE_WARNING       Warning message
 * @field FM_VALIDATE_ERROR         Error message
 */
enum FmValidateSeverity
{
   FM_VALIDATE_INFORMATIONAL = 0,
   FM_VALIDATE_WARNING = 1,
   FM_VALIDATE_ERROR = 2
};

/**
 * @type FmValidateSymbolType
 * @brief Valid symbol type values for symbolType in @link "" "FmValidateVolMsg"
 *
 * @field FM_VALIDATE_SYMBOL_NONE    Symbol has no data
 * @field FM_VALIDATE_SYMBOL_S8      Symbol data is signed 1 byte
 * @field FM_VALIDATE_SYMBOL_U8      Symbol data is unsigned 1 byte
 * @field FM_VALIDATE_SYMBOL_S16     Symbol data is signed 2 bytes
 * @field FM_VALIDATE_SYMBOL_U16     Symbol data is unsigned 2 bytes
 * @field FM_VALIDATE_SYMBOL_S32     Symbol data is signed 4 bytes
 * @field FM_VALIDATE_SYMBOL_U32     Symbol data is unsigned 4 bytes
 * @field FM_VALIDATE_SYMBOL_S64     Symbol data is signed 8 bytes
 * @field FM_VALIDATE_SYMBOL_U64     Symbol data is unsigned 8 bytes
 * @field FM_VALIDATE_SYMBOL_FLOAT32 Symbol data is float 4 bytes
 * @field FM_VALIDATE_SYMBOL_FLOAT64 Symbol data is float 8 bytes
 * @field FM_VALIDATE_SYMBOL_STR     Symbol data is a 0 terminated C string
 *                                   using UTF8.
 */
enum FmValidateSymbolType
{
   FM_VALIDATE_SYMBOL_NONE = 0,
   FM_VALIDATE_SYMBOL_S8 = 1,
   FM_VALIDATE_SYMBOL_U8 = 2,
   FM_VALIDATE_SYMBOL_S16 = 3,
   FM_VALIDATE_SYMBOL_U16 = 4,
   FM_VALIDATE_SYMBOL_S32 = 5,
   FM_VALIDATE_SYMBOL_U32 = 6,
   FM_VALIDATE_SYMBOL_S64 = 7,
   FM_VALIDATE_SYMBOL_U64 = 8,
   FM_VALIDATE_SYMBOL_FLOAT32 = 9,
   FM_VALIDATE_SYMBOL_FLOAT64 = 10,
   FM_VALIDATE_SYMBOL_STR = 11
};

/**
 * @macro FM_GET_LABEL_VOL_REQUEST
 * @brief Get the volume label.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_GET_LABEL_VOL_REQUEST       30118
 *
 * @seealso sendFmGetLabelVolRequest
 */
#define FM_GET_LABEL_VOL_REQUEST  (30118)   /* !-SIGNO(struct FmGetLabelVolRequest)-! */

/**
 * @type FmGetLabelVolRequest
 * @brief Get the volume label.
 * @declaration
 * struct FmGetLabelVolRequest
 * {
 *   SIGSELECT sigNo;
 * };
 *
 * @input sigNo FM_GET_LABEL_VOL_REQUEST
 *
 * @seealso sendFmGetLabelVol
 */

struct FmGetLabelVolRequest
{
   SIGSELECT sigNo;
};

/**
 * @macro sendFmGetLabelVolRequest
 * @brief Get the volume label.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define sendFmGetLabelVolRequest(fmPidMA)
 *
 * @brief Send signal @link "" "FmGetLabelVolRequest" request.
 *
 * @input fmPidMA     PID of volume manager to send the request to.
 *
 * @seealso FmGetLabelVolRequest
 *
 * @example sendFmGetLabelVolRequest(vmPID);
 */
#define sendFmGetLabelVolRequest(fmPidMA) \
do { \
        union SIGNAL *sigMA = alloc(sizeof(struct FmGetLabelVolRequest), \
                                    FM_GET_LABEL_VOL_REQUEST); \
        send(&sigMA, (fmPidMA)); \
} while (0)

/**
 * @macro FM_GET_LABEL_VOL_REPLY
 * @brief Get the volume label.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_GET_LABEL_VOL_REPLY       30119
 *
 * @seealso sendFmGetLabelVolReply
 */
#define FM_GET_LABEL_VOL_REPLY (30119) /* !-SIGNO(struct FmGetLabelVolReply)-! */

/**
 * @type FmGetLabelVolReply
 * @brief Get the volume label.
 * @declaration
 * struct FmGetLabelVolReply
 * {
 *    SIGSELECT sigNo;
 *    EfsStatus status;
 *    char      label[1];
 * }
 *
 * @output  sigNo     FM_GET_LABEL_VOL_REPLY
 * @output  status    EFS_SUCCESS in case of success
 * @output  label     This is a null terminated string that represents
 *                    the volume's label.
 *                         @break This field is of variable size. The size of
 *                         the string(s) must be added when calculating the
 *                         required signal buffer size. Excess space should
 *                         be ignored if the buffer is larger than needed.
 *
 * @seealso sendFmGetLabelVol
 */

struct FmGetLabelVolReply
{
   SIGSELECT sigNo;
   EfsStatus status;
   char      label[1]; /* must be at least 1*/
};


/*
 ****************************************************************************
 * Signals to a Volume Manager, operating on objects.
 ****************************************************************************
 */

/**
 * @signal FmCreateObj
 *
 * @include "ose_spi/fm.sig"
 *
 * @brief Create a new object or a hard link on a volume.
 *
 * @long This signal is sent to a volume manager process. The volume manager
 *       PID can be retrieved by taking the sender of the mount reply signal.
 * @long The object is not open after the creation; thus any open modes in
 *       modes or amode is ignored. Use @link "" "FmOpen" to open it.
 *
 * @member FmCreateObj
 *
 * @seealso sendFmCreateObjRequest
 * @seealso FmOpen
 * @seealso FmMount
 */

/**
 * @macro FM_CREATE_OBJ_REQUEST
 * @brief Create a new object on a volume.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_CREATE_OBJ_REQUEST          30120
 *
 * @seealso FmCreateObj
 */
#define FM_CREATE_OBJ_REQUEST    (30120)    /* !-SIGNO(struct FmCreateObj)-! */

/**
 * @macro FM_CREATE_OBJ_REPLY
 * @brief Create a new object on a volume.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_CREATE_OBJ_REPLY            30121
 *
 * @seealso FmCreateObj
 */
#define FM_CREATE_OBJ_REPLY      (30121)    /* !-SIGNO(struct FmCreateObj)-! */

/**
 * @type FmCreateObj
 *
 * @declaration
 * struct FmCreateObj
 * {
 *   SIGSELECT  sigNo;
 *   EfsStatus  status;
 *   EfsBitmask modes;
 *   EfsBitmask attrs;
 *   U32        name;
 *   U32        amode;
 *   char       s[2];
 * };
 *
 * @long       Note that some field names in this signal may seem confusing.
 *             The naming originates from a time before a Posix interface
 *             was created for the OSE file system.
 *
 * @input sigNo   FM_CREATE_OBJ_REQUEST
 *
 * @input modes       Open mode. Indicates what type of file to create. The
 *                    bits are as defined in @link "" "FmModeEnum".
 *
 * @input attrs       Creation attributes. This is equivalent to "mode"
 *                    argument of the @link "fcntl.h" "creat"() call, but the
 *                    bits are in different positions as defined in
 *                    @link "" "FmAttrEnum". This field is only applicable if
 *                    file creation is requested i the modes field.
 *
 * @input name        Byte offset into string pool where the object name
 *                    of the file to create is stored.
 *                    @break The full path of the object is given. This
 *                    starts with the mount point that is registered in
 *                    the FSS, including the leading slash "/".
 *
 * @input amode       Byte offset into string pool where an optional extra
 *                    string can be stored. @break
 *
 *                    When creating a hard link, i.e. when modes includes
 *                    FM_MODE_HLINK, This string names the preexisting file
 *                    that the new link shall reference.
 *                    @break The full path of the object is given. This
 *                    starts with the mount point that is registered in
 *                    the FSS, including the leading slash "/". @break
 *
 *                    When not creating a link, Most volume managers
 *                    ignore this string or require it to
 *                    be empty. The original intention of this string was to
 *                    support the mode string from the @link "stdio.h" "fopen"()
 *                    call, but there are no OSE volume managers that accept it
 *                    for that purpose.
 *
 * @input s           String pool for variable sized data. The strings
 *                    refered by name and amode are stored here.
 *                    @break This field is of variable size. The size of
 *                    the string(s) must be added when calculating the
 *                    required signal buffer size. Excess space should
 *                    be ignored if the buffer is larger than needed.
 *
 * @output sigNo  FM_CREATE_OBJ_REPLY
 * @output status EFS_SUCCESS if success.
 *                Preferred @link "errno.h" "EROFS" or old
 *                @link "" "FM_EDISK_WRITE_PROTECT" if the
 *                volume is write protected.
 *                Other status codes might appear, e.g. coming from
 *                device driver.
 */
struct FmCreateObj
{
   SIGSELECT  sigNo;
   EfsStatus  status;
   EfsBitmask modes;
   EfsBitmask attrs;
   U32        name;
   U32        amode;
   char       s[2];
};

/**
 * @macro sendFmCreateObjRequest
 *
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define sendFmCreateObjRequest(modesMA, attrsMA, nameMA, amodeMA, fmPidMA)
 *
 * @brief Send signal @link "" "FmCreateObj" request.
 *
 * @input modesMA    Value for the modes field in @link "" "FmCreateObj"
 * @input attrsMA    Value for the attrs field in @link "" "FmCreateObj"
 * @input nameMA     String for the name field in @link "" "FmCreateObj"
 * @input amodeMA    String for the amode field in @link "" "FmCreateObj"
 * @input fmPidMA PID of the volume manager process to send the request to.
 *
 * @seealso FmCreateObj
 *
 * @example
 * sendFmCreateObjRequest(FM_MODE_READ | FM_MODE_WRITE | FM_MODE_TEXT,
 *                        FM_ATTR_READ | FM_ATTR_WRITE,
 *                        "/testfile.txt",
 *                        "r",
 *                        vmPID);
 */
#define sendFmCreateObjRequest(modesMA, attrsMA, nameMA, amodeMA, fmPidMA) \
do { \
	/* The union is a trick to avoid breaking the strict-aliasing rules*/\
	union U \
	{ \
		union SIGNAL *s;\
		struct FmCreateObj *x;\
	} u;\
	struct FmCreateObj *sigMA; \
        const char *_nameMA = (nameMA); \
        const char *_amodeMA = (amodeMA); \
        unsigned l0 = strlen(_nameMA); \
        unsigned l1 = strlen(_amodeMA); \
        u.s = alloc(sizeof(struct FmCreateObj) + l0 + l1, \
                    FM_CREATE_OBJ_REQUEST); \
	sigMA = u.x; \
        sigMA->modes = (modesMA); \
        sigMA->attrs = (attrsMA); \
        sigMA->name = 0; \
        sigMA->amode = l0 + 1; \
        strcpy(&sigMA->s[sigMA->name], (_nameMA)); \
        strcpy(&sigMA->s[sigMA->amode], (_amodeMA)); \
        send(&u.s, (fmPidMA)); \
} while (0)


/**
 * @signal FmRemoveObj
 *
 * @include "ose_spi/fm.sig"
 *
 * @brief Remove an object from a volume.
 *
 * @long This signal is sent to a volume manager process. The volume manager
 *       PID can be retrieved by taking the sender of the mount reply signal.
 * @long The object is typically a file or directory. The object name is a full
 *       path, i.e. starts with "/".
 * @long If force argument is True (1) the object will be removed regardless of
 *       if it is open or write protected. Directories must still be empty even
 *       if force is True.
 *       NOTE: If the volume was mounted read-only objects can not be removed
 *       even if this argument is True.
 *
 * @member FmRemoveObj
 *
 * @seealso FmMount
 * @seealso sendFmRemoveObjRequest
 */

/**
 * @macro FM_REMOVE_OBJ_REQUEST
 * @brief Remove an object from a volume.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_REMOVE_OBJ_REQUEST          30122
 *
 * @seealso FmRemoveObj
 */
#define FM_REMOVE_OBJ_REQUEST    (30122)    /* !-SIGNO(struct FmRemoveObj)-! */

/**
 * @macro FM_REMOVE_OBJ_REPLY
 * @brief Remove an object from a volume.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_REMOVE_OBJ_REPLY            30123
 *
 * @seealso FmRemoveObj
 */
#define FM_REMOVE_OBJ_REPLY      (30123)    /* !-SIGNO(struct FmRemoveObj)-! */

/**
 * @type FmRemoveObj
 *
 * @declaration
 * struct FmRemoveObj
 * {
 *   SIGSELECT  sigNo;
 *   EfsStatus  status;
 *   EfsBitmask types;
 *   EfsBoolean force;
 *   U32        name;
 *   char       s[1];
 * };
 *
 * @input sigNo   FM_REMOVE_OBJ_REQUEST
 * @input types   Type of object to be removed. This value is calculated as
 *                1 << type (@link "" "FmTypeEnum").
 *                Several values can be OR'ed together to select several types.
 * @input force   If True (1) remove object even if it is write protected or
 *                opened. Note that if the volume was mounted read-only objects
 *                can not be removed even if this argument is True.
 * @input name    Byte offset into string pool where the name string of the
 *                object is stored.
 *                         @break The full path of the object is given. This
 *                         starts with the mount point that is registered in
 *                         the FSS, including the leading slash "/".
 * @input s       String pool for variable sized data. The string
 *                refered by name is stored here.
 *                         @break This field is of variable size. The size of
 *                         the string(s) must be added when calculating the
 *                         required signal buffer size. Excess space should
 *                         be ignored if the buffer is larger than needed.
 *
 * @output sigNo  FM_REMOVE_OBJ_REPLY
 * @output status EFS_SUCCESS if success.
 *                Preferred @link "errno.h" "EACCES" or old
 *                @link "" "FM_EOBJECT_IN_USE" if the object has
 *                been  opened and force argument is False (0).
 *                Preferred @link "errno.h" "ENOTEMPTY" or old
 *                @link "" "FM_EDIR_NOT_EMPTY" if trying to
 *                remove a non-empty directory.
 *                Preferred @link "errno.h" "EINVAL" or old
 *                @link "" "FM_EOBJECT_WRONG_TYPE" if name
 *                argument refers to an object of a different type than given
 *                by types argument.
 *                Preferred @link "errno.h" "EROFS" or old
 *                @link "" "FM_EDISK_WRITE_PROTECT" if the
 *                volume is write protected.
 *                Other status codes might appear, e.g. from lower
 *                levels like the device driver.
 */
struct FmRemoveObj
{
   SIGSELECT  sigNo;
   EfsStatus  status;
   EfsBitmask types;
   EfsBoolean force;
   U32        name;
   char       s[1];
};

/**
 * @macro sendFmRemoveObjRequest
 *
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define sendFmRemoveObjRequest(typesMA, forceMA, nameMA, fmPidMA)
 *
 * @brief Send signal @link "" "FmRemoveObj" request.
 *
 * @input typesMA Type of object to be removed, see @link "" "FmTypeEnum".
 *                This value is calculated as 1 << type (@link "" "FmTypeEnum").
 *                Several values can be OR'ed together to select several types.
 * @input forceMA If True (1) remove object even if it is write protected or
 *                opened. Note that if the volume was mounted read-only objects
 *                can not be removed even if this argument is True.
 * @input nameMA  Object name.
 *                         @break The full path of the object is given. This
 *                         starts with the mount point that is registered in
 *                         the FSS, including the leading slash "/".
 * @input fmPidMA PID of volume manager process to send the request to.
 *
 * @seealso FmRemoveObj
 *
 * @example sendFmRemoveObjRequest((1 << FM_TYPE_DIR), True, "/mydir", vmPID);
 */
#define sendFmRemoveObjRequest(typesMA, forceMA, nameMA, fmPidMA) \
do { \
	/* The union is a trick to avoid breaking the strict-aliasing rules*/\
	union U \
	{ \
		union SIGNAL *s;\
		struct FmRemoveObj *x;\
	} u;\
	struct FmRemoveObj *sigMA; \
        const char *_nameMA = (nameMA); \
        u.s = alloc(sizeof(struct FmRemoveObj) + strlen(_nameMA), \
                    FM_REMOVE_OBJ_REQUEST); \
	sigMA = u.x; \
        sigMA->types = (typesMA); \
        sigMA->force = (forceMA); \
        sigMA->name = 0; \
        strcpy(&sigMA->s[sigMA->name], (_nameMA)); \
        send(&u.s, (fmPidMA)); \
} while (0)


/**
 * @signal FmRenameObj
 *
 * @include "ose_spi/fm.sig"
 *
 * @brief Rename an object in a volume.
 *
 * @long This signal is sent to a volume manager process. The volume manager
 *       PID can be retrieved by taking the sender of the mount reply signal.
 * @long Object is typically a directory or a file. Any existing object with
 *       the new name is removed.
 *                         @break The full path of the object is given. This
 *                         starts with the mount point that is registered in
 *                         the FSS, including the leading slash "/".
 * @long This operation cannot move objects between different volumes.
 * @long Some volume managers require that both names are within the same
 *       directory.
 * @long Some volume managers cannot remane objects that are open.
 *
 * @member FmRenameObj
 *
 * @seealso FmMount
 * @seealso sendFmRenameObjRequest
 */

/**
 * @macro FM_RENAME_OBJ_REQUEST
 * @brief Rename an object in a volume.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_RENAME_OBJ_REQUEST          30124
 *
 * @seealso FmRenameObj
 */
#define FM_RENAME_OBJ_REQUEST       (30124)    /* !-SIGNO(struct FmRenameObj)-! */

/**
 * @macro FM_RENAME_OBJ_REPLY
 * @brief Rename an object in a volume.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_RENAME_OBJ_REPLY            30125
 *
 * @seealso FmRenameObj
 */
#define FM_RENAME_OBJ_REPLY         (30125)    /* !-SIGNO(struct FmRenameObj)-! */

/**
 * @type FmRenameObj
 *
 * @declaration
 * struct FmRenameObj
 * {
 *   SIGSELECT sigNo;
 *   EfsStatus status;
 *   U32       oldName;
 *   U32       newName;
 *   char      s[2];
 * };
 *
 * @input sigNo   FM_RENAME_OBJ_REQUEST
 * @input oldName Byte offset into string pool where the old object name is
 *                stored.
 *                         @break The full path of the object is given. This
 *                         starts with the mount point that is registered in
 *                         the FSS, including the leading slash "/".
 * @input newName Byte offset into string pool where the new object name is
 *                stored.
 *                         @break The full path of the object is given. This
 *                         starts with the mount point that is registered in
 *                         the FSS, including the leading slash "/".
 * @input s       String pool for variable sized data. The strings
 *                refered by oldName and newName are stored here.
 *                         @break This field is of variable size. The size of
 *                         the string(s) must be added when calculating the
 *                         required signal buffer size. Excess space should
 *                         be ignored if the buffer is larger than needed.
 *
 * @output sigNo  FM_RENAME_OBJ_REPLY
 * @output status EFS_SUCCESS if success.
 *                Preferred @link "errno.h" "EACCES" or old
 *                @link "" "FM_EOBJECT_IN_USE" if the old object
 *                is open or if the new object exists and is open.
 *                Preferred @link "errno.h" "EINVAL" or old 
 *                @link "" "FM_EINVALID_NAME" if the new or old name are invalid,
 *                maybe not beginning with "/".
 *                Preferred @link "errno.h" "EXDEV" or old
 *                @link "" "FM_EOPER_NOT_IMPLEMENTED" can occur
 *                when rename tries to move the object between volumes or if
 *                volume manager does not support moving the object with rename.
 *                Preferred @link "errno.h" "EROFS" or old
 *                @link "" "FM_EDISK_WRITE_PROTECT" if the volume
 *                is write protected.
 *                Other status codes might appear, e.g. from lower
 *                levels like the device driver.
 */
struct FmRenameObj
{
   SIGSELECT sigNo;
   EfsStatus status;
   U32       oldName;
   U32       newName;
   char      s[2];
};

/**
 * @macro sendFmRenameObjRequest
 *
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define sendFmRenameObjRequest(oldNameMA, newNameMA, fmPidMA)
 *
 * @brief Send signal @link "" "FmRenameObj" request.
 *
 * @input oldNameMA Old object name.
 *                         @break The full path of the object is given. This
 *                         starts with the mount point that is registered in
 *                         the FSS, including the leading slash "/".
 * @input newNameMA New object name.
 *                         @break The full path of the object is given. This
 *                         starts with the mount point that is registered in
 *                         the FSS, including the leading slash "/".
 * @input fmPidMA   PID of volume manager process to send the request to.
 *
 * @seealso FmRenameObj
 *
 * @example sendFmRenameObjRequest("/mydir/text.txt", "/newtext.txt", vmPID);
 */
#define sendFmRenameObjRequest(oldNameMA, newNameMA, fmPidMA) \
do { \
	/* The union is a trick to avoid breaking the strict-aliasing rules*/\
	union U \
	{ \
		union SIGNAL *s;\
		struct FmRenameObj *x;\
	} u;\
	struct FmRenameObj *sigMA; \
        const char *_oldNameMA = (oldNameMA); \
        const char *_newNameMA = (newNameMA); \
        unsigned l0 = strlen(_oldNameMA); \
        unsigned l1 = strlen(_newNameMA); \
        u.s = alloc(sizeof(struct FmRenameObj) + l0 + l1, \
                    FM_RENAME_OBJ_REQUEST); \
	sigMA = u.x; \
        sigMA->oldName = 0; \
        sigMA->newName = l0 + 1; \
        strcpy(&sigMA->s[sigMA->oldName], (_oldNameMA)); \
        strcpy(&sigMA->s[sigMA->newName], (_newNameMA)); \
        send(&u.s, (fmPidMA)); \
} while (0)



/**
 * @signal FmExamineObj
 *
 * @include "ose_spi/fm.sig"
 *
 * @brief Request information about an object in a volume.
 *
 * @long This signal is sent to a volume manager process. The volume manager
 *       PID can be retrieved by taking the sender of the mount reply signal.
 * @long The reply contains an @link "" "FmObjectInfo" structure.
 * @long Object can be examined regardless of whether it is opened or not.
 *
 * @member FmExamineObjRequest
 * @member FmExamineObjReply
 *
 * @seealso sendFmExamineObjRequest
 * @seealso FmMount
 */

/**
 * @macro FM_EXAMINE_OBJ_REQUEST
 * @brief Request information about an object in a volume.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_EXAMINE_OBJ_REQUEST         30126
 *
 * @seealso FmExamineObj
 */
#define FM_EXAMINE_OBJ_REQUEST (30126) /* !-SIGNO(struct FmExamineObjRequest)-! */

/**
 * @macro FM_EXAMINE_OBJ_REPLY
 * @brief Information about an object in a volume.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_EXAMINE_OBJ_REPLY           30127
 *
 * @seealso FmExamineObj
 */
#define FM_EXAMINE_OBJ_REPLY  (30127) /* !-SIGNO(struct FmExamineObjReply)-! */

/**
 * @type FmExamineObjRequest
 *
 * @declaration
 * struct FmExamineObjRequest
 * {
 *   SIGSELECT sigNo;
 *   U32       name;
 *   char      s[1];
 * };
 *
 * @input sigNo FM_EXAMINE_OBJ_REQUEST
 * @input name  Byte offset into string pool where the name of the object is
 *              stored.
 *                         @break The full path of the object is given. This
 *                         starts with the mount point that is registered in
 *                         the FSS, including the leading slash "/".
 * @input s     String pool for variable sized data. The string
 *              refered by name is stored here.
 *                         @break This field is of variable size. The size of
 *                         the string(s) must be added when calculating the
 *                         required signal buffer size. Excess space should
 *                         be ignored if the buffer is larger than needed.
 */
struct FmExamineObjRequest
{
   SIGSELECT sigNo;
   U32       name;
   char      s[1];
};

/**
 * @type FmExamineObjReply
 *
 * @declaration
 * struct FmExamineObjReply
 * {
 *   SIGSELECT           sigNo;
 *   EfsStatus           status;
 *   struct FmObjectInfo info;
 * };
 *
 * @output sigNo  FM_EXAMINE_OBJ_REPLY
 * @output status EFS_SUCCESS if success (other reply data valid).
 *                Other status codes might appear, e.g. coming from
 *                device driver.
 * @output info   An @link "" "FmObjectInfo" structure describing the object.
 *                Only valid if status is EFS_SUCCESS.
 */
struct FmExamineObjReply
{
   SIGSELECT           sigNo;
   EfsStatus           status;
   struct FmObjectInfo info;
};

/**
 * @macro sendFmExamineObjRequest
 *
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define sendFmExamineObjRequest(nameMA, fmPidMA)
 *
 * @brief Send signal @link "" "FmExamineVol" request.
 *
 * @input nameMA  Name of the object.
 *                         @break The full path of the object is given. This
 *                         starts with the mount point that is registered in
 *                         the FSS, including the leading slash "/".
 * @input fmPidMA PID of volume manager process to send the request to.
 *
 * @seealso FmExamineObj
 *
 * @example sendFmExamineObjRequest("/mydir/myfile.txt", vmPID);
 */
#define sendFmExamineObjRequest(nameMA, fmPidMA) \
do { \
	/* The union is a trick to avoid breaking the strict-aliasing rules*/\
	union U \
	{ \
		union SIGNAL *s;\
		struct FmExamineObjRequest *x;\
	} u;\
	struct FmExamineObjRequest *sigMA; \
        const char *_nameMA = (nameMA); \
        u.s = alloc(sizeof(struct FmExamineObjRequest) + strlen(_nameMA), \
                    FM_EXAMINE_OBJ_REQUEST); \
	sigMA = u.x; \
        sigMA->name = 0; \
        strcpy(&sigMA->s[sigMA->name], (_nameMA)); \
        send(&u.s, (fmPidMA)); \
} while (0)



/**
 * @signal FmSetSizeObj
 *
 * @include "ose_spi/fm.sig"
 *
 * @brief Resize object in a volume.
 *
 * @long This signal is sent to a volume manager process. The volume manager
 *       PID can be retrieved by taking the sender of the mount reply signal.
 * @long Object is typically a file. The new size in bytes can be any value
 *       between current size and zero, including. Some volume managers might
 *       support setting a file size larger than the current file size.
 * @long The object must not be opened when resizing.
 *
 * @member FmSetSizeObj
 *
 * @seealso sendFmSetSizeObjRequest
 * @seealso FmMount
 */

/**
 * @macro FM_SET_SIZE_OBJ_REQUEST
 * @brief Resize object in a volume.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_SET_SIZE_OBJ_REQUEST        30130
 *
 * @seealso FmSetSizeObj
 */
#define FM_SET_SIZE_OBJ_REQUEST  (30130)    /* !-SIGNO(struct FmSetSizeObj)-! */

/**
 * @macro FM_SET_SIZE_OBJ_REPLY
 * @brief Resize object in a volume.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_SET_SIZE_OBJ_REPLY          30131
 *
 * @seealso FmSetSizeObj
 */
#define FM_SET_SIZE_OBJ_REPLY    (30131)    /* !-SIGNO(struct FmSetSizeObj)-! */

/**
 * @type FmSetSizeObj
 *
 * @declaration
 * struct FmSetSizeObj
 * {
 *   SIGSELECT sigNo;
 *   EfsStatus status;
 *   U32       size;
 *   U32       name;
 *   char      s[1];
 * };
 *
 * @input sigNo   FM_SET_SIZE_OBJ_REQUEST
 * @input size    New size of the object in number of bytes. Can be anywhere
 *                between current file size and zero, including. Some volume
 *                managers might support larger sizes.
 * @input name    Byte offset into string pool where the name of the object is
 *                stored.
 *                         @break The full path of the object is given. This
 *                         starts with the mount point that is registered in
 *                         the FSS, including the leading slash "/".
 * @input s       String pool for variable sized data. The string
 *                refered by name is stored here.
 *                         @break This field is of variable size. The size of
 *                         the string(s) must be added when calculating the
 *                         required signal buffer size. Excess space should
 *                         be ignored if the buffer is larger than needed.
 *
 * @output sigNo  FM_SET_SIZE_OBJ_REPLY
 * @output status EFS_SUCCESS if success.
 *                Preferred @link "errno.h" "ENOSYS" or old
 *                FM_EOPER_NOT_IMPLEMENTED if trying to extend a file when
 *                the volume manager does not support this.
 *                Other status codes might appear, e.g. coming from
 *                device driver.
 */
struct FmSetSizeObj
{
   SIGSELECT sigNo;
   EfsStatus status;
   U32       size;
   U32       name;
   char      s[1];
};

/**
 * @macro sendFmSetSizeObjRequest
 *
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define sendFmSetSizeObjRequest(sizeMA, nameMA, fmPidMA)
 *
 * @brief Send signal @link "" "FmSyncVol" request.
 *
 * @input sizeMA  New size of the object in number of bytes. Can be anywhere
 *                between current file size and zero, including.
 * @input nameMA  Name of the object to resize.
 *                         @break The full path of the object is given. This
 *                         starts with the mount point that is registered in
 *                         the FSS, including the leading slash "/".
 * @input fmPidMA PID of the volume manager to send the request to.
 *
 * @seealso FmSetSizeObj
 *
 * @example sendFmSetSizeObjRequest(0, "/mytextfile", vmPID);
 */
#define sendFmSetSizeObjRequest(sizeMA, nameMA, fmPidMA) \
do { \
	/* The union is a trick to avoid breaking the strict-aliasing rules*/\
	union U \
	{ \
		union SIGNAL *s;\
		struct FmSetSizeObj *x;\
	} u;\
	struct FmSetSizeObj *sigMA; \
        const char *_nameMA = (nameMA); \
        u.s = alloc(sizeof(struct FmSetSizeObj) + strlen(_nameMA), \
                    FM_SET_SIZE_OBJ_REQUEST); \
	sigMA = u.x; \
        sigMA->size = (sizeMA); \
        sigMA->name = 0; \
        strcpy(&sigMA->s[sigMA->name], (_nameMA)); \
        send(&u.s, (fmPidMA)); \
} while (0)

/**
 * @signal FmSetAttrObj
 *
 * @include "ose_spi/fm.sig"
 *
 * @brief Change permission attribute bits for an object in a volume.
 *
 * @long This signal is sent to a volume manager process. The volume manager
 *       PID can be retrieved by taking the sender of the mount reply signal.
 * @long The mask argument specifies which permission bits to modify, and attrs
 *       argument specifies the new value for these bits.
 *       Depending on the volume manager, only some of the attributes might be
 *       supported.
 *
 * @member FmSetAttrObj
 *
 * @seealso sendFmSetAttrObjRequest
 * @seealso FmMount
 */

/**
 * @macro FM_SET_ATTR_OBJ_REQUEST
 * @brief Change permission attribute bits for an object in a volume.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_SET_ATTR_OBJ_REQUEST        30132
 *
 * @seealso FmSetAttrObj
 */
#define FM_SET_ATTR_OBJ_REQUEST  (30132)    /* !-SIGNO(struct FmSetAttrObj)-! */

/**
 * @macro FM_SET_ATTR_OBJ_REPLY
 * @brief Change permission attribute bits for an object in a volume.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_SET_ATTR_OBJ_REPLY          30133
 *
 * @seealso FmSetAttrObj
 */
#define FM_SET_ATTR_OBJ_REPLY    (30133)    /* !-SIGNO(struct FmSetAttrObj)-! */

/**
 * @type FmSetAttrObj
 *
 * @declaration
 * struct FmSetAttrObj
 * {
 *   SIGSELECT  sigNo;
 *   EfsStatus  status;
 *   EfsBitmask mask;
 *   EfsBitmask attrs;
 *   U32        name;
 *   char       s[1];
 * };
 *
 * @input sigNo   FM_SET_ATTR_OBJ_REQUEST
 * @input mask    A bit field describing which bits in argument attrs will be
 *                changed. If a bit is set (1) the corresponding bit in argument
 *                attrs will be changed to whatever the bit value is in attrs.
 *                Use bit values from @link "" "FmAttrEnum" to select which
 *                attributes to change.
 * @input attrs   Use bit values from @link "" "FmAttrEnum" to select which
 *                attributes to set (1). If a bit value is set in argument mask
 *                but not in attrs, the attribute will be cleared (0), while if
 *                it is set in both mask and attrs, the attribute will be set
 *                (1).
 * @input name    Byte offset into string pool where the object name is stored.
 *                         @break The full path of the object is given. This
 *                         starts with the mount point that is registered in
 *                         the FSS, including the leading slash "/".
 * @input s       String pool for variable sized data. The string
 *                refered by name is stored here.
 *                         @break This field is of variable size. The size of
 *                         the string(s) must be added when calculating the
 *                         required signal buffer size. Excess space should
 *                         be ignored if the buffer is larger than needed.
 *
 * @output sigNo  FM_SET_ATTR_OBJ_REPLY
 * @output status EFS_STATUS if success.
 *                Preferred @link "errno.h" "EROFS" or old
 *                @link "" "FM_EDISK_WRITE_PROTECT" if the volume
 *                is write protected.
 *                Other status codes might appear, e.g. from lower
 *                levels like the device driver.
 */
struct FmSetAttrObj
{
   SIGSELECT  sigNo;
   EfsStatus  status;
   EfsBitmask mask;
   EfsBitmask attrs;
   U32        name;
   char       s[1];
};

/**
 * @macro sendFmSetAttrObjRequest
 *
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define sendFmSetAttrObjRequest(maskMA, attrsMA, nameMA, fmPidMA)
 *
 * @brief Send signal @link "" "FmSetAttrObj" request.
 *
 * @input maskMA  A bit field describing which bits in argument attrs will be
 *                changed. If a bit is set (1) the corresponding bit in argument
 *                attrs will be changed to whatever the bit value is in attrs.
 *                Use bit values from @link "" "FmAttrEnum" to select which
 *                attributes to change.
 * @input attrsMA Use bit values from @link "" "FmAttrEnum" to select which
 *                attributes to set (1). If a bit value is set in argument mask
 *                but not in attrs, the attribute will be cleared (0), while if
 *                it is set in both mask and attrs, the attribute will be set
 *                (1).
 * @input nameMA  Name of object whose attributes is to be changed.
 *                         @break The full path of the object is given. This
 *                         starts with the mount point that is registered in
 *                         the FSS, including the leading slash "/".
 * @input fmPidMA PID of volume manager process to send the request to.
 *
 * @seealso FmSetAttrObj
 *
 * @example
 * // Remove write permission for file myfile.txt, with the PID of the volume
 * // manager process stored in variable vmPID:
 * sendFmSetAttrObjRequest(FM_ATTR_WRITE, 0, "/myfile.txt", vmPID);
 */
#define sendFmSetAttrObjRequest(maskMA, attrsMA, nameMA, fmPidMA) \
do { \
	/* The union is a trick to avoid breaking the strict-aliasing rules*/\
	union U \
	{ \
		union SIGNAL *s;\
		struct FmSetAttrObj *x;\
	} u;\
	struct FmSetAttrObj *sigMA; \
        const char *_nameMA = (nameMA); \
        u.s = alloc(sizeof(struct FmSetAttrObj) + strlen(_nameMA), \
                           FM_SET_ATTR_OBJ_REQUEST); \
	sigMA = u.x; \
        sigMA->mask = (maskMA); \
        sigMA->attrs = (attrsMA); \
        sigMA->name = 0; \
        strcpy(&sigMA->s[sigMA->name], (_nameMA)); \
        send(&u.s, (fmPidMA)); \
} while (0)


/**
 * @signal FmSetTimeObj
 *
 * @include "ose_spi/fm.sig"
 *
 * @brief Change timestamp of an object in a volume.
 *
 * @long This signal is sent to a volume manager process. The volume manager
 *       PID can be retrieved by taking the sender of the mount reply signal.
 * @long The object is usually a file or a directory. The mask can select
 *       several timestamps which means that all these timestamps will be
 *       set with the same time. Depending on the volume manager, only some
 *       of the timestamps might be implemented.
 *
 * @member FmSetTimeObj
 *
 * @seealso sendFmSetTimeObjRequest
 * @seealso FmMount
 */

/**
 * @macro FM_SET_TIME_OBJ_REQUEST
 * @brief Change timestamp of an object in a volume.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_SET_TIME_OBJ_REQUEST        30134
 *
 * @seealso FmSetTimeObj
 */
#define FM_SET_TIME_OBJ_REQUEST  (30134)    /* !-SIGNO(struct FmSetTimeObj)-! */

/**
 * @macro FM_SET_TIME_OBJ_REPLY
 * @brief Change timestamp of an object in a volume.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_SET_TIME_OBJ_REPLY          30135
 *
 * @seealso FmSetTimeObj
 */
#define FM_SET_TIME_OBJ_REPLY    (30135)    /* !-SIGNO(struct FmSetTimeObj)-! */

/**
 * @type FmSetTimeObj
 *
 * @declaration
 * struct FmSetTimeObj
 * {
 *   SIGSELECT  sigNo;
 *   EfsStatus  status;
 *   EfsBitmask mask;
 *   U32        seconds;
 *   U32        micros;
 *   U32        name;
 *   char       s[1];
 * };
 *
 * @input sigNo   FM_SET_TIME_OBJ_REQUEST
 * @input mask    One or several values from @link "" "FmTimeEnum" OR'ed
 *                together to select which timestamp(s) to set.
 * @input seconds The seconds part of the time. Usually starts from 1 January
 *                1970, but this is volume manager dependent.
 * @input micros  The microseconds part of the time. This argument might be
 *                ignored by the volume manager if it does not support this.
 * @input name    Byte offset into string pool where the object name is stored.
 *                         @break The full path of the object is given. This
 *                         starts with the mount point that is registered in
 *                         the FSS, including the leading slash "/".
 * @input s       String pool for variable sized data. The string
 *                refered by name is stored here.
 *                         @break This field is of variable size. The size of
 *                         the string(s) must be added when calculating the
 *                         required signal buffer size. Excess space should
 *                         be ignored if the buffer is larger than needed.
 *
 * @output sigNo  FM_SET_TIME_OBJ_REPLY
 * @output status EFS_SUCCESS if success.
 *                Preferred @link "errno.h" "EROFS" or old
 *                @link "" "FM_EDISK_WRITE_PROTECT" if the
 *                volume is write protected.
 *                Other status codes might appear, e.g. coming from
 *                device driver.
 */
struct FmSetTimeObj
{
   SIGSELECT  sigNo;
   EfsStatus  status;
   EfsBitmask mask;
   U32        seconds;
   U32        micros;
   U32        name;
   char       s[1];
};

/**
 * @macro sendFmSetTimeObjRequest
 *
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define sendFmSetTimeObjRequest(maskMA, secsMA, microsMA, nameMA, fmPidMA)
 *
 * @brief Send signal @link "" "FmSetTimeObj" request.
 *
 * @input maskMA   One or several values from @link "" "FmTimeEnum" OR'ed
 *                 together to select which timestamp(s) to set.
 * @input secsMA   The seconds part of the time. Usually starts from 1 January
 *                 1970, but this is volume manager dependent.
 * @input microsMA The microseconds part of the time. This argument might be
 *                 ignored by the volume manager if it does not support this.
 * @input nameMA   Name of the object.
 *                         @break The full path of the object is given. This
 *                         starts with the mount point that is registered in
 *                         the FSS, including the leading slash "/".
 * @input fmPidMA  PID of volume manager to send the request to.
 *
 * @seealso FmSetTimeObj
 *
 * @example sendFmSetTimeObjRequest(FM_TIME_ACCESS, 20000, 300, "/myfile",
 *                                  vmPID);
 */
#define sendFmSetTimeObjRequest(maskMA, secsMA, microsMA, nameMA, fmPidMA) \
do { \
	/* The union is a trick to avoid breaking the strict-aliasing rules*/\
	union U \
	{ \
		union SIGNAL *s;\
		struct FmSetTimeObj *x;\
	} u;\
	struct FmSetTimeObj *sigMA; \
        const char *_nameMA = (nameMA); \
        u.s = alloc(sizeof(struct FmSetTimeObj) + strlen(_nameMA), \
                           FM_SET_TIME_OBJ_REQUEST); \
	sigMA = u.x; \
        sigMA->mask = (maskMA); \
        sigMA->seconds = (secsMA); \
        sigMA->micros = (microsMA); \
        sigMA->name = 0; \
        strcpy(&sigMA->s[sigMA->name], (_nameMA)); \
        send(&u.s, (fmPidMA)); \
} while (0)

/**
 * @signal FmSetOwnerObj
 *
 * @include "ose_spi/fm.sig"
 *
 * @brief Change owner of an object in a volume.
 *
 * @long This signal is sent to a volume manager process. The volume manager
 *       PID can be retrieved by taking the sender of the mount reply signal.
 * @long The new owner is given as a user ID. Owners might not be implemented
 *       in the volume manager. If owners are not supported by the volume
 *       manager, the result is implementation defined.
 *
 * @member FmSetOwnerObj
 *
 * @seealso sendFmSetOwnerObjRequest
 * @seealso FmMount
 */

/**
 * @macro FM_SET_OWNER_OBJ_REQUEST
 * @brief Change owner of an object in a volume.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_SET_OWNER_OBJ_REQUEST       30136
 *
 * @seealso FmSetOwnerObj
 */
#define FM_SET_OWNER_OBJ_REQUEST    (30136) /* !-SIGNO(struct FmSetOwnerObj)-! */

/**
 * @macro FM_SET_OWNER_OBJ_REPLY
 * @brief Change owner of an object in a volume.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_SET_OWNER_OBJ_REPLY         30137
 *
 * @seealso FmSetOwnerObj
 */
#define FM_SET_OWNER_OBJ_REPLY      (30137) /* !-SIGNO(struct FmSetOwnerObj)-! */

/**
 * @type FmSetOwnerObj
 *
 * @declaration
 * struct FmSetOwnerObj
 * {
 *   SIGSELECT sigNo;
 *   EfsStatus status;
 *   U32       owner;
 *   U32       name;
 *   char      s[1];
 * };
 *
 * @input sigNo   FM_SET_OWNER_OBJ_REQUEST
 * @input owner   User ID of the new owner of the object.
 * @input name    Byte offset into string pool where the object name is stored.
 *                         @break The full path of the object is given. This
 *                         starts with the mount point that is registered in
 *                         the FSS, including the leading slash "/".
 * @input s       String pool for variable sized data. The string
 *                refered by name is stored here.
 *                         @break This field is of variable size. The size of
 *                         the string(s) must be added when calculating the
 *                         required signal buffer size. Excess space should
 *                         be ignored if the buffer is larger than needed.
 *
 * @output sigNo  FM_SET_OWNER_OBJ_REPLY
 * @output status EFS_SUCCESS if success.
 *                Preferred @link "errno.h" "ENOTSUP" or old
 *                @link "" "FM_EOPER_NOT_SUPPORTED" if the
 *                volume does not support ownership.
 *                Preferred @link "errno.h" "EROFS" or old
 *                @link "" "FM_EDISK_WRITE_PROTECT" if the
 *                volume is write protected.
 *                Other status codes might appear, e.g. coming from
 *                device driver.
 */
struct FmSetOwnerObj
{
   SIGSELECT sigNo;
   EfsStatus status;
   U32       owner;
   U32       name;
   char      s[1];
};

/**
 * @macro sendFmSetOwnerObjRequest
 *
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define sendFmSetOwnerObjRequest(ownerMA, nameMA, fmPidMA)
 *
 * @brief Send signal @link "" "FmSetOwnerObj" request.
 *
 * @input ownerMA User ID of the new owner of the object.
 * @input nameMA  Name of object which ownership should be changed.
 *                         @break The full path of the object is given. This
 *                         starts with the mount point that is registered in
 *                         the FSS, including the leading slash "/".
 * @input fmPidMA PID of volume manager to send the request to.
 *
 * @seealso FmSetOwnerObj
 *
 * @example sendFmSetOwnerObjRequest(5, "/myfile", vmPID);
 */
#define sendFmSetOwnerObjRequest(ownerMA, nameMA, fmPidMA) \
do { \
	/* The union is a trick to avoid breaking the strict-aliasing rules*/\
	union U \
	{ \
		union SIGNAL *s;\
		struct FmSetOwnerObj *x;\
	} u;\
	struct FmSetOwnerObj *sigMA; \
        const char *_nameMA = (nameMA); \
        u.s = alloc(sizeof(struct FmSetOwnerObj) + strlen(_nameMA), \
                    FM_SET_OWNER_OBJ_REQUEST); \
	sigMA = u.x; \
        sigMA->owner = (ownerMA); \
        sigMA->name = 0; \
        strcpy(&sigMA->s[sigMA->name], (_nameMA)); \
        send(&u.s, (fmPidMA)); \
} while (0)

/*
 *****************************************************************************
 * Signals to a Volume Manager, operating on open objects by handles.
 *****************************************************************************
 */

/**
 * @signal FmEventRead
 *
 * @include "ose_spi/fm.sig"
 *
 * @brief Request a reply signal when there are bytes available for reading.
 *
 * @long Use signal number FM_EVENT_READ_REQUEST to request an event when
 *       there are a number of bytes available for read. The number of bytes
 *       that must be available for reading before the reply signal is sent
 *       is specified in the extra argument.
 * @long The reply signal number will be FM_EVENT_READ_REPLY, and the extra
 *       argument contains the actual number of bytes available for read.
 * @long Note: This signal is currently supported only by sockets.
 *
 * @member FmEvent
 *
 * @seealso FmEventWrite
 * @seealso FmEventEof
 */

/**
 * @macro FM_EVENT_READ_REQUEST
 * @brief Request a reply signal when there are bytes available for reading.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_EVENT_READ_REQUEST           30138
 *
 * @seealso FmEventRead
 */
#define FM_EVENT_READ_REQUEST           (30138)    /* !-SIGNO(struct FmEvent)-! */

/**
 * @macro FM_EVENT_READ_REPLY
 * @brief Reply signal when there are bytes available for reading.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_EVENT_READ_REPLY             30139
 *
 * @seealso FmEventRead
 */
#define FM_EVENT_READ_REPLY             (30139)    /* !-SIGNO(struct FmEvent)-! */

/**
 * @signal FmEventWrite
 *
 * @include "ose_spi/fm.sig"
 *
 * @brief Request a reply signal when there are bytes available for write.
 *
 * @long Use signal number FM_EVENT_WRITE_REQUEST to request an event when
 *       there are a number of bytes available for write. The number of bytes
 *       that must be available for write before the reply signal is sent
 *       is specified in the extra argument.
 * @long The reply signal number will be FM_EVENT_WRITE_REPLY, and the extra
 *       argument contains the actual number of bytes available for write.
 * @long For non-blocking sockets, if the available space (number of bytes) is
 *       lower than @link "sys/socket.h" "SO_SNDLOWAT", write will anyway fail
 *       with @link "errno.h" "EWOULDBLOCK". In this case, it is recommended to
 *       request at least SO_SNDLOWAT bytes.
 * @long Note: This signal is currently supported only by sockets.
 *
 * @member FmEvent
 *
 * @seealso FmEventRead
 * @seealso FmEventEof
 */

/**
 * @macro FM_EVENT_WRITE_REQUEST
 * @brief Request a reply signal when there are bytes available for write.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_EVENT_WRITE_REQUEST          30140
 *
 * @seealso FmEventWrite
 */
#define FM_EVENT_WRITE_REQUEST          (30140)    /* !-SIGNO(struct FmEvent)-! */

/**
 * @macro FM_EVENT_WRITE_REPLY
 * @brief Reply signal when there are bytes available for write.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_EVENT_WRITE_REPLY            30141
 *
 * @seealso FmEventWrite
 */
#define FM_EVENT_WRITE_REPLY            (30141)    /* !-SIGNO(struct FmEvent)-! */

/**
 * @signal FmEventEof
 *
 * @include "ose_spi/fm.sig"
 *
 * @brief Request a reply signal when an end of file event occurs on a file or
 *        socket.
 *
 * @long Use signal number FM_EVENT_EOF_REQUEST to request an event when
 *       an end of file event occurs. End of file event occurs when the socket
 *       is closed on the remote side.
 * @long The extra argument must be zero.
 * @long The reply signal number will be FM_EVENT_EOF_REPLY, and the extra
 *       argument will be zero.
 * @long Note: This signal is currently supported only by sockets.
 *
 * @member FmEvent
 *
 * @seealso FmEventRead
 * @seealso FmEventWrite
 */

/**
 * @macro FM_EVENT_EOF_REQUEST
 * @brief Request a reply signal when an end of file event occurs on a file or
 *        socket.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_EVENT_EOF_REQUEST            30142
 *
 * @seealso FmEventEof
 */
#define FM_EVENT_EOF_REQUEST            (30142)    /* !-SIGNO(struct FmEvent)-! */

/**
 * @macro FM_EVENT_EOF_REPLY
 * @brief Reply signal when an end of file event occurs on a file or
 *        socket.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_EVENT_EOF_REPLY              30143
 *
 * @seealso FmEventEof
 */
#define FM_EVENT_EOF_REPLY              (30143)    /* !-SIGNO(struct FmEvent)-! */

/**
 * @type FmEvent
 *
 * @declaration
 * struct FmEvent
 * {
 *   SIGSELECT sigNo;
 *   S32       status;
 *   FmHandle  handle;
 *   S32       extra;
 * };
 *
 * @input sigNo   FM_EVENT_READ_REQUEST
 * @input handle  Socket for which to wait for the event on.
 * @input extra   The number of bytes that must be available for reading before
 *                the reply signal is sent
 *
 * @output sigNo  FM_EVENT_READ_REPLY
 * @output status Non-zero status code if the requested event was not accepted.
 * @output extra  The actual number of bytes available for read.
 *
 * @input sigNo   FM_EVENT_WRITE_REQUEST
 * @input handle  Socket for which to wait for the event on.
 * @input extra   The number of bytes that must be available for write before
 *                the reply signal is sent
 *
 * @output sigNo  FM_EVENT_WRITE_REPLY
 * @output status Non-zero status code if the requested event was not accepted.
 * @output extra  The actual number of bytes available for write.
 *
 * @input sigNo   FM_EVENT_EOF_REQUEST
 * @input handle  Socket for which to wait for the event on.
 * @input extra   Must be zero.
 *
 * @output sigNo  FM_EVENT_EOF_REPLY
 * @output status Non-zero status code if the requested event was not accepted.
 * @output extra  Will be zero.
 */
struct FmEvent
{
   SIGSELECT sigNo;
   S32       status;
   FmHandle  handle;
   S32       extra;
};


/**
 * @signal FmOpen
 *
 * @include "ose_spi/fm.sig"
 *
 * @brief Open an object in a volume and return an @link "" "FmHandle".
 *
 * @long This signal is sent to a volume manager process. The volume manager
 *       PID can be retrieved by taking the sender of the mount reply signal.
 * @long The handle can be used to access the object using other signals.
 *       The reply also includes the buffering mode and size.
 *
 * @member FmOpen
 *
 * @seealso sendFmOpenRequest
 * @seealso FmClose
 * @seealso FmMount
 */

/**
 * @macro FM_OPEN_REQUEST
 * @brief Open an object in a volume and return an @link "" "FmHandle".
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_OPEN_REQUEST                30150
 *
 * @seealso FmOpen
 */
#define FM_OPEN_REQUEST                (30150)    /* !-SIGNO(struct FmOpen)-! */

/**
 * @macro FM_OPEN_REPLY
 * @brief Open an object in a volume and return an @link "" "FmHandle".
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_OPEN_REPLY                  30151
 *
 * @seealso FmOpen
 */
#define FM_OPEN_REPLY                  (30151)    /* !-SIGNO(struct FmOpen)-! */

/**
 * @type FmOpen
 *
 * @declaration
 * struct FmOpen
 * {
 *   SIGSELECT  sigNo;
 *   EfsStatus  status;
 *   FmHandle   handle;
 *   U32        bufferMode;
 *   U32        bufferSize;
 *   EfsBitmask modes;
 *   EfsBitmask attrs;
 *   U32        name;
 *   U32        amode;
 *   char       s[2];
 * };
 *
 * @long       Note that some field names in this signal may seem confusing.
 *             The naming originates from a time before a Posix interface
 *             was created for the OSE file system.
 *
 * @input sigNo       FM_OPEN_REQUEST
 *
 * @input modes       Open mode. This is equivalent to the omode argument of
 *                    the @link "fcntl.h" "open"() call, but the bits are in
 *                    different positions as defined in @link "" "FmModeEnum".
 *
 * @input attrs       Creation attributes. This is equivalent to the third
 *                    argument of the @link "fcntl.h" "open"() call, but the
 *                    bits are in different positions as defined in
 *                    @link "" "FmAttrEnum". This field is only applicable if
 *                    file creation is requested i the modes field.
 *
 * @input name        Byte offset into string pool where the object name is
 *                    stored.
 *                    @break The full path of the object is given. This
 *                    starts with the mount point that is registered in
 *                    the FSS, including the leading slash "/".
 *
 * @input amode       Byte offset into string pool where an optional extra
 *                    string can be stored. @break
 *
 *                    Most volume managers ignore this string or require it to
 *                    be empty. The original intention of this string was to
 *                    support the mode string from the @link "stdio.h" "fopen"()
 *                    call, but there are no OSE volume managers that accept it
 *                    for that purpose, and the @link "stdio.h" "fopen"() call
 *                    does not place the regular mode string in this field. The
 *                    mode string is instead translated to open mode flags and
 *                    encoded in the modes field. @break
 *
 *                    There is however a way to provide this string through
 *                    the @link "stdio.h" "fopen"() call: if the regular part
 *                    of the "mode" argument is folowed by a comma and extra
 *                    text, the extra text is passed as the amode string.
 *                    E.g. "w+b,supermode" will make amode reference the
 *                    string "supermode".
 *
 * @input s           String pool for variable sized data. The strings
 *                    refered by name and amode are stored here.
 *                    @break This field is of variable size. The size of
 *                    the string(s) must be added when calculating the
 *                    required signal buffer size. Excess space should
 *                    be ignored if the buffer is larger than needed.
 *
 * @output sigNo      FM_OPEN_REPLY
 * @output status     EFS_SUCCESS if success (other reply data valid).
 *                    Preferred @link "errno.h" "ENOENT" or old
 *                    @link "" "FM_EOBJECT_NOT_FOUND" if the
 *                    object is missing and the access modes do not allow
 *                    creation of a new object.
 *                    Other status codes might appear, e.g. coming from
 *                    device driver.
 * @output handle     Handle of type @link "" "FmHandle" to be used to access
 *                    the object using other signals.
 * @output bufferMode Preferred buffer mode (@link "" "FmBuffEnum"). See
 *                    @link "stdio.h" "setvbuf"() for more information.
 * @output bufferSize Preferred buffer size in bytes. See
 *                    @link "stdio.h" "setvbuf"() for more information.
 */
struct FmOpen
{
   SIGSELECT  sigNo;
   EfsStatus  status;
   FmHandle   handle;
   U32        bufferMode;
   U32        bufferSize;
   EfsBitmask modes;
   EfsBitmask attrs;
   U32        name;
   U32        amode;
   char       s[2];
};

/**
 * @macro sendFmOpenRequest
 *
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define sendFmOpenRequest(modesMA, attrsMA, _nameMA, _amodeMA, fmPid)
 *
 * @brief Send signal @link "" "FmOpen" request.
 *
 * @input modesMA  Value for the modes field in @link "" "FmOpen"
 * @input attrsMA  Value for the attrs field in @link "" "FmOpen"
 * @input _nameMA  String for the name field in @link "" "FmOpen"
 * @input _amodeMA String for the amode field in @link "" "FmOpen"
 * @input fmPid    PID of volume manager to send the request to.
 *
 * @seealso FmOpen
 *
 * @example
 * sendFmOpenRequest(FM_MODE_TEXT | FM_MODE_READ,
 *                   FM_ATTR_READ,
 *                   "/myfile",
 *                   "r",
 *                   vmPID);
 */
#define sendFmOpenRequest(modesMA, attrsMA, _nameMA, _amodeMA, fmPid) \
do { \
	/* The union is a trick to avoid breaking the strict-aliasing rules*/\
	union U \
	{ \
		union SIGNAL *s;\
		struct FmOpen *x;\
	} u;\
	struct FmOpen *sigMA; \
        const char *nameMA = (_nameMA); \
        const char *amodeMA = (_amodeMA); \
        u.s = alloc(sizeof(struct FmOpen) + strlen(nameMA) + strlen(amodeMA), \
                    FM_OPEN_REQUEST); \
	sigMA = u.x; \
        sigMA->modes = (modesMA); \
        sigMA->attrs = (attrsMA); \
        sigMA->name = 0; \
        sigMA->amode = strlen(nameMA) + 1; \
        strcpy(&sigMA->s[sigMA->name], nameMA); \
        strcpy(&sigMA->s[sigMA->amode], amodeMA); \
        send(&u.s, (fmPid)); \
} while (0)


/**
 * @signal FmClose
 *
 * @include "ose_spi/fm.sig"
 *
 * @brief Close an opened object referred by an @link "" "FmHandle" handle.
 *
 * @long This signal is sent to a volume manager process. The volume manager
 *       PID can be retrieved by taking the sender of the mount reply signal.
 * @long This will close the handle, other handles might still remain opened
 *       even if they refer to the same file. Any unwritten output buffers are
 *       written to the volume manager before the handle is closed.
 * @long Any pending operation that is not yet performed will be canceled if
 *       cancel argument is set to True (1), otherwise the object can not be
 *       closed if it has pending operations.
 *
 * @member FmClose
 *
 * @seealso sendFmCloseRequest
 * @seealso FmOpen
 * @seealso FmMount
 */

/**
 * @macro FM_CLOSE_REQUEST
 * @brief Close an opened object referred by an @link "" "FmHandle" handle.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_CLOSE_REQUEST               30152
 *
 * @seealso FmClose
 */
#define FM_CLOSE_REQUEST               (30152)    /* !-SIGNO(struct FmClose)-! */

/**
 * @macro FM_CLOSE_REPLY
 * @brief Close an opened object referred by an @link "" "FmHandle" handle.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_CLOSE_REPLY                 30153
 *
 * @seealso FmClose
 */
#define FM_CLOSE_REPLY                 (30153)    /* !-SIGNO(struct FmClose)-! */

/**
 * @type FmClose
 *
 * @declaration
 * struct FmClose
 * {
 *   SIGSELECT  sigNo;
 *   EfsStatus  status;
 *   FmHandle   handle;
 *   EfsBoolean cancel;
 * };
 *
 * @input sigNo   FM_CLOSE_REQUEST
 * @input handle  @link "" "FmHandle" handle to the opened object. The handle
 *                will be illegal after a successful close.
 * @input cancel  If True (1), it will cancel (and reply to) any pending
 *                read, write or wait operations before closing the handle.
 *                If sending replies it will use the preferred
 *                @link "errno.h" "ECANCELED" or old 
 *                @link "" "FM_EOPER_CANCELLED" status code in the replies.
 *
 * @output sigNo  FM_CLOSE_REPLY
 * @output status EFS_SUCCESS if success.
 *                Preferred @link "errno.h" "EBADF" or old 
 *                @link "" "FM_EINVALID_HANDLE" if handle is not valid, e.g. never
 *                opened or has already been closed.
 *                Other status codes might appear, e.g. coming from
 *                device driver.
 */
struct FmClose
{
   SIGSELECT  sigNo;
   EfsStatus  status;
   FmHandle   handle;
   EfsBoolean cancel;
};

/**
 * @macro sendFmCloseRequest
 *
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define sendFmCloseRequest(handleMA, cancelMA, fmPidMA)
 *
 * @brief Send signal @link "" "FmClose" request.
 *
 * @input handleMA @link "" "FmHandle" handle to the opened object. The handle
 *                 will be illegal after a successful close.
 * @input cancelMA If True (1) will cancel (and reply to) any pending
 *                 read, write or wait operations before closing the handle.
 *                 If sending replies it will use the preferred
 *                 @link "errno.h" "ECANCELED" or old
 *                 @link "" "FM_EOPER_CANCELLED" status code in the replies.
 * @input fmPidMA  PID of volume manager to send the request to.
 *
 * @seealso FmClose
 *
 * @example sendFmCloseRequest(myHandle, True, vmPID);
 */
#define sendFmCloseRequest(handleMA, cancelMA, fmPidMA) \
do { \
	/* The union is a trick to avoid breaking the strict-aliasing rules*/\
	union U \
	{ \
		union SIGNAL *s;\
		struct FmClose *x;\
	} u;\
        struct FmClose *sigMA; \
	u.s = alloc(sizeof(struct FmClose), FM_CLOSE_REQUEST); \
	sigMA = u.x; \
        sigMA->handle = (handleMA); \
        sigMA->cancel = (cancelMA); \
        send(&u.s, (fmPidMA)); \
} while (0)


/**
 * @signal FmSetOwnerProc
 *
 * @include "ose_spi/fm.sig"
 *
 * @brief Changes the owner process for an @link "" "FmHandle" handle.
 *
 * @long This signal is sent to a volume manager process. The volume manager
 *       PID can be retrieved by taking the sender of the mount reply signal.
 *
 * @long This signal is a request to change the process that is supervised
 *       as "owner" for an @link "" "FmHandle" handle. The association
 *       between the handle and the old owner process shall be removed.
 *
 * @long The volume manager is expected to supervise the owner process of an
 *       @link "" "FmHandle" handle, and free the associated resources if
 *       the owner process dies.
 *
 * @long There are two main use-cases for this signal in the CRT:
 *
 * @long When donating a file descriptor to another process, the ownership
 *       is changed to the new process.
 *
 * @long When a program that executes in the "shared" execution model opens a
 *       file, ownership is assigned to a special process that acts as owner
 *       for all the handles that are shared between all the processes of
 *       the program. The intention is that the handle shall be cleaned up
 *       when the program dies, not when an individual process dies.
 *
 * @long Note that there are no access restrictions associated with the
 *       owner process; other processes than the owner are allowed to
 *       access the handle.
 *
 * @long The process that is registered as "sender" of this request signal
 *       shall become the new owner process of the handle. The reply signal
 *       shall be sent to the old owner process. The intended use is that
 *       the old owner of the handle uses the send_w_sender() function to
 *       send the request signal and specify the intended new owner as sender.
 *
 * @long The volume manager can only send the reply signal back to the old
 *       owner process if the handle is valid. Therefore there shall be no
 *       return signal if the handle is invalid, and the old owner process
 *       should use receive_w_tmo() to receive the reply.
 *
 * @long It is the responsibility of the donating process to somehow
 *       transfer the information about the handle to the new process.
 *       The handle can be given to a process in another machine.
 *
 * @member FmSetOwnerProc
 *
 * @seealso sendFmSetOwnerProcRequest
 * @seealso FmOpen
 */

/**
 * @macro FM_SET_OWNER_PROC_REQUEST
 * @brief Changes the process owner for an @link "" "FmHandle" handle.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_SET_OWNER_PROC_REQUEST      30154
 *
 * @seealso FmSetOwnerProc
 */
#define FM_SET_OWNER_PROC_REQUEST  (30154) /* !-SIGNO(struct FmSetOwnerProc)-! */

/**
 * @macro FM_SET_OWNER_PROC_REPLY
 * @brief Changes the process owner for an @link "" "FmHandle" handle.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_SET_OWNER_PROC_REPLY        30155
 *
 * @seealso FmSetOwnerProc
 */
#define FM_SET_OWNER_PROC_REPLY    (30155) /* !-SIGNO(struct FmSetOwnerProc)-! */

/**
 * @type FmSetOwnerProc
 *
 * @declaration
 * struct FmSetOwnerProc
 * {
 *   SIGSELECT sigNo;
 *   EfsStatus status;
 *   FmHandle  handle;
 * };
 *
 * @input sigNo   FM_SET_OWNER_PROC_REQUEST
 * @input handle  @link "" "FmHandle" handle to an opened object.
 *
 * @output sigNo  FM_SET_OWNER_PROC_REPLY
 * @output status EFS_SUCCESS if success.
 *                Other status codes might appear, e.g. coming from
 *                device driver.
 */
struct FmSetOwnerProc
{
   SIGSELECT sigNo;
   EfsStatus status;
   FmHandle  handle;
};

/**
 * @macro sendFmSetOwnerProcRequest
 *
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define sendFmSetOwnerProcRequest(handleMA, newOwnerMA, fmPidMA)
 *
 * @brief Send signal @link "" "FmSetOwnerProc" request.
 *
 * @input handleMA   @link "" "FmHandle" handle to an opened object.
 * @input newOwnerMA PID of new owner.
 * @input fmPidMA    PID of volume manager process to send the request to.
 *
 * @restriction This macro should not be used. It uses send_w_s() function
 * which has a known race condition which can result in the signal being
 * silently lost even though the volume manager still lives. Attaching to new
 * owner does not solve this. Please consider using send_w_sender() instead.
 *
 * @seealso FmSetOwnerProc
 */
#define sendFmSetOwnerProcRequest(handleMA, newOwnerMA, fmPidMA) \
do { \
	/* The union is a trick to avoid breaking the strict-aliasing rules*/\
	union U \
	{ \
		union SIGNAL *s;\
		struct FmSetOwnerProc *x;\
	} u;\
        struct FmSetOwnerProc *sigMA; \
	u.s = alloc(sizeof(struct FmSetOwnerProc), FM_SET_OWNER_PROC_REQUEST); \
	sigMA = u.x; \
        sigMA->handle = (handleMA); \
        send_w_s(&u.s, (newOwnerMA), (fmPidMA)); \
} while (0)


/**
 * @signal FmWrite
 *
 * @include "ose_spi/fm.sig"
 *
 * @brief Write data to an object referred to by an @link "" "FmHandle" handle.
 *
 * @long The intention of this signal is to be a functional match to the
 *       Posix @link "" "write" system call.
 * @long This signal is sent to a volume manager process. The volume manager
 *       PID can be retrieved by taking the sender of the mount reply signal.
 * @long Typically the object is a file or a terminal channel.
 * @long The "position" argument allows to update the file offset of the
 *       object before starting the write. The write shall then start at the
 *       file offset and the file offset shall be incremented by the number
 *       of bytes actually written. Objects that do not support seeking always
 *       write at the current position and ignore this argument. Objects in
 *       append mode always set the file offset to end-of-file before writing
 *       and ignore this argument.
 * @long If the file offset is beyond end-of-file, subsequent reads of data
 *       in the gap shall return bytes with the value 0 until data is actually
 *       written into the gap.
 * @long Note: Some OSE volume managers support seeking but not beyond the
 *       current end-of-file. Such volume managers will fail the write request
 *       if the requested position is beyond end-of-file.
 * @long The "requested" argument defines the number of bytes to attempt to
 *       write.
 * @long With this signal the data that is to be written is contained within
 *       the request signal buffer.
 * @long The @link "" "FmWritePtr" signal provides an alternative way to
 *       transfer data.
 * @long The reply signal will come after transferring the data to the volume
 *       manager. There is no guarantee that the data has reached the device
 *       driver when the reply signal is received, as the volume manager might
 *       cache data.
 *
 * @member FmWriteRequest
 * @member FmWriteReply
 *
 * @seealso sendFmWriteRequest
 * @seealso FmWritePtr
 * @seealso FmOpen
 * @seealso FmMount
 */

/**
 * @macro FM_WRITE_REQUEST
 * @brief Write data to an object referred to by an @link "" "FmHandle" handle.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_WRITE_REQUEST               30156
 *
 * @seealso FmWrite
 */
#define FM_WRITE_REQUEST      (30156)    /* !-SIGNO(struct FmWriteRequest)-! */

/**
 * @macro FM_WRITE_REPLY
 * @brief Write data to an object referred to by an @link "" "FmHandle" handle.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_WRITE_REPLY                 30157
 *
 * @seealso FmWrite
 */
#define FM_WRITE_REPLY        (30157)    /* !-SIGNO(struct FmWriteReply)-! */

/**
 * @type FmWriteRequest
 *
 * @declaration
 * struct FmWriteRequest
 * {
 *   SIGSELECT sigNo;
 *   FmHandle  handle;
 *   EfsEnum   position;
 *   U32       requested;
 *   char      buffer[1];
 * };
 *
 * @input sigNo     FM_WRITE_REQUEST
 * @input handle    @link "" "FmHandle" handle to the opened object to write to.
 * @input position  Option to set the file offset before starting the transfer.
 *                  See @link "" "FmPositionEnum" for instructions.
 * @input requested Number of bytes to write.
 * @input buffer    Buffer with data to write.
 *                         @break This field is of variable size. The size of
 *                         the data must be added when calculating the
 *                         required signal buffer size. Excess space should
 *                         be ignored if the buffer is larger than needed.
 *
 * @seealso FmWrite
 */
struct FmWriteRequest
{
   SIGSELECT sigNo;
   FmHandle  handle;
   EfsEnum   position;
   U32       requested;
   char      buffer[1];
};

/**
 * @type FmWriteReply
 *
 * @declaration
 * struct FmWriteReply
 * {
 *   SIGSELECT sigNo;
 *   EfsStatus status;
 *   FmHandle  handle;
 *   U32       actual;
 * };
 *
 * @output sigNo  FM_WRITE_REPLY
 * @output status EFS_SUCCESS if success (other reply data valid).
 *                Preferred @link "errno.h" "EBADF" or old 
 *                @link "" "FM_EINVALID_HANDLE" if the handle is invalid.
 *                Preferred @link "errno.h" "EACCES" or old
 *                @link "" "FM_EOBJECT_WRITE_PROTECT" if the
 *                object is read-only.
 *                Preferred @link "errno.h" "ECANCELED" or old
 *                @link "" "FM_EOPER_CANCELLED" if the
 *                operation was canceled.
 *                Other status codes might appear, e.g. coming from
 *                device driver.
 * @output handle The @link "" "FmHandle" handle given in the request signal.
 * @output actual Number of bytes actually written.
 *
 * @seealso FmWrite
 */
struct FmWriteReply
{
   SIGSELECT sigNo;
   EfsStatus status;
   FmHandle  handle;
   U32       actual;
};

/**
 * @macro sendFmWriteRequest
 *
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define sendFmWriteRequest(handleMA, posMA, requestedMA, bufferMA, fmPidMA)
 *
 * @brief Send signal @link "" "FmWrite" request.
 *
 * @input handleMA    @link "" "FmHandle" handle to the opened object to
 *                    write to.
 * @input posMA       Option to set the file offset before starting the
 *                    transfer. See @link "" "FmPositionEnum" for instructions.
 * @input requestedMA Number of bytes to write.
 * @input bufferMA    Pointer to the buffer to write to the file.
 * @input fmPidMA     PID of volume manager process to send the request to.
 *
 * @seealso FmWrite
 *
 * @example
 * char mybuf[] = "Write this data to file referred to by myHandle";
 * sendFmWriteRequest(myHandle, FM_POSITION_AT_END, sizeof(mybuf), mybuf,
 *                    vmPID);
 */
#define sendFmWriteRequest(handleMA, posMA, requestedMA, bufferMA, fmPidMA) \
do { \
	/* The union is a trick to avoid breaking the strict-aliasing rules*/\
	union U \
	{ \
		union SIGNAL *s;\
		struct FmWriteRequest *x;\
	} u;\
	struct FmWriteRequest *sigMA; \
        U32 _requestedMA = (requestedMA); \
	u.s = alloc(sizeof(struct FmWriteRequest) + \
                       (_requestedMA), FM_WRITE_REQUEST); \
	sigMA = u.x; \
        sigMA->handle = (handleMA); \
        sigMA->position = (posMA); \
        sigMA->requested = (_requestedMA); \
        memcpy(sigMA->buffer, (bufferMA), (_requestedMA)); \
        send(&u.s, (fmPidMA)); \
} while (0)


/**
 * @signal FmRead
 *
 * @include "ose_spi/fm.sig"
 *
 * @brief Read data from an object referred to by an @link "" "FmHandle" handle.
 *
 * @long The intention of this signal is to be a functional match to the
 *       Posix @link "" "read" system call.
 * @long This signal is sent to a volume manager process. The volume manager
 *       PID can be retrieved by taking the sender of the mount reply signal.
 * @long The object is typically a regular file for reading data, a directory
 *       for reading directory entries, or a communciation channel such as a
 *       terminal or a pipe.
 * @long The "position" argument allows to update the file offset of the object
 *       before starting the read. The read shall then start at the file offset
 *       and the file offset shall be incremented by the number of bytes
 *       actually read. Objects that do not support seeking always read at the
 *       current position and ignore this argument.
 * @long The "requested" argument contains the number of bytes to attempt to
 *       read. No data transfer shall occur past the current end-of-file.
 * @long A reply that contains 0 bytes data indicates end-of-file.
 *       For regular files and directories this happens when a read operation
 *       starts at or beyond end-of-file.
 * @long Objects that represent communication channels only indicate
 *       end-of-file if they are not connected to a data source. Such objects
 *       should block until there is data available if they receive a read
 *       request when no data is available.
 * @long A reply that contains at least one byte but less than the requested
 *       number of bytes is perfectly legal and does NOT indicate end-of-file.
 *       This happens for many reasons, depending on the object type. For
 *       example:
 * @long *  Directory reads are sometimes grouped as full directory entries.
 * @long *  Many communication channels will stop when reaching end of
 *          currently available data.
 * @long *  The console may be configured to deliver individual lines of data.
 * @long With this signal the read data is contained within the reply signal
 *       buffer. It is the callers responsibility to ensure the "requested"
 *       transfer size will not result in a reply signal that is too large
 *       for the system to handle.
 * @long The @link "" "FmReadPtr" signal provides an alternative way to
 *       transfer data.
 *
 * @member FmReadRequest
 * @member FmReadReply
 *
 * @seealso sendFmReadRequest
 * @seealso FmReadPtr
 * @seealso FmOpen
 * @seealso FmMount
 */

/**
 * @macro FM_READ_REQUEST
 * @brief Read data from an object referred to by an @link "" "FmHandle" handle.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_READ_REQUEST                30158
 *
 * @seealso FmRead
 */
#define FM_READ_REQUEST          (30158)    /* !-SIGNO(struct FmReadRequest)-! */

/**
 * @macro FM_READ_REPLY
 * @brief Read data from an object referred to by an @link "" "FmHandle" handle.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_READ_REPLY                  30159
 *
 * @seealso FmRead
 */
#define FM_READ_REPLY            (30159)    /* !-SIGNO(struct FmReadReply)-! */

/**
 * @type FmReadRequest
 *
 * @declaration
 * struct FmReadRequest
 * {
 *   SIGSELECT  sigNo;
 *   FmHandle   handle;
 *   EfsEnum    position;
 *   EfsBoolean block;
 *   U32        requested;
 * };
 *
 * @input sigNo     FM_READ_REQUEST
 * @input handle    @link "" "FmHandle" handle to the opened object to read
 *                  from.
 * @input position  Option to set the file offset before starting the transfer.
 *                  See @link "" "FmPositionEnum" for instructions.
 * @input block     This field is ignored. Should contain 0 or 1 for
 *                  compatibility reasons.
 * @input requested Number of bytes to attempt to read.
 *
 * @seealso FmRead
 */
struct FmReadRequest
{
   SIGSELECT  sigNo;
   FmHandle   handle;
   EfsEnum    position;
   EfsBoolean block;
   U32        requested;
};

/**
 * @type FmReadReply
 *
 * @declaration
 * struct FmReadReply
 * {
 *   SIGSELECT sigNo;
 *   EfsStatus status;
 *   FmHandle  handle;
 *   U32       actual;
 *   char      buffer[1];
 * };
 *
 * @output sigNo  FM_READ_REPLY
 * @output status EFS_SUCCESS if success (other reply data valid).
 *                Preferred @link "errno.h" "EBADF" or old 
 *                @link "" "FM_EINVALID_HANDLE" if the handle is invalid.
 *                Preferred @link "errno.h" "EACCES" or old
 *                @link "" "FM_EOBJECT_READ_PROTECT" if the
 *                object is read protected.
 *                Preferred @link "errno.h" "ECANCELED" or old
 *                @link "" "FM_EOPER_CANCELLED" if the
 *                operation was canceled.
 *                Preferred @link "errno.h" "ENOTSUP" or old
 *                @link "" "FM_EOPER_NOT_SUPPORTED" if reading
 *                is not supported on this volume.
 *                Other status codes might appear, e.g. coming from
 *                device driver.
 * @output handle @link "" "FmHandle" handle used in the request signal.
 * @output actual Actual number of bytes read.
 * @output buffer Variable sized buffer with read data. This buffer contains
 *                the number of valid bytes indicated by the "actual" argument.
 *                         @break This field is of variable size. The size of
 *                         the data must be added when calculating the
 *                         required signal buffer size. Excess space should
 *                         be ignored if the buffer is larger than needed.
 *
 * @seealso FmRead
 */
struct FmReadReply
{
   SIGSELECT sigNo;
   EfsStatus status;
   FmHandle  handle;
   U32       actual;
   char      buffer[1];
};

/**
 * @macro sendFmReadRequest
 *
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define sendFmReadRequest(handleMA, posMA, blockMA, requestedMA, fmPidMA)
 *
 * @brief Send signal @link "" "FmRead" request.
 *
 * @input handleMA    @link "" "FmHandle" handle to the opened object to read
 *                    from.
 * @input posMA       Option to set the file offset before starting the
 *                    transfer. See @link "" "FmPositionEnum" for instructions.
 * @input blockMA     This argument is ignored. Should contain 0 or 1 for
 *                    compatibility reasons.
 * @input requestedMA Number of bytes to attempt to read.
 * @input fmPidMA     PID of volume manager to send the request to.
 *
 * @seealso FmRead
 *
 * @example sendFmReadRequest(myHandle, FM_POSITION_CURRENT, False, 255, vmPID);
 */
#define sendFmReadRequest(handleMA, posMA, blockMA, requestedMA, fmPidMA) \
do { \
	/* The union is a trick to avoid breaking the strict-aliasing rules*/\
	union U \
	{ \
		union SIGNAL *s;\
		struct FmReadRequest *x;\
	} u;\
        struct FmReadRequest *sigMA; \
	u.s = alloc(sizeof(struct FmReadRequest), FM_READ_REQUEST); \
	sigMA = u.x; \
        sigMA->handle = (handleMA); \
        sigMA->position = (posMA); \
        sigMA->block = (blockMA); \
        sigMA->requested = (requestedMA); \
        send(&u.s, (fmPidMA)); \
} while (0)


/**
 * @signal FmWritePtr
 *
 * @include "ose_spi/fm.sig"
 *
 * @brief Write data to an object referred to by an @link "" "FmHandle" handle.
 *
 * @long Except for the details noted below, this signal provides the same
 *       semantics as the @link "" "FmWrite" signal. Refer to that signal
 *       for details.
 * @long This signal contains a pointer directly to a data buffer where data
 *       is to be read from. This data buffer shall be provided by the caller.
 * @long This signal cannot be used if the data is in MAS allocated memory or
 *       if the volume manager is on another machine or core.
 * @long The advantage of using this signal instead of @link "" "FmWrite" is
 *       that it will save one memory copy, and the maximum transfer size is
 *       not limited by the maximum signal size.
 *
 * @member FmWritePtr
 *
 * @seealso FmWrite
 * @seealso sendFmWritePtrRequest
 * @seealso FmOpen
 * @seealso FmMount
 */

/**
 * @macro FM_WRITE_PTR_REQUEST
 * @brief Write data to an object referred to by an @link "" "FmHandle" handle.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_WRITE_PTR_REQUEST           30160
 *
 * @seealso FmWritePtr
 */
#define FM_WRITE_PTR_REQUEST     (30160)    /* !-SIGNO(struct FmWritePtr)-! */

/**
 * @macro FM_WRITE_PTR_REPLY
 * @brief Write data to an object referred to by an @link "" "FmHandle" handle.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_WRITE_PTR_REPLY             30161
 *
 * @seealso FmWritePtr
 */
#define FM_WRITE_PTR_REPLY       (30161)    /* !-SIGNO(struct FmWritePtr)-! */

/**
 * @type FmWritePtr
 *
 * @declaration
 * struct FmWritePtr
 * {
 *   SIGSELECT  sigNo;
 *   EfsStatus  status;
 *   FmHandle   handle;
 *   EfsEnum    position;
 *   U32        requested;
 *   U32        actual;
 *   const char *buffer;
 * };
 *
 * @input sigNo     FM_WRITE_PTR_REQUEST
 * @input handle    @link "" "FmHandle" handle to the opened object to write to.
 * @input position  Option to set the file offset before starting the transfer.
 *                  See @link "" "FmPositionEnum" for instructions.
 * @input requested Number of bytes to write.
 * @input buffer    Pointer to buffer containing data to write to file.
 *
 * @output sigNo    FM_WRITE_PTR_REPLY
 * @output status   EFS_SUCCESS if success (other reply data valid).
 *                  Preferred @link "errno.h" "EBADF" or old 
 *                  @link "" "FM_EINVALID_HANDLE" if the handle is invalid.
 *                  Preferred @link "errno.h" "EACCES" or old
 *                  @link "" "FM_EOBJECT_WRITE_PROTECT" if the
 *                  object is read-only.
 *                  Preferred @link "errno.h" "ECANCELED" or old
 *                  @link "" "FM_EOPER_CANCELLED" if the
 *                  operation was canceled.
 *                  Other status codes might appear coming from e.g.
 *                  device driver.
 * @output handle   @link "" "FmHandle" handle used in the request signal.
 * @output actual   Number of bytes actually written.
 *
 * @seealso FmWritePtr
 */
struct FmWritePtr
{
   SIGSELECT  sigNo;
   EfsStatus  status;
   FmHandle   handle;
   EfsEnum    position;
   U32        requested;
   U32        actual;
   const char *buffer;
};

/**
 * @macro sendFmWritePtrRequest
 *
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define sendFmWritePtrRequest(handleMA, posMA, requestedMA,
 *                               bufferMA, fmPidMA)
 *
 * @brief Send signal @link "" "FmWritePtr" request.
 *
 * @input handleMA    @link "" "FmHandle" handle to the opened object to
 *                    write to.
 * @input posMA       Option to set the file offset before starting the
 *                    transfer. See @link "" "FmPositionEnum" for instructions.
 * @input requestedMA Number of bytes to write.
 * @input bufferMA    Pointer to buffer containing data to write to file.
 * @input fmPidMA     PID of volume manager to send the request to.
 *
 * @seealso FmWritePtr
 */
#define sendFmWritePtrRequest(handleMA, posMA, requestedMA, bufferMA, fmPidMA) \
do { \
	/* The union is a trick to avoid breaking the strict-aliasing rules*/\
	union U \
	{ \
		union SIGNAL *s;\
		struct FmWritePtr *x;\
	} u;\
        struct FmWritePtr *sigMA; \
	u.s = alloc(sizeof(struct FmWritePtr), FM_WRITE_PTR_REQUEST); \
	sigMA = u.x; \
        sigMA->handle = (handleMA); \
        sigMA->position = (posMA); \
        sigMA->requested = (requestedMA); \
        sigMA->buffer = (bufferMA); \
        send(&u.s, (fmPidMA)); \
} while (0)


/**
 * @signal FmReadPtr
 *
 * @include "ose_spi/fm.sig"
 *
 * @brief Read data from an object referred to by an @link "" "FmHandle" handle.
 *
 * @long Except for the details noted below, this signal provides the same
 *       semantics as the @link "" "FmRead" signal. Refer to that signal
 *       for details.
 * @long This signal contains a pointer directly to a data buffer where data
 *       is to be written. This data buffer shall be provided by the caller.
 * @long This signal cannot be used if the data is in MAS allocated memory or
 *       if the volume manager is on another machine or core.
 * @long The advantage of using this signal instead of @link "" "FmRead" is
 *       that it will save one memory copy, and the maximum transfer size is
 *       not limited by the maximum signal size.
 *
 * @member FmReadPtr
 *
 * @seealso FmRead
 * @seealso sendFmReadPtrRequest
 * @seealso FmOpen
 * @seealso FmMount
 */

/**
 * @macro FM_READ_PTR_REQUEST
 * @brief Read data from an object referred to by the @link "" "FmHandle"
 *        handle.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_READ_PTR_REQUEST            30162
 *
 * @seealso FmReadPtr
 */
#define FM_READ_PTR_REQUEST      (30162)    /* !-SIGNO(struct FmReadPtr)-! */

/**
 * @macro FM_READ_PTR_REPLY
 * @brief Read data from an object referred to by the @link "" "FmHandle"
 *        handle.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_READ_PTR_REPLY              30163
 *
 * @seealso FmReadPtr
 */
#define FM_READ_PTR_REPLY        (30163)    /* !-SIGNO(struct FmReadPtr)-! */

/**
 * @type FmReadPtr
 *
 * @declaration
 * struct FmReadPtr
 * {
 *   SIGSELECT  sigNo;
 *   EfsStatus  status;
 *   FmHandle   handle;
 *   EfsEnum    position;
 *   EfsBoolean block;
 *   U32        requested;
 *   U32        actual;
 *   char       *buffer;
 * };
 *
 * @input sigNo     FM_READ_PTR_REQUEST
 * @input handle    @link "" "FmHandle" handle to the opened object to read
 *                  from.
 * @input position  Option to set the file offset before starting the transfer.
 *                  See @link "" "FmPositionEnum" for instructions.
 * @input block     This field is ignored. Should contain 0 or 1 for
 *                  compatibility reasons.
 * @input requested Number of bytes to attempt to read.
 * @input buffer    Pointer to buffer to be filled with read data.
 *
 * @output sigNo    FM_READ_PTR_REPLY
 * @output status   EFS_SUCCESS if success (other reply data valid).
 *                  Preferred @link "errno.h" "EBADF" or old 
 *                  @link "" "FM_EINVALID_HANDLE" if the handle is invalid.
 *                  Preferred @link "errno.h" "EACCES" or old
 *                  @link "" "FM_EOBJECT_READ_PROTECT" if the
 *                  object is read protected.
 *                  Preferred @link "errno.h" "ECANCELED" or old
 *                  @link "" "FM_EOPER_CANCELLED" if the
 *                  operation was canceled.
 *                  Preferred @link "errno.h" "ENOTSUP" or old
 *                  @link "" "FM_EOPER_NOT_SUPPORTED" if reading
 *                  is not supported on this volume.
 *                  Other status codes might appear, e.g. coming from
 *                  device driver.
 * @output handle   @link "" "FmHandle" handle used in the request signal.
 * @output actual   Actual number of bytes read.
 * @output buffer   Points to the user buffer which contains the number of read
 *                  bytes indicated by the actual argument.
 *
 * @seealso FmReadPtr
 */
struct FmReadPtr
{
   SIGSELECT  sigNo;
   EfsStatus  status;
   FmHandle   handle;
   EfsEnum    position;
   EfsBoolean block;
   U32        requested;
   U32        actual;
   char       *buffer;
};

/**
 * @macro sendFmReadPtrRequest
 *
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define sendFmReadPtrRequest(handleMA, posMA, blockMA, requestedMA,
 *                              bufferMA, fmPidMA)
 *
 * @brief Send signal @link "" "FmReadPtr" request.
 *
 * @input handleMA    @link "" "FmHandle" handle to the opened object to read
 *                    from.
 * @input posMA       Option to set the file offset before starting the
 *                    transfer. See @link "" "FmPositionEnum" for instructions.
 * @input blockMA     This argument is ignored. Should contain 0 or 1 for
 *                    compatibility reasons.
 * @input requestedMA Number of bytes to attempt to read.
 * @input bufferMA    Pointer to buffer to be filled with read data.
 * @input fmPidMA     PID of volume manager to send the request to.
 *
 * @seealso FmReadPtr
 */
#define sendFmReadPtrRequest(handleMA, posMA, blockMA, requestedMA, \
                             bufferMA, fmPidMA) \
do { \
	/* The union is a trick to avoid breaking the strict-aliasing rules*/\
	union U \
	{ \
		union SIGNAL *s;\
		struct FmReadPtr *x;\
	} u;\
        struct FmReadPtr *sigMA; \
	u.s = alloc(sizeof(struct FmReadPtr), FM_READ_PTR_REQUEST); \
	sigMA = u.x; \
        sigMA->handle = (handleMA); \
        sigMA->position = (posMA); \
        sigMA->block = (blockMA); \
        sigMA->requested = (requestedMA); \
        sigMA->buffer = (bufferMA); \
        send(&u.s, (fmPidMA)); \
} while (0)


/**
 * @signal FmWait
 *
 * @include "ose_spi/fm.sig"
 *
 * @brief Wait for one or several events for an object referred to by the
 *        @link "" "FmHandle" handle.
 *
 * @long This signal is sent to a volume manager process. The volume manager
 *       PID can be retrieved by taking the sender of the mount reply signal.
 * @long The waitCond is a bitmask where one or more event types can be
 *       OR'ed together to specify which events to wait for.
 * @long In the reply, the readyCond field will indicate the event(s) causing
 *       the reply.
 * @long For a list of possible events, please refer to @link "" "FmCondEnum".
 *
 * @restriction Only confm supports this signal.
 *
 * @member FmWait
 *
 * @seealso FmOpen
 * @seealso FmMount
 */

/**
 * @macro FM_WAIT_REQUEST
 * @brief Wait for one or several events for an object referred to by the
 *        @link "" "FmHandle" handle.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_WAIT_REQUEST                30164
 *
 * @seealso FmWait
 */
#define FM_WAIT_REQUEST                (30164)    /* !-SIGNO(struct FmWait)-! */

/**
 * @macro FM_WAIT_REPLY
 * @brief Wait for one or several events for an object referred to by the
 *        @link "" "FmHandle" handle.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_WAIT_REPLY                  30165
 *
 * @seealso FmWait
 */
#define FM_WAIT_REPLY                  (30165)    /* !-SIGNO(struct FmWait)-! */

/**
 * @type FmWait
 *
 * @declaration
 * struct FmWait
 * {
 *   SIGSELECT  sigNo;
 *   EfsStatus  status;
 *   FmHandle   handle;
 *   EfsBitmask waitCond;
 *   EfsBitmask readyCond;
 * };
 *
 * @input sigNo      FM_WAIT_REQUEST
 * @input handle     @link "" "FmHandle" handle to an opened object.
 * @input waitCond   Event types to wait for (@link "" "FmCondEnum").
 *
 * @output sigNo     FM_WAIT_REPLY
 * @output status    EFS_SUCCESS if success (other reply data valid).
 *                   Preferred @link "errno.h" "ENOSYS" or old
 *                   @link "" "FM_EOPER_NOT_IMPLEMENTED" if the
 *                   operation is not implemented.
 *                   Other status codes might appear, e.g. coming from
 *                   device driver.
 * @output readyCond Which events that has occurred.
 *                   For possible events, see @link "" "FmCondEnum".
 */
struct FmWait
{
   SIGSELECT  sigNo;
   EfsStatus  status;
   FmHandle   handle;
   EfsBitmask waitCond;
   EfsBitmask readyCond;
};

/**
 * @signal FmCancel
 *
 * @include "ose_spi/fm.sig"
 *
 * @brief Cancel pending operations on an object referred to by an
 *        @link "" "FmHandle" handle.
 *
 * @long This signal is sent to a volume manager process. The volume manager
 *       PID can be retrieved by taking the sender of the mount reply signal.
 * @long Either only operations from this process or from all processes are
 *       canceled depending on argument global.
 * @long For a list of operations that can be canceled refer to
 *       @link "" "FmOperEnum".
 * @long When canceled, a reply is sent to the requesting processes for each
 *       pending operation, using the status preferred
 *       @link "errno.h" "ECANCELED" or old @link "" "FM_EOPER_CANCELLED".
 *
 * @member FmCancel
 *
 * @seealso FmOpen
 * @seealso FmMount
 */

/**
 * @macro FM_CANCEL_REQUEST
 * @brief Cancel pending operations on an object referred to by an
 *        @link "" "FmHandle" handle.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_CANCEL_REQUEST              30166
 *
 * @seealso FmCancel
 */
#define FM_CANCEL_REQUEST              (30166)    /* !-SIGNO(struct FmCancel)-! */

/**
 * @macro FM_CANCEL_REPLY
 * @brief Cancel pending operations on an object referred to by an
 *        @link "" "FmHandle" handle.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_CANCEL_REPLY                30167
 *
 * @seealso FmCancel
 */
#define FM_CANCEL_REPLY                (30167)    /* !-SIGNO(struct FmCancel)-! */

/**
 * @type FmCancel
 *
 * @declaration
 * struct FmCancel
 * {
 *   SIGSELECT  sigNo;
 *   EfsStatus  status;
 *   FmHandle   handle;
 *   EfsBitmask operations;
 *   EfsBitmask global;
 *   volatile const char *buffer;
 * };
 *
 * @input sigNo      FM_CANCEL_REQUEST
 * @input handle     @link "" "FmHandle" handle to an opened object.
 * @input operations Type of operations to cancel (@link "" "FmOperEnum").
 * @input global     Scope of operations to cancel (@link "" "FmCancelEnum").
 * @input buffer     Optional pointer to buffer in order to cancel specific
 *                   read or write (pointer) request (must then be != NULL).
 *                   Used when global has the value FM_CANCEL_BYBUFFER.
 *
 * @output sigNo     FM_CANCEL_REPLY
 * @output status    EFS_SUCCESS if success.
 *                   Other status codes might appear, e.g. coming from
 *                   device driver.
 */
struct FmCancel
{
   SIGSELECT  sigNo;
   EfsStatus  status;
   FmHandle   handle;
   EfsBitmask operations;
   EfsBitmask global;
   volatile const char *buffer;
};

/**
 * @signal FmSetPos
 *
 * @include "ose_spi/fm.sig"
 *
 * @brief Set next read/write position on an open object referred to by the
 *        @link "" "FmHandle" handle.
 *
 * @long This signal is sent to a volume manager process. The volume manager
 *       PID can be retrieved by taking the sender of the mount reply signal.
 * @long This new position is used as current position at the next read or write
 *       only if the next access refers to the current position instead of
 *       giving an absolute position.
 *
 * @restriction If a position outside the object is given, the result depends
 *              on the volume manager.
 *
 * @member FmSetPos
 *
 * @seealso FmOpen
 * @seealso FmMount
 */

/**
 * @macro FM_SET_POS_REQUEST
 * @brief Set next read/write position on an open object referred to by the
 *        @link "" "FmHandle" handle.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_SET_POS_REQUEST             30168
 *
 * @seealso FmSetPos
 */
#define FM_SET_POS_REQUEST             (30168)    /* !-SIGNO(struct FmSetPos)-! */

/**
 * @macro FM_SET_POS_REPLY
 * @brief Set next read/write position on an open object referred to by the
 *        @link "" "FmHandle" handle.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_SET_POS_REPLY               30169
 *
 * @seealso FmSetPos
 */
#define FM_SET_POS_REPLY               (30169)    /* !-SIGNO(struct FmSetPos)-! */

/**
 * @type FmSetPos
 *
 * @declaration
 * struct FmSetPos
 * {
 *   SIGSELECT sigNo;
 *   EfsStatus status;
 *   FmHandle  handle;
 *   EfsEnum   whence;
 *   S32       offset;
 *   U32       position;
 * };
 *
 * @input sigNo     FM_SET_POS_REQUEST
 * @input handle    @link "" "FmHandle" handle to an opened object.
 * @input whence    Origin for the new position (@link "" "FmWhenceEnum").
 * @input offset    Signed offset in bytes from origin to the new position.
 *
 * @output sigNo    FM_SET_POS_REPLY
 * @output status   EFS_SUCCESS if success (other reply data valid).
 *                  Preferred @link "errno.h" "EBADF" or old 
 *                  @link "" "FM_EINVALID_HANDLE" if the handle is invalid.
 *                  Other status codes might appear, e.g. coming from
 *                  device driver.
 * @output position Resulting new position in the file, returned as a byte
 *                  offset from the beginning of the file.
 */
struct FmSetPos
{
   SIGSELECT sigNo;
   EfsStatus status;
   FmHandle  handle;
   EfsEnum   whence;
   S32       offset;
   U32       position;
};

/**
 * @macro sendFmSetPosRequest
 *
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define sendFmSetPosRequest(handleMA, whenceMA, offsetMA, fmPidMA)
 *
 * @brief Send signal @link "" "FmSetPos" request.
 *
 * @input handleMA @link "" "FmHandle" handle to the opened object.
 * @input whenceMA Origin for the new position (@link "" "FmWhenceEnum").
 * @input offsetMA Signed offset in bytes from origin for the new position.
 * @input fmPidMA  PID of volume manager to send the request to.
 *
 * @seealso FmSetPos
 */
#define sendFmSetPosRequest(handleMA, whenceMA, offsetMA, fmPidMA) \
do { \
	/* The union is a trick to avoid breaking the strict-aliasing rules*/\
	union U \
	{ \
		union SIGNAL *s;\
		struct FmSetPos *x;\
	} u;\
        struct FmSetPos *sigMA; \
	u.s = alloc(sizeof(struct FmSetPos), FM_SET_POS_REQUEST); \
	sigMA = u.x; \
        sigMA->handle = (handleMA); \
        sigMA->whence = (whenceMA); \
        sigMA->offset = (offsetMA); \
        send(&u.s, (fmPidMA)); \
} while (0)

/**
 * @signal FmDupHandle
 *
 * @include "ose_spi/fm.sig"
 *
 * @brief Duplicate a file handle, optionally giving away the new handle to
 *        another process.
 *
 * @long This signal is sent to a volume manager process. The volume manager
 *       PID can be retrieved by taking the sender of the mount reply signal.
 * @long The sender of this request signal will become the owner of the
 *       newHandle. To duplicate a file handle and give it away, the owner of
 *       the current oldHandle can use the send_w_s() function to send the
 *       request signal and specify a new owner process of newHandle as sender.
 *       The reply signal will anyhow be sent to the owner of the oldHandle,
 *       which is known to the volume manager. If the oldHandle is invalid, it
 *       is not possible for the volume manager to send the reply to the owner.
 *       Therefore, the owner should use receive_w_tmo() to receive the reply.
 *       It is the responsibility of the old owner to somehow transfer the
 *       information about the newHandle to the other process. The owner process
 *       of the newHandle can even be in another machine with OSE.
 *
 * @member FmDupHandle
 *
 * @seealso FmOpen
 * @seealso FmMount
 */

/**
 * @macro FM_DUP_HANDLE_REQUEST
 * @brief Duplicate a file handle, optionally giving away the new handle to
 *        another process.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_DUP_HANDLE_REQUEST          30170
 *
 * @seealso FmDupHandle
 */
#define FM_DUP_HANDLE_REQUEST    (30170)    /* !-SIGNO(struct FmDupHandle)-! */

/**
 * @macro FM_DUP_HANDLE_REPLY
 * @brief Duplicate a file handle, optionally giving away the new handle to
 *        another process.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_DUP_HANDLE_REPLY            30171
 *
 * @seealso FmDupHandle
 */
#define FM_DUP_HANDLE_REPLY      (30171)    /* !-SIGNO(struct FmDupHandle)-! */

/**
 * @type FmDupHandle
 *
 * @declaration
 * struct FmDupHandle
 * {
 *   SIGSELECT sigNo;
 *   EfsStatus status;
 *   FmHandle  oldHandle;
 *   FmHandle  newHandle;
 * };
 *
 * @input sigNo      FM_DUP_HANDLE_REQUEST
 * @input oldHandle  @link "" "FmHandle" handle to the open object.
 *
 * @output sigNo     FM_DUP_HANDLE_REPLY
 * @output status    EFS_SUCCESS if success (other reply data valid).
 *                   Preferred @link "errno.h" "EBADF" or old 
 *                   @link "" "FM_EINVALID_HANDLE" if the handle is invalid.
 *                   Other status codes might appear, e.g. coming from
 *                   device driver.
 * @output newHandle New @link "" "FmHandle" handle to the same open object as
 *                   referred to by oldHandle.
 */
struct FmDupHandle
{
   SIGSELECT sigNo;
   EfsStatus status;
   FmHandle  oldHandle;
   FmHandle  newHandle;
};

/**
 * @macro sendFmDupHandleRequest
 *
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define sendFmDupHandleRequest(oldHandleMA, newOwnerMA, fmPidMA)
 *
 * @brief Send signal @link "" "FmDupHandle" request.
 *
 * @input oldHandleMA @link "" "FmHandle" handle to the open object.
 * @input newOwnerMA  PID of new owner process.
 * @input fmPidMA     PID of volume manager to send the request to.
 *
 * @restriction This macro should not be used. It uses send_w_s() function
 * which has a known race condition which can result in the signal being
 * silently lost even though the volume manager still lives. Attaching to new
 * owner does not solve this. Please consider using send_w_sender() instead.
 *
 * @seealso FmDupHandle
 */
#define sendFmDupHandleRequest(oldHandleMA, newOwnerMA, fmPidMA) \
do { \
	/* The union is a trick to avoid breaking the strict-aliasing rules*/\
	union U \
	{ \
		union SIGNAL *s;\
		struct FmDupHandle *x;\
	} u;\
        struct FmDupHandle *sigMA; \
	u.s = alloc(sizeof(struct FmDupHandle), FM_DUP_HANDLE_REQUEST); \
	sigMA = u.x; \
        sigMA->oldHandle = (oldHandleMA); \
        send_w_s(&u.s, (newOwnerMA), (fmPidMA)); \
} while (0)

/**
 * @signal FmExamine
 *
 * @include "ose_spi/fm.sig"
 *
 * @brief Request information about an open object referred to by an
 *        @link "" "FmHandle" handle.
 *
 * @long This signal is sent to a volume manager process. The volume manager
 *       PID can be retrieved by taking the sender of the mount reply signal.
 * @long The reply contains an @link "" "FmObjectInfo" structure.
 *
 * @member FmExamineRequest
 * @member FmExamineReply
 *
 * @seealso FmOpen
 * @seealso FmMount
 */

/**
 * @macro FM_EXAMINE_REQUEST
 * @brief Request information about an open object referred to by an
 *        @link "" "FmHandle" handle.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_EXAMINE_REQUEST             30172
 *
 * @seealso FmExamineRequest
 */
#define FM_EXAMINE_REQUEST    (30172)    /* !-SIGNO(struct FmExamineRequest)-! */

/**
 * @macro FM_EXAMINE_REPLY
 * @brief Request information about an open object referred to by an
 *        @link "" "FmHandle" handle.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_EXAMINE_REPLY               30173
 *
 * @seealso FmExamineReply
 */
#define FM_EXAMINE_REPLY      (30173)    /* !-SIGNO(struct FmExamineReply)-! */

/**
 * @type FmExamineRequest
 *
 * @declaration
 * struct FmExamineRequest
 * {
 *   SIGSELECT sigNo;
 *   FmHandle  handle;
 * };
 *
 * @input sigNo  FM_EXAMINE_REQUEST
 * @input handle @link "" "FmHandle" handle of the object to examine.
 */
struct FmExamineRequest
{
   SIGSELECT sigNo;
   FmHandle  handle;
};

/**
 * @type FmExamineReply
 *
 * @declaration
 * struct FmExamineReply
 * {
 *   SIGSELECT           sigNo;
 *   EfsStatus           status;
 *   FmHandle            handle;
 *   struct FmObjectInfo info;
 * };
 *
 * @output sigNo   FM_EXAMINE_REPLY
 * @output status  EFS_SUCCESS if success (other reply data valid).
 *                 Other status codes might appear, e.g. coming from
 *                 device driver.
 * @output handle  @link "" "FmHandle" handle of the object examined.
 * @output info    @link "" "FmObjectInfo" structure describing the open object.
 */
struct FmExamineReply
{
   SIGSELECT           sigNo;
   EfsStatus           status;
   FmHandle            handle;
   struct FmObjectInfo info;
};

/**
 * @macro sendFmExamineRequest
 *
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define sendFmExamineRequest(handleMA, fmPidMA)
 *
 * @brief Send signal @link "" "FmExamine" request.
 *
 * @input handleMA @link "" "FmHandle" handle of the object to examine.
 * @input fmPidMA  PID of volume manager to send the request to.
 *
 * @seealso FmExamine
 */
#define sendFmExamineRequest(handleMA, fmPidMA) \
do { \
	/* The union is a trick to avoid breaking the strict-aliasing rules*/\
	union U \
	{ \
		union SIGNAL *s;\
		struct FmExamineRequest *x;\
	} u;\
        struct FmExamineRequest *sigMA; \
	u.s = alloc(sizeof(struct FmExamineRequest), FM_EXAMINE_REQUEST); \
	sigMA = u.x; \
        sigMA->handle = handleMA; \
        send(&u.s, (fmPidMA)); \
} while (0)

/**
 * @signal FmSetSize
 *
 * @include "ose_spi/fm.sig"
 *
 * @brief Resizes the open object referred to by the @link "" "FmHandle" handle.
 *
 * @long This signal is sent to a volume manager process. The volume manager
 *       PID can be retrieved by taking the sender of the mount reply signal.
 * @long The new size (in bytes) can be any value between the current size and
 *       zero, including.
 *
 * @restriction Extending a file beyond its current size is only possible if the
 *              volume manager supports it.
 *              The normal behavior of OSE's volume managers is to support
 *              truncate to length zero or current size.
 *
 * @member FmSetSize
 *
 * @seealso FmOpen
 * @seealso FmMount
 */

/**
 * @macro FM_SET_SIZE_REQUEST
 * @brief Resizes the open object referred to by the @link "" "FmHandle" handle.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_SET_SIZE_REQUEST            30174
 *
 * @seealso FmSetSize
 */
#define FM_SET_SIZE_REQUEST      (30174)    /* !-SIGNO(struct FmSetSize)-! */

/**
 * @macro FM_SET_SIZE_REPLY
 * @brief Resizes the open object referred to by the @link "" "FmHandle" handle.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_SET_SIZE_REPLY              30175
 *
 * @seealso FmSetSize
 */
#define FM_SET_SIZE_REPLY        (30175)    /* !-SIGNO(struct FmSetSize)-! */

/**
 * @type FmSetSize
 *
 * @declaration
 * struct FmSetSize
 * {
 *   SIGSELECT sigNo;
 *   EfsStatus status;
 *   FmHandle  handle;
 *   U32       size;
 * };
 *
 * @input sigNo   FM_SET_SIZE_REQUEST
 * @input handle  @link "" "FmHandle" handle to the open object.
 * @input size    New size of the object in number of bytes.
 *
 * @output sigNo  FM_SET_SIZE_REPLY
 * @output status EFS_SUCCESS if success.
 *                Preferred @link "errno.h" "ENOSYS" or old
 *                FM_EOPER_NOT_IMPLEMENTED e.g. if trying to extend a file
 *                when it is not implemented.
 *                Other status codes might appear, e.g. coming from
 *                device driver.
 */
struct FmSetSize
{
   SIGSELECT sigNo;
   EfsStatus status;
   FmHandle  handle;
   U32       size;
};

/**
 * @macro sendFmSetSizeRequest
 *
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define sendFmSetSizeRequest(handleMA, sizeMA, fmPidMA)
 *
 * @brief Send signal @link "" "FmSetSize" request.
 *
 * @input handleMA @link "" "FmHandle" handle to the open object.
 * @input sizeMA   New size of the object in number of bytes.
 * @input fmPidMA  PID of volume manager to send the request to.
 *
 * @seealso FmSetSize
 */
#define sendFmSetSizeRequest(handleMA, sizeMA, fmPidMA) \
do { \
	/* The union is a trick to avoid breaking the strict-aliasing rules*/\
	union U \
	{ \
		union SIGNAL *s;\
		struct FmSetSize *x;\
	} u;\
        struct FmSetSize *sigMA; \
	u.s = alloc(sizeof(struct FmSetSize), FM_SET_SIZE_REQUEST); \
	sigMA = u.x; \
        sigMA->handle = (handleMA); \
        sigMA->size = (sizeMA); \
        send(&u.s, (fmPidMA)); \
} while (0)

/**
 * @signal FmSetAttr
 *
 * @include "ose_spi/fm.sig"
 *
 * @brief Changes the permission attributes of an open object referred to by an
 *        @link "" "FmHandle" handle.
 *
 * @long This signal is sent to a volume manager process. The volume manager
 *       PID can be retrieved by taking the sender of the mount reply signal.
 * @long The mask and attrs arguments define new values of selected attributes.
 *
 * @restriction Depending on the volume manager, only some of the attributes
 *              might be implemented.
 *
 * @member FmSetAttr
 *
 * @seealso FmOpen
 * @seealso FmMount
 */

/**
 * @macro FM_SET_ATTR_REQUEST
 * @brief Changes the permission attributes of an open object referred to by an
 *        @link "" "FmHandle" handle.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_SET_ATTR_REQUEST            30176
 *
 * @seealso FmSetAttr
 */
#define FM_SET_ATTR_REQUEST      (30176)    /* !-SIGNO(struct FmSetAttr)-! */

/**
 * @macro FM_SET_ATTR_REPLY
 * @brief Changes the permission attributes of an open object referred to by an
 *        @link "" "FmHandle" handle.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_SET_ATTR_REPLY              30177
 *
 * @seealso FmSetAttr
 */
#define FM_SET_ATTR_REPLY        (30177)    /* !-SIGNO(struct FmSetAttr)-! */

/**
 * @type FmSetAttr
 *
 * @declaration
 * struct FmSetAttr
 * {
 *   SIGSELECT  sigNo;
 *   EfsStatus  status;
 *   FmHandle   handle;
 *   EfsBitmask mask;
 *   EfsBitmask attrs;
 * };
 *
 * @input sigNo   FM_SET_ATTR_REQUEST
 * @input handle  @link "" "FmHandle" handle to the open object.
 * @input mask    Which attribute bits (@link "" "FmAttrEnum") to modify (set or
 *                clear according to attrs).
 * @input attrs   New attributes (@link "" "FmAttrEnum") for the object. Only
 *                the bits corresponding to the mask are changed.
 *
 * @output sigNo  FM_SET_ATTR_REPLY
 * @output status EFS_SUCCESS if success.
 *                Preferred @link "errno.h" "EROFS" or old
 *                @link "" "FM_EDISK_WRITE_PROTECT" if the
 *                volume is not writable.
 *                Other status codes might appear, e.g. coming from
 *                device driver.
 */
struct FmSetAttr
{
   SIGSELECT  sigNo;
   EfsStatus  status;
   FmHandle   handle;
   EfsBitmask mask;
   EfsBitmask attrs;
};

/**
 * @signal FmDrain
 *
 * @include "ose_spi/fm.sig"
 *
 * @brief Forces all not-yet-written data buffers for the file referred to by
 *        the @link "" "FmHandle" handle out to the device driver.
 *
 * @long This signal is sent to a volume manager process. The volume manager
 *       PID can be retrieved by taking the sender of the mount reply signal.
 * @long The reply signal is not sent until the data is stored on the physical
 *       storage.
 *
 * @member FmDrain
 *
 * @seealso FmOpen
 * @seealso FmMount
 */

/**
 * @macro FM_DRAIN_REQUEST
 * @brief Forces all not-yet-written data buffers for the file referred to by
 *        the @link "" "FmHandle" handle out to the device driver.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_DRAIN_REQUEST               30178
 *
 * @seealso FmDrain
 */
#define FM_DRAIN_REQUEST               (30178)    /* !-SIGNO(struct FmDrain)-! */

/**
 * @macro FM_DRAIN_REPLY
 * @brief Forces all not-yet-written data buffers for the file referred to by
 *        the @link "" "FmHandle" handle out to the device driver.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_DRAIN_REPLY                 30179
 *
 * @seealso FmDrain
 */
#define FM_DRAIN_REPLY                 (30179)    /* !-SIGNO(struct FmDrain)-! */

/**
 * @type FmDrain
 *
 * @declaration
 * struct FmDrain
 * {
 *   SIGSELECT sigNo;
 *   EfsStatus status;
 *   FmHandle  handle;
 * };
 *
 * @input sigNo   FM_DRAIN_REQUEST
 * @input handle  @link "" "FmHandle" handle to the open object.
 *
 * @output sigNo  FM_DRAIN_REPLY
 * @output status EFS_SUCCESS if success.
 *                @link "" "FM_EDISK_FULL" if there is not space
 *                enough on the device.
 *                Other status codes might appear, e.g. coming from
 *                device driver.
 */
struct FmDrain
{
   SIGSELECT sigNo;
   EfsStatus status;
   FmHandle  handle;
};

/**
 * @macro sendFmDrainRequest
 *
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define sendFmDrainRequest(handleMA, fmPidMA)
 *
 * @brief Send signal @link "" "FmDrain" request.
 *
 * @input handleMA @link "" "FmHandle" handle to the open object.
 * @input fmPidMA  PID of volume manager to send the request to.
 *
 * @seealso FmDrain
 */
#define sendFmDrainRequest(handleMA, fmPidMA) \
do { \
	/* The union is a trick to avoid breaking the strict-aliasing rules*/\
	union U \
	{ \
		union SIGNAL *s;\
		struct FmDrain *x;\
	} u;\
        struct FmDrain *sigMA; \
	u.s = alloc(sizeof(struct FmDrain), FM_DRAIN_REQUEST); \
	sigMA = u.x; \
        sigMA->handle = (handleMA); \
        send(&u.s, (fmPidMA)); \
} while (0)

/**
 * @signal FmLock
 *
 * @include "ose_spi/fm.sig"
 *
 * @brief Performs a lock operation on a part of an open object referred to by
 *        an @link "" "FmHandle" handle. Not yet implemented!
 *
 * @long This signal is sent to a volume manager process. The volume manager
 *       PID can be retrieved by taking the sender of the mount reply signal.
 * @long For a list of possible lock operations, please refer to @link ""
 *       "FmLkopEnum". For a list of possible lock types, please refer to
 *       @link "" "FmLktpEnum".
 *
 * @restriction Not yet implemented!
 *
 * @member FmLock
 *
 * @seealso FmOpen
 * @seealso FmMount
 */

/**
 * @macro FM_LOCK_REQUEST
 * @brief Performs a lock operation on a part of an open object referred to by
 *        an @link "" "FmHandle" handle. Not yet implemented!
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_LOCK_REQUEST                30182
 *
 * @seealso FmLock
 */
#define FM_LOCK_REQUEST                (30182)    /* !-SIGNO(struct FmLock)-! */

/**
 * @macro FM_LOCK_REPLY
 * @brief Performs a lock operation on a part of an open object referred to
 *        by an @link "" "FmHandle" handle. Not yet implemented!
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_LOCK_REPLY                  30183
 *
 * @seealso FmLock
 */
#define FM_LOCK_REPLY                  (30183)    /* !-SIGNO(struct FmLock)-! */

/**
 * @type FmLock
 *
 * @declaration
 * struct FmLock
 * {
 *   SIGSELECT sigNo;
 *   EfsStatus status;
 *   FmHandle  handle;
 *   EfsEnum   oper;
 *   EfsEnum   type;
 *   EfsEnum   whence;
 *   S32       offset;
 *   U32       length;
 *   PROCESS   owner;
 * };
 *
 * @input sigNo   FM_LOCK_REQUEST
 * @input handle  @link "" "FmHandle" handle to the open object.
 * @input oper    Type of locking operation (@link "" "FmLkopEnum").
 * @input type    Type of lock (@link "" "FmLktpEnum").
 * @input whence  Origin for the lock (@link "" "FmWhenceEnum").
 * @input offset  Signed offset in bytes from the origin, indicating at what
 *                position in the file the locked section starts.
 * @input length  Length in bytes of the section with a lock (0 implies
 *                until EOF).
 *
 * @output sigNo  FM_LOCK_REPLY
 * @output status EFS_SUCCESS if success (other reply data valid).
 *                Preferred @link "errno.h" "ENOSYS" or old
 *                @link "" "FM_EOPER_NOT_IMPLEMENTED" if the
 *                operation is not implemented.
 *                Other status codes might appear, e.g. coming from
 *                device driver.
 * @output owner  Owner process ID of the lock reported by FM_LKOP_INFO
 *                (@link "" "FmLkopEnum").
 */
struct FmLock
{
   SIGSELECT sigNo;
   EfsStatus status;
   FmHandle  handle;
   EfsEnum   oper;
   EfsEnum   type;
   EfsEnum   whence;
   S32       offset;
   U32       length;
   PROCESS   owner;
};

/**
 * @macro sendFmLockRequest
 *
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define sendFmLockRequest(handleMA, operMA, typeMA, whenceMA, offsetMA,
 *                           lengthMA, fmPidMA)
 *
 * @brief Send signal @link "" "FmLock" request.
 *
 * @input handleMA @link "" "FmHandle" handle to the open object.
 * @input operMA   Type of locking operation (@link "" "FmLkopEnum").
 * @input typeMA   Type of lock (@link "" "FmLktpEnum").
 * @input whenceMA Origin for the lock (@link "" "FmWhenceEnum").
 * @input offsetMA Signed offset in bytes from the origin, indicating at what
 *                 position in the file the locked section starts.
 * @input lengthMA Length in bytes of the section with a lock (0 implies until
 *                 EOF).
 * @input fmPidMA  PID of volume manager to send the request to.
 *
 * @seealso FmLock
 */
#define sendFmLockRequest(handleMA, operMA, typeMA, whenceMA, offsetMA, \
                         lengthMA, fmPidMA) \
do { \
	/* The union is a trick to avoid breaking the strict-aliasing rules*/\
	union U \
	{ \
		union SIGNAL *s;\
		struct FmLock *x;\
	} u;\
        struct FmLock *sigMA; \
	u.s = alloc(sizeof(struct FmLock), FM_LOCK_REQUEST); \
	sigMA = u.x; \
        sigMA->handle = (handleMA); \
        sigMA->oper = (operMA); \
        sigMA->type = (typeMA); \
        sigMA->whence = (whenceMA); \
        sigMA->offset = (offsetMA); \
        sigMA->length = (lengthMA); \
        send(&u.s, (fmPidMA)); \
} while (0)

/**
 * @signal FmPoll
 *
 * @include "ose_spi/fm.sig"
 *
 * @brief Signal to wait for I/O events on a number of sockets.
 *
 * @long This signal is sent to a volume manager process. The volume manager
 *       PID can be retrieved by taking the sender of the mount reply signal.
 * @long The FmPoll signal provides applications with a mechanism for  waiting
 *       for I/O over a set of file handles. For each entry in the  array  'fds'
 *       the volume manager examines the given file handle for the event(s)
 *       specified in 'events'. The number of pollhandle structures in  the
 *       'fds' array is specified by nfds. The volume manager identifies those
 *       files on which an application can read or write data, or on which
 *        certain events have occurred.
 *
 * @member FmPoll
 *
 * @seealso FmMount
 */

/**
 * @macro FM_POLL_REQUEST
 * @brief Signal to wait for I/O events on a number of sockets.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_POLL_REQUEST                30184
 *
 * @seealso FmPoll
 */
#define FM_POLL_REQUEST                (30184)    /* !-SIGNO(struct FmPoll)-! */

/**
 * @macro FM_POLL_REPLY
 * @brief Signal to wait for I/O events on a number of sockets.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_POLL_REPLY                  30185
 *
 * @seealso FmPoll
 */
#define FM_POLL_REPLY                  (30185)    /* !-SIGNO(struct FmPoll)-! */

/**
 * @macro FM_CANCEL_POLL
 * @brief Cancel waiting for I/O.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_CANCEL_POLL                 30186
 *
 * @seealso FmPoll
 */
#define FM_CANCEL_POLL                 (30186)    /* !-SIGNO(struct FmPoll)-! */

/**
 * @type FmPoll
 *
 * @declaration
 * struct FmPoll
 * {
 *   SIGSELECT          sig_no;
 *   EfsStatus          status;
 *   U32                result;
 *   S32                nfds;
 *   struct pollhandle {
 *      FmHandle        handle;
 *      U16             events;
 *      U16             revents;
 *   } fds[1];
 * };
 *
 * @input sig_no   FM_POLL_REQUEST
 * @input nfds     Number of socket descriptors in 'fds' array.
 * @input fds      Descriptor array (variable size).
 *                         @break This field is of variable size. The size of
 *                         the data must be added when calculating the
 *                         required signal buffer size. Excess space should
 *                         be ignored if the buffer is larger than needed.
 * @input fds.handle   The file handle to be examined.
 * @input fds.events   Requested events for the file handle. events is a bitmask
 *                 combined of event flags defined in sys/poll.h.
 *
 * @output sig_no  FM_POLL_REPLY
 * @output status  The status of the request. Zero on success.
 *                 Negative on error (errno).
 * @output result  Number of ready sockets. Only valid on success.
 * @output fds.handle  The file handle that was examined.
 * @output fds.revents Returned events for the file handle. revents is a bitmask
 *                 combined of event flags defined in sys/poll.h.
 *                 Bits are set in the revents bitmask to indicate which of the
 *                 requested events are  true.
 *
 * @input sig_no   FM_CANCEL_POLL
 * @input nfds     Number of socket descriptors in 'fds' array.
 * @input fds      Descriptor array (variable size).
 * @input fds.handle   The file handle to be examined.
 * @input fds.events   Requested events for the file handle. events is a bitmask
 *                 combined of event flags defined in sys/poll.h.
 */
struct FmPoll
{
   SIGSELECT          sig_no;
   EfsStatus          status;
   U32                result;
   S32                nfds;
   struct pollhandle {
      FmHandle        handle;
      U16             events;
      U16             revents;
   } fds[1];
};

/**
 * @signal FmFcntl
 *
 * @include "ose_spi/fm.sig"
 *
 * @brief Signal provided for future compatibility.
 *
 * @long The function is not implemented in all file systems!
 *
 * @member FmFcntl
 */

/**
 * @macro FM_FCNTL_REQUEST
 * @brief Signal provided for future compatibility.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_FCNTL_REQUEST               30188
 *
 * @seealso FmFcntl
 */
#define FM_FCNTL_REQUEST               (30188)    /* !-SIGNO(struct FmFcntl)-! */

/**
 * @macro FM_FCNTL_REPLY
 * @brief Signal provided for future compatibility.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_FCNTL_REPLY                 30189
 *
 * @seealso FmFcntl
 */
#define FM_FCNTL_REPLY                 (30189)    /* !-SIGNO(struct FmFcntl)-! */

/**
 * @type FmFcntl
 *
 * @declaration
 * struct FmFcntl
 * {
 *   SIGSELECT sigNo;
 *   S32       status;
 *   FmHandle  handle;
 *   S32       cmd;
 *   S32       arg;
 *   S32       result;
 *   S32       reserved[2];
 * };
 *
 * @input sigNo     FM_FCTNL_REQUEST
 * @input handle    File handle.
 * @input cmd       Command code.
 * @input arg       Generic argument.
 * @input reserved  Reserved for future expansion. Must be zero.
 *
 * @output sigNo    FM_FCNTL_REPLY
 * @output status   The status of the request. Zero on success. Negative on
 *                  error (errno).
 * @output handle   File handle.
 * @output result   On success this field hold the non-negative result of the
 *                  operation.
 * @output reserved Reserved for future expansion.
 */
struct FmFcntl
{
   SIGSELECT sigNo;
   S32       status;
   FmHandle  handle;
   S32       cmd;
   S32       arg;
   S32       result;
   S32       reserved[2];
};

/*
 ****************************************************************************
 * Special signals only to a terminal type Volume Manager.
 ****************************************************************************
 */

/**
 * @signal FmFlushTerm
 *
 * @include "ose_spi/fm.sig"
 *
 * @brief Discards any not-yet-written data in the transmit queue and/or
 *        not-yet-read data
 *        in the receive queue for the terminal channel referred to by the
 *        @link "" "FmHandle" handle.
 *
 * @long This signal is sent to a volume manager process. The volume manager
 *       PID can be retrieved by taking the sender of the mount reply signal.
 * @long The queue parameter defines if transmit, receive or both queues shall
 *       be cleared.
 * @long For a list of possible queues, please refer to @link "" "FmQueueEnum".
 *
 * @member FmFlushTerm
 *
 * @seealso FmOpen
 * @seealso FmMount
 */

/**
 * @macro FM_FLUSH_TERM_REQUEST
 * @brief Discards any not-yet-written data in the transmit queue and/or
 *        not-yet-read data
 *        in the receive queue for the terminal channel referred to by the
 *        @link "" "FmHandle" handle.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_FLUSH_TERM_REQUEST          30190
 *
 * @seealso FmFlushTerm
 */
#define FM_FLUSH_TERM_REQUEST    (30190)    /* !-SIGNO(struct FmFlushTerm)-! */

/**
 * @macro FM_FLUSH_TERM_REPLY
 * @brief Discards any not-yet-written data in the transmit queue and/or
 *        not-yet-read data
 *        in the receive queue for the terminal channel referred to by the
 *        @link "" "FmHandle" handle.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_FLUSH_TERM_REPLY            30191
 *
 * @seealso FmFlushTerm
 */
#define FM_FLUSH_TERM_REPLY      (30191)    /* !-SIGNO(struct FmFlushTerm)-! */

/**
 * @type FmFlushTerm
 *
 * @declaration
 * struct FmFlushTerm
 * {
 *   SIGSELECT  sigNo;
 *   EfsStatus  status;
 *   FmHandle   handle;
 *   EfsBitmask queue;
 * };
 *
 * @input sigNo   FM_FLUSH_TERM_REQUEST
 * @input handle  @link "" "FmHandle" handle to the open object.
 * @input queue   Which queue(s) to flush (@link "" "FmQueueEnum").
 *
 * @output sigNo  FM_FLUSH_TERM_REPLY
 * @output status EFS_SUCCESS if success.
 *                Other status codes might appear, e.g. coming from
 *                device driver.
 */
struct FmFlushTerm
{
   SIGSELECT  sigNo;
   EfsStatus  status;
   FmHandle   handle;
   EfsBitmask queue;
};

/**
 * @macro sendFmFlushTermRequest
 *
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define sendFmFlushTermRequest(handleMA, queueMA, fmPidMA)
 *
 * @brief Send signal @link "" "FmFlushTerm" request.
 *
 * @input handleMA @link "" "FmHandle" handle to the open object.
 * @input queueMA  Which queue(s) to flush (@link "" "FmQueueEnum").
 * @input fmPidMA  PID of volume manager to send the request to.
 *
 * @seealso FmFlushTerm
 */
#define sendFmFlushTermRequest(handleMA, queueMA, fmPidMA) \
do { \
	/* The union is a trick to avoid breaking the strict-aliasing rules*/\
	union U \
	{ \
		union SIGNAL *s;\
		struct FmFlushTerm *x;\
	} u;\
        struct FmFlushTerm *sigMA; \
	u.s = alloc(sizeof(struct FmFlushTerm), FM_FLUSH_TERM_REQUEST); \
	sigMA = u.x; \
        sigMA->handle = (handleMA); \
        sigMA->queue = (queueMA); \
        send(&u.s, (fmPidMA)); \
} while (0)

/**
 * @signal FmExamineTerm
 *
 * @include "ose_spi/fm.sig"
 *
 * @brief Request information about an open terminal channel referred to by an
 *        @link "" "FmHandle" handle.
 *
 * @long This signal is sent to a volume manager process. The volume manager
 *       PID can be retrieved by taking the sender of the mount reply signal.
 * @long The reply contains an @link "" "FmTerminalInfo" structure.
 *
 * @member FmExamineTermRequest
 * @member FmExamineTermReply
 *
 * @seealso FmOpen
 * @seealso FmMount
 */

/**
 * @macro FM_EXAMINE_TERM_REQUEST
 * @brief Request information about an open terminal channel referred to by an
 *        @link "" "FmHandle" handle.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_EXAMINE_TERM_REQUEST        30192
 *
 * @seealso FmExamineTermRequest
 */
#define FM_EXAMINE_TERM_REQUEST (30192) /* !-SIGNO(struct FmExamineTermRequest)-! */

/**
 * @macro FM_EXAMINE_TERM_REPLY
 * @brief Request information about an open terminal channel referred to by an
 *        @link "" "FmHandle" handle.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_EXAMINE_TERM_REPLY          30193
 *
 * @seealso FmExamineTermReply
 */
#define FM_EXAMINE_TERM_REPLY   (30193) /* !-SIGNO(struct FmExamineTermReply)-! */

/**
 * @type FmExamineTermRequest
 *
 * @declaration
 * struct FmExamineTermRequest
 * {
 *   SIGSELECT sigNo;
 *   FmHandle  handle;
 * };
 *
 * @input sigNo  FM_EXAMINE_TERM_REQUEST
 * @input handle @link "" "FmHandle" handle of the object (terminal channel)
 *               to examine.
 */
struct FmExamineTermRequest
{
   SIGSELECT sigNo;
   FmHandle  handle;
};

/**
 * @type FmExamineTermReply
 *
 * @declaration
 * struct FmExamineTermReply
 * {
 *   SIGSELECT             sigNo;
 *   EfsStatus             status;
 *   FmHandle              handle;
 *   struct FmTerminalInfo info;
 * };
 *
 * @output sigNo  FM_EXAMINE_TERM_REPLY
 * @output status EFS_SUCCESS if success (other reply data valid).
 *                Other status codes might appear, e.g. coming from
 *                device driver.
 * @output handle @link "" "FmHandle" handle of the object (terminal channel).
 * @output info   @link "" "FmTerminalInfo" structure describing the open
 *                object.
 */
struct FmExamineTermReply
{
   SIGSELECT             sigNo;
   EfsStatus             status;
   FmHandle              handle;
   struct FmTerminalInfo info;
};

/**
 * @macro sendFmExamineTermRequest
 *
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define sendFmExamineTermRequest(handleMA, fmPidMA)
 *
 * @brief Send signal @link "" "FmExamineTerm" request.
 *
 * @input handleMA @link "" "FmHandle" handle of the object (terminal channel)
 *                 to examine.
 * @input fmPidMA  PID of volume manager to send the request to.
 *
 * @seealso FmExamineTerm
 */
#define sendFmExamineTermRequest(handleMA, fmPidMA) \
do { \
	/* The union is a trick to avoid breaking the strict-aliasing rules*/\
	union U \
	{ \
		union SIGNAL *s;\
		struct FmExamineTermRequest *x;\
	} u;\
        struct FmExamineTermRequest *sigMA; \
	u.s = alloc(sizeof(struct FmExamineTermRequest), \
                    FM_EXAMINE_TERM_REQUEST); \
	sigMA = u.x; \
        sigMA->handle = (handleMA); \
        send(&u.s, (fmPidMA)); \
} while (0)

/**
 * @signal FmConfigTerm
 *
 * @include "ose_spi/fm.sig"
 *
 * @brief Changes the terminal attributes for an open terminal channel
 *        referred to by an @link "" "FmHandle" handle.
 *
 * @long This signal is sent to a volume manager process. The volume manager
 *       PID can be retrieved by taking the sender of the mount reply signal.
 * @long The configuration is given in the form of an @link ""
 *       "FmTerminalInfo" structure.
 *
 * @member FmConfigTerm
 *
 * @seealso FmOpen
 * @seealso FmMount
 */

/**
 * @macro FM_CONFIG_TERM_REQUEST
 * @brief Changes the terminal attributes for an open terminal channel
 *        referred to by an @link "" "FmHandle" handle.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_CONFIG_TERM_REQUEST         30194
 *
 * @seealso FmConfigTerm
 */
#define FM_CONFIG_TERM_REQUEST   (30194)    /* !-SIGNO(struct FmConfigTerm)-! */

/**
 * @macro FM_CONFIG_TERM_REPLY
 * @brief Changes the terminal attributes for an open terminal channel
 *        referred to by an @link "" "FmHandle" handle.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_CONFIG_TERM_REPLY           30195
 *
 * @seealso FmConfigTerm
 */
#define FM_CONFIG_TERM_REPLY     (30195)    /* !-SIGNO(struct FmConfigTerm)-! */

/**
 * @type FmConfigTerm
 *
 * @declaration
 * struct FmConfigTerm
 * {
 *   SIGSELECT             sigNo;
 *   EfsStatus             status;
 *   FmHandle              handle;
 *   struct FmTerminalInfo info;
 * };
 *
 * @input sigNo   FM_CONFIG_TERM_REQUEST
 * @input handle  @link "" "FmHandle" handle of the object (terminal channel)
 *                to change.
 * @input info    @link "" "FmTerminalInfo" structure with a new configuration
 *                for the terminal.
 *
 * @output sigNo  FM_CONFIG_TERM_REPLY
 * @output status EFS_SUCCESS if success.
 *                Other status codes might appear, e.g. coming from
 *                device driver.
 */
struct FmConfigTerm
{
   SIGSELECT             sigNo;
   EfsStatus             status;
   FmHandle              handle;
   struct FmTerminalInfo info;
};

/**
 * @macro sendFmConfigTermRequest
 *
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define sendFmConfigTermRequest(handleMA, infoMA, fmPidMA)
 *
 * @brief Send signal @link "" "FmConfigTerm" request.
 *
 * @input handleMA @link "" "FmHandle" handle of the object (terminal channel)
 *                 to change.
 * @input infoMA   @link "" "FmTerminalInfo" structure with a new configuration
 *                 for the terminal.
 * @input fmPidMA  PID of volume manager to send the request to.
 *
 * @seealso FmConfigTerm
 */
#define sendFmConfigTermRequest(handleMA, infoMA, fmPidMA) \
do { \
	/* The union is a trick to avoid breaking the strict-aliasing rules*/\
	union U \
	{ \
		union SIGNAL *s;\
		struct FmConfigTerm *x;\
	} u;\
        struct FmConfigTerm *sigMA; \
	u.s = alloc(sizeof(struct FmConfigTerm), FM_CONFIG_TERM_REQUEST); \
	sigMA = u.x; \
        sigMA->handle = (handleMA); \
        sigMA->info = (infoMA); \
        send(&u.s, (fmPidMA)); \
} while (0)

/**
 * @macro FM_SIG_INTERRUPT
 * @brief  Flag for interrupt event (By default triggered by ctrl-C)
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_SIG_INTERRUPT    0x1
 *
 * @sealso FmSubscribeSignal
 * @sealso FmSignal
 */
#define FM_SIG_INTERRUPT    0x1

/**
 * @macro FM_SIG_SUSPEND
 * @brief  Flag for suspend event (By default triggered by ctrl-Z)
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_SIG_SUSPEND      0x2
 *
 * @sealso FmSubscribeSignal
 * @sealso FmSignal
 */
#define FM_SIG_SUSPEND      0x2

/**
 * @signal FmSubscribeSignal
 * @brief Request to subscribe for notification of asynchronous terminal
 *        events.
 * @include "ose_spi/fm.sig"
 *
 * @long Request to receive @link "" "FmSignal" signals from the terminal
 *       when the terminal encounters special control characters.
 *       The available notifications are:
 *       @link "" "FM_SIG_INTERRUPT" and @link "" "FM_SIG_SUSPEND".
 *
 * @restriction Only intended for use between OSE shell and terminal.
 *              Suitability for other use is not guaranteed.
 *
 * @member FmSubscribeSignal
 */

/**
 * @macro FM_SUBSCRIBE_SIGNAL_REQUEST
 * @brief Request to subscribe for notification of asynchronous terminal
 *        events.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_SUBSCRIBE_SIGNAL_REQUEST 30196
 *
 * @seealso FmSubscribeSignal
 */
#define FM_SUBSCRIBE_SIGNAL_REQUEST (30196) /* !-SIGNO(struct FmSubscribeSignal)-! */

/**
 * @macro FM_SUBSCRIBE_SIGNAL_REPLY
 * @brief Acknowledge of subscription for notification of asynchronous terminal
 *        events.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_SUBSCRIBE_SIGNAL_REPLY   30197
 *
 * @seealso FmSubscribeSignal
 */
#define FM_SUBSCRIBE_SIGNAL_REPLY (30197) /* !-SIGNO(struct FmSubscribeSignal)-! */

/**
 * @type FmSubscribeSignal
 * @brief Request to subscribe for notification of asynchronous terminal
 *        events.
 * @declaration
 * struct FmSubscribeSignal
 * {
 *   SIGSELECT  sigNo;
 *   EfsStatus  status;
 *   FmHandle   handle;
 *   EfsBitmask signals;
 * };
 *
 * @input sigNo   FM_SUBSCRIBE_SIGNAL_REQUEST
 * @input handle  Handle of terminal.
 * @input signals Bit mask of the event signals to enable. Available flags are:
 *                @link "" "FM_SIG_INTERRUPT" and @link "" "FM_SIG_SUSPEND".
 *
 * @output sigNo  FM_SUBSCRIBE_SIGNAL_REPLY
 * @output status EFS_SUCCESS if success.
 *                Other status codes might appear, e.g. coming from
 *                device driver.
 *
 * @seealso FmSubscribeSignal
 */
struct FmSubscribeSignal
{
   SIGSELECT  sigNo;
   EfsStatus  status;
   FmHandle   handle;
   EfsBitmask signals;
};

/**
 * @signal FmSignal
 * @brief Notification of asynchronous terminal event.
 * @include "ose_spi/fm.sig"
 *
 * @long This signal is sent from the terminal to the client that has
 *       requested notification when the terminal detects a special control
 *       character.
 *
 * @long The subscription is activated with the @link "" "FmSubscribeSignal"
 *       signal.
 *
 * @restriction Only intended for use between OSE shell and terminal.
 *              Suitability for other use is not guaranteed.
 *
 * @member FmSignal
 */

/**
 * @macro FM_SIGNAL
 * @brief Notification of asynchronous terminal event.
 * @declaration
 * #define FM_SIGNAL                      30198
 *
 * @seealso FmSignal
 */
#define FM_SIGNAL                      (30198)    /* !-SIGNO(struct FmSignal)-! */

/**
 * @type FmSignal
 * @brief Notification of asynchronous terminal event.
 * @declaration
 * struct FmSignal
 * {
 *   SIGSELECT  sigNo;
 *   FmHandle   handle;
 *   EfsBitmask signal;
 * };
 *
 * @field sigNo  FM_SIGNAL
 * @field handle Handle of terminal.
 * @field signal Bit mask corresponding to the detected event, corresponding
 *               to one of the signals requested via @link "" "FmSubscribeSignal".
 */
struct FmSignal
{
   SIGSELECT  sigNo;
   FmHandle   handle;
   EfsBitmask signal;
};


/*
 ****************************************************************************
 * Special notification signals to a Volume Manager.
 ****************************************************************************
 */

/**
 * @signal FmAddNotification
 *
 * @include "ose_spi/fm.sig"
 *
 * @brief Creates a request to the volume manager to add a notification entry
 *        into the notification list when a process wishes to monitor all
 *        changes to a disk object.
 *
 * @long This signal is sent to a volume manager process. The volume manager
 *       PID can be retrieved by taking the sender of the mount reply signal.
 * @long Notifications on file system events, i.e. changes on files and
 *       directories, is an extension not supported by all volume managers.
 * @long When a process sends an add notification request, the object s path is
 *       checked to
 *       verify that it is a fully qualified path. The root of the object path
 *       must match
 *       the volume name of the volume manager that will receive the request
 *       signal. If
 *       the path is OK, it will then be added to the notification list along
 *       with the requesting process PID and monitored object name.
 * @long NOTE: If the disk object is a directory, then any modification made to
 *       it or its descendants will cause a notification message to be sent.
 *       Objects that do not exist can be added to the notification list.
 *
 * @restriction This signal is only supported by the AFM volume manager.
 *
 * @member FmAddNotification
 *
 * @seealso FmMount
 */

/**
 * @macro FM_ADD_NOTIFICATION_REQUEST
 * @brief A request to the volume manager to add a notification entry
 *        into the notification list.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_ADD_NOTIFICATION_REQUEST    31300
 *
 * @seealso FmAddNotification
 */
#define FM_ADD_NOTIFICATION_REQUEST (31300) /* !-SIGNO(struct FmAddNotification)-! */

/**
 * @macro FM_ADD_NOTIFICATION_REPLY
 * @brief A reply from the volume manager when having added a notification entry
 *        into the notification list.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_ADD_NOTIFICATION_REPLY      31301
 *
 * @seealso FmAddNotification
 */
#define FM_ADD_NOTIFICATION_REPLY (31301) /* !-SIGNO(struct FmAddNotification)-! */

/**
 * @type FmAddNotification
 * @brief The information to the volume manager when wanting to add a
 *        notification entry into the notification list.
 * @declaration
 * struct FmAddNotification
 * {
 *   SIGSELECT sigNo;
 *   EfsStatus status;
 *   U32       objectpath;
 *   char      s[1];
 * };
 *
 * @input sigNo      FM_ADD_NOTIFICATION_REQUEST
 * @input objectpath Byte index into the string pool where the object's fully
 *                   qualified path is stored.
 *                         @break The full path of the object is given. This
 *                         starts with the mount point that is registered in
 *                         the FSS, including the leading slash "/".
 * @input s          String pool for variable sized data. The string
 *                   refered by objectpath is stored here.
 *                         @break This field is of variable size. The size of
 *                         the string(s) must be added when calculating the
 *                         required signal buffer size. Excess space should
 *                         be ignored if the buffer is larger than needed.
 *
 * @output sigNo     FM_ADD_NOTIFICATION_REPLY
 * @output status    EFS_SUCCESS success.
 *                   Preferred @link "errno.h" "EINVAL" or old
 *                   @link "" "FM_ENOT_ABSOLUTE_PATH" if the root
 *                   of the object path does not match the name of the volume.
 *                   Other status codes might appear, e.g. coming from
 *                   device driver.
 */
struct FmAddNotification
{
   SIGSELECT sigNo;
   EfsStatus status;
   U32       objectpath;
   char      s[1];
};

/**
 * @macro sendFmAddNotificationRequest
 * @brief Send a request to the volume manager to add a notification entry
 *        into the notification list.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define sendFmAddNotificationRequest(objectpathMA, fmPidMA)
 *
 * @brief Send signal @link "" "FmAddNotification" request.
 *
 * @input objectpathMA Pointer to the objects fully qualified path string,
 *                     to be added to the notification list.
 * @input fmPidMA      PID of volume manager to send the request to.
 *
 * @seealso FmAddNotification
 */
#define sendFmAddNotificationRequest(objectpathMA, fmPidMA) \
do { \
	/* The union is a trick to avoid breaking the strict-aliasing rules*/\
	union U \
	{ \
		union SIGNAL *s;\
		struct FmAddNotification *x;\
	} u;\
	struct FmAddNotification *sigMA; \
	const char * _objectpathMA = (objectpathMA); \
	U32 objectpathMAlen = strlen(_objectpathMA); \
	u.s = alloc(sizeof(struct FmAddNotification) + objectpathMAlen, \
                    FM_ADD_NOTIFICATION_REQUEST); \
	sigMA = u.x; \
	sigMA->objectpath = 0; \
	strcpy(&sigMA->s[sigMA->objectpath], (_objectpathMA)); \
	send(&u.s, (fmPidMA)); \
} while (0)

/**
 * @signal FmNotification
 *
 * @include "ose_spi/fm.sig"
 *
 * @brief This signal is sent to the monitoring process of an object when the
 *        object on the volume managers notification list is changed, or when a
 *        volume is either unmounted or formatted.
 *
 * @long NOTE: In the case of unmount and format events, all monitoring will
 *       receive an FmNotification signal with an unmount or format event
 *       respectively.
 * @long For a list of notification events, please refer to @link ""
 *       "FmNotificationEnum".
 *
 * @restriction This signal is only supported by the AFM volume manager.
 *
 * @member FmNotification
 *
 * @seealso FmMount
 */

/**
 * @macro FM_NOTIFICATION
 * @brief Sent to the monitoring process of an object when the object
 *        on the volume managers notification list is changed, or when a
 *        volume is either unmounted or formatted.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_NOTIFICATION                31302
 *
 * @seealso FmNotification
 */
#define FM_NOTIFICATION       (31302)    /* !-SIGNO(struct FmNotification)-! */

/**
 * @type FmNotification
 *
 * @declaration
 * struct FmNotification
 * {
 *   SIGSELECT sigNo;
 *   U32       event;
 *   U64       timeStamp;
 *   U32       name;
 *   U32       rename;
 *   char      s[2];
 * };
 *
 * @field sigNo     FM_NOTIFICATION
 * @field event     Type of notification event (@link "" "FmNotificationEnum").
 * @field timeStamp The time when the event occurred.
 * @field name      Byte offset into string pool where the name of the monitored
 *                  object is stored.
 * @field rename    Byte offset into string pool where the new name of the
 *                  renamed object is stored, if the event was a rename.
 * @field s         String pool for variable sized data. The strings
 *                  refered by name and rename are stored here.
 *                         @break This field is of variable size. The size of
 *                         the string(s) must be added when calculating the
 *                         required signal buffer size. Excess space should
 *                         be ignored if the buffer is larger than needed.
 */
struct FmNotification
{
   SIGSELECT sigNo;
   U32       event;
   U64       timeStamp;
   U32       name;
   U32       rename;
   char      s[2];
};

/**
 * @macro sendFmNotification
 *
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define sendFmNotification(eventMA, timeStampMA, nameMA, renameMA, fmPidMA)
 *
 * @brief Send signal @link "" "FmNotification".
 *
 * @field eventMA     Type of notification event (@link ""
 *                    "FmNotificationEnum").
 * @field timeStampMA The time when the event occurred.
 * @field nameMA      Name (string) of the monitored object.
 * @field renameMA    New name (string) of the renamed object, if the event
 *                    was a rename.
 * @input fmPidMA     PID of volume manager to send the request to.
 *
 * @seealso FmNotification
 */
#define sendFmNotification(eventMA, timeStampMA, nameMA, renameMA, fmPidMA) \
do { \
	/* The union is a trick to avoid breaking the strict-aliasing rules*/\
	union U \
	{ \
		union SIGNAL *s;\
		struct FmNotification *x;\
	} u;\
	struct FmNotification *sigMA; \
	const char *_nameMA = (nameMA); \
	const char *_renameMA = (renameMA); \
	unsigned nameMAlen = strlen(_nameMA); \
	unsigned renameMAlen = strlen(_renameMA); \
	u.s = alloc(sizeof(struct FmNotification) + nameMAlen + renameMAlen, \
                    FM_NOTIFICATION); \
	sigMA = u.x; \
	sigMA->name = 0; \
	sigMA->rename = nameMAlen + 1; \
	sigMA->event = eventMA; \
	sigMA->timeStamp = timeStampMA; \
	strcpy(&sigMA->s[sigMA->name], (_nameMA)); \
	strcpy(&sigMA->s[sigMA->rename], (_renameMA)); \
	send(&u.s, (fmPidMA)); \
} while (0)

/**
 * @signal FmRemoveNotification
 *
 * @include "ose_spi/fm.sig"
 *
 * @brief Creates a request to the volume manager to remove a notification entry
 *        from the volume managers notification list.
 *
 * @long This signal is sent to a volume manager process. The volume manager
 *       PID can be retrieved by taking the sender of the mount reply signal.
 * @long Removing notification of an object for a process will terminate the
 *       sending of Notification signals to that process.
 * @long NOTE: The object path must be a fully qualified path, and the root of
 *       the object path must match the volume name of the volume manager.
 *
 * @restriction This signal is only supported by the AFM volume manager.
 *
 * @member FmRemoveNotification
 *
 * @seealso FmMount
 */

/**
 * @macro FM_REMOVE_NOTIFICATION_REQUEST
 * @brief Remove a notification entry from the volume managers notification
 *        list.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_REMOVE_NOTIFICATION_REQUEST 31303
 *
 * @seealso FmRemoveNotification
 */
#define FM_REMOVE_NOTIFICATION_REQUEST (31303)    /* !-SIGNO(struct FmRemoveNotification)-! */

/**
 * @macro FM_REMOVE_NOTIFICATION_REPLY
 * @brief Remove a notification entry from the volume managers notification
 *        list.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_REMOVE_NOTIFICATION_REPLY   31304
 *
 * @seealso FmRemoveNotification
 */
#define FM_REMOVE_NOTIFICATION_REPLY   (31304)    /* !-SIGNO(struct FmRemoveNotification)-! */

/**
 * @type FmRemoveNotification
 * @brief Remove a notification entry from the volume managers notification
 *        list.
 * @declaration
 * struct FmRemoveNotification
 * {
 *   SIGSELECT sigNo;
 *   EfsStatus status;
 *   U32       objectpath;
 *   char      s[1];
 * };
 *
 * @input sigNo      FM_REMOVE_NOTIFICATION_REQUEST
 * @input objectpath Byte index into the string pool where the objects fully
 *                   qualified path is stored.
 *                         @break The full path of the object is given. This
 *                         starts with the mount point that is registered in
 *                         the FSS, including the leading slash "/".
 * @input s          String pool for variable sized data. The string
 *                   refered by objectpath is stored here.
 *                         @break This field is of variable size. The size of
 *                         the string(s) must be added when calculating the
 *                         required signal buffer size. Excess space should
 *                         be ignored if the buffer is larger than needed.
 *
 * @output sigNo     FM_REMOVE_NOTIFICATION_REPLY
 * @output status    EFS_SUCCESS if success.
 *                   Preferred @link "errno.h" "EINVAL" or old
 *                   @link "" "FM_ENOT_ABSOLUTE_PATH" if the root
 *                   of the object path does not match the name of the volume.
 *                   Other status codes might appear, e.g. coming from
 *                   device driver.
 */
struct FmRemoveNotification
{
   SIGSELECT sigNo;
   EfsStatus status;
   U32       objectpath;
   char      s[1];
};

/**
 * @macro sendFmRemoveNotificationRequest
 *
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define sendFmRemoveNotificationRequest(objectpathMA, fmPidMA)
 *
 * @brief Send signal @link "" "FmRemoveNotification" request.
 *
 * @input objectpathMA Pointer to the objects fully qualified path string,
 *                     to be removed from the notification list.
 * @input fmPidMA      PID of volume manager to send the request to.
 *
 * @seealso FmRemoveNotification
 */
#define sendFmRemoveNotificationRequest(objectpathMA, fmPidMA) \
do { \
	/* The union is a trick to avoid breaking the strict-aliasing rules*/\
	union U \
	{ \
		union SIGNAL *s;\
		struct FmRemoveNotification *x;\
	} u;\
	struct FmRemoveNotification *sigMA; \
	const char * _objectpathMA = (objectpathMA); \
	U32 objectpathMAlen = strlen(_objectpathMA); \
	u.s = alloc(sizeof(struct FmRemoveNotification) + objectpathMAlen, \
	FM_REMOVE_NOTIFICATION_REQUEST); \
	sigMA = u.x; \
	sigMA->objectpath = 0; \
	strcpy(&sigMA->s[sigMA->objectpath], (_objectpathMA)); \
	send(&u.s, (fmPidMA)); \
} while (0)


/**
 * @signal FmFallocate
 *
 * @include "ose_spi/fm.sig"
 *
 * @brief Reserve media space for the open object referred to by the
 *        @link "" "FmHandle" handle.
 *
 * @long The signal reserves any required media space for file data from offset
 *       and len bytes forward. On successful return, subsequent writes to the
 *       specified range shall not fail due to lack of free space on storage
 *       media. If the end of the range is beyond the current file size, the
 *       file size shall be adjusted to the range end.
 *
 * @long This signal is sent to a volume manager process. The volume manager
 *       PID can be retrieved by taking the sender of the mount reply signal.
 *
 * @restriction The volume manager may not support this functionality.
 *              It is advisable to use the @link "" "FmInterface" signal
 *              to verify the volume manager's recognized interfaces.
 *
 * @member FmFallocate
 *
 * @seealso FmSetSize
 * @seealso FmFadvise
 * @seealso FmMount
 */

/**
 * @macro FM_FALLOCATE_REQUEST
 * @brief Reserve media space for the open object referred to by the
 *        @link "" "FmHandle" handle.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_FALLOCATE_REQUEST  31310
 *
 * @seealso FmFallocate
 */
#define FM_FALLOCATE_REQUEST     (31310)    /* !-SIGNO(struct FmFallocate)-! */

/**
 * @macro FM_FALLOCATE_REPLY
 * @brief Reserve media space for the open object referred to by the
 *        @link "" "FmHandle" handle.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_FALLOCATE_REPLY    31311
 *
 * @seealso FmFallocate
 */
#define FM_FALLOCATE_REPLY       (31311)    /* !-SIGNO(struct FmFallocate)-! */

/**
 * @type FmFallocate
 *
 * @declaration
 * struct FmFallocate
 * {
 *   SIGSELECT sigNo;
 *   EfsStatus status;
 *   FmHandle  handle;
 *   U32       offset;
 *   U32       length;
 * };
 *
 * @input sigNo   FM_FALLOCATE_REQUEST
 * @input handle  @link "" "FmHandle" handle to the open object.
 * @input offset  Offset of region to reserve space for,
 *                in number of bytes from start of file.
 * @input length  Length of region to reserve space for, in number of bytes.
 *
 * @output sigNo  FM_FALLOCATE_REPLY
 * @output status EFS_SUCCESS if success.
 *                Preferred @link "errno.h" "ENOSYS" or old
 *                @link "" "FM_EOPER_NOT_IMPLEMENTED" if the
 *                operation is not implemented.
 *                Other status codes might appear, e.g. coming from
 *                device driver.
 */
struct FmFallocate
{
   SIGSELECT sigNo;
   EfsStatus status;
   FmHandle  handle;
   U32       offset;
   U32       length;
};

/**
 * @macro sendFmFallocateRequest
 * @brief Send signal @link "" "FmFallocate" request.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define sendFmFallocateRequest(handleMA, offsetMA, lengthMA, vmPidMA)
 *
 * @input handleMA @link "" "FmHandle" handle to the open object.
 * @input offsetMA Offset of region to reserve space for,
 *                 in number of bytes from start of file.
 * @input lengthMA Length of region to reserve space for, in number of bytes.
 * @input vmPidMA  PID of volume manager to send the request to.
 *
 * @seealso FmFallocate
 */
#define sendFmFallocateRequest(handleMA, offsetMA, lengthMA, vmPidMA) \
do { \
	/* The union is a trick to avoid breaking the strict-aliasing rules*/\
	union U \
	{ \
		union SIGNAL *s;\
		struct FmFallocate *x;\
	} u;\
        struct FmFallocate *sigMA; \
	u.s = alloc(sizeof(struct FmFallocate), FM_FALLOCATE_REQUEST); \
	sigMA = u.x; \
        sigMA->handle = (handleMA); \
        sigMA->offset = (offsetMA); \
        sigMA->length = (lengthMA); \
        send(&u.s, (vmPidMA)); \
} while (0)


/**
 * @signal FmFadvise
 *
 * @include "ose_spi/fm.sig"
 *
 * @brief Advise on expected access pattern for given
 *        @link "" "FmHandle" handle.
 *
 * @long The signal advises the volume manager about the application's
 *       expected I/O pattern for the @link "" "FmHandle" handle and within
 *       the specified data range.
 *       The signal shall have no effect on the semantics of other operations
 *       on the specified data, but the volume manager may use the
 *       provided information to optimize handling of the specified data.
 *
 * @long The range need not currently exist in the file.
 *       A zero length specifies all data following the given offset.
 *
 * @long This signal is sent to a volume manager process. The volume manager
 *       PID can be retrieved by taking the sender of the mount reply signal.
 *
 * @restriction The volume manager may not support this functionality.
 *              It is advisable to use the @link "" "FmInterface" signal
 *              to verify the volume manager's recognized interfaces.
 *
 * @member FmFadvise
 *
 * @seealso FmFallocate
 * @seealso FmMount
 */

/**
 * @macro FM_FADVISE_REQUEST
 * @brief Advise on expected access pattern for given
 *        @link "" "FmHandle" handle.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_FADVISE_REQUEST  31312
 *
 * @seealso FmFadvise
 */
#define FM_FADVISE_REQUEST     (31312)    /* !-SIGNO(struct FmFadvise)-! */

/**
 * @macro FM_FADVISE_REPLY
 * @brief Advise on expected access pattern for given
 *        @link "" "FmHandle" handle.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define FM_FADVISE_REPLY    31313
 *
 * @seealso FmFadvise
 */
#define FM_FADVISE_REPLY       (31313)    /* !-SIGNO(struct FmFadvise)-! */

/**
 * @type FmFadvise
 *
 * @input sigNo   FM_FADVISE_REQUEST
 * @input handle  @link "" "FmHandle" handle to the open object.
 * @input offset  Offset of region that advice applies to,
 *                in number of bytes from start of file.
 * @input length  Length in bytes of region that advice applies to.
 * @input advice  Expected access pattern for the handle and data range
 *                (@link "" "FmFadvEnum").
 *
 * @output sigNo  FM_FADVISE_REPLY
 * @output status EFS_SUCCESS if success.
 *                Preferred @link "errno.h" "ENOSYS" or old
 *                @link "" "FM_EOPER_NOT_IMPLEMENTED" if the
 *                operation is not implemented.
 *                Other status codes might appear, e.g. coming from
 *                device driver.
 */
struct FmFadvise
{
   SIGSELECT sigNo;
   EfsStatus status;
   FmHandle  handle;
   U32       offset;
   U32       length;
   EfsEnum   advice;
};

/**
 * @macro sendFmFadviseRequest
 * @brief Send signal @link "" "FmFadvise" request.
 * @include "ose_spi/fm.sig"
 *
 * @declaration
 * #define sendFmFadviseRequest(handleMA, offsetMA, lengthMA, vmPidMA)
 *
 * @input handleMA @link "" "FmHandle" handle to the open object.
 * @input offsetMA Offset of region that advice applies to,
 *                 in number of bytes from start of file.
 * @input lengthMA Length in bytes of region that advice applies to.
 * @input adviceMA Expected access pattern for the handle and data range
 *                 (@link "" "FmFadvEnum").
 * @input vmPidMA  PID of volume manager to send the request to.
 *
 * @seealso FmFadvise
 */
#define sendFmFadviseRequest(handleMA, offsetMA, lengthMA, adviceMA, vmPidMA) \
do { \
	/* The union is a trick to avoid breaking the strict-aliasing rules*/\
	union U \
	{ \
		union SIGNAL *s;\
		struct FmFadvise *x;\
	} u;\
	struct FmFadvise *sigMA; \
	u.s = alloc(sizeof(struct FmFadvise), FM_FADVISE_REQUEST); \
	sigMA = u.x; \
	sigMA->handle = (handleMA); \
	sigMA->offset = (offsetMA); \
	sigMA->length = (lengthMA); \
	sigMA->advice = (adviceMA); \
	send(&u.s, (vmPidMA)); \
} while (0)

/*
 * The following signals was used by EFS notifications, which were introduced
 * in KSSS in OSE 5.4 and removed in OSE 5.5.
 */

#define FM_DEPRECATED_31314 31314
#define FM_DEPRECATED_31315 31315
#define FM_DEPRECATED_31316 31316
#define FM_DEPRECATED_31317 31317
#define FM_DEPRECATED_31318 31318
#define FM_DEPRECATED_31319 31319
#define FM_DEPRECATED_31320 31320
#define FM_DEPRECATED_31321 31321

/* The purpose of this define is to have a top value for the signals in this 
 * range.
 */
#define FM_RANGE_MAX 31399

#ifdef __cplusplus
} /* end of extern "C" */
#endif

#endif /* _FM_SIG */
