/* COPYRIGHT-ENEA-SRC-R2 *
 **************************************************************************
 * Copyright (C) 2004-2010 by Enea Software AB.
 * All rights reserved.
 *
 * This Software is furnished under a software license agreement and
 * may be used only in accordance with the terms of such agreement.
 * Any other use or reproduction is prohibited. No title to and
 * ownership of the Software is hereby transferred.
 *
 * PROPRIETARY NOTICE
 * This Software consists of confidential information.
 * Trade secret law and copyright law protect this Software.
 * The above notice of copyright on this Software does not indicate
 * any actual or intended publication of such Software.
 **************************************************************************
 * COPYRIGHT-END */

/**
 *  @File rmm.c
 *  @brief This is the implementation of the run mode monitor
 *
 *  @long Terminates the monitor*.sig interface used by source code debuggers
 *  and system level debuggers.
 *
 *  @long Replaces LDM and SDM in the OSE4 architecture and runs as part of
 *  the kernel.
 *
 * @long
 *  Internal RMM signals start at signo 35950 and ends at signo 35999,
 *  external at signo 39000.
 *
 *  @long The RMM process (ose_monitor) can only exist in one instance. But
 *  client serving processes is created as part of connect.
 */

#include "stddef.h"
#include "stdlib.h"
#include "assert.h"
#include "string.h"
#include "stdint.h"
#include "stdio.h"
#include "stdbool.h"
#include "compiler.h"

/*
 * NOTE: This redirects system calls to direct function calls to pose.
 * This also bypasses ms to tick translation for calls like receive_w_tmo().
 */
#include "osenames.h"
#include "ose.h"
#include "ose_spi.h"
#include "osetypes.h"
#include "ose_err.h"
#include "osarch.h"
#include "ramlog.h"
#include "sysparam.h"
#include "atomic.h"
#include "ramtrace.h"
#include "rmm_err.h"
#ifdef OSARM
#include "ose_reports/heap_reports.h"
#endif
#ifdef OSE_SMP
#include "spinlock.h"
#endif

#ifdef USE_RAMDUMP
#   include "ramdump.h"
#endif

#ifndef softOSE
#   define USE_EPM
#endif

#include "pose.con"
#include "pose.h"
#include "list.h"

#ifdef USE_EPM
#   include "efmi.h"
#   include "efmi_registers.h"
#   include "epm.h"
#endif

#include "rmm_internal.sig"
#include "mm.sig"
#include "heap.sig"
#include "heap_private.sig"

#include "monitor.sig"
#include "monitor_actions.sig"
#include "monitor_execution.sig"
#include "monitor_events.sig"
#include "monitor_cpu.sig"
#include "monitor_dumps.sig"
#include "monitor_tracepoint.sig"
#include "monitor_counters.sig"

#include "rmm.h"
#include "rmm_ramtrace.h"
#include "rmm_utility.h"
#ifdef OSARM
#include "rmm_internal.h"
#endif
#include "ose_err.h"
#include "cpu_access.h"

#ifndef SMP_CODE_YES_NO
#define SMP_CODE(a)
#define SMP_ARRAY
#define SMP_CODE_YES_NO(a, b) (b)
#endif

#ifndef SMP_INDEX_X
#define SMP_INDEX_X(a)
#define zzget_dp_look_aside(a) zzdp_look_aside
#endif

#ifndef BKL_LOCK_SAVE
#define BKL_LOCK_SAVE(m) LOCK_SAVE((m))
#define BKL_LOCK_RESTORE(m) LOCK_RESTORE((m))
#endif

#ifndef BKL_LOCK
#define BKL_LOCK LOCK
#define BKL_UNLOCK UNLOCK
#endif

#ifndef UINT32_MAX
#define UINT32_MAX 0xffffffff
#endif

#ifndef INT32_MAX
#define INT32_MAX 0x7fffffff
#endif

/*
** Avoid kernel calls via crt, particulary in the warm start sequence,
** where the crt is not yet initialized. Another reason is performance.
*/
#define cancel_tmo(tmoref_p) do{\
   OSTMOREF tmoid = *tmoref_p;\
   *tmoref_p = 0;\
   (void)zzkrn_get_sysconf((OSADDRESS)51,\
                           (OSADDRESS)1,\
                           (OSADDRESS)tmoid,\
                           (OSADDRESS)0);\
   } while(0)
#define ose_bind_process(id, cpu) ((OSBOOLEAN)zzkrn_get_sysconf(75, (OSADDRESS)id, cpu, 0))
#define set_sigsize(sig_h,size) ((OSBOOLEAN)zzkrn_get_sysconf(22, (OSADDRESS)sig_h, size, 0))
#ifdef OSE_SMP
#define ose_cpu_id() zzkrn_cpu_id()
#endif


/*
**===========================================================================
** Defines and macros.
**===========================================================================
*/

/*
** This code heavily uses assignments in conditional expressions. This RVCT
** pragma transforms such compiler warnings into less disturbing remarks.
*/
#if defined(__ARMCC_VERSION)
#pragma diag_remark 1293
#endif

/**
 * @macro CLIENT_MAGIC
 * @brief Magic datastructure identfier for offspring structure.
 */
#define CLIENT_MAGIC (0x681121AA)

#define MAX_MEMORY_XFER_CHUNK (0x100)

#define MAX_IDBUF_SIZE    (32)

#define ROUNDUP(x,m)  (((x)+(m)-1)/(m))

/*
** Pcb debug_flags masks.
**
** The 16 low order bits of the pcb flags are defined in pose.h
** The 16 high order bits are available for use by the LDM.
** These are defined here.
*/

/* The process is a system process. Never to be caught. */
#define SYS_PROCESS (0x10000)

/* The process is present in the break_pid_list. */
#define IN_BREAK_SCOPE (0x20000)

/* The process matches the display scope. */
/* Must update this bit when processes are created or the scope is changed. */
#define IN_DEBUGGER_SCOPE (0x40000)

/* This flag is used for prevent Handle_alloc_free_buf_event function
 *  to execute when already in alloc.
 */
#define IN_ALLOC (0x80000)

/* Action flag for including data */
#define TRACE_WITH_DATA 0x40

/*
** The magic number in the precious area (stored in the trace).
*/
#ifdef OSE_SMP
#   define PRECIOUS_MAGIC (0x681121b9)
#else
#   define PRECIOUS_MAGIC (0x681121b8)
#endif

/* Max name length of process names for monitor reports. */
#define MAX_NAME_LEN 64

/*
** Convert a pcb pointer to a block id.
**
** This operation is used when converting a debug_pcb pointer to the
** id of the block containing the process.
**
** The most significant 16 bits of the block id is the instance number.
** The least significant 16 bits is the block/process index.
*/

#define pcb_to_bid(pcb) ((( (uint32_t) pcb->block->instance) << 16) | pcb->block->pindex)

/* The default values and limits for profiling reports for all report types */
#define MAX_PROF_REPORTS         50
#define MIN_PROF_REPORTS         2
#define MAX_PROF_INTERVAL        0x7FFFFFFF
#define MIN_PROF_INTERVAL        1
#define MAX_PROF_PRIORITY        34 /* Limit priority to 34. */
#define MIN_PROF_PRIORITY        0
#define MIN_PROF_MAXPROCS        0
#define MAX_PROF_MAXPROCS        0x7FFFFFFF
#define MAX_PROF_VALUES          500
#define MIN_PROF_VALUES          0

#define DEF_PROF_REPORTS         50
#define DEF_PROF_MAX_VALUES      8
#define DEF_PROF_INTERVAL        100
#define DEF_PROF_PRIORITY        24
#define DEF_PROF_VALUES          10

#define MIN_PROF_OVERSIZE   2
#define MAX_PROF_OVERSIZE   8
#define DEF_PROF_OVERSIZE   2

/* Hard limits for buffer size used during profiling. The soft limit is in the sysparam */
#define MIN_PROF_BUFFER_LIMIT    0
#define MAX_PROF_BUFFER_LIMIT    0x7FFFFFFF /* signed long in system parameter functions */
#define DEF_PROF_BUFFER_LIMIT    MAX_PROF_BUFFER_LIMIT

/*
 *
 * EVACT Types and Defines
 *
 */

#define RMM_EVACT_ACTION_CATCH          ((RMMEvactAction)  1)
#define RMM_EVACT_ACTION_MONITOR        ((RMMEvactAction)  2)
#define RMM_EVACT_ACTION_TRACE          ((RMMEvactAction)  3)
#define RMM_EVACT_ACTION_TRACE_ON       ((RMMEvactAction)  4)
#define RMM_EVACT_ACTION_TRACE_OFF      ((RMMEvactAction)  5)
#define RMM_EVACT_ACTION_WATCH_ON       ((RMMEvactAction)  7)
#define RMM_EVACT_ACTION_WATCH_OFF      ((RMMEvactAction)  8)
#define RMM_EVACT_ACTION_GOTO           ((RMMEvactAction)  9)
#define RMM_EVACT_ACTION_MONITOR_LONG   ((RMMEvactAction) 10)

#define RMM_EVENT_TYPE_CREATE           ((RMMEventType)    1)
#define RMM_EVENT_TYPE_KILL             ((RMMEventType)    2)
#define RMM_EVENT_TYPE_SWAP             ((RMMEventType)    3)
#define RMM_EVENT_TYPE_SEND             ((RMMEventType)    4)
#define RMM_EVENT_TYPE_RECEIVE          ((RMMEventType)    5)
#define RMM_EVENT_TYPE_OSERROR          ((RMMEventType)    6)
#define RMM_EVENT_TYPE_RESTART          ((RMMEventType)   10)
#define RMM_EVENT_TYPE_ERROR            ((RMMEventType)   11)
#define RMM_EVENT_TYPE_UEVENT           ((RMMEventType)   13)

#define RMM_CATCH_SCOPE_RUNNING         ((RMMCatchScopeStatus) 0)
#define RMM_CATCH_SCOPE_HALTED          ((RMMCatchScopeStatus) 1)


/*
 *
 * Process Stepping Types and Defines
 *
 */


#define RMM_SSTEP_HALT_ON_SEND       ((RMMSStepHaltEventMask) 0x01)
#define RMM_SSTEP_HALT_ON_RECEIVE    ((RMMSStepHaltEventMask) 0x02)
#define RMM_SSTEP_HALT_ON_CREATE     ((RMMSStepHaltEventMask) 0x04)
#define RMM_SSTEP_HALT_ON_KILL       ((RMMSStepHaltEventMask) 0x08)
#define RMM_SSTEP_HALT_ON_OS_ERROR   ((RMMSStepHaltEventMask) 0x10)
#define RMM_SSTEP_HALT_ON_USER_ERROR ((RMMSStepHaltEventMask) 0x20)
#define RMM_SSTEP_HALT_ON_RESTART    ((RMMSStepHaltEventMask) 0x40)


/*
 * Number of replies requested, when using HEAP_BUFFER_INFO_REQUEST
 */
#define HEAP_BUFFER_INFO_BURST 10


/*
 * Last action value, needs to be synchronized with enum MonitorActionValues
 * data structure in monitor_events.sig 
 */
#define HIGHEST_ACTION_VALUE 15


/*
 * Max number of hardware trace functions
 */
#define MAX_NO_HW_TRACE_FUNC 2

/*
**===========================================================================
** Types
**===========================================================================
*/

struct idbuf
{
   int        i0;
   int        i1;
   PROCESS         pid[MAX_IDBUF_SIZE];
};

#define CLIENT_NAME_LEN 80

#ifdef USE_EPM
typedef struct
{
   ListLink        link;
   uint32_t        pid;
   void           *regdump;
} ProcessRegdump;
#endif

typedef enum

{
   GTM_TRACE_CLOSED,
   GTM_TRACE_READING
} GetTraceMultipleState;

/**
 * @type Offspring
 * @brief Hold information about a offspring which is the spawned process that
 * handles the communication with the client.
 * @long Part of a linked list of offsprings.
 *
 * @field link list adm header.
 * @field magic contains CLIENT_MAGIC for assertion
 * @field pid is the pid of the offspring
 * @field client_ is the pid of the real client
 * @field attach_ref attach referens as returned by attach
 * @field flow_ctrl_block number of signals we can send to offspring without blocking
 * @field flow_ctrl_nblock number of signals we can send to offspring without throwing
 * @field application_scope the client's current application scope
 * @field execution_point the current execution point (if any)
 *
 */
typedef struct Offspring
{
   ListLink        link;
   int             magic_head;
   char            name[CLIENT_NAME_LEN + 1];
   PROCESS         pid;
   PROCESS         client_;
   OSATTREF        attach_ref;
   int             flow_ctrl_block;
   int             flow_ctrl_nblock;
   MonitorScope    debug_scope; /* Note: Supports only ID scope types */
   MonitorScope    evact_scope;
   OSBOOLEAN       intercept_enabled;
   OSATTREF        scope_attach_ref;
   int             keep_trace_actions;
   int             handles_new_events;
   TraceEntryState cursor;

#ifdef USE_EPM
   List           *execution_points;
   OSBOOLEAN       scope_intercepted;
   volatile uint32_t regdump_free_count;
   List    *process_regdumps;
#endif

   struct idbuf    notify_pids;

   int             mget_tmo;
   union SIGNAL   *mget_req;
   MonitorGetTraceMultipleReply *mget_reply;
   struct idbuf    mtrace_pids;
   int             mget_read;
   uint32_t        mget_loss;
   int             magic_tail;
} Offspring;

/* Offspring must be defined before */
#include "prof.h"
#include "counter_prof.h"
#include "profiling_counters.h"

/*
** Breakpoint scope pid list entry.
**
** The processes in the breakpoints scope are cached in a linked list.
** This is because the breakpoint scope is sometimes traversed by an
** interrupt process.
**
** The list is organized as a linked list of arrays, each capable of
** holding 128 process id's.
** The list is expanded on demand, but it is never garbage collected.
**
** Unused process id position are set to zero.
**
** See also: break_pid_list.
*/

struct break_pid_entry
{
   struct break_pid_entry *next;        /* Next in list or NULL. */
   uint32_t        size;        /* Total no of pid positions below. Always 128. */
   uint32_t        use_count;   /* No of used pid positions below. */
   PROCESS         pid[128];
};

#ifdef USE_EPM

#include "epm.h"

typedef struct BpInfo
{
   uint32_t        bp_id;
   uint32_t        original_count;
   uint32_t        count;
} BpInfo;

typedef struct SsInfo
{
   uint32_t        type;
   uint32_t        range_start;
   uint32_t        range_end;
} SsInfo;

/* FIXME: This is one ugly struct. Should be reorganized to avoid nonsense like:
 * info = (ExecutionPointInfo *)((uint8_t *)execution_point - offsetof(ExecutionPointInfo, ep));
 */
typedef struct ExecutionPointInfo
{
   ListLink        link;
   enum
   { EP_INVALID, EP_VALID } valid;
   uint32_t        instance;
   struct EpmExecutionPoint ep;
   struct Offspring  *offspring;
   MonitorScope    catch_scope;
   enum
   { SINGLE_STEP = 0, BREAKPOINT = 1 } type;
   union
   {
      BpInfo          bp_info;
      SsInfo          ss_info;
   } ep_type_info;
} ExecutionPointInfo;

#   define RMM_CALLBACK_HIT (35951)
typedef struct RmmCallbackHit
{
   SIGSELECT       sig_no;
   uint32_t        ep_instance; /* Used to validate ep */
   MonitorStatus   status;
   ExecutionPointInfo *info;    /* Freed by callee. */
} RmmCallbackHit;


#endif

/* Magic values for actionpoint structure.
 * ACTION_MAGIC_ALLOCATED an allocated structure under way to
 * be initiated. ACTION_MAGIC_VALID a ready actionpoint and the last
 * byte of the magic is a version number. ACTION_MAGIC_INVALID is a
 * deleted actionpoint.
 */
#define ACTION_MAGIC_ALLOCATED 0xde52ed00
#define ACTION_MAGIC_VALID 0xde52ed02
#define ACTION_MAGIC_INVALID 0xde52edff

/*
** This is an internal representation of an action point.
*/

struct ldefine_action
{
   uint32_t        magic; /* See ACTION_MAGIC_* values above */
   uint32_t        align;
   uint32_t        sig_no;
   struct ldefine_action *next;
   struct ldefine_action *prev;
   struct ldefine_action *orig_action;
   uint64_t        data_range_offset;
   uint64_t        data_range_a;
   uint64_t        data_range_b;
   uint32_t        data_range_size;   
   spinlock_t      action_lock;
   uint32_t        signo_end;
   uint32_t        action_id;
   uint16_t        size;
   uint8_t         dummy;
   uint8_t         event_type;
   uint8_t         action;
   uint8_t         negated;
   uint8_t         event_seq_no;
   uint8_t         goto_list;
   uint32_t        signo;
   uint32_t        sig_from;
   uint32_t        sig_to;
   uint32_t        ecode;
   uint32_t        orig_event_count;
   uint32_t        event_count;
   uint32_t        is_basic_event;

   char            names[1];
};


/*
** This defines user event posts
*/
typedef struct
{
   uint32_t         event_no;
   OSADDRESS        trig_value;
   OSADDRESS        *trig;
} user_event_post;

#define EVENTS_PER_BLOCK 20

typedef struct
{
   ListLink         link;
   PROCESS          sid;
   user_event_post user_event_list[EVENTS_PER_BLOCK];
} blocks_user_events;

/*
** This structure holds a scope definition for either the display scope
** or the breakpoint scope.
**
** Scope entries are allocated from the trace buffer space since
** they must survive a system restart.
**
** count:
**        Number of names in the scope_list.
**
** scope_list[]:
**        Variable size.
**        A packed list of zero terminated name strings.
**        Each string consists of a process name wildcard.
*/

struct scope
{
   uint16_t        count;
   char            scope_list[1];
};

/*
** This structure holds all variables that should survive warm starts.
** The structure is kept in a trace buffer entry so that it can be
** accessed from the outside world.
**
** It is important that the format of the structure is not changed
** very often.
*/

struct precious_info
{
   /* A simple checksum, see calc_checkum for the algorithm. */
   int64_t         checksum;

   /*
    ** Magic number for error checking.
    ** This variable holds the magic number PRECIOUS_MAGIC.
    */

   uint32_t        magic;

   /* Count of evacts used to provide unique ids. */
   uint32_t        evact_count;


   /*
    ** Scope pointers.
    **
    ** These variables hold pointers to the current scope definitions.
    ** The pointers must be valid whenever the global debugger enable flag
    ** is set ( See also: zzset_debug_status ).
    **
    ** Scope entries are allocated from the trace buffer space since
    ** they must survive a system restart.
    */

   struct scope   *debugger_scope;
   struct scope   *breakpoint_scope;

   /*
    ** Master action lists.
    **
    ** These are linked lists of all ldefine_action signals currently known to
    ** the LDM. They are stored in hidden trace buffers in order to survive a
    ** system restart.
    **
    ** The command field in the signal is used as a next pointer.
    ** The count field is used to hold the size of the original define_action
    ** signal.
    ** The sig_no field is used to hold the current count value.
    **
    ** The dispatcher must be locked while traversing these lists in the dbg hooks.
    ** Interrupt processes don't need any precautions.
    ** Otherwise we might get lost if an entry is deleted.
    */

   struct ldefine_action *action_list_create;
   struct ldefine_action *action_list_kill;
   struct ldefine_action *action_list_swap;
   struct ldefine_action *action_list_send;
   struct ldefine_action *action_list_receive;
   struct ldefine_action *action_list_oserror;
   struct ldefine_action *action_list_error;
   struct ldefine_action *action_list_restart;
   struct ldefine_action *action_list_uevent;
#ifdef OSE_SMP
   struct ldefine_action *action_list_bind;
#endif
   struct ldefine_action *action_list_alloc;
   struct ldefine_action *action_list_free_buf;
   struct ldefine_action *action_list_timeout;
   void           *not_used_action_list1;

   /*
    ** Monitor master enable flag.
    **
    ** This flag reflects the monitor state set by the client.
    ** Master_monitor_enabled is modified only with set_monitor_status.
    ** It survives restarts and reconnections, and is present to
    ** allow monitoring to be continued whenever a client is connected.
    */

   uint8_t         master_monitor_enabled;
   uint8_t         pad1;
   uint8_t         pad2;
   uint8_t         pad3;


   /*
    ** This array contains all LDM variables that can be set by
    ** a user.
    ** Variables are accessed with the set_variable and
    ** display_variable signals.
    **
    ** The array contains space for future definitions.
    ** These fields are set to zero.
    **
    ** The following variables are currently defined:
    **   options[0] = max_monitor_size -msb. Def=0.
    **   options[1] = max_monitor_size -lsb. Def=255.
    **    This 16 bit variable limits the number of bytes
    **    sent in a monitor send/receive event to the client.
    **
    **   options[2] = max_trace_size -msb. Def=0.
    **   options[3] = max_trace_size -lsb. Def=255.
    **    This 16 bit variable limits the number of bytes
    **    sent in a trace send/receive event to the client.
    **
    **   options[4] = reconnect_timeout -msb. Def=0.
    **   options[5] = reconnect_timeout -lsb. Def=120.
    **    This 16 bit variable defines how many seconds we are prepared to
    **    wait for a client to reconnect after a system restart or a link lost
    **    before we cancel all client activity.
    **    A value of zero means that we wait indefinitely.
    **
    **   options[6] = clear_event_count_on_restart. Def=1.
    **    This 8 bit variable contains a boolean flag. 0 = no, nonzero = yes.
    **    When set, it causes all event counters in the action list to be reset
    **    on each warm start.
    **    When not set, the counters survive warm restarts.
    **
    **   options[7] = display_cpuregs. Def=0.
    **    This 8 bit variable contains a boolean flag. 0 = no, nonzero = yes.
    **    When set, it allows the client to receive cpu register strings in
    **    the long display process command.
    **    When not set, the strings are removed, saving link bandwidth which
    **    is often convenient on RISC cpu's with many registers.
    **
    **   options[8] = Not used
    **
    **   options[9] = clear_sequence_on_restart. Def=1.
    **    This 8 bit variable contains a boolean flag. 0 = no, nonzero = yes.
    **    When set, it causes the global event sequence number to be reset
    **    on each warm start.
    **    When not set, the sequence number survives warm restarts.
    **
    **   options[10] = error_check_level. Def=fetched from pose.
    **    This 8 bit variable contains the current kernel error check level.
    **    This level can be set or examined with the get_sysconf(3, level, 0, 0) call.
    **    Legal values are 0-3, where 0 is the lowest level of error checks.
    **
    **   options[11] = trace_sysprocs. Def=0.
    **    This 8 bit variable contains a boolean flag. 0 = no, nonzero = yes.
    **    When set, it allows the client to trace, catch and monitor events
    **    associated with system processes as well as other processes.
    **    When not set, tracing is disabled for system processes.
    **    This flag allows debugging of platform products such as linkhandlers and
    **    protocol stacks which are normally hidden to users.
    **
    **   options[12] = flow_ctrl_block. Def=20.
    **    Contains the flow control value for blocking request.
    **
    **   options[13] = flow_ctrl_non_block. Def=10.
    **    Contains the flow control value for notify signals.
    **
    **   options[14] = mget_size -msb
    **   options[15] = mget_size -lsb
    **   options[16] = monitor_sys_error, Def=0
    **    When set at 1, it allows catching of system fatal errors. 
    **   options[17] = monitor_error           Def: 1.
    **   options[18] = Not used.
    **   options[19] = profile_min_sample. Def 1.
    **   options[20] = dynamic_trace_size -msb. Def=0.
    **   options[21] = dynamic_trace_size
    **   options[22] = dynamic_trace_size
    **   options[23] = dynamic_trace_size -lsb.
    **
    **   options[24] = default_proc_interval[0] -msb. Def = DEF_PROF_INTERVAL
    **   options[25] = default_proc_interval[0]
    **   options[26] = default_proc_interval[0]
    **   options[27] = default_proc_interval[0] -lsb.
    **   options[28] = default_proc_reports[0]  -msb. Def = DEF_PROF_REPORTS
    **   options[29] = default_proc_reports[0]
    **   options[30] = default_proc_reports[0]
    **   options[31] = default_proc_reports[0]  -lsb.
    **   options[32] = default_proc_maxprocs[0] -msb. Def = DEF_PROF_MAX_VALUES
    **   options[33] = default_proc_maxprocs[0]
    **   options[34] = default_proc_maxprocs[0]
    **   options[35] = default_proc_maxprocs[0] -lsb.
    **   options[36] = default_proc_interval[1] -msb. Def = DEF_PROF_INTERVAL
    **   options[37] = default_proc_interval[1]
    **   options[38] = default_proc_interval[1]
    **   options[39] = default_proc_interval[1] -lsb.
    **   options[40] = default_proc_reports[1]  -msb. Def = DEF_PROF_REPORTS
    **   options[41] = default_proc_reports[1]
    **   options[42] = default_proc_reports[1]
    **   options[43] = default_proc_reports[1]  -lsb.
    **   options[44] = default_proc_maxprocs[1] -msb. Def = DEF_PROF_MAX_VALUES
    **   options[45] = default_proc_maxprocs[1]
    **   options[46] = default_proc_maxprocs[1]
    **   options[47] = default_proc_maxprocs[1] -lsb.
    **   options[48] = default_pri_interval     -msb. Def = DEF_PROF_INTERVAL
    **   options[49] = default_pri_interval
    **   options[50] = default_pri_interval
    **   options[51] = default_pri_interval     -lsb.
    **   options[52] = default_pri_reports      -msb. Def = DEF_PROF_REPORTS
    **   options[53] = default_pri_reports
    **   options[54] = default_pri_reports
    **   options[55] = default_pri_reports      -lsb.
    **   options[56] = default_cpu_interval     -msb. Def = DEF_PROF_INTERVAL
    **   options[57] = default_cpu_interval
    **   options[58] = default_cpu_interval
    **   options[59] = default_cpu_interval     -lsb.
    **   options[60] = default_cpu_reports.     -msb. Def = DEF_PROF_REPORTS
    **   options[61] = default_cpu_reports
    **   options[62] = default_cpu_reports
    **   options[63] = default_cpu_reports.     -lsb.
    **   options[64] = default_cpu_priority     -msb. Def = DEF_PROF_PRIORITY
    **   options[65] = default_cpu_priority
    **   options[66] = default_cpu_priority
    **   options[67] = default_cpu_priority     -lsb.
    **   options[68] = default_user_interval    -msb. Def = DEF_PROF_INTERVAL
    **   options[69] = default_user_interval
    **   options[70] = default_user_interval
    **   options[71] = default_user_interval    -lsb.
    **   options[72] = default_user_reports     -msb. Def = DEF_PROF_REPORTS
    **   options[73] = default_user_reports
    **   options[74] = default_user_reports
    **   options[75] = default_user_reports     -lsb.
    **   options[76] = default_user_maxvalues   -msb. Def = DEF_PROF_MAX_VALUES
    **   options[77] = default_user_maxvalues
    **   options[78] = default_user_maxvalues
    **   options[79] = default_user_maxvalues   -lsb.
    **   options[80] = swap_interrupt_process
    **
    **  All the options in the chunk above handle monitor CPU load and
    **  user report profiling parameter settings.
    */

    /*
    ** The rest of the option fields are undefined and should be set to 0.
    */

   uint8_t         options[96]; /* Size must correspond with init loop. */

   /*
    ** These two bytes are for future use. For now, they provide alignment
    ** padding, and contain a value of zero.
    */

   uint8_t         pad4;
   uint8_t         pad5;

   /* The fields below are *not* included in the checksum due to frequence
    * of updates in event dbghooks.
    * */

   /*
    ** Trace master enable flag.
    **
    ** This flag reflects the trace state set by a client.
    ** The trace itself is set to reflect this flag at all times except
    ** during a trace readout, when the trace is temporarily disabled.
    */

   uint8_t         trace_enabled;

   /*    This 8 bit variable contains a boolean flag. 0 = run, nonzero = halt.
    **    Current intercept status for the breakpoint scope.
    **    When set, the breakpoint scope is intercepted.
    **    When not set, the breakpoint scope is running.
    **    Default is 0.
    */
   uint8_t         break_scope_status;

   /* The state of the action system and set by the goto action.
    * Was previous known as sequence. */
   uint32_t        action_state;

   /* A number that is *never* initialized and used to obtain a unique
    * id that might survive a warm restart. */
   atomic_t        unique_number;

   uint32_t        magic_tail;
};

/*
**===========================================================================
** The signal union.
**===========================================================================
*/

union SIGNAL
{
   SIGSELECT       sig_no;

   /* Signals to/from system daemon. */
   struct start_modeflag start_modeflag;

   MonitorInterfaceReply interface_reply;
   MonitorNameReply name_reply;
   MonitorConnectRequest connect_request;
   MonitorConnectReply connect_reply;
   MonitorFlowControl flow_control;
   MonitorGetInfoRequest get_info_request;
   MonitorGetProcessInfoReply get_process_info_reply;
   MonitorGetProcessInfoLongReply get_process_info_long_reply;
   MonitorGetInfoEndmark get_info_endmark;
   MonitorGetBlockInfoReply get_block_info_reply;
   MonitorGetEnvVarsReply get_env_vars_reply;
   MonitorGetStackUsageReply get_stack_usage_reply;
   MonitorGetSignalQueueReply get_signal_queue_reply;
   MonitorGetSignalQueueLongReply get_signal_queue_long_reply;
   MonitorGetPoolInfoReply get_pool_info_reply;
   MonitorGetPoolSignalsReply get_pool_signals_reply;
   MonitorGetSegmentInfoReply get_segment_info_reply;
   MonitorActionRequest action_request;
   MonitorActionReply action_reply;
   MonitorCreateProcessRequest create_process_request;
   MonitorCreateProcessReply create_process_reply;
   MonitorSetEnvVarRequest set_env_var_request;
   MonitorSetSysParamRequest set_sys_param_request;
   MonitorSendSignalRequest send_signal_request;
   MonitorSetPriorityRequest set_priority_request;
   MonitorSignalSemaphoreRequest signal_semaphore_request;
   MonitorGetSysParamReply get_sys_param_reply;

   /* Evact signals */
   MonitorSetNotifyEnabledRequest set_notify_enabled_request;
   MonitorSetNotifyEnabledReply set_notify_enabled_reply;
   MonitorSetTraceEnabledRequest set_trace_enabled_request;
   MonitorSetTraceEnabledReply set_trace_enabled_reply;
   MonitorSetInterceptEnabledRequest set_intercept_enabled_request;
   MonitorSetInterceptEnabledReply set_intercept_enabled_reply;
   MonitorGetNotifyEnabledRequest get_notify_enabled_request;
   MonitorGetNotifyEnabledReply get_notify_enabled_reply;
   MonitorGetTraceEnabledRequest get_trace_enabled_request;
   MonitorGetTraceEnabledReply get_trace_enabled_reply;
   MonitorGetInterceptEnabledRequest get_intercept_enabled_request;
   MonitorGetInterceptEnabledReply get_intercept_enabled_reply;
   MonitorSetEventStateRequest set_event_state_request;
   MonitorSetEventStateReply set_event_state_reply;
   MonitorGetEventStateRequest get_event_state_request;
   MonitorGetEventStateReply get_event_state_reply;
   MonitorEventActionpointInfo event_actionpoint_info;
   MonitorSetEventActionpointReply set_event_actionpoint_reply;
   MonitorClearEventActionpointRequest clear_event_actionpoint_request;
   MonitorClearEventActionpointReply clear_event_actionpoint_reply;
   MonitorGetEventActionpointsRequest get_event_actionpoints_request;
   MonitorGetEventActionpointsReply get_event_actionpoints_reply;
   MonitorGetEventActionpointsEndmark get_event_actionpoints_endmark;
   MonitorCreateEventStructRequest create_event_struct_request;

   MonitorSetScopeRequest set_scope_request;
   MonitorSetScopeReply set_scope_reply;
   MonitorProcessInfo process_info;
   MonitorSendReceiveInfo send_receive_info;
   MonitorSwapCreateKillInfo swap_create_kill_info;
   MonitorErrorInfo error_info;
   MonitorResetInfo reset_info;
   MonitorUserInfo user_info;
#ifdef OSE_SMP
   MonitorBindInfo bind_info;
#endif
   MonitorAllocFreeBufInfo alloc_free_buf_info;
#ifdef HAS_TIMEOUT_EVENT
   MonitorTimeoutInfo timeout_info;
#endif
   MonitorLossNotify loss_notify;

   /* Trace signals */
   MonitorGetTraceRequest get_trace_request;
   MonitorGetTraceMultipleRequest get_trace_mult_request;
   MonitorGetTraceEndmark get_trace_endmark;
   MonitorClearTraceRequest clear_trace_request;
   MonitorClearTraceReply clear_trace_reply;

   /* Debugging signals */
   MonitorAttachRequest attach_request;
   MonitorAttachReply attach_reply;
   MonitorDetachRequest detach_request;
   MonitorDetachReply detach_reply;

   MonitorInterceptScopeRequest intercept_scope_request;
   MonitorInterceptScopeReply intercept_scope_reply;
   MonitorRegisterIds register_ids;
   MonitorRegisterValues register_values;
   MonitorGetMemoryRequest get_memory_request;
   MonitorGetMemoryReply get_memory_reply;
   MonitorGetMemoryEndmark get_memory_endmark;
   MonitorSetMemoryRequest set_memory_request;
   MonitorSetMemoryReply set_memory_reply;
   MonitorStepProcessRequest step_process_request;
   MonitorStepProcessNotify step_process_notify;
   MonitorSetBreakpointRequest set_breakpoint_request;
   MonitorSetBreakpointReply set_breakpoint_reply;
   MonitorResumeScopeRequest resume_scope_request;
   MonitorResumeScopeReply resume_scope_reply;
   MonitorBreakpointFiredNotify breakpoint_fired_notify;
   MonitorClearBreakpointRequest clear_breakpoint_request;
   MonitorClearBreakpointReply clear_breakpoint_reply;
   MonitorBasicEventActionRequest basic_event_action_request;
   MonitorBasicEventActionReply basic_event_action_reply;
   MonitorBasicErrorNotify basic_error_notify;
   MonitorBasicCreateKillNotify basic_create_kill_notify;
   MonitorBasicLossNotify basic_loss_notify;

   /* Profiling signals. */
   MonitorSetCPUReportsEnabledRequest set_cpu_reports_enabled_request;
   MonitorSetCPUReportsEnabledReply set_cpu_reports_enabled_reply;
   MonitorGetCPUReportsEnabledRequest get_cpu_reports_enabled_request;
   MonitorGetCPUReportsEnabledReply get_cpu_reports_enabled_reply;
   MonitorGetCPUPriReportsEnabledRequest get_cpu_pri_reports_enabled_request;
   MonitorGetCPUPriReportsEnabledReply get_cpu_pri_reports_enabled_reply;
   MonitorSetCPUPriReportsEnabledRequest set_cpu_pri_reports_enabled_request;
   MonitorSetCPUPriReportsEnabledReply set_cpu_pri_reports_enabled_reply;
   MonitorGetCPUProcessReportsEnabledRequest get_cpu_process_reports_enabled_request;
   MonitorGetCPUProcessReportsEnabledReply get_cpu_process_reports_enabled_reply;
   MonitorSetCPUProcessReportsEnabledRequest set_cpu_process_reports_enabled_request;
   MonitorSetCPUProcessReportsEnabledReply set_cpu_process_reports_enabled_reply;
   MonitorGetCPUReportsRequest get_cpu_reports_request;
   MonitorGetCPUReportsReply get_cpu_reports_reply;
   MonitorGetCPUPriReportsRequest get_cpu_pri_reports_request;
   MonitorGetCPUPriReportsReply get_cpu_pri_reports_reply;
   MonitorGetCPUProcessReportsRequest get_cpu_process_reports_request;
   MonitorGetCPUProcessReportsReply get_cpu_process_reports_reply;
   MonitorCPUProcessReportpointInfo cpu_process_reportpoint_info;
   MonitorSetCPUProcessReportpointReply set_cpu_process_reportpoint_reply;
   MonitorClearCPUProcessReportpointRequest clear_cpu_process_reportpoint_request;
   MonitorClearCPUProcessReportpointReply clear_cpu_process_reportpoint_reply;
   MonitorGetCPUProcessReportpointInfoRequest get_cpu_process_reportpoint_info_request;
   MonitorGetCPUProcessReportpointInfoEndmark get_cpu_process_reportpoint_info_endmark;

   MonitorSetUserReportsEnabledRequest set_user_reports_enabled_request;
   MonitorSetUserReportsEnabledReply set_user_reports_enabled_reply;

   MonitorGetUserReportsEnabledRequest get_user_reports_enabled_request;
   MonitorGetUserReportsEnabledReply get_user_reports_enabled_reply;


   MonitorGetUserReportsRequest get_user_reports_request;
   MonitorGetUserReportsReply get_user_reports_reply;

   MonitorCreateUserReportsStructRequest create_user_reports_struct_request;
   MonitorCreateUserReportsStructReply create_user_reports_struct_reply;

   UserReportsAttachSig user_report_attach;

   MonitorGetHeapInfoReply get_heap_info_reply;
   MonitorGetHeapFragmentInfoReply get_heap_frag_info_reply;
   MonitorGetHeapBuffersReply get_heap_buf_reply;

   MonitorGetProfilingExecutionUnitEnabledRequest get_prof_eu_enabled_request;
   MonitorGetProfilingExecutionUnitEnabledReply get_prof_eu_enabled_reply;
   MonitorSetProfilingExecutionUnitEnabledRequest set_prof_eu_enabled_request;
   MonitorSetProfilingExecutionUnitEnabledReply set_prof_eu_enabled_reply;

   /* Counter profiling signals. */
   MonitorGetCounterTypeEnabledRequest get_counter_type_enabled_request;
   MonitorGetCounterTypeEnabledReply get_counter_type_enabled_reply;
   MonitorSetCounterTypeEnabledRequest set_counter_type_enabled_request;
   MonitorSetCounterTypeEnabledReply set_counter_type_enabled_reply;
   MonitorGetCounterReportsNotifyEnabledRequest get_counter_reports_notify_enabled_request;
   MonitorGetCounterReportsNotifyEnabledReply get_counter_reports_notify_enabled_reply;
   MonitorSetCounterReportsNotifyEnabledRequest set_counter_reports_notify_enabled_request;
   MonitorSetCounterReportsNotifyEnabledReply set_counter_reports_notify_enabled_reply;
   MonitorCounterReportsNotify counter_reports_notify;
   MonitorCounterReportsLossNotify counter_reports_loss_notify;

   /* Dump signals. */
#ifdef USE_RAMDUMP
   MonitorGetRamdumpInfoRequest get_ramdump_info_request;
   MonitorGetRamdumpInfoReply get_ramdump_info_reply;
   MonitorGetRamdumpInfoEndmark get_ramdump_info_endmark;
   MonitorGetRamdumpRequest get_ramdump_request;
   MonitorGetRamdumpReply get_ramdump_reply;
   MonitorGetRamdumpEndmark get_ramdump_endmark;
#endif

   /* Internal signals. */
   UserEventAttachSig user_event_attach_sig;
   RmmOffspringInfo rmm_offspring_info;
   RmmGetOffspringsReply get_offsprings_reply;
#ifdef USE_EPM
   RmmCallbackHit  rmm_callback_hit;
   RmmScopeAttachSig rmm_scope_attach_sig;
#endif


   ProfNewSample   prof_new_sample;

#ifdef OSE_SMP
   ProfCoreFreeBuf prof_core_free_buf;
#endif

   /* LDM event report signal, used by hooks */
   struct event_report_verbose event_report_verbose;

   struct mm_list_regions_reply mm_list_regions_reply;
   struct mm_list_regions_request mm_list_regions_request;
   struct mm_create_region mm_create_region;

   /* heap.sig interface, used by the Heap Browser */
   struct HeapListReply heap_list_rep;
   struct HeapStatsExReply heap_stats_rep;
   struct HeapConfigReply heap_config_rep;
   struct HeapFragsReply heap_frags_rep;
   struct HeapFragInfo heap_frag_info_rep;
   struct HeapBufferInfoReply heap_buf_info_rep;

   /* Internal signal used by Counter profiling. */
   struct BufferFullNotify buffer_full_notify;
};


static void
zzalloc_dbghook(OSBUFSIZE sig_size, void *sig, struct debug_pcb *from,
                 PROCESS poolid);

static void
zzfree_buf_dbghook(OSBUFSIZE sig_size, void *sig, struct debug_pcb *from,
                    PROCESS poolid);
/*
**===========================================================================
** Signal filters.
**===========================================================================
*/

static const SIGSELECT all_sigs_filter[] = { 0 };
static const SIGSELECT no_nblock_sigs_filter[] = { (SIGSELECT)-1,
						   MONITOR_REPORT };

static const SIGSELECT request_sigs[] =
{
   MONITOR_INTERFACE_REQUEST,
   MONITOR_NAME_REQUEST,
   MONITOR_CONNECT_REQUEST,
   MONITOR_DISCONNECT_REQUEST,
   MONITOR_GET_PROCESS_INFO_REQUEST,
   MONITOR_GET_BLOCK_INFO_REQUEST,
   MONITOR_GET_SEGMENT_INFO_REQUEST,
   MONITOR_GET_STACK_USAGE_REQUEST,
   MONITOR_GET_ENV_VARS_REQUEST,
   MONITOR_GET_SIGNAL_QUEUE_REQUEST,
   MONITOR_GET_POOL_INFO_REQUEST,
   MONITOR_GET_POOL_SIGNALS_REQUEST,
   MONITOR_GET_SYS_PARAM_REQUEST,
   MONITOR_CREATE_PROCESS_REQUEST,
   MONITOR_KILL_SCOPE_REQUEST,
   MONITOR_START_SCOPE_REQUEST,
   MONITOR_STOP_SCOPE_REQUEST,
   MONITOR_SET_ENV_VAR_REQUEST,
   MONITOR_SIGNAL_SEMAPHORE_REQUEST,
   MONITOR_SET_PRIORITY_REQUEST,
   MONITOR_SET_SYS_PARAM_REQUEST,
   MONITOR_SEND_SIGNAL_REQUEST,
   MONITOR_SET_EXECUTION_UNIT_REQUEST,
   MONITOR_GET_CPU_REPORTS_ENABLED_REQUEST,
   MONITOR_SET_CPU_REPORTS_ENABLED_REQUEST,
   MONITOR_GET_CPU_PRI_REPORTS_ENABLED_REQUEST,
   MONITOR_SET_CPU_PRI_REPORTS_ENABLED_REQUEST,
   MONITOR_GET_CPU_PROCESS_REPORTS_ENABLED_REQUEST,
   MONITOR_SET_CPU_PROCESS_REPORTS_ENABLED_REQUEST,
   MONITOR_GET_CPU_BLOCK_REPORTS_ENABLED_REQUEST,
   MONITOR_SET_CPU_BLOCK_REPORTS_ENABLED_REQUEST,
   MONITOR_GET_CPU_REPORTS_REQUEST,
   MONITOR_GET_CPU_PRI_REPORTS_REQUEST,
   MONITOR_GET_CPU_PROCESS_REPORTS_REQUEST,
   MONITOR_GET_CPU_BLOCK_REPORTS_REQUEST,
   MONITOR_SET_CPU_PROCESS_REPORTPOINT_REQUEST,
   MONITOR_CLEAR_CPU_PROCESS_REPORTPOINT_REQUEST,
   MONITOR_GET_CPU_PROCESS_REPORTPOINT_INFO_REQUEST,

   MONITOR_GET_USER_REPORTS_REQUEST,
   MONITOR_GET_USER_REPORTS_ENABLED_REQUEST,
   MONITOR_SET_USER_REPORTS_ENABLED_REQUEST,

   MONITOR_SET_NOTIFY_ENABLED_REQUEST,
   MONITOR_GET_NOTIFY_ENABLED_REQUEST,
   MONITOR_SET_INTERCEPT_ENABLED_REQUEST,
   MONITOR_GET_INTERCEPT_ENABLED_REQUEST,
   MONITOR_SET_SEND_ACTIONPOINT_REQUEST,
   MONITOR_SET_RECEIVE_ACTIONPOINT_REQUEST,
   MONITOR_SET_SWAP_ACTIONPOINT_REQUEST,
   MONITOR_SET_CREATE_ACTIONPOINT_REQUEST,
   MONITOR_SET_KILL_ACTIONPOINT_REQUEST,
   MONITOR_SET_ERROR_ACTIONPOINT_REQUEST,
   MONITOR_SET_USER_ACTIONPOINT_REQUEST,
#ifdef OSE_SMP
   MONITOR_SET_BIND_ACTIONPOINT_REQUEST,
#endif
   MONITOR_SET_ALLOC_ACTIONPOINT_REQUEST,
   MONITOR_SET_FREE_BUF_ACTIONPOINT_REQUEST,
#ifdef HAS_TIMEOUT_EVENT
   MONITOR_SET_TIMEOUT_ACTIONPOINT_REQUEST,
#endif
   MONITOR_CLEAR_EVENT_ACTIONPOINT_REQUEST,
   MONITOR_SET_INTERCEPT_SCOPE_REQUEST,
   MONITOR_SET_SCOPE_REQUEST,
   MONITOR_SET_TRACE_ENABLED_REQUEST,
   MONITOR_GET_TRACE_ENABLED_REQUEST,
   MONITOR_CLEAR_TRACE_REQUEST,
   MONITOR_GET_TRACE_REQUEST,
   MONITOR_GET_TRACE_MULTIPLE_REQUEST,

   MONITOR_ATTACH_REQUEST,
   MONITOR_DETACH_REQUEST,
   MONITOR_INTERCEPT_SCOPE_REQUEST,

#ifdef USE_EPM
   MONITOR_GET_REGISTER_INFO_REQUEST,
   MONITOR_GET_REGISTERS_REQUEST,
   MONITOR_SET_REGISTERS_REQUEST,
   MONITOR_STEP_PROCESS_REQUEST,
   MONITOR_SET_BREAKPOINT_REQUEST,
   MONITOR_CLEAR_BREAKPOINT_REQUEST,
#endif

   MONITOR_RESUME_SCOPE_REQUEST,
   MONITOR_GET_MEMORY_REQUEST,
   MONITOR_SET_MEMORY_REQUEST,
   MONITOR_BASIC_ERROR_ACTION_REQUEST,
   MONITOR_BASIC_CREATE_KILL_ACTION_REQUEST,
   MONITOR_SET_EVENT_STATE_REQUEST,
   MONITOR_GET_EVENT_STATE_REQUEST,
   MONITOR_GET_EVENT_ACTIONPOINT_INFO_REQUEST,
   MONITOR_GET_EVENT_ACTIONPOINTS_REQUEST,

#ifdef USE_RAMDUMP
   MONITOR_GET_RAMDUMP_INFO_REQUEST,
   MONITOR_GET_RAMDUMP_REQUEST,
#endif

   MONITOR_GET_TOTAL_CPU_BLOCK_REPORTS_ENABLED_REQUEST,
   MONITOR_SET_TOTAL_CPU_BLOCK_REPORTS_ENABLED_REQUEST,
   MONITOR_GET_TOTAL_CPU_BLOCK_REPORTS_REQUEST,

   MONITOR_GET_HEAP_INFO_REQUEST,
   MONITOR_GET_HEAP_FRAGMENT_INFO_REQUEST,
   MONITOR_GET_HEAP_BUFFERS_REQUEST,
#ifdef OSE_SMP
   MONITOR_GET_PROFILING_EXECUTION_UNIT_ENABLED_REQUEST,
   MONITOR_SET_PROFILING_EXECUTION_UNIT_ENABLED_REQUEST,
#endif
   /* Counter profiling request signals. */
   MONITOR_GET_SUPPORTED_COUNTER_TYPES_REQUEST,
   MONITOR_GET_COUNTER_TYPE_ENABLED_REQUEST,
   MONITOR_SET_COUNTER_TYPE_ENABLED_REQUEST,
   MONITOR_GET_COUNTER_REPORTS_NOTIFY_ENABLED_REQUEST,
   MONITOR_SET_COUNTER_REPORTS_NOTIFY_ENABLED_REQUEST
};

static const SIGSELECT notify_sigs[] =
{
   MONITOR_BASIC_CREATE_KILL_NOTIFY,
   MONITOR_BASIC_ERROR_NOTIFY,
   MONITOR_BASIC_LOSS_NOTIFY,
   MONITOR_LOSS_NOTIFY,
   MONITOR_PROCESS_NOTIFY,
   MONITOR_SEND_NOTIFY,
   MONITOR_RECEIVE_NOTIFY,
   MONITOR_CREATE_NOTIFY,
   MONITOR_KILL_NOTIFY,
   MONITOR_ERROR_NOTIFY,
   MONITOR_USER_NOTIFY,
#ifdef OSE_SMP
   MONITOR_BIND_NOTIFY,
#endif
   MONITOR_ALLOC_NOTIFY,
#ifdef HAS_TIMEOUT_EVENT
   MONITOR_TIMEOUT_NOTIFY,
#endif
   MONITOR_FREE_BUF_NOTIFY
};

static const SIGSELECT intercept_sigs[] =
{
   MONITOR_STEP_PROCESS_NOTIFY,
   MONITOR_BREAKPOINT_FIRED_NOTIFY,
   MONITOR_PROCESS_INTERCEPT_NOTIFY,
   MONITOR_SEND_INTERCEPT_NOTIFY,
   MONITOR_RECEIVE_INTERCEPT_NOTIFY,
   MONITOR_CREATE_INTERCEPT_NOTIFY,
   MONITOR_KILL_INTERCEPT_NOTIFY,
   MONITOR_ERROR_INTERCEPT_NOTIFY,
   MONITOR_USER_INTERCEPT_NOTIFY,
#ifdef OSE_SMP
   MONITOR_BIND_INTERCEPT_NOTIFY,
#endif
   MONITOR_ALLOC_INTERCEPT_NOTIFY,
#ifdef HAS_TIMEOUT_EVENT
   MONITOR_TIMEOUT_INTERCEPT_NOTIFY,
#endif
   MONITOR_FREE_BUF_INTERCEPT_NOTIFY
};

static const SIGSELECT trace_sigs[] =
{
   MONITOR_GET_PROCESS_TRACE_REPLY,
   MONITOR_GET_SEND_TRACE_REPLY,
   MONITOR_GET_RECEIVE_TRACE_REPLY,
   MONITOR_GET_CREATE_TRACE_REPLY,
   MONITOR_GET_KILL_TRACE_REPLY,
   MONITOR_GET_SWAP_TRACE_REPLY,
   MONITOR_GET_ERROR_TRACE_REPLY,
   MONITOR_GET_RESET_TRACE_REPLY,
   MONITOR_GET_USER_TRACE_REPLY,
#ifdef OSE_SMP
   MONITOR_GET_BIND_TRACE_REPLY,
#endif
   MONITOR_GET_ALLOC_TRACE_REPLY,
   MONITOR_GET_FREE_BUF_TRACE_REPLY,
#ifdef HAS_TIMEOUT_EVENT
   MONITOR_GET_TIMEOUT_TRACE_REPLY,
#endif
   MONITOR_GET_TRACE_ENDMARK
};

/*
**===========================================================================
** Variables.
**===========================================================================
*/

#ifdef USE_EPM

/* Global variable for turning on full register saves. */
extern int      zzdebug_regs_on;

/* If zzdebug_regs_on was already set outside the rmm,
   we do not want to clear it. Therefore we need to keep
   track of whether we modified it in the rmm or not. */
static int      modified_zzdebug_regs_on = 0;

#endif


void           *const zzldmvarp[] = {
   (void *) 0,                  /* Zero means table is not supported. */
   (void *) 0,
   (void *) 0
};


typedef struct ReportIdHashItem
{
   OSREPORTID id;       /* The report id that the index in this struct points at */
   uint32_t index;      /* The index in the sample process table*/
   uint32_t sample;     /* Flag indicating the current sample */
} ReportIdHashItem;

#define PROF_USER_REPORT_TRIG 5

#define PROF_USER_REPORT_TRIG_MAGIC 87654321

/* application triggers connected to one report */
typedef struct appl_trig_list
{
   PROCESS sid;                    /* Segment Id */
   OSADDRESS *trig;                /* Pointer to calling application memory */
} ApplTriggers;

typedef struct prof_trig_list
{
   uint32_t magic;
   ApplTriggers app_trig[PROF_USER_REPORT_TRIG];   /* List of trig addresses and block id's */
   struct prof_trig_list *next;              /* Pointer to additional trig addresses if 7 is not enough */
} ProfUserTrigList;


typedef struct prof_user_report_value
{
   OSREPORTID id;
   OSREPORTVAL max;
   OSREPORTVAL min;
   OSREPORTVAL value;
} UserReportValue;

#define PROF_USER_REPORT_SAMPLE_MAGIC 98765432

/* The structure of each sample in the history list. */
typedef struct prof_user_report_sample
{
   uint32_t          magic;
   struct prof_user_report_sample *next;          /* Link in circular list. */
   uint32_t          time_ticks;                  /* Start of sample interval, ticks. */
   uint32_t          time_uticks;                 /* Start of sample interval, microseconds. */
   uint32_t          total_time;                  /* Accumulated length of the interval. */
   uint32_t          valid;                       /* Zero at the end of the list of valid samples. */
   UserReportValue   *report;                     /* Pointer to report data */
   uint32_t          report_count;                /* Number of reports in this sample*/
} ProfUserReportSample;

#define PROF_USER_REPORT_MAGIC 23456789

/* The structure of each created report no */
typedef struct prof_user_report
{
   uint32_t magic;
   struct prof_user_report *next;                     /* Link in circular list. */
   SIGSELECT reportno;                                /* reportno used as identifier for the report class */
   ProfUserTrigList trig_list;                        /* Application trig addresses and block id's*/
   OSBOOLEAN multiple;                                /* if set to one then store multiple values e.g one per process  */
   OSBOOLEAN continuous;                              /* if set to one then copy previous stored value to the next interval */
   OSBOOLEAN maxmin;                                  /* set to one to track the maximum and minimum value during each interval */
#ifdef OSARM
   OSBOOLEAN enabled;                                 /* if set to one, the user
                                                       * report profiling is enabled
                                                       * immediately after receiving
                                                       * create user report request
                                                       */
#endif
   uint32_t interval;                                 /* time interval */
   uint32_t icounter;                                 /* interval counter */
   ProfUserReportSample *prof_user_rpt_samplepool;    /* One sample per interval */
   ProfUserReportSample *prof_user_rpt_firstsample;   /* Points to the first sample in interval */
   ProfUserReportSample *prof_user_rpt_cursample;     /* Points to the current sample in interval */
   uint32_t maxReports;                               /* Max number of reports kept in target */
   uint32_t maxValues;                                /* Max number of values per report */
   uint32_t prof_report_disctimer;                    /* Disconnect timer */
   ReportIdHashItem *report_id_hash_table;            /* hash table */
   ReportIdHashItem *overflow_hash_table;             /* overflow linear table */
   OSREPORTID hash_mask;                              /* hash mask */
   uint32_t hash_size;                                /* size of hash table */
   uint32_t sample_counter;                           /* Counter used to mark the hash table as invalid */
   uint32_t overflow_list_pos_cnt;                    /* Number of report id in overflow list */
} ProfUserReport;

typedef struct UserReportEnabledHashItem
{
   ProfUserReport *enabledUserReport;   /* The index in the report header table*/
} UserReportEnabledHashItem;

/*
 * This struct contains pointers to all action lists (one action list for each
 * event type). The RMM maintain one set of action lists for each core in a SMP
 * system and uses an array of this struct for that in rmm_world.
 *
 * See also smp_distribute_actionpoint() and smp_delete_actionpoint().
 */
struct action_lists
{
   struct ldefine_action *action_list_create;
   struct ldefine_action *action_list_kill;
   struct ldefine_action *action_list_swap;
   struct ldefine_action *action_list_send;
   struct ldefine_action *action_list_receive;
   struct ldefine_action *action_list_oserror;
   struct ldefine_action *action_list_error;
   struct ldefine_action *action_list_restart;
   struct ldefine_action *action_list_uevent;
#ifdef OSE_SMP
   struct ldefine_action *action_list_bind;
#endif
   struct ldefine_action *action_list_alloc;
   struct ldefine_action *action_list_free_buf;
#ifdef HAS_TIMEOUT_EVENT
   struct ldefine_action *action_list_timeout;
#endif
};


/* This struct is used for checking that parameters are within a specified range */
typedef struct parameter_range
{
   uint32_t max;
   uint32_t min;
} parameter_range;


static struct rmm_world
{
   /* Blocking flow control value. */
   int32_t  flow_ctrl_block; /* = 20 @INIT@ */
   /* Non blocking (notify signals) flow control value. */
   int32_t  flow_ctrl_nblock; /* = 10 @INIT@ */

   volatile uint32_t abs_sec;    /* Seconds since 0:0:0 1 jan-1970 GMT. */
   volatile uint32_t abs_tick;   /* The corresponding system tick value. */
   volatile uint32_t abs_usec;   /* The corresponding microseconds subtick value. */
   volatile uint32_t time_update_count; /* Incremented by 2 in two steps when
                                           new time is set. */
   struct abstime *trace_abstime;

   /*
   ** List of blocks user event posts.
   ** User event posts are inserted by the create_event system call to
   ** keep track of block, event number, and trig variable.
   */
   List *user_events;

   /*
   ** This structure holds all variables that should survive warm starts.
   ** The structure is kept in a trace buffer entry so that it can be
   ** accessed from the outside world.
   **
   ** It is important that the format of the structure is not changed
   ** very often.
   **
   ** It is also important to know that this trace entry is the first
   ** one allocated after the trace was created with ose_trace_create().
   */

   struct precious_info *precious;

   /*
   ** Base and size of the trace buffer.
   **
   ** This pointer is passed from the target interface in the
   ** system control block.
   */

   char    *trace;
   uint32_t trace_size;

   /*
   ** Static trace is used whenever dynamic tracing is enabled.
   **
   ** Points to the address of the trace buffer created att RMM startup.
   ** If dynamic tracing is used the trace variable will point to the
   ** dynamic memory. But some functions will still need to use the
   ** static memory. I.e Action points, and proc_info.
   */

   char    *static_trace;
   uint32_t static_trace_size;

   /* Pointer to the OSE block control block, used to check fast if a process
    * is part of the OSE block. */
   struct _block *ose_block_bcb;

   /*
   ** Trace master enable flag.
   **
   ** This flag reflects the trace state set by the client.
   ** The trace itself is set to reflect this flag at all times except
   ** during a trace readout, when the trace is temporarily disabled.
   ** This variable survives warm restarts since it is also in the precious struct.
   */

   OSBOOLEAN trace_enabled;

   OSBOOLEAN master_monitor_enabled;

   /*
   ** Enables interception of system fatal errors.
   ** This is only experimental and should not be used
   */

   uint8_t monitor_sys_error;

   /*
   ** Enables/Disables traped user handled.
   */

   uint8_t monitor_error;

   /*
   ** Current intercept status for the breakpoint scope.
   ** Do we need the "2"?? Why not simply assume a "*" breakpoint scope
   ** during cold start?
   **
   ** Contains the following values:
   **      0 = The scope is running.
   **      1 = The scope is stopped.
   **      2 = All processes except system processes were stopped at
   **          system cold start. The scope definition is unimportant.
   */

   OSBOOLEAN break_scope_status;

   /**
    * @variable
    * @brief Linked list of RMM offsprings.
    */
   List    *offsprings;

   Offspring  *evact_lock_owner;

#ifdef USE_EPM
   List    *ep_free_list;
#endif

   /**
    * @variable
    * @brief Contains the pid of the RMM process.
    */
   PROCESS  zzose_rmm_;

   uint32_t req_disabled_bitmask[64];

#ifdef USE_EPM
   uint32_t next_bp_id;
#endif

/*----------------------------------------------------------------*/

   /* Flag for whenever swap events should be generated for interrupt
    * processes or not. Set from the attribute enable_intproc_swap of the
    * action system parameter. Default is 0, interrupt processes will not
    * generate context switch events by default.
    */
#ifdef HAS_INT_SWAP_HOOK
   uint32_t enable_intproc_swap;
#endif

   /* The maximum number of signal data bytes to trace. */
   uint32_t max_trace_size;

   /* The trace size in dynamic region. */
   uint32_t dynamic_trace_size;

   char *dr_trace;

   /* Actual trace size in dynamic region */
   uint32_t dr_size;

   /* The maximum number of signal data bytes to monitor. */
   uint32_t max_monitor_size;

   /* Whenever monitoring is enabled or not */
   int      monitor_enabled;

   volatile int monitor_overflow;
   volatile int monitor_hook_flow;

   int pending_req;


   /* The client reconnection timeout. */
   U16      reconnect_timeout;

   /*
   ** This flag causes all event counters in the action list to be reset
   ** on each warm start.
   ** When not set, the counters survive warm restarts. This is useful
   ** only in very special cases when it is desired to conduct a debug
   ** session over several system restarts.
   */

   U8       clear_event_count_on_restart;

   /*
   ** The current event sequence number is a global variable managed
   ** by the ldm. When an event occurs in the ldm, only those action
   ** list entries that match the current event sequence number are
   ** honored. Action list entries with sequence number zero are
   ** always honored, regardless of the current event sequence number.
   */

   U8       action_state;

   /*
   ** This flag causes the global event sequence number (action_state) to
   ** be reset on each warm start.
   ** When not set, the counters survive warm restarts. This is useful
   ** only in very special cases when it is desired to conduct a debug
   ** session over several system restarts.
   */

   U8       clear_sequence_on_restart;

   /*
   ** This flag is set to 0 to enable debugger hooks, i.e. tracing,
   ** catching and monitoring for system processes as well.
   ** It is set to SYS_PROCESS to disable it. This is the normal
   ** setting.
   ** Other values are illegal.
   */

   U32      sys_process;


   /*
   ** Breakpoint scope pid list.
   **
   ** The processes in the breakpoints scope are cached in a linked list.
   ** This is because the breakpoint scope is sometimes traversed by an
   ** interrupt process.
   **
   ** Initialized to NULL.
   **
   ** The buffers in this list are signal buffers allocated by either the rmm
   ** or the system daemon. This is ok since they are never freed and those
   ** processes can not die.
   */

   struct break_pid_entry *break_pid_list;

   /*
   ** Scope pointers.
   **
   ** These variables hold pointers to the current scope definitions.
   ** These pointers must be valid whenever the global debugger enable flag
   ** is set ( See also zzset_debug_status ).
   **
   ** Scope entries are allocated from the trace buffer space since
   ** they must survive a system restart.
   ** They survive warm restarts since they are also in the precious struct.
   */

   struct scope *debugger_scope;
   struct scope *breakpoint_scope;

   /*
   ** Scope pointers as wildcard lists.
   **
   ** These structures are present for efficiency. They allow processes to be
   ** quickly tested against the scope definitions by keeping an expanded
   ** list of the debugger_scope and breakpoint_scope variables handy.
   **
   ** Buffers are allocated from the pool, so they don't survive restarts.
   */

   struct procnamelist *quick_debugger_scope;
   struct procnamelist *quick_breakpoint_scope;

   /*
   ** Master action lists.
   **
   ** These are linked lists of all ldefine_action signals currently known to
   ** the RMM. They are stored in hidden trace buffers in order to survive a
   ** system restart.
   ** The list heads survive warm restarts since they are also in the precious struct.
   **
   ** The command field in the signal is used as a next pointer.
   ** The count field is used to hold the size of the original define_action
   ** signal.
   ** The align field is used to hold a verbose wildcard structure of the sig_from
   ** field. It is used to efficiently match processes against wildcard expressions
   ** in debugger hooks.
   ** The ecode field is used to hold a wildcard structure of the sig_to field for
   ** event_types 3, 4, 5 and 14.
   ** The sig_no field is the current event counter.
   **
   ** The dispatcher must be locked while traversing these lists in the dbg hooks.
   ** Interrupt processes don't need any precautions.
   ** Otherwise we might get lost if an entry is deleted.
   */

   struct action_lists actions SMP_ARRAY;

   /*
   ** Catcher management variables.
   **
   ** Catch_enable is nonzero when a catch is allowed. Contains a pointer
   ** to an event_report_verbose signal allocated by the ldm.
   ** This signal is sent to the ldm by the caught process.
   **
   ** caught_pid contains the id of the process that caught the event.
   **
   ** caught_sig is a pointer to a signal pointer, or NULL if no signal caught.
   */

   PROCESS  caught_pid;
   union SIGNAL **caught_sig;
   union SIGNAL *catch_enable;

   /* Used by the send_proc_info and create_truncated_trace_name */
#define TRACE_PROCNAME_MAX_NAMELEN (255)

   /*
   ** ----------------------------------------------------------------
   ** User report concept
   **
   **
   */
   /* This is a pointer that user (application) can set with
    * ose_rmm_set_user_action_handler.
    * It can point to a user specified function that can receive different
    * user reports.
    */
   void (*user_action_function)(union SIGNAL *user_event_report, uint32_t size,
				void *sig_data, OSBUFSIZE data_size);

   /* This is a function pointer that the user (application) can set with
    * ose_rmm_set_user_report_handler function call.
    * If defined, rmm will call the user specified function just before closing a
    * sample interval.
    */
   void (*user_report_function)(SIGSELECT reportno);

   ProfUserReport *prof_first_user_report;        /* Points to the first user report */
   ProfUserReport *prof_cur_user_report;          /* Points to the last(current)user report in link */

#define MAX_NO_OF_ENABLED_USER_REPORTS 50
   UserReportEnabledHashItem *userReportEnabledTable;
   uint32_t userReportEnabledCounter;

   /* Linked list of pending report class requests. */
   union SIGNAL *prof_report_reqlist;

#define USER_RPT_MULTIPLE 1
#define USER_RPT_CONTINUOUS 2
#define USER_RPT_MAXMIN 4

#define TRIG_SID_FOUND     0x01
#define SID_FOUND          0x02
#define EMPTY_SLOT_FOUND   0x10


   uint32_t default_user_interval;
   uint32_t default_user_reports;
   uint32_t default_user_maxvalues;


   /* Ranges for parameters set by host for different profiling actions */
   /* These ranges are set from corresponding system parameters */
   parameter_range proc_interval_range;
   parameter_range proc_reports_range;
   parameter_range proc_maxprocs_range;

   parameter_range block_interval_range;
   parameter_range block_reports_range;
   parameter_range block_maxprocs_range;

   parameter_range pri_interval_range;
   parameter_range pri_reports_range;

   parameter_range cpu_interval_range;
   parameter_range cpu_reports_range;
   parameter_range cpu_priority_range;

   parameter_range user_interval_range;
   parameter_range user_reports_range;
   parameter_range user_maxvalues_range;

   /* Memory limit for profiling buffers */
   uint32_t sample_pool_memory_used;
   uint32_t sample_pool_memory_limit;

   uint32_t last_memory_error;

   /* Default value for reply signal size to MONITOR_GET_TRACE_MULTIPLE in bytes */
   uint32_t def_trace_mget_size; /* @INIT@ = 4096; */
   /* Signal size of MONITOR_GET_TRACE_MULTIPLE_REPLY in units of 4 bytes. */
   uint32_t trace_mget_size;


   /*
   ** These variables contain the system process name list obtained
   ** during system start from the target interface.
   */

   U32      sysproc_count;
   char    *sysprocnames;
   struct procnamelist *expanded_sysprocnames;

   /* Consumed trace memory not used for events (e.g. actionpoints). */
   uint32_t sys_trace_usage;

   /*
   ** This counter shows the total number of processes present in the system.
   ** This variable is used to select the proper trace name buffer mode.
   ** When only a few processes are present, we store a pointer to the
   ** traced name record in each PCB. As the number of processes increase, we
   ** stop saving the name pointer inthe PCB, forcing a new name buffer to
   ** be allocated for each call to the trace hook.
   **
   ** The reason for this is that it is more efficient to save the name pointer,
   ** but as the number of processes rise, the trace becomes cluttered with
   ** non reusable name records, cuasing the trace to trash and finally go
   ** into deadlock.
   **
   ** We currently set the process limit for the mode switch so that not more
   ** than 5% of the trace buffer is filled with names.
   */

   U32      totproc_count;
   U32      trace_mode_limit; /* Not used any more, remove at a later point */

   SMP_CODE(cpuid_t skip_user_prof_lock;) /* Used to allow a tmp recursive lock */
   SMP_CODE(spinlock_t user_prof_lock;)
   SMP_CODE(spinlock_t user_event_lock;)
   SMP_CODE(spinlock_t event_notify_lock;)
   
   /* Start mode, see enum MonitorStartMode for values. */
   uint32_t start_mode;

   /* Enable and disable hardware trace user defined functions. 
    * We assume the compiler does it's job and initialize the 
    * static global rmm_world struct, hence we relay on that these
    * functions pointers are null pointers.
    */
   void(*hw_trace[MAX_NO_HW_TRACE_FUNC])(void);

   /* A structure to maintain a state for Counter/Source Profiling. */
   struct rmm_counter_state counter_state;
} rw;


char **ose_rmm_trace = &rw.trace;
uint32_t *ose_rmm_trace_magic;

#define SYSPARAM_SIGSIZE     0
#define SYSPARAM_ACTION      1
#define SYSPARAM_FLOW_CTRL   2
#define SYSPARAM_PROFILE     3
#define SYSPARAM_INTERFACE   4
#define SYSPARAM_PROF_BLOCK  5
#define SYSPARAM_PROF_PROC   6
#define SYSPARAM_PROF_PRI    7
#define SYSPARAM_PROF_CPU    8
#define SYSPARAM_PROF_USER   9
#define SYSPARAM_SYSTEMNAME 10
#define SYSPARAM_PROF_TOTAL_CPU_BLOCK 11
#define SYSPARAM_PROF_BUFFER_LIMIT 12
#define SYSPARAM_SYS_ERROR  13
#define SYSPARAM_ERROR  14

/* A table of all system parameters */
static const char *rmm_sysparams[] =
{
   "krn/monitor/sigsize",
   "krn/monitor/action",
   "krn/monitor/flow_ctrl",
   "krn/monitor/profile",
   "krn/monitor/interface",
   "krn/monitor/prof/block",
   "krn/monitor/prof/proc",
   "krn/monitor/prof/pri",
   "krn/monitor/prof/cpu",
   "krn/monitor/prof/user",
   "krn/monitor/systemname",
   "krn/monitor/prof/total_cpu_block",
   "krn/monitor/prof/buffer",
   "krn/monitor/sys_error",
   "krn/monitor/error",
   NULL
};

static char        system_name[MAX_NAME_LEN + 1];

extern OSENTRYPOINT ose_rmm_client;

/********* Forward declarations *************/
void free_sample_pool(union SIGNAL **pool);
union SIGNAL *alloc_sample_pool(uint32_t size);

/*----------------------------------------------------------------*/

#if 0                           /* Test code to trace illegal free_bufs. */

#   ifdef free_buf
#      undef free_buf
#   endif

#   define free_buf(s) \
do { \
   ramlog_printf("rmm: freeing signal %x at %d\n", (*((union SIGNAL **)(s)))->sig_no, \
                 __LINE__); \
   zzkrn_free_buf((s)); \
} while (0)
#endif

/*
** Get the priority field from a pcb, selected by a debug_pcb pointer.
** This code is specific for OSE5.5, and depends on the target architecture.
** The reason is the OSE5.5 ARM specific optimization of the pcb.
*/
static OSPRIORITY
get_pcb_priority_from_debug_pcb(const struct debug_pcb *dbg_pcb)
{
#ifdef OSARM
  const int PCB_PRIORITY_OFFSET_FROM_DEBUG_PCB = (-10);
#else
  const int PCB_PRIORITY_OFFSET_FROM_DEBUG_PCB = (-46);
#endif

  /* 0-31=PRI,32=BG,64-95=INT,96-127=TI,128=PHANT */
  return ((OSPRIORITY *)dbg_pcb)[PCB_PRIORITY_OFFSET_FROM_DEBUG_PCB];
}

/*
** This implementation of ose_trace_commit is done for
** performance reason and mainly for the zzfast_swap_trace
** functionality.
** It is compatible with ose_trace_commit in ramtrace.h
** It is therefor essential to keep this up to date with
** ose_trace_commit in ramtrace.h.
*/
static void
ose_trace_commit_inline(void *entry);

static always_inline void
ose_trace_commit_inline(void *entry)
{
   uint8_t *e = entry;
   *(e-16) &= ~1U;   /* the value 16 means that we point to the previous trace_entry */
}

static void
idbuf_init(struct idbuf *buf)
{
   buf->i0 = 0;
   buf->i1 = 0;
#if 0
   ramlog_printf("rmm: idbuf_init\n");
#endif
}


#ifdef USE_EPM
static MonitorStatus
epm_to_rmm_status(EpmStatus epm_status)
{
   switch(epm_status)
   {
      case EPM_OK:
	 return MONITOR_STATUS_OK;
      case EPM_UNKNOWN_FLAGS:
	 return MONITOR_STATUS_INTERNAL_ERROR;
      case EPM_INVALID_ARGUMENTS:           /* No ep is found in list */
	 return MONITOR_STATUS_INTERNAL_ERROR;
      case EPM_INVALID_INSTRUCTION:
	 return MONITOR_STATUS_MEMORY_SET_ERROR;
      case EPM_FUNCTION_NOT_IMPLEMENTED:
	 return MONITOR_STATUS_INTERNAL_ERROR;
      case EPM_INTERNAL_ERROR:
	 return MONITOR_STATUS_KERNEL_ERROR;
      case EPM_NO_UNUSED_HW_BREAKPOINTS:
	 return MONITOR_STATUS_IN_USE;
      case EPM_REQUEST_TO_CLEAR_NONEXISTING_BP:
	  return MONITOR_STATUS_INTERNAL_ERROR;
      case EPM_INIT_FAILED:                /* Present in ose_epm_init_hwbp, which is not used by rmm */
	 return MONITOR_STATUS_EPM_ERROR;
      default:
	 return MONITOR_STATUS_EPM_ERROR;
   }
}

#endif

static          uint32_t
idbuf_contains(struct idbuf *buf, uint32_t pid)
{
   int        i;

   /* Search buffer for 'pid' */
   for (i = buf->i0; i != buf->i1; i = (i + 1) % MAX_IDBUF_SIZE)
   {
      if (buf->pid[i] == pid)
      {
         return 1;
      }
   }

   buf->pid[buf->i1] = pid;
   buf->i1 = (buf->i1 + 1) % MAX_IDBUF_SIZE;

   if (buf->i1 == buf->i0)
   {
      buf->i0 = (buf->i0 + 1) % MAX_IDBUF_SIZE;
   }

   /* Return not found */
   return 0;
}

/*
** These functions facilitate writing and reading 32bit trace entries
** from and to the precious data structure which itself is a 8bit array
*/
static void
set_precious_uint32(uint8_t ms_byte[], uint32_t value)
{
   ms_byte[0] = (value & 0xFF000000);
   ms_byte[1] = (value & 0x00FF0000);
   ms_byte[2] = (value & 0x0000FF00);
   ms_byte[3] = (value & 0x000000FF);
}

static uint32_t
get_precious_uint32(uint8_t ms_byte[])
{
   uint32_t value = 0;

   value |= ms_byte[0] << 24;
   value |= ms_byte[1] << 16;
   value |= ms_byte[2] << 8;
   value |= ms_byte[3];

   return value;
}


/*
** These functions are used to translate a debug_pcb pointer or a raw
** pcb-priority field to a process type suitable for the ptype_from
** and ptype_to fields in the event_report_verbose and related signals.
*/

static          uint32_t
make_ptype_from_pri(uint8_t pri)
{
   static const uint32_t ptype[] = { 0, 64, 128, 256, 512, 0, 0, 0 };
   return ptype[(pri >> 5) & 7];
}

static          uint32_t
make_ptype(const struct debug_pcb *dpcb)
{
   return make_ptype_from_pri(get_pcb_priority_from_debug_pcb(dpcb));
}

/*
 * Expand wild card functions
 *
 */

/*
** This is a wildcard comparison utility.
** It was written by Peter Bergsten in 1995.
**
** The function compares a string against a wildcard that contains
** string fragments mixed with the usual '*' and '?' characters.
** The e parameter contains the wildcard and the c parameter contains
** the non wildcard string to check.
**
** returns zero if the strings match each other and a nonzero value
** if they don't.
**
** Expansion rules: '*' matches zero or more characters.
**                  '?' matches any single character except the string terminator.
**                   Note that "a*" matches "a", but "a?" does not match "a".
**
** The strings are normally traversed until a mismatch occurs or until the
** end of the e-string. If the wildcard ends with a '*' however, the strings
** are only traversed until the end of the wildcard. Thus the "*" wildcard
** causes no loops at all, only three if-statements.
*/

static int
estrcmp(const char *e, const char *c)
{
   const char     *r = (char *) 0;

   while (*e || *c)
   {
      if (*e == '*')
      {
         e++;
         /* Test on the special case of a '*' at the tail. */
         if (!*e)
         {
            return 0;
         }
         r = e;
      }
      else if (*e == '?' && *c)
      {
         e++;
         c++;
      }
      else if (*e == *c)
      {
         e++;
         c++;
      }
      else if (!*c)
         return 1;
      else if (!r)
         return 1;
      else
      {
         if (e != r)
         {
            c -= e - r - 1;
            e = r;
         }
         else
            c++;
      }
   }
   return 0;
}

/* KLUDGE: Same as above but regards ':' as well as '\000' as end of string */
static int
eestrcmp(const char *e, const char *c)
{
   const char     *r = (char *) 0;

   while ((*e && (*e != ':')) || *c)
   {
      if (*e == '*')
      {
         e++;
         /* Test on the special case of a '*' at the tail. */
         if (!*e || *e == ':')
         {
            return 0;
         }
         r = e;
      }
      else if (*e == '?' && *c)
      {
         e++;
         c++;
      }
      else if (*e == *c)
      {
         e++;
         c++;
      }
      else if (!*c)
         return 1;
      else if (!r)
         return 1;
      else
      {
         if (e != r)
         {
            c -= e - r - 1;
            e = r;
         }
         else
            c++;
      }
   }
   return 0;
}


/*
** Convert a string to a 32 bit number.
** Decimal, octal and hexadecimal numbers are accepted as in C constants.
** A leading '-' may be present, which causes a negated result.
** A leading '+' is allowed.
** The string may not contain any characters except the ones that are
** part of the number.
** The result is placed in the variable pointed to by the res parameter.
** The function returns zero if succesful. A nonzero value is returned
** if an incorrectly formed string was found.
** Note that the range check is unsafe for signed numbers.
**
** Requires string.h
*/

static int
strtonum(const char *p, uint32_t * res)
{
   int             negate = 0;
   long            r = 0;
   long            base = 10;

   /* Check for a leading '-' or '+'. */
   if (*p == '-')
   {
      negate = 1;
      p++;
   }
   if (*p == '+')
   {
      p++;
   }

   /* There must be at least one decimal digit present. */
   if ((*p < '0') || (*p > '9'))
      return 1;

   /* Check if octal syntax. */
   if (*p == '0')
   {
      base = 8;
      p++;
      /* Done if there is no more. */
      if (!*p)
      {
         *res = 0L;
         return 0;
      }
      /* Check if hexadecimal syntax. */
      if ((*p == 'x') || (*p == 'X'))
      {
         base = 16;
         p++;
      }
   }

   /* Check if the number is too large. */
   if (base == 16)
   {
      if (strlen((const char *) p) > 8)
         return 1;
   }
   if (base == 10)
   {
      size_t          len = strlen((const char *) p);

      if (len > 10)
         return 1;
      if (len == 10)
         if (strcmp((const char *) p, "4294967295") > 0)
            return 1;
   }
   else
   {
      size_t          len = strlen((const char *) p);

      if (len > 11)
         return 1;
      if (len == 11)
         if (strcmp((const char *) p, "37777777777") > 0)
            return 1;
   }

   /* Read the digits and accumulate them. */
   while (*p)
   {
      r = r * base;
      if ((*p >= '0') && (*p <= '9'))
         r += (long) (*p - '0');
      else if ((*p >= 'a') && (*p <= 'f'))
         r += (long) (*p - 'a' + 10);
      else if ((*p >= 'A') && (*p <= 'F'))
         r += (long) (*p - 'A' + 10);
      else
         return 1;
      p++;
   }

   /* Return the result */
   if (negate)
      *res = (uint32_t) (-r);
   else
      *res = (uint32_t) r;
   return 0;
}

/*
** Subfunction to wildcard_to_plist().
**
** Count the number of colons in a process wildcard string.
*/
static int
count_colons(const char *c)
{
   int             count = 0;

   while (*c)
   {
      if (*c == ':')
         count++;
      c++;
   }
   return count;
}

/*
** Subfunction to wildcard_to_plist().
**
** Split a string to a separate segment, block, and process.
** segment, block, and process are set to NULL if "*" or not present.
*/
static void
split_string(const char **segment, const char **block, const char **process, const char *s)
{
   int colons = count_colons(s);

   if (*s == ':')
   {
   	  s++;
   	  colons--;
   }
   /* Check for no colon. */
   if (colons == 0)
   {
      *segment = NULL;
      *block = NULL;
      *process = s;
      return;
   }
   if (colons == 1)
   /* Split that is either "xxx:xxx" or "xxx:". */
   {
      *segment = NULL;
      *block = s;
      *process = NULL;
      while (*s)
         if (*s++ == ':')
         {
            /* Remove the process if it empty. */
            if (*s)
               *process = s;
            else
               *process = NULL;
            break;
         }
   }
   else if (colons == 2)
   /* Split that is either "xxx:xxx:xxx" or "xxx:xxx:". */
   {
      *segment = s;
      *block = NULL;
      *process = NULL;
      while (*s != ':') s++;
      s++; /* Step over first colon */
      if (*s != ':') *block = s; /* Set only for non empty block */
      while (*s != ':') s++;
      s++; /* Step over second colon */
      if (*s)
         *process = s;
   }

   /* Remove segment that does not contain a "#xxx:". */
   if (*segment && **segment != '#')
   {
      *segment = NULL;
   }

   /* Remove blocks that contain only a "*:". */
   if (*block)
   {
      const char     *q = *block;

      if (*q++ == '*')
         if (*q == ':')
            *block = NULL;
   }

   /* Remove processs that contain only a "*". */
   if (*process)
   {
      const char     *q = *process;

      if (*q++ == '*')
         if (!*q)
            *process = NULL;
   }
}

/*
** Subfunction to wildcard_to_plist().
**
** Get string length up to, but not including the first colon.
*/
static          uint32_t
estrlen(const char *s)
{
   uint32_t        count = 0;

   while (*s)
   {
      if (*s++ == ':')
         return count;
      else
         count++;
   }
   return count;
}

/*
** Subfunction to wildcard_to_plist().
**
** Copy a string until ':' or NUL.
** Add string terminator.
** Return address beyond the terminator.
*/
static char    *
estrcpy(char *dest, const char *src)
{
   for (;;)
   {
      if ((*src == ':') || (!*src))
      {
         *dest++ = 0;
         return dest;
      }
      else
      {
         *dest++ = *src++;
      }
   }
}

/*
** Subfunction to wildcard_to_plist().
**
** This is a general function to sort an array of process id's.
*/

static void
pidsort(PROCESS array[], uint32_t count)
{
   W32             i, j, gap;
   PROCESS        *p1, *p2;
   PROCESS         tmp;

   for (gap = count >> 1; gap > 0; gap >>= 1)
      for (i = gap; i < (W32) count; i++)
         for (j = i - gap; j >= 0; j -= gap)
         {
            p1 = &array[j];
            p2 = &array[j + gap];

            if (*p1 <= *p2)
               break;

            tmp = *p1;
            *p1 = *p2;
            *p2 = tmp;
         }
}

/*
** This function expands a list of process names to a verbose
** procnamelist structure.
**
** The result is returned in a signal buffer obtained from alloc().
** The caller must free this buffer when information is no longer needed.
** The NULL pointer is returned if there is a syntax error in a string
** or if the signal buffer required is too large to be allocated.
**
** We support any flavor of the "block:process" form, including
** the special substring formats "#0x...." and "$0x...." for
** pids and entrypoints.
**
** Some combinations are invalid, so the list of valid strings looks like this:
**      <name>
**      <pid>
**      <entrypoint>
**      <pid>:
**      <pid>:<name>
**      <pid>:<entrypoint>
**      <name>:
**      <name>:<name>
**      <name>:<entrypoint>
*/

static struct procnamelist *
wildcard_to_plist(uint32_t strcount, const char *stringlist)
{
   struct procnamelist *plist = NULL;
   uint32_t        i;
   const char     *c;
   const char     *segment, *head, *tail;
   char           *stringpool;

   /* Structure space requirements. */
   uint32_t        no_of_singlepids = 0;
   uint32_t        no_of_complexpids = 0;
   uint32_t        string_size = 0;

   /*
    ** Scan all strings to calculate space requirements.
    */
   for (i = strcount, c = stringlist; i; i--)
   {
      /* Check that string contains at most two ":". */
      if (count_colons(c) > 2)
      {
      	 ramlog_printf("rmm: scope string syntax error: colons %d\n",
      	    count_colons(c));
         return NULL;
      }

      /* Split the string. */
      split_string(&segment, &head, &tail, c);

      /* Skip to the next string in the list. */
      while (*c++)
      {;                        /* Empty loop body. */
      }

      /* Check for a single pid and update space counters. */
      if ((!segment) && (!head) && tail && (*tail == '#'))
      {
         no_of_singlepids++;
         /* No, need to allocate a complex pid. */
      }
      else
      {
         no_of_complexpids++;
      }

      /* Check for wildcard stringspace and update string pool size. */
      if (head)
      {
         if (*head != '#')
         {
            string_size = string_size + 1 + estrlen(head);
         }
      }

      if (tail)
      {
         if ((*tail != '#') && (*tail != '$'))
         {
            string_size = string_size + 1 + strlen(tail);
         }
      }
   }

   /* Allocate a signal buffer to hold the procnamelist struct. */
   {
      /* Calculate signal buffer size. */
      uint32_t        bufsize = sizeof (struct procnamelist) + string_size +
         no_of_singlepids * sizeof (PROCESS) + no_of_complexpids * sizeof (struct complexpid);

      /* Check that such a signal can be allocated by the LDM. If not, return NULL */
      {
         struct OS_pcb  *pcb;
         OSADDRESS       max_sigsize;

         /* Get pcb info for the rmm. */
         if (!(pcb = get_pcb(current_process())))
         {
            ramlog_printf("rmm: failed to get pcb for rmm\n");
            /* fail if we can't get it. */
            return NULL;
         }
         /* Pick up the largest possible signal buffer in the ldm. */
         max_sigsize = pcb->max_sigsize;
         /* Free the pcb structure. */
         free_buf((union SIGNAL **) &pcb);
         /* Fail if the needed size is larger than the size we can get. */
         if (max_sigsize <= (OSADDRESS) bufsize)
         {
            ramlog_printf("rmm: failed to alloc proc name list max_sigsize %d bufsize %d\n",
               max_sigsize, bufsize);
            return NULL;
         }
      }

      /* Allocate the signal structure and initialize base pointers. */
      plist = (struct procnamelist *) alloc((OSBUFSIZE) bufsize, 0);
      plist->no_of_singlepids = no_of_singlepids;
      plist->no_of_complexpids = no_of_complexpids;
      plist->singlepidlist = (PROCESS *) & plist->complexpidlist[no_of_complexpids];
      stringpool = (char *) (plist->singlepidlist + no_of_singlepids);
   }

   /* Fill the procnamelist structure. */
   {
      struct complexpid *putcomplex = plist->complexpidlist;
      PROCESS        *putsingle = plist->singlepidlist;
      char           *putstr = stringpool;

      for (i = strcount, c = stringlist; i; i--)
      {
         /* Split the string. */
         split_string(&segment, &head, &tail, c);

         /* Skip to the next string in the list. */
         while (*c++)
         {;                     /* Empty loop body. */
         }

         /* Check for a single pid. */
         if ((!segment) && (!head) && tail && (*tail == '#'))
         {
            /* Convert to binary and store it in the single pid array. */
            if (strtonum(tail + 1, (uint32_t *) putsingle++))
            {
               ramlog_printf("rmm: scope string contains badly formated number: %s\n",
                  tail + 1);
               /* If the number is badly formatted then free the buffer and exit. */
               free_buf((union SIGNAL **) &plist);
               return NULL;
            }
            /* Go for the next input string. */
            continue;
         }

         if (segment && *segment == '#')
         {
             static char     qstr[16];        /* Temp so that we can scratch the ':'. */
             char           *q;

             /* make a copy of the string and replace the ':' with a string terminator. */
             strncpy(qstr, segment + 1, 15);
             qstr[15] = 0;
             for (q = qstr; *q; q++)
             {
                if (*q == ':')
                {
                   *q = 0;
                   break;
                }
             }
             /* Store the type for a binary sid and convert the string to binary. */
             putcomplex->segmenttype = 1;
             if (strtonum(qstr, (uint32_t *) & putcomplex->sid))
             {
                ramlog_printf("rmm: scope string contains badly formated number: %s\n",
                  qstr);
                /* If the number is badly formatted then free the buffer and exit. */
                free_buf((union SIGNAL **) &plist);
                return NULL;
             }
         }
         else
         {
            putcomplex->segmenttype = 0;
            putcomplex->sid = 0;
         }

         /* Analyze the blockname substring. */
         if (!head)
         {
            /* No string means that we should match anything. */
            putcomplex->blocknametype = 0;
         }
         else
         {
            /* Strings that start with '#' are converted to a binary pid. */
            if (*head == '#')
            {
               static char     qstr[16];        /* Temp so that we can scratch the ':'. */
               char           *q;

               /* make a copy of the string and replace the ':' with a string terminator. */
               strncpy(qstr, head + 1, 15);
               qstr[15] = 0;
               for (q = qstr; *q; q++)
                  if (*q == ':')
                  {
                     *q = 0;
                     break;
                  }
               /* Store the type for a binary bid and convert the string to binary. */
               putcomplex->blocknametype = 1;
               if (strtonum(qstr, (uint32_t *) & putcomplex->blockptr.bid))
               {
                  ramlog_printf("rmm: scope string contains badly formated number: %s\n",
                     qstr);
                  /* If the number is badly formatted then free the buffer and exit. */
                  free_buf((union SIGNAL **) &plist);
                  return NULL;
               }
            }
            /* Anything else is treated as a wildcard name string. */
            else
            {
               putcomplex->blocknametype = 2;
               putcomplex->blockptr.name = putstr;
               putstr = estrcpy(putstr, head);
            }
         }

         /* Analyze the process name substring. */
         if (!tail)
         {
            /* No string means that we should match anything. */
            putcomplex->procnametype = 0;
         }
         else
         {
            /* Strings that start with '#' are converted to a binary pid. */
            if (*tail == '#')
            {
               putcomplex->procnametype = 1;
               if (strtonum(tail + 1, (uint32_t *) & putcomplex->procptr.pid))
               {
                  ramlog_printf("rmm: scope string contains badly formated number: %s\n",
                     tail + 1);
                  /* If the number is badly formatted then free the buffer and exit. */
                  free_buf((union SIGNAL **) &plist);
                  return NULL;
               }
            }
            /* Strings that start with '$' are converted to a binary entrypoint. */
            else if (*tail == '$')
            {
               putcomplex->procnametype = 3;
               if (strtonum(tail + 1, (uint32_t *) & putcomplex->procptr.entrypoint))
               {
                  ramlog_printf("rmm: scope string contains badly formated number: %s\n",
                     tail + 1);
                  /* If the number is badly formatted then free the buffer and exit. */
                  free_buf((union SIGNAL **) &plist);
                  return NULL;
               }
            }
            /* Anything else is treated as a wildcard name string. */
            else
            {
               putcomplex->procnametype = 2;
               putcomplex->procptr.name = putstr;
               putstr = estrcpy(putstr, tail);
            }
         }

         /* Bump the complex pid pointer. */
         putcomplex++;
      }
   }

   /* Sort the array of single pids. */
   if (no_of_singlepids > 1)
      pidsort(plist->singlepidlist, no_of_singlepids);

   /* Return the created structure. */
   return plist;
}

/**
 * Convert a monitor scope to an old LDM procnamelist
 */
struct procnamelist *
zzrmm_scope_to_plist(MonitorScope * scope, MonitorTag * tags, uint32_t tags_count)
{
   struct procnamelist *plist;
   MonitorTag          *scope_tag = tags;
   OSBUFSIZE            plist_size;

   /* Find relevant tag */
   if (scope->type == MONITOR_SCOPE_NAME_PATTERN || scope->type == MONITOR_SCOPE_ENTRYPOINT)
   {
      MonitorTag     *end_tag = tags + tags_count;

      for (;;)
      {
         if (scope_tag == end_tag)
         {
            return NULL;
         }
         if (scope_tag->tag.type == MONITOR_SCOPE_TAG_NAME || scope_tag->tag.type == MONITOR_SCOPE_TAG_ENTRYPOINT)
         {
            break;
         }
         scope_tag += scope_tag->tag.count;
      }
   }
   
   plist_size = (scope->type == MONITOR_SCOPE_NAME_PATTERN)?
      sizeof (struct procnamelist) + strlen(scope_tag[1].str):
      sizeof (struct procnamelist);
   plist = (void *) alloc(plist_size, 0);
   memset(plist, 0, plist_size);
   plist->no_of_complexpids = 1;

   switch (scope->type)
   {
      case MONITOR_SCOPE_ID:
         plist->complexpidlist[0].procptr.pid = scope->id;
         plist->complexpidlist[0].procnametype = 1;
         break;
      case MONITOR_SCOPE_BLOCK_ID:
         plist->complexpidlist[0].blockptr.bid = scope->id;
         plist->complexpidlist[0].blocknametype = 1;
         break;
      case MONITOR_SCOPE_SEGMENT_ID:
         plist->complexpidlist[0].sid = scope->id;
         plist->complexpidlist[0].segmenttype = 1;
         break;
      case MONITOR_SCOPE_SYSTEM:
         break;
      case MONITOR_SCOPE_NAME_PATTERN:
      {
         const char     *segment;
         const char     *block;
         const char     *proc;
         char           *name = (void *) &plist->complexpidlist[1];

         split_string(&segment, &block, &proc, scope_tag[1].str);
         if (NULL != segment)
         {
         	/* Illegal scope type */
            free_buf((union SIGNAL **) &plist);
            plist = (void *) 0;
            break;
         }
         plist->complexpidlist[0].segmenttype = 0;

         if (block)
         {
            plist->complexpidlist[0].blockptr.name = name;
            plist->complexpidlist[0].blocknametype = 2;
            estrcpy(name, block);
            name += estrlen(block) + 1;
         }
         if (proc)
         {
            plist->complexpidlist[0].procptr.name = name;
            plist->complexpidlist[0].procnametype = 2;
            estrcpy(name, proc);
         }
         break;
      }
      case MONITOR_SCOPE_ENTRYPOINT:
         plist->complexpidlist[0].procptr.entrypoint = (void *) scope_tag[1].num;
         plist->complexpidlist[0].procnametype = 3;
         break;
      default:
         free_buf((union SIGNAL **) &plist);
         plist = (void *) 0;
         break;
   }
   return plist;
}

static          MonitorScopeType
pattern_to_scopetype(const char *str)
{
   const char     *proc;
   const char     *block;
   const char     *segment;
   MonitorScopeType ret = MONITOR_SCOPE_SYSTEM;

   /*
    * Handle classic LDM scopes (with segment ids) that are the result of a
    * conversion from an original structured scope. No mix of IDs and name
    * patterns.
    */
   assert(count_colons(str) < 3);

   split_string(&segment, &block, &proc, str);

   if (segment)
   {
      ret = MONITOR_SCOPE_SEGMENT_ID;
   }

   if (block)
   {
      if (*block == '#')
         ret = MONITOR_SCOPE_BLOCK_ID;
      else
         ret = MONITOR_SCOPE_NAME_PATTERN;
   }

   if (proc)
   {
      if (*proc == '#')
         ret = MONITOR_SCOPE_ID;
      else if (*proc == '$')
         ret = MONITOR_SCOPE_ENTRYPOINT;
      else
         ret = MONITOR_SCOPE_NAME_PATTERN;
   }

   return ret;
}

static void
pattern_to_scope(const char *str, MonitorScope * scope, MonitorTag * tags, uint32_t * tagsCount)
{
   const char     *proc;
   const char     *block;
   const char     *segment;

   assert(str);
   assert(scope);
   assert(tags);
   assert(tagsCount);

   /*
    * Handle classic LDM scopes (with segment ids) that are the result of a
    * conversion from an original structured scope. No mix of IDs and name
    * patterns.
    */
   assert(count_colons(str) < 3);

   split_string(&segment, &block, &proc, str);

   if (proc)
   {
      if (*proc == '#')
      {
         char           *p;
         long            id;

         id = strtoul(proc + 1, &p, 0);

         if (*p == 0)
         {
            scope->type = MONITOR_SCOPE_ID;
            scope->id = id;
         }
         else
         {
            scope->type = MONITOR_SCOPE_VOID;
            scope->id = 0;
         }
      }
      else if (*proc == '$')
      {
         char           *p;
         long            id;

         id = strtoul(proc + 1, &p, 0);

         if (*p == 0)
         {
            scope->type = MONITOR_SCOPE_ENTRYPOINT;
            scope->id = id;
         }
         else
         {
            scope->type = MONITOR_SCOPE_VOID;
            scope->id = 0;
         }
      }
      else
      {
         size_t          len;

         if (block && *block == '#')    /* Ignore BID in BID:NAME */
         {
            str = proc;
         }

         len = 1 + 1 + strlen(str) / sizeof (MonitorTag);

         scope->type = MONITOR_SCOPE_NAME_PATTERN;
         tags->tag.type = MONITOR_SCOPE_TAG_NAME;
         tags->tag.count = len;
         strcpy((tags + 1)->str, str);
         *tagsCount += len;
      }
   }
   else if (block)
   {
      if (*block == '#')
      {
         char           *p;
         long            id;

         id = strtoul(block + 1, &p, 0);

         if (*p == ':')
         {
            scope->type = MONITOR_SCOPE_BLOCK_ID;
            scope->id = id;
         }
         else
         {
            scope->type = MONITOR_SCOPE_VOID;
            scope->id = 0;
         }
      }
      else
      {
         size_t          len = 1 + 1 + strlen(str) / sizeof (MonitorTag);

         scope->type = MONITOR_SCOPE_NAME_PATTERN;
         tags->tag.type = MONITOR_SCOPE_TAG_NAME;
         tags->tag.count = len;
         strcpy((tags + 1)->str, str);
         *tagsCount += len;
      }
   }
   else if (segment)
   {
      if (*segment == '#')
      {
         char           *p;
         long            id;

         id = strtoul(segment + 1, &p, 0);

         if (*p == ':')
         {
            scope->type = MONITOR_SCOPE_SEGMENT_ID;
            scope->id = id;
         }
         else
         {
            scope->type = MONITOR_SCOPE_VOID;
            scope->id = 0;
         }
      }
     else
     {
        scope->type = MONITOR_SCOPE_VOID;
        scope->id = 0;
     }
   }
   else
   {
      scope->type = MONITOR_SCOPE_SYSTEM;
      scope->id = 0;
   }
}

/**
 * Convert a monitor scope to an old LDM wildcard string
 */
static          uint32_t
scope_to_pattern_len(MonitorScope * scope, MonitorTag * tags, uint32_t tags_count, uint32_t ignore)
{
   switch (scope->type)
   {
      case MONITOR_SCOPE_ID:
         return 12;             /* strlen("#0x00000000") + sizeof('\0') */
      case MONITOR_SCOPE_BLOCK_ID:
         return 13;             /* strlen("#0x00000000:") + sizeof('\0') */
      case MONITOR_SCOPE_SEGMENT_ID:
         return 15;             /* strlen("#0x00000000:*:") + sizeof('\0') */
      case MONITOR_SCOPE_SYSTEM:
         return 2;              /* strlen("*") + sizeof('\0') */
      case MONITOR_SCOPE_NAME_PATTERN:
      {
         MonitorTag     *scope_tag = tags;
         MonitorTag     *end_tag = tags + tags_count;

         for (;;)
         {
            if (scope_tag == end_tag)
            {
               return 0;
            }
            if (scope_tag->tag.type == MONITOR_SCOPE_TAG_NAME)
            {
               if (ignore == MONITOR_SCOPE_NAME_PATTERN)
                  ignore = 0;
               else
                  break;
            }
            scope_tag += scope_tag->tag.count;
         }

         return strlen(scope_tag[1].str) + 1;
      }
      case MONITOR_SCOPE_ENTRYPOINT:
         return 12;             /* strlen("$0x00000000")  + sizeof('\0') */
      default:
         return 0;
   }
}

/**
 * Convert a monitor scope to an old LDM wildcard string
 */
static          uint32_t
scope_to_pattern(MonitorScope * scope, MonitorTag * tags, uint32_t tags_count, uint32_t ignore, char *str)
{
   switch (scope->type)
   {
      case MONITOR_SCOPE_ID:
         sprintf(str, "#0x%0.8lx", (long) scope->id);
         break;
      case MONITOR_SCOPE_BLOCK_ID:
         sprintf(str, "#0x%0.8lx:", (long) scope->id);
         break;
      case MONITOR_SCOPE_SEGMENT_ID:
         sprintf(str, "#0x%0.8lx:*:", (long) scope->id);
         break;
      case MONITOR_SCOPE_SYSTEM:
         sprintf(str, "*");
         break;
      case MONITOR_SCOPE_NAME_PATTERN:
      {
         MonitorTag     *scope_tag = tags;
         MonitorTag     *end_tag = tags + tags_count;

         for (;;)
         {
            if (scope_tag == end_tag)
            {
               return 0;
            }
            if (scope_tag->tag.type == MONITOR_SCOPE_TAG_NAME)
            {
               if (ignore == MONITOR_SCOPE_NAME_PATTERN)
                  ignore = 0;
               else
                  break;
            }
            scope_tag += scope_tag->tag.count;
         }
         sprintf(str, "%s", scope_tag[1].str);
         break;
      }
      case MONITOR_SCOPE_ENTRYPOINT:
      {
         MonitorTag     *scope_tag = tags;
         MonitorTag     *end_tag = tags + tags_count;

         for (;;)
         {
            if (scope_tag == end_tag)
            {
               return 0;
            }
            if (scope_tag->tag.type == MONITOR_SCOPE_TAG_ENTRYPOINT)
            {
               if (ignore == MONITOR_SCOPE_ENTRYPOINT)
                  ignore = 0;
               else
                  break;
            }
            scope_tag += scope_tag->tag.count;
         }
         sprintf(str, "$0x%0.8lx", (long) scope_tag[1].num);
         break;
      }
      default:
         return 0;
   }

   /* Double zero termination */
   str[strlen(str) + 1] = 0;
   return 1;
}

static          int64_t
calc_checksum(const void *start, const uint32_t size)
{
   int64_t         sum = 0;
   const char     *c = (const char *) start;
   uint32_t        i;

   for (i = 0; i < size; i++)
   {
      sum += ~((i << 8) + c[i]);
   }

   return sum;
}

static void
update_precious_checksum(void)
{
   uint32_t        size = ((uint32_t) & rw.precious->trace_enabled) - ((uint32_t) & rw.precious->magic);

   rw.precious->checksum = calc_checksum(&rw.precious->magic, size);
}

static int
check_precious_checksum(void)
{
   int64_t         checksum;
   uint32_t        size = ((uint32_t) & rw.precious->trace_enabled) - ((uint32_t) & rw.precious->magic);

   checksum = calc_checksum(&rw.precious->magic, size);
   return checksum == rw.precious->checksum;
}

#define TRACE_USAGE_SIZE(s) ((((s) + 15) & (~0 << 4)) + 16)

static void *
sys_alloc_trace(char *trace, EntrySize size)
{
   void *buf = ose_trace_alloc(trace, size, T_HIDDEN | T_FORCE | T_NOTIME);
   if (buf)
   {
      rw.sys_trace_usage += TRACE_USAGE_SIZE(size);
   }
   return buf;
}

/*
** Subfunction to the LDM.
**
** It is called during the system start procedure to locate the
** ldm variable set and check it.
** Returns zero on success. Nonzero on failure, in which case
** a subsequent create_precious() is expected.
**
** The global precious pointer is set, as well as global copies of the
** variable set.
*/
static int
recover_precious(void)
{
   struct action_lists *actions = &rw.actions SMP_CODE([0]);

   /* Check the magic number. Error if not OK. */
   if (rw.precious->magic != PRECIOUS_MAGIC || rw.precious->magic_tail != PRECIOUS_MAGIC)
   {
      return 1;
   }

   if (!check_precious_checksum())
   {
      ramlog_printf("rmm: precious data not recovered due to checksum error\n");
      return 2;
   }

   /* Temporary destroy magic if we should crash during recovery */
   rw.precious->magic = ~PRECIOUS_MAGIC;

   if (rw.precious->trace_enabled != 0 && rw.precious->trace_enabled != 1)
   {
      ramlog_printf("rmm: precious data not recovered due to illegal"
                    " trace_enabled value\n");
      return 3;
   }
   if (rw.precious->action_state > 255)
   {
      ramlog_printf("rmm: precious data not recovered due to illegal"
                    " action_state value\n");
      return 4;
   }

   /* Set all pointers and flags. */
   rw.debugger_scope = rw.precious->debugger_scope;
   rw.breakpoint_scope = rw.precious->breakpoint_scope;
   if (rw.debugger_scope)
   {
      if (!(rw.quick_debugger_scope = wildcard_to_plist(rw.debugger_scope->count, rw.debugger_scope->scope_list)))
      {
         rw.quick_debugger_scope = wildcard_to_plist(1, "*");
      }
   }
   else
   {
      rw.quick_debugger_scope = wildcard_to_plist(1, "*");
   }
   if (rw.breakpoint_scope)
   {
      if (!(rw.quick_breakpoint_scope = wildcard_to_plist(rw.breakpoint_scope->count, rw.breakpoint_scope->scope_list)))
      {
         rw.quick_breakpoint_scope = wildcard_to_plist(1, "*");
      }
   }
   else
   {
      rw.quick_breakpoint_scope = wildcard_to_plist(1, "*");
   }


   actions->action_list_create = rw.precious->action_list_create;
   actions->action_list_kill = rw.precious->action_list_kill;
   actions->action_list_swap = rw.precious->action_list_swap;
   actions->action_list_send = rw.precious->action_list_send;
   actions->action_list_receive = rw.precious->action_list_receive;
   actions->action_list_oserror = rw.precious->action_list_oserror;
   actions->action_list_error = rw.precious->action_list_error;
   actions->action_list_restart = rw.precious->action_list_restart;
   actions->action_list_uevent = rw.precious->action_list_uevent;
#ifdef OSE_SMP
   actions->action_list_bind = rw.precious->action_list_bind;
#endif
   actions->action_list_alloc = rw.precious->action_list_alloc;
   actions->action_list_free_buf = rw.precious->action_list_free_buf;
#ifdef HAS_TIMEOUT_EVENT
   actions->action_list_timeout = rw.precious->action_list_timeout;
#endif

   rw.trace_enabled = rw.precious->trace_enabled;
   rw.master_monitor_enabled = rw.precious->master_monitor_enabled;

   /* Pick up ldm variables. */
   rw.max_trace_size = (rw.precious->options[0] << 8) + rw.precious->options[1];
   rw.max_monitor_size = (rw.precious->options[2] << 8) + rw.precious->options[3];
   rw.reconnect_timeout = (rw.precious->options[4] << 8) + rw.precious->options[5];
   rw.clear_event_count_on_restart = rw.precious->options[6];
   rw.action_state = rw.precious->action_state;
   rw.clear_sequence_on_restart = rw.precious->options[9];
   rw.sys_process = rw.precious->options[11] ? 0 : SYS_PROCESS;
   rw.break_scope_status = (OSBOOLEAN) (rw.precious->break_scope_status ? 1 : 0);
   rw.flow_ctrl_block = rw.precious->options[12];
   rw.flow_ctrl_nblock = rw.precious->options[13];
   rw.trace_mget_size = (rw.precious->options[14] << 8) + rw.precious->options[15];
   rw.trace_mget_size >>= 2; /* trace_mget_size is in units of 4 bytes */
   rw.monitor_sys_error = rw.precious->options[16];
   zzprof.prof_all_min_us = rw.precious->options[19];
   rw.dynamic_trace_size = rw.precious->options[20] << 24;
   rw.dynamic_trace_size |= rw.precious->options[21] << 16;
   rw.dynamic_trace_size |= rw.precious->options[22] << 8;
   rw.dynamic_trace_size |= rw.precious->options[23];
   rw.monitor_error = rw.precious->options[17];

   /* Retreive the monitor profiling parameters*/
   zzprof.proc[0].default_proc_interval = get_precious_uint32(&rw.precious->options[24]);
   zzprof.proc[0].default_proc_reports = get_precious_uint32(&rw.precious->options[28]);
   zzprof.proc[0].default_proc_maxprocs = get_precious_uint32(&rw.precious->options[32]);
   zzprof.proc[1].default_proc_interval = get_precious_uint32(&rw.precious->options[36]);
   zzprof.proc[1].default_proc_reports = get_precious_uint32(&rw.precious->options[40]);
   zzprof.proc[1].default_proc_maxprocs = get_precious_uint32(&rw.precious->options[44]);
   zzprof.pri.default_pri_interval = get_precious_uint32(&rw.precious->options[48]);
   zzprof.pri.default_pri_reports = get_precious_uint32(&rw.precious->options[52]);
   zzprof.cpu.default_cpu_interval = get_precious_uint32(&rw.precious->options[56]);
   zzprof.cpu.default_cpu_reports = get_precious_uint32(&rw.precious->options[60]);
   zzprof.cpu.default_cpu_priority = get_precious_uint32(&rw.precious->options[64]);
/* until we have a place in the struct */
   zzprof.tcb.default_interval = DEF_PROF_INTERVAL;
   zzprof.tcb.default_reports = DEF_PROF_REPORTS;
   zzprof.tcb.default_priority = DEF_PROF_PRIORITY;
   zzprof.tcb.default_maxblocks = DEF_PROF_MAX_VALUES;
   zzprof.tcb.default_oversize = DEF_PROF_OVERSIZE;
   /* possible ?
    *
    * zzprof.proc[0].default_proc_pri = 33;
    * zzprof.proc[1].default_proc_pri = 33;
    */

   rw.default_user_interval = get_precious_uint32(&rw.precious->options[68]);
   rw.default_user_reports = get_precious_uint32(&rw.precious->options[72]);
   rw.default_user_maxvalues = get_precious_uint32(&rw.precious->options[76]);

#ifdef HAS_INT_SWAP_HOOK
   rw.enable_intproc_swap = rw.precious->options[80];
#endif

   /* Set the error check level on warm start. */
   get_sysconf(3, (OSADDRESS) rw.precious->options[10], 0, 0);

   return 0;
}

static int
get_monitor_sysparam_index(const char *name)
{
   int i;

   for (i = 0; rmm_sysparams[i] != NULL; i++)
   {
      if (0 == strcmp(name, rmm_sysparams[i]))
      {
         return i;
      }
   }

   return -1;
}

static void
read_monitor_sysparam(int param_idx, const char *attr, long *value,
                      long def_value, long min_value, long max_value)
{
   assert(param_idx >= 0);
   assert(param_idx < sizeof(rmm_sysparams) / sizeof(char *));
   assert(attr);
   assert(strlen(rmm_sysparams[param_idx]) <= SYSPARAM_MAX_NAME_LENGTH);
   assert(min_value <= max_value);
   assert(def_value >= min_value);
   assert(def_value <= max_value);

   if (sysparam_readattr_long(rmm_sysparams[param_idx], attr, value))
   {
      *value = def_value;
   }
   else if (*value < min_value)
   {
      ramlog_printf("rmm: system parameter \"%s:%s\" has a too small value: %d\n",
                    rmm_sysparams[param_idx], attr, *value);
      *value = min_value;
   }
   else if (*value > max_value)
   {
      ramlog_printf("rmm: system parameter \"%s:%s\" has a too large value: %d\n",
                    rmm_sysparams[param_idx], attr, *value);
      *value = max_value;
   }
}

/* Check that min value is not greater than max and adjust. */
static void
check_range(const char *parameter, parameter_range *range)
{
   if (range->min > range->max)
   {
      ramlog_printf("rmm: minimum value is greater than maximum value for %s:"
            " [%u, %u]. min value is adjusted to %u\n",
            parameter, range->min, range->max, range->max);
      range->min = range->max;
   }
}

/*
** Function : check_prof_def_value
** Checks the default value of the system parameter and truncates if it is
** outside range
*/
static void
check_prof_def_value (parameter_range range, unsigned int *value)
{
   if (*value < range.min)
   {
      *value = range.min;
   }
   else if (*value > range.max)
   {
      *value = range.max;
   }
}

/* Setup range variables for per process profiling */
static void
setup_parameter_ranges_proc(void)
{
   long value;

   read_monitor_sysparam(SYSPARAM_PROF_PROC, "interval_max", &value,
            MAX_PROF_INTERVAL, MIN_PROF_INTERVAL, MAX_PROF_INTERVAL);
   rw.proc_interval_range.max = value;

   read_monitor_sysparam(SYSPARAM_PROF_PROC, "interval_min", &value,
            MIN_PROF_INTERVAL, MIN_PROF_INTERVAL, MAX_PROF_INTERVAL);
   rw.proc_interval_range.min = value;
   check_range("cpu per proc profiling interval", &rw.proc_interval_range);

   read_monitor_sysparam(SYSPARAM_PROF_PROC, "reports_max", &value,
            MAX_PROF_REPORTS, MIN_PROF_REPORTS, MAX_PROF_REPORTS);
   rw.proc_reports_range.max = value;

   read_monitor_sysparam(SYSPARAM_PROF_PROC, "reports_min", &value,
            MIN_PROF_REPORTS, MIN_PROF_REPORTS, MAX_PROF_REPORTS);
   rw.proc_reports_range.min = value;
   check_range("cpu per proc profiling reports", &rw.proc_reports_range);

   read_monitor_sysparam(SYSPARAM_PROF_PROC, "procs_max", &value,
            zzkrn_get_pid_info(0, 4), MIN_PROF_MAXPROCS, zzkrn_get_pid_info(0,
                     4));
   rw.proc_maxprocs_range.max = value;

   read_monitor_sysparam(SYSPARAM_PROF_PROC, "procs_min", &value,
            MIN_PROF_MAXPROCS, MIN_PROF_MAXPROCS, zzkrn_get_pid_info(0, 4));
   rw.proc_maxprocs_range.min = value;
   check_range("cpu per proc profiling maxprocs", &rw.proc_maxprocs_range);

   /* Check default value of cpu per proc profiling interval */
   check_prof_def_value (rw.proc_interval_range,
      (unsigned int *)&zzprof.proc[0].default_proc_interval);

   /* Check default value of cpu per proc profiling reports */
   check_prof_def_value (rw.proc_reports_range,
      &zzprof.proc[0].default_proc_reports);

   /* Check default value of cpu per proc profiling maxprocs */
   check_prof_def_value (rw.proc_maxprocs_range,
      &zzprof.proc[0].default_proc_maxprocs);
}

/* Setup range variables for per block profiling */
static void
setup_parameter_ranges_block(void)
{
   long value;

   read_monitor_sysparam(SYSPARAM_PROF_BLOCK, "interval_max", &value,
            MAX_PROF_INTERVAL, MIN_PROF_INTERVAL, MAX_PROF_INTERVAL);
   rw.block_interval_range.max = value;

   read_monitor_sysparam(SYSPARAM_PROF_BLOCK, "interval_min", &value,
            MIN_PROF_INTERVAL, MIN_PROF_INTERVAL, MAX_PROF_INTERVAL);
   rw.block_interval_range.min = value;
   check_range("cpu per block profiling interval", &rw.block_interval_range);

   read_monitor_sysparam(SYSPARAM_PROF_BLOCK, "reports_max", &value,
            MAX_PROF_REPORTS, MIN_PROF_REPORTS, MAX_PROF_REPORTS);
   rw.block_reports_range.max = value;

   read_monitor_sysparam(SYSPARAM_PROF_BLOCK, "reports_min", &value,
            MIN_PROF_REPORTS, MIN_PROF_REPORTS, MAX_PROF_REPORTS);
   rw.block_reports_range.min = value;
   check_range("cpu per block profiling reports", &rw.block_reports_range);

   read_monitor_sysparam(SYSPARAM_PROF_BLOCK, "procs_max", &value,
            zzkrn_get_pid_info(0, 4), MIN_PROF_MAXPROCS, zzkrn_get_pid_info(0,
                     4));
   rw.block_maxprocs_range.max = value;

   read_monitor_sysparam(SYSPARAM_PROF_BLOCK, "procs_min", &value,
            MIN_PROF_MAXPROCS, MIN_PROF_MAXPROCS, zzkrn_get_pid_info(0, 4));
   rw.block_maxprocs_range.min = value;
   check_range("cpu per block profiling procs", &rw.block_maxprocs_range);

   /* Check default value of cpu per block profiling interval */
   check_prof_def_value (rw.block_interval_range,
      (unsigned int *)&zzprof.proc[1].default_proc_interval);

   /* Check default value of cpu per block profiling reports */
   check_prof_def_value (rw.block_reports_range,
      &zzprof.proc[1].default_proc_reports);

   /* Check default value of cpu per block profiling maxprocs */
   check_prof_def_value (rw.block_maxprocs_range,
      &zzprof.proc[1].default_proc_maxprocs);
}

/* Setup range variables for per priority profiling */
static void
setup_parameter_ranges_pri(void)
{
   long value;

   read_monitor_sysparam(SYSPARAM_PROF_PRI, "interval_max", &value,
            MAX_PROF_INTERVAL, MIN_PROF_INTERVAL, MAX_PROF_INTERVAL);
   rw.pri_interval_range.max = value;

   read_monitor_sysparam(SYSPARAM_PROF_PRI, "interval_min", &value,
            MIN_PROF_INTERVAL, MIN_PROF_INTERVAL, MAX_PROF_INTERVAL);
   rw.pri_interval_range.min = value;
   check_range("cpu per pri profiling interval", &rw.pri_interval_range);

   read_monitor_sysparam(SYSPARAM_PROF_PRI, "reports_max", &value,
            MAX_PROF_REPORTS, MIN_PROF_REPORTS, MAX_PROF_REPORTS);
   rw.pri_reports_range.max = value;

   read_monitor_sysparam(SYSPARAM_PROF_PRI, "reports_min", &value,
            MIN_PROF_REPORTS, MIN_PROF_REPORTS, MAX_PROF_REPORTS);
   rw.pri_reports_range.min = value;
   check_range("cpu per pri profiling reports", &rw.pri_reports_range);

   /* Check default value of cpu per pri profiling interval */
   check_prof_def_value (rw.pri_interval_range,
      (unsigned int *)&zzprof.pri.default_pri_interval);

   /* Check default value of cpu per pri profiling reports */
   check_prof_def_value (rw.pri_reports_range, &zzprof.pri.default_pri_reports);
}

/* Setup range variables for cpu profiling */
static void
setup_parameter_ranges_cpu(void)
{
   long value;

   read_monitor_sysparam(SYSPARAM_PROF_CPU, "interval_max", &value,
            MAX_PROF_INTERVAL, MIN_PROF_INTERVAL, MAX_PROF_INTERVAL);
   rw.cpu_interval_range.max = value;

   read_monitor_sysparam(SYSPARAM_PROF_CPU, "interval_min", &value,
            MIN_PROF_INTERVAL, MIN_PROF_INTERVAL, MAX_PROF_INTERVAL);
   rw.cpu_interval_range.min = value;
   check_range("cpu profiling interval", &rw.cpu_interval_range);

   read_monitor_sysparam(SYSPARAM_PROF_CPU, "reports_max", &value,
            MAX_PROF_REPORTS, MIN_PROF_REPORTS, MAX_PROF_REPORTS);
   rw.cpu_reports_range.max = value;

   read_monitor_sysparam(SYSPARAM_PROF_CPU, "reports_min", &value,
            MIN_PROF_REPORTS, MIN_PROF_REPORTS, MAX_PROF_REPORTS);
   rw.cpu_reports_range.min = value;
   check_range("cpu profiling reports", &rw.cpu_reports_range);

   read_monitor_sysparam(SYSPARAM_PROF_CPU, "priority_max", &value,
            MAX_PROF_PRIORITY, MIN_PROF_PRIORITY, MAX_PROF_PRIORITY);
   rw.cpu_priority_range.max = value;

   read_monitor_sysparam(SYSPARAM_PROF_CPU, "priority_min", &value,
            MIN_PROF_PRIORITY, MIN_PROF_PRIORITY, MAX_PROF_PRIORITY);
   rw.cpu_priority_range.min = value;
   check_range("cpu profiling priority", &rw.cpu_priority_range);

   /* Check default value of cpu profiling interval */
   check_prof_def_value (rw.cpu_interval_range,
      (unsigned int *)&zzprof.cpu.default_cpu_interval);

   /* Check default value of cpu profiling reports */
   check_prof_def_value (rw.cpu_reports_range, &zzprof.cpu.default_cpu_reports);

   /* Check default value of cpu profiling priority */
   check_prof_def_value (rw.cpu_priority_range,
      &zzprof.cpu.default_cpu_priority);  
}

/* Setup parameter ranges for user profiling */
static void
setup_parameter_ranges_user(void)
{
   long value;

   read_monitor_sysparam(SYSPARAM_PROF_USER, "interval_max", &value,
            MAX_PROF_INTERVAL, MIN_PROF_INTERVAL, MAX_PROF_INTERVAL);
   rw.user_interval_range.max = value;

   read_monitor_sysparam(SYSPARAM_PROF_USER, "interval_min", &value,
            MIN_PROF_INTERVAL, MIN_PROF_INTERVAL, MAX_PROF_INTERVAL);
   rw.user_interval_range.min = value;
   check_range("user profiling interval", &rw.user_interval_range);

   read_monitor_sysparam(SYSPARAM_PROF_USER, "reports_max", &value,
            MAX_PROF_REPORTS, MIN_PROF_REPORTS, MAX_PROF_REPORTS);
   rw.user_reports_range.max = value;

   read_monitor_sysparam(SYSPARAM_PROF_USER, "reports_min", &value,
            MIN_PROF_REPORTS, MIN_PROF_REPORTS, MAX_PROF_REPORTS);
   rw.user_reports_range.min = value;
   check_range("user profiling reports", &rw.user_reports_range);

   read_monitor_sysparam(SYSPARAM_PROF_USER, "maxvalues_max", &value,
            MAX_PROF_VALUES, MIN_PROF_VALUES, MAX_PROF_VALUES);
   rw.user_maxvalues_range.max = value;

   read_monitor_sysparam(SYSPARAM_PROF_USER, "maxvalues_min", &value,
            MIN_PROF_VALUES, MIN_PROF_VALUES, MAX_PROF_VALUES);
   rw.user_maxvalues_range.min = value;
   check_range("user profiling maxvalues", &rw.user_maxvalues_range);

   /* Check default value of user profiling interval */
   check_prof_def_value (rw.user_interval_range,
      (unsigned int *)&rw.default_user_interval);

   /* Check default value of user profiling reports */
   check_prof_def_value (rw.user_reports_range, &rw.default_user_reports);

   /* Check default value of user profiling maxvalues */
   check_prof_def_value (rw.user_maxvalues_range, &rw.default_user_maxvalues);
}

/* Setup total cpu range paramters */
/*static void
 setup_parameter_ranges_total_cpu(void)
 {
 long value;

 read_monitor_sysparam(SYSPARAM_PROF_TOTAL_CPU_BLOCK, "interval_max",
 &value, MAX_PROF_INTERVAL, MIN_PROF_INTERVAL, MAX_PROF_INTERVAL);
 rw.total_cpu_interval_range.max = value;

 read_monitor_sysparam(SYSPARAM_PROF_TOTAL_CPU_BLOCK, "interval_min",
 &value, MIN_PROF_INTERVAL, MIN_PROF_INTERVAL, MAX_PROF_INTERVAL);
 rw.total_cpu_interval_range.min = value;

 read_monitor_sysparam(SYSPARAM_PROF_TOTAL_CPU_BLOCK, "reports_max",
 &value, MAX_PROF_REPORTS, MIN_PROF_REPORTS, MAX_PROF_REPORTS);
 rw.total_cpu_reports_range.max = value;

 read_monitor_sysparam(SYSPARAM_PROF_TOTAL_CPU_BLOCK, "reports_min",
 &value, MIN_PROF_REPORTS, MIN_PROF_REPORTS, MAX_PROF_REPORTS);
 rw.total_cpu_reports_range.min = value;

 read_monitor_sysparam(SYSPARAM_PROF_TOTAL_CPU_BLOCK, "priority_max",
 &value, MAX_PROF_PRIORITY, MIN_PROF_PRIORITY, MAX_PROF_PRIORITY);
 rw.total_cpu_priority_range.max = value;

 read_monitor_sysparam(SYSPARAM_PROF_TOTAL_CPU_BLOCK, "priority_min",
 &value, MIN_PROF_PRIORITY, MIN_PROF_PRIORITY, MAX_PROF_PRIORITY);
 rw.total_cpu_priority_range.min = value;

 read_monitor_sysparam(SYSPARAM_PROF_TOTAL_CPU_BLOCK, "procs_max",
 &value, zzkrn_get_pid_info(0, 4), MIN_PROF_MAXPROCS,
 zzkrn_get_pid_info(0, 4));
 rw.total_cpu_procs_range.max = value;

 read_monitor_sysparam(SYSPARAM_PROF_TOTAL_CPU_BLOCK, "procs_min",
 &value, MIN_PROF_MAXPROCS, MIN_PROF_MAXPROCS,
 zzkrn_get_pid_info(0, 4));
 rw.total_cpu_procs_range.min = value;

 read_monitor_sysparam(SYSPARAM_PROF_TOTAL_CPU_BLOCK, "oversize_max",
 &value, MAX_PROF_OVERSIZE, MIN_PROF_OVERSIZE, MAX_PROF_OVERSIZE);
 rw.total_cpu_oversize_range.max = value;

 read_monitor_sysparam(SYSPARAM_PROF_TOTAL_CPU_BLOCK, "oversize_min",
 &value, MAX_PROF_OVERSIZE, MIN_PROF_OVERSIZE, MAX_PROF_OVERSIZE);
 rw.total_cpu_oversize_range.min = value;

 }
 */

static void
setup_parameter_buffer_limit(void)
{
   long value;

   read_monitor_sysparam(SYSPARAM_PROF_BUFFER_LIMIT, "size", &value,
            DEF_PROF_BUFFER_LIMIT, MIN_PROF_BUFFER_LIMIT, MAX_PROF_BUFFER_LIMIT);
   rw.sample_pool_memory_limit = value;
}

/*
 * Creates the paramater range variables in rmm_world from sysparams
 * and the defined max/min values.
 */
static void
setup_parameter_ranges(void)
{
   setup_parameter_ranges_proc();
   setup_parameter_ranges_block();
   setup_parameter_ranges_pri();
   setup_parameter_ranges_cpu();
   setup_parameter_ranges_user();
   /*setup_parameter_ranges_total_cpu();*/
   setup_parameter_buffer_limit();
}

/* Helper function to facilitate comparison with the parameter_range type */
static OSBOOLEAN
in_range(uint32_t value, parameter_range range)
{
   return (value <= range.max) && (value >= range.min);
}

static void
configure_parameter(int param_idx)
{
   long value;

   switch(param_idx)
   {
      case SYSPARAM_SIGSIZE:
      {
         /*
         ** Get maximum multiple trace size that we can use from max_sigssize
         ** and it if it is larger than 64K we truncate it so it fits into the
         ** two precious bytes that are reserved for it.
         */
         uint32_t max_mget_size = get_pcb(current_process())->max_sigsize - 256;
         max_mget_size = max_mget_size <= 0xffff ? max_mget_size : 0xffff;

         /* Set trace read multiple size: trace_mget_size. */
         read_monitor_sysparam(param_idx, "trace_mget", &value,
                      rw.def_trace_mget_size, 512, max_mget_size); /* def, min, max */

         rw.trace_mget_size = value >> 2; /* trace_mget_size in units of 4 bytes */
         rw.precious->options[14] = (value & 0xff00) >> 8;
         rw.precious->options[15] = (value & 0xff);

         /* Set max_trace_size. */
         read_monitor_sysparam(param_idx, "trace", &value,
                      255, 4, max_mget_size); /* def, min, max */
         rw.max_trace_size = value;
         if (rw.trace_mget_size * 4 < rw.max_trace_size) /* mget_size in 4 byte units*/
         {
            ramlog_printf("rmm: trace_mget < trace. "
               "This may cause lost entries!\n");
         }
         if (rw.trace_size <= rw.max_trace_size )
         {
            ramlog_printf("rmm: tracesize <= trace. "
               "This may cause lost entries!\n");
         }
         rw.precious->options[0] = (value & 0xff00) >> 8;
         rw.precious->options[1] = (value & 0xff);

         /* Set max_monitor_size. */
         read_monitor_sysparam(param_idx, "notify", &value,
                      255, 4, max_mget_size); /* def, min, max */
         rw.max_monitor_size = value;
         rw.precious->options[2] = (value & 0xff00) >> 8;
         rw.precious->options[3] = (value & 0xff);

         /* Don't forget to update precious chksum since we've changed it */
         update_precious_checksum();

         break;
      }
      case SYSPARAM_ACTION:
         read_monitor_sysparam(param_idx, "state", &value,
                      0, 0, 255); /* def, min, max */
         rw.action_state = rw.precious->action_state = value;

         read_monitor_sysparam(param_idx, "sys_processes", &value,
                      0, 0, 1); /* def, min, max */
         rw.precious->options[11] = value & 0x1;
         rw.sys_process = rw.precious->options[11] ? 0 : SYS_PROCESS;
         update_precious_checksum();

         read_monitor_sysparam(param_idx, "dynamic_trace_size", &value,
                     0, 0, 0x7FFFFFFF); /* def, min, max */

         /* Set dynamic trace size */
         if (value > 0 && value < 8192)
         {
               value = 8192; /* Set to default: 8K */
         }
         rw.dynamic_trace_size = value;

         rw.precious->options[20] = (value & 0xff000000) >> 24;
         rw.precious->options[21] = (value & 0x00ff0000) >> 16;
         rw.precious->options[22] = (value & 0x0000ff00) >> 8;
         rw.precious->options[23] = (value & 0x000000ff);

#ifdef HAS_INT_SWAP_HOOK
         read_monitor_sysparam(param_idx, "enable_intproc_swap", &value, 0, 0, 1);
         rw.enable_intproc_swap = rw.precious->options[80] = (value & 0x1);
#endif
         break;

      case SYSPARAM_FLOW_CTRL:
         read_monitor_sysparam(param_idx, "block", &value,
                      20, 5, 255); /* def, min, max */
         rw.flow_ctrl_block = rw.precious->options[12] = (value & 0xff);

         read_monitor_sysparam(param_idx, "notify", &value,
                      10, 2, 255); /* def, min, max */
         rw.flow_ctrl_nblock = rw.precious->options[13] = (value & 0xff);
         update_precious_checksum();
         {
            Offspring *c;
            /* Make sure all offsprings has valid flow control values. */
            for (c = list_head(rw.offsprings); c != NULL; c = list_next(rw.offsprings, &c->link))
            {
               c->flow_ctrl_block = rw.flow_ctrl_block;
               c->flow_ctrl_nblock = rw.flow_ctrl_nblock;
            }
         }
         break;

      case SYSPARAM_PROFILE:
         read_monitor_sysparam(param_idx, "minsample", &value,
         							0,0,255);

         zzprof.prof_all_min_us = value;
         rw.precious->options[19] = value & 0xff;

         break;

      case SYSPARAM_PROF_PROC:
         /* Set up default parameters for per process profiling */
         read_monitor_sysparam(SYSPARAM_PROF_PROC, "interval", &value,
                     DEF_PROF_INTERVAL, MIN_PROF_INTERVAL, MAX_PROF_INTERVAL);

         zzprof.proc[0].default_proc_interval = value;

         read_monitor_sysparam(SYSPARAM_PROF_PROC, "reports", &value,
                     DEF_PROF_REPORTS, MIN_PROF_REPORTS, MAX_PROF_REPORTS);

         zzprof.proc[0].default_proc_reports = value;

         read_monitor_sysparam(SYSPARAM_PROF_PROC, "procs", &value,
                  DEF_PROF_MAX_VALUES, MIN_PROF_MAXPROCS, zzkrn_get_pid_info(0, 4)); /* sys maxprocs */

         zzprof.proc[0].default_proc_maxprocs = value;

         setup_parameter_ranges_proc();

         /* Save corrected values to precious structure */
         set_precious_uint32(&rw.precious->options[24], zzprof.proc[0].default_proc_interval);
         set_precious_uint32(&rw.precious->options[28], zzprof.proc[0].default_proc_reports);
         set_precious_uint32(&rw.precious->options[32], zzprof.proc[0].default_proc_maxprocs);

         break;

      case SYSPARAM_PROF_BLOCK:
         /* Set up default parameters for per block profiling */
         read_monitor_sysparam(SYSPARAM_PROF_BLOCK, "interval", &value,
                     DEF_PROF_INTERVAL, MIN_PROF_INTERVAL, MAX_PROF_INTERVAL);

         zzprof.proc[1].default_proc_interval = value;


         read_monitor_sysparam(SYSPARAM_PROF_BLOCK, "reports", &value,
                     DEF_PROF_REPORTS, MIN_PROF_REPORTS, MAX_PROF_REPORTS);

         zzprof.proc[1].default_proc_reports = value;

         read_monitor_sysparam(SYSPARAM_PROF_BLOCK, "procs", &value,
                     DEF_PROF_MAX_VALUES, MIN_PROF_MAXPROCS, zzkrn_get_pid_info(0, 4)); /* sys maxprocs */

         zzprof.proc[1].default_proc_maxprocs = value;

         setup_parameter_ranges_block();

         /* Save corrected values to precious structure */
         set_precious_uint32(&rw.precious->options[36], zzprof.proc[1].default_proc_interval);
         set_precious_uint32(&rw.precious->options[40], zzprof.proc[1].default_proc_reports);
         set_precious_uint32(&rw.precious->options[44], zzprof.proc[1].default_proc_maxprocs);
         break;

      case SYSPARAM_PROF_PRI:
         /* Set up default parameters for priority profiling */
         read_monitor_sysparam(SYSPARAM_PROF_PRI, "interval", &value,
                     DEF_PROF_INTERVAL, MIN_PROF_INTERVAL, MAX_PROF_INTERVAL);

         zzprof.pri.default_pri_interval = value;

         read_monitor_sysparam(SYSPARAM_PROF_PRI, "reports", &value,
                     DEF_PROF_REPORTS, MIN_PROF_REPORTS, MAX_PROF_REPORTS);

         zzprof.pri.default_pri_reports = value;

         setup_parameter_ranges_pri();

         /* Save corrected values to precious structure */
         set_precious_uint32(&rw.precious->options[48], zzprof.pri.default_pri_interval);
         set_precious_uint32(&rw.precious->options[52], zzprof.pri.default_pri_reports);
         break;

      case SYSPARAM_PROF_CPU:
         /* Set up default parameters for cpu profiling */
         read_monitor_sysparam(SYSPARAM_PROF_CPU, "interval", &value,
                     DEF_PROF_INTERVAL, MIN_PROF_INTERVAL, MAX_PROF_INTERVAL);

         zzprof.cpu.default_cpu_interval = value;

         read_monitor_sysparam(SYSPARAM_PROF_CPU, "reports", &value,
                     DEF_PROF_REPORTS, MIN_PROF_REPORTS, MAX_PROF_REPORTS);

         zzprof.cpu.default_cpu_reports = value;

         read_monitor_sysparam(SYSPARAM_PROF_CPU, "priority", &value,
                     DEF_PROF_PRIORITY, MIN_PROF_PRIORITY, MAX_PROF_PRIORITY);

         zzprof.cpu.default_cpu_priority = value;

         setup_parameter_ranges_cpu();

         /* Save corrected values to precious structure */
         set_precious_uint32(&rw.precious->options[56], zzprof.cpu.default_cpu_interval);
         set_precious_uint32(&rw.precious->options[60], zzprof.cpu.default_cpu_reports);
         set_precious_uint32(&rw.precious->options[64], zzprof.cpu.default_cpu_priority);
         break;

         /************************* ADD new params */
      case SYSPARAM_PROF_TOTAL_CPU_BLOCK:
          /* Set up default parameters for cpu profiling */
          read_monitor_sysparam(SYSPARAM_PROF_TOTAL_CPU_BLOCK, "interval", &value,
                      DEF_PROF_INTERVAL, MIN_PROF_INTERVAL, MAX_PROF_INTERVAL);
          zzprof.tcb.default_interval = value;
          /*    where in precious ???
           * set_precious_uint32(&rw.precious->options[??], zzprof.tcb.default_interval);
           */

          read_monitor_sysparam(SYSPARAM_PROF_TOTAL_CPU_BLOCK, "reports", &value,
                      DEF_PROF_REPORTS, MIN_PROF_REPORTS, MAX_PROF_REPORTS);
          zzprof.tcb.default_reports = value;
             /* where in
              * set_precious_uint32(&rw.precious->options[??], zzprof.tcb.default_reports);
              */
          read_monitor_sysparam(SYSPARAM_PROF_TOTAL_CPU_BLOCK, "priority", &value,
                      DEF_PROF_PRIORITY, MIN_PROF_PRIORITY, MAX_PROF_PRIORITY);

          zzprof.tcb.default_priority = value;
          /* where in precious??
           * set_precious_uint32(&rw.precious->options[64], zzprof.tcb.default_priority);
          */

          read_monitor_sysparam(SYSPARAM_PROF_TOTAL_CPU_BLOCK, "procs", &value,
                   DEF_PROF_MAX_VALUES, MIN_PROF_MAXPROCS, zzkrn_get_pid_info(0, 4));

          zzprof.tcb.default_maxblocks = value;
          /* where in precious??
           * set_precious_uint32(&rw.precious->options[64], zzprof.tcb.default_priority);
           */
          read_monitor_sysparam(SYSPARAM_PROF_TOTAL_CPU_BLOCK, "oversize", &value,
                                DEF_PROF_OVERSIZE, MIN_PROF_OVERSIZE, MAX_PROF_OVERSIZE);

                    zzprof.tcb.default_oversize = value;
          /* where in precious??
           * set_precious_uint32(&rw.precious->options[64], zzprof.tcb.default_priority);
           */

          break;
/************************************* end new */

      case SYSPARAM_PROF_BUFFER_LIMIT:
         setup_parameter_buffer_limit();
         break;

      case SYSPARAM_PROF_USER:
         /* Set up default parameters for user report profiling */
         read_monitor_sysparam(SYSPARAM_PROF_USER, "interval", &value,
                     DEF_PROF_INTERVAL, MIN_PROF_INTERVAL, MAX_PROF_INTERVAL);
         rw.default_user_interval = value;

         read_monitor_sysparam(SYSPARAM_PROF_USER, "reports", &value,
                     DEF_PROF_REPORTS, MIN_PROF_REPORTS, MAX_PROF_REPORTS);
         rw.default_user_reports = value;

         read_monitor_sysparam(SYSPARAM_PROF_USER, "maxvalues", &value,
                     DEF_PROF_MAX_VALUES, MIN_PROF_VALUES, MAX_PROF_VALUES);
         rw.default_user_maxvalues = value;

         setup_parameter_ranges_user();

         /* Save corrected values to precious structure */
         set_precious_uint32(&rw.precious->options[68], rw.default_user_interval);
         set_precious_uint32(&rw.precious->options[72], rw.default_user_reports);
         set_precious_uint32(&rw.precious->options[76], rw.default_user_maxvalues);
         break;

      case SYSPARAM_INTERFACE:
         read_monitor_sysparam(param_idx, "cripple", &value, 0, 0, 1); /* def, min, max */
         if (value)
         {
            /* Mark all request signals as disabled */
            memset(rw.req_disabled_bitmask, 0xFF, sizeof rw.req_disabled_bitmask);
            /* ...but MONITOR_INTERFACE_REQUEST must always be available */
            rw.req_disabled_bitmask[0] &= ~1;
         }
         break;
      case SYSPARAM_SYSTEMNAME:
         if (sysparam_read(rmm_sysparams[SYSPARAM_SYSTEMNAME],
                          (char *)system_name,
                          MAX_NAME_LEN + 1) == 0)
         {
            strcpy(system_name, "");
         }
         break;
       
      case SYSPARAM_SYS_ERROR:
         read_monitor_sysparam(param_idx, "debug", &value, 0, 0, 1);
         rw.monitor_sys_error = value;
         rw.precious->options[16] = (uint8_t) (value & 0xFF);
         break;

      case SYSPARAM_ERROR:
         read_monitor_sysparam(param_idx, "trap_user_handled", &value, 1, 0, 1);
         rw.monitor_error = value;
         rw.precious->options[17] = (uint8_t) (value & 0xFF);
         break;
 
      default:
         ramlog_printf("rmm: internal error: unknown monitor parameter: %d\n",
                       param_idx);
         assert(0);
         break;
   }
}


/*
** Subfunction to the LDM.
**
** It is called during the system start procedure to create the
** ldm variable set.
**
** The global precious pointer is set, as well as global copies of the
** variable set.
*/
static void
create_precious(void)
{
   int             i;
   int             opt_count = sizeof (rw.precious->options) / sizeof (rw.precious->options[0]);

   rw.precious->evact_count = 0;
   /* Clear all pointers. */
   rw.precious->debugger_scope = NULL;
   rw.precious->breakpoint_scope = NULL;
   rw.precious->action_list_create = NULL;
   rw.precious->action_list_kill = NULL;
   rw.precious->action_list_swap = NULL;
   rw.precious->action_list_send = NULL;
   rw.precious->action_list_receive = NULL;
   rw.precious->action_list_oserror = NULL;
   rw.precious->action_list_error = NULL;
   rw.precious->action_list_restart = NULL;
   rw.precious->action_list_uevent = NULL;
#ifdef OSE_SMP
   rw.precious->action_list_bind = NULL;
#endif
   rw.precious->action_list_alloc = NULL;
   rw.precious->action_list_free_buf = NULL;
   rw.precious->action_list_timeout = NULL;
   rw.precious->not_used_action_list1 = NULL;
   rw.precious->trace_enabled = (OSBOOLEAN) 0;
   rw.precious->master_monitor_enabled = (OSBOOLEAN) 0;

   /* Clear all option fields. */
   for (i = 0; i < opt_count; i++)
   {
      rw.precious->options[i] = 0;
   }

   configure_parameter(SYSPARAM_SIGSIZE);
   configure_parameter(SYSPARAM_ACTION);
   configure_parameter(SYSPARAM_FLOW_CTRL);
   configure_parameter(SYSPARAM_PROFILE);
   configure_parameter(SYSPARAM_PROF_PROC);
   configure_parameter(SYSPARAM_PROF_BLOCK);
   configure_parameter(SYSPARAM_PROF_PRI);
   configure_parameter(SYSPARAM_PROF_CPU);
   configure_parameter(SYSPARAM_PROF_USER);
   configure_parameter(SYSPARAM_PROF_TOTAL_CPU_BLOCK);
   configure_parameter(SYSPARAM_PROF_BUFFER_LIMIT);
   configure_parameter(SYSPARAM_SYS_ERROR);
   configure_parameter(SYSPARAM_ERROR);

   /* Set reconnect_timeout. */
   rw.precious->options[4] = 0;
   rw.precious->options[5] = 120;

   /* Set the clear_event_count_on_restart option to "yes". */
   rw.precious->options[6] = 1;

   /* Set the display_cpuregs option to "no". */
   rw.precious->options[7] = 0;

   /* Set the clear_sequence_on_restart option to "yes". */
   rw.precious->options[9] = 1;

   /* Set the break_scope_status flag to "running". */
   rw.precious->break_scope_status = 0;

   /* Clear unused flags for future use. */
   rw.precious->pad1 = 0;
   rw.precious->pad2 = 0;
   rw.precious->pad3 = 0;
   rw.precious->pad4 = 0;
   rw.precious->pad5 = 0;

   /* Read the current error check level on cold start. */
   rw.precious->options[10] = (uint8_t) get_sysconf(3, (OSADDRESS) 1, 0, 0);
   get_sysconf(3, (OSADDRESS) rw.precious->options[10], 0, 0);

   /* Set the magic number last to avoid restart races. */
   rw.precious->magic = PRECIOUS_MAGIC;
   rw.precious->magic_tail = PRECIOUS_MAGIC;

   /* Update the precious checksum */
   update_precious_checksum();

   /* Double check the trace to avoid bugs and set up static vars. */
   if (recover_precious())
      error(OSE_RMM_ECORRUPTED_TRACE_MEMORY);
   else
      rw.precious->magic = PRECIOUS_MAGIC; /* destroyed by recover_precious. */
}

/*
** Copy list pointers to precious structure.
*/
static void
update_precious_lists(void)
{
   struct action_lists *actions = &rw.actions SMP_CODE([0]);

   rw.precious->action_list_create = actions->action_list_create;
   rw.precious->action_list_kill = actions->action_list_kill;
   rw.precious->action_list_swap = actions->action_list_swap;
   rw.precious->action_list_send = actions->action_list_send;
   rw.precious->action_list_receive = actions->action_list_receive;
   rw.precious->action_list_oserror = actions->action_list_oserror;
   rw.precious->action_list_error = actions->action_list_error;
   rw.precious->action_list_restart = actions->action_list_restart;
   rw.precious->action_list_uevent = actions->action_list_uevent;
#ifdef OSE_SMP
   rw.precious->action_list_bind = actions->action_list_bind;
#endif
   rw.precious->action_list_alloc = actions->action_list_alloc;
   rw.precious->action_list_free_buf = actions->action_list_free_buf;
#ifdef HAS_TIMEOUT_EVENT
   rw.precious->action_list_timeout = actions->action_list_timeout;
#endif

   /* Update the checksum */
   update_precious_checksum();
}

/*
**===========================================================================
** System calls. FIXME, they shouldn't be here.
**===========================================================================
*/

/*
** This is the set_abstime() system call.
**
** It is used to make the last known absolute time available to the kernel.
** The call sets a reference point in time, consisting of an absolute time
** and the corresponding kernel relative time obtained from get_systime().
**
** abs_time contains the current time expressed as a number of seconds
** after 00:00:00 1-jan 1970 GMT. (traditional unix format.)
** ticks and microsecs contain the values obtained from get_systime() at
** the time corresponding to abs_time.
*/

void
set_abstime(OSTICK abs_time, OSTICK ticks, OSTICK microsecs)
{
   LOCKSTATE       msr;
   BKL_LOCK_SAVE(msr);

   /* Increment variable so it is odd (bit 0 set). That notes that the time
    * is under update. This is used to implement a lock free read.
    */
   SMP_CODE(rw.time_update_count++;)
   rw.abs_sec = (U32) abs_time;
   rw.abs_tick = (U32) ticks;
   rw.abs_usec = (U32) microsecs;
   /* Update the counter to it is even */
   SMP_CODE(rw.time_update_count++;) /* Even value */

   if (rw.trace_abstime)
   {
      ose_trace_commit_inline(rw.trace_abstime);
      rw.trace_abstime = NULL;
   }
   BKL_LOCK_RESTORE( msr);
}

/*
** This is the get_abstime() system call.
**
** It returns the last time set with set_abstime).
**
** abs_time contains the current time expressed as a number of seconds
** after 00:00:00 1-jan 1970 GMT. (traditional unix format.)
** ticks and microsecs contain the values obtained from get_systime() at
** the time corresponding to abs_time.
**
** If abs_time is zero, then the time was never set.
*/

void
get_abstime(OSTICK * abs_time, OSTICK * ticks, OSTICK * microsecs)
{
   LOCKSTATE       msr;
   SMP_CODE(uint32_t        count;)

   LOCK_SAVE(msr);
   SMP_CODE(do)
   {
      SMP_CODE(count = rw.time_update_count;)
      *abs_time = (OSTICK) rw.abs_sec;
      *ticks = (OSTICK) rw.abs_tick;
      *microsecs = (OSTICK) rw.abs_usec;
      /* For SMP loop until we have read the time without it has been updated
       * by another core. */
   } SMP_CODE(while (count & 1 || count != rw.time_update_count);)
   LOCK_RESTORE(msr);
}

/*
**===========================================================================
** And finally, RMM functions!
**===========================================================================
*/

static void
assert_offspring(const Offspring * offspring)
{
   (void)offspring;
   assert(offspring);
   assert(CLIENT_MAGIC == offspring->magic_head);
   assert(offspring->pid > 0);
   assert(CLIENT_MAGIC == offspring->magic_tail);
}

/*
** utility to copy a scope string set from an input signal to
** a buffer allocated from the trace space.
**
** The string list contains packed, NUL-terminated strings.
**
** Returns a pointer to the new buffer.
** Returns NULL if no space in the trace buffer.
*/

static struct scope *
copy_scope(int scope_count, const char *scope_strings)
{
   int             count;
   struct scope   *scope;
   EntrySize       size;
   const char     *p;
   char           *q;

   /* Calculate the required size of all strings. */
   for (size = 0, count = scope_count, p = scope_strings; count; count--)
   {
      while (*p++)
         size++;
      size++;
   }

   /* @frbr@ Is this sanity-check correct? */
   if ((sizeof (struct scope) + size) > (EntrySize) ~ 0U)
      return NULL;

   /* Allocate a suitable scope buffer from the trace space. */
   scope = sys_alloc_trace(rw.static_trace, (EntrySize) (sizeof (struct scope) + size));

   /* Terminate if the trace is full. */
   if (!scope)
      return scope;

   /* Copy scope strings to the new buffer. */
   scope->count = (U16) scope_count;
   for (count = scope_count, p = scope_strings, q = scope->scope_list; count; count--)
   {
      while ((*q++ = *p++))
      {
         ;                      /* Empty loop body. */
      }
   }

   /* Return the new buffer. */
   return scope;
}

/*
** Utility function to traverse all processes and apply a function to
** each one of them.
**
** It is called in various places to update scopes and action lists.
*/
static void
for_all_processes(void (*apply) (PROCESS pid, U32 misc), U32 data)
{
   struct OS_pid_list *bid_list;
   OSBUFSIZE       bid_count;
   PROCESS        *bid;

   struct OS_pid_list *pid_list;
   OSBUFSIZE       pid_count;
   PROCESS        *pid;

   /* Get all available block id's. */
   if (!(bid_list = get_bid_list((OSUSER) 0)))
      return;

   bid_count = bid_list->count;
   bid = bid_list->list;

   /* For all blocks... */
   for (; bid_count; bid_count--, bid++)
   {
      /* Get all available process id's. */
      if (!(pid_list = get_pid_list(*bid)))
         continue;

      pid_count = pid_list->count;
      pid = pid_list->list;

      /* For all processes... */
      for (; pid_count; pid_count--, pid++)
      {
         /* Apply the operation to the process. */
         (*apply) (*pid, data);
      }

      /* Free the process id list. */
      free_buf((union SIGNAL **) &pid_list);
   }
   /* Free the block id list. */
   free_buf((union SIGNAL **) &bid_list);
}


/*
** utility to match a process/block pid to a set of wildcard names.
** The string list contains packed, NUL-terminated strings.
** Only wildcard strings on the "#0x...." form are considered.
** If the list is empty, a match is assumed. This feature is used when
** the client sends us empty lists in for instance "display process".
** Returns 0 if no match was found.
** Returns 1 if a match was found.
*/

static int
match_pid(PROCESS pid, int wildcard_count, char *wildcard_strings)
{
   uint32_t        digit;
   char           *w = wildcard_strings;

   /* Done if no strings to check. Return a match in this case. */
   if (!wildcard_count)
      return 1;

   /* Check all wildcard strings. */
   while (wildcard_count--)
   {
      /* Skip strings that doesn't start with "#". */
      if (*w == '#')
      {
         /* Skip syntax errors. */
         if (!strtonum(++w, &digit))
         {
            /* Check one string. */
            if (pid == (PROCESS) digit)
               return 1;
         }
      }

      /* Skip to the next string. */
      while (*w++)
      {;                        /* Empty loop body. */
      }
   }

   /* return false if not found. */
   return 0;
}

/*
** utility to match a process/block name to a set of wildcard names.
** The string list contains packed, NUL-terminated strings.
** If the list is empty, a match is assumed. This feature is used when
** the client sends us empty lists in for instance "display process".
** Returns 0 if no match was found.
** Returns 1 if a match was found.
*/

static int
match_wildcard(char *name, int wildcard_count, char *wildcard_strings)
{
   char           *w = wildcard_strings;

   /* Done if no strings to check. Return a match in this case. */
   if (!wildcard_count)
      return 1;

   /* Check all wildcard strings. */
   while (wildcard_count--)
   {
      /* Check one wildcard string. */
      if (!estrcmp(w, name))
         return 1;

      /* Skip to the next string. */
      while (*w++)
      {;                        /* Empty loop body. */
      }
   }

   /* return false if not found. */
   return 0;
}


static struct ldefine_action *
alloc_action(uint16_t size)
{
   struct ldefine_action *action = NULL;

   if (likely(size >= sizeof(struct ldefine_action)))
   {
      action = (struct ldefine_action *)sys_alloc_trace(rw.static_trace, size);
      if (likely(action))
      {
         memset(action, 0, size);
         action->magic = ACTION_MAGIC_ALLOCATED;
         action->orig_action = action;
         action->size = size;
         SMP_CODE(spin_lock_init(&action->action_lock));
      }
      else
      {
         ramlog_printf("rmm: failed to allocate trace memory for an action size=%u\n", size);
      }
   }
   else
   {
      ramlog_printf("rmm: internal error; too small buffer for an action requested\n");
   }

   return action;
}

static void
free_action(struct ldefine_action *action)
{
   if (likely(action && (action->magic == ACTION_MAGIC_ALLOCATED || action->magic == ACTION_MAGIC_VALID)))
   {
#ifdef OSE_SMP
      LOCK();
      SPIN_LOCK(&action->action_lock);
#endif
      action->magic = ACTION_MAGIC_INVALID;
      action->orig_action = NULL;
      action->next = NULL;
      action->prev = NULL;
      SPIN_UNLOCK(&action->action_lock);
      ose_trace_commit_inline(action);
#ifdef OSE_SMP
      UNLOCK();
#endif
   }
   else
   {
      ramlog_printf("rmm: internal error; free of corrupt action point: 0x%x magic %x\n",
                    action, action? action->magic: 0);
      if (action &&
	  (((uint32_t)action) & 0xf) == 0 &&
	  ((char *)action) > ((char *)rw.static_trace) &&
	  ((char *)action) < (char *)(rw.static_trace + rw.static_trace_size))
      {
	 ose_trace_commit_inline(action);
      }
   }
}

/*
** Subfunction to update_hook_bits().
*/

static void
apply_new_flags(PROCESS pid, U32 flags)
{
   struct debug_pcb *deb_pcb;

   /* Get the debugger control block for the process. */
   BKL_LOCK();
   deb_pcb = zzget_debugpcb(pid);

   /* Check that the process is alive and in the display scope. */
   if (deb_pcb)
      if (deb_pcb->debug_flags & IN_DEBUGGER_SCOPE)
      {
         /* Replace the flags that select debugger hooks. */
         deb_pcb->debug_flags = (deb_pcb->debug_flags & ~(0x7fL)) | flags;
      }
   BKL_UNLOCK();
}

/*
** Utility to calculate the current set of debugger hook enable bits
** in each pcb in the display scope.
**
** Only bits 0-6 are calculated. Other bits reflect scope information.
*/

static          uint32_t
current_hook_flags(void)
{
   uint32_t        new_flags = (U32) 0;
   struct action_lists *actions = &rw.actions SMP_CODE([0]);

   /* First calculate the new pcb flag word. */
   if (actions->action_list_swap)
      new_flags |= 0x4;         /* We use only the swap in hook. */
   if (actions->action_list_send)
      new_flags |= 0x1;
   if (actions->action_list_receive)
      new_flags |= 0x2;
   if (actions->action_list_oserror)
      new_flags |= 0x20;
   if (actions->action_list_error)
      new_flags |= 0x10;
#ifdef HAS_TIMEOUT_EVENT
   if (actions->action_list_timeout)
      new_flags |= 0x80;
#endif
#ifdef OSE_SMP
   if (actions->action_list_bind)
      new_flags |= 0x40;
#endif

   /* Return the result. */
   return new_flags;
}

static void
update_int_proc_hook(void)
{

	int hook_bits = 0;
   int prof_enabled = ((!zzprof.cpu.prof_cpu_issampled) && zzprof.cpu.prof_cpu_enabled) ||
      (zzprof.pri.prof_pri_enabled | zzprof.proc[0].prof_proc_enabled | zzprof.proc[1].prof_proc_enabled | zzprof.tcb.enabled );


#ifdef HAS_INT_SWAP_HOOK
	int int_swap_enabled = rw.actions SMP_CODE([0]).action_list_swap && rw.enable_intproc_swap;
#endif

	if (prof_enabled) hook_bits |= 1;
#ifdef HAS_INT_SWAP_HOOK
	if (int_swap_enabled) hook_bits |= 4;
#endif

	zzset_inthook_ldmstatus(hook_bits);
}

static void
update_tick_hook(void)
{
   unsigned long int hook_value;

	if (rw.userReportEnabledCounter)
	{
    	hook_value = 1;
   }
   else
   {
      hook_value = 0;
   }
   zzset_tickhook_ldmstatus(hook_value);
}

/*
** Utility function to update the debugger hook enable bits in all pcb's
** in the display scope.
** The global debugger enable flag is also updated accordingly.
**
** This function is called whenever an action list is updated so that
** it changes state between empty and non empty.
*/

static void
update_hook_bits(void)
{
   U32             new_flags;

   /* First calculate the new pcb flag word. */
   new_flags = current_hook_flags();

   /* If it is zero then disable all debugger hooks. */
   if (!new_flags)
      zzset_debug_status((OSBOOLEAN) 0);

   /* Update all pcb's with the new flag set. */
   for_all_processes(apply_new_flags, new_flags);

   /* If the new pcb flag word != 0 then enable all debugger hooks. */
   if (new_flags)
      zzset_debug_status((OSBOOLEAN) 1);

   update_int_proc_hook();
}


/*
** This function allocates verbose wildcard lists for action list entries.
** These pointers are stored in the align and ecode fields of the
** ldefine_action signal buffers.
**
** If the allow_error flag is set, then errors are handled by fixing
** the structures just enough to avoid crashes. This flag is used
** during the warm start procedure when there is nowhere to report
** errors.
**
** Returns zero if all is well.
** Returns nonzero if there was an error and no structures were allocated.
*/
static int
update_expanded_wildcards(struct ldefine_action *action, int allow_error)
{
   struct procnamelist *align = NULL;
   struct procnamelist *ecode = NULL;

   /* All event types except warm starts have at least one wildcard string. */
   switch (action->event_type)
   {
      case 1:
      case 2:
      case 3:
      case 4:
      case 5:
      case 6:
      case 11:
#ifdef HAS_TIMEOUT_EVENT
      case MONITOR_EVENT_TIMEOUT:
#endif
      case MONITOR_EVENT_USER:
      case MONITOR_EVENT_ALLOC:
      case MONITOR_EVENT_FREE_BUF:
      {
         /* Calculate the number of strings. */
         U32             strcnt = 0;
         char           *s = &action->names[action->sig_from];

         while (*s)
         {
            strcnt++;
            while (*s++)
            {;                  /* Empty loop body. */
            }
         }
         if (strcnt)
         {
            /* Try to parse the wildcard list. */
            align = wildcard_to_plist(strcnt, &action->names[action->sig_from]);
         }
         /* Failed! Abort or allocate a simple emergency entry. */
         if (!align)
         {
            if (allow_error)
               align = wildcard_to_plist(1, "*");
            else
               goto abort;
         }
      }
      default:
         break;
   }

   /* Allocate a second wildcard structure for swap, send and receive events. */
   switch (action->event_type)
   {
         /* The swap event sometimes have a second process name wildcard. */
      case 3:
      {
         if (!action->sig_to)
            break;
         /* Fall into bind/send/receive handling. */
      }
#ifdef OSE_SMP
         /* The bind event only has the "to" process name wildcard. */
      case 14:
#endif /* OSE_SMP */
         /* send and receive events always have a second process name wildcard. */
      case 4:
      case 5:
      {
         /* Calculate the number of strings. */
         U32             strcnt = 0;
         char           *s = &action->names[action->sig_to];

         while (*s)
         {
            strcnt++;
            while (*s++)
            {;                  /* Empty loop body. */
            }
         }
         if (strcnt)
         {
            /* Try to parse the wildcard list. */
            ecode = wildcard_to_plist(strcnt, &action->names[action->sig_to]);
         }
         /* Failed! Abort or allocate a simple emergency entry. */
         if (!ecode)
         {
            if (allow_error)
               ecode = wildcard_to_plist(1, "*");
            else
               goto abort;
         }
      }
      default:
         break;
   }

   /* Everything is OK. Now update the action list entry. */
   if (align)
      action->align = (U32) align;
   if (ecode)
      action->ecode = (U32) ecode;
   return 0;

   /* No, there was a syntax error. Free buffers and return an error flag. */
 abort:
   if (align)
      free_buf((union SIGNAL **) &align);
   if (ecode)
      free_buf((union SIGNAL **) &ecode);
   return 1;
}

/*
** This function frees the verbose wildcard expansion buffers for
** an action list entry.
** This is done whenever an action list entry is deleted.
*/
static void
free_expanded_wildcards(struct ldefine_action *action)
{
   /* All event types except warm starts have at least one wildcard string. */
   switch (action->event_type)
   {
      case 1:
      case 2:
      case 3:
      case 4:
      case 5:
      case 6:
      case 11:
      case 13:
      {
         if (action->align)
            free_buf((union SIGNAL **) &action->align);
         break;
      }
      default:
         break;
   }

   /* There is a second wildcard structure for swap, send and receive events. */
   switch (action->event_type)
   {
         /* The swap event sometimes have only one name wildcard. */
      case 3:
      {
         if (!action->sig_to)
            break;
         /* Fall into bind/send/receive handling. */
      }
#ifdef OSE_SMP
         /* The bind event only has the "to" process name wildcard. */
      case 14:
#endif /* OSE_SMP */
         /* send and receive events always have a second process name wildcard. */
      case 4:
      case 5:
      {
         if (action->ecode)
            free_buf((union SIGNAL **) &action->ecode);
         break;
      }
      default:
         break;
   }
}

/*
** Clean User Event list from triggers that are inactivated!
**
*/

static void
clean_user_event_list(void)
{
   blocks_user_events *user_posts;
   struct ldefine_action *action;
   uint8_t i;
   OSBOOLEAN match;
   uint32_t error_val;
   static uint32_t trigValue = 0;

   if (rw.user_events)
   {
      for (user_posts = list_head(rw.user_events); user_posts != NULL ;
           user_posts = list_next(rw.user_events, &user_posts->link))
      {
         for (i=0; i < EVENTS_PER_BLOCK; i++)
         {
            if (0 != user_posts->user_event_list[i].trig)
            {
               action = rw.actions SMP_CODE([0]).action_list_uevent;
               match = FALSE;

               while ((NULL != action) && (!(match)))
               {
                  if (action->signo <= user_posts->user_event_list[i].event_no &&
                      action->signo_end >= user_posts->user_event_list[i].event_no)
                  {
                     match = TRUE;
                  }
                  action = action->next;
               }

               if (!match)
               {
                  error_val = set_mem(user_posts->sid, &trigValue,
                                      (OSADDRESS)user_posts->user_event_list[i].trig,
                                      sizeof(trigValue));
                  if (error_val)
                  {
                     ramlog_printf("rmm:clean_user_event mem_set ERROR!\n");
                  }
                  else
                  {
                     user_posts->user_event_list[i].trig_value = 0;
                  }
               }
            }
         }
      }
   }
}

#ifdef OSE_SMP

/*
 * This function copies the action point to all other cores on an SMP.
 *
 * Each core has its own action lists and by that we can avoid any big locks
 * in the event hooks.
 * During OSE initialization in SMP systems, all but core 0 are locked so
 * it is not possible nor needed to migrate the rmm process when we distribute
 * actionpoints at initialization.
 */
static void
smp_distribute_actionpoint(struct ldefine_action *action, 
                           struct ldefine_action **action_list, int rmm_init)
{
   cpuid_t cpuid;
   struct ldefine_action *action_copy;
   unsigned long offset = action_list - &rw.actions[0].action_list_create;

   for (cpuid = 1; cpuid < zzkrn_num_cpus(); cpuid++)
   {
      struct ldefine_action **list = &rw.actions [cpuid].action_list_create + offset;
      if (!rmm_init)
      {
         ose_bind_process(current_process(), cpuid);
      }
      action_copy = (struct ldefine_action *)alloc(action->size, 0);
      memcpy(action_copy, action, action->size);

      action_copy->next = NULL;

      LOCK();
      if (*list)
      {
         struct ldefine_action *prev;

         for (prev = *list; prev->next; prev = prev->next)
         {
            ;                     /* Empty loop body. */ 
         }
         prev->next = action_copy;
         action->prev = prev;
      }
      else
      {
         *list = action_copy;
         action_copy->prev = NULL;
      }
      UNLOCK();

   }
   /* Take us back to core 0. */
   if (!rmm_init)
   {
      ose_bind_process(current_process(), 0);
   }
}

/*
 * During OSE initialization in SMP systems, all but core 0 are locked so
 * it is not possible nor needed to migrate the rmm process when we remove
 * actionpoints at initialization.
 */
static void
smp_delete_actionpoint(uint32_t action_id, struct ldefine_action **action_list,
                       int rmm_init)
{
   cpuid_t cpuid;
   struct ldefine_action *action;
   unsigned long offset = action_list - &rw.actions[0].action_list_create;

   for (cpuid = zzkrn_num_cpus() - 1; cpuid > 0; cpuid--)
   {
      struct ldefine_action **list = (&rw.actions [cpuid].action_list_create) + offset;

      if (!rmm_init)
      {
         ose_bind_process(current_process(), cpuid);
      }

      LOCK();

      for (action = *list; action; action = action->next)
      {
         if (action->magic != ACTION_MAGIC_VALID)
         {
            ramlog_printf("rmm: error; illegal magic value (%x) for action at: 0x%x\n",
                  action->magic, action);
            ramlog_printf("rmm: error; id %u action %u next 0x%x prev 0x%x\n",
                  action->action_id, action->action, action->next, action->prev);
            UNLOCK();
            error(OSE_RMM_ECORRUPTED_MEMORY);
            if (!rmm_init)
            {
               ose_bind_process(current_process(), 0);
            }
            return;
         }
         if (action->action_id == action_id)
         {
            break;
         }
      }
      if (action)
      {
         if (action->prev)
         {
            action->prev->next = action->next;
            if (action->next)
            {
               action->next->prev = action->prev;
            }
         }
         else
         {
            *list = action->next;
            if (action->next)
            {
               action->next->prev = NULL;
            }
         }
         action->magic = ACTION_MAGIC_INVALID;
         UNLOCK();
         action->next = NULL;
         action->prev = NULL;
         free_buf((union SIGNAL **)&action);
      }
      else
      {
         UNLOCK();
      }
   }
   /* Take us back to core 0. */
   if (!rmm_init)
   {
      ose_bind_process(current_process(), 0);
   }
}

#endif

/*
** Subfunction to do_delete_action().
** Attempt to delete one action entry from one of the action lists.
**
** Return zero if the entry was not found in the list.
** Return 1 if the entry was found in the list.
** Return 2 if the entry was found and the list is now empty => debug hooks need update.
*/

static int
do_delete_one_action(uint32_t id, struct ldefine_action **action_list, int rmm_init)
{
   struct ldefine_action *action;
   U8              event_type;

   /* Locate the entry that should be deleted. */
   for (action = *action_list; action; action = action->next)
   {
      if (id == action->action_id)
      {
         /* Remember the event_type if create or kill. */
         event_type = action->event_type;

         SMP_CODE(smp_delete_actionpoint(id, action_list, rmm_init);)

         /* Delete the entry. */
         BKL_LOCK();
         if (action->prev)
         {
            action->prev->next = action->next;
         }
         else
         {
            /* Element is first in list */
            assert(*action_list == action);
            *action_list = action->next;
         }
         if (action->next)
         {
            action->next->prev = action->prev;
         }
         BKL_UNLOCK();

         /* Free the wildcard expansion buffers. */
         free_expanded_wildcards(action);

         /* Free the deleted buffer. */
         free_action(action);

         /* Copy list pointers to precious structure. */
         update_precious_lists();
         /* Return a success flag. */
         if (*action_list || event_type <= 2)
         {
            return 1;           /* No update needed if there still are actions or
                                 * the removed event is create or kill. */
         }
         else
         {
            return 2;
         }
      }
   }

   /* The entry was not found. */
   return 0;
}

/*
** Check if we should enable a swap trace optimization directly in the
** dispatcher of pose.c. This is only possible if we are to trace every single
** context switch in the system.
*/
static void
check_enable_swap_trace_opt(Offspring *offspring)
{
   struct ldefine_action *action = rw.actions SMP_CODE([0]).action_list_swap;
   if ((NULL != action) &&
       (NULL == action->next) &&
       (rw.sys_process == 0) && /* Make sure system processes shall be included */
       (!strcmp("*", action->names)) && /* from scope */
       (!strcmp("*", &action->names[3])) && /* to scope */
       ((offspring == NULL) || (offspring->evact_scope.type == MONITOR_SCOPE_SYSTEM)) &&
       (action->orig_event_count == 0) &&
       (action->event_seq_no == 0))
   {
      /* Tell pose to trace all context switches in the system without
       * filters. */
      zzset_trace_swaps_enabled(TRUE);
   }
   else
   {
      /* Turn off trace of all context switches in the system without filters. */
      zzset_trace_swaps_enabled(FALSE);
   }
}

/*
** Subfunction to do_modify().
** Delete one action entry from any of the action lists.
**
** Return zero if the entry was not found in the list.
** Return 1 if the entry was found in the list.
** Return 2 if the entry was found and the list is now empty => debug hooks need update.
*/

static int
do_delete_action(Offspring *offspring, uint32_t id)
{
   struct action_lists *actions = &rw.actions SMP_CODE([0]);
   int             ecode;

   /* scan all lists until found. */
   if ((ecode = do_delete_one_action(id, &actions->action_list_create, False)))
      return ecode;
   if ((ecode = do_delete_one_action(id, &actions->action_list_kill, False)))
      return ecode;
   if ((ecode = do_delete_one_action(id, &actions->action_list_swap, False)))
   {
      check_enable_swap_trace_opt(offspring);
      return ecode;
   }
   if ((ecode = do_delete_one_action(id, &actions->action_list_send, False)))
      return ecode;
   if ((ecode = do_delete_one_action(id, &actions->action_list_receive, False)))
      return ecode;
#ifdef HAS_TIMEOUT_EVENT
   if ((ecode = do_delete_one_action(id, &actions->action_list_timeout, False)))
      return ecode;
#endif
   if ((ecode = do_delete_one_action(id, &actions->action_list_oserror, False)))
      return ecode;
   if ((ecode = do_delete_one_action(id, &actions->action_list_restart, False)))
      return ecode;
   if ((ecode = do_delete_one_action(id, &actions->action_list_uevent, False)))
   {
      clean_user_event_list();
      return ecode;
   }
#ifdef OSE_SMP
   if ((ecode = do_delete_one_action(id, &actions->action_list_bind, False)))
      return ecode;
#endif
   if ((ecode = do_delete_one_action(id, &actions->action_list_alloc, False)))
   {
      if (NULL == actions->action_list_alloc)
      {
         zzkrn_set_handler(12, NULL);
      }
      return ecode;
   }
   if ((ecode = do_delete_one_action(id, &actions->action_list_free_buf, False)))
   {
      if (NULL == actions->action_list_free_buf)
      {
         zzkrn_set_handler(13, NULL);
      }
      return ecode;
   }
   ecode = do_delete_one_action(id, &actions->action_list_error, False);
   return ecode;
}

static int
is_persistent_action(int action)
{
   switch (action & 0xf)
   {
      case 3:                  /* Trace */
      case 4:                  /* Trace On */
      case 5:                  /* Trace Off */
      case 9:                  /* Goto */
         return 1;
      default:
         return 0;
   }
}

static void
do_delete_one_action_list(int save_trace_evacts, struct ldefine_action **action_list, int rmm_init)
{
   struct ldefine_action *action;

   assert(save_trace_evacts == 0 || save_trace_evacts == 0);
   assert(action_list);

   for (action = *action_list; action;)
   {
      if (!(save_trace_evacts && is_persistent_action(action->action)))
      {
         struct ldefine_action *next = action->next;

         do_delete_one_action(action->action_id, action_list, rmm_init);
         action = next;
      }
      else
      {
         /* We should keep this action so just skip to next */
         action = action->next;
      }
   }
}

/*
** Delete all action entries from all action lists with the option to
** save actions related to the trace (trace, trace on, trace off, and
** goto actions).
*/

static void
do_delete_all_actions(int save_trace_evacts, int rmm_init)
{
   struct action_lists *actions = &rw.actions SMP_CODE([0]);

   do_delete_one_action_list(save_trace_evacts, &actions->action_list_create, rmm_init);
   do_delete_one_action_list(save_trace_evacts, &actions->action_list_kill, rmm_init);
   do_delete_one_action_list(save_trace_evacts, &actions->action_list_swap, rmm_init);
   do_delete_one_action_list(save_trace_evacts, &actions->action_list_send, rmm_init);
   do_delete_one_action_list(save_trace_evacts, &actions->action_list_receive, rmm_init);
   do_delete_one_action_list(save_trace_evacts, &actions->action_list_oserror, rmm_init);
   do_delete_one_action_list(save_trace_evacts, &actions->action_list_error, rmm_init);
   do_delete_one_action_list(save_trace_evacts, &actions->action_list_restart, rmm_init);
   do_delete_one_action_list(save_trace_evacts, &actions->action_list_uevent, rmm_init);
#ifdef OSE_SMP
   do_delete_one_action_list(save_trace_evacts, &actions->action_list_bind, rmm_init);
#endif
   do_delete_one_action_list(save_trace_evacts, &actions->action_list_alloc, rmm_init);
   do_delete_one_action_list(save_trace_evacts, &actions->action_list_free_buf, rmm_init);
#ifdef HAS_TIMEOUT_EVENT
   do_delete_one_action_list(save_trace_evacts, &actions->action_list_timeout, rmm_init);
#endif

   check_enable_swap_trace_opt(NULL);
   clean_user_event_list();
}

static int
validate_evact(struct ldefine_action *action)
{
   uint32_t        p = (uint32_t) action;
   uint32_t        ts = (uint32_t) rw.static_trace;
   uint32_t        te = ((uint32_t) rw.static_trace) + rw.static_trace_size - 1;

   if (0 == p || p < ts || p > te)
   {
      ramlog_printf("rmm: illegal action pointer found.\n");
      return 0;
   }
   if (action->magic != ACTION_MAGIC_VALID)
   {
      ramlog_printf("rmm: illegal action magic found.\n");
      return 0;
   }
   p = (uint32_t) action->next;
   if (0 != p && (p < ts || p > te))
   {
      ramlog_printf("rmm: illegal next action pointer found.\n");
      return 0;
   }
#if 0
   p = (uint32_t) action->align;
   if (0 != p && (p < ts || p > te))
   {
      ramlog_printf("rmm: illegal action align pointer found.\n");
      return 0;
   }
#endif
   
   if (0 != action->data_range_size)
   {
      /* The most significant bit of the size is used as a flag for negation */
      uint32_t size = action->data_range_size & 0x7fffffff;
      
      switch (size)
      {
         case 1:
         case 2:
         case 4:
         case 8:
            break;
         default:
            ramlog_printf("rmm: illegal evact action, data range integer size: %u.\n", size);
            return 0;
      }
      if (0 != action->data_range_b && action->data_range_a >= action->data_range_b)
      {
         ramlog_printf("rmm: illegal evact action, data range: %llu %llu.\n",
               action->data_range_a, action->data_range_b);
         return 0;
      }
      if (UINT32_MAX < action->data_range_offset)
      {
         ramlog_printf("rmm: illegal action data range offset %llu.\n",
               action->data_range_offset);
         return 0;
      }
   
   
      if (0 != action->data_range_b && action->data_range_a >= action->data_range_b)
      {
         ramlog_printf("rmm: illegal evact action, data range: %llu %llu.\n",
               action->data_range_a, action->data_range_b);
         return 0;
      }
      if (UINT32_MAX < action->data_range_offset)
      {
         ramlog_printf("rmm: illegal action data range offset %llu.\n",
               action->data_range_offset);
         return 0;
      }
   }
   if ((action->action & ~TRACE_WITH_DATA) > HIGHEST_ACTION_VALUE)
   {
      ramlog_printf("rmm: illegal evact action found.\n");
      return 0;
   }
   if (strlen(action->names) > 1024)
   {
      ramlog_printf("rmm: illegal evact names found.\n");
      return 0;
   }
   return 1;
}

static int
reclaim_one_action_list(struct ldefine_action **actions, int *reclaimed_cnt)
{
   struct ldefine_action *action;

   for (action = *actions; action; action = action->next)
   {
      if (!validate_evact(action))
      {
         return 1;
      }
      action->align = (uint32_t) NULL;
      if (update_expanded_wildcards(action, 0) != 0)
      {
         ramlog_printf("rmm: evacts not recovered due to failure of wild card"
                       " pattern expansion.\n");
         return 1;
      }

      ose_trace_reclaim(action);
      (*reclaimed_cnt)++;
      SMP_CODE(smp_distribute_actionpoint(action, actions, True);)
   }

   return 0;
}

static int
reclaim_all_actions(void)
{
   struct action_lists *actions = &rw.actions SMP_CODE([0]);
   int             result;
   int reclaimed_cnt =0;

   result = reclaim_one_action_list(&actions->action_list_create, &reclaimed_cnt);
   result += reclaim_one_action_list(&actions->action_list_kill, &reclaimed_cnt);
   result += reclaim_one_action_list(&actions->action_list_swap, &reclaimed_cnt);
   result += reclaim_one_action_list(&actions->action_list_send, &reclaimed_cnt);
   result += reclaim_one_action_list(&actions->action_list_receive, &reclaimed_cnt);
   result += reclaim_one_action_list(&actions->action_list_oserror, &reclaimed_cnt);
   result += reclaim_one_action_list(&actions->action_list_error, &reclaimed_cnt);
   result += reclaim_one_action_list(&actions->action_list_restart, &reclaimed_cnt);
   result += reclaim_one_action_list(&actions->action_list_uevent, &reclaimed_cnt);
#ifdef OSE_SMP
   result += reclaim_one_action_list(&actions->action_list_bind, &reclaimed_cnt);
#endif
   result += reclaim_one_action_list(&actions->action_list_alloc, &reclaimed_cnt);
   result += reclaim_one_action_list(&actions->action_list_free_buf, &reclaimed_cnt);
#ifdef HAS_TIMEOUT_EVENT
   result += reclaim_one_action_list(&actions->action_list_timeout, &reclaimed_cnt);
#endif
   ramlog_printf("rmm: reclaimed %d action points\n", reclaimed_cnt);

   return result;
}

#ifdef USE_EPM

/* We should be able to use efmi to retrive registers that have hit an EP
 * and hence do not need to maintain a list of the register dumps anymore.
 * However we lack the functionality to set register by that way, hence
 * we need to keep our regdump administration...
 */
#define USE_REGDUMP_REGISTERS
#ifdef USE_REGDUMP_REGISTERS
#   define REGDUMP_PRE_ALLOCATION_COUNT 2

/* add_regdump()
 * Called from EPM callback hook to register a new regdump for the current
 * process. Note: called with interrups disabled.
 * */
static void
add_regdump(Offspring * offspring, struct cpu_regdump *regdump)
{
   if (offspring->process_regdumps != NULL)
   {
      ProcessRegdump *rd;

      for (rd = list_head(offspring->process_regdumps); rd != NULL; rd = list_next(offspring->process_regdumps, &rd->link))
      {
         if (0 == rd->pid)
         {
            LOCKSTATE       msr;

            LOCK_SAVE(msr);
            if (0 == rd->pid) /* Recheck that it is still free. */
            {
               /* Found a free slot, grab it! */
               offspring->regdump_free_count--;
               rd->pid = current_process();
               LOCK_RESTORE(msr);
               rd->regdump = regdump;
               break;
            }
            LOCK_RESTORE(msr);
         }
      }
   }
}

static struct cpu_regdump *
get_regdump(Offspring * offspring, uint32_t pid)
{
   struct cpu_regdump *dump = NULL;

   if (offspring->process_regdumps != NULL)
   {
      ProcessRegdump *rd;

      for (rd = list_head(offspring->process_regdumps); rd != NULL; rd = list_next(offspring->process_regdumps, &rd->link))
      {
         if (pid == rd->pid)
         {
            dump = rd->regdump;
            break;
         }
      }
   }
   return dump;
}

static void
clear_regdump(Offspring * offspring, uint32_t pid)
{
   if (offspring->process_regdumps != NULL)
   {
      ProcessRegdump *rd;

      for (rd = list_head(offspring->process_regdumps); rd != NULL; rd = list_next(offspring->process_regdumps, &rd->link))
      {
         if (pid == rd->pid)
         {
            LOCKSTATE       msr;

            BKL_LOCK_SAVE(msr);
            offspring->regdump_free_count++;
            rd->regdump = NULL;
            rd->pid = 0;
            /* Make sure empty entries is placed first in the list. */
            list_unlink(rd);
            list_insert_head(offspring->process_regdumps, rd);
            BKL_LOCK_RESTORE(msr);
            break;
         }
      }
   }
}

static void
clear_all_regdumps(Offspring * offspring)
{
   if (offspring->process_regdumps != NULL)
   {
      ProcessRegdump *rd;

      for (rd = list_head(offspring->process_regdumps); rd != NULL; rd = list_next(offspring->process_regdumps, &rd->link))
      {
         if (0 != rd->pid)
         {
            LOCKSTATE       msr;

            BKL_LOCK_SAVE(msr);
            offspring->regdump_free_count++;
            rd->regdump = NULL;
            rd->pid = 0;
            BKL_LOCK_RESTORE(msr);
         }
      }
   }
}

/* ensure_regdump_list_size()
 * Make sure we have at least two free regdump slots to be used by the epm
 * callback hook. This function should be called as soon as the rmm gets
 * notified about that an epm hook has trigged as well when new breakpoints
 * and single steps are requested.
 * */
static void
ensure_regdump_list_size(Offspring * offspring)
{
   if (NULL == offspring->process_regdumps)
   {
      offspring->process_regdumps = list_init((void *) alloc(sizeof (List), 0));
      offspring->regdump_free_count = 0;
   }
   while (offspring->regdump_free_count < REGDUMP_PRE_ALLOCATION_COUNT)
   {
      LOCKSTATE       msr;

      ProcessRegdump *rd = (ProcessRegdump *)alloc(sizeof (ProcessRegdump), 0);

      list_init_link(rd);
      rd->pid = 0;
      rd->regdump = NULL;

      /* The list is traversed by other process in the epm callback hook so
       * lock interrupts while modifying it. */
      BKL_LOCK_SAVE(msr);
      list_insert_head(offspring->process_regdumps, rd);
      offspring->regdump_free_count++;
      BKL_LOCK_RESTORE(msr);
   }
}
#endif

static void
free_ep(ExecutionPointInfo ** epi)
{
   assert(epi);
   assert(*epi);
   assert((*epi)->valid == EP_VALID);
   (*epi)->valid = EP_INVALID;
   (*epi)->offspring = NULL;
   if (NULL == rw.ep_free_list)
   {
      rw.ep_free_list = list_init((void *) alloc(sizeof (List), 0));
   }
   list_init_link(&(*epi)->link);
   list_insert_tail(rw.ep_free_list, (*epi));
   *epi = NULL;
}

static ExecutionPointInfo *
alloc_ep(void)
{
   ExecutionPointInfo *epi;

   if (rw.ep_free_list != NULL && list_length(rw.ep_free_list) > 0)
   {
      epi = (ExecutionPointInfo *) list_head(rw.ep_free_list);
      list_unlink(epi);
      epi->instance++;
   }
   else
   {
      epi = (ExecutionPointInfo *) alloc(sizeof (ExecutionPointInfo), 0);
      epi->instance = 1;
   }
   list_init_link(&epi->link);
   assert(epi);
   return epi;
}

static void
resume_scope(Offspring * offspring, MonitorScope * scope)
{
   assert(scope);

   switch (scope->type)
   {
      case MONITOR_SCOPE_SYSTEM:
      {
         struct OS_pid_list *bid_list;
         OSBUFSIZE       bid_count;
         PROCESS        *bid;

#ifdef USE_REGDUMP_REGISTERS
         LOCK_SWAP();
         clear_all_regdumps(offspring);
         UNLOCK_SWAP();
#endif
         if (!(bid_list = get_bid_list(0)))
            return;

         bid_count = bid_list->count;
         bid = bid_list->list;

         for (; bid_count; bid_count--, bid++)
         {
            resume(*bid);
         }

         free_buf((union SIGNAL **) &bid_list);
      }
         break;

      case MONITOR_SCOPE_SEGMENT_ID:
      {
         struct OS_pid_list *bid_list;
         OSBUFSIZE       bid_count;
         PROCESS        *bid;

         if (!(bid_list = get_bid_list(0)))
            return;

#ifdef USE_REGDUMP_REGISTERS
         LOCK_SWAP();
         /* FIXME we should really only clear the regdumps of this segment. */
         clear_all_regdumps(offspring);
         UNLOCK_SWAP();
#endif
         bid_count = bid_list->count;
         bid = bid_list->list;

         for (; bid_count; bid_count--, bid++)
         {
            if (get_segid(*bid) == scope->id)
            {
               resume(*bid);
            }
         }

         free_buf((union SIGNAL **) &bid_list);
      }
         break;

      case MONITOR_SCOPE_BLOCK_ID:
#ifdef USE_REGDUMP_REGISTERS
         LOCK_SWAP();
         /* FIXME we should really only clear the regdumps of this block. */
         clear_all_regdumps(offspring);
         UNLOCK_SWAP();
#endif
         resume(scope->id);
         break;

      case MONITOR_SCOPE_ID:
#ifdef USE_REGDUMP_REGISTERS
         LOCK_SWAP();
         clear_regdump(offspring, scope->id);
         UNLOCK_SWAP();
#endif
         resume(scope->id);
         break;

      case MONITOR_SCOPE_VOID:
         /* Do nothing. */
         break;

      default:
         assert(False);         /* Illegal scope */
   }
}

#endif

/*
** Utility to intercept the breakpoint scope.
**
** The dispatcher must be locked if a non interrupt process is the caller.
*/
static void
intercept_evact_scope(void)
{
   struct break_pid_entry *bke;
   PROCESS        *pid;
   U32             count;
   OSBOOLEAN      intercept_self = 0;

   /* Set the status flag for the breakpoint scope. */
   rw.break_scope_status = (OSBOOLEAN) 1;
   rw.precious->break_scope_status = (U8) rw.break_scope_status;

   /* Modify all processes in the current breakpoint scope. If we
    * are intercepting a block we want to intercept the process who's
    *  event initiated this last. Otherwise we won't finish unless
    * someone has locked the dispatcher before calling this function. */
   for (bke = rw.break_pid_list; bke; bke = bke->next)
   {
      for (count = bke->size, pid = bke->pid; count--; pid++)
      {
         if (*pid)
         {
            if (*pid != current_process())
            {
               intercept(*pid);
            }
            else
            {
               intercept_self = 1;
            }
         }
      }
   }

   /* Now finally we intercept the process whose context we are in
    * if it is in the breakpoint scope */
   if (intercept_self == 1)
   {
      intercept(current_process());
   }
 }

/*
** Utility to intercept only one process in the breakpoint scope.
*/
static int
intercept_evact_scope_proc(PROCESS interceptee)
{
   struct break_pid_entry *bke;
   PROCESS        *pid;
   U32             count;
   int intercepted = 0;

   if (interceptee == current_process() || 0 == interceptee)
   {
      /* Do not allow anyone to intercept the rmm (but the rmm should not be part of the catch scope???) */
      return 0;
   }

   LOCK_SWAP();
   if (get_ptype(interceptee) == OS_ZOOMBIE)
   {
      /* OK if the process is killed. */
      UNLOCK_SWAP();
      return 1;
   }
   for (bke = rw.break_pid_list; bke; bke = bke->next)
   {
      for (count = bke->size, pid = bke->pid; count--; pid++)
      {
         if (*pid == interceptee)
         {
            intercept(interceptee);
            intercepted = 1;
            break;
         }
      }
   }
   UNLOCK_SWAP();
   if (intercepted) return 1;
   else return 0;
 }

/*
** Utility to resume the breakpoint scope.
**
** The dispatcher must be locked if a non interrupt process is the caller.
*/
static void
resume_evact_scope(void)
{
   struct break_pid_entry *bke;
   PROCESS        *pid;
   U32             count;

   /* Set the status flag for the breakpoint scope. */
   rw.break_scope_status = (OSBOOLEAN) 0;
   rw.precious->break_scope_status = (U8) rw.break_scope_status;

   /* Modify all processes in the current breakpoint scope. */
   for (bke = rw.break_pid_list; bke; bke = bke->next)
   {
      for (count = bke->size, pid = bke->pid; count--; pid++)
      {
         if (*pid)
         {
            resume(*pid);
         }
      }
   }
}

static void
disable_catch(void)
{
   union SIGNAL   *catch_sig = NULL;

   BKL_LOCK();
   if (rw.catch_enable)
   {
      catch_sig = rw.catch_enable;
      rw.catch_enable = NULL;
   }
   if (rw.break_scope_status)
   {
      BKL_UNLOCK();
      resume_evact_scope();
   }
   else
   {
      BKL_UNLOCK();
   }
   if (catch_sig)
   {
      free_buf(&catch_sig);
   }
}

static void
enable_catch(Offspring *offspring)
{
   if (offspring->intercept_enabled)
   {
      OSBOOLEAN       enabled;
      BKL_LOCK();
      enabled = NULL != rw.catch_enable ? 1 : 0;
      BKL_UNLOCK();

      if (!enabled)
      {
         rw.catch_enable = alloc(sizeof (struct event_report_verbose),
                              CATCH_REPORT);
      }
   }
}

static char *
create_dynamic_trace(PROCESS ose_mm_, uint32_t size, uint32_t *actual_size,
                     const char *region_type)
{
   const SIGSELECT all_signals[] = { 0 };
   void           *memory = NULL;
   union SIGNAL   *signal;

   assert(*actual_size == 0);

   signal = alloc(sizeof (struct mm_create_region), MM_CREATE_REGION_REQUEST);

   signal->mm_create_region.size = size;
   signal->mm_create_region.domain = MM_KERNEL_DOMAIN;
   strncpy(signal->mm_create_region.name, "system_trace", MM_NAME_LENGTH);
   strcpy(signal->mm_create_region.phys_mem, MM_INTERNAL_RAM);
   strcpy(signal->mm_create_region.log_mem, region_type);
   signal->mm_create_region.permissions = SU_RW_USR_NA;
   signal->mm_create_region.cache_mode = MM_CACHE_COPY_BACK | MM_SPECULATIVE_ACCESS;
   signal->mm_create_region.base = 0,   /* Optional and not used. */
   signal->mm_create_region.options = 0;
   signal->mm_create_region.align = 32;
   signal->mm_create_region.phys_count = 0;
   signal->mm_create_region.low_limit = 0;
   send(&signal, ose_mm_);

   signal = receive_from(~0UL, all_signals, ose_mm_);
   assert(signal);
   if (signal->sig_no == MM_CREATE_REGION_REPLY)
   {
      if (signal->mm_create_region.status == MM_SUCCESS)
      {
         *actual_size = signal->mm_create_region.size;
         memory = (void *) signal->mm_create_region.base;
      }
      else
      {
         ramlog_printf("rmm: create_region of type \"%s\" and size %u failed "
                       "with status %u\n",
                       region_type, size, signal->mm_create_region.status);
      }
   }
   free_buf(&signal);

   return memory;
}

static void     set_monitor_status(Offspring * offspring, int enabled);

static void
detach_offspring(Offspring * offspring)
{

   assert_offspring(offspring);

   offspring->handles_new_events = 0;

   if (0 != offspring->scope_attach_ref)
   {
      detach(&offspring->scope_attach_ref);
      offspring->scope_attach_ref = 0;
   }

#ifdef USE_EPM
   if (offspring->execution_points != NULL)
   {
      ExecutionPointInfo *ep;
      ExecutionPointInfo *next_ep;

      for (ep = list_head(offspring->execution_points); ep != NULL; ep = next_ep)
      {
         next_ep = list_next(offspring->execution_points, &ep->link);
         assert(ep->offspring == offspring);
         ose_epm_remove_execution_point(&(ep->ep));
         list_unlink(ep);
         free_ep(&ep);
      }
      free_buf((union SIGNAL **)&offspring->execution_points);
      offspring->execution_points = NULL;
   }

#endif

   if (offspring == rw.evact_lock_owner)
   {
      set_monitor_status(offspring, 0);
      rw.evact_lock_owner = NULL;
      do_delete_all_actions(True, False);
      disable_catch();
      offspring->evact_scope.type = MONITOR_SCOPE_VOID;
   }

   /* Just in case the client was reading the trace make sure we don't have
    * any locket entries in the trace. */
   ose_trace_close(&offspring->cursor);

#ifdef USE_EPM
   if (offspring->debug_scope.type != MONITOR_SCOPE_VOID)
   {
      /* Resume entire application scope. */
      resume_scope(offspring, &offspring->debug_scope);
   }

   if (offspring->process_regdumps != NULL)
   {
      ProcessRegdump *rd;
      ProcessRegdump *next_rd;

      for (rd = list_head(offspring->process_regdumps); rd != NULL; rd = next_rd)
      {
         next_rd = list_next(offspring->process_regdumps, &rd->link);
         rd->pid = 0;
         rd->regdump = NULL;
         list_unlink(rd);
         free_buf((union SIGNAL **)&rd);
      }
      free_buf((union SIGNAL **)&offspring->process_regdumps);
   }

   offspring->debug_scope.type = MONITOR_SCOPE_VOID;
   offspring->scope_intercepted = (OSBOOLEAN) 0;
#endif

}


static void
disconnect_offspring(Offspring * offspring, const char *reason)
{
   assert_offspring(offspring);

   ramlog_printf("rmm: disconnecting offspring due to: %s.\n", reason);

   detach_offspring(offspring);

   if (0 != offspring->attach_ref)
   {
      detach(&offspring->attach_ref);
   }
   list_unlink(offspring);
   free_buf((union SIGNAL **)&offspring);

   if (list_head(rw.offsprings) == NULL)
   {
      zzprof_disconnect();
#ifdef USE_EPM
      /* If zzdebug_regs_on was ever modified by rmm and
       * the list of offsprings is empty (which implies
       * that no application scope is currently defined),
       * clear zzdebug_regs_on */
      if (modified_zzdebug_regs_on == 1)
      {
         zzdebug_regs_on = 0;
         modified_zzdebug_regs_on = 0;
      }
#endif
   }
}

/**
 * @function find_offspring_by_pid
 * @brief Look up a offspring by its pid.
 */
static Offspring  *
find_offspring_by_pid(const PROCESS pid)
{
   Offspring         *c;

   for (c = list_head(rw.offsprings); c != NULL; c = list_next(rw.offsprings, &c->link))
   {
      if (pid == c->pid)
      {
         break;
      }
   }
   return c;
}

#if 1/**
 * @function send2offspring
 * @brief send a reply signal that is under flow control.
 * @long returns true if successfull false if offspring was lost.
 */
static          Boolean
send2offspring(union SIGNAL **signal, Offspring * offspring)
{
   assert_offspring(offspring);
   assert(signal && *signal);

   offspring->flow_ctrl_block--;

   if (offspring->flow_ctrl_block == rw.flow_ctrl_block / 2)
   {
      union SIGNAL   *sig = alloc(sizeof (MonitorFlowControl),
                                  MONITOR_BLOCKING_FLOW_CONTROL_REQUEST);

      send(&sig, offspring->pid);
   }

   while (offspring->flow_ctrl_block <= 0)
   {
      Offspring         *c;

#ifdef USE_EPM
      static const SIGSELECT sel_flow_ctrl[] = {
         3,
         OS_ATTACH_SIG,
         MONITOR_BLOCKING_FLOW_CONTROL_REPLY,
         RMM_SCOPE_ATTACH_SIG
      };
#else
      static const SIGSELECT sel_flow_ctrl[] = {
         2,
         OS_ATTACH_SIG,
         MONITOR_BLOCKING_FLOW_CONTROL_REPLY,
      };
#endif

      union SIGNAL   *sig;
      sig = receive(sel_flow_ctrl);
      switch (sig->sig_no)
      {
         case MONITOR_BLOCKING_FLOW_CONTROL_REPLY:
            c = find_offspring_by_pid(sender(&sig));
            if (c != NULL)
            {
               c->flow_ctrl_block += rw.flow_ctrl_block / 2 + 1;
            }
            break;
         case OS_ATTACH_SIG:
            c = find_offspring_by_pid(sender(&sig));
            if (NULL != c)
            {
               c->attach_ref = 0;
               disconnect_offspring(c, "offspring killed");
               if (c == offspring)
               {
                  /* We lost our current offspring */
                  free_buf(&sig);
                  free_buf(signal);
                  return False;
               }
            }
            else
            {
               assert(False);   /* There really should be a offspring for this. */
            }
            break;

#ifdef USE_EPM
         case RMM_SCOPE_ATTACH_SIG:
            c = find_offspring_by_pid(sig->rmm_scope_attach_sig.offspring_pid);
            if (NULL != c)
            {
               c->scope_attach_ref = 0;
               detach_offspring(c);
            }
            else
            {
               assert(False);   /* There really should be a offspring for this. */
            }
            break;
#endif
         default:
            ramlog_printf("rmm: got unknown signal in send2offspring: signo: %u\n", sig->sig_no);
            assert(False);
      }
      free_buf(&sig);
   }
   send(signal, offspring->pid);
   return True;
}
#endif
static void alloc_get_event_actionpoints_reply(union SIGNAL **reply,
                                               MonitorEventActionpointInfo **info,
                                               const uint32_t tagsCount,
                                               const uint32_t wrapsigs)
{
   assert(info);
   assert(reply);

   if (wrapsigs)
   {
      *reply = alloc(sizeof (MonitorGetEventActionpointsReply) + tagsCount * sizeof (MonitorTag), 0);
      (*reply)->sig_no = MONITOR_GET_EVENT_ACTIONPOINTS_REPLY;
      (*reply)->get_event_actionpoints_reply.reserved0 = 0;
      *info = &(*reply)->get_event_actionpoints_reply.info;
   }
   else
   {
      *reply = alloc(sizeof (MonitorEventActionpointInfo) + tagsCount * sizeof (MonitorTag), 0);
      *info = &(*reply)->event_actionpoint_info;
   }
}

static int
send_action_list(Offspring * offspring, struct ldefine_action *action_list, uint32_t wrapsigs)
{
   struct ldefine_action *action;

   /* Traverse the action list. */
   for (action = action_list; action; action = action->next)
   {
      union SIGNAL *reply = NULL;
      MonitorEventActionpointInfo *info = NULL;

      uint32_t        tagsCount = action->orig_event_count ? 2 : 0;
      if (action->data_range_size > 0)
      {
         tagsCount += 8;
      }

      switch (action->event_type)
      {
         case 1:               /* Create */
         case 2:               /* Kill */
         {
            char           *to = &action->names[action->sig_from];

            switch (pattern_to_scopetype(to))
            {
               case MONITOR_SCOPE_NAME_PATTERN:
                  tagsCount += 1 + 1 + strlen(to) / sizeof (MonitorTag);
                  break;
               case MONITOR_SCOPE_ENTRYPOINT:
                  tagsCount += 2;
                  break;
               default:
                  break;
            }

            alloc_get_event_actionpoints_reply(&reply, &info, tagsCount, wrapsigs);

            info->tagsCount = 0;

            if (action->event_type == 1)
               info->sigNo = wrapsigs ? MONITOR_SET_CREATE_ACTIONPOINT_REQUEST
                                      : MONITOR_GET_CREATE_ACTIONPOINT_INFO_REPLY;
            else
               info->sigNo = wrapsigs ? MONITOR_SET_KILL_ACTIONPOINT_REQUEST
                                      : MONITOR_GET_KILL_ACTIONPOINT_INFO_REPLY;

            pattern_to_scope(to, &info->to, &info->tags[info->tagsCount], &info->tagsCount);
            info->from.type = MONITOR_SCOPE_SYSTEM;
            info->from.id = 0;
            break;
         }

         case 3:               /* Swap */
         {
            char           *to = &action->names[action->sig_to];
            char           *from = &action->names[action->sig_from];

            switch (pattern_to_scopetype(from))
            {
               case MONITOR_SCOPE_NAME_PATTERN:
                  tagsCount += 1 + 1 + strlen(from) / sizeof (MonitorTag);
                  break;
               case MONITOR_SCOPE_ENTRYPOINT:
                  tagsCount += 2;
                  break;
               default:
                  break;
            }
            switch (pattern_to_scopetype(to))
            {
               case MONITOR_SCOPE_NAME_PATTERN:
                  tagsCount += 1 + 1 + strlen(to) / sizeof (MonitorTag);
                  break;
               case MONITOR_SCOPE_ENTRYPOINT:
                  tagsCount += 2;
                  break;
               default:
                  break;
            }

            alloc_get_event_actionpoints_reply(&reply, &info, tagsCount, wrapsigs);

            info->tagsCount = 0;
            info->sigNo = wrapsigs ? MONITOR_SET_SWAP_ACTIONPOINT_REQUEST
                                   : MONITOR_GET_SWAP_ACTIONPOINT_INFO_REPLY;

            pattern_to_scope(from, &info->from, &info->tags[info->tagsCount], &info->tagsCount);
            pattern_to_scope(to, &info->to, &info->tags[info->tagsCount], &info->tagsCount);
            break;
         }

         case 4:               /* Send */
         case 5:               /* Receive */
         {
            char           *to = &action->names[action->sig_to];
            char           *from = &action->names[action->sig_from];

            switch (pattern_to_scopetype(from))
            {
               case MONITOR_SCOPE_NAME_PATTERN:
                  tagsCount += 1 + 1 + strlen(from) / sizeof (MonitorTag);
                  break;
               case MONITOR_SCOPE_ENTRYPOINT:
                  tagsCount += 2;
                  break;
               default:
                  break;
            }
            switch (pattern_to_scopetype(to))
            {
               case MONITOR_SCOPE_NAME_PATTERN:
                  tagsCount += 1 + 1 + strlen(to) / sizeof (MonitorTag);
                  break;
               case MONITOR_SCOPE_ENTRYPOINT:
                  tagsCount += 2;
                  break;
               default:
                  break;
            }

            if (action->signo > 0UL || action->signo_end < ~0UL)
            {
               tagsCount += 3;
            }

            alloc_get_event_actionpoints_reply(&reply, &info, tagsCount, wrapsigs);
            info->tagsCount = 0;

            if (action->event_type == 4)
               info->sigNo = wrapsigs ? MONITOR_SET_SEND_ACTIONPOINT_REQUEST
                                      : MONITOR_GET_SEND_ACTIONPOINT_INFO_REPLY;
            else
               info->sigNo = wrapsigs ? MONITOR_SET_RECEIVE_ACTIONPOINT_REQUEST
                                      : MONITOR_GET_RECEIVE_ACTIONPOINT_INFO_REPLY;

            pattern_to_scope(from, &info->from, &info->tags[info->tagsCount], &info->tagsCount);
            pattern_to_scope(to, &info->to, &info->tags[info->tagsCount], &info->tagsCount);

            if (action->signo > 0UL || action->signo_end < ~0UL)
            {
               info->tags[info->tagsCount].tag.type = MONITOR_SIGNAL_FILTER_TAG_NUMBER;
               info->tags[info->tagsCount + 1].num = action->signo;
               info->tags[info->tagsCount + 2].num = action->signo_end + 1;
               info->tags[info->tagsCount].tag.count = 3;
               info->tagsCount += 3;
            }
            break;
         }

#ifdef HAS_TIMEOUT_EVENT
	 case MONITOR_EVENT_TIMEOUT:          /* Receive_w_tmo timed out */
         {
	    char           *from = &action->names[action->sig_from];

	    switch (pattern_to_scopetype(from))
            {
               case MONITOR_SCOPE_NAME_PATTERN:
                  tagsCount += 1 + 1 + strlen(from) / sizeof (MonitorTag);
                  break;
               case MONITOR_SCOPE_ENTRYPOINT:
                  tagsCount += 2;
                  break;
               default:
                  break;
            }

            alloc_get_event_actionpoints_reply(&reply, &info, tagsCount, wrapsigs);
            info->tagsCount = 0;

            info->sigNo = wrapsigs ? MONITOR_SET_TIMEOUT_ACTIONPOINT_REQUEST
                                      : MONITOR_GET_TIMEOUT_ACTIONPOINT_INFO_REPLY;

	    info->to.type = 0;
	    info->to.id = 0;

	    pattern_to_scope(from, &info->from, &info->tags[info->tagsCount], &info->tagsCount);

            break;
         }
#endif

         case 6:               /* Exec error */
         case 11:              /* Error */
         {
            char           *from = &action->names[action->sig_from];

            switch (pattern_to_scopetype(from))
            {
               case MONITOR_SCOPE_NAME_PATTERN:
                  tagsCount += 1 + 1 + strlen(from) / sizeof (MonitorTag);
                  break;
               case MONITOR_SCOPE_ENTRYPOINT:
                  tagsCount += 2;
                  break;
               default:
                  break;
            }

            if (action->ecode > 0UL || action->sig_to < ~0UL)
               tagsCount += 3;
            if (action->event_type == 6)
               tagsCount += 1;

            alloc_get_event_actionpoints_reply(&reply, &info, tagsCount, wrapsigs);

            info->tagsCount = 0;
            info->sigNo = wrapsigs ? MONITOR_SET_ERROR_ACTIONPOINT_REQUEST
                                   : MONITOR_GET_ERROR_ACTIONPOINT_INFO_REPLY;

            pattern_to_scope(from, &info->from, &info->tags[info->tagsCount], &info->tagsCount);
            info->to.type = MONITOR_SCOPE_SYSTEM;
            info->to.id = 0;

            if (action->ecode > 0UL || action->sig_to < ~0UL)
            {
               info->tags[info->tagsCount].tag.type = MONITOR_ERROR_FILTER_TAG_ERROR;
               info->tags[info->tagsCount + 1].num = action->ecode;
               info->tags[info->tagsCount + 2].num = action->sig_to + 1;
               info->tags[info->tagsCount].tag.count = 3;
               info->tagsCount += 3;
            }
            if (action->event_type == 6)
            {
               info->tags[info->tagsCount].tag.type = MONITOR_ERROR_FILTER_TAG_EXEC;
               info->tags[info->tagsCount].tag.count = 1;
               info->tagsCount += 1;
            }
            break;
         }

         case MONITOR_EVENT_USER:                      /* User event */
         {
            if (wrapsigs)              /* Post OSE 5.2 signals must be wrapped */
            {
               char *from = &action->names[action->sig_from];

               switch (pattern_to_scopetype(from))
               {
                  case MONITOR_SCOPE_NAME_PATTERN:
                     tagsCount += 1 + 1 + strlen(from) / sizeof (MonitorTag);
                     break;
                  case MONITOR_SCOPE_ENTRYPOINT:
                     tagsCount += 2;
                     break;
                  default:
                     break;
               }

               if (action->signo > 0UL || action->signo_end < ~0UL)
               {
                  tagsCount += 3;
               }

               alloc_get_event_actionpoints_reply(&reply, &info, tagsCount, 1);

               info->tagsCount = 0;
               info->sigNo = MONITOR_SET_USER_ACTIONPOINT_REQUEST;

               pattern_to_scope(from, &info->from, &info->tags[info->tagsCount], &info->tagsCount);
               info->to.type = MONITOR_SCOPE_SYSTEM;
               info->to.id = 0;

               if (action->signo > 0UL || action->signo_end < ~0UL)
               {
                  info->tags[info->tagsCount].tag.type = MONITOR_USER_FILTER_TAG_NUMBER;
                  info->tags[info->tagsCount + 1].num = action->signo;
                  info->tags[info->tagsCount + 2].num = action->signo_end + 1;
                  info->tags[info->tagsCount].tag.count = 3;
                  info->tagsCount += 3;
               }
            }
            break;
         }

#ifdef OSE_SMP
         case 14:              /* Bind */
         {
            if (wrapsigs)              /* Post OSE 5.2 signals must be wrapped */
            {
               char           *to = &action->names[action->sig_to];

               switch (pattern_to_scopetype(to))
               {
                  case MONITOR_SCOPE_NAME_PATTERN:
                     tagsCount += 1 + 1 + strlen(to) / sizeof (MonitorTag);
                     break;
                  case MONITOR_SCOPE_ENTRYPOINT:
                     tagsCount += 2;
                     break;
                  default:
                     break;
               }

               alloc_get_event_actionpoints_reply(&reply, &info, tagsCount, wrapsigs);

               info->tagsCount = 0;
               info->sigNo = MONITOR_SET_BIND_ACTIONPOINT_REQUEST;

               pattern_to_scope(to, &info->to, &info->tags[info->tagsCount], &info->tagsCount);
               info->from.type = MONITOR_SCOPE_SYSTEM;
               info->from.id = 0;
            }
            break;
         }
#endif /* OSE_SMP */

         case MONITOR_EVENT_ALLOC:  /* Alloc */
         {
            if (wrapsigs)
            {
               char *from = &action->names[action->sig_from];

               switch (pattern_to_scopetype(from))
               {
                  case MONITOR_SCOPE_NAME_PATTERN:
                     tagsCount += 1 + 1 + strlen(from) / sizeof (MonitorTag);
                     break;
                  case MONITOR_SCOPE_ENTRYPOINT:
                     tagsCount += 2;
                     break;
                  default:
                     break;
               }

               if (action->signo > 0UL || action->signo_end < ~0UL)
               {
                  tagsCount += 3;
               }

               alloc_get_event_actionpoints_reply(&reply, &info, tagsCount, 1);

               info->tagsCount = 0;
               info->sigNo = MONITOR_SET_ALLOC_ACTIONPOINT_REQUEST;

               pattern_to_scope(from, &info->from, &info->tags[info->tagsCount],
                                &info->tagsCount);
               info->to.type = MONITOR_SCOPE_SYSTEM;
               info->to.id = 0;

               if (action->signo > 0UL || action->signo_end < ~0UL)
               {
                  info->tags[info->tagsCount].tag.type = MONITOR_SIGNAL_FILTER_TAG_NUMBER;
                  info->tags[info->tagsCount + 1].num = action->signo;
                  info->tags[info->tagsCount + 2].num = action->signo_end + 1;
                  info->tags[info->tagsCount].tag.count = 3;
                  info->tagsCount += 3;
               }
            }
            break;
         }

         case MONITOR_EVENT_FREE_BUF:  /* Free buf */
         {
            if (wrapsigs)
            {
               char *from = &action->names[action->sig_from];

               switch (pattern_to_scopetype(from))
               {
                  case MONITOR_SCOPE_NAME_PATTERN:
                     tagsCount += 1 + 1 + strlen(from) / sizeof (MonitorTag);
                     break;
                  case MONITOR_SCOPE_ENTRYPOINT:
                     tagsCount += 2;
                     break;
                  default:
                     break;
               }

               if (action->signo > 0UL || action->signo_end < ~0UL)
               {
                  tagsCount += 3;
               }

               alloc_get_event_actionpoints_reply(&reply, &info, tagsCount, 1);

               info->tagsCount = 0;
               info->sigNo = MONITOR_SET_FREE_BUF_ACTIONPOINT_REQUEST;

               pattern_to_scope(from, &info->from, &info->tags[info->tagsCount],
                                &info->tagsCount);
               info->to.type = MONITOR_SCOPE_SYSTEM;
               info->to.id = 0;

               if (action->signo > 0UL || action->signo_end < ~0UL)
               {
                  info->tags[info->tagsCount].tag.type = MONITOR_SIGNAL_FILTER_TAG_NUMBER;
                  info->tags[info->tagsCount + 1].num = action->signo;
                  info->tags[info->tagsCount + 2].num = action->signo_end + 1;
                  info->tags[info->tagsCount].tag.count = 3;
                  info->tagsCount += 3;
               }
            }
            break;
         }

         default:
            assert(0);
            break;
      }

      if (reply)
      {
         info->state = action->event_seq_no;
         info->not = action->negated;

         if ((action->action & 0xf) == 10)
         {
            info->action = MONITOR_ACTION_NOTIFY;
            info->parameter = MONITOR_EVENT_ACTIONPOINT_PARAMETER_DATA;
         }
         else
         {
            info->action = action->action & 0xf;
            if (action->action & TRACE_WITH_DATA)
            {
               info->parameter = MONITOR_EVENT_ACTIONPOINT_PARAMETER_DATA;
            }
            else
            {
               info->parameter = 0;
            }
         }

         if ((action->action & 0xf) != 9)
         {
            info->parameter |= MONITOR_EVENT_ACTIONPOINT_PARAMETER_FILE;
         }
         else
         {
            info->parameter = action->goto_list;
         }

         info->intercept.type = MONITOR_SCOPE_SYSTEM;
         info->intercept.id = 0;
         if (NULL != rw.evact_lock_owner)
         {
            info->client = rw.evact_lock_owner->client_;
         }
         else
         {
            info->client = 0;
         }
         info->id = action->action_id;

         if (action->orig_event_count)
            info->count = action->orig_event_count - action->event_count;
         else
            info->count = 0;

         info->reserved0 = 0;
         info->reserved1 = 0;
         info->reserved2 = 0;

         if (action->orig_event_count)
         {
            info->tags[info->tagsCount].tag.type = MONITOR_EVENT_ACTIONPOINT_TAG_IGNORE;
            info->tags[info->tagsCount + 1].num = action->orig_event_count;
            info->tags[info->tagsCount].tag.count = 2;
            info->tagsCount += 2;
         }

         if (action->data_range_size > 0)
         {
            if (0 == (action->data_range_size & (1U << 31)))
            {
               info->tags[info->tagsCount].tag.type = MONITOR_EVENT_ACTIONPOINT_TAG_DATA_RANGE;
            }
            else
            {
               info->tags[info->tagsCount].tag.type = MONITOR_EVENT_ACTIONPOINT_TAG_DATA_RANGE_NOT;               
            }
            info->tags[info->tagsCount].tag.count = 8;
            info->tags[info->tagsCount + 3].num = (uint32_t)(action->data_range_size & INT32_MAX);
#ifdef LITTLE_ENDIAN
            info->tags[info->tagsCount + 1].num = (uint32_t)(action->data_range_offset & UINT32_MAX);
            info->tags[info->tagsCount + 2].num = (uint32_t)(action->data_range_offset >> 32);
            info->tags[info->tagsCount + 4].num = (uint32_t)(action->data_range_a & UINT32_MAX);
            info->tags[info->tagsCount + 5].num = (uint32_t)(action->data_range_a >> 32);
            info->tags[info->tagsCount + 6].num = (uint32_t)(action->data_range_b & UINT32_MAX);
            info->tags[info->tagsCount + 7].num = (uint32_t)(action->data_range_b >> 32);
#else
            info->tags[info->tagsCount + 1].num = (uint32_t)(action->data_range_offset >> 32);
            info->tags[info->tagsCount + 2].num = (uint32_t)(action->data_range_offset & UINT32_MAX);
            info->tags[info->tagsCount + 4].num = (uint32_t)(action->data_range_a >> 32);
            info->tags[info->tagsCount + 5].num = (uint32_t)(action->data_range_a & UINT32_MAX);
            info->tags[info->tagsCount + 6].num = (uint32_t)(action->data_range_b >> 32);
            info->tags[info->tagsCount + 7].num = (uint32_t)(action->data_range_b & UINT32_MAX);
#endif
            info->tagsCount += 8;
         }

         /* Send data to the offspring. Abort if a stop request was received. */
         if (!send2offspring(&reply, offspring))
         {
            return 1;              /* abort. */
         }
      }
   }
   
   return 0;
}

static int
handle_get_event_actionpoints(Offspring * offspring, union SIGNAL *sig)
{
   struct action_lists *actions = &rw.actions SMP_CODE([0]);
   MonitorGetEventActionpointsRequest *req = &sig->get_event_actionpoints_request;
   union SIGNAL   *reply;
   uint32_t wrapsigs;

   assert(req);
   assert_offspring(offspring);

   wrapsigs = (req->sigNo == MONITOR_GET_EVENT_ACTIONPOINTS_REQUEST);

   /* Display all lists. */
   if (send_action_list(offspring, actions->action_list_create, wrapsigs))
      return 0;
   if (send_action_list(offspring, actions->action_list_kill, wrapsigs))
      return 0;
   if (send_action_list(offspring, actions->action_list_swap, wrapsigs))
      return 0;
   if (send_action_list(offspring, actions->action_list_send, wrapsigs))
      return 0;
   if (send_action_list(offspring, actions->action_list_receive, wrapsigs))
      return 0;
   if (send_action_list(offspring, actions->action_list_oserror, wrapsigs))
      return 0;
   if (send_action_list(offspring, actions->action_list_restart, wrapsigs))
      return 0;
   if (send_action_list(offspring, actions->action_list_error, wrapsigs))
      return 0;
#ifdef HAS_TIMEOUT_EVENT
   if (send_action_list(offspring, actions->action_list_timeout, wrapsigs))
      return 0;
#endif
   if (wrapsigs)
   {
      if (send_action_list(offspring, actions->action_list_uevent, wrapsigs))
         return 0;
#ifdef OSE_SMP
      if (send_action_list(offspring, actions->action_list_bind, wrapsigs))
         return 0;
#endif
      if (send_action_list(offspring, actions->action_list_alloc, wrapsigs))
         return 0;
      if (send_action_list(offspring, actions->action_list_free_buf, wrapsigs))
         return 0;
   }

   reply = alloc(sizeof (MonitorGetEventActionpointInfoEndmark),
                 wrapsigs ? MONITOR_GET_EVENT_ACTIONPOINTS_ENDMARK
                 : MONITOR_GET_EVENT_ACTIONPOINT_INFO_ENDMARK);

   reply->get_event_actionpoints_endmark.status = MONITOR_STATUS_OK;
   send(&reply, offspring->pid);
   return 0;
}

static          Boolean
scopes_identical(MonitorScope * a, MonitorScope * b)
{
   if (a->type != b->type)
   {
      return False;
   }

   /* We know now that the scope types are the same! */
   if (a->type == MONITOR_SCOPE_VOID || a->type == MONITOR_SCOPE_SYSTEM || a->id == b->id)
   {
      return True;
   }
   else
   {
      return False;
   }
}

static int
handle_remove_evact(Offspring * offspring, union SIGNAL *sig)
{
   MonitorClearEventActionpointRequest *req = &sig->clear_event_actionpoint_request;
   union SIGNAL   *reply;
   MonitorStatus   status;

   assert_offspring(offspring);
   assert(req);

   if (do_delete_action(offspring, req->id))
   {
      /* Reevaluate action conditions according to the new action list. */
      update_hook_bits();

      /* Found and removed actionpoint */
      status = MONITOR_STATUS_OK;
   }
   else
   {
      status = MONITOR_STATUS_REQUEST_INVALID;
   }

   reply = alloc(sizeof (MonitorClearEventActionpointReply), MONITOR_CLEAR_EVENT_ACTIONPOINT_REPLY);
   reply->clear_event_actionpoint_reply.status = status;
   send(&reply, offspring->pid);
   return 0;
}

static int
is_req_disabled(SIGSELECT signo)
{
   uint32_t index = (signo - MONITOR_INTERFACE_REQUEST) / (8*sizeof(uint32_t));
   uint32_t bit   = (signo - MONITOR_INTERFACE_REQUEST) % (8*sizeof(uint32_t));

   /* Flow control signals should always be received! */
   if ((signo != MONITOR_BLOCKING_FLOW_CONTROL_REPLY) &&
       (signo != MONITOR_NON_BLOCKING_FLOW_CONTROL_REPLY))
      return (rw.req_disabled_bitmask[index] & (1<<bit)) != 0;
   else
      return 0;
}

static void
handle_interface_request(union SIGNAL *sig)
{
   union SIGNAL   *reply;
   unsigned n_supported = 0;
   unsigned i;

   reply = alloc(sizeof (MonitorInterfaceReply)
                 + sizeof (request_sigs),
                 MONITOR_INTERFACE_REPLY);

   *((uint32_t *) reply->interface_reply.endian) = MONITOR_ENDIAN_VALUE;
   reply->interface_reply.reserved0 = 0;
   reply->interface_reply.reserved1 = 0;
   reply->interface_reply.reserved2 = 0;
   reply->interface_reply.osType = MONITOR_OS_OSE_5;

   /* TODO: the cpu class and type should be digged out in some other way. */
#ifdef OSPP
   reply->interface_reply.cpuClass = MONITOR_CPU_PPC;
#elif defined (OSARM)
   reply->interface_reply.cpuClass = MONITOR_CPU_ARM;
#elif defined(OSMIPS)
   reply->interface_reply.cpuClass = MONITOR_CPU_MIPS;
#else
   reply->interface_reply.cpuClass = MONITOR_CPU_UNKNOWN;
#endif

   for (i=0; i<sizeof request_sigs/sizeof request_sigs[0]; i++)
   {
      if (!is_req_disabled(request_sigs[i]))
      {
         /* Source profiling signals should be enabled for
          * boards which have bsp support and disabled for others.
          */
         if (request_sigs[i] == MONITOR_GET_SUPPORTED_COUNTER_TYPES_REQUEST)
         {
#ifndef softOSE
            if (profiling_get_supported_counter_event_types == NULL)
#endif
            {
               /* Skip all counter profiling signals from the request_sigs[].*/
               i += 4;
               continue;
            }
         }
         reply->interface_reply.sigs[n_supported++] = request_sigs[i];
      }
   }
   reply->interface_reply.sigsCount = n_supported;

   send(&reply, sender(&sig));
}

static void
decode_env_value(char **src)
{
   char *value = *src;
   if (strlen(value) == 8
       && (value[4] == 0x40 || value[4] == 0x60)
       && (value[5] == 0x40 || value[5] == 0x60)
       && (value[6] == 0x40 || value[6] == 0x60) && (value[7] == 0x40 || value[7] == 0x60))
   {
      /*
       * The pointer value is obtained by xor-ing the two
       * words in the value.
       */
      OSADDRESS      *mval = (OSADDRESS *) value;
      OSADDRESS       pval = mval[0] ^ mval[1];

      /* Free the envp encoded value string. */
      free_buf((union SIGNAL **) src);

      /*
       * Allocate a new value string and fill it with
       * the decoded envp value.
       */
      *src = (char *) alloc(16, 0);
      sprintf(*src, "0x%X", pval);
   }
}

static void
handle_name_request(union SIGNAL *sig)
{
   union SIGNAL      *reply;
   char              *sysname;
   int                system_name_len;
   static const char *monitor_name = "com.ose.RunModeMonitor";
   int                monitor_name_len = strlen(monitor_name);

   /* Search first for the environment variable */
   sysname = get_env(current_process(), "OSE_MONITOR_SYSTEMNAME");

   if (sysname == NULL)
   {
      sysname = system_name;
   }
   else
   {
      decode_env_value(&sysname);
   }

   system_name_len = strlen(sysname);

   reply = alloc(sizeof (MonitorNameReply) + system_name_len + monitor_name_len + 1, MONITOR_NAME_REPLY);

   reply->name_reply.reserved0 = 0;
   reply->name_reply.reserved1 = 0;
   reply->name_reply.reserved2 = 0;

   reply->name_reply.systemNameOffset = 0;
   strcpy(&reply->name_reply.data[0], sysname);
   reply->name_reply.monitorNameOffset = system_name_len + 1;
   strcpy(&reply->name_reply.data[system_name_len + 1], monitor_name);

   if (sysname != system_name)
   {
      free_buf((union SIGNAL **) &sysname);
   }

   send(&reply, sender(&sig));
}

static void
handle_get_offsprings(union SIGNAL *sig)
{
   union SIGNAL *reply;
   union SIGNAL *endmark;
   Offspring *offspring;
   RmmGetOffspringsReply *r;

   /* Traverse the offspring list. An offspring object exists for each client.*/
   for (offspring = list_head(rw.offsprings); offspring != NULL; offspring
            = list_next(rw.offsprings, &offspring->link))
   {
      size_t name_length;
      size_t sig_size;
      name_length = strlen(offspring->name) + 1;
      sig_size = name_length + sizeof(RmmGetOffspringsReply) - sizeof(uint8_t);

      /* Create and send a reply with offspring information */
      reply = alloc(sig_size, RMM_GET_OFFSPRINGS_REPLY);
      r = &(reply->get_offsprings_reply);
      r->clientPid = offspring->client_;
      r->scope.id = offspring->debug_scope.id;
      r->scope.type = offspring->debug_scope.type;
      r->interceptScope.id = offspring->evact_scope.id;
      r->interceptScope.type = offspring->evact_scope.type;
      r->offspringPid = offspring->pid;
      r->nbrOfBreakpoints = 0;
#ifdef USE_EPM
      if (offspring->execution_points != NULL)
      {
         r->nbrOfBreakpoints = list_length(offspring->execution_points);
      }
#endif
      r->nameLength = name_length;
      strncpy((char *) r->clientName, (const char *) offspring->name, name_length);
      send(&reply, sender(&sig));

   }
   endmark = alloc(sizeof(SIGSELECT), RMM_GET_OFFSPRINGS_ENDMARK);
   send(&endmark, sender(&sig));
}

static void
handle_connect_request(union SIGNAL *sig)
{
   Offspring         *offspring;
   union SIGNAL   *reply;
   OSTICK          utick;
   PROCESS         offspring_;
   union SIGNAL   *connect_info;
   char            name_buf[29];
   OSPRIORITY      offspring_prio;
   PROCESS         client_;
   uint32_t entry_size = 0;

   /* Constant array of the features that this RMM supports. */
   static const MonitorFeature monitor_features[] =
   {
      MONITOR_FEATURE_SCOPE,
      MONITOR_FEATURE_INTERCEPT_SCOPE,
      MONITOR_FEATURE_AUTOMATIC_SCOPE,
      MONITOR_FEATURE_COMPACT_REGISTER_SIGNALS,
      MONITOR_FEATURE_SYSTEM_PROCESSES,
      MONITOR_FEATURE_EVENT_TRACE,
      MONITOR_FEATURE_EVENT_NOTIFY,
      MONITOR_FEATURE_EVENT_INTERCEPT,
      MONITOR_FEATURE_EVENT_STATES,
      MONITOR_FEATURE_PROCESS_FILTER,
      MONITOR_FEATURE_BLOCK_FILTER,
      MONITOR_FEATURE_SIGNAL_FILTER,
      MONITOR_FEATURE_STACK_FILTER,
   #ifdef HAS_UNDO_ACTION
      MONITOR_FEATURE_EVENT_UNDO,
   #endif
   #ifdef HAS_PROF_PRIORITY_THRESHOLD
      MONITOR_FEATURE_PROF_PRIORITY_THRESHOLD,
   #endif
   #ifdef HAS_PROF_ALL_EXECUTION_UNITS
      MONITOR_FEATURE_PROF_ALL_EXECUTION_UNITS,
   #endif
      MONITOR_FEATURE_TOTAL_CPU_BLOCK_MEASURE,
      /* TODO: bugfix see OSECR-8779 MONITOR_FEATURE_PER_CORE_PROFILED_PROCESSES, */
      MONITOR_FEATURE_EVENT_USER_ACTION,
      MONITOR_FEATURE_SEGMENT_INFO_LONG,
      MONITOR_FEATURE_ACTIONPOINT_DATA_RANGE,
      MONITOR_FEATURE_TRACEPOINT_ANNOTATIONS
   #ifdef OSARM
      ,
      MONITOR_FEATURE_ACCUMULATED_HEAP_PROFILING
   #endif
   };


   client_ = sender(&sig);
   switch(get_ptype(client_))
   {
      case OS_PRI_PROC:
         offspring_prio = get_pri(client_);
         break;
      case OS_BG_PROC:
         offspring_prio = 31;
         break;
      default:
         offspring_prio = get_pri(current_process());
         break;
   }

   snprintf(name_buf, 29, "ose_rmm_offspring_0x%x", client_);
   offspring_ = create_process(OS_PRI_PROC, name_buf,
         ose_rmm_client, 4096, offspring_prio, 0, 0, NULL, 0, 0);
   connect_info = alloc(sizeof(RmmOffspringInfo), RMM_OFFSPRING_INFO);
   connect_info->rmm_offspring_info.client_ = client_;
   connect_info->rmm_offspring_info.flow_ctrl_blocking = rw.flow_ctrl_block;
   connect_info->rmm_offspring_info.flow_ctrl_non_blocking = rw.flow_ctrl_nblock;
   memcpy(connect_info->rmm_offspring_info.req_disabled_bitmask,
         rw.req_disabled_bitmask, sizeof(rw.req_disabled_bitmask));
   send(&connect_info, offspring_);
   start(offspring_);

   /* Dynamically add feature MONITOR_FEATURE_EVENT_HW_TRACE, 
    * if enable/disable hardware functions are avaliable. */
   if(rw.hw_trace[0] != NULL)
   {
      entry_size = sizeof(MonitorFeature);
   } 

   reply = alloc(sizeof (MonitorConnectReply) + sizeof (monitor_features) + entry_size - sizeof (MonitorFeature),
                         MONITOR_CONNECT_REPLY);
   /* The connection attempt is accepted. Allocate a control block
    * for it. We need to store for instance the pid of the client. */

   reply->connect_reply.status = MONITOR_STATUS_OK;
   reply->connect_reply.tickLength = system_tick();
   reply->connect_reply.tick = get_systime(&utick);
   reply->connect_reply.ntick = utick;

#ifdef OSPP
   reply->connect_reply.cpuClass = MONITOR_CPU_PPC;
#elif defined (OSARM)
   reply->connect_reply.cpuClass = MONITOR_CPU_ARM;
#elif defined(OSMIPS)
   reply->connect_reply.cpuClass = MONITOR_CPU_MIPS;
#else
   reply->connect_reply.cpuClass = MONITOR_CPU_UNKNOWN;
#endif

#ifdef OSPP
   reply->connect_reply.cpuType = get_pvr(); 
#else
   reply->connect_reply.cpuType = 0;    /* FIXME, try to find reasonable value. */
#endif
   reply->connect_reply.errorHandler = 0;
   reply->connect_reply.ntickFrequency = 1000000;
   reply->connect_reply.euCount = SMP_CODE_YES_NO((uint16_t) zzkrn_num_cpus(), 0);
   reply->connect_reply.startMode = (uint8_t)rw.start_mode;
   reply->connect_reply.reserved0 = 0;
   reply->connect_reply.reserved1 = 0;
   reply->connect_reply.reserved2 = 0;
   reply->connect_reply.reserved3 = 0;
   reply->connect_reply.reserved4 = 0;
   reply->connect_reply.reserved5 = 0;
   reply->connect_reply.reserved6 = 0;
   reply->connect_reply.reserved7 = 0;

   memcpy((void *)reply->connect_reply.features, monitor_features, sizeof (monitor_features));

   /* Dynamically add feature MONITOR_FEATURE_EVENT_HW_TRACE */
   if(entry_size)
   {
      reply->connect_reply.features[sizeof(monitor_features) / sizeof(MonitorFeature)] = MONITOR_FEATURE_EVENT_HW_TRACE;
   }
   reply->connect_reply.featuresCount = (sizeof(monitor_features) + entry_size) / sizeof(MonitorFeature);

   offspring = (Offspring *) list_init_link((void *)alloc(sizeof (Offspring), 0));
   memset(offspring, 0, sizeof (Offspring));
   offspring->magic_head = CLIENT_MAGIC;
   strncpy(offspring->name, sig->connect_request.name, CLIENT_NAME_LEN);
   offspring->pid = offspring_;
   offspring->client_ = client_;
   offspring->attach_ref = attach(NULL, offspring->pid);
   offspring->flow_ctrl_block = rw.flow_ctrl_block;
   offspring->flow_ctrl_nblock = rw.flow_ctrl_nblock;
   offspring->intercept_enabled = (OSBOOLEAN)1;
   offspring->debug_scope.type = MONITOR_SCOPE_VOID;
   offspring->evact_scope.type = MONITOR_SCOPE_VOID;
   offspring->scope_attach_ref = 0;
   offspring->keep_trace_actions = 0;
   offspring->handles_new_events = 0;
   offspring->mget_tmo = 0;
   offspring->mget_read = 1;
   offspring->mget_loss = 0;
   offspring->mget_req = NULL;
   offspring->mget_reply = NULL;
   memset(&offspring->cursor, 0, sizeof(TraceEntryState));
   offspring->magic_tail = CLIENT_MAGIC;
   idbuf_init(&offspring->notify_pids);
   idbuf_init(&offspring->mtrace_pids);
#ifdef USE_EPM
   offspring->scope_intercepted = (OSBOOLEAN) 0;
   offspring->process_regdumps = NULL;
#endif
   list_insert_tail(rw.offsprings, &offspring->link);

   send(&reply, offspring_);
}

static int
handle_set_sys_param_request(Offspring * offspring, union SIGNAL *sig_r)
{
   MonitorStatus   status;
   MonitorSetSysParamRequest *r = &sig_r->set_sys_param_request;

   union SIGNAL   *signal;

   if (strlen(&r->data[r->nameOffset]) <= SYSPARAM_MAX_NAME_LENGTH)
   {
      int rmm_param_idx = get_monitor_sysparam_index(&r->data[r->nameOffset]);
      if (0 == r->reconfigure || rmm_param_idx >= 0)
      {
         if (r->valueOffset == ~0UL)
         {
            /* Remove the system parameter */
            static const char empty_string[] = "";

            if (0 == sysparam_write(&r->data[r->nameOffset], empty_string))
            {
               status = MONITOR_STATUS_OK;
            }
            else
            {
               status = MONITOR_STATUS_SYSTEM_PARAM_ERROR;
            }
         }
         else
         {
            if (strlen(&r->data[r->valueOffset]) <= SYSPARAM_MAX_VALUE_LENGTH)
            {
               if (0 == sysparam_write(&r->data[r->nameOffset],
                                       &r->data[r->valueOffset]))
               {
                  status = MONITOR_STATUS_OK;
               }
               else
               {
                  status = MONITOR_STATUS_SYSTEM_PARAM_ERROR;
               }
            }
            else
            {
               status = MONITOR_STATUS_REQUEST_INVALID;
            }
         }
         if (r->reconfigure && MONITOR_STATUS_OK == status)
         {
            /* Reread and reconfigure the changed sys param. */
            configure_parameter(rmm_param_idx);
         }
      }
      else
      {
         /* Reconfigure requested for unknown system parameter. */
         status = MONITOR_STATUS_REQUEST_INVALID;
      }
   }
   else
   {
      /* Too long name length. */
      status = MONITOR_STATUS_REQUEST_INVALID;
   }

   signal = alloc(sizeof (MonitorActionReply), MONITOR_SET_SYS_PARAM_REPLY);
   signal->action_reply.status = status;
   send(&signal, offspring->pid);
   return 0;
}

OSADDRESS
zzrmm_get_entrypoint(PROCESS pid)
{
   OSADDRESS ep;
   struct debug_pcb *dcb;

   BKL_LOCK();
   dcb = zzget_debugpcb(pid);

   if (dcb)
   {
      ep = (OSADDRESS)dcb->entrypoint;
   }
   else
   {
      ep = (OSADDRESS)NULL;
   }
   BKL_UNLOCK();

   return ep;
}

int
zzrmm_is_sysproc(PROCESS pid)
{
   int is_sysproc;
   struct debug_pcb *dcb;

   BKL_LOCK();
   dcb = zzget_debugpcb(pid);

   if (dcb)
   {
      is_sysproc = (dcb->debug_flags & SYS_PROCESS) != 0;
   }
   else
   {
      is_sysproc = 0;
   }
   BKL_UNLOCK();

   return is_sysproc;
}

/*
** Subfunction to match_proc().
**
** Performs a binary search of a process id table to check if
** a specific pid is present in the table.
*/

static int
binsearch(uint32_t nelem, PROCESS * table, PROCESS key)
{
   PROCESS        *p;
   uint32_t        n;

   for (p = table, n = nelem; n;)
   {
      /* Check midpoint of what is left. */
      uint32_t        pivot = n >> 1;
      PROCESS        *q = p + pivot;

      if (key < *q)
         n = pivot;
      else if (key > *q)
      {
         p = q + 1;
         n -= pivot + 1;
      }
      else
         return 1;
   }

   /* Not found. Return zero. */
   return 0;
}

/*
** utility to match a process to a set of wildcard names.
** The debug_pcb structure represents the process to check
** The procnamelist contains a verbose view of the wildcard list
** to check.
** Returns 0 if no match was found.
** Returns 1 if a match was found.
*/

int
zzrmm_match_proc(struct debug_pcb *proc, struct procnamelist *wildcard)
{
   if (wildcard)
   {
      uint32_t        no_of_complexpids;
      struct complexpid *cp;

      /* Check the list of single pids. */
      if (wildcard->no_of_singlepids)
      {
         if (binsearch(wildcard->no_of_singlepids, wildcard->singlepidlist, proc->pid))
         {
            return 1;
         }
      }

      /* Check the list of complex wildcard expressions. */
      if ((no_of_complexpids = wildcard->no_of_complexpids))
      {
         for (cp = wildcard->complexpidlist; no_of_complexpids; cp++, no_of_complexpids--)
         {
            /* First check the segment pid */
            if (cp->segmenttype)
            {
               if (cp->sid != get_segid(pcb_to_bid(proc)))
               {
                   continue;
                }
            }

            /* Check the bid and blockname. */
            switch (cp->blocknametype)
            {
                  /* Match the block id. */
               case 1:
                  if (cp->blockptr.bid == pcb_to_bid(proc))
                     break;
                  /* Loop if no match. */
                  continue;

                  /* Match the name string. */
               case 2:
                  if (!estrcmp(cp->blockptr.name, proc->blockname))
                     break;
                  /* Loop if no match. */
                  continue;

                  /* Match anything. */
               default:
                  break;
            }

            /* Then check the process name. */
            switch (cp->procnametype)
            {
                  /* Match the process id. */
               case 1:
                  if (cp->procptr.pid == proc->pid)
                     return 1;
                  /* Loop if no match. */
                  continue;

                  /* Match the name string. */
               case 2:
                  if (!estrcmp(cp->procptr.name, proc->name))
                     return 1;
                  /* Loop if no match. */
                  continue;

                  /* Match the entrypoint. */
               case 3:
                  if (cp->procptr.entrypoint == proc->entrypoint)
                     return 1;
                  /* Loop if no match. */
                  continue;

                  /* Match anything. */
               default:
                  return 1;
            }
         }
      }
   }
   /* Not found anywhere. */
   return 0;
}

/* KLUDGE: Same as above, but does not require a debug pcb */
static int
match_proc_data(PROCESS pid, PROCESS bid, const char *tracename, OSADDRESS entrypoint, struct procnamelist *wildcard)
{
   if (wildcard)
   {
      uint32_t        no_of_complexpids;
      struct complexpid *cp;
      const char      *name = tracename;

      assert(count_colons(tracename) == 1);

      while (*name++ != ':') ;

      /* Check the list of single pids. */
      if (wildcard->no_of_singlepids)
      {
         if (binsearch(wildcard->no_of_singlepids, wildcard->singlepidlist, pid))
            return 1;
      }

      /* Check the list of complex wildcard expressions. */
      if ((no_of_complexpids = wildcard->no_of_complexpids))
      {
         for (cp = wildcard->complexpidlist; no_of_complexpids; cp++, no_of_complexpids--)
         {
            /* First check the segment pid */
            if (cp->segmenttype)
            {
               /* This function is used for the trace and we do not
                * have access to a segment pid.
                */
               continue;
            }
            /* Check the bid and blockname. */
            switch (cp->blocknametype)
            {
                  /* Match the block id. */
               case 1:
                  if (cp->blockptr.bid == bid)
                     break;
                  /* Loop if no match. */
                  continue;

                  /* Match the name string. */
               case 2:
                  if (!eestrcmp(cp->blockptr.name, tracename))
                     break;
                  /* Loop if no match. */
                  continue;

                  /* Match anything. */
               default:
                  break;
            }

            /* Then check the process name. */
            switch (cp->procnametype)
            {
                  /* Match the process id. */
               case 1:
                  if (cp->procptr.pid == pid)
                     return 1;
                  /* Loop if no match. */
                  continue;

                  /* Match the name string. */
               case 2:
                  if (!eestrcmp(cp->procptr.name, name))
                     return 1;
                  /* Loop if no match. */
                  continue;

                  /* Match the entrypoint. */
               case 3:
                  if (cp->procptr.entrypoint == (void *) entrypoint)
                     return 1;
                  /* Loop if no match. */
                  continue;

                  /* Match anything. */
               default:
                  return 1;
            }
         }
      }
   }
   /* Not found anywhere. */
   return 0;
}

 /*
 ** Forwards the trace cursor the specified number of trace entries, not
 ** counting AP's and other not readable things of course. If the number of
 ** entries to forward is larger than the number of entries in the trace the
 ** cursor will end up on the last entry, which can then be used by the caller
 ** and the return value will reflect the number of entries forwarded to get
 ** there.
 */
static int
trace_fast_forward(TraceEntryState *current_te, uint32_t nr_entries)
{
   uint32_t forwarded = 0;
   OSADDRESS old_te = (OSADDRESS)current_te->trace_entry;

   while (forwarded < nr_entries && ose_trace_read(current_te))
   {
      /* Skip allocation table entries and unknown trace entries. */
      union TRACE *te = (union TRACE *) current_te->trace_entry;

      if (te->event_no == 0)
         continue;
      if (te->event_no > TRACE_SYSTEM_START)
         continue;

      forwarded++;
   }

   /* If we just re-read the last entry, don't update counter */
   if (old_te == (OSADDRESS)current_te->trace_entry && forwarded == 1)
   {
      forwarded = 0;
   }

   /* Return the number of entries we forwarded the cursor. */
   return forwarded;
}

static int
handle_get_trace(Offspring * offspring, union SIGNAL *sig_r)
{
   uint32_t                count, cur_entry = 0;
   static struct           idbuf ids;
   static TraceEntryState  cur_te;
   union SIGNAL            *sig;
   union TRACE             *te;
   MonitorStatus           status = MONITOR_STATUS_OK;

   assert_offspring(offspring);
   assert(sig_r);
   assert(MONITOR_GET_TRACE_REQUEST == sig_r->sig_no);

   /* Stop the trace. */
   if (rw.trace_enabled)
   {
      ose_trace_disable(rw.trace);
      rw.precious->trace_enabled = rw.trace_enabled = 0;
   }

   /* Open the trace for reading with a cursor and read item #0 */
   if (0 != ose_trace_open(rw.trace, &cur_te) ||
       ose_trace_read(&cur_te) == 0)
   {
      status = MONITOR_STATUS_OK;
      goto abort;
   }

   /* Forward cursor to the first trace entry to display. */
   cur_entry = trace_fast_forward(&cur_te, sig_r->get_trace_request.from);

   /* If from entry is beyond end of trace: get some of the last entries. */
   if (cur_entry < sig_r->get_trace_request.from)
   {
      ose_trace_close(&cur_te);
      if (0 != ose_trace_open(rw.trace, &cur_te) ||
          ose_trace_read(&cur_te) == 0)
      {
         status = MONITOR_STATUS_OK;
         goto abort;
      }

      /* Legacy way of getting a "random" number of entries between 1 and 64. */
      cur_entry = trace_fast_forward(&cur_te, cur_entry - cur_entry%64);
   }

   /* Attempt to get from - to. */
   if (sig_r->get_trace_request.to > cur_entry)
   {
      count = sig_r->get_trace_request.to - cur_entry;
   }
   /* Invalid request: Skip and send endmark */
   else
   {
      status = MONITOR_STATUS_REQUEST_INVALID;
      goto abort;
   }

   idbuf_init(&ids);

   /* Display each entry until end of trace or count exhausted. */
   do
   {
      /* Skip allocation table entries and unknown trace entries. */
      te = (union TRACE *) cur_te.trace_entry;

      if (te->event_no == 0)
         continue;

      /* Check that we have a valid trace entry nbr */
      if (te->event_no > LAST_TRACE_ENTRY_NBR)
         continue;

      /* Build a signal depending on the event. */
      switch (te->event_no)
      {
         case TRACE_SEND:
         case TRACE_RECEIVE:
         {
            struct trace_send_receive *t = &te->trace_send_receive;

            if (!idbuf_contains(&ids, t->from->pid))
            {
               sig = alloc(sizeof (MonitorProcessInfo) + t->from->namelen,
                           MONITOR_GET_PROCESS_TRACE_REPLY);

               sig->process_info.pid = t->from->pid;
               sig->process_info.bid = 0;
               sig->process_info.sid = 0;
               sig->process_info.type = make_ptype_from_pri(t->from->priority);
               sig->process_info.entrypoint = 0;
               sig->process_info.properties = match_proc_data(t->from->pid, 0, t->from->name, 0, rw.expanded_sysprocnames);
               sig->process_info.reserved0 = 0;
               strncpy((char *) &sig->process_info.name, t->from->name, t->from->namelen + 1);

               if (!send2offspring(&sig, offspring))
               {
                  goto fail;    /* we lost our offspring. */
               }
            }
            if (!idbuf_contains(&ids, t->to->pid))
            {
               sig = alloc(sizeof (MonitorProcessInfo) + t->to->namelen, MONITOR_GET_PROCESS_TRACE_REPLY);

               sig->process_info.pid = t->to->pid;
               sig->process_info.bid = 0;
               sig->process_info.sid = 0;
               sig->process_info.type = make_ptype_from_pri(t->to->priority);
               sig->process_info.entrypoint = 0;
               sig->process_info.properties = match_proc_data(t->to->pid, 0, t->to->name, 0, rw.expanded_sysprocnames);
               sig->process_info.reserved0 = 0;
               strncpy((char *) &sig->process_info.name, t->to->name, t->to->namelen + 1);

               if (!send2offspring(&sig, offspring))
               {
                  goto fail;    /* we lost our offspring. */
               }
            }

            sig = alloc(sizeof (MonitorSendReceiveInfo) + t->sigsize + 1,
                        (te->event_no == TRACE_SEND) ? MONITOR_GET_SEND_TRACE_REPLY : MONITOR_GET_RECEIVE_TRACE_REPLY);

            sig->send_receive_info.signalNumber = *((uint32_t *) t->data);
            sig->send_receive_info.signalSender = t->from->pid;
            sig->send_receive_info.signalAddressee = t->to->pid;
            sig->send_receive_info.signalSize = 0;
            sig->send_receive_info.signalAddress = 0;
            sig->send_receive_info.signalId = 0;
            sig->send_receive_info.lineNumber = 0;
            sig->send_receive_info.fileNameOffset = 0;
            sig->send_receive_info.signalDataOffset = 0;
            sig->send_receive_info.signalDataCount = t->sigsize;
            memcpy(sig->send_receive_info.data, t->data, t->sigsize);
            sig->send_receive_info.euId = t->cpu;
            sig->send_receive_info.reserved0 = 0;
            sig->send_receive_info.reserved1 = 0;
            break;
         }

#ifdef HAS_TIMEOUT_EVENT
	 case TRACE_RECEIVE_TMO:
	 {
	    struct trace_receive_tmo *t = &te->trace_receive_tmo;

	    /* If we don't handle new events, just disregard this and continue*/
            if (offspring->handles_new_events == 0)
            {
               continue;
            }

            if (!idbuf_contains(&ids, t->process->pid))
            {
               sig = alloc(sizeof (MonitorProcessInfo) +
                                                  t->process->namelen,
                                                  MONITOR_GET_PROCESS_TRACE_REPLY);

               sig->process_info.pid = t->process->pid;
               sig->process_info.bid = 0;
               sig->process_info.sid = 0;
               sig->process_info.type = make_ptype_from_pri(t->process->priority);
               sig->process_info.entrypoint = 0;
               sig->process_info.properties = match_proc_data(t->process->pid,
                                                        0,
                                                        t->process->name,
                                                        0,
                                                        rw.expanded_sysprocnames);
               sig->process_info.reserved0 = 0;
               strncpy((char *) &sig->process_info.name, t->process->name, t->process->namelen + 1);

               if (!send2offspring(&sig, offspring))
               {
                  goto fail;    /* we lost our offspring. */
               }
            }

	    sig = alloc(sizeof (MonitorTimeoutInfo) + 1, MONITOR_GET_TIMEOUT_TRACE_REPLY);

            sig->timeout_info.signalAddressee = t->process->pid;
	    sig->timeout_info.timeout = t->timeout;
	    sig->timeout_info.tmoSource = 0; /* A receive timed out. */
            sig->timeout_info.lineNumber = 0;
            sig->timeout_info.fileNameOffset = 0;
            sig->timeout_info.euId = t->cpu;
            sig->timeout_info.reserved0 = 0;
	    break;
	 }
#endif
         case TRACE_SWAP:
         {
            struct trace_swap *t = &te->trace_swap;

            if (!idbuf_contains(&ids, t->from_process->pid))
            {
               sig = alloc(sizeof (MonitorProcessInfo) + t->from_process->namelen, MONITOR_GET_PROCESS_TRACE_REPLY);

               sig->process_info.pid = t->from_process->pid;
               sig->process_info.bid = 0;
               sig->process_info.sid = 0;
               sig->process_info.type = make_ptype_from_pri(t->from_process->priority);
               sig->process_info.entrypoint = 0;
               sig->process_info.properties = match_proc_data(t->from_process->pid, 0,
                                                              t->from_process->name, 0, rw.expanded_sysprocnames);
               sig->process_info.reserved0 = 0;
               strncpy((char *) &sig->process_info.name, t->from_process->name, t->from_process->namelen + 1);

               if (!send2offspring(&sig, offspring))
               {
                  goto fail;    /* we lost our offspring. */
               }
            }
            if (!idbuf_contains(&ids, t->to_process->pid))
            {
               sig = alloc(sizeof (MonitorProcessInfo) + t->to_process->namelen, MONITOR_GET_PROCESS_TRACE_REPLY);

               sig->process_info.pid = t->to_process->pid;
               sig->process_info.bid = 0;
               sig->process_info.sid = 0;
               sig->process_info.type = make_ptype_from_pri(t->to_process->priority);
               sig->process_info.entrypoint = 0;
               sig->process_info.properties = match_proc_data(t->to_process->pid, 0,
                                                              t->to_process->name, 0, rw.expanded_sysprocnames);
               sig->process_info.reserved0 = 0;
               strncpy((char *) &sig->process_info.name, t->to_process->name, t->to_process->namelen + 1);

               if (!send2offspring(&sig, offspring))
               {
                  goto fail;    /* we lost our offspring. */
               }
            }

            sig = alloc(sizeof (MonitorSwapCreateKillInfo), MONITOR_GET_SWAP_TRACE_REPLY);

            sig->swap_create_kill_info.from = t->from_process->pid;
            sig->swap_create_kill_info.to = t->to_process->pid;
            sig->swap_create_kill_info.lineNumber = 0;
            sig->swap_create_kill_info.fileNameOffset = 0;
            sig->swap_create_kill_info.euId = t->cpu;
            sig->swap_create_kill_info.reserved0 = 0;
            sig->swap_create_kill_info.reserved1 = 0;
            break;
         }

         case TRACE_USER_ERROR:
         case TRACE_OSERROR:
         {
            struct trace_user_error *t = &te->trace_user_error;

            if (!idbuf_contains(&ids, t->process->pid))
            {
               sig = alloc(sizeof (MonitorProcessInfo) + t->process->namelen, MONITOR_GET_PROCESS_TRACE_REPLY);

               sig->process_info.pid = t->process->pid;
               sig->process_info.bid = 0;
               sig->process_info.sid = 0;
               sig->process_info.type = make_ptype_from_pri(t->process->priority);
               sig->process_info.entrypoint = 0;
               sig->process_info.properties = match_proc_data(t->process->pid, 0,
                                                              t->process->name, 0, rw.expanded_sysprocnames);
               sig->process_info.reserved0 = 0;
               strncpy((char *) &sig->process_info.name, t->process->name, t->process->namelen + 1);

               if (!send2offspring(&sig, offspring))
               {
                  goto fail;    /* we lost our offspring. */
               }
            }

            sig = alloc(sizeof (MonitorErrorInfo), MONITOR_GET_ERROR_TRACE_REPLY);

            sig->error_info.from = t->process->pid;
            sig->error_info.error = t->ecode;
            sig->error_info.extra = t->extra;
            sig->error_info.exec = (te->event_no == TRACE_OSERROR);
            sig->error_info.lineNumber = 0;
            sig->error_info.fileNameOffset = 0;
            sig->error_info.euId = t->cpu;
            sig->error_info.reserved0 = 0;
            sig->error_info.reserved1 = 0;
            break;
         }

         case TRACE_CREATE:
         case TRACE_KILL:
         {
            struct trace_create_kill *t = &te->trace_create_kill;

            if (!idbuf_contains(&ids, t->process->pid))
            {
               sig = alloc(sizeof (MonitorProcessInfo) + t->process->namelen, MONITOR_GET_PROCESS_TRACE_REPLY);

               sig->process_info.pid = t->process->pid;
               sig->process_info.bid = 0;
               sig->process_info.sid = 0;
               sig->process_info.type = make_ptype_from_pri(t->process->priority);
               sig->process_info.entrypoint = 0;
               sig->process_info.properties = match_proc_data(t->process->pid, 0,
                                                              t->process->name, 0, rw.expanded_sysprocnames);
               sig->process_info.reserved0 = 0;
               strncpy((char *) &sig->process_info.name, t->process->name, t->process->namelen + 1);

               if (!send2offspring(&sig, offspring))
               {
                  goto fail;    /* we lost our offspring. */
               }
            }

            sig = alloc(sizeof (MonitorSwapCreateKillInfo),
                        (t->event_no == TRACE_CREATE) ? MONITOR_GET_CREATE_TRACE_REPLY : MONITOR_GET_KILL_TRACE_REPLY);

            sig->swap_create_kill_info.from = 0;
            sig->swap_create_kill_info.to = t->process->pid;
            sig->swap_create_kill_info.lineNumber = 0;
            sig->swap_create_kill_info.fileNameOffset = 0;
            sig->swap_create_kill_info.euId = t->cpu;
            sig->swap_create_kill_info.reserved0 = 0;
            sig->swap_create_kill_info.reserved1 = 0;
            break;
         }

         case TRACE_SYSTEM_START:
         {
            struct trace_system_start *t = &te->trace_system_start;

            sig = alloc(sizeof (MonitorResetInfo), MONITOR_GET_RESET_TRACE_REPLY);

            sig->reset_info.warm = t->warm_flag;
            sig->reset_info.lineNumber = 0;
            sig->reset_info.fileNameOffset = 0;
            sig->reset_info.reserved0 = 0;
            sig->reset_info.reserved1 = 0;
            sig->reset_info.reserved2 = 0;
            break;
         }

         case TRACE_USER_EVENT:
         {
            struct trace_user_event *tue = &te->trace_user_event;

            /* If we don't handle new events, just disregard this and continue*/
            if (offspring->handles_new_events == 0)
            {
               continue;
            }

            if (!idbuf_contains(&ids, tue->process->pid))
            {
               sig = alloc(sizeof (MonitorProcessInfo) + tue->process->namelen,
                                          MONITOR_GET_PROCESS_TRACE_REPLY);

               sig->process_info.pid = tue->process->pid;
               sig->process_info.bid = 0;
               sig->process_info.sid = 0;
               sig->process_info.type = make_ptype_from_pri(tue->process->priority);
               sig->process_info.entrypoint = 0;
               sig->process_info.properties = match_proc_data(tue->process->pid, 0,
                                                              tue->process->name, 0, rw.expanded_sysprocnames);
               sig->process_info.reserved0 = 0;
               strncpy((char *) &sig->process_info.name, tue->process->name, tue->process->namelen + 1);

               if (!send2offspring(&sig, offspring))
               {
                  goto fail;    /* we lost our offspring. */
               }
            }
                 sig = alloc(sizeof (MonitorUserInfo) + tue->event_size,
                        MONITOR_GET_USER_TRACE_REPLY);

            sig->user_info.from = tue->process->pid;
            sig->user_info.eventNumber = tue->event_id;
            sig->user_info.eventDataOffset = 0;
            sig->user_info.eventDataCount = tue->event_size;
            sig->user_info.eventSize = tue->requested_size;
            sig->user_info.lineNumber = 0;
            sig->user_info.fileNameOffset = 0;
            sig->user_info.euId = tue->cpu;
            sig->user_info.reserved0 = 0;

            memcpy(sig->user_info.data, &tue->data, tue->event_size);

            break;
         }

#ifdef OSE_SMP
         case TRACE_BIND:
         {
            struct trace_bind *t = &te->trace_bind;

            /* If we don't handle new events, just disregard this and continue*/
            if (offspring->handles_new_events == 0)
            {
               continue;
            }

            if (!idbuf_contains(&ids, t->process->pid))
            {
               sig = alloc(sizeof (MonitorProcessInfo) + t->process->namelen,
                           MONITOR_GET_PROCESS_TRACE_REPLY);

               sig->process_info.pid = t->process->pid;
               sig->process_info.bid = 0;
               sig->process_info.sid = 0;
               sig->process_info.type = make_ptype_from_pri(t->process->priority);
               sig->process_info.entrypoint = 0;
               sig->process_info.properties = match_proc_data(t->process->pid,
                                                        0,
                                                        t->process->name,
                                                        0,
                                                        rw.expanded_sysprocnames);
               sig->process_info.reserved0 = 0;
               strncpy((char *) &sig->process_info.name, t->process->name, t->process->namelen + 1);

               if (!send2offspring(&sig, offspring))
               {
                  goto fail;    /* we lost our offspring. */
               }
            }

            sig = alloc(sizeof (MonitorBindInfo), MONITOR_GET_BIND_TRACE_REPLY);

            sig->bind_info.reserved0 = 0;
            sig->bind_info.bound = t->process->pid;
            sig->bind_info.fromEuId = t->from_eu;
            sig->bind_info.toEuId = t->to_eu;
            sig->bind_info.lineNumber = 0;
            sig->bind_info.fileNameOffset = 0;
            sig->bind_info.reserved1 = 0;
            break;
         }
#endif /* OSE_SMP */

         case TRACE_ALLOC:
         {
            struct trace_alloc_free_buf *t = &te->trace_alloc_free_buf;

            /* If we don't handle new events, just disregard this and continue*/
            if (offspring->handles_new_events == 0)
            {
               continue;
            }

            if (!idbuf_contains(&ids, t->process->pid))
            {
               sig = alloc(sizeof (MonitorProcessInfo) +
                                                  t->process->namelen,
                                                  MONITOR_GET_PROCESS_TRACE_REPLY);

               sig->process_info.pid = t->process->pid;
               sig->process_info.bid = 0;
               sig->process_info.sid = 0;
               sig->process_info.type = make_ptype_from_pri(t->process->priority);
               sig->process_info.entrypoint = 0;
               sig->process_info.properties = match_proc_data(t->process->pid,
                                                        0,
                                                        t->process->name,
                                                        0,
                                                        rw.expanded_sysprocnames);
               sig->process_info.reserved0 = 0;
               strncpy((char *) &sig->process_info.name, t->process->name, t->process->namelen + 1);

               if (!send2offspring(&sig, offspring))
               {
                  goto fail;    /* we lost our offspring. */
               }
            }

            sig = alloc(sizeof (MonitorAllocFreeBufInfo),
                        MONITOR_GET_ALLOC_TRACE_REPLY);
              sig->alloc_free_buf_info.from = t->process->pid;
                 sig->alloc_free_buf_info.signalNumber = *(uint32_t *)t->data;
                 sig->alloc_free_buf_info.signalSize = t->requested_size;
                 sig->alloc_free_buf_info.signalAddress = t->address;
            sig->alloc_free_buf_info.lineNumber = 0;
            sig->alloc_free_buf_info.fileNameOffset = 0;
            sig->alloc_free_buf_info.euId = t->cpu;
            sig->alloc_free_buf_info.reserved0 = 0;
            break;
         }

         case TRACE_FREE_BUF:
         {
            struct trace_alloc_free_buf *t = &te->trace_alloc_free_buf;

            /* If we don't handle new events, just disregard this and continue*/
            if (offspring->handles_new_events == 0)
            {
               continue;
            }

           if (!idbuf_contains(&ids, t->process->pid))
            {
               sig = alloc(sizeof (MonitorProcessInfo) +
                              t->process->namelen,
                              MONITOR_GET_PROCESS_TRACE_REPLY);

               sig->process_info.pid = t->process->pid;
               sig->process_info.bid = 0;
               sig->process_info.sid = 0;
               sig->process_info.type = make_ptype_from_pri(t->process->priority);
               sig->process_info.entrypoint = 0;
               sig->process_info.properties = match_proc_data(t->process->pid,
                                                        0,
                                                        t->process->name,
                                                        0,
                                                        rw.expanded_sysprocnames);
               sig->process_info.reserved0 = 0;
               strncpy((char *) &sig->process_info.name, t->process->name, t->process->namelen + 1);

               if (!send2offspring(&sig, offspring))
               {
                  goto fail;    /* we lost our offspring. */
               }
            }

            sig = alloc(sizeof (MonitorAllocFreeBufInfo),
                        MONITOR_GET_FREE_BUF_TRACE_REPLY);
            sig->alloc_free_buf_info.from = t->process->pid;
            sig->alloc_free_buf_info.signalNumber = *(uint32_t *)t->data;
            sig->alloc_free_buf_info.signalSize = t->requested_size;
            sig->alloc_free_buf_info.signalAddress = t->address;
            sig->alloc_free_buf_info.lineNumber = 0;
            sig->alloc_free_buf_info.fileNameOffset = 0;
            sig->alloc_free_buf_info.euId = t->cpu;
            sig->alloc_free_buf_info.reserved0 = 0;
            break;
         }

         case TRACE_NOT_USED:
            continue;

         default:
         {
            assert(0);
            /* We won't end up here, but just in case... */
            continue;
         }
      }

      /* fill in generic trace info. */
      sig->send_receive_info.reference = (uint32_t) cur_entry;
      sig->send_receive_info.time.tick = (uint32_t) cur_te.tick;
      sig->send_receive_info.time.ntick = (uint32_t) cur_te.time_count;
      if (te->trace_send_receive.abstime)
      {
         sig->send_receive_info.time.sec = te->trace_send_receive.abstime->abs_sec;
         sig->send_receive_info.time.sectick = te->trace_send_receive.abstime->abs_tick;
      }
      else
      {
         sig->send_receive_info.time.sec = (uint32_t) 0;
         sig->send_receive_info.time.sectick = (uint32_t) 0;
      }
      if (!send2offspring(&sig, offspring))
      {
         goto fail;             /* we lost our offspring. */
      }

      /* Done if count expires. */
      if (!--count)
      {
         break;
      }
      cur_entry++;
   }
   while (ose_trace_read(&cur_te));

 abort:
   sig = alloc(sizeof (MonitorGetTraceEndmark), MONITOR_GET_TRACE_ENDMARK);
   sig->get_trace_endmark.status = status;
   send(&sig, offspring->pid);

 fail:
   ose_trace_close(&cur_te);

   return 0;
}

static MonitorEventInfo *
add_event_info(MonitorGetTraceMultipleReply *reply, SIGSELECT sig_no, size_t size)
{
   MonitorEventInfo *e = (MonitorEventInfo *) &reply->events[reply->reserved0];
   size_t            units;

   /* Add space for header and reserved0 fields */
   size += sizeof(MonitorEventInfoHeader) + sizeof(uint32_t);

   /* Update reserved0 field (in units of MonitorEventInfoHeader:s) When a
    * related group of events (i.e one or two ProcessInfo and one Regular
    * Event) have beed added, a call to commit_event_info() will update the
    * real 'eventsCount' field
    */
   units = ROUNDUP(size, sizeof(MonitorEventInfoHeader));

   reply->reserved0 += units;

   if (reply->reserved0 >= rw.trace_mget_size)
   {
      return NULL;
   }

   /* Initialize common fields */
   e->header.size = units;
   e->reserved0   = 0;
   e->info.sigNo  = sig_no;

   return e;
}

static void
commit_event_info(MonitorGetTraceMultipleReply *reply)
{
   assert(reply->reserved0 <= rw.trace_mget_size);
   assert(reply->eventsCount < reply->reserved0);

   reply->eventsCount = reply->reserved0;
}

static int
check_trace_proc_info(const struct trace_procname *p)
{
   if (p->pid < 0x10000 || (strlen(p->name) > 255))
   {
      ramlog_printf("rmm: illegal process info found in trace\n");
      return 0;
   }
   return 1;
}

static MonitorEventInfo *
add_process_info(MonitorGetTraceMultipleReply *reply,
                 struct idbuf *pid_cache,
                 const struct trace_procname *p)
{
   MonitorEventInfo *e;

   if (!check_trace_proc_info(p))
   {
      reply->status = MONITOR_STATUS_KERNEL_ERROR;
      return NULL;
   }

   e = add_event_info(reply, MONITOR_GET_PROCESS_TRACE_REPLY,
                      sizeof(MonitorProcessInfo) + p->namelen);

   if (e != NULL)
   {
      e->info.process.pid = p->pid;
      e->info.process.bid = 0;
      e->info.process.sid = 0;
      e->info.process.type = make_ptype_from_pri(p->priority);
      e->info.process.entrypoint = 0;
      e->info.process.properties = match_proc_data(p->pid, 0, p->name, 0, rw.expanded_sysprocnames);
      e->info.process.reserved0 = 0;
      strncpy(e->info.process.name, p->name, p->namelen + 1);
      commit_event_info(reply);
#if 0
      ramlog_printf("rmm: sent proc info for 0x%X type %d name %s namelen % d P: 0x%X\n",
         p->pid, e->info.process.type, p->name, p->namelen, p);
#endif
   }
   else
   {
#if 0
      ramlog_printf("rmm: **** failed to add 0x%X len %d name %s P 0x%X\n",
         p->pid, p->namelen, p->name, p);
#endif
      /* Failed so clear the cache. */
      idbuf_init(pid_cache);
   }
   return e;
}

static MonitorEventInfo *
add_send_receive_info(MonitorGetTraceMultipleReply *reply, SIGSELECT sigNo,
                      const struct trace_send_receive *t)
{
   MonitorEventInfo *e;

   e = add_event_info(reply, sigNo, sizeof(MonitorSendReceiveInfo) + t->sigsize + 1);
   if (e != NULL)
   {
      e->info.sendReceive.signalNumber = *((const uint32_t *) t->data);
      e->info.sendReceive.signalSender = t->from->pid;
      e->info.sendReceive.signalAddressee = t->to->pid;
      e->info.sendReceive.signalSize = 0;
      e->info.sendReceive.signalAddress = 0;
      e->info.sendReceive.signalId = 0;
      e->info.sendReceive.lineNumber = 0;
      e->info.sendReceive.fileNameOffset = 0;
      e->info.sendReceive.signalDataOffset = 0;
      e->info.sendReceive.signalDataCount = t->sigsize;
      memcpy(e->info.sendReceive.data, t->data, t->sigsize);
      e->info.sendReceive.euId = t->cpu;
      e->info.sendReceive.reserved0 = 0;
      e->info.sendReceive.reserved1 = 0;
   }
   return e;
}

#ifdef HAS_TIMEOUT_EVENT
static MonitorEventInfo *
add_timeout_info(MonitorGetTraceMultipleReply *reply, SIGSELECT sigNo,
                      const struct trace_receive_tmo *t)
{
   MonitorEventInfo *e;

   e = add_event_info(reply, sigNo, sizeof(MonitorTimeoutInfo) + 1);
   if (e != NULL)
   {
      e->info.timeout.signalAddressee = t->process->pid;
      e->info.timeout.timeout = t->timeout;
      e->info.timeout.tmoSource = 0;
      e->info.timeout.lineNumber = 0;
      e->info.timeout.fileNameOffset = 0;
      e->info.timeout.euId = t->cpu;
      e->info.timeout.reserved0 = 0;
   }
   return e;
}
#endif

static MonitorEventInfo *
add_swap_info(MonitorGetTraceMultipleReply *reply, const struct trace_swap *t)
{
   MonitorEventInfo *e;

   e = add_event_info(reply, MONITOR_GET_SWAP_TRACE_REPLY,
                      sizeof(MonitorSwapCreateKillInfo));
   if (e != NULL)
   {
      e->info.swapCreateKill.from = t->from_process->pid;
      e->info.swapCreateKill.to = t->to_process->pid;
      e->info.swapCreateKill.lineNumber = 0;
      e->info.swapCreateKill.fileNameOffset = 0;
      e->info.swapCreateKill.euId = t->cpu;
      e->info.swapCreateKill.reserved0 = 0;
      e->info.swapCreateKill.reserved1 = 0;
   }
   return e;
}

static MonitorEventInfo *
add_create_kill_info(MonitorGetTraceMultipleReply *reply, SIGSELECT sigNo,
                     const struct trace_create_kill *t)
{
   MonitorEventInfo *e;

   e = add_event_info(reply, sigNo, sizeof(MonitorSwapCreateKillInfo));
   if (e != NULL)
   {
      e->info.swapCreateKill.from = 0;
      e->info.swapCreateKill.to = t->process->pid;
      e->info.swapCreateKill.lineNumber = 0;
      e->info.swapCreateKill.fileNameOffset = 0;
      e->info.swapCreateKill.euId = t->cpu;
      e->info.swapCreateKill.reserved0 = 0;
      e->info.swapCreateKill.reserved1 = 0;
   }
   return e;
}

static MonitorEventInfo *
add_error_info(MonitorGetTraceMultipleReply *reply, int oserror,
               const struct trace_ose_error *t)
{
   MonitorEventInfo *e;

   e = add_event_info(reply, MONITOR_GET_ERROR_TRACE_REPLY,
                      sizeof(MonitorErrorInfo));
   if (e != NULL)
   {
      e->info.error.from = t->process->pid;
      e->info.error.error = t->ecode;
      e->info.error.extra = t->extra;
      e->info.error.exec = oserror;
      e->info.error.lineNumber = 0;
      e->info.error.fileNameOffset = 0;
      e->info.error.euId = t->cpu;
      e->info.error.reserved0 = 0;
      e->info.error.reserved1 = 0;
   }
   return e;
}

static MonitorEventInfo *
add_reset_info(MonitorGetTraceMultipleReply *reply, const struct trace_system_start *t)
{
   MonitorEventInfo *e;

   e = add_event_info(reply, MONITOR_GET_RESET_TRACE_REPLY,
                      sizeof(MonitorResetInfo));
   if (e != NULL)
   {
      e->info.reset.warm = t->warm_flag;
      e->info.reset.lineNumber = 0;
      e->info.reset.fileNameOffset = 0;
      e->info.reset.reserved0 = 0;
      e->info.reset.reserved1 = 0;
      e->info.reset.reserved2 = 0;
   }
   return e;
}

static MonitorEventInfo *
add_alloc_free_buf_info(MonitorGetTraceMultipleReply *reply, SIGSELECT sigNo,
                        const struct trace_alloc_free_buf *t)
{
   MonitorEventInfo *e;

   e = add_event_info(reply, sigNo, sizeof(MonitorAllocFreeBufInfo));
   if (NULL != e)
   {
      e->info.allocFreeBuf.from = t->process->pid;
      e->info.allocFreeBuf.signalNumber = *(const uint32_t *)t->data;
      e->info.allocFreeBuf.signalSize = t->requested_size;
      e->info.allocFreeBuf.signalAddress = t->address;
      e->info.allocFreeBuf.lineNumber = 0;
      e->info.allocFreeBuf.fileNameOffset = 0;
      e->info.allocFreeBuf.euId = t->cpu;
      e->info.allocFreeBuf.reserved0 = 0;
   }
   return e;
}

static MonitorEventInfo *
add_bind_info(MonitorGetTraceMultipleReply *reply, const struct trace_bind *t)
{
   MonitorEventInfo *e;

   e = add_event_info(reply, MONITOR_GET_BIND_TRACE_REPLY,
                      sizeof(MonitorBindInfo));
   if (e != NULL)
   {
      e->info.bind.bound = t->process->pid;
      e->info.bind.fromEuId = t->from_eu;
      e->info.bind.toEuId = t->to_eu;
      e->info.bind.lineNumber = 0;
      e->info.bind.fileNameOffset = 0;
      e->info.bind.reserved0 = 0;
      e->info.bind.reserved1 = 0;
   }
   return e;
}

static MonitorEventInfo *
add_user_info(MonitorGetTraceMultipleReply *reply, const struct trace_user_event *t)
{
   MonitorEventInfo *e;
   size_t size = sizeof(MonitorUserInfo) + t->event_size;

   e = add_event_info(reply, MONITOR_GET_USER_TRACE_REPLY, size);

   if (e != NULL)
   {
      e->info.user.from = t->process->pid;
      e->info.user.eventNumber = t->event_id;
      e->info.user.eventDataOffset = 0;
      e->info.user.eventDataCount = t->event_size;
      e->info.user.eventSize = t->requested_size;
      e->info.user.lineNumber = 0;
      e->info.user.fileNameOffset = 0;
      e->info.user.euId = t->cpu;
      e->info.user.reserved0 = 0;

      memcpy(e->info.user.data, &t->data, t->event_size);
   }
   return e;
}

static void
add_event_common(MonitorEventInfo *event, uint32_t ref, Offspring *offspring, union TRACE *te)
{
   /* Fill in generic trace info. */
   event->info.sendReceive.reference  = ref - 1;
   event->info.sendReceive.time.tick  = (uint32_t) offspring->cursor.tick;
   event->info.sendReceive.time.ntick = (uint32_t) offspring->cursor.time_count;
   if (te->trace_send_receive.abstime)
   {
      event->info.sendReceive.time.sec = te->trace_send_receive.abstime->abs_sec;
      event->info.sendReceive.time.sectick = te->trace_send_receive.abstime->abs_tick;
   }
   else
   {
      event->info.sendReceive.time.sec = (uint32_t) 0;
      event->info.sendReceive.time.sectick = (uint32_t) 0;
   }
}

static int
is_public_trace_event(int event_no)
{
   switch (event_no)
   {
      case TRACE_SEND:
      case TRACE_RECEIVE:
      case TRACE_SWAP:
      case TRACE_USER_ERROR:
      case TRACE_OSERROR:
      case TRACE_CREATE:
      case TRACE_KILL:
      case TRACE_SYSTEM_START:
      case TRACE_USER_EVENT:
#ifdef OSE_SMP
      case TRACE_BIND:
#endif /* OSE_SMP */
         return 1;

      case TRACE_NOT_USED:
      default:
         return 0;
   }
}

#if 0

static OSTIME ticks_to_nticks(OSTICK t0, OSTICK ut0, OSTICK t1, OSTICK ut1, OSTIME tlen);

static uint32_t
real_idbuf_contains(struct idbuf *buf, uint32_t pid)
{
   int        i;

   /* Search buffer for 'pid' */
   for (i = buf->i0; i != buf->i1; i = (i + 1) % MAX_IDBUF_SIZE)
   {
      if (buf->pid[i] == pid)
      {
         return 1;
      }
   }
   return 0;
}

static void
assert_get_trace_multiple_reply_failed(void)
{
   ramlog_printf("rmm: assert_get_trace_multiple_reply_failed\n");
}

static void
assert_get_trace_multiple_reply(MonitorGetTraceMultipleRequest *request,
                                MonitorGetTraceMultipleReply *reply)
{
   static struct idbuf pids;
   static uint32_t reference;
   static uint32_t tick;
   static uint32_t utick;

   MonitorEventInfoHeader *e  = reply->events;
   MonitorEventInfoHeader *e1 = reply->events + reply->eventsCount;

   if (request->handle == 0)
   {
      idbuf_init(&pids);
      reference = 0;
      tick = 0;
      utick = 0;
   }

   while (e < e1)
   {
      MonitorEventInfo *event = (void *)e;

      if (event->info.sigNo == MONITOR_GET_PROCESS_TRACE_REPLY)
      {
         MonitorProcessInfo *info = &event->info.process;

         /* Check for reoccurring processes */
         if (idbuf_contains(&pids, info->pid))
            assert_get_trace_multiple_reply_failed();
      }
      else
      {
         MonitorLossNotify *info = &event->info.loss;

         if (event->info.sigNo == MONITOR_GET_LOSS_TRACE_REPLY)
         {
            idbuf_init(&pids);
         }
         else
         {
            /* Check for timestamps going backwards */
            if (tick && info->time.tick
                && 10000000 < ticks_to_nticks(tick, utick,
                                          info->time.tick, info->time.ntick,
                                          system_tick()))
            {
               assert_get_trace_multiple_reply_failed();
            }

            /* Check for references going backwards */
            if (reference && info->reserved0 < reference)
               assert_get_trace_multiple_reply_failed();

            reference = info->reserved0;

            switch (event->info.sigNo)
            {
               case MONITOR_GET_SEND_TRACE_REPLY:
               case MONITOR_GET_RECEIVE_TRACE_REPLY:
               {
                  MonitorSendReceiveInfo *info = &event->info.sendReceive;

                  if (!real_idbuf_contains(&pids, info->signalSender))
                     assert_get_trace_multiple_reply_failed();
                  if (!real_idbuf_contains(&pids, info->signalAddressee))
                     assert_get_trace_multiple_reply_failed();

                  break;
               }
               case MONITOR_GET_SWAP_TRACE_REPLY:
               {
                  MonitorSwapCreateKillInfo *info = &event->info.swapCreateKill;

                  if (!real_idbuf_contains(&pids, info->from))
                     assert_get_trace_multiple_reply_failed();
                  if (!real_idbuf_contains(&pids, info->to))
                     assert_get_trace_multiple_reply_failed();

                  break;
               }
               case MONITOR_GET_CREATE_TRACE_REPLY:
               case MONITOR_GET_KILL_TRACE_REPLY:
               {
                  MonitorSwapCreateKillInfo *info = &event->info.swapCreateKill;

                  if (!real_idbuf_contains(&pids, info->to))
                     assert_get_trace_multiple_reply_failed();

                  break;
               }
               case MONITOR_GET_ERROR_TRACE_REPLY:
               {
                  MonitorErrorInfo *info = &event->info.error;

                  if (!real_idbuf_contains(&pids, info->from))
                     assert_get_trace_multiple_reply_failed();

                  break;
               }
               case MONITOR_GET_RESET_TRACE_REPLY:
                  break;
               case MONITOR_GET_BIND_TRACE_REPLY:
               {
                  MonitorBindInfo *info = &event->info.bind;

                  if (!real_idbuf_contains(&pids, info->bound))
                     assert_get_trace_multiple_reply_failed();

                  break;
               }
               case MONITOR_GET_ALLOC_TRACE_REPLY:
               case MONITOR_GET_FREE_BUF_TRACE_REPLY:
               {
                  MonitorAllocFreeBufInfo *info = &event->info.allocFreeBuf;

                  if (!real_idbuf_contains(&pids, info->from))
                     assert_get_trace_multiple_reply_failed();

                  break;
               }
               default:
                  assert_get_trace_multiple_reply_failed();
                  break;
            }
         }

         tick = info->time.tick;
         utick = info->time.ntick;
      }

      e += event->header.size;
   }
}

#endif

static int
note_trace_loss(Offspring *offspring, int step, MonitorGetTraceMultipleReply *reply, int cur_entry)
{
   MonitorEventInfo *loss;

   assert(step > 1); /* Should only be called if we have lossed data. */

   /* Clear the trace process cache since we have wrapped and maybe pids
    * have wrapped to. */
   idbuf_init(&offspring->mtrace_pids);

   loss = add_event_info(reply, MONITOR_GET_LOSS_TRACE_REPLY,
                          sizeof(MonitorLossNotify));
   if (loss)
   {
      if (is_public_trace_event(((union TRACE *)offspring->cursor.trace_entry)->event_no))
      {
         loss->info.loss.time.tick  = (uint32_t) offspring->cursor.tick;
         loss->info.loss.time.ntick = (uint32_t) offspring->cursor.time_count;
         loss->info.loss.time.sec = 0; /* FIXME */
         loss->info.loss.time.sectick = 0;
      }
      else
      {
         /* We do not know the time of the loss. */
         loss->info.loss.time.tick = 0;
         loss->info.loss.time.ntick = 0;
         loss->info.loss.time.sec = 0;
         loss->info.loss.time.sectick = 0;
      }
#ifdef HAS_LOSS_REFERENCE
      loss->info.loss.reference = cur_entry - 1;
#else
      loss->info.loss.reserved0 = 0;
#endif
      loss->info.loss.lostSize = step;
      commit_event_info(reply);
      offspring->mget_loss = 0;
      return 1;
   }
   else
   {
      offspring->mget_loss = step;
   }
   return 0;
}

int zzose_trace_read(TraceEntryState *cursor, int wait);

static int
handle_get_trace_multiple(Offspring *offspring, union SIGNAL *sig)
{
   MonitorGetTraceMultipleReply *reply;
   uint32_t                      cur_entry;
   uint32_t                      reply_count;
   int                           step = 1;

   assert(sig);
   assert_offspring(offspring);
   assert(MONITOR_GET_TRACE_MULTIPLE_REQUEST == sig->sig_no);

   if (offspring->mget_reply)
   {
      uint32_t reply_size;

      reply = offspring->mget_reply;

      reply_size = sigsize((union SIGNAL **)&reply);
      reply_count = (reply_size - sizeof(MonitorGetTraceMultipleReply))
                    / sizeof(MonitorEventInfoHeader);
   }
   else
   {
      uint32_t reply_size;

      reply_size = sizeof(MonitorGetTraceMultipleReply) +
                   sizeof(MonitorEventInfoHeader) * rw.trace_mget_size;
      reply_count = rw.trace_mget_size;

      reply = (MonitorGetTraceMultipleReply *)
                 alloc(reply_size, MONITOR_GET_TRACE_MULTIPLE_REPLY);
      reply->reserved0   = 0;
      reply->enabled  = (uint32_t)rw.trace_enabled;
      reply->eventsCount = 0;

      offspring->mget_reply = reply;
   }


   if (0 == sig->get_trace_mult_request.handle)
   {
      /* Just in case the client was reading the trace make sure we don't have
       * any locked entries in the trace. */
      ose_trace_close(&offspring->cursor);

      /* Clear the trace process cache. */
      idbuf_init(&offspring->mtrace_pids);

      /* Do not mind old losses */
      offspring->mget_loss = 0;

      if (ose_trace_open(rw.trace, &offspring->cursor) != 0)
      {
         /* Not a valid trace */
         cur_entry = 0;
         reply->status = MONITOR_STATUS_KERNEL_ERROR;
         goto done;
      }
      else
      {
         cur_entry = 1;
         offspring->mget_read = 1;
         reply->status = MONITOR_STATUS_OK;
      }
   }
   else
   {
	 int loss_got_handle_number = 0;

     if (offspring->mget_loss > 0)
     {
       if (note_trace_loss(offspring, offspring->mget_loss, reply, sig->get_trace_mult_request.handle) > 0)
       {
         /* The loss event gets the first reference number */
         loss_got_handle_number = 1;
       }
     }

      if (NULL == offspring->cursor.base)
      {
         /* The trace is not open for reading. */
         cur_entry = 0;
         reply->status = MONITOR_STATUS_REQUEST_INVALID;
         goto done;
      }
      else
      {
         cur_entry = sig->get_trace_mult_request.handle;
         if ( (cur_entry += loss_got_handle_number) == 0 )
         {
        	 ++cur_entry;
         }
         reply->status = MONITOR_STATUS_OK;
      }
   }

   /* Display each entry until end of trace. */
   for (;;)
   {
      union TRACE      *te;
      MonitorEventInfo *event = NULL;


      if (offspring->mget_read)
      {
         step = zzose_trace_read(&offspring->cursor, 1);

         if (0 == step)
         {
            /* No more entries, is there room for more */
            if (reply->eventsCount < (reply_count - 64))
            {
               if (offspring->mget_tmo == -1
                   || sig->get_trace_mult_request.timeout == 0)
               {
                  /* Timeout expired (or no timeout) so send reply. */
                  goto done;
               }
               else if (offspring->mget_tmo == 0 &&
                        sig->get_trace_mult_request.timeout > 0)
               {
                  /* Initial call and a timeout requested. */
                  if (sig->get_trace_mult_request.timeout > 0x7fffffff)
                  {
                     offspring->mget_tmo = 0x7fffffff;
                  }
                  else
                  {
                     offspring->mget_tmo = sig->get_trace_mult_request.timeout;
                  }
               }
               else
               {
                  if (sig->get_trace_mult_request.timeout == 0xffffffff)
                  {
                     /* Client would like to block forever so reload timeout */
                     offspring->mget_tmo = 0x7fffffff;
                  }
               }
               sig->get_trace_mult_request.handle = cur_entry;
               offspring->mget_req = sig;
               rw.pending_req++;

               /* Return non zero to indicate that we did not send reply */
               return -1;
            }
            else
            {
               /* Reply with what we got */
               goto done;
            }
         } /* if (0 == step) */
         else if (step > 1)
         {
            /* We have losset data */
           if (note_trace_loss(offspring, step, reply, cur_entry)>0)
           {
	         /* Reference shall count events including the loss events */
	         if (++cur_entry == 0)
	         {
	    	   ++cur_entry;
	         }
	       }
         }
      }

      te = (union TRACE *) offspring->cursor.trace_entry;

      /* Build a signal depending on the event. */
      switch (te->event_no)
      {
         case TRACE_SEND:
         case TRACE_RECEIVE:
            {
               struct trace_send_receive *t = &te->trace_send_receive;

               if (!idbuf_contains(&offspring->mtrace_pids, t->from->pid))
               {
                  if (!add_process_info(reply, &offspring->mtrace_pids, t->from))
                  {
                     break;
                  }
               }
               if (!idbuf_contains(&offspring->mtrace_pids, t->to->pid))
               {
                  if (!add_process_info(reply, &offspring->mtrace_pids, t->to))
                  {
                     break;
                  }
               }
               event = add_send_receive_info(reply, te->event_no == TRACE_SEND ?
                                             MONITOR_GET_SEND_TRACE_REPLY :
                                             MONITOR_GET_RECEIVE_TRACE_REPLY, t);
               break;
            }

#ifdef HAS_TIMEOUT_EVENT
         case TRACE_RECEIVE_TMO:
         {
            struct trace_receive_tmo *t = &te->trace_receive_tmo;

            if (!idbuf_contains(&offspring->mtrace_pids, t->process->pid))
            {
               if (!add_process_info(reply, &offspring->mtrace_pids, t->process))
               {
                  break;
               }
            }

            event = add_timeout_info(reply,
                                            MONITOR_GET_TIMEOUT_TRACE_REPLY,
                                            t);
            break;
         }
#endif

         case TRACE_SWAP:
            {
               struct trace_swap *t = &te->trace_swap;

               if (!idbuf_contains(&offspring->mtrace_pids, t->from_process->pid))
               {
                  if (!add_process_info(reply, &offspring->mtrace_pids, t->from_process))
                  {
                     break;
                  }
               }
               if (!idbuf_contains(&offspring->mtrace_pids, t->to_process->pid))
               {
                  if (!add_process_info(reply, &offspring->mtrace_pids, t->to_process))
                  {
                     break;
                  }
               }
               event = add_swap_info(reply, t);
               break;
            }

         case TRACE_USER_ERROR:
         case TRACE_OSERROR:
            {
               struct trace_ose_error *t = &te->trace_ose_error;

               if (!idbuf_contains(&offspring->mtrace_pids, t->process->pid))
               {
                  if (!add_process_info(reply, &offspring->mtrace_pids, t->process))
                  {
                     break;
                  }
               }

               event = add_error_info(reply, (te->event_no == TRACE_OSERROR), t);
               break;
            }

         case TRACE_CREATE:
         case TRACE_KILL:
            {
               struct trace_create_kill *t = &te->trace_create_kill;

               if (!idbuf_contains(&offspring->mtrace_pids, t->process->pid))
               {
                  if (!add_process_info(reply, &offspring->mtrace_pids, t->process))
                  {
                     break;
                  }
               }

               event = add_create_kill_info(reply,
                                            te->event_no == TRACE_CREATE ?
                                            MONITOR_GET_CREATE_TRACE_REPLY :
                                            MONITOR_GET_KILL_TRACE_REPLY,
                                            t);
               break;
            }

         case TRACE_SYSTEM_START:
            {
               struct trace_system_start *t = &te->trace_system_start;
               /* We are now entering a new life so the cached pids aren't
                * valid anymore */
               idbuf_init(&offspring->mtrace_pids);
               event = add_reset_info(reply, t);
               break;
            }

         case TRACE_ALLOC:
         case TRACE_FREE_BUF:
         {
            struct trace_alloc_free_buf *t = &te->trace_alloc_free_buf;

            if (!idbuf_contains(&offspring->mtrace_pids, t->process->pid))
            {
               if (!add_process_info(reply, &offspring->mtrace_pids, t->process))
               {
                  break;
               }
            }

            event = add_alloc_free_buf_info(reply,
                                            te->event_no == TRACE_ALLOC ?
                                            MONITOR_GET_ALLOC_TRACE_REPLY :
                                            MONITOR_GET_FREE_BUF_TRACE_REPLY,
                                            t);
            break;
         }

         case TRACE_BIND:
            {
               struct trace_bind *t = &te->trace_bind;

               if (!idbuf_contains(&offspring->mtrace_pids, t->process->pid))
               {
                  if (!add_process_info(reply, &offspring->mtrace_pids, t->process))
                  {
                     break;
                  }
               }

               event = add_bind_info(reply, t);
               break;
         }

         case TRACE_USER_EVENT:
         {
            struct trace_user_event *t = &te->trace_user_event;

            if (!idbuf_contains(&offspring->mtrace_pids, t->process->pid))
            {
               if (!add_process_info(reply, &offspring->mtrace_pids, t->process))
               {
                  break;
               }
            }

            event = add_user_info(reply, t);

            break;
         }

         case TRACE_NOT_USED:
            continue;

         default:
            /* Ignore unknown entries */
            continue;
      }

      /* Out of room in reply signal */
      if (event == NULL)
      {
         /* Will this event ever fit in a reply signal? */
         if (reply->reserved0 - reply->eventsCount >= rw.trace_mget_size)
         {
            ramlog_printf("rmm: event too large to fit in a multiple trace "
               "reply encountered!\n");

            /* Undo this large event's contribution to the reply size */
            reply->reserved0 = reply->eventsCount;

            /* Note the loss and try adding the next entry to the trace*/
            if (note_trace_loss(offspring, offspring->cursor.entry_size, reply, cur_entry) > 0)
            {
	          /* Reference shall count events including the loss events */
              if (++cur_entry == 0)
              {
                ++cur_entry;
              }
              offspring->mget_read = 1;
              continue;
            }
         }

         /* No room for this event in this reply, so we send the response now */
         offspring->mget_read = 0;
         break;
      }

      add_event_common(event, cur_entry, offspring, te);

      commit_event_info(reply);

      offspring->mget_read = 1;

      if (++cur_entry == 0)
      {
         ++cur_entry;
      }
   }

done:
#if 0
   assert_get_trace_multiple_reply(&sig->get_trace_mult_request, reply);
#endif

   reply->handle = cur_entry;
   set_sigsize((union SIGNAL **)&reply, sizeof(MonitorGetTraceMultipleReply) +
               reply->eventsCount * sizeof(MonitorEventInfoHeader));
   send((union SIGNAL **)&reply, offspring->pid);
   offspring->mget_reply = NULL;
   offspring->mget_req = NULL;
   offspring->mget_tmo = 0;

   return 0;
}


static void
check_pending_mgets(void)
{
   static int last_call_valid;
   static OSTIME last_call;
   OSTIME now = get_ticks();
   OSTIME delta_time;
   Offspring *c;

   if (last_call_valid)
   {
      delta_time = now - last_call;
      delta_time *= (system_tick() / 1000);
   }
   else
   {
      delta_time = 0;
   }
   last_call = now;

   /* Clear pending_req, will be reset in handle_get_trace_multiple */
   rw.pending_req = 0;

   for (c = list_head(rw.offsprings);
        c != NULL;
        c = list_next(rw.offsprings, &c->link))
   {
      if (c->mget_req != NULL)
      {
         union SIGNAL *mget_req = c->mget_req;
         if (c->mget_tmo > 0)
         {
            c->mget_tmo -= delta_time;
            if (c->mget_tmo <= 0) c->mget_tmo = -1; /* signal timeout */
         }
         if (handle_get_trace_multiple(c, mget_req) == 0)
         {
            /* Request completed */
            free_buf(&mget_req);
         }
      }
   }
   last_call_valid = rw.pending_req > 0;
}

static void     set_evact_monitor_scope(int scope_count, const char *scope_strings);
static void     set_evact_catch_scope(int scope_count, const char *scope_strings);

/**
 * @function handle_attach_request
 * @brief Handle a request to set the application scope.
 */
static int
handle_attach_request(Offspring * offspring, union SIGNAL *sig)
{
   uint32_t        i;
   union SIGNAL   *reply;
   int             event_debugger = 0;
   MonitorAttachRequest *req = &sig->attach_request;
   PROCESS        ose_mm_;

   reply = alloc(sizeof (MonitorAttachReply)
                 + sizeof (notify_sigs)
                 + sizeof (intercept_sigs)
                 + sizeof (trace_sigs)
                 - sizeof (uint32_t)
                 , MONITOR_ATTACH_REPLY);

   /* Let the offspring cache one process more than the RMM do due to that there
    * are events that refers to two processes. */
   reply->attach_reply.processCacheCount = MAX_IDBUF_SIZE + 1;
   reply->attach_reply.reserved0 = 0;
   reply->attach_reply.reserved1 = 0;
   reply->attach_reply.reserved2 = 0;

   reply->attach_reply.notifyOffset
      = 0;
   reply->attach_reply.notifyCount
      = sizeof (notify_sigs) / sizeof (SIGSELECT);
   memcpy(&reply->attach_reply.sigs[reply->attach_reply.notifyOffset],
          notify_sigs, sizeof (notify_sigs));
   reply->attach_reply.interceptOffset
      = reply->attach_reply.notifyOffset + reply->attach_reply.notifyCount;
   reply->attach_reply.interceptCount
      = sizeof (intercept_sigs) / sizeof (SIGSELECT);
   memcpy(&reply->attach_reply.sigs[reply->attach_reply.interceptOffset],
          intercept_sigs, sizeof (intercept_sigs));
   reply->attach_reply.traceOffset
      = reply->attach_reply.interceptOffset + reply->attach_reply.interceptCount;
   reply->attach_reply.traceCount
      = sizeof (trace_sigs) / sizeof (SIGSELECT);
   memcpy(&reply->attach_reply.sigs[reply->attach_reply.traceOffset],
          trace_sigs, sizeof (trace_sigs));

   /* Eclipse will clear pid cache on attach, so we should do so too */
   idbuf_init(&offspring->notify_pids);

   /* Make sure the given ID exists */
   switch (req->scope.type)
   {
      case MONITOR_SCOPE_ID:
      {
         enum PROCESS_TYPE ptype = get_ptype(req->scope.id);

         if (ptype != OS_PRI_PROC
             && ptype != OS_BG_PROC && ptype != OS_INT_PROC && ptype != OS_TI_PROC && ptype != OS_PHANTOM)
         {
            reply->attach_reply.status = MONITOR_STATUS_SCOPE_INVALID;
            send(&reply, offspring->pid);
            return 0;
         }
         break;
      }
      case MONITOR_SCOPE_BLOCK_ID:
      {
         enum PROCESS_TYPE ptype = get_ptype(req->scope.id);

         if (ptype != OS_BLOCK)
         {
            reply->attach_reply.status = MONITOR_STATUS_SCOPE_INVALID;
            send(&reply, offspring->pid);
            return 0;
         }
         break;
      }
      case MONITOR_SCOPE_SEGMENT_ID:
      {
         struct OS_pid_list *bids = get_bid_list(0);

         assert(bids);

         for (i = 0; 1; i++)
         {
            if (i == bids->count)
            {
               free_buf((void *) &bids);
               reply->attach_reply.status = MONITOR_STATUS_SCOPE_INVALID;
               send(&reply, offspring->pid);
               return 0;
            }
            if (get_segid(bids->list[i]) == req->scope.id)
            {
               free_buf((void *) &bids);
               break;
            }
         }
         break;
      }
      case MONITOR_SCOPE_SYSTEM:
         break;

      default:
         reply->attach_reply.status = MONITOR_STATUS_SCOPE_INVALID;
         send(&reply, offspring->pid);
         return 0;
   }

   /* Create dynamic region if dynamic trace size is > 0 and if it's not
    * already created.
    */
   if (rw.dynamic_trace_size > 0 && rw.dr_trace == NULL)
   {
      if (hunt("ose_memory_mgr", 0, &ose_mm_, NULL))
      {
      	 /* First try to create a SASE memory region for the trace. */
         rw.dr_trace = create_dynamic_trace(ose_mm_, rw.dynamic_trace_size,
                                         &rw.dr_size, "RAM_SASE");
         if (!rw.dr_trace)
         {
      	    /* If there is no MMU there is no RAM_SASE memory so try a plain
      	     * RAM region. */
            rw.dr_trace = create_dynamic_trace(ose_mm_, rw.dynamic_trace_size,
					       &rw.dr_size, "RAM");
         }

         if (rw.dr_trace != NULL)
         {
            ramlog_printf("rmm: dynamic trace adress = 0x%X and size = %u\n",
                          rw.dr_trace, rw.dr_size);

            /* Create the trace for the dynamic trace region */
            ose_trace_create(rw.dr_trace, rw.dr_size, SYS_TRACE_NAME,
               (GetTimeFuncPointer) get_systime, 1000000);

            /* Move the static trace pointer to the dynamic trace area */
            BKL_LOCK();
            rw.trace = rw.dr_trace;
            rw.trace_size = rw.dr_size;
            BKL_UNLOCK();
         }
         else
         {
            ramlog_printf("rmm: could not create a dynamic trace buffer!\n");
         }
      }
      else
      {
         ramlog_printf("rmm: ose_memory_mgr not found!\n");
      }
   }

   /* Check system or source */
   for (i = 0; i < req->tagsCount; i += req->tags[i].tag.count)
   {
      switch (req->tags[i].tag.type)
      {
         case MONITOR_ATTACH_TAG_EVENT_LOCK:
            event_debugger = 1;
            break;
         case MONITOR_ATTACH_TAG_KEEP_EVENT_TRACEPOINTS:
            offspring->keep_trace_actions = 1;
            break;
         case MONITOR_ATTACH_TAG_EVENT_SIGSELECT:
            offspring->handles_new_events = 1;
         default:
            break;
      }
   }

   /* System event debugger */
   if (event_debugger)
   {
      if (rw.evact_lock_owner == NULL || rw.evact_lock_owner == offspring)
      {
         char            scope[5 + 10 + 1];     /* strlen("*:*:*") + strlen("0x00000000") + 1 */
         char            intercept[5 + 10 + 1]; /* strlen("*:*:*") + strlen("0x00000000") + 1 */

         switch (req->scope.type)
         {
            case MONITOR_SCOPE_ID:
               sprintf(scope, "#%#x", req->scope.id);
               break;
            case MONITOR_SCOPE_BLOCK_ID:
               sprintf(scope, "#%#x:*", req->scope.id);
               break;
            case MONITOR_SCOPE_SEGMENT_ID:
               sprintf(scope, "#%#x:*:*", req->scope.id);
               break;
            case MONITOR_SCOPE_SYSTEM:
               sprintf(scope, "*:*:*");
               break;
            default:
               reply->attach_reply.status = MONITOR_STATUS_SCOPE_INVALID;
               send(&reply, offspring->pid);
               return 0;
         }
         switch (req->intercept.type)
         {
            case MONITOR_SCOPE_ID:
               sprintf(intercept, "#%#x", req->intercept.id);
               break;
            case MONITOR_SCOPE_BLOCK_ID:
               sprintf(intercept, "#%#x:*", req->intercept.id);
               break;
            case MONITOR_SCOPE_SEGMENT_ID:
               sprintf(intercept, "#%#x:*:*", req->intercept.id);
               break;
            case MONITOR_SCOPE_SYSTEM:
               sprintf(intercept, "*:*:*");
               break;
            default:
               reply->attach_reply.status = MONITOR_STATUS_SCOPE_INVALID;
               send(&reply, offspring->pid);
               return 0;
         }

         /* Save the intercept scope */
         offspring->evact_scope = req->intercept;

         /* REVIEW: Should we really wipe actions here? Right now we don't
          * have a choice because the debugserver never reads actions, but
          * actions should survive client connection and disconnection, and
          * be presented to the user upon connection. */
         do_delete_all_actions(offspring->keep_trace_actions, False);
         rw.evact_lock_owner = offspring;
         set_evact_catch_scope(1, intercept);
         set_evact_monitor_scope(1, scope);
         enable_catch(offspring);
         set_monitor_status(offspring, 1);
      }
      else
      {
         reply->attach_reply.status = MONITOR_STATUS_SCOPE_OVERLAP;
         send(&reply, offspring->pid);
         return 0;
      }
   }

   /* Source code debugger */
   else
   {
#ifdef USE_EPM
      Offspring         *c;
      EpmStatus       epm_status;


      /* First, we need to verify that the requested scope does not
       * collide with any existing application scope. As we only allow
       * pure segment, block or process scopes, a scope will either
       * completely overlap or not overlap at all. */
      for (c = list_head(rw.offsprings); c != NULL; c = list_next(rw.offsprings, &c->link))
      {
         if (c == offspring && scopes_identical(&c->debug_scope, &req->scope))
         {
            /* Debugger is trying to set the same scope as it already
             * uses. Return OK and do nothing. */
            reply->attach_reply.status = MONITOR_STATUS_OK;
            send(&reply, offspring->pid);
            return 0;
         }

         /* Don't consider collisions with ourselves. */
         if (c != offspring && zzrmm_check_scope_collision(&c->debug_scope, &req->scope))
         {
            /* We're attempting to set a process application scope that
             * collides with an existing scope. This is an error. */
            reply->attach_reply.status = MONITOR_STATUS_SCOPE_OVERLAP;
            send(&reply, offspring->pid);
            return 0;
         }
      }

      if (MONITOR_SCOPE_SYSTEM == req->scope.type
          || (MONITOR_SCOPE_SEGMENT_ID == req->scope.type && 0 == req->scope.id))
      {
         /* No action needed */
      }
      else if ((MONITOR_SCOPE_SEGMENT_ID == req->scope.type) ||
               (MONITOR_SCOPE_BLOCK_ID == req->scope.type) || (MONITOR_SCOPE_ID == req->scope.type))
      {
         /* Attach to the pid */
         union SIGNAL   *s = alloc(sizeof (RmmScopeAttachSig),
                                   RMM_SCOPE_ATTACH_SIG);

         s->rmm_scope_attach_sig.offspring_pid = offspring->pid;
         offspring->scope_attach_ref = attach(&s, req->scope.id);
      }
      else
      {
         /* Illegal scope type */
         ramlog_printf("zzose_rmm: Detected illegal scope type: %d\n",
             req->scope.type);
         reply->attach_reply.status = MONITOR_STATUS_REQUEST_INVALID;
         send(&reply, offspring->pid);
         return 0;
      }

      /* Okay, there is no collision with any existing application scopes. */

      /* Unless full register saves are already enabled ... */
      if (zzdebug_regs_on == 0)
      {
         /* ... enable full register saves for debugging purposes. */
         zzdebug_regs_on = 1;

         /* Indicate that zzdebug_regs_on was modified by rmm. */
         modified_zzdebug_regs_on = 1;
      }

      offspring->debug_scope = req->scope;
      offspring->scope_intercepted = (OSBOOLEAN) 0;
#ifdef USE_REGDUMP_REGISTERS
      ensure_regdump_list_size(offspring);
#endif
      epm_status = ose_epm_init();
      if (EPM_OK != epm_status)
      {
         ramlog_printf("rmm: epm init failed with status %d\n", epm_status);
      }
#else
      reply->attach_reply.status = MONITOR_STATUS_REQUEST_INVALID;
      send(&reply, offspring->pid);
      return 0;

#endif /* USE_EPM */

   }

   reply->attach_reply.status = MONITOR_STATUS_OK;
   send(&reply, offspring->pid);
   return 0;
}

#ifdef USE_EPM

static void
intercept_scope(MonitorScope * scope)
{
   assert(scope);

   switch (scope->type)
   {
      case MONITOR_SCOPE_SYSTEM:
      {
         struct EfmiSegmentCursor pc;
         struct EfmiSegmentInfo pi;

         EfmiStatus      status = ose_efmi_get_first_segment(&pc,
                                                             &pi,
                                                             sizeof (pi));

         while (EFMI_OK == status)
         {
            zzkrn_dbg_intercept((PROCESS) pi.segment);
            status = ose_efmi_get_next_segment(&pc, &pi);
         }
         break;
      }

      case MONITOR_SCOPE_SEGMENT_ID:
      case MONITOR_SCOPE_BLOCK_ID:
      case MONITOR_SCOPE_ID:
         zzkrn_dbg_intercept(scope->id);
         break;

      case MONITOR_SCOPE_VOID:
         /* Do nothing. */
         break;

      default:
         assert(False);         /* Looks like an illegal scope */
         break;
   }
}

/**
 * @function remove_step
 * @brief Remove all single steps for a client.
 * @long Must be called with the dispatcher locked.
 */
static void
remove_steps(Offspring * offspring)
{
   assert_offspring(offspring);

   if (offspring->execution_points != NULL)
   {
      ExecutionPointInfo *ep;
      ExecutionPointInfo *next_ep;

      for (ep = list_head(offspring->execution_points); ep != NULL; ep = next_ep)
      {
         assert(ep->offspring == offspring);
         next_ep = list_next(offspring->execution_points, &ep->link);
         if (SINGLE_STEP == ep->type)
         {
            ose_epm_remove_execution_point(&(ep->ep));
            list_unlink(ep);
            free_ep(&ep);
         }
      }
   }
}

/**
 * @function remove_step
 * @brief Remove all single steps of a process and a client.
 * @long Must be called with the dispatcher locked.
 */
static void
remove_step(Offspring * offspring, uint32_t pid)
{
   assert_offspring(offspring);

   if (offspring->execution_points != NULL)
   {
      ExecutionPointInfo *ep;
      ExecutionPointInfo *next_ep;

      for (ep = list_head(offspring->execution_points); ep != NULL; ep = next_ep)
      {
         assert(ep->offspring == offspring);
         next_ep = list_next(offspring->execution_points, &ep->link);
         if ((SINGLE_STEP == ep->type) && (pid == ep->ep.pid))
         {
            ose_epm_remove_execution_point(&(ep->ep));
            list_unlink(ep);
            free_ep(&ep);
         }
      }
   }
}

#endif

static int
handle_set_intercept_enabled(Offspring *offspring, union SIGNAL *sig)
{
   union SIGNAL *rep = alloc(sizeof(MonitorSetInterceptEnabledReply),
                             MONITOR_SET_INTERCEPT_ENABLED_REPLY);

   assert_offspring(offspring);
   assert(sig->sig_no == MONITOR_SET_INTERCEPT_ENABLED_REQUEST);

   if (((0 == sig->set_intercept_enabled_request.enabled) ||
        (1 == sig->set_intercept_enabled_request.enabled)))
   {
      if (sig->set_intercept_enabled_request.enabled !=
          offspring->intercept_enabled)
      {
         /* Change the behavior for breakpoints. */
         offspring->intercept_enabled = (OSBOOLEAN)sig->set_intercept_enabled_request.enabled;

         /* If the client use actionpoints then change the behavior. */
         if (offspring == rw.evact_lock_owner)
         {
            if (offspring->intercept_enabled)
            {
               enable_catch(offspring);
            }
            else
            {
               LOCK_SWAP();
               disable_catch();
               UNLOCK_SWAP();
            }
         }
      }
      rep->set_intercept_enabled_reply.status = MONITOR_STATUS_OK;
   }
   else
   {
      rep->set_intercept_enabled_reply.status = MONITOR_STATUS_REQUEST_INVALID;
   }
   send(&rep, offspring->pid);
   return 0;
}

static int
handle_get_intercept_enabled(Offspring *offspring, union SIGNAL *sig)
{
   union SIGNAL *rep = alloc(sizeof(MonitorGetInterceptEnabledReply),
                             MONITOR_GET_INTERCEPT_ENABLED_REPLY);
   (void)sig;
   assert_offspring(offspring);
   assert(sig->sig_no == MONITOR_GET_INTERCEPT_ENABLED_REQUEST);

   rep->get_intercept_enabled_reply.status = MONITOR_STATUS_OK;
   rep->get_intercept_enabled_reply.enabled = offspring->intercept_enabled;
   rep->get_intercept_enabled_reply.reserved0 = 0;
   send(&rep, offspring->pid);
   return 0;
}

/**
 * @function handle_intercept_scope_request
 * @brief Handle a request to intercept a scope.
 */
static int
handle_intercept_scope_request(Offspring * offspring, union SIGNAL *sig)
{
   union SIGNAL   *reply;

   reply = alloc(sizeof (MonitorInterceptScopeReply), MONITOR_INTERCEPT_SCOPE_REPLY);

   if (!offspring->intercept_enabled)
   {
      /* Client disabled intercept. */
      reply->intercept_scope_reply.status = MONITOR_STATUS_REQUEST_INVALID;
   }
   /* System debugger */
   else if (offspring == rw.evact_lock_owner && offspring->debug_scope.type == MONITOR_SCOPE_VOID)
   {
      if ((sig->intercept_scope_request.scope.type == MONITOR_SCOPE_ID) &&
          (offspring->evact_scope.type != sig->intercept_scope_request.scope.type))
      {
         /* intercept the requested process if found in the scope */
         if (intercept_evact_scope_proc(sig->intercept_scope_request.scope.id))
         {
            reply->intercept_scope_reply.status = MONITOR_STATUS_OK;
         }
         else
         {
            /* The process is not, and have never been, part of the scope */
            reply->intercept_scope_reply.status = MONITOR_STATUS_SCOPE_INVALID;
          }
      }
      else
      {
         LOCK_SWAP();
         intercept_evact_scope();
         UNLOCK_SWAP();
         reply->intercept_scope_reply.status = MONITOR_STATUS_OK;
      }
   }
#ifdef USE_EPM
   /* Source code debugger:  */
   else if (offspring->debug_scope.type != MONITOR_SCOPE_VOID)
   {
      MonitorScope    intercepted_scope;
      MonitorScope    scope = sig->intercept_scope_request.scope;

      /* First we need to identify the intersection between between
       * the current application scope and the scope specified in this
       * request. */
      intercepted_scope = zzrmm_get_scope_intersection(&offspring->debug_scope, &scope);

      if (intercepted_scope.type == MONITOR_SCOPE_VOID)
      {
         /* This is an error. */
         reply->intercept_scope_reply.status = MONITOR_STATUS_SCOPE_EMPTY;
         ramlog_printf("rmm: intercept failed due to void scope\n");
         send(&reply, offspring->pid);
         return 0;
      }
      if ((intercepted_scope.type == offspring->debug_scope.type) && (intercepted_scope.id == offspring->debug_scope.id))
      {
         /* The client want to intercept the entire debug scope. */
         LOCK_SWAP();
         remove_steps(offspring);
         intercept_scope(&intercepted_scope);
         offspring->scope_intercepted = (OSBOOLEAN) 1;
         UNLOCK_SWAP();
         reply->intercept_scope_reply.status = MONITOR_STATUS_OK;
      }
      else if (intercepted_scope.type == MONITOR_SCOPE_ID)
      {
         /* The client want to intercept a specific process. */
         LOCK_SWAP();
         remove_step(offspring, intercepted_scope.id);
         intercept_scope(&intercepted_scope);
         UNLOCK_SWAP();
         reply->intercept_scope_reply.status = MONITOR_STATUS_OK;
      }
      else
      {
         /* The client want to intercept a subset of the application scope and
          * more than a single process. However we do not support that due to the
          * complexity of track what is suppoed to be intercepted when a new
          * process is created.
          */
         reply->intercept_scope_reply.status = MONITOR_STATUS_SCOPE_INVALID;
      }
   }
#endif
   /* Unattached client */
   else
   {
      reply->intercept_scope_reply.status = MONITOR_STATUS_REQUEST_INVALID;
   }

   send(&reply, offspring->pid);
   return 0;
}

#ifdef USE_EPM

/**
 * @function handle_get_registers_request
 * @brief Handle a request to read registers.
 */
static int
handle_get_registers_request(Offspring * offspring, union SIGNAL *sig)
{
   uint32_t        i;
   union SIGNAL   *reply;

   reply = alloc(sizeof (MonitorRegisterValues)
                 + sizeof (MonitorRegister) * sig->register_ids.registersCount, MONITOR_GET_REGISTERS_REPLY);

   reply->register_values.status = MONITOR_STATUS_OK;
   reply->register_values.pid = sig->register_ids.pid;
   reply->register_values.registersCount = sig->register_ids.registersCount;

   for (i = 0; i < sig->register_ids.registersCount; i++)
   {
      uint32_t        value;
      uint32_t        error = 1;
      struct cpu_regdump *reg_dump;

      reply->register_values.registers[i].id = sig->register_ids.registers[i];

#ifdef USE_REGDUMP_REGISTERS
      LOCK_SWAP();
      reg_dump = get_regdump(offspring, sig->register_ids.pid);

      if (NULL != reg_dump)
      {
         /* Read reg from dump! */
         error = ose_epm_get_reg(reg_dump, MONITOR_REGISTER_ID(sig->register_ids.registers[i]), (U32 *) & value);
      }
      UNLOCK_SWAP();
      if (error)
#endif
      {
         /* Read reg from EFMI. */
         error = ose_efmi_get_register(sig->register_ids.pid,
                                       MONITOR_REGISTER_ID(sig->register_ids.registers[i]),
                                       (struct EfmiRegisterInfo *) &value);
      }
      if (error)
      {
         reply->register_values.registers[i].id |= MONITOR_REGISTER_INVALID;
         reply->register_values.registers[i].value = 0xffffffff;
      }
      else
      {
         reply->register_values.registers[i].value = value;
      }
   }
   send(&reply, offspring->pid);
   return 0;
}

/**
 * @function handle_set_registers_request
 * @brief Handle a request to write registers.
 */
static int
handle_set_registers_request(Offspring * offspring, union SIGNAL *sig)
{
   uint32_t        i;
   union SIGNAL   *reply;

   reply = alloc(sizeof (MonitorRegisterIds)
                 + sizeof (MonitorRegisterId) * sig->register_values.registersCount, MONITOR_SET_REGISTERS_REPLY);

   reply->register_ids.status = MONITOR_STATUS_OK;
   reply->register_ids.pid = sig->register_values.pid;
   reply->register_ids.registersCount = sig->register_values.registersCount;

   for (i = 0; i < sig->register_ids.registersCount; i++)
   {
      uint32_t        error = 1;
      struct cpu_regdump *reg_dump;

      reply->register_ids.registers[i] = sig->register_values.registers[i].id;

#ifdef USE_REGDUMP_REGISTERS
      LOCK_SWAP();
      reg_dump = get_regdump(offspring, sig->register_values.pid);

      if (NULL != reg_dump)
      {
         error = ose_epm_set_reg(reg_dump,
                                 MONITOR_REGISTER_ID(sig->register_values.registers[i].id),
                                 sig->register_values.registers[i].value);
      }
      UNLOCK_SWAP();
#endif
      if (error)
      {
         reply->register_ids.registers[i] |= MONITOR_REGISTER_INVALID;
      }
   }
   send(&reply, offspring->pid);
   return 0;
}
#endif

/**
 * @function handle_set_memory_request
 * @brief Handle a request to write memory.
 */
static int
handle_set_memory_request(Offspring * offspring, union SIGNAL *sig)
{
   union SIGNAL   *reply;
   PROCESS         pid = sig->set_memory_request.pid;
   uint32_t        address = sig->set_memory_request.address;
   uint32_t        size = sig->set_memory_request.dataCount;
   enum PROCESS_TYPE ptype;

   reply = alloc(sizeof (struct MonitorSetMemoryReply), MONITOR_SET_MEMORY_REPLY);

   reply->set_memory_reply.address = address;

   ptype = (enum PROCESS_TYPE) zzkrn_get_pid_info(pid, 9);

   if (pid != 0 && (ptype == OS_ZOOMBIE || ptype == OS_ILLEGAL))
   {
      reply->set_memory_reply.status = MONITOR_STATUS_SCOPE_INVALID;
      reply->set_memory_reply.size = 0;
   }
   else
   {
      if (set_mem(pid, sig->set_memory_request.data, address, size))
      {
         reply->set_memory_reply.status = MONITOR_STATUS_MEMORY_SET_ERROR;
         reply->set_memory_reply.size = 0;
      }
      else
      {
         reply->set_memory_reply.pid = pid;
         reply->set_memory_reply.address = address;
         reply->set_memory_reply.status = MONITOR_STATUS_OK;
         reply->set_memory_reply.size = size;
      }
   }
   send(&reply, offspring->pid);
   return 0;
}

#ifdef USE_EPM

/**
 * @function epm_callback
 * @brief Called from EPM in context of a process that has hit an
 *  execution point.
 *
 * Note: This function is called with interrupts disabled.
 */
static int
epm_callback(const struct EpmExecutionPoint *execution_point,
             struct cpu_regdump *reg_dump, struct EfmiCurrent *efmi_current)
{
   /* Only consider priority and background processes. */
   if (efmi_current->type <= OS_BG_PROC)
   {
      struct debug_pcb *pcb = zzget_debugpcb(efmi_current->process);

      /* Do not do anything for system processes. */
      if (pcb && !(pcb->debug_flags & SYS_PROCESS))
      {
         ExecutionPointInfo *info;
         union SIGNAL   *sig;
         OSADDRESS       next_ep_pc;
         uint32_t        instance;
         Boolean         trigger_ep = False;
	 PROCESS pid;
	 EpmStatus       epm_status = EPM_OK;

         /* REVIEW: Restructure ExecutionPointInfo to avoid nonsense like this */
         info = (ExecutionPointInfo *) ((OSADDRESS) execution_point - offsetof(ExecutionPointInfo, ep));
         instance = info->instance;
         assert(instance > 0);
         assert(info->offspring);

	 pid = info->offspring->pid;

         /* There may NOT be any system calls before the call to
          * ose_epm_remove_execution_point. This is because interrupts may
          * not be enabled before the execution point removal. This callback
          * function is called with interrupts disabled. */
         if (info->type == SINGLE_STEP)
         {
            epm_status = ose_epm_remove_execution_point(&(info->ep));
            if (epm_status != EPM_OK)
            {
               ramlog_printf("rmm: epm_callback: could not remove EP status: %d\n",
                  epm_status);
               trigger_ep = True;
            }
            else
            {
               info->ep.address = 0;  /* Stepping EP not valid anymore. */

               if ((info->ep_type_info.ss_info.type & MONITOR_STEP_INSTRUCTION) == 0 &&
                   reg_dump->pc >= info->ep_type_info.ss_info.range_start &&
                   reg_dump->pc <= info->ep_type_info.ss_info.range_end)
               {
		  uint32_t type = info->ep_type_info.ss_info.type;

		  UNLOCK();

                  /* Step range requested and we're still within the range to be
                   * stepped over. */
                  epm_status = ose_epm_next_instruction(reg_dump, type, &next_ep_pc);
		  LOCK();
                  if (epm_status != EPM_OK)
                  {
                     /* Failed to get next address. Let RMM call debugger. */
                     ramlog_printf("rmm: epm_callback: failed to get next instruction"
                                   " address for step: status %d pc 0x%x\n", epm_status, reg_dump->pc);
                     trigger_ep = True;
                  }
                  else
                  {
		     /* Check that the offspring is still alive */
		     if (find_offspring_by_pid(pid) == NULL)
		     {
			ramlog_printf("rmm: epm_callback: lost client\n");
			trigger_ep = False;
		     }
		     else
		     {
                     info->ep.address = next_ep_pc;
                     epm_status = ose_epm_add_execution_point(&(info->ep));
                     if (epm_status != EPM_OK)
                     {
                        /* We failed to set the new step breakpoint. Best thing is
                         * to signal to the RMM that the step was completed and let
                         * the debugger attempt to resolve the situation. */
                        ramlog_printf("rmm: epm_callback: could not set intermediate"
                                      " step EP at address 0x%x status %d\n", next_ep_pc, epm_status);
                        trigger_ep = True;
                     }
                  }
               }
               }
               else
               {
                  /* This is a single-instruction single-step operation. Report
                   * to RMM. */
                  trigger_ep = True;
               }
            }
         }
         else
         {
            assert(info->type == BREAKPOINT);
            info->ep_type_info.bp_info.count--;
            if (info->ep_type_info.bp_info.count == 0)
            {
               info->ep_type_info.bp_info.count = info->ep_type_info.bp_info.original_count;
               trigger_ep = info->offspring->intercept_enabled;
            }
         }

         if (trigger_ep)
         {
            SIGSELECT select[] = { 1, RMM_CALLBACK_HIT };
            MonitorScope    scope;

            scope.type = info->catch_scope.type;
            scope.id = info->catch_scope.id;

#ifdef USE_REGDUMP_REGISTERS
            /* Save the register dump for this process. */
            add_regdump(info->offspring, reg_dump);
#endif
            
            if (MONITOR_SCOPE_ID != scope.type)
            {
               LOCK_SWAP();
               /* Intercept all but our self. */
               intercept_scope(&scope);
               info->offspring->scope_intercepted = (OSBOOLEAN) 1;
               UNLOCK_SWAP();
            }

            /* FIXME, possible race condition here, if client is lost before
             * interrupts are turned on, the offspring struct will be removed
             * before intercept below. We may end up in a bunch of
             * intercepted process, that are not debugged. */
            /* Note: The alloc below might turn on interrupts again (if we are a pri or bg proc). */
            sig = alloc(sizeof (RmmCallbackHit), RMM_CALLBACK_HIT);
            sig->rmm_callback_hit.ep_instance = instance;
	    sig->rmm_callback_hit.status = epm_to_rmm_status(epm_status);
            sig->rmm_callback_hit.info = info;

            /* Notify rmm */
            send(&sig, rw.zzose_rmm_);
            sig = receive_from(~0, select, rw.zzose_rmm_);
            /* Do nothing with the signal but free it */
            free_buf(&sig);
         }
      }
   }

   return 0;
}

static          Boolean
get_reg(Offspring * offspring, PROCESS pid, uint32_t reg_id, uint32_t * value)
{
   Boolean         read_ok;
   struct EfmiRegisterInfo reg_info;
   struct cpu_regdump *reg_dump;

#ifdef USE_REGDUMP_REGISTERS
   LOCK_SWAP();
   reg_dump = get_regdump(offspring, pid);
   if (NULL != reg_dump)
   {
      /* Read reg from dump! */
      read_ok = (ose_epm_get_reg(reg_dump, reg_id, (U32 *) value) == EPM_OK);
      UNLOCK_SWAP();
   }
   else
   {
      UNLOCK_SWAP();
#else
   {
#endif
      /* Read reg from EFMI. */
      read_ok = (ose_efmi_get_register(pid, reg_id, &reg_info) == EFMI_OK);
      *value = reg_info.value;
   }
   return read_ok;
}

/**
 * @function handle_step_process_request
 * @brief Handle a request to step a process or entire scope.
 */
static int
handle_step_process_request(Offspring * offspring, union SIGNAL *sig)
{
   union SIGNAL   *reply;
   struct OS_pcb  *pcb;
   OSADDRESS       next_ep_pc;
   EpmStatus       epm_status;
   ExecutionPointInfo *epi;
   MonitorScope    catch_scope;
   MonitorScope    scope = sig->step_process_request.scope;
   struct cpu_regdump *reg_dump;

   pcb = get_pcb(sig->step_process_request.pid);

   if (pcb == NULL)
   {
      reply = alloc(sizeof (struct MonitorStepProcessNotify), MONITOR_STEP_PROCESS_NOTIFY);
      reply->step_process_notify.pid = sig->step_process_request.pid;
      reply->step_process_notify.status = MONITOR_STATUS_ID_INVALID;
      send(&reply, offspring->pid);
      ramlog_printf("rmm: step failed due to illegal pid 0x%x\n", sig->step_process_request.pid);
      return 0;
   }

   if (~pcb->status & OS_INTERCEPTED)
   {
      free_buf((union SIGNAL **) &pcb);

      reply = alloc(sizeof (struct MonitorStepProcessNotify), MONITOR_STEP_PROCESS_NOTIFY);
      reply->step_process_notify.pid = sig->step_process_request.pid;
      reply->step_process_notify.status = MONITOR_STATUS_SCOPE_STATE_INVALID;
      send(&reply, offspring->pid);
      ramlog_printf("rmm: step failed due to process not intercepted\n");
      return 0;
   }
   free_buf((union SIGNAL **) &pcb);

   LOCK_SWAP();
   reg_dump = get_regdump(offspring, sig->step_process_request.pid);

   if (NULL != reg_dump)
   {
      struct cpu_regdump *reg_dump2;
      void *addr = (void *)reg_dump->pc;
      void *addr2;

      UNLOCK_SWAP();

      /* The process to be stepped currently stands on an execution
       * point. We should use the regdump and ose_epm_next_instruction
       * to calculate next address to set an execution point. */
      epm_status = ose_epm_next_instruction(reg_dump, sig->step_process_request.mode, &next_ep_pc);

      LOCK_SWAP();
      /* Recheck that we still have the same reg dump and pc */
      reg_dump2 = get_regdump(offspring, sig->step_process_request.pid);

      addr2 = (reg_dump2 != NULL)? (void *)reg_dump2->pc: NULL;

      UNLOCK_SWAP();

      if (epm_status != EPM_OK || addr != addr2)
      {
         reply = alloc(sizeof (struct MonitorStepProcessNotify), MONITOR_STEP_PROCESS_NOTIFY);
         reply->step_process_notify.pid = sig->step_process_request.pid;
         reply->step_process_notify.status = (epm_status != EPM_OK)?
               epm_to_rmm_status(epm_status): MONITOR_STATUS_SCOPE_STATE_INVALID;

         send(&reply, offspring->pid);
         ramlog_printf("rmm: step failed due to epm error %d or process state changed\n", epm_status);
	 return 0;
      }
   }
   else
   {
      struct EfmiRegisterInfo pc_reg_info;

      /* The process to be stepped is either swapped out or in a
       * system call. Use EFMI to get the user land pc. This will be
       * used to set an ep. */
      ose_efmi_get_register(sig->step_process_request.pid, ZZREG_PC, &pc_reg_info);
      next_ep_pc = (OSADDRESS) pc_reg_info.value;
      UNLOCK_SWAP();
   }

   /* Create the execution point list, if it does not already exist. */
   if (offspring->execution_points == NULL)
   {
      offspring->execution_points = list_init((void *) alloc(sizeof (List), 0));
   }

   for (epi = list_head(offspring->execution_points); epi != NULL; epi = list_next(offspring->execution_points, &epi->link))
   {
      assert(epi->offspring == offspring);
      if (SINGLE_STEP == epi->type && epi->ep.pid == sig->step_process_request.pid)
      {
         /* Remove current single step EP for this process */
         ose_epm_remove_execution_point(&(epi->ep));
         list_unlink(epi);
         free_ep(&epi);
         /* We are done, no more single steps for this process should exists. */
         break;
      }
   }

   /* Identify the intersection between the current application
    * scope and the catch scope specified in this request. */
   catch_scope = zzrmm_get_scope_intersection(&offspring->debug_scope, &scope);

   epi = alloc_ep();
   epi->ep.address = next_ep_pc;
   epi->ep.pid = sig->step_process_request.pid;
   epi->ep.flags = 0;
   epi->ep.cb = epm_callback;
   epi->offspring = offspring;
   epi->catch_scope = catch_scope;
   epi->type = SINGLE_STEP;
   epi->ep_type_info.ss_info.type = sig->step_process_request.mode;
   epi->ep_type_info.ss_info.range_start = (uint32_t) sig->step_process_request.start;
   epi->ep_type_info.ss_info.range_end = (uint32_t) sig->step_process_request.end;
   epi->valid = EP_VALID;

#ifdef USE_REGDUMP_REGISTERS
   ensure_regdump_list_size(offspring);
#endif
   
   epm_status = ose_epm_add_execution_point(&epi->ep);
   if (epm_status != EPM_OK)
   {
      free_ep(&epi);
      reply = alloc(sizeof (struct MonitorStepProcessNotify), MONITOR_STEP_PROCESS_NOTIFY);
      reply->step_process_notify.pid = sig->step_process_request.pid;
      reply->step_process_notify.status = epm_to_rmm_status(epm_status);

      send(&reply, offspring->pid);
      ramlog_printf("rmm: step failed due to epm error %d\n", epm_status);
      return 0;
   }

   LOCK_SWAP();
   list_insert_head(offspring->execution_points, &epi->link);
   UNLOCK_SWAP();
   offspring->scope_intercepted = (OSBOOLEAN) 0;

   resume_scope(offspring, &epi->catch_scope);     /* Resume catch scope. */
   return 0;
}

static void
handle_callback_hit(union SIGNAL *sig)
{
   union SIGNAL   *notify_sig;
   union SIGNAL   *reply_sig;
   uint32_t        pc;
   uint32_t        pid = sender(&sig);

   ExecutionPointInfo *info = sig->rmm_callback_hit.info;

   /* Check that the EP is still valid. */
   if ((EP_VALID == info->valid) && (info->instance == sig->rmm_callback_hit.ep_instance))
   {
      Offspring         *offspring = info->offspring;

      assert(offspring);

      intercept(sender(&sig));

#ifdef USE_REGDUMP_REGISTERS
      ensure_regdump_list_size(offspring);
#endif
      
      if (info->type == SINGLE_STEP)
      {
         /* Step process completed. */
         notify_sig = alloc(sizeof (MonitorStepProcessNotify), MONITOR_STEP_PROCESS_NOTIFY);
         notify_sig->step_process_notify.pid = pid;
         notify_sig->step_process_notify.pc = get_reg(info->offspring, pid, ZZREG_PC, &pc) == True ? pc : 0xFFFFFFFF;
	 notify_sig->step_process_notify.status = sig->rmm_callback_hit.status;

         /* Mark this ep as unused and put it on the ep free list. */
         info->ep.address = 0;
         list_unlink(&info->link);
         free_ep(&info);
      }
      else
      {
         /* Regular breakpoint. */
         notify_sig = alloc(sizeof (MonitorBreakpointFiredNotify), MONITOR_BREAKPOINT_FIRED_NOTIFY);
         notify_sig->breakpoint_fired_notify.id = info->ep_type_info.bp_info.bp_id;
         notify_sig->breakpoint_fired_notify.pid = pid;
      }
      send(&notify_sig, offspring->pid);
   }
   reply_sig = alloc(sizeof(SIGSELECT), RMM_CALLBACK_HIT);
   send(&reply_sig, sender(&sig));
}

/**
 * @function handle_set_bp_request
 * @brief Handle a request to set a breakpoint.
 */
static int
handle_set_breakpoint_request(Offspring * offspring, union SIGNAL *sig)
{
   ExecutionPointInfo *execution_point_info;
   EpmStatus       epm_status;
   union SIGNAL   *reply;
   MonitorScope    bp_scope;
   MonitorScope    catch_scope;
   MonitorScope    req_trig_scope = sig->set_breakpoint_request.scope;
   MonitorScope    req_catch_scope = sig->set_breakpoint_request.intercept;
   uint32_t        i = 0;
   uint32_t        count = 1;   /* Default value for breakpoint hit count. */
   uint32_t        flags = 0;   /* Default value for execution point flags */

   reply = alloc(sizeof (struct MonitorSetBreakpointReply), MONITOR_SET_BREAKPOINT_REPLY);

   /* First we need to identify the intersection between between
    * the current application scope and the scope specified in this
    * request. */
   bp_scope = zzrmm_get_scope_intersection(&offspring->debug_scope, &req_trig_scope);

   /* Identify the intersection between the current application
    * scope and the catch scope specified in this request. */
   catch_scope = zzrmm_get_scope_intersection(&offspring->debug_scope, &req_catch_scope);

   /* Create the execution point list, if it does not already exist. */
   if (offspring->execution_points == NULL)
   {
      offspring->execution_points = list_init((void *) alloc(sizeof (List), 0));
   }

   /* Take care of COUNT tag! */
   while (i < sig->set_breakpoint_request.tagsCount)
   {
      if (sig->set_breakpoint_request.tags[i++].tag.type == MONITOR_BREAKPOINT_TAG_IGNORE)
      {
         count = sig->set_breakpoint_request.tags[i++].num;
      }
   }

   /* If trig scope is system, we need to set an ep flag to
    * indicate that all processes should be stopped. */
   if (bp_scope.type == MONITOR_SCOPE_SYSTEM)
   {
      flags |= EPM_EP_ANY_PROCESS;
   }

   /* Make sure that trig scope is not empty. */
   if (bp_scope.type == MONITOR_SCOPE_VOID)
   {
      reply->set_breakpoint_reply.status = MONITOR_STATUS_SCOPE_EMPTY;
   }
   else
   {
      execution_point_info = alloc_ep();
      execution_point_info->ep.address = sig->set_breakpoint_request.address;
      execution_point_info->ep.pid = bp_scope.id;
      execution_point_info->ep.flags = flags;
      execution_point_info->ep.cb = epm_callback;
      execution_point_info->offspring = offspring;
      execution_point_info->catch_scope = catch_scope;
      execution_point_info->type = BREAKPOINT;
      rw.next_bp_id++;
      execution_point_info->ep_type_info.bp_info.bp_id = rw.next_bp_id;
      execution_point_info->ep_type_info.bp_info.original_count = count;
      execution_point_info->ep_type_info.bp_info.count = count;
      execution_point_info->valid = EP_VALID;

#ifdef USE_REGDUMP_REGISTERS
      ensure_regdump_list_size(offspring);
#endif
      
      epm_status = ose_epm_add_execution_point(&execution_point_info->ep);
      if (epm_status != EPM_OK)
      {
         ramlog_printf("rmm: failed to set breakpoint due to epm error: 0x%X\n", epm_status);
         free_ep(&execution_point_info);
         reply->set_breakpoint_reply.status = epm_to_rmm_status(epm_status);
      }
      else
      {
         LOCK_SWAP();
         list_insert_tail(offspring->execution_points, &execution_point_info->link);
         UNLOCK_SWAP();
         reply->set_breakpoint_reply.status = MONITOR_STATUS_OK;
         reply->set_breakpoint_reply.id = execution_point_info->ep_type_info.bp_info.bp_id;
      }
   }
   send(&reply, offspring->pid);
   return 0;
}

/**
 * @function handle_clear_breakpoint_request
 * @brief Handle a request to clear a breakpoint.
 */
static int
handle_clear_breakpoint_request(Offspring * offspring, union SIGNAL *sig)
{
   ExecutionPointInfo *info;
   union SIGNAL   *reply;

   reply = alloc(sizeof (struct MonitorClearBreakpointReply), MONITOR_CLEAR_BREAKPOINT_REPLY);

   for (info = list_head(offspring->execution_points);
        info != NULL; info = list_next(offspring->execution_points, &info->link))
   {
      if (info->type == BREAKPOINT && info->ep_type_info.bp_info.bp_id == sig->clear_breakpoint_request.id)
      {
         ose_epm_remove_execution_point(&(info->ep));
         list_unlink(info);
         free_ep(&info);
         reply->clear_breakpoint_reply.status = MONITOR_STATUS_OK;
         send(&reply, offspring->pid);
         return 0;
      }
   }

   reply->clear_breakpoint_reply.status = MONITOR_STATUS_REQUEST_INVALID;
   send(&reply, offspring->pid);
   return 0;
}

#endif

/**
 * @function handle_resume_scope_request
 * @brief Handle a request to resume the scope.
 */
static int
handle_resume_scope_request(Offspring * offspring, union SIGNAL *sig)
{
   union SIGNAL   *reply;

   reply = alloc(sizeof (struct MonitorResumeScopeReply), MONITOR_RESUME_SCOPE_REPLY);

   /* System debugger */
   if (offspring == rw.evact_lock_owner && offspring->debug_scope.type == MONITOR_SCOPE_VOID)
   {
      enable_catch(offspring);
      LOCK_SWAP();
      resume_evact_scope();
      UNLOCK_SWAP();
      reply->resume_scope_reply.status = MONITOR_STATUS_OK;
   }
#ifdef USE_EPM
   /* Source code debugger */
   else if (offspring->debug_scope.type != MONITOR_SCOPE_VOID)
   {
      MonitorScope    resumed_scope;
      MonitorScope    req_scope = sig->resume_scope_request.scope;

      /* First we need to identify the intersection between between
       * the current application scope and the scope specified in this
       * request. */
      resumed_scope = zzrmm_get_scope_intersection(&offspring->debug_scope, &req_scope);

      if (resumed_scope.type == MONITOR_SCOPE_VOID)
      {
         /* This is an error. */
         reply->resume_scope_reply.status = MONITOR_STATUS_SCOPE_EMPTY;
         send(&reply, offspring->pid);
         return 0;
      }
      if ((resumed_scope.type == offspring->debug_scope.type) && (resumed_scope.id == offspring->debug_scope.id))
      {
         /* The client want to resume the entire debug scope. */
         LOCK_SWAP();
         remove_steps(offspring);
         offspring->scope_intercepted = (OSBOOLEAN) 0;
         UNLOCK_SWAP();

         /* Reset event system intercept flag (if we own it) */
         if (offspring == rw.evact_lock_owner)
         {
            /* Allow more interceptions */
            enable_catch(offspring);
            /* Set the status flag for the breakpoint scope. */
            rw.break_scope_status = (OSBOOLEAN) 0;
            rw.precious->break_scope_status = (U8) rw.break_scope_status;
         }

         resume_scope(offspring, &resumed_scope);
         reply->resume_scope_reply.status = MONITOR_STATUS_OK;
      }
      else if (resumed_scope.type == MONITOR_SCOPE_ID)
      {
         PROCESS         pid = resumed_scope.id;

         if ((get_ptype(pid) == OS_BG_PROC) || (get_ptype(pid) == OS_PRI_PROC))
         {
            resume_scope(offspring, &resumed_scope);
            reply->resume_scope_reply.status = MONITOR_STATUS_OK;
         }
         else
         {
            reply->resume_scope_reply.status = MONITOR_STATUS_SCOPE_INVALID;
         }
      }
   }
#endif
   /* Unattached client */
   else
   {
      reply->resume_scope_reply.status = MONITOR_STATUS_REQUEST_INVALID;
   }

   send(&reply, offspring->pid);
   return 0;
}

/*
** Returns true or false depending on if it is possible to allocate the memory
** size given as the in-argument in this block
*/
static OSBOOLEAN
get_max_sigsize( uint32_t req_size )
{
  struct OS_pcb *pcb;
  uint32_t sigsize = 0;

  /* Get sigsize and free the block control block if we obtained one. */
  pcb = get_pcb( current_process() );
  if (pcb)
  {
    sigsize = pcb->max_sigsize;
    free_buf((union SIGNAL **) &pcb);
  }

  /* Return dependant on if we can allocate the requested size*/
  if (req_size <= sigsize)
  {
    return (OSBOOLEAN) 1;
  }
  else
  {
    return (OSBOOLEAN) 0;
  }
}


/*
 ** Free sample pool, but only if the signal pointer to the buffer is
 ** not NULL
 */
void
free_sample_pool(union SIGNAL ** pool)
{
   if (*pool != NULL)
   {
      uint32_t size = sigsize(pool);
      /*
       * Keep track of memory allocated in the pool and use the limit to avoid
       * that profiling uses all the memory if certain parameter sets are used.
       */
      atomic_sub(size, (atomic_t *)&rw.sample_pool_memory_used);
      if (((int)rw.sample_pool_memory_used) < 0)
      {
         ramlog_printf("rmm: internal error: rw.sample_pool_memory_used == %d, free size == %u\n",
               rw.sample_pool_memory_used, size);
         rw.sample_pool_memory_used = 0;
      }
      free_buf(pool);
   }
}

uint32_t
get_last_memory_error(void)
{
   return rw.last_memory_error;
}

/*
 ** Allocates signal if it is possible to allocate in the current pool . Otherwise
 **  returns NULL.
 */
union SIGNAL *
alloc_sample_pool(uint32_t size)
{
   if (!get_max_sigsize(size))
   {
      ramlog_printf("\nrmm: alloc_sample_pool failed to allocate %d bytes: No signal large enough\n",
               size);
      rw.last_memory_error = MONITOR_STATUS_MEMORY_EXHAUSTED;

      return NULL;
   }
   else
   {
      /*
       * Keep track of memory allocated in the pool and use the limit to avoid
       * that profiling uses all the memory if certain parameter sets are used.
       *
       * Return NULL, indicating that the allocation failed, if the limit is
       * exceeded
       */
      atomic_add(size, (atomic_t *)&rw.sample_pool_memory_used);
      if (rw.sample_pool_memory_used <= rw.sample_pool_memory_limit)
      {
         return alloc(size, 0);
      }
      else
      {
         ramlog_printf("\nrmm: alloc_sample_pool failed to allocate %u bytes: "
            "Sample pool memory [used: %u] [limit: %u]\n",
            size, rw.sample_pool_memory_used, rw.sample_pool_memory_limit);
         atomic_sub(size, (atomic_t *)&rw.sample_pool_memory_used);
         rw.last_memory_error = MONITOR_STATUS_MEMORY_LIMIT;
         return NULL;
      }
   }
}

/*
** This is a local version of the set_sigsize system call.
** We can't use the real one because that would require
** osemain.c to be present at all times.
**
** This function is needed by the profiling subsystem.
*/

static          OSBOOLEAN
lset_sigsize(union SIGNAL **sig, OSBUFSIZE newsize)
{
   return (OSBOOLEAN) get_sysconf((OSADDRESS) 22, (OSADDRESS) sig, (OSADDRESS) newsize, (OSADDRESS) 0);
}

/* function ose_rmm_set_user_report_handler
 * If defined, rmm will call the user specified function just before closing a
 * sample interval.
 */
void
ose_rmm_set_user_report_handler(void (*handler)(SIGSELECT))
{
   rw.user_report_function = handler;
}


/* function ose_rmm_set_hw_trace_handlers 
 * If defined, rmm will call the user defined functions if action enable/disable hardware 
 * trace is specified. Action enable/disable hardware trace is avaliable for all events. 
 */
uint32_t
ose_rmm_set_hw_trace_handlers(void (*enbl_handler)(void), void (*dsbl_handler)(void))
{
   LOCKSTATE msr;

   if(enbl_handler == NULL || dsbl_handler == NULL)
   {
      return 1; 
   }
   else
   {
      LOCK_SAVE(msr);
      rw.hw_trace[0] = enbl_handler;
      rw.hw_trace[1] = dsbl_handler;
      LOCK_RESTORE(msr);
      return 0;
   } 
}

/*
** Returns the hash table position for a specific report id
*/
static uint32_t
get_reportid_hash(ProfUserReport *cur_report, OSREPORTID id)
{
   return (id & cur_report->hash_mask);
}

/*
**  Add item to the next vacant list position in the overflow list
*/
static OSBOOLEAN
add_reportid_to_overflow_list(ProfUserReport *cur_report, OSREPORTID id, uint32_t *index)
{
   OSBOOLEAN retval;

   if (cur_report->overflow_list_pos_cnt < cur_report->maxValues)
   {
      /* Set data in overflow list */
      *index = cur_report->overflow_hash_table[cur_report->overflow_list_pos_cnt].index = cur_report->prof_user_rpt_cursample->report_count++;

      cur_report->overflow_hash_table[cur_report->overflow_list_pos_cnt].id = id;
      cur_report->overflow_hash_table[cur_report->overflow_list_pos_cnt].sample = cur_report->sample_counter;

      /* Update overflow list counter */
      cur_report->overflow_list_pos_cnt++;

      /* Clear new samples */
#ifdef OSARM
      /* Also clear new samples for report class 52 & 53. We do garbage
       * collection for these report classes.
       */
      if ((!cur_report->continuous) ||
          (cur_report->reportno == HEAP_ACC_PER_PROCESS_REPORT) ||
          (cur_report->reportno == HEAP_ACC_PER_HEAP_REPORT))
      {
         cur_report->prof_user_rpt_cursample->report[*index].value = 0;
      }
      if (cur_report->maxmin)
      {
         cur_report->prof_user_rpt_cursample->report[*index].max = 0;
         cur_report->prof_user_rpt_cursample->report[*index].min = 0;
      }
#else
      if (!cur_report->continuous)
      {
         cur_report->prof_user_rpt_cursample->report[*index].value = 0;

         if (cur_report->maxmin)
         {
            cur_report->prof_user_rpt_cursample->report[*index].max = 0;
            cur_report->prof_user_rpt_cursample->report[*index].min = 0;
         }
      }
#endif

      retval = (OSBOOLEAN) 1;
   }
   else
   {
      /* The overflow list is full !*/
      retval = (OSBOOLEAN) 0;
   }
   return retval;
}

/*
** Tries to add a report id to it's corresponding position in the hash list.
** If the position already is filled for this sample it returns false.
*/
static OSBOOLEAN
add_reportid_to_hash_list(ProfUserReport *cur_report, OSREPORTID id, uint32_t *index)
{
   uint32_t hash_index;

   /* If hash position is vacant, use it for the pid */
   hash_index = get_reportid_hash(cur_report, id);

   if (cur_report->report_id_hash_table[hash_index].sample != cur_report->sample_counter)
   {
      /* Set data in hash list */
      cur_report->report_id_hash_table[hash_index].sample = cur_report->sample_counter;

      *index = cur_report->report_id_hash_table[hash_index].index = cur_report->prof_user_rpt_cursample->report_count++;
      cur_report->report_id_hash_table[hash_index].id = id;

      /* Clear new samples */
#ifdef OSARM
      /* Also clear new samples for report class 52 & 53. We do garbage
       * collection for these report classes.
       */
      if ((!cur_report->continuous) ||
          (cur_report->reportno == HEAP_ACC_PER_PROCESS_REPORT) ||
          (cur_report->reportno == HEAP_ACC_PER_HEAP_REPORT))

      {
         cur_report->prof_user_rpt_cursample->report[*index].value = 0;
      }
      if (cur_report->maxmin)
      {
         cur_report->prof_user_rpt_cursample->report[*index].max = 0;
         cur_report->prof_user_rpt_cursample->report[*index].min = 0;
      }
#else
      if (!cur_report->continuous)
      {
         cur_report->prof_user_rpt_cursample->report[*index].value = 0;

         if (cur_report->maxmin)
         {
            cur_report->prof_user_rpt_cursample->report[*index].max = 0;
            cur_report->prof_user_rpt_cursample->report[*index].min = 0;
         }
      }
#endif

      return (OSBOOLEAN) 1;
   }
   else
   {
      return (OSBOOLEAN) 0;
   }
}

/*
** Adds a report id entry to the sample array and to the indexing look-up lists.
*/
static OSBOOLEAN
add_reportid(ProfUserReport *cur_report, OSREPORTID id, uint32_t *index)
{
   OSBOOLEAN retval = (OSBOOLEAN) 0;

   /* If the process list is full add process to "other" */
   if (cur_report->prof_user_rpt_cursample->report_count < cur_report->maxValues)
   {
      if (!(retval = add_reportid_to_hash_list(cur_report, id, index)))
      {
         if (!(retval = add_reportid_to_overflow_list(cur_report, id, index)))
         {
            /* Error */
           ramlog_printf("rmm: add_reportid(): Both hash table and overflow table are full!\n");
         }
      }
   }
   return retval;
}


/*
** Do a linear search of the overflow list for index of supplied PID nr
*/
static OSBOOLEAN
search_reportid_from_overflow_list(ProfUserReport *cur_report, OSREPORTID id, uint32_t *index)
{
   uint32_t i;
   OSBOOLEAN retval = (OSBOOLEAN) 0;

   for( i = 0; i < cur_report->maxValues; i++ )
   {
      if (cur_report->overflow_hash_table[i].id == id)
      {
         if (cur_report->overflow_hash_table[i].sample == cur_report->sample_counter)
         {
            *index = cur_report->overflow_hash_table[i].index;
            return (OSBOOLEAN) 1;
         }
      }
   }
   return retval;
}

/*
** Looks for the specified reportid at it's hash list position. If that position
** already is occupied by another reportid for this sample, the overflow list
** is searched.
*/
static OSBOOLEAN
get_reportid(ProfUserReport *cur_report, OSREPORTID id, uint32_t *index)
{
   OSBOOLEAN retval = (OSBOOLEAN) 0;
   uint32_t hash_index;

   hash_index = get_reportid_hash(cur_report, id);
   if (cur_report->report_id_hash_table[hash_index].id == id )
   {
      if (cur_report->report_id_hash_table[hash_index].sample == cur_report->sample_counter)
      {
         *index = cur_report->report_id_hash_table[hash_index].index;
         retval = (OSBOOLEAN) 1;
      }
   }
   else
   {
      if (search_reportid_from_overflow_list(cur_report, id, index))
      {
         retval = (OSBOOLEAN) 1;
      }
   }
   return retval;
}

/*
** Sets the global hash size and overflow list sizes so that the selected
** number of Report Ids will fit into them and performance on hashing will be
** good
*/
static void
set_reportIndexing_sizes(ProfUserReport *cur_report)
{
   uint32_t max_nr_reportids = cur_report->maxValues;

   /* Minimum hash list size is set to 32 for performance reasons */
   if (max_nr_reportids < 32)
   {
      max_nr_reportids = 32;
   }

   cur_report->hash_mask = 1;
   /* maxpid is chosen as the power of 2 on or closest above max_nr_pids */
   while(cur_report->hash_mask < max_nr_reportids)
   {
      cur_report->hash_mask = (cur_report->hash_mask << 1) | 1 ;
   }
   cur_report->hash_size = cur_report->hash_mask + 1;
}

/*
** Add reference to the enabled user report.
** The userReportEnabledTable is a way to access enabled reports faster than
** it would have been if searching through all reports in a linear way.
*/
static void
addToUserReportEnableTable(ProfUserReport *enabled_report)
{
   /* Add */
   /* protected by user_prof_lock()*/
   if (rw.userReportEnabledTable == NULL)
   {
      /* Changed from using alloc_sample_pool() to more appropriate alloc() in this case */
      rw.userReportEnabledTable = (void *)alloc(MAX_NO_OF_ENABLED_USER_REPORTS * sizeof(UserReportEnabledHashItem), 0);
      memset(rw.userReportEnabledTable, 0, MAX_NO_OF_ENABLED_USER_REPORTS * sizeof(UserReportEnabledHashItem));
   }
   rw.userReportEnabledTable[rw.userReportEnabledCounter].enabledUserReport = enabled_report;
   rw.userReportEnabledCounter++;
}

/*
** Remove any references to the disabled user report
*  Assume that user_prof_lock is locked with interrupts off
*/
static void
removeFromUserReportEnableTable(ProfUserReport *disabled_report)
{
   uint32_t i, j;

   /* protected by user_prof_lock() */
   for (i = 0; i < rw.userReportEnabledCounter; i++)
   {
      if (rw.userReportEnabledTable[i].enabledUserReport == disabled_report)
      {
         break;
      }
   }
   if (i == rw.userReportEnabledCounter)
   {
      /* Error */
      ramlog_printf("rmm: removeFromUserReportEnableTable(): Did not find report %d to remove!\n", disabled_report->reportno);
      return;
   }
   else
   {
      for (j = i+1; j < rw.userReportEnabledCounter; j++,i++)
      {
         rw.userReportEnabledTable[i].enabledUserReport = rw.userReportEnabledTable[j].enabledUserReport;
      }
      rw.userReportEnabledCounter--;

      if (rw.userReportEnabledCounter == 0)
      {
         /* Free */
         if (rw.zzose_rmm_!= current_process())
         {
            restore((union SIGNAL *)rw.userReportEnabledTable);
         }
         free_sample_pool((union SIGNAL **)&rw.userReportEnabledTable);
      }
   }
}

/*
** Frees the user sample buffer for user reports profiling
*/
static void
free_user_report_bufs(ProfUserReport *cur_report)
{
   if (rw.zzose_rmm_!= current_process())
   {
      restore((union SIGNAL *)cur_report->prof_user_rpt_samplepool);
   }
   free_sample_pool((union SIGNAL **)&cur_report->prof_user_rpt_samplepool);
}

/* Build circular sample list for user report profiling. An extra
** field ( the +1 ) is used to count values for non allocated report id's.
**/
static OSBOOLEAN
create_user_report_bufs(ProfUserReport *cur_report)
{
   uint32_t alloc_size = 0;
   uint32_t sample_data_size = (cur_report->maxValues + 1) * sizeof(UserReportValue);

   /* Allocate size for all samples and their process lists */
   alloc_size = cur_report->maxReports * sizeof (ProfUserReportSample);
   alloc_size += sample_data_size * cur_report->maxReports;

   cur_report->prof_user_rpt_samplepool = (void *) alloc_sample_pool(alloc_size);

   /*
    * From the buffer aquired we construct the sample list by creating
    * maxReports number of sections containing first the sample struct and
    * then a following UserReportValue section where the actual values are stored.
   */
   if (cur_report->prof_user_rpt_samplepool != NULL)
   {
      char *bufp;
      ProfUserReportSample *samplep;
      uint32_t total_sample_size = sizeof(ProfUserReportSample) + sample_data_size;

      cur_report->prof_user_rpt_samplepool->magic = PROF_USER_REPORT_SAMPLE_MAGIC;

      memset(cur_report->prof_user_rpt_samplepool, 0, alloc_size);
      for (bufp = (char*)cur_report->prof_user_rpt_samplepool;
           bufp < (char*)cur_report->prof_user_rpt_samplepool + alloc_size;
           bufp += total_sample_size)
      {
         /* Set the current sample pointer to point at where the buf pointer points */
         samplep = (ProfUserReportSample *)bufp;
         /* next pointer points at the location a full sample ahead of the current*/
         samplep->next = (void *)(bufp + total_sample_size);
         /* report values location at the beginning of the data section */
         samplep->report = (UserReportValue *)( bufp + sizeof(ProfUserReportSample) );
      }

      /* bufp now points at the end of the allocated buffer, so make the last sample's
       * next pointer (which currently is a wild pointer) point at the first sample to
       * create the ring in the ring buffer */
      assert(bufp == ((char*)cur_report->prof_user_rpt_samplepool + alloc_size) );
      samplep = (ProfUserReportSample *)(bufp - total_sample_size);
      samplep->next = cur_report->prof_user_rpt_samplepool;

      /* Reset the buffer pointers and return*/
      cur_report->prof_user_rpt_firstsample = cur_report->prof_user_rpt_cursample = cur_report->prof_user_rpt_samplepool;
      cur_report->prof_user_rpt_firstsample->time_ticks =
         get_systime((OSTICK *) &cur_report->prof_user_rpt_firstsample->time_uticks);
   }
   else
   {
      /* Failed due to too large buffer wanted. Reset variables */
      cur_report->maxReports = 0;
      cur_report->maxValues = 0;
      return 0;
   }
   return 1;
}

/*
** Frees the index buffers for user reports profiling
*/
static void
free_user_rpt_index_bufs(ProfUserReport *cur_report)
{
   if (rw.zzose_rmm_ != current_process())
   {
      restore((union SIGNAL *)cur_report->report_id_hash_table);
      restore((union SIGNAL *)cur_report->overflow_hash_table);
   }
   free_sample_pool((union SIGNAL **)&cur_report->report_id_hash_table);
   free_sample_pool((union SIGNAL **)&cur_report->overflow_hash_table);
}

/*
** Sets up and intializes (or re-initializes existing) indexing
** buffers for the user reports
*/
static OSBOOLEAN
create_user_rpt_index_bufs(ProfUserReport *cur_report)
{
   OSBOOLEAN retval = (OSBOOLEAN) 0;

   /* Calculate and set indexing function sizes */
   set_reportIndexing_sizes(cur_report);

   /* Allocate the hash list search structure and reset it */
   cur_report->report_id_hash_table = (void *)alloc_sample_pool((sizeof(ReportIdHashItem) * cur_report->hash_size));
   /* Allocate the overflow list */
   cur_report->overflow_hash_table = (void *)alloc_sample_pool((sizeof(ReportIdHashItem) * cur_report->maxValues));

   if ((cur_report->report_id_hash_table != NULL) && (cur_report->overflow_hash_table != NULL))
   {
      memset(cur_report->report_id_hash_table, 0, (sizeof(ReportIdHashItem) * cur_report->hash_size));
      memset(cur_report->overflow_hash_table, 0, (sizeof(ReportIdHashItem) * cur_report->maxValues));
      retval = (OSBOOLEAN) 1;
   }
   else
   {
      /* Failed to allocate */
      free_user_rpt_index_bufs(cur_report);
      retval = (OSBOOLEAN) 0;
   }

   /* Reset other indexing variables */
   cur_report->sample_counter = 1;
   cur_report->overflow_list_pos_cnt = 0;

   return retval;
}

/*
** Enable report
** Allocate memory for user report samples and reset counters
*/
static OSBOOLEAN
init_user_report_sample(ProfUserReport *cur_report)
{
   OSBOOLEAN retval = (OSBOOLEAN) 0;

   /* if multiple flag in ProfUserReport is set to 0 then only one instance of UserReportValue should be allocated */
   if (!cur_report->multiple)
   {
      cur_report->maxValues = 1;
   }
   /* create the process report ring buffers*/
   if (create_user_report_bufs(cur_report))
   {
      if (cur_report->multiple)
      {
         /* Set up the indexing buffers for new number of processes */
         if (create_user_rpt_index_bufs(cur_report))
         {
            retval = (OSBOOLEAN) 1;
         }
         else
         {
            free_user_report_bufs(cur_report);
         }
      }
      else
      {
         retval = (OSBOOLEAN) 1;
      }
   }

   /* Reset counters */
   cur_report->sample_counter = 1;
   cur_report->overflow_list_pos_cnt = 0;

   return retval;
}

/*
** Disable report
** Free allocated memory and reset counters
*/
static void
reset_user_report_sample(ProfUserReport *cur_report)
{
   /* First free memory for  ReportId values */
   if (cur_report->prof_user_rpt_samplepool != NULL)
   {
      free_user_report_bufs(cur_report);
   }
   if (cur_report->multiple)
   {
      free_user_rpt_index_bufs(cur_report);
   }

   /* Reset counters */
   cur_report->sample_counter = 1;
   cur_report->overflow_list_pos_cnt = 0;
}

/*
** Enable/Disable application trigger
*/
static void
set_report_trigger (ProfUserReport *cur_report, OSADDRESS trig_value,
      LOCKSTATE *msr)
{
   ProfUserTrigList *trig_list = NULL;
   int i = 0;

   /* Set/Reset all triggers connected to this report. */
   for (trig_list = &cur_report->trig_list; trig_list != NULL; trig_list = trig_list->next)
   {
      for (i = 0; i < PROF_USER_REPORT_TRIG; i++)
      {
         if (trig_list->app_trig[i].trig != NULL)
         {
            PROCESS sid = trig_list->app_trig[i].sid;
            OSADDRESS addr = (OSADDRESS)trig_list->app_trig[i].trig;

            /* Reenable interrupts. */
            SMP_CODE( SPIN_UNLOCK(&rw.user_prof_lock); )
            LOCK_RESTORE(*msr);
            set_mem(sid, &trig_value, addr, sizeof(trig_value));
            LOCK_SAVE(*msr);
            SMP_CODE( SPIN_LOCK(&rw.user_prof_lock);      )
         }
      }
   }
}

/*
** Initialize the trig_list array and pointer
*/
static void
init_report_trig(ProfUserTrigList *trig_list)
{
   int i;
   /* init all entries in trig list */
   for (i = 0; i < PROF_USER_REPORT_TRIG; i++)
   {
      trig_list->app_trig[i].trig = NULL;
      trig_list->app_trig[i].sid = (PROCESS)NULL;
   }
   trig_list->next = NULL;
}



/*
** Search trig_list
**
 */
static uint8_t
search_user_report_trig_list(ProfUserTrigList *trig_list, PROCESS sid, OSADDRESS *trig, int *index)
{
   uint8_t trig_state = 0;
   int i;

   for (i = 0; i < PROF_USER_REPORT_TRIG; i++)
   {
      if (trig_list->app_trig[i].trig == trig && trig_list->app_trig[i].sid == sid)
      {
         return TRIG_SID_FOUND;
      }
      /* Have we already attached to this sid? */
      if (trig_list->app_trig[i].sid == sid)
      {
         trig_state |= SID_FOUND;
      }
      /* Mark the first free slot in trig_list */
      if (trig_list->app_trig[i].trig == NULL && !(trig_state & EMPTY_SLOT_FOUND))
      {
         *index = i;
         trig_state |= EMPTY_SLOT_FOUND;
      }
   }
   return trig_state;
}

/*
** Add the calling application trig address to user report and attach
** to sid if it does not already exist
*/
static void
add_user_report_trig(ProfUserReport *cur_report, PROCESS sid, OSADDRESS *trig)
{
   ProfUserTrigList *search_trig, *prev_trig;
   ProfUserTrigList *add_trig = NULL;
   OSADDRESS trig_value = 0;
   uint8_t found = 0;
   int i = 0;
   int add_index = 0;
   int sid_found = 0;

   /*
    ** Search for trig address in trig_list and keep track of the first
    ** free slot in trig_list. To be used if trig address is not found.
    **/
   for (prev_trig = search_trig = &cur_report->trig_list; search_trig != NULL;
         search_trig = search_trig->next)
   {
      found = search_user_report_trig_list(search_trig, sid, trig, &i);
      if (found & TRIG_SID_FOUND)
      {
         return;
      }
      if (found & SID_FOUND)
      {
    	  sid_found = 1;
      }
      if (found & EMPTY_SLOT_FOUND && add_trig == NULL)
      {
         add_index = i;
         add_trig = search_trig;
      }
      prev_trig = search_trig;
   }

   /* attach to this sid if we are not already attached (no need for sid == 0)*/
   if (sid && !sid_found)
   {
      union SIGNAL *sig;

      sig = alloc(sizeof(UserReportsAttachSig), USER_REPORTS_ATTACH_SIG);
      sig->user_report_attach.sid = sid;
      attach(&sig, sid);
   }

   if (add_trig != NULL)
   {
      add_trig->app_trig[add_index].sid = sid;
      add_trig->app_trig[add_index].trig = trig;
   }
   else
   {
      ProfUserTrigList *new_trig;

      /* allocate memory for new trig post, changed to alloc() from 
         alloc_sample_pool() */
      new_trig = (void *)alloc(sizeof (ProfUserReport), 0);

      new_trig->magic = PROF_USER_REPORT_TRIG_MAGIC;

      init_report_trig(new_trig);

      new_trig->app_trig[0].sid = sid;
      new_trig->app_trig[0].trig = trig;
      prev_trig->next = new_trig;
   }


   /* If Report is already enabled set trig to 1 */
   if (cur_report->icounter)
   {
      trig_value = 1;
   }
   set_mem(sid, &trig_value, (OSADDRESS)trig, sizeof(trig_value));
}

/*
** This function removes all user application address triggers for a
** specific block. It is called when an application have exited
** or crashed.
*/
static void
remove_user_report_trig(PROCESS sid)
{
   ProfUserReport *cur_report = NULL;
   ProfUserTrigList *first_trig = NULL;
   ProfUserTrigList *prev_trig = NULL;
   ProfUserTrigList *cur_trig = NULL;

   for (cur_report = rw.prof_first_user_report; cur_report != NULL; cur_report = cur_report->next)
   {
      first_trig = prev_trig = &cur_report->trig_list;

      /* Remove all triggers for this bid */
      for (cur_trig = first_trig; cur_trig != NULL; cur_trig = cur_trig->next)
      {
         int empty = 1;
         int i;
         for (i = 0; i < PROF_USER_REPORT_TRIG; i++)
         {
            if (cur_trig->app_trig[i].sid == sid)
            {
               cur_trig->app_trig[i].trig = (OSADDRESS)NULL;
               cur_trig->app_trig[i].sid = (PROCESS)0;
            }
            else if (cur_trig->app_trig[i].trig != NULL)
            {
            	empty = 0;
            }
         }
         /* Remove extra allocated memory if trig block is empty*/
         if (cur_trig != first_trig && empty)
         {
            prev_trig->next = cur_trig->next;
            /* Free memory for any empty trig blocks */
            free_buf((union SIGNAL **)&cur_trig);
            /* update cur_trig pointer */
            cur_trig = prev_trig;
         }
         prev_trig = cur_trig;
      }
   }
}

/*
** Check if report already exists
*/
static ProfUserReport *
find_report(SIGSELECT reportno)
{
   ProfUserReport *cur_report = NULL;

   for (cur_report = rw.prof_first_user_report; cur_report; cur_report = cur_report->next)
   {
      if (cur_report->reportno == reportno)
      {
         break;
      }
   }
   return cur_report;
}


/*
 * Handle user reports enabled request signal
 *
 **/
static int
handle_set_user_reports_enabled(Offspring *offspring, union SIGNAL *sig)
{
   union SIGNAL   *sig_r;
   ProfUserReport *cur_report;
   MonitorStatus  status = MONITOR_STATUS_OK;
   LOCKSTATE msr;


   /* Disable interrupts. */
   /* userReportEnabledCounter tests/changes
    * and the tickhook_ldmstatus should be done in an atomic transaction
    * this is protected by the user_prof_lock()
    **/
   LOCK_SAVE(msr);
   SMP_CODE( SPIN_LOCK(&rw.user_prof_lock);      )

   cur_report = find_report(sig->set_user_reports_enabled_request.reportNo);

   if (cur_report && sig->set_user_reports_enabled_request.interval > 0)
   {
      /* Enable reports profiling. */
      if (sig->set_user_reports_enabled_request.enabled)
      {
         /* Check if there is a free entry in array */
         if (rw.userReportEnabledCounter < MAX_NO_OF_ENABLED_USER_REPORTS)
         {
         /* But only if previously disabled. */
            if (!cur_report->icounter)
            {
               cur_report->interval = sig->set_user_reports_enabled_request.interval;
               cur_report->maxReports = sig->set_user_reports_enabled_request.maxReports;
               cur_report->maxValues = sig->set_user_reports_enabled_request.maxValuesReport;

               if (!in_range(cur_report->interval, rw.user_interval_range))
               {
                  ramlog_printf("rmm: user report %d, illegal interval value: %d range [%u:%u]\n",
				cur_report->reportno, sig->set_user_reports_enabled_request.interval,
				rw.user_interval_range.min, rw.user_interval_range.max);
                  cur_report->interval = rw.default_user_interval;
               }
               if (!in_range(cur_report->maxReports, rw.user_reports_range))
               {
                  ramlog_printf("rmm: user report %d, illegal maxReports value: %d range [%u:%u]\n",
                     cur_report->reportno, sig->set_user_reports_enabled_request.maxReports,
                  rw.user_reports_range.min, rw.user_reports_range.max);
                  cur_report->maxReports = rw.default_user_reports;
               }
               if (!in_range(cur_report->maxValues, rw.user_maxvalues_range))
               {
                  ramlog_printf("rmm: user report %d, illegal maxValuesReport value: %d range [%u:%u]\n",
                     cur_report->reportno, sig->set_user_reports_enabled_request.maxValuesReport,
                     rw.user_maxvalues_range.min, rw.user_maxvalues_range.max);
                  cur_report->maxValues = rw.default_user_maxvalues;
               }

               if (!init_user_report_sample(cur_report))
               {
                  status = get_last_memory_error();
               }
               else
               {
                  /* Start the disconnect timer. */
                  cur_report->prof_report_disctimer = 100000000L / zzprof.prof_all_ticklen;    /* 100 sec. */

		  cur_report->icounter = cur_report->interval;

                  /* Reset prof_ */
                  rw.prof_report_reqlist = (void *)0;

                  addToUserReportEnableTable(cur_report);

                  update_tick_hook();

                  /* Should we only update the application trigger if previously disabled ? */
                  set_report_trigger(cur_report, 1, &msr);
               }
            }
            else
            {
               status = MONITOR_STATUS_IN_USE;
            }
         }
         else
         {
            status = MONITOR_STATUS_MEMORY_EXHAUSTED;
         }
      }
      /* Disable reporting profiling. */
      else
      {
         /* But only if previously enabled. */
         if (cur_report->icounter)
         {
            /* Clear all trig connected to this report */
            set_report_trigger(cur_report, 0, &msr);

            removeFromUserReportEnableTable(cur_report);

            /* Free allocated memory */
            reset_user_report_sample(cur_report);

            /* Abort the current sample. */
            cur_report->interval = 0;
            cur_report->icounter = 0;
            cur_report->maxReports = 0;
            cur_report->maxValues = 0;
            cur_report->hash_mask = 0;

            update_tick_hook();
            update_int_proc_hook();
         }
      }
   }
   else
   {
      status = MONITOR_STATUS_REQUEST_INVALID;
   }

   /* Reenable interrupts. */
   SMP_CODE( SPIN_UNLOCK(&rw.user_prof_lock); )
   LOCK_RESTORE(msr);

   /* Send a reply to the offspring. */
   sig_r = alloc(sizeof (MonitorSetUserReportsEnabledReply),
                 MONITOR_SET_USER_REPORTS_ENABLED_REPLY);

   sig_r->set_user_reports_enabled_reply.status = status;
   send(&sig_r, offspring->pid);
   return 0;
}

/*
** This function creates an entry for a profiling user report.
** All memory allocation must be made from within the rmm. Since this
** function is called from an application no memory allocation can
** be done here.
** Send a signal to the rmm with the neccessary values and allocate memory
** from there instead.
** Parameters: rpt_flag containt three boolean flags for multiple, continuous
**             and maxmin values.
*/

OSADDRESS
zzose_create_report_sysconf_78(OSADDRESS reportno, OSADDRESS trig,
                               OSADDRESS rpt_flag)
{
   ProfUserReport *cur_report;
   union SIGNAL *sig;
   static const SIGSELECT sel_reply[] = { 1,
                                          MONITOR_CREATE_USER_REPORTS_STRUCT_REPLY
                                          };
   PROCESS sid;
   OSBOOLEAN multiple = ((rpt_flag & USER_RPT_MULTIPLE) == USER_RPT_MULTIPLE);
   OSBOOLEAN continuous = ((rpt_flag & USER_RPT_CONTINUOUS) == USER_RPT_CONTINUOUS);
   OSBOOLEAN maxmin = ((rpt_flag & USER_RPT_MAXMIN) == USER_RPT_MAXMIN);

   /* Log into ramlog */

   /* Check if report exists and if it is consistent */
   cur_report = find_report(reportno);

   if (cur_report)
   {
      /* Check if report values are consistent */
      if (!(cur_report->multiple == multiple && cur_report->continuous == continuous &&
          cur_report->maxmin == maxmin))
      {
         /* Values are not consistent */
         /* OSE_ERROR */
         zzos_error(0, OSE_CREATE_REPORT | OSE_EBAD_PARAMETER, 0);
         return (OSADDRESS) 0;
      }
   }

   sid = get_segid(current_process());

   sig = alloc(sizeof(MonitorCreateUserReportsStructRequest),
          MONITOR_CREATE_USER_REPORTS_STRUCT_REQUEST);

   sig->create_user_reports_struct_request.reportNo = reportno;
   sig->create_user_reports_struct_request.trig = (uint32_t *)trig;
   sig->create_user_reports_struct_request.multiple = multiple;
   sig->create_user_reports_struct_request.continuous = continuous;
   sig->create_user_reports_struct_request.maxmin = maxmin;
   sig->create_user_reports_struct_request.sid = sid;

   send(&sig, rw.zzose_rmm_);

   while(NIL == (sig = receive_from((OSTIME)~0, sel_reply, rw.zzose_rmm_)));
   free_buf(&sig);

   return (OSADDRESS) 0;
}

#ifdef OSARM
/* Function: enable_user_report
 * Input: reportNo,interval,maxReports,maxValuesReport
 * Output: TRUE - success, FALSE - failure
 * Brief: This function enables user report profiling.
 *        It is used to enable user profiling for those report
 *        classes which needs profiling to be started immediately
 *        after "RMM" receives create user report request
 *        Ex:- User report class 52 & 53
 */
static bool enable_user_report(uint32_t reportNo, uint32_t interval,
                               uint32_t maxReports, uint32_t maxValuesReport)
{
   bool retval = true;
   ProfUserReport *cur_report;
   LOCKSTATE msr;

   /* Disable interrupts. */
   /* userReportEnabledCounter tests/changes
    * and the tickhook_ldmstatus should be done in an atomic transaction
    * this is protected by the user_prof_lock()
    **/
   LOCK_SAVE(msr);
   SMP_CODE( SPIN_LOCK(&rw.user_prof_lock); )

   cur_report = find_report(reportNo);

   if (cur_report && interval > 0)
   {
      /* Check if there is a free entry in array */
      if (rw.userReportEnabledCounter < MAX_NO_OF_ENABLED_USER_REPORTS)
      {
         /* Enable only if it was previously disabled */
         if (!cur_report->icounter)
         {
            cur_report->interval = interval;
            cur_report->maxReports = maxReports;
            cur_report->maxValues = maxValuesReport;

            if (!in_range(cur_report->interval, rw.user_interval_range))
            {
               ramlog_printf("rmm: user report %u, illegal interval value: %d range [%u:%u]\n",
                           cur_report->reportno, interval,
                           rw.user_interval_range.min,
                           rw.user_interval_range.max);

               cur_report->interval = rw.default_user_interval;
            }

            if (!in_range(cur_report->maxReports, rw.user_reports_range))
            {
               ramlog_printf("rmm: user report %u, illegal maxReports value: %d range [%u:%u]\n",
                           cur_report->reportno, maxReports,
                           rw.user_reports_range.min,
                           rw.user_reports_range.max);

               cur_report->maxReports = rw.default_user_reports;
            }
            if (!in_range(cur_report->maxValues, rw.user_maxvalues_range))
            {
               ramlog_printf("rmm: user report %u, illegal maxValuesReport value: %d range [%u:%u]\n",
                           cur_report->reportno, maxValuesReport,
                           rw.user_maxvalues_range.min,
                           rw.user_maxvalues_range.max);

               cur_report->maxValues = rw.default_user_maxvalues;
            }

            if (!init_user_report_sample(cur_report))
            {
               retval = false;
               ramlog_printf("rmm: memory error:%u, user report class %u\n",
                           get_last_memory_error(),reportNo);
            }
            else
            {
               /* This flag will not allow disconnect timer to expire */
               cur_report->enabled = TRUE;
               /* Start the disconnect timer. */
               cur_report->prof_report_disctimer = 100000000 / zzprof.prof_all_ticklen;    /* 100 sec. */

               cur_report->icounter = cur_report->interval;

               addToUserReportEnableTable(cur_report);

               update_tick_hook();

               /* Update the application trigger */
               set_report_trigger(cur_report, 1, &msr);
            }

         } /* if its already enabled retval = true */
      }
      else
      {
         retval = false;
         ramlog_printf("rmm: memory exhausted, user report class %u\n",reportNo);
      }
   }
   else
   {
      retval = false;
   }
   /* Reenable interrupts. */
   SMP_CODE( SPIN_UNLOCK(&rw.user_prof_lock); )
   LOCK_RESTORE(msr);

   return retval;

}


/* Read system parameter "/krn/monitor/enable_report/reportno" */
static bool
read_user_report_system_parameter(SIGSELECT reportno,
                                  OSTICK *interval,
                                  uint32_t *maxReports,
                                  uint32_t *maxValuesReport)
{
   /* User report system parameter string
    * (/krn/monitor/enable_report/ + report no(upto 10 integer characters) + 1)
    */
   char user_rpt_str[26 + 10 + 1];
   Boolean enable = FALSE;

   /* concatenate the report number to the system parameter.
    * assuming that reportno is specified in decimal
    */
   sprintf(user_rpt_str, "krn/monitor/enable_report/%d", reportno);

   if (!sysparam_readattr_bool(user_rpt_str, "enable", &enable))
   {
      if (enable)
      {
         long param;

         /* extract user report attributes */
         if (sysparam_readattr_long(user_rpt_str,"interval", &param))
         {
            /* since this user report class needs to enabled immediately
             * after "RMM" is up, hence log error and continue.If we exit
             * user needs to restart the system.
             * Default interval will be used
             */
            ramlog_printf("rmm: Unable to read 'interval' attribute of %s"
                          " system parameter\n", user_rpt_str);
            return false;

         }
         *interval = (OSTICK)param;

         if (sysparam_readattr_long(user_rpt_str, "reports", &param))
         {
            ramlog_printf("rmm: Unable to read 'reports' attribute of %s"
                          " system parameter\n", user_rpt_str);
            return false;

         }
         *maxReports = (uint32_t)param;

         if (sysparam_readattr_long(user_rpt_str, "maxvalues", &param))
         {
            ramlog_printf("rmm: Unable to read 'maxvalues' attribute of %s"
                          " system parameter\n",user_rpt_str);
            return false;
         }
         *maxValuesReport = (uint32_t)param;

         return true;
      }
   }

   return false;
}
#endif

/*
**  Received signal from ose_create_report function.
**  Allocate and init memory for a user report.
*/
static void
handle_create_user_reports(SIGSELECT reportno, OSADDRESS *trig, OSBOOLEAN multiple,
                              OSBOOLEAN continuous, OSBOOLEAN maxmin, PROCESS sid)
{
   ProfUserReport *cur_report;

   cur_report = find_report(reportno);
   if (cur_report)
   {
      add_user_report_trig(cur_report, sid, trig);
   }
   else
   {
#ifdef OSARM
      OSTICK interval;
      uint32_t maxReports;
      uint32_t maxValuesReport;
#endif
      /* Allocate memory for new report */
      cur_report = (void *) alloc(sizeof (struct prof_user_report),PROF_USER_REPORT_MAGIC);

      /* init report header struct */
      cur_report->reportno = reportno;
      init_report_trig(&cur_report->trig_list);
      cur_report->multiple = multiple;
      cur_report->continuous = continuous;
      cur_report->maxmin = maxmin;
      cur_report->interval = 0;
      cur_report->icounter = 0;
      cur_report->prof_user_rpt_samplepool = NULL;
      cur_report->prof_user_rpt_firstsample = NULL;
      cur_report->prof_user_rpt_cursample = NULL;
      cur_report->maxReports = rw.default_user_reports;
      cur_report->maxValues = rw.default_user_maxvalues;
      cur_report->prof_report_disctimer = 0;
      cur_report->report_id_hash_table = NULL;
      cur_report->overflow_hash_table = NULL;
      cur_report->hash_mask = 0;
      cur_report->hash_size = 0;
      cur_report->sample_counter = 1;
      cur_report->overflow_list_pos_cnt = 0;
      cur_report->next = NULL;
#ifdef OSARM
      cur_report->enabled = FALSE;
#endif

      add_user_report_trig(cur_report, sid, trig);

      if (rw.prof_first_user_report == NULL)
      {
         rw.prof_first_user_report = cur_report;
         rw.prof_cur_user_report = rw.prof_first_user_report;
      }
      else
      {
         rw.prof_cur_user_report->next = cur_report;
         rw.prof_cur_user_report = cur_report;
      }
#ifdef OSARM
      if (true == read_user_report_system_parameter(reportno, &interval,
          &maxReports, &maxValuesReport))
      {
         /* enable profiling for the user report class*/
         if (false == enable_user_report(reportno, interval, maxReports,
             maxValuesReport))
         {
            ramlog_printf("rmm: failed to enable user report number %u"
                          " system parameter \n", reportno);
         }
      }
#endif
   }
}

/*
 * Search only active reports
 */
static ProfUserReport *
find_enabled_report(SIGSELECT reportno)
{
   uint32_t i;

   for (i = 0; i < rw.userReportEnabledCounter; i++)
   {
      if (rw.userReportEnabledTable[i].enabledUserReport->reportno == reportno)
      {
         return rw.userReportEnabledTable[i].enabledUserReport;
      }
   }
   return (ProfUserReport *)NULL;
}

/*
** This function is called from a user application to update
** the value in the current sample. Depending on how the report
** has been configured the change value is interpreted differently.
*/
OSADDRESS
zzose_set_report_val_sysconf_77(OSADDRESS reportno, OSADDRESS id,
                                OSADDRESS change)
{
   LOCKSTATE msr;
   ProfUserReport *cur_report = NULL;

   LOCK_SAVE(msr);
#ifdef OSE_SMP
   if (likely(rw.skip_user_prof_lock != zzkrn_cpu_id()))
   {
      SPIN_LOCK(&rw.user_prof_lock);
   }
#endif

   /* Find enabled report */
   if (rw.userReportEnabledCounter)
   {
      cur_report = find_enabled_report((SIGSELECT) reportno);
   }

   if (cur_report)
   {
      UserReportValue *report_value = cur_report->prof_user_rpt_cursample->report;
      uint32_t idx = 0;

      /*
       * Get the corresponding index for the new report id. If not found, add it.
       */
      if (cur_report->multiple)
      {
         if (!get_reportid(cur_report, (OSREPORTID)id, &idx))
         {
            if (!add_reportid(cur_report, (OSREPORTID)id, &idx))
            {
               /* Add to sumOther field in report_value */
               idx = cur_report->maxValues;
            }
         }
      }
      else
      {
         idx = cur_report->maxValues;
      }

      report_value[idx].id = (OSREPORTID) id;
      report_value[idx].value += (OSREPORTVAL) change;

      /* Track max/min values in this interval? */
      if (cur_report->maxmin)
      {
         if (report_value[idx].value > report_value[idx].max)
         {
            report_value[idx].max = report_value[idx].value;
         }
         if (report_value[idx].value < report_value[idx].min)
         {
            report_value[idx].min = report_value[idx].value;
         }
      }
   }

#ifdef OSE_SMP
   if (likely(rw.skip_user_prof_lock != zzkrn_cpu_id()))
   {
      SMP_CODE( SPIN_UNLOCK(&rw.user_prof_lock);      )
   }
#endif
   LOCK_RESTORE(msr);

   return (OSADDRESS) 0;
}

/**
 * @function handle_report_val
 *
 * @brief Send pending user reports.
 *
 * @long if link_in is true the request will be saved for later use
 * if no pending reports exists that matches the request.
 * @long
 * *rsig will be set to NULL and the caller should not operate on the
 * signal any more.
 *
 */
static void
handle_report_val(union SIGNAL **rsig, int link_in, Offspring * offspring,
      volatile ProfUserReport *cur_report)
{
   union SIGNAL   *sig;
   uint32_t        count, total_report_count;
   uint32_t        tick = (*rsig)->get_user_reports_request.time.tick;
   uint32_t        utick = (*rsig)->get_user_reports_request.time.ntick;
   MonitorUserReport *d;
   MonitorMaxMinUserReport *m;
   ProfUserReportSample *s;
   uint32_t         act_size;
   uint32_t         alloc_size;
   uint32_t i;
   LOCKSTATE msr;


   assert(rsig);
   assert(*rsig);
   assert(link_in == 0 || link_in == 1);
   assert_offspring(offspring);

   retry:
   /* Disable interrupts to avoid conflicts with the tick call etc. */
   LOCK_SAVE(msr);
   SMP_CODE( SPIN_LOCK(&rw.user_prof_lock);)

   /* If measurement is inactive, return requests with the active flag set to zero. */
   /* both branches leave without the user_prof_lock */
   if (!cur_report->icounter)
   {
      SMP_CODE( SPIN_UNLOCK(&rw.user_prof_lock);)
      LOCK_RESTORE(msr);
      sig = alloc(sizeof (MonitorGetUserReportsReply), MONITOR_GET_USER_REPORTS_REPLY);
      sig->get_user_reports_reply.status = MONITOR_STATUS_OK;
      sig->get_user_reports_reply.reportNo = cur_report->reportno;
      sig->get_user_reports_reply.enabled = 0;
      sig->get_user_reports_reply.continuous = cur_report->continuous;
      sig->get_user_reports_reply.maxmin = cur_report->maxmin;
      sig->get_user_reports_reply.multiple = cur_report->multiple;
      sig->get_user_reports_reply.reserved0 = 0;
      sig->get_user_reports_reply.reserved1 = 0;
      sig->get_user_reports_reply.reportsCount = 0;
      sig->get_user_reports_reply.reportValuesCount = 0;

      send(&sig, offspring->pid);
   }
   else                         /* If measurement is active, return any samples found. */
   {
      /* Allocate a worst case response signal. */
      alloc_size = sizeof (MonitorGetUserReportsReply) +
                  cur_report->maxReports * (sizeof (MonitorMaxMinUserReport) +
                                                (cur_report->maxValues - 1) *
                                                sizeof (MonitorMaxMinUserReportValue));
      /* We need to unlock here and when relock and recheck everything. */
      SMP_CODE( SPIN_UNLOCK(&rw.user_prof_lock);)
      LOCK_RESTORE(msr);

      sig = alloc(alloc_size, MONITOR_GET_USER_REPORTS_REPLY);

      LOCK_SAVE(msr);
      SMP_CODE( SPIN_LOCK(&rw.user_prof_lock);)
      if (!cur_report->icounter || alloc_size !=
         sizeof (MonitorGetUserReportsReply) +
         cur_report->maxReports * (sizeof (MonitorMaxMinUserReport) +
         (cur_report->maxValues - 1) *
         sizeof (MonitorMaxMinUserReportValue)))
      {
         /* Oops something has changed, free the signal and restart... */
         SMP_CODE( SPIN_UNLOCK(&rw.user_prof_lock);)
         LOCK_RESTORE(msr);
         free_buf(&sig);
         goto retry;
      }
      sig->get_user_reports_reply.enabled = 1;
      sig->get_user_reports_reply.reserved0 = 0;
      sig->get_user_reports_reply.reserved1 = 0;

      d = (void *)sig->get_user_reports_reply.reportValues;
      m = (void *)sig->get_user_reports_reply.reportValues;

      s = cur_report->prof_user_rpt_firstsample;

      /* Search for the specified time stamp if non zero. */
      if (tick || utick)
      {
         while (s->valid)
         {
            /* Did we find the specified entry. */
            if ((tick == s->time_ticks) && (utick == s->time_uticks))
            {
               /* Found it. Point to the next, yet unread sample. */
               s = s->next;
              break;
            }
            s = s->next;
         }
      }
      if (!s->valid)
      {
         if (link_in)
         {
            /* Link it into the request queue. */
            (*rsig)->sig_no = (SIGSELECT)rw.prof_report_reqlist;
            rw.prof_report_reqlist = *rsig;
            SMP_CODE( SPIN_UNLOCK(&rw.user_prof_lock);)
            LOCK_RESTORE(msr);
            free_buf(&sig);
            *rsig = NULL;
            return;
         }
         else
         {
            /*
             ** If the sample list empty, we will assume that the
             ** input signal contained a bad time stamp and send the entire
             ** backlog. This should make any user look for the bug.
            */
            s = cur_report->prof_user_rpt_firstsample;
         }
      }
      /* Count samples and report id's and copy them into the response. */
      count = total_report_count = 0;
      while (s->valid)
      {
         if (cur_report->maxmin)
         {
            m->time.tick = s->time_ticks;
            m->time.ntick = s->time_uticks;
            m->interval = s->total_time;
            m->sumOther = s->report[cur_report->maxValues].value;
            m->maxOther = s->report[cur_report->maxValues].max;
            m->minOther = s->report[cur_report->maxValues].min;
            m->reserved2 = 0;
            m->reportValuesCount = s->report_count;

            for (i = 0; i < s->report_count; i++)
            {
               m->reportValues[i].reportValue.id = s->report[i].id;
               m->reportValues[i].reportValue.sum = s->report[i].value;
               m->reportValues[i].max = s->report[i].max;
               m->reportValues[i].min = s->report[i].min;
            }
            m = (void *) &m->reportValues[s->report_count];
         }
         else
         {
            d->time.tick = s->time_ticks;
            d->time.ntick = s->time_uticks;
            d->interval = s->total_time;
            d->sumOther = s->report[cur_report->maxValues].value;
            d->reserved0 = 0;
            d->reportValuesCount = s->report_count;


            for (i = 0; i < s->report_count; i++)
            {
               d->reportValues[i].id = s->report[i].id;
               d->reportValues[i].sum = s->report[i].value;
            }
            d = (void *) &d->reportValues[s->report_count];
         }

         ++count;
         total_report_count += s->report_count;

         s = s->next;
      }

      /* Reenable interrupts. */
      SMP_CODE( SPIN_UNLOCK(&rw.user_prof_lock);      )
      LOCK_RESTORE(msr);
      act_size = 0;
      if (count)
      {
         act_size = (count * (cur_report->maxmin ? sizeof(MonitorMaxMinUserReport) : sizeof(MonitorUserReport))) +
                      (((total_report_count - count) *
                           (cur_report->maxmin ? sizeof(MonitorMaxMinUserReportValue) : sizeof(MonitorUserReportValue))));
      }
      /* Finalize the counter and truncate the signal. */
      lset_sigsize(&sig, sizeof (MonitorGetUserReportsReply) + act_size);

      sig->get_user_reports_reply.status = MONITOR_STATUS_OK;
      sig->get_user_reports_reply.reportNo = cur_report->reportno;
      sig->get_user_reports_reply.enabled = cur_report->icounter ? 1 : 0;
      sig->get_user_reports_reply.continuous = cur_report->continuous;
      sig->get_user_reports_reply.maxmin = cur_report->maxmin;
      sig->get_user_reports_reply.multiple = cur_report->multiple;
      sig->get_user_reports_reply.reserved0 = 0;
      sig->get_user_reports_reply.reserved1 = 0;
      sig->get_user_reports_reply.reportsCount = count;
      sig->get_user_reports_reply.reportValuesCount = act_size / sizeof(MonitorUserReportValue); /* total_report_count; */
      send(&sig, offspring->pid);
   }
   free_buf(rsig);
}

#ifdef OSARM
/*
** This function is called from a heap profile plug-in instance when heap is
** terminated. If heap (report id) is part of the current report sample, then
** report value will be forced to zero.
*/
void zzrmm_notify_reportid_killed(OSADDRESS reportno, OSADDRESS id)
{
   LOCKSTATE msr;
   ProfUserReport *cur_report = NULL;

   LOCK_SAVE(msr);
#ifdef OSE_SMP
   if (likely(rw.skip_user_prof_lock != zzkrn_cpu_id()))
   {
      SPIN_LOCK(&rw.user_prof_lock);
   }
#endif

   /* Find enabled report */
   if (rw.userReportEnabledCounter)
   {
      cur_report = find_enabled_report((SIGSELECT) reportno);
   }

   if (cur_report)
   {
      UserReportValue *report_value = cur_report->prof_user_rpt_cursample->report;
      uint32_t idx = 0;

      /*
       * Get index of the terminated heap (report id), if not found return.
       */
      if (cur_report->multiple)
      {
         if (get_reportid(cur_report, (OSREPORTID)id, &idx))
         {
            /* Force report value to zero, indicating heap terminated, it will be
             * garbage collected at close interval
             */
            report_value[idx].value = (OSREPORTVAL) 0;
         }
      }
   }
#ifdef OSE_SMP
   if (likely(rw.skip_user_prof_lock != zzkrn_cpu_id()))
   {
      SMP_CODE( SPIN_UNLOCK(&rw.user_prof_lock);)
   }
#endif
   LOCK_RESTORE(msr);

   return;
}

/* Function: remove_reportid
 * Input: ProfUserReport*, uint32_t
 * Return: None
 * Brief: This function removes the garbage report id from the
 *        current sample and reshuffles hash/overflow table accordingly.
 */
static void remove_reportid(ProfUserReport *cur_report, uint32_t index)
{
   uint32_t j;
   uint32_t i;
   uint32_t hash_index;
   ProfUserReportSample *cur_sample = cur_report->prof_user_rpt_cursample;

   /* Invalidate the report id in the hash table.
    * Note:- report id will definately be part of either hash table or
    * overflow table, and hence no failure condition arises
    */

   hash_index = get_reportid_hash(cur_report, cur_sample->report[index].id);
   if (cur_report->report_id_hash_table[hash_index].id ==  cur_sample->report[index].id)
   {
      if (cur_report->report_id_hash_table[hash_index].sample == cur_report->sample_counter)
      {
         cur_report->report_id_hash_table[hash_index].sample = 0;
      }
   }
   else
   {
      /* check whether report id is part of overflow table */
      for( i = 0; i < cur_report->overflow_list_pos_cnt; i++ )
      {
         if (cur_report->overflow_hash_table[i].id == cur_sample->report[index].id)
         {
            if (cur_report->overflow_hash_table[i].sample == cur_report->sample_counter)
            {
               cur_report->overflow_hash_table[i].sample = 0;
            }
         }
      }
   }

   /* Reshuffle the report ids in the current sample */
   if (cur_report->maxmin)
   {
      for (j = index; j < cur_sample->report_count - 1; j++)
      {
         cur_sample->report[j].id = cur_sample->report[j+1].id;
         cur_sample->report[j].value = cur_sample->report[j+1].value;
         cur_sample->report[j].min = cur_sample->report[j+1].min;
         cur_sample->report[j].max = cur_sample->report[j+1].max;
      }
   }
   else
   {
      for (j = index; j < cur_sample->report_count - 1; j++)
      {
         cur_sample->report[j].id = cur_sample->report[j+1].id;
         cur_sample->report[j].value = cur_sample->report[j+1].value;
      }

   }

   /* Update hash table index values with new indexes.
    * Note cur_sample now contains reshuffled report ids
    * All reshuffled report ids should be part of either hash/overflow table
    * No fail conditions occur
    */
   for (j = index; j < cur_sample->report_count - 1; j++)
   {
      hash_index = get_reportid_hash(cur_report, cur_sample->report[j].id);
      if (cur_report->report_id_hash_table[hash_index].id == cur_sample->report[j].id)
      {
         /* Is this condition required ? */
         if (cur_report->report_id_hash_table[hash_index].sample == cur_report->sample_counter)
         {
            /* Update the index in the hash table */
            cur_report->report_id_hash_table[hash_index].index = j;
         }
      }
      else
      {
         /* check whether report id is part of overflow table */
         for( i = 0; i < cur_report->overflow_list_pos_cnt; i++ )
         {
            if (cur_report->overflow_hash_table[i].id == cur_sample->report[j].id)
            {
               /* Is this condition required ? */
               if (cur_report->overflow_hash_table[i].sample == cur_report->sample_counter)
               {
                  /* Update the index in the overflow table */
                  cur_report->overflow_hash_table[i].index = j;
               }
            }
         }

      }

   }

   /* decrement the current sample report_count */
   cur_sample->report_count--;

   return;

}

/* Function: garbage_collect_reportids
 * input: ProfUserReport*
 * return: None
 * Brief: This function removes garbage report ids from the report.
 *        Garbage report ids are the ones which are terminated or
 *        whose values are zero. Applicable only to report class
 *        52 & 53.
 */
static void garbage_collect_reportids(ProfUserReport *cur_report)
{
   uint32_t report_index = 0;
   ProfUserReportSample *cur_sample = cur_report->prof_user_rpt_cursample;

   if (cur_report->reportno == HEAP_ACC_PER_PROCESS_REPORT)
   {
      while (report_index < cur_sample->report_count)
      {
         /* check whether the process is killed or accumulated heap usage
          * reduced to Zero
          */
         if ((get_ptype(cur_sample->report[report_index].id) == OS_ZOOMBIE) ||
             (cur_sample->report[report_index].value == 0))
         {
            /* remove the process from the report and decrement report count */
            remove_reportid(cur_report, report_index);
         }
         else
         {
            report_index++;
         }
      }
   }
   else
   {
      while (report_index < cur_sample->report_count)
      {
         /* check whether the heap is killed or accumulated heap usage
          * reduced to Zero. For both conditions the report value will be
          * Zero. see zzrmm_notify_heap_killed function.
          */
         if (cur_sample->report[report_index].value == 0)
         {
            /* remove the heap from the report and decrement the report count */
            remove_reportid(cur_report, report_index);
         }
         else
         {
            report_index++;
         }
      }
   }

   return;
}
#endif

#ifdef EXTERN_GET_SYSTIME
OSTICK extern_get_systime(OSTICK *);
#define GET_SYSTIME(t) extern_get_systime(t)
#else
#define GET_SYSTIME(t) get_systime(t)
#endif

/*
** This function processes all enabled user reports.
** At the end of interval the current sample is closed and
** a new sample is opened for the coming interval.
** Mark the current sample as valid. (Available for users).
** update_report_sample() is only called when there are enabled reports.
** The report is set to invalid when the disconnect timer expires and
** all memory allocated for the sample is freed.
*/
static void
update_report_sample(void)
{
   ProfUserReport *cur_report;
   ProfUserReportSample *old, *new;
   uint32_t i;
   int send_notify = 0;
   LOCKSTATE msr;

   LOCK_SAVE(msr);
   SMP_CODE( SPIN_LOCK(&rw.user_prof_lock);      )

   /* Loop through each enabled report and update the sample pointer */
   for (i = 0; i < rw.userReportEnabledCounter; i++)
   {
      cur_report = rw.userReportEnabledTable[i].enabledUserReport;

      if (cur_report->icounter)
      {
         /* Ack. time in the current sample */
         cur_report->prof_user_rpt_cursample->total_time += zzprof.prof_all_ticklen;

         /* Update the counter and see if interval has expired */
         if (!--cur_report->icounter)
         {
            /* Check if callback function registered */
            if (rw.user_report_function != NULL)
            {
               /* This is a special case there the user function is expected to
                * call set_report_val(), which will need the user_prof_lock that
                * we already have. Since we are in a critical region we do not
                * want to unlock it, hence we set a variable with our core id
                * that the set_report_val() can check upon.
                */
               SMP_CODE(rw.skip_user_prof_lock = zzkrn_cpu_id();)
               rw.user_report_function(cur_report->reportno);
               SMP_CODE(rw.skip_user_prof_lock = ~1;)
            }

            /* Restart the counter */
            cur_report->icounter = cur_report->interval;

            /* Close the old sample and allocate a fresh one. */
            old = cur_report->prof_user_rpt_cursample;
            new = old->next;

            if (new == cur_report->prof_user_rpt_firstsample)
            {
               cur_report->prof_user_rpt_firstsample = cur_report->prof_user_rpt_firstsample->next;
            }

            new->valid = 0;
            new->total_time = 0;
            new->report[cur_report->maxValues].id = 0;

            /* Switch from the old to the new sample. */
            old->valid = 1;
            cur_report->prof_user_rpt_cursample = new;

            /* Have the previous values copied to the next interval start */
            if (cur_report->continuous)
            {
               /* Keep the reported values from previous sample */
               for (i = 0; i < old->report_count; i++)
               {
                  cur_report->prof_user_rpt_cursample->report[i].id = old->report[i].id;
                  cur_report->prof_user_rpt_cursample->report[i].value = old->report[i].value;

                  /* Reset maxmin values */
                  if (cur_report->maxmin)
                  {
                     cur_report->prof_user_rpt_cursample->report[i].min = old->report[i].value;
                     cur_report->prof_user_rpt_cursample->report[i].max = old->report[i].value;
                  }
               }
               new->report_count = old->report_count;

               /* Keep "others" value and reset "others" maxmin */
               if (cur_report->maxmin)
               {
                  new->report[cur_report->maxValues].max = old->report[cur_report->maxValues].value;
                  new->report[cur_report->maxValues].min = old->report[cur_report->maxValues].value;
               }
               new->report[cur_report->maxValues].value = old->report[cur_report->maxValues].value;

#ifdef OSARM
               /* Garbage collect report ids, applicable only to report class 52&53
                * (new = cur_report->prof_user_rpt_cursample will be used for
                * garbage collection.
                * Turn off garbage collection if 'other' field is non-zero.
                * Even if 'other' field gets reduced to zero in subsequent samples,
                * it would only mean that all report ids which were part of 'other'
                * field were reduced to zero and turning on garbage collection will
                * still give correct values for new report ids which occupies the
                * empty slots.
                */

               if (((cur_report->reportno == HEAP_ACC_PER_PROCESS_REPORT)||
                   (cur_report->reportno == HEAP_ACC_PER_HEAP_REPORT)) &&
                   (new->report[cur_report->maxValues].value == 0))
               {
                  garbage_collect_reportids(cur_report);
               }
#endif

            }
            else
            {
               new->report_count = 0;

               /* Reset "others" value and reset "others" maxmin */
               if (cur_report->maxmin)
               {
                  new->report[cur_report->maxValues].max = 0;
                  new->report[cur_report->maxValues].min = 0;
               }
               new->report[cur_report->maxValues].value = 0;

               /* Update the samplecounter to mark the hash table as invalid */
               cur_report->overflow_list_pos_cnt = 0;
               cur_report->sample_counter++;
            }

            new->time_ticks = GET_SYSTIME((OSTICK *)&(new->time_uticks));

            /* If a request is pending, send a notification to the RMM. */
            if (rw.prof_report_reqlist)
            {
               send_notify = TRUE;
            }

            /* Disable measurement if no host operation has been seen for a long time. */
            /* Decrement by length of interval. */
            if (cur_report->prof_report_disctimer <= cur_report->interval)
            {
               cur_report->prof_report_disctimer = 0;

               /* Clear all trig connected to this report */
               set_report_trigger(cur_report, 0, &msr);

               removeFromUserReportEnableTable(cur_report);

               /* Free allocated memory */
               reset_user_report_sample(cur_report);

               /* Abort the current sample. */
               cur_report->interval = 0;
               cur_report->icounter = 0;
               cur_report->sample_counter = 1;
               cur_report->overflow_list_pos_cnt = 0;
            }
            else
            {
#ifdef OSARM
               /* If disconnect timer expires the profiling will be disabled,
                * this should not be allowed for user report profiling
                * enabled using system parameter.
                */
               if (!cur_report->enabled)
#endif
               {
                  cur_report->prof_report_disctimer -= cur_report->interval;
               }
            }
         }
      }
   }

   SMP_CODE( SPIN_UNLOCK(&rw.user_prof_lock);      )
   LOCK_RESTORE(msr);

   if (send_notify)
   {
      union SIGNAL   *sig =
         alloc(sizeof (ProfNewSample), PROF_NEW_SAMPLE);
      sig->prof_new_sample.type = REPORT_PROF;
      send(&sig, rw.zzose_rmm_);
   }
}

static void
handle_new_report_sample(void)
{
      while (rw.prof_report_reqlist)
      {
         Offspring         *offspring;
         union SIGNAL   *rsig;

         /* Unlink the next request. */
         rsig = rw.prof_report_reqlist;
         rw.prof_report_reqlist = (union SIGNAL *) rsig->sig_no;     /* Next pointer. */

         offspring = find_offspring_by_pid(sender(&rsig));
         if (NULL != offspring)
         {
            ProfUserReport *requested_report = NULL;

            requested_report = find_report(rsig->get_user_reports_request.reportNo);

            if (requested_report)
            {
               handle_report_val(&rsig, 0, offspring, requested_report);
            }
            else
            {
               free_buf(&rsig);
            }
         }
         else
         {
            /* We have lost the offspring that requested this report. */
            free_buf(&rsig);
         }
      }
}


/* DEBUG */
unsigned long long debug_begin[2];




/*
** Sample interval supervision for the Profiler subsystem.
**
** This function is called from the kernel every system tick in
** the context of the current system timer process.
**
** System calls can be issued as from any interrupt process.
*/

void
zzprof_tickhook(void)
{
   if (rw.userReportEnabledCounter)
   {
      /* Time to update any active user reports */
      update_report_sample();
   }
}


/*-- Evact Functions ----------------------------------------------------*/

/*
** Subfunction to do_set_breakpoint_scope().
**
** It is called for each process when updating the breakpoint scope.
**
** Lock the dispatcher for each process.
** Update pcb flags in all processes.
** Add new processes to the scope cache.
** Intercept new processes if required.
*/

static void
set_one_breakpoint_scope(PROCESS pid, U32 flags)
{
   struct debug_pcb *dcb;

   /* Remove ``not used arg'' compiler warning. */
   (void) flags;

   /* Lock the dispatcher. */
   LOCK_SWAP();

   /* Get the debugger pcb. */
   dcb = zzget_debugpcb(pid);

   /* Skip dead processes. */
   if (!dcb)
   {
      UNLOCK_SWAP();
      return;
   }

   /* Check if the process is a system process. */
   /* Skip any such process. */
   if (!(dcb->debug_flags & SYS_PROCESS))
   {
      /* Match the process name against the breakpoint scope. */
      if (zzrmm_match_proc(dcb, rw.quick_breakpoint_scope) && (!rw.evact_lock_owner /* Inefficient and dangerous kludge to handle segment scopes */
                                                      || rw.evact_lock_owner->debug_scope.type != MONITOR_SCOPE_SEGMENT_ID
                                                      || rw.evact_lock_owner->debug_scope.id == get_segid(pid)))
      {
         /* Done if a matching process is already in the breakpoint scope. */
         if (!(dcb->debug_flags & IN_BREAK_SCOPE))
         {
            register struct break_pid_entry *bke;
            register PROCESS *pidp;

            /* Otherwise Look for a free pid slot. */
            for (bke = rw.break_pid_list; bke; bke = bke->next)
            {
               if (bke->use_count < bke->size)
               {
                  break;
               }
            }

            /* Take it if found. */
            if (bke)
            {
               for (pidp = bke->pid - 1; *(++pidp);)
               {;               /* Empty loop body. */
               }
               BKL_LOCK();
               bke->use_count++;
               *pidp = pid;
               BKL_UNLOCK();
            }
            /* Otherwise allocate a new entry. */
            else
            {
               register PROCESS *pidp;
               register U32    i;
               bke = (struct break_pid_entry *) alloc(sizeof (struct break_pid_entry), (SIGSELECT) 0);
               for (i = 128, pidp = bke->pid; i--;)
                  *pidp++ = (PROCESS) 0;
               bke->size = 128;
               bke->use_count = 1;
               bke->pid[0] = pid;
               BKL_LOCK();
               bke->next = rw.break_pid_list;
               rw.break_pid_list = bke;
               BKL_UNLOCK();
            }

            /* Add a matching process to the breakpoint scope. */
            dcb->debug_flags |= IN_BREAK_SCOPE;

            /* Intercept the new process if required. */
            if (rw.break_scope_status)
               intercept(pid);
         }
      }
      else
      {
         /* Remove any nonmatching process from the breakpoint scope. */
         dcb->debug_flags &= ~IN_BREAK_SCOPE;
      }
   }

   /* Unlock the dispatcher. */
   UNLOCK_SWAP();
}




/*
** Subfunction to do_modify.
**
** It is called when updating the breakpoint scope.
*/

static void
set_evact_catch_scope(int scope_count, const char *scope_strings)
{
   struct scope   *scope;
   struct scope   *oldscope;
   union SIGNAL   *catch_status;
   struct procnamelist *plist;
   struct procnamelist *oldplist;

   /* Parse the process name list. */
   if (scope_count)
      plist = wildcard_to_plist((U32) scope_count, scope_strings);
   else
      plist = wildcard_to_plist((U32) 1, "*");

   /* Done if there was a syntax error. */
   if (!plist)
      return;

   /* Create a new scope buffer. This fails if the trace is horribly full. */
   scope = copy_scope(scope_count, scope_strings);

   /* handle this failure by ignoring the request. */
   if (scope)
   {
      /* Replace the scope buffer pointer. */
      oldscope = rw.breakpoint_scope;
      rw.breakpoint_scope = scope;

      rw.precious->breakpoint_scope = scope;
      /* Update the precious checksum */
      update_precious_checksum();

      oldplist = rw.quick_breakpoint_scope;
      rw.quick_breakpoint_scope = plist;

      /* Free the old scope buffer. */
      ose_trace_commit_inline(oldscope);
      free_buf((union SIGNAL **) &oldplist);

      /* Disable the catcher but save the previous state. */
      BKL_LOCK();
      catch_status = rw.catch_enable;
      rw.catch_enable = NULL;
      BKL_UNLOCK();

      /* Walk through all processes. */
      for_all_processes(set_one_breakpoint_scope, (U32) 0);

      /* Walk through the breakpoint scope cache. */
      {
         register struct break_pid_entry *bke;
         register PROCESS *pid;
         register U32    count;
         struct debug_pcb *dcb;

         for (bke = rw.break_pid_list; bke; bke = bke->next)
            for (count = bke->size, pid = bke->pid; count--; pid++)
               if (*pid)
               {
                  BKL_LOCK();
                  dcb = zzget_debugpcb(*pid);

                  /* Delete dead processes from the cache. */
                  if (!dcb)
                  {
                     *pid = (PROCESS) 0;
                     bke->use_count--;
                     BKL_UNLOCK();
                  }
                  /* Delete processes that are no longer in the scope from the cache. */
                  else if (!(dcb->debug_flags & IN_BREAK_SCOPE))
                  {
                     PROCESS         lpid = *pid;

                     /* Resume removed processes if required. */
                     *pid = (PROCESS) 0;
                     bke->use_count--;
                     BKL_UNLOCK();
                     if (rw.break_scope_status)
                        resume(lpid);
                  }
                  else
                  {
                     BKL_UNLOCK();
                  }
               }
      }

      /* Enable the catcher if previously enabled. */
      rw.catch_enable = catch_status;
   }
   /* Free the plist if we couldn't allocate trace space. */
   else
   {
      free_buf((union SIGNAL **) &plist);
   }
}

/*
** Subfunction to set_evact_monitor_scope.
**
** It is called for each process when updating the debugger scope.
**
** Lock the dispatcher for each process.
** Update pcb flags in all processes.
** Add new processes to the breakpoint scope cache.
** Intercept new processes if required.
*/

static void
set_one_debugger_scope(PROCESS pid, U32 flags)
{
   struct debug_pcb *dcb;

   /* Remove ``not used arg'' compiler warning. */
   (void) flags;

   /* Lock the dispatcher. */
   LOCK_SWAP();

   /* Get the debugger pcb. */
   dcb = zzget_debugpcb(pid);

   /* Skip dead processes. */
   if (!dcb)
   {
      UNLOCK_SWAP();
      return;
   }

   /* Check if the process is visible to the offspring ?? */

   /* Check if the process is in the debugger scope. */
   if (zzrmm_match_proc(dcb, rw.quick_debugger_scope))
   {
      /* Set the debugger scope flag. */
      dcb->debug_flags |= IN_DEBUGGER_SCOPE;
   }
   else
   {
      /* Otherwise remove the process from the display scope required. */
      dcb->debug_flags &= ~IN_DEBUGGER_SCOPE;
   }

   /* Unlock the dispatcher. */
   UNLOCK_SWAP();
}

/*
** It is called when updating the display scope.
*/

static void
set_evact_monitor_scope(int scope_count, const char *scope_strings)
{
   struct scope   *scope;
   struct scope   *oldscope;
   union SIGNAL   *catch_status;
   struct procnamelist *plist;
   struct procnamelist *oldplist;

   /* Parse the process name list. */
   if (scope_count)
      plist = wildcard_to_plist((U32) scope_count, scope_strings);
   else
      plist = wildcard_to_plist((U32) 1, "*");

   /* Done if there was a syntax error. */
   if (!plist)
   {
      return;
   }

   /* Create a new scope buffer. This fails if the trace is horribly full. */
   scope = copy_scope(scope_count, scope_strings);

   /* handle this failure by ignoring the request. */
   if (scope)
   {
      /* Replace the scope buffer pointer. */
      oldscope = rw.debugger_scope;
      rw.debugger_scope = scope;

      rw.precious->debugger_scope = scope;
      /* Update the preciosu checksum */
      update_precious_checksum();

      oldplist = rw.quick_debugger_scope;
      rw.quick_debugger_scope = plist;

      /* Free the old scope buffer. */
      ose_trace_commit_inline(oldscope);
      free_buf((union SIGNAL **) &oldplist);

      /* Disable the catcher but save the previous state. */
      BKL_LOCK();
      catch_status = rw.catch_enable;
      rw.catch_enable = NULL;
      BKL_UNLOCK();

      /* Walk through all processes. */
      for_all_processes(set_one_debugger_scope, (U32) 0);

      /* Enable the catcher if previously enabled. */
      rw.catch_enable = catch_status;
   }
   /* Free the plist if we couldn't allocate trace space. */
   else
   {
      free_buf((union SIGNAL **) &plist);
   }
}

static          uint32_t
get_new_evact_id(void)
{
   rw.precious->evact_count++;
   update_precious_checksum();

   return rw.precious->evact_count;
}

static void
set_event_actionpoint_failed(Offspring * offspring, MonitorStatus status)
{
   union SIGNAL   *reply = alloc(sizeof (MonitorSetEventActionpointReply), MONITOR_SET_EVENT_ACTIONPOINT_REPLY);

   reply->set_event_actionpoint_reply.status = status;
   reply->set_event_actionpoint_reply.id = rw.precious->evact_count + 1;
   send(&reply, offspring->pid);
}

static void
set_event_actionpoint(Offspring * offspring, struct ldefine_action *action, struct ldefine_action **action_list)
{
   uint32_t        id;
   union SIGNAL   *reply;
   OSBOOLEAN       reevaluate;

   assert_offspring(offspring);
   assert(action);
   assert(action_list);

   if (action->magic != ACTION_MAGIC_ALLOCATED)
   {
      ramlog_printf("rmm: internal error; illegal magic actionpoint value\n");
   }

   id = get_new_evact_id();

   action->action_id = id;
   action->next = NULL;
   action->prev = NULL;

   if (update_expanded_wildcards(action, 0))
   {
      ramlog_printf("rmm: got illegal wildcard from client.\n");
      free_action(action);
      set_event_actionpoint_failed(offspring, MONITOR_STATUS_REQUEST_INVALID);
      return;
   }

   action->event_count = action->orig_event_count;

   /* Insert the new entry into the action list. */
   if (*action_list)
   {
      struct ldefine_action *prev;

      reevaluate = (OSBOOLEAN) 0;
      for (prev = *action_list; prev->next; prev = prev->next)
      {
         ;                      /* Empty loop body. */
      }
      prev->next = action;
      action->prev = prev;
   }
   else
   {
      *action_list = action;
      reevaluate = (OSBOOLEAN) 1;
      /* Copy list pointers to precious structure. */
      update_precious_lists();
   }

   action->magic = ACTION_MAGIC_VALID;

   SMP_CODE(smp_distribute_actionpoint(action, action_list, False);)

   /* Enable the OSE debugger hooks in the kernel if required. */
   zzset_debug_status((OSBOOLEAN) 1);

   /* Reevaluate action conditions according to the new action list. */
   if (reevaluate)
   {
      update_hook_bits();
   }

   reply = alloc(sizeof (MonitorSetEventActionpointReply), MONITOR_SET_EVENT_ACTIONPOINT_REPLY);
   reply->set_event_actionpoint_reply.status = MONITOR_STATUS_OK;
   reply->set_event_actionpoint_reply.id = id;
   send(&reply, offspring->pid);
}

static int
check_evact_lock(Offspring * offspring)
{
   if (rw.evact_lock_owner != offspring)
   {
      set_event_actionpoint_failed(offspring, MONITOR_STATUS_REQUEST_INVALID);
      ramlog_printf("rmm: Client with no evact lock tried to set evact.\n");
      return 0;
   }

   return 1;
}

static uint32_t
validate_request(Offspring *offspring, union SIGNAL *sig)
{
   uint32_t signo = (uint32_t)sig->sig_no;
   MonitorEventActionpointInfo *req = &sig->event_actionpoint_info;
   uint32_t action = req->action;

   switch (action)
   {
      case MONITOR_ACTION_INTERCEPT:    
      case MONITOR_ACTION_TRACE:
      case MONITOR_ACTION_ENABLE_TRACE:
      case MONITOR_ACTION_DISABLE_TRACE:
      case MONITOR_ACTION_SET_STATE:
      case MONITOR_ACTION_NOTIFY:
      case MONITOR_ACTION_COUNT:
      case MONITOR_ACTION_USER:
      case MONITOR_ACTION_UNDO:
      case MONITOR_ACTION_ENABLE_HW_TRACE:
      case MONITOR_ACTION_DISABLE_HW_TRACE:
	 break;
      default:
	 set_event_actionpoint_failed(offspring, MONITOR_STATUS_REQUEST_INVALID);
	 ramlog_printf("rmm: action not supported\n");
	 return 0; 
   }   

   if (!check_evact_lock(offspring))
   {
      return 0;
   }

   if ((action == MONITOR_ACTION_ENABLE_HW_TRACE ||
       action == MONITOR_ACTION_DISABLE_HW_TRACE) && rw.hw_trace[0] == NULL)
   {
      set_event_actionpoint_failed(offspring, MONITOR_STATUS_REQUEST_INVALID);
      ramlog_printf("rmm: no functions available for hardware trace\n");
      return 0;
   }

   if ((signo == MONITOR_SET_SWAP_ACTIONPOINT_REQUEST) && 
      (MONITOR_ACTION_NOTIFY == action || MONITOR_ACTION_INTERCEPT == action))
   {
      set_event_actionpoint_failed(offspring, MONITOR_STATUS_REQUEST_INVALID);
      ramlog_printf("rmm: notification/intercept on swap event is not supported\n");
      return 0;
   }

   return 1;
}

static void
set_monitor_status(Offspring * offspring, int enabled)
{
   union SIGNAL   *sig;
   static const SIGSELECT sel[] = { 1, MONITOR_REPORT };

   assert_offspring(offspring);
   assert(0 == enabled || 1 == enabled);

   if (check_evact_lock(offspring))
   {
      if (enabled)
      {
         if (!rw.monitor_enabled)
         {
            SPIN_LOCK(&rw.event_notify_lock);
            rw.precious->master_monitor_enabled = rw.master_monitor_enabled = 1;
            rw.monitor_enabled = 1;
            rw.monitor_hook_flow = rw.flow_ctrl_nblock;
            rw.monitor_overflow = 0;
            SPIN_UNLOCK(&rw.event_notify_lock);
            /* Update the precious checksum */
            update_precious_checksum();
         }
      }
      else
      {
         SPIN_LOCK(&rw.event_notify_lock);
         rw.precious->master_monitor_enabled = rw.master_monitor_enabled = 0;
         rw.monitor_enabled = 0;
         rw.monitor_hook_flow = 0;
         rw.monitor_overflow = 0;
         SPIN_UNLOCK(&rw.event_notify_lock);
         /* Update the precious checksum */
         update_precious_checksum();
         /* Consume any pending monitor reports */
         while ((sig = receive_w_tmo(0, sel)))
         {
            free_buf(&sig);
         }
      }
   }
}

static void
get_actionpoint_ignore(MonitorEventActionpointInfo * req, uint32_t * ignore)
{
   MonitorTag     *t = req->tags;
   MonitorTag     *t1 = req->tags + req->tagsCount;

   while (t < t1)
   {
      if (t->tag.type == MONITOR_EVENT_ACTIONPOINT_TAG_IGNORE)
      {
         *ignore = t[1].num;
         return;
      }
      t += t->tag.count;
   }
   *ignore = 0;
}

static void
get_actionpoint_signo(MonitorEventActionpointInfo * req, uint32_t * signo0, uint32_t * signo1)
{
   MonitorTag     *t = req->tags;
   MonitorTag     *t1 = req->tags + req->tagsCount;

   while (t < t1)
   {
      if (t->tag.type == MONITOR_SIGNAL_FILTER_TAG_NUMBER)
      {
         *signo0 = t[1].num;
         *signo1 = t[2].num;
         return;
      }
      t += t->tag.count;
   }
   *signo0 = 0;
   *signo1 = 0;
}

static bool
get_actionpoint_data_range(MonitorEventActionpointInfo * req,
                           struct ldefine_action *action)
{
   MonitorTag     *t = req->tags;
   MonitorTag     *t1 = req->tags + req->tagsCount;

   while (t < t1)
   {
      if ((t->tag.type == MONITOR_EVENT_ACTIONPOINT_TAG_DATA_RANGE) ||
          (t->tag.type == MONITOR_EVENT_ACTIONPOINT_TAG_DATA_RANGE_NOT))
      {
         if (t->tag.count != 8 || t + t->tag.count > t1)
         {
            /* Illegal count */
            ramlog_printf("rmm: illegal tag count in data range tag"
                  " count %u total tags %u\n",
                  t->tag.count, req->tagsCount);
            return false;
         }
         
         action->data_range_size = t[3].num;
#ifdef LITTLE_ENDIAN
         action->data_range_offset = (((uint64_t)t[2].num) << 32) | t[1].num;
         action->data_range_a = (((uint64_t)t[5].num) << 32) | t[4].num;
         action->data_range_b = (((uint64_t)t[7].num) << 32) | t[6].num;
#else
         action->data_range_offset = (((uint64_t)t[1].num) << 32) | t[2].num;
         action->data_range_a = (((uint64_t)t[4].num) << 32) | t[5].num;
         action->data_range_b = (((uint64_t)t[6].num) << 32) | t[7].num;
#endif

         switch(action->data_range_size)
         {
            case 1:
            case 2:
            case 4:
            case 8:
               break;
            default:
               /* Illegal size */
               ramlog_printf("rmm: illegal size in data range tag %u\n",
                     action->data_range_size);
               return false;
         }
         if (action->data_range_b != 0 && action->data_range_a >= action->data_range_b)
         {
            /* Illegal range */
            ramlog_printf("rmm: illegal data range in tag %llu %llu\n",
                  action->data_range_a,
                  action->data_range_b);
            return false;
         }
         if (t->tag.type == MONITOR_EVENT_ACTIONPOINT_TAG_DATA_RANGE_NOT)
         {
            action->data_range_size |= 1U << 31; /* Let the msb represent negation */
         }
         return true;
      }
      t += t->tag.count;
   }
   action->data_range_offset = 0;
   action->data_range_size = 0;
   action->data_range_a = 0;
   action->data_range_b = 0;
   
   return true;
}

static int
set_alloc_actionpoint(Offspring *offspring, union SIGNAL *sig)
{
   MonitorEventActionpointInfo *req = &sig->event_actionpoint_info;

   assert(req);
   assert(offspring);

   if (validate_request(offspring, sig))
   {
      uint32_t        ignore;
      uint32_t        fromlen;
      uint32_t        signo0, signo1;

      get_actionpoint_ignore(req, &ignore);
      get_actionpoint_signo(req, &signo0, &signo1);

      fromlen = scope_to_pattern_len(&req->from, req->tags, req->tagsCount, 0);

      if (fromlen)
      {
         struct ldefine_action *action;
         uint32_t        actionlen = sizeof (struct ldefine_action) + fromlen + 1;

         action = alloc_action(actionlen);

         if (action)
         {
            assert(req->state <= 255);
            action->event_seq_no = req->state;

            if (MONITOR_ACTION_NOTIFY && (req->parameter & MONITOR_EVENT_ACTIONPOINT_PARAMETER_DATA))
               action->action = 10;
            else
               action->action = req->action;

            action->negated = req->not;

            if (MONITOR_ACTION_SET_STATE == req->action)
               action->goto_list = req->parameter;
            else
               action->goto_list = 0;

            action->is_basic_event = 0;
            action->signo = signo0;
            action->signo_end = signo1 - 1;
            assert(ignore < 65536);
            action->orig_event_count = ignore;
            action->ecode = 0;  /* not valid  */
            action->event_type = MONITOR_EVENT_ALLOC;
            action->sig_from = 0;
            action->sig_to = fromlen + 1;

            scope_to_pattern(&req->from, req->tags, req->tagsCount, 0, action->names);

           if (NULL == rw.actions SMP_CODE([0]).action_list_alloc)
           {
             set_event_actionpoint(offspring, action, &rw.actions SMP_CODE([0]).action_list_alloc);
             zzkrn_set_handler(12,(void *) zzalloc_dbghook);
           }
           else
           {
             set_event_actionpoint(offspring, action, &rw.actions SMP_CODE([0]).action_list_alloc);
           }
         }
         else
         {
            set_event_actionpoint_failed(offspring, MONITOR_STATUS_MEMORY_EXHAUSTED);
         }
      }
      else
      {
         set_event_actionpoint_failed(offspring, MONITOR_STATUS_REQUEST_INVALID);
      }
   }
   return 0;
}

static int
set_free_buf_actionpoint(Offspring *offspring, union SIGNAL *sig)
{
   MonitorEventActionpointInfo *req = &sig->event_actionpoint_info;

   assert(req);
   assert(offspring);

   if (validate_request(offspring, sig))
   {
      uint32_t        ignore;
      uint32_t        fromlen;
      uint32_t        signo0, signo1;

      get_actionpoint_ignore(req, &ignore);
      get_actionpoint_signo(req, &signo0, &signo1);

      fromlen = scope_to_pattern_len(&req->from, req->tags, req->tagsCount, 0);

      if (fromlen)
      {
         struct ldefine_action *action;
         uint32_t        actionlen = sizeof (struct ldefine_action) + fromlen + 1;

         action = alloc_action(actionlen);

         if (action)
         {
            assert(req->state <= 255);
            action->event_seq_no = req->state;

            if (get_actionpoint_data_range(req, action))
            {
              if (MONITOR_ACTION_NOTIFY && (req->parameter & MONITOR_EVENT_ACTIONPOINT_PARAMETER_DATA))
                 action->action = 10;
              else
                 action->action = (uint8_t)req->action;

              action->negated = (uint8_t)req->not;

              if (MONITOR_ACTION_SET_STATE == req->action)
                 action->goto_list = req->parameter;
              else
                 action->goto_list = 0;

              action->is_basic_event = 0;
              action->signo = signo0;
              action->signo_end = signo1 - 1;
              assert(ignore < 65536);
              action->orig_event_count = ignore;
              action->ecode = 0;  /* not valid  */
              action->event_type = MONITOR_EVENT_FREE_BUF;
              action->sig_from = 0;
              action->sig_to = fromlen + 1;

              scope_to_pattern(&req->from, req->tags, req->tagsCount, 0, action->names);

                if (NULL == rw.actions SMP_CODE([0]).action_list_free_buf)
                {
                   set_event_actionpoint(offspring, action, &rw.actions SMP_CODE([0]).action_list_free_buf);
                   zzkrn_set_handler(13,(void *) zzfree_buf_dbghook);
                }
                else
                {
                   set_event_actionpoint(offspring, action, &rw.actions SMP_CODE([0]).action_list_free_buf);
                }
             }
             else
             {
               set_event_actionpoint_failed(offspring, MONITOR_STATUS_REQUEST_INVALID);
             }
         }
         else
         {
            set_event_actionpoint_failed(offspring, MONITOR_STATUS_MEMORY_EXHAUSTED);
         }
      }
      else
      {
         set_event_actionpoint_failed(offspring, MONITOR_STATUS_REQUEST_INVALID);
      }
   }
   return 0;
}

static int
set_send_actionpoint(Offspring * offspring, union SIGNAL *sig)
{
   MonitorEventActionpointInfo *req = &sig->event_actionpoint_info;

   assert(req);
   assert(offspring);

   if (validate_request(offspring, sig))
   {
      uint32_t        ignore;
      uint32_t        fromlen, tolen;
      uint32_t        signo0, signo1;

      get_actionpoint_ignore(req, &ignore);
      get_actionpoint_signo(req, &signo0, &signo1);

      fromlen = scope_to_pattern_len(&req->from, req->tags, req->tagsCount, 0);
      tolen = scope_to_pattern_len(&req->to, req->tags, req->tagsCount, req->from.type);

      if (fromlen && tolen)
      {
         struct ldefine_action *action;
         uint32_t        actionlen = sizeof (struct ldefine_action) + fromlen + 1 + tolen + 1;

         action = alloc_action(actionlen);

         if (action)
         {
            assert(req->state <= 255);
            action->event_seq_no = req->state;

            if ((MONITOR_ACTION_NOTIFY == req->action) &&
                (req->parameter & MONITOR_EVENT_ACTIONPOINT_PARAMETER_DATA))
            {
               action->action = 10;
            }
            else
            {
               action->action = req->action;
               if ((MONITOR_ACTION_TRACE == action->action) &&
                   (req->parameter & MONITOR_EVENT_ACTIONPOINT_PARAMETER_DATA))
               {
                  action->action |= TRACE_WITH_DATA;
               }
            }

            if (get_actionpoint_data_range(req, action))
            {
               action->negated = (uint8_t)req->not;

               if (MONITOR_ACTION_SET_STATE == req->action)
                  action->goto_list = (uint8_t)req->parameter;
               else
                  action->goto_list = 0;

               action->is_basic_event = 0;
               action->signo = signo0;
               action->signo_end = signo1 - 1;
               assert(ignore < 65536);
               action->orig_event_count = ignore;
               action->ecode = 0;  /* not valid  */
               action->event_type = MONITOR_EVENT_SEND;
               action->sig_from = 0;
               action->sig_to = fromlen + 1;

               scope_to_pattern(&req->from, req->tags, req->tagsCount, 0, action->names);
               scope_to_pattern(&req->to, req->tags, req->tagsCount, req->from.type, action->names + fromlen + 1);

               set_event_actionpoint(offspring, action, &rw.actions SMP_CODE([0]).action_list_send);
            }
            else
            {
               set_event_actionpoint_failed(offspring,MONITOR_STATUS_REQUEST_INVALID);
            }
               
         }
         else
         {
            set_event_actionpoint_failed(offspring, MONITOR_STATUS_MEMORY_EXHAUSTED);
         }
      }
      else
      {
         set_event_actionpoint_failed(offspring, MONITOR_STATUS_REQUEST_INVALID);
      }
   }
   return 0;
}

static int
set_receive_actionpoint(Offspring * offspring, union SIGNAL *sig)
{
   MonitorEventActionpointInfo *req = &sig->event_actionpoint_info;

   assert(req);
   assert(offspring);

   if (validate_request(offspring, sig))
   {
      uint32_t        ignore;
      uint32_t        fromlen, tolen;
      uint32_t        signo0, signo1;

      get_actionpoint_ignore(req, &ignore);
      get_actionpoint_signo(req, &signo0, &signo1);

      fromlen = scope_to_pattern_len(&req->from, req->tags, req->tagsCount, 0);
      tolen = scope_to_pattern_len(&req->to, req->tags, req->tagsCount, req->from.type);

      if (fromlen && tolen)
      {
         struct ldefine_action *action;
         uint32_t        actionlen = sizeof (struct ldefine_action) + fromlen + 1 + tolen + 1;

         action = alloc_action(actionlen);

         if (action)
         {
            assert(req->state <= 255);
            action->event_seq_no = (uint8_t)req->state;

            if ((MONITOR_ACTION_NOTIFY == req->action) &&
                (req->parameter & MONITOR_EVENT_ACTIONPOINT_PARAMETER_DATA))
            {
               action->action = 10;
            }
            else
            {
               action->action = (uint8_t)req->action;
               if ((MONITOR_ACTION_TRACE == action->action) &&
                   (req->parameter & MONITOR_EVENT_ACTIONPOINT_PARAMETER_DATA))
               {
                  action->action |= TRACE_WITH_DATA;
               }
            }

            if (get_actionpoint_data_range(req, action))
            {
               action->negated = (uint8_t)req->not;

               if (MONITOR_ACTION_SET_STATE == req->action)
                  action->goto_list = (uint8_t)req->parameter;
               else
                  action->goto_list = 0;

               action->is_basic_event = 0;
               action->signo = signo0;
               action->signo_end = signo1 - 1;
               assert(ignore < 65536);
               action->orig_event_count = ignore;
               action->ecode = 0;  /* not valid  */
               action->event_type = 5;
               action->sig_from = 0;
               action->sig_to = fromlen + 1;

               scope_to_pattern(&req->from, req->tags, req->tagsCount, 0, action->names);
               scope_to_pattern(&req->to, req->tags, req->tagsCount, req->from.type, action->names + fromlen + 1);

               set_event_actionpoint(offspring, action, &rw.actions SMP_CODE([0]).action_list_receive);
            }
            else
            {
               set_event_actionpoint_failed(offspring,MONITOR_STATUS_REQUEST_INVALID);
            }            
         }
         else
         {
            set_event_actionpoint_failed(offspring, MONITOR_STATUS_MEMORY_EXHAUSTED);
         }
      }
      else
      {
         set_event_actionpoint_failed(offspring, MONITOR_STATUS_REQUEST_INVALID);
      }
   }
   return 0;
}

#ifdef HAS_TIMEOUT_EVENT
static int
set_timeout_actionpoint(Offspring * offspring, union SIGNAL *sig)
{
   MonitorEventActionpointInfo *req = &sig->event_actionpoint_info;

   assert(req);
   assert(offspring);

   if (validate_request(offspring, sig))
   {
      uint32_t        ignore;
      uint32_t        fromlen;

      get_actionpoint_ignore(req, &ignore);

      fromlen = scope_to_pattern_len(&req->from, req->tags, req->tagsCount, 0);

      if (fromlen)
      {
         struct ldefine_action *action;
	 uint32_t        actionlen = sizeof (struct ldefine_action) + fromlen + 1;

         action = alloc_action(actionlen);

         if (action)
         {
            assert(req->state <= 255);
            action->event_seq_no = req->state;

            if ((MONITOR_ACTION_NOTIFY == req->action) &&
                (req->parameter & MONITOR_EVENT_ACTIONPOINT_PARAMETER_DATA))
            {
               action->action = 10;
            }
            else
            {
               action->action = req->action;
            }

            action->negated = req->not;

            if (MONITOR_ACTION_SET_STATE == req->action)
               action->goto_list = req->parameter;
            else
               action->goto_list = 0;

            action->is_basic_event = 0;
            action->signo = 0;
            action->signo_end = 0;
            assert(ignore < 65536);
            action->orig_event_count = ignore;
            action->ecode = 0;  /* not valid  */
            action->event_type = MONITOR_EVENT_TIMEOUT;
            action->sig_from = 0;
            action->sig_to = fromlen + 1;

	    scope_to_pattern(&req->from, req->tags, req->tagsCount, 0, action->names);
            set_event_actionpoint(offspring, action, &rw.actions SMP_CODE([0]).action_list_timeout);
         }
         else
         {
            set_event_actionpoint_failed(offspring, MONITOR_STATUS_MEMORY_EXHAUSTED);
         }
      }
      else
      {
         set_event_actionpoint_failed(offspring, MONITOR_STATUS_REQUEST_INVALID);
      }
   }
   return 0;
}
#endif

static int
set_swap_actionpoint(Offspring * offspring, union SIGNAL *sig)
{
   MonitorEventActionpointInfo *req = &sig->event_actionpoint_info;

   assert(req);
   assert(offspring);

   if (validate_request(offspring, sig))
   {
      uint32_t        ignore;
      uint32_t        fromlen, tolen;

      get_actionpoint_ignore(req, &ignore);

      fromlen = scope_to_pattern_len(&req->from, req->tags, req->tagsCount, 0);
      tolen = scope_to_pattern_len(&req->to, req->tags, req->tagsCount, req->from.type);

      if (fromlen && tolen)
      {
         struct ldefine_action *action;
         uint32_t actionlen = sizeof (struct ldefine_action) + fromlen + 1 + tolen + 1;

         action = alloc_action(actionlen);

         if (action)
         {
            assert(req->state <= 255);
            action->event_seq_no = req->state;
            action->action = req->action;
            action->negated = req->not;

            if (MONITOR_ACTION_SET_STATE == req->action)
               action->goto_list = req->parameter;
            else
               action->goto_list = 0;

            action->is_basic_event = 0;
            action->signo = 0;
            action->signo_end = 0;
            assert(ignore < 65536);
            action->orig_event_count = ignore;
            action->ecode = 0;  /* not valid  */
            action->event_type = 3;
            action->sig_from = 0;
            action->sig_to = fromlen + 1;

            scope_to_pattern(&req->from, req->tags, req->tagsCount, 0, action->names);
            scope_to_pattern(&req->to, req->tags, req->tagsCount, req->from.type, action->names + fromlen + 1);

            set_event_actionpoint(offspring, action, &rw.actions SMP_CODE([0]).action_list_swap);
	    check_enable_swap_trace_opt(offspring);
         }
         else
         {
            set_event_actionpoint_failed(offspring, MONITOR_STATUS_MEMORY_EXHAUSTED);
         }
      }
      else
      {
         set_event_actionpoint_failed(offspring, MONITOR_STATUS_REQUEST_INVALID);
      }
   }
   return 0;
}

static int
set_create_actionpoint(Offspring * offspring, union SIGNAL *sig)
{
   MonitorEventActionpointInfo *req = &sig->event_actionpoint_info;

   assert(req);
   assert(offspring);

   if (validate_request(offspring, sig))
   {
      uint32_t        ignore;
      uint32_t        tolen;

      get_actionpoint_ignore(req, &ignore);

      tolen = scope_to_pattern_len(&req->to, req->tags, req->tagsCount, req->from.type);

      if (tolen)
      {
         struct ldefine_action *action;
         uint32_t        actionlen = sizeof (struct ldefine_action) + tolen + 1;

         action = alloc_action(actionlen);

         if (action)
         {
            assert(req->state <= 255);
            action->event_seq_no = req->state;
            action->action = req->action;
            action->negated = req->not;

            if (MONITOR_ACTION_SET_STATE == req->action)
               action->goto_list = req->parameter;
            else
               action->goto_list = 0;

            action->is_basic_event = 0;
            action->signo = 0;
            action->signo_end = 0;
            assert(ignore < 65536);
            action->orig_event_count = ignore;
            action->ecode = 0;  /* not valid  */
            action->event_type = 1;
            action->sig_from = 0;
            action->sig_to = 0;

            scope_to_pattern(&req->to, req->tags, req->tagsCount, 0, action->names);

            set_event_actionpoint(offspring, action, &rw.actions SMP_CODE([0]).action_list_create);
         }
         else
         {
            set_event_actionpoint_failed(offspring, MONITOR_STATUS_MEMORY_EXHAUSTED);
         }
      }
      else
      {
         set_event_actionpoint_failed(offspring, MONITOR_STATUS_REQUEST_INVALID);
      }
   }
   return 0;
}

static int
set_kill_actionpoint(Offspring * offspring, union SIGNAL *sig)
{
   MonitorEventActionpointInfo *req = &sig->event_actionpoint_info;

   assert(req);
   assert(offspring);

   if (validate_request(offspring, sig))
   {
      uint32_t        ignore;
      uint32_t        tolen;

      get_actionpoint_ignore(req, &ignore);

      tolen = scope_to_pattern_len(&req->to, req->tags, req->tagsCount, req->from.type);

      if (tolen)
      {
         struct ldefine_action *action;
         uint32_t        actionlen = sizeof (struct ldefine_action) + tolen + 1;

         action = alloc_action(actionlen);

         if (action)
         {
            assert(req->state <= 255);
            action->event_seq_no = req->state;
            action->action = req->action;
            action->negated = req->not;

            if (MONITOR_ACTION_SET_STATE == req->action)
               action->goto_list = req->parameter;
            else
               action->goto_list = 0;

            action->is_basic_event = 0;
            action->signo = 0;
            action->signo_end = 0;
            assert(ignore < 65536);
            action->orig_event_count = ignore;
            action->ecode = 0;  /* not valid  */
            action->event_type = 2;
            action->sig_from = 0;
            action->sig_to = 0;

            scope_to_pattern(&req->to, req->tags, req->tagsCount, 0, action->names);

            set_event_actionpoint(offspring, action, &rw.actions SMP_CODE([0]).action_list_kill);
         }
         else
         {
            set_event_actionpoint_failed(offspring, MONITOR_STATUS_MEMORY_EXHAUSTED);
         }
      }
      else
      {
         set_event_actionpoint_failed(offspring, MONITOR_STATUS_REQUEST_INVALID);
      }
   }
   return 0;
}

#ifdef OSE_SMP
static int
set_bind_actionpoint(Offspring * offspring, union SIGNAL *sig)
{
   MonitorEventActionpointInfo *req = &sig->event_actionpoint_info;

   assert(req);
   assert(offspring);

   if (validate_request(offspring, sig))
   {
      uint32_t        ignore;
      uint32_t        tolen;

      get_actionpoint_ignore(req, &ignore);

      tolen = scope_to_pattern_len(&req->to, req->tags, req->tagsCount, req->from.type);

      if (tolen)
      {
         struct ldefine_action *action;
         uint32_t        actionlen = sizeof (struct ldefine_action) + tolen + 1;

         action = alloc_action(actionlen);

         if (action)
         {
            assert(req->state <= 255);
            action->event_seq_no = req->state;
            action->action = req->action;
            action->negated = req->not;

            if (MONITOR_ACTION_SET_STATE == req->action)
               action->goto_list = req->parameter;
            else
               action->goto_list = 0;

            action->is_basic_event = 0;
            action->signo = 0;
            action->signo_end = 0;
            assert(ignore < 65536);
            action->orig_event_count = ignore;
            action->ecode = 0;  /* not valid  */
            action->event_type = 14;
            action->sig_from = 0;
            action->sig_to = 0;

            scope_to_pattern(&req->to, req->tags, req->tagsCount, 0, action->names);

            set_event_actionpoint(offspring, action, &rw.actions SMP_CODE([0]).action_list_bind);
         }
         else
         {
            set_event_actionpoint_failed(offspring, MONITOR_STATUS_MEMORY_EXHAUSTED);
         }
      }
      else
      {
         set_event_actionpoint_failed(offspring, MONITOR_STATUS_REQUEST_INVALID);
      }
   }
   return 0;
}
#endif /* OSE_SMP */

static void
get_error_actionpoint_exec(MonitorEventActionpointInfo * req, uint32_t * exec)
{
   uint32_t       not = 0;
   MonitorTag     *t = req->tags;
   MonitorTag     *t1 = req->tags + req->tagsCount;

   while (t < t1)
   {
      if (t->tag.type == MONITOR_ERROR_FILTER_TAG_EXEC)
      {
         *exec = !not;
         return;
      }
      if (t->tag.type == MONITOR_ERROR_FILTER_TAG_NOT)
         not = 1;
      else
         not = 0;

      t += t->tag.count;
   }
   *exec = 0;
}

static void
get_error_actionpoint_error(MonitorEventActionpointInfo * req, uint32_t * error0, uint32_t * error1)
{
   MonitorTag     *t = req->tags;
   MonitorTag     *t1 = req->tags + req->tagsCount;

   while (t < t1)
   {
      if (t->tag.type == MONITOR_ERROR_FILTER_TAG_ERROR)
      {
         *error0 = t[1].num;
         *error1 = t[2].num;
         return;
      }
      t += t->tag.count;
   }
   *error0 = 0;
   *error1 = 0;
}

static int
set_error_actionpoint(Offspring * offspring, union SIGNAL *sig)
{
   MonitorEventActionpointInfo *req = &sig->event_actionpoint_info;

   assert(req);
   assert(offspring);

   if (validate_request(offspring, sig))
   {
      uint32_t        exec;
      uint32_t        error0, error1;
      uint32_t        ignore;
      uint32_t        fromlen;

      get_actionpoint_ignore(req, &ignore);
      get_error_actionpoint_exec(req, &exec);
      get_error_actionpoint_error(req, &error0, &error1);

      fromlen = scope_to_pattern_len(&req->from, req->tags, req->tagsCount, 0);

      if (fromlen)
      {
         struct ldefine_action *action;
         uint32_t        actionlen = sizeof (struct ldefine_action) + fromlen + 1;

         action = alloc_action(actionlen);

         if (action)
         {
            assert(req->state <= 255);
            action->event_seq_no = req->state;
            action->action = req->action;
            action->negated = req->not;

            if (MONITOR_ACTION_SET_STATE == req->action)
               action->goto_list = req->parameter;
            else
               action->goto_list = 0;

            action->is_basic_event = 0;
            action->signo = 0;
            action->signo_end = 0;
            assert(ignore < 65536);
            action->orig_event_count = ignore;
            action->ecode = error0;
            action->event_type = (exec) ? 6 : 11;
            action->sig_from = 0;
            action->sig_to = error1 - 1;        /* Use sig_to for upper error bound */

            scope_to_pattern(&req->from, req->tags, req->tagsCount, 0, action->names);

            set_event_actionpoint(offspring, action, (exec) ? &rw.actions SMP_CODE([0]).action_list_oserror : &rw.actions SMP_CODE([0]).action_list_error);
         }
         else
         {
            set_event_actionpoint_failed(offspring, MONITOR_STATUS_MEMORY_EXHAUSTED);
         }
      }
      else
      {
         set_event_actionpoint_failed(offspring, MONITOR_STATUS_REQUEST_INVALID);
      }
   }
   return 0;
}

static          MonitorStatus
set_basic_evact_scopes(Offspring * offspring)
{
   if (NULL == rw.evact_lock_owner)
   {
      char            str[5 + 10 + 1];  /* strlen("*:*:*") + strlen("0x00000000") + 1 */

      switch (offspring->debug_scope.type)
      {
         case MONITOR_SCOPE_ID:
            sprintf(str, "#%#x", offspring->debug_scope.id);
            break;
         case MONITOR_SCOPE_BLOCK_ID:
            sprintf(str, "#%#x:*", offspring->debug_scope.id);
            break;
         case MONITOR_SCOPE_SEGMENT_ID:
            sprintf(str, "#%#x:*:*", offspring->debug_scope.id);
            break;
         case MONITOR_SCOPE_SYSTEM:
            sprintf(str, "*:*:*");
            break;
         default:
            return MONITOR_STATUS_SCOPE_INVALID;
      }

      do_delete_all_actions(0, False);
      rw.evact_lock_owner = offspring;
      set_evact_catch_scope(1, str);
      set_evact_monitor_scope(1, str);
      enable_catch(offspring);
      set_monitor_status(offspring, 1);
   }

   if (offspring != rw.evact_lock_owner)
   {
      return MONITOR_STATUS_SCOPE_OVERLAP;
   }

   return MONITOR_STATUS_OK;
}

static void
unset_basic_evact_scopes(Offspring * offspring)
{
   if (rw.evact_lock_owner == offspring)
   {
      disable_catch();
      set_monitor_status(offspring, 0);
      rw.evact_lock_owner = NULL;
   }
}

static          MonitorStatus
set_basic_event_action(Offspring * offspring, U8 action_type, U8 event_type)
{
   struct ldefine_action *action;
   struct ldefine_action **action_list;
   uint32_t        reevaluate = 0;
   uint16_t        count;
   char            pattern[] = "*\0";

   (void)offspring;
   switch (event_type)
   {
      case 1:
         action_list = &rw.actions SMP_CODE([0]).action_list_create;
         break;
      case 2:
         action_list = &rw.actions SMP_CODE([0]).action_list_kill;
         break;
      case 6:
         action_list = &rw.actions SMP_CODE([0]).action_list_oserror;
         break;
      case 11:
         action_list = &rw.actions SMP_CODE([0]).action_list_error;
         break;
      default:
         return MONITOR_STATUS_INTERNAL_ERROR;
   }

   count = sizeof (struct ldefine_action) + (sizeof (pattern) - 1) - sizeof (uint32_t);
   action = alloc_action(count + sizeof (uint32_t));

   if (NULL == action)
   {
      return MONITOR_STATUS_MEMORY_EXHAUSTED;
   }

   action->orig_action = action;
   action->next = NULL;
   action->prev = NULL;
   action->signo_end = 0;
   action->action_id = get_new_evact_id();
   action->event_seq_no = 0;
   action->action = action_type;
   action->negated = 0;
   action->goto_list = 0;
   action->signo = 0;
   action->sig_from = 0;        /* Index into names of "*" */
   action->sig_to = ~0;         /* Match all errors */
   action->ecode = 0;           /* Match all */
   action->orig_event_count = 0;
   action->event_count = 0;
   action->event_type = event_type;
   action->is_basic_event = 1;
   memcpy(action->names, pattern, sizeof (pattern));

   if (update_expanded_wildcards(action, 0))
   {
      free_action(action);
      return MONITOR_STATUS_INTERNAL_ERROR;
   }

   action->magic = ACTION_MAGIC_VALID;

   if (*action_list)
   {
      struct ldefine_action *prev;

      for (prev = *action_list; prev->next; prev = prev->next) ;

      prev->next = action;
      action->prev = prev;
   }
   else
   {
      reevaluate = 1;

      *action_list = action;
      update_precious_lists();
   }

   zzset_debug_status((OSBOOLEAN) 1);

   if (reevaluate)
   {
      update_hook_bits();
   }

   SMP_CODE(smp_distribute_actionpoint(action, action_list, False);)

   return MONITOR_STATUS_OK;
}

static int
handle_basic_event_action_request(Offspring * offspring, union SIGNAL *sig)
{
   uint32_t        event1, event2;
   MonitorBasicEventActionReply *rep;
   MonitorBasicEventActionRequest *req;

   req = &sig->basic_event_action_request;
   rep = (void *) alloc(sizeof (MonitorBasicEventActionReply), 0);

   switch (sig->sig_no)
   {
      case MONITOR_BASIC_ERROR_ACTION_REQUEST:
         event1 = 6;            /* OS error */
         event2 = 11;           /* User error */
         rep->sigNo = MONITOR_BASIC_ERROR_ACTION_REPLY;
         break;
      case MONITOR_BASIC_CREATE_KILL_ACTION_REQUEST:
         event1 = 1;            /* Create */
         event2 = 2;            /* Kill */
         rep->sigNo = MONITOR_BASIC_CREATE_KILL_ACTION_REPLY;
         break;
      default:
         assert(0);
         return 0;
   }

   rep->status = set_basic_evact_scopes(offspring);
   if (!rep->status)
   {
      rep->status = set_basic_event_action(offspring, req->action, event1);
      if (!rep->status)
      {
         rep->status = set_basic_event_action(offspring, req->action, event2);
      }
   }
   if (rep->status)
   {
      unset_basic_evact_scopes(offspring);
   }

   send((union SIGNAL **) &rep, offspring->pid);
   return 0;
}

static EntrySize
get_trace_name_length(const struct debug_pcb *proc)
{
   int fullnamelen = strlen(proc->name) + strlen(proc->blockname) + 1;

   return (EntrySize)(sizeof (struct trace_procname) + ((fullnamelen > TRACE_PROCNAME_MAX_NAMELEN)? 
                      TRACE_PROCNAME_MAX_NAMELEN : fullnamelen));
}

static void create_truncated_trace_name(char *d,
                                        const struct debug_pcb *proc,
                                        int namelen)
{
   char *s;

   if (namelen < TRACE_PROCNAME_MAX_NAMELEN)
   {
      for (s = proc->blockname; *s;)
         *d++ = *s++;
      *d++ = ':';
      for (s = proc->name; (*d++ = *s++);)
      {;                        /* Empty loop body. */
      }
   }
   else
   {
      /* Truncate names */
      int             i = 0;

      for (s = proc->blockname; i < 32 && *s; i++)
      {
         *d++ = *s++;
      }
      if (*s)
      {
         /* Note truncation by three dots. */
         *d++ = '.';
         *d++ = '.';
         *d++ = '.';
         i += 3;
      }
      *d++ = ':';
      i++;
      for (s = proc->name; i < 251 && *s; i++)
      {
         *d++ = *s++;
      }
      if (*s)
      {
         /* Note truncation by three dots. */
         *d++ = '.';
         *d++ = '.';
         *d++ = '.';
      }
      *d++ = '\0';
   }
}

/*
** Subfunction to the debugger hooks.
*/
static void
fill_in_trace_name(const struct debug_pcb *proc, void *dest, EntrySize len)
{
   struct trace_procname *proc_name = (struct trace_procname *)dest;
   int namelen = len - sizeof (struct trace_procname);

   /* Create a trace name containing both block and proc name that
    * is truncated if it exceeds 255 characters */
   create_truncated_trace_name(proc_name->name, proc, namelen);

   proc_name->event_no = TRACE_PROCNAME;
   proc_name->namelen = namelen;
   proc_name->pid = proc->pid;

   /* Save the priority field from the pcb. */
   proc_name->priority = get_pcb_priority_from_debug_pcb(proc);
}

/*
** Subfunction to the debugger hooks.
**
** Allocate a name entry in the trace for a process that does not have it yet.
** Returns a pointer to it or NULL if cannot be allocated.
*/

static struct trace_procname *
make_trace_name(struct debug_pcb *proc)
{
   LOCKSTATE       msr;
   struct trace_procname *proc_name;
   EntrySize             len;

   if (unlikely(proc->tracename))
   {
      return proc->tracename;
   }
   len = get_trace_name_length(proc);

   proc_name = ose_trace_alloc(rw.trace, len, T_HIDDEN | T_FORCE | T_NOTIME | T_REFERRED);

   if (unlikely(!proc_name))
   {
      return NULL;
   }
   fill_in_trace_name(proc, proc_name, len);

   BKL_LOCK_SAVE(msr);
   /* Check if the name was allocated by someone else. */
   if (unlikely(proc->tracename))
   {
      /* If so, use that buffer instead. */
      ose_trace_commit_inline((void *) proc_name);
      proc_name = proc->tracename;
      BKL_LOCK_RESTORE(msr);
      return proc_name;
   }

   /* Save the name for future use in the PCB. */
   proc->tracename = proc_name;
   BKL_LOCK_RESTORE(msr);

   return proc_name;
}

/*
** Subfunction to the debugger hooks.
*/
static union TRACE *
trace_alloc_from(EntrySize sz, struct debug_pcb *from,
                 struct trace_procname **from_trace_name)
{
   struct trace_procname *ftn = make_trace_name(from);
   char *e;

   if (!ftn)
   {
      EntrySize size = get_trace_name_length(from);
      EntrySize offset = (sz & 3) == 0? sz: sz + 4 - (sz & 3);

      e = ose_trace_alloc(rw.trace, offset + size, 0);
      if (!e)
         return NULL;
      ftn = (struct trace_procname *)(e + offset);
      fill_in_trace_name(from, ftn, size);
   }
   else
   {
       e = ose_trace_alloc(rw.trace, sz, 0);
   }

   *from_trace_name = ftn;

   return (union TRACE *)e;
}

/*
** Subfunction to the debugger hooks.
*/
static union TRACE *
trace_alloc_from_to(EntrySize sz,
                    struct debug_pcb *from,
                    struct trace_procname **from_trace_name,
                    struct debug_pcb *to,
                    struct trace_procname **to_trace_name)
{
   struct trace_procname *ftn = make_trace_name(to);
   if (!ftn)
   {
      EntrySize size = get_trace_name_length(to);
      EntrySize offset = (sz & 3) == 0? sz: sz + 4 - (sz & 3);
      char *e = (char *)trace_alloc_from(offset + size, from, from_trace_name);

      if (!e)
         return NULL;
      ftn = (struct trace_procname *)(e + offset);
      fill_in_trace_name(to, ftn, size);
      *to_trace_name = ftn;
      return (union TRACE *)e;
   }
   else
   {
      *to_trace_name = ftn;
      return trace_alloc_from(sz, from, from_trace_name);
   }
}

/*
** Subfunction to the debugger hooks.
**
** Allocate an absolute time stamp entry in the trace if there is
** none for the current time.
** Returns a pointer to it or NULL if cannot be allocated.
*/

static struct abstime *
make_trace_abstime(void)
{
   LOCKSTATE       msr;
   struct abstime *abstime;

   abstime = ose_trace_alloc(rw.trace, sizeof(struct abstime), T_HIDDEN
         | T_NOTIME | T_REFERRED);
   if (!abstime)
      return NULL;
   abstime->event_no = TRACE_ABSTIME;
   get_abstime((OSTIME *)&abstime->abs_sec,
               (OSTIME *)&abstime->abs_tick,
               (OSTIME *)&abstime->abs_usec);
   BKL_LOCK_SAVE(msr);
   /* Someone else allocated also? */
   if (rw.trace_abstime)
   {
      /* Yes, use the other one. */
      ose_trace_commit_inline((void *) abstime);
      abstime = rw.trace_abstime;
   }
   else
   {
      rw.trace_abstime = abstime;
   }
   BKL_LOCK_RESTORE(msr);
   return abstime;
}

static always_inline struct abstime *
get_trace_abstime(void)
{
   return rw.trace_abstime? rw.trace_abstime: make_trace_abstime();
}

/* function ose_rmm_set_user_action_handler
 * Used to set user specified functions
 * to receive user reports
*/
void
ose_rmm_set_user_action_handler(void (*handler)(union SIGNAL *, uint32_t, void *, OSBUFSIZE))
{
   rw.user_action_function = handler;
}

/*
 * send_proc_info send information about process to user action handler.
 */
static void
send_proc_info(struct debug_pcb *from)
{
   if (NULL != from)
   {
      EntrySize namelen;
      MonitorProcessInfo event_report_proc_info;
      char event_report_proc_info_name[TRACE_PROCNAME_MAX_NAMELEN + 1];

      event_report_proc_info.sigNo = MONITOR_GET_PROCESS_TRACE_REPLY;
      event_report_proc_info.pid = (U32) from->pid;
      event_report_proc_info.bid = pcb_to_bid(from);
      event_report_proc_info.sid = 0; /* We can not always use get_segid here */
      event_report_proc_info.type = make_ptype(from);
      event_report_proc_info.entrypoint = (OSADDRESS) from->entrypoint;
      event_report_proc_info.properties = (from->debug_flags & SYS_PROCESS) != 0;
      event_report_proc_info.reserved0 = 0;
      event_report_proc_info.name[0] = 0;

      /* Create the block and process name string */
      namelen = get_trace_name_length(from) - sizeof (struct trace_procname);
      create_truncated_trace_name(event_report_proc_info_name,
                                  from,
                                  namelen);

      rw.user_action_function((union SIGNAL *)&event_report_proc_info,
                           offsetof(MonitorProcessInfo, name),
                           event_report_proc_info_name,
                           strlen(event_report_proc_info_name) + 1);
   }
}

static always_inline bool
match_data_range(const uint8_t *data, OSBUFSIZE sig_size, const struct ldefine_action *action)
{
   if (unlikely(action->data_range_size > 0))
   {
      size_t N_size = action->data_range_size & 0xf;

      if (sig_size >= action->data_range_offset + N_size)
      {
         const uint8_t *d = ((const uint8_t *) data) + action->data_range_offset;
         uint64_t N;
         bool result;
         
         switch (N_size)
         {
            case 1:
               N = (uint64_t) *d;
               break;
            case 2:
               N = (uint64_t) *(const uint16_t *) d;
               break;
            case 4:
               N = (uint64_t) *(const uint32_t *) d;
               break;
            case 8:
               N = *(const uint64_t *) d;
               break;
            default:
               assert(false);
               return false;
         }
         result = action->data_range_a <= N
               && (0 == action->data_range_b || N < action->data_range_b);
         
         /* The msb of the size specifies if the result shall ne negated. */
         return (action->data_range_size & (1U << 31)) ? !result: result;
      }
      return false;
   }
   else
   {
      return true; /* No range specified. */
   }
}

/*
**===========================================================================
** Hooks called from POSE. FIXME What to do with these?
**===========================================================================
*/

static void    *
handle_send_receive_event(OSBUFSIZE sig_size, void *sig,
                          struct debug_pcb *from, struct debug_pcb *to,
                          struct ldefine_action *action_list, int event_no)
{
/*
** The actions flag field is defined as follows:
** D0 = 1 => Catch the event.                   Mask = 0x0003
** D1 = 1 => Ignore the catch bit.
** D2 = 1 => Monitor the event.                 Mask = 0x000c
** D3 = 1 => Ignore the monitor bit.
** D4 = 1 => Trace the event.                   Mask = 0x0030
** D5 = 1 => Ignore the trace bit.
** D6 = 1 => Turn on the trace.                 Mask = 0x00c0
** D7 = 1 => Ignore the trace on bit.
** D8 = 1 => Turn off the trace.                Mask = 0x0300
** D9 = 1 => Ignore the trace off bit.
** D10 = 1 => Turn on the watch timer.          Mask = 0x0c00
** D11 = 1 => Ignore the watch on bit.
** D12 = 1 => Turn off the watch timer.         Mask = 0x3000
** D13 = 1 => Ignore the watch off bit.
** D14 = 1 => Goto a new event sequence number. Mask = 0xc000
** D15 = 1 => Ignore the goto bit.
*/
   uint32_t actions = 0;
   struct ldefine_action *action;

   U32             to_pid;      /* Saved in case addressee gets killed. */

   U32 cpu;

   U8              oldevent_seq_no = rw.action_state;
   U8              newevent_seq_no = 0;

   U8              monitor_long_flag = 2;

   uint32_t        interceptionpoint = 0;
   uint32_t        notificationpoint = 0;
   uint32_t        trace_with_data = 0;

   /* Skip any event that involves two system process. */
   if ((from->debug_flags & to->debug_flags) & rw.sys_process)
   {
      return sig;
   }

   /* Skip events that don't involve any process in the display scope. */
   if (!((from->debug_flags | to->debug_flags) & IN_DEBUGGER_SCOPE))
      return sig;

   /* Scan the actions list for actions to perform. */
   /* The dispatcher must be locked while doing this. */
   for (action = action_list; action; action = action->next)
   {
      assert(validate_evact(action));
      /* Check for event sequence match. */
      if (action->event_seq_no && (action->event_seq_no != oldevent_seq_no))
         continue;

      /* Check for sender match. */
      if (!zzrmm_match_proc(from, (struct procnamelist *) action->align))
         continue;

      /* Check for addressee match. */
      if (!zzrmm_match_proc(to, (struct procnamelist *) action->ecode))
         continue;

      /* Check for signal number match. */
      if ((action->signo > *(SIGSELECT *) sig) || (action->signo_end < *(SIGSELECT *) sig))
         continue;

      if (unlikely(!match_data_range((uint8_t *)sig, sig_size, action)))
      {
         continue;
      }
            
      /* Check for an active event counter. */
      if (unlikely(action->orig_event_count))
      {
         /* Update the event counter and ignore the event if the limit is not reached. */
         LOCK();
         SPIN_LOCK(&action->SMP_CODE(orig_action->)action_lock);
         if (--(action->SMP_CODE(orig_action->)event_count))
         {
            SPIN_UNLOCK(&action->SMP_CODE(orig_action->)action_lock);
            UNLOCK();
            continue;
         }
         else
         {
            action->SMP_CODE(orig_action->)event_count = action->orig_event_count;
            SPIN_UNLOCK(&action->SMP_CODE(orig_action->)action_lock);
            UNLOCK();
         }
      }

      /* Decide which action to take and update action flags. */
      {
         U32             newaction = 0;

         switch (action->action & 0xf)
         {
            case 3:
               /* Trace. */
               newaction = 0x0010;
               trace_with_data |= action->action & TRACE_WITH_DATA;
               break;
            case 1:
               /* Catch. */
               interceptionpoint = action->action_id;
               newaction = 0x0001;
               break;
            case 10:
               /* Monitor -long */
               notificationpoint = action->action_id;
               monitor_long_flag = 10;
               newaction = 0x0004;
               break;
            case 2:
               /* Monitor. */
               notificationpoint = action->action_id;
               newaction = 0x0004;
               break;
            case 4:
               /* Trace on. */
               newaction = 0x0040;
               break;
            case 5:
               /* Trace off. */
               newaction = 0x0100;
               break;
            case 7:
               /* Watch on. */
               newaction = 0x0400;
               break;
            case 8:
               /* Watch off. */
               newaction = 0x1000;
               break;
            case 9:
               /* Goto. */
               newaction = 0x4000;
               /* Save the seq number to be for later. */
               newevent_seq_no = action->goto_list;
               break;
            case 12:
               /* User function */
               newaction = 0x10000;
               break;
	    case 13:
               /* Undo */
               newaction = 0x40000;
               break;
            case 14:
               /* Enable hw trace */
               newaction = 0x100000;
               break;
            case 15:
               /* Disable hw trace */
               newaction = 0x400000;
               break;
            default:
               break;
         }
         /* Negate the action if required. */
         if (action->negated)
            newaction <<= 1;
         /* Log the action for later. */
         actions |= newaction;
      }
   }

   /* Done if there are no actions to be performed. */
   if (!actions)
      return sig;

   /* Save the process id of the addressee in case it gets killed. */
   to_pid = to->pid;

   cpu = SMP_CODE_YES_NO(zzkrn_get_pid_info((event_no == TRACE_SEND) ? from->pid : to->pid, 13), 0);

   /* Call user specified function, if it is defined in ose_reg_user_func() */
   if ((rw.user_action_function != NULL) && ((actions & 0x30000) == 0x10000))
   {
      MonitorSendReceiveInfo info;

      info.sigNo = event_no == TRACE_SEND ? MONITOR_GET_SEND_TRACE_REPLY:
         MONITOR_GET_RECEIVE_TRACE_REPLY;

      info.reference = notificationpoint;
      BKL_LOCK();
      info.time.sec = rw.abs_sec;
      info.time.sectick = rw.abs_tick;
      info.time.tick = get_systime((OSTICK *) &info.time.ntick);
      BKL_UNLOCK();
      info.signalSender = (U32) from->pid;
      info.signalNumber = *(SIGSELECT *) sig;
      info.signalSize = sig_size;
      info.signalAddressee = (U32) to->pid;
      info.signalAddress = (U32) sig;
      info.signalId = 0;

      info.lineNumber = 0;
      info.fileNameOffset = 0;
      info.signalDataOffset = 0;
      info.signalDataCount = 0;
      info.euId = cpu;
      info.reserved0 = 0;
      info.reserved1 = 0;
      info.data[0] = 0;

      send_proc_info(from);

      send_proc_info(to);

      rw.user_action_function((union SIGNAL *)&info,
                           offsetof(MonitorSendReceiveInfo,data),
                           sig,
                           sig_size);
   }

   /* Turn on the trace if required. */
   if ((actions & 0x00c0) == 0x0040)
   {
      rw.precious->trace_enabled = rw.trace_enabled = (OSBOOLEAN) 1;
      ose_trace_enable(rw.trace);
   }

   /* Turn on the hardware trace if required. */
   if ((actions & 0x300000) == 0x100000)
   {
      (*rw.hw_trace[0])();
   }

   /* Goto a new event sequence number if required. */
   if ((actions & 0xc000) == 0x4000)
   {
      rw.precious->action_state = rw.action_state = newevent_seq_no;
   }

   /* Trace the event if required. */
   if ((actions & 0x0030) == 0x0010)
   {
      union TRACE    *te;
      struct trace_procname *from_name;
      struct trace_procname *to_name;
      uint32_t tr_sig_size;

      if (trace_with_data)
      {
         /* Trim the size if too large. */
         tr_sig_size = (sig_size > rw.max_trace_size) ? rw.max_trace_size : sig_size;
      }
      else
      {
         tr_sig_size = 4; /* allways include the signal number */
      }

      /* Allocate a trace buffer. */
      te = trace_alloc_from_to((EntrySize) (sizeof (struct trace_send_receive) + tr_sig_size),
                               from, &from_name, to, &to_name);

      /* Fill in the trace buffer. */
      if (te)
      {
         /* Fill the trace buffer. */
         te->trace_send_receive.event_no = (U16) event_no;
         te->trace_send_receive.abstime = get_trace_abstime();;
         te->trace_send_receive.sigsize = (U16) tr_sig_size;
         te->trace_send_receive.from = from_name;
         te->trace_send_receive.to = to_name;
         te->trace_send_receive.cpu = cpu;
         memcpy(te->trace_send_receive.data, sig, tr_sig_size);

         /* Release the trace buffer. */
         ose_trace_commit_inline(te);
      }
   }

   /* Turn off the trace if required. */
   if ((actions & 0x0300) == 0x0100)
   {
      rw.precious->trace_enabled = rw.trace_enabled = (OSBOOLEAN) 0;
      ose_trace_disable(rw.trace);
   }

   /* Turn off the hardware trace if required. */
   if ((actions & 0xc00000) == 0x400000)
   {
      (*rw.hw_trace[1])();
   }


   /* Monitor the event if required. */
   if ((actions & 0x000c) == 0x0004)
   {
      union SIGNAL   *txsig;

      /* Monitor condition was true. Send to ldm if there is space in the mon queue. */
      /* Check that monitoring is enabled and that there is space in the monitor queue. */
      if ((!rw.monitor_enabled) || (rw.monitor_overflow))
      {
         SPIN_LOCK(&rw.event_notify_lock);
         if (rw.monitor_overflow)
         {
            rw.monitor_overflow++;
         }
         SPIN_UNLOCK(&rw.event_notify_lock);
         goto cant_monitor;
      }

      /* Unlock the dispatcher. */
      UNLOCK_SWAP();
      /* Allocate a monitor buffer. */
      set_suid(1);
      if (monitor_long_flag == 10)
      {
         txsig = alloc(sizeof (struct event_report_verbose) +
                       ((sig_size > rw.max_monitor_size) ? rw.max_monitor_size : sig_size), MONITOR_REPORT);
         txsig->event_report_verbose.sig_datasize = (U16) ((sig_size > rw.max_monitor_size) ? rw.max_monitor_size : sig_size);
      }
      else
      {
         txsig = alloc(sizeof (struct event_report_verbose) + 4, MONITOR_REPORT);
         txsig->event_report_verbose.sig_datasize = (U16) 4;
      }

      BKL_LOCK();
      txsig->event_report_verbose.event_abssec = rw.abs_sec;
      txsig->event_report_verbose.event_abstick = rw.abs_tick;
      txsig->event_report_verbose.event_absutick = rw.abs_usec;
      BKL_UNLOCK();
      /* Look out with the parameter type to get_systime()! */
      txsig->event_report_verbose.event_tick = get_systime((OSTICK *) & txsig->event_report_verbose.event_utick);
      txsig->event_report_verbose.sig_from = (U32) from->pid;
      txsig->event_report_verbose.sig_to = (U32) to_pid;
      txsig->event_report_verbose.ecode = 0;
      txsig->event_report_verbose.signo = *(SIGSELECT *) sig;
      txsig->event_report_verbose.cpu = cpu;
      txsig->event_report_verbose.sig_data = 0;

      txsig->event_report_verbose.pid_from = txsig->event_report_verbose.sig_from;
      txsig->event_report_verbose.pid_to = txsig->event_report_verbose.sig_to;
      txsig->event_report_verbose.ptype_from = make_ptype(from);
      txsig->event_report_verbose.ptype_to = make_ptype(to);
      txsig->event_report_verbose.line = (U32) 0;
      txsig->event_report_verbose.file = (U32) 0;
      txsig->event_report_verbose.action_id = notificationpoint;

      txsig->event_report_verbose.event_type = event_no;
      txsig->event_report_verbose.sig_address = (U32) sig;

      txsig->event_report_verbose.action = monitor_long_flag;
      txsig->event_report_verbose.sig_size = sig_size;
      memcpy((void *) txsig->event_report_verbose.data,
             (void *) sig, (size_t) txsig->event_report_verbose.sig_datasize);

      /* Lock the dispatcher again. */
      LOCK_SWAP();
      /* Disable interrupts */
      SPIN_LOCK(&rw.event_notify_lock);
      /* Check again that monitoring is enabled and no overflow present. */
      assert(rw.monitor_hook_flow >= 0);
      if ((rw.monitor_enabled) && (0 == rw.monitor_hook_flow))
      {
         rw.monitor_overflow++;
      }
      if ((!rw.monitor_enabled) || (0 == rw.monitor_hook_flow))
      {
         SPIN_UNLOCK(&rw.event_notify_lock);
         /* Free the buffer and abort if disabled. */
         UNLOCK_SWAP();
         free_buf(&txsig);
         set_suid(0);
         LOCK_SWAP();
         goto cant_monitor;
      }
      /* Take a flow control token . */
      rw.monitor_hook_flow--;
      /* Enable interrupts */
      SPIN_UNLOCK(&rw.event_notify_lock);

      /* Send an event_report_verbose to the ldm. */
      zzsys_send_w_s(&txsig, rw.zzose_rmm_, rw.zzose_rmm_);
      set_suid(0);

   }
 cant_monitor:
   /* Catch the event if required. */
   if ((actions & 0x0003) == 0x0001)
   {
      /* Catch condition was true. Catch it if the scope is not stopped. */
      /* Disable interrupts */
      BKL_LOCK();
      /* Check that the system is not already caught. */
      if (rw.catch_enable)
      {
         union SIGNAL   *txsig;

         /* Disable catcher. */
         txsig = rw.catch_enable;
         rw.catch_enable = NULL;
         /* Store sig pointer and process id for the ldm to access. */
         rw.caught_pid = current_process();
         rw.caught_sig = (union SIGNAL **) &sig;
         txsig->event_report_verbose.event_abssec = rw.abs_sec;
         txsig->event_report_verbose.event_abstick = rw.abs_tick;
         txsig->event_report_verbose.event_absutick = rw.abs_usec;
         /* Enable interrupts. */
         BKL_UNLOCK();
         /* Set up the event_report_verbose signal. Replace process names with id's. */
         txsig->event_report_verbose.sig_no = CATCH_REPORT;
         txsig->event_report_verbose.event_tick = get_systime((OSTICK *) & txsig->event_report_verbose.event_utick);
         txsig->event_report_verbose.sig_from = (U32) from->pid;
         txsig->event_report_verbose.sig_to = (U32) to_pid;
         txsig->event_report_verbose.ecode = 0;
         txsig->event_report_verbose.signo = *(SIGSELECT *) sig;
         txsig->event_report_verbose.cpu = cpu;
         txsig->event_report_verbose.sig_data = 0;
         txsig->event_report_verbose.sig_datasize = (U16) 0;
         txsig->event_report_verbose.event_type = event_no;
         txsig->event_report_verbose.action = 1;        /* Catch action. */
         txsig->event_report_verbose.sig_address = (U32) sig;
         txsig->event_report_verbose.sig_size = sigsize((union SIGNAL **) &sig);

         txsig->event_report_verbose.pid_from = txsig->event_report_verbose.sig_from;
         txsig->event_report_verbose.pid_to = txsig->event_report_verbose.sig_to;
         txsig->event_report_verbose.ptype_from = make_ptype(from);
         txsig->event_report_verbose.ptype_to = make_ptype(to);
         txsig->event_report_verbose.line = (U32) 0;
         txsig->event_report_verbose.file = (U32) 0;
         txsig->event_report_verbose.action_id = interceptionpoint;

         /* Send an event_report to the ldm. */
         set_suid(1);
         restore(txsig);
         zzsys_send_w_s(&txsig, current_process(), rw.zzose_rmm_);
         set_suid(0);

         /* Intercept the breakpoint scope. */
         intercept_evact_scope();

         /* Unlock the dispatcher. */
         UNLOCK_SWAP();

         /* Lock the dispatcher again. */
         LOCK_SWAP();

         /* Clear the caught flag. */
         rw.caught_sig = NULL;
      }
      else
      {
         BKL_UNLOCK();
      }
   }
#ifdef HAS_UNDO_ACTION
   /* Undo, no signal will be sent */
   if ((actions & 0xc0000) == 0x40000)
   {
      free_buf((union SIGNAL **)&sig);
      return NULL;
   }
#endif
   return sig;
}

void           *
zzreceive_dbghook(OSBUFSIZE sig_size, void *sig, struct debug_pcb *from, struct debug_pcb *to)
{
   return handle_send_receive_event(sig_size, sig, from, to, rw.actions SMP_CODE([ose_cpu_id()]).action_list_receive, TRACE_RECEIVE);
}

void           *
zzsend_dbghook(OSBUFSIZE sig_size, void *sig, struct debug_pcb *from, struct debug_pcb *to)
{
   return handle_send_receive_event(sig_size, sig, from, to, rw.actions SMP_CODE([ose_cpu_id()]).action_list_send, TRACE_SEND);
}


/*
**===========================================================================
** Hooks called from POSE. FIXME What to do with these?
**===========================================================================
*/
#ifdef HAS_TIMEOUT_EVENT

void    *
zzreceive_tmo_dbghook(struct debug_pcb *to, OSTIME timeout)
{
/*
** The actions flag field is defined as follows:
** D0 = 1 => Catch the event.                   Mask = 0x0003
** D1 = 1 => Ignore the catch bit.
** D2 = 1 => Monitor the event.                 Mask = 0x000c
** D3 = 1 => Ignore the monitor bit.
** D4 = 1 => Trace the event.                   Mask = 0x0030
** D5 = 1 => Ignore the trace bit.
** D6 = 1 => Turn on the trace.                 Mask = 0x00c0
** D7 = 1 => Ignore the trace on bit.
** D8 = 1 => Turn off the trace.                Mask = 0x0300
** D9 = 1 => Ignore the trace off bit.
** D10 = 1 => Turn on the watch timer.          Mask = 0x0c00
** D11 = 1 => Ignore the watch on bit.
** D12 = 1 => Turn off the watch timer.         Mask = 0x3000
** D13 = 1 => Ignore the watch off bit.
** D14 = 1 => Goto a new event sequence number. Mask = 0xc000
** D15 = 1 => Ignore the goto bit.
*/
   uint32_t actions = 0;
   struct ldefine_action *action;

   U32             process_pid;      /* Saved in case addressee gets killed. */

   U32 cpu;

   U8              oldevent_seq_no = rw.action_state;
   U8              newevent_seq_no = 0;

   U8              monitor_long_flag = 2;

   uint32_t        interceptionpoint = 0;
   uint32_t        notificationpoint = 0;

   /* Skip any event that involves two system process. */
   if (to->debug_flags & rw.sys_process)
   {
      return NULL;
   }

   /* Skip events that don't involve any process in the display scope. */
   if (!(to->debug_flags & IN_DEBUGGER_SCOPE))
      return NULL;

   /* Scan the actions list for actions to perform. */
   /* The dispatcher must be locked while doing this. */
   for (action = rw.actions SMP_CODE([ose_cpu_id()]).action_list_timeout; action; action = action->next)
   {
      assert(validate_evact(action));
      /* Check for event sequence match. */
      if (action->event_seq_no && (action->event_seq_no != oldevent_seq_no))
         continue;

      /* Check for addressee match. */
      if (!zzrmm_match_proc(to, (struct procnamelist *) action->align))
         continue;

      /* Check for an active event counter. */
      if (unlikely(action->orig_event_count))
      {
         /* Update the event counter and ignore the event if the limit is not reached. */
         LOCK();
         SPIN_LOCK(&action->SMP_CODE(orig_action->)action_lock);
         if (--(action->SMP_CODE(orig_action->)event_count))
         {
            SPIN_UNLOCK(&action->SMP_CODE(orig_action->)action_lock);
            UNLOCK();
            continue;
         }
         else
         {
            action->SMP_CODE(orig_action->)event_count = action->orig_event_count;
            SPIN_UNLOCK(&action->SMP_CODE(orig_action->)action_lock);
            UNLOCK();
         }
      }

      /* Decide which action to take and update action flags. */
      {
         U32             newaction = 0;

         switch (action->action & 0xf)
         {
            case 3:
               /* Trace. */
               newaction = 0x0010;
               break;
            case 1:
               /* Catch. */
               interceptionpoint = action->action_id;
               newaction = 0x0001;
               break;
            case 10:
               /* Monitor -long */
               notificationpoint = action->action_id;
               monitor_long_flag = 10;
               newaction = 0x0004;
               break;
            case 2:
               /* Monitor. */
               notificationpoint = action->action_id;
               newaction = 0x0004;
               break;
            case 4:
               /* Trace on. */
               newaction = 0x0040;
               break;
            case 5:
               /* Trace off. */
               newaction = 0x0100;
               break;
            case 7:
               /* Watch on. */
               newaction = 0x0400;
               break;
            case 8:
               /* Watch off. */
               newaction = 0x1000;
               break;
            case 9:
               /* Goto. */
               newaction = 0x4000;
               /* Save the seq number to be for later. */
               newevent_seq_no = action->goto_list;
               break;
            case 12:
               /* User function */
               newaction = 0x10000;
               break;
            case 14:
               /* Enable hw trace */
               newaction = 0x100000;
               break;
            case 15:
               /* Disable hw trace */
               newaction = 0x400000;
               break;
            default:
               break;
         }
         /* Negate the action if required. */
         if (action->negated)
            newaction <<= 1;
         /* Log the action for later. */
         actions |= newaction;
      }
   }

   /* Done if there are no actions to be performed. */
   if (!actions)
      return NULL;

   /* Save the process id of the addressee in case it gets killed. */
   process_pid = to->pid;

   cpu = SMP_CODE_YES_NO(zzkrn_get_pid_info(to->pid, 13), 0);

   /* Call user specified function, if it is defined in ose_reg_user_func() */
   if ((rw.user_action_function != NULL) && ((actions & 0x30000) == 0x10000))
   {
      MonitorTimeoutInfo info;
      info.sigNo = MONITOR_GET_TIMEOUT_TRACE_REPLY;
      info.reference = notificationpoint;
      BKL_LOCK();
      info.time.sec = rw.abs_sec;
      info.time.sectick = rw.abs_tick;
      info.time.tick = get_systime((OSTICK *) &info.time.ntick);
      BKL_UNLOCK();
      info.timeout = (U32)timeout;
      info.tmoSource = 0;    /* A receive timed out. */
      info.signalAddressee = (U32) process_pid;
      info.lineNumber = 0;
      info.fileNameOffset = 0;
      info.euId = cpu;
      info.reserved0 = 0;
      send_proc_info(to);

      rw.user_action_function((union SIGNAL *)&info, 0, NULL, 0);
   }

   /* Turn on the trace if required. */
   if ((actions & 0x00c0) == 0x0040)
   {
      rw.precious->trace_enabled = rw.trace_enabled = (OSBOOLEAN) 1;
      ose_trace_enable(rw.trace);
   }

   /* Goto a new event sequence number if required. */
   if ((actions & 0xc000) == 0x4000)
   {
      rw.precious->action_state = rw.action_state = newevent_seq_no;
   }

   /* Turn on the hardware trace if required. */
   if ((actions & 0x300000) == 0x100000)
   {
      (*rw.hw_trace[0])();
   }

   /* Trace the event if required. */
   if ((actions & 0x0030) == 0x0010)
   {
      union TRACE    *te;
      struct trace_procname *to_name;

      /* Allocate a trace buffer. */
      te = trace_alloc_from(sizeof (struct trace_receive_tmo) + sizeof(SIGSELECT),
			    to, &to_name);

      /* Fill in the trace buffer. */
      if (te)
      {
         /* Fill the trace buffer. */
         te->trace_receive_tmo.event_no = TRACE_RECEIVE_TMO;
	 te->trace_receive_tmo.timeout = (U32)timeout;
         te->trace_receive_tmo.abstime = get_trace_abstime();;
         te->trace_receive_tmo.process = to_name;
         te->trace_receive_tmo.cpu = cpu;

         /* Release the trace buffer. */
         ose_trace_commit_inline(te);
      }
   }

   /* Turn off the trace if required. */
   if ((actions & 0x0300) == 0x0100)
   {
      rw.precious->trace_enabled = rw.trace_enabled = (OSBOOLEAN) 0;
      ose_trace_disable(rw.trace);
   }

   /* Turn off the hardware trace if required. */
   if ((actions & 0xc00000) == 0x400000)
   {
      (*rw.hw_trace[1])();
   }


   /* Monitor the event if required. */
   if ((actions & 0x000c) == 0x0004)
   {
      union SIGNAL   *txsig;

      /* Monitor condition was true. Send to ldm if there is space in the mon queue. */
      /* Check that monitoring is enabled and that there is space in the monitor queue. */
      if ((!rw.monitor_enabled) || (rw.monitor_overflow))
      {
         SPIN_LOCK(&rw.event_notify_lock);
         if (rw.monitor_overflow)
         {
            rw.monitor_overflow++;
         }
         SPIN_UNLOCK(&rw.event_notify_lock);
         goto cant_monitor;
      }

      /* Unlock the dispatcher. */
      UNLOCK_SWAP();
      /* Allocate a monitor buffer. */
      set_suid(1);

      txsig = alloc(sizeof (struct event_report_verbose), MONITOR_REPORT);
      BKL_LOCK();
      txsig->event_report_verbose.event_abssec = rw.abs_sec;
      txsig->event_report_verbose.event_abstick = rw.abs_tick;
      txsig->event_report_verbose.event_absutick = rw.abs_usec;
      BKL_UNLOCK();
      txsig->event_report_verbose.sig_datasize = (U16) 0;
      /* Look out with the parameter type to get_systime()! */
      txsig->event_report_verbose.event_tick = get_systime((OSTICK *) & txsig->event_report_verbose.event_utick);
      txsig->event_report_verbose.sig_from = (U32) process_pid;
      txsig->event_report_verbose.sig_to = (U32) 0;
      txsig->event_report_verbose.ecode = (U32) timeout;
      txsig->event_report_verbose.signo = 0;
      txsig->event_report_verbose.cpu = cpu;
      txsig->event_report_verbose.sig_data = 0;

      txsig->event_report_verbose.pid_from = txsig->event_report_verbose.sig_from;
      txsig->event_report_verbose.pid_to = txsig->event_report_verbose.sig_to;
      txsig->event_report_verbose.ptype_from = make_ptype(to);
      txsig->event_report_verbose.ptype_to = 0;
      txsig->event_report_verbose.line = (U32) 0;
      txsig->event_report_verbose.file = (U32) 0;
      txsig->event_report_verbose.action_id = notificationpoint;

      txsig->event_report_verbose.event_type = TRACE_RECEIVE_TMO;
      txsig->event_report_verbose.sig_address = (U32)0;

      txsig->event_report_verbose.action = monitor_long_flag;
      txsig->event_report_verbose.sig_size = 0;

      /* Lock the dispatcher again. */
      LOCK_SWAP();
      /* Disable interrupts */
      SPIN_LOCK(&rw.event_notify_lock);
      /* Check again that monitoring is enabled and no overflow present. */
      assert(rw.monitor_hook_flow >= 0);
      if ((rw.monitor_enabled) && (0 == rw.monitor_hook_flow))
      {
         rw.monitor_overflow++;
      }
      if ((!rw.monitor_enabled) || (0 == rw.monitor_hook_flow))
      {
         SPIN_UNLOCK(&rw.event_notify_lock);
         /* Free the buffer and abort if disabled. */
         UNLOCK_SWAP();
         free_buf(&txsig);
         set_suid(0);
         LOCK_SWAP();
         goto cant_monitor;
      }
      /* Take a flow control token . */
      rw.monitor_hook_flow--;
      /* Enable interrupts */
      SPIN_UNLOCK(&rw.event_notify_lock);

      /* Send an event_report_verbose to the ldm. */
      zzsys_send_w_s(&txsig, rw.zzose_rmm_, rw.zzose_rmm_);
      set_suid(0);

   }
 cant_monitor:
   /* Catch the event if required. */
   if ((actions & 0x0003) == 0x0001)
   {
      /* Catch condition was true. Catch it if the scope is not stopped. */
      /* Disable interrupts */
      BKL_LOCK();
      /* Check that the system is not already caught. */
      if (rw.catch_enable)
      {
         union SIGNAL   *txsig;

         /* Disable catcher. */
         txsig = rw.catch_enable;
         rw.catch_enable = NULL;
         /* Store sig pointer and process id for the ldm to access. */
         rw.caught_pid = current_process();
         rw.caught_sig = NULL;
         txsig->event_report_verbose.event_abssec = rw.abs_sec;
         txsig->event_report_verbose.event_abstick = rw.abs_tick;
         txsig->event_report_verbose.event_absutick = rw.abs_usec;
         /* Enable interrupts. */
         BKL_UNLOCK();
         /* Set up the event_report_verbose signal. Replace process names with id's. */
         txsig->event_report_verbose.sig_no = CATCH_REPORT;
         txsig->event_report_verbose.event_tick = get_systime((OSTICK *) & txsig->event_report_verbose.event_utick);
         txsig->event_report_verbose.sig_from = (U32) process_pid;
         txsig->event_report_verbose.sig_to = (U32) 0;
         txsig->event_report_verbose.ecode = (U32) timeout;
         txsig->event_report_verbose.signo = 0;
         txsig->event_report_verbose.cpu = cpu;
         txsig->event_report_verbose.sig_data = 0;
         txsig->event_report_verbose.sig_datasize = (U16) 0;
         txsig->event_report_verbose.event_type = TRACE_RECEIVE_TMO;
         txsig->event_report_verbose.action = 1;        /* Catch action. */
         txsig->event_report_verbose.sig_address = 0;
         txsig->event_report_verbose.sig_size = 0;

         txsig->event_report_verbose.pid_from = txsig->event_report_verbose.sig_from;
         txsig->event_report_verbose.pid_to = txsig->event_report_verbose.sig_to;
         txsig->event_report_verbose.ptype_from = make_ptype(to);
         txsig->event_report_verbose.ptype_to = 0;
         txsig->event_report_verbose.line = (U32) 0;
         txsig->event_report_verbose.file = (U32) 0;
         txsig->event_report_verbose.action_id = interceptionpoint;

         /* Send an event_report to the ldm. */
         set_suid(1);
         restore(txsig);
         zzsys_send_w_s(&txsig, current_process(), rw.zzose_rmm_);
         set_suid(0);

         /* Intercept the breakpoint scope. */
         intercept_evact_scope();

         /* Unlock the dispatcher. */
         UNLOCK_SWAP();

         /* Lock the dispatcher again. */
         LOCK_SWAP();

         /* Clear the caught flag. */
         rw.caught_sig = NULL;
      }
      else
      {
         BKL_UNLOCK();
      }
   }
   return NULL;
}
#endif

static void *
handle_int_send_receive_event(OSBUFSIZE sig_size, void *sig,
                              struct debug_pcb *from, struct debug_pcb *to,
                              struct ldefine_action *action_list, int event_no)
{
/*
** The actions flag field is defined as follows:
** D0 = 1 => Catch the event.                   Mask = 0x0003
** D1 = 1 => Ignore the catch bit.
** D2 = 1 => Monitor the event.                 Mask = 0x000c
** D3 = 1 => Ignore the monitor bit.
** D4 = 1 => Trace the event.                   Mask = 0x0030
** D5 = 1 => Ignore the trace bit.
** D6 = 1 => Turn on the trace.                 Mask = 0x00c0
** D7 = 1 => Ignore the trace on bit.
** D8 = 1 => Turn off the trace.                Mask = 0x0300
** D9 = 1 => Ignore the trace off bit.
** D10 = 1 => Turn on the watch timer.          Mask = 0x0c00
** D11 = 1 => Ignore the watch on bit.
** D12 = 1 => Turn off the watch timer.         Mask = 0x3000
** D13 = 1 => Ignore the watch off bit.
** D14 = 1 => Goto a new event sequence number. Mask = 0xc000
** D15 = 1 => Ignore the goto bit.
*/
   LOCKSTATE       msr;
   uint32_t actions = 0;
   struct ldefine_action *action;
   U32             to_pid;      /* Saved in case addressee gets killed. */

   U32 cpu;

   U8              oldevent_seq_no = rw.action_state;
   U8              newevent_seq_no = 0;

   U8              monitor_long_flag = 2;       /* Current monitor action. 2 or 10 */

   uint32_t        interceptionpoint = 0;
   uint32_t        notificationpoint = 0;
   uint32_t        trace_with_data = 0;

   /* Skip any event that involves two system process. */
   if ((from->debug_flags & to->debug_flags) & rw.sys_process)
      return sig;

   /* Skip events that don't involve any process in the display scope. */
   if (!((from->debug_flags | to->debug_flags) & IN_DEBUGGER_SCOPE))
      return sig;

   /* Scan the actions list for actions to perform. */
   for (action = action_list; action; action = action->next)
   {
      /* Check for event sequence match. */
      if (action->event_seq_no && (action->event_seq_no != oldevent_seq_no))
         continue;

      /* Check for sender match. */
      if (!zzrmm_match_proc(from, (struct procnamelist *) action->align))
         continue;

      /* Check for addressee match. */
      if (!zzrmm_match_proc(to, (struct procnamelist *) action->ecode))
         continue;

      /* Check for signal number match. */
      if ((action->signo > *(SIGSELECT *) sig) || (action->signo_end < *(SIGSELECT *) sig))
         continue;

      if (unlikely(!match_data_range((uint8_t *)sig, sig_size, action)))
      {
         continue;
      }

      /* Check for an active event counter. */
      if (unlikely(action->orig_event_count))
      {
         /* Update the event counter and ignore the event if the limit is not reached. */
         SPIN_LOCK_SAVE(&action->SMP_CODE(orig_action->)action_lock, msr);
         if (--(action->SMP_CODE(orig_action->)event_count))
         {
            SPIN_LOCK_RESTORE(&action->SMP_CODE(orig_action->)action_lock, msr);
            continue;
         }
         else
         {
            action->SMP_CODE(orig_action->)event_count = action->orig_event_count;
            SPIN_LOCK_RESTORE(&action->SMP_CODE(orig_action->)action_lock, msr);
         }
      }

      /* Decide which action to take and update action flags. */
      {
         U32             newaction = 0;

         switch (action->action & 0xf)
         {
            case 3:
               /* Trace. */
               newaction = 0x0010;
               trace_with_data |= action->action & TRACE_WITH_DATA;
               break;
            case 1:
               /* Catch. */
               interceptionpoint = action->action_id;
               newaction = 0x0001;
               break;
            case 10:
               /* Monitor -long */
               notificationpoint = action->action_id;
               monitor_long_flag = 10;
               newaction = 0x0004;
               break;
            case 2:
               /* Monitor. */
               notificationpoint = action->action_id;
               newaction = 0x0004;
               break;
            case 4:
               /* Trace on. */
               newaction = 0x0040;
               break;
            case 5:
               /* Trace off. */
               newaction = 0x0100;
               break;
            case 7:
               /* Watch on. */
               newaction = 0x0400;
               break;
            case 8:
               /* Watch off. */
               newaction = 0x1000;
               break;
            case 9:
               /* Goto. */
               newaction = 0x4000;
               /* Save the seq number to be for later. */
               newevent_seq_no = action->goto_list;
               break;
            case 12:
               /* User function */
               newaction = 0x10000;
               break;
	    case 13:
               /* Undo */
               newaction = 0x40000;
               break;
            case 14:
               /* Enable hw trace */
               newaction = 0x100000;
               break;
            case 15:
               /* Disable hw trace */
               newaction = 0x400000;
               break;
            default:
               break;
         }
         /* Negate the action if required. */
         if (action->negated)
            newaction <<= 1;
         /* Log the action for later. */
         actions |= newaction;
      }
   }

   /* Done if there are no actions to be performed. */
   if (!actions)
      return sig;

   /* Save the process id of the addressee in case it gets killed. */
   to_pid = to->pid;

   cpu = SMP_CODE_YES_NO(zzkrn_get_pid_info((event_no == TRACE_SEND) ? from->pid : to->pid, 13), 0);

   /* Call user specified function, if it is defined in ose_reg_user_func() */
   if (rw.user_action_function != NULL && ((actions & 0x30000) == 0x10000))
   {
      MonitorSendReceiveInfo info;

      info.sigNo = event_no == TRACE_SEND ? MONITOR_GET_SEND_TRACE_REPLY:
         MONITOR_GET_RECEIVE_TRACE_REPLY;

      info.reference = notificationpoint;
      BKL_LOCK_SAVE(msr);
      info.time.sec = rw.abs_sec;
      info.time.sectick = rw.abs_tick;
      info.time.tick = get_systime((OSTICK *) &info.time.ntick);
      BKL_LOCK_RESTORE(msr);
      info.signalNumber = *(SIGSELECT *) sig;
      info.signalSender = (U32) from->pid;
      info.signalAddressee = (U32) to->pid;
      info.signalSize = sig_size;
      info.signalAddress = (U32) sig;
      info.signalId = 0;
      info.lineNumber = 0;
      info.fileNameOffset = 0;
      info.signalDataOffset = 0;
      info.signalDataCount = 0;
      info.euId = cpu;
      info.reserved0 = 0;
      info.reserved1 = 0;
      info.data[0] = 0;

      send_proc_info(to);
      send_proc_info(from);

      rw.user_action_function((union SIGNAL *)&info,
                           offsetof(MonitorSendReceiveInfo, data),
                           sig,
                           sig_size);
   }

   /* Turn on the trace if required. */
   if ((actions & 0x00c0) == 0x0040)
   {
      rw.precious->trace_enabled = rw.trace_enabled = (OSBOOLEAN) 1;
      ose_trace_enable(rw.trace);
   }

   /* Turn on the hardware trace if required. */
   if ((actions & 0x300000) == 0x100000)
   {
      (*rw.hw_trace[0])();
   }


   /* Goto a new event sequence number if required. */
   if ((actions & 0xc000) == 0x4000)
   {
      rw.precious->action_state = rw.action_state = newevent_seq_no;
   }
   /* Trace the event if required. */
   if ((actions & 0x0030) == 0x0010)
   {
      union TRACE    *te;
      struct trace_procname *from_name;
      struct trace_procname *to_name;
      uint32_t tr_sig_size;

      if (trace_with_data)
      {
         /* Trim the size if too large. */
         tr_sig_size = (sig_size > rw.max_trace_size) ? rw.max_trace_size : sig_size;
      }
      else
      {
         tr_sig_size = 4; /* allways include the signal number */
      }

      /* Allocate a trace buffer. */
      /* Trim the size if too large. */

      te = trace_alloc_from_to((EntrySize) (sizeof (struct trace_send_receive) + tr_sig_size),
                               from, &from_name, to, &to_name);
      /* Fill in the trace buffer. */
      if (te)
      {
         /* Fill the trace buffer. */
         te->trace_send_receive.event_no = event_no;
         te->trace_send_receive.abstime = get_trace_abstime();;
         te->trace_send_receive.sigsize = (U16) tr_sig_size;
         te->trace_send_receive.from = from_name;
         te->trace_send_receive.to = to_name;
         te->trace_send_receive.cpu = cpu;
         memcpy(te->trace_send_receive.data, sig, tr_sig_size);

         /* Release the trace buffer. */
         ose_trace_commit_inline(te);
      }
   }

   /* Turn off the trace if required. */
   if ((actions & 0x0300) == 0x0100)
   {
      rw.precious->trace_enabled = rw.trace_enabled = (OSBOOLEAN) 0;
      ose_trace_disable(rw.trace);
   }

   /* Turn off the hardware trace if required. */
   if ((actions & 0xc00000) == 0x400000)
   {
      (*rw.hw_trace[1])();
   }

   /* Monitor the event if required. */
   if ((actions & 0x000c) == 0x0004)
   {
      union SIGNAL   *txsig;

      /* Monitor condition was true. Send to ldm if there is space in the mon queue. */
      /* Check that monitoring is enabled and that there is space in the monitor queue. */
      if ((!rw.monitor_enabled) || (rw.monitor_overflow))
      {
         SPIN_LOCK_SAVE(&rw.event_notify_lock, msr);
         if (rw.monitor_overflow)
         {
            rw.monitor_overflow++;
         }
         SPIN_LOCK_RESTORE(&rw.event_notify_lock, msr);
         goto cant_monitor;
      }

      /* Allocate a monitor buffer. */
      set_suid(1);
      if (monitor_long_flag == 10)
      {
         txsig = alloc(sizeof (struct event_report_verbose) +
                       ((sig_size > rw.max_monitor_size) ? rw.max_monitor_size : sig_size), MONITOR_REPORT);
         txsig->event_report_verbose.sig_datasize = (U16) ((sig_size > rw.max_monitor_size) ? rw.max_monitor_size : sig_size);
      }
      else
      {
         txsig = alloc(sizeof (struct event_report_verbose) + 4, MONITOR_REPORT);
         txsig->event_report_verbose.sig_datasize = (U16) 4;
      }

      /* Set up the event_report_verbose signal. Replace process names with id's. */
      LOCK_SAVE(msr);
      SMP_CODE(do)
      {
         txsig->event_report_verbose.event_abssec = rw.abs_sec;
         txsig->event_report_verbose.event_abstick = rw.abs_tick;
         txsig->event_report_verbose.event_absutick = rw.abs_usec;
      }
      SMP_CODE(while(txsig->event_report_verbose.event_abssec != rw.abs_sec || \
         txsig->event_report_verbose.event_abstick != rw.abs_tick || \
         txsig->event_report_verbose.event_absutick != rw.abs_usec);)
      LOCK_RESTORE(msr);
      /* Look out with the parameter type to get_systime()! */
      txsig->event_report_verbose.event_tick = get_systime((OSTICK *) & txsig->event_report_verbose.event_utick);
      txsig->event_report_verbose.sig_from = (U32) from->pid;
      txsig->event_report_verbose.sig_to = (U32) to_pid;
      txsig->event_report_verbose.ecode = 0;
      txsig->event_report_verbose.signo = *(SIGSELECT *) sig;
      txsig->event_report_verbose.cpu = cpu;
      txsig->event_report_verbose.sig_data = 0;
      txsig->event_report_verbose.event_type = event_no;
      txsig->event_report_verbose.sig_address = (U32) sig;

      txsig->event_report_verbose.pid_from = txsig->event_report_verbose.sig_from;
      txsig->event_report_verbose.pid_to = txsig->event_report_verbose.sig_to;
      txsig->event_report_verbose.ptype_from = make_ptype(from);
      txsig->event_report_verbose.ptype_to = make_ptype(to);
      txsig->event_report_verbose.line = (U32) 0;
      txsig->event_report_verbose.file = (U32) 0;
      txsig->event_report_verbose.action_id = notificationpoint;

      txsig->event_report_verbose.action = monitor_long_flag;
      txsig->event_report_verbose.sig_size = sig_size;
      memcpy((void *) txsig->event_report_verbose.data,
             (void *) sig, (size_t) txsig->event_report_verbose.sig_datasize);

      /* Disable interrupts */
      SPIN_LOCK_SAVE(&rw.event_notify_lock, msr);
      /* Check again that monitoring is enabled and no overflow present. */
      if ((rw.monitor_enabled) && (!rw.monitor_hook_flow))
      {
         rw.monitor_overflow++;
      }
      if ((!rw.monitor_enabled) || (!rw.monitor_hook_flow))
      {
         /* Free the buffer and abort if disabled. */
         SPIN_LOCK_RESTORE(&rw.event_notify_lock, msr);
         free_buf(&txsig);
         set_suid(0);
         goto cant_monitor;
      }
      /* Take a flow control token . */
      rw.monitor_hook_flow--;
      /* Enable interrupts */
      SPIN_LOCK_RESTORE(&rw.event_notify_lock, msr);

      /* Send an event_report_verbose to the ose_monitor process. */
      send_w_s(&txsig, rw.zzose_rmm_, rw.zzose_rmm_);
      set_suid(0);

   }
 cant_monitor:
   ;
   /* Catch the event if required. */
   if ((actions & 0x0003) == 0x0001)
   {
      /* Catch condition was true. Catch it if the scope is not stopped. */
      /* Disable interrupts */
      BKL_LOCK_SAVE(msr);
      /* Check that the system is not already caught. */
      if (rw.catch_enable)
      {
         union SIGNAL   *txsig;

         /* Disable catcher. */
         txsig = rw.catch_enable;
         rw.catch_enable = NULL;
         /* Store sig pointer and process id for the ldm to access. */
         rw.caught_pid = current_process();
         rw.caught_sig = (union SIGNAL **) &sig;
         txsig->event_report_verbose.event_abssec = rw.abs_sec;
         txsig->event_report_verbose.event_abstick = rw.abs_tick;
         txsig->event_report_verbose.event_absutick = rw.abs_usec;
         /* Enable interrupts. */
         BKL_LOCK_RESTORE(msr);
         /* Set up the event_report_verbose signal. Replace process names with id's. */
         txsig->event_report_verbose.sig_no = CATCH_REPORT;
         txsig->event_report_verbose.event_tick = get_systime((OSTICK *) & txsig->event_report_verbose.event_utick);
         txsig->event_report_verbose.sig_from = (U32) from->pid;
         txsig->event_report_verbose.sig_to = (U32) to_pid;
         txsig->event_report_verbose.ecode = 0;
         txsig->event_report_verbose.signo = *(SIGSELECT *) sig;
         txsig->event_report_verbose.cpu = cpu;
         txsig->event_report_verbose.sig_data = 0;
         txsig->event_report_verbose.sig_datasize = (U16) 0;
         txsig->event_report_verbose.event_type = event_no;
         txsig->event_report_verbose.action = 1;        /* Catch action. */
         txsig->event_report_verbose.sig_address = (U32) sig;
         txsig->event_report_verbose.sig_size = sigsize((union SIGNAL **) &sig);

         txsig->event_report_verbose.pid_from = txsig->event_report_verbose.sig_from;
         txsig->event_report_verbose.pid_to = txsig->event_report_verbose.sig_to;
         txsig->event_report_verbose.ptype_from = make_ptype(from);
         txsig->event_report_verbose.ptype_to = make_ptype(to);
         txsig->event_report_verbose.line = (U32) 0;
         txsig->event_report_verbose.file = (U32) 0;
         txsig->event_report_verbose.action_id = interceptionpoint;

         /* Send an event_report_verbose to the ldm. */
         set_suid(1);
         restore(txsig);
         send_w_s(&txsig, current_process(), rw.zzose_rmm_);
         set_suid(0);

         /* Intercept the breakpoint scope. */
         intercept_evact_scope();

         /* Clear the caught flag. */
         rw.caught_sig = NULL;
      }
      else
      {
         BKL_LOCK_RESTORE(msr);
      }
   }
#ifdef HAS_UNDO_ACTION
   /* Undo, no signal will be sent */
   if ((actions & 0xc0000) == 0x40000)
   {
      free_buf((union SIGNAL **)&sig);
      return NULL;
   }
#endif
   return sig;
}

#ifdef HAS_UNDO_ACTION
void           *
#else
void
#endif
zzireceive_dbghook(OSBUFSIZE sig_size, void *sig, struct debug_pcb *from, struct debug_pcb *to)
{
#ifdef HAS_UNDO_ACTION
   return
#endif
   handle_int_send_receive_event(sig_size, sig, from, to, rw.actions SMP_CODE([ose_cpu_id()]).action_list_receive, TRACE_RECEIVE);
}

#ifdef HAS_UNDO_ACTION
void           *
#else
void
#endif
zzisend_dbghook(OSBUFSIZE sig_size, void *sig, struct debug_pcb *from, struct debug_pcb *to)
{
#ifdef HAS_UNDO_ACTION
   return
#endif
   handle_int_send_receive_event(sig_size, sig, from, to, rw.actions SMP_CODE([ose_cpu_id()]).action_list_send, TRACE_SEND);
}

static void
handle_alloc_free_buf_event(OSBUFSIZE sig_size, void *sig, struct debug_pcb *from,
                            struct ldefine_action *action_list, int event_no)
{
/*
** The actions flag field is defined as follows:
** D0 = 1 => Catch the event.                   Mask = 0x0003
** D1 = 1 => Ignore the catch bit.
** D2 = 1 => Monitor the event.                 Mask = 0x000c
** D3 = 1 => Ignore the monitor bit.
** D4 = 1 => Trace the event.                   Mask = 0x0030
** D5 = 1 => Ignore the trace bit.
** D6 = 1 => Turn on the trace.                 Mask = 0x00c0
** D7 = 1 => Ignore the trace on bit.
** D8 = 1 => Turn off the trace.                Mask = 0x0300
** D9 = 1 => Ignore the trace off bit.
** D10 = 1 => Turn on the watch timer.          Mask = 0x0c00
** D11 = 1 => Ignore the watch on bit.
** D12 = 1 => Turn off the watch timer.         Mask = 0x3000
** D13 = 1 => Ignore the watch off bit.
** D14 = 1 => Goto a new event sequence number. Mask = 0xc000
** D15 = 1 => Ignore the goto bit.
*/
   LOCKSTATE       msr;
   uint32_t actions = 0;
   struct ldefine_action *action;

   U32 cpu;

   U8              oldevent_seq_no = rw.action_state;
   U8              newevent_seq_no = 0;

   U8              monitor_long_flag = 2;       /* Current monitor action. 2 or 10 */

   uint32_t        interceptionpoint = 0;
   uint32_t        notificationpoint = 0;

   /* Skip any event that involves system process or recursive calls. */
   if (from->debug_flags & rw.sys_process || (from->debug_flags & IN_ALLOC))
      return;
   /* Skip events that don't involve any process in the display scope. */
   if (!(from->debug_flags & IN_DEBUGGER_SCOPE))
      return;

   LOCK_SAVE(msr);
   /* Scan the actions list for actions to perform. */
   for (action = action_list; action; action = action->next)
   {
      /* Check for event sequence match. */
      if (action->event_seq_no && (action->event_seq_no != oldevent_seq_no))
         continue;

      /* Check for sender match. */
      if (!zzrmm_match_proc(from, (struct procnamelist *) action->align))
         continue;

      /* Check for signal number match. */
      if ((action->signo > *(SIGSELECT *) sig) || (action->signo_end < *(SIGSELECT *) sig))
         continue;
      
      if (unlikely(TRACE_FREE_BUF == event_no &&
            !match_data_range((uint8_t *)sig, sig_size, action)))
      {
         continue;
      }

      /* Check for an active event counter. */
      if (unlikely(action->orig_event_count))
      {
         /* Update the event counter and ignore the event if the limit is not reached. */
         SPIN_LOCK(&action->SMP_CODE(orig_action->)action_lock);
         if (--(action->SMP_CODE(orig_action->)event_count))
         {
            SPIN_UNLOCK(&action->SMP_CODE(orig_action->)action_lock);
            continue;
         }
         else
         {
            action->SMP_CODE(orig_action->)event_count = action->orig_event_count;
            SPIN_UNLOCK(&action->SMP_CODE(orig_action->)action_lock);
         }
      }

      /* Decide which action to take and update action flags. */
      {
         U32             newaction = 0;

         switch (action->action)
         {
            case 3:
               /* Trace. */
               newaction = 0x0010;
               break;
            case 1:
               /* Catch. */
               interceptionpoint = action->action_id;
               newaction = 0x0001;
               break;
            case 10:
               /* Monitor -long */
               notificationpoint = action->action_id;
               monitor_long_flag = 10;
               newaction = 0x0004;
               break;
            case 2:
               /* Monitor. */
               notificationpoint = action->action_id;
               newaction = 0x0004;
               break;
            case 4:
               /* Trace on. */
               newaction = 0x0040;
               break;
            case 5:
               /* Trace off. */
               newaction = 0x0100;
               break;
            case 7:
               /* Watch on. */
               newaction = 0x0400;
               break;
            case 8:
               /* Watch off. */
               newaction = 0x1000;
               break;
            case 9:
               /* Goto. */
               newaction = 0x4000;
               /* Save the seq number to be for later. */
               newevent_seq_no = action->goto_list;
               break;
            case 12:
               /* User function */
               newaction = 0x10000;
               break;
            case 14:
               /* Enable hw trace */
               newaction = 0x100000;
               break;
            case 15:
               /* Disable hw trace */
               newaction = 0x400000;
               break;
            default:
               break;
         }
         /* Negate the action if required. */
         if (action->negated)
            newaction <<= 1;
         /* Log the action for later. */
         actions |= newaction;
      }
   }
   LOCK_RESTORE(msr);

   /* Done if there are no actions to be performed. */
   if (!actions)
      return;

   cpu = SMP_CODE_YES_NO(zzkrn_get_pid_info(from->pid, 13), 0);

   /* Call user specified function, if it is defined in ose_reg_user_func() */
   if ((rw.user_action_function != NULL) && ((actions & 0x30000) == 0x10000))
   {
      MonitorAllocFreeBufInfo info;

      info.sigNo = event_no == TRACE_ALLOC ?  MONITOR_GET_ALLOC_TRACE_REPLY:
         MONITOR_GET_FREE_BUF_TRACE_REPLY;

      info.reference = notificationpoint;
      LOCK_SAVE(msr);
      SMP_CODE(do)
      {
         info.time.sec = rw.abs_sec;
         info.time.sectick = rw.abs_tick;
      }
      SMP_CODE(while(info.time.sec != rw.abs_sec || info.time.sectick != rw.abs_tick);)
      info.time.tick = get_systime((OSTICK *) &info.time.ntick);
      LOCK_RESTORE(msr);
      info.from = (U32) from->pid;
      info.signalNumber = *(SIGSELECT *) sig;
      info.signalSize = sig_size;
      info.signalAddress = (U32) sig;
      info.signalId = 0;
      info.lineNumber = 0;
      info.fileNameOffset = 0;
      info.euId = cpu;
      info.reserved0 = 0;
      info.data[0] = 0;

      send_proc_info(from);

      rw.user_action_function((union SIGNAL *)&info,
                           offsetof(MonitorAllocFreeBufInfo, data),
                           NULL,
                           0);
   }

   /* Turn on the trace if required. */
   if ((actions & 0x00c0) == 0x0040)
   {
      rw.precious->trace_enabled = rw.trace_enabled = (OSBOOLEAN) 1;
      ose_trace_enable(rw.trace);
   }

   /* Turn on the hardware trace if required. */
   if ((actions & 0x300000) == 0x100000)
   {
      (*rw.hw_trace[0])();
   }

   /* Goto a new event sequence number if required. */
   if ((actions & 0xc000) == 0x4000)
   {
      rw.precious->action_state = rw.action_state = newevent_seq_no;
   }
   /* Trace the event if required. */
   if ((actions & 0x0030) == 0x0010)
   {
      union TRACE    *te;
      struct trace_procname *from_name;

      /* Allocate a trace buffer. */
      te = trace_alloc_from(sizeof (struct trace_alloc_free_buf) + sizeof(SIGSELECT),
                            from, &from_name);

      if (te)
      {
         /* Fill the trace buffer. */
         te->trace_alloc_free_buf.event_no = event_no;
         te->trace_alloc_free_buf.sigsize = sizeof(SIGSELECT);
         te->trace_alloc_free_buf.abstime = get_trace_abstime();;
         te->trace_alloc_free_buf.requested_size = (U16)sig_size;
         te->trace_alloc_free_buf.process = from_name;
         {
            SIGSELECT *signo = (SIGSELECT *)te->trace_alloc_free_buf.data;
            *signo = *(SIGSELECT *)sig;
         }
         te->trace_alloc_free_buf.address = (U32) sig;
         te->trace_alloc_free_buf.cpu = cpu;

         /* Release the trace buffer. */
         ose_trace_commit_inline(te);
      }
   }

   /* Turn off the trace if required. */
   if ((actions & 0x0300) == 0x0100)
   {
      rw.precious->trace_enabled = rw.trace_enabled = (OSBOOLEAN) 0;
      ose_trace_disable(rw.trace);
   }

   /* Turn off the hardware trace if required. */
   if ((actions & 0xc00000) == 0x400000)
   {
      (*rw.hw_trace[1])();  
   }

   /* Monitor the event if required. */
   if ((actions & 0x000c) == 0x0004)
   {
      union SIGNAL   *txsig;

      /* Monitor condition was true. Send to ldm if there is space in the mon queue. */
      /* Check that monitoring is enabled and that there is space in the monitor queue. */
      if (!rw.monitor_enabled)
      {
         goto cant_monitor;
      }
      SPIN_LOCK_SAVE(&rw.event_notify_lock, msr);
      if (rw.monitor_overflow)
      {
         rw.monitor_overflow++;
         SPIN_LOCK_RESTORE(&rw.event_notify_lock, msr);
         goto cant_monitor;
      }
      SPIN_LOCK_RESTORE(&rw.event_notify_lock, msr);

      /* Allocate a monitor buffer. */
      set_suid(1);
      from->debug_flags |= IN_ALLOC;

      txsig = alloc(sizeof (struct event_report_verbose), MONITOR_REPORT);

      from->debug_flags &= ~IN_ALLOC;

      LOCK_SAVE(msr);
      SMP_CODE(do)
      {
         txsig->event_report_verbose.event_abssec = rw.abs_sec;
         txsig->event_report_verbose.event_abstick = rw.abs_tick;
         txsig->event_report_verbose.event_absutick = rw.abs_usec;
      }
      SMP_CODE(while( \
         txsig->event_report_verbose.event_abssec != rw.abs_sec || \
         txsig->event_report_verbose.event_abstick != rw.abs_tick || \
         txsig->event_report_verbose.event_absutick != rw.abs_usec);)
      LOCK_RESTORE(msr);

      txsig->event_report_verbose.sig_datasize = (U16) 0;
      /* Look out with the parameter type to get_systime()! */
      txsig->event_report_verbose.event_tick = get_systime((OSTICK *) & txsig->event_report_verbose.event_utick);
      txsig->event_report_verbose.sig_from = (U32) from->pid;
      txsig->event_report_verbose.sig_to = 0; /* No to process */
      txsig->event_report_verbose.ecode = 0;
      txsig->event_report_verbose.signo = *(SIGSELECT *) sig;
      txsig->event_report_verbose.cpu = cpu;
      txsig->event_report_verbose.sig_data = 0;
      txsig->event_report_verbose.event_type = event_no;
      txsig->event_report_verbose.sig_address = (U32) sig;

      txsig->event_report_verbose.pid_from = txsig->event_report_verbose.sig_from;
      txsig->event_report_verbose.pid_to = 0; /* No to process */
      txsig->event_report_verbose.ptype_from = make_ptype(from);
      txsig->event_report_verbose.ptype_to = 0; /* No to process */
      txsig->event_report_verbose.line = (U32) 0;
      txsig->event_report_verbose.file = (U32) 0;
      txsig->event_report_verbose.action_id = notificationpoint;

      txsig->event_report_verbose.action = monitor_long_flag;
      txsig->event_report_verbose.sig_size = sig_size;
      /*   memcpy((void *) txsig->event_report_verbose.data,
           (void *) sig, (size_t) txsig->event_report_verbose.sig_datasize);*/

      /* Disable interrupts */
      SPIN_LOCK_SAVE(&rw.event_notify_lock, msr);
      /* Check again that monitoring is enabled and no overflow present. */
      if ((rw.monitor_enabled) && (!rw.monitor_hook_flow))
      {
         rw.monitor_overflow++;
      }

      if ((!rw.monitor_enabled) || (!rw.monitor_hook_flow))
      {
         /* Free the buffer and abort if disabled. */
         from->debug_flags |= IN_ALLOC;
         SPIN_LOCK_RESTORE(&rw.event_notify_lock, msr);

         free_buf(&txsig);

         from->debug_flags &= ~IN_ALLOC;
         set_suid(0);
         goto cant_monitor;
      }
      /* Take a flow control token . */
      rw.monitor_hook_flow--;
      /* Enable interrupts */
      SPIN_LOCK_RESTORE(&rw.event_notify_lock, msr);

      /* Send an event_report_verbose to the ose_monitor process. */
      send_w_s(&txsig, rw.zzose_rmm_, rw.zzose_rmm_);
      set_suid(0);

   }
 cant_monitor:
   ;
   /* Catch the event if required. */
   if ((actions & 0x0003) == 0x0001)
   {
      /* Catch condition was true. Catch it if the scope is not stopped. */
      /* Disable interrupts */
      BKL_LOCK_SAVE(msr);
      /* Check that the system is not already caught. */
      if (rw.catch_enable)
      {
         union SIGNAL   *txsig;

         /* Disable catcher. */
         txsig = rw.catch_enable;
         rw.catch_enable = NULL;
         /* Store sig pointer and process id for the ldm to access. */
         rw.caught_pid = current_process();
         txsig->event_report_verbose.event_abssec = rw.abs_sec;
         txsig->event_report_verbose.event_abstick = rw.abs_tick;
         txsig->event_report_verbose.event_absutick = rw.abs_usec;
         /* Enable interrupts. */
         BKL_LOCK_RESTORE(msr);
         /* Set up the event_report_verbose signal. Replace process names with id's. */
         txsig->event_report_verbose.sig_no = CATCH_REPORT;
         txsig->event_report_verbose.event_tick = get_systime((OSTICK *) & txsig->event_report_verbose.event_utick);
         txsig->event_report_verbose.sig_from = (U32) from->pid;
         txsig->event_report_verbose.sig_to = 0; /* No to process */
         txsig->event_report_verbose.ecode = 0;
         txsig->event_report_verbose.signo = *(SIGSELECT *) sig;
         txsig->event_report_verbose.cpu = cpu;
         txsig->event_report_verbose.sig_data = 0;
         txsig->event_report_verbose.sig_datasize = (U16) 0;
         txsig->event_report_verbose.event_type = event_no;
         txsig->event_report_verbose.action = 1;        /* Catch action. */
         txsig->event_report_verbose.sig_address = (U32) sig;
         txsig->event_report_verbose.sig_size = sigsize((union SIGNAL **) &sig);

         txsig->event_report_verbose.pid_from = txsig->event_report_verbose.sig_from;
         txsig->event_report_verbose.pid_to = 0;
         txsig->event_report_verbose.ptype_from = make_ptype(from);
         txsig->event_report_verbose.ptype_to = 0;
         txsig->event_report_verbose.line = (U32) 0;
         txsig->event_report_verbose.file = (U32) 0;
         txsig->event_report_verbose.action_id = interceptionpoint;

         /* Send an event_report_verbose to the ldm. */
         set_suid(1);
         restore(txsig);
         send_w_s(&txsig, current_process(), rw.zzose_rmm_);
         set_suid(0);

         /* Intercept the breakpoint scope. */
         intercept_evact_scope();

         /* Clear the caught flag. */
         rw.caught_sig = NULL;
      }
      else
      {
         BKL_LOCK_RESTORE(msr);
      }
   }
}

static void
zzalloc_dbghook(OSBUFSIZE sig_size, void *sig, struct debug_pcb *from,
                 PROCESS poolid)
{
   (void)poolid;
      handle_alloc_free_buf_event(sig_size, sig, from, rw.actions SMP_CODE([ose_cpu_id()]).action_list_alloc, TRACE_ALLOC);
}

static void
zzfree_buf_dbghook(OSBUFSIZE sig_size, void *sig, struct debug_pcb *from,
                    PROCESS poolid)
{
   (void)poolid;
      handle_alloc_free_buf_event(sig_size, sig, from, rw.actions SMP_CODE([0]).action_list_free_buf, TRACE_FREE_BUF);
}

void
zzswapin_dbghook(struct debug_pcb *to_proc, struct debug_pcb *from_proc)
{
/*
** The actions flag field is defined as follows:
** D0 = 1 => Catch the event.                   Mask = 0x0003
** D1 = 1 => Ignore the catch bit.
** D2 = 1 => Monitor the event.                 Mask = 0x000c
** D3 = 1 => Ignore the monitor bit.
** D4 = 1 => Trace the event.                   Mask = 0x0030
** D5 = 1 => Ignore the trace bit.
** D6 = 1 => Turn on the trace.                 Mask = 0x00c0
** D7 = 1 => Ignore the trace on bit.
** D8 = 1 => Turn off the trace.                Mask = 0x0300
** D9 = 1 => Ignore the trace off bit.
** D10 = 1 => Turn on the watch timer.          Mask = 0x0c00
** D11 = 1 => Ignore the watch on bit.
** D12 = 1 => Turn off the watch timer.         Mask = 0x3000
** D13 = 1 => Ignore the watch off bit.
** D14 = 1 => Goto a new event sequence number. Mask = 0xc000
** D15 = 1 => Ignore the goto bit.
*/
   uint32_t actions = 0;
   struct ldefine_action *action;
   LOCKSTATE       msr;

   U32 cpu;

   U8              oldevent_seq_no = rw.action_state;
   U8              newevent_seq_no = 0;

   U32             from_sysproc, to_sysproc;

   /*
    ** We won't consider cases when both parties are system processes.
    ** Fixed "set var trace_sysprocs yes" bug.
    */
   from_sysproc = (U32) (from_proc->debug_flags & rw.sys_process);
   to_sysproc = (U32) (to_proc->debug_flags & rw.sys_process);
   if (from_sysproc && to_sysproc)
      return;

   /*
    ** Also, we will skip cases when neither process is in the display scope.
    ** Fixed "set var trace_sysprocs yes" bug.
    */
   from_sysproc = (U32) (from_proc->debug_flags & IN_DEBUGGER_SCOPE);
   to_sysproc = (U32) (to_proc->debug_flags & IN_DEBUGGER_SCOPE);
   if (!(from_sysproc || to_sysproc))
      return;

   LOCK_SAVE(msr);
   /* Scan the actions list for actions to perform. */
   /* The dispatcher must be locked while doing this. */
   for (action = rw.actions SMP_CODE([0]).action_list_swap; action; action = action->next)
   {
      /* Check for event sequence match. */
      if (action->event_seq_no && (action->event_seq_no != oldevent_seq_no))
         continue;

      /*
       ** Action entries have two forms and we have to treat them differently.
       ** Case 1 occurs when action->sig_to == 0. We should then trig if either
       ** from_proc or to_proc matches the action->sig_from field.
       ** In case 2 we should match from_proc with action->sig_from and
       ** to_proc with action->sig_to and trig only if both conditions match.
       */
      if (!action->sig_to)
      {
         if (zzrmm_match_proc(from_proc, (struct procnamelist *) action->align))
            goto match_ok;
         if (!zzrmm_match_proc(to_proc, (struct procnamelist *) action->align))
            continue;
      }
      /*
       ** Case 2:
       */
      else
      {
         if (!zzrmm_match_proc(from_proc, (struct procnamelist *) action->align))
            continue;
         if (!zzrmm_match_proc(to_proc, (struct procnamelist *) action->ecode))
            continue;
      }

      /* Check for an active event counter. */
    match_ok:
    /* Check for an active event counter. */
    if (unlikely(action->orig_event_count))
    {
       /* Update the event counter and ignore the event if the limit is not reached. */
       SPIN_LOCK(&action->SMP_CODE(orig_action->)action_lock);
       if (--(action->SMP_CODE(orig_action->)event_count))
       {
          SPIN_UNLOCK(&action->SMP_CODE(orig_action->)action_lock);
          continue;
       }
       else
       {
          action->SMP_CODE(orig_action->)event_count = action->orig_event_count;
          SPIN_UNLOCK(&action->SMP_CODE(orig_action->)action_lock);
       }
    }

      /* Decide which action to take and update action flags. */
      {
         U32             newaction = 0;

         switch (action->action)
         {
            case 3:
               /* Trace. */
               newaction = 0x0010;
               break;
            case 4:
               /* Trace on. */
               newaction = 0x0040;
               break;
            case 5:
               /* Trace off. */
               newaction = 0x0100;
               break;
            case 7:
               /* Watch on. */
               newaction = 0x0400;
               break;
            case 8:
               /* Watch off. */
               newaction = 0x1000;
               break;
            case 9:
               /* Goto. */
               newaction = 0x4000;
               /* Save the seq number to be for later. */
               newevent_seq_no = action->goto_list;
               break;
            case 1:
               /* Catch. */
               newaction = 0x0001;
               break;
            case 10:
               /* Monitor -long */
               newaction = 0x0004;
               break;
            case 2:
               /* Monitor. */
               newaction = 0x0004;
               break;
            case 12:
               /* User function call */
               newaction = 0x10000;
               break;
            case 14:
               /* Enable hw trace */
               newaction = 0x100000;
               break;
            case 15:
               /* Disable hw trace */
               newaction = 0x400000;
               break;
            default:
               break;
         }
         /* Negate the action if required. */
         if (action->negated)
            newaction <<= 1;
         /* Log the action for later. */
         actions |= newaction;
      }
   }
   LOCK_RESTORE(msr);

   /* Done if there are no actions to be performed. */
   if (!actions)
      return;

   cpu = SMP_CODE_YES_NO(zzkrn_get_pid_info(to_proc->pid, 13), 0);

   /* Call user specified function, if it is defined in ose_reg_user_func() */
   if (rw.user_action_function != NULL && ((actions & 0x30000) == 0x10000))
   {
      MonitorSwapCreateKillInfo info;

      info.sigNo = MONITOR_GET_SWAP_TRACE_REPLY;

      info.reference = 0;
      LOCK_SAVE(msr);
      info.time.sec = rw.abs_sec;
      info.time.sectick = rw.abs_tick;
      info.time.tick = get_systime((OSTICK *) &info.time.ntick);
      LOCK_RESTORE(msr);
      info.from = (U32) from_proc->pid;
      info.to = (U32) to_proc->pid;
      info.lineNumber = 0;
      info.fileNameOffset = 0;
      info.euId = cpu;
      info.reserved0 = 0;
      info.reserved1 = 0;
      info.data[0] = 0;

      send_proc_info(to_proc);

      send_proc_info(from_proc);

      rw.user_action_function((union SIGNAL *)&info,
                           offsetof(MonitorSwapCreateKillInfo, data),
                           NULL,
                           0);
   }


   /* Turn on the trace if required. */
   if ((actions & 0x00c0) == 0x0040)
   {
      rw.precious->trace_enabled = rw.trace_enabled = (OSBOOLEAN) 1;
      ose_trace_enable(rw.trace);
   }

   /* Turn on the hardware trace if required. */
   if ((actions & 0x300000) == 0x100000)
   {  
      (*rw.hw_trace[0])();
   }

   /* Turn on the watch if required. ?? */
   if ((actions & 0x0c00) == 0x0400)
   {
   }

   /* Turn off the watch if required. ?? */
   if ((actions & 0x3000) == 0x1000)
   {
   }

   /* Goto a new event sequence number if required. */
   if ((actions & 0xc000) == 0x4000)
   {
      rw.precious->action_state = rw.action_state = newevent_seq_no;
   }

   /* Trace the event if required. */
   if ((actions & 0x0030) == 0x0010)
   {
      union TRACE    *te;
      struct trace_procname *from_name;
      struct trace_procname *to_name;

      /* Allocate a trace buffer. */
      te = trace_alloc_from_to(sizeof (struct trace_swap),
                               from_proc, &from_name, to_proc, &to_name);

      /* Fill in the trace buffer. */
      if (te)
      {
         /* Fill the trace buffer. */
         te->trace_swap.event_no = TRACE_SWAP;
         te->trace_swap.abstime = rw.trace_abstime ? rw.trace_abstime : make_trace_abstime();
         te->trace_swap.to_process = to_name;
         te->trace_swap.from_process = from_name;
         te->trace_swap.cpu = cpu;

         /* Release the trace buffer. */
         ose_trace_commit_inline(te);
      }
   }

   /* Turn off the trace if required. */
   if ((actions & 0x0300) == 0x0100)
   {
      rw.precious->trace_enabled = rw.trace_enabled = (OSBOOLEAN) 0;
      ose_trace_disable(rw.trace);
   }

   /* Turn off the hardware trace if required. */
   if ((actions & 0xc00000) == 0x400000)
   {
      (*rw.hw_trace[1])();  
   }

#if 0
   /* Monitor the event if required. */
   /* Monitoring can not be supported with this dispatcher. (or ?? ) */
   if ((actions & 0x000c) == 0x0004)
   {
   }

   /* Catch the event if required. */
   /* Catching can not be supported with this dispatcher. (or ?? ) */
   if ((actions & 0x0003) == 0x0001)
   {
   }
#endif
}

/*
** function for the special case when tracing all context switches
** in a system.
**/
void
zzfast_swap_trace(struct debug_pcb *from, struct debug_pcb *to)
{
    union TRACE *te;
    te = ose_trace_alloc(rw.trace, sizeof (struct trace_swap), 0);
    if (likely(te))
    {
      if (unlikely(!to->tracename))
      {
    	 if (unlikely(!make_trace_name(to)))
         {
    	    te->event_no = TRACE_NOT_USED;
    	    ose_trace_commit_inline(te);
            return;
         }
      }
      if (unlikely(!from->tracename))
      {
         if (unlikely(!make_trace_name(from)))
         {
            te->event_no = TRACE_NOT_USED;
            ose_trace_commit_inline(te);
            return;
         }
      }
      /* Fill the trace buffer. */
      te->trace_swap.event_no = TRACE_SWAP;
      te->trace_swap.abstime = get_trace_abstime();;
      te->trace_swap.to_process = (struct trace_procname *)to->tracename;
      te->trace_swap.from_process = (struct trace_procname *)from->tracename;
      te->trace_swap.cpu = SMP_CODE_YES_NO(zzkrn_cpu_id(), 0);

      /* Release the trace buffer. */
      ose_trace_commit_inline(te);
   }
}

OSBOOLEAN
zzcreate_dbghook(struct debug_pcb *proc, OSUSER uid)
{
/*
** The actions flag field is defined as follows:
** D0 = 1 => Catch the event.                   Mask = 0x0003
** D1 = 1 => Ignore the catch bit.
** D2 = 1 => Monitor the event.                 Mask = 0x000c
** D3 = 1 => Ignore the monitor bit.
** D4 = 1 => Trace the event.                   Mask = 0x0030
** D5 = 1 => Ignore the trace bit.
** D6 = 1 => Turn on the trace.                 Mask = 0x00c0
** D7 = 1 => Ignore the trace on bit.
** D8 = 1 => Turn off the trace.                Mask = 0x0300
** D9 = 1 => Ignore the trace off bit.
** D10 = 1 => Turn on the watch timer.          Mask = 0x0c00
** D11 = 1 => Ignore the watch on bit.
** D12 = 1 => Turn off the watch timer.         Mask = 0x3000
** D13 = 1 => Ignore the watch off bit.
** D14 = 1 => Goto a new event sequence number. Mask = 0xc000
** D15 = 1 => Ignore the goto bit.
*/
   uint32_t actions = 0;
   struct ldefine_action *action;

   U32 cpu;

   U8              oldevent_seq_no = rw.action_state;
   U8              newevent_seq_no = 0;

   uint32_t        interceptionpoint = 0;
   uint32_t        notificationpoint = 0;

   /* Remove ``not used arg'' compiler warning. */
   (void) uid;

   /* Update count of all processes. */
   BKL_LOCK();
   rw.totproc_count++;
   BKL_UNLOCK();

   /* Decide if the new process is a system process. */
   if (((struct _block *)proc->block == rw.ose_block_bcb) ||
          (rw.sysproc_count && (match_wildcard(proc->name, rw.sysproc_count, rw.sysprocnames)
                         || match_pid(proc->pid, rw.sysproc_count, rw.sysprocnames)
                         || zzrmm_match_proc(proc, rw.expanded_sysprocnames))))
   {
      /* Set the sysproc flag in the pcb if so. */
      /* Also select the ose and user error hooks. */
      proc->debug_flags = SYS_PROCESS + 0x30;
   }

   /* Lock the dispatcher. */
   LOCK_SWAP();

   /* Check if the new process is visible to the debugger. ?? */

   zzprof_add_to_profiled_processes(proc);

   /* Decide if the new process matches the display scope. */
   if (zzrmm_match_proc(proc, rw.quick_debugger_scope) && (!rw.evact_lock_owner     /* Inefficient and dangerous kludge to handle segment scopes */
                                                  || rw.evact_lock_owner->debug_scope.type != MONITOR_SCOPE_SEGMENT_ID
                                                  || rw.evact_lock_owner->debug_scope.id == get_segid(proc->pid)))
   {
      /* Update debugger hook enable flags and set the display bit in the pcb. */
      proc->debug_flags |= (current_hook_flags() | IN_DEBUGGER_SCOPE);

   }

   /* Decide if the new process matches the breakpoint scope. */
   if (!(proc->debug_flags & SYS_PROCESS))
   {
      if (zzrmm_match_proc(proc, rw.quick_breakpoint_scope) && (!rw.evact_lock_owner        /* Inefficient and dangerous kludge to handle segment scopes */
                                                       || rw.evact_lock_owner->debug_scope.type != MONITOR_SCOPE_SEGMENT_ID
                                                       || rw.evact_lock_owner->debug_scope.id == get_segid(proc->pid)))
      {
         struct break_pid_entry *bke;

         /* If so, add it to the breakpoint scope pid list. */
         /* Look for a free pid slot. */
         for (bke = rw.break_pid_list; bke; bke = bke->next)
         {
            if (bke->use_count < bke->size)
               break;
         }
         /* Take it if found. */
         if (bke)
         {
            PROCESS        *pid;

            for (pid = bke->pid - 1; *(++pid);)
            {;                  /* Empty loop body. */
            }
            BKL_LOCK();
            bke->use_count++;
            *pid = proc->pid;
            BKL_UNLOCK();
         }
         /* Otherwise allocate a new entry. */
         else
         {
            PROCESS        *pid;
            U32             i;
            bke = (struct break_pid_entry *) alloc(sizeof (struct break_pid_entry), (SIGSELECT) 0);
            for (i = 128, pid = bke->pid; i--;)
               *pid++ = (PROCESS) 0;
            bke->size = 128;
            bke->use_count = 1;
            bke->pid[0] = proc->pid;
            BKL_LOCK();
            bke->next = rw.break_pid_list;
            rw.break_pid_list = bke;
            BKL_UNLOCK();
         }
         /* Also set the breakpoint bit in the pcb. */
         proc->debug_flags |= IN_BREAK_SCOPE;

         /* Check if the process should be intercepted. */
         if (rw.break_scope_status)
         {
            intercept(proc->pid);
         }
#ifdef USE_EPM
         else
         {
            Offspring         *c;

            for (c = list_head(rw.offsprings); c != NULL; c = list_next(rw.offsprings, &c->link))
            {
               if (c->scope_intercepted && zzrmm_process_in_scope(proc->pid, &c->debug_scope))
               {
                  zzkrn_dbg_intercept(proc->pid);
                  break;
               }
            }
         }
#endif
      }
   }

   /* No actions should be applied if the process is not in the scope. */
   if ((proc->debug_flags & IN_DEBUGGER_SCOPE))
   {

      /* Scan the actions list for actions to perform. */
      /* The dispatcher must be locked while doing this. */
      for (action = rw.actions SMP_CODE([0]).action_list_create; action; action = action->next)
      {
         /* Check for event sequence match. */
         if (action->event_seq_no && (action->event_seq_no != oldevent_seq_no))
            continue;

         /* Check for process match. */
         if (!zzrmm_match_proc(proc, (struct procnamelist *) action->align))
            continue;

         /* Check for an active event counter. */
         if (unlikely(action->orig_event_count))
         {
            /* Update the event counter and ignore the event if the limit is not reached. */
            LOCK();
            SPIN_LOCK(&action->SMP_CODE(orig_action->)action_lock);
            if (--(action->SMP_CODE(orig_action->)event_count))
            {
               SPIN_UNLOCK(&action->SMP_CODE(orig_action->)action_lock);
               UNLOCK();
               continue;
            }
            else
            {
               action->SMP_CODE(orig_action->)event_count = action->orig_event_count;
               SPIN_UNLOCK(&action->SMP_CODE(orig_action->)action_lock);
               UNLOCK();
            }
         }

         /* Decide which action to take and update action flags. */
         {
            U32             newaction = 0;

            switch (action->action)
            {
               case 3:
                  /* Trace. */
                  newaction = 0x0010;
                  break;
               case 1:
                  /* Catch. */
                  interceptionpoint = action->action_id;
                  newaction = 0x0001;
                  break;
               case 10:
                  /* Monitor -long */
                  notificationpoint = action->action_id;
                  newaction = 0x0004;
                  break;
               case 2:
                  /* Monitor. */
                  notificationpoint = action->action_id;
                  newaction = 0x0004;
                  break;
               case 4:
                  /* Trace on. */
                  newaction = 0x0040;
                  break;
               case 5:
                  /* Trace off. */
                  newaction = 0x0100;
                  break;
               case 7:
                  /* Watch on. */
                  newaction = 0x0400;
                  break;
               case 8:
                  /* Watch off. */
                  newaction = 0x1000;
                  break;
               case 9:
                  /* Goto. */
                  newaction = 0x4000;
                  /* Save the seq number to be for later. */
                  newevent_seq_no = action->goto_list;
                  break;
               case 12:
                  /* User function call */
                  newaction = 0x10000;
                  break;
               case 14:
                  /* Enable hw trace */
                  newaction = 0x100000;
                  break;
               case 15:
                  /* Disable hw trace */
                  newaction = 0x400000;
                  break;
               default:
                  break;
            }
            /* Negate the action if required. */
            if (action->negated)
               newaction <<= 1;
            /* Log the action for later. */
            actions |= newaction;
         }
      }

      /* Done if there are no actions to be performed. */
      if (!actions)
      {
         UNLOCK_SWAP();
         return (OSBOOLEAN) 0;
      }

      cpu = SMP_CODE_YES_NO(zzkrn_get_pid_info(proc->pid, 13), 0);

      /* Call user specified function, if it is defined in ose_reg_user_func() */
      if (rw.user_action_function != NULL && ((actions & 0x30000) == 0x10000))
      {
         MonitorSwapCreateKillInfo info;

         info.sigNo = MONITOR_GET_CREATE_TRACE_REPLY;

         info.reference = notificationpoint;
         BKL_LOCK();
         info.time.sec = rw.abs_sec;
         info.time.sectick = rw.abs_tick;
         info.time.tick = get_systime((OSTICK *) &info.time.ntick);
         BKL_UNLOCK();
         info.from = 0;
         info.to = (U32) proc->pid;
         info.lineNumber = 0;
         info.fileNameOffset = 0;
         info.euId = cpu;
         info.reserved0 = 0;
         info.reserved1 = 0;
         info.data[0] = 0;

         send_proc_info(proc);

         rw.user_action_function((union SIGNAL *)&info,
                        offsetof(MonitorSwapCreateKillInfo, data),
                        NULL,
                        0);
      }


      /* Turn on the trace if required. */
      if ((actions & 0x00c0) == 0x0040)
      {
         rw.precious->trace_enabled = rw.trace_enabled = (OSBOOLEAN) 1;
         ose_trace_enable(rw.trace);
      }

      /* Turn on the hardware trace if required. */
      if ((actions & 0x300000) == 0x100000)
      {
	 (*rw.hw_trace[0])();
      }

      /* Turn on the watch if required. ?? */
      if ((actions & 0x0c00) == 0x0400)
      {
      }

      /* Turn off the watch if required. ?? */
      if ((actions & 0x3000) == 0x1000)
      {
      }

      /* Goto a new event sequence number if required. */
      if ((actions & 0xc000) == 0x4000)
      {
         rw.precious->action_state = rw.action_state = newevent_seq_no;
      }

      /* Trace the event if required. */
      if ((actions & 0x0030) == 0x0010)
      {
         union TRACE    *te;
         struct trace_procname *proc_name;

         /* Allocate a trace buffer. */
         te = trace_alloc_from(sizeof (struct trace_create_kill),
                               proc, &proc_name);

         /* Fill in the trace buffer. */
         if (te)
         {
            /* Fill the trace buffer. */
            te->trace_create_kill.event_no = TRACE_CREATE;
            te->trace_create_kill.abstime = get_trace_abstime();;
            te->trace_create_kill.process = proc_name;
            te->trace_create_kill.cpu = cpu;

            /* Release the trace buffer. */
            ose_trace_commit_inline(te);
         }
      }

      /* Turn off the trace if required. */
      if ((actions & 0x0300) == 0x0100)
      {
         rw.precious->trace_enabled = rw.trace_enabled = (OSBOOLEAN) 0;
         ose_trace_disable(rw.trace);
      }

      /* Turn off the hardware trace if required. */
      if ((actions & 0xc00000) == 0x400000)
      {
	 (*rw.hw_trace[1])();
      }

      /* Monitor the event if required. */
      if ((actions & 0x000c) == 0x0004)
      {
         union SIGNAL   *txsig;

         /* Monitor condition was true. Send to ldm if there is space in the mon queue. */
         /* Check that monitoring is enabled and that there is space in the monitor queue. */
         if ((!rw.monitor_enabled) || (rw.monitor_overflow))
         {
            SPIN_LOCK(&rw.event_notify_lock);
            if (rw.monitor_overflow)
               rw.monitor_overflow++;
            SPIN_UNLOCK(&rw.event_notify_lock);
            goto cant_monitor;
         }

         /* Unlock the dispatcher. */
         UNLOCK_SWAP();
         /* Allocate a monitor buffer. */
         set_suid(1);
         txsig = alloc(sizeof (struct event_report_verbose), MONITOR_REPORT);

         /* Set up the event_report_verbose signal. Replace process names with id's. */
         BKL_LOCK();
         txsig->event_report_verbose.event_abssec = rw.abs_sec;
         txsig->event_report_verbose.event_abstick = rw.abs_tick;
         txsig->event_report_verbose.event_absutick = rw.abs_usec;
         BKL_UNLOCK();
         /* Look out with the parameter type to get_systime()! */
         txsig->event_report_verbose.event_tick = get_systime((OSTICK *) & txsig->event_report_verbose.event_utick);
         txsig->event_report_verbose.sig_from = (U32) proc->pid;
         txsig->event_report_verbose.sig_to = (U32) 0;
         txsig->event_report_verbose.ecode = ((((U32) proc->block->instance) << 16) | proc->block->pindex);
         txsig->event_report_verbose.signo = (SIGSELECT) 0;
         txsig->event_report_verbose.cpu = cpu;
         txsig->event_report_verbose.sig_data = 0;
         txsig->event_report_verbose.sig_datasize = (U16) 0;
         txsig->event_report_verbose.event_type = 1;    /* Process created event. */
         txsig->event_report_verbose.action = 2;        /* Monitor action. */
         txsig->event_report_verbose.sig_address = (U32) 0;
         txsig->event_report_verbose.sig_size = (U32) 0;

         txsig->event_report_verbose.pid_from = txsig->event_report_verbose.sig_from;
         txsig->event_report_verbose.pid_to = txsig->event_report_verbose.sig_to;
         txsig->event_report_verbose.ptype_from = make_ptype(proc);
         txsig->event_report_verbose.ptype_to = (U32) 0;
         txsig->event_report_verbose.line = (U32) 0;
         txsig->event_report_verbose.file = (U32) 0;
         txsig->event_report_verbose.action_id = notificationpoint;

         /* Lock the dispatcher again. */
         LOCK_SWAP();
         /* Disable interrupts */
         SPIN_LOCK(&rw.event_notify_lock);
         /* Check again that monitoring is enabled and no overflow present. */
         if ((rw.monitor_enabled) && (!rw.monitor_hook_flow))
            rw.monitor_overflow++;
         if ((!rw.monitor_enabled) || (!rw.monitor_hook_flow))
         {
            /* Free the buffer and abort if disabled. */
            SPIN_UNLOCK(&rw.event_notify_lock);
            UNLOCK_SWAP();
            free_buf(&txsig);
            set_suid(0);
            LOCK_SWAP();
            goto cant_monitor;
         }
         /* Take a flow control token . */
         rw.monitor_hook_flow--;
         /* Enable interrupts */
         SPIN_UNLOCK(&rw.event_notify_lock);

         /* Send an event_report_verbose to the ldm. */
         zzsys_send_w_s(&txsig, rw.zzose_rmm_, rw.zzose_rmm_);
         set_suid(0);
      }
    cant_monitor:

      /* Catch the event if required. */
      if ((actions & 0x0003) == 0x0001)
      {
         /* Catch condition was true. Catch it if the scope is not stopped. */
         /* Disable interrupts */
         BKL_LOCK();
         /* Check that the system is not already caught. */
         if (rw.catch_enable)
         {
            union SIGNAL   *txsig;

            /* Disable catcher. */
            txsig = rw.catch_enable;
            rw.catch_enable = NULL;
            /* Store sig pointer and process id for the ldm to access. */
            rw.caught_pid = current_process();
            txsig->event_report_verbose.event_abssec = rw.abs_sec;
            txsig->event_report_verbose.event_abstick = rw.abs_tick;
            txsig->event_report_verbose.event_absutick = rw.abs_usec;
            /* Enable interrupts. */
            BKL_UNLOCK();
            /* Set up the event_report_verbose signal. Replace process names with id's. */
            txsig->event_report_verbose.sig_no = CATCH_REPORT;
            txsig->event_report_verbose.event_tick = get_systime((OSTICK *) & txsig->event_report_verbose.event_utick);
            txsig->event_report_verbose.sig_from = (U32) proc->pid;
            txsig->event_report_verbose.sig_to = (U32) 0;
            txsig->event_report_verbose.ecode = ((((U32) proc->block->instance) << 16) | proc->block->pindex);
            txsig->event_report_verbose.signo = (SIGSELECT) 0;
            txsig->event_report_verbose.cpu = cpu;
            txsig->event_report_verbose.sig_data = 0;
            txsig->event_report_verbose.sig_datasize = (U16) 0;
            txsig->event_report_verbose.event_type = 1; /* Process created event. */
            txsig->event_report_verbose.action = 1;     /* Catch action. */
            txsig->event_report_verbose.sig_address = (U32) 0;
            txsig->event_report_verbose.sig_size = 0;

            txsig->event_report_verbose.pid_from = txsig->event_report_verbose.sig_from;
            txsig->event_report_verbose.pid_to = txsig->event_report_verbose.sig_to;
            txsig->event_report_verbose.ptype_from = make_ptype(proc);
            txsig->event_report_verbose.ptype_to = (U32) 0;
            txsig->event_report_verbose.line = (U32) 0;
            txsig->event_report_verbose.file = (U32) 0;
            txsig->event_report_verbose.action_id = interceptionpoint;

            /* Send an event_report_verbose to the ldm. */
            set_suid(1);
            restore(txsig);
            zzsys_send_w_s(&txsig, current_process(), rw.zzose_rmm_);
            set_suid(0);

            /* Intercept the breakpoint scope. */
            intercept_evact_scope();
         }
         else
         {
            BKL_UNLOCK();
         }
      }
   }

   UNLOCK_SWAP();
   return (OSBOOLEAN) 0;
}

void
zzkill_dbghook(struct debug_pcb *proc)
{
/*
** The actions flag field is defined as follows:
** D0 = 1 => Catch the event.                   Mask = 0x0003
** D1 = 1 => Ignore the catch bit.
** D2 = 1 => Monitor the event.                 Mask = 0x000c
** D3 = 1 => Ignore the monitor bit.
** D4 = 1 => Trace the event.                   Mask = 0x0030
** D5 = 1 => Ignore the trace bit.
** D6 = 1 => Turn on the trace.                 Mask = 0x00c0
** D7 = 1 => Ignore the trace on bit.
** D8 = 1 => Turn off the trace.                Mask = 0x0300
** D9 = 1 => Ignore the trace off bit.
** D10 = 1 => Turn on the watch timer.          Mask = 0x0c00
** D11 = 1 => Ignore the watch on bit.
** D12 = 1 => Turn off the watch timer.         Mask = 0x3000
** D13 = 1 => Ignore the watch off bit.
** D14 = 1 => Goto a new event sequence number. Mask = 0xc000
** D15 = 1 => Ignore the goto bit.
*/
   uint32_t actions = 0;
   struct ldefine_action *action;

   U32 cpu;

   U8              oldevent_seq_no = rw.action_state;
   U8              newevent_seq_no = 0;

   uint32_t        interceptionpoint = 0;
   uint32_t        notificationpoint = 0;

   /* Update count of all processes. */
   BKL_LOCK();
   rw.totproc_count--;
   BKL_UNLOCK();

   /* Lock the dispatcher. */
   LOCK_SWAP();

   zzprof_remove_from_profiled_processes(proc);

   /*
    ** If the process is in the breakpoint scope, then remove it.
    ** The process sometimes occur twice since there is a slight race
    ** condition between zzcreate_dbghook() and scope updates from the ldm.
    */
   if (proc->debug_flags & IN_BREAK_SCOPE)
   {
      struct break_pid_entry *bke;
      PROCESS        *pid;
      PROCESS         killed_pid = proc->pid;
      U32             count;

      for (bke = rw.break_pid_list; bke; bke = bke->next)
      {
         for (count = bke->size, pid = bke->pid; count--; pid++)
         {
            if (*pid == killed_pid)
            {
               bke->use_count--;
               *pid = (PROCESS) 0;
            }
         }
      }
   }

   /* If the process owns a caught signal then remove the caught flag. */
   BKL_LOCK();
   if (rw.caught_sig && rw.caught_pid == proc->pid)
      rw.caught_sig = NULL;
   BKL_UNLOCK();

   /* No actions should be applied if the process is not in the scope. */
   if ((proc->debug_flags & IN_DEBUGGER_SCOPE))
   {
      /* Scan the actions list for actions to perform. */
      /* The dispatcher must be locked while doing this. */
      for (action = rw.actions SMP_CODE([0]).action_list_kill; action; action = action->next)
      {
         /* Check for event sequence match. */
         if (action->event_seq_no && (action->event_seq_no != oldevent_seq_no))
            continue;

         /* Check for process match. */
         if (!zzrmm_match_proc(proc, (struct procnamelist *) action->align))
            continue;

         /* Check for an active event counter. */
         if (unlikely(action->orig_event_count))
         {
            /* Update the event counter and ignore the event if the limit is not reached. */
            LOCK();
            SPIN_LOCK(&action->SMP_CODE(orig_action->)action_lock);
            if (--(action->SMP_CODE(orig_action->)event_count))
            {
               SPIN_UNLOCK(&action->SMP_CODE(orig_action->)action_lock);
               UNLOCK();
               continue;
            }
            else
            {
               action->SMP_CODE(orig_action->)event_count = action->orig_event_count;
               SPIN_UNLOCK(&action->SMP_CODE(orig_action->)action_lock);
               UNLOCK();
            }
         }

         /* Decide which action to take and update action flags. */
         {
            U32             newaction = 0;

            switch (action->action)
            {
               case 3:
                  /* Trace. */
                  newaction = 0x0010;
                  break;
               case 1:
                  /* Catch. */
                  interceptionpoint = action->action_id;
                  newaction = 0x0001;
                  break;
               case 10:
                  /* Monitor -long */
                  notificationpoint = action->action_id;
                  newaction = 0x0004;
                  break;
               case 2:
                  /* Monitor. */
                  notificationpoint = action->action_id;
                  newaction = 0x0004;
                  break;
               case 4:
                  /* Trace on. */
                  newaction = 0x0040;
                  break;
               case 5:
                  /* Trace off. */
                  newaction = 0x0100;
                  break;
               case 7:
                  /* Watch on. */
                  newaction = 0x0400;
                  break;
               case 8:
                  /* Watch off. */
                  newaction = 0x1000;
                  break;
               case 9:
                  /* Goto. */
                  newaction = 0x4000;
                  /* Save the seq number to be for later. */
                  newevent_seq_no = action->goto_list;
                  break;
               case 12:
                  /* User function call */
                  newaction = 0x10000;
                  break;
               case 14:
                  /* Enable hw trace */
                  newaction = 0x100000;
                  break;
               case 15:
                  /* Disable hw trace */
                  newaction = 0x400000;
                  break;
               default:
                  break;
            }
            /* Negate the action if required. */
            if (action->negated)
               newaction <<= 1;
            /* Log the action for later. */
            actions |= newaction;
         }
      }

      /* Done if there are no actions to be performed. */
      if (!actions)
      {
         if (proc->tracename)
         {
            ose_trace_commit_inline(proc->tracename);
            proc->tracename = NULL;
         }
         UNLOCK_SWAP();
         return;
      }

      cpu = SMP_CODE_YES_NO(zzkrn_get_pid_info(proc->pid, 13), 0);

      /* Call user specified function, if it is defined in ose_reg_user_func() */
      if (rw.user_action_function != NULL && ((actions & 0x30000) == 0x10000))
      {
         MonitorSwapCreateKillInfo info;

         info.sigNo = MONITOR_GET_KILL_TRACE_REPLY;

         info.reference = notificationpoint;
         BKL_LOCK();
         info.time.sec = rw.abs_sec;
         info.time.sectick = rw.abs_tick;
         info.time.tick = get_systime((OSTICK *) &info.time.ntick);
         BKL_UNLOCK();
         info.from = 0;
         info.to = (U32) proc->pid;
         info.lineNumber = 0;
         info.fileNameOffset = 0;
         info.euId = cpu;
         info.reserved0 = 0;
         info.reserved1 = 0;
         info.data[0] = 0;

         send_proc_info(proc);

         rw.user_action_function((union SIGNAL *)&info,
                        offsetof(MonitorSwapCreateKillInfo, data),
                        NULL,
                        0);
      }

      /* Turn on the trace if required. */
      if ((actions & 0x00c0) == 0x0040)
      {
         rw.precious->trace_enabled = rw.trace_enabled = (OSBOOLEAN) 1;
         ose_trace_enable(rw.trace);
      }

      /* Turn on the hardware trace if required. */
      if ((actions & 0x300000) == 0x100000)
      {
	 (*rw.hw_trace[0])();
      }

      /* Turn on the watch if required. ?? */
      if ((actions & 0x0c00) == 0x0400)
      {
      }

      /* Turn off the watch if required. ?? */
      if ((actions & 0x3000) == 0x1000)
      {
      }

      /* Goto a new event sequence number if required. */
      if ((actions & 0xc000) == 0x4000)
      {
         rw.precious->action_state = rw.action_state = newevent_seq_no;
      }

      /* Trace the event if required. */
      if ((actions & 0x0030) == 0x0010)
      {
         union TRACE    *te;
         struct trace_procname *proc_name;

         /* Allocate a trace buffer. */
         te = trace_alloc_from(sizeof (struct trace_create_kill),
                               proc, &proc_name);

         /* Fill in the trace buffer. */
         if (te)
         {
            /* Fill the trace buffer. */
            te->trace_create_kill.event_no = TRACE_KILL;
            te->trace_create_kill.abstime = get_trace_abstime();;
            te->trace_create_kill.process = proc_name;
            te->trace_create_kill.cpu = cpu;

            /* Release the trace buffer. */
            ose_trace_commit_inline(te);
         }
      }

      /* Turn off the trace if required. */
      if ((actions & 0x0300) == 0x0100)
      {
         rw.precious->trace_enabled = rw.trace_enabled = (OSBOOLEAN) 0;
         ose_trace_disable(rw.trace);
      }

      /* Turn off the hardware trace if required. */
      if ((actions & 0xc00000) == 0x400000)
      {
	 (*rw.hw_trace[1])();
      }

      /* Monitor the event if required. */
      if ((actions & 0x000c) == 0x0004)
      {
         union SIGNAL   *txsig;
         OSBUFSIZE       namelen;

         /* Monitor condition was true. Send to ldm if there is space in the mon queue. */
         /* Check that monitoring is enabled and that there is space in the monitor queue. */
         if ((!rw.monitor_enabled) || (rw.monitor_overflow))
         {
            SPIN_LOCK(&rw.event_notify_lock);
            if (rw.monitor_overflow)
               rw.monitor_overflow++;
            SPIN_UNLOCK(&rw.event_notify_lock);
            goto cant_monitor;
         }

         /* Unlock the dispatcher. */
         UNLOCK_SWAP();
         /* Calculate the size of the total name. I.e. strlen of:
          * "blockname:procname" @frbr@
          */
         namelen = strlen(proc->blockname) + 1 + strlen(proc->name);
         /* Allocate a monitor buffer. */
         set_suid(1);
         /* The variable size array at the end of the struct shall contain:
          * { "-", "blockname:procname" }.
          * I.e. strlen("-") + namelen + 2*sizeof((char))'\0'). There is already
          * space for one string terminating '\0' in the struct. @frbr@
          */
         txsig = alloc(sizeof (struct event_report_verbose) + 1 + namelen + 1, MONITOR_REPORT);

         /* Set up the event_report_verbose signal. Replace process names with id's. */
         BKL_LOCK();
         txsig->event_report_verbose.event_abssec = rw.abs_sec;
         txsig->event_report_verbose.event_abstick = rw.abs_tick;
         txsig->event_report_verbose.event_absutick = rw.abs_usec;
         BKL_UNLOCK();
         /* Look out with the parameter type to get_systime()! */
         txsig->event_report_verbose.event_tick = get_systime((OSTICK *) & txsig->event_report_verbose.event_utick);
         txsig->event_report_verbose.sig_from = (U32) 2;        /* Offset to the name. */
         txsig->event_report_verbose.sig_to = (U32) 0;
         txsig->event_report_verbose.ecode = 0;
         txsig->event_report_verbose.signo = (SIGSELECT) 0;
         txsig->event_report_verbose.cpu = cpu;
         txsig->event_report_verbose.sig_data = 0;
         txsig->event_report_verbose.sig_datasize = (U16) 0;
         txsig->event_report_verbose.event_type = 2;    /* Process killed event. */
         txsig->event_report_verbose.action = 2;        /* Monitor action. */
         txsig->event_report_verbose.sig_address = (U32) 0;
         txsig->event_report_verbose.sig_size = (U32) 0;
         txsig->event_report_verbose.data[0] = '-';
         txsig->event_report_verbose.data[1] = '\0';

         txsig->event_report_verbose.pid_from = (U32) proc->pid;
         txsig->event_report_verbose.pid_to = (U32) 0;
         txsig->event_report_verbose.ptype_from = make_ptype(proc);
         txsig->event_report_verbose.ptype_to = (U32) 0;
         txsig->event_report_verbose.line = (U32) 0;
         txsig->event_report_verbose.file = (U32) 0;
         txsig->event_report_verbose.action_id = notificationpoint;

         /* Create the combined block name and process name. */
         {
            char           *d, *s;

            d = (char *) &txsig->event_report_verbose.data[2];
            for (s = proc->blockname; *s;)
               *d++ = *s++;
            *d++ = ':';
            strcpy(d, proc->name);
         }

         /* Lock the dispatcher again. */
         LOCK_SWAP();
         /* Disable interrupts */
         SPIN_LOCK(&rw.event_notify_lock);
         /* Check again that monitoring is enabled and no overflow present. */
         if ((rw.monitor_enabled) && (!rw.monitor_hook_flow))
            rw.monitor_overflow++;
         if ((!rw.monitor_enabled) || (!rw.monitor_hook_flow))
         {
            /* Free the buffer and abort if disabled. */
            SPIN_UNLOCK(&rw.event_notify_lock);
            UNLOCK_SWAP();
            free_buf(&txsig);
            set_suid(0);
            LOCK_SWAP();
            goto cant_monitor;
         }
         /* Take a flow control token . */
         rw.monitor_hook_flow--;
         /* Enable interrupts */
         SPIN_UNLOCK(&rw.event_notify_lock);

         /* Send an event_report_verbose to the ldm. */
         zzsys_send_w_s(&txsig, rw.zzose_rmm_, rw.zzose_rmm_);
         set_suid(0);
      }
    cant_monitor:

      /* Catch the event if required. */
      if ((actions & 0x0003) == 0x0001)
      {
         OSBUFSIZE       namelen;
         union SIGNAL   *txsig;

         /* Catch condition was true. Catch it if the scope is not stopped. */
         /* Disable interrupts */
         BKL_LOCK();
         /* Check that the system is not already caught. */
         if (rw.catch_enable)
         {
            /* Disable catcher. */
            txsig = rw.catch_enable;
            rw.catch_enable = NULL;
            /* Store sig pointer and process id for the ldm to access. */
            rw.caught_pid = current_process();
            /* Enable interrupts. */
            BKL_UNLOCK();

            /* Intercept the breakpoint scope. */
            intercept_evact_scope();

            /* Free the obtained buffer since we need one with a data space. */
            UNLOCK_SWAP();
            restore(txsig);
            free_buf(&txsig);

            /* Calculate the size of the total name. I.e. strlen of:
             * "blockname:procname" @frbr@
             */
            namelen = 1 + strlen(proc->blockname) + strlen(proc->name);
            /* Allocate a monitor buffer. */
            set_suid(1);
            /* The variable size array at the end of the struct shall contain:
             * { "-", "blockname:procname" }.
             * I.e. strlen("-") + namelen + 2*sizeof((char))'\0').
             * There is already space for one string terminating '\0' in the
             * struct. @frbr@
             */
            txsig = alloc(sizeof (struct event_report_verbose) + 1 + namelen + 1, CATCH_REPORT);

            /* Set up the event_report_verbose signal. Replace process names with id's. */
            BKL_LOCK();
            txsig->event_report_verbose.event_abssec = rw.abs_sec;
            txsig->event_report_verbose.event_abstick = rw.abs_tick;
            txsig->event_report_verbose.event_absutick = rw.abs_usec;
            BKL_UNLOCK();
            txsig->event_report_verbose.event_tick = get_systime((OSTICK *) & txsig->event_report_verbose.event_utick);
            txsig->event_report_verbose.sig_from = (U32) 2;     /* Offset to the name. */
            txsig->event_report_verbose.sig_to = (U32) 0;
            txsig->event_report_verbose.ecode = 0;
            txsig->event_report_verbose.signo = (SIGSELECT) 0;
            txsig->event_report_verbose.cpu = cpu;
            txsig->event_report_verbose.sig_data = 0;
            txsig->event_report_verbose.sig_datasize = (U16) 0;
            txsig->event_report_verbose.event_type = 2; /* Process killed event. */
            txsig->event_report_verbose.action = 1;     /* Catch action. */
            txsig->event_report_verbose.sig_address = (U32) 0;
            txsig->event_report_verbose.sig_size = 0;

            txsig->event_report_verbose.pid_from = (U32) proc->pid;
            txsig->event_report_verbose.pid_to = (U32) 0;
            txsig->event_report_verbose.ptype_from = make_ptype(proc);
            txsig->event_report_verbose.ptype_to = (U32) 0;
            txsig->event_report_verbose.line = (U32) 0;
            txsig->event_report_verbose.file = (U32) 0;
            txsig->event_report_verbose.action_id = interceptionpoint;
            txsig->event_report_verbose.data[0] = '-';
            txsig->event_report_verbose.data[1] = '\0';

            /* Create the combined block name and process name. */
            {
               char           *d, *s;

               d = (char *) &txsig->event_report_verbose.data[2];
               for (s = proc->blockname; *s;)
                  *d++ = *s++;
               *d++ = ':';
               strcpy(d, proc->name);
            }

            /* Send an event_report_verbose to the ldm. */
            send(&txsig, rw.zzose_rmm_);

            /* Balance the UNLOCK_SWAP() at the end of this function. */
            LOCK_SWAP();
         }
         else
         {
            BKL_UNLOCK();
         }
      }
   }

   if (proc->tracename)
   {
      ose_trace_commit_inline(proc->tracename);
      proc->tracename = NULL;
   }
   UNLOCK_SWAP();
}

#ifdef OSE_SMP
void
zzbind_dbghook(struct debug_pcb *proc, cpuid_t old_cpu, cpuid_t new_cpu)
{
/*
** The actions flag field is defined as follows:
** D0 = 1 => Catch the event.                   Mask = 0x0003
** D1 = 1 => Ignore the catch bit.
** D2 = 1 => Monitor the event.                 Mask = 0x000c
** D3 = 1 => Ignore the monitor bit.
** D4 = 1 => Trace the event.                   Mask = 0x0030
** D5 = 1 => Ignore the trace bit.
** D6 = 1 => Turn on the trace.                 Mask = 0x00c0
** D7 = 1 => Ignore the trace on bit.
** D8 = 1 => Turn off the trace.                Mask = 0x0300
** D9 = 1 => Ignore the trace off bit.
** D10 = 1 => Turn on the watch timer.          Mask = 0x0c00
** D11 = 1 => Ignore the watch on bit.
** D12 = 1 => Turn off the watch timer.         Mask = 0x3000
** D13 = 1 => Ignore the watch off bit.
** D14 = 1 => Goto a new event sequence number. Mask = 0xc000
** D15 = 1 => Ignore the goto bit.
*/
   uint32_t actions = 0;
   struct ldefine_action *action;

   U8              oldevent_seq_no = rw.action_state;
   U8              newevent_seq_no = 0;

   uint32_t        interceptionpoint = 0;
   uint32_t        notificationpoint = 0;

   /* Skip any event that involves a system process. */
   if (proc->debug_flags & rw.sys_process)
      return;

   /* Skip events that don't involve a process in the display scope. */
   if (!(proc->debug_flags & IN_DEBUGGER_SCOPE))
      return;

   /* Scan the actions list for actions to perform. */
   /* The dispatcher must be locked while doing this. */
   for (action = rw.actions SMP_CODE([ose_cpu_id()]).action_list_bind; action; action = action->next)
   {
      /* Check for event sequence match. */
      if (action->event_seq_no && (action->event_seq_no != oldevent_seq_no))
         continue;

      /* Check for procedure match. */
      if (!zzrmm_match_proc(proc, (struct procnamelist *) action->ecode))
         continue;

      /* Check for an active event counter. */
      if (unlikely(action->orig_event_count))
      {
         /* Update the event counter and ignore the event if the limit is not reached. */
         LOCK();
         SPIN_LOCK(&action->SMP_CODE(orig_action->)action_lock);
         if (--(action->SMP_CODE(orig_action->)event_count))
         {
            SPIN_UNLOCK(&action->SMP_CODE(orig_action->)action_lock);
            UNLOCK();
            continue;
         }
         else
         {
            action->SMP_CODE(orig_action->)event_count = action->orig_event_count;
            SPIN_UNLOCK(&action->SMP_CODE(orig_action->)action_lock);
            UNLOCK();
         }
      }

      /* Decide which action to take and update action flags. */
      {
         U32             newaction = 0;

         switch (action->action)
         {
            case 3:
               /* Trace. */
               newaction = 0x0010;
               break;
            case 4:
               /* Trace on. */
               newaction = 0x0040;
               break;
            case 5:
               /* Trace off. */
               newaction = 0x0100;
               break;
            case 7:
               /* Watch on. */
               newaction = 0x0400;
               break;
            case 8:
               /* Watch off. */
               newaction = 0x1000;
               break;
            case 9:
               /* Goto. */
               newaction = 0x4000;
               /* Save the seq number to be for later. */
               newevent_seq_no = action->goto_list;
               break;
            case 1:
               /* Catch. */
               interceptionpoint = action->action_id;
               newaction = 0x0001;
               break;
            case 10:
               /* Monitor -long */
               notificationpoint = action->action_id;
               newaction = 0x0004;
               break;
            case 2:
               /* Monitor. */
               notificationpoint = action->action_id;
               newaction = 0x0004;
               break;
            case 12:
               /* User function call */
               newaction = 0x10000;
               break;
            case 14:
               /* Enable hw trace */
               newaction = 0x100000;
               break;
            case 15:
               /* Disable hw trace */
               newaction = 0x400000;
               break;
            default:
               break;
         }
         /* Negate the action if required. */
         if (action->negated)
            newaction <<= 1;
         /* Log the action for later. */
         actions |= newaction;
      }
   }

   /* Done if there are no actions to be performed. */
   if (!actions)
      return;

   /* Call user specified function, if it is defined in ose_reg_user_func() */
   if (rw.user_action_function != NULL && ((actions & 0x30000) == 0x10000))
   {
      MonitorBindInfo info;

      info.sigNo = MONITOR_GET_BIND_TRACE_REPLY;

      info.reference = notificationpoint;
      BKL_LOCK();
      info.time.sec = rw.abs_sec;
      info.time.sectick = rw.abs_tick;
      info.time.tick = get_systime((OSTICK *) &info.time.ntick);
      BKL_UNLOCK();
      info.reserved0 = 0;
      info.bound = (U32) proc->pid;
      info.fromEuId = old_cpu;
      info.toEuId = new_cpu;
      info.lineNumber = 0;
      info.fileNameOffset = 0;
      info.reserved1 = 0;
      info.data[0] = 0;

      send_proc_info(proc);

      rw.user_action_function((union SIGNAL *)&info,
                           offsetof(MonitorBindInfo, data),
                           NULL,
                           0);
   }

   /* Turn on the trace if required. */
   if ((actions & 0x00c0) == 0x0040)
   {
      rw.precious->trace_enabled = rw.trace_enabled = (OSBOOLEAN) 1;
      ose_trace_enable(rw.trace);
   }

   /* Turn on the hardware trace if required. */
   if ((actions & 0x300000) == 0x100000)
   {
      (*rw.hw_trace[0])();
   }

   /* Turn on the watch if required. ?? */
   if ((actions & 0x0c00) == 0x0400)
   {
   }

   /* Turn off the watch if required. ?? */
   if ((actions & 0x3000) == 0x1000)
   {
   }

   /* Goto a new event sequence number if required. */
   if ((actions & 0xc000) == 0x4000)
   {
      rw.precious->action_state = rw.action_state = newevent_seq_no;
   }

   /* Trace the event if required. */
   if ((actions & 0x0030) == 0x0010)
   {
      union TRACE    *te;
      struct trace_procname *proc_name;

      /* Allocate a trace buffer. */
      te = trace_alloc_from(sizeof (struct trace_bind),
                            proc, &proc_name);

      /* Fill in the trace buffer. */
      if (te)
      {
         /* Fill the trace buffer. */
         te->trace_bind.event_no = TRACE_BIND;
         te->trace_bind.abstime = get_trace_abstime();;
         te->trace_bind.process = proc_name;
         te->trace_bind.from_eu = old_cpu;
         te->trace_bind.to_eu = new_cpu;

         /* Release the trace buffer. */
         ose_trace_commit_inline(te);
      }
   }

   /* Turn off the trace if required. */
   if ((actions & 0x0300) == 0x0100)
   {
      rw.precious->trace_enabled = rw.trace_enabled = (OSBOOLEAN) 0;
      ose_trace_disable(rw.trace);
   }

   /* Turn off the hardware trace if required. */
   if ((actions & 0xc00000) == 0x400000)
   {
      (*rw.hw_trace[1])();
   }   

   /* Monitor the event if required. */
   if ((actions & 0x000c) == 0x0004)
   {
      union SIGNAL   *txsig;

      /* Monitor condition was true. Send to ldm if there is space in the mon queue. */
      /* Check that monitoring is enabled and that there is space in the monitor queue. */
      if ((!rw.monitor_enabled) || (rw.monitor_overflow))
      {
         SPIN_LOCK(&rw.event_notify_lock);
         if (rw.monitor_overflow)
         {
            rw.monitor_overflow++;
         }
         SPIN_UNLOCK(&rw.event_notify_lock);
         goto cant_monitor;
      }

      /* Allocate a monitor buffer. */
      set_suid(1);
      txsig = alloc(sizeof (struct event_report_verbose), MONITOR_REPORT);

      BKL_LOCK();
      txsig->event_report_verbose.event_abssec = rw.abs_sec;
      txsig->event_report_verbose.event_abstick = rw.abs_tick;
      txsig->event_report_verbose.event_absutick = rw.abs_usec;
      BKL_UNLOCK();
      /* Look out with the parameter type to get_systime()! */
      txsig->event_report_verbose.event_tick = get_systime((OSTICK *) & txsig->event_report_verbose.event_utick);
      txsig->event_report_verbose.sig_from = (U32) 0;
      txsig->event_report_verbose.sig_to = (U32) proc->pid;
      txsig->event_report_verbose.ecode = old_cpu;
      txsig->event_report_verbose.signo = new_cpu;
      txsig->event_report_verbose.cpu = 0;
      txsig->event_report_verbose.sig_data = 0;
      txsig->event_report_verbose.sig_datasize = 0;

      txsig->event_report_verbose.event_type = 14;   /* Process bind event. */
      txsig->event_report_verbose.action = 2;        /* Monitor action. */

      txsig->event_report_verbose.sig_address = 0;
      txsig->event_report_verbose.sig_size = 0;

      txsig->event_report_verbose.pid_from = 0;
      txsig->event_report_verbose.pid_to = txsig->event_report_verbose.sig_to;
      txsig->event_report_verbose.ptype_from = 0;
      txsig->event_report_verbose.ptype_to = make_ptype(proc);
      txsig->event_report_verbose.line = 0;
      txsig->event_report_verbose.file = 0;
      txsig->event_report_verbose.action_id = notificationpoint;

      /* Disable interrupts */
      SPIN_LOCK(&rw.event_notify_lock);
      /* Check again that monitoring is enabled and no overflow present. */
      assert(rw.monitor_hook_flow >= 0);
      if ((rw.monitor_enabled) && (0 == rw.monitor_hook_flow))
      {
         rw.monitor_overflow++;
      }
      if ((!rw.monitor_enabled) || (0 == rw.monitor_hook_flow))
      {
         SPIN_UNLOCK(&rw.event_notify_lock);
         /* Free the buffer and abort if disabled. */
         free_buf(&txsig);
         set_suid(0);
         goto cant_monitor;
      }
      /* Take a flow control token . */
      rw.monitor_hook_flow--;
      /* Enable interrupts */
      SPIN_UNLOCK(&rw.event_notify_lock);

      /* Send an event_report_verbose to the ldm. */
      zzsys_send_w_s(&txsig, rw.zzose_rmm_, rw.zzose_rmm_);
      set_suid(0);
   }
 cant_monitor:

   /* Catch the event if required. */
   if ((actions & 0x0003) == 0x0001)
   {
      /* Catch condition was true. Catch it if the scope is not stopped. */
      /* Disable interrupts */
      BKL_LOCK();
      /* Check that the system is not already caught. */
      if (rw.catch_enable)
      {
         union SIGNAL   *txsig;

         /* Disable catcher. */
         txsig = rw.catch_enable;
         rw.catch_enable = NULL;
         /* Store sig pointer and process id for the ldm to access. */
         rw.caught_pid = current_process();

         txsig->event_report_verbose.event_abssec = rw.abs_sec;
         txsig->event_report_verbose.event_abstick = rw.abs_tick;
         txsig->event_report_verbose.event_absutick = rw.abs_usec;
         /* Enable interrupts. */
         BKL_UNLOCK();
         /* Set up the event_report_verbose signal. Replace process names with id's. */
         txsig->event_report_verbose.sig_no = CATCH_REPORT;
         txsig->event_report_verbose.event_tick = get_systime((OSTICK *) & txsig->event_report_verbose.event_utick);
         txsig->event_report_verbose.sig_from = (U32) 0;
         txsig->event_report_verbose.sig_to = (U32) proc->pid;
         txsig->event_report_verbose.ecode = old_cpu;
         txsig->event_report_verbose.signo = new_cpu;
         txsig->event_report_verbose.cpu = 0;
         txsig->event_report_verbose.sig_data = 0;
         txsig->event_report_verbose.sig_datasize = (U16) 0;

         txsig->event_report_verbose.event_type = 14;   /* Process bind event. */
         txsig->event_report_verbose.action = 1;        /* Catch action. */

         txsig->event_report_verbose.sig_address = 0;
         txsig->event_report_verbose.sig_size = 0;

         txsig->event_report_verbose.pid_from = 0;
         txsig->event_report_verbose.pid_to = txsig->event_report_verbose.sig_to;
         txsig->event_report_verbose.ptype_from = 0;
         txsig->event_report_verbose.ptype_to = make_ptype(proc);
         txsig->event_report_verbose.line = 0;
         txsig->event_report_verbose.file = 0;
         txsig->event_report_verbose.action_id = interceptionpoint;

         /* Send an event_report to the ldm. */
         set_suid(1);
         restore(txsig);
         zzsys_send_w_s(&txsig, current_process(), rw.zzose_rmm_);
         set_suid(0);

         /* Intercept the breakpoint scope. */
         intercept_evact_scope();

         /* Clear the caught flag. */
         rw.caught_sig = NULL;
      }
      else
      {
         BKL_UNLOCK();
      }
   }
}
#endif /* OSE_SMP */

static void
handle_error_event(struct debug_pcb *proc, OSBOOLEAN user_called, OSERRCODE ecode,
                   OSERRCODE extra, int interrupt, int called_cnt)
{
/*
** The actions flag field is defined as follows:
** D0 = 1 => Catch the event.                   Mask = 0x0003
** D1 = 1 => Ignore the catch bit.
** D2 = 1 => Monitor the event.                 Mask = 0x000c
** D3 = 1 => Ignore the monitor bit.
** D4 = 1 => Trace the event.                   Mask = 0x0030
** D5 = 1 => Ignore the trace bit.
** D6 = 1 => Turn on the trace.                 Mask = 0x00c0
** D7 = 1 => Ignore the trace on bit.
** D8 = 1 => Turn off the trace.                Mask = 0x0300
** D9 = 1 => Ignore the trace off bit.
** D10 = 1 => Turn on the watch timer.          Mask = 0x0c00
** D11 = 1 => Ignore the watch on bit.
** D12 = 1 => Turn off the watch timer.         Mask = 0x3000
** D13 = 1 => Ignore the watch off bit.
** D14 = 1 => Goto a new event sequence number. Mask = 0xc000
** D15 = 1 => Ignore the goto bit.
*/
   LOCKSTATE       msr;
   uint32_t actions = 0;
   struct ldefine_action *action;
   uint8_t event_type;

   U32 cpu;

   U8              oldevent_seq_no = rw.action_state;
   U8              newevent_seq_no = 0;

   uint32_t        interceptionpoint = 0;
   uint32_t        notificationpoint = 0;

   /* Skip events that don't involve any process in the display scope. */
   if (!(proc->debug_flags & IN_DEBUGGER_SCOPE))
      return;

   /* Choose between trace OSE or USER error request. */
   if (user_called)
   {
      event_type = TRACE_USER_ERROR;
      action = rw.actions SMP_CODE([0]).action_list_error;
   }
   else
   {
      event_type = TRACE_OSERROR;
      action = rw.actions SMP_CODE([0]).action_list_oserror;
   }

   /* Scan the actions list for actions to perform. */
   for (; action; action = action->next)
   {
      /* Check for event sequence match. */
      if (action->event_seq_no && (action->event_seq_no != oldevent_seq_no))
         continue;

      /* Check for process match. */
      if (!zzrmm_match_proc(proc, (struct procnamelist *) action->align))
         continue;

      /* Check for error code match. */
      if (ecode < action->ecode || ecode > action->sig_to)
         continue;

      /* OSECR-13648: When the error is handled by a process then
       * basic events should be ignored by first call so
       * that debugger will not intercept when the system parameter
       * krn/monitor/error=trap_user_handled:0.
       */
      if (action->is_basic_event && called_cnt == 1 && !rw.monitor_error)
      {
          continue;
      }
      if(!action->is_basic_event && called_cnt == 2 && !rw.monitor_error)
      {
         continue;
      }

      /* Check for an active event counter. */
      if (unlikely(action->orig_event_count))
      {
         /* Update the event counter and ignore the event if the limit is not reached. */
         SPIN_LOCK_SAVE(&action->SMP_CODE(orig_action->)action_lock, msr);
         if (--(action->SMP_CODE(orig_action->)event_count))
         {
            SPIN_LOCK_RESTORE(&action->SMP_CODE(orig_action->)action_lock, msr);
            continue;
         }
         else
         {
            action->SMP_CODE(orig_action->)event_count = action->orig_event_count;
            SPIN_LOCK_RESTORE(&action->SMP_CODE(orig_action->)action_lock, msr);
         }
      }

      /* Decide which action to take and update action flags. */
      {
         U32             newaction = 0;

         switch (action->action)
         {
            case 3:
               /* Trace. */
               newaction = 0x0010;
               break;
            case 1:
               /* Catch. */
               interceptionpoint = action->action_id;
               newaction = 0x0001;
               break;
            case 10:
               /* Monitor -long */
               notificationpoint = action->action_id;
               newaction = 0x0004;
               break;
            case 2:
               /* Monitor. */
               notificationpoint = action->action_id;
               newaction = 0x0004;
               break;
            case 4:
               /* Trace on. */
               newaction = 0x0040;
               break;
            case 5:
               /* Trace off. */
               newaction = 0x0100;
               break;
            case 7:
               /* Watch on. */
               newaction = 0x0400;
               break;
            case 8:
               /* Watch off. */
               newaction = 0x1000;
               break;
            case 9:
               /* Goto. */
               newaction = 0x4000;
               /* Save the seq number to be for later. */
               newevent_seq_no = action->goto_list;
               break;
            case 12:
               /* User function */
               newaction = 0x10000;
               break;
            case 14:
               /* Enable hw trace */
               newaction = 0x100000;
               break;
            case 15:
               /* Disable hw trace */
               newaction = 0x400000;
               break;
            default:
               break;
         }
         /* Negate the action if required. */
         if (action->negated)
            newaction <<= 1;
         /* Log the action for later. */
         actions |= newaction;
      }
   }

   /* Done if there are no actions to be performed. */
   if (!actions)
   {
      return;
   }

   cpu = SMP_CODE_YES_NO(zzkrn_get_pid_info(proc->pid, 13), 0);

   /* If event occurs outside the break scope we clear the catch action */
   if (!(proc->debug_flags & IN_BREAK_SCOPE))
   {
      actions &= ~0x0003; /*Skip the catch action. */
   }

   /* Skip some operations for fatal errors. */
   if ((event_type == 6) && (ecode & OSE_EFATAL_MASK))
   {
      if (rw.monitor_sys_error == 0)
      {
         actions &= ~0x0003;       /* Skip the catch action. */
      }
      actions &= ~0x000c;       /* Skip the monitor action. */
      actions &= ~0x0c00;       /* Skip the watch on action. */
      actions &= ~0x3000;       /* Skip the watch off action. */
   }

   /*
    ** Skip some operations for out of pool errors (which ought to be
    ** fatal).
    */
   if (event_type == 6)
   {
      switch (ecode & OSE_EERROR_MASK)
      {
         case OSE_ENO_KERN_SPACE:
         case OSE_ENO_USER_SIGSPACE:
         case OSE_ENO_USER_STACKSPACE:
            if (rw.monitor_sys_error == 0)
            {
               actions &= ~0x0003; /* Skip the catch action. */
            }
            actions &= ~0x000c; /* Skip the monitor action. */
            actions &= ~0x0c00; /* Skip the watch on action. */
            actions &= ~0x3000; /* Skip the watch off action. */
            break;
         default:;
      }
   }

   /* Call user specified function, if it is defined in ose_reg_user_func() */
   if (rw.user_action_function != NULL && ((actions & 0x30000) == 0x10000))
   {
      MonitorErrorInfo info;

      info.sigNo = MONITOR_GET_ERROR_TRACE_REPLY;

      info.reference = notificationpoint;
      BKL_LOCK_SAVE(msr);
      info.time.sec = rw.abs_sec;
      info.time.sectick = rw.abs_tick;
      info.time.tick = get_systime((OSTICK *) &info.time.ntick);
      BKL_LOCK_RESTORE(msr);
      info.from = (U32) proc->pid;
      info.exec = (event_type == 6);
      info.error = ecode;
      info.extra = (SIGSELECT) extra;
      info.lineNumber = 0;
      info.fileNameOffset = 0;
      info.euId = cpu;
      info.reserved0 = 0;
      info.reserved1 = 0;
      info.data[0] = 0;

      send_proc_info(proc);

      rw.user_action_function((union SIGNAL *)&info,
                           offsetof(MonitorErrorInfo, data),
                           NULL,
                           0);
   }

   /* Turn on the trace if required. */
   if ((actions & 0x00c0) == 0x0040)
   {
      rw.precious->trace_enabled = rw.trace_enabled = (OSBOOLEAN) 1;
      ose_trace_enable(rw.trace);
   }

   /* Turn on the hardware trace if required. */
   if ((actions & 0x300000) == 0x100000)
   {
      (*rw.hw_trace[0])();
   } 

   /* Goto a new event sequence number if required. */
   if ((actions & 0xc000) == 0x4000)
   {
      rw.precious->action_state = rw.action_state = newevent_seq_no;
   }

   /* Trace the event if required. */
   if ((actions & 0x0030) == 0x0010)
   {
      union TRACE    *te;
      struct trace_procname *proc_name;

      /* Allocate a trace buffer. */
      te = trace_alloc_from(sizeof (struct trace_user_error),
                            proc, &proc_name);
      /* Fill in the trace buffer. */
      if (te)
      {
         /* Fill the trace buffer. */
         te->trace_user_error.event_no = event_type;
         te->trace_user_error.abstime = get_trace_abstime();;
         te->trace_user_error.process = proc_name;
         te->trace_user_error.ecode = ecode;
         te->trace_user_error.extra = extra;
         te->trace_user_error.cpu = cpu;

         /* Release the trace buffer. */
         ose_trace_commit_inline(te);
      }
   }

   /* Turn off the trace if required. */
   if ((actions & 0x0300) == 0x0100)
   {
      rw.precious->trace_enabled = rw.trace_enabled = (OSBOOLEAN) 0;
      ose_trace_disable(rw.trace);
   }

   /* Turn off the hardware trace if required. */
   if ((actions & 0xc00000) == 0x400000)
   {
      (*rw.hw_trace[1])();
   }


   /* Monitor the event if required. */
   if ((actions & 0x000c) == 0x0004)
   {
      union SIGNAL   *txsig;

      /* Monitor condition was true. Send to ldm if there is space in the mon queue. */
      /* Check that monitoring is enabled and that there is space in the monitor queue. */
      if ((!rw.monitor_enabled) || (rw.monitor_overflow))
      {
         SPIN_LOCK_SAVE(&rw.event_notify_lock, msr);
         if (rw.monitor_overflow)
         {
            rw.monitor_overflow++;
         }
         SPIN_LOCK_RESTORE(&rw.event_notify_lock, msr);
         goto cant_monitor;
      }

      if (!interrupt)
      {
         /* Unlock the dispatcher. */
         UNLOCK_SWAP();
      }

      /* Allocate a monitor buffer. */
      set_suid(1);
      txsig = alloc(sizeof (struct event_report_verbose), MONITOR_REPORT);

      /* Set up the event_report_verbose signal. Replace process names with id's. */
      BKL_LOCK_SAVE(msr);
      txsig->event_report_verbose.event_abssec = rw.abs_sec;
      txsig->event_report_verbose.event_abstick = rw.abs_tick;
      txsig->event_report_verbose.event_absutick = rw.abs_usec;
      BKL_LOCK_RESTORE(msr);
      /* Look out with the parameter type to get_systime()! */
      txsig->event_report_verbose.event_tick = get_systime((OSTICK *) & txsig->event_report_verbose.event_utick);
      txsig->event_report_verbose.sig_from = (U32) proc->pid;
      txsig->event_report_verbose.sig_to = (U32) 0;
      txsig->event_report_verbose.ecode = ecode;
      txsig->event_report_verbose.signo = (SIGSELECT) extra;
      txsig->event_report_verbose.cpu = cpu;
      txsig->event_report_verbose.sig_data = 0;
      txsig->event_report_verbose.sig_datasize = (U16) 0;
      txsig->event_report_verbose.event_type = event_type;
      txsig->event_report_verbose.action = 2;   /* Monitor action. */
      txsig->event_report_verbose.sig_address = (U32) 0;
      txsig->event_report_verbose.sig_size = (U32) 0;

      txsig->event_report_verbose.pid_from = txsig->event_report_verbose.sig_from;
      txsig->event_report_verbose.pid_to = txsig->event_report_verbose.sig_to;
      txsig->event_report_verbose.ptype_from = make_ptype(proc);
      txsig->event_report_verbose.ptype_to = (U32) 0;
      txsig->event_report_verbose.line = (U32) 0;
      txsig->event_report_verbose.file = (U32) 0;
      txsig->event_report_verbose.action_id = notificationpoint;

      if (!interrupt)
      {
         LOCK_SWAP();
      }
      /* Disable interrupts */
      SPIN_LOCK_SAVE(&rw.event_notify_lock, msr);
      /* Check again that monitoring is enabled and no overflow present. */
      if ((rw.monitor_enabled) && (!rw.monitor_hook_flow))
      {
         rw.monitor_overflow++;
      }
      if ((!rw.monitor_enabled) || (!rw.monitor_hook_flow))
      {
         /* Free the buffer and abort if disabled. */
         SPIN_LOCK_RESTORE(&rw.event_notify_lock, msr);
         free_buf(&txsig);
         set_suid(0);
         goto cant_monitor;
      }
      /* Take a flow control token . */
      rw.monitor_hook_flow--;
      /* Enable interrupts */
      SPIN_LOCK_RESTORE(&rw.event_notify_lock, msr);

      /* Send an event_report_verbose to the ldm. */
      if (interrupt)
      {
         send_w_s(&txsig, rw.zzose_rmm_, rw.zzose_rmm_);
      }
      else
      {
         zzsys_send_w_s(&txsig, rw.zzose_rmm_, rw.zzose_rmm_);
      }
      set_suid(0);
   }
 cant_monitor:
   /* Catch the event if required. */
   if ((actions & 0x0003) == 0x0001)
   {
      /* Catch condition was true. Catch it if the scope is not stopped. */
      /* Disable interrupts */
      BKL_LOCK_SAVE(msr);
      /* Check that the system is not already caught. */
      if (rw.catch_enable)
      {
         union SIGNAL   *txsig;

         /* Disable catcher. */
         txsig = rw.catch_enable;
         rw.catch_enable = NULL;
         /* Store sig pointer and process id for the ldm to access. */
         rw.caught_pid = current_process();
         txsig->event_report_verbose.event_abssec = rw.abs_sec;
         txsig->event_report_verbose.event_abstick = rw.abs_tick;
         txsig->event_report_verbose.event_absutick = rw.abs_usec;
         /* Enable interrupts. */
         BKL_LOCK_RESTORE(msr);
         /* Set up the event_report_verbose signal. Replace process names with id's. */
         txsig->event_report_verbose.sig_no = CATCH_REPORT;
         txsig->event_report_verbose.event_tick = get_systime((OSTICK *) & txsig->event_report_verbose.event_utick);
         txsig->event_report_verbose.sig_from = (U32) proc->pid;
         txsig->event_report_verbose.sig_to = (U32) 0;
         txsig->event_report_verbose.ecode = ecode;
         txsig->event_report_verbose.signo = (SIGSELECT) extra;
         txsig->event_report_verbose.cpu = cpu;
         txsig->event_report_verbose.sig_data = 0;
         txsig->event_report_verbose.sig_datasize = (U16) 0;
         txsig->event_report_verbose.event_type = event_type;
         txsig->event_report_verbose.action = 1;        /* Catch action. */
         txsig->event_report_verbose.sig_address = (U32) 0;
         txsig->event_report_verbose.sig_size = 0;

         txsig->event_report_verbose.pid_from = txsig->event_report_verbose.sig_from;
         txsig->event_report_verbose.pid_to = txsig->event_report_verbose.sig_to;
         txsig->event_report_verbose.ptype_from = make_ptype(proc);
         txsig->event_report_verbose.ptype_to = (U32) 0;
         txsig->event_report_verbose.line = (U32) 0;
         txsig->event_report_verbose.file = (U32) 0;
         txsig->event_report_verbose.action_id = interceptionpoint;

         /* Send an event_report_verbose to the ldm. */
         set_suid(1);
         restore(txsig);
         if (interrupt)
         {
            send_w_s(&txsig, current_process(), rw.zzose_rmm_);
            set_suid(0);
            /* Intercept the breakpoint scope. */
            intercept_evact_scope();
         }
         else
         {
            zzsys_send_w_s(&txsig, current_process(), rw.zzose_rmm_);
            set_suid(0);
            LOCK_SWAP();
            /* Intercept the breakpoint scope. */
            intercept_evact_scope();
            UNLOCK_SWAP();
         }
      }
      else
      {
         BKL_LOCK_RESTORE(msr);
      }
   }
}

void
zzerror_dbghook(struct debug_pcb *proc, OSERRCODE ecode, OSERRCODE extra)
{
   /* Lock the dispatcher. */
   LOCK_SWAP();
   handle_error_event(proc, TRUE, ecode, extra, 0, 1);
   UNLOCK_SWAP();
}

void
zzierror_dbghook(struct debug_pcb *proc, OSERRCODE ecode, OSERRCODE extra)
{
   handle_error_event(proc, TRUE, ecode, extra, 1, 1);
}

void
zzoseerror_dbghook(struct debug_pcb *proc, OSERRCODE ecode, OSERRCODE extra)
{
   Offspring         *c;

   /* Lock the dispatcher. */
   LOCK_SWAP();

#ifdef USE_EPM
#ifdef USE_REGDUMP_REGISTERS
   if ((ecode & OSE_EERROR_MASK) == OSE_EUNEXPECTED_EXCEPTION_REGDUMP)
   {
      /* find the offspring that intercepts this process */
      for (c = list_head(rw.offsprings); c != NULL; c = list_next(rw.offsprings, &c->link))
      {
         if (zzrmm_process_in_scope(proc->pid, &c->debug_scope))
         {
        	LOCKSTATE       msr;
        	BKL_LOCK_SAVE(msr);
        	/* Save the register dump for this process. */
        	add_regdump(c, (struct cpu_regdump *)extra);
        	BKL_LOCK_RESTORE(msr);
            break;
         }
      }
   }
#endif
#endif

   handle_error_event(proc, FALSE, ecode, extra, 0, 1);

#ifdef USE_EPM
#ifdef USE_REGDUMP_REGISTERS
   if ((ecode & OSE_EERROR_MASK) == OSE_EUNEXPECTED_EXCEPTION_REGDUMP)
   {
      /* find the offspring that intercepts this process */
      for (c = list_head(rw.offsprings); c != NULL; c = list_next(rw.offsprings, &c->link))
      {
         if (zzrmm_process_in_scope(proc->pid, &c->debug_scope))
         {
        	/* Clear the register dump for this process. */
        	clear_regdump(c, proc->pid);
            break;
         }
      }
   }
#endif
#endif

   UNLOCK_SWAP();
}

void
zzioseerror_dbghook(struct debug_pcb *proc, OSERRCODE ecode, OSERRCODE extra)
{
   handle_error_event(proc, FALSE, ecode, extra, 1, 1);
}

static void
handle_event_report(struct event_report_verbose *event)
{
   Offspring         *offspring;

   assert(event);
   assert(rw.offsprings);

   offspring = rw.evact_lock_owner;

   if (offspring != NULL)
   {
      uint32_t        isig;
      union SIGNAL   *sig[3];
      char           *from = NULL;
      char           *to = NULL;
      OSBUFSIZE       from_len = 0;
      OSBUFSIZE       to_len = 0;
      OSADDRESS       from_ep = 0;
      OSADDRESS       to_ep = 0;
      PROCESS         from_bid = 0;
      PROCESS         to_bid = 0;
      uint32_t        from_props = 0;
      uint32_t        to_props = 0;
      OSBUFSIZE       file_len = 0;
      PROCESS         from_sid = 0;
      PROCESS         to_sid = 0;

      if (event->line)
      {
         file_len = strlen((char *)&event->data[event->file]);
      }

      /* Name already filled in for kill events */
      if (event->event_type != 2)
      {
         char           *s;
         OSBUFSIZE       tmp_len;
         static char     from_buf[MAX_NAME_LEN + 1];
         static char     to_buf[MAX_NAME_LEN + 1];
         struct debug_pcb *dcb;

         from = from_buf;
         to = to_buf;

         /* Get names of processes. */
         LOCK_SWAP();

         dcb = zzget_debugpcb((PROCESS) event->sig_from);
         if (dcb != NULL)
         {
            from_props = (dcb->debug_flags & SYS_PROCESS) != 0;
            from_ep = (OSADDRESS) dcb->entrypoint;
            from_bid = pcb_to_bid(dcb);
            {
               PROCESS         sid = get_segid(from_bid);

               from_sid = (from_bid != sid) ? sid : 0;
            }
            from_len = strlen(dcb->name);
            if (from_len < MAX_NAME_LEN)
            {
               tmp_len = strlen(dcb->blockname) + 1;
               if ((from_len + tmp_len) < MAX_NAME_LEN)
               {
                  from_len += tmp_len;
                  for (s = dcb->blockname; *s;)
                  {
                     *from++ = *s++;
                  }
                  *from++ = ':';
                  strcpy(from, dcb->name);
                  from = from_buf;
               }
               else
               {
                  strcpy(from, dcb->name);
               }
            }
            else
            {
               strcpy(from, dcb->name + (from_len - MAX_NAME_LEN));
               from_len = MAX_NAME_LEN;
            }
         }
         else
         {
            from_len = 1;
            strcpy(from, "-");
         }

         dcb = zzget_debugpcb((PROCESS) event->sig_to);
         if (dcb != NULL)
         {
            to_props = (dcb->debug_flags & SYS_PROCESS) != 0;
            to_ep = (OSADDRESS) dcb->entrypoint;
            to_bid = pcb_to_bid(dcb);
            {
               PROCESS         sid = get_segid(to_bid);

               to_sid = (to_bid != sid) ? sid : 0;
            }
            to_len = strlen(dcb->name);
            if (to_len < MAX_NAME_LEN)
            {
               tmp_len = strlen(dcb->blockname) + 1;
               if ((to_len + tmp_len) < MAX_NAME_LEN)
               {
                  to_len += tmp_len;
                  for (s = dcb->blockname; *s;)
                  {
                     *to++ = *s++;
                  }
                  *to++ = ':';
                  strcpy(to, dcb->name);
                  to = to_buf;
               }
               else
               {
                  strcpy(to, dcb->name);
               }
            }
            else
            {
               strcpy(to, dcb->name + (to_len - MAX_NAME_LEN));
               to_len = MAX_NAME_LEN;
            }
         }
         else
         {
            to_len = 1;
            strcpy(to, "-");
         }

         UNLOCK_SWAP();
      }
      else
      {
         from = (char *)&event->data[event->sig_from];
         from_len = strlen(from);

         from_props = match_proc_data(event->pid_from, from_bid, from, 0, rw.expanded_sysprocnames);
      }

      switch (event->action)
      {
         case 1:               /* Catch */
            isig = 0;
            sig[0] = sig[1] = sig[2] = NULL;

            if (!offspring->intercept_enabled)
            {
               LOCK_SWAP();
               resume_evact_scope();
               UNLOCK_SWAP();
               break;
            }

            switch (event->event_type)
            {
               case 1:         /* Create */
               case 2:         /* Kill */
                  /* Basic event reports to source code debuggers */
                  if (rw.evact_lock_owner->debug_scope.type != MONITOR_SCOPE_VOID)
                  {
                	 const char    *s, *b, *p;

                     sig[isig] = alloc(sizeof (MonitorBasicCreateKillNotify) + from_len,
                                       MONITOR_BASIC_CREATE_KILL_NOTIFY);
                     sig[isig]->basic_create_kill_notify.pid = event->pid_from;
                     sig[isig]->basic_create_kill_notify.bid = from_bid;
                     sig[isig]->basic_create_kill_notify.sid = from_sid;
                     sig[isig]->basic_create_kill_notify.type = event->ptype_from;

                     if (event->event_type == 2)
                        sig[isig]->basic_create_kill_notify.state = MONITOR_PROCESS_STATE_KILLED;
                     else
                        sig[isig]->basic_create_kill_notify.state = MONITOR_PROCESS_STATE_READY;

                     sig[isig]->basic_create_kill_notify.properties = from_props;
                     sig[isig]->basic_create_kill_notify.entrypoint = from_ep;

                     split_string(&s, &b, &p, from);
                     strcpy(sig[isig]->basic_create_kill_notify.name, p);
                  }
                  /* Full event reports to Illuminator */
                  else
                  {
                     if (!idbuf_contains(&offspring->notify_pids, event->pid_from))
                     {
                        sig[isig] = alloc(sizeof (MonitorProcessInfo) + from_len, MONITOR_PROCESS_INTERCEPT_NOTIFY);

                        sig[isig]->process_info.pid = event->pid_from;
                        sig[isig]->process_info.bid = from_bid;
                        sig[isig]->process_info.sid = from_sid;
                        sig[isig]->process_info.type = event->ptype_from;
                        sig[isig]->process_info.entrypoint = from_ep;
                        sig[isig]->process_info.properties = from_props;
                        sig[isig]->process_info.reserved0 = 0;
                        strcpy(sig[isig]->process_info.name, from);

                        isig++;
                     }

                     sig[isig] = alloc(sizeof (MonitorSwapCreateKillInfo) + file_len,
                                       (event->event_type == 1)
                                       ? MONITOR_CREATE_INTERCEPT_NOTIFY : MONITOR_KILL_INTERCEPT_NOTIFY);

                     sig[isig]->swap_create_kill_info.reference = event->action_id;
                     sig[isig]->swap_create_kill_info.time.tick = event->event_tick;
                     sig[isig]->swap_create_kill_info.time.ntick = event->event_utick;
                     sig[isig]->swap_create_kill_info.time.sec = event->event_abssec;
                     sig[isig]->swap_create_kill_info.time.sectick = event->event_abstick;
                     sig[isig]->swap_create_kill_info.from = 0; /* TODO: Creator / Killer */
                     sig[isig]->swap_create_kill_info.to = event->pid_from;
                     sig[isig]->swap_create_kill_info.lineNumber = event->line;
                     sig[isig]->swap_create_kill_info.fileNameOffset = 0;
                     if (event->line)
                     {
                        strcpy((char *)sig[isig]->swap_create_kill_info.data, (char *)&event->data[event->file]);
                     }
                     sig[isig]->swap_create_kill_info.euId = event->cpu;
                     sig[isig]->swap_create_kill_info.reserved0 = 0;
                     sig[isig]->swap_create_kill_info.reserved1 = 0;
                  }
                  break;

               case 4:         /* send */
               case 5:         /* receive */
                  if (!idbuf_contains(&offspring->notify_pids, event->pid_from))
                  {
                     sig[isig] = alloc(sizeof (MonitorProcessInfo) + from_len, MONITOR_PROCESS_INTERCEPT_NOTIFY);

                     sig[isig]->process_info.pid = event->pid_from;
                     sig[isig]->process_info.bid = from_bid;
                     sig[isig]->process_info.sid = from_sid;
                     sig[isig]->process_info.type = event->ptype_from;
                     sig[isig]->process_info.entrypoint = from_ep;
                     sig[isig]->process_info.properties = from_props;
                     sig[isig]->process_info.reserved0 = 0;
                     strcpy(sig[isig]->process_info.name, from);

                     isig++;
                  }
                  if (!idbuf_contains(&offspring->notify_pids, event->pid_to))
                  {
                     sig[isig] = alloc(sizeof (MonitorProcessInfo) + to_len, MONITOR_PROCESS_INTERCEPT_NOTIFY);

                     sig[isig]->process_info.pid = event->pid_to;
                     sig[isig]->process_info.bid = to_bid;
                     sig[isig]->process_info.sid = to_sid;
                     sig[isig]->process_info.type = event->ptype_to;
                     sig[isig]->process_info.entrypoint = to_ep;
                     sig[isig]->process_info.properties = from_props;
                     sig[isig]->process_info.reserved0 = 0;
                     strcpy(sig[isig]->process_info.name, to);

                     isig++;
                  }

                  sig[isig] = alloc(sizeof (MonitorSendReceiveInfo) + file_len,
                                    (event->event_type == 4)
                                    ? MONITOR_SEND_INTERCEPT_NOTIFY : MONITOR_RECEIVE_INTERCEPT_NOTIFY);

                  sig[isig]->send_receive_info.reference = event->action_id;
                  sig[isig]->send_receive_info.time.tick = event->event_tick;
                  sig[isig]->send_receive_info.time.ntick = event->event_utick;
                  sig[isig]->send_receive_info.time.sec = event->event_abssec;
                  sig[isig]->send_receive_info.time.sectick = event->event_abstick;
                  sig[isig]->send_receive_info.signalNumber = event->signo;
                  sig[isig]->send_receive_info.signalSender = event->pid_from;
                  sig[isig]->send_receive_info.signalAddressee = event->pid_to;
                  sig[isig]->send_receive_info.signalSize = event->sig_size;
                  sig[isig]->send_receive_info.signalAddress = event->sig_address;
                  sig[isig]->send_receive_info.signalId = 0;
                  sig[isig]->send_receive_info.signalDataOffset = 0;
                  sig[isig]->send_receive_info.signalDataCount = event->sig_datasize;
                  if (event->sig_datasize)
                  {
                     memcpy(sig[isig]->send_receive_info.data, &event->data[event->sig_data], event->sig_datasize);
                  }
                  sig[isig]->send_receive_info.lineNumber = event->line;
                  sig[isig]->send_receive_info.fileNameOffset = event->sig_datasize;
                  if (event->line)
                  {
                     strcpy((char *)&sig[isig]->send_receive_info.data[event->sig_datasize], (char *)&event->data[event->file]);
                  }
                  sig[isig]->send_receive_info.euId = event->cpu;
                  sig[isig]->send_receive_info.reserved0 = 0;
                  sig[isig]->send_receive_info.reserved1 = 0;
                  break;

#ifdef HAS_TIMEOUT_EVENT
	       case MONITOR_EVENT_TIMEOUT:  /* timeout in receive_w_tmo */

		  if (offspring->handles_new_events)
                  {

                     if (!idbuf_contains(&offspring->notify_pids, event->pid_from))
                     {
                        sig[isig] = alloc(sizeof (MonitorProcessInfo) + from_len, MONITOR_PROCESS_INTERCEPT_NOTIFY);

                        sig[isig]->process_info.pid = event->pid_from;
                        sig[isig]->process_info.bid = from_bid;
                        sig[isig]->process_info.sid = from_sid;
                        sig[isig]->process_info.type = event->ptype_from;
                        sig[isig]->process_info.entrypoint = from_ep;
                        sig[isig]->process_info.properties = from_props;
                        sig[isig]->process_info.reserved0 = 0;
                        strcpy(sig[isig]->process_info.name, from);

                        isig++;
                     }

                     sig[isig] = alloc(sizeof (MonitorTimeoutInfo) + file_len,
                                       MONITOR_TIMEOUT_INTERCEPT_NOTIFY);

                     sig[isig]->timeout_info.reference = event->action_id;
                     sig[isig]->timeout_info.time.tick = event->event_tick;
                     sig[isig]->timeout_info.time.ntick = event->event_utick;
                     sig[isig]->timeout_info.time.sec = event->event_abssec;
                     sig[isig]->timeout_info.time.sectick = event->event_abstick;
                     sig[isig]->timeout_info.signalAddressee = event->sig_from;
                     sig[isig]->timeout_info.timeout = event->ecode;
                     sig[isig]->timeout_info.lineNumber = event->line;
                     sig[isig]->timeout_info.fileNameOffset = 0;
                     if (event->line)
                     {
                        assert(file_len > 0);
                        strcpy((char *)&sig[isig]->timeout_info.data[0], (char *)&event->data[event->file]);
                     }

                     sig[isig]->timeout_info.euId = event->cpu;
                     sig[isig]->timeout_info.reserved0 = 0;

                  }

                  break;
#endif

               case MONITOR_EVENT_USER:  /* User event */
                  if (offspring->handles_new_events)
                  {
                     if (!idbuf_contains(&offspring->notify_pids, event->pid_from))
                     {
                        sig[isig] = alloc(sizeof (MonitorProcessInfo) + from_len, MONITOR_PROCESS_INTERCEPT_NOTIFY);

                        sig[isig]->process_info.pid = event->pid_from;
                        sig[isig]->process_info.bid = from_bid;
                        sig[isig]->process_info.sid = from_sid;
                        sig[isig]->process_info.type = event->ptype_from;
                        sig[isig]->process_info.entrypoint = from_ep;
                        sig[isig]->process_info.properties = from_props;
                        sig[isig]->process_info.reserved0 = 0;
                        strcpy(sig[isig]->process_info.name, from);

                        isig++;
                     }

                     sig[isig] = alloc(sizeof (MonitorUserInfo) + file_len,
                                       MONITOR_USER_INTERCEPT_NOTIFY);

                     sig[isig]->user_info.reference = event->action_id;
                     sig[isig]->user_info.time.tick = event->event_tick;
                     sig[isig]->user_info.time.ntick = event->event_utick;
                     sig[isig]->user_info.time.sec = event->event_abssec;
                     sig[isig]->user_info.time.sectick = event->event_abstick;
                     sig[isig]->user_info.eventNumber = event->signo;
                     sig[isig]->user_info.eventDataOffset = 0;
                     sig[isig]->user_info.eventDataCount = event->sig_datasize;
                     sig[isig]->user_info.from = event->sig_from;
                     sig[isig]->user_info.eventSize = event->sig_size;

                     if (event->sig_datasize)
                     {
                        memcpy(sig[isig]->user_info.data, &event->data[event->sig_data], event->sig_datasize);
                     }
                     sig[isig]->user_info.lineNumber = event->line;
                     if (event->line)
                     {
                        assert(file_len > 0);
                        sig[isig]->user_info.fileNameOffset = event->sig_datasize;
                        strcpy((char *)&sig[isig]->user_info.data[event->sig_datasize], (char *)&event->data[event->file]);
                     }
                     else
                     {
                        sig[isig]->user_info.fileNameOffset = 0;
                     }
                     sig[isig]->user_info.euId = event->cpu;
                     sig[isig]->user_info.reserved0 = 0;
                  }
                  break;

               case MONITOR_EVENT_ALLOC:  /* Alloc */
                  if (offspring->handles_new_events)
                  {
                     if (!idbuf_contains(&offspring->notify_pids, event->pid_from))
                     {
                        sig[isig] = alloc(sizeof (MonitorProcessInfo) + from_len, MONITOR_PROCESS_INTERCEPT_NOTIFY);

                        sig[isig]->process_info.pid = event->pid_from;
                        sig[isig]->process_info.bid = from_bid;
                        sig[isig]->process_info.sid = from_sid;
                        sig[isig]->process_info.type = event->ptype_from;
                        sig[isig]->process_info.entrypoint = from_ep;
                        sig[isig]->process_info.properties = from_props;
                        sig[isig]->process_info.reserved0 = 0;
                        strcpy(sig[isig]->process_info.name, from);

                        isig++;
                     }

                     sig[isig] = alloc(sizeof (MonitorAllocFreeBufInfo) + file_len,
                                       MONITOR_ALLOC_INTERCEPT_NOTIFY);


                     sig[isig]->alloc_free_buf_info.reference = event->action_id;
                     sig[isig]->alloc_free_buf_info.time.tick = event->event_tick;
                     sig[isig]->alloc_free_buf_info.time.ntick = event->event_utick;
                     sig[isig]->alloc_free_buf_info.time.sec = event->event_abssec;
                     sig[isig]->alloc_free_buf_info.time.sectick = event->event_abstick;
                     sig[isig]->alloc_free_buf_info.from = event->sig_from;
                     sig[isig]->alloc_free_buf_info.signalNumber = event->signo;
                     sig[isig]->alloc_free_buf_info.signalSize = event->sig_size;
                     sig[isig]->alloc_free_buf_info.signalAddress = event->sig_address;
                     sig[isig]->alloc_free_buf_info.signalId = 0;
                     sig[isig]->alloc_free_buf_info.lineNumber = event->line;
                     sig[isig]->alloc_free_buf_info.fileNameOffset = 0;
                     if (event->line)
                     {
                        assert(file_len > 0);
                        strcpy((char *)&sig[isig]->alloc_free_buf_info.data[0], (char *)&event->data[event->file]);
                     }

                     sig[isig]->alloc_free_buf_info.euId = event->cpu;
                     sig[isig]->alloc_free_buf_info.reserved0 = 0;
                  }
                  break;

               case MONITOR_EVENT_FREE_BUF:  /* Free buf */
                  if (offspring->handles_new_events)
                  {
                     if (!idbuf_contains(&offspring->notify_pids, event->pid_from))
                     {
                        sig[isig] = alloc(sizeof (MonitorProcessInfo) + from_len, MONITOR_PROCESS_INTERCEPT_NOTIFY);

                        sig[isig]->process_info.pid = event->pid_from;
                        sig[isig]->process_info.bid = from_bid;
                        sig[isig]->process_info.sid = from_sid;
                        sig[isig]->process_info.type = event->ptype_from;
                        sig[isig]->process_info.entrypoint = from_ep;
                        sig[isig]->process_info.properties = from_props;
                        sig[isig]->process_info.reserved0 = 0;
                        strcpy(sig[isig]->process_info.name, from);

                        isig++;
                     }

                     sig[isig] = alloc(sizeof (MonitorAllocFreeBufInfo) + file_len,
                                       MONITOR_FREE_BUF_INTERCEPT_NOTIFY);


                     sig[isig]->alloc_free_buf_info.reference = event->action_id;
                     sig[isig]->alloc_free_buf_info.time.tick = event->event_tick;
                     sig[isig]->alloc_free_buf_info.time.ntick = event->event_utick;
                     sig[isig]->alloc_free_buf_info.time.sec = event->event_abssec;
                     sig[isig]->alloc_free_buf_info.time.sectick = event->event_abstick;
                     sig[isig]->alloc_free_buf_info.from = event->sig_from;
                     sig[isig]->alloc_free_buf_info.signalNumber = event->signo;
                     sig[isig]->alloc_free_buf_info.signalSize = event->sig_size;
                     sig[isig]->alloc_free_buf_info.signalAddress = event->sig_address;
                     sig[isig]->alloc_free_buf_info.signalId = 0;
                     sig[isig]->alloc_free_buf_info.lineNumber = event->line;
                     sig[isig]->alloc_free_buf_info.fileNameOffset = 0;
                     if (event->line)
                     {
                        assert(file_len > 0);
                        strcpy((char *)&sig[isig]->alloc_free_buf_info.data[0], (char *)&event->data[event->file]);
                     }

                     sig[isig]->alloc_free_buf_info.euId = event->cpu;
                     sig[isig]->alloc_free_buf_info.reserved0 = 0;
                  }
                  break;

               case 6:
               case 11:
                  /* Basic event reports to source code debuggers */
                  if (rw.evact_lock_owner->debug_scope.type != MONITOR_SCOPE_VOID)
                  {
                     sig[isig] = alloc(sizeof (MonitorBasicErrorNotify), MONITOR_BASIC_ERROR_NOTIFY);
                     sig[isig]->basic_error_notify.pid = event->pid_from;

                     if (event->event_type == 6)
                        sig[isig]->basic_error_notify.exec = 1;
                     else
                        sig[isig]->basic_error_notify.exec = 0;

                     sig[isig]->basic_error_notify.error = event->ecode;
                     sig[isig]->basic_error_notify.extra = event->signo;
                     sig[isig]->basic_error_notify.reserved0 = 0;
                  }
                  /* Full event reports to Illuminator */
                  else
                  {
                     if (!idbuf_contains(&offspring->notify_pids, event->pid_from))
                     {
                        sig[isig] = alloc(sizeof (MonitorProcessInfo) + from_len, MONITOR_PROCESS_INTERCEPT_NOTIFY);

                        sig[isig]->process_info.pid = event->pid_from;
                        sig[isig]->process_info.bid = from_bid;
                        sig[isig]->process_info.sid = from_sid;
                        sig[isig]->process_info.type = event->ptype_from;
                        sig[isig]->process_info.entrypoint = from_ep;
                        sig[isig]->process_info.properties = from_props;
                        sig[isig]->process_info.reserved0 = 0;
                        strcpy(sig[isig]->process_info.name, from);

                        isig++;
                     }

                     sig[isig] = alloc(sizeof (MonitorErrorInfo) + file_len, MONITOR_ERROR_INTERCEPT_NOTIFY);

                     sig[isig]->error_info.reference = event->action_id;
                     sig[isig]->error_info.time.tick = event->event_tick;
                     sig[isig]->error_info.time.ntick = event->event_utick;
                     sig[isig]->error_info.time.sec = event->event_abssec;
                     sig[isig]->error_info.time.sectick = event->event_abstick;
                     sig[isig]->error_info.from = event->pid_from;
                     sig[isig]->error_info.exec = (event->event_type == 6);
                     sig[isig]->error_info.error = event->ecode;
                     sig[isig]->error_info.extra = event->signo;
                     sig[isig]->error_info.lineNumber = event->line;
                     sig[isig]->error_info.fileNameOffset = 0;
                     if (event->line)
                     {
                        strcpy((char *)sig[isig]->error_info.data, (char *)&event->data[event->file]);
                     }
                     sig[isig]->error_info.euId = event->cpu;
                     sig[isig]->error_info.reserved0 = 0;
                     sig[isig]->error_info.reserved1 = 0;
                  }
                  break;

#ifdef OSE_SMP
               case 14:        /* Bind */
                  if (offspring->handles_new_events)
                  {
                     if (!idbuf_contains(&offspring->notify_pids, event->pid_to))
                     {
                        sig[isig] = alloc(sizeof (MonitorProcessInfo) + to_len, MONITOR_PROCESS_INTERCEPT_NOTIFY);

                        sig[isig]->process_info.pid = event->pid_to;
                        sig[isig]->process_info.bid = to_bid;
                        sig[isig]->process_info.sid = to_sid;
                        sig[isig]->process_info.type = event->ptype_to;
                        sig[isig]->process_info.entrypoint = to_ep;
                        sig[isig]->process_info.properties = to_props;
                        sig[isig]->process_info.reserved0 = 0;
                        strcpy(sig[isig]->process_info.name, to);

                        isig++;
                     }

                     sig[isig] = alloc(sizeof (MonitorBindInfo) + file_len,
                                       MONITOR_BIND_INTERCEPT_NOTIFY);

                     sig[isig]->bind_info.reference = event->action_id;
                     sig[isig]->bind_info.time.tick = event->event_tick;
                     sig[isig]->bind_info.time.ntick = event->event_utick;
                     sig[isig]->bind_info.time.sec = event->event_abssec;
                     sig[isig]->bind_info.time.sectick = event->event_abstick;
                     sig[isig]->bind_info.reserved0 = 0;
                     sig[isig]->bind_info.bound = event->pid_to;
                     sig[isig]->bind_info.fromEuId = event->ecode;
                     sig[isig]->bind_info.toEuId = event->signo;
                     sig[isig]->bind_info.lineNumber = event->line;
                     sig[isig]->bind_info.fileNameOffset = 0;
                     if (event->line)
                     {
                        strcpy((char *)sig[isig]->bind_info.data, (char *)&event->data[event->file]);
                     }
                     sig[isig]->bind_info.reserved1 = 0;
                  }
                  break;
#endif /* OSE_SMP */

               default:
                  assert(0);
                  break;
            }
            if (sig[0])
            {
               send(&sig[0], offspring->pid);

               if (sig[1])
               {
                  send(&sig[1], offspring->pid);

                  if (sig[2])
                  {
                     send(&sig[2], offspring->pid);
                  }
               }
            }
            break;

         case 2:               /* Monitor */
         case 10:              /* Monitor long */
            if (rw.monitor_enabled)
            {
               isig = 0;
               sig[0] = sig[1] = sig[2] = NULL;

               switch (event->event_type)
               {
                  case 1:      /* Create */
                  case 2:      /* Kill */
                     /* Basic event reports to source code debuggers */
                     if (rw.evact_lock_owner->debug_scope.type != MONITOR_SCOPE_VOID)
                     {
                        const char *s, *b, *p;

                        sig[isig] = alloc(sizeof (MonitorBasicCreateKillNotify) + from_len,
                                          MONITOR_BASIC_CREATE_KILL_NOTIFY);
                        sig[isig]->basic_create_kill_notify.pid = event->pid_from;
                        sig[isig]->basic_create_kill_notify.bid = from_bid;
                        sig[isig]->basic_create_kill_notify.sid = from_sid;
                        sig[isig]->basic_create_kill_notify.type = event->ptype_from;

                        if (event->event_type == 2)
                           sig[isig]->basic_create_kill_notify.state = MONITOR_PROCESS_STATE_KILLED;
                        else
                           sig[isig]->basic_create_kill_notify.state = MONITOR_PROCESS_STATE_READY;

                        sig[isig]->basic_create_kill_notify.properties = from_props;
                        sig[isig]->basic_create_kill_notify.entrypoint = from_ep;
                        split_string(&s, &b, &p, from);
                        strcpy(sig[isig]->basic_create_kill_notify.name, p);
                     }
                     /* Full event reports to Illuminator */
                     else
                     {
                        if (!idbuf_contains(&offspring->notify_pids, event->pid_from))
                        {
                           sig[isig] = alloc(sizeof (MonitorProcessInfo) + from_len, MONITOR_PROCESS_NOTIFY);

                           sig[isig]->process_info.pid = event->pid_from;
                           sig[isig]->process_info.bid = from_bid;
                           sig[isig]->process_info.sid = from_sid;
                           sig[isig]->process_info.type = event->ptype_from;
                           sig[isig]->process_info.entrypoint = from_ep;
                           sig[isig]->process_info.properties = from_props;
                           sig[isig]->process_info.reserved0 = 0;
                           strcpy(sig[isig]->process_info.name, from);

                           isig++;
                        }

                        sig[isig] = alloc(sizeof (MonitorSwapCreateKillInfo) + file_len,
                                          (event->event_type == 1) ? MONITOR_CREATE_NOTIFY : MONITOR_KILL_NOTIFY);

                        sig[isig]->swap_create_kill_info.reference = event->action_id;
                        sig[isig]->swap_create_kill_info.time.tick = event->event_tick;
                        sig[isig]->swap_create_kill_info.time.ntick = event->event_utick;
                        sig[isig]->swap_create_kill_info.time.sec = event->event_abssec;
                        sig[isig]->swap_create_kill_info.time.sectick = event->event_abstick;
                        sig[isig]->swap_create_kill_info.from = 0;      /* TODO: Creator / Killer */
                        sig[isig]->swap_create_kill_info.to = event->pid_from;
                        sig[isig]->swap_create_kill_info.lineNumber = event->line;
                        sig[isig]->swap_create_kill_info.fileNameOffset = 0;
                        if (event->line)
                        {
                           strcpy((char *)sig[isig]->swap_create_kill_info.data, (char *)&event->data[event->file]);
                        }
                        sig[isig]->swap_create_kill_info.euId = event->cpu;
                        sig[isig]->swap_create_kill_info.reserved0 = 0;
                        sig[isig]->swap_create_kill_info.reserved1 = 0;
                     }
                     break;

                  case 4:      /* send */
                  case 5:      /* receive */
                     if (!idbuf_contains(&offspring->notify_pids, event->pid_from))
                     {
                        sig[isig] = alloc(sizeof (MonitorProcessInfo) + from_len, MONITOR_PROCESS_NOTIFY);

                        sig[isig]->process_info.pid = event->pid_from;
                        sig[isig]->process_info.bid = from_bid;
                        sig[isig]->process_info.sid = from_sid;
                        sig[isig]->process_info.type = event->ptype_from;
                        sig[isig]->process_info.entrypoint = from_ep;
                        sig[isig]->process_info.properties = from_props;
                        sig[isig]->process_info.reserved0 = 0;
                        strcpy(sig[isig]->process_info.name, from);

                        isig++;
                     }
                     if (!idbuf_contains(&offspring->notify_pids, event->pid_to))
                     {
                        sig[isig] = alloc(sizeof (MonitorProcessInfo) + to_len, MONITOR_PROCESS_NOTIFY);

                        sig[isig]->process_info.pid = event->pid_to;
                        sig[isig]->process_info.bid = to_bid;
                        sig[isig]->process_info.sid = to_sid;
                        sig[isig]->process_info.type = event->ptype_to;
                        sig[isig]->process_info.entrypoint = to_ep;
                        sig[isig]->process_info.properties = to_props;
                        sig[isig]->process_info.reserved0 = 0;
                        strcpy(sig[isig]->process_info.name, to);

                        isig++;
                     }

                     sig[isig] = alloc(sizeof (MonitorSendReceiveInfo) + file_len +
                                       event->sig_datasize,
                                       (event->event_type == 4) ? MONITOR_SEND_NOTIFY : MONITOR_RECEIVE_NOTIFY);

                     sig[isig]->send_receive_info.reference = event->action_id;
                     sig[isig]->send_receive_info.time.tick = event->event_tick;
                     sig[isig]->send_receive_info.time.ntick = event->event_utick;
                     sig[isig]->send_receive_info.time.sec = event->event_abssec;
                     sig[isig]->send_receive_info.time.sectick = event->event_abstick;
                     sig[isig]->send_receive_info.signalNumber = event->signo;
                     sig[isig]->send_receive_info.signalSender = event->pid_from;
                     sig[isig]->send_receive_info.signalAddressee = event->pid_to;
                     sig[isig]->send_receive_info.signalSize = event->sig_size;
                     sig[isig]->send_receive_info.signalAddress = event->sig_address;
                     sig[isig]->send_receive_info.signalId = 0;
                     sig[isig]->send_receive_info.signalDataOffset = 0;
                     sig[isig]->send_receive_info.signalDataCount = event->sig_datasize;
                     if (event->sig_datasize)
                     {
                        memcpy(sig[isig]->send_receive_info.data, &event->data[event->sig_data], event->sig_datasize);
                     }
                     sig[isig]->send_receive_info.lineNumber = event->line;
                     sig[isig]->send_receive_info.fileNameOffset = event->sig_datasize;
                     if (event->line)
                     {
                        strcpy((char *)&sig[isig]->send_receive_info.data[event->sig_datasize],
                        		(char *)&event->data[event->file]);
                     }
                     sig[isig]->send_receive_info.euId = event->cpu;
                     sig[isig]->send_receive_info.reserved0 = 0;
                     sig[isig]->send_receive_info.reserved1 = 0;
                     break;

#ifdef HAS_TIMEOUT_EVENT
		  case MONITOR_EVENT_TIMEOUT:  /* timeout in receive_w_tmo */
		     if (offspring->handles_new_events)
		     {
			if (!idbuf_contains(&offspring->notify_pids, event->pid_from))
			{
			   sig[isig] = alloc(sizeof (MonitorProcessInfo) + from_len, MONITOR_PROCESS_NOTIFY);

			   sig[isig]->process_info.pid = event->pid_from;
			   sig[isig]->process_info.bid = from_bid;
			   sig[isig]->process_info.sid = from_sid;
			   sig[isig]->process_info.type = event->ptype_from;
			   sig[isig]->process_info.entrypoint = from_ep;
			   sig[isig]->process_info.properties = from_props;
			   sig[isig]->process_info.reserved0 = 0;
			   strcpy(sig[isig]->process_info.name, from);

			   isig++;
			}

			sig[isig] = alloc(sizeof (MonitorTimeoutInfo) + file_len,
                                       MONITOR_TIMEOUT_NOTIFY);

			sig[isig]->timeout_info.reference = event->action_id;
			sig[isig]->timeout_info.time.tick = event->event_tick;
			sig[isig]->timeout_info.time.ntick = event->event_utick;
			sig[isig]->timeout_info.time.sec = event->event_abssec;
			sig[isig]->timeout_info.time.sectick = event->event_abstick;
			sig[isig]->timeout_info.signalAddressee = event->sig_from;
			sig[isig]->timeout_info.timeout = event->ecode;
			sig[isig]->timeout_info.tmoSource = 0;  /* A receive timeout */
			sig[isig]->timeout_info.lineNumber = event->line;
			sig[isig]->timeout_info.fileNameOffset = 0;
			if (event->line)
			{
			   assert(file_len > 0);
			   strcpy((char *)&sig[isig]->timeout_info.data[0], (char *)&event->data[event->file]);
			}

			sig[isig]->timeout_info.euId = event->cpu;
			sig[isig]->timeout_info.reserved0 = 0;
		     }
		     break;
#endif

                  case MONITOR_EVENT_USER: /* User events */
                     if (offspring->handles_new_events)
                     {
                        if (!idbuf_contains(&offspring->notify_pids, event->pid_from))
                        {
                           sig[isig] = alloc(sizeof (MonitorProcessInfo) + from_len, MONITOR_PROCESS_NOTIFY);

                           sig[isig]->process_info.pid = event->pid_from;
                           sig[isig]->process_info.bid = from_bid;
                           sig[isig]->process_info.sid = from_sid;
                           sig[isig]->process_info.type = event->ptype_from;
                           sig[isig]->process_info.entrypoint = from_ep;
                           sig[isig]->process_info.properties = from_props;
                           sig[isig]->process_info.reserved0 = 0;

                           strcpy(sig[isig]->process_info.name, from);

                           isig++;
                        }
                        sig[isig] = alloc(sizeof (MonitorUserInfo) + file_len +
                                          event->sig_datasize, MONITOR_USER_NOTIFY);

                        sig[isig]->user_info.reference = event->action_id;
                        sig[isig]->user_info.time.tick = event->event_tick;
                        sig[isig]->user_info.time.ntick = event->event_utick;
                        sig[isig]->user_info.time.sec = event->event_abssec;
                        sig[isig]->user_info.time.sectick = event->event_abstick;
                        sig[isig]->user_info.eventNumber = event->signo;
                        sig[isig]->user_info.eventDataOffset = 0;
                        sig[isig]->user_info.eventDataCount = event->sig_datasize;
                        sig[isig]->user_info.from = event->sig_from;
                        sig[isig]->user_info.eventSize = event->sig_size;

                        if (event->sig_datasize)
                        {
                           memcpy(sig[isig]->user_info.data, &event->data[event->sig_data], event->sig_datasize);
                        }
                        sig[isig]->user_info.lineNumber = event->line;
                        if (event->line)
                        {
                           assert(file_len > 0);
                           sig[isig]->user_info.fileNameOffset = event->sig_datasize;
                           strcpy((char *)&sig[isig]->user_info.data[event->sig_datasize], (char *)&event->data[event->file]);
                        }
                        else
                        {
                           sig[isig]->user_info.fileNameOffset = 0;
                        }
                        sig[isig]->user_info.euId = event->cpu;
                        sig[isig]->user_info.reserved0 = 0;
                     }
                     break;

                  case MONITOR_EVENT_ALLOC:  /* Alloc */
                     if (offspring->handles_new_events)
                     {
                        if (!idbuf_contains(&offspring->notify_pids, event->pid_from))
                        {
                           sig[isig] = alloc(sizeof (MonitorProcessInfo) + from_len, MONITOR_PROCESS_NOTIFY);

                           sig[isig]->process_info.pid = event->pid_from;
                           sig[isig]->process_info.bid = from_bid;
                           sig[isig]->process_info.sid = from_sid;
                           sig[isig]->process_info.type = event->ptype_from;
                           sig[isig]->process_info.entrypoint = from_ep;
                           sig[isig]->process_info.properties = from_props;
                           sig[isig]->process_info.reserved0 = 0;
                           strcpy(sig[isig]->process_info.name, from);

                           isig++;
                        }

                        sig[isig] = alloc(sizeof (MonitorAllocFreeBufInfo) + file_len,
                                          MONITOR_ALLOC_NOTIFY);


                        sig[isig]->alloc_free_buf_info.reference = event->action_id;
                        sig[isig]->alloc_free_buf_info.time.tick = event->event_tick;
                        sig[isig]->alloc_free_buf_info.time.ntick = event->event_utick;
                        sig[isig]->alloc_free_buf_info.time.sec = event->event_abssec;
                        sig[isig]->alloc_free_buf_info.time.sectick = event->event_abstick;
                        sig[isig]->alloc_free_buf_info.from = event->sig_from;
                        sig[isig]->alloc_free_buf_info.signalNumber = event->signo;
                        sig[isig]->alloc_free_buf_info.signalSize = event->sig_size;
                        sig[isig]->alloc_free_buf_info.signalAddress = event->sig_address;
                        sig[isig]->alloc_free_buf_info.signalId = 0;

                        sig[isig]->alloc_free_buf_info.euId = event->cpu;
                        sig[isig]->alloc_free_buf_info.reserved0 = 0;
                        sig[isig]->alloc_free_buf_info.lineNumber = event->line;
                        sig[isig]->alloc_free_buf_info.fileNameOffset = 0;
                        if (event->line)
                        {
                           assert(file_len > 0);
                           strcpy((char *)&sig[isig]->alloc_free_buf_info.data[0], (char *)&event->data[event->file]);
                        }
                     }
                     break;

                  case MONITOR_EVENT_FREE_BUF:  /* Free buf */
                     if (offspring->handles_new_events)
                     {
                        if (!idbuf_contains(&offspring->notify_pids, event->pid_from))
                        {
                           sig[isig] = alloc(sizeof (MonitorProcessInfo) + from_len, MONITOR_PROCESS_NOTIFY);

                           sig[isig]->process_info.pid = event->pid_from;
                           sig[isig]->process_info.bid = from_bid;
                           sig[isig]->process_info.sid = from_sid;
                           sig[isig]->process_info.type = event->ptype_from;
                           sig[isig]->process_info.entrypoint = from_ep;
                           sig[isig]->process_info.properties = from_props;
                           sig[isig]->process_info.reserved0 = 0;
                           strcpy(sig[isig]->process_info.name, from);

                           isig++;
                        }

                        sig[isig] = alloc(sizeof (MonitorAllocFreeBufInfo) + file_len,
                                          MONITOR_FREE_BUF_NOTIFY);


                        sig[isig]->alloc_free_buf_info.reference = event->action_id;
                        sig[isig]->alloc_free_buf_info.time.tick = event->event_tick;
                        sig[isig]->alloc_free_buf_info.time.ntick = event->event_utick;
                        sig[isig]->alloc_free_buf_info.time.sec = event->event_abssec;
                        sig[isig]->alloc_free_buf_info.time.sectick = event->event_abstick;
                        sig[isig]->alloc_free_buf_info.from = event->sig_from;
                        sig[isig]->alloc_free_buf_info.signalNumber = event->signo;
                        sig[isig]->alloc_free_buf_info.signalSize = event->sig_size;
                        sig[isig]->alloc_free_buf_info.signalAddress = event->sig_address;
                        sig[isig]->alloc_free_buf_info.signalId = 0;

                        sig[isig]->alloc_free_buf_info.lineNumber = event->line;
                        if (event->line)
                        {
                           assert(file_len > 0);
                           sig[isig]->alloc_free_buf_info.fileNameOffset = event->sig_datasize;
                           strcpy((char *)&sig[isig]->alloc_free_buf_info.data[event->sig_datasize],
                        		   (char *)&event->data[event->file]);
                        }
                        else
                        {
                           sig[isig]->alloc_free_buf_info.fileNameOffset = 0;
                        }
                        sig[isig]->alloc_free_buf_info.euId = event->cpu;
                        sig[isig]->alloc_free_buf_info.reserved0 = 0;
                     }
                     break;

                  case 6:
                  case 11:
                     /* Basic event reports to source code debuggers */
                     if (rw.evact_lock_owner->debug_scope.type != MONITOR_SCOPE_VOID)
                     {
                        sig[isig] = alloc(sizeof (MonitorBasicErrorNotify), MONITOR_BASIC_ERROR_NOTIFY);
                        sig[isig]->basic_error_notify.pid = event->pid_from;

                        if (event->event_type == 6)
                           sig[isig]->basic_error_notify.exec = 1;
                        else
                           sig[isig]->basic_error_notify.exec = 0;

                        sig[isig]->basic_error_notify.error = event->ecode;
                        sig[isig]->basic_error_notify.extra = event->signo;
                        sig[isig]->basic_error_notify.reserved0 = 0;
                     }
                     else
                     {
                        if (!idbuf_contains(&offspring->notify_pids, event->pid_from))
                        {
                           sig[isig] = alloc(sizeof (MonitorProcessInfo) + from_len, MONITOR_PROCESS_NOTIFY);

                           sig[isig]->process_info.pid = event->pid_from;
                           sig[isig]->process_info.bid = from_bid;
                           sig[isig]->process_info.sid = from_sid;
                           sig[isig]->process_info.type = event->ptype_from;
                           sig[isig]->process_info.entrypoint = from_ep;
                           sig[isig]->process_info.properties = from_props;
                           sig[isig]->process_info.reserved0 = 0;
                           strcpy(sig[isig]->process_info.name, from);

                           isig++;
                        }

                        sig[isig] = alloc(sizeof (MonitorErrorInfo) + file_len, MONITOR_ERROR_NOTIFY);

                        sig[isig]->error_info.reference = event->action_id;
                        sig[isig]->error_info.time.tick = event->event_tick;
                        sig[isig]->error_info.time.ntick = event->event_utick;
                        sig[isig]->error_info.time.sec = event->event_abssec;
                        sig[isig]->error_info.time.sectick = event->event_abstick;
                        sig[isig]->error_info.from = event->pid_from;
                        sig[isig]->error_info.exec = (event->event_type == 6);
                        sig[isig]->error_info.error = event->ecode;
                        sig[isig]->error_info.extra = event->signo;
                        sig[isig]->error_info.lineNumber = event->line;
                        sig[isig]->error_info.fileNameOffset = 0;
                        if (event->line)
                        {
                           strcpy((char *)sig[isig]->error_info.data, (char *)&event->data[event->file]);
                        }
                        sig[isig]->error_info.euId = event->cpu;
                        sig[isig]->error_info.reserved0 = 0;
                        sig[isig]->error_info.reserved1 = 0;
                     }
                     break;

#ifdef OSE_SMP
                  case 14:        /* Bind */
                     if (offspring->handles_new_events)
                     {
                        if (!idbuf_contains(&offspring->notify_pids, event->pid_to))
                        {
                           sig[isig] = alloc(sizeof (MonitorProcessInfo) + to_len, MONITOR_PROCESS_NOTIFY);

                           sig[isig]->process_info.pid = event->pid_to;
                           sig[isig]->process_info.bid = to_bid;
                           sig[isig]->process_info.sid = to_sid;
                           sig[isig]->process_info.type = event->ptype_to;
                           sig[isig]->process_info.entrypoint = to_ep;
                           sig[isig]->process_info.properties = to_props;
                           sig[isig]->process_info.reserved0 = 0;
                           strcpy(sig[isig]->process_info.name, to);

                           isig++;
                        }

                        sig[isig] = alloc(sizeof (MonitorBindInfo) + file_len,
                                          MONITOR_BIND_NOTIFY);

                        sig[isig]->bind_info.reference = event->action_id;
                        sig[isig]->bind_info.time.tick = event->event_tick;
                        sig[isig]->bind_info.time.ntick = event->event_utick;
                        sig[isig]->bind_info.time.sec = event->event_abssec;
                        sig[isig]->bind_info.time.sectick = event->event_abstick;
                        sig[isig]->bind_info.reserved0 = 0;
                        sig[isig]->bind_info.bound = event->pid_to;
                        sig[isig]->bind_info.fromEuId = event->ecode;
                        sig[isig]->bind_info.toEuId = event->signo;
                        sig[isig]->bind_info.lineNumber = event->line;
                        sig[isig]->bind_info.fileNameOffset = 0;
                        if (event->line)
                        {
                           strcpy((char *)sig[isig]->bind_info.data, (char *)&event->data[event->file]);
                        }
                        sig[isig]->bind_info.reserved1 = 0;
                     }
                     break;
#endif /* OSE_SMP */

                  default:
                     assert(0);
                     break;
               }                /* switch(event->event_type) */

               if (sig[0])
               {
                  uint32_t        i;
                  uint32_t        n = (sig[0] != 0) + (sig[1] != 0) + (sig[2] != 0);

                  /*
                   * We cheat a little with the flow control here, since we
                   * do not count sig[0..2] as separate signals. However the
                   * client has no counters of its own and only responds to
                   * our flow control requests, so we should be OK.
                   */
                  if (offspring->flow_ctrl_nblock > 0)
                  {
                     offspring->flow_ctrl_nblock--;

                     if (offspring->flow_ctrl_nblock == rw.flow_ctrl_nblock / 2)
                     {
                        union SIGNAL   *fsig = alloc(sizeof (MonitorFlowControl),
                                                     MONITOR_NON_BLOCKING_FLOW_CONTROL_REQUEST);

                        send_w_s(&fsig, offspring->pid, offspring->client_);
                     }
                     for (i = 0; i < n; i++)
                     {
                        send_w_s(&sig[i], offspring->pid, offspring->client_);
                     }
                  }
                  else
                  {
                     for (i = 0; i < n; i++)
                     {
                        free_buf(&sig[i]);
                     }
                  }
               }

               SPIN_LOCK(&rw.event_notify_lock);
               rw.monitor_hook_flow++;
               /*
                * Get out of overflow and notify the client, but not until
                * we have sent all the event reports waiting in our inqueue
                */
               if (rw.monitor_overflow && rw.monitor_hook_flow == rw.flow_ctrl_nblock)
               {
                  union SIGNAL   *fsig;

                  SPIN_UNLOCK(&rw.event_notify_lock);

                  /* Basic event reports to source code debuggers */
                  if (offspring->debug_scope.type != MONITOR_SCOPE_VOID)
                  {
                     fsig = alloc(sizeof (MonitorBasicLossNotify), MONITOR_BASIC_LOSS_NOTIFY);
                     send_w_s(&fsig, offspring->pid, offspring->client_);
                  }
                  else
                  {
                     fsig = alloc(sizeof (MonitorLossNotify), MONITOR_LOSS_NOTIFY);

#ifdef HAS_LOSS_REFERENCE
                     fsig->loss_notify.reference = 0;
#else
                     fsig->loss_notify.reserved0 = 0;
#endif
                     fsig->loss_notify.time.tick = get_systime((OSTICK *) & fsig->loss_notify.time.ntick);
                     BKL_LOCK();
                     fsig->loss_notify.time.sec = rw.abs_sec;
                     fsig->loss_notify.time.sectick = rw.abs_tick;
                     BKL_UNLOCK();
                     fsig->loss_notify.lostSize = 0;
                     send_w_s(&fsig, offspring->pid, offspring->client_);
                  }
                  rw.monitor_overflow = 0;
               }
               else
               {
                  SPIN_UNLOCK(&rw.event_notify_lock);
               }
            }                   /* if (monitor_enabled) */
            break;

         default:
            assert(0);
            break;
      }                         /* switch(event->action) */
   }                            /* if (offspring != NULL) */
}

/*
 * Update trig variables according to new actionpoint.
 */
static void
set_user_event_trig_var(uint32_t start, uint32_t end)
{
   blocks_user_events *user_posts;

   assert(start <= end);

   for(user_posts = list_head(rw.user_events); user_posts != NULL ;
       user_posts = list_next(rw.user_events, &user_posts->link))
   {
      uint8_t i;
      uint32_t error_val;
      static uint32_t trigValue = 1;

      LOCK();
      SMP_CODE(SPIN_LOCK(&rw.user_event_lock);)

      for(i=0;( i < EVENTS_PER_BLOCK && ( 0 != user_posts->user_event_list[i].trig)); i++)
      {
         if (start <= user_posts->user_event_list[i].event_no &&
            end >= user_posts->user_event_list[i].event_no)
         {
            volatile OSADDRESS addr = (OSADDRESS)user_posts->user_event_list[i].trig;

            SMP_CODE(SPIN_UNLOCK(&rw.user_event_lock);)
            UNLOCK();
            error_val = set_mem(user_posts->sid, &trigValue,
                                addr,
                                sizeof(trigValue));
            if (error_val)
            {
               ramlog_printf("rmm: set_mem return %u sid: 0x%x addr: 0x%x\n",
                     error_val, user_posts->sid, (OSADDRESS)user_posts->user_event_list[i].trig);
            }
            /* if no error! */
            user_posts->user_event_list[i].trig_value = trigValue;
            LOCK();
            SMP_CODE(SPIN_LOCK(&rw.user_event_lock);)
         }
      }

      SMP_CODE(SPIN_UNLOCK(&rw.user_event_lock);)
      UNLOCK();
   }
}

void
create_event(SIGSELECT eventno, OSADDRESS *trig)
{
   if (trig == NULL)
   {
      zzos_error(0, OSE_CREATE_EVENT | OSE_EBAD_PARAMETER, (OSERRCODE)trig);
      return;
   }
   {
      struct ldefine_action *action;
      OSADDRESS trigValue = 0;
      uint32_t error_val;

      PROCESS sid;
      OSBOOLEAN sid_exist = FALSE;
      blocks_user_events *user_posts;
      SMP_CODE(cpuid_t cpu;)

      sid = get_segid(current_process());

      LOCK();
      SMP_CODE(SPIN_LOCK(&rw.user_event_lock);)

      if (rw.user_events != NULL)
      {
         for(user_posts = list_head(rw.user_events);
             user_posts != NULL;
             user_posts = list_next(rw.user_events, &user_posts->link))
         {
            if (user_posts->sid == sid)
            {
               uint8_t i;
               sid_exist = TRUE;

               for(i = 0; i < EVENTS_PER_BLOCK; i++)
               {
                  /* Trig variable already registered */
                  if (user_posts->user_event_list[i].trig == trig)
                  {
                     SMP_CODE(SPIN_UNLOCK(&rw.user_event_lock);)
                     UNLOCK();
                     return;
                  }
                  /* Unused slot which we can use */
                  if (user_posts->user_event_list[i].trig == 0)
                  {
                     user_posts->user_event_list[i].event_no = eventno;
                     user_posts->user_event_list[i].trig_value = trigValue;
                     user_posts->user_event_list[i].trig = trig;

                     SMP_CODE(SPIN_UNLOCK(&rw.user_event_lock);)
                     UNLOCK();
                     goto enable_trig_monitor;
                  }
               }
            }
         }
      }
      SMP_CODE(SPIN_UNLOCK(&rw.user_event_lock);)
      UNLOCK();

      /* No empty slot found, create new user post struct */
      {
         union SIGNAL *sig;
         static const SIGSELECT sel_reply[] = { 1, MONITOR_CREATE_EVENT_STRUCT_REPLY};

         sig = alloc(sizeof(MonitorCreateEventStructRequest),
                  MONITOR_CREATE_EVENT_STRUCT_REQUEST);

         sig->create_event_struct_request.sid = sid;
         sig->create_event_struct_request.event_id = eventno;
         sig->create_event_struct_request.trig = trig;
         sig->create_event_struct_request.trig_value = trigValue;
         sig->create_event_struct_request.sid_exist = sid_exist;

         send(&sig, rw.zzose_rmm_);

         while(NIL == (sig = receive_from((OSTIME)~0,sel_reply ,rw.zzose_rmm_)));
         free_buf(&sig);
      }
 
 enable_trig_monitor:

      LOCK();
      SMP_CODE(cpu = ose_cpu_id();)
      /* Go through the list of user event actionpoints
       * and see if we should enable the trig variable. */
      if (NULL != rw.actions SMP_CODE([cpu]).action_list_uevent)
      {
         for (action = rw.actions SMP_CODE([cpu]).action_list_uevent; action; action = action->next)
         {
            assert(validate_evact(action));

            /* Check for event number match. */
            if ((action->signo <= eventno) && (action->signo_end >= eventno))
            {
               trigValue = 1;
               break;
            }
         }
      }
      UNLOCK();

      error_val = set_mem(sid, &trigValue, (OSADDRESS)trig, sizeof(trigValue));

      if (error_val)
      {
         /* FIXME report error and get an error code from ose_err.h */
         ramlog_printf("rmm: ERROR in create_event set_mem (%d)\n", error_val);
      }
   }
}

static void
add_new_user_event_struct(MonitorCreateEventStructRequest *req, PROCESS sig_sender)
{

   blocks_user_events *new_block;

   new_block = (blocks_user_events *)list_init_link(
      (void *)alloc(sizeof (blocks_user_events),0));

   memset(&new_block->user_event_list, 0,
          sizeof(user_event_post) * EVENTS_PER_BLOCK);

   new_block->sid = req->sid;
   new_block->user_event_list[0].event_no = req->event_id;
   new_block->user_event_list[0].trig = req->trig;
   new_block->user_event_list[0].trig_value = req->trig_value;

   LOCK();
   SMP_CODE(SPIN_LOCK(&rw.user_event_lock);)
   list_insert_head(rw.user_events, new_block);
   SMP_CODE(SPIN_UNLOCK(&rw.user_event_lock);)
   UNLOCK();

   if (!req->sid_exist && req->sid)
   {
      union SIGNAL *sig_attach;
      sig_attach = alloc(sizeof (UserEventAttachSig), USER_EVENT_ATTACH_SIG);

      sig_attach->user_event_attach_sig.sid = req->sid;
      attach(&sig_attach, req->sid);
   }
   {
      /* Reply to create_event */
      union SIGNAL *sig;
      sig = alloc(0, MONITOR_CREATE_EVENT_STRUCT_REPLY);
      send(&sig, sig_sender);
   }
}

/*
 * Implementation of the system call signal_event.
 */
void
zzsignal_event(SIGSELECT event_id, const void *data, OSBUFSIZE size)
{
/*
** The actions flag field is defined as follows:
** D0 = 1 => Catch the event.                   Mask = 0x0003
** D1 = 1 => Ignore the catch bit.
** D2 = 1 => Monitor the event.                 Mask = 0x000c
** D3 = 1 => Ignore the monitor bit.
** D4 = 1 => Trace the event.                   Mask = 0x0030
** D5 = 1 => Ignore the trace bit.
** D6 = 1 => Turn on the trace.                 Mask = 0x00c0
** D7 = 1 => Ignore the trace on bit.
** D8 = 1 => Turn off the trace.                Mask = 0x0300
** D9 = 1 => Ignore the trace off bit.
** D10 = 1 => Turn on the watch timer.          Mask = 0x0c00
** D11 = 1 => Ignore the watch on bit.
** D12 = 1 => Turn off the watch timer.         Mask = 0x3000
** D13 = 1 => Ignore the watch off bit.
** D14 = 1 => Goto a new event sequence number. Mask = 0xc000
** D15 = 1 => Ignore the goto bit.
*/
   LOCKSTATE       msr;
   uint32_t actions = 0;
   struct ldefine_action *action;

   U8        oldevent_seq_no = rw.action_state;
   U8        newevent_seq_no = 0;

   U8        monitor_long_flag = 2;

   uint32_t  interceptionpoint = 0;
   uint32_t  notificationpoint = 0;
   uint32_t  trace_with_data = 0;
   int       cpu_id;

   struct debug_pcb *from = zzget_debugpcb(current_process());

   /*   uint16_t event_no = (uint16_t)TRACE_USER_EVENT;*/
   struct ldefine_action *action_list;

   if (data == NULL && size != 0)
   {
      /* data == NULL but a size specified. */
      zzos_error(0, OSE_EBAD_PARAMETER, size);
      size = 0;
   }

   /* Skip any event that involves system process */
   if (from->debug_flags & rw.sys_process)
      return;

   /* Skip events that don't involve any process in the display scope. */
   if (!(from->debug_flags & IN_DEBUGGER_SCOPE))
      return;

   LOCK_SAVE(msr);
   cpu_id = SMP_CODE_YES_NO(zzkrn_cpu_id(), 0);
   action_list = rw.actions SMP_CODE([cpu_id]).action_list_uevent;

   for (action = action_list; action; action = action->next)
   {
      assert(validate_evact(action));

      /* Check for event sequence match. */
      if (action->event_seq_no && (action->event_seq_no != oldevent_seq_no))
         continue;

      /* Check for sender match. */
      if (!zzrmm_match_proc(from, (struct procnamelist *) action->align))
         continue;

      /* Check for event number match. */
      if ((action->signo > event_id) || (action->signo_end < event_id))
         continue;

      if (unlikely(!match_data_range((const uint8_t *)data, size, action)))
      {
         continue;
      }

      /* Check for an active event counter. */
      if (unlikely(action->orig_event_count))
      {
         /* Update the event counter and ignore the event if the limit is not reached. */
         SPIN_LOCK(&action->SMP_CODE(orig_action->)action_lock);
         if (--(action->SMP_CODE(orig_action->)event_count))
         {
            SPIN_UNLOCK(&action->SMP_CODE(orig_action->)action_lock);
            continue;
         }
         else
         {
            action->SMP_CODE(orig_action->)event_count = action->orig_event_count;
            SPIN_UNLOCK(&action->SMP_CODE(orig_action->)action_lock);
         }
      }

      /* Decide which action to take and update action flags. */
      {
         U32 newaction = 0;

         switch (action->action & 0xf)
         {
            case 3:
               /* Trace */
               newaction = 0x0010;
               trace_with_data |= action->action & TRACE_WITH_DATA;
               break;
            case 1:
               /* Catch */
               interceptionpoint = action->action_id;
               newaction = 0x0001;
               break;
            case 10:
               /* Monitor -long */
               notificationpoint = action->action_id;
               monitor_long_flag = 10;
               newaction = 0x0004;
               break;
            case 2:
               /* Monitor */
               notificationpoint = action->action_id;
               newaction = 0x0004;
               break;
            case 4:
               /* Trace on */
               newaction = 0x0040;
               break;
            case 5:
               /* Trace off. */
               newaction = 0x0100;
               break;
            case 7:
               /* Watch on */
               newaction = 0x0400;
               break;
            case 8:
               /* watch off */
               newaction =0x1000;
               break;
            case 9:
               /* Goto */
               newaction = 0x4000;
               /* Save the seq number to be for later. */
               newevent_seq_no = action->goto_list;
               break;
            case 12:
               /* User function */
               newaction = 0x10000;
               break;
            case 14:
               /* Enable hw trace */
               newaction = 0x100000;
               break;
            case 15:
               /* Disable hw trace */
               newaction = 0x400000;
               break;
            default:
               break;
         }
         /* Negate the action if required. */
         if (action->negated)
            newaction <<=1;
         /* Log the action for later. */
         actions |= newaction;
      }
   }
   LOCK_RESTORE(msr);
   if (!actions)
   {
      return;
   }

   /* Call user specified function, if it is defined in ose_reg_user_func() */
   if (rw.user_action_function != NULL && ((actions & 0x30000) == 0x10000))
   {
      MonitorUserInfo info;

      info.sigNo = MONITOR_GET_USER_TRACE_REPLY;
      info.reference = notificationpoint;
      BKL_LOCK_SAVE(msr);
      info.time.sec = rw.abs_sec;
      info.time.sectick = rw.abs_tick;
      info.time.tick = get_systime((OSTICK *) &info.time.ntick);
      BKL_LOCK_RESTORE(msr);
      info.from = (U32) from->pid;
      info.eventNumber = event_id;
      info.eventSize = size;
      info.lineNumber = 0;
      info.fileNameOffset = 0;
      info.eventDataOffset = 0;
      info.eventDataCount = 0;
      info.euId = cpu_id;
      info.reserved0 = 0;
      info.data[0] = 0;

      BKL_LOCK_SAVE(msr);
      send_proc_info(from);

      rw.user_action_function((union SIGNAL *)&info,
                           offsetof(MonitorUserInfo, data),
                           (void *)(OSADDRESS)data,
                           size);
      BKL_LOCK_RESTORE(msr);
   }

   /* Turn on the trace if required. */
   if ((actions & 0x00c0) == 0x0040)
   {
      BKL_LOCK_SAVE(msr);
      rw.precious->trace_enabled = rw.trace_enabled = (OSBOOLEAN)1;
      ose_trace_enable(rw.trace);
      BKL_LOCK_RESTORE(msr);
   }

   /* Turn on the hardware trace if required. */
   if ((actions & 0x300000) == 0x100000)
   {
      (*rw.hw_trace[0])();
   }

   /* Goto a new event state if required. */
   if ((actions & 0xc000) == 0x4000)
   {
      BKL_LOCK_SAVE(msr);
      rw.precious->action_state = rw.action_state = newevent_seq_no;
      BKL_LOCK_RESTORE(msr);
   }

   /* Trace the event if required */
   if ((actions & 0x0030) == 0x0010)
   {
      OSBUFSIZE       esize;
      struct trace_user_event *ue;
      struct trace_procname *proc_name;

      if (trace_with_data)
      {
         /* Trim the size if too large. */
         esize = (size > rw.max_trace_size) ? rw.max_trace_size : size;
      }
      else
      {
         esize = 0;
      }
      ue = (struct trace_user_event *)
         trace_alloc_from(sizeof (struct trace_user_event) + esize,
                          from, &proc_name);
      if (NULL != ue)
      {
         ue->event_no = TRACE_USER_EVENT;
         ue->event_size = esize;
         ue->requested_size = size;
         ue->abstime = get_trace_abstime();;
         ue->process = proc_name;
         ue->event_id = event_id;
         ue->cpu = SMP_CODE_YES_NO(zzkrn_cpu_id(), 0);

         if (data != NULL && trace_with_data)
         {
            memcpy(ue->data, data, esize);
         }

         ose_trace_commit_inline(ue);
      }
   }

   /* Turn off the trace if required */
   if ((actions & 0x0300) == 0x0100)
   {
      rw.precious->trace_enabled = rw.trace_enabled = (OSBOOLEAN) 0;
      ose_trace_disable(rw.trace);
   }

   /* Turn off the hardware trace if required. */
   if ((actions & 0xc00000) == 0x400000)
   {
      (*rw.hw_trace[1])();
   }


   /* Monitor the event if required */
   if ((actions & 0x000c) == 0x0004)
   {
      union SIGNAL *txsig;

      /* Monitor condition was true. Send to ldm if there is space in the mon queue. */
      /* Check that monitoring is enabled and that there is space in the monitor queue. */

      SPIN_LOCK_SAVE(&rw.event_notify_lock, msr);
      if ((!rw.monitor_enabled) || (rw.monitor_overflow))
      {
        if (rw.monitor_overflow)
        {
	   rw.monitor_overflow++;
        }
        SPIN_LOCK_RESTORE(&rw.event_notify_lock, msr);
        goto cant_monitor;
      }
      SPIN_LOCK_RESTORE(&rw.event_notify_lock, msr);

      /* Allocate a monitor buffer */
      set_suid(1);
      if (monitor_long_flag == 10)
      {
        txsig = alloc(sizeof (struct event_report_verbose) + ((size > rw.max_monitor_size) ? rw.max_monitor_size : size), MONITOR_REPORT);
        txsig->event_report_verbose.sig_datasize = (U16) ((size > rw.max_monitor_size) ? rw.max_monitor_size : size);
      }
      else
      {
        txsig = alloc(sizeof (struct event_report_verbose) + 0, MONITOR_REPORT);
        txsig ->event_report_verbose.sig_datasize = (U16) 0;
      }

      BKL_LOCK_SAVE(msr);
      txsig->event_report_verbose.event_abssec = rw.abs_sec;
      txsig->event_report_verbose.event_abstick = rw.abs_tick;
      txsig->event_report_verbose.event_absutick = rw.abs_usec;
      txsig->event_report_verbose.event_tick = get_systime((OSTICK *) & txsig->event_report_verbose.event_utick);
      BKL_LOCK_RESTORE(msr);
      /* Look out with the parameter type to get_systime()! */
      txsig->event_report_verbose.sig_from = (U32) from->pid;
      txsig->event_report_verbose.sig_to = 0; /* No "to" in user event! */
      txsig->event_report_verbose.ecode = 0;
      txsig->event_report_verbose.signo = event_id;
      txsig->event_report_verbose.cpu = SMP_CODE_YES_NO(zzkrn_cpu_id(), 0);
      txsig->event_report_verbose.sig_data = 0;

      txsig->event_report_verbose.pid_from = txsig->event_report_verbose.sig_from;
      txsig->event_report_verbose.pid_to = 0;
      txsig->event_report_verbose.ptype_from = make_ptype(from);
      txsig->event_report_verbose.ptype_to = 0; /* No "to" in user event */
      txsig->event_report_verbose.line = (U32) 0;
      txsig->event_report_verbose.file = (U32) 0;
      txsig->event_report_verbose.action_id = notificationpoint;

      txsig->event_report_verbose.event_type = MONITOR_EVENT_USER;
      txsig->event_report_verbose.sig_address = 0;

      txsig->event_report_verbose.action = monitor_long_flag;
      txsig->event_report_verbose.sig_size = size; /*size of requested log size */
      memcpy((void *) txsig->event_report_verbose.data,
             (const void *) data, (size_t) txsig->event_report_verbose.sig_datasize);

      SPIN_LOCK_SAVE(&rw.event_notify_lock, msr);
      /* Check again that monitoring is enabled and no overflow present. */
      assert(rw.monitor_hook_flow >= 0);
      if ((rw.monitor_enabled) && (0 == rw.monitor_hook_flow))
      {
         rw.monitor_overflow++;
      }
      if ((!rw.monitor_enabled) || (0 == rw.monitor_hook_flow))
      {
         SPIN_LOCK_RESTORE(&rw.event_notify_lock, msr);
         /* Free the buffer and abort if disabled. */
         free_buf(&txsig);
         set_suid(0);
         goto cant_monitor;
      }
      /* Take a flow control token . */
      rw.monitor_hook_flow--;
      /* Enable interrupts */
      SPIN_LOCK_RESTORE(&rw.event_notify_lock, msr);

      /* Send an event_report_verbose to the ldm. */
      zzsys_send_w_s(&txsig, rw.zzose_rmm_, rw.zzose_rmm_);
      set_suid(0);
   }

  cant_monitor:
   /* Catch the event if required. */

   if ((actions & 0x0003) == 0x0001)
   {
      /* Catch condition was true. Catch it if the scope is not stopped. */
      /* Disable interrupts */
      BKL_LOCK_SAVE(msr);
      /* Check that the system is not already caught. */
      if (rw.catch_enable)
      {
         union SIGNAL   *txsig;

         /* Disable catcher. */
         txsig = rw.catch_enable;
         rw.catch_enable = NULL;
         /* Store process id for the rmm to access. */
         rw.caught_pid = current_process();
         txsig->event_report_verbose.event_abssec = rw.abs_sec;
         txsig->event_report_verbose.event_abstick = rw.abs_tick;
         txsig->event_report_verbose.event_absutick = rw.abs_usec;
         txsig->event_report_verbose.event_tick = get_systime((OSTICK *) & txsig->event_report_verbose.event_utick);
         /* Enable interrupts. */
         BKL_LOCK_RESTORE(msr);
         /* Set up the event_report_verbose signal. Replace process names with id's. */
         txsig->event_report_verbose.sig_no = CATCH_REPORT;
         txsig->event_report_verbose.sig_from = (U32) from->pid;
         txsig->event_report_verbose.sig_to = (U32) 0; /* No "to" in user event */
         txsig->event_report_verbose.ecode = 0;
         txsig->event_report_verbose.signo = event_id;
         txsig->event_report_verbose.cpu = SMP_CODE_YES_NO(zzkrn_cpu_id(), 0);
         txsig->event_report_verbose.sig_data = 0;
         txsig->event_report_verbose.sig_datasize = (U16) 0;
         txsig->event_report_verbose.event_type = MONITOR_EVENT_USER;
         txsig->event_report_verbose.action = 1;        /* Catch action. */
         txsig->event_report_verbose.sig_address = 0;
         txsig->event_report_verbose.sig_size = 0;

         txsig->event_report_verbose.pid_from = txsig->event_report_verbose.sig_from;
         txsig->event_report_verbose.pid_to = 0;
         txsig->event_report_verbose.ptype_from = make_ptype(from);
         txsig->event_report_verbose.ptype_to = 0;
         txsig->event_report_verbose.line = (U32) 0;
         txsig->event_report_verbose.file = (U32) 0;
         txsig->event_report_verbose.action_id = interceptionpoint;

         /* Send an event_report to the ldm. */
         set_suid(1);
         restore(txsig);
         zzsys_send_w_s(&txsig, current_process(), rw.zzose_rmm_);
         set_suid(0);

         /* Intercept the breakpoint scope. */
         intercept_evact_scope();
      }
      else
      {
         BKL_LOCK_RESTORE(msr);
      }
   }
   return;
}

/* Get the event number range tag */
static void
get_actionpoint_ue(MonitorEventActionpointInfo *req,
                   uint32_t *eventno0, uint32_t *eventno1)
{
   MonitorTag     *t = req->tags;
   MonitorTag     *t1 = req->tags + req->tagsCount;

   while (t < t1)
   {
      if (t->tag.type == MONITOR_USER_FILTER_TAG_NUMBER)
      {
         *eventno0 = t[1].num;
         *eventno1 = t[2].num;
         return;
      }
      t += t->tag.count;
   }
   *eventno0 = 0;
   *eventno1 = 0;
}

static int
set_ue_actionpoint(Offspring *offspring, union SIGNAL *sig)
{
   MonitorEventActionpointInfo *req = &sig->event_actionpoint_info;

   assert(req);
   assert(offspring);

   if (validate_request(offspring, sig))
   {
      uint32_t   ignore;
      uint32_t   fromlen;
      uint32_t   eventno0, eventno1;

      get_actionpoint_ignore(req, &ignore);
      get_actionpoint_ue(req, &eventno0, &eventno1);

      fromlen = scope_to_pattern_len(&req->from, req->tags,req->tagsCount, 0);

      if (fromlen)
      {
         struct ldefine_action *action;
         uint32_t   actionlen =
            sizeof (struct ldefine_action) + fromlen + 1;

         action = alloc_action(actionlen);

         if (action)
         {
            assert(req->state <= 255);
            action->event_seq_no = req->state;

            /* Check whenever user event data should be included */
            if ((req->action == MONITOR_ACTION_NOTIFY) &&
                     (req->parameter & MONITOR_EVENT_ACTIONPOINT_PARAMETER_DATA))
               action->action = 10;
            else
             {
               action->action = req->action;
               if ((action->action == MONITOR_ACTION_TRACE) &&
                   (req->parameter & MONITOR_EVENT_ACTIONPOINT_PARAMETER_DATA))
               {
                  action->action |= TRACE_WITH_DATA;
               }
             }

            if (get_actionpoint_data_range(req, action))
            {
               action->negated = (uint8_t)req->not;

               /* Set the state (or sequence) of the event action point */
               if (MONITOR_ACTION_SET_STATE == req->action)
                  action->goto_list = (uint8_t)req->parameter;
               else
                  action->goto_list = 0;

               action->is_basic_event = 0;
               action->signo = eventno0;
               action->signo_end = eventno1 - 1;
               assert(ignore < 65536);
               action->orig_event_count = ignore;
               action->ecode = 0; /* not valid */
               action->event_type = MONITOR_EVENT_USER;
               action->sig_from = 0;
               action->sig_to = 0; /* not used */

               scope_to_pattern(&req->from, req->tags, req->tagsCount, 0,
                                action->names);
               /* There is no to process for user event */

               set_user_event_trig_var(eventno0, (eventno1 - 1));
               set_event_actionpoint(offspring, action, &rw.actions SMP_CODE([0]).action_list_uevent);
            }
            else
            {
               set_event_actionpoint_failed(offspring,MONITOR_STATUS_REQUEST_INVALID);
            }            
         }
         else
         {
            set_event_actionpoint_failed(offspring, MONITOR_STATUS_MEMORY_EXHAUSTED);
         }
      }
      else
      {
         set_event_actionpoint_failed(offspring, MONITOR_STATUS_REQUEST_INVALID);
      }
   }
   return 0;
}

static void
disable_evacts_in_segment(UserEventAttachSig *req)
{
   blocks_user_events *user_posts, *continue_next, *remove;
   PROCESS sid = req->sid;

   user_posts = list_head(rw.user_events);

   while (NULL != user_posts)
   {
      if (user_posts->sid == sid)
      {
         continue_next = list_next(rw.user_events, &user_posts->link);
         LOCK_SWAP();
         remove = (blocks_user_events *)list_unlink(user_posts);
         UNLOCK_SWAP();
         free_buf((union SIGNAL **)&remove);

         user_posts = continue_next;
      }
      else
      {
         user_posts = list_next(rw.user_events, &user_posts->link);
      }
   }
}

/*
**
** Called during the system start procedure to locate all current
** processes and give them system process status.
** At this point, only the system processes and zzmain exist.
*/
static void
init_system_processes(void)
{
   struct OS_pid_list *bid_list;
   OSBUFSIZE       bid_count = 0;
   register PROCESS *bid = NULL;
   struct OS_pid_list *pid_list;
   OSBUFSIZE       pid_count = 0;
   register PROCESS *pid = NULL;

   /* Get all available block id's. */
   if ((bid_list = get_bid_list((OSUSER) 0)))
   {
      bid_count = bid_list->count;
      bid = bid_list->list;
   }

   /* For all blocks... */
   for (; bid_count; bid_count--, bid++)
   {
      /* Get all available process id's. */
      if ((pid_list = get_pid_list(*bid)))
      {
         pid_count = pid_list->count;
         pid = pid_list->list;
      }

      /* For all processes... */
      for (; pid_count; pid_count--, pid++)
      {
         struct debug_pcb *dcb;

         /* Set the sysproc flag in the pcb. */
         /* Also select the ose and user error hooks. */
         dcb = zzget_debugpcb(*pid);
         dcb->debug_flags = SYS_PROCESS + 0x30;
         /* Cancel the sysproc flag again for the hunt daemon phantom. */
         if (!strcmp((const char *) dcb->name, "ose_huntd"))
         {
            struct OS_pcb  *pcb;

            pcb = get_pcb(dcb->pid);
            if (pcb)
            {
               if (pcb->type == OS_PHANTOM)
               {
                  dcb->debug_flags &= ~SYS_PROCESS;
               }
               free_buf((union SIGNAL **) &pcb);
            }
         }
      }
      if (pid_list)
      {
         free_buf((union SIGNAL **) &pid_list);
      }
   }
   if (bid_list)
   {
      free_buf((union SIGNAL **) &bid_list);
   }
}

static int
handle_set_event_state_request(Offspring * offspring, union SIGNAL *sig)
{
   MonitorSetEventStateRequest *req = &sig->set_event_state_request;
   union SIGNAL   *reply;

   reply = alloc(sizeof (MonitorSetEventStateReply), MONITOR_SET_EVENT_STATE_REPLY);


   if (offspring == rw.evact_lock_owner && req->state < 255)
   {
      rw.precious->action_state = rw.action_state = (uint8_t) req->state;
      reply->set_event_state_reply.status = MONITOR_STATUS_OK;
   }
   else
   {
      reply->set_event_state_reply.status = MONITOR_STATUS_REQUEST_INVALID;
   }
   send(&reply, offspring->pid);
   return 0;
}

static int
handle_get_event_state_request(Offspring * offspring, union SIGNAL *sig)
{
   MonitorGetEventStateRequest *req = &sig->get_event_state_request;
   union SIGNAL   *reply;

   (void)req;

   reply = alloc(sizeof (MonitorGetEventStateReply), MONITOR_GET_EVENT_STATE_REPLY);

   if (offspring == rw.evact_lock_owner)
   {
      reply->get_event_state_reply.state = rw.action_state;
      reply->get_event_state_reply.status = MONITOR_STATUS_OK;
   }
   else
   {
      reply->get_event_state_reply.status = MONITOR_STATUS_REQUEST_INVALID;
   }
   send(&reply, offspring->pid);
   return 0;
}




/*
 * Dump related functions.
 */

#ifdef USE_RAMDUMP

/**
 * @function ramdump_status
 * @brief Map a ramdump status value to a monitor status value
 */
static          MonitorStatus
ramdump_status(RamdumpStatus status)
{
   switch (status)
   {
      case RAMDUMP_SUCCESS:
         return MONITOR_STATUS_OK;
      case RAMDUMP_EMEMORY_ACCESS_NOK:
         return MONITOR_STATUS_MEMORY_GET_ERROR;
      case RAMDUMP_EPOS_OR_DUMP_NOT_AVAILABLE:
      case RAMDUMP_EINVALID_ARGUMENT:
      case RAMDUMP_ERAMDUMP_NOT_INITIALIZED:
         return MONITOR_STATUS_REQUEST_INVALID;
      default:
         return MONITOR_STATUS_INTERNAL_ERROR;
   }
}

/**
 * @function handle_get_ramdump_info_request
 * @brief Handle a request to get RAM dump information
 */
static int
handle_get_ramdump_info_request(Offspring * offspring, union SIGNAL *sig)
{
   MonitorGetRamdumpInfoRequest *req = &sig->get_ramdump_info_request;
   union SIGNAL   *signal;
   struct RamdumpDumpInfo dump_info;
   RamdumpStatus   status = RAMDUMP_SUCCESS;
   uint32_t        i;

   assert_offspring(offspring);

   /* Get dump info for dumps in requested interval */
   for (i = req->from; i < req->to; i++)
   {
      /* Retrieve dump id corresponding to index */
      RamdumpDumpId   dump_id = ramdump_get_dump_id(i);

      if (!dump_id)
      {
         /* End of dumps, break out of loop. */
         break;
      }

      /* Retrieve information corresponding to dump id */
      status = ramdump_get_dump_info(dump_id, &dump_info);

      if (status == RAMDUMP_SUCCESS)
      {
         /* Fill out reply signal fields */
         signal = alloc(sizeof (MonitorGetRamdumpInfoReply), MONITOR_GET_RAMDUMP_INFO_REPLY);
         signal->get_ramdump_info_reply.dumpId = dump_info.dump_id;
         signal->get_ramdump_info_reply.size = dump_info.size;
         signal->get_ramdump_info_reply.sec = dump_info.clock;
         signal->get_ramdump_info_reply.microsec = dump_info.microseconds;

         if (!send2offspring(&signal, offspring))
         {
            /* Lost offspring */
            return 0;
         }
      }
      else
      {
         /* Ramdump info retrieval failed for some reason.
          * Break out of loop. */
         break;
      }
   }

   /* Send endmark */
   signal = alloc(sizeof (MonitorGetRamdumpInfoEndmark), MONITOR_GET_RAMDUMP_INFO_ENDMARK);
   signal->get_ramdump_info_endmark.status = ramdump_status(status);
   send(&signal, offspring->pid);
   return 0;
}


/**
 * @function handle_get_ramdump_request
 * @brief Handle a request to get RAM dump contents
 */
static int
handle_get_ramdump_request(Offspring * offspring, union SIGNAL *sig)
{
   MonitorGetRamdumpRequest *req = &sig->get_ramdump_request;
   union SIGNAL   *signal;
   struct RamdumpDumpInfo dump_info;
   RamdumpStatus   dump_status;
   MonitorStatus   mon_status = MONITOR_STATUS_OK;

   assert_offspring(offspring);

   /* Get information for requested dump */
   dump_status = ramdump_get_dump_info(req->dumpId, &dump_info);

   if (dump_status == RAMDUMP_SUCCESS)
   {
      /* Dump exists */

      uint32_t        offset = 0;
      uint32_t        byte_count = dump_info.size;

      while (byte_count > 0)
      {
         int             bytes_read;
         uint32_t        chunk_size;

         /* There are more bytes left to read, keep reading... */

         /* Read memory in chunks of at most MAX_MEMORY_XFER_CHUNK bytes */
         chunk_size = (byte_count > MAX_MEMORY_XFER_CHUNK) ? MAX_MEMORY_XFER_CHUNK : byte_count;

         /* Allocate signal and read dump data directly to signal buffer */
         signal = alloc(sizeof (struct MonitorGetRamdumpReply) + chunk_size, MONITOR_GET_RAMDUMP_REPLY);

         bytes_read = ramdump_read_dump(req->dumpId, offset, (void *) signal->get_ramdump_reply.data, chunk_size);

         if (bytes_read == RAMDUMP_EMEMORY_ACCESS_NOK
             || bytes_read == RAMDUMP_EPOS_OR_DUMP_NOT_AVAILABLE
             || bytes_read == RAMDUMP_ERAMDUMP_NOT_INITIALIZED || bytes_read < 0)
         {
            /* Unable to read dump, error reported */
            mon_status = ramdump_status(bytes_read);
            free_buf(&signal);
            break;
         }

         if (bytes_read <= 0 || chunk_size < (uint32_t) bytes_read)
         {
            /* No bytes were read, or too many */
            mon_status = MONITOR_STATUS_MEMORY_GET_ERROR;
            free_buf(&signal);
            break;
         }

         /* Fill out the remaining fields of the reply signal */
         signal->get_ramdump_reply.dumpId = req->dumpId;
         signal->get_ramdump_reply.offset = offset;
         signal->get_ramdump_reply.dataCount = bytes_read;

         /* Update loop counters */
         offset += (uint32_t) bytes_read;
         byte_count -= (uint32_t) bytes_read;

         if (!send2offspring(&signal, offspring))
         {
            /* Lost offspring */
            return 0;
         }
      }
   }
   else
   {
      /* Unable to get dump info for specified dump_id */
      mon_status = ramdump_status(dump_status);
   }

   /* Send endmark */
   signal = alloc(sizeof (MonitorGetRamdumpEndmark), MONITOR_GET_RAMDUMP_ENDMARK);
   signal->get_ramdump_endmark.status = mon_status;
   send(&signal, offspring->pid);
   return 0;
}

#endif

#ifdef RMM_USE_PERSISTENT_REGIONS

static void    *
create_reg(PROCESS ose_mm_, const char *name, uint32_t size, uint32_t * actual_size, uint32_t options)
{
   const SIGSELECT all_signals[] = { 0 };
   void           *memory = NULL;
   union SIGNAL   *signal;

   assert(*actual_size == 0);

   signal = alloc(sizeof (struct mm_create_region), MM_CREATE_REGION_REQUEST);

   signal->mm_create_region.size = size;
   signal->mm_create_region.domain = MM_KERNEL_DOMAIN;
   strncpy(signal->mm_create_region.name, name, MM_NAME_LENGTH);
   strcpy(signal->mm_create_region.phys_mem, MM_INTERNAL_RAM);
   strcpy(signal->mm_create_region.log_mem, "RAM_SASE");
   signal->mm_create_region.permissions = SU_RW_USR_NA;
   signal->mm_create_region.cache_mode = MM_CACHE_COPY_BACK | MM_SPECULATIVE_ACCESS;
   signal->mm_create_region.base = 0,   /* Optional and not used. */
      signal->mm_create_region.options = options;
   signal->mm_create_region.align = 32;
   signal->mm_create_region.phys_count = 0;
   signal->mm_create_region.low_limit = 0;
   send(&signal, ose_mm_);

   signal = receive_from(~0UL, all_signals, ose_mm_);
   assert(signal);
   if (signal->sig_no == MM_CREATE_REGION_REPLY)
   {
      if (signal->mm_create_region.status == MM_SUCCESS)
      {
         *actual_size = signal->mm_create_region.size;
         memory = (void *) signal->mm_create_region.base;
         ramlog_printf("rmm: create_region success, address: %u\n", memory);
      }
      else
      {
         ramlog_printf("rmm: create_region failed options: %u, status %u\n", options, signal->mm_create_region.status);
      }
   }
   free_buf(&signal);

   return memory;
}

static void    *
get_persistent_memory(const char *name, uint32_t size, uint32_t * actual_size)
{
   void           *memory = NULL;
   PROCESS         ose_mm_;

   *actual_size = 0;

   ramlog_printf("rmm:     get_persistent_memory name %s size %u\n", name, size);
   if (hunt("ose_memory_mgr", 0, &ose_mm_, NULL))
   {
      union SIGNAL   *reply;
      const SIGSELECT all_signals[] = { 0 };
      struct mm_list_regions_request *list_req;

      list_req = (struct mm_list_regions_request *) alloc(sizeof (struct mm_list_regions_request),
                                                          MM_LIST_REGIONS_REQUEST);
      list_req->domain = MM_KERNEL_DOMAIN;
      send((union SIGNAL **) &list_req, ose_mm_);

      reply = receive_from(~0UL, all_signals, ose_mm_);
      assert(reply);
      if (MM_LIST_REGIONS_REPLY == reply->sig_no)
      {
         if (reply->mm_list_regions_reply.status == MM_SUCCESS)
         {
            int             i;

            ramlog_printf("rmm: got region list reply count = %u\n", reply->mm_list_regions_reply.count);
            for (i = 0; i < reply->mm_list_regions_reply.count; i++)
            {
               union SIGNAL   *info_req = alloc(sizeof (struct mm_create_region),
                                                MM_GET_REGION_INFO_REQUEST);
               union SIGNAL   *info_rep;

               info_req->mm_create_region.domain = MM_KERNEL_DOMAIN;
               info_req->mm_create_region.base = reply->mm_list_regions_reply.regions[i];
               send(&info_req, ose_mm_);

               info_rep = receive_from(~0UL, all_signals, ose_mm_);
               assert(info_rep);
               if (MM_GET_REGION_INFO_REPLY == info_rep->sig_no)
               {
                  if (info_rep->mm_create_region.status == MM_SUCCESS)
                  {
                     ramlog_printf("rmm: got region info for region at 0x%X\n",
                                   reply->mm_list_regions_reply.regions[i]);
                     if ((info_rep->mm_create_region.options & MM_CREATE_REGION_SPECIFY_NAME) &&
                         (strcmp(info_rep->mm_create_region.name, name) == 0))
                     {
                        /* We found our persistent region */
                        if (size <= info_rep->mm_create_region.size)
                        {
                           memory = (void *) info_rep->mm_create_region.base;
                           *actual_size = info_rep->mm_create_region.size;
                           break;
                        }
                        else
                        {
                           /* We found the region but it is too small so delete it
                            * so that we can creat it again */
                           /* FIXME !!! */
                           ramlog_printf("rmm: we *should* delete the existing region that is too small here!\n");
                        }
                     }
                  }
                  else
                  {
                     ramlog_printf("rmm: failed to get region info for region at 0x%X with status %u\n",
                                   reply->mm_list_regions_reply.regions[i], info_rep->mm_create_region.status);
                  }
               }
               else
               {
                  ramlog_printf("rmm: got unexpected signal from mm: signo: %u\n", info_rep->sig_no);
                  assert(0);
               }
               free_buf(&info_rep);
            }
         }
         else
         {
            ramlog_printf("rmm: got error status from MM while listing regions: %u\n",
                          reply->mm_list_regions_reply.status);
         }
      }
      else
      {
         ramlog_printf("rmm: got unexpected signal from mm: signo: %u\n", reply->sig_no);
         assert(0);
      }
      free_buf(&reply);

      if (memory == NULL)
      {
         /* Something failed with retriving an existing persistent region (maybe it did not exists)
          * so try to create a new persistent region. */
         memory = create_reg(ose_mm_, name, size, actual_size,
                             MM_CREATE_REGION_SPECIFY_NAME | MM_CREATE_REGION_SPECIFY_ALIGN |
                             MM_CREATE_REGION_PERSISTENT);
         if (!memory)
         {
            /* That failed too so retry without requesting it to be persistent. */
            memory = create_reg(ose_mm_, name, size, actual_size,
                                MM_CREATE_REGION_SPECIFY_NAME | MM_CREATE_REGION_SPECIFY_ALIGN);

         }
      }
   }
   else
   {
      ramlog_printf("rmm: ose_mm not found!\n");
   }
   return memory;
}

#endif

void
ose_rmm_set_req_enabled(SIGSELECT signo, int enabled)
{
   uint32_t index = (signo - MONITOR_INTERFACE_REQUEST) / (8*sizeof(uint32_t));
   uint32_t bit   = (signo - MONITOR_INTERFACE_REQUEST) % (8*sizeof(uint32_t));
   PROCESS  caller_pid = current_process();

   if (! zzkrn_get_pid_info(caller_pid, 10 /* Supervisor mode */)
       || get_uid(caller_pid) != 0)
   {
      /* Must be supervisor mode process with user ID zero */
      zzos_error(0, OSE_EILLEGAL_USER_CALL, 0);
      return;
   }

   if (enabled)
   {
      /* Clear disabled bit */
      rw.req_disabled_bitmask[index] &= ~(1<<bit);
   }
   else
   {
      /* Set disabled bit */
      rw.req_disabled_bitmask[index] |= (1<<bit);
   }
}

static int
handle_disconnect(Offspring *offspring, union SIGNAL *sig)
{
   (void)sig;
#ifndef softOSE
   /* Stop all active profiling counters */
   handle_disconnect_request_counter(offspring, sig);
#endif
   disconnect_offspring(offspring, "client requested disconnect");
   return 0;
}

static int
handle_blocking_flow_control(Offspring *offspring, union SIGNAL *sig)
{
   (void)sig;
   offspring->flow_ctrl_block += rw.flow_ctrl_block / 2 + 1;
   return 0;
}

static int
handle_nonblocking_flow_control(Offspring *offspring, union SIGNAL *sig)
{
   (void)sig;
   offspring->flow_ctrl_nblock += rw.flow_ctrl_nblock / 2 + 1;
   return 0;
}

static int
handle_get_user_reports_enabled(Offspring *offspring, union SIGNAL *sig)
{
   union SIGNAL   *sig_r;
   ProfUserReport *cur_report;
   cur_report = find_report(sig->get_user_reports_enabled_request.reportNo);

   /* Send a reply to the offspring. */
   sig_r = alloc(sizeof (MonitorGetUserReportsEnabledReply) + 1,
                 MONITOR_GET_USER_REPORTS_ENABLED_REPLY);
   if (cur_report)
   {
      sig_r->get_user_reports_enabled_reply.reportNo = cur_report->reportno;
      sig_r->get_user_reports_enabled_reply.enabled = cur_report->icounter ? 1 : 0;

      if (sig_r->get_user_reports_enabled_reply.enabled)
      {
         sig_r->get_user_reports_enabled_reply.interval = cur_report->interval;
         sig_r->get_user_reports_enabled_reply.maxReports = cur_report->maxReports;
         sig_r->get_user_reports_enabled_reply.maxValuesReport = cur_report->maxValues;
      }
      else
      {
         sig_r->get_user_reports_enabled_reply.interval = rw.default_user_interval;
         sig_r->get_user_reports_enabled_reply.maxReports = rw.default_user_reports;
         sig_r->get_user_reports_enabled_reply.maxValuesReport = rw.default_user_maxvalues;
      }
      get_abstime((OSTICK *)&sig_r->get_user_reports_enabled_reply.sec,
           (OSTICK *)&sig_r->get_user_reports_enabled_reply.sectick,
           (OSTICK *)&sig_r->get_user_reports_enabled_reply.secntick);
      sig_r->get_user_reports_enabled_reply.reserved0 = 0;
   }
   else
   {
      sig_r->get_user_reports_enabled_reply.reportNo = sig->get_user_reports_enabled_request.reportNo;
      sig_r->get_user_reports_enabled_reply.enabled = 0;
      sig_r->get_user_reports_enabled_reply.interval = 0;
      sig_r->get_user_reports_enabled_reply.maxReports = 0;
      sig_r->get_user_reports_enabled_reply.maxValuesReport = 0;
      sig_r->get_user_reports_enabled_reply.reserved0 = 0;
      sig_r->get_user_reports_enabled_reply.sec = 0;
      sig_r->get_user_reports_enabled_reply.sectick = 0;
      sig_r->get_user_reports_enabled_reply.secntick = 0;
   }
   send(&sig_r, offspring->pid);
   return 0;
}

static int
handle_get_user_reports(Offspring *offspring, union SIGNAL *sig)
{
   union SIGNAL   *sig_r;
   ProfUserReport *cur_report = NULL;

   /* Get requested report */
   cur_report = find_report(sig->get_user_reports_request.reportNo);

   if (cur_report)
   {
      /* Restart the disconnect timer. */
      cur_report->prof_report_disctimer = 100000000L / zzprof.prof_all_ticklen;   /* 100 sec. */
      handle_report_val(&sig, 1, offspring, cur_report);
      /* Do not free_buf in main signal loop */
      /* handle_report_val takes care of that */
      return 1;
   }
   else
   {
      /* Send a not ok responde to offspring and break out of case*/
      sig_r = alloc(sizeof (MonitorGetUserReportsReply),
                    MONITOR_GET_USER_REPORTS_REPLY);

      sig_r->get_user_reports_reply.status = MONITOR_STATUS_REQUEST_INVALID;
      send(&sig_r, offspring->pid);
      return 0;
   }
}

static int
handle_set_notify_enabled(Offspring *offspring, union SIGNAL *sig)
{
   union SIGNAL   *rep;

   rep = alloc(sizeof (MonitorSetNotifyEnabledReply), MONITOR_SET_NOTIFY_ENABLED_REPLY);
   rep->set_notify_enabled_reply.status = MONITOR_STATUS_OK;
   set_monitor_status(offspring, sig->set_notify_enabled_request.enabled);
   send(&rep, offspring->pid);
   return 0;
}

static int
handle_get_notify_enabled(Offspring *offspring, union SIGNAL *sig)
{
   union SIGNAL   *rep;

   (void)sig;
   rep = alloc(sizeof (MonitorGetNotifyEnabledReply), MONITOR_GET_NOTIFY_ENABLED_REPLY);
   rep->get_notify_enabled_reply.status = MONITOR_STATUS_OK;
   rep->get_notify_enabled_reply.enabled = rw.monitor_enabled;
   send(&rep, offspring->pid);
   return 0;
}

static int
handle_set_intercept_scope(Offspring *offspring, union SIGNAL *sig)
{
   if (rw.evact_lock_owner == offspring)
   {
      MonitorSetScopeRequest *req = &sig->set_scope_request;
      MonitorStatus   status = MONITOR_STATUS_OK;
      char            scope[5 + 10 + 1]; /* strlen("*:*:*") + strlen("0x00000000") + 1 */
      union SIGNAL   *reply = alloc(sizeof (MonitorSetScopeReply),
                                    MONITOR_SET_INTERCEPT_SCOPE_REPLY);

      switch (req->scope.type)
      {
         case MONITOR_SCOPE_ID:
            sprintf(scope, "#%#x", req->scope.id);
            break;
         case MONITOR_SCOPE_BLOCK_ID:
            sprintf(scope, "#%#x:*", req->scope.id);
            break;
         case MONITOR_SCOPE_SEGMENT_ID:
            sprintf(scope, "#%#x:*:*", req->scope.id);
            break;
         case MONITOR_SCOPE_SYSTEM:
            sprintf(scope, "*:*:*");
            break;
         default:
            status = MONITOR_STATUS_SCOPE_INVALID;
            break;
      }
      set_evact_catch_scope(1, scope);
      reply->set_scope_reply.status = status;
      send(&reply, offspring->pid);
   }
   else
   {
      ramlog_printf("rmm: error: set evact catch scope" " without evact lock\n");
   }

   return 0;
}

static int
handle_set_scope(Offspring *offspring, union SIGNAL *sig)
{
   if (rw.evact_lock_owner == offspring)
   {
      MonitorSetScopeRequest *req = &sig->set_scope_request;
      MonitorStatus   status = MONITOR_STATUS_OK;
      char            scope[5 + 10 + 1]; /* strlen("*:*:*") + strlen("0x00000000") + 1 */
      union SIGNAL   *reply = alloc(sizeof (MonitorSetScopeReply),
                                    MONITOR_SET_SCOPE_REPLY);

      switch (req->scope.type)
      {
         case MONITOR_SCOPE_ID:
            sprintf(scope, "#%#x", req->scope.id);
            break;
         case MONITOR_SCOPE_BLOCK_ID:
            sprintf(scope, "#%#x:*", req->scope.id);
            break;
         case MONITOR_SCOPE_SEGMENT_ID:
            sprintf(scope, "#%#x:*:*", req->scope.id);
            break;
         case MONITOR_SCOPE_SYSTEM:
            sprintf(scope, "*:*:*");
            break;
         default:
            status = MONITOR_STATUS_SCOPE_INVALID;
            break;
      }
      set_evact_monitor_scope(1, scope);
      reply->set_scope_reply.status = status;
      send(&reply, offspring->pid);
   }
   else
   {
      ramlog_printf("rmm: error: set evact monitor/trace scope"
         " without evact lock\n");
   }

   return 0;
}

static int
handle_set_trace_enabled(Offspring *offspring, union SIGNAL *sig)
{
   union SIGNAL   *rep;

   rep = alloc(sizeof (MonitorSetTraceEnabledReply), MONITOR_SET_TRACE_ENABLED_REPLY);
   rep->set_trace_enabled_reply.status = MONITOR_STATUS_OK;
   if (sig->set_trace_enabled_request.enabled)
   {
      ose_trace_enable(rw.trace);
      rw.trace_enabled = 1;
   }
   else
   {
      ose_trace_disable(rw.trace);
      rw.trace_enabled = 0;
   }
   rw.precious->trace_enabled = rw.trace_enabled;
   send(&rep, offspring->pid);
   return 0;
}

static int
handle_get_trace_enabled(Offspring *offspring, union SIGNAL *sig)
{
   union SIGNAL   *rep;

   (void)sig;
   rep = alloc(sizeof (MonitorGetTraceEnabledReply), MONITOR_GET_TRACE_ENABLED_REPLY);
   rep->get_trace_enabled_reply.status = MONITOR_STATUS_OK;
   rep->get_trace_enabled_reply.enabled = rw.trace_enabled;
   send(&rep, offspring->pid);
   return 0;
}

static int
handle_clear_trace(Offspring *offspring, union SIGNAL *sig)
{
   union SIGNAL   *rep;
   (void)sig;

   if (offspring->mget_req != NULL)
   {
      /* Send what we got now and force the client to restart. */
      union SIGNAL *mget_req = offspring->mget_req;

      offspring->mget_tmo = -1; /* signal timeout */
      handle_get_trace_multiple(offspring, mget_req);

      /* Request completed */
      free_buf(&mget_req);
   }
   rep = alloc(sizeof (MonitorClearTraceReply), MONITOR_CLEAR_TRACE_REPLY);
   rep->clear_trace_reply.status = MONITOR_STATUS_OK;
   ose_trace_clear(rw.trace);
   ose_trace_close(&offspring->cursor);
   offspring->cursor.base = NULL;
   offspring->mget_read = 1;
   idbuf_init(&offspring->mtrace_pids);
   send(&rep, offspring->pid);

   return 0;
}

static int
handle_detach(Offspring *offspring, union SIGNAL *sig)
{
   union SIGNAL   *rep;

   (void)sig;
   rep = alloc(sizeof (MonitorDetachReply), MONITOR_DETACH_REPLY);
   rep->detach_reply.status = MONITOR_STATUS_OK;
   detach_offspring(offspring);
   send(&rep, offspring->pid);
   return 0;
}

#ifdef USE_EPM
#   define EPM_REQ(x) (x)
#else
#   define EPM_REQ(x) NULL
#endif

/* Return 1 if the signal should be kept, 0 if the signal should be freed */
typedef int (*signal_request_t)(Offspring *, union SIGNAL *);

static const signal_request_t sig_handler_table[] = {
   NULL,                               /* [0] MONITOR_INTERFACE_REQUEST (39000) */
   NULL,
   NULL,                               /* [2] MONITOR_OBSOLETE_CONNECT_REQUEST (39002) */
   NULL,
   handle_disconnect,                  /* [4] MONITOR_DISCONNECT_REQUEST (39004) */
   NULL,
   NULL,                               /* [6] MONITOR_NAME_REQUEST (39006) */
   NULL,
   NULL,                               /* [8] MONITOR_CONNECT_REQUEST (39008) */
   NULL,
   NULL,                               /* [10] MONITOR_GET_PROCESS_INFO_REQUEST (39010) */
   NULL, NULL, NULL,
   NULL,                               /* [14] MONITOR_GET_BLOCK_INFO_REQUEST (39014) */
   NULL, NULL,
   NULL,                               /* [17] MONITOR_GET_SEGMENT_INFO_REQUEST (39017) */
   NULL, NULL, NULL, NULL, NULL, NULL, NULL,
   NULL,                               /* [25] MONITOR_GET_STACK_USAGE_REQUEST (39025) */
   NULL, NULL, NULL, NULL,
   NULL,                               /* [30] MONITOR_GET_ENV_VARS_REQUEST (39030) */
   NULL, NULL, NULL, NULL,
   NULL,                               /* [35] MONITOR_GET_SIGNAL_QUEUE_REQUEST (39035) */
   NULL, NULL, NULL, NULL,
   NULL,                               /* [40] MONITOR_GET_ATTACH_LIST_REQUEST (39040) */
   NULL, NULL, NULL, NULL,
   NULL,                               /* [45] MONITOR_GET_SIGNAL_REDIRECTION_REQUEST (39045) */
   NULL, NULL, NULL, NULL,
   NULL,                               /* [50] MONITOR_GET_POOL_INFO_REQUEST (39050) */
   NULL, NULL,
   NULL,                               /* [53] MONITOR_GET_POOL_SIGNALS_REQUEST (39053) */
   NULL, NULL, NULL, NULL, NULL, NULL,
   NULL,                               /* [60] MONITOR_GET_SYS_PARAM_REQUEST (39060) */
   NULL, NULL, NULL, NULL,
   NULL,                               /* [65] MONITOR_BLOCKING_FLOW_CONTROL_REQUEST (39065) */
   handle_blocking_flow_control,       /* [66] MONITOR_BLOCKING_FLOW_CONTROL_REPLY (39066) */
   NULL,                               /* [67] MONITOR_NON_BLOCKING_FLOW_CONTROL_REQUEST (39067) */
   handle_nonblocking_flow_control,    /* [68] MONITOR_NON_BLOCKING_FLOW_CONTROL_REPLY (39068) */
   NULL, NULL, NULL, NULL, NULL, NULL,
   NULL,                               /* [75] MONITOR_CREATE_PROCESS_REQUEST (39075) */
   NULL,
   NULL,                               /* [77] MONITOR_KILL_SCOPE_REQUEST (39077) */
   NULL,
   NULL,                               /* [79] MONITOR_START_SCOPE_REQUEST (39079) */
   NULL,
   NULL,                               /* [81] MONITOR_STOP_SCOPE_REQUEST (39081) */
   NULL,
   NULL,                               /* [83] MONITOR_SET_ENV_VAR_REQUEST (39083) */
   NULL,
   NULL,                               /* [85] MONITOR_SEND_SIGNAL_REQUEST (39085) */
   NULL,
   NULL,                               /* [87] MONITOR_SIGNAL_SEMAPHORE_REQUEST (39087) */
   NULL,
   NULL,                               /* [89] MONITOR_SET_PRIORITY_REQUEST (39089) */
   NULL,
   handle_set_sys_param_request,       /* [91] MONITOR_SET_SYS_PARAM_REQUEST (39091) */
   NULL, 
   NULL,                               /* [93] MONITOR_SET_EXECUTION_UNIT_REQUEST (39093) */ 
   NULL, 
   NULL, NULL, NULL, NULL, NULL,
   NULL,                               /* [100] MONITOR_GET_MEMORY_REQUEST (39100) */
   NULL, NULL,
   handle_set_memory_request,          /* [103] MONITOR_SET_MEMORY_REQUEST (39103) */
   NULL,
   EPM_REQ(handle_get_registers_request), /* [105] MONITOR_GET_REGISTERS_REQUEST (39105) */
   NULL,
   EPM_REQ(handle_set_registers_request), /* [107] MONITOR_SET_REGISTERS_REQUEST (39107) */
   NULL,
   handle_attach_request,              /* [109] MONITOR_ATTACH_REQUEST (39109) */
   NULL,
   handle_detach,                      /* [111] MONITOR_DETACH_REQUEST (39111) */
   NULL,
   handle_basic_event_action_request,  /* [113] MONITOR_BASIC_ERROR_ACTION_REQUEST (39113) */
   NULL,
   handle_basic_event_action_request,  /* [115] MONITOR_BASIC_CREATE_KILL_ACTION_REQUEST (39115) */
   NULL, NULL, NULL, NULL, NULL,
   NULL,                               /* [121] MONITOR_GET_REGISTER_INFO_REQUEST (39121) */
   EPM_REQ(handle_set_breakpoint_request),   /* [122] MONITOR_SET_BREAKPOINT_REQUEST (39122) */
   NULL, NULL,
   EPM_REQ(handle_clear_breakpoint_request), /* [125] MONITOR_CLEAR_BREAKPOINT_REQUEST (39125) */
   NULL,
   EPM_REQ(handle_step_process_request),     /* [127] MONITOR_STEP_PROCESS_REQUEST (39127) */
   NULL, NULL,
   handle_intercept_scope_request,     /* [130] MONITOR_INTERCEPT_SCOPE_REQUEST (39130) */
   NULL,
   handle_resume_scope_request,        /* [132] MONITOR_RESUME_SCOPE_REQUEST (39132) */
   NULL,NULL,NULL,NULL,NULL,NULL,NULL,
   NULL,                               /* [140] MONITOR_GET_HEAP_INFO_REQUEST (39140) */
   NULL,NULL,
   NULL,                               /* [143] MONITOR_GET_HEAP_FRAGMENT_INFO_REQUEST (39143) */
   NULL,NULL,
   NULL,                               /* [146] MONITOR_GET_HEAP_BUFFERS_REQUEST (39146) */
   NULL,NULL,NULL,
   set_send_actionpoint,               /* [150] MONITOR_SET_SEND_ACTIONPOINT_REQUEST (39150) */
   set_receive_actionpoint,            /* [151] MONITOR_SET_RECEIVE_ACTIONPOINT_REQUEST (39151) */
   set_swap_actionpoint,               /* [152] MONITOR_SET_SWAP_ACTIONPOINT_REQUEST (39152) */
   set_create_actionpoint,             /* [153] MONITOR_SET_CREATE_ACTIONPOINT_REQUEST (39153) */
   set_kill_actionpoint,               /* [154] MONITOR_SET_KILL_ACTIONPOINT_REQUEST (39154) */
   set_error_actionpoint,              /* [155] MONITOR_SET_ERROR_ACTIONPOINT_REQUEST (39155) */
   set_ue_actionpoint,                 /* [156] MONITOR_SET_USER_ACTIONPOINT_REQUEST (39156) */
#ifdef OSE_SMP
   set_bind_actionpoint,               /* [157] MONITOR_SET_BIND_ACTIONPOINT_REQUEST (39157) */
#else
   NULL,                               /* [157] MONITOR_SET_BIND_ACTIONPOINT_REQUEST (39157) */
#endif
   set_alloc_actionpoint,              /* [158] MONITOR_SET_ALLOC_ACTIONPOINT_REQUEST (39158) */
   set_free_buf_actionpoint,           /* [159] MONITOR_SET_FREE_BUF_ACTIONPOINT_REQUEST (39159) */
   NULL,
   handle_remove_evact,                /* [161] MONITOR_CLEAR_EVENT_ACTIONPOINT_REQUEST (39161) */
   NULL,
   handle_get_event_actionpoints,      /* [163] MONITOR_GET_EVENT_ACTIONPOINT_INFO_REQUEST (39163) */
   NULL, NULL, NULL, NULL, NULL, NULL,
   handle_get_event_actionpoints,      /* [170] MONITOR_GET_EVENT_ACTIONPOINTS_REQUEST (39170) */
   NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
   handle_set_event_state_request,     /* [180] MONITOR_SET_EVENT_STATE_REQUEST (39180) */
   NULL,
   handle_get_event_state_request,     /* [182] MONITOR_GET_EVENT_STATE_REQUEST (39182) */
   NULL, NULL,
   handle_set_notify_enabled,          /* [185] MONITOR_SET_NOTIFY_ENABLED_REQUEST (39185) */
   NULL,
   handle_get_notify_enabled,          /* [187] MONITOR_GET_NOTIFY_ENABLED_REQUEST (39187) */
   NULL, NULL,
   NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
   NULL, NULL, NULL, NULL, NULL,
   handle_set_trace_enabled,           /* [205] MONITOR_SET_TRACE_ENABLED_REQUEST (39205) */
   NULL,
   handle_get_trace_enabled,           /* [207] MONITOR_GET_TRACE_ENABLED_REQUEST (39207) */
   NULL,
   handle_clear_trace,                 /* [209] MONITOR_CLEAR_TRACE_REQUEST (39209) */
   NULL,
   handle_get_trace_multiple,          /* [211] MONITOR_GET_TRACE_MULTIPLE_REQUEST (39211) */
   NULL, NULL, NULL,
   handle_get_trace,                   /* [215] MONITOR_GET_TRACE_REQUEST (39215) */
   NULL, NULL, NULL, NULL,
   NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
   NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
   handle_set_scope,                   /* [240] MONITOR_SET_SCOPE_REQUEST (39240) */
   NULL,
   handle_set_intercept_scope,         /* [242] MONITOR_SET_INTERCEPT_SCOPE_REQUEST (39242) */
   NULL, NULL, NULL,
   handle_set_intercept_enabled,       /* [246] MONITOR_SET_INTERCEPT_ENABLED_REQUEST (39246) */
   NULL,
   handle_get_intercept_enabled,       /* [248] MONITOR_GET_INTERCEPT_ENABLED_REQUEST (39248) */
   NULL,
   zzprof_handle_get_cpu_reports_enabled,     /* [250] MONITOR_GET_CPU_REPORTS_ENABLED_REQUEST (39250) */
   NULL,
   zzprof_handle_set_cpu_reports_enabled,     /* [252] MONITOR_SET_CPU_REPORTS_ENABLED_REQUEST (39252) */
   NULL,
   zzprof_handle_get_cpu_pri_reports_enabled, /* [254] MONITOR_GET_CPU_PRI_REPORTS_ENABLED_REQUEST (39254) */
   NULL,
   zzprof_handle_set_cpu_pri_reports_enabled, /* [256] MONITOR_SET_CPU_PRI_REPORTS_ENABLED_REQUEST (39256) */
   NULL,
   zzprof_handle_get_cpu_process_reports_enabled, /* [258] MONITOR_GET_CPU_PROCESS_REPORTS_ENABLED_REQUEST (39258) */
   NULL,
   zzprof_handle_set_cpu_process_reports_enabled, /* [260] MONITOR_SET_CPU_PROCESS_REPORTS_ENABLED_REQUEST (39260) */
   NULL,
   zzprof_handle_set_cpu_process_reportpoint, /* [262] MONITOR_SET_CPU_PROCESS_REPORTPOINT_REQUEST (39262) */
   NULL,
   zzprof_handle_clear_cpu_process_reportpoint, /* [264] MONITOR_CLEAR_CPU_PROCESS_REPORTPOINT_REQUEST (39264) */
   NULL,
   zzprof_handle_get_cpu_process_reportpoint_info, /* [266] MONITOR_GET_CPU_PROCESS_REPORTPOINT_INFO_REQUEST (39266) */
   NULL, NULL, NULL,
   zzprof_handle_get_cpu_reports,             /* [270] MONITOR_GET_CPU_REPORTS_REQUEST (39270) */
   NULL,
   zzprof_handle_get_cpu_pri_reports,         /* [272] MONITOR_GET_CPU_PRI_REPORTS_REQUEST (39272) */
   NULL,
   zzprof_handle_get_cpu_process_reports,     /* [274] MONITOR_GET_CPU_PROCESS_REPORTS_REQUEST (39274) */
   NULL,
   zzprof_handle_get_cpu_block_reports,       /* [276] MONITOR_GET_CPU_BLOCK_REPORTS_REQUEST (39276) */
   NULL,
   handle_get_user_reports_enabled,            /* [278] MONITOR_GET_USER_REPORTS_ENABLED_REQUEST (39278) */
   NULL,
   handle_set_user_reports_enabled,             /* [280] MONITOR_SET_USER_REPORTS_ENABLED_REQUEST (39280) */
   NULL,
   handle_get_user_reports,                      /* [282] MONITOR_GET_USER_REPORTS_REQUEST (39282) */
   NULL,
   zzprof_handle_get_total_cpu_block_reports_enabled, /* [284] MONITOR_GET_TOTAL_CPU_BLOCK_REPORTS_ENABLED_REQUEST (39284) */
   NULL,
   zzprof_handle_set_total_cpu_block_reports_enabled, /* [286] MONITOR_SET_TOTAL_CPU_BLOCK_REPORTS_ENABLED_REQUEST (39286) */
   NULL,
   zzprof_handle_get_total_cpu_block_reports, /* [288]  MONITOR_GET_TOTAL_CPU_BLOCK_REPORTS_REQUEST (39288) */
   NULL,
   zzprof_handle_get_cpu_process_reports_enabled, /* [290] MONITOR_GET_CPU_BLOCK_REPORTS_ENABLED_REQUEST (39290) */
   NULL,
   zzprof_handle_set_cpu_process_reports_enabled, /* [292] MONITOR_SET_CPU_BLOCK_REPORTS_ENABLED_REQUEST (39292) */
   NULL,
#ifdef HAS_TIMEOUT_EVENT
   set_timeout_actionpoint,            /* [294] MONITOR_SET_TIMEOUT_ACTIONPOINT_REQUEST (39294) */
#else
   NULL,
#endif
   NULL, NULL, NULL, NULL, NULL,
   NULL, NULL, NULL, NULL, NULL,
   NULL, NULL, NULL, NULL, NULL,
   NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
   NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
   NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
   NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
   NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
   NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
   NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
   NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
   NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,

   NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
   NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
   NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
   NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
   NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
   NULL,/* [450] MONITOR_GET_PROFILING_EXECUTION_UNIT_ENABLED_REQUEST */
   NULL,
   NULL,/* [452] MONITOR_SET_PROFILING_EXECUTION_UNIT_ENABLED_REQUEST */
   NULL,
   NULL,
   NULL, NULL, NULL, NULL, NULL,
   NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
   NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
   NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
   NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,

#ifdef USE_RAMDUMP
   handle_get_ramdump_info_request,    /* [500] MONITOR_GET_RAMDUMP_INFO_REQUEST (39500) */
   NULL, NULL,
   handle_get_ramdump_request          /* [503] MONITOR_GET_RAMDUMP_REQUEST (39503) */
#else
   NULL, NULL, NULL, NULL
#endif
};

#define SIGTABLE_LENGTH (sizeof sig_handler_table / sizeof sig_handler_table[0])

static const signal_request_t counter_prof_sig_handler[] = {
   NULL,                                                 /* [0] MONITOR_GET_SUPPORTED_COUNTER_TYPES_REPLY (40781) */
   NULL,
#ifndef softOSE
   handle_get_counter_type_enabled_request,              /* [3] MONITOR_GET_COUNTER_TYPE_ENABLED_REQUEST (40783) */
   NULL,
   handle_set_counter_type_enabled_request,              /* [5] MONITOR_SET_COUNTER_TYPE_ENABLED_REQUEST (40785) */
   NULL,
   handle_get_counter_reports_notify_enabled_request,    /* [7] MONITOR_GET_COUNTER_REPORTS_NOTIFY_ENABLED_REQUEST (40787) */
   NULL,
   handle_set_counter_reports_notify_enabled_request,    /* [9] MONITOR_SET_COUNTER_REPORTS_NOTIFY_ENABLED_REQUEST (40789) */
#endif
   NULL,
   NULL,
   NULL

};

#define COUNTER_PROF_SIG_LENGTH (sizeof counter_prof_sig_handler / sizeof counter_prof_sig_handler[0])

#include "prof.c"

 /* SFK does not support weak references for linking functions
  * and source profiling feature can be disabled for soft kernels. */
#ifndef softOSE
#include "rmm_counterprof.c"
#endif

OSENTRYPOINT zzose_rmm;

/**
 * @function zzose_rmm
 *
 * @brief The rmm process entry point.
 */
OS_PROCESS(zzose_rmm)
{
   static const SIGSELECT sel_start[] = { 1, START_MODEFLAG };
   OSBOOLEAN       warm_start;
   union SIGNAL   *sig;

   rw.flow_ctrl_block = 20;
   rw.flow_ctrl_nblock = 10;
   rw.def_trace_mget_size = 4096;
   rw.monitor_sys_error = 0;

   rw.monitor_error = 1;

   rw.totproc_count = 0;

   /* Initialize user function to NULL */
   rw.user_action_function = NULL;
   rw.user_report_function = NULL;

   rw.user_events = NULL;

   rw.zzose_rmm_ = current_process();
   rw.ose_block_bcb = ((struct _pcb *) get_pid_info(rw.zzose_rmm_, 7))->block;

   sig = receive(sel_start);

   init_system_processes();

#ifndef softOSE
   /* Initialize source/counter profiler. */
   rmm_counter_prof_init();
#endif

   /* Dummy code to make sure the alignment variables is used and included */
   SMP_CODE(zzprof.pre_alignment_dummy[0] = zzprof.post_alignment_dummy[0];)

   /*
    ** Read the base and size of the trace buffer provided by the target interface
    ** in the system control block.
    */
   {
      struct sysctlblk *p;

      p = zzget_sysctlblk();
      rw.trace = (char *) ((p->pool_base + (POOL_ALIGNMENT - 1)) /*& (OSADDRESS) ~ (POOL_ALIGNMENT - 1)*/);
      if (p->tracesize < 8192)
         rw.trace_size = (uint32_t) 8192;
      else
         rw.trace_size = (uint32_t) (p->tracesize /*& (OSADDRESS) ~ (POOL_ALIGNMENT - 1)*/);
   }

   {
      int             psize = sizeof (struct precious_info);

      /* Break precious from the start of the trace and move the trace start. */
      rw.precious = (struct precious_info *) rw.trace;
      psize = (psize & 0xffc0) == psize ? psize : (psize + 0x40) & 0xffc0;
      rw.trace = (((char *) rw.trace) + psize);
      rw.trace_size -= psize;
      ose_rmm_trace_magic = &(rw.precious->magic);
      /*
       ** If use of dynamic trace the trace variable will point to the
       ** dynamic allocated trace region after running the mon_attach
       ** command.
       ** Save address and size of original trace (static trace).
       ** Always to be used when calling sys_alloc_trace() and
       ** ose_trace_alloc()in make_trace_name().
       */
      rw.static_trace = rw.trace;
      rw.static_trace_size = rw.trace_size;
   }

   /* Always try to recover trace the trace */
   if (!ose_trace_recover(rw.trace, rw.trace_size, SYS_TRACE_NAME,
                          (GetTimeFuncPointer)get_systime, 1000000)
       && !recover_precious()
       && reclaim_all_actions() == 0)
   {
      /* Done recovering so restore magic */
      rw.precious->magic = PRECIOUS_MAGIC;
      warm_start = True;
      ramlog_printf("rmm: Trace recovered.\n");
   }
   else
   {
      /* Recover failed, the trace was damaged. Make it a cold start. */
      warm_start = False;
      ramlog_printf("rmm: Trace could not be recovered.\n");
   }

   configure_parameter(SYSPARAM_INTERFACE);
   configure_parameter(SYSPARAM_SYSTEMNAME);
   configure_parameter(SYSPARAM_SYS_ERROR);
   ramlog_printf("rmm: krn/monitor/sys_error=debug:%d\n", rw.monitor_sys_error);

   configure_parameter(SYSPARAM_ERROR);
   /*
    ** Obtain from the target interface a list of system process names.
    ** These should be excluded from the breakpoint scope and
    ** from event analysis.
    **
    ** We also attempt to support the new "block:proc" convention by
    ** creating a wildcard expansion buffer in expanded_sysprocnames.
    ** We won't bother about syntax errors since those will just give
    ** us the NULL pointer instead.
    */

   rw.sysproc_count = zzget_sysprocnames(&rw.sysprocnames);

   rw.expanded_sysprocnames = NULL;
   if (rw.sysproc_count)
   {
      rw.expanded_sysprocnames = wildcard_to_plist(rw.sysproc_count, rw.sysprocnames);
   }

   /* Init the offspring list, needed before create_precious. */
   rw.offsprings = list_init((void *) alloc(sizeof (List), 0));

   /*
    ** We should not create the trace until all variables have been initialized.
    ** Otherwise we might see a warm start if restarted before init is complete.
    */

   if (!warm_start)
   {
      do_delete_all_actions(False, True); /* Remove all actions. If some went OK in reclaim_all_actions! */
      ose_trace_create(rw.trace, rw.trace_size, SYS_TRACE_NAME,
                       (GetTimeFuncPointer) get_systime, 1000000);
      /* Initialize the ldm variable set. */
      create_precious();
   }
   else
   {
      do_delete_all_actions(True, True);      /* Remove all non trace actions. */

      /* Read sysparam to setup ranges for the profiling parameters */
      /* (Create precious does the same inside configure_parameter) */
      setup_parameter_ranges();
   }

   /*
    ** On warm start, take the buffers in the trace that contains
    ** scope lists.
    ** These were unintentionally freed by recover_trace().
    **
    ** If there are no buffers then allocate some. This can happen if cold start
    ** or if a restart occurred during a previous system start.
    ** Set default scopes to "*".
    */
   if (rw.debugger_scope)
      ose_trace_reclaim(rw.debugger_scope);
   else
   {
      rw.precious->debugger_scope = rw.debugger_scope = copy_scope(1, "*");
      update_precious_checksum();
      rw.quick_debugger_scope = wildcard_to_plist(1, "*");
   }

   if (rw.breakpoint_scope)
      ose_trace_reclaim(rw.breakpoint_scope);
   else
   {
      rw.precious->breakpoint_scope = rw.breakpoint_scope = copy_scope(1, "*");
      update_precious_checksum();
      rw.quick_breakpoint_scope = wildcard_to_plist(1, "*");
   }

   update_hook_bits();

   zzprof_init();
   SMP_CODE(rw.skip_user_prof_lock = ~1;)
   SMP_CODE(spin_lock_init(&rw.user_prof_lock));
   SMP_CODE(spin_lock_init(&rw.user_event_lock));
   SMP_CODE(spin_lock_init(&rw.event_notify_lock));

   rw.user_events = list_init((void *) alloc(sizeof (List), 0));
   
   /* Install alloc hook if needed */
   if (NULL != rw.actions SMP_CODE([0]).action_list_alloc)
   {
      zzkrn_set_handler(12,(void *) zzalloc_dbghook);
   }
   /* Install free_buf hook if needed */
   if (NULL != rw.actions SMP_CODE([0]).action_list_free_buf)
   {
      zzkrn_set_handler(13,(void *) zzfree_buf_dbghook);
   }

   rw.start_mode = warm_start? MONITOR_START_MODE_WARM: MONITOR_START_MODE_COLD;
   
   ose_trace_enable(rw.trace);
   {
      union TRACE    *te;
      te = ose_trace_alloc(rw.trace, sizeof (struct trace_system_start), 0);
      /* Fill in the trace buffer. */
      if (te)
      {
         te->trace_system_start.event_no = TRACE_SYSTEM_START;
         te->trace_system_start.warm_flag = warm_start;
         te->trace_system_start.reserved = 0;
         te->trace_system_start.abstime = NULL;
         ose_trace_commit_inline(te);
      }
   }
   ose_trace_disable(rw.trace);

   /* On warm start, enable the trace if it was enabled before the restart. */
   if (warm_start)
   {
      if (rw.trace_enabled)
      {
         ose_trace_enable(rw.trace);
      }
   }

   /* Send back the mode sig to the system daemon. */
   send(&sig, sender(&sig));

#ifdef RMM_USE_PERSISTENT_REGIONS
   {
      OSPRIORITY oldpri = get_pri(current_process());
      uint32_t        s;

      set_pri(31);              /* FIXME A temp yield */
      delay(3000);              /* FIXME! Wait synchronize with MM */
      void           *p = get_persistent_memory("system-trace", 8192, &s);

      if (p != NULL)
      {
         assert(s >= trace_size);
         trace = p;
         trace_size = s;
      }
      else
      {
         assert(s == 0);
      }
      set_pri(oldpri);
   }
#endif

   ramlog_printf("rmm: Up and running\n");

   for (;;)
   {
      const SIGSELECT *select = (rw.evact_lock_owner && rw.evact_lock_owner->flow_ctrl_nblock <= 0)?
         no_nblock_sigs_filter: all_sigs_filter;

      for (sig = NULL;
           NULL == sig && rw.pending_req > 0;
           sig = receive_w_tmo(10, select)) /* block max 10 ticks */
      {
         check_pending_mgets();
      }

      if (NULL == sig)
      {
         sig = receive(select);
      }

      if (sig->sig_no < MONITOR_INTERFACE_REQUEST)
      {
         /* Internal signals */
         if (BUFFER_FULL_NOTIFY == sig->sig_no)
         {
#ifndef softOSE
            handle_buffer_full_notify(&sig->buffer_full_notify);
#endif
         }
         else if (MONITOR_REPORT == sig->sig_no)
         {
            handle_event_report(&sig->event_report_verbose);
         }
         else if (PROF_NEW_SAMPLE == sig->sig_no)
         {
        	 if (sig->prof_new_sample.type == REPORT_PROF )
        	 {
        		 handle_new_report_sample();
        	 }
        	 else {
        		 zzprof_handle_new_sample(sig->prof_new_sample.type );
        	 }
         }
#ifdef OSE_SMP
         else if (PROF_CORE_FREE_BUF == sig->sig_no)
         {
        	 zzprof_handle_prof_core_free_buf(sig->prof_core_free_buf.cpu );
         }
#endif
         else if (CATCH_REPORT == sig->sig_no)
         {
            handle_event_report(&sig->event_report_verbose);
         }
#ifdef USE_EPM
         else if (RMM_CALLBACK_HIT == sig->sig_no)
         {
            handle_callback_hit(sig);
         }
         else if (RMM_SCOPE_ATTACH_SIG == sig->sig_no)
         {
            Offspring *c = find_offspring_by_pid(sig->rmm_scope_attach_sig.offspring_pid);

            if (c != NULL)
            {
               c->scope_attach_ref = 0;
               detach_offspring(c);
            }
         }
#endif
         else if (MONITOR_CREATE_EVENT_STRUCT_REQUEST == sig->sig_no)
         {
            PROCESS sig_sender = sender(&sig);
            add_new_user_event_struct(&sig->create_event_struct_request, sig_sender);
         }
         else if (USER_EVENT_ATTACH_SIG == sig->sig_no)
         {
            disable_evacts_in_segment(&sig->user_event_attach_sig);
         }
         else if (MONITOR_CREATE_USER_REPORTS_STRUCT_REQUEST == sig->sig_no)
         {
            union SIGNAL *sig_r;

            handle_create_user_reports(sig->create_user_reports_struct_request.reportNo,
                                       (OSADDRESS *)sig->create_user_reports_struct_request.trig,
                                       sig->create_user_reports_struct_request.multiple,
                                       sig->create_user_reports_struct_request.continuous,
                                       sig->create_user_reports_struct_request.maxmin,
                                       (PROCESS)sig->create_user_reports_struct_request.sid);


            sig_r = alloc(0, MONITOR_CREATE_USER_REPORTS_STRUCT_REPLY);
            send(&sig_r, sender(&sig));
         }
         else if(RMM_GET_OFFSPRINGS_REQUEST == sig->sig_no)
         {
           handle_get_offsprings(sig);
         }
         else if (USER_REPORTS_ATTACH_SIG == sig->sig_no)
         {
            remove_user_report_trig(sig->user_report_attach.sid);
         }
         else if (OS_ATTACH_SIG == sig->sig_no)
         {
            Offspring *offspring = find_offspring_by_pid(sender(&sig));
#ifndef softOSE
            /* Stop all active profiling counters */
            handle_disconnect_request_counter(offspring, sig);
#endif
            assert_offspring(offspring);
            offspring->attach_ref = 0;
            disconnect_offspring(offspring, "client killed");
         }
      }
      else
      {
         /* External signal interface */

         if (MONITOR_INTERFACE_REQUEST == sig->sig_no)
         {
            handle_interface_request(sig);
         }
         else if (MONITOR_CONNECT_REQUEST == sig->sig_no &&
                  !is_req_disabled(MONITOR_CONNECT_REQUEST))
         {
            handle_connect_request(sig);
         }
         else if (MONITOR_NAME_REQUEST == sig->sig_no &&
                  !is_req_disabled(MONITOR_NAME_REQUEST))
         {
            handle_name_request(sig);
         }
         else
         {
            Offspring *offspring = find_offspring_by_pid(sender(&sig));

            if (NULL == offspring)
            {
               /* Got signal from unknown process so ignore it. */
               ramlog_printf("rmm: Signal from unknown process (pid: 0x%X sig_no: %d)\n",
                             sender(&sig), sig->sig_no);
            }
            else
            {
               uint32_t index = sig->sig_no - MONITOR_INTERFACE_REQUEST;
               signal_request_t sig_handler;
               uint32_t counter_sig_index = sig->sig_no -
                                      MONITOR_GET_SUPPORTED_COUNTER_TYPES_REPLY;

               assert_offspring(offspring);

               if (index >= SIGTABLE_LENGTH && counter_sig_index >= COUNTER_PROF_SIG_LENGTH)
               {
                  ramlog_printf("rmm: *** Unknown signal %u\n", sig->sig_no);
               }
               else if (is_req_disabled(sig->sig_no))
               {
                  ramlog_printf("rmm: *** Disabled signal %u\n", sig->sig_no);
               }
               else if (((index > SIGTABLE_LENGTH || (sig_handler = sig_handler_table[index]) == NULL)) &&
                        (counter_sig_index > COUNTER_PROF_SIG_LENGTH ||
                          (sig_handler = counter_prof_sig_handler[counter_sig_index]) == NULL))
               {
                  ramlog_printf("rmm: *** Illegal request signal %u\n", sig->sig_no);
               }
               else
               {
                  /* Call handler function for this signal */
                  if (sig_handler(offspring, sig) != 0)
                  {
                     /* If zero is not returned the handler freed the signal */
                     continue;
                  }
               }
            }
         }
      }

      /* Finally free the signal we received */
      free_buf(&sig);

   } /* for(;;) */
}

uint32_t
ose_rmm_get_unique_number(void)
{
   return atomic_inc_return(&rw.precious->unique_number);
}

void
ose_rmm_init_unique_number(uint32_t start_number)
{
   uint32_t number;
   do
   {
      number = atomic_get(&rw.precious->unique_number);
   } while(number != atomic_cas(&rw.precious->unique_number, number, start_number));
}

/*
** Debugger hook for system error events.
** This hook is called for both kernel and user errors, when not ignored
** by any local process or block error handler.
** This hook is called from interrupt, prioritized and background processes.
** The purpose is primarily to catch system errors in the debugger.
*/

void
zzsystem_error_dbghook(struct debug_pcb *proc, int as_intproc,
		       OSBOOLEAN user_called, OSERRCODE ecode, OSERRCODE extra)
{
   if (!as_intproc)
   {
      LOCK_SWAP();
   }

   handle_error_event(proc, user_called, ecode, extra, as_intproc, 2);

   if (!as_intproc)
   {
      UNLOCK_SWAP();
   }
}
